/*----------------------------------------------------------------------------
NAME:    api_char_lamp_detector_response

PURPOSE: Process the post warmup lamp data to get the radiance statistics 
         for each detector in one SCA. Store the detector level statistics 
         results into database and optionally output them into the report 
         file by calling the char_lamp_detector_response algorithm code.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "ias_logging.h"
#include "ias_threadpool.h"
#include "ias_miscellaneous.h"
#include "ias_const.h"
#include "ias_cpf.h"
#include "rps_local.h"
#include "ias_l0r.h"
#include "ias_math.h"
#include "ias_satellite_attributes.h"

/* Note: The compute_mean and compute_stdev routines below are used because
         the corresponding ias_math_compute_mean and ias_math_compute_stdev
         library routines only handle double data type, precision will be 
         decreased if IAS Math library routines are used. Later with the 
         modification of the library routines, the two internal routines 
         below could be removed */  
int compute_mean
(
    const float *sample,       /* I: pointer to the data */
    int nsamps,                /* I: number of samples */
    float *mean                /* O: mean of the data */
);

int compute_stdev
(
    const float *sample,   /* I: pointer to the data */
    int nsamps,            /* I: number of samples */
    float mean,            /* I: mean of the samples */
    float *std_dev         /* O: standard deviation of the data */
);

void clean_up_lamp_telemetry
(
    float *lamp_current_a,
    float *lamp_voltage_a,
    float *photodiode_micro_amp_a,
    float *lamp_housing_temp_a,
    float *monitor_diode_temp_a,
    float *lamp_current_b,
    float *lamp_voltage_b,
    float *photodiode_micro_amp_b,
    float *lamp_housing_temp_b,
    float *monitor_diode_temp_b
);

int api_char_lamp_detector_response
(
    BLOB *blob,         /* I: Miscellaneous parameters */
    int   band_index,   /* I: */
    int   sca_index,    /* I: */
    int   algo_index    /* I: Algorithm index */
)
{
    IAS_RPS_HISTOGRAM_INTERMEDIATE_RESULTS *statistics_results_l0r = NULL;
    IAS_RPS_HISTOGRAM_INTERMEDIATE_RESULTS *statistics_results_l1r= NULL;
    RPS_BAND_ATTRIBUTES *band_attributes = get_band_attributes(band_index);
    int band_number = band_attributes->band_number;
    int sca_number = sca_index + 1;

    int num_lines = band_attributes->number_lines;
    int num_samps = band_attributes->number_detectors;

    DB_CHAR_LAMP_RESPONSE_DETECTOR_RESULTS *db_l0r_detector_results = NULL;
    DB_CHAR_LAMP_RESPONSE_DETECTOR_RESULTS *db_l1r_detector_results = NULL;
    DB_CHAR_LAMP_RESPONSE_ONE_SCA_RESULTS *db_l0r_one_sca_results = NULL;
    DB_CHAR_LAMP_RESPONSE_ONE_SCA_RESULTS *db_l1r_one_sca_results = NULL;

    IAS_REPORT *report_information = NULL;
    int status;

    /* The blob->lamp_processing be set in preprocess, if it's not set,
       the lamp data processing will be stopped */
    if (!blob->lamp_processing)
    {
        IAS_LOG_ERROR("The lamp processing flag in the blob is not set");
        return ERROR;
    }

    /* If the processing band is a TIRS thermal band, then return 
       to make sure we are processing only OLI band */
    if (band_attributes->spectral_type == IAS_SPECTRAL_THERMAL)
    {
        return SUCCESS;
    }

    /* Note: It isn't necessary to decrement the usage of the memory buffer
       on error conditions from this point on.  Since the memory may not be
       freed.  The main thread will clean up allocated buffers when error
       conditions dictate.  It doesn't hurt but it isn't necessary. */

    /* Set the MM_L0R_SCA_STATISTICS buffer */
    db_l0r_one_sca_results = (DB_CHAR_LAMP_RESPONSE_ONE_SCA_RESULTS *)
        get_memory_buffer(MM_L0R_LAMP_ONE_SCA_STATISTICS, SETTER,
        band_index, sca_index);
    if (!db_l0r_one_sca_results)
    {
        IAS_LOG_ERROR("Allocating memory for MM_L0R_ONE_SCA_STATISTICS "
                      "structure");
        return ERROR;
    }

    /* Allocate memory for the DB_CHAR_LAMP_RESPONSE_DETECTOR_RESULTS
       structure */
    db_l0r_detector_results = rps_db_allocate_lamp_detector_results( 
                              num_samps);
    if (!db_l0r_detector_results)
    {
        IAS_LOG_ERROR("Allocating memory for detector response structure");
        return ERROR;
    }

    /* Set the reporting information */
    if (blob->generate_report)
        report_information = &blob->report_information;

    /* Get the L0R intermediate statistics results buffer */
    statistics_results_l0r = (IAS_RPS_HISTOGRAM_INTERMEDIATE_RESULTS *)
        get_memory_buffer(MM_L0R_INTERMEDIATE_STATISTICS, CONSUMER,
        band_index, sca_index);
    if (!statistics_results_l0r)
    {
        IAS_LOG_ERROR("Getting l0r intermediate statistics memory buffer");
        rps_db_free_lamp_detector_results(db_l0r_detector_results);
        return ERROR;
    }
    else
    {
        /* Set the common values for db_detector_results that are known here */
        db_l0r_detector_results->dbase = blob->database_information.db;
        strncpy(db_l0r_detector_results->rad_common_id, 
            blob->database_information.rad_common_id, 
            sizeof(db_l0r_detector_results->rad_common_id));
        db_l0r_detector_results->rad_common_id[sizeof(
            db_l0r_detector_results->rad_common_id) - 1] = '\0';
        db_l0r_detector_results->band = band_number;
        db_l0r_detector_results->sca = sca_number;
        db_l0r_detector_results->num_detectors = num_samps;
        db_l0r_detector_results->processing_level 
                 = strdup(statistics_results_l0r[0].processing_level);
        if (!db_l0r_detector_results->processing_level)
        {
            IAS_LOG_ERROR("Duplicating processing level string for "
                          "L0R db_detector_results");
            rps_db_free_lamp_detector_results(db_l0r_detector_results);
            return ERROR;
        }

        /* Set the common values for db_l0r_one_sca_results that are known 
           here */
        strncpy(db_l0r_one_sca_results->rad_common_id, 
            blob->database_information.rad_common_id, 
            sizeof(db_l0r_one_sca_results->rad_common_id));
        db_l0r_one_sca_results->rad_common_id[sizeof(
            db_l0r_one_sca_results->rad_common_id) - 1] = '\0';
        db_l0r_one_sca_results->band = band_number;
        db_l0r_one_sca_results->sca = sca_number;
        db_l0r_one_sca_results->processing_level = 
                strdup(db_l0r_detector_results->processing_level);
        if (!db_l0r_one_sca_results->processing_level)
        {
            IAS_LOG_ERROR("Duplicating processing level string for "
                          "L0R db_l0r_one_sca_results");
            rps_db_free_lamp_detector_results(db_l0r_detector_results);
            return ERROR;
        }

        status = char_lamp_detector_response(band_number, sca_number, 
            num_samps, num_lines, blob->stim_lamp_pair_id, blob->lamp_on_time, 
            blob->post_warmup_start_time, blob->post_warmup_stop_time, 
            blob->post_warmup_start_frame, blob->post_warmup_stop_frame,
            blob->pixel_mask[band_index][sca_index],
            statistics_results_l0r, report_information,
            db_l0r_detector_results, db_l0r_one_sca_results);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("From call to detector response char (lamp) "
                "algorithm for band_number: %d, SCA_number %d", band_number, 
                sca_number );
            rps_db_free_lamp_detector_results(db_l0r_detector_results);
            return ERROR;
        }

        /* Tell the memory management that this buffer is now valid */
        set_buffer_filled_flag( MM_L0R_LAMP_ONE_SCA_STATISTICS, band_index,
            sca_index );

        if (blob->store_in_database)
        {
            /* Insert the L0R part LAMP_RESPONSE_DETECTOR table into database */
            status = ias_threadpool_add_to_work_queue(blob->threadpool,
                rps_store_lamp_detector_in_database, db_l0r_detector_results);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Submitting lamp detector DB insertion to work "
                    "queue band_number: %d, SCA_number: %d", 
                    band_number, sca_number);
                /* Before freeing memory, ensure there is more than one thread
                   running.  A double free of the memory may occur if only one
                   thread is running and an internal database error occurs. */
                if ( ias_threadpool_get_thread_count( blob->threadpool ) > 1 )
                    rps_db_free_lamp_detector_results(db_l0r_detector_results);
                return ERROR;
            }
            db_l0r_detector_results = NULL;
        }
    }

    /* Free the allocated memory for L0R data */
    if (db_l0r_detector_results != NULL)
        rps_db_free_lamp_detector_results(db_l0r_detector_results);

    /* Tell memory management that the L0R related data was used */
    status = decrement_consumer_count(MM_L0R_INTERMEDIATE_STATISTICS,
        band_index, sca_index);
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Decrementing memory buffer");
        return ERROR;
    }
    status = decrement_consumer_count(MM_L0R_LAMP_ONE_SCA_STATISTICS,
        band_index, sca_index);
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Decrementing memory buffer");
        return ERROR;
    }

    /* Get the MM_L1R_SCA_STATISTICS buffer */
    db_l1r_one_sca_results = (DB_CHAR_LAMP_RESPONSE_ONE_SCA_RESULTS *)
        get_memory_buffer(MM_L1R_LAMP_ONE_SCA_STATISTICS, SETTER,
        band_index, sca_index);
    if (!db_l1r_one_sca_results)
    {
        IAS_LOG_ERROR("Getting memory buffer for "
            "MM_L1R_LAMP_ONE_SCA_STATISTICS");
        return ERROR;
    }

    /* Allocate memory for the DB_CHAR_LAMP_RESPONSE_DETECTOR_RESULTS
       structure */
    db_l1r_detector_results = rps_db_allocate_lamp_detector_results( 
                              num_samps);
    if (!db_l1r_detector_results)
    {
        IAS_LOG_ERROR("Allocating memory for detector response structure");
        return ERROR;
    }

    /* Get the L1R intermediate statistics results buffer */
    statistics_results_l1r = (IAS_RPS_HISTOGRAM_INTERMEDIATE_RESULTS *)
        get_memory_buffer(MM_L1R_INTERMEDIATE_STATISTICS, CONSUMER,
        band_index, sca_index);
    if (!statistics_results_l1r)
    {
        IAS_LOG_ERROR("Getting l1r intermediate statistics memory buffer");
        rps_db_free_lamp_detector_results(db_l1r_detector_results);
        return ERROR;
    }
    else
    {
        /* Set the common values for db_detector_results that are known here */
        db_l1r_detector_results->dbase = blob->database_information.db;
        strncpy(db_l1r_detector_results->rad_common_id, 
            blob->database_information.rad_common_id, 
            sizeof(db_l1r_detector_results->rad_common_id));
        db_l1r_detector_results->rad_common_id[sizeof(
            db_l1r_detector_results->rad_common_id) - 1] = '\0';
        db_l1r_detector_results->band = band_number;
        db_l1r_detector_results->sca = sca_number;
        db_l1r_detector_results->num_detectors = num_samps;
        db_l1r_detector_results->processing_level 
                 = strdup(statistics_results_l1r[0].processing_level);
        if (!db_l1r_detector_results->processing_level)
        {
            IAS_LOG_ERROR("Duplicating processing level string for "
                          "L1R db_detector_results");
            rps_db_free_lamp_detector_results(db_l1r_detector_results);
            return ERROR;
        }

        /* Set the common values for db_l1r_one_sca_results that are known 
           here */
        strncpy(db_l1r_one_sca_results->rad_common_id, 
            blob->database_information.rad_common_id, 
            sizeof(db_l1r_one_sca_results->rad_common_id));
        db_l1r_one_sca_results->rad_common_id[sizeof(
            db_l1r_one_sca_results->rad_common_id) - 1] = '\0';
        db_l1r_one_sca_results->band = band_number;
        db_l1r_one_sca_results->sca = sca_number;
        db_l1r_one_sca_results->processing_level = 
               strdup(db_l1r_detector_results->processing_level);
        if (!db_l1r_one_sca_results->processing_level)
        {
            IAS_LOG_ERROR("Duplicating processing level string for "
                          "L0R db_l1r_one_sca_results");
            rps_db_free_lamp_detector_results(db_l1r_detector_results);
            return ERROR;
        }

        /* Call the char_lamp_detector_response algorithm */
        status = char_lamp_detector_response( 
            band_number, sca_number,  
            num_samps, num_lines, blob->stim_lamp_pair_id, blob->lamp_on_time, 
            blob->post_warmup_start_time, blob->post_warmup_stop_time, 
            blob->post_warmup_start_frame, blob->post_warmup_stop_frame,
            blob->pixel_mask[band_index][sca_index], 
            statistics_results_l1r, report_information,
            db_l1r_detector_results, db_l1r_one_sca_results);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("From call to detector response char (lamp) "
                          "algorithm for band_number: %d, SCA_number %d", 
                          band_number, sca_number);
            rps_db_free_lamp_detector_results(db_l1r_detector_results);
            return ERROR;
        }

        /* Tell the memory management that this buffer is now valid */
        set_buffer_filled_flag( MM_L1R_LAMP_ONE_SCA_STATISTICS, band_index,
            sca_index );

        if (blob->store_in_database)
        {
            /* Insert the L1R part LAMP_RESPONSE_DETECTOR table into database */
            status = ias_threadpool_add_to_work_queue(blob->threadpool,
                rps_store_lamp_detector_in_database, db_l1r_detector_results);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Submitting lamp detector DB insertion to work "
                    "queue band_number: %d, SCA_number: %d", 
                    band_number, sca_number);
                /* Before freeing memory, ensure there is more than one thread
                   running.  A double free of the memory may occur if only one
                   thread is running and an internal database error occurs. */
                if ( ias_threadpool_get_thread_count( blob->threadpool ) > 1 )
                    rps_db_free_lamp_detector_results(db_l1r_detector_results);
                return ERROR;
             }
             db_l1r_detector_results = NULL;
        }
    }

    /* Tell memory management that the data was used */
    status = decrement_consumer_count(MM_L1R_INTERMEDIATE_STATISTICS,
        band_index, sca_index);
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Decrementing memory buffer");
        return ERROR;
    }
    if ( band_attributes->vrp_band_number > 0 )
    {
        status = decrement_consumer_count(MM_L0R_INTERMEDIATE_STATISTICS,
            band_attributes->vrp_band_index, sca_index);
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }
        status = decrement_consumer_count(MM_L1R_INTERMEDIATE_STATISTICS,
            band_attributes->vrp_band_index, sca_index);
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }
    }
    if ( band_attributes->blind_band_number > 0 
        && band_attributes->process_blind_band_flag )
    {
        status = decrement_consumer_count(MM_L0R_INTERMEDIATE_STATISTICS,
            band_attributes->blind_band_index, sca_index);
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }
        status = decrement_consumer_count(MM_L1R_INTERMEDIATE_STATISTICS,
            band_attributes->blind_band_index, sca_index);
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }
    }

    /* Free the allocated memory for L1R data */
    if (db_l1r_detector_results != NULL)
        rps_db_free_lamp_detector_results(db_l1r_detector_results);

    status = decrement_consumer_count(MM_L1R_LAMP_ONE_SCA_STATISTICS,
        band_index, sca_index);
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Decrementing memory buffer");
        return ERROR;
    }

    return status;
}  /* END api_char_lamp_detector_response */


/*----------------------------------------------------------------------------
NAME:  char_lamp_detector_response_preprocess  

PURPOSE: Increase all the memory consumer counts and calculate the statistics
         for the lamp focal plane temperature data

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int char_lamp_detector_response_preprocess
(
    BLOB *blob          /* I: Miscellaneous parameters */
)
{
    int status;                    /* Return status */
    /* Define a list of memory buffers needed by the algorithm */
    MEMORY_BUFFER_TYPE memory_buffers[] =
    {
        MM_L0R_INTERMEDIATE_STATISTICS,
        MM_L1R_INTERMEDIATE_STATISTICS,
        MM_L0R_LAMP_ONE_SCA_STATISTICS,
        MM_L1R_LAMP_ONE_SCA_STATISTICS,
        MM_L0R_LAMP_ONE_SCA_STATISTICS,
        MM_L1R_LAMP_ONE_SCA_STATISTICS,
    };
    int number_of_memory_buffers = sizeof(memory_buffers) /
        sizeof(MEMORY_BUFFER_TYPE);
    int i;

    /* Increment the consumer counts for the memory buffers */
    for ( i = 0; i < number_of_memory_buffers; i++ )
    {
        status = increment_consumer_count( memory_buffers[i] );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Incrementing %s consumer count",
                get_memory_buffer_string( memory_buffers[i] ));
            return ERROR;
        }
    }

    /* Set the blob->lamp_processing flag */
    blob->lamp_processing = 1;

    /* Delete any pre-existing records for lamp_response_detector table 
       if database storage was requested*/
    if ( blob->store_in_database )
    {
        status = delete_db_table_records( blob->database_information,
            "lamp_response_detector");
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Returned from delete_db_table_data "
                          " for lamp_response_detector table");
            return ERROR;
        }

        /* Delete any pre-existing records for lamp_response_sca table */
        status = delete_db_table_records( blob->database_information,
            "lamp_response_sca");
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Returned from delete_db_table_data "
                          " for lamp_response_sca table");
            return ERROR;
        }

        /* Delete any pre-existing records for lamp_response_band table */
        status = delete_db_table_records( blob->database_information,
            "lamp_response_band");
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Returned from delete_db_table_data "
                          " for lamp_response_band table");
            return ERROR;
        }

        /* Delete any pre-existing records for lamp_response table */
        status = delete_db_table_records( blob->database_information,
            "lamp_response");
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Returned from delete_db_table_data "
                          " for lamp_response table");
            return ERROR;
        }

        /* Delete any pre-existing records for lamp_telemetry table */
        status = delete_db_table_records( blob->database_information,
            "lamp_telemetry");
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Returned from delete_db_table_data "
                          " for lamp_telemetry table");
            return ERROR;
        }
    }

    /* Set the stim lamp pair id in the blob */
    status = read_stim_lamp_pair_id(blob);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading stim lamp pair id and setting them in the "
                      "Blob");
        return ERROR;
    }

    /* Set the lamp response data in the blob */
    status = read_lamp_response(blob);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading lamp response data and setting them in the "
                      "Blob");
        return ERROR;
    }

    /* The call to read_lamp_response found the starting and ending line for
       the post-warmup.  Use these values to override the starting line and
       number of lines in any band to be processed. */
    for ( i = 0; i < blob->number_of_bands; i++ )
    {
        int band_index = blob->list_of_band_indices[i];
        RPS_BAND_ATTRIBUTES *band_atts = get_band_attributes( band_index );
        int starting_line;
        int number_lines;
        int lines_per_frame = ias_sat_attr_get_lines_per_frame(
            band_atts->band_number );
        if ( lines_per_frame == ERROR )
        {
            IAS_LOG_ERROR("Getting lines per frame for band %d",
                band_atts->band_number );
            return ERROR;
        }
        starting_line = blob->post_warmup_start_frame * lines_per_frame;
        number_lines = blob->post_warmup_nframes * lines_per_frame;

        /* Set the normal band starting and number lines */
        band_atts->starting_line = starting_line;
        band_atts->number_lines = number_lines;

        /* Set the VPR */
        if ( band_atts->vrp_band_number > 0 )
        {
            RPS_BAND_ATTRIBUTES *vrp_band_atts = get_band_attributes(
                band_atts->vrp_band_index );
            vrp_band_atts->starting_line = starting_line;
            vrp_band_atts->number_lines = number_lines;
        }

        /* Set the blind band if applicable */
        if ( band_atts->blind_band_number > 0
            && band_atts->process_blind_band_flag )
        {
            RPS_BAND_ATTRIBUTES *blind_band_atts = get_band_attributes(
                band_atts->blind_band_index );
            blind_band_atts->starting_line = starting_line;
            blind_band_atts->number_lines = number_lines;
        }
    }

    /* Sanity checks to verify that enough data and warmup time is present. */
    if (blob->post_warmup_start_time - blob->lamp_on_time
        < blob->lamp_warmup_time)
    {
        /* Spit out some relevant information for what was requested and the
           failure values in an error message. */
        IAS_LOG_ERROR("Insufficient lamp warmup time; Found "
                      "(post_warmup_start_time - lamp_on_time) = %f;\n"
                      "\t\tlamp_on_time = %f; "
                      "lamp_warmup_time(requested) = %f; "
                      "post_warmup_start_time = %f; "
                      "post_warmup_nframes(requested) = %d",
                      blob->post_warmup_start_time - blob->lamp_on_time,
                      blob->lamp_on_time, blob->lamp_warmup_time,
                      blob->post_warmup_start_time,
                      blob->post_warmup_nframes);
        return ERROR;
    }

    return SUCCESS;
}  /* END char_lamp_detector_response_preprocess */

/*----------------------------------------------------------------------------
NAME:    char_lamp_detector_response_postprocess

PURPOSE: Process the results from the api code to get the radiance statistics 
         for each SCA and each band. Store the statistics results together 
         with lamp telemetry and lamp response data into database and 
         optionally output them into the report file. 

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int char_lamp_detector_response_postprocess
(
    BLOB *blob          /* I: Miscellaneous parameters */
)
{
    DB_CHAR_LAMP_RESPONSE_ONE_SCA_RESULTS *db_one_sca_results = NULL;
    DB_CHAR_LAMP_RESPONSE_SCA_RESULTS *db_sca_results = NULL;
    DB_CHAR_LAMP_RESPONSE_BAND_RESULTS *db_band_results = NULL;
    DB_CHAR_LAMP_RESPONSE_RESULTS *db_response_results = NULL;
    DB_CHAR_LAMP_TELEMETRY_RESULTS *db_telemetry_results = NULL;
    IAS_REPORT *report_information = NULL;
    int band_number;                    /* Band number */  
    int band_index;                     /* Band index */  
    int sca_index = 0;                  /* SCA index */
    int index;                          /* Index used for accessing the 
                                           database structure */
    int sca_counter;                    /* SCA counter */
    int status;                         /* Return status */
    float *focal_plane_temp1;           /* Focal plane temperature 1 list */
    float *focal_plane_temp2;           /* Focal plane temperature 2 list */
    int num_focal_temp1;                /* Number of scene focal plane module 7
                                           temperature */
    int num_focal_temp2;                /* Number of scene focal plane module
                                           14 interface temperature */
    float mean_fp_temp1;                /* Averaged focal plane temperature 1 */
    float mean_fp_temp2;                /* Averaged focal plane temperature 2 */
    float stdev_fp_temp1;               /* Stdev of focal plane temperature 1 */
    float stdev_fp_temp2;               /* Stdev of focal plane temperature 2 */
    float *lamp_current_a = NULL;       /* Group A lamp current data */
    float *lamp_voltage_a = NULL;       /* Group A lamp voltage data */
    float *photodiode_micro_amp_a = NULL;/* Group A photodiode current data */ 
    float *lamp_housing_temp_a = NULL;  /* Group A lamp housing temperature
                                           data */
    float *monitor_diode_temp_a = NULL; /* Group A monitor diode 
                                           temperature data */
    float *lamp_current_b = NULL;       /* Group B lamp current data */
    float *lamp_voltage_b = NULL;       /* Group B lamp voltage data */
    float *photodiode_micro_amp_b = NULL; /* Group B photodiode current data */
    float *lamp_housing_temp_b = NULL;  /* Group B lamp housing temperature
                                           data */
    float *monitor_diode_temp_b = NULL; /* Group B monitor diode 
                                           temperature data */
    int num_telemetry_data;             /* Number of lamp telemetry data */

    /* Allocate memory for the DB_CHAR_LAMP_SCA_DETECTOR_RESULTS
       structure */
    db_sca_results = rps_db_allocate_lamp_sca_results( 
                     blob-> max_number_scas);
    if (!db_sca_results)
    {
        IAS_LOG_ERROR("Allocating the SCA response structure");
        return ERROR;
    }

    /* Allocate memory for DB_CHAR_LAMP_RESPONSE_DETECTOR_RESULTS
       structure */
    db_band_results = rps_db_allocate_lamp_band_results( 
                      blob->number_of_bands);
    if (!db_band_results)
    {
        IAS_LOG_ERROR("Allocating the band response structure");
        rps_db_free_lamp_sca_results(db_sca_results);
        return ERROR;
    }

    /* Set the common values for db_band_results that are known here */
    db_band_results->dbase = blob->database_information.db;
    strncpy(db_band_results->rad_common_id, 
        blob->database_information.rad_common_id, 
        sizeof(db_band_results->rad_common_id));
    db_band_results->rad_common_id[sizeof(db_band_results->rad_common_id) - 1] 
        = '\0';
    db_band_results->num_bands = blob->number_of_bands;

    /* Set the common values for db_sca_results that are known here */
    db_sca_results->dbase = blob->database_information.db;
    strncpy(db_sca_results->rad_common_id, 
        blob->database_information.rad_common_id, 
        sizeof(db_sca_results->rad_common_id));
    db_sca_results->rad_common_id[sizeof(db_sca_results->rad_common_id) - 1] 
        = '\0';
    db_sca_results->num_scas = blob->max_number_scas;

    /* Get the report information */
    if (blob->generate_report)
        report_information  = &blob->report_information;

    /* Make sure the MM_L0R_INTERMEDIATE_STATISTICS buffer is valid, otherwise
       skip the call to the algorithm to make the workflow working if cal/val
       people adjust the location of the histogram statistics algorithm in the
       workflow */
    if ( get_buffer_filled_flag( MM_L0R_INTERMEDIATE_STATISTICS,
         blob->list_of_band_indices[0], 0 ) == 0 )
    {
        IAS_LOG_WARNING("Attempting to use unfilled buffer %s", 
            get_memory_buffer_string(MM_L0R_INTERMEDIATE_STATISTICS));
    }
    else
    {
        /* Loop through each band of the input data */
        for (index= 0; index < blob->number_of_bands; index++)
        {
            band_index = blob->list_of_band_indices[index];
            band_number = blob->list_of_bands[index];
            /* Initialize the db_band_results mean and stdev values */
            db_band_results->band[index] = band_number;
            db_band_results->db_mean[index] = 0;
            db_band_results->db_stdev[index] = 0;
            db_band_results->db_corrected[index] = 0;
            sca_counter = 0;

            db_sca_results->band = band_number;

            /* Loop through the SCA for L0R lamp data */
            for (sca_index = 0; sca_index < blob->max_number_scas; sca_index++)
            {
                /* Get the L0R SCA statistics results buffer */
                db_one_sca_results = 
                    (DB_CHAR_LAMP_RESPONSE_ONE_SCA_RESULTS *)
                    get_memory_buffer(MM_L0R_LAMP_ONE_SCA_STATISTICS,
                    CONSUMER, band_index, sca_index);
                if (!db_one_sca_results)
                {
                    IAS_LOG_ERROR("Getting MM_L0R_SCA_STATISTICS "
                                  "memory buffer");
                    rps_db_free_lamp_sca_results(db_sca_results);
                    rps_db_free_lamp_band_results(db_band_results);
                    return ERROR;
                }

                /* Populate processing_level for db_band_results and 
                   db_sca_results structures */
                if (index == 0 && sca_index == 0)
                {
                    db_band_results->processing_level 
                        = strdup(db_one_sca_results->processing_level);
                    db_sca_results->processing_level 
                        = strdup(db_one_sca_results->processing_level);
                    if (!db_band_results->processing_level ||
                        !db_sca_results->processing_level)
                    {
                        IAS_LOG_ERROR("Duplicating processing_level string "
                                      "for db_band_results and db_sca_results");
                        rps_db_free_lamp_sca_results(db_sca_results);
                        rps_db_free_lamp_band_results(db_band_results);
                        return ERROR;
                    }
                }

                /* Calculate the statistics for the db_band_results structure */
                db_band_results->db_mean[index] += 
                                   db_one_sca_results->db_mean;
                db_band_results->db_stdev[index] += 
                                   db_one_sca_results->db_stdev;
                db_band_results->db_corrected[index] += 
                                   db_one_sca_results->db_corrected;
                sca_counter++;

                /* Populate db_one_sca_results into the db_sca_results 
                   structure */
                db_sca_results->sca[sca_index] = sca_index + 1;
                db_sca_results->db_mean[sca_index] = 
                                db_one_sca_results->db_mean;
                db_sca_results->db_stdev[sca_index] = 
                                db_one_sca_results->db_stdev;
                db_sca_results->db_corrected[sca_index] = 
                                db_one_sca_results->db_corrected;

                /* The normal memory management mechanism has no provision for
                   freeing structure members that have been allocated
                   internally. The manager only has a void pointer so it cannot
                   free the string dupped processing level. Free the string and
                   set the pointers to NULL.  This is a post-process routine so
                   there is no risk of another thread trying to access the
                   memory. */
                free( db_one_sca_results->processing_level );
                db_one_sca_results->processing_level = NULL;
                db_one_sca_results = NULL;
            }
    
            /* Calculate the statistics for the db_band_results structure */
            db_band_results->db_mean[index] = 
                   db_band_results->db_mean[index] / sca_counter;
            db_band_results->db_stdev[index] = 
                   db_band_results->db_stdev[index] / sca_counter;
            db_band_results->db_corrected[index] = 
                   db_band_results->db_corrected[index] / sca_counter;  

            /* Write out L0R SCA level lamp statistics data */
            if (blob->generate_report)
            {
                status = rps_output_lamp_sca_results(band_number, 
                    db_sca_results, report_information, 
                    blob->stim_lamp_pair_id, blob->lamp_on_time, 
                    blob->post_warmup_start_time, blob->post_warmup_stop_time, 
                    blob->post_warmup_start_frame, 
                    blob->post_warmup_stop_frame);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR("Calling rps_output_lamp_sca_results");
                    rps_db_free_lamp_sca_results(db_sca_results);
                    rps_db_free_lamp_band_results(db_band_results);
                    return ERROR;
                } 
            }

            /* Store the L0R part LAMP_RESPONSE_SCA table into database */
            if (blob->store_in_database)
            {
                status = rps_store_lamp_sca_in_database(db_sca_results);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR("Inserting l0r db_sca_results into DB for "
                                  "band_number: %d", band_number);
                    rps_db_free_lamp_sca_results(db_sca_results);
                    rps_db_free_lamp_band_results(db_band_results);
                    return ERROR;
                }
            }
        }

        /* Write out L0R band level lamp statistics data */
        if (blob->generate_report)
        {
            status = rps_output_lamp_band_results(db_band_results, 
                report_information, blob->stim_lamp_pair_id,
                blob->lamp_on_time,
                blob->post_warmup_start_time, blob->post_warmup_stop_time, 
                blob->post_warmup_start_frame, blob->post_warmup_stop_frame);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Calling rps_output_lamp_band_results");
                rps_db_free_lamp_sca_results(db_sca_results);
                rps_db_free_lamp_band_results(db_band_results);
                return ERROR;
            } 
        }

        /* Store the L0R part LAMP_RESPONSE_BAND table into database */
        if (blob->store_in_database)
        {
            /* Insert the L0R part LAMP_RESPONSE_BAND table into database */
            status = rps_store_lamp_band_in_database(db_band_results);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Inserting l0r db_band_results into DB");
                rps_db_free_lamp_sca_results(db_sca_results);
                rps_db_free_lamp_band_results(db_band_results);
                return ERROR;
            }
        }
    }

    /* Loop through each band of the input data */
    for (index= 0; index < blob->number_of_bands; index++)
    {
        band_index = blob->list_of_band_indices[index];
        band_number = blob->list_of_bands[index];
        /* Loop through the SCA for L0R lamp data */
        for (sca_index = 0; sca_index < blob->max_number_scas; sca_index++)
        {
            /* Decrease the MM_L0R_ONE_SCA_STATISTICS memory counter */
            status = decrement_consumer_count(
                    MM_L0R_LAMP_ONE_SCA_STATISTICS, band_index, sca_index);
            if ( status != SUCCESS )
            {
                IAS_LOG_ERROR("Decrementing memory buffer");
                return ERROR;
            }
        }
    }

    /* Make sure the MM_L1R_INTERMEDIATE_STATISTICS buffer is valid, otherwise
       skip the call to the algorithm to make the workflow working if cal/val
       people adjust the location of the histogram statistics algorithm in the
       workflow */
    if ( get_buffer_filled_flag( MM_L1R_INTERMEDIATE_STATISTICS,
         blob->list_of_band_indices[0], 0 ) == 0 )
    {
        IAS_LOG_WARNING("Attempting to use unfilled buffer %s", 
            get_memory_buffer_string(MM_L1R_INTERMEDIATE_STATISTICS));
    }
    else
    {
        /* The string duped processing levels need to be freed before the
           results are used again, otherwise the memory will be leaked */
        free( db_band_results->processing_level );
        db_band_results->processing_level = NULL;
        free( db_sca_results->processing_level );
        db_sca_results->processing_level = NULL;

        /* Loop through each band of the input data */
        for (index = 0; index < blob->number_of_bands; index++)
        {
            band_index = blob->list_of_band_indices[index];
            band_number = blob->list_of_bands[index];
            /* Initialize the db_band_results mean and stdev values */
            db_band_results->band[index] = band_number;
            db_band_results->db_mean[index] = 0;
            db_band_results->db_stdev[index] = 0;
            db_band_results->db_corrected[index] = 0;
            sca_counter = 0;

            /* Populate band number in db_sca_results */
            db_sca_results->band = band_number;

            /* Loop through the SCA for L1R data */
            for (sca_index = 0; sca_index < blob->max_number_scas; sca_index++)
            {
                /* Get one L1R SCA statistics results buffer */
                db_one_sca_results = 
                    (DB_CHAR_LAMP_RESPONSE_ONE_SCA_RESULTS *)
                    get_memory_buffer( MM_L1R_LAMP_ONE_SCA_STATISTICS,
                    CONSUMER, band_index, sca_index);
                if (!db_one_sca_results)
                {
                    IAS_LOG_ERROR("Getting MM_L1R_SCA_STATISTICS "
                                  "memory buffer");
                    rps_db_free_lamp_sca_results(db_sca_results);
                    rps_db_free_lamp_band_results(db_band_results);
                    return ERROR;
                }

                /* Populate processing_level for db_band_results and 
                   db_sca_results structures */
                if (index == 0 && sca_index == 0)
                {
                    db_band_results->processing_level 
                        = strdup(db_one_sca_results->processing_level);
                    db_sca_results->processing_level 
                        = strdup(db_one_sca_results->processing_level);
                    if (!db_band_results->processing_level ||
                        !db_sca_results->processing_level)
                    {
                        IAS_LOG_ERROR("Duplicating processing_level string "
                                      "for db_band_results and db_sca_results");
                        rps_db_free_lamp_sca_results(db_sca_results);
                        rps_db_free_lamp_band_results(db_band_results);
                       return ERROR;
                    }
                }

                /* Calculate the statistics for the db_sca_results structure */
                db_band_results->db_mean[index] += db_one_sca_results->db_mean;
                db_band_results->db_stdev[index] += 
                              db_one_sca_results->db_stdev;
                db_band_results->db_corrected[index] += 
                              db_one_sca_results->db_corrected;
                sca_counter++;

                /* Populate db_one_sca_results into multiple SCA
                   db_sca_results calculate */
                db_sca_results->sca[sca_index] = sca_index + 1;
                db_sca_results->db_mean[sca_index] = 
                              db_one_sca_results->db_mean;
                db_sca_results->db_stdev[sca_index] = 
                              db_one_sca_results->db_stdev;
                db_sca_results->db_corrected[sca_index] =
                              db_one_sca_results->db_corrected;

                /* The normal memory management mechanism has no provision for
                   freeing structure members that have been allocated
                   internally. The manager only has a void pointer so it cannot
                   free the string dupped processing level. Free the string and
                   set the pointers to NULL.  This is a post-process routine so
                   there is no risk of another thread trying to access the
                   memory. */
                free( db_one_sca_results->processing_level );
                db_one_sca_results->processing_level = NULL;
                db_one_sca_results = NULL;
            }
    
            /* Calculate the statistics for the db_band_results structure */
            db_band_results->db_mean[index] = 
                   db_band_results->db_mean[index] / sca_counter;
            db_band_results->db_stdev[index] = 
                   db_band_results->db_stdev[index] / sca_counter;
            db_band_results->db_corrected[index] = 
                   db_band_results->db_corrected[index] / sca_counter;  

            /* Write out L1R SCA level lamp statistics data */
            if (blob->generate_report)
            {
                status = rps_output_lamp_sca_results(band_number, 
                    db_sca_results, report_information, blob->stim_lamp_pair_id,
                    blob->lamp_on_time, blob->post_warmup_start_time,
                    blob->post_warmup_stop_time, blob->post_warmup_start_frame,
                    blob->post_warmup_stop_frame);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR("Calling rps_output_lamp_sca_results");
                    rps_db_free_lamp_sca_results(db_sca_results);
                    rps_db_free_lamp_band_results(db_band_results);
                    return ERROR;
                } 
            }

            /* Store the L1R part LAMP_RESPONSE_SCA table into database */
            if (blob->store_in_database)
            {
                status = rps_store_lamp_sca_in_database(db_sca_results);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR("Inserting l1r db_sca_results into DB for "
                                  "band_number: %d", band_number);
                    rps_db_free_lamp_sca_results(db_sca_results);
                    rps_db_free_lamp_band_results(db_band_results);
                    return ERROR;
                }
            }
        }

        /* Write out L1R band level lamp statistics data */
        if (blob->generate_report)
        {
            status = rps_output_lamp_band_results(db_band_results, 
                report_information, blob->stim_lamp_pair_id, blob->lamp_on_time,
                blob->post_warmup_start_time, blob->post_warmup_stop_time, 
                blob->post_warmup_start_frame, blob->post_warmup_stop_frame);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Calling rps_output_lamp_band_results");
                rps_db_free_lamp_sca_results(db_sca_results);
                rps_db_free_lamp_band_results(db_band_results);
                return ERROR;
            } 
        }

        /* Store the L1R part LAMP_RESPONSE_BAND table into database */
        if (blob->store_in_database)
        {
            status = rps_store_lamp_band_in_database(db_band_results);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Inserting l1r db_band_results into DB");
                rps_db_free_lamp_sca_results(db_sca_results);
                rps_db_free_lamp_band_results(db_band_results);
                return ERROR;
            }
        }
    }

    /* Loop through each band of the input data */
    for (index= 0; index < blob->number_of_bands; index++)
    {
        band_index = blob->list_of_band_indices[index];
        band_number = blob->list_of_bands[index];
        /* Loop through the SCA for L0R lamp data */
        for (sca_index = 0; sca_index < blob->max_number_scas; sca_index++)
        {
            /* Decrease the MM_L1R_ONE_SCA_STATISTICS memory counter */
            status = decrement_consumer_count(
                    MM_L1R_LAMP_ONE_SCA_STATISTICS, band_index, sca_index);
            if ( status != SUCCESS )
            {
                IAS_LOG_ERROR("Decrementing memory buffer");
                return ERROR;
            }
        }
    }

    /* Free the db_sca_results structure */
    rps_db_free_lamp_sca_results(db_sca_results);

    /* Free the db_band_results structure */
    rps_db_free_lamp_band_results(db_band_results);

    /* Allocate memory for the DB_CHAR_LAMP_RESPONSE_RESULTS
       structure */
    db_response_results = malloc(sizeof(DB_CHAR_LAMP_RESPONSE_RESULTS));
    if ( !db_response_results )
    {
        IAS_LOG_ERROR("Allocating the lamp response structure");
        return ERROR;
    }

    /* Get the db_response_results data from the blob */
    db_response_results->dbase = blob->database_information.db;
    strncpy(db_response_results->rad_common_id, 
        blob->database_information.rad_common_id, 
        sizeof(db_response_results->rad_common_id));
    db_response_results->rad_common_id[sizeof(
        db_response_results->rad_common_id) - 1] = '\0';
    db_response_results->lamp_pair_id = blob->stim_lamp_pair_id;
    db_response_results->lamp_on_time = blob->lamp_on_time;
    db_response_results->interval_start_line = blob->post_warmup_start_frame;
    db_response_results->interval_stop_line = blob->post_warmup_stop_frame;
    db_response_results->lamp_warmup_time = blob->lamp_warmup_time;
    db_response_results->lamp_warmup_start_line 
                        = blob->lamp_warmup_start_frame;
    db_response_results->lamp_warmup_stop_line 
                        = blob->lamp_warmup_stop_frame;

    db_response_results->interval_start_time 
        = blob->post_warmup_start_time;
    db_response_results->interval_stop_time 
        = blob->post_warmup_stop_time;

    /* Write out lamp response statistics data */
    if (blob->generate_report)
    {
        status = rps_output_lamp_response_results(db_response_results, 
                      report_information);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Calling rps_output_lamp_response_results");
            free(db_response_results);
            return ERROR;
        } 
    }

    /* Store Lamp Response data into DB here */
    if(blob->store_in_database)
    {
        status = rps_store_lamp_response_in_database(db_response_results);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Inserting db_response_results into DB");
            free(db_response_results);
            return ERROR;
        }
    }

    /* Free the allocated db_response_results memory */
    free(db_response_results);

    /* Allocate memory for the DB_CHAR_LAMP_TELEMETRY_RESULTS
       structure */
    db_telemetry_results = malloc(sizeof(
                           DB_CHAR_LAMP_TELEMETRY_RESULTS));
    if (!db_telemetry_results)
    {
        IAS_LOG_ERROR("Allocating the lamp telemetry structure");
        return ERROR;
    }

    /* Populate the common parameters for the db_telemetry_results */
    db_telemetry_results->dbase = blob->database_information.db;
    strncpy(db_telemetry_results->rad_common_id, 
        blob->database_information.rad_common_id, 
        sizeof(db_telemetry_results->rad_common_id));
    db_telemetry_results->rad_common_id[sizeof(
        db_telemetry_results->rad_common_id) - 1] = '\0';

    /* Read in the post warmup focal plane temperatures */
    status = read_post_warmup_focal_plane_temperatures( 
                &focal_plane_temp1, &focal_plane_temp2, &num_focal_temp1);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading post warmup focal plane temperatures");
        free(db_telemetry_results);
        return ERROR;
    }    
    num_focal_temp2 = num_focal_temp1;

    /* Now the mean and stdev for focal_plane_temp1 */
    status = compute_mean(focal_plane_temp1, num_focal_temp1, 
                          &mean_fp_temp1);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating focal plane mean temperature1");
        free(focal_plane_temp1);
        free(focal_plane_temp2);
        free(db_telemetry_results);
        return ERROR;
    }

    status = compute_stdev(focal_plane_temp1, num_focal_temp1, 
                           mean_fp_temp1, &stdev_fp_temp1);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating stdev of focal plane temperature1");
        free(focal_plane_temp1);
        free(focal_plane_temp2);
        free(db_telemetry_results);
        return ERROR;
    }

    /* Now the mean and stdev for focal_plane_temp2 */
    status = compute_mean(focal_plane_temp2, num_focal_temp2, 
                          &mean_fp_temp2);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating focal plane mean temperature2");
        free(focal_plane_temp1);
        free(focal_plane_temp2);
        free(db_telemetry_results);
        return ERROR;
    }

    status = compute_stdev(focal_plane_temp2, num_focal_temp2, 
                           mean_fp_temp2, &stdev_fp_temp2);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating stdev of focal plane temperature2");
        free(focal_plane_temp1);
        free(focal_plane_temp2);
        free(db_telemetry_results);
        return ERROR;
    }

    /* Populate in focal plane temperature results */
    db_telemetry_results->fp_temp_7_mean = mean_fp_temp1; 
    db_telemetry_results->fp_temp_7_stdev = stdev_fp_temp1; 
    db_telemetry_results->fp_temp_14_mean = mean_fp_temp2; 
    db_telemetry_results->fp_temp_14_stdev = stdev_fp_temp2; 

    /* Clean the tocal_plane_temp data buffer */
    free(focal_plane_temp1);
    free(focal_plane_temp2);

    /* Read in the post warmup lamp telemetry data */
    status = read_post_warmup_lamp_telemetry(
        &lamp_current_a, &lamp_voltage_a, &photodiode_micro_amp_a, 
        &lamp_housing_temp_a, &monitor_diode_temp_a, &lamp_current_b, 
        &lamp_voltage_b, &photodiode_micro_amp_b, &lamp_housing_temp_b, 
        &monitor_diode_temp_b, &num_telemetry_data); 
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading in the post warmup lamp telemetry data");
        clean_up_lamp_telemetry(lamp_current_a, lamp_voltage_a, 
            photodiode_micro_amp_a, lamp_housing_temp_a, monitor_diode_temp_a, 
            lamp_current_b, lamp_voltage_b, photodiode_micro_amp_b, 
            lamp_housing_temp_b, monitor_diode_temp_b);
        free(db_telemetry_results);
        return ERROR;
    }

    /* Calculate mean and std dev for the lamp current data */
    status = compute_mean(lamp_current_a, num_telemetry_data, 
                           &db_telemetry_results->lamp_current_a_mean);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating mean of lamp current A");
        clean_up_lamp_telemetry(lamp_current_a, lamp_voltage_a, 
            photodiode_micro_amp_a, lamp_housing_temp_a, monitor_diode_temp_a, 
            lamp_current_b, lamp_voltage_b, photodiode_micro_amp_b, 
            lamp_housing_temp_b, monitor_diode_temp_b);
        free(db_telemetry_results);
        return ERROR;
    }

    status = compute_stdev(lamp_current_a, num_telemetry_data, 
                         db_telemetry_results->lamp_current_a_mean, 
                         &db_telemetry_results->lamp_current_a_stdev);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating stdev of lamp current A");
        clean_up_lamp_telemetry(lamp_current_a, lamp_voltage_a, 
            photodiode_micro_amp_a, lamp_housing_temp_a, monitor_diode_temp_a, 
            lamp_current_b, lamp_voltage_b, photodiode_micro_amp_b, 
            lamp_housing_temp_b, monitor_diode_temp_b);
        free(db_telemetry_results);
        return ERROR;
    }

    status = compute_mean(lamp_current_b, num_telemetry_data, 
                           &db_telemetry_results->lamp_current_b_mean);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating mean of lamp current B");
        clean_up_lamp_telemetry(lamp_current_a, lamp_voltage_a, 
            photodiode_micro_amp_a, lamp_housing_temp_a, monitor_diode_temp_a, 
            lamp_current_b, lamp_voltage_b, photodiode_micro_amp_b, 
            lamp_housing_temp_b, monitor_diode_temp_b);
        free(db_telemetry_results);
        return ERROR;
    }

    status = compute_stdev(lamp_current_b, num_telemetry_data, 
                         db_telemetry_results->lamp_current_b_mean, 
                         &db_telemetry_results->lamp_current_b_stdev);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating stdev of lamp current B");
        clean_up_lamp_telemetry(lamp_current_a, lamp_voltage_a, 
            photodiode_micro_amp_a, lamp_housing_temp_a, monitor_diode_temp_a, 
            lamp_current_b, lamp_voltage_b, photodiode_micro_amp_b, 
            lamp_housing_temp_b, monitor_diode_temp_b);
        free(db_telemetry_results);
        return ERROR;
    }

    /* Calculate mean and std dev for the lamp voltage data */
    status = compute_mean(lamp_voltage_a, num_telemetry_data, 
                          &db_telemetry_results->lamp_voltage_a_mean);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating mean of lamp voltage A");
        clean_up_lamp_telemetry(lamp_current_a, lamp_voltage_a, 
            photodiode_micro_amp_a, lamp_housing_temp_a, monitor_diode_temp_a, 
            lamp_current_b, lamp_voltage_b, photodiode_micro_amp_b, 
            lamp_housing_temp_b, monitor_diode_temp_b);
        free(db_telemetry_results);
        return ERROR;
    }

    status = compute_stdev(lamp_voltage_a, num_telemetry_data, 
                          db_telemetry_results->lamp_voltage_a_mean, 
                          &db_telemetry_results->lamp_voltage_a_stdev);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating stdev of lamp voltage A");
        clean_up_lamp_telemetry(lamp_current_a, lamp_voltage_a, 
            photodiode_micro_amp_a, lamp_housing_temp_a, monitor_diode_temp_a, 
            lamp_current_b, lamp_voltage_b, photodiode_micro_amp_b, 
            lamp_housing_temp_b, monitor_diode_temp_b);
        free(db_telemetry_results);
        return ERROR;
    }
             
    status = compute_mean(lamp_voltage_b, num_telemetry_data, 
                          &db_telemetry_results->lamp_voltage_b_mean);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating mean of lamp voltage B");
        clean_up_lamp_telemetry(lamp_current_a, lamp_voltage_a, 
            photodiode_micro_amp_a, lamp_housing_temp_a, monitor_diode_temp_a, 
            lamp_current_b, lamp_voltage_b, photodiode_micro_amp_b, 
            lamp_housing_temp_b, monitor_diode_temp_b);
        free(db_telemetry_results);
        return ERROR;
    }

    status = compute_stdev(lamp_voltage_b, num_telemetry_data, 
                          db_telemetry_results->lamp_voltage_b_mean, 
                          &db_telemetry_results->lamp_voltage_b_stdev);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating stdev of lamp voltage B");
        clean_up_lamp_telemetry(lamp_current_a, lamp_voltage_a, 
            photodiode_micro_amp_a, lamp_housing_temp_a, monitor_diode_temp_a, 
            lamp_current_b, lamp_voltage_b, photodiode_micro_amp_b, 
            lamp_housing_temp_b, monitor_diode_temp_b);
        free(db_telemetry_results);
        return ERROR;
    }
             
    /* Calculate mean and std dev for the photodiode current data */
    status = compute_mean(photodiode_micro_amp_a, num_telemetry_data, 
                       &db_telemetry_results->photodiode_micro_amp_a_mean);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating mean of photodiode current A");
        clean_up_lamp_telemetry(lamp_current_a, lamp_voltage_a, 
            photodiode_micro_amp_a, lamp_housing_temp_a, monitor_diode_temp_a, 
            lamp_current_b, lamp_voltage_b, photodiode_micro_amp_b, 
            lamp_housing_temp_b, monitor_diode_temp_b);
        free(db_telemetry_results);
        return ERROR;
    }

    status = compute_stdev(photodiode_micro_amp_a, num_telemetry_data, 
                       db_telemetry_results->photodiode_micro_amp_a_mean, 
                       &db_telemetry_results->photodiode_micro_amp_a_stdev);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating stdev of photodiode current A");
        clean_up_lamp_telemetry(lamp_current_a, lamp_voltage_a, 
            photodiode_micro_amp_a, lamp_housing_temp_a, monitor_diode_temp_a, 
            lamp_current_b, lamp_voltage_b, photodiode_micro_amp_b, 
            lamp_housing_temp_b, monitor_diode_temp_b);
        free(db_telemetry_results);
        return ERROR;
    }
             
    status = compute_mean(photodiode_micro_amp_b, num_telemetry_data, 
                       &db_telemetry_results->photodiode_micro_amp_b_mean);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating mean of photodiode current B");
        clean_up_lamp_telemetry(lamp_current_a, lamp_voltage_a, 
            photodiode_micro_amp_a, lamp_housing_temp_a, monitor_diode_temp_a, 
            lamp_current_b, lamp_voltage_b, photodiode_micro_amp_b, 
            lamp_housing_temp_b, monitor_diode_temp_b);
        free(db_telemetry_results);
        return ERROR;
    }

    status = compute_stdev(photodiode_micro_amp_b, num_telemetry_data, 
                       db_telemetry_results->photodiode_micro_amp_b_mean, 
                       &db_telemetry_results->photodiode_micro_amp_b_stdev);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating stdev of photodiode current B");
        clean_up_lamp_telemetry(lamp_current_a, lamp_voltage_a, 
            photodiode_micro_amp_a, lamp_housing_temp_a, monitor_diode_temp_a, 
            lamp_current_b, lamp_voltage_b, photodiode_micro_amp_b, 
            lamp_housing_temp_b, monitor_diode_temp_b);
        free(db_telemetry_results);
        return ERROR;
    }
             
    /* Calculate mean and std dev for the lamp housing temperature data */
    status = compute_mean(lamp_housing_temp_a, num_telemetry_data, 
                    &db_telemetry_results->lamp_housing_temp_a_mean);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating mean of lamp housing temperature A");
        clean_up_lamp_telemetry(lamp_current_a, lamp_voltage_a, 
            photodiode_micro_amp_a, lamp_housing_temp_a, monitor_diode_temp_a, 
            lamp_current_b, lamp_voltage_b, photodiode_micro_amp_b, 
            lamp_housing_temp_b, monitor_diode_temp_b);
        free(db_telemetry_results);
        return ERROR;
    }

    status = compute_stdev(lamp_housing_temp_a, num_telemetry_data, 
                    db_telemetry_results->lamp_housing_temp_a_mean, 
                    &db_telemetry_results->lamp_housing_temp_a_stdev);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating stdev of lamp housing temperature A");
        clean_up_lamp_telemetry(lamp_current_a, lamp_voltage_a, 
            photodiode_micro_amp_a, lamp_housing_temp_a, monitor_diode_temp_a, 
            lamp_current_b, lamp_voltage_b, photodiode_micro_amp_b, 
            lamp_housing_temp_b, monitor_diode_temp_b);
        free(db_telemetry_results);
        return ERROR;
    }
             
    status = compute_mean(lamp_housing_temp_b, num_telemetry_data, 
                    &db_telemetry_results->lamp_housing_temp_b_mean);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating mean of lamp housing temperature B");
        clean_up_lamp_telemetry(lamp_current_a, lamp_voltage_a, 
            photodiode_micro_amp_a, lamp_housing_temp_a, monitor_diode_temp_a, 
            lamp_current_b, lamp_voltage_b, photodiode_micro_amp_b, 
            lamp_housing_temp_b, monitor_diode_temp_b);
        free(db_telemetry_results);
        return ERROR;
    }

    status = compute_stdev(lamp_housing_temp_b, num_telemetry_data, 
                    db_telemetry_results->lamp_housing_temp_b_mean, 
                    &db_telemetry_results->lamp_housing_temp_b_stdev);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating stdev of lamp housing temperature B");
        clean_up_lamp_telemetry(lamp_current_a, lamp_voltage_a, 
            photodiode_micro_amp_a, lamp_housing_temp_a, monitor_diode_temp_a, 
            lamp_current_b, lamp_voltage_b, photodiode_micro_amp_b, 
            lamp_housing_temp_b, monitor_diode_temp_b);
        free(db_telemetry_results);
        return ERROR;
    }
             
    /* Calculate mean and std dev for the monitor diode temperature data */
    status = compute_mean(monitor_diode_temp_a, num_telemetry_data, 
                    &db_telemetry_results->monitor_diode_temp_a_mean);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating mean of monitor diode temperature A");
        clean_up_lamp_telemetry(lamp_current_a, lamp_voltage_a, 
            photodiode_micro_amp_a, lamp_housing_temp_a, monitor_diode_temp_a, 
            lamp_current_b, lamp_voltage_b, photodiode_micro_amp_b, 
            lamp_housing_temp_b, monitor_diode_temp_b);
        free(db_telemetry_results);
        return ERROR;
    }

    status = compute_stdev(monitor_diode_temp_a, num_telemetry_data, 
                    db_telemetry_results->monitor_diode_temp_a_mean, 
                    &db_telemetry_results->monitor_diode_temp_a_stdev);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating stdev of monitor diode temperature A");
        clean_up_lamp_telemetry(lamp_current_a, lamp_voltage_a, 
            photodiode_micro_amp_a, lamp_housing_temp_a, monitor_diode_temp_a, 
            lamp_current_b, lamp_voltage_b, photodiode_micro_amp_b, 
            lamp_housing_temp_b, monitor_diode_temp_b);
        free(db_telemetry_results);
        return ERROR;
    }

    status = compute_mean(monitor_diode_temp_b, num_telemetry_data, 
                    &db_telemetry_results->monitor_diode_temp_b_mean);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating mean of monitor diode temperature B");
        clean_up_lamp_telemetry(lamp_current_a, lamp_voltage_a, 
            photodiode_micro_amp_a, lamp_housing_temp_a, monitor_diode_temp_a, 
            lamp_current_b, lamp_voltage_b, photodiode_micro_amp_b, 
            lamp_housing_temp_b, monitor_diode_temp_b);
        free(db_telemetry_results);
        return ERROR;
    }

    status = compute_stdev(monitor_diode_temp_b, num_telemetry_data, 
                    db_telemetry_results->monitor_diode_temp_b_mean, 
                    &db_telemetry_results->monitor_diode_temp_b_stdev);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating stdev of monitor diode temperature B");
        clean_up_lamp_telemetry(lamp_current_a, lamp_voltage_a, 
            photodiode_micro_amp_a, lamp_housing_temp_a, monitor_diode_temp_a, 
            lamp_current_b, lamp_voltage_b, photodiode_micro_amp_b, 
            lamp_housing_temp_b, monitor_diode_temp_b);
        free(db_telemetry_results);
        return ERROR;
    }

    /* Clean up lamp telemetry data buffers */
    clean_up_lamp_telemetry(lamp_current_a, lamp_voltage_a, 
        photodiode_micro_amp_a, lamp_housing_temp_a, monitor_diode_temp_a, 
        lamp_current_b, lamp_voltage_b, photodiode_micro_amp_b, 
        lamp_housing_temp_b, monitor_diode_temp_b);

    /* Write out lamp telemetry statistics data */
    if (blob->generate_report)
    {
        status = rps_output_lamp_telemetry_results(db_telemetry_results, 
                 report_information, blob->stim_lamp_pair_id, 
                 blob->lamp_on_time, blob->post_warmup_start_time, 
                 blob->post_warmup_stop_time, blob->post_warmup_start_frame, 
                 blob->post_warmup_stop_frame);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Calling rps_output_lamp_telemetry_results");
            free(db_telemetry_results);
            return ERROR;
        } 
    }

    /* Store Lamp Telemetry data into DB here */
    if(blob->store_in_database)
    {
        status = rps_store_lamp_telemetry_in_database(db_telemetry_results);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Inserting db_telemetry_results into DB");
            free(db_telemetry_results);
            return ERROR;
        }
    }

    /* Free the allocated memory */
    free(db_telemetry_results);

    return SUCCESS;
}  /* END char_lamp_detector_response_postprocess */

/*----------------------------------------------------------------------------
NAME: compute_mean  

PURPOSE: Mean calculation for floating point numbers

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int compute_mean
(
    const float *sample,       /* I: pointer to the data */
    int nsamps,                /* I: number of samples */
    float *mean                /* O: mean of the data */
)
{
    int i;                               /* loop variable */
    float sum=0.0;             /* sum of the samples */

    if (nsamps <= 0)
    {
        IAS_LOG_ERROR("Sample size is invalid.");
        *mean = 0.0;
        return ERROR;
    }

    for (i = 0; i < nsamps; i++)
        sum += sample[i];

    *mean = sum / nsamps;

    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME: compute_stdev  

PURPOSE: Standard deviation calculation for floating point numbers

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int compute_stdev
(
    const float *sample,   /* I: pointer to the data */
    int nsamps,            /* I: number of samples */
    float mean,            /* I: mean of the samples */
    float *std_dev         /* O: standard deviation of the data */
)
{
    int i;                 /* loop variable */
    float sum_sqr=0.0;     /* sum of the squared difference from mean */

    /* Standard deviation calculation.  If number of samples is invalid
       return ERROR. */
    if (nsamps <= 0)
    {
        IAS_LOG_ERROR("Sample size is invalid.");
        return ERROR;
    }

    if (nsamps == 1)
    {
        *std_dev = 0.0;
        return SUCCESS; 
    }

    for (i = 0; i < nsamps; i++)
        sum_sqr += sample[i] * sample[i];  /* sum of squares */
        
    sum_sqr -= nsamps * (mean * mean);     /* subtract mean squared for each
                                              sample */
    /* Use the absolute value to avoid round off errors */
    *std_dev = sqrt(fabs(sum_sqr) / (nsamps-1));

    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME:  clean_up_lamp_telemetry

PURPOSE: Free the allocated lamp telemetry data buffers

RETURNS: None
-----------------------------------------------------------------------------*/
void clean_up_lamp_telemetry
(
    float *lamp_current_a,
    float *lamp_voltage_a,
    float *photodiode_micro_amp_a,
    float *lamp_housing_temp_a,
    float *monitor_diode_temp_a,
    float *lamp_current_b,
    float *lamp_voltage_b,
    float *photodiode_micro_amp_b,
    float *lamp_housing_temp_b,
    float *monitor_diode_temp_b
)
{
    if (lamp_current_a != NULL)
        free(lamp_current_a);
    if (lamp_voltage_a != NULL)
        free(lamp_voltage_a);
    if (photodiode_micro_amp_a != NULL)
        free(photodiode_micro_amp_a);
    if (lamp_housing_temp_a != NULL)
        free(lamp_housing_temp_a);
    if (monitor_diode_temp_a != NULL)
        free(monitor_diode_temp_a);
    if (lamp_current_b != NULL)
        free(lamp_current_b);
    if (lamp_voltage_b != NULL)
        free(lamp_voltage_b);
    if (photodiode_micro_amp_b != NULL)
        free(photodiode_micro_amp_b);
    if (lamp_housing_temp_b != NULL)
        free(lamp_housing_temp_b);
    if (monitor_diode_temp_b != NULL)
        free(monitor_diode_temp_b);
}

