/*----------------------------------------------------------------------------
NAME:    convert_to_float

PURPOSE: Convert integer image and associated buffers to floating point.
         Due to the simplicity of the conversion, there is no corresponding
         algorithm for this API.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
#include "ias_logging.h"
#include "rps_local.h"

static int convert_to_float
(
    BLOB *blob,         /* I: Miscellaneous parameters */
    int   band_index,   /* I: */
    int   sca_index,    /* I: */
    int   shift_flag    /* I: Flag to specify whether the data should be
                              shifted when converted: 1 (true) shift the
                              data, 0 (false) don't shift */
)
{
    uint16_t *input_buffer;  /* Level 0 input buffer in DN */
    float *output_buffer;    /* Level 0 output buffer in floating point DN */
    unsigned int i;          /* Plain old loop counter */
    RPS_BAND_ATTRIBUTES *band_attributes = get_band_attributes( band_index );
    int band_number = band_attributes->band_number;
    int num_lines = band_attributes->number_lines;
    int num_samps = band_attributes->number_detectors;
    unsigned int buffer_size;
    int status;

    /* Get the buffers */
    input_buffer = (uint16_t *)get_memory_buffer( IMAGE_L0R, CONSUMER,
        band_index, sca_index );

    output_buffer = (float *)get_memory_buffer( IMAGE_L1R, SETTER,
        band_index, sca_index );
    if ( !input_buffer || !output_buffer )
    {
        IAS_LOG_ERROR("Getting memory buffers for band %d SCA %d",
            band_number, sca_index + 1 );
        return ERROR;
    }


    /* Convert the DNs to floating point.  Also, if the data is the upper 12
       bits, shift the data.  The truncation setting from the L0R image
       header will be 0 if the upper 12 bits were transmitted. There are errors
       associated with A/D conversion and truncation of the lower two bits of 
       non-barrel shifted data.  The addition of 1.5 reduces the error on 
       average to the same error of the barrel shifted data that bias is 
       derived from. */
    buffer_size = num_lines * num_samps;
    if ( shift_flag )
    {
        for ( i = 0; i < buffer_size; i++ )
            output_buffer[i] = (float)( (input_buffer[i] << 2 ) + 1.5 );

        if (set_rps_level(band_index, sca_index, 
                IAS_RPS_LEVEL_CONVERT_FLOAT_CORRECT) != SUCCESS)
        {
            IAS_LOG_ERROR("RPS Level update failed");
            return ERROR;
        }
    }
    else
    {
        for ( i = 0; i < buffer_size; i++ )
            output_buffer[i] = (float)input_buffer[i];

        if (set_rps_level(band_index, sca_index, 
                IAS_RPS_LEVEL_CONVERT_FLOAT_NOSHIFT) != SUCCESS)
        {
            IAS_LOG_ERROR("RPS Level update failed");
            return ERROR;
        }
    }

    /* Tell the memory management that this buffer is now valid */
    set_buffer_filled_flag( IMAGE_L1R, band_index, sca_index );

    /* Tell memory management that the data was used */
    status = decrement_consumer_count( IMAGE_L0R, band_index, sca_index );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Decrementing memory buffer");
        return ERROR;
    }
    status = decrement_consumer_count( IMAGE_L1R, band_index, sca_index );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Decrementing memory buffer");
        return ERROR;
    }

    return SUCCESS;
}  /* END convert_to_float */


/*----------------------------------------------------------------------------
NAME:    api_convert_to_float

PURPOSE: Call the internal convert_to_float with the image band and its 
         associated bands.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int api_convert_to_float
(
    BLOB *blob,         /* I: Miscellaneous parameters */
    int   band_index,   /* I: */
    int   sca_index,    /* I: */
    int   algo_index    /* I: Algorithm index */
)
{
    int status;
    int shift_flag = 0; /* Initialize the flag to false */
    RPS_BAND_ATTRIBUTES *band_attributes = get_band_attributes( band_index );
    int band_number = band_attributes->band_number;
    int sca_number = sca_index + 1;

    /* Set the shift flag to true if the image data has been truncated.
       Only set the flag for OLI bands. */
    if ( band_attributes->sensor_id == IAS_OLI )
    {
        if ( blob->image_data_truncation_setting == 0 )
            shift_flag = 1;
    }
    else
    {
        shift_flag = 0;
    }

    /* Convert the image data */
    status = convert_to_float( blob, band_index, sca_index, shift_flag );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Return from convert to float for band %d SCA %d",
            band_number, sca_number );
        return ERROR;
    }

    /* VRP and blind data are never truncated */
    shift_flag = 0;

    /* Convert the VRP data if this band has a VRP associated with it */
    if ( band_attributes->vrp_band_number > 0 )
    {
        status = convert_to_float( blob, band_attributes->vrp_band_index,
            sca_index, shift_flag );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Return from convert to float for band %d SCA %d",
                band_attributes->vrp_band_number, sca_number );
            return ERROR;
        }
    }

    /* Convert the blind band associated with the image band but only if
       it is to be processed by this band */
    if ( band_attributes->blind_band_number > 0 
        && band_attributes->process_blind_band_flag )
    {
        status = convert_to_float( blob, band_attributes->blind_band_index,
            sca_index, shift_flag );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Return from convert to float for band %d SCA %d",
                band_attributes->blind_band_number, sca_number );
            return ERROR;
        }
    }

    return SUCCESS;
}  /* END api_convert_to_float */


/*-------------------------------------------------------------------------
NAME:    convert_to_float_preprocess

PURPOSE: Register the memory buffers needed by this algorithm

RETURNS: SUCCESS or ERROR
--------------------------------------------------------------------------*/
int convert_to_float_preprocess
(
    BLOB *blob          /* I: Miscellaneous parameters */
)
{
    /* Define a list of memory buffers needed by the algorithm */
    MEMORY_BUFFER_TYPE memory_buffers[] =
    {
        IMAGE_L0R,
        IMAGE_L1R
    };
    int number_of_memory_buffers = sizeof(memory_buffers) /
        sizeof(MEMORY_BUFFER_TYPE);
    int i;
    int status;

    /* Increment the consumer counts for the memory buffers */
    for ( i = 0; i < number_of_memory_buffers; i++ )
    {
        status = increment_consumer_count( memory_buffers[i] );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Incrementing %s consumer count",
                get_memory_buffer_string( memory_buffers[i] ));
            return ERROR;
        }
    }

    return SUCCESS;
}  /* END convert_to_float_preprocess */
