/*****************************************************************************
NAME:      api_dropped_frame_characterization

PURPOSE:   Gather and package input data for the dropped frame
           characterization algorithm, then execute the algorithm.

RETURNS:   algorithm return status

ROUTINES:  dropped_frame_characterization     (internal routine)
           api_dropped_frame_characterization
           dropped_frame_characterization_preprocess
*****************************************************************************/
#include "ias_logging.h"
#include "ias_radiometry.h"
#include "ias_satellite_attributes.h"
#include "ias_threadpool.h"
#include "rps_local.h"
#include "rps_blob.h"

static int number_of_memory_buffers;
static MEMORY_BUFFER_TYPE memory_buffers[] =
{
    IMAGE_L0R,
    MM_L0R_TIRS_FRAME_HEADER,
    MM_L0R_OLI_FRAME_HEADER
};

/***************************************************************************
 NAME:     dropped_frame_characterization

 PURPOSE:  Generic "wrapper" function around RPS library call

 RETURNS:  Integer status code of SUCCESS or ERROR
****************************************************************************/
static int dropped_frame_characterization
(
    BLOB *blob,              /* I: Various input parameters */
    int band_index,          /* I: */
    int sca_index            /* I: */
)
{
    RPS_BAND_ATTRIBUTES *band_attributes = get_band_attributes(band_index);
    const IAS_BAND_ATTRIBUTES *ias_atts = NULL;
    const void *l0r_frame_headers = NULL;    /* Frame header array */
    int band_number = band_attributes->band_number;
    int number_lines = band_attributes->number_lines;
    int number_detectors = band_attributes->number_detectors;
    int number_frame_headers; /* Number of existing frame headers */
    int i;                    /* Loop counter */
    int status;
    int lines_per_frame;
    int frames_to_process;
    int starting_frame;


    /* Get band_number attribute information */
    ias_atts = ias_sat_attr_get_band_attributes(band_number);
    if (ias_atts == NULL)
    {
        IAS_LOG_ERROR("Cannot initialize band attributes information");
        return ERROR;
    }

    /* The image must be covered by the actual frames in the image */
    lines_per_frame = ias_sat_attr_get_lines_per_frame(
        band_attributes->band_number );
    if ( lines_per_frame == ERROR )
    {
        IAS_LOG_ERROR("Getting lines per frame for band %d",
            band_attributes->band_number );
        return ERROR;
    }

    /* Get the array of frame header data for the current band and SCA.
       In case the image start and/or number of lines aren't for the full
       image the pointer needs to be adjusted.

       This needs to be cast to a (void *) pointer--it will be explicitly
       cast to the proper type in ias_rps_dropped_frame_characterization
       for the sensor in question. */
    starting_frame = band_attributes->starting_line / lines_per_frame;
    if (ias_atts->sensor_id == IAS_OLI)
    {
        IAS_L0R_OLI_FRAME_HEADER *oli_header =
            (IAS_L0R_OLI_FRAME_HEADER *) get_const_memory_buffer(
            MM_L0R_OLI_FRAME_HEADER, band_index, sca_index);
        number_frame_headers = get_memory_buffer_number_values(
            MM_L0R_OLI_FRAME_HEADER, band_index, sca_index);
        oli_header += starting_frame;
        l0r_frame_headers = (void *)oli_header;
    }
    else if (ias_atts->sensor_id == IAS_TIRS)
    {
        IAS_L0R_TIRS_FRAME_HEADER *tirs_header =
            (IAS_L0R_TIRS_FRAME_HEADER *)get_const_memory_buffer(
            MM_L0R_TIRS_FRAME_HEADER, band_index, sca_index);
        number_frame_headers = get_memory_buffer_number_values(
            MM_L0R_TIRS_FRAME_HEADER, band_index, sca_index);
        tirs_header += starting_frame;
        l0r_frame_headers = (void *)tirs_header;
    }
    else   /* Should never occur, but in case it does... */
    {
        IAS_LOG_ERROR("Invalid sensor ID for band %d", band_number);
        return ERROR;
    }
    if (l0r_frame_headers == NULL)
    {
        IAS_LOG_ERROR("Getting frame header data buffer");
        return ERROR;
    }

    /* Update the frames to process */
    frames_to_process = band_attributes->number_lines / lines_per_frame;
    if ( frames_to_process > number_frame_headers )
    {
        IAS_LOG_WARNING("Number of image lines %d exceeds the number of "
            "L0R frame headers %d", number_lines, number_frame_headers);
    }

    /* Call the dropped frame characterization routine */
    status = ias_rps_dropped_frame_characterization(l0r_frame_headers,
        ias_atts->sensor_id, band_number, number_detectors, 
        frames_to_process, blob->pixel_mask[band_index][sca_index]);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("From ias_rps_dropped_frame_characterization "
            "band_number %d sca_number %d", band_number, sca_index + 1);
        return ERROR;
    }

    /* Let RPS know we're done with this data */
    for (i = 0; i < number_of_memory_buffers; i++)
    {
        decrement_consumer_count(memory_buffers[i], band_index, sca_index);
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }
    }

    return SUCCESS;
}  /* END -- dropped_frame_characterization */



int api_dropped_frame_characterization
(
    BLOB *blob,              /* I: Various input parameters */
    int band_index,          /* I: */
    int sca_index,           /* I: */
    int algorithm_index      /* I: Algorithm index */
)
{
    RPS_BAND_ATTRIBUTES *band_atts = get_band_attributes(band_index);
    int vrp_band_number = band_atts->vrp_band_number;
    int blind_band_number = band_atts->blind_band_number;
    int process_blind_band_flag = band_atts->process_blind_band_flag;
    int vrp_band_index = band_atts->vrp_band_index;
    int blind_band_index = band_atts->blind_band_index;
    int status;

    /* Call the wrapper function for a "normal" imaging band */
    status = dropped_frame_characterization(blob, band_index, sca_index);
    if (status != SUCCESS)
       return ERROR;

    /* Call the wrapper for a VRP band */
    if (vrp_band_number > 0)
    {
        status = dropped_frame_characterization(blob, vrp_band_index,
            sca_index );
        if (status != SUCCESS)
            return ERROR;
    }

    /* Call the wrapper for a blind band */
    if ((blind_band_number > 0) && process_blind_band_flag)
    {
        status = dropped_frame_characterization(blob, blind_band_index,
            sca_index);
        if (status != SUCCESS)
            return ERROR;
    }

    return SUCCESS;
}   /* END -- api_dropped_frame_characterization */



/*************************************************************************
 NAME:     dropped_frame_characterization_preprocess

 PURPOSE:  Increments all the required consumer counts

 RETURNS:  Integer status code of SUCCESS or ERROR
**************************************************************************/
int dropped_frame_characterization_preprocess
(
    BLOB *blob          /* I: Various parameters */
)
{
    int i;
    int status;

    number_of_memory_buffers = sizeof(memory_buffers) /
        sizeof(MEMORY_BUFFER_TYPE);

    /* Increment the consumer counts for the memory buffers */
    for ( i = 0; i < number_of_memory_buffers; i++ )
    {
        status = increment_consumer_count(memory_buffers[i]);

        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Incrementing %s consumer count",
                get_memory_buffer_string(memory_buffers[i]));
            return ERROR;
        }
    }

    return SUCCESS;
}  /* END dropped_frame_characterization_preprocess */
