/*----------------------------------------------------------------------------
NAME:    api_dump_imagery

PURPOSE: API for the write_l1r_image_data routine that will write an SCA of
         data to a file in the intermediate_dumps directory.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>

#include "ias_l1r.h"
#include "ias_logging.h"
#include "ias_types.h"
#include "rps_local.h"

static int num_dump_algorithms;
static L1RIO **l1r_file_pointers;
/* This is a 2-d array of L1R band structures.  There is one per algorithm
   (first dimension) with DUMP modifier and output band (second dimension).
   The band order is the same as the blob->list_of_bands array.  That
   means 2nd dimension of this array isn't accessed using a band_index:
   band numbers and indices must be obtained from the blob lists. */
static L1R_BAND_IO ***l1r_bands;
static int number_allocated;


/*----------------------------------------------------------------------------
NAME:    api_dump_imagery

PURPOSE: Retrieve the appropriate memory buffer and call the
         write_l1r_image_data routine

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int api_dump_imagery
(
    BLOB *blob,           /* I: Miscellaneous parameters */
    int   band_index,     /* I: */
    int   sca_index,      /* I: */
    int   algo_index_parm /* I: Algorithm index */
)
{
    int status;            /* Routine return status */
    float *l1r_image_data; /* Image data buffer */
    int algo_index;        /* Algorithm index */
    int dump_index;        /* Unmapped dump algorithm index */
    int index;
    RPS_BAND_ATTRIBUTES *band_attributes = get_band_attributes( band_index );
    int band_number = band_attributes->band_number;

    /* Retrieve the image buffer */
    l1r_image_data = (float *)get_memory_buffer( IMAGE_L1R, CONSUMER,
        band_index, sca_index );
    if ( !l1r_image_data )
    {
        IAS_LOG_ERROR("Retrieving L1R image buffer");
        return ERROR;
    }

    /* Unmap the algorithm index.  Since the dump algorithm list only has
       entries for the dump algorithms, the current algorithm must be mapped
       to the appropriate one. */
    dump_index = -1;
    for ( algo_index = 0; algo_index < num_dump_algorithms; algo_index++ )
    {
        if ( blob->dump_algorithm_mapping[algo_index] == algo_index_parm )
        {
            dump_index = algo_index;
            break;
        }
    }

    /* This shouldn't happen but, if it does, it's an error */
    if ( dump_index < 0 || algo_index == num_dump_algorithms )
    {
        IAS_LOG_ERROR("Couldn't map algorithm index back to dump index");
        return ERROR;
    }

    /* Find this band's index into the processing band list, this isn't the
       index associated with the band number */
    for ( index = 0; index < blob->number_of_bands; index++ )
    {
        if ( blob->list_of_bands[index] == band_number )
            break;
    }

    status = write_l1r_image_data( l1r_bands[dump_index][index],
        band_index, sca_index, l1r_image_data );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Writing L1R image buffer for: %s",
            blob->dump_algorithm_string_list[dump_index] );
        /* Don't return ERROR here because we need to decrement the 
           consumer count. */
    }

    status = decrement_consumer_count( IMAGE_L1R, band_index, sca_index );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Decrementing memory buffer");
        return ERROR;
    }
    if ( band_attributes->vrp_band_number > 0 )
    {
        status = decrement_consumer_count( IMAGE_L1R,
            band_attributes->vrp_band_index, sca_index );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }
    }
    if ( band_attributes->blind_band_number > 0
        && band_attributes->process_blind_band_flag )
    {
        status = decrement_consumer_count( IMAGE_L1R,
            band_attributes->blind_band_index, sca_index );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }
    }
    return status;
}  /* END api_dump_imagery */



/*----------------------------------------------------------------------------
NAME:    dump_preprocess

PURPOSE: Registers the memory buffers needed by the algorithm; allocates
         memory for the output band images; opens output images for all
         algorithms that have been marked for an intermediate dump.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int dump_preprocess
(
    BLOB *blob      /* I: Miscellaneous parameters */
)
{
    int algo_index;          /* Algorithm index */
    int band_index;          /* Band index used in for loop */
    int band_number;         /* 1-based band number */
    int index;
    char file_name[PATH_MAX];/* Name of the output file */
    char l1r_path[PATH_MAX]; /* Path to the output directory (working dir) */
    char dir_name[PATH_MAX]; /* sub-directory for dump files */
    char *p_char;            /* character pointer used for directory parsing */
    struct stat stat_buf;    /* Structure used for determining existence of 
                                sub-directory */
    int open_error = 0;      /* Flag for file open errors */
    int alloc_error = 0;     /* Flag for allocation errors */
    int status;              /* Routine return status */
    IAS_DATA_TYPE data_type = IAS_R4;   /* Data type of the output image */
    const IAS_SATELLITE_ATTRIBUTES *sat_attributes;

    num_dump_algorithms = blob->number_dump_algorithms;

    /* Set the memory management layer usage */
    for ( algo_index = 0; algo_index < num_dump_algorithms; algo_index++ )
    {
        status = increment_consumer_count( IMAGE_L1R );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Incrementing consumer count");
            return ERROR;
        }
    }

    /* Get the path to the L1R output file.  If no path was specified use
       the current directory.  Assuming that if there is a / in the file
       name that it is a full path. */
    strcpy( l1r_path,  blob->l1r_file_name );
    p_char = strrchr( l1r_path, '/' );
    if ( p_char )
        p_char[0] = '\0';
    if ( !p_char || l1r_path[0] == '\0' )
    {
        l1r_path[0] = '.';
        l1r_path[1] = '\0';
    }
    sprintf( dir_name, "%s/%s", l1r_path, blob->dump_directory );

    /* Create the directory that the imagery will be dumped unless it
       already exists.  If it exists the most likely case is that the work
       order has already been run.  If so, the image files will simply be
       overwritten. */
    if ( stat( dir_name, &stat_buf ))
    {
        int  the_error = errno;
        char *error_string;

        /* Check errno.  ENOENT ``No such file or directory'' so we will
           assume the directory we want to create -- doesn't exist. */
        if ( the_error == ENOENT )
        {
            if ( mkdir( dir_name, 0777 ))
            {
                IAS_LOG_ERROR("Creating intermediate dump directory");
                return ERROR;
            }
        }
        else
        {
            /* Something else happened */
            error_string = strerror( the_error );
            IAS_LOG_ERROR("Creating %s", dir_name );
            IAS_LOG_ERROR("%s", error_string);
            return ERROR;
        }
    }


    /* Allocate memory for the output file and band pointers */
    sat_attributes = ias_sat_attr_get_attributes();
    number_allocated = sat_attributes->total_bands;
    l1r_file_pointers = calloc( num_dump_algorithms, sizeof(L1RIO *));
    if ( !l1r_file_pointers )
    {
        IAS_LOG_ERROR("Allocating L1R pointers");
        return ERROR;
    }
    l1r_bands = calloc( num_dump_algorithms, sizeof(L1R_BAND_IO **));
    if ( !l1r_bands )
    {
        IAS_LOG_ERROR("Allocating L1R band pointers");
        return ERROR;
    }
    for ( algo_index = 0; algo_index < num_dump_algorithms; algo_index++ )
    {
        l1r_bands[algo_index] = calloc( number_allocated,
            sizeof(L1R_BAND_IO *));
        if ( !l1r_bands[algo_index] )
        {
            alloc_error = 1;
        }
    }
    if ( alloc_error )
    {
        IAS_LOG_ERROR("Allocating L1R_BAND_IO pointers");
        for ( index = 0; index < number_allocated; index++ )
            free( l1r_bands[index] );
        free( l1r_file_pointers );
        return ERROR;
    }

    /* Open the output files */
    for ( algo_index = 0; algo_index < num_dump_algorithms; algo_index++ )
    {
        sprintf( file_name, "%s/%s_%02d.h5", dir_name,
            blob->dump_algorithm_string_list[algo_index], algo_index );

        l1r_file_pointers[algo_index] = ias_l1r_open_image( file_name,
            IAS_WRITE );
        if ( !l1r_file_pointers[algo_index] )
        {
            IAS_LOG_ERROR("Opening output file: %s", file_name );
            open_error = 1;
        }

        for ( index = 0; index < blob->number_of_bands; index++ )
        {
            band_index = blob->list_of_band_indices[index];
            RPS_BAND_ATTRIBUTES *band_attributes = get_band_attributes(
                band_index );
            /* Since we are writing to a file the returned values aren't set
               use temporary variables. */
            int lines;
            int samples;
            int number_of_scas;

            /* Get the 1-based band number and band information */
            band_number = band_attributes->band_number;
            lines = band_attributes->number_lines;
            samples = band_attributes->number_detectors;
            number_of_scas = band_attributes->number_scas;
            /* Open it */
            l1r_bands[algo_index][index] = ias_l1r_open_band(
                l1r_file_pointers[algo_index], band_number, &data_type,
                &number_of_scas, &lines, &samples );
        }
    }

    /* If there was an error opening one of the output, close any successful
       file opens. */
    if ( open_error )
    {
        for ( algo_index = 0; algo_index < num_dump_algorithms; algo_index++ )
        {
            if ( l1r_file_pointers[algo_index] )
            {
                ias_l1r_close_image( l1r_file_pointers[algo_index] );
                l1r_file_pointers[algo_index] = NULL;
            }
        }
        free( l1r_file_pointers );
        l1r_file_pointers = NULL;
        return ERROR;
    }

    return SUCCESS;
}  /* END dump_preprocess */


/*----------------------------------------------------------------------------
NAME:    dump_postprocess

PURPOSE: Close all the intermediate dump output images and free memory
         allocated in the pre-process routine

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int dump_postprocess
(
    BLOB *blob           /* I: Miscellaneous parameters */
)
{
    int index;           /* Index into the L1R bands structure */
    int algo_index;      /* Algorithm index */
    int status;          /* Return status */
    int close_error = 0; /* Flag to capture errors */

    /* Close the output images */
    for ( algo_index = 0; algo_index < num_dump_algorithms; algo_index++ )
    {
        /* Set the L1R metadata for the current intermediate output being
           dumped */
        status = set_l1r_metadata( blob, l1r_file_pointers[algo_index] );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("L1R metadata not properly set");
            close_error = 1;
        }


        /* Close the bands */
        for ( index = 0; index < number_allocated; index++ )
        {
            if ( l1r_bands[algo_index][index] != NULL )
            {
                status = ias_l1r_close_band(
                    l1r_bands[algo_index][index] );
                if ( status != SUCCESS )
                {
                    close_error = 1;
                    IAS_LOG_ERROR("Closing %s band %d",
                        blob->dump_algorithm_string_list[algo_index],
                        blob->list_of_bands[index] );
                }
            }
        }

        /* Close the file */
        status = ias_l1r_close_image( l1r_file_pointers[algo_index] );
        if ( status != SUCCESS )
        {
            close_error = 1;
            IAS_LOG_ERROR("Closing %s",
                blob->dump_algorithm_string_list[algo_index] );
        }
    }

    /* Free the allocated structures */
    for ( algo_index = 0; algo_index < num_dump_algorithms; algo_index++ )
        free( l1r_bands[algo_index] );
    free( l1r_bands );
    free( l1r_file_pointers );

    if ( close_error )
        return ERROR;
    return SUCCESS;
}  /* END dump_postprocess */
