/*----------------------------------------------------------------------------
NAME:    api_gain_application

PURPOSE: Gather data needed for the gain application routine and call it.  

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
#include <stdlib.h>
#include <string.h>
#include "ias_logging.h"
#include "rps_local.h"

int api_gain_application
(
    BLOB *blob,         /* I: Miscellaneous parameters */
    int   band_index,   /* I: */
    int   sca_index,    /* I: */
    int   algo_index    /* I: Algorithm index */
)
{
    float *data_buffer;                   /* Level 1 data buffer */
    RPS_BAND_ATTRIBUTES *band_attributes = get_band_attributes( band_index );
    int band_number = band_attributes->band_number;
    int num_lines = band_attributes->number_lines;
    int num_samps = band_attributes->number_detectors;
    const double *absolute_gain = NULL;   /* SCA level absolute gain */
    const double dummy[1] = {0.0};        /* Dummy array for use when no
                                             absolute gain is being applied */
    double *relative_gain = NULL;         /* detector level relative gain */
    const double *temp_correction_factors = NULL;/* detector level temperature 
                                             sensitivity correction factors */
    int perform_abs_gain;                 /* Flag for the absolute gain 
                                             correction*/
    int perform_relative_gain;            /* Flag for the relative gain 
                                             correction*/
    int perform_temp_sens;                /* Flag for the temperature 
                                             sensitivity correction */
    int status;

    /* Add the ID for temperature sensitivity correction for gain application 
       to the RPS processing level if blob->perform_temp_sens is 1 */
    if (blob->perform_temp_sens)
    {
        if (set_rps_level(band_index, sca_index, 
            IAS_RPS_LEVEL_TEMP_SENSITIVITY_GAIN) == ERROR)
        {
            IAS_LOG_ERROR("Updating the RPS processing level");
            return ERROR;
        }
    }

    if (set_rps_level(band_index, sca_index, 
            IAS_RPS_LEVEL_GAIN_APPLICATION) == ERROR)
    {
        IAS_LOG_ERROR("Updating the RPS processing level");
        return ERROR;
    }

    /* This algorithm doesn't process OLI blind or VRP bands therefore the
        processing level *must* be set to skip */
    if ( band_attributes->vrp_band_index >= 0 )
    {
        if (set_rps_level(band_attributes->vrp_band_index, sca_index,
            IAS_RPS_LEVEL_SKIP) == ERROR)
        {
            IAS_LOG_ERROR("Updating RPS processing level");
            return ERROR;
        }
    }
    if ( band_attributes->blind_band_index >= 0 &&
         band_attributes->process_blind_band_flag )
    {
        if (set_rps_level(band_attributes->blind_band_index, sca_index,
            IAS_RPS_LEVEL_SKIP) == ERROR)
        {
            IAS_LOG_ERROR("Updating RPS processing level");
            return ERROR;
        }
    }

    perform_abs_gain = blob->apply_absolute_gain;
    perform_relative_gain = blob->apply_relative_gain;
    perform_temp_sens = blob->perform_temp_sens;

    /* Get the data buffer */
    data_buffer = (float *)get_memory_buffer(IMAGE_L1R, CONSUMER, band_index,
        sca_index);
    if (!data_buffer)
    {
        IAS_LOG_ERROR("Getting image memory buffer");
        return ERROR;
    }

    /* Get the temperature sensitivity coeffs buffer */
    if (perform_temp_sens)
    {
        temp_correction_factors = (const double *)get_memory_buffer(
            MM_TEMP_CORRECTION_FACTOR, CONSUMER, band_index, sca_index);
        if (!temp_correction_factors)
        {
            IAS_LOG_ERROR("Getting temperature sensitivity correction "
                "factors buffer");
            return ERROR;
        }
    }

    /* Point the absolute gain variable, which is dereferenced in 
       the call to the algorithm, to the dummy array. */
    if (perform_abs_gain == 0)
    {
        absolute_gain = dummy;
    }

    /* Get the absolute gain */
    if (perform_abs_gain)
    {
        absolute_gain = (const double *)get_const_memory_buffer(
            MM_ABSOLUTE_GAINS, band_index, sca_index);
        if (!absolute_gain)
        {
            IAS_LOG_ERROR("Getting absolute gains buffer");
            return ERROR;
        }
    }

    /* Get the relative gains */
    if (perform_relative_gain)
    {
        relative_gain = (double *)get_const_memory_buffer(MM_RELATIVE_GAINS, 
                                                    band_index, sca_index);
        if (!relative_gain)
        {
            IAS_LOG_ERROR("Getting relative gains buffer");
            return ERROR;
        }
    }

    /* Make sure the temperature sensitivity correction factor buffer is valid 
       if perform_temp_sens is 1*/
    if ( perform_temp_sens == 1)
    {
        if ( get_buffer_filled_flag( MM_TEMP_CORRECTION_FACTOR,
             band_index, sca_index ) == 0 )
        {
            IAS_LOG_ERROR("Attempting to use uninitialized buffer %s for band "
                "%d SCA %d", get_memory_buffer_string(
                MM_TEMP_CORRECTION_FACTOR ), band_number, sca_index + 1);
            return ERROR;
        }
    }

    /* Note: It isn't necessary to decrement the usage of the memory buffer
       on error conditions from this point on.  Since the memory may not be
       freed.  The main thread will clean up allocated buffers when error
       conditions dictate.  It doesn't hurt but it isn't necessary. */

    gain_application(data_buffer, band_number, sca_index + 1, num_lines,
        num_samps, absolute_gain[0], relative_gain, temp_correction_factors,
        perform_abs_gain, perform_relative_gain, perform_temp_sens);

    /* Set the band metadata */
    if ( sca_index == 0 && perform_temp_sens )
    {
        blob->l1r_band_metadata[band_index]->
            gain_temp_sensitivity_correction_applied = 1;
    }

    /* Tell memory management that the data was used */
    status = decrement_consumer_count(IMAGE_L1R, band_index, sca_index);
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Decrementing memory buffer");
        return ERROR;
    }
    if ( band_attributes->vrp_band_number > 0 )
    {
        status = decrement_consumer_count( IMAGE_L1R,
            band_attributes->vrp_band_index, sca_index );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }
    }
    if ( band_attributes->blind_band_number > 0 
        && band_attributes->process_blind_band_flag )
    {
        status = decrement_consumer_count( IMAGE_L1R,
            band_attributes->blind_band_index, sca_index );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }
    }
    status = decrement_consumer_count(MM_TEMP_CORRECTION_FACTOR, band_index,
        sca_index);
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Decrementing memory buffer");
        return ERROR;
    }
    status = decrement_consumer_count(MM_ABSOLUTE_GAINS, band_index, sca_index);
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Decrementing memory buffer");
        return ERROR;
    }
    status = decrement_consumer_count(MM_RELATIVE_GAINS, band_index, sca_index);
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Decrementing memory buffer");
        return ERROR;
    }

    return SUCCESS;
}  /* END api_gain_application */


/*----------------------------------------------------------------------------
NAME:    gain_application_preprocess

PURPOSE: Do preprocessing for the algorithm by incrementing the consumer
         counts.  

RETURNS: SUCCESS or ERROR

-----------------------------------------------------------------------------*/
int gain_application_preprocess
(
    BLOB *blob          /* I: Miscellaneous parameters */
)
{
    /* Define a list of memory buffers needed by the algorithm */
    MEMORY_BUFFER_TYPE memory_buffers[] =
    {
        IMAGE_L1R,
        MM_TEMP_CORRECTION_FACTOR,
        MM_ABSOLUTE_GAINS,
        MM_RELATIVE_GAINS
    };
    int number_of_memory_buffers = sizeof(memory_buffers) /
        sizeof(MEMORY_BUFFER_TYPE);
    int i;
    int status;

    /* Increment the consumer counts for the memory buffers */
    for ( i = 0; i < number_of_memory_buffers; i++ )
    {
        status = increment_consumer_count( memory_buffers[i] );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Incrementing %s consumer count",
                get_memory_buffer_string( memory_buffers[i] ));
            return ERROR;
        }
    }

    /* Set L1R file metadata */
    if ( blob->apply_absolute_gain )
        blob->l1r_file_metadata.band_average_gain_applied = 1;
    if ( blob->apply_relative_gain )
        blob->l1r_file_metadata.relative_gain_applied = 1;

    return SUCCESS;
}  /* END gain_application_preprocess */
