/*----------------------------------------------------------------------------
NAME:    histogram_statistics_characterization

PURPOSE: Gather data needed for the histogram statistics characterization
         routine and call it.  If database storage is desired, submit the
         results to the routine to store them.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
#include <stdlib.h>
#include <string.h>

#include "ias_logging.h"
#include "ias_threadpool.h"
#include "ias_radiometry.h"
#include "rps_local.h"
#include "ias_math.h"
#include "ias_satellite_attributes.h"
#include "ias_db_insert.h"

/* Define a list of memory buffers needed by the algorithm */
static MEMORY_BUFFER_TYPE memory_buffers[] =
{
    IMAGE_L1R,
    MM_HIST_FRAMES_TO_SKIP_TOP,
    MM_HIST_FRAMES_TO_SKIP_BOTTOM,
    MM_L0R_INTERMEDIATE_STATISTICS,
    MM_L1R_INTERMEDIATE_STATISTICS,
    MM_SPATIAL_DETECTOR_OFFSETS
};
static int number_of_memory_buffers = sizeof(memory_buffers) /
    sizeof(MEMORY_BUFFER_TYPE);

/*----------------------------------------------------------------------------
 NAME:    decrement_counters

 PURPOSE: Internal routine to decrement the consumer counts of all required
          memory buffers

 RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
static int decrement_counters
(
    int band_index,     /* I: */
    int sca_index       /* I: */
)
{
    int i;                                  /* Local loop counter */

    /* Decrement the consumer counts for the memory buffers */
    for (i = 0; i < number_of_memory_buffers; i++)
    {
        int status = decrement_consumer_count(memory_buffers[i], band_index,
            sca_index);
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }
    }
    return SUCCESS;
}   /* END decrement_counters */

/*----------------------------------------------------------------------------
 NAME:    free_report_information

 PURPOSE: Internal routine to free the local copy of report_information 
          structure

 RETURNS: Nothing
-----------------------------------------------------------------------------*/
static void free_report_information
(
    IAS_REPORT *report_information
)
{
    if ( report_information )
    {
        if ( report_information->header )
            ias_misc_free_report_header( report_information->header );
        if ( report_information->directory )
            free( report_information->directory );
        if ( report_information->rps_level )
            free( report_information->rps_level );
        if ( report_information->rlut_version )
            free( report_information->rlut_version );
        free( report_information );
    }
}   /* END free_report_information */

/*----------------------------------------------------------------------------
NAME:    histogram_statistics_characterization

PURPOSE: Internal routine to do the bulk of the API work.  Called from the
         externally available OLI and TIRS APIs

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
static int histogram_statistics_characterization
(
    BLOB *blob,         /* I: Miscellaneous parameters */
    int   band_index,   /* I: */
    int   sca_index     /* I: */
)
{
    float *data_buffer;    /* Level 1 data buffer */
    RPS_BAND_ATTRIBUTES *band_attributes = get_band_attributes( band_index );
    int band_number = band_attributes->band_number;
    int num_lines = band_attributes->number_lines;
    int num_samps = band_attributes->number_detectors;
    const IAS_BAND_ATTRIBUTES *ias_band_attributes =
        ias_sat_attr_get_band_attributes( band_number );
    int pan_band = 0;
    IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS *db_results = NULL;
    IAS_RPS_HISTOGRAM_INTERMEDIATE_RESULTS *intermediate_results = NULL;
    IAS_REPORT *copy_of_report_information = NULL;
    const int *cpf_frames_to_skip_top = NULL;
    const int *cpf_frames_to_skip_bottom = NULL;
    char *local_rps_level = NULL;
    int *detector_offsets;
    int det;
    int status;
    int level0_stats = 0;

    /* Get the CPF number of frames to skip at the top and/or bottom of the
       image */
    cpf_frames_to_skip_top = (const int *)get_const_memory_buffer(
        MM_HIST_FRAMES_TO_SKIP_TOP, band_index, sca_index);
    cpf_frames_to_skip_bottom = (const int *)get_const_memory_buffer(
        MM_HIST_FRAMES_TO_SKIP_BOTTOM, band_index, sca_index);
    if ((cpf_frames_to_skip_top == NULL)
        || (cpf_frames_to_skip_bottom == NULL))
    {
        IAS_LOG_ERROR("Retrieving data buffers for CPF number of frames to "
            "skip at image top and/or bottom for band %d SCA %d",
            band_number, (sca_index + 1));
        return ERROR;
    }
    
    /* Get the number of lines per frame of image data for current band */
    int lines_per_frame = ias_sat_attr_get_lines_per_frame(band_number);

    /* Determine the number of lines to skip at the top and bottom of the
       image based on the CPF frames-to-skip values.  In general, these will
       be the values to use for TIRS and OLI non-lamp collects */
    int lines_to_skip_top    = *cpf_frames_to_skip_top * lines_per_frame;
    int lines_to_skip_bottom = *cpf_frames_to_skip_bottom * lines_per_frame;

    /* Getting the rps level */
    local_rps_level = get_rps_level( band_index, sca_index );
    if (local_rps_level == NULL)
    {
        IAS_LOG_ERROR("Getting RPS level for Band %d SCA %d", band_number,
            sca_index + 1 );
        return ERROR;
    }

    /* Check to see the processing band is a TIRS band or OLI band */
    if (band_attributes->spectral_type == IAS_SPECTRAL_THERMAL)
    {
        /* Need blind band statistics only if processing the blind band
           for TIRS dark response processing */
        if (blob->process_tirs_blind_band)
        {
            if (rps_level_is_l0r(local_rps_level))
            {
                intermediate_results =
                    (IAS_RPS_HISTOGRAM_INTERMEDIATE_RESULTS *)
                    get_memory_buffer(MM_L0R_INTERMEDIATE_STATISTICS,
                    SETTER, band_index, sca_index);
                if (intermediate_results == NULL)
                {
                    IAS_LOG_ERROR("Getting intermediate results for L0R");
                    free( local_rps_level );
                    return ERROR;
                }
            }
            else
            {
                IAS_LOG_ERROR("Unexpected RPS processing level %s", 
                    local_rps_level);
                free( local_rps_level );
                return ERROR;
            }
        }
    }
    else
    {
        /* For OLI bands, check to see if the lamp data are to be processed and
           therefore the statistics intermediate results are needed */
        if ( blob->lamp_processing && band_attributes->is_normal_band )
        {
            /* Due to the way memory management works, data buffers like the
               intermediate results aren't allocated until they are retrieved
               the first time, this includes setting the consumer count.
               Therefore both buffers need to be retrieved so that the proper
               decrements of consumer counts can be made.  Otherwise the
               memory will be leaked until the final memory management clean
               up. */
            IAS_RPS_HISTOGRAM_INTERMEDIATE_RESULTS *l0rbuf =
                (IAS_RPS_HISTOGRAM_INTERMEDIATE_RESULTS *)get_memory_buffer(
                MM_L0R_INTERMEDIATE_STATISTICS, SETTER, band_index,
                sca_index );
            if ( !l0rbuf )
            {
                IAS_LOG_ERROR("Getting intermediate_results for L0R");
                free( local_rps_level );
                return ERROR;
            }
            IAS_RPS_HISTOGRAM_INTERMEDIATE_RESULTS *l1rbuf = 
                (IAS_RPS_HISTOGRAM_INTERMEDIATE_RESULTS *)get_memory_buffer(
                MM_L1R_INTERMEDIATE_STATISTICS, SETTER, band_index,
                sca_index );
            if ( !l1rbuf )
            {
                IAS_LOG_ERROR("Getting intermediate_results for L1R");
                free( local_rps_level );
                return ERROR;
            }
            /* The lamp algorithm set up the starting line and number of
               lines to process so this algorithm should process the entire
               image that's in memory.  Override the lines to skip at top and
               bottom. */
            lines_to_skip_top = 0;
            lines_to_skip_bottom = 0;

            /* The lamps process uses the L0rc (bias-subtracted and 
               linearized DNs), and L1R (Radiance Units, after gain 
               application). Therefore, "G" represent gain application is
               used for distinguish rps_levels for L0R and L1R */
            char *end_rps_level;
            end_rps_level = strrchr(local_rps_level, 'G');
            if (end_rps_level != NULL)
            {
                /* If the L1R buffer has already been filled, issue a warning 
                   and let the workflow continue in case cal/val team wants to 
                   adjust the location of histogram statistics algorithm in the
                   workflow */  
                if ( get_buffer_filled_flag( MM_L1R_INTERMEDIATE_STATISTICS,
                     band_index, sca_index ) == 0 )
                {
                    intermediate_results = l1rbuf;
                    level0_stats = 0;
                }
                else 
                {
                    IAS_LOG_WARNING("The buffer %s has already been filled for "
                                " band %d, SCA %d", get_memory_buffer_string(
                                MM_L1R_INTERMEDIATE_STATISTICS),
                                band_number, sca_index + 1);
                    intermediate_results = NULL;
                }
            }
            else
            {
                /* If the L0R buffer has already been filled, issue a warning 
                   and let the workflow continue (the database insertion will
                   be done) in case cal/val team wants to adjust the location 
                   of histogram statistics algorithm in the workflow */  
                if ( get_buffer_filled_flag( MM_L0R_INTERMEDIATE_STATISTICS,
                     band_index, sca_index ) == 0 )
                {
                    intermediate_results = l0rbuf;
                    level0_stats = 1;
                }
                else 
                {
                    IAS_LOG_WARNING("The level 0 report statistics for buffer "
                        "%s band %d SCA %d have alradey been computed. The "
                        "statistics for RPS level %s will only be inserted "
                        "into database", get_memory_buffer_string(
                        MM_L0R_INTERMEDIATE_STATISTICS), band_number, 
                                    sca_index + 1, local_rps_level);
                    intermediate_results = NULL;
                }
            }
        }
    }

    /* Set the reporting and database structures */
    if ( ias_band_attributes->band_type == IAS_PAN_BAND )
        pan_band = 1;
    if ( blob->generate_report )
    {
        /* Allocate memory for copy_of_report_information */
        copy_of_report_information = malloc(sizeof(IAS_REPORT));
        if ( !copy_of_report_information )
        {
            IAS_LOG_ERROR( "Allocating memory for copy_of_report_information");
            free( local_rps_level );
            return ERROR;
        }

        /* Initialize the standard report header */
        copy_of_report_information->header = ias_misc_copy_report_header(
            blob->report_information.header);
        if ( !copy_of_report_information->header )
        {
            IAS_LOG_ERROR( "Creating local copy of report header" );
            free_report_information( copy_of_report_information );
            free( local_rps_level );
            return ERROR;
        }

        copy_of_report_information->directory = 
            strdup(blob->report_information.directory);
        if ( !copy_of_report_information->directory )
        {
            IAS_LOG_ERROR( "Duplicating string for "
                           "blob->report_information.directory" );
            free_report_information( copy_of_report_information );
            free( local_rps_level );
            return ERROR;
        }

        copy_of_report_information->rps_level = strdup(local_rps_level);
        if ( !copy_of_report_information->rps_level )
        {
            IAS_LOG_ERROR( "Duplicating string for local_rps_level" );
            free_report_information( copy_of_report_information );
            free( local_rps_level );
            return ERROR;
        }

        if ( !blob->report_information.rlut_version )
        {
            copy_of_report_information->rlut_version = strdup("N/A");
        }
        else
        {
            copy_of_report_information->rlut_version = 
                strdup(blob->report_information.rlut_version);
        }
        if ( !copy_of_report_information->rlut_version )
        {
            IAS_LOG_ERROR( "Duplicating string for "
                           "blob->report_information.rlut_version" );
            free_report_information( copy_of_report_information );
            free( local_rps_level );
            return ERROR;
        }

        copy_of_report_information->oli_frame_rate = 
            blob->report_information.oli_frame_rate;
        copy_of_report_information->tirs_frame_rate = 
            blob->report_information.tirs_frame_rate;
    }

    if ( blob->store_in_database )
    {
        db_results = ias_rps_db_allocate_histogram_results( num_samps,
            pan_band );
        if ( !db_results )
        {
            IAS_LOG_ERROR("Allocating memory for the database results");
            free_report_information( copy_of_report_information );
            free( local_rps_level );
            return ERROR;
        }
        
        /* Set the common values that are known here */
        db_results->dbase = blob->database_information.db;
        strncpy(db_results->rad_common_id, 
            blob->database_information.rad_common_id, 
            sizeof(db_results->rad_common_id));
        db_results->rad_common_id[sizeof(db_results->rad_common_id) - 1] = '\0';
        db_results->band_number = band_number;
        db_results->sca = sca_index + 1;
        db_results->num_detectors = num_samps;
        db_results->is_odd_even_data_split = pan_band;
        db_results->processing_level = strdup(local_rps_level);
        if(!db_results->processing_level)
        {
            IAS_LOG_ERROR("Duplicating string for "
                "db_results->processing_level");
            free_report_information(copy_of_report_information);
            free(local_rps_level);
            ias_rps_db_free_histogram_results(db_results);
            return ERROR;
        }

        strncpy(db_results->collection_type, 
               blob->l1r_file_metadata.collect_type, 
               sizeof(db_results->collection_type));
        db_results->collection_type[sizeof(db_results->collection_type) - 1]
            = '\0';
    } /* END if store in database block */

    /* Populate the processing_level for the intermediate_results 
       structure */
    if (blob->lamp_processing && band_attributes->is_normal_band &&
        intermediate_results != NULL)
    {
        for (det = 0; det < num_samps; det++)
        {
            status = snprintf( &intermediate_results[det].processing_level[0],
                IAS_RPS_PROCESSING_LEVEL_LENGTH, "%s", local_rps_level );
            if ( status >= IAS_RPS_PROCESSING_LEVEL_LENGTH || status < 0 )
            {
                IAS_LOG_ERROR("Copying processing level to intermediate "
                      "results structure");
                ias_rps_db_free_histogram_results( db_results );
                free_report_information( copy_of_report_information );
                free( local_rps_level );
                return ERROR;
            }
        }
    }

    /* Note: It isn't necessary to decrement the usage of the memory buffer
       on error conditions from this point on.  Since the memory may not be
       freed.  The main thread will clean up allocated buffers when error
       conditions dictate.  It doesn't hurt but it isn't necessary. */
    /* Get the data buffer */
    data_buffer = (float *)get_memory_buffer( IMAGE_L1R, CONSUMER, band_index,
        sca_index );
    if ( !data_buffer )
    {
        IAS_LOG_ERROR("Getting image memory buffer");
        ias_rps_db_free_histogram_results( db_results );
        free_report_information( copy_of_report_information );
        free( local_rps_level );
        return ERROR;
    }

    /* Get the detector offsets */
    detector_offsets = (int *)get_memory_buffer(
        MM_SPATIAL_DETECTOR_OFFSETS, CONSUMER, band_index, sca_index );
    if ( !detector_offsets )
    {
        IAS_LOG_ERROR("Getting %s for band %d SCA %d",
            get_memory_buffer_string(MM_SPATIAL_DETECTOR_OFFSETS),
            band_number, sca_index + 1);
        ias_rps_db_free_histogram_results( db_results );
        free_report_information( copy_of_report_information );
        free( local_rps_level );
        return ERROR;
    }

    status = ias_rps_histogram_statistics_characterization( data_buffer,
        IAS_R4, band_number, sca_index + 1, num_samps, num_lines,
        detector_offsets, lines_to_skip_top, lines_to_skip_bottom,
        blob->pixel_mask[band_index][sca_index],
        copy_of_report_information, db_results, intermediate_results );
    
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("From ias_rps_histogram_statistics_characterization"
            " band: %d, SCA %d", band_number, sca_index + 1 );
        ias_rps_db_free_histogram_results( db_results );
        free_report_information( copy_of_report_information );
        free( local_rps_level );
        return ERROR;
    }

    /* Calculate and store the scene summary statistics */
    if (blob->populate_scene_summary_stats_flag &&
        ias_band_attributes->band_classification == IAS_NORMAL_BAND)
    {
        int offset; /* Loop offset variable */

        /* Make sure list has been initialized */
        if (!blob->scene_statistics_list) 
        {
            IAS_LOG_ERROR("Scene statistics list not initialized ");
            ias_rps_db_free_histogram_results(db_results);
            free_report_information(copy_of_report_information);
            free(local_rps_level);
            return ERROR;
        }

        /* Execute stats collection then database insert on single thread */
        for (offset = 0; offset <= db_results->is_odd_even_data_split; offset++)
        {
            DB_SCENE_SUMMARY_STATISTICS *node; /* Current node */
            
            /* Allocated memory for node */
            node = malloc(sizeof(DB_SCENE_SUMMARY_STATISTICS));
            if (!node)
            {
                IAS_LOG_ERROR("Error allocating scene summary stats node");
                return ERROR;
            } 

            /* Copy information to node from db_results */
            node->band_number = db_results->band_number;
            node->sca_number = db_results->sca;

            /* Determine the pan split value for storing in the database */
            node->lineset = IAS_RPS_NORMAL_BAND_LINESET;
            if (db_results->is_odd_even_data_split)
            {
                node->lineset = offset;
            }
            
            /* Calculate the SCA summary stats */
            status = ias_math_merge_mean_and_stdev_samples(
                db_results->num_values[offset], db_results->mean[offset], 
                db_results->stdev[offset], db_results->num_detectors, 
                &node->stdev, &node->mean, &node->frames);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Calculating SCA summary stats for Band %d"
                    " SCA %d", node->band_number, node->sca_number);
                free(node);
                return ERROR;
            }

            /* Store the SCA summary stats into the database */
            status = ias_threadpool_add_to_work_queue(blob->threadpool,
                store_histogram_statistics_per_sca_to_db, node);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Could not store statistics for band %d SCA %d",
                    band_number, node->sca_number);
                /* Before freeing memory, ensure there is more than one thread
                   running.  A double free of the memory may occur if only one
                   thread is running and an internal database error occurs. */
                if ( ias_threadpool_get_thread_count( blob->threadpool ) > 1 )
                {
                    free(node);
                }
                return ERROR;
            }
        }   
    }
   
    /* Free the allocated memory in the local report_information structure */
    free_report_information( copy_of_report_information );

    if ( blob->store_in_database )
    {
        status = ias_threadpool_add_to_work_queue( blob->threadpool,
            ias_rps_store_histogram_stats_in_database, db_results );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Submitting histogram statistics to work queue "
                "band: %d, SCA %d", band_number, sca_index + 1 );
            /* Before freeing memory, ensure there is more than one thread
               running.  A double free of the memory may occur if only one
               thread is running and an internal database error occurs. */
            if ( ias_threadpool_get_thread_count( blob->threadpool ) > 1 )
                ias_rps_db_free_histogram_results( db_results );
            free( local_rps_level );
            return ERROR;
        }
    }

    /* Let the memory management know that the buffer is now valid.  Only
       set the one that was filled. */
    if ( intermediate_results )
    {
        if ( level0_stats )
        {
            set_buffer_filled_flag( MM_L0R_INTERMEDIATE_STATISTICS,
                band_index, sca_index );
        }
        else
        {
            set_buffer_filled_flag( MM_L1R_INTERMEDIATE_STATISTICS,
                band_index, sca_index );
        }
    }

    /* Tell memory management that the data was used */
    status = decrement_counters( band_index, sca_index );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Decrementing memory buffer");
        free( local_rps_level );
        return ERROR;
    }

    free( local_rps_level );

    return status;
}  /* END histogram_statistics_characterization */


/*--------------------------------------------------------------------------
NAME:    api_tirs_histogram_statistics_characterization

PURPOSE: Call the internal routine for TIRS bands

RETURNS: SUCCESS or ERROR
---------------------------------------------------------------------------*/
int api_tirs_histogram_statistics_characterization
(
    BLOB *blob,         /* I: Miscellaneous parameters */
    int   band_index,   /* I: */
    int   sca_index,    /* I: */
    int   algo_index    /* I: Algorithm index */
)
{
    RPS_BAND_ATTRIBUTES *band_attributes = get_band_attributes( band_index );
    int status;

    /* If called with other than a thermal band return */
    if ( band_attributes->spectral_type != IAS_SPECTRAL_THERMAL )
    {
        status = decrement_counters( band_index, sca_index );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }
        if ( band_attributes->vrp_band_number > 0 )
        {
            status = decrement_counters( band_attributes->vrp_band_index,
                sca_index );
            if ( status != SUCCESS )
            {
                IAS_LOG_ERROR("Decrementing memory buffer");
                return ERROR;
            }
        }
        if ( band_attributes->blind_band_number > 0 
            && band_attributes->process_blind_band_flag )
        {
            status = decrement_counters( band_attributes->blind_band_index,
                sca_index );
            if ( status != SUCCESS )
            {
                IAS_LOG_ERROR("Decrementing memory buffer");
                return ERROR;
            }
        }
        return SUCCESS;
    }

    /* We've got a TIRS band.  If we are doing TIRS dark response
       processing, skip the normal TIRS band. */
    if (!blob->process_tirs_blind_band)
    {
        status = histogram_statistics_characterization(blob, band_index,
            sca_index);
        if (status != SUCCESS)
            return ERROR;
    }
    else
    {
        status = decrement_counters(band_index, sca_index);
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }
    }

    /* Process the associated blind band if this band is is supposed to
       process it */
    if ( band_attributes->blind_band_number > 0
        && band_attributes->process_blind_band_flag )
    {
        status = histogram_statistics_characterization( blob,
            band_attributes->blind_band_index, sca_index );
        if ( status != SUCCESS )
            return ERROR;
    }

    return SUCCESS;
}


/*--------------------------------------------------------------------------
NAME:    api_oli_histogram_statistics_characterization

PURPOSE: Call the internal routine for OLI bands

RETURNS: SUCCESS or ERROR
---------------------------------------------------------------------------*/
int api_oli_histogram_statistics_characterization
(
    BLOB *blob,         /* I: Miscellaneous parameters */
    int   band_index,   /* I: */
    int   sca_index,    /* I: */
    int   algo_index    /* I: Algorithm index */
)
{
    RPS_BAND_ATTRIBUTES *band_attributes = get_band_attributes( band_index );
    int status;

    /* If called with a thermal band return */
    if ( band_attributes->spectral_type == IAS_SPECTRAL_THERMAL )
    {
        status = decrement_counters( band_index, sca_index );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }
        if ( band_attributes->blind_band_number > 0 
            && band_attributes->process_blind_band_flag )
        {
            status = decrement_counters( band_attributes->blind_band_index,
                sca_index );
            if ( status != SUCCESS )
            {
                IAS_LOG_ERROR("Decrementing memory buffer");
                return ERROR;
            }
        }
        return SUCCESS;
    }

    status = histogram_statistics_characterization( blob, band_index,
        sca_index );
    if ( status != SUCCESS )
        return ERROR;

    /* Process the VRP data if this band has a VRP associated with it */
    if ( band_attributes->vrp_band_number > 0 )
    {
        status = histogram_statistics_characterization( blob,
            band_attributes->vrp_band_index, sca_index );
        if ( status != SUCCESS )
            return ERROR;
    }

    /* Process the blind band associated with the image band but only if
       it's to be processed by this band */
    if ( band_attributes->blind_band_number > 0
        && band_attributes->process_blind_band_flag )
    {
        status = histogram_statistics_characterization( blob,
            band_attributes->blind_band_index, sca_index );
        if ( status != SUCCESS )
            return ERROR;
    }

    return SUCCESS;
}  /* END api_oli_histogram_statistics_characterization */


/*--------------------------------------------------------------------------
NAME:    histogram_statistics_characterization_preprocess

PURPOSE: Registers memory buffers for the algorithm

RETURNS: SUCCESS or ERROR
---------------------------------------------------------------------------*/
int histogram_statistics_characterization_preprocess
(
    BLOB *blob          /* I: Miscellaneous parameters */
)
{
    int i;
    int status;

    /* Increment the consumer counts for the memory buffers */
    for ( i = 0; i < number_of_memory_buffers; i++ )
    {
        status = increment_consumer_count( memory_buffers[i] );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Incrementing %s consumer count",
                get_memory_buffer_string( memory_buffers[i] ));
            return ERROR;
        }
    }

    /* Determine if Earth collection and if populate stats flag is set*/
    if (blob->populate_scene_summary_stats_flag && 
        strcmp(blob->l1r_file_metadata.collect_type, 
        IAS_RPS_EARTH_IMAGING_COLLECT) != 0)
        
    {
        IAS_LOG_WARNING("Trying to collect summary stats from non-earth"
            "collection, summary statistics will not be generated");
        blob->populate_scene_summary_stats_flag = 0;
    }

    /* If store_in_database is not set then unset scene stats flag*/
    if (!blob->store_in_database && blob->populate_scene_summary_stats_flag)
    {
         blob->populate_scene_summary_stats_flag = 0;
    }

    /* Delete any pre-existing records for this work order if database
       storage was requested */
    if (blob->store_in_database)
    {
        const char *hist_stats_table_name;

        /* Get the correct histogram statistics table name according to the
           collection type provided. */
        hist_stats_table_name = ias_rps_get_histogram_statistics_table_name
                                (blob->l1r_file_metadata.collect_type);

        status = delete_db_table_records( blob->database_information,
            hist_stats_table_name);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Returned from delete_db_table_data");
            return ERROR;
        }
        if (blob->populate_scene_summary_stats_flag)
        {
            status = delete_db_table_records(blob->database_information,
                "hist_scene_summary_stats");
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Returned from delete_db_table_data "
                    "hist_scene_summary_stats");
                return ERROR;
            }
        }
    }

    return SUCCESS;
}  /* END histogram_statistics_characterization_preprocess */

/*--------------------------------------------------------------------------
NAME:    histogram_statistics_characterization_postprocess

PURPOSE: Runs checks after histogram_statistics_characterization runs

RETURNS: SUCCESS or ERROR
---------------------------------------------------------------------------*/
int histogram_statistics_characterization_postprocess
(
    BLOB *blob          /* I: Miscellaneous parameters */
)
{
    int num_list_nodes;

    num_list_nodes = ias_linked_list_count_nodes(blob->scene_statistics_list);
    if (num_list_nodes > 0)
    {
        IAS_LOG_WARNING("There were %d SCAs not inserted into the database",
            num_list_nodes);
    }

    return SUCCESS;
}

/*--------------------------------------------------------------------------
NAME:    store_histogram_statistics_per_sca_to_db

PURPOSE: Calculates & stores per SCA summary statistics to the database. 
         If all SCAs are present in linked list it will calculate the per
         band summary statistics and store them into the database as well.

RETURNS: SUCCESS or ERROR
---------------------------------------------------------------------------*/
int store_histogram_statistics_per_sca_to_db
(
    void * message  /* IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS for SCA */
)
{   
    int status;                 /* Status variable */
    int scas_per_band;          /* Hold number of SCAs for current band */
    int index;                  /* Loop variable for SCA number */
    int all_scas_present = TRUE;/* Boolean value for all sca stats stored */
    DB_SCENE_SUMMARY_STATISTICS *statistics; /* SCA struct */
    IAS_LINKED_LIST_NODE *head; /* Pointer to linked list head */
    DB_SCENE_SUMMARY_STATISTICS *current_node; /* Loop pointer */

    /* Addresses of current band nodes */
    DB_SCENE_SUMMARY_STATISTICS *sca_addr[IAS_MAX_NSCAS + 1] = {0}; 

    /* Cast message to correct structure pointer */
    statistics = (DB_SCENE_SUMMARY_STATISTICS *) message;
    if (!statistics)
    {
        IAS_LOG_ERROR("Message passed is NULL");
        return ERROR;
    }

    /* Get the scene summary list */
    head = get_scene_summary_list();

    /* Initialize the node and add it to the head of the list */
    ias_linked_list_initialize_node(&statistics->node);
    ias_linked_list_add_head(head, &statistics->node);

    /* Get number of scas per band */
    scas_per_band = ias_sat_attr_get_scas_per_band(statistics->band_number);
    if (scas_per_band == ERROR)
    {
        IAS_LOG_ERROR("Getting SCAs per band for Band %d", 
            statistics->band_number);
        return ERROR;
    }

    /* Loop getting all SCAs with same band number, lineset */
    GET_OBJECT_FOR_EACH_ENTRY(current_node, head, DB_SCENE_SUMMARY_STATISTICS, 
        node)
    {
        int band_match = (statistics->band_number == current_node->band_number);
        int lineset_match = (statistics->lineset == current_node->lineset);
        if (band_match && lineset_match)
        {
            sca_addr[current_node->sca_number - 1] = current_node;
        }
    }

    /* Check if all SCAs are present*/
    for (index = 0; index < scas_per_band; index++)
    {
        if (!sca_addr[index])
        {
            all_scas_present = FALSE;
            break;
        }
    }

    if (all_scas_present)
    {
        int frames[IAS_MAX_NSCAS]; /* Keeps track of frames */
        float mean[IAS_MAX_NSCAS]; /* Keeps track of means */
        float stdev[IAS_MAX_NSCAS];/* Keeps track of stdevs */
        DB_SCENE_SUMMARY_STATISTICS sca_summary;/* Summary SCA */
                
        /* Copy Frames, Mean, Stdev */
        for (index = 0; index < scas_per_band; index++)
        {
            frames[index] = sca_addr[index]->frames; 
            mean[index] = sca_addr[index]->mean; 
            stdev[index] = sca_addr[index]->stdev;
        }
         
        /* Calculate the SCA summary stats */
        status = ias_math_merge_mean_and_stdev_samples(frames, mean, stdev,
            scas_per_band, &sca_summary.stdev, &sca_summary.mean, 
            &sca_summary.frames);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Calculating the summary stats for band %d",
                statistics->band_number);
            return ERROR;
        }

        /* Set the summary stats band number and SCA (0 for the band summary) */
        sca_summary.band_number = statistics->band_number;
        sca_summary.sca_number = 0;
        
        /* Store summary stats address in sca_address array */
        sca_addr[scas_per_band] = &sca_summary;

        /* Store the results in the database */
        status = store_histogram_summary_stats_in_database(sca_addr,
            scas_per_band + 1);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Storing SCA summary statistics to database for Band "
                "%d", statistics->band_number);
            return ERROR;
        }

        /* Free up all the memory for current band, lineset, rad id */
        for (index = 0; index < scas_per_band; index++)
        {
            ias_linked_list_remove_node(&sca_addr[index]->node);
            free(sca_addr[index]);
        }
    }

    return SUCCESS;
}

/*-------------------------------------------------------------------------
   NAME:    store_histogram_summary_stats_in_database

   PURPOSE: Stores the histogram summary statistics in the database

   RETURNS: SUCCESS/ERROR
-------------------------------------------------------------------------*/
int store_histogram_summary_stats_in_database
(
    DB_SCENE_SUMMARY_STATISTICS **statistics,/* I: Stats array ptr */
    const int num_scas                       /* I: Total SCAs */
)
{
    int number_db_columns;      /* Number of columns */
    int status;                 /* Status holder */
    int sca_numbers[num_scas];  /* SCA number array */
    float stdevs[num_scas];     /* Stdev array */
    float means[num_scas];      /* Means array */
    float frames[num_scas];     /* Frames array */
    int index;                  /* Generic loop variable */
    char rad_common_id[IAS_CHAR_ID_SIZE]; /* Rad common ID */

    /* Copy the rad_common_id from blob */
    strcpy(rad_common_id, get_blob_rad_common_id());

    /* Copy the SCA, stdev, means, and frames calculations */
    for (index = 0; index < num_scas; index++)
    {
        sca_numbers[index] = statistics[index]->sca_number;
        stdevs[index] = statistics[index]->stdev;
        means[index] = statistics[index]->mean;
        frames[index] = statistics[index]->frames;   
    }
    
    /* Create the table to be inserted */
    IAS_DB_INSERT_TABLE table[]=
    {
        IAS_DB_TABLE_STRING("radiometric_wo_common_id", rad_common_id),
        IAS_DB_TABLE_INT("band_id", statistics[0]->band_number),
        IAS_DB_TABLE_INT_ARRAY("sensor_chip_assembly_id", sca_numbers),
        IAS_DB_TABLE_FLOAT_ARRAY("scene_stdev", stdevs),
        IAS_DB_TABLE_FLOAT_ARRAY("scene_mean", means),
        IAS_DB_TABLE_FLOAT_ARRAY("scene_number_frames", frames),
        IAS_DB_TABLE_INT("lineset" , statistics[0]->lineset)
    };

    /* Get the number of columns inserted into the table and insert the
       data into the database */
    number_db_columns = sizeof(table) / sizeof(IAS_DB_INSERT_TABLE);

    /* Store the table to the DB */
    status = ias_db_insert_using_table(get_blob_db_connection(),
        "hist_scene_summary_stats", table, number_db_columns, num_scas);
    if (status)
    {
        IAS_LOG_ERROR("Inserting histogram summary statistics for band %d, "
            "SCA %d", statistics[0]->band_number, statistics[0]->sca_number);
        return ERROR;
    }

    return SUCCESS;
}
