/*------------------------------------------------------------------------
NAME:    api_impulse_noise_characterization

PURPOSE: Gather data for the algorithm and call it

RETURNS: SUCCESS or ERROR
-------------------------------------------------------------------------*/
#include <string.h>
#include "ias_logging.h"
#include "ias_radiometry.h"
#include "ias_threadpool.h"
#include "rps_local.h"

/* Define a list of memory buffers needed by the algorithm */
static int number_of_memory_buffers;
static MEMORY_BUFFER_TYPE memory_buffers[] =
{
    IMAGE_L0R,
    MM_IMPULSE_NOISE_LIMIT,
    MM_IMPULSE_NOISE_FILTER_WIDTH,
    MM_DETECTOR_NOISE
};


/*------------------------------------------------------------------------
NAME:    impulse_noise_characterization

PURPOSE: Gather data for the algorithm and call it

RETURNS: SUCCESS or ERROR
-------------------------------------------------------------------------*/
static int impulse_noise_characterization
(
    BLOB *blob,         /* I: Miscellaneous parameters */
    int   band_index,   /* I: */
    int   sca_index     /* I: */
)
{
    uint16_t *data_buffer;    /* Level 0 data buffer */
    int i;
    int status;
    RPS_BAND_ATTRIBUTES *band_attributes = get_band_attributes( band_index );
    int band_number = band_attributes->band_number;
    int number_lines = band_attributes->number_lines;
    int number_dets = band_attributes->number_detectors;
    const int *median_filter_width;
    const int *impulse_noise_limit;
    const double *detector_noise;
    IAS_REPORT *report_info = NULL;
    IAS_RPS_DB_IMPULSE_NOISE_CHAR_RESULTS *db_results = NULL;

    /* Retrieve the buffers needed by the algorithm */
    data_buffer = (uint16_t *)get_memory_buffer( IMAGE_L0R, CONSUMER,
        band_index, sca_index );
    if ( ! data_buffer )
    {
        IAS_LOG_ERROR("Getting %s memory buffer", get_memory_buffer_string(
            IMAGE_L0R ));
        return ERROR;
    }
    impulse_noise_limit = (const int *)get_const_memory_buffer(
        MM_IMPULSE_NOISE_LIMIT, band_index, sca_index );
    if ( !impulse_noise_limit )
    {
        IAS_LOG_ERROR("Getting %s memory buffer", get_memory_buffer_string(
            MM_IMPULSE_NOISE_LIMIT ));
        return ERROR;
    }
    median_filter_width = (const int *)get_const_memory_buffer(
        MM_IMPULSE_NOISE_FILTER_WIDTH, band_index, sca_index );
    if ( !median_filter_width )
    {
        IAS_LOG_ERROR("Getting %s memory buffer", get_memory_buffer_string(
            MM_IMPULSE_NOISE_FILTER_WIDTH ));
        return ERROR;
    }
    detector_noise = (const double *)get_const_memory_buffer( MM_DETECTOR_NOISE,
        band_index, sca_index );
    if ( !detector_noise )
    {
        IAS_LOG_ERROR("Getting %s memory buffer", get_memory_buffer_string(
            MM_DETECTOR_NOISE ));
        return ERROR;
    }

    /* If a report was specified get the report information from the blob */
    if ( blob->generate_report )
        report_info = &( blob->report_information );

    /* Initialize database and information and allocate memory if requested */
    if ( blob->store_in_database )
    {
        db_results = ias_rps_db_allocate_impulse_noise_results( number_dets );
        if ( !db_results )
        {
            IAS_LOG_ERROR("Allocating memory for the database results");
            return ERROR;
        }
        db_results->dbase = blob->database_information.db;
        strncpy(db_results->rad_common_id, 
            blob->database_information.rad_common_id, 
            sizeof(db_results->rad_common_id));
        db_results->rad_common_id[sizeof(db_results->rad_common_id) - 1] = '\0';
        db_results->band = band_number;
        db_results->sca  = sca_index + 1;
        db_results->num_detectors = number_dets;
    }

    status = ias_rps_impulse_noise_characterization( data_buffer,
        band_number, sca_index + 1, number_dets, number_lines,
        median_filter_width[0], detector_noise, impulse_noise_limit[0],
        blob->pixel_mask[band_index][sca_index],
        report_info, db_results );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Characterizing impulse noise for band %d SCA %d",
            band_number, sca_index + 1 );
        ias_rps_db_free_impulse_noise_results( db_results );
        return ERROR;
    }

    /* Store data in the database if requested */
    if ( blob->store_in_database )
    {
        status = ias_threadpool_add_to_work_queue( blob->threadpool,
            ias_rps_store_impulse_noise_counts_in_database, db_results );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Submitting impulse noise results to work queue "
                "for band %d, SCA %d", band_number, sca_index + 1 );
            /* Before freeing memory, ensure there is more than one thread
               running.  A double free of the memory may occur if only one
               thread is running and an internal database error occurs. */
            if ( ias_threadpool_get_thread_count( blob->threadpool ) > 1 )
                ias_rps_db_free_impulse_noise_results( db_results );
            return ERROR;
        }
    }

    /* Release the memory buffers */
    for ( i = 0; i < number_of_memory_buffers; i++ )
    {
        decrement_consumer_count( memory_buffers[i], band_index, sca_index );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }
    }

    return SUCCESS;
}  /* END impulse_noise_characterization */


/*------------------------------------------------------------------------
NAME:    api_impulse_noise_characterization

PURPOSE: External API to call the internal routine.  It will call the 
         internal routine with the current band then with its associated VRP
         and blind band if needed.

RETURNS: SUCCESS or ERROR
-------------------------------------------------------------------------*/
int api_impulse_noise_characterization
(
    BLOB *blob,         /* I: Miscellaneous parameters */
    int   band_index,   /* I: */
    int   sca_index,    /* I: */
    int   algo_index    /* I: Algorithm index */
)
{
    RPS_BAND_ATTRIBUTES *band_attributes = get_band_attributes( band_index );
    int status;

    status = impulse_noise_characterization( blob, band_index,
        sca_index );
    if ( status != SUCCESS )
        return ERROR;

    if ( band_attributes->vrp_band_number > 0 )
    {
        status = impulse_noise_characterization( blob,
            band_attributes->vrp_band_index, sca_index );
        if ( status != SUCCESS )
            return ERROR;
    }

    if ( band_attributes->blind_band_number > 0 
        && band_attributes->process_blind_band_flag )
    {
        status = impulse_noise_characterization( blob,
            band_attributes->blind_band_index, sca_index );
        if ( status != SUCCESS )
            return ERROR;
    }

    return SUCCESS;
}  /* END api_impulse_noise_characterization */


/*------------------------------------------------------------------------
NAME:    impulse_noise_characterization_preprocess

PURPOSE: Register memory buffers for the algorithm

RETURNS: SUCCESS or ERROR
-------------------------------------------------------------------------*/
int impulse_noise_characterization_preprocess
(
    BLOB *blob          /* I: Miscellaneous parameters */
)
{
    int i;
    int status;

    number_of_memory_buffers = sizeof(memory_buffers) /
        sizeof(MEMORY_BUFFER_TYPE);

    /* Increment the consumer counts for the memory buffers */
    for ( i = 0; i < number_of_memory_buffers; i++ )
    {
        status = increment_consumer_count( memory_buffers[i] );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Incrementing %s consumer count",
                get_memory_buffer_string( memory_buffers[i] ));
            return ERROR;
        }
    }

    /* Delete any pre-existing records for this work order if database
       storage was requested */
    if ( blob->store_in_database )
    {
        status = delete_db_table_records( blob->database_information,
            "impulse_noise");
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Returned from delete_db_table_data");
            return ERROR;
        }
    }

    return SUCCESS;
}  /* END impulse_noise_characterization_preprocess */
