/*----------------------------------------------------------------------------
NAME:    api_inoperable_detector_identification

PURPOSE: Gather data needed for the inoperable detector identification
         routine and call it.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
#include "ias_logging.h"
#include "rps_local.h"

int api_inoperable_detector_identification
(
    BLOB *blob,         /* I: Miscellaneous parameters */
    int   band_index,   /* I: */
    int   sca_index,    /* I: */
    int   algo_index    /* I: Algorithm index */
)
{
    RPS_BAND_ATTRIBUTES *band_attributes = get_band_attributes( band_index );
    int band_number = band_attributes->band_number;
    int sca_number = sca_index + 1;
    int num_lines = band_attributes->number_lines;
    int num_sca_detectors = band_attributes->number_detectors;
    const int *inoperable_detector_list;
    int number_inoperable_detectors;
    int status;

    /* Get the number of values in each list */
    number_inoperable_detectors = get_memory_buffer_number_values(
        MM_INOPERABLE_DETECTORS, band_index, sca_index );

    /* If there are inoperable detectors, get the CPF inoperable detectors
       list and set the pixel mask. */
    if ( number_inoperable_detectors > 0 )
    {
        inoperable_detector_list = (const int *)get_const_memory_buffer(
            MM_INOPERABLE_DETECTORS, band_index, sca_index );
        if ( !inoperable_detector_list )
        {
            IAS_LOG_ERROR("Retrieving memory buffer MM_INOPERABLE_DETECTORS "
                "for band %d SCA %d", band_number, sca_number );
            return ERROR;
        }

        /* Call the library routine to set the pixel mask */
        status = ias_rps_inoperable_detector_identification(
            inoperable_detector_list, number_inoperable_detectors,
            band_number, sca_number, num_sca_detectors, num_lines,
            blob->pixel_mask[band_index][sca_index] );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Returned from "
                "ias_rps_inoperable_detector_identification for "
                "band %d SCA %d", band_number, sca_number );
            return ERROR;
        }
    }

    /* Let memory management know the data was used.  Regardless on whether
       the data was actually used or not, the count must be decremented. */
    status = decrement_consumer_count( MM_INOPERABLE_DETECTORS, band_index,
        sca_index );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Decrementing memory buffer");
        return ERROR;
    }

    return SUCCESS;
}  /* END api_inoperable_detector_identification */


/*----------------------------------------------------------------------------
NAME:    inoperable_detector_identification_preprocess

PURPOSE: Run pre-process steps

RETURNS: SUCCESS or ERROR
----------------------------------------------------------------------------*/
int inoperable_detector_identification_preprocess
(
    BLOB *blob          /* I: Miscellaneous parameters */
)
{
    /* Define a list of memory buffers needed by the algorithm */
    MEMORY_BUFFER_TYPE memory_buffers[] =
    {
        MM_INOPERABLE_DETECTORS,
    };
    int number_of_memory_buffers = sizeof(memory_buffers) /
        sizeof(MEMORY_BUFFER_TYPE);
    int i;
    int status;

    /* Increment the consumer counts for the memory buffers */
    for ( i = 0; i < number_of_memory_buffers; i++ )
    {
        status = increment_consumer_count( memory_buffers[i] );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Incrementing %s consumer count",
                get_memory_buffer_string( memory_buffers[i] ));
            return ERROR;
        }
    }

    return SUCCESS;
}  /* END inoperable_detector_identification_preprocess */
