/**************************************************************************
 NAME:     api_inoperable_detectors_fill

 PURPOSE:  Gathers information needed to run the inoperable_detectors_fill
           routine, then calls the routine

 RETURNS:  Integer status code of SUCCESS or ERROR
***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ias_logging.h"
#include "rps_local.h"


/* Define a globally applicable list of memory buffers needed by the
   application and determine the overall size of the required data
   structure */
static MEMORY_BUFFER_TYPE memory_buffers[] =
{
    IMAGE_L1R,
    MM_SPATIAL_DETECTOR_OFFSETS
};
static int number_of_memory_buffers = sizeof(memory_buffers)
    / sizeof(MEMORY_BUFFER_TYPE);


/*-------------------------------------------------------------------------
 NAME:     decrement_counters

 PURPOSE:  Internal routine to decrement the consumer counts of all
           required memory buffers

 RETURNS:  SUCCESS or ERROR
--------------------------------------------------------------------------*/
static int decrement_counters
(
    int band_index,            /* I: Index to current band */
    int sca_index              /* I: Index to current SCA */
)
{
    /* Get the attributes for the current band */
    RPS_BAND_ATTRIBUTES *band_atts = get_band_attributes(band_index);

    int index;                     /* Loop counter for consumer counts */
    int status;


    /* Decrement the consumer counts for the VRP and blind bands as well
       so memory management can free these when nobody else needs them */
    for (index = 0; index < number_of_memory_buffers; index++)
    {
        status = decrement_consumer_count(memory_buffers[index], band_index,
            sca_index);
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }

        if (band_atts->vrp_band_number > 0)
        {
            status = decrement_consumer_count(memory_buffers[index],
                band_atts->vrp_band_index, sca_index);
            if ( status != SUCCESS )
            {
                IAS_LOG_ERROR("Decrementing memory buffer");
                return ERROR;
            }
        }
        if ((band_atts->blind_band_number > 0)
            && (band_atts->process_blind_band_flag))
        {
            status = decrement_consumer_count(memory_buffers[index],
                band_atts->blind_band_index, sca_index);
            if ( status != SUCCESS )
            {
                IAS_LOG_ERROR("Decrementing memory buffer");
                return ERROR;
            }
        }
    }
    return SUCCESS;
}   /* END -- decrement_counters */



/*************************************************************************
 NAME:     api_inoperable_detectors_fill

 PURPOSE:  Main API routine

 RETURNS:  Integer status code of SUCCESS or ERROR
**************************************************************************/
int api_inoperable_detectors_fill
(
    BLOB *blob,                /* I: Misc. processing parameters */
    int  band_index,           /* I: Current band */
    int  sca_index,            /* I: Current SCA */
    int  algo_index            /* I: Index into algorithm list */
)
{
    float    *l1r_buffer = NULL;  /* Image data buffer */
    int *normalized_detector_offsets = NULL;
                                  /* Integerized CPF detector offsets
                                     normalized such that the "smallest"
                                     offset is 0 */

    RPS_BAND_ATTRIBUTES *band_atts = get_band_attributes(band_index);
    int  band_number      = band_atts->band_number;
    int  number_detectors = band_atts->number_detectors;
    int  number_lines     = band_atts->number_lines;
    int  sca_number       = (sca_index + 1);
    int  status;

    /* Add the ID for inoperable detectors fill to RPS processing level */
    if (set_rps_level(band_index, sca_index, 
            IAS_RPS_LEVEL_INOPERABLE_DETECTORS) == ERROR)
    {
        IAS_LOG_ERROR("Updating the RPS processing level");
        return ERROR;
    }

    /* This algorithm doesn't process OLI blind or VRP bands therefore the
        processing level *must* be set to skip */
    if ( band_atts->vrp_band_index >= 0 )
    {
        if (set_rps_level(band_atts->vrp_band_index, sca_index,
            IAS_RPS_LEVEL_SKIP) == ERROR)
        {
            IAS_LOG_ERROR("Updating RPS processing level");
            return ERROR;
        }
    }
    if ( band_atts->blind_band_index >= 0 &&
         band_atts->process_blind_band_flag )
    {
        if (set_rps_level(band_atts->blind_band_index, sca_index,
            IAS_RPS_LEVEL_SKIP) == ERROR)
        {
            IAS_LOG_ERROR("Updating RPS processing level");
            return ERROR;
        }
    }

    /* Get the image data */
    l1r_buffer = (float *)get_memory_buffer(IMAGE_L1R, CONSUMER, band_index,
        sca_index);
    if (l1r_buffer == NULL)
    {
        IAS_LOG_ERROR("Retrieving band %d SCA %d image data", band_number,
            sca_number);
        return ERROR;
    }

    /* Grab the normalized CPF (integer) detector offsets */
    normalized_detector_offsets = (int *)get_memory_buffer(
        MM_SPATIAL_DETECTOR_OFFSETS, CONSUMER, band_index, sca_index);
    if (normalized_detector_offsets == NULL)
    {
        IAS_LOG_ERROR("Retrieving band %d SCA %d integer detector offsets",
            band_number, sca_number);
        return ERROR;
    }

    /* Call the algorithm */
    status = inoperable_detectors_fill(l1r_buffer, number_detectors,
        number_lines, normalized_detector_offsets,
        blob->pixel_mask[band_index][sca_index]);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("in band %d SCA %d call to inoperable_detectors_fill "
            "routine", band_number, sca_number);
        return ERROR;
    }

    /* Decrement the consumer counters */
    status = decrement_counters(band_index, sca_index);
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Decrementing memory buffer");
        return ERROR;
    }

    return SUCCESS;
}   /*  END -- api_inoperable_detectors_fill */



/*************************************************************************
 NAME:     inoperable_detectors_fill_preprocess

 PURPOSE:  Increments the consumer counts for all memory buffers required
           by the application

 RETURNS:  Integer status code of SUCCESS or ERROR
**************************************************************************/
int inoperable_detectors_fill_preprocess
(
    BLOB  *blob                    /* I: Misc. processing parameters */
)
{
    int i;                         /* Loop counter */
    int status;


    for (i = 0; i < number_of_memory_buffers; i++)
    {
        status = increment_consumer_count(memory_buffers[i]);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Incrementing %s consumer count",
                get_memory_buffer_string(memory_buffers[i]));
            return ERROR;
        }
    }

    /* Set the L1R file metadata */
    blob->l1r_file_metadata.inoperable_detector_fill_applied = 1;

    return SUCCESS;
}   /* END -- inoperable_detectors_fill_preprocess */
