/***************************************************************************
NAME:      api_oli_bias_determination

PURPOSE:   Routines needed to gather data for the OLI bias determination
           algorithm and decide which memory buffers are needed based on
           selections in the blob set from input parameters.

RETURNS:   SUCCESS or ERROR

NOTE:      The CPF_BIAS option is removed because the bias removal algorithm 
           reads the CPF per detector bias directly to save memory and 
           execution time if CPF_BIAS option is selected (No bias determination
           algorithm is needed to run).(Refer: Bug tracker 3906)
****************************************************************************/
#include <stdlib.h>
#include <string.h>

#include "ias_logging.h"
#include "rps_local.h"
#include "ias_satellite_attributes.h"

/* Define a list of memory buffers needed by the algorithm */
static MEMORY_BUFFER_TYPE memory_buffers[] =
{
    IMAGE_L0R,
    MM_NON_PAN_PRE_SHUTTER,
    MM_NON_PAN_POST_SHUTTER,
    MM_NON_PAN_MEAN_SHUTTER,
    MM_NON_PAN_CPF_BIAS,
    MM_PAN_ODD_PRE_SHUTTER,
    MM_PAN_ODD_POST_SHUTTER,
    MM_PAN_ODD_MEAN_SHUTTER,
    MM_PAN_ODD_CPF_BIAS,
    MM_PAN_EVEN_PRE_SHUTTER,
    MM_PAN_EVEN_POST_SHUTTER,
    MM_PAN_EVEN_MEAN_SHUTTER,
    MM_PAN_EVEN_CPF_BIAS,
    MM_NON_PAN_BIAS_A1_PARAMETER,
    MM_PAN_ODD_BIAS_A1_PARAMETER,
    MM_PAN_EVEN_BIAS_A1_PARAMETER,
    MM_NON_PAN_BIAS_C1_PARAMETER,
    MM_PAN_ODD_BIAS_C1_PARAMETER,
    MM_PAN_EVEN_BIAS_C1_PARAMETER,
    MM_NON_PAN_BIAS_A0_PARAMETER,
    MM_PAN_ODD_BIAS_A0_PARAMETER,
    MM_PAN_EVEN_BIAS_A0_PARAMETER,
    MM_NON_PAN_PERFRAME_BIAS,
    MM_PAN_PERFRAME_BIAS
};
static int number_of_memory_buffers = sizeof(memory_buffers) /
        sizeof(MEMORY_BUFFER_TYPE);

/**************************************************************************
NAME:     decrement_counters

PURPOSE:  Internal routine to decrement the consumer counts of all required
          memory buffers

RETURNS:  SUCCESS or ERROR
***************************************************************************/
static int decrement_counters
(
    int band_index,     /* I: 0-based band index */
    int sca_index       /* I: 0-based SCA index */
)
{
    RPS_BAND_ATTRIBUTES *band_attributes;   /* RPS band attributes */
    int index;                              /* Local loop counter */
    int status;

    /* Decrement the consumer counts for the memory buffers */
    for (index = 0; index < number_of_memory_buffers; index++)
    {
        status = decrement_consumer_count(memory_buffers[index], band_index,
            sca_index);
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }
    }
    band_attributes = get_band_attributes(band_index);

    /* Decrement IMAGE_L0R for vrp band if the band has an associated VRP
       band */
    if (band_attributes->vrp_band_number > 0)
    {
        status = decrement_consumer_count(IMAGE_L0R,
            band_attributes->vrp_band_index, sca_index);
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }
    }

    /* If the band has an associated blind band and it's set to process the
       blind band, decrement the user count */
    if (band_attributes->blind_band_number > 0
        && band_attributes->process_blind_band_flag )
    {
        status = decrement_consumer_count( IMAGE_L0R, 
            band_attributes->blind_band_index, sca_index );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }
    }

    return SUCCESS;
}   /* END decrement_counters */

/*----------------------------------------------------------------------------
NAME:    api_oli_bias_determination

PURPOSE: Gather data needed for the OLI bias determination algorithm routine 
         and call the routine 

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int api_oli_bias_determination
(
    BLOB *blob,         /* I: Miscellaneous parameters */
    int   band_index,   /* I: */
    int   sca_index,    /* I: */
    int   algo_index    /* I: Algorithm index */
)
{
    RPS_BAND_ATTRIBUTES *band_attributes = get_band_attributes(band_index);
    int band_number = band_attributes->band_number;
    int sca_number = sca_index + 1;
    RPS_BAND_ATTRIBUTES *vrp_band_attributes; /* RPS VRP band attributes */
    BIAS_SPECTRAL_TYPE bias_spectral_type;    
        /* Flag to indicate the spectral type for bias application, NON_PAN =0,
           PAN_ODD =1, PAN_EVEN = 2 */
    uint16_t *vrp_buffer = NULL;   /* Level 0 VRP data buffer */
    int number_lines;              /* Number of lines in the image */
    int number_detectors;          /* Number of detectors for the SCA */
    int vrp_detectors;             /* Number of detectors for VRP band */  
    int vrp_lines;                 /* Number of lines for VRP band */  
    IAS_REPORT *report_information = NULL; 
    double *bias_from_file = NULL; /* I: Per detector bias from BPF/CPF */
    double *a1_parameter = NULL;   /* A1_parameter pointer */
    double *c1_parameter = NULL;   /* C1_parameter pointer */
    double *a0_parameter = NULL;   /* A0_parameter pointer */
    double a0_value = 0.0;         /* A0_parameter value for NON_PAN, and 
                                      PAN ODD bands */
    double *perframe_bias = NULL;  /* Pointer to hold result output bias */
    int status = SUCCESS;          /* Routine routine status */
    int loop_index;                /* Loop index */
    int loop_number;               /* Number of iteration for loop */ 
    int a1_c1_estimate_flag = 0;   /* Flag indicating detector level bias 
                                      estimated with a1/C1 parameters */ 
    int bias_source_to_buffer_map[3][4] = {{MM_NON_PAN_PRE_SHUTTER, 
           MM_NON_PAN_POST_SHUTTER, MM_NON_PAN_MEAN_SHUTTER,
           MM_NON_PAN_CPF_BIAS}, {MM_PAN_ODD_PRE_SHUTTER, 
           MM_PAN_ODD_POST_SHUTTER, MM_PAN_ODD_MEAN_SHUTTER,
           MM_PAN_ODD_CPF_BIAS}, {MM_PAN_EVEN_PRE_SHUTTER, 
           MM_PAN_EVEN_POST_SHUTTER, MM_PAN_EVEN_MEAN_SHUTTER,
           MM_PAN_EVEN_CPF_BIAS}};
        /* Array for bias_spectral_type and memory buffer map */
    int perframe_bias_to_buffer_map[3] = {MM_NON_PAN_PERFRAME_BIAS, 
           MM_PAN_PERFRAME_BIAS, MM_PAN_PERFRAME_BIAS};
        /* Array for perframe_bias and memory buffer map */
    int a0_parameter_to_buffer_map[3] = {MM_NON_PAN_BIAS_A0_PARAMETER, 
           MM_PAN_ODD_BIAS_A0_PARAMETER, MM_PAN_EVEN_BIAS_A0_PARAMETER};
        /* Array for a0_parameter and memory buffer map */
    int a1_parameter_to_buffer_map[3] = {MM_NON_PAN_BIAS_A1_PARAMETER, 
           MM_PAN_ODD_BIAS_A1_PARAMETER, MM_PAN_EVEN_BIAS_A1_PARAMETER};
        /* Array for a1_parameter and memory buffer map */
    int c1_parameter_to_buffer_map[3] = {MM_NON_PAN_BIAS_C1_PARAMETER, 
           MM_PAN_ODD_BIAS_C1_PARAMETER, MM_PAN_EVEN_BIAS_C1_PARAMETER};
        /* Array for c1_parameter and memory buffer map */
    MEMORY_BUFFER_TYPE bias_from_file_type = MM_NON_PAN_MEAN_SHUTTER;       
        /* Needed memory buffer type for bias_from_file */
    MEMORY_BUFFER_TYPE a0_parameter_type = MM_NON_PAN_BIAS_A0_PARAMETER;
        /* Needed memory buffer type for a0_parameter */
    MEMORY_BUFFER_TYPE a1_parameter_type = MM_NON_PAN_BIAS_A0_PARAMETER;
        /* Needed memory buffer type for a1_parameter */
    MEMORY_BUFFER_TYPE c1_parameter_type = MM_NON_PAN_BIAS_A0_PARAMETER;
        /* Needed memory buffer type for c1_parameter */
    MEMORY_BUFFER_TYPE perframe_bias_type = MM_NON_PAN_PERFRAME_BIAS;          
        /* Needed memory buffer type for output per frame bias */

    /* Return SUCCESS if CPF_BIAS is selected for BIAS_SOURCE parameter,
       meaning the bias determination algorithm will be skipped if CPF_BIAS is 
       selected for BIAS_SOURCE, and the bias removal algorithm will directly
       read bias from CPF and do the needed bias removal */
    if (blob->bias_source == CPF_BIAS && blob->per_frame_bias_flag == 0)
    {
        status = decrement_counters(band_index, sca_index);
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }
        return SUCCESS;
    } 

    /* Get number_lines and number_detectors for the band and number_detectors
       for the corresponding VRP band, the number_lines for VRP band is the 
       same as the number_lines for the corresponding normal band */
    number_lines = band_attributes->number_lines;
    number_detectors = band_attributes->number_detectors;
    if (band_attributes->spectral_type == IAS_SPECTRAL_THERMAL)
    {
        status = decrement_counters(band_index, sca_index);
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }
        return SUCCESS;
    }
    else
    {
        int vrp_index = band_attributes->vrp_band_index;
        vrp_band_attributes = get_band_attributes(vrp_index);
        vrp_detectors = vrp_band_attributes->number_detectors;
        vrp_lines = vrp_band_attributes->number_lines;
    }

    /* Set the reporting information */
    if (blob->generate_report)
        report_information = &blob->report_information;

    /* Set the spectral_index, loop_number, and bias_spectral_type
       flag based on band spectral type */
    switch(band_attributes->spectral_type)
    {
        case IAS_SPECTRAL_VNIR:
        case IAS_SPECTRAL_SWIR:
            bias_spectral_type = NON_PAN;
            loop_number = 1;
            break;

        case IAS_SPECTRAL_PAN:
            bias_spectral_type = PAN_ODD;
            loop_number = 2;
            break;

        default:
            IAS_LOG_ERROR("Unsupported spectral_type");
            return ERROR;
    }

    /* Check to make sure blob->bias_source value be correctly set */
    if (blob->bias_source < PRE_ACQUISITION_SHUTTER 
        || blob->bias_source > A1_C1_ESTIMATION)
    {
        IAS_LOG_ERROR("Incorrect blob->bias_source value");
        return ERROR;
    }

    /* Get the output per-frame bias buffer
        Note: Only need one output per-frame bias buffer for PAN band */
    perframe_bias_type = perframe_bias_to_buffer_map[bias_spectral_type];

    /* Get the bias results buffer */
    perframe_bias = (double *)get_memory_buffer(perframe_bias_type, SETTER, 
        band_index, sca_index);
    if (!perframe_bias)
    {
        IAS_LOG_ERROR("Getting bias buffer: %s for band %d SCA %d", 
            get_memory_buffer_string(perframe_bias_type), 
            band_index, sca_number);
        return ERROR;
    }

    /* Note: It isn't necessary to decrement the usage of the memory buffer
       on error conditions from this point on.  The main thread will clean
       up allocated buffers when error conditions dictate.  It doesn't hurt
       but it isn't necessary. */
    for (loop_index = 0; loop_index < loop_number; loop_index++)
    {
        /* Get the PAN_EVEN needed memory buffer indices for all needed buffer 
           types */
        if (loop_index != 0)
            bias_spectral_type = PAN_EVEN;

        /* Get the needed memory buffer types */
        if (blob->bias_source != A1_C1_ESTIMATION)
        {
            bias_from_file_type = bias_source_to_buffer_map[bias_spectral_type]
                                  [blob->bias_source];
            a1_c1_estimate_flag = 0;
        }
        else
        {
            a1_parameter_type = a1_parameter_to_buffer_map[bias_spectral_type];
            c1_parameter_type = c1_parameter_to_buffer_map[bias_spectral_type];
            a1_c1_estimate_flag = 1;
        }

        /* Get the bias data from BPF via memory buffer */
        if (blob->bias_source == PRE_ACQUISITION_SHUTTER
            || blob->bias_source == POST_ACQUISITION_SHUTTER
            || blob->bias_source == MEAN_ACQUISITION_SHUTTER)
        {
            bias_from_file = (double *)get_memory_buffer(
                   bias_from_file_type, CONSUMER, band_index, sca_index);
            if (!bias_from_file)
            {
                IAS_LOG_ERROR("Getting memory buffer: %s for band %d SCA %d", 
                    get_memory_buffer_string(bias_from_file_type),
                    band_number, sca_number);
                return ERROR;
            }
        }
        /* Get the bias data from CPF via memory buffer */
        else if (blob->bias_source == CPF_BIAS)
        {
            bias_from_file = (double *)get_const_memory_buffer(
                   bias_from_file_type, band_index, sca_index);
            if (!bias_from_file)
            {
                IAS_LOG_ERROR("Getting memory buffer: %s for band %d SCA %d", 
                    get_memory_buffer_string(bias_from_file_type),
                    band_number, sca_number);
                return ERROR;
            }
        }
        else
        {
            /* Get the a1 parameter data buffer */
            a1_parameter = (double *)get_memory_buffer(a1_parameter_type, 
                           CONSUMER, band_index, sca_index);
            if (!a1_parameter)
            {
                IAS_LOG_ERROR("Getting A1 parameter: %s for band %d SCA %d",
                get_memory_buffer_string(a1_parameter_type), 
                band_number, sca_number);
                return ERROR;
            }  

            /* Get the c1 parameter data buffer */
            c1_parameter = (double *)get_memory_buffer(c1_parameter_type, 
                           CONSUMER, band_index, sca_index);
            if (!c1_parameter)
            {
                IAS_LOG_ERROR("Getting C1 parameter: %s for band %d SCA %d",
                get_memory_buffer_string(c1_parameter_type), 
                band_number, sca_number);
                return ERROR;
            }  
        }

        /* Get the VRP band data and A0 parameters if needed */
        switch(blob->per_frame_bias_flag)
        {
            case PER_FRAME_BIAS:
                /* Get the VRP band data */
                vrp_buffer = (uint16_t *)get_memory_buffer(IMAGE_L0R,
                    CONSUMER, band_attributes->vrp_band_index, sca_index);
                if (!vrp_buffer)
                {
                    IAS_LOG_ERROR("Getting VRP band IMAGE_L0R memory buffer");
                    return ERROR;
                }

                /* Set the a0_parameter memory buffer type */
                a0_parameter_type = 
                    a0_parameter_to_buffer_map[bias_spectral_type];

                /* Get the a0 parameter data buffer */
                a0_parameter = (double *)get_memory_buffer(a0_parameter_type, 
                    CONSUMER, band_index, sca_index);
                if (!a0_parameter)
                {
                    IAS_LOG_ERROR("Getting A0 parameter: %s for band %d SCA %d",
                        get_memory_buffer_string(a0_parameter_type), 
                        band_number, sca_number);
                    return ERROR;
                }  
                a0_value = *a0_parameter;
                break;

            case PER_DETECTOR_BIAS:
                if (blob->bias_source == A1_C1_ESTIMATION)
                {
                    /* Get the VRP band data */
                    vrp_buffer = (uint16_t *)get_memory_buffer(IMAGE_L0R,
                        CONSUMER, band_attributes->vrp_band_index, sca_index);
                    if (!vrp_buffer)
                    {
                        IAS_LOG_ERROR("Getting VRP band IMAGE_L0R memory "
                                      "buffer");
                        return ERROR;
                    }
                }
                break;

            default:
                IAS_LOG_ERROR("Unsupported blob->per_frame_bias_flag value");
                return ERROR;
        }

        /* Call the oli_bias_determination algorithm routine */
        status = oli_bias_determination(vrp_buffer, band_number, sca_number, 
                 number_detectors, number_lines, vrp_detectors, vrp_lines, 
                 blob->per_frame_bias_flag, a1_c1_estimate_flag, 
                 bias_spectral_type, bias_from_file, a1_parameter, 
                 c1_parameter, a0_value,
                 blob->pixel_mask[band_index][sca_index], report_information, 
                 perframe_bias);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Call oli_bias_determination band_number: %d, "
                          "SCA_number: %d", band_number, sca_number);
            return ERROR; 
        }

        /* Let consumers know that the per-frame buffer has been set */
        set_buffer_filled_flag( perframe_bias_type, band_index, sca_index );
    }

    /* Decrement all memory consumer counters */
    status = decrement_counters(band_index, sca_index);
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Decrementing memory buffer");
        return ERROR;
    }

    return status;
} 


/*----------------------------------------------------------------------------
NAME:  oli_bias_determination_preprocess  

PURPOSE: Increase all the memory consumer counts 

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int oli_bias_determination_preprocess
(
    BLOB *blob          /* I: Miscellaneous parameters */
)
{
    int index;
    int status;

    /* Increment the consumer counts for the memory buffers */
    for (index = 0; index < number_of_memory_buffers; index++)
    {
        status = increment_consumer_count(memory_buffers[index]);
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Incrementing %s consumer count",
                get_memory_buffer_string(memory_buffers[index]));
            return ERROR;
        }
    }

    return SUCCESS;
}  /* END oli_bias_determination_preprocess */
