/*****************************************************************************
 NAME:     api_oli_bias_removal

 PURPOSE:  Gather and package input data for the OLI bias removal algorithm,
           then execute the algorithm.

 RETURNS:  Integer status code of SUCCESS or ERROR
*****************************************************************************/
#include <string.h>

#include "ias_logging.h"
#include "ias_satellite_attributes.h"
#include "rps_local.h"


/* Define a list of memory buffers needed by the algorithm and the overall
   size of this data structure */
static MEMORY_BUFFER_TYPE memory_buffers[] =
{
    IMAGE_L1R,
    MM_TEMP_CORRECTION_FACTOR,
    MM_NON_PAN_CPF_BIAS,
    MM_PAN_EVEN_CPF_BIAS,
    MM_PAN_ODD_CPF_BIAS,
    MM_NON_PAN_PERFRAME_BIAS,
    MM_PAN_PERFRAME_BIAS
};

static int number_of_memory_buffers =
    sizeof(memory_buffers) / sizeof(MEMORY_BUFFER_TYPE);


/**************************************************************************
 NAME:     decrement_counters

 PURPOSE:  Internal routine to decrement the consumer counts of all required
           memory buffers

 RETURNS:  SUCCESS or ERROR
***************************************************************************/
static int decrement_counters
(
    int band_index,     /* I: */
    int sca_index       /* I: */
)
{
    int i;                                  /* Local loop counter */
    RPS_BAND_ATTRIBUTES *band_attributes = get_band_attributes( band_index );
    int vrp_index = band_attributes->vrp_band_index;
    int blind_index = band_attributes->blind_band_index;
    int status;
    if ( band_attributes->process_blind_band_flag == 0 )
        blind_index = -1;

    /* Decrement the consumer counts for the memory buffers */
    for (i = 0; i < number_of_memory_buffers; i++)
    {
        status = decrement_consumer_count(memory_buffers[i], band_index,
            sca_index);
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }
        if ( vrp_index >= 0 )
        {
            status = decrement_consumer_count(memory_buffers[i], vrp_index,
                sca_index);
            if ( status != SUCCESS )
            {
                IAS_LOG_ERROR("Decrementing memory buffer");
                return ERROR;
            }
        }
        if ( blind_index >= 0 )
        {
            status = decrement_consumer_count(memory_buffers[i], blind_index,
                sca_index);
            if ( status != SUCCESS )
            {
                IAS_LOG_ERROR("Decrementing memory buffer");
                return ERROR;
            }
        }
    }
    return SUCCESS;
}   /* END decrement_counters */



int api_oli_bias_removal
(
    BLOB *blob,         /* I: Various input parameters */
    int band_index,     /* I: */
    int sca_index,      /* I: */
    int algorithm_index /* I: Algorithm index */
)
{
    RPS_BAND_ATTRIBUTES *band_attributes = get_band_attributes(band_index);
    int band_number = band_attributes->band_number;
    int sca_number = sca_index + 1;

    int pan_flag = 0;                    /* PAN band flag (0=not PAN band,
                                            1=PAN band) */
    int status;

    float *image_buffer = NULL;           /* Per-frame image data buffer */

    const double *per_frame_bias_buffer = NULL; /* Per-frame bias data buffer */
    const double *cpf_bias_buffer_1 = NULL;     /* CPF bias data buffer #1 */
    const double *cpf_bias_buffer_2 = NULL;     /* CPF bias data buffer #2 */
    double *temp_cf_buffer = NULL;        /* Temperature correction factor
                                             data buffer */

    /* If we have a TIRS band or an OLI blind/VRP band, decrement all ]
       consumer counts and return SUCCESS */
    if ((band_attributes->spectral_type == IAS_SPECTRAL_THERMAL)
        || (band_attributes->band_classification != IAS_NORMAL_BAND))
    {
        status = decrement_counters(band_index, sca_index);
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }
        return SUCCESS;
    }

    /* Add the ID for temperature sensitivity correction for OLI bias removal 
       to the RPS processing level if blob->perform_temp_sens is 1.  */
    if (blob->perform_temp_sens)
    {
        if (set_rps_level(band_index, sca_index, 
            IAS_RPS_LEVEL_TEMP_SENSITIVITY_BIAS) == ERROR)
        {
            IAS_LOG_ERROR("Updating the RPS processing level");
            return ERROR;
        }
    }

    /* Add the ID for bias removal to the RPS processing level */
    if (set_rps_level(band_index, sca_index, 
            IAS_RPS_LEVEL_BIAS_REMOVAL) == ERROR)
    {
        IAS_LOG_ERROR("Updating RPS processing level");
        return ERROR;
    }

    /* This algorithm doesn't process OLI blind or VRP bands therefore the
        processing level *must* be set to skip */
    if ( band_attributes->vrp_band_index >= 0 )
    {
        if (set_rps_level(band_attributes->vrp_band_index, sca_index,
            IAS_RPS_LEVEL_SKIP) == ERROR)
        {
            IAS_LOG_ERROR("Updating RPS processing level");
            return ERROR;
        }
    }
    if ( band_attributes->blind_band_index >= 0 &&
         band_attributes->process_blind_band_flag )
    {
        if (set_rps_level(band_attributes->blind_band_index, sca_index,
            IAS_RPS_LEVEL_SKIP) == ERROR)
        {
            IAS_LOG_ERROR("Updating RPS processing level");
            return ERROR;
        }
    }

    /* Get the appropriate bias data */
    switch (band_attributes->spectral_type)
    {
        case IAS_SPECTRAL_VNIR:
        case IAS_SPECTRAL_SWIR:
        {
            if (blob->bias_source == CPF_BIAS && 
                blob->per_frame_bias_flag == 0)
            {
                cpf_bias_buffer_1 = (const double *)get_const_memory_buffer(
                    MM_NON_PAN_CPF_BIAS, band_index, sca_index);
                if (cpf_bias_buffer_1 == NULL)
                {
                    IAS_LOG_ERROR("Getting VNIR current launch bias for "
                        "band %d SCA %d", band_number, sca_number);
                    return ERROR;
                }
            }
            else
            {
                per_frame_bias_buffer = (const double *)get_memory_buffer(
                    MM_NON_PAN_PERFRAME_BIAS, CONSUMER, band_index, sca_index);
                if (per_frame_bias_buffer == NULL)
                {
                    IAS_LOG_ERROR("Getting VNIR per-frame bias for "
                        "band %d SCA %d", band_number, sca_number);
                    return ERROR;
                }
                /* Check the validity of the buffer */
                if ( get_buffer_filled_flag( MM_NON_PAN_PERFRAME_BIAS,
                    band_index, sca_index ) == 0 )
                {
                    IAS_LOG_ERROR("Attempting to use uninitialized buffer %s "
                        "for band %d SCA %d", get_memory_buffer_string(
                        MM_NON_PAN_PERFRAME_BIAS ), band_number, sca_number );
                    return ERROR;
                }
            }
            break;
        }
        case IAS_SPECTRAL_PAN:
        {
            /* Set the flag to indicate we're processing the PAN band */
            pan_flag = 1;

            if (blob->bias_source == CPF_BIAS && 
                blob->per_frame_bias_flag == 0)
            {
                cpf_bias_buffer_1 = (const double *)get_const_memory_buffer(
                    MM_PAN_ODD_CPF_BIAS, band_index, sca_index);
                cpf_bias_buffer_2 = (const double *)get_const_memory_buffer(
                    MM_PAN_EVEN_CPF_BIAS, band_index, sca_index);
                if ((cpf_bias_buffer_1 == NULL) || (cpf_bias_buffer_2 == NULL))
                {
                    IAS_LOG_ERROR("Getting PAN current launch bias for SCA %d",
                        sca_number);
                    return ERROR;
                }
            }
            else
            {
                per_frame_bias_buffer = (const double *)get_memory_buffer(
                    MM_PAN_PERFRAME_BIAS, CONSUMER, band_index, sca_index);
                if (per_frame_bias_buffer == NULL)
                {
                    IAS_LOG_ERROR("Getting PAN per-frame bias in SCA %d",
                        sca_number);
                    return ERROR;
                }
                /* Check the validity of the buffer */
                if ( get_buffer_filled_flag( MM_PAN_PERFRAME_BIAS, band_index,
                    sca_index ) == 0 )
                {
                    IAS_LOG_ERROR("Attempting to use uninitialized buffer %s "
                        "for band %d SCA %d", get_memory_buffer_string(
                        MM_PAN_PERFRAME_BIAS ), band_number, sca_number );
                    return ERROR;
                }
            }
            break;
        }
        default:
        {
            IAS_LOG_ERROR("Invalid spectral type %d for band %d SCA %d",
                band_attributes->spectral_type, band_number, sca_index);
            return ERROR;
        }
    }

    /* Get the image data for the current band/SCA */
    image_buffer = (float *)get_memory_buffer(IMAGE_L1R, CONSUMER, band_index,
        sca_index);
    if (image_buffer == NULL)
    {
        IAS_LOG_ERROR("Getting L1R image data buffer for band %d, SCA %d",
            band_number, sca_number);
        return ERROR;
    }

    /* Finally, get the temperature correction factor coefficients for the
       current band/SCA IF temperature sensitivity correction has been
       selected */
    if (blob->perform_temp_sens)
    {
        temp_cf_buffer = (double *)get_memory_buffer(
            MM_TEMP_CORRECTION_FACTOR, CONSUMER, band_index, sca_index);
        if (temp_cf_buffer == NULL)
        {
            IAS_LOG_ERROR("Getting temperature correction factors for "
                    "band %d, SCA %d", band_number, sca_number);
            return ERROR;
        }

        /* Make sure the temperature sensitivity correction factor buffer is 
           valid */
        if ( get_buffer_filled_flag( MM_TEMP_CORRECTION_FACTOR,
             band_index, sca_index ) == 0 )
        {
            IAS_LOG_ERROR("Attempting to use uninitialized buffer %s for band "
                "%d SCA %d", get_memory_buffer_string(
                MM_TEMP_CORRECTION_FACTOR ), band_number, sca_index + 1);
            return ERROR;
        }
    }

    /* Call the algorithm */
    status = oli_bias_removal(image_buffer, cpf_bias_buffer_1,
        cpf_bias_buffer_2, per_frame_bias_buffer, blob->bias_source,
        blob->per_frame_bias_flag, pan_flag, band_attributes->number_detectors,
        band_attributes->number_lines, temp_cf_buffer);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Removing bias for band %d SCA %d", band_number,
            sca_number);
        return ERROR;
    }

    /* Set the bias source and bias applied flag in the band metadata */
    if ( sca_index == 0 )
    {
        switch ( blob->bias_source )
        {
            case PRE_ACQUISITION_SHUTTER :
                strcpy( blob->l1r_band_metadata[band_index]->bias_source,
                    "BPF, Pre");
                break;
            case POST_ACQUISITION_SHUTTER :
                strcpy( blob->l1r_band_metadata[band_index]->bias_source,
                    "BPF, Post");
                break;
            case MEAN_ACQUISITION_SHUTTER :
                strcpy( blob->l1r_band_metadata[band_index]->bias_source,
                    "BPF, Mean");
                break;
            case CPF_BIAS :
                strcpy( blob->l1r_band_metadata[band_index]->bias_source,
                    "CPF");
                break;
            case A1_C1_ESTIMATION :
                strcpy( blob->l1r_band_metadata[band_index]->bias_source,
                    "A1 C1 est.");
                break;
        }
        if ( blob->per_frame_bias_flag == 1 )
        {
            strcat( blob->l1r_band_metadata[band_index]->bias_source,
                ", Per-frame" );
        }
        if ( blob->perform_temp_sens )
        {
            blob->l1r_band_metadata[band_index]->
                bias_temp_sensitivity_correction_applied = 1;
        }

        /* Set the bias correction flag */
        blob->l1r_band_metadata[band_index]->bias_correction_applied = 1;
    }

    /* Decrement memory usage counters */
    status = decrement_counters(band_index, sca_index);
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Decrementing memory buffer");
        return ERROR;
    }

    return SUCCESS;
}   /* END api_oli_bias_removal */



/**************************************************************************
 NAME:     oli_bias_removal_preprocess

 PURPOSE:  Sets up memory management for the OLI bias removal algorithm

 RETURNS:  Integer status code of SUCCESS or ERROR
***************************************************************************/
int oli_bias_removal_preprocess
(
    BLOB *blob          /* I: Miscellaneous parameters
                           O: Updated bias_correction_applied flag */
)
{
    int i;
    int status;

    /* Increment the consumer counts for the memory buffers */
    for (i = 0; i < number_of_memory_buffers; i++)
    {
        status = increment_consumer_count(memory_buffers[i]);
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Incrementing %s consumer count",
                get_memory_buffer_string(memory_buffers[i]));
            return ERROR;
        }
    }

    return SUCCESS;
}  /* END oli_bias_removal_preprocess */
