/*----------------------------------------------------------------------------
NAME:    api_reflectance_conversion

PURPOSE: Gather data needed for the reflectance conversion routine and call it.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
#include <stdlib.h>
#include <string.h>
#include "ias_logging.h"
#include "ias_const.h"
#include "rps_local.h"
#include "ias_satellite_attributes.h"
#include "ias_structures.h"
#include "ias_miscellaneous.h"
#include "ias_math.h"

/* Define a list of memory buffers needed by the algorithm */
static MEMORY_BUFFER_TYPE memory_buffers[] =
{
    IMAGE_L1R,
    MM_REFLECTANCE_CONVERSION_COEFFICIENTS
};
static int number_of_memory_buffers = sizeof(memory_buffers) /
    sizeof(MEMORY_BUFFER_TYPE);

/*----------------------------------------------------------------------------
 NAME:    decrement_counters

 PURPOSE: Internal routine to decrement the consumer counts of all required
          memory buffers

 RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
static int decrement_counters
(
    int band_index,     /* I: */
    int sca_index       /* I: */
)
{
    int i;                                  /* Local loop counter */
    RPS_BAND_ATTRIBUTES *band_attributes = get_band_attributes( band_index );
    int vrp_index = band_attributes->vrp_band_index;
    int blind_index = band_attributes->blind_band_index;
    int status;
    if ( ! band_attributes->process_blind_band_flag )
        blind_index = -1;

    /* Decrement the consumer counts for the memory buffers */
    for (i = 0; i < number_of_memory_buffers; i++)
    {
        status = decrement_consumer_count(memory_buffers[i], band_index,
            sca_index);
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }
        if ( vrp_index >= 0 )
        {
            status = decrement_consumer_count(memory_buffers[i], vrp_index,
                sca_index);
            if ( status != SUCCESS )
            {
                IAS_LOG_ERROR("Decrementing memory buffer");
                return ERROR;
            }
        }
        if ( blind_index >= 0 )
        {
            status = decrement_consumer_count(memory_buffers[i], blind_index,
                sca_index);
            if ( status != SUCCESS )
            {
                IAS_LOG_ERROR("Decrementing memory buffer");
                return ERROR;
            }
        }
    }
    return SUCCESS;
}   /* END decrement_counters */


int api_reflectance_conversion
(
    BLOB *blob,         /* I: Miscellaneous parameters */
    int   band_index,   /* I: */
    int   sca_index,    /* I: */
    int   algo_index    /* I: Algorithm index */
)
{
    RPS_BAND_ATTRIBUTES *band_attributes = get_band_attributes(band_index);
    int num_lines = band_attributes->number_lines;
    int num_samps = band_attributes->number_detectors;
    float *data_buffer;              /* Level 1 data buffer */
    int status;
    double earth_sun_distance;       /* earth sun distance in astronomy unit
                                        from NOVAS library calculation */
    const double *rad_to_reflect_conv_coeff; /* band level radiance to
                                                reflectance conversion
                                                coefficient */
    double reflect_to_rad_conv_coeff; /* band level reflectance to radiance
                                         conversion coefficient */   

    /* This algorithm doesn't process OLI blind or VRP bands therefore the
        processing level *must* be set to skip */
    if ( band_attributes->vrp_band_index >= 0 )
    {
        if (set_rps_level(band_attributes->vrp_band_index, sca_index,
            IAS_RPS_LEVEL_SKIP) == ERROR)
        {
            IAS_LOG_ERROR("Updating RPS processing level");
            return ERROR;
        }
    }
    if ( band_attributes->blind_band_index >= 0 &&
         band_attributes->process_blind_band_flag )
    {
        if (set_rps_level(band_attributes->blind_band_index, sca_index,
            IAS_RPS_LEVEL_SKIP) == ERROR)
        {
            IAS_LOG_ERROR("Updating RPS processing level");
            return ERROR;
        }
    }

    /* If called with a thermal band return */
    if ( band_attributes->spectral_type == IAS_SPECTRAL_THERMAL )
    {
        status = decrement_counters( band_index, sca_index );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }
        return SUCCESS;
    }

    /* Add the ID for reflectance conversion to the RPS processing level */
    if (set_rps_level(band_index, sca_index, 
            IAS_RPS_LEVEL_REFLECTANCE) == ERROR)
    {
        IAS_LOG_ERROR("Updating RPS processing level");
        return ERROR;
    }

    /* Get the earth-sun distance, since the value has been calculated in
       intialize_io_routines call to get_earth_sun_distance, NULL is used for
       date_acquired pointer */
    status = get_earth_sun_distance(NULL, &earth_sun_distance);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Getting earth-sun distance");
        return ERROR;
    }

    /* Note: It isn't necessary to decrement the usage of the memory buffer
       on error conditions. The main thread will clean up allocated buffers
       when error conditions dictate */

    /* Get the  band level radiance to reflectance conversion coefficient
       from CPF */
    rad_to_reflect_conv_coeff = (const double *)
        get_const_memory_buffer(MM_REFLECTANCE_CONVERSION_COEFFICIENTS, 
        band_index, sca_index);
    if (!rad_to_reflect_conv_coeff)
    {
        IAS_LOG_ERROR("Getting radiance to reflectance "
                      "conversion coefficient");
        return ERROR;
    }

    /* Get the data buffer for IMAGE_L1R */
    data_buffer = (float *)get_memory_buffer(IMAGE_L1R, CONSUMER,
        band_index, sca_index);
    if ( !data_buffer )
    {
        IAS_LOG_ERROR("Getting image memory buffer");
        return ERROR;
    }

    /* Call the reflectance conversion algorithm */
    status = reflectance_conversion(data_buffer, num_samps, num_lines,
        earth_sun_distance, rad_to_reflect_conv_coeff[0],
        &reflect_to_rad_conv_coeff);
    if (status !=SUCCESS)
    {
        IAS_LOG_ERROR("Calling reflectance_conversion algorithm");
        return ERROR;    
    }

    /* Store the reflectance to radiance conversion coefficient in the
       L1R metadata */
    blob->l1r_band_metadata[band_index][0].
        reflectance_to_radiance_coefficient = reflect_to_rad_conv_coeff;

    /* Tell memory management that the data was used */
    status = decrement_counters( band_index, sca_index );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Decrementing memory buffer");
        return ERROR;
    }

    return SUCCESS;
}  /* END api_reflectance_conversion */


/*----------------------------------------------------------------------------
NAME:    reflectance_conversion_preprocess

PURPOSE: Registers memory for the algorithm and sets the sun earth distance
         in the blob

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int reflectance_conversion_preprocess
(
    BLOB *blob          /* I: Miscellaneous parameters */
)
{
    int i;
    int status;

    /* Increment the consumer counts for the memory buffers */
    for ( i = 0; i < number_of_memory_buffers; i++ )
    {
        status = increment_consumer_count( memory_buffers[i] );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Incrementing %s consumer count",
                get_memory_buffer_string( memory_buffers[i] ));
            return ERROR;
        }
    }

    /* Set the L1R file metadata */
    blob->l1r_file_metadata.reflectance_conversion_applied = 1;

    return SUCCESS;
}  /* END reflectance_conversion_preprocess */
