/*****************************************************************************
 NAME:     api_residual_striping_correction

 PURPOSE:  Gather the input data for the residual striping correction
           algorithm and then execute the algorithm.

 RETURNS:  Integer status code of SUCCESS or ERROR
*****************************************************************************/
#include "ias_logging.h"
#include "rps_local.h"

/* Define a list of memory buffers needed by the algorithm and the overall
   size of this data structure */
static MEMORY_BUFFER_TYPE memory_buffers[] =
{
    IMAGE_L1R,
    MM_SCENE_STRIPING_CORRECTION_MATRIX
};

static int number_of_memory_buffers =
    sizeof(memory_buffers) / sizeof(MEMORY_BUFFER_TYPE);

/*----------------------------------------------------------------------------
NAME:    decrement_counters

PURPOSE:  Internal routine to decrement the consumer counts of all required
          memory buffers

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
static int decrement_counters
(
    int band_index,     /* I: 0-based band index */
    int sca_index       /* I: 0-based SCA index */
)
{
    RPS_BAND_ATTRIBUTES *band_attributes = get_band_attributes( band_index );
    int index;
    int status;

    /* Decrement the consumer counts for the memory buffers */
    for ( index = 0; index < number_of_memory_buffers; index++ )
    {
        status = decrement_consumer_count( memory_buffers[index], band_index,
            sca_index );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }

        if ( band_attributes->vrp_band_number > 0 )
        {
            status = decrement_consumer_count( memory_buffers[index],
                band_attributes->vrp_band_index, sca_index );
            if ( status != SUCCESS )
            {
                IAS_LOG_ERROR("Decrementing memory buffer");
                return ERROR;
            }
        }

        if ( band_attributes->blind_band_number > 0
            && band_attributes->process_blind_band_flag )
        {
            status = decrement_consumer_count( memory_buffers[index],
                band_attributes->blind_band_index, sca_index );
            if ( status != SUCCESS )
            {
                IAS_LOG_ERROR("Decrementing memory buffer");
                return ERROR;
            }
        }
    }
    return SUCCESS;
}  /* END decrement_counters */

/**************************************************************************
 NAME:     api_residual_striping_correction

 PURPOSE:  Main API call

 RETURNS:  Integer status code of SUCCESS or ERROR
***************************************************************************/
int api_residual_striping_correction
(
    BLOB *blob,         /* I: Various input parameters */
    int band_index,     /* I: 0-based band index */
    int sca_index,      /* I: 0-based SCA index */
    int algorithm_index /* I: Algorithm index */
)
{
    RPS_BAND_ATTRIBUTES *band_attributes = get_band_attributes(band_index);
    int band_number = band_attributes->band_number;
    int sca_number = sca_index + 1;
    float *data_buffer = NULL;          /* Level 1 data buffer */
    double *striping_buffer = NULL;     /* Striping correction matrix buffer */
    int num_lines = band_attributes->number_lines;
    int num_detectors = band_attributes->number_detectors;
    int status;

    /* Add the ID for residual striping to the RPS processing level */
    if (set_rps_level(band_index, sca_index, 
            IAS_RPS_LEVEL_RESIDUAL_STRIPING) == ERROR)
    {
        IAS_LOG_ERROR("Updating the RPS processing level");
        return ERROR;
    }

    /* This algorithm doesn't process OLI blind or VRP bands therefore the
        processing level *must* be set to skip */
    if ( band_attributes->vrp_band_index >= 0 )
    {
        if (set_rps_level(band_attributes->vrp_band_index, sca_index,
            IAS_RPS_LEVEL_SKIP) == ERROR)
        {
            IAS_LOG_ERROR("Updating RPS processing level");
            return ERROR;
        }
    }
    if ( band_attributes->blind_band_index >= 0 &&
         band_attributes->process_blind_band_flag )
    {
        if (set_rps_level(band_attributes->blind_band_index, sca_index,
            IAS_RPS_LEVEL_SKIP) == ERROR)
        {
            IAS_LOG_ERROR("Updating RPS processing level");
            return ERROR;
        }
    }

    /* Get the image data buffer */
    data_buffer = (float *)get_memory_buffer(IMAGE_L1R, CONSUMER, band_index,
        sca_index);
    if (data_buffer == NULL)
    {
        IAS_LOG_ERROR("Getting image memory buffer for band %d SCA %d",
            band_number, sca_number);
        return ERROR;
    }

    /* Get the striping correction matrix buffer */
    striping_buffer = (double *)get_memory_buffer(
                      MM_SCENE_STRIPING_CORRECTION_MATRIX, 
                      CONSUMER, band_index, sca_index);
    if (striping_buffer == NULL)
    {
        IAS_LOG_ERROR("Getting striping correction matrix buffer for band %d "
                      "SCA %d", band_number, sca_number);
        return ERROR;
    }
    /* Make sure the buffer is valid */
    if ( get_buffer_filled_flag( MM_SCENE_STRIPING_CORRECTION_MATRIX,
        band_index, sca_index ) == 0 )
    {
        IAS_LOG_ERROR("Attempting to use uninitialized buffer %s for band "
            "%d SCA %d", get_memory_buffer_string(
            MM_SCENE_STRIPING_CORRECTION_MATRIX ), band_number, sca_number);
        return ERROR;
    }

    /* Call the residual striping correction algorithm */
    residual_striping_correction(data_buffer, striping_buffer, 
                                 num_detectors, num_lines);

    /* Decrement memory usage counters */
    status = decrement_counters( band_index, sca_index );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Decrementing memory buffer");
        return ERROR;
    }

    return SUCCESS;
}   /* END api_residual_striping_correction */



/**************************************************************************
 NAME:     api_residual_striping_correction_preprocess

 PURPOSE:  Sets up memory management for the residual striping correction 
           algorithm

 RETURNS:  Integer status code of SUCCESS or ERROR
***************************************************************************/
int residual_striping_correction_preprocess
(
    BLOB *blob          /* I: Miscellaneous parameters */
)
{
    int i;              /* Loop counter */
    int status;         /* Return variable */

    /* Increment the consumer counts for the memory buffers */
    for (i = 0; i < number_of_memory_buffers; i++)
    {
        status = increment_consumer_count(memory_buffers[i]);
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Incrementing %s consumer count",
                get_memory_buffer_string(memory_buffers[i]));
            return ERROR;
        }
    }

    /* Set the L1R file metadata */
    blob->l1r_file_metadata.residual_striping_correction_applied = 1;

    return SUCCESS;
}  /* END residual_striping_correction_preprocess */
