/*----------------------------------------------------------------------------
NAME:    api_saturated_pixel_characterization

PURPOSE: Gather data needed for the saturated pixel characterization
         routine and call it.  If database storage is desired, submit the
         results to the routine to store them. If write report is desired,
         the output report is written.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
#include <stdlib.h>
#include <string.h>

#include "ias_logging.h"
#include "ias_threadpool.h"
#include "ias_const.h"
#include "ias_cpf.h"
#include "rps_local.h"
#include "ias_l0r.h"
#include "ias_satellite_attributes.h"

/* Define a list of memory buffers needed by the algorithm */
static int number_of_memory_buffers;
static MEMORY_BUFFER_TYPE memory_buffers[] =
    {
        IMAGE_L0R,
        MM_ANALOG_HIGH_SAT,
        MM_ANALOG_LOW_SAT,
        MM_DIGITAL_HIGH_SAT,
        MM_DIGITAL_LOW_SAT
    };

/**************************************************************************
 NAME:     decrement_counters

 PURPOSE:  Internal routine to decrement the consumer counts of all required
           memory buffers

 RETURNS:  SUCCESS or ERROR
***************************************************************************/
static int decrement_counters
(
    int band_index,     /* I: Current band */
    int sca_index       /* I: SCA routine */
)
{
    int i;              /* Local loop counter */
    int status;

    /* Decrement the consumer counts for the memory buffers */
    for (i = 0; i < number_of_memory_buffers; i++)
    {
        status = decrement_consumer_count(memory_buffers[i], band_index,
            sca_index);
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }
    }
    return SUCCESS;
}


/*------------------------------------------------------------------------
NAME:   saturated_pixel_characterization 

PURPOSE: Internal routine for gathering data for the algorithm and call it

RETURNS: SUCCESS or ERROR
-------------------------------------------------------------------------*/
static int saturated_pixel_characterization
(
    BLOB *blob,         /* I: Miscellaneous parameters */
    int   band_index,   /* I: */
    int   sca_index     /* I: */
)
{
    uint16_t *data_buffer;    /* Level 0 data buffer */
    RPS_BAND_ATTRIBUTES *band_attributes = get_band_attributes( band_index );
    int band_number = band_attributes->band_number;
    int num_lines = band_attributes->number_lines;
    int num_samps = band_attributes->number_detectors;
    IAS_RPS_DB_SATURATED_CHARACTERIZATION_RESULTS *db_results = NULL;
    IAS_REPORT *report_information = NULL;
    int status;
    const int *digital_low_saturated = NULL; /* I - Digital low saturated level
                                                    from CPF */
    const int *digital_high_saturated = NULL;/* I - Digital high saturated
                                                    level from CPF */
    const int *analog_low_saturated = NULL;  /* I - Analog low saturated level 
                                                    from CPF */
    const int *analog_high_saturated = NULL; /* I - Analog high saturated level
                                                    from CPF */

    /* Get the analog high saturated pixel level from CPF for normal band */
    analog_high_saturated = (int *)
        get_const_memory_buffer(MM_ANALOG_HIGH_SAT, band_index, sca_index);

    /* Get the analog low saturated pixel level from CPF for normal band */
    analog_low_saturated = (int *)
        get_const_memory_buffer(MM_ANALOG_LOW_SAT, band_index, sca_index);

    /* Get the digital high saturated pixel level from CPF for normal band */
    digital_high_saturated = (int *)
        get_const_memory_buffer(MM_DIGITAL_HIGH_SAT, band_index, sca_index);

    /* Get the analog high saturated pixel level from CPF for normal band */
    digital_low_saturated = (int *)
        get_const_memory_buffer(MM_DIGITAL_LOW_SAT, band_index, sca_index);

    /* Set the reporting and database structures */
    if ( blob->generate_report )
        report_information  = &blob->report_information;

    /* Allocate memory for the database if needed and populate its members.
       Dynamic allocation is used here for threading purposes. */
    if ( blob->store_in_database )
    {
        db_results = ias_rps_db_allocate_saturated_results( num_samps );
        if ( !db_results )
        {
            IAS_LOG_ERROR("Allocating the saturated result structure");
            return ERROR;
        }

        /* Set the common values that are known here */
        db_results->dbase = blob->database_information.db;
        strncpy(db_results->rad_common_id, 
            blob->database_information.rad_common_id, 
            sizeof(db_results->rad_common_id));
        db_results->rad_common_id[sizeof(db_results->rad_common_id) - 1] = '\0';
        db_results->band = band_number;
        db_results->sca = sca_index + 1;
        db_results->num_detectors = num_samps;
    }

    /* Get the data buffer */
    data_buffer = (uint16_t *)get_memory_buffer( IMAGE_L0R, CONSUMER,
        band_index, sca_index );
    if ( !data_buffer )
    {
        IAS_LOG_ERROR("Getting image memory buffer");
        ias_rps_db_free_saturated_results( db_results );
        return ERROR;
    }
    /* Note: It isn't necessary to decrement the usage of the memory buffer
       on error conditions from this point on.  Since the memory may not be
       freed.  The main thread will clean up allocated buffers when error
       conditions dictate.  It doesn't hurt but it isn't necessary. */

    status = ias_rps_saturated_pixel_characterization( data_buffer,
        band_number, sca_index + 1, num_samps, num_lines,
        digital_low_saturated, digital_high_saturated, analog_low_saturated,
        analog_high_saturated, blob->pixel_mask[band_index][sca_index],
        report_information, db_results);
    if ( status )
    {
        IAS_LOG_ERROR("From ias_rps_saturated_pixel_characterization"
            " band: %d, SCA %d", band_number, sca_index + 1 );
        ias_rps_db_free_saturated_results( db_results );
        return ERROR;
    }

    if ( blob->store_in_database )
    {
        status = ias_threadpool_add_to_work_queue( blob->threadpool,
            ias_rps_store_saturated_pixels_in_database, db_results );
        if ( status )
        {
            IAS_LOG_ERROR("Submitting saturated pixel to work queue "
                "band: %d, SCA %d", band_number, sca_index + 1 );
            /* Before freeing memory, ensure there is more than one thread
               running.  A double free of the memory may occur if only one
               thread is running and an internal database error occurs. */
            if ( ias_threadpool_get_thread_count( blob->threadpool ) > 1 )
                ias_rps_db_free_saturated_results( db_results );
            return ERROR;
        }
    }

    /* Release the memory buffers */
    status = decrement_counters(band_index, sca_index);
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Decrementing memory buffer");
        return ERROR;
    }

    return status;
}  /* END saturated_pixel_characterization */


/*----------------------------------------------------------------------------
NAME:    api_saturated_pixel_characterization

PURPOSE: External API to call the internal routine.  It will call the 
         internal routine with the current band then with its associated VRP
         and blind band if needed.

RETURNS: SUCCESS or ERROR
--------------------------------------------------------------------------*/
int api_saturated_pixel_characterization
(
    BLOB *blob,         /* I: Miscellaneous parameters */
    int   band_index,   /* I: */
    int   sca_index,    /* I: */
    int   algo_index    /* I: Algorithm index */
)
{
    int status;
    RPS_BAND_ATTRIBUTES *band_attributes = get_band_attributes( band_index );

    status = saturated_pixel_characterization( blob, band_index, sca_index );
    if ( status != SUCCESS )
        return ERROR;

    if ( band_attributes->vrp_band_number > 0 )
    {
        status = saturated_pixel_characterization( blob,
            band_attributes->vrp_band_index, sca_index );
        if ( status != SUCCESS )
            return ERROR;
    }

    if ( band_attributes->blind_band_number > 0 
        && band_attributes->process_blind_band_flag )
    {
        status = decrement_counters( band_attributes->blind_band_index,
            sca_index );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }
    }

    return SUCCESS;
}  /* END api_saturated_pixel_characterization */

/*-------------------------------------------------------------------------
NAME:    saturated_pixel_characterization_preprocess

PURPOSE: Registers memory for the algorithm

RETURNS: SUCCESS or ERROR
--------------------------------------------------------------------------*/
int saturated_pixel_characterization_preprocess
(
    BLOB *blob          /* I: Miscelaneous parameters */
)
{
    int i;
    int status;

    number_of_memory_buffers = sizeof(memory_buffers) /
        sizeof(MEMORY_BUFFER_TYPE);

    /* Increment the consumer counts for the memory buffers */
    for ( i = 0; i < number_of_memory_buffers; i++ )
    {
        status = increment_consumer_count( memory_buffers[i] );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Incrementing %s consumer count",
                get_memory_buffer_string( memory_buffers[i] ));
            return ERROR;
        }
    }

    /* Delete any pre-existing records for this work order if database
       storage was requested */
    if ( blob->store_in_database )
    {
        status = delete_db_table_records( blob->database_information,
            "saturated_pixels");
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Returned from delete_db_table_data");
            return ERROR;
        }
    }

    return SUCCESS;
}  /* END saturated_pixel_characterization_preprocess */
