/*----------------------------------------------------------------------------
NAME:    api_saturated_pixel_replacement

PURPOSE: Replace low and high saturated pixels with the low and high
         radiance saturation value from the CPF.  Use the Label Mask 
         to determine if a pixel is saturated.  Flags are used 
         to determine which pixels will be replaced.  Options are 
         Low and High, Analog and Digital Saturated pixels.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
#include <stdlib.h>
#include <string.h>
#include "ias_logging.h"
#include "rps_local.h"

/* Define a list of memory buffers needed by the algorithm */
MEMORY_BUFFER_TYPE memory_buffers[] =
{
    IMAGE_L1R,
    MM_HIGH_RADIANCE_SAT,
    MM_LOW_RADIANCE_SAT
};

int number_of_memory_buffers = sizeof(memory_buffers) /
        sizeof(MEMORY_BUFFER_TYPE);

/**************************************************************************
 NAME:     decrement_counters

 PURPOSE:  Internal routine to decrement the consumer counts of all required
           memory buffers

 RETURNS: SUCCESS or ERROR
***************************************************************************/
static int decrement_counters
(
    int band_index,     /* I: */
    int sca_index       /* I: */
)
{
   /* Get the attributes for the current band */
    RPS_BAND_ATTRIBUTES *band_atts = get_band_attributes(band_index);

    int index;                     /* Loop counter for consumer counts */
    int status;


    /* Decrement the consumer counts for the VRP and blind bands
       to avoid memory/resource leaks */
    for (index = 0; index < number_of_memory_buffers; index++)
    {
        status = decrement_consumer_count(memory_buffers[index], band_index,
            sca_index);
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }

        if (band_atts->vrp_band_number > 0)
        {
            status = decrement_consumer_count(memory_buffers[index],
                band_atts->vrp_band_index, sca_index);
            if ( status != SUCCESS )
            {
                IAS_LOG_ERROR("Decrementing memory buffer");
                return ERROR;
            }
        }
        if ((band_atts->blind_band_number > 0)
            && (band_atts->process_blind_band_flag))
        {
            status = decrement_consumer_count(memory_buffers[index],
                band_atts->blind_band_index, sca_index);
            if ( status != SUCCESS )
            {
                IAS_LOG_ERROR("Decrementing memory buffer");
                return ERROR;
            }
        }
    }
    return SUCCESS;
}   /* END decrement_counters */

/****************************************************************************
Name:  api_saturated_pixel_replacement

PURPOSE:  call the algorithm with the blob settings
****************************************************************************/
int api_saturated_pixel_replacement
(
    BLOB *blob,         /* I: Miscellaneous parameters */
    int band_index,     /* I: Array Index for band */
    int sca_index,      /* I: Array Index for SCA */
    int algo_index      /* I: Algorithm index */
)
{
    float *data_buffer;                   /* Level 1 data buffer */
    RPS_BAND_ATTRIBUTES *band_attributes = get_band_attributes( band_index );
    int num_lines = band_attributes->number_lines;
    int num_samps = band_attributes->number_detectors;

    int digital_high_flag;          /* Replace high digital saturated pixels */
    int digital_low_flag;           /* Replace low digital saturated pixels */
    int analog_high_flag;           /* Replace high analog saturated pixels */
    int analog_low_flag;            /* Replace low analog saturated pixels */
    const double *high_ptr = NULL;
    const double *low_ptr = NULL;
    double high_replacement = 0;    /* Value to replace high saturated pixels */
    double low_replacement = 0;     /* Value to replace low saturated pixels */
    int status;

    digital_high_flag = blob->digital_high_saturation_flag;
    digital_low_flag = blob->digital_low_saturation_flag;
    analog_high_flag = blob->analog_high_saturation_flag;
    analog_low_flag = blob->analog_low_saturation_flag;

    /* Note: It isn't necessary to decrement the usage of the memory buffer
       on error conditions from this point on.  Since the memory may not be
       freed.  The main thread will clean up allocated buffers when error
       conditions dictate.  It doesn't hurt but it isn't necessary. */

    /* Add the ID for saturated pixel to the RPS processing level */
    if (set_rps_level(band_index, sca_index, 
            IAS_RPS_LEVEL_SATURATED_PIXEL) == ERROR)
    {
        IAS_LOG_ERROR("Updating the RPS processing level");
        return ERROR;
    }

    /* This algorithm doesn't process OLI blind or VRP bands therefore the
        processing level *must* be set to skip */
    if ( band_attributes->vrp_band_index >= 0 )
    {
        if (set_rps_level(band_attributes->vrp_band_index, sca_index,
            IAS_RPS_LEVEL_SKIP) == ERROR)
        {
            IAS_LOG_ERROR("Updating RPS processing level");
            return ERROR;
        }
    }
    if ( band_attributes->blind_band_index >= 0 &&
         band_attributes->process_blind_band_flag )
    {
        if (set_rps_level(band_attributes->blind_band_index, sca_index,
            IAS_RPS_LEVEL_SKIP) == ERROR)
        {
            IAS_LOG_ERROR("Updating RPS processing level");
            return ERROR;
        }
    }

    /* Get the data buffer */
    data_buffer = (float *)get_memory_buffer(IMAGE_L1R, CONSUMER, band_index,
        sca_index);
    if (!data_buffer)
    {
        IAS_LOG_ERROR("Getting image memory buffer");
        return ERROR;
    }

    /* Get the cpf replacement value if necessary */

    /* If either high flag is set, get the high radiance value from cpf */
    if ((digital_high_flag) || (analog_high_flag))
    {
        high_ptr = (double *)get_const_memory_buffer(MM_HIGH_RADIANCE_SAT, 
                                                    band_index, sca_index);
        if (!high_ptr)
        {
            IAS_LOG_ERROR("Could not read the High Radiance Saturated pixel "
                        "values");
            return ERROR;
        }
        high_replacement = *high_ptr;
    }

    /* If either low flag is set, get the low radiance value from cpf */
    if ((digital_low_flag) || (analog_low_flag))
    {
        low_ptr = (double *)get_const_memory_buffer(MM_LOW_RADIANCE_SAT, 
                                                    band_index, sca_index);
        if (!low_ptr)
        {
            IAS_LOG_ERROR("Could not read the Low Radiance Saturated pixel "
                        "values");
            return ERROR;
        }
        low_replacement = *low_ptr;
    }

    if (saturated_pixel_replacement(data_buffer, num_samps, num_lines, 
                blob->pixel_mask[band_index][sca_index],
                (float)high_replacement, (float)low_replacement,
                digital_high_flag, digital_low_flag,
                analog_high_flag, analog_low_flag) != SUCCESS)
    {
        IAS_LOG_ERROR("Call to saturated_pixel_replacement failed "
                "for band number %d SCA number %d",
                band_attributes->band_number, sca_index+1);
        return ERROR;
    }

    /* Decrement the counters */
    status = decrement_counters(band_index, sca_index);
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Decrementing memory buffer");
        return ERROR;
    }

    return SUCCESS;
}  /* END api_saturated_pixel_replacement */


/*----------------------------------------------------------------------------
NAME:    saturated_pixel_replacement_preprocess

PURPOSE: Do preprocessing for the algorithm by incrementing the consumer
         counts.  

RETURNS: SUCCESS or ERROR

-----------------------------------------------------------------------------*/
int saturated_pixel_replacement_preprocess
(
    BLOB *blob          /* I: Miscellaneous parameters */
)
{
    int i;
    int status;

    /* Increment the consumer counts for the memory buffers */
    for ( i = 0; i < number_of_memory_buffers; i++ )
    {
        status = increment_consumer_count( memory_buffers[i] );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Incrementing %s consumer count",
                get_memory_buffer_string( memory_buffers[i] ));
            return ERROR;
        }
    }

    /* Set the L1R file metadata */
    blob->l1r_file_metadata.saturated_pixel_correction_applied = 1;

    return SUCCESS;
}  /* END replace_saturated_pixels_preprocess */
