/***************************************************************************
 NAME:      api_sca_discontinuity_correction

 PURPOSE: Gather data needed for the sca_discontinuity_correction application 
          routine and call the routine.
  
 RETURNS: SUCCESS or ERROR
****************************************************************************/
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include "ias_logging.h"
#include "ias_satellite_attributes.h"
#include "rps_local.h"

static FILE *create_report_file
(
    IAS_REPORT *report_info          /* I: IAS report structure pointer */
);

static int close_report_file
(
     FILE *report_file               /* Output file pointer */
);

/* Define a list of memory buffers needed by the algorithm and the overall
   size of this data structure */
static MEMORY_BUFFER_TYPE memory_buffers[] =
{
    IMAGE_L1R,
    MM_SCA_CORRECTION_FACTOR_CPF
};
static int number_of_memory_buffers = sizeof(memory_buffers) /
        sizeof(MEMORY_BUFFER_TYPE);

/*----------------------------------------------------------------------------
NAME:    decrement_counters

PURPOSE:  Internal routine to decrement the consumer counts of all required
          memory buffers

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
static int decrement_counters
(
    int band_index,     /* I: 0-based band index */
    int sca_index       /* I: 0-based SCA index */
)
{
    RPS_BAND_ATTRIBUTES *band_attributes = get_band_attributes( band_index );
    int index;
    int status;

    /* Decrement the consumer counts for the memory buffers */
    for ( index = 0; index < number_of_memory_buffers; index++ )
    {
        status = decrement_consumer_count( memory_buffers[index], band_index,
            sca_index );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }

        if ( band_attributes->vrp_band_number > 0 )
        {
            status = decrement_consumer_count( memory_buffers[index],
                band_attributes->vrp_band_index, sca_index );
            if ( status != SUCCESS )
            {
                IAS_LOG_ERROR("Decrementing memory buffer");
                return ERROR;
            }
        }

        if ( band_attributes->blind_band_number > 0
            && band_attributes->process_blind_band_flag )
        {
            status = decrement_consumer_count( memory_buffers[index],
                band_attributes->blind_band_index, sca_index );
            if ( status != SUCCESS )
            {
                IAS_LOG_ERROR("Decrementing memory buffer");
                return ERROR;
            }
        }
    }
    return SUCCESS;
}  /* END decrement_counters */

/**************************************************************************
 NAME:     api_sca_discontinuity_correction

 PURPOSE:  Main API call

 RETURNS:  Integer status code of SUCCESS or ERROR
***************************************************************************/
int api_sca_discontinuity_correction
(
    BLOB *blob,         /* I: Miscellaneous parameters */
    int band_index,     /* I: 0-based band index */
    int sca_index,      /* I: 0-based SCA index */
    int algo_index      /* I: Algorithm index */
)
{
    RPS_BAND_ATTRIBUTES *band_attributes = get_band_attributes(band_index);
    int band_number = band_attributes->band_number;
    int sca_number = sca_index + 1;
    float *data_buffer = NULL;               /* Level 1 data buffer */
    double *sca_correction_factor = NULL;    /* Normalized correction factor */
    int num_lines = band_attributes->number_lines;
    int num_detectors = band_attributes->number_detectors;
    int status;                    /* Return status */

    /* Add the ID for SCA discontinuity to the RPS processing level */
    if (set_rps_level(band_index, sca_index, 
            IAS_RPS_LEVEL_SCA_DISCONTINUITY) == ERROR)
    {
        IAS_LOG_ERROR("Updating the RPS processing level");
        return ERROR;
    }

    /* This algorithm doesn't process OLI blind or VRP bands therefore the
        processing level *must* be set to skip */
    if ( band_attributes->vrp_band_index >= 0 )
    {
        if (set_rps_level(band_attributes->vrp_band_index, sca_index,
            IAS_RPS_LEVEL_SKIP) == ERROR)
        {
            IAS_LOG_ERROR("Updating RPS processing level");
            return ERROR;
        }
    }
    if ( band_attributes->blind_band_index >= 0 &&
         band_attributes->process_blind_band_flag )
    {
        if (set_rps_level(band_attributes->blind_band_index, sca_index,
            IAS_RPS_LEVEL_SKIP) == ERROR)
        {
            IAS_LOG_ERROR("Updating RPS processing level");
            return ERROR;
        }
    }

    /* Get the image data buffer */
    data_buffer = (float *)get_memory_buffer(IMAGE_L1R, CONSUMER, band_index,
        sca_index);
    if (data_buffer == NULL)
    {
        IAS_LOG_ERROR("Getting image memory buffer for band %d SCA %d",
            band_number, sca_number);
        return ERROR;
    }

    /* Get SCA Correction Factor from CPF */
    sca_correction_factor = (double *)get_memory_buffer(
        MM_SCA_CORRECTION_FACTOR_CPF, CONSUMER, band_index, sca_index);
    if (!sca_correction_factor)
    {
        IAS_LOG_ERROR("Getting memory buffer: %s for band %d SCA %d", 
            get_memory_buffer_string(MM_SCA_CORRECTION_FACTOR_CPF),
                      band_number, sca_number);
        return ERROR;
    }

    /* Set the discontinuity_factor_source field in the band metadata to
       indicate the CPF is the source */
    strcpy(blob->l1r_band_metadata[band_index][0].discontinuity_factor_source,
        "CPF");

    /* Do the SCA discontinuity correction */
    if ( ias_log_message_level == IAS_LOG_LEVEL_DEBUG )
    {
        struct timeval tv;
        gettimeofday( &tv, NULL );
        IAS_LOG_DEBUG("%ld.%ld Executing SCA discontinuity correction for "
            "band %d SCA %d", tv.tv_sec, tv.tv_usec, band_number, sca_number );
    }
    status = sca_discontinuity_correction(data_buffer, num_detectors,
        num_lines, blob->pixel_mask[band_index][sca_index],
        sca_correction_factor[0]);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("From call to sca_discontinuity_correction algorithm "
            "for band_number: %d, SCA_number %d", band_number, sca_number);
        return ERROR;
    }

    /* Decrement the memory management counters */
    status = decrement_counters( band_index, sca_index );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Decrementing memory buffer");
        return ERROR;
    }

    /* Free the allocated memories */
    return SUCCESS;
}


/***************************************************************************
 NAME:    sca_discontinuity_correction_preprocess

 PURPOSE: Do preprocessing for the algorithm by incrementing the memory
          buffer consumer counts.

 RETURNS: SUCCESS or ERROR
****************************************************************************/
int sca_discontinuity_correction_preprocess
(
    BLOB *blob                 /* I: Miscellaneous parameters */
)
{
    int i;                     /* Local loop counter */
    int status;                /* Status flag */    

    /* Increment the memory buffer consumer counts */
    for (i = 0; i < number_of_memory_buffers; i++)
    {
        status = increment_consumer_count(memory_buffers[i]);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Incrementing %s consumer count",
                get_memory_buffer_string(memory_buffers[i]));
            return ERROR;
        }
    }

    /* Set the L1R file metadata */
    blob->l1r_file_metadata.sca_discontinuity_correction_applied = 1;

    /* Increment the memory buffer consumer counts for postprocess */
    status = increment_consumer_count(MM_SCA_CORRECTION_FACTOR_CPF);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Incrementing %s consumer count",
                get_memory_buffer_string(MM_SCA_CORRECTION_FACTOR_CPF));
        return ERROR;
    }
    
    return SUCCESS;
}  

/*----------------------------------------------------------------------------
NAME:  sca_discontinuity_correction_postprocess  

PURPOSE: Output report file if blob->generate_report is true

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int sca_discontinuity_correction_postprocess
(
    BLOB *blob          /* I: Miscellaneous parameters */
)
{
    IAS_REPORT *report_information = NULL;
    FILE *report_file = NULL;    /* File handle for the report */
    double *correction_factor = NULL;
    int band_number;                    /* Band number */  
    int band_index;                     /* Band index */  
    int sca_index;                      /* SCA index */
    int i;                              /* Loop counter */
    int status;
 
    /* Get the report information */
    if (blob->generate_report)
    {
        /* Get the report information */
        report_information  = &blob->report_information;
        report_file = create_report_file(report_information);
        if (!report_file)
        {
            IAS_LOG_ERROR("Opening report file");
            return ERROR;
        }

        /* Print out the correction factor data */
        for (i = 0; i < blob->number_of_bands; i++)
        {
            band_number = blob->list_of_bands[i];
            band_index = blob->list_of_band_indices[i];  
            RPS_BAND_ATTRIBUTES *band_attributes = 
                get_band_attributes(band_index);
            IAS_LOG_DEBUG("Executing discontinuity post-process for "
                "band %2d", band_number );
            
            /* Print out the column title line */
            status = fprintf(report_file, "Band Number   SCA Number   "
                                  "SCA Correction Factor\n");
            if (status < 0)
            {
                IAS_LOG_ERROR("Writing report column title line");
                return ERROR;
            }

            for (sca_index = 0; sca_index < 
                 band_attributes->number_scas; sca_index++)
            {
                correction_factor = (double *)get_memory_buffer(
                    MM_SCA_CORRECTION_FACTOR_CPF, CONSUMER, band_index,
                    sca_index);
                if (!correction_factor)
                {
                    IAS_LOG_ERROR("Getting memory buffer: %s for band %d "
                        "SCA %d", get_memory_buffer_string(
                    MM_SCA_CORRECTION_FACTOR_CPF), band_number, 
                    sca_index + 1);
                    return ERROR;
                }
  
                status = fprintf( report_file, "%6d %12d %18.6lf\n",
                    band_number, sca_index + 1, *correction_factor);
                if (status < 0)    
                {
                    IAS_LOG_ERROR("Writing report data");
                    return ERROR;
                }

                /* Decremet the consumer counts for  
                   MM_SCA_CORRECTION_FACTOR_CPF */
                status = decrement_consumer_count(MM_SCA_CORRECTION_FACTOR_CPF, 
                                         band_index, sca_index); 
                if ( status != SUCCESS )
                {
                    IAS_LOG_ERROR("Decrementing memory buffer");
                    return ERROR;
                }
            }
        }
        status = close_report_file(report_file);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Closing report file");
            return ERROR;
        }        
    }
    else
    {
        for (i = 0; i < blob->number_of_bands; i++)
        {
            band_number = blob->list_of_bands[i];
            band_index = blob->list_of_band_indices[i];  
            RPS_BAND_ATTRIBUTES *band_attributes = 
                get_band_attributes(band_index);

            for (sca_index = 0; sca_index < 
                band_attributes->number_scas; sca_index++)
            {
                /* Decrement the consumer counts */
                status = decrement_consumer_count(MM_SCA_CORRECTION_FACTOR_CPF, 
                                         band_index, sca_index); 
                if ( status != SUCCESS )
                {
                    IAS_LOG_ERROR("Decrementing memory buffer");
                    return ERROR;
                }
            }
        }
    }

    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME: create_report_file

PURPOSE: Open the report file and allocate temporary storage variables.

RETURNS: Success or Error
-----------------------------------------------------------------------------*/
static FILE *create_report_file
(
    IAS_REPORT *report_info        /* I: IAS report structure pointer */
)
{
    FILE *report_file = NULL;
    char file_base_name[] = "sca_discontinuity_correction_factor";
    char *file_name = NULL;
    IAS_REPORT_HEADER *copy_of_report_header;

    /* Get output report file name */
    file_name = ias_rps_get_report_file_name(report_info, -1, -1, 
                                             file_base_name);
    if (!file_name)
    {
        IAS_LOG_ERROR("Getting report filename %s", file_name);
        return NULL;        
    }

    /* Open the report file */
    report_file = fopen(file_name, "w");
    if (!report_file)
    {
        IAS_LOG_ERROR("Opening report file %s", file_name);
        free(file_name);
        return NULL;
    }

    /* Get the local copy of the report header since the "standard" header
       includes non-standard things such as band, SCA and algorithm name.
       In a threaded environment the input header cannot be manipulated. */
    copy_of_report_header = ias_misc_copy_report_header(report_info->header);
    if (!copy_of_report_header)
    {
        IAS_LOG_ERROR("Creating local copy of report header");
        fclose(report_file);
        free(file_name);
        return NULL;
    }

    /* Set the band number and SCA number to 0 in the report header--
       they won't be written out.  Also, set the algorithm name in
       the report header */
    ias_misc_set_header_band_number(copy_of_report_header, 0);
    ias_misc_set_header_sca_number(copy_of_report_header, 0);
    ias_misc_set_header_algorithm_name(copy_of_report_header,
        "SCA Discontinuity Correction");

    if (ias_misc_write_report_header(report_file, copy_of_report_header))
    {
        ias_misc_free_report_header(copy_of_report_header);
        fclose(report_file);
        free(file_name);
        return NULL;
    }

    ias_misc_free_report_header(copy_of_report_header);
    free(file_name);

    return report_file;
}

/*----------------------------------------------------------------------------
NAME: close_report_file

PURPOSE: Close the report file.

RETURNS: None
-----------------------------------------------------------------------------*/
static int close_report_file
(
     FILE *report_file                /* Output file pointer */
)
{
    if (!report_file)
        return ERROR;

    if (fclose( report_file ))
    {
        IAS_LOG_ERROR("Closing report file");
        return ERROR;
    }

    return SUCCESS;
}

