/*----------------------------------------------------------------------------
NAME:    api_striping_characterization.c

         Contains:
             api_striping_characterization
             striping_characterization_preprocess
             striping_characterization_postprocess

PURPOSE: Gather data for the striping characterization algorithm.
         The post-process routine gathers the detector striping metric for
         each SCA of a band and calls the algorithm to compute the overall
         striping metric for the band.

-----------------------------------------------------------------------------*/
#include <stdlib.h>
#include <string.h>

#include "ias_logging.h"
#include "rps_local.h"
#include "ias_satellite_attributes.h"

/* Define a list of memory buffers needed by the algorithm */
static MEMORY_BUFFER_TYPE memory_buffers[] =
{   
    IMAGE_L1R,
    MM_SCENE_STRIPING_CORRECTION_MATRIX,
    MM_DETECTOR_STRIPING_METRIC,
    MM_SPATIAL_DETECTOR_OFFSETS,
    MM_MAX_SPATIAL_DETECTOR_OFFSETS,
    MM_ABSOLUTE_GAINS,
    MM_STRIPING_METRIC_CUTOFF
};
static int number_of_memory_buffers = sizeof(memory_buffers) /
    sizeof(MEMORY_BUFFER_TYPE);

/*----------------------------------------------------------------------------
NAME:    decrement_counters

PURPOSE:  Internal routine to decrement the consumer counts of all required
          memory buffers

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
static int decrement_counters
(
    int band_index,     /* I: */
    int sca_index       /* I: */
)
{
    RPS_BAND_ATTRIBUTES *band_attributes = get_band_attributes( band_index );
    int vrp_band_index = band_attributes->vrp_band_index;
    int blind_band_index = band_attributes->blind_band_index;
    int blind_vrp_band_index = -1;
    int index;
    int status;

    if ( blind_band_index >= 0 )
    {
        blind_vrp_band_index = get_band_attributes(
            blind_band_index)->vrp_band_index;
    }

    /* Decrement the consumer counts for the memory buffers */
    for ( index = 0; index < number_of_memory_buffers; index++ )
    {
        status = decrement_consumer_count( memory_buffers[index], band_index,
            sca_index );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }

        if ( vrp_band_index >= 0 )
        {
            status = decrement_consumer_count( memory_buffers[index],
                vrp_band_index, sca_index );
            if ( status != SUCCESS )
            {
                IAS_LOG_ERROR("Decrementing memory buffer");
                return ERROR;
            }
        }

        if ( blind_band_index >= 0
            && band_attributes->process_blind_band_flag )
        {
            status = decrement_consumer_count( memory_buffers[index],
                blind_band_index, sca_index );
            if ( status != SUCCESS )
            {
                IAS_LOG_ERROR("Decrementing memory buffer");
                return ERROR;
            }
            if ( blind_vrp_band_index >= 0 )
            {
                status = decrement_consumer_count( memory_buffers[index],
                    blind_vrp_band_index, sca_index );
                if ( status != SUCCESS )
                {
                    IAS_LOG_ERROR("Decrementing memory buffer");
                    return ERROR;
                }
            }
        }
    }
    return SUCCESS;
}  /* END decrement_counters */



/*----------------------------------------------------------------------------
NAME:    api_striping_characterization

PURPOSE: Gather data needed by the algorithm and call it.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int api_striping_characterization
(
    BLOB *blob,         /* I: Miscellaneous parameters */
    int   band_index,   /* I: */
    int   sca_index,    /* I: */
    int   algo_index    /* I: Algorithm index */
)
{
    RPS_BAND_ATTRIBUTES *band_attributes = get_band_attributes( band_index );
    int band_number = band_attributes->band_number;
    int sca_number = sca_index + 1;
    int number_lines = band_attributes->number_lines;
    int number_detectors = band_attributes->number_detectors;
    int status;
    float *image_buffer;
    int *detector_offsets;
    int *max_offset;
    int dump_matrix_flag = 0;
    IAS_REPORT *report_info = NULL;
    double *detector_striping_metric;
    double *striping_correction_matrix;
    double *striping_metric_cutoff;

    if ( blob->generate_report )
    {
        report_info = &blob->report_information;
        dump_matrix_flag = 1;
    }

    /* Get the Level 1R image data */
    image_buffer = (float *)get_memory_buffer( IMAGE_L1R, CONSUMER, band_index,
        sca_index );
    if ( !image_buffer )
    {
        IAS_LOG_ERROR("Getting %s for band %d SCA %d",
            get_memory_buffer_string(IMAGE_L1R), band_number, sca_number );
        return ERROR;
    }

    /* Get the detector offsets and the max offset value */
    detector_offsets = (int *)get_memory_buffer(
        MM_SPATIAL_DETECTOR_OFFSETS, CONSUMER, band_index, sca_index );
    if ( !detector_offsets )
    {
        IAS_LOG_ERROR("Getting %s for band %d SCA %d",
            get_memory_buffer_string(MM_SPATIAL_DETECTOR_OFFSETS),
            band_number, sca_number );
        return ERROR;
    }
    max_offset = (int *)get_memory_buffer( MM_MAX_SPATIAL_DETECTOR_OFFSETS,
        CONSUMER, band_index, sca_index );
    if ( !max_offset )
    {
        IAS_LOG_ERROR("Getting %s for band %d SCA %d",
            get_memory_buffer_string(MM_MAX_SPATIAL_DETECTOR_OFFSETS),
            band_number, sca_number );
        return ERROR;
    }

    /* Get the CPF value for striping metric cutoff */
    striping_metric_cutoff = (double *)get_const_memory_buffer(
        MM_STRIPING_METRIC_CUTOFF, band_index, sca_index );
    if ( !striping_metric_cutoff )
    {
        IAS_LOG_ERROR("Getting %s for band %d SCA %d",
            get_memory_buffer_string(MM_STRIPING_METRIC_CUTOFF),
            band_number, sca_number );
        return ERROR;
    }

    /* Get the two output buffers for the algorithm: correction matrix and
       detector metric */
    detector_striping_metric = (double *)get_memory_buffer(
        MM_DETECTOR_STRIPING_METRIC, SETTER, band_index, sca_index );
    if ( !detector_striping_metric )
    {
        IAS_LOG_ERROR("Getting %s for band %d SCA %d",
            get_memory_buffer_string(MM_DETECTOR_STRIPING_METRIC),
            band_number, sca_number );
        return ERROR;
    }
    striping_correction_matrix = (double *)get_memory_buffer(
        MM_SCENE_STRIPING_CORRECTION_MATRIX, SETTER, band_index, sca_index );
    if ( !striping_correction_matrix )
    {
        IAS_LOG_ERROR("Getting %s for band %d SCA %d",
            get_memory_buffer_string(MM_SCENE_STRIPING_CORRECTION_MATRIX),
            band_number, sca_number );
        return ERROR;
    }

    status = striping_characterization( image_buffer, band_number,
        sca_number, number_detectors, number_lines, detector_offsets,
        max_offset[0], blob->pixel_mask[band_index][sca_index], 
        striping_metric_cutoff[0], dump_matrix_flag, report_info,
        detector_striping_metric, striping_correction_matrix );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Returned from striping_characterization for band "
            "%d SCA %d", band_number, sca_number );
        return ERROR;
    }

    /* Let consumers know that the striping matrices are valid */
    set_buffer_filled_flag( MM_DETECTOR_STRIPING_METRIC, band_index,
        sca_index );
    set_buffer_filled_flag( MM_SCENE_STRIPING_CORRECTION_MATRIX, band_index,
        sca_index );

    status = decrement_counters( band_index, sca_index );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Decrementing memory buffer");
        return ERROR;
    }

    return SUCCESS;
}  /* END api_striping_characterization */



/*----------------------------------------------------------------------------
NAME:    striping_characterization_preprocess

PURPOSE: Increment the memory buffers

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int striping_characterization_preprocess
(
    BLOB *blob          /* I: Miscellaneous parameters */
)
{
    int index;
    int status;

    /* Register the main algorithms memory needs */
    for ( index = 0; index < number_of_memory_buffers; index++ )
    {
        status = increment_consumer_count( memory_buffers[index] );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Incrementing %s consumer count",
                get_memory_buffer_string( memory_buffers[index] ));
            return ERROR;
        }
    }

    /* The post-process routine needs the detector striping metric and the
       absolute gains so increment them here */
    status = increment_consumer_count( MM_DETECTOR_STRIPING_METRIC );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Incrementing %s consumer count",
            get_memory_buffer_string( MM_DETECTOR_STRIPING_METRIC ));
        return ERROR;
    }

    status = increment_consumer_count( MM_ABSOLUTE_GAINS );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Incrementing %s consumer count",
            get_memory_buffer_string( MM_ABSOLUTE_GAINS ));
        return ERROR;
    }

    /* Clean up the database records if needed */
    if ( blob->store_in_database )
    {
        status = delete_db_table_records( blob->database_information,
            "striping" );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Returned from delete_db_table_data");
            return ERROR;
        }
    }

    return SUCCESS;
}  /* END striping_characterization_preprocess */



/*----------------------------------------------------------------------------
NAME:    striping_characterization_postprocess

PURPOSE: Compute the overall striping metric for all bands

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int striping_characterization_postprocess
(
    BLOB *blob          /* I: Miscellaneous parameters */
)
{
    int status;
    int i;
    double *combined_dsm = NULL;
    size_t size_of_combined_dsm = 0;
    RPS_BAND_ATTRIBUTES *band_attributes;
    int number_scas;
    int band_index;
    int sca_index;
    DB_OVERALL_STRIPING_METRIC_RESULTS *results;

    /* If no reporting or database storage is desired there is no point in
       running the algorithm */
    if ( blob->generate_report == 0 && blob->store_in_database == 0 )
    {
        IAS_LOG_INFO("Overall striping metric not calculated due to report "
           "and database storage flags being false");

        /* Decrement the counters */
        for ( i = 0; i < blob->number_of_bands; i++ )
        {
            band_index = blob->list_of_band_indices[i];
            band_attributes = get_band_attributes( band_index );
            number_scas = band_attributes->number_scas;
            for ( sca_index = 0; sca_index < number_scas; sca_index++ )
            {
                status = decrement_consumer_count( MM_DETECTOR_STRIPING_METRIC,
                    band_index, sca_index );
                if ( status != SUCCESS )
                {
                    IAS_LOG_ERROR("Decrementing memory buffer");
                    return ERROR;
                }
                status = decrement_consumer_count( MM_ABSOLUTE_GAINS,
                    band_index, sca_index );
                if ( status != SUCCESS )
                {
                    IAS_LOG_ERROR("Decrementing memory buffer");
                    return ERROR;
                }
            }
        }

        return SUCCESS;
    }

    /* Since we got here we have database or reporting to do.  Use the
       same structure for both. */
    results = db_allocate_overall_striping_metric_results(
        blob->number_of_bands );
    if ( !results )
    {
        IAS_LOG_ERROR("Allocating the overall striping results structure");
        return ERROR;
    }
    

    /* Loop through the list of bands and compute the overall striping
       metric.  This is a full band metric so the individual SCA detector
       striping metrics are combined. */
    for ( i = 0; i < blob->number_of_bands; i++ )
    {
        band_index = blob->list_of_band_indices[i];
        band_attributes = get_band_attributes( band_index );
        int band_number = band_attributes->band_number;
        number_scas = band_attributes->number_scas;
        int number_detectors = band_attributes->number_detectors;
        size_t size_of_one_sca = sizeof(double) * number_detectors;
        int total_number_detectors = number_detectors * number_scas;
        size_t size_of_data = sizeof(double) * total_number_detectors;
        double *data_ptr;
        double mean_absolute_gain;
        double osm_radiance;
        double osm_dn;

        /* Make sure the buffer is big enough for the combined buffer */
        if ( size_of_data > size_of_combined_dsm )
        {
            double *temp = realloc( combined_dsm, size_of_data );
            if ( !temp )
            {
                IAS_LOG_ERROR("Reallocating the combined DSM buffer "
                    "for Band %d", band_number );
                db_free_overall_striping_metric_results( results );
                free( combined_dsm );
                return ERROR;
            }
            combined_dsm = temp;
        }

        /* Get the detector striping metric for each SCA and copy it to the
           combined buffer.  Also compute the mean of the SCA absolute
           gains. */
        data_ptr = combined_dsm;
        mean_absolute_gain = 0.0;
        for ( sca_index = 0; sca_index < number_scas; sca_index++ )
        {
            const double *absolute_gain;
            double *dsm = (double *)get_memory_buffer(
                MM_DETECTOR_STRIPING_METRIC, CONSUMER, band_index, sca_index );
            if ( !dsm )
            {
                IAS_LOG_ERROR("Getting the detector striping metric for "
                    "Band %d DCA %d", band_number, sca_index + 1 );
                db_free_overall_striping_metric_results( results );
                free( combined_dsm );
                return ERROR;
            }
            memcpy( data_ptr, dsm, size_of_one_sca );

            /* Update the pointer to the start of the next SCA */
            data_ptr += number_detectors;

            absolute_gain = (const double *)get_const_memory_buffer(
                MM_ABSOLUTE_GAINS, band_index, sca_index );
            if ( !absolute_gain )
            {
                IAS_LOG_ERROR("Getting the absolute gain for Band %d "
                    "SCA %d", band_number, sca_index + 1 );
                db_free_overall_striping_metric_results( results );
                free( combined_dsm );
                return ERROR;
            }
            /* Add this SCA absolute gain into the mean */
            mean_absolute_gain += absolute_gain[0];

            /* Release the memory */
            status = decrement_consumer_count( MM_DETECTOR_STRIPING_METRIC,
                band_index, sca_index );
            if ( status != SUCCESS )
            {
                IAS_LOG_ERROR("Decrementing memory buffer");
                db_free_overall_striping_metric_results( results );
                free( combined_dsm );
                return ERROR;
            }
            status = decrement_consumer_count( MM_ABSOLUTE_GAINS,
                band_index, sca_index );
            if ( status != SUCCESS )
            {
                IAS_LOG_ERROR("Decrementing memory buffer");
                db_free_overall_striping_metric_results( results );
                free( combined_dsm );
                return ERROR;
            }
        }
        /* Compute the mean */
        mean_absolute_gain /= number_scas;

        /* Call the algorithm */
        status = calculate_overall_striping_metric( combined_dsm,
            total_number_detectors, mean_absolute_gain, &osm_radiance,
            &osm_dn );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Calculating overall striping metric for Band %d",
                band_number );
            db_free_overall_striping_metric_results( results );
            free( combined_dsm );
            return ERROR;
        }

        /* Store the results for this band */
        results->band_numbers[i] = band_number;
        results->osm_radiance[i] = osm_radiance;
        results->osm_dn[i] = osm_dn;
    }  /* END for i loop */

    /* Write out the report if desired */
    if ( blob->generate_report )
    {
        status = generate_overall_striping_metric_report(
            &blob->report_information, results );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Writing overall striping metric report");
            db_free_overall_striping_metric_results( results );
            free( combined_dsm );
            return ERROR;
        }
    }

    /* Store in database if desired */
    if ( blob->store_in_database )
    {
        results->dbase = blob->database_information.db;
        strncpy( results->rad_common_id,
            blob->database_information.rad_common_id,
            sizeof(results->rad_common_id));
        status = db_store_overall_striping_metric_in_database( results );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Storing Overall striping metric to database");
            db_free_overall_striping_metric_results( results );
            free( combined_dsm );
            return ERROR;
        }
    }

    /* Clean up */
    db_free_overall_striping_metric_results( results );
    free( combined_dsm );
    return SUCCESS;
}  /* END striping_characterization_postprocess */
