/*----------------------------------------------------------------------------
NAME:    api_temperature_sensitivity_correction

PURPOSE: Gather data needed for the temperature_sensitivity_correction
         routine and call it.  

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
#include <stdlib.h>
#include "ias_logging.h"
#include "rps_local.h"
#include "ias_satellite_attributes.h"

#define CELSIUS_TO_KELVIN_CONVERTOR 273.15

int api_temperature_sensitivity_correction
(
    BLOB *blob,         /* I: Miscellaneous parameters */
    int   band_index,   /* I: */
    int   sca_index,    /* I: */
    int   algo_index    /* I: Algorithm index */
)
{
    RPS_BAND_ATTRIBUTES *band_attributes = get_band_attributes(band_index);
    int num_samps = band_attributes->number_detectors;
    const double *temp_sens_coeff;  /* Temperature sensitivity coefficients */
    const double *ref_temp;         /* Refernce temperature */
    double *temp_correction_factor; /* Temperature correction factor */
    float mean_focal_temp;          /* Mean focal plane temperature */
    int status;

    /* This algorithm doesn't process OLI & TIRS blind bands or OLI VRP bands 
       therefore the processing level *must* be set to skip */
    if ( band_attributes->vrp_band_index >= 0 )
    {
        if (set_rps_level(band_attributes->vrp_band_index, sca_index,
            IAS_RPS_LEVEL_SKIP) == ERROR)
        {
            IAS_LOG_ERROR("Updating RPS processing level");
            return ERROR;
        }
    }
    if ( band_attributes->blind_band_index >= 0 &&
         band_attributes->process_blind_band_flag )
    {
        if (set_rps_level(band_attributes->blind_band_index, sca_index,
            IAS_RPS_LEVEL_SKIP) == ERROR)
        {
            IAS_LOG_ERROR("Updating RPS processing level");
            return ERROR;
        }
    }

    /* Get the correct mean_focal_plane temperature based on spectral type */
    if ( band_attributes->spectral_type == IAS_SPECTRAL_THERMAL )
    {
        /* Get the TIRS mean focal plane temperature */
        status = get_tirs_focal_plane_temperature( &mean_focal_temp );
        if ( status != SUCCESS )
        {
           IAS_LOG_ERROR("Getting TIRS focal plane temperature");
            return ERROR;
        }
    }
    else
    {
        /* Get the OLI mean focal plane temperature */
        status = get_oli_focal_plane_temperature( &mean_focal_temp );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Getting OLI focal plane temperature");
            return ERROR;
        }
    }

    /* Get the sensitivity reference temperature */
    ref_temp = (const double *)get_const_memory_buffer(
        MM_TEMP_SENSITIVITY_REFERENCE_TEMP, band_index, sca_index);
    if (!ref_temp)
    {
        IAS_LOG_ERROR("Getting sensitivity reference temperature");
        return ERROR;
    }

    /* Get the temperature sensitivity coefficient */
    temp_sens_coeff = (const double *)get_const_memory_buffer(
        MM_TEMP_SENSITIVITY_COEFFICIENT, band_index, sca_index);
    if (!temp_sens_coeff)
    {
        IAS_LOG_ERROR("Getting temperature sensitivity coefficient");
        return ERROR;
    }

    /* Get the temp_correction_factor memory buffer */
    temp_correction_factor = get_memory_buffer(MM_TEMP_CORRECTION_FACTOR,
        SETTER, band_index, sca_index);
    if (!temp_correction_factor)
    {
        IAS_LOG_ERROR("Setting the temp_correction_factor");
        return ERROR;
    }

    temperature_sensitivity_correction(num_samps, mean_focal_temp, 
             temp_sens_coeff, ref_temp[0], temp_correction_factor);

    /* Let consumers know that the buffer is valid */
    set_buffer_filled_flag( MM_TEMP_CORRECTION_FACTOR, band_index, sca_index );

    /* Tell memory management that the data was used */
    status = decrement_consumer_count(MM_TEMP_SENSITIVITY_REFERENCE_TEMP,
        band_index, sca_index);
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Decrementing memory buffer");
        return ERROR;
    }
    status = decrement_consumer_count(MM_TEMP_SENSITIVITY_COEFFICIENT,
        band_index, sca_index);
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Decrementing memory buffer");
        return ERROR;
    }
    status = decrement_consumer_count(MM_TEMP_CORRECTION_FACTOR, band_index,
        sca_index);
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Decrementing memory buffer");
        return ERROR;
    }

    return SUCCESS;
}  /* END api_temperature_sensitivity_correction */


/*------------------------------------------------------------------------
NAME:    temperature_sensitivity_correction_preprocess

PURPOSE: Registers the memory needed by the algorithm.  Calculates the
         mean focal plane temperature and stores in a memory buffer for
         use by the algorithm.

RETURNS: SUCCESS or ERROR
-------------------------------------------------------------------------*/
int temperature_sensitivity_correction_preprocess
(
    BLOB *blob           /* I: Miscellaneous parameters */
)
{
    const struct IAS_CPF_TEMP_SENSITIVITY *temp_sensitivity;
    float *oli_fp_temp1 = NULL;   /* OLI focal plane temperature 1 list */
    float *oli_fp_temp2 = NULL;   /* OLI focal plane temperature 2 list */
    float *tirs_fp_temp1 = NULL;  /* TIRS focal plane temperature 1 list */
    float *tirs_fp_temp2 = NULL;  /* TIRS focal plane temperature 2 list */
    float *tirs_fp_temp3 = NULL;  /* TIRS focal plane temperature 1 list */
    float *tirs_fp_temp4 = NULL;  /* TIRS focal plane temperature 2 list */
    int num_oli_fp_temps;         /* Number of OLI focal plane temperatures */
    int num_tirs_fp_temps;        /* Number of TIRS focal plane temperatures */
    int index;                    /* Loop counter */
    double mean_oli_fp_temp1;     /* averaged OLI focal plane temperature 1 */
    double mean_oli_fp_temp2;     /* averaged OLI focal plane temperature 2 */
    double mean_tirs_fp_temp1;    /* averaged TIRS focal plane temperature 1 */
    double mean_tirs_fp_temp2;    /* averaged TIRS focal plane temperature 2 */
    double mean_tirs_fp_temp3;    /* averaged TIRS focal plane temperature 3 */
    double mean_tirs_fp_temp4;    /* averaged TIRS focal plane temperature 4 */
    double sum_oli_temp;          /* OLI focal plane temperature sum */
    double sum_tirs_temp;         /* TIRS focal plane temperature sum */
    float mean_oli_fp_temp;       /* Mean OLI focal plane temperature */
    float mean_tirs_fp_temp;      /* Mean TIRS focal plane temperature */

    /* Define a list of memory buffers needed by the algorithm */
    MEMORY_BUFFER_TYPE memory_buffers[] =
    {
        MM_TEMP_SENSITIVITY_REFERENCE_TEMP,
        MM_TEMP_SENSITIVITY_COEFFICIENT,
        MM_TEMP_CORRECTION_FACTOR
    };
    int number_of_memory_buffers = sizeof(memory_buffers) /
        sizeof(MEMORY_BUFFER_TYPE);
    int i;
    int status;

    /* Increment the consumer counts for the memory buffers */
    for ( i = 0; i < number_of_memory_buffers; i++ )
    {
        status = increment_consumer_count( memory_buffers[i] );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Incrementing %s consumer count",
                get_memory_buffer_string( memory_buffers[i] ));
            return ERROR;
        }
    }

    /* Get the OLI and TIRS thermistor_flag values from CPF */
    temp_sensitivity = ias_cpf_get_temp_sensitivity( get_blob_cpf() );
    if ( !temp_sensitivity )
    {
        IAS_LOG_ERROR("Getting the CPF temperature sensitivity "
            "thermistor flags");
        return ERROR;
    }


    /* Get the OLI focal plane temperatures if any OLI band appears in the
       blob->list_of_bands to process*/
    if (blob->process_oli_bands)
    {
        status = read_oli_focal_plane_temperatures( &oli_fp_temp1,
            &oli_fp_temp2, &num_oli_fp_temps );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Getting the OLI focal plane temperatures");
            return ERROR;
        }

        /* Now calculate the mean focal temperature based on oli_fp_temp1
           and oli_fp_temp2*/
        sum_oli_temp = 0;
        for (index = 0; index < num_oli_fp_temps; index++)
            sum_oli_temp += oli_fp_temp1[index];
        mean_oli_fp_temp1 = sum_oli_temp / num_oli_fp_temps; 

        sum_oli_temp = 0;
        for (index = 0; index < num_oli_fp_temps; index++)
            sum_oli_temp += oli_fp_temp2[index];
        mean_oli_fp_temp2 = sum_oli_temp / num_oli_fp_temps; 

        mean_oli_fp_temp = (mean_oli_fp_temp1 * 
           temp_sensitivity->oli_thermistor_flag[0] + mean_oli_fp_temp2 * 
           temp_sensitivity->oli_thermistor_flag[1]) / 2.0;

        /* Convert the mean OLI focal plane temperature from Celsius to 
           Kelvin */
        mean_oli_fp_temp += CELSIUS_TO_KELVIN_CONVERTOR;

        /* Set the mean OLI focal plane temperature in the blob */
        set_oli_focal_plane_temperature( mean_oli_fp_temp );

        free( oli_fp_temp1 );
        free( oli_fp_temp2 );
    }

    /* Get the TIRS focal plane temperatures */
    if (blob->process_tirs_bands)
    {
        status = read_tirs_focal_plane_temperatures( &tirs_fp_temp1,
            &tirs_fp_temp2, &tirs_fp_temp3, &tirs_fp_temp4, 
            &num_tirs_fp_temps );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Getting the TIRS focal plane temperatures");
            return ERROR;
        }

        /* Now calculate the mean focal temperature based on tirs_fp_temp1,
           tirs_fp_temp2, tirs_fp_temp3, and tirs_fp_temp4 */
        sum_tirs_temp = 0;
        for (index = 0; index < num_tirs_fp_temps; index++)
            sum_tirs_temp += tirs_fp_temp1[index];
        mean_tirs_fp_temp1 = sum_tirs_temp / num_tirs_fp_temps; 

        sum_tirs_temp = 0;
        for (index = 0; index < num_tirs_fp_temps; index++)
            sum_tirs_temp += tirs_fp_temp2[index];
        mean_tirs_fp_temp2 = sum_tirs_temp / num_tirs_fp_temps; 

        sum_tirs_temp = 0;
        for (index = 0; index < num_tirs_fp_temps; index++)
            sum_tirs_temp += tirs_fp_temp3[index];
        mean_tirs_fp_temp3 = sum_tirs_temp / num_tirs_fp_temps; 

        sum_tirs_temp = 0;
        for (index = 0; index < num_tirs_fp_temps; index++)
            sum_tirs_temp += tirs_fp_temp4[index];
        mean_tirs_fp_temp4 = sum_tirs_temp / num_tirs_fp_temps; 

        mean_tirs_fp_temp = (mean_tirs_fp_temp1 * 
           temp_sensitivity->tirs_thermistor_flag[0] + mean_tirs_fp_temp2 * 
           temp_sensitivity->tirs_thermistor_flag[1] + mean_tirs_fp_temp3 * 
           temp_sensitivity->tirs_thermistor_flag[2] + mean_tirs_fp_temp4 * 
           temp_sensitivity->tirs_thermistor_flag[3]) / 4.0;

        /* Convert the mean TIRS focal plane temperature from Celsius to 
           Kelvin */
        mean_tirs_fp_temp += CELSIUS_TO_KELVIN_CONVERTOR;

        /* Set the mean TIRS focal plane temperature in the blob */
        set_tirs_focal_plane_temperature( mean_tirs_fp_temp );

        free( tirs_fp_temp1 );
        free( tirs_fp_temp2 );
        free( tirs_fp_temp3 );
        free( tirs_fp_temp4 );
    }

    return SUCCESS;
}  /* END temperature_sensitivity_correction_preprocess */
