/*----------------------------------------------------------------------------
NAME:    api_tirs_bias_removal

PURPOSE: Gather data needed for the TIRS bial removal application routine and 
         call the routine.  

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
#include <stdlib.h>
#include <string.h>
#include "ias_logging.h"
#include "rps_local.h"

/* Define a list of memory buffers needed by the algorithm */
static MEMORY_BUFFER_TYPE memory_buffers[] =
{
    IMAGE_L1R,
    MM_TIRS_DARK_RESPONSE,
    MM_TIRS_BACKGROUND_RESPONSE,
    MM_BPF_BACKGROUND_RESPONSE
};

static int number_of_memory_buffers =
    sizeof(memory_buffers) / sizeof(MEMORY_BUFFER_TYPE);

/**************************************************************************
 NAME:     decrement_counters

 PURPOSE:  Internal routine to decrement the consumer counts of all required
           memory buffers

 RETURNS: SUCCESS or ERROR
***************************************************************************/
static int decrement_counters
(
    int band_index,     /* I: */
    int sca_index       /* I: */
)
{
    RPS_BAND_ATTRIBUTES *band_attributes = get_band_attributes( band_index );
    int i;              /* Local loop counter */
    int status;

    /* Decrement the consumer counts for the memory buffers */
    for (i = 0; i < number_of_memory_buffers; i++)
    {
        status = decrement_consumer_count(memory_buffers[i], band_index,
            sca_index);
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }

        /* If the band has an associated VRP band decrement the user count */
        if (band_attributes->vrp_band_number > 0)
        {
            status = decrement_consumer_count(memory_buffers[i],
                band_attributes->vrp_band_index, sca_index );
            if ( status != SUCCESS )
            {
                IAS_LOG_ERROR("Decrementing memory buffer");
                return ERROR;
            }
        }

        /* If the band has an associated blind band and it's set to process the
           blind band, decrement the user count */
        if (band_attributes->blind_band_number > 0
            && band_attributes->process_blind_band_flag )
        {
            status = decrement_consumer_count(memory_buffers[i],
                band_attributes->blind_band_index, sca_index);
            if ( status != SUCCESS )
            {
                IAS_LOG_ERROR("Decrementing memory buffer");
                return ERROR;
            }
        }
    }
    return SUCCESS;
}   /* END decrement_counters */

/*----------------------------------------------------------------------------
NAME:    api_tirs_bias_removal

PURPOSE: Gather data needed for the TIRS bias removal application routine and 
         call the routine.  

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int api_tirs_bias_removal
(
    BLOB *blob,         /* I: Miscellaneous parameters */
    int   band_index,   /* I: */
    int   sca_index,    /* I: */
    int   algo_index    /* I: Algorithm index */
)
{
    RPS_BAND_ATTRIBUTES *band_attributes = get_band_attributes(band_index);
    int band_number = band_attributes->band_number;
    int number_lines = band_attributes->number_lines;
    int number_detectors = band_attributes->number_detectors;
    float *data_buffer;            /* Level 1 data buffer */
    const double *cpf_background_response = NULL;  
                                   /* Background response values from CPF */
    double *background_response = NULL; /* Background/Dark response values */
    int det;                       /* detector index counter */
    int status;

    /* This is only to be run for TIRS bands */
    if (band_attributes->spectral_type != IAS_SPECTRAL_THERMAL)
    {
        status = decrement_counters(band_index, sca_index);
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }
        return SUCCESS;
    }

    /* Add the ID for bias removal to the RPS processing level */
    if (set_rps_level(band_index, sca_index, 
            IAS_RPS_LEVEL_BIAS_REMOVAL) == ERROR)
    {
        IAS_LOG_ERROR("Updating RPS processing level");
        return ERROR;
    }

    /* This algorithm doesn't process TIRS blind bands therefore the
        processing level *must* be set to skip */
    if ( band_attributes->blind_band_index >= 0 &&
         band_attributes->process_blind_band_flag )
    {
        if (set_rps_level(band_attributes->blind_band_index, sca_index,
            IAS_RPS_LEVEL_SKIP) == ERROR)
        {
            IAS_LOG_ERROR("Updating RPS processing level");
            return ERROR;
        }
    }

    /* Note: It isn't necessary to decrement the usage of the memory buffer
       on error conditions from this point on.  The main thread will clean
       up allocated buffers when error conditions dictate.  It doesn't hurt
       but it isn't necessary. */

    /* Get the data buffer */
    data_buffer = (float *)get_memory_buffer(IMAGE_L1R, CONSUMER, band_index,
        sca_index);
    if (!data_buffer)
    {
        IAS_LOG_ERROR("Getting image memory buffer");
        return ERROR;
    }

    /* Set the appropriate tirs bias data */
    switch (blob->background_response_selection)
    {
        case PRE_ACQUISITION_DEEP_SPACE:
        case POST_ACQUISITION_DEEP_SPACE:
        case MEAN_ACQUISITION_DEEP_SPACE:
            /* Get the pre-aquisition deep space averages */
            background_response = (double *)get_memory_buffer(
                MM_BPF_BACKGROUND_RESPONSE, CONSUMER, band_index, sca_index);
            if (!background_response)
            {
                IAS_LOG_ERROR("Getting memory buffer %s for band %d SCA %d",
                    get_memory_buffer_string(MM_BPF_BACKGROUND_RESPONSE), 
                    band_number, sca_index + 1);
                return ERROR;
            }
            break;

        case CPF_DARK_AND_BACKGROUND_RESPONSE:
            /* Get the dark response value, use background_response 
               variable to make it consistence with BPF option */
            background_response = (double *)get_memory_buffer(
                MM_TIRS_DARK_RESPONSE, CONSUMER, band_index, sca_index);
            if (!background_response)
            {
             IAS_LOG_ERROR("Getting memory buffer %s for band %d SCA %d",
                    get_memory_buffer_string(MM_TIRS_DARK_RESPONSE), 
                    band_number, sca_index + 1);
                return ERROR;
            }

            /* Get the background response value */
            cpf_background_response = (const double *)get_const_memory_buffer(
                MM_TIRS_BACKGROUND_RESPONSE, band_index, sca_index);
            if (!cpf_background_response)
            {
             IAS_LOG_ERROR("Getting memory buffer %s for band %d SCA %d",
                    get_memory_buffer_string(MM_TIRS_BACKGROUND_RESPONSE), 
                    band_number, sca_index + 1);
                return ERROR;
            }
            for (det = 0; det < number_detectors; det++)
                background_response[det] += cpf_background_response[det];
            break;

        default:
        {
            IAS_LOG_ERROR("Invalid BACKGROUND_RESPONSE_SELECTION %d for band "
                "%d SCA %d", blob->background_response_selection, band_number, 
                sca_index + 1);
            return ERROR;
        }
    }

    /* Do the bias removal */
    generic_parameter_removal(data_buffer, number_detectors, number_lines, 
                      background_response);

    /* Set the band metadata */
    if ( sca_index == 0 )
    {
        switch (blob->background_response_selection)
        {
            case PRE_ACQUISITION_DEEP_SPACE:
                strcpy( blob->l1r_band_metadata[band_index]->bias_source,
                    "BPF, Pre");
                break;
            case POST_ACQUISITION_DEEP_SPACE:
                strcpy( blob->l1r_band_metadata[band_index]->bias_source,
                    "BPF, Post");
                break;
            case MEAN_ACQUISITION_DEEP_SPACE:
                strcpy( blob->l1r_band_metadata[band_index]->bias_source,
                    "BPF, Mean");
                break;
            case CPF_DARK_AND_BACKGROUND_RESPONSE:
                strcpy( blob->l1r_band_metadata[band_index]->bias_source,
                    "CPF");
                break;
        }

        /* Set the bias correction flag */
        blob->l1r_band_metadata[band_index]->bias_correction_applied = 1;
    }

    /* Decrement the memory management counters */
    status = decrement_counters(band_index, sca_index);
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Decrementing memory buffer");
        return ERROR;
    }

    return SUCCESS;
}  /* END api_tirs_bias_removal */


/*----------------------------------------------------------------------------
NAME:    tirs_bias_removal_preprocess

PURPOSE: Do preprocessing for the algorithm by incrementing the consumer
         counts.  

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int tirs_bias_removal_preprocess
(
    BLOB *blob          /* I: Miscellaneous parameters */
)
{
    int i;
    int status;

    /* Increment the consumer counts for the memory buffers */
    for ( i = 0; i < number_of_memory_buffers; i++ )
    {
        status = increment_consumer_count( memory_buffers[i] );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Incrementing %s consumer count",
                get_memory_buffer_string( memory_buffers[i] ));
            return ERROR;
        }
    }

    return SUCCESS;
}  /* END tirs_bias_removal_preprocess */
