/*************************************************************************
 NAME:     api_tirs_dark_response_determination

 PURPOSE:  Gather and package input data for the TIRS dark response
           determination algorithm, then execute the algorithm

 RETURNS:  Integer status code of SUCCESS or ERROR

**************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <limits.h>
#include "ias_logging.h"
#include "ias_satellite_attributes.h"
#include "rps_local.h"
#include "ias_rps_histogram_statistics_characterization.h"
#include "ias_rps_tirs_dark_response_determination.h"


static MEMORY_BUFFER_TYPE memory_buffers[] =
{
    MM_L0R_INTERMEDIATE_STATISTICS,
    MM_ABSOLUTE_GAINS,
    MM_RELATIVE_GAINS,
    MM_TIRS_DARK_RESPONSE,
    MM_TIRS_DARK_RESPONSE_BLIND,
    MM_TIRS_ABSOLUTE_GAINS_BLIND,
    MM_TIRS_RELATIVE_GAINS_BLIND,
    MM_TIRS_SCA_DARK_RESPONSE_DIFFERENCE
};

static int number_of_memory_buffers =
    sizeof(memory_buffers) / sizeof(MEMORY_BUFFER_TYPE);


/**************************************************************************
 NAME:     decrement_counters

 PURPOSE:  Internal routine to decrement the consumer counts of all required
           memory buffers

 RETURNS: SUCCESS or ERROR
***************************************************************************/
static int decrement_counters
(
    int band_index,     /* I: */
    int sca_index       /* I: */
)
{
    RPS_BAND_ATTRIBUTES *band_attributes = get_band_attributes( band_index );
    int i;              /* Local loop counter */
    int status;

    /* Decrement the consumer counts for the memory buffers */
    for (i = 0; i < number_of_memory_buffers; i++)
    {
        status = decrement_consumer_count(memory_buffers[i], band_index,
            sca_index);
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }

        /* If the band has an associated blind band and it's set to process the
           blind band, decrement the user count */
        if (band_attributes->blind_band_number > 0
            && band_attributes->process_blind_band_flag )
        {
            status = decrement_consumer_count(memory_buffers[i],
                band_attributes->blind_band_index, sca_index);
            if ( status != SUCCESS )
            {
                IAS_LOG_ERROR("Decrementing memory buffer");
                return ERROR;
            }
        }
    }
    return SUCCESS;
}   /* END decrement_counters */



/************************** MAIN API FUNCTION **************************/
int api_tirs_dark_response_determination
(
    BLOB *blob,         /* I: Various input parameters */
    int band_index,     /* I: */
    int sca_index,      /* I: */
    int algorithm_index /* I: Index to algorithm list */
)
{
    /* RPS-specific band attributes information */
    RPS_BAND_ATTRIBUTES *rps_atts = get_band_attributes(band_index);
    int band_number = rps_atts->band_number;
    int sca_number = sca_index + 1;

    /* Histogram statistics intermediate results for the blind band */
    IAS_RPS_HISTOGRAM_INTERMEDIATE_RESULTS *blind_stats = NULL;

    /* CPF quantities */
    double *dark_response = NULL;           /* CPF baseline dark response
                                               of all detectors in the
                                               current band/SCA */
    const double *relative_gains = NULL;    /* Relative gains for the
                                               current band/SCA */
    const double *absolute_gain = NULL;     /* Pointer for absolute gain
                                               the current band/SCA */

    /* Memory buffer to hold the mean blind band dark response difference
       value for the current SCA */
    double *sca_dark_response_difference = NULL;

    int blind_band_number;                  /* TIRS blind band number */
    int number_detectors;                   /* Number of detectors in the
                                               current SCA */
    int is_blind_band = 0;                  /* Blind band flag */
    int curr_band_number;                   /* Current band number of
                                               processing band */
    int curr_band_index;
    int status;

    /* Make sure we've got the correct spectral type AND classification */
    if (rps_atts->spectral_type != IAS_SPECTRAL_THERMAL)
        return SUCCESS;

    if ((rps_atts->band_classification != IAS_NORMAL_BAND)
          && (rps_atts->band_classification != IAS_BLIND_BAND))
    {
        status = decrement_counters(band_index, sca_index);
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }
        return SUCCESS;
    }

    /* Initialize curr_band_number to ... the current input band number */
    curr_band_number = band_number;
    curr_band_index = band_index;

    /* NOTE: It isn't necessary to decrement the usage of the memory buffer
       on error conditions.  The main thread will clean up allocated buffers
       when error conditions dictate.  It doesn't hurt but it isn't
       necessary. */

    /* Set the number of detectors */
    number_detectors = rps_atts->number_detectors;

    /* Get the blind band number */
    blind_band_number = rps_atts->blind_band_number;

    /* Determine if we're processing the blind band.  This flag will be
       set when we process the blind band in the first pass of the
       characterization */
    if (blob->process_tirs_blind_band)
    {
        /* Set the flag indicating we're processing the blind band */
        is_blind_band = 1;

        /* "Reset" the current band number to the blind band number */
        curr_band_number = blind_band_number;
        curr_band_index = rps_atts->blind_band_index;

        /* Get the blind band gains */
        absolute_gain = (const double *)get_const_memory_buffer(
            MM_TIRS_ABSOLUTE_GAINS_BLIND, curr_band_index, sca_index);

        relative_gains = (const double *)get_const_memory_buffer(
            MM_TIRS_RELATIVE_GAINS_BLIND, curr_band_index, sca_index);

        /* Get the blind band CPF dark response */
        dark_response = (double *)get_memory_buffer(
            MM_TIRS_DARK_RESPONSE_BLIND, CONSUMER, curr_band_index, sca_index);

        /* Get the blind band histogram statistics scene means for the
           current SCA as well */
        blind_stats =
            (IAS_RPS_HISTOGRAM_INTERMEDIATE_RESULTS *)get_memory_buffer(
                 MM_L0R_INTERMEDIATE_STATISTICS, CONSUMER, curr_band_index,
                 sca_index);
        if (blind_stats == NULL)
        {
            IAS_LOG_ERROR("Retrieving TIRS blind band scene means data "
                "buffer for SCA %d", sca_number);
            return ERROR;
        }

        /* Get the mean dark response difference for the current SCA
           (or an empty buffer for it) */
        sca_dark_response_difference = (double *)get_memory_buffer(
            MM_TIRS_SCA_DARK_RESPONSE_DIFFERENCE, SETTER, curr_band_index,
            sca_index);
    }
    else
    {
        /* Get the imaging band gains */
        absolute_gain = (const double *)get_const_memory_buffer(
            MM_ABSOLUTE_GAINS, band_index, sca_index);

        relative_gains = (const double *)get_const_memory_buffer(
            MM_RELATIVE_GAINS, band_index, sca_index);

        /* Get the imaging band CPF dark response */
        dark_response = (double *)get_memory_buffer(
            MM_TIRS_DARK_RESPONSE, CONSUMER, band_index, sca_index);

        /* Get the mean dark response difference for the current SCA
           (or an empty buffer for it) */
        sca_dark_response_difference = (double *)get_memory_buffer(
            MM_TIRS_SCA_DARK_RESPONSE_DIFFERENCE, CONSUMER, band_index,
            sca_index);
    }

    /* Make sure we've got valid data buffers */
    if ((absolute_gain == NULL) || (relative_gains == NULL)
            || (dark_response == NULL))
    {
        IAS_LOG_ERROR("Retrieving absolute/relative gains and/or "
            "baseline dark response for band %d SCA %d from CPF",
            band_number, sca_number);
        return ERROR;
    }
    if (sca_dark_response_difference == NULL)
    {
        IAS_LOG_ERROR("Retrieving SCA dark response difference data "
            "buffer for SCA %d", sca_number);
        return ERROR;
    }

    /* Call the algorithm for the current band/SCA if an adjusted dark
       response should be determined (BLOB flag indicating to use the
       CPF baseline dark response is not set).  Otherwise, this will just
       pass the CPF baseline dark response */
    if (!blob->use_tirs_cpf_baseline_response)
    {
        status = ias_rps_tirs_dark_response_determination(
            blind_stats, dark_response, sca_dark_response_difference,
            absolute_gain[0], relative_gains, is_blind_band,
            number_detectors,
            blob->pixel_mask[curr_band_index][sca_index]);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Determining adjusted dark response");
            return ERROR;
        }
        set_buffer_filled_flag( MM_TIRS_SCA_DARK_RESPONSE_DIFFERENCE,
            curr_band_index, sca_index );
    }

    /* If specified to do so in the blob, write the dark response to an
       ODL-formatted text file.  This allows for the possibility of 
       copying the baseline CPF dark response to a file for easier 
       viewing, etc */
    if (blob->write_tirs_dark_response_to_file)
    {
        status = ias_rps_tirs_dark_response_write_to_group_file(
            dark_response, is_blind_band, curr_band_number, sca_number,
            number_detectors);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Writing band %d SCA %d adjusted dark response "
                "to group file", curr_band_number, sca_number);
            return ERROR;
        }
    }

    /* Decrement the consumer counts */
    status = decrement_counters(band_index, sca_index);
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Decrementing memory buffer");
        return ERROR;
    }

    return SUCCESS;
}   /* END api_tirs_dark_response_determination */




/*-----------------------------------------------------------------------
 NAME:           tirs_dark_response_determination_preprocess

 PURPOSE:  Sets up memory management for the TIRS dark response
           determination algorithm.  For the imaging bands, it also
           reads the SCA dark response files produced during blind
           band processing.  For the blind band, no action is taken

 RETURNS:  Integer status code of SUCCESS or ERROR
-------------------------------------------------------------------------*/
int tirs_dark_response_determination_preprocess
(
    BLOB *blob          /* I - Miscellaneous parameters */
)
{
    char sca_mean_response_diff_filename[PATH_MAX] = "";
    double *sca_dark_response_difference = NULL;
    double value ;
    FILE *response_fptr = NULL;
    int band_list[IAS_MAX_NBANDS];
    int band_number;
    int num_image_bands;
    int status = SUCCESS;
    int sca_index;
    int sca_number;                /* Current SCA number */
    int num_scas;
    int sca;                       /* SCA number read from response
                                      difference file */
    int i;                         /* General loop counter */


    /* Increment the consumer counts for the memory buffers */
    for (i = 0; i < number_of_memory_buffers; i++)
    {
        status = increment_consumer_count(memory_buffers[i]);
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Incrementing %s consumer count",
                get_memory_buffer_string(memory_buffers[i]));
            return ERROR;
        }
    }

    /* Increment the consumer count for MM_TIRS_SCA_DARK_RESPONSE_DIFFERENCE 
       for post-process use but only if it will be used */
    if (blob->process_tirs_blind_band)
    {
        status = increment_consumer_count(MM_TIRS_SCA_DARK_RESPONSE_DIFFERENCE);
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Incrementing %s consumer count",
                get_memory_buffer_string(MM_TIRS_SCA_DARK_RESPONSE_DIFFERENCE));
            return ERROR;
        }   
    }

    /* Set the flag in the blob to indicate we're processing TIRS
       dark response data */
    blob->tirs_dark_response_processing = TRUE;

    /* Get the list of imaging bands */
    status = ias_sat_attr_get_sensor_band_numbers(IAS_TIRS, IAS_NORMAL_BAND,
        0, band_list, IAS_MAX_NBANDS, &num_image_bands);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving TIRS image band list");
        return ERROR;
    }

    /* Get the number of SCAs */
    num_scas = ias_sat_attr_get_sensor_sca_count(IAS_TIRS);
    if (num_scas == ERROR)
    {
        IAS_LOG_ERROR("Retrieving number of TIRS SCAs");
        return ERROR;
    }

    /* If we're not processing the blind band, get the adjusted dark
       response differences for each SCA.  We'll need to do this for
       all normal imaging bands */
    if (!blob->process_tirs_blind_band)
    {
        /* Process each SCA */
        for (sca_index = 0; sca_index < num_scas; sca_index++)
        {
            sca_number = sca_index + 1;

            /* Open the individual dark response file for the current SCA
               and get its data */
            status = snprintf(sca_mean_response_diff_filename,
                sizeof(sca_mean_response_diff_filename),
                "response_diff_SCA%02d.txt", sca_number);
            if ((status < 0)
                || (status >= sizeof(sca_mean_response_diff_filename)))
            {
                IAS_LOG_ERROR("Constructing SCA %d response difference "
                    "file name", sca_number);
                return ERROR;
            }
            response_fptr = fopen(sca_mean_response_diff_filename, "r");
            if (response_fptr == NULL)
            {
                IAS_LOG_ERROR("Cannot open SCA dark response difference "
                    "file %s", sca_mean_response_diff_filename);
                return ERROR;
            }
            status = fscanf(response_fptr, "%d,%lf\n", &sca, &value);
            fclose(response_fptr);
            if (status < 0)
            {
                IAS_LOG_ERROR("Reading blind band dark response "
                    "for SCA %d", sca_number);
                return ERROR;
            }

            /* Load up the proper dark response difference value for each
               band/SCA */
            for (i = 0; i < num_image_bands; i++)
            {
                /* Get the correct band number for the current imaging band */
                band_number = band_list[i];
                int band_index = ias_sat_attr_convert_band_number_to_index(
                    band_number );
                if ( band_index == ERROR )
                {
                    IAS_LOG_ERROR("Getting band index for band %d",
                        band_number );
                    return ERROR;
                }

                /* Get an empty memory buffer */
                sca_dark_response_difference =
                    (double *)preprocess_get_memory_buffer(
                    MM_TIRS_SCA_DARK_RESPONSE_DIFFERENCE, band_index,
                    sca_index);
                if (sca_dark_response_difference == NULL)
                {
                    IAS_LOG_ERROR("Obtaining memory buffer for band %d "
                       "to read SCA %d coefficient", band_number,
                       sca_number);
                    return ERROR;
                }

                /* Copy the value to the memory buffer */
                *sca_dark_response_difference = value;
                set_buffer_filled_flag( MM_TIRS_SCA_DARK_RESPONSE_DIFFERENCE,
                    band_index, sca_index );

            }   /* END band loop */
        }       /* END SCA loop */
    }           /* END IF block */

    return SUCCESS;
}   /* END -- tirs_dark_response_determination_preprocess */




/*-----------------------------------------------------------------------
 NAME:           tirs_dark_response_determination_postprocess

 PURPOSE:  Performs required post-processing actions for TIRS
           dark response determination processing.  For the blind band,
           it creates the SCA dark response difference files needed for
           processing by the imaging bands.  For the image bands,
           no action is taken

 RETURNS:  Integer status code of SUCCESS or ERROR
-------------------------------------------------------------------------*/
int tirs_dark_response_determination_postprocess
(
    BLOB *blob          /* I - Miscellaneous parameters */
)
{
    char sca_mean_response_diff_filename[PATH_MAX] = "";
    char band_name[IAS_BAND_NAME_SIZE] = "";
    double *sca_dark_response_difference = NULL;
    FILE *response_fptr = NULL;
    IAS_BAND_TYPE band_type = IAS_THERMAL1_BAND;
    IAS_BAND_CLASSIFICATION band_classification = IAS_BLIND_BAND;
    int blind_band_number;
    int status;
    int sca_index;
    int sca_number;
    int num_scas;


    /* Get the blind band name and number */
    status = ias_sat_attr_get_band_name_from_type_and_class(band_type,
        band_classification,  band_name);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving current TIRS blind band name");
        return ERROR;
    }
    blind_band_number = ias_sat_attr_convert_band_name_to_number(
        band_name);
    if (blind_band_number == ERROR)
    {
        IAS_LOG_ERROR("Retrieving current TIRS blind band number");
        return ERROR;
    }
    /* Get the number of SCAs */
    num_scas = ias_sat_attr_get_scas_per_band(blind_band_number);
    if (num_scas == ERROR)
    {
        IAS_LOG_ERROR("Retrieving number of SCAs for the blind band");
        return ERROR;
    }

    /* Are we processing the blind band?  If so, write the dark
       response difference of each SCA to text files read during
       preprocessing of the image bands.  This assumes a separate
       run of the RPS driver to process the blind band by itself */
    if (blob->process_tirs_blind_band)
    {
        int blind_band_index = ias_sat_attr_convert_band_number_to_index(
            blind_band_number );
        if ( blind_band_index == ERROR )
        {
            IAS_LOG_ERROR("Getting the index for the TIRS blind band");
            return ERROR;
        }

        for (sca_index = 0; sca_index < num_scas; sca_index++)
        {
            sca_number = (sca_index + 1);

            /* Create the response file */
            status = snprintf(sca_mean_response_diff_filename,
                sizeof(sca_mean_response_diff_filename),
                "response_diff_SCA%02d.txt", sca_number);
            if ((status < 0)
                || (status >= sizeof(sca_mean_response_diff_filename)))
            {
                IAS_LOG_ERROR("Creating SCA %d response difference file "
                    "name", sca_number);
                return ERROR;
            }
            response_fptr = fopen(sca_mean_response_diff_filename, "w");
            if (response_fptr == NULL)
            {
                IAS_LOG_ERROR("Creating SCA dark response difference file %s",
                    sca_mean_response_diff_filename);
                return ERROR;
            }

            /* Get the dark response difference for each SCA */
            sca_dark_response_difference = (double *)get_memory_buffer(
                MM_TIRS_SCA_DARK_RESPONSE_DIFFERENCE, CONSUMER,
                blind_band_index, sca_index);
            if (sca_dark_response_difference == NULL)
            {
                IAS_LOG_ERROR("Retrieving blind band SCA dark response "
                    "difference for SCA %d", sca_number);
                fclose(response_fptr);
                return ERROR;
            }

            /* Write it to the file */
            status = fprintf(response_fptr, "%d,%lf\n", sca_number,
                *sca_dark_response_difference);
            if (status < 0)
            {
                IAS_LOG_ERROR("Writing SCA %d dark response difference to "
                    "response difference file %s",
                    sca_number, sca_mean_response_diff_filename);
                fclose(response_fptr);
                return ERROR;
            }

            /* Decrement the consumer count */
            status = decrement_consumer_count(
                MM_TIRS_SCA_DARK_RESPONSE_DIFFERENCE, blind_band_index,
                sca_index);
            if ( status != SUCCESS )
            {
                IAS_LOG_ERROR("Decrementing memory buffer");
                fclose(response_fptr);
                return ERROR;
            }

            /* Close the file */
            if (fclose(response_fptr))
            {
                IAS_LOG_ERROR("Properly closing response file %s",
                    sca_mean_response_diff_filename);
                return ERROR;
            }
        }
    }

    return SUCCESS;
}   /* END -- tirs_dark_response_determination_postprocess */
