/*----------------------------------------------------------------------------
NAME:    api_tirs_secondary_linearization

PURPOSE: Gather data needed for the TIRS secondary linearization algorithm
         and call the routine

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
#include <stdlib.h>
#include <string.h>
#include "ias_logging.h"
#include "ias_rlut.h"
#include "rps_local.h"

/* Define a list of memory buffers needed by the algorithm */
static MEMORY_BUFFER_TYPE memory_buffers[] =
{
    IMAGE_L1R,
    MM_TIRS_SECONDARY_LINEARIZATION_PARAMS
};

static int number_of_memory_buffers =
    sizeof(memory_buffers) / sizeof(MEMORY_BUFFER_TYPE);


/*---------------------------------------------------------------------------
 NAME:     decrement_counters

 PURPOSE:  Internal routine to decrement the consumer counts of all required
           memory buffers

 RETURNS: SUCCESS or ERROR
---------------------------------------------------------------------------*/
static int decrement_counters
(
    int band_index,     /* I: */
    int sca_index       /* I: */
)
{
    RPS_BAND_ATTRIBUTES *band_attributes = get_band_attributes( band_index );
    int i;              /* Local loop counter */
    int status;

    /* Decrement the consumer counts for the memory buffers */
    for (i = 0; i < number_of_memory_buffers; i++)
    {
        status = decrement_consumer_count(memory_buffers[i], band_index,
            sca_index);
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }

        /* If the band has an associated VRP band decrement the user count */
        if (band_attributes->vrp_band_number > 0)
        {
            if ( band_attributes->vrp_band_number > 0 )
            {
                decrement_consumer_count(memory_buffers[i],
                    band_attributes->vrp_band_index, sca_index );
            }
        }

        /* If the band has an associated blind band and it's set to process the
           blind band, decrement the user count */
        if (band_attributes->blind_band_number > 0
            && band_attributes->process_blind_band_flag )
        {
            status = decrement_consumer_count(memory_buffers[i],
                band_attributes->blind_band_index, sca_index);
            if ( status != SUCCESS )
            {
                IAS_LOG_ERROR("Decrementing memory buffer");
                return ERROR;
            }
        }
    }
    return SUCCESS;
}   /* END decrement_counters */

/*----------------------------------------------------------------------------
NAME:    api_tirs_secondary_linearization

PURPOSE: Gather data needed for the TIRS secondary linearization algorithm
         and call the routine

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int api_tirs_secondary_linearization
(
    BLOB *blob,         /* I: Miscellaneous parameters */
    int   band_index,   /* I: */
    int   sca_index,    /* I: */
    int   algo_index    /* I: Algorithm index */
)
{
    RPS_BAND_ATTRIBUTES *band_attributes = get_band_attributes(band_index);
    int band_number = band_attributes->band_number;
    int number_lines = band_attributes->number_lines;
    float *data_buffer;            /* Level 1 data buffer */
    int status;
    IAS_RLUT_LOOKUP_TABLE *params;

    /* This is only to be run for TIRS bands */
    if ( band_attributes->spectral_type != IAS_SPECTRAL_THERMAL )
    {
        status = decrement_counters( band_index, sca_index );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }
        return SUCCESS;
    }

    /* Set the processing level for the current band / SCA */
    if ( set_rps_level( band_index, sca_index, 
            IAS_RPS_LEVEL_TIRS_2ND_LINEARIZATION ) == ERROR )
    {
        IAS_LOG_ERROR("Updating RPS processing level");
        return ERROR;
    }

    /* This algorithm doesn't process TIRS blind bands therefore the
        processing level for the blind band *must* be set to skip */
    if ( band_attributes->blind_band_index >= 0 &&
         band_attributes->process_blind_band_flag )
    {
        if ( set_rps_level( band_attributes->blind_band_index, sca_index,
            IAS_RPS_LEVEL_SKIP ) == ERROR )
        {
            IAS_LOG_ERROR("Updating RPS processing level");
            return ERROR;
        }
    }

    /* Get the data buffer */
    data_buffer = (float *)get_memory_buffer( IMAGE_L1R, CONSUMER, band_index,
        sca_index );
    if ( ! data_buffer )
    {
        IAS_LOG_ERROR("Getting image memory buffer");
        return ERROR;
    }

    /* Get the parameters */
    params = (IAS_RLUT_LOOKUP_TABLE *)
        get_memory_buffer( MM_TIRS_SECONDARY_LINEARIZATION_PARAMS, CONSUMER,
        band_index, sca_index );
    if ( ! params )
    {
        IAS_LOG_ERROR("Getting TIRS secondary linearization parameters for "
            "band %d SCA %d", band_number, sca_index + 1 );
        return ERROR;
    }

    /* Apply the linearization using the alternate response linearization
       library routine */
    status = ias_rps_lookup_response_linearization( data_buffer, params,
        number_lines );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("TIRS secondary linearization for band %d SCA %d",
            band_number, sca_index + 1 );
        return ERROR;
    }

    /* Decrement the memory management counters */
    status = decrement_counters(band_index, sca_index);
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Decrementing memory buffer");
        return ERROR;
    }

    return SUCCESS;
}  /* END api_tirs_secondary_linearization */


/*----------------------------------------------------------------------------
NAME:    tirs_secondary_linearization_preprocess

PURPOSE: Do preprocessing for the algorithm by incrementing the consumer
         counts.  

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int tirs_secondary_linearization_preprocess
(
    BLOB *blob          /* I: Miscellaneous parameters */
)
{
    int i;
    int status;

    /* Ensure the RLUT is open and open it if not */
    if ( blob->rlut == NULL )
    {
        blob->rlut = ias_rlut_open_file( blob->rlut_filename, IAS_READ );
        if ( ! blob->rlut )
        {
            IAS_LOG_ERROR("Opening RLUT file %s to read",
                blob->rlut_filename );
            return ERROR;
        }
    }

    /* Increment the consumer counts for the memory buffers */
    for ( i = 0; i < number_of_memory_buffers; i++ )
    {
        status = increment_consumer_count( memory_buffers[i] );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Incrementing %s consumer count",
                get_memory_buffer_string( memory_buffers[i] ));
            return ERROR;
        }
    }

    return SUCCESS;
}  /* END tirs_secondary_linearization_preprocess */
