/* Standard Library Includes */
#include <stdlib.h>
#include <string.h>

/* IAS Library Includes */
#include "ias_logging.h"
#include "ias_l1r.h"

/* Local Includes */
#include "rps_local.h"
#include "rps_algorithm.h"

/*-------------------------------------------------------------------------
NAME:    tirs_stray_light_correction_preprocess

PURPOSE: Register the memory buffers needed by this algorithm

RETURNS: SUCCESS or ERROR
--------------------------------------------------------------------------*/
int tirs_stray_light_correction_preprocess
(
    BLOB *blob          /* I: Miscellaneous parameters */
)
{
    /* Increment the L1R consumer count */
    if (increment_consumer_count(IMAGE_L1R) != SUCCESS)
    {
        IAS_LOG_ERROR("Incrementing %s consumer count",
            get_memory_buffer_string(IMAGE_L1R));
        return ERROR;
    }

    return SUCCESS;
}  /* END tirs_stray_light_correction_preprocess */

/*----------------------------------------------------------------------------
NAME:    api_tirs_stray_light_correction

PURPOSE: Subtracts the 'ghost' image from the L1R output imagery for TIRS bands

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int api_tirs_stray_light_correction
(
    BLOB *blob,         /* I: Miscellaneous parameters */
    int   band_index,   /* I: */
    int   sca_index,    /* I: */
    int   algo_index    /* I: Algorithm index */
)
{
    RPS_BAND_ATTRIBUTES *band_attributes = get_band_attributes(band_index);

    /* Check if current band needs stray light correction */
    if (band_attributes->sensor_id == IAS_TIRS)
    {
        IAS_THREAD_MUTEX_TYPE *hdf_io_mutex = get_hdf_io_mutex();
        int band_number = band_attributes->band_number;
        int sca_number = sca_index + 1;
        int num_lines = band_attributes->number_lines;
        int num_samps = band_attributes->number_detectors;
        int samp_start = 0;
        int line_start = 0;
        int buffer_size = num_lines * num_samps;
        float *tirs_stray_light_buffer;
        float *l1r_buffer;
        int index;

        tirs_stray_light_buffer = malloc(sizeof(float) * buffer_size);
        if (!tirs_stray_light_buffer)
        {
            IAS_LOG_ERROR("Allocating memory for band number %d SCA "
                "number %d", band_number, sca_number);
            return ERROR;
        }

        /* Lock the HDF io mutex */
        if (IAS_THREAD_LOCK_MUTEX(hdf_io_mutex))
        {
            IAS_LOG_ERROR("Locking hdf io mutex for band number %d SCA number "
                "%d", band_number, sca_number);
            free(tirs_stray_light_buffer);
            return ERROR;
        }

        /* Read in the current SCA stray light correction data */
        if (ias_l1r_read_image(blob->tirs_stray_light_band_io[band_index],
            sca_index, line_start, samp_start, num_lines, num_samps,
            tirs_stray_light_buffer) != SUCCESS)
        {
            IAS_LOG_ERROR("Reading stray light correction band number %d "
                "SCA number %d", band_number, sca_number);
            IAS_THREAD_UNLOCK_MUTEX(hdf_io_mutex);
            free(tirs_stray_light_buffer);
            return ERROR;
        }

        /* Unlock the HDF io mutex */
        if (IAS_THREAD_UNLOCK_MUTEX(hdf_io_mutex))
        {
            IAS_LOG_ERROR("Unlocking hdf io mutex for band number %d SCA "
                "number %d", band_number, sca_number);
            free(tirs_stray_light_buffer);
            return ERROR;
        }

        /* Retrieve the L1R buffer for this Band/SCA index combination */
        l1r_buffer = (float *)get_memory_buffer(IMAGE_L1R, SETTER,
            band_index, sca_index);
        if (!l1r_buffer)
        {
            IAS_LOG_ERROR("Retrieving output L1R buffer for band number %d "
                "SCA number %d", band_number, sca_number);
            free(tirs_stray_light_buffer);
            return ERROR;
        }

        /* Subtract the stray light correction data from the output L1R */
        for (index = 0; index < buffer_size; index++)
        {
            l1r_buffer[index] -= tirs_stray_light_buffer[index];
        }

        /* Free the stray light buffer */
        free(tirs_stray_light_buffer);
    }

    /* Decrement the memory management consumer counts */
    if (decrement_consumer_count(IMAGE_L1R, band_index, sca_index) != SUCCESS)
    {
        IAS_LOG_ERROR("Decrementing memory buffer");
        return ERROR;
    }
    if (band_attributes->vrp_band_number > 0)
    {
        if (decrement_consumer_count(IMAGE_L1R, band_attributes->vrp_band_index,
            sca_index) != SUCCESS)
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }
    }
    if (band_attributes->blind_band_number > 0
        && band_attributes->process_blind_band_flag)
    {
        if (decrement_consumer_count(IMAGE_L1R,
            band_attributes->blind_band_index, sca_index) != SUCCESS)
        {
            IAS_LOG_ERROR("Decrementing memory buffer");
            return ERROR;
        }
    }

    return SUCCESS;
}  /* END api_tirs_stray_light_correction */
