/*----------------------------------------------------------------------------
NAME:    assign_work
         initialize_work_assignment
         cleanup_work_assignment
         synchronize_query
         synchronize_broadcast
         set_work_assignment_error_condition

         Internal routines for outputting debug information.
             set_state_complete
             open_debug_output_file
             close_debug_output_file
             print_debug_information

PURPOSE: Determine a work assignment for a processing thread including
         routines to initialize and clean up.  The set_next_algorithm
         routine allows external routines to atomically set state
         structure members.  Finally two routines to synchronize the
         threadpool are included for serialized routine support.

RETURNS: various
-----------------------------------------------------------------------------*/
#include <stdlib.h>

#define IAS_LOG_CHANNEL "assign_work"

#include "ias_logging.h"
#include "ias_threadpool.h"
#include "ias_threadsync.h"
#include "rps_local.h"

/* Structure to hold the info for the SYNC_NEED_ADJACENT_SCAS condition */
typedef struct adjacent_sca_condition
{
    int algo_index;     /* algorithm index of the synchronized algorithm */
    int first_algorithm_for_band; /* flag that indicates this is the first
                                     algorithm for the band */
} ADJACENT_SCA_CONDITION;

/* Mutex to control access to condition variable, work state structure
   and the other variables needed to maintain the current state */
static IAS_THREAD_MUTEX_TYPE work_assignment_mutex;

/* Condition variables which are used to control threads that encounter
   serialized algorithms.  These array is the number of bands to process in
   size. */
static int number_allocated = 0;
static IAS_THREAD_COND *work_assignment_condition = NULL;

static int number_algorithms;  /* Total number of algorithms */
static int number_state_struc; /* Number algorithms * number bands */
/* This array of structures is sized as number_state_struc.  Internal to the
   structure is a IAS_MAX_NSCAS sized state array.  It's purpose is to store
   the work state for each algorithm, band and SCA combination. */
static WORK_STATE_STRUCT *work_state;

/* Any thread encountering an error will set this flag to 1 (true).  The
   assign_work routine will return with no work when the flag is true. */
static int error_condition_encountered = 0;

/* Two arrays for finding a the band index that can be used for determining
   an index into the work_state structure.  The band LUT is a copy of the
   list of indices passed into the initialization routine.  The reverse LUT
   gives an index into the band LUT given a band index.  For example if the
   band numbers to process are 1, 3, 5, 8 and 10, the two arrays look like
   the following.  The u stands for unused.

             index 0 1 2 3 4 5 6 7 8 9 10 ...
          band_lut 0 2 4 7 9 u ...
   band_reverselut 0 u 1 u 2 u u 3 u 4 u ...

   Given a band index of 7 the reverse lut says the work state index is 3. */
static int *band_lut;
static int *band_reverselut;

/* Local copy of the thread pool.  It is used to check the thread pool for
   error conditions. */
static struct ias_threadpool *local_pool;

/* File descriptor for the debug output file and prototypes for the debug
   routines */
static char debug_filename[] = "assign_work_debug.txt";
static FILE *debug_fd = NULL;
static int open_debug_output_file(void);
static void close_debug_output_file(void);
#if 0
/* Disable this since it isn't currently being used */
static int print_debug_information
(
    int band_index,     /* I: */
    int sca_index,      /* I: */
    int algo_index      /* I: */
);
#endif

/*------------------------------------------------------------------------
NAME:    check_function

PURPOSE: Tests for the condition equal to zero.  A pointer to this function
         will be passed to the thread library routine
         ias_thread_wait_on_condition() routine which will call this routine.

RETURNS: 1 (true) the condition is met
         0 (false) the condition isn't met
------------------------------------------------------------------------*/
static int check_function
(
    void *check_parameter       /* I: Pointer to the parameter to check */
)
{
    char *state = (char *)check_parameter;

    if ( *state == COMPLETE || error_condition_encountered )
        return 1;
    else
        return 0;
}


/*------------------------------------------------------------------------
NAME:    set_work_assignment_error_condition

PURPOSE: This routine sets the error encountered flag to 1 (true)

RETURNS: none
------------------------------------------------------------------------*/
void set_work_assignment_error_condition( void )
{
    int index;

    error_condition_encountered = 1;

    /* Broadcast to all the condition variables to make sure they wake up.
       Don't bother checking for errors since there isn't any more that can be
       done about them now. */
    IAS_THREAD_LOCK_MUTEX( &work_assignment_mutex );

    for (index = 0; index < number_allocated; index++)
    {
        ias_thread_broadcast_condition( &work_assignment_mutex,
            &work_assignment_condition[index] );
    }

    IAS_THREAD_UNLOCK_MUTEX( &work_assignment_mutex );
}


/*------------------------------------------------------------------------
NAME:    set_state_complete

PURPOSE: Set the SCAs algorithm index in the state structure to the value
         passed in and mark it as complete.

RETURNS: SUCCESS or ERROR
------------------------------------------------------------------------*/
static int set_state_complete
(
    int band_index,     /* I: Band */
    int sca_index,      /* I: SCA */
    int algorithm_index /* I: Index of algorithm */
)
{
    int work_array_index;  /* Index into the work state structure */

    /* Prevent band_index errors */
    if ( band_reverselut[band_index] == -1 )
    {
        IAS_LOG_ERROR("Access the reverse look up table resulted in a "
            "a negative index: band_index = %d", band_index );
        return ERROR;
    }
    
    work_array_index = ( band_reverselut[band_index] * number_algorithms ) +
        algorithm_index;

    work_state[work_array_index].state[sca_index] = COMPLETE;

    return SUCCESS;
}

/*------------------------------------------------------------------------
NAME:    set_algorithm_sync_type

PURPOSE: Set the sync type for an algorithm.

RETURNS: SUCCESS or ERROR
------------------------------------------------------------------------*/
int set_algorithm_sync_type
(
    int algorithm_index,           /* I: Index of algorithm */
    ALGORITHM_SYNC_TYPE sync_type  /* I: Sync type to use */
)
{
    int band_index;
    int band_count = number_state_struc / number_algorithms;

    /* Make sure the algorithm index is legal */
    if (algorithm_index > number_algorithms)
    {
        IAS_LOG_ERROR("Invalid algorithm index %d when algorithm count it %d",
            algorithm_index, number_algorithms);
        return ERROR;
    }

    /* Set the sync type for each of the bands in the work state array for
       the indicated algorithm */
    for (band_index = 0; band_index < band_count; band_index++)
    {
        int work_array_index;  /* Index into the work state structure */

        work_array_index = ( band_index * number_algorithms )
                + algorithm_index;

        work_state[work_array_index].sync_type = sync_type;
    }

    return SUCCESS;
}

/*------------------------------------------------------------------------
NAME: are_adjacent_scas_done

PURPOSE: Test whether the adjacent SCAs for the indicated algorithm index
    have been completed.

RETURNS:
    1 = adjacent SCAs are done
    0 = adjacent SCAs are not done
------------------------------------------------------------------------*/
static int are_adjacent_scas_done
(
    int algo_index,     /* I: algorithm index to check */
    int sca_index       /* I: starting SCA index to check (checks this index
                              and index + 1) */
)
{
    if ((work_state[algo_index].state[sca_index] == COMPLETE)
        && (work_state[algo_index].state[sca_index + 1] == COMPLETE))
    {
        return 1;
    }
    else
    {
        return 0;
    }
}

/*------------------------------------------------------------------------
NAME: count_adjacent_scas_ready

PURPOSE: Count the number of adjacent SCA pairs that have completed the
    previous algorithm and are ready for the current algorithm to be run.

RETURNS: Count of adjacent SCA pairs that are ready
------------------------------------------------------------------------*/
static int count_adjacent_scas_ready
(
    int first_algorithm_for_band,   /* I: flag to indicate this is the first
                                          algorithm for the band */
    int algo_index                  /* I: algorithm index to count */
)
{
    int prev_algo = algo_index - 1; /* previous algorithm index */
    int count = 0;
    int i;

    /* Look over the SCAs present for this algorithm index */
    for (i = 0; i < work_state[algo_index].nscas - 1; i++)
    {
        /* If the two SCAs are marked as complete in the previous algorithm, it
           can be worked on (or if it is the first algorithm for a band) */
        if (first_algorithm_for_band || are_adjacent_scas_done(prev_algo, i))
        {
            /* If this SCA is not yet assigned, count it as one that is ready */
            if (work_state[algo_index].state[i] == NOT_ASSIGNED)
                count++;
        }
    }

    return count;
}

/*------------------------------------------------------------------------
NAME:    check_adjacent_sca_condition

PURPOSE: Checks for adjacent SCAs to be available, or the entire algorithm
         to be completed.  A pointer to this function will be passed to the
         thread library routine ias_thread_wait_on_condition() routine which
         will call this routine.

RETURNS: 1 (true) the condition is met
         0 (false) the condition isn't met
------------------------------------------------------------------------*/
static int check_adjacent_sca_condition
(
    void *check_parameter       /* I: Pointer to the parameter to check */
)
{
    /* For this routine, the parameter is a pointer to an
       ADJACENT_SCA_CONDITION structure */
    const ADJACENT_SCA_CONDITION *parm
        = (const ADJACENT_SCA_CONDITION *)check_parameter;
    int algo_index = parm->algo_index;
    int i;

    /* Consider the condition met if an error happened */
    if (error_condition_encountered)
        return 1;

    /* If any SCAs are ready, the condition is met */
    if (count_adjacent_scas_ready(parm->first_algorithm_for_band, algo_index)
        > 0)
    {
        return 1;
    }

    /* If all the scas for this algorithm are completed, the condition is met */
    for (i = 0; i < work_state[algo_index].nscas; i++)
    {
        /* If the state is not complete, condition is not met */
        if (work_state[algo_index].state[i] != COMPLETE)
        {
            return 0;
        }
    }

    /* All the SCAs are completed, so the condition has been met */
    return 1;
}

/*------------------------------------------------------------------------
NAME: wait_for_adjacent_scas

PURPOSE: Waits for adjacent SCAs to be available, or the entire algorithm
     to be completed.  Should be called with the work_assignment_mutex locked.

RETURNS: SUCCESS or ERROR
------------------------------------------------------------------------*/
static int wait_for_adjacent_scas
(
    int first_algorithm_for_band,   /* I: flag to indicate this is the first
                                          algorithm for the band */
    int algo_index                  /* I: algorithm index to count */
)
{
    /* Need to sleep, so wait on the condition variable for this band's work
       order state */
    int condition_index = band_reverselut[work_state[algo_index].band];
    ADJACENT_SCA_CONDITION cond;

    cond.algo_index = algo_index;
    cond.first_algorithm_for_band = first_algorithm_for_band;

    /* Wait on the condition variable.  Note that this call will not return
       until the wait is over.  That is why the "cond" variable declared local
       (on the stack) here is valid to pass into this routine. */
    return ias_thread_wait_on_condition( &work_assignment_mutex,
                &work_assignment_condition[condition_index],
                check_adjacent_sca_condition, &cond );
}

/*------------------------------------------------------------------------
NAME:    assign_work

PURPOSE: This routine determines a work assignment for the calling thread 
         On input the indices indicate the work just completed by the
         thread and on output they are the assignment.  The initial call to
         this routine should be called with a band index of -1 to indicate
         that no work has been completed.

RETURNS: The return status of this routine is 
         1 (true) successfully found work for the caller
         0 (false) either an error condition or no work found
------------------------------------------------------------------------*/
int assign_work
(
    int *band_index,            /* I/O: Completed/Assigned band */
    int *sca_index,             /* I/O: Completed/Assigned SCA */
    int *algorithm_index        /* I/O: Completed/Assigned algorithm */
)
{
    int algo_count;             /* Algoritym loop counter */
    int sca_count;              /* SCA loop counter */
    int work_found = 0;         /* Flag to indicate whether work is found */
    int algo_increment = 1;     /* Increment size to use when stepping through
                                   algorithms.  It is normally one, but can be
                                   set to zero if the same algorithm needs to
                                   be re-evaluated on the next pass through the
                                   loop */

    /* Return no work to the caller if errors have occurred */
    if ( error_condition_encountered 
        || ias_threadpool_get_worker_return_status( local_pool ) != SUCCESS )
    {
        IAS_LOG_ERROR("Encountered error, setting the threads free");
        set_work_assignment_error_condition();
        return 0;
    }

    if ( IAS_THREAD_LOCK_MUTEX( &work_assignment_mutex ))
    {
        IAS_LOG_ERROR("Locking work assignment mutex");
        set_work_assignment_error_condition();
        return 0;
    }

    /* Set the work passed in as complete */
    if ( *band_index >= 0 )
    {
        int work_array_index = band_reverselut[*band_index] * number_algorithms
                + *algorithm_index;

        if (work_state[work_array_index].sync_type == SYNC_SINGLE_THREAD)
        {
            /* For the single threaded algorithms, set the state to complete
               for all the SCAs since one thread processes them all */
            int i;
            int condition_index = band_reverselut[*band_index];
            for (i = 0; i < work_state[work_array_index].nscas; i++)
            {
                set_state_complete( *band_index, i, *algorithm_index );
            }

            /* Wake up any threads that were waiting for this algorithm to
               complete */
            if (ias_thread_broadcast_condition( &work_assignment_mutex,
                    &work_assignment_condition[condition_index] ) != SUCCESS)
            {
                IAS_LOG_ERROR("Broadcasting condition update");
                IAS_THREAD_UNLOCK_MUTEX( &work_assignment_mutex );
                set_work_assignment_error_condition();
                return 0;
            }
        }
        else if (work_state[work_array_index].sync_type
                 == SYNC_NEED_ADJACENT_SCAS)
        {
            /* For algorithms that need the adjacent SCAs from the previous
               algorithm, the last SCA completing indicates that any threads
               waiting can be released */

            IAS_LOG_DEBUG("SYNC_NEED_ADJACENT_SCAS: setting band %d, sca %d,"
                "algorithm %d completed", *band_index, *sca_index,
                *algorithm_index);

            /* Complete the current band/sca/algorithm */
            set_state_complete( *band_index, *sca_index, *algorithm_index );

            /* If this is the last SCA for the algorithm, the algorithm is
               finished and any waiting threads can now be released to continue
               with any other work available */
            if (work_state[work_array_index].nscas - 1 == *sca_index)
            {
                int condition_index = band_reverselut[*band_index];

                IAS_LOG_DEBUG("SYNC_NEED_ADJACENT_SCAS: last sca done, "
                        "releasing threads");

                /* Wake up any threads that were waiting for this algorithm to
                   complete */
                if (ias_thread_broadcast_condition( &work_assignment_mutex,
                        &work_assignment_condition[condition_index] )
                        != SUCCESS)
                {
                    IAS_LOG_ERROR("Broadcasting condition update");
                    IAS_THREAD_UNLOCK_MUTEX( &work_assignment_mutex );
                    set_work_assignment_error_condition();
                    return 0;
                }
            }
        }
        else
        {
            /* For normal sync types, just set the state to complete */
            set_state_complete( *band_index, *sca_index, *algorithm_index );
        }

    }

    /* Loop through the algorithm state array */
    for ( algo_count = 0; algo_count < number_state_struc;
          algo_count += algo_increment )
    {
        int first_algorithm_for_band = 0;

        /* Assume the algorithm increment on the next iteration should be 1 */
        algo_increment = 1;

        /* If this is the first algorithm or the band number for the previous
           algorithm is less than the band number for the current algorithm,
           this is the first algorithm for the band */
        if ((algo_count == 0) || (band_reverselut[work_state[algo_count-1].band]
                 < band_reverselut[work_state[algo_count].band]))
        {
            first_algorithm_for_band = 1;
        }

        /* Loop through the SCAs looking for work */
        for ( sca_count = 0; sca_count < work_state[algo_count].nscas;
              sca_count++ )
        {
            /* If an error has been encountered in another thread, give up
               trying to assign work */
            if ( error_condition_encountered )
            {
                IAS_LOG_DEBUG("Giving up in assign work since an error was "
                        "detected");
                IAS_THREAD_UNLOCK_MUTEX( &work_assignment_mutex );
                return 0;
            }

            if ( work_state[algo_count].state[sca_count] == NOT_ASSIGNED )
            {
                /* If this algorithm is supposed to be single threaded, make
                   sure that all the previous algorithms have completed.  If
                   they haven't, just put this thread to sleep.  After the
                   thread wakes up, this algorithm will have been completed by
                   another thread, so it just needs to break out of the loop
                   and the next algorithm will be checked for work. */
                if (work_state[algo_count].sync_type == SYNC_SINGLE_THREAD)
                {
                    int go_to_sleep = 0;
                    int i;

                    /* If a thread is already working on this algorithm, 
                       we need to sleep */
                    for (i = 0; i < work_state[algo_count].nscas; i++)
                    {
                        if (work_state[algo_count].state[i] == WORKING)
                        {
                            go_to_sleep = 1;
                            break;
                        }
                    }

                    /* If this is not the first algorithm for the band and the
                       previous one is not complete, we need to sleep */
                    if ( !go_to_sleep && (algo_count == 0 ||
                         band_reverselut[work_state[algo_count-1].band] ==
                         band_reverselut[work_state[algo_count].band]))
                    {
                        for (i = 0; i < work_state[algo_count-1].nscas; i++)
                        {
                            if (work_state[algo_count-1].state[i] != COMPLETE)
                            {
                                go_to_sleep = 1;
                                break;
                            }
                        }
                    }

                    if (go_to_sleep)
                    {
                        /* Need to sleep, so wait on the condition variable
                           for this band's work order state */
                        int condition_index
                            = band_reverselut[work_state[algo_count].band];

                        if (ias_thread_wait_on_condition( 
                                &work_assignment_mutex,
                                &work_assignment_condition[condition_index],
                                check_function,
                                &work_state[algo_count].state[0] ) != SUCCESS)
                        {
                            IAS_LOG_ERROR("Broadcasting condition update");
                            IAS_THREAD_UNLOCK_MUTEX( &work_assignment_mutex );
                            set_work_assignment_error_condition();
                            return 0;
                        }

                        /* Break out of the SCA loop since all of them have
                           been processed when this thread wakes up */
                        break;
                    }
                }
                else if (work_state[algo_count].sync_type
                            == SYNC_NEED_ADJACENT_SCAS)
                {
                    int go_to_sleep = 0;
                    int can_assign = 0;
                    int i;
                    int condition_index
                            = band_reverselut[work_state[algo_count].band];
                    int adjacent_ready_to_start = 0;

                    /* Count how many SCAs are ready to start */
                    adjacent_ready_to_start = count_adjacent_scas_ready(
                            first_algorithm_for_band, algo_count);

                    IAS_LOG_DEBUG("SYNC_NEED_ADJACENT_SCAS: %d SCA pairs ready",
                            adjacent_ready_to_start);

                    /* The last SCA needs special handling since it can only
                       run after all the other SCAs are completed */
                    if (sca_count == work_state[algo_count].nscas - 1)
                    {
                        IAS_LOG_DEBUG("SYNC_NEED_ADJACENT_SCAS: Checking last "
                            "SCA for scheduling");

                        /* Look for any earlier SCAs that are not completed */
                        for (i = 0; i < work_state[algo_count].nscas - 1; i++)
                        {
                            if (work_state[algo_count].state[i] != COMPLETE)
                            {
                                IAS_LOG_DEBUG("SYNC_NEED_ADJACENT_SCAS: "
                                    "Checking last SCA - SCA index "
                                    "%d not done, so sleeping", i);

                                /* Found a previous SCA that is not completed,
                                   so go to sleep */
                                go_to_sleep = 1;
                                break;
                            }
                        }
                        if (!go_to_sleep)
                        {
                            IAS_LOG_DEBUG("SYNC_NEED_ADJACENT_SCAS: "
                                "assigning last SCA");

                            can_assign = 1;
                        }
                    }
                    else if (adjacent_ready_to_start == 0)
                    {
                        IAS_LOG_DEBUG("SYNC_NEED_ADJACENT_SCAS: "
                            "Can't start any SCAs.  Sleeping");

                        /* need to sleep if none are ready to start */
                        go_to_sleep = 1;
                    }
                    else if (first_algorithm_for_band 
                         || are_adjacent_scas_done(algo_count - 1, sca_count))
                    {
                        /* Flag that this one can be assigned */
                        can_assign = 1;

                        IAS_LOG_DEBUG("SYNC_NEED_ADJACENT_SCAS: "
                            "assigning sca %d", sca_count);

                        /* If there are more than one ready to start, wake up
                           any sleeping threads to let them work */
                        if (adjacent_ready_to_start > 1)
                        {
                            IAS_LOG_DEBUG("SYNC_NEED_ADJACENT_SCAS: "
                                "Waking up other threads to deal with other "
                                "adjacent SCAs that are ready");

                            /* This might wake up more threads than needed, but
                               any extras will go back to sleep again */
                            if (ias_thread_broadcast_condition(
                                    &work_assignment_mutex,
                                    &work_assignment_condition[condition_index])
                                    != SUCCESS)
                            {
                                IAS_LOG_ERROR("Broadcasting condition update");
                                IAS_THREAD_UNLOCK_MUTEX(&work_assignment_mutex);
                                set_work_assignment_error_condition();
                                return 0;
                            }
                        }
                    }

                    /* Go to sleep to wait for conditions to be right to
                       resume work */
                    if (go_to_sleep)
                    {
                        IAS_LOG_DEBUG("SYNC_NEED_ADJACENT_SCAS: sleeping");

                        /* Need to sleep, so wait on the condition variable
                           for this band's work order state */
                        if (wait_for_adjacent_scas(first_algorithm_for_band,
                                algo_count) != SUCCESS)
                        {
                            IAS_LOG_ERROR("Waiting on condition");
                            IAS_THREAD_UNLOCK_MUTEX( &work_assignment_mutex );
                            set_work_assignment_error_condition();
                            return 0;
                        }

                        IAS_LOG_DEBUG("SYNC_NEED_ADJACENT_SCAS: woke up");

                        /* After being woken, it is possible that an SCA
                           from this algorithm index can now be assigned.
                           So, change the algorithm increment to zero so that
                           the same algorithm is checked again.  Then break
                           out of the SCA loop to restart from the first
                           SCA to evaluate which SCA can be assigned. */
                        algo_increment = 0;
                        break;
                    }
                    if (!can_assign)
                    {
                        /* Can't assign this SCA, so continue the loop to see
                           if a later SCA can be worked on */
                        continue;
                    }

                    /* Getting here means the SCA can be worked on.  Falling
                       out of this if block will let the work be assigned for
                       this SCA. */
                }

                /* The previous algorithm must be complete before selecting
                   this SCA for work unless it is the first algorithm for the
                   band */
                if ( first_algorithm_for_band 
                     || work_state[algo_count-1].state[sca_count] == COMPLETE )
                {
                    work_state[algo_count].state[sca_count] = WORKING;
                    *band_index = work_state[algo_count].band;
                    *sca_index = sca_count;
                    *algorithm_index = work_state[algo_count].algorithm_index;
                    work_found = 1;
                    break;  /* from for sca_count loop */
                }
            }

            /* If this algorithm is one that needs to be synchronized on the
               last SCA and we reach the last SCA without assigning work,
               we need to wait for things to complete */
            else if ((work_state[algo_count].sync_type
                        == SYNC_NEED_ADJACENT_SCAS)
                     && ( work_state[algo_count].state[sca_count] == WORKING ))
            {
                if (sca_count == work_state[algo_count].nscas - 1)
                {
                    IAS_LOG_DEBUG("Putting thread to sleep until last "
                        "SCA done");

                    /* Need to sleep until the last SCA is finished */
                    if (wait_for_adjacent_scas(first_algorithm_for_band,
                            algo_count) != SUCCESS)
                    {
                        IAS_LOG_ERROR("Waiting on condition");
                        IAS_THREAD_UNLOCK_MUTEX( &work_assignment_mutex );
                        set_work_assignment_error_condition();
                        return 0;
                    }
                    IAS_LOG_DEBUG("Last thread must be done");
                }
            }
        }  /* END sca_count loop */

        if ( work_found )
            break;  /* from for algo_count loop */
    }  /* END for algo_count loop */

    if ( IAS_THREAD_UNLOCK_MUTEX( &work_assignment_mutex ))
    {
        IAS_LOG_ERROR("Unlocking work assignment mutex");
        set_work_assignment_error_condition();
        return 0;
    }

    /* If an error condition has occurred (which could happen if this thread
       ended up waiting), force no work to be found */
    if ( error_condition_encountered 
        || ias_threadpool_get_worker_return_status( local_pool ) != SUCCESS )
    {
        set_work_assignment_error_condition();
        work_found = 0;
    }

    if ( !work_found )
        return 0;
    return 1;
}  /* END assign_work */




/*------------------------------------------------------------------------
NAME:    initialize_work_assignment

PURPOSE: Create the mutex and initialize the condition variable and state
         structure 

RETURNS: SUCCESS or ERROR
------------------------------------------------------------------------*/
int initialize_work_assignment
(
    int  total_algorithms,         /* I: Total number of algorithms */
    int  nbands,                   /* I: Number of bands to process */
    int  list_of_band_indices[],   /* I: List of band indices to process */
    struct ias_threadpool *pool    /* I: Threadpool */
)
{
    int status;     /* Routine return status */
    int nscas;      /* Number of SCAs for a band */
    int algo_index; /* Loop index for the algorithms */
    int band_index; /* Loop index for bands */
    int sca_index;  /* Loop index for SCAs */
    int work_array_index; /* Index into the work state structure */
    RPS_BAND_ATTRIBUTES *band_attributes;
    const IAS_SATELLITE_ATTRIBUTES *sat_atts;

    /* Check the log level, if it's IAS_LOG_LEVEL_DEBUG then call the
       routine to set up the debug file to hold work state information */
    if ( ias_log_message_level == IAS_LOG_LEVEL_DEBUG )
    {
        status = open_debug_output_file();
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Opening debugging file");
            return ERROR;
        }
    }

    local_pool = pool;

    sat_atts = ias_sat_attr_get_attributes();
    if ( !sat_atts )
    {
        IAS_LOG_ERROR("Getting satellite attributes");
        return ERROR;
    }

    /* Allocate some memory */
    band_lut = malloc( sizeof(int) * sat_atts->total_bands );
    band_reverselut = malloc( sizeof(int) * ( sat_atts->total_bands ));
    if ( !band_lut || !band_reverselut )
    {
        free( band_lut );
        free( band_reverselut );
        IAS_LOG_ERROR("Allocating memory for assign work");
        return ERROR;
    }

    /* Initialize the variables and state structure */
    number_algorithms = total_algorithms;
    number_state_struc = number_algorithms * nbands;

    /* Set up band look up tables */
    for ( band_index = 0; band_index < sat_atts->total_bands; band_index++ )
        band_reverselut[band_index] = -1;
    for ( band_index = 0; band_index < nbands; band_index++ )
    {
        band_lut[band_index] = list_of_band_indices[band_index];
        band_reverselut[band_lut[band_index]] = band_index;
    }

    /* Allocate the work state array */
    work_state = malloc( sizeof(WORK_STATE_STRUCT) * number_state_struc );
    if ( !work_state )
    {
        IAS_LOG_ERROR("Allocating memory for work state");
        free( band_lut );
        free( band_reverselut );
        return ERROR;
    }

    /* Initialize the work state struct */
    for ( band_index = 0; band_index < nbands; band_index++ )
    {
        band_attributes = get_band_attributes( band_lut[band_index] );
        nscas = band_attributes->number_scas;

        for ( algo_index = 0; algo_index < number_algorithms; algo_index++ )
        {
            work_array_index = ( band_index * number_algorithms ) + algo_index;
            work_state[work_array_index].algorithm_index = algo_index;
            work_state[work_array_index].band = band_lut[band_index];
            work_state[work_array_index].nscas = nscas;
            work_state[work_array_index].sync_type = SYNC_NONE;
            for ( sca_index = 0; sca_index < nscas; sca_index++ )
            {
                if ( band_attributes->one_sca_present )
                {
                    /* For single SCA case mark all SCAs that aren't to be
                       processed as complete.  The assign_work routine will
                       see them as complete and not assign them to the
                       threads.  Only mark the SCA to be processed as
                       NOT_ASSIGNED. */
                    if ( band_attributes->sca_number == sca_index+1 )
                    {
                        work_state[work_array_index].state[sca_index] =
                            NOT_ASSIGNED;
                    }
                    else
                    {
                        work_state[work_array_index].state[sca_index] =
                            COMPLETE;
                    }
                }
                else
                {
                    work_state[work_array_index].state[sca_index] =
                        NOT_ASSIGNED;
                }
            }  /* END for sca_index loop */
        }  /* END for algo_index loop */
    }  /* END for band_index loop */

    /* Allocate memory for the condition variables and counts */
    number_allocated = nbands;
    work_assignment_condition = malloc( sizeof(*work_assignment_condition) *
        number_allocated );
    if ( !work_assignment_condition )
    {
        IAS_LOG_ERROR("Creating the work assignment condition variables");
        free( band_lut );
        free( band_reverselut );
        return ERROR;
    }

    /* Create the mutex and condition variable */
    status = IAS_THREAD_CREATE_MUTEX( &work_assignment_mutex );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Creating the work assignment mutex");
        free( band_lut );
        free( band_reverselut );
        free( work_assignment_condition );
        return ERROR;
    }
    for ( band_index = 0; band_index < number_allocated; band_index++ )
    {
        status = IAS_THREAD_CREATE_COND(
            &work_assignment_condition[band_index] );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Creating the work assignment condition variable "
                "number %d", band_index + 1 );
            free( band_lut );
            free( band_reverselut );
            free( work_assignment_condition );
            return ERROR;
        }
    }

    return SUCCESS;
}  /* END initialize_work_assignment */



/*------------------------------------------------------------------------
NAME:    cleanup_work_assignment

PURPOSE: Destroy the mutex variables and free local memory

RETURNS: none
------------------------------------------------------------------------*/
void cleanup_work_assignment( void )
{
    int index;

    free( band_lut );
    band_lut = NULL;
    free( band_reverselut );
    band_reverselut = NULL;
    free( work_state );
    work_state = NULL;
    IAS_THREAD_DESTROY_MUTEX( &work_assignment_mutex );

    for ( index = 0; index < number_allocated; index++ )
        IAS_THREAD_DESTROY_COND( &work_assignment_condition[index] );
    free( work_assignment_condition );
    work_assignment_condition = NULL;

    /* Check the log level, if it's IAS_LOG_LEVEL_DEBUG then call the
       routine to close the output debug file */
    if ( ias_log_message_level == IAS_LOG_LEVEL_DEBUG )
        close_debug_output_file();
}


/*--------------------------- Test Function ------------------------------
   The routines in this section are turned on by setting IAS_LOG_LEVEL to
   DEBUG.  The output is an ASCII text file that contains a dump of the
   condition counts and work state each time the synchronize_query routine
   is called.  Note the synchronize_query routine is only called for
   serialized algorithms.  The output file name is assign_work_debug.txt and
   is delimited to allow easy parsing.  First line of a block is:

-------------------- 9 2 11

   where 9 2 11 are band, SCA and algorithm indices.  The end of the block
   is:

+++

   Using the command:
       sed -n '/--* 10 0/,/+++/p' assign_work_debug.txt
   will output the block for band_index 10 sca_index 0.  Partial output:

-------------------- 10 0 11
Condition counts:  0  0  0  0  0  0  0  3  1 
aindex:   0 band:  0  algo:  0  nscas: 14 state: C C C C C C C C C C C C C C 
aindex:   1 band:  0  algo:  1  nscas: 14 state: C C C C C C C C C C C C C C 
aindex:   2 band:  0  algo:  2  nscas: 14 state: C C C C C C C C C C C C C C 
aindex:   3 band:  0  algo:  3  nscas: 14 state: C C C C C C C C C C C C C C 
    ...
aindex: 114 band:  9  algo:  9  nscas:  3 state: C C C 
aindex: 115 band:  9  algo: 10  nscas:  3 state: C C C 
aindex: 116 band:  9  algo: 11  nscas:  3 state: W W W 
aindex: 117 band:  9  algo: 12  nscas:  3 state: . . . 
aindex: 118 band:  9  algo: 13  nscas:  3 state: . . . 
    ...
aindex: 129 band: 10  algo:  9  nscas:  3 state: C C C 
aindex: 130 band: 10  algo: 10  nscas:  3 state: C C C 
aindex: 131 band: 10  algo: 11  nscas:  3 state: W . . 
aindex: 132 band: 10  algo: 12  nscas:  3 state: . . . 
aindex: 133 band: 10  algo: 13  nscas:  3 state: . . . 
aindex: 134 band: 10  algo: 14  nscas:  3 state: . . . 
+++

   The above output shows that band index 9 condition is met since 3 ==
   number of SCAs for the band and is therefore running the algorithm.  It
   also shows that band index 10 SCA index 0 called the algorithm and the
   condition count for that band was incremented.
------------------------------------------------------------------------*/
/*--------------------------- Test Function ------------------------------
NAME:    open_debug_output_file

PURPOSE: Opens the debugging output file

RETURNS: SUCCESS or ERROR
------------------------------------------------------------------------*/
static int open_debug_output_file(void)
{

    debug_fd = fopen( debug_filename, "w" );
    if ( !debug_fd )
    {
        IAS_LOG_ERROR("Opening debug output file: %s", debug_filename );
        return ERROR;
    }
    return SUCCESS;
}

/*--------------------------- Test Function ------------------------------
NAME:    close_debug_output_file

PURPOSE: Closes the debugging output file

RETURNS: none
------------------------------------------------------------------------*/
static void close_debug_output_file(void)
{
    int status = fclose( debug_fd );
    if ( status )
    {
        IAS_LOG_ERROR("Closing debug output file: %s", debug_filename );
    }
}

/*--------------------------- Test Function ------------------------------
NAME:    print_debug_information

PURPOSE: Prints debug information to output file

RETURNS: SUCCESS or ERROR
------------------------------------------------------------------------*/
#if 0
/* Disable this routine for now since it isn't being used */
static int print_debug_information
(
    int band_index,     /* I: */
    int sca_index,      /* I: */
    int algo_index      /* I: */
)
{
    int status;
    int i;

    /* Identify the beginning of a block of output */
    status = fprintf( debug_fd, "-------------------- %d %d %d\n",
        band_index, sca_index, algo_index);
    if ( status < 0 )
        return ERROR;

    /* Print the work state structure */
    for ( i = 0; i < number_state_struc; i++ )
    {
        int j;
        status = fprintf( debug_fd, "aindex: %3d band: %2d  algo: %2d  nscas: "
            "%2d  sync type: %d  state: ", i, work_state[i].band,
            work_state[i].algorithm_index, work_state[i].nscas,
            work_state[i].sync_type );
        if ( status < 0 )
            return ERROR;
        for ( j = 0; j < work_state[i].nscas; j++ )
        {
            char s;
            switch ( work_state[i].state[j] )
            {
                case NOT_ASSIGNED :
                    s = '.';
                    break;
                case WORKING :
                    s = 'W';
                    break;
                case COMPLETE :
                    s = 'C';
                    break;
                default:
                    s = '?';
            }
            status = fprintf( debug_fd, "%c ", s );
            if ( status < 0 )
                return ERROR;
        }
        status = fprintf( debug_fd, "\n" );
        if ( status < 0 )
            return ERROR;
    }

    /* Identify the end of a block of output */
    status = fprintf( debug_fd, "+++\n");
    if ( status < 0 )
        return ERROR;

    return SUCCESS;
}  /* END print_debug_information */
#endif
