/*----------------------------------------------------------------------------
NAME:    blob_interface

PURPOSE: Collection of routines to create information in, or access
         information from, the blob
-----------------------------------------------------------------------------*/
#include <stdlib.h>
#include <string.h>

#include "ias_cpf.h"
#include "ias_logging.h"
#include "ias_miscellaneous.h"
#include "ias_model_io.h"
#include "rps_local.h"

#define LINKED_LIST_DELETE( list ) \
    ias_linked_list_delete( (list), \
    offsetof(DB_SCENE_SUMMARY_STATISTICS, node ));

/*----------------------------------------------------------------------------
NAME:    allocate_blob

PURPOSE: Allocates memory for the blob.  This is intended to be called from
         main

RETURNS: Pointer to the blob or NULL on error
-----------------------------------------------------------------------------*/
static BLOB *blob = NULL;
BLOB *allocate_blob( void )
{
    blob = calloc( 1, sizeof( BLOB ));
    return blob;
}


/*----------------------------------------------------------------------------
NAME:    free_the_blob

PURPOSE: Frees all the allocated resources in the blob then the blob itself

RETURNS: none
-----------------------------------------------------------------------------*/
void free_the_blob( void )
{
    int i;
    int band_index;
    int sca_index;
    RPS_BAND_ATTRIBUTES *atts;
    if ( !blob )
        return;

    if ( blob->pixel_mask )
    {
        for ( band_index = 0; band_index < blob->total_bands; band_index++ )
        {
            if ( blob->pixel_mask[band_index] == NULL )
                continue;

            atts = get_band_attributes( band_index );
            for ( sca_index = 0; sca_index < atts->number_scas; sca_index++ )
            {
                if ( blob->pixel_mask[band_index][sca_index] != NULL )
                    ias_pm_destroy( blob->pixel_mask[band_index][sca_index] );
            }

            free( blob->pixel_mask[band_index] );
        }
        free( blob->pixel_mask );
    }
    free( blob->algorithm_pointer_list );

    /* Free the parameter strings that were allocated in the read parameter
       routine but only if its been allocated */
    if ( blob->algorithm_wo_parameter_list )
    {
        for ( i = 0; i < blob->number_of_wo_parm_strings_allocated; i++ )
            free( blob->algorithm_wo_parameter_list[i] );
    }
    free( blob->algorithm_wo_parameter_list );

    /* These is simply an array of pointers that contain copies of the
       algorithm_wo_parameter_list pointers just freed above so just free
       the memory */
    free( blob->algorithm_string_list );
    free( blob->dump_algorithm_string_list );

    free( blob->pre_proc_pointer_list );
    free( blob->post_proc_pointer_list );
    free( blob->dump_algorithm_mapping );
    free( blob->list_of_bands );
    free( blob->list_of_band_indices );
    if ( blob->band_information )
    {
        for ( i = 0; i <= blob->total_bands; i++ )
        {
            if ( blob->band_information[i] )
                free( blob->band_information[i] );
        }
    }
    if ( blob->l1r_band_metadata )
    {
        for ( i = 0; i <= blob->total_bands; i++ )
        {
            if ( blob->l1r_band_metadata[i] )
                free( blob->l1r_band_metadata[i] );
        }
    }
    free( blob->l1r_band_metadata );
    ias_misc_free_report_header( blob->report_information.header );
    free( blob->report_information.directory );
    free( blob->report_information.rlut_version );
    free( blob->band_information );
    if ( blob->cpf )
        ias_cpf_free( blob->cpf );
    if ( blob->bpf_oli )
        ias_bpf_free( blob->bpf_oli );
    if ( blob->bpf_tirs )
        ias_bpf_free( blob->bpf_tirs );
    if ( blob->rlut )
    {
        if ( ias_rlut_close_file( blob->rlut ) != SUCCESS )
            IAS_LOG_ERROR("Closing RLUT file");
    }

    if (blob->scene_statistics_list)
    {
        LINKED_LIST_DELETE(blob->scene_statistics_list);
        free(blob->scene_statistics_list);
    }

    free( blob );
    blob = NULL;
}  /* END free_the_blob */



/*------------------------------------------------------------------------
NAME:    create_blob_pixel_mask

PURPOSE: creates a pixel mask in the blob based on the number of bands

RETURNS: SUCCESS or ERROR
------------------------------------------------------------------------*/
int create_blob_pixel_mask( void )
{
    int band_index;
    int sca_index;
    RPS_BAND_ATTRIBUTES *atts;

    /* Create the pixel mask */
    for ( band_index = 0; band_index < blob->total_bands; band_index++ )
    {
        atts = get_band_attributes( band_index );
        blob->pixel_mask[band_index] = calloc( atts->number_scas,
            sizeof( IAS_PIXEL_MASK * ));
        if ( !blob->pixel_mask[band_index] )
        {
            IAS_LOG_ERROR("Allocating SCAs in the pixel mask");
            return ERROR;
        }
        
        for ( sca_index = 0; sca_index < atts->number_scas; sca_index++ )
        {
            blob->pixel_mask[band_index][sca_index] = ias_pm_create(
                atts->band_number, sca_index + 1, atts->number_detectors,
                atts->number_lines );
            if (blob->pixel_mask[band_index][sca_index] == NULL)
            {
                IAS_LOG_ERROR("Creating empty pixel mask for band index %d "
                    "SCA %d", band_index, (sca_index + 1));
                return ERROR;
            }
        }
    }

    return SUCCESS;
}  /* END create_blob_pixel_mask */


/*----------------------------------------------------------------------------
NAME:    finish_blob_initialization

PURPOSE: Complete the initialization of the blob.  Includes: allocating
         data structures based on bands present; setting up the report
         structure (if needed); opening CPF

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int finish_blob_initialization( void )
{
    int i;
    const IAS_BAND_ATTRIBUTES *band_atts;

    blob->num_normal_bands = ias_sat_attr_get_normal_band_count();
    if (blob->num_normal_bands == ERROR)
    {
        IAS_LOG_ERROR("Getting normal band count");
        return ERROR;
    }

    blob->max_number_scas = 0;
    for ( i = 1; i <= blob->total_bands; i++ )
    {
        band_atts = ias_sat_attr_get_band_attributes( i );
        if ( band_atts->scas > blob->max_number_scas )
            blob->max_number_scas = band_atts->scas;
    }

    /* Allocate memory */
    blob->list_of_bands = calloc( blob->total_bands, sizeof(int) );
    if ( !blob->list_of_bands )
    {
        IAS_LOG_ERROR("Allocating list of bands");
        free_the_blob();
        return ERROR;
    }
    blob->list_of_band_indices = calloc( blob->total_bands, sizeof(int) );
    if ( !blob->list_of_band_indices )
    {
        IAS_LOG_ERROR("Allocating list of bands indices");
        free_the_blob();
        return ERROR;
    }
    blob->band_information = calloc(( blob->total_bands + 1),
        sizeof( RPS_BAND_ATTRIBUTES * ));
    if ( !blob->band_information )
    {
        IAS_LOG_ERROR("Allocating band_information structure array");
        free_the_blob();
        return ERROR;
    }
    for ( i = 0; i <= blob->total_bands; i++ )
        blob->band_information[i] = calloc( 1, sizeof(RPS_BAND_ATTRIBUTES));
    blob->l1r_band_metadata = calloc(( blob->total_bands + 1),
        sizeof( IAS_L1R_BAND_METADATA * ));
    if ( !blob->l1r_band_metadata )
    {
        IAS_LOG_ERROR("Allocating L1R metadata structure");
        free_the_blob();
        return ERROR;
    }
    for ( i = 0; i <=  blob->total_bands; i++ )
        blob->l1r_band_metadata[i] = calloc( 1, sizeof(IAS_L1R_BAND_METADATA));

    blob->pixel_mask = calloc( blob->total_bands, sizeof(IAS_PIXEL_MASK **));
    if ( !blob->pixel_mask )
    {
        IAS_LOG_ERROR("Allocating pixel mask");
        free_the_blob();
        return ERROR;
    }

    /* Set up the report structure if needed */
    if ( blob->generate_report )
    {
        IAS_REPORT *report_info = &blob->report_information;

        report_info->directory = malloc(sizeof(blob->report_directory));
        if ( !report_info->directory )
        {
            IAS_LOG_ERROR("Allocating the report directory");
            free_the_blob();
            return ERROR;
        }
        else
            strcpy( report_info->directory, blob->report_directory );

        /* Initialize the standard report header */
        report_info->header = ias_misc_initialize_report_header(
            blob->l0r_file_name,
            blob->work_order_id,
            "" );
        if ( !report_info->header )
        {
            IAS_LOG_ERROR( "Initializing report header" );
            free_the_blob();
            return ERROR;
        }
    }

    /* Read the CPF */
    blob->cpf = ias_cpf_read( blob->cpf_filename );
    if ( !blob->cpf )
    {
        IAS_LOG_ERROR("Reading CPF: %s", blob->cpf_filename );
        free_the_blob();
        return ERROR;
    }

    /* Allocate memory for linked list head */
    blob->scene_statistics_list = malloc(sizeof(IAS_LINKED_LIST_NODE));
    if (!blob->scene_statistics_list)
    {
        IAS_LOG_ERROR("Initializing scene statistics list");
        return ERROR;
    }
    
    /* Initialize the linked list */ 
    ias_linked_list_initialize_node(blob->scene_statistics_list);

    return SUCCESS;
}  /* END finish_blob_initialization */

/*----------------------------------------------------------------------------
NAME:    read_oli_tirs_bpf

PURPOSE: Opening OLI and/or TIRS BPF if OLI or TIRS band is present and
         OLI or TIRS BPF filename is avaliable

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int read_oli_tirs_bpf
(  
    BLOB *blob          /* I: Contains OLI and TIRS band present info and
                              OLI and TIRS BPF filenames */
)
{
    /* Read the OLI BPF if OLI band is present for processing and OLI BPF  
       filename is available */
    if ((blob->process_oli_bands == 1) 
       && (strcmp(blob->bpf_filename_oli, "") != 0))
    {
        blob->bpf_oli = ias_bpf_read( blob->bpf_filename_oli );
        if ( !blob->bpf_oli )
        {
            IAS_LOG_ERROR("Reading BPF: %s", blob->bpf_filename_oli );
            free_the_blob();
            return ERROR;
        }
    }

    /* Read the TIRS BPF if TIRS band is present for processing and TIRS BPF  
       filename is available */
    if ((blob->process_tirs_bands == 1) 
         && (strcmp(blob->bpf_filename_tirs, "") != 0))
    {
        blob->bpf_tirs = ias_bpf_read( blob->bpf_filename_tirs );
        if ( !blob->bpf_tirs )
        {
            IAS_LOG_ERROR("Reading BPF: %s", blob->bpf_filename_tirs );
            free_the_blob();
            return ERROR;
        }
    }

    return SUCCESS;
}

/*------------------------ BLOB Getter Routines ------------------------*/
/* Return pointer to the CPF */
IAS_CPF *get_blob_cpf( void )
{
    return blob->cpf;
}

/* Return pointer to the OLI BPF */
IAS_BPF *get_blob_bpf_oli( void )
{
    return blob->bpf_oli;
}

/* Return pointer to the TIRS BPF */
IAS_BPF *get_blob_bpf_tirs( void )
{
    return blob->bpf_tirs;
}

/* Get a pointer to the RLUT file name */
char *get_blob_rlut_filename(void)
{
    return blob->rlut_filename;
}

/* Return the OLI scene start time */
double get_blob_oli_start_time( void )
{
    return blob->oli_start_time;
}

/* Return the OLI scene stop time */
double get_blob_oli_stop_time( void )
{
    return blob->oli_stop_time;
}

/* Return the TIRS scene start time */
double get_blob_tirs_start_time( void )
{
    return blob->tirs_start_time;
}

/* Return the TIRS scene stop time */
double get_blob_tirs_stop_time( void )
{
    return blob->tirs_stop_time;
}

/* Returns a structure containing information about a band */
RPS_BAND_ATTRIBUTES *get_band_attributes
(
    int band_index            /* I: */
)
{
    return blob->band_information[band_index];
}

/* Return the maximum number of SCAs */
int get_blob_max_nscas( void )
{
    return blob->max_number_scas;
}

/*  Returns the pointer to the database connection */
struct ias_db_connection *get_blob_db_connection(void)
{
    return blob->database_information.db;
}

/* Returns the radiometric work order common ID */
const char *get_blob_rad_common_id(void)
{
    return blob->database_information.rad_common_id;
}

/* Returns the work order string */
char *get_blob_work_order_id(void)
{
    return blob->work_order_id;
}

/* Returns stim lamp post warmup number of frames of lamp data */
int get_blob_post_warmup_nframes( void )
{
    return blob->post_warmup_nframes;
}

/* Returns stim lamp post warmup number of lines */
double get_blob_lamp_warmup_time_period( void )
{
    return blob->lamp_warmup_time_period;
}

/* Returns the number of bands to be processed */
int get_blob_number_of_bands(void)
{
    return blob->number_of_bands;
}

/* Returns the list of bands to be processed */
int *get_blob_list_of_bands_to_process(void)
{
    return blob->list_of_bands;
}

/* Returns the list of band indices to be processed */
int *get_blob_list_of_band_indices_to_process(void)
{
    return blob->list_of_band_indices;
}

/* Returns a pointer to the RLUT */
IAS_RLUT_IO *get_blob_rlut( void )
{
    return blob->rlut;
}

/* Returns the value of the flag */
int get_band_fields_initialized( void )
{
    return blob->band_fields_initialized;
}

/* Get the algorithm index for a particular algorithm based on the function
   pointer to it */
int get_algorithm_index
(
    void *func_ptr
)
{
    int i;

    for (i = 0; i < blob->number_algorithms; i++)
    {
        if (blob->algorithm_pointer_list[i] == func_ptr)
            return i;
    }

    return -1;
}

int get_background_response_selection(void)
{
    return blob->background_response_selection;
}

/* Set the OLI focal plane temperature and the flag to indicate it's been
   set */
void set_oli_focal_plane_temperature
(
    float temperature /* I: OLI mean focal plane temperature */
)
{
    blob->oli_mean_focal_plane_temperature = temperature;
    blob->oli_mean_focal_plane_temperature_valid = 1;
}

/* Set the TIRS focal plane temperature and the flag to indicate it's been
   set */
void set_tirs_focal_plane_temperature
(
    float temperature /* I: TIRS mean focal plane temperature */
)
{
    blob->tirs_mean_focal_plane_temperature = temperature;
    blob->tirs_mean_focal_plane_temperature_valid = 1;
}

/* Returns SUCCESS and the focal plane temperature if the temperature has
   been previously set or ERROR if it hasn't been set */
int get_oli_focal_plane_temperature
(
    float *temperature /* O: OLI mean focal plane temperature */
)
{
    if ( ! blob->oli_mean_focal_plane_temperature_valid )
        return ERROR;

    *temperature = blob->oli_mean_focal_plane_temperature;
    return SUCCESS;
}

/* Returns SUCCESS and the focal plane temperature if the temperature has
   been previously set or ERROR if it hasn't been set */
int get_tirs_focal_plane_temperature
(
    float *temperature /* O: TIRS mean focal plane temperature */
)
{
    if ( ! blob->tirs_mean_focal_plane_temperature_valid )
        return ERROR;

    *temperature = blob->tirs_mean_focal_plane_temperature;
    return SUCCESS;
}

/* Returns pointer to head of scene summary list */
IAS_LINKED_LIST_NODE *get_scene_summary_list(void)
{
    return blob->scene_statistics_list;
}
