/*----------------------------------------------------------------------------
NAME:    create_algorithm_pointer_list

PURPOSE: Create a list of function pointers based on a list of function
         names.  Also create lists for pre- and post-processing algorithms.
         There are two algorithm modifiers that can be added to the work
         order algorithm name: '_SKIP' and '_DUMP'.  This routine will
         remove the algorithms marked as '_SKIP'.  The '_DUMP' modifier is
         used to ask for an image dump after the algorithm.  The '_DUMP'
         modifier is removed and the special dump algorithm inserted in the
         list.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
#include "stdlib.h"
#include "string.h"

#include "ias_logging.h"
#include "rps_local.h"

static void free_lists( BLOB * );

#define PREREQUISITE_ALGORITHM  "convert_to_float"
/* This array contains the entire catalog of available algorithms that
   the RPS is capable of running.  Although not needed here, the dump
   imagery algorithm is included for completeness.  The last field in the
   structure is 'must_run_after_prerequisite' and is used to ensure that
   these algorithms follow the required convert_to_float algorithm. */
static ALG_TYPE available_algorithms[] =
{
    /* Add optional algorithms below here as they become available */
    { "dropped_frame_characterization",
       api_dropped_frame_characterization,
       dropped_frame_characterization_preprocess, NULL, 0, 0 },
    { "saturated_pixel_characterization",
       api_saturated_pixel_characterization, 
       saturated_pixel_characterization_preprocess, NULL, 0, 0 },
    { "oli_bias_removal",
       api_oli_bias_removal,
       oli_bias_removal_preprocess, NULL, 0, 1 },
    { "temperature_sensitivity_correction", 
       api_temperature_sensitivity_correction,
       temperature_sensitivity_correction_preprocess, NULL, 0, 1 },
    { "gain_application",
       api_gain_application,
       gain_application_preprocess, NULL, 0, 1 },
    { "tirs_histogram_statistics_characterization",
       api_tirs_histogram_statistics_characterization,
       histogram_statistics_characterization_preprocess,
       histogram_statistics_characterization_postprocess, 0, 1 },
    { "oli_histogram_statistics_characterization",
       api_oli_histogram_statistics_characterization,
       histogram_statistics_characterization_preprocess,
       histogram_statistics_characterization_postprocess, 0, 1 },
    { "tirs_3q_response_linearization",
       api_tirs_3q_response_linearization,
       response_linearization_3q_preprocess, NULL, 0, 1 },
    { "oli_3q_response_linearization",
       api_oli_3q_response_linearization,
       response_linearization_3q_preprocess, NULL, 0, 1 },
    { "tirs_lookup_response_linearization",
       api_tirs_lookup_response_linearization,
       lookup_response_linearization_preprocess, NULL, 0, 1 },
    { "oli_lookup_response_linearization",
       api_oli_lookup_response_linearization,
       lookup_response_linearization_preprocess, NULL, 0, 1 },
    { "tirs_bias_removal",
       api_tirs_bias_removal,
       tirs_bias_removal_preprocess, NULL, 0, 1 },
    { "oli_bias_determination",
       api_oli_bias_determination,
       oli_bias_determination_preprocess, NULL, 0, 1 },
    { "char_lamp_detector_response",
       api_char_lamp_detector_response,
       char_lamp_detector_response_preprocess, 
       char_lamp_detector_response_postprocess, 0, 1 },
    { "tirs_dark_response_determination",
       api_tirs_dark_response_determination,
       tirs_dark_response_determination_preprocess,
       tirs_dark_response_determination_postprocess, 0, 1},
    { "reflectance_conversion",
       api_reflectance_conversion,
       reflectance_conversion_preprocess, NULL, 0, 1},
    { "inoperable_detector_identification",
       api_inoperable_detector_identification,
       inoperable_detector_identification_preprocess, NULL, 0, 0},
    { "impulse_noise_characterization",
       api_impulse_noise_characterization,
       impulse_noise_characterization_preprocess, NULL, 0, 0 },
    { "sca_discontinuity_correction",
       api_sca_discontinuity_correction,
       sca_discontinuity_correction_preprocess, 
       sca_discontinuity_correction_postprocess, 0, 1 },
    { "saturated_pixel_replacement",
       api_saturated_pixel_replacement,
       saturated_pixel_replacement_preprocess, NULL, 0, 1 },
    { "residual_striping_correction",
       api_residual_striping_correction,
       residual_striping_correction_preprocess, NULL, 0, 1 },
    { "inoperable_detectors_fill",
      api_inoperable_detectors_fill,
      inoperable_detectors_fill_preprocess, NULL, 0, 1 },
    { "striping_characterization",
      api_striping_characterization, striping_characterization_preprocess,
      striping_characterization_postprocess, 0, 1 },
    { "tirs_secondary_linearization",
      api_tirs_secondary_linearization,
      tirs_secondary_linearization_preprocess, NULL, 0, 1 },
    { "tirs_gain_offset_removal",
      api_tirs_gain_offset_removal,
      tirs_gain_offset_removal_preprocess, NULL, 0, 1 },
    { "tirs_stray_light_correction",
      api_tirs_stray_light_correction,
      tirs_stray_light_correction_preprocess, NULL, 0, 1 },

    /* Add required algorithms below here as they become available */
    { PREREQUISITE_ALGORITHM, api_convert_to_float,
       convert_to_float_preprocess, NULL, 1, 0 },

    /* NOTE 1: Although there are pre- and post-process algorithms for
               dump_imagery, we only want to call them once.  That is why
               the pointers are NULL.  This is a special case and is handled
               in the code.  
       
       NOTE 2: This must be last so that callers of get_available_algorithms
               won't see it as an externally available algorithm.  The dump
               algorithm is accessed externally by addition of the _DUMP
               attribute / modifier. */
    { "dump_imagery", api_dump_imagery, NULL, NULL, 0, 0 }
};


/*----------------------------------------------------------------------------
NAME:    get_available_algorithms

PURPOSE: Return a list and count of the available algorithms.  This list
         does not include the dump routine in the count.

RETURNS: pointer to the list of algorithms
-----------------------------------------------------------------------------*/
const ALG_TYPE *get_available_algorithms
(
    int *number_algorithms  /* O: Number of externally available algorithms */
)
{
    /* Return a count of the algorithm list less the special case
       dump_imagery which is an internal algorithm invoked by the
       attribute/modifier "_DUMP" */
    *number_algorithms = sizeof(available_algorithms) / sizeof(ALG_TYPE) - 1;
    return &available_algorithms[0];
}


/*----------------------------------------------------------------------------
NAME:    create_algorithm_pointer_list

PURPOSE: Create a list of function pointers based on a list of function
         names.  Also create lists for pre- and post-processing algorithms.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int create_algorithm_pointer_list
(
    BLOB *blob     /* I: List of algorithms as strings
                      I: Number of algorithms
                      O: List of algorithms as pointers to API
                      O: Updated number of algorithms
                      O: List of pre-processing algorithms
                      O: List of post-processing algorithms
                      O: List of algorithms with dump modifier
                      O: Number of algorithms with dump modifier */
)
{
    int i;                /* Plain old loop counter */
    int number_allocated; /* The correct number of algorithms to allocate */
    int dump_count;       /* Temporary counter */
    char *p_modifier;     /* Pointer to a modifier string SKIP or DUMP */
    int dump_mod_present; /* Flag indicating presense of dump modifier */
    int algo_index;       /* Index into available algorithm list */
    int pre_algo_index;   /* Index into the pre-processing algorithm list */
    int post_algo_index;  /* Index into the post-processing algorithm list */
    int ap_index;         /* Index into the blob algorithm lists */
    int number_available; /* Total number of available algorithms */
    int num_of_dump_algos;/* Running count of dump algorithms added to list */
    int original_number_algorithms; /* Original number of algorithms that
                                       was passed in.  Need to keep track
                                       since the blob value may be modified. */
    int missing_required;  /* Flag to track whether there are any missing
                              but required algorithms */
    int prerequisite_missing; /* Flag to indicate that convert_to_float
                                 algorithm hasn't been encountered in the
                                 input list */
    int string_list_index; /* Index to track access to the list of algorithm
                              strings.  This is needed to handle an empty
                              string when the DUMP modifier is encountered.
                              */

    /* Set the number of available algorithms based on what has been
       initialized above. */
    number_available = sizeof(available_algorithms) / sizeof(ALG_TYPE);

    /* Need to keep track of the original number of algorithms, it will be
       used to step through the algorithm list since there may be additions
       due to DUMP and deletions due to SKIP. */
    original_number_algorithms = blob->number_algorithms;
    /* Determine the number of algorithms to allocate this is:
       number allocated = number of input algos - skip algos + dump algos
       This will also be used as a crosscheck before returning */
    number_allocated = blob->number_algorithms;
    for ( i = 0, dump_count = 0; i < blob->number_algorithms; i++ )
    {
        p_modifier = strstr( blob->algorithm_wo_parameter_list[i], "_SKIP" );
        if ( p_modifier )
        {
            number_allocated--;
        }
        /* If the modifier wasn't SKIP check for DUMP.  Need to skip this
           if the SKIP modifier was found and DUMP was also added. */
        else
        {
            p_modifier = strstr( blob->algorithm_wo_parameter_list[i],
                "_DUMP" );
            if ( p_modifier )
            {
                number_allocated++;
                dump_count++;
            }
        }
    }
    /* A negative number is probably impossible but check just in case.
       However, the zero case is possible, i.e. all algorithms marked as
       SKIP. */
    if ( number_allocated <= 0 )
    {
        IAS_LOG_ERROR( "Number of algorithms is %d", number_allocated );
        return ERROR;
    }
    /* Return the correct number of algorithms */
    blob->number_algorithms = number_allocated;
    blob->number_dump_algorithms = dump_count;

    /* Allocate the pointer lists.  Set the pointers to NULL so that the
       free routine will work without problems. */
    blob->algorithm_pointer_list = NULL;
    blob->pre_proc_pointer_list = NULL;
    blob->post_proc_pointer_list = NULL;
    blob->dump_algorithm_string_list = NULL;
    blob->dump_algorithm_mapping = NULL;
    blob->algorithm_string_list = NULL;
    blob->algorithm_pointer_list = calloc( number_allocated, sizeof(int *));
    if ( ! blob->algorithm_pointer_list )
    {
        IAS_LOG_ERROR( "Allocating algorithm_pointer_list" );
        return ERROR;
    }
    /* Add one to this to allow the dump pre-process algorithm */
    blob->pre_proc_pointer_list = calloc( number_allocated + 1, sizeof(int *));
    if ( ! blob->pre_proc_pointer_list )
    {
        IAS_LOG_ERROR( "Allocating pre_proc_pointer_list" );
        free_lists( blob );
        return ERROR;
    }
    /* Add one to this to allow the dump post-process algorithm */
    blob->post_proc_pointer_list = calloc( number_allocated + 1, sizeof(int *));
    if ( ! blob->post_proc_pointer_list )
    {
        IAS_LOG_ERROR( "Allocating post_proc_pointer_list" );
        free_lists( blob );
        return ERROR;
    }
    blob->dump_algorithm_string_list = calloc( dump_count, sizeof(char *));
    if ( ! blob->dump_algorithm_string_list )
    {
        IAS_LOG_ERROR( "Allocating dump_algorithm_string_list" );
        free_lists( blob );
        return ERROR;
    }
    blob->dump_algorithm_mapping = calloc( dump_count, sizeof(int));
    if ( ! blob->dump_algorithm_mapping )
    {
        IAS_LOG_ERROR( "Allocating dump_algorithm_mapping" );
        free_lists( blob );
        return ERROR;
    }
    blob->algorithm_string_list = calloc( number_allocated, sizeof(char *));
    if ( ! blob->algorithm_string_list )
    {
        IAS_LOG_ERROR( "Allocating algorithm_string_list" );
        free_lists( blob );
        return ERROR;
    }

    /* Set the flag indicating the prerequisite hasn't been found */
    prerequisite_missing = 1;

    /* Set the index to manage the algorithm string list */
    string_list_index = 0;

    /* Process through the input algorithm list */
    ap_index = pre_algo_index = post_algo_index = num_of_dump_algos = 0;
    for ( i = 0; i < original_number_algorithms; i++ )
    {
        /* First thing to check is for the SKIP modifier.  If it's present
           continue to next loop iteration.  Basically ignore it. */
        p_modifier = strstr( blob->algorithm_wo_parameter_list[i], "_SKIP" );
        if ( p_modifier )
            continue;

        /* Check to see if the DUMP modifier is present.  If found set a
           flag indicating its presense and remove the modifier.  The
           modifier must be removed so that later comparison with the list
           of available algorithms will match. */
        p_modifier = strstr( blob->algorithm_wo_parameter_list[i], "_DUMP" );
        if ( p_modifier )
        {
            dump_mod_present = 1;
            /* If the modifier was found remove it from the parameter name
               so that it can be found in the list of available algorithms */
            p_modifier[0] = '\0';
        }
        else
        {
            dump_mod_present = 0;
        }

        /* Find the matching algorithm from the available algorithm list */
        for ( algo_index = 0; algo_index < number_available; algo_index++ )
        {
            if ( strcasecmp(  blob->algorithm_wo_parameter_list[i],
                 available_algorithms[algo_index].algorithm_name ) == 0 )
            {
                /* Matching algorithm found so add it's API function pointer */
                blob->algorithm_pointer_list[ap_index] =
                    available_algorithms[algo_index].api_algorithm;

                /* Add the algorithm string to the blob for use in error
                   messages and the L1R metadata */
                blob->algorithm_string_list[string_list_index] =
                    blob->algorithm_wo_parameter_list[i];
    
                /* Add pre-and post- proceeing algorithm pointers if they
                   are available.  Even though the pointers will be NULL in
                   the available algorithm list, this check needs to be done
                   so the indices are updated properly. */
                if ( available_algorithms[algo_index].pre_process_algorithm )
                {
                    blob->pre_proc_pointer_list[pre_algo_index] =
                        available_algorithms[algo_index].pre_process_algorithm;
                    pre_algo_index++;
                }
                if ( available_algorithms[algo_index].post_process_algorithm )
                {
                    blob->post_proc_pointer_list[post_algo_index] =
                        available_algorithms[algo_index].post_process_algorithm;
                    post_algo_index++;
                }

                /* Automatically set the required flag to false indicating
                   this algorithm has been added to the list.  Later the
                   list of available algorithms will be searched for any
                   true value. */
                available_algorithms[algo_index].is_required = 0;

                /* If this is the convert_to_float algorithm unset the 
                   flag */
                if ( strcasecmp( blob->algorithm_string_list[string_list_index],
                     PREREQUISITE_ALGORITHM ) == 0 )
                {
                    prerequisite_missing = 0;
                }
                /* Otherwise check the algorithm to see if it needs the 
                   prerequisite and error out if not */
                else if (  prerequisite_missing &&
                    available_algorithms[algo_index].
                    must_run_after_prerequisite )
                {
                    IAS_LOG_ERROR("Algorithm %s must be run before %s, but "
                        "is missing from the algorithm list",
                        PREREQUISITE_ALGORITHM,
                        blob->algorithm_string_list[string_list_index] );
                    free_lists( blob );
                    return ERROR;
                }

                /* Increment the running algorithm pointer index and break
                   out of the loop.  We break so the current for loop 
                   algo_index value may be used to check for no matching
                   algorithm. */
                ap_index++;
                string_list_index++;
                break;
            }
        }  /* END for algo_index loop */
        /* Make sure a match was found.  If no match was found the loop
           counter will have been incremented to the value of number
           available. */
        if ( algo_index == number_available )
        {
            IAS_LOG_ERROR( "No matching algorithm found for: %s",
                blob->algorithm_wo_parameter_list[i] );
            free_lists( blob );
            return ERROR;
        }

        /* Add the dump routine if the modifier was present */
        if ( dump_mod_present )
        {
            /* Sanity check -- only dump_count dumps were allocated. */
            if ( num_of_dump_algos == dump_count )
            {
                IAS_LOG_ERROR(
                    "Attempting to add unallocated dump algorithm for: %s",
                    blob->algorithm_wo_parameter_list[i] );
                free_lists( blob );
                return ERROR;
            }

            /* Add the algorithm */
            blob->algorithm_pointer_list[ap_index] = api_dump_imagery;
            /* Map the algorithm index to the dump algorithm index */
            blob->dump_algorithm_mapping[num_of_dump_algos] = ap_index;
            /* And the name of the algorithm requesting the dump */
            blob->dump_algorithm_string_list[num_of_dump_algos] =
                blob->algorithm_wo_parameter_list[i];
            /* Increment the indices */
            ap_index++;
            num_of_dump_algos++;
        }
    }  /* END main for i loop */

    /* If there were dump algorithms specified add the dump pre- and post-
       process routines.  This is done here because there may have been
       multiple dumps specified and they can only be run once each. */
    if ( num_of_dump_algos > 0 )
    {
        blob->pre_proc_pointer_list[pre_algo_index] =
            dump_preprocess;
        blob->post_proc_pointer_list[post_algo_index] =
            dump_postprocess;
    }

    /* Make sure the algorithm list is the size we expected */
    if ( ap_index > number_allocated )
    {
        IAS_LOG_ERROR( "Number of algorithms exceeds number allocated" );
        free_lists( blob );
        return ERROR;
    }

    /* Finally make sure that required algorithms are in the list.  During
       the for algo_index loop the is_required flag was set to false for all
       algorithms encountered.  If a pass through the list encounters a true
       that ``required'' algorithm was not listed or was SKIPped. */
    for ( i = 0, missing_required = 0; i < number_available; i++ )
    {
        if ( available_algorithms[i].is_required )
        {
            IAS_LOG_ERROR(
                "Algorithm %s is required but not present in "
                "algorithm list", available_algorithms[i].algorithm_name );
            missing_required = 1;
        }
    }
    /* It is an ERROR if any required algorithms are missing */
    if ( missing_required )
    {
        free_lists( blob );
        return ERROR;
    }

    return SUCCESS;
}  /* END create_algorithm_pointer_list */


/*------------------------------------------------------------------------
NAME:    free_lists

PURPOSE: Local routine to free algorithm pointer lists and set them to NULL

RETURNS: none
------------------------------------------------------------------------*/
static void free_lists( BLOB *blob )
{
    free( blob->algorithm_pointer_list );
    free( blob->pre_proc_pointer_list );
    free( blob->post_proc_pointer_list );
    free( blob->dump_algorithm_string_list );
    free( blob->algorithm_string_list );
    blob->algorithm_pointer_list = NULL;
    blob->pre_proc_pointer_list = NULL;
    blob->post_proc_pointer_list = NULL;
    blob->dump_algorithm_string_list = NULL;
    blob->algorithm_string_list = NULL;
}
