/*----------------------------------------------------------------------------
NAME:     database_management.c

ROUTINES: establish_db_connection
          close_db_connection
          delete_db_table_records

PURPOSE:  Collection of routines to manage a database connection.
-----------------------------------------------------------------------------*/
#include <stdlib.h>

#include "ias_db.h"
#include "ias_db_get_connect_info.h"
#include "ias_logging.h"
#include "rps_local.h"

/*----------------------------------------------------------------------------
NAME:    establish_db_connection

PURPOSE: Connect to the database

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int establish_db_connection
(
    BLOB *blob                  /* O: Database connection information */
)
{
    int status;
    char *database_name = NULL;  /* Database name to connect to */
    char *user_name = NULL;      /* user name to use for the connection */
    char *password = NULL;       /* user's password for the database */
    char *host = NULL;           /* host name where the database is located */


    /* Get the database connection info */
    status = ias_db_get_connect_info( &database_name, &user_name, &password,
        &host );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Getting database connection information");
        return ERROR;
    }

    /* Connect to the database with a transaction */
    blob->database_information.db = ias_db_connect_with_transaction(
        database_name, user_name, password, host );

    /* Free the resources before checking the connection */
    free( database_name );
    free( user_name );
    free( password );
    free( host );

    /* Check the connection */
    if ( blob->database_information.db == NULL )
    {
        IAS_LOG_ERROR("Connecting to database");
        return ERROR;
    }

    return SUCCESS;
}  /* END establish_db_connection */


/*----------------------------------------------------------------------------
NAME:    close_db_connection

PURPOSE: Disconnect from the database either with a commit or a rollback

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int close_db_connection
(
    BLOB *blob,         /* I: database connection information */
    int  commit         /* I: flag to indicate commiting the transaction */
)
{
    int status;

    if ( !blob->database_information.db )
        return SUCCESS;

    status = ias_db_disconnect_with_transaction(
        blob->database_information.db, commit );
    if ( status != SUCCESS )
    {
        if ( commit )
        {
            IAS_LOG_ERROR("Commiting and closing database connection");
            return ERROR;
        }
        else
        {
            IAS_LOG_ERROR("Rolling back and closing database connection");
        }
    }
    return SUCCESS;
}  /* END close_db_connection */



/*----------------------------------------------------------------------------
NAME:    delete_db_table_records

PURPOSE: Deletes records from the table passed in.  It only deletes those
         data for the current radiometric work order ID.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int delete_db_table_records
(
    RPS_DATABASE_INFORMATION db_info, /* I: Database connection and
                                            rad_common_id */
    const char *table_name            /* I: Table to delete data from */
)
{
    char query_string[256];
    struct ias_db_query *query = NULL;
    char error_message[IAS_SQL_RTN_MSGLEN];
    int status;

    /* Create the delete string */
    status = snprintf( query_string, sizeof(query_string), "delete from %s "
        "where radiometric_wo_common_id=%s", table_name,
        db_info.rad_common_id );
    if ( status < 0 || status >= sizeof(query_string) )
    {
        IAS_LOG_ERROR("Creating string to delete records from %s",
            table_name );
        return ERROR;
    }

    /* Delete previous records */
    query = ias_db_perform_query( db_info.db, query_string );
    if ( !query )
    {
        IAS_LOG_ERROR("Returned from ias_db_perform_query for: %s",
            query_string );
        return ERROR;
    }

    /* Check the result of the query */
    if ( ias_db_query_was_successful( query ))
    {
        status = SUCCESS;
    }
    else
    {
        ias_db_query_get_error_message( query, error_message,
            IAS_SQL_RTN_MSGLEN );
        IAS_LOG_ERROR("Deleting records from %s where "
            "radiometric_wo_common_id=%s", table_name, db_info.rad_common_id );
        IAS_LOG_ERROR("Message is: %s", error_message );
        status = ERROR;
    }
    ias_db_query_close( query );

    return status;
}  /* END delete_db_table_records */
