/*----------------------------------------------------------------------------
NAME:    image_io_routines.c
         contains:
             initialize_io_routines
             open_l1r_bands
             cleanup_io_routines
             set_l1r_metadata
             read_l0r_image_data
             read_l0r_frame_headers
             read_oli_focal_plane_temperatures
             read_tirs_focal_plane_temperatures
             read_post_warmup_focal_plane_temperatures
             read_lamp_response
             read_post_warmup_lamp_telemetry
             read_stim_lamp_pair_id
             write_l1r_output_image_data
             write_l1r_image_data

PURPOSE: Collection of routines to read L0R and write L1R image data.
         There are external interfaces to retrieve pointers to the data.
-----------------------------------------------------------------------------*/
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <limits.h>

#include "ias_types.h"
#include "ias_threadpool.h"
#include "ias_l0r.h"
#include "ias_l1r.h"
#include "ias_logging.h"
#include "ias_satellite_attributes.h"
#include "rps_local.h"
#include "ias_ancillary.h"
#include "ias_miscellaneous.h" /* for software version */


/* Local lamp-related #defines.  The RPS_STL_MAIN_POWER_STATUS_BIT and
   RPS_STL_PAIR_ID_BITMASK #defines relate to specific bits in the
   'stl_flags' field of the L0R OLI telemetry_header_4 group information. */
#define RPS_STL_POWER_ON               0  /* Lamp pair is powered */
#define RPS_STL_POWER_OFF              1  /* Lamp pair is not powered */
#define RPS_STL_MAIN_POWER_STATUS_BIT  2  /* Bit to check for lamps being
                                             powered */
#define RPS_STL_PAIR_ID_BITMASK        3  /* Mask to isolate the bits that
                                             identify a lamp pair */


/* Mutex for controlling access to the HDF library */
static IAS_THREAD_MUTEX_TYPE hdf_io_mutex;

/* Pointer to the input image */
static int number_of_bands;
static L0RIO *l0r_image = NULL;
static int *l0r_band_open = NULL;

/* Pointer to the output image and output bands */
static L1RIO *l1r_image = NULL;
static L1R_BAND_IO **l1r_band;

/* Pointers for the frame header data */
static IAS_L0R_OLI_FRAME_HEADER *oli_fh = NULL;
static IAS_L0R_TIRS_FRAME_HEADER *tirs_fh = NULL;
static int num_oli_frame_headers;
static int num_tirs_frame_headers;

IAS_THREAD_MUTEX_TYPE *get_hdf_io_mutex( void )
{
    return &hdf_io_mutex;
}

/*-------------------------------------------------------------------------
NAME:    close_the_l0r_imagery

PURPOSE: Local routine to close the L0R imagery

RETURNS: SUCCESS or ERROR
--------------------------------------------------------------------------*/
static int close_the_l0r_imagery( void )
{
    int band_index;  /* Loop index for the bands */
    int status;      /* routine return status */
    int error_occurred = 0; /* Flag to indicate error */

    if ( l0r_image == NULL )
        return SUCCESS;

    /* Close the L0R ancillary data */
    status = ias_l0r_close_ancillary(l0r_image);
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Closing the L0R ancillary data");
        error_occurred = 1;
    }

    /* Close miscellaneous open stuff */
    if ( ias_l0r_close_metadata( l0r_image ))
    {
        IAS_LOG_ERROR("Closing L0R metadata");
        error_occurred = 1;
    }
    if ( ias_l0r_close_header( l0r_image ))
    {
        IAS_LOG_ERROR("Closing L0R header");
        error_occurred = 1;
    }

    /* Close the open band files */
    if ( l0r_band_open != NULL )
    {
        for ( band_index = 0; band_index < number_of_bands; band_index++ )
        {
            if ( l0r_band_open[band_index] )
            {
                int band_number =
                    get_band_attributes( band_index )->band_number;
                status = ias_l0r_close_band( l0r_image, band_number );
                if ( status != SUCCESS )
                {
                    IAS_LOG_ERROR("Closing L0R band %d", band_number );
                    error_occurred = 1;
                }
                else
                {
                    l0r_band_open[band_index] = 0;
                }
            }
        }
    }

    /* Close the L0R */
    status = ias_l0r_close( l0r_image );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Closing L0R file");
        error_occurred = 1;
    }
    else
    {
        l0r_image = NULL;
    }

    /* Free memory used by the l0r_band_open buffer */
    free( l0r_band_open );
    l0r_band_open = NULL;

    /* Return appropriate status */
    if ( error_occurred )
        return ERROR;
    return SUCCESS;
}  /* END close_the_l0r_imagery */


/*------------------------------------------------------------------------
NAME:    get_nth_bit

PURPOSE: Private routine to check if the nth bit value of an integer is 
         set.

RETURNS: 1 (true) the bit is set
         0 (false) the bit isn't set

NOTES:   Bit numbering is defined as bit 0 being the least significant
         bit
------------------------------------------------------------------------*/
static int get_nth_bit
(
    int value,              /* I: Source value for bit information */
    int N                   /* I: Desired bit position */
)
{
    int mask;

    mask = 1 << N;
    if (value & mask)
        return 1;
    return 0;
}


/*-------------------------------------------------------------------------
NAME:    get_post_warmup_start_end_line_numbers

PURPOSE: Local routine to get both the telemetry data starting
         and ending line numbers and the image data starting and ending
         frame numbers and other lamp response data. Since some of them can be 
         obtained with get_blob_lamp_warmup_time_period and 
         get_blob_post_warmup_nframes, only minimum required values are 
         selected as output parameters

RETURNS: SUCCESS or ERROR
---------------------------------------------------------------------------*/
static int get_post_warmup_start_end_line_numbers
( 
    int post_warmup_nframes,       /* I: Number of lines for post warmup 
                                        processing */ 
    float warmup_time_period,     /* I: Time period for lamp warmup */
    int *telemetry_start_line,    /* O: Telemetry data post warmup start line
                                        number */
    int *telemetry_stop_line,     /* O: Telemetry data post warmup stop line
                                        number */
    int *post_warmup_start_frame, /* O: Lamp data post warmup start frame
                                        number */
    int *warmup_start_frame,      /* O: Lamp data warmup start frame number */
    int *warmup_stop_frame,       /* O: Lamp data warmup stop frame number*/
    float *lamp_on_time,          /* O: Lamp turn on time */
    float *post_warmup_start_time,/* O: Lamp data post warmup start time */
    float *post_warmup_stop_time  /* O: Lamp data post warmup stop time */
)
{
    static int lamp_response_data_set = 0;
    static int start_post_warmup_telemetry_rec;
    static int end_post_warmup_telemetry_rec;
    static int lamp_post_warmup_start_frame;
    static int lamp_warmup_start_frame;
    static int lamp_warmup_stop_frame; 
    static float lamp_on_time_in_seconds;
    static float lamp_post_warmup_start_time;
    static float lamp_post_warmup_stop_time; 

    int status;                          /* routine return status */
    int count_of_ancillary_data;         /* Number of ancillary telemetry
                                            records */
    IAS_L0R_OLI_TELEMETRY *telemetry;
    int rec;                             /* loop counter */
    int stim_power_status = RPS_STL_POWER_OFF;
                                         /* Stim lamp power status */
    IAS_L0R_OLI_FRAME_HEADER l0r_one_frame_headers;
    IAS_L0R_OLI_FRAME_HEADER *l0r_frame_headers;
    int num_frame_headers;               /* Actual number of OLI or TIRS
                                            frame headers */
    size_t buffer_size;                  /* Actual buffer size */
    int post_warmup_stop_frame;

    /* Only calculate the outputs when they heven't been calculated before,
       therefore, only one time calculation is needed */
    if (!lamp_response_data_set)
    {
        double lamp_on_time_j2000 = 0.0;
        double lamp_off_time_j2000 = 0.0;
        double lamp_post_start_time_j2000;
        double image_post_stop_time_j2000;
        double record_time_j2000;
        int lamp_on_record_number = 0;

        /* Find out how many records there are in the telemetry data */
        status = ias_l0r_get_ancillary_oli_telemetry_records_count( l0r_image,
            &count_of_ancillary_data );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Getting size of ancillary data");
            return ERROR;
        }

        /* Allocate memory for telemetry */
        telemetry = malloc( count_of_ancillary_data * 
                            sizeof(IAS_L0R_OLI_TELEMETRY) );
        if ( !telemetry )

        {
            IAS_LOG_ERROR("Allocating memory for telemetry group");
            return ERROR;
        }

        /* Read out the whole telemetry data set */
        status = ias_l0r_get_ancillary_oli_telemetry( l0r_image, 0, 
                 count_of_ancillary_data, telemetry );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Getting ancillary oli telemetry data");
            free(telemetry);
            return ERROR;
        }
  
        /* Find the lamp_on_time_in_seconds 
           Note: The lamp is normally on shortly after the lamp data collection
                 starts, therefore, there is no need to use binary search */
        for ( rec = 0; rec <= count_of_ancillary_data - 1; rec++ )
        {
            int stl_flags =
                telemetry[rec].oli_telemetry_header_4.stl_flags;

            stim_power_status = get_nth_bit(stl_flags,
                RPS_STL_MAIN_POWER_STATUS_BIT);
            if (stim_power_status == RPS_STL_POWER_ON)
            {
                lamp_on_time_j2000 =
                   IAS_L0R_CONVERT_TIME_TO_SECONDS_SINCE_J2000(
                        telemetry[rec].oli_telemetry_header_3.l0r_time);
                lamp_on_record_number = rec;
                lamp_on_time_in_seconds = telemetry[rec].
                    oli_telemetry_header_3.l0r_time.seconds_of_day;
                break;   
            }
        }

        /* Return an error if the previous loop didn't detect that the
           lamp was turned on */
        if (stim_power_status == RPS_STL_POWER_OFF)
        {
            IAS_LOG_ERROR("Lamps do not appear to have been turned on");
            free(telemetry);
            return ERROR;
        }

        /* Add the warmup time period to the lamp on time */
        lamp_post_start_time_j2000 = lamp_on_time_j2000 + warmup_time_period;


        /* Find the post warmup telemetry data start line 
           Note: No need to use binary search since it should break out
                 in a few iterations */
        start_post_warmup_telemetry_rec = -1;
        for ( rec = lamp_on_record_number + 1;
              rec < count_of_ancillary_data; rec++ )
        {

            record_time_j2000 = IAS_L0R_CONVERT_TIME_TO_SECONDS_SINCE_J2000(
                telemetry[rec].oli_telemetry_header_3.l0r_time);
            if ( record_time_j2000 >= lamp_post_start_time_j2000 )
            {
                start_post_warmup_telemetry_rec = rec;            
                break;
            }                      
        }

        /* Ensure the data wasn't exhausted before finding the post warmup
           record */
        if ( start_post_warmup_telemetry_rec == -1 )
        {
            IAS_LOG_ERROR("Finding the post warmup telemetry record");
            free(telemetry);
            return ERROR;
        }

        /* Get the whole OLI frame headers */
        l0r_frame_headers = (IAS_L0R_OLI_FRAME_HEADER *)read_l0r_frame_headers(
                             MM_L0R_OLI_FRAME_HEADER, &num_frame_headers, 
                             &buffer_size);
        if (l0r_frame_headers == NULL)
        {
            IAS_LOG_ERROR("Retrieving L0R frame headers");
            free(telemetry);
            return ERROR;
        }

        /* Find the image post warmup start time and line number using the
           binary search idea */
        int low = 0;
        int high = num_frame_headers -1;

        while ( low <= high )
        {       
            int middle = low + ( high - low )/2;
            record_time_j2000 = IAS_L0R_CONVERT_TIME_TO_SECONDS_SINCE_J2000(
                l0r_frame_headers[middle].l0r_time );
            if ( record_time_j2000 < lamp_post_start_time_j2000 )
            {
                low = middle + 1;
            }
            else if ( record_time_j2000 > lamp_post_start_time_j2000 )
            {
                if ( IAS_L0R_CONVERT_TIME_TO_SECONDS_SINCE_J2000(
                    l0r_frame_headers[middle-1].l0r_time ) > 
                    lamp_post_start_time_j2000 ) 
                    {
                        high = middle - 1;
                    }
                else
                {
                    lamp_post_warmup_start_frame = 
                        l0r_frame_headers[middle].frame_number;
                    lamp_post_warmup_start_time = l0r_frame_headers[middle].
                                      l0r_time.seconds_of_day;
                    break;
                }
            }
            else
            {
                lamp_post_warmup_start_frame = 
                    l0r_frame_headers[middle].frame_number;
                lamp_post_warmup_start_time = l0r_frame_headers[middle].
                                  l0r_time.seconds_of_day;
                break;
            }
        }

        /* Get the image post_warmup_stop_frame */
        post_warmup_stop_frame = lamp_post_warmup_start_frame + 
            post_warmup_nframes - 1;
        if ( post_warmup_stop_frame > num_frame_headers )
        {
            IAS_LOG_ERROR("There is not enough image data for %d lines of "
                "post-warmup data.  Number of image frames is %d and "
                "calculated frame is %d.", post_warmup_nframes,
                num_frame_headers, post_warmup_stop_frame );
            free(telemetry);
            return ERROR;
        }

        /* Find the image post warmup stop time */
        status = ias_l0r_get_oli_frame_headers(l0r_image,
            post_warmup_stop_frame, 1, &l0r_one_frame_headers);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading OLI frame header data");
            free(telemetry);
            return ERROR;
        }
        image_post_stop_time_j2000 =
            IAS_L0R_CONVERT_TIME_TO_SECONDS_SINCE_J2000(
                l0r_one_frame_headers.l0r_time);
        lamp_post_warmup_stop_time = 
             l0r_one_frame_headers.l0r_time.seconds_of_day; 

        /* Find the telemetry post_warmup_stop_frame, use last line of telemetry
           data as telemetry post_warmup_stop_frame if the last line recording 
           time is less than post_warmup_stop_time 
           Note: No need to use binary search since it should break out
                 in a few iterations */
        end_post_warmup_telemetry_rec = -1;
        for ( rec = start_post_warmup_telemetry_rec; 
              rec <= count_of_ancillary_data - 1; rec++ )
        {
            record_time_j2000 = IAS_L0R_CONVERT_TIME_TO_SECONDS_SINCE_J2000(
                telemetry[rec].oli_telemetry_header_3.l0r_time);

            end_post_warmup_telemetry_rec = rec; 
            if ( record_time_j2000 >= image_post_stop_time_j2000 )
                break;
        }

        /* Find the image warmup start line number */
        for ( rec = 0; rec <= num_frame_headers - 1; rec++ )
        {
            record_time_j2000 = IAS_L0R_CONVERT_TIME_TO_SECONDS_SINCE_J2000(
                l0r_frame_headers[rec].l0r_time);
            if (record_time_j2000 >= lamp_on_time_j2000)
            {
                lamp_warmup_start_frame =
                    l0r_frame_headers[rec].frame_number; 
                break;
            }                      
        }

        /* Initialize the lamp off time to 0 */
        lamp_off_time_j2000 = 0;

        /* Find the lamp off time.  If the lamp is always on, the off time
           is already set to the proper value.  */
        for ( rec = end_post_warmup_telemetry_rec;
            rec <= count_of_ancillary_data - 1; rec++ )
        {
            int stl_flags = telemetry[rec].oli_telemetry_header_4.stl_flags;

            stim_power_status = get_nth_bit(stl_flags,
                RPS_STL_MAIN_POWER_STATUS_BIT);
            if (stim_power_status == RPS_STL_POWER_OFF)
            {
                lamp_off_time_j2000 = 
                    IAS_L0R_CONVERT_TIME_TO_SECONDS_SINCE_J2000(
                        telemetry[rec].oli_telemetry_header_3.l0r_time);
                break;   
            }
        }

        /* Find the image warmup stop line number */
        for ( rec = lamp_post_warmup_start_frame; rec <= num_frame_headers - 1;
              rec++ )
        {
            record_time_j2000 = IAS_L0R_CONVERT_TIME_TO_SECONDS_SINCE_J2000(
                l0r_frame_headers[rec].l0r_time);
            if (record_time_j2000 > lamp_off_time_j2000)
            {
                lamp_warmup_stop_frame =
                    l0r_frame_headers[rec].frame_number; 
                break;
            }                      
        }

        /* If lamp is always on for the scene, last record number is used
           for as the lamp_warmup_stop_line */
        if (lamp_off_time_j2000 == 0)
        {
            lamp_warmup_stop_frame =
                l0r_frame_headers[num_frame_headers - 1].frame_number; 
        }

        /* Set the lamp_response_data_set be 1 */
        lamp_response_data_set = 1;

        /* Free the telemetery memory */
        free(telemetry);

        /* The l0r_frame_headers will be freed somewhere else */
    }  /* END if lamp_response_data_set not set */

    /* Populate the output parameters with local variable values */
    *telemetry_start_line = start_post_warmup_telemetry_rec;
    *telemetry_stop_line = end_post_warmup_telemetry_rec;
    *post_warmup_start_frame = lamp_post_warmup_start_frame;
    *warmup_start_frame = lamp_warmup_start_frame;
    *warmup_stop_frame = lamp_warmup_stop_frame;
    *lamp_on_time = lamp_on_time_in_seconds;
    *post_warmup_start_time = lamp_post_warmup_start_time;
    *post_warmup_stop_time = lamp_post_warmup_stop_time;

    return SUCCESS;
}  /* END get_post_warmup_start_end_line_numbers */


/*------------------------------------------------------------------------
NAME:    get_oli_tirs_frame_rates

PURPOSE: Get the OLI and TIRS frame rates if they are available and set
         them in the blob

RETURNS: SUCCESS or ERROR
------------------------------------------------------------------------*/
int get_oli_tirs_frame_rates
(
    BLOB *blob          /* O: Set OLI and TIRS frame rates */
)
{
    IAS_L0R_OLI_FRAME_HEADER *oli_frame_headers = NULL;
    IAS_L0R_TIRS_FRAME_HEADER *tirs_frame_headers = NULL;
    int num_oli_frame_headers;
    int num_tirs_frame_headers;
    double oli_start_frame_time;     /* OLI start frame time */
    double oli_stop_frame_time;      /* OLI stop frame time */
    double tirs_start_frame_time;    /* OLI start frame index */
    double tirs_stop_frame_time;     /* OLI stop frame index */
    int oli_start_frame_index;       /* TIRS start frame time */
    int oli_stop_frame_index;        /* TIRS stop frame time */
    int tirs_start_frame_index;      /* TIRS start frame index */
    int tirs_stop_frame_index;       /* TIRS start frame index */
    int status;

    /* Get a count of OLI frame headers */
    status = ias_l0r_get_oli_frame_headers_records_count( l0r_image,
        &num_oli_frame_headers );
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Obtaining number of OLI frame headers");
        return ERROR;
    }

    /* Get the count of TIRS frame headers */
    status = ias_l0r_get_tirs_frame_headers_records_count( l0r_image,
        &num_tirs_frame_headers );
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Obtaining number of TIRS frame headers");
        return ERROR;
    }

    /* If there are OLI frame headers, use them to get the start and stop
       times */
    if (num_oli_frame_headers > 0)
    {
        /* Allocate space for the OLI frame header */
        oli_frame_headers = malloc(sizeof(*oli_frame_headers) * 
                               num_oli_frame_headers);
        if (!oli_frame_headers)
        {
            IAS_LOG_ERROR("Allocating memory for OLI frame headers");
            return ERROR;
        }

        /* Get the OLI frame headers */
        status = ias_l0r_get_oli_frame_headers(l0r_image, 0, 
                    num_oli_frame_headers, oli_frame_headers);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading OLI frame headers");
            free(oli_frame_headers);
            return ERROR;
        }
    }

    if (num_tirs_frame_headers > 0)
    {
        /* Allocate space for the TIRS frame header */
        tirs_frame_headers = malloc(sizeof(*tirs_frame_headers) * 
                               num_tirs_frame_headers);
        if (!tirs_frame_headers)
        {
            IAS_LOG_ERROR("Allocating memory for TIRS frame headers");
            if (oli_frame_headers)
                free(oli_frame_headers);   
            return ERROR;
        }

        /* Get the TIRS frame headers */
        status = ias_l0r_get_tirs_frame_headers(l0r_image, 0, 
                    num_tirs_frame_headers, tirs_frame_headers);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading TIRS frame headers");
            if (oli_frame_headers)
                free(oli_frame_headers);   
            free(tirs_frame_headers);
            return ERROR;
        }
    }

    /* Get the OLI and TIRS first valid and last vaild frame times and
       their associated index numbers */
    status = ias_ancillary_get_start_stop_frame_times(oli_frame_headers, 
        tirs_frame_headers, num_oli_frame_headers, num_tirs_frame_headers, 
        &oli_start_frame_time, &oli_stop_frame_time, &tirs_start_frame_time, 
        &tirs_stop_frame_time, &oli_start_frame_index, &oli_stop_frame_index, 
        &tirs_start_frame_index, &tirs_stop_frame_index);
    if (status != SUCCESS)
    {
            IAS_LOG_ERROR("Getting OLI/TIRS start/stop frame times");
            if (oli_frame_headers)
                free(oli_frame_headers);   
            if (tirs_frame_headers)
                free(tirs_frame_headers);
            return ERROR;
    }

    /* Calculate the OLI frame_rate and populate the blob elements */
    if (oli_frame_headers)
    {
        if (oli_stop_frame_index != oli_start_frame_index)
        {    
            blob->report_information.oli_frame_rate = (oli_stop_frame_time - 
                oli_start_frame_time) / (oli_stop_frame_index - 
                oli_start_frame_index);
        }
        else
        {
            blob->report_information.oli_frame_rate = -1.0;
        }

        blob->oli_start_time = oli_start_frame_time;
        blob->oli_stop_time = oli_stop_frame_time;
    }
    else
    {
        blob->report_information.oli_frame_rate = -1.0;
        blob->oli_start_time = -1.0;
        blob->oli_stop_time = -1.0;
    }

    /* Calculate the TIRS frame_rate and populate the blob elements */
    if (tirs_frame_headers)
    {
        if (tirs_stop_frame_index != tirs_start_frame_index)
        {
            blob->report_information.tirs_frame_rate = (tirs_stop_frame_time - 
                tirs_start_frame_time) / (tirs_stop_frame_index - 
                tirs_start_frame_index);
        }
        else
        {
            blob->report_information.tirs_frame_rate = -1.0;
        }

        blob->tirs_start_time = tirs_start_frame_time;
        blob->tirs_stop_time = tirs_stop_frame_time;
    }
    else
    {
        blob->report_information.tirs_frame_rate = -1.0;
        blob->tirs_start_time = -1.0;
        blob->tirs_stop_time = -1.0;
    }

    /* Free the allocated memory */
    if (oli_frame_headers)
        free(oli_frame_headers);   
    if (tirs_frame_headers)
        free(tirs_frame_headers);

    return SUCCESS;
}

/*------------------------------------------------------------------------
NAME:    open_tirs_ghost_image_bands

PURPOSE: This routine opens the tirs ghost image input bands.  It uses the band
         attributes in case the number of output lines has been changed
         due to algorithm pre-process routines or other parameter changes.

RETURNS: SUCCESS or ERROR
------------------------------------------------------------------------*/
static int open_tirs_ghost_image_bands
(
    BLOB *blob  /* I/O: list of bands, indices, number of bands, band
                      information */
)
{
    int band_index; /* Band index loop variable */
    IAS_L1R_FILE_METADATA file_metadata; /* L1R file metadata */

    /* Set the pointers to 0 to get the defined behavior on errors */
    memset(blob->tirs_stray_light_band_io, 0, sizeof(L1R_BAND_IO *)        
        * IAS_MAX_NBANDS);

    /* Open the L1R TIRS stray light image */
    blob->tirs_stray_light_l1r_io = ias_l1r_open_image(
        blob->tirs_stray_light_correction_filename, IAS_READ);
    if (!blob->tirs_stray_light_l1r_io)
    {
        IAS_LOG_ERROR("Opening stray light correction image %s", 
            blob->tirs_stray_light_correction_filename);
        return ERROR;
    }

    /* Get the stray light file metadata */
    if (ias_l1r_get_file_metadata(blob->tirs_stray_light_l1r_io, &file_metadata)
        != SUCCESS)
    {
        IAS_LOG_ERROR("Reading tirs stray light file metadata");         
        ias_l1r_close_image(blob->tirs_stray_light_l1r_io);
        return ERROR;
    }
    
    /* Copy the stray light source */
    strcpy(blob->stray_light_correction_source, file_metadata.collect_type);

    /* Allocate the io bands for TIRS stray light */
    for (band_index = 0; band_index < IAS_MAX_NBANDS; band_index++)
    {
        IAS_DATA_TYPE data_type;    /* Input L1R data type */
        int number_scas;            /* Input L1R number of SCAs */
        int number_lines;           /* Input L1R number of lines */
        int number_samples;         /* Input L1R number of samps */
        int band_check = SUCCESS;   /* Band consistency check flag */
        RPS_BAND_ATTRIBUTES *band_attributes
            = get_band_attributes(band_index);

        /* Only allocate memory for TIRS bands that are being processed */
        if (band_attributes->sensor_id != IAS_TIRS
            || !blob->band_information[band_index]->process_band)
        {
            continue;
        }

        /* Ensure that the band is present in L1R */
        if (!ias_l1r_is_band_present(blob->tirs_stray_light_l1r_io,
            band_attributes->band_number))
        {
            IAS_LOG_ERROR("Band %d is not present in stray light file",
                band_attributes->band_number);
            band_check = ERROR;
        }
        else
        {
            /* Open this L1R band for reading */
            blob->tirs_stray_light_band_io[band_index] = ias_l1r_open_band(
                blob->tirs_stray_light_l1r_io, band_attributes->band_number,
                &data_type, &number_scas, &number_lines, &number_samples);
            if (!blob->tirs_stray_light_band_io[band_index])
            {
                IAS_LOG_ERROR("Opening stray light correction band number %d",
                    band_attributes->band_number);
                band_check = ERROR;
            }
            /* Make some consistency checks between input and output image */
            else if (data_type != IAS_R4)
            {
                IAS_LOG_ERROR("Bad datatype for stray light correction image");
                band_check = ERROR;
            }
            else if (number_scas != band_attributes->number_scas)
            {
                IAS_LOG_ERROR("Stray light correction band number %d has a "
                    "different number of SCAs (%d) than the output image (%d)",
                    band_attributes->band_number, number_scas,
                    band_attributes->number_scas);
                band_check = ERROR;
            }
            else if (number_lines != band_attributes->number_lines)
            {
                IAS_LOG_ERROR("Stray light correction band number %d has a "
                    "different number of lines (%d) than the output image (%d)",
                    band_attributes->band_number, number_lines,
                    band_attributes->number_lines);
                band_check = ERROR;
            }
            else if (number_samples != band_attributes->number_detectors)
            {
                IAS_LOG_ERROR("Stray light correction band number %d has a "
                    "different number of samples (%d) than the output image "
                    "(%d)", band_attributes->band_number, number_samples,
                    band_attributes->number_detectors);
                band_check = ERROR;
            }
        }

        /* Make sure no errors occurred */
        if (band_check != SUCCESS)
        {
            int index;
            IAS_LOG_ERROR("Stray light correction band consistency check "
                "failed for band number %d", band_attributes->band_number);
        
            /* Close all the ghost image bands */
            for (index = 0; index < band_index; index++)
            {
                if (blob->tirs_stray_light_band_io[index])
                {
                    ias_l1r_close_band(blob->tirs_stray_light_band_io[index]);
                    blob->tirs_stray_light_band_io[index] = NULL;
                }
            }
            
            /* Close the ghost image */
            ias_l1r_close_image(blob->tirs_stray_light_l1r_io);
            blob->tirs_stray_light_l1r_io = NULL;

            return ERROR;
        }
    }
    
    return SUCCESS;
}

/*------------------------------------------------------------------------
NAME:    initialize_io_routines

PURPOSE: Public routine to initialize the image I/O routines.  This includes
         opening the files and gathering metadata from the L0R.

         NOTE: This routine doesn't open the L1R bands since the number of
               lines to process may not be the same as the number of L0R
               image lines.

RETURNS: SUCCESS or ERROR
------------------------------------------------------------------------*/
int initialize_io_routines
(
    BLOB *blob          /* I: Contains L0R and L1R file names */
                        /* O: Metadata obtained from L0R file */
)
{
    int band_number;
    int band_index;
    int index;          /* loop counter */
    int status;         /* function return status */
    int count;          /* number of l0r metadata records */
    char *filename;     /* pointer to l0r file name */
    char path_to_filename[PATH_MAX];  /* l0r path without file name */
    IAS_L0R_INTERVAL_METADATA l0r_interval_metadata;
    IAS_L0R_SCENE_METADATA scene_metadata;
    IAS_L0R_OLI_IMAGE_HEADER l0r_image_header;
    int list_of_l0r_bands[blob->total_bands];  
    int number_of_l0r_bands_present; /* Count of bands in L0R input files */
    const IAS_BAND_ATTRIBUTES *standard_atts; /* Band attributes structure
                                                 (single band) */
    int size;           /* Image size returned from L0R library */
    int process_blind_band[IAS_MAX_TOTAL_BANDS];
    int use_oli_subinterval_frames_flag = FALSE;
    int use_tirs_subinterval_frames_flag = FALSE;
    int oli_target_scene_frames = 0;
    int oli_target_scene_start_frame = 0;
    int tirs_target_scene_frames = 0;
    int tirs_target_scene_start_frame = 0;

    number_of_bands = blob->total_bands;

    /* Create a mutex for control of the HDF library */
    status = IAS_THREAD_CREATE_MUTEX( &hdf_io_mutex );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Couldn't create HDF I/O mutex");
        return ERROR;
    }

    /* -------------------- Initialize the L0R library -------------------- */
    /* The l0r_image structure is passed to every routine to identify the
       data.  The L0R file name in the blob needs to be split for the open
       call. */
    strcpy( path_to_filename, blob->l0r_file_name );
    filename = strrchr( path_to_filename, '/' );
    if ( !filename )
    {
        filename = path_to_filename;
    }
    else
    {
        *filename = '\0';
        filename++;
    }
    l0r_image = ias_l0r_open( filename, path_to_filename, 0 );
    if ( !l0r_image )
    {
        IAS_LOG_ERROR("Initializing the L0R data structure");
        return ERROR;
    }

    /* Open the metadata */
    status = ias_l0r_open_metadata( l0r_image, IAS_READ );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Opening the metadata file");
        close_the_l0r_imagery();
        return ERROR;
    }

    /* Get a count of the interval metadata, if there aren't any records
       it's an error. */
    status = ias_l0r_get_interval_metadata_records_count( l0r_image, &count );
    if ( status != SUCCESS || count <= 0 )
    {
        IAS_LOG_ERROR("There is no interval metadata record in L0R");
        close_the_l0r_imagery();
        return ERROR;
    }

    /* Read the interval metadata */
    status = ias_l0r_get_interval_metadata( l0r_image, &l0r_interval_metadata );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Reading L0R interval metadata");
        close_the_l0r_imagery();
        return ERROR;
    }

    /* Get a count of the scene metadata.  If there is a single record, get the
       date_acquired from it.  If there are multiple scenes find the correct
       metadata using the scene id.  If no scene id is present then get the
       date_acquired from the interval metadata. */
    status = ias_l0r_get_scene_metadata_records_count( l0r_image, &count );
    if ( status == SUCCESS && count == 1 )
    {
        status = ias_l0r_get_scene_metadata( l0r_image, 0, 1, &scene_metadata );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Reading the scene metadata for %s", filename );
            close_the_l0r_imagery();
            return ERROR;
        }

        if ( !strcasecmp(blob->date_acquired, "" ))
        {
            strncpy( blob->date_acquired, scene_metadata.date_acquired, 
                     IAS_L0R_DATE_LENGTH );
            blob->date_acquired[IAS_L0R_DATE_LENGTH] = '\0';
        }
    }
    else if ( status == SUCCESS && count >= 0 )
    {
        /* If a specific scene was requested identify the frames that 
           cover it */
        if ( count > 0 && strlen(blob->landsat_scene_id) > 0 )
        {
            IAS_L0R_SCENE_METADATA *scene_metadata; /* Scene metadata pointer */
            int index;                /* Generic loop variable */
            int oli_start_index = 0;  /* OLI scene start index */ 
            int tirs_start_index = 0; /* TIRS scene start index */

            /* Allocate memory for the scene metadata */
            scene_metadata = malloc(count * sizeof(*scene_metadata));
            if (!scene_metadata)
            {
                IAS_LOG_ERROR("Failed allocating memory for %d records of "
                    "scene metadata", count);
                close_the_l0r_imagery();
                return ERROR;
            }

            /* Read the scene metadata */
            if (ias_l0r_get_scene_metadata(l0r_image, 0, count, scene_metadata)
                != SUCCESS)
            {
                IAS_LOG_ERROR("Reading L0R scene metadata");
                free(scene_metadata);
                close_the_l0r_imagery();
                return ERROR;
            }

            /* Determine the TIRS and OLI index starts */
            for (oli_start_index = 0; oli_start_index < count; 
                 oli_start_index++)
            {
                if (scene_metadata[oli_start_index].sensor_present_oli)
                    break;
            }
            if (oli_start_index < count && !blob->use_three_tirs_scenes_flag)
                use_oli_subinterval_frames_flag = TRUE;

            for (tirs_start_index = 0; tirs_start_index < count; 
                 tirs_start_index++)
            {
                if (scene_metadata[tirs_start_index].sensor_present_tirs)
                    break;           
            }
            if (tirs_start_index < count && !blob->use_three_tirs_scenes_flag)
                use_tirs_subinterval_frames_flag = TRUE;

            /* Find the target scene */
            for (index = 0; index < count; index++)
            {
                if (strcmp(blob->landsat_scene_id, 
                    scene_metadata[index].landsat_scene_id) == 0)
                    break;
            }
            if (index >= count)
            {
                IAS_LOG_ERROR("Scene ID %s not found in the L0R", 
                    blob->landsat_scene_id);
                free(scene_metadata);
                close_the_l0r_imagery();
                return ERROR;
            }
        
            /* Store the scene acquired data */
            if (!strcasecmp(blob->date_acquired, ""))
            {
                strncpy(blob->date_acquired, 
                    scene_metadata[index].date_acquired, IAS_L0R_DATE_LENGTH);
                blob->date_acquired[IAS_L0R_DATE_LENGTH] = '\0';
            }
            
            /* If only part of the interval is being processed adjust the frames
               to be read */
            if (use_tirs_subinterval_frames_flag)
            {
                tirs_target_scene_start_frame
                    = scene_metadata[index].scene_start_frame_tirs;
                tirs_target_scene_frames = 
                    scene_metadata[index].scene_stop_frame_tirs
                    - tirs_target_scene_start_frame + 1;
                tirs_target_scene_start_frame -= 
                    scene_metadata[tirs_start_index].scene_start_frame_tirs;
            }
            if (use_oli_subinterval_frames_flag)
            {
                oli_target_scene_start_frame
                    = scene_metadata[index].scene_start_frame_oli;
                oli_target_scene_frames = 
                    scene_metadata[index].scene_stop_frame_oli
                    - oli_target_scene_start_frame + 1;
                oli_target_scene_start_frame -= 
                    scene_metadata[oli_start_index].scene_start_frame_oli;
            }

            /* Free the scene metadata */
            free(scene_metadata);
        }
        /* If multiple scenes were encountered issue a warning */
        else if (count > 0)
        {
            IAS_LOG_WARNING("%d scenes were encountered -- using interval "
                "metadata for date acquired", count );
        }
        if ( !strcasecmp(blob->date_acquired, "" ))
        {
            strncpy( blob->date_acquired, l0r_interval_metadata.date_acquired, 
                    IAS_L0R_DATE_LENGTH);
            blob->date_acquired[IAS_L0R_DATE_LENGTH] = '\0';
        }
    }
    else
    {
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Error getting scene metadata count for %s",
                          filename);
        }
        else
        {
            IAS_LOG_ERROR("Found %d scene metadata records when expecting 1", 
                          count);
        }
        close_the_l0r_imagery();
        return ERROR;
    }

    /* Open the L0R header */
    status = ias_l0r_open_header( l0r_image, IAS_READ );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Opening the L0R header file");
        close_the_l0r_imagery();
        return ERROR;
    }

    /* Allocate a list of 1-based bands present in the L0R */
    l0r_band_open = calloc( sizeof(int), number_of_bands );
    if ( !l0r_band_open )
    {
        IAS_LOG_ERROR("Allocating band open array");
        close_the_l0r_imagery();
        return ERROR;
    }

    /* Determine what bands are present in the L0R imagery */
    index = number_of_l0r_bands_present = 0;
    for ( band_index = 0; band_index < number_of_bands; band_index++ )
    {
        int is_present;
        int i;

        /* At this time we have no idea what the band numbers are since the
           there is no guarantee of the band sequence numbers.  However band
           indices are supposed to start at 0 and increment by one.  Get a
           band number for the band_index */
        band_number = ias_sat_attr_convert_band_index_to_number( band_index );
        if ( band_number == ERROR )
        {
            IAS_LOG_ERROR("Getting a band number for index %d", band_index );
            close_the_l0r_imagery();
            return ERROR;
        }

        /* See if the band is present */
        status = ias_l0r_get_band_present( l0r_image, band_number,
            &is_present );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Checking for presence of band %d", band_number );
            close_the_l0r_imagery();
            return ERROR;
        }
        else if ( is_present )
        {
            blob->band_information[band_index]->band_present = 1;
            list_of_l0r_bands[index] = band_number;
            index++;
            number_of_l0r_bands_present++;
        }
        else
        {
            blob->band_information[band_index]->process_band = 0;
            blob->band_information[band_index]->band_present = 0;
        }

        /* Throw a warning if the band isn't present in the imagery but is
           in the list of bands to process */
        for ( i = 0; i < blob->number_of_bands; i++ )
        {
            if ( blob->list_of_bands[i] == band_number )
            {
                if ( ! is_present )
                {
                    int j;
                    /* Remove the band from the list of bands by moving the
                       remaining bands up one and decrementing the number of
                       bands and throw a warning */
                    for ( j = i; j < blob->number_of_bands; j++ )
                    {
                        blob->list_of_bands[j] = blob->list_of_bands[j + 1];
                        blob->list_of_band_indices[j] =
                            blob->list_of_band_indices[j + 1];
                    }
                    blob->number_of_bands--;
                    IAS_LOG_WARNING("Band %d is in the list of bands to "
                        "process but isn't in the L0R imagery.  The band "
                        "is being removed from the processing list.",
                        band_number );
                }
            }
        }
    }

    /* The blob list of bands may have been pared down due to bands that
       aren't present in the imagery.  Make sure there is at least one band
       in the list. */
    if ( blob->number_of_bands < 1 )
    {
        IAS_LOG_ERROR("The L0R imagery doesn't contain any of the bands in "
            "the input band list");
        close_the_l0r_imagery();
        return ERROR;
    }

    /* Now go through the list of bands in the L0R opening the bands and
       populating the band information structure.  Directly access the blob
       attributes structure since it's available. */
    for ( index = 0; index < IAS_MAX_TOTAL_BANDS; index++ )
        process_blind_band[index] = 0;
    for ( index = 0; index < number_of_l0r_bands_present; index++ )
    {
        int band_index;
        int vrp_band_number = 0;
        int vrp_band_index;
        int blind_band_number = 0;
        int blind_band_index;
        int i;

        /* Get the standard attributes for the band */
        band_number = list_of_l0r_bands[index];
        standard_atts = ias_sat_attr_get_band_attributes( band_number );
        if ( !standard_atts )
        {
            IAS_LOG_ERROR("Getting satellite attributes for band %d",
                band_number );
            close_the_l0r_imagery();
            return ERROR;
        }
        band_index = standard_atts->band_index;

        /* Open the band data file */
        status = ias_l0r_open_band( l0r_image, band_number, IAS_READ );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Opening band %d", band_number );
            close_the_l0r_imagery();
            return ERROR;
        }
        l0r_band_open[band_index] = 1;

        /* Get the number of lines in the image from the L0R file */
        status = ias_l0r_get_band_records_count( l0r_image, band_number,
            &size );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Getting number of lines for band %d", band_number );
            close_the_l0r_imagery();
            return ERROR;
        }
        
        /* Determine the number of band lines */
        blob->band_information[band_index]->number_lines = size;
        if ( use_tirs_subinterval_frames_flag
             && standard_atts->sensor_id == IAS_TIRS )
        {
            blob->band_information[band_index]->number_lines 
                = tirs_target_scene_frames * standard_atts->lines_per_frame;
            blob->band_information[band_index]->starting_line
                = tirs_target_scene_start_frame
                * standard_atts->lines_per_frame;
        }
        else if (use_oli_subinterval_frames_flag 
            && standard_atts->sensor_id == IAS_OLI)
        {   
            blob->band_information[band_index]->number_lines 
                = oli_target_scene_frames * standard_atts->lines_per_frame;
            blob->band_information[band_index]->starting_line
                = oli_target_scene_start_frame * standard_atts->lines_per_frame;
        }
        
        IAS_LOG_DEBUG("LEVEL1 Band: %d  size %d", band_number, size );

        /* See if the band is in the list to process and set the processing
           flag and also store its index */
        blob->band_information[band_index]->process_band = 0;
        for ( i = 0; i < blob->number_of_bands; i++ )
        {
            if ( blob->list_of_bands[i] == band_number )
            {
                blob->band_information[band_index]->process_band = 1;
                break;
            }
        }

        /* Set the standard band information */
        blob->band_information[band_index]->band_number = band_number;
        vrp_band_number = standard_atts->vrp_band_number;
        if ( vrp_band_number <= 0 )
            blob->band_information[band_index]->vrp_band_index = -1;
        blind_band_number = standard_atts->blind_band_number;
        if ( blind_band_number <= 0 )
            blob->band_information[band_index]->blind_band_index = -1;
        if ( standard_atts->band_classification & IAS_NORMAL_BAND )
            blob->band_information[band_index]->is_normal_band = 1;
        else
            blob->band_information[band_index]->is_normal_band = 0;
        blob->band_information[band_index]->vrp_band_number = vrp_band_number;
        blob->band_information[band_index]->blind_band_number =
            blind_band_number;
        blob->band_information[band_index]->number_detectors =
            standard_atts->detectors_per_sca;
        blob->band_information[band_index]->number_scas = standard_atts->scas;
        blob->band_information[band_index]->band_classification =
            standard_atts->band_classification;
        blob->band_information[band_index]->sensor_id =
            standard_atts->sensor_id;
        blob->band_information[band_index]->spectral_type =
            standard_atts->spectral_type;
        blob->band_information[band_index]->one_sca_present = 0;
        blob->band_information[band_index]->sca_number = 0;

        if ( vrp_band_number > 0 )
        {
            /* Set the standard VRP information */
            standard_atts 
                = ias_sat_attr_get_band_attributes( vrp_band_number );
            if ( !standard_atts )
            {
                IAS_LOG_ERROR("Getting satellite attributes for band %d",
                    vrp_band_number );
                close_the_l0r_imagery();
                return ERROR;
            }
            vrp_band_index = standard_atts->band_index;
            blob->band_information[vrp_band_index]->band_number =
                vrp_band_number;
            blob->band_information[vrp_band_index]->is_normal_band = 0;
            blob->band_information[vrp_band_index]->number_detectors =
                standard_atts->detectors_per_sca;
            blob->band_information[vrp_band_index]->number_scas =
                standard_atts->scas;
            blob->band_information[vrp_band_index]->band_classification =
                standard_atts->band_classification;
            blob->band_information[vrp_band_index]->spectral_type =
                standard_atts->spectral_type;
            blob->band_information[vrp_band_index]->one_sca_present = 0;
            blob->band_information[vrp_band_index]->sca_number = 0;
            blob->band_information[band_index]->vrp_band_index =
                vrp_band_index;

            /* Only the normal band has the blind band or VRP associated
               so set these appropriately */
            blob->band_information[vrp_band_index]->vrp_band_number = -1;
            blob->band_information[vrp_band_index]->blind_band_number = -1;
            blob->band_information[vrp_band_number]->
                process_blind_band_flag = 0;

            /* Get the number of lines in the VRP */
            status = ias_l0r_get_band_records_count( l0r_image,
                vrp_band_number, &size );
            if ( status != SUCCESS )
            {
                IAS_LOG_ERROR("Getting number of lines for band %d",
                    vrp_band_number );
                close_the_l0r_imagery();
                return ERROR;
            }
            blob->band_information[vrp_band_index]->number_lines = size;
            if (use_tirs_subinterval_frames_flag 
                && standard_atts->sensor_id == IAS_TIRS)
            {
                blob->band_information[vrp_band_index]->number_lines 
                    = tirs_target_scene_frames
                    * standard_atts->lines_per_frame;
                blob->band_information[vrp_band_index]->starting_line
                    = tirs_target_scene_start_frame
                    * standard_atts->lines_per_frame;
            }
            else if (use_oli_subinterval_frames_flag
                && standard_atts->sensor_id == IAS_OLI)
            {
                blob->band_information[vrp_band_index]->number_lines 
                    = oli_target_scene_frames
                    * standard_atts->lines_per_frame;
                blob->band_information[vrp_band_index]->starting_line
                    = oli_target_scene_start_frame
                    * standard_atts->lines_per_frame;
            }            
        }

        if ( blob->band_information[band_index]->process_band 
            && blind_band_number > 0 )
        {
            /* Check to ensure the blind band is present */
            int idx;
            int blind_band_present = 0;
            for ( idx = 0; idx < number_of_l0r_bands_present; idx++ )
            {
                if ( list_of_l0r_bands[idx] == blind_band_number )
                    blind_band_present = 1;
            }

            /* Set the standard blind band information */
            standard_atts 
                = ias_sat_attr_get_band_attributes( blind_band_number );
            if ( !standard_atts )
            {
                IAS_LOG_ERROR("Getting satellite attributes for band %d",
                    blind_band_number );
                return ERROR;
            }
            blind_band_index = standard_atts->band_index;

            /* If the blind band isn't present mark it as such and also
               set its associated band to not process it */
            if ( ! blind_band_present )
            {
                blob->band_information[band_index]->blind_band_number = -1;
                blob->band_information[band_index]->blind_band_index = -1;
                blob->band_information[band_index]->process_blind_band_flag = 0;
                process_blind_band[blind_band_index] = 0;
            }
            else
            {
                blob->band_information[blind_band_index]->band_number =
                    blind_band_number;
                blob->band_information[blind_band_index]->is_normal_band = 0;
                blob->band_information[band_index]->blind_band_index =
                    blind_band_index;

                /* We only want one band to process the blind band through
                   algorithms.  If this is the first encounter with this band
                   set the process flag. */
                if ( process_blind_band[blind_band_index] == 0 )
                {
                    blob->band_information[band_index]->blind_band_number =
                        blind_band_number;
                    blob->band_information[band_index]->
                        process_blind_band_flag = 1;
                    process_blind_band[blind_band_index] = 1;
                }

                blob->band_information[blind_band_index]->number_detectors =
                    standard_atts->detectors_per_sca;
                blob->band_information[blind_band_index]->number_scas =
                    standard_atts->scas;
                blob->band_information[blind_band_index]->band_classification =
                    standard_atts->band_classification;
                blob->band_information[blind_band_index]->spectral_type =
                    standard_atts->spectral_type;
                blob->band_information[blind_band_index]->one_sca_present = 0;
                blob->band_information[blind_band_index]->sca_number = 0;
            }

            /* Only the normal band has the blind band or VRP associated
               so set these appropriately */
            blob->band_information[blind_band_index]->vrp_band_number = -1;
            blob->band_information[blind_band_index]->blind_band_number = -1;
            blob->band_information[blind_band_index]->
                process_blind_band_flag = 0;
        }
    }  /* END for 0 to number_of_l0r_bands_present loop */

    /* Set the flag indicating the band information is set */
    blob->band_fields_initialized = 1;

    /* If we are processing OLI any OLI bands get the data truncation flag */
    for ( index = 0; index < blob->number_of_bands; index++ )
    {
        band_index = blob->list_of_band_indices[index];
        if ( blob->band_information[band_index]->sensor_id == IAS_OLI &&
             blob->band_information[band_index]->process_band )
        { 
            /* See if the image header is present, if not it's an error */
            status = ias_l0r_is_oli_image_header_present( l0r_image, &count );
            if ( status != SUCCESS || count <= 0 )
            {
               IAS_LOG_ERROR("There is no header record in L0R");
               close_the_l0r_imagery();
               return ERROR;
            }

            /* Read the image header */
            status = ias_l0r_get_oli_image_header( l0r_image,
                &l0r_image_header );
            if ( status != SUCCESS )
            {
                IAS_LOG_ERROR("Reading the L0R image header");
                close_the_l0r_imagery();
                return ERROR;
            }

            blob->image_data_truncation_setting = 
                l0r_image_header.image_data_truncation_setting;
            break;
        }
    }

    /* -------------------- Initialize the L1R image -------------------- */
    l1r_image = ias_l1r_open_image( blob->l1r_file_name, IAS_WRITE );
    if ( !l1r_image )
    {
        IAS_LOG_ERROR("Opening Level 1R file for output");
        close_the_l0r_imagery();
        return ERROR;
    }

    /* Set a few file metadata members from interval metadata. */
    memset( &blob->l1r_file_metadata, 0, sizeof(IAS_L1R_FILE_METADATA));
    strcpy( blob->l1r_file_metadata.software_version,
        ias_misc_get_software_version());

    /* Add information from the L0R file */
    strcpy( blob->l1r_file_metadata.ingest_software_version,
        l0r_interval_metadata.is_version );
    strcpy( blob->l1r_file_metadata.collect_type,
        l0r_interval_metadata.collection_type );

    /* Get the earth_sun_distance and set it in the blob->l1r_file_metadata*/
    status = get_earth_sun_distance(blob->date_acquired, 
             &blob->l1r_file_metadata.earth_sun_distance);    
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Getting the earth_sun_distance");
        close_the_l0r_imagery();
        return ERROR;
    }
    
    /* Open the L0R ancillary data */
    status = ias_l0r_open_ancillary(l0r_image, IAS_READ);
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Opening the L0R ancillary data");
        close_the_l0r_imagery();
        return ERROR;
    }

    /* Get the OLI and TIRS frame rates and set them in the blob */
    status = get_oli_tirs_frame_rates(blob);
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Getting the OLI/TIRS frame rates");
        close_the_l0r_imagery();
        return ERROR;
    }

    /* Check if the ghost image will be used during processing */
    if (blob->use_tirs_ghost_image_flag)
    {
        status = open_tirs_ghost_image_bands(blob);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Opening the TIRS ghost image I/O bands");
            close_the_l0r_imagery();
            return ERROR;
        }
    }

    /* Allocate the output band */
    l1r_band = calloc( number_of_bands, sizeof(L1R_BAND_IO *));
    if ( !l1r_band )
    {
        IAS_LOG_ERROR("Allocating memory for output file pointers");
        close_the_l0r_imagery();
        return ERROR;
    }
    
    return SUCCESS;
}  /* END initialize_io_routines */

/*------------------------------------------------------------------------
NAME:    open_l1r_bands

PURPOSE: This routine opens the L1R output bands.  It uses the band
         attributes in case the number of output lines has been changed
         due to algorithm pre-process routines or other parameter changes.

RETURNS: SUCCESS or ERROR
------------------------------------------------------------------------*/
int open_l1r_bands
(
    BLOB *blob  /* I: list of bands, indices, number of bands, band
                      information */
)
{
    int index;
    IAS_DATA_TYPE data_type = IAS_R4;   /* Data type of the output image */
    
    for ( index = 0; index < blob->number_of_bands; index++ )
    {
        int lines;
        int samples;
        int number_of_scas;
        int band_index;
        int band_number;

        /* Get the band number and index */
        band_index = blob->list_of_band_indices[index];
        band_number = blob->list_of_bands[index];

        /* Only open normal bands */
        if ( blob->band_information[band_index]->is_normal_band == 0 )
        {
            l1r_band[band_index] = NULL;
            continue;
        }
        /* Or are in the list to process */
        else if ( blob->band_information[band_index]->process_band == 0 )
        {
            l1r_band[band_index] = NULL;
            continue;
        }

        /* Since we are writing to a file the returned values aren't set
           use temporary variables. */
        lines = blob->band_information[band_index]->number_lines;
        samples = blob->band_information[band_index]->number_detectors;
        number_of_scas = blob->band_information[band_index]->number_scas;
        
        l1r_band[band_index] = ias_l1r_open_band( l1r_image, band_number,
            &data_type, &number_of_scas, &lines, &samples );
        if ( !l1r_band[band_index] )
        {
            IAS_LOG_ERROR("Couldn't open L1R band %d", band_number );
            ias_l1r_close_image( l1r_image );
            l1r_image = NULL;
            close_the_l0r_imagery();
            return ERROR;
        }
    }

    return SUCCESS;
}  /* END open_l1r_bands */



/*------------------------------------------------------------------------
NAME:    cleanup_io_routines

PURPOSE: This routine sets the file and band metadata in the L1R file;
         closes the l1r bands and closes both the l0r and l1r files.  It
         also frees the OLI and/or TIRS L0R frame header data buffers.
         This routine isn't called in a threaded environment so the HDF
         library mutex isn't locked.

RETURNS: SUCCESS or ERROR
------------------------------------------------------------------------*/
int cleanup_io_routines
(
    BLOB *blob    /* I/O: Contains number of bands */
)
{
    int status;         /* Routine return status */
    int band_index;     /* Band loop index */
    int error = 0;      /* Flag to indicate an error occurred closing an
                           image */

    /* Close the open bands */
    for ( band_index = 0; band_index < number_of_bands; band_index++ )
    {
        /* Make sure the band is still open before attempting to close it. */
        if ( l1r_band[band_index] != NULL )
        {
            status = ias_l1r_close_band( l1r_band[band_index] );
            l1r_band[band_index] = NULL;
            if ( status != SUCCESS )
            {
                IAS_LOG_ERROR("Closing L1R band %d", 
                    get_band_attributes( band_index )->band_number );
                error = 1;
            }
        }
    }

    /* Close both input and output files,  Before closing the file make sure
       there is something to close.  Trying to close a NULL pointer is bad.
       Then set the pointers to NULL so the check will actually be
       meaningful. */
    if ( l1r_image == NULL )
    {
        IAS_LOG_WARNING("Attempting to close an L1R file that is not open");
    }
    else
    {
        status = ias_l1r_close_image( l1r_image );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Closing L1R image");
            error = 1;
        }
        l1r_image = NULL;
    }
    if ( l0r_image == NULL )
    {
        IAS_LOG_WARNING("Attempting to close an L0R file that is not open");
    }
    else
    {
        status = close_the_l0r_imagery();
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Closing L0R imagery");
            error = 1;
        }
    }

    free( l1r_band );

    /* Free up the base L0R frame header data buffers */
    free(oli_fh);
    oli_fh = NULL;
    free(tirs_fh);
    tirs_fh = NULL;

    if (blob->use_tirs_ghost_image_flag)
    {
        /* Free up the tirs stray light data */
        for (band_index = 0; band_index < IAS_MAX_NBANDS; band_index++)
        {
            RPS_BAND_ATTRIBUTES *band_attributes 
                = get_band_attributes(band_index);

            /* Check the band io pointers */
            if (blob->tirs_stray_light_band_io[band_index])
            {
                if (ias_l1r_close_band(
                    blob->tirs_stray_light_band_io[band_index]) != SUCCESS)
                {
                    IAS_LOG_ERROR("Closing stray light correction band "
                        "number %d", band_attributes->band_number);
                    error = 1;
                } 
                blob->tirs_stray_light_band_io[band_index] = NULL;
            }
        }

        /* Free up the L1R IO pointer */
        if (blob->tirs_stray_light_l1r_io)
        {
            if (ias_l1r_close_image(blob->tirs_stray_light_l1r_io) != SUCCESS)
            {
                IAS_LOG_ERROR("Closing stray light correction image");
                error = 1;
            } 
            blob->tirs_stray_light_l1r_io = NULL;
        }
    }

    /* Return the appropriate status */
    if ( error )
        return ERROR;

    return SUCCESS;
}  /* END cleanup_io_routines */


/*------------------------------------------------------------------------
NAME:    set_l1r_metadata

PURPOSE: Set the L1R metadata into the hd5 output file.

RETURNS: SUCCESS or ERROR
------------------------------------------------------------------------*/
int set_l1r_metadata
(
    const BLOB *blob,   /* I: Contains file metadata structure */
    L1RIO *out_image    /* I: Image to set the metadata to.  If NULL, 
                              default to l1r_image */
)
{
    int status;         /* Routine return status */
    int index;          /* Loop index */
    int error = 0;      /* Flag to indicate an error occurred closing an
                           image */

    /* If the input image is not set, set it to the l1r_image by default */
    if (!out_image)
        out_image = l1r_image;

    if ( IAS_THREAD_LOCK_MUTEX( &hdf_io_mutex ))
    {
        IAS_LOG_ERROR("Locking HDF library mutex");
        return ERROR;
    }

    /* Set the file metadata */
    status = ias_l1r_set_file_metadata( out_image, &blob->l1r_file_metadata );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Setting L1R file metadata");
        error = 1;
    }

    /* Set the band metadata */
    for ( index = 0; index < blob->number_of_bands; index++ )
    {
        int band_number = blob->list_of_bands[index];
        int band_index = blob->list_of_band_indices[index];
        status = ias_l1r_set_band_metadata( out_image,
            blob->l1r_band_metadata[band_index], 1 );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Setting L1R band metadata for band %d",
                band_number );
            error = 1;
        }
    }

    if ( IAS_THREAD_UNLOCK_MUTEX( &hdf_io_mutex ))
    {
        IAS_LOG_ERROR("Unlocking HDF library mutex");
        return ERROR;
    }

    if ( error )
        return ERROR;

    return SUCCESS;
}  /* END set_l1r_metadata */


/*------------------------------------------------------------------------
NAME:    read_l0r_image_data

PURPOSE: Read an SCA of data from the L0R input file.  The return buffer
         is allocated here.  The memory management routines will free it
         when no longer needed.

RETURNS: SUCCESS or ERROR
------------------------------------------------------------------------*/
uint16_t *read_l0r_image_data
(
    int band_index,    /* I: Band to retrieve */
    int sca_index      /* I: SCA to retrieve */
)
{
    int status;  /* Routine return status */
    uint16_t *l0r_data = NULL;  /* Pointer to input data buffer */
    RPS_BAND_ATTRIBUTES *band_attributes = get_band_attributes( band_index );
    int band_number = band_attributes->band_number;
    int starting_line = band_attributes->starting_line;
    int number_lines = band_attributes->number_lines;
    int number_detectors = band_attributes->number_detectors;

    /* Allocate the memory */
    l0r_data = malloc( sizeof(uint16_t) * number_lines * number_detectors );
    if ( !l0r_data )
    {
        IAS_LOG_DEBUG("Allocating memory for L0R data");
        return NULL;
    }

    if ( IAS_THREAD_LOCK_MUTEX( &hdf_io_mutex ))
    {
        IAS_LOG_ERROR("Locking HDF library mutex");
        return NULL;
    }

    IAS_LOG_DEBUG("Read L0R band %d SCA %d start", band_number,
        sca_index + 1);
    status = ias_l0r_get_band_lines_sca( l0r_image, band_number, sca_index,
        starting_line, 0, number_lines, number_detectors, l0r_data );
    if ( status != SUCCESS )
    {
        IAS_LOG_DEBUG("Getting L0R image data for band %d SCA %d",
            band_number, sca_index + 1 );
        free( l0r_data );
        l0r_data = NULL;
    }

    if ( IAS_THREAD_UNLOCK_MUTEX( &hdf_io_mutex ))
    {
        IAS_LOG_ERROR("Unlocking HDF library mutex");
        return NULL;
    }
    IAS_LOG_DEBUG("Read L0R band %d SCA %d done", band_number,
        sca_index + 1 );

    /* Tell the memory management that the buffer is now valid */
    set_buffer_filled_flag( IMAGE_L0R, band_index, sca_index );

    return l0r_data;
}  /* END read_l0r_image_data */

/*------------------------------------------------------------------------
NAME:     read_l0r_frame_headers

PURPOSE:  Reads OLI/TIRS L0R frame header information from the
          ancillary data file.  Note: this is called from the memory
          management routines via an increment consumer count and therefore
          won't be called in a threaded environment hence no mutex lock.

RETURNS:  Pointer to data buffer if successful; NULL pointer if there's
          an error
--------------------------------------------------------------------------*/
void *read_l0r_frame_headers
(
    MEMORY_BUFFER_TYPE buffer_type,           /* I: Memory buffer type */
    int *num_headers,                         /* O: Actual number of L0R
                                                 frame headers */
    size_t *buffer_size                       /* O: Actual buffer size */
)
{
    void *l0r_frame_headers = NULL;
    int status;


    /* If the L0R image data buffer hasn't been initialized yet, that's an
       error */
    if (l0r_image == NULL)
    {
        IAS_LOG_ERROR("L0R image data not initialized/read yet");
        return NULL;
    }

    /* For either buffer type, check to see if the corresponding frame
       header buffer has been allocated and populated.  If it has,
       we're set--get it, the number of frame headers, recompute the
       required buffer size, and return.  Otherwise, determine the number of
       frame headers, calculate the required buffer size, allocate the buffer
       and populate it, then return */
    if (buffer_type == MM_L0R_OLI_FRAME_HEADER)
    {
        /* If the base pointer has already been allocated, return with
           it */
        if (oli_fh != NULL)
        {
            *num_headers = num_oli_frame_headers;
            *buffer_size =
                num_oli_frame_headers * sizeof(IAS_L0R_OLI_FRAME_HEADER);
            return (void *)oli_fh;
        }

        /* Get the actual number of OLI frame headers */
        status = ias_l0r_get_oli_frame_headers_records_count(l0r_image,
            &num_oli_frame_headers);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Obtaining number of OLI frame headers");
            return NULL;
        }
        *num_headers = num_oli_frame_headers;

        /* Allocate the frame header data buffer based on the actual number
           of frame headers */
        *buffer_size =
            num_oli_frame_headers * sizeof(IAS_L0R_OLI_FRAME_HEADER);
        oli_fh = malloc(*buffer_size);
        if (oli_fh == NULL)
        {
            IAS_LOG_ERROR("Allocating OLI frame header data buffer");
            return NULL;
        }

        /* Read the OLI frame header data */
        status = ias_l0r_get_oli_frame_headers(l0r_image, 0,
            num_oli_frame_headers, oli_fh);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading OLI frame header data");
            return NULL;
        }
        l0r_frame_headers = (void *)oli_fh;
    }   /* END -- OLI frame header block */
    else
    {
        if (tirs_fh != NULL)
        {
            *num_headers = num_tirs_frame_headers;
            *buffer_size =
                num_tirs_frame_headers * sizeof(IAS_L0R_TIRS_FRAME_HEADER);
            return (void *)tirs_fh;
        }

        /* Get the actual number of TIRS frame headers */
        status = ias_l0r_get_tirs_frame_headers_records_count(l0r_image,
            &num_tirs_frame_headers);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Obtaining number of TIRS frame headers");
            return NULL;
        }
        *num_headers = num_tirs_frame_headers;

        /* Allocate the frame header data buffer based on the actual number
           of frame headers */
        *buffer_size =
            num_tirs_frame_headers * sizeof(IAS_L0R_TIRS_FRAME_HEADER);
        tirs_fh = malloc(*buffer_size);
        if (tirs_fh == NULL)
        {
            IAS_LOG_ERROR("Allocating OLI frame header data buffer");
            return NULL;
        }

        /* Read the TIRS frame header data */
        status = ias_l0r_get_tirs_frame_headers(l0r_image, 0,
            num_tirs_frame_headers, tirs_fh);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Reading TIRS frame header data");
            return NULL;
        }
        l0r_frame_headers = (void *)tirs_fh;
    }   /* END -- TIRS frame header block */

    return l0r_frame_headers;
}   /* END read_l0r_frame_headers */


/*------------------------------------------------------------------------
NAME:    read_oli_focal_plane_temperatures

PURPOSE: Read the OLI focal plane temperatures

RETURNS: SUCCESS or ERROR
------------------------------------------------------------------------*/
int read_oli_focal_plane_temperatures
(
    float **temperature7,      /* O: temperature 7 */
    float **temperature14,     /* O: Temperature 14 */
    int *number_temps          /* O: Number of temperatures */
)
{
    int status;                 /* routine return status */
    int count_of_ancillary_data;/* Number of ancillary telemetry records */
    int num_of_temps;           /* local number of temperatures for the
                                   scene */
    int i;                      /* array index */
    int rec;                    /* loop counter */
    int start_rec;              /* starting record number for the scene */
    int end_rec;                /* ending record number for the scene */
    IAS_L0R_OLI_TELEMETRY *telemetry = NULL; /* the ancillary telemetry data */
    float *local_temp7;
    float *local_temp14;
    double scene_start_time_j2000 = get_blob_oli_start_time();
    double scene_stop_time_j2000 = get_blob_oli_stop_time();
    double record_time_j2000;

    if ( IAS_THREAD_LOCK_MUTEX( &hdf_io_mutex ))
    {
        IAS_LOG_ERROR("Locking HDF library mutex");
        return ERROR;
    }

    /* Find out how many records there are in the telemetry data */
    status = ias_l0r_get_ancillary_oli_telemetry_records_count( l0r_image,
        &count_of_ancillary_data );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Getting size of ancillary data");
        if (IAS_THREAD_UNLOCK_MUTEX(&hdf_io_mutex))
            IAS_LOG_ERROR("Unlocking HDF library mutex");
        return ERROR;
    }

    if ( count_of_ancillary_data == 0 )
    {
        IAS_LOG_ERROR("There are no ancillary records");
        if (IAS_THREAD_UNLOCK_MUTEX(&hdf_io_mutex))
            IAS_LOG_ERROR("Unlocking HDF library mutex");
        return ERROR;
    }

    /* Allocate a buffer to contain all the telemetry records */
    telemetry = malloc(count_of_ancillary_data * sizeof(IAS_L0R_OLI_TELEMETRY));
    if (telemetry == NULL)
    {
        IAS_LOG_ERROR("Unable to allocate memory for OLI ancillary data");
        if (IAS_THREAD_UNLOCK_MUTEX(&hdf_io_mutex))
            IAS_LOG_ERROR("Unlocking HDF library mutex");
        return ERROR;
    }

    /* For simplicity, grab all the telemetry records. */
    status = ias_l0r_get_ancillary_oli_telemetry( l0r_image, 0,
        count_of_ancillary_data, telemetry );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Cannot obtain OLI ancillary telemetry data");
        if (IAS_THREAD_UNLOCK_MUTEX(&hdf_io_mutex))
            IAS_LOG_ERROR("Unlocking HDF library mutex");
        free(telemetry);
        return ERROR;
    }

    if ( IAS_THREAD_UNLOCK_MUTEX( &hdf_io_mutex ))
    {
        IAS_LOG_ERROR("Unlocking HDF library mutex");
        free(telemetry);
        return ERROR;
    }

    /* Loop through the telemetry records to find the telemetry points that
       bracket the image start and end frame times.  The logic that follows
       ensures one telemetry point is obtained even if the telemetry doesn't
       overlap any image frame time.  This can easily happen in very short
       collects. */
    start_rec = end_rec = 0;
    for ( i = 0; i < count_of_ancillary_data; i++ )
    {
        record_time_j2000 = IAS_L0R_CONVERT_TIME_TO_SECONDS_SINCE_J2000(
                telemetry[i].oli_telemetry_header_3.l0r_time);

        /* If the time is greater than the scene stop time use this record
           as the stop time.  This ensures at least one telemetry point
           after the last image frame. */
        if (record_time_j2000 > scene_stop_time_j2000)
        {
            end_rec = i;
            break; /* There is no need to search any further. */
        }

        /* Update the start record index as long as we haven't found a
           time after the scene start time.  This ensures at least one
           telemetry point before the first image frame. */
        if ( record_time_j2000 < scene_start_time_j2000 )
            start_rec = i;

        /* Still haven't found a time greater than the scene stop time so
           update the end record index */
        end_rec = i;
    }

    /* Determine how many temperature records there are and allocate memory
       for them. */
    num_of_temps = end_rec - start_rec + 1;
    local_temp7 = malloc( num_of_temps * sizeof(float));
    local_temp14 = malloc( num_of_temps * sizeof(float));
    if ( !local_temp7 || !local_temp14 )
    {
        IAS_LOG_ERROR("Allocating OLI temperature arrays");
        free( local_temp7 );
        free( local_temp14 );
        free(telemetry);
        local_temp7 = NULL;
        local_temp14 = NULL;
        *number_temps = 0;
        return ERROR;
    }

    /* Get the ancillary records for the scene and extract the two
       temperatures */
    for ( rec = start_rec, i = 0; rec <= end_rec; rec++, i++ )
    {
        local_temp7[i] =
            telemetry[rec].oli_telemetry_header_3.fpm_7_temp_celsius;
        local_temp14[i] =
            telemetry[rec].oli_telemetry_header_3.fpm_14_temp_celsius;
    }
    free(telemetry);

    *number_temps = num_of_temps;
    *temperature7 = local_temp7;
    *temperature14 = local_temp14;

    return SUCCESS;
}  /* END read_oli_focal_plane_temperatures */


/*------------------------------------------------------------------------
NAME:    read_tirs_focal_plane_temperatures

PURPOSE: Read the TIRS focal plane temperatures

RETURNS: SUCCESS or ERROR
------------------------------------------------------------------------*/
int read_tirs_focal_plane_temperatures
(
    float **f2_temperature,    /* O: fp_f2_fine_sensor_1_celsius temperature */
    float **f4_temperature,    /* O: fp_f4_fine_sensor_3_celsius temperature */
    float **f6_temperature,    /* O: fp_f6_fine_sensor_1_celsius temperature */
    float **f7_temperature,    /* O: fp_f7_fine_sensor_2_celsius temperature */
    int *number_temps          /* O: Number of temperatures */
)
{
    int status;                 /* routine return status */
    int count_of_ancillary_data;/* Number of ancillary telemetry records */
    int num_of_temps;           /* local number of temperatures for the
                                   scene */
    int i;                      /* array index */
    int rec;                    /* loop counter */
    int start_rec;              /* starting record number for the scene */
    int end_rec;                /* ending record number for the scene */
    IAS_L0R_TIRS_TELEMETRY *telemetry = NULL; 
                                /* TIRS ancillary telemetry data */
    float *local_f2_temp;
    float *local_f4_temp;
    float *local_f6_temp;
    float *local_f7_temp;
    double scene_start_time_j2000 = get_blob_tirs_start_time();
    double scene_stop_time_j2000 = get_blob_tirs_stop_time();
    double record_time_j2000;

    if ( IAS_THREAD_LOCK_MUTEX( &hdf_io_mutex ))
    {
        IAS_LOG_ERROR("Locking HDF library mutex");
        return ERROR;
    }

    /* Find out how many records there are in the TIRS telemetry data */
    status = ias_l0r_get_ancillary_tirs_telemetry_records_count( l0r_image,
        &count_of_ancillary_data );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Getting size of ancillary data");
        if (IAS_THREAD_UNLOCK_MUTEX(&hdf_io_mutex))
            IAS_LOG_ERROR("Unlocking HDF library mutex");
        return ERROR;
    }

    if ( count_of_ancillary_data == 0 )
    {
        IAS_LOG_ERROR("There are no ancillary records");
        if (IAS_THREAD_UNLOCK_MUTEX(&hdf_io_mutex))
            IAS_LOG_ERROR("Unlocking HDF library mutex");
        return ERROR;
    }

    /* Allocate a buffer to contain all the telemetry records */
    telemetry = malloc(count_of_ancillary_data * 
                       sizeof(IAS_L0R_TIRS_TELEMETRY));
    if (telemetry == NULL)
    {
        IAS_LOG_ERROR("Unable to allocate memory for TIRS ancillary data");
        if (IAS_THREAD_UNLOCK_MUTEX(&hdf_io_mutex))
            IAS_LOG_ERROR("Unlocking HDF library mutex");
        return ERROR;
    }

    /* For simplicity, grab all the telemetry records. */
    status = ias_l0r_get_ancillary_tirs_telemetry( l0r_image, 0,
        count_of_ancillary_data, telemetry );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Cannot obtain TIRS ancillary telemetry data");
        if (IAS_THREAD_UNLOCK_MUTEX(&hdf_io_mutex))
            IAS_LOG_ERROR("Unlocking HDF library mutex");
        free(telemetry);
        return ERROR;
    }

    if ( IAS_THREAD_UNLOCK_MUTEX( &hdf_io_mutex ))
    {
        IAS_LOG_ERROR("Unlocking HDF library mutex");
        free(telemetry);
        return ERROR;
    }

    /* Loop through the telemetry records to find the telemetry points that
       bracket the image start and end frame times.  The logic that follows
       ensures one telemetry point is obtained even if the telemetry doesn't
       overlap any image frame time.  This can easily happen in very short
       collects. */
    start_rec = end_rec = 0;
    for ( i = 0; i < count_of_ancillary_data; i++ )
    {
        record_time_j2000 = IAS_L0R_CONVERT_TIME_TO_SECONDS_SINCE_J2000(
                telemetry[i].l0r_time);

        /* If the time is greater than the scene stop time use this record
           as the stop time.  This ensures at least one telemetry point
           after the last image frame. */
        if (record_time_j2000 > scene_stop_time_j2000)
        {
            end_rec = i;
            break; /* There is no need to search any further. */
        }

        /* Update the start record index as long as we haven't found a
           time after the scene start time.  This ensures at least one
           telemetry point before the first image frame. */
        if ( record_time_j2000 < scene_start_time_j2000 )
            start_rec = i;

        /* Still haven't found a time greater than the scene stop time so
           update the end record index */
        end_rec = i;
    }

    /* Determine how many temperature records there are and allocate memory
       for them. */
    num_of_temps = end_rec - start_rec + 1;
    local_f2_temp = malloc( num_of_temps * sizeof(float));
    local_f4_temp = malloc( num_of_temps * sizeof(float));
    local_f6_temp = malloc( num_of_temps * sizeof(float));
    local_f7_temp = malloc( num_of_temps * sizeof(float));
    if ( !local_f2_temp || !local_f4_temp || !local_f6_temp || !local_f7_temp  )
    {
        IAS_LOG_ERROR("Allocating TIRS temperature arrays");
        free( local_f2_temp );
        free( local_f4_temp );
        free( local_f6_temp );
        free( local_f7_temp );
        free(telemetry);
        local_f2_temp = NULL;
        local_f4_temp = NULL;
        local_f6_temp = NULL;
        local_f7_temp = NULL;
        *number_temps = 0;
        return ERROR;
    }

    /* Get the ancillary records for the scene and extract the two
       temperatures */
    for ( rec = start_rec, i = 0; rec <= end_rec; rec++, i++ )
    {
        local_f2_temp[i] =
            telemetry[rec].temperature.fp_f2_fine_sensor_1_celsius;
        local_f4_temp[i] =
            telemetry[rec].temperature.fp_f4_fine_sensor_3_celsius;
        local_f6_temp[i] =
            telemetry[rec].temperature.fp_f6_fine_sensor_1_celsius;
        local_f7_temp[i] =
            telemetry[rec].temperature.fp_f7_fine_sensor_2_celsius;
    }
    free(telemetry);

    *number_temps = num_of_temps;
    *f2_temperature = local_f2_temp;
    *f4_temperature = local_f4_temp;
    *f6_temperature = local_f6_temp;
    *f7_temperature = local_f7_temp;

    return SUCCESS;
}  /* END read_tirs_focal_plane_temperatures */


/*------------------------------------------------------------------------
NAME:    read_post_warmup_focal_plane_temperatures

PURPOSE: Read the post warmup focal plane temperatures

RETURNS: SUCCESS or ERROR
------------------------------------------------------------------------*/
int read_post_warmup_focal_plane_temperatures
(
    float **temperature7,      /* O: temperature 7 */
    float **temperature14,     /* O: Temperature 14 */
    int *number_temps          /* O: Number of temperatures */
)
{
    int status;                 /* routine return status */
    int start_post_warmup_telemetry_rec;              
                                /* starting post warmup telemetry number */
    int end_post_warmup_telemetry_rec;                
                                /* ending post warmup telemetry number */
    int i;                      /* array index */
    int num_of_temps;           /* local number of temperatures for the
                                   scene */
    IAS_L0R_OLI_TELEMETRY *telemetry;  /* the ancillary telemetry data */
    int post_warmup_start_frame;       /* Image data post warmup start line
                                            number */
    int post_warmup_nframes;    /* Lamp warmup time period */
    float warmup_time_period;   /* Lamp warmup time period */
    float lamp_on_time;                /* Lamp turn on time */
    int warmup_start_frame;            /* Lamp warmup start frame number */
    int warmup_stop_frame;             /* Lamp warmup stop frame number */
    float post_warmup_start_time;      /* Image post warmup start time */
    float post_warmup_stop_time;       /* Image post warmup stop time */
    float *local_temp7;
    float *local_temp14;


    if ( IAS_THREAD_LOCK_MUTEX( &hdf_io_mutex ))
    {
        IAS_LOG_ERROR("Locking HDF library mutex");
        return ERROR;
    }

    /* Get the post_warmup_nframes and warmup_time_period from the blob */
    warmup_time_period = get_blob_lamp_warmup_time_period();
    post_warmup_nframes = get_blob_post_warmup_nframes();

    /* Get the telemetry data starting and ending record numbers */
    status = get_post_warmup_start_end_line_numbers(post_warmup_nframes, 
             warmup_time_period, &start_post_warmup_telemetry_rec, 
             &end_post_warmup_telemetry_rec, &post_warmup_start_frame, 
             &warmup_start_frame, &warmup_stop_frame, &lamp_on_time, 
             &post_warmup_start_time, &post_warmup_stop_time);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Getting scene starting and ending record numbers"); 
        if (IAS_THREAD_UNLOCK_MUTEX(&hdf_io_mutex))
            IAS_LOG_ERROR("Unlocking HDF library mutex");
        return ERROR;
    }

    /* Set the *temperature7 & *temperature14 be NULL for error check */
    *temperature7 = NULL;
    *temperature14 = NULL;

    /* Determine how many post warmup temperature records there are */
    num_of_temps = end_post_warmup_telemetry_rec - 
                   start_post_warmup_telemetry_rec + 1;


    /* Allocate memory to contain all the post warmup telemetry records */
    telemetry = malloc(num_of_temps * sizeof(IAS_L0R_OLI_TELEMETRY));
    if (telemetry == NULL)
    {
        IAS_LOG_ERROR("Unable to allocate memory for OLI ancillary data");
        if (IAS_THREAD_UNLOCK_MUTEX(&hdf_io_mutex))
            IAS_LOG_ERROR("Unlocking HDF library mutex");
        return ERROR;
    }

    /* For simplicity, grab all the post warmup telemetry records. */
    status = ias_l0r_get_ancillary_oli_telemetry( l0r_image, 
             start_post_warmup_telemetry_rec, num_of_temps, telemetry );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Cannot obtain OLI ancillary telemetry data");
        if (IAS_THREAD_UNLOCK_MUTEX(&hdf_io_mutex))
            IAS_LOG_ERROR("Unlocking HDF library mutex");
        free(telemetry);
        return ERROR;
    }

    if ( IAS_THREAD_UNLOCK_MUTEX( &hdf_io_mutex ))
    {
        IAS_LOG_ERROR("Unlocking HDF library mutex");
        free(telemetry);
        return ERROR;
    }

    /* Allocate memory for the local post warmup temperatures */
    local_temp7 = malloc( num_of_temps * sizeof(float));
    local_temp14 = malloc( num_of_temps * sizeof(float));
    if ( !local_temp7 || !local_temp14 )
    {
        free( local_temp7 );
        free( local_temp14 );
        local_temp7 = NULL;
        local_temp14 = NULL;
        *number_temps = 0;
        IAS_LOG_ERROR( "Allocating temperature arrays" );
        return ERROR;
    }

    /* Get the ancillary records for the scene and extract the two
       temperatures */
    for ( i = 0; i < num_of_temps; i++ )
    {
        local_temp7[i] = 
            telemetry[i].oli_telemetry_header_3.fpm_7_temp_celsius;
        local_temp14[i] = 
            telemetry[i].oli_telemetry_header_3.fpm_14_temp_celsius;
    }

    *number_temps = num_of_temps;
    *temperature7 = local_temp7;
    *temperature14 = local_temp14;

    /* Clearup */
    free(telemetry);

    return SUCCESS;
}  /* END read_post_warmup_focal_plane_temperatures */


/*------------------------------------------------------------------------
NAME:    read_lamp_response

PURPOSE: Read the lamp response data and set the lamp response data in 
         the blob.  Note: This routine assume ias_l0r_open and
         ias_l0r_open_ancillary routines have already been executed 

RETURNS: SUCCESS or ERROR
------------------------------------------------------------------------*/
int read_lamp_response
(
    BLOB *blob                      /* I: Get post warmup lines */
                                    /* O: Set lamp response data */
)
{
    int status;                 /* routine return status */
    int start_post_warmup_telemetry_rec;              
                                /* starting post warmup telemetry number */
    int end_post_warmup_telemetry_rec;                
                                /* ending post warmup telemetry number */
    int post_warmup_start_frame;         /* Image data post warmup start line
                                            number */
    int post_warmup_nframes;    /* Lamp warmup time period */
    float warmup_time_period;   /* Lamp warmup time period */
    float lamp_on_time;                /* Lamp turn on time */
    int warmup_start_frame;            /* Lamp warmup start frame number */
    int warmup_stop_frame;             /* Lamp warmup stop frame number */
    float post_warmup_start_time;      /* Image post warmup start time */
    float post_warmup_stop_time;       /* Image post warmup stop time */

    if ( IAS_THREAD_LOCK_MUTEX( &hdf_io_mutex ))
    {
        IAS_LOG_ERROR("Locking HDF library mutex");
        return ERROR;
    }

    /* Get the post_warmup_nframes and warmup_time_period from the blob */
    warmup_time_period = get_blob_lamp_warmup_time_period();
    post_warmup_nframes = get_blob_post_warmup_nframes();

    /* Get some of the lamp response data */
    status = get_post_warmup_start_end_line_numbers(post_warmup_nframes, 
             warmup_time_period, &start_post_warmup_telemetry_rec, 
             &end_post_warmup_telemetry_rec, &post_warmup_start_frame, 
             &warmup_start_frame, &warmup_stop_frame, &lamp_on_time, 
             &post_warmup_start_time, &post_warmup_stop_time);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Getting scene starting and ending record numbers"); 
        if (IAS_THREAD_UNLOCK_MUTEX( &hdf_io_mutex ))
            IAS_LOG_ERROR("Unlocking HDF library mutex");
        return ERROR;
    }

    /* Populate some lamp response data into the blob */
    blob->lamp_on_time = lamp_on_time;
    blob->lamp_warmup_time = warmup_time_period;
    blob->post_warmup_start_time = post_warmup_start_time;
    blob->post_warmup_stop_time = post_warmup_stop_time;
    blob->post_warmup_start_frame = post_warmup_start_frame;
    blob->post_warmup_stop_frame =
        post_warmup_start_frame + post_warmup_nframes;
    blob->lamp_warmup_start_frame = warmup_start_frame;
    blob->lamp_warmup_stop_frame = warmup_stop_frame;

    /* Release the mutex -- we don't need it anymore. */
    if (IAS_THREAD_UNLOCK_MUTEX(&hdf_io_mutex))
    {
        IAS_LOG_ERROR("Unlocking HDF library mutex");
        return ERROR;
    }

    return SUCCESS;

    /* The L0R image file will be closed somewhere else */
}  /* END read_lamp_response */


/*------------------------------------------------------------------------
NAME:    read_post_warmup_lamp_telemetry

PURPOSE: Read the post warmup lamp telemetry data 

         Note: This routine assume ias_l0r_open and ias_l0r_open_ancillary 
               routines have already been executed 
         Note: This routine currently assume the photodiode_micro_amp and 
               monitor_diode_temp are sampled at the same rate because
               they are currently in the different structure 
               (IAS_L0R_OLI_TELEMETRY and IAS_L0R_OLI_TIRS_TEMPERATURE).
               If it is not the case, the num_recs has to be calculated
               for monitor_diode_temp based on L0r_time in 
               IAS_L0R_OLI_TIRS_TEMPERATURE structure and the
               monitor_diode_temp has to be read out in a separate loop and
               one new output variable int *number_diode_temp has to be
               added for the read_post_warmup_lamp_telemetry routine.

RETURNS: SUCCESS or ERROR
------------------------------------------------------------------------*/
int read_post_warmup_lamp_telemetry
(
    float **lamp_current_a,         /* O: Group A lamp current values */
    float **lamp_voltage_a,         /* O: Group A lamp voltage values */
    float **photodiode_micro_amp_a, /* O: Group A lamp photodiode current */
    float **lamp_housing_temp_a,    /* O: Group A lamp housing temperatures */
    float **monitor_diode_temp_a,   /* O: Group A lamp monitor diode 
                                          temperatures */
    float **lamp_current_b,         /* O: Group B lamp current values */
    float **lamp_voltage_b,         /* O: Group B lamp voltage values */
    float **photodiode_micro_amp_b, /* O: Group B lamp photodiode current */
    float **lamp_housing_temp_b,    /* O: Group B lamp housing temperatures */
    float **monitor_diode_temp_b,   /* O: Group B lamp monitor diode 
                                          temperatures */
    int *number_lamp_telemetry      /* O: Number of lamp telemetry data */
)
{
    int status;                 /* routine return status */
    int start_post_warmup_telemetry_rec;              
                                /* starting post warmup telemetry number */
    int end_post_warmup_telemetry_rec;                
                                /* ending post warmup telemetry number */
    int i;                      /* array index */
    int num_of_telemetry;       /* local number of lamp telemetry for the
                                   scene */
    IAS_L0R_OLI_TELEMETRY *telemetry;  /* the ancillary telemetry data */
    IAS_L0R_OLI_TIRS_TEMPERATURE *oli_tirs_temperatures;
    int post_warmup_start_frame; /* Image data post warmup start frame
                                    number */
    int post_warmup_nframes;     /* Lamp warmup time period */
    float warmup_time_period;    /* Lamp warmup time period */
    float lamp_on_time;                /* Lamp turn on time */
    int warmup_start_frame;            /* Lamp warmup start frame number */
    int warmup_stop_frame;             /* Lamp warmup stop frame number */
    float post_warmup_start_time;      /* Image post warmup start time */
    float post_warmup_stop_time;       /* Image post warmup stop time */
    float *local_lamp_current_a;
    float *local_lamp_voltage_a;
    float *local_photodiode_micro_amp_a;
    float *local_lamp_housing_temp_a;
    float *local_monitor_diode_temp_a;
    float *local_lamp_current_b;
    float *local_lamp_voltage_b;
    float *local_photodiode_micro_amp_b;
    float *local_lamp_housing_temp_b;
    float *local_monitor_diode_temp_b;

    if ( IAS_THREAD_LOCK_MUTEX( &hdf_io_mutex ))
    {
        IAS_LOG_ERROR("Locking HDF library mutex");
        return ERROR;
    }

    /* Get the post_warmup_nframes and warmup_time_period from the blob */
    warmup_time_period = get_blob_lamp_warmup_time_period();
    post_warmup_nframes = get_blob_post_warmup_nframes();

    /* Get the telemetry data starting and ending record numbers */
    status = get_post_warmup_start_end_line_numbers(post_warmup_nframes, 
             warmup_time_period, &start_post_warmup_telemetry_rec, 
             &end_post_warmup_telemetry_rec, &post_warmup_start_frame, 
             &warmup_start_frame, &warmup_stop_frame, &lamp_on_time, 
             &post_warmup_start_time, &post_warmup_stop_time);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Getting scene starting and ending record numbers"); 
        if (IAS_THREAD_UNLOCK_MUTEX(&hdf_io_mutex))
            IAS_LOG_ERROR("Unlocking HDF library mutex");
        return ERROR;
    }

    /* Set the output double pointers be NULL for error check */
    *lamp_current_a = NULL;
    *lamp_voltage_a = NULL;
    *photodiode_micro_amp_a = NULL;
    *lamp_housing_temp_a = NULL;
    *monitor_diode_temp_a = NULL;
    *lamp_current_b = NULL;
    *lamp_voltage_b = NULL;
    *photodiode_micro_amp_b = NULL;
    *lamp_housing_temp_b = NULL;
    *monitor_diode_temp_b = NULL;

    /* Determine how many post warmup temperature records there are */
    num_of_telemetry = end_post_warmup_telemetry_rec - 
                       start_post_warmup_telemetry_rec + 1;

    /* Allocate memory to contain all the post warmup telemetry records */
    telemetry = malloc(num_of_telemetry * sizeof(IAS_L0R_OLI_TELEMETRY));
    if (telemetry == NULL)
    {
        IAS_LOG_ERROR("Unable to allocate memory for OLI ancillary data");
        if (IAS_THREAD_UNLOCK_MUTEX(&hdf_io_mutex))
            IAS_LOG_ERROR("Unlocking HDF library mutex");
        return ERROR;
    }

    /* For simplicity, grab all the post warmup telemetry records. */
    status = ias_l0r_get_ancillary_oli_telemetry( l0r_image, 
             start_post_warmup_telemetry_rec, num_of_telemetry, telemetry );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Cannot obtain OLI ancillary telemetry data");
        if (IAS_THREAD_UNLOCK_MUTEX(&hdf_io_mutex))
            IAS_LOG_ERROR("Unlocking HDF library mutex");
        free(telemetry);
        return ERROR;
    }

    /* Allocate memory to contain all the post warmup temperature records */
    oli_tirs_temperatures = malloc(num_of_telemetry * 
        sizeof(IAS_L0R_OLI_TIRS_TEMPERATURE));
    if (oli_tirs_temperatures == NULL)
    {
        IAS_LOG_ERROR("Unable to allocate memory for OLI ancillary data");
        free(telemetry);
        if (IAS_THREAD_UNLOCK_MUTEX(&hdf_io_mutex))
            IAS_LOG_ERROR("Unlocking HDF library mutex");
        return ERROR;
    }

    /* For simplicity, grab all the post warmup oli_tirs_temperatures records */
    status = ias_l0r_get_ancillary_oli_tirs_temperature( l0r_image,  
             start_post_warmup_telemetry_rec, num_of_telemetry, 
             oli_tirs_temperatures );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Cannot obtain OLI ancillary telemetry data");
        if (IAS_THREAD_UNLOCK_MUTEX(&hdf_io_mutex))
            IAS_LOG_ERROR("Unlocking HDF library mutex");
        free(telemetry);
        free(oli_tirs_temperatures);
        return ERROR;
    }

    if ( IAS_THREAD_UNLOCK_MUTEX( &hdf_io_mutex ))
    {
        IAS_LOG_ERROR("Unlocking HDF library mutex");
        free(telemetry);
        free(oli_tirs_temperatures);
        return ERROR;
    }

    /* Allocate memory for the local post warmup lamp telemetry record */
    local_lamp_current_a = malloc( num_of_telemetry * sizeof(float));
    local_lamp_voltage_a = malloc( num_of_telemetry * sizeof(float));
    local_photodiode_micro_amp_a = malloc( num_of_telemetry * sizeof(float));
    local_lamp_housing_temp_a = malloc( num_of_telemetry * sizeof(float));
    local_monitor_diode_temp_a = malloc( num_of_telemetry * sizeof(float));
    local_lamp_current_b = malloc( num_of_telemetry * sizeof(float));
    local_lamp_voltage_b = malloc( num_of_telemetry * sizeof(float));
    local_photodiode_micro_amp_b = malloc( num_of_telemetry * sizeof(float));
    local_lamp_housing_temp_b = malloc( num_of_telemetry * sizeof(float));
    local_monitor_diode_temp_b = malloc( num_of_telemetry * sizeof(float));
    if ( !local_lamp_current_a || !local_lamp_voltage_a 
         || !local_photodiode_micro_amp_a || !local_lamp_housing_temp_a 
         || !local_monitor_diode_temp_a || !local_lamp_current_b 
         || !local_lamp_voltage_b || !local_photodiode_micro_amp_b 
         || !local_lamp_housing_temp_b || !local_monitor_diode_temp_b )
    {
        free(local_lamp_current_a);
        free(local_lamp_voltage_a);
        free(local_photodiode_micro_amp_a);
        free(local_lamp_housing_temp_a);
        free(local_monitor_diode_temp_a);
        free(local_lamp_current_b);
        free(local_lamp_voltage_b);
        free(local_photodiode_micro_amp_b);
        free(local_lamp_housing_temp_b);
        free(local_monitor_diode_temp_b);
        local_lamp_current_a = NULL;
        local_lamp_voltage_a = NULL;
        local_photodiode_micro_amp_a = NULL;
        local_lamp_housing_temp_a = NULL;
        local_monitor_diode_temp_a = NULL;
        local_lamp_current_b = NULL;
        local_lamp_voltage_b = NULL;
        local_photodiode_micro_amp_b = NULL;
        local_lamp_housing_temp_b = NULL;
        local_monitor_diode_temp_b = NULL;
        *number_lamp_telemetry = 0;
        free(telemetry);
        free(oli_tirs_temperatures);
        IAS_LOG_ERROR("Allocating lamp telemetry arrays");
        return ERROR;
    }

    /* Get the ancillary records for the scene and extract the two
       temperatures */
    for ( i = 0; i < num_of_telemetry; i++ )
    {
        /* Populate the lamp telemetry data from payload structure */
        local_lamp_current_a[i] = 
            telemetry[i].oli_telemetry_header_3.stim_lamp_output_current_amps;
        local_lamp_voltage_a[i] = 
            telemetry[i].oli_telemetry_header_3.stim_lamp_bulb_a_volts;
        local_photodiode_micro_amp_a[i] = 
            telemetry[i].oli_telemetry_header_3.
            stim_lamp_photodiode1_micro_amps;
        local_lamp_housing_temp_a[i] = 
            telemetry[i].oli_telemetry_header_3.stim_lamp_thermistor1_celsius;
        local_monitor_diode_temp_a[i] = 
            oli_tirs_temperatures[i].temperatures_celsius
               [OLI_STIMULATION_LAMP_1_DIODE_BOARD];
        local_lamp_current_b[i] = 
            telemetry[i].oli_telemetry_header_3.stim_lamp_output_current_amps;
        local_lamp_voltage_b[i] = 
            telemetry[i].oli_telemetry_header_3.stim_lamp_bulb_b_volts;
        local_photodiode_micro_amp_b[i] = 
            telemetry[i].oli_telemetry_header_3.
            stim_lamp_photodiode2_micro_amps;
        local_lamp_housing_temp_b[i] = 
            telemetry[i].oli_telemetry_header_3.stim_lamp_thermistor2_celsius;
        local_monitor_diode_temp_b[i] = 
            oli_tirs_temperatures[i].temperatures_celsius
            [OLI_STIMULATION_LAMP_2_DIODE_BOARD];
    }

    *number_lamp_telemetry = num_of_telemetry;
    *lamp_current_a = local_lamp_current_a;
    *lamp_voltage_a = local_lamp_voltage_a;
    *photodiode_micro_amp_a = local_photodiode_micro_amp_a;
    *lamp_housing_temp_a = local_lamp_housing_temp_a;
    *monitor_diode_temp_a = local_monitor_diode_temp_a;
    *lamp_current_b = local_lamp_current_b;
    *lamp_voltage_b = local_lamp_voltage_b;
    *photodiode_micro_amp_b = local_photodiode_micro_amp_b;
    *lamp_housing_temp_b = local_lamp_housing_temp_b;
    *monitor_diode_temp_b = local_monitor_diode_temp_b;

    /* Clearup */
    free(telemetry);
    free(oli_tirs_temperatures);

    return SUCCESS;

   /* The L0R image file will be closed somewhere else */
}  /* END read_post_warmup_lamp_telemetry */


/*------------------------------------------------------------------------
NAME:    read_stim_lamp_pair_id

PURPOSE: Get the stim lamp pair id

RETURNS: Integer status code of SUCCESS or ERROR
------------------------------------------------------------------------*/
int read_stim_lamp_pair_id
(
    BLOB *blob                       /* O: Set stim lamp pair id */
)
{
    IAS_L0R_OLI_TELEMETRY *telemetry = NULL;
    int count_of_ancillary_data;         /* Number of ancillary telemetry
                                            records */
    int lamp_pair_id = -1;
    int stim_power_status = RPS_STL_POWER_OFF;
                                         /* Stim lamp power status */
    int status;
    int rec;
    
    if ( IAS_THREAD_LOCK_MUTEX( &hdf_io_mutex ))
    {
        IAS_LOG_ERROR("Locking HDF library mutex");
        return ERROR;
    }

    /* Find out how many records there are in the telemetry data */
    status = ias_l0r_get_ancillary_oli_telemetry_records_count( l0r_image,
        &count_of_ancillary_data );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Getting size of ancillary data");
        if (IAS_THREAD_UNLOCK_MUTEX(&hdf_io_mutex))
            IAS_LOG_ERROR("Unlocking HDF library mutex");
        return ERROR;
    }

    /* Allocate a buffer to contain all the telemetry records */
    telemetry = malloc(count_of_ancillary_data * sizeof(IAS_L0R_OLI_TELEMETRY));
    if (telemetry == NULL)
    {
        IAS_LOG_ERROR("Unable to allocate memory for OLI ancillary data");
        if (IAS_THREAD_UNLOCK_MUTEX(&hdf_io_mutex))
            IAS_LOG_ERROR("Unlocking HDF library mutex");
        return ERROR;
    }


    /* For simplicity, grab all the telemetry records. */
    status = ias_l0r_get_ancillary_oli_telemetry( l0r_image, 0,
        count_of_ancillary_data, telemetry );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Cannot obtain OLI ancillary telemetry data");
        if (IAS_THREAD_UNLOCK_MUTEX(&hdf_io_mutex))
            IAS_LOG_ERROR("Unlocking HDF library mutex");
        free(telemetry);
        return ERROR;
    }

    if ( IAS_THREAD_UNLOCK_MUTEX( &hdf_io_mutex ))
    {
        IAS_LOG_ERROR("Unlocking HDF library mutex");
        free(telemetry);
        return ERROR;
    }

    /* Find the lamp_pair_id after the lamp is powered on  */
    for ( rec = 0; rec <= count_of_ancillary_data - 1; rec++ )
    {
        int stl_flags;

        stl_flags = telemetry[rec].oli_telemetry_header_4.stl_flags;

        /* Check to see if the current record indicates a lamp pair is
           powered on. */
        stim_power_status = get_nth_bit(stl_flags,
            RPS_STL_MAIN_POWER_STATUS_BIT);
        if (stim_power_status == RPS_STL_POWER_ON)
        {
           /* Get the pair ID value.  */
           lamp_pair_id = stl_flags & RPS_STL_PAIR_ID_BITMASK;

           /* We only need to check the first record where the lamp is
              powered on. */
           break;
        }
    }

    /* Suppose we have a lamp collect, but the telemetry data never shows any
       lamp pair was powered on.  This shouldn't ever occur in practice.  If
       it does occur, something is seriously wrong elsewhere that needs to be
       looked at.  */
    if (stim_power_status == RPS_STL_POWER_OFF)
    {
        IAS_LOG_ERROR("Cannot determine lamp pair ID used for this lamp "
            "collect");
        free(telemetry);
        return ERROR;
    }

    /* Set stim lamp pair id in the blob */
    blob->stim_lamp_pair_id = lamp_pair_id;

    /* Free the allocated memory */
    free(telemetry);

    return SUCCESS;
    /* The L0R image file will be closed somewhere else */
}  /* END read_stim_lamp_pair_id */


/*------------------------------------------------------------------------
NAME:    write_l1r_output_image_data

PURPOSE: Wrapper routine to call write_l1r_image_data which writes a
         single SCA of image data to the L1R output image.

         Note: This wrapper is needed since external callers don't know
               the L1R band structure for the output file

RETURNS: SUCCESS or ERROR
------------------------------------------------------------------------*/
int write_l1r_output_image_data
(
    int band_index,     /* I: */
    int sca_index,      /* I: */
    float *l1r_data     /* I: Pointer to the image data */
)
{
    if ( ! l1r_band[band_index] )
    {
        IAS_LOG_ERROR("Band %d hasn't been opened for writing",
            get_band_attributes( band_index )->band_number );
        return ERROR;
    }

    /* Call write_intermediate_l1r_image_data with the correct band */
    return write_l1r_image_data( l1r_band[band_index],
        band_index, sca_index, l1r_data );
}  /* END write_l1r_image_data */


/*------------------------------------------------------------------------
NAME:    write_l1r_image_data

PURPOSE: Write an SCA of image data to the L1R output file.

         Note: This is provided for external callers (dump_imagery) who
               have their own L1R band structure

RETURNS: SUCCESS or ERROR
------------------------------------------------------------------------*/
int write_l1r_image_data
(
    L1R_BAND_IO *band_struct, /* I: Pointer to image band structure */
    int band_index,           /* I: Band to write */
    int sca_index,            /* I: SCA to write */
    float *l1r_data           /* I: Pointer to the image data */
)
{
    RPS_BAND_ATTRIBUTES *band_attributes = get_band_attributes( band_index );
    int number_lines = band_attributes->number_lines;
    int number_samples = band_attributes->number_detectors;
    int starting_line = 0;
    int starting_sample = 0;
    int status;

    if ( IAS_THREAD_LOCK_MUTEX( &hdf_io_mutex ))
    {
        IAS_LOG_ERROR("Locking HDF library mutex");
        return ERROR;
    }

    status =  ias_l1r_write_image( band_struct, sca_index, starting_line,
        starting_sample, number_lines, number_samples, (void *)l1r_data );

    if ( IAS_THREAD_UNLOCK_MUTEX( &hdf_io_mutex ))
    {
        IAS_LOG_ERROR("Unlocking HDF library mutex");
        return ERROR;
    }

    return status;
}  /* END write_l1r_image_data */
