/*----------------------------------------------------------------------------
NAME:     memory_management.c

ROUTINES: initialize_memory_management_buffers
          cleanup_memory_management_buffers
          increment_consumer_count
          decrement_consumer_count
          get_memory_buffer
          get_const_memory_buffer
          get_memory_buffer_number_values
          set_consumer_counts
          set_buffer_filled_flag
          get_buffer_filled_flag

PURPOSE:  Collection of routines to manage memory.  In general a routine is
          called to retrieve a pointer to some data, l0r data for example.
          If the requested memory is NULL a routine will be called
          to allocate, and possibly populate, and return the pointer.  The
          pointer will be stored and a count of the number of consumers is
          initialized.  When the current consumer no longer needs the data it
          decrements the count.  When the count is zero the memory is freed.
          Special cases such as CPF and BPF manage their own memory -- that
          memory is not freed here.
-----------------------------------------------------------------------------*/
#include <stdlib.h> 
#include <stdint.h>
#include <string.h>
#include <math.h>

#include "ias_threadsync.h"
#include "ias_logging.h"
#include "ias_rlut.h"
#include "ias_satellite_attributes.h"
#include "ias_radiometry.h"
#include "rps_local.h"

/* Mutex for controlling access to the BPF I/O library */
static IAS_THREAD_MUTEX_TYPE bpf_io_mutex;

/* ----------------- Local CPF prototypes ----------------- */
static int get_cpf_absolute_gains( void );
static int get_cpf_relative_gains( void );
static int get_cpf_saturation_levels( MEMORY_BUFFER_TYPE );
static int get_cpf_temp_sensitivity_reference_temp( void );
static int get_cpf_temp_sensitivity_coefficient( void );
static int get_cpf_reflectance_conversion( void );
static int get_cpf_lamp_radiance( MEMORY_BUFFER_TYPE buffer_type );
static int get_cpf_focal_plane( MEMORY_BUFFER_TYPE buffer_type );
static int get_cpf_inoperable_detectors( );
static int get_cpf_tirs_background_response( void );
static int get_cpf_tirs_gain_offsets( void );
static int get_cpf_tirs_dark_response( void );
static int get_cpf_tirs_dark_response_blind( void );
static int get_cpf_tirs_relative_gains_blind( void );
static int get_cpf_tirs_absolute_gains_blind( void );
static int get_cpf_bias( MEMORY_BUFFER_TYPE buffer_type );
static int get_cpf_impulse_noise( MEMORY_BUFFER_TYPE buffer_type );
static int get_cpf_detector_noise( void );
static int get_cpf_sca_parameters( MEMORY_BUFFER_TYPE buffer_type );
static int get_cpf_histogram_char_parameters(MEMORY_BUFFER_TYPE buffer_type);
static int get_sca_correction_factor(void);
static int get_cpf_discontinuity_ratio( void );

/* ----------------- Normalized detector offset prototype --------------*/
static int get_spatial_detector_offsets( MEMORY_BUFFER_TYPE buffer_type );

/* ----------------- Frame Header prototypes ----------- */
static int get_l0r_frame_headers(MEMORY_BUFFER_TYPE buffer_type);

/* ----------------- Local BPF prototypes ----------------- */
static int read_bpf_bias_model_parameters( int band_number, int sca_number,
           MEMORY_BUFFER_TYPE buffer_type, void ** pre_shutter,
           void ** post_shutter, void ** mean_shutter, void **a1_parameter, 
           void **c1_parameter, void **a0_parameter, int *number_detectors );

static int read_tirs_bpf_bias_model_parameters( int band_number, 
           int sca_number, void ** background_response, int *number_detectors );

/* ----------------- Local RLUT prototypes ---------------- */
static int get_rlut_linearization_parameters
(
    MEMORY_BUFFER_TYPE buffer_type,
    int band_number,
    int sca_number
);

static int get_rlut_lookup_linearization_parameters
(
    MEMORY_BUFFER_TYPE buffer_type,
    int band_number,
    int sca_number
);

static int get_rlut_tirs_secondary_linearization_parameters
(
    MEMORY_BUFFER_TYPE buffer_type,
    int band_number,
    int sca_number
);

/* -------------------- Memory buffers -------------------- */
static MEMORY_BUFFER **buffers[NUM_MEMORY_BUFFERS];
static int number_consumers[NUM_MEMORY_BUFFERS];
static int total_bands;

/* Processing band list -- contains all the band indicess, including VRP and
   blind bands.  Although order isn't important the indices have been sorted
   in increasing order. */
static int *mm_list_of_band_indices = NULL;
static int mm_number_of_bands = 0;

static int set_consumer_counts_called_flag = 0;


/*------------------------------------------------------------------------
NAME:    get_buffer_size

PURPOSE: This private routine calculates the total size of a memory buffer.
         For each MEMORY_BUFFER_TYPE there must be a corresponding case in
         the switch statement.

RETURNS: total size of the buffer
------------------------------------------------------------------------*/
static size_t get_buffer_size
(
    MEMORY_BUFFER_TYPE buffer_type, /* I: Memory type */
    int band_index,                 /* I: */
    int *number_values              /* O: */
)
{
    RPS_BAND_ATTRIBUTES *band_attributes =
        get_band_attributes( band_index );
    size_t buffer_size = 0;
    size_t num_lines = band_attributes->number_lines;
    size_t num_samps = band_attributes->number_detectors;
    int value_count = 0;

    switch ( buffer_type )
    {
        case IMAGE_L0R :
            value_count = num_lines * num_samps;
            buffer_size = value_count * sizeof(uint16_t);
            break;
        case IMAGE_L1R :
            value_count = num_lines * num_samps;
            buffer_size = value_count * sizeof(float);
            break;

        /* Output is per-frame.  Can either be per-detector bias
           duplicated across all frames, or actual per-detector/per-frame
           bias derived from a model estimate */
        case MM_NON_PAN_PERFRAME_BIAS:
        case MM_PAN_PERFRAME_BIAS:
        case MM_SCENE_STRIPING_CORRECTION_MATRIX :
            value_count = num_lines * num_samps;
            buffer_size = value_count * sizeof(double);
            break;

        /* Shutter statistics. Always per-detector */
        case MM_NON_PAN_PRE_SHUTTER:
        case MM_PAN_ODD_PRE_SHUTTER:
        case MM_PAN_EVEN_PRE_SHUTTER:
        case MM_NON_PAN_POST_SHUTTER:
        case MM_PAN_ODD_POST_SHUTTER:
        case MM_PAN_EVEN_POST_SHUTTER:
        case MM_NON_PAN_MEAN_SHUTTER:
        case MM_PAN_ODD_MEAN_SHUTTER:
        case MM_PAN_EVEN_MEAN_SHUTTER:
        case MM_NON_PAN_BIAS_A1_PARAMETER:
        case MM_PAN_ODD_BIAS_A1_PARAMETER:
        case MM_PAN_EVEN_BIAS_A1_PARAMETER:
        case MM_NON_PAN_BIAS_C1_PARAMETER:
        case MM_PAN_ODD_BIAS_C1_PARAMETER:
        case MM_PAN_EVEN_BIAS_C1_PARAMETER:
        case MM_NON_PAN_CPF_BIAS:
        case MM_PAN_ODD_CPF_BIAS:
        case MM_PAN_EVEN_CPF_BIAS:
        case MM_DETECTOR_STRIPING_METRIC :
        case MM_BPF_BACKGROUND_RESPONSE :
            value_count = num_samps;
            buffer_size = value_count * sizeof(double);
            break;

        /* Linearization parameters.  Also per-detector */
        case MM_3Q_RESPONSE_LINEARIZATION_PARAMS:
            value_count = num_samps;
            buffer_size = value_count * sizeof(IAS_RLUT_LINEARIZATION_PARAMS);
            break;

        /* Lookup table linearization parameters */
        case MM_LOOKUP_RESPONSE_LINEARIZATION_PARAMS:
        case MM_TIRS_SECONDARY_LINEARIZATION_PARAMS :
            value_count = 1;
            buffer_size = sizeof(IAS_RLUT_LOOKUP_TABLE *);
            break;

        case MM_NON_PAN_BIAS_A0_PARAMETER:
        case MM_PAN_ODD_BIAS_A0_PARAMETER:
        case MM_PAN_EVEN_BIAS_A0_PARAMETER:
            value_count = 1;
            buffer_size = value_count * sizeof(double);
            break;

        case MM_NORMALIZED_DETECTOR_OFFSETS :
        case MM_SPATIAL_DETECTOR_OFFSETS :
            value_count = num_samps;
            buffer_size = value_count * sizeof(int);
            break;

        case MM_ANALOG_HIGH_SAT :
        case MM_ANALOG_LOW_SAT :
        case MM_DIGITAL_HIGH_SAT :
        case MM_DIGITAL_LOW_SAT :
            value_count = num_samps;
            buffer_size = value_count * sizeof(int);
            break;

        case MM_RELATIVE_GAINS:
        case MM_TIRS_RELATIVE_GAINS_BLIND:
        case MM_TEMP_SENSITIVITY_COEFFICIENT :
        case MM_TEMP_CORRECTION_FACTOR :
        case MM_DETECTOR_NOISE :
            value_count = num_samps;
            buffer_size = value_count * sizeof(double);
            break;

        case MM_TIRS_DARK_RESPONSE:
        case MM_TIRS_DARK_RESPONSE_BLIND:
        case MM_TIRS_BACKGROUND_RESPONSE :
        case MM_TIRS_GAIN_OFFSETS :
            value_count = num_samps;
            buffer_size = value_count * sizeof(double);
            break;

        /* These are for single value entities of double */
        case MM_ABSOLUTE_GAINS:
        case MM_TIRS_ABSOLUTE_GAINS_BLIND:
        case MM_TEMP_SENSITIVITY_REFERENCE_TEMP :
        case MM_SCA_OVERLAP :
        case MM_TIRS_SCA_DARK_RESPONSE_DIFFERENCE:
        case MM_STRIPING_METRIC_CUTOFF :
        case MM_SCA_CORRECTION_FACTOR:
        case MM_SCA_CORRECTION_FACTOR_CPF:
        case MM_SCA_OVERLAP_THRESHOLD :
        case MM_CPF_DISCONTINUITY_RATIO :
            value_count = 1;
            buffer_size = value_count * sizeof(double);
            break;

        /* These are for single value entities of int */
        case MM_HIST_FRAMES_TO_SKIP_TOP :
        case MM_HIST_FRAMES_TO_SKIP_BOTTOM :
        case MM_IMPULSE_NOISE_LIMIT :
        case MM_IMPULSE_NOISE_FILTER_WIDTH :
        case MM_SCA_VALID_NEIGHBOR_SEGMENTS :
        case MM_SCA_CORRELATION_SHIFT :
        case MM_MAX_NORMALIZED_DETECTOR_OFFSET :
        case MM_MAX_SPATIAL_DETECTOR_OFFSETS :
        case MM_HIGH_RADIANCE_SAT:
        case MM_LOW_RADIANCE_SAT:
        case MM_CPF_SCA_OFFSET :
            value_count = 1;
            buffer_size = value_count * sizeof(int);
            break;

        /* For Detector Response Characterization for Lamp */
        case MM_L0R_INTERMEDIATE_STATISTICS :
        case MM_L1R_INTERMEDIATE_STATISTICS :
            value_count = num_samps;
            buffer_size =
                value_count * sizeof(IAS_RPS_HISTOGRAM_INTERMEDIATE_RESULTS);
            break;

        /* These cases are for buffer sizes that are dynamic and determined
           when they are filled.  A good example of this are the focal plane
           temperatures from the ancillary data. */
        case MM_REFLECTANCE_CONVERSION_COEFFICIENTS :
        case MM_INOPERABLE_DETECTORS :
            value_count = 1;
            buffer_size = value_count * sizeof(float);
            break;

        /* For Detector Response Characterization for Lamp */
        case MM_L0R_LAMP_ONE_SCA_STATISTICS :
        case MM_L1R_LAMP_ONE_SCA_STATISTICS :
            value_count = 1;
            buffer_size = value_count * 
                        sizeof(DB_CHAR_LAMP_RESPONSE_ONE_SCA_RESULTS);
            break;

        case MM_WORKING_LAMP_RADIANCE :
        case MM_BACKUP_LAMP_RADIANCE :
        case MM_PRISTINE_LAMP_RADIANCE :
            value_count = num_samps;
            buffer_size = value_count * sizeof(double);
            break;

        /* OLI/TIRS L0R frame headers (for dropped frame characterization).
           The ultimate size of these buffers depends on the number of
           frame headers */
        case MM_L0R_OLI_FRAME_HEADER:
            value_count = 1;
            buffer_size = value_count * sizeof(IAS_L0R_OLI_FRAME_HEADER);
            break;
        case MM_L0R_TIRS_FRAME_HEADER:
            value_count = 1;
            buffer_size = value_count * sizeof(IAS_L0R_TIRS_FRAME_HEADER);
            break;

        default :
            IAS_LOG_ERROR(
                /* Don't attempt to use get_memory_buffer_string()
                   here since the buffer type is invalid */
                "Attempting to get buffer size for invalid memory type: %d "
                "Valid values are 0 - %d", buffer_type,
                NUM_MEMORY_BUFFERS - 1 );
            buffer_size = 0;
            value_count = 0;
    }

    *number_values = value_count;
    return buffer_size;
}  /* END get_buffer_size */


/*------------------------------------------------------------------------
NAME:    initialize_memory_management_buffers

PURPOSE: This routine ensures the buffers are NULL so the
         increment_consumer_count routine will work.  This should be called
         prior to any pre-processing steps.

RETURNS: none
------------------------------------------------------------------------*/
int initialize_memory_management_buffers( void )
{
    MEMORY_BUFFER_TYPE type;
    const IAS_SATELLITE_ATTRIBUTES *sat_atts;
    int status;
    int *blob_list_of_band_indices;
    int number_bands;
    int i;

    sat_atts = ias_sat_attr_get_attributes();
    total_bands = sat_atts->total_bands;
    for ( type = 0; type < NUM_MEMORY_BUFFERS; type++ )
    {
        buffers[type] = NULL;
        number_consumers[type] = 0;
    }

    /* Create a mutex for control of the BPF library */
    status = IAS_THREAD_CREATE_MUTEX( &bpf_io_mutex );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Couldn't create BPF I/O mutex");
        return ERROR;
    }

    /* Set up the local list of band indices.  This will include VRP and
       blind bands. */
    if ( !get_band_fields_initialized() )
    {
        IAS_LOG_ERROR("The band information fields aren't initialized");
        return ERROR;
    }

    mm_list_of_band_indices = malloc( sizeof(int) * total_bands );
    if ( !mm_list_of_band_indices )
    {
        IAS_LOG_ERROR("Allocating memory management list of band indices");
        return ERROR;
    }

    number_bands = get_blob_number_of_bands();
    blob_list_of_band_indices = get_blob_list_of_band_indices_to_process();
    mm_number_of_bands = 0;
    for ( i = 0; i < number_bands; i++ )
    {
        int band_index = blob_list_of_band_indices[i];
        RPS_BAND_ATTRIBUTES *band_atts = get_band_attributes( band_index );
        int vrp_band_index = band_atts->vrp_band_index;
        int blind_band_index = band_atts->blind_band_index;

        mm_list_of_band_indices[mm_number_of_bands++] = band_index;

        if ( vrp_band_index >= 0 )
            mm_list_of_band_indices[mm_number_of_bands++] = vrp_band_index;

        if ( blind_band_index >= 0
            && band_atts->process_blind_band_flag )
        {
            mm_list_of_band_indices[mm_number_of_bands++] = blind_band_index;
        }
    }
    ias_math_insertion_sort_integer_array( mm_number_of_bands,
        mm_list_of_band_indices );

    return SUCCESS;
}  /* END initialize_memory_management_buffers */


/*------------------------------------------------------------------------
NAME:    cleanup_memory_management_buffers

PURPOSE: This frees all MEMORY_BUFFER allocations with the exception of CPF

RETURNS: none
------------------------------------------------------------------------*/
void cleanup_memory_management_buffers( int display_warning )
{
    MEMORY_BUFFER_TYPE type;
    int band_index;
    int sca_index;

    for ( type = 0; type < NUM_MEMORY_BUFFERS; type++ )
    {
        if ( buffers[type] != NULL )
        {
            for ( band_index = 0; band_index < total_bands; band_index++ )
            {
                /* Since there are non-constant buffers read from the CPF
                   and copied into buffers for all bands there may be
                   residual memory hanging around if that band wasn't
                   processed.  Free the data. */
                for ( sca_index = 0; sca_index < get_blob_max_nscas();
                    sca_index++ )
                {
                    /* Only free non-constant data */
                    if ( ! buffers[type][band_index][sca_index].is_const_data
                      && buffers[type][band_index][sca_index].data_pointer )
                    {
                        /* If this buffer wasn't freed prior issue warning */
                        if ( buffers[type][band_index][sca_index].data_pointer )
                        {
                            /* If the display warning flag is not set to
                               the defined value ENABLE_WARNING_DISPLAY,
                               we're at this point because we're cleaning
                               up from an error condition.  If the flag
                               is set to ENABLE_WARNING_DISPLAY, we're
                               cleaning up after processing successfully
                               completes; we'll want to know whether we're
                               leaking memory in this case. */
                            if (display_warning == ENABLE_WARNING_DISPLAY)
                            {
                                IAS_LOG_WARNING("Band %2d SCA %2d buffer %s "
                                    "was leaked, %d of %d remaining",
                                    band_index + 1, sca_index + 1,
                                    get_memory_buffer_string(type),
                                    buffers[type][band_index][sca_index]
                                        .user_count,
                                    number_consumers[type]);
                            }
                        }
                        free( buffers[type][band_index][sca_index].
                            data_pointer );
                        buffers[type][band_index][sca_index].data_pointer =
                            NULL;
                    }
                }
                free( buffers[type][band_index] );
            }
            free( buffers[type] );
        }
    }

    free( mm_list_of_band_indices );

    IAS_THREAD_DESTROY_MUTEX( &bpf_io_mutex );
}  /* END cleanup_memory_management_buffers */


/*------------------------------------------------------------------------
NAME:    set_consumer_counts

PURPOSE: This routine sets all the consumer counts to the values accumulated
         in the pre-process increments.  It's only allowed to be called
         once.  Normal get memory buffer calls won't be allowed until this
         routine is called.  Once this routine is called, pre-process get
         memory buffer calls will no longer be allowed.

         This routine should be called just prior to process collection
         setting the threads free to process the SCAs.

RETURNS: SUCCESS or ERROR
------------------------------------------------------------------------*/
int set_consumer_counts( void )
{
    MEMORY_BUFFER_TYPE buffer_type;
    int band_index;
    int sca_index;
    int max_nscas = get_blob_max_nscas();

    if ( set_consumer_counts_called_flag )
    {
        IAS_LOG_ERROR("Set consumer counts has already been called");
        return ERROR;
    }

    set_consumer_counts_called_flag = 1;

    for ( buffer_type = 0; buffer_type < NUM_MEMORY_BUFFERS; buffer_type++ )
    {
        if ( buffers[buffer_type] != NULL )
        {
            for ( band_index = 0; band_index < total_bands; band_index++ )
            {
                for ( sca_index = 0; sca_index < max_nscas; sca_index++ )
                {
                    buffers[buffer_type][band_index][sca_index].user_count =
                        number_consumers[buffer_type];
                    buffers[buffer_type][band_index][sca_index].consumer_set =
                        1;
                }
            }
        }
    }
    return SUCCESS;
}  /* END set_consumer_counts */


/*------------------------------------------------------------------------
NAME:    increment_consumer_count

PURPOSE: This routine increments the consumer count for a memory buffer.
         It is intended to be called by the pre-process routines that need
         a memory buffer.  This count will not be altered once processing
         is begun.

         The buffers may or may not be initialized by this routine.

RETURNS: the status from any internally called routine
         SUCCESS or ERROR
------------------------------------------------------------------------*/
int increment_consumer_count
(
    MEMORY_BUFFER_TYPE buffer_type  /* I: Index into number consumers array */
)
{
    int band_index;
    int sca_index;
    int max_num_scas = get_blob_max_nscas();
    int number_values;
    int status = SUCCESS;

    if ( set_consumer_counts_called_flag )
    {
        IAS_LOG_ERROR("Increment can't be called after set consumer counts "
            "has been called.  Called with Buffer: %s",
            get_memory_buffer_string( buffer_type ));
        return ERROR;
    }

    /* If this buffer hasn't been allocated yet do the allocation */
    if ( buffers[buffer_type] == NULL )
    {
        /* Allocate an array of pointer ensuring that they are initialized
           to NULL.  They must be NULL so the free routine will work
           correctly. */
        buffers[buffer_type] = calloc( total_bands, sizeof(MEMORY_BUFFER *));
        /* Error message is output by caller */
        if ( !buffers[buffer_type] )
            return ERROR;

        /* Now allocate space for all bands and SCAs and set the values to
           NULL and zero */
        for ( band_index = 0; band_index < total_bands; band_index++ )
        {
            buffers[buffer_type][band_index] =
                malloc( sizeof(MEMORY_BUFFER) * max_num_scas );
            if ( !buffers[buffer_type][band_index] )
            {
                /* Error message is output by caller */
                cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
                return ERROR;
            }
            for ( sca_index = 0; sca_index < max_num_scas; sca_index++ )
            {
                buffers[buffer_type][band_index][sca_index].user_count = 0;
                buffers[buffer_type][band_index][sca_index].is_const_data = 0;
                buffers[buffer_type][band_index][sca_index].data_pointer =
                    NULL;
                buffers[buffer_type][band_index][sca_index].const_data_pointer
                    = NULL;
                buffers[buffer_type][band_index][sca_index].size = 
                    get_buffer_size( buffer_type, band_index, &number_values );
                buffers[buffer_type][band_index][sca_index].number_values = 
                    number_values;
                buffers[buffer_type][band_index][sca_index].buffer_allocated
                    = 0;
                buffers[buffer_type][band_index][sca_index].consumer_set = 0;
                buffers[buffer_type][band_index][sca_index].buffer_filled = 0;
            }
        }

        /* These are special case CPF values that need to be copied into
           locally allocated buffers since they are modified.  The
           routine to read the buffer allocates memory; stores the
           memory pointer in the data_pointer member; and copies the
           data to the newly allocated memory.  The is_const_data is
           left as 0. */
        switch ( buffer_type )
        {
            case MM_TIRS_DARK_RESPONSE:
                status = get_cpf_tirs_dark_response();
                break;
            case MM_TIRS_DARK_RESPONSE_BLIND:
                status = get_cpf_tirs_dark_response_blind();
                break;
            case MM_SCA_CORRECTION_FACTOR_CPF :
                status = get_sca_correction_factor();
                break;
            default:
                break;/* No default action -- here for compiler warning */
        }

        /* The CPF manages its own memory and isn't thread safe so if this
           is a CPF buffer, call the associated routine to populate the
           pointers.  There *must* be a corresponding case added to the
           switch in the decrement_consumer_count to prevent that routine
           from trying to free CPF memory.  This section also applies to the
           BPF or other memory buffers that manage their own memory. */
        switch ( buffer_type )
        {
            /* Populate the buffer based on type */
            case MM_ANALOG_HIGH_SAT :
            case MM_ANALOG_LOW_SAT :
            case MM_DIGITAL_HIGH_SAT :
            case MM_DIGITAL_LOW_SAT :
            case MM_HIGH_RADIANCE_SAT :
            case MM_LOW_RADIANCE_SAT :
                status = get_cpf_saturation_levels( buffer_type );
                break;
            case MM_NON_PAN_CPF_BIAS:
            case MM_PAN_ODD_CPF_BIAS:
            case MM_PAN_EVEN_CPF_BIAS:
                status = get_cpf_bias(buffer_type);
                break;
            case MM_ABSOLUTE_GAINS :
                status = get_cpf_absolute_gains();
                break;
            case MM_TIRS_ABSOLUTE_GAINS_BLIND:
                status = get_cpf_tirs_absolute_gains_blind();
                break;
            case MM_RELATIVE_GAINS :
                status = get_cpf_relative_gains();
                break;
            case MM_TIRS_RELATIVE_GAINS_BLIND:
                status = get_cpf_tirs_relative_gains_blind();
                break;
            case MM_TEMP_SENSITIVITY_REFERENCE_TEMP :
                status = get_cpf_temp_sensitivity_reference_temp();
                break;
            case MM_TEMP_SENSITIVITY_COEFFICIENT :
                status = get_cpf_temp_sensitivity_coefficient();
                break;
            case MM_SCA_OVERLAP :
            case MM_CPF_SCA_OFFSET :
                status = get_cpf_focal_plane( buffer_type );
                break;
            case MM_REFLECTANCE_CONVERSION_COEFFICIENTS :
                status = get_cpf_reflectance_conversion();
                break;
            case MM_WORKING_LAMP_RADIANCE :
            case MM_BACKUP_LAMP_RADIANCE :
            case MM_PRISTINE_LAMP_RADIANCE :
                status = get_cpf_lamp_radiance( buffer_type );
                break;
            case MM_INOPERABLE_DETECTORS :
                status = get_cpf_inoperable_detectors( );
                break;
            case MM_TIRS_BACKGROUND_RESPONSE :
                status = get_cpf_tirs_background_response();
                break;
            case MM_TIRS_GAIN_OFFSETS :
                status = get_cpf_tirs_gain_offsets();
                break;
            case MM_IMPULSE_NOISE_LIMIT :
            case MM_IMPULSE_NOISE_FILTER_WIDTH :
                status = get_cpf_impulse_noise( buffer_type );
                break;
            case MM_DETECTOR_NOISE :
                status = get_cpf_detector_noise();
                break;
            case MM_SCA_CORRELATION_SHIFT :
            case MM_SCA_VALID_NEIGHBOR_SEGMENTS :
            case MM_STRIPING_METRIC_CUTOFF :
            case MM_SCA_OVERLAP_THRESHOLD :
                status = get_cpf_sca_parameters( buffer_type );
                break;
            case MM_CPF_DISCONTINUITY_RATIO :
                status = get_cpf_discontinuity_ratio();
                break;
            case MM_HIST_FRAMES_TO_SKIP_TOP :
            case MM_HIST_FRAMES_TO_SKIP_BOTTOM :
                status = get_cpf_histogram_char_parameters(buffer_type);
                break;

            default:
                break; /* No default action -- here for compiler warning */
        }

        /* The L0R frame header data buffers are treated similarly to the
           CPF data buffers defined above.  These are put into a separate 
           group in case others in the future are defined with similar
           behavior */
        switch (buffer_type)
        {
            case MM_L0R_OLI_FRAME_HEADER:
            case MM_L0R_TIRS_FRAME_HEADER:
                status = get_l0r_frame_headers(buffer_type);
                break;

            /* Get the set of normalized detector offsets or the
               maximum normalized detector offset here.*/
            case MM_SPATIAL_DETECTOR_OFFSETS:
            case MM_MAX_SPATIAL_DETECTOR_OFFSETS:
                status = get_spatial_detector_offsets( buffer_type );
                break;

            default:
                break;/* No default action -- here for compiler warning */
        }
    }

    number_consumers[buffer_type]++;
    /* If an error occurred the message is output by caller */
    return status;
}  /* END increment_consumer_count */


/*------------------------------------------------------------------------
NAME:    decrement_consumer_count

PURPOSE: This routine decrements the user count indicating it is done using
         the data.  When the user count reaches zero, no more algorithms
         need the data so it will be freed.  There are exceptions to the
         freeing for those buffers managed by external routines such as
         CPF and BPF.

RETURNS: SUCCESS or ERROR
------------------------------------------------------------------------*/
int decrement_consumer_count
(
    MEMORY_BUFFER_TYPE buffer_type, /* I: Index into static buffer arrays */
    int band_index,                 /* I: */
    int sca_index                   /* I: */
)
{
    if ( buffers[buffer_type][band_index][sca_index].consumer_set == 0 )
    {
        IAS_LOG_ERROR("Buffer %s for band %d SCA %d is being decremented "
            "before consumer counts have been set --  this could be an error",
            get_memory_buffer_string( buffer_type ),
            get_band_attributes(band_index)->band_number, sca_index + 1 );
        return ERROR;
    }

    buffers[buffer_type][band_index][sca_index].user_count--;

    if ( buffers[buffer_type][band_index][sca_index].user_count <= 0 )
    {
        /* Indicate the buffer was previously allocated to help detect bugs */
        buffers[buffer_type][band_index][sca_index].buffer_allocated = 1;

        switch ( buffer_type )
        {
            /* Case for lookup table response linearization */
            case MM_LOOKUP_RESPONSE_LINEARIZATION_PARAMS :
            {
                IAS_RLUT_LOOKUP_TABLE *tempptr;
                buffers[buffer_type][band_index][sca_index].user_count = 0;
                tempptr = 
                    buffers[buffer_type][band_index][sca_index].data_pointer;
                ias_rlut_free_lookup_table(tempptr);
                tempptr = NULL;
                buffers[buffer_type][band_index][sca_index].data_pointer = NULL;
            }
            break;

            /* CPF case -- do not free the memory since the CPF memory is
               managed by the blob.  Only set the consumer count to zero and
               set the data pointer NULL. */
            case MM_ANALOG_HIGH_SAT :
            case MM_ANALOG_LOW_SAT :
            case MM_DIGITAL_HIGH_SAT :
            case MM_DIGITAL_LOW_SAT :
            case MM_HIGH_RADIANCE_SAT :
            case MM_LOW_RADIANCE_SAT :
            case MM_ABSOLUTE_GAINS:
            case MM_TIRS_ABSOLUTE_GAINS_BLIND:
            case MM_RELATIVE_GAINS:
            case MM_TIRS_RELATIVE_GAINS_BLIND:
            case MM_TEMP_SENSITIVITY_REFERENCE_TEMP :
            case MM_TEMP_SENSITIVITY_COEFFICIENT :
            case MM_REFLECTANCE_CONVERSION_COEFFICIENTS :
            case MM_WORKING_LAMP_RADIANCE:
            case MM_BACKUP_LAMP_RADIANCE:
            case MM_PRISTINE_LAMP_RADIANCE:
            case MM_SCA_OVERLAP :
            case MM_CPF_SCA_OFFSET :
            case MM_INOPERABLE_DETECTORS :
            case MM_TIRS_BACKGROUND_RESPONSE :
            case MM_TIRS_GAIN_OFFSETS :
            case MM_NON_PAN_CPF_BIAS:
            case MM_PAN_ODD_CPF_BIAS:
            case MM_PAN_EVEN_CPF_BIAS:
            case MM_HIST_FRAMES_TO_SKIP_TOP :
            case MM_HIST_FRAMES_TO_SKIP_BOTTOM :
            case MM_IMPULSE_NOISE_LIMIT :
            case MM_IMPULSE_NOISE_FILTER_WIDTH :
            case MM_DETECTOR_NOISE :
            case MM_SCA_CORRELATION_SHIFT :
            case MM_SCA_VALID_NEIGHBOR_SEGMENTS :
            case MM_STRIPING_METRIC_CUTOFF :
            case MM_SCA_OVERLAP_THRESHOLD :
            case MM_CPF_DISCONTINUITY_RATIO :
                buffers[buffer_type][band_index][sca_index].
                    user_count = 0;
                buffers[buffer_type][band_index][sca_index].
                    data_pointer = NULL;
                break;

            /* The L0R frame headers should be treated similarly
               to the previous CPF data buffers.  The only difference is
               that one allocation of the data buffer is performed for
               all bands of a given sensor.  cleanup_io_routines
               frees the "core" memory block */
            case MM_L0R_OLI_FRAME_HEADER:
            case MM_L0R_TIRS_FRAME_HEADER:
                buffers[buffer_type][band_index][sca_index]
                    .user_count = 0;
                buffers[buffer_type][band_index][sca_index]
                    .data_pointer = NULL;
                break;

            /* Special case of the TIRS secondary linearization params */
            case MM_TIRS_SECONDARY_LINEARIZATION_PARAMS :
                ias_rlut_free_lookup_table(
                    buffers[buffer_type][band_index][sca_index].data_pointer );
                buffers[buffer_type][band_index][sca_index]
                    .user_count = 0;
                buffers[buffer_type][band_index][sca_index]
                    .data_pointer = NULL;
                break;

            /* These are special case CPF values that will need to be freed
               here -- these cases fall through to the default case */
            case MM_TIRS_DARK_RESPONSE:
            case MM_TIRS_DARK_RESPONSE_BLIND:

            /* These are special case BPF values that will need to be freed
               here -- these cases fall through to the default case */
            case MM_NON_PAN_PRE_SHUTTER:
            case MM_PAN_ODD_PRE_SHUTTER:
            case MM_PAN_EVEN_PRE_SHUTTER:
            case MM_NON_PAN_POST_SHUTTER:
            case MM_PAN_ODD_POST_SHUTTER:
            case MM_PAN_EVEN_POST_SHUTTER:
            case MM_NON_PAN_MEAN_SHUTTER:
            case MM_PAN_ODD_MEAN_SHUTTER:
            case MM_PAN_EVEN_MEAN_SHUTTER:
            case MM_NON_PAN_BIAS_A0_PARAMETER:
            case MM_PAN_ODD_BIAS_A0_PARAMETER:
            case MM_PAN_EVEN_BIAS_A0_PARAMETER:
            case MM_NON_PAN_BIAS_A1_PARAMETER:
            case MM_PAN_ODD_BIAS_A1_PARAMETER:
            case MM_PAN_EVEN_BIAS_A1_PARAMETER:
            case MM_NON_PAN_BIAS_C1_PARAMETER:
            case MM_PAN_ODD_BIAS_C1_PARAMETER:
            case MM_PAN_EVEN_BIAS_C1_PARAMETER:
            case MM_BPF_BACKGROUND_RESPONSE: 

            /* Special case RLUT values need to be freed here via the
               default mechanism */
            case MM_3Q_RESPONSE_LINEARIZATION_PARAMS :

            /* Normalized detector offset data buffers should also be
               handled by the default mechanism */
            case MM_SPATIAL_DETECTOR_OFFSETS:
            case MM_MAX_SPATIAL_DETECTOR_OFFSETS :

            /* NORMALIZED CORRECTION FACTOR from CPF should also be
               handled by the default mechanism */
            case MM_SCA_CORRECTION_FACTOR_CPF :

            /* Normal case where individual SCA buffers were allocated */
            default :
                buffers[buffer_type][band_index][sca_index].user_count = 0;

                if (buffers[buffer_type][band_index][sca_index].is_const_data)
                {
                    free((void *)buffers[buffer_type][band_index]
                        [sca_index].const_data_pointer);
                    buffers[buffer_type][band_index][sca_index]
                        .const_data_pointer = NULL;
                }
                else
                {
                    free( buffers[buffer_type][band_index]
                        [sca_index].data_pointer );
                    buffers[buffer_type][band_index][sca_index]
                        .data_pointer = NULL;
                }
        }  /* END switch buffer_type */
    }  /* END if user count <= 0 */

    return SUCCESS;
}  /* END decrement_consumer_count */



/*------------------------------------------------------------------------
NAME:    get_const_memory_buffer

PURPOSE: Routine to get a pointer to a const memory buffer.  It is the
         caller's responsibility to cast the void pointer to the correct
         type.  This routine returns const pointers that are read from
         things such as the CPF.

RETURNS: pointer to the buffer
------------------------------------------------------------------------*/
const void *get_const_memory_buffer
(
    MEMORY_BUFFER_TYPE buffer_type, /* I: Index into static buffer arrays */
    int band_index,                 /* I: */
    int sca_index                   /* I: */
)
{
    /* If the data is not a const pointer calling this routine is an error */
    if ( !buffers[buffer_type][band_index][sca_index].is_const_data )
    {
        IAS_LOG_ERROR("Attempting to read a non-const buffer type: %s for "
            "band %d SCA %d", get_memory_buffer_string( buffer_type ),
            get_band_attributes(band_index)->band_number, sca_index + 1 );
        return NULL;
    }

    /* If the consumer count hasn't been set it's an error */
    if ( buffers[buffer_type][band_index][sca_index].consumer_set == 0 )
    {
        IAS_LOG_ERROR("Attempting to read const memory buffer %s for band "
            "%d SCA %d without consumer count set",
            get_memory_buffer_string( buffer_type ),
            get_band_attributes(band_index)->band_number, sca_index + 1 );
        return NULL;
    }

    /* If the buffer hasn't been marked as allocated it's an error */
    if ( buffers[buffer_type][band_index][sca_index].buffer_allocated == 0 )
    {
        IAS_LOG_ERROR("Attempting to read const memory buffer %s for band "
            "%d SCA %d that hasn't been allocated",
            get_memory_buffer_string( buffer_type ),
            get_band_attributes(band_index)->band_number, sca_index + 1 );
        return NULL;
    }

    return buffers[buffer_type][band_index][sca_index].const_data_pointer;
}  /* END get_const_memory_buffer */



/*------------------------------------------------------------------------
NAME:    internal_get_memory_buffer

PURPOSE: Internal routine to get a pointer to a memory buffer.  It is the
         caller's responsibility to cast the void pointer to the correct
         type.  If the buffer isn't available it will be allocated and
         possibly filled with values.  For example: if the data type is
         L0R, the sca will be read.

RETURNS: pointer to the buffer
         NULL on error
------------------------------------------------------------------------*/
static void *internal_get_memory_buffer
(
    MEMORY_BUFFER_TYPE buffer_type, /* I: Index into static buffer arrays */
    MEMORY_USER_TYPE consumer_type, /* I: The type of user:
                                          CONSUMER - the caller uses the
                                            data so the buffer must be valid
                                          SETTER - the caller will fill the
                                            buffer and set the buffer filled
                                            flag */
    int band_index,                 /* I: */
    int sca_index                   /* I: */
)
{
    int status;
    int band_number = get_band_attributes(band_index)->band_number;
    int sca_number = sca_index + 1;

    /* If the data is a const pointer calling this routine is an error */
    if ( buffers[buffer_type][band_index][sca_index].is_const_data )
    {
        IAS_LOG_ERROR("Attempting to read a const buffer type: %s for "
            "band %d SCA %d", get_memory_buffer_string( buffer_type ),
            band_number, sca_number );
        return NULL;
    }

    /* Return immediately if the buffer has already been allocated */
    if ( buffers[buffer_type][band_index][sca_index].data_pointer )
        return buffers[buffer_type][band_index][sca_index].data_pointer;

    /* Verify the buffer has never been allocated before to catch errors with
       buffer management */
    if (buffers[buffer_type][band_index][sca_index].buffer_allocated)
    {
        IAS_LOG_WARNING("Band number %d, SCA number %d: Buffer %s previously "
            "allocated and is being allocated again.  This may be a bug.",
            band_number, sca_number, get_memory_buffer_string( buffer_type ));
    }

    /* Allocate and possibly fill if not already in use.  In most cases there
       is no need to check the status of the allocation or read since the
       pointer will be NULL if errors were encountered.  The caller is
       responsible for checking for NULL pointer before using the buffer. */
    switch ( buffer_type )
    {
        case IMAGE_L0R :
            /* No need to allocate the memory, the read does that and is
               based on actual size of data. */
            buffers[buffer_type][band_index][sca_index].data_pointer =
                read_l0r_image_data( band_index, sca_index );
            break;

        case MM_NON_PAN_PRE_SHUTTER:
        case MM_NON_PAN_POST_SHUTTER:
        case MM_NON_PAN_MEAN_SHUTTER:
        case MM_NON_PAN_BIAS_A0_PARAMETER:
        case MM_NON_PAN_BIAS_A1_PARAMETER:
        case MM_NON_PAN_BIAS_C1_PARAMETER:
            /* This call retrieves all buffers but only sets the number of
               values for the first buffer */
            status = read_bpf_bias_model_parameters( band_number, sca_number, 
                buffer_type, &buffers[MM_NON_PAN_PRE_SHUTTER][band_index]
                [sca_index].data_pointer, &buffers[MM_NON_PAN_POST_SHUTTER]
                [band_index][sca_index].data_pointer,
                &buffers[MM_NON_PAN_MEAN_SHUTTER][band_index][sca_index].
                data_pointer,
                &buffers[MM_NON_PAN_BIAS_A1_PARAMETER][band_index][sca_index].
                data_pointer,
                &buffers[MM_NON_PAN_BIAS_C1_PARAMETER][band_index][sca_index].
                data_pointer,
                &buffers[MM_NON_PAN_BIAS_A0_PARAMETER][band_index][sca_index].
                data_pointer,
                &buffers[MM_NON_PAN_PRE_SHUTTER][band_index][sca_index].
                number_values );
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Reading BPF bias model parameters for NON_PAN "
                              "bands");
                return NULL;
            }    

            /* Set the number of values for the rest buffers */
            buffers[MM_NON_PAN_POST_SHUTTER][band_index][sca_index].
                number_values = buffers[MM_NON_PAN_PRE_SHUTTER][band_index]
                [sca_index].number_values;
            buffers[MM_NON_PAN_MEAN_SHUTTER][band_index][sca_index].
                number_values = buffers[MM_NON_PAN_PRE_SHUTTER][band_index]
                [sca_index].number_values;
            buffers[MM_NON_PAN_BIAS_A1_PARAMETER][band_index][sca_index].
                number_values = 1;
            buffers[MM_NON_PAN_BIAS_C1_PARAMETER][band_index][sca_index].
                number_values = 1;
            buffers[MM_NON_PAN_BIAS_A0_PARAMETER][band_index][sca_index].
                number_values = 1;
            break;

        case MM_PAN_ODD_PRE_SHUTTER:
        case MM_PAN_ODD_POST_SHUTTER:
        case MM_PAN_ODD_MEAN_SHUTTER:
        case MM_PAN_ODD_BIAS_A0_PARAMETER:
        case MM_PAN_ODD_BIAS_A1_PARAMETER:
        case MM_PAN_ODD_BIAS_C1_PARAMETER:
            /* This call retrieves all buffers but only sets the number of
               values for the first buffer */
            status = read_bpf_bias_model_parameters( band_number, sca_number, 
                buffer_type, &buffers[MM_PAN_ODD_PRE_SHUTTER][band_index]
                [sca_index].data_pointer, &buffers[MM_PAN_ODD_POST_SHUTTER]
                [band_index][sca_index].data_pointer,
                &buffers[MM_PAN_ODD_MEAN_SHUTTER][band_index][sca_index].
                data_pointer,
                &buffers[MM_PAN_ODD_BIAS_A1_PARAMETER][band_index][sca_index].
                data_pointer,
                &buffers[MM_PAN_ODD_BIAS_C1_PARAMETER][band_index][sca_index].
                data_pointer,
                &buffers[MM_PAN_ODD_BIAS_A0_PARAMETER][band_index][sca_index].
                data_pointer,
                &buffers[MM_PAN_ODD_PRE_SHUTTER][band_index][sca_index].
                number_values );
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Reading BPF bias model parameters for PAN_ODD "
                              "band");
                return NULL;
            }    

            /* Set the number of values for the rest buffers */
            buffers[MM_PAN_ODD_POST_SHUTTER][band_index][sca_index].
                number_values = buffers[MM_PAN_ODD_PRE_SHUTTER][band_index]
                [sca_index].number_values;
            buffers[MM_PAN_ODD_MEAN_SHUTTER][band_index][sca_index].
                number_values = buffers[MM_PAN_ODD_PRE_SHUTTER][band_index]
                [sca_index].number_values;
            buffers[MM_PAN_ODD_BIAS_A1_PARAMETER][band_index][sca_index].
                number_values = 1;
            buffers[MM_PAN_ODD_BIAS_C1_PARAMETER][band_index][sca_index].
                number_values = 1;
            buffers[MM_PAN_ODD_BIAS_A0_PARAMETER][band_index][sca_index].
                number_values = 1;
            break;

        case MM_PAN_EVEN_PRE_SHUTTER:
        case MM_PAN_EVEN_POST_SHUTTER:
        case MM_PAN_EVEN_MEAN_SHUTTER:
        case MM_PAN_EVEN_BIAS_A0_PARAMETER:
        case MM_PAN_EVEN_BIAS_A1_PARAMETER:
        case MM_PAN_EVEN_BIAS_C1_PARAMETER:
            /* This call retrieves all buffers but only sets the number of
               values for the first buffer */
            status = read_bpf_bias_model_parameters( band_number, sca_number, 
                buffer_type, &buffers[MM_PAN_EVEN_PRE_SHUTTER][band_index]
                [sca_index].data_pointer, &buffers[MM_PAN_EVEN_POST_SHUTTER]
                [band_index][sca_index].data_pointer,
                &buffers[MM_PAN_EVEN_MEAN_SHUTTER][band_index][sca_index].
                data_pointer,
                &buffers[MM_PAN_EVEN_BIAS_A1_PARAMETER][band_index][sca_index].
                data_pointer,
                &buffers[MM_PAN_EVEN_BIAS_C1_PARAMETER][band_index][sca_index].
                data_pointer,
                &buffers[MM_PAN_EVEN_BIAS_A0_PARAMETER][band_index][sca_index].
                data_pointer,
                &buffers[MM_PAN_EVEN_PRE_SHUTTER][band_index][sca_index].
                number_values );
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Reading BPF bias model parameters for PAN_EVEN "
                              "band");
                return NULL;
            }    

            /* Set the number of values for the rest buffers */
            buffers[MM_PAN_EVEN_POST_SHUTTER][band_index][sca_index].
                number_values = buffers[MM_PAN_EVEN_PRE_SHUTTER][band_index]
                [sca_index].number_values;
            buffers[MM_PAN_EVEN_MEAN_SHUTTER][band_index][sca_index].
                number_values = buffers[MM_PAN_EVEN_PRE_SHUTTER][band_index]
                [sca_index].number_values;
            buffers[MM_PAN_EVEN_BIAS_A1_PARAMETER][band_index][sca_index].
                number_values = 1;
            buffers[MM_PAN_EVEN_BIAS_C1_PARAMETER][band_index][sca_index].
                number_values = 1;
            buffers[MM_PAN_EVEN_BIAS_A0_PARAMETER][band_index][sca_index].
                number_values = 1;
            break;

        case MM_BPF_BACKGROUND_RESPONSE:
            /* This call retrieves all buffers but only sets the number of
               values for the first buffer */
            status = read_tirs_bpf_bias_model_parameters( band_number, 
                sca_number, &buffers[MM_BPF_BACKGROUND_RESPONSE][band_index]
                [sca_index].data_pointer, 
                &buffers[MM_BPF_BACKGROUND_RESPONSE][band_index][sca_index].
                number_values );
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Reading TIRS BPF bias model parameters");
                return NULL;
            }    
            break;

        case MM_3Q_RESPONSE_LINEARIZATION_PARAMS:
            status = get_rlut_linearization_parameters( buffer_type,
                band_number, sca_number );
            break;

        /* Get the lookup linearization RLUT parameters for linearity */
        case MM_LOOKUP_RESPONSE_LINEARIZATION_PARAMS:
            status = get_rlut_lookup_linearization_parameters(
                    buffer_type, band_number, sca_number );
            break;

        /* Get the TIRS secondary linearization parameters */
        case MM_TIRS_SECONDARY_LINEARIZATION_PARAMS :
            /* No need to get or check the return status since the pointer
               returned to the caller will be NULL */
            get_rlut_tirs_secondary_linearization_parameters( buffer_type,
                band_number, sca_number );
            break;

        /* Allocate memory for perframe bias by default case below for 
           use in the OLI bias determination algorithm */
        case MM_NON_PAN_PERFRAME_BIAS:
        case MM_PAN_PERFRAME_BIAS:

        /* Allocate memory buffer by default case below for use in
           the Residual Striping Correction API code */
        case MM_SCENE_STRIPING_CORRECTION_MATRIX:

        /* Allocate memory buffer by default case below for use in
           the temperature_sensitivity_correction API code */
        case MM_TEMP_CORRECTION_FACTOR :

        /* Allocate memory buffers by default case below for use in
           the sca_discontinuity_correction API code */
        case MM_SCA_CORRECTION_FACTOR :

        default:
            buffers[buffer_type][band_index][sca_index].data_pointer =
                malloc( buffers[buffer_type][band_index][sca_index].size );
    }

    /* Indicate the buffer has been allocated to help detect bugs */
    buffers[buffer_type][band_index][sca_index].buffer_allocated = 1;

    /* Check to see if the buffer is valid for consumers.  This check can't
       happen in the get_memory_buffer routine since there are buffers that
       are filled by the routines called by this internal routine.  Such
       buffers as the IMAGE_L0R, and BPF buffers are among them.  */
    if ( consumer_type == CONSUMER && get_buffer_filled_flag( buffer_type,
        band_index, sca_index ) == 0 )
    {
        IAS_LOG_ERROR("User type is CONSUMER however buffer %s hasn't "
            "been filled", get_memory_buffer_string( buffer_type ));
        return NULL;
    }

    return buffers[buffer_type][band_index][sca_index].data_pointer;
}  /* END internal_get_memory_buffer */



/*------------------------------------------------------------------------
NAME:    get_memory_buffer

PURPOSE: Routine to get a pointer to a memory buffer after the consumer
         counts have been set.  It calls the internal routine to get the
         buffer. It is the caller's responsibility to cast the void pointer
         to the correct type.

RETURNS: pointer to the buffer
         NULL on error
------------------------------------------------------------------------*/
void *get_memory_buffer
(
    MEMORY_BUFFER_TYPE buffer_type, /* I: Index into static buffer arrays */
    MEMORY_USER_TYPE consumer_type, /* I: The type of user:
                                          CONSUMER - the caller uses the
                                            data so the buffer must be valid
                                          SETTER - the caller will fill the
                                            buffer and set the buffer filled
                                            flag */
    int band_index,                 /* I: */
    int sca_index                   /* I: */
)
{
    if ( ! set_consumer_counts_called_flag )
    {
        IAS_LOG_ERROR("Attempting to get buffer %s before the consumer "
            "counts have been set", get_memory_buffer_string( buffer_type ));
        return NULL;
    }

    /* Also verify that the consumer count specifically for this buffer has
       been set to catch errors with buffer management */
    if ( buffers[buffer_type][band_index][sca_index].consumer_set == 0 )
    {
        IAS_LOG_ERROR("Band number %d, SCA number %d: Buffer %s consumer "
            "count hasn't been set.",
            get_band_attributes(band_index)->band_number,  sca_index + 1,
            get_memory_buffer_string( buffer_type ));
        return NULL;
    }

    return internal_get_memory_buffer( buffer_type, consumer_type,
        band_index, sca_index );
}  /* END get_memory_buffer */


/*------------------------------------------------------------------------
NAME:    preprocess_get_memory_buffer

PURPOSE: Routine to get a pointer to a memory buffer after the consumer
         counts have been set.  It calls the internal routine to get the
         buffer. It is the caller's responsibility to cast the void pointer
         to the correct type.

RETURNS: pointer to the buffer
------------------------------------------------------------------------*/
void *preprocess_get_memory_buffer
(
    MEMORY_BUFFER_TYPE buffer_type, /* I: Index into static buffer arrays */
    int band_index,                 /* I: */
    int sca_index                   /* I: */
)
{
    if ( set_consumer_counts_called_flag )
    {
        IAS_LOG_ERROR("Attempting to get buffer %s using pre-process get "
            "routine after the consumer " "counts being set",
            get_memory_buffer_string( buffer_type ));
        return NULL;
    }

    /* Calling from pre-process is considered as a setter routine even if
       the data isn't actually set -- in most, if not all, the data will be
       set. */
    return internal_get_memory_buffer( buffer_type, SETTER, band_index,
        sca_index );
}  /* END preprocess_get_memory_buffer */


/*------------------------------------------------------------------------
NAME:    get_memory_buffer_number_values

PURPOSE: Public routine to access the number of values in a buffer.  This 
         is provided as a convenience for those buffers that have unknown
         size.  A good example is number of telemetry lines.

RETURNS: number of values in the buffer
------------------------------------------------------------------------*/
int get_memory_buffer_number_values
(
    MEMORY_BUFFER_TYPE buffer_type, /* I: Index into static buffer arrays */
    int band_index,                 /* I: */
    int sca_index                   /* I: */
)
{
    return buffers[buffer_type][band_index][sca_index].number_values;
}  /* END get_memory_buffer_number_values */


/*------------------------------------------------------------------------
NAME:    set_buffer_filled_flag

PURPOSE: Public routine to set a flag indicating the buffer has been filled
         and is therefore safe to use.

RETURNS: none
------------------------------------------------------------------------*/
void set_buffer_filled_flag
(
    MEMORY_BUFFER_TYPE buffer_type, /* I: Index into static buffer arrays */
    int band_index,                 /* I: */
    int sca_index                   /* I: */
)
{
    buffers[buffer_type][band_index][sca_index].buffer_filled = 1;
}  /* END set_buffer_filled_flag */


/*------------------------------------------------------------------------
NAME:    get_buffer_filled_flag

PURPOSE: Public routine to get the flag indicating the buffer has been filled
         and therefore safe to use.

RETURNS: 1 - the buffer has been filled
         0 - the buffer is uninitialized
------------------------------------------------------------------------*/
int get_buffer_filled_flag
(
    MEMORY_BUFFER_TYPE buffer_type, /* I: Index into static buffer arrays */
    int band_index,                 /* I: */
    int sca_index                   /* I: */
)
{
    return buffers[buffer_type][band_index][sca_index].buffer_filled;
}  /* END get_buffer_filled_flag */



/*------------------------------------------------------------------------
   Private CPF routines ...
   Since the CPF getter routines aren't thread safe the pre-process
   routines will call the routines that follow to load up the memory
   management pointers with pointers into the CPF data structure.

RETURNS: each routine returns SUCCESS or ERROR
------------------------------------------------------------------------*/
static int get_cpf_reflectance_conversion( void )
{
    const struct IAS_CPF_REFLECTANCE_CONVERSION *reflectance_conversion;

    int max_num_scas = get_blob_max_nscas();
    int band;
    int sca;

    reflectance_conversion = ias_cpf_get_reflect_conv( get_blob_cpf() );
    if ( !reflectance_conversion )
    {
        IAS_LOG_ERROR("Getting the CPF reflectance coefficients");
        return ERROR;
    }
    for ( band = 0; band < total_bands; band++ )
    {
        for ( sca = 0; sca < max_num_scas; sca++ )
        {
            buffers[MM_REFLECTANCE_CONVERSION_COEFFICIENTS][band][sca].
                is_const_data = 1;
            buffers[MM_REFLECTANCE_CONVERSION_COEFFICIENTS][band][sca].
                const_data_pointer = 
                &reflectance_conversion->reflect_conv_coeff[band];
            buffers[MM_REFLECTANCE_CONVERSION_COEFFICIENTS][band][sca].
                buffer_allocated = 1;
        }
    }

    return SUCCESS;
}  /* END get_cpf_reflectance_conversion */

static int get_cpf_temp_sensitivity_reference_temp( void )
{
    const struct IAS_CPF_TEMP_SENSITIVITY *temp_sensitivity;

    int max_num_scas = get_blob_max_nscas();
    int band_list[IAS_MAX_NBANDS];
    int num_bands;
    int status;
    int band;
    int sca;

    temp_sensitivity = ias_cpf_get_temp_sensitivity( get_blob_cpf() );
    if ( !temp_sensitivity )
    {
        IAS_LOG_ERROR("Getting the CPF temperature sensitivity reference "
            "temperature");
        return ERROR;
    }

    /* Get a list of the OLI and TIRS normal bands */
    status = ias_sat_attr_get_sensor_band_numbers( IAS_MAX_SENSORS, 
        IAS_NORMAL_BAND, IAS_SECONDARY_BAND, band_list, IAS_MAX_NBANDS, 
        &num_bands);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving the list of normal bands");
        return ERROR;
    }

    for ( band = 0; band < num_bands; band++ )
    {
        int band_index = ias_sat_attr_convert_band_number_to_index(
            band_list[band] );
        if ( band_index == ERROR )
        {
            IAS_LOG_ERROR("Getting index for band %d", band_list[band] );
            return ERROR;
        }
        RPS_BAND_ATTRIBUTES *band_attributes = 
                                get_band_attributes( band_index );

        for ( sca = 0; sca < max_num_scas; sca++ )
        {
            buffers[MM_TEMP_SENSITIVITY_REFERENCE_TEMP][band_index][sca].
                is_const_data = 1;
            if (band_attributes->spectral_type == IAS_SPECTRAL_THERMAL)
            {
                buffers[MM_TEMP_SENSITIVITY_REFERENCE_TEMP][band_index][sca].
                    const_data_pointer = &temp_sensitivity->tirs_reference_temp;
            }
            else
            {
                buffers[MM_TEMP_SENSITIVITY_REFERENCE_TEMP][band_index][sca].
                    const_data_pointer = &temp_sensitivity->oli_reference_temp;
            }
            buffers[MM_TEMP_SENSITIVITY_REFERENCE_TEMP][band_index][sca].
                buffer_allocated = 1;
        }
    }
    return SUCCESS;
}  /* END get_cpf_temp_sensitivity_reference_temp */

static int get_cpf_temp_sensitivity_coefficient( void )
{
    const struct IAS_CPF_TEMP_SENSITIVITY *temp_sensitivity;
    int max_num_scas = get_blob_max_nscas();
    int band_list[IAS_MAX_NBANDS];
    int num_bands;
    int status;
    int band;
    int sca;

    temp_sensitivity = ias_cpf_get_temp_sensitivity( get_blob_cpf() );
    if ( !temp_sensitivity )
    {
        IAS_LOG_ERROR("Getting the CPF temperature sensitivity "
            "coefficients");
        return ERROR;
    }

    /* Get a list of the OLI and TIRS normal bands */
    status = ias_sat_attr_get_sensor_band_numbers( IAS_MAX_SENSORS, 
        IAS_NORMAL_BAND, IAS_SECONDARY_BAND, band_list, IAS_MAX_NBANDS, 
        &num_bands);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving the list of normal bands");
        return ERROR;
    }

    for ( band = 0; band < num_bands; band++ )
    {
        int band_index = ias_sat_attr_convert_band_number_to_index(
            band_list[band] );
        if ( band_index == ERROR )
        {
            IAS_LOG_ERROR("Getting index for band %d", band_list[band] );
            return ERROR;
        }

        for ( sca = 0; sca < max_num_scas; sca++ )
        {
            buffers[MM_TEMP_SENSITIVITY_COEFFICIENT][band_index][sca].
                is_const_data = 1;
            buffers[MM_TEMP_SENSITIVITY_COEFFICIENT][band_index][sca].
                const_data_pointer =
                &temp_sensitivity->temp_sensitivity_coeff[band_index][sca][0];
            buffers[MM_TEMP_SENSITIVITY_COEFFICIENT][band_index][sca].
                buffer_allocated = 1;
        }
    }
    return SUCCESS;
}  /* END get_cpf_temp_sensitivity_coefficient */

static int get_cpf_absolute_gains( void )
{
    const struct IAS_CPF_ABSOLUTE_GAINS *absolute_gains;

    int max_num_scas = get_blob_max_nscas();
    int band;
    int sca;

    absolute_gains = ias_cpf_get_abs_gains( get_blob_cpf() );
    if ( !absolute_gains )
    {
        IAS_LOG_ERROR("Getting the CPF absolute gains");
        return ERROR;
    }
    for ( band = 0; band < total_bands; band++ )
    {
        for ( sca = 0; sca < max_num_scas; sca++ )
        {
            buffers[MM_ABSOLUTE_GAINS][band][sca].is_const_data = 1;
            buffers[MM_ABSOLUTE_GAINS][band][sca].const_data_pointer =
                &absolute_gains->gain[band][sca];
            buffers[MM_ABSOLUTE_GAINS][band][sca].buffer_allocated = 1;
        }
            
    }
    return SUCCESS;
}  /* END get_cpf_absolute_gains */

static int get_cpf_tirs_absolute_gains_blind( void )
{
    const struct IAS_CPF_ABSOLUTE_GAINS *blind_absolute_gains;

    int tirs_band_list[IAS_MAX_NBANDS];
    int num_blind_bands = 0;
    int number_scas = 0;
    int band_index = 0;
    int index = 0;
    int band_number = 0;
    int status = 0;
    int sca = 0;

    /* Get the number of blind bands and the corresponding list */
    memset(tirs_band_list, 0, IAS_MAX_NBANDS * sizeof(int));
    status = ias_sat_attr_get_sensor_band_numbers(IAS_TIRS, IAS_BLIND_BAND,
        IAS_SECONDARY_BAND, tirs_band_list, IAS_MAX_NBANDS,
        &num_blind_bands);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Getting the CPF TIRS absolute gains");
        return ERROR;
    }

    /* Get the total number of TIRS SCAs */
    number_scas = ias_sat_attr_get_sensor_sca_count(IAS_TIRS);
    if (number_scas == ERROR)
    {
        IAS_LOG_ERROR("Getting sensor SCA count for TIRS");
        return ERROR;
    }

    /* GET the TIRS_ABSOLUTE_GAINS_BLIND group data  */
    blind_absolute_gains = ias_cpf_get_tirs_abs_gains_blind( get_blob_cpf());
    if (blind_absolute_gains == NULL)
    {
        IAS_LOG_ERROR("Getting the CPF TIRS blind band absolute gains");
        return ERROR;
    }
 
    /* Get the proper pointers */
    for (index = 0; index < num_blind_bands; index++)
    {
        band_number = tirs_band_list[index];
        band_index = ias_sat_attr_convert_band_number_to_index(band_number);
        if (band_index == ERROR)
        {
            IAS_LOG_ERROR("Converting band number %d to index", band_number );
            return ERROR;
        }

        for (sca = 0; sca < number_scas; sca++)
        {
            buffers[MM_TIRS_ABSOLUTE_GAINS_BLIND][band_index][sca].
                is_const_data = 1;
            buffers[MM_TIRS_ABSOLUTE_GAINS_BLIND][band_index][sca].
                const_data_pointer = &blind_absolute_gains->gain[index][sca];
            buffers[MM_TIRS_ABSOLUTE_GAINS_BLIND][band_index][sca].
                buffer_allocated = 1;
        }
    }

    return SUCCESS;
}  /* END get_cpf_tirs_absolute_gains_blind */

static int get_cpf_relative_gains( void )
{
    const struct IAS_CPF_RELATIVE_GAINS *relative_gains;

    int max_num_scas = get_blob_max_nscas();
    int band;
    int sca;

    relative_gains = ias_cpf_get_relative_gains( get_blob_cpf() );
    if ( !relative_gains )
    {
        IAS_LOG_ERROR("Getting the CPF relative gains");
        return ERROR;
    }
    for ( band = 0; band < total_bands; band++ )
    {
        for ( sca = 0; sca < max_num_scas; sca++ )
        {
            buffers[MM_RELATIVE_GAINS][band][sca].is_const_data = 1;
            buffers[MM_RELATIVE_GAINS][band][sca].const_data_pointer =
                &relative_gains->per_detector[band][sca][0];
            buffers[MM_RELATIVE_GAINS][band][sca].buffer_allocated = 1;
        }
            
    }
    return SUCCESS;
}  /* END get_cpf_relative_gains */

static int get_cpf_tirs_relative_gains_blind( void )
{
    const struct IAS_CPF_RELATIVE_GAINS *blind_relative_gains;

    int number_scas;
    int index;
    int band_index;
    int band_number;
    int sca;
    int num_blind_bands;
    int tirs_band_list[IAS_MAX_NBANDS];
    int status;


    /* Get the index value into the band array */
    status = ias_sat_attr_get_sensor_band_numbers(IAS_TIRS, IAS_BLIND_BAND,
        IAS_SECONDARY_BAND, tirs_band_list, IAS_MAX_NBANDS,
        &num_blind_bands);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Getting the TIRS blind band numbers");
        return ERROR;
    }

    /* Get the total number of TIRS SCAs */
    number_scas = ias_sat_attr_get_sensor_sca_count(IAS_TIRS);
    if (number_scas == ERROR)
    {
        IAS_LOG_ERROR("Getting the TIRS sensor SCA count");
        return ERROR;
    }

    /* GET the TIRS_RELATIVE_GAINS group data  */
    blind_relative_gains = ias_cpf_get_tirs_rel_gains_blind( get_blob_cpf() );
    if (blind_relative_gains == NULL)
    {
        IAS_LOG_ERROR("Getting the CPF TIRS blind band relative gains");
        return ERROR;
    }
 
    /* Get the proper pointers */
    for (index = 0; index < num_blind_bands; index++)
    {
        band_number = tirs_band_list[index];
        band_index = ias_sat_attr_convert_band_number_to_index(band_number);
        if (band_index == ERROR)
        {
            IAS_LOG_ERROR("Converting band number %d to index", band_number );
            return ERROR;
        }

        for (sca = 0; sca < number_scas; sca++)
        {
            buffers[MM_TIRS_RELATIVE_GAINS_BLIND][band_index][sca].
                is_const_data = 1;
            buffers[MM_TIRS_RELATIVE_GAINS_BLIND][band_index][sca].
                const_data_pointer = &(blind_relative_gains->
                per_detector[index][sca][0]);
            buffers[MM_TIRS_RELATIVE_GAINS_BLIND][band_index][sca].
                buffer_allocated = 1;
        }
    }

    return SUCCESS;
}  /* END get_cpf_tirs_relative_gains_blind */

static int get_cpf_lamp_radiance(MEMORY_BUFFER_TYPE buffer_type)
{
    const struct IAS_CPF_LAMP_RADIANCE *lamp_radiance;

    int max_num_scas = get_blob_max_nscas();
    int band;
    int sca;

    lamp_radiance = ias_cpf_get_lamp_radiance( get_blob_cpf() );
    if (!lamp_radiance)
    {
        IAS_LOG_ERROR("Getting the CPF lamp radiance");
        return ERROR;
    }
    for ( band = 0; band < total_bands; band++ )
    {
        for ( sca = 0; sca < max_num_scas; sca++ )
        {
            /* Set the data type as const */
            buffers[buffer_type][band][sca].is_const_data = 1;
            buffers[buffer_type][band][sca].buffer_allocated = 1;

            switch ( buffer_type )
            {
                case MM_WORKING_LAMP_RADIANCE:
                    buffers[buffer_type][band][sca].const_data_pointer =
                        &lamp_radiance->effective_rad_working[band][sca][0];
                    break;
                case MM_BACKUP_LAMP_RADIANCE:
                    buffers[buffer_type][band][sca].const_data_pointer =
                        &lamp_radiance->effective_rad_backup[band][sca][0];
                    break;
                case MM_PRISTINE_LAMP_RADIANCE:
                    buffers[buffer_type][band][sca].const_data_pointer =
                        &lamp_radiance->effective_rad_pristine[band][sca][0];
                    break;

                default:
                    break; /* No default action -- here for compiler warning */
            }
        }
    }

    return SUCCESS;
}  /* END get_cpf_lamp_radiance */

static int get_cpf_focal_plane( MEMORY_BUFFER_TYPE buffer_type )
{
    const struct IAS_CPF_FOCAL_PLANE *focal_plane;
    int max_num_scas = get_blob_max_nscas();
    int band;
    int sca;

    focal_plane = ias_cpf_get_focal_plane( get_blob_cpf() );
    if ( !focal_plane )
    {
        IAS_LOG_ERROR("Getting the CPF focal plane parameters");
        return ERROR;
    }

    for ( band = 0; band < total_bands; band++ )
    {
        for ( sca = 0; sca < max_num_scas; sca++ )
        {
            /* Set the data type as const */
            buffers[buffer_type][band][sca].is_const_data = 1;
            buffers[buffer_type][band][sca].buffer_allocated = 1;

            switch ( buffer_type )
            {
                case MM_SCA_OVERLAP :
                    buffers[buffer_type][band][sca].const_data_pointer =
                        &focal_plane->sca_overlap[band];
                    break;

                case MM_CPF_SCA_OFFSET :
                    buffers[buffer_type][band][sca].const_data_pointer =
                        &focal_plane->sca_offset[band][sca];
                    break;

                default:
                    break; /* No default action -- here for compiler warning */
            }
        }
    }

    return SUCCESS;
}  /* END get_cpf_focal_plane */

static int get_cpf_inoperable_detectors( )
{
    const struct IAS_CPF_DETECTOR_STATUS *detector_status;
    int max_num_scas = get_blob_max_nscas();
    int band_index;
    int sca;

    detector_status = ias_cpf_get_det_status( get_blob_cpf() );
    if ( !detector_status )
    {
        IAS_LOG_ERROR("Getting the CPF detector status");
        return ERROR;
    }

    for ( band_index = 0; band_index < total_bands; band_index++ )
    {
        for ( sca = 0; sca < max_num_scas; sca++ )
        {
            buffers[MM_INOPERABLE_DETECTORS][band_index][sca].is_const_data = 
                1;
            buffers[MM_INOPERABLE_DETECTORS][band_index][sca].
                const_data_pointer =
                &detector_status->inoperable[band_index][sca][0];
            buffers[MM_INOPERABLE_DETECTORS][band_index][sca].
                number_values = detector_status->
                inoperable_count[band_index][sca];
            buffers[MM_INOPERABLE_DETECTORS][band_index][sca].
                buffer_allocated = 1;
        }
    }

    return SUCCESS;
}  /* END get_cpf_inoperable_detectors */

static int get_cpf_bias( MEMORY_BUFFER_TYPE buffer_type )
{
    const struct IAS_CPF_AVERAGE_BIAS *oli_avg_bias;
    int status;
    int band_list[IAS_MAX_NBANDS];
    int number_of_bands = 0;
    int nbands_returned;
    int i;

    oli_avg_bias = ias_cpf_get_oli_avg_bias( get_blob_cpf() );
    if ( !oli_avg_bias )
    {
        IAS_LOG_ERROR("Getting the CPF OLI average bias");
        return ERROR;
    }

    status = ias_sat_attr_get_sensor_band_numbers( IAS_OLI, IAS_NORMAL_BAND,
        0, &band_list[0], IAS_MAX_NBANDS, &nbands_returned );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Getting the OLI normal band list");
        return ERROR;
    }
    number_of_bands += nbands_returned;

    for ( i = 0; i < number_of_bands; i++ )
    {
        int band_number = band_list[i];
        int band_index = ias_sat_attr_convert_band_number_to_index(
            band_number );
        if ( band_index == ERROR )
        {
            IAS_LOG_ERROR("Getting the band index for band %d", band_number );
        }
        int sca_index;
        RPS_BAND_ATTRIBUTES *band_attributes =
            get_band_attributes( band_index );

        for ( sca_index = 0; sca_index < band_attributes->number_scas;
            sca_index++ )
        {
            /* Set the data type as const */
            buffers[buffer_type][band_index][sca_index].is_const_data = 1;
            buffers[buffer_type][band_index][sca_index].buffer_allocated = 1;

            switch ( buffer_type )
            {
                case MM_NON_PAN_CPF_BIAS:
                    if (band_attributes->spectral_type == IAS_SPECTRAL_SWIR)
                    { 
                        buffers[buffer_type][band_index][sca_index].
                            const_data_pointer =
                            &oli_avg_bias->bias_swir[band_index][sca_index][0];
                    }
                    if (band_attributes->spectral_type == IAS_SPECTRAL_VNIR)
                    { 
                        buffers[buffer_type][band_index][sca_index].
                            const_data_pointer =
                            &oli_avg_bias->bias_vnir[band_index][sca_index][0];
                    }
                    break;
                case MM_PAN_ODD_CPF_BIAS:
                    if (band_attributes->spectral_type == IAS_SPECTRAL_PAN)
                    { 
                        buffers[buffer_type][band_index][sca_index].
                            const_data_pointer =
                            &oli_avg_bias->bias_odd_pan[sca_index][0];
                    }
                    break;
                case MM_PAN_EVEN_CPF_BIAS:
                    if (band_attributes->spectral_type == IAS_SPECTRAL_PAN)
                    { 
                        buffers[buffer_type][band_index][sca_index].
                            const_data_pointer =
                            &oli_avg_bias->bias_even_pan[sca_index][0];
                    }
                    break;
                default:
                    break; /* No default action -- here for compiler warning */
            }  /* END switch */
        }  /* END for sca loop */
    }  /* END for band loop */

    return SUCCESS;
}  /* END get_cpf_bias */

static int get_cpf_tirs_background_response( void )
{
    RPS_BAND_ATTRIBUTES *band_atts;
    const struct IAS_CPF_TIRS_DETECTOR_RESPONSE *tirs_det_response;
    int band;
    int sca;
    int band_number;
    int tirs_band_list[IAS_MAX_NBANDS];
    int number_of_bands;
    int nbands_returned;
    int status;

    /* Get a list of bands to process -- get the normal bands */
    number_of_bands = 0;
    status = ias_sat_attr_get_sensor_band_numbers( IAS_TIRS, IAS_NORMAL_BAND,
        0, /* Don't exclude any normal bands */
        &tirs_band_list[0], IAS_MAX_NBANDS, &nbands_returned );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Getting the TIRS normal band list");
        return ERROR;
    }
    number_of_bands += nbands_returned;

    /* Add the TIRS blind band */
    status = ias_sat_attr_get_sensor_band_numbers( IAS_TIRS, IAS_BLIND_BAND,
        IAS_SECONDARY_BAND, /* Exclude the redundant band */
        &tirs_band_list[number_of_bands], IAS_MAX_NBANDS - number_of_bands,
        &nbands_returned );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Getting the TIRS blind band numbers");
        return ERROR;
    }
    number_of_bands += nbands_returned;

    tirs_det_response = ias_cpf_get_tirs_det_response(get_blob_cpf());
    if ( !tirs_det_response )
    {
        IAS_LOG_ERROR("Getting the CPF TIRS detector response");
        return ERROR;
    }
    for ( band = 0; band < number_of_bands; band++ )
    {
        /* Get the band attributes */
        band_number = tirs_band_list[band];
        int band_index = ias_sat_attr_convert_band_number_to_index(
            band_number );
        if ( band_index == ERROR )
        {
            IAS_LOG_ERROR("Getting index for band %d", band_number );
            return ERROR;
        }
        band_atts = get_band_attributes( band_index );

        for ( sca = 0; sca < band_atts->number_scas; sca++ )
        {
            /* Must use the band_index based on band number to access the
               MM buffers however the CPF memory is laid out starting at an
               index of zero so use the loop counter band to access the
               CPF background response */
            buffers[MM_TIRS_BACKGROUND_RESPONSE][band_index][sca].
                is_const_data = 1;
            buffers[MM_TIRS_BACKGROUND_RESPONSE][band_index][sca].
                const_data_pointer =
                &tirs_det_response->background_response[band][sca][0];
            buffers[MM_TIRS_BACKGROUND_RESPONSE][band_index][sca].
                buffer_allocated = 1;
        }            
    }

    return SUCCESS;
}  /* END get_cpf_tirs_background_response */

static int get_cpf_tirs_gain_offsets( void )
{
    RPS_BAND_ATTRIBUTES *band_atts;
    const struct IAS_CPF_TIRS_DETECTOR_RESPONSE *tirs_det_response;
    int band;
    int sca;
    int band_number;
    int tirs_band_list[IAS_MAX_NBANDS];
    int number_of_bands;
    int nbands_returned;
    int status;

    /* Get a list of bands to process -- get the normal bands */
    number_of_bands = 0;
    status = ias_sat_attr_get_sensor_band_numbers( IAS_TIRS, IAS_NORMAL_BAND,
        0, /* Don't exclude any normal bands */
        &tirs_band_list[0], IAS_MAX_NBANDS, &nbands_returned );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Getting the TIRS normal band list");
        return ERROR;
    }
    number_of_bands += nbands_returned;

    /* Add the TIRS blind band */
    status = ias_sat_attr_get_sensor_band_numbers( IAS_TIRS, IAS_BLIND_BAND,
        IAS_SECONDARY_BAND, /* Exclude the redundant band */
        &tirs_band_list[number_of_bands], IAS_MAX_NBANDS - number_of_bands,
        &nbands_returned );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Getting the TIRS blind band numbers");
        return ERROR;
    }
    number_of_bands += nbands_returned;

    tirs_det_response = ias_cpf_get_tirs_det_response(get_blob_cpf());
    if ( !tirs_det_response )
    {
        IAS_LOG_ERROR("Getting the CPF TIRS detector response");
        return ERROR;
    }
    for ( band = 0; band < number_of_bands; band++ )
    {
        /* Get the band attributes */
        band_number = tirs_band_list[band];
        int band_index = ias_sat_attr_convert_band_number_to_index(
            band_number );
        if ( band_index == ERROR )
        {
            IAS_LOG_ERROR("Getting index for band %d", band_number );
            return ERROR;
        }
        band_atts = get_band_attributes( band_index );

        for ( sca = 0; sca < band_atts->number_scas; sca++ )
        {
            /* Must use the band_index based on band number to access the
               MM buffers however the CPF memory is laid out starting at an
               index of zero so use the loop counter band to access the
               CPF background response */
            buffers[MM_TIRS_GAIN_OFFSETS][band_index][sca].
                is_const_data = 1;
            buffers[MM_TIRS_GAIN_OFFSETS][band_index][sca].
                const_data_pointer =
                &tirs_det_response->gain_offsets[band][sca][0];
            buffers[MM_TIRS_GAIN_OFFSETS][band_index][sca].
                buffer_allocated = 1;
        }            
    }

    return SUCCESS;
}  /* END get_cpf_tirs_gain_offsets */

static int get_cpf_impulse_noise( MEMORY_BUFFER_TYPE buffer_type )
{
    const IAS_CPF_IMPULSE_NOISE *impulse_noise;
    const int *ptr_impulse_noise;
    int band;
    int sca;
    int band_list[IAS_MAX_NBANDS];
    int num_bands;
    int status;

    /* Get a list of the normal bands */
    status = ias_sat_attr_get_sensor_band_numbers( IAS_MAX_SENSORS,
        IAS_NORMAL_BAND, IAS_SECONDARY_BAND, band_list, IAS_MAX_NBANDS,
        &num_bands);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Retrieving the list of normal bands");
        return ERROR;
    }

    /* Read the impulse noise from the CPF */
    impulse_noise = ias_cpf_get_impulse_noise( get_blob_cpf() );
    if ( !impulse_noise )
    {
        IAS_LOG_ERROR("Getting the CPF impulse noise");
        return ERROR;
    }

    /* Point to the right location in the impulse noise for the buffer type */
    if ( buffer_type == MM_IMPULSE_NOISE_LIMIT )
        ptr_impulse_noise = impulse_noise->in_limit;
    else 
        ptr_impulse_noise = impulse_noise->median_filter_width;

    /* Initialize the buffers for the bands that will be processed */
    for ( band = 0; band < mm_number_of_bands; band++ )
    {
        int band_index = mm_list_of_band_indices[band];
        for ( sca = 0; sca < IAS_MAX_NSCAS; sca++ )
        {
            buffers[buffer_type][band_index][sca].is_const_data = 1;
            buffers[buffer_type][band_index][sca].const_data_pointer = 
                &ptr_impulse_noise[band_index];
            buffers[buffer_type][band_index][sca].buffer_allocated = 1;
        }
    }

    return SUCCESS;
}  /* END get_cpf_impulse_noise */

static int get_cpf_detector_noise( void )
{
    const struct IAS_CPF_DETECTOR_NOISE *detector_noise;
    int band;
    int sca;
    int band_list[IAS_MAX_NBANDS];
    int num_bands;
    int status;

    /* Get a list of the normal bands */
    status = ias_sat_attr_get_sensor_band_numbers( IAS_MAX_SENSORS,
        IAS_NORMAL_BAND, IAS_SECONDARY_BAND, band_list, IAS_MAX_NBANDS,
        &num_bands);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Retrieving the list of normal bands");
        return ERROR;
    }

    
    detector_noise = ias_cpf_get_det_noise( get_blob_cpf() );
    if ( !detector_noise )
    {
        IAS_LOG_ERROR("Getting the CPF detector noise");
        return ERROR;
    }

    /* Initialize the buffers for the bands that will be processed */
    for ( band = 0; band < mm_number_of_bands; band++ )
    {
        int band_index = mm_list_of_band_indices[band];
        for ( sca = 0; sca < IAS_MAX_NSCAS; sca++ )
        {
            buffers[MM_DETECTOR_NOISE][band_index][sca].is_const_data = 1;
            buffers[MM_DETECTOR_NOISE][band_index][sca].const_data_pointer = 
                &detector_noise->per_detector[band_index][sca][0];
            buffers[MM_DETECTOR_NOISE][band_index][sca].buffer_allocated = 1;
        }
    }

    return SUCCESS;
}  /* END get_cpf_detector_noise */

static int get_cpf_histogram_char_parameters(MEMORY_BUFFER_TYPE buffer_type)
{
    const struct IAS_CPF_HISTOGRAM_CHAR *hist_char = NULL;
    int *num_frames_to_skip = NULL;
    int index;                     /* Index to current band */


    /* Get the CPF parameters.  There are only two values in this group
       currently--it's assumed that the number of lines to skip at the
       top and/or bottom will be the same for all bands/SCAs.  We may
       need to revisit that assumption in a future release... */
    hist_char = ias_cpf_get_histogram_char(get_blob_cpf());
    if (hist_char == NULL)
    {
        IAS_LOG_ERROR("Getting histogram characterization parameters from "
            "CPF");
        return ERROR;
    }

    /* Select the "top" or "bottom" number of frames to skip depending
       on the buffer type */
    if (buffer_type == MM_HIST_FRAMES_TO_SKIP_TOP)
        num_frames_to_skip = (int *)&(hist_char->frames_to_skip_top);
    else
        num_frames_to_skip = (int *)&(hist_char->frames_to_skip_bottom);

    /* Get the values assigned to all bands/SCAs */
    for (index = 0; index < total_bands; index++)
    {
        const RPS_BAND_ATTRIBUTES *band_atts = get_band_attributes(index);
        int num_scas  = band_atts->number_scas;
        int sca_index;
        for (sca_index = 0; sca_index < num_scas; sca_index++)
        {
            buffers[buffer_type][index][sca_index].is_const_data = 1;
            buffers[buffer_type][index][sca_index].const_data_pointer =
                num_frames_to_skip;
            buffers[buffer_type][index][sca_index].buffer_allocated = 1;
        }
    }

    return SUCCESS;
}   /* END -- get_cpf_histogram_char_parameters */

static int get_cpf_sca_parameters( MEMORY_BUFFER_TYPE buffer_type )
{
    RPS_BAND_ATTRIBUTES *band_atts;
    int band_number;
    int band_index;
    int sca;
    int band_list[IAS_MAX_NBANDS];
    int num_bands;
    int status;

    /* Get the SCA_PARAMETERS group data */
    const struct IAS_CPF_SCA_PARAMETERS *sca_parameters = 
        ias_cpf_get_sca_parameters( get_blob_cpf() );
    if ( !sca_parameters )
    {
        IAS_LOG_ERROR("Getting the SCA parameters from the CPF");
        return ERROR;
    }

    /* Get a list of the normal bands */
    status = ias_sat_attr_get_sensor_band_numbers( IAS_MAX_SENSORS,
        IAS_NORMAL_BAND, IAS_SECONDARY_BAND, band_list, IAS_MAX_NBANDS,
        &num_bands);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Retrieving the list of normal bands");
        return ERROR;
    }

    for ( band_index = 0; band_index < num_bands; band_index++ )
    {
        band_number = band_list[band_index];
        int band_index = ias_sat_attr_convert_band_number_to_index(
            band_number );
        if ( band_index == ERROR )
        {
            IAS_LOG_ERROR("Getting index for band %d", band_number );
            return ERROR;
        }
        band_atts = get_band_attributes( band_index );

        for ( sca = 0; sca < band_atts->number_scas; sca++ )
        {
            buffers[buffer_type][band_index][sca].is_const_data = 1;
            buffers[buffer_type][band_index][sca].buffer_allocated = 1;

            switch ( buffer_type )
            {
                case MM_SCA_CORRELATION_SHIFT :
                    buffers[buffer_type][band_index][sca].const_data_pointer =
                        &sca_parameters->
                        max_valid_correlation_shift[band_index];
                    break;

                case MM_SCA_VALID_NEIGHBOR_SEGMENTS :
                    buffers[buffer_type][band_index][sca].const_data_pointer =
                        &sca_parameters->
                        min_valid_neighbor_segments[band_index];
                    break;

                case MM_STRIPING_METRIC_CUTOFF :
                    buffers[buffer_type][band_index][sca].const_data_pointer =
                        &sca_parameters->stripe_cutoff[band_index];
                    break;

                case MM_SCA_OVERLAP_THRESHOLD :
                    buffers[buffer_type][band_index][sca].const_data_pointer =
                        &sca_parameters->sca_overlap_threshold[band_index];
                    break;

                default:
                    break; /* No default action -- here for compiler warning */
            }
        }
    }
    
    return SUCCESS;
}  /* END get_cpf_sca_parameters */

static int get_cpf_discontinuity_ratio( void )
{
    int index;
    MEMORY_BUFFER_TYPE buffer_type = MM_CPF_DISCONTINUITY_RATIO;
    const struct IAS_CPF_SCA_PARAMETERS *sca_parameters =
        ias_cpf_get_sca_parameters( get_blob_cpf() );
    if ( ! sca_parameters )
    {
        IAS_LOG_ERROR("Getting the SCA parameters from the CPF");
        return ERROR;
    }

    for ( index = 0; index < mm_number_of_bands; index++ )
    {
        int band_index = mm_list_of_band_indices[index];
        int sca_index;
        RPS_BAND_ATTRIBUTES *band_atts = get_band_attributes( band_index );
        int number_values = band_atts->number_scas - 1;

        /* Only have data for the normal bands */
        if ( ! band_atts->is_normal_band )
            continue;

        for ( sca_index = 0; sca_index < number_values; sca_index++ )
        {
            buffers[buffer_type][band_index][sca_index].is_const_data = 1;
            buffers[buffer_type][band_index][sca_index].const_data_pointer =
                &sca_parameters->discontinuity_ratio[band_index][sca_index];
            buffers[buffer_type][band_index][sca_index].buffer_allocated = 1;
        }
    }
    return SUCCESS;
}

static int get_sca_correction_factor(void)
{
    int index;
    MEMORY_BUFFER_TYPE buffer_type = MM_SCA_CORRECTION_FACTOR_CPF;

    /* If any of these buffers have been allocated then the rest have as
       well so return */
    for ( index = 0; index < mm_number_of_bands; index++ )
    {
        int i = mm_list_of_band_indices[index];
        if ( buffers[buffer_type][i][0].buffer_allocated )
            return SUCCESS;
    }

    /* Get the SCA_PARAMETERS group data */
    const struct IAS_CPF_SCA_PARAMETERS *sca_parameters = 
        ias_cpf_get_sca_parameters( get_blob_cpf() );
    if ( !sca_parameters )
    {
        IAS_LOG_ERROR("Getting the SCA parameters from the CPF");
        return ERROR;
    }

    for ( index = 0; index < mm_number_of_bands; index++ )
    {
        int status;
        int sca_index;
        double correction_factors[IAS_MAX_NSCAS];
        double discontinuity_ratio[IAS_MAX_NSCAS];
        int band_index = mm_list_of_band_indices[index];
        RPS_BAND_ATTRIBUTES *band_atts = get_band_attributes( band_index );
        int number_scas = band_atts->number_scas;

        /* Only have data for the normal bands */
        if ( ! band_atts->is_normal_band )
            continue;

        /* Copy this band's ratios into local buffer, there are number SCAs
           minus one stored in the CPF hence the sca_index - 1 access of the
           CPF value.  The first ratio is always 1. */
        discontinuity_ratio[0] = 1.0;
        for ( sca_index = 1; sca_index < number_scas; sca_index++ )
        {
            discontinuity_ratio[sca_index] =
                sca_parameters->discontinuity_ratio[band_index][sca_index - 1];
        }
        
        status = calc_normalized_correction_factor( discontinuity_ratio,
            number_scas, correction_factors );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Calculating normalized correction factors");
            return ERROR;
        }

        /* Allocate the memory and copy the ratios */
        for ( sca_index = 0; sca_index < number_scas; sca_index++ )
        {
            double *data_ptr;
            buffers[buffer_type][band_index][sca_index].is_const_data = 0;
            buffers[buffer_type][band_index][sca_index].data_pointer = 
                malloc( buffers[buffer_type][band_index][sca_index].size );
            if (!buffers[buffer_type][band_index][sca_index].data_pointer)
            {
                IAS_LOG_ERROR("Allocating the SCA correction factor for "
                    "Band %d SCA %d", band_atts->band_number, 
                    sca_index + 1 );
                return ERROR;
            }
            buffers[buffer_type][band_index][sca_index].buffer_allocated = 1;

            /* Assign the normalized value */
            data_ptr = buffers[buffer_type][band_index][sca_index].data_pointer;
            data_ptr[0] = correction_factors[sca_index];
        }
    }

    return SUCCESS;
}  /* END get_sca_correction_factor */

static int get_cpf_saturation_levels(  MEMORY_BUFFER_TYPE buffer_type )
{
    RPS_BAND_ATTRIBUTES *band_attributes;
    int sca;
    int band;
    int num_bands;
    int status;
    int band_list[IAS_MAX_NBANDS];
    const struct IAS_CPF_SATURATION_LEVEL *saturation_levels;

    /* Get a list of the normal bands */
    status = ias_sat_attr_get_sensor_band_numbers( IAS_MAX_SENSORS,
        IAS_NORMAL_BAND, IAS_SECONDARY_BAND, band_list, IAS_MAX_NBANDS,
        &num_bands);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Retrieving the list of normal bands");
        return ERROR;
    }

    /* Read the saturation levels from the CPF */
    saturation_levels = ias_cpf_get_saturation_level( get_blob_cpf() );
    if ( !saturation_levels )
    {
        IAS_LOG_ERROR("Getting the CPF saturation levels");
        return ERROR;
    }

    /* Loop through the normal bands only */
    for ( band = 0; band < num_bands; band++ )
    {
        int band_index = ias_sat_attr_convert_band_number_to_index(
            band_list[band] );
        if ( band_index == ERROR )
        {
            IAS_LOG_ERROR("Getting index for band %d", band_list[band] );
            return ERROR;
        }

        band_attributes = get_band_attributes( band_index );
        int vrp_band = band_attributes->vrp_band_number;
        int vrp_index = -1;
        if ( vrp_band > 0 )
            vrp_index = band_attributes->vrp_band_index;

        for ( sca = 0; sca < band_attributes->number_scas; sca++ )
        {
            /* Set the data type as const */
            /* VRP bands are not supported for Radiance values so only 
               handle the other buffer types */
            switch ( buffer_type )
            {
                case MM_ANALOG_LOW_SAT :
                    buffers[buffer_type][band_index][sca].is_const_data = 1;
                    buffers[buffer_type][band_index][sca].
                        const_data_pointer = &saturation_levels->
                        analog_low_saturation_level[band_index][sca][0];
                    buffers[buffer_type][band_index][sca].buffer_allocated =
                        1;
                    if (vrp_index >= 0)
                    {
                        buffers[buffer_type][vrp_index][sca].is_const_data = 1;
                        buffers[buffer_type][vrp_index][sca].
                            const_data_pointer = &saturation_levels->
                            analog_low_saturation_level[vrp_index][sca][0];
                        buffers[buffer_type][vrp_index][sca].
                            buffer_allocated = 1;
                    }
                    break;

                case MM_ANALOG_HIGH_SAT :
                    buffers[buffer_type][band_index][sca].is_const_data = 1;
                    buffers[buffer_type][band_index][sca].
                        const_data_pointer = &saturation_levels->
                        analog_high_saturation_level[band_index][sca][0];
                    buffers[buffer_type][band_index][sca].buffer_allocated =
                        1;
                    if (vrp_index >= 0)
                    {
                        buffers[buffer_type][vrp_index][sca].is_const_data = 1;
                        buffers[buffer_type][vrp_index][sca].
                            const_data_pointer = &saturation_levels->
                            analog_high_saturation_level[vrp_index][sca][0];
                        buffers[buffer_type][vrp_index][sca].
                            buffer_allocated = 1;
                    }
                    break;

                case MM_DIGITAL_LOW_SAT :
                    buffers[buffer_type][band_index][sca].is_const_data = 1;
                    buffers[buffer_type][band_index][sca].
                        const_data_pointer = &saturation_levels->
                        digital_low_saturation_level[band_index][sca][0];
                    buffers[buffer_type][band_index][sca].buffer_allocated =
                        1;
                    if (vrp_index >= 0)
                    {
                        buffers[buffer_type][vrp_index][sca].is_const_data = 1;
                        buffers[buffer_type][vrp_index][sca].
                            const_data_pointer = &saturation_levels->
                            digital_low_saturation_level[vrp_index][sca][0];
                        buffers[buffer_type][vrp_index][sca].
                            buffer_allocated = 1;
                    }
                    break;

                case MM_DIGITAL_HIGH_SAT :
                    buffers[buffer_type][band_index][sca].is_const_data = 1;
                    buffers[buffer_type][band_index][sca].
                        const_data_pointer = &saturation_levels->
                        digital_high_saturation_level[band_index][sca][0];
                    buffers[buffer_type][band_index][sca].buffer_allocated =
                        1;
                    if (vrp_index >= 0)
                    {
                        buffers[buffer_type][vrp_index][sca].is_const_data = 1;
                        buffers[buffer_type][vrp_index][sca].
                            const_data_pointer = &saturation_levels->
                            digital_high_saturation_level[vrp_index][sca][0];
                        buffers[buffer_type][vrp_index][sca].
                            buffer_allocated = 1;
                    }
                    break;

                case MM_HIGH_RADIANCE_SAT :
                    /* The Radiance Saturation values only
                       exist for normal bands */
                    buffers[buffer_type][band_index][sca].is_const_data = 1;
                    buffers[buffer_type][band_index][sca].
                        const_data_pointer = &saturation_levels->
                        high_radiance_saturation[band_index];
                    buffers[buffer_type][band_index][sca].buffer_allocated =
                        1;
                    break;

                case MM_LOW_RADIANCE_SAT :
                    /* The Radiance Saturation values only
                       exist for normal bands */
                    buffers[buffer_type][band_index][sca].is_const_data = 1;
                    buffers[buffer_type][band_index][sca].
                        const_data_pointer = &saturation_levels->
                        low_radiance_saturation[band_index];
                    buffers[buffer_type][band_index][sca].buffer_allocated =
                        1;
                    break;

                default:
                    break; /* No default action -- here for compiler warning */
            }
        }
    }
                        
    return SUCCESS;
}  /* END get_cpf_saturation_levels */

/*---------------------------------------------------------------------------
The routines in this section are special case CPF buffers that need to be
manipulated in the individual APIs or algorithms.  Therefore local copies
of the buffers are made and handled by the normal methods.

The values need to be modified before use so memory must be allocated and
the data copied.  The newly allocated memory will be freed in the decrement
routine.
---------------------------------------------------------------------------*/
static int get_cpf_tirs_dark_response()
{
    const struct IAS_CPF_TIRS_DETECTOR_RESPONSE *tirs_det_response;
    MEMORY_BUFFER_TYPE buffer_type = MM_TIRS_DARK_RESPONSE;

    int tirs_band_list[IAS_MAX_NBANDS];
    int index;
    int num_scas;
    int sca;
    int num_image_bands;
    int status;

    /* Get the list of TIRS imaging bands */
    status = ias_sat_attr_get_sensor_band_numbers(IAS_TIRS, IAS_NORMAL_BAND,
        0, tirs_band_list, IAS_MAX_NBANDS, &num_image_bands);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Getting the TIRS normal band numbers");
        return ERROR;
    }

    /* Get the TIRS_DETECTOR_RESPONSE groups */
    tirs_det_response = ias_cpf_get_tirs_det_response(get_blob_cpf());
    if (tirs_det_response == NULL)
    {
        IAS_LOG_ERROR("Getting the CPF TIRS detector response");
        return ERROR;
    }

    /* Get the number of TIRS SCAs */
    num_scas = ias_sat_attr_get_sensor_sca_count(IAS_TIRS);
    if (num_scas == ERROR)
    {
        IAS_LOG_ERROR("Getting the TIRS sensor SCA count");
        return ERROR;
    }

    /* Get the pointers */
    for (index = 0; index < num_image_bands; index++)
    {
        /* Get the band attributes */
        int band_number = tirs_band_list[index];
        int band_index = ias_sat_attr_convert_band_number_to_index(
            band_number );
        if ( band_index == ERROR )
        {
            IAS_LOG_ERROR("Getting index for band %d", band_number );
            return ERROR;
        }
        RPS_BAND_ATTRIBUTES *band_atts = get_band_attributes( band_index );
        size_t buffer_size = sizeof(double) * band_atts->number_detectors;
        int i;
        int process_band = 0;

        /* Check to see if this band is to be processed */
        for ( i = 0; i < mm_number_of_bands; i++ )
        {
            if ( band_index == mm_list_of_band_indices[i] )
            {
                process_band = 1;
                break;
            }
        }
        /* Continue to the next band if the band isn't to be processed */
        if ( ! process_band )
            continue;

        for (sca = 0; sca < num_scas; sca++)
        {
            buffers[buffer_type][band_index][sca].is_const_data = 0;
            buffers[buffer_type][band_index][sca].data_pointer =
                malloc( buffer_size );
            if ( !buffers[buffer_type][band_index][sca].data_pointer )
            {
                IAS_LOG_ERROR("Allocating data buffer %s for band %d "
                    "SCA %d", get_memory_buffer_string( buffer_type ),
                    tirs_band_list[band_index], sca + 1 );
                return ERROR;
            }
            buffers[buffer_type][band_index][sca].buffer_allocated = 1;

            memcpy( buffers[buffer_type][band_index][sca].data_pointer,
                tirs_det_response->baseline_dark_response[index][sca],
                buffer_size );
        }
    }

    return SUCCESS;
}   /* END get_cpf_tirs_dark_response */

static int get_cpf_tirs_dark_response_blind()
{
    const struct IAS_CPF_TIRS_DETECTOR_RESPONSE *tirs_det_response_blind;
    MEMORY_BUFFER_TYPE buffer_type = MM_TIRS_DARK_RESPONSE_BLIND;

    int tirs_band_list[IAS_MAX_NBANDS];
    int index;
    int num_scas;
    int sca;
    int num_blind_bands;
    int status;

    /* Get the list of TIRS blind bands */
    status = ias_sat_attr_get_sensor_band_numbers(IAS_TIRS, IAS_BLIND_BAND,
        IAS_SECONDARY_BAND, tirs_band_list, IAS_MAX_NBANDS,
        &num_blind_bands);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Getting the TIRS blind band numbers");
        return ERROR;
    }

    /* Get the TIRS_DETECTOR_RESPONSE groups */
    tirs_det_response_blind = ias_cpf_get_tirs_det_response_blind(
        get_blob_cpf() );
    if (tirs_det_response_blind == NULL)
    {
        IAS_LOG_ERROR("Getting the CPF TIRS blind band detector response");
        return ERROR;
    }

    /* Get the number of TIRS SCAs */
    num_scas = ias_sat_attr_get_sensor_sca_count(IAS_TIRS);
    if (num_scas == ERROR)
    {
        IAS_LOG_ERROR("Getting the TIRS blind band SCA count");
        return ERROR;
    }

    /* Get the pointers */
    for (index = 0; index < num_blind_bands; index++)
    {
        /* Get the band attributes */
        int band_number = tirs_band_list[index];
        int band_index = ias_sat_attr_convert_band_number_to_index(band_number);
        if (band_index == ERROR)
        {
            IAS_LOG_ERROR("Converting band number %d to index", band_number );
            return ERROR;
        }
        RPS_BAND_ATTRIBUTES *band_atts = get_band_attributes( band_index );
        size_t buffer_size = sizeof(double) * band_atts->number_detectors;

        for (sca = 0; sca < num_scas; sca++)
        {
            buffers[buffer_type][band_index][sca].is_const_data = 0;
            buffers[buffer_type][band_index][sca].data_pointer =
                malloc( buffer_size );
            if ( !buffers[buffer_type][band_index][sca].data_pointer )
            {
                IAS_LOG_ERROR("Allocating data buffer %s for band %d "
                    "SCA %d", get_memory_buffer_string( buffer_type ),
                    band_number, sca + 1 );
                return ERROR;
            }
            buffers[buffer_type][band_index][sca].buffer_allocated = 1;

            memcpy( buffers[buffer_type][band_index][sca].data_pointer,
                tirs_det_response_blind->baseline_dark_response[index][sca],
                buffer_size );
        }
    }

    return SUCCESS;
}   /* END get_cpf_tirs_dark_response_blind */

/*--------------------------------------------------------------------------
NAME:     get_spatial_detector_offsets

PURPOSE:  Retrieves along-track detector offset (actually detector delays)
          data from the CPF and converts them to spatial offsets.  The
          delays are integerized and normalized with the library routine.
          The smallest detector offset is 0 and the largest offset is
          stored in the MM_MAX_SPATIAL_DETECTOR_OFFSETS buffer.

RETURNS:  SUCCESS or ERROR
----------------------------------------------------------------------------*/
static int get_spatial_detector_offsets( MEMORY_BUFFER_TYPE buffer_type )
{
    int band_index = mm_list_of_band_indices[0];
    const struct IAS_CPF_DETECTOR_OFFSETS  *det_offsets;
    int index;
    int num_allocated = 0;
    int *temp_buffer = NULL; /* Initialized to NULL since it may not be
                                     allocated but will be freed */

    /* Since the first call to this routine initializes both buffer types
       for all bands and SCAs, check the first band in the list.  If that
       has been allocated return. */
    if ( buffers[buffer_type][band_index][0].buffer_allocated )
        return SUCCESS;

    /* Get the DETECTOR_OFFSETS group data */
    det_offsets = ias_cpf_get_det_offsets(get_blob_cpf());
    if (det_offsets == NULL)
    {
        IAS_LOG_ERROR("Getting the CPF detector offsets");
        return ERROR;
    }

    for ( index = 0; index < mm_number_of_bands; index++ )
    {
        /* Get the attributes for the current band */
        int band_index = mm_list_of_band_indices[index];
        RPS_BAND_ATTRIBUTES *band_atts = get_band_attributes( band_index );
        int band_number = band_atts->band_number;
        int number_scas = band_atts->number_scas;
        int number_detectors = band_atts->number_detectors;
        int sca_index;
        IAS_LOG_DEBUG("Allocating spatial offsets for band %d index %d",
            band_number, mm_list_of_band_indices[index] );

        /* We need temporary storage if getting the maximum value */
        if ( buffer_type == MM_MAX_SPATIAL_DETECTOR_OFFSETS
            && num_allocated < number_detectors )
        {
            int *temp_ptr;
            num_allocated = number_detectors;
            temp_ptr = realloc( temp_buffer, num_allocated *
                sizeof(int));
            if ( !temp_ptr )
            {
                IAS_LOG_ERROR("Reallocating temporary buffer");
                free( temp_buffer );
                return ERROR;
            }
            temp_buffer = temp_ptr;
        }

        for ( sca_index = 0; sca_index < number_scas; sca_index++ )
        {
            /* Allocate the buffer for the normal band */
            MEMORY_BUFFER *memory_buffer =
                &buffers[buffer_type][band_index][sca_index];
            memory_buffer->data_pointer = malloc( memory_buffer->size );
            if ( !memory_buffer->data_pointer )
            {
                IAS_LOG_ERROR("Allocating %s for band %d SCA %d",
                    get_memory_buffer_string( buffer_type), band_number,
                    sca_index + 1 );
                free( temp_buffer );
                return ERROR;
            }
            memory_buffer->buffer_allocated = 1;
            memory_buffer->is_const_data = 0;

            switch ( buffer_type )
            {
                case MM_SPATIAL_DETECTOR_OFFSETS :
                {
                    if ( band_atts->is_normal_band )
                    {
                        ias_rps_convert_detector_delays2spatial_offsets(
                            det_offsets->along_per_detector[band_index]
                            [sca_index], number_detectors,
                            memory_buffer->data_pointer, NULL );
                    }
                    else
                    {
                        /* The offsets for VRP and blind bands are 0 */
                        memset( memory_buffer->data_pointer, 0,
                            memory_buffer->size );
                    }
                    break;
                }

                case MM_MAX_SPATIAL_DETECTOR_OFFSETS :
                {
                    int max = 0;
                    int *data_pointer = (int *)memory_buffer->data_pointer;
                    if ( band_atts->is_normal_band )
                    {
                        ias_rps_convert_detector_delays2spatial_offsets(
                            det_offsets->
                            along_per_detector[band_index][sca_index],
                            number_detectors, temp_buffer, &max );
                    }
                    data_pointer[0] = (int)max;
                    break;
                }
                default :
                    ; /* No action */
            }
        }  /* END for sca_index loop */
    }  /* END for index loop */

    free( temp_buffer );
    return SUCCESS;
}  /* END get_spatial_detector_offsets */


/*--------------------------------------------------------------------------
 NAME:     get_l0r_frame_headers

 PURPOSE:  Obtains an array of OLI L0R frame headers

 RETURNS:  Integer status code of SUCCESS or ERROR
---------------------------------------------------------------------------*/
static int get_l0r_frame_headers
(
    MEMORY_BUFFER_TYPE buffer_type
)
{
    void *l0r_frame_headers = NULL;
    const RPS_BAND_ATTRIBUTES *band_atts = NULL;
    int *list_of_bands = NULL;
    int num_bands_to_process;
    int num_frame_headers;
    int band_index;
    int band_number;
    int vrp_band_number;
    int blind_band_number;
    int index;
    int sca_index;
    size_t buffer_size;


    /* Get the number of bands to process in the band list and the list
       itself */
    num_bands_to_process = get_blob_number_of_bands();
    list_of_bands = get_blob_list_of_bands_to_process();
    if (list_of_bands == NULL)
    {
        IAS_LOG_ERROR("Obtaining list of bands to process");
        return ERROR;
    }

    /* Count the number of OLI and TIRS bands that are supposed to be
       processed for the ensuing logic */
    int num_tirs_bands = 0;
    int num_oli_bands = 0;
    for (index = 0; index < num_bands_to_process; index++)
    {
        band_number = list_of_bands[index];
        band_index = ias_sat_attr_convert_band_number_to_index(
            band_number );
        if ( band_index == ERROR )
        {
            IAS_LOG_ERROR("Getting index for band %d", band_number );
            return ERROR;
        }
        band_atts = get_band_attributes(band_index);
        if (!band_atts->band_present)
            continue;

        /* Count the number of OLI bands and the number of TIRS bands
           in the processing list */
        if (band_atts->spectral_type == IAS_SPECTRAL_THERMAL)
            num_tirs_bands++;
        else
            num_oli_bands++;
    }

    /* Since the frame headers apply to multiple bands this routine will
       be called from the pre-process increment.  Pre-process routines don't
       take into account bands to process so this routine will be called
       once for OLI and again for TIRS.

       The purpose of the following logic is to prevent calling the routine
       to read the frame headers for the sensor if that sensor isn't to be
       processed.  Since it isn't an error to call this routine for a sensor
       that isn't in the processing list, return SUCCESS. */
    if ((buffer_type == MM_L0R_TIRS_FRAME_HEADER)
        && (num_tirs_bands == 0))
    {
        return SUCCESS;
    }
    else if ((buffer_type == MM_L0R_OLI_FRAME_HEADER)
        && (num_oli_bands == 0))
    {
        return SUCCESS;
    }

    /* Read the buffer the first time */
    l0r_frame_headers = read_l0r_frame_headers(buffer_type, &num_frame_headers,
        &buffer_size);
    if (l0r_frame_headers == NULL)
    {
        IAS_LOG_ERROR("Retrieving L0R frame headers");
        return ERROR;
    }

    /* Copy the information to every band/SCA to be processed */
    for (index = 0; index < num_bands_to_process; index++)
    {
        /* Get the (normal) band number and corresponding index */
        band_number = list_of_bands[index];
        band_index = ias_sat_attr_convert_band_number_to_index(
            band_number );
        if ( band_index == ERROR )
        {
            IAS_LOG_ERROR("Getting index for band %d", band_number );
            return ERROR;
        }

        /* Get the attributes */
        band_atts = get_band_attributes(band_index);

        /* Get the VRP and blind band numbers for later use */
        vrp_band_number = band_atts->vrp_band_number;
        blind_band_number = band_atts->blind_band_number;

        /* Copy the base data pointer, the number of frame headers,
           and the buffer size to the appropriate band/SCA in the
           specified buffer array */
        if ((buffer_type == MM_L0R_TIRS_FRAME_HEADER)
           && (band_atts->spectral_type == IAS_SPECTRAL_THERMAL))
        {
            for (sca_index = 0; sca_index < band_atts->number_scas;
                 sca_index++)
            {
                buffers[buffer_type][band_index][sca_index].is_const_data = 1;
                buffers[buffer_type][band_index][sca_index].
                    const_data_pointer = l0r_frame_headers;
                buffers[buffer_type][band_index][sca_index].number_values =
                    num_frame_headers;
                buffers[buffer_type][band_index][sca_index].size = buffer_size;
                buffers[buffer_type][band_index][sca_index].
                    buffer_allocated = 1;
            }
        }
        else if ((buffer_type == MM_L0R_OLI_FRAME_HEADER)
            && (band_atts->spectral_type != IAS_SPECTRAL_THERMAL))
        {
            for (sca_index = 0; sca_index < band_atts->number_scas;
                 sca_index++)
            {
                buffers[buffer_type][band_index][sca_index].is_const_data = 1;
                buffers[buffer_type][band_index][sca_index].
                    const_data_pointer = l0r_frame_headers;
                buffers[buffer_type][band_index][sca_index].number_values =
                    num_frame_headers;
                buffers[buffer_type][band_index][sca_index].size = buffer_size;
                buffers[buffer_type][band_index][sca_index].
                    buffer_allocated = 1;
            }
        }

        /* Don't forget the VRP bands or the blind bands! */
        if (vrp_band_number > 0)
        {
            band_index = band_atts->vrp_band_index;

            if ((buffer_type == MM_L0R_OLI_FRAME_HEADER)
                && (band_atts->spectral_type != IAS_SPECTRAL_THERMAL))
            {
                for (sca_index = 0; sca_index < band_atts->number_scas;
                     sca_index++)
                {
                    buffers[buffer_type][band_index][sca_index].
                        is_const_data = 1;
                    buffers[buffer_type][band_index][sca_index]
                        .const_data_pointer = l0r_frame_headers;
                    buffers[buffer_type][band_index][sca_index]
                        .number_values = num_frame_headers;
                    buffers[buffer_type][band_index][sca_index].size =
                        buffer_size;
                    buffers[buffer_type][band_index][sca_index].
                        buffer_allocated = 1;
                }
            }
        }
        if ((blind_band_number > 0) && band_atts->process_blind_band_flag)
        {
            band_index = band_atts->blind_band_index;

            if ((buffer_type == MM_L0R_OLI_FRAME_HEADER)
                && (band_atts->spectral_type != IAS_SPECTRAL_THERMAL))
            {
                for (sca_index = 0; sca_index < band_atts->number_scas;
                     sca_index++)
                {
                    buffers[buffer_type][band_index][sca_index].
                        is_const_data = 1;
                    buffers[buffer_type][band_index][sca_index]
                        .const_data_pointer = l0r_frame_headers;
                    buffers[buffer_type][band_index][sca_index]
                        .number_values = num_frame_headers;
                    buffers[buffer_type][band_index][sca_index].size =
                        buffer_size;
                    buffers[buffer_type][band_index][sca_index].
                        buffer_allocated = 1;
                }
            }
            else if ((buffer_type == MM_L0R_TIRS_FRAME_HEADER)
                && (band_atts->spectral_type == IAS_SPECTRAL_THERMAL))
            {
                for (sca_index = 0; sca_index < band_atts->number_scas;
                     sca_index++)
                {
                    buffers[buffer_type][band_index][sca_index].
                        is_const_data = 1;
                    buffers[buffer_type][band_index][sca_index]
                        .const_data_pointer = l0r_frame_headers;
                    buffers[buffer_type][band_index][sca_index]
                        .number_values = num_frame_headers;
                    buffers[buffer_type][band_index][sca_index].size =
                        buffer_size;
                    buffers[buffer_type][band_index][sca_index].
                        buffer_allocated = 1;
                }
            }
        }
    }

    return SUCCESS;
}   /* END -- get_l0r_frame_headers */


/*------------------------------------------------------------------------
NAME:    get_rlut_linearization_parameters

PURPOSE: Loads response linearization parameters into memory for a band/SCA

RETURNS: SUCCESS or ERROR
--------------------------------------------------------------------------*/
int get_rlut_linearization_parameters
(
    MEMORY_BUFFER_TYPE buffer_type,
    int band_number,
    int sca_number
)
{
    RPS_BAND_ATTRIBUTES *band_atts;
    int sca_index = sca_number - 1;
    IAS_THREAD_MUTEX_TYPE *hdf_io_mutex = get_hdf_io_mutex();
    int band_index = ias_sat_attr_convert_band_number_to_index( band_number );
    if ( band_index == ERROR )
    {
        IAS_LOG_ERROR("Getting index for band %d", band_number );
        return ERROR;
    }

    /* Get the band attributes */
    band_atts = get_band_attributes( band_index );
    
    buffers[buffer_type][band_index][sca_index].is_const_data = 0;
    buffers[buffer_type][band_index][sca_index].buffer_allocated = 1;

    if( IAS_THREAD_LOCK_MUTEX( hdf_io_mutex ))
    {
        IAS_LOG_ERROR("Locking the BPF I/O mutex");
        return ERROR;
    }

    buffers[buffer_type][band_index][sca_index].data_pointer =
        ias_rlut_read_linearization_params( get_blob_rlut(), band_number,
        sca_number, band_atts->number_detectors);
    if ( buffers[buffer_type][band_index][sca_index].data_pointer == NULL )
    {
        IAS_LOG_ERROR("Retrieving band %d SCA %d linearization "
            "parameters from RLUT", band_number, sca_number );
        IAS_THREAD_UNLOCK_MUTEX( hdf_io_mutex );
        return ERROR;
    }

    if ( IAS_THREAD_UNLOCK_MUTEX( hdf_io_mutex ))
    {
        IAS_LOG_ERROR("Unlocking the RLUT I/O mutex");
        return ERROR;
    }

    set_buffer_filled_flag( buffer_type, band_index, sca_index );

    return SUCCESS;
}   /* END get_rlut_linearization_parameters */

/*------------------------------------------------------------------------
NAME:    get_rlut_lookup_linearization_parameters

PURPOSE: Loads the lookup table response linearization for one band/SCA

RETURNS: SUCCESS or ERROR
--------------------------------------------------------------------------*/
int get_rlut_lookup_linearization_parameters
(
    MEMORY_BUFFER_TYPE buffer_type,
    int band_number,
    int sca_number
)
{
    IAS_RLUT_LOOKUP_TABLE *lin_params = NULL; 
    int sca_index = sca_number - 1;
    IAS_THREAD_MUTEX_TYPE *hdf_io_mutex = get_hdf_io_mutex();
    int band_index = ias_sat_attr_convert_band_number_to_index( band_number );
    IAS_RLUT_IO *rlut_ptr;     /* Local pointer to the RLUT */

    if ( band_index == ERROR )
    {
        IAS_LOG_ERROR("Getting index for band %d", band_number );
        return ERROR;
    }

    buffers[buffer_type][band_index][sca_index].is_const_data = 0;

    if( IAS_THREAD_LOCK_MUTEX( hdf_io_mutex ))
    {
        IAS_LOG_ERROR("Locking the RLUT I/O mutex");
        return ERROR;
    }

    /* Read the linearization rlut file for the linearity parameters. */
    rlut_ptr = get_blob_rlut();
    if (rlut_ptr == NULL)
    {
        IAS_LOG_ERROR("Retrieving the RLUT");
        return ERROR;
    }

    /* Read the linearity lookup table from the RLUT */
    lin_params = ias_rlut_read_lookup_table( rlut_ptr,
        IAS_RLUT_LINEARITY_LOOKUP, band_number, sca_number);
    if (lin_params == NULL)
    {
        IAS_LOG_ERROR("Retrieving band %d SCA %d linearization "
            "parameters from RLUT", band_number, sca_number );
        IAS_THREAD_UNLOCK_MUTEX( hdf_io_mutex );
        return ERROR;
    }

    buffers[buffer_type][band_index][sca_index].data_pointer = lin_params;

    if ( IAS_THREAD_UNLOCK_MUTEX( hdf_io_mutex ))
    {
        IAS_LOG_ERROR("Unlocking the RLUT I/O mutex");
        return ERROR;
    }

    set_buffer_filled_flag( buffer_type, band_index, sca_index );

    /* Indicate that the buffer retrieval was successful */
    buffers[buffer_type][band_index][sca_index].buffer_allocated = 1;

    return SUCCESS;
}   /* END get_rlut_lookup_linearization_parameters */


/*------------------------------------------------------------------------
------------------------------------------------------------------------*/
static int get_rlut_tirs_secondary_linearization_parameters
(
    MEMORY_BUFFER_TYPE buffer_type, /* I: */
    int band_number,    /* I: */
    int sca_number      /* I: */
)
{
    IAS_RLUT_LOOKUP_TABLE *params;
    IAS_THREAD_MUTEX_TYPE *hdf_io_mutex = get_hdf_io_mutex();
    int sca_index = sca_number - 1;
    int band_index = ias_sat_attr_convert_band_number_to_index( band_number );

    if ( band_index == ERROR )
    {
        IAS_LOG_ERROR("Getting index for band %d", band_number );
        return ERROR;
    }

    if( IAS_THREAD_LOCK_MUTEX( hdf_io_mutex ))
    {
        IAS_LOG_ERROR("Locking the RLUT I/O mutex");
        return ERROR;
    }
    
    /* Read the paramaters */
    params = ias_rlut_read_lookup_table( get_blob_rlut(),
         IAS_RLUT_TIRS_SECONDARY_LOOKUP, band_number, sca_number );
    if ( ! params )
    {
        IAS_THREAD_UNLOCK_MUTEX( hdf_io_mutex );
        IAS_LOG_ERROR("Reading the TIRS secondary linearization parameters "
            "for Band %d SCA %d", band_number, sca_number );
        return ERROR;
    }
    
    if ( IAS_THREAD_UNLOCK_MUTEX( hdf_io_mutex ))
    {
        IAS_LOG_ERROR("Unlocking the RLUT I/O mutex");
        return ERROR;
    }

    /* data_pointer is NULL until this point */
    buffers[buffer_type][band_index][sca_index].data_pointer = params;
    buffers[buffer_type][band_index][sca_index].is_const_data = 0;
    buffers[buffer_type][band_index][sca_index].buffer_allocated = 1;
    set_buffer_filled_flag( buffer_type, band_index, sca_index );

    return SUCCESS;
}  /* END get_rlut_tirs_secondary_linearization_parameters */

/*------------------------------------------------------------------------
Name:    read_bpf_bias_model_parameters

PURPOSE: Read in the needed bpf bias data and bias model data for the
         OLI bias determination algorithm

NOTE:    Due to the current BPF library routines, the BPF has to be handled
         similar to the image data

RETURNS: each routine returns SUCCESS or ERROR
------------------------------------------------------------------------*/
int read_bpf_bias_model_parameters( int band_number, int sca_number,
           MEMORY_BUFFER_TYPE buffer_type, void ** pre_shutter,
           void ** post_shutter, void ** mean_shutter, void ** a1_parameter, 
           void ** c1_parameter, void **a0_parameter, int *number_detectors )
{
    RPS_BAND_ATTRIBUTES *band_attributes = NULL;
    static struct IAS_BPF_BIAS_MODEL *bias_model = NULL;
    /* The local pre-acquisition shutter average pointer */
    double *local_pre_shutter = NULL;
    /* The local post-acquisition shutter average pointer */
    double *local_post_shutter = NULL;
    /* The local mean shutter average pointer */
    double *local_mean_shutter = NULL;
    /* The local pointer for the a1_parameter */
    double *local_a1_parameter = NULL;
    /* The local pointer for the c1_parameter */
    double *local_c1_parameter = NULL;
    /* The local pointer for the a0_parameter */
    double *local_a0_parameter = NULL;

    /* Flag to determine whether to read ODD or EVEN bias state */
    int pan_evenodd_flag = IAS_BPF_PAN_ODD;
    int status;
    int band_index;
    int det_index;
    int sca_index = sca_number - 1;

    /* Set the pan_evenodd_flag based on buffer_type */
    switch ( buffer_type )
    {
        case MM_PAN_EVEN_PRE_SHUTTER:
        case MM_PAN_EVEN_POST_SHUTTER:
        case MM_PAN_EVEN_MEAN_SHUTTER:
        case MM_PAN_EVEN_BIAS_A1_PARAMETER:
        case MM_PAN_EVEN_BIAS_C1_PARAMETER:
        case MM_PAN_EVEN_BIAS_A0_PARAMETER:
            pan_evenodd_flag = IAS_BPF_PAN_EVEN;
            break; 

        default:
            break; /* No default action -- here for compiler warning */ 
    }

    /* Allocate memory for local shutter buffers and a0_parameter,
       here we have to use detector number for pan band to allocate memory */
    band_index = ias_sat_attr_convert_band_number_to_index( band_number );
    if ( band_index == ERROR )
    {
        IAS_LOG_ERROR("Getting index for band %d", band_number );
        return ERROR;
    }
    band_attributes = get_band_attributes( band_index );
    local_pre_shutter = malloc(
        band_attributes->number_detectors * sizeof(double));
    local_post_shutter = malloc(
        band_attributes->number_detectors * sizeof(double));
    local_mean_shutter = malloc(
        band_attributes->number_detectors * sizeof(double));
    local_a1_parameter = malloc(
        band_attributes->number_detectors * sizeof(double));
    local_c1_parameter = malloc(
        band_attributes->number_detectors * sizeof(double));
    local_a0_parameter = malloc(sizeof(double));
    if (!local_pre_shutter || !local_post_shutter
        || !local_mean_shutter || !local_a1_parameter
        || !local_c1_parameter || !local_a0_parameter)
    {
        free(local_pre_shutter);
        free(local_post_shutter);
        free(local_mean_shutter);
        free(local_a1_parameter);
        free(local_c1_parameter);
        free(local_a0_parameter);
        return ERROR;
    }

    if( IAS_THREAD_LOCK_MUTEX( &bpf_io_mutex ))
    {
        IAS_LOG_ERROR("Locking the BPF I/O mutex");
        free(local_pre_shutter);
        free(local_post_shutter);
        free(local_mean_shutter);
        free(local_a1_parameter);
        free(local_c1_parameter);
        free(local_a0_parameter);
        return ERROR;
    }

    /* Get the BPF bias_model structure for OLI */
    if ( !bias_model )
    {
        bias_model = (struct IAS_BPF_BIAS_MODEL *)ias_bpf_get_bias_model(
            get_blob_bpf_oli());
        if ( !bias_model )
        {
            IAS_THREAD_UNLOCK_MUTEX( &bpf_io_mutex );
            IAS_LOG_ERROR("Getting the BPF bias model");
            free(local_pre_shutter);
            free(local_post_shutter);
            free(local_mean_shutter);
            free(local_a1_parameter);
            free(local_c1_parameter);
            free(local_a0_parameter);
            return ERROR;
        }
    }

    /* Getting the shutter data and a0_parameter from BPF */
    status = ias_bpf_get_model_parameters(bias_model,
                     band_attributes->spectral_type, band_number, 
                     sca_number, band_attributes->number_detectors,
                     pan_evenodd_flag, local_pre_shutter, local_post_shutter, 
                     local_a1_parameter, local_c1_parameter, 
                     local_a0_parameter);
    if (status != SUCCESS)
    {
        IAS_THREAD_UNLOCK_MUTEX( &bpf_io_mutex );
        IAS_LOG_ERROR("Getting the BPF model parameters");
        free(local_pre_shutter);
        free(local_post_shutter);
        free(local_mean_shutter);
        free(local_a1_parameter);
        free(local_c1_parameter);
        free(local_a0_parameter);
        return ERROR;
    }

    for (det_index = 0; det_index < band_attributes->number_detectors;
         det_index++)
    {
        local_mean_shutter[det_index] = 0.5 * (local_pre_shutter[det_index] +
            local_post_shutter[det_index]);
    }

    if ( IAS_THREAD_UNLOCK_MUTEX( &bpf_io_mutex ))
    {
        IAS_LOG_ERROR("Unlocking the BPF I/O mutex");
        free(local_pre_shutter);
        free(local_post_shutter);
        free(local_mean_shutter);
        free(local_a1_parameter);
        free(local_c1_parameter);
        free(local_a0_parameter);
        return ERROR;
    }

    /* Populate the output pointers */ 
    *number_detectors = band_attributes->number_detectors;
    *pre_shutter = (void *)local_pre_shutter;
    *post_shutter = (void *)local_post_shutter;
    *mean_shutter = (void *)local_mean_shutter;
    *a1_parameter = (void *)local_a1_parameter;
    *c1_parameter = (void *)local_c1_parameter;
    *a0_parameter = (void *)local_a0_parameter;

    /* Since the library for OLI BPF was written so that all of the
       parameters are retrieved at the same time -- even if not needed --
       they have been set.  This mess is to accomodate setting the filled
       flags for each of the different buffers. */
    switch ( buffer_type )
    {
        case MM_PAN_EVEN_PRE_SHUTTER:
        case MM_PAN_EVEN_POST_SHUTTER:
        case MM_PAN_EVEN_MEAN_SHUTTER:
        case MM_PAN_EVEN_BIAS_A1_PARAMETER:
        case MM_PAN_EVEN_BIAS_C1_PARAMETER:
        case MM_PAN_EVEN_BIAS_A0_PARAMETER:
            set_buffer_filled_flag( MM_PAN_EVEN_PRE_SHUTTER, band_index,
                sca_index );
            set_buffer_filled_flag( MM_PAN_EVEN_POST_SHUTTER, band_index,
                sca_index );
            set_buffer_filled_flag( MM_PAN_EVEN_MEAN_SHUTTER, band_index,
                sca_index );
            set_buffer_filled_flag( MM_PAN_EVEN_BIAS_A1_PARAMETER, band_index,
                sca_index );
            set_buffer_filled_flag( MM_PAN_EVEN_BIAS_C1_PARAMETER, band_index,
                sca_index );
            set_buffer_filled_flag( MM_PAN_EVEN_BIAS_A0_PARAMETER, band_index,
                sca_index );
            break;

        case MM_PAN_ODD_PRE_SHUTTER:
        case MM_PAN_ODD_POST_SHUTTER:
        case MM_PAN_ODD_MEAN_SHUTTER:
        case MM_PAN_ODD_BIAS_A1_PARAMETER:
        case MM_PAN_ODD_BIAS_C1_PARAMETER:
        case MM_PAN_ODD_BIAS_A0_PARAMETER:
            set_buffer_filled_flag( MM_PAN_ODD_PRE_SHUTTER, band_index,
                sca_index );
            set_buffer_filled_flag( MM_PAN_ODD_POST_SHUTTER, band_index,
                sca_index );
            set_buffer_filled_flag( MM_PAN_ODD_MEAN_SHUTTER, band_index,
                sca_index );
            set_buffer_filled_flag( MM_PAN_ODD_BIAS_A1_PARAMETER, band_index,
                sca_index );
            set_buffer_filled_flag( MM_PAN_ODD_BIAS_C1_PARAMETER, band_index,
                sca_index );
            set_buffer_filled_flag( MM_PAN_ODD_BIAS_A0_PARAMETER, band_index,
                sca_index );
            break;

        case MM_NON_PAN_PRE_SHUTTER :
        case MM_NON_PAN_POST_SHUTTER :
        case MM_NON_PAN_MEAN_SHUTTER :
        case MM_NON_PAN_BIAS_A0_PARAMETER :
        case MM_NON_PAN_BIAS_A1_PARAMETER :
        case MM_NON_PAN_BIAS_C1_PARAMETER :
        default :
            set_buffer_filled_flag( MM_NON_PAN_PRE_SHUTTER, band_index,
                sca_index );
            set_buffer_filled_flag( MM_NON_PAN_POST_SHUTTER, band_index,
                sca_index );
            set_buffer_filled_flag( MM_NON_PAN_MEAN_SHUTTER, band_index,
                sca_index );
            set_buffer_filled_flag( MM_NON_PAN_BIAS_A0_PARAMETER, band_index,
                sca_index );
            set_buffer_filled_flag( MM_NON_PAN_BIAS_A1_PARAMETER, band_index,
                sca_index );
            set_buffer_filled_flag( MM_NON_PAN_BIAS_C1_PARAMETER, band_index,
                sca_index );
            break;
    }

    return SUCCESS;
}  /* END read_bpf_bias_model_parameters */

/*------------------------------------------------------------------------
Name:    read_tirs_bpf_bias_model_parameters

PURPOSE: Read in the needed bpf bias data and bias model data for the
         TIRS bias removal algorithm

NOTE:    Due to the current BPF library routines, the BPF has to be handled
         similar to the image data

RETURN: SUCCESS or ERROR
------------------------------------------------------------------------*/
int read_tirs_bpf_bias_model_parameters( int band_number, int sca_number,
           void ** background_response, int *number_detectors )
{
    RPS_BAND_ATTRIBUTES *band_attributes = NULL;
    static const struct IAS_BPF_BIAS_MODEL *bias_model = NULL;
    /* The local pre-acquisition deep space average pointer */
    double *local_pre_acq = NULL;
    /* The local post-acquisition deep space average pointer */
    double *local_post_acq = NULL;
    int local_background_response_selection = 
        get_background_response_selection();
    int status;
    int band_index;
    int det_index;

    /* Allocate memory for local acquisition buffers, here we have to use 
       detector number for pan band to allocate memory */
    band_index = ias_sat_attr_convert_band_number_to_index( band_number );
    if ( band_index == ERROR )
    {
        IAS_LOG_ERROR("Getting index for band %d", band_number );
        return ERROR;
    }
    band_attributes = get_band_attributes( band_index );
    local_pre_acq = malloc(
        band_attributes->number_detectors * sizeof(double));
    local_post_acq = malloc(
        band_attributes->number_detectors * sizeof(double));
    if (!local_pre_acq || !local_post_acq )
    {
        free(local_pre_acq);
        free(local_post_acq);
        return ERROR;
    }

    if( IAS_THREAD_LOCK_MUTEX( &bpf_io_mutex ))
    {
        IAS_LOG_ERROR("Locking the BPF I/O mutex");
        free(local_pre_acq);
        free(local_post_acq);
        return ERROR;
    }

    /* Get the BPF bias_model structure for TIRS */
    if ( !bias_model )
    {
        IAS_BPF *bpf = get_blob_bpf_tirs();
        if ( !bpf )
        {
            IAS_THREAD_UNLOCK_MUTEX( &bpf_io_mutex );
            IAS_LOG_ERROR("TIRS BPF not available");
            free(local_pre_acq);
            free(local_post_acq);
            return ERROR;
        }

        bias_model = ias_bpf_get_bias_model(bpf);
        if ( !bias_model )
        {
            IAS_THREAD_UNLOCK_MUTEX( &bpf_io_mutex );
            IAS_LOG_ERROR("Getting the BPF bias model");
            free(local_pre_acq);
            free(local_post_acq);
            return ERROR;
        }
    }

    /* Getting the shutter data and a0_parameter from BPF */
    status = ias_bpf_get_model_parameters(bias_model,
                     band_attributes->spectral_type, band_number, 
                     sca_number, band_attributes->number_detectors,
                     0, local_pre_acq, local_post_acq, NULL, NULL, NULL);
    if (status != SUCCESS)
    {
        IAS_THREAD_UNLOCK_MUTEX( &bpf_io_mutex );
        IAS_LOG_ERROR("Getting the BPF model parameters");
        free(local_pre_acq);
        free(local_post_acq);
        return ERROR;
    }

    if ( IAS_THREAD_UNLOCK_MUTEX( &bpf_io_mutex ))
    {
        IAS_LOG_ERROR("Unlocking the BPF I/O mutex");
        free(local_pre_acq);
        free(local_post_acq);
        return ERROR;
    }

    /* Populate the output pointers */ 
    *number_detectors = band_attributes->number_detectors;
    if (local_background_response_selection ==
        PRE_ACQUISITION_DEEP_SPACE)
    {
        *background_response = (void *)local_pre_acq;
        free(local_post_acq);
    }
    else if (local_background_response_selection ==
             POST_ACQUISITION_DEEP_SPACE)
    {
        *background_response = (void *)local_post_acq;
        free(local_pre_acq);
    }
    else if (local_background_response_selection ==
        MEAN_ACQUISITION_DEEP_SPACE)
    {
        for (det_index = 0; det_index < band_attributes->number_detectors;
             det_index++)
        {
            local_post_acq[det_index] = 0.5 * (local_post_acq[det_index] +
                local_pre_acq[det_index]);
        }
        *background_response = (void *)local_post_acq;
        free(local_pre_acq);
    }
    else
    {
        IAS_LOG_ERROR("Only BPF BACKGROUND_RESPONSE_SELECTION %d for band "
                      "%d SCA %d can be called here", 
                      local_background_response_selection,
                      band_number, sca_number);
        free(local_pre_acq);
        free(local_post_acq);
        return ERROR;
    }
    set_buffer_filled_flag( MM_BPF_BACKGROUND_RESPONSE, band_index,
        sca_number - 1 );

    return SUCCESS;
}  /* END read_tirs_bpf_bias_model_parameters */

/*--------------------------- Test Function ------------------------------
   This function is called by the test suite and is used to track  
   changes in the internal state of this set of routines.  It is NOT
   intended to be used by any other routine.
------------------------------------------------------------------------*/
void get_mem_mgmt_struct_dont_use_me
(
    MEMORY_BUFFER_TYPE type,    /* I: Memory type */
    int band_index,             /* I: */
    int sca_index,              /* I: */
    int *num_users,             /* O: Number users*/
    MEMORY_BUFFER *buf          /* O: Copy of buf */
)
{
    buf->user_count = buffers[type][band_index][sca_index].user_count;
    buf->data_pointer = buffers[type][band_index][sca_index].data_pointer;
    *num_users = number_consumers[type];
}
