/*----------------------------------------------------------------------------
NAME:     miscellany.c

ROUTINES: get_rps_level
          initialize_rps_level
          cleanup_rps_level
          update_omf
          get_earth_sun_distance
          rps_level_is_l0r

PURPOSE:  Collection of routines that don't fit into any other areas.
-----------------------------------------------------------------------------*/
#include <stdlib.h>
#include <string.h>

#include "ias_types.h"
#include "ias_logging.h"
#include "ias_parm.h"
#include "ias_miscellaneous.h"
#include "ias_structures.h"
#include "ias_pixel_mask.h"
#include "ias_math.h"
#include "ias_geo.h"
#include "ias_parm_standard_parameters.h"
#include "rps_local.h"

static char* **rps_level_array = NULL;
static int rps_level_nbands = 0;
static int rps_level_nscas = 0;
static int earth_sun_distance_initialized = 0;
static double distance_from_earth_to_sun;   /* Earth-sun distance in AU */

/*------------------------------------------------------------------------
NAME:    initialize_rps_level

PURPOSE: Allocates the RPS level array and initializes it to zero.  This
         array is used to track the current RPS level for each band SCA
         combination.  The RPS level is used in certain database tables
         to allow a single algorithm to be run multiple times.  A good
         example of this is the histogram statistics table.

RETURNS: SUCCESS or ERROR
------------------------------------------------------------------------*/
int initialize_rps_level
(
    int number_of_bands,        /* I: */
    int number_of_scas          /* I: */
)
{
    int band_index;
    int sca_index;
    int status = SUCCESS;

    rps_level_array = malloc( sizeof(int *) * number_of_bands );
    if ( !rps_level_array )
    {
        IAS_LOG_ERROR("Allocating the RPS level Array");
        return ERROR;
    }

    for ( band_index = 0; band_index < number_of_bands; band_index++ )
    {
        rps_level_array[band_index] = malloc( sizeof(int *) * number_of_scas );
        if ( rps_level_array[band_index] )
        {
            for ( sca_index = 0; sca_index < number_of_scas; sca_index++ )
            {
                rps_level_array[band_index][sca_index] = 
                    malloc(sizeof(char) * (IAS_RPS_PROCESSING_LEVEL_LENGTH+1));

                if ( rps_level_array[band_index][sca_index] )
                {
                    rps_level_array[band_index][sca_index][0] = '\0';
                }
                else
                {
                    status = ERROR;
                }
            }
        }
        else
        {
            status = ERROR;
        }
    }
    rps_level_nbands = number_of_bands;
    rps_level_nscas = number_of_scas;

    if ( status != SUCCESS )
    {
        cleanup_rps_level();
        IAS_LOG_ERROR("Allocating the RPS level Array");
    }
    return status;
}  /* END initialize_rps_level */


/*------------------------------------------------------------------------
NAME:    cleanup_rps_level

PURPOSE: Frees memory associated with the routine and sets variables to
         uninitialized state

RETURNS: none
------------------------------------------------------------------------*/
void cleanup_rps_level( void )
{
    int band_index;
    int sca_index;

    if ( rps_level_array )
    {
        for ( band_index = 0; band_index < rps_level_nbands; band_index++ )
        {
            if ( rps_level_array[band_index] )
            {
                for ( sca_index = 0; sca_index < rps_level_nscas; sca_index++ )
                    free( rps_level_array[band_index][sca_index] );
            }

            free( rps_level_array[band_index] );
        }
    }
    free( rps_level_array );
    rps_level_nbands = 0;
    rps_level_nscas = 0;
    rps_level_array = NULL;
}  /* END cleanup_rps_level */


/*------------------------------------------------------------------------
NAME:    get_rps_level

PURPOSE: Return a string representation of the RPS level.  Each algorithm
         that needs the RPS level will call this routine.  The caller is
         responsible for freeing the memory allocated.

RETURNS: pointer to the string
------------------------------------------------------------------------*/
char *get_rps_level
(
    int band_index,           /* I: */
    int sca_index             /* I: */
)
{
    int length;
    char *rps_level;
    int status;

    if ( rps_level_array == NULL )
    {
        IAS_LOG_ERROR("RPS level routines are not initialized");
        return NULL;
    }
    if ( band_index < 0 || band_index > rps_level_nbands - 1 )
    {
        IAS_LOG_ERROR("RPS level band index: %d is not in 1...%d",
            band_index, rps_level_nbands );
        return NULL;
    }
    if ( sca_index < 0 || sca_index > rps_level_nscas - 1)
    {
        IAS_LOG_ERROR("RPS level SCA: %d is not in 1...%d",
            sca_index + 1, rps_level_nscas );
        return NULL;
    }

    /* Get the length of the rps_level string */
    length = strlen( rps_level_array[band_index][sca_index] );

    rps_level = malloc( sizeof(char) * (length + 1) );
    if ( !rps_level )
    {
        IAS_LOG_ERROR("Allocating the RPS level string");
        return NULL;
    }

    status = snprintf( rps_level, (length + 1), "%s", 
        rps_level_array[band_index][sca_index] );
    if ( status < 0 || status >= (length + 1) )
    {
        IAS_LOG_ERROR("Getting RPS level for band %d SCA %d",
            get_band_attributes( band_index )->band_number, sca_index + 1);
        free( rps_level );
        return NULL;
    }

    return rps_level;
}  /* END get_rps_level */

/*------------------------------------------------------------------------
NAME:    set_rps_level

PURPOSE: Appends process_id to the rps_processing_level string for band_index 
         and sca_index.
   
         Note: This routine is not actually thread safe.  The work assignment
               algorithm *is* mutex controlled and can't assign algorithm B
               before algorithm A is complete.  Therefore the RPS level for
               each band SCA combination can't be set by multiple threads.

RETURNS: SUCCESS or ERROR
------------------------------------------------------------------------*/
int set_rps_level
(
    int band_index,         /* I: */
    int sca_index,          /* I: */
    char process_id         /* I: */
)
{
    int status;
    int length;
    char temp_str[IAS_RPS_PROCESSING_LEVEL_LENGTH + 1];

    if ( rps_level_array == NULL )
    {
        IAS_LOG_ERROR("RPS level routines are not initialized");
        return ERROR;
    }
    if ( band_index < 0 || band_index > rps_level_nbands - 1 )
    {
        IAS_LOG_ERROR("RPS level band index: %d is not in 1...%d",
            band_index, rps_level_nbands );
        return ERROR;
    }
    if ( sca_index < 0 || sca_index > rps_level_nscas - 1)
    {
        IAS_LOG_ERROR("RPS level SCA: %d is not in 1...%d",
            sca_index + 1, rps_level_nscas );
        return ERROR;
    }

    /* Check if the rps processing level has reached its maximum size */
    length = strlen( rps_level_array[band_index][sca_index] );
    if ( length >= IAS_RPS_PROCESSING_LEVEL_LENGTH)
    {
        IAS_LOG_ERROR("RPS level exceeds maximum length.");
        return ERROR;
    }

    strcpy(temp_str, rps_level_array[band_index][sca_index]);

    /* Append the new process_id to the RPS level */
    status = snprintf(rps_level_array[band_index][sca_index], 
        IAS_RPS_PROCESSING_LEVEL_LENGTH, "%s%c", temp_str, process_id);

    if ( status != (length + 1) )
    {
        strcpy(rps_level_array[band_index][sca_index], temp_str);
        IAS_LOG_ERROR("Error appending process_id to RPS level");
        return ERROR;
    }

    return SUCCESS;
}  /* END set_rps_level */

/*------------------------------------------------------------------------
NAME:    rps_level_is_l0r

PURPOSE: Check if the processing level string represents L0R data. 
   
RETURNS: TRUE or FALSE
------------------------------------------------------------------------*/
int rps_level_is_l0r
(
    const char *rps_level      /* I: RPS processing level */
)
{
    while (*rps_level != '\0')
    {
        /* For L0R data the only processing level characters present will be 
           convert_to_float or the skipped algorithm character */
        if (*rps_level != IAS_RPS_LEVEL_SKIP 
                && *rps_level != IAS_RPS_LEVEL_CONVERT_FLOAT_NOSHIFT
                && *rps_level != IAS_RPS_LEVEL_CONVERT_FLOAT_CORRECT)
        {
            return FALSE;
        }

        rps_level++;
    }

    return TRUE;
}

/*------------------------------------------------------------------------
NAME:    update_omf

PURPOSE: Updates the OMF with the L1R file name
   
RETURNS: SUCCESS or ERROR
------------------------------------------------------------------------*/
int update_omf
(
    BLOB *blob         /* I: contains the OMF and L1R file names */
)
{
    int count;
    int status;

    /* Set up a parameter table */
    IAS_PARM_DECLARE_TABLE( omf_parameter_table, 3 );
    IAS_PARM_L1R_FILENAME( omf_parameter_table, IAS_PARM_REQUIRED,
        blob->l1r_file_name, sizeof(blob->l1r_file_name), 0 );
    IAS_PARM_PIXEL_MASK_FILENAME( omf_parameter_table, IAS_PARM_REQUIRED,
        blob->pixel_mask_filename, sizeof(blob->pixel_mask_filename), 0 );
    IAS_PARM_TIRS_STRAY_LIGHT_CORRECTION_SOURCE( omf_parameter_table, 
        IAS_PARM_OPTIONAL, blob->stray_light_correction_source, 
        sizeof(blob->stray_light_correction_source), 0 );

    IAS_PARM_SET_COUNT_WRITE( omf_parameter_table, L1R_FILENAME, 1 );
    IAS_PARM_SET_COUNT_WRITE( omf_parameter_table, PIXEL_MASK_FILENAME, 1 );

    /* Write out the stray light correction source if present */
    IAS_PARM_SET_COUNT_WRITE( omf_parameter_table, 
        TIRS_STRAY_LIGHT_CORRECTION_SOURCE, 1 );
    if (strlen(blob->stray_light_correction_source) <= 0)
        IAS_PARM_POP_LAST_PARAMETER( omf_parameter_table );

    /* Get the parameter write count */
    count = IAS_PARM_GET_TABLE_SIZE( omf_parameter_table );

    status = ias_parm_update( blob->omf_file_name , omf_parameter_table,
        count );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Cannot write L1R filename to OMF %s",
            blob->omf_file_name  );
        return ERROR;
    }

    return SUCCESS;
}  /* END update_omf */


/*------------------------------------------------------------------------
NAME:    get_earth_sun_distance

PURPOSE: Get the earth-sun distance by using the jpl library
   
RETURNS: SUCCESS or ERROR
------------------------------------------------------------------------*/
int get_earth_sun_distance
(
    char *date_acquired,        /* I: contains the date_required */
    double *earth_sun_distance  /* O: earth-sun distance */
)
{
    double julian_date;              /* Julian ephemeris date */
    IAS_DATETIME datetime;           /* Data structure containing date/time 
                                        elements as integer values */
    double secs_in_day;              /* Seconds in the day */
    double right_ascension;          /* Right ascension (radian), relative 
                                        to "true" equator and equinox */
    double declination;              /* Declination (degrees), relative to 
                                        "true" equator and equinox */
    int status;                      /* Return status */

    /* Initialize the JPL library when it hasn't been initialized */
    if (!earth_sun_distance_initialized)
    {
        if (!date_acquired)
        {
            IAS_LOG_ERROR("date_acquired pointer is empty");
            return ERROR;
        }

        if (ias_geo_initialize_novas() != SUCCESS)
        {
            IAS_LOG_ERROR("Initializing NOVAS library");
            return ERROR;
        }

        /* Parse the datetime string, so that we can use the components to
           calculate a Julian date */
        memset(&datetime, 0, sizeof(IAS_DATETIME));
        status = ias_misc_parse_datetime_string(date_acquired, 
                 IAS_DATETIME_L0R_FORMAT, &datetime);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Parsing date/time string");
            return ERROR;
        }

        /* Calculate the total seconds corresponding to the time in the
           current day */
        secs_in_day
            = datetime.hour * 3600 + datetime.minute * 60 + datetime.second;

        /* Compute the full Julian date */
        julian_date = ias_math_compute_full_julian_date(datetime.year,
            datetime.month, datetime.day_of_month, secs_in_day);

        /* Compute the Earth-sun distance in AU. Values for right ascension and
           declination are determined as part of this calculation, so 
           variables for them must be provided as parameters. However, their 
           values are not used in this context. */
        status = ias_geo_compute_sun_position(julian_date, &right_ascension, 
                   &declination, &distance_from_earth_to_sun);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Calculating Earth-Sun distance using JPL "
                      "NOVAS library routines");
            return ERROR;
        }

        /* Set the earth_sun_distance_initialized value be true */
        earth_sun_distance_initialized = 1;

        /* Close out the JPL library */
        if (ias_geo_shutdown_novas() != SUCCESS)
        {
            IAS_LOG_ERROR("Shutting down NOVAS library");
            return ERROR;
        }
    }

    /* Return the Earth-sun distance */
    *earth_sun_distance = distance_from_earth_to_sun;

    return SUCCESS;
}  /* END get_earth_sun_distance */


/**************************************************************************
 NAME:     calc_normalized_correction_factor

 PURPOSE:  Calculate correction factors and normalize them

 RETURNS:  Integer status code of SUCCESS or ERROR
***************************************************************************/
int calc_normalized_correction_factor
(
    const double *input_vector,/* I: Vector values (nscas size) from which
                                     normalized factors are calculated.
                                     Element 0 is expected to be pre-
                                     populated with 1.0 and the remainder
                                     with SCA-SCA ratios: element 1 is the
                                     ratio between SCAs 1 and 2, and so on. */
    int nscas,                 /* I: Number of SCAs per band */
    double *correction_factor  /* O: Array of normalized correction factors 
                                     with NSCAS numbers */
)
{
    float sum_correction_factor;
    float inverse_mean_correction_factor;
    int index;

    /* Calculate correction factor.  This is a cumulative product of the
       elements along the vector.  The first element in the vector is a
       simple copy of the input vector ie. multiplied by a non-existant 1.0
       to the left of it. */
    correction_factor[0] = input_vector[0];
    for (index = 1; index < nscas; index++)
    {
        correction_factor[index] = correction_factor[index - 1] * 
            input_vector[index];
    }

    /* Calculate the inverse mean of the cumulative product */
    sum_correction_factor = 0.0;
    for (index = 0; index < nscas; index++)
        sum_correction_factor += correction_factor[index];

    if (sum_correction_factor == 0)
    {
        IAS_LOG_ERROR("Correction factors are invalid.");
        return ERROR;
    }
    inverse_mean_correction_factor = nscas / sum_correction_factor;

    /* Normalize the values */
    for (index = 0; index < nscas; index++)
        correction_factor[index] *= inverse_mean_correction_factor;

    return SUCCESS;
}

