/*----------------------------------------------------------------------------
NAME:    process_collection

PURPOSE: Run any pre-process algorithms; initialize the thread pool and
         release it to operate on process_sca.  Also enables a work queue
         for database transactions.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include <limits.h>

#include "ias_threadpool.h"
#include "ias_threadsync.h"
#include "ias_pixel_mask.h"
#include "ias_logging.h"
#include "rps_local.h"


int process_collection
(
    BLOB *blob          /* I: Various input parameters:
                              pre-process pointer list
                              number of SCAs
                              number of bands
                              band information
                              list of bands to process
                              post-process pointer list */
)
{
    int i;              /* Generic loop counter */
    int status;         /* Routine return status */
    struct ias_threadpool *threadpool;   /* threadpool for processing */
    int index;          /* Loop counter */
    int error_occurred = 0;
    char *algorithm_list; /* Convenience pointer */

    /* If there are reports to generate make sure the directory is available */
    if ( blob->generate_report )
    {
        struct stat stat_buf;
        if ( stat( blob->report_information.directory, &stat_buf ))
        {
            int  the_error = errno;
            char *error_string;

            /* Check errno.  ENOENT ``No such file or directory'' so we will
               assume the directory we want to create -- doesn't exist. */
            if ( the_error == ENOENT )
            {
                if ( mkdir( blob->report_information.directory, 0755 ))
                {
                    IAS_LOG_ERROR("Creating report files directory");
                    return ERROR;
                }
            }
            else
            {
                /* Something else happened */
                error_string = strerror( the_error );
                IAS_LOG_ERROR("%s", error_string );
                return ERROR;
            }
        }
    }

    /* Ensure the band information has been set before continuing.  No other
       checks will be made after this point. */
    if ( !get_band_fields_initialized() )
    {
        IAS_LOG_ERROR("The band fields in the BLOB haven't been initialized");
        return ERROR;
    }

    /* Initialize the thread pool */
    threadpool = ias_threadpool_initialize( blob->procs_to_use );
    if ( !threadpool )
    {
        IAS_LOG_ERROR("Creating thread pool");
        return ERROR;
    }

    /* Enable the work queue */
    status = ias_threadpool_enable_work_queue( threadpool );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Enabling the work queue");
        ias_threadpool_destroy( threadpool );
        return ERROR;
    }
    blob->threadpool = threadpool;

    /* Initialize the work assignment routine */
    status = initialize_work_assignment( blob->number_algorithms,
        blob->number_of_bands, blob->list_of_band_indices, threadpool );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Initializing work assignment");
        ias_threadpool_destroy( threadpool );
        return ERROR;
    }

    /* Execute any pre-process routines */
    i = 0;
    while ( blob->pre_proc_pointer_list[i] != NULL )
    {
        status = blob->pre_proc_pointer_list[i]( blob );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Executing pre-process routine for algorithm %s",
                blob->algorithm_string_list[i] );
            cleanup_work_assignment();
            ias_threadpool_destroy( threadpool );
            return ERROR;
        }
        i++;
    }

    /* Opening the output L1R bands has been delayed to this point in case
       any pre-process routine or parameter changes have altered the
       starting line and number of lines to process */
    status = open_l1r_bands( blob );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Opening the L1R bands");
        cleanup_work_assignment();
        ias_threadpool_destroy( threadpool );
        return ERROR;
    }

    /* Process SCAs needs to access the L1R data for writing so increment
       the consumer count.  Call the routine to set all the consumer counts
       in memory management.  No more increments will be allowed after this
       point. */
    increment_consumer_count( IMAGE_L1R );
    status = set_consumer_counts();
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Setting consumer counts");
        cleanup_work_assignment();
        ias_threadpool_destroy( threadpool );
        return ERROR;
    }

    /* Initialize the RPS level, this will set the level for all bands and
       SCAs to 0 */
    status = initialize_rps_level( blob->total_bands,
        blob->max_number_scas );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Initializing RPS level");
        cleanup_work_assignment();
        ias_threadpool_destroy( threadpool );
        return ERROR;
    }

    /* Open the pixel mask file and create a mutex to control access to
       the write routine */
    blob->pixel_mask_file_handle = ias_pm_open_pixel_mask(
        blob->pixel_mask_filename, IAS_WRITE );
    if ( ! blob->pixel_mask_file_handle )
    {
        IAS_LOG_ERROR("Opening pixel mask file");
        cleanup_work_assignment();
        ias_threadpool_destroy( threadpool );
        cleanup_io_routines( blob );
        cleanup_rps_level();
        return ERROR;
    }
    status = IAS_THREAD_CREATE_MUTEX( &blob->pixel_mask_io_mutex );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Creating mutex for pixel mask access");
        cleanup_work_assignment();
        ias_threadpool_destroy( threadpool );
        cleanup_io_routines( blob );
        cleanup_rps_level();
        ias_pm_close_pixel_mask( blob->pixel_mask_file_handle );
        return ERROR;
    }

    /* Turn the thread pool loose */
    status = ias_threadpool_run_function( threadpool, process_scas, blob );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Processing SCAs");
        cleanup_work_assignment();
        ias_threadpool_destroy( threadpool );
        cleanup_io_routines( blob );
        cleanup_rps_level();
        ias_pm_close_pixel_mask( blob->pixel_mask_file_handle );
        IAS_THREAD_DESTROY_MUTEX( &blob->pixel_mask_io_mutex );
        return ERROR;
    }

    /* Destroy the thread pool */
    ias_threadpool_destroy( threadpool );
    threadpool = NULL;

    /* Close the pixel mask and destroy the mutex.  Do not error here just
       throw the warning. */
    status = ias_pm_close_pixel_mask( blob->pixel_mask_file_handle );
    if ( status != SUCCESS )
    {
        IAS_LOG_WARNING("The pixel mask file wasn't closed properly");
        error_occurred = 1;
    }
    blob->pixel_mask_file_handle = NULL;
    status = IAS_THREAD_DESTROY_MUTEX( &blob->pixel_mask_io_mutex );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Destroying the pixel mask mutex");
        cleanup_work_assignment();
        cleanup_rps_level();
        return ERROR;
    }

    /* Add the list of algorithms to the metadata as a comma delimited
       list */
    algorithm_list = blob->l1r_file_metadata.
        custom_rad_processing_steps_applied;
    algorithm_list[0] = '\0';
    for ( i = 0; i < blob->number_algorithms - blob->number_dump_algorithms;
        i++ )
    {
        if ( i != 0 )
        {
            strncat( algorithm_list, ",", IAS_L1R_MAX_RAD_STEPS_SIZE - 1 -
                strlen(algorithm_list));
        }
        /* Let the next strncat catch any errors */

        strncat( algorithm_list, blob->algorithm_string_list[i],
            IAS_L1R_MAX_RAD_STEPS_SIZE - 1 - strlen(algorithm_list));
        if ( strlen( algorithm_list ) == IAS_L1R_MAX_RAD_STEPS_SIZE - 1 
            && i <= blob->number_algorithms - 1 )
        {
            IAS_LOG_ERROR("The metadata string for custom radiometric "
                "processing steps size of %d was exceeded",
                IAS_L1R_MAX_RAD_STEPS_SIZE );
            error_occurred = 1;
            break;
        }
    }

    /* Add the band specific values to the metadata */
    for ( index = 0; index < blob->number_of_bands; index++ )
    {
        int band_number = blob->list_of_bands[index];
        int band_index = blob->list_of_band_indices[index];
        const IAS_BAND_ATTRIBUTES *band_atts =
            ias_sat_attr_get_band_attributes( band_number );
        if ( !band_atts )
        {
            IAS_LOG_ERROR("Getting band attributes for band %d, L1R "
                "metadata not properly set",
                band_number );
            error_occurred = 1;
            break;
        }

        blob->l1r_band_metadata[band_index]->band_number = band_number;
        strcpy(blob->l1r_band_metadata[band_index]->band_name, 
            ias_sat_attr_convert_band_number_to_name( band_number ));
        strcpy(blob->l1r_band_metadata[band_index]->instrument_source,
            ias_sat_attr_get_sensor_name( band_atts->sensor_id ));
    }

    /* Execute any post-process routines */
    i = 0;
    while ( blob->post_proc_pointer_list[i] != NULL )
    {
        status = blob->post_proc_pointer_list[i]( blob );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Executing post-process routine");
            cleanup_work_assignment();
            ias_threadpool_destroy( threadpool );
            cleanup_io_routines( blob );
            cleanup_rps_level();
            return ERROR;
        }
        i++;
    }

    /* Set the L1R metadata for the final L1R Product.  Don't return on 
       ERROR since there are more things to clean up. */
    status = set_l1r_metadata( blob, NULL );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("L1R metadata not properly set");
        error_occurred = 1;
    }

    cleanup_rps_level();

    /* Close the input and output file, again don't return on ERROR */
    status = cleanup_io_routines( blob );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Cleaning up I/O routines");
        error_occurred = 1;
    }

    /* Clean up the work assignment routine */
    cleanup_work_assignment();

    if ( error_occurred )
        return ERROR;

    return SUCCESS;
}  /* END process_collection */
