/*----------------------------------------------------------------------------
NAME:    process_scas

PURPOSE: Process all SCAs through the algorithm list

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
#include <sys/time.h>
#include "ias_logging.h"
#include "rps_local.h"


int process_scas
(
    void *input_parms,  /* I: Collection of input parameters, including
                              algorithm list and pointer to pixel mask */
    int thread_num      /* I: running thread number */
)
{
    /* Cast input parameters to correct structure */
    BLOB *blob = input_parms;
    int band_number;
    int band_index;
    int sca_number;
    int sca_index;
    int algorithm_index;/* Index of the algorithm to execute */
    float *image_buffer;/* Pointer to the L1R image data */
    int status = 0;     /* Function return status.  Initialized to suppress
                           compiler warning. */

    /* Loop until done.  Termination logic is handled within the loop.  Set
       the band index to -1 for the first call to assign_work so that no
       work is set to complete. */
    band_index = -1;
    while ( 1 )
    {
        /* Get a work assignment.  Return if no work was found. */
        if ( ! assign_work( &band_index, &sca_index, &algorithm_index ))
            return SUCCESS;
        sca_number = sca_index + 1;
        band_number = get_band_attributes( band_index )->band_number;

        /* To provide some comfort factor to those that need it, output a
           processing message when SCA 1 is encountered */
        if ( sca_number == 1 && algorithm_index == 0 )
            IAS_LOG_INFO("Processing band %d", band_number );
        if ( ias_log_message_level == IAS_LOG_LEVEL_DEBUG )
        {
            struct timeval tv;
            gettimeofday( &tv, NULL );
            IAS_LOG_DEBUG("Processing %ld.%-6ld %s for Band %2d SCA %2d",
                tv.tv_sec, tv.tv_usec,
                blob->algorithm_string_list[algorithm_index], band_number,
                sca_number );
        }

        /* Call the algorithm API */
        status = blob->algorithm_pointer_list[algorithm_index]( blob, 
            band_index, sca_index, algorithm_index );
        if ( status != SUCCESS )
        {
            set_work_assignment_error_condition();
            IAS_LOG_ERROR( "Returned from: %s for band %d SCA %d",
                blob->algorithm_string_list[algorithm_index], band_number,
                sca_number );
            return ERROR;
        }

        /* Check to see if this is the last algorithm to be run.  If it is,
           write and release the L1R imagery.  Write out the pixel mask for
           this SCA.  Since the index is 0-based check against one less than
           total number of algorithms. */
        if ( algorithm_index == blob->number_algorithms - 1 )
        {
            image_buffer = (float *)get_memory_buffer( IMAGE_L1R, CONSUMER,
                band_index, sca_index );
            if ( !image_buffer )
            {
                set_work_assignment_error_condition();
                IAS_LOG_ERROR("Didn't get an L1R image buffer");
                return ERROR;
            }
            else
            {
                /* Variables to manage freeing the pixel masks associated
                   with the current band */
                IAS_PIXEL_MASK *masks_to_free[4];
                int mask_count;
                int i;

                RPS_BAND_ATTRIBUTES *band_attributes =
                    get_band_attributes( band_index );

                /* Only write normal bands */
                if ( band_attributes->is_normal_band )
                {
                    status = write_l1r_output_image_data( band_index,
                        sca_index, image_buffer );
                    if ( status != SUCCESS )
                    {
                        set_work_assignment_error_condition();
                        IAS_LOG_ERROR(
                            "Writing L1R image buffer for band %d SCA %d",
                            band_number, sca_number );
                        return ERROR;
                    }
                }
                status = decrement_consumer_count( IMAGE_L1R, band_index,
                    sca_index );
                if ( status != SUCCESS )
                {
                    set_work_assignment_error_condition();
                    IAS_LOG_ERROR("Decrementing memory buffer");
                    return ERROR;
                }
                if ( band_attributes->vrp_band_number > 0 )
                {
                    status = decrement_consumer_count( IMAGE_L1R,
                        band_attributes->vrp_band_index, sca_index );
                    if ( status != SUCCESS )
                    {
                        set_work_assignment_error_condition();
                        IAS_LOG_ERROR("Decrementing memory buffer");
                        return ERROR;
                    }
                }
                if ( band_attributes->blind_band_number > 0
                    && band_attributes->process_blind_band_flag )
                {
                    status = decrement_consumer_count( IMAGE_L1R,
                        band_attributes->blind_band_index, sca_index );
                    if ( status != SUCCESS )
                    {
                        IAS_LOG_ERROR("Decrementing memory buffer");
                        set_work_assignment_error_condition();
                        return ERROR;
                    }
                }

                /* Empty the array of masks */
                for ( i = 0; i < 4; i++ )
                    masks_to_free[i] = NULL;
                mask_count = 0;

                /* Write the pixel mask -- This has to be done for
                   associated bands as well */
                if( IAS_THREAD_LOCK_MUTEX( &blob->pixel_mask_io_mutex ))
                {
                    IAS_LOG_ERROR("Locking the pixel mask mutex");
                    set_work_assignment_error_condition();
                    return ERROR;
                }
                status = ias_pm_write_single_mask_to_file(
                    blob->pixel_mask_file_handle,
                    blob->pixel_mask[band_index][sca_index] );
                if ( status != SUCCESS )
                {
                    IAS_LOG_ERROR("Writing pixel mask for band %d "
                        "SCA %d", band_number, sca_number );
                    set_work_assignment_error_condition();
                    IAS_THREAD_UNLOCK_MUTEX( &blob->pixel_mask_io_mutex );
                    return ERROR;
                }

                /* The purpose of writing the mask above is to prevent the
                   situation where there are a huge number of artifacts and
                   memory usage gets out of control.  We need to free this
                   pixel mask but won't do so until we unlock the mutex.
                   Save a pointer to the mask in a local variable and count
                   it.  It is important to set the pixel mask pointer to
                   NULL since freeing the blob will clean up on error
                   conditions.  The same will be done for each associated
                   mask below. */
                masks_to_free[mask_count++] =
                    blob->pixel_mask[band_index][sca_index];
                blob->pixel_mask[band_index][sca_index] = NULL;
                /* Mask clean up must occur on error from this point on */


                /* VRP pixel mask */
                if ( band_attributes->vrp_band_number > 0 )
                {
                    status = ias_pm_write_single_mask_to_file(
                        blob->pixel_mask_file_handle,
                        blob->pixel_mask[band_attributes->vrp_band_index]
                        [sca_index] );
                    if ( status != SUCCESS )
                    {
                        IAS_LOG_ERROR("Writing pixel mask for band %d "
                            "SCA %d", band_attributes->vrp_band_number,
                            sca_number );
                        set_work_assignment_error_condition();
                        IAS_THREAD_UNLOCK_MUTEX( &blob->pixel_mask_io_mutex );
                        for ( i = 0; i < mask_count; i++ )
                            ias_pm_destroy( masks_to_free[i] );
                        return ERROR;
                    }
                    /* Save the pointer and set the mask to NULL */
                    masks_to_free[mask_count++] =
                        blob->pixel_mask[band_attributes->vrp_band_index]
                        [sca_index];
                    blob->pixel_mask [band_attributes->vrp_band_index]
                        [sca_index] = NULL;
                }
                /* Blind pixel mask */
                if ( band_attributes->blind_band_number > 0 
                    && band_attributes->process_blind_band_flag )
                {
                    const IAS_BAND_ATTRIBUTES *atts =
                        ias_sat_attr_get_band_attributes(
                        band_attributes->blind_band_number );
                    int vrp_number;
                    int vrp_index = -1;

                    if ( ! atts )
                    {
                        IAS_LOG_ERROR("Getting attributes for band %d",
                            band_attributes->blind_band_number );
                        set_work_assignment_error_condition();
                        IAS_THREAD_UNLOCK_MUTEX( &blob->pixel_mask_io_mutex );
                        for ( i = 0; i < mask_count; i++ )
                            ias_pm_destroy( masks_to_free[i] );
                        return ERROR;
                    }
                    vrp_number = atts->vrp_band_number;
                    if ( vrp_number > 0 )
                    {
                        vrp_index = ias_sat_attr_convert_band_number_to_index(
                            vrp_number );
                        if ( vrp_index < 0 )
                        {
                            IAS_LOG_ERROR("Converting band number to index"
                                " for band %d", vrp_number );
                            set_work_assignment_error_condition();
                            IAS_THREAD_UNLOCK_MUTEX(
                                &blob->pixel_mask_io_mutex );
                            for ( i = 0; i < mask_count; i++ )
                                ias_pm_destroy( masks_to_free[i] );
                            return ERROR;
                        }
                    }

                    status = ias_pm_write_single_mask_to_file(
                        blob->pixel_mask_file_handle,
                        blob->pixel_mask[band_attributes->blind_band_index]
                        [sca_index] );
                    if ( status != SUCCESS )
                    {
                        IAS_LOG_ERROR("Writing pixel mask for band %d "
                            "SCA %d", band_attributes->blind_band_number,
                            sca_number );
                        set_work_assignment_error_condition();
                        IAS_THREAD_UNLOCK_MUTEX( &blob->pixel_mask_io_mutex );
                        for ( i = 0; i < mask_count; i++ )
                            ias_pm_destroy( masks_to_free[i] );
                        return ERROR;
                    }
                    /* Save the pointer and set the mask to NULL */
                    masks_to_free[mask_count++] =
                        blob->pixel_mask[band_attributes->blind_band_index]
                        [sca_index];
                    blob->pixel_mask [band_attributes->blind_band_index]
                        [sca_index] = NULL;

                    /* If there is an associated VRP for this blind band
                       it also needs to be written out */
                    if ( vrp_number > 0 )
                    {
                        status = ias_pm_write_single_mask_to_file(
                            blob->pixel_mask_file_handle,
                            blob->pixel_mask[vrp_index][sca_index] );
                        if ( status != SUCCESS )
                        {
                            IAS_LOG_ERROR("Writing pixel mask for band %d "
                                "SCA %d", vrp_number, sca_number );
                            set_work_assignment_error_condition();
                            IAS_THREAD_UNLOCK_MUTEX(
                                &blob->pixel_mask_io_mutex );
                            for ( i = 0; i < mask_count; i++ )
                                ias_pm_destroy( masks_to_free[i] );
                            return ERROR;
                        }
                        /* Save the pointer and set the mask to NULL */
                        masks_to_free[mask_count++] =
                            blob->pixel_mask[vrp_index][sca_index];
                        blob->pixel_mask[vrp_index][sca_index] = NULL;
                    }
                }
                if ( IAS_THREAD_UNLOCK_MUTEX( &blob->pixel_mask_io_mutex ))
                {
                    IAS_LOG_ERROR("Unlocking pixel mask mutex");
                    set_work_assignment_error_condition();
                    for ( i = 0; i < mask_count; i++ )
                        ias_pm_destroy( masks_to_free[i] );
                    return ERROR;
                }

                /* Free the pixel masks */
                for ( i = 0; i < mask_count; i++ )
                    ias_pm_destroy( masks_to_free[i] );
            }
        }
    }  /* END infinite while loop */

    return SUCCESS;
}  /* END process_scas */
