/*----------------------------------------------------------------------------
NAME:    read_parameters

PURPOSE: Read the input ODL file and populate the blob with parameters

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>

#include "ias_logging.h"
#include "ias_parm.h"
#include "ias_parm_standard_parameters.h"
#include "ias_satellite_attributes.h"
#include "rps_local.h"

static int number_algorithms;
static int total_valid_algorithms;

#define MODIFIER_LENGTH  6 /* Length of the dump and skip modifiers */

/*----------------------------------------------------------------------------
NAME:    free_valid_algorithm_list

PURPOSE: Local routine to free the list of valid algorithms

RETURNS: none
-----------------------------------------------------------------------------*/
void free_valid_algorithm_list
(
    char **valid_algorithms /* I: Pointer to list to be freed */
)
{
    int i;
    for ( i = 0; i < total_valid_algorithms; i++ )
        free( valid_algorithms[i] );
    free( valid_algorithms );
}

/*----------------------------------------------------------------------------
NAME:    create_valid_algorithm_list

PURPOSE: Local routine to create a list of valid algorithms.  It uses the
         data structure defined elsewhere to create the list.

RETURNS: Pointer to the list
-----------------------------------------------------------------------------*/
char **create_valid_algorithm_list( void )
{
    int i;
    int index;
    const ALG_TYPE *available_algorithms = get_available_algorithms(
        &number_algorithms );
    char **valid_algorithms;
    char skip[MODIFIER_LENGTH] = "_SKIP";
    char dump[MODIFIER_LENGTH] = "_DUMP";

    /* Allocate the memory, there are two modifiers so allocate three times
       as many pointers */
    total_valid_algorithms = 3 * number_algorithms;
    valid_algorithms = calloc( total_valid_algorithms, sizeof(char *));
    if ( !valid_algorithms )
        return NULL;

    index = 0;
    for ( i = 0; i < number_algorithms; i++ )
    {
        const char *algorithm = available_algorithms[i].algorithm_name;
        size_t length = strlen( algorithm ) + MODIFIER_LENGTH;

        /* The structure member algorithm_name contains the base name of
           the algorithm.  Duplicate this name in the local array. */
        valid_algorithms[index] =
            strdup( available_algorithms[i].algorithm_name );
        if ( valid_algorithms[index] == NULL )
        {
            free_valid_algorithm_list( valid_algorithms );
            return NULL;
        }

        /* There are two modifiers to the base algorithm name so allocate
           memory for the base name plus the modifier and copy the base name
           then cat the appropriate modifier.  This section creates the name
           with the skip modifier. */
        index++;
        valid_algorithms[index] = malloc( length * sizeof(char));
        if ( valid_algorithms[index] == NULL )
        {
            free_valid_algorithm_list( valid_algorithms );
            return NULL;
        }
        strcpy( valid_algorithms[index], algorithm );
        strcat( valid_algorithms[index], skip );

        /* This section creates the name with the dump modifier */
        index++;
        valid_algorithms[index] = malloc( length * sizeof(char));
        if ( valid_algorithms[index] == NULL )
        {
            free_valid_algorithm_list( valid_algorithms );
            return NULL;
        }
        strcpy( valid_algorithms[index], algorithm );
        strcat( valid_algorithms[index], dump );

        index++;
    }

    return valid_algorithms;
}


/*----------------------------------------------------------------------------
NAME:    read_parameters

PURPOSE: Read the input ODL file and populate the blob with parameters

RETURNS: exits with EXIT_SUCCESS when caller asks for help and help call
             succeeded
         otherwise SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int read_parameters
(
    BLOB *blob   /* I: parameter and OMF file names
                    O: parameters listed in input files */
)
{
    /* Set of flags indicating a band number is already in the band list */
    int band_present_in_list[IAS_MAX_NBANDS] = {0};
    int status;  /* Routine return status */
    int count;   /* Count of elements in the parameter definition tables */
    int i;       /* Plain old loop counter */
    int band_index;
    int generate_report = 0;
    int apply_absolute_gain = 0;
    int post_warmup_nframes;
    double warmup_time_period;
    int apply_linearization_correction = 0;
    int apply_relative_gain = 0;
    int perform_temp_sens = 0;
    int skip_time_limit_check = 0;
    int store_in_database = 0;
    int use_tirs_cpf_baseline_response = 1;
    int write_tirs_dark_response_to_file = 0;
    const IAS_SATELLITE_ATTRIBUTES *sat_attributes;
    int list_of_bands[50];
    int satellite_number;
    IAS_SATELLITE_ID satellite_id;
    int stray_light_filename_present;
    int local_algorithm_count;
    int algorithm_index;
    char local_tirs_stray_light_correction_filename[PATH_MAX];

    /* Local copy of the string list.  We have this because we don't know
       how many algorithms have been specified and it isn't directly
       obtainable until the list has been read.  This has been sized so that
       there should never be a problem in production. */
    char *algorithm_string_list[1000];

    /* In order to provide a list of the externally available algorithms as
       the valid options for the ALGORITHMS parameter, get a pointer to the
       algorithm list array and set the valid strings. */
    char **valid_algorithms = create_valid_algorithm_list();
    if ( valid_algorithms == NULL )
    {
        IAS_LOG_ERROR("Creating valid algorithm list");
        return ERROR;
    }


    /*-----------------------------------------------------------------*/
    /* This is the table definition for things from the parameter file */
    /*-----------------------------------------------------------------*/
    IAS_PARM_DECLARE_TABLE( parms, 30 );

    IAS_PARM_WORK_ORDER_ID( parms, blob->work_order_id,
        sizeof(blob->work_order_id), 1 );

    /* Add the Algorithms */
    IAS_PARM_ADD_STRING_ARRAY( parms, ALGORITHMS, "List of RPS algorithms",
        IAS_PARM_REQUIRED,
        total_valid_algorithms, (const char **)valid_algorithms,/* valid list */
        0, NULL, /* no default */
        algorithm_string_list, sizeof(algorithm_string_list), 0 );

    IAS_PARM_BAND_LIST( parms, IAS_PARM_OPTIONAL, list_of_bands,
        sizeof(list_of_bands), 1 );

    /* Add the L1R file name */
    const char *default_l1r_file_name[] = {"l1r_image.h5"};
    IAS_PARM_ADD_STRING( parms, L1R_FILENAME, "L1R image file name",
        IAS_PARM_OPTIONAL,
        0, NULL, /* no restrictions */
        1, default_l1r_file_name, /* Default file name */
        blob->l1r_file_name, sizeof(blob->l1r_file_name), 0 );

    /* Add the dump output directory */
    const char *default_dump_directory[] = {"intermediate_dumps"};
    IAS_PARM_ADD_STRING( parms, DUMP_DIRECTORY, "L1R image dump file directory",
        IAS_PARM_OPTIONAL,
        0, NULL, /* no restrictions */
        1, default_dump_directory, /* Default file name */
        blob->dump_directory, sizeof(blob->dump_directory), 0 );

    /* Add the report output directory */
    const char *default_report_directory[] = {"report_files"};
    IAS_PARM_ADD_STRING( parms, REPORT_DIRECTORY, 
        "Output report file directory", IAS_PARM_OPTIONAL,
        0, NULL, /* no restrictions */
        1, default_report_directory, /* Default file name */
        blob->report_directory, 
        sizeof(blob->report_directory), 0 );

    /* Add the pixel mask file name */
    const char *default_pixel_mask_filename[] = {"pixel_mask.dat"};
    IAS_PARM_ADD_STRING( parms, PIXEL_MASK_FILENAME, "Pixel mask file name",
        IAS_PARM_OPTIONAL,
        0, NULL, /* no restrictions */
        1, default_pixel_mask_filename, /* Default file name */
        blob->pixel_mask_filename, sizeof(blob->pixel_mask_filename), 0 );

    /* Disable maximum time limit of the L0R. (Intended for use in Dev only) */
    int disable_l1r_size_limit_default = 0;
    IAS_PARM_ADD_BOOLEAN(parms, DISABLE_TIME_LIMIT_CHECK,
        "Skip checking the time limit. 1 = Skip check 0 = Perform check",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        &disable_l1r_size_limit_default,
        &skip_time_limit_check,
        sizeof(skip_time_limit_check), 0);

    /* Linearization parameter correction */
    int linearization_correction_default = 0;
    IAS_PARM_ADD_BOOLEAN(parms, APPLY_LINEARIZATION_CORRECTION,
        "Apply correction for non-linear detector response: 1 = apply, "
        "0 = don't apply",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        &linearization_correction_default,
        &apply_linearization_correction,
        sizeof(apply_linearization_correction), 0);

    /* Gain application parameters */
    int abs_gain_default = 1;
    IAS_PARM_ADD_BOOLEAN( parms, APPLY_ABSOLUTE_GAIN,
        "Apply absolute gain correction: 1 = apply, 0 = don't apply",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        &abs_gain_default, &apply_absolute_gain,
        sizeof(apply_absolute_gain) , 0 );

    int relative_gain_default = 1;
    IAS_PARM_ADD_BOOLEAN( parms, APPLY_RELATIVE_GAIN,
        "Apply relative gain correction: 1 = apply, 0 = don't apply",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        &relative_gain_default, &apply_relative_gain,
        sizeof(apply_relative_gain) , 0 );

    int temp_sens_default = 0;
    IAS_PARM_ADD_BOOLEAN( parms, PERFORM_TEMP_SENS,
        "Perform temperature sensitivity correction: 1 = perform, 0 = don't "
        "perform", IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        &temp_sens_default, &perform_temp_sens,
        sizeof(perform_temp_sens) , 0 );

    /* Detector response characterization for lamp parameters, specifically
       the total number of FRAMES of post-warmup lamp data */
    int post_warmup_nframes_default = 512;
    IAS_PARM_ADD_INT( parms, POST_WARMUP_NUMBER_FRAMES,
        "Number of post-warmup lamp frames to process", IAS_PARM_OPTIONAL,
        IAS_PARM_NOT_ARRAY, 0, 0, 0, 1, &post_warmup_nframes_default,
        &post_warmup_nframes,
        sizeof(post_warmup_nframes) , 0 );

    /* Detector response characterization for lamp parameters */
    double warmup_time_period_default = 150.0;
    IAS_PARM_ADD_DOUBLE( parms, WARMUP_TIME_PERIOD,
        "Time period, in seconds, for lamp warmup", IAS_PARM_OPTIONAL,
        IAS_PARM_NOT_ARRAY, 0, 0, 0, 1, &warmup_time_period_default,
        &warmup_time_period,
        sizeof(warmup_time_period) , 0 );

    /* TIRS dark response determination parameters */
    int use_tirs_cpf_baseline_response_default = 1;
    IAS_PARM_ADD_BOOLEAN( parms, USE_TIRS_CPF_BASELINE_RESPONSE,
        "Use TIRS CPF baseline dark response: 1 = use, 0 = don't use",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        &use_tirs_cpf_baseline_response_default,
        &use_tirs_cpf_baseline_response,
        sizeof(use_tirs_cpf_baseline_response), 1 );

    int write_tirs_dark_response_to_file_default = 0;
    IAS_PARM_ADD_BOOLEAN( parms, WRITE_TIRS_DARK_RESPONSE_FILE,
        "Write TIRS dark response to file: 1 = write file, 0 = don't write",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        &write_tirs_dark_response_to_file_default,
        &write_tirs_dark_response_to_file,
        sizeof(write_tirs_dark_response_to_file), 1 );

    int process_tirs_blind_band_default = 0;
    IAS_PARM_ADD_BOOLEAN( parms, PROCESS_TIRS_BLIND_BAND,
        "Process the TIRS blind band for dark response determination: "
        "1 = process, 0 = don't process",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        &process_tirs_blind_band_default,
        &blob->process_tirs_blind_band,
        sizeof(blob->process_tirs_blind_band), 1 );

    /* Report and database options */
    int report_default = 0;
    IAS_PARM_ADD_BOOLEAN( parms, GENERATE_REPORT,
        "1 = generate report files, 0 = don't generate reports",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, &report_default,
        &generate_report, sizeof(generate_report), 0 );

    IAS_PARM_STORE_CHARACTERIZATION_IN_DB( parms, IAS_PARM_OPTIONAL,
        &store_in_database, sizeof(store_in_database), 0);

    /* For OLI bias determination algorithm */
    int bias_source_default = 2;
    IAS_PARM_ADD_INT( parms, BIAS_SOURCE, "Source of per-detector bias "
        "selected: 0 = pre acquisition shutter, 1 = post acquisition shutter, "
        "2 = mean acquisition shutter, 3 = bias from cpf, 4 = a1_c1_estimation",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 1, 0, 4, 1,
        &bias_source_default, &blob->bias_source,
        sizeof(blob->bias_source), 0 );

    int per_frame_bias_flag_default = 0;
    IAS_PARM_ADD_BOOLEAN( parms, PER_FRAME_BIAS_FLAG, "The type of bias "
        "selected in OLI Bias Determination: 0 = per detector bias, "
        "1 = per frame bias", IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 
        &per_frame_bias_flag_default, &blob->per_frame_bias_flag, 
        sizeof(blob->per_frame_bias_flag), 0 );     

    /* For TIRS Bias Removal algorithm */
    int background_response_selection_default = 3;
    IAS_PARM_ADD_INT( parms, BACKGROUND_RESPONSE_SELECTION, "Selection for the"
        " background response removal: 1 = BPF pre-acquisition deep space, "
        "2 = BPF post-acquisition deep space, 3 BPF mean deep space, "
        "4 = CPF dark and background response", 
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 1, 1, 4, 1, 
        &background_response_selection_default, 
        &blob->background_response_selection,
        sizeof(blob->background_response_selection), 0 );

    /* For Replace saturated pixel algorithm */
    int analog_high_saturation_flag_default = 0;
    IAS_PARM_ADD_BOOLEAN( parms, ANALOG_HIGH_SATURATION_FLAG,
        "Replace the Analog High Saturated Pixels: "
        "1 = replace with High_Radiance_Saturation value, 0 = don't replace",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        &analog_high_saturation_flag_default,
        &blob->analog_high_saturation_flag,
        sizeof(blob->analog_high_saturation_flag), 1 );

    int analog_low_saturation_flag_default = 0;
    IAS_PARM_ADD_BOOLEAN( parms, ANALOG_LOW_SATURATION_FLAG,
        "Replace the Analog Low Saturated Pixels: "
        "1 = replace with Low_Radiance_Saturation value, 0 = don't replace",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        &analog_low_saturation_flag_default,
        &blob->analog_low_saturation_flag,
        sizeof(blob->analog_low_saturation_flag), 1 );

    int digital_high_saturation_flag_default = 1;
    IAS_PARM_ADD_BOOLEAN( parms, DIGITAL_HIGH_SATURATION_FLAG,
        "Replace the Digital High Saturated Pixels: "
        "1 = replace with High_Radiance_Saturation value, 0 = don't replace",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        &digital_high_saturation_flag_default,
        &blob->digital_high_saturation_flag,
        sizeof(blob->digital_high_saturation_flag), 1 );

    int digital_low_saturation_flag_default = 1;
    IAS_PARM_ADD_BOOLEAN( parms, DIGITAL_LOW_SATURATION_FLAG,
        "Replace the Digital Low Saturated Pixels: "
        "1 = replace with Low_Radiance_Saturation value, 0 = don't replace",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        &digital_low_saturation_flag_default,
        &blob->digital_low_saturation_flag,
        sizeof(blob->digital_low_saturation_flag), 1 );
    
    /* Populate database with summary statistics */
    int default_populate_scene_summary_stats_flag = 0;
    IAS_PARM_ADD_BOOLEAN(parms, POPULATE_SCENE_SUMMARY_STATS_FLAG,
        "Populate the histogram scene summary statistics for Earth Collections"
        " 1 = populate, 0 = don't populate", IAS_PARM_OPTIONAL,
        IAS_PARM_NOT_ARRAY, &default_populate_scene_summary_stats_flag,
        &blob->populate_scene_summary_stats_flag,
        sizeof(blob->populate_scene_summary_stats_flag), 0);

    /* Flag to indicate that the tirs scenes will consist of 3 scenes */
    int default_use_three_tirs_scenes_flag[] = {0};
    IAS_PARM_USE_THREE_TIRS_SCENES_FLAG(parms, IAS_PARM_OPTIONAL,
        default_use_three_tirs_scenes_flag, &blob->use_three_tirs_scenes_flag,
        sizeof(blob->use_three_tirs_scenes_flag), 1);

    /* Add the stray light correction filename */
    IAS_PARM_STRAY_LIGHT_CORRECTION_FILENAME(parms, IAS_PARM_OPTIONAL,
        local_tirs_stray_light_correction_filename,
        sizeof(local_tirs_stray_light_correction_filename), 0);

    /*----------------------------------------------------------------*/
    /*----- This is the table definition for things from the OMF -----*/
    /*----------------------------------------------------------------*/
    IAS_PARM_DECLARE_TABLE( omf_parms, 9 );

    IAS_PARM_L0R_FILENAME( omf_parms, IAS_PARM_REQUIRED, blob->l0r_file_name,
        sizeof(blob->l0r_file_name), 0 );
    IAS_PARM_SATELLITE( omf_parms, &satellite_number,
        sizeof(satellite_number), 1 );
    IAS_PARM_CPF( omf_parms, 1, blob->cpf_filename,
        sizeof(blob->cpf_filename), 1 );
    IAS_PARM_BPF_OLI( omf_parms, 0, blob->bpf_filename_oli,
        sizeof(blob->bpf_filename_oli), 1 );
    IAS_PARM_BPF_TIRS( omf_parms, 0, blob->bpf_filename_tirs,
        sizeof(blob->bpf_filename_tirs), 1 );
    IAS_PARM_RLUT(omf_parms, 1, blob->rlut_filename,
        sizeof(blob->rlut_filename), 1);
    strcpy(blob->landsat_scene_id, "");
    IAS_PARM_LANDSAT_SCENE_ID( omf_parms, IAS_PARM_OPTIONAL, 
        blob->landsat_scene_id, sizeof(blob->landsat_scene_id), 1 );
    IAS_PARM_MAX_PROCESSORS_TO_USE( omf_parms, IAS_PARM_OPTIONAL,
        &blob->procs_to_use, sizeof(blob->procs_to_use), 1 );

    /* Add the stray light correction filename */
    IAS_PARM_STRAY_LIGHT_CORRECTION_FILENAME(omf_parms, IAS_PARM_OPTIONAL,
        blob->tirs_stray_light_correction_filename,
        sizeof(blob->tirs_stray_light_correction_filename), 0);

    /* See if the user wanted help or a template */
    status = ias_parm_provide_help( blob->parameter_file_name, parms,
        IAS_PARM_GET_TABLE_SIZE( parms ), IAS_INPUT_PARAMETERS);
    if ( status == 1 )
    {
        /* If the status of the provide help is 1 the user asked for help.
           Since the parameter file name contains the help request, copy it
           to the omf file name so the next call will ask for the same help */
        strcpy( blob->omf_file_name, blob->parameter_file_name );
    }
    else if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Returned from parameter file help");
        free_valid_algorithm_list( valid_algorithms );
        return ERROR;
    }
    else
    {
        /* Read the parameter file */
        count = IAS_PARM_GET_TABLE_SIZE(parms);
        status = ias_parm_read( blob->parameter_file_name, parms, count );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR( "Reading parameters from: %s",
                blob->parameter_file_name );
            free_valid_algorithm_list( valid_algorithms );
            return ERROR;
        }

        /* Build the omf file name */
        sprintf( blob->omf_file_name, "%s.omf", blob->work_order_id );
    }

    /* See if the user wanted help or a template this time with the OMF file */
    status = ias_parm_provide_help( blob->omf_file_name, omf_parms,
        IAS_PARM_GET_TABLE_SIZE( omf_parms ), IAS_OMF_PARAMETERS);
    if ( status == 1 )
    {
        exit( EXIT_SUCCESS );
    }
    else if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Returned from OMF parameter file help");
        free_valid_algorithm_list( valid_algorithms );
        return ERROR;
    }

    /* Read the OMF file */
    count = IAS_PARM_GET_TABLE_SIZE( omf_parms );
    status = ias_parm_read( blob->omf_file_name, omf_parms, count );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Reading parameters from OMF file: %s",
            blob->omf_file_name );
        free_valid_algorithm_list( valid_algorithms );
        return ERROR;
    }

    /* The algorithm list is only needed for the provide help routines 
       so it can be freed now */
    free_valid_algorithm_list( valid_algorithms );
    valid_algorithms = NULL;

    /* Set the report and database flags.  Must be done prior to doing the
       final stages of blob initialization. */
    if ( generate_report )
        blob->generate_report = 1;
    else
        blob->generate_report = 0;
    if ( store_in_database )
        blob->store_in_database = 1;
    else
        blob->store_in_database = 0;

    /* Get the platform name and initialize the satellite attributes. */
    satellite_id =
        ias_sat_attr_get_satellite_id_from_satellite_number( satellite_number );
    if ( ias_sat_attr_initialize( satellite_id ))
    {
        IAS_LOG_ERROR("Initializing satellite attributes");
        return ERROR;
    }

    /* Now that we know the sensor, finish up the blob initialization */
    sat_attributes = ias_sat_attr_get_attributes();
    blob->total_bands = sat_attributes->total_bands;
    status = finish_blob_initialization();
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Completing blob initialization");
        return ERROR;
    }

    /* Populate some blob structure fields */
    blob->skip_time_limit_check = skip_time_limit_check;
    blob->apply_linearization_correction = apply_linearization_correction;

    /* Add warning message if absolute gain is turned off and relative gain is
       turned on */
    if (apply_absolute_gain == 0 && apply_relative_gain == 1)
    {
        IAS_LOG_WARNING("Absolute gain off, relative gain on");
    }

    if ( apply_absolute_gain )
        blob->apply_absolute_gain = 1;
    else
        blob->apply_absolute_gain = 0;

    if ( apply_relative_gain )
        blob->apply_relative_gain = 1;
    else
        blob->apply_relative_gain = 0;

    if ( perform_temp_sens )
        blob->perform_temp_sens = 1;
    else
        blob->perform_temp_sens = 0;

    stray_light_filename_present = IAS_PARM_GET_COUNT_READ(omf_parms,
        STRAY_LIGHT_CORRECTION_FILENAME );
    blob->use_tirs_ghost_image_flag = 0;

    /* Copy the override filename and notify the user */
    if (IAS_PARM_GET_COUNT_READ(parms,STRAY_LIGHT_CORRECTION_FILENAME) > 0)
    {
        strcpy(blob->tirs_stray_light_correction_filename,
            local_tirs_stray_light_correction_filename);
        stray_light_filename_present = TRUE;
        IAS_LOG_INFO("Overriding stray light correction filename "
            "to %s", blob->tirs_stray_light_correction_filename);
    }

    /* Find out how many algorithms were specified.  Allocate the blob
       string list and copy the string pointers to it. */
    blob->number_algorithms = IAS_PARM_GET_COUNT_READ( parms, ALGORITHMS );
    local_algorithm_count = blob->number_algorithms;
    blob->number_of_wo_parm_strings_allocated = blob->number_algorithms;
    blob->algorithm_wo_parameter_list = malloc( sizeof(char *) *
        blob->number_of_wo_parm_strings_allocated );
    if ( !blob->algorithm_wo_parameter_list )
    {
        IAS_LOG_ERROR("Allocating the blob's algorithm string list");
        return ERROR;
    }

    for ( i = 0, algorithm_index = 0; i < local_algorithm_count; i++ )
    {
        if (strcmp(algorithm_string_list[i], "tirs_stray_light_correction") 
            == 0)
        {
            if (!stray_light_filename_present)
            {
                IAS_LOG_INFO("Stray light correction filename missing. "
                    "Skipping tirs_stray_light_correction algorithm");

                /* Perform cleanup of unused algorithm */
                free(algorithm_string_list[i]);
                algorithm_string_list[i] = NULL;
                blob->number_algorithms--;
                blob->number_of_wo_parm_strings_allocated--;
                continue;
            }
            blob->use_tirs_ghost_image_flag = 1;
        }
        
        /* Add the algorithm to the list */
        blob->algorithm_wo_parameter_list[algorithm_index] 
            = algorithm_string_list[i];
        algorithm_index++;
    }

    /* Set how many bands there are to process and set the flags for
       processing them; make sure the current band has not already been
       found in the band list.  Processing options are set only for the
       image bands, including blind band, not the VRPs.  The VRPs will be
       handled as part of the band processing. */
    blob->number_of_bands = IAS_PARM_GET_COUNT_READ( parms, BAND_LIST );
    if (blob->number_of_bands == 0)
    {
        IAS_LOG_ERROR("No bands found in list");
        return ERROR;
    }

    /* Initialize processing band flags */
    for ( band_index = 0; band_index <= blob->total_bands; band_index++ ) 
        blob->band_information[band_index]->process_band = 0;

    for ( i = 0; i < blob->number_of_bands; i++ )
    {
        blob->list_of_bands[i] = list_of_bands[i];
        blob->list_of_band_indices[i] =
            ias_sat_attr_convert_band_number_to_index(list_of_bands[i]);
        if (blob->list_of_band_indices[i] == ERROR)
        {
            IAS_LOG_ERROR("Converting band %d to corresponding index",
                list_of_bands[i]);
            return ERROR;
        }
        else if (band_present_in_list[blob->list_of_band_indices[i]])
        {
            IAS_LOG_ERROR("Duplicate band %d found in list",
                list_of_bands[i]);
            return ERROR;
        }
        else
        {
            /* First occurrence of band in list, flag it as 'found' */
            band_present_in_list[blob->list_of_band_indices[i]]++;

            /* Set flag to indicate this band should be processed */
            blob->band_information[list_of_bands[i]]->process_band = 1;
        }
    }

    /* Set the number of the stim lamp post warmup frames */
    blob->post_warmup_nframes = post_warmup_nframes;

    /* Set the lamp warmup time period */
    blob->lamp_warmup_time_period = warmup_time_period;

    /* Set the TIRS dark response parameters */
    if (use_tirs_cpf_baseline_response)
        blob->use_tirs_cpf_baseline_response = 1;
    else
        blob->use_tirs_cpf_baseline_response = 0;

    if (write_tirs_dark_response_to_file)
        blob->write_tirs_dark_response_to_file = 1;
    else
        blob->write_tirs_dark_response_to_file = 0;

    return SUCCESS;
}  /* END read_parameters */
