#ifndef RPS_BLOB_H
#define RPS_BLOB_H

#include <limits.h>

#include "ias_cpf.h"
#include "ias_bpf.h"
#include "ias_l0r.h"
#include "ias_l1r.h"
#include "ias_pixel_mask.h"
#include "ias_rlut.h"
#include "ias_satellite_attributes.h"
#include "ias_radiometry.h"

/* This structure contains the basic information about a band that RPS needs
   for processing.  It is used instead of the IAS attributes because of the 
   need to know other bits of information such as number of lines and whether
   or not to process the band.  Also used to track if only one SCA is
   present and what that SCA number is. */
typedef struct rps_band_attributes
{
    int band_present;    /* true indicates the band is in in L0R data */
    int process_band;    /* true indicates the RPS should process the band */
    int one_sca_present; /* true means only one SCA is present, false means
                            all SCAs present */
    int is_normal_band;  /* true indicates this is a normal band */
    int sca_number;      /* If only one SCA is present this is its number */
    int band_number;
    int vrp_band_number;
    int vrp_band_index;
    int blind_band_number;
    int blind_band_index;
    int process_blind_band_flag; /* Since there is a one to many association
                                    with the blind band, only one image band
                                    should process the blind band.  The flag
                                    will be set to 1 (true) by the first
                                    associated band to indicate it will
                                    process the band. */
    int starting_line;   /* The starting line to process.  This will default
                            to 0.  For special processing it will be set to
                            the actual starting line to process. */
    int number_lines;    /* Number of lines to process.  This may be the
                            actual number of lines in the image or it may be
                            a subset. */
    int number_detectors;
    int number_scas;
    IAS_SENSOR_ID sensor_id;
    IAS_BAND_CLASSIFICATION band_classification;
    IAS_SPECTRAL_TYPE spectral_type;
} RPS_BAND_ATTRIBUTES;

typedef struct blob
{
    /* Common fields */
    char parameter_file_name[PATH_MAX];
    char omf_file_name[PATH_MAX];
    char l0r_file_name[PATH_MAX];
    char l1r_file_name[PATH_MAX];
    char bpf_filename_oli[PATH_MAX];
    char bpf_filename_tirs[PATH_MAX];
    char cpf_filename[PATH_MAX];
    char rlut_filename[PATH_MAX];
    char landsat_scene_id[IAS_L0R_SCENE_ID_LENGTH + 1];
    char work_order_id[IAS_WORKORDER_ID_SIZE];
    char pixel_mask_filename[PATH_MAX];
    char dump_directory[PATH_MAX];
    char report_directory[PATH_MAX];
    char tirs_stray_light_correction_filename[PATH_MAX];

    /* Image specific details */
    double oli_start_time;
    double oli_stop_time;
    double tirs_start_time;
    double tirs_stop_time;
    int skip_time_limit_check;

    /* Calibration Parameter File */
    IAS_CPF *cpf;

    /* Bias Parameter File */
    IAS_BPF *bpf_oli;
    IAS_BPF *bpf_tirs;

    /* Radiometric Look up table */
    IAS_RLUT_IO *rlut;

    /* Algorithm list specific parameters */
    int number_of_wo_parm_strings_allocated; /* Number of strings allocated
                                                in the wo parameter list */
    char **algorithm_wo_parameter_list;
    char **algorithm_string_list; /* This list contains pointers from the wo
                                     parameter list but it excludes any that
                                     have the SKIP modifier -- so it is the
                                     list of algorithms actually executed.
                                     Its size is number_algorithms -
                                     number_dump_algorithms. */
    int  number_algorithms;       /* Total number of algorithms */
    int number_dump_algorithms;   /* Number of dump algorithms */
    int (**algorithm_pointer_list)();  /* Contains a sequential list of
                                          pointers to all algorithms to run
                                          including the dump algorithms */
    char **dump_algorithm_string_list; /* These are the algorithms that have
                                          associated DUMP modifiers */
    int *dump_algorithm_mapping;  /* Contains the algorithm indices for the
                                     associated that have DUMP modifiers and
                                     is used to map the algorithm index to
                                     the proper dump algorithm index */
    int number_pre_proc;              /* Number of pre-process algorithms */
    int (**pre_proc_pointer_list)();  /*   and the list of pointers       */
    int number_post_proc;             /* Number of post-process algorithms */ 
    int (**post_proc_pointer_list)(); /*   and the list of pointers        */

    /* Thread specific fields */
    struct ias_threadpool *threadpool; /* Threadpool being used */
    int thread_error;       /* Thread number that encountered an error */
    int procs_to_use;       /* Number of threads to use default is 0 meaning
                               run single threaded */

    /* Band fields */
    int  total_bands;
    int  max_number_scas;
    /* This flag is set in initialize_io_routines and should be checked
       prior to using any of the band lists or information prior to
       processing the collect. */
    int  band_fields_initialized;
    int  number_of_bands;   /* This is the count of the bands in the list
                               of bands and indices */
    int  *list_of_bands;    /* This is a list of the normal bands to process */
    int  *list_of_band_indices; /* and their indices */
    int  num_normal_bands;
    int  process_oli_bands; 
    int  process_tirs_bands;
    /* This band information is set up during initialization of the I/O.
       It contains attributes such as band type, associated bands as well as
       some things specific to the L0R imagery such as number lines.  It is
       initialized with information about *all* bands that are present in
       the L0R. */
    RPS_BAND_ATTRIBUTES **band_information;

    /* L1R metadata */
    IAS_L1R_FILE_METADATA l1r_file_metadata;
    IAS_L1R_BAND_METADATA **l1r_band_metadata;

    /* L0R information - The data truncation setting comes from the L0R
       image header.  If this is set (1) then the image data was truncated
       to 12 bit on the spacecraft and needs to be shifted back to 14 bits */
    uint8_t image_data_truncation_setting;

    /* Pixel mask - This is a 2-d array of pixel mask structures it is
       accessed as [band_index][sca_index] */
    IAS_PIXEL_MASK ***pixel_mask;

    /* Pixel mask mutex and file handle -- the pixel mask is written to disk
       in a threaded routine and that access needs to be atomic.  The file
       is opened and the mutex created prior to firing off the threads and
       passed to the threads through the blob. */
    IAS_THREAD_MUTEX_TYPE pixel_mask_io_mutex;
    IAS_PIXEL_MASK_IO *pixel_mask_file_handle;

    /* Flag to indicated if ghost image will be used during processing */
    int use_tirs_ghost_image_flag;

    /* TIRS stray light band io and l1r io */
    L1RIO *tirs_stray_light_l1r_io;
    L1R_BAND_IO *tirs_stray_light_band_io[IAS_MAX_NBANDS]; 
    char stray_light_correction_source[IAS_COLLECT_TYPE_SIZE];
    int use_three_tirs_scenes_flag; /* Flag indicating use the full interval */

    /* Database and reporting structures */
    RPS_DATABASE_INFORMATION database_information;
    IAS_REPORT report_information;
    IAS_LINKED_LIST_NODE *scene_statistics_list;
    int connected_to_database;

    /* Mean OLI nad TIRS focal plane temperatures */
    int oli_mean_focal_plane_temperature_valid;
    int tirs_mean_focal_plane_temperature_valid;
    float oli_mean_focal_plane_temperature;
    float tirs_mean_focal_plane_temperature;

    /* --------------------------------------------------------------- */
    /* -------------------- Work Order parameters -------------------- */
    /* Common parms */
    int generate_report;
    int store_in_database;
    int populate_scene_summary_stats_flag;

    /* Linearization correction parm */
    int apply_linearization_correction;
    /* Gain application parms */
    int apply_absolute_gain;
    int apply_relative_gain;
    int perform_temp_sens;
    /* Image acquired date timecode for the center line */
    char date_acquired[IAS_L0R_TIME_LENGTH + 1];

    /* TIRS dark response determination*/
    int use_tirs_cpf_baseline_response;
    int write_tirs_dark_response_to_file;
    int process_tirs_blind_band;

    /* Detector Response Characterization for lamp */
    float lamp_warmup_time_period;
    int post_warmup_nframes;
    int stim_lamp_pair_id;
    float lamp_on_time;
    float post_warmup_start_time;
    float post_warmup_stop_time;
    float lamp_warmup_time;
    int lamp_warmup_start_frame;
    int lamp_warmup_stop_frame;
    int post_warmup_start_frame;
    int post_warmup_stop_frame;

    /* Histogram statistics characterization */
    int lamp_processing;
    int tirs_dark_response_processing;

    /* OLI Bias Determination */
    int bias_source;
    int per_frame_bias_flag;

    /* TIRS Bias Removal */
    int background_response_selection;

    /* Replace Saturation Pixels */
    int digital_high_saturation_flag;
    int digital_low_saturation_flag;
    int analog_high_saturation_flag;
    int analog_low_saturation_flag;
} BLOB;

#endif
