#ifndef RPS_LOCAL_H
#define RPS_LOCAL_H

#include "ias_const.h"
#include "ias_threadsync.h"
#include "ias_linked_list.h"

/* #defined constants for use by cleanup_memory_management_buffers() to
   enable/disable display of memory leak warning message.  If
   cleanup_memory_management_buffers() is called following detection of
   an error, the warning messsage is suppressed.  If
   cleanup_memory_management_buffers() is called when processing completes
   successfully, the warning message is displayed. */
#define DISABLE_WARNING_DISPLAY  0
#define ENABLE_WARNING_DISPLAY   1


/* ------------------------------------------------------------------------- */
/* --------------------------- General Structures -------------------------- */
/* These macros allow the creation of an enumerated type definition from
   the following included file.  The include contains the list of memory
   management buffers. */
#define MM_TYPE_BEGIN(name)  typedef enum
#define MM_TYPE_ZERO(name)   name = 0
#define MM_TYPE_NORM(name)   name
#define MM_TYPE_END(name)    name
#include "memory_management.h"

typedef struct alg_type
{
    const char *algorithm_name;
    int (*api_algorithm)();
    int (*pre_process_algorithm)();
    int (*post_process_algorithm)();
    int is_required;
    int must_run_after_prerequisite;
} ALG_TYPE;

typedef struct rps_database_information
{
    struct ias_db_connection *db;
    char rad_common_id[IAS_CHAR_ID_SIZE];  
    struct ias_threadpool *threadpool;
    /* Additional data needed for histogram stats */
    char *rps_level;
    char *rlut_version;
} RPS_DATABASE_INFORMATION;

/* Scene summary stats structure used by RPS */
typedef struct db_scene_summary_statistics
{
    int band_number;                      /* Band number of statistics */
    int sca_number;                       /* SCA number of statistics */
    int lineset;                          /* Lineset of statistics */
    float frames;                         /* Calculated frames */
    float mean;                           /* Calculated mean */
    float stdev;                          /* Calculated stdev */
    IAS_LINKED_LIST_NODE node;            /* Node for statistics list */
} DB_SCENE_SUMMARY_STATISTICS;

/* The BLOB needs the above structures defined which is why it's included
   here */
#include "rps_blob.h"

/* An array of IAS_RLUT_LINEARIZATION_PARAMS data structures is required
   as an input parameter to the response linearization algorithm, which is
   defined in ias_rlut.h.  So include it here */
#include "ias_rlut.h"

#include "rps_algorithm.h"


/* ------------------------------------------------------------------------- */
/* ------------------ Work Management Structures and enums ----------------- */

/* Define an algorithm synchronization type.  Currently, there are only the
   "none" and "single thread" types, but it would be possible to define other
   types such as one to wait until all the previous algorithms are complete
   then allow threading to resume (i.e. a barrier between algorithms); or
   one to wait until all the previous algorithms are complete, allow multiple
   threads to operate on the algorithm, but wait at the end of it until all
   of them are done. */
typedef enum sync_type
{
    SYNC_NONE,          /* No synchronization needed for the algorithm */
    SYNC_NEED_ADJACENT_SCAS, /* The algorithm needs two adjacent SCAs to be
                                processed through all the previous algorithms
                                before it can be run.  For example, to run the
                                algorithm on SCA 1, SCAs 1 & 2 need to be run
                                through all earlier algorithms.  Also, the last
                                SCA needs to be run last after all the lower
                                numbered SCAs have completed the algorithm. */
    SYNC_SINGLE_THREAD  /* The algorithm needs to be run on a single thread
                           with all the previous algorithms completed and no
                           algorithms after it running until this one
                           completes */
} ALGORITHM_SYNC_TYPE;

typedef struct work_state_struct
{
    int  algorithm_index;       /* 0-based algorithm index */
    int  band;                  /* 1-based band number */
    int  nscas;                 /* Number of SCAs for this band */
    char state[IAS_MAX_NSCAS];  /* State of work for the SCAs */
    ALGORITHM_SYNC_TYPE sync_type; /* The type of synchronization needed by
                                      this algorithm */
} WORK_STATE_STRUCT;

typedef enum working_state
{
    NOT_ASSIGNED,
    WORKING,
    COMPLETE
} WORKING_STATE;


/* ----------------------------------------------------------------------- */
/* ----------------- Memory Management Structure and enum ---------------- */

/* Note this structure isn't available globally since it's declared static
   to the memory management file */
typedef struct memory_buffer
{
    int user_count;    /* Count of users that still need access to buffer  */
    int is_const_data; /* Flag to indicate the type of data pointer to use */
    void *data_pointer;            /* Regular data if (is_const_data == 0) */
    const void *const_data_pointer;/*  Const  data if (is_const_data == 1) */
    size_t size;       /* Total size in bytes of the buffer */
    int number_values; /* Total number of values in in the buffer */
    int buffer_allocated; /* Flag to indicate the buffer has been allocated at
                             some point for this buffer.  This is used to
                             detect errors in buffer management. */
    int consumer_set;  /* Flag to indicate if the buffer consumer count has
                          been set.  This is used to ensure that no
                          decrement occurs before all consumers have been
                          registered. */
    int buffer_filled; /* Flag to indicate if the buffer has been filled.
                          If the flag is 1 it has been filled and is safe to
                          use.  If the flag is 0 it is either uninitialized
                          or the initializing routine didn't set the flag.  */
} MEMORY_BUFFER;

typedef enum memory_user_type
{
    CONSUMER,
    SETTER
} MEMORY_USER_TYPE;


/* ---------------------------------------------------------------------- */
/* ------------------ Driver General Routine Prototypes ----------------- */
int read_parameters
(
    BLOB *blob   /* I: parameter and OMF file names
                    O: parameters listed in input files */
);

int create_algorithm_pointer_list
(
    BLOB *blob     /* I: List of algorithms as strings
                      I: Number of algorithms
                      O: List of algorithms as pointers to API
                      O: Updated number of algorithms
                      O: List of pre-processing algorithms
                      O: List of post-processing algorithms
                      O: List of algorithms with dump modifier
                      O: Number of algorithms with dump modifier */
);

const ALG_TYPE *get_available_algorithms
(
    int *number_algorithms  /* O: Number of externally available algorithms */
);

int process_collection
(
    BLOB *blob          /* I: Various input parameters:
                              pre-process pointer list
                              number of SCAs
                              number of bands
                              band information
                              SCAs to process
                              list of bands to process
                              post-process pointer list */
);

int process_scas
(
    void *input_parms,  /* I: Collection of input parameters, including
                              algorithm list and pointer to pixel mask */
    int thread_num      /* I: running thread number */
);


/* ----- Work assignment Prototypes ----- */
int assign_work
(
    int *band_index,            /* O: Assigned band */
    int *sca_index,             /* O: Assigned SCA */
    int *algorithm_index        /* O: Assigned algorithm */
);

void cleanup_work_assignment( void );

int initialize_work_assignment
(
    int  total_algorithms,         /* I: Total number of algorithms */
    int  nbands,                   /* I: Number of bands to process */
    int  list_of_band_indices[],   /* I: List of band indices to process */
    struct ias_threadpool *pool    /* I: Threadpool */
);

void set_work_assignment_error_condition( void );

int set_algorithm_sync_type
(
    int algorithm_index,           /* I: Index of algorithm */
    ALGORITHM_SYNC_TYPE sync_type  /* I: Sync type to use */
);

/* ----- Image I/O Prototypes ----- */
uint16_t *read_l0r_image_data
(
    int band_index,      /* I: Band to retrieve */
    int sca_index        /* I: SCA to retrieve */
);

int write_l1r_output_image_data
(
    int band_index,     /* I: Band to write */
    int sca_index,      /* I: SCA to write */
    float *l1r_data     /* I: Pointer to the image data */
);

int write_l1r_image_data
(
    L1R_BAND_IO *band_struct, /* I: Pointer to image band structure */
    int band_index,           /* I: Band to write */
    int sca_index,            /* I: SCA to write */
    float *l1r_data           /* I: Pointer to the image data */
);

int initialize_io_routines
(
    BLOB *blob          /* I: Contains L0R and L1R file names */
                        /* O: Metadata obtained from L0R file */
);

int read_oli_tirs_bpf
(  
    BLOB *blob          /* I: Contains OLI and TIRS band present info and
                              OLI and TIRS BPF filenames */
);

int open_l1r_bands
(
    BLOB *blob  /* I: list of bands, indices, number of bands, band
                      information */
);

int cleanup_io_routines
(
    BLOB *blob    /* I/O: Contains file metadata structure */
);

int set_l1r_metadata
(
    const BLOB *blob,       /* I: Contains file metadata structure */
    L1RIO *out_image        /* I: The image in which to set metadata */
);

void *read_l0r_frame_headers
(
    MEMORY_BUFFER_TYPE buffer_type,           /* I: Memory buffer type */
    int *num_headers,                         /* O: Actual number of L0R
                                                 frame headers */
    size_t *buffer_size                       /* O: Actual buffer size */
);

int read_oli_focal_plane_temperatures
(
    float **temp1,             /* O: temperature 1 */
    float **temp2,             /* O: Temperature 2 */
    int *number_temps          /* O: Number of temperatures */
);

int read_tirs_focal_plane_temperatures
(
    float **f2_temperature,    /* O: fp_f2_fine_sensor_1_celsius temperature */
    float **f4_temperature,    /* O: fp_f4_fine_sensor_3_celsius temperature */
    float **f6_temperature,    /* O: fp_f6_fine_sensor_1_celsius temperature */
    float **f7_temperature,    /* O: fp_f7_fine_sensor_2_celsius temperature */
    int *number_temps          /* O: Number of temperatures */
);

int read_post_warmup_focal_plane_temperatures
(
    float **temperature7,      /* O: temperature 7 */
    float **temperature14,     /* O: Temperature 14 */
    int *number_temps          /* O: Number of temperatures */
);

int read_lamp_response
(
    BLOB *blob                      /* I: Get post warmup lines */
                                    /* O: Set lamp response data */
);

int read_post_warmup_lamp_telemetry
(
    float **lamp_current_a,         /* O: Group A lamp current values */
    float **lamp_voltage_a,         /* O: Group A lamp voltage values */
    float **photodiode_micro_amp_a, /* O: Group A lamp photodiode current */
    float **lamp_housing_temp_a,    /* O: Group A lamp housing temperatures */
    float **monitor_diode_temp_a,   /* O: Group A lamp monitor diode 
                                          temperatures */
    float **lamp_current_b,         /* O: Group B lamp current values */
    float **lamp_voltage_b,         /* O: Group B lamp voltage values */
    float **photodiode_micro_amp_b, /* O: Group B lamp photodiode current */
    float **lamp_housing_temp_b,    /* O: Group B lamp housing temperatures */
    float **monitor_diode_temp_b,   /* O: Group B lamp monitor diode 
                                          temperatures */
    int *number_lamp_telemetry      /* O: Number of lamp telemetry data */
);

int read_stim_lamp_pair_id
(
    BLOB *blob                       /* O: Set stim lamp pair id */
);

IAS_THREAD_MUTEX_TYPE *get_hdf_io_mutex( void );

/* ----- Memory Management Prototypes ----- */
int initialize_memory_management_buffers( void );

void cleanup_memory_management_buffers(int display_warning);

int increment_consumer_count
(
    MEMORY_BUFFER_TYPE buffer_type  /* I: Index into number consumers array */
);

int decrement_consumer_count
(
    MEMORY_BUFFER_TYPE buffer_type, /* I: Index into static buffer arrays */
    int band_index,                 /* I: */
    int sca_index                   /* I: */
);

void *get_memory_buffer
(
    MEMORY_BUFFER_TYPE buffer_type, /* I: Index into static buffer arrays */
    MEMORY_USER_TYPE consumer_flag, /* I: The type of user:
                                          CONSUMER - the caller uses the
                                            data so the buffer must be valid
                                          SETTER - the caller will fill the
                                            buffer and set the buffer filled
                                            flag */
    int band_index,                 /* I: */
    int sca_index                   /* I: */
);

void *preprocess_get_memory_buffer
(
    MEMORY_BUFFER_TYPE buffer_type, /* I: Index into static buffer arrays */
    int band_index,                 /* I: */
    int sca_index                   /* I: */
);

const void *get_const_memory_buffer
(
    MEMORY_BUFFER_TYPE buffer_type, /* I: Index into static buffer arrays */
    int band_index,                 /* I: */
    int sca_index                   /* I: */
);

int get_memory_buffer_number_values
(
    MEMORY_BUFFER_TYPE buffer_type, /* I: Index into static buffer arrays */
    int band_index,                 /* I: */
    int sca_index                   /* I: */
);

const char *get_memory_buffer_string
(
    MEMORY_BUFFER_TYPE buffer_type  /* I: Index into static string array */
);

int set_consumer_counts(void);

void set_buffer_filled_flag
(
    MEMORY_BUFFER_TYPE buffer_type, /* I: Index into static buffer arrays */
    int band_index,                 /* I: */
    int sca_index                   /* I: */
);

int get_buffer_filled_flag
(
    MEMORY_BUFFER_TYPE buffer_type, /* I: Index into static buffer arrays */
    int band_index,                 /* I: */
    int sca_index                   /* I: */
);


/* ----- BLOB interface routines ----- */
BLOB *allocate_blob( void );
void free_the_blob( void );
int finish_blob_initialization( void );
int get_blob_max_nscas( void );
struct ias_db_connection *get_blob_db_connection( void );
const char *get_blob_rad_common_id( void );
char *get_blob_work_order_id( void );
char *get_blob_rlut_filename( void );
int get_blob_post_warmup_nframes( void );
double get_blob_lamp_warmup_time_period( void );
IAS_CPF *get_blob_cpf( void );
IAS_BPF *get_blob_bpf_oli( void );
IAS_BPF *get_blob_bpf_tirs( void );
IAS_RLUT_IO *get_blob_rlut(void);
double get_blob_oli_start_time( void );
double get_blob_oli_stop_time( void );
double get_blob_tirs_start_time( void );
double get_blob_tirs_stop_time( void );
int create_blob_pixel_mask( void );
int get_blob_number_of_bands( void );
int *get_blob_list_of_bands_to_process( void );
int *get_blob_list_of_band_indices_to_process( void );
int get_blob_process_blind_band_flag( void );
IAS_RLUT_IO *get_blob_rlut( void );
int get_band_fields_initialized( void );
int get_algorithm_index(void *);
int get_background_response_selection(void);
void set_oli_focal_plane_temperature( float temperature );
void set_tirs_focal_plane_temperature( float temperature );
int get_oli_focal_plane_temperature( float *temperature );
int get_tirs_focal_plane_temperature( float *temperature );
IAS_LINKED_LIST_NODE *get_scene_summary_list(void);

RPS_BAND_ATTRIBUTES *get_band_attributes
(
    int band_index      /* I: */
);


/* ----- database interface routines ----- */
int establish_db_connection
(
    BLOB *blob                  /* O: Database connection information */
);

int close_db_connection
(
    BLOB *blob,         /* I: Database connection information */
    int  commit         /* I: Flag to indicate committing the transaction */
);

int delete_db_table_records
(
    RPS_DATABASE_INFORMATION db_info, /* I: Database connection and
                                            rad_common_id */
    const char *table_name            /* I: Table to delete data from */
);


/* ----- miscellany ----- */
char *get_rps_level
(
    int band_index,     /* I: */
    int sca_index       /* I: */
);

int set_rps_level
(
    int band_index,     /* I: */
    int sca_index,      /* I: */
    char process_id     /* I: */
);

int initialize_rps_level
(
    int number_of_bands,        /* I: */
    int number_of_scas          /* I: */
);

void cleanup_rps_level( void );

int rps_level_is_l0r
(
    const char *rps_level  /* I: */
);

int update_omf
(
    BLOB *blob          /* I: contains the OMF and L1R file names */
);

int get_earth_sun_distance
(
    char *date_acquired,        /* I: contains the date_required */
    double *earth_sun_distance  /* O: earth-sun distance */
);

int calc_normalized_correction_factor
(
    const double *sca_ratio,   /* I: Array of SCA_SCA ratios with NSCAS 
                                     numbers*/
    int nscas,                 /* I: Number of SCAs per band */
    double *correction_factor  /* O: Array of normalized correction factors 
                                     with NSCAS numbers */
);

/* Prototypes for test routines.  These are only used by the test suite. */
void get_mem_mgmt_struct_dont_use_me( MEMORY_BUFFER_TYPE,int,int,int
    *,MEMORY_BUFFER * );

#endif
