/*----------------------------------------------------------------------------
NAME:    rps_main

PURPOSE: Execute the radiometric characterization, correction and other
         alogorithms specified by the input ODL file.

RETURNS: EXIT_SUCCESS or EXIT_FAILURE
-----------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libgen.h>

#include "ias_logging.h"
#include "rps_local.h"
#include "ias_l0r.h"

#define MAXIMUM_L0R_SECONDS_PROCESSED 270.0

int main( int argc, char *argv[] )
{
    int  status;
    int error_status = 0;
    BLOB *blob;
    char *app_name;
    int display_warning = ENABLE_WARNING_DISPLAY;
    int band_index;
    int index;

    /* Strip off the path to the application so that the logging library
       doesn't log it and initialize the library */
    app_name = strdup( basename( argv[0] ));
    if ( !app_name )
    {
        IAS_LOG_ERROR("Allocating memory for application name");
        exit(EXIT_FAILURE);
    }
    ias_log_initialize( app_name );
    free( app_name );

    /* Everything in the RPS needs the blob so allocate it right up front.
       Note that the blob isn't complete at this point.  We need more
       information -- the satellite number for instance -- from the input
       parameters before doing some more initialization. */
    blob = allocate_blob();
    if ( !blob )
    {
        IAS_LOG_ERROR("Allocating the BLOB");
        exit(EXIT_FAILURE);
    }

    /* There must be one command line agrument: ODL parameter file */
    if ( argc != 2 )
    {
        IAS_LOG_ERROR("Missing input command line argument" );
        free_the_blob();
        exit(EXIT_FAILURE);
    }
    /* Store the file name in the blob */
    strcpy( blob->parameter_file_name, argv[1] );

    /* Get input parameters from input files */
    status = read_parameters( blob );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Reading input parameters");
        free_the_blob();
        exit(EXIT_FAILURE);
    }

    IAS_LOG_INFO("Initiated create_l1r");

    /* Create the algorithm lists */
    status = create_algorithm_pointer_list( blob );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Creating algorithm pointer list");
        free_the_blob();
        exit(EXIT_FAILURE);
    }

    /* Database connect */
    if ( blob->store_in_database )
    {
        status = establish_db_connection( blob );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Establishing database connection");
            free_the_blob();
            exit(EXIT_FAILURE);
        }
        blob->connected_to_database = 1;

        /* Get or create a radiometric common ID */
        status = ias_misc_get_char_id( blob->database_information.db,
                "radiometric_work_order_common", "radiometric_wo_common_id",
                blob->work_order_id, blob->database_information.rad_common_id);
        if ( status == ERROR )
        {
            close_db_connection( blob, IAS_DB_ROLLBACK );
            free_the_blob();
            exit(EXIT_FAILURE);
        }
    }
    else
    {
        blob->connected_to_database = 0;
    }

    /* Initialize band I/O and memory management */
    status = initialize_io_routines( blob );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Initializing the band I/O routines");
        if ( blob->connected_to_database )
            close_db_connection( blob, IAS_DB_ROLLBACK );
        free_the_blob();
        exit(EXIT_FAILURE);
    }

    /* Check to see if any OLI bands appear in the bands to process */
    blob->process_oli_bands = 0;
    for ( index = 0; index < blob->number_of_bands; index++ )
    {
        band_index = blob->list_of_band_indices[index];
        if ( blob->band_information[band_index]->sensor_id == IAS_OLI &&
             blob->band_information[band_index]->process_band )
        {
            blob->process_oli_bands = 1;
            break;
        }
    }

    /* Check to see if any TIRS bands appear in the bands to process */
    blob->process_tirs_bands = 0;
    for ( index = 0; index < blob->number_of_bands; index++ )
    {
        band_index = blob->list_of_band_indices[index];
        if ( blob->band_information[band_index]->sensor_id == IAS_TIRS &&
             blob->band_information[band_index]->process_band )
        {
            blob->process_tirs_bands = 1;
            break;
        }
    }

    /* Read OLI and/or TIRS BPF as needed */
    status = read_oli_tirs_bpf( blob );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Reading OLI and/or TIRS BPF");
        if ( blob->connected_to_database )
            close_db_connection( blob, IAS_DB_ROLLBACK );
        free_the_blob();
        exit(EXIT_FAILURE);
    }

    /* Check L0R data duration, if it's longer than 4.5 minutes, exit with
       an error since any longer than that will currently exceed the memory
       available on the server (unless we have been instructed to bypass this
       check). */
    if (blob->process_oli_bands)
    {
        if ((blob->oli_stop_time - blob->oli_start_time > 
         MAXIMUM_L0R_SECONDS_PROCESSED) &&
         !blob->skip_time_limit_check)
        {  
            IAS_LOG_ERROR("The L0R imagery is %f seconds which is longer than "
                      "the %f limit allowed",  blob->oli_stop_time - 
                      blob->oli_start_time, MAXIMUM_L0R_SECONDS_PROCESSED);
            if ( blob->connected_to_database )
                close_db_connection( blob, IAS_DB_ROLLBACK );
            free_the_blob();
            exit(EXIT_FAILURE);
        }
    }

    if (blob->process_tirs_bands)    
    {
        if ((blob->tirs_stop_time - blob->tirs_start_time > 
         MAXIMUM_L0R_SECONDS_PROCESSED) &&
         !blob->skip_time_limit_check)
        {  
            IAS_LOG_ERROR("The L0R imagery is %f seconds which is longer than "
                      "the %f limit allowed",  blob->tirs_stop_time - 
                      blob->tirs_start_time, MAXIMUM_L0R_SECONDS_PROCESSED);
            if ( blob->connected_to_database )
                close_db_connection( blob, IAS_DB_ROLLBACK );
            free_the_blob();
            exit(EXIT_FAILURE);
        }
    }

    status = initialize_memory_management_buffers();
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Initializing the memory management routines");
        if ( blob->connected_to_database )
            close_db_connection( blob, IAS_DB_ROLLBACK );
        cleanup_io_routines( blob );
        free_the_blob();
        exit(EXIT_FAILURE);
    }

    status = create_blob_pixel_mask();
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Creating the pixel mask");
        if ( blob->connected_to_database )
            close_db_connection( blob, IAS_DB_ROLLBACK );
        cleanup_io_routines( blob );
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        exit(EXIT_FAILURE);
    }

    /* Process the collection */
    status = process_collection( blob );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Processing collection");
        if ( blob->connected_to_database )
            close_db_connection( blob, IAS_DB_ROLLBACK );
        /* Continue on to clean up ... */
        error_status = 1;
        display_warning = DISABLE_WARNING_DISPLAY;
    }
    else
    {
        /* Close the database if connected */
        if ( blob->connected_to_database )
        {
            status = close_db_connection( blob, IAS_DB_COMMIT );
            if ( status != SUCCESS )
            {
                IAS_LOG_ERROR("Closing database connection");
                error_status = 1;
                display_warning = DISABLE_WARNING_DISPLAY;
            }
        }
    }

    /* Write the L1R file name to the OMF */
    status = update_omf( blob );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Updating OMF");
        error_status = 1;
        display_warning = DISABLE_WARNING_DISPLAY;
    }

    /* Free up memory */
    cleanup_memory_management_buffers(display_warning);
    free_the_blob();

    /* Return the proper exit status and appropriate message */
    if ( error_status )
    {
        IAS_LOG_ERROR("did not complete successfully");
        exit(EXIT_FAILURE);
    }
    else
    {
        IAS_LOG_INFO("Successful completion");
        return EXIT_SUCCESS;
    }
}  /* END main */
