#!/bin/csh -f
# PURPOSE: This test executes create_l1r with 4 threads with debugging turned
#          on.  With debugging on the synchronize_query call outputs a txt
#          file assign_work_debug.txt with internal work assignment state.
#          After execution the resulting file is searched for the work
#          states for one OLI band and two TIRS bands.  The condition counts
#          are examined to ensure the algorithm actually executed for the
#          bands.  This test was specifically written to test that band 11
#          gets run through the serialized algorithm (tracker 3787).
#
# RETURNS: 0 if no errors occurred
#          count of the failures on error
#---------------------------------------------------------------------------

# Initialize the counts of passes and failures
set passes = 0
set failures = 0

# Ensure the unit test directory is set
if ( ${?IAS_UNIT_TEST_DATA_DIR} == 0 ) then
    echo "IAS_UNIT_TEST_DATA_DIR isn't set"
    @ failures++
    echo "Completed ${0}: $passes passes,  $failures failure"
    exit 1
endif
@ passes++

echo "Running $0"

# Since the OMF gets updated make a copy of it from the template and expand
# the IAS_UNIT_TEST_DATA_DIR
sed "s|IAS_UNIT_TEST_DATA_DIR|$IAS_UNIT_TEST_DATA_DIR|" \
    assign_work_template.omf > AW1.omf
if ( $status ) then
    echo "Making working copy of OMF"
    @ failures++
    echo "Completed ${0}: $passes passes,  $failures failure"
    exit 1
endif
@ passes++

# Execute the driver
setenv IAS_LOG_LEVEL DEBUG
../create_l1r assign_work.odl >& assign_work_out.txt
if ( $status ) then
    echo "Error executing: ../create_l1r assign_work.odl"
    @ failures++
    echo "Completed ${0}: $passes passes,  $failures failure"
    exit 1
endif
@ passes++
# Partial clean up
rm -f pixel_mask.dat assign_work_l1r_image.h5 AW1.omf

# The algorithm we're interested in is index 5 and there are 4 bands in the
# band list: BAND_LIST = (1, 9, 10, 11)
# Check these three bands
set band_index = (8 9 10)
set num_scas = (14 3 3)

set i = 1
foreach bi ( $band_index )
    set ns = $num_scas[$i]
    set si = 0
    while ( $si <  $ns )
        # grab the three condition counts of interest
        set cond = `sed -n "/--* $bi $si /,/+++/p" assign_work_debug.txt | \
            grep Condition | awk '{print $4,$5,$6}'`
        # Get the one for this band index
        @ one_cond = $cond[$i]

        # Grab the line for the serialized algorithm.  For this test case
        # that is 'algo:  5'
        set a = `sed -n "/--* $bi $si /,/+++/p" assign_work_debug.txt | \
             grep "algo:  5" | grep "band: *$bi" | awk -F: '{print $6}'`

        # Count the number of entries that are W or C
        set w_entry_count = 0
        set c_entry_count = 0
        set loop_index = 0
        while ( $loop_index < $ns )
            @ loop_index++
            if ( $a[$loop_index] == 'W' ) then
                @ w_entry_count++
            endif
            if ( $a[$loop_index] == 'C' ) then
                @ c_entry_count++
            endif
        end

        # The condition count should be the count of C entries plus 1 -- at
        # least that is what I've observed.  If that isn't always true this
        # should be looked at again.
        @ entry_count = $c_entry_count + 1

        # Compare the entry count with the condition count -- they should match
        if ( $one_cond != $entry_count ) then
            echo "Band index $bi SCA index $si - [$one_cond] should " \
                "match [$entry_count]"
            @ failures++
        else
            @ passes++
        endif
        @ si++
    end
    @ i++
end

echo "Completed ${0}: $passes passes, $failures failures"
# If any were in error exit without cleaning up
if ( $failures ) then
    exit $failures
endif

# Finish Clean up
rm -f assign_work_out.txt assign_work_debug.txt
exit 0
