#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "ias_pixel_mask.h"
#include "ias_satellite_attributes.h"
#include "cgreen/cgreen.h"
#include "../rps_local.h"
#include "ias_logging.h"
#include "ias_const.h"
#include "ias_miscellaneous.h"
#include "ias_radiometry.h"

#define NUM_BANDS 9
#define NUM_SCAS 14
#define NUM_DETECTORS 494
#define NUM_LINES 512
#define NUM_PAN_DETECTORS 988
#define NUM_PAN_LINES 1024


/* Path to test data and file names */
static char data_dir[] = "ias_base/rps/driver/lamp";
static char data_file[] = "Level0_test6.DAT";


Ensure char_lamp_detector_response_nasa_data_test(void);

int main(int argc, char *argv[])
{
    /* Initialize the log library */
    ias_log_initialize("lamp_nasa_data_test");

    TestSuite *suite = create_test_suite();
    add_test(suite, char_lamp_detector_response_nasa_data_test);

    return run_test_suite(suite, create_text_reporter());
}


Ensure char_lamp_detector_response_nasa_data_test(void)
{
    IAS_PIXEL_MASK *pm;
    IAS_REPORT report_info;
    IAS_REPORT_HEADER *report_header = NULL;
    char work_order_id[IAS_WORKORDER_ID_SIZE];
    char routine_name[] = "Lamp Char tests";
    int status;

    /* Read in NASA lamp data file */
    int nfov = NUM_DETECTORS;
    int nframe_ms = NUM_LINES;
    FILE *l0_f = NULL;
    char *test_directory;       /* test environment value */
    char l0r_file[PATH_MAX];
    short int *buffer;
    short int *pan_buffer;

    double sca_mean[NUM_SCAS][NUM_BANDS];
    double sca_stdev[NUM_SCAS][NUM_BANDS];

    IAS_RPS_HISTOGRAM_INTERMEDIATE_RESULTS *statistics_results_l0r = NULL;
    IAS_RPS_HISTOGRAM_INTERMEDIATE_RESULTS *statistics_results_pan_l0r = NULL;
    DB_CHAR_LAMP_RESPONSE_DETECTOR_RESULTS *db_detector_results = NULL;
    DB_CHAR_LAMP_RESPONSE_DETECTOR_RESULTS *db_detector_pan_results = NULL;
    DB_CHAR_LAMP_RESPONSE_ONE_SCA_RESULTS db_one_sca_l0r_results;
    DB_CHAR_LAMP_RESPONSE_SCA_RESULTS *db_l0r_sca_results = NULL;
    DB_CHAR_LAMP_RESPONSE_BAND_RESULTS *db_l0r_band_results = NULL;
    int number_of_scas = NUM_SCAS;
    int number_of_bands = NUM_BANDS;
    int band_num;
    int band_index;
    int band_actual_num;
    int sca_num;
    int sca_index;
    int num_lines = NUM_LINES;
    int num_samps = NUM_DETECTORS;
    int generate_report = 1;
    int sca_counter;
    FILE *inf = NULL;
    float temp;
    char rps_level[] = "C";
    int lamp_pair_id = 1;
    float lamp_on_time = 5.0;
    float post_warmup_start_time = 155.0;
    float post_warmup_stop_time = 157.3;
    int post_warmup_start_frame = 12345;
    int post_warmup_stop_frame = 12857;
    int detector_offsets[494];
    int pan_detector_offsets[988];
    int det_index;

    /* Get the environment variable for the test data */
    test_directory = getenv("IAS_UNIT_TEST_DATA_DIR");
    if ( !test_directory )
    {
        assert_true_with_message(  0, "Getting IAS_UNIT_TEST_DATA_DIR");
        return;
    }

    /* Initialize detector_offsets and pan_detector_offsets */
    for (det_index = 0; det_index < NUM_DETECTORS; det_index++)
        detector_offsets[det_index] = 0;
    for (det_index = 0; det_index < NUM_PAN_DETECTORS; det_index++)
        pan_detector_offsets[det_index] = 0;

    /* Get the input L0r data file name and path */
    sprintf(l0r_file, "%s/%s/%s", test_directory, data_dir, data_file );

    /* Initialize the attributes lib */
    status = ias_sat_attr_initialize(IAS_L8);
    if (status)
    {
        assert_true_with_message( 0,
            "ias_sat_attr_initialize returned ERROR");
        return;
    }

    /* Set up a standard header */
    setenv("PCS_WOID", "I4582", 1);
    strcpy(work_order_id, "I4582");
    report_header = ias_misc_initialize_report_header("",
        work_order_id, routine_name);
    if (!report_header)
    {
        assert_true_with_message( 0, "Error initializing report header");
        return;
    }
    ias_misc_set_header_pathrow(report_header, 0, 0);
    report_info.header = report_header;
    report_info.directory = NULL;

    /* Allocate memory for the 
                   DB_CHAR_LAMP_RESPONSE_DETECTOR_RESULTS structure */
    db_detector_results = rps_db_allocate_lamp_detector_results(num_samps);
    if (!db_detector_results)
    {
        assert_true_with_message( 0, "Error allocating detector "
                                        "results database memory");
        ias_misc_free_report_header(report_header);
        return;
    }

    /* Allocate memory for the 
                   DB_CHAR_LAMP_RESPONSE_DETECTOR_RESULTS structure */
    db_detector_pan_results = rps_db_allocate_lamp_detector_results(
                              2*num_samps);
    if (!db_detector_pan_results)
    {
        assert_true_with_message( 0, "Error allocating detector "
                                        "results database memory");
        ias_misc_free_report_header(report_header);
        rps_db_free_lamp_detector_results(db_detector_results);
        return;
    }

    /* Allocate memory for the DB_CHAR_LAMP_SCA_DETECTOR_RESULTS
       structure */
    db_l0r_sca_results = rps_db_allocate_lamp_sca_results( 
                     number_of_scas);
    if (!db_l0r_sca_results)
    {
        assert_true_with_message( 0, "Error allocating sca "
                                        "results database memory");
        ias_misc_free_report_header(report_header);
        rps_db_free_lamp_detector_results(db_detector_results);
        rps_db_free_lamp_detector_results(db_detector_pan_results);
        return;
    }

    /* Allocate memory for DB_CHAR_LAMP_RESPONSE_BAND_RESULTS
       structure */
    db_l0r_band_results = rps_db_allocate_lamp_band_results( 
                      number_of_bands);
    if (!db_l0r_band_results)
    {
        assert_true_with_message( 0, "Error allocating sca "
                                        "results database memory");
        ias_misc_free_report_header(report_header);
        rps_db_free_lamp_detector_results(db_detector_results);
        rps_db_free_lamp_detector_results(db_detector_pan_results);
        rps_db_free_lamp_sca_results(db_l0r_sca_results);
        return;
    }


    /* Allocate memory for the statistics_results_l0r and 
       statistics_results_pan_l0r structures */
    statistics_results_l0r 
        = malloc(num_samps * sizeof(IAS_RPS_HISTOGRAM_INTERMEDIATE_RESULTS));
    statistics_results_pan_l0r 
        = malloc(2 * num_samps * 
                 sizeof(IAS_RPS_HISTOGRAM_INTERMEDIATE_RESULTS));
    if ( !statistics_results_l0r || !statistics_results_pan_l0r )
    {
        assert_true_with_message( 0, "Error allocating statistics "
                                        "results memory");
        ias_misc_free_report_header(report_header);
        rps_db_free_lamp_detector_results(db_detector_results);
        rps_db_free_lamp_detector_results(db_detector_pan_results);
        rps_db_free_lamp_sca_results(db_l0r_sca_results);
        rps_db_free_lamp_band_results(db_l0r_band_results);
        free(statistics_results_l0r);
        free(statistics_results_pan_l0r);
        return;
    }

    /* Read in the sca_mean and sca_stdev based on NASA idl outputs */
    inf = fopen("char_lamp_sca_mean.txt", "r");
    if (!inf)
    {
        assert_true_with_message( 0, "Error opening: "
                                        "char_lamp_sca_mean.txt");
        ias_misc_free_report_header( report_header );
        rps_db_free_lamp_detector_results(db_detector_results);
        rps_db_free_lamp_detector_results(db_detector_pan_results);
        rps_db_free_lamp_sca_results(db_l0r_sca_results);
        rps_db_free_lamp_band_results(db_l0r_band_results);
        free(statistics_results_l0r);
        free(statistics_results_pan_l0r);
        return;
    }

    for (band_index = 0; band_index < number_of_bands; band_index++)
    {
        for (sca_index = 0; sca_index < number_of_scas; sca_index++)
        {
            fscanf(inf, "%f14.8", &temp);
            sca_mean[sca_index][band_index] = temp;
        }
    }
    fclose(inf);

    inf = fopen("char_lamp_sca_stdev.txt", "r");
    if (!inf)
    {
        assert_true_with_message( 0, "Error opening: "
                                        "char_lamp_sca_stdev.txt");
        ias_misc_free_report_header( report_header );
        rps_db_free_lamp_detector_results(db_detector_results);
        rps_db_free_lamp_detector_results(db_detector_pan_results);
        rps_db_free_lamp_sca_results(db_l0r_sca_results);
        rps_db_free_lamp_band_results(db_l0r_band_results);
        free(statistics_results_l0r);
        free(statistics_results_pan_l0r);
        return;
    }

    for (band_index = 0; band_index < number_of_bands; band_index++)
    {
        for (sca_index = 0; sca_index < number_of_scas; sca_index++)
        {
            fscanf(inf, "%f14.8", &temp);
            sca_stdev[sca_index][band_index] = temp;
        }
    }
    fclose(inf);

    /* Allocate memory for buffer and pan_buffer */
    buffer = malloc(NUM_DETECTORS * NUM_LINES * sizeof(short int));
    pan_buffer = malloc(NUM_PAN_DETECTORS * NUM_PAN_LINES * sizeof(short int));
    if (!buffer || !pan_buffer)
    {
        assert_true_with_message( 0, "Error allocating buffer "
                                        "and pan_buffer memory");
        ias_misc_free_report_header( report_header );
        rps_db_free_lamp_detector_results(db_detector_results);
        rps_db_free_lamp_detector_results(db_detector_pan_results);
        rps_db_free_lamp_sca_results(db_l0r_sca_results);
        rps_db_free_lamp_band_results(db_l0r_band_results);
        free(statistics_results_l0r);
        free(statistics_results_pan_l0r);
    }

    /* Initialize the processing level in db_detector_results and
       db_detector_pan_results */ 
    db_detector_results->processing_level 
                  = malloc(sizeof(rps_level) + 1);
    strcpy(db_detector_results->processing_level, rps_level);
    db_detector_pan_results->processing_level 
                  = strdup(db_detector_results->processing_level);
    db_l0r_sca_results->processing_level 
                  = strdup(db_detector_results->processing_level);
    db_l0r_band_results->processing_level 
                  = strdup(db_detector_results->processing_level);
    if (!db_detector_pan_results->processing_level || 
        !db_l0r_sca_results->processing_level ||
        !db_l0r_band_results->processing_level)
    {
        assert_true_with_message( 0, "Error allocating buffer "
                                        "and pan_buffer memory");
        ias_misc_free_report_header( report_header );
        rps_db_free_lamp_detector_results(db_detector_results);
        rps_db_free_lamp_detector_results(db_detector_pan_results);
        rps_db_free_lamp_sca_results(db_l0r_sca_results);
        rps_db_free_lamp_band_results(db_l0r_band_results);
        free(statistics_results_l0r);
        free(statistics_results_pan_l0r);
    }

    for (band_num = 1; band_num <= number_of_bands; band_num++)
    {
        /* Initialize the db_band_results mean and stdev values */
        db_l0r_band_results->num_bands = number_of_bands;
        db_l0r_band_results->band[band_num-1] = band_num;
        db_l0r_band_results->db_mean[band_num-1] = 0;
        db_l0r_band_results->db_stdev[band_num-1] = 0;
        db_l0r_band_results->db_corrected[band_num-1] = 0;

        sca_counter = 0;

        for (sca_num = 1; sca_num <= number_of_scas; sca_num++)
        {
            /* Create the pixel mask 
               Note: In test data file from NASA, pan band is band 9 instead
                     of band 8 as defined in the landsat 8 */
            if (band_num != 9)
            {
                pm = ias_pm_create(band_num, sca_num, NUM_DETECTORS, NUM_LINES);
                assert_true(pm != NULL);
                if (pm == NULL)
                {
                    assert_true_with_message( 0,"Error from creating Pixel "
                                                   "Mask for normal band");
                    ias_misc_free_report_header( report_header );
                    rps_db_free_lamp_detector_results(db_detector_results);
                    rps_db_free_lamp_detector_results(db_detector_pan_results);
                    rps_db_free_lamp_sca_results(db_l0r_sca_results);
                    rps_db_free_lamp_band_results(db_l0r_band_results);
                    free(buffer);
                    free(pan_buffer);
                    free(statistics_results_l0r);
                    free(statistics_results_pan_l0r);
                    return;
                }                
            }
            else
            {
                /* Create the pan band pixel mask */
                pm = ias_pm_create(band_num, sca_num, NUM_PAN_DETECTORS, 
                                   NUM_PAN_LINES);
                assert_true(pm != NULL);
                if (pm == NULL)
                {
                    assert_true_with_message( 0,"Error from creating Pixel "
                                             "Mask for pan band");
                    ias_misc_free_report_header( report_header );
                    rps_db_free_lamp_detector_results(db_detector_results);
                    rps_db_free_lamp_detector_results(db_detector_pan_results);
                    rps_db_free_lamp_sca_results(db_l0r_sca_results);
                    rps_db_free_lamp_band_results(db_l0r_band_results);
                    free(buffer);
                    free(pan_buffer);
                    free(statistics_results_l0r);
                    free(statistics_results_pan_l0r);
                    return;
                }                
            }

            /* Read in NASA Lamp L0R data file, the file is a rewritten
               binary file which has skipped all the warmup lines of data */
            l0_f=fopen(l0r_file, "rb");
            if (l0_f == NULL)
            {
                assert_true_with_message( 0,"Error from open l0r_file");
                ias_misc_free_report_header( report_header );
                rps_db_free_lamp_detector_results(db_detector_results);
                rps_db_free_lamp_detector_results(db_detector_pan_results);
                rps_db_free_lamp_sca_results(db_l0r_sca_results);
                rps_db_free_lamp_band_results(db_l0r_band_results);
                free(buffer);
                free(pan_buffer);
                free(statistics_results_l0r);
                free(statistics_results_pan_l0r);
                return;
            }

            if (band_num != 9)
            {
                fseek(l0_f, sizeof(short int) * ((band_num-1) * number_of_scas 
                                   + sca_num-1) * nfov * nframe_ms, SEEK_SET);
                fread(&buffer[0], nfov*sizeof(short int), nframe_ms, l0_f);
            }
            else
            {
                fseek(l0_f, sizeof(short int) * ((band_num-1)
                      * number_of_scas) * nfov * nframe_ms, SEEK_SET);
                fseek(l0_f, sizeof(short int) * (sca_num-1) * nfov * 2
                      * nframe_ms * 2, SEEK_CUR);
                fread(&pan_buffer[0], 
                          nfov * 2 * sizeof(short int), nframe_ms * 2, l0_f);
            }
            
            fclose(l0_f);

            if (band_num != 9)
            {
                if (band_num != 8)
                    band_actual_num = band_num;
                else
                    band_actual_num = band_num + 1; 

                /* IAS_UI2 */
                status = ias_rps_histogram_statistics_characterization(
                         &buffer[0], IAS_UI2, band_actual_num, sca_num, 
                         NUM_DETECTORS, NUM_LINES, detector_offsets, 0, 
                         0, pm, NULL, NULL, statistics_results_l0r);
                if ( status )
                {
                    assert_true_with_message( 0, "Error from first "
                    "ias_rps_histogram_statistics_characterization");
                    ias_misc_free_report_header(report_header);
                    rps_db_free_lamp_detector_results(db_detector_results);
                    rps_db_free_lamp_detector_results(
                                    db_detector_pan_results);
                    rps_db_free_lamp_sca_results(db_l0r_sca_results);
                    rps_db_free_lamp_band_results(db_l0r_band_results);
                    free(buffer);
                    free(pan_buffer);
                    free(statistics_results_l0r);
                    free(statistics_results_pan_l0r);
                    return;
                }
            }
            else
            {
                band_actual_num = band_num - 1; 
                /* IAS_UI2 */
                status = ias_rps_histogram_statistics_characterization(
                    &pan_buffer[0], IAS_UI2, band_actual_num, sca_num, 
                    NUM_PAN_DETECTORS, NUM_PAN_LINES, pan_detector_offsets, 0, 
                    0, pm, NULL, NULL, statistics_results_pan_l0r);
                if ( status )
                {
                    assert_true_with_message(  0, "Error from first "
                    "ias_rps_histogram_statistics_characterization" );
                    ias_misc_free_report_header( report_header );
                    rps_db_free_lamp_detector_results(db_detector_results);
                    rps_db_free_lamp_detector_results(
                            db_detector_pan_results);
                    rps_db_free_lamp_sca_results(db_l0r_sca_results);
                    rps_db_free_lamp_band_results(db_l0r_band_results);
                    free(buffer);
                    free(pan_buffer);
                    free(statistics_results_l0r);
                    free(statistics_results_pan_l0r);
                    return;
                }
            }

            /* Set the common values for db_detector_results that are known
               here */
            db_detector_results->band = band_num;
            db_detector_results->sca = sca_num;
            db_detector_results->num_detectors = num_samps;

            db_detector_pan_results->band = band_num;
            db_detector_pan_results->sca = sca_num;
            db_detector_pan_results->num_detectors = 2 * num_samps;

            /* Set the common values for db_one_sca_l0r_results that are known
               here */
            db_one_sca_l0r_results.band = band_num;
            db_one_sca_l0r_results.sca = sca_num;

            /* Call char_lamp_detector_response algorithm code. Because the 
               test input data file from NASA make PAN band the band 9 instead
               of 8 as defined in Landsat 8, the below band swap is needed */
            if (band_num != 9)
            {
                if (band_num != 8)
                    band_actual_num = band_num;
                else
                    band_actual_num = band_num + 1; 

                status = char_lamp_detector_response(
                    band_actual_num, sca_num, num_samps, 
                    num_lines, lamp_pair_id, lamp_on_time, 
                    post_warmup_start_time, post_warmup_stop_time, 
                    post_warmup_start_frame, post_warmup_stop_frame, pm, 
                    statistics_results_l0r, &report_info, 
                    db_detector_results, &db_one_sca_l0r_results);
                assert_equal(status, SUCCESS);
            }
            else
            {
                band_actual_num = band_num - 1; 

                status = char_lamp_detector_response(band_actual_num, sca_num, 
                    2 * num_samps, 2 * num_lines, lamp_pair_id, lamp_on_time, 
                    post_warmup_start_time, post_warmup_stop_time, 
                    2 * post_warmup_start_frame, 2 * post_warmup_stop_frame,
                    pm, statistics_results_pan_l0r, 
                    &report_info, db_detector_pan_results, 
                    &db_one_sca_l0r_results);
                assert_equal(status, SUCCESS);
            }

            /* Set the common values for db_sca_results that are known here */
            db_l0r_sca_results->num_scas = number_of_scas;
            db_l0r_sca_results->band = band_num;

            /* Calculate the statistics for the db_band_results structure */
            db_l0r_band_results->db_mean[band_num-1] 
                       += db_one_sca_l0r_results.db_mean;
            db_l0r_band_results->db_stdev[band_num-1] += 
                                       db_one_sca_l0r_results.db_stdev;
            db_l0r_band_results->db_corrected[band_num-1] += 
                                       db_one_sca_l0r_results.db_corrected;

            sca_counter++;

            /* Populate db_one_sca_results into multiple SCA
               db_sca_results calculate */
            db_l0r_sca_results->sca[sca_num-1] = sca_num;
            db_l0r_sca_results->db_mean[sca_num-1] 
                   = db_one_sca_l0r_results.db_mean;
            db_l0r_sca_results->db_stdev[sca_num-1] 
                   = db_one_sca_l0r_results.db_stdev;
            db_l0r_sca_results->db_corrected[sca_num-1] 
                   = db_one_sca_l0r_results.db_mean;

            /* Destroy the pixel mask */
            ias_pm_destroy(pm);

            /* Unit test sca level results with NASA idl outputs. 
               Now the 1/sqrt(N) is used in histogram statistics mean and 
               stdev calculation and 1/sqrt(N-1) is used in idl mean and
               stdev calculation, with the large number of N (number of data
               points), the difference is ignorable as I was told */
            assert_true(fabs(db_l0r_sca_results->db_mean[sca_num-1] 
                             -sca_mean[sca_num-1][band_num-1]) < 1e-2);
            assert_true(fabs(db_l0r_sca_results->db_stdev[sca_num-1] 
                             -sqrt(sca_stdev[sca_num-1][band_num-1])) < 1e-2);
        }
        /* Write out SCA level lamp statistics data */
        if (generate_report)
        {
            /* Write L0R SCA level statistics to output report file */
            if (rps_output_lamp_sca_results(band_num, 
                    db_l0r_sca_results, &report_info,
                    LAMP_WORKING_PAIR, lamp_on_time, 
                    post_warmup_start_time, post_warmup_stop_time, 
                    post_warmup_start_frame, post_warmup_stop_frame) != SUCCESS)
            {
                assert_true_with_message(  0, "Error from first "
                    "write_lamp_response_sca_report_file" );
                ias_misc_free_report_header( report_header );
                rps_db_free_lamp_detector_results(db_detector_results);
                rps_db_free_lamp_detector_results(db_detector_pan_results);
                rps_db_free_lamp_sca_results(db_l0r_sca_results);
                rps_db_free_lamp_band_results(db_l0r_band_results);
                free(buffer);
                free(pan_buffer);
                free(statistics_results_l0r);
                free(statistics_results_pan_l0r);
                return;
            }
        }

        /* Calculate the statistics for the db_band_results structure */
        db_l0r_band_results->db_mean[band_num-1] = 
                   db_l0r_band_results->db_mean[band_num-1] / sca_counter;
        db_l0r_band_results->db_stdev[band_num-1] = 
                   db_l0r_band_results->db_stdev[band_num-1] / sca_counter;
        db_l0r_band_results->db_corrected[band_num-1] = 
                   db_l0r_band_results->db_corrected[band_num-1] / sca_counter;
    }

    /* Write out band level lamp statistics data */
    if (generate_report)
    {
        /* Write band level statistics to output report file */
        if (rps_output_lamp_band_results(db_l0r_band_results, 
                            &report_info, LAMP_WORKING_PAIR, lamp_on_time, 
                    post_warmup_start_time, post_warmup_stop_time, 
                    post_warmup_start_frame, post_warmup_stop_frame) != SUCCESS)
        {
            assert_true_with_message(  0, "Error from first "
                "write_lamp_response_sca_report_file" );
            rps_db_free_lamp_band_results(db_l0r_band_results);
            ias_misc_free_report_header( report_header );
            rps_db_free_lamp_detector_results(db_detector_results);
            rps_db_free_lamp_detector_results(db_detector_pan_results);
            rps_db_free_lamp_sca_results(db_l0r_sca_results);
            rps_db_free_lamp_band_results(db_l0r_band_results);
            free(buffer);
            free(pan_buffer);
            free(statistics_results_l0r);
            free(statistics_results_pan_l0r);
            return;
        }
    }

    /* Free allocated memory for buffer and pan_buffer */
    free(buffer);
    free(pan_buffer);

    /* Free db_detector_results structure */
    rps_db_free_lamp_detector_results(db_detector_results);
    rps_db_free_lamp_detector_results(db_detector_pan_results);

    /* Free statistics_results_l0r and statistics_results_pan_l0r 
       structure */
    free(statistics_results_l0r);      
    free(statistics_results_pan_l0r);      


    /* Free db_l0r_sca_results structure */
    rps_db_free_lamp_sca_results(db_l0r_sca_results);

    /* Free db_l0r_band_results structure */
    rps_db_free_lamp_band_results(db_l0r_band_results);

    ias_misc_free_report_header( report_header );

}

