#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "cgreen/cgreen.h"
#include "ias_satellite_attributes.h"
#include "ias_db_get_connect_info.h"
#include "ias_db_insert.h"
#include "ias_db.h"
#include "ias_db_query.h"
#include "ias_logging.h"
#include "../rps_local.h"
#include "../rps_blob.h"

#define STRLEN 64

static int test_database = 0;

/* Function prototypes */
struct ias_db_connection *db_open( void );
void db_close(struct ias_db_connection *);
int db_check_lamp_detector_in_database();
int db_check_lamp_response_in_database();

Ensure the_lamp_test_through_driver();


/*   Main test program */
int main( int argc, char *argv[] )
{
    char *enable_db_tests;

    setenv( "PCS_WOID", "L20091", 0 );
    TestSuite *suite = create_test_suite();
    add_test( suite, the_lamp_test_through_driver );

    /* See if the database tests should be enabled */
    enable_db_tests = getenv( "ENABLE_DB_TESTS" );
    if ( enable_db_tests )
    {
        test_database = 1;
        printf( "\n\tDatabase checks have been enabled\n" );
    }
    else
        printf( "\n\t***** Database checks will not be made *****\n" );

    return run_single_test( suite, "the_lamp_test_through_driver",
        create_text_reporter() );
    return run_test_suite( suite, create_text_reporter() );
}

Ensure the_lamp_test_through_driver()
{
    int status;
    struct stat buf;
    char l1r_output_name[] = "lamp2.h5";
    char cmd[256];
    static char *unit_test_dir = NULL;
    char command[PATH_MAX];

    /* Get the IAS_UNIT_TEST_DATA_DIR environment variable */
    unit_test_dir = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (unit_test_dir == NULL)
    {
        printf("The IAS_UNIT_TEST_DATA_DIR environment variable is not set\n");
        exit(EXIT_FAILURE);
    }

    status = snprintf(command, PATH_MAX,
            "sed 's|$IAS_UNIT_TEST_DATA_DIR|%s|' template_I15920.omf > "
            "I15920.omf", unit_test_dir);
    if (status < 0 || PATH_MAX < status)
    {
        assert_true_with_message(0, "Error preparing system command");
        return;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Error preparing OMF file for unit "
                "test data directory path setup");
        return;
    }

    /* Make sure the L1R output file is not present */
    status = stat( l1r_output_name, &buf );
    if ( status == 0 )
    {
        /* Remove the file */
        sprintf( cmd, "rm -f %s", l1r_output_name );
        status = system( cmd );
        if ( status )
        {
            assert_true_with_message( 0,
                "Error removing file: %s", l1r_output_name );
        }
        else
            printf("Removed file: %s\n", l1r_output_name );
    }

    /* Create the command for running create_l1r */
    if ( test_database )
    {
        status = snprintf(command, PATH_MAX, "export IAS_LOG_LEVEL=WARN; "
            "nice --adjustment=20 ../create_l1r band_lamp_test.odl");
    }
    else
    {
        status = snprintf(command, PATH_MAX, "export IAS_LOG_LEVEL=WARN; "
            "nice --adjustment=20 ../create_l1r band_lamp_test_no_db.odl");
    }
    if (status < 0 || PATH_MAX < status)
    {
        assert_true_with_message(0, "Error preparing system command for "
            "executing create_l1r");
        return;
    }
    
    /* Execute create_l1r */
    printf("Executing: %s\n", command);
    status = system( command );
    printf("Done: %s\n", command);
    assert_equal( status, 0 );
    if ( status )
    {
        /* Clean up and return */
        printf("Error executing: %s\n", command);
        sprintf( cmd, "nice --adjustment=20 rm -rf %s", l1r_output_name );
        status = system( cmd );
        if ( status )
        {
            assert_true_with_message( 0,
                "Error executing system command: %s", cmd );
        }
        return;
    }

    /* Clean up the L1R output file */
    sprintf( cmd, "nice --adjustment=20 rm -rf %s", l1r_output_name );
    status = system( cmd );
    if ( status )
    {
        assert_true_with_message( 0,
        "Error executing system command: %s", cmd );
    }

    /* Check the results from database reading */
    if ( test_database )
    {
        /* Check detector response information */
        status = db_check_lamp_detector_in_database();
        if (status != SUCCESS)
        {
            assert_true_with_message(  0, "check lamp_detector_response "
                "data in the database" );
        }

        /* Check lamp response information */
        status = db_check_lamp_response_in_database();
        if (status != SUCCESS)
        {
            assert_true_with_message(0, "check 'lamp_response' data in the "
                "database");
        }
    }
}

struct ias_db_connection *db_open(void)
{
    char *database_name = NULL;
    char *user_name = NULL;
    char *password = NULL;
    char *host = NULL;
    struct ias_db_connection *db = NULL;
    int status;

    /* Initialize the satellite attributes library */
    ias_sat_attr_initialize(IAS_L8);

    status = ias_db_initialize_database_lib( );
    if (status)
    {
        assert_true_with_message( 0, "Error from "
            "ias_db_initialize_database_lib");
        return NULL;
    }
    status = ias_db_get_connect_info(&database_name, &user_name,
        &password, &host);
    if (status)
    {
        assert_true_with_message( 0, "Error from "
            "ias_db_get_connect_info");
        free(database_name);
        free(user_name);
        free(password);
        free(host);
        ias_db_close_database_lib();
        return NULL;
    }

    db = ias_db_connect_to_database(database_name, user_name, password,
        host);
    free(database_name);
    free(user_name);
    free(password);
    free(host);
    if (!db)
    {
        assert_true_with_message( 0, "Error from "
            "ias_db_connect_to_database");
        ias_db_close_database_lib();
        return NULL;
    }

    return db;
}

void db_close(struct ias_db_connection *db)
{
    ias_db_close_connection(db);
    ias_db_close_database_lib();
}

/* This test confirms that data have been correctly written to the
   LAMP_RESPONSE table.  */
int db_check_lamp_response_in_database()
{
    char query_string[1000];
    char db_error_message[256];
    char rad_wo_common_id[IAS_CHAR_ID_SIZE];  
    char work_order_id[] = "I15920";
    struct ias_db_connection *db;
    struct ias_db_query *query = NULL;    /* Query data */
    double interval_start_time;
    double interval_stop_time;
    double lamp_on_time;
    double lamp_warmup_time;
    int interval_start_line;
    int interval_stop_line;
    int lamp_pair_id;
    int lamp_warmup_start_line;
    int lamp_warmup_stop_line;
    int status;

    /* Get a database connection */
    db = db_open();
    if ( !db )
    {
        assert_true_with_message(0, "Cannot connect to database");
        return ERROR;
    }

    /* Get the radiometric work order ID key for the specified work order
       string */
    status = ias_misc_get_char_id(db,
        "RADIOMETRIC_WORK_ORDER_COMMON", "radiometric_wo_common_id",
        work_order_id, rad_wo_common_id);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
    {
        assert_true_with_message( 0, "Invalid rad_wo_common_id key "
               "value for specified work order ID %s", work_order_id);
        db_close(db);
        return ERROR;
    }

    /* Get values from the LAMP_RESPONSE table */
    status = snprintf(query_string, sizeof(query_string),
        "SELECT lamp_pair_id, lamp_on_time, "
        "       interval_start_time, interval_stop_time, "
        "       interval_start_line, interval_stop_line, "
        "       lamp_warmup_time, lamp_warmup_start_line, "
        "       lamp_warmup_stop_line "
        "FROM   LAMP_RESPONSE "
        "WHERE  radiometric_wo_common_id=%s;", rad_wo_common_id);
    if ((status < 0) || (status >= sizeof(query_string)))
    {
        assert_true_with_message(0, "Error constructing LAMP_RESPONSE "
            "SQL string");
        db_close(db);
        return ERROR;
    }

    /* Execute the query and verify it worked */
    query = ias_db_perform_query(db, query_string);
    if (!ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, db_error_message,
            sizeof(db_error_message));
        assert_true_with_message(0, "Error from select query:  %s",
            db_error_message);
        ias_db_query_close(query);
        db_close(db);
        return ERROR;
    }

    /* Move DB "pointer" to the first valid record. */
    if (!ias_db_query_next(query))
    {
        assert_true_with_message(0, "Error reading LAMP_RESPONSE table");
        ias_db_query_close(query);
        db_close(db);
        return ERROR;
    }

    /* Start checking values from the query against the corresponding
       expected results */
    status = ias_db_query_int_value(query, &lamp_pair_id, 0);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error reading 'lamp_pair_id' from "
            "LAMP_RESPONSE table query" );
        ias_db_query_close(query);
        db_close(db);
        return ERROR;
    }
    assert_equal(lamp_pair_id, 2);

    status = ias_db_query_double_value(query, &lamp_on_time, 1);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error reading 'lamp_on_time' from "
            "LAMP_RESPONSE table query" );
        ias_db_query_close(query);
        db_close(db);
        return ERROR;
    }
    assert_double_equal(lamp_on_time, 2000.59998);

    status = ias_db_query_double_value(query, &interval_start_time, 2);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error reading 'interval_start_time' "
            "from LAMP_RESPONSE table query");
        ias_db_query_close(query);
        db_close(db);
        return ERROR;
    }
    assert_double_equal(interval_start_time, 2240.07349);

    status = ias_db_query_double_value(query, &interval_stop_time, 3);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error reading 'interval_stop_time' "
            "from LAMP_RESPONSE table query");
        ias_db_query_close(query);
        db_close(db);
        return ERROR;
    }
    assert_double_equal(interval_stop_time, 2242.27515);

    status = ias_db_query_int_value(query, &interval_start_line, 4);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error reading 'interval_start_line' from "
            "from LAMP_RESPONSE table query");
        ias_db_query_close(query);
        db_close(db);
        return ERROR;
    }
    assert_equal(interval_start_line, 55832);

    status = ias_db_query_int_value(query, &interval_stop_line, 5);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error reading 'interval_stop_line' from "
            "LAMP_RESPONSE table query");
        ias_db_query_close(query);
        db_close(db);
        return ERROR;
    }
    assert_equal(interval_stop_line, 56344);

    status = ias_db_query_double_value(query, &lamp_warmup_time, 6);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error reading 'lamp_warmup_time' from "
            "LAMP_RESPONSE table query");
        ias_db_query_close(query);
        db_close(db);
        return ERROR;
    }
    assert_double_equal(lamp_warmup_time, 239.473007);

    status = ias_db_query_int_value(query, &lamp_warmup_start_line, 7);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error reading 'lamp_warmup_start_line' "
            "from LAMP_RESPONSE table query");
        ias_db_query_close(query);
        db_close(db);
        return ERROR;
    }
    assert_equal(lamp_warmup_start_line, 141);

    status = ias_db_query_int_value(query, &lamp_warmup_stop_line, 8);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error reading 'lamp_warmup_stop_line' "
            "from LAMP_RESPONSE table query");
        ias_db_query_close(query);
        db_close(db);
        return ERROR;
    }
    assert_equal(lamp_warmup_stop_line, 56374);

    /* Release query resources */
    ias_db_query_close(query);
    db_close( db );

    return SUCCESS;
}

int db_check_lamp_detector_in_database()
{
    int status;
    struct ias_db_connection *db;
    char db_error_message[256];  /* Query error message buffer */
    char query_string[200];
    char rad_wo_common_id[IAS_CHAR_ID_SIZE];  
    char work_order_id[] = "I15920";
    int band_num = 1; 
    int sca_num = 1; 
    int sca_num2 = 2; 
    int det_num = 1; 
    char processing_level1[] = "CB";  
    char processing_level2[] = "CBG";  
    double db_mean;
    double db_stdev;
    double db_skewness;
    double db_kurtosis;
    struct ias_db_query *query = NULL;    /* Query data */

    db = db_open();
    if ( !db )
        return 1;

    /* Get the radiometric work order ID key for the specified work order
       string */
    status = ias_misc_get_char_id(db,
        "RADIOMETRIC_WORK_ORDER_COMMON", "radiometric_wo_common_id",
        work_order_id, rad_wo_common_id);
    assert_not_equal(status, ERROR);
    if (status == ERROR)
    {
        assert_true_with_message( 0, "Invalid rad_wo_common_id key "
               "value for specified work order ID %s", work_order_id);
        ias_db_close_connection(db);
        ias_db_close_database_lib();
        return 1;
    }

    /* Check a L0R band/SCA/detctor results from the database */
    sprintf(query_string, "SELECT mean, stdev, skewness, kurtosis FROM "
            "LAMP_RESPONSE_DETECTOR WHERE radiometric_wo_common_id=%s "
            "AND band_id = %d AND sensor_chip_assembly_id = %d "
            "AND detector_id = %d AND processing_level='%s';",
            rad_wo_common_id, band_num, sca_num, det_num, processing_level1);

    /* Execute the query and verify it worked */
    query = ias_db_perform_query(db, query_string);

    if (!ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, db_error_message,
                sizeof(db_error_message));
        assert_true_with_message(  0, "Error from "
            "select query" );
        ias_db_query_close(query);
        db_close( db );
        return 1;
    }

    /* Move DB "pointer" to the first valid record. */
    if (!ias_db_query_next(query))
    {
        ias_db_query_get_error_message(query, db_error_message,
                sizeof(db_error_message));

        assert_true_with_message(  0, "Error reading lamp detector table "
            "from expected results" );
        ias_db_query_close(query);
        db_close( db );
        return 1;
    }

    status = ias_db_query_double_value(query, &db_mean, 0);
    if (status != SUCCESS)
    {
        assert_true_with_message(  0, "Error obtaining mean "
            "from lamp_response_table" );
        ias_db_query_close(query);
        db_close( db );
        return 1;
    }
    assert_double_equal(db_mean, 2917.65942);

    status = ias_db_query_double_value(query, &db_stdev, 1);
    if (status != SUCCESS)
    {
        assert_true_with_message(  0, "Error obtaining stdev "
            "from lamp_response_table" );
        ias_db_query_close(query);
        db_close( db );
        return 1;
    }
    assert_double_equal(db_stdev, 11.3193932);

    status = ias_db_query_double_value(query, &db_skewness, 2);
    if (status != SUCCESS)
    {
        assert_true_with_message(  0, "Error obtaining skewness "
            "from lamp_response_table" );
        ias_db_query_close(query);
        db_close( db );
        return 1;
    }
    assert_double_equal(db_skewness, -0.088616289);

    status = ias_db_query_double_value(query, &db_kurtosis, 3);
    if (status != SUCCESS)
    {
        assert_true_with_message(  0, "Error obtaining kurtosis "
            "from lamp_response_table" );
        ias_db_query_close(query);
        db_close( db );
        return 1;
    }
    assert_double_equal(db_kurtosis, -1.16790104);

    /* Check a L1R band/SCA/detctor results from the database 
       note: based on current CPF used, the absolute_gain value for 
             band 1 and SCA 1 is 0, and for band 1 and SCA 2 is -5.96 */
    sprintf(query_string, "SELECT mean, stdev, skewness, kurtosis FROM "
            "LAMP_RESPONSE_DETECTOR WHERE radiometric_wo_common_id=%s "
            "AND band_id = %d AND sensor_chip_assembly_id = %d "
            "AND detector_id = %d AND processing_level='%s';",
            rad_wo_common_id, band_num, sca_num2, det_num, processing_level2);

    /* Execute the query and verify it worked */
    query = ias_db_perform_query(db, query_string);

    if (!ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, db_error_message,
                sizeof(db_error_message));
        assert_true_with_message(  0, "Error from "
            "select query" );
        ias_db_query_close(query);
        db_close( db );
        return 1;
    }

    /* Move DB "pointer" to the first valid record. */
    if (!ias_db_query_next(query))
    {
        ias_db_query_get_error_message(query, db_error_message,
                sizeof(db_error_message));

        assert_true_with_message(  0, "Error reading lamp detector table "
            "from expected results" );
        ias_db_query_close(query);
        db_close( db );
        return 1;
    }

    status = ias_db_query_double_value(query, &db_mean, 0);
    if (status != SUCCESS)
    {
        assert_true_with_message(  0, "Error obtaining mean "
            "from lamp_response_table" );
        ias_db_query_close(query);
        db_close( db );
        return 1;
    }
    assert_double_equal(db_mean, 181.761154);

    status = ias_db_query_double_value(query, &db_stdev, 1);
    if (status != SUCCESS)
    {
        assert_true_with_message(  0, "Error obtaining stdev "
            "from lamp_response_table" );
        ias_db_query_close(query);
        db_close( db );
        return 1;
    }
    assert_double_equal(db_stdev, 0.686773598);

    status = ias_db_query_double_value(query, &db_skewness, 2);
    if (status != SUCCESS)
    {
        assert_true_with_message(  0, "Error obtaining skewness "
            "from lamp_response_table" );
        ias_db_query_close(query);
        db_close( db );
        return 1;
    }
    assert_double_equal(db_skewness, 0.03492621);

    status = ias_db_query_double_value(query, &db_kurtosis, 3);
    if (status != SUCCESS)
    {
        assert_true_with_message(  0, "Error obtaining kurtosis "
            "from lamp_response_table" );
        ias_db_query_close(query);
        db_close( db );
        return 1;
    }
    assert_double_equal(db_kurtosis, -1.15159106);

    /* Release query resources */
    ias_db_query_close(query);
    db_close( db );

    return status;
}

