#include <stdio.h>
#include <string.h>
#include <math.h>

#include "ias_pixel_mask.h"
#include "ias_satellite_attributes.h"
#include "ias_db_get_connect_info.h"
#include "ias_db_insert.h"
#include "ias_db.h"
#include "ias_db_query.h"
#include "cgreen/cgreen.h"
#include "../rps_local.h"
#include "ias_logging.h"
#include "ias_const.h"
#include "ias_miscellaneous.h"

#define STRLEN 64
#define NUMBERS_OF_BANDS 11

Ensure char_lamp_response_band_test( void );

/* database prototypes */
struct ias_db_connection *db_open( void );
void db_close( struct ias_db_connection * );
int db_store_lamp_response_band_results
          (DB_CHAR_LAMP_RESPONSE_BAND_RESULTS *);
int db_check_lamp_band_in_database
(
    DB_CHAR_LAMP_RESPONSE_BAND_RESULTS *db_band_results
);

static int test_database = 0;

int main( int argc, char *argv[] )
{
    char *enable_db_tests = NULL;

    /* Suppress any IAS_LOG_DEBUG messages */
    ias_log_initialize("lamp_band_test");

    /* See if the database tests should be enabled */
    enable_db_tests = getenv( "ENABLE_DB_TESTS" );
    if ( enable_db_tests )
    {
        test_database = 1;
        printf( "\n\tDatabase checks have been enabled\n" );
    }
    else
        printf( "\n\t***** Database checks will not be made *****\n" );

    TestSuite *suite = create_test_suite();
    add_test( suite, char_lamp_response_band_test );

    return run_test_suite( suite, create_text_reporter() );
}


Ensure char_lamp_response_band_test( void )
{
    IAS_REPORT report_info;
    IAS_REPORT_HEADER *report_header = NULL;
    char work_order_id[IAS_WORKORDER_ID_SIZE];
    char routine_name[] = "Lamp Char tests - BAND level statistics";
    int status;

    /* Set up a standard header */
    setenv("PCS_WOID", "I4582", 1);
    strcpy(work_order_id, "I4582");
    report_header = ias_misc_initialize_report_header("",
        work_order_id, routine_name);
    if (!report_header)
    {
        assert_true_with_message( 0, "Error initializing report header");
        ias_misc_free_report_header(report_header);
        return;
    }
    ias_misc_set_header_pathrow(report_header, 0, 0);
    report_info.header = report_header;
    report_info.directory = NULL;

    DB_CHAR_LAMP_RESPONSE_BAND_RESULTS *db_band_results = NULL;
    int num_bands = NUMBERS_OF_BANDS;
    int band;
    float lamp_on_time = 5.0;
    float post_warmup_start_time = 155.0;
    float post_warmup_stop_time = 157.3;
    int post_warmup_start_frame = 12345;
    int post_warmup_stop_frame = 12857;

    /* Allocate memory for the 
                   DB_CHAR_LAMP_RESPONSE_DETECTOR_RESULTS structure */
    db_band_results = rps_db_allocate_lamp_band_results(num_bands);
    if (!db_band_results)
    {
        assert_true_with_message( 0, "Error allocating detector "
                                        "results database memory");
        ias_misc_free_report_header(report_header);
        return;
    }

    /* Set the common values for db_detector_results that are known
       here */
    db_band_results->num_bands = num_bands;
    db_band_results->processing_level = strdup("C");
    if (!db_band_results->processing_level)
    {
        assert_true_with_message( 0, "Error allocating band "
                                        "level processing level");
        ias_misc_free_report_header(report_header);
        return;
    }

    for (band = 0; band < num_bands; band++)
    {

            db_band_results->band[band] = band + 1;
            db_band_results->db_mean[band] = 100.0;
            db_band_results->db_stdev[band] = 2.0;
            db_band_results->db_corrected[band] = 100.0;
    }

    /* Store the stats in the database */
    if (test_database)
    {
        if (db_store_lamp_response_band_results( db_band_results))
        {
            assert_true_with_message( 0, "Error storing results to the"
                " database");
        }
        else
        {
            db_check_lamp_band_in_database(db_band_results);
        }
    }

    /* Write band level statistics to output report file */
    status = rps_output_lamp_band_results(db_band_results, 
                    &report_info, LAMP_WORKING_PAIR, lamp_on_time, 
                    post_warmup_start_time, post_warmup_stop_time, 
                    post_warmup_start_frame, post_warmup_stop_frame);
    assert_equal(status, SUCCESS);
  
    db_band_results->processing_level = strdup("CT");
    if (!db_band_results->processing_level)
    {
        assert_true_with_message( 0, "Error allocating band "
                                        "level processing level");
        ias_misc_free_report_header(report_header);
        return;
    }


    /* Store the stats in the database */
    if (test_database)
    {
        if ( db_store_lamp_response_band_results(db_band_results))
        {
            assert_true_with_message( 0, "Error storing results to the"
                " database");
        }
        else
        {
            db_check_lamp_band_in_database(db_band_results);
        }
    }

    /* Write band level statistics to output report file */
    status = rps_output_lamp_band_results(db_band_results, 
                    &report_info, LAMP_WORKING_PAIR, lamp_on_time, 
                    post_warmup_start_time, post_warmup_stop_time, 
                    post_warmup_start_frame, post_warmup_stop_frame);
    assert_equal(status, SUCCESS);

    rps_db_free_lamp_band_results(db_band_results);
    ias_misc_free_report_header( report_header );

}

struct ias_db_connection *db_open(void)
{
    char *database_name = NULL;
    char *user_name = NULL;
    char *password = NULL;
    char *host = NULL;
    struct ias_db_connection *db = NULL;
    int status;

    /* Initialize the satellite attributes library */
    ias_sat_attr_initialize(IAS_L8);

    status = ias_db_initialize_database_lib( );
    if (status)
    {
        assert_true_with_message( 0, "Error from "
            "ias_db_initialize_database_lib");
        return NULL;
    }
    status = ias_db_get_connect_info(&database_name, &user_name,
        &password, &host);
    if (status)
    {
        assert_true_with_message( 0, "Error from "
            "ias_db_get_connect_info");
        free(database_name);
        free(user_name);
        free(password);
        free(host);
        ias_db_close_database_lib();
        return NULL;
    }

    db = ias_db_connect_to_database(database_name, user_name, password,
        host);
    free(database_name);
    free(user_name);
    free(password);
    free(host);
    if (!db)
    {
        assert_true_with_message( 0, "Error from "
            "ias_db_connect_to_database");
        ias_db_close_database_lib();
        return NULL;
    }

    return db;
}

void db_close(struct ias_db_connection *db)
{
    ias_db_close_connection(db);
    ias_db_close_database_lib();
}

int db_store_lamp_response_band_results
(
    DB_CHAR_LAMP_RESPONSE_BAND_RESULTS *db_band_results
)
{
    int status;
    struct ias_db_connection *db;
    char db_error_message[IAS_SQL_RTN_MSGLEN]; /* Query error message buffer */
    struct ias_db_query *query = NULL;    /* IAS DB query handle data */
    char query_string[] = "select radiometric_wo_common_id from "
        "radiometric_work_order_common where work_order_id='HS8899'";
    char insert_string[] = "insert into radiometric_work_order_common "
        "(radiometric_wo_common_id, work_order_id) "
        "values (8899, 'HS8899');";
    char rad_common_id[IAS_CHAR_ID_SIZE];  

    db = db_open();
    if ( !db )
        return 1;

    status = ias_db_start_transaction( db );
    if ( status )
    {
        assert_true_with_message(  0, "Error from "
            "ias_db_start_transaction" );
        db_close( db );
        return 1;
    }

    /* Get the rad common ID */
    query = ias_db_perform_query(db, query_string);
    if ( !ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, db_error_message,
            sizeof(db_error_message));
        assert_true_with_message(  0, "Error from "
            "ias_db_perform_query: %s", db_error_message );
        ias_db_query_close(query);
        db_close( db );
        return 1;
    }
    else
    {
        /* If it isn't there create it */
        if (!ias_db_query_next(query))
        {
            /* Need to close the previous query before doing the insert */
            ias_db_query_close(query);
            query = ias_db_perform_query(db, insert_string);
            if ( ias_db_query_was_successful(query))
            {
                strcpy(rad_common_id, "8899");
            }
            else
            {
                ias_db_query_get_error_message(query, db_error_message,
                    sizeof(db_error_message));
                assert_true_with_message( 0, "Error from "
                    "ias_db_perform_query: %s", db_error_message);
                ias_db_query_close(query);
                db_close(db);
                return 1;
            }
        }
        /* If it was there use the ID */
        else
        {
            ias_db_query_string_value(query, 0, rad_common_id, 
                sizeof(rad_common_id));
            if ( strcmp(rad_common_id, "") == 0 )
            {
                assert_true_with_message( 0, "Empty string returned from "
                    "ias_db_query_string_value, couldn't get ID" );
                ias_db_query_close(query);
                db_close( db );
                return 1;
            }
        }
    }
    ias_db_query_close(query);

    /* Now clean out the lamp response detector table at the beginning of 
       db insertion */
    if (strcmp(db_band_results->processing_level, "C") == 0)
    {
        status = ias_db_delete_records(db, "LAMP_RESPONSE_BAND",
                             "radiometric_wo_common_id", "8899");
        if (status)
        {
            assert_true_with_message(  0, "Error from "
                    "ias_db_delete_records" );
            ias_db_query_close(query);
            db_close( db );
            return 1;
        }
    }
    
    /* Change the results for the new rad_common_id */
    strncpy(db_band_results->rad_common_id, rad_common_id, 
        sizeof(db_band_results->rad_common_id));
    db_band_results->rad_common_id[sizeof(
        db_band_results->rad_common_id) - 1] = '\0';

    /* Store the stats in the database */
    db_band_results->dbase = db;
    status = rps_store_lamp_band_in_database(
                  db_band_results);
    if ( status )
    {
        ias_db_rollback_transaction( db );
        assert_true_with_message(  0, "Error from "
            "rps_store_lamp_response_band_in_database" );
    }
    else
    {
        ias_db_commit_transaction( db );
    }

    db_close( db );
    return status;
}

int db_check_lamp_band_in_database
(
    DB_CHAR_LAMP_RESPONSE_BAND_RESULTS *db_band_results
)
{
    int status;
    struct ias_db_connection *db;
    char db_error_message[256];  /* Query error message buffer */
    char query_string[200];
    int rad_wo_common_id = 8899;
    int band; 
    int band_id;   
    double db_mean;
    double db_stdev;
    double db_corrected;
    struct ias_db_query *query = NULL;    /* Query data */

    sprintf(query_string, "SELECT band_id, "
            "mean, stdev, corrected FROM "
            "LAMP_RESPONSE_BAND WHERE radiometric_wo_common_id=%d "
            "AND processing_level='%s' ORDER BY "
            "band_id;", rad_wo_common_id, db_band_results->processing_level);

    db = db_open();
    if ( !db )
        return 1;

    /* Execute the query and verify it worked */
    query = ias_db_perform_query(db, query_string);

    if (!ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, db_error_message,
                sizeof(db_error_message));
        assert_true_with_message(  0, "Error from "
            "select query" );
        ias_db_query_close(query);
        db_close( db );
        return 1;
    }

    /* Move DB "pointer" to the first valid record. */
    if (!ias_db_query_next(query))
    {
        ias_db_query_get_error_message(query, db_error_message,
                sizeof(db_error_message));

        assert_true_with_message(  0, "Error reading lamp band table "
            "from expected results" );
        ias_db_query_close(query);
        db_close( db );
        return 1;
    }

    /* Check values from the table records */
    for (band = 0; band < db_band_results->num_bands; band++)
    {
        status = ias_db_query_int_value(query, &band_id, 0);
        if (status != SUCCESS)
        {
            assert_true_with_message(  0, "Error obtaining band_id "
            "from lamp_response_table" );
            ias_db_query_close(query);
            db_close( db );
            return 1;
        }
        assert_equal(band_id, db_band_results->band[band]);

        status = ias_db_query_double_value(query, &db_mean, 1);
        if (status != SUCCESS)
        {
            assert_true_with_message(  0, "Error obtaining mean "
            "from lamp_response_table" );
            ias_db_query_close(query);
            db_close( db );
            return 1;
        }
        assert_double_equal(db_mean, db_band_results->db_mean[band]);

        status = ias_db_query_double_value(query, &db_stdev, 2);
        if (status != SUCCESS)
        {
            assert_true_with_message(  0, "Error obtaining stdev "
            "from lamp_response_table" );
            ias_db_query_close(query);
            db_close( db );
            return 1;
        }
        assert_double_equal(db_stdev, db_band_results->db_stdev[band]);

        status = ias_db_query_double_value(query, &db_corrected, 3);
        if (status != SUCCESS)
        {
            assert_true_with_message(  0, "Error obtaining corrected "
            "from lamp_response_table" );
            ias_db_query_close(query);
            db_close( db );
            return 1;
        }
        assert_double_equal(db_corrected, 
                             db_band_results->db_corrected[band]);

        /* Advance DB "pointer" to next record as long as we're
           not at the last band */
        if (band < (db_band_results->num_bands - 1))
        {
            if (!ias_db_query_next(query))
            {
                assert_true_with_message(  0,"Cannot move to next "
                                          "detector's records");
                ias_db_query_close(query);
                return 1;
            }
        }
    }

    /* Release query resources */
    ias_db_query_close(query);
    db_close( db );

    return status;
}

