#include <stdio.h>
#include <string.h>
#include <math.h>

#include "ias_pixel_mask.h"
#include "ias_satellite_attributes.h"
#include "ias_db_get_connect_info.h"
#include "ias_db_insert.h"
#include "ias_db.h"
#include "ias_db_query.h"
#include "cgreen/cgreen.h"
#include "../rps_local.h"
#include "ias_logging.h"
#include "ias_const.h"
#include "ias_miscellaneous.h"

#define STRLEN 64

Ensure char_lamp_response_detector_test(void);

/* database prototypes */
int db_store_lamp_response_detector_results
    (DB_CHAR_LAMP_RESPONSE_DETECTOR_RESULTS *);
struct ias_db_connection *db_open( void );
void db_close(struct ias_db_connection *);
int db_check_lamp_detector_in_database
(DB_CHAR_LAMP_RESPONSE_DETECTOR_RESULTS *);

static int test_database = 0;

int main(int argc, char *argv[])
{
    char *enable_db_tests = NULL;

    /* Suppress any IAS_LOG_DEBUG messages */
    ias_log_initialize("lamp_detector_test");

    /* See if the database tests should be enabled */
    enable_db_tests = getenv("ENABLE_DB_TESTS");
    if (enable_db_tests)
    {
        test_database = 1;
        printf("\n\tDatabase checks have been enabled\n");
    }
    else
        printf("\n\t***** Database checks will not be made *****\n");

    TestSuite *suite = create_test_suite();
    add_test(suite, char_lamp_response_detector_test);

    return run_test_suite(suite, create_text_reporter());
}


Ensure char_lamp_response_detector_test(void)
{
    IAS_PIXEL_MASK *pm = NULL;
    IAS_REPORT report_information;
    IAS_REPORT_HEADER *report_header = NULL;
    char work_order_id[IAS_WORKORDER_ID_SIZE];
    char routine_name[] = "Lamp Char tests";
    int status;
    unsigned int det;

    /* Initialize the attributes lib */
    status = ias_sat_attr_initialize(IAS_L8);
    if (status)
    {
        assert_true_with_message( 0,
            "ias_sat_attr_initialize returned ERROR");
        return;
    }

    /* Create the pixel mask */
    pm = ias_pm_create(1, 1, 8, 14);
    assert_true(pm != NULL);
    if (pm == NULL)
    {
        assert_true_with_message( 0, "Error from creating Pixel Mask");
        return;
    }

    /* Add the inoperable detectors */
    assert_false( ias_pm_add_pixels(pm, 3-1, 1-1, 14, IAS_RPS_INOP_DETECTOR));

    /* Set up a standard header */
    setenv("PCS_WOID", "I4582", 1);
    strcpy(work_order_id, "I4582");
    report_header = ias_misc_initialize_report_header("",
        work_order_id, routine_name);
    if (!report_header)
    {
        assert_true_with_message( 0, "Error initializing report header");
        return;
    }
    ias_misc_set_header_pathrow(report_header, 0, 0);
    report_information.header = report_header;
    report_information.directory = NULL;

    /* Test API part */
    IAS_RPS_HISTOGRAM_INTERMEDIATE_RESULTS *statistics_results_l0r = NULL;
    IAS_RPS_HISTOGRAM_INTERMEDIATE_RESULTS *statistics_results_l1r = NULL;
    DB_CHAR_LAMP_RESPONSE_DETECTOR_RESULTS *db_detector_l0r_results = NULL;
    DB_CHAR_LAMP_RESPONSE_DETECTOR_RESULTS *db_detector_l1r_results = NULL;
    DB_CHAR_LAMP_RESPONSE_DETECTOR_RESULTS *db_detector_check_results = NULL;
    DB_CHAR_LAMP_RESPONSE_ONE_SCA_RESULTS db_one_sca_l0r_results;
    DB_CHAR_LAMP_RESPONSE_ONE_SCA_RESULTS db_one_sca_l1r_results;
    int number_of_scas = 4;
    int number_of_bands = 8;
    int band_num;
    int sca_num;
    int det_index;
    int num_lines = 14;
    int num_samps = 8;
    int lamp_pair_id = 1;
    float lamp_on_time = 5.0;
    float post_warmup_start_time = 155.0;
    float post_warmup_stop_time = 157.3;
    int post_warmup_start_frame = 12345;
    int post_warmup_stop_frame = 12857;

    /* Allocate memory for the statistics_results structures */
    statistics_results_l0r = malloc(num_samps 
               * sizeof(IAS_RPS_HISTOGRAM_INTERMEDIATE_RESULTS));
    statistics_results_l1r = malloc(num_samps 
               * sizeof(IAS_RPS_HISTOGRAM_INTERMEDIATE_RESULTS));
    if (!statistics_results_l0r || !statistics_results_l1r)
    {
        ias_misc_free_report_header( report_header );
        free(statistics_results_l0r);
        free(statistics_results_l1r);
        assert_true_with_message(  0, "Error from "
                             "allocating statistics results structures" );
        return;
    }

    for (band_num = 1; band_num <= number_of_bands; band_num++)
    {
        for (sca_num = 1; sca_num <= number_of_scas; sca_num++)
        {
            for (det_index = 0; det_index < num_samps; det_index++)
            {
                /* Set needed member values for the 
                   L0R IAS_HISTOGRAM_INTERMEDIATE_RESULTS structure */
                statistics_results_l0r[det_index].mean[0] = 1.1;
                statistics_results_l0r[det_index].mean[1] = 1.9;
                statistics_results_l0r[det_index].standard_deviation[0] = 0.1;
                statistics_results_l0r[det_index].standard_deviation[1] = 0.9;
                statistics_results_l0r[det_index].skewness[0] = 0.01;
                statistics_results_l0r[det_index].skewness[1] = 0.09;
                statistics_results_l0r[det_index].kurtosis[0] = 0.02;
                statistics_results_l0r[det_index].kurtosis[1] = 0.08;

                /* Set needed member values for the 
                   L1R IAS_HISTOGRAM_INTERMEDIATE_RESULTS structure */
                statistics_results_l1r[det_index].mean[0] = 2.1;
                statistics_results_l1r[det_index].mean[1] = 2.9;
                statistics_results_l1r[det_index].standard_deviation[0] = 0.2;
                statistics_results_l1r[det_index].standard_deviation[1] = 0.8;
                statistics_results_l1r[det_index].skewness[0] = 0.02;
                statistics_results_l1r[det_index].skewness[1] = 0.08;
                statistics_results_l1r[det_index].kurtosis[0] = 0.03;
                statistics_results_l1r[det_index].kurtosis[1] = 0.07;
            }

            /* Allocate memory for the 
                   DB_CHAR_LAMP_RESPONSE_DETECTOR_RESULTS structure */
            db_detector_l0r_results = rps_db_allocate_lamp_detector_results(
                       num_samps);
            if (!db_detector_l0r_results)
            {
                assert_true_with_message( 0, "Error allocating detector "
                                        "results database memory");
                ias_misc_free_report_header( report_header );
                free(statistics_results_l0r);
                free(statistics_results_l1r);
                return;
            }

            /* Set the common values for db_detector_l0r_results that are known
               here */
            db_detector_l0r_results->band = band_num;
            db_detector_l0r_results->sca = sca_num;
            db_detector_l0r_results->num_detectors = num_samps;
            db_detector_l0r_results->processing_level = strdup("C");
            if (db_detector_l0r_results->processing_level == NULL)
            {
                assert_true_with_message( 0, "Error from duplicating "
                             "processing level for db_band_results");
                ias_misc_free_report_header( report_header );
                free(statistics_results_l0r);
                free(statistics_results_l1r);
                rps_db_free_lamp_detector_results(db_detector_l0r_results);
                return;
            }

            /* Set the common values for db_one_sca_l0r_results that are known
               here */
            db_one_sca_l0r_results.band = band_num;
            db_one_sca_l0r_results.sca = sca_num;
            db_one_sca_l0r_results.processing_level = strdup("C");
            if (db_one_sca_l0r_results.processing_level == NULL)
            {
                assert_true_with_message( 0, "Error from duplicating "
                             "processing level for db_one_sca_l0r_results");
                ias_misc_free_report_header( report_header );
                free(statistics_results_l0r);
                free(statistics_results_l1r);
                rps_db_free_lamp_detector_results(db_detector_l0r_results);
                return;
            }

            /* Call char_lamp_detector_response algorithm code */
            status = char_lamp_detector_response(band_num, sca_num, 
                num_samps, num_lines, lamp_pair_id, lamp_on_time, 
                    post_warmup_start_time, post_warmup_stop_time, 
                    post_warmup_start_frame, post_warmup_stop_frame, pm,
                statistics_results_l0r, &report_information, 
                db_detector_l0r_results, &db_one_sca_l0r_results);
            if (status != SUCCESS)
            {
                assert_true_with_message( 0, "Error from allocating "
                             "individual statistics results structures");
                ias_misc_free_report_header( report_header );
                free(statistics_results_l0r);
                free(statistics_results_l1r);
                rps_db_free_lamp_detector_results(db_detector_l0r_results);
                return;
            }

            /* Store the stats in the database */
            if (test_database)
            {
                /* Allocate memory for db_detector_check_results structure 
                   Note: the db_detector_results is used because the 
                         db_detector_l1r_results are freed in the 
                         rps_store_lamp_telemetry_in_database routine*/
                db_detector_check_results 
                    = rps_db_allocate_lamp_detector_results(num_samps);
                if (!db_detector_check_results)
                {
                      assert_true_with_message( 0, "Error allocating "
                                  "check detector results database memory");
                      ias_misc_free_report_header( report_header );
                      free(statistics_results_l0r);
                      free(statistics_results_l1r);
                      rps_db_free_lamp_detector_results(
                              db_detector_l0r_results);
                      return;
                }

                /* Copy the db_detector_l1r_results structure to 
                   db_detector_check_results structure */
                db_detector_check_results->band 
                           = db_detector_l0r_results->band; 
                db_detector_check_results->sca 
                           = db_detector_l0r_results->sca; 
                db_detector_check_results->num_detectors 
                           = db_detector_l0r_results->num_detectors; 
                db_detector_check_results->processing_level =
                    strdup(db_detector_l0r_results->processing_level);
                if (!db_detector_check_results->processing_level)
                {
                      assert_true_with_message( 0, "Error allocating "
                                  "check detector results database memory");
                      ias_misc_free_report_header( report_header );
                      free(statistics_results_l0r);
                      free(statistics_results_l1r);
                      rps_db_free_lamp_detector_results(
                              db_detector_l0r_results);
                      return;
                }

                for (det = 0; det < db_detector_l0r_results->num_detectors; 
                     det++)
                {
                    db_detector_check_results->detector[det] = det + 1;
                    db_detector_check_results->db_mean[det] 
                              = db_detector_l0r_results->db_mean[det];
                    db_detector_check_results->db_stdev[det] 
                              = db_detector_l0r_results->db_stdev[det];
                    db_detector_check_results->db_skewness[det] 
                              = db_detector_l0r_results->db_skewness[det];
                    db_detector_check_results->db_kurtosis[det] 
                              = db_detector_l0r_results->db_kurtosis[det];
                    db_detector_check_results->db_corrected[det] 
                              = db_detector_l0r_results->db_corrected[det];
                }

                if ( db_store_lamp_response_detector_results( 
                                        db_detector_l0r_results ))
                {
                    assert_true_with_message( 0, "Error storing results to "
                                             "the database");
                }
                else
                {
                    db_check_lamp_detector_in_database(
                                        db_detector_check_results);
                }

                /* The use of strdup needs to be freed each time it's used */ 
                free(db_detector_check_results->processing_level);
            
            }
            else
            {
                rps_db_free_lamp_detector_results(db_detector_l0r_results);
            }

            /* Check the returned db_one_sca_l0r_results */
            assert_equal(db_one_sca_l0r_results.band, band_num);
            assert_equal(db_one_sca_l0r_results.sca, sca_num);
            assert_string_equal(db_one_sca_l0r_results.processing_level, 
                                "C");

            /* Allocate memory for the 
                   DB_CHAR_LAMP_RESPONSE_DETECTOR_RESULTS structure */
            db_detector_l1r_results = rps_db_allocate_lamp_detector_results(
                       num_samps);
            if (!db_detector_l1r_results)
            {
                assert_true_with_message( 0, "Error allocating detector "
                                        "results database memory");
                ias_misc_free_report_header( report_header );
                free(statistics_results_l0r);
                free(statistics_results_l1r);
                return;
            }

            /* Set the common values for db_detector_l1r_results that are known
               here */
            db_detector_l1r_results->band = band_num;
            db_detector_l1r_results->sca = sca_num;
            db_detector_l1r_results->num_detectors = num_samps;
            db_detector_l1r_results->processing_level = strdup("CT");
            if (db_detector_l1r_results->processing_level == NULL)
            {
                assert_true_with_message( 0, "Error from duplicating "
                             "processing level for db_band_results");
                ias_misc_free_report_header( report_header );
                free(statistics_results_l0r);
                free(statistics_results_l1r);
                rps_db_free_lamp_detector_results(db_detector_l1r_results);
                return;
            }

            /* Set the common values for db_one_sca_l1r_results that are known
               here */
            db_one_sca_l1r_results.band = band_num;
            db_one_sca_l1r_results.sca = sca_num;
            db_one_sca_l1r_results.processing_level = strdup("CT"); 
            if (db_one_sca_l1r_results.processing_level == NULL)
            {
                assert_true_with_message( 0, "Error from duplicating "
                             "processing level for db_one_sca_l1r_results");
                ias_misc_free_report_header( report_header );
                free(statistics_results_l0r);
                free(statistics_results_l1r);
                rps_db_free_lamp_detector_results(db_detector_l1r_results);
                return;
            }

            status = char_lamp_detector_response(band_num, sca_num, 
                     num_samps, num_lines, lamp_pair_id, lamp_on_time, 
                     post_warmup_start_time, post_warmup_stop_time, 
                     post_warmup_start_frame, post_warmup_stop_frame, pm,
                     statistics_results_l1r, &report_information,
                     db_detector_l1r_results, &db_one_sca_l1r_results);
            assert_equal(status, SUCCESS);

            /* Store the stats in the database */
            if (test_database)
            {
                /* Copy the db_detector_l1r_results structure to 
                   db_detector_check_results structure */
                db_detector_check_results->band 
                           = db_detector_l1r_results->band; 
                db_detector_check_results->sca 
                           = db_detector_l1r_results->sca; 
                db_detector_check_results->num_detectors 
                           = db_detector_l1r_results->num_detectors; 
                db_detector_check_results->processing_level =
                    strdup(db_detector_l1r_results->processing_level);
                if (!db_detector_check_results->processing_level)
                {
                      assert_true_with_message( 0, "Error allocating "
                                  "check detector results database memory");
                      ias_misc_free_report_header( report_header );
                      free(statistics_results_l0r);
                      free(statistics_results_l1r);
                      rps_db_free_lamp_detector_results(
                               db_detector_l1r_results);
                      return;
                }

                for (det = 0; det < db_detector_l1r_results->num_detectors; 
                     det++)
                {
                    db_detector_check_results->detector[det] = det + 1;
                    db_detector_check_results->db_mean[det] 
                              = db_detector_l1r_results->db_mean[det];
                    db_detector_check_results->db_stdev[det] 
                              = db_detector_l1r_results->db_stdev[det];
                    db_detector_check_results->db_skewness[det] 
                              = db_detector_l1r_results->db_skewness[det];
                    db_detector_check_results->db_kurtosis[det] 
                              = db_detector_l1r_results->db_kurtosis[det];
                    db_detector_check_results->db_corrected[det] 
                              = db_detector_l1r_results->db_corrected[det];
                }

                if ( db_store_lamp_response_detector_results( 
                                        db_detector_l1r_results ))
                {
                    assert_true_with_message( 0, "Error storing results to "
                                             "the database");
                }
                else
                {
                    db_check_lamp_detector_in_database(
                                        db_detector_check_results);
                    rps_db_free_lamp_detector_results(
                                        db_detector_check_results);
                }
            }
            else
            {
                rps_db_free_lamp_detector_results(db_detector_l1r_results);
            }

            /* Check the returned db_one_sca_l1r_results */
            assert_equal(db_one_sca_l1r_results.band, band_num);
            assert_equal(db_one_sca_l1r_results.sca, sca_num);
            assert_string_equal(db_one_sca_l1r_results.processing_level, 
                                "CT");

            /* The use of strdup needs to be freed each time it's used */ 
            free(db_one_sca_l0r_results.processing_level);
            free(db_one_sca_l1r_results.processing_level);
        }
    }

    /* Free the allocated memory */
    free(statistics_results_l0r); 
    free(statistics_results_l1r); 

    ias_pm_destroy(pm);
    ias_misc_free_report_header( report_header );
}

struct ias_db_connection *db_open(void)
{
    char *database_name = NULL;
    char *user_name = NULL;
    char *password = NULL;
    char *host = NULL;
    struct ias_db_connection *db = NULL;
    int status;

    status = ias_db_initialize_database_lib( );
    if (status)
    {
        assert_true_with_message( 0, "Error from "
            "ias_db_initialize_database_lib");
        return NULL;
    }
    status = ias_db_get_connect_info(&database_name, &user_name,
        &password, &host);
    if (status)
    {
        assert_true_with_message( 0, "Error from "
            "ias_db_get_connect_info");
        free(database_name);
        free(user_name);
        free(password);
        free(host);
        ias_db_close_database_lib();
        return NULL;
    }

    db = ias_db_connect_to_database(database_name, user_name, password,
        host);
    free(database_name);
    free(user_name);
    free(password);
    free(host);
    if (!db)
    {
        assert_true_with_message( 0, "Error from "
            "ias_db_connect_to_database");
        ias_db_close_database_lib();
        return NULL;
    }

    return db;
}

void db_close(struct ias_db_connection *db)
{
    ias_db_close_connection(db);
    ias_db_close_database_lib();
}

int db_store_lamp_response_detector_results
(
    DB_CHAR_LAMP_RESPONSE_DETECTOR_RESULTS *db_detector_results
)
{
    int status;
    struct ias_db_connection *db;
    char db_error_message[IAS_SQL_RTN_MSGLEN]; /* Query error message buffer */
    struct ias_db_query *query = NULL;    /* IAS DB query handle data */
    char query_string[] = "select radiometric_wo_common_id from "
        "radiometric_work_order_common where work_order_id='HS8899'";
    char insert_string[] = "insert into radiometric_work_order_common "
        "(radiometric_wo_common_id, work_order_id) "
        "values (8899, 'HS8899');";
    char rad_common_id[IAS_CHAR_ID_SIZE];  

    db = db_open();
    if ( !db )
        return 1;

    status = ias_db_start_transaction( db );
    if ( status )
    {
        assert_true_with_message(  0, "Error from "
            "ias_db_start_transaction" );
        db_close( db );
        return 1;
    }

    /* Get the rad common ID */
    query = ias_db_perform_query(db, query_string);
    if ( !ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, db_error_message,
            sizeof(db_error_message));
        assert_true_with_message(  0, "Error from "
            "ias_db_perform_query: %s", db_error_message );
        ias_db_query_close(query);
        db_close( db );
        return 1;
    }
    else
    {
        /* If it isn't there create it */
        if (!ias_db_query_next(query))
        {
            /* Need to close the previous query before doing the insert */
            ias_db_query_close(query);
            query = ias_db_perform_query(db, insert_string);
            if ( ias_db_query_was_successful(query))
            {
                strcpy(rad_common_id, "8899");
            }
            else
            {
                ias_db_query_get_error_message(query, db_error_message,
                    sizeof(db_error_message));
                assert_true_with_message( 0, "Error from "
                    "ias_db_perform_query: %s", db_error_message);
                ias_db_query_close(query);
                db_close(db);
                return 1;
            }
        }
        /* If it was there use the ID */
        else
        {
            ias_db_query_string_value(query, 0, rad_common_id, 
                sizeof(rad_common_id));
            if ( strcmp(rad_common_id, "") == 0 )
            {
                assert_true_with_message( 0, "Empty string returned from "
                    "ias_db_query_string_value, couldn't get ID" );
                ias_db_query_close(query);
                db_close( db );
                return 1;
            }
        }
    }
    ias_db_query_close(query);

    /* Now clean out the lamp response detector table at the beginning */
    if (db_detector_results->band == 1 && db_detector_results->sca == 1
        && strcmp(db_detector_results->processing_level, "C") == 0)
    {
        status = ias_db_delete_records(db, "LAMP_RESPONSE_DETECTOR",
                             "radiometric_wo_common_id", "8899");
        if (status)
        {
            assert_true_with_message(  0, "Error from "
                    "ias_db_delete_records" );
            ias_db_query_close(query);
            db_close( db );
            return 1;
        }
    }
    
    /* Change the results for the new rad_common_id */
    strncpy(db_detector_results->rad_common_id, rad_common_id, 
        sizeof(db_detector_results->rad_common_id));;
    db_detector_results->rad_common_id[sizeof(
        db_detector_results->rad_common_id) - 1] = '\0';

    /* Store the stats in the database */
    db_detector_results->dbase = db;
    status = rps_store_lamp_detector_in_database(
                  db_detector_results);
    if ( status )
    {
        ias_db_rollback_transaction( db );
        assert_true_with_message(  0, "Error from "
            "rps_store_lamp_response_band_in_database" );
    }
    else
    {
        ias_db_commit_transaction( db );
    }

    db_close( db );
    return status;
}

int db_check_lamp_detector_in_database
(
    DB_CHAR_LAMP_RESPONSE_DETECTOR_RESULTS *db_detector_results
)
{
    int status;
    struct ias_db_connection *db;
    char db_error_message[256];  /* Query error message buffer */
    char query_string[200];
    int rad_wo_common_id = 8899;
    int det; 
    int detector_id;   
    double db_mean;
    double db_stdev;
    double db_skewness;
    double db_kurtosis;
    double db_corrected;
    struct ias_db_query *query = NULL;    /* Query data */

    sprintf(query_string, "SELECT detector_id, "
            "mean, stdev, skewness, kurtosis, corrected FROM "
            "LAMP_RESPONSE_DETECTOR WHERE radiometric_wo_common_id=%d "
            "AND band_id = %d AND sensor_chip_assembly_id = %d "
            "AND processing_level='%s' ORDER BY detector_id;",
            rad_wo_common_id, db_detector_results->band, 
            db_detector_results->sca, db_detector_results->processing_level);

    db = db_open();
    if ( !db )
        return 1;

    /* Execute the query and verify it worked */
    query = ias_db_perform_query(db, query_string);

    if (!ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, db_error_message,
                sizeof(db_error_message));
        assert_true_with_message(  0, "Error from "
            "select query" );
        ias_db_query_close(query);
        db_close( db );
        return 1;
    }

    /* Move DB "pointer" to the first valid record. */
    if (!ias_db_query_next(query))
    {
        ias_db_query_get_error_message(query, db_error_message,
                sizeof(db_error_message));

        assert_true_with_message(  0, "Error reading lamp detector table "
            "from expected results" );
        ias_db_query_close(query);
        db_close( db );
        return 1;
    }

    /* Check values from the table records */
    for (det = 0; det < db_detector_results->num_detectors; det++)
    {
        status = ias_db_query_int_value(query, &detector_id, 0);
        if (status != SUCCESS)
        {
            assert_true_with_message(  0, "Error obtaining detector_id "
            "from lamp_response_table" );
            ias_db_query_close(query);
            db_close( db );
            return 1;
        }
        assert_equal(detector_id, db_detector_results->detector[det]);

        status = ias_db_query_double_value(query, &db_mean, 1);
        if (status != SUCCESS)
        {
            assert_true_with_message(  0, "Error obtaining mean "
            "from lamp_response_table" );
            ias_db_query_close(query);
            db_close( db );
            return 1;
        }
        assert_double_equal(db_mean, db_detector_results->db_mean[det]);

        status = ias_db_query_double_value(query, &db_stdev, 2);
        if (status != SUCCESS)
        {
            assert_true_with_message(  0, "Error obtaining stdev "
            "from lamp_response_table" );
            ias_db_query_close(query);
            db_close( db );
            return 1;
        }
        assert_double_equal(db_stdev, db_detector_results->db_stdev[det]);

        status = ias_db_query_double_value(query, &db_skewness, 3);
        if (status != SUCCESS)
        {
            assert_true_with_message(  0, "Error obtaining skewness "
            "from lamp_response_table" );
            ias_db_query_close(query);
            db_close( db );
            return 1;
        }
        assert_double_equal(db_skewness, db_detector_results->db_skewness[det]);

        status = ias_db_query_double_value(query, &db_kurtosis, 4);
        if (status != SUCCESS)
        {
            assert_true_with_message(  0, "Error obtaining kurtosis "
            "from lamp_response_table" );
            ias_db_query_close(query);
            db_close( db );
            return 1;
        }
        assert_double_equal(db_kurtosis, db_detector_results->db_kurtosis[det]);

        status = ias_db_query_double_value(query, &db_corrected, 5);
        if (status != SUCCESS)
        {
            assert_true_with_message(  0, "Error obtaining corrected "
            "from lamp_response_table" );
            ias_db_query_close(query);
            db_close( db );
            return 1;
        }
        assert_double_equal(db_corrected, 
                             db_detector_results->db_corrected[det]);

        /* Advance DB "pointer" to next record as long as we're
           not at the last band */
        if (det < (db_detector_results->num_detectors - 1))
        {
            if (!ias_db_query_next(query))
            {
                assert_true_with_message(  0,"Cannot move to next "
                                          "detector's records");
                ias_db_query_close(query);
                return 1;
            }
        }
    }

    /* Release query resources */
    ias_db_query_close(query);
    db_close( db );

    return status;
}

