#include <stdio.h>
#include <string.h>
#include <math.h>

#include "ias_pixel_mask.h"
#include "ias_satellite_attributes.h"
#include "ias_db_get_connect_info.h"
#include "ias_db_insert.h"
#include "ias_db.h"
#include "ias_db_query.h"
#include "cgreen/cgreen.h"
#include "../rps_local.h"
#include "ias_logging.h"
#include "ias_const.h"
#include "ias_miscellaneous.h"

#define STRLEN 64

Ensure char_lamp_response_test(void);

/* database prototypes */
struct ias_db_connection *db_open(void);
void db_close( struct ias_db_connection * );
int db_store_lamp_response_results(DB_CHAR_LAMP_RESPONSE_RESULTS *);
int db_check_lamp_response_in_database(DB_CHAR_LAMP_RESPONSE_RESULTS *);

static int test_database = 0;

int main(int argc, char *argv[])
{
    char *enable_db_tests = NULL;

    /* Suppress any IAS_LOG_DEBUG messages */
    ias_log_initialize("lamp_response_test");

    /* See if the database tests should be enabled */
    enable_db_tests = getenv("ENABLE_DB_TESTS");
    if ( enable_db_tests )
    {
        test_database = 1;
        printf("\n\tDatabase checks have been enabled\n");
    }
    else
        printf("\n\t***** Database checks will not be made *****\n");

    TestSuite *suite = create_test_suite();
    add_test(suite, char_lamp_response_test);

    return run_test_suite(suite, create_text_reporter());
}


Ensure char_lamp_response_test(void)
{
    IAS_REPORT report_info;
    IAS_REPORT_HEADER *report_header = NULL;
    char work_order_id[IAS_WORKORDER_ID_SIZE];
    char routine_name[] = "Lamp Char tests - response";
    int status;

    /* Set up a standard header */
    setenv("PCS_WOID", "I4582", 1);
    strcpy(work_order_id, "I4582");
    report_header = ias_misc_initialize_report_header("",
        work_order_id, routine_name);
    if (!report_header)
    {
        assert_true_with_message(0, "Error initializing report header" );
        return;
    }
    ias_misc_set_header_pathrow(report_header, 0, 0);
    report_info.header = report_header;
    report_info.directory = NULL;

    /* Test lamp_response data part */
    DB_CHAR_LAMP_RESPONSE_RESULTS *db_response_results = NULL;

    /* Allocate memory for db_response_results */
    db_response_results = malloc(sizeof(DB_CHAR_LAMP_RESPONSE_RESULTS));
    assert_not_equal(db_response_results, NULL);

    /* Set the db_response_results data from the blob */
    db_response_results->lamp_pair_id = 1;
    db_response_results->lamp_on_time = 1234.5;
    db_response_results->interval_start_line = 100;
    db_response_results->interval_stop_line = 200;
    db_response_results->lamp_warmup_time = 240.0;
    db_response_results->lamp_warmup_start_line 
                            = 5;
    db_response_results->lamp_warmup_stop_line 
                            = 99;
    float post_warmup_start_time = 86398.0;
    float post_warmup_stop_time = 86458.0;

    db_response_results->interval_start_time =  
           db_response_results->lamp_on_time + post_warmup_start_time;
    db_response_results->interval_stop_time =  
           db_response_results->lamp_on_time + post_warmup_stop_time;

    /* Insert Lamp Response data into DB here */
    if (test_database)
    {
        if ( db_store_lamp_response_results(db_response_results))
        {
            assert_true_with_message(0, "Error storing results to the"
                " database");
        }
        else
        {
            db_check_lamp_response_in_database(db_response_results);
        }
    }

    /* Write lamp response data to output report file */
    status = rps_output_lamp_response_results(db_response_results, 
                      &report_info);
    assert_equal(status, SUCCESS);

    free(db_response_results);
    ias_misc_free_report_header(report_header);
}

struct ias_db_connection *db_open(void)
{
    char *database_name = NULL;
    char *user_name = NULL;
    char *password = NULL;
    char *host = NULL;
    struct ias_db_connection *db = NULL;
    int status;

    /* Initialize the satellite attributes library */
    ias_sat_attr_initialize(IAS_L8);

    status = ias_db_initialize_database_lib( );
    if (status)
    {
        assert_true_with_message(0, "Error from "
            "ias_db_initialize_database_lib");
        return NULL;
    }
    status = ias_db_get_connect_info(&database_name, &user_name,
        &password, &host);
    if (status)
    {
        assert_true_with_message(0, "Error from "
            "ias_db_get_connect_info");
        free(database_name);
        free(user_name);
        free(password);
        free(host);
        ias_db_close_database_lib();
        return NULL;
    }

    db = ias_db_connect_to_database(database_name, user_name, password,
        host);
    free(database_name);
    free(user_name);
    free(password);
    free(host);
    if (!db)
    {
        assert_true_with_message(0, "Error from "
            "ias_db_connect_to_database");
        ias_db_close_database_lib();
        return NULL;
    }

    return db;
}

void db_close(struct ias_db_connection *db)
{
    ias_db_close_connection(db);
    ias_db_close_database_lib();
}

int db_store_lamp_response_results
(
    DB_CHAR_LAMP_RESPONSE_RESULTS *db_response_results
)
{
    int status;
    struct ias_db_connection *db;
    char db_error_message[256];  /* Query error message buffer */
    struct ias_db_query *query = NULL;    /* IAS DB query handle data */
    char query_string[] = "select radiometric_wo_common_id from "
        "radiometric_work_order_common where work_order_id='HS8899'";
    char insert_string[] = "insert into radiometric_work_order_common "
        "(radiometric_wo_common_id, work_order_id) "
        "values (8899, 'HS8899');";
    char rad_common_id[IAS_CHAR_ID_SIZE];  

    db = db_open();
    if (!db)
        return 1;

    status = ias_db_start_transaction(db);
    if (status)
    {
        assert_true_with_message(0, "Error from "
            "ias_db_start_transaction");
        db_close(db);
        return 1;
    }

    /* Get the rad common ID */
    query = ias_db_perform_query(db, query_string);
    if ( !ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, db_error_message,
            sizeof(db_error_message));
        assert_true_with_message(0, "Error from "
            "ias_db_perform_query: %s", db_error_message);
        ias_db_query_close(query);
        db_close(db);
        return 1;
    }
    else
    {
        /* If it isn't there create it */
        if (!ias_db_query_next(query))
        {
            /* Need to close the previous query before doing the insert */
            ias_db_query_close(query);
            query = ias_db_perform_query(db, insert_string);
            if ( ias_db_query_was_successful(query))
            {
                strcpy(rad_common_id, "8899");
            }
            else
            {
                ias_db_query_get_error_message(query, db_error_message,
                    sizeof(db_error_message));
                assert_true_with_message(0, "Error from "
                    "ias_db_perform_query: %s", db_error_message);
                ias_db_query_close(query);
                db_close(db);
                return 1;
            }
        }
        /* If it was there use the ID */
        else
        {
            ias_db_query_string_value(query, 0, rad_common_id, 
                sizeof(rad_common_id));
            if ( strcmp(rad_common_id, "") == 0 )
            {
                assert_true_with_message( 0, "Empty string returned from "
                    "ias_db_query_string_value, couldn't get ID" );
                ias_db_query_close(query);
                db_close(db);
                return 1;
            }
        }
    }
    ias_db_query_close(query);

    /* Now clean out the lamp response detector table */
    status = ias_db_delete_records(db, "LAMP_RESPONSE",
                             "radiometric_wo_common_id",
                             "8899");
    
    /* Change the results for the new rad_common_id */
    strncpy(db_response_results->rad_common_id, rad_common_id, 
        sizeof(db_response_results->rad_common_id));
    db_response_results->rad_common_id[sizeof(
        db_response_results->rad_common_id) - 1] = '\0';

    /* Store the stats in the database */
    db_response_results->dbase = db;
    status = rps_store_lamp_response_in_database(
                  db_response_results);
    if (status)
    {
        ias_db_rollback_transaction(db);
        assert_true_with_message(0, "Error from "
            "rps_store_lamp_response_band_in_database");
    }
    else
    {
        ias_db_commit_transaction(db);
    }

    db_close(db);
    return status;
}

int db_check_lamp_response_in_database
(
    DB_CHAR_LAMP_RESPONSE_RESULTS *db_response_results
)
{
    int status;
    struct ias_db_connection *db;
    char db_error_message[256];  /* Query error message buffer */
    char query_string[200];
    int rad_wo_common_id = 8899;
    int lamp_pair_id;   
    double lamp_on_time;
    double interval_start_time;
    double interval_stop_time;
    int interval_start_line;
    int interval_stop_line;
    double lamp_warmup_time;
    int lamp_warmup_start_line;
    int lamp_warmup_stop_line;
    struct ias_db_query *query = NULL;    /* Query data */

    sprintf(query_string, "SELECT lamp_pair_id, lamp_on_time, "
            "interval_start_time, interval_stop_time, interval_start_line, "
            "interval_stop_line, lamp_warmup_time, lamp_warmup_start_line, "
            "lamp_warmup_stop_line FROM LAMP_RESPONSE "
            "WHERE radiometric_wo_common_id=%d;", rad_wo_common_id);

    db = db_open();
    if (!db)
        return 1;

    /* Execute the query and verify it worked */
    query = ias_db_perform_query(db, query_string);

    if (!ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, db_error_message,
                sizeof(db_error_message));
        assert_true_with_message(0, "Error from lamp response "
            "select query");
        ias_db_query_close(query);
        db_close( db );
        return 1;
    }

    /* Move DB "pointer" to the first valid record. */
    if (!ias_db_query_next(query))
    {
        ias_db_query_get_error_message(query, db_error_message,
                sizeof(db_error_message));

        assert_true_with_message(0, "Error reading lamp response table "
            "from expected results");
        ias_db_query_close(query);
        db_close(db);
        return 1;
    }

    /* Check values from the table records */
    status = ias_db_query_int_value(query, &lamp_pair_id, 0);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error obtaining lamp_pair_id"
        "from lamp_response_table");
        ias_db_query_close(query);
        db_close(db);
        return 1;
    }
    assert_equal(lamp_pair_id, db_response_results->lamp_pair_id);

    status = ias_db_query_double_value(query, &lamp_on_time, 1);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error obtaining lamp_on_time "
            "from lamp_response_table");
        ias_db_query_close(query);
        db_close(db);
        return 1;
    }
    assert_double_equal(lamp_on_time, db_response_results->lamp_on_time);

    status = ias_db_query_double_value(query, &interval_start_time, 2);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error obtaining interval_start_time "
            "from lamp_response_table");
        ias_db_query_close(query);
        db_close(db);
        return 1;
    }
    assert_double_equal(interval_start_time, 
                        db_response_results->interval_start_time);

    status = ias_db_query_double_value(query, &interval_stop_time, 3);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error obtaining interval_stop_time "
            "from lamp_response_table");
        ias_db_query_close(query);
        db_close(db);
        return 1;
    }
    assert_double_equal(interval_stop_time, 
                        db_response_results->interval_stop_time);

    status = ias_db_query_int_value(query, &interval_start_line, 4);
    if (status != SUCCESS)
    {
        assert_true_with_message( 0, "Error obtaining interval_start_line "
        "from lamp_response_table" );
        ias_db_query_close(query);
        db_close(db);
        return 1;
    }
    assert_equal(interval_start_line, db_response_results->interval_start_line);

    status = ias_db_query_int_value(query, &interval_stop_line, 5);
    if (status != SUCCESS)
    {
        assert_true_with_message( 0, "Error obtaining interval_stop_line "
        "from lamp_response_table" );
        ias_db_query_close(query);
        db_close(db);
        return 1;
    }
    assert_equal(interval_stop_line, db_response_results->interval_stop_line);

    status = ias_db_query_double_value(query, &lamp_warmup_time, 6);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error obtaining lamp_warmup_time "
            "from lamp_response_table");
        ias_db_query_close(query);
        db_close(db);
        return 1;
    }
    assert_double_equal(lamp_warmup_time, 
                        db_response_results->lamp_warmup_time);

    status = ias_db_query_int_value(query, &lamp_warmup_start_line, 7);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error obtaining "
        "lamp_warmup_start_line from lamp_response_table");
        ias_db_query_close(query);
        db_close(db);
        return 1;
    }
    assert_equal(lamp_warmup_start_line, 
                 db_response_results->lamp_warmup_start_line);

    status = ias_db_query_int_value(query, &lamp_warmup_stop_line, 8);
    if (status != SUCCESS)
    {
        assert_true_with_message( 0, "Error obtaining "
        "lamp_warmup_stop_line from lamp_response_table" );
        ias_db_query_close(query);
        db_close(db);
        return 1;
    }
    assert_equal(lamp_warmup_stop_line, 
                 db_response_results->lamp_warmup_stop_line);

    /* Release query resources */
    ias_db_query_close(query);
    db_close(db);

    return status;
}

