#include <stdio.h>
#include <string.h>
#include <math.h>

#include "ias_pixel_mask.h"
#include "ias_satellite_attributes.h"
#include "ias_db_get_connect_info.h"
#include "ias_db_insert.h"
#include "ias_db.h"
#include "ias_db_query.h"
#include "cgreen/cgreen.h"
#include "ias_logging.h"
#include "ias_const.h"
#include "ias_math.h"
#include "ias_miscellaneous.h"
#include "../rps_local.h"

#define STRLEN 64

Ensure char_lamp_telemetry_test(void);

/* database prototypes */
struct ias_db_connection *db_open(void);
void db_close(struct ias_db_connection *);
int db_store_lamp_telemetry_results(DB_CHAR_LAMP_TELEMETRY_RESULTS *);
int db_check_lamp_telemetry_in_database(DB_CHAR_LAMP_TELEMETRY_RESULTS *);

int math_compute_mean
(
    const float *sample,       /* I: pointer to the data */
    int nsamps,                 /* I: number of samples */
    float *mean                /* O: mean of the data */
);

int math_compute_stdev
(
    const float *sample,   /* I: pointer to the data */
    int nsamps,             /* I: number of samples */
    float mean,            /* I: mean of the samples */
    float *std_dev         /* O: standard deviation of the data */
);

static int test_database = 0;

int main(int argc, char *argv[])
{
    char *enable_db_tests = NULL;

    /* Suppress any IAS_LOG_DEBUG messages */
    ias_log_initialize("lamp_telemetry_test");

    /* See if the database tests should be enabled */
    enable_db_tests = getenv("ENABLE_DB_TESTS");
    if ( enable_db_tests )
    {
        test_database = 1;
        printf("\n\tDatabase checks have been enabled\n");
    }
    else
        printf("\n\t***** Database checks will not be made *****\n");

    TestSuite *suite = create_test_suite();
    add_test(suite, char_lamp_telemetry_test);

    return run_test_suite(suite, create_text_reporter());
}


Ensure char_lamp_telemetry_test(void)
{
    IAS_REPORT report_info;
    IAS_REPORT_HEADER *report_header = NULL;
    char work_order_id[IAS_WORKORDER_ID_SIZE];
    char routine_name[] = "Lamp Char tests - telemetry";
    int status;

    /* Set up a standard header */
    setenv("PCS_WOID", "I4582", 1);
    strcpy(work_order_id, "I4582");
    report_header = ias_misc_initialize_report_header("",
        work_order_id, routine_name);
    if (!report_header)
    {
        assert_true_with_message(0, "Error initializing report header");
        return;
    }
    ias_misc_set_header_pathrow(report_header, 0, 0);
    report_info.header = report_header;
    report_info.directory = NULL;

    /* Test preprocess part */
    float focal_plane_temp1[] = {209.09,209.09,209.25};      
                                    /* Focal plane temperature 1 list */
    float focal_plane_temp2[] = {209.01,210.09,210.00};       
                                    /* Focal plane temperature 2 list */
    int num_focal_temp1 = 3;        /* Number of scene focal plane module 7 
                                       temperature */
    int num_focal_temp2 = 3;        /* Number of scene focal plane module 14 
                                       interface temperature */
    float mean_fp_temp1;           /* averaged focal plane temperature 1 */
    float mean_fp_temp2;           /* averaged focal plane temperature 2 */
    float stdev_fp_temp1;          /* stdev of focal plane temperature 1 */
    float stdev_fp_temp2;          /* stdev of focal plane temperature 2 */
    float mean_stdev_focal_temp[4];/* Mean and stdev of focal temperatures */

    /* Now the mean and stdev for focal_plane_temp1 */
    status = math_compute_mean(focal_plane_temp1, num_focal_temp1, 
                                   &mean_fp_temp1);
    if (status != SUCCESS)
    {
        assert_true_with_message( 0, "Error from calculating"
                             "focal plane mean temperature1" );
        return;
    }

    status = math_compute_stdev(focal_plane_temp1, num_focal_temp1, 
                                    mean_fp_temp1, &stdev_fp_temp1);
    if (status != SUCCESS)
    {
        assert_true_with_message( 0, "Error from calculating"
                             "stdev of  plane mean temperature1" );
        return;
    }

    /* Now the mean and stdev for focal_plane_temp2 */
    status = math_compute_mean(focal_plane_temp2, num_focal_temp2, 
                                   &mean_fp_temp2);
    if (status != SUCCESS)
    {
        assert_true_with_message( 0, "Error from calculating"
                             "focal plane mean temperature2" );
        return;
    }

    status = math_compute_stdev(focal_plane_temp2, num_focal_temp2, 
                                    mean_fp_temp2, &stdev_fp_temp2);
    if (status != SUCCESS)
    {
        assert_true_with_message( 0, "Error from calculating"
                             "stdev of  plane mean temperature2" );
        return;
    }

    /* Allocate memory for mean_stdev_focal_temp pointer and populate values 
       for it */
    mean_stdev_focal_temp[0] = mean_fp_temp1;
    mean_stdev_focal_temp[1] = stdev_fp_temp1;
    mean_stdev_focal_temp[2] = mean_fp_temp2;
    mean_stdev_focal_temp[3] = stdev_fp_temp2;

    /* Test lamp_response and lamp_telemetry data part */
    DB_CHAR_LAMP_TELEMETRY_RESULTS *db_telemetry_results = NULL;
    float lamp_current[] = {1.1, 1.2, 1.3};
    float lamp_voltage[] = {10.1, 10.2, 10.3};
    float photodiode_micro_amp[] = {11.1, 11.2, 11.3};
    float lamp_housing_temp[] = {100.1, 100.2, 100.3};
    float monitor_diode_temp[] = {110.1, 110.2, 110.3};
    int data_size = 3;
    float lamp_on_time = 5.0;
    float post_warmup_start_time = 155.0;
    float post_warmup_stop_time = 157.3;
    int post_warmup_start_frame = 12345;
    int post_warmup_stop_frame = 12857;

    /* Allocate memory for db_telemetry_results */
    db_telemetry_results = malloc(sizeof(DB_CHAR_LAMP_TELEMETRY_RESULTS));
    assert_not_equal(db_telemetry_results, NULL);

    /* Calculate mean and std dev for the lamp telemetry data */
    status = math_compute_mean(lamp_current, data_size, 
                           &db_telemetry_results->lamp_current_a_mean);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error calculating mean of "
                                        "lamp current");
        return;
    }

    status = math_compute_stdev(lamp_current, data_size, 
                         db_telemetry_results->lamp_current_a_mean, 
                         &db_telemetry_results->lamp_current_a_stdev);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error calculating stdev of "
                                        "lamp current");
        return;
    }
    db_telemetry_results->lamp_current_b_mean = 
        db_telemetry_results->lamp_current_a_mean;
    db_telemetry_results->lamp_current_b_stdev = 
        db_telemetry_results->lamp_current_a_stdev;
             
    /* Calculate mean and std dev for the lamp voltage data */
    status = math_compute_mean(lamp_voltage, data_size, 
                          &db_telemetry_results->lamp_voltage_a_mean);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error calculating mean of "
                                        "lamp voltage");
        return;
    }

    status = math_compute_stdev(lamp_voltage, data_size, 
                          db_telemetry_results->lamp_voltage_a_mean, 
                          &db_telemetry_results->lamp_voltage_a_stdev);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error calculating stdev of "
                                        "lamp voltage");
        return;
    }
    db_telemetry_results->lamp_voltage_b_mean = 
        db_telemetry_results->lamp_voltage_a_mean;
    db_telemetry_results->lamp_voltage_b_stdev = 
        db_telemetry_results->lamp_voltage_a_stdev;

             
    /* Calculate mean and std dev for the photodiode current data */
    status = math_compute_mean(photodiode_micro_amp, data_size, 
                       &db_telemetry_results->photodiode_micro_amp_a_mean);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error calculating mean of "
                                        "photodiode current");
        return;
    }

    status = math_compute_stdev(photodiode_micro_amp, data_size, 
                       db_telemetry_results->photodiode_micro_amp_a_mean, 
                       &db_telemetry_results->photodiode_micro_amp_a_stdev);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error calculating stdev of "
                                        "photodiode current");
        return;
    }
    db_telemetry_results->photodiode_micro_amp_b_mean = 
        db_telemetry_results->photodiode_micro_amp_a_mean;
    db_telemetry_results->photodiode_micro_amp_b_stdev = 
        db_telemetry_results->photodiode_micro_amp_a_stdev;
             
    /* Calculate mean and std dev for the lamp housing temperature data */
    status = math_compute_mean(lamp_housing_temp, data_size, 
                    &db_telemetry_results->lamp_housing_temp_a_mean);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error calculating mean of "
                                        "lamp housing temperature");
        return;
    }

    status = math_compute_stdev(lamp_housing_temp, data_size, 
                    db_telemetry_results->lamp_housing_temp_a_mean, 
                    &db_telemetry_results->lamp_housing_temp_a_stdev);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error calculating stdev of "
                                        "lamp housing temperature");
        return;
    }
    db_telemetry_results->lamp_housing_temp_b_mean = 
        db_telemetry_results->lamp_housing_temp_a_mean;
    db_telemetry_results->lamp_housing_temp_b_stdev = 
        db_telemetry_results->lamp_housing_temp_a_stdev;
             
    /* Calculate mean and std dev for the monitor diode temperature data */
    status = math_compute_mean(monitor_diode_temp, data_size, 
                    &db_telemetry_results->monitor_diode_temp_a_mean);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error calculating mean of "
                                        "monitor diode temperature");
        return;
    }

    status = math_compute_stdev(monitor_diode_temp, data_size, 
                    db_telemetry_results->monitor_diode_temp_a_mean, 
                    &db_telemetry_results->monitor_diode_temp_a_stdev);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error calculating stdev of "
                                        "lamp housing temperature");
        return;
    }
    db_telemetry_results->monitor_diode_temp_b_mean = 
        db_telemetry_results->monitor_diode_temp_a_mean;
    db_telemetry_results->monitor_diode_temp_b_stdev = 
        db_telemetry_results->monitor_diode_temp_a_stdev;

    db_telemetry_results->fp_temp_7_mean = mean_stdev_focal_temp[0];
    db_telemetry_results->fp_temp_7_stdev = mean_stdev_focal_temp[1];
    db_telemetry_results->fp_temp_14_mean = mean_stdev_focal_temp[2];
    db_telemetry_results->fp_temp_14_stdev = mean_stdev_focal_temp[3];

    /* Insert Lamp Telemetry data into DB here */
    if (test_database)
    {
        if (db_store_lamp_telemetry_results(db_telemetry_results))
        {
            assert_true_with_message(0, "Error storing results to the"
                " database");
        }
        else
        {
            db_check_lamp_telemetry_in_database(db_telemetry_results);
        }
    }

    /* Write out band lamp telemetry data */
    status = rps_output_lamp_telemetry_results(db_telemetry_results, 
                    &report_info, LAMP_WORKING_PAIR, lamp_on_time, 
                    post_warmup_start_time, post_warmup_stop_time, 
                    post_warmup_start_frame, post_warmup_stop_frame);
    assert_equal(status, SUCCESS);
  
    free(db_telemetry_results);
    ias_misc_free_report_header( report_header );
}

struct ias_db_connection *db_open(void)
{
    char *database_name = NULL;
    char *user_name = NULL;
    char *password = NULL;
    char *host = NULL;
    struct ias_db_connection *db = NULL;
    int status;

    /* Initialize the satellite attributes library */
    ias_sat_attr_initialize(IAS_L8);

    status = ias_db_initialize_database_lib( );
    if (status)
    {
        assert_true_with_message(0, "Error from "
            "ias_db_initialize_database_lib");
        return NULL;
    }
    status = ias_db_get_connect_info(&database_name, &user_name,
        &password, &host);
    if (status)
    {
        assert_true_with_message(0, "Error from "
            "ias_db_get_connect_info");
        free(database_name);
        free(user_name);
        free(password);
        free(host);
        ias_db_close_database_lib();
        return NULL;
    }

    db = ias_db_connect_to_database(database_name, user_name, password,
        host);
    free(database_name);
    free(user_name);
    free(password);
    free(host);
    if (!db)
    {
        assert_true_with_message(0, "Error from "
            "ias_db_connect_to_database");
        ias_db_close_database_lib();
        return NULL;
    }

    return db;
}

void db_close(struct ias_db_connection *db)
{
    ias_db_close_connection(db);
    ias_db_close_database_lib();
}

/* Open the report file and allocate temporary storage variables */

int db_store_lamp_telemetry_results
(
    DB_CHAR_LAMP_TELEMETRY_RESULTS *db_telemetry_results
)
{
    int status;
    struct ias_db_connection *db;
    char db_error_message[256];  /* Query error message buffer */
    struct ias_db_query *query = NULL;    /* IAS DB query handle data */
    char query_string[] = "select radiometric_wo_common_id from "
        "radiometric_work_order_common where work_order_id='HS8899'";
    char insert_string[] = "insert into radiometric_work_order_common "
        "(radiometric_wo_common_id, work_order_id) "
        "values (8899, 'HS8899');";
    char rad_common_id[IAS_CHAR_ID_SIZE];  

    db = db_open();
    if (!db)
        return 1;

    status = ias_db_start_transaction(db);
    if ( status )
    {
        assert_true_with_message(0, "Error from "
            "ias_db_start_transaction");
        db_close(db);
        return 1;
    }

    /* Get the rad common ID */
    query = ias_db_perform_query(db, query_string);
    if (!ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, db_error_message,
            sizeof(db_error_message));
        assert_true_with_message(0, "Error from "
            "ias_db_perform_query: %s", db_error_message);
        ias_db_query_close(query);
        db_close(db);
        return 1;
    }
    else
    {
        /* If it isn't there create it */
        if (!ias_db_query_next(query))
        {
            /* Need to close the previous query before doing the insert */
            ias_db_query_close(query);
            query = ias_db_perform_query(db, insert_string);
            if ( ias_db_query_was_successful(query))
            {
                strcpy(rad_common_id, "8899");
            }
            else
            {
                ias_db_query_get_error_message(query, db_error_message,
                    sizeof(db_error_message));
                assert_true_with_message(0, "Error from "
                    "ias_db_perform_query: %s", db_error_message);
                ias_db_query_close(query);
                db_close(db);
                return 1;
            }
        }
        /* If it was there use the ID */
        else
        {
            ias_db_query_string_value(query, 0, rad_common_id, 
                sizeof(rad_common_id));
            if ( strcmp(rad_common_id, "") == 0 )
            {
                assert_true_with_message( 0, "Empty string returned from "
                    "ias_db_query_string_value, couldn't get ID" );
                ias_db_query_close(query);
                db_close(db);
                return 1;
            }
        }
    }
    ias_db_query_close(query);

    /* Clear up the lamp_telemetry table */
    status = ias_db_delete_records(db, "LAMP_TELEMETRY",
                             "radiometric_wo_common_id",
                             "8899");

    /* Change the results for the new rad_common_id */
    strncpy(db_telemetry_results->rad_common_id, rad_common_id, 
        sizeof(db_telemetry_results->rad_common_id));
    db_telemetry_results->rad_common_id[sizeof(
        db_telemetry_results->rad_common_id) - 1] = '\0';

    /* Store the stats in the database */
    db_telemetry_results->dbase = db;
    status = rps_store_lamp_telemetry_in_database(
                  db_telemetry_results);
    if (status)
    {
        ias_db_rollback_transaction(db);
        assert_true_with_message(0, "Error from "
            "rps_store_lamp_telemetry_in_database");
    }
    else
    {
        ias_db_commit_transaction(db);
    }

    db_close(db);
    return status;
}

int db_check_lamp_telemetry_in_database
(
    DB_CHAR_LAMP_TELEMETRY_RESULTS *db_telemetry_results
)
{
    int status;
    struct ias_db_connection *db;
    char db_error_message[256];  /* Query error message buffer */
    char query_string[800];
    int rad_wo_common_id = 8899;
    double lamp_current_a_mean;
    double lamp_current_a_stdev;
    double lamp_current_b_mean;
    double lamp_current_b_stdev;
    double lamp_voltage_a_mean;
    double lamp_voltage_a_stdev;
    double lamp_voltage_b_mean;
    double lamp_voltage_b_stdev;
    double photodiode_micro_amp_a_mean;
    double photodiode_micro_amp_a_stdev;
    double photodiode_micro_amp_b_mean;
    double photodiode_micro_amp_b_stdev;
    double lamp_housing_temp_a_mean;
    double lamp_housing_temp_a_stdev;
    double lamp_housing_temp_b_mean;
    double lamp_housing_temp_b_stdev;
    double monitor_diode_temp_a_mean;
    double monitor_diode_temp_a_stdev;
    double monitor_diode_temp_b_mean;
    double monitor_diode_temp_b_stdev;
    double fp_temp_7_mean;
    double fp_temp_7_stdev;
    double fp_temp_14_mean;
    double fp_temp_14_stdev;
    struct ias_db_query *query = NULL;    /* Query data */

    status = snprintf(query_string, sizeof(query_string),
            "SELECT lamp_current_a_mean, "
            "lamp_current_a_stdev, lamp_current_b_mean, lamp_current_b_stdev, "
            "lamp_voltage_a_mean, lamp_voltage_a_stdev, lamp_voltage_b_mean, "
            "lamp_voltage_b_stdev, photodiode_current_a_mean, "
            "photodiode_current_a_stdev, photodiode_current_b_mean, "
            "photodiode_current_b_stdev, lamp_housing_temp_a_mean, "
            "lamp_housing_temp_a_stdev, lamp_housing_temp_b_mean,"
            "lamp_housing_temp_b_stdev, monitor_diode_temp_a_mean, "
            "monitor_diode_temp_a_stdev, monitor_diode_temp_b_mean, "
            "monitor_diode_temp_b_stdev, fp_temp_7_mean, "
            "fp_temp_7_stdev, fp_temp_14_mean, fp_temp_14_stdev "
            "FROM LAMP_TELEMETRY "
            "WHERE radiometric_wo_common_id=%d;", rad_wo_common_id);
    if (status < 0 || status >= sizeof(query_string))
    {
        assert_true_with_message(0, "query_string buffer too small");
        return 1;
    }

    db = db_open();
    if (!db)
        return 1;

    /* Execute the query and verify it worked */
    query = ias_db_perform_query(db, query_string);

    if (!ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, db_error_message,
                sizeof(db_error_message));
        assert_true_with_message(0, "Error from lamp telemetry "
            "select query");
        ias_db_query_close(query);
        db_close( db );
        return 1;
    }

    /* Move DB "pointer" to the first valid record. */
    if (!ias_db_query_next(query))
    {
        ias_db_query_get_error_message(query, db_error_message,
                sizeof(db_error_message));

        assert_true_with_message(0, "Error reading lamp_telemetry table "
            "from expected results");
        ias_db_query_close(query);
        db_close(db);
        return 1;
    }

    /* Check values from the table records */
    status = ias_db_query_double_value(query, &lamp_current_a_mean, 0);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error obtaining lamp_current_a_mean "
            "from lamp_telemetry table");
        ias_db_query_close(query);
        db_close(db);
        return 1;
    }
    assert_double_equal(lamp_current_a_mean, 
        db_telemetry_results->lamp_current_a_mean);

    status = ias_db_query_double_value(query, &lamp_current_a_stdev, 1);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error obtaining lamp_current_a_stdev "
            "from lamp_telemetry table");
        ias_db_query_close(query);
        db_close(db);
        return 1;
    }
    assert_double_equal(lamp_current_a_stdev, 
        db_telemetry_results->lamp_current_a_stdev);

    status = ias_db_query_double_value(query, &lamp_current_b_mean, 2);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error obtaining lamp_current_b_mean "
            "from lamp_telemetry table");
        ias_db_query_close(query);
        db_close(db);
        return 1;
    }
    assert_double_equal(lamp_current_b_mean, 
        db_telemetry_results->lamp_current_b_mean);

    status = ias_db_query_double_value(query, &lamp_current_b_stdev, 3);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error obtaining lamp_current_b_stdev "
            "from lamp_telemetry table");
        ias_db_query_close(query);
        db_close(db);
        return 1;
    }
    assert_double_equal(lamp_current_b_stdev, 
        db_telemetry_results->lamp_current_b_stdev);

    status = ias_db_query_double_value(query, &lamp_voltage_a_mean, 4);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error obtaining lamp_voltage_a_mean "
            "from lamp_telemetry table");
        ias_db_query_close(query);
        db_close(db);
        return 1;
    }
    assert_double_equal(lamp_voltage_a_mean, 
        db_telemetry_results->lamp_voltage_a_mean);

    status = ias_db_query_double_value(query, &lamp_voltage_a_stdev, 5);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error obtaining lamp_voltage_a_stdev "
            "from lamp_telemetry table");
        ias_db_query_close(query);
        db_close(db);
        return 1;
    }
    assert_double_equal(lamp_voltage_a_stdev, 
        db_telemetry_results->lamp_voltage_a_stdev);

    status = ias_db_query_double_value(query, &lamp_voltage_b_mean, 6);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error obtaining lamp_voltage_b_mean "
            "from lamp_telemetry table");
        ias_db_query_close(query);
        db_close(db);
        return 1;
    }
    assert_double_equal(lamp_voltage_b_mean, 
        db_telemetry_results->lamp_voltage_b_mean);

    status = ias_db_query_double_value(query, &lamp_voltage_b_stdev, 7);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error obtaining lamp_voltage_b_stdev "
            "from lamp_telemetry table");
        ias_db_query_close(query);
        db_close(db);
        return 1;
    }
    assert_double_equal(lamp_voltage_b_stdev, 
        db_telemetry_results->lamp_voltage_b_stdev);

    status = ias_db_query_double_value(query, &photodiode_micro_amp_a_mean, 8);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error obtaining "
            "photodiode_micro_amp_a_mean from lamp_telemetry table");
        ias_db_query_close(query);
        db_close(db);
        return 1;
    }
    assert_double_equal(photodiode_micro_amp_a_mean, 
        db_telemetry_results->photodiode_micro_amp_a_mean);

    status = ias_db_query_double_value(query, &photodiode_micro_amp_a_stdev, 9);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error obtaining "
            "photodiode_micro_amp_a_stdev from lamp_telemetry table");
        ias_db_query_close(query);
        db_close(db);
        return 1;
    }
    assert_double_equal(photodiode_micro_amp_a_stdev, 
        db_telemetry_results->photodiode_micro_amp_a_stdev);

    status = ias_db_query_double_value(query, &photodiode_micro_amp_b_mean, 10);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error obtaining photodiode_micro_amp_b_mean "
            "from lamp_telemetry table");
        ias_db_query_close(query);
        db_close(db);
        return 1;
    }
    assert_double_equal(photodiode_micro_amp_b_mean, 
        db_telemetry_results->photodiode_micro_amp_b_mean);

    status = ias_db_query_double_value(query, &photodiode_micro_amp_b_stdev, 11);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error obtaining "
            "photodiode_micro_amp_b_stdev from lamp_telemetry table");
        ias_db_query_close(query);
        db_close(db);
        return 1;
    }
    assert_double_equal(photodiode_micro_amp_b_stdev, 
        db_telemetry_results->photodiode_micro_amp_b_stdev);

    status = ias_db_query_double_value(query, &lamp_housing_temp_a_mean, 12);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error obtaining lamp_housing_temp_a_mean "
            "from lamp_telemetry table");
        ias_db_query_close(query);
        db_close(db);
        return 1;
    }
    assert_double_equal(lamp_housing_temp_a_mean, 
        db_telemetry_results->lamp_housing_temp_a_mean);

    status = ias_db_query_double_value(query, &lamp_housing_temp_a_stdev, 13);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error obtaining lamp_housing_temp_a_stdev"
            " from lamp_telemetry table");
        ias_db_query_close(query);
        db_close(db);
        return 1;
    }
    assert_double_equal(lamp_housing_temp_a_stdev, 
        db_telemetry_results->lamp_housing_temp_a_stdev);

    status = ias_db_query_double_value(query, &lamp_housing_temp_b_mean, 14);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error obtaining lamp_housing_temp_b_mean "
            "from lamp_telemetry table");
        ias_db_query_close(query);
        db_close(db);
        return 1;
    }
    assert_double_equal(lamp_housing_temp_b_mean, 
        db_telemetry_results->lamp_housing_temp_b_mean);

    status = ias_db_query_double_value(query, &lamp_housing_temp_b_stdev, 15);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error obtaining lamp_housing_temp_b_stdev"
            " from lamp_telemetry table");
        ias_db_query_close(query);
        db_close(db);
        return 1;
    }
    assert_double_equal(lamp_housing_temp_b_stdev, 
        db_telemetry_results->lamp_housing_temp_b_stdev);

    status = ias_db_query_double_value(query, &monitor_diode_temp_a_mean, 16);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error obtaining monitor_diode_temp_a_mean"
            " from lamp_telemetry table");
        ias_db_query_close(query);
        db_close(db);
        return 1;
    }
    assert_double_equal(monitor_diode_temp_a_mean, 
        db_telemetry_results->monitor_diode_temp_a_mean);

    status = ias_db_query_double_value(query, &monitor_diode_temp_a_stdev, 17);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error obtaining "
            "monitor_diode_temp_a_stdev from lamp_telemetry table");
        ias_db_query_close(query);
        db_close(db);
        return 1;
    }
    assert_double_equal(monitor_diode_temp_a_stdev, 
        db_telemetry_results->monitor_diode_temp_a_stdev);

    status = ias_db_query_double_value(query, &monitor_diode_temp_b_mean, 18);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error obtaining monitor_diode_temp_b_mean"
            " from lamp_telemetry table");
        ias_db_query_close(query);
        db_close(db);
        return 1;
    }
    assert_double_equal(monitor_diode_temp_b_mean, 
        db_telemetry_results->monitor_diode_temp_b_mean);

    status = ias_db_query_double_value(query, &monitor_diode_temp_b_stdev, 19);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error obtaining "
            "monitor_diode_temp_b_stdev from lamp_telemetry table");
        ias_db_query_close(query);
        db_close(db);
        return 1;
    }
    assert_double_equal(monitor_diode_temp_b_stdev, 
        db_telemetry_results->monitor_diode_temp_b_stdev);

    status = ias_db_query_double_value(query, &fp_temp_7_mean, 20);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error obtaining fp_temp_7_mean "
            "from lamp_telemetry table");
        ias_db_query_close(query);
        db_close(db);
        return 1;
    }
    assert_double_equal(fp_temp_7_mean, 
        db_telemetry_results->fp_temp_7_mean);

    status = ias_db_query_double_value(query, &fp_temp_7_stdev, 21);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error obtaining fp_temp_7_stdev "
            "from lamp_telemetry table");
        ias_db_query_close(query);
        db_close(db);
        return 1;
    }
    assert_double_equal(fp_temp_7_stdev, 
        db_telemetry_results->fp_temp_7_stdev);

    status = ias_db_query_double_value(query, &fp_temp_14_mean, 22);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error obtaining fp_temp_14_mean "
            "from lamp_telemetry table");
        ias_db_query_close(query);
        db_close(db);
        return 1;
    }
    assert_double_equal(fp_temp_14_mean, 
        db_telemetry_results->fp_temp_14_mean);

    status = ias_db_query_double_value(query, &fp_temp_14_stdev, 23);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error obtaining fp_temp_14_stdev "
            "from lamp_telemetry table");
        ias_db_query_close(query);
        db_close(db);
        return 1;
    }
    assert_double_equal(fp_temp_14_stdev, 
        db_telemetry_results->fp_temp_14_stdev);

    /* Release query resources */
    ias_db_query_close(query);
    db_close(db);

    return status;
}

int math_compute_mean
(
    const float *sample,       /* I: pointer to the data */
    int nsamps,                /* I: number of samples */
    float *mean                /* O: mean of the data */
)
{
    int i;                               /* loop variable */
    float sum=0.0;             /* sum of the samples */

/*---------------------------------------------------------------------
    Mean calculation.  If number of samples is invalid return ERROR.
---------------------------------------------------------------------*/
    if (nsamps <= 0)
    {
        IAS_LOG_ERROR("Sample size is invalid.");
        *mean = 0.0;
        return(ERROR);
    }

    for (i = 0; i < nsamps; i++)
        sum += sample[i];

    *mean = sum / nsamps;

    return SUCCESS;
}

int math_compute_stdev
(
    const float *sample,   /* I: pointer to the data */
    int nsamps,            /* I: number of samples */
    float mean,            /* I: mean of the samples */
    float *std_dev         /* O: standard deviation of the data */
)
{
    int i;                  /* loop variable */
    float sum_sqr=0.0;      /* sum of the squared difference from mean */

    /* Standard deviation calculation.  If number of samples is invalid
       return ERROR. */
    if (nsamps <= 1)
    {
        IAS_LOG_ERROR("Sample size is invalid.");
        *std_dev = 0.0;
        return(ERROR);
    }
    
    for (i = 0; i < nsamps; i++)
        sum_sqr += sample[i] * sample[i];  /* sum of squares */
        
    sum_sqr -= nsamps * (mean * mean);     /* subtract mean squared for each
                                              sample */
    /* Use the absolute value to avoid round off errors */
    *std_dev = sqrt(fabs(sum_sqr) / (nsamps-1));

    return(SUCCESS);
}

