#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include "cgreen/cgreen.h"
#include "ias_satellite_attributes.h"
#include "ias_logging.h"
#include "ias_threadpool.h"
#include "../rps_local.h"
#include "../rps_algorithm.h"
#include "../rps_blob.h"

static char test_dir[PATH_MAX];
static char l0rname[] = "/ias_base/rps/driver/L0R/LO82220010052011265EDC00";
static char l1rname[] = "L1R_output_file";
static char omfname[] = "I82345.omf";
static char cpfname[] = "ias_lib/cpf/L8CPF20090101_20090331.01";
static char bpfnameoli[] 
                = "ias_lib/bpf/test_LO8BPF20140310103310_20140310103345.06";
static char l0r_file[PATH_MAX];


Ensure test_read_parameters()
{
    BLOB *blob;
    int  status;

    /* Some of the blob stuff needs to be set up for this test */
    blob = allocate_blob();
    if ( !blob )
    {
        assert_true_with_message( 0, "allocating blob" );
        return;
    }
    status = ias_sat_attr_initialize( IAS_L8 );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0,
            "ias_sat_attr_initialize returned an error" );
        return;
    }
    blob->total_bands = 30;
    sprintf( blob->cpf_filename, "%s/%s", test_dir, cpfname );
    sprintf( blob->bpf_filename_oli, "%s/%s", test_dir, bpfnameoli );
    strcpy( blob->bpf_filename_tirs, "" );

    strcpy( blob->parameter_file_name, "parameter.odl" );
    status = read_parameters( blob );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error reading input parameters" );
        free_the_blob();
        return;
    }
    assert_string_equal( blob->omf_file_name, omfname );
    assert_string_equal( blob->l0r_file_name, l0r_file );
    assert_string_equal( blob->l1r_file_name, l1rname );
    assert_string_equal( blob->work_order_id, "I82345" );
    assert_equal( blob->number_algorithms, 4 );
    if ( blob->number_algorithms != 4 )
    {
        assert_true_with_message( 0,
            "Error: read %d algorithms instead of 4", blob->number_algorithms );
        return;
    }
    assert_string_equal( blob->algorithm_wo_parameter_list[1],
        "saturated_pixel_characterization" );
    assert_string_equal( blob->algorithm_wo_parameter_list[2],
        "oli_bias_removal_DUMP" );
    assert_string_equal( blob->algorithm_wo_parameter_list[3],
        "oli_bias_determination" );

    free_the_blob();
}


Ensure test_read_parameters_duplicate_band_list()
{
    BLOB *blob;
    int  status;

    /* Some of the blob stuff needs to be set up for this test */
    blob = allocate_blob();
    if ( !blob )
    {
        assert_true_with_message( 0, "allocating blob" );
        return;
    }
    status = ias_sat_attr_initialize( IAS_L8 );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0,
            "ias_sat_attr_initialize returned an error" );
        return;
    }
    blob->total_bands = 30;

    strcpy( blob->parameter_file_name, "parameter_duplicate_bands.odl" );
    ias_log_set_output_level( IAS_LOG_LEVEL_DISABLE );
    status = read_parameters( blob );
    ias_log_set_output_level( IAS_LOG_LEVEL_WARN );
    assert_equal_with_message(status, ERROR,
        "Duplicate bands in band list, should have returned an ERROR");
    if (status != ERROR)
    {
        free_the_blob();
        return;
    }

    free_the_blob();
}


Ensure test_create_algorithm_pointer_list()
{
    BLOB *blob;
    int status;
    int (*api_algorithm)() = api_dump_imagery;
    int i;
    int count;

    /* Some of the blob stuff needs to be set up for this test */
    blob = allocate_blob();
    if ( !blob )
    {
        assert_true_with_message( 0, "allocating blob" );
        return;
    }
    status = ias_sat_attr_initialize( IAS_L8 );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0,
            "ias_sat_attr_initialize returned an error" );
        return;
    }
    blob->total_bands = 30;
    sprintf( blob->cpf_filename, "%s/%s", test_dir, cpfname );
    sprintf( blob->bpf_filename_oli, "%s/%s", test_dir, bpfnameoli );
    strcpy( blob->bpf_filename_tirs, "" );

    strcpy( blob->parameter_file_name, "create_algo_test.odl" );
    status = read_parameters( blob );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error reading input parameters" );
        free_the_blob();
        return;
    }
    assert_string_equal( blob->omf_file_name, omfname );
    assert_equal( blob->number_algorithms, 6 );
    if ( blob->number_algorithms != 6 )
    {
        assert_true_with_message( 0,
            "Error: read %d algorithms instead of 6", blob->number_algorithms );
        free_the_blob();
        return;
    }

    status = create_algorithm_pointer_list( blob );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0,
            "Error returned from create_algorithm_pointer_list" );
        free_the_blob();
        return;
    }
    assert_equal( blob->number_dump_algorithms, 2 );
    assert_equal( blob->number_algorithms, 7 );

    /* Make sure the WO algorithms are what is expected */
    assert_string_equal( blob->algorithm_wo_parameter_list[0],
        "convert_to_float" );
    assert_string_equal( blob->algorithm_wo_parameter_list[1],
        "gain_application_SKIP" );
    assert_string_equal( blob->algorithm_wo_parameter_list[2],
        "saturated_pixel_characterization" );
    assert_string_equal( blob->algorithm_wo_parameter_list[3],
        "oli_bias_removal" );
    assert_string_equal( blob->algorithm_wo_parameter_list[4],
        "oli_bias_determination" );
    assert_string_equal( blob->algorithm_wo_parameter_list[5],
        "oli_bias_removal" );

    /* Now ensure the algorithm strings are correct */
    /* This should match the parameter algorithm index 0 */
    assert_string_equal( blob->algorithm_string_list[0],
        "convert_to_float" );
    /* And this one shouldn't because it was skipped */
    assert_string_not_equal( blob->algorithm_string_list[1],
        "gain_application_SKIP" );
    /* the rest should match but since the gain_application was skipped one
       less index */
    assert_string_equal( blob->algorithm_string_list[1],
        "saturated_pixel_characterization" );
    assert_string_equal( blob->algorithm_string_list[2],
        "oli_bias_removal" );
    assert_string_equal( blob->algorithm_string_list[3],
        "oli_bias_determination" );
    assert_string_equal( blob->algorithm_string_list[4],
        "oli_bias_removal" );

    /* Make sure the dump algorithms are correct */
    assert_string_equal( blob->dump_algorithm_string_list[0],
        "oli_bias_removal" );
    assert_string_equal( blob->dump_algorithm_string_list[1],
        "oli_bias_determination" );

    /* Make sure the algorithm pointers are correct */
    i = 0;
    api_algorithm = api_convert_to_float;
    assert_equal( api_algorithm, blob->algorithm_pointer_list[i++] );
    api_algorithm = api_saturated_pixel_characterization;
    assert_equal( api_algorithm, blob->algorithm_pointer_list[i++] );
    api_algorithm = api_oli_bias_removal;
    assert_equal( api_algorithm, blob->algorithm_pointer_list[i++] );
    api_algorithm = api_dump_imagery;
    assert_equal( api_algorithm, blob->algorithm_pointer_list[i++] );
    api_algorithm = api_oli_bias_determination;
    assert_equal( api_algorithm, blob->algorithm_pointer_list[i++] );
    api_algorithm = api_dump_imagery;
    assert_equal( api_algorithm, blob->algorithm_pointer_list[i++] );
    api_algorithm = api_oli_bias_removal;
    assert_equal( api_algorithm, blob->algorithm_pointer_list[i++] );

    for ( count = 0; blob->pre_proc_pointer_list[count] != NULL; count++ );
    assert_equal( count, 6 );
    for ( count = 0; blob->post_proc_pointer_list[count] != NULL; count++ );
    assert_equal( count, 1 );

    free_the_blob();
}


Ensure test_memory_management()
{
    int i;
    int status;
    uint16_t *image;
    MEMORY_BUFFER buf;
    int num_consumers;
    double *rthebuf;
    BLOB *blob;
    char *unit_test_dir;
    const double *gain;
    const double *temperature;
    int *analog_low_sat;

    blob = allocate_blob();
    if ( !blob )
    {
        assert_true_with_message( 0, "allocating blob" );
        return;
    }


    unit_test_dir = getenv( "IAS_UNIT_TEST_DATA_DIR" );
    if ( !unit_test_dir )
    {
        assert_true_with_message( 0,
            "Environment variable IAS_UNIT_TEST_DATA_DIR not set" );
        return;
    }
    strcpy( blob->l0r_file_name, l0r_file );
    strcpy( blob->l1r_file_name, l1rname );
    strcpy( blob->cpf_filename, unit_test_dir );
    strcat( blob->cpf_filename, "/ias_base/rps/driver/partial.cpf" );
    strcpy( blob->bpf_filename_oli, unit_test_dir );
    strcat( blob->bpf_filename_oli,
        "/ias_lib/bpf/test_LO8BPF20140310103310_20140310103345.06" );
    strcpy( blob->bpf_filename_tirs, "" );

    /* Initialize the file I/O routines */
    status = ias_sat_attr_initialize( IAS_L8 );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0,
            "ias_sat_attr_initialize returned an error" );
        return;
    }
    blob->total_bands = 30;
    if ( finish_blob_initialization())
    {
        assert_true_with_message( 0, "initializing blob" );
        return;
    }
    blob->number_of_bands = 1;
    blob->list_of_bands[0] = 1;

    ias_log_set_output_level( IAS_LOG_LEVEL_DISABLE );
    status = initialize_io_routines( blob );
    ias_log_set_output_level( IAS_LOG_LEVEL_INFO );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0,
            "initialize_io_routines returned an error" );
        free_the_blob();
        return;
    }

    /* Initialize management for three L0R consumers.  This algorithm
       doesn't need any memory management buffers that require the list of
       bands to be initialized -- number of bands set to 0 for that reason. */
    blob->number_of_bands = 0;
    blob->band_fields_initialized = 1;
    status = initialize_memory_management_buffers();
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error initializing memory management");
        free_the_blob();
        return;
    }

    for ( i = 0; i < 3; i++ )
    {
        status = increment_consumer_count( IMAGE_L0R );
        if ( status != SUCCESS )
        {
            assert_true_with_message( 0,
                "Error returned from increment_consumer_count" );
            cleanup_io_routines( blob );
            cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
            free_the_blob();
            return;
        }
    }
    get_mem_mgmt_struct_dont_use_me( IMAGE_L0R, 3, 8, &num_consumers, &buf );
    assert_equal( buf.user_count, 0 );
    assert_equal( buf.data_pointer, NULL );
    assert_equal( num_consumers, 3 );

    /* Attempt to retrieve a buffer.  It is an error to call for a buffer
       at this time with get_memory_buffer */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    image = (uint16_t *)get_memory_buffer( IMAGE_L0R, CONSUMER, 3, 8 );
    ias_log_set_output_level(IAS_LOG_LEVEL_WARN);
    if ( image )
    {
        assert_true_with_message( 0,
            "get_memory_buffer didn't return a NULL pointer as it should");
    }
    /* and not an error with the pre-process call */
    image = (uint16_t *)preprocess_get_memory_buffer( IMAGE_L0R, 3, 8 );
    if ( ! image )
    {
        assert_true_with_message( 0,
            "get_memory_buffer returned a NULL pointer" );
        cleanup_io_routines( blob );
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* The image buffer should contain the value 4 */
    for ( i = 0, status = 0; i < 16*494; i++ )
    {
        if ( image[i] != 4 )
            status++;
    }
    assert_equal( status, 0 );

    /* There should be three consumers and the buffers should have been
       allocated */
    for ( i = 3; i > 0; i-- )
    {
        get_mem_mgmt_struct_dont_use_me( IMAGE_L0R, 3, 8,
            &num_consumers, &buf );
        assert_not_equal( buf.data_pointer, NULL );
        assert_equal( num_consumers, 3 );
    }

    /* Set up a non-image buffer */
    for ( i = 0; i < 10; i++ )
    {
        status = increment_consumer_count( MM_NON_PAN_PERFRAME_BIAS );
        if ( status != SUCCESS )
        {
            assert_true_with_message( 0,
                "Error returned from increment_consumer_count" );
            cleanup_io_routines( blob );
            cleanup_memory_management_buffers(
                DISABLE_WARNING_DISPLAY);
            free_the_blob();
            return;
        }
    }
    get_mem_mgmt_struct_dont_use_me( MM_NON_PAN_PERFRAME_BIAS, 8, 11,
        &num_consumers, &buf );
    assert_equal( buf.user_count, 0 );
    assert_equal( buf.data_pointer, NULL );
    assert_equal( num_consumers, 10 );
    /* Request the pointer */
    rthebuf = (double *)preprocess_get_memory_buffer( MM_NON_PAN_PERFRAME_BIAS,
        8, 13 );
    if ( !rthebuf )
    {
        assert_true_with_message( 0,
            "get_memory_buffer returned a NULL pointer" );
        cleanup_io_routines( blob );
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }
    /* 8, 11 should still be NULL */
    get_mem_mgmt_struct_dont_use_me( MM_NON_PAN_PERFRAME_BIAS, 8, 11,
        &num_consumers, &buf );
    assert_equal( buf.user_count, 0 );
    assert_equal( buf.data_pointer, NULL );
    assert_equal( num_consumers, 10 );
    /* However 8, 13 should not */
    get_mem_mgmt_struct_dont_use_me( MM_NON_PAN_PERFRAME_BIAS, 8, 13,
        &num_consumers, &buf );
    assert_not_equal( buf.data_pointer, NULL );
    assert_equal( num_consumers, 10 );

    /* Set up a constant buffer */
    status = increment_consumer_count( MM_ABSOLUTE_GAINS );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0,
            "Error returned from increment_consumer_count for absolute"
            "gains" );
        cleanup_io_routines( blob );
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    status = increment_consumer_count( MM_RELATIVE_GAINS );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0,
            "Error returned from increment_consumer_count for relative "
            "gains" );
        cleanup_io_routines( blob );
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    status = increment_consumer_count( MM_ANALOG_LOW_SAT );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error from increment_consumer_count "
            "for MM_ANALOG_LOW_SAT");
        cleanup_io_routines( blob );
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    status = increment_consumer_count( MM_TEMP_SENSITIVITY_REFERENCE_TEMP );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0,
            "Error from increment_consumer_count for temperature sensitivity");
        cleanup_io_routines( blob );
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    status = increment_consumer_count( MM_TEMP_SENSITIVITY_COEFFICIENT );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0,
            "Error from increment_consumer_count for temperature sensitivity");
        cleanup_io_routines( blob );
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* Set the consumer counts */
    status = set_consumer_counts();
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error setting the consumer counts");
        cleanup_io_routines( blob );
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* It's an error to call the pre-process get memory buffer after counts
       have been set */
    rthebuf = (double *)preprocess_get_memory_buffer( MM_NON_PAN_PERFRAME_BIAS,
        8, 13 );
    if ( rthebuf )
    {
        assert_true_with_message( 0, "preprocess_get_memory_buffer didn't "
            "return a NULL pointer as it should");
    }

    /* It's an error to call increment after counts have been set */
    status = increment_consumer_count( MM_TEMP_SENSITIVITY_COEFFICIENT );
    if ( status != ERROR )
    {
        assert_true_with_message( 0, "increment_consumer_count returned "
            "SUCCESS when it shouldn't have");
    }

    /* It's an error to call set_consumer_counts more than once */
    status = set_consumer_counts(); 
    if ( status != ERROR )
    {
        assert_true_with_message( 0, "Second call to set_consumer_counts "
            "didn't return an ERROR");
    }
    ias_log_set_output_level(IAS_LOG_LEVEL_WARN);

    /* Consume the level 0R data buffer */
    for ( i = 3; i > 0; i-- )
    {
        get_mem_mgmt_struct_dont_use_me( IMAGE_L0R, 3, 8,
            &num_consumers, &buf );
        assert_equal( buf.user_count, i );
        assert_not_equal( buf.data_pointer, NULL );
        assert_equal( num_consumers, 3 );
        status = decrement_consumer_count( IMAGE_L0R, 3, 8 );
        assert_equal_with_message( status, SUCCESS,
            "Error returned from decrement_consumer_count");
    }
    get_mem_mgmt_struct_dont_use_me( IMAGE_L0R, 3, 8, &num_consumers, &buf );
    assert_equal( buf.user_count, 0 );
    assert_equal( buf.data_pointer, NULL );
    assert_equal( num_consumers, 3 );

    /* Test the consumer counts have been set for this buffer and consume 
       the buffer */
    get_mem_mgmt_struct_dont_use_me( MM_NON_PAN_PERFRAME_BIAS, 8, 13,
        &num_consumers, &buf );
    assert_equal( buf.user_count, 10 );
    for ( i = 10; i > 0; i-- )
    {
        get_mem_mgmt_struct_dont_use_me( MM_NON_PAN_PERFRAME_BIAS, 8, 13,
            &num_consumers, &buf );
        assert_equal( buf.user_count, i );
        assert_not_equal( buf.data_pointer, NULL );
        assert_equal( num_consumers, 10 );
        status = decrement_consumer_count( MM_NON_PAN_PERFRAME_BIAS, 8, 13 );
        assert_equal_with_message( status, SUCCESS,
            "Error returned from decrement_consumer_count");
    }
    /* The data has now been consumed and the pointer should be NULL and
       no consumers.  The num_consumers shoudn't have changed. */
    get_mem_mgmt_struct_dont_use_me( MM_NON_PAN_PERFRAME_BIAS, 8, 13,
        &num_consumers, &buf );
    assert_equal( buf.user_count, 0 );
    assert_equal( buf.data_pointer, NULL );
    assert_equal( num_consumers, 10 );

    /* The constant buffers don't have a pre-process get buffer call so 
       check these buffers now */
    gain = get_const_memory_buffer( MM_ABSOLUTE_GAINS, 0, 9 );
    if ( !gain )
    {
        assert_true_with_message( 0,
            "Error retrieving the absolute gain for 1, 10");
        cleanup_io_routines( blob );
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }
    assert_double_equal( gain[0], 9.286 );
    gain = get_const_memory_buffer( MM_RELATIVE_GAINS, 0, 0 );
    if ( !gain )
    {
        assert_true_with_message( 0,
            "Error retrieving the relative gain for 1, 1");
        cleanup_io_routines( blob );
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }
    assert_double_equal( gain[0], 0.111000 );
    assert_double_equal( gain[1], 0.222000 );
    assert_double_equal( gain[2], 0.333000 );
    assert_double_equal( gain[3], 0.444000 );
    assert_double_equal( gain[4], 0.0 );

    analog_low_sat = (int *)get_const_memory_buffer( MM_ANALOG_LOW_SAT, 0, 3);
    if ( !analog_low_sat )
    {
        assert_true_with_message( 0, "Error getting MM_ANALOG_LOW_SAT buffer");
        cleanup_io_routines( blob );
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }
    assert_equal( analog_low_sat[0], 55 );
    assert_equal( analog_low_sat[1], 2 );
    assert_equal( analog_low_sat[493], 44 );

    temperature = get_const_memory_buffer( MM_TEMP_SENSITIVITY_REFERENCE_TEMP,
        0, 4 );
    if ( !temperature )
    {
        assert_true_with_message( 0,
            "Error retrieving the temperature sensitivity reference temp");
        cleanup_io_routines( blob );
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }
    assert_double_equal( temperature[0], 210.0 );

    temperature = get_const_memory_buffer( MM_TEMP_SENSITIVITY_COEFFICIENT,
        7, 12 );
    if ( !temperature )
    {
        assert_true_with_message( 0,
            "Error retrieving the temperature sensitivity reference temp");
        cleanup_io_routines( blob );
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }
    assert_double_equal( temperature[490], 0.005000 );
    assert_double_equal( temperature[491], 0.005000 );
    assert_double_equal( temperature[492], 0.005000 );
    assert_double_equal( temperature[493], 0.005000 );

    cleanup_io_routines( blob );
    cleanup_memory_management_buffers(ENABLE_WARNING_DISPLAY);
    //cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
    free_the_blob();
}


Ensure test_miscellany()
{
    char *rps_level;
    int status;

    /* Verify get/set rps level return the correct values if rps_level has 
       not been initialized. Disable logging so errors aren't printed */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);  

    status = set_rps_level( 1, 2, 'T' );
    assert_equal( status, ERROR );

    rps_level = get_rps_level( 1, 2 );
    assert_equal( rps_level, NULL );
    free( rps_level );
    
    /* Restore the warning level */
    ias_log_set_output_level(IAS_LOG_LEVEL_WARN);

    status = initialize_rps_level( 5, 10 );

    /* Verify the level was initialized to an empty string */    
    rps_level = get_rps_level( 4, 9 );
    assert_string_equal( rps_level, "" );
    free( rps_level );

    /* Test set_rps_level */
    status = set_rps_level( 4, 9, 'T' );
    rps_level = get_rps_level( 4, 9 );
    assert_string_equal( rps_level, "T" );
    free( rps_level );

    /* Verify set_rps_level updates only the provided band/sca */
    rps_level = get_rps_level( 3, 3 );
    assert_string_equal( rps_level, "" );
    free( rps_level );

    /* Test set_rps_level appends additional process Ids */
    status = set_rps_level( 4, 9, 'E' );
    rps_level = get_rps_level( 4, 9 );
    assert_string_equal( rps_level, "TE" );
    free( rps_level );

    /* Test the rps_level_is_l0r function. Use an unused band/SCA for the 
       tests */
    rps_level = get_rps_level( 3, 1 );
    assert_equal( rps_level_is_l0r(rps_level), TRUE );
    free( rps_level );

    status = set_rps_level( 3, 1, IAS_RPS_LEVEL_CONVERT_FLOAT_NOSHIFT );
    rps_level = get_rps_level( 3, 1 );
    assert_equal( rps_level_is_l0r(rps_level), TRUE );
    free( rps_level );

    status = set_rps_level( 3, 2, IAS_RPS_LEVEL_CONVERT_FLOAT_CORRECT );
    rps_level = get_rps_level( 3, 2 );
    assert_equal( rps_level_is_l0r(rps_level), TRUE );
    free( rps_level );

    status = set_rps_level( 3, 2, IAS_RPS_LEVEL_SKIP );
    rps_level = get_rps_level( 3, 2 );
    assert_equal( rps_level_is_l0r(rps_level), TRUE );
    free( rps_level );

    status = set_rps_level( 3, 2, IAS_RPS_LEVEL_BIAS_REMOVAL );
    rps_level = get_rps_level( 3, 2 );
    assert_equal( rps_level_is_l0r(rps_level), FALSE );
    free( rps_level );

    /* Theses are all errors so disable the error logging so they doesn't
       show up as ERROR in the output */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    rps_level = get_rps_level( -1, 9 );
    assert_equal( rps_level, NULL );
    free( rps_level );
    rps_level = get_rps_level( 5, 9 );
    assert_equal( rps_level, NULL );
    free( rps_level );
    rps_level = get_rps_level( 5, 10 );
    assert_equal( rps_level, NULL );
    free( rps_level );

    /* Restore the warning level */
    ias_log_set_output_level(IAS_LOG_LEVEL_WARN);

    cleanup_rps_level();
}


int main( int argc, char *argv[] )
{
    int status;
    char *unit_test_dir;
    FILE *omf;
    TestSuite *suite = create_test_suite();

    add_test( suite, test_read_parameters );
    add_test( suite, test_read_parameters_duplicate_band_list );
    add_test( suite, test_create_algorithm_pointer_list );
    add_test( suite, test_memory_management );
    add_test( suite, test_miscellany );
    
    /* Create an omf file that points to the correct unit test dir */
    unit_test_dir = getenv( "IAS_UNIT_TEST_DATA_DIR" );
    if ( !unit_test_dir )
    {
        printf("Environment variable IAS_UNIT_TEST_DATA_DIR is not set" );
        return 1;
    }
    strcpy( test_dir, unit_test_dir );
    sprintf( l0r_file, "%s/%s", unit_test_dir, l0rname );
    omf = fopen( omfname, "w" );
    if ( !omf )
    {
        printf("Unable to create OMF file: %s\n", omfname );
        return 1;
    }
    fprintf( omf, "OBJECT = OMF\n" );
    fprintf( omf, "L0R_FILENAME = \"%s\"\nL1R_FILENAME = \"L1R_output_file\"",
        l0r_file );
    fprintf( omf, "\nSENSOR = \"L8_OLITIRS\"");
    fprintf( omf, "\nCAL_PARM_FILENAME = \"%s/ias_lib/cpf/"
        "L8CPF20090101_20090331.01\"", unit_test_dir );
    fprintf( omf, "\nBIAS_PARM_FILENAME_OLI = \"%s/ias_lib/bpf/"
        "test_LO8BPF20140310103310_20140310103345.06\"", unit_test_dir );
    fprintf( omf, "\nBIAS_PARM_FILENAME_TIRS = \"\"");
    fprintf( omf, "\nRLUT_FILENAME = \"%s/ias_lib/rlut/test_linearization_rlut"
        ".h5\"\n", unit_test_dir );
    fprintf( omf, "END_OBJECT = OMF\n" );
    fprintf( omf, "END\n" );
    fclose( omf );

    if ( argc > 1 )
        status = run_single_test( suite, argv[1], create_text_reporter() );
    else
        status = run_test_suite( suite, create_text_reporter() );

    remove( omfname );
    remove( l1rname );
    return status;
}
