/***************************************************************************
 NAME:      dropped_frame_test

 PURPOSE:   CGreen unit testing for the OLI/TIRS dropped frame
            characterization algorithm called from the RPS Driver

 RETURNS:   CGreen unit test status

 ROUTINES:  test_dropped_frame_char_through_api
            main

 NOTES:     The majority of code in this test relates to the correct
            processing of the L0R frame header data.  As indicated below,
            VRP bands and blind bands associated with a normal OLI/TIRS
            band being processed through the RPS driver will also
            be characterized
****************************************************************************/
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "cgreen.h"
#include "ias_const.h"
#include "ias_l0r_constants.h"
#include "ias_l0r_structures.h"
#include "ias_logging.h"
#include "ias_pixel_mask.h"
#include "ias_radiometry.h"
#include "ias_satellite_attributes.h"
#include "../rps_local.h"
#include "../rps_blob.h"


#define DROPPED_FRAME_LINE 4
#define GOOD_LINE         96


static char *test_data_path = NULL;
static char l0r_data_dir[] = "ias_base/rps/driver/dropped_frame_char";
static char l0r_basename[] = "LO80390360362003148EDC00";
static char cpf_data_dir[] = "ias_lib/cpf";
static char cpf_basename[] = "L8CPF20090101_20090331.01";
static char bpf_data_dir[] = "ias_lib/bpf";
static char bpf_basename_oli[]  = "test_LO8BPF20140310103310_20140310103345.06";
static char rlut_data_dir[] = "ias_lib/rlut";
static char rlut_basename[] = "test_linearization_rlut.h5";
static BLOB *blob = NULL;


Ensure test_oli_dropped_frame_char_through_api()
{
    const RPS_BAND_ATTRIBUTES *band_atts = NULL;
    IAS_PIXEL_MASK_ITERATOR *it = NULL;
    IAS_PIXEL_MASK_SPAN span;
    const IAS_L0R_OLI_FRAME_HEADER *oli_fh = NULL;
    int span_start_index_oli[5] = {1, 4, 7, 20, 6044};
                                        /* Indices in frame header array
                                           where a dropped frame span
                                           begins.  The last two starting
                                           indices are for OLI and TIRS
                                           frame header sets, respectively */
    int dropped_frame_error_count = 0;  /* Dropped frame identification
                                           error counter */
    int det_index;
    int normal_band_index;
    int blind_band_index;
    int vrp_band_index;
    int sca_index;
    int status;
    int ii;                             /* Local band loop counter */
    int jj;                             /* Local SCA loop counter */
    int process_blind_band_flag = 0;


    printf("Testing OLI dropped frame characterization run from RPS "
        "driver...\n");

    /* Initialize the logging library in case any library call issues
       log messages */
    ias_log_initialize("api_dropped_frame_test_oli");

    /* Allocate BLOB data block */
    blob = allocate_blob();
    if (blob == NULL)
    {
        assert_true_with_message(FALSE, "Error allocating BLOB base data "
            "block");
        return;
    }

    /* Initialize the BLOB */
    blob->max_number_scas = 14;
    blob->total_bands = 30;
    blob->generate_report = 0;
    blob->store_in_database = 0;

    sprintf(blob->l0r_file_name, "%s/%s/%s", test_data_path, l0r_data_dir,
        l0r_basename);
    sprintf(blob->l1r_file_name, "%s", "L1R_output_file");
    sprintf(blob->cpf_filename, "%s/%s/%s", test_data_path, cpf_data_dir,
        cpf_basename);
    sprintf(blob->bpf_filename_oli, "%s/%s/%s", test_data_path, bpf_data_dir,
        bpf_basename_oli);
    strcpy(blob->bpf_filename_tirs, "");
    sprintf(blob->rlut_filename, "%s/%s/%s", test_data_path, rlut_data_dir,
        rlut_basename);

    status = finish_blob_initialization();
    assert_equal_with_message(status, SUCCESS, "Error completing test blob "
        "initialization");
    if (status != SUCCESS)
    {
        free_the_blob();
        return;
    }
    blob->number_of_bands = 2;
    blob->list_of_bands[0] = 6;           /* OLI band 6 */
    blob->list_of_bands[1] = 10;          /* TIRS band 10 */
    blob->list_of_band_indices[0] =
        ias_sat_attr_convert_band_number_to_index( blob->list_of_bands[0] );
    blob->list_of_band_indices[1] =
        ias_sat_attr_convert_band_number_to_index( blob->list_of_bands[1] );
    if ( blob->list_of_band_indices[0] == ERROR ||
         blob->list_of_band_indices[1] == ERROR )
    {
        assert_true_with_message( 0, "Converting band number to index");
        free_the_blob();
        return;
    }

    /* Initialize the L0R image data */
    ias_log_set_output_level( IAS_LOG_LEVEL_DISABLE );
    status = initialize_io_routines(blob);
    ias_log_set_output_level( IAS_LOG_LEVEL_INFO );
    assert_equal_with_message(status, SUCCESS, "Error initializing image "
        "data");
    if (status != SUCCESS)
    {
        free_the_blob();
        return;
    }

    /* Create the blob pixel mask array */
    status = create_blob_pixel_mask();
    assert_equal_with_message(status, SUCCESS, "Error creating blob pixel "
        "mask array");
    if (status != SUCCESS)
    {
        cleanup_io_routines(blob);
        free_the_blob();
        return;
    }

    /* Initialize memory management */
    status = initialize_memory_management_buffers();
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error initializing memory management");
        cleanup_io_routines( blob );
        free_the_blob();
        return;
    }

    status = dropped_frame_characterization_preprocess(blob);
    assert_equal_with_message(status, SUCCESS, "Error in dropped_frame_char_"
        "preprocess");
    if (status != SUCCESS)
    {
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Set the band index and SCA index for OLI band 6, SCA 1 */
    normal_band_index = blob->list_of_band_indices[0];
    sca_index = 0;

    /* Get a copy of the memory buffer for the frame header and check the
       first couple of values.  As currently constructed, dropped frame
       characterization will also be performed on the VRP and blind bands
       for an OLI/TIRS normal band being processed if the band number exists
       and--for the blind band--the
       process_blind_band_flag in the attributes is set.  Consequently,
       the increment_consumer_count routine assigns the frame header
       buffer pointer in all SCAs to be processed in those bands in
       addition to the normal band  */
    status = increment_consumer_count(MM_L0R_OLI_FRAME_HEADER);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error incrementing OLI frame "
            "header consumer count");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }
    /* It is important that the consumer counts get set *after* all the
       increments */
    status = set_consumer_counts();
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error setting consumer counts");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Run the test for OLI band 6 SCA 1 */
    band_atts = get_band_attributes(normal_band_index);

    oli_fh = (const IAS_L0R_OLI_FRAME_HEADER *)get_const_memory_buffer(
        MM_L0R_OLI_FRAME_HEADER, normal_band_index, sca_index);
    assert_not_equal_with_message(oli_fh, NULL, "Error obtaining OLI "
        "L0R frame header data buffer");
    if (oli_fh == NULL)
    {
        for (ii = 0; ii < blob->number_of_bands; ii++)
        {
            normal_band_index = blob->list_of_band_indices[ii];
            band_atts = get_band_attributes(normal_band_index);
            vrp_band_index = band_atts->vrp_band_index;
            blind_band_index = band_atts->blind_band_index;
            process_blind_band_flag = band_atts->process_blind_band_flag;
            if (band_atts->spectral_type != IAS_SPECTRAL_THERMAL)
            {
                for (jj = 0; jj < band_atts->number_scas; jj++)
                {
                    status = decrement_consumer_count(MM_L0R_OLI_FRAME_HEADER,
                            normal_band_index, jj);
                    assert_equal_with_message( status, SUCCESS,
                            "Error returned from decrement consumer count");
                }
                if (vrp_band_index >= 0)
                {
                    for (jj = 0; jj < band_atts->number_scas; jj++)
                    {
                        status = decrement_consumer_count(
                                MM_L0R_OLI_FRAME_HEADER,
                                vrp_band_index, jj);
                        assert_equal_with_message( status, SUCCESS,
                                "Error returned from decrement consumer count");
                    }
                }
                if ((blind_band_index >= 0) && process_blind_band_flag)
                {
                    for (jj = 0; jj < band_atts->number_scas; jj++)
                    {
                        status = decrement_consumer_count(
                                MM_L0R_OLI_FRAME_HEADER,
                                blind_band_index, jj);
                        assert_equal_with_message( status, SUCCESS,
                                "Error returned from decrement consumer count");
                    }
                }
            }
            else
            {
                for (jj = 0; jj < band_atts->number_scas; jj++)
                {
                    status = decrement_consumer_count(MM_L0R_TIRS_FRAME_HEADER,
                            normal_band_index, jj);
                    assert_equal_with_message( status, SUCCESS,
                            "Error returned from decrement consumer count");
                }
                if ((blind_band_index >= 0) && process_blind_band_flag)
                {
                    for (jj = 0; jj < band_atts->number_scas; jj++)
                    {
                        status = decrement_consumer_count(
                                MM_L0R_TIRS_FRAME_HEADER,
                                blind_band_index, jj);
                        assert_equal_with_message( status, SUCCESS,
                                "Error returned from decrement consumer count");
                    }
                }
            }
        }
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    assert_equal(oli_fh[0].frame_status, GOOD_LINE);
    assert_equal(oli_fh[1].frame_status, DROPPED_FRAME_LINE);
    assert_equal(oli_fh[2].frame_status, GOOD_LINE);

    /* Call the API function */
    status = api_dropped_frame_characterization(blob, normal_band_index,
        sca_index, 0);
    assert_equal_with_message(status, SUCCESS, "API run of dropped frame "
        "characterization failed");
    if (status != SUCCESS)
    {
        for (ii = 0; ii < blob->number_of_bands; ii++)
        {
            normal_band_index = blob->list_of_band_indices[ii];
            band_atts = get_band_attributes(normal_band_index);
            vrp_band_index = band_atts->vrp_band_index;
            blind_band_index = band_atts->blind_band_index;
            process_blind_band_flag = band_atts->process_blind_band_flag;
            if (band_atts->spectral_type != IAS_SPECTRAL_THERMAL)
            {
                for (jj = 0; jj < band_atts->number_scas; jj++)
                {
                    status = decrement_consumer_count(MM_L0R_OLI_FRAME_HEADER,
                            normal_band_index, jj);
                    assert_equal_with_message( status, SUCCESS,
                            "Error returned from decrement consumer count");
                }
                if (vrp_band_index >= 0)
                {
                    for (jj = 0; jj < band_atts->number_scas; jj++)
                    {
                        status = decrement_consumer_count(
                                MM_L0R_OLI_FRAME_HEADER,
                                vrp_band_index, jj);
                        assert_equal_with_message( status, SUCCESS,
                                "Error returned from decrement consumer count");
                    }
                }
                if ((blind_band_index >= 0) && process_blind_band_flag)
                {
                    for (jj = 0; jj < band_atts->number_scas; jj++)
                    {
                        status = decrement_consumer_count(
                                MM_L0R_OLI_FRAME_HEADER,
                                blind_band_index, jj);
                        assert_equal_with_message( status, SUCCESS,
                                "Error returned from decrement consumer count");
                    }
                }
            }
            else
            {
                for (jj = 0; jj < band_atts->number_scas; jj++)
                {
                    status = decrement_consumer_count(MM_L0R_TIRS_FRAME_HEADER,
                            normal_band_index, jj);
                    assert_equal_with_message( status, SUCCESS,
                            "Error returned from decrement consumer count");
                }
                if ((blind_band_index >= 0) && process_blind_band_flag)
                {
                    for (jj = 0; jj < band_atts->number_scas; jj++)
                    {
                        status = decrement_consumer_count(
                                MM_L0R_TIRS_FRAME_HEADER,
                                blind_band_index, jj);
                        assert_equal_with_message( status, SUCCESS,
                                "Error returned from decrement consumer count");
                    }
                }
            }
        }
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Verify the blob pixel mask in SCA 1 has the correct dropped frame
       spans marked for all detectors */
    it = ias_pm_get_iterator( blob->pixel_mask[normal_band_index][0],
        IAS_PM_EXCLUDE, 0x00);
    if (it == NULL)
    {
        for (ii = 0; ii < blob->number_of_bands; ii++)
        {
            normal_band_index = blob->list_of_band_indices[ii];
            band_atts = get_band_attributes(normal_band_index);
            vrp_band_index = band_atts->vrp_band_index;
            blind_band_index = band_atts->blind_band_index;
            process_blind_band_flag = band_atts->process_blind_band_flag;
            if (band_atts->spectral_type != IAS_SPECTRAL_THERMAL)
            {
                for (jj = 0; jj < band_atts->number_scas; jj++)
                {
                    status = decrement_consumer_count(MM_L0R_OLI_FRAME_HEADER,
                            normal_band_index, jj);
                    assert_equal_with_message( status, SUCCESS,
                            "Error returned from decrement consumer count");
                }
                if (vrp_band_index >= 0)
                {
                    for (jj = 0; jj < band_atts->number_scas; jj++)
                    {
                        status = decrement_consumer_count(
                                MM_L0R_OLI_FRAME_HEADER,
                                vrp_band_index, jj);
                        assert_equal_with_message( status, SUCCESS,
                                "Error returned from decrement consumer count");
                    }
                }
                if ((blind_band_index >= 0) && process_blind_band_flag)
                {
                    for (jj = 0; jj < band_atts->number_scas; jj++)
                    {
                        status = decrement_consumer_count(
                                MM_L0R_OLI_FRAME_HEADER,
                                blind_band_index, jj);
                        assert_equal_with_message( status, SUCCESS,
                                "Error returned from decrement consumer count");
                    }
                }
            }
            else
            {
                for (jj = 0; jj < band_atts->number_scas; jj++)
                {
                    status = decrement_consumer_count(MM_L0R_TIRS_FRAME_HEADER,
                            normal_band_index, jj);
                    assert_equal_with_message( status, SUCCESS,
                            "Error returned from decrement consumer count");
                }
                if ((blind_band_index >= 0) && process_blind_band_flag)
                {
                    for (jj = 0; jj < band_atts->number_scas; jj++)
                    {
                        status = decrement_consumer_count(
                                MM_L0R_TIRS_FRAME_HEADER,
                                blind_band_index, jj);
                        assert_equal_with_message( status, SUCCESS,
                                "Error returned from decrement consumer count");
                    }
                }
            }
        }
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }
    for (det_index = 0; det_index < band_atts->number_detectors;
         det_index++)
    {
        /* Check the first span--1 frame, starting at frame 2 */
        status = ias_pm_get_next_span(it, &span);
        if (status == 1)
        {
            if ((span.starting_pixel_index != span_start_index_oli[0])
                || (span.length_of_span != 1)
                || (span.detector_index != det_index)
                || (span.pixel_mask != IAS_RPS_DROPLINE))
            {
                dropped_frame_error_count++;
            }
        }

        /* Check the second span--1 frame, starting at frame 5 */
        status = ias_pm_get_next_span(it, &span);
        if (status == 1)
        {
            if ((span.starting_pixel_index != span_start_index_oli[1])
                || (span.length_of_span != 1)
                || (span.detector_index != det_index)
                || (span.pixel_mask != IAS_RPS_DROPLINE))
            {
                dropped_frame_error_count++;
            }
        }

        /* Check the third span--5 frames, starting at frame 8 */
        status = ias_pm_get_next_span(it, &span);
        if (status == 1)
        {
            if ((span.starting_pixel_index != span_start_index_oli[2])
                || (span.length_of_span != 5)
                || (span.detector_index != det_index)
                || (span.pixel_mask != IAS_RPS_DROPLINE))
            {
                dropped_frame_error_count++;
            }
        }

        /* Check the fourth span-- 8 frames, starting at frame 21 */
        status = ias_pm_get_next_span(it, &span);
        if (status == 1)
        {
            if ((span.starting_pixel_index != span_start_index_oli[3])
                || (span.length_of_span != 8)
                || (span.detector_index != det_index)
                || (span.pixel_mask != IAS_RPS_DROPLINE))
            {
                dropped_frame_error_count++;
            }
        }

        /* Check the final span--6 frames, starting at frame 6044*/
        status = ias_pm_get_next_span(it, &span);
        if (status == 1)
        {
            if ((span.starting_pixel_index != span_start_index_oli[4])
                || (span.length_of_span != 6)
                || (span.detector_index != det_index)
                || (span.pixel_mask != IAS_RPS_DROPLINE))
            {
                dropped_frame_error_count++;
            }
        }
    }
    assert_equal_with_message(dropped_frame_error_count, 0, "%d errors "
        "flagging dropped frames in pixel mask recorded",
        dropped_frame_error_count);

    /* Cleanup */
    for (ii = 0; ii < blob->number_of_bands; ii++)
    {
        normal_band_index = blob->list_of_bands[ii];
        band_atts = get_band_attributes(normal_band_index);
        vrp_band_index = band_atts->vrp_band_index;
        blind_band_index = band_atts->blind_band_index;
        process_blind_band_flag = band_atts->process_blind_band_flag;
        if (band_atts->spectral_type != IAS_SPECTRAL_THERMAL)
        {
            for (jj = 0; jj < band_atts->number_scas; jj++)
            {
                status = decrement_consumer_count(MM_L0R_OLI_FRAME_HEADER,
                        normal_band_index, jj);
                assert_equal_with_message( status, SUCCESS,
                        "Error returned from decrement consumer count");
            }
            if (vrp_band_index > 0)
            {
                for (jj = 0; jj < band_atts->number_scas; jj++)
                {
                    status = decrement_consumer_count(MM_L0R_OLI_FRAME_HEADER,
                            vrp_band_index, jj);
                    assert_equal_with_message( status, SUCCESS,
                            "Error returned from decrement consumer count");
                }
            }
            if ((blind_band_index >= 0) && process_blind_band_flag)
            {
                for (jj = 0; jj < band_atts->number_scas; jj++)
                {
                    status = decrement_consumer_count(MM_L0R_OLI_FRAME_HEADER,
                            blind_band_index, jj);
                    assert_equal_with_message( status, SUCCESS,
                            "Error returned from decrement consumer count");
                }
            }
        }
        else
        {
            for (jj = 0; jj < band_atts->number_scas; jj++)
            {
                status = decrement_consumer_count(MM_L0R_TIRS_FRAME_HEADER,
                        normal_band_index, jj);
                assert_equal_with_message( status, SUCCESS,
                        "Error returned from decrement consumer count");
            }
            if ((blind_band_index >= 0) && process_blind_band_flag)
            {
                for (jj = 0; jj < band_atts->number_scas; jj++)
                {
                    status = decrement_consumer_count(MM_L0R_TIRS_FRAME_HEADER,
                            blind_band_index, jj);
                    assert_equal_with_message( status, SUCCESS,
                            "Error returned from decrement consumer count");
                }
            }
        }
    }

    cleanup_io_routines(blob);
    cleanup_memory_management_buffers(ENABLE_WARNING_DISPLAY);
    ias_pm_destroy_iterator(it);
    free_the_blob();

}   /* END test_oli_dropped_frame_char_through_api */



Ensure test_tirs_dropped_frame_char_through_api()
{
    const RPS_BAND_ATTRIBUTES *band_atts = NULL;
    IAS_PIXEL_MASK_ITERATOR *it = NULL;
    IAS_PIXEL_MASK_SPAN span;
    const IAS_L0R_TIRS_FRAME_HEADER *tirs_fh = NULL;
    int span_start_index_tirs[5] = {1, 4, 7, 21, 1893};
                                        /* Indices in frame header array
                                           where a dropped frame span
                                           begins.  The last two starting
                                           indices are for OLI and TIRS
                                           frame header sets, respectively */
    int dropped_frame_error_count = 0;  /* Dropped frame identification
                                           error counter */
    int det_index;
    int normal_band_index;
    int blind_band_index;
    int vrp_band_index;
    int sca_index;
    int status;
    int ii;                             /* Local band loop counter */
    int jj;                             /* Local SCA loop counter */
    int process_blind_band_flag;


    printf("Testing TIRS dropped frame characterization run from RPS "
        "driver...\n");

    /* Initialize the logging library in case any library call issues
       log messages */
    ias_log_initialize("api_dropped_frame_test_tirs");

    /* Allocate BLOB data block */
    blob = allocate_blob();
    if (blob == NULL)
    {
        assert_true_with_message(FALSE, "Error allocating BLOB base data "
            "block");
        return;
    }

    /* Initialize the BLOB */
    blob->max_number_scas = 14;
    blob->total_bands = 30;
    blob->generate_report = 0;
    blob->store_in_database = 0;

    sprintf(blob->l0r_file_name, "%s/%s/%s", test_data_path, l0r_data_dir,
        l0r_basename);
    sprintf(blob->l1r_file_name, "%s", "L1R_output_file");
    sprintf(blob->cpf_filename, "%s/%s/%s", test_data_path, cpf_data_dir,
        cpf_basename);
    sprintf(blob->bpf_filename_oli, "%s/%s/%s", test_data_path, bpf_data_dir,
        bpf_basename_oli);
    strcpy(blob->bpf_filename_tirs, "");
    sprintf(blob->rlut_filename, "%s/%s/%s", test_data_path, rlut_data_dir,
        rlut_basename);

    status = finish_blob_initialization();
    assert_equal_with_message(status, SUCCESS, "Error completing test blob "
        "initialization");
    if (status != SUCCESS)
    {
        free_the_blob();
        return;
    }
    blob->number_of_bands = 2;
    blob->list_of_bands[0] = 6;           /* OLI band 6 */
    blob->list_of_bands[1] = 10;          /* TIRS band 10 */
    blob->list_of_band_indices[0] =
        ias_sat_attr_convert_band_number_to_index( blob->list_of_bands[0] );
    blob->list_of_band_indices[1] =
        ias_sat_attr_convert_band_number_to_index( blob->list_of_bands[1] );
    if ( blob->list_of_band_indices[0] == ERROR ||
         blob->list_of_band_indices[1] == ERROR )
    {
        assert_true_with_message( 0, "Converting band number to index");
        free_the_blob();
        return;
    }

    /* Initialize the L0R image data */
    ias_log_set_output_level( IAS_LOG_LEVEL_DISABLE );
    status = initialize_io_routines(blob);
    ias_log_set_output_level( IAS_LOG_LEVEL_INFO );
    assert_equal_with_message(status, SUCCESS, "Error initializing image "
        "data");
    if (status != SUCCESS)
    {
        free_the_blob();
        return;
    }

    /* Create the blob pixel mask array */
    status = create_blob_pixel_mask();
    assert_equal_with_message(status, SUCCESS, "Error creating blob pixel "
        "mask array");
    if (status != SUCCESS)
    {
        cleanup_io_routines(blob);
        free_the_blob();
        return;
    }

    /* Initialize memory management */
    status = initialize_memory_management_buffers();
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error initializing memory management");
        cleanup_io_routines( blob );
        free_the_blob();
        return;
    }

    status = dropped_frame_characterization_preprocess(blob);
    assert_equal_with_message(status, SUCCESS, "Error in dropped_frame_char_"
        "preprocess");
    if (status != SUCCESS)
    {
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Set the band index and SCA index for TIRS band 10, SCA 3 */
    normal_band_index = blob->list_of_band_indices[1];
    sca_index = 2;

    /* Get a copy of the memory buffer for the frame header and check the
       first couple of values.  As currently constructed, dropped frame
       characterization will also be performed on the VRP and blind bands
       for an OLI/TIRS normal band being processed if the band number exists
       and--for the blind band--the
       process_blind_band_flag in the attributes is set.  Consequently,
       the increment_consumer_count routine assigns the frame header
       buffer pointer in all SCAs to be processed in those bands in
       addition to the normal band  */
    status = increment_consumer_count(MM_L0R_TIRS_FRAME_HEADER);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error incrementing OLI frame "
            "header consumer count");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }
    /* It is important that the consumer counts get set *after* all the
       increments */
    status = set_consumer_counts();
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error setting consumer counts");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Run the test for OLI band 10 SCA 3 */
    band_atts = get_band_attributes(normal_band_index);

    tirs_fh = (const IAS_L0R_TIRS_FRAME_HEADER *)get_const_memory_buffer(
        MM_L0R_TIRS_FRAME_HEADER, normal_band_index, sca_index);
    assert_not_equal_with_message(tirs_fh, NULL, "Error obtaining TIRS "
        "L0R frame header data buffer");
    if (tirs_fh == NULL)
    {
        for (ii = 0; ii < blob->number_of_bands; ii++)
        {
            normal_band_index = blob->list_of_band_indices[ii];
            band_atts = get_band_attributes(normal_band_index);
            vrp_band_index = band_atts->vrp_band_index;
            blind_band_index = band_atts->blind_band_index;
            process_blind_band_flag = band_atts->process_blind_band_flag;
            if (band_atts->spectral_type != IAS_SPECTRAL_THERMAL)
            {
                for (jj = 0; jj < band_atts->number_scas; jj++)
                {
                    status = decrement_consumer_count(MM_L0R_OLI_FRAME_HEADER,
                            normal_band_index, jj);
                    assert_equal_with_message( status, SUCCESS,
                            "Error returned from decrement consumer count");
                }
                if (vrp_band_index >= 0)
                {
                    for (jj = 0; jj < band_atts->number_scas; jj++)
                    {
                        status = decrement_consumer_count(
                                MM_L0R_OLI_FRAME_HEADER,
                                vrp_band_index, jj);
                        assert_equal_with_message( status, SUCCESS,
                                "Error returned from decrement consumer count");
                    }
                }
                if ((blind_band_index >= 0) && process_blind_band_flag)
                {
                    for (jj = 0; jj < band_atts->number_scas; jj++)
                    {
                        status = decrement_consumer_count(
                                MM_L0R_OLI_FRAME_HEADER,
                                blind_band_index, jj);
                        assert_equal_with_message( status, SUCCESS,
                                "Error returned from decrement consumer count");
                    }
                }
            }
            else
            {
                for (jj = 0; jj < band_atts->number_scas; jj++)
                {
                    status = decrement_consumer_count(MM_L0R_TIRS_FRAME_HEADER,
                            normal_band_index, jj);
                    assert_equal_with_message( status, SUCCESS,
                            "Error returned from decrement consumer count");
                }
                if ((blind_band_index >= 0) && process_blind_band_flag)
                {
                    for (jj = 0; jj < band_atts->number_scas; jj++)
                    {
                        status = decrement_consumer_count(
                                MM_L0R_TIRS_FRAME_HEADER,
                                blind_band_index, jj);
                        assert_equal_with_message( status, SUCCESS,
                                "Error returned from decrement consumer count");
                    }
                }
            }
        }
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    assert_equal(tirs_fh[0].frame_status, GOOD_LINE);
    assert_equal(tirs_fh[1].frame_status, DROPPED_FRAME_LINE);
    assert_equal(tirs_fh[2].frame_status, GOOD_LINE);

    /* Call the API function */
    status = api_dropped_frame_characterization(blob, normal_band_index,
        sca_index, 0);
    assert_equal_with_message(status, SUCCESS, "API run of dropped frame "
        "characterization failed");
    if (status != SUCCESS)
    {
        for (ii = 0; ii < blob->number_of_bands; ii++)
        {
            normal_band_index = blob->list_of_band_indices[ii];
            band_atts = get_band_attributes(normal_band_index);
            vrp_band_index = band_atts->vrp_band_index;
            blind_band_index = band_atts->blind_band_index;
            process_blind_band_flag = band_atts->process_blind_band_flag;
            if (band_atts->spectral_type != IAS_SPECTRAL_THERMAL)
            {
                for (jj = 0; jj < band_atts->number_scas; jj++)
                {
                    status = decrement_consumer_count(MM_L0R_OLI_FRAME_HEADER,
                            normal_band_index, jj);
                    assert_equal_with_message( status, SUCCESS,
                            "Error returned from decrement consumer count");
                }
                if (vrp_band_index >= 0)
                {
                    for (jj = 0; jj < band_atts->number_scas; jj++)
                    {
                        status = decrement_consumer_count(
                                MM_L0R_OLI_FRAME_HEADER,
                                vrp_band_index, jj);
                        assert_equal_with_message( status, SUCCESS,
                                "Error returned from decrement consumer count");
                    }
                }
                if ((blind_band_index >= 0) && process_blind_band_flag)
                {
                    for (jj = 0; jj < band_atts->number_scas; jj++)
                    {
                        status = decrement_consumer_count(
                                MM_L0R_OLI_FRAME_HEADER,
                                blind_band_index, jj);
                        assert_equal_with_message( status, SUCCESS,
                                "Error returned from decrement consumer count");
                    }
                }
            }
            else
            {
                for (jj = 0; jj < band_atts->number_scas; jj++)
                {
                    status = decrement_consumer_count(MM_L0R_TIRS_FRAME_HEADER,
                            normal_band_index, jj);
                    assert_equal_with_message( status, SUCCESS,
                            "Error returned from decrement consumer count");
                }
                if ((blind_band_index >= 0) && process_blind_band_flag)
                {
                    for (jj = 0; jj < band_atts->number_scas; jj++)
                    {
                        status = decrement_consumer_count(
                                MM_L0R_TIRS_FRAME_HEADER,
                                blind_band_index, jj);
                        assert_equal_with_message( status, SUCCESS,
                                "Error returned from decrement consumer count");
                    }
                }
            }
        }
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Verify the blob pixel mask in SCA 1 has the correct dropped frame
       spans marked for all detectors */
    it = ias_pm_get_iterator( blob->pixel_mask[normal_band_index][0],
        IAS_PM_EXCLUDE, 0x00);
    if (it == NULL)
    {
        for (ii = 0; ii < blob->number_of_bands; ii++)
        {
            normal_band_index = blob->list_of_band_indices[ii];
            band_atts = get_band_attributes(normal_band_index);
            vrp_band_index = band_atts->vrp_band_index;
            blind_band_index = band_atts->blind_band_index;
            process_blind_band_flag = band_atts->process_blind_band_flag;
            if (band_atts->spectral_type != IAS_SPECTRAL_THERMAL)
            {
                for (jj = 0; jj < band_atts->number_scas; jj++)
                {
                    status = decrement_consumer_count(MM_L0R_OLI_FRAME_HEADER,
                            normal_band_index, jj);
                    assert_equal_with_message( status, SUCCESS,
                            "Error returned from decrement consumer count");
                }
                if (vrp_band_index >= 0)
                {
                    for (jj = 0; jj < band_atts->number_scas; jj++)
                    {
                        status = decrement_consumer_count(
                                MM_L0R_OLI_FRAME_HEADER,
                                vrp_band_index, jj);
                        assert_equal_with_message( status, SUCCESS,
                                "Error returned from decrement consumer count");
                    }
                }
                if ((blind_band_index >= 0) && process_blind_band_flag)
                {
                    for (jj = 0; jj < band_atts->number_scas; jj++)
                    {
                        status = decrement_consumer_count(
                                MM_L0R_OLI_FRAME_HEADER,
                                blind_band_index, jj);
                        assert_equal_with_message( status, SUCCESS,
                                "Error returned from decrement consumer count");
                    }
                }
            }
            else
            {
                for (jj = 0; jj < band_atts->number_scas; jj++)
                {
                    status = decrement_consumer_count(MM_L0R_TIRS_FRAME_HEADER,
                            normal_band_index, jj);
                    assert_equal_with_message( status, SUCCESS,
                            "Error returned from decrement consumer count");
                }
                if ((blind_band_index >= 0) && process_blind_band_flag)
                {
                    for (jj = 0; jj < band_atts->number_scas; jj++)
                    {
                        status = decrement_consumer_count(
                                MM_L0R_TIRS_FRAME_HEADER,
                                blind_band_index, jj);
                        assert_equal_with_message( status, SUCCESS,
                                "Error returned from decrement consumer count");
                    }
                }
            }
        }
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }
    for (det_index = 0; det_index < band_atts->number_detectors;
         det_index++)
    {
        /* Check the first span--1 frame, starting at frame 2 */
        status = ias_pm_get_next_span(it, &span);
        if (status == 1)
        {
            if ((span.starting_pixel_index != span_start_index_tirs[0])
                || (span.length_of_span != 1)
                || (span.detector_index != det_index)
                || (span.pixel_mask != IAS_RPS_DROPLINE))
            {
                dropped_frame_error_count++;
            }
        }

        /* Check the second span--1 frame, starting at frame 5 */
        status = ias_pm_get_next_span(it, &span);
        if (status == 1)
        {
            if ((span.starting_pixel_index != span_start_index_tirs[1])
                || (span.length_of_span != 1)
                || (span.detector_index != det_index)
                || (span.pixel_mask != IAS_RPS_DROPLINE))
            {
                dropped_frame_error_count++;
            }
        }

        /* Check the third span--5 frames, starting at frame 8 */
        status = ias_pm_get_next_span(it, &span);
        if (status == 1)
        {
            if ((span.starting_pixel_index != span_start_index_tirs[2])
                || (span.length_of_span != 5)
                || (span.detector_index != det_index)
                || (span.pixel_mask != IAS_RPS_DROPLINE))
            {
                dropped_frame_error_count++;
            }
        }

        /* Check the fourth span-- 8 frames, starting at frame 21 */
        status = ias_pm_get_next_span(it, &span);
        if (status == 1)
        {
            if ((span.starting_pixel_index != span_start_index_tirs[3])
                || (span.length_of_span != 8)
                || (span.detector_index != det_index)
                || (span.pixel_mask != IAS_RPS_DROPLINE))
            {
                dropped_frame_error_count++;
            }
        }

        /* Check the final span--6 frames, starting at frame 6044*/
        status = ias_pm_get_next_span(it, &span);
        if (status == 1)
        {
            if ((span.starting_pixel_index != span_start_index_tirs[4])
                || (span.length_of_span != 6)
                || (span.detector_index != det_index)
                || (span.pixel_mask != IAS_RPS_DROPLINE))
            {
                dropped_frame_error_count++;
            }
        }
    }
    assert_equal_with_message(dropped_frame_error_count, 0, "%d errors "
        "flagging dropped frames in pixel mask recorded",
        dropped_frame_error_count);

    /* Cleanup */
    for (ii = 0; ii < blob->number_of_bands; ii++)
    {
        normal_band_index = blob->list_of_band_indices[ii];
        band_atts = get_band_attributes(normal_band_index);
        vrp_band_index = band_atts->vrp_band_index;
        blind_band_index = band_atts->blind_band_index;
        process_blind_band_flag = band_atts->process_blind_band_flag;
        if (band_atts->spectral_type != IAS_SPECTRAL_THERMAL)
        {
            for (jj = 0; jj < band_atts->number_scas; jj++)
            {
                status = decrement_consumer_count(MM_L0R_OLI_FRAME_HEADER,
                        normal_band_index, jj);
                assert_equal_with_message( status, SUCCESS,
                        "Error returned from decrement consumer count");
            }
            if (vrp_band_index >= 0)
            {
                for (jj = 0; jj < band_atts->number_scas; jj++)
                {
                    status = decrement_consumer_count(MM_L0R_OLI_FRAME_HEADER,
                            vrp_band_index, jj);
                    assert_equal_with_message( status, SUCCESS,
                            "Error returned from decrement consumer count");
                }
            }
            if ((blind_band_index >= 0) && process_blind_band_flag)
            {
                for (jj = 0; jj < band_atts->number_scas; jj++)
                {
                    status = decrement_consumer_count(MM_L0R_OLI_FRAME_HEADER,
                            blind_band_index, jj);
                    assert_equal_with_message( status, SUCCESS,
                            "Error returned from decrement consumer count");
                }
            }
        }
        else
        {
            for (jj = 0; jj < band_atts->number_scas; jj++)
            {
                status = decrement_consumer_count(MM_L0R_TIRS_FRAME_HEADER,
                        normal_band_index, jj);
                assert_equal_with_message( status, SUCCESS,
                        "Error returned from decrement consumer count");
            }
            if ((blind_band_index >= 0) && process_blind_band_flag)
            {
                for (jj = 0; jj < band_atts->number_scas; jj++)
                {
                    status = decrement_consumer_count(MM_L0R_TIRS_FRAME_HEADER,
                            blind_band_index, jj);
                    assert_equal_with_message( status, SUCCESS,
                            "Error returned from decrement consumer count");
                }
            }
        }
    }

    cleanup_io_routines(blob);
    cleanup_memory_management_buffers(ENABLE_WARNING_DISPLAY);
    ias_pm_destroy_iterator(it);
    free_the_blob();

}   /* END test_tirs_dropped_frame_char_through_api */



int main(int argc, char *argv[])
{
    int status;


    /* Look for the test data path */
    test_data_path = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (test_data_path == NULL)
    {
        printf("Cannot locate IAS test data directory\n");
        return -1;
    }

    /* Initialize the satellite attributes */
    status = ias_sat_attr_initialize(IAS_L8);
    if (status != SUCCESS)
    {
        printf("Initializing Satellite Attributes Library\n");
        return -1;
    }

    /* Build the "test suite" */
    TestSuite *suite = create_test_suite();
    add_test(suite, test_oli_dropped_frame_char_through_api);
    add_test(suite, test_tirs_dropped_frame_char_through_api);

    /* Run the test */
    if (argc > 1)
        return run_single_test(suite, argv[1], create_text_reporter());
    return run_test_suite(suite, create_text_reporter());
}   /* END main */
