#include <stdio.h>
#include <string.h>
#include <math.h>

#include "ias_satellite_attributes.h"
#include "cgreen/cgreen.h"
#include "ias_logging.h"
#include "ias_radiometry.h"
#include "../rps_local.h"

#define NLINES 64
#define NDETS  64

Ensure gain_application_test_all_corrections( void );
Ensure gain_application_test_absolute_relative_gain_corrections( void );
Ensure gain_application_test_relative_gain_correction( void );
Ensure gain_application_test_no_correction( void );
Ensure test_algorithm_thru_api(void);

static char data_dir[] = "ias_base/rps/driver/gain_application";
static char cpf_dir[] = "ias_lib/cpf";
static char cpf_file[] = "L8CPF20090101_20090331.01";
static char expected_file1[] = "gain_application_expected_results1.txt";
static char expected_file2[] = "gain_application_expected_results2.txt";
static char expected_file3[] = "gain_application_expected_results3.txt";
static char expected_file4[] = "gain_application_expected_results4.txt";

int main( int argc, char *argv[] )
{
    TestSuite *suite = create_test_suite();
    add_test(suite, gain_application_test_all_corrections);
    add_test(suite, gain_application_test_absolute_relative_gain_corrections);
    add_test(suite, gain_application_test_relative_gain_correction);
    add_test(suite, gain_application_test_no_correction);
    add_test(suite, test_algorithm_thru_api);

    if (argc > 1) {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}

/*--------------------------------------------------------------------------
  read_expected_results
--------------------------------------------------------------------------*/
static int read_expected_results
(
    const char *file_name,          /* I: File name to open */
    float *buffer                   /* O: Result buffer */
)
{
    FILE *inf;                         /* File pointer for all reads */
    unsigned int det, frame;           /* loop variables for det and frame */

    /* Read in the expected results */
    inf = fopen( file_name, "r" );
    if (!inf)
    {
        assert_true_with_message(0, "Error opening %s", file_name);
        return 0;
    }

    for (det = 0; det < NDETS; det++)
    {
        for (frame = 0; frame < NLINES; frame++)
        {
            fscanf(inf, "%f\n", &buffer[det * NLINES + frame]);  
        }
    }
    fclose(inf);
    return 1;    
}

Ensure gain_application_test_all_corrections(void)
{
    float *buffer;                     /* pointer to the l1r data */
    float l1r_data[64][64];            /* input L1r image data */
    float expected_results[64][64];    /* expected output L1r image data */
    unsigned int det, frame;           /* loop variables for det and frame */

    double absolute_gain = 0.25;       /* Absolute gain */
    double relative_gain[64];          /* relative gain */
    double temp_sens_coeffs[64];       /* temperature sensitivity coeffcients */
    int perform_abs_gain = 1;          /* Correct the absolute gain */
    int perform_relative_gain = 1;     /* Correct the relative gain */
    int perform_temp_sens = 1;         /* Correct the temperature sensitivity */
    char *test_directory;              /* Test environment value */
    char file_name[PATH_MAX];          /* File name to open */
    int status;
    
    /* Get the environment variable for the test data */
    test_directory = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (!test_directory)
    {
        assert_true_with_message(0, "Getting IAS_UNIT_TEST_DATA_DIR");
        return;
    }

    /* Compose the path and file for expected result files */ 
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, expected_file1);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building the expected "
            "result file name");
        return;
    }

    /* Loop through the detector */
    for (det = 0; det < NDETS; det++)
    {
         relative_gain[det] = 1.05;
         temp_sens_coeffs[det] = 0.95;
    }

    /* Loop through the detector */
    for (det = 0; det < NDETS; det++)
    {
        /* Loop through the frame */
        for (frame = 0; frame < NLINES; frame++)
        {
            l1r_data[det][frame] = (float) (det * frame);
        }
    }

    /* Create a pointer link to the input l1r image data */
    buffer = (float *)l1r_data;

    /* Call the gain application algorithm */
    gain_application( buffer, 1, 1, NDETS, NLINES, absolute_gain, 
                &relative_gain[0], &temp_sens_coeffs[0], perform_abs_gain, 
                perform_relative_gain, perform_temp_sens);

    /* Read in the expected results */
    status = read_expected_results(file_name, &expected_results[0][0]);
    assert_equal_with_message(status, 1, "Error reading the expected results");

    /* Compare the results */
    int difference_count = 0;
    for (det = 0; det < NDETS; det++)
    {
        for (frame = 0; frame < NLINES; frame++)
        {
        if (fabsf(l1r_data[det][frame] - expected_results[det][frame]) > 1e-5)
            difference_count++;
        }
    }
    assert_equal_with_message(difference_count, 0, "There were %d "
        "differences out of %d values", difference_count, NDETS *
        NLINES);

}  /* END gain_application_test_all_corrections */

Ensure gain_application_test_absolute_relative_gain_corrections(void)
{
    float *buffer;                     /* pointer to the l1r data */
    float l1r_data[64][64];            /* input L1r image data */
    float expected_results[64][64];    /* expected output L1r image data */
    unsigned int det, frame;           /* loop variables for det and frame */

    double absolute_gain = 0.25;       /* Absolute gain */
    double relative_gain[64];          /* relative gain */
    double temp_sens_coeffs[64];       /* temperature sensitivity coeffcients */
    int perform_abs_gain = 1;          /* Correct the absolute gain */
    int perform_relative_gain = 1;     /* Correct the relative gain */
    int perform_temp_sens = 0;         /* Correct the temperature sensitivity */
    char *test_directory;              /* Test environment value */
    char file_name[PATH_MAX];          /* File name to open */
    int status;
    
    /* Get the environment variable for the test data */
    test_directory = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (!test_directory)
    {
        assert_true_with_message(0, "Getting IAS_UNIT_TEST_DATA_DIR");
        return;
    }

    /* Compose the path and file for expected result files */ 
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, expected_file2);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building the expected "
            "result file name");
        return;
    }

    /* Loop through the detector */
    for (det = 0; det < NDETS; det++)
    {
         relative_gain[det] = 1.05;
         temp_sens_coeffs[det] = 0.95;
    }

    /* Loop through the detector */
    for (det = 0; det < NDETS; det++)
    {
        /* Loop through the frame */
        for (frame = 0; frame < NLINES; frame++)
        {
            l1r_data[det][frame] = (float) (det * frame);
        }
    }

    /* Create a pointer link to the input l1r image data */
    buffer = (float *)l1r_data;

    /* Call the gain application algorithm */
    gain_application(buffer, 1, 1, NDETS, NLINES, absolute_gain, 
                &relative_gain[0], &temp_sens_coeffs[0], perform_abs_gain, 
                perform_relative_gain, perform_temp_sens);

    /* Read in the expected results */
    status = read_expected_results(file_name, &expected_results[0][0]);
    assert_equal_with_message(status, 1, "Error reading the expected results");

    /* Compare the results */
    int difference_count = 0;
    for (det = 0; det < NDETS; det++)
    {
        for (frame = 0; frame < NLINES; frame++)
        {
        if (fabsf(l1r_data[det][frame] - expected_results[det][frame]) > 1e-5)
            difference_count++;
        }
    }
    assert_equal_with_message(difference_count, 0, "There were %d "
        "differences out of %d values", difference_count, NDETS *
        NLINES);

}  /* END gain_application_test_absolute_relative_gain_corrections */

Ensure gain_application_test_relative_gain_correction(void)
{
    float *buffer;                     /* pointer to the l1r data */
    float l1r_data[64][64];            /* input L1r image data */
    float expected_results[64][64];    /* expected output L1r image data */
    unsigned int det, frame;           /* loop variables for det and frame */

    double absolute_gain = 0.25;       /* Absolute gain */
    double relative_gain[64];          /* relative gain */
    double temp_sens_coeffs[64];       /* temperature sensitivity coeffcients */
    int perform_abs_gain = 0;          /* Correct the absolute gain */
    int perform_relative_gain = 1;     /* Correct the relative gain */
    int perform_temp_sens = 0;         /* Correct the temperature sensitivity */
    char *test_directory;              /* Test environment value */
    char file_name[PATH_MAX];          /* File name to open */
    int status;
    
    /* Get the environment variable for the test data */
    test_directory = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (!test_directory)
    {
        assert_true_with_message(0, "Getting IAS_UNIT_TEST_DATA_DIR");
        return;
    }

    /* Compose the path and file for expected result files */ 
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, expected_file3);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building the expected "
            "result file name");
        return;
    }

    /* Loop through the detector */
    for (det = 0; det < NDETS; det++)
    {
         relative_gain[det] = 1.05;
         temp_sens_coeffs[det] = 0.95;
    }

    /* Loop through the detector */
    for (det = 0; det < NDETS; det++)
    {
        /* Loop through the frame */
        for (frame = 0; frame < NLINES; frame++)
        {
            l1r_data[det][frame] = (float) (det * frame);
        }
    }

    /* Create a pointer link to the input l1r image data */
    buffer = (float *)l1r_data;

    /* Call the gain application algorithm */
    gain_application(buffer, 1, 1, NDETS, NLINES, absolute_gain, 
                &relative_gain[0], &temp_sens_coeffs[0], perform_abs_gain, 
                perform_relative_gain, perform_temp_sens);

    /* Read in the expected results */
    status = read_expected_results(file_name, &expected_results[0][0]);
    assert_equal_with_message(status, 1, "Error reading the expected results");

    /* Compare the results */
    int difference_count = 0;
    for (det = 0; det < NDETS; det++)
    {
        for (frame = 0; frame < NLINES; frame++)
        {
        if (fabsf(l1r_data[det][frame] - expected_results[det][frame]) > 1e-5)
            difference_count++;
        }
    }
    assert_equal_with_message(difference_count, 0, "There were %d "
        "differences out of %d values", difference_count, NDETS *
        NLINES);

}  /* END gain_application_test_relative_gain_correction */

Ensure gain_application_test_no_correction(void)
{
    float *buffer;                     /* pointer to the l1r data */
    float l1r_data[64][64];            /* input L1r image data */
    float expected_results[64][64];    /* expected output L1r image data */
    unsigned int det, frame;           /* loop variables for det and frame */

    double absolute_gain = 0.25;       /* absolute gain */
    double relative_gain[64];          /* relative gain */
    double temp_sens_coeffs[64];       /* temperature sensitivity coeffcients */
    int perform_abs_gain = 0;          /* flag for the absolute gain 
                                           correction*/
    int perform_relative_gain = 0;     /* flag for the relative gain 
                                           correction */
    int perform_temp_sens = 0;         /* flag for the temperature 
                                           sensitivity correction */
    char *test_directory;              /* Test environment value */
    char file_name[PATH_MAX];          /* File name to open */
    int status;
    
    /* Get the environment variable for the test data */
    test_directory = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (!test_directory)
    {
        assert_true_with_message(0, "Getting IAS_UNIT_TEST_DATA_DIR");
        return;
    }

    /* Compose the path and file for expected result files */ 
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, expected_file4);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building the expected "
            "result file name");
        return;
    }

    /* Loop through the detector */
    for (det = 0; det < NDETS; det++)
    {
         relative_gain[det] = 1.05;
         temp_sens_coeffs[det] = 0.95;
    }

    /* Loop through the detector */
    for (det = 0; det < NDETS; det++)
    {
        /* Loop through the frame */
        for (frame = 0; frame < NLINES; frame++)
        {
            l1r_data[det][frame] = (float) (det * frame);
        }
    }

    /* Create a pointer link to the input l1r image data */
    buffer = (float *)l1r_data;

    /* Call the gain application algorithm */
    gain_application( buffer, 1, 1, NDETS, NLINES, absolute_gain, 
                &relative_gain[0], &temp_sens_coeffs[0], perform_abs_gain, 
                perform_relative_gain, perform_temp_sens);

    /* Read in the expected results */
    status = read_expected_results(file_name, &expected_results[0][0]);
    assert_equal_with_message(status, 1, "Error reading the expected results");

    /* Compare the results */
    int difference_count = 0;
    for (det = 0; det < NDETS; det++)
    {
        for (frame = 0; frame < NLINES; frame++)
        {
        if (fabsf(l1r_data[det][frame] - expected_results[det][frame]) > 1e-5)
            difference_count++;
        }
    }
    assert_equal_with_message(difference_count, 0, "There were %d "
        "differences out of %d values", difference_count, NDETS *
        NLINES);

}  /* END gain_application_test_no_correction */

/*--------------------------------------------------------------------------
  local_free_the_blob

  This a local override of the routine to free the blob.  Since not all
  members of the blob are initialized -- specifically the pixel mask -- the
  normal routine cannot be called.
--------------------------------------------------------------------------*/
static void local_free_the_blob(BLOB *blob)
{
    int i;
    
    if ( !blob )
        return;

    if ( blob->l1r_band_metadata )
    {
        for ( i = 0; i < blob->total_bands; i++ )
            free( blob->l1r_band_metadata[i] );
        free( blob->l1r_band_metadata );
    }

    if ( blob->band_information )
    {
        for ( i = 0; i < blob->total_bands; i++ )
        {
            if ( blob->band_information[i] )
                free( blob->band_information[i] );
        }
    }
    free( blob->band_information );

    if ( blob->cpf )
        ias_cpf_free( blob->cpf );
    free( blob );
}

/*--------------------------------------------------------------------------
  test_algorithm_thru_api

  This test sets up for a call to the algorithm through the API.  Also calls
  the pre-process routine to set up memory management. The input ASCII files 
  are extracted from calval binary test files with only one band and one sca 
  used here. The compared result file is generated from cal/val IDL code.
--------------------------------------------------------------------------*/
Ensure test_algorithm_thru_api(void)
{
    const IAS_BAND_ATTRIBUTES *band_atts;
    char *test_directory;          /* test environment value */
    char file_name[PATH_MAX];      /* File name to open */
    BLOB *blob;                    /* Declare a blob structure */
    float *l1r_input_buffer = NULL;/* Image data for the current band/SCA */
    int det;                       /* Detector loop counter */
    int line;                      /* Frame loop counter */
    int band_number = 1;           /* 1-based band number */
    int status;                    /* Return variable */
    int band_index;                /* 0-based band number */
    int list_of_bands[1] = {1};         /* List of one test band */
    int list_of_band_indices[1] = {0};  /* List of one test band index */
    int num_bands = 1;                      /* Number of bands for testing */
    double *absolute_gain = NULL;   /* SCA level absolute gain */
    double *relative_gain = NULL;   /* detector level relative gain */
    double *temp_correction_factors = NULL;/* detector level temperature 
                                             sensitivity */

    /* Get the environment variable for the test data */
    test_directory = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (!test_directory)
    {
        assert_true_with_message(0, "Getting IAS_UNIT_TEST_DATA_DIR");
        return;
    }

    /* The attributes must be initialized for memory management to work */
    ias_sat_attr_initialize(IAS_L8);

    /* Allocate a blob and populate the minimum members for the test
       to succeed */
    blob = allocate_blob();
    blob->max_number_scas = 14;
    blob->total_bands = 30;
    blob->band_information = calloc(blob->total_bands, 
                                    sizeof( RPS_BAND_ATTRIBUTES *));
    blob->l1r_band_metadata = calloc( blob->total_bands,
        sizeof( IAS_L1R_BAND_METADATA * ));
    if ( ! blob->l1r_band_metadata )
    {
        assert_true_with_message(0, "Error allocating band metadata");
        local_free_the_blob( blob );
        return;
    }
    for ( band_index = 0; band_index < blob->total_bands; band_index++ )
    {
        blob->l1r_band_metadata[band_index] = calloc( 1,
            sizeof(IAS_L1R_BAND_METADATA));
        if ( ! blob->l1r_band_metadata[band_index] )
        {
            assert_true_with_message(0, "Error allocating band metadata");
            local_free_the_blob( blob );
            return;
        }
    }

    /* Local initialize the blob */
    for (band_number = 1; band_number <= blob->total_bands; band_number++)
    {
        int i;
        band_atts = ias_sat_attr_get_band_attributes(band_number);
        if (band_atts == NULL)
        {
            assert_true_with_message(FALSE, "Error obtaining RPS attribute "
                "information for band %d", band_number);
            local_free_the_blob(blob);
            return;
        }

        int band_index = band_atts->band_index;

        blob->band_information[band_index] = calloc(1,
            sizeof(RPS_BAND_ATTRIBUTES));
        if (!blob->band_information[band_index])
        {
            assert_true_with_message(0, "Error allocating band "
                "attribute memory");
            local_free_the_blob(blob);
            return;
        }
        blob->band_information[band_index]->band_number = band_number;
        blob->band_information[band_index]->number_lines = NLINES;
        blob->band_information[band_index]->number_detectors = NDETS;
        blob->band_information[band_index]->number_scas = band_atts->scas;
        blob->band_information[band_index]->band_classification =
            band_atts->band_classification;
        blob->band_information[band_index]->spectral_type =
            band_atts->spectral_type;

        for ( i = 0; i < num_bands; i++ )
        {
            if ( band_number == list_of_bands[i] )
            {
                list_of_band_indices[i] = band_index;
                break; /* Since there are no repeated bands */
            }
        }
    }

    /* Set gain application flags */
    blob->apply_absolute_gain = 1;
    blob->apply_relative_gain = 1;
    blob->perform_temp_sens = 1;

    /* Open the CPF */
    sprintf(file_name, "%s/%s/%s", test_directory, cpf_dir, cpf_file);
    blob->cpf = ias_cpf_read(file_name);
    if (!blob->cpf)
    {
        assert_true_with_message(0, "Error opening CPF: %s", file_name);
        local_free_the_blob(blob);
        return;
    }

    /* Diable logging to suppress the memory management warnings */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* This initializes the memory management and registers buffers through
       the algorithms pre-process routine.  This algorithm doesn't need any
       memory management buffers that require the list of bands to be
       initialized -- number of bands set to 0 for that reason. */
    blob->number_of_bands = 0;
    blob->band_fields_initialized = 1;
    status = initialize_memory_management_buffers();
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error initializing memory management");
        local_free_the_blob(blob);
        return;
    }

    /* Initialize the RPS process level arrays */
    status = initialize_rps_level(1, 1);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error calling initialize_rps_level");
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    status = gain_application_preprocess(blob);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error returned from: "
            "gain_application_preprocess");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    /* Register the L1R buffer to this routine so that the API can't
       destroy the memory before we can retrieve the output */
    status = increment_consumer_count(IMAGE_L1R);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error returned from: "
            "increment_consumer_count IMAGE_L1R");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    /* Get the image buffer for band 1, SCA 1 this simply returns an unfilled,
       allocated buffer */
    l1r_input_buffer = (float *)preprocess_get_memory_buffer(IMAGE_L1R,
        list_of_band_indices[0], 0);
    if (!l1r_input_buffer)
    {
        assert_true_with_message(0, "Getting IMAGE_L1R buffer");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    /* Set the l1r_input_buffer values */
    /* Loop through the detector */
    for (det = 0; det < NDETS; det++)
    {
        /* Loop through the frame */
        for (line = 0; line < NLINES; line++)
        {
            l1r_input_buffer[det * NLINES + line] = 
                (float) (det + line * NDETS);
        }
    }
    set_buffer_filled_flag( IMAGE_L1R, list_of_band_indices[0], 0);

    /* Check an "empty" buffer for the temp_correction_factors */
    temp_correction_factors = (double *)preprocess_get_memory_buffer(
        MM_TEMP_CORRECTION_FACTOR, list_of_band_indices[0], 0);
    if (!temp_correction_factors)
    {
        assert_true_with_message(0, "Error getting "
            "MM_TEMP_CORRECTION_FACTOR buffer");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    /* Set the temperature sensivity correction value */
    for (det = 0; det < NDETS; det++)
        temp_correction_factors[det] = 0.95;
    set_buffer_filled_flag( MM_TEMP_CORRECTION_FACTOR,
        list_of_band_indices[0], 0 );    


    /* It is important that the consumer counts get set *after* all the
       increments */
    status = set_consumer_counts();
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error setting consumer counts");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    /* Check an "empty" buffer for the absolute_gain */
    absolute_gain = (double *)get_const_memory_buffer(
            MM_ABSOLUTE_GAINS, list_of_band_indices[0], 0);
    if (!absolute_gain)
    {
        assert_true_with_message(0, "Error getting "
            "MM_ABSOLUTE_GAINS buffer");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    /* Set the absolute gain value */
    *absolute_gain = 0.25;

    /* Check an "empty" buffer for the relative_gain */
    relative_gain = (double *)get_const_memory_buffer(
            MM_RELATIVE_GAINS, list_of_band_indices[0], 0);
    if (!relative_gain)
    {
        assert_true_with_message(0, "Error getting "
            "MM_RELATIVE_GAINS buffer");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    /* Set the relative gain value */
    for (det = 0; det < NDETS; det++)
    {
         relative_gain[det] = 1.05;
    }

    /* Call the API with band 1 and SCA 1 */
    band_number = list_of_bands[0];
    band_index = list_of_band_indices[0];
    status = api_gain_application( blob, band_index, 0, 0 );
    assert_equal_with_message( status,SUCCESS, "Error returned from "
        "api_gain_application");
    if ( status != SUCCESS )
    {
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob( blob );
        return;
    }

    /* Check some outputs with the expected values */
    assert_double_equal(l1r_input_buffer[64], 3.61904764);
    assert_double_equal(l1r_input_buffer[64 * 2], 7.23809528);
    assert_double_equal(l1r_input_buffer[64 * 64 - 1], 14820.0);

    /* Free the image memory */
    status = decrement_consumer_count(IMAGE_L1R, 0, 0);
    assert_equal_with_message( status, SUCCESS,
        "Error returned from decrement consumer count");

    cleanup_rps_level();
    cleanup_memory_management_buffers(ENABLE_WARNING_DISPLAY);
    local_free_the_blob(blob);
} 

