/*----------------------------------------------------------------------------
NAME:    impulse_noise_charcterization_test.c

PURPOSE: Test the impulse_noise_charcaterization algorithm by testing it 
         through API
----------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "cgreen/cgreen.h"
#include "ias_logging.h"
#include "../rps_local.h"
#include "ias_l0r.h"

#define NDETS 8
#define NLINES 12

Ensure test_algorithm_thru_api(void);

int main(int argc, char *argv[])
{
    TestSuite *suite = create_test_suite();
    add_test(suite, test_algorithm_thru_api);
    if ( argc == 2 )
        return run_single_test(suite, argv[1], create_text_reporter());
    return run_test_suite(suite, create_text_reporter());
}

/*--------------------------------------------------------------------------
  test_algorithm_thru_api

  This test sets up for a call to the algorithm through the API.  Also calls
  the pre-process routine to set up memory management. The test data used
  is the same as that in impulse_noise_charcterization test code under 
  ias_lib/rps/tests directory
--------------------------------------------------------------------------*/
Ensure test_algorithm_thru_api(void)
{
    const IAS_BAND_ATTRIBUTES *band_atts; /* RPS band attributes */
    char *test_directory;          /* test environment value */
    int det;                       /* Detector loop counter */
    int line;                      /* Frame loop counter */
    int i, j;
    IAS_REPORT_HEADER *report_header = NULL;
    char work_order_id[IAS_WORKORDER_ID_SIZE];
    char routine_name[] = "Impulse noise characterization";

    BLOB *blob;                    /* Declare a blob structure */
    int band_number = 1;           /* 1-based band number */
    int status;
    int band_index;
    PIXEL_MASK_TYPE mask;
    const IAS_SATELLITE_ATTRIBUTES *sat_attributes;
    int *median_filter_width;
    int *impulse_noise_limit;
    double *detector_noise;
    uint16_t *l0r_input_buffer = NULL;
    uint16_t l0r_data[12][8] =
    {
        { 400,650,399,2  ,402,400,2  ,400 },
        { 399,400,400,2  ,412,408,2  ,401 },
        { 500,399,399,2  ,399,400,2  ,394 },
        { 400,600,401,2  ,401,399,2  ,400 },
        { 4  ,4  ,4  ,2  ,4  ,4  ,2  ,4   },
        { 401,400,401,2  ,400,200,2  ,400 },
        { 401,399,300,2  ,280,400,2  ,401 },
        { 550,520,399,2  ,400,500,2  ,600 },
        { 550,402,400,2  ,280,150,2  ,420 },
        { 399,400,401,50 ,401,400,2  ,399 },
        { 4  ,4  ,80 ,2  ,4  ,4  ,2  ,4   },
        { 400,400,400,2  ,400,60 ,2  ,400 }
    };
    double detector_noise_value[] =
        { 0.78, 0.81, 0.86, 0.82, 0.79, 0.85, 0.91, 0.80 };

    static char bpf_dir[] = "ias_lib/bpf";
    static char cpf_dir[] = "ias_lib/cpf";
    static char rlut_dir[] = "ias_lib/rlut";
    static char l0r_dir[] = 
           "ias_base/rps/driver/LC81800382002045SGS00";
    static char bpf_name_oli[] = "test_LO8BPF20140310103310_20140310103345.06";
    static char cpf_name[] = "L8CPF20090101_20090331.01";
    static char rlut_name[] = "test_linearization_rlut.h5";
    static char l0r_name[] = "LC81800382002045SGS00";
    int vrp_number = -1;
    int vrp_index = -1;

    /* Get the environment variable for the test data */
    test_directory = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (!test_directory)
    {
        assert_true_with_message(0, "Getting IAS_UNIT_TEST_DATA_DIR");
        return;
    }

    /* The attributes must be initialized for memory management to work */
    ias_sat_attr_initialize(IAS_L8);

    /* Allocate a blob and populate the minimum members for the test
       to succeed */
    blob = allocate_blob();
    if ( !blob )
    {
        assert_true_with_message( 0, "allocating blob" );
        return;
    }

    /* Initialize blob */
    status = snprintf(blob->cpf_filename, sizeof(blob->cpf_filename), 
             "%s/%s/%s", test_directory, cpf_dir, cpf_name);
    if ((status < 0) || (status >= sizeof(blob->cpf_filename)))
    {
        assert_true_with_message(FALSE, "Error building CPF file name");
        return;
    }
    status = snprintf(blob->bpf_filename_oli, sizeof(blob->bpf_filename_oli), 
             "%s/%s/%s", test_directory, bpf_dir, bpf_name_oli);
    if ((status < 0) || (status >= sizeof(blob->bpf_filename_oli)))
    {
        assert_true_with_message(FALSE, "Error building OLI BPF file name");
        return;
    }
    strcpy(blob->bpf_filename_tirs, "");
    status = snprintf(blob->rlut_filename, sizeof(blob->rlut_filename), 
             "%s/%s/%s", test_directory, rlut_dir, rlut_name);
    if ((status < 0) || (status >= sizeof(blob->rlut_filename)))
    {
        assert_true_with_message(FALSE, "Error building RLUT file name");
        return;
    }
    sat_attributes = ias_sat_attr_get_attributes();
    blob->total_bands = sat_attributes->total_bands;
    blob->number_of_bands = 1;

    if ( finish_blob_initialization())
    {
        assert_true_with_message( 0, "initializing blob" );
        return;
    }
    blob->list_of_bands[0] = 1;
    blob->list_of_band_indices[0] =
        ias_sat_attr_convert_band_number_to_index( blob->list_of_bands[0] );
    if ( blob->list_of_band_indices[0] == ERROR )
    {
        assert_true_with_message( 0, "Converting band number to index");
        free_the_blob();
        return;
    }

    /* Initialize I/O routines */
    status = snprintf(blob->l0r_file_name, sizeof(blob->l0r_file_name), 
             "%s/%s/%s", test_directory, l0r_dir, l0r_name);
    if ((status < 0) || (status >= sizeof(blob->l0r_file_name)))
    {
        assert_true_with_message(FALSE, "Error building L0R data file name");
        return;
    }
    status = snprintf(blob->l1r_file_name, sizeof(blob->l1r_file_name), "%s", 
             "temp_l1r");
    if ((status < 0) || (status >= sizeof(blob->l1r_file_name)))
    {
        assert_true_with_message(FALSE, "Error building L1R data file name");
        return;
    }
    ias_log_set_output_level( IAS_LOG_LEVEL_DISABLE );
    status = initialize_io_routines( blob );
    if ( status )
    {
        free_the_blob();
        assert_true_with_message( 0,
            "initialize_io_routines returned an error" );
        return;
    }
    ias_log_set_output_level( IAS_LOG_LEVEL_INFO );

    /* Set the detector and frame numbers in the blob */
    for (band_number = 1; band_number <= blob->total_bands; band_number++)
    {
        band_atts = ias_sat_attr_get_band_attributes(band_number);
        if (band_atts == NULL)
        {
            assert_true_with_message(FALSE, "Error obtaining RPS attribute "
                "information for band %d", band_number);
            free_the_blob();
            return;
        }
        band_index = band_atts->band_index;

        if ( band_index == blob->list_of_band_indices[0] )
            vrp_number = band_atts->vrp_band_number;
            

        blob->band_information[band_index]->number_lines = NLINES;
        blob->band_information[band_index]->number_detectors = NDETS;
    }
    if ( vrp_number > 0 )
    {
        vrp_index = ias_sat_attr_convert_band_number_to_index( vrp_number );
        if ( vrp_index == ERROR )
        {
            assert_true_with_message( 0, "Converting band number to index");
            free_the_blob();
            return;
        }
        
    }
    /* Override the VRP, because there is no reason to check it */
    blob->band_information[0]->vrp_band_number = -1;
    blob->band_information[0]->vrp_band_index = -1;

    blob->store_in_database = 0;
    blob->generate_report = 1;

    /* Create blob pixel_masks */
    status = create_blob_pixel_mask();
    assert_equal_with_message(status, SUCCESS, "Error creating blob pixel "
        "masks");
    if (status != SUCCESS)
    {
        cleanup_io_routines(blob);
        free_the_blob();
        return;
    }

    /* Set up a standard header */
    setenv("PCS_WOID", "I4582", 1);
    strcpy(work_order_id, "I4582");
    report_header = ias_misc_initialize_report_header("",
        work_order_id, routine_name);
    if (!report_header)
    {
        assert_true_with_message(0, "Error initializing report header");
        cleanup_io_routines(blob);
        free_the_blob();
        return;
    }
    ias_misc_set_header_pathrow(report_header, 0, 0);
    blob->report_information.header = report_header;
    blob->report_information.directory = NULL;

    /* Add the inoperable detectors to pixel mask */
    assert_equal(SUCCESS, ias_pm_add_pixels(blob->pixel_mask[0][0], 
                 4-1, 1-1, 12, IAS_RPS_INOP_DETECTOR));
    assert_equal(SUCCESS, ias_pm_add_pixels(blob->pixel_mask[0][0], 
                 7-1, 1-1, 12, IAS_RPS_INOP_DETECTOR));

    /* Add the dropped frames to pixel mask */
    for ( det = 1-1; det <= 8-1; det++ )
    {
        assert_equal(SUCCESS, ias_pm_add_pixels(blob->pixel_mask[0][0], det, 
                     5-1, 1, IAS_RPS_DROPLINE));
    }

    for ( det = 1-1; det <= 8-1; det++ )
    {
        assert_equal(SUCCESS, ias_pm_add_pixels(blob->pixel_mask[0][0], det, 
                     11-1, 1, IAS_RPS_DROPLINE));
    }

    /* This initializes the memory management and registers buffers through
       the algorithms pre-process routine */
    status = initialize_memory_management_buffers();
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error initializing memory management");
        cleanup_io_routines(blob);
        free_the_blob();
        return;
    }

    status = impulse_noise_characterization_preprocess(blob);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error returned from: "
            "impulse_noise_characterization_preprocess");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Register the L0R buffer to this routine so that the API can't
       destroy the memory before we can retrieve the output */
    status = increment_consumer_count(IMAGE_L0R);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error returned from: "
            "increment_consumer_count IMAGE_L0R");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Get the image buffer for band 1, SCA 1 this simply returns an unfilled,
       allocated buffer */
    l0r_input_buffer = (uint16_t *)preprocess_get_memory_buffer(IMAGE_L0R, 
                       blob->list_of_band_indices[0], 0);
    if (!l0r_input_buffer)
    {
        assert_true_with_message(0, "Getting IMAGE_L0R buffer");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Set the l0r_input_buffer values */
    /* Loop through the frame */
    for (line = 0; line < NLINES; line++)
    {
        /* Loop through the detector */
        for (det = 0; det < NDETS; det++)
        {
            l0r_input_buffer[line * NDETS + det] = 
                 l0r_data[line][det];
        }
    }
    set_buffer_filled_flag( IMAGE_L0R, blob->list_of_band_indices[0], 0);

    /* It is important that the consumer counts get set *after* all the
       increments */
    status = set_consumer_counts();
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error setting consumer counts");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Check an "empty" buffer for the median_filter_width */
    median_filter_width = (int *)get_const_memory_buffer(
            MM_IMPULSE_NOISE_FILTER_WIDTH, blob->list_of_band_indices[0], 0);
    if (!median_filter_width)
    {
        assert_true_with_message(0, "Error getting "
            "MM_IMPULSE_NOISE_FILTER_WIDTH buffer");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Set the median filter width value */
    *median_filter_width = 5;

    /* Check an "empty" buffer for the impulse_noise_limit */
    impulse_noise_limit = (int *)get_const_memory_buffer(
            MM_IMPULSE_NOISE_LIMIT, blob->list_of_band_indices[0], 0);
    if (!impulse_noise_limit)
    {
        assert_true_with_message(0, "Error getting "
            "MM_IMPULSE_NOISE_LIMIT buffer");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Set the impulse noise limit value */
    *impulse_noise_limit = 8;

    /* Check an "empty" buffer for the detector_noise */
    detector_noise = (double *)get_const_memory_buffer(
            MM_DETECTOR_NOISE, blob->list_of_band_indices[0], 0);
    if (!detector_noise)
    {
        assert_true_with_message(0, "Error getting "
            "MM_DETECTOR_NOISE buffer");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Set the detector noise values */
    for (det = 0; det < NDETS; det++)
    {
         detector_noise[det] = detector_noise_value[det];
    }

    /* Call the API */
    band_number = blob->list_of_bands[0];
    band_index = blob->list_of_band_indices[0];
    status = api_impulse_noise_characterization(blob, band_index, 0, 0);
    assert_equal_with_message(status, SUCCESS, "Error returned from "
        "api_impulse_noise_characterization with band = 1");
    if (status != SUCCESS)
    {
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Make sure the correct number of impulse noise locations were
       identified */
    int count1 = 0;
    for ( i = 0; i < NDETS; i++ )
    {
        for ( j = 0; j < NLINES; j++ )
        {
           ias_pm_get_mask_at( blob->pixel_mask[0][0], i, j, &mask );
            if ( mask & IAS_RPS_IMPNOISE )
                count1++;
        }
    }
    assert_equal( count1, 10 );

    /* Check each of the pixel locations that should be impulse noise */
    int in_det2[]  = {0,1,1,1,2,4,4,5,5,7};
    int in_line2[] = {2,0,3,7,6,6,8,1,5,7};
    for ( i = 0; i < 10; i++ )
    {
        ias_pm_get_mask_at( blob->pixel_mask[0][0], in_det2[i], in_line2[i], 
                            &mask );
        assert_true_with_message( mask & IAS_RPS_IMPNOISE,
            "i = %d  Detector %d  Line %d  Mask %d", in_det2[i],
                in_line2[i], mask );
    }

    /* Free the image memory */
    status = decrement_consumer_count(IMAGE_L0R, blob->list_of_band_indices[0],
        0);
    assert_equal_with_message( status, SUCCESS,
        "Error returned from decrement consumer count");
    /* Free the associated VRP if there is one */
    if ( vrp_index >= 0 )
    {
        status = decrement_consumer_count(IMAGE_L0R, vrp_index, 0);
        assert_equal_with_message( status, SUCCESS,
            "Error returned from decrement consumer count");
    }

    cleanup_io_routines(blob);
    cleanup_memory_management_buffers(ENABLE_WARNING_DISPLAY);
    free_the_blob();
}

