#include "string.h"
#include "cgreen/cgreen.h"
#include "ias_logging.h"
#include "../rps_local.h"

/* Define DEBUG to run single test */
#define DEBUG

Ensure inop_detector_test(void);

int main( int argc, char *argv[] )
{
    TestSuite *suite = create_test_suite();
    add_test(suite, inop_detector_test );

#ifdef DEBUG
    return run_single_test( suite, "inop_detector_test",
        create_text_reporter());
#endif
    return run_test_suite( suite, create_text_reporter());
}

static char test_dir[] = ".";
static char bpf_dir[] = "ias_lib/bpf";
static char rlut_dir[] = "ias_lib/rlut";
static char data_dir[] = "ias_base/rps/driver/L0R";
static char l0r_file[] = "LO82220010052011265EDC00";
static char l1rname[] = "L1R_output_file";
static char bpf_name_oli[] = "test_LO8BPF20140310103310_20140310103345.06";
static char rlut_name[] = "test_linearization_rlut.h5";

Ensure inop_detector_test(void)
{
    BLOB *blob;
    int *detector_list;
    int number_of_detectors;
    int status;
    IAS_PIXEL_MASK *pm;
    IAS_PIXEL_MASK_ITERATOR *it;
    IAS_PIXEL_MASK_SPAN span;
    char *test_directory;
    char l0r_file_name[512];
    int band_index;

    test_directory = getenv("IAS_UNIT_TEST_DATA_DIR");
    if ( !test_directory )
    {
        assert_true_with_message( 0, "Getting IAS_UNIT_TEST_DATA_DIR");
        return;
    }
    sprintf( l0r_file_name, "%s/%s/%s", test_directory, data_dir, l0r_file );

    blob = allocate_blob();
    if ( !blob )
    {
        assert_true_with_message( 0, "allocating blob" );
        return;
    }
    blob->total_bands = 30;
    blob->number_of_bands = 2;
    strcpy( blob->l0r_file_name, l0r_file_name );
    strcpy( blob->l1r_file_name, l1rname );
    sprintf( blob->cpf_filename, "%s/cpf_detector_group", test_dir );
    sprintf( blob->bpf_filename_oli, "%s/%s/%s", test_directory, bpf_dir, 
             bpf_name_oli );
    strcpy( blob->bpf_filename_tirs, "" );
    sprintf( blob->rlut_filename, "%s/%s/%s", test_directory, rlut_dir, 
             rlut_name );

    /* Several things must be done to ensure that all blob members are
       properly initialized.  We need the attributes library initialized
       before completing the blob initialization.  Then the I/O routines
       must be initialized to populate the band present stuff. */
    if ( ias_sat_attr_initialize( IAS_L8 ))
    {
        assert_true_with_message( 0, "Error initializing satellite "
            "attributes library.");
        free_the_blob();
        return;
    }

    if ( finish_blob_initialization())
    {
        assert_true_with_message( 0, "initializing blob -- %s",
            blob->cpf_filename );
        free_the_blob();
        return;
    }
    blob->list_of_bands[0] = 1;
    blob->list_of_bands[1] = 2;
    blob->list_of_band_indices[0] =
        ias_sat_attr_convert_band_number_to_index( blob->list_of_bands[0] );
    blob->list_of_band_indices[1] =
        ias_sat_attr_convert_band_number_to_index( blob->list_of_bands[1] );
    if ( blob->list_of_band_indices[0] == ERROR ||
         blob->list_of_band_indices[1] == ERROR )
    {
        assert_true_with_message( 0, "Converting band number to index");
        free_the_blob();
        return;
    }

    ias_log_set_output_level( IAS_LOG_LEVEL_DISABLE );
    if ( initialize_io_routines( blob ))
    {
        assert_true_with_message( 0,
            "initialize_io_routines returned an error" );
        free_the_blob();
        return;
    }
    ias_log_set_output_level( IAS_LOG_LEVEL_INFO );

    /* Now that the I/O routines are initialized we know what bands are
       present so create the pixel mask. */
    if ( create_blob_pixel_mask())
    {
        assert_true_with_message( 0, "Error returned from "
            "create_blob_pixel_mask");
        cleanup_io_routines( blob );
        free_the_blob();
        return;
    }

    /* Initialize the memory management layer and register the inoperable
       detector memory needed by calling the pre-process routine */
    status = initialize_memory_management_buffers();
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error initializing memory management");
        cleanup_io_routines( blob );
        free_the_blob();
        return;
    }

    status = inoperable_detector_identification_preprocess( blob );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error returned from "
            "inoperable_detector_identification_preprocess");
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        cleanup_io_routines( blob );
        free_the_blob();
        return;
    }
    inoperable_detector_identification_preprocess( blob );
    /* It is important that the consumer counts get set *after* all the
       increments */
    status = set_consumer_counts();
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error setting consumer counts");
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        cleanup_io_routines( blob );
        free_the_blob();
        return;
    }

    /* Perform a few sanity checks to make sure we get some valid values
       from the CPF:
       band 1 SCA 1 should be none. */
    number_of_detectors = get_memory_buffer_number_values(
        MM_INOPERABLE_DETECTORS, 0, 0 );
    assert_equal( number_of_detectors, 0 );

    /* Band 2 SCA 5 should be 5 */
    number_of_detectors = get_memory_buffer_number_values(
        MM_INOPERABLE_DETECTORS, 1, 4 );
    assert_equal( number_of_detectors, 5 );
    detector_list = (int *)get_const_memory_buffer( MM_INOPERABLE_DETECTORS, 
                                                    2, 4 );

    /* Band 25 -- the VRP for band 7 -- SCA 12 should be 1 */
    number_of_detectors = get_memory_buffer_number_values(
        MM_INOPERABLE_DETECTORS, 24, 11 );
    assert_equal( number_of_detectors, 1 );

    /* Call the API for inoperable detectors several times to set the pixel
       mask. */
    status = 0;
    band_index = blob->list_of_band_indices[0];
    if ( api_inoperable_detector_identification( blob, band_index, 1-1, 0 ))
        status = 1;
    band_index = blob->list_of_band_indices[1];
    if ( api_inoperable_detector_identification( blob, band_index, 5-1, 0 ))
        status = 1;
    if ( status )
    {
        assert_true_with_message( 0, "Error returned from "
            "api_inoperable_detector_identification");
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        cleanup_io_routines( blob );
        free_the_blob();
        return;
    }

    /* Check the state of the pixel mask -- this test shouldn't return
       any inoperable detectors. */
    pm = blob->pixel_mask[0][0];
    it = ias_pm_get_iterator( pm, IAS_PM_INCLUDE, IAS_RPS_INOP_DETECTOR );
    if ( !it )
    {
        assert_true_with_message( 0, "No iterator returned for "
            "pixel mask band 1, SCA 1");
    }
    else
    {
        status = ias_pm_get_next_span( it, &span );
        assert_false_with_message( status, "Span was returned for band 1, "
            "SCA 1 and it shouldn't have.");
    }
    ias_pm_destroy_iterator( it );

    /* This test should return 5 spans */
    number_of_detectors = get_memory_buffer_number_values(
        MM_INOPERABLE_DETECTORS, 1, 4 );
    detector_list = (int *)get_const_memory_buffer(
        MM_INOPERABLE_DETECTORS, 1, 4 );
    pm = blob->pixel_mask[1][4];
    it = ias_pm_get_iterator( pm, IAS_PM_INCLUDE, IAS_RPS_INOP_DETECTOR );
    if ( !it )
    {
        assert_true_with_message( 0, "No iterator returned for "
            "pixel mask band 2, SCA 5");
    }
    else
    {
        int count = 0;
        int temp;

        /* The inop det list is not sorted but we need it to be.  Swap the
           3rd and 4th entries. */
        temp = detector_list[2];
        detector_list[2] = detector_list[3];
        detector_list[3] = temp;

        while ( ias_pm_get_next_span( it, &span ))
        {
            if ( count == number_of_detectors )
            {
                assert_true_with_message( 0, "Number of spans returned "
                    "is larger than expected number of inoperable detectors %d",
                    number_of_detectors );
                break;
            }
            assert_equal( span.detector_index + 1, detector_list[count] );
            assert_equal( span.starting_pixel_index, 0 );
            assert_equal( span.length_of_span, 16 );
            count++;
        }

        /* Five spans should be returned */
        assert_equal( count, 5 );
    }
    ias_pm_destroy_iterator( it );

    cleanup_memory_management_buffers(ENABLE_WARNING_DISPLAY);
    cleanup_io_routines( blob );
    free_the_blob();
    remove( l1rname );
}
