/***************************************************************************
 NAME:      inoperable_detectors_fill_test

 PURPOSE:   CGreen unit testing for the OLI/TIRS inoperable detector fill
            algorithm

 RETURNS:   CGreen test status

 ROUTINES:  -compare_data_buffers            (internal)
            -load_pixel_mask                 (internal -- algorithm tests)
            -run_inopdetfill_algorithm_test  (internal -- algorithm tests)
            -local_free_the_blob             (internal -- API test)
            -test_inopdetfill01_from_algorithm
            -test_inopdetfill02_from_algorithm
            -test_inopdetfill03_from_algorithm
            -test_inopdetfill_custom_from_algorithm
            -test_inopdetfill01_from_api
****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>
#include "ias_const.h"
#include "ias_cpf.h"
#include "ias_logging.h"
#include "ias_pixel_mask.h"
#include "ias_radiometry.h"
#include "ias_satellite_attributes.h"
#include "../rps_local.h"
#include "../rps_algorithm.h"
#include "cgreen.h"

/* Header file with 'custom' input test data and expected results for
   the 6 cases of detector offsets checked in the
   'test_inopdetfill_custom_from_algorithm' routine  */
#include "inopdetfill_custom_test_data.h"


#define NUMBER_DETECTORS   16
#define NUMBER_LINES       12


/* The assumed detector 'offsets' used for the prototype code unit tests have
   the following values:

   static int prototype_det_offsets[NUMBER_DETECTORS] = {0, 2, 0, 2, 0,
    2, 0, 2, 0, 2, 0, 2, 0, 2, 0, 2};

   In the core algorithm, the starting lines for the replacement detectors
   are calculated as follows:

   left_line = left_det_offset - center_det_offset
   right_line = right_det_offset - center_det_offset

   For example, for detector 2 in prototype test dataset #1, the
   starting lines for the replacement detectors given the assumed offsets
   would be determined as follows:

   left_line = XXX (not used -- no left detector)
   right_line = 0 - 2 = -2 (two lines "above" the first pixel)

   If used as THE spatial offsets (as would be returned from the
   RPS library routine 'ias_rps_convert_detector_delays2spatial_offsets'),
   the test results WILL NOT match the expected prototype results and API
   results (as the API result is generated using prototype dataset #1).
   For example, for inoperable detector 2 in prototype test dataset #1,
   the expected fill given the current algorithm implementation should be
   this:

   detector_1[NUMBER_LINES] = {0.0, 0.0, 6.1734, 6.4147, 7.0, 6.0, 6.1873,
       7.3266, 6.1664, 6.0704, 6.7889, 6.0448};

   rather than this using the assumed offsets:

   detector_1[NUMBER_LINES] = {7.0, 6.0, 6.1873, 7.3266, 6.1664, 6.0704,
       6.7889, 6.0448, 5.6635, 6.4659, 0.0, 0.0};

   However, if the prototype 'offsets' are in fact considered instead to
   be CPF-based detector delays, the corresponding spatial offsets are as
   follows:

   static int prototype_det_offsets[NUMBER_DETECTORS] = {2, 0, 2, 0, 2,
       0, 2, 0, 2, 0, 2, 0, 2, 0, 2, 0};

   In this case, the algorithm and API unit tests WILL match the prototype
   expected results.  */
static int prototype_det_offsets[NUMBER_DETECTORS] = {2, 0, 2, 0, 2, 0,
    2, 0, 2, 0, 2, 0, 2, 0, 2, 0};

static char *test_data_path = NULL;


/***************** COMMON INTERNAL ROUTINES ******************************/

/*------------------------------------------------------------------------
 NAME:     compare_data_buffers

 PURPOSE:  Compares the pixel values in the reference and test data
           buffers

 RETURNS:  The number of differences found
--------------------------------------------------------------------------*/
static int compare_data_buffers
(
    float *test_buffer,                     /* I: Data buffer generated from
                                               running algorithm */
    float *ref_buffer,                      /* I: Data buffer with
                                               'expected' results */
    int   num_detectors,                    /* I: Number of detectors */
    int   num_lines,                        /* I: Number of lines */
    float threshold                         /* I: Threshold for max.
                                               allowable difference in
                                               pixel values */
)
{
    int num_differences = 0;                /* Number of pixels that
                                               are significantly different */
    int line_index;
    int det_index;


    /* Calculate the total number of pixels */
    for (line_index = 0; line_index < num_lines; line_index++)
    {
        for (det_index = 0; det_index < num_detectors; det_index++)
        {
            int offset = line_index * num_detectors + det_index;
            float pixel_diff =
                fabsf(ref_buffer[offset] - test_buffer[offset]);
            if (pixel_diff > threshold)
            {
                num_differences++;
                printf("Line %d, det %d:  test_buffer=%f, "
                "ref_buffer=%f\n", line_index + 1,
                det_index + 1, test_buffer[offset],
                ref_buffer[offset]);
            }
        }
    }

    return num_differences;
}   /* END -- compare_data_buffers */



/***************** API TEST INTERNAL ROUTINES ****************************/
/*------------------------------------------------------------------------
 NAME:     local_free_the_blob

 PURPOSE:  Local override of RPS driver routine to free the blob.  Since
           not all members of the blob are initialized, the normal routine
           cannot be called

 RETURNS:  Nothing
--------------------------------------------------------------------------*/
static void local_free_the_blob(BLOB *blob)
{
    RPS_BAND_ATTRIBUTES *atts = NULL;
    int band_index;
    int sca_index;


    if (blob == NULL)
        return;

    /* Free up the band list arrays */
    free(blob->list_of_bands);
    free(blob->list_of_band_indices);

    /* Free up the allocated data blocks in all bands */
    for (band_index = 0; band_index < blob->total_bands; band_index++)
    {
        if (blob->pixel_mask[band_index] == NULL)
            continue;

        atts = get_band_attributes(band_index);
        for (sca_index = 0; sca_index < atts->number_scas; sca_index++)
        {
            if (blob->pixel_mask[band_index][sca_index] != NULL)
                ias_pm_destroy(blob->pixel_mask[band_index][sca_index]);
        }

        /* Free the band-level pixel mask data blocks */
        free(blob->pixel_mask[band_index]);

        /* Free the band information data blocks */
        free(blob->band_information[band_index]);
    }

    /* Free the base pixel mask and band information data blocks */
    free(blob->pixel_mask);
    free(blob->band_information);

    /* Free the main BLOB data block */
    free(blob);
}   /* END -- local_free_the_blob */



/**************** ALGORITHM TEST INTERNAL ROUTINES ***********************/


/*------------------------------------------------------------------------
 NAME:     load_pixel_mask

 PURPOSE:  Takes the list of inoperable detectors and creates a pixel
           mask where they're populated

 RETURNS:  Integer status code of SUCCESS or ERROR
--------------------------------------------------------------------------*/
static IAS_PIXEL_MASK *load_pixel_mask
(
    int *inoperable_detector_list,             /* I: Inoperable detector
                                                  list */
    int number_inop_detectors,                 /* I: Number of inoperable
                                                  detectors */
    int number_detectors,                      /* I: Total number of
                                                  detectors */
    int number_lines                           /* I: Total number of lines */
)
{
    IAS_PIXEL_MASK *pm = NULL;

    int det_index;
    int status;

    /* Create the pixel mask */
    pm = ias_pm_create(1, 1, number_detectors, number_lines);
    if (pm == NULL)
        return NULL;

    /* Mark the inoperable detectors */
    for (det_index = 0; det_index < number_inop_detectors; det_index++)
    {
        status = ias_pm_add_pixels(pm,
            (inoperable_detector_list[det_index] - 1), 0, number_lines,
            IAS_RPS_INOP_DETECTOR);
        if (status != SUCCESS)
        {
            printf("Error adding inoperative detector %d to pixel mask",
                inoperable_detector_list[det_index]);
            ias_pm_destroy(pm);
            return NULL;
        }
    }

    return pm;
}   /* END -- load_pixel_mask */



/*------------------------------------------------------------------------
 NAME:     run_inopdetfill_algorithm_test

 PURPOSE:  Sets up data buffers and calls the inoperable_detectors_fill
           algorithm

 RETURNS:  Integer status code of SUCCESS or ERROR

 NOTES:    If the offset_type parameter is set to USE_CUSTOM_OFFSETS,
           the detector offsets can be set up however desired.  Make sure
           the 'reference' filled dataset is updated accordingly
--------------------------------------------------------------------------*/
static int run_inopdetfill_algorithm_test
(
    float *test_buffer,                /* I: Input test data
                                          O: Processed test data*/
    float *ref_buffer,                 /* I: Expected results */
    int   *inopdet_list,               /* I: Inoperable detector list */
    int *det_offsets,                  /* I: Desired set of detector
                                          offsets */
    int   number_inoperable_detectors, /* I: Total number of inoperable
                                          detectors */
    int   number_detectors,            /* I: Number of detectors in image */
    int   number_lines                 /* I: Total number of lines of image
                                          data */
)
{
    IAS_PIXEL_MASK *pmask = NULL;
    int      status;
    int      num_major_differences;


    /* Create the test pixel mask */
    pmask = load_pixel_mask(inopdet_list, number_inoperable_detectors,
        number_detectors, number_lines);
    if (pmask == NULL)
    {
        assert_true_with_message(FALSE, "Error creating test pixel mask");
        return ERROR;
    }

    /* Run the test image through processing */
    status = inoperable_detectors_fill(test_buffer, number_detectors,
        number_lines, det_offsets, pmask);
    assert_equal_with_message(status, 0, "Error running inoperable "
        "detector fill correction");
    if (status != SUCCESS)
    {
        ias_pm_destroy(pmask);
        return ERROR;
    }
    /* Compare the test and reference image data buffers */
    num_major_differences = compare_data_buffers(test_buffer,
        ref_buffer, number_detectors, number_lines, 1e-5);
    assert_equal_with_message(num_major_differences, 0, "%d differences "
        "greater than 1e-5 found", num_major_differences);

    /* Free pixel mask resources */
    ias_pm_destroy(pmask);

    return status;
}   /* END -- run_inopdetfill_test */



/***************** ALGORITHM UNIT TESTS ***********************************/
Ensure test_inopdetfill01_from_algorithm()
{
    /* Prototype test dataset #1, contained in the binary file
       '$UNIT_TEST_DATA_DIR/ias_base/rps/driver/inoperable_detectors_fill/
       testim_inopdetfill01_na' */
    float test_buffer1[NUMBER_LINES * NUMBER_DETECTORS] = {1.80726361,
    2, 6.17340422, 8, 6.16524172, 1, 5.30539846, 5, 6.21433163,
    2, 1.55584431, 6, 6.01985121, 4, 1.72543621, 2, 1.52886128, 2.5,
    6.41467619, 8.5, 6.23790312, 1.5, 5.32965851, 5.5, 5.33769655, 2.5,
    3.07041478, 6.5, 5.31004429, 4.5, 2.14186597, 2.5, 2, 1.95226336, 7,
    6.43812513, 8, 2.03341365, 6, 5.01313114, 9, 2.27266526, 1.5,
    6.69284534, 7, 1.56524122, 1.5, 2.02655745, 1, 2.19381142, 6,
    6.40397406, 7, 2.74729466, 5, 5.80406046, 8, 2.12546515, 0.5,
    6.6089077, 6, 1.6026597, 0.5, 1.88149047, 1.94379115, 1.51684272,
    6.18730831, 5.19833326, 6.00959396, 1.46376789, 6.18984175,
    5.66164446, 6.555336, 1.8036263, 2.71748924, 5.25247478, 4.99925232,
    2.34425116, 2.62553644, 1.96863198, 1.74074531, 2.75459313, 7.32660103,
    4.81897402, 5.56998634, 2.91167235, 5.54933405, 5.99198961, 4.94606924,
    1.68897903, 2.61670065, 6.01864147, 6.41774845, 2.79283643, 3.01235247,
    2.63555455, 2.26749134, 2.20191526, 6.16635704, 5.64914989, 5.98852539,
    1.39578545, 5.90160275, 6.25758743, 5.72508144, 1.40476167, 1.62117767,
    6.40143538, 5.82859087, 2.62508941, 1.9805516, 2.11056924, 2.02870393,
    1.78892934, 6.07038546, 6.82595158, 5.69884157, 1.96730423, 6.15347052,
    6.22241497, 6.04713058, 1.0607667, 2.36928916, 6.48692656, 5.76101875,
    1.94219828, 2.4992919, 2.8320117, 2.53290868, 1.16298795, 6.78891754,
    6.11753654, 6.43495083, 1.83658183, 6.08372831, 6.57044935, 5.98089027,
    1.78799582, 1.44281936, 6.7803669, 5.55544949, 1.33411515, 1.62133312,
    1.97852004, 2.8103683, 1.65618491, 6.04477406, 5.92412376, 5.71459389,
    1.4254247, 6.89403963, 6.22384071, 6.94146347, 2.38862085, 1.1470654,
    6.79312468, 6.63171434, 0.828597188, 2.29807258, 1.93338227,
    2.06092691, 1.48640728, 5.66350794, 5.92205572, 6.65233994,
    1.34597909, 5.68795872, 6.15772009, 6.02773046, 1.64303696,
    2.33058119, 6.42814875, 6.1916008, 1.53671837, 3.06160855, 2.39660001,
    1.38097954, 1.75372124, 6.46594667, 6.51908255, 5.97867775, 2.3231914,
    5.92501974, 6.47279596, 5.69305515, 2.79231977, 1.13517904,
    5.28775835, 5.94056654, 2.56478381, 2.65586257, 2.2131362};

    /* Prototype result dataset #1, contained in the binary file
       '$UNIT_TEST_DATA_DIR/ias_base/rps/driver/inoperable_detectors_fill/
       resultim_inopdetfill01_na' */
    float ref_buffer1[NUMBER_LINES * NUMBER_DETECTORS] = {6.17340422, 7,
    6.17340422, 8, 6.16524172, 7, 5.30539846, 5, 6.21433163,
    8, 2.0714438, 6, 6.01985121, 0, 0, 0, 6.41467619, 6, 6.41467619, 8.5,
    6.23790312, 6, 5.32965851, 5.5, 5.33769655, 7.5, 1.77923214, 6.5,
    5.31004429, 0, 0, 0, 7, 6.18730831, 7, 6.43812513, 8, 6.09971809, 6,
    5.01313114, 9, 6.60117245, 7, 6.69284534, 7, 0, 0, 0, 6, 7.32660103, 6,
    6.40397406, 7, 5.55966043, 5, 5.80406046, 8, 5.50034857, 7, 6.6089077,
    6, 0, 0, 0, 6.18730831, 6.16635704, 6.18730831, 5.19833326, 6.00959396,
    5.94506407, 6.18984175, 5.66164446, 6.555336, 5.56754589, 6.6470089,
    5.25247478, 4.99925232, 0, 0, 0, 7.32660103, 6.07038546, 7.32660103,
    4.81897402, 5.56998634, 5.92615604, 5.54933405, 5.99198961, 4.94606924,
    6.0376339, 6.05462837, 6.01864147, 6.41774845, 0, 0, 0, 6.16635704,
    6.78891754, 6.16635704, 5.64914989, 5.98852539, 6.25933981, 5.90160275,
    6.25758743, 5.72508144, 6.12107182, 5.41001034, 6.40143538,
    5.82859087, 0, 0, 0, 6.07038546, 6.04477406, 6.07038546, 6.82595158,
    5.69884157, 6.30431652, 6.15347052, 6.22241497, 6.04713058,
    6.78995132, 6.02813768, 6.48692656, 5.76101875, 0, 0, 0, 6.78891754,
    5.66350794, 6.78891754, 6.11753654, 6.43495083, 6.17014933,
    6.08372831, 6.57044935, 5.98089027, 6.27860928, 6.26125383, 6.7803669,
    5.55544949, 0, 0, 0, 6.04477406, 6.46594667, 6.04477406, 5.92412376,
    5.71459389, 5.95184851, 6.89403963, 6.22384071, 6.94146347,
    6.05974483, 6.63843918, 6.79312468, 6.63171434, 0, 0, 0, 5.66350794,
    0, 5.66350794, 5.92205572, 6.65233994, 0, 5.68795872, 6.15772009,
    6.02773046, 2.14271641, 6.52948809, 6.42814875, 6.1916008, 0, 0, 0,
    6.46594667, 0, 6.46594667, 6.51908255, 5.97867775, 0, 5.92501974,
    6.47279596, 5.69305515, 1.76258612, 6.42643499, 5.28775835,
    5.94056654, 0, 0, 0};

    int inop_list1[8] = {1, 2, 6, 10, 11, 14, 15, 16};
    int num_inop_detectors = 8;
    int status;


    printf("Verifying inoperable detector fill algorithm for prototype test "
        "image #1...\n");

    /* Run dataset inopdetfill01 */
    status = run_inopdetfill_algorithm_test(test_buffer1, ref_buffer1,
        inop_list1, prototype_det_offsets, num_inop_detectors,
        NUMBER_DETECTORS, NUMBER_LINES);
    assert_equal_with_message(status, 0, "Inoperable detector fill test "
        "failed");

}   /* END -- test_inopdetfill01_from_algorithm */



Ensure test_inopdetfill02_from_algorithm()
{
    /* Prototype test dataset #2, contained in the binary file
       '$UNIT_TEST_DATA_DIR/ias_base/rps/driver/inoperable_detectors_fill/
       testim_inopdetfill02_na' */
    float test_buffer2[NUMBER_LINES * NUMBER_DETECTORS] = {1.80726361,
    2, 2.17340422, 4, 6.16524172, 5, 5.30539846, 1, 6.21433163,
    6, 1.55584431, 6, 6.01985121, 8, 5.72543621, 2, 1.52886128, 2.5,
    2.41467619, 4.5, 6.23790312, 5.5, 5.32965851, 1.5, 5.33769655, 6.5,
    3.07041478, 6.5, 5.31004429, 8.5, 6.14186573, 2.5, 2, 1.95226336, 3,
    2.43812513, 8, 6.03341389, 6, 1.01313126, 9, 6.2726655, 1.5,
    6.69284534, 7, 5.56524134, 5.5, 2.02655745, 1, 2.19381142, 2,
    2.40397406, 7, 6.7472949, 5, 1.8040607, 8, 6.12546539, 0.5,
    6.6089077, 6, 5.6026597, 4.5, 1.88149047, 1.94379115, 1.51684272,
    2.18730831, 1.19833302, 6.00959396, 5.46376801, 6.18984175,
    1.66164434, 6.555336, 5.80362606, 2.71748924, 5.25247478, 4.99925232,
    6.34425116, 6.62553644, 1.96863198, 1.74074531, 2.75459313, 3.32660103,
    0.818974078, 5.56998634, 6.91167212, 5.54933405, 1.99198949,
    4.94606924, 5.68897915, 2.61670065, 6.01864147, 6.41774845,
    6.79283667, 7.01235247, 2.63555455, 2.26749134, 2.20191526,
    2.16635704, 1.64914989, 5.98852539, 5.39578533, 5.90160275,
    2.25758743, 5.72508144, 5.40476179, 1.62117767, 6.40143538,
    5.82859087, 6.62508965, 5.98055172, 2.11056924, 2.02870393,
    1.78892934, 2.07038546, 2.82595158, 5.69884157, 5.96730423,
    6.15347052, 2.22241473, 6.04713058, 5.0607667, 2.36928916,
    6.48692656, 5.76101875, 5.94219828, 6.4992919, 2.8320117,
    2.53290868, 1.16298795, 2.78891754, 2.11753631, 6.43495083,
    5.83658171, 6.08372831, 2.57044911, 5.98089027, 5.78799582,
    1.44281936, 6.7803669, 5.55544949, 5.33411503, 5.62133312,
    1.97852004, 2.8103683, 1.65618491, 2.04477382, 1.92412364,
    5.71459389, 5.42542458, 6.89403963, 2.22384048, 6.94146347,
    6.38862085, 1.1470654, 6.79312468, 6.63171434, 4.82859707,
    6.29807234, 1.93338227, 2.06092691, 1.48640728, 1.66350818,
    1.92205572, 6.65233994, 5.34597921, 5.68795872, 2.15772009,
    6.02773046, 5.64303684, 2.33058119, 6.42814875, 6.1916008,
    5.53671837, 7.06160831, 2.39660001, 1.38097954, 1.75372124,
    2.46594667, 2.51908231, 5.97867775, 6.32319164, 5.92501974,
    2.47279572, 5.69305515, 6.79231977, 1.13517904, 5.28775835,
    5.94056654, 6.56478357, 6.65586281, 2.2131362};

    /* Prototype result dataset #2, contained in the binary file
       '$UNIT_TEST_DATA_DIR/ias_base/rps/driver/inoperable_detectors_fill/
       resultim_inopdetfill02_na' */
    float ref_buffer2[NUMBER_LINES * NUMBER_DETECTORS] = {0, 0, 0, 0,
    6.16524172, 5, 5.30539846, 7.5, 6.21433163, 6, 0, 6, 6.01985121, 8,
    5.72543621, 5.5, 0, 0, 0, 0, 6.23790312, 5.5, 5.32965851, 6.5,
    5.33769655, 6.5, 0, 6.5, 5.31004429, 8.5, 6.14186573, 4.5, 0, 0, 0, 0,
    8, 6.03341389, 6, 6.37258911, 9, 6.2726655, 6, 6.69284534, 7,
    5.56524134, 5.5, 6.62553644, 0, 0, 0, 0, 7, 6.7472949, 5, 5.24770164,
    8, 6.12546539, 6.5, 6.6089077, 6, 5.6026597, 4.5, 7.01235247, 0, 0, 0,
    0, 6.00959396, 5.46376801, 6.18984175, 5.81334209, 6.555336,
    5.80362606, 6.48275518, 5.25247478, 4.99925232, 6.34425116,
    6.62553644, 5.98055172, 0, 0, 0, 0, 5.56998634, 6.91167212,
    5.54933405, 6.10030079, 4.94606924, 5.68897915, 6.36718655,
    6.01864147, 6.41774845, 6.79283667, 7.01235247, 6.4992919, 0, 0, 0, 0,
    5.98852539, 5.39578533, 5.90160275, 6.03230906, 5.72508144,
    5.40476179, 5.52805042, 6.40143538, 5.82859087, 6.62508965,
    5.98055172, 5.62133312, 0, 0, 0, 0, 5.69884157, 5.96730423,
    6.15347052, 6.91775131, 6.04713058, 5.0607667, 5.85381031, 6.48692656,
    5.76101875, 5.94219828, 6.4992919, 6.29807234, 0, 0, 0, 0, 6.43495083,
    5.83658171, 6.08372831, 5.85784435, 5.98089027, 5.78799582,
    5.90309858, 6.7803669, 5.55544949, 5.33411503, 5.62133312, 7.06160831,
    0, 0, 0, 0, 5.71459389, 5.42542458, 6.89403963, 5.80903721,
    6.94146347, 6.38862085, 5.77384663, 6.79312468, 6.63171434,
    4.82859707, 6.29807234, 6.65586281, 0, 0, 0, 0, 6.65233994,
    5.34597921, 5.68795872, 0, 6.02773046, 5.64303684, 6.28418159,
    6.42814875, 6.1916008, 5.53671837, 7.06160831, 0, 0, 0, 0, 0,
    5.97867775, 6.32319164, 5.92501974, 0, 5.69305515, 6.79231977,
    6.59087276, 5.28775835, 5.94056654, 6.56478357, 6.65586281, 0};

    int inop_list2[7] = {1, 2, 3, 4, 8, 11, 16};
    int num_inop_detectors = 7;
    int status;


    printf("Verifying inoperable detector fill algorithm for prototype test "
        "image #2...\n");

    /* Run dataset inopdetfill02 */
    status = run_inopdetfill_algorithm_test(test_buffer2, ref_buffer2,
        inop_list2, prototype_det_offsets, num_inop_detectors,
        NUMBER_DETECTORS, NUMBER_LINES);
    assert_equal_with_message(status, 0, "Inoperable detector fill test "
        "failed");

}   /* END -- test_inopdetfill02_from_algorithm */



Ensure test_inopdetfill03_from_algorithm()
{
    /* Prototype test dataset #3, contained in the binary file
       '$UNIT_TEST_DATA_DIR/ias_base/rps/driver/inoperable_detectors_fill/
       testim_inopdetfill03_na' */
    float test_buffer3[NUMBER_LINES * NUMBER_DETECTORS] = {1.80726361,
    6, 6.17340422, 8, 2.16524196, 1, 1.30539823, 1, 6.21433163,
    6, 5.55584431, 2, 6.01985121, 8, 1.72543621, 2, 1.52886128, 6.5,
    6.41467619, 8.5, 2.23790336, 1.5, 1.32965875, 1.5, 5.33769655, 6.5,
    7.07041454, 2.5, 5.31004429, 8.5, 2.14186597, 2.5, 2, 5.95226336, 7,
    6.43812513, 4, 2.03341365, 2, 1.01313126, 9, 6.2726655, 5.5,
    2.69284534, 7, 5.56524134, 1.5, 2.02655745, 1, 6.19381142, 6,
    6.40397406, 3, 2.74729466, 1, 1.8040607, 8, 6.12546539, 4.5,
    2.60890794, 6, 5.6026597, 0.5, 1.88149047, 1.94379115, 5.51684284,
    6.18730831, 5.19833326, 2.0095942, 1.46376789, 2.18984199, 1.66164434,
    6.555336, 5.80362606, 6.71748924, 1.25247455, 4.99925232, 6.34425116,
    2.62553644, 1.96863198, 1.74074531, 6.75459337, 7.32660103,
    4.81897402, 1.56998634, 2.91167235, 1.54933417, 1.99198949,
    4.94606924, 5.68897915, 6.61670065, 2.01864147, 6.41774845,
    6.79283667, 3.01235247, 2.63555455, 2.26749134, 6.20191526,
    6.16635704, 5.64914989, 1.98852527, 1.39578545, 1.90160286,
    2.25758743, 5.72508144, 5.40476179, 5.62117767, 2.40143538,
    5.82859087, 6.62508965, 1.9805516, 2.11056924, 2.02870393, 5.78892946,
    6.07038546, 6.82595158, 1.69884133, 1.96730423, 2.15347075,
    2.22241473, 6.04713058, 5.0607667, 6.36928892, 2.48692679, 5.76101875,
    5.94219828, 2.4992919, 2.8320117, 2.53290868, 5.16298771, 6.78891754,
    6.11753654, 2.43495083, 1.83658183, 2.08372831, 2.57044911, 5.98089027,
    5.78799582, 5.4428196, 2.7803669, 5.55544949, 5.33411503, 1.62133312,
    1.97852004, 2.8103683, 5.65618515, 6.04477406, 5.92412376, 1.71459377,
    1.4254247, 2.89403939, 2.22384048, 6.94146347, 6.38862085, 5.14706564,
    2.79312468, 6.63171434, 4.82859707, 2.29807258, 1.93338227, 2.06092691,
    5.48640728, 5.66350794, 5.92205572, 2.6523397, 1.34597909, 1.68795872,
    2.15772009, 6.02773046, 5.64303684, 6.33058119, 2.42814898, 6.1916008,
    5.53671837, 3.06160855, 2.39660001, 1.38097954, 5.75372124, 6.46594667,
    6.51908255, 1.97867787, 2.3231914, 1.9250195, 2.47279572, 5.69305515,
    6.79231977, 5.13517904, 1.28775847, 5.94056654, 6.56478357, 2.65586257,
    2.2131362};

    /* Prototype filled dataset #3, contained in the binary file
       '$UNIT_TEST_DATA_DIR/ias_base/rps/driver/inoperable_detectors_fill/
       resultim_inopdetfill03_na' */
    float ref_buffer3[NUMBER_LINES * NUMBER_DETECTORS] = {0, 6,
    6.17340422, 8, 0, 0, 0, 0, 6.21433163, 6, 5.55584431, 6.25, 6.01985121,
    8, 0, 8, 0, 6.5, 6.41467619, 8.5, 0, 0, 0, 0, 5.33769655, 6.5, 7.07041454,
    5.25, 5.31004429, 8.5, 0, 8.5, 6, 5.95226336, 7, 6.43812513, 0, 0, 0, 0,
    9, 6.2726655, 5.5, 5.85837078, 7, 5.56524134, 8, 5.56524134, 6.5,
    6.19381142, 6, 6.40397406, 0, 0, 0, 0, 8, 6.12546539, 4.5, 6.51722431,
    6, 5.6026597, 8.5, 5.6026597, 5.95226336, 5.51684284, 6.18730831,
    5.19833326, 0, 0, 0, 0, 6.555336, 5.80362606, 6.71748924, 5.72488451,
    4.99925232, 6.34425116, 5.56524134, 6.34425116, 6.19381142,
    6.75459337, 7.32660103, 4.81897402, 0, 0, 0, 0, 4.94606924,
    5.68897915, 6.61670065, 6.06515408, 6.41774845, 6.79283667, 5.6026597,
    6.79283667, 5.51684284, 6.20191526, 6.16635704, 5.64914989, 0, 0, 0,
    0, 5.72508144, 5.40476179, 5.62117767, 5.49913454, 5.82859087,
    6.62508965, 6.34425116, 6.62508965, 6.75459337, 5.78892946,
    6.07038546, 6.82595158, 0, 0, 0, 0, 6.04713058, 5.0607667,
    6.36928892, 5.88938999, 5.76101875, 5.94219828, 6.79283667,
    5.94219828, 6.20191526, 5.16298771, 6.78891754, 6.11753654, 0, 0, 0,
    0, 5.98089027, 5.78799582, 5.4428196, 6.26109076, 5.55544949,
    5.33411503, 6.62508965, 5.33411503, 5.78892946, 5.65618515,
    6.04477406, 5.92412376, 0, 0, 0, 0, 6.94146347, 6.38862085,
    5.14706564, 5.53787279, 6.63171434, 4.82859707, 5.94219828,
    4.82859707, 5.16298771, 5.48640728, 5.66350794, 5.92205572, 0,
    0, 0, 0, 6.02773046, 5.64303684, 6.33058119, 0, 6.1916008, 5.53671837,
    5.33411503, 5.53671837, 5.65618515, 5.75372124, 6.46594667,
    6.51908255, 0, 0, 0, 0, 5.69305515, 6.79231977, 5.13517904, 0,
    5.94056654, 6.56478357, 4.82859707, 6.56478357};

    int inop_list3[8] = {1, 5, 6, 7, 8, 12, 15, 16};
    int num_inop_detectors = 8;
    int status;


    printf("Verifying inoperable detector fill algorithm for prototype test "
        "image #3...\n");

    /* Run dataset inopdetfill03 */
    status = run_inopdetfill_algorithm_test(test_buffer3, ref_buffer3,
        inop_list3, prototype_det_offsets, num_inop_detectors,
        NUMBER_DETECTORS, NUMBER_LINES);
    assert_equal_with_message(status, 0, "Inoperable detector fill test "
        "failed");

}   /* END -- test_inopdetfill03_from_algorithm */



/* This test checks the algorithm in the case that an inoperable detector
   and its associated replacement detectors all have non-zero spatial
   offsets.  Detector 6 is the inoperable detector, detectors 5 and 7
   are the replacement data detectors  */
Ensure test_inopdetfill_custom_from_algorithm()
{
    float test_buffer[72];     /* Working test data buffer; the input
                                  data buffer contents are (re)copied
                                  here for each test */

    int det_offsets1[8] = {0, 2, 0, 2, 2, 1, 3, 2};
    int det_offsets2[8] = {0, 2, 0, 2, 3, 1, 2, 2};
    int det_offsets3[8] = {0, 2, 0, 2, 1, 3, 2, 2};
    int det_offsets4[8] = {0, 2, 0, 2, 2, 3, 1, 2};
    int det_offsets5[8] = {0, 2, 0, 2, 3, 2, 1, 2};
    int det_offsets6[8] = {0, 2, 0, 2, 1, 2, 3, 2};

    int inop_list4[1] = {6};
    int num_inop_detectors = 1;
    int num_detectors = 8;
    int num_lines = 9;
    int status;


    printf("Verifying inoperable detector fill algorithm for a custom "
           "9 line x 8 detector test data set using consecutive nonzero "
           "offsets for detectors 5, 6, and 7.  Detector 6 is to be "
           "considered inoperable...\n");

    /* Run the data sets */
    memcpy(test_buffer, input_buffer, num_detectors*num_lines*sizeof(float));
    status = run_inopdetfill_algorithm_test(test_buffer, ref_buffer1_custom,
        inop_list4, det_offsets1, num_inop_detectors,
        num_detectors, num_lines);
    assert_equal_with_message(status, 0, "Inoperable detector fill test "
        "failed");

    memcpy(test_buffer, input_buffer, num_detectors*num_lines*sizeof(float));
    status = run_inopdetfill_algorithm_test(test_buffer, ref_buffer2_custom,
        inop_list4, det_offsets2, num_inop_detectors,
        num_detectors, num_lines);
    assert_equal_with_message(status, 0, "Inoperable detector fill test "
        "failed");

    memcpy(test_buffer, input_buffer, num_detectors*num_lines*sizeof(float));
    status = run_inopdetfill_algorithm_test(test_buffer, ref_buffer3_custom,
        inop_list4, det_offsets3, num_inop_detectors,
        num_detectors, num_lines);
    assert_equal_with_message(status, 0, "Inoperable detector fill test "
        "failed");

    memcpy(test_buffer, input_buffer, num_detectors*num_lines*sizeof(float));
    status = run_inopdetfill_algorithm_test(test_buffer, ref_buffer4_custom,
        inop_list4, det_offsets4, num_inop_detectors,
        num_detectors, num_lines);
    assert_equal_with_message(status, 0, "Inoperable detector fill test "
        "failed");

    memcpy(test_buffer, input_buffer, num_detectors*num_lines*sizeof(float));
    status = run_inopdetfill_algorithm_test(test_buffer, ref_buffer5_custom,
        inop_list4, det_offsets5, num_inop_detectors,
        num_detectors, num_lines);
    assert_equal_with_message(status, 0, "Inoperable detector fill test "
        "failed");

    memcpy(test_buffer, input_buffer, num_detectors*num_lines*sizeof(float));
    status = run_inopdetfill_algorithm_test(test_buffer, ref_buffer6_custom,
        inop_list4, det_offsets6, num_inop_detectors,
        num_detectors, num_lines);
    assert_equal_with_message(status, 0, "Inoperable detector fill test "
        "failed");

}   /* END -- test_inopdetfill_custom_from_algorithm */



/************************* API UNIT TEST ***********************************/
Ensure test_inopdetfill01_from_api()
{
    float test_data[NUMBER_DETECTORS * NUMBER_LINES] = {1.80726361, 2,
    6.17340422, 8, 6.16524172, 1, 5.30539846, 5, 6.21433163,
    2, 1.55584431, 6, 6.01985121, 4, 1.72543621, 2, 1.52886128, 2.5,
    6.41467619, 8.5, 6.23790312, 1.5, 5.32965851, 5.5, 5.33769655, 2.5,
    3.07041478, 6.5, 5.31004429, 4.5, 2.14186597, 2.5, 2, 1.95226336, 7,
    6.43812513, 8, 2.03341365, 6, 5.01313114, 9, 2.27266526, 1.5,
    6.69284534, 7, 1.56524122, 1.5, 2.02655745, 1, 2.19381142, 6,
    6.40397406, 7, 2.74729466, 5, 5.80406046, 8, 2.12546515, 0.5,
    6.6089077, 6, 1.6026597, 0.5, 1.88149047, 1.94379115, 1.51684272,
    6.18730831, 5.19833326, 6.00959396, 1.46376789, 6.18984175,
    5.66164446, 6.555336, 1.8036263, 2.71748924, 5.25247478, 4.99925232,
    2.34425116, 2.62553644, 1.96863198, 1.74074531, 2.75459313, 7.32660103,
    4.81897402, 5.56998634, 2.91167235, 5.54933405, 5.99198961, 4.94606924,
    1.68897903, 2.61670065, 6.01864147, 6.41774845, 2.79283643, 3.01235247,
    2.63555455, 2.26749134, 2.20191526, 6.16635704, 5.64914989, 5.98852539,
    1.39578545, 5.90160275, 6.25758743, 5.72508144, 1.40476167, 1.62117767,
    6.40143538, 5.82859087, 2.62508941, 1.9805516, 2.11056924, 2.02870393,
    1.78892934, 6.07038546, 6.82595158, 5.69884157, 1.96730423, 6.15347052,
    6.22241497, 6.04713058, 1.0607667, 2.36928916, 6.48692656, 5.76101875,
    1.94219828, 2.4992919, 2.8320117, 2.53290868, 1.16298795, 6.78891754,
    6.11753654, 6.43495083, 1.83658183, 6.08372831, 6.57044935, 5.98089027,
    1.78799582, 1.44281936, 6.7803669, 5.55544949, 1.33411515, 1.62133312,
    1.97852004, 2.8103683, 1.65618491, 6.04477406, 5.92412376, 5.71459389,
    1.4254247, 6.89403963, 6.22384071, 6.94146347, 2.38862085, 1.1470654,
    6.79312468, 6.63171434, 0.828597188, 2.29807258, 1.93338227,
    2.06092691, 1.48640728, 5.66350794, 5.92205572, 6.65233994,
    1.34597909, 5.68795872, 6.15772009, 6.02773046, 1.64303696,
    2.33058119, 6.42814875, 6.1916008, 1.53671837, 3.06160855, 2.39660001,
    1.38097954, 1.75372124, 6.46594667, 6.51908255, 5.97867775, 2.3231914,
    5.92501974, 6.47279596, 5.69305515, 2.79231977, 1.13517904,
    5.28775835, 5.94056654, 2.56478381, 2.65586257, 2.2131362};

    float ref_l1r_buffer[NUMBER_DETECTORS * NUMBER_LINES] = {6.17340422, 7,
    6.17340422, 8, 6.16524172, 7, 5.30539846, 5, 6.21433163,
    8, 2.0714438, 6, 6.01985121, 0, 0, 0, 6.41467619, 6, 6.41467619, 8.5,
    6.23790312, 6, 5.32965851, 5.5, 5.33769655, 7.5, 1.77923214, 6.5,
    5.31004429, 0, 0, 0, 7, 6.18730831, 7, 6.43812513, 8, 6.09971809, 6,
    5.01313114, 9, 6.60117245, 7, 6.69284534, 7, 0, 0, 0, 6, 7.32660103, 6,
    6.40397406, 7, 5.55966043, 5, 5.80406046, 8, 5.50034857, 7, 6.6089077,
    6, 0, 0, 0, 6.18730831, 6.16635704, 6.18730831, 5.19833326, 6.00959396,
    5.94506407, 6.18984175, 5.66164446, 6.555336, 5.56754589, 6.6470089,
    5.25247478, 4.99925232, 0, 0, 0, 7.32660103, 6.07038546, 7.32660103,
    4.81897402, 5.56998634, 5.92615604, 5.54933405, 5.99198961, 4.94606924,
    6.0376339, 6.05462837, 6.01864147, 6.41774845, 0, 0, 0, 6.16635704,
    6.78891754, 6.16635704, 5.64914989, 5.98852539, 6.25933981, 5.90160275,
    6.25758743, 5.72508144, 6.12107182, 5.41001034, 6.40143538,
    5.82859087, 0, 0, 0, 6.07038546, 6.04477406, 6.07038546, 6.82595158,
    5.69884157, 6.30431652, 6.15347052, 6.22241497, 6.04713058,
    6.78995132, 6.02813768, 6.48692656, 5.76101875, 0, 0, 0, 6.78891754,
    5.66350794, 6.78891754, 6.11753654, 6.43495083, 6.17014933,
    6.08372831, 6.57044935, 5.98089027, 6.27860928, 6.26125383, 6.7803669,
    5.55544949, 0, 0, 0, 6.04477406, 6.46594667, 6.04477406, 5.92412376,
    5.71459389, 5.95184851, 6.89403963, 6.22384071, 6.94146347,
    6.05974483, 6.63843918, 6.79312468, 6.63171434, 0, 0, 0, 5.66350794,
    0, 5.66350794, 5.92205572, 6.65233994, 0, 5.68795872, 6.15772009,
    6.02773046, 2.14271641, 6.52948809, 6.42814875, 6.1916008, 0, 0, 0,
    6.46594667, 0, 6.46594667, 6.51908255, 5.97867775, 0, 5.92501974,
    6.47279596, 5.69305515, 1.76258612, 6.42643499, 5.28775835,
    5.94056654, 0, 0, 0};

    char cpf_filename[PATH_MAX];
    char cpf_dir[]        = "ias_base/rps/driver/inoperable_detectors_fill";
    char cpf_name[]       = "L8CPF20111001_20111231.01";
    const IAS_BAND_ATTRIBUTES *band_atts = NULL;
    float *test_l1r_buffer;
    BLOB *blob;
    int *normalized_detector_offsets;
    int  inoperable_detector_list[8] = {1, 2, 6, 10, 11, 14, 15, 16};
    int  num_inoperable_detectors = 8;
    int  num_major_differences = 0;
    int  num_bands = 30;
    int  max_num_scas = 14;
    int  band_index;
    int  band_number;
    int  curr_band_number;
    int  curr_band_index;
    int  status;
    int  i;            /* Generic loop counter */
    int  test_bands[1] = {1};
    int  test_sca_number = 1;
    int  test_sca_index = test_sca_number - 1;
    int  sca_index;


    printf("Verifying inoperable detector fill algorithm called from "
        "RPS driver...\n");

    /* Initialize the satellite attributes */
    status = ias_sat_attr_initialize(IAS_L8);
    if (status == ERROR)
    {
        assert_true_with_message(FALSE, "Cannot initialize satellite "
            "attributes");
        return;
    }

    /* Allocate the BLOB and initialize it */
    blob = allocate_blob();
    if (blob == NULL)
    {
        assert_true_with_message(FALSE, "Cannot allocate base BLOB "
            "data block.");
        return;
    }
    blob->total_bands = num_bands;
    blob->max_number_scas = max_num_scas;

    /* Allocate the base pixel mask data block within the BLOB */
    blob->pixel_mask = calloc(blob->total_bands, sizeof(IAS_PIXEL_MASK **));
    if (blob->pixel_mask == NULL)
    {
        assert_true_with_message(FALSE, "Cannot allocate BLOB pixel mask data "
            "block.");
        local_free_the_blob(blob);
        return;
    }

    /* Allocate the band information blocks within the BLOB, then
       initialize them */
    blob->number_of_bands = 1;
    blob->list_of_bands = calloc(blob->number_of_bands, sizeof(int));
    blob->list_of_band_indices = calloc(blob->number_of_bands, sizeof(int));
    if ((blob->list_of_bands == NULL)
        || (blob->list_of_band_indices == NULL))
    {
        assert_true_with_message(FALSE, "Error allocating BLOB band "
            "list arrays");
        local_free_the_blob(blob);
        return;
    }
    for (i = 0; i < blob->number_of_bands; i++)
        blob->list_of_bands[i] = test_bands[i];

    blob->band_information =
        calloc(num_bands, sizeof(RPS_BAND_ATTRIBUTES *));
    if (blob->band_information == NULL)
    {
        assert_true_with_message(FALSE, "Cannot allocate band information "
            "data blocks within main BLOB structure.");
        local_free_the_blob(blob);
        return;
    }
    for (band_index = 0; band_index < num_bands; band_index++)
    {
        band_number =
            ias_sat_attr_convert_band_index_to_number(band_index);
        if (band_number== ERROR)
        {
            assert_true_with_message(FALSE, "Cannot obtain band number for "
                "index %d.", band_index);
            local_free_the_blob(blob);
            return;
        }
        blob->band_information[band_index] =
            calloc(1, sizeof(RPS_BAND_ATTRIBUTES));
        if (blob->band_information[band_index] == NULL)
        {
            assert_true_with_message(FALSE, "Cannot allocate band %d "
                "information data block.", band_number);
            local_free_the_blob(blob);
            return;
        }

        /* Initialize the attributes */
        band_atts = ias_sat_attr_get_band_attributes(band_number);
        blob->band_information[band_index]->band_number = band_number;
        blob->band_information[band_index]->number_lines = NUMBER_LINES;
        blob->band_information[band_index]->vrp_band_number =
            band_atts->vrp_band_number;
        if (band_atts->vrp_band_number != 0)
        {
            blob->band_information[band_index]->vrp_band_index =
                ias_sat_attr_convert_band_number_to_index(
                band_atts->vrp_band_number);
        }
        blob->band_information[band_index]->blind_band_number =
            band_atts->blind_band_number;
        if (band_atts->blind_band_number != 0)
        {
            blob->band_information[band_index]->blind_band_index =
                ias_sat_attr_convert_band_number_to_index(
                band_atts->blind_band_number);
            blob->band_information[band_index]->process_blind_band_flag = 1;
        }
        blob->band_information[band_index]->number_scas = band_atts->scas;
        blob->band_information[band_index]->number_detectors =
            NUMBER_DETECTORS;
        blob->band_information[band_index]->band_classification =
            band_atts->band_classification;
        blob->band_information[band_index]->spectral_type =
            band_atts->spectral_type;
        if (band_atts->band_classification == IAS_NORMAL_BAND)
            blob->band_information[band_index]->is_normal_band = 1;

        /* Create the list of band numbers and indices to process */
        for (i = 0; i < blob->number_of_bands; i++)
        {
            if (band_number == blob->list_of_bands[i])
            {
                blob->list_of_band_indices[i] = band_index;
                break;
            }
        }
    }
    /* Make sure we indicate the band field information has in fact been
       initialized */
    if (!blob->band_fields_initialized)
        blob->band_fields_initialized = 1;

    /* Create a set of empty pixel masks in the BLOB */
    status = create_blob_pixel_mask();
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error creating empty pixel mask "
            "set in BLOB");
        local_free_the_blob(blob);
        return;
    }

    /* Populate the band 1 SCA 1 mask with the inoperable detector
       information given above */
    curr_band_number = blob->list_of_bands[0];
    curr_band_index  = blob->list_of_band_indices[0];

    status = ias_rps_inoperable_detector_identification(
        (const int *)inoperable_detector_list, num_inoperable_detectors,
        curr_band_number, test_sca_number, NUMBER_DETECTORS, NUMBER_LINES,
        blob->pixel_mask[curr_band_index][test_sca_index]);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error populating inoperable "
            "detector information into desired pixel mask");
        local_free_the_blob(blob);
        return;
    }

    /* Open the CPF and load it into the BLOB */
    status = snprintf(cpf_filename, sizeof(cpf_filename), "%s/%s/%s",
        test_data_path, cpf_dir, cpf_name);
    if ((status < 0) || (status >= sizeof(cpf_filename)))
    {
        assert_true_with_message(FALSE, "Error constructing CPF file name.");
        local_free_the_blob(blob);
        return;
    }
    blob->cpf = ias_cpf_read(cpf_filename);
    if (blob->cpf == NULL)
    {
        assert_true_with_message(FALSE, "Error opening CPF file %s.",
            cpf_filename);
        local_free_the_blob(blob);
        return;
    }

    /* Initialize the memory management system */
    status = initialize_memory_management_buffers();
    assert_equal_with_message(status, SUCCESS, "Error initializing "
        "memory management");
    if (status != SUCCESS)
    {
        ias_cpf_free(blob->cpf);
        local_free_the_blob(blob);
        return;
    }
    
    /* Initialize the RPS level arrays */
    status = initialize_rps_level(num_bands, max_num_scas);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error initializing RPS level");
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        ias_cpf_free(blob->cpf);
        local_free_the_blob(blob);
    }

    /* Call the preprocessing routine to do initial setup */
    status = inoperable_detectors_fill_preprocess(blob);
    assert_equal_with_message(status, SUCCESS, "Error in 'inoperable_"
        "detectors_fill_preprocess' step");
    if (status != SUCCESS)
    {
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        ias_cpf_free(blob->cpf);
        local_free_the_blob(blob);
        return;
    }

    /* Register the image data buffer and the detector offsets data buffer
       so they don't get destroyed before their output can be modified */
    status = increment_consumer_count(IMAGE_L1R);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error in 'inoperable_detectors_"
            "fill_preprocess' step");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        ias_cpf_free(blob->cpf);
        local_free_the_blob(blob);
        return;
    }

    /* Get an empty image data buffer for band 1, SCA 1 */
    test_l1r_buffer = (float *)preprocess_get_memory_buffer(IMAGE_L1R,
        curr_band_index, test_sca_index);
    if (test_l1r_buffer == NULL)
    {
        assert_true_with_message(FALSE, "Error retrieving L1R image buffer");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        ias_cpf_free(blob->cpf);
        local_free_the_blob(blob);
        return;
    }
    set_buffer_filled_flag( IMAGE_L1R, curr_band_index, test_sca_index );

    /* It is important that the consumer counts get set *after* all the
       increments */
    status = set_consumer_counts();
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error setting consumer counts");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        ias_cpf_free(blob->cpf);
        local_free_the_blob(blob);
        return;
    }

    /* Copy the test data into the empty memory buffer */
    memcpy(test_l1r_buffer, test_data,
        (NUMBER_DETECTORS * NUMBER_LINES * sizeof(float)));

    /* Get the normalized detector offsets for band 1, SCA 1 */
    normalized_detector_offsets = (int *)get_memory_buffer(
        MM_SPATIAL_DETECTOR_OFFSETS, CONSUMER, blob->list_of_band_indices[0],
        test_sca_index);
    if (normalized_detector_offsets == NULL)
    {
        assert_true_with_message(FALSE, "Error retrieving normalized "
            "detector offsets data buffer");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        ias_cpf_free(blob->cpf);
        local_free_the_blob(blob);
        return;
    }

    /* Process the test band */
    status = api_inoperable_detectors_fill(blob, curr_band_index,
        test_sca_index, 0);
    assert_equal_with_message(status, SUCCESS, "Should have exited with "
        "SUCCESS having completed processing");
    if (status != SUCCESS)
    {
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        ias_cpf_free(blob->cpf);
        local_free_the_blob(blob);
        return;
    }

    /* Compare the test and reference data buffers--if all went as expected,
       they should be the same (to minor differences in values on the order
       of less than 1e-5) */
    num_major_differences = compare_data_buffers(test_l1r_buffer,
        ref_l1r_buffer, NUMBER_DETECTORS, NUMBER_LINES, 1e-5);
    assert_equal_with_message(num_major_differences, 0, "%d differences "
        "greater than 1e-5 found", num_major_differences);

    /* Final cleanup.  SCA's 2-14 and the VRP band 19, SCAs 2-14 have
       detector offset data buffers allocated for them as well.  These
       counts need to be decremented in order to properly free their
       memory.  Start the SCA index loop at 1--the first SCA was taken
       care of when the API call cleaned up.  This assumes, of course,
       that test_sca_index is 0 (for the first SCA); if the test
       SCA changes, this will need to be reworked to decrement the
       counts from SCA 1 to (test_sca_index+1), and from (test_sca_index+2)
       to 14*/
    status = decrement_consumer_count(IMAGE_L1R, curr_band_index,
        test_sca_index);
    assert_equal_with_message( status, SUCCESS,
        "Error returned from decrement consumer count");
    for (sca_index = 1;
         sca_index < blob->band_information[curr_band_index]->number_scas;
         sca_index++)
    {
        status = decrement_consumer_count(MM_SPATIAL_DETECTOR_OFFSETS,
            curr_band_index, sca_index);
        assert_equal_with_message( status, SUCCESS,
            "Error returned from decrement consumer count");
        if (blob->band_information[curr_band_index]->vrp_band_number > 0)
        {
            status = decrement_consumer_count(MM_SPATIAL_DETECTOR_OFFSETS,
                blob->band_information[curr_band_index]->vrp_band_index,
                sca_index);
            assert_equal_with_message( status, SUCCESS,
                "Error returned from decrement consumer count");
        }
    }
    cleanup_rps_level();
    cleanup_memory_management_buffers(ENABLE_WARNING_DISPLAY);
    ias_cpf_free(blob->cpf);
    local_free_the_blob(blob);

}   /* END -- test_inopdetfill01_from_api */



/********************* RUN UNIT TEST(S) ***********************************/
int main(int argc, char *argv[])
{
    int status;

    /* Get the test data path.  Quit here if it isn't defined */
    test_data_path = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (test_data_path == NULL)
    {
        printf("Cannot find main IAS test data directory\n");
        exit(EXIT_FAILURE);
    }

    /* Create the main test suite */
    TestSuite *suite = create_test_suite();

    /* Add tests here */
    add_test(suite, test_inopdetfill01_from_algorithm);
    add_test(suite, test_inopdetfill02_from_algorithm);
    add_test(suite, test_inopdetfill03_from_algorithm);
    add_test(suite, test_inopdetfill_custom_from_algorithm);
    add_test(suite, test_inopdetfill01_from_api);

    /* Run the test(s).  Disable IAS logging of DEBUG messages */
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);
    if (argc > 1)
        status = run_single_test(suite, argv[1], create_text_reporter());
    else
        status = run_test_suite(suite, create_text_reporter());
    ias_log_set_output_level(IAS_LOG_LEVEL_DEBUG);

    return status;
}   /* END -- main */
