/*----------------------------------------------------------------------------
NAME:    oli_bias_determination_test.c

PURPOSE: Test the OLI bias determination routine by testing it directly and 
         testing it through API. 

USAGE:   oli_bias_determination_test [test name]

         When run without a command line argument all tests will be run.
         Pass the name of a test to run the single test -- useful for
         debugging.

Note: The algorithm and API implementation is based on ADD V3.1c, but some
      unit tests are using old IDL code because the Cal/Val V3.3 (IDL) 
      prototype code has multiple PM artifact files. Some test are using the 
      new Cal/Val V3.4 (C) prototype code.  
      The algorithm and API code haven been updated according to ADD 3.1e
      and then with that "1.5 bias subtraction" removed (by adding 1.5
      for truncation (upper 12 bits) data in "covert_to_float".  
----------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "cgreen/cgreen.h"
#include "ias_logging.h"
#include "../rps_local.h"
#include "ias_l0r.h"

#define NLINES 255
#define NDETS  494
#define NVRPDETS  12

#define NPANLINES 510
#define NPANDETS  988
#define NPANVRPDETS  24

#define PAN_BAND_NUMBER 8

/* Path to test data and file names */
static char data_dir[] = "ias_base/rps/driver/oli_bias_determination";
static char vrp_file[] = "Qvrp_b01_s01.txt";
static char vrp_pan_file[] = "Qvrp_b08_s01.txt";
static char bias_from_shutter_file[] = "bavg_b01_s01.txt";
static char bias_from_shutter_file2[] = "bavg_b01_s01_2.txt";
static char shutter_pan_file[] = "bavg_b08_s01.txt";
static char dropped_frames_file[] = "dropped_frames_b01_s01.txt";
static char dropped_frames_pan_file[] = "dropped_frames_b08_s01.txt";
static char impulse_noise_file[] = "IN_b01_s01.txt";
static char impulse_noise_pan_file[] = "IN_b08_s01.txt";
static char inop_dets_file[] = "inop_dets_b01_s01.txt";
static char inop_dets_pan_file[] = "inop_dets_b08_s01.txt";
static char sat_pixel_file[] = "sat_pix_b01_s01.txt";
static char sat_pixel_pan_file[] = "sat_pix_b08_s01.txt";
static char bias_file[] = "bias_b01_s01.txt";
static char bias_file2[] = "bias_b01_s01_2.txt";
static char bias_pan_file[] = "bias_b08_s01.txt";
static char dropped_frames_binary_file[] = "DroppedFrames1.dat";
static char vrp_binary_file[] = "Avrp1.dat";
static char a1_parameter_binary_file[] = "A11.dat";
static char c1_parameter_binary_file[] = "C11.dat";
static char bias_binary_file[] = "Bias1_5_0.dat";
static char bias_model_binary_file[] = "Bias1_5_1.dat";

/* Test prototypes */
Ensure test_algorithm_directly_non_pan_band(void);
Ensure test_algorithm_directly_non_pan_band2(void);
Ensure test_algorithm_directly_pan_band(void);
Ensure test_algorithm_thru_api(void);
Ensure test_algorithm_pan_band_thru_api(void);
Ensure test_algorithm_directly_non_pan_band_a1_c1(void);
Ensure test_algorithm_thru_api_a1_c1(void);

int main(int argc, char *argv[])
{
    TestSuite *suite = create_test_suite();
    add_test(suite, test_algorithm_directly_non_pan_band);
    add_test(suite, test_algorithm_directly_non_pan_band2);
    add_test(suite, test_algorithm_directly_pan_band);
    add_test(suite, test_algorithm_thru_api);
    add_test(suite, test_algorithm_pan_band_thru_api);
    add_test(suite, test_algorithm_directly_non_pan_band_a1_c1);
    add_test(suite, test_algorithm_thru_api_a1_c1);

    if (argc == 2)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}

/*--------------------------------------------------------------------------
   test_algorithm_directly_non_pan_band

   This test executes the algorithm with a subset of the Cal/Val prototype
   input and output. The input ASCII files are extracted from calval binary 
   test files with only one band and one sca used one at a time. A non_pan 
   band SCA are tested. The bias from shutter file used is the average of the 
   pre- and post-acquisition shutter average. The per-frame bias are calculated
   with model parameters (A0 and VRP band data). The compared result file is 
   generated from cal/val V3.3 code. 
---------------------------------------------------------------------------*/
Ensure test_algorithm_directly_non_pan_band(void)
{
    FILE *inf;                     /* File pointer for all reads */
    char *test_directory;          /* Test environment value */
    char file_name[PATH_MAX];      /* File name to open */
    uint16_t *vrp_buffer;          /* Buffer containing the input vrp data */
    double *bias_from_shutter_buffer;/* Per detector bias from file buffer */
    double *bias_buffer;           /* Calculated bias data buffer */
    double *result_buffer;         /* Bias data buffer from calval test */
    double *a1_parameter;          /* a1_parameter buffer from calval test */
    double *c1_parameter;          /* c1_parameter buffer from calval test */
    int temp_int;                  /* Temporary value read from file */
    double temp_double;            /* Temporary value read from file */
    int det;                       /* Detector loop counter */
    int line;                      /* Frame loop counter */
    int pixel;                     /* Pixel loop counter */
    int difference_count;          /* Numbers of differences between the 
                                      calculated value and expected results */
    char work_order_id[IAS_WORKORDER_ID_SIZE];
    char routine_name[] = "OLI bias determination tests";
    int return_value;              /* Return values from fscanf read */ 
    int a1_c1_estimate_flag = 0;   /* Not using a1/C1 estimation option */
    double a0_b01_s01 = 0.624321;  /* A0 coefficient for band 1 and SCA 1 */
    IAS_PIXEL_MASK *pm;            /* Pixel mask for VRP data */
    int band_number;               /* 1-based band number */
    int sca_number;                /* 1-based SCA number */
    IAS_REPORT report_info;
    IAS_REPORT_HEADER *report_header = NULL;
    int status;

    /* Get the environment variable for the test data */
    test_directory = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (!test_directory)
    {
        assert_true_with_message(0, "Getting IAS_UNIT_TEST_DATA_DIR");
        return;
    }

    /* Set up a standard header */
    setenv("PCS_WOID", "I4582", 1);
    strcpy(work_order_id, "I4582");
    report_header = ias_misc_initialize_report_header("",
        work_order_id, routine_name);
    if (!report_header)
    {
        assert_true_with_message(0, "Error initializing report header");
        return;
    }
    ias_misc_set_header_pathrow(report_header, 0, 0);
    report_info.header = report_header;
    report_info.directory = NULL;

    /* Create one pixel mask */
    pm = ias_pm_create(1, 1, NVRPDETS, NLINES);
    assert_not_equal(pm, NULL);

    /* Open and read the inoperable detector pixel mask */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory, 
             data_dir, inop_dets_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building the inoperable "
            "detector pixel mask file name");
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    for (det = 0; det < NVRPDETS; det++)
    {
        return_value = fscanf(inf, "%d", &temp_int);
        if (return_value == EOF)
        {
            assert_true_with_message(0, "Reading %s", file_name);
            return;        
        }        
        /* Add the inoperable detectors to pixel mask */
        if (temp_int == 0)
        {
            assert_false(ias_pm_add_pixels(pm, det, 1-1, NLINES, 
                         IAS_RPS_INOP_DETECTOR));
        }
    }
    fclose(inf);
    /* Open and read the dropped frame pixel mask 
       Note: In calval test file, the whole frame is dropped if the frame
             is labeled (0 in value), therefore, the read in dropped frames
             pixel mask is one value per frame in each SCA and each band */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory, 
             data_dir, dropped_frames_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building the dropped frame "
                                 "pixel mask file name");
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    for (line = 0; line < NLINES; line++)
    {
        return_value = fscanf(inf, "%d", &temp_int);
        if (return_value == EOF)
        {
            assert_true_with_message(0, "Reading %s", file_name);
            ias_pm_destroy(pm);
            ias_misc_free_report_header(report_header);
            return;        
        }
        /* Add the dropped frames to pixel mask */
        if (temp_int == 0)
        {
            for (det = 0; det < NVRPDETS; det++)
            {
                assert_false(ias_pm_add_pixels(pm, det, line, 1, 
                             IAS_RPS_DROPLINE));
            }
        }
    }
    fclose(inf);

    /* Open and read the impulse noise pixel mask */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory, 
             data_dir, impulse_noise_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building impulse noise pixel "
                                 "mask file name");
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    for (line = 0; line < NLINES; line++)
    {
        for (det = 0; det < NVRPDETS; det++)
        {
            return_value = fscanf(inf, "%d", &temp_int);
            if (return_value == EOF)
            {
                assert_true_with_message(0, "Reading %s", file_name);
                ias_pm_destroy(pm);
                ias_misc_free_report_header(report_header);
                return;        
            }        
            /* Add the impulse noise to pixel mask */
            if (temp_int == 0)
             {
                assert_false(ias_pm_add_pixels(pm, det, line, 1, 
                             IAS_RPS_IMPNOISE));
             }
        }
    }
    fclose(inf);

    /* Open and read the saturated pixel pixel mask */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory, 
             data_dir, sat_pixel_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building saturated pixel pixel "
                                 "mask file name");
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    for (line = 0; line < NLINES; line++)
    {
        for (det = 0; det < NVRPDETS; det++)
        {
            return_value = fscanf(inf, "%d", &temp_int);
            if (return_value == EOF)
            {
                assert_true_with_message(0, "Reading %s", file_name);
                ias_pm_destroy(pm);
                ias_misc_free_report_header(report_header);
                return;        
            }        
            /* Add the saturated pixel to pixel mask
               Note: The calval saturated pixel mask file doesn't distinguish 
                     the saturated pixel type, therefore, 
                     IAS_RPS_ANALOG_SAT_HI type is used since it really doesn't
                     matter */
            if (temp_int == 0)
            {
                assert_false(ias_pm_add_pixels(pm, det, line, 1, 
                             IAS_RPS_ANALOG_SAT_HI));
            }
        }
    }
    fclose(inf);

    /* Open and read the detector level bias_from_shutter_file buffer */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, bias_from_shutter_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building shutter buffer file "
                                 "name");
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    bias_from_shutter_buffer = malloc(NDETS * sizeof(double));
    if (!bias_from_shutter_buffer)
    {
        assert_true_with_message(0, "Allocating memory for "
                                 "bias_from_shutter_buffer");
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        free(bias_from_shutter_buffer);
        ias_misc_free_report_header(report_header);
        ias_pm_destroy(pm);
        return;
    }

    for (det = 0; det < NDETS; det++)
    {
        return_value = fscanf(inf, "%lf", &temp_double);
        if (return_value == EOF)
        {
            assert_true_with_message(0, "Reading %s", file_name);
            free(bias_from_shutter_buffer);
            ias_misc_free_report_header(report_header);
            ias_pm_destroy(pm);
            return;        
        }        
        bias_from_shutter_buffer[det] = temp_double;
    }
    fclose(inf);

    /* Open and read the VRP data */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, vrp_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building VRP data file name");
        free(bias_from_shutter_buffer);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }
    vrp_buffer = malloc(NVRPDETS * NLINES * sizeof(uint16_t));
    if (!vrp_buffer)
    {
        assert_true_with_message(0, "Allocating memory for dark_buffer" );
        free(bias_from_shutter_buffer);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        free(bias_from_shutter_buffer);
        free(vrp_buffer);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    for (line = 0; line < NLINES; line++)
    {
        for (det = 0; det < NVRPDETS; det++)
        {
            return_value = fscanf(inf, "%d", &temp_int);
            if (return_value == EOF)
            {
                assert_true_with_message(0, "Reading %s", file_name);
                free(bias_from_shutter_buffer);
                free(vrp_buffer);
                ias_pm_destroy(pm);
                ias_misc_free_report_header(report_header);
                return;        
            }        
            vrp_buffer[line * NVRPDETS + det] = (uint16_t)temp_int;
        }
    }
    fclose(inf);

    /* Set band_number and sca_number */
    band_number = 1;
    sca_number = 1;

    /* Allocate memory for a1_parameter */
    a1_parameter = malloc(NDETS * sizeof(double));
    if (!a1_parameter)
    {
        assert_true_with_message(0, "Allocating memory for bias" );
        free(bias_from_shutter_buffer);
        free(vrp_buffer);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    /* Polulate a1_parameter with 0.236531 for now */
    for (det = 0; det < NDETS; det++)
    {
        a1_parameter[det] = 0.236531;
    }

    /* Allocate memory for c1_parameter */
    c1_parameter = malloc(NDETS * sizeof(double));
    if (!c1_parameter)
    {
        assert_true_with_message(0, "Allocating memory for bias" );
        free(bias_from_shutter_buffer);
        free(vrp_buffer);
        free(a1_parameter);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    /* Polulate c1_parameter with 65.932693 for now */
    for (det = 0; det < NDETS; det++)
    {
        c1_parameter[det] = 65.932693;
    }

    /* Allocate memory for bias */
    bias_buffer = malloc(NDETS * NLINES * sizeof(double));
    if (!bias_buffer)
    {
        assert_true_with_message(0, "Allocating memory for bias" );
        free(bias_from_shutter_buffer);
        free(vrp_buffer);
        free(a1_parameter);
        free(c1_parameter);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    /* Call the algorithm and test it for band 1 and SCA 1 */
    status = oli_bias_determination(vrp_buffer, band_number, sca_number, NDETS,
             NLINES, NVRPDETS, NLINES, PER_FRAME_BIAS, a1_c1_estimate_flag, 
             NON_PAN, bias_from_shutter_buffer, a1_parameter, c1_parameter, 
             a0_b01_s01, pm, &report_info, bias_buffer);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Calling oli bias determination" );
        free(bias_from_shutter_buffer);
        free(vrp_buffer);
        free(bias_buffer);
        free(a1_parameter);
        free(c1_parameter);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    /* Open and read the bias result file generated from cal/val code */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, bias_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building result data file name");
        free(bias_from_shutter_buffer);
        free(vrp_buffer);
        free(bias_buffer);
        free(a1_parameter);
        free(c1_parameter);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }
    result_buffer = malloc(NDETS * NLINES * sizeof(double));
    if (!result_buffer)
    {
        assert_true_with_message(0, "Allocating memory for bias " 
                                         "buffer");
        free(bias_from_shutter_buffer);
        free(vrp_buffer);
        free(bias_buffer);
        free(a1_parameter);
        free(c1_parameter);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        free(bias_from_shutter_buffer);
        free(vrp_buffer);
        free(bias_buffer);
        free(result_buffer);
        free(a1_parameter);
        free(c1_parameter);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    for (line = 0; line < NLINES; line++)
    {
        for (det = 0; det < NDETS; det++)
        {
            return_value = fscanf(inf, "%lf", &temp_double);
            if (return_value == EOF)
            {
                assert_true_with_message(0, "Reading %s", file_name);
                free(bias_from_shutter_buffer);
                free(vrp_buffer);
                free(bias_buffer);
                free(result_buffer);
                free(a1_parameter);
                free(c1_parameter);
                ias_pm_destroy(pm);
                ias_misc_free_report_header(report_header);
                return;        
            }        
            result_buffer[line * NDETS + det] = temp_double;
        }
    }
    fclose(inf);
    
    /* Compare the results.  Loop through the results and count the
       differences.  This comparison isn't done with asserts since there is
       no need for 494*255 assert statements which cost message buffer space.
       One assert for pass fail with a count is good enough. */
    difference_count = 0;
    for (pixel = 0; pixel < NDETS * NLINES; pixel++)
    {
        if (fabsf(bias_buffer[pixel] - result_buffer[pixel]) > 1e-2)
            difference_count++;
    }
    assert_equal_with_message(difference_count, 0, "There were %d "
        "differences out of %d values", difference_count, NDETS * NLINES);

    /* Free the allocated memory */
    ias_pm_destroy(pm);
    free(bias_from_shutter_buffer);
    free(vrp_buffer);
    free(bias_buffer);
    free(result_buffer);
    free(a1_parameter);
    free(c1_parameter);
    ias_misc_free_report_header( report_header );
} 

/*--------------------------------------------------------------------------
   test_algorithm_directly_non_pan_band2

   This test executes the algorithm with a subset of the Cal/Val prototype
   input and output. The input ASCII files are extracted from calval binary 
   test files with only one band and one sca used one at a time. A non_pan 
   band SCA are tested. The bias from shutter file used is the average of the 
   pre- and post-acquisition shutter average. No model coefficients are used.
   The per-frame bias are simply extension of the per-detector bias. The 
   compared result file is generated from cal/val V3.3 code. 
---------------------------------------------------------------------------*/
Ensure test_algorithm_directly_non_pan_band2(void)
{
    FILE *inf;                     /* File pointer for all reads */
    char *test_directory;          /* Test environment value */
    char file_name[PATH_MAX];      /* File name to open */
    uint16_t *vrp_buffer = NULL;   /* Buffer containing the input vrp data */
    double *bias_from_shutter_buffer; /* Per detector bias from file buffer */
    double *bias_buffer;           /* Calculated bias data buffer */
    double *result_buffer;         /* Bias data buffer from calval code test */
    double *a1_parameter = NULL;   /* a1_parameter buffer from calval test */
    double *c1_parameter = NULL;   /* c1_parameter buffer from calval test */
    double temp_double;            /* Temporary value read from file */
    int det;                       /* Detector loop counter */
    int line;                      /* Frame loop counter */
    int pixel;                     /* Pixel loop counter */
    int difference_count;          /* Numbers of differences between the 
                                      calculated value and expected results */
    char work_order_id[IAS_WORKORDER_ID_SIZE];
    char routine_name[] = "OLI bias determination tests";
    int return_value;              /* Return values from fscanf read */ 
    int a1_c1_estimate_flag = 0;   /* Not using a1/C1 estimation option */
    double a0_b01_s01 = 0.624321;  /* A0 coefficient for band 1 and SCA 1 */
    IAS_PIXEL_MASK *pm;            /* Pixel mask for VRP data */
    int band_number;               /* 1-based band number */
    int sca_number;                /* 1-based SCA number */
    IAS_REPORT report_info;
    IAS_REPORT_HEADER *report_header = NULL;
    int status;

    /* Get the environment variable for the test data */
    test_directory = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (!test_directory)
    {
        assert_true_with_message(0, "Getting IAS_UNIT_TEST_DATA_DIR");
        return;
    }

    /* Set up a standard header */
    setenv("PCS_WOID", "I4582", 1);
    strcpy(work_order_id, "I4582");
    report_header = ias_misc_initialize_report_header("",
        work_order_id, routine_name);
    if (!report_header)
    {
        assert_true_with_message(0, "Error initializing report header");
        return;
    }
    ias_misc_set_header_pathrow(report_header, 0, 0);
    report_info.header = report_header;
    report_info.directory = NULL;

    /* Create one pixel mask */
    pm = ias_pm_create(1, 1, NVRPDETS, NLINES);
    assert_not_equal(pm, NULL);

    /* Open and read the detector level shutter buffer */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, bias_from_shutter_file2);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building shutter file name");
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }
    bias_from_shutter_buffer = malloc(NDETS * sizeof(double));
    if (!bias_from_shutter_buffer)
    {
        assert_true_with_message(0, "Allocating memory for "
                                 "bias_from_shutter_buffer");
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        free(bias_from_shutter_buffer);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    for (det = 0; det < NDETS; det++)
    {
        return_value = fscanf(inf, "%lf", &temp_double);
        if (return_value == EOF)
        {
            assert_true_with_message(0, "Reading %s", file_name);
            free(bias_from_shutter_buffer);
            ias_misc_free_report_header(report_header);
            ias_pm_destroy(pm);
            return;        
        }        
        bias_from_shutter_buffer[det] = temp_double;
    }
    fclose(inf);

    /* Set band_number and sca_number */
    band_number = 1;
    sca_number = 1;

    /* Allocate memory for bias */
    bias_buffer = malloc(NDETS * NLINES * sizeof(double));
    if (!bias_buffer)
    {
        assert_true_with_message(0, "Allocating memory for bias" );
        free(bias_from_shutter_buffer);
        ias_misc_free_report_header(report_header);
        ias_pm_destroy(pm);
        return;
    }

    /* Call the algorithm and test it for band 1 and SCA 1 */
    status = oli_bias_determination(vrp_buffer, band_number, sca_number, NDETS,
             NLINES, NVRPDETS, NLINES, PER_DETECTOR_BIAS, a1_c1_estimate_flag, 
             NON_PAN, bias_from_shutter_buffer, a1_parameter, c1_parameter, 
             a0_b01_s01, pm, &report_info, bias_buffer);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Calling oli bias determination" );
        free(bias_from_shutter_buffer);
        free(bias_buffer);
        ias_misc_free_report_header(report_header);
        ias_pm_destroy(pm);
        return;
    }

    /* Open and read the bias result file generated from cal/val code */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, bias_file2);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building result file name");
        free(bias_from_shutter_buffer);
        free(bias_buffer);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    /* Allocate memory for result_buffer */
    result_buffer = malloc(NDETS * NLINES * sizeof(double));
    if (!result_buffer)
    {
        assert_true_with_message(0, "Allocating memory for bias " 
                                         "buffer");
        free(bias_from_shutter_buffer);
        free(bias_buffer);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        free(bias_from_shutter_buffer);
        free(bias_buffer);
        free(result_buffer);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    for (line = 0; line < NLINES; line++)
    {
        for (det = 0; det < NDETS; det++)
        {
            return_value = fscanf(inf, "%lf", &temp_double);
            if (return_value == EOF)
            {
                assert_true_with_message(0, "Reading %s", file_name);
                free(bias_from_shutter_buffer);
                free(bias_buffer);
                free(result_buffer);
                ias_pm_destroy(pm);
                ias_misc_free_report_header(report_header);
                return;        
            }        
            result_buffer[line * NDETS + det] = temp_double;
        }
    }
    fclose(inf);
    
    /* Compare the results.  Loop through the results and count the
       differences.  This comparison isn't done with asserts since there is
       no need for 494*255 assert statements which cost message buffer space.
       One assert for pass fail with a count is good enough. */
    difference_count = 0;
    for (pixel = 0; pixel < NDETS * NLINES; pixel++)
    {
        if (fabsf(bias_buffer[pixel] - result_buffer[pixel]) > 1e-5)
            difference_count++;
    }
    assert_equal_with_message(difference_count, 0, "There were %d "
        "differences out of %d values", difference_count, NDETS * NLINES);

    /* Free the allocated memory */
    ias_pm_destroy(pm);
    free(bias_from_shutter_buffer);
    free(bias_buffer);
    free(result_buffer);
    ias_misc_free_report_header(report_header);
} 

/*--------------------------------------------------------------------------
   test_algorithm_directly_pan_band

   This test executes the algorithm with a subset of the Cal/Val prototype
   input and output. The input ASCII files are extracted from calval binary 
   test files with only pan band and one sca used one at a time. A pan 
   band SCA are tested. The bias from shutter file used is the average of the 
   pre- and post-acquisition shutter average. The per-frame bias are calculated
   from model parameters (A0 and VRP band data). The compared result file is 
   generated from cal/val V3.3 code. 
---------------------------------------------------------------------------*/
Ensure test_algorithm_directly_pan_band(void)
{
    FILE *inf;                     /* File pointer for all reads */
    char *test_directory;          /* Test environment value */
    char file_name[PATH_MAX];      /* File name to open */
    uint16_t *vrp_buffer;          /* Buffer containing the input vrp data */
    double *bias_from_shutter_buffer;/* Per detector bias from file buffer */
    double *bias_buffer;           /* Calculated bias data buffer */
    double *result_buffer;         /* Bias data buffer from calval code test */
    double *a1_parameter;          /* a1_parameter buffer from calval test */
    double *c1_parameter;          /* c1_parameter buffer from calval test */
    int temp_int;                  /* Temporary value read from file */
    double temp_double;            /* Temporary value read from file */
    int det;                       /* Detector loop counter */
    int line;                      /* Frame loop counter */
    int pixel;                     /* Pixel loop counter */
    int difference_count;          /* Numbers of differences between the 
                                      calculated value and expected results */
    int return_value;              /* Return values from fscanf read */ 
    char work_order_id[IAS_WORKORDER_ID_SIZE];
    char routine_name[] = "OLI bias determination tests";
    int a1_c1_estimate_flag = 0;   /* Not using a1/C1 estimation option */
    double a0_b08_s01_even = 0.448819; /* A0 coefficient for band 8 (pan band 
                                         even) and SCA 1 */
    double a0_b08_s01_odd = 0.559437;  /* A0 coefficient for band 8 (pan band
                                         odd) and SCA 1 */
    IAS_PIXEL_MASK *pm;            /* Pixel mask for VRP data */
    int band_number;               /* 1-based band number */
    int sca_number;                /* 1-based SCA number */
    IAS_REPORT report_info;
    IAS_REPORT_HEADER *report_header = NULL;
    int status;

    /* Get the environment variable for the test data */
    test_directory = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (!test_directory)
    {
        assert_true_with_message(0, "Getting IAS_UNIT_TEST_DATA_DIR");
        return;
    }

    /* Set up a standard header */
    setenv("PCS_WOID", "I4582", 1);
    strcpy(work_order_id, "I4582");
    report_header = ias_misc_initialize_report_header("",
        work_order_id, routine_name);
    if (!report_header)
    {
        assert_true_with_message(0, "Error initializing report header");
        return;
    }
    ias_misc_set_header_pathrow(report_header, 0, 0);
    report_info.header = report_header;
    report_info.directory = NULL;

    /* Create one pixel mask 
       Note: The cal/val test file only have 12 detectors for pan band,
             have to double it here */
    pm = ias_pm_create(PAN_BAND_NUMBER, 1, NPANVRPDETS, NPANLINES);
    assert_not_equal(pm, NULL);

    /* Open and read the inoperable detector pixel mask */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, inop_dets_pan_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building inoperable detector "
                                 "pixel maskfile name");
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    /* Note: The cal/val test data file only has 12 VRP dets for the pan band
             and pixel mask data, therefore, these 12 dets data will be used 
             for odd and even VRP detectors separately. Not having 24 dets VRP 
             data available is due to the use of a flawed data converter based
             on cal/val response */
    for (det = 0; det < NVRPDETS; det++)
    {
        return_value = fscanf(inf, "%d", &temp_int);
        if (return_value == EOF)
        {
            assert_true_with_message(0, "Reading %s", file_name);
            return;        
        }        
        /* Add the inoperable detectors to pixel mask */
        if (temp_int == 0)
        {
            assert_false(ias_pm_add_pixels(pm, det, 1-1, NPANLINES, 
                         IAS_RPS_INOP_DETECTOR));
            assert_false(ias_pm_add_pixels(pm, NVRPDETS + det, 1-1, NPANLINES, 
                         IAS_RPS_INOP_DETECTOR));
        }
    }
    fclose(inf);

    /* Open and read the dropped frame pixel mask 
       Note: In calval test file, the whole frame is dropped if the frame
             is labeled (0 in value), therefore, the read in dropped frames
             pixel mask is one value per frame in each SCA and each band */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, dropped_frames_pan_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building dropped frame pixel "
                                 "maskfile name");
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    for (line = 0; line < NPANLINES; line++)
    {
        return_value = fscanf(inf, "%d", &temp_int);
        if (return_value == EOF)
        {
            assert_true_with_message(0, "Reading %s", file_name);
            return;        
        }
        /* Add the dropped frames to pixel mask */
        if (temp_int == 0)
        {
            for (det = 0; det < NVRPDETS; det++)
            {
                assert_false(ias_pm_add_pixels(pm, det, line, 1, 
                             IAS_RPS_DROPLINE));
                assert_false(ias_pm_add_pixels(pm, NVRPDETS + det, line, 1, 
                             IAS_RPS_DROPLINE));
            }
        }
    }
    fclose(inf);

    /* Open and read the impulse noise pixel mask */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, impulse_noise_pan_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building impulse noise pixel "
                                 "mask file name");
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    for (line = 0; line < NPANLINES; line++)
    {
        for (det = 0; det < NVRPDETS; det++)
        {
            return_value = fscanf(inf, "%d", &temp_int);
            if (return_value == EOF)
            {
                assert_true_with_message(0, "Reading %s", file_name);
                return;        
            }        
            /* Add the impulse noise to pixel mask */
            if (temp_int == 0)
            {
                ias_pm_add_pixels(pm, det, line, 1, 
                             IAS_RPS_IMPNOISE);
                ias_pm_add_pixels(pm, NVRPDETS + det, line, 1, 
                             IAS_RPS_IMPNOISE);
            }
        }
    }
    fclose(inf);

    /* Open and read the saturated pixels pixel mask */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, sat_pixel_pan_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building saturated pixels "
                                 "pixel mask file name");
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    for (line = 0; line < NPANLINES; line++)
    {
        for (det = 0; det < NVRPDETS; det++)
        {
            return_value = fscanf(inf, "%d", &temp_int);
            if (return_value == EOF)
            {
                assert_true_with_message(0, "Reading %s", file_name);
                ias_pm_destroy(pm);
                ias_misc_free_report_header(report_header);
                return;        
            }        
            /* Add the saturated pixel to pixel mask
               Note: The calval saturated pixel mask file doesn't distinguish 
                     the saturated pixel type, therefore, 
                     IAS_RPS_ANALOG_SAT_HI type is used since it really doesn't
                     matter */
            if (temp_int == 0)
            {
                ias_pm_add_pixels(pm, det, line, 1, 
                             IAS_RPS_ANALOG_SAT_HI);
                ias_pm_add_pixels(pm, NVRPDETS + det, line, 1, 
                             IAS_RPS_ANALOG_SAT_HI);
            }
        }
    }
    fclose(inf);

    /* Open and read the detector level shutter buffer */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, shutter_pan_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building shutter file name");
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }
    bias_from_shutter_buffer = malloc(NPANDETS * sizeof(double));
    if (!bias_from_shutter_buffer)
    {
        assert_true_with_message(0, "Allocating memory for "
                                 "bias_from_shutter_buffer");
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        free(bias_from_shutter_buffer);
        ias_misc_free_report_header(report_header);
        return;
    }

    for (det = 0; det < NPANDETS; det++)
    {
        return_value = fscanf(inf, "%lf", &temp_double);
        if (return_value == EOF)
        {
            assert_true_with_message(0, "Reading %s", file_name);
            free(bias_from_shutter_buffer);
            ias_pm_destroy(pm);
            ias_misc_free_report_header(report_header);
            return;        
        }        
        bias_from_shutter_buffer[det] = temp_double;
    }
    fclose(inf);

    /* Open and read the VRP data */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, vrp_pan_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building VRP data file name");
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }
    vrp_buffer = malloc(NPANVRPDETS * NPANLINES * sizeof(uint16_t));
    if (!vrp_buffer)
    {
        assert_true_with_message(0, "Allocating memory for dark_buffer" );
        free(bias_from_shutter_buffer);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        free(bias_from_shutter_buffer);
        free(vrp_buffer);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    for (line = 0; line < NPANLINES; line++)
    {
        for (det = 0; det < NVRPDETS; det++)
        {
            return_value = fscanf(inf, "%d", &temp_int);
            if (return_value == EOF)
            {
                assert_true_with_message(0, "Reading %s", file_name);
                free(bias_from_shutter_buffer);
                free(vrp_buffer);
                ias_pm_destroy(pm);
                ias_misc_free_report_header(report_header);
                return;        
            }        
            vrp_buffer[line * NPANVRPDETS + det] = (uint16_t)temp_int;
            vrp_buffer[line * NPANVRPDETS + NVRPDETS + det ] 
                = (uint16_t)temp_int;
        }
    }
    fclose(inf);

    /* Set band_number and sca_number */
    band_number = PAN_BAND_NUMBER;
    sca_number = 1;

    /* Allocate memory for a1_parameter */
    a1_parameter = malloc(NPANDETS * sizeof(double));
    if (!a1_parameter)
    {
        assert_true_with_message(0, "Allocating memory for bias" );
        free(bias_from_shutter_buffer);
        free(vrp_buffer);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    /* Polulate a1_parameter with 0.236531 for now */
    for (det = 0; det < NPANDETS; det++)
    {
        a1_parameter[det] = 0.236531;
    }

    /* Allocate memory for c1_parameter */
    c1_parameter = malloc(NPANDETS * sizeof(double));
    if (!c1_parameter)
    {
        assert_true_with_message(0, "Allocating memory for bias" );
        free(bias_from_shutter_buffer);
        free(vrp_buffer);
        free(a1_parameter);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    /* Polulate c1_parameter with 65.932693 for now */
    for (det = 0; det < NPANDETS; det++)
    {
        c1_parameter[det] = 65.932693;
    }

    /* Allocate memory for bias */
    bias_buffer = malloc(NPANDETS * NPANLINES * sizeof(double));
    if (!bias_buffer)
    {
        assert_true_with_message(0, "Allocating memory for bias" );
        free(bias_from_shutter_buffer);
        free(vrp_buffer);
        free(a1_parameter);
        free(c1_parameter);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    /* Call the algorithm and test it for band 8 (pan_odd) and SCA 1 */
    status = oli_bias_determination(vrp_buffer, band_number, sca_number, 
             NPANDETS, NPANLINES, NPANVRPDETS, NPANLINES, PER_FRAME_BIAS, 
             a1_c1_estimate_flag, PAN_ODD, bias_from_shutter_buffer, 
             a1_parameter, c1_parameter, a0_b08_s01_odd, pm, &report_info, 
             bias_buffer);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Calling oli bias determination" );
        free(bias_from_shutter_buffer);
        free(vrp_buffer);
        free(bias_buffer);
        free(a1_parameter);
        free(c1_parameter);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    /* Call the algorithm and test it for band 8 (pan even) and SCA 1 */
    status = oli_bias_determination(vrp_buffer, band_number, sca_number, 
             NPANDETS, NPANLINES, NPANVRPDETS, NPANLINES, PER_FRAME_BIAS, 
             a1_c1_estimate_flag, PAN_EVEN, bias_from_shutter_buffer, 
             a1_parameter, c1_parameter, a0_b08_s01_even, pm, &report_info, 
             bias_buffer);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Calling oli bias determination" );
        free(bias_from_shutter_buffer);
        free(vrp_buffer);
        free(bias_buffer);
        free(a1_parameter);
        free(c1_parameter);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    /* Open and read the bias result file generated from cal/val code
       */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, bias_pan_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building result data file name");
        free(bias_from_shutter_buffer);
        free(vrp_buffer);
        free(bias_buffer);
        free(a1_parameter);
        free(c1_parameter);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }
    result_buffer = malloc(NPANDETS * NPANLINES * sizeof(double));
    if (!result_buffer)
    {
        assert_true_with_message(0, "Allocating memory for bias " 
                                         "buffer");
        free(bias_from_shutter_buffer);
        free(vrp_buffer);
        free(bias_buffer);
        free(a1_parameter);
        free(c1_parameter);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        free(bias_from_shutter_buffer);
        free(vrp_buffer);
        free(bias_buffer);
        free(result_buffer);
        free(a1_parameter);
        free(c1_parameter);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    for (line = 0; line < NPANLINES; line++)
    {
        for (det = 0; det < NPANDETS; det++)
        {
            return_value = fscanf(inf, "%lf", &temp_double);
            if (return_value == EOF)
            {
                assert_true_with_message(0, "Reading %s", file_name);
                free(bias_from_shutter_buffer);
                free(vrp_buffer);
                free(bias_buffer);
                free(result_buffer);
                free(a1_parameter);
                free(c1_parameter);
                ias_pm_destroy(pm);
                ias_misc_free_report_header(report_header);
                return;        
            }        
            result_buffer[line * NPANDETS + det] = temp_double;
        }
    }
    fclose(inf);
    
    /* Compare the results.  Loop through the results and count the
       differences.  This comparison isn't done with asserts since there is
       no need for 988*510 assert statements which cost message buffer space.
       One assert for pass fail with a count is good enough. */
    difference_count = 0;
    for (pixel = 0; pixel < NPANDETS * NPANLINES; pixel++)
    {
        if (fabsf(bias_buffer[pixel] - result_buffer[pixel]) > 1e-2)
        {
            difference_count++;
        }
    }
    assert_equal_with_message(difference_count, 0, "There were %d "
        "differences out of %d values", difference_count, NPANDETS * NPANLINES);

    /* Free the allocated memory */
    ias_pm_destroy(pm);
    free(bias_from_shutter_buffer);
    free(vrp_buffer);
    free(bias_buffer);
    free(result_buffer);
    free(a1_parameter);
    free(c1_parameter);
    ias_misc_free_report_header(report_header);
} 

/*--------------------------------------------------------------------------
  test_algorithm_thru_api

  This test sets up for a call to the algorithm through the API.  Also calls
  the pre-process routine to set up memory management. The input ASCII files 
  are extracted from calval binary test files with only one band and one sca 
  used here. The bias from shutter file used is the average of the 
  pre- and post-acquisition shutter average. The per-frame bias are calculated
  with model parameters (A0 and VRP band data). The compared result file is 
  generated from cal/val V3.3 code. 
--------------------------------------------------------------------------*/
Ensure test_algorithm_thru_api(void)
{
    FILE *inf;                     /* File pointer for all reads */
    char *test_directory;          /* test environment value */
    char file_name[PATH_MAX];      /* File name to open */
    uint16_t *vrp_buffer;          /* Buffer containing the input vrp data */
    double *bias_from_shutter_buffer;/* Per detector bias from file buffer */
    double *bias_buffer;           /* Calculated bias data buffer */
    double *result_buffer;         /* Bias data buffer from calval code test */
    int temp_int;                  /* Temporary value read from file */
    double temp_double;            /* Temporary value read from file */
    int det;                       /* Detector loop counter */
    int line;                      /* Frame loop counter */
    int pixel;                     /* Pixel loop counter */
    int difference_count;          /* Numbers of differences between the 
                                      calculated value and expected results */
    int return_value;              /* Return values from fscanf read */ 
    char work_order_id[IAS_WORKORDER_ID_SIZE];
    char routine_name[] = "OLI bias determination tests";

    double a0_b01_s01 = 0.624321;  /* A0 coefficient for band 1 and SCA 1 */
    double *a0_parameter;          /* Buffer containing A0 coefficient  */
    double *a1_parameter;          /* a1_parameter buffer from calval test */
    double *c1_parameter;          /* c1_parameter buffer from calval test */
    BLOB *blob;                    /* Declare a blob structure */
    IAS_REPORT_HEADER *report_header = NULL;
    int status;
    int band_index;
    int test_band_index;
    int test_vrp_index;
    const IAS_SATELLITE_ATTRIBUTES *sat_attributes;

    static char bpf_dir[] = "ias_lib/bpf";
    static char cpf_dir[] = "ias_lib/cpf";
    static char rlut_dir[] = "ias_lib/rlut";
    static char l0r_dir[] = "ias_base/rps/driver/output";
    static char bpf_name_oli[] = "test_LO8BPF20140310103310_20140310103345.07";
    static char cpf_name[] = "L8CPF20090101_20090331.01";
    static char rlut_name[] = "test_linearization_rlut.h5";
    static char l0r_name[] = "LO82220010052011265EDC00";
    static char l1r_name[] = "temp_l1r";

    /* Get the environment variable for the test data */
    test_directory = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (!test_directory)
    {
        assert_true_with_message(0, "Getting IAS_UNIT_TEST_DATA_DIR");
        return;
    }

    /* The attributes must be initialized for memory management to work */
    ias_sat_attr_initialize(IAS_L8);

    /* Allocate a blob and populate the minimum members for the test
       to succeed */
    blob = allocate_blob();
    if ( !blob )
    {
        assert_true_with_message( 0, "allocating blob" );
        return;
    }

    /* Initialize blob */
    status = snprintf(blob->cpf_filename, sizeof(blob->cpf_filename), 
             "%s/%s/%s", test_directory, cpf_dir, cpf_name);
    if ((status < 0) || (status >= sizeof(blob->cpf_filename)))
    {
        assert_true_with_message(FALSE, "Error building CPF file name");
        return;
    }
    status = snprintf(blob->bpf_filename_oli, sizeof(blob->bpf_filename_oli), 
             "%s/%s/%s", test_directory, bpf_dir, bpf_name_oli);
    if ((status < 0) || (status >= sizeof(blob->bpf_filename_oli)))
    {
        assert_true_with_message(FALSE, "Error building OLI BPF file name");
        return;
    }
    strcpy(blob->bpf_filename_tirs, "");
    status = snprintf(blob->rlut_filename, sizeof(blob->rlut_filename), 
             "%s/%s/%s", test_directory, rlut_dir, rlut_name);
    if ((status < 0) || (status >= sizeof(blob->rlut_filename)))
    {
        assert_true_with_message(FALSE, "Error building RLUT file name");
        return;
    }
    sat_attributes = ias_sat_attr_get_attributes();
    blob->total_bands = sat_attributes->total_bands;
    blob->number_of_bands = 3;
    blob->band_fields_initialized = 1;

    if ( finish_blob_initialization())
    {
        assert_true_with_message( 0, "initializing blob" );
        return;
    }
    blob->list_of_bands[0] = 3;
    blob->list_of_bands[1] = 11;
    blob->list_of_bands[2] = 1;
    blob->list_of_band_indices[0] =
        ias_sat_attr_convert_band_number_to_index( blob->list_of_bands[0] );
    blob->list_of_band_indices[1] =
        ias_sat_attr_convert_band_number_to_index( blob->list_of_bands[1] );
    blob->list_of_band_indices[2] =
        ias_sat_attr_convert_band_number_to_index( blob->list_of_bands[2] );
    if ( blob->list_of_band_indices[0] == ERROR ||
         blob->list_of_band_indices[1] == ERROR ||
         blob->list_of_band_indices[2] == ERROR )
    {
        assert_true_with_message( 0, "Converting band number to index");
        free_the_blob();
        return;
    }
    blob->bias_source = MEAN_ACQUISITION_SHUTTER;
    blob->per_frame_bias_flag = PER_FRAME_BIAS;
    strcpy(blob->date_acquired, "2010:205:16:15:27.04520");
    blob->process_oli_bands = 1;
    blob->process_tirs_bands = 0;

    /* Initialize I/O routines */
    status = snprintf(blob->l0r_file_name, sizeof(blob->l0r_file_name), 
             "%s/%s/%s", test_directory, l0r_dir, l0r_name);
    if ((status < 0) || (status >= sizeof(blob->l0r_file_name)))
    {
        assert_true_with_message(FALSE, "Error building L0R data file name");
        free_the_blob();
        return;
    }
    status = snprintf(blob->l1r_file_name, sizeof(blob->l1r_file_name), "%s", 
             l1r_name);
    if ((status < 0) || (status >= sizeof(blob->l1r_file_name)))
    {
        assert_true_with_message(FALSE, "Error building L1R data file name");
        free_the_blob();
        return;
    }
    ias_log_set_output_level( IAS_LOG_LEVEL_DISABLE );
    status = initialize_io_routines( blob );
    if ( status )
    {
        free_the_blob();
        assert_true_with_message( 0,
            "initialize_io_routines returned an error" );
        return;
    }
    ias_log_set_output_level( IAS_LOG_LEVEL_INFO );

    /* Read OLI BPF */
    status = read_oli_tirs_bpf(blob);
    assert_equal_with_message(status, SUCCESS, "Error reading OLI BPF "
        "masks");
    if (status != SUCCESS)
    {
        cleanup_io_routines(blob);
        free_the_blob();
        return;
    }

    /* Create blob pixel_masks */
    status = create_blob_pixel_mask();
    assert_equal_with_message(status, SUCCESS, "Error creating blob pixel "
        "masks");
    if (status != SUCCESS)
    {
        cleanup_io_routines(blob);
        free_the_blob();
        return;
    }

    /* Set up a standard header */
    setenv("PCS_WOID", "I4582", 1);
    strcpy(work_order_id, "I4582");
    report_header = ias_misc_initialize_report_header("",
        work_order_id, routine_name);
    if (!report_header)
    {
        assert_true_with_message(0, "Error initializing report header");
        cleanup_io_routines(blob);
        free_the_blob();
        return;
    }
    ias_misc_set_header_pathrow(report_header, 0, 0);
    blob->report_information.header = report_header;
    blob->report_information.directory = NULL;

    /* Open and read the inoperable detector pixel mask */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, inop_dets_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building inoperable detector "
                                 "pixel mask file name");
        cleanup_io_routines(blob);
        free_the_blob();
        return;
    }

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        cleanup_io_routines(blob);
        free_the_blob();
        return;
    }

    for (det = 0; det < NVRPDETS; det++)
    {
        return_value = fscanf(inf, "%d", &temp_int);
        if (return_value == EOF)
        {
            assert_true_with_message(0, "Reading %s", file_name);
            cleanup_io_routines(blob);
            free_the_blob();
            return;        
        }        
        /* Add the inoperable detectors to pixel mask */
        if (temp_int == 0)
        {
            assert_false(ias_pm_add_pixels(blob->pixel_mask[0][0], det, 1-1, 
                         NLINES, IAS_RPS_INOP_DETECTOR));
        }
    }
    fclose(inf);

    /* Open and read the dropped frame pixel mask 
       Note: In calval test file, the whole frame is dropped if the frame
             is labeled (0 in value), therefore, the read in dropped frames
             pixel mask is one value per frame in each SCA and each band */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, dropped_frames_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building dropped frame pixel "
                                 "maskfile name");
        cleanup_io_routines(blob);
        free_the_blob();
        return;
    }

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        cleanup_io_routines(blob);
        free_the_blob();
        return;
    }

    for (line = 0; line < NLINES; line++)
    {
        return_value = fscanf(inf, "%d", &temp_int);
        if (return_value == EOF)
        {
            assert_true_with_message(0, "Reading %s", file_name);
            cleanup_io_routines(blob);
            free_the_blob();
            return;        
        }
        /* Add the dropped frames to pixel mask */
        if (temp_int == 0)
        {
            for (det = 0; det < NVRPDETS; det++)
            {
                assert_false(ias_pm_add_pixels(blob->pixel_mask[0][0], det, 
                             line, 1, IAS_RPS_DROPLINE));
            }
        }
    }
    fclose(inf);

    /* Open and read the impulse noise pixel mask */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, impulse_noise_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building impulse noise pixel "
                                 "maskfile name");
        cleanup_io_routines(blob);
        free_the_blob();
        return;
    }

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        cleanup_io_routines(blob);
        free_the_blob();
        return;
    }

    for (line = 0; line < NLINES; line++)
    {
        for (det = 0; det < NVRPDETS; det++)
        {
            return_value = fscanf(inf, "%d", &temp_int);
            if (return_value == EOF)
            {
                assert_true_with_message(0, "Reading %s", file_name);
                return;        
            }        
            /* Add the impulse noise to pixel mask */
            if (temp_int == 0)
            {
                assert_false(ias_pm_add_pixels(blob->pixel_mask[0][0], det, 
                             line, 1, IAS_RPS_IMPNOISE));
            }
        }
    }
    fclose(inf);

    /* Open and read the saturated pixel pixel mask */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, sat_pixel_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building saturated pixel "
                                 "pixel mask file name");
        cleanup_io_routines(blob);
        free_the_blob();
        return;
    }

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        cleanup_io_routines(blob);
        free_the_blob();
        return;
    }

    for (line = 0; line < NLINES; line++)
    {
        for (det = 0; det < NVRPDETS; det++)
        {
            return_value = fscanf(inf, "%d", &temp_int);
            if (return_value == EOF)
            {
                assert_true_with_message(0, "Reading %s", file_name);
                cleanup_io_routines(blob);
                free_the_blob();
                return;        
            }        
            /* Add the saturated pixel to pixel mask
               Note: The calval saturated pixel mask file doesn't distinguish 
                     the saturated pixel type, therefore, 
                     IAS_RPS_ANALOG_SAT_HI type is used since it really doesn't
                     matter */
            if (temp_int == 0)
            {
                assert_false(ias_pm_add_pixels(blob->pixel_mask[0][0], det, 
                             line, 1, IAS_RPS_ANALOG_SAT_HI));
            }
        }
    }
    fclose(inf);

    /* This initializes the memory management and registers buffers through
       the algorithms pre-process routine */
    status = initialize_memory_management_buffers();
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error initializing memory management");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }
    status = oli_bias_determination_preprocess(blob);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error returned from: "
            "tirs_response_linearization_preprocess");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Register the MM_NON_PAN_PERFRAME_BIAS MM buffer to this routine so that 
       the API can't destroy the memory before we can retrieve the output */
    status = increment_consumer_count(MM_NON_PAN_PERFRAME_BIAS);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error incrementing "
            "MM_NON_PAN_PERFRAME_BIAS buffer");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Get the index of the test band which is band 1 and is 3rd in the list */
    test_band_index = blob->list_of_band_indices[2];
    test_vrp_index = blob->band_information[test_band_index]->vrp_band_index;

    /* Check an "empty" buffer for the A0 parameter */
    a0_parameter = (double *)preprocess_get_memory_buffer(
        MM_NON_PAN_BIAS_A0_PARAMETER, test_band_index, 0);
    if (!a0_parameter)
    {
        assert_true_with_message(0, "Error getting "
            "MM_NON_PAN_BIAS_A0_PARAMETER buffer");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }
    a0_parameter[0] = a0_b01_s01;
    set_buffer_filled_flag( MM_NON_PAN_BIAS_A0_PARAMETER, test_band_index, 0);

    /* Check an "empty" buffer for the shutter buffer */
    bias_from_shutter_buffer = (double *)preprocess_get_memory_buffer(
        MM_NON_PAN_MEAN_SHUTTER, test_band_index, 0);
    if (!bias_from_shutter_buffer)
    {
        assert_true_with_message(0, "Error getting "
            "MM_PAN_MEAN_SHUTTER buffer");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Open and read the detector level shutter buffer */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, bias_from_shutter_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building shutter file name");
        free(bias_from_shutter_buffer);
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        free(bias_from_shutter_buffer);
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    for (det = 0; det < NDETS; det++)
    {
        return_value = fscanf(inf, "%lf", &temp_double);
        if (return_value == EOF)
        {
            assert_true_with_message(0, "Reading %s", file_name);
            free(bias_from_shutter_buffer);
            cleanup_io_routines(blob);
            cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
            free_the_blob();
            return;        
        }        
        bias_from_shutter_buffer[det] = (double)temp_double;
    }
    fclose(inf);
    set_buffer_filled_flag( MM_NON_PAN_MEAN_SHUTTER, test_band_index, 0);

    /* Check an "empty" buffer for the VRP buffer */
    vrp_buffer = (uint16_t *)preprocess_get_memory_buffer(
                 IMAGE_L0R, test_vrp_index, 0);
    if (!vrp_buffer)
    {
        assert_true_with_message(0, "Error getting "
            "IMAGE_L0R VRP buffer");
        free(bias_from_shutter_buffer);
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Open and read the VRP data */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, vrp_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building VRP data file name" );
        free(bias_from_shutter_buffer);
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        free(vrp_buffer);
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    for (line = 0; line < NLINES; line++)
    {
        for (det = 0; det < NVRPDETS; det++)
        {
            return_value = fscanf(inf, "%d", &temp_int);
            if (return_value == EOF)
            {
                assert_true_with_message(0, "Reading %s", file_name);
                free(vrp_buffer);
                cleanup_io_routines(blob);
                cleanup_memory_management_buffers(
                    DISABLE_WARNING_DISPLAY);
                free_the_blob();
                return;        
            }        
            vrp_buffer[line * NVRPDETS + det] = (uint16_t)temp_int;
        }
    }
    fclose(inf);
    set_buffer_filled_flag( IMAGE_L0R, test_vrp_index, 0);

    /* Check an "empty" buffer for the a1_parameters */
    a1_parameter = (double *)preprocess_get_memory_buffer(
        MM_NON_PAN_BIAS_A1_PARAMETER, test_band_index, 0);
    if (!a1_parameter)
    {
        assert_true_with_message(0, "Error getting "
            "MM_NON_PAN_BIAS_A1_PARAMETER buffer");
        free(vrp_buffer);
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Populate a1_parameter with 0.236531 for now */
    for (det = 0; det < NDETS; det++)
        a1_parameter[det] = 0.236531;
    set_buffer_filled_flag( MM_NON_PAN_BIAS_A1_PARAMETER, test_band_index, 0);

    /* Check an "empty" buffer for the c1_parameters */
    c1_parameter = (double *)preprocess_get_memory_buffer(
        MM_NON_PAN_BIAS_A1_PARAMETER, test_band_index, 0);
    if (!c1_parameter)
    {
        assert_true_with_message(0, "Error getting "
            "MM_NON_PAN_BIAS_C1_PARAMETER buffer");
        free(vrp_buffer);
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Populate a1_parameter with 65.932693 for now */
    for (det = 0; det < NDETS; det++)
        c1_parameter[det] = 65.932693;
    set_buffer_filled_flag( MM_NON_PAN_BIAS_A1_PARAMETER, test_band_index, 0);

    /* Check an "empty" buffer for the bias buffer */
    bias_buffer = (double *)preprocess_get_memory_buffer(
        MM_NON_PAN_PERFRAME_BIAS, test_band_index, 0);
    if (!bias_buffer)
    {
        assert_true_with_message(0, "Error getting "
            "MM_NON_PAN_PERFRAME_BIAS buffer");
        free(vrp_buffer);
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* It is important that the consumer counts get set *after* all the
       increments */
    status = set_consumer_counts();
    if ( status != SUCCESS )
    {
        assert_true_with_message(0, "Error setting consumer counts");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Call the API with another band.  The API should return immediately.
       The only way to really check this is to ensure the data buffer wasn't
       changed at all. */

    /* Initialize a few bias_buffer for checking purpose here */
    bias_buffer[0] = 0.0;
    bias_buffer[494] = 494.0;
    bias_buffer[494*2] = 988.0;

    band_index = blob->list_of_band_indices[0];
    status = api_oli_bias_determination( blob, band_index, 0, 0 );
    assert_equal_with_message( status,SUCCESS, "Error returned from "
        "api_oli_bias_determination with band = 3");
    if ( status != SUCCESS )
    {
        free(vrp_buffer);
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Check some outputs with the original inputs, no changes */
    assert_double_equal(bias_buffer[0], 0.0);
    assert_double_equal(bias_buffer[494], 494.0);
    assert_double_equal(bias_buffer[494*2], 988.0);

    /* Call the API with TIRS band.  The API should return immediately.
       The only way to really check this is to ensure the data buffer wasn't
       changed at all. */
    band_index = blob->list_of_band_indices[1];
    status = api_oli_bias_determination( blob, band_index, 0, 0 );
    assert_equal_with_message( status,SUCCESS, "Error returned from "
        "api_oli_bias_determination with band = 11");
    if ( status != SUCCESS )
    {
        free(vrp_buffer);
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Check some outputs with the original inputs, no changes */
    assert_double_equal(bias_buffer[0], 0.0);
    assert_double_equal(bias_buffer[494], 494.0);
    assert_double_equal(bias_buffer[494*2], 988.0);

    /* Call the API with correct band */
    band_index = blob->list_of_band_indices[2];
    status = api_oli_bias_determination(blob, band_index, 0, 0);
    assert_equal_with_message(status, SUCCESS, "Error returned from "
        "api_tirs_bias_removal with band = 1");
    if (status != SUCCESS)
    {
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Open and read the bias result file generated from cal/val code
       */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, bias_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building result file name");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }
    result_buffer = malloc(NDETS * NLINES * sizeof(double));
    if (!result_buffer)
    {
        assert_true_with_message(0, "Allocating memory for bias " 
                                         "buffer");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        free(result_buffer);
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        ias_misc_free_report_header(report_header);
        return;
    }

    for (line = 0; line < NLINES; line++)
    {
        for (det = 0; det < NDETS; det++)
        {
            return_value = fscanf(inf, "%lf", &temp_double);
            if (return_value == EOF)
            {
                assert_true_with_message(0, "Reading %s", file_name);
                free(result_buffer);
                cleanup_io_routines(blob);
                cleanup_memory_management_buffers(
                    DISABLE_WARNING_DISPLAY);
                free_the_blob();
                return;        
            }        
            result_buffer[line * NDETS + det] = temp_double;
        }
    }
    fclose(inf);

    /* Compare the results.  Loop through the results and count the
       differences.  This comparison isn't done with asserts since there is
       no need for 494*255 assert statements which cost message buffer space.
       One assert for pass fail with a count is good enough. */
    difference_count = 0;
    for (pixel = 0; pixel < NDETS * NLINES; pixel++)
    {
        if (fabsf(bias_buffer[pixel] - result_buffer[pixel]) > 1e-2)
            difference_count++;
    }
    assert_equal_with_message(difference_count, 0, "There were %d "
        "differences out of %d values", difference_count, NDETS * NLINES);

    /* Free the image memory */
    status = decrement_consumer_count(MM_NON_PAN_PERFRAME_BIAS,
        test_band_index, 0);
    assert_equal_with_message( status, SUCCESS,
        "Error returned from decrement consumer count");
    status = decrement_consumer_count(MM_NON_PAN_PERFRAME_BIAS,
        blob->list_of_band_indices[0], 0);
    assert_equal_with_message( status, SUCCESS,
        "Error returned from decrement consumer count");

    free(result_buffer);
    cleanup_io_routines(blob);
    cleanup_memory_management_buffers(ENABLE_WARNING_DISPLAY);
    free_the_blob();
}

/*--------------------------------------------------------------------------
  test_algorithm_pan_band_thru_api

  This test sets up for a call to the algorithm through the API for pan band.  
  Also calls the pre-process routine to set up memory management. The input 
  ASCII files are extracted from calval binary test files with only pan band 
  and one sca used here. The bias from shutter file used is the average of the 
  pre- and post-acquisition shutter average. The per-frame bias are calculated
  from model parameters (A0 and VRP band data). The compared result file is 
  generated from cal/val V3.3 code. 
--------------------------------------------------------------------------*/
Ensure test_algorithm_pan_band_thru_api(void)
{
    FILE *inf;                     /* File pointer for all reads */
    char *test_directory;          /* test environment value */
    char file_name[PATH_MAX];      /* File name to open */
    uint16_t *vrp_buffer;          /* Buffer containing the input vrp data */
    double *bias_buffer;           /* Calculated bias data buffer */
    double *result_buffer;         /* Bias data buffer from calval code test */
    int temp_int;                  /* Temporary value read from file */
    double temp_double;            /* Temporary value read from file */
    int det;                       /* Detector loop counter */
    int line;                      /* Frame loop counter */
    int pixel;                     /* Pixel loop counter */
    int difference_count;          /* Numbers of differences between the 
                                      calculated value and expected results */
    int return_value;              /* Return values from fscanf read */ 
    char work_order_id[IAS_WORKORDER_ID_SIZE];
    char routine_name[] = "OLI bias determination tests";

    double a0_b08_s01_even = 0.448819; /* A0 coefficient for band 8 (pan band 
                                         even) and SCA 1 */
    double a0_b08_s01_odd = 0.559437;  /* A0 coefficient for band 8 (pan band
                                         odd) and SCA 1 */
    double *a0_pan_odd_parameter;      /* Pan odd A0 coefficient buffer */
    double *a0_pan_even_parameter;     /* Pan even A0 coefficient buffer */
    double *a1_pan_odd_parameter;      /* Pan odd A1 coefficient buffer */
    double *a1_pan_even_parameter;     /* Pan even A1 coefficient buffer */
    double *c1_pan_odd_parameter;      /* Pan odd C1 coefficient buffer  */
    double *c1_pan_even_parameter;     /* Pan even C1 coefficient buffer */
    double *shutter_pan_odd;           /* Pan odd shutter buffer  */
    double *shutter_pan_even;          /* Pan even shutter buffer */
    BLOB *blob;                    /* Declare a blob structure */
    IAS_REPORT_HEADER *report_header = NULL;
    int status;
    int band_index;
    int test_band_index;
    int test_vrp_index;
    const IAS_SATELLITE_ATTRIBUTES *sat_attributes;

    static char bpf_dir[] = "ias_lib/bpf";
    static char cpf_dir[] = "ias_lib/cpf";
    static char rlut_dir[] = "ias_lib/rlut";
    static char l0r_dir[] = "ias_base/rps/driver/output";
    static char bpf_name_oli[] = "test_LO8BPF20140310103310_20140310103345.07";
    static char cpf_name[] = "L8CPF20090101_20090331.01";
    static char rlut_name[] = "test_linearization_rlut.h5";
    static char l0r_name[] = "LO82220010052011265EDC00";
    static char l1r_name[] = "temp_l1r";

    /* Get the environment variable for the test data */
    test_directory = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (!test_directory)
    {
        assert_true_with_message(0, "Getting IAS_UNIT_TEST_DATA_DIR");
        return;
    }

    /* The attributes must be initialized for memory management to work */
    ias_sat_attr_initialize(IAS_L8);

    /* Allocate a blob and populate the minimum members for the test
       to succeed */
    blob = allocate_blob();
    if ( !blob )
    {
        assert_true_with_message( 0, "allocating blob" );
        return;
    }

    /* Initialize blob */
    status = snprintf(blob->cpf_filename, sizeof(blob->cpf_filename), 
             "%s/%s/%s", test_directory, cpf_dir, cpf_name);
    if ((status < 0) || (status >= sizeof(blob->cpf_filename)))
    {
        assert_true_with_message(FALSE, "Error building CPF file name");
        return;
    }
    status = snprintf(blob->bpf_filename_oli, sizeof(blob->bpf_filename_oli), 
             "%s/%s/%s", test_directory, bpf_dir, bpf_name_oli);
    if ((status < 0) || (status >= sizeof(blob->bpf_filename_oli)))
    {
        assert_true_with_message(FALSE, "Error building OLI BPF file name");
        return;
    }
    strcpy(blob->bpf_filename_tirs, "");
    status = snprintf(blob->rlut_filename, sizeof(blob->rlut_filename), 
             "%s/%s/%s", test_directory, rlut_dir, rlut_name);
    if ((status < 0) || (status >= sizeof(blob->rlut_filename)))
    {
        assert_true_with_message(FALSE, "Error building RLUT file name");
        return;
    }
    sat_attributes = ias_sat_attr_get_attributes();
    blob->total_bands = sat_attributes->total_bands;
    blob->number_of_bands = 1;
    blob->band_fields_initialized = 1;

    if ( finish_blob_initialization())
    {
        assert_true_with_message( 0, "initializing blob" );
        return;
    }
    blob->list_of_bands[0] = PAN_BAND_NUMBER;
    blob->list_of_band_indices[0] =
        ias_sat_attr_convert_band_number_to_index( blob->list_of_bands[0] );
    if ( blob->list_of_band_indices[0] == ERROR )
    {
        assert_true_with_message( 0, "Converting band number to index");
        free_the_blob();
        return;
    }
    blob->bias_source = MEAN_ACQUISITION_SHUTTER;
    blob->per_frame_bias_flag = PER_FRAME_BIAS;
    strcpy(blob->date_acquired, "2010:205:16:15:27.04520");
    blob->process_oli_bands = 1;
    blob->process_tirs_bands = 0;

    /* Initialize I/O routines */
    status = snprintf(blob->l0r_file_name, sizeof(blob->l0r_file_name), 
             "%s/%s/%s", test_directory, l0r_dir, l0r_name);
    if ((status < 0) || (status >= sizeof(blob->l0r_file_name)))
    {
        assert_true_with_message(FALSE, "Error building L0R data file name");
        free_the_blob();
        return;
    }
    status = snprintf(blob->l1r_file_name, sizeof(blob->l1r_file_name), "%s", 
             l1r_name);
    if ((status < 0) || (status >= sizeof(blob->l1r_file_name)))
    {
        assert_true_with_message(FALSE, "Error building L1R data file name");
        free_the_blob();
        return;
    }
    ias_log_set_output_level( IAS_LOG_LEVEL_DISABLE );
    status = initialize_io_routines( blob );
    if ( status )
    {
        free_the_blob();
        assert_true_with_message( 0,
            "initialize_io_routines returned an error" );
        return;
    }
    ias_log_set_output_level( IAS_LOG_LEVEL_INFO );

    /* Read OLI BPF */
    status = read_oli_tirs_bpf(blob);
    assert_equal_with_message(status, SUCCESS, "Error reading OLI BPF "
        "masks");
    if (status != SUCCESS)
    {
        cleanup_io_routines(blob);
        free_the_blob();
        return;
    }

    /* Create blob pixel_masks */
    status = create_blob_pixel_mask();
    assert_equal_with_message(status, SUCCESS, "Error creating blob pixel "
        "masks");
    if (status != SUCCESS)
    {
        cleanup_io_routines(blob);
        free_the_blob();
        return;
    }

    /* Set up a standard header */
    setenv("PCS_WOID", "I4582", 1);
    strcpy(work_order_id, "I4582");
    report_header = ias_misc_initialize_report_header("",
        work_order_id, routine_name);
    if (!report_header)
    {
        assert_true_with_message(0, "Error initializing report header");
        cleanup_io_routines(blob);
        free_the_blob();
        return;
    }
    ias_misc_set_header_pathrow(report_header, 0, 0);
    blob->report_information.header = report_header;
    blob->report_information.directory = NULL;

    /* Open and read the inoperable detector pixel mask */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, inop_dets_pan_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building inoperable detector "
                                 "pixel mask file name");
        cleanup_io_routines(blob);
        free_the_blob();
        return;
    }

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        cleanup_io_routines(blob);
        free_the_blob();
        return;
    }

    /* Note: The cal/val test data file only has 12 VRP dets for the pan band
             and pixel mask data, therefore, these 12 dets data will be used 
             for odd and even VRP detectors separately. Not having 24 dets VRP 
             data available is due to the use of a flawed data converter based
             on cal/val response */
    for (det = 0; det < NVRPDETS; det++)
    {
        return_value = fscanf(inf, "%d", &temp_int);
        if (return_value == EOF)
        {
            assert_true_with_message(0, "Reading %s", file_name);
            cleanup_io_routines(blob);
            free_the_blob();
            return;        
        }        
        /* Add the inoperable detectors to pixel mask */
        if (temp_int == 0)
        {
            assert_false(ias_pm_add_pixels(blob->pixel_mask[7][0], det, 1-1, 
                         NPANLINES, IAS_RPS_INOP_DETECTOR));
            assert_false(ias_pm_add_pixels(blob->pixel_mask[7][0], 
                         NVRPDETS + det, 1-1, NPANLINES, 
                         IAS_RPS_INOP_DETECTOR));
        }
    }
    fclose(inf);

    /* Open and read the dropped frame pixel mask 
       Note: In calval test file, the whole frame is dropped if the frame
             is labeled (0 in value), therefore, the read in dropped frames
             pixel mask is one value per frame in each SCA and each band */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, dropped_frames_pan_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building dropped frame pixel "
                                 "maskfile name");
        cleanup_io_routines(blob);
        free_the_blob();
        return;
    }

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        cleanup_io_routines(blob);
        free_the_blob();
        return;
    }

    for (line = 0; line < NPANLINES; line++)
    {
        return_value = fscanf(inf, "%d", &temp_int);
        if (return_value == EOF)
        {
            assert_true_with_message(0, "Reading %s", file_name);
            cleanup_io_routines(blob);
            free_the_blob();
            return;        
        }
        /* Add the dropped frames to pixel mask */
        if (temp_int == 0)
        {
            for (det = 0; det < NVRPDETS; det++)
            {
                assert_false(ias_pm_add_pixels(blob->pixel_mask[7][0], det, 
                             line, 1, IAS_RPS_DROPLINE));
                assert_false(ias_pm_add_pixels(blob->pixel_mask[7][0], 
                             NVRPDETS + det, line, 1, IAS_RPS_DROPLINE));
            }
        }
    }
    fclose(inf);

    /* Open and read the impulse noise pixel mask */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, impulse_noise_pan_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building impulse noise pixel "
                                 "mask file name");
        cleanup_io_routines(blob);
        free_the_blob();
        return;
    }

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        return;
    }

    for (line = 0; line < NPANLINES; line++)
    {
        for (det = 0; det < NVRPDETS; det++)
        {
            return_value = fscanf(inf, "%d", &temp_int);
            if (return_value == EOF)
            {
                assert_true_with_message(0, "Reading %s", file_name);
                return;        
                cleanup_io_routines(blob);
                free_the_blob();
            }        
            /* Add the impulse noise to pixel mask */
            if (temp_int == 0)
            {
                ias_pm_add_pixels(blob->pixel_mask[7][0], det, line, 1, 
                             IAS_RPS_IMPNOISE);
                ias_pm_add_pixels(blob->pixel_mask[7][0], NVRPDETS + det, line,
                             1, IAS_RPS_IMPNOISE);
            }
        }
    }
    fclose(inf);

    /* Open and read the saturated pixels pixel mask */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, sat_pixel_pan_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building saturated pixels "
                                 "pixel mask file name");
        cleanup_io_routines(blob);
        free_the_blob();
        return;
    }

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        cleanup_io_routines(blob);
        free_the_blob();
        return;
    }

    for (line = 0; line < NPANLINES; line++)
    {
        for (det = 0; det < NVRPDETS; det++)
        {
            return_value = fscanf(inf, "%d", &temp_int);
            if (return_value == EOF)
            {
                assert_true_with_message(0, "Reading %s", file_name);
                cleanup_io_routines(blob);
                free_the_blob();
                return;        
            }        
            /* Add the saturated pixel to pixel mask
               Note: The calval saturated pixel mask file doesn't distinguish 
                     the saturated pixel type, therefore, 
                     IAS_RPS_ANALOG_SAT_HI type is used since it really doesn't
                     matter */
            if (temp_int == 0)
            {
                ias_pm_add_pixels(blob->pixel_mask[7][0], det, line, 1, 
                             IAS_RPS_ANALOG_SAT_HI);
                ias_pm_add_pixels(blob->pixel_mask[7][0], NVRPDETS + det, 
                             line, 1, IAS_RPS_ANALOG_SAT_HI);
            }
        }
    }
    fclose(inf);

    /* This initializes the memory management and registers buffers through
       the algorithms pre-process routine */
    status = initialize_memory_management_buffers();
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error initializing memory management");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }
    status = oli_bias_determination_preprocess(blob);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error returned from: "
            "tirs_response_linearization_preprocess");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Register the MM_NON_PAN_PERFRAME_BIAS MM buffer to this routine so that 
       the API can't destroy the memory before we can retrieve the output */
    status = increment_consumer_count(MM_NON_PAN_PERFRAME_BIAS);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error incrementing "
            "MM_NON_PAN_PERFRAME_BIAS buffer");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }
    /* Register the MM_PAN_PERFRAME_BIAS MM buffer to this routine so that 
       the API can't destroy the memory before we can retrieve the output */
    status = increment_consumer_count(MM_PAN_PERFRAME_BIAS);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error incrementing "
            "MM_PAN_PERFRAME_BIAS buffer");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Get the index of the test band which is band 1 and is 3rd in the list */
    test_band_index = blob->list_of_band_indices[0];
    test_vrp_index = blob->band_information[test_band_index]->vrp_band_index;

    /* Check an "empty" buffer for the a0_pan_odd_parameter */
    a0_pan_odd_parameter = (double *)preprocess_get_memory_buffer(
        MM_PAN_ODD_BIAS_A0_PARAMETER, test_band_index, 0);
    if (!a0_pan_odd_parameter)
    {
        assert_true_with_message(0, "Error getting "
            "MM_PAN_ODD_BIAS_A0_PARAMETER buffer");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }
    a0_pan_odd_parameter[0] = a0_b08_s01_odd;
    set_buffer_filled_flag( MM_PAN_ODD_BIAS_A0_PARAMETER, test_band_index, 0);

     /* Check an "empty" buffer for the a0_pan_even_parameter */
    a0_pan_even_parameter = (double *)preprocess_get_memory_buffer(
        MM_PAN_EVEN_BIAS_A0_PARAMETER, test_band_index, 0);
    if (!a0_pan_even_parameter)
    {
        assert_true_with_message(0, "Error getting "
            "MM_PAN_ODD_BIAS_A0_PARAMETER buffer");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }
    a0_pan_even_parameter[0] = a0_b08_s01_even;
    set_buffer_filled_flag( MM_PAN_EVEN_BIAS_A0_PARAMETER, test_band_index, 0);

    /* Check an "empty" buffer for the pan odd shutter buffer */
    shutter_pan_odd = (double *)preprocess_get_memory_buffer(
        MM_PAN_ODD_MEAN_SHUTTER, test_band_index, 0);
    if (!shutter_pan_odd)
    {
        assert_true_with_message(0, "Error getting "
            "MM_PAN_ODD_MEAN_SHUTTER buffer");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Check an "empty" buffer for the pan odd shutter buffer */
    shutter_pan_even = (double *)preprocess_get_memory_buffer(
        MM_PAN_EVEN_MEAN_SHUTTER, test_band_index, 0);
    if (!shutter_pan_even)
    {
        assert_true_with_message(0, "Error getting "
            "MM_PAN_EVEN_MEAN_SHUTTER buffer");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        ias_misc_free_report_header(report_header);
        return;
    }

    /* Open and read the detector level shutter buffer */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, shutter_pan_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building shutter file name");
        return;
    }

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        ias_misc_free_report_header(report_header);
        return;
    }

    for (det = 0; det < NPANDETS; det++)
    {
        return_value = fscanf(inf, "%lf", &temp_double);
        if (return_value == EOF)
        {
            assert_true_with_message(0, "Reading %s", file_name);
            cleanup_io_routines(blob);
            cleanup_memory_management_buffers(
                DISABLE_WARNING_DISPLAY);
            free_the_blob();
            ias_misc_free_report_header(report_header);
            return;        
        }        
        shutter_pan_odd[det] = temp_double;
        shutter_pan_even[det] = temp_double;
    }
    fclose(inf);
    set_buffer_filled_flag( MM_PAN_ODD_MEAN_SHUTTER, test_band_index, 0);
    set_buffer_filled_flag( MM_PAN_EVEN_MEAN_SHUTTER, test_band_index, 0);

    /* Check an "empty" buffer for the VRP buffer */
    vrp_buffer = (uint16_t *)preprocess_get_memory_buffer(
                 IMAGE_L0R, test_vrp_index, 0);
    if (!vrp_buffer)
    {
        assert_true_with_message(0, "Error getting "
            "IMAGE_L0R VRP buffer");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Open and read the VRP data */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, vrp_pan_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building VRP data file name");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }
    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    for (line = 0; line < NPANLINES; line++)
    {
        for (det = 0; det < NVRPDETS; det++)
        {
            return_value = fscanf(inf, "%d", &temp_int);
            if (return_value == EOF)
            {
                assert_true_with_message(0, "Reading %s", file_name);
                cleanup_io_routines(blob);
                cleanup_memory_management_buffers(
                    DISABLE_WARNING_DISPLAY);
                free_the_blob();
                return;        
            }        
            vrp_buffer[line * NPANVRPDETS + det] = (uint16_t)temp_int;
            vrp_buffer[line * NPANVRPDETS + NVRPDETS + det ] 
                = (uint16_t)temp_int;
        }
    }
    fclose(inf);
    set_buffer_filled_flag( IMAGE_L0R, test_vrp_index, 0);

    /* Check an "empty" buffer for the a1_pan_odd_parameters */
    a1_pan_odd_parameter = (double *)preprocess_get_memory_buffer(
        MM_PAN_ODD_BIAS_A1_PARAMETER, test_band_index, 0);
    if (!a1_pan_odd_parameter)
    {
        assert_true_with_message(0, "Error getting "
            "MM_PAN_ODD_BIAS_A1_PARAMETER buffer");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Check an "empty" buffer for the a1_pan_even_parameters */
    a1_pan_even_parameter = (double *)preprocess_get_memory_buffer(
        MM_PAN_EVEN_BIAS_A1_PARAMETER, test_band_index, 0);
    if (!a1_pan_even_parameter)
    {
        assert_true_with_message(0, "Error getting "
            "MM_PAN_EVEN_BIAS_A1_PARAMETER buffer");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Populate a1_parameter with 0.236531 for now */
    for (det = 0; det < NDETS; det++)
    {
        a1_pan_odd_parameter[det] = 0.236531;
        a1_pan_even_parameter[det] = 0.236531;
    }
    set_buffer_filled_flag( MM_PAN_ODD_BIAS_A1_PARAMETER, test_band_index, 0);
    set_buffer_filled_flag( MM_PAN_EVEN_BIAS_A1_PARAMETER, test_band_index, 0);

    /* Check an "empty" buffer for the c1_pan_odd_parameters */
    c1_pan_odd_parameter = (double *)preprocess_get_memory_buffer(
        MM_PAN_ODD_BIAS_C1_PARAMETER, test_band_index, 0);
    if (!c1_pan_odd_parameter)
    {
        assert_true_with_message(0, "Error getting "
            "MM_PAN_ODD_BIAS_C1_PARAMETER buffer");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Check an "empty" buffer for the c1_pan_even_parameters */
    c1_pan_even_parameter = (double *)preprocess_get_memory_buffer(
        MM_PAN_EVEN_BIAS_C1_PARAMETER, test_band_index, 0);
    if (!c1_pan_even_parameter)
    {
        assert_true_with_message(0, "Error getting "
            "MM_PAN_EVEN_BIAS_C1_PARAMETER buffer");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Populate a1_parameter with 65.932693 for now */
    for (det = 0; det < NDETS; det++)
    {
        c1_pan_odd_parameter[det] = 65.932693;
        c1_pan_even_parameter[det] = 65.932693;
    }
    set_buffer_filled_flag( MM_PAN_ODD_BIAS_C1_PARAMETER, test_band_index, 0);
    set_buffer_filled_flag( MM_PAN_EVEN_BIAS_C1_PARAMETER, test_band_index, 0);

    /* Check an "empty" buffer for the bias buffer */
    bias_buffer = (double *)preprocess_get_memory_buffer( MM_PAN_PERFRAME_BIAS,
        test_band_index, 0);
    if (!bias_buffer)
    {
        assert_true_with_message(0, "Error getting "
            "MM_PAN_PERFRAME_BIAS buffer");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* It is important that the consumer counts get set *after* all the
       increments */
    status = set_consumer_counts();
    if ( status != SUCCESS )
    {
        assert_true_with_message(0, "Error setting consumer counts");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Call the API with correct band */
    band_index = blob->list_of_band_indices[0];
    status = api_oli_bias_determination(blob, band_index, 0, 0);
    assert_equal_with_message(status, SUCCESS, "Error returned from "
        "api_tirs_bias_removal with band = 1");
    if (status != SUCCESS)
    {
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Open and read the bias result file generated from cal/val code
       */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, bias_pan_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building result file name");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }
    result_buffer = malloc(NPANDETS * NPANLINES * sizeof(double));
    if (!result_buffer)
    {
        assert_true_with_message(0, "Allocating memory for bias " 
                                         "buffer");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        free(result_buffer);
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    for (line = 0; line < NPANLINES; line++)
    {
        for (det = 0; det < NPANDETS; det++)
        {
            return_value = fscanf(inf, "%lf", &temp_double);
            if (return_value == EOF)
            {
                assert_true_with_message(0, "Reading %s", file_name);
                free(result_buffer);
                cleanup_io_routines(blob);
                cleanup_memory_management_buffers(
                    DISABLE_WARNING_DISPLAY);
                free_the_blob();
                return;        
            }        
            result_buffer[line * NPANDETS + det] = temp_double;
        }
    }
    fclose(inf);

    /* Compare the results.  Loop through the results and count the
       differences.  This comparison isn't done with asserts since there is
       no need for 494*255 assert statements which cost message buffer space.
       One assert for pass fail with a count is good enough. */
    difference_count = 0;
    for (pixel = 0; pixel < NPANDETS * NPANLINES; pixel++)
    {
        if (fabsf(bias_buffer[pixel] - result_buffer[pixel]) > 1e-2)
            difference_count++;
    }
    assert_equal_with_message(difference_count, 0, "There were %d "
        "differences out of %d values", difference_count, NDETS * NLINES);

    /* Free the image memory */
    status = decrement_consumer_count(MM_PAN_PERFRAME_BIAS, test_band_index, 0);
    assert_equal_with_message( status, SUCCESS,
        "Error returned from decrement consumer count");
    status = decrement_consumer_count(MM_PAN_PERFRAME_BIAS,
        blob->list_of_band_indices[0], 0);
    assert_equal_with_message( status, SUCCESS,
        "Error returned from decrement consumer count");

    free(result_buffer);
    cleanup_io_routines(blob);
    cleanup_memory_management_buffers(ENABLE_WARNING_DISPLAY);
    free_the_blob();
}

/*--------------------------------------------------------------------------
   test_algorithm_directly_non_pan_band_a1_c1

   This test executes the algorithm with a subset of the Cal/Val prototype
   input and output. The input binary files are extracted from calval binary 
   test files with only one band and one sca used one at a time. A non_pan 
   band SCA are tested. The detector level bias are estimated with a1/C1 
   parameters (option 5 in the updated ADD 3.1c and the per-frame bias are 
   calculated with model parameters (A0 and VRP band data). The compared result
   file is generated from cal/val prototype V3.4 code.
---------------------------------------------------------------------------*/
Ensure test_algorithm_directly_non_pan_band_a1_c1(void)
{
    FILE *inf;                     /* File pointer for all reads */
    char *test_directory;          /* Test environment value */
    char file_name[PATH_MAX];      /* File name to open */
    short *vrp_buffer;             /* Buffer containing the input vrp data */
    double *bias_buffer;           /* Calculated bias data buffer */
    double *result_buffer;         /* Bias data buffer from calval test */
    double *a1_parameter;          /* a1_parameter buffer from calval test */
    double *c1_parameter;          /* c1_parameter buffer from calval test */
    int det;                       /* Detector loop counter */
    int line;                      /* Frame loop counter */
    int pixel;                     /* Pixel loop counter */
    int difference_count;          /* Numbers of differences between the 
                                      calculated value and expected results */
    char work_order_id[IAS_WORKORDER_ID_SIZE];
    char routine_name[] = "OLI bias determination tests";
    double a0_b01_s01 = 0.163879;  /* A0 coefficient for band 1 and SCA 1 */
    IAS_PIXEL_MASK *pm;            /* Pixel mask for VRP data */
    int band_number;               /* 1-based band number */
    int sca_number;                /* 1-based SCA number */
    IAS_REPORT report_info;
    IAS_REPORT_HEADER *report_header = NULL;
    int status;
    short *dropped_frame;
    double * bias_from_shutter_buffer = NULL;
    int a1_c1_estimate_flag = 1;   /* Using a1/C1 estimation option */

    /* Get the environment variable for the test data */
    test_directory = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (!test_directory)
    {
        assert_true_with_message(0, "Getting IAS_UNIT_TEST_DATA_DIR");
        return;
    }

    /* Set up a standard header */
    setenv("PCS_WOID", "I4582", 1);
    strcpy(work_order_id, "I4582");
    report_header = ias_misc_initialize_report_header("",
        work_order_id, routine_name);
    if (!report_header)
    {
        assert_true_with_message(0, "Error initializing report header");
        return;
    }
    ias_misc_set_header_pathrow(report_header, 0, 0);
    report_info.header = report_header;
    report_info.directory = NULL;

    /* Create one pixel mask */
    pm = ias_pm_create(1, 1, NVRPDETS, NLINES);
    assert_not_equal(pm, NULL);

    /* Open and read the dropped frame pixel mask */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory, 
             data_dir, dropped_frames_binary_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building the dropped frame "
                                 "pixel mask file name");
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    /* Allocate memory for dropped frame */
    dropped_frame = malloc(NVRPDETS * NLINES * sizeof(short));
    if (!dropped_frame)
    {
        assert_true_with_message(0, "Allocating memory for dropped frame" );
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    inf = fopen(file_name, "rb");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        free(dropped_frame);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    fread(dropped_frame, NVRPDETS * NLINES * sizeof(short), 1, inf);
    fclose(inf);

    for (det = 0; det < NVRPDETS; det++)
    {
        for (line = 0; line < NLINES; line++)
        {
            /* Add the dropped frames to pixel mask if a frame_drooped pixel 
               is found */
            if (dropped_frame[det * NLINES + line] != 1)
                assert_false(ias_pm_add_pixels(pm, det, line, 1, 
                             IAS_RPS_DROPLINE));
        }
    }

    /* Open and read the VRP data */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, vrp_binary_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building VRP data file name");
        free(dropped_frame);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }
    vrp_buffer = malloc(NVRPDETS * NLINES * sizeof(short));
    if (!vrp_buffer)
    {
        assert_true_with_message(0, "Allocating memory for dark_buffer" );
        free(dropped_frame);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    inf = fopen(file_name, "rb");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        free(dropped_frame);
        free(vrp_buffer);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    fread(vrp_buffer, NVRPDETS * NLINES * sizeof(short), 1, inf);
    fclose(inf);

    /* Set band_number and sca_number */
    band_number = 1;
    sca_number = 1;

    /* Allocate memory for a1_parameter */
    a1_parameter = malloc(NDETS * sizeof(double));
    if (!a1_parameter)
    {
        assert_true_with_message(0, "Allocating memory for bias" );
        free(vrp_buffer);
        free(dropped_frame);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    /* Open and read the a1_parameter buffer */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, a1_parameter_binary_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building a1_parameter file "
                                 "name");
        free(vrp_buffer);
        free(dropped_frame);
        free(a1_parameter);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    /* Polulate a1_parameter */
    inf = fopen(file_name, "rb");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        free(vrp_buffer);
        free(dropped_frame);
        free(a1_parameter);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    fread(a1_parameter, NDETS * sizeof(double), 1, inf);
    fclose(inf);

    /* Allocate memory for c1_parameter */
    c1_parameter = malloc(NDETS * sizeof(double));
    if (!c1_parameter)
    {
        assert_true_with_message(0, "Allocating memory for bias" );
        free(vrp_buffer);
        free(dropped_frame);
        free(a1_parameter);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    /* Open and read the c1_parameter buffer */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, c1_parameter_binary_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building c1_parameter file "
                                 "name");
        free(vrp_buffer);
        free(dropped_frame);
        free(a1_parameter);
        free(c1_parameter);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    /* Polulate c1_parameter */
    inf = fopen(file_name, "rb");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        free(vrp_buffer);
        free(dropped_frame);
        free(a1_parameter);
        free(c1_parameter);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    fread(c1_parameter, NDETS * sizeof(double), 1 , inf);
    fclose(inf);

    /* Allocate memory for bias */
    bias_buffer = malloc(NDETS * NLINES * sizeof(double));
    if (!bias_buffer)
    {
        assert_true_with_message(0, "Allocating memory for bias" );
        free(vrp_buffer);
        free(dropped_frame);
        free(a1_parameter);
        free(c1_parameter);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    /* Call the algorithm and test it for band 1 and SCA 1 */
    status = oli_bias_determination((uint16_t *)vrp_buffer, band_number, 
             sca_number, NDETS, NLINES, NVRPDETS, NLINES, PER_FRAME_BIAS, 
             a1_c1_estimate_flag, NON_PAN, bias_from_shutter_buffer, 
             a1_parameter, c1_parameter, a0_b01_s01, pm, &report_info, 
             bias_buffer);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Calling oli bias determination" );
        free(dropped_frame);
        free(vrp_buffer);
        free(bias_buffer);
        free(a1_parameter);
        free(c1_parameter);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    /* Open and read the bias results generated from cal/val prototype code */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, bias_binary_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building result data file name");
        free(dropped_frame);
        free(vrp_buffer);
        free(bias_buffer);
        free(a1_parameter);
        free(c1_parameter);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }
    result_buffer = malloc(NDETS * NLINES * sizeof(double));
    if (!result_buffer)
    {
        assert_true_with_message(0, "Allocating memory for bias " 
                                         "buffer");
        free(dropped_frame);
        free(vrp_buffer);
        free(bias_buffer);
        free(a1_parameter);
        free(c1_parameter);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    inf = fopen(file_name, "rb");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        free(dropped_frame);
        free(vrp_buffer);
        free(bias_buffer);
        free(result_buffer);
        free(a1_parameter);
        free(c1_parameter);
        ias_pm_destroy(pm);
        ias_misc_free_report_header(report_header);
        return;
    }

    fread(result_buffer, NDETS * NLINES * sizeof(double), 1, inf);
    fclose(inf);

    /* Compare the results.  Loop through the results and count the
       differences.  This comparison isn't done with asserts since there is
       no need for 494*255 assert statements which cost message buffer space.
       One assert for pass fail with a count is good enough. */
    difference_count = 0;
    for (pixel = 0; pixel < NDETS * NLINES; pixel++)
    {
        if (fabsf(bias_buffer[pixel] - result_buffer[pixel]) > 1e-4)
            difference_count++;
    }
    assert_equal_with_message(difference_count, 0, "There were %d "
        "differences out of %d values", difference_count, NDETS * NLINES);

    /* Free the allocated memory */
    ias_pm_destroy(pm);
    free(dropped_frame);
    free(vrp_buffer);
    free(bias_buffer);
    free(result_buffer);
    free(a1_parameter);
    free(c1_parameter);
    ias_misc_free_report_header( report_header );
} 

/*--------------------------------------------------------------------------
  test_algorithm_thru_api_a1_c1

  This test sets up for a call to the algorithm through the API.  Also calls
  the pre-process routine to set up memory management. The input binary files 
  are extracted from calval binary test files with only one band and one sca 
  used here. A non_pan band SCA are tested. The bias are estimated with a1/C1 
  parameters (option 5 in the updated ADD 3.1c) and the per-frame bias are 
  calculated with model parameters (A0 and VRP band data). The compared result 
  file is generated from cal/val V3.4 code. 
--------------------------------------------------------------------------*/
Ensure test_algorithm_thru_api_a1_c1(void)
{
    FILE *inf;                     /* File pointer for all reads */
    char *test_directory;          /* test environment value */
    char file_name[PATH_MAX];      /* File name to open */
    uint16_t *vrp_buffer;          /* Buffer containing the input vrp data */
    double *bias_buffer;           /* Calculated bias data buffer */
    double *result_buffer;         /* Bias data buffer from calval code test */
    int det;                       /* Detector loop counter */
    int line;                      /* Frame loop counter */
    int pixel;                     /* Pixel loop counter */
    int difference_count;          /* Numbers of differences between the 
                                      calculated value and expected results */
    char work_order_id[IAS_WORKORDER_ID_SIZE];
    char routine_name[] = "OLI bias determination tests";

    double a0_b01_s01 = 0.163879;  /* A0 coefficient for band 1 and SCA 1 */
    double *a0_parameter;          /* Buffer containing A0 coefficient  */
    double *a1_parameter;          /* a1_parameter buffer from calval test */
    double *c1_parameter;          /* c1_parameter buffer from calval test */
    BLOB *blob;                    /* Declare a blob structure */
    IAS_REPORT_HEADER *report_header = NULL;
    int status;
    int band_index;
    int test_band_index;
    int test_vrp_index;
    const IAS_SATELLITE_ATTRIBUTES *sat_attributes;
    short *dropped_frame;

    static char bpf_dir[] = "ias_lib/bpf";
    static char cpf_dir[] = "ias_lib/cpf";
    static char rlut_dir[] = "ias_lib/rlut";
    static char l0r_dir[] = "ias_base/rps/driver/output";
    static char bpf_name_oli[] = "test_LO8BPF20140310103310_20140310103345.07";
    static char cpf_name[] = "L8CPF20090101_20090331.01";
    static char rlut_name[] = "test_linearization_rlut.h5";
    static char l0r_name[] = "LO82220010052011265EDC00";
    static char l1r_name[] = "temp_l1r";

    /* Get the environment variable for the test data */
    test_directory = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (!test_directory)
    {
        assert_true_with_message(0, "Getting IAS_UNIT_TEST_DATA_DIR");
        return;
    }

    /* The attributes must be initialized for memory management to work */
    ias_sat_attr_initialize(IAS_L8);

    /* Allocate a blob and populate the minimum members for the test
       to succeed */
    blob = allocate_blob();
    if ( !blob )
    {
        assert_true_with_message( 0, "allocating blob" );
        return;
    }

    /* Initialize blob */
    status = snprintf(blob->cpf_filename, sizeof(blob->cpf_filename), 
             "%s/%s/%s", test_directory, cpf_dir, cpf_name);
    if ((status < 0) || (status >= sizeof(blob->cpf_filename)))
    {
        assert_true_with_message(FALSE, "Error building CPF file name");
        return;
    }
    status = snprintf(blob->bpf_filename_oli, sizeof(blob->bpf_filename_oli), 
             "%s/%s/%s", test_directory, bpf_dir, bpf_name_oli);
    if ((status < 0) || (status >= sizeof(blob->bpf_filename_oli)))
    {
        assert_true_with_message(FALSE, "Error building OLI BPF file name");
        return;
    }
    strcpy(blob->bpf_filename_tirs, "");
    status = snprintf(blob->rlut_filename, sizeof(blob->rlut_filename), 
             "%s/%s/%s", test_directory, rlut_dir, rlut_name);
    if ((status < 0) || (status >= sizeof(blob->rlut_filename)))
    {
        assert_true_with_message(FALSE, "Error building RLUT file name");
        return;
    }
    sat_attributes = ias_sat_attr_get_attributes();
    blob->total_bands = sat_attributes->total_bands;
    blob->number_of_bands = 3;
    blob->band_fields_initialized = 1;

    if ( finish_blob_initialization())
    {
        assert_true_with_message( 0, "initializing blob" );
        return;
    }
    blob->list_of_bands[0] = 3;
    blob->list_of_bands[1] = 11;
    blob->list_of_bands[2] = 1;
    blob->list_of_band_indices[0] =
        ias_sat_attr_convert_band_number_to_index( blob->list_of_bands[0] );
    blob->list_of_band_indices[1] =
        ias_sat_attr_convert_band_number_to_index( blob->list_of_bands[1] );
    blob->list_of_band_indices[2] =
        ias_sat_attr_convert_band_number_to_index( blob->list_of_bands[2] );
    if ( blob->list_of_band_indices[0] == ERROR ||
         blob->list_of_band_indices[1] == ERROR ||
         blob->list_of_band_indices[2] == ERROR )
    {
        assert_true_with_message( 0, "Converting band number to index");
        free_the_blob();
        return;
    }
    blob->bias_source = MEAN_ACQUISITION_SHUTTER;
    blob->per_frame_bias_flag = PER_FRAME_BIAS;
    strcpy(blob->date_acquired, "2010:205:16:15:27.04520");
    blob->process_oli_bands = 1;
    blob->process_tirs_bands = 0;

    /* Initialize I/O routines */
    status = snprintf(blob->l0r_file_name, sizeof(blob->l0r_file_name), 
             "%s/%s/%s", test_directory, l0r_dir, l0r_name);
    if ((status < 0) || (status >= sizeof(blob->l0r_file_name)))
    {
        assert_true_with_message(FALSE, "Error building L0R data file name");
        free_the_blob();
        return;
    }
    status = snprintf(blob->l1r_file_name, sizeof(blob->l1r_file_name), "%s", 
             l1r_name);
    if ((status < 0) || (status >= sizeof(blob->l1r_file_name)))
    {
        assert_true_with_message(FALSE, "Error building L1R data file name");
        return;
    }
    ias_log_set_output_level( IAS_LOG_LEVEL_DISABLE );
    status = initialize_io_routines( blob );
    if ( status )
    {
        free_the_blob();
        assert_true_with_message( 0,
            "initialize_io_routines returned an error" );
        return;
    }
    ias_log_set_output_level( IAS_LOG_LEVEL_INFO );

    /* Read OLI BPF */
    status = read_oli_tirs_bpf(blob);
    assert_equal_with_message(status, SUCCESS, "Error reading OLI BPF "
        "masks");
    if (status != SUCCESS)
    {
        cleanup_io_routines(blob);
        free_the_blob();
        return;
    }

    /* Create blob pixel_masks */
    status = create_blob_pixel_mask();
    assert_equal_with_message(status, SUCCESS, "Error creating blob pixel "
        "masks");
    if (status != SUCCESS)
    {
        cleanup_io_routines(blob);
        free_the_blob();
        return;
    }

    /* Set up a standard header */
    setenv("PCS_WOID", "I4582", 1);
    strcpy(work_order_id, "I4582");
    report_header = ias_misc_initialize_report_header("",
        work_order_id, routine_name);
    if (!report_header)
    {
        assert_true_with_message(0, "Error initializing report header");
        cleanup_io_routines(blob);
        free_the_blob();
        return;
    }
    ias_misc_set_header_pathrow(report_header, 0, 0);
    blob->report_information.header = report_header;
    blob->report_information.directory = NULL;

    /* Open and read the dropped frame pixel mask */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, dropped_frames_binary_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building dropped frame pixel "
                                 "maskfile name");
        cleanup_io_routines(blob);
        free_the_blob();
        return;
    }

    inf = fopen(file_name, "rb");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        cleanup_io_routines(blob);
        free_the_blob();
        return;
    }

    dropped_frame = malloc(NVRPDETS * NLINES * sizeof(short));
    if (!dropped_frame)
    {
        assert_true_with_message(0, "Allocating memory for dark_buffer" );
        cleanup_io_routines(blob);
        free_the_blob();
        return;
    }

    fread(dropped_frame, NVRPDETS * NLINES * sizeof(short), 1, inf);
    fclose(inf);

    for (det = 0; det < NVRPDETS; det++)
    {
        for (line = 0; line < NLINES; line++)
        {
            /* Add the dropped frames to pixel mask if a frame_drooped pixel 
               is found */
            if (dropped_frame[det * NLINES + line] != 1)
            assert_false(ias_pm_add_pixels(blob->pixel_mask[0][0], det, line, 
                         1, IAS_RPS_DROPLINE));
        }
    }

    /* This initializes the memory management and registers buffers through
       the algorithms pre-process routine */
    status = initialize_memory_management_buffers();
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error initializing memory management");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }
    status = oli_bias_determination_preprocess(blob);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error returned from: "
            "tirs_response_linearization_preprocess");
        free(dropped_frame);
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Register the MM_NON_PAN_PERFRAME_BIAS MM buffer to this routine so that 
       the API can't destroy the memory before we can retrieve the output */
    status = increment_consumer_count(MM_NON_PAN_PERFRAME_BIAS);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error incrementing "
            "MM_NON_PAN_PERFRAME_BIAS buffer");
        free(dropped_frame);
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Get the index of the test band which is band 1 and is 3rd in the list */
    test_band_index = blob->list_of_band_indices[2];
    test_vrp_index = blob->band_information[test_band_index]->vrp_band_index;

    /* Check an "empty" buffer for the A0 parameter */
    a0_parameter = (double *)preprocess_get_memory_buffer(
        MM_NON_PAN_BIAS_A0_PARAMETER, test_band_index, 0);
    if (!a0_parameter)
    {
        assert_true_with_message(0, "Error getting "
            "MM_NON_PAN_BIAS_A0_PARAMETER buffer");
        free(dropped_frame);
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }
    a0_parameter[0] = a0_b01_s01;
    set_buffer_filled_flag( MM_NON_PAN_BIAS_A0_PARAMETER, test_band_index, 0);

    /* Check an "empty" buffer for the VRP buffer */
    vrp_buffer = (uint16_t *)preprocess_get_memory_buffer(
                 IMAGE_L0R, test_vrp_index, 0);
    if (!vrp_buffer)
    {
        assert_true_with_message(0, "Error getting "
            "IMAGE_L0R VRP buffer");
        free(dropped_frame);
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Open and read the VRP data */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, vrp_binary_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building VRP data file name" );
        free(dropped_frame);
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    inf = fopen(file_name, "rb");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        free(dropped_frame);
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    fread(vrp_buffer, NVRPDETS * NLINES * sizeof(short), 1, inf);
    fclose(inf);
    set_buffer_filled_flag( IMAGE_L0R, test_vrp_index, 0);

    /* Check an "empty" buffer for the a1_parameters */
    a1_parameter = (double *)preprocess_get_memory_buffer(
        MM_NON_PAN_BIAS_A1_PARAMETER, test_band_index, 0);
    if (!a1_parameter)
    {
        assert_true_with_message(0, "Error getting "
            "MM_NON_PAN_BIAS_A1_PARAMETER buffer");
        free(dropped_frame);
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Open and read the a1_parameter shutter buffer */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, a1_parameter_binary_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building a1_parameter file "
                                 "name");
        free(dropped_frame);
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Polulate a1_parameter */
    inf = fopen(file_name, "rb");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        free(dropped_frame);
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    fread(a1_parameter, NDETS * sizeof(double), 1, inf);
    fclose(inf);
    set_buffer_filled_flag( MM_NON_PAN_BIAS_A1_PARAMETER, test_band_index, 0);

    /* Check an "empty" buffer for the c1_parameters */
    c1_parameter = (double *)preprocess_get_memory_buffer(
        MM_NON_PAN_BIAS_A1_PARAMETER, test_band_index, 0);
    if (!c1_parameter)
    {
        assert_true_with_message(0, "Error getting "
            "MM_NON_PAN_BIAS_C1_PARAMETER buffer");
        free(dropped_frame);
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Open and read the a1_parameter shutter buffer */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, c1_parameter_binary_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building a1_parameter file "
                                 "name");
        free(dropped_frame);
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Polulate a1_parameter */
    inf = fopen(file_name, "rb");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        free(dropped_frame);
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    fread(c1_parameter, NDETS * sizeof(double), 1, inf);
    fclose(inf);
    set_buffer_filled_flag( MM_NON_PAN_BIAS_A1_PARAMETER, test_band_index, 0);

    /* Check an "empty" buffer for the bias buffer */
    bias_buffer = (double *)preprocess_get_memory_buffer(
        MM_NON_PAN_PERFRAME_BIAS, test_band_index, 0);
    if (!bias_buffer)
    {
        assert_true_with_message(0, "Error getting "
            "MM_NON_PAN_PERFRAME_BIAS buffer");
        free(dropped_frame);
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* It is important that the consumer counts get set *after* all the
       increments */
    status = set_consumer_counts();
    if ( status != SUCCESS )
    {
        assert_true_with_message(0, "Error setting consumer counts");
        free(dropped_frame);
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Call the API with another band.  The API should return immediately.
       The only way to really check this is to ensure the data buffer wasn't
       changed at all. */

    /* Initialize a few bias_buffer for checking purpose here */
    bias_buffer[0] = 0.0;
    bias_buffer[494] = 494.0;
    bias_buffer[494*2] = 988.0;

    band_index = blob->list_of_band_indices[0];
    status = api_oli_bias_determination( blob, band_index, 0, 0 );
    assert_equal_with_message( status,SUCCESS, "Error returned from "
        "api_oli_bias_determination with band = 3");
    if ( status != SUCCESS )
    {
        free(dropped_frame);
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Check some outputs with the original inputs, no changes */
    assert_double_equal(bias_buffer[0], 0.0);
    assert_double_equal(bias_buffer[494], 494.0);
    assert_double_equal(bias_buffer[494*2], 988.0);

    /* Call the API with TIRS band.  The API should return immediately.
       The only way to really check this is to ensure the data buffer wasn't
       changed at all. */
    band_index = blob->list_of_band_indices[1];
    status = api_oli_bias_determination( blob, band_index, 0, 0 );
    assert_equal_with_message( status,SUCCESS, "Error returned from "
        "api_oli_bias_determination with band = 11");
    if ( status != SUCCESS )
    {
        free(dropped_frame);
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Check some outputs with the original inputs, no changes */
    assert_double_equal(bias_buffer[0], 0.0);
    assert_double_equal(bias_buffer[494], 494.0);
    assert_double_equal(bias_buffer[494*2], 988.0);

    /* Call the API with correct band */
    band_index = blob->list_of_band_indices[2];
    status = api_oli_bias_determination(blob, band_index, 0, 0);
    assert_equal_with_message(status, SUCCESS, "Error returned from "
        "api_tirs_bias_removal with band = 1");
    if (status != SUCCESS)
    {
        free(dropped_frame);
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Open and read the bias results generated from cal/val prototype code */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, bias_model_binary_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building result file name");
        free(dropped_frame);
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }
    result_buffer = malloc(NDETS * NLINES * sizeof(double));
    if (!result_buffer)
    {
        assert_true_with_message(0, "Allocating memory for bias " 
                                         "buffer");
        free(dropped_frame);
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    inf = fopen(file_name, "rb");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        free(dropped_frame);
        free(result_buffer);
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    fread(result_buffer, NDETS * NLINES * sizeof(double), 1, inf);
    fclose(inf);

    /* Compare the results.  Loop through the results and count the
       differences.  This comparison isn't done with asserts since there is
       no need for 494*255 assert statements which cost message buffer space.
       One assert for pass fail with a count is good enough. */
    difference_count = 0;
    for (pixel = 0; pixel < NDETS * NLINES; pixel++)
    {
        if (fabsf(bias_buffer[pixel] - result_buffer[pixel]) > 1e-4)
            difference_count++;
    }
    assert_equal_with_message(difference_count, 0, "There were %d "
        "differences out of %d values", difference_count, NDETS * NLINES);

    /* Free the image memory */
    status = decrement_consumer_count(MM_NON_PAN_PERFRAME_BIAS,
        test_band_index, 0);
    assert_equal_with_message( status, SUCCESS,
        "Error returned from decrement consumer count");
    status = decrement_consumer_count(MM_NON_PAN_PERFRAME_BIAS,
        blob->list_of_band_indices[0], 0);
    assert_equal_with_message( status, SUCCESS,
        "Error returned from decrement consumer count");

    free(dropped_frame);
    free(result_buffer);
    cleanup_io_routines(blob);
    cleanup_memory_management_buffers(ENABLE_WARNING_DISPLAY);
    free_the_blob();
}

