/***************************************************************************
 NAME:      oli_bias_removal_test

 PURPOSE:   CGreen unit testing for the OLI bias removal algorithm

 RETURNS:   CGreen test status

 ROUTINES:  local_free_the_blob                      (internal routine)
            test_oli_bias_removal_direct_algorithm
            test_oli_bias_removal_from_api
****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "ias_const.h"
#include "ias_logging.h"
#include "cgreen.h"
#include "../rps_local.h"


/* Path to test data and file names */
static char cpf_dir[] = "ias_lib/cpf";
static char cpf_file[] = "L8CPF20090101_20090331.01";


/**************************************************************************
 NAME:     local_free_the_blob

 PURPOSE:  Local override of the routine to free the blob.  Since not all
           members of the blob are initialized the normal routine cannot
           be called

 RETURNS:  Nothing
***************************************************************************/
static void local_free_the_blob(BLOB *blob)
{
    int band_index;

    if ( blob->l1r_band_metadata )
    {
        for ( band_index = 0; band_index < blob->total_bands; band_index++ )
            free( blob->l1r_band_metadata[band_index] );
        free( blob->l1r_band_metadata );
    }

    for (band_index = 0; band_index < 30; band_index++)
        free(blob->band_information[band_index]);
    free(blob->band_information);

    free(blob);
}   /* END local_free_the_blob */


Ensure test_oli_bias_removal_direct_algorithm()
{
    /* 6 x 6 prototype "test" PAN bias data expressed as a 1D array */
    double pan_test_bias[36] = {1.6, 0.1, 6.4, 2.5, 8.1, 1.6,
        1.6, 0.1, 6.4, 2.5, 8.1, 1.6, 1.6, 0.1, 6.4, 2.5, 8.1, 1.6,
        1.6, 0.1, 6.4, 2.5, 8.1, 1.6, 1.6, 0.1, 6.4, 2.5, 8.1, 1.6,
        1.6, 0.1, 6.4, 2.5, 8.1, 1.6};

    /* PAN band prototype "test" temperature correction factors */
    double pan_test_temp_cfs[6] = {0.4, 0.1, 0.8, 0.5, 0.9, 0.4};

    /* 3 x 3 prototype "test" bias data expressed as a 1D array */
    double test_bias[9] = {4.0, 1.0, 8.0, 5.0, 9.0, 4.0, 7.0, 1.0, 7.0};

    /* Prototype "test" temperature correction factor coefficients */
    double test_temp_cfs[3] = {0.5, 0.9, 0.4};

    /* 6 x 6 prototype "test" PAN band image expressed as a 1D array.
       oli_bias_removal will modify the original values to the bias-
       corrected values */
    float pan_test_image[36] = {4.0, 1.0, 8.0, 5.0, 9.0, 4.0,
        7.0, 1.0, 7.0, 7.0, 4.0, 6.0, 9.0, 5.0, 7.0, 2.0, 3.0, 8.0,
        8.0, 9.0, 1.0, 3.0, 9.0, 3.0, 5.0, 4.0, 2.0, 5.0, 8.0, 8.0,
        0.0, 10.0, 7.0, 8.0, 1.0, 4.0};
    float pan_corrected_image1[36];
    float pan_corrected_image2[36];

    /* This is what the bias-corrected test image (no temp correction)
       should look like */
    float expected_pan_corrected_image1[36] = {2.4, 0.9, 1.6, 2.5, 0.9, 2.4,
        5.4, 0.9, 0.6, 4.5, -4.1, 4.4, 7.4, 4.9, 0.6, -0.5, -5.1, 6.4,
        6.4, 8.9, -5.4, 0.5, 0.9, 1.4, 3.4, 3.9, -4.4, 2.5, -0.1, 6.4,
        -1.6, 9.9, 0.6, 5.5, -7.1, 2.4};

    /* This is what the bias-corrected test image (with temp correction)
       should look like */
    float expected_pan_corrected_image2[36] = {3.36, 0.99, 2.88, 3.75, 1.71,
        3.36, 6.36, 0.99, 1.88, 5.75, -3.29, 5.36, 8.36, 4.99, 1.88, 0.75,
        -4.29, 7.36, 7.36, 8.99, -4.12, 1.75, 1.71, 2.36, 4.36, 3.99, -3.12,
        3.75, 0.71, 7.36, -0.64, 9.99, 1.88, 6.75, -6.29, 3.36};

    /* 3 x 3 prototype "test" image expressed as a 1D array.
       oli_bias_removal will modify the original values to the
       bias-corrected values */
    float test_image[9] = {34.0, 19.0, 32.0, 44.0, 26.0, 33.0, 12.0,
        13.0, 38.0};
    float corrected_image1[9];
    float corrected_image2[9];

    /* This is what the bias-corrected test image (no temp correction)
       should be */
    float expected_test_corrected_image1[9] = {30.0, 18.0, 24.0, 39.0, 17.0,
        29.0, 5.0, 12.0, 31.0};

    /* This is what the bias-corrected test image (temp correction)
       should be */
    float expected_test_corrected_image2[9] = {32.0, 18.1, 28.8, 41.5,
        17.9, 31.4, 8.5, 12.1, 35.2};

    float diff;                   /* Difference between expected and
                                     actual values */
    int number_pan_detectors = 6; /* Number of detectors in PAN "test"
                                     image */
    int number_pan_lines = 6;     /* Number of lines in PAN "test" image */
    int number_detectors = 3;     /* Number of detectors in "test" image */
    int number_lines = 3;         /* Number of lines in "test" image */
    int ii;                       /* Local loop counter */


    printf("Testing OLI bias removal algorithm directly...\n");

    /* Do bias removal for the uncorrected bias case */
    memcpy(corrected_image1, test_image, 9*sizeof(float));
    oli_bias_removal(corrected_image1, NULL, NULL,
       test_bias, MEAN_ACQUISITION_SHUTTER, 0, 0, 
       number_detectors, number_lines, NULL);

    /* Now check the results */
    for (ii = 0; ii < number_detectors * number_lines; ii++)
    {
        assert_double_equal(expected_test_corrected_image1[ii],
           corrected_image1[ii]);
    }

    /* Do bias removal for the corrected bias case */
    memcpy(corrected_image2, test_image, 9*sizeof(float));
        oli_bias_removal(corrected_image2, NULL, NULL,
        test_bias, MEAN_ACQUISITION_SHUTTER, 0, 0,
        number_detectors, number_lines, test_temp_cfs);

    /* Now check the results.  I've switched to testing the true
       difference to 5 decimal places rather than strict equality to 8
       decimal places using assert_double_equal--with this test data,
       the corrected pixel test_image[4] (17.900000) differed from the
       expected pixel value (17.900002) by more than 1e-8,
       causing assert_double_equal() to fail.  A "doubles vs floats"
       issue, I think... */
    for (ii = 0; ii < number_detectors * number_lines; ii++)
    {
        diff = 
            fabs(expected_test_corrected_image2[ii] - corrected_image2[ii]);
        assert_true_with_message((diff < 1e-5), "Pixel %d:  expected=%f, "
            "returned=%f\n", (ii + 1), expected_test_corrected_image2[ii],
            corrected_image2[ii]);
    }

    /* Do bias removal on the pan band (uncorrected) bias case */
    memcpy(pan_corrected_image1, pan_test_image, 36*sizeof(float));
    oli_bias_removal(pan_corrected_image1, NULL, NULL, pan_test_bias,
        MEAN_ACQUISITION_SHUTTER, 0, 0, number_pan_detectors, number_pan_lines,
        NULL);

    /* Check the results */
    for (ii = 0; ii < number_pan_detectors * number_pan_lines; ii++)
    {
        diff =
            fabs(expected_pan_corrected_image1[ii] - pan_corrected_image1[ii]);
        assert_true_with_message((diff < 1e-5), "Pixel %d:  expected=%f, "
            "returned=%f\n", (ii + 1), expected_pan_corrected_image1[ii],
            pan_corrected_image1[ii]);
    }

    /* Do bias removal on the pan band (temp corrected) bias case */
    memcpy(pan_corrected_image2, pan_test_image, 36*sizeof(float));
    oli_bias_removal(pan_corrected_image2, NULL, NULL,
        pan_test_bias, MEAN_ACQUISITION_SHUTTER, 0, 0,
        number_pan_detectors, number_pan_lines,
        pan_test_temp_cfs);

    /* Check the results */
    for (ii = 0; ii < number_pan_detectors * number_pan_lines; ii++)
    {
        diff =
            fabs(expected_pan_corrected_image2[ii] - pan_corrected_image2[ii]);
        assert_true_with_message((diff < 1e-5), "Pixel %d:  expected=%f, "
            "returned=%f\n", (ii + 1), expected_pan_corrected_image2[ii],
            pan_corrected_image2[ii]);
    }

    /* Test the new IS_ODD_LINE() macro*/
    int even_flag = 1;
    int odd_flag = 0;
    int test_flag;
    int line_index;

    for (line_index = 0; line_index < 6; line_index++)
    {
        test_flag = IAS_IS_EVEN_LINE(line_index);
        if ((line_index % 2) == 0)
        {
            assert_equal_with_message(test_flag, odd_flag, "Line should be"
                "flagged as odd");
        }
        else
        {
            assert_equal_with_message(test_flag, even_flag, "Line should be "
                "flagged as even");
        }
    }
}   /* End test_oli_bias_removal_direct_algorithm */




Ensure test_oli_bias_removal_from_api()
{
    char file_name[PATH_MAX];          /* File name to open */
    char *test_directory = NULL;       /* test environment value */
    const IAS_BAND_ATTRIBUTES *band_atts = NULL;
    BLOB *blob = NULL;                 /* "Test"BLOB data structure */
    double *bias_buffer = NULL;              /* Test bias data buffer */
    double bias_data[9] = {4.0, 1.0, 8.0, 5.0, 9.0, 4.0, 7.0, 1.0, 7.0};
    float *image_buffer = NULL;              /* Test image data buffer */
    float image_data[9] = {34.0, 19.0, 32.0, 44.0, 26.0, 33.0, 12.0,
        13.0, 38.0};
    float bias_removed_image[9] = {30.0, 18.0, 24.0, 39.0, 17.0,
        29.0, 5.0, 12.0, 31.0};
    int status;
    int band_index;
    int band_number;
    int pixel;
    int num_lines = 3;
    int num_dets = 3;
    int num_pixels = num_dets * num_lines;
    int list_of_bands[3] = {27, 10, 7};
    int list_of_band_indices[3];
    int num_bands = 3;
    int test_band_index;


    printf("Testing OLI bias removal algorithm through API interface...\n");

    /* Get the environment variable for the test data */
    test_directory = getenv("IAS_UNIT_TEST_DATA_DIR");
    if ( !test_directory )
    {
        assert_true_with_message( 0, "Getting IAS_UNIT_TEST_DATA_DIR");
        return;
    }

    /* Initialize the attributes.  This is required to get the memory
       management system to work */
    status = ias_sat_attr_initialize(IAS_L8);
    if (status == ERROR)
    {
        assert_true_with_message(0, "Cannot initialize satellite "
            "attributes");
        return;
    }

    /* Allocate BLOB data block */
    blob = allocate_blob();
    if (blob == NULL)
    {
        assert_true_with_message(0, 
            "Error allocating BLOB base data block\n");
        return;
    }

    /* Initialize the BLOB */
    blob->max_number_scas = 14;
    blob->total_bands = 30;
    blob->l1r_band_metadata = calloc( blob->total_bands,
        sizeof( IAS_L1R_BAND_METADATA * ));
    if ( ! blob->l1r_band_metadata )
    {
        assert_true_with_message(0, "Error allocating band metadata");
        local_free_the_blob( blob );
        return;
    }
    for ( band_index = 0; band_index < blob->total_bands; band_index++ )
    {
        blob->l1r_band_metadata[band_index] = calloc( 1,
            sizeof(IAS_L1R_BAND_METADATA));
        if ( ! blob->l1r_band_metadata[band_index] )
        {
            assert_true_with_message(0, "Error allocating band metadata");
            local_free_the_blob( blob );
            return;
        }
    }

    blob->band_information = calloc(30, sizeof(RPS_BAND_ATTRIBUTES *));
    if (blob->band_information == NULL)
    {
        assert_true_with_message(0, "Error allocating BLOB band "
            "attributes data block");
        local_free_the_blob(blob);
        return;
    }
    for (band_number = 1; band_number <= 30; band_number++)
    {
        int i;

        band_atts = ias_sat_attr_get_band_attributes(band_number);
        band_index = band_atts->band_index;

        blob->band_information[band_index] =
            calloc(1, sizeof(RPS_BAND_ATTRIBUTES));
        if (blob->band_information[band_index] == NULL)
        {
            assert_true_with_message(0, "Error allocating band %d "
                "attribute data block", band_index);
            local_free_the_blob(blob);
            return;
        }

        blob->band_information[band_index]->band_number = band_number;
        blob->band_information[band_index]->number_lines = 3;
        blob->band_information[band_index]->vrp_band_number = 
            band_atts->vrp_band_number;
        blob->band_information[band_index]->number_detectors = 3;
        blob->band_information[band_index]->number_scas = band_atts->scas;
        blob->band_information[band_index]->band_classification =
            band_atts->band_classification;
        blob->band_information[band_index]->spectral_type =
            band_atts->spectral_type;
        for ( i = 0; i < num_bands; i++ )
        {
            if ( band_number == list_of_bands[i] )
            {
                list_of_band_indices[i] = band_index;
                break;
            }
        }
    }

    /* Set the test band index, it's the 3rd one in the list */
    test_band_index = list_of_band_indices[2];

    /* The blob->l1r_file_metadata.bias_source generation is moved to
       oli_bias_removal_preprocess, so blob->output_bias_type needs 
       to be set */
    blob->per_frame_bias_flag =PER_DETECTOR_BIAS; 

    /* Open the CPF */
    sprintf(file_name, "%s/%s/%s", test_directory, cpf_dir, cpf_file);
    blob->cpf = ias_cpf_read(file_name);
    if (!blob->cpf)
    {
        assert_true_with_message(0, "Error opening CPF: %s", file_name);
        local_free_the_blob(blob);
        return;
    }

     /* Turn off temperature correction */
     blob->perform_temp_sens = 0;

    /* Disable the logging so that the memory management warnings are
       supressed */
    ias_log_set_output_level( IAS_LOG_LEVEL_DISABLE );

    /* Initialize the memory management system.  This algorithm doesn't need
       any memory management buffers that require the list of bands to be
       initialized -- number of bands set to 0 for that reason. */
    blob->number_of_bands = 0;
    blob->band_fields_initialized = 1;
    status = initialize_memory_management_buffers();
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error initializing memory management");
        ias_cpf_free(blob->cpf);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    /* Initialize the RPS processing level arrays */
    status = initialize_rps_level(30, 1);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error calling initialize_rps_level");
        ias_cpf_free(blob->cpf);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    status = oli_bias_removal_preprocess(blob);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error in oli_bias_removal preprocess");
        cleanup_rps_level();
        ias_cpf_free(blob->cpf);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    /* Register the image data buffer so it can't be destroyed before
       the output can be checked */
    status = increment_consumer_count(IMAGE_L1R);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error in oli_bias_removal preprocess");
        cleanup_rps_level();
        ias_cpf_free(blob->cpf);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    status = increment_consumer_count(MM_NON_PAN_PERFRAME_BIAS);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error in oli_bias_removal preprocess");
        cleanup_rps_level();
        ias_cpf_free(blob->cpf);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    /* Get an "empty" image buffer from one of the NON_PAN bands (say 7) */
    image_buffer = (float *)preprocess_get_memory_buffer(IMAGE_L1R,
        test_band_index, 0);
    if (!image_buffer)
    {
        assert_true_with_message(0, "Error getting IMAGE_L1R buffer");
        cleanup_rps_level();
        ias_cpf_free(blob->cpf);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    /* Get an "empty" bias data buffer for band 7 */
    bias_buffer = (double *)preprocess_get_memory_buffer(
        MM_NON_PAN_PERFRAME_BIAS, test_band_index, 0);
    if (!bias_buffer)
    {
        assert_true_with_message(0, "Error getting MM_NON_PAN_PERFRAME_BIAS "
            "buffer");
        cleanup_rps_level();
        ias_cpf_free(blob->cpf);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    /* Populate the data buffers */
    for (pixel = 0; pixel < num_pixels; pixel++)
    {
        image_buffer[pixel] = image_data[pixel];
        bias_buffer[pixel] = bias_data[pixel];
    }
    set_buffer_filled_flag( IMAGE_L1R, test_band_index, 0);
    set_buffer_filled_flag( MM_NON_PAN_PERFRAME_BIAS, test_band_index, 0 );

    /* It is important that the consumer counts get set *after* all the
       increments */
    status = set_consumer_counts();
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error setting consumer counts");
        cleanup_rps_level();
        ias_cpf_free(blob->cpf);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    /* Call the API in three cases.  First two cases should return
       SUCCESS with nothing done */

    /* First case:  band 27 */
    band_index = list_of_band_indices[0];
    status = api_oli_bias_removal(blob, band_index, 0, 0);
    assert_equal_with_message(status, SUCCESS, "Should not be processing "
        "band 27");
    if (status != SUCCESS)
    {
        cleanup_rps_level();
        ias_cpf_free(blob->cpf);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    /* Check the data buffers for unexpected changes.  The bias data buffer
       should not change at all, for any reason; if it does, something's
       seriously wrong somewhere.  The image data buffer shouldn't change,
       either, if we have to exit either because we're processing a TIRS
       band, or any kind of "non-normal" band...*/
    for (pixel = 0; pixel < num_pixels; pixel++)
    {
        assert_double_equal(image_buffer[pixel], image_data[pixel]);
        assert_double_equal(bias_buffer[pixel], bias_data[pixel]);
    }

    /* Second case: band 10 (TIRS band) */
    band_index = list_of_band_indices[1];
    status = api_oli_bias_removal(blob, band_index, 0, 0);
    assert_equal_with_message(status, SUCCESS, "Should have returned SUCCESS "
        "for TIRS bands");
    if (status != SUCCESS)
    {
        cleanup_rps_level();
        ias_cpf_free(blob->cpf);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    /* Check the data buffers for unexpected changes */
    for (pixel = 0; pixel < num_pixels; pixel++)
    {
        assert_double_equal(image_buffer[pixel], image_data[pixel]);
        assert_double_equal(bias_buffer[pixel], bias_data[pixel]);
    }

    /* Third case:  band 7 (our OLI band) */
    band_index = list_of_band_indices[2];
    status = api_oli_bias_removal(blob, band_index, 0, 0);
    assert_equal_with_message(status, SUCCESS, "Should have successfully "
        "processed this band");
    if (status != SUCCESS)
    {
        cleanup_rps_level();
        ias_cpf_free(blob->cpf);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    /* Checks go here */
    float diff;
    for (pixel = 0; pixel < num_pixels; pixel++)
    {
        diff = fabs(image_buffer[pixel] - bias_removed_image[pixel]);
        if (diff >= 1e-5)
        {
            assert_true_with_message(0, "Difference between corrected and "
                "expected test data at least 1e-5 or greater");
        }
    }

    /* Free all allocated memory */
    status = decrement_consumer_count(IMAGE_L1R, test_band_index, 0);
    assert_equal_with_message( status, SUCCESS,
        "Error returned from decrement consumer count");
    status = decrement_consumer_count(MM_NON_PAN_PERFRAME_BIAS,
        test_band_index, 0);
    assert_equal_with_message( status, SUCCESS,
        "Error returned from decrement consumer count");
    cleanup_rps_level();
    cleanup_memory_management_buffers(ENABLE_WARNING_DISPLAY);
    ias_cpf_free(blob->cpf);
    local_free_the_blob(blob);
}



int main(int argc, char *argv[])
{
    /* Create the main test suite */
    TestSuite *suite = create_test_suite();

    add_test(suite, test_oli_bias_removal_direct_algorithm);
    add_test(suite, test_oli_bias_removal_from_api);

    /* Run the tests */
    if (argc > 1)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}
