#include <stdio.h>
#include "string.h"
#include "ias_satellite_attributes.h"
#include "cgreen/cgreen.h"
#include "ias_logging.h"
#include "ias_radiometry.h"
#include "../rps_local.h"

#define NLINES 4
#define NDETS  3

static char l0r_data_dir[] = 
       "ias_base/rps/driver/LC81800382002045SGS00";
static char l0r_basename[] = "LC81800382002045SGS00";
static char cpf_data_dir[] =
       "ias_base/rps/driver/temperature_sensitivity_correction";
static char cpf_basename[] = "L8CPF20020701_20020930.01";
static char bpf_data_dir[] = 
       "ias_base/rps/driver/temperature_sensitivity_correction";
static char bpf_basename_oli[] = "LO8BPF20020802230930_20020802234142.01";
static char rlut_data_dir[] = 
       "ias_base/rps/driver/temperature_sensitivity_correction";
static char rlut_basename[] = "L8RLUT19990101_20111231v01.h5";

Ensure reflectance_conversion_test( void );
Ensure earth_sun_distance_test( void );
Ensure test_algorithm_thru_api(void);

int main( int argc, char *argv[] )
{
    TestSuite *suite = create_test_suite();
    add_test( suite, reflectance_conversion_test );
    add_test( suite, earth_sun_distance_test );

    add_test(suite, test_algorithm_thru_api);

    if (argc > 1) {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite( suite, create_text_reporter() );
}


Ensure reflectance_conversion_test( void )
{
    float *buffer;                     /* pointer to the l1r data */
    float l1r_data[3][4];              /* input L1r image data */
    unsigned int det, frame;           /* loop variables for det and frame */

    int num_detectors = 3;             /* number of detectors */
    int num_frames = 4;                /* number of frames */
    double earth_sun_distance;         /* earth sun distance in astronomy unit
                                          from NOVAS library calculation */
    double rad_to_reflect_conv_coeff;  /* band level radiance to reflectance
                                          conversion coefficient */
    double reflect_to_rad_conv_coeff;  /* band level reflectance to radiance
                                          conversion coefficient */   
    double conversion_factor;   /* intermediate reflectance to radiance 
                                    conversion coefficient factor */
    int status;

    float conv_rad[] = {0.000000000, 0.005165853, 0.010331707, 0.015497560,
                        0.001721951, 0.006887804, 0.012053658, 0.017219518,
                        0.003443902, 0.008609755, 0.013775609, 0.018941462};

    /* Initialize the parameters */
    earth_sun_distance = 1.0154351;
    rad_to_reflect_conv_coeff = 0.00167;

    conversion_factor = 1 / 
                    ((earth_sun_distance * earth_sun_distance) 
                     * rad_to_reflect_conv_coeff);

    /* Loop through the detector */
    for (det = 0; det < num_detectors; det++)
    {
        /* Loop through the frame */
        for (frame = 0; frame < num_frames; frame++)
        {
            l1r_data[det][frame] = (float) (det + frame * num_detectors);
        }
    }

    /* Create a pointer link to the input l1r image data */
    buffer = &l1r_data[0][0];

    /* Call the gain application algorithm */
    status = reflectance_conversion(buffer, 3, 4, 
                      earth_sun_distance, rad_to_reflect_conv_coeff, 
                      &reflect_to_rad_conv_coeff);
    if ( status )
    {
        assert_true_with_message( 0, "Error from "
            "reflectance_conversion" );
    }

    /* Check the result  for band level reflectance to radiance
       conversion coefficient */
    assert_double_equal(reflect_to_rad_conv_coeff, conversion_factor);

    /* Loop through the detector */
    for (det = 0; det < num_detectors; det++)
    {
        /* Loop through the frame */
        for (frame = 0; frame < num_frames; frame++)
        {
            /* Check the algorithm results */
         assert_double_equal(l1r_data[det][frame],
                             (float)conv_rad[frame + det * num_frames]);
        } /* End frame loop */
    }  /* End detector loop */
}  /* END reflectance_conversion_test */

Ensure earth_sun_distance_test( void )
{
    double earth_sun_distance;  /* Returned earth_sun_distance from API call */
    double dummy_distance;      /* Returned earth_sun_distance from preprocess 
                                   call */
    double expected_earth_sun_distance = 1.01576752;
    int status;

    /* Set the date_acquired */
    char date_acquired[] = "2010:205:16:15:27.04520";


    /* Call get_earth_sun_distance, this is the simulation of call to
       initialize the JPL NOVAS library in case multiple threads are used */
    status = get_earth_sun_distance(date_acquired, &dummy_distance);    
    if ( status )
    {
        assert_true_with_message( 0, "Error in simulation of call to "
            "get_earth_sun_distance in preprocess" );
    }

    /* Call get_earth_sun_distance, this is the simulation of call in API 
       in case multiple threads are used */
    status = get_earth_sun_distance(NULL, &earth_sun_distance);    
    if ( status )
    {
        assert_true_with_message( 0, "Error in simulation of call to"
            "get_earth_sun_distance in API" );
    }

    assert_double_equal(earth_sun_distance, expected_earth_sun_distance);

}

/*--------------------------------------------------------------------------
  test_algorithm_thru_api

  This test sets up for a call to the algorithm through the API.  Also calls
  the pre-process routine to set up memory management. The input ASCII files 
  are extracted from calval binary test files with only one band and one sca 
  used here. The compared result file is generated from cal/val IDL code.
--------------------------------------------------------------------------*/
Ensure test_algorithm_thru_api(void)
{
    const IAS_BAND_ATTRIBUTES *band_atts;
    char *test_directory;          /* test environment value */
    BLOB *blob;                    /* Declare a blob structure */
    float *l1r_input_buffer = NULL;/* Image data for the current band/SCA */
    int i;                         /* Loop variable */
    int band_number = 1;           /* 1-based band number */
    int status;                    /* Return variable */
    int band_index;                /* 0-based band number */
    int list_of_bands[2] = {11, 1};         /* List of two test bands */
    int list_of_band_indices[2] = {10, 0};  /* List of two test band indices */
    int num_bands = 2;                      /* Number of bands for testing */
    double *rad_to_reflect_conv_coeff = NULL;  
    double expected_rad_to_reflect_conv_coeff = 613.922724;
    float input_data[] = {0.0, 3.0, 6.0, 9.0, 1.0, 4.0, 7.0, 10.0, 2.0,
                          5.0, 8.0, 11.0};
    float conv_rad[] = {0.000000000, 0.004886609, 0.009773217, 0.014659826,
                        0.001628869, 0.006515478, 0.011402086, 0.016288696,
                        0.003257739, 0.008144348, 0.013030956, 0.017917564};

    /* Get the environment variable for the test data */
    test_directory = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (!test_directory)
    {
        assert_true_with_message(0, "Getting IAS_UNIT_TEST_DATA_DIR");
        return;
    }

    /* The attributes must be initialized for memory management to work */
    ias_sat_attr_initialize(IAS_L8);

    /* Allocate a blob and populate the minimum members for the test
       to succeed */
    blob = allocate_blob();
    blob->max_number_scas = 14;
    blob->total_bands = 30;

    sprintf(blob->l0r_file_name, "%s/%s/%s", test_directory, l0r_data_dir,
        l0r_basename);
    sprintf(blob->l1r_file_name, "%s", "temp_l1r");
    sprintf(blob->cpf_filename, "%s/%s/%s", test_directory, cpf_data_dir,
        cpf_basename);
    sprintf(blob->bpf_filename_oli, "%s/%s/%s", test_directory, bpf_data_dir,
        bpf_basename_oli);
    strcpy(blob->bpf_filename_tirs, "");
    sprintf(blob->rlut_filename, "%s/%s/%s", test_directory, rlut_data_dir,
        rlut_basename);

    status = finish_blob_initialization();
    assert_equal_with_message(status, SUCCESS, "Error completing test blob "
        "initialization");
    if (status != SUCCESS)
    {
        free_the_blob();
        return;
    }

    /* Set some blob information */
    blob->list_of_bands[0] = list_of_bands[0];
    blob->list_of_bands[1] = list_of_bands[1];
    blob->number_of_bands = 2;
    for (band_number = 1; band_number <= blob->total_bands; band_number++)
    {
        band_atts = ias_sat_attr_get_band_attributes(band_number);
        if (band_atts == NULL)
        {
            assert_true_with_message(FALSE, "Error obtaining RPS attribute "
                "information for band %d", band_number);
            free_the_blob();
            return;
        }
        band_index = band_atts->band_index;

        blob->band_information[band_index]->band_number = band_number;
        blob->band_information[band_index]->number_lines = NLINES;
        blob->band_information[band_index]->number_detectors = NDETS;
        blob->band_information[band_index]->number_scas = band_atts->scas;
        blob->band_information[band_index]->band_classification =
            band_atts->band_classification;
        blob->band_information[band_index]->spectral_type =
            band_atts->spectral_type;
        for ( i = 0; i < num_bands; i++ )
        {
            if ( band_number == list_of_bands[i] )
            {
                list_of_band_indices[i] = band_index;
                break;
            }
        }

        /* Set the reflectance_to_radiance_coefficient be zero 
           in L1R metadata in order to be compared with TIRS band call */
        blob->l1r_band_metadata[band_index][0].
            reflectance_to_radiance_coefficient = 0.0;
    }

    /* Diable logging to suppress the memory management warnings */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* Initialize the L0R image data */
    status = initialize_io_routines(blob);
    ias_log_set_output_level( IAS_LOG_LEVEL_INFO );
    assert_equal_with_message(status, SUCCESS, "Error initializing image "
        "data");
    if (status != SUCCESS)
    {
        cleanup_io_routines(blob);
        free_the_blob();
        return;
    }

    /* This initializes the memory management and registers buffers through
       the algorithms pre-process routine */
    status = initialize_memory_management_buffers();
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error initializing memory management");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Initialize the RPS processing level arrays */
    status = initialize_rps_level(30, 14);
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error calling initialize_rps_level" );
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    status = reflectance_conversion_preprocess(blob);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error returned from: "
            "reflectance_conversion_preprocess");
        cleanup_rps_level();
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Register the L1R buffer to this routine so that the API can't
       destroy the memory before we can retrieve the output */
    status = increment_consumer_count(IMAGE_L1R);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error returned from: "
            "increment_consumer_count IMAGE_L1R");
        cleanup_rps_level();
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Get the image buffer for band 1, SCA 1 this simply returns an unfilled,
       allocated buffer */
    l1r_input_buffer = (float *)preprocess_get_memory_buffer(IMAGE_L1R, 
        list_of_band_indices[1], 0);
    if (!l1r_input_buffer)
    {
        assert_true_with_message(0, "Getting IMAGE_L1R buffer");
        cleanup_rps_level();
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Set the l1r_input_buffer values */
    for (i = 0; i < NDETS * NLINES; i++)
        l1r_input_buffer[i] = input_data[i];
    set_buffer_filled_flag( IMAGE_L1R, list_of_band_indices[1], 0);

    /* It is important that the consumer counts get set *after* all the
       increments */
    status = set_consumer_counts();
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error setting consumer counts");
        cleanup_rps_level();
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Check an "empty" buffer for the reflectance conversion coefficient */
    rad_to_reflect_conv_coeff = (double *)get_const_memory_buffer(
        MM_REFLECTANCE_CONVERSION_COEFFICIENTS, list_of_band_indices[1], 0);
    if (!rad_to_reflect_conv_coeff)
    {
        assert_true_with_message(0, "Error getting "
            "MM_REFLECTANCE_CONVERSION_COEFFICIENTS buffer");
        cleanup_rps_level();
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Set the reflectance conversion coefficient value */
    *rad_to_reflect_conv_coeff = 0.00167;

    /* Call the API with band 11 and SCA 1 and the Reflectance Conversion
       algorithm shouldn't be executed for TIRS bands */
    band_number = list_of_bands[0];
    band_index = list_of_band_indices[0];
    status = api_reflectance_conversion( blob, band_index, 0, 0 );
    assert_equal_with_message( status,SUCCESS, "Error returned from "
          "api_reflectance_conversion for band_index %d", band_index);
    if ( status != SUCCESS )
    {
        cleanup_rps_level();
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob( );
        return;
    }

    /* The reflectance_to_radiance_coefficient should be zero since no 
       calculation is done for TIRS bands  */
    assert_double_equal(blob->l1r_band_metadata[band_index][0].
        reflectance_to_radiance_coefficient, 0.0);

    /* Check outputs, values shouldn't have any changes */
    for (i = 0; i < NDETS * NLINES; i++)
        assert_double_equal(l1r_input_buffer[i], input_data[i]);

    /* Call the API with band 1 and SCA 1 */
    band_number = list_of_bands[1];
    band_index = list_of_band_indices[1];
    status = api_reflectance_conversion( blob, band_index, 0, 0 );
    assert_equal_with_message( status,SUCCESS, "Error returned from "
           "api_reflectance_conversion for band_index %d", band_index);
    if ( status != SUCCESS )
    {
        cleanup_rps_level();
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob( );
        return;
    }

    /* The reflectance_to_radiance_coefficient should be changed to the 
       expected value */
    assert_double_equal(blob->l1r_band_metadata[band_index][0].
        reflectance_to_radiance_coefficient, 
        expected_rad_to_reflect_conv_coeff);

    /* Check outputs, values should have been changed to expected values */
    for (i = 0; i < NDETS * NLINES; i++)
        assert_double_equal(l1r_input_buffer[i], conv_rad[i]);

    /* Free the image memory */
    status = decrement_consumer_count(IMAGE_L1R, 0, 0);
    assert_equal_with_message( status, SUCCESS,
        "Error returned from decrement consumer count");

    cleanup_rps_level();
    cleanup_io_routines(blob);
    cleanup_memory_management_buffers(ENABLE_WARNING_DISPLAY);
    free_the_blob();
} 

