/*----------------------------------------------------------------------------
NAME:    residual_striping_correction_test.c

PURPOSE: Test the residual_striping_correction routine by testing it directly 
         and testing it through API

USAGE:   residual_striping_correction_test [test name]

         When run without a command line argument all tests will be run.
         Pass the name of a test to run the single test -- useful for
         debugging.
----------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "cgreen/cgreen.h"
#include "ias_logging.h"
#include "../rps_local.h"

#define NLINES 6357
#define NDETS  320

/* Path to test data and file names */
static char data_dir[] = "ias_base/rps/driver/residual_striping_correction";
static char scene_input_file[] = "old_scene.txt";
static char scene_output_file[] = "new_scene.txt";
static char striping_matrix_file[] = "striping_matrix.txt";

/* Test prototypes */
Ensure test_algorithm_directly(void);
Ensure test_algorithm_thru_api(void);

int main(int argc, char *argv[])
{
    TestSuite *suite = create_test_suite();
    add_test(suite, test_algorithm_directly);
    add_test(suite, test_algorithm_thru_api);

    if (argc == 2)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}


/*----------------------------------------------------------------------------
NAME: compute_float_mean  

PURPOSE: Mean calculation for floating point numbers

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int compute_float_mean
(
    const float *sample,       /* I: pointer to the data */
    int nsamps,                /* I: number of samples */
    float *mean                /* O: mean of the data */
)
{
    int i;                     /* loop variable */
    float sum = 0.0;           /* sum of the samples */

    if (nsamps <= 0)
    {
        IAS_LOG_ERROR("Sample size is invalid.");
        *mean = 0.0;
        return ERROR;
    }

    for (i = 0; i < nsamps; i++)
        sum += sample[i];

    *mean = sum / nsamps;

    return SUCCESS;
}

/*--------------------------------------------------------------------------
   test_algorithm_directly

   This test executes the algorithm with a subset of the Cal/Val prototype
   input and output. The input and compared result files are generated from 
   cal/val SDSU matlab code.
---------------------------------------------------------------------------*/
Ensure test_algorithm_directly(void)
{
    FILE *inf;                     /* File pointer for all reads */
    char *test_directory;          /* Test environment value */
    char file_name[PATH_MAX];      /* File name to open */
    float *l1r_input_buffer;       /* Image data for the current band/SCA */
    float *l1r_output_buffer;      /* Image data for the current band/SCA */
    double *striping_buffer;       /* Striping matrix for the current 
                                      band/SCA */
    float temp_float;              /* Temporary value read from file */
    double temp_double;            /* Temporary value read from file */
    float input_scene_mean;        /* Mean of the input test scene */
    float output_scene_mean;       /* Mean of the output test scene */
    float expected_scene_mean = 128.26;     
                                   /* Expected mean value from ADD */
    int difference_count;          /* Numbers of differences between the 
                                      calculated value and expected results */
    int return_value;              /* Return values from fscanf read */ 
    int line;                      /* Line loop counter */                     
    int det;                       /* Detector loop counter */
    int pixel;                     /* Pixel loop counter */
    int status;                    /* Return variable */

    /* Get the environment variable for the test data */
    test_directory = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (!test_directory)
    {
        assert_true_with_message(0, "Getting IAS_UNIT_TEST_DATA_DIR");
        return;
    }

    /* Open and read the input 1 SCA data */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, scene_input_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building L1r input data file "
                                 "name");
        return;
    }
    /* Allocate memory for input L1R data buffer */
    l1r_input_buffer = malloc(NDETS * NLINES * sizeof(float));
    if (!l1r_input_buffer)
    {
        assert_true_with_message(0, "Allocating memory for l1r_input_buffer");
        return;
    }

    /* Read the input l1r buffer */
    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        free(l1r_input_buffer);
        return;
    }

    for (line = 0; line < NLINES; line++)
    {
        for (det = 0; det < NDETS; det++)
        {
            return_value = fscanf(inf, "%f", &temp_float);
            if (return_value == EOF)
            {
                assert_true_with_message(0, "Reading %s", file_name);
                free(l1r_input_buffer);
                return;        
            }        
            l1r_input_buffer[line * NDETS + det] = temp_float;
        }
    }
    fclose(inf);

    /* Open and read the striping matrix data */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, striping_matrix_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building striping matrix data "
                                 "file name");
        free(l1r_input_buffer);
        return;
    }

    /* Allocate memory for input striping correction matrix buffer */
    striping_buffer = malloc(NDETS * NLINES * sizeof(double));
    if (!striping_buffer)
    {
        assert_true_with_message(0, "Allocating memory for striping_buffer" );
        free(l1r_input_buffer);
        return;
    }

    /* Read the input striping correction matrix buffer */
    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        free(l1r_input_buffer);
        free(striping_buffer);
        return;
    }

    for (line = 0; line < NLINES; line++)
    {
        for (det = 0; det < NDETS; det++)
        {
            return_value = fscanf(inf, "%lf", &temp_double);
            if (return_value == EOF)
            {
                assert_true_with_message(0, "Reading %s", file_name);
                free(l1r_input_buffer);
                free(striping_buffer);
                return;        
            }        
            striping_buffer[line * NDETS + det] = temp_double;
        }
    }
    fclose(inf);

    /* Calculate scene mean for input data and compare the result with
       expected from ADD or SDSU matlab code value */
    status = compute_float_mean(l1r_input_buffer, NDETS * NLINES, 
                                &input_scene_mean);
    assert_equal_with_message(status, SUCCESS, "Error returned from "
        "compute_float_mean");
    if (status != SUCCESS)
    {
        free(l1r_input_buffer);
        free(striping_buffer);
        return;        
    }
    assert_true((fabs(input_scene_mean - expected_scene_mean) /
                 expected_scene_mean) < 1e-4);

    /* Call the algorithm and test it for band 1 and SCA 1 */
    residual_striping_correction(l1r_input_buffer, striping_buffer, 
                                 NDETS, NLINES);

    /* Calculate scene mean for output data and compare the result with
       expected from ADD or SDSU matlab code value */
    status = compute_float_mean(l1r_input_buffer, NDETS * NLINES, 
                                &output_scene_mean);
    assert_equal_with_message(status, SUCCESS, "Error returned from "
        "compute_float_mean");
    if (status != SUCCESS)
    {
        free(l1r_input_buffer);
        free(striping_buffer);
        return;        
    }

    /* The striping correction is supposed to adjust the image data without
       changing the overall scene mean value */
    assert_true((fabs(output_scene_mean - expected_scene_mean) /
                 expected_scene_mean) < 1e-4);

    /* Open and read the result file generated from SDSU matlab code */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, scene_output_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building result data file name");
        free(l1r_input_buffer);
        free(striping_buffer);
        return;
    }
    l1r_output_buffer = malloc(NDETS * NLINES * sizeof(float));
    if (!l1r_output_buffer)
    {
        assert_true_with_message(0, "Allocating memory for bias " 
                                         "buffer");
        free(l1r_input_buffer);
        free(striping_buffer);
        return;
    }

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        free(l1r_input_buffer);
        free(striping_buffer);
        free(l1r_output_buffer);
        return;
    }

    for (line = 0; line < NLINES; line++)
    {
        for (det = 0; det < NDETS; det++)
        {
            return_value = fscanf(inf, "%f", &temp_float);
            if (return_value == EOF)
            {
                assert_true_with_message(0, "Reading %s", file_name);
                free(l1r_input_buffer);
                free(l1r_output_buffer);
                free(striping_buffer);
                return;        
            }        
            l1r_output_buffer[line * NDETS + det] = temp_float;
        }
    }
    fclose(inf);
    
    /* Compare the results.  Loop through the results and count the
       differences.  This comparison isn't done with asserts since there is
       no need for 320*6578 assert statements which cost message buffer space.
       One assert for pass fail with a count is good enough. */
    difference_count = 0;
    for (pixel = 0; pixel < NDETS * NLINES; pixel++)
    {
        if ((fabsf(l1r_input_buffer[pixel] - l1r_output_buffer[pixel])) > 1e-4)
            difference_count++;
    }
    assert_equal_with_message(difference_count, 0, "There were %d "
        "differences out of %d values", difference_count, NDETS * NLINES);

    /* Free the allocated memory */
    free(l1r_input_buffer);
    free(striping_buffer);
    free(l1r_output_buffer);
}

/*--------------------------------------------------------------------------
  local_free_the_blob

  This a local override of the routine to free the blob.  Since not all
  members of the blob are initialized -- specifically the pixel mask -- the
  normal routine cannot be called.
--------------------------------------------------------------------------*/
static void local_free_the_blob(BLOB *blob)
{
    int i;                  /* Loop counter */
    
    if ( !blob )
        return;

    if ( blob->band_information )
    {
        for ( i = 0; i < blob->total_bands; i++ )
        {
            if ( blob->band_information[i] )
                free( blob->band_information[i] );
        }
    }
    free( blob->band_information );

    free( blob );
}

/*--------------------------------------------------------------------------
  test_algorithm_thru_api

  This test sets up for a call to the algorithm through the API.  Also calls
  the pre-process routine to set up memory management. The input ASCII files 
  are extracted from calval binary test files with only one band and one sca 
  used here. The compared result file is generated from cal/val IDL code.
--------------------------------------------------------------------------*/
Ensure test_algorithm_thru_api(void)
{
    const IAS_BAND_ATTRIBUTES *band_atts;
    FILE *inf;                     /* File pointer for all reads */
    char *test_directory;          /* test environment value */
    char file_name[PATH_MAX];      /* File name to open */
    BLOB *blob;                    /* Declare a blob structure */
    float *l1r_input_buffer;       /* Image data for the current band/SCA */
    float *l1r_output_buffer;      /* Image data for the current band/SCA */
    double *striping_matrix_buffer;/* Striping matrix for the current 
                                      band/SCA */
    float input_scene_mean;        /* Mean of the input test scene */
    float output_scene_mean;       /* Mean of the output test scene */
    float expected_scene_mean = 128.26;     
                                   /* Expected mean value from ADD */
    float temp_float;              /* Temporary value read from file */
    double temp_double;            /* Temporary value read from file */
    int return_value;              /* Return values from fscanf read */ 
    int det;                       /* Detector loop counter */
    int line;                      /* Frame loop counter */
    int pixel;                     /* Pixel loop counter */
    int difference_count;          /* Numbers of differences between the 
                                      calculated value and expected results */
    int band_number = 1;           /* 1-based band number */
    int status;                    /* Return variable */
    int band_index;                /* 0-based band number */
    int list_of_bands[1] = {1};    /* List of one test band */
    int list_of_band_indices[1] = {0};   /* List of one test band index */
    int num_bands = 1;             /* Number of bands for testing */

    /* Get the environment variable for the test data */
    test_directory = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (!test_directory)
    {
        assert_true_with_message(0, "Getting IAS_UNIT_TEST_DATA_DIR");
        return;
    }

    /* The attributes must be initialized for memory management to work */
    ias_sat_attr_initialize(IAS_L8);

    /* Allocate a blob and populate the minimum members for the test
       to succeed */
    blob = allocate_blob();
    blob->max_number_scas = 14;
    blob->total_bands = 30;
    blob->band_information = calloc(blob->total_bands, 
                                    sizeof( RPS_BAND_ATTRIBUTES *));

    /* Local initialize the blob */
    for (band_number = 1; band_number <= blob->total_bands; band_number++)
    {
        int i;
        band_atts = ias_sat_attr_get_band_attributes(band_number);
        if (band_atts == NULL)
        {
            assert_true_with_message(FALSE, "Error obtaining RPS attribute "
                "information for band %d", band_number);
            local_free_the_blob(blob);
            return;
        }

        int band_index = band_atts->band_index;

        blob->band_information[band_index] = calloc(1,
            sizeof(RPS_BAND_ATTRIBUTES));
        if (!blob->band_information[band_index])
        {
            assert_true_with_message(0, "Error allocating band "
                "attribute memory");
            local_free_the_blob(blob);
            return;
        }
        blob->band_information[band_index]->band_number = band_number;
        blob->band_information[band_index]->number_lines = NLINES;
        blob->band_information[band_index]->number_detectors = NDETS;
        blob->band_information[band_index]->number_scas = band_atts->scas;
        blob->band_information[band_index]->band_classification =
            band_atts->band_classification;
        if (band_atts->band_classification & IAS_NORMAL_BAND)
            blob->band_information[band_index]->is_normal_band = 1;
        else
            blob->band_information[band_index]->is_normal_band = 0;
        blob->band_information[band_index]->spectral_type =
            band_atts->spectral_type;

        for ( i = 0; i < num_bands; i++ )
        {
            if ( band_number == list_of_bands[i] )
            {
                list_of_band_indices[i] = band_index;
                break; /* Since there are no repeated bands */
            }
        }
    }

    /* Diable logging to suppress the memory management warnings */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* This initializes the memory management and registers buffers through
       the algorithms pre-process routine.  This algorithm doesn't need
       any memory management buffers that require the list of bands to be
       initialized -- number of bands set to 0 for that reason. */
    blob->number_of_bands = 0;
    blob->band_fields_initialized = 1;
    status = initialize_memory_management_buffers();
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error initializing memory management");
        local_free_the_blob(blob);
        return;
    }

    /* Initialize the RPS processing level arrays */
    status = initialize_rps_level(15, 1);
    if (status != SUCCESS)
    {
        assert_true_with_message( 0, "Error calling initialize_rps_level");
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    status = residual_striping_correction_preprocess(blob);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error returned from: "
            "sca_discontinuity_correction_preprocess");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    /* Register the L1R buffer to this routine so that the API can't
       destroy the memory before we can retrieve the output */
    status = increment_consumer_count(IMAGE_L1R);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error returned from: "
            "increment_consumer_count IMAGE_L1R");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    /* Get the image buffer for band 1, SCA 1 this simply returns an unfilled,
       allocated buffer */
    l1r_input_buffer = (float *)preprocess_get_memory_buffer(IMAGE_L1R, 
        list_of_band_indices[0], 0);
    if (!l1r_input_buffer)
    {
        assert_true_with_message(0, "Getting IMAGE_L1R buffer");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    /* Open and read the input 1 SCA data */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, scene_input_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building L1R input data file "
                                 "name");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    for (line = 0; line < NLINES; line++)
    {
        for (det = 0; det < NDETS; det++)
        {
            return_value = fscanf(inf, "%f", &temp_float);
            if (return_value == EOF)
            {
                assert_true_with_message(0, "Reading %s", file_name);
                cleanup_rps_level();
                cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
                local_free_the_blob(blob);
                return;        
            }        
            l1r_input_buffer[line * NDETS + det] = temp_float;
        }
    }
    fclose(inf);
    set_buffer_filled_flag( IMAGE_L1R, list_of_band_indices[0], 0);

    /* Check an "empty" buffer for the striping matrix */
    striping_matrix_buffer = (double *)preprocess_get_memory_buffer(
        MM_SCENE_STRIPING_CORRECTION_MATRIX, list_of_band_indices[0], 0);
    if (!striping_matrix_buffer)
    {
        assert_true_with_message(0, "Error getting "
            "MM_SCENE_STRIPING_CORRECTION_MATRIX buffer");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    /* Open and read the striping matrix data */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, striping_matrix_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building striping matrix data "
                                 "file name");
        free(l1r_input_buffer);
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    /* Read the input striping correction matrix buffer */
    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        free(l1r_input_buffer);
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    for (line = 0; line < NLINES; line++)
    {
        for (det = 0; det < NDETS; det++)
        {
            return_value = fscanf(inf, "%lf", &temp_double);
            if (return_value == EOF)
            {
                assert_true_with_message(0, "Reading %s", file_name);
                free(l1r_input_buffer);
                cleanup_rps_level();
                cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
                local_free_the_blob(blob);
                return;        
            }        
            striping_matrix_buffer[line * NDETS + det] = temp_double;
        }
    }
    fclose(inf);
    set_buffer_filled_flag( MM_SCENE_STRIPING_CORRECTION_MATRIX,
        list_of_band_indices[0], 0 );

    /* Calculate scene mean for input data and compare the result with
       expected from ADD or SDSU matlab code value */
    status = compute_float_mean(l1r_input_buffer, NDETS * NLINES, 
                                &input_scene_mean);
    assert_equal_with_message(status, SUCCESS, "Error returned from "
        "compute_float_mean");
    if (status != SUCCESS)
    {
        free(l1r_input_buffer);
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob( blob );
        return;        
    }
    assert_true((fabs(input_scene_mean - expected_scene_mean) /
                 expected_scene_mean) < 1e-4);

    /* It is important that the consumer counts get set *after* all the
       increments */
    status = set_consumer_counts();
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error setting consumer counts");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    /* Call the API with band 1 and SCA 1 */
    band_number = list_of_bands[0];
    band_index = list_of_band_indices[0];
    status = api_residual_striping_correction( blob, band_index, 0, 0 );
    assert_equal_with_message( status,SUCCESS, "Error returned from "
        "api_residual_striping_correction");
    if ( status != SUCCESS )
    {
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob( blob );
        return;
    }

    /* Calculate scene mean for output data and compare the result with
       expected from ADD or SDSU matlab code value */
    status = compute_float_mean(l1r_input_buffer, NDETS * NLINES, 
                                &output_scene_mean);
    assert_equal_with_message(status, SUCCESS, "Error returned from "
        "compute_float_mean");
    if (status != SUCCESS)
    {
        free(l1r_input_buffer);
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob( blob );
        return;        
    }
    assert_true((fabs(output_scene_mean - expected_scene_mean) /
                 expected_scene_mean) < 1e-4);

    /* Open and read the result file generated from SDSU matlab code */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, scene_output_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building result data file "
            "name");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob( blob );
        return;
    }
    l1r_output_buffer = malloc(NDETS * NLINES * sizeof(float));
    if (!l1r_output_buffer)
    {
        assert_true_with_message(0, "Allocating memory for bias " 
                                         "buffer");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob( blob );
        return;
    }

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        free(l1r_output_buffer);
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob( blob );
        return;
    }

    for (line = 0; line < NLINES; line++)
    {
        for (det = 0; det < NDETS; det++)
        {
            return_value = fscanf(inf, "%f", &temp_float);
            if (return_value == EOF)
            {
                assert_true_with_message(0, "Reading %s", file_name);
                free(l1r_output_buffer);
                cleanup_rps_level();
                cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
                local_free_the_blob( blob );
                return;        
            }        
            l1r_output_buffer[line * NDETS + det] = temp_float;
        }
    }
    fclose(inf);
    
    /* Compare the results.  Loop through the results and count the
       differences.  This comparison isn't done with asserts since there is
       no need for 320*6578 assert statements which cost message buffer space.
       One assert for pass fail with a count is good enough. */
    difference_count = 0;
    for (pixel = 0; pixel < NDETS * NLINES; pixel++)
    {
        if ((fabsf(l1r_input_buffer[pixel] - l1r_output_buffer[pixel])) > 1e-4)
            difference_count++;
    }
    assert_equal_with_message(difference_count, 0, "There were %d "
        "differences out of %d values", difference_count, NDETS * NLINES);

    /* Free the image memory */
    status = decrement_consumer_count(IMAGE_L1R, 0, 0);
    assert_equal_with_message( status, SUCCESS,
        "Error returned from decrement consumer count");

    free(l1r_output_buffer);
    cleanup_rps_level();
    cleanup_memory_management_buffers(ENABLE_WARNING_DISPLAY);
    local_free_the_blob(blob);
} 
