/*----------------------------------------------------------------------------
NAME:    response_linearization_3q_test

PURPOSE: Test the OLI/TIRS response linearization routine and its API(s)

USAGE:   response_linearization_3q_test [test name]

         When run without a command line argument all tests will be run.
         Pass the name of a test to run the single test -- useful for
         debugging.
----------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "cgreen.h"
#include "ias_logging.h"
#include "../rps_local.h"

#define NLINES 10
#define NDETS  640


/* Path to test data and file names */
static char *test_data_path = NULL;
static char data_dir[] = "ias_base/rps/driver/3q_response_linearization";

/* Test RLUT file.  The file is a "complete" RLUT file (all bands/SCAs present).
   However, the "valid" parameters are for band 10 SCA 2, which were used in
   the cal/val prototype code.  The values for the other bands/SCAs were just
   copied from this set */
static char rlut_basename[] = "test_linearization_rlut_v2.h5";

/* Binary data files containing test input data and expected correction results
   used in the cal/val prototype code */
static char data_file[] = "dataval_S2_R0.dat";
static char results_file[] = "resultval_S2_R0.dat";


/*--------------------------------------------------------------------------
 NAME:     local_free_the_blob

 PURPOSE:  This is a local override of the routine to free the blob.
           Since not all members of the blob are initialized -- specifically
           the pixel mask -- the normal routine cannot be called.

 RETURNS:  Nothing
--------------------------------------------------------------------------*/
static void local_free_the_blob(BLOB *blob)
{
    int band_index;

    for (band_index = 0; band_index < 30; band_index++)
        free(blob->band_information[band_index]);
    free(blob->band_information);
    if (blob->list_of_bands != NULL)
        free(blob->list_of_bands);
    if (blob->list_of_band_indices != NULL)
        free(blob->list_of_band_indices);
    if ( blob->rlut )
    {
        if ( ias_rlut_close_file( blob->rlut ) != SUCCESS )
        {
            assert_true_with_message(FALSE, "Error closing RLUT file %s",
                blob->rlut_filename );
        }
    }
    free(blob->report_information.rlut_version);
    free(blob->report_information.directory);
    free(blob->report_information.rps_level);
    free(blob);
}



/*--------------------------------------------------------------------------
  test_algorithm_through_api

  This test sets up for a call to the algorithm through the OLI/TIRS 
  wrapper APIs.  Also calls the pre-process routine to set up memory
  management.
--------------------------------------------------------------------------*/
Ensure test_algorithm_through_api()
{
    char input_filename[PATH_MAX];          /* Input image data file name */
    BLOB *blob = NULL;                      /* Simulated RPS "blob" */
    const IAS_BAND_ATTRIBUTES *band_atts = NULL;
                                            /* IAS attribute information for
                                               the current band */
    FILE *input_fptr = NULL;                /* Input data file pointer */
    float *l1r_data_buffer = NULL;          /* Image data buffer */
    float *expected_results = NULL;         /* Results data buffer */
    int band_number;                        /* Current band number */
    int band_index;                         /* Band loop counter */
    int difference_count = 0;               /* Count of detector-specific
                                               differences between test
                                               corrected and "reference"
                                               corrected test image data */
    int i;                                  /* Generic loop counter */
    int n;                                  /* Number of values read from
                                               input data file */
    int num_values;                         /* Total number of pixels in
                                               test "image" */
    int sca_number;                         /* Current SCA number */
    int status;
    int list_of_bands[3] = {10, 2, 10};
    int list_of_band_indices[3];
    int num_bands = 3;


    printf("Testing linearization correction APIs...\n");

    /* Allocate a blob and populate the information required for the test
       to succeeed */
    blob = allocate_blob();
    if (blob == NULL)
    {
        assert_true_with_message(FALSE, "Error allocating and/or populating "
            "BLOB data structure");
        return;
    }
    blob->max_number_scas = 14;
    blob->total_bands = 30;

    /* Populate the BLOB band attribute information */
    blob->band_information = calloc((blob->total_bands),
        sizeof(RPS_BAND_ATTRIBUTES *));
    if (blob->band_information == NULL)
    {
        assert_true_with_message(FALSE, "Error populating BLOB band "
            "attribute information");
        local_free_the_blob(blob);
        return;
    }
    blob->number_of_bands = 1;
    blob->list_of_bands = calloc(blob->number_of_bands,
        sizeof(*blob->list_of_bands));
    if (blob->list_of_bands == NULL)
    {
        assert_true_with_message(FALSE, "Error allocating band list");
        local_free_the_blob(blob);
        return;
    }
    blob->list_of_bands[0] = 10;
    blob->list_of_band_indices = calloc(blob->number_of_bands,
        sizeof(*blob->list_of_band_indices));
    if (blob->list_of_band_indices == NULL)
    {
        assert_true_with_message(FALSE, "Error allocating band list of "
            "indices");
        local_free_the_blob(blob);
        return;
    }
    blob->list_of_band_indices[0] = 3;

    /* Populate the RPS band attributes information in the BLOB  */
    for (band_number = 1; band_number <= blob->total_bands; band_number++)
    {
        int i;
        band_atts = ias_sat_attr_get_band_attributes(band_number);
        if (band_atts == NULL)
        {
            assert_true_with_message(FALSE, "Error obtaining RPS attribute "
                "information for band %d", band_number);
            local_free_the_blob(blob);
            return;
        }
        band_index = band_atts->band_index;

        blob->band_information[band_index] = calloc(1,
            sizeof(RPS_BAND_ATTRIBUTES));
        if (blob->band_information[band_index] == NULL)
        {
           assert_true_with_message(FALSE, "Error allocating band %d "
               "attribute data buffer", band_index);
           local_free_the_blob(blob);
           return;
        }
        blob->band_information[band_index]->band_number = band_number;
        blob->band_information[band_index]->number_lines = NLINES;
        blob->band_information[band_index]->number_scas = band_atts->scas;
        blob->band_information[band_index]->number_detectors =
            band_atts->detectors_per_sca;
        blob->band_information[band_index]->band_classification =
            band_atts->band_classification;
        blob->band_information[band_index]->spectral_type =
            band_atts->spectral_type;
        for ( i = 0; i < num_bands; i++ )
        {
            if ( band_number == list_of_bands[i] )
            {
                list_of_band_indices[i] = band_index;
                /* Don't break since a band is repeated in the band list */
            }
        }
    }

    /* Get the RLUT file name into the BLOB */
    status = snprintf(blob->rlut_filename, sizeof(blob->rlut_filename),
        "%s/%s/%s", test_data_path, data_dir, rlut_basename);
    if ((status < 0) || (status >= sizeof(blob->rlut_filename)))
    {
        assert_true_with_message(FALSE, "Error building RLUT file name");
        local_free_the_blob(blob);
        return;
    }

    /* Open the RLUT */
    blob->rlut = ias_rlut_open_file( blob->rlut_filename, IAS_READ );
    if ( !blob->rlut )
    {
        assert_true_with_message(FALSE, "Error opening RLUT file %s",
            blob->rlut_filename );
        local_free_the_blob(blob);
        return;
    }

    /* Disable the logging to supress the memory management warnings */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* Inititalize the memory management and register the data buffers.
       This algorithm doesn't need any memory management buffers that
       require the list of bands to be initialized -- number of bands set
       to 0 for that reason. */
    blob->number_of_bands = 0;
    blob->band_fields_initialized = 1;
    status = initialize_memory_management_buffers();
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error initializing memory management");
        local_free_the_blob(blob);
        return;
    }

    /* Initialize the RPS processing level arrays */
    status = initialize_rps_level(10, 14);
    if (status != SUCCESS)
    {
        assert_true_with_message( 0, "Error calling initialize_rps_level" );
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
    }

    status = response_linearization_3q_preprocess(blob);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error returned from "
            "response_linearization_3q_preprocess");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    /* Register the L1R image data buffer so the API can't destroy it
       before we can check the output */
    status = increment_consumer_count(IMAGE_L1R);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error returned from increment_"
            "consumer_count IMAGE_L1R");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }
    status = increment_consumer_count(IMAGE_L1R);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error returned from increment_"
            "consumer_count IMAGE_L1R");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }
    /* It is important that the consumer counts get set *after* all the
       increments */
    status = set_consumer_counts();
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error setting consumer counts");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    /* Open the input image data file */
    status = snprintf(input_filename, sizeof(input_filename), "%s/%s/%s",
        test_data_path, data_dir, data_file);
    if ((status < 0) || (status >= sizeof(input_filename)))
    {
        assert_true_with_message(FALSE, "Error building input data filename");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
    }
    input_fptr = fopen(input_filename, "r");
    if (input_fptr == NULL)
    {
        assert_true_with_message(FALSE, "Error opening input data file %s",
            input_filename);
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }
   
    /* Get an empty image data buffer for band 10 */
    band_number = list_of_bands[0];
    band_index = list_of_band_indices[0];
    blob->list_of_bands[0] = band_number;
    sca_number = 2;
    l1r_data_buffer = (float *)get_memory_buffer(IMAGE_L1R, SETTER, band_index,
        sca_number - 1);
    if (l1r_data_buffer == NULL)
    {
        assert_true_with_message(FALSE, "Error getting IMAGE_L1R buffer");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        fclose(input_fptr);
        return;
    }
    set_buffer_filled_flag( IMAGE_L1R, band_index, sca_number - 1);

    /* Read the test input data */
    num_values = NLINES * NDETS;
    n = fread(l1r_data_buffer, sizeof(float), num_values, input_fptr);
    fclose(input_fptr);
    if (n != num_values)
    {
        assert_true_with_message(FALSE, "Error reading input data file %s",
            input_filename);
        status = decrement_consumer_count(IMAGE_L1R, band_index,
            sca_number - 1);
        assert_equal_with_message( status, SUCCESS,
            "Error returned from decrement consumer count");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    /* Call the OLI API "wrapper" with the current band/SCA numbers.
       Since band 10 is a TIRS band, SUCCESS should be returned, and
       no buffers should be modified/freed.  Confirm the image data hasn't
       been touched by checking the first detector's output */
    status = api_oli_3q_response_linearization(blob, band_index,
        sca_number - 1, 0);
    assert_equal_with_message(status, SUCCESS, "SUCCESS should have been "
        "returned from an OLI API call with a TIRS band number");
    if (status != SUCCESS)
    {
        status = decrement_consumer_count(IMAGE_L1R, band_index,
            sca_number - 1);
        assert_equal_with_message( status, SUCCESS,
            "Error returned from decrement consumer count");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }
    assert_double_equal((double)l1r_data_buffer[0], 235.6784667969);
    assert_double_equal((double)l1r_data_buffer[640], 412.2143554688);
    assert_double_equal((double)l1r_data_buffer[640 * 2], 597.5356445312);
    assert_double_equal((double)l1r_data_buffer[640 * 3], 1084.5803222656);
    assert_double_equal((double)l1r_data_buffer[640 * 4], 1424.8750000000);
    assert_double_equal((double)l1r_data_buffer[640 * 5], 1710.7231445312);
    assert_double_equal((double)l1r_data_buffer[640 * 6], 1989.7856445312);
    assert_double_equal((double)l1r_data_buffer[640 * 7], 2540.1250000000);
    assert_double_equal((double)l1r_data_buffer[640 * 8], 3612.6340332031);
    assert_double_equal((double)l1r_data_buffer[640 * 9], 4096.0000000000);

    /* Switch the band number to an OLI band number, say 2.  Then call the
       TIRS API "wrapper".  Again, SUCCESS should be returned, and the
       input data shouldn't be touched */
    band_number = list_of_bands[1];
    band_index = list_of_band_indices[1];
    status = api_tirs_3q_response_linearization(blob, band_index,
        sca_number - 1, 0);
    assert_equal_with_message(status, SUCCESS, "SUCCESS should have been "
        "returned from TIRS API call with an OLI band number");
    if (status != SUCCESS)
    {
        band_number = list_of_bands[0];
        status = decrement_consumer_count(IMAGE_L1R, band_index,
            sca_number - 1);
        assert_equal_with_message( status, SUCCESS,
            "Error returned from decrement consumer count");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    assert_double_equal((double)l1r_data_buffer[0], 235.6784667969);
    assert_double_equal((double)l1r_data_buffer[640], 412.2143554688);
    assert_double_equal((double)l1r_data_buffer[640 * 2], 597.5356445312);
    assert_double_equal((double)l1r_data_buffer[640 * 3], 1084.5803222656);
    assert_double_equal((double)l1r_data_buffer[640 * 4], 1424.8750000000);
    assert_double_equal((double)l1r_data_buffer[640 * 5], 1710.7231445312);
    assert_double_equal((double)l1r_data_buffer[640 * 6], 1989.7856445312);
    assert_double_equal((double)l1r_data_buffer[640 * 7], 2540.1250000000);
    assert_double_equal((double)l1r_data_buffer[640 * 8], 3612.6340332031);
    assert_double_equal((double)l1r_data_buffer[640 * 9], 4096.0000000000);

    /* Reset the band number to 10 and call the TIRS "wrapper".  Check the
       output of the first detector to make sure the correction was
       performed */
    band_number = list_of_bands[2];
    band_index = list_of_band_indices[2];
    status = api_tirs_3q_response_linearization(blob, band_index,
        sca_number - 1, 0);
    assert_equal_with_message(status, SUCCESS, "Error processing TIRS band "
        "%d test image data", band_number);
    if (status != SUCCESS)
    {
        status = decrement_consumer_count(IMAGE_L1R, band_index,
            sca_number - 1);
        assert_equal_with_message( status, SUCCESS,
            "Error returned from decrement consumer count");
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        cleanup_rps_level();
        local_free_the_blob(blob);
        return;
    }

    /* Open and read the expected results */
    status = snprintf(input_filename, sizeof(input_filename), "%s/%s/%s",
        test_data_path, data_dir, results_file);
    if ((status < 0) || (status >= sizeof(input_filename)))
    {
        assert_true_with_message(FALSE, "Error building results file name");
        status = decrement_consumer_count(IMAGE_L1R, band_index,
            sca_number - 1);
        assert_equal_with_message( status, SUCCESS,
            "Error returned from decrement consumer count");
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }
    num_values = NLINES * NDETS;
    expected_results = malloc(num_values * sizeof(float));
    if (expected_results == NULL)
    {
        assert_true_with_message( FALSE, "Error allocating memory for the "
           "expected results data buffer");
        status = decrement_consumer_count(IMAGE_L1R, band_index,
            sca_number - 1);
        assert_equal_with_message( status, SUCCESS,
            "Error returned from decrement consumer count");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }
    input_fptr = fopen(input_filename, "r");
    if (input_fptr == NULL)
    {
        assert_true_with_message(FALSE, "Opening expected results file %s",
            input_filename);
        status = decrement_consumer_count(IMAGE_L1R, band_index,
            sca_number - 1);
        assert_equal_with_message( status, SUCCESS,
            "Error returned from decrement consumer count");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        free(expected_results);
        return;
    }
    n = fread(expected_results, sizeof(float), num_values, input_fptr);
    fclose(input_fptr);
    if (n != num_values)
    {
        assert_true_with_message(FALSE, "Unable to read results file %s",
           input_filename);
        status = decrement_consumer_count(IMAGE_L1R, band_index,
            sca_number - 1);
        assert_equal_with_message( status, SUCCESS,
            "Error returned from decrement consumer count");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        free(expected_results);
        return;
    }

    /* Compare the results.  First, do a buffer-to-buffer comparison; if
       the buffers have the same contents, we're done.  If they don't...
       Loop through the results and count the differences.  This comparison
       isn't done with asserts since there is no need for 6400 assert
       statements which cost message buffer space.  One assert for pass
       fail with a count is good enough. */
    num_values = NLINES * NDETS;

    status = memcmp(l1r_data_buffer, expected_results,
        (num_values * sizeof(float)));
    assert_equal_with_message(status, 0, "Buffer comparison failed:  "
        "possible differences found between the test corrected data and "
        "expected values");
    if (status != 0)
    {
        for (i = 0; i < num_values; i++)
        {
            if (l1r_data_buffer[i] != expected_results[i])
                difference_count++;
        }
        if (difference_count == 0)
        {
            assert_true_with_message(FALSE, "Anomalous failure of memcmp() "
                "function--no differences found out of %d values", num_values);
        }
        else    /* Most likely reason for intial failure */
        {
            assert_true_with_message(FALSE, "Found %d differences out of "
                "%d values ", difference_count, num_values);
        }
    }

    /* Cleanup.  First the image data buffer, then ALL of the
       linearization parameter data buffers--increment_consumer_count
       reads all of them at one time */
    status = decrement_consumer_count(IMAGE_L1R, band_index, sca_number - 1);
    assert_equal_with_message( status, SUCCESS,
        "Error returned from decrement consumer count");
    cleanup_rps_level();
    cleanup_memory_management_buffers(ENABLE_WARNING_DISPLAY);
    local_free_the_blob(blob);
    free(expected_results);
}   /* END test_algorithm_through_api */



int main(int argc, char *argv[])
{
    /* Get the test data directory */
    test_data_path = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (test_data_path == NULL)
    {
        printf("Cannot find test data directory location\n");
        return 1;
    }

    /* Initialize the satellite attributes */
    int status = ias_sat_attr_initialize(IAS_L8);
    if (status != SUCCESS)
    {
        printf("Cannot initialize IAS satellite attributes information\n");
        return 1;
    }

    /* Build the test suite */
    TestSuite *suite = create_test_suite();
    add_test(suite, test_algorithm_through_api);

    /* Run either a single test or the entire suite */
    if (argc == 2)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    else
        return run_test_suite(suite, create_text_reporter());
}   /* END unit test main */
