/***************************************************************************
 NAME:      saturated_pixel_char_test

 PURPOSE:   CGreen unit testing for the OLI/TIRS saturated pixel
            charaterization api call.

 RETURNS:   CGreen test status

 ROUTINES:  -load_pixel_mask                (internal -- algorithm tests)
            -saturated_pixel_char_test      (internal -- algorithm tests)
            -local_free_the_blob            (internal -- API test)
            -test_inopdetfill01_from_algorithm
            -test_inopdetfill02_from_algorithm
            -test_inopdetfill03_from_algorithm
            -test_inopdetfill01_from_api
****************************************************************************/
#include "string.h"
#include "cgreen/cgreen.h"
#include "ias_logging.h"
#include "../rps_local.h"

/* Define DEBUG to run single test */
#define DEBUG

static char test_dir[] = ".";
static char bpf_dir[] = "ias_lib/bpf";
static char rlut_dir[] = "ias_lib/rlut";
static char data_dir[] = "ias_base/rps/driver/L0R";
static char l0r_file[] = "LO82220010052011265EDC00";
static char l1rname[] = "L1R_output_file";
static char bpf_name_oli[]  = "test_LO8BPF20140310103310_20140310103345.06";
static char rlut_name[] = "test_linearization_rlut.h5";
char *test_directory;


/****************************************************************************
NAME: saturated_pixel_char_test_from_api

PURPOSE:  To test the api call to saturated pixel characterization

*****************************************************************************/

Ensure saturated_pixel_char_test_from_api(void)
{
    int status;
    IAS_PIXEL_MASK *pm;
    IAS_PIXEL_MASK_ITERATOR *it;
    IAS_PIXEL_MASK_SPAN span;
    char l0r_file_name[512];
    int band_index;             /* The index used in arrays for band */
    int band_number = 0;        /* The actual band number */
    int i;                      /* General loop control variable */
    BLOB *blob;

    /* Initialize the logging library in case any library call issues
       log messages */
    ias_log_initialize("api_saturated_pixel_characterization");

    if ( ias_sat_attr_initialize( IAS_L8 ))
    {
        assert_true_with_message( 0, "Error initializing satellite "
            "attributes library.");
        return;
    }

    /* Initiate the blob */
    sprintf( l0r_file_name, "%s/%s/%s", test_directory, data_dir, l0r_file );
    blob = allocate_blob();
    if ( !blob )
    {
        assert_true_with_message( 0, "allocating blob" );
        return;
    }

    blob->generate_report = 0;
    blob->store_in_database = 0;

    blob->total_bands = 30;
    blob->number_of_bands = 1;
    strcpy( blob->l0r_file_name, l0r_file_name );
    strcpy( blob->l1r_file_name, l1rname );
    sprintf( blob->cpf_filename, "%s/cpf_saturation_level", test_dir );
    sprintf( blob->bpf_filename_oli, "%s/%s/%s", test_directory, bpf_dir, 
             bpf_name_oli );
    strcpy( blob->bpf_filename_tirs, "" );
    sprintf( blob->rlut_filename, "%s/%s/%s", test_directory, rlut_dir, 
             rlut_name );

    /* Several things must be done to ensure that all blob members are
       properly initialized.  We need the attributes library initialized
       before completing the blob initialization.  Then the I/O routines
       must be initialized to populate the band present stuff. */
    if ( finish_blob_initialization())
    {
        assert_true_with_message( 0, "initializing blob -- %s",
            blob->cpf_filename );
        free_the_blob();
        return;
    }
    blob->list_of_bands[0] = 1;
    blob->list_of_band_indices[0] =
        ias_sat_attr_convert_band_number_to_index( blob->list_of_bands[0] );
    if ( blob->list_of_band_indices[0] == ERROR )
    {
        assert_true_with_message( 0, "Converting band number to index");
        free_the_blob();
        return;
    }

    ias_log_set_output_level( IAS_LOG_LEVEL_DISABLE );
    if ( initialize_io_routines( blob ))
    {
        assert_true_with_message( 0,
            "initialize_io_routines returned an error" );
        free_the_blob();
        return;
    }
    ias_log_set_output_level( IAS_LOG_LEVEL_INFO );

    /* Now that the I/O routines are initialized we know what bands are
       present so create the pixel mask. */
    if ( create_blob_pixel_mask())
    {
        assert_true_with_message( 0, "Error returned from "
            "create_blob_pixel_mask");
        cleanup_io_routines( blob );
        free_the_blob();
        return;
    }

    /* Initialize the memory management layer and register the saturated
       pixel characterization memory by calling the pre-process routine */
    status = initialize_memory_management_buffers();
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error initializing memory management");
        cleanup_io_routines( blob );
        free_the_blob();
        return;
    }

    if (saturated_pixel_characterization_preprocess( blob ))
    {
        assert_true_with_message( 0, "Error returned from "
            "saturated_pixel_characterization_preprocess");
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        cleanup_io_routines( blob );
        free_the_blob();
        return;
    }
    /* It is important that the consumer counts get set *after* all the
       increments */
    status = set_consumer_counts();
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error setting consumer counts");
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        cleanup_io_routines( blob );
        free_the_blob();
        return;
    }

    /* Call saturated pixel characterization api for normal band 1 */
    /* The pixel mask will be populated for both band 1 and 19 since vrp band 19
       is associated with band 1 */
    status = 0;
    band_index = blob->list_of_band_indices[0];
    if ( api_saturated_pixel_characterization( blob, band_index, 0, 0 ))
        status = 1;

    /* Check if status fail for either call */
    if ( status )
    {
        assert_true_with_message( 0, "Error returned from "
            "api_saturated_pixel_characterization");
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        cleanup_io_routines( blob );
        free_the_blob();
        return;
    }

    /* Check the state of the pixel mask -- this test shouldn't return
       any inoperable detectors. */
    for (i = 0; i < 2; i++)
    {
        int expected_digital_high;
        int expected_digital_low;
        int expected_analog_high;
        int expected_analog_low;

        /* First iteration, Normal band 1 */
        if (i == 0)
        {
            pm = blob->pixel_mask[0][0];
            band_number = 1;
            expected_digital_high = 5;
            expected_digital_low = 4;
            expected_analog_high = 3;
            expected_analog_low = 2;
        }
        /* Second iteration VRP band 19 */
        else
        {
            pm = blob->pixel_mask[18][0];
            band_number = 19;
            expected_digital_high = 9;
            expected_digital_low = 8;
            expected_analog_high = 7;
            expected_analog_low = 6;
        }

        /* Get iterator for the Digital Low Pixels */
        it = ias_pm_get_iterator( pm, IAS_PM_INCLUDE, IAS_RPS_ADCONV_SAT_LO );
        if ( !it )
        {
            assert_true_with_message( 0, "No Digital Low Saturated Pixel"
                " returned for pixel mask band %d, SCA 1", band_number);
        }
        else
        {
            status = ias_pm_get_next_span( it, &span );
            if (status == 0)
                assert_true_with_message( status, 
                    "No Digital Low Saturated Pixel"
                    " retrieved for pixel mask band %d, SCA 1", band_number);
            else
            {
                assert_equal_with_message(expected_digital_low, 
                    span.detector_index,
                    " Digital Low band %d, SCA 1 found" 
                    " detector %d but expected %d",
                    band_number, span.detector_index, expected_digital_low);
            }
            ias_pm_destroy_iterator( it );
        }

        /* Get iterator for the Digital High Pixels */
        it = ias_pm_get_iterator( pm, IAS_PM_INCLUDE, IAS_RPS_ADCONV_SAT_HI );
        if ( !it )
        {
            assert_true_with_message( 0, "No Digital High Saturated Pixel"
                " returned for pixel mask band %d, SCA 1", band_number);
        }
        else
        {
            status = ias_pm_get_next_span( it, &span );
            if (status == 0)
                assert_true_with_message( status, 
                    "No Digital High Saturated Pixel"
                    " retrieved for pixel mask band %d, SCA 1", band_number);
            else
            {
                assert_equal_with_message(expected_digital_high, 
                    span.detector_index,
                    " Digital High band %d, SCA 1 found" 
                    " detector %d but expected %d",
                    band_number, span.detector_index, expected_digital_high);
            }
            ias_pm_destroy_iterator( it );
        }

        /* Get iterator for the Analog Low Pixels */
        it = ias_pm_get_iterator( pm, IAS_PM_INCLUDE, IAS_RPS_ANALOG_SAT_LO );
        if ( !it )
        {
            assert_true_with_message( 0, "No Analog Low Saturated Pixel"
                " returned for pixel mask band %d, SCA 1", band_number);
        }
        else
        {
            status = ias_pm_get_next_span( it, &span );
            if (status == 0)
                assert_true_with_message( status, 
                    "No Analog Low Saturated Pixel"
                    " retrieved for pixel mask band %d, SCA 1", band_number);
            else
            {
                assert_equal_with_message(expected_analog_low, 
                    span.detector_index,
                    " Analog Low band %d, SCA 1 found" 
                    " detector %d but expected %d",
                    band_number, span.detector_index, expected_analog_low);
            }
            ias_pm_destroy_iterator( it );
        }

        /* Get iterator for the Analog High Pixels */
        it = ias_pm_get_iterator( pm, IAS_PM_INCLUDE, IAS_RPS_ANALOG_SAT_HI );
        if ( !it )
        {
            assert_true_with_message( 0, "No Analog High Saturated Pixel"
                " returned for pixel mask band %d, SCA 1", band_number);
        }
        else
        {
            status = ias_pm_get_next_span( it, &span );
            if (status == 0)
                assert_true_with_message( status, 
                    "No Analog High Saturated Pixel"
                    " retrieved for pixel mask band %d, SCA 1", band_number);
            else
            {
                assert_equal_with_message(expected_analog_high, 
                    span.detector_index,
                    " Analog High band %d, SCA 1 found" 
                    " detector %d but expected %d",
                    band_number, span.detector_index, expected_analog_high);
            }
            ias_pm_destroy_iterator( it );
        }
    }

    cleanup_memory_management_buffers(ENABLE_WARNING_DISPLAY);
    cleanup_io_routines( blob );
    free_the_blob();
    remove( l1rname );
}

/********************* RUN UNIT TEST(S) ***********************************/
int main(int argc, char *argv[])
{
    /* Get the test data path.  Quit here if it isn't defined */
    test_directory = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (test_directory == NULL)
    {
        printf("Cannot find main IAS test data directory\n");
        exit(EXIT_FAILURE);
    }

    /* Create the main test suite */
    TestSuite *suite = create_test_suite();

    add_test(suite, saturated_pixel_char_test_from_api);

    /* Run the test(s) */
    if (argc > 1)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}

