/*----------------------------------------------------------------------------
NAME:    sca_discontinuity_correction_test.c

PURPOSE: Test the sca_discontinuity_correction routine by testing it directly 
         and testing it through API

USAGE:   sca_discontinuity_correction_test [test name]

         When run without a command line argument all tests will be run.
         Pass the name of a test to run the single test -- useful for
         debugging.
----------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "cgreen/cgreen.h"
#include "ias_logging.h"
#include "ias_pixel_mask.h"
#include "../rps_local.h"

#define NLINES 6578
#define NDETS  320

#define MLINES 4
#define MDETS  3

/* Path to test data and file names */
static char data_dir[] = "ias_base/rps/driver/sca_discontinuity_correction";
static char sca1_input_file[] = "sca1_input.txt";
static char sca1_output_file[] = "sca1_output.txt";

static char cpf_dir[] = "ias_lib/cpf";
static char cpf_file[] = "L8CPF20090101_20090331.01";

/* Test prototypes */
Ensure test_algorithm_directly(void);
Ensure test_algorithm_directly_with_artifacts(void);
Ensure test_calc_normalized_correction_factor(void);
Ensure test_algorithm_thru_api(void);

int main(int argc, char *argv[])
{
    TestSuite *suite = create_test_suite();
    add_test(suite, test_algorithm_directly);
    add_test(suite, test_algorithm_directly_with_artifacts);
    add_test(suite, test_calc_normalized_correction_factor);
    add_test(suite, test_algorithm_thru_api);

    if (argc == 2)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}


/*--------------------------------------------------------------------------
   test_algorithm_directly

   This test executes the algorithm with a subset of the Cal/Val prototype
   inputs and outputs. The outputs are the results of the SDSU matlab code
   run. 
---------------------------------------------------------------------------*/
Ensure test_algorithm_directly(void)
{
    FILE *inf;                     /* File pointer for all reads */
    char *test_directory;          /* Test environment value */
    char file_name[PATH_MAX];      /* File name to open */
    IAS_PIXEL_MASK *pm;            /* Pixel mask for VRP data */
    float *l1r_input_buffer;       /* Image data for the current band/SCA */
    float *l1r_output_buffer;      /* Image data for the current band/SCA */
    float temp_float;              /* Temporary value read from file */
    double normalized_correction_factor = 1.0192;
    int difference_count;          /* Numbers of differences between the 
                                      calculated value and expected results */
    int return_value;              /* Return values from fscanf read */ 
    int line;                      
    int det;
    int pixel;
    int status;

    /* Get the environment variable for the test data */
    test_directory = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (!test_directory)
    {
        assert_true_with_message(0, "Getting IAS_UNIT_TEST_DATA_DIR");
        return;
    }

    /* Create one pixel mask */
    pm = ias_pm_create(1, 1, NDETS, NLINES);
    assert_not_equal(pm, NULL);
    if (!pm)
    {
        assert_true_with_message(FALSE, "Error creating pixel mask");
        return;
    }


    /* Open and read the input 1 SCA data */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, sca1_input_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building shutter buffer file "
                                 "name");
        ias_pm_destroy(pm);
        return;
    }

    /* Allocate memory for input buffer */
    l1r_input_buffer = malloc(NDETS * NLINES * sizeof(float));
    if (!l1r_input_buffer)
    {
        assert_true_with_message(0, "Allocating memory for dark_buffer" );
        ias_pm_destroy(pm);
        return;
    }

    /* Read the input l1r buffer */
    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        ias_pm_destroy(pm);
        free(l1r_input_buffer);
        return;
    }

    for (line = 0; line < NLINES; line++)
    {
        for (det = 0; det < NDETS; det++)
        {
            return_value = fscanf(inf, "%f", &temp_float);
            if (return_value != 1)
            {
                assert_true_with_message(0, "Reading %s", file_name);
                ias_pm_destroy(pm);
                free(l1r_input_buffer);
                return;        
            }        
            l1r_input_buffer[line * NDETS + det] = (float)temp_float;
        }
    }
    fclose(inf);

    /* Call the algorithm and test it for band 1 and SCA 1 */
    status = sca_discontinuity_correction(l1r_input_buffer, NDETS, NLINES, 
                                          pm, normalized_correction_factor);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Calling sca_discontinuity_correction" );
        ias_pm_destroy(pm);
        free(l1r_input_buffer);
        return;
    }

    /* Open and read the result file generated from SDSU matlab code */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, sca1_output_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building result data file name");
        ias_pm_destroy(pm);
        free(l1r_input_buffer);
        return;
    }
    l1r_output_buffer = malloc(NDETS * NLINES * sizeof(float));
    if (!l1r_output_buffer)
    {
        assert_true_with_message(0, "Allocating memory for bias " 
                                         "buffer");
        ias_pm_destroy(pm);
        free(l1r_input_buffer);
        return;
    }

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        ias_pm_destroy(pm);
        free(l1r_input_buffer);
        free(l1r_output_buffer);
        return;
    }

    for (line = 0; line < NLINES; line++)
    {
        for (det = 0; det < NDETS; det++)
        {
            return_value = fscanf(inf, "%f", &temp_float);
            if (return_value != 1)
            {
                assert_true_with_message(0, "Reading %s", file_name);
                ias_pm_destroy(pm);
                free(l1r_input_buffer);
                free(l1r_output_buffer);
                return;        
            }        
            l1r_output_buffer[line * NDETS + det] = temp_float;
        }
    }
    fclose(inf);
    
    /* Compare the results.  Loop through the results and count the
       differences.  This comparison isn't done with asserts since there is
       no need for 320*6578 assert statements which cost message buffer space.
       One assert for pass fail with a count is good enough. */
    difference_count = 0;
    for (pixel = 0; pixel < NDETS * NLINES; pixel++)
    {
        if ((fabsf(l1r_input_buffer[pixel] - l1r_output_buffer[pixel]) /
            l1r_output_buffer[pixel]) > 1e-4)
            difference_count++;
    }
    assert_equal_with_message(difference_count, 0, "There were %d "
        "differences out of %d values", difference_count, NDETS * NLINES);

    /* Free the allocated memory */
    ias_pm_destroy(pm);
    free(l1r_input_buffer);
    free(l1r_output_buffer);
}

/*--------------------------------------------------------------------------
   test_algorithm_directly_with_artifacts

   This test executes the algorithm directly with some artifacts set in the
   pixel_mask, the values of artifact pixels (inop or drop frame pixels here) 
   doesn't change while the other pixel values changed to the expexted values. 
----------------------------------------------------------------------------*/
Ensure test_algorithm_directly_with_artifacts(void)
{
    float *l1r_buffer;                     /* pointer to the l1r data */
    IAS_PIXEL_MASK *pm;                /* Pixel mask for VRP data */
    double normalized_correction_factor = 1.0192;
    float l1r_data[3][4];              /* input L1r image data */
    float expected_results[] = 
                       {0.000000000, 3.000000000, 6.115200043, 9.000000000,
                        1.000000000, 4.000000000, 7.134399891, 10.000000000,
                        2.038399935, 5.096000195, 8.000000000, 11.211199760};
    int frame;                      
    int det;
    int status;

    /* Create one pixel mask */
    pm = ias_pm_create(1, 1, MDETS, MLINES);
    assert_not_equal(pm, NULL);
    if (!pm)
    {
        assert_true_with_message(FALSE, "Error creating pixel mask");
        return;
    }

    /* Add the inoperable detectors to pixel mask */
    assert_equal(SUCCESS, ias_pm_add_pixels(pm, 
                 2-1, 1-1, 4, IAS_RPS_INOP_DETECTOR));
    /* Add the dropped frames to pixel mask */
    for ( det = 1-1; det <= 3-1; det++ )
    {
        assert_equal(SUCCESS, ias_pm_add_pixels(pm, det, 
                     2-1, 1, IAS_RPS_DROPLINE));
    }


    /* Loop through the detector */
    for (det = 0; det < MDETS; det++)
    {
        /* Loop through the frame */
        for (frame = 0; frame < MLINES; frame++)
        {
            l1r_data[det][frame] = (float) (det + frame * MDETS);
        }
    }

    /* Create a pointer link to the input l1r image data */
    l1r_buffer = &l1r_data[0][0];

    /* Call the algorithm and test it for band 1 and SCA 1 */
    status = sca_discontinuity_correction(l1r_buffer, MDETS, MLINES, 
                                          pm, normalized_correction_factor);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Calling sca_discontinuity_correction" );
        ias_pm_destroy(pm);
        return;
    }

    /* Check the results, the values of artifact pixels (inop or drop frame 
       pixels here) doesn't change while the other pixel values changed */  
    /* Loop through the detector */
    for (det = 0; det < MDETS; det++)
    {
        /* Loop through the frame */
        for (frame = 0; frame < MLINES; frame++)
        {
            /* Check the algorithm results */
            assert_double_equal(l1r_data[det][frame],
                expected_results[frame + det * MLINES]);
        }
    } 

    /* Free the allocated memory */
    ias_pm_destroy(pm);
}

/*--------------------------------------------------------------------------
   test_calc_normalized_correction_factor

   This test tests the execution of calc_normalized_correction_factor routine
   to get the normalized sca correction factor and compare results with SDSU
   matlab code results.
---------------------------------------------------------------------------*/
Ensure test_calc_normalized_correction_factor(void)
{
    double sca_ratio[] = {1.0,0.8497, 1.1724, 1.0830};
    double correction_factor[4];
    double expected_correction_factor[] =
        {1.019171, 0.865989, 1.015286, 1.099554};
    int nscas = 4;
    int difference_count;          /* Numbers of differences between the 
                                      calculated value and expected results */
    int sca_index;
    int status;

    status = calc_normalized_correction_factor(&sca_ratio[0], nscas, 
                                               &correction_factor[0]);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error returned from: "
            "calc_normalized_correction_factor");
        return;
    }
    
    /* Compare results for MM_SCA_CORRECTION_FACTOR */
    for (sca_index = 0; sca_index < nscas; sca_index++)
    {
        /* Check the correction factor with the correct_factor calculated
           from SDSU matlab code */
        difference_count = 0;
        if (fabsf(correction_factor[sca_index] - 
            expected_correction_factor[sca_index]) > 1e-4)
            difference_count++;
    }
    assert_equal_with_message(difference_count, 0, "There were %d "
        "differences out of %d values", difference_count, nscas);
}
 
/*--------------------------------------------------------------------------
  local_free_the_blob

  This a local override of the routine to free the blob.  Since not all
  members of the blob are initialized -- specifically the pixel mask -- the
  normal routine cannot be called.
--------------------------------------------------------------------------*/
static void local_free_the_blob(BLOB *blob)
{
    RPS_BAND_ATTRIBUTES *atts;
    int band_index;
    int sca_index;
    int i;
    
    if ( !blob )
        return;

    for ( band_index = 0; band_index < blob->total_bands; band_index++ )
    {
        if ( blob->pixel_mask[band_index] == NULL )
            continue;

        atts = get_band_attributes( band_index );
        for ( sca_index = 0; sca_index < atts->number_scas; sca_index++ )
        {
            if ( blob->pixel_mask[band_index][sca_index] != NULL )
                ias_pm_destroy( blob->pixel_mask[band_index][sca_index] );
        }

        free( blob->pixel_mask[band_index] );
    }
    free( blob->pixel_mask );
    free( blob->list_of_bands );
    free( blob->list_of_band_indices );

    if ( blob->l1r_band_metadata )
    {
        for ( i = 0; i <= blob->total_bands; i++ )
        {
            if ( blob->l1r_band_metadata[i] )
                free( blob->l1r_band_metadata[i] );
        }
    }
    free( blob->l1r_band_metadata );

    if ( blob->band_information )
    {
        for ( i = 0; i < blob->total_bands; i++ )
        {
            if ( blob->band_information[i] )
                free( blob->band_information[i] );
        }
    }
    free( blob->band_information );

    ias_misc_free_report_header( blob->report_information.header );
    free( blob->report_information.directory );

    if ( blob->cpf )
        ias_cpf_free( blob->cpf );
    free( blob );
}

/*--------------------------------------------------------------------------
  test_algorithm_thru_api

  This test sets up for a call to the algorithm through the API.  Also calls
  the preprocess routine to set up memory management and calls the postprocess 
  routine to generate the output report file. The compared result file is 
  generated from SDSU matlab code.
--------------------------------------------------------------------------*/
Ensure test_algorithm_thru_api(void)
{
    const IAS_BAND_ATTRIBUTES *band_atts;
    FILE *inf;                     /* File pointer for all reads */
    char *test_directory;          /* test environment value */
    char file_name[PATH_MAX];      /* File name to open */
    BLOB *blob;                    /* Declare a blob structure */
    float *l1r_input_buffer;       /* Image data for the current band/SCA */
    float *l1r_output_buffer;      /* Image data for the current band/SCA */
    float temp_float;              /* Temporary value read from file */
    double *correction_factor = NULL;
    double sca_correction_factor = 1.0192;
    int return_value;              /* Return values from fscanf read */ 
    int band;                      /* Band loop counter */
    int sca;                       /* SCA loop counter */
    int det;                       /* Detector loop counter */
    int line;                      /* Frame loop counter */
    int pixel;                     /* Pixel loop counter */
    int difference_count;          /* Numbers of differences between the 
                                      calculated value and expected results */
    int band_number = 1;           /* 1-based band number */
    int status;
    char work_order_id[IAS_WORKORDER_ID_SIZE];
    char routine_name[] = "SCA Discontinuity Correction";
    IAS_REPORT_HEADER *report_header = NULL;
    int band_index;                /* 0-based band number */

    /* Get the environment variable for the test data */
    test_directory = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (!test_directory)
    {
        assert_true_with_message(0, "Getting IAS_UNIT_TEST_DATA_DIR");
        return;
    }

    /* The attributes must be initialized for memory management to work */
    ias_sat_attr_initialize(IAS_L8);

    /* Allocate a blob and populate the minimum members for the test
       to succeed */
    blob = allocate_blob();
    blob->max_number_scas = 14;
    blob->total_bands = 30;
    blob->band_information = calloc(blob->total_bands, 
                                    sizeof( RPS_BAND_ATTRIBUTES *));

    /* Create a set of pixel masks.  Assume all detectors are perfectly
       operable */
    blob->pixel_mask = calloc(blob->total_bands, sizeof(IAS_PIXEL_MASK **));
    if (blob->pixel_mask == NULL)
    {
        assert_true_with_message(FALSE, "Cannot allocate blob pixel mask "
            "base array");
        local_free_the_blob(blob);
        return;
    }

    /* Allocate memory blob->l1r_band_metadata */
    blob->l1r_band_metadata = calloc((blob->total_bands + 1),
        sizeof(IAS_L1R_BAND_METADATA *));
    if (!blob->l1r_band_metadata)
    {
        assert_true_with_message(0, "Allocating blob->l1r_band_metadata "
            "memory");
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    /* Allocate memory for blob->list_of_bands & blob->list_of_band_indices */
    blob->list_of_bands = calloc(blob->total_bands, sizeof(int));
    if (!blob->list_of_bands)
    {
        assert_true_with_message(0, "Allocating blob->list_of_bands memory");
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }
    blob->list_of_band_indices = calloc(blob->total_bands, sizeof(int));
    if (!blob->list_of_band_indices)
    {
        assert_true_with_message(0, "Allocating blob->list_of_bands memory");
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    /* Initialize the RPS level arrays */
    status = initialize_rps_level(15, 1);
    if (status != SUCCESS)
    {   
        assert_true_with_message(0, "Calling initialize_rps_level");
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    /* Local initialize the blob */
    for (band_number = 1; band_number <= blob->total_bands; band_number++)
    {
        band_atts = ias_sat_attr_get_band_attributes(band_number);
        if (band_atts == NULL)
        {
            assert_true_with_message(FALSE, "Error obtaining RPS attribute "
                "information for band %d", band_number);
            cleanup_rps_level();
            local_free_the_blob(blob);
            return;
        }

        int band_index = band_atts->band_index;

        blob->band_information[band_index] = calloc(1,
            sizeof(RPS_BAND_ATTRIBUTES));
        if (!blob->band_information[band_index])
        {
            assert_true_with_message(0, "Error allocating band "
                "attribute memory");
            cleanup_rps_level();
            local_free_the_blob(blob);
            return;
        }
        blob->band_information[band_index]->band_number = band_number;
        blob->band_information[band_index]->number_lines = NLINES;
        blob->band_information[band_index]->number_detectors = NDETS;
        blob->band_information[band_index]->number_scas = band_atts->scas;
        blob->band_information[band_index]->band_classification =
            band_atts->band_classification;
        if (band_atts->band_classification & IAS_NORMAL_BAND)
            blob->band_information[band_index]->is_normal_band = 1;
        else
            blob->band_information[band_index]->is_normal_band = 0;
        blob->band_information[band_index]->spectral_type =
            band_atts->spectral_type;

        blob->l1r_band_metadata[band_index] =
            calloc(1, sizeof(IAS_L1R_BAND_METADATA));
        if (!blob->l1r_band_metadata[band_index])
        {
            assert_true_with_message(0, "Error allocating L1R band "
                "metadata structure");
            cleanup_rps_level();
            local_free_the_blob(blob);
            return;
        }
    }
    blob->band_information[0]->number_scas = 1;
    blob->generate_report = 1;
    blob->number_of_bands = 2;

    blob->list_of_bands[0] = 11;
    blob->list_of_bands[1] = 1;
    blob->list_of_band_indices[0] =
        ias_sat_attr_convert_band_number_to_index( blob->list_of_bands[0] );
    blob->list_of_band_indices[1] =
        ias_sat_attr_convert_band_number_to_index( blob->list_of_bands[1] );
    if ( blob->list_of_band_indices[0] == ERROR ||
         blob->list_of_band_indices[1] == ERROR )
    {
        assert_true_with_message( 0, "Converting band number to index");
        cleanup_rps_level();
        free_the_blob();
        return;
    }
    blob->number_of_bands = 2;
    blob->band_fields_initialized = 1;

    /* Initialize the L1R band metadata for the two test bands */
    blob->l1r_band_metadata[blob->list_of_band_indices[0]][0].band_number =
        blob->list_of_bands[0];
    strcpy(blob->l1r_band_metadata[blob->list_of_band_indices[0]][0].band_name,
        ias_sat_attr_convert_band_number_to_name(blob->list_of_bands[0]));
    strcpy(blob->l1r_band_metadata[blob->list_of_band_indices[0]][0]
        .instrument_source, "VARIOUS");

    blob->l1r_band_metadata[blob->list_of_band_indices[1]][0].band_number =
        blob->list_of_bands[1];
    strcpy(blob->l1r_band_metadata[blob->list_of_band_indices[1]][0].band_name,
        ias_sat_attr_convert_band_number_to_name(blob->list_of_bands[1]));
    strcpy(blob->l1r_band_metadata[blob->list_of_band_indices[1]][0]
        .instrument_source, "VARIOUS");

    /* Create blob pixel_masks */
    status = create_blob_pixel_mask();
    assert_equal_with_message(status, SUCCESS, "Error creating blob pixel "
        "masks");
    if (status != SUCCESS)
    {
        cleanup_rps_level();
        local_free_the_blob(blob);
        return;
    }

    /* Set up a standard header */
    setenv("PCS_WOID", "I4582", 1);
    strcpy(work_order_id, "I4582");
    report_header = ias_misc_initialize_report_header("",
        work_order_id, routine_name);
    if (!report_header)
    {
        assert_true_with_message(0, "Error initializing report header");
        cleanup_rps_level();
        local_free_the_blob(blob);
        return;
    }
    ias_misc_set_header_pathrow(report_header, 0, 0);
    blob->report_information.header = report_header;
    blob->report_information.directory = NULL;

    /* Open the CPF */
    sprintf(file_name, "%s/%s/%s", test_directory, cpf_dir, cpf_file);
    blob->cpf = ias_cpf_read(file_name);
    if (!blob->cpf)
    {
        assert_true_with_message(0, "Error opening CPF: %s", file_name);
        cleanup_rps_level();
        local_free_the_blob(blob);
        return;
    }

    /* This initializes the memory management and registers buffers through
       the algorithms pre-process routine */
    status = initialize_memory_management_buffers();
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error initializing memory management");
        cleanup_rps_level();
        local_free_the_blob(blob);
        return;
    }

    status = sca_discontinuity_correction_preprocess(blob);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error returned from: "
            "sca_discontinuity_correction_preprocess");
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        cleanup_rps_level();
        local_free_the_blob(blob);
        return;
    }

    /* Register the L1R buffer to this routine so that the API can't
       destroy the memory before we can retrieve the output */
    status = increment_consumer_count(IMAGE_L1R);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error returned from: "
            "increment_consumer_count IMAGE_L1R");
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        cleanup_rps_level();
        local_free_the_blob(blob);
        return;
    }
    status = set_consumer_counts();
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error setting consumer counts");
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        cleanup_rps_level();
        local_free_the_blob(blob);
        return;
    }


    /* Initialize the memory buffers */
    for (band = 0; band < blob->number_of_bands; band++)
    {
        for (sca = 0; sca < blob->band_information[
            blob->list_of_band_indices[band]]->number_scas; sca++)
        {
            /* Set MM_SCA_CORRECTION_FACTOR_CPF buffer be 1.0192 */
            correction_factor = (double *)get_memory_buffer(
                MM_SCA_CORRECTION_FACTOR_CPF, SETTER,
                blob->list_of_band_indices[band], sca);
            if (!correction_factor)
            {
                assert_true_with_message(0, "Error getting "
                    "MM_SCA_CORRECTION_FACTOR_CPF buffer");
                cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
                cleanup_rps_level();
                local_free_the_blob(blob);
                return;
            }
            correction_factor[0] = sca_correction_factor;
            set_buffer_filled_flag( MM_SCA_CORRECTION_FACTOR_CPF,
                blob->list_of_band_indices[band], sca );
        }
    }

    /* Get the image buffer for band 1, SCA 1 this simply returns an unfilled,
       allocated buffer */
    l1r_input_buffer = (float *)get_memory_buffer(IMAGE_L1R, SETTER, 0, 0);
    if (!l1r_input_buffer)
    {
        assert_true_with_message(0, "Getting IMAGE_L1R buffer");
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        cleanup_rps_level();
        local_free_the_blob(blob);
        return;
    }

    /* Open and read the input 1 SCA data */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, sca1_input_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building shutter buffer file "
                                 "name");
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        cleanup_rps_level();
        local_free_the_blob(blob);
        return;
    }

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        cleanup_rps_level();
        local_free_the_blob(blob);
        return;
    }

    for (line = 0; line < NLINES; line++)
    {
        for (det = 0; det < NDETS; det++)
        {
            return_value = fscanf(inf, "%f", &temp_float);
            if (return_value != 1)
            {
                assert_true_with_message(0, "Reading %s", file_name);
                cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
                cleanup_rps_level();
                local_free_the_blob(blob);
                return;        
            }        
            l1r_input_buffer[line * NDETS + det] = (float)temp_float;
        }
    }
    fclose(inf);
    set_buffer_filled_flag( IMAGE_L1R, 0, 0 );

    /* Diable logging to suppress info and debug output */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* INFO: removed a useless test ... The API will *never* be called with
             a non-normal band *ever* */

    /* Call the API with band 1 and SCA 1 */
    band_index = blob->list_of_band_indices[1];
    status = api_sca_discontinuity_correction( blob, band_index, 0, 0 );
    assert_equal_with_message( status,SUCCESS, "Error returned from "
        "api_sca_discontinuity_correction");
    if ( status != SUCCESS )
    {
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        cleanup_rps_level();
        local_free_the_blob( blob );
        return;
    }
    ias_log_set_output_level(IAS_LOG_LEVEL_WARN);

    /* Open and read the result file generated from SDSU matlab code */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_directory,
             data_dir, sca1_output_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building result data file "
            "name");
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        cleanup_rps_level();
        local_free_the_blob( blob );
        return;
    }
    l1r_output_buffer = malloc(NDETS * NLINES * sizeof(double));
    if (!l1r_output_buffer)
    {
        assert_true_with_message(0, "Allocating memory for bias " 
                                         "buffer");
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        cleanup_rps_level();
        local_free_the_blob( blob );
        return;
    }

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        free(l1r_output_buffer);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        cleanup_rps_level();
        local_free_the_blob( blob );
        return;
    }

    for (line = 0; line < NLINES; line++)
    {
        for (det = 0; det < NDETS; det++)
        {
            return_value = fscanf(inf, "%f", &temp_float);
            if (return_value != 1)
            {
                assert_true_with_message(0, "Reading %s", file_name);
                free(l1r_output_buffer);
                cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
                cleanup_rps_level();
                local_free_the_blob( blob );
                return;        
            }        
            l1r_output_buffer[line * NDETS + det] = temp_float;
        }
    }
    fclose(inf);
    
    /* Compare the results.  Loop through the results and count the
       differences.  This comparison isn't done with asserts since there is
       no need for 320*6578 assert statements which cost message buffer space.
       One assert for pass fail with a count is good enough. */
    difference_count = 0;
    for (pixel = 0; pixel < NDETS * NLINES; pixel++)
    {
        if ((fabsf(l1r_input_buffer[pixel] - l1r_output_buffer[pixel]) /
            l1r_output_buffer[pixel]) > 1e-4)
            difference_count++;
    }
    assert_equal_with_message(difference_count, 0, "There were %d "
        "differences out of %d values", difference_count, NDETS * NLINES);

    /* Check the L1R band metadata to ensure a discontinuity factor
       source was recorded for this band */
    assert_string_not_equal(
        blob->l1r_band_metadata[band_index][0].discontinuity_factor_source,
        "");
    printf("Discontinuity correction source factor:  %s\n",
        blob->l1r_band_metadata[band_index][0].discontinuity_factor_source);

    /* Call the postprocess routine to test that it works and can generate the 
       output report file accordingly. For this test the TIRS band (band 11)
       correction factors are zeros for all SCA and the number_scas is
       set to be 1 for band 1 for this tests case */
    status = sca_discontinuity_correction_postprocess(blob);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error returned from: "
            "sca_discontinuity_correction_postprocess");
        cleanup_rps_level();
        free(l1r_output_buffer);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    /* Free the image memory */
    status = decrement_consumer_count(IMAGE_L1R, 0, 0);
    assert_equal_with_message( status, SUCCESS,
        "Error returned from decrement consumer count");

    free(l1r_output_buffer);
    cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
    cleanup_rps_level();
    local_free_the_blob(blob);
} 
