/* Standard Library Includes */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

/* IAS Includes */
#include "cgreen/cgreen.h"
#include "ias_satellite_attributes.h"
#include "ias_db_get_connect_info.h"
#include "ias_db_insert.h"
#include "ias_db.h"
#include "ias_db_query.h"
#include "ias_logging.h"
#include "../rps_local.h"
#include "../rps_blob.h"

/* Global variables */
static int test_database = 0;

/* Function prototypes */
struct ias_db_connection *db_open(void);
void db_close(struct ias_db_connection *);
Ensure the_correct_summary_statistics_through_driver();
int db_check_statistics_in_database();

/*   Main test program */
int main( int argc, char *argv[] )
{
    char *enable_db_tests;

    TestSuite *suite = create_test_suite();
    add_test(suite, the_correct_summary_statistics_through_driver);

    /* See if the database tests should be enabled */
    enable_db_tests = getenv("ENABLE_DB_TESTS");
    if (enable_db_tests)
    {
        test_database = 1;
        printf( "\n\tDatabase checks have been enabled\n" );
    }
    else
    {
        printf("\n\t***** Database checks not enabled *****\n");
        return EXIT_SUCCESS;
    }

    if (argc > 1)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    else
    {
        return run_test_suite(suite, create_text_reporter());
    }
}

/* Test the create_l1r runs correctly with the summary stats enabled */
Ensure the_correct_summary_statistics_through_driver()
{
    int status;                         /* Status placeholder */
    struct stat buf;                    /* Stats buffer struct */
    static char *unit_test_dir = NULL;  /* Unit test directory */
    char command[PATH_MAX];             /* Command buffer */
    char l1r_output_name[] = "summary_statistics.h5"; /* L1R file */

    /* Get the IAS_UNIT_TEST_DATA_DIR environment variable */
    unit_test_dir = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (unit_test_dir == NULL)
    {
        printf("The IAS_UNIT_TEST_DATA_DIR environment variable is not set\n");
        exit(EXIT_FAILURE);
    }

    /* Prepare the sed command */
    status = snprintf(command, PATH_MAX,
            "sed 's|$IAS_UNIT_TEST_DATA_DIR|%s|' template_I16000.omf > "
            "I16000.omf", unit_test_dir);
    if (status < 0 || PATH_MAX <= status)
    {
        assert_true_with_message(0, "Error preparing system command");
        return;
    }

    /* Run the sed command */
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Error preparing OMF file for unit "
                "test data directory path setup");
        return;
    }

    /* Make sure the L1R output file is not present */
    status = stat(l1r_output_name, &buf);
    if (status == 0)
    {
        /* Remove the file */
        sprintf(command, "rm -f %s", l1r_output_name);
        status = system(command);
        if ( status )
        {
            assert_true_with_message(0,
                "Error removing file: %s", l1r_output_name);
        }
        else
            printf("Removed file: %s\n", l1r_output_name);
    }

    /* Create the command for running create_l1r */
    if (test_database)
    {
        status = snprintf(command, PATH_MAX, "export IAS_LOG_LEVEL=WARN; "
            "nice --adjustment=20 ../create_l1r scene_summary_statistics.odl");
    }
    else
    {
        status = snprintf(command, PATH_MAX, "export IAS_LOG_LEVEL=WARN; "
            "nice --adjustment=20 ../create_l1r "
            "scene_summary_statistics_no_db.odl");
    }
    if (status < 0 || PATH_MAX <= status)
    {
        assert_true_with_message(0, "Error preparing system command for "
            "executing create_l1r");
        return;
    }
    
    /* Execute create_l1r */
    printf("Executing: %s\n", command);
    status = system(command);
    printf("Done: %s\n", command);
    assert_equal(status, 0);
    if (status)
    {
        /* Clean up and return */
        printf("Error executing: %s\n", command);
        sprintf(command, "nice --adjustment=20 rm -f %s", l1r_output_name);
        status = system(command);
        if (status)
        {
            assert_true_with_message(0,
                "Error executing system command: %s", command);
        }
        return;
    }

    /* Clean up the L1R output file */
    sprintf(command, "nice --adjustment=20 rm -f %s", l1r_output_name);
    status = system(command);
    if (status)
    {
        assert_true_with_message(0,
        "Error executing system command: %s", command);
    }

    /* Check the results from database reading */
    if (test_database)
    {
        /* Check detector response information */
        status = db_check_statistics_in_database();
        if (status != SUCCESS)
        {
            assert_true_with_message(0, "check scene statistics "
                "data in the database");
        }
    }
}

/* Opens up database connection to be used with testing */
struct ias_db_connection *db_open(void)
{
    char *database_name = NULL; /* Database name */
    char *user_name = NULL;     /* Username for database */
    char *password = NULL;      /* Password for database */
    char *host = NULL;          /* Host name for database */
    struct ias_db_connection *db = NULL;    /* Database pointer */
    int status;                 /* Status placeholder */

    /* Initialize the satellite attributes library */
    status = ias_sat_attr_initialize(IAS_L8);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error initializing satellite"
            "library");
        return NULL;
    }

    /* Initialize the db library */
    status = ias_db_initialize_database_lib();
    if (status)
    {
        assert_true_with_message( 0, "Error from "
            "ias_db_initialize_database_lib");
        return NULL;
    }

    /* Get database connection info */
    status = ias_db_get_connect_info(&database_name, &user_name,
        &password, &host);
    if (status)
    {
        assert_true_with_message( 0, "Error from "
            "ias_db_get_connect_info");
        ias_db_close_database_lib();
        return NULL;
    }

    /* Connect to the database */
    db = ias_db_connect_to_database(database_name, user_name, password,
        host);
    
    /* Free database support info before checking return */
    free(database_name);
    free(user_name);
    free(password);
    free(host);

    /* Check if connected to the database */
    if (!db)
    {
        assert_true_with_message( 0, "Error from "
            "ias_db_connect_to_database");
        ias_db_close_database_lib();
        return NULL;
    }

    return db;
}

/* Closes the database connection and database library */
void db_close(struct ias_db_connection *db)
{
    ias_db_close_connection(db);
    ias_db_close_database_lib();
}

/* This test confirms that data have been correctly written to the
   HIST_SCENE_SUMMARY_STATS table.  */
int db_check_statistics_in_database()
{
    char query_string[1000];                /* Query Buffer */
    char db_error_message[256];             /* Error Buffer */
    char rad_wo_common_id[IAS_CHAR_ID_SIZE];/* WO ID */
    char work_order_id[] = "I16000";        /* Work order to run */
    struct ias_db_connection *db;           /* Database pointer */
    struct ias_db_query *query = NULL;      /* Query data */
    int status;                             /* Status placeholder */
    int band_id;                            /* Band id from database */
    int sca_id;                             /* SCA id from database */
    double scene_mean;                      /* Scene mean from database */
    double scene_stdev;                     /* Scene stdev from database */
    double scene_number_frames;             /* Scene frames from database */
    int lineset;                            /* Scene lineset from database */
    int sca;                                /* Loop variable for scas */
    int band_index;                         /* Loop variable for bands */
    int band;                               /* Hold current band */
    int index = 0;                          /* Generic loop variable */
    double means[] = {723.547729, 627.473816, 636.519592, 581.406555,650.947144,
        686.994629, 845.26239, 750.981262, 874.208313, 883.315186, 863.526062, 
        853.499084, 682.740723, 840.82959, 351.964142, 847.37677, 733.075317, 
        727.355103, 684.131287, 741.183167, 794.508362, 960.1745, 898.738892, 
        1027.51257, 1021.35632, 993.532166, 1013.45844, 839.188416, 969.01355, 
        460.046753, 781.503662, 686.163086, 588.414612, 616.842041, 617.437988, 
        762.876587, 850.715088, 843.671631, 943.579834, 959.540649, 866.563782, 
        971.651794, 757.468445, 990.040894, 486.085022, 774.565735, 696.420837, 
        550.232605, 578.076294, 597.523926, 755.765137, 824.487793, 823.741699, 
        899.599121, 977.350586,  873.99707, 967.38678, 750.052917, 974.260071, 
        575.025696, 2156.73706, 1883.03162, 1489.76514, 1669.51282, 1629.33765, 
        1998.82349, 2215.83521, 2355.61475, 2394.97729, 2681.70947, 2429.93359, 
        2626.94043, 2250.72534, 2560.19043, 2007.92029, 1107.1178, 748.42395, 
        719.316467, 884.408813, 774.181885, 1106.24854, 1034.61487, 1185.35339, 
        1209.90417, 1476.48474, 1296.84558, 1470.10559, 1094.32141, 1414.35291, 
        1085.08752, 1038.04626, 770.634766, 654.493408, 776.938293, 715.650024, 
        1048.62427, 1000.15173, 1068.79565, 1161.43152, 1376.03113, 1270.97058, 
        1361.52148, 1000.23694, 1336.01355, 991.153809, 1030.70422, 896.928467, 
        756.011047, 798.649353, 861.628906,  939.837158, 1107.37854, 1164.14075,
        1206.5083,  1302.13562, 1198.03259, 1287.80176, 1127.23645, 1255.75745, 
        527.812317, 1031.24634, 897.394653, 756.58606, 799.261108, 862.140869,
        940.416992, 1107.94006, 1164.64758, 1207.0675, 1302.66382,  1198.50073, 
        1288.37903, 1127.7926, 1256.37939, 528.277283, 681.114075, 391.6362, 
        444.759827, 540.599121, 480.256561, 704.789307, 691.149109, 775.328552, 
        740.24176, 937.230713, 796.636169, 936.705811, 593.078247, 856.736877, 
        646.448669};
    double stdevs[] = {430.627747, 236.998611, 241.80307, 252.86116, 250.279633,
        309.27774, 469.081238, 495.1922, 450.654419, 394.510986, 363.513367, 
        351.682037, 332.336639, 368.39328, 797.463867, 497.330841, 279.620087, 
        285.269928, 295.552856, 297.058838, 367.477936, 558.662659, 586.403625, 
        534.229187, 467.233093, 428.889709, 424.850372, 399.647339, 432.141998, 
        872.833313, 618.084839, 492.663544, 470.879822,  523.015503, 495.384186,
        563.720398, 684.080627, 712.955139, 646.241577, 590.76416, 533.450806, 
        556.6875, 541.854736, 592.265991, 850.510742, 716.622131, 642.11792, 
        613.797668, 684.972778, 658.183105, 711.054321, 750.14679, 782.600586, 
        715.013184, 698.525024, 631.270569, 665.869812, 655.869385, 704.084656, 
        857.142578, 2047.57251, 1902.96814, 1858.83899, 2036.28418, 1978.77222, 
        2066.50708, 2147.94727, 2161.53101, 2048.85229, 2056.12915, 1851.41663, 
        1966.48975, 1953.50269, 2011.07422, 2120.10229, 1213.62659, 1313.36841, 
        1082.43506, 1230.32422, 1133.57898, 1283.39148, 1154.47314, 1177.12439, 
        1201.63086, 1229.86353, 1123.55078, 1177.6145, 1148.82202, 1179.14319, 
        1176.07898, 1107.5105, 1161.75793, 995.65979, 1112.53723, 1037.27209, 
        1165.88367, 1074.90417, 1075.61914, 1089.96924, 1109.34729, 1073.40454, 
        1064.34448, 1042.45374, 1076.005, 1059.98181, 969.765076, 834.333435, 
        825.616516, 892.080994, 905.039185, 913.907471, 998.346558, 1051.78381, 
        957.019409, 935.303162, 838.721802, 892.859131, 893.59259, 933.03363,  
        1247.28589, 969.812683, 834.383667, 825.669739,892.111816, 905.101624,
        913.917114, 998.357239, 1051.78357, 957.034363,935.317078,838.72168,
        892.854431, 893.624329, 933.069275,1247.55542, 705.813171, 852.706177, 
        606.231079, 715.568054, 651.9151, 739.424988, 666.40271, 650.135376, 
        694.888, 700.109985, 645.18103, 665.293884, 648.425537, 654.259094, 
        672.296997};
    int expected_bands[] = {1, 2, 3, 4, 5, 6, 7, 8, 8, 9};
    /* Open the database connection */
    db = db_open();
    if (!db)
    {
        assert_true_with_message(0, "Cannot connect to database");
        return ERROR;
    }

    /* Get the radiometric work order ID key for the specified work order
       string */
    status = ias_misc_get_char_id(db,
        "RADIOMETRIC_WORK_ORDER_COMMON", "radiometric_wo_common_id",
        work_order_id, rad_wo_common_id);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Invalid rad_wo_common_id key "
               "value for specified work order ID %s", work_order_id);
        db_close(db);
        return ERROR;
    }

    /* Get values from the HIST_SCENE_SUMMARY_STATS table */
    status = snprintf(query_string, sizeof(query_string),
        "SELECT band_id, "
        "       sensor_chip_assembly_id, "
        "       scene_mean, "
        "       scene_stdev, "
        "       scene_number_frames, "
        "       lineset "
        "FROM   HIST_SCENE_SUMMARY_STATS "
        "WHERE  radiometric_wo_common_id=%s "
        "order by band_id, lineset, sensor_chip_assembly_id",
        rad_wo_common_id);
    if ((status < 0) || (status >= sizeof(query_string)))
    {
        assert_true_with_message(0, "Error constructing "
            "HIST_SCENE_SUMMARY_STATS SQL string");
        db_close(db);
        return ERROR;
    }

    /* Execute the query and verify that it worked */
    query = ias_db_perform_query(db, query_string);
    if (!ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, db_error_message,
            sizeof(db_error_message));
        assert_true_with_message(0, "Error from select query:  %s",
            db_error_message);
        ias_db_query_close(query);
        db_close(db);
        return ERROR;
    }
    
    for (band_index = 0; band_index <= 9; band_index++)
    {
        band = expected_bands[band_index];
        for (sca = 0; sca <= 14; sca++, index++)
        {
            double current_stdev = stdevs[index];
            double current_mean = means[index];
            /* Move DB "pointer" to the first valid record. */
            if (!ias_db_query_next(query))
            {
                assert_true_with_message(0, "Error reading HIST_SCENE_"
                    "SUMMARY_STATS table");
                ias_db_query_close(query);
                db_close(db);
                return ERROR;
            }

            /* Start checking values from the query against the corresponding
            expected results */
            status = ias_db_query_int_value(query, &band_id, 0);
            if (status != SUCCESS)
            {
                assert_true_with_message(0, "Error reading 'band_id' from "
                "HIST_SCENE_SUMMARY_STATS table query");
                ias_db_query_close(query);
                db_close(db);
                return ERROR;
            }
            assert_equal(band_id, band);

            /* Check that the SCA is matches the expected value */
            status = ias_db_query_int_value(query, &sca_id, 1);
            if (status != SUCCESS)
            {
                assert_true_with_message(0, "Error reading 'SCA id' from "
                    "HIST_SCENE_SUMMARY_STATS table query");
                ias_db_query_close(query);
                db_close(db);
                return ERROR;
            }
            assert_equal(sca_id, sca);

            /* Check that the scene mean matches the expected value */
            status = ias_db_query_double_value(query, &scene_mean, 2);
            if (status != SUCCESS)
            {
                assert_true_with_message(0, "Error reading 'scene_mean' "
                    "from HIST_SCENE_SUMMARY_STATS table query");
                ias_db_query_close(query);
                db_close(db);
                return ERROR;
            }
            assert_double_equal(scene_mean, current_mean);

            /* Check that the stdev matches the expected value */
            status = ias_db_query_double_value(query, &scene_stdev, 3);
            if (status != SUCCESS)
            {
                assert_true_with_message(0, "Error reading 'scene_stdev' "
                    "from HIST_SCENE_SUMMARY_STATS table query");
                ias_db_query_close(query);
                db_close(db);
                return ERROR;
            }
            assert_double_equal(scene_stdev, current_stdev);

            /* Check that the frames matches the expected value */
            status = ias_db_query_double_value(query, &scene_number_frames, 4);
            if (status != SUCCESS)
            {
                assert_true_with_message(0, "Error reading 'scene_number_"
                    "frames' from from HIST_SCENE_SUMMARY_STATS table query");
                ias_db_query_close(query);
                db_close(db);
                return ERROR;
            }
            assert_double_equal(scene_number_frames, 7002.0);

            /* Check that the lineset matches the expected value */
            status = ias_db_query_int_value(query, &lineset, 5);
            if (status != SUCCESS)
            {
                assert_true_with_message(0, "Error reading 'lineset' from "
                    "HIST_SCENE_SUMMARY_STATS table query");
                ias_db_query_close(query);
                db_close(db);
                return ERROR;
            }
    
            if (band == 8 && band_index == 7)
            {
                assert_equal(lineset, 0);
            }
            else if(band == 8 && band_index == 8)
            {
                assert_equal(lineset, 1);
            }
            else
            {   
                assert_equal(lineset, 2);
            }
        }
    }
    /* Release query resources */
    ias_db_query_close(query);
    db_close(db);

    return SUCCESS;
}

