/*----------------------------------------------------------------------------
NAME:    striping_characterization_test.c

PURPOSE: Test the striping characterization routines.

USAGE:   response_linearization_test [TESTNAME]

         Run the single test TESTNAME or all tests if TESTNAME isn't
         provided.
-----------------------------------------------------------------------------*/
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <math.h>

#include "ias_pixel_mask.h"
#include "ias_logging.h"
#include "cgreen/cgreen.h"
#include "../rps_local.h"

/* The test image size is */
#define DETECTORS 320
#define LINES     6584

/* This contains the expected results */
#include "detector_striping_metric.h"

/* Path to test data and file names */
static char *test_data_path = NULL;
static char data_dir[] = "ias_base/rps/driver/striping_characterization";
static char image1[] = "brookings_b1s1.bin";
static char exp_scm[] = "binary_striping_characterization_scm_01_01.rpt";
static char exp_ssm[] = "binary_striping_characterization_ssm_01_01.rpt";
static char exp_dsm[] = "striping_characterization_dsm_01_01.rpt";
static char exp_osm[] = "striping_characterization_osm.rpt";

static int test_database = 0;

Ensure striping_characterization_test( void )
{
    float *image_buffer;
    int band_number = 1;
    int sca_number = 1;
    int number_detectors = DETECTORS;
    int number_lines = LINES;
    int image_size;
    char sys_command[PATH_MAX];
    /* These offsets are derived from the CPF parameter
       Along_Detector_Offsets_B02_SCA01.  That group was chosen because it
       has a negative offset as well as a +4 offset.  That gives an offset
       range of 6. */
    int detector_offsets[] = {
        2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4, 
        2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4, 
        2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4, 
        2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4, 
        2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4, 
        2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4, 
        2,  4,  2,  4,  2,  4,  0,  4,  2,  4,  2,  4,  2,  4,  2,  4, 
        2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4, 
        2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4, 
        2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4, 
        2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4, 
        2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4, 
        2,  6,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4, 
        2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4, 
        2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4, 
        2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4, 
        2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4, 
        2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4, 
        2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4, 
        2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4,  2,  4 }; 
    int offset_range = 6;
    IAS_PIXEL_MASK *pixel_mask;
    int inop_detector_index = 257; /* 258 1-based */
    float striping_metric_cutoff = 4.27;
    int dump_matrix_flag = 1;
    double *detector_striping_metric;
    double *striping_correction_matrix;
    short *inbuf; /* Buffer for reading input data */
    FILE *inf;
    int i;
    char filename[PATH_MAX];
    int status;
    double osm_radiance; /* Overall Striping Metric values in radiance */
    double osm_dn;       /*   and in DN                                */
    IAS_REPORT report_info;
    IAS_REPORT_HEADER *report_header;

    /* Set up a report header */
    report_header = ias_misc_initialize_report_header("", "Dummy",
        "striping_characterization_test" );
    if ( !report_header )
    {
        assert_true_with_message(0, "Error initializing report header");
        return;
    }
    ias_misc_set_header_pathrow(report_header, 0, 0);
    report_info.header = report_header;
    report_info.directory = NULL;


    /* Allocate buffers for the input data */
    image_size = number_detectors * number_lines;
    inbuf = malloc( sizeof(short) * image_size );
    image_buffer = malloc( sizeof(float) * image_size );
    if ( !inbuf || !image_buffer )
    {
        free( inbuf );
        free( image_buffer );
        assert_true_with_message( 0, "Error allocating memory for image");
        return;
    }

    /* Open the image file */
    sprintf( filename, "%s/%s/%s", test_data_path, data_dir, image1 );
    inf = fopen( filename, "r" );
    if ( !inf )
    {
        free( inbuf );
        free( image_buffer );
        assert_true_with_message( 0, "Error opening: %s", filename );
        return;
    }

    /* Read the data and convert it to float */
    i = fread( inbuf, sizeof(short), image_size, inf );
    if ( i != image_size )
    {
        free( inbuf );
        free( image_buffer );
        assert_true_with_message( 0, "Error reading: %s, read %d of %d",
            filename, i, image_size );
        return;
    }
    for ( i = 0; i < image_size; i++ )
        image_buffer[i] = (float)inbuf[i];
    free( inbuf );
    status = fclose( inf );
    assert_false_with_message( status, "Error closing: %s", filename );

    /* Set up the pixel mask */
    pixel_mask = ias_pm_create( band_number, sca_number, number_detectors,
        number_lines );
    if ( !pixel_mask )
    {
        free( image_buffer );
        assert_true_with_message( 0, "Error creating pixel mask");
        return;
    }

    /* The prototype only has one artifact the inop detector */
    status = ias_pm_add_pixels( pixel_mask, inop_detector_index, 0,
        number_lines, IAS_RPS_INOP_DETECTOR );
    if ( status != SUCCESS )
    {
        free( image_buffer );
        ias_pm_destroy( pixel_mask );
        assert_true_with_message( 0, "Error adding dead detector to pixel "
            "mask");
        return;
    }

    /* Allocate the striping correction matrix and detector striping metric */
    striping_correction_matrix = malloc( sizeof(double) * image_size );
    detector_striping_metric = malloc( sizeof(double) * number_detectors );
    if ( !striping_correction_matrix || !detector_striping_metric )
    {
        free( detector_striping_metric );
        free( striping_correction_matrix );
        free( image_buffer );
        ias_pm_destroy( pixel_mask );
        assert_true_with_message( 0, "Error allocating the correction matrix");
        return;
    }

    /* Call the algorithm */
    status = striping_characterization( image_buffer, band_number,
        sca_number, number_detectors, number_lines, detector_offsets,
        offset_range, pixel_mask, striping_metric_cutoff, dump_matrix_flag,
        &report_info, detector_striping_metric, striping_correction_matrix );
    if ( status != SUCCESS )
    {
        free( image_buffer );
        ias_pm_destroy( pixel_mask );
        free( striping_correction_matrix );
        free( detector_striping_metric );
        assert_true_with_message( 0, "Error returned from striping_"
            "characterization");
        return;
    }
    /* No longer need these */
    free( image_buffer );
    ias_pm_destroy( pixel_mask );
    ias_misc_free_report_header( report_header );

    /* Compare the detector striping metric with the expected results */
    for ( i = 0; i < number_detectors; i++ )
    {
        assert_double_equal( detector_striping_metric[i],
            expected_dsm_results[i] );
    }

    /* The striping_correction_matrix and the scene striping metric are both
       full number detectors by number line double values.  Both are stored
       as flat files. */
    status = snprintf( sys_command, PATH_MAX, "diff -q %s/%s/%s .",
        test_data_path, data_dir, exp_scm );
    if ( status < 0 || status >= PATH_MAX )
    {
        assert_true_with_message( 0, "Error creating system command to "
            "compare %s, comparison skipped", exp_scm );
    }
    else
    {
        status = system( sys_command );
        assert_equal( status, 0 );
    }
    status = snprintf( sys_command, PATH_MAX, "diff -q %s/%s/%s .",
        test_data_path, data_dir, exp_ssm );
    if ( status < 0 || status >= PATH_MAX )
    {
        assert_true_with_message( 0, "Error creating system command to "
            "compare %s, comparison skipped", exp_ssm );
    }
    else
    {
        status = system( sys_command );
        assert_equal( status, 0 );
    }

    /* Check the report */
    status = snprintf( sys_command, PATH_MAX, "sed '/^Detector/,$!'d %s | "
        "diff -q - %s/%s/%s", exp_dsm, test_data_path, data_dir, exp_dsm );
    if ( status < 0 || status >= PATH_MAX )
    {
        assert_true_with_message( 0, "Error creating system command to "
            "compare %s, comparison skipped", exp_dsm );
    }
    else
    {
        status = system( sys_command );
        assert_equal( status, 0 );
    }


    /* Get the overall striping metric, the first and last values in the DSM
       are zero and aren't taken into account in the prototype algorithm so
       adjust the start of the DSM and number of detectors.  Also pass a
       nominal band 1 absolute gain. */
    status = calculate_overall_striping_metric( &detector_striping_metric[1],
        number_detectors - 2, 15.6012, &osm_radiance, &osm_dn );
    if ( status != SUCCESS )
    {
        free( striping_correction_matrix );
        free( detector_striping_metric );
        assert_true_with_message( 0, "Error returned from "
            "calculate_detector_striping_metric");
        return;
    }
    assert_double_equal( osm_radiance, 0.03567765 );
    assert_double_equal( osm_dn, 0.55661422 );

    free( striping_correction_matrix );
    free( detector_striping_metric );

    /* Check the striping correction matrix and scene striping metric */
    status = snprintf( sys_command, PATH_MAX, "diff -q %s/%s/binary_"
        "striping_characterization_ssm_01_01.rpt .", test_data_path,
        data_dir );
    if ( status < 0 || status >= PATH_MAX )
    {
        assert_true_with_message( 0, "Error creating system command to "
            "compare binary_striping_characterization_ssm_01_01.rpt, "
            "comparison skipped" );
    }
    else
    {
        status = system( sys_command );
        assert_equal( status, 0 );
    }

    status = snprintf( sys_command, PATH_MAX, "diff -q %s/%s/binary_"
        "striping_characterization_scm_01_01.rpt .", test_data_path,
        data_dir );
    if ( status < 0 || status >= PATH_MAX )
    {
        assert_true_with_message( 0, "Error creating system command to "
            "compare binary_striping_characterization_scm_01_01.rpt, "
            "comparison skipped" );
    }
    else
    {
        status = system( sys_command );
        assert_equal( status, 0 );
    }
}  /* END striping_characterization_test */


Ensure api_striping_characterization_test( void )
{
    int status;
    int inop_detector_index = 257; /* 258 1-based */
    double *detector_striping_metric;
    double *scene_correction_matrix;
    RPS_BAND_ATTRIBUTES *band_attributes;
    BLOB *blob;
    char l0rname[] = "ias_base/rps/driver/LC81800382002045SGS00/"
        "LC81800382002045SGS00";
    char cpfname[] = "ias_base/rps/driver/LC81800382002045SGS00/"
        "L8CPF20020101_20020331.01";
    uint16_t *image_l0r;
    float *image_l1r;
    double *double_ptr;
    int i;
    int off[3] = { 2, 0, 2 };
    int line;
    int non_zero_count;
    int buffer_length;
    int band_index = 1;

    /* The attributes must be initialized for memory management to work */
    ias_sat_attr_initialize( IAS_L8 );

    /* Allocate a blob and populate the minimum members for the test
       to succeed */
    blob = allocate_blob();
    sprintf( blob->l0r_file_name, "%s/%s", test_data_path, l0rname );
    sprintf( blob->cpf_filename, "%s/%s", test_data_path, cpfname );
    strcpy( blob->l1r_file_name, "dummy_striping.h5" );
    blob->max_number_scas = 14;
    blob->total_bands = 30;
    blob->number_of_bands = 1;
    blob->generate_report = 1;
    status = finish_blob_initialization();
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error initializing the blob");
        free_the_blob();
        return;
    }
    /* We want the report files in the current directory so free the
       directory in the blob and set it to NULL */
    free( blob->report_information.directory );
    blob->report_information.directory = NULL;
    band_attributes = get_band_attributes( band_index );
    blob->list_of_bands[0] = band_attributes->band_number;
    blob->list_of_band_indices[0] = band_index;

    /* Use a level 0 image to initialize attributes structure and get access
       to the imagery */
    ias_log_set_output_level( IAS_LOG_LEVEL_DISABLE );
    status = initialize_io_routines( blob );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error initializing I/O routines");
        free_the_blob();
        return;
    }
    ias_log_set_output_level( IAS_LOG_LEVEL_INFO );

    /* Create a pixel mask and set some artifacts */
    status = create_blob_pixel_mask();
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error creating pixel mask");
        cleanup_io_routines( blob );
        free_the_blob();
        return;
    }

    status = ias_pm_add_pixels( blob->pixel_mask[band_index][0],
        inop_detector_index, 0, band_attributes->number_lines,
        IAS_RPS_INOP_DETECTOR );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error adding dead detector to pixel "
            "mask");
        cleanup_io_routines( blob );
        free_the_blob();
        return;
    }

    status = ias_pm_add_pixels( blob->pixel_mask[band_index][0],
        320, 223, 3, IAS_RPS_ADCONV_SAT_HI );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error adding saturation to pixel "
            "mask");
        cleanup_io_routines( blob );
        free_the_blob();
        return;
    }

    status = ias_pm_add_pixels( blob->pixel_mask[band_index][0],
        20, 254, 1, IAS_RPS_ADCONV_SAT_HI );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error adding saturation to pixel "
            "mask");
        cleanup_io_routines( blob );
        free_the_blob();
        return;
    }

    /* Add artifacts along the aligned line 203.  The purpose of this is
       to ensure that detector offsets are properly handled in the
       algorithm. */
    for ( i = 0; i < 3; i++ )
    {
        status = ias_pm_add_pixels( blob->pixel_mask[band_index][0],
            i + 202, 203 + off[i], 1, IAS_RPS_ADCONV_SAT_HI );
        if ( status != SUCCESS )
        {
            assert_true_with_message( 0, "Error adding saturation to pixel "
                "mask");
            cleanup_io_routines( blob );
            free_the_blob();
            return;
        }
    }

    /* Finish up initialization */
    status = initialize_memory_management_buffers();
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error initializing memory management");
        cleanup_io_routines( blob );
        free_the_blob();
        return;
    }

    /* Register the detector striping metric and scene striping correction
       matrix so that they will be available after the API call */
    status = increment_consumer_count( MM_DETECTOR_STRIPING_METRIC );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error incrementing the detector "
            "striping metric");
        cleanup_io_routines( blob );
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }
    status = increment_consumer_count( MM_SCENE_STRIPING_CORRECTION_MATRIX );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error incrementing the scene striping "
            "correction matrix");
        cleanup_io_routines( blob );
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Call pre-process to register the algorithms memory needs */
    status = striping_characterization_preprocess( blob );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error returned from pre-process");
        cleanup_io_routines( blob );
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Since we are bypassing the normal algorithm flow, register and
       get an SCA of the level 0 image.  Also register the scene striping
       correction matrix so that it can be checked after the call to the
       API. */
    status = increment_consumer_count( IMAGE_L0R );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error incrementing Level 0r buffer");
        cleanup_io_routines( blob );
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }
    status = increment_consumer_count( MM_SCENE_STRIPING_CORRECTION_MATRIX );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error incrementing scene striping "
            "correction matrix");
        cleanup_io_routines( blob );
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    image_l0r = (uint16_t *)preprocess_get_memory_buffer( IMAGE_L0R, 1, 0 );
    if ( !image_l0r )
    {
        assert_true_with_message( 0, "Error getting L0R image");
        cleanup_io_routines( blob );
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* The level 1 buffer was already registered by pre-process so get it */
    image_l1r = (float *)preprocess_get_memory_buffer( IMAGE_L1R,
        blob->list_of_band_indices[0], 0 );
    if ( !image_l1r )
    {
        assert_true_with_message( 0, "Error getting L1R image");
        cleanup_io_routines( blob );
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* It is important that the consumer counts get set *after* all the
       increments */
    status = set_consumer_counts();
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error setting consumer counts");
        cleanup_io_routines( blob );
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Get the SCA attributes and convert the level 0 to level 1 */
    band_attributes = get_band_attributes( blob->list_of_band_indices[0] );
    buffer_length = band_attributes->number_detectors *
        band_attributes->number_lines;
    for ( i = 0; i < buffer_length; i++ )
        image_l1r[i] = (float)image_l0r[i];

    /* Release the level 0 buffer */
    status = decrement_consumer_count( IMAGE_L0R,
        blob->list_of_band_indices[0], 0 );
    assert_equal_with_message( status, SUCCESS,
        "Error returned from decrement consumer count");

    /* Call the API */
    status = api_striping_characterization( blob,
        blob->list_of_band_indices[0], 0, 0 );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error returned from API call");
        cleanup_io_routines( blob );
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Get the detector striping metric and compare it with the expected
       results */
    detector_striping_metric = (double *)get_memory_buffer(
        MM_DETECTOR_STRIPING_METRIC, CONSUMER, blob->list_of_band_indices[0],
        0 );
    if ( !detector_striping_metric )
    {
        assert_true_with_message( 0, "Error getting detector striping metric");
        cleanup_io_routines( blob );
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    for ( i = 0; i < DETECTORS; i++ )
    {
        double diff = fabs( detector_striping_metric[i] -
            api_expected_dsm_results[i] );
        assert_true_with_message ( diff < 1e-5, "%f should match %f",
            detector_striping_metric[i], api_expected_dsm_results[i] );
    }

    /* Get the correction matrix */
    scene_correction_matrix = (double *)get_memory_buffer(
        MM_SCENE_STRIPING_CORRECTION_MATRIX, CONSUMER,
        blob->list_of_band_indices[0], 0 );
    if ( !scene_correction_matrix )
    {
        assert_true_with_message( 0, "Error getting scene striping "
            "correction matrix");
        cleanup_io_routines( blob );
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Check that artifacts were properly excluded - the inoperable detector
       and its two neighbors should be zero */
    non_zero_count = 0;
    double_ptr = &scene_correction_matrix[inop_detector_index - 1];
    for ( line = 0; line < band_attributes->number_lines; line++ )
    {
        if ( double_ptr[0] != 0.0 )
            non_zero_count++;
        if ( double_ptr[1] != 0.0 )
            non_zero_count++;
        if ( double_ptr[2] != 0.0 )
            non_zero_count++;
        double_ptr += band_attributes->number_detectors;
    }
    assert_equal_with_message( non_zero_count, 0, "There were %d non-zero "
        "values for the inoperable detector", non_zero_count );

    /* For artifacts other than inoperable detectors the detector offsets
       have to be taken into account.  The following three sets of artifacts
       have considered the offsets.  I'm showing more detail on the first
       set since it is the more complicated.  Handling the offsets in the
       other two cases is more simple.

       There are three artifacts on aligned line 203, detectors 202, 203,
       and 204, the offsets need to be taken into account.  The values look
       like the following in the correction matrix without the artifacts (the
       values in parens are the ones that are affected):

       2.291998 -0.270036 -0.739499  1.550454 -0.120821 -0.212684 -0.016243
       0.734873 -0.013720  0.688205  0.340404  0.305963 -0.846012 -0.183378
       0.079352 -0.003581 -0.078831  1.220777  0.117957 -2.906980 -0.193681
      -0.000917 -0.005533 -0.078140 -0.362536  0.147806 -1.916386 -0.032071
       0.000418 -0.002426 -0.440873  0.142230  0.282978 -1.939887 -0.028084
       0.001037 -0.000646 -0.041622  0.008064  0.327814 -0.219735 -0.077034
       0.001813  0.000049 -0.038344 -0.000846 -0.095116  0.000503 -0.076134

       These are the resulting values:

       2.291998 -0.270036 -0.739499  1.550454 -0.120821 -0.212684 -0.016243
       0.734873 -0.013720  0.688205 (0.000000) 0.305963 -0.846012 -0.183378
       0.079352 (0.000000)-0.078831 (0.000000) 0.117957 (0.000000)-0.193681
      -0.000917 -0.005533 (0.000000)(0.000000)(0.000000)-1.916386 -0.032071
       0.000418 -0.002426 (0.000000) 0.142230 (0.000000)-1.939887 -0.028084
       0.001037 -0.000646 (0.000000) 0.008064 (0.000000)-0.219735 -0.077034
       0.001813  0.000049 -0.038344 -0.000846 -0.095116  0.000503 -0.076134

       The offsets are 0 2 0 2 0 */
    double_ptr = scene_correction_matrix + 201 + (
        band_attributes->number_detectors * ( 201 ));
    assert_double_equal( double_ptr[2], 0.0 );
    double_ptr += band_attributes->number_detectors;
    assert_double_equal( double_ptr[0], 0.0 );
    assert_double_equal( double_ptr[2], 0.0 );
    assert_double_equal( double_ptr[4], 0.0 );
    double_ptr += band_attributes->number_detectors;
    assert_double_equal( double_ptr[1], 0.0 );
    assert_double_equal( double_ptr[2], 0.0 );
    assert_double_equal( double_ptr[3], 0.0 );
    double_ptr += band_attributes->number_detectors;
    assert_double_equal( double_ptr[1], 0.0 );
    assert_double_equal( double_ptr[3], 0.0 );
    double_ptr += band_attributes->number_detectors;
    assert_double_equal( double_ptr[1], 0.0 );
    assert_double_equal( double_ptr[3], 0.0 );
     
    /* There is a single artifact at detector 20 line 254, offset 2.
       Offsets are 0 for the neighbor detectors. */
    double_ptr = scene_correction_matrix + 19 + (
        band_attributes->number_detectors * 251 );
    assert_double_equal( double_ptr[0], 0.0 );
    assert_double_equal( double_ptr[2], 0.0 );
    double_ptr += band_attributes->number_detectors;
    assert_double_equal( double_ptr[1], 0.0 );
    double_ptr += band_attributes->number_detectors;
    assert_double_equal( double_ptr[1], 0.0 );
    double_ptr += band_attributes->number_detectors;
    assert_double_equal( double_ptr[1], 0.0 );

    /* There is a span of three artifacts for detector 320 starting on line
       223.  Offsets are 0 2 0. */
    double_ptr = scene_correction_matrix + 319 + (
        band_attributes->number_detectors * 220 );
    assert_double_equal( double_ptr[0], 0.0 );
    assert_double_equal( double_ptr[2], 0.0 );
    double_ptr += band_attributes->number_detectors;
    assert_double_equal( double_ptr[0], 0.0 );
    assert_double_equal( double_ptr[1], 0.0 );
    assert_double_equal( double_ptr[2], 0.0 );
    double_ptr += band_attributes->number_detectors;
    assert_double_equal( double_ptr[0], 0.0 );
    assert_double_equal( double_ptr[1], 0.0 );
    assert_double_equal( double_ptr[2], 0.0 );
    double_ptr += band_attributes->number_detectors;
    assert_double_equal( double_ptr[1], 0.0 );
    double_ptr += band_attributes->number_detectors;
    assert_double_equal( double_ptr[1], 0.0 );
    double_ptr += band_attributes->number_detectors;
    assert_double_equal( double_ptr[1], 0.0 );


    /* Clean up */
    status = decrement_consumer_count( MM_SCENE_STRIPING_CORRECTION_MATRIX,
        blob->list_of_band_indices[0], 0 );
    assert_equal_with_message( status, SUCCESS,
        "Error returned from decrement consumer count");
    status = decrement_consumer_count( MM_DETECTOR_STRIPING_METRIC,
        blob->list_of_band_indices[0], 0 );
    assert_equal_with_message( status, SUCCESS,
        "Error returned from decrement consumer count");
    for ( i = 0; i < band_attributes->number_scas; i++ )
    {
        int vrp_index = band_attributes->vrp_band_index;
        status = decrement_consumer_count( MM_MAX_SPATIAL_DETECTOR_OFFSETS,
            blob->list_of_band_indices[0], i );
        assert_equal_with_message( status, SUCCESS,
            "Error returned from decrement consumer count");
        status = decrement_consumer_count( MM_MAX_SPATIAL_DETECTOR_OFFSETS,
            vrp_index, i );
        assert_equal_with_message( status, SUCCESS,
            "Error returned from decrement consumer count");
        status = decrement_consumer_count( MM_SPATIAL_DETECTOR_OFFSETS,
            blob->list_of_band_indices[0], i );
        assert_equal_with_message( status, SUCCESS,
            "Error returned from decrement consumer count");
        status = decrement_consumer_count( MM_SPATIAL_DETECTOR_OFFSETS,
            vrp_index, i );
        assert_equal_with_message( status, SUCCESS,
            "Error returned from decrement consumer count");
        status = decrement_consumer_count( MM_SCENE_STRIPING_CORRECTION_MATRIX,
            blob->list_of_band_indices[0], i );
        assert_equal_with_message( status, SUCCESS,
            "Error returned from decrement consumer count");
        status = decrement_consumer_count( MM_DETECTOR_STRIPING_METRIC,
            blob->list_of_band_indices[0], i );
        assert_equal_with_message( status, SUCCESS,
            "Error returned from decrement consumer count");
    }
    remove( blob->l1r_file_name );
    cleanup_io_routines( blob );
    cleanup_memory_management_buffers(ENABLE_WARNING_DISPLAY);
    free_the_blob();
}  /* END api_striping_characterization_test */


Ensure api_striping_characterization_postproccess_test( void )
{
    int status;
    RPS_BAND_ATTRIBUTES *band_attributes;
    BLOB *blob;
    char l0rname[] = "ias_base/rps/driver/LC81800382002045SGS00/"
        "LC81800382002045SGS00";
    char cpfname[] = "ias_base/rps/driver/LC81800382002045SGS00/"
        "L8CPF20020101_20020331.01";
    int band_index = 1;
    int sca_index;
    char sys_command[PATH_MAX];

    /* The attributes must be initialized for memory management to work */
    ias_sat_attr_initialize( IAS_L8 );

    /* Allocate a blob and populate the minimum members for the test
       to succeed */
    blob = allocate_blob();
    sprintf( blob->l0r_file_name, "%s/%s", test_data_path, l0rname );
    sprintf( blob->cpf_filename, "%s/%s", test_data_path, cpfname );
    sprintf( blob->work_order_id, "StripingTest" );
    strcpy( blob->l1r_file_name, "dummy_striping.h5" );
    blob->max_number_scas = 14;
    blob->total_bands = 30;
    blob->number_of_bands = 1;
    blob->store_in_database = 1;
    blob->generate_report = 1;
    status = finish_blob_initialization();
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error initializing the blob");
        free_the_blob();
        return;
    }
    /* We want the report files in the current directory so free the
       directory in the blob and set it to NULL */
    free( blob->report_information.directory );
    blob->report_information.directory = NULL;

    /* Use a level 0 image to initialize attributes structure and get access
       to the imagery */
    ias_log_set_output_level( IAS_LOG_LEVEL_DISABLE );
    status = initialize_io_routines( blob );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error initializing I/O routines");
        free_the_blob();
        return;
    }
    ias_log_set_output_level( IAS_LOG_LEVEL_INFO );
    band_attributes = get_band_attributes( band_index );
    blob->list_of_bands[0] = band_attributes->band_number;
    blob->list_of_band_indices[0] = band_index;

    /* Create a pixel mask */
    status = create_blob_pixel_mask();
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error creating pixel mask");
        cleanup_io_routines( blob );
        free_the_blob();
        return;
    }

    /* Get a database connection */
    if ( test_database == 0 )
    {
        blob->store_in_database = 0;
    }
    else
    {
        status = establish_db_connection( blob );
        if ( status != SUCCESS )
        {
            assert_true_with_message( 0, "Error establishing database "
                "connection");
            blob->store_in_database = 0;
        }
        else
        {
            status = ias_misc_get_char_id( blob->database_information.db,
                "radiometric_work_order_common", "radiometric_wo_common_id",
                blob->work_order_id, blob->database_information.rad_common_id );
            if ( status != SUCCESS )
            {
                assert_true_with_message( 0, "Error getting a radiometric "
                    "characterization ID");
                close_db_connection( blob, IAS_DB_ROLLBACK );
                blob->store_in_database = 0;
            }
            status = delete_db_table_records( blob->database_information,
                "striping" );
            if ( status != SUCCESS )
            {
                assert_true_with_message( 0, "Error deleting records from the "
                    "striping table");
                close_db_connection( blob, IAS_DB_ROLLBACK );
                blob->store_in_database = 0;
            }

        }
    }

    /* Finish up initialization */
    status = initialize_memory_management_buffers();
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error initializing memory management");
        cleanup_io_routines( blob );
        free_the_blob();
        return;
    }

    /* Call pre-process to register the algorithms memory needs */
    status = convert_to_float_preprocess( blob );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error returned from pre-process");
        cleanup_io_routines( blob );
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }
    status = striping_characterization_preprocess( blob );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error returned from pre-process");
        cleanup_io_routines( blob );
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }
    /* It is important that the consumer counts get set *after* all the
       increments */
    status = set_consumer_counts();
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error setting consumer counts");
        cleanup_io_routines( blob );
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Initialize the RPS level arrays */
    status = initialize_rps_level(30, 14);
    if (status != SUCCESS)
    {
        assert_true_with_message( 0, "Error initializing RPS level arrray" );
        cleanup_io_routines( blob );
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    blob->generate_report = 0; /* Don't need the reports for now */
    band_attributes = get_band_attributes( blob->list_of_band_indices[0] );
    for ( sca_index = 0; sca_index < band_attributes->number_scas;
        sca_index++ )
    {
        /* Convert the L0 buffer to floating point */
        status = api_convert_to_float( blob, blob->list_of_band_indices[0],
            sca_index, 0 );
        if ( status != SUCCESS )
        {
            assert_true_with_message( 0, "Error returned from convert to "
                "float call");
            cleanup_rps_level();
            cleanup_io_routines( blob );
            cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
            free_the_blob();
            return;
        }

        /* Call the API */
        status = api_striping_characterization( blob,
            blob->list_of_band_indices[0], sca_index, 0 );
        if ( status != SUCCESS )
        {
            assert_true_with_message( 0, "Error returned from striping "
                "characterization call");
            cleanup_rps_level();
            cleanup_io_routines( blob );
            cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
            free_the_blob();
            return;
        }
    }

    blob->generate_report = 1; /* Now we want the report */
    status = striping_characterization_postprocess( blob );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error returned from striping "
            "characterization post-process call");
    }

    /* Check the report */
    status = snprintf( sys_command, PATH_MAX, "sed '/Sensor:/,$!'d %s | "
        "diff -q - %s/%s/%s", exp_osm, test_data_path, data_dir, exp_osm );
    if ( status < 0 || status >= PATH_MAX )
    {
        assert_true_with_message( 0, "Error creating system command to "
            "compare %s, comparison skipped", exp_dsm );
    }
    else
    {
        status = system( sys_command );
        assert_equal_with_message( status, 0, sys_command );
    }

    if ( blob->store_in_database == 1 )
    {
        status = close_db_connection( blob, IAS_DB_COMMIT );
        if ( status != SUCCESS )
        {
            assert_true_with_message( 0, "Error closing database connection");
            blob->store_in_database = 0;
        }
    }

    /* Only check the database if there was a successful database close,
       this is indicated by the flag still being 1 */
    if ( blob->store_in_database == 1 )
    {
        char select[] = "select band_id, striping_factor_radiance, "
            "striping_factor_dn from striping where "
            "radiometric_wo_common_id=";

        status = snprintf( sys_command, PATH_MAX, "echo \"%s%s;\" | "
            "sqlplus $IAS_DB_L8_OLITIRS | sed -n '/^--*/{n;p;}'", select,
            blob->database_information.rad_common_id );
        if ( status < 0 || status >= PATH_MAX )
        {
            assert_true_with_message( 0, "Error creating system command to "
                "query for record");
        }
        else
        {
            FILE *inf = popen( sys_command, "r" );
            if ( !inf )
            {
                assert_true_with_message( 0, "Error opening pipe");
            }
            else
            {
                int band = 0;
                double osm_rad = 0.0;
                double osm_dn = 0.0;
                int n = fscanf( inf, "%d %lf %lf", &band, &osm_rad, &osm_dn );
                if ( n < 3 )
                    assert_true_with_message( 0, "Error reading from pipe");
                assert_equal( band, 2 );
                assert_double_equal( osm_rad, 1.02729456 );
                assert_double_equal( osm_dn, 19.29319120 );
                status = pclose( inf );
                if ( status )
                    assert_true_with_message( 0, "Error closing pipe");
            }
        }

        /* Delete the record */
        status = system( "echo \"delete from radiometric_work_order_common "
            "where work_order_id='StripingTest';\" "
            "| sqlplus $IAS_DB_L8_OLITIRS > /dev/null" );
        assert_equal_with_message( status, 0, "Error deleting the records "
            "for this test work_order_id='StripingTest'");
    }

    /* Clean up */
    cleanup_rps_level();
    remove( blob->l1r_file_name );
    cleanup_io_routines( blob );
    cleanup_memory_management_buffers(ENABLE_WARNING_DISPLAY);
    free_the_blob();
}  /* END api_striping_characterization_postproccess_test */

int main( int argc, char *argv[] )
{
    char *enable_db_tests = NULL;
    TestSuite *suite = create_test_suite();

    /* Get the test data directory */
    test_data_path = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (test_data_path == NULL)
    {
        printf("Cannot find test data directory location\n");
        return 1;
    }

    /* See if the database tests should be enabled */
    enable_db_tests = getenv( "ENABLE_DB_TESTS" );
    if ( enable_db_tests )
    {
        test_database = 1;
        printf( "\n\tDatabase checks have been enabled\n" );
    }
    else
    {
        test_database = 0;
        printf( "\n\t***** Database checks will not be made *****\n" );
    }

    add_test( suite, striping_characterization_test );
    add_test( suite, api_striping_characterization_test );
    add_test( suite, api_striping_characterization_postproccess_test );

    /* Run either a single test or the entire suite */
    if (argc == 2)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite( suite, create_text_reporter() );
}
