#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "cgreen/cgreen.h"
#include "ias_logging.h"
#include "../rps_local.h"

static char l0r_data_dir[] = "ias_base/rps/driver";
static char l0r_basename[] = "LC81800382002045SGS00";
static char cpf_data_dir[] =
       "ias_base/rps/driver/temperature_sensitivity_correction";
static char cpf_basename[] = "L8CPF20020701_20020930.01";
static char bpf_data_dir[] = 
       "ias_base/rps/driver/temperature_sensitivity_correction";
static char bpf_basename_oli[] = "LO8BPF20020802230930_20020802234142.01";
static char rlut_data_dir[] = 
       "ias_base/rps/driver/temperature_sensitivity_correction";
static char rlut_basename[] = "L8RLUT19990101_20111231v01.h5";

Ensure oli_temperature_sensitivity_correction_test(void);
Ensure tirs_temperature_sensitivity_correction_test(void);
Ensure test_algorithm_thru_api(void);
Ensure test_read_temperatures(void);

int main(int argc, char *argv[])
{
    TestSuite *suite = create_test_suite();
    add_test(suite, oli_temperature_sensitivity_correction_test);
    add_test(suite, tirs_temperature_sensitivity_correction_test);
    add_test(suite, test_algorithm_thru_api);
    add_test(suite, test_read_temperatures);

    /* Run the tests */
    if (argc > 1)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}


Ensure oli_temperature_sensitivity_correction_test(void)
{
    float mean_oli_fp_temp = 209.42166667;       
                                  /* Mean OLI focal plane temperature */
    int det;                      /* Loop counter for the detectors */
    int oli_num_dets = 506;       /* number of detectors */
    double oli_temp_sens_coeff[506]; /* Temperature sensitivity coefficients */
    double oli_ref_temp;             /* Reference temperature */
    double *temp_correction_factor;  /* Temperature correction factor */
    double expected_band1_correction_factor = 0.998589418;
    double expected_band8_correction_factor = 0.998589418;

    /* Allocate memory for the output temperature correction factor */
    temp_correction_factor = malloc(sizeof(double) * oli_num_dets);
    if (!temp_correction_factor)
    {
        assert_true_with_message(0, "Error allocating memory "
            "temperature correction factor");
        return;
    }

    for (det = 0; det < oli_num_dets; det++)
    {
        oli_temp_sens_coeff[det] = 0.005;
    }
    oli_ref_temp = 210.00;

    /* Test with the identical calval test sample code with 506 detectors */ 
    /* Call the temperature sensitivity correction algorithm */
    temperature_sensitivity_correction( 
              oli_num_dets, mean_oli_fp_temp, oli_temp_sens_coeff, 
              oli_ref_temp, temp_correction_factor);

    /* Loop through the detector */
    for (det = 0; det < oli_num_dets; det++)
    {
         /* Check the algorithm results */
         assert_double_equal(temp_correction_factor[det], 
                             expected_band1_correction_factor);
     }

    /* Free the allocated memory */
    free(temp_correction_factor);    

    /* Test with the identical calval test sample code with 1012 detectors */ 
    double oli_pan_temp_sens_coeffs[1012]; 
    int num_oli_pan_dets = 1012;

    /* Allocate memory for the output temperature correction factor */
    temp_correction_factor = malloc(sizeof(double) * num_oli_pan_dets);
    if (!temp_correction_factor)
    {
        assert_true_with_message(0, "Error allocating memory "
            "temperature correction factor");
        return;
    }

    for (det = 0; det < num_oli_pan_dets; det++)
    {
        oli_pan_temp_sens_coeffs[det] = 0.005;
    }
    
    /* Call the temperature sensitivity correction algorithm */
    temperature_sensitivity_correction( 
              num_oli_pan_dets, mean_oli_fp_temp, oli_pan_temp_sens_coeffs, 
              oli_ref_temp, temp_correction_factor);

   /* Loop through the detector */
   for (det = 0; det < num_oli_pan_dets; det++)
   {
        /* Check the algorithm results */
        assert_double_equal(temp_correction_factor[det], 
                            expected_band8_correction_factor);
   }

    /* Free the allocated memory */
    free(temp_correction_factor);    
  
}  /* END temperature_sensitivity_correction_test */

Ensure tirs_temperature_sensitivity_correction_test(void)
{
    float mean_tirs_fp_temp = 41.6842249878;      
                                  /* Mean TIRS focal plane temperature */
    int det;                      /* Loop counter for the detectors */
    int tirs_num_dets = 640;      /* number of detectors */
    double tirs_temp_sens_coeff[640];/* Temperature sensitivity coefficients */
    double tirs_ref_temp;            /* Reference temperature */
    double *temp_correction_factor;     /* Temperature correction factor */
    double expected_band10_correction_factor = 1.004852647;

    /* Allocate memory for the output temperature correction factor */
    temp_correction_factor = malloc(sizeof(double) * tirs_num_dets);
    if (!temp_correction_factor)
    {
        assert_true_with_message(0, "Error allocating memory "
            "temperature correction factor");
        return;
    }

    for (det = 0; det < tirs_num_dets; det++)
    {
        tirs_temp_sens_coeff[det] = 0.01;
    }
    tirs_ref_temp = 41.00;

    /* Call the temperature sensitivity correction algorithm */
    temperature_sensitivity_correction( 
              tirs_num_dets, mean_tirs_fp_temp, tirs_temp_sens_coeff, 
              tirs_ref_temp, temp_correction_factor);

    /* Loop through the detector */
    for (det = 0; det < tirs_num_dets; det++)
    {
         /* Check the algorithm results */
         assert_double_equal(temp_correction_factor[det], 
                             expected_band10_correction_factor);
    }

    /* Free the allocated memory */
    free(temp_correction_factor);    

}  /* END temperature_sensitivity_correction_test */

/*--------------------------------------------------------------------------
  test_algorithm_thru_api

  This test sets up for a call to the algorithm through the API.  Also calls
  the pre-process routine to set up memory management. The input ASCII files 
  are extracted from calval binary test files with only one band and one sca 
  used here. 
--------------------------------------------------------------------------*/
Ensure test_algorithm_thru_api(void)
{
    char *test_directory;       /* test environment value */
    int status;                 /* Routine return status */
    BLOB *blob;                 /* Declare a blob structure */
    int band_index;
    int list_of_bands[2] = {11, 1};         /* List of two test bands */
    int list_of_band_indices[2] = {10, 0};  /* List of two test band indices */
    int num_bands = 2;                      /* Number of bands for testing */
    double *oli_temp_correction_factor;     /* OLI temperature correction 
                                               factor */
    double *tirs_temp_correction_factor;    /* TIRS temperature correction 
                                               factor */
    double expected_band1_correction_factor = 1.374232818;
    double expected_band11_correction_factor = 6.046738998;
    int det;
    int i;

    /* Get the environment variable for the test data */
    test_directory = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (!test_directory)
    {
        assert_true_with_message(0, "Getting IAS_UNIT_TEST_DATA_DIR");
        return;
    }

    /* The attributes must be initialized for memory management to work */
    ias_sat_attr_initialize(IAS_L8);

    /* Allocate a blob and populate the minimum members for the test
       to succeed */
    blob = allocate_blob();
    blob->max_number_scas = 14;
    blob->total_bands = 30;

    sprintf(blob->l0r_file_name, "%s/%s/%s/%s", test_directory, l0r_data_dir,
        l0r_basename, l0r_basename);
    sprintf(blob->l1r_file_name, "%s", l0r_basename);
    sprintf(blob->cpf_filename, "%s/%s/%s", test_directory, cpf_data_dir,
        cpf_basename);
    sprintf(blob->bpf_filename_oli, "%s/%s/%s", test_directory, bpf_data_dir,
        bpf_basename_oli);
    strcpy(blob->bpf_filename_tirs, "");
    sprintf(blob->rlut_filename, "%s/%s/%s", test_directory, rlut_data_dir,
        rlut_basename);

    status = finish_blob_initialization();
    assert_equal_with_message(status, SUCCESS, "Error completing test blob "
        "initialization");
    if (status != SUCCESS)
    {
        free_the_blob();
        return;
    }

    for ( i = 0; i < num_bands; i++ )
    {
        list_of_band_indices[i] = 
            ias_sat_attr_convert_band_number_to_index(list_of_bands[i]);
    }
    blob->list_of_bands[0] = list_of_bands[0];
    blob->list_of_band_indices[0] = list_of_band_indices[0];
    blob->list_of_bands[1] = list_of_bands[1];
    blob->list_of_band_indices[1] = list_of_band_indices[1];
    blob->number_of_bands = 2;
    blob->process_oli_bands = 1;
    blob->process_tirs_bands = 1;

    /* Diable logging to suppress the memory management warnings */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* Initialize the L0R image data */
    status = initialize_io_routines(blob);
    ias_log_set_output_level( IAS_LOG_LEVEL_INFO );
    assert_equal_with_message(status, SUCCESS, "Error initializing image "
        "data");
    if (status != SUCCESS)
    {
        cleanup_io_routines(blob);
        free_the_blob();
        return;
    }

    /* This initializes the memory management and registers buffers through
       the algorithms pre-process routine */
    status = initialize_memory_management_buffers();
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error initializing memory management");
        cleanup_io_routines(blob);
        free_the_blob();
        return;
    }

    /* Initialize the RPS process level arrays */
    status = initialize_rps_level(30, 14);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error calling initialize_rps_level");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    status = temperature_sensitivity_correction_preprocess(blob);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error returned from: "
            "tirs_response_linearization_preprocess");
        cleanup_rps_level();
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Register the MM_TEMP_CORRECTION_FACTOR to this routine so that the API 
       can't destroy the memory before we can compare the output */
    status = increment_consumer_count(MM_TEMP_CORRECTION_FACTOR);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error returned from: "
            "increment_consumer_count MM_TEMP_CORRECTION_FACTOR");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }
    /* It is important that the consumer counts get set *after* all the
       increments */
    status = set_consumer_counts();
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error setting consumer counts");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Set the temp_correction_factor into memory buffer */
    oli_temp_correction_factor = get_memory_buffer(MM_TEMP_CORRECTION_FACTOR,
        SETTER, 0, 0);
    for (det = 0; det < blob->band_information[0]->number_detectors; det++)
        oli_temp_correction_factor[0] = 0.0;
    set_buffer_filled_flag( MM_TEMP_CORRECTION_FACTOR, 0, 0 );

    /* Set the temp_correction_factor into memory buffer */
    tirs_temp_correction_factor = get_memory_buffer(MM_TEMP_CORRECTION_FACTOR,
        SETTER, 10, 0);
    for (det = 0; det < blob->band_information[10]->number_detectors; det++)
        tirs_temp_correction_factor[0] = 0.0;
    set_buffer_filled_flag( MM_TEMP_CORRECTION_FACTOR, 10, 0 );

    /* Call the API for band 11 and SCA 1*/
    band_index = list_of_band_indices[0];
    status = api_temperature_sensitivity_correction(blob, band_index, 0, 0);
    assert_equal_with_message(status, SUCCESS, "Error returned from "
        "api_temperature_sensitivity_correction");
    if (status != SUCCESS)
    {
        cleanup_rps_level();
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* The temperature correction factor should be zero since no calculation
       is done for TIRS bands  */
    assert_double_equal(tirs_temp_correction_factor[0], 
                        expected_band11_correction_factor);

    /* Call the API for band 1 and SCA 1*/
    band_index = list_of_band_indices[1];
    status = api_temperature_sensitivity_correction(blob, band_index, 0, 0);
    assert_equal_with_message(status, SUCCESS, "Error returned from "
        "api_temperature_sensitivity_correction");
    if (status != SUCCESS)
    {
        cleanup_rps_level();
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    /* Compare the results.  */
    assert_double_equal(oli_temp_correction_factor[0], 
                        expected_band1_correction_factor);

    /* Free the image memory */
    status = decrement_consumer_count(MM_TEMP_CORRECTION_FACTOR, 0, 0);
    assert_equal_with_message( status, SUCCESS,
        "Error returned from decrement consumer count");
    status = decrement_consumer_count(MM_TEMP_CORRECTION_FACTOR, 10, 0);
    assert_equal_with_message( status, SUCCESS,
        "Error returned from decrement consumer count");

    cleanup_rps_level();
    cleanup_io_routines(blob);
    cleanup_memory_management_buffers(ENABLE_WARNING_DISPLAY);
    free_the_blob();
}  /* END test_algorithm_thru_api */


/*--------------------------------------------------------------------------
  test_read_temperatures

  This test attempts to read temperature information from a short collect
  where the telemetry and/or ancillary data do not overlap the image data. 
--------------------------------------------------------------------------*/
Ensure test_read_temperatures(void)
{
    char test_data_dir[] = "temperature_sensitivity_correction/short_solar";
    char *test_directory;       /* test environment value */
    int status;                 /* Routine return status */
    BLOB *blob;                 /* Declare a blob structure */
    int list_of_bands[2] = {11, 1};         /* List of two test bands */
    int list_of_band_indices[2] = {10, 0};  /* List of two test band indices */
    int num_bands = 2;                      /* Number of bands for testing */
    int i;
    char file_name[] = "temp_l0r";
    float *fp_temp1;
    float *fp_temp2;
    int num_temps;

    /* Get the environment variable for the test data */
    test_directory = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (!test_directory)
    {
        assert_true_with_message(0, "Getting IAS_UNIT_TEST_DATA_DIR");
        return;
    }

    /* The attributes must be initialized for memory management to work */
    ias_sat_attr_initialize(IAS_L8);

    /* Allocate a blob and populate the minimum members for the test
       to succeed */
    blob = allocate_blob();
    blob->max_number_scas = 14;
    blob->total_bands = 30;

    sprintf(blob->l0r_file_name, "%s/%s/%s/%s", test_directory, l0r_data_dir,
        test_data_dir, file_name );
    sprintf(blob->l1r_file_name, "%s", file_name );
    sprintf(blob->cpf_filename, "%s/%s/%s", test_directory, cpf_data_dir,
        cpf_basename);
    sprintf(blob->bpf_filename_oli, "%s/%s/%s", test_directory, bpf_data_dir,
        bpf_basename_oli);
    strcpy(blob->bpf_filename_tirs, "");
    sprintf(blob->rlut_filename, "%s/%s/%s", test_directory, rlut_data_dir,
        rlut_basename);

    status = finish_blob_initialization();
    assert_equal_with_message(status, SUCCESS, "Error completing test blob "
        "initialization");
    if (status != SUCCESS)
    {
        free_the_blob();
        return;
    }

    for ( i = 0; i < num_bands; i++ )
    {
        list_of_band_indices[i] = 
            ias_sat_attr_convert_band_number_to_index(list_of_bands[i]);
    }
    blob->list_of_bands[0] = list_of_bands[0];
    blob->list_of_band_indices[0] = list_of_band_indices[0];
    blob->list_of_bands[1] = list_of_bands[1];
    blob->list_of_band_indices[1] = list_of_band_indices[1];
    blob->number_of_bands = 2;

    /* Diable logging to suppress the memory management warnings */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* Initialize the L0R image data */
    status = initialize_io_routines(blob);
    ias_log_set_output_level( IAS_LOG_LEVEL_INFO );
    assert_equal_with_message(status, SUCCESS, "Error initializing image "
        "data");
    if (status != SUCCESS)
    {
        cleanup_io_routines(blob);
        free_the_blob();
        return;
    }

    status = read_oli_focal_plane_temperatures( &fp_temp1, &fp_temp2,
        &num_temps );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0,
            "Error reading OLI focal plane temperatures");
        cleanup_io_routines(blob);
        free_the_blob();
        return;
    }
    assert_equal( num_temps, 1 );
    assert_double_equal( fp_temp1[0], 20.0 );
    assert_double_equal( fp_temp2[0], 30.0 );

    free( fp_temp1 );
    free( fp_temp2 );
    cleanup_io_routines(blob);
    free_the_blob();
}  /* END test_read_temperatures */
