#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "cgreen/cgreen.h"
#include "ias_satellite_attributes.h"
#include "../rps_local.h"
#include "../rps_blob.h"

/*******************************************************************************
  This programs verifies histogram stat report and L1R creation through the
  driver
*******************************************************************************/

Ensure the_driver_works();

static char *unit_test_dir = NULL;

int main( int argc, char *argv[] )
{
    TestSuite *suite = create_test_suite();
    unit_test_dir = getenv("IAS_UNIT_TEST_DATA_DIR");
    if ( unit_test_dir == NULL )
    {
        printf("The IAS_UNIT_TEST_DATA_DIR environment variable is not set\n");
        exit(EXIT_FAILURE);
    }
    setenv( "PCS_WOID", "L20091", 0 );
    add_test( suite, the_driver_works );

    if ( argc > 1 )
    {
        return run_single_test( suite, "the_driver_works",
            create_text_reporter() );
    }
    return run_test_suite( suite, create_text_reporter() );
}

Ensure the_driver_works()
{
    int status;
    struct stat buf;
    int band;
    int sca;
    char report_dir_name[] = "./report_files";
    char hist_report_root[] = "histogram_statistics_characterization";
    char l1r_output_name[] = "test_driver.h5";
    char data_dir[] = "ias_base/rps/driver/sca_overlap";
    char report_name[100];
    char cmd[256];
    char command[PATH_MAX];
    int error_happened = 0;

    /* Make sure the report files directory is not present */
    status = stat(report_dir_name, &buf );
    if ( status == 0 )
    {
        /* Remove the directory */
        sprintf( cmd, "rm -rf %s", report_dir_name );
        status = system( cmd );
        if ( status )
        {
            assert_true_with_message(0,
                "Error removing report files directory");
        }
        else
            printf("Removed directory: %s\n", report_dir_name );
    }

    /* Make sure the L1R output file is not present */
    status = stat( l1r_output_name, &buf );
    if ( status == 0 )
    {
        /* Remove the file */
        sprintf( cmd, "rm -f %s", l1r_output_name );
        status = system( cmd );
        if ( status )
        {
            assert_true_with_message(0,
                "Error removing file: %s", l1r_output_name );
        }
        else
            printf("Removed file: %s\n", l1r_output_name );
    }

    /* Since the OMF gets updated make a copy of it from the template and
       the IAS_UNIT_TEST_DATA_DIR must be expanded */
    status = snprintf( command, PATH_MAX,
        "sed 's|$IAS_UNIT_TEST_DATA_DIR|%s|' I14920_template.omf > "
        "I14920.omf", unit_test_dir );
    status = system( command );
    if ( status )
    {
        assert_true_with_message( 0, "Error making copy of I14920.omf");
        return;
    }

    /* Execute the driver */
    status = system("export IAS_LOG_LEVEL=WARN; ../create_l1r test_driver.odl");
    assert_equal( status, 0 );
    if ( status )
    {
        /* Clean up and return */
        printf("Error executing: ../create_l1r test_driver.odl\n");
        sprintf( cmd, "rm -rf %s %s", l1r_output_name, report_dir_name );
        status = system( cmd );
        if ( status )
        {
            assert_true_with_message(0,
                "Error executing system command: %s", cmd );
        }
        return;
    }

    /* See if the report directory exists */
    status = stat( report_dir_name, &buf );
    if ( status )
    {
        assert_true_with_message(0,
            "The report directory was not created");
        error_happened = 1;
    }
    else
    {
        /* Ensure the histogram reports were created */
        for ( band = 1; band < 20; band += 18 )
        {
            for ( sca = 1; sca <= 14; sca++ )
            {
                if (band == 1)
                {
                    sprintf( report_name, "%s/%s_%s_%02d_%02d.rpt",
                         report_dir_name, hist_report_root, "CBL", band, sca );
                }
                else
                {
                    sprintf( report_name, "%s/%s_%s_%02d_%02d.rpt",
                         report_dir_name, hist_report_root, "0..", band, sca );
                }
                status = stat( report_name, &buf );
                if ( status )
                {
                    assert_true_with_message(0,
                        "Report not created: %s", report_name );
                    error_happened = 1;
                }
                else
                    assert_equal( status, 0 );
            }
        }
    }

    /* Spot check a report file for reasonable values */
    sprintf( report_name, "%s_%s_01_08.rpt", hist_report_root, "CBL" );
    status = snprintf( command, PATH_MAX, "sed '/ Detector Valid/,$!'d %s/%s | "
        "diff -q - %s/%s/%s", report_dir_name, report_name, unit_test_dir,
        data_dir, report_name );
    if ( status < 0 || status >= PATH_MAX )
    {
        assert_true_with_message( 0, "Error creating system command to "
            "compare %s, comparison skipped",  report_name );
        error_happened = 1;
    }
    else
    {
        status = system( command );
        if ( status )
            error_happened = 1;
        assert_equal( status, 0 );
    }


    /* Make sure the L1R file was created */
    status = stat( l1r_output_name, &buf );
    if ( status )
    {
        assert_true_with_message(0,
            "The output file, %s, was not created", l1r_output_name );
        error_happened = 1;
    }
    else
    {
        /* The file size should be larger than ... */
        assert_true( buf.st_size > 183000000 );
    }

    /* Don't clean up if error happened */
    if ( error_happened )
    {
        printf("An error occurred: returning without cleaning up\n");
        return;
    }

    /* Clean up */
    sprintf( cmd, "rm -rf %s %s", l1r_output_name, report_dir_name );
    status = system( cmd );
    if ( status )
    {
        assert_true_with_message(0,
            "Error executing system command: %s", cmd );
    }
}
