/*----------------------------------------------------------------------------
NAME:    tirs_bias_removal_full_test.c

PURPOSE: Test TIRS bias  removal with all four options and test them through 
         driver with different input parameter ODL files 

USAGE:   oli_bias_determination_test [test name]

         When run without a command line argument all tests will be run.
         Pass the name of a test to run the single test -- useful for
         debugging.
----------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include "cgreen/cgreen.h"
#include "ias_logging.h"
#include "../rps_local.h"
#include "ias_l0r.h"

Ensure tirs_bias_removal_full_test();

static char *unit_test_dir = NULL;

int main( int argc, char *argv[] )
{
    unit_test_dir = getenv("IAS_UNIT_TEST_DATA_DIR");
    if ( unit_test_dir == NULL )
    {
        printf("The IAS_UNIT_TEST_DATA_DIR environment variable is not set\n");
        exit(EXIT_FAILURE);
    }
    setenv( "PCS_WOID", "L2000248", 0 );
    TestSuite *suite = create_test_suite();
    add_test( suite, tirs_bias_removal_full_test );

    if ( argc > 1 )
    {
     return run_single_test( suite, argv[1],
               create_text_reporter() );
    }
    return run_test_suite( suite, create_text_reporter() );
}

Ensure tirs_bias_removal_full_test()
{
    int status;
    int i;
    struct stat buf;
    char l1r_output_name[] = "l1r_image.h5";
    char cmd[256];
    static char *unit_test_dir = NULL;
    char command[PATH_MAX];
    int error_happened = 0;
    char odl_file[40];
    char *p_char;            /* character pointer used for filename parsing */
    char *odl_files[] = {"tirs_bias_source5.odl", 
         "tirs_bias_source6.odl", "tirs_bias_source7.odl", 
         "tirs_bias_source8.odl"};
    int num_odl_files = 4;
    char *dump = "$COTS/hdf5/bin/h5dump --binary=LE -o b10_partial_image "
                    "--dataset=\"B10\" -s \"0,1,1\" -c \"1,480,275\" "
                    "l1r_image.h5 >& /dev/null";

    /* Get the IAS_UNIT_TEST_DATA_DIR environment variable */
    unit_test_dir = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (unit_test_dir == NULL)
    {
        assert_true_with_message( 0, "Erroe: The IAS_UNIT_TEST_DATA_DIR "
                                  "environment variable is not set\n");
        return;
    }

    /* Loop through the ODL files in the text file and get the file names */
    for (i = 0; i < num_odl_files; i++)
    {
        /* Make sure the L1R output file is not present */
        status = stat( l1r_output_name, &buf );
        if ( status == 0 )
        {
            /* Remove the file */
            sprintf( cmd, "rm -f %s", l1r_output_name );
            status = system( cmd );
            if ( status )
            {
                assert_true_with_message(0,
                    "Error removing file: %s", l1r_output_name );
            }
            else
                printf("Removed file: %s\n", l1r_output_name );
        }

        /* Since the OMF gets updated make a copy of it from the template and
           the IAS_UNIT_TEST_DATA_DIR must be expanded */
        status = snprintf(command, PATH_MAX,
                "sed 's|$IAS_UNIT_TEST_DATA_DIR|%s|' template_I16960.omf > "
                "I16960.omf", unit_test_dir);
        if (status < 0 || PATH_MAX < status)
        {
            assert_true_with_message(0, "Error preparing system command");
            return;
        }
        status = system(command);
        if (status != 0)
        {
            assert_true_with_message(0, "Error preparing OMF file for unit "
                    "test data directory path setup");
            return;
        }

        /* Create the command for running create_l1r */
        status = snprintf(command, PATH_MAX, "export IAS_LOG_LEVEL=WARN; "
                          "../create_l1r %s", odl_files[i]);
        if (status < 0 || status > PATH_MAX)
        {
            assert_true_with_message(0, "Error preparing system command for "
                "executing create_l1r");
            return;
        }
    
        /* Execute create_l1r */
        printf("Executing: %s\n", command);
        status = system( command );
        printf("Done: %s\n", command);
        assert_equal( status, 0 );
        if ( status )
        {
            /* Clean up and return */
            printf("Error executing: %s\n", command);
            sprintf( cmd, "rm -rf %s", l1r_output_name );
            status = system( cmd );
            if ( status )
            {
                assert_true_with_message( 0,
                    "Error executing system command: %s", cmd );
            }
            return;
        }

        /* Make sure the L1R file was created */
        status = stat( l1r_output_name, &buf );
        if ( status )
        {
            assert_true_with_message(0,
                "The output file, %s, was not created", l1r_output_name );
            error_happened = 1;
        }

        /* Get the odl_filename without suffix */
        strcpy(odl_file, odl_files[i]);
        p_char = strchr( odl_file, '.' );
        if ( p_char )
            p_char[0] = '\0';

        /* Do a partial diff on dumped partial L1R image on band 10 */
        if ( system( dump ))
        {
            assert_true_with_message(0, "Error dumped image");
            return;
        }
        sprintf(cmd, "diff -q b10_partial_image "
            "%s/ias_base/rps/driver/tirs_bias_removal/%s/"
            "b10_partial_image", unit_test_dir, odl_file);
        if ( system( cmd ))
        {
             assert_true_with_message(0, 
                    "Differences occur for %s", odl_files[i]);
             return;
        }

        /* Don't clean up if error happened */
        if ( error_happened )
        {
            printf("An error occurred: returning without cleaning up\n");
            return;
        }

        /* Clean up */
        sprintf( cmd, "rm -rf %s %s %s", l1r_output_name, "b10_partial_image", 
                 "I16960.omf" );
        status = system( cmd );
        if ( status )
        {
            assert_true_with_message( 0,
            "Error executing system command: %s", cmd );
        }
    }
}
