/*----------------------------------------------------------------------------
NAME:    tirs_bias_removal_test.c

PURPOSE: Test the TIRS bias removal routine by testing it directly and testing
         it through API

USAGE:   tirs_bias_removal_test [test name]

         When run without a command line argument all tests will be run.
         Pass the name of a test to run the single test -- useful for
         debugging.
----------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "ias_logging.h"
#include "cgreen/cgreen.h"
#include "rps_local.h"
#include "../../../config.h"

#define NLINES 10
#define NDETS  640

/* Path to test data and file names */
static char cpf_dir[] = "ias_lib/cpf";
static char bpf_dir[] = "ias_lib/bpf";
static char data_dir[] = "ias_base/rps/driver/tirs_bias_removal";
static char data_file[] = "input_scene.txt";
static char dark_file[] = "dark_scene.txt";
static char background_file[] = "background_scene.txt";
static char pre_acq_file[] = "pre_acq.txt";
static char results_file6[] = "output_scene6.txt";
static char results_file5[] = "output_scene5.txt";
static char cpf_file[] = "L08CPF20090101_20090331.01";
static char tirs_bpf_file[] = "test_LT8BPF20140310103310_20140310103345.01";

/* Test prototypes */
Ensure test_algorithm_directly(void);
Ensure test_algorithm_thru_api_bpf(void);
Ensure test_algorithm_thru_api_cpf(void);

int main(int argc, char *argv[])
{
    TestSuite *suite = create_test_suite();
    add_test(suite, test_algorithm_directly);
    add_test(suite, test_algorithm_thru_api_bpf);
    add_test(suite, test_algorithm_thru_api_cpf);

    if (argc == 2)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}


/*--------------------------------------------------------------------------
   test_algorithm_directly

   This test executes the algorithm with a subset of the Cal/Val prototype
   input and output. The input Ascii files are extracted from calval binary 
   test files with only one band and one sca used here. 
---------------------------------------------------------------------------*/
Ensure test_algorithm_directly(void)
{
    FILE *inf;                     /* File pointer for all reads */
    char *test_directory;          /* Test environment value */
    char file_name[PATH_MAX];      /* File name to open */
    float *data_buffer;            /* Buffer containing the input data */
    double *dark_buffer;           /* Dark response data buffer */
    double *background_buffer;     /* Background response data buffer */
    float *results_buffer;         /* Buffer for expected results from cal/val
                                      code */
    int pixel;                     /* Pixel loop counter */
    int num_values;                /* Number of values for malloc and reads */
    int difference_count;          /* Nount of differences between the 
                                      calculated value and expected results */
    int return_value;              /* Return values from fscanf read */ 
    int det, line;
    float temp_float;

    /* Get the environment variable for the test data */
    test_directory = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (!test_directory)
    {
        assert_true_with_message(0, "Getting IAS_UNIT_TEST_DATA_DIR");
        return;
    }

    /* Open and read the data buffer, only the first sca and first row 
       640x500 data will be used out for testing purpose */
    sprintf(file_name, "%s/%s/%s", test_directory, data_dir, data_file);
    num_values = NLINES * NDETS;
    data_buffer = malloc(num_values * sizeof(float));
    if ( !data_buffer )
    {
        assert_true_with_message(0, "Allocating memory for data_buffer");
        return;
    }

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        free( data_buffer );
        return;
    }

    for (det = 0; det < NDETS; det++)
    {
        for (line = 0; line < NLINES; line++)
        {
            return_value = fscanf(inf, "%f", &temp_float);
            if (return_value != 1)
            {
                assert_true_with_message(0, "Reading %s", file_name);
                free( data_buffer );
                return;        
            } 
            data_buffer[det * NLINES + line] = temp_float;       
        }
    }
    fclose(inf);

    /* Open and read the dark response parameters */
    sprintf(file_name, "%s/%s/%s", test_directory, data_dir, dark_file);
    dark_buffer = malloc(NDETS * sizeof(double));
    if (!dark_buffer)
    {
        assert_true_with_message(0, "Allocating memory for dark_buffer" );
        free(data_buffer);
        return;
    }

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        free(data_buffer);
        free(dark_buffer);
        return;
    }

    for (det = 0; det < NDETS; det++)
    {
        return_value = fscanf(inf, "%f", &temp_float);
        if (return_value != 1)
        {
            assert_true_with_message(0, "Reading %s", file_name);
            free( data_buffer );
            free(dark_buffer);
            return;        
        } 
        dark_buffer[det] = (double)temp_float;       
    }
    fclose(inf);

    /* Open and read the background response parameters */
    sprintf(file_name, "%s/%s/%s", test_directory, data_dir, background_file);
    background_buffer = malloc(NDETS * sizeof(double));
    if (!background_buffer)
    {
        assert_true_with_message(0, "Allocating memory for background " 
                                         "buffer");
        free(data_buffer);
        free(dark_buffer);
        return;
    }

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        free(data_buffer);
        free(dark_buffer);
        free(background_buffer);
        return;
    }

    for (det = 0; det < NDETS; det++)
    {
        return_value = fscanf(inf, "%f", &temp_float);
        if (return_value != 1)
        {
            assert_true_with_message(0, "Reading %s", file_name);
            free( data_buffer );
            free(dark_buffer);
            free(background_buffer);
            return;        
        } 
        background_buffer[det] = (double)temp_float;       
    }
    fclose(inf);

    for (det = 0; det < NDETS; det++)
        background_buffer[det] += dark_buffer[det];

    /* Call the algorithm */
    generic_parameter_removal(&data_buffer[0], NDETS, NLINES,
        background_buffer);

    /* Open and read the results file */
    sprintf(file_name, "%s/%s/%s", test_directory, data_dir, results_file5);
    results_buffer = malloc(num_values * sizeof(float));
    if (!results_buffer)
    {
        assert_true_with_message(0, "Allocating memory for background " 
                                         "buffer");
        free(data_buffer);
        free(dark_buffer);
        free(background_buffer);
        return;
    }

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        free(data_buffer);
        free(dark_buffer);
        free(background_buffer);
        free(results_buffer);
        return;
    }

    for (det = 0; det < NDETS; det++)
    {
        for (line = 0; line < NLINES; line++)
        {
            return_value = fscanf(inf, "%f", &temp_float);
            if (return_value != 1)
            {
                assert_true_with_message(0, "Reading %s", file_name);
                free( data_buffer );
                free(dark_buffer);
                free(background_buffer);
                free(results_buffer);
                return;        
            } 
            results_buffer[det * NLINES + line] = temp_float;       
        }
    }
    fclose(inf);
    
    /* Compare the results.  Loop through the results and count the
       differences.  This comparison isn't done with asserts since there is
       no need for 320000 assert statements which cost message buffer space.
       One assert for pass fail with a count is good enough. */
    difference_count = 0;
    for (pixel = 0; pixel < NDETS * NLINES; pixel++)
    {
        if (fabsf(data_buffer[pixel] - results_buffer[pixel]) > 1e-3)
            difference_count++;
    }

    assert_equal_with_message(difference_count, 0, "There were %d "
        "differences out of %d values", difference_count, NDETS*NLINES);


    /* Free the allocated memory */
    free(data_buffer);
    free(dark_buffer);
    free(background_buffer);
    free(results_buffer);
} 

/*--------------------------------------------------------------------------
  local_free_the_blob

  This a local override of the routine to free the blob.  Since not all
  members of the blob are initialized -- specifically the pixel mask -- the
  normal routine cannot be called.
--------------------------------------------------------------------------*/
static void local_free_the_blob(BLOB *blob)
{
    int band_index;
    
    if ( blob->l1r_band_metadata )
    {
        for ( band_index = 0; band_index < blob->total_bands; band_index++ )
            free( blob->l1r_band_metadata[band_index] );
        free( blob->l1r_band_metadata );
    }

    for (band_index = 0; band_index < 30; band_index++)
        free( blob->band_information[band_index] );
    free( blob->band_information );
    if ( blob->cpf )
        ias_cpf_free( blob->cpf );
    if ( blob->bpf_tirs )
        ias_bpf_free( blob->bpf_tirs );
    free( blob );
}


/*--------------------------------------------------------------------------
  test_algorithm_thru_api_bpf

  This test sets up for a call to the algorithm through the API.  Also calls
  the pre-process routine to set up memory management. The input Ascii files 
  are extracted from calval binary test files with only one band and one sca 
  used here. The BACKGROUND_RESPONSE_SELECTION is 1, that's pre-aquisition 
  deep space averages from BPF.
--------------------------------------------------------------------------*/
Ensure test_algorithm_thru_api_bpf(void)
{
    FILE *inf;                  /* File pointer for all reads */
    char *test_directory;       /* test environment value */
    char file_name[PATH_MAX];   /* File name to open */
    float *l1r_buffer;          /* buffer containing the input data */
    double *pre_acq_buffer;     /* Pre-acquisition deep space data buffer */
    float *results_buffer;      /* Buffer for expected results from cal/val
                                   code */
    int pixel;                  /* Loop parameter for pixel */
    int status;                 /* Routine return status */
    BLOB *blob;                 /* Declare a blob structure */
    const IAS_BAND_ATTRIBUTES *band_atts;
    int band_number;
    int difference_count;       /* count of differences between the 
                                   calculated value and expected results */
    int return_value;           /* Return values from fscanf read */ 
    int list_of_bands[2] = {1, 11};
    int list_of_band_indices[2];
    int num_bands = 2;
    int band_index;
    int det, line;
    float temp_float;

    /* Get the environment variable for the test data */
    test_directory = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (!test_directory)
    {
        assert_true_with_message(0, "Getting IAS_UNIT_TEST_DATA_DIR");
        return;
    }

    /* The attributes must be initialized for memory management to work */
    ias_sat_attr_initialize(IAS_L8);

    /* Allocate a blob and populate the minimum members for the test
       to succeed */
    blob = allocate_blob();
    blob->max_number_scas = 14;
    blob->total_bands = 30;
    blob->l1r_band_metadata = calloc( blob->total_bands,
        sizeof( IAS_L1R_BAND_METADATA * ));
    if ( ! blob->l1r_band_metadata )
    {
        assert_true_with_message(0, "Error allocating band metadata");
        local_free_the_blob( blob );
        return;
    }
    for ( band_index = 0; band_index < blob->total_bands; band_index++ )
    {
        blob->l1r_band_metadata[band_index] = calloc( 1,
            sizeof(IAS_L1R_BAND_METADATA));
        if ( ! blob->l1r_band_metadata[band_index] )
        {
            assert_true_with_message(0, "Error allocating band metadata");
            local_free_the_blob( blob );
            return;
        }
    }

    blob->band_information = calloc(30, sizeof( RPS_BAND_ATTRIBUTES *));
    for (band_number = 1; band_number <= 30; band_number++)
    {
        int i;
        int band_index;

        band_atts = ias_sat_attr_get_band_attributes(band_number);
        if (band_atts == NULL)
        {
            assert_true_with_message(FALSE, "Error obtaining RPS attribute "
                "information for band %d", band_number);
            local_free_the_blob(blob);
            return;
        }

        band_index = band_atts->band_index;
        blob->band_information[band_index] = calloc(1,
            sizeof(RPS_BAND_ATTRIBUTES));
        if (!blob->band_information[band_index])
        {
            assert_true_with_message(0, "Error allocating band "
                "attribute memory");
            local_free_the_blob(blob);
            return;
        }
        blob->band_information[band_index]->band_number = band_number;
        blob->band_information[band_index]->number_lines = 10;
        blob->band_information[band_index]->number_detectors = 640;
        blob->band_information[band_index]->number_scas = band_atts->scas;
        blob->band_information[band_index]->band_classification =
            band_atts->band_classification;
        blob->band_information[band_index]->spectral_type =
            band_atts->spectral_type;
        for ( i = 0; i < num_bands; i++ )
        {
            if ( band_number == list_of_bands[i] )
            {
                list_of_band_indices[i] = band_index;
                break; /* Since there are no repeated bands */
            }
        }
    }
    blob->background_response_selection = 1;
    blob->process_oli_bands = 1;
    blob->process_tirs_bands = 0;

    /* Open the CPF */
    sprintf(file_name, "%s/%s/%s", test_directory, cpf_dir, cpf_file);
    blob->cpf = ias_cpf_read(file_name);
    if (!blob->cpf)
    {
        assert_true_with_message(0, "Error opening CPF: %s", file_name);
        local_free_the_blob(blob);
        return;
    }

    /* Open the TIRS_BPF */
    sprintf(file_name, "%s/%s/%s", test_directory, bpf_dir, tirs_bpf_file);
    blob->bpf_tirs = ias_bpf_read(file_name);
    if (!blob->bpf_tirs)
    {
        assert_true_with_message(0, "Error opening TIRS BPF: %s", file_name);
        local_free_the_blob(blob);
        return;
    }

    /* Disable logging to suppress the memory management warnings */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* This initializes the memory management and registers buffers through
       the algorithms pre-process routine.  This algorithm doesn't need any
       memory management buffers that require the list of bands to be
       initialized -- number of bands set to 0 for that reason. */
    blob->number_of_bands = 0;
    blob->band_fields_initialized = 1;
    status = initialize_memory_management_buffers();
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error initializing memory management");
        local_free_the_blob(blob);
        return;
    }

    /* Initialize the RPS process level arrays */
    status = initialize_rps_level(30, 14);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error calling initialize_rps_level");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    status = tirs_bias_removal_preprocess(blob);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error returned from: "
            "tirs_bias_removal_preprocess");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    /* Register the L1R buffer to this routine so that the API can't
       destroy the memory before we can retrieve the output */
    status = increment_consumer_count(IMAGE_L1R);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error returned from: "
            "increment_consumer_count IMAGE_L1R");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }
    /* It is important that the consumer counts get set *after* all the
       increments */
    set_consumer_counts();

    /* Check an "empty" buffer for the pre-acquistion, the TIRS band
       is the second one in the list */
    pre_acq_buffer = (double *)get_memory_buffer( MM_BPF_BACKGROUND_RESPONSE,
        SETTER, list_of_band_indices[1], 0);
    if (!pre_acq_buffer)
    {
        assert_true_with_message(0, "Error getting "
            "MM_BPF_BACKGROUND_RESPONSE buffer");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    /* Open and read the pre_acquisition from pre_acquisition data file */
    sprintf(file_name, "%s/%s/%s", test_directory, data_dir, pre_acq_file);

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    for (det = 0; det < NDETS; det++)
    {
        return_value = fscanf(inf, "%f", &temp_float);
        if (return_value != 1)
        {
            assert_true_with_message(0, "Reading %s", file_name);
            cleanup_rps_level();
            cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
            local_free_the_blob(blob);
            return;        
        } 
        pre_acq_buffer[det] = (double)temp_float;       
    }
    fclose(inf);
    set_buffer_filled_flag( MM_BPF_BACKGROUND_RESPONSE,
        list_of_band_indices[1], 0);

    /* Get the image buffer for band 11, this simply returns an unfilled,
       allocated buffer, band 11 is second in the list */
    l1r_buffer = (float *)get_memory_buffer(IMAGE_L1R, SETTER,
        list_of_band_indices[1], 0);
    if (!l1r_buffer)
    {
        assert_true_with_message(0, "Getting IMAGE_L1R buffer");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    /* Open and read the data buffer */
    sprintf(file_name, "%s/%s/%s", test_directory, data_dir, data_file);

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    for (det = 0; det < NDETS; det++)
    {
        for (line = 0; line < NLINES; line++)
        {
            return_value = fscanf(inf, "%f", &temp_float);
            if (return_value != 1)
            {
                assert_true_with_message(0, "Reading %s", file_name);
                cleanup_rps_level();
                cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
                local_free_the_blob(blob);
                return;        
            } 
            l1r_buffer[det * NLINES + line] = temp_float;       
        }
    }
    fclose(inf);
    set_buffer_filled_flag( IMAGE_L1R, list_of_band_indices[1], 0);

    /* Call the API with a non-TIRS band.  The API should return immediately.
       The only way to really check this is to ensure the data buffer wasn't
       changed at all. */
    band_number = list_of_bands[0];
    band_index = list_of_band_indices[0];
    status = api_tirs_bias_removal( blob, band_index, 0, 0 );
    assert_equal_with_message( status,SUCCESS, "Error returned from "
        "api_bias_removal with band = 1");
    if ( status != SUCCESS )
    {
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob( blob );
        return;
    }

    /* Check some outputs with the original inputs, no changes */
    assert_double_equal(l1r_buffer[0], 4326.954102);
    assert_double_equal(l1r_buffer[640], 4330.139648);
    assert_double_equal(l1r_buffer[640*2], 4326.954102);

    /* Call the API */
    band_number = list_of_bands[1];
    band_index = list_of_band_indices[1];
    status = api_tirs_bias_removal(blob, band_index, 0, 0);
    assert_equal_with_message(status, SUCCESS, "Error returned from "
        "api_tirs_bias_removal with band = 11");
    if (status != SUCCESS)
    {
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    /* Open and read the results file */
    sprintf(file_name, "%s/%s/%s", test_directory, data_dir, results_file6);
    results_buffer = malloc(NDETS * NLINES * sizeof(float));
    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    for (det = 0; det < NDETS; det++)
    {
        for (line = 0; line < NLINES; line++)
        {
            return_value = fscanf(inf, "%f", &temp_float);
            if (return_value != 1)
            {
                assert_true_with_message(0, "Reading %s", file_name);
                cleanup_rps_level();
                cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
                local_free_the_blob(blob);
                return;        
            } 
            results_buffer[det * NLINES + line] = temp_float;       
        }
    }
    fclose(inf);

    /* Compare the results.  Loop through the results and count the
       differences.  This comparison isn't done with asserts since there is
       no need for 320000 assert statements which cost message buffer space.
       One assert for pass fail with a count is good enough. */
    difference_count = 0;
    for (pixel = 0; pixel < NDETS * NLINES; pixel++)
    {
        if (fabsf(l1r_buffer[pixel] - results_buffer[pixel]) > 1e-3)
            difference_count++;
    }

    assert_equal_with_message(difference_count, 0, "There were %d "
        "differences out of %d values", difference_count, NDETS*NLINES);

    /* Free the image memory */
    status = decrement_consumer_count(IMAGE_L1R, list_of_band_indices[1], 0);
    assert_equal_with_message( status, SUCCESS,
        "Error returned from decrement consumer count");

    free(results_buffer);
    cleanup_rps_level();
    cleanup_memory_management_buffers(ENABLE_WARNING_DISPLAY);
    local_free_the_blob(blob);
}
 
/*--------------------------------------------------------------------------
  test_algorithm_thru_api_cpf

  This test sets up for a call to the algorithm through the API.  Also calls
  the pre-process routine to set up memory management. The input Ascii files 
  are extracted from calval binary test files with only one band and one sca 
  used here. The BACKGROUND_RESPONSE_SELECTION is 4, that's to use dark and
  background responses from CPF.
--------------------------------------------------------------------------*/
Ensure test_algorithm_thru_api_cpf(void)
{
    FILE *inf;                  /* File pointer for all reads */
    char *test_directory;       /* test environment value */
    char file_name[PATH_MAX];   /* File name to open */
    float *l1r_buffer;          /* buffer containing the input data */
    double *dark_buffer;        /* Dark response data buffer */
    double *background_buffer;  /* Background response data buffer */
    float *results_buffer;      /* Buffer for expected results from cal/val
                                   code */
    int pixel;                  /* Loop parameter for pixel */
    int status;                 /* Routine return status */
    BLOB *blob;                 /* Declare a blob structure */
    const IAS_BAND_ATTRIBUTES *band_atts;
    int band_number;
    int difference_count;       /* count of differences between the 
                                   calculated value and expected results */
    int return_value;           /* Return values from fscanf read */ 
    int list_of_bands[2] = {1, 11};
    int list_of_band_indices[2];
    int num_bands = 2;
    int band_index;
    int det, line;
    float temp_float;

    /* Get the environment variable for the test data */
    test_directory = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (!test_directory)
    {
        assert_true_with_message(0, "Getting IAS_UNIT_TEST_DATA_DIR");
        return;
    }

    /* The attributes must be initialized for memory management to work */
    ias_sat_attr_initialize(IAS_L8);

    /* Allocate a blob and populate the minimum members for the test
       to succeed */
    blob = allocate_blob();
    blob->max_number_scas = 14;
    blob->total_bands = 30;
    blob->l1r_band_metadata = calloc( blob->total_bands,
        sizeof( IAS_L1R_BAND_METADATA * ));
    if ( ! blob->l1r_band_metadata )
    {
        assert_true_with_message(0, "Error allocating band metadata");
        local_free_the_blob( blob );
        return;
    }
    for ( band_index = 0; band_index < blob->total_bands; band_index++ )
    {
        blob->l1r_band_metadata[band_index] = calloc( 1,
            sizeof(IAS_L1R_BAND_METADATA));
        if ( ! blob->l1r_band_metadata[band_index] )
        {
            assert_true_with_message(0, "Error allocating band metadata");
            local_free_the_blob( blob );
            return;
        }
    }

    blob->band_information = calloc(30, sizeof( RPS_BAND_ATTRIBUTES *));
    for (band_number = 1; band_number <= 30; band_number++)
    {
        int i;
        int band_index;

        band_atts = ias_sat_attr_get_band_attributes(band_number);
        if (band_atts == NULL)
        {
            assert_true_with_message(FALSE, "Error obtaining RPS attribute "
                "information for band %d", band_number);
            local_free_the_blob(blob);
            return;
        }

        band_index = band_atts->band_index;
        blob->band_information[band_index] = calloc(1,
            sizeof(RPS_BAND_ATTRIBUTES));
        if (!blob->band_information[band_index])
        {
            assert_true_with_message(0, "Error allocating band "
                "attribute memory");
            local_free_the_blob(blob);
            return;
        }
        blob->band_information[band_index]->band_number = band_number;
        blob->band_information[band_index]->number_lines = 10;
        blob->band_information[band_index]->number_detectors = 640;
        blob->band_information[band_index]->number_scas = band_atts->scas;
        blob->band_information[band_index]->band_classification =
            band_atts->band_classification;
        blob->band_information[band_index]->spectral_type =
            band_atts->spectral_type;
        for ( i = 0; i < num_bands; i++ )
        {
            if ( band_number == list_of_bands[i] )
            {
                list_of_band_indices[i] = band_index;
                break; /* Since there are no repeated bands */
            }
        }
    }
    blob->background_response_selection = 4;

    /* Open the CPF */
    sprintf(file_name, "%s/%s/%s", test_directory, cpf_dir, cpf_file);
    blob->cpf = ias_cpf_read(file_name);
    if (!blob->cpf)
    {
        assert_true_with_message(0, "Error opening CPF: %s", file_name);
        local_free_the_blob(blob);
        return;
    }

    /* Open the TIRS_BPF */
    sprintf(file_name, "%s/%s/%s", test_directory, bpf_dir, tirs_bpf_file);
    blob->bpf_tirs = ias_bpf_read(file_name);
    if (!blob->bpf_tirs)
    {
        assert_true_with_message(0, "Error opening TIRS BPF: %s", file_name);
        local_free_the_blob(blob);
        return;
    }

    /* Disable logging to suppress the memory management warnings */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* This initializes the memory management and registers buffers through
       the algorithms pre-process routine.  This algorithm doesn't need any
       memory management buffers that require the list of bands to be
       initialized -- number of bands set to 0 for that reason. */
    blob->number_of_bands = 0;
    blob->band_fields_initialized = 1;
    status = initialize_memory_management_buffers();
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error initializing memory management");
        local_free_the_blob(blob);
        return;
    }

    /* Initialize the RPS process level arrays */
    status = initialize_rps_level(30, 14);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error calling initialize_rps_level");
        cleanup_io_routines(blob);
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        free_the_blob();
        return;
    }

    status = tirs_bias_removal_preprocess(blob);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error returned from: "
            "tirs_bias_removal_preprocess");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    /* Register the L1R buffer to this routine so that the API can't
       destroy the memory before we can retrieve the output */
    status = increment_consumer_count(IMAGE_L1R);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error returned from: "
            "increment_consumer_count IMAGE_L1R");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }
    /* It is important that the consumer counts get set *after* all the
       increments */
    set_consumer_counts();

    /* Check an "empty" buffer for the dark response, the TIRS band
       is the second one in the list */
    dark_buffer = (double *)get_memory_buffer( MM_TIRS_DARK_RESPONSE, 
        SETTER, list_of_band_indices[1], 0);
    if (!dark_buffer)
    {
        assert_true_with_message(0, "Error getting "
            "MM_TIRS_DARK_RESPONSE buffer");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    /* Open and read the dark response buffer from dark response data file */
    sprintf(file_name, "%s/%s/%s", test_directory, data_dir, dark_file);

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    for (det = 0; det < NDETS; det++)
    {
        return_value = fscanf(inf, "%f", &temp_float);
        if (return_value != 1)
        {
            assert_true_with_message(0, "Reading %s", file_name);
            cleanup_rps_level();
            cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
            local_free_the_blob(blob);
            return;        
        } 
        dark_buffer[det] = (double)temp_float;       
    }
    fclose(inf);
    set_buffer_filled_flag( MM_TIRS_DARK_RESPONSE, list_of_band_indices[1],
        0);

    /* Check an "empty" buffer for the background response, the TIRS band
       is the second one in the list */
    background_buffer = (double *)get_const_memory_buffer(
        MM_TIRS_BACKGROUND_RESPONSE, list_of_band_indices[1], 0);
    if (!background_buffer)
    {
        assert_true_with_message(0, "Error getting "
            "MM_TIRS_BACKGROUND_RESPONSE buffer");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    /* Open and read background buffer */
    sprintf(file_name, "%s/%s/%s", test_directory, data_dir, background_file);

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    for (det = 0; det < NDETS; det++)
    {
        return_value = fscanf(inf, "%f", &temp_float);
        if (return_value != 1)
        {
            assert_true_with_message(0, "Reading %s", file_name);
            cleanup_rps_level();
            cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
            local_free_the_blob(blob);
            return;        
        } 
        background_buffer[det] = (double)temp_float;       
    }
    fclose(inf);

    /* Get the image buffer for band 11, this simply returns an unfilled,
       allocated buffer, band 11 is second in the list */
    l1r_buffer = (float *)get_memory_buffer(IMAGE_L1R, SETTER,
        list_of_band_indices[1], 0);
    if (!l1r_buffer)
    {
        assert_true_with_message(0, "Getting IMAGE_L1R buffer");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    /* Open and read the data buffer */
    sprintf(file_name, "%s/%s/%s", test_directory, data_dir, data_file);

    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    for (det = 0; det < NDETS; det++)
    {
        for (line = 0; line < NLINES; line++)
        {
            return_value = fscanf(inf, "%f", &temp_float);
            if (return_value != 1)
            {
                assert_true_with_message(0, "Reading %s", file_name);
                cleanup_rps_level();
                cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
                local_free_the_blob(blob);
                return;        
            } 
            l1r_buffer[det * NLINES + line] = temp_float;       
        }
    }
    fclose(inf);
    set_buffer_filled_flag( IMAGE_L1R, list_of_band_indices[1], 0);

    /* Call the API with a non-TIRS band.  The API should return immediately.
       The only way to really check this is to ensure the data buffer wasn't
       changed at all. */
    band_number = list_of_bands[0];
    band_index = list_of_band_indices[0];
    status = api_tirs_bias_removal( blob, band_index, 0, 0 );
    assert_equal_with_message( status,SUCCESS, "Error returned from "
        "api_bias_removal with band = 1");
    if ( status != SUCCESS )
    {
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob( blob );
        return;
    }

    /* Check some outputs with the original inputs, no changes */
    assert_double_equal(l1r_buffer[0], 4326.954102);
    assert_double_equal(l1r_buffer[640], 4330.139648);
    assert_double_equal(l1r_buffer[640*2], 4326.954102);

    /* Call the API */
    band_number = list_of_bands[1];
    band_index = list_of_band_indices[1];
    status = api_tirs_bias_removal(blob, band_index, 0, 0);
    assert_equal_with_message(status, SUCCESS, "Error returned from "
        "api_tirs_bias_removal with band = 11");
    if (status != SUCCESS)
    {
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    /* Open and read the results file */
    sprintf(file_name, "%s/%s/%s", test_directory, data_dir, results_file5);
    results_buffer = malloc(NDETS * NLINES * sizeof(float));
    inf = fopen(file_name, "r");
    if (!inf)
    {
        assert_true_with_message(0, "Opening %s", file_name);
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    for (det = 0; det < NDETS; det++)
    {
        for (line = 0; line < NLINES; line++)
        {
            return_value = fscanf(inf, "%f", &temp_float);
            if (return_value != 1)
            {
                assert_true_with_message(0, "Reading %s", file_name);
                cleanup_rps_level();
                cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
                local_free_the_blob(blob);
                return;        
            } 
            results_buffer[det * NLINES + line] = temp_float;       
        }
    }
    fclose(inf);

    /* Compare the results.  Loop through the results and count the
       differences.  This comparison isn't done with asserts since there is
       no need for 320000 assert statements which cost message buffer space.
       One assert for pass fail with a count is good enough. */
    difference_count = 0;
    for (pixel = 0; pixel < NDETS * NLINES; pixel++)
    {
        if (fabsf(l1r_buffer[pixel] - results_buffer[pixel]) > 1e-3)
            difference_count++;
    }

    assert_equal_with_message(difference_count, 0, "There were %d "
        "differences out of %d values", difference_count, NDETS*NLINES);

    /* Free the image memory */
    status = decrement_consumer_count(IMAGE_L1R, list_of_band_indices[1], 0);
    assert_equal_with_message( status, SUCCESS,
        "Error returned from decrement consumer count");

    /* Due to the lack of the real CPF used here, the 
       MM_TIRS_BACKGROUND_RESPONSE memory is decreased here to avoid memory
       loss in the test code */
    status = decrement_consumer_count(MM_TIRS_BACKGROUND_RESPONSE,
        list_of_band_indices[1], 0);
    assert_equal_with_message( status, SUCCESS,
        "Error returned from decrement consumer count");

    free(results_buffer);
    cleanup_rps_level();
    cleanup_memory_management_buffers(ENABLE_WARNING_DISPLAY);
    local_free_the_blob(blob);
} 


