/**************************************************************************
 NAME:      tirs_dark_response_determination_test

 PURPOSE:   Perform CGreen unit testing of the TIRS dark response
            determination algorithm flow as implemented for per-scene
            processing through the RPS driver

 RETURNS:   CGreen unit test status

 ROUTINES:  local_create_blob_pixel_masks (internal routine)
            local_free_the_blob           (internal routine)
            local_allocate_the_blob       (internal routine)
            check_group_files             (internal routine)
            test_algorithm_api_blind_band_char
            test_algorithm_api_image_band_char

 NOTES:     The core algorithm (ias_rps_tirs_dark_response_determination)
            has been tested in the RPS library.

            The test data used here is based on the test data set used
            to prototype the dark response determination algorithm.  The
            two "rows" of baseline dark response and relative gain data in
            each band/SCA were averaged to obtain an overall baseline
            dark response and relative gain for each detector.

            DO NOT run test_algorithm_api_image_band_char BEFORE
            test_algorithm_api_blind_band_char!  The image band
            characterization requires three local files produced
            (one for each SCA) by the blind band characterization
            which contain an average dark response difference
            needed to adjust the baseline dark response.

***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include "ias_const.h"
#include "ias_pixel_mask.h"
#include "ias_miscellaneous.h"
#include "ias_satellite_attributes.h"
#include "ias_logging.h"
#include "cgreen.h"
#include "../rps_local.h"
#include "../rps_blob.h"
#include "ias_rps_histogram_statistics_characterization.h"
#include "ias_rps_tirs_dark_response_determination.h"


#define NUM_LINES 100

#ifndef FALSE
#define FALSE 0
#endif

#ifndef TRUE
#define TRUE 1
#endif


/* Test data directory location strings and command string buffer used in
   multiple tests */
static char *test_data_path = NULL;
static char test_data_directory[] = "ias_base/rps/driver/"
    "tirs_dark_response_determination_driver";
static char cpf_data_directory[] = "ias_lib/cpf";


/*------------------------------------------------------------------------
 NAME:     local_create_blob_pixel_masks

 PURPOSE:  Allocate the BLOB pixel mask data structure for each
           band

 RETURNS:  Integer status code of SUCCESS or ERROR

--------------------------------------------------------------------------*/
static int local_create_blob_pixel_masks(BLOB *blob)
{
    int band_number;                    /* Band loop counter */
    int band_index;
    int sca = 0;                        /* SCA loop counter */
    int sca_number;
    int num_detectors;
    int num_scas;
    int num_lines;


    for (band_index = 0; band_index < blob->total_bands; band_index++)
    {
        band_number = blob->band_information[band_index]->band_number;
        num_scas = blob->band_information[band_index]->number_scas;
        num_detectors = blob->band_information[band_index]->number_detectors;
        num_lines = blob->band_information[band_index]->number_lines;

        blob->pixel_mask[band_index] = calloc(num_scas,
            sizeof(IAS_PIXEL_MASK *));
        if (blob->pixel_mask[band_index] == NULL)
        {
            printf("Error allocating pixel mask array for band %d",
                band_number);
            return ERROR;
        }
        for (sca = 0; sca < num_scas; sca++)
        {
            sca_number = sca + 1;
            blob->pixel_mask[band_index][sca] = ias_pm_create(
                band_number, sca_number, num_detectors, num_lines);
            if (blob->pixel_mask[band_index][sca] == NULL)
            {
                printf("Error creating empty pixel mask for band %d SCA %d",
                    band_number, sca_number);
                return ERROR;
            }
        }
    }

    /* Done */
    return SUCCESS;
}   /* END -- local_create_blob_pixel_masks */



/*--------------------------------------------------------------------------
  NAME:               local_free_the_blob

  PURPOSE:  This a local version of the routine to free the blob.  Since
            not all members of the blob may be initialized, the  normal
            routine should not be called.
--------------------------------------------------------------------------*/
static void local_free_the_blob(BLOB *blob)
{
    int band_index;                           /* Band loop counter */
    int sca;                            /* SCA loop counter */
    int num_scas;


    /* Free the pixel mask array */
    if (blob->pixel_mask != NULL)
    {
        for (band_index = 0; band_index < blob->total_bands; band_index++)
        {
            num_scas = blob->band_information[band_index]->number_scas;
            if (blob->pixel_mask[band_index] != NULL)
            {
                for (sca = 0; sca < num_scas; sca++)
                {
                    if (blob->pixel_mask[band_index][sca] != NULL)
                        ias_pm_destroy(blob->pixel_mask[band_index][sca]);
                }
                free(blob->pixel_mask[band_index]);
            }
        }
        free(blob->pixel_mask);
    }

    /* Free the band information array */
    if (blob->band_information != NULL)
    {
        for (band_index = 0; band_index < blob->total_bands; band_index++)
        {
            if (blob->band_information[band_index] != NULL)
                free(blob->band_information[band_index]);
        }
        free(blob->band_information);
    }

    /* If we have CPF data, free those references */
    if (blob->cpf != NULL)
        ias_cpf_free(blob->cpf);

    /* Free the proc level string */
    free(blob->database_information.rps_level);

    /* Free the blob band list */
    free(blob->list_of_bands);

    /* Free the base data structure */
    free(blob);
}



/*------------------------------------------------------------------------
 NAME:     local_allocate_the_blob

 PURPOSE:  Allocates and populates a BLOB data structure

 RETURNS:  Allocated and populated BLOB data structure if successful,
           NULL pointer if there's an error

--------------------------------------------------------------------------*/
static BLOB *local_allocate_the_blob()
{
    BLOB *blob = NULL;              /* BLOB data structure pointer to
                                       return */
    int band_number;                /* Band loop counter */
    int band_index;
    int ii;                         /* General loop counter */
    int vrp_band_number;
    int blind_band_number;
    int status = SUCCESS;

    const IAS_BAND_ATTRIBUTES *band_atts = NULL;

    char cpf_basename[] = "eval_L8CPF20090101_20090331.01";


    /* Allocate the blob for the purposes of this test */
    blob = allocate_blob();
    if (blob == NULL)
    {
        assert_true_with_message(FALSE, "Cannot allocate blob data block");
        return NULL;
    }

    /* Start populating the BLOB */

    /* Set sensor information */
    blob->max_number_scas = 14;
    blob->total_bands = 30;
    blob->band_information = calloc(30, sizeof (RPS_BAND_ATTRIBUTES *));
    assert_not_equal_with_message(blob->band_information, NULL, "Cannot "
        "allocate blob band information data block");
    if (blob->band_information == NULL)
    {
        free(blob);
        return NULL;
    }
    for (band_number = 1; band_number <= blob->total_bands; band_number++)
    {
        int blind_band_index;
        band_atts = ias_sat_attr_get_band_attributes(band_number);
        band_index = band_atts->band_index;

        blob->band_information[band_index] = calloc(1,
            sizeof(RPS_BAND_ATTRIBUTES));
        assert_not_equal_with_message(blob->band_information[band_index],
            NULL, "Cannot allocate band information element for band %d",
            band_number);
        if (blob->band_information[band_index] == NULL)
        {
            assert_true_with_message(FALSE, "Error allocating band "
                "information element for band %d", band_number);
            for (ii = 0; ii < band_index; ii++)
                free(blob->band_information[ii]);
            free(blob->band_information);
            free(blob);
            return NULL;
        }

        /* Get the index of the blind band no need to error check since the
           value is just assigned to the band information as is.  Logging is
           also disabled since the satellite library throws a warning for
           most of these calls. */
        ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
        blind_band_index = ias_sat_attr_convert_band_number_to_index(
            band_atts->blind_band_number );
        ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

        /* Get IAS attributes for the current band number */
        vrp_band_number = band_atts->vrp_band_number;
        blind_band_number = band_atts->blind_band_number;
        blob->band_information[band_index]->band_number = band_number;
        blob->band_information[band_index]->band_present = 1;
        blob->band_information[band_index]->vrp_band_number =
            vrp_band_number;
        blob->band_information[band_index]->blind_band_number =
            blind_band_number;
        blob->band_information[band_index]->blind_band_index =
            blind_band_index;
        blob->band_information[band_index]->number_lines = NUM_LINES;
        blob->band_information[band_index]->number_detectors =
            band_atts->detectors_per_sca;
        blob->band_information[band_index]->number_scas = band_atts->scas;
        blob->band_information[band_index]->band_classification =
            band_atts->band_classification;
        if (band_atts->band_classification & IAS_NORMAL_BAND)
            blob->band_information[band_index]->is_normal_band = 1;
        else
            blob->band_information[band_index]->is_normal_band = 0;
        blob->band_information[band_index]->spectral_type =
            band_atts->spectral_type;
    }

    /* Create a set of pixel masks.  Assume all detectors are perfectly
       operable */
    blob->pixel_mask = calloc(blob->total_bands, sizeof(IAS_PIXEL_MASK **));
    if (blob->pixel_mask == NULL)
    {
        assert_true_with_message(FALSE, "Cannot allocate blob pixel mask "
            "base array");
        local_free_the_blob(blob);
        return NULL;
    }
    status = local_create_blob_pixel_masks(blob);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error creating blob pixel masks");
        local_free_the_blob(blob);
        return NULL;
    }

    /* Open the CPF */
    sprintf(blob->cpf_filename, "%s/%s/%s", test_data_path,
        cpf_data_directory, cpf_basename);
    blob->cpf = ias_cpf_read(blob->cpf_filename);
    assert_not_equal_with_message(blob->cpf, NULL, "Error loading data "
        "from CPF %s", blob->cpf_filename);
    if (blob->cpf == NULL)
    {
        local_free_the_blob(blob);
        return NULL;
    }


    /* Allocate the RPS proc level string buffer, and set it to '0' */
    blob->database_information.rps_level = strdup("0");
    assert_not_equal_with_message(blob->database_information.rps_level,
        NULL, "Cannot set RPS processing level information in BLOB");
    if (blob->database_information.rps_level == NULL)
    {
        local_free_the_blob(blob);
        return NULL;
    }

    /* Allocate space for the band list and indices */
    blob->list_of_bands = calloc(IAS_MAX_NBANDS,
        sizeof(*(blob->list_of_bands)));
    blob->list_of_band_indices = calloc(IAS_MAX_NBANDS,
        sizeof(*(blob->list_of_band_indices)));
    if (blob->list_of_bands == NULL || blob->list_of_band_indices == NULL)
    {
        assert_true_with_message(FALSE, "Cannot allocate BLOB band lists "
            "array");
        local_free_the_blob(blob);
        return NULL;
    }

    /* Done */
    return blob;
}



/*-------------------------------------------------------------------------
 NAME:                 check_group_files

 PURPOSE:  Internal routine used to compare an adjusted dark response
           output file produced by the TIRS dark response determination
           algorithm to a "reference" output file

 RETURNS:  Integer status code of SUCCESS or ERROR.  ERROR is only
           returned if the test and/or reference file cannot be opened
           or a line cannot be read

--------------------------------------------------------------------------*/
static int check_group_files(const char *test_filename,
    const char *reference_group_filename, double pass_threshold)
{
    char test_buff[1000] = "";             /* String buffer used for
                                              "test" data file read */
    char reference_buff[1000] = "";        /* String buffer used for
                                              "reference" data file read */

    FILE *test_fptr = NULL;                /* "Test" data file */
    FILE *reference_fptr = NULL;           /* "reference data file for
                                              comparison */

    double diff = 0.0;                     /* Absolute difference between
                                              corresponding test and
                                              reference data */
    double test_response[5] = {0.0, 0.0, 0.0, 0.0, 0.0};
                                           /* Array of test dark responses
                                              from a line in the test
                                              output file */
    double reference_response[5] = {0.0, 0.0, 0.0, 0.0, 0.0};
                                           /* Array of reference dark
                                              responses from a line in the
                                              reference output file */

    int ii = 0;                            /* Loop counter */
    int jj = 0;                            /* Loop counter */
    int response_diffs = 0;                /* # of differences in dark
                                              response (for a given detector
                                              between the "test" data and the
                                              "reference" data) exceeding
                                              the specified threshold
                                              value*/
    int status = SUCCESS;


    /* Open the group files */
    test_fptr = fopen(test_filename, "r");
    reference_fptr = fopen(reference_group_filename, "r");

    assert_not_equal_with_message(test_fptr, NULL, "Cannot open test "
        "group file %s", test_filename);
    assert_not_equal_with_message(reference_fptr, NULL, "Cannot open "
        "reference group file %s", reference_group_filename);
    if ((test_fptr == NULL) || (reference_fptr == NULL))
    {
        if (test_fptr)
            fclose(test_fptr);
        if (reference_fptr)
            fclose(reference_fptr);
        return ERROR;
    }

    /* Check the GROUP header and parameter names to make sure the
       test one is correctly written */
    fgets(test_buff, 1000, test_fptr);
    fgets(reference_buff, 1000, reference_fptr);
    assert_not_equal_with_message(test_buff, NULL, "Cannot populate "
        "test string buffer");
    assert_not_equal_with_message(reference_buff, NULL, "Cannot populate "
        "reference string buffer");
    if ((test_buff == NULL) || (reference_buff == NULL))
    {
        fclose(test_fptr);
        fclose(reference_fptr);
        return ERROR;
    }
    assert_string_equal_with_message(test_buff, reference_buff,
        "GROUP header statement difference");

    fgets(test_buff, 1000, test_fptr);
    fgets(reference_buff, 1000, reference_fptr);
    assert_not_equal_with_message(test_buff, NULL, "Cannot populate "
        "test string buffer");
    assert_not_equal_with_message(reference_buff, NULL, "Cannot populate "
        "reference string buffer");
    if ((test_buff == NULL) || (reference_buff == NULL))
    {
        fclose(test_fptr);
        fclose(reference_fptr);
        return ERROR;
    }
    assert_string_equal_with_message(test_buff, reference_buff,
        "Parameter name difference");

    /* Check each line of data.  As currently implemented,
       there are 5 dark response values per line.  For 640
       detectors, that translates into 128 lines of dark response
       values */
    for (ii = 0; ii < 128; ii++)
    {
        status = fscanf(test_fptr,
            "%lf,%lf,%lf,%lf,%lf%*[ ,)]\n",
            &test_response[0], &test_response[1],
            &test_response[2], &test_response[3],
            &test_response[4]);
        if (status < 0)
        {
            assert_true_with_message(FALSE, "Cannot read test line %d",
                (ii + 1));
            fclose(test_fptr);
            fclose(reference_fptr);
            return ERROR;
        }

        status = fscanf(reference_fptr,
            "%lf,%lf,%lf,%lf,%lf%*[ ,)]\n",
            &reference_response[0], &reference_response[1],
            &reference_response[2], &reference_response[3],
            &reference_response[4]);
        if (status < 0)
        {
            assert_true_with_message(FALSE, "Cannot read test line %d",
                (ii + 1));
            fclose(test_fptr);
            fclose(reference_fptr);
            return ERROR;
        }

        /* Check the absolute difference for each set of dark responses.
           Count those that are greater than the passing threshold */
        for (jj = 0; jj < 5; jj++)
        {
            diff = fabs(test_response[jj] - reference_response[jj]);
            if (diff > pass_threshold)
            {
                response_diffs++;
            }
        }
    }

    assert_equal_with_message(response_diffs, 0,
        "%d differences of %lf or greater detected", response_diffs,
            pass_threshold);

    /* Check the END_GROUP and END statements */
    fgets(test_buff, 1000, test_fptr);
    fgets(reference_buff, 1000, reference_fptr);
    assert_not_equal_with_message(test_buff, NULL, "Cannot populate "
        "test string buffer");
    assert_not_equal_with_message(reference_buff, NULL, "Cannot populate "
        "reference string buffer");
    if ((test_buff == NULL) || (reference_buff == NULL))
    {
        fclose(test_fptr);
        fclose(reference_fptr);
        return ERROR;
    }
    assert_string_equal_with_message(test_buff, reference_buff,
        "END_GROUP header statement difference");

    fgets(test_buff, 1000, test_fptr);
    fgets(reference_buff, 1000, reference_fptr);
    assert_not_equal_with_message(test_buff, NULL, "Cannot populate "
        "test string buffer");
    assert_not_equal_with_message(reference_buff, NULL, "Cannot populate "
        "reference string buffer");
    if ((test_buff == NULL) || (reference_buff == NULL))
    {
        fclose(test_fptr);
        fclose(reference_fptr);
        return ERROR;
    }
    assert_string_equal_with_message(test_buff, reference_buff,
        "END closing statement difference");

    /* Close the files */
    fclose(test_fptr);
    fclose(reference_fptr);

    /* Done */
    return SUCCESS;
}



/************************ API TESTS **************************************/
Ensure test_algorithm_api_blind_band_char()
{
    BLOB *blob = NULL;

    const RPS_BAND_ATTRIBUTES *rps_atts = NULL;
    IAS_RPS_HISTOGRAM_INTERMEDIATE_RESULTS *test_blind_stats = NULL;

    /* Names of data files containing "test" input data and "reference"
       data used for comparison */
    char test_filename[PATH_MAX] = "";
    char reference_group_filename[PATH_MAX] = "";

    /* File name for test blind band histogram statistics data */
    char test_stats_filename[PATH_MAX] = "";
    int blind_band_number;         /* Current blind band number */
    int band_number = 10;          /* Current band number */
    int sca_number = 1;            /* Current (1-based) SCA number */
    int num_scas = 3;
    int number_detectors = 640;
    int detector_index = 0;        /* Detector loop counter */
    int status = SUCCESS;
    int sca_index = 0;             /* SCA loop counter */
    int ii = 0;                    /* Generic loop counter */

    FILE *test_fptr = NULL;        /* Open test data file */

    struct stat buff;              /* Used for finding the
                                      concatenated file */

    double value = 0.0;            /* Data value read from file */
    int band_index;
    int blind_band_index;


    printf("Testing TIRS blind band characterization flow...\n");

    /* Allocate/populate the BLOB */
    blob = local_allocate_the_blob();
    if (blob == NULL)
    {
        assert_true_with_message(FALSE, "Cannot allocate/populate BLOB "
            "structure");
        return;
    }

    /* Get the band index -- this call does need to be error checked */
    band_index = ias_sat_attr_convert_band_number_to_index( band_number );
    if ( band_index == ERROR )
    {
        assert_true_with_message(FALSE, "Getting the band index for band %d",
            band_number );
        return;
    }

    /* Set the flags to process the TIRS blind band and write the
       adjusted dark response CPF groups to file */
    blob->use_tirs_cpf_baseline_response = 0;
    blob->process_tirs_blind_band = 1;
    blob->write_tirs_dark_response_to_file = 1;

    /* Initialize the blob band list and get the blind band number */
    blob->list_of_bands[0] = band_number;
    blob->list_of_band_indices[0] = band_index;
    blob->number_of_bands = 1;
    blob->band_fields_initialized = 1;
    rps_atts = get_band_attributes(band_index);
    blind_band_number = rps_atts->blind_band_number;
    blind_band_index = rps_atts->blind_band_index;

    /* Initialize memory management and register buffers */
    status = initialize_memory_management_buffers();
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error initializing memory management");
        local_free_the_blob(blob);
        return;
    }

    /* Call the preprocess routine to set up memory management */
    status = tirs_dark_response_determination_preprocess(blob);
    assert_equal_with_message(status, SUCCESS, "Blind band pre-processing "
        "failed");
    if (status != SUCCESS)
    {
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }

    status = increment_consumer_count(MM_L0R_INTERMEDIATE_STATISTICS);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Cannot get proper increment count "
            "for blind band stats data buffer");
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }
    /* It is important that the consumer counts get set *after* all the
       increments */
    set_consumer_counts();

    for (sca_index = 0; sca_index < num_scas; sca_index++)
    {
        sca_number = (sca_index + 1);

        test_blind_stats = 
            (IAS_RPS_HISTOGRAM_INTERMEDIATE_RESULTS *)get_memory_buffer(
                MM_L0R_INTERMEDIATE_STATISTICS, SETTER, blind_band_index,
                sca_index);
        assert_not_equal_with_message(test_blind_stats, NULL,
            "Cannot obtain intermediate histogram results data buffer");
        if (test_blind_stats == NULL)
        {
            status = decrement_consumer_count(MM_L0R_INTERMEDIATE_STATISTICS,
                blind_band_index, sca_index);
            assert_equal_with_message( status, SUCCESS,
                "Error returned from decrement consumer count");
            cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
            local_free_the_blob(blob);
            return;
        }

        /* Open the test data and read in the prototype test data */
        sprintf(test_stats_filename, "%s/%s/sc_B%02d_SCA%02d.txt",
           test_data_path, test_data_directory, blind_band_number,
           sca_number);
        test_fptr = fopen(test_stats_filename, "r");
        assert_not_equal_with_message(test_fptr, NULL, "Cannot open "
           "test blind band scene stats file %s", test_stats_filename);
        if (test_fptr == NULL)
        {
            status = decrement_consumer_count(MM_L0R_INTERMEDIATE_STATISTICS,
                blind_band_index, sca_index);
            assert_equal_with_message( status, SUCCESS,
                "Error returned from decrement consumer count");
            cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
            local_free_the_blob(blob);
            return;
        }
        for (detector_index = 0;
             detector_index < number_detectors;
             detector_index++)
        {
            fscanf(test_fptr, "%lf\n", &value);
            test_blind_stats[detector_index].mean[0] = value;
        }
        fclose(test_fptr);
        set_buffer_filled_flag( MM_L0R_INTERMEDIATE_STATISTICS,
            blind_band_index, sca_index );

        /* Call the API.  Make sure we're set to calculate the adjusted
          dark response */
        if (!blob->use_tirs_cpf_baseline_response)
        {
            status = api_tirs_dark_response_determination(blob,
                blind_band_index, sca_number-1, 0);
            if (status != SUCCESS)
            {
                assert_true_with_message(FALSE, "API call failed:  SCA %d",
                    sca_number);
                for (ii = 0; ii < num_scas; ii++)
                {
                    status = decrement_consumer_count(
                        MM_TIRS_SCA_DARK_RESPONSE_DIFFERENCE,
                        blind_band_index, ii);
                    assert_equal_with_message( status, SUCCESS,
                        "Error returned from decrement consumer count");
                }
                status = decrement_consumer_count(
                    MM_L0R_INTERMEDIATE_STATISTICS,
                    blind_band_index, sca_index);
                assert_equal_with_message( status, SUCCESS,
                    "Error returned from decrement consumer count");
                cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
                local_free_the_blob(blob);
                return;
            }
        }

        /* Does the test output file for the current SCA exist? */
        sprintf(test_filename,
            "cpf_dark_response_group_B%02d_SCA%02d.odl",
            blind_band_number, sca_number);
        status = stat(test_filename, &buff);
        if (status != 0)
        {
            assert_true_with_message(FALSE, "Dark response output file for "
                "SCA %d does not exist", sca_number);
            for (ii = 0; ii < num_scas; ii++)
            {
                status = decrement_consumer_count(
                    MM_TIRS_SCA_DARK_RESPONSE_DIFFERENCE,
                    blind_band_index, ii);
                assert_equal_with_message( status, SUCCESS,
                    "Error returned from decrement consumer count");
            }
            status = decrement_consumer_count(MM_L0R_INTERMEDIATE_STATISTICS,
                blind_band_index, sca_index);
            assert_equal_with_message( status, SUCCESS,
                "Error returned from decrement consumer count");
            cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
            local_free_the_blob(blob);
            return;
        }

        sprintf(reference_group_filename,
            "%s/%s/cpf_adrgroup_B%02d_SCA%02d.odl",
            test_data_path, test_data_directory, blind_band_number,
            sca_number);
        status = check_group_files(test_filename, reference_group_filename,
            1e-3);
        if (status != SUCCESS)
        {
            assert_true_with_message(FALSE, "Cannot complete check for band "
                " %d SCA %d", blind_band_number, sca_number);
        }

        /* Decrement the consumer count for the current SCA */
        status = decrement_consumer_count(MM_L0R_INTERMEDIATE_STATISTICS,
            blind_band_index, sca_index);
        assert_equal_with_message( status, SUCCESS,
            "Error returned from decrement consumer count");
    }

    /* Write the SCA dark response data to the files, then check to make 
       sure the files were created */
    status = tirs_dark_response_determination_postprocess(blob);
    assert_equal_with_message(status, SUCCESS, "Could not run blind "
        "band postprocessing");

    if (status == SUCCESS)
    {
        for (sca_index = 0; sca_index < num_scas; sca_index++)
        {
            sca_number = (sca_index + 1);

            sprintf(test_filename, "response_diff_SCA%02d.txt",
                sca_number);
            status = stat(test_filename, &buff);
            if (status != 0)
            {
                assert_true_with_message(FALSE, "Cannot find response "
                    "difference file for SCA %d", sca_number);
            }
        }
    }

    /* Do memory management cleanup.  The warning is disabled for this
       cleanup since it's known that SCA 3 wasn't decremented above and the
       blind band wasn't processed. */
    cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);

    /* Do final cleanup of the blob data */
    local_free_the_blob(blob);
}




Ensure test_algorithm_api_image_band_char()
{
    BLOB *blob = NULL;
    const RPS_BAND_ATTRIBUTES *rps_atts = NULL;

    /* Names of data files containing "test" input data and "reference"
       data used for comparison */
    char test_filename[PATH_MAX] = "";
    char reference_group_filename[PATH_MAX] = "";
    int blind_band_index;         /* Current blind band index */
    int band_number = 10;          /* Current band number */
    int sca_number = 1;            /* Current (1-based) SCA number */
    int num_scas = 3;
    int status = SUCCESS;
    int sca_index = 0;             /* SCA loop counter */
    int ii = 0;                    /* Generic loop counter */
    int band_index;


    printf("Testing TIRS image band characterization flow...\n");

    /* Allocate/populate the BLOB */
    blob = local_allocate_the_blob();
    if (blob == NULL)
    {
        assert_true_with_message(FALSE, "Cannot allocate/populate BLOB "
            "structure");
        return;
    }

    /* Get the band index -- this call does need to be error checked */
    band_index = ias_sat_attr_convert_band_number_to_index( band_number );
    if ( band_index == ERROR )
    {
        assert_true_with_message(FALSE, "Getting the band index for band %d",
            band_number );
        return;
    }

    /* Set the flags to NOT process the TIRS blind band, but write the
       adjusted dark response CPF groups to file */
    blob->process_tirs_blind_band = 0;
    blob->use_tirs_cpf_baseline_response = 0;
    blob->write_tirs_dark_response_to_file = 1;

    /* Initialize the blob band list and get the blind band number */
    blob->list_of_bands[0] = band_number;
    blob->list_of_band_indices[0] = band_index;
    blob->number_of_bands = 1;
    blob->band_fields_initialized = 1;
    rps_atts = get_band_attributes(band_index);
    blind_band_index = rps_atts->blind_band_index;

    /* Initialize memory management and register buffers */
    status = initialize_memory_management_buffers();
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Initializing memory management");
        local_free_the_blob(blob);
        return;
    }

    /* Call the preprocess routine to set up memory management */
    status = tirs_dark_response_determination_preprocess(blob);
    assert_equal_with_message(status, SUCCESS, "Band %d pre-processing "
        "failed", band_number);
    if (status != SUCCESS)
    {
        for (sca_index = 0; sca_index < num_scas; sca_index++)
        {
            sca_number = (sca_index + 1);
            status = decrement_consumer_count(
                MM_TIRS_SCA_DARK_RESPONSE_DIFFERENCE,
                blind_band_index, sca_index);
            assert_equal_with_message( status, SUCCESS,
                "Error returned from decrement consumer count");
        }
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        local_free_the_blob(blob);
        return;
    }
    /* It is important that the consumer counts get set *after* all the
       increments */
    set_consumer_counts();

    /* Call the algorithm */
    if (!blob->use_tirs_cpf_baseline_response)
    {
        /* SCA 3's test data was found to have a problem with the
           relative gain for detector 289--one of the two rows had a
           negative gain.  The average gain value was garbage.  The overall 
           effect is that the adjusted dark responses calculated for
           the remaining detectors differ from the estimated adjusted
           response by more than the threshold value, failing the test */
        for (sca_index = 0; sca_index < 2; sca_index++)
        {
            sca_number = (sca_index + 1);

            status = api_tirs_dark_response_determination(blob,
                band_index, sca_number-1, 0);
            assert_equal_with_message(status, SUCCESS, "API call failed");
            if (status != SUCCESS)
            {
                for (ii = 0; ii < num_scas; ii++)
                {
                    status = decrement_consumer_count(
                        MM_TIRS_SCA_DARK_RESPONSE_DIFFERENCE, blind_band_index,
                        ii);
                    assert_equal_with_message( status, SUCCESS,
                            "Error returned from decrement consumer count");
                }
                cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
                free_the_blob();
                return;
            }

            /* Compare the output for the band 10 test file against the
               corresponding "reference" file */
            sprintf(test_filename,
                "cpf_dark_response_group_B%02d_SCA%02d.odl",
                band_number, sca_number);
            sprintf(reference_group_filename,
                "%s/%s/cpf_adrgroup_B%02d_SCA%02d.odl",
                test_data_path, test_data_directory, band_number,
                sca_number);

            status = check_group_files(test_filename,
                reference_group_filename, 1e-3);
            if (status != SUCCESS)
            {
                assert_true_with_message(FALSE, "Cannot complete check for "
                    "band %d SCA %d", band_number, sca_number);
            }
        }
    }

    /* Decrement the remaining consumer counts for the SCA dark response
       difference */
    for (sca_index = 0; sca_index < num_scas; sca_index++)
    {
        sca_number = (sca_index + 1);
        status = decrement_consumer_count(
            MM_TIRS_SCA_DARK_RESPONSE_DIFFERENCE, blind_band_index,
            sca_index);
        assert_equal_with_message( status, SUCCESS,
            "Error returned from decrement consumer count");
    }

    /* Do memory management cleanup.  The warning is disabled for this
       cleanup since it's known that SCA 3 wasn't decremented above and the
       blind band wasn't processed. */
    cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);

    /* Do final cleanup of the blob data */
    local_free_the_blob(blob);
}   /* END -- test_algorithm_api_image_band_char */




/************************* MAIN TEST PROGRAM *******************************/
int main(int argc, char *argv[])
{

    /* Get the test data directory */
    test_data_path = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (test_data_path == NULL)
    {
        printf("Cannot find test data directory path\n");
        exit(100);
    }

    /* Initialize the satellite attributes so that memory management
       works properly */
    ias_sat_attr_initialize(IAS_L8);

    /* Create the test suite */
    TestSuite *suite = create_test_suite();

    add_test(suite, test_algorithm_api_blind_band_char);
    add_test(suite, test_algorithm_api_image_band_char);

    if (argc > 1)
        return run_single_test(suite, argv[1], create_text_reporter());
    else
        return run_test_suite(suite, create_text_reporter());

}
