/*--------------------------------------------------------------------------
NAME:    tirs_secondary_linearization_test

PURPOSE: Test the TIRS secondary linearization API.  The algorithm is
         located in the library so it isn't tested here.
--------------------------------------------------------------------------*/
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "ias_satellite_attributes.h"
#include "cgreen/cgreen.h"
#include "ias_logging.h"
#include "ias_radiometry.h"
#include "rps_local.h"

#define NLINES 64
#define NDETS  640

static char test_subdir[] = "ias_lib/rps/lookup_response_linearization";
static char rlut_file[] = "eval_L8RLUT20130211_20431231v07.h5";

Ensure test_algorithm_thru_api(void);

int main( int argc, char *argv[] )
{
    TestSuite *suite = create_test_suite();
    add_test(suite, test_algorithm_thru_api);

    if (argc > 1) {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}


/*--------------------------------------------------------------------------
  local_free_the_blob

  This a local override of the routine to free the blob.  Since not all
  members of the blob are initialized -- specifically the pixel mask -- the
  normal routine cannot be called.
--------------------------------------------------------------------------*/
static void local_free_the_blob(BLOB *blob)
{
    int i;
    
    if ( !blob )
        return;

    if ( blob->l1r_band_metadata )
    {
        for ( i = 0; i < blob->total_bands; i++ )
            free( blob->l1r_band_metadata[i] );
        free( blob->l1r_band_metadata );
    }

    if ( blob->band_information )
    {
        for ( i = 0; i < blob->total_bands; i++ )
        {
            if ( blob->band_information[i] )
                free( blob->band_information[i] );
        }
    }
    free( blob->band_information );

    if ( blob->cpf )
        ias_cpf_free( blob->cpf );
    free( blob );
}

/*--------------------------------------------------------------------------
  test_algorithm_thru_api

  This test sets up for a call to the algorithm through the API.  Also calls
  the pre-process routine to set up memory management. The input ASCII files 
  are extracted from calval binary test files with only one band and one sca 
  used here. The compared result file is generated from cal/val IDL code.
--------------------------------------------------------------------------*/
Ensure test_algorithm_thru_api(void)
{
    const IAS_BAND_ATTRIBUTES *band_atts;
    char *test_directory;          /* test environment value */
    char file_name[PATH_MAX];      /* File name to open */
    BLOB *blob;                    /* Declare a blob structure */
    float *l1r_input_buffer = NULL;/* Image data for the current band/SCA */
    int band_number = 1;           /* 1-based band number */
    int status;                    /* Return variable */
    int band_index;                /* 0-based band number */
    int list_of_bands[1] = {10};        /* List of one test band */
    int list_of_band_indices[1] = {9};  /* List of one test band index */
    int num_bands = 1;                      /* Number of bands for testing */
    int i;

    /* Get the environment variable for the test data */
    test_directory = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (!test_directory)
    {
        assert_true_with_message(0, "Getting IAS_UNIT_TEST_DATA_DIR");
        return;
    }

    /* The attributes must be initialized for memory management to work */
    ias_sat_attr_initialize(IAS_L8);

    /* Allocate a blob and populate the minimum members for the test
       to succeed */
    blob = allocate_blob();
    blob->max_number_scas = 14;
    blob->total_bands = 30;
    blob->band_information = calloc(blob->total_bands, 
                                    sizeof( RPS_BAND_ATTRIBUTES *));
    blob->l1r_band_metadata = calloc( blob->total_bands,
        sizeof( IAS_L1R_BAND_METADATA * ));
    if ( ! blob->l1r_band_metadata )
    {
        assert_true_with_message(0, "Error allocating band metadata");
        local_free_the_blob( blob );
        return;
    }
    for ( band_index = 0; band_index < blob->total_bands; band_index++ )
    {
        blob->l1r_band_metadata[band_index] = calloc( 1,
            sizeof(IAS_L1R_BAND_METADATA));
        if ( ! blob->l1r_band_metadata[band_index] )
        {
            assert_true_with_message(0, "Error allocating band metadata");
            local_free_the_blob( blob );
            return;
        }
    }

    /* Local initialize the blob */
    for (band_number = 1; band_number <= blob->total_bands; band_number++)
    {
        int i;
        band_atts = ias_sat_attr_get_band_attributes(band_number);
        if (band_atts == NULL)
        {
            assert_true_with_message(FALSE, "Error obtaining RPS attribute "
                "information for band %d", band_number);
            local_free_the_blob(blob);
            return;
        }

        int band_index = band_atts->band_index;

        blob->band_information[band_index] = calloc(1,
            sizeof(RPS_BAND_ATTRIBUTES));
        if (!blob->band_information[band_index])
        {
            assert_true_with_message(0, "Error allocating band "
                "attribute memory");
            local_free_the_blob(blob);
            return;
        }
        blob->band_information[band_index]->band_number = band_number;
        blob->band_information[band_index]->number_lines = NLINES;
        blob->band_information[band_index]->number_detectors = NDETS;
        blob->band_information[band_index]->number_scas = band_atts->scas;
        blob->band_information[band_index]->band_classification =
            band_atts->band_classification;
        blob->band_information[band_index]->spectral_type =
            band_atts->spectral_type;

        for ( i = 0; i < num_bands; i++ )
        {
            if ( band_number == list_of_bands[i] )
            {
                list_of_band_indices[i] = band_index;
                break; /* Since there are no repeated bands */
            }
        }
    }

    /* Open the RLUT */
    status = snprintf(file_name, sizeof( file_name ), "%s/%s/%s",
        test_directory, test_subdir, rlut_file);
    if ( status < 0 || status >= sizeof( file_name ))
    {
        assert_true_with_message(0, "Building RLUT file name string");
        local_free_the_blob(blob);
        return;
    }
    blob->rlut = ias_rlut_open_file( file_name, IAS_READ );
    if ( ! blob->rlut )
    {
        assert_true_with_message(0, "Error opening %s", file_name );
        local_free_the_blob(blob);
        return;
    }

    /* Diable logging to suppress the memory management warnings and leave
       it off to supress debug in the algorithm */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* This initializes the memory management and registers buffers through
       the algorithms pre-process routine.  This algorithm doesn't need any
       memory management buffers that require the list of bands to be
       initialized -- number of bands set to 0 for that reason. */
    blob->number_of_bands = 0;
    blob->band_fields_initialized = 1;
    status = initialize_memory_management_buffers();
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error initializing memory management");
        ias_rlut_close_file( blob->rlut );
        local_free_the_blob(blob);
        return;
    }

    /* Initialize the RPS process level arrays need the TIRS blind band
       initialized */
    status = initialize_rps_level(14, 1);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error calling initialize_rps_level");
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        ias_rlut_close_file( blob->rlut );
        local_free_the_blob(blob);
        return;
    }

    /* Run the preprocess to register the memory buffers for the algorithm */
    status = tirs_secondary_linearization_preprocess(blob);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error returned from: "
            "tirs_secondary_linearization_preprocess");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        ias_rlut_close_file( blob->rlut );
        local_free_the_blob(blob);
        return;
    }

    /* Register the L1R buffer to this routine so that the API can't
       destroy the memory before we can retrieve the output */
    status = increment_consumer_count(IMAGE_L1R);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error returned from: "
            "increment_consumer_count IMAGE_L1R");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        ias_rlut_close_file( blob->rlut );
        local_free_the_blob(blob);
        return;
    }

    /* Get the image buffer for a band and SCA, this simply returns an
       unfilled, allocated buffer */
    l1r_input_buffer = (float *)preprocess_get_memory_buffer(IMAGE_L1R,
        list_of_band_indices[0], 0);
    if (!l1r_input_buffer)
    {
        assert_true_with_message(0, "Getting IMAGE_L1R buffer");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        ias_rlut_close_file( blob->rlut );
        local_free_the_blob(blob);
        return;
    }

    /* Set the l1r_input_buffer values.  Note: the starting value of this
       data ramp *must* be large enough to be within the data range of the
       DN look up and not blow over the top. */
    for ( i = 0; i < NDETS * NLINES; i++ )
        l1r_input_buffer[i] = 1400.0 + i;
    set_buffer_filled_flag( IMAGE_L1R, list_of_band_indices[0], 0);

    /* It is important that the consumer counts get set *after* all the
       increments */
    status = set_consumer_counts();
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error setting consumer counts");
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        ias_rlut_close_file( blob->rlut );
        local_free_the_blob(blob);
        return;
    }

    /* Call the API with a band and SCA */
    band_number = list_of_bands[0];
    band_index = list_of_band_indices[0];
    status = api_tirs_secondary_linearization( blob, band_index, 0, 0 );
    assert_equal_with_message( status,SUCCESS, "Error returned from "
        "api_tirs_secondary_linearization");
    if ( status != SUCCESS )
    {
        cleanup_rps_level();
        cleanup_memory_management_buffers(DISABLE_WARNING_DISPLAY);
        ias_rlut_close_file( blob->rlut );
        local_free_the_blob( blob );
        return;
    }

    /* Check some outputs with the expected values */
    assert_double_equal(l1r_input_buffer[64], 1481.384766);
    assert_double_equal(l1r_input_buffer[64 * 2], 1545.996094);
    assert_double_equal(l1r_input_buffer[64 * 64 - 1], 5585.947266);

    /* Free the image memory */
    status = decrement_consumer_count(IMAGE_L1R, list_of_band_indices[0], 0);
    assert_equal_with_message( status, SUCCESS,
        "Error returned from decrement consumer count");

    cleanup_rps_level();
    cleanup_memory_management_buffers(ENABLE_WARNING_DISPLAY);
    ias_rlut_close_file( blob->rlut );
    local_free_the_blob(blob);
}  /* END test_algorithm_thru_api */
