/* Standard Library Includes */
#include <sys/stat.h>
#include <stdio.h>
#include <limits.h>
#include <unistd.h>

/* COTS Library Includes */
#include "cgreen/cgreen.h"

Ensure tirs_stray_light_correction_test()
{
    int status;                         /* Status placeholder */
    struct stat buf;                    /* Stats buffer struct */
    static char *unit_test_dir;         /* Unit test directory */
    char command[PATH_MAX];             /* Command buffer */
    const char *l1r_output_name = "tirs_stray_light.h5"; /* L1R file */

    /* Get the IAS_UNIT_TEST_DATA_DIR environment variable */
    unit_test_dir = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (!unit_test_dir)
    {
        assert_true_with_message(0, "The IAS_UNIT_TEST_DATA_DIR environment "
            "variable is not set\n");
        return;
    }

    /* Prepare the sed command for the OMF file */
    status = snprintf(command, sizeof(command), 
        "sed 's|$IAS_UNIT_TEST_DATA_DIR|%s|' "
        "template_I13000.omf > I13000.omf", unit_test_dir);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Preparing OMF system command");
        return;
    }

    /* Run the sed command */
    if (system(command) != 0)
    {
        assert_true_with_message(0, "Preparing OMF file for unit "
            "test data directory path setup");
        return;
    }

    /* Prepare the sed command for the ODL file */
    status = snprintf(command, sizeof(command), 
        "sed 's|$IAS_UNIT_TEST_DATA_DIR|%s|' "
        "template_tirs_stray_light_correction.odl > tirs_stray_light_"
        "correction.odl", unit_test_dir);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Preparing ODL system command");
        return;
    }

    /* Run the sed command */
    if (system(command) != 0)
    {
        assert_true_with_message(0, "Preparing ODL file for unit "
            "test data directory path setup");
        return;
    }

    /* Make sure the L1R output file is not present */
    if (stat(l1r_output_name, &buf) == 0)
    {
        /* Remove the file */
        if (unlink(l1r_output_name))
        {
            assert_true_with_message(0, "Removing file: %s", l1r_output_name);
        }
        else
            printf("Removed file: %s\n", l1r_output_name);
    }

    /* Create the command for running create_l1r */
    status = snprintf(command, sizeof(command), "export IAS_LOG_LEVEL=WARN; "
        "nice --adjustment=20 ../create_l1r tirs_stray_light_correction.odl");
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Preparing system command for "
            "executing create_l1r");
        return;
    }
    
    /* Execute create_l1r */
    printf("Executing: %s\n", command);
    status = system(command);
    assert_equal(status, 0);
    if (status != 0)
    {
        /* Clean up and return */
        assert_true_with_message(0, "Executing system command: %s",
            command);

        /* Clean up bad L1R */
        if (unlink(l1r_output_name))
        {
            assert_true_with_message(0, "Removing file %s", l1r_output_name);
        }
        return;
    }
    printf("Done: %s\n", command);

    /* Create the diff command */
    status = snprintf(command, sizeof(command), "nice --adjustment=20 h5diff "
        "%s %s/ias_base/rps/driver/tirs_stray_light_correction/%s", 
        l1r_output_name, unit_test_dir, l1r_output_name);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(0, "Preparing diff command for l1r");
        return;
    }

    /* Execute diff command */
    status = system(command);
    assert_equal(status, 0);
    if (status != 0)
    {
        assert_true_with_message(0, "Executing system command: %s", command);
        return;
    }

    /* Clean up the L1R output file */
    if (unlink(l1r_output_name))
    {
        assert_true_with_message(0, "Removing file %s", l1r_output_name);
        return;
    }
}

/*   Main test program */
int main( int argc, char *argv[] )
{
    TestSuite *suite = create_test_suite();
    add_test(suite, tirs_stray_light_correction_test);
    
    if (setenv("IAS_VERSION_OVERRIDE", "Y", 1) != 0)
    {
        printf("Error: Setting the IAS_VERSION_OVERRIDE environment variable");
        return 1;
    }

    /* Run the tests */
    if (argc > 1)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}
