/*--------------------------------------------------------------------------
 NAME:     1lr_sca_stitching

 PURPOSE:  Utility to combine individual L1R SCA images for a given band
           into a single image.  Parameters establish how the individual
           SCA images will be combined

 RETURNS:  EXIT_SUCCESS or EXIT_FAILURE

 ALGORITHM REFERENCES:  Version 3.3 of L1R SCA Stitching Utility ADD

---------------------------------------------------------------------------*/
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include "ias_cpf.h"
#include "ias_l1r.h"
#include "ias_logging.h"
#include "l1r_sca_stitching.h"
#include "ias_satellite_attributes.h" 

int main( int argc, char *argv[] )
{
    char output_l1r_filename[PATH_MAX];     /* File name of output image */
    IAS_L1R_FILE_METADATA input_file_metadata; /* L1R file metadata for
                                                  input image */
    IAS_L1R_FILE_METADATA output_file_metadata;/* L1R file metadata for
                                                  output image */
    L1RIO *input_image_ptr   = NULL;        /* L1R IO structure for input */
    L1RIO *output_image_ptr  = NULL;        /* L1R IO structure for output */
    IAS_CPF *cpf = NULL;                    /* CPF data */
    const struct IAS_CPF_FOCAL_PLANE *cpf_focal_plane = NULL;
    const struct IAS_CPF_DETECTOR_OFFSETS *cpf_det_offsets = NULL;
    int max_combined_band_sca_offset = 0;   /* Largest (band + SCA) offset 
                                               (in frames) for all the bands */
    int start_line_adjustment = 0;          /* Value that to adjust the top
                                              of the image by to account for
                                              detector offsets */
    int status;                             /* Function return status */
    SCA_STITCH_PARAMS  params;              /* Processing parameter data
                                               structure */
    int *band_offsets = NULL;                                    
    int *sca_offsets = NULL;                                    
    int *calculated_band_offsets = NULL;                                    
    int *calculated_sca_offsets = NULL;                                    
    const IAS_BAND_ATTRIBUTES *band_attributes = NULL; /* Attribute info
                                                        for current band */
    int band_number;
    int band_index;
    int sca_index;
    int index;                       /* Simple loop control */

    /* Initialize logging */
    ias_log_initialize( argv[0] );

    /* Only one command-line argument: the parameter file */
    if ( argc != 2 )
    {
        IAS_LOG_ERROR("Usage: %s <parameter_file>\n", argv[0] );
        return EXIT_FAILURE;
    }

    /* Read the parameter file */
    status = read_stitch_parameters( argv[1], &params );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Reading parameters");
        return EXIT_FAILURE;
    }

    /* Output the start up message */
    IAS_LOG_INFO("L1R SCA Image Stitching Starts ...\n");

   /* Read the CPF */
    cpf = ias_cpf_read( params.cpf_filename );
    if ( !cpf )
    {
        IAS_LOG_ERROR("Opening CPF file %s", params.cpf_filename );
        return EXIT_FAILURE;
    }

    /* Get the focal plane parameter group from the CPF */
    cpf_focal_plane = ias_cpf_get_focal_plane( cpf );
    if ( !cpf_focal_plane )
    {
        IAS_LOG_ERROR("Getting focal plane parameters from CPF");
        ias_cpf_free( cpf );
        return EXIT_FAILURE;
    }

    if (params.apply_det_offset_flag)
    {
        /* Read the detector offsets from the CPF */
        cpf_det_offsets = ias_cpf_get_det_offsets(cpf);
        if (cpf_det_offsets == NULL)
        {
            IAS_LOG_ERROR("Getting detector offsets from CPF");
            ias_cpf_free( cpf );
            return EXIT_FAILURE;
        }
    }

    /* Open the input files */
    input_image_ptr = ias_l1r_open_image( params.l1r_filename, IAS_READ );
    if ( input_image_ptr == NULL )
    {
        IAS_LOG_ERROR("Cannot open input file %s for reading",
            params.l1r_filename );
        free(calculated_sca_offsets);
        ias_cpf_free( cpf );
        return EXIT_FAILURE;
    }

    /* Get the metadata from the input file */
    status = ias_l1r_get_file_metadata( input_image_ptr,
        &input_file_metadata );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Getting the input file metadata from %s",
            params.l1r_filename );
        ias_l1r_close_image( input_image_ptr );
        ias_cpf_free( cpf );
        return EXIT_FAILURE;
    }

    /* The calculated_band_offsets and calculated_sca_offsets are only done for
       Earth image */
    if(strstr(input_file_metadata.collect_type, "EARTH") != NULL)
    {
        /* Earth Image */
        if (params.scene_center_latitude_odl_flag == 0 )
        {
            /* The scene center latitude from the input parameter file is not 
               available, so derive scene center latitude from L0R metadata */
            if (strcasecmp(params.l0r_filename,"dummy"))
            {
                status = get_l0r_scene_center_latitude(params.l0r_filename, 
                    &params.scene_center_latitude);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR("Getting the scene center latitude from L0R");
                    ias_l1r_close_image( input_image_ptr );
                    ias_cpf_free( cpf );
                    return EXIT_FAILURE;
                }
            }
            else
            {
                IAS_LOG_ERROR("Scene center latitude is not available from "
                    "either the SCENE_CENTER_LATITUDE parameter or the L0R");
                ias_l1r_close_image( input_image_ptr );
                ias_cpf_free( cpf );
                return EXIT_FAILURE;
            }
        }

        /* Allocate memory for calculated_band_offsets */
        calculated_band_offsets = calloc(params.total_normal_bands, 
                                         sizeof(int));
        if(!calculated_band_offsets)
        {
            IAS_LOG_ERROR("Allocating memory for calculated_band_offsets");
            ias_l1r_close_image( input_image_ptr );
            ias_cpf_free( cpf );
            return EXIT_FAILURE;
        }

        /* Allocate memory for calculated_sca_offsets */
        calculated_sca_offsets = calloc(params.total_normal_bands * 
                                        IAS_MAX_NSCAS, sizeof(int));
        if (!calculated_sca_offsets)
        {
            IAS_LOG_ERROR("Allocating memory for calculated_sca_offsets");
            free(calculated_band_offsets);
            ias_l1r_close_image( input_image_ptr );
            ias_cpf_free( cpf );
            return EXIT_FAILURE;
        }

        /* Loop as many times as there are bands to process */
        for ( index = 0; index < params.total_normal_bands; index++ )
        {
            band_number = ias_sat_attr_convert_band_index_to_number( index );
            band_attributes = ias_sat_attr_get_band_attributes( band_number );
            if ( band_attributes == NULL )
            {
                IAS_LOG_ERROR("Cannot retrieve band %d attributes",
                    band_number );
                free(calculated_sca_offsets);
                free(calculated_band_offsets);
                ias_l1r_close_image( input_image_ptr );
                ias_cpf_free( cpf );
                return ERROR;
            }
            band_index = band_attributes->band_index;

            /* Get the calculated SCA offsets from SCA Alignment */
            status = get_calculated_sca_offsets(cpf, band_attributes->sensor_id,
                     band_index, band_attributes->scas, 
                     params.scene_center_latitude, 
                     params.nominal_scene_elevation,
                     &calculated_band_offsets[band_index], 
                     &calculated_sca_offsets[band_index * IAS_MAX_NSCAS]);
            if ( status != SUCCESS )
            {
                IAS_LOG_ERROR("Getting the OLI calculated SCA offsets");
                ias_l1r_close_image( input_image_ptr );
                ias_cpf_free( cpf );
                free(calculated_sca_offsets);
                free(calculated_band_offsets);
                return EXIT_FAILURE;
            }
        }
    }

    /* Allocate memory for band_offsets */
    band_offsets = calloc(params.total_normal_bands, sizeof(int));
    if(!band_offsets)
    {
        IAS_LOG_ERROR("Allocating memory for band_offsets");
        free(calculated_band_offsets);
        free(calculated_sca_offsets);
        ias_l1r_close_image( input_image_ptr );
        ias_cpf_free( cpf );
        return EXIT_FAILURE;
    }

    /* Allocate memory for sca_offsets */
    sca_offsets = calloc(params.total_normal_bands * 
                         IAS_MAX_NSCAS, sizeof(int));
    if (!sca_offsets)
    {
        IAS_LOG_ERROR("Allocating memory for sca_offsets");
        free(calculated_band_offsets);
        free(calculated_sca_offsets);
        ias_l1r_close_image( input_image_ptr );
        free(band_offsets);
        ias_cpf_free( cpf );
        return EXIT_FAILURE;
    }

    /* Loop as many times as there are bands to process */
    for ( index = 0; index < params.total_normal_bands; index++ )
    {
        band_number = ias_sat_attr_convert_band_index_to_number( index );
        band_attributes = ias_sat_attr_get_band_attributes(
            band_number );
        if ( band_attributes == NULL )
        {
            IAS_LOG_ERROR("Cannot retrieve band %d attributes",
                band_number );
            free(calculated_band_offsets);
            free(calculated_sca_offsets);
            ias_l1r_close_image( input_image_ptr );
            free(band_offsets);
            free(sca_offsets);
            ias_cpf_free( cpf );
            return ERROR;
        }
        band_index = band_attributes->band_index;

        /* Get the the band offsets from the CPF and added in
           the calculated band offset if it's available */
        band_offsets[band_index] = cpf_focal_plane->band_offset[band_index];
        if (calculated_band_offsets)
        {
            band_offsets[band_index] += 
                calculated_band_offsets[band_index];
        }

        for (sca_index = 0; sca_index < band_attributes->scas; sca_index++)
        {
            /* Get the SCA offsets from the CPF and added in
               the calculated band offset if it's available */
            sca_offsets[band_index * IAS_MAX_NSCAS + sca_index] = 
                cpf_focal_plane->sca_offset[band_index][sca_index];
            if (calculated_sca_offsets)
            {
                sca_offsets[band_index * IAS_MAX_NSCAS + sca_index] += 
                    calculated_sca_offsets[band_index * 
                    IAS_MAX_NSCAS + sca_index];
            }
        }
    }

    /* Free the allocated memory for calculated band/sca offsets */
    free(calculated_band_offsets);
    calculated_band_offsets = NULL;
    free(calculated_sca_offsets);
    calculated_sca_offsets = NULL;

    /* Get the largest (band + SCA) offset (in frames) found in the CPF.
       The detector offset will optionally be determined also.
       NOTE:
       If TIRS and OLI bands are processed at the same time, it's
       possible that separate offsets will be needed.  This 
       implementation currently assumes it won't be... */
    status = get_max_combined_band_sca_offset( cpf_det_offsets, band_offsets, 
        sca_offsets, params.total_normal_bands, params.apply_det_offset_flag,
        &max_combined_band_sca_offset, &start_line_adjustment );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Cannot find largest combined (band + SCA) offset");
        ias_cpf_free( cpf );
        ias_l1r_close_image( input_image_ptr );
        free(calculated_sca_offsets);
        free(calculated_band_offsets);
        free(sca_offsets);
        free(band_offsets);
        return EXIT_FAILURE;
    }

    /* Open the output files */
    output_image_ptr = ias_l1r_open_image( params.output_l1r_filename,
        IAS_WRITE );
    if ( output_image_ptr == NULL )
    {
        IAS_LOG_ERROR("Cannot open output file %s for writing",
            output_l1r_filename );
        ias_l1r_close_image( input_image_ptr );
        ias_cpf_free( cpf );
        free(calculated_sca_offsets);
        free(calculated_band_offsets);
        free(sca_offsets);
        free(band_offsets);
        return EXIT_FAILURE;
    }

    /* Process the bands */
    status = process_bands( input_image_ptr, output_image_ptr, params,
        cpf_focal_plane, max_combined_band_sca_offset,
        band_offsets, sca_offsets, params.apply_det_offset_flag,
        cpf_det_offsets, start_line_adjustment );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Processing bands");
        ias_l1r_close_image( input_image_ptr );
        ias_l1r_close_image( output_image_ptr );
        ias_cpf_free( cpf );
        free(calculated_sca_offsets);
        free(calculated_band_offsets);
        free(sca_offsets);
        free(band_offsets);
        return EXIT_FAILURE;
    }

    /* Free the allocated memory for band/sca offsets */
    free(band_offsets);
    band_offsets = NULL;
    free(sca_offsets);
    sca_offsets = NULL;

    /* Update the metadata with the stitching method */
    status = update_l1r_metadata( &input_file_metadata, params.stitch_method,
            &output_file_metadata );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Cannot update L1R metadata in %s with stitching "
            "method information", params.output_l1r_filename );
        ias_l1r_close_image( input_image_ptr );
        ias_l1r_close_image( output_image_ptr );
        ias_cpf_free( cpf );
        return EXIT_FAILURE;
    }

    /* Write the file metadata */
    status = ias_l1r_set_file_metadata( output_image_ptr,
        &output_file_metadata );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Cannot write metadata to stitched image file");
        ias_l1r_close_image( input_image_ptr );
        ias_l1r_close_image( output_image_ptr );
        ias_cpf_free( cpf );
        return EXIT_FAILURE;
    }

    /* Cleanup */
    ias_l1r_close_image( input_image_ptr );
    ias_l1r_close_image( output_image_ptr );
    ias_cpf_free( cpf );

    IAS_LOG_INFO("%s completed successfully", argv[0] );
    return EXIT_SUCCESS;
}  /* END main */
