#ifndef    L1R_SCA_STITCHING_H
#define    L1R_SCA_STITCHING_H

#include <limits.h>

#include "ias_cpf.h"
#include "ias_l1r.h"

/* The stitch method is passed in as integers.  This enumeration is used to
   make internal code more readable. */
typedef enum stitch_method_type
{
    SCA_STITCH_NO_OVERLAP,       /* No overlap method */
    SCA_STITCH_LEFT_OVERLAP,     /* Left overlap method */
    SCA_STITCH_RIGHT_OVERLAP,    /* Right overlap method */
    SCA_STITCH_HALF_OVERLAP      /* Left-right overlap method */
} STITCH_METHOD_TYPE;

/* This structure contains the data read from the input parameter file and
   OMF.  Used to conveniently pass values around. */
typedef struct sca_stitch_params
{
    char omf_filename[PATH_MAX];
    char work_order_id[IAS_WORKORDER_ID_SIZE];
    char l0r_filename[PATH_MAX];
    char l1r_filename[PATH_MAX];
    char output_l1r_filename[PATH_MAX];
    int apply_det_offset_flag;
    char cpf_filename[PATH_MAX];         /* CPF filename / location */
    int num_bands_to_process;            /* Total number of bands to process */
    int band_number_list[IAS_MAX_NBANDS];/* List of bands to process */
    int total_normal_bands;              /* Number of normal bands */
    int scene_center_latitude_odl_flag;  /* Flag to indicate if 
                                            scene_center_latitude is specified 
                                            in the input ODL file */    
    double scene_center_latitude;       
    double nominal_scene_elevation;     
    STITCH_METHOD_TYPE stitch_method;    /* Flag value representing stitching
                                            method */
} SCA_STITCH_PARAMS;

/* Prototypes */
int read_stitch_parameters
(
    const char *stitch_params_filename,    /* I: Name of processing
                                                 parameter file */
    SCA_STITCH_PARAMS *stitch_param_data   /* O: Processing parameters */
);

int get_cpf_sca_offset
(
    const struct IAS_CPF_FOCAL_PLANE *focal_plane, 
                                           /* I: CPF focal plane parameter */
    int   band_index,                      /* I: 0-based index for the band */
    int   sca_index                        /* I: 0-based index for the SCA */
);

int get_max_combined_band_sca_offset
(
    const struct IAS_CPF_DETECTOR_OFFSETS *det_offsets,
                            /* I: CPF along track detector offset parameters */
    const int *band_offsets,            /* I: Band offsets */
    const int *sca_offsets,             /* I: SCA offsets */
    int num_bands,                      /* I: Total number of bands */
    int apply_det_offsets_flag,         /* I: Flag to apply the detector 
                                                  offsets */
    int *max_combined_band_sca_offset,  /* O: Largest (band + sca) offset */
    int *det_offset_adjustment          /* O: Value that to adjust the top
                                              of the image by to account for
                                              detector offsets */
);

int update_l1r_metadata
(
    const IAS_L1R_FILE_METADATA *in_fmd,/* I: Input L1r file metadata */
    STITCH_METHOD_TYPE stitch_method,   /* I: SCA stitching method */
    IAS_L1R_FILE_METADATA *out_fmd      /* O: Stitched l1r image data
                                              buffer */
);

int stitch_image
(
    L1R_BAND_IO *input_band_io, /* I: Input image data set for this band */
    STITCH_METHOD_TYPE stitch_method, /* I: Stitching information */
    int band_offset,            /* I: Offset from top of image for the band */
    const int *sca_offsets,     /* I: Offsets for each SCA in the band */
    int sca_overlap,            /* I: SCA Overlap width for each the band */
    int num_scas,               /* I: Number of SCAs in the input band */
    int number_input_lines,     /* I: Number of lines in the input image */
    int number_input_samples,   /* I: Number of samples per line in the
                                      input image */
    int number_output_samples,  /* I: Number of samples per line in the
                                      stitched output image */
    int apply_det_offsets_flag, /* I: Flag to apply the detector offsets */
    int band_index,             /* I: Band index for max det offset arr */

    const struct IAS_CPF_DETECTOR_OFFSETS *cpf_det_offset,
    int start_line_adjustment,  /* I: This value is added to the starting
                                      line of the output image to adjust
                                      to negative detector offsets */
    float *output_buffer        /* I/O: Populated stitched image buffer */
);

int process_bands
(
    L1RIO *input_image_ptr,     /* I: Pointer to input image */
    L1RIO *output_image_ptr,    /* I: Pointer to output image */
    SCA_STITCH_PARAMS  params,  /* I: Miscellaneous parameters */
    const struct IAS_CPF_FOCAL_PLANE *cpf_focal_plane,
                                /* I: Focal plane offsets and overlap widths */
    int max_combined_band_sca_offset,  /* I: Largest (band + SCA) offset */
    const int *band_offsets,           /* I: Band offsets */
    const int *sca_offsets,            /* I: SCA offsets */
    int apply_det_offsets_flag, /* I: Flag to apply the detector offsets */
    const struct IAS_CPF_DETECTOR_OFFSETS *cpf_det_offsets,
                                /* I: Detector offsets */
    int start_line_adjustment   /* I: This value is added to the starting
                                      line of the output image to adjust
                                      to negative detector offsets */
);

int get_calculated_sca_offsets
(
    IAS_CPF  *cpf,                 /* I: IAS_CPF pointer */
    int sensor_id,                 /* I: Sensor ID ("IAS_OLI", "IAS_TIRS") */
    int        band_index,         /* I: 0-based current band index */
    int        number_scas,        /* I: Number of SCAs in current band */
    double scene_center_latitude,  /* I: Scene center latitude */     
    double nominal_scene_elevation,/* Nominal scene elevation */     
    int        *band_offset,       /* O: Band offset in frames */
    int        *sca_offsets        /* O: SCA offsets in frames */
);

int get_l0r_scene_center_latitude
(
    const char *l0r_file_name,        /* I: L0R image filename with path */
    double *scene_center_latitude     /* O: Scene center latitude */
);

#endif

