/*--------------------------------------------------------------------------
 NAME:     process_bands

 PURPOSE:  Utility to combine individual L1R SCA images for a given band
           into a single image.  Parameters establish how the individual
           SCA images will be combined

 RETURNS:  SUCCESS or ERROR

---------------------------------------------------------------------------*/
#include <stdlib.h>
#include <string.h>

#include "ias_cpf.h"
#include "ias_l1r.h"
#include "ias_logging.h"
#include "l1r_sca_stitching.h"
#include "ias_satellite_attributes.h" 

/* Local routine to close the bands */
static int close_bands
(
    L1R_BAND_IO *input_band,    /* I: Input band structure */
    L1R_BAND_IO *output_band,   /* I: Output band structure */
    int band_number             /* I: Band number for error message */
)
{
    int status;
    int return_status = SUCCESS;

    if ( input_band != NULL )
    {
        status = ias_l1r_close_band( input_band );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Closing input band %d", band_number );
            return_status = ERROR;
        }
    }
    if ( output_band != NULL )
    {
        status = ias_l1r_close_band( output_band );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Closing output band %d", band_number );
            return_status = ERROR;
        }
    }

    return return_status;
}  /* END close_bands */

/****************************************************************************
Name: process_bands

***************************************************************************/

int process_bands
(
    L1RIO *input_image_ptr,     /* I: Pointer to input image */
    L1RIO *output_image_ptr,    /* I: Pointer to output image */
    SCA_STITCH_PARAMS  params,  /* I: Miscellaneous parameters */
    const struct IAS_CPF_FOCAL_PLANE *cpf_focal_plane,/* I: Focal plane offsets
                                                         and overlap widths */
    int max_combined_band_sca_offset,  /* I: Largest (band + SCA) offset */
    const int *band_offsets,           /* I: Band offsets */
    const int *sca_offsets,            /* I: SCA offsets */
    int apply_det_offsets_flag, /* I: Flag to apply the detector offsets */
    const struct IAS_CPF_DETECTOR_OFFSETS *cpf_det_offsets,
                                /* I: Detector offsets */
    int start_line_adjustment   /* I: This value is added to the starting
                                      line of the output image to adjust
                                      to negative detector offsets */
)
{
    int status;                 /* Routine return status */
    int number_output_lines;    /* Calculated number of output lines */
    int number_output_samples;  /* Calculated number of output samples */
    int number_output_scas;     /* Number of SCAs in stitched output image */
    int detectors_per_band;     /* Number of detectors in the band */
    int net_sca_overlap_width;  /* Final overlap between consecutive SCAs */
    int number_input_samples;   /* Number samples read from L1R band data */
    int number_input_scas;      /* Number of SCAs read from L1R band data */
    int band_offset;            /* Current band offset from CPF */
    IAS_DATA_TYPE data_type;    /* Data type for current band's image data */
    int number_input_lines;     /* Number of lines in the input image */
    int band_number;            /* Current band number */
    int band_index;             /* Index of the band, it's always one less
                                   than band_number -- used for convenience */
    IAS_L1R_BAND_METADATA  band_metadata; /* Band metadata read from input
                                             image and written to output */
    const IAS_BAND_ATTRIBUTES *band_attributes = NULL; /* Attribute info
                                                        for current band */
    L1R_BAND_IO *input_band_io  = NULL; /* Input L1R IO structure */
    L1R_BAND_IO *output_band_io = NULL; /* Output L1R IO structure */
    int sca_overlap;             /* Overlap width within the band */
    float *output_buffer = NULL; /*  Data buffer for stitched output image */
    int i;                       /* Simple loop control */

    /* Loop as many times as there are bands to process */
    for ( i = 0; i < params.num_bands_to_process; i++ )
    {
        /* Get the band number */
        band_number = params.band_number_list[i];

        /* Output Stitching band message */
        IAS_LOG_INFO("Stitching band %d ...\n", band_number);

        /* Get the sensor attributes for the current band */
        band_attributes = ias_sat_attr_get_band_attributes(
            band_number );
        if ( band_attributes == NULL )
        {
            IAS_LOG_ERROR("Cannot retrieve band %d attributes",
                band_number );
            return ERROR;
        }

        /* For convenience get the index in a local variable */
        band_index = band_attributes->band_index;

        /* Determine the total number of detectors in the band */
        detectors_per_band = band_attributes->scas *
            band_attributes->detectors_per_sca;

        /* Get the offset for the current band from the CPF */
        band_offset = band_offsets[band_index];

        /* Get the SCA overlap widths for the current band from the CPF */
        sca_overlap = cpf_focal_plane->sca_overlap[band_index];

        /* Calculate the "net" SCA overlap width.  This should take into
           account any observed differences in overlap width between
           the SCAs.  If the overlaps are identical, the total overlap
           will be (NSCAs-1) * SCA_overlap_width */
        net_sca_overlap_width = 0;
        switch ( params.stitch_method )
        {
            case SCA_STITCH_NO_OVERLAP :
                net_sca_overlap_width = 0;
                break;

            case SCA_STITCH_LEFT_OVERLAP :
            case SCA_STITCH_RIGHT_OVERLAP :
            case SCA_STITCH_HALF_OVERLAP :
                net_sca_overlap_width = ( band_attributes->scas - 1 ) *
                    sca_overlap;
                break;

            default :
                IAS_LOG_ERROR("Invalid stitching identifier");
                return ERROR;
        }

        /* Set up reading the image data */
        input_band_io = ias_l1r_open_band( input_image_ptr,
            band_number, &data_type, &number_input_scas,
            &number_input_lines, &number_input_samples );
        if ( input_band_io == NULL )
        {
            IAS_LOG_ERROR("Opening input band %d image data",
                band_number );
            return ERROR;
        }

        /* Get the band metadata */
        status = ias_l1r_get_band_metadata( input_image_ptr,
            band_number, &band_metadata );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Reading band metadata for band %d",
                band_number );
            close_bands( input_band_io, NULL, band_number );
            return ERROR;
        }

        /* Validate the correct data type for a L1R product.  It should
           be float */
        if ( data_type != IAS_R4 )
        {
            IAS_LOG_ERROR("Invalid data type for band %d L1R product",
                band_number );
            close_bands( input_band_io, NULL, band_number );
            return ERROR;
        }

        /* Calculate required output image dimensions given the
           selected stitching method.  The offset is in frames and needs to
           be converted to lines. */
        number_output_lines = number_input_lines + 
            ( max_combined_band_sca_offset *
            band_attributes->lines_per_frame );
        number_output_samples = detectors_per_band - net_sca_overlap_width;

        /* Allocate the output image buffer, calloc is used here since the 
           stitching process doesn't fill every pixel in the output buffer */
        output_buffer = calloc(( number_output_lines * number_output_samples ),
                sizeof(*output_buffer));
        if ( output_buffer == NULL )
        {
            IAS_LOG_ERROR("Cannot allocate image data buffers");
            close_bands( input_band_io, NULL, band_number );
            return ERROR;
        }

        /* Open the output image band.  Since this is a stitched image there
           is only one SCA to write. */
        number_output_scas = 1;
        output_band_io = ias_l1r_open_band( output_image_ptr,
            band_number, &data_type, &number_output_scas,
            &number_output_lines, &number_output_samples );
        if ( output_band_io == NULL )
        {
            IAS_LOG_ERROR("Opening output band %d stitched image data",
                band_number );
            close_bands( input_band_io, NULL, band_number );
            free( output_buffer );
            return ERROR;
        }

        /* Create the stitched image */
        status = stitch_image( input_band_io, params.stitch_method,
            band_offset, &sca_offsets[band_index * IAS_MAX_NSCAS], 
            sca_overlap, number_input_scas, number_input_lines, 
            number_input_samples, number_output_samples,
            apply_det_offsets_flag, band_index,
            cpf_det_offsets, start_line_adjustment, output_buffer );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Creating stitched image data buffer for band %d",
                band_number );
            close_bands( input_band_io, output_band_io, band_number );
            free( output_buffer );
            return ERROR;
        }

        /* Write the stitched image data buffer to the output file.  The
           entire buffer is written as one SCA. */
        status = ias_l1r_write_image( output_band_io, 0, 0, 0,
            number_output_lines, number_output_samples, output_buffer );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Writing band %d stitched image to file",
                band_number );
            close_bands( input_band_io, output_band_io, band_number );
            return ERROR;
        }

        /* Write (some people call this set, as the library routine implies,
           but it's really a write ;-) the band metadata */
        status = ias_l1r_set_band_metadata( output_image_ptr,
            &band_metadata, 1 );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Setting band metadata for band %d",
                band_number );
            close_bands( input_band_io, output_band_io, band_number );
            return ERROR;
        }

        /* Close the input and output image bands, and free up the
           output data buffer */
        free( output_buffer );
        status = close_bands( input_band_io, output_band_io, band_number );
        if ( status )
        {
            IAS_LOG_ERROR("Closing bands at end of band loop");
            return ERROR;
        }
    }  /* END for band_index loop */

    return SUCCESS;
}  /* END process_bands */
