/*---------------------------------------------------------------------------
 NAME:     read_stitch_parameters

 PURPOSE:  Reads the input parameter and OMF files and populates the
           parameter data structure

 RETURNS:  SUCCESS or ERROR
           Exits with EXIT_SUCCESS if the user supplied --help, --template or
           --loadtable on the command line

---------------------------------------------------------------------------*/
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include "ias_parm.h"
#include "ias_parm_standard_parameters.h"
#include "l1r_sca_stitching.h"

int read_stitch_parameters
(
    const char *parameter_filename,         /* I: Parameter filename */
    SCA_STITCH_PARAMS *parameter_data       /* O: Stitching parameters */
)
{
    int local_stitch_method;    /* Integer stitch method from parameters */
    int local_det_off_flag;     /* Apply Detector offset flag */
    int satellite_number;       /* Satellite number being processed */
    int status;                 /* Status for each read operation */
    int read_count;             /* Count of parameters successfully read */
    int num_bands_to_process;   /* Number of bands read from parameter file */
    int i;                      /* simple loop counter */
    IAS_SATELLITE_ID satellite_id;    /* The ID for the satellite attributes
                                         library initialization */
    const char *default_output_file_name[] = {"stitched_l1r_file.h5"};
    int min_elevation;
    int max_elevation;

    /* This is the default stitching method if parameter isn't specified */
    int default_stitch_method = SCA_STITCH_NO_OVERLAP;

    /* This local band list is set to a large number of entries since it's
       unknown at this point how many bands will be passed in */
    int local_band_list[IAS_MAX_NBANDS];
    /* Set up an array for checking if the bands are listed more than once.
       Array is sized one larger to be accessed with a band number. */
    int is_band_already_listed[IAS_MAX_NBANDS + 1] = {0};

    /* Table definition for the input parameters */
    IAS_PARM_DECLARE_TABLE( parameter_table, 6 );

    IAS_PARM_ADD_STRING( parameter_table, WORK_ORDER_ID, "Work order ID",
        IAS_PARM_REQUIRED, 0, NULL, 0, NULL,
        parameter_data->work_order_id,
        sizeof(parameter_data->work_order_id), 0 );

    IAS_PARM_ADD_INT( parameter_table, BAND_LIST, "List of bands: "
        "The list defaults to the normal image bands, if a non-normal band "
        "is passed in an error will be issued.",
        IAS_PARM_OPTIONAL, IAS_PARM_ARRAY,
        0, 0, 0, 0, NULL, local_band_list,
        sizeof(local_band_list), 1 );

    IAS_PARM_ADD_INT( parameter_table, STITCH_METHOD, "Stitching method: "
        "no overlap = 0, left = 1, right = 2, half = 3",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 1,
        SCA_STITCH_NO_OVERLAP, SCA_STITCH_HALF_OVERLAP, 1,
        &default_stitch_method, &local_stitch_method,
        sizeof(local_stitch_method), 1 );

    IAS_PARM_ADD_STRING( parameter_table, STITCHED_FILENAME,
        "File name for the stitched image", IAS_PARM_OPTIONAL, 0, NULL, 1,
        default_output_file_name, parameter_data->output_l1r_filename,
        sizeof(parameter_data->output_l1r_filename), 1 );

    int apply_detector_offsets_flag_default = 1;
    IAS_PARM_ADD_BOOLEAN( parameter_table, APPLY_DETECTOR_OFFSETS_FLAG,
        "Apply the detector offsets to the stitched image: "
        "1 = apply offsets, 0 = don't apply offsets",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        &apply_detector_offsets_flag_default,
        &local_det_off_flag, 
        sizeof(local_det_off_flag), 1 );

    IAS_PARM_ADD_DOUBLE( parameter_table, SCENE_CENTER_LATITUDE,
        "Scene center latitude in degree", IAS_PARM_OPTIONAL,
        IAS_PARM_NOT_ARRAY, 1, -90.0, 90.0, 0, NULL,
        &parameter_data->scene_center_latitude,
        sizeof(parameter_data->scene_center_latitude) , 1 );

    /* Table definition for the OMF */
    IAS_PARM_DECLARE_TABLE( omf_parameter_table, 6 );

    IAS_PARM_SATELLITE( omf_parameter_table, &satellite_number,
        sizeof(satellite_number), 1 );

    IAS_PARM_CPF(omf_parameter_table, IAS_PARM_REQUIRED,
        parameter_data->cpf_filename, sizeof(parameter_data->cpf_filename), 1);

    const char *default_l0r_filename[] = {"dummy"};
    IAS_PARM_ADD_STRING( omf_parameter_table, L0R_FILENAME,
        "L0R file name", IAS_PARM_OPTIONAL,
        0, NULL, 1, default_l0r_filename, parameter_data->l0r_filename,
        sizeof(parameter_data->l0r_filename), 0 );

    IAS_PARM_ADD_STRING( omf_parameter_table, L1R_FILENAME,
        "L1R file name", IAS_PARM_REQUIRED,
        0, NULL, 0, NULL, parameter_data->l1r_filename,
        sizeof(parameter_data->l1r_filename), 0 );

    /* Get the DEM elevation information */
    int min_elevation_default = 0;
    IAS_PARM_MIN_ELEVATION(omf_parameter_table, IAS_PARM_OPTIONAL, 1,
        &min_elevation_default, &min_elevation, sizeof(min_elevation), 0);

    int max_elevation_default = 0;
    IAS_PARM_MAX_ELEVATION(omf_parameter_table, IAS_PARM_OPTIONAL, 1,
        &max_elevation_default, &max_elevation, sizeof(max_elevation), 0);

    /* See if the user wanted help or template */
    read_count = IAS_PARM_GET_TABLE_SIZE( parameter_table );
    status = ias_parm_provide_help( parameter_filename, parameter_table,
            read_count, IAS_INPUT_PARAMETERS );
    if ( status == ERROR )
        return ERROR;
    else if ( status == 1 )
    {
        /* If the status of the provide help is 1, the user asked for help.
           Since the parameter file name contains the help request, copy it
           to the omf file name so the next call will ask for the same help */
           strcpy( parameter_data->omf_filename, parameter_filename );
    }
    else
    {
        /* Read the parameter file */
        status = ias_parm_read( parameter_filename, parameter_table,
            read_count );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR( "Reading parameters from: %s",
                parameter_filename );
            return ERROR;
        }

        /* Build the OMF file name */
        sprintf( parameter_data->omf_filename, "%s.omf",
            parameter_data->work_order_id );
    }

    /* Again check to see if the user wanted help */
    read_count = IAS_PARM_GET_TABLE_SIZE( omf_parameter_table );
    status = ias_parm_provide_help( parameter_data->omf_filename,
        omf_parameter_table, read_count, IAS_OMF_PARAMETERS );
    if ( status == ERROR )
        return ERROR;
    else if ( status == 1 )
        exit( EXIT_SUCCESS );

    /* Read the OMF file */
    read_count = IAS_PARM_GET_TABLE_SIZE( omf_parameter_table );
    status = ias_parm_read( parameter_data->omf_filename,
        omf_parameter_table, read_count );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Reading parameters from OMF file: %s",
            parameter_data->omf_filename );
        return ERROR;
    }

    /* Check if only one of MIN_ELEVATION and MAX_ELEVATION present */
    if (IAS_PARM_GET_COUNT_READ(omf_parameter_table, MIN_ELEVATION) != 
        IAS_PARM_GET_COUNT_READ(omf_parameter_table, MAX_ELEVATION))
    {
        IAS_LOG_ERROR("Only one of the elevation is present");
        return ERROR;
    }

    /* Store the mean elevation in the param structure -- the OMF has the 
       values in meters and the SCA alignment needs the nominal elevation in
       kilometers hence the divde by 2000. */
    parameter_data->nominal_scene_elevation =
        (double)( min_elevation + max_elevation) / 2000.0;

    /* Initialize the Satellite Attributes Library */
    satellite_id = ias_sat_attr_get_satellite_id_from_satellite_number(
        satellite_number );
    status = ias_sat_attr_initialize( satellite_id );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Initializing IAS Satellite Attributes Library" );
        return ERROR;
    }

    /* Get the total number of bands to process.  If the count is zero the
       BANDS parameter wasn't specified so populate the band list using the
       normal band list. */
    num_bands_to_process = IAS_PARM_GET_COUNT_READ( parameter_table,
        BAND_LIST );
    if ( num_bands_to_process == 0 )
    {
        if (ias_sat_attr_get_sensor_band_numbers(IAS_MAX_SENSORS, 
                IAS_NORMAL_BAND, 0, parameter_data->band_number_list,
                IAS_MAX_NBANDS, &parameter_data->num_bands_to_process)
                != SUCCESS)
        {
            IAS_LOG_ERROR("Getting the list of normal band numbers");
            return ERROR;
        }
    }

    /* Otherwise make sure the band list doesn't contain any
       "abbie-normal" bands */
    else
    {
        parameter_data->num_bands_to_process = num_bands_to_process;
        for ( i = 0; i < num_bands_to_process; i++ )
        {
            if (!ias_sat_attr_band_classification_matches(local_band_list[i],
                    IAS_NORMAL_BAND))
            {
                IAS_LOG_ERROR("Only normal image bands can be processed. "
                    " Band %d is not in supported", local_band_list[i]);
                return ERROR;
            }
            else
            {
                /* Make sure this isn't a duplicate band */
                if ( is_band_already_listed[local_band_list[i]] )
                {
                    IAS_LOG_ERROR("Band %d specified more than once",
                        local_band_list[i] );
                    return ERROR;
                }
                /* Mark that the band is listed and save to data structure */
                is_band_already_listed[local_band_list[i]] = 1;
                parameter_data->band_number_list[i] = local_band_list[i];
            }
        }
    }
    parameter_data->total_normal_bands = ias_sat_attr_get_normal_band_count();
    if (parameter_data->total_normal_bands == ERROR)
    {
        IAS_LOG_ERROR("Failed to retrieve the number of normal bands");
        return ERROR;
    }

    /* Pass the stitch method along */
    parameter_data->stitch_method = local_stitch_method;

    /* Pass the detector offset method along */
    parameter_data->apply_det_offset_flag = local_det_off_flag;

    /* Check if the SCENE_CENTER_LATITUDE is specified in the input ODL file */
    if (IAS_PARM_GET_COUNT_READ(parameter_table, SCENE_CENTER_LATITUDE) == 0)
        parameter_data->scene_center_latitude_odl_flag = 0;
    else
        parameter_data->scene_center_latitude_odl_flag = 1;

    /* Clean up */
    return SUCCESS;
}  /* END read_stitch_parameters */
