/*--------------------------------------------------------------------------
 NAME:     sca_stitch_utility_funcs

 PURPOSE:  Implements internal routines used by the l1r_sca_stitching
           utility application

---------------------------------------------------------------------------*/
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>

#include "ias_const.h"
#include "ias_cpf.h"
#include "ias_logging.h"
#include "ias_satellite_attributes.h"
#include "l1r_sca_stitching.h"
#include "ias_radiometry.h"
#include "ias_l0r.h"

/*-------------------------------------------------------------------------
 NAME:                   get_max_combined_band_sca_offset

 PURPOSE:  Scans the band_offset and sca_offset array's in a populated CPF
           data structure and finds the largest combined offset for the
           specified bands.  The offset returned is in frames.
           Optionally add the maximum detector off set.  The starting
           line adjustment is found at this time as well.  This accounts for
           the detectors that are shifted up.

 RETURNS:  SUCCESS or ERROR

 NOTES:    The offsets may need to be differentiated between the OLI bands
           and the TIRS bands.
---------------------------------------------------------------------------*/
int get_max_combined_band_sca_offset
(
    const struct IAS_CPF_DETECTOR_OFFSETS *det_offsets, 
                            /* I: CPF along track detector offset parameters */
    const int *band_offsets,            /* I: Band offsets */
    const int *sca_offsets,             /* I: SCA offsets */
    int num_bands,                      /* I: Total number of bands */
    int apply_det_offsets_flag,         /* I: Flag to apply the detector 
                                              Offsets */
    int *max_combined_band_sca_offset,  /* O: Largest (band + sca) offset */
    int *start_line_adjustment          /* O: Value that to adjust the top
                                              of the image by to account for
                                              detector offsets */
)
{
    int band_index;         /* Band loop counter */
    int sca_index;          /* Index of first SCA with non-zero offset */
    int det_index;          /* Detector index for the loop */
    double max_sca_offset = INT_MIN;
    double max_band_offset = INT_MIN;
    double max_det_offset = INT_MIN;
    double min_det_offset = INT_MAX;
    double sca_band_offset = 0;/* Add the band and sca offset */
    double absolute_min_det = 0;/* Absolute value of the min detector */
    double temp_adjustment = 0;  /* Adjustment to compare with current */
    double det_offset;      /* Temporary offset */
    double long_offset_adjustment = 0;
    const IAS_BAND_ATTRIBUTES *band_attributes;

    /* Look through all the bands and SCAs to find the largest offset.  If it
       can't be found, return an "obviously invalid" flag value. */
    for ( band_index = 0; band_index < num_bands; band_index++ )
    {
        band_attributes = ias_sat_attr_get_band_attributes( band_index + 1 );
        /* Convert the offset to frames since they are stored as lines in
           the CPF */
        double band_offset = (double)band_offsets[band_index];
        band_offset /= (double)band_attributes->lines_per_frame;
        if ( band_offset > max_band_offset )
        {
            max_band_offset = band_offset;
        }

        for ( sca_index = 0; sca_index < band_attributes->scas;
              sca_index++ )
        {
            /* Convert the offset to frames since they are stored as lines in
               the CPF */
            double sca_offset = (double) sca_offsets[band_index * 
                                IAS_MAX_NSCAS + sca_index];
            sca_offset /= (double)band_attributes->lines_per_frame;
            if ( sca_offset > max_sca_offset )
            {
                max_sca_offset = sca_offset;
            }

            /* Add the detector offsets and get the starting line adjustment */
            if (apply_det_offsets_flag)
            {
                /* Reset the min det offset for this sca */
                min_det_offset = INT_MAX;

                for (det_index = 0; 
                        det_index < band_attributes->detectors_per_sca;
                        det_index++)
                {
                    /* Retrieve the offsets and compare to the min/max */
                    det_offset = det_offsets->
                        along_per_detector[band_index][sca_index][det_index];
                    if (det_offset > max_det_offset )
                    {
                        max_det_offset = det_offset/
                                    (double)band_attributes->lines_per_frame;
                    }
                    if (det_offset < min_det_offset)
                    {
                        min_det_offset = det_offset/
                                    (double)band_attributes->lines_per_frame;
                    }
                }
                /* Determine if there will be an adjustment with the 
                   starting line due to a negative detector adjustment 
                   and record the value.
                   A negative detector offset will push the image sample 
                   up and may requre an adjustment to where the image 
                   data starts in the output image */
                sca_band_offset = sca_offset + band_offset;

                if (min_det_offset < 0)
                {
                    absolute_min_det = fabs(min_det_offset);
                    temp_adjustment = absolute_min_det - sca_band_offset;
                    if (temp_adjustment > long_offset_adjustment)
                    {
                        long_offset_adjustment = temp_adjustment;
                    }
                }
                else /* zero min_det_offset does not effect
                        the top of the image.  Update the offset adjustment
                        if necessary. */
                {
                    temp_adjustment = -sca_band_offset;
                    if (temp_adjustment > long_offset_adjustment)
                    {
                        long_offset_adjustment = temp_adjustment;
                    }
                }
            }
            else /* max_band_offset and max_offset are used here for the 
                    calculation of the adjustment where the imagedata starts 
                    in the output image if no detecor offsets are appled */
            {
                sca_band_offset = sca_offset + band_offset;
                temp_adjustment = -sca_band_offset;

                if (temp_adjustment > long_offset_adjustment)
                {
                    long_offset_adjustment = temp_adjustment;
                }
            }
        }
    }

    if ( max_band_offset == INT_MIN )
    {
        IAS_LOG_ERROR("Could not determine max band offset");
        return ERROR;
    }
    if ( max_sca_offset == INT_MIN )
    {
        IAS_LOG_ERROR("Could not determine max SCA offset");
        return ERROR;
    }

    /* No actions will be necessary if the max det offset is negative or
       if there are no positive maximum offsets or no negative minimum
       offsets */
    if (max_det_offset < 0)
    {
        max_det_offset = 0.0;
    }

    /* Set the starting to adjust to detector offsets that go above the image */
    *start_line_adjustment = (int)floor(long_offset_adjustment +.5);

    /* Add the band and sca offset, the starting line adjustment, and the
       max detector offset. */
    *max_combined_band_sca_offset = 
                (int)ceil(max_band_offset + max_sca_offset) + 
                (int)(max_det_offset + .5) +
                *start_line_adjustment;

    /* Done */
    return SUCCESS;
}  /* END get_max_combined_band_sca_offset */

/*------------------------------------------------------------------------
 NAME:                  update_l1r_metadata

 PURPOSE:  Copies the file metadata from the input image, and appends a
           string specifying the SCA stitching step and the stitching
           method to the custom processing field

 RETURNS:  Integer status code of SUCCESS or ERROR
--------------------------------------------------------------------------*/
int update_l1r_metadata
(
    const IAS_L1R_FILE_METADATA *in_fmd, /* I: Input L1R file metadata */
    STITCH_METHOD_TYPE stitch_method,    /* I: SCA stitching method */
    IAS_L1R_FILE_METADATA *out_fmd       /* O: Updated l1R file metadata */ 
)
{
    /* Local strings for "English" stitching method description */
    char *stitch_method_str;
    char *stitch_method_strings[4] = 
    {
        "SCA stitching -- no SCA overlap",
        "SCA stitching -- left SCA overlap",
        "SCA stitching -- right SCA overlap",
        "SCA stitching -- half SCA overlap"
    };
    /* Buffer sizes */
    size_t fmd_size = sizeof(IAS_L1R_FILE_METADATA);
    size_t old_custom_proc_size;
    size_t new_custom_proc_size;

    /* Copy the existing input metadata to the output metadata structure */
    memcpy( out_fmd, in_fmd, fmd_size );

    /* Point to the appropriate string describing the stitch method */
    stitch_method_str = stitch_method_strings[stitch_method];

    /* Determine the size of the existing processing string and the new one
       containing the string to add.  Add one to the new string to
       accomodate the separating comma. */
    old_custom_proc_size =
        strlen( in_fmd->custom_rad_processing_steps_applied );
    new_custom_proc_size =
        old_custom_proc_size + strlen( stitch_method_str ) + 1;

    /* Can't add onto the list as it's already full */
    if ( new_custom_proc_size >= IAS_L1R_MAX_RAD_STEPS_SIZE )
    {
        IAS_LOG_ERROR("Cannot append stitching method information -- "
                "custom processing list is already full");
        return ERROR;
    }
    else if ( old_custom_proc_size == 0 )  /* Clean list -- just add the
                                            stitching method used */
    {
        strcpy( out_fmd->custom_rad_processing_steps_applied,
            stitch_method_str );
    }
    else
    {

        if ( snprintf( out_fmd->custom_rad_processing_steps_applied,
                IAS_L1R_MAX_RAD_STEPS_SIZE, "%s,%s",
                in_fmd->custom_rad_processing_steps_applied,
                stitch_method_str ) < 0 )
        {
            IAS_LOG_ERROR("Writing stitching method to "
                    "existing custom processing list");
            return ERROR;
        }
    }

    return SUCCESS;
}  /* END update_l1r_metadata */

/*-------------------------------------------------------------------------
NAME:                  get_calculated_sca_offsets

PURPOSE: Gather the needed inputs and call the calculate_sca_offsets
         routine to get the SCA offsets for one band

RETURNS: SUCCESS or ERROR
---------------------------------------------------------------------------*/
int get_calculated_sca_offsets
(
    IAS_CPF  *cpf,                 /* I: IAS_CPF pointer */
    int sensor_id,                 /* I: Sensor ID ("IAS_OLI", "IAS_TIRS") */
    int        band_index,         /* I: 0-based current band index */
    int        number_scas,        /* I: Number of SCAs in current band */
    double scene_center_latitude,  /* I: Scene center latitude */     
    double nominal_scene_elevation,/* Nominal scene elevation */     
    int        *band_offset,       /* O: Band offset in frames */
    int        *sca_offsets        /* O: SCA offsets in frames */
)
{
    const struct IAS_CPF_FOCAL_PLANE *focal_plane;
    const struct IAS_CPF_ORBIT_PARAMETERS *orbit_parameters;
    const struct IAS_CPF_EARTH_CONSTANTS *earth_constants;
    const struct IAS_CPF_OLI_PARAMETERS *oli_parameters;
    const struct IAS_CPF_TIRS_PARAMETERS *tirs_parameters;
    int band_number = ias_sat_attr_convert_band_index_to_number( band_index );
    int lines_per_frame = ias_sat_attr_get_lines_per_frame( band_number );
    double *legendre_coeffs = NULL;
    double nominal_frame_time;

    /* Get the focal plane parameters for the Legendre coefficients */
    focal_plane = ias_cpf_get_focal_plane( cpf );
    if ( !focal_plane )
    {
        IAS_LOG_ERROR("Getting CPF focal plane parameters");
        return ERROR;
    }

    /* Need the orbit parameters for inclination angle, nominal orbit radius
       and the orbital period */
    orbit_parameters = ias_cpf_get_orbit( cpf );
    if ( !orbit_parameters )
    {
        IAS_LOG_ERROR("Getting CPF orbit parameters");
        return ERROR;
    }

    /* Need the earth constants for semi major and semi minor axis, earth
       angular velocity and the eccentricity */
    earth_constants = ias_cpf_get_earth_const( cpf );
    if ( !earth_constants )
    {
        IAS_LOG_ERROR("Getting CPF Earth constants");
        return ERROR;
    }

    /* Now get the sensor specific parameter group.  Pull the nominal
       frame time from that group, and be sure to convert it to seconds
       (as it's in units of milliseconds in the CPF) */
    if (sensor_id == IAS_OLI)
    {
        oli_parameters = ias_cpf_get_oli_parameters(cpf);
        if (oli_parameters == NULL)
        {
            IAS_LOG_ERROR("Getting CPF OLI parameters");
            return ERROR;
        }
        nominal_frame_time = 0.001 * oli_parameters->nominal_frame_time;
    }
    else if (sensor_id == IAS_TIRS)
    {
        tirs_parameters = ias_cpf_get_tirs_parameters(cpf);
        if (tirs_parameters == NULL)
        {
            IAS_LOG_ERROR("Getting CPF TIRS parameters");
            return ERROR;
        }
        nominal_frame_time = 0.001 * tirs_parameters->nominal_frame_time;
    }
    else
    {
        IAS_LOG_ERROR("Invalid sensor ID %d, "
            "should be either 0 for IAS_OLI or 1 for IAS_TIRS", sensor_id);
        return ERROR;
    }

    if (lines_per_frame != 0)
    {
        nominal_frame_time /= lines_per_frame;
    }
    else
    {
        IAS_LOG_ERROR("lines_per_frame is zero which shouldn't be");
        return ERROR;
    }

    /* Allocate the Legendre coefficients data buffer, then load it */
    legendre_coeffs = malloc(number_scas * IAS_LOS_LEGENDRE_TERMS *
        sizeof(double));
    if (legendre_coeffs == NULL)
    {
        IAS_LOG_ERROR("Error allocating Legendre coefficient data buffer");
        return ERROR;
    }

    int sca_index;
    for (sca_index = 0; sca_index < number_scas; sca_index++)
    {
        int i;
        for (i = 0; i < IAS_LOS_LEGENDRE_TERMS; i++)
        {
            int coeff_offset = sca_index * IAS_LOS_LEGENDRE_TERMS + i;
            legendre_coeffs[coeff_offset] =
                focal_plane->along_los_legendre[band_index][sca_index][i];
        }
    }

    /* Call the library routine */
    ias_rps_legendre_calculate_sca_offsets(legendre_coeffs, number_scas,
        scene_center_latitude, nominal_scene_elevation, nominal_frame_time,
        earth_constants->eccentricity,
        0.001 * earth_constants->semi_major_axis,
        0.001 * earth_constants->semi_minor_axis,
        earth_constants->earth_angular_velocity,
        orbit_parameters->inclination_angle,
        orbit_parameters->nominal_orbit_radius,
        orbit_parameters->orbital_period,
        band_offset, sca_offsets);

    /* Clean up */
    free(legendre_coeffs);

    return SUCCESS;
}

/*-------------------------------------------------------------------------
NAME:                  get_l0r_scene_center_latitude

PURPOSE: Gather the needed inputs and call the calculate_sca_offsets
         routine to get the SCA offsets for one band

RETURNS: SUCCESS or ERROR
---------------------------------------------------------------------------*/
int get_l0r_scene_center_latitude
(
    const char *l0r_file_name,        /* I: L0R image filename with path */
    double *scene_center_latitude     /* O: Scene center latitude */
)
{
    L0RIO *l0r_image = NULL;
    int status;         /* function return status */
    int count;          /* number of l0r metadata records */
    char *filename;     /* pointer to l0r file name */
    char path_to_filename[PATH_MAX];  /* l0r path without file name */
    IAS_L0R_SCENE_METADATA scene_metadata;

    /* -------------------- Initialize the L0R library -------------------- */
    /* The l0r_image structure is passed to every routine to identify the
       data.  The L0R file name in the blob needs to be split for the open
       call. */
    strcpy( path_to_filename, l0r_file_name );
    filename = strrchr( path_to_filename, '/' );
    if ( !filename )
    {
        filename = path_to_filename;
    }
    else
    {
        *filename = '\0';
        filename++;
    }

    /* Open the L0R image */
    l0r_image = ias_l0r_open( filename, path_to_filename, 0 );
    if ( !l0r_image )
    {
        IAS_LOG_ERROR("Initializing the L0R data structure");
        return ERROR;
    }

    /* Open the metadata */
    status = ias_l0r_open_metadata( l0r_image, IAS_READ );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Opening the metadata file");
        ias_l0r_close( l0r_image );
        return ERROR;
    }

    /* Get a count of the scene metadata.  If there is 1 scene record, get the
       scene center latitude from the scene metadata. */
    status = ias_l0r_get_scene_metadata_records_count( l0r_image, &count );
    if ( status == SUCCESS && count == 1 )
    {
        status = ias_l0r_get_scene_metadata( l0r_image, 0, 1, &scene_metadata );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Reading the scene metadata for %s", filename );
            ias_l0r_close_metadata( l0r_image );
            ias_l0r_close( l0r_image );
            return ERROR;
        }
        *scene_center_latitude = scene_metadata.scene_center.lat;
    }
    /* There were no scene records or multiple scene records, so derive the 
       center latitude from the interval data */
    else if ( status == SUCCESS && (count == 0 || count > 1))
    {
        int sensor_present_oli = FALSE;
        int sensor_present_tirs = FALSE;

        IAS_L0R_INTERVAL_METADATA l0r_interval_metadata;

        status = ias_l0r_get_interval_metadata( l0r_image,
            &l0r_interval_metadata );
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Reading the interval metadata for %s", filename );
            ias_l0r_close_metadata( l0r_image );
            ias_l0r_close( l0r_image );
            return ERROR;
        }

        /* Find out which sensors are present */
        if (strcmp(l0r_interval_metadata.sensor_id, "OLI_TIRS") == 0)
        {
            sensor_present_oli = TRUE;
            sensor_present_tirs = TRUE;
        }
        else if (strcmp(l0r_interval_metadata.sensor_id, "OLI") == 0)
        {
            sensor_present_oli = TRUE;
            sensor_present_tirs = FALSE;
        }
        else if (strcmp(l0r_interval_metadata.sensor_id, "TIRS") == 0)
        {
            sensor_present_oli = FALSE;
            sensor_present_tirs = TRUE;
        }

        /* Estimate scene center latitude by dividing sum of 4 corner 
           latitudes by 4.  Arbitrarily use OLI if both are present */
        if (sensor_present_oli)
        {
            *scene_center_latitude = 
                (l0r_interval_metadata.corners_oli.upleft.lat
                + l0r_interval_metadata.corners_oli.upright.lat
                + l0r_interval_metadata.corners_oli.loright.lat
                + l0r_interval_metadata.corners_oli.loleft.lat) / 4.0;
        }
        else if (sensor_present_tirs)
        {
            *scene_center_latitude = 
                (l0r_interval_metadata.corners_tirs.upleft.lat
                + l0r_interval_metadata.corners_tirs.upright.lat
                + l0r_interval_metadata.corners_tirs.loright.lat
                + l0r_interval_metadata.corners_tirs.loleft.lat) / 4.0;
        }
        else
        {
            IAS_LOG_ERROR("Neither OLI nor TIRS interval data is present for "
                "%s", filename);
            ias_l0r_close_metadata( l0r_image );
            ias_l0r_close( l0r_image );
            return ERROR;
        }
    }
    else
    {
        if ( status != SUCCESS )
        {
            IAS_LOG_ERROR("Error getting scene metadata count for %s",
                          filename);
        }
        else /* A negative count of scene metadata records was returned */ 
        {
            IAS_LOG_ERROR("Found %d scene metadata records", count);
        }
        ias_l0r_close_metadata( l0r_image );
        ias_l0r_close( l0r_image );
        return ERROR;
    }

    /* Close miscellaneous open stuff */
    if ( ias_l0r_close_metadata( l0r_image ))
    {
        IAS_LOG_ERROR("Closing L0R metadata");
        ias_l0r_close( l0r_image );
        return ERROR;
    }
    status = ias_l0r_close( l0r_image );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Closing L0R file");
        return ERROR;
    }
    else
    {
        l0r_image = NULL;
    }

    return SUCCESS;
}
