/*--------------------------------------------------------------------------
 NAME:     stitch_image

 PURPOSE:  Creates the combined SCA image for the given band according
           to the input/output dimensions and the stitching method
           specified in the processing parameters

 RETURNS:  SUCCESS or ERROR

---------------------------------------------------------------------------*/
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "ias_logging.h"
#include "l1r_sca_stitching.h"

/*--------------------------------------------------------------------------
 NAME:    copy_data

 PURPOSE: Local routine to copy from one buffer to the other
---------------------------------------------------------------------------*/
static void copy_data
(
    float *input_buffer,        /* I: Description should be */
    float *output_buffer,       /* I: obvious from the      */
    int band_offset,            /* I: names                 */
    int sca_offset,             /* I:                       */
    int starting_input_sample,  /* I:                       */
    int number_input_samples,   /* I:                       */
    int number_input_lines,     /* I:                       */
    int number_output_samples,  /* I:                       */
    int starting_output_sample, /* I:                       */
    int apply_det_offsets_flag, /* I:                       */
    const struct IAS_CPF_DETECTOR_OFFSETS *cpf_det_offsets,
    int band_index,
    int sca_index
)
{
    float *input_buffer_ptr;  /* Pointer to one column of input data */
    float *output_buffer_ptr; /* Pointer to one column of output data */
    int input_line;           /* Line index in input buffer */
    int output_line;          /* Line index in output buffer */
    int local_current_output_sample = starting_output_sample;
    int number_bytes = ( number_input_samples - starting_input_sample ) *
        sizeof(float);
    int det_index;
    int input_sample;
    int det_offset;

    input_sample = starting_input_sample;
    output_line = band_offset + sca_offset;

    for ( input_line = 0; input_line < number_input_lines; input_line++ )
    {

        /* Apply detector offsets */
        if (apply_det_offsets_flag)
        {
            for (det_index = 0; det_index < 
                            number_input_samples - starting_input_sample; 
                                            det_index++)
            {
                /* Point to the proper line and sample in the buffers */
                det_offset = (int)(floor(cpf_det_offsets->
                    along_per_detector[band_index][sca_index]
                            [det_index + input_sample] + 0.5));

                input_buffer_ptr = input_buffer  + starting_input_sample + 
                        (det_index + ( input_line * number_input_samples ));

                output_buffer_ptr = output_buffer + 
                        (local_current_output_sample + det_index) +
                        ((output_line + det_offset) * 
                        number_output_samples);
            
                memcpy( output_buffer_ptr, input_buffer_ptr, sizeof(float));
            }
        }
        /* Do not apply detector offsets */
        else
        {
            input_buffer_ptr = input_buffer + starting_input_sample +
                (input_line * number_input_samples);
            output_buffer_ptr = output_buffer + local_current_output_sample +
                (output_line * number_output_samples);

            /* Copy the data */
            memcpy( output_buffer_ptr, input_buffer_ptr, number_bytes );
        }

        output_line++;
    }
}  /* END copy_data */

/****************************************************************************
NAME:  stitch_image

****************************************************************************/

int stitch_image
(
    L1R_BAND_IO *input_band_io,  /* I: Input image data set for this band */
    STITCH_METHOD_TYPE stitch_method, /* I: Stitching information */
    int band_offset,             /* I: Offset from top of image for the band */
    const int *sca_offsets,      /* I: Offsets for each SCA in the band */
    int sca_overlap,             /* I: SCA Overlap width for each the band */
    int num_scas,                /* I: Number of SCAs in the input band */
    int number_input_lines,      /* I: Number of lines in the input image */
    int number_input_samples,    /* I: Number of samples per line in the
                                       input image */
    int number_output_samples,   /* I: Number of samples per line in the
                                       stitched output image */
    int apply_det_offsets_flag, /* I: Flag to apply the detector offsets */
    int band_index,             /* I: Band index for max det offset arr */
    const struct IAS_CPF_DETECTOR_OFFSETS *cpf_det_offsets,
    int start_line_adjustment,   /* I: This value is added to the starting
                                      line of the output image to adjust
                                      to negative detector offsets */
    float *output_buffer         /* I/O: Populated stitched image buffer */
)
{
    float *input_buffer;         /* One SCA of L1R data */
    int sca_index;               /* SCA loop counter */
    int starting_output_sample;  /* Starting sample in the output buffer */
    int starting_input_sample;   /* Starting sample in the input buffer */
    int status;                  /* Return status */
    int half_overlap_remainder = 0;/* Remainder of overlap divided by 2 */

    /* Allocate an image buffer for the SCA input image */
    input_buffer = malloc( number_input_lines * number_input_samples *
        sizeof(*input_buffer));
    if ( input_buffer == NULL )
    {
        IAS_LOG_ERROR("Cannot allocate input image buffer for SCA data");
        return ERROR;
    }

    starting_input_sample = 0;
    starting_output_sample = 0;

    /* The following blocks stitch the SCAs together */
    if ( stitch_method == SCA_STITCH_LEFT_OVERLAP )
    {
        /* The left overlap case copies the last SCA to the output buffer
           first followed by the next to last SCA and so on.  Hence the left
           SCA overlaps the right SCA. */
        for ( sca_index = num_scas - 1; sca_index >= 0; sca_index-- )
        {
            /* Read the current SCA's input image */
            status = ias_l1r_read_image( input_band_io, sca_index, 0, 0,
                number_input_lines, number_input_samples, input_buffer );
            if ( status != SUCCESS )
            {
                IAS_LOG_ERROR("Cannot read SCA %d input image",
                        sca_index + 1 );
                free( input_buffer );
                return ERROR;
            }

            /* Adjust the starting output sample for this SCA.  The amount
               of SCA overlap is subtracted to start copying it so it will
               be overlapped by the left SCA. */
            starting_output_sample = sca_index *
                ( number_input_samples - sca_overlap );

            copy_data( input_buffer, output_buffer, band_offset,
                sca_offsets[sca_index] + start_line_adjustment,
                starting_input_sample,
                number_input_samples, number_input_lines,
                number_output_samples, starting_output_sample,
                apply_det_offsets_flag, cpf_det_offsets, 
                band_index, sca_index );
        }
    }
    else
    {
        /* For the rest of the cases, the SCAs are all copied left to right
           so the right SCA overlaps the left except for the no overlap case
           where, obviously, no overlap is made. */
        for ( sca_index = 0; sca_index < num_scas; sca_index++ )
        {
            /* Read the current SCA's input image */
            status = ias_l1r_read_image( input_band_io, sca_index, 0, 0,
                number_input_lines, number_input_samples, input_buffer );
            if ( status != SUCCESS )
            {
                IAS_LOG_ERROR("Cannot read SCA %d input image",
                        sca_index + 1 );
                free( input_buffer );
                return ERROR;
            }

            copy_data( input_buffer, output_buffer, band_offset,
                sca_offsets[sca_index] + start_line_adjustment, 
                starting_input_sample,
                number_input_samples, number_input_lines,
                number_output_samples, starting_output_sample,
                apply_det_offsets_flag, cpf_det_offsets, 
                band_index, sca_index );

            /* For this method back the starting sample up by subtracting
               the overlap so the next SCA will overlap */
            if ( stitch_method == SCA_STITCH_RIGHT_OVERLAP )
            {
                starting_output_sample += number_input_samples - sca_overlap;
            }

            /* For this overlap method we need to adjust the starting
               input sample to half the overlap. */
            else if ( stitch_method == SCA_STITCH_HALF_OVERLAP )
            {
                /* Add the number of samples just copied to output */
                starting_output_sample += number_input_samples -
                        starting_input_sample;

                /* Calculate the new starting sample in the input buffer
                   For odd numbers, capture the remainder for starting output */
                starting_input_sample = ( sca_overlap / 2);
                half_overlap_remainder = sca_overlap % 2;

                /* And subtract that value from the starting output sample
                   so that the right SCA will overlap the left */
                starting_output_sample -= (starting_input_sample + 
                        half_overlap_remainder);
            }
            else
            {
                /* Otherwise just increment the output sample by the number
                   of input samples so there is no overlap */
                starting_output_sample += number_input_samples;
            }
        }
    }

    /* Clean up */
    free( input_buffer );

    return SUCCESS;
}  /* END stitch_image */
