#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include "ias_l1r.h"
#include "ias_types.h"

int main(int argc, char *argv[])
{
    const char input_directory[PATH_MAX] = 
        "./input";
    const char base_output_directory[PATH_MAX] = 
        "./output";

    char full_input_name[PATH_MAX];
    char full_display_name[PATH_MAX];
    char full_output_name[PATH_MAX];

    FILE *in_fptr = NULL;
    FILE *out_fptr = NULL;

    L1RIO *out_image          = NULL;
    L1R_BAND_IO  *out_band_io = NULL;
    IAS_L1R_FILE_METADATA file_metadata;
    IAS_L1R_BAND_METADATA band_metadata[1];

    int num_bands = 3;
    int num_scas  = 6;
    int num_detectors = 16;
    int num_lines = 80;
    int band_index;
    int sca_index;
    IAS_DATA_TYPE dtype = IAS_R4;
    int out_nl = 80;
    int out_ns = 16;
    int status;
    
    float *out_buffer = NULL;

    /* Allocate the input and output buffers */
    out_buffer = calloc((num_lines * num_detectors), sizeof(*out_buffer));
    if (out_buffer == NULL)
    {
        printf("Cannot allocate input and/or output data buffers\n");
        free(out_buffer);
        exit(1);
    }

    /* Set up the output image for the current band */
    sprintf( full_output_name, "%s/l1r_stitch_test_input_image.h5",
            base_output_directory );

    out_image = ias_l1r_open_image(full_output_name, IAS_WRITE);
    if (out_image == NULL)
    {
        printf("Cannot open %s for writing\n", full_output_name);
        free(out_buffer);
        exit(1);
    }

    /* Process each image */
    for (band_index = 0; band_index < num_bands; band_index++)
    {
        printf("Generating band %d output file...\n", (band_index + 1));

        out_band_io = ias_l1r_open_band(out_image, (band_index + 1),
                                        &dtype, &num_scas, &out_nl,
                                        &out_ns);
        if (out_band_io == NULL)
        {
            printf("Cannot set up band %d for writing\n",
                   (band_index + 1));
            ias_l1r_close_image(out_image);
            free(out_buffer);
            exit(1);
        }

        /* Get the current SCA for the current band, and write it to the
           output image */
        for (sca_index = 0; sca_index < num_scas; sca_index++)
        {
            printf("Reading SCA %d input file...\n", (sca_index + 1));

            sprintf(full_input_name, "%s/testim_stitchB%dS%d",
                    input_directory, (band_index + 1), (sca_index + 1));
            in_fptr = fopen(full_input_name, "r");
            if (in_fptr == NULL)
            {
                printf("Cannot open file %s for reading\n",
                       full_input_name);
                ias_l1r_close_band(out_band_io);
                ias_l1r_close_image(out_image);
                free(out_buffer);
                exit(1);
            }
            fread(out_buffer, sizeof(float), (out_nl * out_ns), in_fptr);

            sprintf(full_display_name, "%s/testim_B%dS%d_display",
                    base_output_directory, (band_index + 1),
                    (sca_index + 1));
            out_fptr = fopen(full_display_name, "w");
            if (out_fptr == NULL)
            {
                printf("Cannot write display image for band %d SCA %d\n",
                       (band_index + 1), (sca_index + 1));
            }
            else
            {
                fwrite(out_buffer, sizeof(float), (out_nl * out_ns),
                       out_fptr);
            }
            fclose(out_fptr);

            /* Write the current SCA's image data to the HDF file */
            status = ias_l1r_write_image(out_band_io, sca_index, 0, 0,
                                         out_nl, out_ns, out_buffer);

            if (status)
            {
                printf("Cannot write SCA image %d to output file\n",
                       (sca_index + 1));
                ias_l1r_close_band(out_band_io);
                ias_l1r_close_image(out_image);
                free(out_buffer);
                exit(1);
            }

            /* Clean out the image buffer */
            memset(out_buffer, 0, (out_nl * out_ns * sizeof(float)));
        }

        /* Create some dummy metadata */
        band_metadata[0].band_number = band_index + 1;
        strcpy( band_metadata[0].band_name, "Dummy band name" );
        strcpy( band_metadata[0].instrument_source, "dummy source" );
        strcpy( band_metadata[0].discontinuity_factor_source, "CPF" );
        band_metadata[0].reflectance_to_radiance_coefficient = 9.9;
        ias_l1r_set_band_metadata( out_image, band_metadata, 1 );

        /* Close up the current band */
        ias_l1r_close_band(out_band_io);
    }

    /* Set some file metadata */
    strcpy( file_metadata.software_version, "SW V-0.0" );
    strcpy( file_metadata.ingest_software_version, "V-1.1" );
    strcpy( file_metadata.collect_type, "EARTH" );
    file_metadata.linearization_correction_applied = 1;
    file_metadata.band_average_gain_applied = 0;
    file_metadata.relative_gain_applied = 1;
    file_metadata.gain_temp_sensitivity_correction_applied = 0;
    file_metadata.sca_discontinuity_correction_applied = 1;
    file_metadata.inoperable_detector_fill_applied = 0;
    file_metadata.saturated_pixel_correction_applied = 1;
    file_metadata.residual_striping_correction_applied = 0;
    file_metadata.reflectance_conversion_applied = 1;
    file_metadata.earth_sun_distance = 1.0856;
    strcpy( file_metadata.custom_rad_processing_steps_applied, "" );
    strcpy( file_metadata.custom_rad_processing_steps_applied,
        "histogram_statistics_characterization, sca_overlap_characterization" );
    status = ias_l1r_set_file_metadata( out_image, &file_metadata );
    if ( status )
        printf("Error setting file metadata");

    ias_l1r_close_image(out_image);

    /* Free remaining allocated resources */
    free(out_buffer);

    return 0;
}
