/*-------------------------------------------------------------------------
  NAME:    l1r_sca_stitching_stitch_image_test

  PURPOSE: Test stitching all four methods with all three bands and
           compare with images generated by Cal/Val prototype code.  There
           is one supporting script, compare_image_results.csh, that does
           a diff of the fields written by the tests against the expected
           results files.
-------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "ias_l1r.h"
#include "ias_miscellaneous.h"
#include "ias_satellite_attributes.h"
#include "ias_const.h"
#include "cgreen.h"
#include "../l1r_sca_stitching.h"


/* Simulated CPF data */
static int num_bands = 3;
static int num_scas = 6;
static int num_detectors = 16;
static int sca_overlaps[] = {4, 4, 4, 4, 4, 4};
static int bands[]   = {1, 2, 3};

static char *test_data_dir = NULL;

/* Allocate and populate an array of offsets */
static int **allocate_sca_offset_array( int nrows, int ncols )
{
    int **array_ptr = NULL;
    int row;
    int col;

    array_ptr = (int **)ias_misc_allocate_2d_array( nrows, ncols,
        sizeof(int));
    if ( array_ptr == NULL )
    {
        assert_true_with_message( 0,
            "Memory allocation failed for SCA offset array");
        return NULL;
    }

    /* Populate the array */
    for ( row = 0; row < nrows; row++ )
    {
        for ( col = 0; col < ncols; col++ )
        {
            if (( col % 2 ) == 0 )
            {
                array_ptr[row][col] = 4 + ( row * 12 );
            }
            else
            {
                array_ptr[row][col] = 0;
            }
        }
    }

    /* Done */
    return array_ptr;
}  /* END allocate_sca_offset_array */


/* This routine takes input from the various tests, opens the band
   and calls the stitch routine.  It then writes out a flat binary
   file for comparison to the Cal/Val test output. */
void test_stitch_image
(
    int band_number,
    const char *output_basename,
    STITCH_METHOD_TYPE stitch_method,
    int number_output_samples
)
{
    const char input_basename[] = "testim";


    char full_input_name[PATH_MAX];
    char full_output_name[PATH_MAX];

    int l1rp_length;
    int sca_index;
    int status;

    L1RIO *in_image          = NULL;
    L1R_BAND_IO *in_band_io  = NULL;
    FILE *out_fptr = NULL;
    float *output_buffer = NULL;

    int **sca_offsets = NULL;
    int band_sca_offsets[6];
    int num_lines = 80;
    int max_sca_offset = 40;
    int band_offsets[] = {18, 12, 6};
    IAS_DATA_TYPE dtype = IAS_R8;


    /* Create the SCA offsets array */
    sca_offsets = allocate_sca_offset_array( num_bands, num_scas );
    if ( sca_offsets == NULL )
    {
        assert_true_with_message( 0, "Cannot create SCA offsets array");
        return;
    }

    /* Allocate the final output buffer */
    l1rp_length = num_lines + max_sca_offset;
    output_buffer = calloc(( l1rp_length * number_output_samples ),
                            sizeof(*output_buffer));
    assert_true( output_buffer != NULL );
    if ( output_buffer == NULL )
    {
        assert_true_with_message( 0, "Cannot allocate output data buffer");
        ias_misc_free_2d_array((void **)sca_offsets );
        return;
    }

    /* Build the input file name */
    sprintf( full_input_name, "%s/ias_base/rps/l1r_sca_stitching/%s_B%d.h5",
            test_data_dir, input_basename, band_number );

    /* Open the input image */
    in_image = ias_l1r_open_image( full_input_name, IAS_READ );
    if ( in_image == NULL )
    {
        assert_true_with_message( 0, "Cannot open %s for reading",
            full_input_name );
        ias_l1r_close_image( in_image );
        ias_misc_free_2d_array((void **)sca_offsets );
        free( output_buffer );
        return;
    }

    /* Open up the current band */
    in_band_io = ias_l1r_open_band( in_image, band_number, &dtype,
                                   &num_scas, &num_lines, &num_detectors );
    if ( in_band_io == NULL )
    {
        assert_true_with_message( 0,
            "Cannot set up band %d information in input image", band_number );
        ias_l1r_close_band( in_band_io );
        ias_l1r_close_image( in_image );
        ias_misc_free_2d_array((void **)sca_offsets );
        free( output_buffer );
        return;
    }

    /* Assemble the band_sca_offsets array */
    for ( sca_index = 0; sca_index < num_scas; sca_index++ )
        band_sca_offsets[sca_index] = sca_offsets[band_number-1][sca_index];

    /* Not used, declare the cpf detector offsets variable */
    struct IAS_CPF_DETECTOR_OFFSETS *cpf_det_offsets = NULL;

    /* Now stitch the test image together */
    status = stitch_image( in_band_io, stitch_method,
        band_offsets[band_number-1], band_sca_offsets, sca_overlaps[0],
        num_scas, num_lines, num_detectors, number_output_samples,
        0, band_number - 1, cpf_det_offsets, 0,
        output_buffer );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Could not generate stitched image");
        ias_l1r_close_band( in_band_io );
        ias_l1r_close_image( in_image );
        free( output_buffer );
        ias_misc_free_2d_array((void **)sca_offsets );
        return;
    }

    /* Output the stitched image in raw form so it can be compared to the
       expected results files */
    sprintf( full_output_name, "./out_stitch_B%d_%s.img", band_number,
        output_basename );
    out_fptr = fopen( full_output_name, "w" );
    if ( out_fptr == NULL )
    {
        assert_true_with_message( 0, "Error opening %s", full_output_name );
        ias_l1r_close_band( in_band_io );
        ias_l1r_close_image( in_image );
        free( output_buffer );
        ias_misc_free_2d_array((void **)sca_offsets );
        return;
    }
    else
    {
        fwrite( output_buffer, sizeof(float), ( l1rp_length *
        number_output_samples ),
               out_fptr );
        if ( fclose( out_fptr ) != 0 )
        {
            assert_true_with_message( 0, "Error closing file %s",
            full_output_name );
        }
    }

    /* Clean up */
    ias_l1r_close_band( in_band_io );
    ias_l1r_close_image( in_image );
    free( output_buffer );
    ias_misc_free_2d_array((void **)sca_offsets );
}  /* END test_stitch_image */


/* The following four routines setup the specific parameters for a given
   stitch method and loop through the bands calling the test routine */
Ensure test_stitch_image_no_overlap( void )
{
    const char output_basename[] = "no_overlap";
    STITCH_METHOD_TYPE stitch_method = SCA_STITCH_NO_OVERLAP;
    int number_output_samples = num_scas * num_detectors;
    int i;

    for ( i = 0; i < num_bands; i++ )
    {
        test_stitch_image( bands[i], output_basename, stitch_method,
            number_output_samples );
    }
}  /* END test_stitch_image_no_overlap */


Ensure test_stitch_image_left_overlap( void )
{
    const char output_basename[] = "left_overlap";
    STITCH_METHOD_TYPE stitch_method = SCA_STITCH_LEFT_OVERLAP;
    int number_output_samples;
    int i;

    number_output_samples = ( num_scas * num_detectors ) -
        (( num_scas - 1 ) * sca_overlaps[0] );

    for ( i = 0; i < num_bands; i++ )
    {
        test_stitch_image( bands[i], output_basename, stitch_method,
            number_output_samples );
    }
}  /* END test_stitch_image_left_overlap */


Ensure test_stitch_image_right_overlap( void )
{
    const char output_basename[] = "right_overlap";
    STITCH_METHOD_TYPE stitch_method = SCA_STITCH_RIGHT_OVERLAP;
    int number_output_samples;
    int i;

    number_output_samples = ( num_scas * num_detectors ) -
        (( num_scas - 1 ) * sca_overlaps[0] );

    for ( i = 0; i < num_bands; i++ )
    {
        test_stitch_image( bands[i], output_basename, stitch_method,
            number_output_samples );
    }
}  /* END test_stitch_image_right_overlap */


Ensure test_stitch_image_half_overlap( void )
{
    const char output_basename[] = "half_overlap";
    STITCH_METHOD_TYPE stitch_method = SCA_STITCH_HALF_OVERLAP;
    int number_output_samples;
    int half_overlap = ( sca_overlaps[0] + 1 ) / 2;
    int i;

    number_output_samples = ( 2 * ( num_detectors - half_overlap )) +
        (( num_scas - 2 ) * ( num_detectors - half_overlap -
        half_overlap ));

    for ( i = 0; i < num_bands; i++ )
    {
        test_stitch_image( bands[i], output_basename, stitch_method,
            number_output_samples );
    }
}  /* END test_stitch_image_half_overlap */


/* This routine calls the supporting shell script to compare the output
   with the expected results. */
Ensure compare_results( void )
{
    char command[] = "./compare_image_results.csh";
    char report_file[] = "./difference_report";
    struct stat buf;
    int status;

    /* Figure out if the output files are good */
    if ( system( command ))
        assert_true_with_message( 0, "Error executing: %s", command );

    /* If the report file exists some error occurred */
    status = stat( report_file, &buf );
    assert_false_with_message( status == 0, "Check the file %s for errors",
        report_file );
}  /* END compare_results */


int main( int argc, char *argv[] )
{
    TestSuite *suite = create_test_suite();

    test_data_dir = getenv("IAS_UNIT_TEST_DATA_DIR");
    if ( !test_data_dir )
    {
        printf( 
            "Could not read environment variable IAS_UNIT_TEST_DATA_DIR\n");
        return EXIT_FAILURE;
    }

    add_test( suite, test_stitch_image_no_overlap );
    add_test( suite, test_stitch_image_left_overlap );
    add_test( suite, test_stitch_image_right_overlap );
    add_test( suite, test_stitch_image_half_overlap );
    add_test( suite, compare_results );

    if ( argc > 1 )
        return run_single_test( suite, argv[1], create_text_reporter());
    else
        return run_test_suite( suite, create_text_reporter());
}
