/*-------------------------------------------------------------------------
  NAME:    l1r_sca_stitching_update_l1r_metadata_test

  PURPOSE: Test updating the L1R metadata with the stitch method
-------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include "ias_l1r.h"
#include "cgreen.h"
#include "../l1r_sca_stitching.h"


Ensure test_update_l1r_metadata()
{
    const char stitch_params_basename[] = "test_stitch_params.odl";

    char *test_data_path = NULL;

    char full_stitch_params_filename[PATH_MAX];
    char stitch_method[256];

    IAS_L1R_FILE_METADATA   in_fmd;
    IAS_L1R_FILE_METADATA  out_fmd;

    SCA_STITCH_PARAMS stitch_params;

    int status;


    /* Get the path to the test data directory */
    test_data_path = getenv("IAS_UNIT_TEST_DATA_DIR");
    assert_true( test_data_path != NULL );
    if ( test_data_path == NULL )
    {
        assert_true_with_message( 0, 
            "Could not read environment variable IAS_UNIT_TEST_DATA_DIR");
        return;
    }

    /* Build the name of the parameter file */
    sprintf( full_stitch_params_filename,
            "%s/ias_base/rps/l1r_sca_stitching/%s", test_data_path,
            stitch_params_basename );

    /* Read the parameter file */
    memset( &stitch_params, 0, sizeof(SCA_STITCH_PARAMS));
    status
        = read_stitch_parameters( full_stitch_params_filename,
                                 &stitch_params );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0,
            "Cannot read processing parameter information from file " "%s",
            full_stitch_params_filename );
        return;
    }

    /* Gin up an input metadata file */
    memset( &in_fmd, 0, sizeof(IAS_L1R_FILE_METADATA));

    sprintf( in_fmd.collect_type, "SOLAR");
    
    in_fmd.band_average_gain_applied = 1;
    in_fmd.relative_gain_applied = 1;
    in_fmd.sca_discontinuity_correction_applied = 1;
    in_fmd.inoperable_detector_fill_applied = 1;
    in_fmd.saturated_pixel_correction_applied = 1;
    in_fmd.residual_striping_correction_applied = 1;

    /* Get the stitching method from the parameter file */
    switch ( stitch_params.stitch_method )
    {
        case SCA_STITCH_NO_OVERLAP:
        {
            snprintf( stitch_method, sizeof(stitch_method),
                     "SCA stitching -- no SCA overlap");
            break;
        }
        case SCA_STITCH_LEFT_OVERLAP:
        {
            snprintf( stitch_method, sizeof(stitch_method),
                     "SCA stitching -- left SCA overlap");
            break;
        }
        case SCA_STITCH_RIGHT_OVERLAP:
        {
            snprintf( stitch_method, sizeof(stitch_method),
                     "SCA stitching -- right SCA overlap");
            break;
        }
        case SCA_STITCH_HALF_OVERLAP:
        {
            snprintf( stitch_method, sizeof(stitch_method),
                     "SCA stitching -- half SCA overlap");
            break;
        }
        default:
        {
            assert_true_with_message( 0, 
                "Invalid stitching method identifier");
            return;
        }
    }

    /* Update the metadata information */
    status = update_l1r_metadata( &in_fmd, stitch_params.stitch_method,
                                 &out_fmd );
    assert_true( status == SUCCESS );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Metadata update failed");
        return;
    }

    /* Check the data */
    assert_string_equal( out_fmd.collect_type, in_fmd.collect_type );

    assert_equal( out_fmd.band_average_gain_applied,
                 in_fmd.band_average_gain_applied );
    assert_equal( out_fmd.relative_gain_applied,
                 in_fmd.relative_gain_applied );

    assert_equal( out_fmd.sca_discontinuity_correction_applied,
                 in_fmd.sca_discontinuity_correction_applied );

    assert_equal( out_fmd.inoperable_detector_fill_applied,
                 in_fmd.inoperable_detector_fill_applied );
    assert_equal( out_fmd.saturated_pixel_correction_applied,
                 in_fmd.saturated_pixel_correction_applied );
    assert_equal( out_fmd.residual_striping_correction_applied,
                 in_fmd.residual_striping_correction_applied );

    /* Now the big one... Check the custom_rad_processing_steps_applied
       field in the output metadata against the string generated
       locally... */
    assert_string_equal( out_fmd.custom_rad_processing_steps_applied,
                        stitch_method );
}  /* END test_update_l1r_metadata */


int main( int argc, char *argv[] )
{
    TestSuite *suite = create_test_suite();

    add_test( suite, test_update_l1r_metadata );

    if ( argc > 1 )
    {
        return run_single_test( suite, argv[1], create_text_reporter());
    }
    return run_test_suite( suite, create_text_reporter());
}
