/***************************************************************************
NAME:    test_l1r_sca_stitching

PURPOSE: Test the entire l1r_sca_stitching application. There is one
    supporting routine, compare_results.csh, that extracts the raw binary
    data from the HDF file and subsets it. After that it does a binary diff
    with the expected results file.
****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <limits.h>

#include "cgreen.h"
#include "../l1r_sca_stitching.h"

static char test_directory[PATH_MAX];

/***************************************************************************
NAME:  create_odl_omf_file

PURPOSE: Creates a parameter odl and OMF file pair with caller specified
    names. It's important that they have specific names so they can be used
    with others simultaneously in the cgreen threaded environment.
****************************************************************************/
int create_odl_omf_file
(
    char *parameter_file_name,
    char *output_file_name,
    STITCH_METHOD_TYPE stitch_method,
    int apply_det_offset_flag,
    double scene_center_latitude,
    int scene_center_latitude_flag,
    int *band_numbers,
    int number_of_bands,
    char *l0r_filename,
    char *l1r_filename,
    char *sensor,
    char *cpf_filename
)
{
    FILE *ouf;
    char temp_file_name[PATH_MAX];
    int i;
    int status;

    /* Create the parameter file */
    status = snprintf(temp_file_name, sizeof(temp_file_name), "%s.odl",
            parameter_file_name);
    if (status < 0 || status >= sizeof(temp_file_name))
    {
        assert_true_with_message(false, "String length exceeds buffer size");
        return true;
    }
    ouf = fopen(temp_file_name, "w");
    if (!ouf)
    {
        assert_true_with_message(false, "Error opening: %s",
                parameter_file_name);
        return true;
    }

    fprintf(ouf, "OBJECT = %s\n", parameter_file_name);
    fprintf(ouf, "WORK_ORDER_ID = \"%s\"\n"
            "SCRIPT_NAME = L1R_SCA_STITCHING\nBAND_LIST = (",
            parameter_file_name);
    for (i = 0; i < number_of_bands - 1; i++)
        fprintf(ouf, "%d,", band_numbers[i]);
    fprintf(ouf, "%d)\n", band_numbers[i]);
    fprintf(ouf, "STITCH_METHOD = %d\n", stitch_method);
    fprintf(ouf, "STITCHED_FILENAME = \"%s\"\n", output_file_name);
    fprintf(ouf, "APPLY_DETECTOR_OFFSETS_FLAG = %d\n",
            apply_det_offset_flag);
    if (scene_center_latitude_flag)
    {
        fprintf(ouf, "SCENE_CENTER_LATITUDE = %f\n", scene_center_latitude);
    }
    fprintf(ouf, "END_OBJECT = %s\n", parameter_file_name);
    fprintf(ouf, "END\n");
    fclose(ouf);

    /* Create the OMF file */
    status = snprintf(temp_file_name, sizeof(temp_file_name), "%s.omf",
            parameter_file_name);
    if (status < 0 || status >= sizeof(temp_file_name))
    {
        assert_true_with_message(false, "String length exceeds buffer size");
        return true;
    }
    ouf = fopen(temp_file_name, "w");
    if (!ouf)
    {
        assert_true_with_message(false, "Error opening: test_omf.omf");
        return true;
    }

    fprintf(ouf, "OBJECT = OMF\n");
    fprintf(ouf, "L0R_FILENAME = \"%s/%s\"\n", test_directory, l0r_filename);
    fprintf(ouf, "L1R_FILENAME = \"%s/%s\"\n", test_directory, l1r_filename);
    fprintf(ouf, "CAL_PARM_FILENAME = \"%s/%s\"\n"
            "SENSOR = %s\n", test_directory, cpf_filename, sensor);
    fprintf(ouf, "END_OBJECT = OMF\n");
    fprintf(ouf, "END\n");
    fclose(ouf);

    return false;
}  /* END create_odl_omf_file */

/***************************************************************************
NAME:  test_the_application

PURPOSE: Actually execute the application with the specified parameter file
****************************************************************************/
int test_the_application
(
    char *parameter_file_name
)
{
    char run_app[PATH_MAX];
    int status;

    /* Execute the app */
    status = snprintf(run_app, sizeof(run_app), "../l1r_sca_stitching %s.odl",
            parameter_file_name);
    if (status < 0 || status >= sizeof(run_app))
    {
        assert_true_with_message(false, "String length exceeds buffer size");
        return true;
    }
    if (system(run_app))
    {
        assert_true_with_message(false, "Error executing: %s", run_app);
        return true;
    }

    /* Clean up */
    return false;
}  /* END test_the_application */

/***************************************************************************
  The main test suite of routines are below. In the following four test
  cases it's important that parameter ODL and OMF file names, as well as
  output file names, are unique to the test. They are written separately
  so they can be run simultaneously by the cgreen threads.

  Each test calls the supporting shell script to extract the raw binary
  data from the HDF, subset that to a size suitable for comparing with
  the expected results in the unit test directory. If errors occur in
  those comparisons, the offending files aren't removed. In the absence
  of errors all files are cleaned up.
****************************************************************************/

/***************************************************************************
NAME:  no_overlap_test

PURPOSE: This test is for the no overlap case. It also test that a single
    band can be stitched.
****************************************************************************/
Ensure no_overlap_test()
{
    char parameter_file[] = "no_overlap";
    char output_file[] = "no_overlap.h5";
    char command[3 * PATH_MAX];
    STITCH_METHOD_TYPE method = SCA_STITCH_NO_OVERLAP;
    double scene_center_latitude = 45.0;
    int scene_center_latitude_flag = 1;
    int band_list[] = {2};
    int nbands = sizeof(band_list) / sizeof(int);
    int apply_det_offset_flag = 0;
    char l0r_filename[] = "";
    char l1r_filename[] = "l1r_stitch_test_input_image.h5";
    char sensor[] = "L8_OLITIRS";
    char cpf_filename[] = "partial_cpf";
    int status;

    /* Set up the input parameter file */
    if (create_odl_omf_file(parameter_file, output_file, method,
                apply_det_offset_flag, scene_center_latitude, 
                scene_center_latitude_flag, band_list, nbands, l0r_filename, 
                l1r_filename, sensor, cpf_filename))
    {
        assert_true_with_message(false, "Error creating parameters");
        return;
    }

    /* Call the test */
    if (test_the_application(parameter_file))
    {
        assert_true_with_message(false, "Error from test_the_application");
        return;
    }
    else
    {
        assert_true(true);
    }

    /* Check the return */
    status = snprintf(command, sizeof(command),
            "./compare_results.csh no %s %d %d %d", output_file, band_list[0],
            96, 120);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(false, "String length exceeds buffer size");
        return;
    }
    if (system(command))
    {
        assert_true_with_message(false, "Error in no overlap case B%02d",
                band_list[0]);
        return;
    }
    else
    {
        assert_true(true);
    }

    /* Clean up */
    status = snprintf(command, sizeof(command), "rm -rf %s.odl %s.omf %s",
            parameter_file, parameter_file, output_file);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(false, "String length exceeds buffer size");
        return;
    }
    if (system(command))
    {
        assert_true_with_message(false, "Error cleaning up files");
        return;
    }
    else
    {
        assert_true(true);
    }
}  /* END no_overlap_test */

/***************************************************************************
NAME:  left_overlap_test

PURPOSE: This test is for left overlap. It tests creation of two bands in
    an arbitrary order.
****************************************************************************/
Ensure left_overlap_test()
{
    char parameter_file[] = "left_overlap";
    char output_file[] = "left_overlap_stitch.h5";
    char command[3 * PATH_MAX];
    STITCH_METHOD_TYPE method = SCA_STITCH_LEFT_OVERLAP;
    double scene_center_latitude = 45.0;
    int scene_center_latitude_flag = 1;
    int band_list[] = {3, 1};
    int nbands = sizeof(band_list) / sizeof(int);
    int i;
    int apply_det_offset_flag = 0;
    char l0r_filename[] = "";
    char l1r_filename[] = "l1r_stitch_test_input_image.h5";
    char sensor[] = "L8_OLITIRS";
    char cpf_filename[] = "partial_cpf";
    int status;

    /* Set up the input parameter file */
    if (create_odl_omf_file(parameter_file, output_file, method,
                apply_det_offset_flag, scene_center_latitude, 
                scene_center_latitude_flag, band_list, nbands, l0r_filename, 
                l1r_filename, sensor, cpf_filename))
    {
        assert_true_with_message(false, "Error creating parameters");
        return;
    }

    /* Call the test */
    if (test_the_application(parameter_file))
    {
        assert_true_with_message(false, "Error from test_the_application");
        return;
    }

    /* Check the return */
    for (i = 0; i < nbands; i++)
    {
        status = snprintf(command, sizeof(command),
                "./compare_results.csh left %s %d %d %d", output_file,
                band_list[i], 76, 120);
        if (status < 0 || status >= sizeof(command))
        {
            assert_true_with_message(false,
                    "String length exceeds buffer size");
            return;
        }
        if (system(command))
        {
            assert_true_with_message(false, "Error in left overlap case B%02d",
                    band_list[i]);
            return;
        }
        else
        {
            assert_true(true);
        }
    }

    /* Clean up */
    status = snprintf(command, sizeof(command), "rm -rf %s.odl %s.omf %s",
            parameter_file, parameter_file, output_file);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(false, "String length exceeds buffer size");
        return;
    }
    if (system(command))
    {
        assert_true_with_message(false, "Error cleaning up files");
        return;
    }
    else
    {
        assert_true(true);
    }
}  /* END left_overlap_test */

/***************************************************************************
NAME:  right_overlap_test

PURPOSE: Test the right overlap method and use all three bands.
****************************************************************************/
Ensure right_overlap_test()
{
    char parameter_file[] = "right_overla"; /* used as WOID, <= 12 chars long */
    char output_file[] = "right_overlap_stitch.h5";
    char command[3 * PATH_MAX];
    STITCH_METHOD_TYPE method = SCA_STITCH_RIGHT_OVERLAP;
    double scene_center_latitude = 45.0;
    int scene_center_latitude_flag = 1;
    int band_list[] = {1, 2, 3};
    int nbands = sizeof(band_list) / sizeof(int);
    int i;
    int apply_det_offset_flag = 0;
    char l0r_filename[] = "";
    char l1r_filename[] = "l1r_stitch_test_input_image.h5";
    char sensor[] = "L8_OLITIRS";
    char cpf_filename[] = "partial_cpf";
    int status;

    /* Set up the input parameter file */
    if (create_odl_omf_file(parameter_file, output_file, method,
                apply_det_offset_flag, scene_center_latitude, 
                scene_center_latitude_flag, band_list, nbands, l0r_filename, 
                l1r_filename, sensor, cpf_filename))
    {
        assert_true_with_message(false, "Error creating parameters");
        return;
    }

    /* Call the test */
    if (test_the_application(parameter_file))
    {
        assert_true_with_message(false, "Error from test_the_application");
        return;
    }

    /* Check the return */
    for (i = 0; i < nbands; i++)
    {
        status = snprintf(command, sizeof(command),
                "./compare_results.csh right %s %d %d %d", output_file,
                band_list[i], 76, 120);
        if (status < 0 || status >= sizeof(command))
        {
            assert_true_with_message(false,
                    "String length exceeds buffer size");
            return;
        }
        if (system(command))
        {
            assert_true_with_message(false, "Error in right overlap case B%02d",
                    band_list[i]);
            return;
        }
        else
        {
            assert_true(true);
        }
    }

    /* Clean up */
    status = snprintf(command, sizeof(command), "rm -rf %s.odl %s.omf %s",
            parameter_file, parameter_file, output_file);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(false, "String length exceeds buffer size");
        return;
    }
    if (system(command))
    {
        assert_true_with_message(false, "Error cleaning up files");
        return;
    }
    else
    {
        assert_true(true);
    }
}  /* END right_overlap_test */

/***************************************************************************
NAME:  half_overlap_test

PURPOSE: Test the half overlap using all three bands
****************************************************************************/
Ensure half_overlap_test()
{
    char parameter_file[] = "half_overlap";
    char output_file[] = "half_overlap_stitch.h5";
    char command[3 * PATH_MAX];
    STITCH_METHOD_TYPE method = SCA_STITCH_HALF_OVERLAP;
    double scene_center_latitude = 45.0;
    int scene_center_latitude_flag = 1;
    int band_list[] = {1, 2, 3};
    int nbands = sizeof(band_list) / sizeof(int);
    int i;
    int apply_det_offset_flag = 0;
    char l0r_filename[] = "";
    char l1r_filename[] = "l1r_stitch_test_input_image.h5";
    char sensor[] = "L8_OLITIRS";
    char cpf_filename[] = "partial_cpf";
    int status;

    /* Set up the input parameter file */
    if (create_odl_omf_file(parameter_file, output_file, method,
                apply_det_offset_flag, scene_center_latitude, 
                scene_center_latitude_flag, band_list, nbands, l0r_filename, 
                l1r_filename, sensor, cpf_filename))
    {
        assert_true_with_message(false, "Error creating parameters");
        return;
    }

    /* Call the test */
    if (test_the_application(parameter_file))
    {
        assert_true_with_message(false, "Error from test_the_application");
        return;
    }

    /* Check the return */
    for (i = 0; i < nbands; i++)
    {
        status = snprintf(command, sizeof(command),
                "./compare_results.csh half %s %d %d %d", output_file,
                band_list[i], 76, 120);
        if (status < 0 || status >= sizeof(command))
        {
            assert_true_with_message(false,
                    "String length exceeds buffer size");
            return;
        }
        if (system(command))
        {
            assert_true_with_message(false, "Error in half overlap case B%02d",
                    band_list[i]);
            return;
        }
        else
        {
            assert_true(true);
        }
    }

    /* Clean up */
    status = snprintf(command, sizeof(command), "rm -rf %s.odl %s.omf %s",
            parameter_file, parameter_file, output_file);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(false, "String length exceeds buffer size");
        return;
    }
    if (system(command))
    {
        assert_true_with_message(false, "Error cleaning up files");
        return;
    }
    else
    {
        assert_true(true);
    }
}  /* END half_overlap_test */

/***************************************************************************
NAME:  overlap_spacing_test

PURPOSE: Test that the overlap spacing is consistent for all stich types.
    This will happen if the last sca is the same for all types. This test
    takes the same sample of the image for each type and diffs the results
    with each other. They should be identical.
****************************************************************************/
Ensure overlap_spacing_test()
{
    char parameter_file[PATH_MAX];  /* Attach type to base parameter file */
    char stitching_output[PATH_MAX];
    char binary_output[PATH_MAX];
    char stitch_type[3][10];
    char cmd[3 * PATH_MAX];
    int type_index;
    int status;

    strcpy(stitch_type[0], "half");
    strcpy(stitch_type[1], "right");
    strcpy(stitch_type[2], "left");

    /* Run the l1r_sca_stitching routine on each stitch type */
    for (type_index = 0; type_index < 3; type_index++)
    {
        status = snprintf(parameter_file, sizeof(parameter_file),
                "overlap_spacing_test_%s", stitch_type[type_index]);
        if (status < 0 || status >= sizeof(parameter_file))
        {
            assert_true_with_message(false,
                    "String length exceeds buffer size");
            return;
        }
        /* Run the stitcher for this stitch type */
        if (test_the_application(parameter_file))
        {
            assert_true_with_message(false, "Error from test_the_application");
            return;
        }

        status = snprintf(binary_output, sizeof(binary_output),
                "./create_image/output/bin_overlap_spacing_%s",
                stitch_type[type_index]);
        if (status < 0 || status >= sizeof(binary_output))
        {
            assert_true_with_message(false,
                    "String length exceeds buffer size");
            return;
        }
        status = snprintf(stitching_output, sizeof(stitching_output),
                "./create_image/output/%s.h5", parameter_file);
        if (status < 0 || status >= sizeof(stitching_output))
        {
            assert_true_with_message(false,
                    "String length exceeds buffer size");
            return;
        }

        /* Clip a section of the last sca for comparison. Band 4 was selected
           at random. The odl files will be set accordingly. */
        status = snprintf(cmd, sizeof(cmd),
                "./get_binary_section.csh %s %s %s %d %d %d %d", binary_output,
                stitching_output, "B04", 6250, 275, 5750, 480);
        if (status < 0 || status >= sizeof(cmd))
        {
            assert_true_with_message(false,
                    "String length exceeds buffer size");
            return;
        }
        if (system(cmd))
        {
            assert_true_with_message(false, "Error cropping out image section");
            return;
        }
        else
        {
            assert_true(true);
        }
    }

    /* Compare the binary files. They should be the same. */
    strcpy(cmd, "./compare_binary_files.csh "
            "./create_image/output/bin_overlap_spacing_half "
            "./create_image/output/bin_overlap_spacing_right");
    if (system(cmd))
    {
        assert_true_with_message(false,
                "Differences occur between stitch types");
        return;
    }
    else
    {
        assert_true(true);
    }

    strcpy(cmd, "./compare_binary_files.csh "
            "./create_image/output/bin_overlap_spacing_half "
            "./create_image/output/bin_overlap_spacing_left");
    if (system(cmd))
    {
        assert_true_with_message(false,
                "Differences occur between stitch types");
        return;
    }
    else
    {
        assert_true(true);
    }

    /* Clean up overlap spacing test files */
    status = snprintf(cmd, sizeof(cmd),
            "rm -rf ./create_image/output/*overlap_spacing*");
    if (status < 0 || status >= sizeof(cmd))
    {
        assert_true_with_message(false, "String length exceeds buffer size");
        return;
    }
    if (system(cmd))
    {
        assert_true_with_message(false, "Error cleaning up files");
        return;
    }
    else
    {
        assert_true(true);
    }

    return;
}  /* END overlap_spacing_test */

/***************************************************************************
NAME:  detector_offset_off_test

PURPOSE: Run l1r_sca_stitching for one band with the detector offset off.
    Then compare a section with the expected results. Use no overlap and
    right overlap as random stitch types. The input image will be contain
    valid data and has been visually inspected to verify that the results
    are correct.
****************************************************************************/
Ensure detector_offset_off_test()
{
    char parameter_file[PATH_MAX];  /* Attach type to base parameter file */
    char stitching_output[PATH_MAX];
    char binary_output[PATH_MAX];
    char stitch_type[2][11];
    char cmd[3 * PATH_MAX];
    int type_index;
    int status;

    strcpy(stitch_type[0], "no_overlap");
    strcpy(stitch_type[1], "right");

    /* Run the l1r_sca_stitching routine on each stitch type */
    for (type_index = 0; type_index < 2; type_index++)
    {
        status = snprintf(parameter_file, sizeof(parameter_file),
                "no_detector_test_%s", stitch_type[type_index]);
        if (status < 0 || status >= sizeof(parameter_file))
        {
            assert_true_with_message(false,
                    "String length exceeds buffer size");
            return;
        }

        /* Run the stitcher for this stitch type */
        if (test_the_application(parameter_file))
        {
            assert_true_with_message(false, "Error from test_the_application");
            return;
        }

        status = snprintf(binary_output, sizeof(binary_output),
                "./create_image/output/bin_no_detector_%s",
                stitch_type[type_index]);
        if (status < 0 || status >= sizeof(binary_output))
        {
            assert_true_with_message(false,
                    "String length exceeds buffer size");
            return;
        }
        status = snprintf(stitching_output, sizeof(stitching_output),
                "./create_image/output/%s.h5", parameter_file);
        if (status < 0 || status >= sizeof(stitching_output))
        {
            assert_true_with_message(false,
                    "String length exceeds buffer size");
            return;
        }

        /* Clip a sample section for comparison. Band 5 was selected
           at random. The odl files will be set accordingly. */
        status = snprintf(cmd, sizeof(cmd),
                "./get_binary_section.csh %s %s %s %d %d %d %d", binary_output,
                stitching_output, "B05", 0, 1550, 0, 7600);
        if (status < 0 || status >= sizeof(cmd))
        {
            assert_true_with_message(false,
                    "String length exceeds buffer size");
            return;
        }
        if (system(cmd))
        {
            assert_true_with_message(false, "Error cropping out image section");
            return;
        }
        else
        {
            assert_true(true);
        }

        /* Compare the binary files. They should be the same. */
        status = snprintf(cmd, sizeof(cmd), "./compare_binary_files.csh "
                "%s/expected_results/bin_no_detector_%s "
                "./create_image/output/bin_no_detector_%s", test_directory,
                stitch_type[type_index], stitch_type[type_index]);
        if (status < 0 || status >= sizeof(cmd))
        {
            assert_true_with_message(false,
                    "String length exceeds buffer size");
            return;
        }
        if (system(cmd))
        {
            assert_true_with_message(false,
                    "Differences occur between stitch types");
            return;
        }
        else
        {
            assert_true(true);
        }
    }

    /* Clean up overlap spacing test files */
    status = snprintf(cmd, sizeof(cmd),
            "rm -rf ./create_image/output/*no_detector*");
    if (status < 0 || status >= sizeof(cmd))
    {
        assert_true_with_message(false, "String length exceeds buffer size");
        return;
    }
    if (system(cmd))
    {
        assert_true_with_message(false, "Error cleaning up files");
        return;
    }
    else
    {
        assert_true(true);
    }

    return;
}  /* END detector_offset_off_test */

/***************************************************************************
NAME:  detector_on_half_test

PURPOSE: Run l1r_sca_stitching for three bands with the detector offset on.
    Then compare a section with the expected results. Use half stitch types.
    The input image will contain valid data and has been visually inspected,
    including band to band comparisons, to verify that the results are correct.
****************************************************************************/
Ensure detector_on_half_test()
{
    char parameter_file[] = "detector_on_half_test";
    char stitching_output[PATH_MAX];
    char binary_output[] = "./create_image/output/bin_detector_on_half";
    char cmd[3 * PATH_MAX];
    int band_index;
    int status;

    /* Run the stitcher for this stitch type */
    if (test_the_application(parameter_file))
    {
        assert_true_with_message(false, "Error from test_the_application");
        return;
    }

    /* Set up the file names */
    status = snprintf(stitching_output, sizeof(stitching_output),
            "./create_image/output/%s.h5", parameter_file);
    if (status < 0 || status >= sizeof(stitching_output))
    {
        assert_true_with_message(false, "String length exceeds buffer size");
        return;
    }

    /* Step through the bands and get section to test */
    for (band_index = 1; band_index <= 3; band_index++)
    {
        /* Clip a sample section for comparison. Bands 1, 2, and 3 were
           selected at random. The odl file will be set accordingly. */
        status = snprintf(cmd, sizeof(cmd),
                "./get_binary_section.csh %s_%02d %s B%02d %d %d %d %d",
                binary_output, band_index, stitching_output, band_index, 0,
                1550, 0, 7600);
        if (status < 0 || status >= sizeof(cmd))
        {
            assert_true_with_message(false,
                    "String length exceeds buffer size");
            return;
        }
        if (system(cmd))
        {
            assert_true_with_message(false, "Error cropping out image section");
            return;
        }
        else
        {
            assert_true(true);
        }

        /* Compare the binary files with the expected results. */
        status = snprintf(cmd, sizeof(cmd), "./compare_binary_files.csh "
                "%s/expected_results/bin_detector_on_half_%02d "
                "./create_image/output/bin_detector_on_half_%02d",
                test_directory, band_index, band_index);
        if (status < 0 || status >= sizeof(cmd))
        {
            assert_true_with_message(false,
                    "String length exceeds buffer size");
            return;
        }
        if (system(cmd))
        {
            assert_true_with_message(false, "Differences occur between "
                    "detector corrected half stitch type");
            return;
        }
        else
        {
            assert_true(true);
        }
    }

    /* Clean up overlap spacing test files */
    status = snprintf(cmd, sizeof(cmd),
            "rm -rf ./create_image/output/*detector_on_half*");
    if (status < 0 || status >= sizeof(cmd))
    {
        assert_true_with_message(false, "String length exceeds buffer size");
        return;
    }
    if (system(cmd))
    {
        assert_true_with_message(false, "Error cleaning up files");
        return;
    }
    else
    {
        assert_true(true);
    }

    return;
}  /* END detector_on_half_test */

/***************************************************************************
NAME:  detector_on_left_test

PURPOSE: Run l1r_sca_stitching for two bands (including pan band) with the
    detector offset on. Use left stitch types. Then compare a section with
    the expected results. The input image will contain valid data and has
    been visually inspected to verify that the results are correct.
****************************************************************************/
Ensure detector_on_left_test()
{
    char parameter_file[] = "detector_on_left_test";
    char stitching_output[PATH_MAX];
    char binary_output[] = "./create_image/output/bin_detector_on_left";
    char cmd[3 * PATH_MAX];
    int band_index;
    int sample_count;
    int line_count;
    int status;

    /* Run the stitcher for this stitch type */
    if (test_the_application(parameter_file))
    {
        assert_true_with_message(false, "Error from test_the_application");
        return;
    }

    /* Set up the file names */
    status = snprintf(stitching_output, sizeof(stitching_output),
            "./create_image/output/%s.h5", parameter_file);
    if (status < 0 || status >= sizeof(stitching_output))
    {
        assert_true_with_message(false, "String length exceeds buffer size");
        return;
    }

    /* Step through bands 8 and 9 and get section to test */
    for (band_index = 7; band_index <= 8; band_index++)
    {
        /* Set the size of the sample, band 8 is pan band so it's larger */
        if (band_index == 8)
        {
            sample_count = 3000;
            line_count = 14600;
        }
        else
        {
            sample_count = 1550;
            line_count = 7600;
        }

        /* Clip a sample section for comparison. Bands 7 and Pan band 8 were
           selected at random. The odl file will be set accordingly. */
        status = snprintf(cmd, sizeof(cmd),
                "./get_binary_section.csh %s_%02d %s B%02d %d %d %d %d",
                binary_output, band_index, stitching_output, band_index, 0,
                sample_count, 0, line_count);
        if (status < 0 || status >= sizeof(cmd))
        {
            assert_true_with_message(false,
                    "String length exceeds buffer size");
            return;
        }
        if (system(cmd))
        {
            assert_true_with_message(false, "Error cropping out image section");
            return;
        }
        else
        {
            assert_true(true);
        }

        /* Compare the binary files with the expected results. */
        status = snprintf(cmd, sizeof(cmd), "./compare_binary_files.csh "
                "%s/expected_results/bin_detector_on_left_%02d "
                "./create_image/output/bin_detector_on_left_%02d",
                test_directory, band_index, band_index);
        if (status < 0 || status >= sizeof(cmd))
        {
            assert_true_with_message(false,
                    "String length exceeds buffer size");
            return;
        }
        if (system(cmd))
        {
            assert_true_with_message(false, "Differences occur between "
                    "detector corrected left stitch type");
            return;
        }
        else
        {
            assert_true(true);
        }
    }

    /* Clean up overlap spacing test files */
    status = snprintf(cmd, sizeof(cmd),
            "rm -rf ./create_image/output/*detector_on_left*");
    if (status < 0 || status >= sizeof(cmd))
    {
        assert_true_with_message(false, "String length exceeds buffer size");
        return;
    }
    if (system(cmd))
    {
        assert_true_with_message(false, "Error cleaning up files");
        return;
    }
    else
    {
        assert_true(true);
    }

    return;
}  /* END detector_on_left_test */

/***************************************************************************
NAME:  detector_on_right_test

PURPOSE: Run l1r_sca_stitching for two bands (including TIRS band) with the
    detector offset on. Use right stitch types. Then compare a section with
    the expected results. The input image will contain valid data and has
    been visually inspected to verify that the results are correct.
****************************************************************************/
Ensure detector_on_right_test()
{
    char parameter_file[] = "detector_on_right_test";
    char stitching_output[PATH_MAX];
    char binary_output[] = "./create_image/output/bin_detector_on_right";
    char cmd[3 * PATH_MAX];
    int band_index;
    int start_sample;
    int sample_count;
    int start_line;
    int line_count;
    int status;

    /* Run the stitcher for this stitch type */
    if (test_the_application(parameter_file))
    {
        assert_true_with_message(false, "Error from test_the_application");
        return;
    }

    /* Set up the file names */
    status = snprintf(stitching_output, sizeof(stitching_output),
            "./create_image/output/%s.h5", parameter_file);
    if (status < 0 || status >= sizeof(stitching_output))
    {
        assert_true_with_message(false, "String length exceeds buffer size");
        return;
    }

    /* Step through bands 9 and 10 and get section to test */
    for (band_index = 9; band_index <= 10; band_index++)
    {
        /* Set the size of the sample, band 10 is pan band so it's smaller */
        if (band_index == 10)
        {
            start_line = 0;
            sample_count = 1849;
            start_sample = 0;
            line_count = 3320;
        }
        else
        {
            start_sample = 4650;
            sample_count = 1940;
            start_line = 0;
            line_count = 7705;
        }

        /* Clip a sample section for comparison. Bands 9 and TIRS band 10
           were selected at random. The odl file will be set accordingly. */
        status = snprintf(cmd, sizeof(cmd),
                "./get_binary_section.csh %s_%02d %s B%02d %d %d %d %d",
                binary_output, band_index, stitching_output, band_index,
                start_sample, sample_count, start_line, line_count);
        if (status < 0 || status >= sizeof(cmd))
        {
            assert_true_with_message(false,
                    "String length exceeds buffer size");
            return;
        }
        if (system(cmd))
        {
            assert_true_with_message(false, "Error cropping out image section");
            return;
        }
        else
        {
            assert_true(true);
        }

        /* Compare the binary files with the expected results. */
        status = snprintf(cmd, sizeof(cmd), "./compare_binary_files.csh "
                "%s/expected_results/bin_detector_on_right_%02d "
                "./create_image/output/bin_detector_on_right_%02d",
                test_directory, band_index, band_index);
        if (status < 0 || status >= sizeof(cmd))
        {
            assert_true_with_message(false,
                    "String length exceeds buffer size");
            return;
        }
        if (system(cmd))
        {
            assert_true_with_message(false, "Differences occur between "
                    "detector corrected right stitch type");
            return;
        }
        else
        {
            assert_true(true);
        }
    }

    /* Clean up overlap spacing test files */
    status = snprintf(cmd, sizeof(cmd),
            "rm -rf ./create_image/output/*detector_on_right*");
    if (status < 0 || status >= sizeof(cmd))
    {
        assert_true_with_message(false, "String length exceeds buffer size");
        return;
    }
    if (system(cmd))
    {
        assert_true_with_message(false, "Error cleaning up files");
        return;
    }
    else
    {
        assert_true(true);
    }

    return;
}  /* END detector_on_right_test */

/***************************************************************************
NAME:  detector_on_no_overlap_test

PURPOSE: Run l1r_sca_stitching for three bands with the detector offset on.
    Use no_overlap stitch types. Then compare a section with the expected
    results. The input image will contain valid data and has been visually
    inspected to verify that the results are correct.
****************************************************************************/
Ensure detector_on_no_overlap_test()
{
    char parameter_file[] = "detector_on_no_overlap_test";
    char stitching_output[PATH_MAX];
    char binary_output[] = "./create_image/output/bin_detector_on_no_overlap";
    char cmd[3 * PATH_MAX];
    int band_index;
    int start_sample = 2860;
    int sample_count = 1210;
    int start_line = 0;
    int line_count = 7670;
    int status;

    /* Run the stitcher for this stitch type */
    if (test_the_application(parameter_file))
    {
        assert_true_with_message(false, "Error from test_the_application");
        return;
    }

    /* Set up the file names */
    status = snprintf(stitching_output, sizeof(stitching_output),
            "./create_image/output/%s.h5", parameter_file);
    if (status < 0 || status >= sizeof(stitching_output))
    {
        assert_true_with_message(false, "String length exceeds buffer size");
        return;
    }

    /* Step through bands 4 through 6 and get section to test */
    for (band_index = 4; band_index <= 6; band_index++)
    {
        /* Clip a sample section for comparison. Bands 4, 5, and 6 were
           selected at random. The odl file will be set accordingly. */
        status = snprintf(cmd, sizeof(cmd),
                "./get_binary_section.csh %s_%02d %s B%02d %d %d %d %d",
                binary_output, band_index, stitching_output, band_index,
                start_sample, sample_count, start_line, line_count);
        if (status < 0 || status >= sizeof(cmd))
        {
            assert_true_with_message(false,
                    "String length exceeds buffer size");
            return;
        }
        if (system(cmd))
        {
            assert_true_with_message(false, "Error cropping out image section");
            return;
        }
        else
        {
            assert_true(true);
        }

        /* Compare the binary files with the expected results. */
        status = snprintf(cmd, sizeof(cmd), "./compare_binary_files.csh "
                "%s/expected_results/bin_detector_on_no_overlap_%02d "
                "./create_image/output/bin_detector_on_no_overlap_%02d",
                test_directory, band_index, band_index);
        if (status < 0 || status >= sizeof(cmd))
        {
            assert_true_with_message(false,
                    "String length exceeds buffer size");
            return;
        }
        if (system(cmd))
        {
            assert_true_with_message(false, "Differences occur between "
                    "detector corrected no overlap stitch type");
            return;
        }
        else
        {
            assert_true(true);
        }
    }

    /* Clean up overlap spacing test files */
    status = snprintf(cmd, sizeof(cmd),
            "rm -rf ./create_image/output/*detector_on_no_overlap*");
    if (status < 0 || status >= sizeof(cmd))
    {
        assert_true_with_message(false, "String length exceeds buffer size");
        return;
    }
    if (system(cmd))
    {
        assert_true_with_message(false, "Error cleaning up files");
        return;
    }
    else
    {
        assert_true(true);
    }

    return;
}  /* END detector_on_no_overlap_test */

/***************************************************************************
NAME:  interval_test

PURPOSE: This test is for the multiple-scene interval case.  It also tests
that if the parameter file scene center latitude is not specified, the scene
center is derived from the L0R.
****************************************************************************/
Ensure interval_test()
{
    char parameter_file[] = "interval";
    char output_file[] = "interval.h5";
    char command[3 * PATH_MAX];
    STITCH_METHOD_TYPE method = SCA_STITCH_NO_OVERLAP;
    double scene_center_latitude = 71.0;
    int scene_center_latitude_flag = 0;
    int band_list[] = {10,11};
    int nbands = sizeof(band_list) / sizeof(int);
    int apply_det_offset_flag = 0;
    char l0r_filename[] = "LT80150080112013066LGN00/LT80150080112013066LGN00";
        /* This is a multiscene interval, so if the L0R file is used instead
           of the parameter file, scene center latitude must be derived from
           interval data */
    char l1r_filename[] = "l1r_stitch_test_input_image_multiscene.h5"; /* This
        is an Earth image so it will try to get the scene center latitude from
        the parameter file and then the L0R file */
    char sensor[] = "L8_TIRS";
    char cpf_filename[] = "partial_cpf_tirs_interval";
    int status;

    /* Set up the input parameter file */
    if (create_odl_omf_file(parameter_file, output_file, method,
                apply_det_offset_flag, scene_center_latitude, 
                scene_center_latitude_flag, band_list, nbands, l0r_filename, 
                l1r_filename, sensor, cpf_filename))
    {
        assert_true_with_message(false, "Error creating parameters");
        return;
    }

    /* Call the test */
    if (test_the_application(parameter_file))
    {
        assert_true_with_message(false, "Error from test_the_application");
        return;
    }
    else
    {
        assert_true(true);
    }

    /* Check the return */
    status = snprintf(command, sizeof(command),
            "./compare_results.csh no %s %d %d %d", output_file, band_list[0],
            96, 120);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(false, "String length exceeds buffer size");
        return;
    }
    if (system(command))
    {
        assert_true_with_message(false, "Error in multiscene interval case "
                "B%02d", band_list[0]);
        return;
    }
    else
    {
        assert_true(true);
    }

    /* Clean up */
    status = snprintf(command, sizeof(command), "rm -rf %s.odl %s.omf %s",
            parameter_file, parameter_file, output_file);
    if (status < 0 || status >= sizeof(command))
    {
        assert_true_with_message(false, "String length exceeds buffer size");
        return;
    }
    if (system(command))
    {
        assert_true_with_message(false, "Error cleaning up files");
        return;
    }
    else
    {
        assert_true(true);
    }
}  /* END interval_test */

/***************************************************************************
Name:  main
***************************************************************************/
int main(int argc, char *argv[])
{
    TestSuite *suite = create_test_suite();
    char *test_data_dir;
    int status;

    test_data_dir = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (!test_data_dir)
    {
        printf("Could not read environment variable IAS_UNIT_TEST_DATA_DIR\n");
        return EXIT_FAILURE;
    }
    status = snprintf(test_directory, sizeof(test_directory),
            "%s/ias_base/rps/l1r_sca_stitching", test_data_dir);
    if (status < 0 || status >= sizeof(test_directory))
    {
        assert_true_with_message(false, "String length exceeds buffer size");
        return EXIT_FAILURE;
    }

    add_test(suite, no_overlap_test);
    add_test(suite, left_overlap_test);
    add_test(suite, right_overlap_test);
    add_test(suite, half_overlap_test);
    add_test(suite, overlap_spacing_test);
    add_test(suite, detector_offset_off_test);
    add_test(suite, detector_on_half_test);
    add_test(suite, detector_on_left_test);
    add_test(suite, detector_on_right_test);
    add_test(suite, detector_on_no_overlap_test);
    add_test(suite, interval_test);

    /* Allow a command line argument to run a single test. The argument
       is the name of the test. */
    if (argc == 2)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }

    return run_test_suite(suite, create_text_reporter());
}
