#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "angular_rate_valid_fix.h"
#include "ias_logging.h"
#include "ias_satellite_attributes.h"

int main( int argc, char *argv[] )
{
    int error_status = SUCCESS;
    L0RIO *l0r_data_input = NULL;

    if (argc != 2)
    {
        IAS_LOG_ERROR( "Usage: %s /path/to/Ancillary_File_Base ", argv[0]);
        return ERROR;
    }

    /* Initialize the logging library */
    ias_log_initialize( "angular_rate_valid_fix" );
    ias_log_set_output_level(IAS_LOG_LEVEL_DEBUG);
    ias_sat_attr_initialize(IAS_L8);

    /* Open the Input dataset */
    l0r_data_input = ias_l0r_open_from_filename(argv[1],
        IAS_L0R_COMPRESSION_ON);
    if(l0r_data_input == NULL)
    {
        IAS_LOG_ERROR( "Unable to open dataset: %s",
            argv[1]);
        return ERROR;
    }
    error_status = ias_l0r_open_ancillary(l0r_data_input, IAS_WRITE);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to open Ancillary file");
        return ERROR;
    }

    error_status = fix_angular_rate(l0r_data_input);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to generate dataset");
        return ERROR;
    }

    /*Close files */
    error_status = ias_l0r_close_ancillary(l0r_data_input);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to close Ancillary file");
        return ERROR;
    }
    error_status = ias_l0r_close(l0r_data_input);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to close L0R file");
        return ERROR;
    }

    return 0;
}  /* END main */

int fix_angular_rate(L0RIO *ptr)
{
    IAS_L0R_IMU *imu_data = NULL;
    int fix_count = 0;
    int error_status;
    int imu_count;
    int i;
    int j;

    error_status = ias_l0r_get_ancillary_imu_records_count(ptr, &imu_count);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to get imu count");
        return ERROR;
    }

    imu_data = calloc(imu_count, sizeof(IAS_L0R_IMU));
    if ( imu_data == NULL )
    {
        IAS_LOG_ERROR( "Unable to allocate memory for imu Data");
        return ERROR;
    }

    error_status = ias_l0r_get_ancillary_imu(ptr, 0, imu_count, imu_data);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to get imu data");
        free(imu_data);
        return ERROR;
    }

    /* ---- Perform the Fix ----- */
    for(i=0; i < imu_count; i++)
    {
        for(j=0; j < IAS_L0R_NUM_SAMPLES_PER_IMU_RECORD; j++)
        {
            if ( imu_data[i].gyro_samples[j].angular_rate_valid == 0x0F)
            {
                imu_data[i].gyro_samples[j].angular_rate_valid = 0xF0;
                fix_count++;
            }
        }
    }

    IAS_LOG_INFO("Performed the fix (0x0F to 0xF0) %i times", fix_count);

    /* Now write the data back out, overtop the old */
    error_status = ias_l0r_set_ancillary_imu(ptr, 0, imu_count, imu_data);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to get imu data");
        free(imu_data);
        return ERROR;
    }

    free(imu_data);

    return SUCCESS;
}
