#include <hdf5.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "generate_corrupted_ancillary.h"
#include "ias_logging.h"
#include "ias_satellite_attributes.h"

const int SEC_IN_DAY = 86400;

int main( int argc, char *argv[] )
{
    struct ANCILLARY_POINTERS anc_pointers;
    int error_status = SUCCESS;
    int oli_frame_count;
    int tirs_frame_count;
    L0RIO *l0r_data_input = NULL;

    if (argc != 2)
    {
        IAS_LOG_ERROR( "Usage: %s /path/to/Ancillary_File_Base ", argv[0]);
        return ERROR;
    }
    anc_pointers.data_location = argv[1];


   /* Initialize the logging library */
    ias_log_initialize( "generate_corrupted_ancillary" );
    ias_log_set_output_level(IAS_LOG_LEVEL_DEBUG);
    ias_sat_attr_initialize(IAS_L8);

    /* Open the Input dataset */
    l0r_data_input = ias_l0r_open_from_filename(anc_pointers.data_location,
        IAS_L0R_COMPRESSION_ON);
    if(l0r_data_input == NULL)
    {
        IAS_LOG_ERROR( "Unable to open dataset: %s",
            anc_pointers.data_location);
        return ERROR;
    }
    error_status = ias_l0r_open_ancillary(l0r_data_input, IAS_READ);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to open Ancillary file");
        return ERROR;
    }

    error_status = ias_l0r_open_header(l0r_data_input, IAS_READ);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to open Ancillary file");
        return ERROR;
    }

    error_status = ias_l0r_open_metadata(l0r_data_input, IAS_READ);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to open Ancillary file");
        return ERROR;
    }

    error_status = ias_l0r_get_ancillary_attitude_records_count(l0r_data_input,
        &anc_pointers.attitude_count);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to get Attitude count");
        return ERROR;
    }

    anc_pointers.attitude = calloc(anc_pointers.attitude_count,
        sizeof(IAS_L0R_ATTITUDE));
    if ( anc_pointers.attitude == NULL )
    {
        IAS_LOG_ERROR( "Unable to allocate memory for Attitude Data");
        return ERROR;
    }

    error_status = ias_l0r_get_ancillary_ephemeris_records_count(l0r_data_input,
        &anc_pointers.ephemeris_count);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to get Ephemeris count");
        return ERROR;
    }

    anc_pointers.ephemeris = calloc(anc_pointers.ephemeris_count,
        sizeof(IAS_L0R_EPHEMERIS));
    if ( anc_pointers.ephemeris == NULL )
    {
        IAS_LOG_ERROR( "Unable to allocate memory for ephemeris Data");
        return ERROR;
    }

    error_status = ias_l0r_get_ancillary_imu_records_count(l0r_data_input,
        &anc_pointers.imu_count);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to get imu count");
        return ERROR;
    }

    anc_pointers.imu = calloc(anc_pointers.imu_count,
        sizeof(IAS_L0R_IMU));
    if ( anc_pointers.imu == NULL )
    {
        IAS_LOG_ERROR( "Unable to allocate memory for imu Data");
        return ERROR;
    }

    error_status = ias_l0r_get_oli_frame_headers_records_count(l0r_data_input,
        &oli_frame_count);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to get OLI Frame header count");
        return ERROR;
    }

    anc_pointers.oli_frame_headers = calloc(oli_frame_count,
        sizeof(IAS_L0R_OLI_FRAME_HEADER));
    if(anc_pointers.oli_frame_headers == NULL)
    {
        IAS_LOG_ERROR("Unable allocate memory for OLI Frame Headers");
        return ERROR;
    }

    error_status = ias_l0r_get_tirs_frame_headers_records_count(l0r_data_input,
        &tirs_frame_count);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to get TIRS Frame header count");
        return ERROR;
    }

    anc_pointers.tirs_frame_headers = calloc(tirs_frame_count,
        sizeof(IAS_L0R_TIRS_FRAME_HEADER));
    if(anc_pointers.tirs_frame_headers == NULL)
    {
        IAS_LOG_ERROR("Unable allocate memory for TIRS Frame Headers");
        return ERROR;
    }

    /* Load each of the datasets into the allocated memory */
    error_status = ias_l0r_get_ancillary_attitude( l0r_data_input,
        0, anc_pointers.attitude_count, anc_pointers.attitude);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to get attitude records");
        return ERROR;
    }

    error_status = ias_l0r_get_ancillary_ephemeris( l0r_data_input,
        0, anc_pointers.ephemeris_count, anc_pointers.ephemeris);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to get ephemeris records");
        return ERROR;
    }

    error_status = ias_l0r_get_ancillary_imu( l0r_data_input,
        0, anc_pointers.imu_count, anc_pointers.imu);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to get imu records");
        return ERROR;
    }

    error_status = ias_l0r_get_oli_frame_headers( l0r_data_input,
        0, oli_frame_count, anc_pointers.oli_frame_headers);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to get OLI frame header records");
        return ERROR;
    }

    error_status = ias_l0r_get_tirs_frame_headers( l0r_data_input,
        0, tirs_frame_count, anc_pointers.tirs_frame_headers);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to get TIRS frame header records");
        return ERROR;
    }

    error_status = ias_l0r_get_ancillary_tirs_telemetry_records_count(
        l0r_data_input, &anc_pointers.tirs_telemetry_count);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to get TIRS telemetry count");
        return ERROR;
    }

    anc_pointers.tirs_telemetry = calloc(anc_pointers.tirs_telemetry_count,
        sizeof(IAS_L0R_TIRS_TELEMETRY));
    if(anc_pointers.tirs_telemetry == NULL)
    {
        IAS_LOG_ERROR("Unable allocate memory for TIRS Telemetry");
        return ERROR;
    }

    error_status = ias_l0r_get_ancillary_tirs_telemetry( l0r_data_input,
        0, anc_pointers.tirs_telemetry_count, anc_pointers.tirs_telemetry);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to get TIRS Telemetry records");
        return ERROR;
    }

    error_status = ias_l0r_get_oli_image_header( l0r_data_input,
        &anc_pointers.oli_image_header);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to get OLI image header records");
        return ERROR;
    }

    error_status = ias_l0r_get_interval_metadata( l0r_data_input,
        &anc_pointers.interval_metadata);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to get Interval Metadata");
        return ERROR;
    }

    /* Record the image start and stop times */
    anc_pointers.start_image_time = 
        (anc_pointers.oli_frame_headers[0].l0r_time.days_from_J2000 
        * SEC_IN_DAY) +
        anc_pointers.oli_frame_headers[0].l0r_time.seconds_of_day;

    anc_pointers.stop_image_time = 
        (anc_pointers.oli_frame_headers[oli_frame_count-1].
            l0r_time.days_from_J2000
        * SEC_IN_DAY) +
         anc_pointers.oli_frame_headers[oli_frame_count-1].
            l0r_time.seconds_of_day;

    /* Generate each of the datasets */
    error_status = remove_attitude_back(anc_pointers);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to generate dataset");
        return ERROR;
    }
    error_status = remove_attitude_front(anc_pointers);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to generate dataset");
        return ERROR;
    }
    error_status = remove_ephemeris_back(anc_pointers);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to generate dataset");
        return ERROR;
    }
    error_status = remove_ephemeris_front(anc_pointers);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to generate dataset");
        return ERROR;
    }
    error_status = remove_imu_back(anc_pointers);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to generate dataset");
        return ERROR;
    }
    error_status = remove_imu_front(anc_pointers);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to generate dataset");
        return ERROR;
    }
    error_status = set_imu_zeros(anc_pointers);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to generate dataset");
        return ERROR;
    }
    error_status = set_imu_invalid_flags(anc_pointers);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to generate dataset");
        return ERROR;
    }
    error_status = set_imu_invalid_sync_times(anc_pointers);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to generate dataset");
        return ERROR;
    }
    error_status = set_zeroed_integration_times(anc_pointers);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to generate dataset");
        return ERROR;
    }
    error_status = set_zeroed_image_header(anc_pointers);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to generate dataset");
        return ERROR;
    }
    error_status = set_non_matching_tirs_det_id(anc_pointers);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to generate dataset");
        return ERROR;
    }
    error_status = set_miscellaneous_corruptions(anc_pointers);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to generate dataset");
        return ERROR;
    }

    /*Close files and free memory */
    error_status = ias_l0r_close_ancillary(l0r_data_input);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to close Ancillary file");
        return ERROR;
    }
    error_status = ias_l0r_close_header(l0r_data_input);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to close Header file");
        return ERROR;
    }
    error_status = ias_l0r_close(l0r_data_input);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to close L0R file");
        return ERROR;
    }

    free(anc_pointers.attitude);
    free(anc_pointers.ephemeris);
    free(anc_pointers.imu);
    free(anc_pointers.oli_frame_headers);
    free(anc_pointers.tirs_frame_headers);
    return 0;
}  /* END main */

int remove_attitude_back(struct ANCILLARY_POINTERS pointers)
{
    int cut_off_point = 1;
    int i = 0;
    int error_status = 0;
    L0RIO *l0r_data_output = NULL;
    char output_loc[1024];
    char full_anc_path[1024];
    char sys_command[2048];
    
    for (i=pointers.attitude_count-1; i>=0; i--)
    {
        cut_off_point = i;
        if(((pointers.attitude[i].l0r_time.days_from_J2000 * SEC_IN_DAY) +
            pointers.attitude[i].l0r_time.seconds_of_day) < 
            pointers.stop_image_time)
        {
            break;
        }
    }
    /* Move a few back from the cutoff */
    cut_off_point -= 15;
    if (cut_off_point < 0)
        cut_off_point = 0;
    IAS_LOG_DEBUG("Calculated cut off point is %i", cut_off_point);

    snprintf(output_loc, 1024, "%s_05", pointers.data_location);
    snprintf(full_anc_path, 1024, "%s_ANC.h5", output_loc);
    snprintf(sys_command, 2048, "cp -f %s_ANC.h5 %s",
        pointers.data_location,
        full_anc_path);
    error_status = system(sys_command);
    if (error_status != 0)
    {
        IAS_LOG_ERROR("Copying of Ancillary file failed");
        return ERROR;
    }

    /* First we need to "notch out" the dataset we're changing */
    notch_dataset(full_anc_path, "Spacecraft/ACS/Attitude");

    /* Open the Output dataset */
    IAS_LOG_DEBUG("Opening %s", output_loc);
    l0r_data_output = ias_l0r_open_from_filename(output_loc,
        IAS_L0R_COMPRESSION_ON);
    if(l0r_data_output == NULL)
    {
        IAS_LOG_ERROR( "Unable to open dataset: %s",
            output_loc);
        return ERROR;
    }
    error_status = ias_l0r_open_ancillary(l0r_data_output, IAS_WRITE);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to open Ancillary file");
        return ERROR;
    }

    /* Add the trimmed dataset */
    error_status = ias_l0r_set_ancillary_attitude(l0r_data_output,
        0, cut_off_point, pointers.attitude);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to open write dataset");
        return ERROR;
    } 

    error_status = ias_l0r_close_ancillary(l0r_data_output);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to close Ancillary file");
        return ERROR;
    }
    error_status = ias_l0r_close(l0r_data_output);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to close dataset");
        return ERROR;
    }

    return SUCCESS;
}

int set_miscellaneous_corruptions(struct ANCILLARY_POINTERS pointers)
{
    int error_status = 0;
    L0RIO *l0r_data_output = NULL;
    char output_loc[1024];
    char full_anc_path[1024];
    char sys_command[2048];
    int record_index;

    snprintf(output_loc, 1024, "%s_02", pointers.data_location);
    snprintf(full_anc_path, 1024, "%s_ANC.h5", output_loc);
    snprintf(sys_command, 2048, "cp -f %s_ANC.h5 %s",
        pointers.data_location,
        full_anc_path);
    error_status = system(sys_command);
    if (error_status != 0)
    {
        IAS_LOG_ERROR("Copying of Ancillary file failed");
        return ERROR;
    }

    IAS_LOG_DEBUG("Opening %s", output_loc);

    /* Open the Output dataset */
    l0r_data_output = ias_l0r_open_from_filename(output_loc,
        IAS_L0R_COMPRESSION_ON);
    if(l0r_data_output == NULL)
    {
        IAS_LOG_ERROR( "Unable to open dataset: %s",
            output_loc);
        return ERROR;
    }
    error_status = ias_l0r_open_header(l0r_data_output, IAS_WRITE);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to open Ancillary file");
        return ERROR;
    }

    /* Create corruption in the OLI frame header structure */
    IAS_LOG_DEBUG("oli_frame_headers[1].milliseconds = 22182872");
    pointers.oli_frame_headers[1].milliseconds = 22182872;
    IAS_LOG_DEBUG("oli_frame_headers[9].milliseconds = 22182906");
    pointers.oli_frame_headers[9].milliseconds = 22182906;
    IAS_LOG_DEBUG("oli_frame_headers[17].milliseconds = 22182940");
    pointers.oli_frame_headers[17].milliseconds = 22182940;
    IAS_LOG_DEBUG("oli_frame_headers[25].milliseconds = 22182973");
    pointers.oli_frame_headers[25].milliseconds = 22182973;
    IAS_LOG_DEBUG("oli_frame_headers[33].milliseconds = 22183007");
    pointers.oli_frame_headers[33].milliseconds = 22183007;
    IAS_LOG_DEBUG("oli_frame_headers[41].milliseconds = 22183041");
    pointers.oli_frame_headers[41].milliseconds = 22183041;
    IAS_LOG_DEBUG("oli_frame_headers[49].milliseconds = 22183075");
    pointers.oli_frame_headers[49].milliseconds = 22183075;
    IAS_LOG_DEBUG("oli_frame_headers[57].milliseconds = 22183109");
    pointers.oli_frame_headers[57].milliseconds = 22183109;
    IAS_LOG_DEBUG("oli_frame_headers[65].milliseconds = 22183143");
    pointers.oli_frame_headers[65].milliseconds = 22183143;
    IAS_LOG_DEBUG("oli_frame_headers[73].milliseconds = 22183177");
    pointers.oli_frame_headers[73].milliseconds = 22183177;
    IAS_LOG_DEBUG("oli_frame_headers[2332].microseconds = 388");
    pointers.oli_frame_headers[2332].microseconds = 388;
    IAS_LOG_DEBUG("oli_frame_headers[2333].milliseconds = 22192702");
    pointers.oli_frame_headers[2333].milliseconds = 22192702;

    error_status = ias_l0r_set_oli_frame_headers(l0r_data_output,
        0, 2334, pointers.oli_frame_headers);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to write OLI frame header dataset");
        return ERROR;
    } 

    /* Create corruption in the TIRS frame header structure */
    IAS_LOG_DEBUG("tirs_frame_headers[1].milliseconds = 22178981");
    pointers.tirs_frame_headers[1].milliseconds = 22178981;
    IAS_LOG_DEBUG("tirs_frame_headers[1].integration_duration = 3.026");
    pointers.tirs_frame_headers[1].integration_duration = 3.026;
    IAS_LOG_DEBUG("tirs_frame_headers[2].integration_duration = 3.027");
    pointers.tirs_frame_headers[2].integration_duration = 3.027;
    IAS_LOG_DEBUG("tirs_frame_headers[3].integration_duration = 3.028");
    pointers.tirs_frame_headers[3].integration_duration = 3.028;
    IAS_LOG_DEBUG("tirs_frame_headers[5].integration_duration = 4.0");
    pointers.tirs_frame_headers[5].integration_duration = 4.0;
    IAS_LOG_DEBUG("tirs_frame_headers[8].integration_duration = 4.0");
    pointers.tirs_frame_headers[8].integration_duration = 4.0;
    IAS_LOG_DEBUG("tirs_frame_headers[9].milliseconds = 22179095");
    pointers.tirs_frame_headers[9].milliseconds = 22179095;
    IAS_LOG_DEBUG("tirs_frame_headers[17].milliseconds = 22179209");
    pointers.tirs_frame_headers[17].milliseconds = 22179209;
    IAS_LOG_DEBUG("tirs_frame_headers[25].milliseconds = 22179324");
    pointers.tirs_frame_headers[25].milliseconds = 22179324;
    IAS_LOG_DEBUG("tirs_frame_headers[33].milliseconds = 22179438");
    pointers.tirs_frame_headers[33].milliseconds = 22179438;
    IAS_LOG_DEBUG("tirs_frame_headers[41].milliseconds = 22179552");
    pointers.tirs_frame_headers[41].milliseconds = 22179552;
    IAS_LOG_DEBUG("tirs_frame_headers[49].milliseconds = 22179666");
    pointers.tirs_frame_headers[49].milliseconds = 22179666;
    IAS_LOG_DEBUG("tirs_frame_headers[57].milliseconds = 22179781");
    pointers.tirs_frame_headers[57].milliseconds = 22179781;
    IAS_LOG_DEBUG("tirs_frame_headers[65].milliseconds = 22179895");
    pointers.tirs_frame_headers[65].milliseconds = 22179895;
    IAS_LOG_DEBUG("tirs_frame_headers[73].milliseconds = 22180009");
    pointers.tirs_frame_headers[73].milliseconds = 22180009;
    IAS_LOG_DEBUG("tirs_frame_headers[1310].microseconds = 544");
    pointers.tirs_frame_headers[1310].microseconds = 544;
    IAS_LOG_DEBUG("tirs_frame_headers[1311].milliseconds = 22197647");
    pointers.tirs_frame_headers[1311].milliseconds = 22197647;
    IAS_LOG_DEBUG("tirs_frame_headers[1312].day = 579");
    pointers.tirs_frame_headers[1312].day = 579;
    IAS_LOG_DEBUG("tirs_frame_headers[1313].microseconds = 402");
    pointers.tirs_frame_headers[1313].microseconds = 402;
    IAS_LOG_DEBUG("tirs_frame_headers[1314].milliseconds = 22197690");
    pointers.tirs_frame_headers[1314].milliseconds = 22197690;

    error_status = ias_l0r_set_tirs_frame_headers(l0r_data_output,
        0, 1315, pointers.tirs_frame_headers);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to write TIRS frame header dataset");
        return ERROR;
    } 

    /* Create corruption in the OLI image header structure */
    IAS_LOG_DEBUG("oli_image_header.ms_integration_time = 3615");
    pointers.oli_image_header.ms_integration_time = 3615;
    IAS_LOG_DEBUG("oli_image_header.pan_integration_time = 675");
    pointers.oli_image_header.pan_integration_time = 675;

    error_status = ias_l0r_set_oli_image_header(l0r_data_output,
        &pointers.oli_image_header);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to write OLI image header dataset");
        return ERROR;
    }

    error_status = ias_l0r_close_header(l0r_data_output);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to close Ancillary file header");
        return ERROR;
    }

    error_status = ias_l0r_open_ancillary(l0r_data_output, IAS_WRITE);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to open Ancillary file");
        return ERROR;
    }

    /* Create corruption in the spacecraft attitude structure */
    record_index = 10;
    IAS_LOG_DEBUG("attitude[%d].inertial_to_body.vector.x = 0",
        record_index);
    pointers.attitude[record_index].inertial_to_body.vector.x = 0;
    IAS_LOG_DEBUG("attitude[%d].inertial_to_body.vector.y = 0",
        record_index);
    pointers.attitude[record_index].inertial_to_body.vector.y = 0;
    IAS_LOG_DEBUG("attitude[%d].inertial_to_body.vector.z = 0",
        record_index);
    pointers.attitude[record_index].inertial_to_body.vector.z = 0;
    IAS_LOG_DEBUG("attitude[%d].inertial_to_body.scalar = 0",
        record_index);
    pointers.attitude[record_index].inertial_to_body.scalar = 0;

    error_status = ias_l0r_set_ancillary_attitude(l0r_data_output, 0, 11,
        pointers.attitude);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to write spacecraft attitude dataset");
        return ERROR;
    }

    /* Create corruption in the spacecraft ephemeris structure */
    IAS_LOG_DEBUG("ephemeris[0].time_tag_sec_orig -= 3");
    pointers.ephemeris[0].time_tag_sec_orig -= 3;
    IAS_LOG_DEBUG("ephemeris[1].time_tag_sec_orig += 5");
    pointers.ephemeris[1].time_tag_sec_orig += 5;
    IAS_LOG_DEBUG("ephemeris[2].time_tag_sec_orig += 9");
    pointers.ephemeris[2].time_tag_sec_orig += 9;
    IAS_LOG_DEBUG("ephemeris[3].time_tag_sec_orig -= 6");
    pointers.ephemeris[3].time_tag_sec_orig -= 6;
    IAS_LOG_DEBUG("ephemeris[4].time_tag_sec_orig += 4");
    pointers.ephemeris[4].time_tag_sec_orig += 4;
    IAS_LOG_DEBUG("ephemeris[5].time_tag_sec_orig -= 8");
    pointers.ephemeris[5].time_tag_sec_orig -= 8;
    IAS_LOG_DEBUG("ephemeris[6].time_tag_sec_orig -= 3");
    pointers.ephemeris[6].time_tag_sec_orig -= 3;
    IAS_LOG_DEBUG("ephemeris[7].time_tag_sec_orig += 5");
    pointers.ephemeris[7].time_tag_sec_orig += 5;
    IAS_LOG_DEBUG("ephemeris[8].time_tag_sec_orig += 9");
    pointers.ephemeris[8].time_tag_sec_orig += 9;
    IAS_LOG_DEBUG("ephemeris[9].time_tag_sec_orig -= 6");
    pointers.ephemeris[9].time_tag_sec_orig -= 6;

    IAS_LOG_DEBUG("ephemeris[2].ecef_position_meters.z += 50000");
    pointers.ephemeris[2].ecef_position_meters.z += 50000;
    IAS_LOG_DEBUG("ephemeris[3].ecef_velocity_meters_per_sec.x += 500");
    pointers.ephemeris[3].ecef_velocity_meters_per_sec.x += 500;
    IAS_LOG_DEBUG("ephemeris[4].ecef_velocity_meters_per_sec.y += 5000");
    pointers.ephemeris[4].ecef_velocity_meters_per_sec.y += 5000;

    error_status = ias_l0r_set_ancillary_ephemeris(l0r_data_output,
        0, 10, pointers.ephemeris);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to write spacecraft ephemeris dataset");
        return ERROR;
    } 

    /* Create corruption in the TIRS Telemetry structure */
    record_index = 10;
    IAS_LOG_DEBUG("tirs_telemetry[%d].command.ssm_position_sel ="
        " IAS_L0R_TIRS_SSM_SAFE_POSITION", record_index);
    pointers.tirs_telemetry[record_index].command.ssm_position_sel =
        IAS_L0R_TIRS_SSM_SAFE_POSITION;

    error_status = ias_l0r_set_ancillary_tirs_telemetry(l0r_data_output,
        0, pointers.tirs_telemetry_count, pointers.tirs_telemetry);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to write TIRS telemetry dataset");
        return ERROR;
    } 

    error_status = ias_l0r_close_ancillary(l0r_data_output);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to close Ancillary file");
        return ERROR;
    }

    error_status = ias_l0r_close(l0r_data_output);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to close dataset");
        return ERROR;
    }

    return SUCCESS;
}

int remove_attitude_front(struct ANCILLARY_POINTERS pointers)
{
    int cut_off_point = 1;
    int i = 0;
    int error_status = 0;
    L0RIO *l0r_data_output = NULL;
    char output_loc[1024];
    char full_anc_path[1024];
    char sys_command[2048];
    
    for (i=0; i<pointers.attitude_count; i++)
    {
        cut_off_point = i;
        if(((pointers.attitude[i].l0r_time.days_from_J2000 * SEC_IN_DAY) +
            pointers.attitude[i].l0r_time.seconds_of_day) > 
            pointers.start_image_time)
        {
            break;
        }
    }
    /* Move a few forward from the cutoff */
    cut_off_point += 15;
    if (cut_off_point > pointers.attitude_count)
        cut_off_point = pointers.attitude_count;
    IAS_LOG_DEBUG("Calculated cut off point is %i", cut_off_point);

    snprintf(output_loc, 1024, "%s_06", pointers.data_location);
    snprintf(full_anc_path, 1024, "%s_ANC.h5", output_loc);
    snprintf(sys_command, 2048, "cp -f %s_ANC.h5 %s",
        pointers.data_location,
        full_anc_path);
    error_status = system(sys_command);
    if (error_status != 0)
    {
        IAS_LOG_ERROR("Copying of Ancillary file failed");
        return ERROR;
    }

    /* First we need to "notch out" the dataset we're changing */
    notch_dataset(full_anc_path, "Spacecraft/ACS/Attitude");

    /* Open the Output dataset */
    IAS_LOG_DEBUG("Opening %s", output_loc);
    l0r_data_output = ias_l0r_open_from_filename(output_loc,
        IAS_L0R_COMPRESSION_ON);
    if(l0r_data_output == NULL)
    {
        IAS_LOG_ERROR( "Unable to open dataset: %s",
            output_loc);
        return ERROR;
    }
    error_status = ias_l0r_open_ancillary(l0r_data_output, IAS_WRITE);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to open Ancillary file");
        return ERROR;
    }

    /* Add the trimed dataset */
    error_status = ias_l0r_set_ancillary_attitude(l0r_data_output,
        0, pointers.attitude_count-cut_off_point+1,
        &pointers.attitude[cut_off_point-1]);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to open write dataset");
        return ERROR;
    } 

    error_status = ias_l0r_close_ancillary(l0r_data_output);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to close Ancillary file");
        return ERROR;
    }
    error_status = ias_l0r_close(l0r_data_output);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to close dataset");
        return ERROR;
    }

    return SUCCESS;
}


int remove_ephemeris_back(struct ANCILLARY_POINTERS pointers)
{
    int cut_off_point = 1;
    int i = 0;
    int error_status = 0;
    L0RIO *l0r_data_output = NULL;
    char output_loc[1024];
    char full_anc_path[1024];
    char sys_command[2048];
    
    for (i=pointers.ephemeris_count-1; i>=0; i--)
    {
        cut_off_point = i;
        if(((pointers.ephemeris[i].l0r_time.days_from_J2000 * SEC_IN_DAY) +
            pointers.ephemeris[i].l0r_time.seconds_of_day) < 
            pointers.stop_image_time)
        {
            break;
        }
    }
    /* Move a few back from the cutoff */
    cut_off_point -= 15;
    if (cut_off_point < 0)
        cut_off_point = 0;
    IAS_LOG_DEBUG("Calculated cut off point is %i", cut_off_point);

    snprintf(output_loc, 1024, "%s_03", pointers.data_location);
    snprintf(full_anc_path, 1024, "%s_ANC.h5", output_loc);
    snprintf(sys_command, 2048, "cp -f %s_ANC.h5 %s",
        pointers.data_location,
        full_anc_path);
    error_status = system(sys_command);
    if (error_status != 0)
    {
        IAS_LOG_ERROR("Copying of Ancillary file failed");
        return ERROR;
    }

    /* First we need to "notch out" the dataset we're changing */
    notch_dataset(full_anc_path, "Spacecraft/Ephemeris");

    /* Open the Output dataset */
    IAS_LOG_DEBUG("Opening %s", output_loc);
    l0r_data_output = ias_l0r_open_from_filename(output_loc,
        IAS_L0R_COMPRESSION_ON);
    if(l0r_data_output == NULL)
    {
        IAS_LOG_ERROR( "Unable to open dataset: %s",
            output_loc);
        return ERROR;
    }
    error_status = ias_l0r_open_ancillary(l0r_data_output, IAS_WRITE);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to open Ancillary file");
        return ERROR;
    }

    /* Add the trimed dataset */
    error_status = ias_l0r_set_ancillary_ephemeris(l0r_data_output,
        0, cut_off_point, pointers.ephemeris);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to open write dataset");
        return ERROR;
    } 

    error_status = ias_l0r_close_ancillary(l0r_data_output);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to close Ancillary file");
        return ERROR;
    }
    error_status = ias_l0r_close(l0r_data_output);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to close dataset");
        return ERROR;
    }

    return SUCCESS;
}

int remove_ephemeris_front(struct ANCILLARY_POINTERS pointers)
{
    int cut_off_point = 1;
    int i = 0;
    int error_status = 0;
    L0RIO *l0r_data_output = NULL;
    char output_loc[1024];
    char full_anc_path[1024];
    char sys_command[2048];
    
    for (i=0; i<pointers.ephemeris_count; i++)
    {
        cut_off_point = i;
        if(((pointers.ephemeris[i].l0r_time.days_from_J2000 * SEC_IN_DAY) +
            pointers.ephemeris[i].l0r_time.seconds_of_day) > 
            pointers.start_image_time)
        {
            break;
        }
    }
    /* Move a few forward from the cutoff */
    cut_off_point += 15;
    if (cut_off_point > pointers.ephemeris_count)
        cut_off_point = pointers.ephemeris_count;
    IAS_LOG_DEBUG("Calculated cut off point is %i", cut_off_point);

    snprintf(output_loc, 1024, "%s_04", pointers.data_location);
    snprintf(full_anc_path, 1024, "%s_ANC.h5", output_loc);
    snprintf(sys_command, 2048, "cp -f %s_ANC.h5 %s",
        pointers.data_location,
        full_anc_path);
    error_status = system(sys_command);
    if (error_status != 0)
    {
        IAS_LOG_ERROR("Copying of Ancillary file failed");
        return ERROR;
    }

    /* First we need to "notch out" the dataset we're changing */
    notch_dataset(full_anc_path, "Spacecraft/Ephemeris");

    /* Open the Output dataset */
    IAS_LOG_DEBUG("Opening %s", output_loc);
    l0r_data_output = ias_l0r_open_from_filename(output_loc,
        IAS_L0R_COMPRESSION_ON);
    if(l0r_data_output == NULL)
    {
        IAS_LOG_ERROR( "Unable to open dataset: %s",
            output_loc);
        return ERROR;
    }
    error_status = ias_l0r_open_ancillary(l0r_data_output, IAS_WRITE);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to open Ancillary file");
        return ERROR;
    }

    /* Add the trimed dataset */
    error_status = ias_l0r_set_ancillary_ephemeris(l0r_data_output,
        0, pointers.ephemeris_count-cut_off_point+1,
        &pointers.ephemeris[cut_off_point-1]);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to open write dataset");
        return ERROR;
    } 

    error_status = ias_l0r_close_ancillary(l0r_data_output);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to close Ancillary file");
        return ERROR;
    }
    error_status = ias_l0r_close(l0r_data_output);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to close dataset");
        return ERROR;
    }

    return SUCCESS;
}

int remove_imu_back(struct ANCILLARY_POINTERS pointers)
{
    int cut_off_point = 1;
    int i = 0;
    int error_status = 0;
    L0RIO *l0r_data_output = NULL;
    char output_loc[1024];
    char full_anc_path[1024];
    char sys_command[2048];
    
    for (i=pointers.imu_count-1; i>=0; i--)
    {
        cut_off_point = i;
        if(((pointers.imu[i].l0r_time.days_from_J2000 * SEC_IN_DAY) +
            pointers.imu[i].l0r_time.seconds_of_day) < 
            pointers.stop_image_time)
        {
            break;
        }
    }
    /* Move a few back from the cutoff */
    cut_off_point -= 15;
    if (cut_off_point < 0)
        cut_off_point = 0;
    IAS_LOG_DEBUG("Calculated cut off point is %i", cut_off_point);

    snprintf(output_loc, 1024, "%s_07", pointers.data_location);
    snprintf(full_anc_path, 1024, "%s_ANC.h5", output_loc);
    snprintf(sys_command, 2048, "cp -f %s_ANC.h5 %s",
        pointers.data_location,
        full_anc_path);
    error_status = system(sys_command);
    if (error_status != 0)
    {
        IAS_LOG_ERROR("Copying of Ancillary file failed");
        return ERROR;
    }

    /* First we need to "notch out" the dataset we're changing */
    notch_dataset(full_anc_path, "Spacecraft/IMU/Gyro");

    /* Open the Output dataset */
    IAS_LOG_DEBUG("Opening %s", output_loc);
    l0r_data_output = ias_l0r_open_from_filename(output_loc,
        IAS_L0R_COMPRESSION_ON);
    if(l0r_data_output == NULL)
    {
        IAS_LOG_ERROR( "Unable to open dataset: %s",
            output_loc);
        return ERROR;
    }
    error_status = ias_l0r_open_ancillary(l0r_data_output, IAS_WRITE);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to open Ancillary file");
        return ERROR;
    }

    /* Add the trimed dataset */
    error_status = ias_l0r_set_ancillary_imu(l0r_data_output,
        0, cut_off_point, pointers.imu);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to open write dataset");
        return ERROR;
    } 

    error_status = ias_l0r_close_ancillary(l0r_data_output);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to close Ancillary file");
        return ERROR;
    }
    error_status = ias_l0r_close(l0r_data_output);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to close dataset");
        return ERROR;
    }

    return SUCCESS;
}


int remove_imu_front(struct ANCILLARY_POINTERS pointers)
{
    int cut_off_point = 1;
    int i = 0;
    int error_status = 0;
    L0RIO *l0r_data_output = NULL;
    char output_loc[1024];
    char full_anc_path[1024];
    char sys_command[2048];
    
    for (i=0; i<pointers.imu_count; i++)
    {
        cut_off_point = i;
        if(((pointers.imu[i].l0r_time.days_from_J2000 * SEC_IN_DAY) +
            pointers.imu[i].l0r_time.seconds_of_day) > 
            pointers.start_image_time)
        {
            break;
        }
    }
    /* Move a few forward from the cutoff */
    cut_off_point += 15;
    if (cut_off_point > pointers.imu_count)
        cut_off_point = pointers.imu_count;
    IAS_LOG_DEBUG("Calculated cut off point is %i", cut_off_point);

    snprintf(output_loc, 1024, "%s_08", pointers.data_location);
    snprintf(full_anc_path, 1024, "%s_ANC.h5", output_loc);
    snprintf(sys_command, 2048, "cp -f %s_ANC.h5 %s",
        pointers.data_location,
        full_anc_path);
    error_status = system(sys_command);
    if (error_status != 0)
    {
        IAS_LOG_ERROR("Copying of Ancillary file failed");
        return ERROR;
    }

    /* First we need to "notch out" the dataset we're changing */
    notch_dataset(full_anc_path, "Spacecraft/IMU/Gyro");

    /* Open the Output dataset */
    IAS_LOG_DEBUG("Opening %s", output_loc);
    l0r_data_output = ias_l0r_open_from_filename(output_loc,
        IAS_L0R_COMPRESSION_ON);
    if(l0r_data_output == NULL)
    {
        IAS_LOG_ERROR( "Unable to open dataset: %s",
            output_loc);
        return ERROR;
    }
    error_status = ias_l0r_open_ancillary(l0r_data_output, IAS_WRITE);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to open Ancillary file");
        return ERROR;
    }

    /* Add the trimed dataset */
    error_status = ias_l0r_set_ancillary_imu(l0r_data_output,
        0, pointers.imu_count-cut_off_point+1, &pointers.imu[cut_off_point-1]);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to open write dataset");
        return ERROR;
    } 

    error_status = ias_l0r_close_ancillary(l0r_data_output);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to close Ancillary file");
        return ERROR;
    }
    error_status = ias_l0r_close(l0r_data_output);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to close dataset");
        return ERROR;
    }

    return SUCCESS;
}

/*
    - Modify the data to generate a warning message about consecutive invalid
      gryo samples
    - Also force to terminate by the quaternions not covering the start of the
      IMU data
*/
int set_imu_zeros(struct ANCILLARY_POINTERS pointers)
{
    int index;
    int gyro_index ;
    int error_status;
    int start_point;
    int cut_off_point;
    L0RIO *l0r_data_output = NULL;
    IAS_L0R_IMU *temp_IMU;
    IAS_L0R_ATTITUDE *temp_ATT;
    char output_loc[1024];
    char full_anc_path[1024];
    char sys_command[2048];

    /* Copy the data, so we can change it*/
    temp_IMU = calloc(pointers.imu_count, sizeof(IAS_L0R_IMU));
    memcpy(temp_IMU, pointers.imu, pointers.imu_count * sizeof(IAS_L0R_IMU));
    temp_ATT = calloc(pointers.attitude_count, sizeof(IAS_L0R_ATTITUDE));
    memcpy(temp_ATT, pointers.attitude,
        pointers.attitude_count * sizeof(IAS_L0R_ATTITUDE));

    start_point = 50;
    cut_off_point = pointers.attitude_count - start_point;

    /* Modify the data in memory */
    for (index = 0; index < pointers.imu_count; index++)
    {
        for (gyro_index = 0;
             gyro_index < IAS_L0R_NUM_SAMPLES_PER_IMU_RECORD;
             gyro_index++)
        {
            temp_IMU[index].gyro_samples[gyro_index].angular_rate_valid = 0x00;
        }
    }
    for (index = 0; index < cut_off_point; index++)
    {
        temp_ATT[index] = temp_ATT[index+start_point];
    }
    
    snprintf(output_loc, 1024, "%s_09", pointers.data_location);
    snprintf(full_anc_path, 1024, "%s_ANC.h5", output_loc);
    snprintf(sys_command, 2048, "cp -f %s_ANC.h5 %s",
        pointers.data_location,
        full_anc_path);
    error_status = system(sys_command);
    if (error_status != 0)
    {
        IAS_LOG_ERROR("Copying of Ancillary file failed");
        return ERROR;
    }

    /* First we need to "notch out" the datasets we're changing */
    notch_dataset(full_anc_path, "Spacecraft/IMU/Gyro");
    notch_dataset(full_anc_path, "Spacecraft/ACS/Attitude");

    /* Open the Output dataset */
    IAS_LOG_DEBUG("Opening %s", output_loc);
    l0r_data_output = ias_l0r_open_from_filename(output_loc,
        IAS_L0R_COMPRESSION_ON);
    if(l0r_data_output == NULL)
    {
        IAS_LOG_ERROR( "Unable to open dataset: %s",
            output_loc);
        return ERROR;
    }
    error_status = ias_l0r_open_ancillary(l0r_data_output, IAS_WRITE);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to open Ancillary file");
        return ERROR;
    }

    /* Add the trimed datasets */
    error_status = ias_l0r_set_ancillary_imu(l0r_data_output,
        0, pointers.imu_count, temp_IMU);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to open write dataset");
        return ERROR;
    }
    error_status = ias_l0r_set_ancillary_attitude(l0r_data_output,
        0, cut_off_point, temp_ATT);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to open write dataset");
        return ERROR;
    } 

    free(temp_IMU);
    free(temp_ATT);

    error_status = ias_l0r_close_ancillary(l0r_data_output);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to close Ancillary file");
        return ERROR;
    }
    error_status = ias_l0r_close(l0r_data_output);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to close dataset");
        return ERROR;
    }

    return SUCCESS;
}

int set_imu_invalid_flags(struct ANCILLARY_POINTERS pointers)
{
    int i = 0;
    int j = 0;
    int error_status = 0;
    L0RIO *l0r_data_output = NULL;
    IAS_L0R_IMU *temp_IMU;
    char output_loc[1024];
    char full_anc_path[1024];
    char sys_command[2048];

    /* Copy the data, so we can change it*/
    temp_IMU = calloc(pointers.imu_count, sizeof(IAS_L0R_IMU));
    memcpy(temp_IMU, pointers.imu, pointers.imu_count * sizeof(IAS_L0R_IMU));

    /* Modify the data in memory */
    for (i=0; i < pointers.imu_count; i++)
    {
        for(j=0; j < IAS_L0R_NUM_SAMPLES_PER_IMU_RECORD; j++)
        {
            temp_IMU[i].gyro_samples[j].saturation_and_scaling = 0xFF;
            temp_IMU[i].gyro_samples[j].angular_rate_valid = 0x00;
        }
    }
    
    snprintf(output_loc, 1024, "%s_10", pointers.data_location);
    snprintf(full_anc_path, 1024, "%s_ANC.h5", output_loc);
    snprintf(sys_command, 2048, "cp -f %s_ANC.h5 %s",
        pointers.data_location,
        full_anc_path);
    error_status = system(sys_command);
    if (error_status != 0)
    {
        IAS_LOG_ERROR("Copying of Ancillary file failed");
        return ERROR;
    }

    /* First we need to "notch out" the dataset we're changing */
    notch_dataset(full_anc_path, "Spacecraft/IMU/Gyro");

    /* Open the Output dataset */
    IAS_LOG_DEBUG("Opening %s", output_loc);
    l0r_data_output = ias_l0r_open_from_filename(output_loc,
        IAS_L0R_COMPRESSION_ON);
    if(l0r_data_output == NULL)
    {
        IAS_LOG_ERROR( "Unable to open dataset: %s",
            output_loc);
        return ERROR;
    }
    error_status = ias_l0r_open_ancillary(l0r_data_output, IAS_WRITE);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to open Ancillary file");
        return ERROR;
    }

    /* Add the trimed dataset */
    error_status = ias_l0r_set_ancillary_imu(l0r_data_output,
        0, pointers.imu_count, temp_IMU);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to open write dataset");
        return ERROR;
    } 

    free(temp_IMU);

    error_status = ias_l0r_close_ancillary(l0r_data_output);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to close Ancillary file");
        return ERROR;
    }
    error_status = ias_l0r_close(l0r_data_output);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to close dataset");
        return ERROR;
    }

    return SUCCESS;
}

int set_imu_invalid_sync_times(struct ANCILLARY_POINTERS pointers)
{
    int i = 0;
    int j = 0;
    int error_status = 0;
    L0RIO *l0r_data_output = NULL;
    IAS_L0R_IMU *temp_IMU;
    char output_loc[1024];
    char full_anc_path[1024];
    char sys_command[2048];

    /* Copy the data, so we can change it*/
    temp_IMU = calloc(pointers.imu_count, sizeof(IAS_L0R_IMU));
    memcpy(temp_IMU, pointers.imu, pointers.imu_count * sizeof(IAS_L0R_IMU));

    /* Modify the data in memory */
    for (i=0; i < pointers.imu_count; i++)
    {
        for(j=0; j < IAS_L0R_NUM_SAMPLES_PER_IMU_RECORD; j++)
        {
            temp_IMU[i].gyro_samples[j].sync_event_time_tag = -1 * 0x8000;
        }
    }
    
    snprintf(output_loc, 1024, "%s_11", pointers.data_location);
    snprintf(full_anc_path, 1024, "%s_ANC.h5", output_loc);
    snprintf(sys_command, 2048, "cp -f %s_ANC.h5 %s",
        pointers.data_location,
        full_anc_path);
    error_status = system(sys_command);
    if (error_status != 0)
    {
        IAS_LOG_ERROR("Copying of Ancillary file failed");
        return ERROR;
    }

    /* First we need to "notch out" the dataset we're changing */
    notch_dataset(full_anc_path, "Spacecraft/IMU/Gyro");

    /* Open the Output dataset */
    IAS_LOG_DEBUG("Opening %s", output_loc);
    l0r_data_output = ias_l0r_open_from_filename(output_loc,
        IAS_L0R_COMPRESSION_ON);
    if(l0r_data_output == NULL)
    {
        IAS_LOG_ERROR( "Unable to open dataset: %s",
            output_loc);
        return ERROR;
    }
    error_status = ias_l0r_open_ancillary(l0r_data_output, IAS_WRITE);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to open Ancillary file");
        return ERROR;
    }

    /* Add the trimed dataset */
    error_status = ias_l0r_set_ancillary_imu(l0r_data_output,
        0, pointers.imu_count, temp_IMU);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to open write dataset");
        return ERROR;
    } 

    free(temp_IMU);

    error_status = ias_l0r_close_ancillary(l0r_data_output);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to close Ancillary file");
        return ERROR;
    }
    error_status = ias_l0r_close(l0r_data_output);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to close dataset");
        return ERROR;
    }

    return SUCCESS;
}

int set_zeroed_integration_times(struct ANCILLARY_POINTERS pointers)
{
    int error_status = 0;
    L0RIO *l0r_data_output = NULL;
    IAS_L0R_OLI_IMAGE_HEADER temp_oli_image_header;
    char output_loc[1024];
    char full_anc_path[1024];
    char sys_command[2048];

    /* Copy the data, so we can change it*/
    memcpy(&temp_oli_image_header, &pointers.oli_image_header,
        sizeof(IAS_L0R_OLI_IMAGE_HEADER));

    /* Modify the data in memory */
    temp_oli_image_header.ms_integration_time = 0;
    temp_oli_image_header.pan_integration_time = 0;

    snprintf(output_loc, 1024, "%s_12", pointers.data_location);
    snprintf(full_anc_path, 1024, "%s_ANC.h5", output_loc);
    snprintf(sys_command, 2048, "cp -f %s_ANC.h5 %s",
        pointers.data_location,
        full_anc_path);
    error_status = system(sys_command);
    if (error_status != 0)
    {
        IAS_LOG_ERROR("Copying of Ancillary file failed");
        return ERROR;
    }

    /* Open the Output dataset */
    IAS_LOG_DEBUG("Opening %s", output_loc);
    l0r_data_output = ias_l0r_open_from_filename(output_loc,
        IAS_L0R_COMPRESSION_ON);
    if(l0r_data_output == NULL)
    {
        IAS_LOG_ERROR( "Unable to open dataset: %s",
            output_loc);
        return ERROR;
    }
    error_status = ias_l0r_open_header(l0r_data_output, IAS_WRITE);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to open Ancillary file");
        return ERROR;
    }

    error_status = ias_l0r_set_oli_image_header(l0r_data_output,
        &temp_oli_image_header);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to open write dataset");
        return ERROR;
    } 

    error_status = ias_l0r_close_header(l0r_data_output);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to close Ancillary file");
        return ERROR;
    }
    error_status = ias_l0r_close(l0r_data_output);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to close dataset");
        return ERROR;
    }

    return SUCCESS;
}

int set_zeroed_image_header(struct ANCILLARY_POINTERS pointers)
{
    int error_status = 0;
    L0RIO *l0r_data_output = NULL;
    IAS_L0R_OLI_IMAGE_HEADER temp_oli_image_header;
    char output_loc[1024];
    char full_anc_path[1024];
    char sys_command[2048];

    /* Fill the temp image header with all 0s */
    memset(&temp_oli_image_header, 0,
        sizeof(IAS_L0R_OLI_IMAGE_HEADER));

    /* Mark the image header as filled */
    temp_oli_image_header.frame_header.frame_status = IAS_L0R_FILLED;

    snprintf(output_loc, 1024, "%s_13", pointers.data_location);
    snprintf(full_anc_path, 1024, "%s_ANC.h5", output_loc);
    snprintf(sys_command, 2048, "cp -f %s_ANC.h5 %s",
        pointers.data_location,
        full_anc_path);
    error_status = system(sys_command);
    if (error_status != 0)
    {
        IAS_LOG_ERROR("Copying of Ancillary file failed");
        return ERROR;
    }

    /* Open the Output dataset */
    IAS_LOG_DEBUG("Opening %s", output_loc);
    l0r_data_output = ias_l0r_open_from_filename(output_loc,
        IAS_L0R_COMPRESSION_ON);
    if(l0r_data_output == NULL)
    {
        IAS_LOG_ERROR( "Unable to open dataset: %s",
            output_loc);
        return ERROR;
    }
    error_status = ias_l0r_open_header(l0r_data_output, IAS_WRITE);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to open Ancillary file");
        return ERROR;
    }

    error_status = ias_l0r_set_oli_image_header(l0r_data_output,
        &temp_oli_image_header);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to open write dataset");
        return ERROR;
    } 

    error_status = ias_l0r_close_header(l0r_data_output);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to close Ancillary file");
        return ERROR;
    }
    error_status = ias_l0r_close(l0r_data_output);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to close dataset");
        return ERROR;
    }

    return SUCCESS;
}

int set_non_matching_tirs_det_id(struct ANCILLARY_POINTERS pointers)
{
    int error_status = 0;
    L0RIO *l0r_data_output = NULL;
    IAS_L0R_INTERVAL_METADATA temp_metadata;
    char output_loc[1024];
    char full_anc_path[1024];
    char sys_command[2048];

    /* Copy the data, so we can change it*/
    memcpy(&temp_metadata, &pointers.interval_metadata,
        sizeof(IAS_L0R_INTERVAL_METADATA));

    /* Set the detector map id to a value that will differ from the CPF value */
    temp_metadata.detector_map_id_tirs = 999999;

    snprintf(output_loc, 1024, "%s_14", pointers.data_location);
    snprintf(full_anc_path, 1024, "%s_MTA.h5", output_loc);
    snprintf(sys_command, 2048, "cp -f %s_MTA.h5 %s",
        pointers.data_location,
        full_anc_path);
    error_status = system(sys_command);
    if (error_status != 0)
    {
        IAS_LOG_ERROR("Copying of Metadata file failed");
        return ERROR;
    }

    /* Open the Output dataset */
    IAS_LOG_DEBUG("Opening %s", output_loc);
    l0r_data_output = ias_l0r_open_from_filename(output_loc,
        IAS_L0R_COMPRESSION_ON);
    if(l0r_data_output == NULL)
    {
        IAS_LOG_ERROR( "Unable to open dataset: %s",
            output_loc);
        return ERROR;
    }
    error_status = ias_l0r_open_metadata(l0r_data_output, IAS_WRITE);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to open Metadata file");
        return ERROR;
    }

    error_status = ias_l0r_set_interval_metadata(l0r_data_output,
        &temp_metadata);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to open write dataset");
        return ERROR;
    } 

    error_status = ias_l0r_close_metadata(l0r_data_output);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to close Metadata file");
        return ERROR;
    }
    error_status = ias_l0r_close(l0r_data_output);
    if(error_status != SUCCESS)
    {
        IAS_LOG_ERROR( "Unable to close dataset");
        return ERROR;
    }

    return SUCCESS;
}

int notch_dataset(char *file_location, char *dataset_name)
{
    hid_t hdf_file_id;
    hid_t hdf_dataset_id;
    herr_t status;
    hsize_t zero[] = {0};

    hdf_file_id = H5Fopen(file_location, H5F_ACC_RDWR, H5P_DEFAULT);
    if (hdf_file_id < 0)
    {
        IAS_LOG_ERROR("Unable to open HDF file");
        return ERROR;
    }
    hdf_dataset_id = H5Dopen1(hdf_file_id, dataset_name);
    if (hdf_dataset_id < 0)
    {
        IAS_LOG_ERROR("Unable to open dataset");
        return ERROR;
    }
    status = H5Dset_extent(hdf_dataset_id, zero);
    if (status < 0)
    {
        IAS_LOG_ERROR("Unable to shrink dataset to 0");
        return ERROR;
    }

    status = H5Dclose(hdf_dataset_id);
    if (status < 0)
        return ERROR;
    status = H5Fclose(hdf_file_id);
    if (status < 0)
        return ERROR;
    return SUCCESS;
}
