#!/usr/bin/perl
#------------------------------------------------------------------------------
# Name: audit_system_libraries.pl
#
# Description: Finds Perl modules that are referenced by Perl scripts by
# loading that module into the current workspace.  It also finds shared object
# library references for binary files by using the ldd program.  This script
# will report the RPM package that provides the libraries or the path to the
# library if it is not provided by an RPM.
#------------------------------------------------------------------------------
use strict;
use warnings;
use Getopt::Long;
use File::Basename;

use constant {
    BFILE => 1,
    TFILE => 2,
};
use constant {
    TRUE => 1,
    FALSE => 0,
};
# Exit statuses
use constant {
    SUCCESS => 0,
    HELP_INVOKED => 1,
    INVALID_OPTION => 2,
};

# Forward declarations
sub GetFiles();
sub GetDirectoryFiles($$);
sub IsAcceptableFile($$);
sub ProcessTextFiles();
sub ParseTextFile($);
sub ProcessBinaryFiles();
sub ParseBinaryFile($);
sub ProcessOptions();
sub ShowUsage($);

# Global main variables
my $verbose = 0;
my $requestType = 0;
my $perlScripts = 0;
my %moduleList = (); # Keep track of all Perl modules that have been loaded.
my %soList = (); # Keep track of all shared libraries found
my @files; # List of files (TEXT and/or BINARY, depending on request type)

# Let's begin by getting the options
ProcessOptions();

# Parse the rest of the arguments for the files (or directories) to be scanned
@files = @{GetFiles()};
# Find Perl module references
ProcessTextFiles() if ($requestType & TFILE);
# Find shared library references
ProcessBinaryFiles() if ($requestType & BFILE);
# Let them know if it ran through without finding anything
if (($perlScripts == 0) && (!scalar %soList))
{
    print "No references were found.\n";
}
exit SUCCESS;

#------------------------------------------------------------------------------
# Name: GetFiles()
#
# Description: Returns a global list of files (see main @files) using the
# argument file list supplied on the command line.  If the argument list is
# empty, the current directory is assumed.  If a directory is supplied, it
# attempts to process all the files in given directory.  (However, it is not
# recursive.  That is, it does not attempt to process child directories for
# additional files, etc.).  Files can also be directly supplied on the command
# line.
#------------------------------------------------------------------------------
sub GetFiles()
{
    my @files = ();
    if ($#ARGV == -1)
    {
        push (@ARGV, '.'); # Use the current directory as the default
    }
    foreach my $file (@ARGV)
    {
        if (!-e $file)
        {
            print STDERR "$file: No such file or directory\n";
            next;
        }
        if (-d $file)
        {
            push (@files, @{GetDirectoryFiles($file, $requestType)});
        }
        elsif (-f $file)
        {
            if (IsAcceptableFile($file, $requestType))
            {
                push (@files, $file);
            }
        } 
    }
    return \@files;
}

#------------------------------------------------------------------------------
# Name: GetDirectoryFiles(DIR, TYPE)
#
# Description: Returns a list of files in the supplied directory that match the
# type (or types) of files requested.
#------------------------------------------------------------------------------
sub GetDirectoryFiles($$)
{
    my ($dir, $type) = @_;
    my @files = ();
    # Remove any trailing slashes in the directory (for readability)
    $dir = $1 if ($dir =~/(.*)\/+$/);
    if (!opendir(DIR, $dir))
    {
        print STDERR "opendir failed for $dir. $!\n";
        return \@files;
    }
    while (my $file = readdir(DIR))
    {
        my $filepath = "$dir/$file";
        if (IsAcceptableFile($filepath, $type))
        {
            push(@files, $filepath);
        }
    }
    closedir(DIR);
    return \@files;
}

#------------------------------------------------------------------------------
# Name: IsAcceptableFile(FILE, TYPE)
#
# Description: Returns TRUE, if the file matches the type (BINARY and/or TEXT) 
# of the file requested.  If it starts with a period, isn't a file, or does not
# have the executable bit on, it returns FALSE, indicating the file should
# be skipped.  (Note: IAS Constants TRUE and FALSE are not used, since this
# could create an unnecessary dependency on ias_const.pm.)
#------------------------------------------------------------------------------
sub IsAcceptableFile($$)
{
    use File::Basename;
    my ($file, $type) = @_;
    my $basename = basename($file);
    # Skip files beginning with a period
    return FALSE if ($basename =~ m/^\./);
    # File test to ensure that it is a file
    return FALSE if (!-f $file);
    # File test to see if the file has the executable bit on
    return FALSE if (!-x $file);
    # File test to see if the file is a binary file
    if (-B $file)
    {
        if ($type & BFILE)
        {
            return TRUE;
        }
    }
    elsif (-T $file)
    {
        if ($type & TFILE)
        {
            return TRUE;
        }
    }
}

#------------------------------------------------------------------------------
# Name: ProcessTextFiles()
#
# Description: Iterates through the @files array asking ParseTextFile to
# parse for use clauses on the text file (if it is a Perl file).  Later, the
# entire list of Perl modules references directly (or indirectly) by the
# Perl scripts (a subset of the @files array) are printed to STDOUT.
#------------------------------------------------------------------------------
sub ProcessTextFiles()
{
    # Perl portion
    foreach my $file (@files)
    {
        # If this is an ASCII text file ...
        ParseTextFile($file) if (-T $file);
    }

    if ($perlScripts > 0)
    {
        # filter out the modules that are part of the install
        my $installPath = $ENV{IAS_BIN};
        $installPath =~ s/\/$//;
        $installPath = dirname($installPath);

        # Reverse the hash to get the full paths to the modules
        my %modulePaths = reverse %INC;

        # Manually add the DBD::Oracle module to the list since it is used but
        # only loaded dynamically when used (which this script doesn't force)
        $modulePaths{"DBD::Oracle"} = 1;

        # Get the package version numbers
        my %sysPackages;
        my %customPackages;
        foreach (sort keys %modulePaths)
        {
            # Skip modules installed from our source code
            next if ($_ =~ m/^$installPath/);

            # Determine whether any packages provide this module
            my $command = "rpm -q --whatprovides $_";
            my $sysPackage = `$command`;
            chomp $sysPackage;
            if (($sysPackage =~ m/not owned by any package/)
                 || ($sysPackage =~ m/no package provides/))
            {
                # Get the module name in a form that can be "used"
                my $module = basename($_, ".pm");

                # Build a command to run perl with the module loaded and 
                # print the module version number
                my $command = "perl -M$module -e 'print \"\$" .
                    $module . "::VERSION\"'";

                # Get the version
                my $version = `$command`;
                my $status = $? >> 8;
                if ($status != 0)
                {
                    print STDERR "Error returned from: $command\n";
                    return;
                }

                # Save the custom package info
                my $customPackage = "$module $version";
                $customPackages{$customPackage} = 1;
            }
            else
            {
                $sysPackages{$sysPackage} = 1;
            }
        }

        # Print the unique system packages
        print "System Perl modules that were referenced:\n";
        foreach (sort keys %sysPackages)
        {
            print "   $_\n";
        }

        # Print custom packages
        print "Custom Perl modules that were referenced:\n";
        foreach (sort keys %customPackages)
        {
            print "   $_\n";
        }

    }
}

#------------------------------------------------------------------------------
# Name: ParseTextFile(FILE)
#
# Description: Reads the text file, determines if it is a perl file by using the
# directive on the first line of the file.  (Note: This line is assumed to be
# there).  If it is a Perl file, it is scanned (using regex) for use module
# clauses.  Those clauses are ran through an eval method to attempt to load
# those modules into the current workspace.  Once loaded, the %INC hash, will
# be updated with the needed perl module(s) for that use clause.
#------------------------------------------------------------------------------
sub ParseTextFile($)
{
    my $file = shift;
    my $line;
    if (!open (INPUT, $file))
    {
        print STDERR "Could not open file: $file. $!\n";
        return;
    }

    my $counter = 0;
    while ($line = <INPUT>)
    {
        $counter++;
        chomp $line;
        if ($counter == 1)
        {
            # Does the first line of the file ask for perl?
            # Note: $^X is the path to the perl interpreter (e.g. /usr/bin/perl)
            if ($line =~m/^#!\s*$^X($|\s+)/)
            {
                $perlScripts++;
                if ($verbose)
                {
                    # Print header on first occurrence
                    print "Perl scripts that were analyzed:\n"
                        if ($perlScripts == 1);
                    print "   $file\n";
                }
            }
            else
            {
                last;
            }
        }
        # Perl use clause
        if ($line =~m/^\s*use\s+(\w+(::\w+)*)/)
        {
            # The code is assuming the format of the use statement is in this
            # style:
            # (optional white space)use module[::submodule]

            # We could miss it, if the module, say, follows on the next line,
            # or two use statements are on the same line, etc. However, it is
            # likely that style of coding would be caught in a peer review.

            # We could gather a false positive, if the phrase 'use something '
            # happened to be embedded in PERL pod or a PERL HERE-document.
            # Again, that would be a possible, but an unlikely occurrence.

            my $module = $1;
            if ($module ne 'constant' && $module ne 'vars')
            {
                if (!exists $moduleList{$module})
                {
                    my $status;
                    $moduleList{$module} = 1;
                    # When the use clause is evaluated, Perl will attempt to
                    # load that module.  If that module references other
                    # modules (by use clauses), those modules are loaded, as
                    # well. Think of it as a recursive process. When the eval
                    # function returns, Perl will have updated the %INC hash
                    # with the paths to all of the modules referenced. In
                    # effect, we need not look through other IAS Perl libraries,
                    # as long as we are parsing the main Perl scripts.
                    {
                        no warnings;
                        # Suppress Exporter warning(s)
                        $status = eval "use $module; 1";
                    }
                    # The 1 is needed to discern whether the expression
                    # evaluated completely.  If left out, eval would return 
                    # undef whether it worked or not.
                    if (!defined $status)
                    {
                        print STDERR
                            "Cannot find module '$module' used in $file.\n";
                    }
                }
            }
        }
    }
    close (INPUT);
}

#------------------------------------------------------------------------------
# Name: ParseBinaryFile(FILE)
#
# Description: Iterates through the list of files (@files) searching for binary
# files.  On these, it calls ParseBinaryFile($file) to generate a hash of
# shared object files that are referenced.  Then, it prints that list out.
#------------------------------------------------------------------------------
sub ProcessBinaryFiles()
{
    # Binary / Shared library portion
    my $binaryCount = 0;
    foreach my $file (@files)
    {
        if (-B $file)
        {
            $binaryCount++;
            if ($verbose)
            {
                # Print header on first occurrence
                print "Binary files that were analyzed:\n"
                    if ($binaryCount == 1);
                print "   $file\n";
            }
            ParseBinaryFile($file)
        }
    }
    if (scalar %soList)
    {
        # filter out the modules that are part of the install
        my $installPath = $ENV{IAS_BIN};
        $installPath =~ s/\/$//;
        $installPath = dirname($installPath);

        my %sysPackages;
        my %customLibraries;

        foreach (sort keys %soList)
        {
            # Skip modules installed from our source code
            next if ($_ =~ m/^$installPath/);

            # Determine whether any packages provide this library
            my $command = "rpm -q --whatprovides $_";
            my $sysPackage = `$command`;
            chomp $sysPackage;
            if ($sysPackage =~ m/not owned by any package/)
            {
                # If it is a locally installed component of Qt, reduce it
                # to just be one entry
                my $filename = $soList{$_};
                if ($filename =~ m/libQt.*\.so\.([0-9\.]*)/)
                {
                    my $name = "Qt $1";
                    $customLibraries{$name} = 1;
                }
                else
                {
                    # It isn't Qt and isn't provided by a package so just track
                    # the libaries
                    $customLibraries{$_} = 1;
                }
            }
            else
            {
                $sysPackages{$sysPackage} = 1;
            }
        }

        # Print the unique system packages
        print "System library packages that were referenced:\n";
        foreach (sort keys %sysPackages)
        {
            print "   $_\n";
        }

        # Print the custom libraries that aren't part of a package
        print "Custom libraries that were referenced:\n";
        foreach (sort keys %customLibraries)
        {
            if ($soList{$_})
            {
                print "   $_ ($soList{$_})\n";
            }
            else
            {
                print "   $_\n";
            }
        }
    }
}

#------------------------------------------------------------------------------
# Name: ParseBinaryFile(FILE)
#
# Description: Uses ldd to find all the shared object library references that
# this binary executable FILE references.  These are stored in the hash
# %soList and later reported as output.  Any errors with references that are
# 'not found' are reported to STDERR.
#------------------------------------------------------------------------------
sub ParseBinaryFile($)
{
    my $file = shift;
    my $output = `ldd $file`;
    my $ldd_status = $? >> 8;
    if ($ldd_status != 0)
    {
        print STDERR "ldd failed with exit status $ldd_status " .
            "for file: $file.\n";
        return;
    }
    my @lines = split(/\n/,$output);
    foreach my $line (@lines)
    {
        # Example line(s):
        #        libz.so.1 => /usr/lib64/libz.so.1 (0x0000003f9f000000)
        #        /lib64/ld-linux-x86-64.so.2 (0x0000003f9dc00000)
        if ($line =~ m/^\s*(|(lib\S+\.so\.\d+) => )(\S+) \([0-9a-fx]+\)$/)
        {
            my $soFile = $3;
            if (!exists $soList{$soFile})
            {
                $soList{$soFile} = ''; # Symbolic link value, if any
                if (-l $soFile)
                {
                    my $target = readlink $soFile;
                    if (defined $target)
                    {
                        $soList{$soFile} = $target;
                    }
                }
            }
        }
        else
        {
            # Print message that the shared library wasn't found.
            # Remove leading whitespace, and ' => '.
            $line =~s/^\s+//;
            $line =~s/ =>//;
            print STDERR "Error: Shared library $line for $file.\n";
        }
    }
}

#------------------------------------------------------------------------------
# Name: ProcessOptions()
#
# Description:  Using Getopt::Long, this routine parses the argument list,
# flagging any options found in the parameter list.  If an invalid option is
# given or the help option is requested, a usage message is shown and the
# program exits immediately.
#------------------------------------------------------------------------------
sub ProcessOptions()
{
    my $help = '';
    my $sharedOnly = '';
    my $perlOnly = '';
    my $result = GetOptions(
        "help|h" => \$help,
        "sharedOnly|s" => \$sharedOnly,
        "perlOnly|p" => \$perlOnly,
        "verbose|v" => \$verbose,
    );
    ShowUsage(INVALID_OPTION) if (!$result);
    ShowUsage(HELP_INVOKED) if $help;

    if (!$sharedOnly && !$perlOnly)
    {
        $requestType = BFILE | TFILE;
    }
    else
    {
        $requestType |= BFILE if ($sharedOnly);
        $requestType |= TFILE if ($perlOnly);
    }
}

#------------------------------------------------------------------------------
# Name: ShowUsage(EXITVALUE)
#
# Description: Shows the help or usage directives for the program.
#
# Exits with parameter value.
#------------------------------------------------------------------------------
sub ShowUsage($)
{
    my $exitvalue = shift; # I: Exit status
    print << "USAGE";
find_references.pl [Options] [Files]
Lists all references used by the files (the current directory by default).
Options:
  --help (-h)
        Prints a help summary
  --sharedOnly (-s)
        Prints all shared libraries referenced by the files
  --perlOnly (-p)
        Prints all Perl modules used by the files
  --verbose (-v)
        Prints the application names that were analyzed
USAGE
    exit $exitvalue;
}
