/****************************************************************************
Name: calc_offsets

Description: This is a quick and dirty hack to calculate the band and sca
    offset values for the CPF based on the legendre coefficients.
****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include "ias_satellite_attributes.h"
#include "ias_math.h"
#include "ias_cpf.h"

int main(int argc, char *argv[])
{
    IAS_CPF *cpf;
    int band_index;
    int sca_index;
    int band_numbers[IAS_MAX_NBANDS];
    double ifovs[IAS_MAX_NBANDS];
    double band_offsets[IAS_MAX_NBANDS];
    double sca_offsets[IAS_MAX_NBANDS][IAS_MAX_NSCAS];
    double min_sca_offsets[IAS_MAX_NBANDS];
    int sca_counts[IAS_MAX_NBANDS];
    const IAS_CPF_OLI_PARAMETERS* oli_parameters = NULL;
    const IAS_CPF_TIRS_PARAMETERS* tirs_parameters = NULL;
    const IAS_CPF_FOCAL_PLANE *focal_plane = NULL;
    double ref_legendre;

    /* Get the CPF name from the command line parameters */
    if (argc != 2)
    {
        printf("Usage: calc_offsets CPF_filename\n");
        return EXIT_FAILURE;
    }

    /* Initialize the satellite attributes */
    if (ias_sat_attr_initialize(IAS_L8) != SUCCESS)
    {
        printf("Error: Failed initializing the satellite attributes library\n");
        return EXIT_FAILURE;
    }

    /* Initialize the array of band numbers */
    for (band_index = 0; band_index < IAS_MAX_NBANDS; band_index++)
    {
        int band_number;

        band_number = ias_sat_attr_convert_band_index_to_number(band_index);
        if (band_number == ERROR)
        {
            printf("Error converting band index %d to a band number\n",
                band_index);
            return EXIT_FAILURE;
        }
        band_numbers[band_index] = band_number;
    }

    /* Read the CPF */
    cpf = ias_cpf_read(argv[1]);
    if (!cpf)
    {
        printf("Error: unable to read CPF %s\n", argv[1]);
        return EXIT_FAILURE;
    }

    /* Get the IFOV of the bands */
    oli_parameters = ias_cpf_get_oli_parameters(cpf);
    if (!oli_parameters)
    {
        printf("Error getting the OLI parameters\n");
        return EXIT_FAILURE;
    }
    tirs_parameters = ias_cpf_get_tirs_parameters(cpf);
    if (!tirs_parameters)
    {
        printf("Error getting the TIRS parameters\n");
        return EXIT_FAILURE;
    }

    /* Get the focal plane data from the CPF */
    focal_plane = ias_cpf_get_focal_plane(cpf);
    if (!focal_plane)
    {
        printf("Error getting the focal plane parameters\n");
        return EXIT_FAILURE;
    }

    /* Build an array of IFOVs for each band */
    for (band_index = 0; band_index < IAS_MAX_NBANDS; band_index++)
    {
        IAS_SPECTRAL_TYPE spec = 
            ias_sat_attr_get_spectral_type_from_band_number(
                band_numbers[band_index]);
        if (spec == ERROR)
        {
            printf("Error getting the spectral type for band number %d\n",
                band_numbers[band_index]);
            return EXIT_FAILURE;
        }
        if (spec == IAS_SPECTRAL_PAN)
            ifovs[band_index] = oli_parameters->along_ifov_pan;
        else if (spec == IAS_SPECTRAL_THERMAL)
            ifovs[band_index] = tirs_parameters->along_ifov_thermal;
        else
            ifovs[band_index] = oli_parameters->along_ifov_ms;
    }
    
    /* Calculate the reference legendre value for the worst TIRS band */
    ref_legendre = ias_math_eval_legendre(0.0,
            focal_plane->along_los_legendre[8][0], 4);

    /* Calculate the band offsets relative to the first band by using the first
       SCA's legendre coefficients */
    for (band_index = 0; band_index < IAS_MAX_NBANDS; band_index++)
    {
        double legendre;

        /* Calculate the current band's legendre value */
        legendre = ias_math_eval_legendre(0.0,
            focal_plane->along_los_legendre[band_index][0], 4);

        /* Now calculate the band offset relative to the reference band */
        band_offsets[band_index] = (ref_legendre - legendre)
                / ifovs[band_index];
    }

    /* Calculate the SCA offsets for each band using the legendre
       coefficients */
    for (band_index = 0; band_index < IAS_MAX_NBANDS; band_index++)
    {
        int sca_count;

        /* Get the number of SCAs for this band */
        sca_count = ias_sat_attr_get_scas_per_band(band_numbers[band_index]);
        if (sca_count == ERROR)
        {
            printf("Error getting SCA count for band number %d\n", 
                band_numbers[band_index]);
            return EXIT_FAILURE;
        }
        sca_counts[band_index] = sca_count;

        /* Calculate the reference legendre value for this band */
        ref_legendre = ias_math_eval_legendre(0.0,
            focal_plane->along_los_legendre[band_index][0], 4);

        /* initialize the min SCA offset for later normalizing them to zero */
        min_sca_offsets[band_index] = 1000000.0;

        /* Calculate the offset relative to the first SCA */
        for (sca_index = 0; sca_index < sca_count; sca_index++)
        {
            double legendre;
            double sca_offset;

            /* Calculate the current band's legendre value */
            legendre = ias_math_eval_legendre(0.0,
                focal_plane->along_los_legendre[band_index][sca_index], 4);

            /* Now calculate the band offset relative to the reference band */
            sca_offset = (legendre - ref_legendre) / ifovs[band_index];

            sca_offsets[band_index][sca_index] = sca_offset;

            if (sca_offset < min_sca_offsets[band_index])
                min_sca_offsets[band_index] = sca_offset;
        }

        /* Adjust the offsets to be relative to the top-most SCA */
        for (sca_index = 0; sca_index < sca_count; sca_index++)
        {
            sca_offsets[band_index][sca_index] -= min_sca_offsets[band_index];
        }
    }

    /* Free the cpf since it is no longer needed */
    ias_cpf_free(cpf);
    cpf = NULL;

    /* Print out the groups for the CPF */
    printf("GROUP = OLI_FOCAL_PLANE\n");
    printf("  Band_Offset = (");
    for (band_index = 0; band_index < 9; band_index++)
    {
        int offset = (int)(band_offsets[band_index] + 0.5);
        if (band_index != 0)
            printf(", %d", offset);
        else
            printf("%d", offset);
    }

    printf(")\n");

    for (band_index = 0; band_index < 9; band_index++)
    {
        printf("  SCA_Offset_B%02d = (", band_numbers[band_index]);
        for (sca_index = 0; sca_index < sca_counts[band_index]; sca_index++)
        {
            int offset = (int)(sca_offsets[band_index][sca_index] + 0.5);
            if (sca_index != 0)
                printf(", %d", offset);
            else
                printf("%d", offset);
        }
        printf(")\n");
    }

    printf("END_GROUP = OLI_FOCAL_PLANE\n");

    printf("GROUP = TIRS_FOCAL_PLANE\n");
    printf("  Band_Offset = (");
    for (band_index = 9; band_index < 11; band_index++)
    {
        /* for the TIRS bands, just make them relative to the first TIRS band */
        int offset = (int)(band_offsets[band_index] - band_offsets[9] + 0.5);

        if (band_index != 9)
            printf(", %d", offset);
        else
            printf("%d", offset);
    }
    printf(")\n");

    for (band_index = 9; band_index < 11; band_index++)
    {
        printf("  SCA_Offset_B%02d = (", band_numbers[band_index]);
        for (sca_index = 0; sca_index < sca_counts[band_index]; sca_index++)
        {
            int offset = (int)(sca_offsets[band_index][sca_index] + 0.5);
            if (sca_index != 0)
                printf(", %d", offset);
            else
                printf("%d", offset);
        }
        printf(")\n");
    }

    printf("END_GROUP = OLI_FOCAL_PLANE\n");
    
    return EXIT_SUCCESS;
}
