#!/usr/bin/perl -w
############################################################################
#   NAME:  check_parameter_changes.pl
#
#   PURPOSE: The check parameter work orders routine will compare the
#           new loadtable output with the controlled previous output.
#           If a change exists, check the differences and make corrections
#           to the loader files in the dpas_database
#
#   PARAMETERS: None
#
#   OUTPUT: Error log messages
#           File Differences
#
# TODO write a new note!!
#   NOTE:   All parameters are optional.  If the user enters no parameters,
#           the output files should be identical and this program will
#           return successful.  In order to give this routine flexibility,
#           The user is allowed to enter parameters.  The work order entered
#           will reflect the entered parameters but will likely cause 
#           an error upon return.
#
############################################################################

use strict;

use File::Compare;
use Getopt::Long;
use Logger;

# include global routines
use ias_create_work_order qw
{
    ias_create_work_order
};

use ias_const;
use ias_processing_system;

# declare some constants
use constant PROG       => "createwotest";

# debug flag is a method of having select information printed to standard out
# set to 1 to see debug output, set to 0 shut off debug output
use constant DEBUG      =>  0;      # Debug flag

#############################################################################
# Variable declarations
#############################################################################
my $sql;                # SQL statement
my $wo_id;              # Subroutine return value
my $scene_id;           # Scene ID for the product
my $product_id;         # Product ID for work order

my $help;               # Validate parameters, print usage and exit
my $wo_comment = "";    # Comment to be attached to the work order
my $rtn;                # Return value from subroutines
my $delete_flag_str;    # String entered by user as a parameter
my $delete_flag;        # True/False to delete database entries
my $g_Logger;           # The logger object to use for log messages.

# retrieve the optional command-line parameters
#TODO change the optional parameters!!!!
my %optctl = (
    delete_flag => \$delete_flag_str,
    help => \$help
    );

my $opt_status = GetOptions (\%optctl,
     "delete_flag=s",
     "help",
     );


# initialize the ias_ProcSys module
if(ias_init_system() == ERROR)
{
    print STDERR PROG . "$ias_error_msg\n";
    exit FAILURE;
}

# initialize the logger object
$g_Logger = Logger->new($ias_log_filename);
if(!$g_Logger)
{
    print STDERR PROG . ": ERROR: Failed to establish logger object.\n";

    exit FAILURE;
}

set_default_log($g_Logger);
$g_Logger->set_module_name(PROG);

if ($opt_status == 0)
{
    usage();
    $g_Logger->fatal_msg(__FILE__, __LINE__,
                        "Unable to process arguments");
    $g_Logger->close_log();
    exit FAILURE;
}

if (DEBUG)
{
    print "HELP: $help\n";
}

if ($help)
{
    usage();
    $g_Logger->info_msg(__FILE__, __LINE__,
                        "Usage help completed");
    $g_Logger->close_log();
    exit SUCCESS;
}

my @app_list = qw (
    ancillary_data_preprocessor
    b2b_char
    bandcal
    char_nonuniformity
    char_rad_stability
    correct_los_model
    create_grid
    create_l1r
    create_los_model
    focal_plane_legendre
    format_and_package
    gcpcorrelate
    geodetic
    geometric
    geomresample
    i2i_char
    makegeomgrid
    resample
    retrieve_elevation
    setup_work_order
    solarcal_oli
    tdist
    terrain_occlusion
    );

my $systemcall;
my $diff_file;

foreach my $app (@app_list)
{
    #my $loadtablefile = $app_list[$i] + "_loadtable.txt";
    my $loadtablefile = $app . "_loadtableorig.txt";
    my $tmploadtable = $app . "_loadtabletmp.txt";

    # Remove the messages from the output and write to the temp file
    $systemcall = 
       "$app --loadtable | grep -v -e INFO -e WARN -e ERROR > $tmploadtable";
    system($systemcall);

    #TODO check if system call failed

    if (compare($tmploadtable, $loadtablefile) == 0)
    {
        print "$app has not changed\n";
    }
    else
    {
        $diff_file = $app . "_loadtablediff.txt";
        $systemcall = "diff $tmploadtable $loadtablefile > $diff_file";
        system($systemcall);
        print "$app is different, please investigate diff file\n";
    }
}

$g_Logger->close_log();

##############################################################################
#               Subroutines defined here                                     #
##############################################################################


##############################################################################
# Name: usage
#
# Purpose: Display command line options and general description of program
#
##############################################################################
sub usage
{
    print << "HELP";
    USAGE:
    $0 [-help]

    -help           - Display this usage message.  No processing will be
                      performed (optional)

    check_parameter_changes.pl is a command line interface to test 
    if the parameters have changed from the version controlled output.
    This routine compares the differences to the output when the 
    flag -loadtable is used when running an application process, such 
    as <process> -loadtable.

    The results are printed to standard out.

    A file with the current output will be written to the current working
    directory.  It will be named <process>_loadtabletmp.txt

    If the new output is different from the old output, a difference file
    will be printed called <process>_loadtablediff.txt.

    This routine is intended to run with no arguments.

    Examples:
    check_parameter_changes.pl
 
    A list of the results will be printed to standard out.  An example of
    the output looks something like this:

        ancillary_data_preprocessor has not changed
        b2b_char has not changed
        bandcal has not changed
        char_rad_stability is different, please investigate diff file
        correct_los_model has not changed
        create_grid has not changed
        create_l1r is different, please investigate diff file
        create_los_model has not changed
        focal_plane_legendre has not changed
        format_and_package is different, please investigate diff file
        gcpcorrelate has not changed
        geodetic has not changed
        geomresample has not changed
        i2i_char has not changed
        makegeomgrid has not changed
        resample has not changed
        retrieve_elevation has not changed
        setup_work_order is different, please investigate diff file
        solarcal_oli is different, please investigate diff file
        tdist has not changed

    In this case, investigations are necessary for char_rad_stability,
    create_l1r, format_and_package, setup_work_order, and solarcal_oli.
    The parameters in the other routines have not changed.

HELP
}

