#!/usr/local/bin/python3
""" The scripts dumps out database tables as Oracle sqlldr files
"""
from ias.db import database as db
from ias.log.logger import log_exceptions
from ias.log import logger
from ias.misc.const import SUCCESS, ERROR, EXIT_SUCCESS, EXIT_FAILURE
from ias.misc import satellite_attributes as sat_attr
import argparse
import re
import sys
import os.path


class SqlldrError(Exception):
    pass


class UnsupportedColumnType(SqlldrError):
    pass


class CreateSqlldr():
    """ Class to create Oracle sqlldr files from database contents
    """
    class LoaderColumn:
        """ Class representing a database column
        """
        def __init__(self, column_name_in, type_in, size_in=0):
            """ Construct a new LoaderColumn object.

                Parameters:
                    column_name_in: Name of the column
                    type_in: Data type of column (DATE, NUMBER, VARCHAR2, or
                             TIMESTAMP)
                    size_in: Size of a VARCHAR2 string, 0 for all other types
            """
            self.column_name = column_name_in
            self.column_type = type_in
            # Note the size is only used for VARCHAR2 types, it is 0 for
            # other types
            self.column_size = size_in

        def __str__(self):
            """ Return a string representation of a LoaderColumn object.
                The string consists of name and type separated by a ':'.
                Additionally for VARCHAR2 types, the size is included in
                parentheses following the type.
            """
            if self.column_size > 0:
                line = "{0}:{1}({2})".format(self.column_name,
                                             self.column_type,
                                             self.column_size)
            else:
                line = "{0}:{1}".format(self.column_name,
                                        self.column_type)
            return line

    def __init__(self, db_in, table_name_in, dir_in=".", where_in=None,
                 skip_in=False, load_in="append"):
        """ Construct a new CreateSqlldr object.

            Parameters:
                db_in: Already created database connection
                table_name_in: Name of table to create files for
                dir_in: Location to create files. Defaults to CWD.
                where_in: Optional WHERE clause to filter results
                skip_in: If true unsupported columns will not cause
                         a failure and will be ignored. If false,
                         the default, unsupported columns cause
                         an exception.
        """
        self.db = db_in

        # Check if a schema is specified with the table name
        schema_match = re.match(r'(\w+)\.(\w+)', table_name_in)
        if schema_match:
            self.schema_name = schema_match.group(1).upper()
            self.table_name = schema_match.group(2).upper()
        else:
            self.table_name = table_name_in.upper()
            self.schema_name = None

        self.columns = []
        if not os.path.isdir(dir_in):
            raise SqlldrError("{0} is not a dir".format(dir_in))
        self.output_dir = dir_in
        self.where = where_in
        self.skip = skip_in
        self.load_method = load_in.upper()

    def _populate_columns(self):
        """ Populates columns member of this object

            Parameters:
                None

            Returns:
                None
        """
        # Look in user_tab_columns if the table is in our schema
        if self.schema_name is None:
            select_statement = (
                "SELECT column_name, data_type, data_length FROM "
                "user_tab_columns WHERE table_name = '{0}'".format(
                                                             self.table_name))
        # Look in all_table_columns if the table is in another schema
        else:
            select_statement = (
                "SELECT column_name, data_type, data_length FROM "
                "all_tab_columns WHERE table_name = '{0}' and "
                "owner = '{1}'".format(self.table_name, self.schema_name))
        cursor = self.db.get_cursor()
        cursor.execute(select_statement)
        rows = cursor.fetchall()
        cursor.close()
        unsupported_columns = []
        if len(rows) < 1:
            raise SqlldrError("No columns found for {0}. Does table exist?".
                              format(self.table_name))
        for row in rows:
            # The TIMESTAMP returned by Oracle contains more info that we want
            # (w/ or w/o TIME ZONE, precision, etc.)
            if re.match(r"TIMESTAMP", row[1]):
                # TIMESTAMP (w/o the extra info) is stored as the internal type
                column = self.LoaderColumn(row[0], "TIMESTAMP")
                self.columns.append(column)
            elif row[1] == "VARCHAR2":
                column = self.LoaderColumn(row[0], row[1], row[2])
                self.columns.append(column)
            elif row[1] == "NUMBER" or row[1] == "DATE":
                column = self.LoaderColumn(row[0], row[1])
                self.columns.append(column)
            else:
                msg = ("Type of {0} for column {1} is not supported").format(
                       row[1], row[0])
                unsupported_columns.append(msg)

        # Check we had at least one supported column type
        if len(self.columns) < 1:
            raise SqlldrError("No supported column types in table")

        # Check if we had any unsupported column types
        if len(unsupported_columns) > 0:
            raise UnsupportedColumnType(". ".join(unsupported_columns))

    def _create_ctl_file(self):
        """ Creates the .ctl file for the table associated with this object

            Parameters:
                None

            Returns:
                None
        """
        content = """OPTIONS (ERRORS=50)
        LOAD DATA
        INFILE 'ABC12345_TABLE.ldr' "str '@@@\\n'"
        LOAD12345
        CONTINUEIF NEXT(1:1) = '#'
        INTO TABLE "ABC12345_TABLE"
        FIELDS TERMINATED BY'|'
        OPTIONALLY ENCLOSED BY '"' AND '"'
        TRAILING NULLCOLS (
        """
        # remove the indents
        content = re.sub(r'^ +', '', content, flags=re.MULTILINE)
        # replace the dummy name with the real table name
        content = re.sub(r'ABC12345_TABLE', self.table_name, content)
        # replace the dummy load method with the one of the options
        content = re.sub(r'LOAD12345', self.load_method, content)

        # Add in the columns
        for col in self.columns:
            if col.column_type == "NUMBER":
                content += '"{0}",\n'.format(col.column_name)
            elif col.column_type == "VARCHAR2":
                content += '"{0}" CHAR({1}),\n'.format(col.column_name,
                                                       col.column_size)
            elif col.column_type == "DATE":
                content += '"{0}" DATE "YYYY-MM-DD HH24:MI:SS",\n'.format(
                                                            col.column_name)
            elif col.column_type == "TIMESTAMP":
                content += \
                    '"{0}" TIMESTAMP "YYYY-MM-DD HH24:MI:SS.FF",\n'.format(
                                                             col.column_name)
            else:
                # This indicates a programming error, throw a generic exception
                # and not UnsupportedColumnType
                raise Exception("Unknown column type")
        # Remove final ","
        content = re.sub(r',$', '', content)
        content += ")\n"

        filename = "{0}/{1}.ctl".format(self.output_dir,
                                        self.table_name)
        with open(filename, "w") as ctl_file:
            ctl_file.write(content)

    def _create_ldr_file(self):
        """ Creates the .ldr file for the table associated with this object

            Parameters:
                None

            Returns:
                None
        """
        columns = [col.column_name for col in self.columns]
        if self.schema_name:
            select_statement = (
                 "SELECT {0} FROM {1}.{2}".format(", ".join(columns),
                                                  self.schema_name,
                                                  self.table_name))
        else:
            select_statement = "SELECT {0} FROM {1}".format(", ".join(columns),
                                                            self.table_name)
        if self.where:
            select_statement += " WHERE " + self.where
        cursor = self.db.get_cursor()
        cursor.arraysize = 1000
        cursor.execute(select_statement)

        filename = "{0}/{1}.ldr".format(self.output_dir,
                                        self.table_name)
        with open(filename, "w") as ldr_file:
            while 1:
                rows = cursor.fetchmany()
                if (len(rows) <= 0):
                    break
                for r in rows:
                    # Oracle weirdness, each row starts with a space
                    row_str = " "
                    for i in range(len(self.columns)):
                        if r[i] is None:
                            # If the row is NULL, put nothing in between the
                            # pipes
                            row_str += '|'
                        elif self.columns[i].column_type == "VARCHAR2":
                            # Quote the strings
                            row_str += '"' + str(r[i]) + '"|'
                        elif self.columns[i].column_type == "TIMESTAMP":
                            # Can't use the default str with a timestamp,
                            # since it prints differently for w/ or w/o
                            # sub-seconds w/ and w/o time zone
                            row_str += ('"' + r[i].strftime("%Y-%m-%d %X.%f") +
                                        '"|')
                        else:
                            # Print other datatypes as is
                            row_str += str(r[i]) + '|'
                    # @@@ is part of the record separator, so we can have
                    # newlines within the records
                    ldr_file.write(row_str + '@@@\n')

    def dump(self):
        """ Main function which populates the columns, and creates the files
            for the table assoicated with this object

            Parameters:
                None

            Returns:
                None
        """
        try:
            self._populate_columns()
        except UnsupportedColumnType as Ex:
            if self.skip:
                # Just print the exception if the user set the skip option
                print(Ex)
            else:
                # Re-raise if skip is not set
                raise(Ex)
        self._create_ctl_file()
        self._create_ldr_file()


def connect_to_database():
    """ Connect to the database using the currently configured satellite/sensor
        schema.  The caller is expected to have properly initialized the
        satellite attributes library prior to calling this routine.
    """
    # Get the satellite attributes
    sa = sat_attr.get_default()

    # Connect to the database sensor schema
    db_connection = db.DatabaseConnection()
    db_connection.set_login_using_satellite_sensor(
        sa.get_satellite_name(), sa.get_satellite_sensor_name())
    db_connection.connect()

    # return the DB connection
    return db_connection


def parse_parameters():
    """ Parses command line arguments for this script

            Parameters:
                arguments: List of args to parse

            Returns:
                Namespace populated with argument values
    """
    argparser = argparse.ArgumentParser(
        description='Dump table as sqlldr files')

    argparser.add_argument(
        'table_name', nargs='+',
        help="Table name(s) to dump. A schema name can optionally precede "
             "the table name in front of a '.'.")

    argparser.add_argument(
        '--path', default='.', help='Output the sqlldr files to this path')

    argparser.add_argument(
        '--where', default=None, help='Filter the table with a WHERE clause. '
        'This option is limited to dumping one table at a time')

    argparser.add_argument(
        '--skip_unsupported', action='store_true',  default=False,
        help='Remove unsupported column types from output')

    argparser.add_argument(
        '--load_method', choices=['append', 'truncate', 'replace'],  default='append',
        help='Type of load to specify in control file')

    par = argparser.parse_args()

    # If a WHERE clause is passed, limit to dumping one table
    if par.where and len(par.table_name) > 1:
        raise SqlldrError("Only one table can be dumped at time when "
                          "passing a WHERE clause")
    return par


def main():
    status = EXIT_SUCCESS
    db_conn = None
    try:
        # Initialize the logger
        log = logger.Logger("Create sqlldr files")

        # Init satellite attributes
        sat_attr.set_default(sat_attr.get_satellite_id_from_number(8))

        parsed_args = parse_parameters()

        # Connect to the database
        db_conn = connect_to_database()

        for tab in parsed_args.table_name:
            dumper = CreateSqlldr(db_conn, tab, parsed_args.path,
                                  parsed_args.where,
                                  parsed_args.skip_unsupported,
                                  parsed_args.load_method)
            dumper.dump()
    except SqlldrError as ex:
        log.error("Exception running the dumper: %s", ex)
        status = EXIT_FAILURE
    finally:
        if db_conn is not None:
            db_conn.close(db.DISCONNECT_ROLLBACK)
        log.close()

    exit(status)

# Run the main routine if called from command line
if __name__ == "__main__":
    main()
