#!/usr/bin/perl -w
#------------------------------------------------------------------------------
# NAME: dev_tram_loader.pl
#
# PURPOSE: Load a developers TRAM tables with orders specified in a SOAP
#          Envelope formatted file.
#
# PARAMETERS: SOAP Envelope file
#
# EXAMPLE USAGE:
#     dev_tram_loader.pl test_case_004_soapenv.xml
#
# OUTPUT: SUCCESS
#         FAILURE
#------------------------------------------------------------------------------

use strict;
use DBI;
use POSIX;
use IO::File;

use ias_const;
use ias_processing_system;

## Sub Declarations
sub initDb();
sub closeDb();
sub loadWoFile($$);
sub createTramOrder($$);
sub createTramUnit($$$$$$);

## Database configuration
my $tramDbh;
my $woDbh;

##############################################################################
## Main
##############################################################################

if (not defined $ARGV[0])
{
    die ("Error: Missing command line argument.");
}

my $outputFilename = undef;
if (defined $ARGV[1])
{
    $outputFilename = $ARGV[1];
}

print "Processing $ARGV[0]\n";

if (ias_init_system() != SUCCESS)
{
    print "ERROR: could not initialize ias system: $ias_error_msg\n";
    exit FAILURE;
}

# Open database connections
if (initDb() == ERROR)
{
    print "ERROR: could not connect to database\n";
    exit FAILURE;
}

# Read the input
my $order_hash = {};
if (loadWoFile($ARGV[0], $order_hash) != SUCCESS)
{
    print "ERROR: could not load wo file $ARGV[0]\n";
    exit FAILURE;
}

# Open the output file
if (defined $outputFilename)
{
    open(OUTPUT, ">", $outputFilename) 
        or die "Error: Cannot open $outputFilename: $!";
}

# Create an order number by using current time in seconds
my $orderNumber = sprintf("%013d", time);
my $priority = $order_hash->{"priority"};

if(createTramOrder($orderNumber, $priority) != SUCCESS)
{
    die "ERROR: could not create test orders in TRAM database";
}

# Loop through each unit and call createTramUnit
my $unitCount = 1; # Start with unit 1
for my $scene (@{$order_hash->{"units"}})
{
    my $unitNumber = sprintf("%05d", $unitCount);
    my $sceneId = $scene->{"sceneId"};
    my $productCode = $scene->{"productName"};
    my $priority = $order_hash->{"priority"};
    my $comment = $scene->{"unitComment"};

    print "Creating $orderNumber $unitNumber\n";
    if (defined $outputFilename)
    {
        print OUTPUT "$orderNumber\_$unitNumber $sceneId\n";
    }

    if(createTramUnit($orderNumber, $unitCount, $sceneId,
                      $productCode, $priority, $comment) != SUCCESS)
    {
        $tramDbh->rollback();
        die "ERROR: could not create test orders in TRAM database";
    }

    #Advance to the next unit
    $unitCount++;
}

if (defined $outputFilename)
{
    close(OUTPUT);
}

closeDb();

print "Exiting dev_tram_loader.pl\n";

exit SUCCESS;

##############################################################################
## Subroutines
##############################################################################

##############################################################################
# Name: initDb
# Purpose: Initiate database connections to TRAM and WO_COMMON
# Returns: SUCCESS - database connections were successfully opended
#          ERROR - error occurred
# Parameters: None
##############################################################################
sub initDb()
{
    my $tramEnv = $ENV{"IAS_DB_TRAM"};
    my $woEnv = $ENV{"IAS_DB_COM"};
    my $sid = $ENV{"ORACLE_SID"};
    my $username;
    my $password;

    # verfiy environment variables were set
    if (!defined $tramEnv)
    {
        print "ERROR: IAS_DB_TRAM not defined\n";
        return ERROR;
    }

    if (!defined $woEnv)
    {
        print "ERROR: IAS_DB_COM not defined\n";
        return ERROR;
    }

    if (!defined $sid)
    {
        print "ERROR: ORACLE_SID not defined\n";
        return ERROR;
    }

    ## Open TRAM database connection
    ($username, $password) = split(/\//,$tramEnv);

    $tramDbh = DBI->connect("DBI:Oracle:$sid", $username, $password,
                           {AutoCommit => 0, PrintError =>0});

    if (!$tramDbh)
    {
        print "ERROR: could not connect to TRAM db: $DBI::errstr\n";
        return ERROR;
    }

    ## Open work order database connection
    ($username, $password) = split(/\//,$woEnv);

    $woDbh = DBI->connect("DBI:Oracle:$sid", $username, $password,
                           {AutoCommit => 0, PrintError =>0});

    if (!$woDbh)
    {
        print "ERROR: could not connect to WO_COMMON: $DBI::errstr\n";
        return ERROR;
    }
    
    return SUCCESS;
}

##############################################################################
# Name: closeDb
# Purpose: Close database connections to TRAM and WO_COMMON
# Returns: None
# Parameters: None
##############################################################################
sub closeDb()
{
    $tramDbh->disconnect();
    $woDbh->disconnect();
}

##############################################################################
# Name: createTramOrder
# Purpose: Create an order in the TRAM tables
# Returns: SUCCESS - Order was successfully created
#          ERROR - Order was not created
# Parameters: orderNbr - TRAM order number
#             priority - 
##############################################################################
sub createTramOrder($$)
{
    my ($orderNbr, $priority) = @_;
    my $insertCnt;

    my $orderSql = "INSERT INTO tram_orders "
                 . "            (order_nbr, "
                 . "             order_status, "
                 . "             date_entered, "
                 . "             priority) "
                 . "     VALUES ('$orderNbr', "
                 . "             'Q', "
                 . "             sysdate, "
                 . "             '$priority')";

    my $sth = $tramDbh->prepare($orderSql);

    $insertCnt = $sth->execute();
    $sth->finish();
    
    if ($insertCnt != 1)
    {
        $tramDbh->rollback();
        return ERROR;
    } 

    return SUCCESS;
}

##############################################################################
# Name: createTramUnit
# Purpose: Create an order in the TRAM tables
# Returns: SUCCESS - Order was successfully created
#          ERROR - Order was not created
# Parameters: orderNbr - TRAM order number
#             unitNbr - TRAM unit number
#             landsatSceneId -
#             prodCode - 
#             priority - 
# Notes: Right now the unitkey and lpgsondkey are hard-coded to '1'.  In TRAM
#   these are primary key and should be unique, but replicated TRAM in IAS
#   doesn't have these constraints.  DCO does not validate the acquistion date
#   against the scene Id, so the acquisition_date in the tram_lpgs_orders table
#   can be set to sysdate. 
##############################################################################
sub createTramUnit($$$$$$)
{
    my ($orderNbr, $unitNbr, $landsatSceneId,
        $prodCode, $priority, $comment) = @_;
    my $insertCnt;

    my $unitSql = "INSERT INTO tram_units "
                . "            (unitkey, "
                . "             order_nbr, "
                . "             unit_nbr, "
                . "             unit_status, "
                . "             ordering_id, "
                . "             prod_code, "
                . "             unit_comment, "
                . "             date_queued) " 
                . "     VALUES (1, "
                . "             '$orderNbr', "
                . "             $unitNbr, "
                . "             'Q', "
                . "             '$landsatSceneId', "
                . "             '$prodCode', "
                . "             '$comment', "
                . "             sysdate)";

    my $sth = $tramDbh->prepare($unitSql);

    $insertCnt = $sth->execute();
    $sth->finish();
    if ($insertCnt != 1)
    {
        $tramDbh->rollback();
        return ERROR;
    } 

    # Parse the path/row and acquisition date
    my $wrsPath = substr($landsatSceneId, 3, 3);
    my $wrsRow = substr($landsatSceneId, 6, 3);

    my $lpgsSql = "INSERT INTO tram_lpgs_orders "
                . "            (lpgsondkey, "
                . "             order_nbr, "
                . "             unit_nbr, "
                . "             satellite_id, "
                . "             wrs_path, "
                . "             wrs_row, "
                . "             landsat_scene_id, "
                . "             acquisition_date, "
                . "             priority, "
                . "             prod_code) "
                . "     VALUES (1, "
                . "             '$orderNbr', "
                . "             $unitNbr, "
                . "             'L8ORDER', "
                . "             $wrsPath, "
                . "             $wrsRow, "
                . "             '$landsatSceneId', "
                . "             sysdate, "
                . "             $priority, "
                . "             '$prodCode')";

    $sth = $tramDbh->prepare($lpgsSql);

    $insertCnt = $sth->execute();
    $sth->finish();
    if ($insertCnt != 1)
    {
        $tramDbh->rollback();
        return ERROR;
    } 

    if (!$tramDbh->commit())
    {
        print "ERROR: Create Tram unit commit failed: $DBI::errstr\n";
        return ERROR;
    }
    return SUCCESS;
}

sub loadWoFile($$)
{
    my ($filename, $hash_ref) = @_;

    # Open the input file contains the order details
    my $fh = IO::File->new($filename, 'r');
    if (not defined $fh)
    {
        die("Error: openning $filename failed");
    }

    # create a new unit array in the order hash reference
    my $units_list = [];

    my $section = "";
    while (<$fh>)
    {
        chomp;

        # skip comments and blank lines
        if ($_ =~ /^\s*(#|$)/) { next; }

        # check for a new section
        if ($_ =~ /^\[([\w\s]+)\]$/)
        {
            $section = $1;
            next;
        }

        if ($section eq "order" || $section eq "service")
        {
            my ($key, $value) = split("=", $_);
            $hash_ref->{$key} = $value;
        }
        elsif ($section eq "units")
        {
            my ($scene_id, $product, $comment, $recipe) = split(/\|/, $_);

            my $unit_hash = {
                "sceneId" => $scene_id,
                "productName" => $product,
                "unitComment" => $comment,
                "recipeId" => $recipe
            };

            push(@{$units_list}, $unit_hash);
        }
        else
        {
            print "Error: invalid section header [$section] in $filename\n";
            $fh->close();
            return ERROR;
        }
    }
    $fh->close();

    $hash_ref->{"units"} = $units_list;

    return SUCCESS;
}
