#!/bin/csh -f

##############################################################################
#
# Purpose:
#    The purpose of this script is to dump L1R & L1G images into each band
#    specified file with an ENVI needed header file for each band to make them
#    viewable in ENVI and keep the maximum capabitities     
#
# INPUT ARGUMENTS: 
#       - Prefix to use for the output files
#       - Name of the input L1R or L1G file
#       - Band number to dump (optional - dumps all bands without it)
#
# RETURNS: 0 for successful completed and 1 for exit with error
#
# NOTE: Argument 3 is an optional argument, if it's present, it's a single
#       band number for single band processing purpose  
##############################################################################
#Make sure the correct number of arguments are provided
if ( ${#argv} < 2 || ${#argv} > 4) then
    echo "Usage: $0 <dump file prefix> <input l1r or l1g filename> \
          <optional band number>"
    exit 1
endif

# Make sure the input file exists
if (! -e $2) then
    echo "Input file $2 does not exist"
    exit 1    
endif

if ( ${#argv} == 3 ) then
    if ( $3 == "10" || $3 == "11" ) then
        set bands = "B"$3
    else
        set bands = "B0"$3
    endif
else
    # Get band list from the L1R or L1G image
    set bands = `h5ls $2 | grep "^B[01]" | awk '{print $1}'`
    if ( $status != 0 ) then
       echo "Error getting the $2 file content"
       exit 1    
    endif
endif

# Loop through bands
foreach band ( $bands )
    echo "Extracting band $band"
    # Get the SIMPLE dataspace 
    set extents = `h5dump -b LE -d /$band -o ${1}_$band $2 | grep SIMPLE`
    # For multiband processing, if there was an error it generally means the 
    # band wasn't present, so remove the empty file created and continue to 
    # the next band. For single band, just error out
    if ( $status != 0 ) then
        if ( ${#argv} == 3 ) then
            exit 1    
        else
            rm ${1}_$band
            continue
        endif
    endif

    # Get number of bands of the image, for single band selection, number 
    # of bands is set to be number of SCAs to enable multiple SCAs viewed
    # together in ENVI for one band
    set number_bands = \
        `echo $extents | awk -F',' '{print $1}' | awk '{print $5}'`
    # Get the numbers of lines and samples for the band
    set samples = `echo $extents | awk -F',' '{print $3}' | awk '{print $1}'`
    set lines = `echo $extents | awk -F',' '{print $2}' | awk '{print $1}'`

    # Get the DATATYPE dataspace 
    set extents = `h5dump -b LE -d /$band -o ${1}_$band $2 | grep DATATYPE`
    # For multiband processing, if there was an error it generally means the 
    # band wasn't present, so remove the empty file created and continue to 
    # the next band. For single band, just error out
    if ( $status != 0 ) then
        if ( ${#argv} == 3 ) then
            exit 1    
        else
            rm ${1}_$band
            continue
        endif
    endif

    # Get the HDF5 image band data type 
    set datatype = `echo $extents | awk -F',' '{print $1}' | awk '{print $2}'`

    # Convert the HDF5 data type to the corresponding envi data type number
    switch ($datatype)
        case H5T_IEEE_F32LE
             set envitype = 4
             breaksw
        case H5T_STD_I32LE
             set envitype = 3
             breaksw
        case H5T_STD_I16LE
             set envitype = 2
             breaksw
        case H5T_STD_U8LE
             set envitype = 1
             breaksw
        case H5T_STD_U16LE
             set envitype = 12
             breaksw
        default
             echo "Unsupported data type"
    endsw

    # Create the header for the band
    set header = ${1}_${band}.hdr
    echo "ENVI" > $header
    echo "description = {" >> $header
    echo "  File Imported into ENVI.}" >> $header
    echo "samples = $samples" >> $header
    echo "lines   = $lines" >> $header
    echo "bands   = $number_bands" >> $header
    echo "header offset = 0" >> $header
    echo "file type = ENVI Standard" >> $header
    echo "data type = $envitype" >> $header
    echo "interleave = bsq" >> $header
    echo "sensor type = Unknown" >> $header
    echo "byte order = 0" >> $header
    echo "wavelength units = Unknown" >> $header
    
    echo "Band $band has been extracted successfully"
end
exit 0
