#!/bin/csh -f

#############################################################################
# NAME:     get_procedure_parameters.csh
#
# PURPOSE:  Queries the 'sub_modules' and 'def_parms' tables to list the
#           scripts and associated parameter values for the specified
#           procedure.
#
# RETURNS:  Integer status of 0 or 1 (if command-line error)
#############################################################################

# Check the command line to ensure we have exactly one parameter -- the
# procedure name.
if ( ${#argv} != 1 ) then
    echo
    echo "USAGE:    $0 <procedure_name>"
    echo "EXAMPLE:  $0 "'"L1G Lunar Product"'
    echo
    exit 1
endif

# Set up and run the query.  The desired columns are as follows:
#
# N:           seq_nbr field in the 'sub_modules' table
# SUB_MOD_ID:  script ID field in the 'sub_modules' table
# PARM_NAME:   parm_name field in the 'def_parms' table
# OCCUR:       parm_occurrence field in the 'def_parms' table
# R:           required field in the 'def_parms' table
# VALUE:       parm_value field in the 'def_parms' table
#
# Print a description of the column headings.
echo
echo "                       KEY TO COLUMN HEADINGS"
echo
echo "N          -- database sequence number representing the execution "
echo "              order of a sub-module (processing script) within a "
echo "              module (procedure)."
echo "SUB_MOD_ID -- the name of a processing script."
echo "PARM_NAME  -- the name of a processing parameter in a given script."
echo "OCCUR      -- If the parameter is an array, its element number (i.e. "
echo "              1 for the first element, 2 for the second element, etc)."
echo "R          -- Flag indicating whether a parameter value is required."
echo "              0=parameter value is optional, 1=parameter value is "
echo "              required."
echo "VALUE      -- Default parameter value"
echo

sqlplus -s ${IAS_DB_L8_OLITIRS} << END
   set tab off
   set pagesize 50000
   set linesize 160
   set colsep ' '
   column SEQ_NBR    format a1
   column SUB_MOD_ID format a20
   column PARM_NAME  format a30
   column OCCUR      format a5
   column R          format a1
   column VALUE      format a35
   select cast(sm.seq_nbr as varchar(1)) as N, sm.sub_module_id as SUB_MOD_ID,
   dp.parm_name as PARM_NAME, cast(dp.parm_occurrence as varchar(5))as OCCUR,
   cast(dp.required as varchar(1)) as REQUIRED, dp.parm_value as VALUE
   from sub_modules sm, def_parms dp
   where sm.module_id=dp.module_id
   and dp.module_id='$1'
   and sm.sub_module_id=dp.sub_module_id
   order by sm.seq_nbr, dp.parm_name, dp.parm_occurrence;
   exit
END

exit 0
