/*==============================================================================
NAME:
    import_rlut2

PURPOSE:
    This application will read an ODL file that contains DN lookup and
    correction values for the alternative response linearization lookup table
    algorithm and write those values to an RLUT2 HDF5 file.

    The format of the ODL file is expected to match the ODL file that was
    supplied with the ADD in
    svn://edclxs131/l8/tags/calval/phase3/alt_linearization_3.0.  Processing
    this file with the IAS ODL library takes a VERY LONG time, approximately
    30-45 minutes.

RETURNS:
    Creates an HDF5 RLUT2 file
==============================================================================*/

/* System Includes */
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

/* IAS Includes */
#include "ias_logging.h"
#include "ias_odl.h"
#include "ias_rlut.h"
#include "ias_rlut2.h"
#include "ias_satellite_attributes.h"

/* Local Includes */


#define GROUP_NAME_LENGTH 256
#define ODL_ROOT_NAME     "ALTLIN_PARAMETERS"

/* Define the maximum DN value that the satellite can detect */
#define MAX_DN 16383    /* 2^14 - 1 = 16383 */

/* Global variables that allow the user to subset what they process */
int gOnlyProcessBandIndex = -1;
int gOnlyProcessScaIndex  = -1;


/*------------------------------------------------------------------------------
NAME: convert_band_to_odl_group

PURPOSE:
    Convert a band number to a ODL group name.

RETURNS:
    This function will return a status of SUCCESS if the band number is
    successfully converted to a name, ERROR otherwise.
------------------------------------------------------------------------------*/
int convert_band_to_odl_group (
    int band,                                 /* I: Band number */
    char nameLinearity[GROUP_NAME_LENGTH],    /* O: Linearity ODL group */
    char nameNonuniformity[GROUP_NAME_LENGTH] /* O: Nonuniformity ODL group */
    )
{
    if (band == 1)
    {
        strcpy(nameLinearity, "CA_LIN");      /* BAND 01 Coastal Aerosol */
        strcpy(nameNonuniformity, "CA_NUC");
    }
    else if (band == 2)
    {
        strcpy(nameLinearity, "B_LIN");       /* BAND 02 Blue */
        strcpy(nameNonuniformity, "B_NUC");
    }
    else if (band == 3)
    {
        strcpy(nameLinearity, "G_LIN");       /* BAND 03 Green */
        strcpy(nameNonuniformity, "G_NUC");
    }
    else if (band == 4)
    {
        strcpy(nameLinearity, "R_LIN");       /* BAND 04 Red */
        strcpy(nameNonuniformity, "R_NUC");
    }
    else if (band == 5)
    {
        strcpy(nameLinearity, "NIR_LIN");     /* BAND 05 NIR */
        strcpy(nameNonuniformity, "NIR_NUC");
    }
    else if (band == 6)
    {
        strcpy(nameLinearity, "1SW_LIN");     /* BAND 06 SWIR 1 */
        strcpy(nameNonuniformity, "1SW_NUC");
    }
    else if (band == 7)
    {
        strcpy(nameLinearity, "2SW_LIN");     /* BAND 07 SWIR 2 */
        strcpy(nameNonuniformity, "2SW_NUC");
    }
    else if (band == 8)
    {
        strcpy(nameLinearity, "P_LIN");       /* BAND 08 PAN */
        strcpy(nameNonuniformity, "P_NUC");
    }
    else if (band == 9)
    {
        strcpy(nameLinearity, "CRS_LIN");     /* BAND 09 Cirrus */
        strcpy(nameNonuniformity, "CRS_NUC");
    }
    else
    {
        IAS_LOG_WARNING("Band %d is not supported yet", band);
        return ERROR;
    }
    return SUCCESS;
}


/*------------------------------------------------------------------------------
NAME: print_example

PURPOSE:
    Print an example of the ODL file
------------------------------------------------------------------------------*/
void print_example ()
{
    char nameLinearity[GROUP_NAME_LENGTH];
    char nameNonuniformity[GROUP_NAME_LENGTH];
    int  band;
    int  i;

    printf("GROUP = %s\n", ODL_ROOT_NAME);
    for (i = 0; i < IAS_MAX_NBANDS; i++)
    {
        /* Convert this band index into a band number */
        band = ias_sat_attr_convert_band_index_to_number(i);
        if (band == ERROR)
        {
            IAS_LOG_ERROR("Converting band index %d to band number", i);
            break;
        }

        if (convert_band_to_odl_group(band, nameLinearity, nameNonuniformity)
            != SUCCESS)
        {
            IAS_LOG_ERROR("Converting band number to ODL group name");
            break;
        }

        printf("  GROUP = %s  /* BAND %02d %s */\n", nameLinearity, band,
            ias_sat_attr_convert_band_number_to_name(band));
        if (band == 1)
        {
            printf("    FPM00D000_INPUT = (0.000, 283.000, 566.000)\n");
            printf("    FPM00D000_VAL   = (0.000,   7.688,  14.779)\n");
            printf("    ...\n");
            printf("    FPM13D493_INPUT = (0.000, 282.000, 563.000)\n");
            printf("    FPM13D493_VAL   = (0.000,   5.226,  11.537)\n");
        }
        else
            printf("    ...\n");
        printf("  END_GROUP = %s\n", nameLinearity);
    }
    printf("END_GROUP = %s\n", ODL_ROOT_NAME);
    exit(0);
}


/*------------------------------------------------------------------------------
NAME: process_sca

PURPOSE:
    Process all of the detector values for a single SCA of a specified band.
    This function will examine all of the DN lookup and correction values to
    determine the maximum size of each of these arrays.  It will then allocate
    an RLUT2 structure to the maximum size and fill the structure with data for
    each detector.

    Once the RLUT2 structure has been filled with data, it will be written to
    the output RLUT2 HDF5 file that will be used by the RPS driver to perform
    detector response linearization corrections.

RETURNS:
    This function will return a status of SUCCESS if all of the detector are for
    the SCA are successfully processed and written to the output file.  It will
    return a status of ERROR if an error is encountered.
------------------------------------------------------------------------------*/
int process_sca (
    const char * objectName, /* I:  Name of the object within the ODL file */
    int band,                /* I:  Band number to process */
    int scaIndex,            /* I:  SCA index to process */
    int nDetectors,          /* I:  Number of detectors per SCA */
    int isLinearity,         /* I:  TRUE=linearity, FALSE=nonuniformity */
    IAS_OBJ_DESC * odl,      /* I:  Root object from ODL file */
    IAS_RLUT2_IO * fd        /* IO: File descriptor to RLUT2 HDF5 file */
    )
{
    IAS_RLUT2_TABLE * rlut2Table       = NULL;
    char              nameParm[256];
    float             values[MAX_DN+1];
    int               detector;
    int               i;
    int               inputValue;
    int               maxNValues       = 0;
    int               nValues;
    int               status;

    /*--------------------------------------------------------------------------
     * We are either working with Linearity (LIN) correction values or
     * Nonuniformity (NUC) correction values.  Within the specified ODL object
     * we need to look at every parameter that equates to a single detector for
     * this SCA.  As we walk through both the DN lookup sequences and the
     * correction sequences we need to find the length of the sequence with the
     * most entries.
     * 
     * Once we locate the longest sequence, we'll allocate an IAS_RLUT2_TABLE
     * object and then load all of these values into 2-dimensional arrays of
     * floating point values.  These values will then be written to the output
     * RLUT2 HDF5 file.
     * 
     * The array of floating point values is sized so it can contain each of the
     * MAX_DN (16384) possible values that the satellite can detect.  We are
     * expecting a very small percentage of these values to appear in the look
     * up values but our array is sized so that we could handle all distinct
     * values if they did show up.
     *------------------------------------------------------------------------*/

    /* Find the longest sequence */
    for (detector = 0; detector < nDetectors; detector++)
    {
        for (inputValue = 0; inputValue < 2; inputValue++)
        {
            /* Create a parameter name for the current SCA (some call it a Focal
               Plane Module (FPM)) and detector.  We'll need to examine both the
               DN lookup (e.g. INPUT) and the correction value (e.g. VAL). */
            sprintf(nameParm, "FPM%02dD%03d_%s", scaIndex, detector,
                ((inputValue == 0) ? "INPUT" : "VAL"));

            /* Read the values for this parameter into an array */
            status = ias_odl_get_field(values, sizeof(values),
                IAS_ODL_Float, odl, objectName, nameParm, &nValues);
            if (status != SUCCESS)
            {
                char errString[256];

                if (status == IAS_ODL_NOT_ENOUGH_MEMORY_SUPPLIED)
                    strcpy(errString, "not enough memory available");
                else if (status == IAS_ODL_NOT_FOUND)
                    strcpy(errString, "requested object not found");
                else if (status == IAS_ODL_INVALID_DATA_TYPE)
                    strcpy(errString, "data type mismatch");
                else
                    strcpy(errString, "unknown ODL error encountered");

                IAS_LOG_ERROR("Retrieving '%s' from '%s'; %s", nameParm, objectName,
                    errString);
                return ERROR;
            }

            /* Keep track of the maximum number of values that we encounter
               through all of these parameters.  We'll use this to size the
               IAS_RLUT2_TABLE appropriately. */
            if (nValues > maxNValues)
                maxNValues = nValues;
        }
    }

    /* Allocate the RLUT2 table for this band and SCA */
    status = ias_rlut2_realloc_table(band, scaIndex+1, maxNValues, &rlut2Table);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Allocating RLUT2 table for band %d, SCA %d, object %s "
            "with %d entries", band, scaIndex+1, objectName, maxNValues);
        return ERROR;
    }

    /* Read the values and pack them into the RLUT2 table */
    for (detector = 0; detector < nDetectors; detector++)
    {
        for (inputValue = 0; inputValue < 2; inputValue++)
        {
            /* Create a parameter name for the current SCA (some call it a Focal
               Plane Module (FPM)) and detector.  We'll need to examine both the
               DN lookup (e.g. INPUT) and the correction value (e.g. VAL). */
            sprintf(nameParm, "FPM%02dD%03d_%s", scaIndex, detector,
                ((inputValue == 0) ? "INPUT" : "VAL"));

            /* Read the values for this parameter into an array */
            status = ias_odl_get_field(values, sizeof(values),
                IAS_ODL_Float, odl, objectName, nameParm, &nValues);
            if (status != SUCCESS)
            {
                char errString[256];

                if (status == IAS_ODL_NOT_ENOUGH_MEMORY_SUPPLIED)
                    strcpy(errString, "not enough memory available");
                else if (status == IAS_ODL_NOT_FOUND)
                    strcpy(errString, "requested object not found");
                else if (status == IAS_ODL_INVALID_DATA_TYPE)
                    strcpy(errString, "data type mismatch");
                else
                    strcpy(errString, "unknown ODL error encountered");

                IAS_LOG_ERROR("Retrieving '%s' from '%s'; %s", nameParm, objectName,
                    errString);
                ias_rlut2_free_table(rlut2Table);
                rlut2Table = NULL;
                return ERROR;
            }

            /* Copy the array of values into the table */
            if (inputValue == 0)
            {
                int offset = detector * maxNValues;

                /* The DN lookup table must begin with a 0.0 value and all of
                   the remaining values in a detector's row must be set to the
                   maximum DN value that the satellite can report.  This is done
                   to ensure all lookups will succeed and all different search
                   algorithms to work properly. */
                for (i = 0; i < nValues; i++)
                    rlut2Table->input_dn[offset + i] = values[i];
                while (i < maxNValues)
                    rlut2Table->input_dn[offset + i++] = (float)MAX_DN;
            }
            else
            {
                int offset = detector * maxNValues;

                /* Any additional values in the correction array are simply zero
                   filled */
                for (i = 0; i < nValues; i++)
                    rlut2Table->output_correction[offset + i] = values[i];
                while (i < maxNValues)
                    rlut2Table->output_correction[offset + i++] = 0.0;
            }
        }
    }

    /* Write the filled out table to the RLUT2 HDF5 file and free the memory
       that was dynamically allocated for the table. */
    status = ias_rlut2_write(fd, isLinearity, rlut2Table);
    ias_rlut2_free_table(rlut2Table);
    rlut2Table = NULL;
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Processing band %d, SCA %d, %s", band, scaIndex+1,
            objectName);
        return ERROR;
    }

    IAS_LOG_INFO("Processed band %d, SCA %2d, %s: nValues %d", band, scaIndex+1,
        objectName, maxNValues);
    return SUCCESS;
}


/*------------------------------------------------------------------------------
NAME: process_band

PURPOSE:
    Process one band of data from the input ODL file and write it to the HDF5
    RLUT2 file.

RETURNS:
    This function returns a status of SUCCESS if the band is successfully
    converted, ERROR if an error is encountered.
------------------------------------------------------------------------------*/
int process_band (
    int band,                     /* I:  Band number to process */
    IAS_OBJ_DESC * odlRootObject, /* I:  Root object from ODL file */
    IAS_RLUT2_IO * fd             /* IO: File descriptor to RLUT2 HDF5 file */
    )
{
    IAS_OBJ_DESC * odlBandObj;
    char           objectNameLin[GROUP_NAME_LENGTH];
    char           objectNameNuc[GROUP_NAME_LENGTH];
    int            nDetectors;
    int            nScas;
    int            sca;
    int            status;

    nDetectors = ias_sat_attr_get_detectors_per_sca(band);
    if (nDetectors == ERROR)
    {
        IAS_LOG_ERROR("Obtaining number of detectors/SCA for band %d", band);
        return ERROR;
    }
    nScas = ias_sat_attr_get_scas_per_band(band);
    if (nScas == ERROR)
    {
        IAS_LOG_ERROR("Obtaining number of SCAs/band for band %d", band);
        return ERROR;
    }

    if (convert_band_to_odl_group(band, objectNameLin, objectNameNuc) !=SUCCESS)
    {
        IAS_LOG_ERROR("Converting band number to ODL group name");
        return ERROR;
    }

    /* Parse out the band's Linearity correction object */
    odlBandObj = ias_odl_get_group(odlRootObject, objectNameLin);
    if (odlBandObj == NULL)
    {
        IAS_LOG_ERROR("Band group/object '%s' not found", objectNameLin);
        return ERROR;
    }

    /* Process the Linearity correction object for this band's SCAs */
    for (sca = 0; sca < nScas; sca++)
    {
        /* If the user specified a specific SCA to process, only process that
           one SCA */
        if ((gOnlyProcessScaIndex >= 0) && (sca != gOnlyProcessScaIndex))
            continue;

        status = process_sca(objectNameLin, band, sca, nDetectors, TRUE,
            odlBandObj, fd);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Processing Band %02d SCA %02d", band, sca+1);
            return ERROR;
        }
    }

    /* Parse out the band's Nonuniformity correction object */
    odlBandObj = ias_odl_get_group(odlRootObject, objectNameNuc);
    if (odlBandObj == NULL)
    {
        IAS_LOG_ERROR("Band group/object '%s' not found", objectNameNuc);
        return ERROR;
    }

    /* Process the Nonuniformity correction object for this band's SCAs */
    for (sca = 0; sca < nScas; sca++)
    {
        /* If the user specified a specific SCA to process, only process that
           one SCA */
        if ((gOnlyProcessScaIndex >= 0) && (sca != gOnlyProcessScaIndex))
            continue;

        status = process_sca(objectNameNuc, band, sca, nDetectors, FALSE,
            odlBandObj, fd);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Processing Band %02d SCA %02d", band, sca+1);
            return ERROR;
        }
    }

    return SUCCESS;
}


/*------------------------------------------------------------------------------
NAME: usage

PURPOSE:
    Print a usage message to the user
------------------------------------------------------------------------------*/
void usage ()
{
    printf("\n");
    printf("usage: import_rlut2 [-b <band>] [-l <sat>] [-s <sca>] [-x] -i <file> -o <file>\n");
    printf("\n");
    printf("    This application will import an ODL file that contains DN lookup/linearity  \n");
    printf("    pairs and DN lookup/nonuniformity pairs that are used in the alternative    \n");
    printf("    response linearization lookup table (RLUT2) algorithm to correct the non-   \n");
    printf("    linear response of the Landsat 8 detectors.                                 \n");
    printf("                                                                                \n");
    printf("    -b    Only process the specified band number                                \n");
    printf("    -i    Specify the input ODL file that contains the RLUT values              \n");
    printf("    -l    Specify a satellite number, default is 8.  Position within the        \n");
    printf("          argument list is important.  This option should be specified prior to \n");
    printf("          the -b and -s options.                                                \n");
    printf("    -o    Specify the output RLUT2 HDF5 file name                               \n");
    printf("    -s    Only process the specified SCA number for each band.  If combined with\n");
    printf("          the -b option, the user may limit the processing to a single SCA for a\n");
    printf("          specific band.                                                        \n");
    printf("    -x    Print an example ODL file format                                      \n");
    printf("\n");
    exit(EXIT_FAILURE);
}


int main (int argc, char * argv[])
{
    IAS_OBJ_DESC * odlFile;
    IAS_OBJ_DESC * odlRootObject;
    IAS_RLUT2_IO * fd;
    char           pathIn[PATH_MAX]  = "";
    char           pathOut[PATH_MAX] = "";
    int            band;
    int            i;
    int            status            = SUCCESS;
    int            tmp;
    time_t         timeStart;
    IAS_RLUT_FILE_ATTRIBUTES file_attributes;

    /* Initialize the satellite attributes library to our default satellite.  If
       the user wishes to use another satellite, the library will be
       reinitialized when that command line argument is processed. */
    if (ias_sat_attr_initialize(IAS_L8) != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing the satellite attributes library");
        return EXIT_FAILURE;
    }

    for (i = 1; i < argc; i++)
    {
        if ((strcmp(argv[i], "-b") == 0) && ((i+1) < argc) &&
            (sscanf(argv[i+1], "%d", &tmp) == 1))
        {
            gOnlyProcessBandIndex =
                ias_sat_attr_convert_band_number_to_index(tmp);
            if (gOnlyProcessBandIndex == ERROR)
            {
                IAS_LOG_ERROR("Invalid band number specified: %d", tmp);
                return EXIT_FAILURE;
            }
            i++;
        }
        else if (((strcmp(argv[i], "-i") == 0) ||
                  (strcmp(argv[i], "-o") == 0)) &&
                 ((i+1) < argc))
        {
            if (strlen(argv[++i]) >= PATH_MAX)
            {
                IAS_LOG_ERROR("File path too long: %s", argv[i]);
                return EXIT_FAILURE;
            }
            if (strcmp(argv[i-1], "-i") == 0)
                strcpy(pathIn, argv[i]);
            else
                strcpy(pathOut, argv[i]);
        }
        else if ((strcmp(argv[i], "-l") == 0) && ((i+1) < argc) &&
            (sscanf(argv[i+1], "%d", &tmp) == 1))
        {
            IAS_SATELLITE_ID satId;

            satId = ias_sat_attr_get_satellite_id_from_satellite_number(tmp);
            if (satId == ERROR)
            {
                IAS_LOG_ERROR("Invalid satellite number: %d", tmp);
                return EXIT_FAILURE;
            }
            i++;

            if (ias_sat_attr_initialize(satId) != SUCCESS)
            {
                IAS_LOG_ERROR("Initializing the satellite attributes library");
                return EXIT_FAILURE;
            }
        }
        else if ((strcmp(argv[i], "-s") == 0) && ((i+1) < argc) &&
                 (sscanf(argv[i+1], "%d", &tmp) == 1))
        {
            if ((tmp < 1) || (tmp > IAS_MAX_NSCAS))
            {
                IAS_LOG_ERROR("Invalid SCA index specified: %d", tmp);
                return EXIT_FAILURE;
            }
            gOnlyProcessScaIndex = tmp - 1;
            i++;
        }
        else if (strcmp(argv[i], "-x") == 0)
            print_example();
        else
            usage();
    }

    /* The user must specify the -f option for this application to run */
    if ((strlen(pathIn) == 0) || (strlen(pathOut) == 0))
        usage();
    if (access(pathIn, R_OK))
    {
        IAS_LOG_ERROR("Access denied: %s", pathIn);
        return EXIT_FAILURE;
    }

    /* Open and parse the ODL file */
    IAS_LOG_INFO("Parsing ODL file.  This may take several minutes...");
    timeStart = time(NULL);
    odlFile = ias_odl_read_tree(pathIn);
    if (odlFile == NULL)
    {
        IAS_LOG_ERROR("Reading ODL file: %s", pathIn);
        return EXIT_FAILURE;
    }
    tmp = (int)(time(NULL) - timeStart);
    IAS_LOG_INFO("Parsed ODL file in %d seconds", tmp);

    /* Parse out the expected root level object */
    odlRootObject = ias_odl_get_group(odlFile, ODL_ROOT_NAME);
    if (odlRootObject == NULL)
    {
        IAS_LOG_ERROR("Root object '%s' not found", ODL_ROOT_NAME);
        ias_odl_free_tree(odlFile);
        return EXIT_FAILURE;
    }

    /* Create the RLUT2 HDF5 file */
    fd = ias_rlut2_open_write(pathOut);
    if (fd == NULL)
    {
        IAS_LOG_ERROR("Creating RLUT2 HDF5 file: %s", pathOut);
        ias_odl_free_tree(odlFile);
        return EXIT_FAILURE;
    }

    /* Set up the file attributes in the RLUT */
    memset(&file_attributes, 0, sizeof(file_attributes));
    strcpy(file_attributes.file_source, "CalVal ODL file");
    strcpy(file_attributes.effective_date_begin,"2012-01-04T00:00:00");
    strcpy(file_attributes.effective_date_end,"2012-01-04T00:00:00");
    strcpy(file_attributes.status,"ACTIVE");
    strcpy(file_attributes.baseline_date,"2012-01-04T00:00:00");
    strcpy(file_attributes.description, "Imported from ODL");
    file_attributes.version = 1;

    if (ias_rlut2_write_file_attributes(fd, &file_attributes) != SUCCESS)
    {
        IAS_LOG_ERROR("Writing file attributes to: %s", pathOut);
        ias_odl_free_tree(odlFile);
        ias_rlut2_close_file(fd);
        return EXIT_FAILURE;
    } 

    /* Process all of the known bands for this satellite */
    for (i = 0; i < IAS_MAX_NBANDS; i++)
    {
        /* If the user specified a specific band to process, only process that
           one band */
        if ((gOnlyProcessBandIndex >= 0) && (i != gOnlyProcessBandIndex))
            continue;

        /* Convert this band index into a band number */
        band = ias_sat_attr_convert_band_index_to_number(i);
        if (band == ERROR)
        {
            IAS_LOG_ERROR("Converting band index %d to band number", i);
            break;
        }

        /* Convert one band of information from the ODL file into the RLUT2 file
           */
        status = process_band(band, odlRootObject, fd);
        if (band >= 10)
        {
            /* TODO: Remove this if statement once TIRS data is supplied */
            IAS_LOG_WARNING("TODO: Remove this once TIRS data is supplied.  "
                "Band %d not processed", band);
            status = SUCCESS;
            continue;
        }
        else if (status != SUCCESS)
            break;
    }

    /* Close the output file */
    if (ias_rlut2_close_file(fd) != SUCCESS)
        IAS_LOG_ERROR("Closing RLUT2 HDF5 file: %s", pathOut);

    /* Free the memory allocated for the ODL file and root object */
    ias_odl_free_tree(odlFile);

    if (status != SUCCESS)
        return EXIT_FAILURE;

    tmp = (int)(time(NULL) - timeStart);
    IAS_LOG_INFO("Total runtime: %d seconds", tmp);
    return 0;
}
