#!/usr/bin/perl -w
#
# Purpose: Selects L0 scene files to call with the ias_create_work_order module.
#
#
# Process:  1. Open input file containing scene names/IDs
#           2. For each scene:
#              a. Call ias_create_work_order 
#              b. If an output file was specified, write out the work order
#                 id returned by ias_create_work_order.
#
#  Inputs:  1. The name of the input file
#           2. The name of the work order id output file.
#
# Outputs:  1. File containing work order id's. (If specified)
#
# System Requirements:
#           1. IAS system installed.
#
# Input File Format:
#           1. Text file with a scene ID/Name, product id, work order
#              comment, L0R data file, and script parameters on each line
#              (one line per work order) separated by commas.
#              - The work order comment, L0R data file, and script
#                parameters are optional. 
#              - Commas for each field are required. (see example)
#              - The script parameters are to be specified in this format:
#                SCRIPT1_NAME:PARM1_NAME=value SCRIPT2_NAME:PARM1_NAME=value
#                * script and parameter names use all uppercase
#                * spacing is the delimiter between parameter definitions
#                * define common/global parameters through the use of 'script'
#                  name "COMMON"
#                * file names must contain the fully qualified path and file
#                  name
#                * specifying a parameter name with absence of a 'value' has
#                  the effect of setting the parameter value to NULL in DB
#                * use '_ELEM#' (1-based) appended to parameter name to set
#                  the value of a specific element for list parameters
#                * lists can be given for a parameter with (x,y,z) format
#                   > either all list elements must be specified or
#                   > a subset of list elements can be specified as the first
#                     'n' elements of the list; if all possible list values are
#                     not given, tailing elements of the list not specified are
#                     considered to have been left blank (set to NULL in DB);
#                     e.g. (10,11) given for a BAND_LIST parameter has the
#                     effect of setting band list element 1 being '10',
#                     element 2 being '11', and all remaining 9 elements of
#                     the band list not specified are set to NULL in DB
#                * the number of elements for a given parameter can vary in
#                  definition across procedures and cannot be increased by
#                  providing more elements than defined for each procedure
#                * invalid parameter definitions cause fatal error
#
#           2. Examples: (lines should NOT start with spaces)
#
# LC80772082000150EDC00,L1T Product,Test Case 036 (Earth),L0Rfile
# LC80400392001213EDC00,L1T Product,Test Case 036 (Earth),L0Rfile
# LC81012421999300SGS00,L1T Product,Test Case 036,L0Rfile
# LC80400381999272EDC00,L1T Product,Test Case 036,L0Rfile
# LO80920832001001EDC00,L1T Product
# LT81510302000189SGS00,LIT Product,,L0Rfile
# LT81510302000189SGS00,LIT Product,,L0Rfile,DGC:GCP_TYPE=GLS L8RPS:GENERATE_REPORT=1 L8RPS:BAND_LIST_ELEM11=
# LT81510302000189SGS00,LIT Product,,,DGC:GCP_TYPE=GLS L8RPS:BAND_LIST=(1,2,3,4)
# LO80920832001001EDC00,L1T Product,,,COMMON:STORE_CHARACTERIZATION_IN_DB=0
# LO80920832001001EDC00,L1T Product,,,L8RPS:ALGORITHMS_ELEM16=reflectance_conversion_SKIP
#
# Documentation: No other documentation exists for this script
#

use strict;
use warnings;

use Logger;
use ias_create_work_order qw
{
    ias_create_work_order
};
use ias_const;

use constant PROG     => "load_work_orders.pl";

my $g_Logger;         # The logger object to use for log messages

# initialize logger
$g_Logger = Logger->new("STDOUT", PROG);
if (!$g_Logger)
{
    print STDERR PROG . ": ERROR: Failed to establish logger object.\n";

    exit FAILURE;
}
set_default_log($g_Logger);

# Get input and output file names from command line argument(s)
my $numArgs = $#ARGV + 1;
if ($numArgs < 1)
{
    $g_Logger->fatal_msg("Please provide an input file name.");
    exit FAILURE;
}

my $inputFileName = $ARGV[0];
my $outputFileName = $ARGV[1];

# Open input file and populate an array with scene names
$g_Logger->info_msg("Opening input file:  $inputFileName");
open FILE, "<", $inputFileName or die;
my @sceneIds = <FILE>;
close FILE;

# Open output file to record work order id's
if ($outputFileName)
{
    open FILE, ">", $outputFileName or die $!;
}
if (defined $outputFileName)
{
    $g_Logger->info_msg("Output file: $outputFileName");
}

# Grab the test data directory path from the environment
my $test_data_path = $ENV{IAS_IT_TEST_DATA_DIR};

# Call ias_create_work_order for each scene
foreach (@sceneIds)
{
    # declare locally to help prevent any values set for a prior work order
    # from being carried over in the creation of a subsequent work order
    my $sceneId = "";
    my $procedureId = "";
    my $woComment = undef;
    my $l0r_data_path = undef;
    my $script_params = '';
    my @script_params = ();
    my %args;                  # args to pass to ias_create_work_order
    my $wo_id;                 # work order id returned by ias_create_work_order

    # remove trailing newline character from file line
    chomp($_);

    # split file line into individual fields
    ($sceneId, $procedureId, $woComment, $l0r_data_path, $script_params)
        = split( ',', $_, 5 );

    if (!defined $procedureId) # procedure id missing
    {
        $procedureId = '';
    }
    if (!defined $l0r_data_path) # no data path provided
    {
        $l0r_data_path = '';
    }
    else
    {
        # fix the L0R data path, by replacing the IAS_IT_TEST_DATA_DIR with the
        # value from the environment
        $l0r_data_path =~ s#IAS_IT_TEST_DATA_DIR#$test_data_path#;
        $args{'l0r_data_path'} = $l0r_data_path;
    }

    $args{'wo_comment'} = $woComment;

    # splitting on the space delimiting from the input file format, put the
    # parameters into a list - this implementation obviously will break if
    # any of the script or parameter name or the parameter value contain a
    # space, which is not known to be a valid possibility at this time so is
    # not supported
    if (defined($script_params))
    {
        @script_params = split /\s/, $script_params;
        $args{'parms'} = \@script_params;
    }

    $wo_id = ias_create_work_order($procedureId, $sceneId, %args);

    if (!defined($wo_id))
    {
        $g_Logger->fatal_msg("Error: Loading scene: $sceneId");
    }
    else
    {
        # write work order id to file
        if ($outputFileName)
        {
            print FILE "$wo_id\n";
        }
    }
}

if ($outputFileName)
{
    close FILE;
}

$g_Logger->info_msg("Script complete");
