#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ias_pixel_mask.h"
#include "ias_const.h"
#include "ias_logging.h"
#include "ias_radiometry.h"

void print_mask(void)
{
    printf("The RPS pixel masks are defined as:\n\n");
    printf("    IAS_RPS_NOVALUE             0x%02x\n"
           "    IAS_RPS_DROPLINE            0x%02x\n"
           "    IAS_RPS_OUTOFSPEC_DETECTOR  0x%02x\n"
           "    IAS_RPS_IMPNOISE            0x%02x\n"
           "    IAS_RPS_ADCONV_SAT_LO       0x%02x\n"
           "    IAS_RPS_ADCONV_SAT_HI       0x%02x\n"
           "    IAS_RPS_ANALOG_SAT_LO       0x%02x\n"
           "    IAS_RPS_ANALOG_SAT_HI       0x%02x\n"
           "    IAS_RPS_INOP_DETECTOR       0x%02x\n\n",
        IAS_RPS_NOVALUE, IAS_RPS_DROPLINE, IAS_RPS_OUTOFSPEC_DETECTOR,
        IAS_RPS_IMPNOISE, IAS_RPS_ADCONV_SAT_LO, IAS_RPS_ADCONV_SAT_HI,
        IAS_RPS_ANALOG_SAT_LO, IAS_RPS_ANALOG_SAT_HI,
        IAS_RPS_INOP_DETECTOR);
}

int main( int argc, char *argv[] )
{
    PIXEL_MASK_TYPE arr[8] = {IAS_RPS_INOP_DETECTOR, IAS_RPS_ANALOG_SAT_HI,
        IAS_RPS_ANALOG_SAT_LO, IAS_RPS_ADCONV_SAT_HI, IAS_RPS_ADCONV_SAT_LO,
        IAS_RPS_IMPNOISE, IAS_RPS_OUTOFSPEC_DETECTOR, IAS_RPS_DROPLINE};
    PIXEL_MASK_TYPE mask;
    IAS_PIXEL_MASK_IO *pm_file = NULL;
    IAS_PIXEL_MASK_BAND_SCA_LIST *pm_band_sca_list = NULL;
    IAS_PIXEL_MASK_ITERATOR *it;
    IAS_PIXEL_MASK_SPAN span;

    const char *pixel_mask_filename = NULL;
    int i;
    int count;
    int band_number, sca_number;
    int status;

    if ( argc != 5 )
    {
        printf("\nUsage: %s filename band sca mask_value\nWhere mask_value "
            "value is in:\n "
            "    IAS_RPS_DROPLINE, IAS_RPS_OUTOFSPEC_DETECTOR, "
            "IAS_RPS_IMPNOISE ... \n\n",
            argv[0] );
        printf("Passing 0 for the band argument will give a list of Band/SCA "
            "combinations that are available in the pixel mask\n\n");
        print_mask();
        return EXIT_FAILURE;
    }
    /* Initialize the logging library in case any library call issues a
    log message */
    ias_log_initialize( "read_mask_value" );

    status = sscanf( argv[2], "%d", &band_number );
    if (status != 1)
    {
        printf("Error reading band number from command line parameters\n");
        return EXIT_FAILURE;
    }
    status = sscanf( argv[3], "%d", &sca_number );
    if (status != 1)
    {
        printf("Error reading SCA number from command line parameters\n");
        return EXIT_FAILURE;
    }

    /* Set up the mask value to be checked. */
    if ( strcmp( argv[4], "IAS_RPS_NOVALUE" ) == 0 )
        mask = IAS_RPS_NOVALUE;
    else if ( strcmp( argv[4], "IAS_RPS_DROPLINE" ) == 0 )
        mask = IAS_RPS_DROPLINE;
    else if ( strcmp( argv[4], "IAS_RPS_OUTOFSPEC_DETECTOR" ) == 0 )
        mask = IAS_RPS_OUTOFSPEC_DETECTOR;
    else if ( strcmp( argv[4], "IAS_RPS_IMPNOISE" ) == 0 )
        mask = IAS_RPS_IMPNOISE;
    else if ( strcmp( argv[4], "IAS_RPS_ADCONV_SAT_LO" ) == 0 )
        mask = IAS_RPS_ADCONV_SAT_LO;
    else if ( strcmp( argv[4], "IAS_RPS_ADCONV_SAT_HI" ) == 0 )
        mask = IAS_RPS_ADCONV_SAT_HI;
    else if ( strcmp( argv[4], "IAS_RPS_ANALOG_SAT_LO" ) == 0 )
        mask = IAS_RPS_ANALOG_SAT_LO;
    else if ( strcmp( argv[4], "IAS_RPS_ANALOG_SAT_HI" ) == 0 )
        mask = IAS_RPS_ANALOG_SAT_HI;
    else if ( strcmp( argv[4], "IAS_RPS_INOP_DETECTOR" ) == 0 )
        mask = IAS_RPS_INOP_DETECTOR;
    else
    {
        printf("%s isn't a valid mask type\n", argv[4] );
        print_mask();
        return EXIT_FAILURE;
    }

    /* Open the pixel mask file. */
    pixel_mask_filename = argv[1];
    pm_file = ias_pm_open_pixel_mask(pixel_mask_filename, IAS_READ);
    if (pm_file == NULL)
    {
        printf("Error opening pixel mask file %s\n", pixel_mask_filename);
        return EXIT_FAILURE;
    }

    /* Get the list of bands/SCAs present in the file. */
    status = ias_pm_get_file_band_sca_list(pm_file, &pm_band_sca_list,
        &count);
    if (status != SUCCESS)
    {
        printf("Error etrieving band/SCA list from pixel mask file "
            "%s", pixel_mask_filename);
        ias_pm_close_pixel_mask(pm_file);
        return EXIT_FAILURE;
    }

    /* If the band passed was 0 print out the contents of the pixel mask
       and exit */
    if ( band_number == 0 )
    {
        printf("Read %d masks from file, they are:\n", count );
        for ( i = 0; i < count; i++ )
        {
            printf("index %5d  B%02d  S%02d\n", i,
                pm_band_sca_list[i].band_number,
                pm_band_sca_list[i].sca_number);
        }
    }
    /* Otherwise search for the band and SCA requested */
    else
    {
        IAS_PIXEL_MASK *pm = NULL;
        int index = -1;

        for ( i = 0; i < count; i++ )
        {
            if ((pm_band_sca_list[i].band_number == band_number)
                && (pm_band_sca_list[i].sca_number == sca_number))
            {
                index = i;
                break;
            }
        }

        /* If the band/SCA combination wasn't found exit */
        if ( index < 0 )
        {
            printf("Couldn't find band %d SCA %d in the pixel mask\n", 
                band_number, sca_number );
            ias_pm_close_pixel_mask(pm_file);
            free(pm_band_sca_list);
            return EXIT_FAILURE;
        }

        /* Print out the current set of values */
        print_mask();

        /* Get the pixel mask for the specified band/SCA number. */
        pm = ias_pm_read_single_mask_from_file(pm_file, band_number,
            sca_number);
        if (pm == NULL)
        {
            printf("Error reading band %d SCA %d pixel mask from file",
                band_number, sca_number);
            ias_pm_close_pixel_mask(pm_file);
            free(pm_band_sca_list);
            return EXIT_FAILURE;
        }

        /* Get an iterator for the mask specified on the command line */
        it = ias_pm_get_iterator( pm, IAS_PM_INCLUDE, mask );
        if (it == NULL)
        {
            printf("Error getting a pixel mask iterator");
            free(pm_band_sca_list);
            ias_pm_close_pixel_mask(pm_file);
            return EXIT_FAILURE;
        }

        /* Loop through all the spans and print stuff */
        printf("Band %d  SCA %d\n", band_number, sca_number );
        printf("Pixel mask index: %d\n", index);
        printf("Spans for mask value: %s\n", argv[4] );
        printf(" Det   Start     Len Mask  Bit field\n");
        while ( ias_pm_get_next_span( it, &span ))
        {
            int c;
            /* Print the span information */
            printf( " %3d  %6d  %6d  %d  ", span.detector_index,
                span.starting_pixel_index, span.length_of_span,
                span.pixel_mask );

            /* Print a bit representation of the mask */
            for ( c = 0; c < 8; c++ )
            {
                if ( span.pixel_mask & arr[c] )
                    printf("1");
                else
                    printf("0");
            }
            printf("\n");
        }

        /* Clean up the pixel mask resources */
        ias_pm_destroy_iterator( it );
        ias_pm_destroy(pm);
    }

    /* Residual clean up */
    ias_pm_close_pixel_mask(pm_file);
    free(pm_band_sca_list);

    return EXIT_SUCCESS;
}  /* END main */
