/**************************************************************************
 NAME:     write_mask_image

 PURPOSE:  Utility that reads a pixel mask file and writes an L1R format file 
           for the pixel mask data for bands and/or SCAs specified by the
           user.

 RETURNS:  system status EXIT_SUCCESS if successful
           system status EXIT_FAILURE on error

 NOTES:    This utility creates an mask pixel mask image for all bands/SCAs 
           within the pixel mask file by default.  It can also create this
           file for any available SCA(s) in a single band selected by the
           user, as well as an individual band/SCA combination selected by
           the user.
**************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "ias_const.h"
#include "ias_logging.h"
#include "ias_pixel_mask.h"
#include "ias_radiometry.h"
#include "ias_satellite_attributes.h"

#include "write_mask_image.h"



/**************************************************************************
 NAME:     write_mask_legend_file

 PURPOSE:  Creates a file describing the various pixel mask values, and how
           they can be combined

 RETURNS:  SUCCESS if successful
           ERROR on error
**************************************************************************/
static int write_mask_legend_file
(
    const char *mask_image_filename      /* I: Name of mask image file */
)
{
    char *file_ext = NULL;
    char mask_image_basename[PATH_MAX];
    char legend_file_name[PATH_MAX];
    FILE *fptr;
    int status;

    /* Strip off the image file base name and append '_image_legend.txt' to
       it */
    strncpy(mask_image_basename, mask_image_filename,
        sizeof(mask_image_basename));
    mask_image_basename[sizeof(mask_image_basename) - 1] = '\0';
    file_ext = strrchr(mask_image_basename, '.');
    if (file_ext != NULL)
    {
        *file_ext = '\0';
    }

    status = snprintf(legend_file_name, sizeof(legend_file_name),
        "%s_legend.txt", mask_image_basename);
    if ((status < 0) || (status >= sizeof(legend_file_name)))
    {
        IAS_LOG_ERROR("Constructing legend file name");
        return ERROR;
    }

    /* Create the legend file */
    fptr = fopen(legend_file_name, "w");
    if (fptr == NULL)
    {
        IAS_LOG_ERROR("Creating legend file %s", legend_file_name);
        return ERROR;
    }

    status = fprintf(fptr, "The RPS pixel masks are currently defined as:\n\n"
        "    IAS_RPS_NOVALUE               0x%02x ('%d')\n"
        "    IAS_RPS_DROPLINE              0x%02x ('%d')\n"
        "    IAS_RPS_OUTOFSPEC_DETECTOR    0x%02x ('%d')\n"
        "    IAS_RPS_IMPNOISE              0x%02x ('%d')\n"
        "    IAS_RPS_ADCONV_SAT_LO         0x%02x ('%d')\n"
        "    IAS_RPS_ADCONV_SAT_HI         0x%02x ('%d')\n"
        "    IAS_RPS_ANALOG_SAT_LO         0x%02x ('%d')\n"
        "    IAS_RPS_ANALOG_SAT_HI         0x%02x ('%d')\n"
        "    IAS_RPS_INOP_DETECTOR         0x%02x ('%d')\n\n\n",
        IAS_RPS_NOVALUE, IAS_RPS_NOVALUE,
        IAS_RPS_DROPLINE, IAS_RPS_DROPLINE,
        IAS_RPS_OUTOFSPEC_DETECTOR, IAS_RPS_OUTOFSPEC_DETECTOR,
        IAS_RPS_IMPNOISE, IAS_RPS_IMPNOISE,
        IAS_RPS_ADCONV_SAT_LO, IAS_RPS_ADCONV_SAT_LO,
        IAS_RPS_ADCONV_SAT_HI, IAS_RPS_ADCONV_SAT_HI,
        IAS_RPS_ANALOG_SAT_LO, IAS_RPS_ANALOG_SAT_LO,
        IAS_RPS_ANALOG_SAT_HI, IAS_RPS_ANALOG_SAT_HI,
        IAS_RPS_INOP_DETECTOR, IAS_RPS_INOP_DETECTOR);
    if (status < 0)
    {
        IAS_LOG_ERROR("Error printing specific pixel mask values to file %s",
            legend_file_name);
        fclose(fptr);
        remove(legend_file_name);
        return ERROR;
    }

    status = fprintf(fptr, "For a specific pixel location, multiple "
        "artifacts can be flagged; the\noverall mask value is then the "
        "sum of the individual mask values.  For\nexample, the mask value "
        "'%d' represents flagging of an inoperable\ndetector AND a dropped "
        "line.\n\n"
        "If all artifacts have been flagged, the mask value is '%d'.\n"
        "If no artifacts have been flagged, the mask value is '0'.\n",
        IAS_RPS_DROPLINE + IAS_RPS_INOP_DETECTOR,
        IAS_RPS_DROPLINE + IAS_RPS_OUTOFSPEC_DETECTOR + IAS_RPS_IMPNOISE +
        IAS_RPS_ADCONV_SAT_LO + IAS_RPS_ADCONV_SAT_HI + IAS_RPS_ANALOG_SAT_LO +
        IAS_RPS_ANALOG_SAT_HI + IAS_RPS_INOP_DETECTOR);
    if (status < 0)
    {
        IAS_LOG_ERROR("Printing final legend information to file %s",
            legend_file_name);
        fclose(fptr);
        remove(legend_file_name);
        return ERROR;
    }

    status = fclose(fptr);
    if (status != 0)
    {
        IAS_LOG_ERROR("Closing legend file %s", legend_file_name);
        remove(legend_file_name);
        return ERROR;
    }

    return SUCCESS;
}   /* END -- write_mask_legend_file */



/**************************************************************************
    main routine
**************************************************************************/
int main(int argc, char *argv[])
{
    IAS_PIXEL_MASK_IO *pm_file = NULL;
    const char *pixel_mask_filename = NULL;
                                        /* Name of pixel mask file */
    const char *mask_image_filename = NULL;
                                        /* Copy of name of output mask
                                           image file */
    int band_number = 0;                /* Band number read from command
                                           line */
    int sca_number = 0;                 /* SCA number read from command
                                           line */
    int status;                         /* Function return status */

    /* Check the number of arguments.  There should be a minimum of two --
       the name of a pixel mask file, and the name of the output mask image
       file.  The remaining parameters are optional selections. */
    if ((argc < 3) || (argc > 5))
    {
        printf("\nUSAGE:  write_mask_image\n"
        "        <name of pixel_mask_file>\n"
        "        <output mask file name>\n"
        "        [<band_number>] <[sca_number]>\n\n"
        "        In all cases, the pixel mask file name and the name\n"
        "        of the output mask image file are required parameters.\n\n"
        "        Valid values for the optional <band_number> parameter\n"
        "        are either 0 (all bands present) or a band number between\n"
        "        1 and IAS_MAX_TOTAL_BANDS.\n\n"
        "        Valid values for the optional <sca_number> parameter\n"
        "        that can be specified on the commmand line are between 1\n"
        "        and the maximum number of SCAs present on either of the\n"
        "        sensors.  If not specified on the command line,\n"
        "        <sca_number> will default to 0, meaning all SCAs for the\n"
        "        current band will be written to the image file.\n\n"
        "OPTIONAL PARAMETER SPECIFICATION AND APPLICATION BEHAVIOR:\n"
        "        The application behaves as follows, according to how\n"
        "        the optional parameters are specified on the command\n"
        "        line:\n\n"
        "        1.  if neither <band_number> nor <sca_number> are\n"
        "        specified (the default case), all masks found in the\n"
        "        pixel mask array will be written to the image file.\n\n"
        "        2.  if <band_number> is specified, but <sca_number>\n"
        "        is not, all masks found in the pixel mask array with the\n"
        "        specified band number will be written to the image file.\n"
        "        'Missing' SCAs for a specific band number will be\n"
        "        treated as though they were 'empty' pixel masks (i.e.\n"
        "        masks with no artifacts flagged).\n\n"
        "        3.  if <band_number> is specified as 0, but <sca_number>\n"
        "        is specified with a valid SCA number, all masks found in\n"
        "        the pixel mask array with the specified SCA number will be\n"
        "        written to the image file; 'missing' bands will NOT be\n"
        "        included.  If <sca_number> is not specified in this case,\n"
        "        the default behavior will occur (i.e. all masks found in\n"
        "        the pixel mask array will be written to the image file,\n"
        "        with 'missing' SCAs written as 'empty' pixel masks).\n\n"
        "        4.  If both <band_number> and <sca_number> are specified,\n"
        "        the single mask with the specified band and SCA number\n"
        "        will be written to the image file.\n\n"
        "EXAMPLES:\n"
        "           i) to write all masks from the mask file\n"
        "              'pixel_mask.dat' to the image file "
        "'l1r_mask_image_full.h5':\n\n"
        "              ./write_mask_image pixel_mask.dat "
        "l1r_mask_image_full.h5\n\n"
        "          ii) to write the masks for band 5 from the mask file\n"
        "              'pixel_mask.dat' to the image file "
        "'l1r_mask_image_B05.h5':\n\n"
        "              ./write_mask_image pixel_mask.dat l1r_image_B05.h5 "
        "5\n\n"
        "         iii) to write the masks for all bands present in SCA 5\n"
        "              from the mask file 'lunar_mask.dat' to the image file\n"
        "              'l1r_lunar_mask_SCA05.h5':\n\n"
        "              ./write_mask_image lunar_mask.dat "
        "l1r_lunar_mask_SCA05.h5 0 5\n\n"
        "          iv) to write the mask for band 1 SCA 10 from the mask\n"
        "              file 'pixel_mask.dat' to the image file\n"
        "              'l1r_mask_B01_SCA10.h5':\n\n"
        "              ./write_mask_image pixel_mask.dat "
        "l1r_mask_B01_SCA10.h5 1 10\n\n");
        return EXIT_FAILURE;
    }

    /* Initialize the logging library in case any library call issues a
       log message */
    ias_log_initialize("write_mask_image");

    /* Get the name of the pixel mask file from the command line. */
    pixel_mask_filename = argv[1];

    /* Get the name of the output mask image file from the command line */
    mask_image_filename = argv[2];

    /* Process the optional parameters.  The 'default' case is where
       the optional band/SCA parameters aren't specified (argc = 3) */
    if (argc >= 4)
    {
        /* We've specified an optional band number.  This can be any of the
           band numbers in the PM, or 0.  If the specified band number is 0, 
           masks from all bands present in the mask array should be written 
           to the image file */

        /* Get the band number from the command line */
        status = sscanf(argv[3], "%d", &band_number);
        if (status != 1)
        {
            IAS_LOG_ERROR("Reading band number from command line "
                "parameters");
            return EXIT_FAILURE;
        }

        /* Check the band number */
        if ((band_number < 0) || (band_number > IAS_MAX_TOTAL_BANDS))
        {
            IAS_LOG_ERROR("Invalid band number %d", band_number);
            return EXIT_FAILURE;
        }

        /* We've also specified an SCA number on the command line */
        if (argc == 5)
        {
            /* Get the SCA number from the command line */
            status = sscanf(argv[4], "%d", &sca_number);
            if (status != 1)
            {
                IAS_LOG_ERROR("Reading SCA number from command line ");
                return EXIT_FAILURE;
            }

            /* Check the SCA number */
            if ((sca_number < 0) || (sca_number > IAS_MAX_NSCAS))
            {
                IAS_LOG_ERROR("Invalid SCA number %d", sca_number);
                return EXIT_FAILURE;
            }
        }
    }

    /* Open the specified pixel mask file */
    pm_file = ias_pm_open_pixel_mask(pixel_mask_filename, IAS_READ);
    if (pm_file == NULL)
    {
        IAS_LOG_ERROR("Opening pixel mask file %s for reading",
            pixel_mask_filename);
        return EXIT_FAILURE;
    }

    /* Write the specified mask(s) to an image file */
    if (band_number && sca_number)
    {
        /* Write a single mask to the image file */
        status = write_single_mask_image_file(pm_file,
            mask_image_filename, band_number, sca_number);
    }
    else
    {
        /* Two cases to consider:
           i)  band_number = 0, sca_number != 0)
          ii)  band_number = X (don't care), sca_number = 0) */
        if (sca_number)
        {
            /* Band number = 0, SCA number != 0 */
            status = write_mask_image_file_by_sca(pm_file,
                mask_image_filename, sca_number);
        }
        else
        {
            /* Band number = X (don't care), SCA number = 0 */
            status = write_mask_image_file_by_band(pm_file,
                mask_image_filename, band_number);
        }
    }

    /* Close the pixel mask file */
    ias_pm_close_pixel_mask(pm_file);

    /* Check whether creating the image file succeeded */
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Creating pixel mask image file");
        return EXIT_FAILURE;
    }

    /* Write the mask legend file */
    IAS_LOG_INFO("Creating pixel mask values legend file..");
    status = write_mask_legend_file(mask_image_filename);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Creating pixel mask values legend file");
        return EXIT_FAILURE;
    }

    IAS_LOG_INFO("Creation of pixel mask image file %s completed "
        "successfully", mask_image_filename);
    return EXIT_SUCCESS;

}   /* END -- write_mask_image MAIN */
