/***************************************************************************
 NAME:       write_multiple_mask_image_file

 PURPOSE:    Module implementing routines that write more than one pixel
             mask to an image file
****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include "ias_l1r.h"
#include "ias_logging.h"
#include "ias_pixel_mask.h"
#include "ias_satellite_attributes.h"

#include "write_mask_image.h"


/***************************************************************************
 NAME:     generate_band_list

 PURPOSE:  Scans the pixel mask array and obtains the list of bands
           that are present.

 RETURNS:  Pointer to a populated band list array if successful,
           NULL if there is an error.
****************************************************************************/
static int *generate_band_list
(
    const IAS_PIXEL_MASK_BAND_SCA_LIST *pm_band_sca_list,
                                        /* I: Band/SCA list information from
                                           pixel mask file */
    int read_count,                     /* I: Number of masks in array */
    int *number_of_bands_found     /* O: Number of bands in band list */
)
{
    int *band_list = NULL;
    int mask_index;
    int band_count = 0;
    int number_of_available_bands;

    /* assume all pixel masks are unique band numbers */
    number_of_available_bands = read_count;  

    /* Allocate the band list data buffer */
    band_list = malloc(number_of_available_bands * sizeof(*band_list));
    if (band_list == NULL)
    {
        IAS_LOG_ERROR("Allocating band list data buffer");
        return NULL;
    }

    /* Parse the mask band/SCA list information to construct the output
       band list */
    for (mask_index = 0; mask_index < read_count; mask_index++)
    {
        int found_band = 0;
        int bandlist_index;

        /* Get the band number for the current mask */
        int band_number = pm_band_sca_list[mask_index].band_number;

        /* If the band hasn't been found before, add it to the band list
           and increment the band count */
        for (bandlist_index = 0; bandlist_index < band_count; bandlist_index++)
        {
            if (band_number == band_list[bandlist_index])
            {
                found_band = 1;
                break;
            }
        }

        /* Only include normal bands in the output image */
        if (!found_band && (band_number <= IAS_MAX_NBANDS))
        {
            band_list[band_count] = band_number;
            band_count++;
        }
    }

    /* If no bands were found, that should be considered an error condition */
    if (band_count == 0)
    {
        IAS_LOG_ERROR("No band numbers found in the pixel mask array");
        free(band_list);
        return NULL;
    }

    /* Save the final band count */
    *number_of_bands_found = band_count;

    return band_list;
}   /* END -- generate_band_list */


/***************************************************************************
 NAME:     get_highest_sca_number

 PURPOSE:  Scans the pixel mask band/SCA list and determines the highest
           numbered SCA for the given band to be checked

 RETURNS:  Largest SCA number found for the current band, or -1 if no
           SCAs were found for the specified band number.
****************************************************************************/
static int get_highest_sca_number
(
    const IAS_PIXEL_MASK_BAND_SCA_LIST *pm_band_sca_list,
                                        /* I: Band/SCA list information from
                                           pixel mask file */
    int read_count,                     /* I: Number of masks in file */
    int band_number                     /* I: Band number specified on
                                              command line */
)
{
    int mask_index;
    int highest_sca_number = -1;

    for (mask_index = 0; mask_index < read_count; mask_index++)
    {
        /* Get the band number for the current mask */
        int list_band_number = pm_band_sca_list[mask_index].band_number;

        /* just look for the max SCA number in mask */
        if (list_band_number == band_number)
        {
            int sca_num = pm_band_sca_list[mask_index].sca_number;
            if( sca_num > highest_sca_number )
               highest_sca_number = sca_num;
        }
    }

    return highest_sca_number;
}


/***************************************************************************
 NAME:     write_mask_image_file_by_band

 PURPOSE:  Converts the pixel masks (for all SCA(s) of each band in a
           specified band list) to images, then writes the images to
           an L1R image file

 RETURNS:  Integer status code of SUCCESS or FAILURE
****************************************************************************/
int write_mask_image_file_by_band
(
    const IAS_PIXEL_MASK_IO *pm_file,      /* I: Open pixel mask file */
    const char *L1R_filename,              /* I: Name of L1R image file */
    int band_number                        /* I: Band number specified on
                                              command line */
)
{
    IAS_PIXEL_MASK_BAND_SCA_LIST *pm_band_sca_list = NULL;
                                           /* List of bands/SCAs in mask
                                              file */
    L1RIO *output_image = NULL;            /* L1R image pointer */
    PIXEL_MASK_TYPE *mask_image = NULL; /* Mask image data buffer */
    int *band_list = NULL;         /* List of band(s) to process */
    int number_of_bands = 0;       /* Number of bands in band list */
    int list_index;                /* Index into band list */
    int status;                    /* Function return status */
    int read_count;                /* Number of masks in the file */


    /* Get the band/SCA list information from the mask file. */
    status = ias_pm_get_file_band_sca_list(pm_file, &pm_band_sca_list,
        &read_count);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving band/SCA list information from mask file");
        return ERROR;
    }

    /* Generate the band list to process.  If the band number is 0,
       get all the bands found in the pixel mask array.  Otherwise, build
       a band list array with a single element */
    if (band_number == 0)
    {
        band_list = generate_band_list(pm_band_sca_list, read_count,
            &number_of_bands);
        if (band_list == NULL)
        {
            IAS_LOG_ERROR("Generating band list from pixel mask array");
            free(pm_band_sca_list);
            return ERROR;
        }
    }
    else
    {
        number_of_bands = 1;
        band_list = malloc(number_of_bands * sizeof(*band_list));
        if (band_list == NULL)
        {
            IAS_LOG_ERROR("Allocating single-element band list array");
            free(pm_band_sca_list);
            return ERROR;
        }
        band_list[0] = band_number;
    }

    /* Open the output image file */
    output_image = ias_l1r_open_image(L1R_filename, IAS_WRITE);
    if (output_image == NULL)
    {
        IAS_LOG_ERROR("Opening L1R mask image file %s", L1R_filename);
        free(band_list);
        free(pm_band_sca_list);
        return ERROR;
    }

    /* Get the required image dimensions for each band in the band list.
       The image dimensions (number_of_lines, number_of_detectors) will be
       obtained from the mask with the first SCA number found for the band */
    for (list_index = 0; list_index < number_of_bands; list_index++)
    {
        int list_band_number = band_list[list_index];
        int number_of_scas;
        int sca_index;
        L1R_BAND_IO *output_band_io = NULL;    /* L1R band pointer  */
        int first_number_of_lines = 0;
        int first_number_of_detectors = 0;

        /* Get the highest SCA number present in the pixel mask file for the
           band */
        number_of_scas = get_highest_sca_number(pm_band_sca_list, read_count,
            list_band_number);

        for (sca_index = 0; sca_index < number_of_scas; sca_index++)
        {
            int number_of_detectors;
            int number_of_lines;
            int sca_number = (sca_index + 1);
            int mask_size = 0;

            /* Get the mask for the current band/SCA (if it exists).  The
               first mask found in the current band will provide the
               image dimensions (i.e. number of detectors and number of
               lines). */
            if (ias_pm_is_band_sca_present(pm_file, list_band_number,
                sca_number))
            {
                IAS_PIXEL_MASK *pm;

                /* Read the mask with the specified band/SCA from the file. */
                pm = ias_pm_read_single_mask_from_file(pm_file,
                    list_band_number, sca_number);
                if (pm == NULL)
                {
                    IAS_LOG_ERROR("Reading band number %d SCA number %d "
                        "pixel mask from file", list_band_number,
                        sca_number);
                    if (output_band_io != NULL)
                        ias_l1r_close_band(output_band_io);
                    ias_l1r_close_image(output_image);
                    free(pm_band_sca_list);
                    free(band_list);
                    free(mask_image);
                    return ERROR;
                }
                number_of_detectors = ias_pm_get_number_of_detectors(pm);
                number_of_lines = ias_pm_get_number_of_pixels(pm);
                if ((number_of_detectors == ERROR)
                    || (number_of_lines == ERROR))
                {
                    IAS_LOG_ERROR("Getting number of detectors and/or "
                        "number of image pixels dimensions for band number "
                        "%d", list_band_number);
                    if (output_band_io != NULL)
                        ias_l1r_close_band(output_band_io);
                    ias_l1r_close_image(output_image);
                    free(pm_band_sca_list);
                    free(band_list);
                    ias_pm_destroy(pm);
                    free(mask_image);
                    return ERROR;
                }

                /* If the L1R band isn't open yet, open it now and catch up
                   on zero-filling any SCAs that preceded this one.  Otherwise
                   confirm the number of lines and detectors match like they
                   should */
                if (output_band_io == NULL)
                {
                    IAS_DATA_TYPE dtype = IAS_PIXEL_MASK_DATA_TYPE;
                                        /* Data type of mask pixels */

                    first_number_of_lines = number_of_lines;
                    first_number_of_detectors = number_of_detectors;

                    /* Open the current band */
                    output_band_io = ias_l1r_open_band(output_image, 
                        list_band_number, &dtype, &number_of_scas,
                        &number_of_lines, &number_of_detectors);
                    if (output_band_io == NULL)
                    {
                        IAS_LOG_ERROR("Opening band number %d of output L1R "
                            "image '%s'", list_band_number, L1R_filename);
                        ias_l1r_close_image(output_image);
                        free(band_list);
                        free(pm_band_sca_list);
                        ias_pm_destroy(pm);
                        return ERROR;
                    }

                    /* Allocate memory for the mask image */
                    mask_size = number_of_detectors * number_of_lines
                         * sizeof(PIXEL_MASK_TYPE);
                    mask_image = malloc(mask_size);
                    if (mask_image == NULL)
                    {
                        IAS_LOG_ERROR("Allocating mask image data buffer for "
                            "band number %d", list_band_number);
                        free(band_list);
                        free(pm_band_sca_list);
                        ias_l1r_close_image(output_image);
                        ias_pm_destroy(pm);
                        return ERROR;
                    }

                    /* If this isn't the first SCA of the band, zero fill the
                       previous SCAs */
                    if (sca_number > 1)
                    {
                        int write_index;
                        memset(mask_image, 0, mask_size);
                        for (write_index = 1; write_index < sca_index;
                             write_index++)
                        {
                            status = ias_l1r_write_image(output_band_io,
                                write_index, 0, 0, number_of_lines,
                                number_of_detectors, mask_image);
                            if (status != SUCCESS)
                            {
                                IAS_LOG_ERROR("Writing band number %d SCA %d "
                                    "pixel mask image to L1R file '%s'",
                                    list_band_number, (write_index + 1),
                                    L1R_filename);
                                ias_l1r_close_band(output_band_io);
                                ias_l1r_close_image(output_image);
                                free(mask_image);
                                free(band_list);
                                free(pm_band_sca_list);
                                ias_pm_destroy(pm);
                                return ERROR;
                            }
                        }
                    }
                }
                else
                {
                    if ((number_of_lines != first_number_of_lines)
                        || (number_of_detectors != first_number_of_detectors))
                    {
                        IAS_LOG_ERROR("The pixel masks for band number %d "
                            "are different sizes", list_band_number);
                        ias_l1r_close_band(output_band_io);
                        ias_l1r_close_image(output_image);
                        free(mask_image);
                        free(band_list);
                        free(pm_band_sca_list);
                        ias_pm_destroy(pm);
                        return ERROR;
                    }
                }

                /* Convert the pixel mask to an image */
                status = ias_pm_get_image(pm, mask_image);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR("Reading pixel mask image for band number "
                        "%d SCA number %d", list_band_number, sca_number);
                    if (output_band_io != NULL)
                        ias_l1r_close_band(output_band_io);
                    ias_l1r_close_image(output_image);
                    free(mask_image);
                    free(band_list);
                    free(pm_band_sca_list);
                    ias_pm_destroy(pm);
                    return ERROR;
                }

                IAS_LOG_INFO("Writing band number %d SCA number %d mask to "
                    "image file", list_band_number, sca_number);

                /* Free the mask memory */
                ias_pm_destroy(pm);
                pm = NULL;
            } /* band/SCA present */
            else if (mask_image != NULL)
            {
                /* The band/SCA isn't present, but the mask_image has been
                   allocated, so zero fill the buffer so it can be written
                   to the image file */
                memset(mask_image, 0, mask_size);
            }

            /* Write the mask image data to the image file if the band has
               been opened.  If the band hasn't been opened yet, it means
               the output image size has not yet been determined so we can't
               zero fill a missing band/SCA. */
            if (output_band_io != NULL)
            {
                status = ias_l1r_write_image(output_band_io, sca_index,
                    0, 0, number_of_lines, number_of_detectors, mask_image);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR("Writing band number %d SCA %d pixel mask "
                        "image to L1R file '%s'", list_band_number,
                        (sca_index + 1), L1R_filename);
                    ias_l1r_close_band(output_band_io);
                    ias_l1r_close_image(output_image);
                    free(mask_image);
                    free(band_list);
                    free(pm_band_sca_list);
                    return ERROR;
                }
            }
        } /* SCA loop */

        /* Free the mask image for the current band */
        free(mask_image);
        mask_image = NULL;

        /* Close the current band in the image file */
        status = ias_l1r_close_band(output_band_io);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Closing band number %d of L1R mask image file '%s'",
                list_band_number, L1R_filename);
            ias_l1r_close_image(output_image);
            free(band_list);
            free(pm_band_sca_list);
            return ERROR;
        }
    }   /* END -- band list processing */

    /* Close the L1R file */
    status = ias_l1r_close_image(output_image);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Closing final L1R mask image file '%s'",
            L1R_filename);
        free(band_list);
        free(pm_band_sca_list);
        return ERROR;
    }

    /* Final cleanup */
    free(band_list);
    band_list = NULL;
    free(pm_band_sca_list);
    pm_band_sca_list = NULL;

    return SUCCESS;
}   /* END -- write_mask_image_file_by_band */



/***************************************************************************
 NAME:     write_mask_image_file_by_sca

 PURPOSE:  Converts the pixel masks (for all bands(s) of the specified
           SCA) to images, then writes the images to an L1R image file

 RETURNS:  Integer status code of SUCCESS or FAILURE

 NOTES:    This function is intended to be used primarily for analysis of
           lunar mask images, where one SCA is expected to image the entire
           lunar disk
****************************************************************************/
int write_mask_image_file_by_sca
(
    const IAS_PIXEL_MASK_IO *pm_file,      /* I: Open pixel mask file */
    const char *L1R_filename,              /* I: Name of L1R image file */
    int sca_number                         /* I: SCA number to process */
)
{
    IAS_PIXEL_MASK_BAND_SCA_LIST *pm_band_sca_list = NULL;
                                           /* List of bands/SCAs in mask
                                              file */
    L1RIO *output_image         = NULL;    /* L1R image pointer */
    L1R_BAND_IO *output_band_io = NULL;    /* L1R band pointer  */
    IAS_DATA_TYPE dtype = IAS_PIXEL_MASK_DATA_TYPE;
                                           /* Data type of mask pixels */
    PIXEL_MASK_TYPE *mask_image = NULL;    /* Mask image data buffer */
    int   *band_list = NULL;               /* List of bands to process */
    int   number_of_bands = 0;             /* Number of bands in list */
    int   list_index;                      /* Index into band list */
    int   number_of_detectors = 0;         /* Number of detectors in mask */
    int   number_of_lines = 0;             /* Number of lines in mask */
    int   number_of_scas = 1;              /* Number of SCAs in output
                                              band */
    int   status;                          /* Function return status */
    int   band_number = 0;                 /* Current band number */
    int   read_count;                      /* Number of masks in file */
    size_t buffer_size;                    /* Total size of mask image
                                              buffer in bytes */


    /* Get the band/SCA list information from the mask file. */
    status = ias_pm_get_file_band_sca_list(pm_file, &pm_band_sca_list,
        &read_count);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving band/SCA list information from mask file");
        return ERROR;
    }

    /* Generate the band list to process.  If the band number is 0,
       get all the bands found in the pixel mask array. */
    band_list = generate_band_list(pm_band_sca_list, read_count,
        &number_of_bands);
    if (band_list == NULL)
    {
        IAS_LOG_ERROR("Generating band list from pixel mask array");
        free(pm_band_sca_list);
        return ERROR;
    }

    /* The band/SCA list is no longer needed, so free it */
    free(pm_band_sca_list);
    pm_band_sca_list = NULL;

    /* Open the output image file */
    output_image = ias_l1r_open_image(L1R_filename, IAS_WRITE);
    if (output_image == NULL)

    {
        IAS_LOG_ERROR("Opening L1R mask image file %s", L1R_filename);
        free(band_list);
        return ERROR;
    }

    /* Process all the bands in the list */
    for (list_index = 0; list_index < number_of_bands; list_index++)
    {
        /* Get the current band number from the list */
        band_number = band_list[list_index];

        /* See if the mask is present */
        if (ias_pm_is_band_sca_present(pm_file, band_number, sca_number))
        {
            /* Get the number of lines and detectors for the mask image */
            IAS_PIXEL_MASK *pm = ias_pm_read_single_mask_from_file(pm_file,
                band_number, sca_number);
            if (pm == NULL)
            {
                IAS_LOG_ERROR("Reading band number %d SCA number %d pixel "
                    "mask from file", band_number, sca_number);
                ias_l1r_close_image(output_image);
                free(band_list);
                return ERROR;
            }
            number_of_detectors = ias_pm_get_number_of_detectors(pm);
            number_of_lines = ias_pm_get_number_of_pixels(pm);
            if ((number_of_detectors == ERROR) || (number_of_lines == ERROR))
            {
                IAS_LOG_ERROR("Retrieving dimensions for band number %d "
                    "SCA number %d mask", band_number, sca_number);
                ias_l1r_close_image(output_image);
                ias_pm_destroy(pm);
                free(band_list);
                return ERROR;
            }

            /* Open the L1R band  */
            output_band_io = ias_l1r_open_band(output_image, band_number,
                &dtype, &number_of_scas, &number_of_lines,
                &number_of_detectors);
            if (output_band_io == NULL)
            {
                IAS_LOG_ERROR("Opening band number %d of L1R pixel mask file "
                    "'%s'", band_number, L1R_filename);
                ias_l1r_close_image(output_image);
                ias_pm_destroy(pm);
                free(band_list);
                return ERROR;
            }

            /* Allocate the image data buffer */
            buffer_size = number_of_lines * number_of_detectors
                * sizeof(PIXEL_MASK_TYPE);
            mask_image = malloc(buffer_size);
            if (mask_image == NULL)
            {
                IAS_LOG_ERROR("Allocating band number %d mask image data "
                   "buffer", band_number);
                ias_l1r_close_band(output_band_io);
                ias_l1r_close_image(output_image);
                ias_pm_destroy(pm);
                free(band_list);
                return ERROR;
            }

            /* Populate the image data buffer */
            status = ias_pm_get_image(pm, mask_image);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Reading pixel mask image for band number %d "
                    "SCA number %d", band_number, sca_number);
                ias_l1r_close_band(output_band_io);
                ias_l1r_close_image(output_image);
                ias_pm_destroy(pm);
                free(mask_image);
                free(band_list);
                return ERROR;
            }

            /* Free up the mask. */
            ias_pm_destroy(pm);
            pm = NULL;

            /* Write it to file */
            IAS_LOG_INFO("Writing band number %d SCA number %d mask to "
                "image file", band_number, sca_number);

            status = ias_l1r_write_image(output_band_io, 0, 0, 0,
                number_of_lines, number_of_detectors, mask_image);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Writing band number %d SCA %d pixel mask "
                    "image to L1R file '%s'", band_number, sca_number,
                    L1R_filename);
                ias_l1r_close_band(output_band_io);
                ias_l1r_close_image(output_image);
                free(mask_image);
                free(band_list);
                return ERROR;
            }

            /* Clean up for this band */
            free(mask_image);
            mask_image = NULL;

            status = ias_l1r_close_band(output_band_io);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Closing band number %d of L1R mask image file "
                    "'%s'", band_number, L1R_filename);
                ias_l1r_close_image(output_image);
                free(band_list);
                return ERROR;
            }
        }
    }   /* END -- band list processing */

    /* Free up the band list array */
    free(band_list);
    band_list = NULL;

    /* Close the L1R file */
    status = ias_l1r_close_image(output_image);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Closing final L1R mask image file '%s'",
            L1R_filename);
        return ERROR;
    }

    return SUCCESS;
}   /* END -- write_mask_image_file_by_sca */
