/***************************************************************************
 NAME:     write_single_mask_image_file

 PURPOSE:  Converts a single band/SCA pixel mask to image, then writes it
           to an L1R image file

 RETURNS:  Integer status code of SUCCESS or FAILURE
****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>

#include "ias_l1r.h"
#include "ias_logging.h"
#include "ias_pixel_mask.h"

#include "write_mask_image.h"


int write_single_mask_image_file
(
    const IAS_PIXEL_MASK_IO *pm_file,      /* I: Open pixel mask file */
    const char *L1R_filename,              /* I: Name of output L1R image
                                              file */
    int band_number,                       /* I: Band number */
    int sca_number                         /* I: SCA number */
)
{
    IAS_PIXEL_MASK *pm = NULL;             /* Individual pixel mask */

    L1RIO *output_image = NULL;            /* L1R image pointer */
    L1R_BAND_IO *output_band_io = NULL;    /* L1R band pointer  */
    IAS_DATA_TYPE dtype = IAS_PIXEL_MASK_DATA_TYPE;
                                           /* Data type of mask pixels */
    PIXEL_MASK_TYPE *mask_image = NULL;
                                           /* Mask image data buffer */
    int mask_number_of_detectors;          /* Number of detectors in mask */
    int mask_number_of_lines;              /* Number of lines in mask */
    int number_of_scas = 1;                /* Number of SCAs in output
                                              band */
    int status;                            /* Function return status */
    size_t buffer_size;                    /* Total size of mask image
                                              buffer in bytes */

    /* Make sure the specified band/SCA combination is present. */
    if (!ias_pm_is_band_sca_present(pm_file, band_number, sca_number))
    {
        IAS_LOG_ERROR("Mask with band number %d SCA number %d not "
            "found in pixel mask file", band_number, sca_number);
        return ERROR;
    }

    /* Get the mask from the file. */
    pm = ias_pm_read_single_mask_from_file(pm_file, band_number, sca_number);
    if (pm == NULL)
    {
        IAS_LOG_ERROR("Reading band number %d SCA number %d mask from pixel "
            "mask file", band_number, sca_number);
        return ERROR;
    }

    /* Get the number of detectors and number of lines.  These determine the
       image dimensions  */
    mask_number_of_detectors = ias_pm_get_number_of_detectors(pm);
    mask_number_of_lines = ias_pm_get_number_of_pixels(pm);
    if ((mask_number_of_detectors == ERROR) || (mask_number_of_lines == ERROR))
    {
        IAS_LOG_ERROR("Retrieving number of detectors and/or number of lines "
            "for band %d SCA %d mask", band_number, sca_number);
        ias_pm_destroy(pm);
        return ERROR;
    }

    /* Allocate a buffer for the mask image and populate it */
    buffer_size = mask_number_of_lines * mask_number_of_detectors
        * sizeof(PIXEL_MASK_TYPE);
    mask_image = malloc(buffer_size);
    if (mask_image == NULL)
    {
        IAS_LOG_ERROR("Allocating mask image data buffer");
        ias_pm_destroy(pm);
        return ERROR;
    }
    status = ias_pm_get_image(pm, mask_image);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Reading band number %d SCA number %d mask image",
            band_number, sca_number);
        ias_pm_destroy(pm);
        free(mask_image);
        return ERROR;
    }
    ias_pm_destroy(pm);
    pm = NULL;

    /* Open the L1R image file */
    output_image = ias_l1r_open_image(L1R_filename, IAS_WRITE);
    if (output_image == NULL)
    {
        IAS_LOG_ERROR("Opening pixel mask image file %s", L1R_filename);
        free(mask_image);
        return ERROR;
    }

    /* Open the L1R band */
    output_band_io = ias_l1r_open_band(output_image, band_number, &dtype,
        &number_of_scas, &mask_number_of_lines, &mask_number_of_detectors);
    if (output_band_io == NULL)
    {
        IAS_LOG_ERROR("Opening band %d of L1R pixel mask file %s",
            band_number, L1R_filename);
        ias_l1r_close_image(output_image);
        free(mask_image);
        return ERROR;
    }

    /* Write the image to the file */
    IAS_LOG_INFO("Writing band number %d SCA number %d mask to image file",
        band_number, sca_number);

    status = ias_l1r_write_image(output_band_io, 0, 0, 0,
        mask_number_of_lines, mask_number_of_detectors, mask_image);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing pixel mask image to L1R file %s for "
            "band %d SCA %d", L1R_filename, band_number, sca_number);
        ias_l1r_close_band(output_band_io);
        ias_l1r_close_image(output_image);
        free(mask_image);
        return ERROR;
    }

    /* Clean up */
    free(mask_image);
    mask_image = NULL;

    status = ias_l1r_close_band(output_band_io);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Closing band %d of L1R pixel mask image file %s",
            band_number, L1R_filename);
        ias_l1r_close_image(output_image);
        return ERROR;
    }
    status = ias_l1r_close_image(output_image);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Closing pixel mask image file %s", L1R_filename);
        return ERROR;
    }

    return SUCCESS;
}   /* END -- write_single_mask_image_file */
