#!/usr/bin/perl

##############################################################################
#
# Purpose:
#    The purpose of this script is to simulate successful subsets from the
#    subsetter for work orders in the IAS.  It also simulates successfully
#    releasing the L0R from the subsetter.
#
# Usage:
#    simulate_subsetter_for_work_order.pl work_order_id
#
#    The update will only happen if the status in the subsetter is "REQ-NEW".  
#
# Notes:
#    The $IAS_IT_TEST_DATA_DIR environment variable can be overridden by
#    adding a second parameter with the directory to use
##############################################################################

use warnings;
use strict;
use DBI;
use POSIX;
use File::Find;

use ias_const;
use ias_wo_const;
use ias_processing_system;

# Subroutine Declarations
sub initDb();
sub closeDb();

# Database Configuration
my $woDbh;

##############################################################################
# Main
##############################################################################

if (not defined $ARGV[0])
{
    print ("Error: Missing command line argument.");
    exit FAILURE;
}

my $iasItTestDataDir = $ENV{"IAS_IT_TEST_DATA_DIR"};
if (not defined $iasItTestDataDir)
{
    print "ERROR: IAS_IT_TEST_DATA_DIR not defined\n";
    return FAILURE;
}

if (defined $ARGV[1])
{
    $iasItTestDataDir = $ARGV[1];
}

if (ias_init_system() != SUCCESS)
{
    print "ERROR: could not initialize ias system: $ias_error_msg\n";
    exit FAILURE;
}

# Open database connections
if (initDb() == ERROR)
{
    print "ERROR: could not connect to database\n";
    exit FAILURE;
}

my $work_order_id = $ARGV[0];

# Find the scene id requested by the work order
my $sql = "SELECT scene_id, work_order_id"
        . " FROM work_orders"
        . " WHERE work_order_id = '$work_order_id'"
        . " AND state = '" . DB_WO_L0R_REQUESTED . "'";

my $dbError = "";

my $selectSth = $woDbh->prepare($sql);

if (not defined $selectSth)
{
    print "ERROR: SQL PREPARE ERROR : $DBI::errstr";
    closeDb();
    exit FAILURE;
}
elsif (!$selectSth->execute())
{
    print "ERROR: SQL EXECUTE ERROR : $DBI::errstr";
    $selectSth->finish();
    closeDb();
    exit FAILURE;
}

# Process/Update each work order ID
while (my $dbRowRef = $selectSth->fetchrow_hashref())
{
    my $sceneId =  $dbRowRef->{SCENE_ID};
    my $woId =  $dbRowRef->{WORK_ORDER_ID};
    my $use_three_tirs_scenes_flag = 0;

    $sql = "SELECT PARM_VALUE"
        . " FROM WO_PARMS"
        . " WHERE work_order_id = '$woId'"
        . " AND script_id = 'DSW'"
        . " AND parm_name = 'USE_THREE_TIRS_SCENES_FLAG'";

    my $sth = $woDbh->prepare($sql);
    if (not defined $sth)
    {
        print "ERROR: SQL PREPARE ERROR : $DBI::errstr";
        exit FAILURE;
    }
    elsif (!$sth->execute())
    {
        print "ERROR: SQL EXECUTE ERROR : $DBI::errstr";
        $sth->finish();
        exit FAILURE;
    }

    # Check if the parameter is defined
    my ($parm_value) = $sth->fetchrow_array();
    if (defined $parm_value)
    {
        $use_three_tirs_scenes_flag = $parm_value;
    }
    $sth->finish();

    # Determine the L0R location
    my $l0rpPath = "$iasItTestDataDir/l0r/single_scene/$sceneId";
    if ($use_three_tirs_scenes_flag)
    {
        $l0rpPath = "$iasItTestDataDir/l0r/three_tirs_scenes/$sceneId";
    }

    # If the scene doesn't exit search for it
    if (!(-e $l0rpPath))
    {
        my @found_scenes;
        find( sub{  return unless -d;
                    return unless /$sceneId/;
                    push @found_scenes, $File::Find::name;
                 }, "$iasItTestDataDir/l0r/");

        # Check if a scene was found that matches
        if (scalar @found_scenes == 0)
        {
            print "WARN: No scene was found for scene id: $sceneId in " .
                "the directories under $iasItTestDataDir/l0r/\n";
            next;
        }
        $l0rpPath = $found_scenes[0];
    }

    $sql = "UPDATE subsetter_request"
         . " SET l0rp_location = '$l0rpPath',"
         . " status = 'L0Rp-READY'"
         . " WHERE work_order_id = '$work_order_id'"
         . " AND status = 'REQ-NEW'";

    my $rows = $woDbh->do($sql);

    if (not defined $rows)
    {
        print "ERROR: SQL UPDATE ERROR : $DBI::errstr\n";
        $woDbh->rollback();
        closeDb();
        exit FAILURE;
    }
}

$selectSth->finish();

# Commit the current updates to the database
if (!$woDbh->commit())
{
    print "ERROR: SQL COMMIT ERROR: $DBI::errstr\n";
    closeDb();
    exit FAILURE;
}

closeDb();

print "Successfully simulated subsetter for work order $work_order_id\n";

exit 0;

##############################################################################
# Subroutines
##############################################################################

##############################################################################
# Name: initDb
# Purpose: Initiate database connections to TRAM and WO_COMMON
# Returns: SUCCESS - database connections were successfully opened
#          ERROR - error occurred
# Parameters: None
##############################################################################
sub initDb()
{
    my $woEnv = $ENV{"IAS_DB_COM"};
    my $sid = $ENV{"ORACLE_SID"};
    my $username;
    my $password;

    # verify environment variables were set
    if (not defined $woEnv)
    {
        print "ERROR: IAS_DB_COM not defined\n";
        return ERROR;
    }

    if (not defined $sid)
    {
        print "ERROR: ORACLE_SID not defined\n";
        return ERROR;
    }

    ## Open work order database connection
    ($username, $password) = split(/\//,$woEnv);

    $woDbh = DBI->connect("DBI:Oracle:$sid", $username, $password,
                           {AutoCommit => 0, PrintError =>0});

    if (!$woDbh)
    {
        print "ERROR: could not connect to WO_COMMON: $DBI::errstr\n";
        return ERROR;
    }
    
    return SUCCESS;
}

##############################################################################
# Name: closeDb
# Purpose: Close database connections to TRAM and WO_COMMON
# Returns: None
# Parameters: None
##############################################################################
sub closeDb()
{
    $woDbh->disconnect();
}

