#!/usr/bin/perl -w

##
# update_gcplib.pl
#
# Usage: update_gcplib.pl <directory>
#
# <directory> should be a directory contain GCP directories.  Each GCP 
# directory should be name with format <path>_<row> and should contain the 
# GCPLib file.
#
# The script will strip existing comments from GCPLib and add a commnent for
# the zone number. A backup name GCPLib.orig is created before modifying 
# GCPLib.
##

use strict;
use DBI;

### DATABASE CONFIGURATION
my $dbh;

my $TRUE    = 1;
my $FALSE   = 0;
my $SUCCESS = 0;
my $ERROR   = -1;
my $USER    = $ENV{USER};

### SUBROUTINE DECLARATIONS
###############################################################################
sub usage();
sub db_connect();
sub db_close();
sub get_zone($$);

### MAIN
###############################################################################

### GET BASE DIRECTORY FROM COMMAND LINE ARGS
my $base_directory;
$base_directory = shift(@ARGV) or usage();

print "Base directory: $base_directory\n";

### LOAD ALL CHIP DIRECTORIES UNDER BASE DIRECTORY
my @chip_directories;
if ( -d $base_directory )
{
    opendir(my $temp_dh, $base_directory);

    while (my $t = readdir($temp_dh))
    {
        if ( -d "$base_directory/$t" && $t ne "." && $t ne "..")
        {
            push(@chip_directories, $t);
        }
    } 
}
else
{
    print "$base_directory does not exists.\n";
    exit 1;
}

## Open the database connection
my $result = db_connect();
if($result == $ERROR)
{
    print "ERROR: connecting to database\n";
    exit 1;
}

for my $dir (@chip_directories)
{
    ### PARSE THE PATH/ROW FROM THE DIRECTORY NAME
    my ($path, $row) = split("_", $dir);

    print "Process path/row $path/$row\n";

    ### LOOKUP THE ZONE USING PATH ROW
    my $zone = get_zone($path, $row);

    print "$path/$row = $zone\n";

    ### UPDATE THE GCPLib FILE UNDER THE CHIP DIRECTORY
    my $backup_file = "$base_directory/$dir/GCPLib.orig";
    my $lib_file = "$base_directory/$dir/GCPLib";

    # COPY GCPLib to GCPLib.orig
    if(!rename($lib_file, $backup_file))
    {   
        print "ERROR: creating GCPLib.orig\n";
        db_close();
        exit 1;
    }

    open(my $BAKFH, "<", $backup_file);
    open(my $LIBFH, ">", $lib_file);

    my $printZone = $TRUE;
    while(<$BAKFH>)
    {
        if($_ =~ m/^#/)
        {
            #Already has zone code, print the comment
            if($_ =~ m/Zone code:/)
            {
                $printZone = $FALSE;
                print $LIBFH $_;
            }
            next;
        }

        #Print zone after other comments and if Zone code comment
        #does not already exist;
        if($printZone == $TRUE)
        {
            print $LIBFH "#    Zone code: $zone\n";
            $printZone = $FALSE;
            next;
        }

        #Not a comment print the lin
        print $LIBFH $_;
    }

    close($BAKFH);
    close($LIBFH);
}

## close the database connection
db_close();

exit 0;

### SUBROUTINES
###############################################################################
sub usage()
{
    print "usage: update_gcplib.pl <base_directory>\n";
    exit 1;
}

sub db_connect()
{
    my $L8_DB = $ENV{"IAS_DB_L8_OLITIRS"};
    my $SID   = $ENV{"ORACLE_SID"};
    my ($username, $password) = split("/", $L8_DB);

    if (!defined $L8_DB)
    {
        print "ERROR: IAS_DB_L8 not defined\n";
        return $ERROR;
    }
    
    if (!defined $SID)
    {
        print "ERROR: ORACLE_SID not defined\n";
        return $ERROR;
    }

    $dbh = DBI->connect("DBI:Oracle:$SID", $username, $password,
                        {AutoCommit => 0, PrintError => 0});

    if(!$dbh)
    {
        print "ERROR: could not connect to IAS database: ".$DBI::errstr."\n";
        return $ERROR;
    }
    
    return $SUCCESS;
}

sub db_close()
{
    $dbh->disconnect();
}

# Convert wrs_path/wrs_row to UTM zone.
# @returns UTM zone or $ERROR
sub get_zone($$)
{
    my ($path, $row) = @_;

    my $sql = "SELECT zone_nbr FROM datasets_ias.path_row_to_utm_zone "
            . " WHERE wrs_path = $path AND wrs_row = $row";

    my $sth = $dbh->prepare($sql);
    if(!$sth)
    {
        print "ERROR: parsing ZONE_NBR select statement".$DBI::errstr."\n";
        return $ERROR;
    }

    $sth->execute();

    my $zone_nbr;
    if (!(($zone_nbr) = $sth->fetchrow_array()))
    {
        print "ERROR: Zone not found for $path/$row\n";
        $zone_nbr = $ERROR;
    }
    $sth->finish();
    
    return $zone_nbr;
}
