#!/bin/csh -f
################################################################################
# Name: run_in_instance.csh
#
# Purpose: This script will source a setup file (passed as parameter/looked up)
#          for a different instance and run the commands passed as arguments
#
# Example: run_in_instance.csh [--allow_path|-a] Collection1 resample
#                              RESAMPLE.odl
#
################################################################################

# Define the constants and alias to be used
set ERROR = 1
set DEBUG = 0

alias usage 'echo;echo "USAGE:"; echo "    run_in_instance'\
            '[--allow_path|-a] <instance_name/path_to_setup_file>'\
            '<command_line_arguments>";echo;'

alias error_msg "echo;echo `date`  'run_in_instance'  $$  ERROR: \!:1"

alias debug_msg "if ($DEBUG) echo; \\
                 if ($DEBUG) echo `date` 'run_in_instance' $$ DEBUG: \!:1;"

# Check that the number of arguments is not less than 2
if ($#argv < 2) then
    usage
    exit $ERROR
endif

# Check if allow_path is set
set allow_path = 0
if ("$1" == "--allow_path" || "$1" == "-a") then
    set allow_path = 1
    shift
    debug_msg "Overriding to allow using paths"
endif

# Get the collection and remove it from the arguments
set collection = $1
shift
debug_msg "Running using instance: $collection"

# Setup the command to be run
set run_cmd = "$*"
debug_msg "Command to run: $run_cmd"

# Check if the collection instance passed is a path or if it needs to be looked
# up in the processing instances table
if ($collection =~ '*/*') then
    debug_msg "Contains path: $collection"
    set path_to_setup = $collection
else
    # Retrieve the path to the setup file
    debug_msg "Retrieving instance from database: $collection"
    set path_to_setup = `sqlplus -s $IAS_DB_COM` << END1
        set head off;
        set feedback off;
        SELECT
            SETUP_FILE
        FROM
            PROCESSING_INSTANCES
        WHERE
            INSTANCE_NAME = '$collection';
        exit;
END1

    # Check that the processing instance was found
    if (! $?path_to_setup || "$path_to_setup" =~ '*no rows*' \
        || $path_to_setup == "") then
        error_msg "Could not find instance $collection in PROCESSING_INSTANCES"
        usage
        exit $ERROR
    endif
endif
debug_msg "Using $path_to_setup"

# Ensure that the setup file exist before sourcing it
if ( ! -f $path_to_setup ) then
    error_msg "Path to setup file does not exist: $path_to_setup"
    usage
    exit $ERROR
endif

# Check that the executable doesn't contain a path (ran from sourced path)
if (! $allow_path && $1 =~ '*/*') then
    error_msg "Path should not be part of script ran: $1"
    usage
    exit $ERROR
endif

# Setup the environment variables to be reset
set env_variables = `printenv | awk -F "=" '{print $1}'`

# Reset any enviroment variables that contain a path
foreach variable ($env_variables)
    if ($variable =~ {'*PATH*','*BUILD*','*SRCROOT*','*LIB*','*DEV*'}) then
        unsetenv $variable
    endif
end
setenv PATH "/usr/local/bin:/bin:/usr/bin"

# Source the setup file
source $path_to_setup

debug_msg "Running from $PATH - $run_cmd"
$run_cmd

exit $status
