#!/usr/bin/perl -w
#############################################################################
# Name: skip_precision_for_tirs.pl
#
# Description: This is a utility to update an existing work order to skip
#   processing scripts if the data being processed is TIRS-only.  The list of
#   scripts to skip is provided in an ODL parameter file.  The OMF file is also
#   read and potentially updated.
#
# Parameters:
#   Parameter file name which contains: 
#       - WORK_ORDER_ID parameter
#       - SCRIPT_LIST parameter
#       - MODEL_PROCESSING_PASS parameter
#       - FORCE_SKIP_PRECISION_FLAG parameter
#   OMF file which contains:
#       - SENSOR parameter
#       - MODEL_FILENAME_PASS_## parameter
#
# Outputs:
#   If the SENSOR is TIRS, the work order script states are updated to skip
#   the list of provided scripts.  The OMF is also updated to populate the
#   following parameters:
#       PRECISION_SUCCEEDED_FLAG_PASS_## (set to zero to indicate systematic)
#       MODEL_FILENAME_PASS_## (Set to the filename from MODEL_FILENAME_PASS_##)
#
#############################################################################

use strict;
use DBI;
use File::Basename;
use File::Copy;
use Logger;         # Perl module to issue log messages
use ias_const;      # Various IAS constants
use ias_odl;        # Module to parse ODL files
use ias_db;         # Database support routines
use flow_utilities; # Importing skip_work_order_scripts

# Declare a constant for the program name
use constant PROG => basename($0);

main_routine();
exit SUCCESS;

#############################################################################
# Name: main_routine
#
# Description: Main routine for the script.  It reads the input parameter file
#   and calls the routine to update the script states.
#############################################################################
sub main_routine
{
    # Create the logger
    my $logger = Logger->new("STDOUT");
    if (!$logger)
    {
        print STDERR PROG . ": ERROR: Failed to establish logger object for"
            . " STDOUT.\n";
        exit FAILURE;
    }
    $logger->set_module_name(PROG);
    set_default_log($logger);

    # Verify a single parameter was provided and get it
    if (scalar @ARGV != 1)
    {
        $logger->fatal_msg("Missing parameter file name parameter");
        $logger->close_log();
        exit FAILURE;
    }

    # Get the parameter file that was passed in
    my $parameter_filename = $ARGV[0];

    # Parse the parameter file
    my $work_order_id;
    my @script_names;
    my $sensor_name;
    my $model_filename;
    my $model_pass;
    my $force_skip_flag;
    my $status = parse_parameter_file($parameter_filename, \$work_order_id,
        \@script_names, \$sensor_name, \$model_filename, \$model_pass,
        \$force_skip_flag);
    if ($status != SUCCESS)
    {
        $logger->fatal_msg("Error parsing parameters");
        $logger->close_log();
        exit FAILURE;
    }

    # Force skipping precision if requested
    if (!$force_skip_flag)
    {
        # If the sensor isn't just TIRS, nothing more to do
        if ($sensor_name ne "TIRS")
        {
            $logger->info_msg("The sensor is $sensor_name, so not skipping "
                . "precision processing steps");
            $logger->close_log();
            return;
        }
    }
    else
    {
        $logger->info_msg("The force skip precision flag is enabled so "
            . "skipping precision processing steps");
    }

    # Get the username and password for the COMMON database
    my ($user_id, $password) = ias_db_GetDBConnectionInfo('COMMON');

    # Connect to the database
    my $dbh = ias_db_ConnectToDB($user_id, $password);
    if (!defined $dbh)
    {
        $logger->fatal_msg("Unable to connect to database for user $user_id");
        $logger->close_log();
        exit FAILURE;
    }

    # Update the script states
    if (skip_work_order_scripts($dbh, $work_order_id, \@script_names)
        != SUCCESS)
    {
        $dbh->rollback();
        $dbh->disconnect();
        $logger->fatal_msg("Failed updating work order to skip scripts");
        $logger->close_log();
        exit FAILURE;
    }

    # Update the OMF file
    if (update_omf($work_order_id, $model_filename, $model_pass) != SUCCESS)
    {
        $dbh->rollback();
        $dbh->disconnect();
        $logger->fatal_msg("Failed updating the OMF file for work order "
            . "$work_order_id");
        $logger->close_log();
        exit FAILURE;
    }

    # Successfully updated the script states, so commit the changes
    if (!$dbh->commit())
    {
        $dbh->rollback();
        $dbh->disconnect();
        $logger->fatal_msg("Failed committing database updates");
        $logger->close_log();
        exit FAILURE;
    }

    # Disconnect from the database
    $dbh->disconnect();

    $logger->info_msg("Successfully updated work order $work_order_id to "
        . "skip processing scripts");

    # Close the log file
    $logger->close_log();
}

#############################################################################
# Name: parse_parameter_file
#
# Description: Parses the application parameter file.  The parameters read
#   include:
#       WORK_ORDER_ID
#       SCRIPT_LIST
#
# Parameters:
#   Name of the parameter file
#   Reference to return the work order id
#   Reference to the array for the script names
#   Reference to return the sensor name
#   Reference to return the model filename
#   Reference to return the model processing pass
#
# Returns:
#   SUCCESS/ERROR
#   Work order ID
#   Reference to the array of script names
#
#############################################################################
sub parse_parameter_file
{
    my $filename = shift;
    my $work_order_id_ref = shift;
    my $script_names_ref = shift;
    my $sensor_name_ref = shift;
    my $model_filename_ref = shift;
    my $model_pass_ref = shift;
    my $force_skip_flag_ref = shift;

    my $logger = get_default_log();

    # Default the force skip flag to not skip
    $$force_skip_flag_ref = 0;

    # Create an ODL reader
    my $parameters = new ias_odl();

    # Read the parameter file
    my $status = $parameters->load_odl($filename);
    if ($status != SUCCESS)
    {
        $logger->fatal_msg("Failed reading parameters from $filename: "
            . $parameters->{error_string});
        return ERROR;
    }

    # Get the group name
    my @groups = $parameters->groups();
    if (!@groups)
    {
        $logger->fatal_msg("Failed reading parameters from $filename: "
            . $parameters->{error_string});
        return ERROR;
    }
    if (scalar @groups != 1)
    {
        $logger->fatal_msg("Multiple groups found in $filename; only one "
            . "allowed");
        return ERROR;
    }
    my $group = $groups[0];

    # Get the work order id
    my $work_order_id
        = $parameters->get_parameter_value("$group.WORK_ORDER_ID");
    if (!defined $work_order_id)
    {
        $logger->fatal_msg("Failed reading WORK_ORDER_ID parameter from "
            . "$filename: " . $parameters->{error_string});
        return ERROR;
    }
    $$work_order_id_ref = $work_order_id;

    # Get the script list string
    my $script_string = $parameters->get_parameter_value("$group.SCRIPT_LIST");
    if (!defined $script_string)
    {
        $logger->fatal_msg("Failed reading SCRIPT_LIST parameter from "
            . "$filename: " . $parameters->{error_string});
        return ERROR;
    }

    # Get the skip flag if passed
    my $force_skip_flag = $parameters->get_parameter_value(
        "$group.FORCE_SKIP_PRECISION_FLAG");
    if (defined $force_skip_flag)
    {
        $$force_skip_flag_ref = $force_skip_flag;
    }

    # Split the script list string into individual names and eliminate any
    # leading or trailing spaces
    my @scripts = split(',', $script_string);
    foreach my $script (@scripts)
    {
        $script =~ s/^\s*//;
        $script =~ s/\s*$//;
        push (@$script_names_ref, $script);
    }

    # Get the model processing pass
    my $model_processing_pass 
        = $parameters->get_parameter_value("$group.MODEL_PROCESSING_PASS");
    if (!defined $model_processing_pass)
    {
        $logger->fatal_msg("Failed reading MODEL_PROCESSING_PASS parameter from"
            . " $filename: " . $parameters->{error_string});
        return ERROR;
    }
    $$model_pass_ref = $model_processing_pass;

    # Read the OMF file
    my $omf = new ias_odl();
    my $omf_name = $work_order_id . ".omf";
    $status = $omf->load_odl($omf_name);
    if ($status != SUCCESS)
    {
        $logger->fatal_msg("Failed reading OMF from $omf_name: "
            . $omf->{error_string});
        return ERROR;
    }

    # Get the sensor parameter
    my $sensor = $omf->get_parameter_value("OMF.SENSOR");
    if (!defined $sensor)
    {
        $logger->fatal_msg("Failed reading SENSOR parameter from $omf_name: "
            . $omf->{error_string});
        return ERROR;
    }
    $$sensor_name_ref = $sensor;

    # Get the MODEL_FILENAME_PASS_## parameter
    my $model_filename = $omf->get_parameter_value
        ("OMF.MODEL_FILENAME_PASS_$model_processing_pass");
    if (!defined $model_filename)
    {
        $logger->fatal_msg("Failed reading " 
            . "MODEL_FILENAME_PASS_$model_processing_pass parameter "
            . "from $omf_name: " . $omf->{error_string});
        return ERROR;
    }
    $$model_filename_ref = $model_filename;

    return SUCCESS;
}

#############################################################################
# Name: update_omf
#
# Description: Updates the OMF file to add or update the
#   PRECISION_SUCCEEDED_FLAG_PASS_## and MODEL_FILENAME_PASS_## parameters.  The
#   precision succeeded flag is set to zero to indicate it did not succeed
#   (since it wasn't run) and the model filename is set the same as the pass 1
#   model filename.
#
# Parameters:
#   Work order ID
#   Model filename
#   Model processing pass
#############################################################################
sub update_omf
{
    my $work_order_id = shift;
    my $model_filename = shift;
    my $model_pass = shift;

    # Increment the model pass to be used but save the original pass number 
    # to use for the precision succeeded flag pass number
    my $flag_pass = $model_pass;
    $model_pass = $model_pass + 1;

    # Build the OMF filename using the work order ID
    my $omf_name = $work_order_id . ".omf";
    my $output_omf_name = $omf_name . ".temp";

    my $logger = get_default_log();

    # Read the original OMF file
    my $status = open (INPUTOMF, "< $omf_name");
    if (!defined($status))
    {
        $logger->fatal_msg("Can't open input OMF file $omf_name for reading");
        return ERROR;
    }
    my @lines = <INPUTOMF>;
    close(INPUTOMF);

    # Open the output file
    $status = open (OUTPUTOMF, "> $output_omf_name");
    if (!defined($status))
    {
        $logger->fatal_msg("Can't open input OMF file $output_omf_name"
            . " for writing");
        return ERROR;
    }

    # Echo the input file lines to the output file, updating the two parameters
    # to make it look like precision correction ran but decided to use the
    # systematic model
    foreach my $line (@lines)
    {
        # Watch for end of parameters
        if ($line =~ /^END_OBJECT/)
        {
            if (! (print OUTPUTOMF "  PRECISION_SUCCEEDED_FLAG_PASS_$flag_pass "
                . "= 0\n  MODEL_FILENAME_PASS_$model_pass = "
                . "\"$model_filename\"\n"))
            {
                $logger->fatal_msg("Failed writing updated OMF parameters");
                close(OUTPUTOMF);
                return ERROR;
            }
        }

        # Don't write out previous versions of the
        # PRECISION_SUCCEEDED_FLAG_PASS_## or MODEL_FILENAME_PASS_##
        if ($line !~ /PRECISION_SUCCEEDED_FLAG_PASS_$flag_pass/ 
            && $line !~ /MODEL_FILENAME_PASS_$model_pass/)
        {
            if (!print OUTPUTOMF $line)
            {
                $logger->fatal_msg("Failed writing OMF parameters");
                close(OUTPUTOMF);
                return ERROR;
            }
        }
    }

    # Close the OMF file.
    $status = close(OUTPUTOMF);
    if (!$status)
    {
        $logger->fatal_msg("Output OMF file $output_omf_name did not close"
            . " properly");
        return  ERROR;
    }

    # Rename the new OMF file to the correct name
    if (!move($output_omf_name, $omf_name))
    {
        $logger->fatal_msg("Failed renaming $output_omf_name to $omf_name");
        return  ERROR;
    }

    return SUCCESS;
}
