#!/usr/bin/perl -w
#############################################################################
# Name: skip_work_order_scripts.pl
#
# Description: This is a utility to update an existing work order to skip
#   processing scripts if the data being processed is TIRS, OLI, or OLITIRS.
#   The list of scripts to skip is provided in an ODL parameter file.
#
# Parameters:
#   Parameter file name which contains:
#       - WORK_ORDER_ID parameter
#       - SCRIPT_LIST parameter
#       - SKIP_SCRIPTS_FOR_SENSOR parameter used as skip flag
#       - FORCE_SKIP_SCRIPTS_FLAG parameter used to force skipping of specified
#         scripts
#   OMF file which contains:
#       - SENSOR parameter
#
# Outputs:
#   If the SENSOR (OMF) is SKIP_SCRIPTS_FOR_SENSOR (Parameter File),
#   the work order script states are updated to skip the list of
#   provided scripts.
#############################################################################

use strict;
use DBI;
use File::Basename;
use Logger;         # Perl module to issue log messages
use ias_const;      # Various IAS constants
use ias_odl;        # Module to parse ODL files
use ias_db;         # Database support routines
use flow_utilities; # Importing skip_work_order_scripts

# Declare a constant for the program name
use constant PROG => basename($0);

main_routine();
exit SUCCESS;

#############################################################################
# Name: main_routine
#
# Description: Main routine for the script.  It reads the input parameter file
#   and calls the routine to update the script states.
#############################################################################
sub main_routine
{
    # Create the logger
    my $logger = Logger->new("STDOUT");
    if (!$logger)
    {
        print STDOUT PROG . ": ERROR: Failed to establish logger object for"
            . " STDOUT.\n";
        exit FAILURE;
    }
    $logger->set_module_name(PROG);
    set_default_log($logger);

    # Verify a single parameter was provided and get it
    if (scalar @ARGV != 1)
    {
        $logger->fatal_msg("Missing parameter file name");
        $logger->close_log();
        exit FAILURE;
    }

    # Get the parameter file that was passed in
    my $parameter_filename = $ARGV[0];

    # Parse the parameter file
    my $work_order_id;
    my @script_names;
    my $sensor_name;
    my $skip_sensor;
    my $force_skip_scripts_flag;

    my $status = parse_parameter_file($parameter_filename, \$work_order_id,
        \@script_names, \$sensor_name, \$skip_sensor,
        \$force_skip_scripts_flag);
    if ($status != SUCCESS)
    {
        $logger->fatal_msg("Error parsing parameters");
        $logger->close_log();
        exit FAILURE;
    }

    # If the sensors do not match, nothing more to do
    if ($force_skip_scripts_flag)
    {
        $logger->info_msg("The force skip flag is enabled. Skipping the "
            . "processing steps");
    }
    elsif ($sensor_name ne $skip_sensor)
    {
        $logger->info_msg("The sensor is $sensor_name, so not skipping "
            . "the processing steps");
        $logger->close_log();
        return;
    }

    # Get the username and password for the COMMON database
    my ($user_id, $password) = ias_db_GetDBConnectionInfo('COMMON');

    # Connect to the database
    my $dbh = ias_db_ConnectToDB($user_id, $password);
    if (!defined $dbh)
    {
        $logger->fatal_msg("Unable to connect to database for user $user_id");
        $logger->close_log();
        exit FAILURE;
    }

    # Update the script states
    if (skip_work_order_scripts($dbh, $work_order_id, \@script_names)
        != SUCCESS)
    {
        $dbh->rollback();
        $dbh->disconnect();
        $logger->fatal_msg("Failed updating work order to skip scripts");
        $logger->close_log();
        exit FAILURE;
    }

    # Successfully updated the script states, so commit the changes
    if (!$dbh->commit())
    {
        $dbh->rollback();
        $dbh->disconnect();
        $logger->fatal_msg("Failed committing database updates");
        $logger->close_log();
        exit FAILURE;
    }

    # Disconnect from the database
    $dbh->disconnect();

    $logger->info_msg("Successfully updated work order $work_order_id to "
        . "skip processing scripts");

    # Close the log file
    $logger->close_log();
}

#############################################################################
# Name: parse_parameter_file
#
# Description: Parses the application parameter file.  The parameters read
#   include:
#       WORK_ORDER_ID
#       SCRIPT_LIST
#
# Parameters:
#   Name of the parameter file
#   Reference to return the work order id
#   Reference to the array for the script names
#   Reference to return the sensor name
#   Reference to return the skip sensor name
#
# Returns:
#   SUCCESS/ERROR
#   Work order ID
#   Reference to the array of script names
#
#############################################################################
sub parse_parameter_file
{
    my $filename = shift;
    my $work_order_id_ref = shift;
    my $script_names_ref = shift;
    my $sensor_name_ref = shift;
    my $skip_sensor_ref = shift;
    my $force_skip_scripts_flag_ref = shift;
    my $logger = get_default_log();

    # Create an ODL reader
    my $parameters = new ias_odl();

    # Read the parameter file
    my $status = $parameters->load_odl($filename);
    if ($status != SUCCESS)
    {
        $logger->fatal_msg("Failed reading parameters from $filename: "
            . $parameters->{error_string});
        return ERROR;
    }

    # Get the group name
    my @groups = $parameters->groups();
    if (!@groups)
    {
        $logger->fatal_msg("Failed reading parameters from $filename: "
            . $parameters->{error_string});
        return ERROR;
    }
    if (scalar @groups != 1)
    {
        $logger->fatal_msg("Multiple groups found in $filename; only one "
            . "allowed");
        return ERROR;
    }
    my $group = $groups[0];

    # Get the work order id
    my $work_order_id
        = $parameters->get_parameter_value("$group.WORK_ORDER_ID");
    if (!defined $work_order_id)
    {
        $logger->fatal_msg("Failed reading WORK_ORDER_ID parameter from "
            . "$filename: " . $parameters->{error_string});
        return ERROR;
    }
    $$work_order_id_ref = $work_order_id;

    # Get the script list string
    my $script_string = $parameters->get_parameter_value("$group.SCRIPT_LIST");
    if (!defined $script_string)
    {
        $logger->fatal_msg("Failed reading SCRIPT_LIST parameter from "
            . "$filename: " . $parameters->{error_string});
        return ERROR;
    }

    my $skip_sensor
        = $parameters->get_parameter_value("$group.SKIP_SCRIPTS_FOR_SENSOR");
    if (!defined $skip_sensor)
    {
        $logger->fatal_msg("Failed reading SKIP_SCRIPTS_FOR_SENSOR parameter "
            . "from $filename: " . $parameters->{error_string});
        return ERROR;
    }
    $$skip_sensor_ref = $skip_sensor;

    my $force_skip_scripts_flag
        = $parameters->get_parameter_value("$group.FORCE_SKIP_SCRIPTS_FLAG");
    if (!defined $force_skip_scripts_flag)
    {
        # Default the flag to be zero if not present
        $force_skip_scripts_flag = 0;
    }
    $$force_skip_scripts_flag_ref = $force_skip_scripts_flag;

    # Split the script list string into individual names and eliminate any
    # leading or trailing spaces
    my @scripts = split(',', $script_string);
    foreach my $script (@scripts)
    {
        $script =~ s/^\s*//;
        $script =~ s/\s*$//;
        push (@$script_names_ref, $script);
    }

    # Read the OMF file
    my $omf = new ias_odl();
    my $omf_name = $work_order_id . ".omf";
    $status = $omf->load_odl($omf_name);
    if ($status != SUCCESS)
    {
        $logger->fatal_msg("Failed reading OMF from $omf_name: "
            . $omf->{error_string});
        return ERROR;
    }

    # Get the sensor parameter
    my $sensor = $omf->get_parameter_value("OMF.SENSOR");
    if (!defined $sensor)
    {
        $logger->fatal_msg("Failed reading SENSOR parameter from $omf_name: "
            . $omf->{error_string});
        return ERROR;
    }
    $$sensor_name_ref = $sensor;

    return SUCCESS;
}
