#!/usr/bin/perl -w
##############################################################################
# Name: test_skip_precision_for_tirs.pl
#
# Description: Test the skip_precision_for_tirs.pl script
#
##############################################################################

use strict;
use ias_const;
use ias_db;

my $instance = $ENV{IAS_INSTANCE_NAME};
if (!defined $instance)
{
    print "Set IAS_INSTANCE_NAME to enable work order insertion\n";
    exit FAILURE;
}

my $ldr_file = "work_orders_table.ldr";
my $template = $ldr_file . ".template";

# Replace text in the omf file
if (system("sed -e 's|INSTANCE_NAME_REPLACE|$instance|' $template > $ldr_file")
    != 0)
{
    print "Unable to create the work orders ldr file\n";
    exit FAILURE;
}

# Get the username and password for the COMMON database
my ($user_id, $password) = ias_db_GetDBConnectionInfo('COMMON');

# Connect to the database
my $dbh = ias_db_ConnectToDB($user_id, $password);
if (!defined $dbh)
{
    print "Unable to connect to database for user $user_id\n";
    exit FAILURE;
}

# Make sure the dummy work order is not already in the database
remove_wo_from_db($dbh);
$dbh->commit();

# Load a dummy order in the database
if (create_wo_in_db() != SUCCESS)
{
    $dbh->disconnect();
    exit FAILURE;
}

# Run a test case with the OLITIRS sensor.  No changes should occur in the
# database or OMF file.
if (run_test_case($dbh, "OLITIRS") != SUCCESS)
{
    $dbh->disconnect();
    exit FAILURE;
}

# Run a test case with the OLI sensor.  No changes should occur in the database
# or OMF file.
if (run_test_case($dbh, "OLI") != SUCCESS)
{
    $dbh->disconnect();
    exit FAILURE;
}

# Run a test case with the TIRS sensor.  The OMF should be updated to make it
# look like precision correction decided to make a systematic product and the
# correct scripts set to the skip state in the database.
if (run_test_case($dbh, "TIRS") != SUCCESS)
{
    $dbh->disconnect();
    exit FAILURE;
}

# Delete the dummy order from the database
if (remove_wo_from_db($dbh) != SUCCESS)
{
    $dbh->disconnect();
    exit FAILURE;
}
$dbh->commit();

# disconnect from the database
$dbh->disconnect();

exit SUCCESS;


##############################################################################
# Name: run_test_case
#
# Description: Perform one run of the test case
#
# Parameters:
#   Database connection
#   Sensor
#
# Returns
#   SUCCESS/ERROR
##############################################################################
sub run_test_case
{
    my $dbh = shift;
    my $sensor = shift;

    print "------ Testing $sensor -----\n";

    # Create an OMF with the correct sensor
    my $status = system("sed -e \"s/__SENSOR__/$sensor/\" T1234.omf.template "
        . " > T1234.omf");
    if ($status)
    {
        print "Error creating OMF file for the $sensor test case\n";
        return ERROR;
    }

    # Run the application
    $status = system("../skip_precision_for_tirs.pl test.odl");
    if ($status)
    {
        print "Error returned from skip_precision_for_tirs.pl for the $sensor "
            . "test case\n";
        return ERROR;
    }

    # Verify the OMF is as expected
    $status = system("diff -q $sensor.omf T1234.omf");
    if ($status)
    {
        print "The T1234.omf does not agree with the expected results in "
            . "$sensor.omf\n";
        return ERROR;
    }

    # Verify the database is as expected
    if (dump_wo_script_from_db($dbh, "db_dump.txt") != SUCCESS)
    {
        print "Error creating dump file for the $sensor test case\n";
        return ERROR;
    }
    $status = system("diff -q db_dump.txt $sensor.dump");
    if ($status)
    {
        print "The database contents in db_dump.txt do not agree with the "
            ." expected results in $sensor.dump\n";
        return ERROR;
    }

    print "Success\n";
    return SUCCESS;
}

##############################################################################
# Name: dump_wo_script_from_db
#
# Description: Dumps the wo_scripts table entries from the database for the
#   test work order id
#
# Parameters:
#   Database connection
#   Filename to dump to
#
# Returns:
#   SUCCESS/ERROR
##############################################################################
sub dump_wo_script_from_db
{
    my $dbh = shift;
    my $dump_name = shift;

    # Open a file to dump the results to
    if (!open(DUMP, "> $dump_name"))
    {
        print "Error opening $dump_name for the database dump\n";
        return ERROR;
    }
    # Select the script IDs and states for the test work order
    my $sql = "select SCRIPT_ID, STATE from WO_SCRIPTS where "
        . "WORK_ORDER_ID = 'T1234' ORDER BY SEQ_NBR";
    my $sth = $dbh->prepare($sql);
    if (!$sth)
    {
        print "Error preparing SQL: $sql\n";
        close(DUMP);
        return ERROR;
    }
    if (!$sth->execute())
    {
        $sth->finish();
        print "Error executing SQL: $sql\n";
        return ERROR;
    }
    while (my $row_hash = $sth->fetchrow_hashref())
    {
        print DUMP $row_hash->{SCRIPT_ID} . "|" . $row_hash->{STATE} . "\n";
    }
    $sth->finish();

    close(DUMP);

    return SUCCESS;
}

##############################################################################
# Name: create_wo_in_db
#
# Description: Createa a dummy work order in the database.  Exits if an error
#   occurs.
#
# Parameters:
#   none
#
# Returns:
#   SUCCESS/ERROR
##############################################################################
sub create_wo_in_db
{
    my $status = system("sqlldr \$IAS_DB_COM work_orders_table.ldr");
    if ($status)
    {
        print "Error loading work orders table\n";
        return ERROR;
    }

    $status = system("sqlldr \$IAS_DB_COM wo_scripts_table.ldr");
    if ($status)
    {
        print "Error loading work orders table\n";
        return ERROR;
    }

    return SUCCESS;
}

##############################################################################
# Name: remove_wo_from_db
#
# Description: Removes the dummy work order from the database.  Exits if an
#   error occurs.
#
# Parameters:
#   Database connection
#
# Returns:
#   SUCCESS/ERROR
##############################################################################
sub remove_wo_from_db
{
    my $dbh = shift;

    # Remove the work order from the work orders table.  It cascade deletes
    # the wo_scripts table.
    my $sql = "delete from WORK_ORDERS where WORK_ORDER_ID = 'T1234'";
    my $sth = $dbh->prepare($sql);
    if (!$sth)
    {
        print "Error preparing SQL: $sql\n";
        return ERROR;
    }
    if (!$sth->execute())
    {
        $sth->finish();
        print "Error executing SQL: $sql\n";
        return ERROR;
    }
    $sth->finish();

    return SUCCESS;
}
