/******************************************************************************
Name: ias_ancillary_get_attitude_at_time

Purpose: Find the roll, pitch and yaw at requested time using the ancillary
    data.  Values are interpolated between two closet neighbors (roll, pitch,
    yaw) of requested time. 

Notes:
    - There is a similar routine named ias_sc_model_find_attitude_at_time that
      returns the attitude at a given time. It should be used whenever the 
      LOS model is available.
    - This routine is provided for the Ingest subsystem to be able to obtain
      the pitch/roll/yaw without needing to create a LOS model.

RETURNS: SUCCESS or ERROR

******************************************************************************/
#include <stdio.h>
#include <math.h>
#include "ias_logging.h"
#include "ias_ancillary.h"

int ias_ancillary_get_attitude_at_time
(
   const IAS_ANC_ATTITUDE_DATA *att,    /* I: Attitude structure */
   double att_time,                     /* I: Actual time (sec from epoch) */
   double *roll,                        /* O: Output roll (radians) */
   double *pitch,                       /* O: Output pitch (radians) */
   double *yaw                          /* O: Output yaw (radians) */
)
{
    int index1;                  /* Index to the "left of" att_time */
    int index2;                  /* Index to the "right of" att_time */
    double w;                    /* Interpolation weight */
    double first_time;           /* first attitude record time from epoch */
    double last_time;            /* last attitude record time from epoch */
    double time_diff;            /* time difference between records */
    const IAS_ANC_ATTITUDE_RECORD *records; /* pointer to attitude records */

    records = att->records;
    first_time = att->records[0].seconds_from_epoch;
    last_time = att->records[att->number_of_samples - 1].seconds_from_epoch;

    /* Make sure there are enough points to interpolate */
    if (att->number_of_samples < 2)
    {
        IAS_LOG_ERROR("Insufficient points to interpolate between (%d)", 
            att->number_of_samples);
        return ERROR;
    }

    /* Make sure the requested time is within the available times */
    if (att_time < first_time)
    {
        IAS_LOG_ERROR("Requested attitude time %f comes before the starting "
            "record time of %f", att_time, first_time);
        return ERROR;
    }
    else if (att_time > last_time)
    {
        IAS_LOG_ERROR("Requested attitude time %f falls after the ending "
            "record time of %f", att_time, last_time);
        return ERROR;
    }
    /* Find the attitude records on either side of the requested time.  Note
       that no assumption is made here that the samples are regularly spaced
       even though they likely are.  Start by finding the index of the record
       right before the requested time. */
    index1 = 0;
    while (index1 < att->number_of_samples - 1
           && att_time >= records[index1].seconds_from_epoch)
    {
        index1++;
    }
    /* back up one record to be below the requested time, if possible */
    if (index1 > 0)
        index1--;

    /* The next index should be greater than or equal to the target time */
    index2 = index1 + 1;

    /* Do some checks on the indices to make sure something didn't go wrong */
    if (index1 < 0 || index2 >= att->number_of_samples)
    {
        IAS_LOG_ERROR("Calculated attitude record indices are invalid for "
            "time %f.  Start index: %d, End index: %d, Max allowed index: %d",
            att_time, index1, index2, att->number_of_samples - 1);
        return ERROR;
    }

    /* Verify the end time makes sense */
    if (records[index2].seconds_from_epoch < att_time)
    {
        /* It should not be possible for this to happen, but just in case */
        IAS_LOG_ERROR("Ending time found (%f) is before the requested "
            "time (%f)", records[index2].seconds_from_epoch, att_time);
        return ERROR;
    }

    /* Use linear interpolation between the points */
    time_diff = records[index2].seconds_from_epoch
              - records[index1].seconds_from_epoch;

    w = (att_time - records[index1].seconds_from_epoch) / time_diff;

    *roll  =  records[index1].roll + (records[index2].roll 
           -  records[index1].roll) * w;
    *pitch =  records[index1].pitch + (records[index2].pitch 
           -  records[index1].pitch) * w;
    *yaw   =  records[index1].yaw + (records[index2].yaw 
           -  records[index1].yaw) * w;

    return SUCCESS;
}
