#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>              /* PATH_MAX definition */

#include "ias_gcc_pragmas.h"
#include "ias_types.h"
#include "ias_logging.h"
#include "ias_cpf.h"
#include "ias_math.h"
#include "ias_ancillary.h"
#include "ias_ancillary_private.h"

#include "cgreen.h"

#define ADP_STATIC_TEST_DATA 1

IAS_CPF *cpf = NULL;
/*
The correct CPF. Currently there are issues with UT1_Modified_Julian parameter.
#define CPF "/ias_lib/cpf/L8CPF20090101_20090331.01"
*/
#define CPF "/ias_lib/ancillary/L8CPF20090101_20090331.01"

/* Disable the command line specified optimization
   It is re-enabled below
   We disable the optimization because the inclusion of headers which contain
   structures defining huge amounts of testing data, cause very very very long
   compiles.  And it is the optimization in the newer compilers which cause
   the lengthy compile time with the include files. */
IAS_DISABLE_OPTIMIZATION

Ensure test_ias_identify_quaternion_outliers()
{
    int status = 0;
    int index = 0;
    int interpolate_quaternions_flag = 0;
    int quat_count;

    #include "test_headers/test_data_identify_quaternion_outliers_in.h"
    #include "test_headers/test_data_identify_quaternion_outliers_out.h"

    /* Allocate structure that will contain valid quaternion data */
    quat_count = l0r_attitude_count + IAS_L0R_NUM_SAMPLES_PER_IMU_RECORD;
    quaternion_data = (IAS_QUATERNION *)malloc(sizeof(IAS_QUATERNION)
        * quat_count);
    if (quaternion_data == NULL)
    {
        IAS_LOG_ERROR("Allocating quaternions");
        exit(EXIT_FAILURE);
    }

    /* Allocate structure that will contain valid quaternion time data */
    quaternion_time_data = (double *)malloc(sizeof(double) * quat_count);
    if (quaternion_time_data == NULL)
    {
        IAS_LOG_ERROR("Allocating quaternion time");
        free(quaternion_data);
        exit(EXIT_FAILURE);
    }

    /* Allocate array for quaternion quality flags */
    valid_quaternion_flag = (int *)malloc(sizeof(int) * quat_count);
    if (valid_quaternion_flag == NULL)
    {
        IAS_LOG_ERROR("Allocating quaternion quality flag");
        free(quaternion_data);
        free(quaternion_time_data);
        exit(EXIT_FAILURE);
    }

    status = ias_ancillary_identify_quaternion_outliers(cpf, l0r_attitude,
                 l0r_attitude_count, quaternion_time_data, quaternion_data,
                 valid_quaternion_flag, quat_count, &valid_quaternion_count, 
                 &interpolate_quaternions_flag);

    /* verify that the return value is correct */
    assert_equal(status, SUCCESS);
    /* no missing records, so the valid count should be the same as the
       original record count and no interpolation forced due to missing
       records */
    assert_equal(valid_quaternion_count, l0r_attitude_count);
    assert_equal(interpolate_quaternions_flag, 0);

    valid_quaternion_count = l0r_attitude_count;
    for (index = 0; index < valid_quaternion_count; index++)
    {
        if (!doubles_are_equal(quaternion_time_data[index],
                        output_quaternion_time_data[index]))
        {
            assert_double_equal(quaternion_time_data[index],
                  output_quaternion_time_data[index]);
        }

        if (!doubles_are_equal(quaternion_data[index].vector.x,
                        output_quaternion_data[index].vector.x))
        {
            assert_double_equal(quaternion_data[index].vector.x,
                  output_quaternion_data[index].vector.x);
        }

        if (!doubles_are_equal(quaternion_data[index].vector.y,
                        output_quaternion_data[index].vector.y))
        {
            assert_double_equal(quaternion_data[index].vector.y,
                  output_quaternion_data[index].vector.y);
        }

        if (!doubles_are_equal(quaternion_data[index].vector.z,
                        output_quaternion_data[index].vector.z))
        {
            assert_double_equal(quaternion_data[index].vector.z,
                  output_quaternion_data[index].vector.z);
        }

        if (!doubles_are_equal(quaternion_data[index].scalar,
                        output_quaternion_data[index].scalar))
        {
            assert_double_equal(quaternion_data[index].scalar,
                  output_quaternion_data[index].scalar);
        }

        if (valid_quaternion_flag[index] != output_valid_quaternion_flag[index])
        {
            assert_equal(valid_quaternion_flag[index],
                  output_valid_quaternion_flag[index]);
        }
    }

    free(quaternion_data);
    free(quaternion_time_data);
    free(valid_quaternion_flag);

    ias_cpf_free(cpf);
    cpf = NULL;
}

Ensure test_ias_convert_imu_to_acs()
{
    int status = 0;
    int index = 0;
    int error_count = 0;
    int imu_valid = 0;
    int imu_count;

    #include "test_headers/test_data_convert_imu_to_acs_in.h"
    #include "test_headers/test_data_convert_imu_to_acs_out.h"

    /* To test the IRU_Outlier_Threshold application, the threshold in the 
       CPF was tweaked to make a couple entries fail.  So, override the two
       expected failures to indicate they should fail.  If the input test
       data changes, this may also need to change. */
    output_valid_imu_flag[25593] = 0;
    output_valid_imu_flag[25594] = 0;

    /* Patch up the angular_rate_valid flag since the valid value changed from
       0x0F to 0xF0 after the test data was created.  If the test data is
       ever regenerated with the correct values, this block of code can be
       removed. */
    for (index = 0; index < l0r_imu_count; index++)
    {
        int gyro;
        for (gyro = 0; gyro < IAS_L0R_NUM_SAMPLES_PER_IMU_RECORD; gyro++)
        {
            if (l0r_imu[index].gyro_samples[gyro].angular_rate_valid == 0x0F)
                l0r_imu[index].gyro_samples[gyro].angular_rate_valid = 0xF0;
        }
    }

    /* allocate structure that will contain valid IMU */
    imu_data = (IAS_VECTOR *)malloc(sizeof(IAS_VECTOR)
            * l0r_imu_count * IAS_L0R_NUM_SAMPLES_PER_IMU_RECORD);
    if (imu_data == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for imu_data");
        exit(EXIT_FAILURE);
    }

    /* allocate structure that will contain valid IMU time stamps */
    imu_time_data = (double *)malloc(sizeof(double)
            * l0r_imu_count * IAS_L0R_NUM_SAMPLES_PER_IMU_RECORD);
    if (imu_time_data == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for imu_time_data");
        free(imu_data);
        exit(EXIT_FAILURE);
    }

    /* Allocate structure that will contain IMU quality flag. */
    valid_imu_flag = (int *)malloc(sizeof(int) * l0r_imu_count
            * IAS_L0R_NUM_SAMPLES_PER_IMU_RECORD);
    if (valid_imu_flag == NULL)
    {
        IAS_LOG_ERROR("Allocating IMU quality flag");
        free(imu_data);
        free(imu_time_data);
        exit(EXIT_FAILURE);
    }

    /* Allocate dummy structure for quaternion time data */
    double quaternion_time_data[200];

    status = ias_ancillary_convert_imu_to_acs(
                 cpf, l0r_imu, l0r_imu_count, 400, 
                 quaternion_time_data, imu_data, imu_time_data, valid_imu_flag,
                 &imu_valid, &imu_count);

    /* verify that the return value is correct */
    assert_equal(status, SUCCESS);

    /* test the output values here. */
    assert_equal(imu_count, output_imu_count);

    for (index = 0; index < imu_count; index++)
    {
        if (!doubles_are_equal(imu_time_data[index],
                        output_imu_time_data[index]))
        {
            assert_double_equal(imu_time_data[index],
                output_imu_time_data[index]);
            error_count++;
        }

        if (!doubles_are_equal(imu_data[index].x, output_imu_data[index].x))
        {
            assert_double_equal(imu_data[index].x, output_imu_data[index].x);
            error_count++;
        }

        if (!doubles_are_equal(imu_data[index].y, output_imu_data[index].y))
        {
            assert_double_equal(imu_data[index].y, output_imu_data[index].y);
            error_count++;
        }

        if (!doubles_are_equal(imu_data[index].z, output_imu_data[index].z))
        {
            assert_double_equal(imu_data[index].z, output_imu_data[index].z);
            error_count++;
        }

        if (valid_imu_flag[index] != output_valid_imu_flag[index])
        {
            assert_equal(valid_imu_flag[index], output_valid_imu_flag[index]);
            error_count++;
        }

        if (error_count > 1000)
        {
            printf("Too many errors - cutting test short\n");
            break;
        }
    }

    free(imu_data);
    free(imu_time_data);
    free(valid_imu_flag);

    ias_cpf_free(cpf);
    cpf = NULL;
}

Ensure test_ias_extract_valid_imu_data_window()
{
    int status = 0;
    int index = 0;

    #include "test_headers/test_data_extract_valid_imu_data_window_in.h"
    #include "test_headers/test_data_extract_valid_imu_data_window_out.h"

    status = ias_ancillary_extract_valid_imu_data_window(
                 anc_ephemeris_data, imu_time_data, imu_data,
                 valid_imu_flag, &imu_count, imu_epoch, &eph_start_time,
                 &eph_stop_time, &invalid_imu_count);

    /* verify that the return value is correct */
    assert_equal(status, SUCCESS);

    /* test the output values here. */
    assert_equal(imu_count, output_imu_count);
    assert_equal(invalid_imu_count, output_invalid_imu_count);

    for (index = 0; index < imu_count; index++)
    {
        if (!doubles_are_equal(imu_time_data[index],
                        output_imu_time_data[index]))
        {
            assert_double_equal(imu_time_data[index],
                output_imu_time_data[index]);
        }

        if (!doubles_are_equal(imu_data[index].x, output_imu_data[index].x))
            assert_double_equal(imu_data[index].x, output_imu_data[index].x);

        if (!doubles_are_equal(imu_data[index].y, output_imu_data[index].y))
            assert_double_equal(imu_data[index].y, output_imu_data[index].y);

        if (!doubles_are_equal(imu_data[index].z, output_imu_data[index].z))
            assert_double_equal(imu_data[index].z, output_imu_data[index].z);

        if (valid_imu_flag[index] != output_valid_imu_flag[index])
            assert_equal(valid_imu_flag[index], output_valid_imu_flag[index]);
    }

    assert_equal(imu_epoch[0], output_imu_epoch[0]);
    assert_equal(imu_epoch[1], output_imu_epoch[1]);
    assert_equal(imu_epoch[2], output_imu_epoch[2]);

    assert_equal(eph_start_time, output_eph_start_time);
    assert_equal(eph_stop_time, output_eph_stop_time);

    /* this was malloc'd in the header file */
    free(anc_ephemeris_data);

    ias_cpf_free(cpf);
    cpf = NULL;
}

Ensure test_ias_extract_valid_quaternion_window()
{
    int status = 0;
    int index = 0;

    #include "test_headers/test_data_extract_valid_quaternion_window_in.h"
    #include "test_headers/test_data_extract_valid_quaternion_window_out.h"

    status = ias_ancillary_extract_valid_quaternion_window(
                 imu_time_data, imu_count, quaternion_data,
                 quaternion_time_data, valid_quaternion_flag,
                 &valid_quaternion_count, &invalid_quaternion_count);

    /* verify that the return value is correct */
    assert_equal(status, SUCCESS);

    /* test the output values here. */
    assert_equal(valid_quaternion_count, output_valid_quaternion_count);
    assert_equal(invalid_quaternion_count, output_invalid_quaternion_count);

    for (index = 0; index < valid_quaternion_count; index++)
    {
        if (!doubles_are_equal(quaternion_time_data[index],
                        output_quaternion_time_data[index]))
        {
            assert_double_equal(quaternion_time_data[index],
                  output_quaternion_time_data[index]);
        }

        if (!doubles_are_equal(quaternion_data[index].vector.x,
                        output_quaternion_data[index].vector.x))
        {
            assert_double_equal(quaternion_data[index].vector.x,
                  output_quaternion_data[index].vector.x);
        }

        if (!doubles_are_equal(quaternion_data[index].vector.y,
                        output_quaternion_data[index].vector.y))
        {
            assert_double_equal(quaternion_data[index].vector.y,
                  output_quaternion_data[index].vector.y);
        }

        if (!doubles_are_equal(quaternion_data[index].vector.z,
                        output_quaternion_data[index].vector.z))
        {
            assert_double_equal(quaternion_data[index].vector.z,
                  output_quaternion_data[index].vector.z);
        }

        if (!doubles_are_equal(quaternion_data[index].scalar,
                        output_quaternion_data[index].scalar))
        {
            assert_double_equal(quaternion_data[index].scalar,
                  output_quaternion_data[index].scalar);
        }

        if (valid_quaternion_flag[index] != output_valid_quaternion_flag[index])
        {
            assert_equal(valid_quaternion_flag[index],
                  output_valid_quaternion_flag[index]);
        }
    }

    ias_cpf_free(cpf);
    cpf = NULL;
}

Ensure test_ias_compute_spacecraft_attitude()
{
    int status = 0;
    int index = 0;
    double image_start_time[3];
    double image_stop_time[3];
    int interpolate_quaternions_flag = 0;

    #include "test_headers/test_data_compute_spacecraft_attitude_in.h"
    #include "test_headers/test_data_compute_spacecraft_attitude_out.h"

    /* Make sure the image start/stop times are inside the data */
    image_start_time[0] = anc_ephemeris_data->utc_epoch_time[0];
    image_start_time[1] = anc_ephemeris_data->utc_epoch_time[1];
    image_start_time[2] = anc_ephemeris_data->utc_epoch_time[2] + 9;
    image_stop_time[0] = anc_ephemeris_data->utc_epoch_time[0];
    image_stop_time[1] = anc_ephemeris_data->utc_epoch_time[1];
    image_stop_time[2] = anc_ephemeris_data->utc_epoch_time[2] + 11;

    status = ias_ancillary_compute_spacecraft_attitude(
                 anc_ephemeris_data, &imu_count, quaternion_data,
                 valid_quaternion_count, eph_start_time, eph_stop_time,
                 image_start_time, image_stop_time, acq_type, 1,
                 interpolate_quaternions_flag, imu_epoch, imu_time_data,
                 quaternion_time_data, valid_imu_flag, valid_quaternion_flag,
                 imu_data, &imu_start_time);

    /* verify that the return value is correct */
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
        return;

    /* test the output values here. */
    assert_equal(imu_start_time, output_imu_start_time);

    for (index = 0; index < valid_quaternion_count; index++)
    {
        if (valid_quaternion_flag[index] != output_valid_quaternion_flag[index])
        {
            assert_equal(valid_quaternion_flag[index],
                         output_valid_quaternion_flag[index]);
        }
    }

    for (index = 0; index < imu_count; index++)
    {
        /* round the imu_time_data to 6 significant digits so that it matches
           the %lf format used for generating the output_imu_time_data
           values */
        imu_time_data[index] =
            ((long int)(imu_time_data[index] * 1.0e6 + 0.5)) * 1.0e-6;
        if (!doubles_are_equal(imu_time_data[index],
                        output_imu_time_data[index]))
        {
            printf("\nimu_time_data[%d]: %.18e %.18e\n", index,
                   imu_time_data[index], output_imu_time_data[index]);
            assert_double_equal(imu_time_data[index],
                output_imu_time_data[index]);
        }

        if (!doubles_are_equal(imu_data[index].x, output_imu_data[index].x))
        {
            printf("\nimu data x[%d]: %.18e %.18e\n", index,
                   imu_data[index].x, output_imu_data[index].x);
            assert_double_equal(imu_data[index].x, output_imu_data[index].x);
        }

        if (!doubles_are_equal(imu_data[index].y, output_imu_data[index].y))
        {
            printf("\nimu data y[%d]: %.18e %.18e\n", index,
                   imu_data[index].y, output_imu_data[index].y);
            assert_double_equal(imu_data[index].y, output_imu_data[index].y);
        }

        if (!doubles_are_equal(imu_data[index].z, output_imu_data[index].z))
        {
            printf("\nimu data z[%d]: %.18e %.18e\n", index,
                   imu_data[index].z, output_imu_data[index].z);
            assert_double_equal(imu_data[index].z, output_imu_data[index].z);
        }
    }

    /* Confirm that the data not covering the image is caught as an error */
    image_start_time[0] = anc_ephemeris_data->utc_epoch_time[0];
    image_start_time[1] = anc_ephemeris_data->utc_epoch_time[1];
    image_start_time[2] = anc_ephemeris_data->utc_epoch_time[2] + 2;
    image_stop_time[0] = anc_ephemeris_data->utc_epoch_time[0];
    image_stop_time[1] = anc_ephemeris_data->utc_epoch_time[1];
    image_stop_time[2] = anc_ephemeris_data->utc_epoch_time[2] + 11;

    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = ias_ancillary_compute_spacecraft_attitude(
                 anc_ephemeris_data, &imu_count, quaternion_data,
                 valid_quaternion_count, eph_start_time, eph_stop_time,
                 image_start_time, image_stop_time, acq_type, 1,
                 interpolate_quaternions_flag, imu_epoch, imu_time_data,
                 quaternion_time_data, valid_imu_flag, valid_quaternion_flag,
                 imu_data, &imu_start_time);
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

    /* verify that the return value is correct */
    assert_equal(status, ERROR);

    /* this was malloc'd in the header file */
    free(anc_ephemeris_data);

    ias_cpf_free(cpf);
    cpf = NULL;
}

Ensure test_ias_convert_imu_to_attitude()
{
    int status = 0;
    int index_1 = 0;
    int index_2 = 0;

    #include "test_headers/test_data_convert_imu_to_attitude_in.h"
    #include "test_headers/test_data_convert_imu_to_attitude_out.h"

    /* Allocate ancillary data structure */
    anc_attitude_data = ias_ancillary_allocate_attitude(imu_count);
    if (anc_attitude_data == NULL)
    {
        IAS_LOG_ERROR("Allocating ancillary attitude records");
        exit(EXIT_FAILURE);
    }

    status = ias_ancillary_convert_imu_to_attitude( cpf, IAS_EARTH,
                 anc_ephemeris_data, imu_epoch, imu_data, imu_time_data,
                 imu_count, imu_start_time, anc_attitude_data);

    /* verify that the return value is correct */
    assert_equal(status, SUCCESS);

    /* test the output values here. */
    assert_double_equal(anc_attitude_data->utc_epoch_time[0],
                 output_anc_attitude_data->utc_epoch_time[0]);
    assert_double_equal(anc_attitude_data->utc_epoch_time[1],
                 output_anc_attitude_data->utc_epoch_time[1]);
    assert_double_equal(anc_attitude_data->utc_epoch_time[2],
                 output_anc_attitude_data->utc_epoch_time[2]);

    assert_equal(anc_attitude_data->number_of_samples,
                 output_anc_attitude_data->number_of_samples);

    for (index_1 = 0; index_1 < anc_attitude_data->number_of_samples;
         index_1++)
    {
        if (!(index_1 % 20))
        {
            if (!doubles_are_equal(
                anc_attitude_data->records[index_1].seconds_from_epoch,
                output_anc_attitude_data->records[index_1].seconds_from_epoch))
            {
                assert_double_equal(
                 anc_attitude_data->records[index_1].seconds_from_epoch,
                 output_anc_attitude_data->records[index_1].seconds_from_epoch);
            }

            for (index_2 = 0; index_2 < 4; index_2++)
            {
                if (!doubles_are_equal(
                    anc_attitude_data->records[index_1].eci_quaternion[index_2],
                    output_anc_attitude_data->records[index_1]
                        .eci_quaternion[index_2]))
                {
                    assert_double_equal(
                    anc_attitude_data->records[index_1].eci_quaternion[index_2],
                    output_anc_attitude_data->records[index_1].
                        eci_quaternion[index_2]);
                }
            }
    
            for (index_2 = 0; index_2 < 4; index_2++)
            {
                if (!doubles_are_equal(
                   anc_attitude_data->records[index_1].ecef_quaternion[index_2],
                   output_anc_attitude_data->records[index_1]
                    .ecef_quaternion[index_2]))
                {
                    assert_double_equal(
                        anc_attitude_data->records[index_1]
                            .ecef_quaternion[index_2],
                        output_anc_attitude_data->records[index_1]
                            .ecef_quaternion[index_2]);
                }
            }
    
            if (!doubles_are_equal(anc_attitude_data->records[index_1].roll,
                output_anc_attitude_data->records[index_1].roll))
            {
                assert_double_equal(anc_attitude_data->records[index_1].roll,
                    output_anc_attitude_data->records[index_1].roll);
            }
            if (!doubles_are_equal(
                 anc_attitude_data->records[index_1].roll_rate,
                 output_anc_attitude_data->records[index_1].roll_rate))
            {
                assert_double_equal(
                   anc_attitude_data->records[index_1].roll_rate,
                    output_anc_attitude_data->records[index_1].roll_rate);
            }

            if (!doubles_are_equal(anc_attitude_data->records[index_1].pitch,
                output_anc_attitude_data->records[index_1].pitch))
            {
                assert_double_equal(anc_attitude_data->records[index_1].pitch,
                    output_anc_attitude_data->records[index_1].pitch);
            }
            if (!doubles_are_equal(
                   anc_attitude_data->records[index_1].pitch_rate,
                   output_anc_attitude_data->records[index_1].pitch_rate))
            {
                assert_double_equal(
                    anc_attitude_data->records[index_1].pitch_rate,
                    output_anc_attitude_data->records[index_1].pitch_rate);

            }
            if (!doubles_are_equal(anc_attitude_data->records[index_1].yaw,
                output_anc_attitude_data->records[index_1].yaw))
            {
                assert_double_equal(anc_attitude_data->records[index_1].yaw,
                    output_anc_attitude_data->records[index_1].yaw);
            }
            if (!doubles_are_equal(anc_attitude_data->records[index_1].yaw_rate,
                output_anc_attitude_data->records[index_1].yaw_rate))
            {
                assert_double_equal(
                    anc_attitude_data->records[index_1].yaw_rate,
                    output_anc_attitude_data->records[index_1].yaw_rate);
            }
        }
    }

    free(anc_attitude_data);

    /* this was malloc'd in the header file */
    free(anc_ephemeris_data);
    free(output_anc_attitude_data);

    ias_cpf_free(cpf);
    cpf = NULL;
}

Ensure test_ias_smooth_ephemeris()
{
    int status = 0;
    int index = 0;

    #include "test_headers/test_data_smooth_ephemeris_in.h"
    #include "test_headers/test_data_smooth_ephemeris_out.h"

    /* Allocate memory for the ephemeris seconds */
    ephemeris_seconds_since_j2000 =
        (double *)malloc(sizeof(double) * l0r_ephemeris_count);
    if (ephemeris_seconds_since_j2000 == NULL)
    {
        IAS_LOG_ERROR("Allocating ephemeris_seconds_since_j2000");
        exit(EXIT_FAILURE);
    }

    smoothed_eph_pos =
        (IAS_VECTOR *)malloc(sizeof(IAS_VECTOR) * l0r_ephemeris_count);
    if (smoothed_eph_pos == NULL)
    {
        IAS_LOG_ERROR("Allocating smoothed_eph_pos");
        free(ephemeris_seconds_since_j2000);
        exit(EXIT_FAILURE);
    }

    smoothed_eph_vel =
        (IAS_VECTOR *)malloc(sizeof(IAS_VECTOR) * l0r_ephemeris_count);
    if (smoothed_eph_vel == NULL)
    {
        IAS_LOG_ERROR("Allocating smoothed_eph_vel");
        free(ephemeris_seconds_since_j2000);
        free(smoothed_eph_pos);
        exit(EXIT_FAILURE);
    }

    status = ias_ancillary_smooth_ephemeris(
                 IAS_EARTH, cpf, l0r_ephemeris, l0r_ephemeris_count,
                 interval_start_time, interval_stop_time,
                 &valid_ephemeris_count, &invalid_ephemeris_count,
                 ephemeris_seconds_since_j2000,
                 smoothed_eph_pos, smoothed_eph_vel);

    /* verify that the return value is correct */
    assert_equal(status, SUCCESS);

    /* test the output values here. */
    assert_equal(valid_ephemeris_count, o_valid_ephemeris_count);
    assert_equal(invalid_ephemeris_count, o_invalid_ephemeris_count);

    /* test the output data */
    for (index = 0; index < valid_ephemeris_count; index++)
    {
        if (!doubles_are_equal(ephemeris_seconds_since_j2000[index],
                             o_ephemeris_seconds_since_j2000[index]))
        {
            assert_double_equal(ephemeris_seconds_since_j2000[index],
                       o_ephemeris_seconds_since_j2000[index]);
        }
        if (!doubles_are_equal(smoothed_eph_pos[index].x,
                             o_smoothed_eph_pos[index].x))
        {
            assert_double_equal(smoothed_eph_pos[index].x,
                       o_smoothed_eph_pos[index].x);
        }
        if (!doubles_are_equal(smoothed_eph_pos[index].y,
                             o_smoothed_eph_pos[index].y))
        {
            assert_double_equal(smoothed_eph_pos[index].y,
                       o_smoothed_eph_pos[index].y);
        }
        if (!doubles_are_equal(smoothed_eph_pos[index].z,
                             o_smoothed_eph_pos[index].z))
        {
            assert_double_equal(smoothed_eph_pos[index].z,
                       o_smoothed_eph_pos[index].z);
        }
        if (!doubles_are_equal(smoothed_eph_vel[index].x,
                             o_smoothed_eph_vel[index].x))
        {
            assert_double_equal(smoothed_eph_vel[index].x,
                       o_smoothed_eph_vel[index].x);
        }
        if (!doubles_are_equal(smoothed_eph_vel[index].y,
                             o_smoothed_eph_vel[index].y))
        {
            assert_double_equal(smoothed_eph_vel[index].y,
                       o_smoothed_eph_vel[index].y);
        }
        if (!doubles_are_equal(smoothed_eph_vel[index].z,
                             o_smoothed_eph_vel[index].z))
        {
            assert_double_equal(smoothed_eph_vel[index].z,
                       o_smoothed_eph_vel[index].z);
        }
    } /* END - index for loop */

    free(ephemeris_seconds_since_j2000);
    free(smoothed_eph_pos);
    free(smoothed_eph_vel);

    ias_cpf_free(cpf);
    cpf = NULL;
}

Ensure test_ias_build_ephemeris()
{
    int status = 0;
    int index_1 = 0;
    int index_2 = 0;

    #include "test_headers/test_data_build_ephemeris_in.h"
    #include "test_headers/test_data_build_ephemeris_out.h"

    /* Allocate the ancillary data structure */
    anc_ephemeris_data =
        ias_ancillary_allocate_ephemeris(valid_ephemeris_count);
    if (anc_ephemeris_data == NULL)
    {
        IAS_LOG_ERROR("Allocating ancillary ephemeris records");
        exit(EXIT_FAILURE);
    }

    status = ias_ancillary_build_ephemeris(
                 cpf, valid_ephemeris_count, ephemeris_seconds_since_j2000,
                 smoothed_eph_pos, smoothed_eph_vel, anc_ephemeris_data);

    /* verify that the return value is correct */
    assert_equal(status, SUCCESS);

    /* test the output values here. */
    assert_double_equal(anc_ephemeris_data->utc_epoch_time[0],
               o_anc_ephemeris_data->utc_epoch_time[0]);
    assert_double_equal(anc_ephemeris_data->utc_epoch_time[1],
               o_anc_ephemeris_data->utc_epoch_time[1]);
    assert_double_equal(anc_ephemeris_data->utc_epoch_time[2],
               o_anc_ephemeris_data->utc_epoch_time[2]);

    assert_equal(anc_ephemeris_data->number_of_samples,
               o_anc_ephemeris_data->number_of_samples);

    for (index_1 = 0;
         index_1 < anc_ephemeris_data->number_of_samples;
         index_1++)
    {
        if (!doubles_are_equal(
           anc_ephemeris_data->records[index_1].seconds_from_epoch,
           o_anc_ephemeris_data->records[index_1].seconds_from_epoch))
        {
            assert_double_equal(
                anc_ephemeris_data->records[index_1].seconds_from_epoch,
                o_anc_ephemeris_data->records[index_1].seconds_from_epoch);
        }

        for (index_2 = 0; index_2 < 3; index_2++)
        {
            if (!doubles_are_equal(
                anc_ephemeris_data->records[index_1].eci_position[index_2],
                o_anc_ephemeris_data->records[index_1].eci_position[index_2]))
            {
                assert_double_equal(
                  anc_ephemeris_data->records[index_1].eci_position[index_2],
                  o_anc_ephemeris_data->records[index_1].eci_position[index_2]);
            }

            if (!doubles_are_equal(
                anc_ephemeris_data->records[index_1].eci_velocity[index_2],
                o_anc_ephemeris_data->records[index_1].eci_velocity[index_2]))
            {
                assert_double_equal(
                  anc_ephemeris_data->records[index_1].eci_velocity[index_2],
                  o_anc_ephemeris_data->records[index_1].eci_velocity[index_2]);
            }

            if (!doubles_are_equal(
                anc_ephemeris_data->records[index_1].ecef_position[index_2],
                o_anc_ephemeris_data->records[index_1].ecef_position[index_2]))
            {
                assert_double_equal(
                 anc_ephemeris_data->records[index_1].ecef_position[index_2],
                 o_anc_ephemeris_data->records[index_1].ecef_position[index_2]);
            }

            if (!doubles_are_equal(
                anc_ephemeris_data->records[index_1].ecef_velocity[index_2],
                o_anc_ephemeris_data->records[index_1].ecef_velocity[index_2]))
            {
                assert_double_equal(
                 anc_ephemeris_data->records[index_1].ecef_velocity[index_2],
                 o_anc_ephemeris_data->records[index_1].ecef_velocity[index_2]);
            }
        } /* END - index_2 for loop */
    } /* END - index_1 for loop */

    free(anc_ephemeris_data);

    /* this was malloc'd in the header file */
    free(o_anc_ephemeris_data);

    ias_cpf_free(cpf);
    cpf = NULL;
}

int initialize()
{
    double init_time_seconds = 0.0;
    const struct IAS_CPF_EARTH_CONSTANTS *earth_constants = NULL;
    int status;
    char *test_data_dir;           /* Holds test data dir env var */
    char cpf_filename[PATH_MAX];   /* CPF filename */

    /* Initialize the satellite attributes library */
    status = ias_sat_attr_initialize(IAS_L8);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing sat_attr");
        return ERROR;
    }

    /* Get IAS_UNIT_TEST_DATA_DIR environment variable */
    test_data_dir = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (test_data_dir == NULL)
    {
        IAS_LOG_ERROR("Getting IAS_UNIT_TEST_DATA_DIR env variable");
        return ERROR;
    }
    else
    {
        status = snprintf(cpf_filename, sizeof(cpf_filename), 
            "%s%s", test_data_dir, CPF);
        if (status >= sizeof(cpf_filename))
        {
            IAS_LOG_ERROR("Test data dir string truncated by snprintf");
            return ERROR;
        }
    }

    cpf = ias_cpf_read(cpf_filename);
    if (cpf == NULL)
    {
        IAS_LOG_ERROR("Reading CPF");
        return ERROR;
    }

    /* Get the CPF earth constants */
    earth_constants = ias_cpf_get_earth_const(cpf);
    if (earth_constants == NULL)
    {
        IAS_LOG_ERROR("Reading CPF earth constants");
        ias_cpf_free(cpf);
        return ERROR;
    }

    /* Initialize the leap seconds */
    init_time_seconds = 107417131.550800; /* from the calval test data */
    if(ias_math_init_leap_seconds(init_time_seconds, 
       &earth_constants->leap_seconds_data))
    {
        IAS_LOG_ERROR("Computing leap second conversion for l0r_header");
        return ERROR;
    }

    return SUCCESS;
}

/* Re-enable the command line specified optimization */
IAS_RESTORE_OPTIMIZATION

int main(int argc, char **argv)
{
    int status = 0;
    char prog_name[PATH_MAX];          /* Name of the program */

    /* Get the name of the program */  
    strcpy(prog_name, argv[0]);

    /* Initialize the IAS logging library */
    status = ias_log_initialize(prog_name);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing log");
        exit(EXIT_FAILURE);
    }

    if (initialize() != SUCCESS)
    {
        exit(EXIT_FAILURE);
    }

    TestSuite *suite = create_test_suite();

    add_test(suite, test_ias_smooth_ephemeris);
    add_test(suite, test_ias_build_ephemeris);
    add_test(suite, test_ias_identify_quaternion_outliers);
    add_test(suite, test_ias_convert_imu_to_acs);
    add_test(suite, test_ias_extract_valid_imu_data_window);
    add_test(suite, test_ias_extract_valid_quaternion_window);
    add_test(suite, test_ias_compute_spacecraft_attitude);
    add_test(suite, test_ias_convert_imu_to_attitude);

    if (argc > 1)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }

    status = run_test_suite(suite, create_text_reporter());

    ias_cpf_free(cpf);
    cpf = NULL;

    return status;
}

