#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>              /* PATH_MAX definition */

#include "ias_gcc_pragmas.h"
#include "ias_types.h"
#include "ias_logging.h"
#include "ias_cpf.h"
#include "ias_math.h"
#include "ias_ancillary.h"
#include "ias_ancillary_private.h"

#include "cgreen.h"

IAS_CPF *cpf = NULL;
#define CPF "/ias_lib/ancillary/L8CPF20090101_20090331.01"

/* Disable the command line specified optimization
   It is re-enabled below
   We disable the optimization because the inclusion of headers which contain
   structures defining huge amounts of testing data, cause very very very long
   compiles.  And it is the optimization in the newer compilers which cause
   the lengthy compile time with the include files. */
IAS_DISABLE_OPTIMIZATION

Ensure test_ias_convert_imu_to_acs_anomaly_handling()
{
    int index = 0;
    int imu_valid = 0;
    IAS_L0R_IMU *corrupt_imu = NULL;
    IAS_L0R_IMU *fixed_imu = NULL;
    int last_record_index;

    #include "test_headers/test_data_convert_imu_to_acs_in.h"

    /* To test for the spacecraft anomalies, corrupt two records of the data
       with the known patterns */
    corrupt_imu = malloc(l0r_imu_count * sizeof(*corrupt_imu));
    if (corrupt_imu == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for corrupt_imu");
        exit(EXIT_FAILURE);
    }
    memcpy(corrupt_imu, l0r_imu, l0r_imu_count * sizeof(*corrupt_imu));

    /* Corrupt record 4 with a type 2 anomaly */
    corrupt_imu[3].gyro_samples[0].time_tag 
        = corrupt_imu[2].gyro_samples[0].time_tag;
    corrupt_imu[3].gyro_samples[1].time_tag 
        = corrupt_imu[2].gyro_samples[1].time_tag;

    /* Corrupt record 8 with a type 3 anomaly */
    last_record_index = IAS_L0R_NUM_SAMPLES_PER_IMU_RECORD - 1;
    corrupt_imu[7].gyro_samples[last_record_index].time_tag 
        = corrupt_imu[6].gyro_samples[last_record_index].time_tag;
    corrupt_imu[7].gyro_samples[last_record_index - 1].time_tag 
        = corrupt_imu[6].gyro_samples[last_record_index - 1].time_tag;

    /* allocate structure that will contain valid IMU */
    imu_data = (IAS_VECTOR *)malloc(sizeof(IAS_VECTOR)
            * l0r_imu_count * IAS_L0R_NUM_SAMPLES_PER_IMU_RECORD);
    if (imu_data == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for imu_data");
        free(corrupt_imu);
        exit(EXIT_FAILURE);
    }

    /* allocate structure that will contain valid IMU time stamps */
    imu_time_data = (double *)malloc(sizeof(double)
            * l0r_imu_count * IAS_L0R_NUM_SAMPLES_PER_IMU_RECORD);
    if (imu_time_data == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for imu_time_data");
        free(imu_data);
        free(corrupt_imu);
        exit(EXIT_FAILURE);
    }

    /* Allocate structure that will contain IMU quality flag. */
    valid_imu_flag = (int *)malloc(sizeof(int) * l0r_imu_count
            * IAS_L0R_NUM_SAMPLES_PER_IMU_RECORD);
    if (valid_imu_flag == NULL)
    {
        IAS_LOG_ERROR("Allocating IMU quality flag");
        free(imu_data);
        free(imu_time_data);
        free(corrupt_imu);
        exit(EXIT_FAILURE);
    }

    /* Process the IMU times */
    fixed_imu = ias_ancillary_process_imu_times(corrupt_imu, 4.0e-6,
                    l0r_imu_count, &imu_valid, imu_time_data);

    /* Confirm a fixed array of IMU data was returned */
    assert_true(fixed_imu != NULL);
    if (fixed_imu == NULL)
    {
        free(imu_data);
        free(imu_time_data);
        free(valid_imu_flag);
        free(corrupt_imu);
        exit(EXIT_FAILURE);
    }
    assert_equal(imu_valid, 1);

    /* Confirm the 4th record has the first 7 samples invalidated and the
       8th record has the last 45 samples invalidated. */
    for (index = 0; index <= last_record_index; index++)
    {
        if (index < 7)
        {
            assert_equal(fixed_imu[3].gyro_samples[index].angular_rate_valid,
                0);
        }
        else
        {
            assert_not_equal(fixed_imu[3].gyro_samples[index]
                .angular_rate_valid, 0);
        }

        if (index > 4)
        {
            assert_equal(fixed_imu[7].gyro_samples[index].angular_rate_valid,
                0);
        }
        else
        {
            assert_not_equal(fixed_imu[7].gyro_samples[index]
                .angular_rate_valid, 0);
        }
    }

    /* Free the fixed array */
    free(fixed_imu);
    fixed_imu = NULL;

    /* Now make record index 9 record disappear from the IMU data to test
       for the anomaly where an entire record is missed */
    memcpy(corrupt_imu, l0r_imu, 8 * sizeof(*corrupt_imu));
    memcpy(&corrupt_imu[8], &l0r_imu[9],
        (l0r_imu_count - 9) * sizeof(*corrupt_imu));

    /* make sure an array of fixed data is still returned, the valid flag
       is not set */
    fixed_imu = ias_ancillary_process_imu_times(corrupt_imu, 4.0e-6,
        l0r_imu_count - 1, &imu_valid, imu_time_data);

    /* Confirm the imu data was flagged as invalid and pointer to an array of
       fixed array of IMU data was returned */
    assert_equal(imu_valid, 0);
    assert_true(fixed_imu != NULL);
    if (fixed_imu == NULL)
    {
        free(imu_data);
        free(imu_time_data);
        free(corrupt_imu);
        free(valid_imu_flag);
        exit(EXIT_FAILURE);
    }
    free(fixed_imu);
    fixed_imu = NULL;

    free(imu_data);
    free(imu_time_data);
    free(valid_imu_flag);
    free(corrupt_imu);

    ias_cpf_free(cpf);
    cpf = NULL;
}


static void verify_quaternions
(
    const double *times,
    const IAS_QUATERNION *quats,
    const int *valid_flags,
    const double *expected_times,
    const IAS_QUATERNION *expected_quats,
    const int *expected_valid_flags,
    int valid_quaternion_count
)
{
    int index;

    for (index = 0; index < valid_quaternion_count; index++)
    {
        if (!doubles_are_equal(times[index], expected_times[index]))
        {
            assert_double_equal(times[index], expected_times[index]);
        }

        if (!doubles_are_equal(quats[index].vector.x,
                        expected_quats[index].vector.x))
        {
            assert_double_equal(quats[index].vector.x,
                  expected_quats[index].vector.x);
        }

        if (!doubles_are_equal(quats[index].vector.y,
                        expected_quats[index].vector.y))
        {
            assert_double_equal(quats[index].vector.y,
                  expected_quats[index].vector.y);
        }

        if (!doubles_are_equal(quats[index].vector.z,
                        expected_quats[index].vector.z))
        {
            assert_double_equal(quats[index].vector.z,
                  expected_quats[index].vector.z);
        }

        if (!doubles_are_equal(quats[index].scalar,
                        expected_quats[index].scalar))
        {
            assert_double_equal(quats[index].scalar,
                  expected_quats[index].scalar);
        }

        if (valid_flags[index] != expected_valid_flags[index])
        {
            assert_equal(valid_flags[index], expected_valid_flags[index]);
        }
    }
}

Ensure test_ias_identify_quaternion_outliers_anomaly_handling()
{
    int status = 0;
    int index = 0;
    IAS_L0R_ATTITUDE *corrupt_att = NULL;
    int samples_per_record = IAS_L0R_NUM_SAMPLES_PER_IMU_RECORD;
    int quat_count;
    int interpolate_quaternions_flag = 0;
    IAS_QUATERNION *modified_output = NULL;
    int *modified_valid_flag = NULL;

    /* reuse the test data from the normal unit test for this routine */
    #include "test_headers/test_data_identify_quaternion_outliers_in.h"
    #include "test_headers/test_data_identify_quaternion_outliers_out.h"

    /* Set the number of records to allocate for the quaternion buffers.
       Normally, we would allocate an extra IAS_L0R_NUM_SAMPLES_PER_IMU_RECORD
       samples, but since we reduce the number of L0R records by that much when
       setting up the test data later in the routine, use the L0R attitude
       count gets the extra records needed */
    quat_count = l0r_attitude_count;

    /* Allocate structure that will contain valid quaternion data */
    quaternion_data = (IAS_QUATERNION *)malloc(sizeof(IAS_QUATERNION)
        * quat_count);
    if (quaternion_data == NULL)
    {
        IAS_LOG_ERROR("Allocating quaternions");
        exit(EXIT_FAILURE);
    }

    /* Allocate structure that will contain valid quaternion time data */
    quaternion_time_data = (double *)malloc( sizeof(double) * quat_count );
    if (quaternion_time_data == NULL)
    {
        IAS_LOG_ERROR("Allocating quaternion time");
        free(quaternion_data);
        exit(EXIT_FAILURE);
    }

    /* Allocate array for quaternion quality flags */
    valid_quaternion_flag = (int *)malloc(sizeof(int) * quat_count);
    if (valid_quaternion_flag == NULL)
    {
        IAS_LOG_ERROR("Allocating quaternion quality flag");
        free(quaternion_data);
        free(quaternion_time_data);
        exit(EXIT_FAILURE);
    }

    /* Allocate array to create a corrupted copy of the attitude data */
    corrupt_att = malloc(sizeof(*corrupt_att) * l0r_attitude_count);
    if (corrupt_att == NULL)
    {
        IAS_LOG_ERROR("Allocating quaternion quality flag");
        free(quaternion_data);
        free(quaternion_time_data);
        free(valid_quaternion_flag);
        exit(EXIT_FAILURE);
    }
    memcpy(corrupt_att, l0r_attitude, 
        sizeof(*corrupt_att) * l0r_attitude_count);
    
    /* The anomalies are detected using the times, so corrupt those in the
       L0R.  First create a type 2 anomaly in the 4th record index. */
    index = 5 * samples_per_record - 1;
    corrupt_att[index].l0r_time
        = corrupt_att[index - samples_per_record].l0r_time;
    corrupt_att[index - 1].l0r_time
        = corrupt_att[index - samples_per_record - 1].l0r_time;

    /* Corrupt the 8th record index with a type 4 anomaly */
    index = 8 * samples_per_record;
    corrupt_att[index].l0r_time
        = corrupt_att[index - samples_per_record].l0r_time;
    corrupt_att[index + 1].l0r_time
        = corrupt_att[index - samples_per_record + 1].l0r_time;

    /* Allocate space for a copy of the expected output data that will be
       modified */
    modified_output = malloc(sizeof(*modified_output) * l0r_attitude_count);
    if (modified_output == NULL)
    {
        IAS_LOG_ERROR("Allocating quaternion quality flag");
        free(corrupt_att);
        free(quaternion_data);
        free(quaternion_time_data);
        free(valid_quaternion_flag);
        exit(EXIT_FAILURE);
    }
    memcpy(modified_output, output_quaternion_data,
        sizeof(*modified_output) * l0r_attitude_count);

    modified_valid_flag = malloc(sizeof(*modified_valid_flag) 
        * l0r_attitude_count);
    if (modified_valid_flag == NULL)
    {
        IAS_LOG_ERROR("Allocating quaternion quality flag");
        free(modified_output);
        free(corrupt_att);
        free(quaternion_data);
        free(quaternion_time_data);
        free(valid_quaternion_flag);
        exit(EXIT_FAILURE);
    }
    memcpy(modified_valid_flag, output_valid_quaternion_flag,
        sizeof(*modified_valid_flag) * l0r_attitude_count);

    /* Update the expected output with the values expected for the corrupted
       samples.  First the last 22 samples of the 4th record index. */
    for (index = 4 * samples_per_record + 28; index < 5 * samples_per_record;
         index++)
    {
        modified_output[index].vector.x = 0.0;
        modified_output[index].vector.y = 0.0;
        modified_output[index].vector.z = 0.0;
        modified_output[index].scalar = 0.0;
        modified_valid_flag[index] = 0;
    }
    /* Now the first 29 samples of record index 8 */
    for (index = 8 * samples_per_record; index < 8 * samples_per_record + 29;
         index++)
    {
        modified_output[index].vector.x = 0.0;
        modified_output[index].vector.y = 0.0;
        modified_output[index].vector.z = 0.0;
        modified_output[index].scalar = 0.0;
        modified_valid_flag[index] = 0;
    }

    /* Call the routine with the corrupted data */
    status = ias_ancillary_identify_quaternion_outliers(cpf, corrupt_att,
                 l0r_attitude_count, quaternion_time_data, quaternion_data,
                 valid_quaternion_flag, quat_count, &valid_quaternion_count,
                 &interpolate_quaternions_flag);

    /* verify that the return value is correct */
    assert_equal(status, SUCCESS);
    /* no quaternions should have been inserted */
    assert_equal(valid_quaternion_count, l0r_attitude_count);
    assert_equal(interpolate_quaternions_flag, 0);

    verify_quaternions(quaternion_time_data, quaternion_data, 
        valid_quaternion_flag, output_quaternion_time_data,
        modified_output, modified_valid_flag, valid_quaternion_count);

    /* Now test the case where an entire record is missing by making record
       index 8 disappear from the data. Allocate the corrupt_att buffer with 
       a full record fewer entries */
    free(corrupt_att);
    corrupt_att = malloc(sizeof(*corrupt_att)
        * (l0r_attitude_count - samples_per_record));
    if (corrupt_att == NULL)
    {
        IAS_LOG_ERROR("Allocating quaternion quality flag");
        free(quaternion_data);
        free(quaternion_time_data);
        free(valid_quaternion_flag);
        exit(EXIT_FAILURE);
    }

    /* Copy all the records except the 9th one */
    memcpy(corrupt_att, l0r_attitude, 
        sizeof(*corrupt_att) * 8 * samples_per_record);
    memcpy(&corrupt_att[8 * samples_per_record], 
           &l0r_attitude[9 * samples_per_record], 
           sizeof(*corrupt_att) * (l0r_attitude_count
            - 9 * samples_per_record));

    /* Now set the expected output for the missing record */
    memcpy(modified_output, output_quaternion_data,
        sizeof(*modified_output) * l0r_attitude_count);
    memcpy(modified_valid_flag, output_valid_quaternion_flag,
        sizeof(*modified_valid_flag) * l0r_attitude_count);
    for (index = 8 * samples_per_record; index < 9 * samples_per_record;
         index++)
    {
        modified_output[index].vector.x = 0.0;
        modified_output[index].vector.y = 0.0;
        modified_output[index].vector.z = 0.0;
        modified_output[index].scalar = 0.0;
        modified_valid_flag[index] = 0;
    }

    /* Call the routine with the corrupted data */
    status = ias_ancillary_identify_quaternion_outliers(cpf, corrupt_att,
                 l0r_attitude_count - samples_per_record, quaternion_time_data,
                 quaternion_data, valid_quaternion_flag, quat_count,
                 &valid_quaternion_count, &interpolate_quaternions_flag);

    /* verify that the return value is correct */
    assert_equal(status, SUCCESS);

    /* quaternions should have been inserted */
    assert_equal(valid_quaternion_count, l0r_attitude_count);
    assert_equal(interpolate_quaternions_flag, 1);

    verify_quaternions(quaternion_time_data, quaternion_data, 
        valid_quaternion_flag, output_quaternion_time_data,
        modified_output, modified_valid_flag, valid_quaternion_count);

    /* Now test that we won't write past the end of the buffer if there isn't
       enough buffer space.  One of the buffers is reallocated smaller so
       it can be used to trigger errors in tools like valgrind if the buffer
       extents are exceeded. */
    free(quaternion_data);
    quaternion_data = (IAS_QUATERNION *)malloc(sizeof(IAS_QUATERNION)
        * (quat_count - samples_per_record));
    if (quaternion_data == NULL)
    {
        IAS_LOG_ERROR("Allocating quaternions");
        exit(EXIT_FAILURE);
    }
    status = ias_ancillary_identify_quaternion_outliers(cpf, corrupt_att,
                l0r_attitude_count - samples_per_record, quaternion_time_data,
                quaternion_data, valid_quaternion_flag, 
                quat_count - samples_per_record, &valid_quaternion_count,
                &interpolate_quaternions_flag);

    /* verify that the return value is correct */
    assert_equal(status, SUCCESS);

    /* quaternions should have been inserted */
    assert_equal(valid_quaternion_count, quat_count - samples_per_record);
    assert_equal(interpolate_quaternions_flag, 1);

    verify_quaternions(quaternion_time_data, quaternion_data, 
        valid_quaternion_flag, output_quaternion_time_data,
        modified_output, modified_valid_flag, valid_quaternion_count);

    /* Free the allocated memory */
    free(modified_valid_flag);
    free(modified_output);
    free(corrupt_att);
    free(quaternion_data);
    free(quaternion_time_data);
    free(valid_quaternion_flag);

    ias_cpf_free(cpf);
    cpf = NULL;
}


int initialize()
{
    double init_time_seconds = 0.0;
    const struct IAS_CPF_EARTH_CONSTANTS *earth_constants = NULL;
    int status;
    char *test_data_dir;           /* Holds test data dir env var */
    char cpf_filename[PATH_MAX];   /* CPF filename */

    /* Initialize the satellite attributes library */
    status = ias_sat_attr_initialize(IAS_L8);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing sat_attr");
        return ERROR;
    }

    /* Get IAS_UNIT_TEST_DATA_DIR environment variable */
    test_data_dir = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (test_data_dir == NULL)
    {
        IAS_LOG_ERROR("Getting IAS_UNIT_TEST_DATA_DIR env variable");
        return ERROR;
    }
    else
    {
        status = snprintf(cpf_filename, sizeof(cpf_filename), 
            "%s%s", test_data_dir, CPF);
        if (status >= sizeof(cpf_filename))
        {
            IAS_LOG_ERROR("Test data dir string truncated by snprintf");
            return ERROR;
        }
    }

    cpf = ias_cpf_read(cpf_filename);
    if (cpf == NULL)
    {
        IAS_LOG_ERROR("Reading CPF");
        return ERROR;
    }

    /* Get the CPF earth constants */
    earth_constants = ias_cpf_get_earth_const(cpf);
    if (earth_constants == NULL)
    {
        IAS_LOG_ERROR("Reading CPF earth constants");
        ias_cpf_free(cpf);
        return ERROR;
    }

    /* Initialize the leap seconds */
    init_time_seconds = 107417131.550800; /* from the calval test data */
    if(ias_math_init_leap_seconds(init_time_seconds, 
       &earth_constants->leap_seconds_data))
    {
        IAS_LOG_ERROR("Computing leap second conversion for l0r_header");
        return ERROR;
    }

    return SUCCESS;
}

/* Re-enable the command line specified optimization */
IAS_RESTORE_OPTIMIZATION

int main(int argc, char **argv)
{
    int status = 0;
    char prog_name[PATH_MAX];          /* Name of the program */

    /* Get the name of the program */  
    strcpy(prog_name, argv[0]);

    /* Initialize the IAS logging library */
    status = ias_log_initialize(prog_name);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing log");
        exit(EXIT_FAILURE);
    }

    if (initialize() != SUCCESS)
    {
        exit(EXIT_FAILURE);
    }

    TestSuite *suite = create_test_suite();

    add_test(suite, test_ias_convert_imu_to_acs_anomaly_handling);
    add_test(suite, test_ias_identify_quaternion_outliers_anomaly_handling);

    if (argc > 1)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }

    status = run_test_suite(suite, create_text_reporter());

    ias_cpf_free(cpf);
    cpf = NULL;

    return status;
}

