#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ias_logging.h"
#include "ias_ancillary.h"

#include "cgreen.h"

#define RECORDS 10

Ensure test_get_attitude()
{
    IAS_ANC_ATTITUDE_DATA *att;
    int i;
    int status;
    double roll;
    double pitch;
    double yaw;
    IAS_LOG_MESSAGE_LEVEL old_level;

    /* Allocate space for some dummied up attitude records */
    att = ias_ancillary_allocate_attitude(RECORDS);
    assert_true(att != NULL);
    if (att == NULL)
        return;

    /* Set a dummy epoch time since this test doesn't care about it */
    att->utc_epoch_time[0] = 2012.0;
    att->utc_epoch_time[1] = 2.0;
    att->utc_epoch_time[2] = 0.0;

    /* populate the attitude records */
    for (i = 0; i < RECORDS; i++)
    {
        int j;
        double seconds;
        double roll;
        double pitch;
        double yaw;

        /* Calculate the record time, including a jump half way through the
           records */
        if (i <= RECORDS/2)
            j = i;
        else
            j = i + 1;

        seconds = j * 0.02;
        roll = 5.0 + j * 0.5;
        pitch = -6.0 + j * 0.6;
        yaw = 7.0 - j * 0.7;

        att->records[i].seconds_from_epoch = seconds;
        att->records[i].roll = roll;
        att->records[i].pitch = pitch;
        att->records[i].yaw = yaw;
    }

    /* Test times that should fail since they are outside of the attitude
       time range */
    old_level = ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = ias_ancillary_get_attitude_at_time(att, -0.001, &roll, &pitch,
        &yaw);
    assert_true(status == ERROR);
    status = ias_ancillary_get_attitude_at_time(att, 0.201, &roll, &pitch,
        &yaw);
    assert_true(status == ERROR);
    ias_log_set_output_level(old_level);

    /* Test times that should pass */
    /* Test time matching the first record */
    status = ias_ancillary_get_attitude_at_time(att, 0.0, &roll, &pitch, &yaw);
    assert_true(status == SUCCESS);
    assert_double_equal(roll, 5.0);
    assert_double_equal(pitch, -6.0);
    assert_double_equal(yaw, 7.0);

    /* Test time just after the first record */
    status = ias_ancillary_get_attitude_at_time(att, 0.01, &roll, &pitch, &yaw);
    assert_true(status == SUCCESS);
    assert_double_equal(roll, 5.25);
    assert_double_equal(pitch, -5.7);
    assert_double_equal(yaw, 6.65);

    /* Test time equal to the second record */
    status = ias_ancillary_get_attitude_at_time(att, 0.02, &roll, &pitch, &yaw);
    assert_true(status == SUCCESS);
    assert_double_equal(roll, 5.5);
    assert_double_equal(pitch, -5.4);
    assert_double_equal(yaw, 6.3);

    /* Test time between two records */
    status = ias_ancillary_get_attitude_at_time(att, 0.049, &roll, &pitch,
        &yaw);
    assert_true(status == SUCCESS);
    assert_double_equal(roll, 6.225);
    assert_double_equal(pitch, -4.53);
    assert_double_equal(yaw, 5.285);

    /* Test time at start of time jump */
    status = ias_ancillary_get_attitude_at_time(att, 0.10, &roll, &pitch, &yaw);
    assert_true(status == SUCCESS);
    assert_double_equal(roll, 7.5);
    assert_double_equal(pitch, -3.0);
    assert_double_equal(yaw, 3.5);

    /* Test time between records with a time jump */
    status = ias_ancillary_get_attitude_at_time(att, 0.12, &roll, &pitch, &yaw);
    assert_true(status == SUCCESS);
    assert_double_equal(roll, 8.0);
    assert_double_equal(pitch, -2.4);
    assert_double_equal(yaw, 2.8);

    /* Test time at the end of the time jump */
    status = ias_ancillary_get_attitude_at_time(att, 0.14, &roll, &pitch, &yaw);
    assert_true(status == SUCCESS);
    assert_double_equal(roll, 8.5);
    assert_double_equal(pitch, -1.8);
    assert_double_equal(yaw, 2.1);

    /* Test time just before the last record */
    status = ias_ancillary_get_attitude_at_time(att, 0.199, &roll, &pitch,
        &yaw);
    assert_true(status == SUCCESS);
    assert_double_equal(roll, 9.975);
    assert_double_equal(pitch, -0.03);
    assert_double_equal(yaw, 0.035);

    /* Test time equal to the last record */
    status = ias_ancillary_get_attitude_at_time(att, 0.2, &roll, &pitch, &yaw);
    assert_true(status == SUCCESS);
    assert_double_equal(roll, 10.0);
    assert_double_equal(pitch, 0.0);
    assert_double_equal(yaw, 0.0);

    /* Free the attitude records */
    ias_ancillary_free_attitude(att);
}

/**************************************************************************/

int main(int argc, char **argv)
{
    int status = 0;

    /* Initialize the IAS logging library */
    status = ias_log_initialize(argv[0]);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing log");
        exit(EXIT_FAILURE);
    }

    TestSuite *suite = create_test_suite();

    add_test(suite, test_get_attitude);

    if (argc > 1)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }

    status = run_test_suite(suite, create_text_reporter());

    return status;
}
