#include "ias_types.h"
#include "wedge_grid.h"
#include "calval_band_map.h"

/*****************************************************************************/
int copy_base_grid
(
    const oli_GRID_TYPE *old_grid,
    IAS_GRID_TYPE *grid,
    int calval_version
)
{
    int index;

    /* copy the straight-forward items */
    strcpy(grid->gridname, old_grid->gridname);
    grid->dtype = old_grid->dtype;
    grid->satnum = old_grid->satnum;
    grid->frame_type = old_grid->frame_type;
    strcpy(grid->units, old_grid->units);
    grid->code = old_grid->code;
    strcpy(grid->datum, old_grid->datum);
    grid->spheroid = old_grid->spheroid;
    grid->zone = old_grid->zone;
    for (index = 0; index < IAS_PROJ_PARAM_SIZE; index++)
        grid->projprms[index] = old_grid->projprms[index];
    grid->corners.upleft.x = old_grid->corners.upleft.x;
    grid->corners.upleft.y = old_grid->corners.upleft.y;
    grid->corners.upright.x = old_grid->corners.upright.x;
    grid->corners.upright.y = old_grid->corners.upright.y;
    grid->corners.loleft.x = old_grid->corners.loleft.x;
    grid->corners.loleft.y = old_grid->corners.loleft.y;
    grid->corners.loright.x = old_grid->corners.loright.x;
    grid->corners.loright.y = old_grid->corners.loright.y;
    grid->center_pt.lat = old_grid->center_pt.lat;
    grid->center_pt.lng = old_grid->center_pt.lng;
    grid->proj_center_pt.x = old_grid->proj_center_pt.x;
    grid->proj_center_pt.y = old_grid->proj_center_pt.y;
    grid->rotate_angle = old_grid->rotate_angle;
    grid->orient_angle = old_grid->orient_angle;

    /* translate the correction type */
    switch (old_grid->correction_type)
    {
        case PRECISION:
            grid->correction_type = IAS_PRECISION;
            break;
        case TERRAIN:
            grid->correction_type = IAS_TERRAIN;
            break;
        default:
            grid->correction_type = IAS_SYSTEMATIC;
            break;
    }
    /* assume earth acquisition */
    grid->acquisition_type = IAS_EARTH;

    /* translate the things that aren't quite the same */
    /*-------------------------------------------------*/
    /* initialize the bands available and present */
    for (index = 0; index < IAS_MAX_NBANDS; index++)
    {
        grid->bands_available[index] = 0;
        grid->bands_present[index] = 0;
        /* hard code the OLI/TIRS bands */
        if (index < 9)
            grid->scas_per_band[index] = old_grid->nscas;
        else
            grid->scas_per_band[index] = 3;
    }

    /* set up the bands that are available */
    for (index = 0; index < old_grid->nbands; index++)
    {
        int target_band_index = convert_calval_band_index_to_oli(
                calval_version, old_grid->bands[index]);

        grid->bands_available[target_band_index] = 1;
    }

    /* copy the active corners */
    for (index = 0; index < old_grid->nbands; index++)
    {
        int target_band_index = convert_calval_band_index_to_oli(
                calval_version, old_grid->bands[index]);

        grid->active_corners[target_band_index].upleft.x
            = old_grid->actarea[index].upleft.x;
        grid->active_corners[target_band_index].upleft.y
            = old_grid->actarea[index].upleft.y;
        grid->active_corners[target_band_index].upright.x
            = old_grid->actarea[index].upright.x;
        grid->active_corners[target_band_index].upright.y
            = old_grid->actarea[index].upright.y;
        grid->active_corners[target_band_index].loright.x
            = old_grid->actarea[index].loright.x;
        grid->active_corners[target_band_index].loright.y
            = old_grid->actarea[index].loright.y;
        grid->active_corners[target_band_index].loleft.x
            = old_grid->actarea[index].loleft.x;
        grid->active_corners[target_band_index].loleft.y
            = old_grid->actarea[index].loleft.y;
    }

    return SUCCESS;
}

/*****************************************************************************/
int copy_gridbands
(
    const oli_GRID_TYPE *old_grid,
    IAS_GRID_TYPE *grid,
    int calval_version
)
{
    int band_index;
    int sca_index;

    for (band_index = 0; band_index < grid->nbands; band_index++)
    {
        int band_number;
        int old_band_index;
        int band_available = 0;

        for (sca_index = 0; sca_index < grid->scas_per_band[band_index];
             sca_index++)
        {
            memset(&grid->gridbands[band_index][sca_index], 0, 
                   sizeof(IAS_GRID_BAND_TYPE));
        }

        band_number = ias_sat_attr_convert_band_index_to_number(band_index);

        /* find the matching gridband in the old grid */
        for (old_band_index = 0; old_band_index < old_grid->nbands;
             old_band_index++)
        {
            if (convert_calval_band_index_to_oli(calval_version,
                    old_grid->bands[old_band_index]) == band_index)
            {
                band_available = 1;
                break;
            }
        }

        for (sca_index = 0; sca_index < grid->scas_per_band[band_index];
             sca_index++)
        {
            IAS_GRID_BAND_TYPE *gridband 
                = &grid->gridbands[band_index][sca_index];

            gridband->band_index = band_index;
            gridband->sca_number = sca_index + 1;

            if (band_available)
            {
                const oli_GRID_BAND_TYPE *ogb 
                    = &old_grid->gridbands[sca_index]
                                          [old_grid->bands[old_band_index]];

                gridband->grid_cell_size.line = ogb->grid_cell_size.line;
                gridband->grid_cell_size.samp = ogb->grid_cell_size.samp;
                gridband->inv_cell_lines = ogb->inv_cell_lines;
                gridband->inv_cell_samps = ogb->inv_cell_samps;
                gridband->pixsize = ogb->pixsize;
                gridband->zspacing = ogb->zspacing;
                gridband->lines = ogb->lines;
                gridband->samps = ogb->samps;
                gridband->ngrid_lines = ogb->ngrid_lines;
                gridband->ngrid_samps = ogb->ngrid_samps;
                gridband->nzplanes = ogb->nzplanes;
                gridband->zeroplane = ogb->zeroplane;
                gridband->degree = ogb->degree;
            }
        }

        if (band_available)
        {
            if (ias_grid_malloc_band(grid, band_number) != SUCCESS)
            {
                printf("Error allocating gridband for band number %d\n",
                       band_number);
                return ERROR;
            }
            grid->bands_present[band_index] = 1;

            for (sca_index = 0; sca_index < grid->scas_per_band[band_index];
                 sca_index++)
            {
                IAS_GRID_BAND_TYPE *gridband 
                    = &grid->gridbands[band_index][sca_index];
                int size;
                const oli_GRID_BAND_TYPE *ogb 
                    = &old_grid->gridbands[sca_index]
                                          [old_grid->bands[old_band_index]];

                /* copy the data across */
                memcpy(gridband->in_lines, ogb->in_lines, 
                       gridband->ngrid_lines * sizeof(int));
                memcpy(gridband->in_samps, ogb->in_samps, 
                       gridband->ngrid_samps * sizeof(int));
                size = gridband->ngrid_lines * gridband->ngrid_samps 
                     * gridband->nzplanes * sizeof(double);
                memcpy(gridband->out_lines, ogb->out_lines, size);
                memcpy(gridband->out_samps, ogb->out_samps, size);
                size = 2 * gridband->ngrid_lines * gridband->ngrid_samps
                     * sizeof(double);
                memcpy(gridband->delta_line_oe, ogb->delta_line_oe, size);
                memcpy(gridband->delta_samp_oe, ogb->delta_samp_oe, size);
                size = (gridband->ngrid_lines - 1) * (gridband->ngrid_samps - 1)
                     * gridband->nzplanes * sizeof(IAS_COEFFICIENTS);
                memcpy(gridband->sattoproj, ogb->sattoproj, size);
                memcpy(gridband->projtosat, ogb->projtosat, size);
                size = (gridband->degree + 1) * (gridband->degree + 1)
                     * gridband->nzplanes * sizeof(double);
                memcpy(gridband->poly_lines, ogb->poly_lines, size);
                memcpy(gridband->poly_samps, ogb->poly_samps, size);
                size = gridband->ngrid_lines * gridband->ngrid_samps 
                     * gridband->nzplanes * sizeof(IAS_VECTOR);
                memcpy(gridband->line_sensitivity, ogb->line_sens, size);
                memcpy(gridband->samp_sensitivity, ogb->samp_sens, size);
            }
        }
    }

    return SUCCESS;
}

/*****************************************************************************/
int wedge_grid
(
    const oli_GRID_TYPE *old_grid,
    IAS_GRID_TYPE *grid,
    int calval_version      /* I: version of calval band number mapping */
)
{
    /* initialize the target grid */
    if (ias_grid_initialize(grid, IAS_MAX_NBANDS, IAS_MAX_NSCAS) != SUCCESS)
    {
        printf("Error initializing grid structure\n");
        return 0;
    }

    /* copy the base grid structure */
    copy_base_grid(old_grid, grid, calval_version);

    /* copy the non-allocated members of the gridbands */
    if (copy_gridbands(old_grid, grid, calval_version) != SUCCESS)
    {
        printf("Error copying gridbands\n");
        return 0;
    }

    return 1;
}
