
/*!
    \file IASDatabaseErrorDialog.cpp
    \brief Provides the implementation for IASDatabaseErrorDialog.

    See IASDatabaseErrorDialog.
*/

/* System Includes */
#include <iostream>

/* Qt Includes */
#include <QtCore>
#include <QtGui>
#include <QtSql>

/* Local Includes */
#include "IASDatabaseErrorDialog.h"

/*!
    This constructor allows the user to create a IASDatabaseErrorDialog
    object.  It constructs a specific type of error dialog for database
    errors.

    All file and/or database operations that need to be performed to
    record the error are handled by the base class constructor.

    If a pointer to a record set is passed in, the details of that record
    set's variables and members are printed in the details box.  If no
    pointer to a record set is passed in, then the details of the global
    database object are used when displaying the extra details info.

    \param parent Input: The parent widget from which to create the
                         IASDatabaseErrorDialog object.

    \param strErrorText Input: A short description of the error.

    \param strFileName Input: The name of the invoking file.
                              (Use __FILE__)

    \param lineNum Input: The line number from the invoking file.
                          (Use __LINE__)

    \param rsDetails Input: The pointer to the QSqlQuery that returned
                            the error.

    \param fl Input: Qt uses this for internal purposes.
*/
IASDatabaseErrorDialog::IASDatabaseErrorDialog(QWidget* parent,
                                         const QString strErrorText,
                                         const QString strFileName,
                                         const int lineNum,
                                         QSqlQuery *rsDetails,
                                         Qt::WindowFlags fl)
: IASErrorDialog(parent,strErrorText,fl)
{
    //change the size policy of the frame holding the error text to not
    //grow vertically when resized
    frameErrorInfo->setSizePolicy( QSizePolicy( QSizePolicy::Preferred, 
                QSizePolicy::Fixed ) );

    frameDetails->show();

    createDetails(strFileName,lineNum,rsDetails);
}

/*!
    This method creates any objects that are needed in the details frame.

    \param strFileName Input: The name of the invoking file.

    \param lineNum Input: The line number from the invoking file.

    \param rsDetails Input: The pointer to the QSqlQuery that returned
                            the error.
*/
void IASDatabaseErrorDialog::createDetails(QString strFileName,
                                        int lineNum,
                                        QSqlQuery *rsDetails)
{
    QSqlError lastSQLError;
    
    if(rsDetails !=NULL)	//RecordSet caused the error
    {
        lastSQLError = rsDetails->lastError();

        addDetailLine("QSqlQuery Error");
        addDetailLine("The error occurred in " + strFileName +
                      " at line " "number: " +
                      QString::number(lineNum) + "\n");
        addDetailLine("SQL Statement: " + rsDetails->lastQuery() + "\n" );
        addDetailLine("Oracle Error Message: " + 
                      lastSQLError.databaseText() + "\n");
    }
}

