 
/*!
    \file IASDatabaseLogin.cpp
    \brief Provides the implementation for IASDatabaseLogin.

    See IASDatabaseLogin.
*/

// System Includes
#include <stdlib.h>

// Qt Includes
#include <QtCore>
#include <QtGui>
#include <QtSql>

// Local Includes
#include "IASDatabaseLogin.h"

/*!
    This constructor allows the user to create a IASDatabaseLogin object.

    \param parent Input: The parent widget from which to create the
                         IASDatabaseLogin object.

    \param environmentVariable Input: The name of the environment variable to
                                      use for figuring out the username and
                                      password.

    \param fl Input: Qt uses this for internal purposes.
*/
IASDatabaseLogin::IASDatabaseLogin(QWidget* parent,
                             QString environmentVariable,
                             Qt::WindowFlags fl)
: QDialog(parent, fl),
  myPrePopulated(false),
  myConnected(false)
{
    // Set up the user interface.
    setupUi(this);

    // In the constructor set the focus initially to the user name
    // line edit and get the database TNS name from the TWO_TASK 
    // environment variable if it is set.

    leUserName->setFocus();

    QString twoTask = getenv("TWO_TASK");

    if( !twoTask.isEmpty() )
    {
        leHostString->setText(twoTask);
    }

    // Try the environment.
    if( !environmentVariable.isEmpty() )
    {
        QString usernamePassword = getenv(environmentVariable.toAscii());

        if( !usernamePassword.isEmpty() )
        {
            // Remove any spaces.
            usernamePassword = usernamePassword.simplified();

            QStringList list = usernamePassword.split(QRegExp("/"));
            if( list.count() == 2 )
            {
                if( !twoTask.isEmpty() &&
                    !list[0].isEmpty() &&
                    !list[1].isEmpty() )
                {
                    leUserName->setText(list[0]);
                    lePassword->setText(list[1]);

                    myPrePopulated = true;
                }
            }
        }
    }

    // Connect all the signals and slots.
    connect(pbCancel, SIGNAL(clicked()), this, SLOT(reject()));
    connect(pbOK, SIGNAL(clicked()), this, SLOT(tryConnect()));
}

/*!
    Execute the dialog.  If the user has an environment variable setup
    correctly try that first before displaying the interface and prompting
    the user.
*/
void IASDatabaseLogin::exec()
{
    if( myPrePopulated && !myConnected)
    {
        tryConnect();

        if(result() != QDialog::Accepted)
        {
            QDialog::exec();
        }
    }
    else
    {
        QDialog::exec();
    }
}

/*!
    Attempts to connect to the database using the information entered
    by the operator or through the environment.
*/
void IASDatabaseLogin::tryConnect()
{
    QString strUserName=leUserName->text().simplified();
    QString strHostName=leHostString->text().simplified();
    QString strPassword=lePassword->text().simplified();
    
    // Fist, check to make sure a host name (TNS) is entered.  If one 
    // is not, the default TNS name on the machine is used.  
    // This may not be desired.
    //--------------------------------------------------------------
    if( strHostName.isEmpty() || strHostName.isNull() )
    {
        QMessageBox::warning(0,"Database Connection Error",
                 "Please enter a host name (TNS name) to connect to.",
                 QMessageBox::Ok,0);
        return;
    }
    
    QSqlError connectionError;

    theDatabase.setDatabaseName(strHostName);
    theDatabase.setUserName(strUserName);
    theDatabase.setPassword(strPassword);
	
    if( !theDatabase.open() )
    {
        connectionError = theDatabase.lastError();
        QMessageBox::warning(0,"Database Connection Error",
              "An error occurred attempting to connect to the"
              " database. \n"
              + connectionError.databaseText().simplified(),
              QMessageBox::Ok,0);
    }
    else    //The database is connected to and open
    {
        myConnected = true;
        accept();
    }
}

/*!
    Sets the database object to use when connecting to the database.
*/
void IASDatabaseLogin::setDatabase(QSqlDatabase newDatabase)	
{    
    theDatabase = newDatabase;
}

