
/*!
    \file IASListViewItem.cpp
    \brief Provides the implementation for IASListViewItem.

    See IASListViewItem.
*/

/* System Includes */
#include <iostream>

/* Qt Includes */
#include <QtCore>
#include <QtGui>

/* Local Includes */
#include "IASUtilities.h"
#include "IASListViewItem.h"

/*!
    This constructor allows the user to create a IASListViewItem object.
*/
IASListViewItem::IASListViewItem()
: QTreeWidgetItem()
{
}

/*!
    This constructor allows the user to create a IASListViewItem object.

    \param list Input: The columnDataTypesList to use for sorting.
*/
IASListViewItem::IASListViewItem(columnDataTypesList &list)
: QTreeWidgetItem()
{
    m_listColumnType = list;
}

/*!
    This constructor allows the user to create a IASListViewItem object.

    \param parent Input: The parent QTreeWidget object from which to create
                         the IASListViewItem object.
*/
IASListViewItem::IASListViewItem(QTreeWidget *parent)
: QTreeWidgetItem(parent)
{
}

/*!
    This constructor allows the user to create a IASListViewItem object
    and pass in a columnDataTypesList to use for sorting.

    \param parent Input: The parent QTreeWidget object from which to create
                         the IASListViewItem object.

    \param list Input: The columnDataTypesList to use for sorting.
*/
IASListViewItem::IASListViewItem(QTreeWidget *parent,
                                 columnDataTypesList &list)
: QTreeWidgetItem(parent)
{
    m_listColumnType = list;
}

/*!
    This method overrides the base less-than operator to compare data
    according to the type.

    This method returns true if this item is less than the other item, and
    false if it is equal to or greater-than the other item.

    \param item Input: The IASListViewItem to compare against.
*/
bool IASListViewItem::operator<(const QTreeWidgetItem &item) const
{
    double cNum, oNum;
    QDateTime cDateTime, oDateTime;
    QDate cDate, oDate;
    QTime cTime, oTime;
    int col = treeWidget()->sortColumn();
    QString tempStr;
    QRegExp regExp = QRegExp("[A-Za-z]+");

    if (m_listColumnType.isEmpty())
        return false;
    else
    {
        columnDataTypes type = m_listColumnType[col];

        switch (type)
        {
            case Id:
                tempStr = text(col);
                tempStr.remove(regExp); // Remove leading characters.
                if (tempStr.isEmpty())
                    return false;
                cNum = tempStr.toDouble();

                tempStr = item.text(col);
                tempStr.remove(regExp); // Remove leading characters.
                if (tempStr.isEmpty())
                    return false;
                oNum = tempStr.toDouble();

                if (cNum == oNum)
                    return false;
                else
                    return cNum < oNum ? true : false;
            case Number:
            case Float:
                cNum = text(col).toDouble(NULL);
                oNum = item.text(col).toDouble(NULL);
                if (cNum == oNum)
                    return false;
                else
                    return cNum < oNum ? true : false;
            case DateTime:
            case DateYMD:
                cDateTime = IASUtilities::getDateTime(text(col));
                oDateTime = IASUtilities::getDateTime(item.text(col));
                if (cDateTime == oDateTime)
                    return false;
                else
                    return cDateTime < oDateTime ? true : false;
            case DateMDY:
                qDebug() << __FILE__ << ":" << __LINE__
                         << ":Error: DateMDY not supported";
                return false;
            case DateDMMMY:
                cDate = IASUtilities::getDate(text(col),type);
                oDate = IASUtilities::getDate(item.text(col),type);
                if (cDate == oDate)
                    return false;
                else
                    return cDate < oDate ? true : false;
            case Time:
                cTime = IASUtilities::getTime(text(col));
                oTime = IASUtilities::getTime(item.text(col));
                if (cTime == oTime)
                    return false;
                else
                    return cTime < oTime ? true : false;
            default:
                return text(col) < item.text(col);
        }
    }

    return false; /* Return to satisfy compiler warnings. Shouldn't get here. */
}

/*!
    Sets the tag associated with this listviewitem.

    \param newTag Input: The tag value to use.
*/
void IASListViewItem::setTag(QString newTag)
{
    m_Tag = newTag;

    return;
}

/*!
    Returns the tag associated with this listviewitem.
*/
QString IASListViewItem::getTag()
{
    return (m_Tag);
}

