
/* System Includes */

/* Qt Includes */
#include <QtCore>
#include <QtGui>

/* Local Includes */
#include "IASParameterDelegate.h"
#include "IASParameterItem.h"
#include "IASParameterModel.h"

/* ===========================================================================
  IASParameterDelegate

  Constructor for the object.
=========================================================================== */
IASParameterDelegate::IASParameterDelegate(QObject* parent)
: QStyledItemDelegate(parent)
{
}

/* ===========================================================================
  createEditor

  Creates and initializes the editor to be used for modifying the model data.
=========================================================================== */
QWidget* IASParameterDelegate::createEditor(
    QWidget* parent,
    const QStyleOptionViewItem&,
    const QModelIndex& index) const
{
    // Grab the item from the model so we can use it's type.
    IASParameterModel* model = (IASParameterModel*)index.model();
    IASParameterItem* item = model->getItem(index.row(), index.column());

    // The widget object to return.
    QWidget* editor = NULL;

    // Based on the type, create and set up the specific editor.
    switch (item->type())
    {
        case (IASParameterItem::TextSelection):
        {
            QComboBox* cbEditor = new QComboBox(parent);
            cbEditor->addItems(item->options());
            cbEditor->setCurrentIndex(
                cbEditor->findText(item->data().toString()));
            editor = cbEditor;
            break;
        }

        case (IASParameterItem::IntegerRange):
        {
            QSpinBox* sbEditor = new QSpinBox(parent);
            if (item->minimum().toInt() != item->maximum().toInt())
            {
                sbEditor->setMinimum(item->minimum().toInt());
                sbEditor->setMaximum(item->maximum().toInt());
            }
            /* We don't set the min and max if the values are equal,
               because we want an unlimited range.  If you don't want to
               edit the item set the editable flag correctly. */
            sbEditor->setValue(item->data().toInt());
            editor = sbEditor;
            break;
        }

        case (IASParameterItem::FloatRange):
        {
            QDoubleSpinBox* dsbEditor = new QDoubleSpinBox(parent);
            int decimals = item->decimals().toInt();
            if (item->minimum().toDouble() != item->maximum().toDouble())
            {
                dsbEditor->setMinimum(item->minimum().toDouble());
                dsbEditor->setMaximum(item->maximum().toDouble());
                // "decimals" is the precision.
                dsbEditor->setDecimals(decimals);
            }
            /* We don't set the min and max if the values are equal,
               because we want an unlimited range.  If you don't want to
               edit the item set the editable flag correctly. */
            dsbEditor->setValue(item->data().toDouble());
            editor = dsbEditor;
            break;
        }

        case (IASParameterItem::Float):
        {
            QLineEdit* leEditor = new QLineEdit(parent);
            int decimals = item->decimals().toInt();
            QString tmpStr = QString::number(item->data().toDouble(),
                                     'f', decimals);
            leEditor->setText(tmpStr);
            editor = leEditor;
            break;
        }

        case (IASParameterItem::Text):
        case (IASParameterItem::Integer):
        default:
        {
            QLineEdit* leEditor = new QLineEdit(parent);
            leEditor->setText(item->data().toString());
            leEditor->setSelection(0, item->data().toString().length());
            editor = leEditor;
            break;
        }
    }

    return editor;
}

/* ===========================================================================
  setEditorData

  Sets the starting value in the editor with the value from the model.
=========================================================================== */
void IASParameterDelegate::setEditorData(
    QWidget* editor,
    const QModelIndex& index)
{
    // Grab the item from the model so we can use its type.
    IASParameterModel* model = (IASParameterModel*)index.model();
    IASParameterItem* item = model->getItem(index.row(), index.column());

    // The switch causes these to be declared here. 
    QSpinBox* sbEditor = NULL;
    QDoubleSpinBox* dsbEditor = NULL;
    QComboBox* cbEditor = NULL;
    QLineEdit* leEditor = NULL;
    QString tmpStr("");
    int decimals = item->decimals().toInt();

    // Based on the type, perform actions using the specific editor.
    switch (item->type())
    {
        case (IASParameterItem::TextSelection):
            cbEditor = static_cast<QComboBox*>(editor);
            cbEditor->setCurrentIndex(
                cbEditor->findText(item->data().toString()));
            break;

        case (IASParameterItem::IntegerRange):
            sbEditor = static_cast<QSpinBox*>(editor);
            sbEditor->setValue(item->data().toInt());
            break;

        case (IASParameterItem::FloatRange):
            dsbEditor = static_cast<QDoubleSpinBox*>(editor);
            dsbEditor->setValue(item->data().toDouble());
            break;

        case (IASParameterItem::Float):
            leEditor = static_cast<QLineEdit*>(editor);
            tmpStr = QString::number(item->data().toDouble(),
                                     'f', decimals);
            leEditor->setText(tmpStr);
            break;

        case (IASParameterItem::Text):
        case (IASParameterItem::Integer):
        default:
            leEditor = static_cast<QLineEdit*>(editor);
            leEditor->setText(item->data().toString());
            break;
    }
}

/* ===========================================================================
  setModelData

  Updates the model with the contents/selection from the editor.
=========================================================================== */
void IASParameterDelegate::setModelData(
    QWidget* editor,
    QAbstractItemModel* model,
    const QModelIndex& index) const
{
    // Grab the item from the model so we can use its type.
    IASParameterModel* cModel = (IASParameterModel*)model;
    IASParameterItem* item = cModel->getItem(index.row(), index.column());

    // Variable to hold the data retrieved from the editor.
    QVariant data("");

    // The switch causes these to be declared here. 
    QSpinBox* sbEditor = NULL;
    QDoubleSpinBox* dsbEditor = NULL;
    QComboBox* cbEditor = NULL;
    QLineEdit* leEditor = NULL;

    // Based on the type, perform actions using the specific editor.
    switch (item->type())
    {
        case (IASParameterItem::TextSelection):
            cbEditor = static_cast<QComboBox*>(editor);
            data = cbEditor->currentText();
            break;

        case (IASParameterItem::IntegerRange):
            sbEditor = static_cast<QSpinBox*>(editor);
            data = sbEditor->value();
            break;

        case (IASParameterItem::FloatRange):
            dsbEditor = static_cast<QDoubleSpinBox*>(editor);
            data = dsbEditor->value();
            break;

        case (IASParameterItem::Text):
        case (IASParameterItem::Integer):
        case (IASParameterItem::Float):
        default:
            leEditor = static_cast<QLineEdit*>(editor);
            data = leEditor->text();
            break;
    }

    // Only modify the model if the data has been changed.
    if (item->data() != data)
    {
        cModel->setModified(true);
        cModel->setData(index, data, Qt::EditRole);
    }
}

/* ===========================================================================
  updateEditoryGeometry

  Updates the editors geometry with information from the style.
=========================================================================== */
void IASParameterDelegate::updateEditorGeometry(
    QWidget *editor,
    const QStyleOptionViewItem &option,
    const QModelIndex &) const
{
    editor->setGeometry(option.rect);
}

