
/* System Includes */

/* Qt Includes */
#include <QtCore>
#include <QtGui>

/* Local Includes */
#include "IASParameterModel.h"

QVariant PROG_ERROR = "PROG_ERROR";

/* ===========================================================================
  IASParameterModel

  The constructor for the model object.
=========================================================================== */
IASParameterModel::IASParameterModel(QObject* parent)
: QAbstractTableModel(parent),
  m_rows(0),
  m_columns(0)
{
    // Provide default initialization.

    m_modified = false;

    // We have to Create a dummy item to return since we are using "take",
    // because the caller assumes ownership.
    m_dummyIASParameterItem = new IASParameterItem;
    m_dummyIASParameterItem->setData(PROG_ERROR);
}

/* ===========================================================================
  flags

  Returns all flags associated with the item/cell.
=========================================================================== */
Qt::ItemFlags IASParameterModel::flags(const QModelIndex& index) const
{
    // Grab the default flags for the cell.
    Qt::ItemFlags flags = QAbstractItemModel::flags(index);

    // If the index is not valid return the default flags.
    if (!index.isValid())
    {
        return flags;
    }

    // Figure out the row and column for the cell.
    int row = index.row();
    int column = index.column();

    // Grab the specific item for the cell.
    IASParameterItemRow* itemRow = m_data[row];
    IASParameterItem* item = itemRow->value(column);

    // Determine if the item can be edited.
    if (item->editable())
    {
        // Add the editable flag to the flag object.
        flags |= Qt::ItemIsEditable;
    }

    return flags;
}

/* ===========================================================================
  data

  Returns the data associated with the cell.
=========================================================================== */
QVariant IASParameterModel::data(
    const QModelIndex& index,
    int role) const
{
    // If the index is not valid or the table is empty, we want to return an
    // empty data object.
    if (!index.isValid() || m_rows == 0)
    {
        return QVariant();
    }

    // Figure out the row and column for the cell.
    int row = index.row();
    int column = index.column();

    // Grab the specific item for the cell.
    IASParameterItemRow* itemRow = m_data[row];
    IASParameterItem* item = itemRow->value(column);

    // Depending on the roles we use, return the correct data.
    if (role == Qt::TextAlignmentRole)
    {
        return int(Qt::AlignVCenter | item->alignment());
    }
    else if (role == Qt::DisplayRole || /* To display the data. */
             role == Qt::EditRole) /* To set the data in the editor. */
    {
        QVariant data;

        switch (item->type())
        {
            case (IASParameterItem::Float):
            case (IASParameterItem::FloatRange):
            {
                int decimals = item->decimals().toInt();
                QString dataStr = QString::number(item->data().toDouble(),
                                          'f', decimals);
                data = QVariant(dataStr);
                break;
            }

            default:
            {
                /* All of the other types are currently covered by this
                   default case.  This is because they do not require anything
                   special to be done like the floating point values do. */
                data = item->data();
                break;
            }
        }

        return data;
    }

    return QVariant();
}

/* ===========================================================================
  headerData

  Return information about the headers.
=========================================================================== */
QVariant IASParameterModel::headerData(
    int section,
    Qt::Orientation,
    int role) const
{
    // If we are asking for the display role, return the label for the
    // specific section.
    if (role == Qt::DisplayRole)
    {
        return m_horizontalHeaderLabels[section];
    }

    // Default to an empty data object.
    return QVariant();
}

/* ===========================================================================
  rowCount

  Returns the number of rows in the table/data.
=========================================================================== */
int IASParameterModel::rowCount(const QModelIndex&) const
{
    return m_rows;
}

/* ===========================================================================
  setColumncount

  Sets the number of columns in the table/data.
=========================================================================== */
void IASParameterModel::setColumnCount(int columns)
{
    m_columns = columns;
}

/* ===========================================================================
  columnCount

  Returns the number of columns in the table/data.
=========================================================================== */
int IASParameterModel::columnCount(const QModelIndex&) const
{
    return m_columns;
}

/* ===========================================================================
  setData

  Sets the data for the specific item in the table.

  Returns true if the item was modified and false if the item was not
  modified.
=========================================================================== */
bool IASParameterModel::setData(
    const QModelIndex& index,
    const QVariant& value,
    int role)
{
    // If the index is not valid we want to return failure.
    if (!index.isValid())
    {
        return false;
    }

    // Figure out the row and column for the cell.
    int row = index.row();
    int column = index.column();

    // Grab the specific item for the cell.
    IASParameterItemRow* itemRow = m_data[row];
    IASParameterItem* item = itemRow->value(column);

    // We only want to mess with the item if it is editable.
    if (role == Qt::EditRole && item->editable())
    {
        // Convert to a local variable.
        QVariant data = value;

        // Set the items data object.
        item->setData(data);

        // Required to emit this signal.
        emit dataChanged(index, index);

        if (m_modified)
        {
            emit itemDataChanged(index);
            m_modified = false;
        }

        return true;
    }

    return false;
}

/* ===========================================================================
  setModelData

  Sets the data for the model to the data passed in.
=========================================================================== */
void IASParameterModel::setModelData(IASParameterItemRows& data)
{
    m_data = data;

    m_rows = data.count();
    m_columns = data[0]->count();
}

/* ===========================================================================
  modelData

  Returns a reference to all of the model data.
=========================================================================== */
IASParameterItemRows& IASParameterModel::modelData()
{
    return m_data;
}

/* ===========================================================================
  insertRow

  Add a row of data to the model.
=========================================================================== */
void IASParameterModel::insertRow(
    const QModelIndex& index,
    IASParameterItemRow* row)
{
    // Default to appending the new row.
    int insertPos = m_data.count();

    if (index.isValid())
    {
        insertPos = index.row();
    }

    beginInsertRows(QModelIndex(), m_rows, m_rows);

    // Add the row to the data.
    m_data.insert(insertPos, row);

    // Update the number of row in the model.
    m_rows = m_data.count();

    // Set the number of columns to the number provided in the first row
    // of data given to the model.  This model prevents columns from being
    // added to the model/table.
    if (m_columns == 0)
    {
        m_columns = row->count();
    }

    endInsertRows();
}

/* ===========================================================================
  appendRow

  Add a row of data to the model.
=========================================================================== */
void IASParameterModel::appendRow(IASParameterItemRow* row)
{
    beginInsertRows(QModelIndex(), m_rows, m_rows);

    // Add the row to the data.
    m_data.append(row);

    // Update the number of rows in the model.
    m_rows = m_data.count();

    // Set the number of columns to the number provided in the first row
    // of data given to the model.  This model prevents columns from being
    // added to the model/table.
    if (m_columns == 0)
    {
        m_columns = row->count();
    }

    endInsertRows();
}

/* ===========================================================================
  takeRow

  This removes a row from the models data.  It does not delete it, but
  instead returns it to the calling code.
=========================================================================== */
IASParameterItemRow* IASParameterModel::takeRow(const QModelIndex& index)
{
    int row = index.row();

    // Make sure the row exists in the data/table.
    if (row < 0 || row >= m_rows)
    {
        qDebug() << __FILE__ << ":" << __LINE__ << ":Error: Retrieving data";

        // We have to Create a dummy item here to return since we are using
        // "take", because the caller assumes ownership.
        IASParameterItemRow *dummyRow = new IASParameterItemRow;

        return dummyRow;
    }

    beginRemoveRows(QModelIndex(), row, row);

    // Grab the row.
    IASParameterItemRow* itemRow = m_data.takeAt(row);

    // Fix the number of rows.
    m_rows = m_data.count();

    endRemoveRows();

    return itemRow;
}

/* ===========================================================================
  deleteRow

  Delete a row of data from the model.
=========================================================================== */
void IASParameterModel::deleteRow(int row)
{
    // Make sure the row exists in the data/table.
    if (row < 0 || row >= m_rows)
    {
        return;
    }

    beginRemoveRows(QModelIndex(), row, row);

    // Grab the row.
    IASParameterItemRow* itemRow = m_data.takeAt(row);

    // Delete each column in the row.
    for (int idx = 0; idx < itemRow->count(); idx++)
    {
        // Grab the column/item from the row.
        IASParameterItem* item = itemRow->takeAt(0);
        delete item;
    }

    delete itemRow;

    // Fix the number of rows.
    m_rows = m_data.count();

    endRemoveRows();
}

/* ===========================================================================
  setItemData

  Sets the data for the specified row and column.

  The item is required to have already been inserted into the model
  and table using any of the add row type functions. (Such as appendRow)
=========================================================================== */
bool IASParameterModel::setItemData(
    const int row,
    const int column,
    const QVariant& value)
{
    // Make sure the row and column exists in the data/table.
    if ((row < 0 || row >= m_rows) ||
        (column < 0 || column >= m_columns))
    {
        return false;
    }

    IASParameterItemRow* itemRow = m_data[row];
    IASParameterItem* item = itemRow->value(column);

    QVariant data = value;
    item->setData(data);

    QModelIndex tIndex = index(row, column);

    emit dataChanged(tIndex, tIndex);

    return true;
}

/* ===========================================================================
  deleteData

  Delete all the data from the model.
=========================================================================== */
void IASParameterModel::deleteData()
{
    beginRemoveRows(QModelIndex(), 0, m_rows-1);

    // Delete each row.
    for (int row = 0; row < m_rows; row++)
    {
        // Grab the row.
        IASParameterItemRow* itemRow = m_data.takeAt(0);

        // Delete each column in the row.
        for (int idx = 0; idx < itemRow->count(); idx++)
        {
            // Grab the column/item from the row.
            IASParameterItem* item = itemRow->takeAt(0);
            delete item;
        }

        delete itemRow;
    }

    m_rows = m_data.count();

    endRemoveRows();
}

/* ===========================================================================
  getItem

  Returns the item associated with the specified row and column.
=========================================================================== */
IASParameterItem* IASParameterModel::getItem(int row, int column)
{
    // Make sure the row and column exists in the data/table.
    if ((row < 0 || row >= m_rows) ||
        (column < 0 || column >= m_columns))
    {
        qDebug() << __FILE__ << ":" << __LINE__ << ":Error: Retrieving data";
        return m_dummyIASParameterItem;
    }

    // Grab the row.
    IASParameterItemRow* itemRow = m_data[row];

    // Grab the column/item from the row.
    IASParameterItem* item = itemRow->value(column);

    return item;
}

/* ===========================================================================
  setModified

  Sets the modified state.
=========================================================================== */
void IASParameterModel::setModified(bool state)
{
    m_modified = state;
}

/* ===========================================================================
  modified

  Returns the modified state back to the calling code.
=========================================================================== */
bool IASParameterModel::modified()
{
    return m_modified;
}

/* ===========================================================================
  setHorizontalHeaderLabels

  Set the header labels to eventually be displayed by the table.
=========================================================================== */
void IASParameterModel::setHorizontalHeaderLabels(QStringList& labels)
{
    m_horizontalHeaderLabels = labels;
}

