
/*!
    \file IASProcessExecution.cpp
    \brief Provides the implementation for IASProcessExecution.

    See IASProcessExecution.
*/

/* System Includes */

/* Qt Includes */
#include <QtCore>
#include <QtGui>

/* Local Includes */
#include "IASUtilities.h"
#include "IASProcessExecution.h"

IASProcessExecution::IASProcessExecution(QWidget * parent,
                                   Qt::WindowFlags f)
: QWidget(parent, f),
  myProcess(NULL)
{
    myProcess = new QProcess(this);

    // Connect the outputs so we can read them.
    connect(myProcess, SIGNAL(readyReadStandardOutput()),
            this, SLOT(readFromStdoutSlot()));
    connect(myProcess, SIGNAL(readyReadStandardError()),
            this, SLOT(readFromStderrSlot()));
}

IASProcessExecution::~IASProcessExecution()
{
    delete(myProcess);
}

bool IASProcessExecution::execute(QString & cmd,
                               bool wait)
{
    QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));

    QStringList list = cmd.split(QRegExp("\\s+"));

    // Remove environment variables from the path string.
    list[0] = IASUtilities::unEnvCmd(list[0]);

    myStdout.clear();
    myStderr.clear();

    if(list[0].isEmpty())
    {
        myStderr.append("Invalid environment variable in path.");
        QApplication::restoreOverrideCursor();
        return false;
    }

    QString prog = list[0];
    QStringList argList = list;
    argList.removeFirst();

    myProcess->start(prog, argList);

    if(!myProcess->waitForStarted())
    {
        QApplication::restoreOverrideCursor();
        return false;
    }

    // We need to wait for the process to finish so that we can
    // get the results.
    while(myProcess->state() == QProcess::Running && wait)
    {
        qApp->processEvents(); // Process other events if there are any.
    }

    if(!myStderr.isEmpty())
    {
        QApplication::restoreOverrideCursor();
        return false;
    }

    QApplication::restoreOverrideCursor();
    return true;
}

bool IASProcessExecution::executeDetached(QString & cmd)
{
    bool status;

    QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));

    QStringList list = cmd.split(QRegExp("\\s+"));

    // Remove environment variables from the path string.
    list[0] = IASUtilities::unEnvCmd(list[0]);

    myStdout.clear();
    myStderr.clear();

    if(list[0].isEmpty())
    {
        myStderr.append("Invalid environment variable in path.");
        QApplication::restoreOverrideCursor();
        return false;
    }

    QString prog = list[0];
    QStringList argList = list;
    argList.removeFirst();

    // By using startDetached, we don't have to maintain any knowledge about
    // the child process or use "wait" to reap it later.  Qt takes care of
    // setting everything up correctly for us.
    status = QProcess::startDetached(prog, argList);

    QApplication::restoreOverrideCursor();

    return status;
}

bool IASProcessExecution::normalExit()
{
    if(myProcess->exitStatus() == QProcess::CrashExit)
    {
        return false;
    }

    return true;
}

int IASProcessExecution::exitStatus()
{
    return myProcess->exitCode();
}

QStringList IASProcessExecution::getStdout()
{
    return myStdout;
}

QStringList IASProcessExecution::getStderr()
{
    return myStderr;
}

void IASProcessExecution::readFromStdoutSlot()
{
    // Grab the raw byte stream output from stdout and put it into a string.
    QString output = myProcess->readAllStandardOutput();

    // Split the output string by the newline(s) and put the
    // resulting substring(s) into the list.
    myStdout.append(output.split('\n', QString::SkipEmptyParts));
}

void IASProcessExecution::readFromStderrSlot()
{
    // Grab the raw byte stream output from stderr and put it into a string.
    QString output = myProcess->readAllStandardError();

    // Split the output string by the newline(s) and put the
    // resulting substring(s) into the list.
    myStderr.append(output.split('\n', QString::SkipEmptyParts));
}
