
/*****************************************************************************
    Name: iasQtSystemStatus.cpp

    Purpose: To provide the implementation for the SystemStatus class.
 ****************************************************************************/

/* System Includes */

/* Qt Includes */
#include <QtCore>
#include <QtGui>

/* Local Includes */
#include "IASProcessExecution.h"
#include "IASQtSystemStatus.h"

/*****************************************************************************
    Name: isTaskRunning

    Purpose: To take a task name and a PID and compare them to the processes
             running on the system.  If a match is found return true else
             return false.

    Return Value: bool
         true:  A match was found.
        false:  No match was found or error.
 ****************************************************************************/
bool SystemStatus::isTaskRunning(const QString &taskName,
                                 const QString &processId)
{
    QString errorMessage("");

    /* Get the system user ID from the environment */
    QString systemUserId = getenv("IAS_SYSTEM_USER_ID");
    if (systemUserId.isEmpty())
    {
        errorMessage = "Failed to get IAS_SYSTEM_USER_ID from environment.\n";
        QMessageBox::critical(0, "System Status - Error", errorMessage);
        return false;
    }

    /* Build the command line.  Truncate the task name to 15 characters. */
    QString localTaskName(taskName);
    localTaskName.truncate(15);
    QString cmd = "pgrep -u " + systemUserId + " " + localTaskName;

    /* Execute the command using the IASProcessExecution class so that we can
       capture the STDOUT from the pgrep executable. */
    IASProcessExecution pe(0);
    if (!pe.execute(cmd))
    {
        errorMessage = "Failed to execute pgrep command.\n" + cmd;
        QMessageBox::critical(0, "System Status - Error", errorMessage);
        return false;
    }

    /* Determine if it ran successfully. */
    if (!pe.normalExit())
    {
        int result = pe.exitStatus();
        errorMessage = "pgrep command\n" + cmd + "\n"
                       "returned " + QString::number(WEXITSTATUS(result));
        QMessageBox::critical(0, "System Status - Error", errorMessage);
        return false;
    }

    /*************************************************************************
        Read the PIDs from the STDOUT and do a quick error check.

       - If no PIDs are returned, then the task is not running.
       - If one PID is returned, then the task is running.
       - If more than one PID is returned, then something is wrong so return
         false.
     ************************************************************************/
    QStringList pids = pe.getStdout();
    int pidCount = pids.size();

    if (pidCount == 0)
    {
        /* System isn't running */
        return false;
    }
    else if (pidCount > 1)
    {
        errorMessage = "pgrep returned more than 1 match for " + taskName;
        QMessageBox::critical(0, "System Status - Error", errorMessage);
        return false;
    }

    /* Verify that the provided PID matches the PID from the pgrep command. */
    if (processId == pids[0])
    {
        return true;
    }

    return false;
}

