/******************************************************************************
               Image Assessment System (IAS) Software
 ******************************************************************************
 *
 * Name:   IASSchemaConfig.cpp
 *
 * PURPOSE:
 * This is the definition of the IASSchemaConfig class which is used to 
 * get the schema name from database and return satellite and sensor value.
 *
 *****************************************************************************/

// System Includes

// QT Includes
#include <QtCore>
#include <QtGui>

// Local Includes
#include "IASSqlQuery.h"
#include "IASSchemaConfig.h"

// Default the instance variable to NULL.
IASSchemaConfig * IASSchemaConfig::myInstance = NULL;

// Don't do anything.
IASSchemaConfig::IASSchemaConfig()
{
}

// Used to create or get the instance.
IASSchemaConfig * IASSchemaConfig::instance()
{
    // If NULL then create one.
    if(myInstance == NULL)
    {
        myInstance = new IASSchemaConfig();

        // Initialize it.
        if(!myInstance->initialize())
        {
            // If failure delete the memory and initialize the pointer
            // back to NULL.
            delete myInstance;
            myInstance = NULL;
        }
    }

    return myInstance;
}

// Used to populate the internal class attributes.
bool IASSchemaConfig::initialize()
{
    IASSqlQuery query;
    QString errorMessage("");

    /* Figure out and save the common schema.
       It is not saved in the schema map because that is reserved for the
       satellite sensor schemas. */
    myCommonSchema = "";

    QString commonEnv = getenv("IAS_DB_COM");
    if(commonEnv.isEmpty())
    {
        errorMessage = "Environment Error:\n"
            "Failed to get IAS_DB_COM from environment";
        QMessageBox::critical(0, "System Status - Error", errorMessage);
        return false;
    }

    QStringList list = commonEnv.split(QRegExp("\\/"));
    if(list[0].isEmpty() || (list.count() != 2))
    {
        // Bad environment variable.
        errorMessage = "Environment Error:\n"
            "Invalid IAS_DB_COM environment variable";
        QMessageBox::critical(0, "System Status - Error", errorMessage);
        return false;
    }

    myCommonSchema = list[0];
    QString tableName = myCommonSchema + ".satellite_sensor_cfg";

    // This is where the system is configured for multiple schemas.
    QString sqlStmt = "select satellite, sensor, to_char(epoch, 'YYYY-MM-DD') "
                      "from " + tableName;

    if(!query.exec(sqlStmt, __FILE__, __LINE__))
    {
        errorMessage = "Database Error:\n"
            "Failed accessing table " + tableName;
        QMessageBox::critical(0, "System Status - Error", errorMessage);
        return false;
    }

    // While we have database results to process

    bool returnValue = false;
    while(query.next())
    {
        if(!query.isValid())
        {
            // Some kind of database error occurred.
            errorMessage = "Database Error:\n Query is invalid";
            QMessageBox::critical(0, "System Status - Error", errorMessage);
            return false;
        }

        QString sat = query.value(0).toString();
        QString sen = query.value(1).toString();
        QDate epoch = query.value(2).toDate();

        QString envName = "IAS_DB_" + sat + "_" + sen;

        QString envValue = getenv(envName.toAscii());
        if(!envValue.isEmpty())
        {
            list = envValue.split(QRegExp("\\/"));
            if(list[0].isEmpty() || (list.count() != 2))
            {
                // Bad environment variable.
                errorMessage = "Environment Error:\n"
                    "Invalid " + envName + " environment variable";
                QMessageBox::critical(0, "System Status - Error", errorMessage);
                return false;
            }
            mySchemaMap.insert(sat + "_" + sen, list[0] + ".");
            myEpochBySchemaMap.insert(list[0], epoch);
            returnValue = true;
        }
        else
        {
            // Missing environment variable.
            errorMessage = "Environment Error:\n"
                "Failed to get " + envName + " from environment";
            QMessageBox::critical(0, "System Status - Error", errorMessage);
            return false;
        }
    }

    if (!returnValue)
    {
        errorMessage = "Database Error:\n" + tableName + " not populated";
        QMessageBox::critical(0, "System Status - Error", errorMessage);
    }

    return returnValue;
}

// Used to get the schema name from an L0R_PRODUCT_ID.
QString IASSchemaConfig::schemaNameFromL0rProductId(QString productId)
{
    IASSqlQuery query;

    // Get the satellite and sensor.
    QString sqlStmt = "select satellite, sensor from " + myCommonSchema +
                      ".work_orders ";
    sqlStmt += "where l0r_product_id = '" + productId + "'";

    if(!query.exec(sqlStmt, __FILE__, __LINE__))
    {
        return QString("");
    }

    query.next();

    if(!query.isValid())
    {
        return QString("");
    }

    QString sat = query.value(0).toString();
    QString sen = query.value(1).toString();

    // Return the schema name stored in the map.
    return schemaName(sat, sen);
}

// Used to get the schema name from a ORDER_ID.
QString IASSchemaConfig::schemaNameFromOrderId(QString orderId)
{
    IASSqlQuery query;

    // Get the satellite and sensor.
    QString sqlStmt = "select satellite, sensor from " + myCommonSchema +
                      ".work_orders ";
    sqlStmt += "where order_id = '" + orderId + "'";

    if(!query.exec(sqlStmt, __FILE__, __LINE__))
    {
        return QString("");
    }

    query.next();

    if(!query.isValid())
    {
        return QString("");
    }

    QString sat = query.value(0).toString();
    QString sen = query.value(1).toString();

    // Return the schema name stored in the map.
    return schemaName(sat, sen);
}

// Used to get the schema name when the key is known.
QString IASSchemaConfig::schemaNameFromKey(QString key)
{
    return mySchemaMap.value(key, "");
}


// Used to figure out the schema from a WORK_ORDER_ID.
QString IASSchemaConfig::schemaNameFromWOId(QString woId)
{
    QString satSen = satSenFromWOId(woId);

    // This can be called because <sat>_<sen> is the key and that is what
    // is placed into the database and retrieved from the above call.
    if(!satSen.isEmpty())
    {
        return schemaNameFromKey(satSen);
    }

    return QString("");
}

// Used to get the schema name when the satellite and sensor are known.
QString IASSchemaConfig::schemaName(QString sat,
                                 QString sen)
{
    // Return the schema name stored in the map.
    return schemaNameFromKey(sat + "_" + sen);
}

// Used to get the common schema name.
QString IASSchemaConfig::commonSchemaName()
{
    // Return the common schema name.
    return myCommonSchema;
}

// Used to create a list of all the schemas.
QStringList IASSchemaConfig::allSchemas()
{
    QStringList values = mySchemaMap.values();

    return values;
}

// Used to create a list of all the schema keys.
QStringList IASSchemaConfig::allKeys()
{
    QStringList keys = mySchemaMap.keys();

    return keys;
}

// Used to figure out the satellite and sensor from a WORK_ORDER_ID.
QString IASSchemaConfig::satSenFromWOId(QString woId)
{
    IASSqlQuery query;

    // Setup the SQL to get the satellite and sensor.
    QString sqlStmt = "select satellite, sensor from " + myCommonSchema +
                      ".work_orders ";
    sqlStmt += "where work_order_id = '" + woId + "'";

    if(!query.exec(sqlStmt, __FILE__, __LINE__))
    {
        return QString("");
    }

    query.next();

    if(!query.isValid())
    {
        return QString("");
    }

    QString satSen = query.value(0).toString(); // Satellite
    satSen += "_";
    satSen += query.value(1).toString(); // Sensor

    // Return the satellite sensor value stored in the database.
    return satSen;
}

// Used to get the schema name when the key is known.
QDate IASSchemaConfig::epochDateFromSchemaName(QString schemaName)
{
    return myEpochBySchemaMap.value(schemaName, QDate());
}

