
/*!
    \file IASSqlQuery.cpp
    \brief Provides the implementation for IASSqlQuery.

    See IASSqlQuery.
*/

// Qt Includes
#include <QtCore>
#include <QtGui>
#include <QtSql>

// Local Includes
#include "IASSqlQuery.h"
#include "IASErrorDialog.h"

/*!
    This constructor allows the user to create a IASSqlQuery object.

    \param query Input: The SQL statement to execute.  If the query is a
                        null string then the query will not be executed.

    \param db Input: The database to execute queries against.
*/
IASSqlQuery::IASSqlQuery(const QString& query,
                         QSqlDatabase db)
: QSqlQuery(query, db)
{
}

/*!
    This method executes a query and checks the return code.  If an error
    occured the IASErrorDialog class will be used to prompt the operator.

    \param query Input: The SQL statement to execute.

    \param fileName Input: The name of the invoking file.
                           (Use __FILE__)

    \param lineNum Input: The line number from the invoking file.
                          (Use __LINE__)
*/
bool IASSqlQuery::exec(const QString &query,
                       QString fileName,
                       int lineNum)
{
    QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));
    bool returnVal = QSqlQuery::exec(query);    
    QApplication::restoreOverrideCursor();

    if(!returnVal)
    {
        IASErrorDialog::databaseError(0, "A database error occurred while "
                                      "trying to execute an SQL statement",
                                   fileName, lineNum, this);
    }
				   
    return returnVal;
}

/*!
    This method implements a variation of QSqlQuery::value that return the 
    value for a given column name.

    \param columnName Input: The column name
*/
QVariant IASSqlQuery::value(const QString &columnName) const
{
    // if the query isn't positioned on valid record return an invalid QVariant
    if (!isActive() || !isValid())
        return QVariant(QVariant::Invalid);

    return QSqlQuery::value(record().indexOf(columnName));
}

/*!
    Return a QStringList contain the values of each column in the query.
*/
QStringList IASSqlQuery::toStringList() const
{
    // if the query isn't positioned on valid record return an empty list
    if (!isActive() || !isValid())
        return QStringList();

    QStringList values;
    int numColumns = record().count();

    for (int col = 0; col < numColumns; col++)
        values << QSqlQuery::value(col).toString();

    return values;
}
