
/*!
    \file IASUtilities.cpp
    \brief Provides the implementation for IASUtilities.

    See IASUtilities.
*/

/* System Includes */
#include <stdio.h>
#include <stdlib.h>
#include <float.h>
#include <math.h>
#include <iostream>

/* Qt Includes */
#include <QtCore>

/* Local Includes */
#include "iasQtWoConst.h"
#include "IASSchemaConfig.h"
#include "IASUtilities.h"
#include "IASSqlQuery.h"
#include "IASDatabaseErrorDialog.h"

/*!
    This function will take a Date/Time string and convert it into a
    QDateTime object.  The format is assumed to be "YYYY-MM-DD[/|T]HH:MM:SS".
    If the string is empty a default date and time of
    "1980-01-01/00:00:00" will be used.

    \param str Input: The Date/Time string to use.
*/
QDateTime IASUtilities::getDateTime(const QString str)
{
    QString strDate;
    QString strTime;
    QChar sep;

    // Find the Date and Time sections by looking for the '/' or 'T' separator
    if (str.isEmpty())
    {
        strDate = "1980-01-01";
        strTime = "00:00:00";
    }
    else
    {
        if (str.count('T') == 1)
            sep = 'T';
        else
            sep = '/';
        strDate = str.section(sep,0,0);
        strTime = str.section(sep,1,1);
    }

    QDateTime tmpDateTime(QDate::fromString(strDate,Qt::ISODate),
                          QTime::fromString(strTime,Qt::ISODate));

    return tmpDateTime;
}

/*!
    This function will take a Date string and convert it to a QDate
    object.  It is meant to be used for IASListView objects where columns
    can have a specific type.  It can convert dates of the following
    types: DateYMD, DateMDY, DateDMMMY, and DateYYJJJ

    \param str Input: The Date/Time string to use.

    \param columnType Input: The type to use during the conversion.
*/
QDate IASUtilities::getDate(const QString str,
                            columnDataTypes columnType)
{
    QString strDate,year,month,day,julian;
    int year_num,julian_num,i;
    QDate tmpDate;

    strDate = str;

    tmpDate.setYMD(80,1,1);

    if (!strDate.isEmpty())
    {
        switch (columnType)
        {
            case DateYMD:
                tmpDate = QDate::fromString(strDate,Qt::ISODate);
                break;

            case DateMDY:
                year = strDate.section('-',2,2);
                month = strDate.section('-',1,1);
                day = strDate.section('-',0,0);
                tmpDate.setYMD(year.toInt(NULL),month.toInt(NULL),
                               day.toInt(NULL)); 
                break;

            case DateDMMMY:
                year = strDate.section('-',2,2);
                month = strDate.section('-',1,1);
                day = strDate.section('-',0,0);
                tmpDate.setYMD(year.toInt(NULL),getMonth(month),
                               day.toInt(NULL));
                break;

            case DateYYJJJ:
                {
                    int non_leap[13] 
                        = {0,31,59,90,120,151,181,212,243,273,304,334,365};
                    int leap[13] 
                        = {0,31,60,91,121,152,182,213,244,274,305,335,366};

                    year = strDate.section('-',0,0);
                    year_num = year.toLong(NULL);
                    julian = strDate.section('-',1,1);
                    julian_num = julian.toLong(NULL);

                    /* Leap year
                       ------------*/
                    if (QDate::isLeapYear(year_num))
                    {
                        for (i=1; i < 13; i++)
                        {
                            if (julian_num <= leap[i])
                            {
                                tmpDate.setYMD(year_num,julian_num - leap[i],i);
                                break;
                            }
                        }
                    }
                    else
                    {
                        /* Non-leap year
                           ----------------*/
                        for (i=1; i < 13; i++)
                        {
                            if (julian_num <= non_leap[i])
                            {
                                tmpDate.setYMD(year_num,
                                               julian_num - non_leap[i], i);
                                break;
                            }
                        }
                    }
                }
                break;

            default:
                qDebug() << __FILE__ << ":" << __LINE__ <<
                         ":Error: Calling getDate with invalid column type" <<
                         " for a date";
                break;
        }
    }

    return tmpDate;
}

/*!
    This function will take a Time string and convert it to a QTime
    object.  It accepts input in the format "HH:MM:SS".

    \param str Input: The input time string to use.
*/
QTime IASUtilities::getTime(const QString str)
{
    QString strTime;
    QTime tmpTime;

    strTime = str;

    if (strTime.isEmpty())
        tmpTime.setHMS(00,00,00);
    else
        tmpTime = QTime::fromString(strTime,Qt::ISODate);

    return tmpTime;
}

/*!
    This function will take a Name of a month and return the number of
    the month.  The month must be given the format "MMM".

    \param month Input: The name of the month to convert.
*/
int IASUtilities::getMonth(const QString month)
{
    if (month == "JAN")
        return 1;
    else if (month == "FEB")
        return 2;
    else if (month == "MAR")
        return 3;
    else if (month == "APR")
        return 4;
    else if (month == "MAY")
        return 5;
    else if (month == "JUN")
        return 6;
    else if (month == "JUL")
        return 7;
    else if (month == "AUG")
        return 8;
    else if (month == "SEP")
        return 9;
    else if (month == "OCT")
        return 10;
    else if (month == "NOV")
        return 11;
    else if (month == "DEC")
        return 12;
    else
        return 0;
}

/*!
    This converts the number of seconds from an epoch into a date format

    \param secs Input: The number of seconds from the epoch date.
    \param epochDate Input: The epoch date to use.
*/
QString IASUtilities::changeSecsToDate(int secs,
                                       QDate epochDate)
{
    QString str_date;
    QDate epoch(epochDate);

    int total_secs,total_days,time_of_day,hrs,min,seconds;
    QString ch_hrs;
    QString ch_min;
    QString ch_sec;

    if (secs < 1) return "";

    // Compute the total number of days
    total_secs=secs;
    total_days=(total_secs/86400);

    str_date = epoch.addDays(total_days).toString(Qt::ISODate);

    // For appearance issues, add a space to the end of the date
    str_date += "/";

    // Compute the hour, minutes, seconds
    time_of_day=(secs%SEC_IN_DAY);
    hrs=(time_of_day / 3600);
    min=(time_of_day / 60) - (hrs * 60);
    seconds=(time_of_day%60);

    ch_hrs=QString::number(hrs,10);
    ch_min=QString::number(min,10);
    ch_sec=QString::number(seconds,10);
    if (ch_hrs.length() == 1)
        str_date += "0";
    str_date += ch_hrs + ":";

    if (ch_min.length() == 1)
        str_date += "0";
    str_date += ch_min + ":";

    if (ch_sec.length() == 1)
        str_date += "0";
    str_date += ch_sec;

    return str_date;
}

/*!
    This accesses the database and determines if the given work order ID
    exists in the work_orders table.  Returns true if the work order ID was
    found and false if it was not found.

    \param woId Input: The Work Order ID to find.
*/
bool IASUtilities::dbValidWoId(const QString& woId)
{
    QString sqlStmt("");
    IASSqlQuery query;
    QString result("");

    QString comName = IASSchemaConfig::instance()->commonSchemaName();

    sqlStmt = "select work_order_id "
                  "from " + comName + ".work_orders "
                  "where work_order_id = '" + woId + "'";

    if(query.exec(sqlStmt, __FILE__, __LINE__))
    {
        if(query.next())
        {
            result = query.value(0).toString();
        }
    }

    if(result != woId)
    {
        return false;
    }

    return true;
}

/*!
    This public function updates the given work order ID and any failed scripts
    to pending if the current state of the work order is halted.

    If the work order ID cannot be found, or for some reason cannot be updated,
    false is returned.

    \param woID Input: The Work Order ID to reset to a pending state
    \param errorMessage Output: Returns the reason why the function failed
    \param displayMessage Input: Flags whether a dialog or messagebox should
    be shown detailing the error.
*/

bool IASUtilities::resetWOPending(
    const QString &woID,
    QString &errorMessage,
    const bool &displayMessage=false)
{
    bool errorShown;  /* This tells us if we have already displayed the
    Database Error Dialog */
    bool result;
    result = resetWOPending(woID, errorMessage, displayMessage, errorShown);
    if (!result && displayMessage && !errorShown)
    {
         QMessageBox::critical(0, "Resume Error",
            "Work order '" + woID + "' could not be reset back to pending due "
            "to:\n" + errorMessage, QMessageBox::Ok, 0);
    }
    return result;
}

/*!
    This private function actually does the work of updating the work order.
    It is identical to the public function, except it has an additional
    parameter that indicates whether the database error dialog was shown.

    \param woID Input: The Work Order ID to reset to a pending state
    \param errorMessage Output: Returns the reason why the function failed
    \param displayMessage Input: Flags whether a database error dialog should
    be shown detailing the error.
    \param errorShown Output: Whether the database error dialog was shown.

*/
bool IASUtilities::resetWOPending(
    const QString &woID,
    QString &errorMessage,
    const bool &displayMessage,
    bool &errorShown)
{
    errorShown = false;
    const QString genericSQLError =
        "A database error occurred while trying to execute an SQL statement";

    QString nodeClause(" ");
    QString systemID = getenv("IAS_SYSTEM_ID");
    if (systemID.isEmpty())
    {
        errorMessage = "Environment variable: IAS_SYSTEM_ID is not set.";
        return false;
    }
    if (systemID == "L")
    {
        nodeClause = ", processing_node = NULL ";
    }
    // Use the default database connection
    QSqlDatabase database = QSqlDatabase::database();
    if (!database.isValid())
    {
        errorMessage = database.lastError().text();
        return false;
    }
    if (!database.transaction())
    {
        errorMessage = "Transaction cannot be started or is unsupported.";
        return false;
    }

    QSqlQuery query;
    query.prepare("UPDATE work_orders "
                    "SET state = '" + DB_WO_PENDING + "'" +
                    nodeClause +
                    "WHERE work_order_id = :woID " +
                    "AND state = '" + DB_WO_HALTED + "'");
    query.bindValue(":woID", woID);
    if (!query.exec())
    {
        errorMessage = query.lastError().text();
        database.rollback();
        if (displayMessage)
        {
            IASErrorDialog::databaseError(0, genericSQLError,
                __FILE__, __LINE__, &query);
            errorShown = true; /* Flag caller that there's no need to
            display the error, now that we already did. */
        }
       return false;
    }
    if (query.numRowsAffected() != 1)
    {
        errorMessage = "Work order '" + woID +"' could not be found.";
        database.rollback();
        return false;
    }
    query.prepare("UPDATE wo_scripts"
                  " SET state = '" + DB_WO_SCRIPT_PENDING + "',"
                      " compl_status = NULL"
                  " WHERE work_order_id = :woID"
                      + " AND compl_status = '"
                          + DB_WO_SCRIPT_STATUS_FAILURE + "'");
    query.bindValue(":woID", woID);
    if (!query.exec())
    {
        errorMessage = query.lastError().text();
        database.rollback();
        if (displayMessage)
        {
            IASErrorDialog::databaseError(0, genericSQLError,
                                   __FILE__, __LINE__, &query);
            errorShown = true;
        }
        return false;
    }
    if (!database.commit())
    {
        errorMessage = database.lastError().text();
        database.rollback();
        return false;
    }
    return true;
}
/*!
    This reads a "key = value" input file and returns a KeyValueMap, to
    the caller, populated with the results of the grep/search.  If a key
    is not found in the file the key will not be in the map returned to
    the caller.  All greps/searches are performed using upper case.  The
    keys returned to the caller will all be upper case.

    \param filename Input: The full path to the file to grep/search.
    \param list Input: The list of uppercase keys to search for.
*/
KeyValueMap IASUtilities::readKeyValueFile(const QString& filename,
                                           QStringList& list)
{
    KeyValueMap kvm;
    QStringList upperList;
    QStringList::iterator itr;
    QFile fh(filename);

    kvm.clear();
    upperList.clear();

    if(!fh.open(QIODevice::ReadOnly))
    {
        return kvm;
    }

    /* Make sure we have an uppercase list to search for. */
    for(itr = list.begin(); itr != list.end(); ++itr)
    {
        upperList.append((*itr).toUpper());
    }

    QTextStream ts(&fh);
    QString tempStr("");
    while(!ts.atEnd())
    {
        tempStr = ts.readLine().simplified();
        QStringList lineList = tempStr.split(QRegExp(" "));

        QString value = lineList[0].toUpper();
        if(upperList.contains(value))
        {
            kvm[value] = lineList[2];
        }
    }

    fh.close();

    return kvm;
}

/*!
    This processes a command line and replaces environment variables with
    the values of the variables.  This will also remove duplicate /'s.

    \param cmd Input: The command line to search for environment variables
                      and replace them with the values.
*/
QString IASUtilities::unEnvCmd(const QString& cmd)
{
    QString inputCmd = cmd;
    QString outputCmd("");
    QString part("");

    QStringList splitCmd = inputCmd.split(QRegExp("\\/"));

    QStringList::iterator itr;
    for(itr = splitCmd.begin(); itr != splitCmd.end(); ++itr)
    {
        part = (*itr);

        if(part.isEmpty())
        {
            // Skip empty slashes "/////" that are in the path.
            continue;
        }

        /* Only add /'s if we are after the first part and have more parts.
           This will also catch a leading /. */
        if(itr != splitCmd.begin())
        {
            outputCmd += "/";
        }

        if(part.contains(QRegExp("\\$")))
        {
            part.replace('$', ""); // Remove the $ so getenv works.
            part = getenv(part.toAscii());

            if(part.isEmpty())
            {
                return "";
            }
        }

        outputCmd += part;
    }

    /* If the last part was empty, then we had a trailing / */
    if(part.isEmpty())
        outputCmd += "/";

    return outputCmd;
}

/* This method will compare two numerical data QStrings using less than.
   This method allows for an empty QString to be sorted correctly.
   This method is intended to be used by qSort. */
bool IASUtilities::numericLessThan(const QString &first,
                                   const QString &second)
{
    if(first == "")
    {
        return true;
    }

    if(second == "" && first != "")
    {
        return false;
    }

    if(second == "" && first == "")
    {
        return false;
    }

    if(first.toDouble() < second.toDouble())
    {
        return true;
    }

    return false;
}

/* These are used in the state convertion routines below. */
KeyValueMap stateValueToText;
KeyValueMap stateTextToValue;

void initializeStates()
{
    stateValueToText[DB_WO_L0R_NEW] = UI_WO_L0R_NEW_STR;
    stateValueToText[DB_WO_REQUEST_FAILED] = UI_WO_REQUEST_FAILED_STR;
    stateValueToText[DB_WO_L0R_REQUESTED] = UI_WO_L0R_REQUESTED_STR;
    stateValueToText[DB_WO_SUBSETTER_FAILED] = UI_WO_SUBSETTER_FAILED_STR;
    stateValueToText[DB_WO_L0R_READY] = UI_WO_L0R_READY_STR;
    stateValueToText[DB_WO_INGEST_FAILED] = UI_WO_INGEST_FAILED_STR;
    stateValueToText[DB_WO_PENDING] = UI_WO_PENDING_STR;
    stateValueToText[DB_WO_RUNNING] = UI_WO_RUNNING_STR;
    stateValueToText[DB_WO_HALTED] = UI_WO_HALTED_STR;
    stateValueToText[DB_WO_EVALUATE] = UI_WO_EVALUATE_STR;
    stateValueToText[DB_WO_INACTIVE] = UI_WO_INACTIVE_STR;
    stateValueToText[DB_WO_DIAGNOSTIC_COMPLETE] = UI_WO_DIAGNOSTIC_COMPLETE_STR;
    stateValueToText[DB_WO_READY_TO_SHIP] = UI_WO_READY_TO_SHIP_STR;
    stateValueToText[DB_WO_TRANSFER_FAILED] = UI_WO_TRANSFER_FAILED_STR;
    stateValueToText[DB_WO_ABORT_WO] = UI_WO_ABORT_WO_STR;
    stateValueToText[DB_WO_COMPLETE_WO] = UI_WO_COMPLETE_WO_STR;
    stateValueToText[DB_WO_RELEASE_ABORT] = UI_WO_RELEASE_ABORT_STR;
    stateValueToText[DB_WO_RELEASE_COMPLETE] = UI_WO_RELEASE_COMPLETE_STR;
    stateValueToText[DB_WO_ABORTED] = UI_WO_ABORTED_STR;
    stateValueToText[DB_WO_COMPLETED] = UI_WO_COMPLETED_STR;

    stateTextToValue[UI_WO_L0R_NEW_STR] = DB_WO_L0R_NEW;
    stateTextToValue[UI_WO_REQUEST_FAILED_STR] = DB_WO_REQUEST_FAILED;
    stateTextToValue[UI_WO_L0R_REQUESTED_STR] = DB_WO_L0R_REQUESTED;
    stateTextToValue[UI_WO_SUBSETTER_FAILED_STR] = DB_WO_SUBSETTER_FAILED;
    stateTextToValue[UI_WO_L0R_READY_STR] = DB_WO_L0R_READY;
    stateTextToValue[UI_WO_INGEST_FAILED_STR] = DB_WO_INGEST_FAILED;
    stateTextToValue[UI_WO_PENDING_STR] = DB_WO_PENDING;
    stateTextToValue[UI_WO_RUNNING_STR] = DB_WO_RUNNING;
    stateTextToValue[UI_WO_HALTED_STR] = DB_WO_HALTED;
    stateTextToValue[UI_WO_EVALUATE_STR] = DB_WO_EVALUATE;
    stateTextToValue[UI_WO_INACTIVE_STR] = DB_WO_INACTIVE;
    stateTextToValue[UI_WO_DIAGNOSTIC_COMPLETE_STR] = DB_WO_DIAGNOSTIC_COMPLETE;
    stateTextToValue[UI_WO_READY_TO_SHIP_STR] = DB_WO_READY_TO_SHIP;
    stateTextToValue[UI_WO_TRANSFER_FAILED_STR] = DB_WO_TRANSFER_FAILED;
    stateTextToValue[UI_WO_ABORT_WO_STR] = DB_WO_ABORT_WO;
    stateTextToValue[UI_WO_COMPLETE_WO_STR] = DB_WO_COMPLETE_WO;
    stateTextToValue[UI_WO_RELEASE_ABORT_STR] = DB_WO_RELEASE_ABORT;
    stateTextToValue[UI_WO_RELEASE_COMPLETE_STR] = DB_WO_RELEASE_COMPLETE;
    stateTextToValue[UI_WO_ABORTED_STR] = DB_WO_ABORTED;
    stateTextToValue[UI_WO_COMPLETED_STR] = DB_WO_COMPLETED;
}

QString IASUtilities::getStateValue(const QString& stateText)
{
    if (stateValueToText.isEmpty()) /* If one is empty they both are. */
    {
        initializeStates();
    }

    return stateTextToValue[stateText];
}

QString IASUtilities::getStateText(const QString& stateValue)
{
    if (stateValueToText.isEmpty()) /* If one is empty they both are. */
    {
        initializeStates();
    }

    return stateValueToText[stateValue];
}

QStringList IASUtilities::getListOfStateText()
{
    if (stateValueToText.isEmpty()) /* If one is empty they both are. */
    {
        initializeStates();
    }

    return stateValueToText.values();
}

/*!
    Checks the IAS_SYS_DIR and HOME directories for a plugin file.  Calls
    loadPluginFile for each plugin file found.  All plugins found will be
	added to menuBar and all QActions mapped to their command using the
	mapper. 

    \param menuBar Input: Pointer to a QMenuBar
    \param mapper Input: Pointer to a QSignalMapper
    \param filename Input: Plugin filename (without path)
*/
void IASUtilities::addPlugins(QMenuBar *menuBar, 
                              QSignalMapper *mapper,
                              QString filename)
{
    if (menuBar == NULL)
        return;

    if (mapper == NULL)
        return;

    // Check for the system level plugin file
    QString sys_dir = getenv("IAS_SYS_DIR");
    if (!sys_dir.isEmpty())
    {
        QFile file(sys_dir + "/etc/" + filename);
        if (file.exists())
        {
            loadPluginFile(menuBar, mapper, file.fileName());
        }
    }

    // Check for user plugin file in user's HOME directory
    QString homeDir = getenv("HOME");
    if (!homeDir.isEmpty())
    {
        QFile file(homeDir + "/." + filename);
        if (file.exists())
        {
            loadPluginFile(menuBar, mapper, file.fileName());
        }
    }

    return;
}

/*!
    Reads a plugin file and calls insertPlugin for each plugin listed. All
	plugins found in filename will be added to menuBar and the QActions
	mapped to their commands using the mapper.

    \param menuBar Input: Pointer to a QMenuBar
    \param mapper Input: Pointer to a QSignalMapper
    \param filename Input: Fully qualified plugin filename
*/
void IASUtilities::loadPluginFile(QMenuBar *menuBar, 
                                  QSignalMapper *mapper,
                                  QString filename)
{
    if (menuBar == NULL) 
        return;

    if (mapper == NULL)
        return;

    QFile file(filename);
    if (file.open(QIODevice::ReadOnly))
    {
        QTextStream stream(&file);
        QString line;
        while (!stream.atEnd())
        {
            line = stream.readLine();
            if (!line.startsWith("PLUGIN"))
                continue;

            // A plugin line has three parameters: PLUGIN|<Menu Path>|<command>
            QStringList parms = line.split(QRegExp("\\|"));
            if (parms.count() < 3)
                continue;

            // The menu path is comma delimited
            QStringList path = parms[1].split(QRegExp("\\,")); 
            
            QString cmd = parms[2].simplified();
            
            // add plugin
            insertPlugin(path, cmd, mapper, menuBar);
        }
    }

    return;
}

/*!
    A function recursively builds a QMenu tree under the parent object. A
    QAction will be created for the last item in the menuPath.  The mapper
    is setup to associate the QAction with the cmd.

    \param menuPath Input: QStringList containing the path for plugin.
    \param cmd Input: Command associated with the plugin.
    \param mapper Input: Pointer to a QSignalMapper.
    \param parent Input: QObject point to either a QMenu or QMenuBar. A QAction
        for the menuPath will be created under parent.
*/
void IASUtilities::insertPlugin(QStringList menuPath, 
                                QString cmd,
                                QSignalMapper *mapper, 
                                QObject *parent)
{
    if (parent == NULL) 
        return;

    if (menuPath.isEmpty()) 
        return;

    QStringList newPath = menuPath;

    // Take to first item in the menu path
    QString menuText = newPath.takeFirst().simplified();

    QString objectName = menuText;
    objectName.replace("&", "");
    objectName.replace(" ", "");

    // Pointers to various QT types
    QMenuBar *tmpMenuBar = NULL;
    QMenu *tmpMenu       = NULL;
    QMenu *menuItem      = NULL;
    QAction *actionItem  = NULL;

    // Attempt to find the current path item in the parent object
    tmpMenu = qobject_cast<QMenu*>(parent);
    if (tmpMenu == NULL)
    {
        tmpMenuBar = qobject_cast<QMenuBar*>(parent);
        if (tmpMenuBar != NULL)
            menuItem = tmpMenuBar->findChild<QMenu*>(objectName);
    }
    else
    {
        menuItem = tmpMenu->findChild<QMenu*>(objectName);
    }

    if (menuItem != NULL)
    {
        // A menu item already exists for the current object in the 
        // menuPath. Move on to the next item in the path using the current 
        // menuItem as the parent.
        insertPlugin(newPath, cmd, mapper, menuItem);
    }
    else
    {
        // A menu does not yet exist for the current item in the path.

        if (tmpMenu != NULL)
        {
            if (newPath.count() >= 1)
            {
                menuItem = new QMenu(menuText, tmpMenu);
                menuItem->setObjectName(objectName);
                tmpMenu->addMenu(menuItem);

                insertPlugin(newPath, cmd, mapper, menuItem);
            }
            else
            {
                actionItem = new QAction(menuText, tmpMenu);
                tmpMenu->addAction(actionItem);
                
                mapper->setMapping(actionItem, cmd);
                QObject::connect(actionItem, SIGNAL(triggered(bool)),
                    mapper, SLOT(map())); 
            }
        }
        else if (tmpMenuBar != NULL)
        {
            if (newPath.count() >= 1)
            {
                menuItem = new QMenu(menuText, tmpMenuBar);
                menuItem->setObjectName(objectName);
                tmpMenuBar->addMenu(menuItem);

                insertPlugin(newPath, cmd, mapper, menuItem);
            }
            else
            {
                actionItem = new QAction(menuText, tmpMenuBar);
                tmpMenuBar->addAction(actionItem);
                
                mapper->setMapping(actionItem, cmd);
                QObject::connect(actionItem, SIGNAL(triggered(bool)),
                    mapper, SLOT(map())); 
            }
        }
    }

    return;
}
