/*-----------------------------------------------------------------------------

NAME: ias_l0r_mta_private.h

PURPOSE: Interval functions for reading and writing metadata contained in the 
    L0R data.

ALGORITHM REFERENCES: NONE

NOTES:

-----------------------------------------------------------------------------*/
#include <string.h>

/* project included headers */
#include "ias_l0r.h"
#include "ias_l0r_hdf.h"
#include "ias_l0r_mta_private.h"
#include "ias_logging.h"


/*******************************************************************************
 private metadata routines
*******************************************************************************/
/******************************************************************************
 NAME: ias_l0r_establish_mta_file

 PURPOSE: Initializes access to or creates the specified metadata file

 RETURNS: SUCCESS- Access had been established
          ERROR- Could not establish access
******************************************************************************/
int ias_l0r_mta_establish_file
(
    HDFIO *hdfio_ptr, /* I: Pointer to structure used in I/O */
    int create_if_absent /* I: Create the file if it does not exist */
)
{
    int status = 0;

    char metadata_filename[IAS_L0R_FILE_NAME_LENGTH];
    char *metadata_appendix = "_MTA.h5\0";

    if (hdfio_ptr->access_mode_metadata == -1)
    {
        IAS_LOG_ERROR("Metadata file is not open");
        return ERROR;
    }
   
    status = snprintf(metadata_filename, IAS_L0R_FILE_NAME_LENGTH, "%s%s",
        hdfio_ptr->l0r_name_prefix,
        metadata_appendix);
  
    if (status < 0 || status >= IAS_L0R_FILE_NAME_LENGTH)
    {
        IAS_LOG_ERROR("Metadata file name of %s%s is too long. Max is %i",
            hdfio_ptr->l0r_name_prefix,metadata_appendix,
            IAS_L0R_FILE_NAME_LENGTH);
        return ERROR;
    }

    return ias_l0r_hdf_establish_versioned_file(hdfio_ptr->path,
        metadata_filename, &hdfio_ptr->file_id_metadata, 
        &hdfio_ptr->metadata_lib_version, hdfio_ptr->access_mode_metadata,
        create_if_absent);
}

/* subroutine used in reading metadata */
/******************************************************************************
 NAME: ias_l0r_mta_read

 PURPOSE: Internal routines to read the mta from the table name specified

 RETURNS: SUCCESS- Data was read in the buffer
          FAILURE- Unable to read into the buffer
******************************************************************************/
int ias_l0r_mta_read
(
    HDFIO *hdfio_ptr, /* I: Pointer to structure used in I/O */
    const int index, /* I: First record to read */
    const int count, /* I: Number of records to read */
    const char *table_name, /* I: Table name to read from */
    const size_t size, /* I: Size of the structure which stores the data */
    const size_t *offset, /* I: List of the offsets for each member */
    const size_t *size_list, /* I: List of the sizes of each member */
    void *buffer /* I: Data Buffer */
)
{
    int temp_size = 0;
    herr_t status;

    /* clear the buffer to initialize structure members that are not in the
       table, useful for reading older files that may not contain new fields */
    memset(buffer, 0, size * count );

    status = ias_l0r_mta_establish_file(hdfio_ptr, FALSE);
    if (status == ERROR || hdfio_ptr->file_id_metadata <= 0)
    {
        IAS_LOG_ERROR("Error establishing file for read");
        return ERROR;
    }
 
    /* verify there are an adequate number of records for the read */
    if (ias_l0r_mta_table_records_count(hdfio_ptr, table_name,
         &temp_size) == ERROR)
    {
        IAS_LOG_ERROR("Unable to get size of %s", table_name);
        return ERROR;
    }
    if (temp_size < (index + count) )
    {
        IAS_LOG_ERROR("Insufficient records to read from %s"
            " Attempting to read %d "
            "records out of %d from %d", table_name, count, temp_size, index);
        return ERROR;
    }

    status = H5TBread_records(hdfio_ptr->file_id_metadata,
        table_name, index, count, size,
        offset, size_list, buffer);
    if (status < 0)
    {
        IAS_LOG_ERROR("Error reading from %s", table_name);
        return ERROR;
    }

    return SUCCESS;
}

/******************************************************************************
 NAME: ias_l0r_mta_table_records_count

 PURPOSE: Internal routine to read the size of mta table with the name
          specified

 RETURNS: SUCCESS- Size was read
          FAILURE- Unable to read size
******************************************************************************/
int ias_l0r_mta_table_records_count
(
    HDFIO *hdfio_ptr, /* I: Pointer to structure used in I/O */
    const char *table, /* I: Table name to read from */
    int *size          /* O: Number of records in the table */
)
{
    int status;
    status = ias_l0r_mta_establish_file(hdfio_ptr, FALSE);
    if (status == ERROR)
    {
        *size = 0;
        return ERROR;
    }
    /* if there was no file opened but the routine was successful the file 
       does not exist, the case for size 0 */
    else if (hdfio_ptr->file_id_metadata < 1)
    {
        *size = 0;
        return SUCCESS;
    }

    /* if the file exists and there wasn't an error check the size
       of the table */
    status = ias_l0r_hdf_table_records_count(hdfio_ptr->file_id_metadata,
         table, size);
    if (status == ERROR)
    {
        *size = 0;
        return ERROR;
    }

    return SUCCESS;
}

