/*-----------------------------------------------------------------------------

NAME: ias_l0r_mta_v5.c

PURPOSE: Implementation of the routines that provided backwards compatibility
    with metadata created prior to L0R format version 6.

ALGORITHM REFERENCES: NONE

NOTES:

-----------------------------------------------------------------------------*/
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

/* project included headers */
#include "ias_const.h"
#include "ias_l0r.h"
#include "ias_l0r_hdf.h"
#include "ias_l0r_version.h"
#include "ias_l0r_mta_private.h"
#include "ias_l0r_mta_v5.h"
#include "ias_logging.h"
#include "ias_structures.h"
#include "ias_types.h"

#define DATASET_NAME_MTA_INTERVAL "Interval"
#define TABLE_NAME_INTERVAL_METADATA "Interval"
#define TABLE_SIZE_INTERVAL_METADATA 63

#define COPY_CORNERS(cd, cs) \
    cd.upleft.lat = cs.upleft.lat; \
    cd.upleft.lng = cs.upleft.lng;  \
    cd.upright.lat = cs.upright.lat; \
    cd.upright.lng = cs.upright.lng; \
    cd.loleft.lat = cs.loleft.lat; \
    cd.loleft.lng = cs.loleft.lng; \
    cd.loright.lat = cs.loright.lat; \
    cd.loright.lng = cs.loright.lng;

/*******************************************************************************
*Metadata structures

Time Format:  For fields present in the metadata which indicate a time, shall
              in the following format: YYYY:DDD:HH:MM:SS.SSSSSSS
              where:
              YYYY = four-digit Julian year,
              DDD = day (001-366*),
              HH = hours (00-23),
              MM = minutes (00-59),
              SS = seconds (00-59),
              SSSSSSS = fractional seconds (0-9999999)
              * For cases when active imaging occurs at the end of a leap year.
              Time is in TAI from the spacecraft.

*******************************************************************************/
/**
*This struct contains the interval level metadata
*
*Contents are listed in alphabetical order as found in the LMDD
*/
typedef struct IAS_L0R_INTERVAL_METADATA_V5
{
    char ancillary_start_time[IAS_L0R_DATE_LENGTH+1];
    /* Year, Day of year and spacecraft start time
        associated with the first group ancillary data received */
    char ancillary_stop_time[IAS_L0R_DATE_LENGTH+1];
    /* Year, Day of year and spacecraft start time
        associated with the last group ancillary data received */
    uint32_t attitude_points; /*good spacecraft attitude points */
    uint32_t attitude_points_missing; /*spacecraft attitude points missing*/
    uint32_t attitude_points_rejected; /*spacecraft attitude points rejected*/
    char collection_type[IAS_L0R_COLLECTION_TYPE_LENGTH+1];
    IAS_L0R_CORNERS corners_oli; /* OLI corner points of the interval */
    IAS_L0R_CORNERS corners_tirs; /* TIRS corner points of the interval */
    char cpf_name[IAS_L0R_CPF_NAME_LENGTH_V5+1];  /* CPF used for processing */
    uint32_t crc_errors_oli; /* Number of frame CRC failures for OLI */
    uint32_t crc_errors_tirs; /* Number of frame CRC failures for TIRS */
    char data_type[IAS_L0R_DATA_TYPE_LENGTH+1]; /*data type identifier string */
    char date_acquired[IAS_L0R_DATE_LENGTH+1];
    /* Year, Day of year and GMT that this scene was imaged */
    uint32_t ephemeris_points; /* good ephemeris data points received */
    uint32_t ephemeris_points_missing; /* ephemeris data points found missing */
    uint32_t ephemeris_points_rejected;/* ephemeris data points found to fail */
    uint32_t frames_filled_oli;/*number of OLI frames that were filled*/
    uint32_t frames_filled_tirs;/*number of TIRs frames that were filled*/
    char hostname[IAS_L0R_HOSTNAME_LENGTH+1];
    /* machine where data was processed*/
    int8_t image_quality_oli;
    int8_t image_quality_tirs;
    uint32_t interval_frames_oli; /*number of OLI frames reported
         in this interval file */
    uint32_t interval_frames_tirs; /*number of TIRS frames reported
         in this interval file */
    uint8_t interval_number;/*Sequence of the interval within the ingest file */
    uint8_t interval_version;/*Processed version of the interval, starts at 0*/
    char is_version[IAS_SOFTWARE_VERSION_SIZE+1];
    /* Version number of the software installed on the IS when a metadata file
    * and its associated Level 0R files were generated */
    char landsat_cal_interval_id[IAS_L0R_INTERVAL_ID_LENGTH+1];
    /* Unique Landsat calibration interval identifier*/
    char landsat_interval_id[IAS_L0R_INTERVAL_ID_LENGTH+1];
    char nadir_offnadir[IAS_L0R_NADIR_LENGTH+1];
    /* Nadir or Off-Nadir condition of the interval */
    char quality_algorithm[IAS_L0R_ALGORITHM_NAME_LENGTH+1];
    /*Algorithm (date, name, and version number) used to calculate the
        quality for this interval*/
    float roll_angle; /*amount of spacecraft roll at interval center */
    uint8_t satellite; /* Landsat vehicle number that captured the data */
    char sensor_id[IAS_L0R_SENSOR_ID_LENGTH+1]; /* name of the sensor  */
    char spacecraft_id[IAS_L0R_SPACECRAFT_ID_LENGTH+1];
    /*name of the spacecraft*/
    char start_time_oli[IAS_L0R_DATE_LENGTH+1];
    /* Year, Day of year and GMT spacecraft start time of the first OLI frame
         of the interval */
    char start_time_tirs[IAS_L0R_DATE_LENGTH+1];
    /* Year, Day of year and GMT spacecraft start time of the first TIRS frame
        of the interval */
    char station_id[IAS_L0R_STATION_ID_LENGTH+1]; /* receiving group
        station code */
    char stop_time_oli[IAS_L0R_DATE_LENGTH+1];
    /* Year, Day of year and spacecraft start time of the last OLI frame
        of the interval */
    char stop_time_tirs[IAS_L0R_DATE_LENGTH+1];
    /* Year, Day of year and spacecraft start time of the last TIRS frame
        of the interval */
    uint32_t time_code_errors_oli;/*number of OLI time code
        errors in the interval */
    uint32_t time_code_errors_tirs;/*number of TIRS time code
        errors in the interval */
    uint32_t detector_map_id_tirs; /* Detector map id for TIRS retrieved from
        the CPF. */
    uint8_t wrs_ending_row;
    /* WRS row number for the last reported scene in this interval */
    uint8_t wrs_scenes; /* total number of full & partial scenes
        in the interval  */
    uint8_t wrs_scenes_full;    /* number of full scenes in the interval */
    uint8_t wrs_scenes_partial; /* number of partial scenes in the interval */
    uint8_t wrs_starting_path;  /* starting WRS2 path for scenes */
    uint8_t wrs_starting_row;   /* WRS2 number row for the first scene */
    uint8_t wrs_type;           /* WRS which applies to the interval, WRS-2 */
}IAS_L0R_INTERVAL_METADATA_V5;

static const IAS_L0R_INTERVAL_METADATA_V5 *ias_l0r_mta_interval_ptr;

static const size_t IAS_L0R_INTERVAL_METADATA_SIZE =
    sizeof(IAS_L0R_INTERVAL_METADATA_V5);
static const size_t IAS_L0R_INTERVAL_METADATA_SIZES
    [TABLE_SIZE_INTERVAL_METADATA] =
{
    sizeof( (*ias_l0r_mta_interval_ptr).ancillary_start_time),
    sizeof( (*ias_l0r_mta_interval_ptr).ancillary_stop_time),
    sizeof( (*ias_l0r_mta_interval_ptr).attitude_points),
    sizeof( (*ias_l0r_mta_interval_ptr).attitude_points_missing),
    sizeof( (*ias_l0r_mta_interval_ptr).attitude_points_rejected),
    sizeof( (*ias_l0r_mta_interval_ptr).collection_type),
    sizeof( (*ias_l0r_mta_interval_ptr).corners_oli.upleft.lat),
    sizeof( (*ias_l0r_mta_interval_ptr).corners_oli.upleft.lng),
    sizeof( (*ias_l0r_mta_interval_ptr).corners_oli.upright.lat),
    sizeof( (*ias_l0r_mta_interval_ptr).corners_oli.upright.lng),
    sizeof( (*ias_l0r_mta_interval_ptr).corners_oli.loleft.lat),
    sizeof( (*ias_l0r_mta_interval_ptr).corners_oli.loleft.lng),
    sizeof( (*ias_l0r_mta_interval_ptr).corners_oli.loright.lat),
    sizeof( (*ias_l0r_mta_interval_ptr).corners_oli.loright.lng),
    sizeof( (*ias_l0r_mta_interval_ptr).corners_tirs.upleft.lat),
    sizeof( (*ias_l0r_mta_interval_ptr).corners_tirs.upleft.lng),
    sizeof( (*ias_l0r_mta_interval_ptr).corners_tirs.upright.lat),
    sizeof( (*ias_l0r_mta_interval_ptr).corners_tirs.upright.lng),
    sizeof( (*ias_l0r_mta_interval_ptr).corners_tirs.loleft.lat),
    sizeof( (*ias_l0r_mta_interval_ptr).corners_tirs.loleft.lng),
    sizeof( (*ias_l0r_mta_interval_ptr).corners_tirs.loright.lat),
    sizeof( (*ias_l0r_mta_interval_ptr).corners_tirs.loright.lng),
    sizeof( (*ias_l0r_mta_interval_ptr).cpf_name),
    sizeof( (*ias_l0r_mta_interval_ptr).crc_errors_oli),
    sizeof( (*ias_l0r_mta_interval_ptr).crc_errors_tirs),
    sizeof( (*ias_l0r_mta_interval_ptr).data_type),
    sizeof( (*ias_l0r_mta_interval_ptr).date_acquired),
    sizeof( (*ias_l0r_mta_interval_ptr).ephemeris_points),
    sizeof( (*ias_l0r_mta_interval_ptr).ephemeris_points_missing),
    sizeof( (*ias_l0r_mta_interval_ptr).ephemeris_points_rejected),
    sizeof( (*ias_l0r_mta_interval_ptr).frames_filled_oli),
    sizeof( (*ias_l0r_mta_interval_ptr).frames_filled_tirs),
    sizeof( (*ias_l0r_mta_interval_ptr).hostname),
    sizeof( (*ias_l0r_mta_interval_ptr).image_quality_oli),
    sizeof( (*ias_l0r_mta_interval_ptr).image_quality_tirs),
    sizeof( (*ias_l0r_mta_interval_ptr).interval_frames_oli),
    sizeof( (*ias_l0r_mta_interval_ptr).interval_frames_tirs),
    sizeof( (*ias_l0r_mta_interval_ptr).interval_number),
    sizeof( (*ias_l0r_mta_interval_ptr).interval_version),
    sizeof( (*ias_l0r_mta_interval_ptr).is_version),
    sizeof( (*ias_l0r_mta_interval_ptr).landsat_cal_interval_id),
    sizeof( (*ias_l0r_mta_interval_ptr).landsat_interval_id),
    sizeof( (*ias_l0r_mta_interval_ptr).nadir_offnadir),
    sizeof( (*ias_l0r_mta_interval_ptr).quality_algorithm),
    sizeof( (*ias_l0r_mta_interval_ptr).roll_angle ),
    sizeof( (*ias_l0r_mta_interval_ptr).satellite ),
    sizeof( (*ias_l0r_mta_interval_ptr).sensor_id ),
    sizeof( (*ias_l0r_mta_interval_ptr).spacecraft_id ),
    sizeof( (*ias_l0r_mta_interval_ptr).start_time_oli),
    sizeof( (*ias_l0r_mta_interval_ptr).start_time_tirs),
    sizeof( (*ias_l0r_mta_interval_ptr).station_id ),
    sizeof( (*ias_l0r_mta_interval_ptr).stop_time_oli),
    sizeof( (*ias_l0r_mta_interval_ptr).stop_time_tirs),
    sizeof( (*ias_l0r_mta_interval_ptr).time_code_errors_oli),
    sizeof( (*ias_l0r_mta_interval_ptr).time_code_errors_tirs),
    sizeof( (*ias_l0r_mta_interval_ptr).detector_map_id_tirs),
    sizeof( (*ias_l0r_mta_interval_ptr).wrs_ending_row),
    sizeof( (*ias_l0r_mta_interval_ptr).wrs_scenes),
    sizeof( (*ias_l0r_mta_interval_ptr).wrs_scenes_full),
    sizeof( (*ias_l0r_mta_interval_ptr).wrs_scenes_partial),
    sizeof( (*ias_l0r_mta_interval_ptr).wrs_starting_path),
    sizeof( (*ias_l0r_mta_interval_ptr).wrs_starting_row),
    sizeof( (*ias_l0r_mta_interval_ptr).wrs_type)
};

static const size_t OFFSET_INTERVAL_METADATA[TABLE_SIZE_INTERVAL_METADATA] =
{
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,ancillary_start_time),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,ancillary_stop_time),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,attitude_points),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,attitude_points_missing),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,attitude_points_rejected),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,collection_type),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,corners_oli.upleft.lat),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,corners_oli.upleft.lng),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,corners_oli.upright.lat),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,corners_oli.upright.lng),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,corners_oli.loleft.lat),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,corners_oli.loleft.lng),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,corners_oli.loright.lat),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,corners_oli.loright.lng),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,corners_tirs.upleft.lat),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,corners_tirs.upleft.lng),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,corners_tirs.upright.lat),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,corners_tirs.upright.lng),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,corners_tirs.loleft.lat),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,corners_tirs.loleft.lng),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,corners_tirs.loright.lat),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,corners_tirs.loright.lng),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,cpf_name),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,crc_errors_oli),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,crc_errors_tirs),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,data_type),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,date_acquired),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,ephemeris_points),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,ephemeris_points_missing),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,ephemeris_points_rejected),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,frames_filled_oli),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,frames_filled_tirs),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,hostname),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,image_quality_oli),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,image_quality_tirs),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,interval_frames_oli),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,interval_frames_tirs),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,interval_number),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,interval_version),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,is_version),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,landsat_cal_interval_id),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,landsat_interval_id),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,nadir_offnadir),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,quality_algorithm),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,roll_angle),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,satellite),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,sensor_id),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,spacecraft_id),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,start_time_oli),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,start_time_tirs),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,station_id),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,stop_time_oli),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,stop_time_tirs),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,time_code_errors_oli),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,time_code_errors_tirs),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,detector_map_id_tirs),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,wrs_ending_row),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,wrs_scenes),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,wrs_scenes_full),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,wrs_scenes_partial),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,wrs_starting_path),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,wrs_starting_row),
   HOFFSET( IAS_L0R_INTERVAL_METADATA_V5,wrs_type)
};


/******************************************************************************
 NAME: ias_l0r_get_interval_metadata_v5

 PURPOSE: Read the interval level metadata from the file provided

 RETURNS: SUCCESS- Metadata retrieved
          ERROR- Metadata could not be retrieved
******************************************************************************/
int ias_l0r_copy_interval_metadata_v5
(
    IAS_L0R_INTERVAL_METADATA_V5 *old_metadata, /*I: Metadata to copy */
    IAS_L0R_INTERVAL_METADATA *new_metadata /*O: Metadata copied */
)
{
    int i = 0;

    /*Clear the structure to initialize*/
    memset(new_metadata, 0, sizeof(IAS_L0R_INTERVAL_METADATA));

    memcpy(new_metadata->ancillary_start_time,
        old_metadata->ancillary_start_time,
        sizeof(new_metadata->ancillary_start_time));
    new_metadata->ancillary_start_time[IAS_L0R_DATE_LENGTH] = '\0';
    memcpy(new_metadata->ancillary_stop_time, old_metadata->ancillary_stop_time,
        sizeof(new_metadata->ancillary_stop_time));
    new_metadata->ancillary_stop_time[IAS_L0R_DATE_LENGTH] = '\0';
    new_metadata->attitude_points = old_metadata->attitude_points;
    new_metadata->attitude_points_missing = 
        old_metadata->attitude_points_missing;
    new_metadata->attitude_points_rejected = 
        old_metadata->attitude_points_rejected;
    memcpy(new_metadata->collection_type, old_metadata->collection_type,
           sizeof(new_metadata->collection_type));

    /* Replace spaces with underscores for Epoch 3*/
    for (i=0; i < IAS_L0R_COLLECTION_TYPE_LENGTH-1; i++)
        if(new_metadata->collection_type[i] == ' ')
            new_metadata->collection_type[i] = '_';
    new_metadata->collection_type[IAS_L0R_COLLECTION_TYPE_LENGTH] = '\0';

    COPY_CORNERS(new_metadata->corners_oli, old_metadata->corners_oli);
    COPY_CORNERS(new_metadata->corners_tirs, old_metadata->corners_tirs);

    memcpy(new_metadata->cpf_name, old_metadata->cpf_name, 
        sizeof(new_metadata->cpf_name));

    strncpy(new_metadata->cpf_name, old_metadata->cpf_name, 
        IAS_L0R_CPF_NAME_LENGTH_V5);

    new_metadata->crc_errors_oli = old_metadata->crc_errors_oli;
    new_metadata->crc_errors_tirs = old_metadata->crc_errors_tirs;
    memcpy(new_metadata->data_type, old_metadata->data_type, 
        sizeof(new_metadata->data_type));
    new_metadata->data_type[IAS_L0R_DATA_TYPE_LENGTH] = '\0';
    memcpy(new_metadata->date_acquired, old_metadata->date_acquired,
        sizeof(new_metadata->date_acquired));
    new_metadata->date_acquired[IAS_L0R_DATE_LENGTH] = '\0';
    new_metadata->ephemeris_points = old_metadata->ephemeris_points;
    new_metadata->ephemeris_points_missing = 
        old_metadata->ephemeris_points_missing;
    new_metadata->ephemeris_points_rejected = 
        old_metadata->ephemeris_points_rejected;
    memcpy(new_metadata->hostname, old_metadata->hostname, 
        sizeof(new_metadata->hostname));
    new_metadata->hostname[IAS_L0R_HOSTNAME_LENGTH] = '\0';
    new_metadata->image_quality_oli = old_metadata->image_quality_oli;
    new_metadata->image_quality_tirs = old_metadata->image_quality_tirs;
    new_metadata->interval_frames_oli = old_metadata->interval_frames_oli;
    new_metadata->interval_frames_tirs = old_metadata->interval_frames_tirs;
    new_metadata->interval_number = old_metadata->interval_number;
    new_metadata->interval_version = old_metadata->interval_version;
    memcpy(new_metadata->is_version, old_metadata->is_version, 
        sizeof(new_metadata->is_version));
    new_metadata->is_version[IAS_SOFTWARE_VERSION_SIZE - 1] = '\0';

    memcpy(new_metadata->landsat_cal_interval_id, 
        old_metadata->landsat_cal_interval_id,
        sizeof(new_metadata->landsat_cal_interval_id));
    new_metadata->landsat_cal_interval_id[IAS_L0R_INTERVAL_ID_LENGTH] = '\0';
    memcpy(new_metadata->landsat_interval_id, old_metadata->landsat_interval_id,
        sizeof(new_metadata->landsat_interval_id));
    new_metadata->landsat_interval_id[IAS_L0R_INTERVAL_ID_LENGTH] = '\0';
    new_metadata->frames_filled_oli = old_metadata->frames_filled_oli;
    new_metadata->frames_filled_tirs = old_metadata->frames_filled_tirs;
    memcpy(new_metadata->nadir_offnadir, old_metadata->nadir_offnadir, 
        sizeof(new_metadata->nadir_offnadir));
    new_metadata->nadir_offnadir[IAS_L0R_NADIR_LENGTH] = '\0';
    memcpy(new_metadata->quality_algorithm, old_metadata->quality_algorithm,
           sizeof(new_metadata->quality_algorithm));
    new_metadata->quality_algorithm[IAS_L0R_ALGORITHM_NAME_LENGTH] = '\0';
    new_metadata->roll_angle = old_metadata->roll_angle;
    new_metadata->satellite = old_metadata->satellite;
    memcpy(new_metadata->sensor_id, old_metadata->sensor_id,
        sizeof(new_metadata->sensor_id));
    new_metadata->sensor_id[IAS_L0R_SENSOR_ID_LENGTH] = '\0';
    memcpy(new_metadata->spacecraft_id, old_metadata->spacecraft_id,
        sizeof(new_metadata->spacecraft_id));
    new_metadata->spacecraft_id[IAS_L0R_SPACECRAFT_ID_LENGTH] = '\0';
    memcpy(new_metadata->start_time_oli, old_metadata->start_time_oli,
        sizeof(new_metadata->start_time_oli));
    new_metadata->start_time_oli[IAS_L0R_DATE_LENGTH] = '\0';
    memcpy(new_metadata->start_time_tirs, old_metadata->start_time_tirs,
        sizeof(new_metadata->start_time_tirs));
    memcpy(new_metadata->start_time_tirs, old_metadata->start_time_tirs,
        sizeof(new_metadata->start_time_tirs));
    new_metadata->start_time_tirs[IAS_L0R_DATE_LENGTH] = '\0';
    memcpy(new_metadata->station_id, old_metadata->station_id, 
        sizeof(new_metadata->station_id));
    new_metadata->station_id[IAS_L0R_STATION_ID_LENGTH] = '\0';
    memcpy(new_metadata->stop_time_oli, old_metadata->stop_time_oli, 
        sizeof(new_metadata->stop_time_oli));
    new_metadata->stop_time_oli[IAS_L0R_DATE_LENGTH] = '\0';
    memcpy(new_metadata->stop_time_tirs, old_metadata->stop_time_tirs,
        sizeof(new_metadata->stop_time_tirs));
    new_metadata->stop_time_tirs[IAS_L0R_DATE_LENGTH] = '\0';
    new_metadata->time_code_errors_oli = old_metadata->time_code_errors_oli;
    new_metadata->time_code_errors_tirs = old_metadata->time_code_errors_tirs;
    new_metadata->detector_map_id_tirs = old_metadata->detector_map_id_tirs;
    new_metadata->wrs_ending_row = old_metadata->wrs_ending_row;
    new_metadata->wrs_scenes = old_metadata->wrs_scenes;
    new_metadata->wrs_scenes_full = old_metadata->wrs_scenes_full;
    new_metadata->wrs_scenes_partial = old_metadata->wrs_scenes_partial;
    new_metadata->wrs_starting_path = old_metadata->wrs_starting_path;
    new_metadata->wrs_starting_row = old_metadata->wrs_starting_row;
    new_metadata->wrs_type = old_metadata->wrs_type;

    return SUCCESS;
}

/******************************************************************************
 NAME: ias_l0r_get_interval_metadata_v5

 PURPOSE: Read the interval level metadata from the file provided

 RETURNS: SUCCESS- Metadata retrieved
          ERROR- Metadata could not be retrieved
******************************************************************************/
int ias_l0r_get_interval_metadata_v5
(
    L0RIO *l0r, /* I: structure used with the l0r data */
    IAS_L0R_INTERVAL_METADATA *mta_info_interval /* O: Metadata read */
)
{
    int status = 0;

    IAS_L0R_INTERVAL_METADATA_V5 interval_v5;

    if (l0r == NULL)
    {
        IAS_LOG_ERROR("Error NULL pointer received");
        return ERROR;
    }

    HDFIO *hdfio_ptr =  l0r;

    status = ias_l0r_mta_read(hdfio_ptr, 0, 1,
        TABLE_NAME_INTERVAL_METADATA,
        IAS_L0R_INTERVAL_METADATA_SIZE, OFFSET_INTERVAL_METADATA,
        IAS_L0R_INTERVAL_METADATA_SIZES, &interval_v5);

    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Read of interval metadata has failed.");
        return ERROR;
    }

    /*Copy the interval metadata to the new format*/
    return ias_l0r_copy_interval_metadata_v5(&interval_v5, mta_info_interval);
}

