#include <string.h>
#include <stdio.h>
#include <time.h>
#include <sys/stat.h>

#include "ias_logging.h"
#include "ias_l0r.h"
#include "ias_satellite_attributes.h"
#include "cgreen/cgreen.h"

/**
*test_image_create creates an image line for each band file writes i
* and then reads it
*
*
*/
static void test_image_create(void)
{
    int band_number = 0;
    int i = 0;
    int j = 0;
    int status;

    uint16_t *image = NULL;
    uint16_t *image_from_file = NULL;

    L0RIO *l0r_data = NULL;
    const IAS_BAND_ATTRIBUTES *band_attributes;
    const IAS_SATELLITE_ATTRIBUTES *landsat8_attributes;

    srand(time(NULL));

    /* open access to L0R data */
    l0r_data = ias_l0r_open("test", "./", IAS_L0R_COMPRESSION_ON);
    assert_true(l0r_data != NULL);

    /* No files should exist for the data yet */
    if(ias_l0r_present("./test") == TRUE)
    {
        assert_true(system("rm ./test*.h5") == 0);
    }

    ias_sat_attr_initialize(IAS_L8);
    landsat8_attributes = ias_sat_attr_get_attributes();
    assert_true(landsat8_attributes != NULL);

    /* for all of the bands */
    for (band_number = 1; band_number <= landsat8_attributes->total_bands;
        band_number++)
    {
        int image_size;
        status = ias_l0r_open_band(l0r_data, band_number, IAS_WRITE);
        assert_true(status == SUCCESS);
        band_attributes = ias_sat_attr_get_band_attributes(band_number);
        assert_true(band_attributes != NULL);
        if (band_attributes == NULL)
        {
            IAS_LOG_ERROR("Unable to get band attributes for band %d",
                          band_number);
        }

        image_size = band_attributes->detectors_per_sca * sizeof(uint16_t);
        image = malloc(image_size);
        assert_true(image != NULL);

        image_from_file = malloc(band_attributes->detectors_per_sca *
            sizeof(uint16_t));
        assert_true(image != NULL);
        
        /* Create random data and test file write and read on a per sca basis*/
        for(i = 0; i < band_attributes->scas; i++)
        {
            for(j = 0; j < band_attributes->detectors_per_sca; j++)
            {
                image[j] = rand()%UINT16_MAX;
            }
            status = ias_l0r_set_band_lines_sca(l0r_data, band_number, i, 0, 0,
                1, band_attributes->detectors_per_sca, image);
            assert_true(status == SUCCESS);

            status = ias_l0r_close_band(l0r_data, band_number);
            assert_true(status == SUCCESS);

            status = ias_l0r_open_band(l0r_data, band_number, IAS_WRITE);
            assert_true(status == SUCCESS);
 
            status = ias_l0r_get_band_lines_sca(l0r_data,
                band_number, i, 0, 0, 1,
                band_attributes->detectors_per_sca,
                image_from_file);
            assert_true(status == SUCCESS);

            assert_true(memcmp(image, image_from_file, image_size) == 0);
        }

        /* We should have band files created at this point */
        assert_true(ias_l0r_present("./test") == TRUE);

        status = ias_l0r_close_band(l0r_data, band_number);
        assert_true(status == SUCCESS);
        /* Cleanup the resources used for this band */
        free(image);
        image = NULL;
        free(image_from_file);
        image_from_file = NULL;
    }
    
    /*close access to the L0R data */
    status = ias_l0r_close(l0r_data);
    assert_true(status == SUCCESS);
}

/**
*test_image_create_random_order
*
*
*/
static void test_image_create_random_order(void)
{
    const int number_of_lines = 4;
    int band_number = 0;
    int error_count = 0;
    int i = 0;
    int j = 0;
    int k = 0;
    int line = 0;
    int line_count = 1;
    int status;

    uint16_t *image = NULL;
    uint16_t *image_from_file = NULL;
    uint16_t *top_offsets = NULL;
    uint16_t *bottom_offsets = NULL;
    uint16_t *top_offsets_from_file = NULL;
    uint16_t *bottom_offsets_from_file = NULL;
    uint16_t *top_offsets_from_file_sca = NULL;
    uint16_t *bottom_offsets_from_file_sca = NULL;

    L0RIO *l0r_data = NULL;
    const IAS_BAND_ATTRIBUTES *band_attributes;
    IAS_L0R_BAND_ATTRIBUTES image_attributes;
    const IAS_SATELLITE_ATTRIBUTES *landsat8_attributes;

    srand(time(NULL));

    /* open access to L0R data */
    l0r_data = ias_l0r_open("test", "./", IAS_L0R_COMPRESSION_ON);
    assert_true(l0r_data != NULL);

    ias_sat_attr_initialize(IAS_L8);
    landsat8_attributes = ias_sat_attr_get_attributes();

    /* open all the bands */
    for (band_number = 1;
         band_number <= landsat8_attributes->total_bands;
         band_number++)
    {
        band_attributes = ias_sat_attr_get_band_attributes(band_number);
        status = ias_l0r_open_band(l0r_data, band_number, IAS_WRITE);
        assert_true(status == SUCCESS);
    }

    for (k = 0; k < 99; k++)
    {
        int image_size;
        int sca_size;

        do
        {
            band_number = ((rand()%landsat8_attributes->total_bands)+1);
        }while(!(band_number >= 1 && band_number <=
            landsat8_attributes->total_bands));

        band_attributes = ias_sat_attr_get_band_attributes(band_number);
 
        image_size = band_attributes->detectors_per_sca * sizeof(uint16_t);
        image = malloc(image_size);
        assert_true(image != NULL);

        image_from_file = malloc(band_attributes->detectors_per_sca *
            sizeof(uint16_t));
        assert_true(image_from_file != NULL);

        line = rand()%number_of_lines;

        /* Create random data and test file write and read on a per sca basis*/
        for(i = 0; i < band_attributes->scas; i++)
        {
            for(j = 0; j < band_attributes->detectors_per_sca; j++)
            {
                image[j] = rand()%UINT16_MAX;
            }
            
            status = ias_l0r_set_band_lines_sca(l0r_data, band_number, i, line,
                 0, line_count, band_attributes->detectors_per_sca, image);
            assert_true(status == SUCCESS);
             
            /* Occasionally flush the file by close/open and check contents */
            if (rand()%2 == 0)
            {
                error_count = 0;

                status = ias_l0r_close_band(l0r_data, band_number);
                if(status != SUCCESS)
                    error_count++;
                status = ias_l0r_open_band(l0r_data, band_number, IAS_WRITE);
                if(status != SUCCESS)
                    error_count++;

                status = ias_l0r_get_band_lines_sca(l0r_data,
                   band_number, i, line, 0,
                   line_count,
                   band_attributes->detectors_per_sca,
                   image_from_file);
                if(status != SUCCESS)
                    error_count++;

                if(memcmp(image,image_from_file, image_size) != 0)
                    error_count++;
                assert_true(error_count == 0);
            }
 
        }

        /* Cleanup the resources used with sca read/write */
        free(image);
        image = NULL;
        free(image_from_file);
        image_from_file = NULL;

        image_size = band_attributes->detectors_per_sca
            * band_attributes->scas * sizeof(uint16_t);
        image = malloc(image_size);
        assert_true(image != NULL);

        top_offsets = malloc(image_size);
        assert_true(top_offsets != NULL);

        bottom_offsets = malloc(image_size);
        assert_true(bottom_offsets != NULL);
 
        image_from_file = malloc(image_size);
        assert_true(image != NULL);
 
        top_offsets_from_file = malloc(image_size);
        assert_true(top_offsets_from_file != NULL);

        bottom_offsets_from_file = malloc(image_size);
        assert_true(bottom_offsets_from_file != NULL);

        sca_size = band_attributes->detectors_per_sca * sizeof(uint16_t);
        top_offsets_from_file_sca = malloc(sca_size);
        assert_true(top_offsets_from_file_sca != NULL);

        bottom_offsets_from_file_sca = malloc(sca_size);
        assert_true(bottom_offsets_from_file_sca != NULL);
   
        /* Fill the line data */
        for(j = 0; j < (band_attributes->detectors_per_sca *
            band_attributes->scas); j++)
        {
            image[j] = rand()%UINT16_MAX;
            top_offsets[j] = rand()%UINT16_MAX;
            bottom_offsets[j] = rand()%UINT16_MAX;
        }

        status = ias_l0r_set_band_lines(l0r_data, band_number, line,
            line_count, image);
        assert_true(status == SUCCESS);

        /* Don't do offset for VRPs bands, they don't have any */
        if (ias_sat_attr_band_classification_matches(band_number,
            IAS_NORMAL_BAND))
        {
            status = ias_l0r_set_top_detector_offsets(l0r_data, band_number,
                top_offsets);
            assert_true(status == SUCCESS);

            status = ias_l0r_set_bottom_detector_offsets(l0r_data, band_number,
                bottom_offsets);
            assert_true(status == SUCCESS);
        }

        /* Occasionally flush the file by close/open and check contents */
        if (k % 8 == 0)
        {
            status = ias_l0r_close_band(l0r_data, band_number);
            assert_true(status == SUCCESS);

            status = ias_l0r_open_band(l0r_data, band_number, IAS_WRITE);
            assert_true(status == SUCCESS);
        
            status = ias_l0r_get_band_lines(l0r_data, band_number, line,
                      line_count, band_attributes->detectors_per_sca *
            band_attributes->scas, image_from_file);
            assert_true(status == SUCCESS);
            assert_true(memcmp(image, image_from_file, image_size) == 0);

            /* Don't do offset for VRPs bands, they don't have any */
            if (ias_sat_attr_band_classification_matches(band_number,
                IAS_NORMAL_BAND))
            {
                status = ias_l0r_get_top_detector_offsets(l0r_data,
                    band_number, top_offsets_from_file,
                    (band_attributes->scas *
                    band_attributes->detectors_per_sca));
                assert_true(status == SUCCESS);
                assert_true(memcmp(top_offsets, top_offsets_from_file,
                    image_size) == 0);

                status = ias_l0r_get_bottom_detector_offsets(l0r_data,
                     band_number, bottom_offsets_from_file,
                     (band_attributes->scas *
                     band_attributes->detectors_per_sca));
                assert_true(status == SUCCESS);
                assert_true(memcmp(bottom_offsets, bottom_offsets_from_file,
                    image_size) == 0);
            }
       
            /* Don't do offset for VRPs bands, they don't have any */
            int s;
            for(s = 0; s < band_attributes->scas &&
                ias_sat_attr_band_classification_matches(band_number,
                    IAS_NORMAL_BAND); s++)
            {
                error_count = 0;

                status = ias_l0r_get_top_detector_offsets_sca(l0r_data,
                        band_number, s, top_offsets_from_file_sca,
                        (band_attributes->detectors_per_sca));
                if(status != SUCCESS)
                    error_count++;
                status = ias_l0r_get_bottom_detector_offsets_sca(l0r_data,
                        band_number, s, bottom_offsets_from_file_sca,
                        (band_attributes->detectors_per_sca));
                if(status != SUCCESS)
                    error_count++;

                if(memcmp(&top_offsets[s * band_attributes->detectors_per_sca],
                   top_offsets_from_file_sca, sca_size) != 0)
                        error_count++;
                if(memcmp(&bottom_offsets[
                    (s * band_attributes->detectors_per_sca)],
                    bottom_offsets_from_file_sca, sca_size) != 0)
                        error_count++;
                assert_true(error_count == 0);
            }
 
            status = ias_l0r_get_band_attributes(l0r_data, band_number,
                &image_attributes);
            assert_true(status == SUCCESS);

            assert_true(image_attributes.sca_count == band_attributes->scas);
            assert_true(image_attributes.detectors_per_sca ==
                band_attributes->detectors_per_sca);
        }

        /* Cleanup the resources used with line read/write */
        free(image);
        image = NULL;
        free(image_from_file);
        image_from_file = NULL;
        free(top_offsets);
        top_offsets = NULL;
        free(bottom_offsets);
        bottom_offsets = NULL;
        free(top_offsets_from_file);
        top_offsets_from_file = NULL;
        free(bottom_offsets_from_file);
        bottom_offsets_from_file = NULL;
        free(top_offsets_from_file_sca);
        top_offsets_from_file_sca = NULL;
        free(bottom_offsets_from_file_sca);
        bottom_offsets_from_file_sca = NULL;
    }

    /*close access to the L0R data */
    status = ias_l0r_close(l0r_data);
    assert_true(status == SUCCESS);
}

/* test the get_band_present routine.  This requires that the create test
   was run before this one. */
static void test_get_band_present()
{
    int status;
    L0RIO *l0r_file;
    int band_present;
    int count;

    /* open L0R file */
    l0r_file = ias_l0r_open("test", "./", IAS_L0R_COMPRESSION_DEFAULT);
    assert_true(l0r_file != NULL);

    status = ias_l0r_open_header(l0r_file, IAS_READ);
    assert_equal(status, SUCCESS);

    status = ias_l0r_is_oli_image_header_present(l0r_file, &count);
    assert_equal(status, SUCCESS);

    status = ias_l0r_get_band_present(l0r_file, 17, &band_present);
    assert_equal(status, SUCCESS);
    assert_equal(band_present, 1);

    status = ias_l0r_get_band_present(l0r_file, 20, &band_present);
    assert_equal(status, SUCCESS);
    assert_equal(band_present, 1);

    status = ias_l0r_close_header(l0r_file);
    assert_equal(status, SUCCESS);

    status = ias_l0r_close(l0r_file);
    assert_true(status == SUCCESS);
}

/**
*test_image_truncate
*
*
*/
static void test_image_truncate(void)
{
    int band_number = 0;
    int size = 0;
    int size_2 = 0;
    int status;

    uint16_t *image = NULL;

    L0RIO *l0r_data = NULL;
    const IAS_BAND_ATTRIBUTES *band_attributes;
    const IAS_SATELLITE_ATTRIBUTES *landsat8_attributes;

    srand(time(NULL));

    /* open access to L0R data */
    l0r_data = ias_l0r_open("test", "./", IAS_L0R_COMPRESSION_ON);
    assert_true(l0r_data != NULL);

    ias_sat_attr_initialize(IAS_L8);
    landsat8_attributes = ias_sat_attr_get_attributes();
    assert_true(landsat8_attributes != NULL);

    /* for all of the bands */
    for (band_number = 1; band_number <= landsat8_attributes->total_bands;
        band_number++)
    {
        status = ias_l0r_open_band(l0r_data, band_number, IAS_WRITE);
        assert_true(status == SUCCESS);
        band_attributes = ias_sat_attr_get_band_attributes(band_number);
        assert_true(band_attributes != NULL);
        if (band_attributes == NULL)
        {
            IAS_LOG_ERROR("Unable to get band attributes for band %d",
                          band_number);
        }
        
        status = ias_l0r_get_band_records_count(l0r_data, band_number, &size);
        assert_true(status == SUCCESS);
        assert_true(size >= 1);
        
        image = calloc(size, band_attributes->detectors_per_sca *
            band_attributes->scas * sizeof(uint16_t));
        assert_true(image != NULL);
        
        /* Store the image data to memory, so it can be written back to file */
        status = ias_l0r_get_band_lines(l0r_data, band_number, 0, size,
        	band_attributes->detectors_per_sca * band_attributes->scas,
        	image);
    
   		status = ias_l0r_truncate_band_lines(l0r_data, band_number);
   		assert_true(status == SUCCESS);
   		
   		/* Close and re-open the band to ensure data is flushed to file */
        status = ias_l0r_close_band(l0r_data, band_number);
        assert_true(status == SUCCESS);
        status = ias_l0r_open_band(l0r_data, band_number, IAS_WRITE);
        assert_true(status == SUCCESS);
   		
   		/* Verify the lines were truncated to 1 line (not 0, HDF doesn't
            allow setting extent to 0 ) */
   		status = ias_l0r_get_band_records_count(l0r_data, band_number, &size_2);
   		assert_true(status == SUCCESS);
   		assert_true(size_2 <= 1);
   		
   		 /* Write stored data back to file */ 

        status = ias_l0r_set_band_lines(l0r_data, band_number, 0, size,
        	image);
        
        status = ias_l0r_close_band(l0r_data, band_number);
        assert_true(status == SUCCESS);
        /* Cleanup the resources used for this band */
        free(image);
        image = NULL;
    }
    
    /*close access to the L0R data */
    status = ias_l0r_close(l0r_data);
    assert_true(status == SUCCESS);
}
	
int main(int argc, char **argv)
{
    int status = 0;

    status = ias_log_initialize("L0R Band Test");
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Error encountered when initializing the ias logging "
            "library");
        return ERROR;
    }
    if (ias_sat_attr_initialize(IAS_L8) != SUCCESS)
    {
        IAS_LOG_ERROR("Error initializing attributes");
        return ERROR;
    }
    ias_log_set_output_level(IAS_LOG_LEVEL_ERROR);

    TestSuite *suite = create_test_suite();
    TestReporter *reporter = create_text_reporter();

    if (suite == NULL || reporter == NULL)
    {
        return ERROR;
    }

    srand(time(NULL));

    ias_sat_attr_initialize(IAS_L8);

    /* Add tests for creating image data */
    add_test(suite, test_image_create);
    add_test(suite, test_image_create_random_order);
    add_test(suite, test_get_band_present);
    add_test(suite, test_image_truncate);

    return run_test_suite(suite, reporter);
}
