#include <inttypes.h>
#include <string.h>
#include <stdio.h>
#include <sys/stat.h>
#include <time.h>

#include "ias_logging.h"
#include "ias_l0r.h"
#include "ias_satellite_attributes.h"
#include "cgreen/cgreen.h"

void print_metadata_file_structure(const IAS_L0R_FILE_METADATA *metadata)
{
    int i = 0; /* loop iterator */

    printf("#################################\n");

    printf("name: %s\n", metadata->ancillary_file.name);

    for (i = 0; i < IAS_L0R_MAX_BAND_FILES; i++)
    {
        printf("name: %s\n", metadata->band_files[i].name);
    }

    printf("name: %s\n", metadata->checksum_file.name);

    printf("name: %s\n", metadata->metadata_file.name);

    printf("interval_files %d\n", metadata->interval_files);


    printf("#################################\n");
}

static void test_metadata_create_file_level(void)
{
    int count = 0;
    int i = 0;
    int j = 0;
    int status;
    int different_elements_count;
    unsigned int filenames_generated = 0;
    L0RIO *l0r_data = NULL;
    IAS_L0R_FILE_METADATA file_metadata;
    IAS_L0R_FILE_METADATA file_metadata_from_file;

    srand(time(NULL));

    memset(&file_metadata, 0, sizeof(IAS_L0R_FILE_METADATA));
    memset(&file_metadata_from_file, 0, sizeof(IAS_L0R_FILE_METADATA));

    for (i = 0; i < IAS_L0R_MAX_BAND_FILES; i++)
    {
        /*randomly fill the name*/
        for (j = 0; j < IAS_L0R_FILE_NAME_LENGTH; j++)
        {
            file_metadata.band_files[i].name[j] = 'A' + rand()%52;
        }
        /* if there was a filename generated */
        if (strlen(file_metadata.band_files[i].name) > 0)
        {
            filenames_generated++;
        }
    }

    /*randomly fill the name for the ancillary file*/
    for (j = 0; j < IAS_L0R_FILE_NAME_LENGTH; j++)
    {
        file_metadata.ancillary_file.name[j] = 'A' + rand()%52;
    }
    /* if there was a filename generated */
    if (strlen(file_metadata.ancillary_file.name) > 0)
    {
        filenames_generated++;
    }

    /* randomly fill metadata filename */
    for (j = 0; j < IAS_L0R_FILE_NAME_LENGTH; j++)
    {
        file_metadata.metadata_file.name[j] = 'A' + rand()%52;
    }
    /* if there was a filename generated */
    if (strlen(file_metadata.metadata_file.name) > 0)
    {
        filenames_generated++;
    }

    /*randomly fill the name for the checksum file*/
    for (j = 0; j < IAS_L0R_FILE_NAME_LENGTH; j++)
    {
        file_metadata.checksum_file.name[j] = 'A' + rand()%52;
    }
    /* if there was a filename generated */
    if (strlen(file_metadata.checksum_file.name) > 0)
    {
        filenames_generated++;
    }

    /* open access to L0R data */
    l0r_data = ias_l0r_open("test", "./", IAS_L0R_COMPRESSION_ON);
    assert_true(l0r_data != NULL);

    ias_l0r_open_metadata(l0r_data, IAS_WRITE);

    /* No files should exist for the data yet */
    if (ias_l0r_present("./test") == TRUE)
    {
        assert_true(system("rm ./test*.h5") == 0);
    }
 
    status = ias_l0r_set_file_metadata(l0r_data, &file_metadata);
    assert_true(status == SUCCESS);
 
    status = ias_l0r_get_file_metadata(l0r_data, &file_metadata_from_file);
    assert_true(status == SUCCESS);
 
    /* Verify that the values read from file match the values written */
    status = ias_l0r_mta_compare_file(&file_metadata,
        &file_metadata_from_file, &different_elements_count);
    assert_true(status == SUCCESS);
    assert_true(different_elements_count == 0);
    if (different_elements_count != 0)
    {
        print_metadata_file_structure(&file_metadata);
        print_metadata_file_structure(&file_metadata_from_file);
    }

    status = ias_l0r_get_filename_count(l0r_data, &count);
    assert_true(status == SUCCESS);
    /* Assert there are file names for band files + ancillary + checksum */
    if (count != filenames_generated)
    {
        assert_true_with_message(count == filenames_generated,
            "The number of files returned by the lib is %d for %d files",
            count, filenames_generated);
    }

    /* We should have an MTA file created at this point */
    assert_true(ias_l0r_present("./test") == TRUE);

    status = ias_l0r_close_metadata(l0r_data);
    assert_true(status == SUCCESS);

    /*close access to the L0R data */
    status = ias_l0r_close(l0r_data);
    assert_true(status == SUCCESS);
}

static void test_metadata_create_interval_level(void)
{
    int i = 0;
    int status;
    int different_elements_count;
    L0RIO *l0r_data = NULL;
    IAS_L0R_INTERVAL_METADATA interval_metadata;
    IAS_L0R_INTERVAL_METADATA interval_metadata_from_file;

    srand(time(NULL));

    memset(&interval_metadata, 0, sizeof(IAS_L0R_INTERVAL_METADATA));
    memset(&interval_metadata_from_file, 0, sizeof(IAS_L0R_INTERVAL_METADATA));

    /* Generate random interval metadata values for the unit test */
    for (i =0 ; i<IAS_L0R_INTERVAL_ID_LENGTH; i++)
    {
        interval_metadata.landsat_interval_id[i] = 'A' + rand()%52;
    }
    for (i =0 ; i<IAS_L0R_SPACECRAFT_ID_LENGTH; i++)
    {
        interval_metadata.spacecraft_id[i] = 'A' + rand()%52;
    }
    for (i =0 ; i<IAS_L0R_SENSOR_ID_LENGTH; i++)
    {
        interval_metadata.sensor_id[i] = 'A' + rand()%52;
    }
    interval_metadata.satellite = rand()%UINT8_MAX;
    for (i =0 ; i<IAS_L0R_STATION_ID_LENGTH; i++)
    {
        interval_metadata.station_id[i] = 'A' + rand()%52;
    }
    for (i =0 ; i<IAS_L0R_DATA_TYPE_LENGTH; i++)
    {
        interval_metadata.data_type[i] = 'A' + rand()%52;
    }
    for (i =0 ; i<IAS_L0R_COLLECTION_TYPE_LENGTH; i++)
    {
        interval_metadata.collection_type[i] = 'A' + rand()%52;
    }
    interval_metadata.wrs_starting_path = rand()%UINT8_MAX;
    interval_metadata.wrs_starting_row = rand()%UINT8_MAX;
    interval_metadata.wrs_ending_row = rand()%UINT8_MAX;
    for (i =0 ; i<IAS_L0R_DATE_LENGTH; i++)
    {
        interval_metadata.ancillary_start_time[i] = 'A' + rand()%52;
    }
    for (i =0 ; i<IAS_L0R_DATE_LENGTH; i++)
    {
        interval_metadata.ancillary_stop_time[i] = 'A' + rand()%52;
    }
    for (i =0 ; i<IAS_L0R_DATE_LENGTH; i++)
    {
        interval_metadata.start_time_oli[i] = 'A' + rand()%52;
    }
    for (i =0 ; i<IAS_L0R_DATE_LENGTH; i++)
    {
        interval_metadata.start_time_tirs[i] = 'A' + rand()%52;
    }
    for (i =0 ; i<IAS_L0R_DATE_LENGTH; i++)
    {
        interval_metadata.stop_time_oli[i] = 'A' + rand()%52;
    }
    for (i =0 ; i<IAS_L0R_DATE_LENGTH; i++)
    {
        interval_metadata.stop_time_tirs[i] = 'A' + rand()%52;
    }
    for (i =0 ; i<IAS_L0R_CPF_NAME_LENGTH; i++)
    {
        interval_metadata.cpf_name[i] = 'A' + rand()%52;
    }
    interval_metadata.crc_errors_oli = rand()%UINT32_MAX;
    interval_metadata.crc_errors_tirs = rand()%UINT32_MAX;
    interval_metadata.image_quality_oli = rand()%UINT8_MAX;
    interval_metadata.image_quality_tirs = rand()%UINT8_MAX;
    interval_metadata.interval_frames_oli = rand()%UINT32_MAX;
    interval_metadata.interval_frames_tirs = rand()%UINT32_MAX;
    interval_metadata.corners_oli.upleft.lat = (rand() % UINT8_MAX) + 0.1 /
         ((rand() % UINT16_MAX) + 1.0);
    interval_metadata.corners_oli.upleft.lng = (rand() % UINT8_MAX) + 0.1 /
         ((rand() % UINT16_MAX) + 1.0);
    interval_metadata.corners_oli.upright.lat = (rand() % UINT8_MAX) + 0.1 /
         ((rand() % UINT16_MAX) + 1.0);
    interval_metadata.corners_oli.upright.lng = (rand() % UINT8_MAX) + 0.1 /
         ((rand() % UINT16_MAX) + 1.0);
    interval_metadata.corners_oli.loleft.lat = (rand() % UINT8_MAX) + 0.1 /
         ((rand() % UINT16_MAX) + 1.0);
    interval_metadata.corners_oli.loleft.lng = (rand() % UINT8_MAX) + 0.1 /
         ((rand() % UINT16_MAX) + 1.0);
    interval_metadata.corners_oli.loright.lat = (rand() % UINT8_MAX) + 0.1 /
         ((rand() % UINT16_MAX) + 1.0);
    interval_metadata.corners_oli.loright.lng = (rand() % UINT8_MAX) + 0.1 /
         ((rand() % UINT16_MAX) + 1.0);
    interval_metadata.corners_tirs.upleft.lat = (rand() % UINT8_MAX) + 0.1 /
         ((rand() % UINT16_MAX) + 1.0);
    interval_metadata.corners_tirs.upleft.lng = (rand() % UINT8_MAX) + 0.1 /
         ((rand() % UINT16_MAX) + 1.0);
    interval_metadata.corners_tirs.upright.lat = (rand() % UINT8_MAX) + 0.1 /
         ((rand() % UINT16_MAX) + 1.0);
    interval_metadata.corners_tirs.upright.lng = (rand() % UINT8_MAX) + 0.1 /
         ((rand() % UINT16_MAX) + 1.0);
    interval_metadata.corners_tirs.loleft.lat = (rand() % UINT8_MAX) + 0.1 /
         ((rand() % UINT16_MAX) + 1.0);
    interval_metadata.corners_tirs.loleft.lng = (rand() % UINT8_MAX) + 0.1 /
         ((rand() % UINT16_MAX) + 1.0);
    interval_metadata.corners_tirs.loright.lat = (rand() % UINT8_MAX) + 0.1 /
         ((rand() % UINT16_MAX) + 1.0);
    interval_metadata.corners_tirs.loright.lng = (rand() % UINT8_MAX) + 0.1 /
         ((rand() % UINT16_MAX) + 1.0);
    interval_metadata.attitude_points = rand()%UINT32_MAX;
    interval_metadata.attitude_points_missing = rand()%UINT32_MAX;
    interval_metadata.attitude_points_rejected = rand()%UINT32_MAX;
    interval_metadata.ephemeris_points = rand()%UINT32_MAX;
    interval_metadata.ephemeris_points_missing = rand()%UINT32_MAX;
    interval_metadata.ephemeris_points_rejected = rand()%UINT32_MAX;
    interval_metadata.detector_map_id_tirs = rand()%UINT32_MAX;
    interval_metadata.wrs_type = rand()%UINT8_MAX;
    interval_metadata.wrs_scenes = rand()%UINT8_MAX;
    interval_metadata.wrs_scenes_full = rand()%UINT8_MAX;
    interval_metadata.wrs_scenes_partial = rand()%UINT8_MAX;
    for (i =0 ; i<IAS_SOFTWARE_VERSION_SIZE; i++)
    {
        interval_metadata.is_version[i] = 'A' + rand()%52;
    }
    for (i =0 ; i<IAS_L0R_HOSTNAME_LENGTH; i++)
    {
        interval_metadata.hostname[i] = 'A' + rand()%52;
    }

   /* open access to L0R data */
    l0r_data = ias_l0r_open("test", "./", IAS_L0R_COMPRESSION_ON);
    assert_true(l0r_data != NULL);

    ias_l0r_open_metadata(l0r_data, IAS_WRITE);

    status = ias_l0r_set_interval_metadata(l0r_data, &interval_metadata);
    assert_true(status == SUCCESS);

    status = ias_l0r_get_interval_metadata(l0r_data,
            &interval_metadata_from_file);
    assert_true(status == SUCCESS);

    /* Verify that the values read from file match the values written */
    status = ias_l0r_mta_compare_interval(&interval_metadata,
        &interval_metadata_from_file, &different_elements_count);
    assert_true(status == SUCCESS);
    assert_true(different_elements_count == 0);

    status = ias_l0r_close_metadata(l0r_data);
    assert_true(status == SUCCESS);

    /*close access to the L0R data */
    status = ias_l0r_close(l0r_data);
    assert_true(status == SUCCESS);
}

/**
*test_metadata_create_scene_level() tests creating the scene table for metadata
*
*/
static void test_metadata_create_scene_level(void)
{
    int i = 0;
    int status;
    int different_elements_count;
    L0RIO *l0r_data = NULL;
    IAS_L0R_SCENE_METADATA scene_metadata;
    IAS_L0R_SCENE_METADATA scene_metadata_from_file;

    srand(time(NULL));

    memset(&scene_metadata, 0, sizeof(IAS_L0R_SCENE_METADATA));
    memset(&scene_metadata_from_file, 0, sizeof(IAS_L0R_SCENE_METADATA));

    scene_metadata.attitude_points = rand() % UINT16_MAX;
    scene_metadata.attitude_points_missing = rand() % UINT16_MAX;
    scene_metadata.attitude_points_rejected = rand() % UINT16_MAX;
    scene_metadata.corners_oli.upleft.lat = (rand() % UINT8_MAX) + 0.1 /
         ((rand() % UINT16_MAX) + 1.0);
    scene_metadata.corners_oli.upleft.lng = (rand() % UINT8_MAX) + 0.1 /
         ((rand() % UINT16_MAX) + 1.0);
    scene_metadata.corners_oli.upright.lat = (rand() % UINT8_MAX) + 0.1 /
         ((rand() % UINT16_MAX) + 1.0);
    scene_metadata.corners_oli.upright.lng = (rand() % UINT8_MAX) + 0.1 /
         ((rand() % UINT16_MAX) + 1.0);
    scene_metadata.corners_oli.loleft.lat = (rand() % UINT8_MAX) + 0.1 /
         ((rand() % UINT16_MAX) + 1.0);
    scene_metadata.corners_oli.loleft.lng = (rand() % UINT8_MAX) + 0.1 /
         ((rand() % UINT16_MAX) + 1.0);
    scene_metadata.corners_oli.loright.lat = (rand() % UINT8_MAX) + 0.1 /
         ((rand() % UINT16_MAX) + 1.0);
    scene_metadata.corners_oli.loright.lng = (rand() % UINT8_MAX) + 0.1 /
         ((rand() % UINT16_MAX) + 1.0);
    scene_metadata.corners_tirs.upleft.lat = (rand() % UINT8_MAX) + 0.1 /
         ((rand() % UINT16_MAX) + 1.0);
    scene_metadata.corners_tirs.upleft.lng = (rand() % UINT8_MAX) + 0.1 /
         ((rand() % UINT16_MAX) + 1.0);
    scene_metadata.corners_tirs.upright.lat = (rand() % UINT8_MAX) + 0.1 /
         ((rand() % UINT16_MAX) + 1.0);
    scene_metadata.corners_tirs.upright.lng = (rand() % UINT8_MAX) + 0.1 /
         ((rand() % UINT16_MAX) + 1.0);
    scene_metadata.corners_tirs.loleft.lat = (rand() % UINT8_MAX) + 0.1 /
         ((rand() % UINT16_MAX) + 1.0);
    scene_metadata.corners_tirs.loleft.lng = (rand() % UINT8_MAX) + 0.1 /
         ((rand() % UINT16_MAX) + 1.0);
    scene_metadata.corners_tirs.loright.lat = (rand() % UINT8_MAX) + 0.1 /
         ((rand() % UINT16_MAX) + 1.0);
    scene_metadata.corners_tirs.loright.lng = (rand() % UINT8_MAX) + 0.1 /
         ((rand() % UINT16_MAX) + 1.0);
    scene_metadata.crc_errors = rand() % UINT32_MAX;
    for (i =0 ; i<IAS_L0R_TIME_LENGTH; i++)
    {
        scene_metadata.date_acquired[i] = 'A' + rand()%52;
    }
    for (i =0 ; i<IAS_L0R_DAY_NIGHT_LENGTH; i++)
    {
        scene_metadata.day_night[i] = 'A' + rand()%52;
    }
    scene_metadata.ephemeris_points = rand() % UINT16_MAX;
    scene_metadata.ephemeris_points_missing = rand() % UINT16_MAX;
    scene_metadata.ephemeris_points_rejected = rand() % UINT16_MAX;
    for (i =0 ; i<IAS_L0R_SCENE_FULL_PARTIAL_LENGTH; i++)
    {
        scene_metadata.full_partial_scene[i] = 'A' + rand()%52;
    }
    for (i=0; i<IAS_L0R_HOSTNAME_LENGTH; i++)
    {
        scene_metadata.hostname[i] = 'A' + rand()%52;
    }
    scene_metadata.image_quality_oli = rand() % INT8_MAX;
    scene_metadata.image_quality_tirs = rand() % INT8_MAX;
    for (i =0 ; i<IAS_L0R_SCENE_ID_LENGTH; i++)
    {
        scene_metadata.landsat_scene_id[i] = 'A' + rand()%52;
    }
    scene_metadata.missing_frames = rand() % UINT16_MAX;
    for (i=0; i<IAS_L0R_NADIR_LENGTH; i++)
    {
        scene_metadata.nadir_offnadir[i] = 'A' + rand()%52;
    }
    scene_metadata.roll_angle = (rand() % UINT8_MAX) + 0.1 /
         ((rand() % UINT16_MAX) + 1.0);
    scene_metadata.scene_center.lat =  (rand() % UINT8_MAX) + 0.1 /
         ((rand() % UINT16_MAX) + 1.0);
    scene_metadata.scene_center.lng =  (rand() % UINT8_MAX) + 0.1 /
         ((rand() % UINT16_MAX) + 1.0);
    scene_metadata.scene_center_shift = rand() % INT32_MAX;
    scene_metadata.scene_start_frame_oli = rand() % UINT32_MAX;
    scene_metadata.scene_stop_frame_oli = rand() % UINT32_MAX;
    scene_metadata.scene_start_frame_tirs = rand() % UINT32_MAX;
    scene_metadata.scene_stop_frame_tirs = rand() % UINT32_MAX;
    scene_metadata.sensor_present_oli = rand() % UINT8_MAX;
    scene_metadata.sensor_present_tirs = rand() % UINT8_MAX;
    for (i =0 ; i<IAS_L0R_TIME_LENGTH; i++)
    {
        scene_metadata.start_time[i] = 'A' + rand()%52;
    }

    for (i =0 ; i<IAS_L0R_TIME_LENGTH; i++)
    {
        scene_metadata.stop_time[i] = 'A' + rand()%52;
    }
    for (i =0 ; i< IAS_SOFTWARE_VERSION_SIZE; i++)
    {
        scene_metadata.subsetter_version_l0rp[i] = 'A' + rand()%52;
    }
    scene_metadata.sun_azimuth = (rand() % UINT8_MAX) + 0.1 /
         ((rand() % UINT16_MAX) + 1.0);
    scene_metadata.sun_elevation = (rand() % UINT8_MAX) + 0.1 /
         ((rand() % UINT16_MAX) + 1.0);
    scene_metadata.target_wrs_path = rand() % UINT8_MAX;
    scene_metadata.target_wrs_row = rand() % UINT16_MAX;
    scene_metadata.time_code_errors = rand() % UINT16_MAX;
    scene_metadata.wrs_path = rand() % UINT16_MAX;
    scene_metadata.wrs_row = rand() % UINT8_MAX;
    scene_metadata.wrs_scene_number = rand() % UINT8_MAX;

    /* open access to L0R data */
    l0r_data = ias_l0r_open("test", "./", IAS_L0R_COMPRESSION_ON);
    assert_true(l0r_data != NULL);

    ias_l0r_open_metadata(l0r_data, IAS_WRITE);

    status = ias_l0r_set_scene_metadata(l0r_data, 0, 1, &scene_metadata);
    assert_true(status == SUCCESS);

    status = ias_l0r_get_scene_metadata(l0r_data,
        0, 1, &scene_metadata_from_file);
    assert_true(status == SUCCESS);

    /* Verify that the values read from file match the values written */
    status = ias_l0r_mta_compare_scene(&scene_metadata,
        &scene_metadata_from_file, &different_elements_count);
    assert_true(status == SUCCESS);
    assert_true(different_elements_count == 0);

    memset(&scene_metadata_from_file, 0, sizeof(IAS_L0R_SCENE_METADATA));
    /* Verify the read by ID is working */
    status = ias_l0r_get_scene_metadata_by_id(l0r_data,
        scene_metadata.landsat_scene_id,
        &scene_metadata_from_file);
    assert_true(status == SUCCESS);
    status = ias_l0r_mta_compare_scene(&scene_metadata,
        &scene_metadata_from_file, &different_elements_count);
    assert_true(different_elements_count == 0);

    status = ias_l0r_close_metadata(l0r_data);
    assert_true(status == SUCCESS);

    /*close access to the L0R data */
    status = ias_l0r_close(l0r_data);
    assert_true(status == SUCCESS);
}

/**
*test_metadata_delete_scene_level() tests removing the scene table
*
*/
static void test_metadata_delete_scene_level(void)
{
    int status;
    int size;
    L0RIO *l0r_data = NULL;
    IAS_L0R_SCENE_METADATA scene_metadata;
    
    /* open access to L0R data */
    l0r_data = ias_l0r_open("test", "./", IAS_L0R_COMPRESSION_ON);
    assert_true(l0r_data != NULL);

    status = ias_l0r_open_metadata(l0r_data, IAS_WRITE);
    assert_true(status == SUCCESS);

    /* Store the scene metadata, so we can put it back later */
    status = ias_l0r_get_scene_metadata(l0r_data,
        0, 1, &scene_metadata);
    assert_true(status == SUCCESS);
    
    /* Call the delete function */
    status = ias_l0r_delete_scene_metadata(l0r_data);
    assert_true(status == SUCCESS);

    /*Close and open file to ensure file is flushed */
    status = ias_l0r_close_metadata(l0r_data);
    assert_true(status == SUCCESS);
    status = ias_l0r_open_metadata(l0r_data, IAS_WRITE);
    assert_true(status == SUCCESS);
 
    status = ias_l0r_get_scene_metadata_records_count(l0r_data, &size);
    assert_true(status == SUCCESS);
    assert_true(size == 0);
    
    /* Write the stored data back to the file */
    status = ias_l0r_set_scene_metadata(l0r_data, 0, 1, &scene_metadata);
    assert_true(status == SUCCESS);
    
    status = ias_l0r_close_metadata(l0r_data);
    assert_true(status == SUCCESS);

    /*close access to the L0R data */
    status = ias_l0r_close(l0r_data);
    assert_true(status == SUCCESS);
}    

/* Test the backwards compatability for reading version 5 interval metadata*/
static void test_metadata_read_interval_v5(void)
{
    int status;
    int different_elements_count;
    L0RIO *l0r_data = NULL;
    L0RIO *l0r_data_v6 = NULL;
    IAS_L0R_INTERVAL_METADATA interval_metadata_v5;
    IAS_L0R_INTERVAL_METADATA interval_metadata_v6;

    ias_log_set_output_level(IAS_LOG_LEVEL_ERROR);

    /* open access to L0R data */
    l0r_data = ias_l0r_open("regression_metadata_v5", "./", 
        IAS_L0R_COMPRESSION_ON);
    assert_true(l0r_data != NULL);

    ias_l0r_open_metadata(l0r_data, IAS_READ);

    status = ias_l0r_get_interval_metadata(l0r_data, &interval_metadata_v5);
    assert_true(status == SUCCESS);

    status = ias_l0r_close_metadata(l0r_data);
    assert_true(status == SUCCESS);

    /*close access to the L0R data */
    status = ias_l0r_close(l0r_data);
    assert_true(status == SUCCESS);

    /* open access to L0R data */
    l0r_data_v6 = ias_l0r_open("regression_metadata_v6", "./",
        IAS_L0R_COMPRESSION_ON);
    assert_true(l0r_data_v6 != NULL);

    ias_l0r_open_metadata(l0r_data_v6, IAS_READ);

    status = ias_l0r_get_interval_metadata(l0r_data_v6, &interval_metadata_v6);
    assert_true(status == SUCCESS);

    /* Verify that the values read from file match the values written */
    status = ias_l0r_mta_compare_interval(&interval_metadata_v5,
        &interval_metadata_v6, &different_elements_count);
    assert_true(status == SUCCESS);
    assert_true(different_elements_count == 0);

    status = ias_l0r_close_metadata(l0r_data_v6);
    assert_true(status == SUCCESS);

    /*close access to the L0R data */
    status = ias_l0r_close(l0r_data_v6);
    assert_true(status == SUCCESS);
}


int main(int argc, char **argv)
{
    TestSuite *suite = create_test_suite();
    TestReporter *reporter = create_text_reporter();
    int status = 0;

    status = ias_log_initialize("L0R Metadata Test");
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Error encountered when initializing the ias logging "
            "library");
        return ERROR;
    }
    if (ias_sat_attr_initialize(IAS_L8) != SUCCESS)
    {
        IAS_LOG_ERROR("Error initializing attributes");
        return ERROR;
    }
    ias_log_set_output_level(IAS_LOG_LEVEL_ERROR);

    /* Perform tests for metadata */
    add_test(suite, test_metadata_create_file_level);
    add_test(suite, test_metadata_create_interval_level);
    add_test(suite, test_metadata_read_interval_v5);
    add_test(suite, test_metadata_create_scene_level);
    add_test(suite, test_metadata_delete_scene_level);

    return run_test_suite(suite, reporter);
}
