#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include "ias_l1g.h"
#include "ias_types.h" /* IAS_DATA_TYPE, IAS_ACCESS_MODE definitions */
#include "ias_const.h" /* SUCCESS/ERROR definitions */
#include "ias_logging.h"
#include "cgreen/cgreen.h"

void dump_band_metadata(const IAS_L1G_BAND_METADATA *bmd)
{
    printf("band number: %d\n", bmd->band_number);
    printf("band name: %s\n", bmd->band_name);
    printf("uly: %f\n", bmd->upper_left_y);
    printf("ulx: %f\n", bmd->upper_left_x);
    printf("ury: %f\n", bmd->upper_right_y);
    printf("urx: %f\n", bmd->upper_right_x);
    printf("lly: %f\n", bmd->lower_left_y);
    printf("llx: %f\n", bmd->lower_left_x);
    printf("lry: %f\n", bmd->lower_right_y);
    printf("lrx: %f\n", bmd->lower_right_x);
    printf("proj dist y: %f\n", bmd->projection_distance_y);
    printf("proj dist x: %f\n", bmd->projection_distance_x);
    printf("max pixel value: %f\n", bmd->maximum_pixel_value);
    printf("min pixel value: %f\n", bmd->minimum_pixel_value);
    printf("pixel range valid: %d\n", bmd->pixel_range_valid);
    printf("max radiance: %f\n", bmd->maximum_radiance);
    printf("min radiance: %f\n", bmd->minimum_radiance);
    printf("scaling offset: %f\n", bmd->spectral_radiance_scaling_offset);
    printf("scaling gain: %f\n", bmd->spectral_radiance_scaling_gain);
    printf("scaline valid: %d\n", bmd->radiance_valid);
    printf("instrument source: %s\n", bmd->instrument_source);
}

Ensure test_image_open_close()
{
    L1GIO *img;
    IAS_L1G_FILE_METADATA fmd;
    int status;
    const int BAND_NUMBER_LIST_SIZE=10;
    int band_number_list[BAND_NUMBER_LIST_SIZE];
    int number_of_bands;

    img = ias_l1g_open_image("test.h5", IAS_WRITE);
    assert_true(img != NULL);

    /* make sure the file is L1G file */
    status = ias_l1g_is_l1g_file("test.h5");
    assert_equal(status, 1);

    /* make sure no bands are found */
    assert_equal(ias_l1g_get_band_count(img), 0);
    assert_equal(ias_l1g_is_band_present(img, 1), 0);
    status=ias_l1g_get_band_list(img, band_number_list, BAND_NUMBER_LIST_SIZE, 
                                 &number_of_bands );
    assert_equal(status, SUCCESS);
    assert_equal(number_of_bands, 0);

    status = ias_l1g_close_image(img);
    assert_equal(status, SUCCESS);

    /* reopen the image and make sure it doesn't complain when trying
       to set the file metadata */
    img = ias_l1g_open_image("test.h5", IAS_UPDATE);
    assert_true(img != NULL);

    memset(&fmd, 0, sizeof(fmd));
    status = ias_l1g_set_file_metadata(img, &fmd);
    assert_equal(status, SUCCESS);

    status = ias_l1g_close_image(img);
    assert_equal(status, SUCCESS);
}

Ensure test_band_io()
{
    L1GIO *img;
    int number_of_scas = 2;
    int status;
    int i;
    IAS_L1G_FILE_METADATA fmd;
    IAS_L1G_FILE_METADATA fmd_read;
    IAS_L1G_BAND_METADATA bmd;

    /* open the test image */
    img = ias_l1g_open_image("test.h5", IAS_WRITE);
    assert_true(img != NULL);

    if (img != NULL)
    {
        int lines = 20;
        int samples = 10;
        IAS_DATA_TYPE data_type = IAS_I2;
        L1G_BAND_IO *l1g_band;
        const int BAND_NUMBER_LIST_SIZE=10;
        int band_number_list[BAND_NUMBER_LIST_SIZE];
        int number_of_bands;

        assert_equal(ias_l1g_is_band_present(img, 1), 0);

        /* open the test band */
        l1g_band = ias_l1g_open_band(img, 1, &data_type, &number_of_scas,
                                   &lines, &samples);
        assert_true(l1g_band != NULL);

        if (l1g_band != NULL)
        {
            short int data[30];
            short int data_read[30];
            short int data2[30];
            short int data_read2[30];
            int i;
            int old_log_level;
            for (i = 0; i < 30; i++)
            {
                data[i] = 30 + i;
                data2[i] = 60 + i;
            }

            assert_equal(ias_l1g_get_band_count(img), 1);
            assert_equal(ias_l1g_is_band_present(img, 1), 1);
            assert_equal(ias_l1g_is_band_present(img, 2), 0);
            status=ias_l1g_get_band_list(img, band_number_list, 
                                         BAND_NUMBER_LIST_SIZE, 
                                         &number_of_bands );
            assert_equal(status, SUCCESS);
            assert_equal(number_of_bands, 1);
            assert_equal(band_number_list[0],1);

            /* temporarily turn off ias log messages since we expect a 
               failure */
            old_log_level = ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

            /* test for a list size too small */
            status=ias_l1g_get_band_list(img, band_number_list, 
                                         0, &number_of_bands );
            assert_equal(status, ERROR);

            /* restore normal logging level */
            ias_log_set_output_level(old_log_level);

            /* test writing the imagery to sca 0 */
            status = ias_l1g_write_image(l1g_band, 0, 2, 3, 5, 6, data);
            assert_equal(status, SUCCESS);

            /* test writing another chunk to sca 0 */
            status = ias_l1g_write_image(l1g_band, 0, 12, 3, 5, 6, data2);
            assert_equal(status, SUCCESS);

            /* test writing the imagery to sca 1 */
            status = ias_l1g_write_image(l1g_band, 1, 4, 5, 6, 5, data2);
            assert_equal(status, SUCCESS);

            /* read the imagery from the file */
            status = ias_l1g_read_image(l1g_band, 0, 2, 3, 5, 6, data_read);
            assert_equal(status, SUCCESS);

            /* verify the data read is the same as the data written */
            assert_equal(memcmp(data, data_read, sizeof(data)), 0);

            /* read the imagery from the file */
            status = ias_l1g_read_image(l1g_band, 0, 12, 3, 5, 6, data_read);
            assert_equal(status, SUCCESS);

            /* verify the data read is the same as the data written */
            assert_equal(memcmp(data2, data_read, sizeof(data2)), 0);

            /* verify sca 1 reads correctly */
            status = ias_l1g_read_image(l1g_band, 1, 4, 5, 6, 5, data_read2);
            assert_equal(status, SUCCESS);

            /* verify the data read is the same as the data written */
            assert_equal(memcmp(data2, data_read2, sizeof(data2)), 0);

            /* make sure the full image can be written and read */
            status = ias_l1g_write_image(l1g_band, 0, lines - 1, 0, 1, samples, 
                                         data);
            assert_equal(status, SUCCESS);

            status = ias_l1g_read_image(l1g_band, 0, lines - 1, 0, 1, samples,
                                        data_read);
            assert_equal(status, SUCCESS);
            assert_equal(memcmp(data, data_read, samples * sizeof(short int)),
                         0);

            /* temporarily turn off ias log messages since we expect a 
               failure */
            ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

            /* attempt to write to illegal locations */
            /* test bad sca numbers */
            status = ias_l1g_write_image(l1g_band, -1, 0, 0, 1, 1, data);
            assert_equal(status, ERROR);
            status = ias_l1g_write_image(l1g_band, number_of_scas + 1, 0, 0, 1,
                                         1, data);
            assert_equal(status, ERROR);

            /* test bad lines */
            status = ias_l1g_write_image(l1g_band, 0, -1, 0, 1, 1, data);
            assert_equal(status, ERROR);
            status = ias_l1g_write_image(l1g_band, 0, 1, 0, lines, 1, data);
            assert_equal(status, ERROR);

            /* test bad samples */
            status = ias_l1g_write_image(l1g_band, 0, 0, -1, 1, 1, data);
            assert_equal(status, ERROR);
            status = ias_l1g_write_image(l1g_band, 0, 0, 1, 1, samples, data);
            assert_equal(status, ERROR);

            /* attempt to read an illegal location in the file to make sure
               it gets an error */
            /* test bad scas */
            status = ias_l1g_read_image(l1g_band, -1, 0, 0, 1, 1, data_read);
            assert_equal(status, ERROR);
            status = ias_l1g_read_image(l1g_band, number_of_scas + 1, 0, 0, 1,
                                        1, data_read);
            assert_equal(status, ERROR);

            /* test bad lines */
            status = ias_l1g_read_image(l1g_band, 0, -1, 0, 1, 1, data_read);
            assert_equal(status, ERROR);
            status = ias_l1g_read_image(l1g_band, 0, 1, 0, lines, 1, data_read);
            assert_equal(status, ERROR);

            /* test bad samples */
            status = ias_l1g_read_image(l1g_band, 0, 0, -1, 1, 1, data_read);
            assert_equal(status, ERROR);
            status = ias_l1g_read_image(l1g_band, 0, 0, 1, 1, samples,
                                        data_read);
            assert_equal(status, ERROR);

            /* restore normal logging level */
            ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

            /* close the test band */
            status = ias_l1g_close_band(l1g_band);
            assert_equal(status, SUCCESS);
            l1g_band = NULL;
        }

        /* write some file metadata */
        memset(&fmd, 0, sizeof(fmd));
        fmd.projection_code = 1;
        fmd.zone_code = 2;
        strcpy(fmd.datum, "WGS84");
        fmd.spheroid_code = 3;
        strcpy(fmd.projection_units, "METERS");
        for (i = 0; i < 15; i++)
            fmd.projection_parameters[i] = i + 1.1;
        fmd.wrs_path = 29;
        fmd.wrs_row = 30;
        fmd.roll_angle = 3.1;
        strcpy(fmd.spacecraft, "OLI");
        strcpy(fmd.collection_type, "EARTH");
        strcpy(fmd.capture_direction, "Descending and Down");
        strcpy(fmd.capture_date, "2008-11-12");
        strcpy(fmd.capture_time, "10:30:02");
        strcpy(fmd.correction_type, "L1T");
        strcpy(fmd.resample_type, "CC");
        strcpy(fmd.software_version, "1.2.3");
        strcpy(fmd.ingest_software_version, "1.2.3");
        strcpy(fmd.sun_angle_correction, "PER_PIXEL");
        fmd.earth_sun_distance = 1.03786321;
        fmd.sun_azimuth = 141.14;
        fmd.sun_elevation = 53.35;
        fmd.sun_angles_valid = 1;
        status = ias_l1g_set_file_metadata(img, &fmd);
        assert_equal(status, SUCCESS);

        /* write some band metadata */
        bmd.band_number = 1;
        strcpy(bmd.band_name, "RED");
        bmd.upper_left_y = 100.1;
        bmd.upper_left_x = 200.1;
        bmd.upper_right_y = 100.1;
        bmd.upper_right_x = 250.1;
        bmd.lower_left_y = 150.1;
        bmd.lower_left_x = 200.1;
        bmd.lower_right_y = 150.1;
        bmd.lower_right_x = 250.1;
        bmd.projection_distance_x = 10.0;
        bmd.projection_distance_y = 10.1;
        bmd.maximum_pixel_value = 50.0;
        bmd.minimum_pixel_value = 0.0;
        bmd.pixel_range_valid = 1;
        bmd.maximum_radiance = 50.5;
        bmd.minimum_radiance = 2.1;
        bmd.spectral_radiance_scaling_offset = 0.2;
        bmd.spectral_radiance_scaling_gain = 0.99;
        bmd.radiance_valid = 1;
        bmd.reflectance_scaling_offset = 0.1;
        bmd.reflectance_scaling_gain = 1.1;
        bmd.reflectance_valid = 1;
        strcpy(bmd.instrument_source, "OLI Red Band");
        status = ias_l1g_set_band_metadata(img, &bmd, 1);
        assert_equal(status, SUCCESS);

        /* read the file metadata and make sure it is identical to the 
           written data */
        status = ias_l1g_get_file_metadata(img, &fmd_read);
        assert_equal(status,SUCCESS);
        assert_equal(memcmp(&fmd, &fmd_read, sizeof(fmd)), 0);

        /* close the test image */
        status = ias_l1g_close_image(img);
        assert_equal(status, SUCCESS);
    }
}

Ensure test_band_read()
{
    L1GIO *img;
    L1G_BAND_IO *l1g_band;
    int number_of_scas = 0;
    int status;
    IAS_L1G_BAND_METADATA bmd;
    IAS_L1G_FILE_METADATA fmd;
    const int BAND_NUMBER_LIST_SIZE=10;
    int band_number_list[BAND_NUMBER_LIST_SIZE];
    int number_of_bands;

    /* open the test image */
    img = ias_l1g_open_image("test.h5", IAS_READ);
    assert_true(img != NULL);

    if (img != NULL)
    {
        int lines;
        int samples;
        IAS_DATA_TYPE data_type;

        assert_equal(ias_l1g_get_band_count(img), 1);
        status=ias_l1g_get_band_list(img, band_number_list, 
                                         BAND_NUMBER_LIST_SIZE, 
                                         &number_of_bands );
        assert_equal(status, SUCCESS);
        assert_equal(number_of_bands, 1);
        assert_equal(band_number_list[0],1);

        /* open the band */
        l1g_band = ias_l1g_open_band(img, 1, &data_type, &number_of_scas,
                                   &lines, &samples);
        assert_true(l1g_band != NULL);

        if (l1g_band != NULL)
        {
            short int data[1] = {1};
            short int data_read[30];
            int index;

            assert_equal(data_type, IAS_I2);
            assert_equal(number_of_scas, 2);
            assert_equal(lines, 20);
            assert_equal(samples, 10);

            /* temporarily turn off ias log messages since we expect a 
               failure */
            ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

            /* verify that trying to write to the band fails */
            status = ias_l1g_write_image(l1g_band, 0, 1, 1, 1, 1, data);
            assert_equal(status, ERROR);

            /* restore normal logging level */
            ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

            /* make sure reading works as expected since it takes a slightly
               different path in the library than the test_band_io test */
            status = ias_l1g_read_image(l1g_band, 0, 2, 3, 5, 6, data_read);
            assert_equal(status, SUCCESS);
            for (index = 0; index < sizeof(data_read)/sizeof(data_read[0]);
                 index++)
            {
                assert_equal(data_read[index], 30 + index);
            }

            /* close the band */
            status = ias_l1g_close_band(l1g_band);
            assert_equal(status, SUCCESS);
        }

        /* get the metadata and make sure it is set to the previously written
           value */
        status = ias_l1g_get_band_metadata(img, 1, &bmd);
        assert_equal(status, SUCCESS);

        /* verify the values read are correct */
        assert_equal(bmd.band_number, 1);
        assert_string_equal(bmd.band_name, "RED");
        assert_double_equal(bmd.upper_left_y, 100.1);
        assert_double_equal(bmd.upper_left_x, 200.1);
        assert_double_equal(bmd.upper_right_y, 100.1);
        assert_double_equal(bmd.upper_right_x, 250.1);
        assert_double_equal(bmd.lower_left_y, 150.1);
        assert_double_equal(bmd.lower_left_x, 200.1);
        assert_double_equal(bmd.lower_right_y, 150.1);
        assert_double_equal(bmd.lower_right_x, 250.1);
        assert_double_equal(bmd.projection_distance_x, 10.0);
        assert_double_equal(bmd.projection_distance_y, 10.1);
        assert_double_equal(bmd.maximum_pixel_value, 50.0);
        assert_double_equal(bmd.minimum_pixel_value, 0.0);
        assert_equal(bmd.pixel_range_valid, 1);
        assert_double_equal(bmd.maximum_radiance, 50.5);
        assert_double_equal(bmd.minimum_radiance, 2.1);
        assert_double_equal(bmd.spectral_radiance_scaling_offset, 0.2);
        assert_double_equal(bmd.spectral_radiance_scaling_gain, 0.99);
        assert_equal(bmd.radiance_valid, 1);
        assert_double_equal(bmd.reflectance_scaling_offset, 0.1);
        assert_double_equal(bmd.reflectance_scaling_gain, 1.1);
        assert_equal(bmd.reflectance_valid, 1);
        assert_string_equal(bmd.instrument_source, "OLI Red Band");

        /* temporarily turn off ias log messages since we expect a failure */
        ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

        /* attempt to write the metadata back to the file which should fail */
        status = ias_l1g_set_band_metadata(img, &bmd, 1);
        assert_equal(status, ERROR);

        /* attempt to write the file metadata which should fail */
        status = ias_l1g_set_file_metadata(img, &fmd);
        assert_equal(status, ERROR);

        /* restore normal logging level */
        ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

        /* close the test image */
        status = ias_l1g_close_image(img);
        assert_equal(status, SUCCESS);
    }
}

Ensure test_band_update()
{
    L1GIO *img;
    L1G_BAND_IO *l1g_band;
    int status;
    const int BAND_NUMBER_LIST_SIZE=10;
    int band_number_list[BAND_NUMBER_LIST_SIZE];
    int number_of_bands;

    /* open the test image */
    img = ias_l1g_open_image("test.h5", IAS_UPDATE);
    assert_true(img != NULL);

    if (img)
    {
        int number_of_scas;
        int lines;
        int samples;
        IAS_DATA_TYPE data_type;
        IAS_L1G_BAND_METADATA bmd[2];
        IAS_L1G_BAND_METADATA read_bmd;
        IAS_L1G_FILE_METADATA fmd;

        /* open the band */
        l1g_band = ias_l1g_open_band(img, 1, &data_type, &number_of_scas,
                                   &lines, &samples);
        assert_true(l1g_band != NULL);

        if (l1g_band != NULL)
        {
            short int data[6] = {1, 2, 3, 4, 5, 6};
            short int data_read[6];

            assert_equal(ias_l1g_get_band_count(img), 1);
            status=ias_l1g_get_band_list(img, band_number_list, 
                                         BAND_NUMBER_LIST_SIZE, 
                                         &number_of_bands );
            assert_equal(status, SUCCESS);
            assert_equal(number_of_bands, 1);
            assert_equal(band_number_list[0],1);
            assert_equal(data_type, IAS_I2);
            assert_equal(number_of_scas, 2);
            assert_equal(lines, 20);
            assert_equal(samples, 10);

            /* now write an update to the open band and verify it
               is written successfully */
            status = ias_l1g_write_image(l1g_band, 0, 0, 1, 2, 3, data);
            assert_equal(status, SUCCESS);

            /* read the imagery from the file */
            status = ias_l1g_read_image(l1g_band, 0, 0, 1, 2, 3, data_read);
            assert_equal(status, SUCCESS);

            /* verify the data read is the same as the data written */
            assert_equal(memcmp(data, data_read, sizeof(data)), 0);

            /* close the band */
            status = ias_l1g_close_band(l1g_band);
            assert_equal(status, SUCCESS);
        }

        /* create a new band in the file and verify it can be written with
           the float data type */
        number_of_scas = 1;
        lines = 40;
        samples = 20;
        data_type = IAS_R4;
        l1g_band = ias_l1g_open_band(img, 2, &data_type, &number_of_scas,
                                   &lines, &samples);
        assert_true(l1g_band != NULL);

        if (img)
        {
            float data[8] = {1.0, 1.2, 1.4, 1.6, 2.2, 2.4, 2.6, 2.8};
            float data_read[8];

            assert_equal(ias_l1g_get_band_count(img), 2);
            assert_equal(ias_l1g_is_band_present(img, 2), 1);
            assert_equal(ias_l1g_is_band_present(img, 20), 0);
            status=ias_l1g_get_band_list(img, band_number_list, 
                                         BAND_NUMBER_LIST_SIZE, 
                                         &number_of_bands );
            assert_equal(status, SUCCESS);
            assert_equal(number_of_bands, 2);
            assert_equal(band_number_list[0],1);
            assert_equal(band_number_list[1],2);

            /* write data to the band, read it, and make sure it is the same */
            status = ias_l1g_write_image(l1g_band, 0, 1, 0, 2, 4, data);
            assert_equal(status, SUCCESS);

            status = ias_l1g_read_image(l1g_band, 0, 1, 0, 2, 4, data_read);
            assert_equal(status, SUCCESS);

            assert_equal(memcmp(data, data_read, sizeof(data)), 0);

            /* close the band */
            status = ias_l1g_close_band(l1g_band);
            assert_equal(status, SUCCESS);
        }

        /* add some band metadata for the blue band to the file */
        memset(bmd, 0, sizeof(bmd));
        bmd[0].band_number = 2;
        strcpy(bmd[0].band_name, "BLUE");
        bmd[0].upper_left_y = 100.1;
        bmd[0].upper_left_x = 200.1;
        bmd[0].upper_right_y = 100.1;
        bmd[0].upper_right_x = 250.1;
        bmd[0].lower_left_y = 150.1;
        bmd[0].lower_left_x = 200.1;
        bmd[0].lower_right_y = 150.1;
        bmd[0].lower_right_x = 250.1;
        bmd[0].projection_distance_x = 110.0;
        bmd[0].projection_distance_y = 110.1;
        bmd[0].maximum_pixel_value = 150.0;
        bmd[0].minimum_pixel_value = 10.0;
        bmd[0].pixel_range_valid = 1;
        bmd[0].maximum_radiance = 150.5;
        bmd[0].minimum_radiance = 12.1;
        bmd[0].spectral_radiance_scaling_offset = 0.1;
        bmd[0].spectral_radiance_scaling_gain = 1.01;
        bmd[0].radiance_valid = 1;
        bmd[0].reflectance_scaling_offset = 0.3;
        bmd[0].reflectance_scaling_gain = 1.04;
        bmd[0].reflectance_valid = 1;
        strcpy(bmd[0].instrument_source, "OLI Blue Band");
        memcpy(&bmd[1], &bmd[0], sizeof(bmd[1]));
        bmd[1].band_number = 3;
        bmd[1].radiance_valid = 0;
        bmd[1].reflectance_valid = 0;
        status = ias_l1g_set_band_metadata(img, bmd, 2);
        assert_equal(status, SUCCESS);

        /* change one piece of metadata and set the same bands again */
        bmd[0].maximum_radiance = 152.3;
        status = ias_l1g_set_band_metadata(img, bmd, 2);
        assert_equal(status, SUCCESS);

        /* read the file metadata, change a field, then write it back to make
           sure it can be updated */
        status = ias_l1g_get_file_metadata(img, &fmd);
        assert_equal(status, SUCCESS);
        fmd.projection_parameters[1] = 6.5;
        strcpy(fmd.capture_direction, "Descending");
        status = ias_l1g_set_file_metadata(img, &fmd);
        assert_equal(status, SUCCESS);
        memset(&fmd, 0, sizeof(fmd));
        status = ias_l1g_get_file_metadata(img, &fmd);
        assert_equal(status, SUCCESS);
        assert_double_equal(fmd.projection_parameters[1], 6.5);
        assert_equal(strcmp(fmd.capture_direction, "Descending"), 0);

        /* read back the band metadata and verify it is correct */
        status = ias_l1g_get_band_metadata(img, 2, &read_bmd);
        assert_equal(status, SUCCESS);
        assert_true(memcmp(&bmd[0], &read_bmd, sizeof(read_bmd)) == 0);
        status = ias_l1g_get_band_metadata(img, 3, &read_bmd);
        assert_equal(status, SUCCESS);
        assert_double_equal(read_bmd.maximum_radiance, 0.0);
        assert_double_equal(read_bmd.minimum_radiance, 0.0);
        assert_double_equal(read_bmd.spectral_radiance_scaling_offset, 0.0);
        assert_double_equal(read_bmd.spectral_radiance_scaling_gain, 0.0);
        assert_double_equal(read_bmd.reflectance_scaling_offset, 0.0);
        assert_double_equal(read_bmd.reflectance_scaling_gain, 0.0);
        /* reset the invalid values to zero */
        bmd[1].maximum_radiance = 0.0;
        bmd[1].minimum_radiance = 0.0;
        bmd[1].spectral_radiance_scaling_offset = 0.0;
        bmd[1].spectral_radiance_scaling_gain = 0.0;
        bmd[1].reflectance_scaling_offset = 0.0;
        bmd[1].reflectance_scaling_gain = 0.0;
        assert_true(memcmp(&bmd[1], &read_bmd, sizeof(read_bmd)) == 0);

        /* close the test image */
        status = ias_l1g_close_image(img);
        assert_equal(status, SUCCESS);
        img = NULL;

        /* reopen the test image to make sure the right band metadata record
           is on disk.  This check verifies that existing records on disk are
           updated when they should be instead of writing new ones */
        img = ias_l1g_open_image("test.h5", IAS_UPDATE);
        assert_true(img != NULL);

        status = ias_l1g_get_band_metadata(img, 2, &read_bmd);
        assert_equal(status, SUCCESS);
        assert_double_equal(read_bmd.maximum_radiance, 152.3);

        ias_l1g_close_image(img);
        assert_equal(status, SUCCESS);
    }
}

Ensure test_open_multiple_bands()
{
    L1GIO *img;
    L1G_BAND_IO *l1g_band;
    L1G_BAND_IO *l1g_band_2;
    int status;
    const int BAND_NUMBER_LIST_SIZE=10;
    int band_number_list[BAND_NUMBER_LIST_SIZE];
    int number_of_bands;

    /* open the test image */
    img = ias_l1g_open_image("test.h5", IAS_UPDATE);
    assert_true(img != NULL);

    if (img)
    {
        int number_of_scas;
        int lines;
        int samples;
        IAS_DATA_TYPE data_type;

        /* open the band */
        l1g_band = ias_l1g_open_band(img, 1, &data_type, &number_of_scas,
                                   &lines, &samples);
        assert_true(l1g_band != NULL);

        if (l1g_band != NULL)
        {
            short int data[6] = {1, 2, 3, 4, 5, 6};
            short int data_read[6];

            assert_equal(ias_l1g_get_band_count(img), 2);
            status=ias_l1g_get_band_list(img, band_number_list, 
                                         BAND_NUMBER_LIST_SIZE, 
                                         &number_of_bands );
            assert_equal(status, SUCCESS);
            assert_equal(number_of_bands, 2);
            assert_equal(band_number_list[0],1);
            assert_equal(band_number_list[1],2);
            assert_equal(data_type, IAS_I2);
            assert_equal(number_of_scas, 2);
            assert_equal(lines, 20);
            assert_equal(samples, 10);

            /* now write an update to the open band and verify it
               is written successfully */
            status = ias_l1g_write_image(l1g_band, 0, 0, 1, 2, 3,
                                         data);
            assert_equal(status, SUCCESS);

            /* read the imagery from the file */
            status = ias_l1g_read_image(l1g_band, 0, 0, 1, 2, 3,
                                        data_read);
            assert_equal(status, SUCCESS);

            /* verify the data read is the same as the data written */
            assert_equal(memcmp(data, data_read, sizeof(data)), 0);
        }


        /* create a new band in the file and verify it can be written with
           the float data type */
        number_of_scas = 1;
        lines = 40;
        samples = 20;
        data_type = IAS_R4;
        l1g_band_2 = ias_l1g_open_band(img, 2, &data_type, &number_of_scas,
                                   &lines, &samples);
        assert_true(l1g_band_2 != NULL);

        if (l1g_band_2 != NULL)
        {
            float data[8] = {1.0, 1.2, 1.4, 1.6, 2.2, 2.4, 2.6, 2.8};
            float data_read[8];

            assert_equal(ias_l1g_get_band_count(img), 2);
            assert_equal(ias_l1g_is_band_present(img, 2), 1);
            assert_equal(ias_l1g_is_band_present(img, 20), 0);
            status=ias_l1g_get_band_list(img, band_number_list, 
                                         BAND_NUMBER_LIST_SIZE, 
                                         &number_of_bands );
            assert_equal(status, SUCCESS);
            assert_equal(number_of_bands, 2);
            assert_equal(band_number_list[0],1);
            assert_equal(band_number_list[1],2);

            /* write data to the band, read it, and make sure it is the same */
            status = ias_l1g_write_image(l1g_band_2, 0, 1, 0, 2, 4,
                                         data);
            assert_equal(status, SUCCESS);

            status = ias_l1g_read_image(l1g_band_2, 0, 1, 0, 2, 4,
                                        data_read);
            assert_equal(status, SUCCESS);

            assert_equal(memcmp(data, data_read, sizeof(data)), 0);

            /* close the 2nd band */
            status = ias_l1g_close_band(l1g_band_2);
            assert_equal(status, SUCCESS);
        }

        /* close the 1st band */
        /* NOTE: disabling this close to allow testing the code that will
           automatically close open bands when close_image is called
        status = ias_l1g_close_band(l1g_band);
        assert_equal(status, SUCCESS); */

        /* close the test image */
        status = ias_l1g_close_image(img);
        assert_equal(status, SUCCESS);
    }
}

Ensure test_u16_bit_band()
{
    int number_of_scas;
    int lines;
    int samples;
    IAS_DATA_TYPE data_type;
    L1GIO *img;
    L1G_BAND_IO *l1g_band;
    int status;
    const int BAND_NUMBER_LIST_SIZE=10;
    int band_number_list[BAND_NUMBER_LIST_SIZE];
    int number_of_bands;

    /* open the test image */
    img = ias_l1g_open_image("test.h5", IAS_UPDATE);
    assert_true(img != NULL);

    if (img == NULL)
    {
        return;
    }

    /* create a new band in the file and verify it can be written with
       the unsigned 16 bit data type */
    number_of_scas = 1;
    lines = 40;
    samples = 20;
    data_type = IAS_UI2;
    l1g_band = ias_l1g_open_band(img, 3, &data_type, &number_of_scas,
                               &lines, &samples);
    assert_true(l1g_band != NULL);

    if (l1g_band != NULL)
    {
        unsigned short int data[8] = {5010, 40512, 54020, 1600, 22, 24, 26, 28};
        unsigned short int data_read[8];

        assert_equal(ias_l1g_get_band_count(img), 3);
        assert_equal(ias_l1g_is_band_present(img, 3), 1);
        assert_equal(ias_l1g_is_band_present(img, 20), 0);
        status=ias_l1g_get_band_list(img, band_number_list, 
                                     BAND_NUMBER_LIST_SIZE, 
                                     &number_of_bands );
        assert_equal(status, SUCCESS);
        assert_equal(number_of_bands, 3);
        assert_equal(band_number_list[0],1);
        assert_equal(band_number_list[1],2);
        assert_equal(band_number_list[2],3);

        /* write data to the band, read it, and make sure it is the same */
        status = ias_l1g_write_image(l1g_band, 0, 1, 0, 2, 4,
                                     data);
        assert_equal(status, SUCCESS);

        status = ias_l1g_read_image(l1g_band, 0, 1, 0, 2, 4,
                                    data_read);
        assert_equal(status, SUCCESS);

        assert_equal(memcmp(data, data_read, sizeof(data)), 0);

        /* close the band */
        status = ias_l1g_close_band(l1g_band);
        assert_equal(status, SUCCESS);

        status = ias_l1g_close_image(img);
        assert_equal(status, SUCCESS);
    }

}

/* the following test was created to help verify that multiple calls to
   set the band metadata results in a correct output file */
Ensure test_multiple_set_band_metadata_calls()
{
    L1GIO *img;
    IAS_L1G_BAND_METADATA bmd;
    IAS_L1G_BAND_METADATA read_bmd;
    int status;
    const int BAND_NUMBER_LIST_SIZE=10;
    int band_number_list[BAND_NUMBER_LIST_SIZE];
    int number_of_bands;

    img = ias_l1g_open_image("test_bmd.h5", IAS_WRITE);
    assert_true(img != NULL);

    /* make sure no bands are found */
    assert_equal(ias_l1g_get_band_count(img), 0);
    assert_equal(ias_l1g_is_band_present(img, 1), 0);
    status=ias_l1g_get_band_list(img, band_number_list, 
                                     BAND_NUMBER_LIST_SIZE, 
                                     &number_of_bands );
    assert_equal(status, SUCCESS);
    assert_equal(number_of_bands, 0);

    memset(&bmd, 0, sizeof(bmd));
    bmd.band_number = 1;
    status = ias_l1g_set_band_metadata(img, &bmd, 1);
    assert_equal(status, SUCCESS);

    bmd.band_number = 2;
    status = ias_l1g_set_band_metadata(img, &bmd, 1);
    assert_equal(status, SUCCESS);

    status = ias_l1g_close_image(img);
    assert_equal(status, SUCCESS);

    /* open the image */
    img = ias_l1g_open_image("test_bmd.h5", IAS_READ);
    assert_true(img != NULL);

    /* make sure the reading doesn't detect any errors */
    status = ias_l1g_get_band_metadata(img, 1, &read_bmd);
    assert_equal(status, SUCCESS);

    status = ias_l1g_close_image(img);
    assert_equal(status, SUCCESS);
}

Ensure test_read_write_version_1_file_metadata()
{
    L1GIO *img;
    IAS_L1G_FILE_METADATA fmd;
    int status;

    /* Copy the template file to be modified */
    if (system("cp template_test_ver_1.h5 test_ver_1.h5") != 0)
    {
        assert_true_with_message(0, "Copying template_test_ver_1.h5 "
            "to test_ver_1.h5");
        return;
    }

    img = ias_l1g_open_image("test_ver_1.h5", IAS_READ);
    assert_true(img != NULL);

    /* Make sure the file is L1G file */
    status = ias_l1g_is_l1g_file("test_ver_1.h5");
    assert_equal(status, 1);

    status = ias_l1g_get_file_metadata(img, &fmd);
    assert_equal(status, SUCCESS);
    
    /* Make sure the newer version fields are not set */
    assert_equal(fmd.earth_sun_distance, 0.0);
    assert_string_equal(fmd.sun_angle_correction, "");
    
    status = ias_l1g_close_image(img);
    assert_equal(status, SUCCESS);

    img = ias_l1g_open_image("test_ver_1.h5", IAS_UPDATE);
    assert_true(img != NULL);

    strcpy(fmd.sun_angle_correction, "PER_PIXEL");
    fmd.earth_sun_distance = 1.03786321;

    status = ias_l1g_set_file_metadata(img, &fmd);
    assert_equal(status, SUCCESS);

    status = ias_l1g_close_image(img);
    assert_equal(status, SUCCESS);

    /* Compare the metadata files to make sure no changes were made */
    if (system("h5diff template_test_ver_1.h5 test_ver_1.h5") != 0)
    {
        assert_true_with_message(0, "h5diff template_test_ver_1.h5 "
            "test_ver_1.h5 failed");
        return;
    }

    assert_equal(unlink("test_ver_1.h5"), 0);
}

int main(int argc, char **argv)
{
    TestSuite *suite = create_test_suite();
    add_test(suite, test_image_open_close);
    add_test(suite, test_band_io);
    add_test(suite, test_band_read);
    add_test(suite, test_band_update);
    add_test(suite, test_open_multiple_bands);
    add_test(suite, test_multiple_set_band_metadata_calls);
    add_test(suite, test_u16_bit_band);
    add_test(suite, test_read_write_version_1_file_metadata);
    if (argc > 1) {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}
