#include <string.h>
#include <stdio.h>
#include <limits.h>
#include "ias_l1r.h"
#include "ias_types.h"
#include "ias_const.h"
#include "ias_logging.h"
#include "cgreen/cgreen.h"

void dump_band_metadata(const IAS_L1R_BAND_METADATA *bmd)
{
    printf("band number: %d\n", bmd->band_number);
    printf("band name: %s\n", bmd->band_name);
    printf("instrument source: %s\n", bmd->instrument_source);
    printf("reflectance to radiance coefficient: %f\n", 
            bmd->reflectance_to_radiance_coefficient);
}

Ensure test_image_open_close()
{
    L1RIO *img;
    IAS_L1R_FILE_METADATA fmd;
    int status;
    const int BAND_LIST_SIZE=10;
    int band_number_list[BAND_LIST_SIZE];
    int band_count;

    img = ias_l1r_open_image("test.h5", IAS_WRITE);
    assert_true(img != NULL);

    /* make sure the file is an L1R file */
    status = ias_l1r_is_l1r_file("test.h5");
    assert_equal(status, 1);

    /* make sure no bands are found */
    assert_equal(ias_l1r_get_band_count(img), 0);
    assert_equal(ias_l1r_is_band_present(img, 1), 0);
    status = ias_l1r_get_band_list(img, band_number_list, 
                                    BAND_LIST_SIZE, &band_count);
    assert_equal(status,SUCCESS);
    assert_equal(band_count,0);

    status = ias_l1r_close_image(img);
    assert_equal(status, SUCCESS);

    /* reopen the image and make sure it doesn't complain when trying
       to set the file metadata */
    img = ias_l1r_open_image("test.h5", IAS_UPDATE);
    assert_true(img != NULL);

    memset(&fmd, 0, sizeof(fmd));
    status = ias_l1r_set_file_metadata(img, &fmd);
    assert_equal(status, SUCCESS);

    status = ias_l1r_close_image(img);
    assert_equal(status, SUCCESS);
}

Ensure test_band_io()
{
    L1RIO *img;
    int number_of_scas = 2;
    int status;
    IAS_L1R_FILE_METADATA fmd;
    IAS_L1R_FILE_METADATA fmd_read;
    IAS_L1R_BAND_METADATA bmd;

    /* open the test image */
    img = ias_l1r_open_image("test.h5", IAS_WRITE);
    assert_true(img != NULL);

    if (img != NULL)
    {
        int lines = 20;
        int samples = 10;
        IAS_DATA_TYPE data_type = IAS_I2;
        L1R_BAND_IO *l1r_band;

        assert_equal(ias_l1r_is_band_present(img, 1), 0);

        /* open the test band */
        l1r_band = ias_l1r_open_band(img, 1, &data_type, &number_of_scas,
                                   &lines, &samples);
        assert_true(l1r_band != NULL);

        if (l1r_band != NULL)
        {

            short int data[30];
            short int data2[30];
            short int data_read[30];
            short int data_read2[30];
            const int BAND_LIST_SIZE=10;
            int band_number_list[BAND_LIST_SIZE];
            int band_count;
            int i;
            int old_log_level;

            for (i = 0; i < 30; i++)
            {
                data[i] = 30 + i;
                data2[i] = 60 + i;
            }

            assert_equal(ias_l1r_get_band_count(img), 1);
            assert_equal(ias_l1r_is_band_present(img, 1), 1);
            assert_equal(ias_l1r_is_band_present(img, 2), 0);
            status = ias_l1r_get_band_list(img, band_number_list, 
                                    BAND_LIST_SIZE, &band_count);
            assert_equal(status,SUCCESS);
            assert_equal(band_count,1);
            assert_equal(band_number_list[0],1);

            /* temporarily turn off ias log messages since we expect a 
               failure */
            old_log_level = ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
            /* test for a list size too small */
            status = ias_l1r_get_band_list(img, band_number_list, 
                                    0, &band_count);
            assert_equal(status,ERROR);
            /* restore previous log level */
            ias_log_set_output_level(old_log_level);

            /* test writing the imagery to sca 0 */
            status = ias_l1r_write_image(l1r_band, 0, 2, 3, 5, 6, data);
            assert_equal(status, SUCCESS);

            /* test writing another chunk to sca 0 */
            status = ias_l1r_write_image(l1r_band, 0, 12, 3, 5, 6, data2);
            assert_equal(status, SUCCESS);

            /* test writing the imagery to sca 1 */
            status = ias_l1r_write_image(l1r_band, 1, 4, 5, 6, 5, data2);
            assert_equal(status, SUCCESS);

            /* read the imagery from the file */
            status = ias_l1r_read_image(l1r_band, 0, 2, 3, 5, 6, data_read);
            assert_equal(status, SUCCESS);

            /* verify the data read is the same as the data written */
            assert_equal(memcmp(data, data_read, sizeof(data)), 0);

            /* read the imagery from the file */
            status = ias_l1r_read_image(l1r_band, 0, 12, 3, 5, 6, data_read);
            assert_equal(status, SUCCESS);

            /* verify the data read is the same as the data written */
            assert_equal(memcmp(data2, data_read, sizeof(data2)), 0);

            /* verify sca 1 reads correctly */
            status = ias_l1r_read_image(l1r_band, 1, 4, 5, 6, 5, data_read2);
            assert_equal(status, SUCCESS);

            /* verify the data read is the same as the data written */
            assert_equal(memcmp(data2, data_read2, sizeof(data2)), 0);

            /* make sure the full image can be written and read */
            status = ias_l1r_write_image(l1r_band, 0, lines - 1, 0, 1,
                                         samples, data);
            assert_equal(status, SUCCESS);

            status = ias_l1r_read_image(l1r_band, 0, lines - 1, 0, 1, 
                                        samples, data_read);
            assert_equal(status, SUCCESS);
            assert_equal(memcmp(data, data_read, samples * sizeof(short int)),
                         0);

            /* temporarily turn off ias log messages since we expect a 
               failure */
            ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

            /* attempt to write to an illegal location */
            /* test bad sca numbers */
            status = ias_l1r_write_image(l1r_band, -1, 0, 0, 1, 1, data);
            assert_equal(status, ERROR);
            status = ias_l1r_write_image(l1r_band, number_of_scas + 1, 0, 0, 1,
                                         1, data); 
            assert_equal(status, ERROR);

            /* test bad lines */
            status = ias_l1r_write_image(l1r_band, 0, -1, 0, 1, 1, data);
            assert_equal(status, ERROR);
            status = ias_l1r_write_image(l1r_band, 0, 1, 0, lines, 1, data);
            assert_equal(status, ERROR);

            /* test bad samples */
            status = ias_l1r_write_image(l1r_band, 0, 0, -1, 1, 1, data);
            assert_equal(status, ERROR);
            status = ias_l1r_write_image(l1r_band, 0, 0, 1, 1, samples, data);
            assert_equal(status, ERROR);

            /* attempt to read an illegal location in the file to make sure
               it gets an error */
            /* test bad scas */
            status = ias_l1r_read_image(l1r_band, -1, 0, 0, 1, 1, data_read);
            assert_equal(status, ERROR);
            status = ias_l1r_read_image(l1r_band, number_of_scas + 1, 0, 0, 1,
                                        1, data_read);
            assert_equal(status, ERROR);

            /* test bad lines */
            status = ias_l1r_read_image(l1r_band, 0, -1, 0, 1, 1, data_read);
            assert_equal(status, ERROR);
            status = ias_l1r_read_image(l1r_band, 0, 1, 0, lines, 1, data_read);
            assert_equal(status, ERROR);

            /* test bad samples */
            status = ias_l1r_read_image(l1r_band, 0, 0, -1, 1, 1, data_read);
            assert_equal(status, ERROR);
            status = ias_l1r_read_image(l1r_band, 0, 0, 1, 1, samples,
                                        data_read);
            assert_equal(status, ERROR);

            /* restore normal logging level */
            ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

            /* close the test band */
            status = ias_l1r_close_band(l1r_band);
            assert_equal(status, SUCCESS);
        }

        /* write some file metadata */
        memset(&fmd, 0, sizeof(fmd));
        strcpy(fmd.software_version, "1.2.3");
        strcpy(fmd.ingest_software_version, "1.2.3");
        strcpy(fmd.collect_type, "EARTH");
        fmd.band_average_gain_applied = 1;
        fmd.relative_gain_applied = 1;
        fmd.sca_discontinuity_correction_applied = 1;
        fmd.inoperable_detector_fill_applied = 1;
        fmd.saturated_pixel_correction_applied = 1;
        fmd.residual_striping_correction_applied = 1;
        fmd.reflectance_conversion_applied = 1;
        fmd.earth_sun_distance = 1.0856;
        strcpy(fmd.custom_rad_processing_steps_applied, "ONE, TWO");
        status = ias_l1r_set_file_metadata(img, &fmd);
        assert_equal(status, SUCCESS);

        /* write some band metadata */
        bmd.band_number = 1;
        strcpy(bmd.band_name, "RED");
        strcpy(bmd.instrument_source, "OLI Red Band");
        strcpy(bmd.discontinuity_factor_source, "CPF");
        bmd.reflectance_to_radiance_coefficient = 10.1;
        bmd.bias_temp_sensitivity_correction_applied = 1;
        bmd.bias_correction_applied = 1;
        strcpy(bmd.linearization, "Standard");
        bmd.gain_temp_sensitivity_correction_applied = 1;
        status = ias_l1r_set_band_metadata(img, &bmd, 1);
        assert_equal(status, SUCCESS);

        /* read the file metadata and make sure it is identical to the 
           written data */
        status = ias_l1r_get_file_metadata(img, &fmd_read);
        assert_equal(status,SUCCESS);
        assert_equal(memcmp(&fmd, &fmd_read, sizeof(fmd)), 0);

        /* close the test image */
        status = ias_l1r_close_image(img);
        assert_equal(status, SUCCESS);
    }
}

Ensure test_band_read()
{
    L1RIO *img;
    L1R_BAND_IO *l1r_band;
    int number_of_scas = 0;
    int status;
    IAS_L1R_BAND_METADATA bmd;
    IAS_L1R_FILE_METADATA fmd;

    /* open the test image */
    img = ias_l1r_open_image("test.h5", IAS_READ);
    assert_true(img != NULL);

    if (img != NULL)
    {
        int lines;
        int samples;
        IAS_DATA_TYPE data_type;
        const int BAND_LIST_SIZE=10;
        int band_number_list[BAND_LIST_SIZE];
        int band_count;
        int old_log_level;

        assert_equal(ias_l1r_get_band_count(img), 1);
        status = ias_l1r_get_band_list(img, band_number_list, 
                                    BAND_LIST_SIZE, &band_count);
        assert_equal(status,SUCCESS);
        assert_equal(band_count,1);
        assert_equal(band_number_list[0],1);

        /* open the band */
        l1r_band = ias_l1r_open_band(img, 1, &data_type, &number_of_scas,
                                   &lines, &samples);
        assert_true(l1r_band != NULL);

        if (l1r_band != NULL)
        {
            short int data[1] = {1};
            short int data_read[30];
            int index;
            int old_log_level;

            assert_equal(data_type, IAS_I2);
            assert_equal(number_of_scas, 2);
            assert_equal(lines, 20);
            assert_equal(samples, 10);

            /* temporarily turn off ias log messages since we expect a 
               failure */
            old_log_level = ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
            /* verify that trying to write to the band fails */
            status = ias_l1r_write_image(l1r_band, 0, 1, 1, 1, 1, 
                                         data);
            assert_equal(status, ERROR);
            /* restore previous log level */
            ias_log_set_output_level(old_log_level);

            /* make sure reading works as expected since it takes a slightly
               different path in the library than the test_band_io test */
            status = ias_l1r_read_image(l1r_band, 0, 2, 3, 5, 6, data_read);
            assert_equal(status, SUCCESS);
            for (index = 0; index < sizeof(data_read)/sizeof(data_read[0]);
                 index++)
            {
                assert_equal(data_read[index], 30 + index);
            }

            /* close the band */
            status = ias_l1r_close_band(l1r_band);
            assert_equal(status, SUCCESS);
        }

        /* get the metadata and make sure it is set to the previously written
           value */
        status = ias_l1r_get_band_metadata(img, 1, &bmd);
        assert_equal(status, SUCCESS);

        /* verify the values read are correct */
        assert_equal(bmd.band_number, 1);
        assert_string_equal(bmd.band_name, "RED");
        assert_string_equal(bmd.instrument_source, "OLI Red Band");
        assert_string_equal(bmd.discontinuity_factor_source, "CPF");
        assert_double_equal(bmd.reflectance_to_radiance_coefficient, 10.1);

        /* temporarily turn off ias log messages since we expect a 
           failure */
        old_log_level = ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
        /* attempt to write the metadata back to the file which should fail */
        status = ias_l1r_set_band_metadata(img, &bmd, 1);
        assert_equal(status, ERROR);

        /* attempt to write the file metadata which should fail */
        status = ias_l1r_set_file_metadata(img, &fmd);
        assert_equal(status, ERROR);
        /* restore previous log level */
        ias_log_set_output_level(old_log_level);

        /* close the test image */
        status = ias_l1r_close_image(img);
        assert_equal(status, SUCCESS);
    }
}

Ensure test_band_update()
{
    L1RIO *img;
    L1R_BAND_IO *l1r_band;
    int status;

    /* open the test image */
    img = ias_l1r_open_image("test.h5", IAS_UPDATE);
    assert_true(img != NULL);

    if (img)
    {
        int number_of_scas;
        int lines;
        int samples;
        IAS_DATA_TYPE data_type;
        IAS_L1R_BAND_METADATA bmd[2];
        IAS_L1R_BAND_METADATA read_bmd;
        IAS_L1R_FILE_METADATA fmd;

        /* open the band */
        l1r_band = ias_l1r_open_band(img, 1, &data_type, &number_of_scas,
                                   &lines, &samples);
        assert_true(l1r_band != NULL);

        if (l1r_band != NULL)
        {
            short int data[6] = {1, 2, 3, 4, 5, 6};
            short int data_read[6];
            const int BAND_LIST_SIZE=10;
            int band_number_list[BAND_LIST_SIZE];
            int band_count;

            assert_equal(ias_l1r_get_band_count(img), 1);
            status = ias_l1r_get_band_list(img, band_number_list, 
                                    BAND_LIST_SIZE, &band_count);
            assert_equal(status,SUCCESS);
            assert_equal(band_count,1);
            assert_equal(band_number_list[0],1);
            assert_equal(data_type, IAS_I2);
            assert_equal(number_of_scas, 2);
            assert_equal(lines, 20);
            assert_equal(samples, 10);

            /* now write an update to the open band and verify it
               is written successfully */
            status = ias_l1r_write_image(l1r_band, 0, 0, 1, 2, 3, 
                                         data);
            assert_equal(status, SUCCESS);

            /* read the imagery from the file */
            status = ias_l1r_read_image(l1r_band, 0, 0, 1, 2, 3, 
                                        data_read);
            assert_equal(status, SUCCESS);

            /* verify the data read is the same as the data written */
            assert_equal(memcmp(data, data_read, sizeof(data)), 0);

            /* close the band */
            status = ias_l1r_close_band(l1r_band);
            assert_equal(status, SUCCESS);
        }

        /* create a new band in the file and verify it can be written with
           the float data type */
        number_of_scas = 1;
        lines = 40;
        samples = 20;
        data_type = IAS_R4;
        l1r_band = ias_l1r_open_band(img, 2, &data_type, &number_of_scas,
                                   &lines, &samples);
        assert_true(l1r_band != NULL);

        if (img)
        {
            float data[8] = {1.0, 1.2, 1.4, 1.6, 2.2, 2.4, 2.6, 2.8};
            float data_read[8];
            const int BAND_LIST_SIZE=10;
            int band_number_list[BAND_LIST_SIZE];
            int band_count;

            assert_equal(ias_l1r_get_band_count(img), 2);
            assert_equal(ias_l1r_is_band_present(img, 2), 1);
            assert_equal(ias_l1r_is_band_present(img, 20), 0);
            status = ias_l1r_get_band_list(img, band_number_list, 
                                    BAND_LIST_SIZE, &band_count);
            assert_equal(status,SUCCESS);
            assert_equal(band_count,2);
            assert_equal(band_number_list[0],1);
            assert_equal(band_number_list[1],2);

            /* write data to the band, read it, and make sure it is the same */
            status = ias_l1r_write_image(l1r_band, 0, 1, 0, 2, 4, 
                                         data);
            assert_equal(status, SUCCESS);
            status = ias_l1r_read_image(l1r_band, 0, 1, 0, 2, 4, 
                                        data_read);
            assert_equal(status, SUCCESS);
            assert_equal(memcmp(data, data_read, sizeof(data)), 0);

            /* close the band */
            status = ias_l1r_close_band(l1r_band);
            assert_equal(status, SUCCESS);
        }

        /* add some band metadata for the blue band to the file */
        memset(bmd, 0, sizeof(bmd));
        bmd[0].band_number = 2;
        strcpy(bmd[0].band_name, "BLUE");
        strcpy(bmd[0].instrument_source, "OLI Blue Band");
        strcpy(bmd[0].discontinuity_factor_source, "SCA_OVERLAP_CHAR");
        bmd[0].reflectance_to_radiance_coefficient = 12.1;
        bmd[0].gain_temp_sensitivity_correction_applied = 1;
        bmd[0].bias_correction_applied = 1;
        memcpy(&bmd[1], &bmd[0], sizeof(bmd[1]));
        bmd[1].band_number = 3;
        status = ias_l1r_set_band_metadata(img, bmd, 2);
        assert_equal(status, SUCCESS);

        /* read the file metadata, change a field, then write it back to make
           sure it can be updated */
        status = ias_l1r_get_file_metadata(img, &fmd);
        assert_equal(status, SUCCESS);
        assert_double_equal(fmd.earth_sun_distance, 1.0856);
        fmd.earth_sun_distance = 1.01533;
        status = ias_l1r_set_file_metadata(img, &fmd);
        assert_equal(status, SUCCESS);
        memset(&fmd, 0, sizeof(fmd));
        status = ias_l1r_get_file_metadata(img, &fmd);
        assert_equal(status, SUCCESS);
        assert_double_equal(fmd.earth_sun_distance, 1.01533);

        /* read back the band metadata and verify it is correct */
        status = ias_l1r_get_band_metadata(img, 2, &read_bmd);
        assert_equal(status, SUCCESS);
        assert_true(memcmp(&bmd[0], &read_bmd, sizeof(read_bmd)) == 0);
        status = ias_l1r_get_band_metadata(img, 3, &read_bmd);
        assert_equal(status, SUCCESS);

        /* reset the invalid values to zero */
        assert_true(memcmp(&bmd[1], &read_bmd, sizeof(read_bmd)) == 0);

        /* close the test image */
        status = ias_l1r_close_image(img);
        assert_equal(status, SUCCESS);

        img = NULL;

        /* reopen the test image to make sure the right band metadata record
           is on disk.  This check verifies that existing records on disk are
           updated when they should be instead of writing new ones */
        img = ias_l1r_open_image("test.h5", IAS_UPDATE);
        assert_true(img != NULL);

        status = ias_l1r_get_band_metadata(img, 2, &read_bmd);
        assert_equal(status, SUCCESS);

        ias_l1r_close_image(img);
        assert_equal(status, SUCCESS);
    }
}

/* test multi-sca functionality */
Ensure test_band_update_multi_sca()
{
    L1RIO *img;
    L1R_BAND_IO *l1r_band;
    int status;
    int i, j;
    int number_of_scas = 14;

    /* open the test image */
    img = ias_l1r_open_image("test2.h5", IAS_WRITE);
    assert_true(img != NULL);

    if (img != NULL)
    {
        int lines = 20;
        int samples = 10;
        IAS_DATA_TYPE data_type = IAS_I2;
        const int BAND_LIST_SIZE=10;
        int band_number_list[BAND_LIST_SIZE];
        int band_count;

        assert_equal(ias_l1r_is_band_present(img, 1), 0);

        /* open the band */
        l1r_band = ias_l1r_open_band(img, 1, &data_type, &number_of_scas,
                                   &lines, &samples);
        assert_true(l1r_band != NULL);

        if (l1r_band != NULL)
        {

            short int data_read[14][6];
            short int data[14][6];

            for (i = 0; i < 14; i++)
                for (j = 0; j < 6; j++)
                    data[i][j] = j + i;

            assert_equal(ias_l1r_get_band_count(img), 1);
            status = ias_l1r_get_band_list(img, band_number_list, 
                                    BAND_LIST_SIZE, &band_count);
            assert_equal(status,SUCCESS);
            assert_equal(band_count,1);
            assert_equal(band_number_list[0],1);
            assert_equal(data_type, IAS_I2);
            assert_equal(number_of_scas, 14);
            assert_equal(lines, 20);
            assert_equal(samples, 10);

            for (i = 0; i < 14; i++)
            {
                /* now write an update to the open band and verify it
                   is written successfully */
                status = ias_l1r_write_image(l1r_band, i, 0, 1, 2, 3, 
                                             data[i]);
                assert_equal(status, SUCCESS);

                /* read the imagery from the file */
                status = ias_l1r_read_image(l1r_band, i, 0, 1, 2, 3, 
                                            data_read[i]);
                assert_equal(status, SUCCESS);

                /* verify the data read is the same as the data written */
                assert_equal(memcmp(data[i], data_read[i], sizeof(data[i])), 0);
            }

            /* close the band */
            status = ias_l1r_close_band(l1r_band);
            assert_equal(status, SUCCESS);
        }
    }
    /* close the test image */
    status = ias_l1r_close_image(img);
    assert_equal(status, SUCCESS);
}

Ensure test_open_multiple_bands()
{
    L1RIO *img;
    L1R_BAND_IO *l1r_band;
    L1R_BAND_IO *l1r_band_2;
    int status;

    /* open the test image */
    img = ias_l1r_open_image("test.h5", IAS_UPDATE);
    assert_true(img != NULL);

    if (img)
    {
        int number_of_scas;
        int lines;
        int samples;
        IAS_DATA_TYPE data_type;
        const int BAND_LIST_SIZE=10;
        int band_number_list[BAND_LIST_SIZE];
        int band_count;

        /* open the band */
        l1r_band = ias_l1r_open_band(img, 1, &data_type, &number_of_scas,
                                   &lines, &samples);
        assert_true(l1r_band != NULL);

        if (l1r_band != NULL)
        {
            short int data[6] = {1, 2, 3, 4, 5, 6};
            short int data_read[6];

            assert_equal(ias_l1r_get_band_count(img), 2);
            status = ias_l1r_get_band_list(img, band_number_list, 
                                    BAND_LIST_SIZE, &band_count);
            assert_equal(status,SUCCESS);
            assert_equal(band_count,2);
            assert_equal(band_number_list[0],1);
            assert_equal(band_number_list[1],2);
            assert_equal(data_type, IAS_I2);
            assert_equal(number_of_scas, 2);
            assert_equal(lines, 20);
            assert_equal(samples, 10);

            /* now write an update to the open band and verify it
               is written successfully */
            status = ias_l1r_write_image(l1r_band, 0, 0, 1, 2, 3, 
                                         data);
            assert_equal(status, SUCCESS);

            /* read the imagery from the file */
            status = ias_l1r_read_image(l1r_band, 0, 0, 1, 2, 3, 
                                        data_read);
            assert_equal(status, SUCCESS);

            /* verify the data read is the same as the data written */
            assert_equal(memcmp(data, data_read, sizeof(data)), 0);
        }

        /* create a new band in the file and verify it can be written with
           the float data type */
        number_of_scas = 1;
        lines = 40;
        samples = 20;
        data_type = IAS_R4;
        l1r_band_2 = ias_l1r_open_band(img, 2, &data_type, &number_of_scas,
                                   &lines, &samples);
        assert_true(l1r_band_2 != NULL);

        if (l1r_band_2 != NULL)
        {
            float data[8] = {1.0, 1.2, 1.4, 1.6, 2.2, 2.4, 2.6, 2.8};
            float data_read[8];

            assert_equal(ias_l1r_get_band_count(img), 2);
            assert_equal(ias_l1r_is_band_present(img, 2), 1);
            assert_equal(ias_l1r_is_band_present(img, 20), 0);
            status = ias_l1r_get_band_list(img, band_number_list, 
                                    BAND_LIST_SIZE, &band_count);
            assert_equal(status,SUCCESS);
            assert_equal(band_count,2);
            assert_equal(band_number_list[0],1);
            assert_equal(band_number_list[1],2);

            /* write data to the band, read it, and make sure it is the same */
            status = ias_l1r_write_image(l1r_band_2, 0, 1, 0, 2, 4, 
                                         data);
            assert_equal(status, SUCCESS);

            status = ias_l1r_read_image(l1r_band_2, 0, 1, 0, 2, 4, 
                                        data_read);
            assert_equal(status, SUCCESS);

            assert_equal(memcmp(data, data_read, sizeof(data)), 0);

            /* close the 2nd band */
            status = ias_l1r_close_band(l1r_band_2);
            assert_equal(status, SUCCESS);
        }

        /* close the 1st band */
        /* NOTE: disabling this close to allow testing the code that will
           automatically close open bands when close_image is called
        status = ias_l1r_close_band(l1r_band);
        assert_equal(status, SUCCESS); */

        /* close the test image */
        status = ias_l1r_close_image(img);
        assert_equal(status, SUCCESS);
    }
}

Ensure test_close_image_with_band_open()
{
    L1RIO *img;
    int status;
    L1R_BAND_IO *l1r_band;
    IAS_DATA_TYPE data_type = IAS_I2;
    int lines = 20;
    int samples = 10;
    int number_of_scas = 1;

    /* open the image */
    img = ias_l1r_open_image("test.h5", IAS_READ);
    assert_true(img != NULL);

    /* open the test band */
    l1r_band = ias_l1r_open_band(img, 1, &data_type, &number_of_scas,
                                 &lines, &samples);
    assert_true(l1r_band != NULL);

    /* close the image with band still open */
    status = ias_l1r_close_image(img);
    assert_equal(status, SUCCESS);
}

/* the following test was created to help verify that multiple calls to
   set the band metadata results in a correct output file */
Ensure test_multiple_set_band_metadata_calls()
{
    L1RIO *img;
    IAS_L1R_BAND_METADATA bmd;
    IAS_L1R_BAND_METADATA read_bmd;
    int status;
    const int BAND_LIST_SIZE=10;
    int band_number_list[BAND_LIST_SIZE];
    int band_count;

    img = ias_l1r_open_image("test_bmd.h5", IAS_WRITE);
    assert_true(img != NULL);

    /* make sure no bands are found */
    assert_equal(ias_l1r_get_band_count(img), 0);
    status = ias_l1r_get_band_list(img, band_number_list, 
                                    BAND_LIST_SIZE, &band_count);
    assert_equal(status,SUCCESS);
    assert_equal(band_count,0);
    assert_equal(ias_l1r_is_band_present(img, 1), 0);

    memset(&bmd, 0, sizeof(bmd));
    bmd.band_number = 1;
    status = ias_l1r_set_band_metadata(img, &bmd, 1);
    assert_equal(status, SUCCESS);

    bmd.band_number = 2;
    status = ias_l1r_set_band_metadata(img, &bmd, 1);
    assert_equal(status, SUCCESS);

    status = ias_l1r_close_image(img);
    assert_equal(status, SUCCESS);

    /* open the image */
    img = ias_l1r_open_image("test_bmd.h5", IAS_READ);
    assert_true(img != NULL);

    /* make sure the reading doesn't detect any errors */
    status = ias_l1r_get_band_metadata(img, 1, &read_bmd);
    assert_equal(status, SUCCESS);

    status = ias_l1r_close_image(img);
    assert_equal(status, SUCCESS);
}

Ensure test_read_v1_file_metadata()
{
    char v1_image_filename[PATH_MAX] = "test_v1.h5";
    L1RIO *l1r_file;
    IAS_L1R_FILE_METADATA fmd;
    IAS_L1R_FILE_METADATA expected_fmd;
    int status;


    printf("Checking read of version 1 file metadata...\n");

    /* Initialize */
    memset(&fmd, 0, sizeof(IAS_L1R_FILE_METADATA));
    memset(&expected_fmd, 0, sizeof(IAS_L1R_FILE_METADATA));

    strcpy(expected_fmd.software_version, "1.2.3");
    strcpy(expected_fmd.ingest_software_version, "1.2.3");
    strcpy(expected_fmd.collect_type, "EARTH");
    expected_fmd.band_average_gain_applied = 1;
    expected_fmd.relative_gain_applied = 1;
    expected_fmd.sca_discontinuity_correction_applied = 1;
    expected_fmd.inoperable_detector_fill_applied = 1;
    expected_fmd.saturated_pixel_correction_applied = 1;
    expected_fmd.residual_striping_correction_applied = 1;
    expected_fmd.reflectance_conversion_applied = 1;
    expected_fmd.earth_sun_distance = 1.0856;
    strcpy(expected_fmd.custom_rad_processing_steps_applied, "ONE, TWO");

    /* open the test image for reading */
    l1r_file = ias_l1r_open_image(v1_image_filename, IAS_READ);
    assert_not_equal(NULL, l1r_file);

    /* read the file metadata */
    status = ias_l1r_get_file_metadata(l1r_file, &fmd);
    assert_equal_with_message(status, SUCCESS, "Error reading version 1 "
        "file metadata");
    if (status != SUCCESS)
    {
        ias_l1r_close_image(l1r_file);
        return;
    }

    /* Compare the read values to the expected values */
    assert_string_equal(fmd.software_version, expected_fmd.software_version);
    assert_string_equal(fmd.ingest_software_version,
        expected_fmd.ingest_software_version);
    assert_string_equal(fmd.collect_type, expected_fmd.collect_type);
    assert_equal(fmd.band_average_gain_applied,
        expected_fmd.band_average_gain_applied);
    assert_equal(fmd.relative_gain_applied,
        expected_fmd.relative_gain_applied);
    assert_equal(fmd.sca_discontinuity_correction_applied,
        expected_fmd.sca_discontinuity_correction_applied);
    assert_equal(fmd.inoperable_detector_fill_applied,
        expected_fmd.inoperable_detector_fill_applied);
    assert_equal(fmd.saturated_pixel_correction_applied,
        expected_fmd.saturated_pixel_correction_applied);
    assert_equal(fmd.residual_striping_correction_applied,
        expected_fmd.residual_striping_correction_applied);
    assert_equal(fmd.reflectance_conversion_applied,
        expected_fmd.reflectance_conversion_applied);
    assert_double_equal(fmd.earth_sun_distance,
        expected_fmd.earth_sun_distance);
    assert_string_equal(fmd.custom_rad_processing_steps_applied,
        expected_fmd.custom_rad_processing_steps_applied);

    /* Clean up */
    status = ias_l1r_close_image(l1r_file);
    assert_equal(status, SUCCESS);
}

Ensure test_read_v1_band_metadata()
{
    char v1_image_filename[PATH_MAX] = "test_v1.h5";
    L1RIO *l1r_file;
    IAS_L1R_BAND_METADATA bmd;
    IAS_L1R_BAND_METADATA expected_b1_bmd;
    IAS_L1R_BAND_METADATA expected_b2_bmd;
    IAS_L1R_BAND_METADATA expected_b3_bmd;
    int status;

    /* Initialize */
    memset(&bmd, 0, sizeof(IAS_L1R_BAND_METADATA));
    memset(&expected_b1_bmd, 0, sizeof(IAS_L1R_BAND_METADATA));
    memset(&expected_b2_bmd, 0, sizeof(IAS_L1R_BAND_METADATA));
    memset(&expected_b3_bmd, 0, sizeof(IAS_L1R_BAND_METADATA));

    expected_b1_bmd.band_number = 1;
    strcpy(expected_b1_bmd.band_name, "RED");
    strcpy(expected_b1_bmd.instrument_source, "OLI Red Band");
    expected_b1_bmd.reflectance_to_radiance_coefficient = 10.1;
    expected_b1_bmd.bias_temp_sensitivity_correction_applied = 0;
    expected_b1_bmd.gain_temp_sensitivity_correction_applied = 0;
    strcpy(expected_b1_bmd.linearization, "");

    expected_b2_bmd.band_number = 2;
    strcpy(expected_b2_bmd.band_name, "BLUE");
    strcpy(expected_b2_bmd.instrument_source, "OLI Blue Band");
    expected_b2_bmd.reflectance_to_radiance_coefficient = 12.1;

    expected_b3_bmd.band_number = 3;
    strcpy(expected_b3_bmd.band_name, "BLUE");
    strcpy(expected_b3_bmd.instrument_source, "OLI Blue Band");
    expected_b3_bmd.reflectance_to_radiance_coefficient = 12.1;

    /* Open the image for reading */
    l1r_file = ias_l1r_open_image(v1_image_filename, IAS_READ);
    assert_not_equal(NULL, l1r_file);

    /* Get the band 1 metadata and compare to the expected metadata */
    status = ias_l1r_get_band_metadata(l1r_file, 1, &bmd);
    assert_equal_with_message(status, SUCCESS, "Error getting band 1 "
       "metadata");
    if (status != SUCCESS)
    {
        ias_l1r_close_image(l1r_file);
        return;
    }

    assert_equal(bmd.band_number, expected_b1_bmd.band_number);
    assert_string_equal(bmd.band_name, expected_b1_bmd.band_name);
    assert_string_equal(bmd.instrument_source,
        expected_b1_bmd.instrument_source);
    assert_double_equal(bmd.reflectance_to_radiance_coefficient,
        expected_b1_bmd.reflectance_to_radiance_coefficient);
    assert_equal(bmd.bias_temp_sensitivity_correction_applied,
        expected_b1_bmd.bias_temp_sensitivity_correction_applied);
    assert_equal(bmd.gain_temp_sensitivity_correction_applied,
        expected_b1_bmd.gain_temp_sensitivity_correction_applied);
    assert_string_equal(bmd.linearization, expected_b1_bmd.linearization);

    /* Get the band 2 metadata and compare to the expected metadata */
    status = ias_l1r_get_band_metadata(l1r_file, 2, &bmd);
    assert_equal_with_message(status, SUCCESS, "Error getting band 2 "
       "metadata");
    if (status != SUCCESS)
    {
        ias_l1r_close_image(l1r_file);
        return;
    }

    assert_equal(bmd.band_number, expected_b2_bmd.band_number);
    assert_string_equal(bmd.band_name, expected_b2_bmd.band_name);
    assert_string_equal(bmd.instrument_source,
        expected_b2_bmd.instrument_source);
    assert_double_equal(bmd.reflectance_to_radiance_coefficient,
        expected_b2_bmd.reflectance_to_radiance_coefficient);

    /* Get the band 3 metadata and compare to the expected metadata.
       Since it's identical to the band 2 metadata, clear the test
       metadata structure before getting the metadata */
    memset(&bmd, 0, sizeof(IAS_L1R_BAND_METADATA));
    status = ias_l1r_get_band_metadata(l1r_file, 3, &bmd);
    assert_equal_with_message(status, SUCCESS, "Error getting band 3 "
       "metadata");
    if (status != SUCCESS)
    {
        ias_l1r_close_image(l1r_file);
        return;
    }

    assert_equal(bmd.band_number, expected_b3_bmd.band_number);
    assert_string_equal(bmd.band_name, expected_b3_bmd.band_name);
    assert_string_equal(bmd.instrument_source,
        expected_b3_bmd.instrument_source);
    assert_double_equal(bmd.reflectance_to_radiance_coefficient,
        expected_b3_bmd.reflectance_to_radiance_coefficient);

    /* Cleanup */
    status = ias_l1r_close_image(l1r_file);
    assert_equal(status, SUCCESS);
}

int main(int argc, char **argv)
{
    TestSuite *suite = create_test_suite();
    add_test(suite, test_image_open_close);
    add_test(suite, test_band_io);
    add_test(suite, test_band_read);
    add_test(suite, test_band_update);
    add_test(suite, test_band_update_multi_sca);
    add_test(suite, test_open_multiple_bands);
    add_test(suite, test_close_image_with_band_open);
    add_test(suite, test_multiple_set_band_metadata_calls);
    add_test(suite, test_read_v1_file_metadata);
    add_test(suite, test_read_v1_band_metadata);
    if (argc > 1) 
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}
