/************************************************************************

NAME:     READWRITE_TESTS

PURPOSE:  Use CGreen to perform unit testing on IAS Ancillary Data
          Processor (ADP) IO Library to write attitude and ephemeris
          data records into tables contained in the ancillary data file.

*************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ias_const.h"
#include "cgreen.h"

#include "../ias_ancillary_io.h"
#include "../ias_ancillary_io_private.h"


/*************************************************************************
           WRITE TEST
**************************************************************************/
Ensure test_ias_ancillary_write()
{
    /* Simulated attitude and ephemeris data.   */
    IAS_ANC_ATTITUDE_DATA *attitude_data = NULL;
    IAS_ANC_EPHEMERIS_DATA *ephemeris_data = NULL;

    /* Attitude and ephemeris test data */
    double attitude_epoch_time[3] = {1970., 1., 10.};
    double attitude_seconds_from_epoch[2] = {1000000, 1000400};
    double test_eci[2][4] = {{1.000, 2.000, 3.000, 4.000},
                             {2.000, 4.000, 6.000, 8.000}};
    double test_ecef[2][4] = {{5.000, 6.000, 7.000, 8.000},
                             {10.000, 12.000, 14.000, 16.000}};
    double roll[2] = {1.000, -1.000};
    double roll_rate[2] = {-0.555, -0.557};
    double pitch[2] = {2.000, -0.500};
    double pitch_rate[2] = {0.001, -0.001};
    double yaw[2] = {6.000, 6.005};
    double yaw_rate[2] = {-0.044, 0.044};
    double ephemeris_epoch_time[3] = {1980., 6., 10.};
    double ephemeris_seconds_from_epoch[2] = {1010400, 1010455};
    double test_eci_position[2][3] = {{1.000, 2.000, 3.000},
                                      {2.000, 4.000, 6.000}};
    double test_eci_velocity[2][3] = {{-0.5, -0.6, -0.7},
                                      {-0.25, -0.3, -0.35}};
    double test_ecef_position[2][3] = {{1.000, 2.000, 3.000},
                                       {3.000, -1.000, -2.000}};
    double test_ecef_velocity[2][3] = {{10.0, -15.0, 5.0},
                                      {5.0, 10.0, -15.0}};

    /* "Shared" variables */
    int num_records = 2;
    int status = SUCCESS;
    int rec = 0;               /* Record loop counter */
    int field = 0;             /* Field loop counter */


    printf("Testing write to ancillary data file...\n");

    /* Allocate and populate attitude and ephemeris data structures. */
    attitude_data = ias_ancillary_allocate_attitude(num_records);
    ephemeris_data = ias_ancillary_allocate_ephemeris(num_records);
    assert_not_equal_with_message(attitude_data, NULL, "Cannot allocate "
        "attitude data buffer");
    assert_not_equal_with_message(ephemeris_data, NULL, "Cannot allocate "
        "ephemeris data buffer");
    if ((attitude_data == NULL) || (ephemeris_data == NULL))
    {
        ias_ancillary_free_attitude(attitude_data);
        ias_ancillary_free_ephemeris(ephemeris_data);
        return;
    }
    else
    {
        /* Copy the test data into the data buffers */
        memcpy(attitude_data->utc_epoch_time, attitude_epoch_time,
               sizeof(attitude_epoch_time));
        memcpy(ephemeris_data->utc_epoch_time, ephemeris_epoch_time,
               sizeof(ephemeris_epoch_time));
        for (rec = 0; rec < num_records; rec++)
        {
            attitude_data->records[rec].seconds_from_epoch =
               attitude_seconds_from_epoch[rec];
            memcpy(attitude_data->records[rec].eci_quaternion,
                   &(test_eci[rec][0]), 4*sizeof(double));
            memcpy(attitude_data->records[rec].ecef_quaternion,
                   &(test_ecef[rec][0]), 4*sizeof(double));
            attitude_data->records[rec].roll = roll[rec];
            attitude_data->records[rec].roll_rate = roll_rate[rec];
            attitude_data->records[rec].pitch = pitch[rec];
            attitude_data->records[rec].pitch_rate = pitch_rate[rec];
            attitude_data->records[rec].yaw = yaw[rec];
            attitude_data->records[rec].yaw_rate = yaw_rate[rec];

            ephemeris_data->records[rec].seconds_from_epoch = 
               ephemeris_seconds_from_epoch[rec];
            memcpy(ephemeris_data->records[rec].eci_position,
                   &(test_eci_position[rec][0]), 3*sizeof(double));
            memcpy(ephemeris_data->records[rec].eci_velocity,
                   &(test_eci_velocity[rec][0]), 3*sizeof(double));
            memcpy(ephemeris_data->records[rec].ecef_position,
                   &(test_ecef_position[rec][0]), 3*sizeof(double));
            memcpy(ephemeris_data->records[rec].ecef_velocity,
                   &(test_ecef_velocity[rec][0]), 3*sizeof(double));
        }

        /* DEBUGGING AID -- verify the memcpy's and assignments
           wrote the right data to the right structure members */
        for (field = 0; field < 3; field++)
        {
            assert_double_equal(attitude_data->utc_epoch_time[field],
               attitude_epoch_time[field]);
            assert_double_equal(ephemeris_data->utc_epoch_time[field],
               ephemeris_epoch_time[field]);
        }
        for (rec = 0; rec < num_records; rec++)
        {
            assert_double_equal(
                attitude_data->records[rec].seconds_from_epoch,
                attitude_seconds_from_epoch[rec]);
            assert_double_equal(
                ephemeris_data->records[rec].seconds_from_epoch,
                ephemeris_seconds_from_epoch[rec]);

            assert_double_equal(attitude_data->records[rec].roll,
                roll[rec]);
            assert_double_equal(attitude_data->records[rec].roll_rate,
                roll_rate[rec]);
            assert_double_equal(attitude_data->records[rec].pitch,
                pitch[rec]);
            assert_double_equal(attitude_data->records[rec].pitch_rate,
                pitch_rate[rec]);
            assert_double_equal(attitude_data->records[rec].yaw, yaw[rec]);
            assert_double_equal(attitude_data->records[rec].yaw_rate,
                yaw_rate[rec]);

            for (field = 0; field < 4; field++)
            {
                assert_double_equal(
                    attitude_data->records[rec].eci_quaternion[field],
                    test_eci[rec][field]);
                assert_double_equal(
                    attitude_data->records[rec].ecef_quaternion[field],
                    test_ecef[rec][field]);
            }

            for (field = 0; field < 3; field++)
            {
                assert_double_equal(
                    ephemeris_data->records[rec].eci_position[field],
                    test_eci_position[rec][field]);
                assert_double_equal(
                    ephemeris_data->records[rec].eci_velocity[field],
                    test_eci_velocity[rec][field]);
                assert_double_equal(
                    ephemeris_data->records[rec].ecef_position[field],
                    test_ecef_position[rec][field]);
                assert_double_equal(
                    ephemeris_data->records[rec].ecef_velocity[field],
                    test_ecef_velocity[rec][field]);
            }
        }

        /* Write the simulated data to the file. */
        status = ias_ancillary_write("TestAncillaryDataFile.h5",
                                     attitude_data, ephemeris_data);
        assert_equal_with_message(status, SUCCESS, "Write to test "
            "ancillary data file failed");

        /* Clean up */
        ias_ancillary_free_attitude(attitude_data);
        ias_ancillary_free_ephemeris(ephemeris_data);
    }
}



/************************************************************************
             READ TEST
*************************************************************************/
Ensure test_ias_ancillary_read()
{

    /* Output pointers with data read from the test ancillary file. */
    IAS_ANC_ATTITUDE_DATA *attitude_read = NULL;
    IAS_ANC_EPHEMERIS_DATA *ephemeris_read = NULL;

    double attitude_epoch_time[3] = {1970., 1., 10.};
    double attitude_seconds_from_epoch[2] = {1000000, 1000400};
    double test_eci[2][4] = {{1.000, 2.000, 3.000, 4.000},
                             {2.000, 4.000, 6.000, 8.000}};
    double test_ecef[2][4] = {{5.000, 6.000, 7.000, 8.000},
                             {10.000, 12.000, 14.000, 16.000}};
    double roll[2] = {1.000, -1.000};
    double roll_rate[2] = {-0.555, -0.557};
    double pitch[2] = {2.000, -0.500};
    double pitch_rate[2] = {0.001, -0.001};
    double yaw[2] = {6.000, 6.005};
    double yaw_rate[2] = {-0.044, 0.044};
    double ephemeris_epoch_time[3] = {1980., 6., 10.};
    double ephemeris_seconds_from_epoch[2] = {1010400, 1010455};
    double test_eci_position[2][3] = {{1.000, 2.000, 3.000},
                                      {2.000, 4.000, 6.000}};
    double test_eci_velocity[2][3] = {{-0.5, -0.6, -0.7},
                                      {-0.25, -0.3, -0.35}};
    double test_ecef_position[2][3] = {{1.000, 2.000, 3.000},
                                       {3.000, -1.000, -2.000}};
    double test_ecef_velocity[2][3] = {{10.0, -15.0, 5.0},
                                      {5.0, 10.0, -15.0}};

    /* "Shared" variables. */
    int num_records = 2;
    int status = SUCCESS;
    int rec = 0;                  /* Record loop counter */
    int field = 0;                /* Field loop counter */


    printf("Testing read from ancillary data file...\n");

    status = ias_ancillary_read("TestAncillaryDataFile.h5",
        &attitude_read, &ephemeris_read);

    /* Check the return status */
    assert_equal_with_message(status, SUCCESS, "Cannot read "
        "TestAncillaryDataFile.h5");

    /* Check the returned pointers to make sure they were allocated
       and populated */
    assert_not_equal_with_message(attitude_read, NULL, "Attitude data "
        "not read");
    assert_not_equal_with_message(ephemeris_read, NULL, "Ephemeris data "
        "not read");

    /* If anything failed to read, quit right here */
    if ((status != SUCCESS) || (attitude_read == NULL)
            || (ephemeris_read == NULL))
    {
        ias_ancillary_free_attitude(attitude_read);
        ias_ancillary_free_ephemeris(ephemeris_read);
        return;
    }

    /* Check the number of attitude records */
    assert_equal_with_message(attitude_read->number_of_samples, num_records,
        "Read number of samples %s does not equal initial "
        "number of samples %s", attitude_read->number_of_samples,
        num_records);

    for (field = 0; field < 3; field++)
    {
        assert_double_equal_with_message(
            attitude_read->utc_epoch_time[field],
            attitude_epoch_time[field],
            "Read time field value %lf does not equal initial "
            "time field value %lf",
            attitude_read->utc_epoch_time[field],
            attitude_epoch_time[field]);
    }
    for (rec = 0; rec < num_records; rec++)
    {
        /* Check seconds_from_epoch */
        assert_double_equal_with_message(
            attitude_read->records[rec].seconds_from_epoch,
            attitude_seconds_from_epoch[rec],
            "Read seconds from epoch %lf does not match initial "
            "seconds from epoch %ls",
            attitude_read->records[rec].seconds_from_epoch,
            attitude_seconds_from_epoch[rec]);

        /*Check roll/roll rate, pitch/pitch rate, and
          yaw/yaw rate */
        assert_double_equal_with_message(attitude_read->records[rec].roll,
            roll[rec], "Read roll value %lf does not match initial roll "
            "value %lf", attitude_read->records[rec].roll, roll[rec]);
        assert_double_equal_with_message(
            attitude_read->records[rec].roll_rate, roll_rate[rec],
            "Read roll rate value %lf does not match initial roll rate "
            "value %lf", attitude_read->records[rec].roll_rate,
            roll_rate[rec]);

        assert_double_equal_with_message(attitude_read->records[rec].pitch,
            pitch[rec], "Read pitch value %lf does not match initial "
            "pitch value %lf",
            attitude_read->records[rec].pitch, pitch[rec]);
        assert_double_equal_with_message(
            attitude_read->records[rec].pitch_rate, pitch_rate[rec],
            "Read pitch rate value %lf does not match initial pitch rate "
            "value %lf", attitude_read->records[rec].pitch_rate,
            pitch_rate[rec]);

        assert_double_equal_with_message(attitude_read->records[rec].yaw,
            yaw[rec], "Read yaw value %lf does not match initial yaw "
            "value %lf", attitude_read->records[rec].yaw, yaw[rec]);
        assert_double_equal_with_message(
            attitude_read->records[rec].yaw_rate, yaw_rate[rec],
            "Read yaw rate value %lf does not match initial yaw rate "
            "value %lf", attitude_read->records[rec].yaw_rate,
            yaw_rate[rec]);

        /* Check the ECI and ECEF quaternions */
        for (field = 0; field < 4; field++)
        {
            assert_double_equal_with_message(
                attitude_read->records[rec].eci_quaternion[field],
                test_eci[rec][field],
                "Read ECI quaternion value %lf does not match "
                "initial ECI quaternion value %lf",
                attitude_read->records[rec].eci_quaternion[field],
                test_eci[rec][field]);

            assert_double_equal_with_message(
                attitude_read->records[rec].ecef_quaternion[field],
                test_ecef[rec][field],
                "Read ECEF quaternion value %lf does not match "
                "initial ECEF quaternion value %lf",
                attitude_read->records[rec].ecef_quaternion[field],
                test_ecef[rec][field]);
        }
    }

    /* Check the number of ephemeris records */
    assert_equal_with_message(ephemeris_read->number_of_samples, num_records,
        "Read ephemeris number of samples %d does not match initial "
        "number of samples %d", ephemeris_read->number_of_samples,
        num_records);

    /* Check the ephemeris epoch time */
    for (field = 0; field < 3; field++)
    {
        assert_double_equal_with_message(
            ephemeris_read->utc_epoch_time[field],
            ephemeris_epoch_time[field],
            "Ephemeris epoch time field value %lf does not match "
            "initial time field value %lf",
            ephemeris_read->utc_epoch_time[field],
            ephemeris_epoch_time[field]);
    }

    for (rec = 0; rec < num_records; rec++)
    {
        /* Check the seconds from epoch for each record */
        assert_double_equal_with_message(
            ephemeris_read->records[rec].seconds_from_epoch,
            ephemeris_seconds_from_epoch[rec],
            "Ephemeris seconds from epoch %lf does not match "
            "initial seconds from epoch %lf",
            ephemeris_read->records[rec].seconds_from_epoch,
            ephemeris_seconds_from_epoch[rec]);

        /* Check the ECI and ECEF position and velocity vectors */
        for (field = 0; field < 3; field++)
        {
            assert_double_equal_with_message(
                ephemeris_read->records[rec].eci_position[field],
                test_eci_position[rec][field],
                "Read ECI position component %lf does not match initial "
                "ECI position component %lf",
                ephemeris_read->records[rec].eci_position[field],
                test_eci_position[rec][field]);

            assert_double_equal_with_message(
                ephemeris_read->records[rec].eci_velocity[field],
                test_eci_velocity[rec][field],
                "Read ECI velocity component %lf does not match "
                "initial ECI velocity component %lf",
                ephemeris_read->records[rec].eci_velocity[field],
                test_eci_velocity[rec][field]);

            assert_double_equal_with_message(
                ephemeris_read->records[rec].ecef_position[field],
                test_ecef_position[rec][field],
                "Read ECEF position component %lf does not match "
                "initial ECEF position component %lf",
                ephemeris_read->records[rec].ecef_position[field],
                test_ecef_position[rec][field]);

             assert_double_equal_with_message(
                ephemeris_read->records[rec].ecef_velocity[field],
                test_ecef_velocity[rec][field],
                "Read ECEF velocity component %lf does not match "
                "initial ECEF velocity component %lf",
                ephemeris_read->records[rec].ecef_velocity[field],
                test_ecef_velocity[rec][field]);
        }
    }

    /* Clean up */
    ias_ancillary_free_attitude(attitude_read);
    ias_ancillary_free_ephemeris(ephemeris_read);
}




/************************************************************************
             MAIN "DRIVER"
*************************************************************************/
int main(int argc, char **argv)
{
    /* Create the test suite */
    TestSuite *suite = create_test_suite();
    add_test(suite, test_ias_ancillary_write);
    add_test(suite, test_ias_ancillary_read);

    /* Run the test */
    if (argc > 1)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }

    return run_test_suite(suite, create_text_reporter());
}



