/************************************************************************
 NAME:       check_bpf_insert_results

 PURPOSE:    Checks selected BPF parameter values following insertion
             into the BPF database.

 RETURNS:    Integer status code of SUCCESS or ERROR
*************************************************************************/
#include <string.h>
#include "ias_db.h"
#include "ias_db_query.h"
#include "ias_logging.h"
#include "ias_types.h"
#include "ias_const.h"
#include "ias_bpf.h"
#include "../ias_bpf_database.h"
#include "../ias_bpf_database_local.h"
#include "../../../misc/database_access/local.h"
#include "cgreen/cgreen.h"


int check_bpf_insert_results
(
   struct ias_db_connection *conn,   /* I: Open database connection */
   IAS_BPF *bpf                      /* I: BPF data structure */
)
{
    char sql[512];
    char group_name[80];
    char parameter_value[80];
    char read_sensor_name[20];
    char db_sensor_id_string[IAS_BPF_SENSOR_NAME_LENGTH];
    struct ias_db_query *query = NULL;
    struct PARAMETER_NAME_ID *parameter_ids = NULL;
    const struct IAS_BPF_FILE_ATTRIBUTES *file_attrs = NULL;
    double d_parm_value;
    IAS_SENSOR_ID sensor_id;
    int    band;
    int    bpfid;
    int    det;
    int    max_det = 0;
    int    equal_items;
    int    i;
    int    group_index;
    int    num_parameter_ids;
    int    parameter_id;
    int    p_check;
    int    sca;
    int    status;

    /* Get the sensor ID number from the BPF data. */
    sensor_id = ias_bpf_get_sensor_id(bpf);
    if (sensor_id == IAS_INVALID_SENSOR_ID)
    {
        assert_true_with_message(FALSE, "Error retrieving sensor ID ");
        return ERROR;
    }

    /* Get the file attributes. */
    file_attrs = ias_bpf_get_file_attributes(bpf);
    if (file_attrs == NULL)
    {
        assert_true_with_message(FALSE, "Error retrieving BPF file "
            "attributes");
        return ERROR;
    }

    /* Get the maximum number of detectors for the specified sensor. */
    if (sensor_id == IAS_OLI)
    {
        max_det = IAS_BPF_PAN_MAX_NDET;
        sprintf(db_sensor_id_string, "OLI");
    }
    else
    {
        max_det = IAS_BPF_THERMAL_MAX_NDET;
        sprintf(db_sensor_id_string, "TIRS");
    }

    /* Query some things back out of the BPF_MAIN table to
       ensure it stored correctly. */
    status = snprintf(sql, sizeof(sql), "select bpf_main_id, sensor_id "
        "from bpf_main where file_name='%s';", file_attrs->file_name);
    if ((status < 0) || (status >= sizeof(sql)))
    {
        assert_true_with_message(FALSE, "Error constructing SQL string");
        return ERROR;
    }

    query = ias_db_perform_query(conn, sql);
    assert_not_equal_with_message(query, NULL, "BPF_MAIN query failed to "
        "execute");
    if (query == NULL)
    {
        return ERROR;
    }

    /* ias_db_query_was_successful returns 1 if the query succeeded, 0
       otherwise. So assert that the return value is nonzero. */
    status = ias_db_query_was_successful(query);
    assert_not_equal_with_message(status, 0, "BPF_MAIN query failed");
    if (status == 0)
    {
        ias_db_query_close(query);
        return ERROR;
    }

    /* ias_db_query_next returns 1 if the first record can be accessed,
       0 otherwise.  So assert that the return value is nonzero. */
    status = ias_db_query_next(query);
    assert_not_equal_with_message(status, 0, "BPF_MAIN query failed to find "
        "first record");
    if (status == 0)
    {
        ias_db_query_close(query);
        return ERROR;
    }

    status = ias_db_query_int_value(query, &bpfid, 0);
    assert_equal_with_message(status, SUCCESS, "Error retrieving BPF ID "
        "number from database");
    ias_db_query_string_value(query, 1, read_sensor_name,
        sizeof(read_sensor_name));
    assert_string_equal_with_message(read_sensor_name, db_sensor_id_string,
        "BPF sensor name '%s' does not match sensor name '%s' read "
        "from database", db_sensor_id_string, read_sensor_name);
    ias_db_query_close(query);

    /* Get the parameter IDs for items we're checking. */
    parameter_ids = ias_bpf_db_get_parameter_ids(conn, &num_parameter_ids);
    assert_not_equal_with_message(parameter_ids, NULL, "Error retrieving "
            "parameter ID information");
    if (parameter_ids == NULL)
    {
        return ERROR;
    }

    /* Set up to get the actual parameter values and the group they're in.
       The space at the tail end of the SQL statement (following
       'a.bpf_parameter_id=?') is needed to get the query to function
       correctly.  */
    status = snprintf(sql, sizeof(sql), "select value1, value2, value3, "
        "value4, value5, value6, value7, name "
        "from bpf_parameter_values_data a, bpf_main_group b, bpf_group c, "
        "bpf_main d "
        "where a.bpf_group_id=b.bpf_group_id "
        "and b.bpf_group_id=c.bpf_group_id "
        "and d.bpf_main_id=b.bpf_main_id "
        "and b.bpf_main_id=? "
        "and a.bpf_parameter_id=? ");
    if ((status < 0) || (status >= sizeof(sql)))
    {
        assert_true_with_message(FALSE, "Error constructing SQL string");
        return ERROR;
    }

    query = ias_db_prepare_query(conn, sql);
    assert_not_equal_with_message(query, NULL, "Error executing check "
        "query");
    if (query == NULL)
    {
        return ERROR;
    }

    status = ias_db_query_was_successful(query);
    assert_not_equal_with_message(status, 0, "Check query failed ");
    if (status == 0)
    {
        ias_db_query_close(query);
        return ERROR;
    }

    /* Bind the BPF ID number and parameter mode values to the overall
       check query. */
    status = ias_db_bind_int_by_index(query, 0, IAS_DB_PARAMETER_MODE_INPUT,
        &bpfid, NULL);
    assert_equal_with_message(status, SUCCESS, "Error binding BPF ID to "
        "check query");
    if (status != SUCCESS)
    {
        ias_db_query_close(query);
        return ERROR;
    }
    status = ias_db_bind_int_by_index(query, 1, IAS_DB_PARAMETER_MODE_INPUT,
        &parameter_id, NULL);
    assert_equal_with_message(status, SUCCESS, "Error binding parameter ID "
        "to check query");
    if (status != SUCCESS)
    {
        ias_db_query_close(query);
        return ERROR;
    }

    /* Check special parameters, and every 10th one otherwise. */
    for (i = 0; i < num_parameter_ids; i++)
    {
        if (strcmp(parameter_ids[i].parameter_name, "File_Name") == 0)
        {
            parameter_id = parameter_ids[i].parameter_id;
            status = ias_db_exec_prepared_query(query);
            assert_equal_with_message(status, SUCCESS, "Query for 'File_'"
                "Name' parameter failed");
            assert_true_with_message(ias_db_query_next(query),
                "Error advancing to current parameter record");
            ias_db_query_string_value(query, 0, parameter_value, 
                sizeof(parameter_value));
            assert_string_equal(parameter_value, file_attrs->file_name); 
            status = ias_db_query_end(query);
            assert_equal_with_message(status, SUCCESS, "Error ending "
                "'File_Name' parameter query");
        }
        else if (strcmp(parameter_ids[i].parameter_name, "Orbit_Number")
            == 0)
        {
            parameter_id = parameter_ids[i].parameter_id;
            status = ias_db_exec_prepared_query(query);
            assert_equal_with_message(status, SUCCESS, "Query for 'Orbit_'"
                "Number' parameter failed");
            assert_true_with_message(ias_db_query_next(query),
                "Error advancing to current parameter record");
            ias_db_query_string_value(query, 0, parameter_value, 
                sizeof(parameter_value));
            assert_string_equal(parameter_value, "100"); 
            status = ias_db_query_end(query);
            assert_equal_with_message(status, SUCCESS, "Error ending "
                "'Orbit_Number' parameter query");
        }
        else if ((sensor_id == IAS_OLI)
            && (strcmp(parameter_ids[i].parameter_name, "A0_Coefficient")
                == 0))
        {
            parameter_id = parameter_ids[i].parameter_id;
            status = ias_db_exec_prepared_query(query);
            assert_equal_with_message(status, SUCCESS, "Query for 'A0_"
                "Coefficient' parameter failed");

            /* Loop through all the groups that have this parameter */
            for (group_index = 0; group_index < 140; group_index++)
            {
                assert_true_with_message(ias_db_query_next(query),
                    "Error advancing to next record");

                /* Extract the band & SCA number from the group
                   name, used to compute the expected value. */
                ias_db_query_string_value(query, 7, group_name, 
                    sizeof(group_name));
                sscanf(group_name, "%*[^0-9]%02d_SCA%02d", &band, &sca);

                /* Check the first value */
                ias_db_query_string_value(query, 0, parameter_value,
                    sizeof(parameter_value));
                sscanf(parameter_value, "%lf", &d_parm_value);
                equal_items = (d_parm_value == (sca * 1e-6));

                /* Check the 2nd, 3rd, 4th, 5th, 6th, and 7th values -- these
                   should all be NULL */
                ias_db_query_string_value(query, 1, parameter_value,
                    sizeof(parameter_value));
                equal_items += 
                    (strcmp(parameter_value, "NULL") == 0);

                ias_db_query_string_value(query, 2, parameter_value,
                    sizeof(parameter_value));
                equal_items += 
                    (strcmp(parameter_value, "NULL") == 0);

                ias_db_query_string_value(query, 3, parameter_value,
                    sizeof(parameter_value));
                equal_items += 
                    (strcmp(parameter_value, "NULL") == 0);

                ias_db_query_string_value(query, 4, parameter_value,
                    sizeof(parameter_value));
                equal_items += 
                    (strcmp(parameter_value, "NULL") == 0);

                ias_db_query_string_value(query, 5, parameter_value,
                    sizeof(parameter_value));
                equal_items += 
                    (strcmp(parameter_value, "NULL") == 0);

                ias_db_query_string_value(query, 6, parameter_value,
                    sizeof(parameter_value));
                equal_items += 
                    (strcmp(parameter_value, "NULL") == 0);

                /* ensure all items were equal */
                assert_equal_with_message(equal_items, 7,
                    "Checking values of %s in %s", 
                    parameter_ids[i].parameter_name, group_name); 
            }

            status = ias_db_query_end(query);
            assert_equal_with_message(status, SUCCESS, "Error ending "
                "'A0_Coefficient' parameter query");
        }
        /* Do every 10th parameter otherwise */
        else if ((i % 10 == 0) && (i < max_det))
        {
            int group_index_max;

            parameter_id = parameter_ids[i].parameter_id;
            status = ias_db_exec_prepared_query(query);
            assert_equal_with_message(status, SUCCESS, "Query for 'Dddd' "
                "parameter failed");

            /* If this is a Dxxx parameter, check it */
            status = sscanf(parameter_ids[i].parameter_name, "D%03d", &det);
            if (status != 1) 
                continue;

            /* Set this parameter depending on which sensor's BPF we're
               dealing with.  The value depends on the
               number of band/SCA combinations for the sensor; for
               TIRS, set it to 6 (2 bands x 3 SCAs per band).  For OLI, the
               maximum value is 140 (8 bands x 14 SCAs per band +
               2 x 14 SCAs/band for the pan band), but set it to 20 to keep
               the processing time "reasonable". */
            if (sensor_id == IAS_OLI)
                group_index_max = 20;
            else
                group_index_max = 6;

            /* Loop through the first few groups that have this parameter */
            for (group_index = 0; group_index < group_index_max;
                 group_index++)
            {
                assert_true_with_message(ias_db_query_next(query),
                    "Error advancing to next record");

                /* Extract the band & SCA number from the group
                   name, used to compute the expected value */
                ias_db_query_string_value(query, 7, group_name, 
                    sizeof(group_name));
                sscanf(group_name, "%*[^0-9]%02d_SCA%02d", &band, &sca);

                /* check the 1st value */
                ias_db_query_string_value(query, 0, parameter_value, 
                    sizeof(parameter_value));
                sscanf(parameter_value, "%lf", &d_parm_value);
                equal_items = 
                    ((int)d_parm_value == (band * 100000 + sca * 1000 + det));

                /* check the 2nd value */
                ias_db_query_string_value(query, 1, parameter_value, 
                    sizeof(parameter_value));
                sscanf(parameter_value, "%lf", &d_parm_value);
                equal_items += 
                    ((int)d_parm_value 
                     == (band * 100000 + sca * 1000 + det + 1));

                /* For OLI BPFs, check the 3rd and 4th parameter values.
                   FOR TIRS BPFs, check the 3rd and 4th parameter values
                   to ensure they're NULL.  */
                if (sensor_id == IAS_OLI)
                {
                    /* check the 3rd value */
                    ias_db_query_string_value(query, 2, parameter_value, 
                        sizeof(parameter_value));
                    sscanf(parameter_value, "%lf", &d_parm_value);
                    equal_items += 
                        ((int)d_parm_value 
                         == (band * 100000 + sca * 1000 + det + 2));

                    /* check the 4th value */
                    ias_db_query_string_value(query, 3, parameter_value, 
                        sizeof(parameter_value));
                    sscanf(parameter_value, "%lf", &d_parm_value);
                    equal_items += 
                        ((int)d_parm_value 
                         == (band * 100000 + sca * 1000 + det + 3));
                }
                else
                {
                    ias_db_query_string_value(query, 2, parameter_value,
                        sizeof(parameter_value));
                    equal_items += (strcmp(parameter_value, "NULL") == 0);

                    ias_db_query_string_value(query, 3, parameter_value,
                        sizeof(parameter_value));
                    equal_items += (strcmp(parameter_value, "NULL") == 0);
                }

                /* There should not be a 5th, 6th, or 7th value for
                   either an OLI or TIRS BPF, so check one at random. */
                p_check = ((i / 1000) % 3) + 4;
                ias_db_query_string_value(query, p_check, parameter_value, 
                    sizeof(parameter_value));
                equal_items += 
                    (strcmp(parameter_value, "NULL") == 0);

                /* ensure all items were equal */
                assert_equal_with_message(equal_items, 5,
                    "Checking values of %s in %s", 
                    parameter_ids[i].parameter_name, group_name); 
            }

            status = ias_db_query_end(query);
            assert_equal_with_message(status, SUCCESS, "Error ending 'Dddd' "
                "parameter query");
        }
    }

    /* Clean up. */
    ias_db_query_close(query);
    free(parameter_ids);

    return SUCCESS;

}   /* END -- check_bpf_insert_results */
