/*************************************************************************

NAME: database_test_bpf.c

PURPOSE: Using Cgreen for testing bpf database library routines 

Algorithm References: None

**************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <time.h>
#include <sys/timeb.h>

#include "ias_db.h"
#include "ias_db_get_connect_info.h"
#include "ias_db_insert.h"
#include "ias_db_query.h"
#include "ias_db_stored_proc.h"
#include "ias_logging.h"
#include "ias_types.h"
#include "ias_const.h"
#include "ias_satellite_attributes.h"
#include "ias_bpf.h"
#include "../ias_bpf_database.h"
#include "../ias_bpf_database_local.h"
#include "../../../misc/database_access/local.h"
#include "cgreen/cgreen.h"

/*
#define DEBUG 1 
*/

/* Global variables used by tests, and teardown */
IAS_BPF *good_oli_bpf = NULL; /* used to test a successful insert */
IAS_BPF *good_tirs_bpf = NULL;
IAS_BPF *bad_oli_bpf = NULL;  /* used to test various error conditions */
IAS_BPF *bad_tirs_bpf = NULL;
time_t timestamp;         /* timestamp of this run */

extern int check_bpf_insert_results
(
   struct ias_db_connection *conn,   /* I: Open database connection */
   IAS_BPF *bpf                      /* I: BPF data */
);

/* helper function used to allocate & initialize a BPF structure with 
 * default values */
void setup_bpf(int do_good, IAS_SENSOR_ID sensor_id)
{
    time_t ts;         
    struct tm start_time;
    struct tm end_time;
    char   tmpstring[256];
    char   sensor_char[2];
    double params[7];
    int    band_list[IAS_MAX_NBANDS];
    int    bandno;
    int    detno;
    int    index;
    int    maxdet;
    int    maxsca;
    int    mindet;
    int    minsca;
    int    number_of_bands;
    int    scano;
    int    status;
    int    paramno;
    int    version = 1;
    IAS_SPECTRAL_TYPE band_spectral_type;
    IAS_BPF *bpf;  /* temporary pointer to the BPF being set up */
    if (do_good)
    {
        if (sensor_id == IAS_OLI)
        {
            if (good_oli_bpf == NULL)
            {
                good_oli_bpf = ias_bpf_allocate();
            }
            bpf = good_oli_bpf;
        }
        else
        {
            if (good_tirs_bpf == NULL)
            {
                good_tirs_bpf = ias_bpf_allocate();
            }
            bpf = good_tirs_bpf;
        }
    }
    else
    {
        if (sensor_id == IAS_OLI)
        {
            if (bad_oli_bpf == NULL)
            {
                bad_oli_bpf = ias_bpf_allocate();
            }
            bpf = bad_oli_bpf;
        }
        else
        {
            if (bad_tirs_bpf == NULL)
            {
                bad_tirs_bpf = ias_bpf_allocate();
            }
            bpf = bad_tirs_bpf;
        }
    }

    /* Get the sensor_specific band list. */
    status = ias_sat_attr_get_sensor_band_numbers(sensor_id,
        IAS_NORMAL_BAND, 0, band_list, IAS_MAX_NBANDS, &number_of_bands);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error retrieving sensor-specific "
            "band list");
        return;
    }

    /* set up bpf structure */
#ifdef DEBUG
    printf("\n\nSetting up bpf data...\n");
#endif

    /* FILE_ATTRIBUTES */
    ts = timestamp;
    gmtime_r(&ts, &start_time);
    ts += (90 * 60); /* add 90 min for end*/
    gmtime_r(&ts, &end_time);

    ias_bpf_set_file_attributes_file_source(bpf, "test");
    sprintf(tmpstring,
        "%04d-%02d-%02dT%02d:%02d:%02d.00000",
        start_time.tm_year+1900, start_time.tm_mon+1,
        start_time.tm_mday,      start_time.tm_hour,
        start_time.tm_min,       start_time.tm_sec);
    ias_bpf_set_file_attributes_effective_date_begin(bpf, tmpstring);
    sprintf(tmpstring,
        "%04d-%02d-%02dT%02d:%02d:%02d.00000",
        end_time.tm_year+1900,   end_time.tm_mon+1,
        end_time.tm_mday,        end_time.tm_hour,
        end_time.tm_min,         end_time.tm_sec);
    ias_bpf_set_file_attributes_effective_date_end(bpf, tmpstring);
    ias_bpf_set_file_attributes_spacecraft_name(bpf, "Landsat_8");

    /* Set information for the specified sensor ID. */
    if (sensor_id == IAS_OLI)
    {
        ias_bpf_set_file_attributes_sensor_name(bpf, "Operational Land "
            "Imager");
        sprintf(sensor_char, "O");
    }
    else
    {
        ias_bpf_set_file_attributes_sensor_name(bpf, "TIRS");
        sprintf(sensor_char, "T");
    }

    ias_bpf_set_file_attributes_baseline_date(bpf,
        "2010-03-21T00:00:00.00000");
    ias_bpf_set_file_attributes_description(bpf, "created by test tool");
    ias_bpf_set_file_attributes_version(bpf, version);

    /* Construct the file name depending on the sensor ID */
    status = snprintf(tmpstring, sizeof(tmpstring),
        "eval_L%s8BPF%04d%02d%02d%02d%02d%02d_%04d%02d%02d%02d%02d%02d.%02d",
        sensor_char,
        start_time.tm_year+1900, start_time.tm_mon+1,
        start_time.tm_mday,      start_time.tm_hour,
        start_time.tm_min,       start_time.tm_sec,
        end_time.tm_year+1900,   end_time.tm_mon+1,
        end_time.tm_mday,        end_time.tm_hour,
        end_time.tm_min,         end_time.tm_sec,
        version);
    if ((status < 0) || (status >= sizeof(tmpstring)))
    {
        assert_true_with_message(FALSE, "Error constructing BPF file "
            "name for specified sensor");
        return;
    }
    ias_bpf_set_file_attributes_file_name(bpf, tmpstring);

    /* ORBIT_PARAMETERS */
    ias_bpf_set_orbit_parameters_begin_orbit_number(bpf, 100);

    /* BIAS_MODEL */
    /* Set the coefficients to identifiable test values */
    minsca = 1;
    mindet = 1;
    for (index = 0; index < number_of_bands; index++)
    {
        bandno = band_list[index];

#ifdef DEBUG
        printf(" Adding parameters for band %d\n", bandno);
#endif

        /* Set the band number and spectral type, and get the number
           of SCAs and detectors per SCA. */
        ias_bpf_set_bias_model_band_number(bpf, bandno);
        band_spectral_type = 
            ias_sat_attr_get_spectral_type_from_band_number(bandno);
        ias_bpf_set_bias_model_spectral_type(bpf, bandno, 
            band_spectral_type);
        maxsca = ias_sat_attr_get_scas_per_band(bandno);
        maxdet = ias_sat_attr_get_detectors_per_sca(bandno);

        if (sensor_id == IAS_OLI)
        {
            for (scano = minsca; scano <= maxsca; scano++) 
            {
                /* Set the A0_Coefficient parameter value. */
                ias_bpf_set_bias_model_a0_coefficient(bpf, bandno,
                   (scano - 1), IAS_BPF_PAN_EVEN, (scano * 1e-6));
                ias_bpf_set_bias_model_a0_coefficient(bpf, bandno,
                   (scano - 1), IAS_BPF_PAN_ODD, (scano * 1e-6));

                /* Set the detector parameter values. */
                for (detno = mindet; detno <= maxdet; detno++)
                {
                    if (band_spectral_type == IAS_SPECTRAL_VNIR)
                    {
                        for (paramno = 0;
                             paramno < IAS_BPF_VNIR_VALUES;
                             paramno++)
                        {
                            params[paramno] = bandno * 100000 + scano * 1000 
                                + detno + paramno;
                        }

                        ias_bpf_set_bias_model_detector_coefficients(
                            bpf, bandno, (scano - 1), (detno - 1),
                            IAS_BPF_PAN_EVEN, params,
                            IAS_BPF_VNIR_VALUES);
                    }
                    else if (band_spectral_type == IAS_SPECTRAL_SWIR)
                    {
                        for (paramno = 0;
                             paramno < IAS_BPF_SWIR_VALUES;
                             paramno++)
                        {
                            params[paramno] = bandno * 100000 + scano * 1000 
                                + detno + paramno;
                        }

                        ias_bpf_set_bias_model_detector_coefficients(
                            bpf, bandno, (scano - 1), (detno - 1),
                            IAS_BPF_PAN_EVEN, params,
                            IAS_BPF_SWIR_VALUES);
                    }
                    else    /*  PAN band  */
                    {
                        for (paramno = 0;
                             paramno < IAS_BPF_PAN_VALUES;
                             paramno++)
                        {
                            params[paramno] = bandno * 100000 + scano * 1000 
                                + detno + paramno;
                        }

                        ias_bpf_set_bias_model_detector_coefficients(
                            bpf, bandno, (scano - 1), (detno - 1),
                            IAS_BPF_PAN_EVEN, params,
                            IAS_BPF_PAN_VALUES);

                        ias_bpf_set_bias_model_detector_coefficients(
                            bpf, bandno, (scano - 1), (detno - 1),
                            IAS_BPF_PAN_ODD, params,
                            IAS_BPF_PAN_VALUES);
                    }
                }
            }
        }
        else
        {
            for (scano = minsca; scano <= maxsca; scano++)
            {
                /* Set the detector parameter values. */
                for (detno = mindet; detno <= maxdet; detno++)
                {
                    for (paramno = 0;
                         paramno < IAS_BPF_THERMAL_VALUES;
                         paramno++)
                    {
                        params[paramno] = bandno * 100000 + scano * 1000
                            + detno + paramno;
                    }

                    ias_bpf_set_bias_model_detector_coefficients(
                        bpf, bandno, (scano - 1), (detno - 1),
                        IAS_BPF_PAN_EVEN, params,
                        IAS_BPF_THERMAL_VALUES);
                }
            }
        }
    }
#ifdef DEBUG
    printf("Done setting up bpf data...\n");
#endif
    return;
}

static void initialize()
{
    /* reset logging */
    ias_log_initialize("database_test_bpf");

    /* initialize the data in the good BPFs */
    timestamp = time(NULL);
    setup_bpf(1, IAS_OLI);

    timestamp = time(NULL);
    setup_bpf(1, IAS_TIRS);
}


static void cleanup()
{
    ias_bpf_free(good_oli_bpf);
    good_oli_bpf = NULL;
    ias_bpf_free(bad_oli_bpf);
    bad_oli_bpf = NULL;

    ias_bpf_free(good_tirs_bpf);
    good_tirs_bpf = NULL;
    ias_bpf_free(bad_tirs_bpf);
    bad_tirs_bpf = NULL;
}

/* Test to insert & update a BPF using the BPF DB I/O routine */
Ensure test_bpf_insert()
{
    /* Variables used to query data back out of the database */
    int status;
    char *db;
    char *user;
    char *pwd;
    char *host;
    struct ias_db_connection *conn = NULL;

#ifdef DEBUG
    struct timeb tp;   /* time structure, used to time the insert operations */
#endif

    /* Open DB connection */
    ias_db_initialize_database_lib();
    status = ias_db_get_connect_info(&db, &user, &pwd, &host);
    assert_equal(status, SUCCESS);
    conn = ias_db_connect_to_database(db, user, pwd, host);
    assert_not_equal(conn, NULL);
    free(db);
    free(user);
    free(pwd);
    free(host);
    if (conn == NULL) return;

    /* Insert an OLI BPF. */
    assert_not_equal(good_oli_bpf, NULL);
    if (good_oli_bpf == NULL) return;

    /* Test inserting this (OLI) BPF */
#ifdef DEBUG
    ftime(&tp);
    printf("Before insert: %ld %d\n",(long)tp.time,tp.millitm);
#endif
    status = ias_bpf_db_insert(conn, good_oli_bpf);
    assert_equal(status, SUCCESS);
#ifdef DEBUG
    ftime(&tp);
    printf("After insert: %ld %d\n", (long)tp.time,tp.millitm);
#endif

    /* Insert a TIRS BPF. */
    assert_not_equal(good_tirs_bpf, NULL);
    if (good_tirs_bpf == NULL)
        return;

   /* Test inserting this (TIRS) BPF */
#ifdef DEBUG
    ftime(&tp);
    printf("Before TIRS insert:  %ld %d\n", (long)tp.time, tp.millitm);
#endif
    status = ias_bpf_db_insert(conn, good_tirs_bpf);
    assert_equal(status, SUCCESS);
#ifdef  DEBUG
    ftime(&tp);
    printf("After TIRS insert:  %ld %d\n", (long)tp.time, tp.millitm);
#endif

    /* Check selected values written for the OLI BPF. */
    status = check_bpf_insert_results(conn, good_oli_bpf);
    assert_equal_with_message(status, SUCCESS, "OLI BPF confirmation "
        "check failed");

    /* Check selected values written to the TIRS BPF. */
    status = check_bpf_insert_results(conn, good_tirs_bpf);
    assert_equal_with_message(status, SUCCESS, "TIRS BPF confirmation "
        "check failed");

    /* 
     * Error cases ---------------------------------------------------------
     * don't show expected error messages
     */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE); 

    /* Copy the bpf structure, and make it invalid in various ways
     Same file - should be an error trying to insert a duplicate file */
    setup_bpf(0, IAS_OLI); /*initialize the bad BPF */
    status = ias_bpf_db_insert(conn, bad_oli_bpf);
    assert_equal_with_message(status, ERROR, "Inserting duplicate");

    /* Invalid file name */
    setup_bpf(0, IAS_TIRS); /*initialize the bad BPF */
    ias_bpf_set_file_attributes_file_name(bad_tirs_bpf, "test");
    status = ias_bpf_db_insert(conn, bad_tirs_bpf);
    assert_equal_with_message(status, ERROR, "Invalid file name");

    /* Version doesn't match file name */
    setup_bpf(0, IAS_OLI);  /*initialize the bad BPF */
    ias_bpf_set_file_attributes_version(bad_oli_bpf, 6);
    status = ias_bpf_db_insert(conn, bad_oli_bpf);
    assert_equal_with_message(status, ERROR, "Version doesn't match file "
        "name");

    /* invalid date formats */
    setup_bpf(0, IAS_TIRS); /* initialize the bad BPF */
    ias_bpf_set_file_attributes_effective_date_begin(bad_tirs_bpf, "1/1/2004");
    status = ias_bpf_db_insert(conn, bad_tirs_bpf);
    assert_equal_with_message(status, ERROR, "Invalid begin date");

    setup_bpf(0, IAS_OLI);  /*initialize the bad BPF */
    ias_bpf_set_file_attributes_effective_date_end(bad_oli_bpf, "2004-01-01");
    status = ias_bpf_db_insert(conn, bad_oli_bpf);
    assert_equal_with_message(status, ERROR, "Invalid end date");

    setup_bpf(0, IAS_TIRS);  /* initialize the bad BPF */
    ias_bpf_set_file_attributes_baseline_date(bad_tirs_bpf, "01012004");
    status = ias_bpf_db_insert(conn, bad_tirs_bpf);
    assert_equal_with_message(status, ERROR, "Invalid baseline date");

    /* end date < start date */
    setup_bpf(0, IAS_OLI);  /* initialize the bad BPF */
    ias_bpf_set_file_attributes_effective_date_end(bad_oli_bpf, 
        "1990-03-21T00:00:00.00000");
    status = ias_bpf_db_insert(conn, bad_oli_bpf);
    assert_equal_with_message(status, ERROR, "End date before start date");

    /* 'Garbage' sensor name.  */
    setup_bpf(0, IAS_OLI);    /* initialize the bad BPF */
    ias_bpf_set_file_attributes_sensor_name(bad_oli_bpf, "Fred");
    status = ias_bpf_db_insert(conn, bad_oli_bpf);
    assert_equal_with_message(status, ERROR, "Attempting to insert invalid "
        "OLI sensor name");

    setup_bpf(0, IAS_TIRS);   /* Initialize the bad BPF */
    ias_bpf_set_file_attributes_sensor_name(bad_tirs_bpf, "");
    status = ias_bpf_db_insert(conn, bad_tirs_bpf);
    assert_equal_with_message(status, ERROR, "Attempting to insert invalid "
        "TIRS sensor name");

    /* Invalid orbit number */
    setup_bpf(0, IAS_TIRS);  /* initialize the bad BPF */
    ias_bpf_set_orbit_parameters_begin_orbit_number(bad_tirs_bpf, -99);
    status = ias_bpf_db_insert(conn, bad_tirs_bpf);
    assert_equal_with_message(status, ERROR, "Invalid orbit number");

    /* Invalid band bias model data */
    setup_bpf(0, IAS_OLI);
        /* initialize the bad BPF set a VNIR band to SWIR */
    ias_bpf_set_bias_model_spectral_type(bad_oli_bpf, 1, 99); 
    status = ias_bpf_db_insert(conn, bad_oli_bpf);
    assert_equal_with_message(status, ERROR, "Invalid bias model data");

    /* Re-enable logging for the remaining tests */
    ias_log_initialize("database_test_bpf");

    /* Set some fields blank to let the library fill them in
      increment the version to create another unique BPF */
    setup_bpf(0, IAS_TIRS);
        /* initialize the bad BPF, in this case it has some empty fields  */
    ias_bpf_set_file_attributes_version(bad_tirs_bpf, 2);
    ias_bpf_set_file_attributes_file_name(bad_tirs_bpf, "");
    ias_bpf_set_file_attributes_description(bad_tirs_bpf, "");
    ias_bpf_set_file_attributes_sensor_name(bad_tirs_bpf, "TIRS");
    ias_bpf_set_file_attributes_spacecraft_name(bad_tirs_bpf, "");

#ifdef DEBUG
    ftime(&tp);
    printf("Before insert: %ld %d\n",(long)tp.time,tp.millitm);
#endif

    status = ias_bpf_db_insert(conn, bad_tirs_bpf);
    assert_equal_with_message(status, SUCCESS, "Error inserting into BPF "
        "database");
#ifdef DEBUG
    ftime(&tp);
    printf("After insert: %ld %d\n",(long)tp.time,tp.millitm);
#endif
    ias_db_close_connection(conn);
    ias_db_close_database_lib();
}


int main(int argc, char **argv)
{
    /* Initialize the satellite attributes library */
    int status = ias_sat_attr_initialize(IAS_L8);
    if (status != SUCCESS)
    {
        printf("Error initializing Satellite Attributes Library.\n");
        exit(1);
    }

    /* Initialize the test suite. */
    TestSuite *suite = create_test_suite();

    /* Set up the setup & teardown for all the tests */
    setup(suite, initialize);
    teardown(suite, cleanup);

    if (getenv("ENABLE_DB_TESTS"))
    {
        add_test(suite, test_bpf_insert);
    } 
    else 
    {
        printf("Database tests are not enabled\n");
        exit(EXIT_SUCCESS);
    }

    if (argc > 1) 
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}
