/*************************************************************************
NAME: database_test_bpf_version.c

PURPOSE: Test the ability to find the version used for a BPF.

Algorithm References: None
**************************************************************************/
#include <stdio.h>
#include <stdlib.h>

#include "cgreen/cgreen.h"
#include "ias_bpf.h"
#include "ias_bpf_database.h"
#include "ias_db_get_connect_info.h"
#include "ias_db_insert.h"
#include "ias_db_query.h"
#include "ias_db_stored_proc.h"
#include "ias_logging.h"
#include "ias_miscellaneous.h"

Ensure test_bpf_version_empty()
{
    /* The version numbering for the BPF starts at 1 */
    const int starting_version = 1;

    char *db = NULL;
    char *user = NULL;
    char *pwd = NULL;
    char *host = NULL;
    char tmpstring[256];
   
    int status = 0;

    /* Global variables used by tests, and teardown */
    IAS_BPF *bpf = NULL; /* used to test a successful insert */
    const struct IAS_BPF_FILE_ATTRIBUTES *attributes = NULL;
    struct ias_db_connection *conn = NULL;

    /* Open DB connection */
    ias_db_initialize_database_lib();
    status = ias_db_get_connect_info(&db, &user, &pwd, &host);
    assert_equal(status, SUCCESS);

    conn = ias_db_connect_to_database(db, user, pwd, host);
    assert_not_equal(conn, NULL);
    free(db);
    free(user);
    free(pwd);
    free(host);
    if (conn == NULL)
    { 
        assert_true(0); 
        return;
    }

    bpf = ias_bpf_allocate();
    if (bpf == NULL)
    { 
        assert_true(0);
        return;
    }

    /*Use a date where something should not exist, a check could be performed*/
    sprintf(tmpstring,
        "%04d-%02d-%02dT%02d:%02d:%02d.00000",
        1772, 1, 1, 1, 1, 1);
    ias_bpf_set_file_attributes_effective_date_begin(bpf, tmpstring); 
    
    sprintf(tmpstring,
        "%04d-%02d-%02dT%02d:%02d:%02d.00000",
        1772, 1, 2, 1, 1, 1);
    ias_bpf_set_file_attributes_effective_date_end(bpf, tmpstring);

    /* Choose a specific sensor. */
    ias_bpf_set_file_attributes_sensor_name(bpf, "OLI");

    /* Populate the initial version in the database. */
    status = ias_bpf_db_populate_version(conn, bpf);
    assert_equal(status, SUCCESS);

    /* Check that the version populated was 1 */
    attributes = ias_bpf_get_file_attributes(bpf);
    assert_equal(attributes->version, starting_version);

    ias_bpf_free(bpf);
    bpf = NULL;
}

Ensure test_bpf_version_increment_oli()
{
    const int max_clause = 1023;

    char new_file_source[] = "test";
    char new_status[] = "INACTIVE";
    char new_description[] = "test";
    char new_date[] = "2011-01-01";
    char new_file_name[80]; /* the file name must be unique */

    /* Use the database-centric sensor name (shorter form) */
    char sensor_id_string[] = "OLI";
    char sensor_id_char[2] = "O";
    char *db = NULL;
    char *user = NULL;
    char *pwd = NULL;
    char *host = NULL;
    char start_string[256];
    char end_string[256];
    char where_clause[max_clause + 1];

    /* The version numbering for the BPF starts at 1 */
    const int starting_version = 1;

    int  new_orbit = 1;
    int files_available = 0;
    int max_version = 0;
    int status = 0;
    int table_size = 0;

    /* Global variables used by tests, and teardown */
    IAS_BPF *bpf = NULL; /* used to test a successful insert */
    const struct IAS_BPF_FILE_ATTRIBUTES *attributes = NULL;
    struct ias_db_connection *conn = NULL;
    IAS_DATETIME effective_date_begin;  /* begin date of this bpf */
    IAS_DATETIME effective_date_end;    /* end date of this bpf */

    /* Open DB connection */
    ias_db_initialize_database_lib();
    status = ias_db_get_connect_info(&db, &user, &pwd, &host);
    assert_equal(status, SUCCESS);

    /* Build a query to check if there is a version to update or
        if an insert is required */
    IAS_DB_QUERY_TABLE table[] =
    {
        {"COUNT(*)", &files_available,
            sizeof(int), IAS_QUERY_INT, 1},
        {"MAX(VERSION_NUMBER)", &max_version, sizeof(int),
            IAS_QUERY_INT, 1}
    };

    bpf = ias_bpf_allocate();
    if (bpf == NULL)
    {
        IAS_LOG_ERROR("Input BPF structure is NULL");
        assert_true(0);
        return;
    }

    conn = ias_db_connect_to_database(db, user, pwd, host);
    assert_not_equal(conn, NULL);
    free(db);
    free(user);
    free(pwd);
    free(host);

    if (conn == NULL)
    {
        IAS_LOG_ERROR("Input ias_db_connection structure is NULL");
        assert_true(0);
        return;
    }

    /*Use a date where something should not exist, a check could be performed*/
    sprintf(start_string,
        "%04d-%02d-%02dT%02d:%02d:%02d.00000",
        1771, 1, 1, 1, 1, 1);
    ias_bpf_set_file_attributes_effective_date_begin(bpf, start_string);

    sprintf(end_string,
        "%04d-%02d-%02dT%02d:%02d:%02d.00000",
        1771, 1, 2, 1, 1, 1);
    ias_bpf_set_file_attributes_effective_date_end(bpf, end_string);

    /* Set the sensor ID string in the file attributes. */
    ias_bpf_set_file_attributes_sensor_name(bpf, sensor_id_string);

    IAS_LOG_DEBUG("Parsing effective start/end times");

    /* Get the FILE_ATTRIBUTES group */
    attributes = ias_bpf_get_file_attributes(bpf);
    if (attributes == NULL)
    {
        IAS_LOG_ERROR("Cannot retrieve IAS_BPF_FILE_ATTRIBUTES group");
        assert_true(0);
        return;
    }

    /* effective_date_begin: must be valid date format */
    status = ias_misc_parse_datetime_string(
        attributes->effective_date_begin,
        IAS_DATETIME_CPF_FORMAT, &effective_date_begin);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Format of effective_date_begin (%s)",
            attributes->effective_date_begin);
        assert_true(0);
        return;
    }

    /* effective_date_end: must be valid date format and must be 
     * equal to or later to begin date */
    status = ias_misc_parse_datetime_string(
        attributes->effective_date_end,
        IAS_DATETIME_CPF_FORMAT, &effective_date_end);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Format of effective_date_end (%s)",
            attributes->effective_date_end);
        assert_true(0);
        return;
    }

    IAS_LOG_DEBUG("Building the query");

    /* Build the where clause for the query */
    status = snprintf(where_clause, max_clause,
        "EFFECTIVE_DATE_BEGIN <= "
        "to_date('%04d-%02d-%02d %02d.%02d.%02d','yyyy-MM-dd hh24.mi.ss') "
        "AND EFFECTIVE_DATE_END >= "
        "to_date('%04d-%02d-%02d %02d.%02d.%02d','yyyy-mm-dd hh24.mi.ss') "
        "AND sensor_id = '%s'",
        effective_date_begin.year,         effective_date_begin.month,
        effective_date_begin.day_of_month, effective_date_begin.hour,
        effective_date_begin.minute,       (int)effective_date_begin.second,
        effective_date_end.year,           effective_date_end.month,
        effective_date_end.day_of_month,   effective_date_end.hour,
        effective_date_end.minute,         (int)effective_date_end.second,
        sensor_id_string);
    if (status > max_clause)
    {
        IAS_LOG_ERROR("Query for the number of previous collections was"
            " unsuccessful.\n where clause: %s", where_clause);
        assert_true(0);
        return;
    }

    table_size = sizeof(table)/sizeof(IAS_DB_QUERY_TABLE);

    IAS_LOG_DEBUG("Performing query");

    /* Perform the query */
    status = ias_db_query_using_table(conn, "BPF_MAIN",
        table, table_size, where_clause, 1);

    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Unable to perform query using where clause %s",
            where_clause);
        assert_true(0);
        return;
    }

    IAS_LOG_DEBUG("Query for max version complete: %d", max_version);

    /* Set the version based on the version found */
    if (files_available < 1)
    {
        IAS_LOG_DEBUG("No version currently available performing insert");
        max_version = starting_version;
    }
    else
    {
        max_version += 1;
        IAS_LOG_DEBUG("Inserting version %d", max_version);
    }

    sprintf(new_file_name, "L%s8BPF%s_%s.%02d", sensor_id_char, start_string,
        end_string, max_version);
    IAS_DB_INSERT_TABLE insert_table[] =
    {
        IAS_DB_TABLE_DATETIME("effective_date_begin" , start_string),
        IAS_DB_TABLE_DATETIME("effective_date_end" , end_string),
        IAS_DB_TABLE_INT("version_number" , max_version),
        IAS_DB_TABLE_STRING("sensor_id", sensor_id_string),
        /* required fields that don't impact ias_bpf_db_populate_version */
        IAS_DB_TABLE_DATETIME("date_entered" , new_date),
        IAS_DB_TABLE_DATETIME("date_created" , new_date),
        IAS_DB_TABLE_STRING("file_name" , new_file_name),
        IAS_DB_TABLE_STRING("file_source" , new_file_source),
        IAS_DB_TABLE_STRING("description" , new_description),
        IAS_DB_TABLE_STRING("status" , new_status),
        IAS_DB_TABLE_INT("orbit_begin" , new_orbit),
    };
    table_size  = sizeof(insert_table)/sizeof(IAS_DB_INSERT_TABLE);

    status = ias_db_insert_using_table(conn, "BPF_MAIN",
        insert_table, table_size, 1);

    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Unable to insert BPF record");
        assert_true(0);
        return;
    }

    /* Test the function use to populate the version for use */
    status = ias_bpf_db_populate_version(conn, bpf); 
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Unable to insert BPF");
        assert_true(0);
        return;
    }

    /* Check that the version was populated correctly */
    attributes = ias_bpf_get_file_attributes(bpf);
    assert_equal(attributes->version, max_version + 1);

    ias_bpf_free(bpf);
    bpf = NULL;
}

Ensure test_bpf_version_increment_tirs()
{
    const int max_clause = 1023;

    char new_file_source[] = "test";
    char new_status[] = "INACTIVE";
    char new_description[] = "test";
    char new_date[] = "2011-01-01";
    char new_file_name[80]; /* the file name must be unique */
    char db_sensor_id_string[IAS_BPF_SENSOR_NAME_LENGTH];

    /* Use the DFCB-specified sensor name (longer form) */
    char sensor_id_string[] = "Thermal Infrared Sensor";
    char sensor_id_char[2] = "T";
    char *db = NULL;
    char *user = NULL;
    char *pwd = NULL;
    char *host = NULL;
    char start_string[256];
    char end_string[256];
    char where_clause[max_clause + 1];

    /* The version numbering for the BPF starts at 1 */
    const int starting_version = 1;

    int  new_orbit = 1;
    int files_available = 0;
    int max_version = 0;
    int status = 0;
    int table_size = 0;

    /* Global variables used by tests, and teardown */
    IAS_BPF *bpf = NULL; /* used to test a successful insert */
    const struct IAS_BPF_FILE_ATTRIBUTES *attributes = NULL;
    struct ias_db_connection *conn = NULL;
    IAS_DATETIME effective_date_begin;  /* begin date of this bpf */
    IAS_DATETIME effective_date_end;    /* end date of this bpf */

    /* Open DB connection */
    ias_db_initialize_database_lib();
    status = ias_db_get_connect_info(&db, &user, &pwd, &host);
    assert_equal(status, SUCCESS);

    /* Build a query to check if there is a version to update or
        if an insert is required */
    IAS_DB_QUERY_TABLE table[] =
    {
        {"COUNT(*)", &files_available,
            sizeof(int), IAS_QUERY_INT, 1},
        {"MAX(VERSION_NUMBER)", &max_version, sizeof(int),
            IAS_QUERY_INT, 1}
    };

    bpf = ias_bpf_allocate();
    if (bpf == NULL)
    {
        IAS_LOG_ERROR("Input BPF structure is NULL");
        assert_true(0);
        return;
    }

    conn = ias_db_connect_to_database(db, user, pwd, host);
    assert_not_equal(conn, NULL);
    free(db);
    free(user);
    free(pwd);
    free(host);

    if (conn == NULL)
    {
        IAS_LOG_ERROR("Input ias_db_connection structure is NULL");
        assert_true(0);
        return;
    }

    /*Use a date where something should not exist, a check could be performed*/
    sprintf(start_string,
        "%04d-%02d-%02dT%02d:%02d:%02d.00000",
        1771, 1, 1, 1, 1, 1);
    ias_bpf_set_file_attributes_effective_date_begin(bpf, start_string);

    sprintf(end_string,
        "%04d-%02d-%02dT%02d:%02d:%02d.00000",
        1771, 1, 2, 1, 1, 1);
    ias_bpf_set_file_attributes_effective_date_end(bpf, end_string);

    /* Set the sensor ID string in the file attributes. */
    ias_bpf_set_file_attributes_sensor_name(bpf, sensor_id_string);

    IAS_LOG_DEBUG("Parsing effective start/end times");

    /* Get the FILE_ATTRIBUTES group */
    attributes = ias_bpf_get_file_attributes(bpf);
    if (attributes == NULL)
    {
        IAS_LOG_ERROR("Cannot retrieve IAS_BPF_FILE_ATTRIBUTES group");
        assert_true(0);
        return;
    }

    /* effective_date_begin: must be valid date format */
    status = ias_misc_parse_datetime_string(
        attributes->effective_date_begin,
        IAS_DATETIME_CPF_FORMAT, &effective_date_begin);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Format of effective_date_begin (%s)",
            attributes->effective_date_begin);
        assert_true(0);
        return;
    }

    /* effective_date_end: must be valid date format and must be 
     * equal to or later to begin date */
    status = ias_misc_parse_datetime_string(
        attributes->effective_date_end,
        IAS_DATETIME_CPF_FORMAT, &effective_date_end);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Format of effective_date_end (%s)",
            attributes->effective_date_end);
        assert_true(0);
        return;
    }

    IAS_LOG_DEBUG("Building the query");

    /* Get the sensor name compatible with the BPF_MAIN sensor_id string. */
    sprintf(db_sensor_id_string, "TIRS");

    /* Build the where clause for the query */
    status = snprintf(where_clause, max_clause,
        "EFFECTIVE_DATE_BEGIN <= "
        "to_date('%04d-%02d-%02d %02d.%02d.%02d','yyyy-MM-dd hh24.mi.ss') "
        "AND EFFECTIVE_DATE_END >= "
        "to_date('%04d-%02d-%02d %02d.%02d.%02d','yyyy-mm-dd hh24.mi.ss') "
        "AND sensor_id = '%s'",
        effective_date_begin.year,         effective_date_begin.month,
        effective_date_begin.day_of_month, effective_date_begin.hour,
        effective_date_begin.minute,       (int)effective_date_begin.second,
        effective_date_end.year,           effective_date_end.month,
        effective_date_end.day_of_month,   effective_date_end.hour,
        effective_date_end.minute,         (int)effective_date_end.second,
        db_sensor_id_string);
    if (status > max_clause)
    {
        IAS_LOG_ERROR("Query for the number of previous collections was"
            " unsuccessful.\n where clause: %s", where_clause);
        assert_true(0);
        return;
    }

    table_size = sizeof(table)/sizeof(IAS_DB_QUERY_TABLE);

    IAS_LOG_DEBUG("Performing query");

    /* Perform the query */
    status = ias_db_query_using_table(conn, "BPF_MAIN",
        table, table_size, where_clause, 1);

    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Unable to perform query using where clause %s",
            where_clause);
        assert_true(0);
        return;
    }

    IAS_LOG_DEBUG("Query for max version complete: %d", max_version);

    /* Set the version based on the version found */
    if (files_available < 1)
    {
        IAS_LOG_DEBUG("No version currently available performing insert");
        max_version = starting_version;
    }
    else
    {
        max_version += 1;
        IAS_LOG_DEBUG("Inserting version %d", max_version);
    }

    sprintf(new_file_name, "L%s8BPF%s_%s.%02d", sensor_id_char, start_string,
        end_string, max_version);
    IAS_DB_INSERT_TABLE insert_table[] =
    {
        IAS_DB_TABLE_DATETIME("effective_date_begin" , start_string),
        IAS_DB_TABLE_DATETIME("effective_date_end" , end_string),
        IAS_DB_TABLE_INT("version_number" , max_version),
        IAS_DB_TABLE_STRING("sensor_id", db_sensor_id_string),
        /* required fields that don't impact ias_bpf_db_populate_version */
        IAS_DB_TABLE_DATETIME("date_entered" , new_date),
        IAS_DB_TABLE_DATETIME("date_created" , new_date),
        IAS_DB_TABLE_STRING("file_name" , new_file_name),
        IAS_DB_TABLE_STRING("file_source" , new_file_source),
        IAS_DB_TABLE_STRING("description" , new_description),
        IAS_DB_TABLE_STRING("status" , new_status),
        IAS_DB_TABLE_INT("orbit_begin" , new_orbit),
    };
    table_size  = sizeof(insert_table)/sizeof(IAS_DB_INSERT_TABLE);

    status = ias_db_insert_using_table(conn, "BPF_MAIN",
        insert_table, table_size, 1);

    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Unable to insert BPF record");
        assert_true(0);
        return;
    }

    /* Test the function use to populate the version for use */
    status = ias_bpf_db_populate_version(conn, bpf); 
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Unable to insert BPF");
        assert_true(0);
        return;
    }

    /* Check that the version was populated correctly */
    attributes = ias_bpf_get_file_attributes(bpf);
    assert_equal(attributes->version, max_version + 1);

    ias_bpf_free(bpf);
    bpf = NULL;
}

int main(int argc, char **argv)
{
    TestSuite *suite = create_test_suite();

    if (getenv("ENABLE_DB_TESTS"))
    {
        add_test(suite, test_bpf_version_empty);
        add_test(suite, test_bpf_version_increment_oli);
        add_test(suite, test_bpf_version_increment_tirs);
    }
    else
    {
        printf("Database tests are not enabled\n");
    }

    /* Turn off DEBUG messages */
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

    /* Initialize the satellite attributes library */
    ias_sat_attr_initialize(IAS_L8);

    return run_test_suite(suite, create_text_reporter());
}
