/*************************************************************************
 NAME:                        test_bpf_file

 PURPOSE: CGreen unit testing of BPF File I/O library routines

 RETURNS:  CGreen unit status

**************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <time.h>
#include <unistd.h>

#include "ias_logging.h"
#include "ias_types.h"
#include "ias_satellite_attributes.h"
#include "bpf_local_defines.h"
#include "ias_bpf.h"

#include "cgreen.h"


static char bpf_dir[] = "ias_lib/bpf";
static char oli_bpf_name[] = "test_LO8BPF20140310103310_20140310103345.07";
static char tirs_bpf_name[] = "test_LT8BPF20140310103310_20140310103345.01";


static void check_bpf_file_attributes(const struct IAS_BPF_FILE_ATTRIBUTES
    *expected_file_atts, const struct IAS_BPF_FILE_ATTRIBUTES *test_file_atts)
{
    /* Check the BPF values against the test values */
    assert_string_equal(expected_file_atts->effective_date_begin,
        test_file_atts->effective_date_begin);
    assert_string_equal(expected_file_atts->effective_date_end,
        test_file_atts->effective_date_end);
    assert_string_equal(expected_file_atts->baseline_date,
        test_file_atts->baseline_date);
    assert_string_equal(expected_file_atts->file_name,
        test_file_atts->file_name);
    assert_string_equal(expected_file_atts->file_source,
        test_file_atts->file_source);
    assert_string_equal(expected_file_atts->spacecraft_name,
        test_file_atts->spacecraft_name);
    assert_string_equal(expected_file_atts->sensor_name,
        test_file_atts->sensor_name);
    assert_string_equal(expected_file_atts->description,
        test_file_atts->description);
    assert_equal(expected_file_atts->version, test_file_atts->version);
}


/* Local static routine to open a BPF file */
static IAS_BPF *open_bpf(const char *bpf_name)
{
    char *test_data_path = NULL;
    char bpf_filename[PATH_MAX];
    IAS_BPF *bpf = NULL;
    int status;

    /* Get the test data directory */
    test_data_path = getenv("IAS_UNIT_TEST_DATA_DIR");
    assert_not_equal_with_message(test_data_path, NULL, "Cannot find test "
        "data directory on system");
    if (test_data_path == NULL)
    {
        return NULL;
    }

    /* Build the BPF file name */
    status = snprintf(bpf_filename, sizeof(bpf_filename), "%s/%s/%s",
        test_data_path, bpf_dir, bpf_name);
    if ((status < 0) || (status >= sizeof(bpf_filename)))
    {
        assert_true_with_message(FALSE, "Error constructing test OLI BPF "
            "file name");
        return NULL;
    }

    /* Open the test BPF */
    bpf = ias_bpf_read(bpf_filename);
    assert_not_equal_with_message(bpf, NULL, "Cannot read BPF contents "
        "from file %s", bpf_filename);
    if (bpf == NULL)
    {
        return NULL;
    }

    /* Done */
    return bpf;
}


/* Test to allocate a BPF structure */
Ensure test_ias_bpf_allocate()
{
    IAS_BPF *test_bpf = NULL;

    test_bpf = ias_bpf_allocate();
    assert_not_equal(test_bpf, NULL);
    if (test_bpf == NULL)
    {
        printf("Cannot allocate test BPF data block\n");
        return;
    }

    ias_bpf_free(test_bpf);
}


/* Routines to extract the specified data substructures from the BPF
   data structure */

Ensure test_read_file_attributes()
{
    IAS_BPF *bpf = NULL;
    const struct IAS_BPF_FILE_ATTRIBUTES *oli_file_attributes = NULL;
    const struct IAS_BPF_FILE_ATTRIBUTES *tirs_file_attributes = NULL;
    struct IAS_BPF_FILE_ATTRIBUTES expected_oli_file_attributes;
    struct IAS_BPF_FILE_ATTRIBUTES expected_tirs_file_attributes;

    /* Clear out the memory space for the expected values data structures. */
    memset(&expected_oli_file_attributes, 0,
        sizeof(expected_oli_file_attributes));
    memset(&expected_tirs_file_attributes, 0,
        sizeof(expected_tirs_file_attributes));

    /* Populate the OLI expected file attribute values. */
    strcpy(expected_oli_file_attributes.effective_date_begin,
        "2014-03-10T10:33:10");
    strcpy(expected_oli_file_attributes.effective_date_end,
        "2014-03-10T10:33:45");
    strcpy(expected_oli_file_attributes.baseline_date,
        "2014-03-14T10:00:00");
    strcpy(expected_oli_file_attributes.file_name,
        "test_LO8BPF20140310103310_20140310103345.06");
    strcpy(expected_oli_file_attributes.file_source, "test");
    strcpy(expected_oli_file_attributes.spacecraft_name, "Landsat_8");
    strcpy(expected_oli_file_attributes.sensor_name,
        "Operational Land Imager");
    strcpy(expected_oli_file_attributes.description, "Created for testing");
    expected_oli_file_attributes.version = 6;

    /* Populate the TIRS expected file attribute values. */
    strcpy(expected_tirs_file_attributes.effective_date_begin,
        expected_oli_file_attributes.effective_date_begin);
    strcpy(expected_tirs_file_attributes.effective_date_end,
        expected_oli_file_attributes.effective_date_end);
    strcpy(expected_tirs_file_attributes.baseline_date,
        "2014-03-25T12:50:55");
    strcpy(expected_tirs_file_attributes.file_name,
        "test_LT8BPF20140310103310_20140310103345.01");
    strcpy(expected_tirs_file_attributes.file_source, "test");
    strcpy(expected_tirs_file_attributes.spacecraft_name, "Landsat_8");
    strcpy(expected_tirs_file_attributes.sensor_name,
        "Thermal Infrared Sensor");
    strcpy(expected_tirs_file_attributes.description, "Created for testing");
    expected_tirs_file_attributes.version = 1;

    printf("Checking OLI BPF FILE_ATTRIBUTES group...\n");

    /* Open/read the OLI BPF */
    bpf = open_bpf(oli_bpf_name);
    if (bpf == NULL)
    {
        assert_true_with_message(FALSE, "OLI BPF file '%s' not read",
            oli_bpf_name);
        return;
    }

    /* Get the OLI FILE_ATTRIBUTES group data */
    oli_file_attributes = ias_bpf_get_file_attributes(bpf);
    assert_true(oli_file_attributes != NULL);
    if (oli_file_attributes == NULL)
    {
        printf("OLI FILE_ATTRIBUTES group not extracted\n");
        ias_bpf_free(bpf);
        return;
    }

    /* Check the read file attributes vs. the expected values. */
    check_bpf_file_attributes(
        (const struct IAS_BPF_FILE_ATTRIBUTES *)&expected_oli_file_attributes,
        oli_file_attributes);

    /* clean up */
    ias_bpf_free(bpf);
    bpf = NULL;

    /* Open/read the TIRS BPF */
    printf("Checking TIRS BPF FILE_ATTRIBUTES group...\n");

    bpf = open_bpf(tirs_bpf_name);
    if (bpf == NULL)
    {
        assert_true_with_message(FALSE, "TIRS BPF file '%s' not read",
            tirs_bpf_name);
        return;
    }

    /* Get the TIRS FILE_ATTRIBUTES group data */
    tirs_file_attributes = ias_bpf_get_file_attributes(bpf);
    assert_true(tirs_file_attributes != NULL);
    if (tirs_file_attributes == NULL)
    {
        printf("TIRS FILE_ATTRIBUTES group not extracted\n");
        ias_bpf_free(bpf);
        return;
    }

    /* Check the read file attributes vs. the expected values. */
    check_bpf_file_attributes(
        (const struct IAS_BPF_FILE_ATTRIBUTES *)&expected_tirs_file_attributes,
        tirs_file_attributes);

    /* clean up */
    ias_bpf_free(bpf);
}



Ensure test_read_orbit_parameters()
{
    int oli_orbit_number = 15345;
    int tirs_orbit_number = 12384;
    IAS_BPF *bpf = NULL;
    const struct IAS_BPF_ORBIT_PARAMETERS *oli_orbit_parameters = NULL;
    const struct IAS_BPF_ORBIT_PARAMETERS *tirs_orbit_parameters = NULL;


    printf("Checking OLI ORBIT_PARAMETERS group...\n");

    /* Open/read the OLI BPF. */
    bpf = open_bpf(oli_bpf_name);
    if (bpf == NULL)
    {
        assert_true_with_message(FALSE, "OLI BPF file not read");
        return;
    }

    /* Get the OLI ORBIT_PARAMETERS group data. */
    oli_orbit_parameters = ias_bpf_get_orbit_parameters(bpf);
    assert_true(oli_orbit_parameters != NULL);
    if (oli_orbit_parameters == NULL)
    {
        printf("OLI ORBIT_PARAMETERS group not extracted\n");
        ias_bpf_free(bpf);
        return;
    }

    /* Check the expected values against the test values */
    assert_equal(oli_orbit_number, oli_orbit_parameters->begin_orbit_number);

    /* Clean up. */
    ias_bpf_free(bpf);
    bpf = NULL;

    /* Open/read the TIRS BPF. */
    printf("Checking TIRS ORBIT_PARAMETERS group...\n");

    bpf = open_bpf(tirs_bpf_name);
    if (bpf == NULL)
    {
        assert_true_with_message(FALSE, "TIRS BPF file not read");
        return;
    }

    /* Get the TIRS ORBIT_PARAMETERS group data. */
    tirs_orbit_parameters = ias_bpf_get_orbit_parameters(bpf);
    assert_true(tirs_orbit_parameters != NULL);
    if (tirs_orbit_parameters == NULL)
    {
        printf("TIRS ORBIT_PARAMETERS group not extracted\n");
        ias_bpf_free(bpf);
        return;
    }

    /* Check the values against the test values. */
    assert_equal(tirs_orbit_number, tirs_orbit_parameters->begin_orbit_number);

    /* Clean up. */
    ias_bpf_free(bpf);
}



Ensure test_read_tirs_bias_model()
{
    IAS_BPF *bpf = NULL;
    const struct IAS_BPF_BIAS_MODEL *bias_model = NULL;

    double preacq_deepspace[IAS_BPF_THERMAL_MAX_NDET];
    double postacq_deepspace[IAS_BPF_THERMAL_MAX_NDET];
    double thermal_bpf_values[IAS_BPF_THERMAL_VALUES] = {1.00, 3.00};
    double thermal_bpf_model_coeffs[IAS_BPF_THERMAL_VALUES] = {0.0, 0.0};

    int status;
    int test_band;                     /* Band number to test */
    int test_sca;                      /* SCA number to test */
    int test_det;                      /* Detector number to test */
    int ii = 0;                        /* General loop counter */


    /* Open/read the BPF */
    printf("Checking TIRS BIAS_MODEL group...\n");

    bpf = open_bpf(tirs_bpf_name);
    if (bpf == NULL)
    {
        assert_true_with_message(FALSE, "Error reading TIRS BPF '%s",
            tirs_bpf_name);
        return;
    }

    /* Get the BIAS_MODEL group data */
    bias_model = ias_bpf_get_bias_model(bpf);
    assert_not_equal_with_message(bias_model, NULL, "BIAS_MODEL "
        "data group not extracted");
    if (bias_model == NULL)
    {
        ias_bpf_free(bpf);
        return;
    }

    /* Get the bias parameters for band 10, SCA 2, detector 500 */
    test_band = 10;
    test_sca = 2;
    test_det = 500;

    status = ias_bpf_get_model_parameters(bias_model,
        IAS_SPECTRAL_THERMAL, test_band, test_sca, IAS_BPF_THERMAL_MAX_NDET,
        IAS_BPF_PAN_EVEN, preacq_deepspace, postacq_deepspace,
        NULL, NULL, NULL);
    assert_equal_with_message(status, SUCCESS, "Cannot retrieve band "
        "%d SCA %d parameters", test_band, test_sca);
    if (status != SUCCESS)
    {
        ias_bpf_free(bpf);
        return;
    }
    thermal_bpf_model_coeffs[0] = preacq_deepspace[test_det - 1];
    thermal_bpf_model_coeffs[1] = postacq_deepspace[test_det - 1];

    /* Check the THERMAL read */
    for (ii = 0; ii < IAS_BPF_THERMAL_VALUES; ii++)
    {
        assert_double_equal(thermal_bpf_values[ii],
            thermal_bpf_model_coeffs[ii]);
    }

    /* Clean up. */
    ias_bpf_free(bpf);
}



Ensure test_read_oli_bias_model()
{
    IAS_BPF *bpf = NULL;
    const struct IAS_BPF_BIAS_MODEL *bias_model = NULL;

    /* Known test data */
    
    /* Vectors to hold bias model pre-acquisition/post-acquisition/mean
       average responses, respectively */
    double pan_preacq_shutter[IAS_BPF_PAN_MAX_NDET];
    double pan_postacq_shutter[IAS_BPF_PAN_MAX_NDET];
    double pan_a1_coefficient[IAS_BPF_PAN_MAX_NDET];
    double pan_c1_coefficient[IAS_BPF_PAN_MAX_NDET];
    double preacq_shutter[IAS_BPF_MS_MAX_NDET];
    double postacq_shutter[IAS_BPF_MS_MAX_NDET];
    double a1_coefficient[IAS_BPF_MS_MAX_NDET];
    double c1_coefficient[IAS_BPF_MS_MAX_NDET];

    double vnir_bpf_model_coeffs[IAS_BPF_VNIR_VALUES + 1] =
        {0.0, 0.0, 0.0, 0.0, 0.0};
    double pan_odd_bpf_model_coeffs[IAS_BPF_PAN_VALUES + 1] =
        {0.0, 0.0, 0.0, 0.0, 0.0};
    double pan_even_bpf_model_coeffs[IAS_BPF_PAN_VALUES + 1] =
        {0.0, 0.0, 0.0, 0.0, 0.0};
    double swir_bpf_model_coeffs[IAS_BPF_SWIR_VALUES + 1] =
        {0.0, 0.0, 0.0, 0.0, 0.0};

    /* Test data vectors.  These data are populated in the
       prototype BPF file under test */
    double vnir_bpf_values[IAS_BPF_VNIR_VALUES] = 
           {1171.94, 1171.69, 0.236531, 65.932693};
    double vnir_a0 = 0.690801;

    double swir_bpf_values[IAS_BPF_SWIR_VALUES] = 
           {1127.69, 1127.80, 0.236531, 65.932693};
    double swir_a0 = 0.649777;

    double pan_even_bpf_values[IAS_BPF_PAN_VALUES] = 
           {1110.22, 1110.06, 0.236531, 65.932693};
    double pan_even_a0 = 0.581586;

    double pan_odd_bpf_values[IAS_BPF_PAN_VALUES] = 
           {1098.65, 1098.51, 0.236531, 65.932693};
    double pan_odd_a0 = 0.579791;


    int status;
    int test_band;                     /* Band number to test */
    int test_sca;                      /* SCA number to test */
    int test_det;                      /* Detector number to test */
    int ii = 0;                        /* General loop counter */


    printf("Checking OLI BIAS_MODEL group...\n");

    /* Open/read the BPF */
    bpf = open_bpf(oli_bpf_name);
    assert_not_equal_with_message(bpf, NULL, "Error reading BPF file");
    if (bpf == NULL)
    {
        return;
    }

    /* Get the BIAS_MODEL group data.  */
    bias_model = ias_bpf_get_bias_model(bpf);
    assert_not_equal_with_message(bias_model, NULL, "BIAS_MODEL "
        "data group not extracted");
    if (bias_model == NULL)
    {
        ias_bpf_free(bpf);
        return;
    }

    /* Get some VNIR band bias parameters */
    test_band = 1;
    test_sca = 1;
    test_det = 1;

    status = ias_bpf_get_model_parameters(bias_model,
        IAS_SPECTRAL_VNIR, test_band, test_sca, IAS_BPF_MS_MAX_NDET,
        IAS_BPF_PAN_EVEN, preacq_shutter, postacq_shutter, a1_coefficient,
        c1_coefficient, &vnir_bpf_model_coeffs[4]);

    assert_equal_with_message(status, SUCCESS, "Cannot retrieve VNIR band "
        "%d SCA %d parameters", test_band, test_sca);
    if (status != SUCCESS)
    {
        ias_bpf_free(bpf);
        return;
    }
    vnir_bpf_model_coeffs[0] = preacq_shutter[test_det - 1];
    vnir_bpf_model_coeffs[1] = postacq_shutter[test_det - 1];
    vnir_bpf_model_coeffs[2] = a1_coefficient[test_det - 1];
    vnir_bpf_model_coeffs[3] = c1_coefficient[test_det - 1];

    /* Check the VNIR read */
    for (ii = 0; ii < IAS_BPF_VNIR_VALUES; ii++)
        assert_double_equal(vnir_bpf_values[ii], vnir_bpf_model_coeffs[ii]);

    assert_double_equal(vnir_bpf_model_coeffs[IAS_BPF_VNIR_VALUES],
        vnir_a0);

    /* Get some SWIR band bias parameters */
    test_band = 7;
    test_sca = 10;
    test_det = 150;

    status = ias_bpf_get_model_parameters(bias_model,
        IAS_SPECTRAL_SWIR, test_band, test_sca, IAS_BPF_MS_MAX_NDET,
        IAS_BPF_PAN_EVEN, preacq_shutter, postacq_shutter, a1_coefficient,
        c1_coefficient, &swir_bpf_model_coeffs[4]);

    assert_equal_with_message(status, SUCCESS, "Cannot retrieve SWIR band "
        "%d SCA %d parameters", test_band, test_sca);
    if (status != SUCCESS)
    {
        ias_bpf_free(bpf);
        return;
    }

    swir_bpf_model_coeffs[0] = preacq_shutter[test_det - 1];
    swir_bpf_model_coeffs[1] = postacq_shutter[test_det - 1];
    swir_bpf_model_coeffs[2] = a1_coefficient[test_det - 1];
    swir_bpf_model_coeffs[3] = c1_coefficient[test_det - 1];

    /* Check the SWIR read */
    for (ii = 0; ii < IAS_BPF_SWIR_VALUES; ii++)
        assert_double_equal(swir_bpf_values[ii], swir_bpf_model_coeffs[ii]);

    assert_double_equal(swir_bpf_model_coeffs[IAS_BPF_SWIR_VALUES],
        swir_a0);


    /* Get some PAN band bias parameters */
    test_band = 8;
    test_sca = 6;
    test_det = 487;

    status = ias_bpf_get_model_parameters(bias_model,
        IAS_SPECTRAL_PAN, test_band, test_sca, IAS_BPF_PAN_MAX_NDET,
        IAS_BPF_PAN_EVEN, pan_preacq_shutter, pan_postacq_shutter,
        pan_a1_coefficient, pan_c1_coefficient,
        &pan_even_bpf_model_coeffs[4]);
    assert_equal_with_message(status, SUCCESS, "Cannot retrieve even PAN "
        "band %d SCA %d parameters", test_band, test_sca);
    if (status != SUCCESS)
    {
        ias_bpf_free(bpf);
        return;
    }

    pan_even_bpf_model_coeffs[0] = pan_preacq_shutter[test_det - 1];
    pan_even_bpf_model_coeffs[1] = pan_postacq_shutter[test_det - 1];
    pan_even_bpf_model_coeffs[2] = pan_a1_coefficient[test_det - 1];
    pan_even_bpf_model_coeffs[3] = pan_c1_coefficient[test_det - 1];

    /* Read the odd parameters */
    status = ias_bpf_get_model_parameters(bias_model,
        IAS_SPECTRAL_PAN, test_band, test_sca, IAS_BPF_PAN_MAX_NDET,
        IAS_BPF_PAN_ODD, pan_preacq_shutter, pan_postacq_shutter,
        pan_a1_coefficient, pan_c1_coefficient,
        &pan_odd_bpf_model_coeffs[4]);
    assert_equal_with_message(status, SUCCESS, "Cannot retrieve odd PAN "
        "band %d SCA %d parameters", test_band, test_sca);
    if (status != SUCCESS)
    {
        ias_bpf_free(bpf);
        return;
    }

    pan_odd_bpf_model_coeffs[0] = pan_preacq_shutter[test_det - 1];
    pan_odd_bpf_model_coeffs[1] = pan_postacq_shutter[test_det - 1];
    pan_odd_bpf_model_coeffs[2] = pan_a1_coefficient[test_det - 1];
    pan_odd_bpf_model_coeffs[3] = pan_c1_coefficient[test_det - 1];

    /* Check the PAN read */
    for (ii = 0; ii < IAS_BPF_PAN_VALUES; ii++)
    {
        assert_double_equal(pan_odd_bpf_values[ii],
           pan_odd_bpf_model_coeffs[ii]);
        assert_double_equal(pan_even_bpf_values[ii],
           pan_even_bpf_model_coeffs[ii]);
    }

    assert_double_equal(pan_even_bpf_model_coeffs[IAS_BPF_PAN_VALUES],
        pan_even_a0);
    assert_double_equal(pan_odd_bpf_model_coeffs[IAS_BPF_PAN_VALUES],
        pan_odd_a0);

    /* Free BPF resources */
    ias_bpf_free(bpf);
}



/* Test of BPF set/get functions applied to a TIRS BPF */
Ensure test_set_tirs_bpf_parameter_values()
{
    const struct IAS_BPF_FILE_ATTRIBUTES *file_attributes = NULL;
    const struct IAS_BPF_ORBIT_PARAMETERS *orbit_parameters = NULL;
    const struct IAS_BPF_BIAS_MODEL *bias_model = NULL;
    char date_string[IAS_L0R_DATE_LENGTH];
    IAS_BPF *test_bpf = NULL;
    int status;
    int sca_index;             /* SCA loop counter */
    int det_index;             /* Detector loop counter */
    int num_dets = 10;
    int num_scas = 2;
    int band_number = 10;
    int band_index;
    int param_count;           /* Bias parameter loop counter */
    double thermal_params[IAS_BPF_THERMAL_VALUES];
                               /* Array containing deepspace averages */
    double test_param_value;


    /* Check setting parameters for a TIRS BPF. */
    printf("Testing TIRS manual BPF editing...\n");

    test_bpf = ias_bpf_allocate();
    assert_not_equal_with_message(test_bpf, NULL, "Cannot allocate BPF "
        "data block");
    if (test_bpf == NULL)
    {
        return;
    }

    /* Set the members of the FILE_ATTRIBUTES group */
    ias_bpf_set_file_attributes_file_name(test_bpf,
        "eval_LT8BPF2010819170045_20100819180045.01");

    strcpy(date_string, "2010-08-19T08:25:20.12345");
    ias_bpf_set_file_attributes_effective_date_begin(test_bpf, date_string);

    strcpy(date_string, "2010-08-19T09:25:20.12345");
    ias_bpf_set_file_attributes_effective_date_end(test_bpf, date_string);

    ias_bpf_set_file_attributes_baseline_date(test_bpf,
        "2010-03-21T00:00:00.00000");    
    ias_bpf_set_file_attributes_file_source(test_bpf, "test");
    ias_bpf_set_file_attributes_spacecraft_name(test_bpf, "Landsat 8");
    ias_bpf_set_file_attributes_sensor_name(test_bpf, "TIRS");
    ias_bpf_set_file_attributes_description(test_bpf, "created by test tool");
    ias_bpf_set_file_attributes_version(test_bpf, 1);

    /* Set the members of the ORBIT_PARAMETERS group */
    ias_bpf_set_orbit_parameters_begin_orbit_number(test_bpf, 100);

    /* Set band 10 of the BIAS_MODEL group.  Set it as a THERMAL band;
       the THERMAL band model coefficient data block will also be allocated
       at this time */
    band_number = 10;
    status = ias_bpf_set_bias_model_band_number(test_bpf, band_number);
    assert_equal_with_message(status, SUCCESS, "Error setting band number");
    if (status != SUCCESS)
    {
        ias_bpf_free(test_bpf);
        return;
    }

    status = ias_bpf_set_bias_model_spectral_type(test_bpf, band_number,
        IAS_SPECTRAL_THERMAL);
    assert_equal_with_message(status, SUCCESS, "Error setting band "
        "type and/or allocating correct bias model data block");
    if (status != SUCCESS)
    {
        ias_bpf_free(test_bpf);
        return;
    }

    /* Populate SCA 1 and 2, detectors 1-10 */
    for (sca_index = 0; sca_index < num_scas; sca_index++)
    {
        for (det_index = 0; det_index < num_dets; det_index++)
        {
            for (param_count = 0; param_count < IAS_BPF_THERMAL_VALUES;
                 param_count++)
            {
                test_param_value = 100000 + (1000 * sca_index)
                    + (100 * det_index) + (10 * param_count);
                thermal_params[param_count] = test_param_value;
            }
            status = ias_bpf_set_bias_model_detector_coefficients(
                test_bpf, band_number, sca_index, det_index,
                IAS_BPF_PAN_EVEN, thermal_params, IAS_BPF_THERMAL_VALUES);
            if (status != SUCCESS)
            {
                assert_true_with_message(0, "Error setting detector-"
                    "specific THERMAL band model parameters");
                ias_bpf_free(test_bpf);
                return;
            }
        }
    }

    /* Read back FILE_ATTRIBUTES */
    file_attributes = ias_bpf_get_file_attributes(test_bpf);
    if (file_attributes == NULL)
    {
        IAS_LOG_ERROR("Retrieving TIRS FILE_ATTRIBUTES BPF information");
        ias_bpf_free(test_bpf);
        return;
    }
    assert_string_equal(file_attributes->file_name,
        "eval_LT8BPF2010819170045_20100819180045.01");
    assert_string_equal(file_attributes->file_source, "test");
    assert_string_equal(file_attributes->spacecraft_name, "Landsat 8");
    assert_string_equal(file_attributes->sensor_name, "TIRS");
    assert_string_equal(file_attributes->baseline_date,
        "2010-03-21T00:00:00.00000");
    assert_string_equal(file_attributes->effective_date_begin,
        "2010-08-19T08:25:20.12345");
    assert_string_equal(file_attributes->effective_date_end,
        "2010-08-19T09:25:20.12345");
    assert_string_equal(file_attributes->description,
        "created by test tool");
    assert_equal(file_attributes->version, 1);

    /* Read back ORBIT_PARAMETERS */
    orbit_parameters = ias_bpf_get_orbit_parameters(test_bpf);
    if (file_attributes == NULL)
    {
        IAS_LOG_ERROR("Retrieving TIRS ORBIT_PARAMETERS BPF information");
        ias_bpf_free(test_bpf);
        return;
    }
    assert_equal(orbit_parameters->begin_orbit_number, 100);

    /* Make sure only the THERMAL band array has been allocated and
       populated */
    band_index = ias_sat_attr_convert_band_number_to_index(band_number);
    assert_not_equal_with_message(band_index, ERROR, "Cannot convert band "
        "number %d to correct index value", band_number);
    if (band_index == ERROR)
    {
        ias_bpf_free(test_bpf);
        return;
    }

    bias_model = ias_bpf_get_bias_model(test_bpf);
    if (bias_model == NULL)
    {
        IAS_LOG_ERROR("Retrieving TIRS BIAS_MODEL BPF information");
        ias_bpf_free(test_bpf);
        return;
    }
    assert_equal(bias_model[band_index].band_number, band_number);
    assert_equal(bias_model[band_index].spectral_type,
        IAS_SPECTRAL_THERMAL);
    assert_equal(bias_model[band_index].vnir, NULL);
    assert_equal(bias_model[band_index].pan, NULL);
    assert_equal(bias_model[band_index].swir, NULL);
    assert_not_equal(bias_model[band_index].thermal, NULL);

    if ((bias_model[band_index].vnir != NULL)
        || (bias_model[band_index].pan != NULL)
        || (bias_model[band_index].swir != NULL)
        || (bias_model[band_index].thermal == NULL))
    {
        printf("THERMAL band buffer should be allocated, VNIR, PAN, and/or "
            "SWIR band buffer pointers should be NULL\n");
        ias_bpf_free(test_bpf);
        return;
    }

    /* Read back BIAS_MODEL */
    for (sca_index = 0; sca_index < num_scas; sca_index++)
    {
        for (det_index = 0; det_index < num_dets; det_index++)
        {
            for (param_count = 0; param_count < IAS_BPF_THERMAL_VALUES;
                 param_count++)
            {
                test_param_value =
                    100000 + (1000 * sca_index) + (100 * det_index)
                        + (10 * param_count);
                assert_double_equal(bias_model[band_index]
                    .thermal->det_params[sca_index][det_index][param_count],
                    test_param_value);
            }
        }
    }

    /* Clean up */
    ias_bpf_free(test_bpf);
}


/* Test of BPF set/get functions applied to an OLI BPF */
Ensure test_set_oli_bpf_parameter_values()
{
    const struct IAS_BPF_FILE_ATTRIBUTES *file_attributes = NULL;
    const struct IAS_BPF_ORBIT_PARAMETERS *orbit_parameters = NULL;
    const struct IAS_BPF_BIAS_MODEL *bias_model = NULL;

    char date_string[IAS_L0R_DATE_LENGTH];

    IAS_BPF *test_bpf = NULL;

    int status;
    int sca_index;             /* SCA loop counter */
    int det_index;             /* Detector loop counter */
    int num_dets = 10;
    int num_scas = 2;
    int band_number = 1;
    int band_index;
    int param_count;           /* Bias parameter loop counter */

    double vnir_params[IAS_BPF_VNIR_VALUES];
                               /* Array containing shutter averages +
                                  model coefficient */
    double a0_coeff = 0.0;
    double test_param_value = 0.0;


    printf("Testing OLI manual BPF editing...\n");

    /* Allocate a test BPF data block */
    test_bpf = ias_bpf_allocate();
    assert_not_equal_with_message(test_bpf, NULL, "Cannot allocate BPF "
        "data block");
    if (test_bpf == NULL)
    {
        return;
    }

    /* Set the members of the FILE_ATTRIBUTES group */
    ias_bpf_set_file_attributes_file_name(test_bpf,
        "eval_LO8BPF2010819170045_20100819180045.01");

    sprintf(date_string, "2010-08-19T08:25:20.12345");
    ias_bpf_set_file_attributes_effective_date_begin(test_bpf, date_string);

    sprintf(date_string, "2010-08-19T09:25:20.12345");
    ias_bpf_set_file_attributes_effective_date_end(test_bpf, date_string);

    ias_bpf_set_file_attributes_baseline_date(test_bpf,
        "2010-03-21T00:00:00.00000");    
    ias_bpf_set_file_attributes_file_source(test_bpf, "test");
    ias_bpf_set_file_attributes_spacecraft_name(test_bpf, "Landsat 8");
    ias_bpf_set_file_attributes_sensor_name(test_bpf, "OLI");
    ias_bpf_set_file_attributes_description(test_bpf, "created by test tool");
    ias_bpf_set_file_attributes_version(test_bpf, 1);

    /* Set the members of the ORBIT_PARAMETERS group */
    ias_bpf_set_orbit_parameters_begin_orbit_number(test_bpf, 100);

    /* Set band 1 of the BIAS_MODEL group.  Set it as a VNIR band;
       the VNIR band model coefficient data block will also be allocated
       at this time */
    status = ias_bpf_set_bias_model_band_number(test_bpf, band_number);
    assert_equal_with_message(status, SUCCESS, "Error setting band number");
    if (status != SUCCESS)
    {
        ias_bpf_free(test_bpf);
        return;
    }

    status = ias_bpf_set_bias_model_spectral_type(test_bpf, band_number,
        IAS_SPECTRAL_VNIR);
    assert_equal_with_message(status, SUCCESS, "Error setting band "
        "type and/or allocating correct bias model data block");
    if (status != SUCCESS)
    {
        ias_bpf_free(test_bpf);
        return;
    }

    /* Populate SCA 1 and 2, detectors 1-14 */
    for (sca_index = 0; sca_index < num_scas; sca_index++)
    {
        a0_coeff = sca_index * 1e-6;

        for (det_index = 0; det_index < num_dets; det_index++)
        {
            for (param_count = 0;
                 param_count < IAS_BPF_VNIR_VALUES;
                 param_count++)
            {
                test_param_value = 100000 + (1000 * sca_index)
                    + (100 * det_index) + (10 * param_count);
                vnir_params[param_count] = test_param_value;
            }
            status = ias_bpf_set_bias_model_detector_coefficients(
                test_bpf, band_number, sca_index, det_index,
                IAS_BPF_PAN_EVEN, vnir_params,
                IAS_BPF_VNIR_VALUES);
            if (status != SUCCESS)
            {
                assert_true_with_message(0, "Error setting detector-"
                    "specific VNIR band model parameters");
                ias_bpf_free(test_bpf);
                return;
            }
        }

        /* Set the A0 model coefficient */
        status = ias_bpf_set_bias_model_a0_coefficient(test_bpf,
            band_number, sca_index, IAS_BPF_PAN_EVEN, a0_coeff);
        if (status != SUCCESS)
        {
            assert_true_with_message(FALSE, "Error setting VNIR a0_"
                "coefficient parameter");
            ias_bpf_free(test_bpf);
            return;
        }
    }

    /* Pull the edited groups from the new BPF */
    file_attributes = ias_bpf_get_file_attributes(test_bpf);
    assert_not_equal_with_message(file_attributes, NULL, "Error retrieving "
        "file attributes");
    if (file_attributes == NULL)
    {
        ias_bpf_free(test_bpf);
        return;
    }
    orbit_parameters = ias_bpf_get_orbit_parameters(test_bpf);
    assert_not_equal_with_message(orbit_parameters, NULL, "Error "
        "retrieving orbit parameters");
    if (orbit_parameters == NULL)
    {
        ias_bpf_free(test_bpf);
        return;
    }
    bias_model = ias_bpf_get_bias_model(test_bpf);
    assert_not_equal_with_message(bias_model, NULL, "Error retrieving "
         "band bias model parameters");
    if (bias_model == NULL)
    {
        ias_bpf_free(test_bpf);
        return;
    }

    /* Read back FILE_ATTRIBUTES */
    assert_string_equal(file_attributes->file_name,
        "eval_LO8BPF2010819170045_20100819180045.01");
    assert_string_equal(file_attributes->file_source, "test");
    assert_string_equal(file_attributes->spacecraft_name, "Landsat 8");
    assert_string_equal(file_attributes->sensor_name, "OLI");
    assert_string_equal(file_attributes->baseline_date,
        "2010-03-21T00:00:00.00000");
    assert_string_equal(file_attributes->effective_date_begin,
        "2010-08-19T08:25:20.12345");
    assert_string_equal(file_attributes->effective_date_end,
        "2010-08-19T09:25:20.12345");
    assert_string_equal(file_attributes->description,
        "created by test tool");
    assert_equal(file_attributes->version, 1);

    /* Read back ORBIT_PARAMETERS */
    assert_equal(orbit_parameters->begin_orbit_number, 100);

    /* Read back BIAS_MODEL */
    band_index = ias_sat_attr_convert_band_number_to_index(band_number);
    assert_not_equal_with_message(band_index, ERROR, "Cannot convert band "
        "number %d to correct index value", band_number);
    if (band_index == ERROR)
    {
        ias_bpf_free(test_bpf);
        return;
    }

    /* Make sure only the VNIR band array has been allocated and
       populated */
    assert_equal(bias_model[band_index].band_number, band_number);
    assert_equal(bias_model[band_index].spectral_type,
        IAS_SPECTRAL_VNIR);
    assert_not_equal(bias_model[band_index].vnir, NULL);
    assert_equal(bias_model[band_index].pan, NULL);
    assert_equal(bias_model[band_index].swir, NULL);
    assert_equal(bias_model[band_index].thermal, NULL);

    if ((bias_model[band_index].vnir == NULL)
        || (bias_model[band_index].pan != NULL)
        || (bias_model[band_index].swir != NULL)
        || (bias_model[band_index].thermal != NULL))
    {
        printf("VNIR band buffer should be allocated, PAN, SWIR, and/or "
            "THERMAL band buffer pointers should be NULL\n");
        ias_bpf_free(test_bpf);
        return;
    }

    /* Read back the test values */
    for (sca_index = 0; sca_index < num_scas; sca_index++)
    {
        for (det_index = 0; det_index < num_dets; det_index++)
        {
            for (param_count = 0;
                 param_count < IAS_BPF_VNIR_VALUES;
                 param_count++)
            {
                test_param_value =
                    100000 + (1000 * sca_index) + (100 * det_index)
                        + (10 * param_count);
                assert_double_equal(bias_model[band_index]
                    .vnir->det_params[sca_index][det_index][param_count],
                    test_param_value);
            }
        }

        assert_double_equal(bias_model[band_index]
            .vnir->a0_coefficient[sca_index], sca_index * 1e-6);
    }

    /* Clean up */
    ias_bpf_free(test_bpf);
}



/*------------------------------------------------------------------------------
 * Test the BPF write function.  This function creates our own internal
 * representation of a BPF that will be written to disk and then immediately
 * read again.  The values read from disk will then be compared to the values
 * that were generated in memory.
 * 
 * This approach has been taken because the formatting of the variables in a BPF
 * retrieved from the CPS does not match the variable formatting performed by
 * the ODL library which the ias_bpf_write() function uses.  For example a value
 * of 0.0000 in a BPF retrieved from the CPS will be represented simply as 0
 * when using the ODL library.
 * 
 * In addition, rounding differences for double values can result in slightly
 * different numbers being stored in the ODL file.  Therefore, attempting to use
 * diff to compare two files isn't appropriate.
 */
Ensure test_write_tirs_bpf()
{
    const char kBpfName[] = "./tmpTIRSBpfWriteTest.odl";

    /* Pointers to data structures that contain returned data read from the
       temporary file. */
    const struct IAS_BPF_BIAS_MODEL       *read_bias_model = NULL;
    const struct IAS_BPF_FILE_ATTRIBUTES  *read_file_atts = NULL;
    const struct IAS_BPF_ORBIT_PARAMETERS *read_orbit_params = NULL;

    /* Data structures containing data written to the temporary file. */
    struct IAS_BPF_FILE_ATTRIBUTES    file_atts;
    struct IAS_BPF_ORBIT_PARAMETERS   orbit_params;
    struct IAS_BPF_THERMAL_BIAS_MODEL thermal;

    IAS_BPF   *bpf = NULL;
    int       band_list[IAS_MAX_NBANDS];
    int       band_index;
    int       band_number;
    int       number_of_bands;
    int       detector;
    int       i;
    int       index;
    int       sca;
    int       status;
    struct    tm date;


    printf("Generating TIRS BPF to test ias_bpf_write()...\n");

    /* Get the current band list for TIRS */
    memset(band_list, 0, IAS_MAX_NBANDS * sizeof(int));
    status = ias_sat_attr_get_sensor_band_numbers(IAS_TIRS, IAS_NORMAL_BAND,
        0, band_list, IAS_MAX_NBANDS, &number_of_bands);
    assert_equal_with_message(status, SUCCESS, "Error obtaining TIRS band "
        "information");
    if (status != SUCCESS)
        return;

    /* Fill the file attributes structure */
    strftime(file_atts.effective_date_begin, IAS_L0R_DATE_LENGTH, "%FT%T",
        &date);
    strftime(file_atts.effective_date_end, IAS_L0R_DATE_LENGTH, "%FT%T",
        &date);
    strftime(file_atts.baseline_date, IAS_L0R_DATE_LENGTH, "%FT%T", &date);
    strcpy(file_atts.file_name, "LT8TestBPF");
    strcpy(file_atts.file_source, "TIRS BPF Unit Test");
    strcpy(file_atts.spacecraft_name, "Landsat_8");
    strcpy(file_atts.sensor_name, "Thermal Infrared Sensor");
    strcpy(file_atts.description, "This is the description");
    file_atts.version = 10;

    /* Fill the orbit parameter */
    orbit_params.begin_orbit_number = 5280;

    /* Fill the bias model data. */
    for (sca = 0; sca < IAS_BPF_THERMAL_MAX_NSCAS; sca++)
    {
        for (detector = 0; detector < IAS_BPF_THERMAL_MAX_NDET; detector++)
        {
            for (i = 0; i < IAS_BPF_THERMAL_VALUES; i++)
            {
                thermal.det_params[sca][detector][i] = (2.0 * i) + 1;
            }
        }
    }

    /* Create an empty BPF structure, and load it with the file attributes
       and orbit parameter information created above.  */
    bpf = ias_bpf_allocate();
    if (bpf == NULL)
    {
        assert_true_with_message(FALSE, "Error allocating BPF structure");
        return;
    }
    memset(bpf, 0, sizeof(IAS_BPF));
    memcpy(&bpf->file_attributes, &file_atts, sizeof(file_atts));
    memcpy(&bpf->orbit_parameters, &orbit_params, sizeof(orbit_params));

    /* Load the bias model information created above. */
    for (index = 0; index < number_of_bands; index++)
    {
        band_number = band_list[index];
        band_index = ias_sat_attr_convert_band_number_to_index(band_number);
        if (band_index == ERROR)
        {
            assert_true_with_message(FALSE, "Error retrieving band index "
                "corresponding to band number %d", band_number);
            ias_bpf_free(bpf);
            return;
        }

        bpf->bias_model[band_index].band_number =  band_number;
        bpf->bias_model[band_index].spectral_type =
            ias_sat_attr_get_spectral_type_from_band_number(band_number);
        if (bpf->bias_model[band_index].spectral_type == ERROR)
        {
            assert_true_with_message(FALSE, "Error retrieving spectral "
                "type for band %d", band_number);
            ias_bpf_free(bpf);
            return;
        }
        bpf->bias_model[band_index].thermal = malloc(sizeof(thermal));
        if (bpf->bias_model[band_index].thermal == NULL)
        {
            assert_true_with_message(FALSE, "Error allocating THERMAL "
                "band bias model data buffer");
            ias_bpf_free(bpf);
            return;
        }
        memcpy(bpf->bias_model[band_index].thermal, &thermal,
            sizeof(thermal));
    }

    /* Alright, our manually generated BPF is ready to be written to a file.
       Once written to the file, free all of the memory used by the BPF */
    printf("Writing BPF to %s...\n", kBpfName);
    status = ias_bpf_write(bpf, kBpfName);
    ias_bpf_free(bpf);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "ERROR writing TIRS BPF to temporary "
            "file %s", kBpfName);
        unlink(kBpfName);
        return;
    }
    bpf = NULL;

    /* Read the generated BPF back from disk into a brand new BPF structure */
    printf("Reading BPF from %s...\n", kBpfName);
    bpf = ias_bpf_read(kBpfName);
    /*unlink(kBpfName);*/
    if (bpf == NULL)
    {
        assert_true_with_message(FALSE, "ERROR reading TIRS BPF from "
            "temporary file %s", kBpfName);
        return;
    }
    read_file_atts = ias_bpf_get_file_attributes(bpf);
    assert_not_equal_with_message(read_file_atts, NULL,
        "ERROR obtaining file attributes from BPF");

    read_orbit_params = ias_bpf_get_orbit_parameters(bpf);
    assert_not_equal_with_message(read_orbit_params, NULL,
        "ERROR obtaining orbit parameters from BPF");

    read_bias_model = ias_bpf_get_bias_model(bpf);
    assert_not_equal_with_message(read_bias_model, NULL,
        "ERROR obtaining bias model from BPF");

    if ((read_file_atts == NULL) || (read_orbit_params == NULL)
        || (read_bias_model == NULL))
    {
        ias_bpf_free(bpf);
        return;
    }

    printf("Comparing data read to data written...\n");

    /* Compare the file attributes read from the file against the fake data we
       generated above.  It must all match. */
    assert_string_equal_with_message(file_atts.effective_date_begin,
        read_file_atts->effective_date_begin,
        "ERROR effective date begin mismatch");
    assert_string_equal_with_message(file_atts.effective_date_end,
        read_file_atts->effective_date_end,
        "ERROR effective date end mismatch");
    assert_string_equal_with_message(file_atts.baseline_date,
        read_file_atts->baseline_date, "ERROR baseline date mismatch");
    assert_string_equal_with_message(file_atts.file_name,
        read_file_atts->file_name, "ERROR file name mismatch");
    assert_string_equal_with_message(file_atts.file_source,
        read_file_atts->file_source, "ERROR file source mismatch");
    assert_string_equal_with_message(file_atts.spacecraft_name,
        read_file_atts->spacecraft_name, "ERROR spacecraft name mismatch");
    assert_string_equal_with_message(file_atts.sensor_name,
        read_file_atts->sensor_name, "ERROR sensor name mismatch");
    assert_string_equal_with_message(file_atts.description,
        read_file_atts->description, "ERROR description mismatch");
    assert_equal_with_message(file_atts.version, read_file_atts->version,
        "ERROR version mismatch");

    /* Compare the orbit parameters */
    assert_equal_with_message(orbit_params.begin_orbit_number,
        read_orbit_params->begin_orbit_number, "ERROR orbit number mismatch");

    /* Finally, compare all of the bias model information for each band
       against the data generated above. */
    for (index = 0; index < number_of_bands; index++)
    {
        band_number = band_list[index];
        band_index = ias_sat_attr_convert_band_number_to_index(band_number);
        if (band_index == ERROR)
        {
            assert_true_with_message(FALSE, "Error converting band number "
                "%d to its corresponding index value", band_number);
            ias_bpf_free(bpf);
            return;
        }
        assert_equal_with_message(band_number,
            bpf->bias_model[band_index].band_number,  "ERROR bias_model"
                "[%d] band number mismatch", band_index);
        assert_equal_with_message(bpf->bias_model[band_index].spectral_type,
            ias_sat_attr_get_spectral_type_from_band_number(band_number),
            "ERROR bias_model[%d] spectral type mismatch", band_index);

        for (sca = 0; sca < IAS_BPF_THERMAL_MAX_NSCAS; sca++)
        {
            int number_of_bias_model_mismatches = 0;
            for (detector = 0; detector < IAS_BPF_THERMAL_MAX_NDET;
                 detector++)
            {
                for (i = 0; i < IAS_BPF_THERMAL_VALUES; i++)
                {
                    if (bpf->bias_model[band_index].thermal
                        ->det_params[sca][detector][i]
                        != thermal.det_params[sca][detector][i])
                    {
                        number_of_bias_model_mismatches++;
                    }
                }
            }
            assert_equal_with_message(number_of_bias_model_mismatches, 0,
                "ERROR bias_model[%d] mismatches in SCA %d",
                band_index, (sca + 1));
        }
    }

    /* Clean up. */
    ias_bpf_free(bpf);
}



Ensure test_write_oli_bpf()
{
    /* Name of the temporary BPF file that is generated */
    const char kBpfName[] = "./tmpOLIBpfWriteTest.odl";

    /* These structures are filled with generated data, the values will be
       copied into a BPF structure which is then written to disk.  After the
       BPF is created, the structure is freed and the ias_bpf_read() method is
       used to re-read all of the data. */
    struct IAS_BPF_FILE_ATTRIBUTES  fileAttr;
    struct IAS_BPF_ORBIT_PARAMETERS orbitParms;
    struct IAS_BPF_PAN_BIAS_MODEL   pan;
    struct IAS_BPF_SWIR_BIAS_MODEL  swir;
    struct IAS_BPF_VNIR_BIAS_MODEL  vnir;

    /* These pointers point to the data structures that are returned when a BPF
       is read from a file.  The data that they point to is then compared to
       the data contained in the structures above to ensure the ias_bpf_read()
       function returns the same data. */
    const struct IAS_BPF_BIAS_MODEL        *pBiasModel  = NULL;
    const struct IAS_BPF_FILE_ATTRIBUTES   *pFileAttr   = NULL;
    const struct IAS_BPF_ORBIT_PARAMETERS  *pOrbitParms = NULL;

    IAS_BPF  *bpf;
    int       bandIndex;
    int       bandNum;
    int       detector;
    int       i;
    int       sca;
    struct tm date;
    time_t    timeInSecs = time(NULL);

    printf("Generating OLI BPF to test ias_bpf_write()...\n");

    /* Fill the file attributes structure */
    localtime_r(&timeInSecs, &date);
    strftime(fileAttr.effective_date_begin, IAS_L0R_DATE_LENGTH, "%FT%T",
        &date);
    strftime(fileAttr.effective_date_end, IAS_L0R_DATE_LENGTH, "%FT%T", &date);
    strftime(fileAttr.baseline_date, IAS_L0R_DATE_LENGTH, "%FT%T", &date);
    strcpy(fileAttr.file_name, "LO8TestBPF");
    strcpy(fileAttr.file_source, "OLI BPF Unit Test");
    strcpy(fileAttr.spacecraft_name, "Landsat_8");
    strcpy(fileAttr.sensor_name, "Operational Land Imager");
    strcpy(fileAttr.description, "This is the description");
    fileAttr.version = 25;

    /* Fill the orbit parameter */
    orbitParms.begin_orbit_number = 5280;


    /* Fill the data for each of the different spectral types */
    for (sca = 0; sca < IAS_MAX_NSCAS; sca++)
    {
        for (detector = 0; detector < IAS_BPF_PAN_MAX_NDET; detector++)
        {
            for (i = 0; i < IAS_BPF_PAN_VALUES; i++)
            {
                pan.det_params_even[sca][detector][i] = 1.0;
                pan.det_params_odd[sca][detector][i] = 3.0;
            }
        }
        pan.a0_coefficient_even[sca] = 2.0;
        pan.a0_coefficient_odd[sca] = 4.0;
    }

    for (sca = 0; sca < IAS_MAX_NSCAS; sca++)
    {
        for (detector = 0; detector < IAS_BPF_MS_MAX_NDET; detector++)
        {
            for (i = 0; i < IAS_BPF_SWIR_VALUES; i++)
                swir.det_params[sca][detector][i] = 5.0;
        }
        swir.a0_coefficient[sca] = 6.0;
    }

    for (sca = 0; sca < IAS_MAX_NSCAS; sca++)
    {
        for (detector = 0; detector < IAS_BPF_MS_MAX_NDET; detector++)
        {
            for (i = 0; i < IAS_BPF_VNIR_VALUES; i++)
                vnir.det_params[sca][detector][i] = 7.0;
        }
        vnir.a0_coefficient[sca] = 8.0;
    }

    /* Now we'll create a BPF structure and load it will all of the data that
       was generated above.  The bias model information is duplicated multiple
       times for bands of the same spectral type. */
    bpf = ias_bpf_allocate();
    memcpy(&bpf->file_attributes, &fileAttr, sizeof(fileAttr));
    memcpy(&bpf->orbit_parameters, &orbitParms, sizeof(orbitParms));
    for (bandIndex = 0; bandIndex < IAS_BPF_OLI_NBANDS; bandIndex++)
    {
        /* Convert the band index into a band number */
        bandNum = ias_sat_attr_convert_band_index_to_number(bandIndex);
        if (bandNum == ERROR)
        {
            assert_true_with_message(FALSE,
                "ERROR received from satellite attributes library");
            return;
        }

        /* Populate the data for the current band */
        bpf->bias_model[bandIndex].band_number = bandNum;
        bpf->bias_model[bandIndex].spectral_type =
            ias_sat_attr_get_spectral_type_from_band_number(bandNum);
        switch (bpf->bias_model[bandIndex].spectral_type)
        {
            case IAS_SPECTRAL_PAN:
                bpf->bias_model[bandIndex].pan = malloc(sizeof(pan));
                if (bpf->bias_model[bandIndex].pan == NULL)
                {
                    assert_true_with_message(FALSE, "ERROR allocating memory");
                    return;
                }
                memcpy(bpf->bias_model[bandIndex].pan, &pan, sizeof(pan));
                break;

            case IAS_SPECTRAL_SWIR:
                bpf->bias_model[bandIndex].swir = malloc(sizeof(swir));
                if (bpf->bias_model[bandIndex].swir == NULL)
                {
                    assert_true_with_message(FALSE, "ERROR allocating memory");
                    return;
                }
                memcpy(bpf->bias_model[bandIndex].swir, &swir, sizeof(swir));
                break;

            case IAS_SPECTRAL_VNIR:
                bpf->bias_model[bandIndex].vnir = malloc(sizeof(vnir));
                if (bpf->bias_model[bandIndex].vnir == NULL)
                {
                    assert_true_with_message(FALSE, "ERROR allocating memory");
                    return;
                }
                memcpy(bpf->bias_model[bandIndex].vnir, &vnir, sizeof(vnir));
                break;

            default:
                assert_true_with_message(FALSE,
                    "ERROR: unexpected spectral type encountered for band %d",
                    bandNum);
                return;
        }
    }

    /* Alright, our manually generated BPF is ready to be written to a file.
       Once written to the file, free all of the memory used by the BPF */
    printf("Writing BPF to %s...\n", kBpfName);
    if (ias_bpf_write(bpf, kBpfName) != SUCCESS)
    {
        assert_true_with_message(FALSE,
            "ERROR writing BPF to temporary file %s", kBpfName);
        unlink(kBpfName);
        return;
    }
    ias_bpf_free(bpf);

    /* Read the generated BPF back from disk into a brand new BPF structure */
    printf("Reading BPF from %s...\n", kBpfName);
    bpf = ias_bpf_read(kBpfName);
    unlink(kBpfName);
    if (bpf == NULL)
    {
        assert_true_with_message(FALSE, "ERROR reading BPF from file %s",
            kBpfName);
        return;
    }
    pFileAttr = ias_bpf_get_file_attributes(bpf);
    assert_not_equal_with_message(pFileAttr, NULL,
        "ERROR obtaining file attributes from BPF");

    pOrbitParms = ias_bpf_get_orbit_parameters(bpf);
    assert_not_equal_with_message(pOrbitParms, NULL,
        "ERROR obtaining orbit parameters from BPF");

    pBiasModel = ias_bpf_get_bias_model(bpf);
    assert_not_equal_with_message(pOrbitParms, NULL,
        "ERROR obtaining bias model from BPF");

    if ((pFileAttr == NULL) || (pOrbitParms == NULL) || (pBiasModel == NULL))
    {
        ias_bpf_free(bpf);
        return;
    }

    printf("Comparing data written to data read...\n");

    /* Compare the file attributes read from the file against the fake data we
       generated above.  It must all match. */
    assert_string_equal_with_message(fileAttr.effective_date_begin,
        pFileAttr->effective_date_begin, "ERROR effective date begin mismatch");
    assert_string_equal_with_message(fileAttr.effective_date_end,
        pFileAttr->effective_date_end, "ERROR effective date end mismatch");
    assert_string_equal_with_message(fileAttr.baseline_date,
        pFileAttr->baseline_date, "ERROR baseline date mismatch");
    assert_string_equal_with_message(fileAttr.file_name,
        pFileAttr->file_name, "ERROR file name mismatch");
    assert_string_equal_with_message(fileAttr.file_source,
        pFileAttr->file_source, "ERROR file source mismatch");
    assert_string_equal_with_message(fileAttr.spacecraft_name,
        pFileAttr->spacecraft_name, "ERROR spacecraft name mismatch");
    assert_string_equal_with_message(fileAttr.sensor_name,
        pFileAttr->sensor_name, "ERROR sensor name mismatch");
    assert_string_equal_with_message(fileAttr.description,
        pFileAttr->description, "ERROR description mismatch");
    assert_equal_with_message(fileAttr.version, pFileAttr->version,
        "ERROR description mismatch");

    /* Compare the orbit parameter */
    assert_equal_with_message(orbitParms.begin_orbit_number,
        pOrbitParms->begin_orbit_number, "ERROR orbit number mismatch");

    /* Finally, compare all of the bias model information for each band against
       the data generated above. */
    for (bandIndex = 0; bandIndex < IAS_BPF_OLI_NBANDS; bandIndex++)
    {
        bandNum = ias_sat_attr_convert_band_index_to_number(bandIndex);
        if (bandNum == ERROR)
        {
            assert_true_with_message(FALSE,
                "ERROR received from satellite attributes library");
            ias_bpf_free(bpf);
            return;
        }

        assert_equal_with_message(bpf->bias_model[bandIndex].band_number,
            bandNum, "ERROR bias_model[%d] band number mismatch", bandIndex);
        assert_equal_with_message(bpf->bias_model[bandIndex].spectral_type,
            ias_sat_attr_get_spectral_type_from_band_number(bandNum),
            "ERROR bias_model[%d] spectral type mismatch", bandIndex);

        if (bpf->bias_model[bandIndex].spectral_type == IAS_SPECTRAL_PAN)
        {
            for (sca = 0; sca < IAS_MAX_NSCAS; sca++)
            {
                for (detector = 0; detector < IAS_BPF_PAN_MAX_NDET; detector++)
                {
                    for (i = 0; i < IAS_BPF_PAN_VALUES; i++)
                    {
                        if ((bpf->bias_model[bandIndex].pan->
                            det_params_even[sca][detector][i] !=
                            pan.det_params_even[sca][detector][i]) ||
                            (bpf->bias_model[bandIndex].pan->
                            det_params_odd[sca][detector][i] !=
                            pan.det_params_odd[sca][detector][i]))
                        {
                            assert_true_with_message(FALSE,
                                "ERROR Bias Model mismatch at bandIndex %d, "
                                "SCA %d, detector %d, value %d", bandIndex,
                                sca, detector, i);
                        }
                    }
                }
                assert_double_equal(
                    bpf->bias_model[bandIndex].pan->a0_coefficient_even[sca],
                    pan.a0_coefficient_even[sca]);
                assert_double_equal(
                    bpf->bias_model[bandIndex].pan->a0_coefficient_odd[sca],
                    pan.a0_coefficient_odd[sca]);
            }
        }
        else if (bpf->bias_model[bandIndex].spectral_type == IAS_SPECTRAL_SWIR)
        {
            for (sca = 0; sca < IAS_MAX_NSCAS; sca++)
            {
                for (detector = 0; detector < IAS_BPF_MS_MAX_NDET; detector++)
                {
                    for (i = 0; i < IAS_BPF_SWIR_VALUES; i++)
                    {
                        if (bpf->bias_model[bandIndex].swir->
                            det_params[sca][detector][i] !=
                            swir.det_params[sca][detector][i])
                        {
                            assert_true_with_message(FALSE,
                                "ERROR Bias Model mismatch at bandIndex %d, "
                                "SCA %d, detector %d, value %d", bandIndex,
                                sca, detector, i);
                        }
                    }
                }
                assert_double_equal(
                    bpf->bias_model[bandIndex].swir->a0_coefficient[sca],
                    swir.a0_coefficient[sca]);
            }
        }
        else if (bpf->bias_model[bandIndex].spectral_type == IAS_SPECTRAL_VNIR)
        {
            for (sca = 0; sca < IAS_MAX_NSCAS; sca++)
            {
                for (detector = 0; detector < IAS_BPF_MS_MAX_NDET; detector++)
                {
                    for (i = 0; i < IAS_BPF_VNIR_VALUES; i++)
                    {
                        if (bpf->bias_model[bandIndex].vnir->
                            det_params[sca][detector][i] !=
                            vnir.det_params[sca][detector][i])
                        {
                            assert_true_with_message(FALSE,
                                "ERROR Bias Model mismatch at bandIndex %d, "
                                "SCA %d, detector %d, value %d", bandIndex,
                                sca, detector, i);
                        }
                    }
                }
                assert_double_equal(
                    bpf->bias_model[bandIndex].vnir->a0_coefficient[sca],
                    vnir.a0_coefficient[sca]);
            }
        }
        else
        {
            assert_true_with_message(FALSE,
                "ERROR: unexpected spectral type encountered for band %d",
                bandNum);
        }
    }

    /* Clean up. */
    ias_bpf_free(bpf);
}



int main(int argc, char **argv)
{
    TestSuite *suite = create_test_suite();

    int status;

    /* Initialize the Satellite Attributes Library */
    status = ias_sat_attr_initialize(IAS_L8);
    if (status != SUCCESS)
    {
        printf("Cannot initialize Satellite Attributes Library\n");
        exit(100);
    }

    add_test(suite, test_ias_bpf_allocate);
    add_test(suite, test_read_file_attributes);
    add_test(suite, test_read_orbit_parameters);
    add_test(suite, test_read_oli_bias_model);
    add_test(suite, test_read_tirs_bias_model);
    add_test(suite, test_set_oli_bpf_parameter_values);
    add_test(suite, test_set_tirs_bpf_parameter_values);
    add_test(suite, test_write_tirs_bpf);
    add_test(suite, test_write_oli_bpf);
    
    if (argc > 1) 
        return run_single_test(suite, argv[1], create_text_reporter());
    return run_test_suite(suite, create_text_reporter());
}
