/*************************************************************************

NAME: ias_cpf_parse_cloud_cover_assessment

PURPOSE: Retrieve the cloud cover assessment data group from the
         odl tree and insert into the internal cpf strucure.

RETURN VALUE: SUCCESS or ERROR

******************************************************************************/

#include <stdlib.h>
#include <string.h>
#include "ias_cpf.h"
#include "ias_logging.h"
#include "local_defines.h"

int ias_cpf_parse_cloud_cover_assessment
(
    const IAS_CPF *cpf,             /* I: CPF structure */
    struct IAS_CPF_CLOUD_COVER_ASSESSMENT *cc_assessment
                                    /* O: CPF cloud cover params */
)
{
    int status;                     /* Status of return from function */
    char group_name[] = "CLOUD_COVER_ASSESSMENT";
                                    /* Group to retrieve from the CPF */
    IAS_OBJ_DESC *odl_tree;         /* ODL tree */
    int nelements;                  /* Number of returned attribute values */   

    /* get the odl tree */
    GET_GROUP_FROM_CACHE(cpf, group_name, odl_tree);

    /* Retrieve the Cirrus Threshold parameter */
    status = ias_odl_get_field(cc_assessment->cirrus_threshold,
        sizeof(double) * IAS_CPF_CIRRUS_THRESHOLD_SIZE, IAS_ODL_Double,
        odl_tree, group_name, "Cirrus_Threshold", &nelements);

    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving the cirrus threshold");
        DROP_ODL_TREE(odl_tree);
        return ERROR;
    }

    /* Error check to make sure either one or IAS_CPF_CIRRUS_THRESHOLD_SIZE 
       values were read for the cirrus threshold */
    if (nelements != IAS_CPF_CIRRUS_THRESHOLD_SIZE && nelements != 1)
    {
        IAS_LOG_ERROR("Read in %d values for Cirrus Threshold. "
            "1 or %d expected", nelements, IAS_CPF_CIRRUS_THRESHOLD_SIZE);
        DROP_ODL_TREE(odl_tree);
        return ERROR; 
    }

    /* If only 1 parameter was read in for the Cirrus Threshold, print a warning
       message, and set A, B, and C to default values */
    else if (nelements == 1)
    {
        IAS_LOG_WARNING("Only 1 parameter read in from the CPF "
            "for Cirrus Threshold");

        /* Set appropriate default values */
        /* Default values: A = 0, B = 0, C = 1000, D = always present */
        /* C is 1000 in order to avoid a divide by zero error later in the
           application */
        cc_assessment->cirrus_threshold[IAS_CPF_CIRRUS_THRESHOLD_SIZE-1] = 
            cc_assessment->cirrus_threshold[0];
        
        /* Set A, B, and C respectively */
        cc_assessment->cirrus_threshold[0] = 0; /* A */
        cc_assessment->cirrus_threshold[1] = 0; /* B */
        cc_assessment->cirrus_threshold[2] = 1000; /* C */
    }
    
    /* Retrieve the optional Minimum_Sun_Elevation parameter and default it to
       zero if not found */
    status = ias_odl_get_field(&cc_assessment->minimum_sun_elevation_degrees,
        sizeof(cc_assessment->minimum_sun_elevation_degrees), IAS_ODL_Double,
        odl_tree, group_name, "Minimum_Sun_Elevation", &nelements);

    /* Verify the return value and the correct number of parameters was read */
    if (status == IAS_ODL_NOT_FOUND && nelements != 1)
    {
        IAS_LOG_WARNING(
            "Minimum_Sun_Elevation not in CPF %s group, defaulting to zero",
            group_name);
        cc_assessment->minimum_sun_elevation_degrees = 0.0;
    }
    else if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving Minimum_Sun_Elevation from the ODL");
        DROP_ODL_TREE(odl_tree);
        return ERROR;
    }

    DROP_ODL_TREE(odl_tree);

    return SUCCESS;
}
