/*************************************************************************

NAME: ias_cpf_parse_framing_parameters

PURPOSE: Retrieve the framing parameters group from the odl tree and insert
    into the internal cpf strucure.  The same routine can be used for the OLI
    and TIRS groups.
        
RETURN VALUE: SUCCESS or ERROR

******************************************************************************/

#include "ias_cpf.h"
#include "ias_logging.h"
#include "local_defines.h"

int ias_cpf_parse_framing_parameters
(
    const IAS_CPF *cpf,                 /* I: CPF structure */
    IAS_SENSOR_ID sensor,               /* I: OLI/TIRS selection */
    struct IAS_CPF_FRAMING_PARAMETERS *framing_parameters
                                        /* O: framing parameters */
)
{
    int status;                     /* Status of return from function */
    int count = 0;                  /* Number of items in CPF_LIST_TYPE */
    const char *sensor_name;        /* Name of the requested sensor */
    char group_name[100];           /* Group to retrieve from the CPF */
    IAS_OBJ_DESC *odl_tree;         /* ODL tree */

    ODL_LIST_TYPE list[] =          /* Struct to contain group attributes */
    {
        {group_name, "Frames_Per_Scene", &framing_parameters->frames_per_scene,
         sizeof(framing_parameters->frames_per_scene), IAS_ODL_Int, 1},

        {group_name, "Min_Scene_Overlap_Frames", 
         &framing_parameters->min_scene_overlap_frames,
         sizeof(framing_parameters->min_scene_overlap_frames), IAS_ODL_Int, 1}
    };

    /* Get the sensor name */
    sensor_name = ias_sat_attr_get_sensor_name(sensor);

    /* Create the group name based on the sensor.  Consider it an error if
       the name isn't at least one character or needs more buffer space than
       is provided. */
    status = snprintf(group_name, sizeof(group_name), "%s_FRAMING_PARAMETERS",
        sensor_name);
    if (status <= 0 || status >= sizeof(group_name))
    {
        IAS_LOG_ERROR("Creating group name for %s framing parameters",
            sensor_name);
        return ERROR;
    }

    /* Calculate the number of attributes to retrieve */
    count = sizeof(list) / sizeof(ODL_LIST_TYPE);

    GET_GROUP_FROM_CACHE(cpf, group_name, odl_tree);

    /* Populate the list from the odl tree */
    status = ias_odl_get_field_list(odl_tree, list, count);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Getting group: %s from CPF", group_name);
        DROP_ODL_TREE(odl_tree);
        return ERROR;
    }

    DROP_ODL_TREE(odl_tree);

    return SUCCESS;
}
