#ifndef IAS_GCP_H
#define IAS_GCP_H

#include <limits.h>                      /* PATH_MAX definition */
#include <stdio.h>
#include "ias_types.h"
#include "ias_structures.h"

#define IAS_GCP_ID_SIZE 16               /* Length of a point id */
#define IAS_GCP_UPDATE_REASON_SIZE 11    /* Size of the update reason string */
#define IAS_GCP_CHIP_NAME_SIZE 19        /* Length of chip name */
#define IAS_GCP_SOURCE_SIZE 5            /* Length of control source */
#define IAS_GCP_TYPE_SIZE 11             /* Length of control type */
#define IAS_GCP_PROJECTION_SIZE 14       /* Length of control projection */
#define IAS_GCP_SEASON_LEN 7             /* Length of season */
#define IAS_GCP_NUM_SEASONS 4            /* User may enter 4 seasons */
#define IAS_GCP_NUM_CHIP_SOURCES 10      /* User may enter 10 sources */
#define IAS_GCP_DATE_FORMAT "%02d-%02d-%04d"
#define IAS_GCP_DATE_LEN ((int)sizeof("mm-dd-yyyy")) /* Length of date based on
                                            on date format */
#define IAS_GCP_ABS_REL_SIZE 4           /* Length of ABS or REL value */
#define IAS_GCP_DATA_TYPE_MAX_SIZE 10    /* Max size a string buffer for the
                                            data type needs to be */
#define IAS_GCP_SENSOR_SIZE 11           /* Size of the chip sensor field */
#define IAS_GCP_SELECT_METHOD_SIZE 11    /* Size of the select method field */
#define IAS_GCP_YN_SIZE 2                /* Size of the Y/N flag fields */

#define IAS_GCP_NO_CONVERSION 0          /* Flag to do no conversions in the
                                            chip reading routine */
#define IAS_GCP_CONVERT_TO_FLOAT 1       /* Flag to convert to float in the 
                                            chip reading routine */
#define IAS_GCP_IMAGE_CHIP_FILENAME "image_chips" /* Standard base filename 
                                            for an image chip tarball file */

typedef struct ias_gcp_record
{
    char point_id[IAS_GCP_ID_SIZE];         /* Ground control point ID */
    char point_active[IAS_GCP_YN_SIZE];     /* Y/N flag indicating the point is
                                               active (valid to use) */
    char point_date_updated[IAS_GCP_DATE_LEN]; /* Date the point info was last
                                                  updated (mm-dd-yyyy) */
    char update_reason[IAS_GCP_UPDATE_REASON_SIZE]; 
                                            /* Reason for the last update */
    int point_version;                      /* GCP point version number */
    int point_wrs_path;                     /* GCP WRS path */
    int point_wrs_row;                      /* GCP WRS row */
    char source[IAS_GCP_SOURCE_SIZE];       /* GCP source */
    double latitude;                        /* True latitude of reference point
                                               in degrees */
    double longitude;                       /* True longitude of reference point
                                               in degrees */
    double elevation;                       /* Elevation at the GCP in meters */
    char chip_type[IAS_GCP_TYPE_SIZE];      /* GCP type */
    char chip_name[IAS_GCP_CHIP_NAME_SIZE]; /* Name of the chip image */
    char chip_active[IAS_GCP_YN_SIZE];      /* Y/N flag indicating the chip is
                                               active (valid to use) */
    char chip_date_updated[IAS_GCP_DATE_LEN]; /* Date the image chip was last
                                                 updated (mm-dd-yyyy) */
    int chip_version;                       /* image chip version number */
    int chip_wrs_path;                      /* image chip WRS path */
    int chip_wrs_row;                       /* image chip WRS row */
    double reference_line;                  /* Reference line */
    double reference_sample;                /* Reference sample */
    double projection_x;                    /* Projection x value of reference
                                               point in meters */
    double projection_y;                    /* Projection y value of reference
                                               point in meters */
    double pixel_size_x;                    /* Pixel size x value in meters */
    double pixel_size_y;                    /* Pixel size y value in meters */
    double chip_size_lines;                 /* Chip size in lines */
    double chip_size_samples;               /* Chip size in samples */
    char chip_sensor[IAS_GCP_SENSOR_SIZE];  /* Sensor that provided the 
                                               image chip pixels */
    char chip_select_method[IAS_GCP_SELECT_METHOD_SIZE];
                                    /* Method used to select the image chip */
    int moravec_rank;                       /* Moravec rank of the chip if
                                               the select method is MORAVEC */
    char projection[IAS_GCP_PROJECTION_SIZE]; /* Chip projection */ 
    int zone;                               /* Projection UTM zone */
    char absolute_or_relative[IAS_GCP_ABS_REL_SIZE]; /* ABS or REL */
    char date[IAS_GCP_DATE_LEN];            /* Acquisition date form 
                                               mm-dd-yyyy */
    IAS_DATA_TYPE chip_data_type;           /* Image chip data type */
} IAS_GCP_RECORD;

typedef struct ias_gcp_results
{
    char point_id[IAS_GCP_ID_SIZE];     /* Ground Control Point ID */
    double reference_line;              /* Reference line */
    double reference_sample;            /* Reference sample */
    double latitude;                    /* True latitude */
    double longitude;                   /* True longitude */
    double elevation;                   /* Ground Control Point elevation */
    IAS_VECTOR gcp_position;            /* Satellite position in cartesian 
                                           space */
    double predicted_search_line;       /* Predicted line on the 1G image */
    double predicted_search_sample;     /* Predicted sample on the 1G image */
    double fit_line_offset;             /* Best-fit vertical (line)
                                           offsets of correlation peak */
    double fit_sample_offset;           /* Best-fit horizontal (sample) 
                                           offsets of correlation peak */
    int accept_flag;                    /* Acceptance flag of the ground 
                                           control point */
    double correlation_coefficient;     /* Correlation coefficient */
    int search_band;                    /* Band number of the search image */
    int search_sca;                     /* SCA number, 1 for SCA-combined 
                                           images */
    char chip_source[IAS_GCP_SOURCE_SIZE]; /* The chip source (DOQ, GLS) */
    int reference_band;                 /* Reference image band number */
    int reference_sca;                  /* Reference SCA number */
    char reference_name[PATH_MAX];      /* Reference image name with path */
    char search_name[PATH_MAX];         /* Full search image name with path */
    IAS_DATA_TYPE chip_data_type;       /* Image chip data type */
} IAS_GCP_RESULTS;

/* Forward declaration to the IAS_GCP_IMAGE_CHIP_IO type to keep the definition
   hidden outside the library */
typedef struct ias_gcp_image_chip_io IAS_GCP_IMAGE_CHIP_IO;

/* Define the supported types of containers for image chips */
typedef enum ias_gcp_container_type
{
    IAS_GCP_CONTAINER_DIR,
    IAS_GCP_CONTAINER_TARBALL,
    IAS_GCP_CONTAINER_SEARCH
} IAS_GCP_CONTAINER_TYPE;

int ias_gcp_read_gcplib
(
    const char *gcplib_file_name, /* I: Name of the GCPLIB file */
    IAS_GCP_RECORD **gcp_lib,     /* O: Structure of chip information */
    int *num_gcp,                 /* O: Number of ground control points */
    int *overall_gcp_version      /* O: Overal GCP version number */

);

int ias_gcp_read_gcplib_filtered
(
    const char *gcplib_file_name, /* I: Name of the GCPLIB file */
    /* For dates, 1 = January, etc. Year is YYYY */
    const int *begin_date,        /* I: Beginning date [0] = month [1] = year */
    const int *end_date,          /* I: Ending date [0] = month [1] = year */
                                  /* I: Season of chip */
    char season[IAS_GCP_NUM_SEASONS][IAS_GCP_SEASON_LEN],
                                  /* I: Source of chip */
    char chip_source[IAS_GCP_NUM_CHIP_SOURCES][IAS_GCP_SOURCE_SIZE],
    const char *chip_type,        /* I: Type of chip */
    IAS_GCP_RECORD **gcp_lib,     /* O: Structure of chip information */
    int *num_gcp,                 /* O: Number of ground control points */
    int *overall_gcp_version      /* O: Overal GCP version number */

);

int ias_gcp_write_gcplib
(
    const char *gcplib_filename,       /* I: Output GCP filename */
    const IAS_GCP_RECORD *gcp_records, /* I: Structure of chip information */
    int number_of_gcps,                /* I: Number of ground control points */
    int overall_gcp_version            /* I: Overall GCP version */
);

int ias_gcp_read_correlation_results
(
    const char *gcp_data_filename,  /* I: GCP data file name */
    const double semi_major_axis,   /* I: CPF earth semi major axis constant */
    const double ellipticity,       /* I: CPF earth ellipticity constant */
    IAS_GCP_RESULTS **gcp_data,     /* O: GCP data records */
    int *num_gcp                    /* O: Number of GCPs read */
);

int ias_gcp_write_correlation_results 
(
    const char *gcp_data_filename,    /* I: Name of GCP data file to add to */
    const IAS_GCP_RESULTS gcp_data[], /* I: GCP data structures */
    int num_gcp                       /* I: Number of GCPs */
);

IAS_GCP_IMAGE_CHIP_IO *ias_gcp_open_image_chip_container
(
    const char *container_name,     /* I: Name of container to open */
    IAS_GCP_CONTAINER_TYPE type,    /* I: Type of container */
    IAS_ACCESS_MODE access_mode     /* I: Access mode to use when opening */
);

int ias_gcp_close_image_chip_container
(
    IAS_GCP_IMAGE_CHIP_IO *chip_io  /* I: Chip container to close */
);

const char *ias_gcp_get_image_chip_container_name
(
    const IAS_GCP_IMAGE_CHIP_IO *chip_io  /* I: Chip container to close */
);

IAS_GCP_CONTAINER_TYPE ias_gcp_get_image_chip_container_type
(
    const IAS_GCP_IMAGE_CHIP_IO *chip_io  /* I: Chip container to close */
);

int ias_gcp_read_image_chip
(
    const IAS_GCP_IMAGE_CHIP_IO *chip_io, /* I: Chip container to read */
    const char *chip_name,          /* I: Image chip name to read */
    int chip_size_lines,            /* I: Expected number of chip lines */
    int chip_size_samples,          /* I: Expected number of chip samples */
    IAS_DATA_TYPE chip_data_type,   /* I: Image chip data type */
    int convert_output_to_float,    /* I: flag to convert the output to
                                          floating point (1 = convert to float,
                                          0 = no conversion) */
    void *output_chip_buffer        /* O: Buffer of chip image data converted
                                          to float (R*4) if requested, 
                                          otherwise the same type contained in
                                          the file */
);

int ias_gcp_write_image_chip
(
    IAS_GCP_IMAGE_CHIP_IO *chip_io, /* I/O: Chip container to write */
    const char *chip_name,      /* I: Name of the image chip file */
    int chip_size_lines,        /* I: Number of chip lines in data buffer */
    int chip_size_samples,      /* I: Number of chip samples in data buffer */
    IAS_DATA_TYPE dtype,        /* I: Data type for byte swapping if needed */
    const void *buf             /* I: Image chip data buffer to write to file */
);

#endif
