package ias_gcp;

use strict;
use warnings;
use Logger;                 # Logging routines
use ias_const;              # Defines SUCCESS and ERROR
use Switch;                 # Allows for use of switch statements
use Archive::Tar;           # Provides Tarball functionality
use Time::Local;            # Getting current date/time
use Digest::MD5 qw(md5_hex);# MD5 to verify image chips
use File::Basename;         # Basename routine

# Define Constants 
use constant IAS_GCP_BZIP2 => 0;    # BZIP2 Tarball Compression method
use constant IAS_GCP_GZIP => 1;     # GZIP  Tarball Compression method
use constant IAS_GCP_NO_COMPRESSION => 2; # Do not compress the tarball
use constant IAS_GCP_IMAGE_CHIPS_FILENAME => 'image_chips'; # Default filename

# Flag to not generate MD5 sums for image chips when creating the tarball
use constant IAS_GCP_NO_MD5SUM => 0;
# Flag to generate MD5 sums for image chips when creating the tarball
use constant IAS_GCP_DO_MD5SUM => 1;

our (@ISA, @EXPORT_OK);
BEGIN
{
    require Exporter;
    @ISA = qw(Exporter);

    # Symbols to export on request
    @EXPORT_OK = qw(    IAS_GCP_IMAGE_CHIPS_FILENAME
                        IAS_GCP_BZIP2
                        IAS_GCP_GZIP
                        IAS_GCP_NO_COMPRESSION
                        IAS_GCP_NO_MD5SUM
                        IAS_GCP_DO_MD5SUM
                        ias_gcp_read_gcplib 
                        ias_gcp_write_gcplib
                        ias_gcp_get_sizeof_data_type                 
                        ias_gcp_tar_image_chips
                        ias_gcp_create_tar_filename
                    );
}

#############################################################################
# 
# NAME:    ias_gcp_read_gcplib
# 
# PURPOSE: Read GCP metadata from disk. 
# 
# Parameters:
#   - GCPLib filename
#
# RETURNS: SUCCESS, Overall GCP Version, and GCP data structure on success
#          ERROR, undef, and undef on error
# 
#############################################################################
sub ias_gcp_read_gcplib
{
    my $gcp_filename = shift;              # GCP data file name
    my $overall_gcp_version;               # Overall GCP Version
    my @gcp_data;                          # array of hashes

    # Define the list of fields to read from the file and the order that they
    # should be in
    my @fields = qw(point_id point_active point_date_updated update_reason
        point_version point_path point_row source latitude longitude elevation
        point_type chip_name chip_active chip_date_updated chip_version
        chip_path chip_row reference_line reference_sample projection_x
        projection_y pixel_size_x pixel_size_y chip_size_lines
        chip_size_samples chip_sensor chip_select_method moravec_rank
        projection zone abs_rel_flag chip_acq_date chip_data_type);

    # Get the default logger object from the library 
    my $error_log = get_default_log();

    # Verify filename is defined and is not blank 
    if ((!defined($gcp_filename)) || ($gcp_filename eq ""))
    {
        $error_log->fatal_msg("Empty Filename");
        return ERROR, undef, undef;
    }
    
    # Open GCPLIB file
    my $status = open GCP_FILE, "$gcp_filename";
    if (!defined($status))
    {
        $error_log->fatal_msg("Opening $gcp_filename");
        return ERROR, undef, undef;
    }

    my $expected_record_count = 0;
    my $found_begin = 0;
    my $line_number = 0;
    while (my $line = <GCP_FILE>)
    {
        my $gcp_record = {};
        chomp $line;

        $line_number++;

        # Skip comments
        next if $line =~ /^#/;

        # Look for the Version data start marker
        if ($line =~ /^VERSION\s*=\s*(\d+)\s*$/)
        {
            # Get overall GCP version number
            $overall_gcp_version = $1;
            next;
        }

        if ($line =~ /^BEGIN$/)
        {
            # Read the next line for the record count
            $line = <GCP_FILE>;
            $line_number += 1;
            chomp $line;
            if ($line =~ /^\d+$/)
            {
                $expected_record_count = $line;
                $found_begin = 1;
                next;
            }
            else
            {
                # Invalid record count line
                $error_log->fatal_msg("Invalid record count following BEGIN "
                    . "in $gcp_filename");
                close GCP_FILE;
                return ERROR, undef, undef;
            }
        }

        # Verify the begin marker was found.  If it hasn't, there is something
        # wrong with the file
        if ($found_begin != 1)
        {
            $error_log->fatal_msg("BEGIN record not found before records start"
                . " in $gcp_filename");
            close GCP_FILE;
            return ERROR, undef, undef;
        }

        # Split the read line on whitespace
        my @values = split(/\s/, $line);

        # Verify the correct number of values were found
        if (scalar @fields != scalar @values)
        {
            $error_log->fatal_msg("$gcp_filename has an invalid number of "
                . "fields on line number $line_number : $line");
            close GCP_FILE;
            return ERROR, undef, undef;
        }

        # Assign the values to the corresponding key in the record
        for (my $index = 0; $index <= $#fields; $index++)
        {
            my $field = $fields[$index];
            $gcp_record->{$field} = $values[$index];
        }

        # Verify the GCP entries
        $status = check_gcp($gcp_record, $error_log);
        if ($status != SUCCESS)
        {
            $error_log->fatal_msg("Checking GCP record from line number "
                . "$line_number in $gcp_filename");
            close GCP_FILE;
            return ERROR, undef, undef;
        }

        # Add gcp_record hash to gcp_data array
        push @gcp_data, $gcp_record;
    }

    close GCP_FILE;

    # Verify gcp_data array of hashes has the number of expected records
    my $total = $#gcp_data + 1;
    if ($total != $expected_record_count)
    {
        $error_log->fatal_msg("Expected $expected_record_count GCP records in "
            . "$gcp_filename, but found $total");
        return ERROR, undef, undef;
    }
    
    # Verify that the version was found in the GCPLib file
    if (!defined $overall_gcp_version)
    {
        $error_log->fatal_msg("Version not found in GCPLib file $gcp_filename");
        return ERROR, undef, undef;
    }

    return SUCCESS, $overall_gcp_version, @gcp_data;
}

#############################################################################
# 
# NAME:    ias_gcp_write_gcplib
# 
# PURPOSE: The GCP Retrieval Module Detail Design describes a Perl script
#          that acquires GCP data. This Perl module writes the GCP 
#          metadata to disk. 
# 
# Parameters:
#   - GCPLib filename
#   - Overall GCP version number to write to the file
#   - Reference to an array of GCPLib records to write
#
# RETURNS: SUCCESS/ERROR
# 
#############################################################################
sub ias_gcp_write_gcplib
{
    my $gcp_filename = shift;              # GCP data file name
    my $overall_gcp_version = shift;       # Overall GCP version
    my $gcp_data_reference = shift;        # Array reference

    # Get the default logger object from the library 
    my $error_log = get_default_log();

    # Verify gcp_data_reference is defined
    if (!defined($gcp_data_reference))
    {
        $error_log->fatal_msg("gcp_data_reference is not defined");
        return ERROR;
    }

    # Verify filename is defined and is not blank 
    if ((!defined($gcp_filename)) || ($gcp_filename eq ""))
    {
        $error_log->fatal_msg("Empty Filename");
        return ERROR;
    }

    # Verify that the gcp version is defined
    if (!defined $overall_gcp_version)
    {
        $error_log->fatal_msg("overall_gcp_version in not defined");
        return ERROR;
    }

    my @gcp_data = @{$gcp_data_reference}; # Array of hashes containing GCP data

    # Verify gcp_data array of hashes has at least one record (hash)
    my $total = $#gcp_data + 1;
    if ($total == 0)
    {
        $error_log->warn_msg("No GCPLIB records");
    }

    # Open the file
    my $status = open OUT, ">$gcp_filename";
    if (!defined($status))
    {
        $error_log->fatal_msg("Opening $gcp_filename");
        return ERROR;
    }

    # Print the header 
    $status = print OUT "# Unique Ground Control Point (GCP) ID - integer\n"
        . "# GCP Active Flag - string (Y/N)\n"
        . "# GCP date updated - string in form mm-dd-yyyy\n"
        . "# GCP update reason - string\n"
        . "# GCP version number - integer\n"
        . "# GCP WRS-2 path - integer\n"
        . "# GCP WRS-2 row - integer\n"
        . "# GCP source - string\n"
        . "# GCP latitude of reference point in degrees - double\n"
        . "# GCP longitude of reference point in degrees - double\n"
        . "# GCP elevation of reference point in meters - double\n"
        . "# GCP type -string (CONTROL or VALIDATION)\n"
        . "# Image chip ID - string\n"
        . "# Image chip Active Flag - string (Y/N)\n"
        . "# Image chip date updated - string in form mm-dd-yyyy\n"
        . "# Image chip version number - integer\n"
        . "# Image chip source WRS-2 path - integer\n"
        . "# Image chip source WRS-2 row - integer\n"
        . "# Image chip reference line - double\n"
        . "# Image chip reference sample - double\n"
        . "# Image chip reference point projection X in meters - double\n"
        . "# Image chip reference point projection Y in meters - double\n"
        . "# Image chip pixel size x value in meters - double\n"
        . "# Image chip pixel size y value in meters - double\n"
        . "# Image chip size in lines - double\n"
        . "# Image chip size in samples - double\n"
        . "# Image chip sensor - string\n"
        . "# Image chip selection method - string\n"
        . "# Image chip Moravec rank if selection method is MORAVEC, "
                . "zero otherwise - integer\n"
        . "# Image chip projection - string\n"
        . "# Image chip projection UTM zone - integer\n"
        . "# Image chip type - string ABS (Absolute) or REL (relative)\n"
        . "# Image chip acquisition date - string in form mm-dd-yyyy\n"
        . "# Image chip data type - string\n"
        . "VERSION = ".$overall_gcp_version."\n";
    if ($status == 0)
    {
        $error_log->fatal_msg("Writing header comment to $gcp_filename");
        close OUT;
        return ERROR;
    }

    # Print the data start marker
    $status = print OUT "BEGIN\n";
    if ($status == 0)
    {
        $error_log->fatal_msg("Writing BEGIN to $gcp_filename");
        close OUT;
        return ERROR;
    }

    # Print the number of GCPs
    $status = print OUT "$total\n";
    if ($status == 0)
    {
        $error_log->fatal_msg("Writing number of GCPs to $gcp_filename");
        close OUT;
        return ERROR;
    }

    for my $gcp_record ( @gcp_data )
    {
        # Verify the GCP entries
        $status = check_gcp($gcp_record, $error_log);
        if ($status != SUCCESS)
        {
            $error_log->fatal_msg("Found while checking GCPs in $gcp_filename");
            close OUT;
            return ERROR;
        }

        # Print out the values to the file, start with GCP items
        my $status = printf OUT "%s %s %s %s %d %d %d %s %.7f %.7f %.3f %s "
                # then image chip ID through chip size
                . "%s %s %s %d %d %d %lf %lf %.3f %.3f %lf %lf %lf %lf "
                # and chip sensor to the end
                . "%s %s %d %s %d %s %s %s\n",
                # GCP related fields
                $gcp_record->{'point_id'},
                $gcp_record->{'point_active'},
                $gcp_record->{'point_date_updated'},
                $gcp_record->{'update_reason'},
                $gcp_record->{'point_version'},
                $gcp_record->{'point_path'},
                $gcp_record->{'point_row'},
                $gcp_record->{'source'},
                $gcp_record->{'latitude'},
                $gcp_record->{'longitude'},
                $gcp_record->{'elevation'},
                $gcp_record->{'point_type'},
                # Start of image chip information
                $gcp_record->{'chip_name'},
                $gcp_record->{'chip_active'},
                $gcp_record->{'chip_date_updated'},
                $gcp_record->{'chip_version'},
                $gcp_record->{'chip_path'},
                $gcp_record->{'chip_row'},
                $gcp_record->{'reference_line'},
                $gcp_record->{'reference_sample'},
                $gcp_record->{'projection_x'},
                $gcp_record->{'projection_y'},
                $gcp_record->{'pixel_size_x'},
                $gcp_record->{'pixel_size_y'},
                $gcp_record->{'chip_size_lines'},
                $gcp_record->{'chip_size_samples'},
                $gcp_record->{'chip_sensor'},
                $gcp_record->{'chip_select_method'},
                $gcp_record->{'moravec_rank'},
                $gcp_record->{'projection'},
                $gcp_record->{'zone'},
                $gcp_record->{'abs_rel_flag'},
                $gcp_record->{'chip_acq_date'},
                $gcp_record->{'chip_data_type'};
        if ($status == 0)
        {
            $error_log->fatal_msg("Writing values to $gcp_filename");
            close OUT;
            return ERROR;
        }
    }

    if (!close OUT)
    {
        $error_log->fatal_msg("Closing $gcp_filename");
        return ERROR;
    }

    return SUCCESS;
}

#############################################################################
# 
# NAME: check_gcp
# 
# PURPOSE: Does rudamentary checks on the GCP fields to make sure they 
#   are reasonable.
# 
# RETURNS: SUCCESS or ERROR
# 
#############################################################################
sub check_gcp
{
    my $gcp_record = shift;
    my $error_log = shift;
    
    # Set up the lists hash members that should be strings, integers,
    # doubles, and dates
    my @string_list = qw(point_active update_reason source chip_active
        chip_name point_type chip_sensor chip_select_method projection
        chip_data_type abs_rel_flag);
    my @int_list = qw(point_id point_version point_path point_row chip_version
        chip_path chip_row zone moravec_rank);
    my @double_list = qw(latitude longitude elevation reference_line
        reference_sample projection_x projection_y pixel_size_x pixel_size_y
        chip_size_lines chip_size_samples);
    my @date_list = qw(point_date_updated chip_date_updated chip_acq_date);

    # Cache the GCP ID for error messages
    my $gcp_id = $gcp_record->{'point_id'};

    # Check the integer fields to make sure they are defined and integers
    for my $field (@int_list)
    {
        if ((!defined $gcp_record->{$field})
            || $gcp_record->{$field} !~ /^-?\d+$/)
        {
            my $val = "'undefined'";
            $val = $gcp_record->{$field} if (defined $gcp_record->{$field});

            $error_log->fatal_msg("GCP ID $gcp_id $field has a value of "
                . "$val when it should be an integer");
            return ERROR;
        }
    }

    # Check the double fields to make sure they are defined and doubles.
    # Note that integer values are accepted as doubles since the database
    # queries only include the decimal point when needed.
    for my $field (@double_list)
    {
        if ((!defined $gcp_record->{$field})
            || $gcp_record->{$field} !~ /^-?\d+\.?\d*$/)
        {
            my $val = "'undefined'";
            $val = $gcp_record->{$field} if (defined $gcp_record->{$field});

            $error_log->fatal_msg("GCP ID $gcp_id $field has a value of "
                . "$val when it should be a double");
            return ERROR;
        }
    }

    # Check the string fields to make sure they are defined and have a value
    for my $field (@string_list)
    {
        if ((!defined $gcp_record->{$field})
            || $gcp_record->{$field} !~ /^\S+$/)
        {
            $error_log->fatal_msg("GCP ID $gcp_id does not have a value for "
                . $field);
            return ERROR;
        }
        # Do an additional check for active flags to make sure they are Y/N
        elsif ($field =~ /_active$/
            && $gcp_record->{$field} !~ /^[YN]$/)
        {
            $error_log->fatal_msg("GCP ID $gcp_id $field has a value of "
                . $gcp_record->{$field} ." when it should be 'Y' or 'N'");
            return ERROR;
        }

    }

    # Check the date fields to make sure they are defined and have the correct
    # format
    for my $field (@date_list)
    {
        if ((!defined $gcp_record->{$field})
            || $gcp_record->{$field} !~ /^\d{2}-\d{2}-\d{4}$/)
        {
            my $val = "'undefined'";
            $val = $gcp_record->{$field} if (defined $gcp_record->{$field});

            $error_log->fatal_msg("GCP ID $gcp_id $field has a value of "
                . "$val when it should be a date in mm-dd-yyyy format");
            return ERROR;
        }
    }

    # To double-check this routine to make sure all the fields were checked
    # the number of keys in the hash should match the sum of the list sizes
    my $list_size = scalar @string_list + scalar @int_list
          + scalar @double_list + scalar @date_list;
    if ($list_size != scalar keys %$gcp_record)
    {
        # This is a programming error
        $error_log->fatal_msg("Not all GCP record entries are being checked");
        return ERROR;
    }

    return SUCCESS;
}

#############################################################################
# 
# NAME: ias_gcp_get_sizeof_data_type
# 
# PURPOSE:  Using the passed in data type string returns the number of bytes
#           for that data type.
#
# RETURNS: Data type size / ERROR
# 
#############################################################################
sub ias_gcp_get_sizeof_data_type
{
    # Using passed string return data type size
    switch (uc(shift))
    {
        case "BYTE" { return 1; }
        case "CHAR" { return 1; }
        case "I*2"  { return 2; }
        case "UI*2" { return 2; }
        case "I*4"  { return 4; }
        case "UI*4" { return 4; }
        case "R*4"  { return 4; }
        case "R*8"  { return 8; }
        else        { return ERROR; }
    }
}

##############################################################################
# Name: ias_gcp_tar_image_chips
#
# Description: Will use GCP record array with attached image chip
#              data and package up the image chips into a tarball
#
# Parameters:
#   - GCP records Array reference
#   - Tarball filename
#   - Compression Type (BZIP2, GZIP, or NO_COMPRESSION)
#   - MD5Sum Flag (Determines if md5sum is run)
#
# Returns:
#   SUCCESS - on successful completion
#   ERROR   - if error occurs during execution
##############################################################################
sub ias_gcp_tar_image_chips
{
    my $gcp_records = shift;        # List of files to tar up
    my $filename = shift;           # Tarball filename
    my $compression = shift;        # Tar Compression Type
    my $md5sum_flag = shift;        # MD5Sum Flag

    # Set up needed variables
    my $logger = get_default_log(); # Get the default logger   
    my $tar = Archive::Tar->new;    # Create the new tarball
    my $status;                     # Status holder
 
    # Create the md5 file for image chips
    if ($md5sum_flag)
    {
        my ($base, $directory, $extension) = fileparse($filename, qr/\..*/);
        my $md5sum_filename = $directory . $base . ".md5";

        $status = open(MD5FILE, ">" , $md5sum_filename);
        if (!defined $status)
        {
            $logger->fatal_msg("Unable to open MD5 file $md5sum_filename");
            return ERROR;
        }
    } 

    # Add each filename and chip data to tarball
    foreach my $chip (@{$gcp_records})
    {
        # Reformat the date so it works with tar
        my @date = split('-', $chip->{'chip_date_updated'});
        my $tar_date = timelocal(0, 0, 0, $date[1], $date[0] - 1, $date[2]);

        # Create the options used for the current file
        my $options = {gid => 0, uid => 0, uname => "0", gname => "0",
            mode => 0644, mtime => $tar_date};

        # Add the file to the tarball
        $status = $tar->add_data($chip->{'chip_name'}, $chip->{'image_chip'},
            $options);
        if (!defined $status)
        {
            $logger->fatal_msg("Unable to add $chip->{'filename'} to tarball");
            return ERROR;
        }

        # Calculate MD5Sum and add to record
        if ($md5sum_flag)
        {
            $chip->{'MD5'} = md5_hex($chip->{'image_chip'});

            # Print the key to to the MD5 file
            $status = print MD5FILE "$chip->{'MD5'}  $chip->{'chip_name'}\n";
            if (!$status)
            {
                $logger->fatal_msg("Unable to print to MD5 file ");
                return ERROR;
            }
        }
    }
    
    # Close MD5 file
    if ($md5sum_flag && !close(MD5FILE))
    {
        $logger->fatal_msg("Unable to close MD5 file");
        return ERROR;
    }
    
    # Write out the tarball with compression specified
    if ($compression == IAS_GCP_NO_COMPRESSION)
    {
        $status = $tar->write($filename);
    }
    elsif ($compression == IAS_GCP_BZIP2)
    {
        $status = $tar->write($filename, COMPRESS_BZIP);
    }
    elsif ($compression == IAS_GCP_GZIP)
    {
        $status = $tar->write($filename, COMPRESS_GZIP);
    }
    else
    {
        $logger->fatal_msg("Unrecognized compression type");
        return ERROR;
    }

    # Ensure the tarball was written to file
    if (!defined $status)
    {
        $logger->fatal_msg("Unable to write the tarball to file $filename");
        return ERROR;
    }

    return SUCCESS;
}

#############################################################################
# 
# NAME:    ias_gcp_create_tar_filename
# 
# PURPOSE: Using the provided input returns the correct filename to use for
#          taring. 
# 
# Parameters:
#   - Base filename 
#   - Compression type
#
# RETURNS: Filename on success
#          Undef on error
# 
#############################################################################
sub ias_gcp_create_tar_filename
{
    my $filename = shift;           # Base filename
    my $compression = shift;        # Compression type
    my $logger = get_default_log(); # Get the default logger   

    if ($compression == IAS_GCP_BZIP2)
    {
        $filename .= ".tbz";
    }
    elsif ($compression == IAS_GCP_GZIP)
    {
        $filename .= ".tgz";
    }
    elsif ($compression == IAS_GCP_NO_COMPRESSION)
    {
        $filename .= ".tar";
    }
    else
    {
        $logger->fatal_msg("Unrecognized compression type");
        return undef;
    }
    
    return $filename;
}
