/******************************************************************************
Name: ias_gcp_image_chip_io

Purpose: Implements an image chip container IO interface.  The image chip
    container can be a directory or tarball.

******************************************************************************/
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include "config.h"
#ifdef HAVE_ARCHIVE_LIB
#include <archive.h>
#include <archive_entry.h>
#endif
#include "ias_const.h"
#include "ias_logging.h"
#include "ias_miscellaneous.h"
#include "ias_gcp.h"
#include "local_defines.h"


/* Define the standard names for the image chips tar file */
#define CHIP_TARBALL IAS_GCP_IMAGE_CHIP_FILENAME".tar"
#define CHIP_BZIP_TARBALL IAS_GCP_IMAGE_CHIP_FILENAME".tbz"

/******************************************************************************
Name: add_chip

Purpose: Adds an image chip to an IAS_GCP_IMAGE_CHIP_IO structure.  The
    image_chips array in the structure is grown as needed to contain the image
    chips added.

Returns: SUCCESS/ERROR
******************************************************************************/
static int add_chip
(
    IAS_GCP_IMAGE_CHIP_IO *chip_io, /* I/O: IO structure to add the chip to */
    const char *chip_name,          /* I: Name of chip to add */
    const unsigned char *chip_data, /* I: Image data for the chip */
    int chip_bytes                  /* I: Bytes of chip data */
)
{
    IMAGE_CHIP *chip;

    if (strlen(chip_name) >= IAS_GCP_CHIP_NAME_SIZE)
    {
        IAS_LOG_ERROR("Chip name %s is too long to be a legal chip name",
                chip_name);
        return ERROR;
    }

    /* Grow the image chip array if needed */
    if (chip_io->image_chip_count >= chip_io->allocated_chip_count)
    {
        IMAGE_CHIP *new_array;
        int new_count;

        /* Start with 200 records if none allocated yet, otherwise
           double the number of records */
        if (chip_io->allocated_chip_count == 0)
            new_count = 200;
        else
            new_count = 2 * chip_io->allocated_chip_count;

        /* Reallocate the buffer */
        new_array = realloc(chip_io->image_chips,
                        new_count * sizeof(*new_array));
        if (!new_array)
        {
            /* Error resizing the array, so return an error */
            IAS_LOG_ERROR("Resizing allocation for image chip array");
            return ERROR;
        }
        chip_io->image_chips = new_array;
        chip_io->allocated_chip_count = new_count;
    }

    /* Get a pointer to the chip to populate in the array */
    chip = &chip_io->image_chips[chip_io->image_chip_count];

    /* allocate memory for the chip imagery */
    chip->chip = malloc(chip_bytes * sizeof(unsigned char));
    if (!chip->chip)
    {
        IAS_LOG_ERROR("Allocating memory for an image chip");
        return ERROR;
    }

    /* Populate the next chip index */
    strcpy(chip->name, chip_name);
    chip->bytes = chip_bytes;

    /* Copy the chip data to the allocated memory */
    memcpy(chip->chip, chip_data, chip_bytes);

    chip_io->image_chip_count++;

    return SUCCESS;
}

/******************************************************************************
Name: free_chip_array

Purpose: Frees the image chips stored in the IAS_GCP_IMAGE_CHIP_IO image chips
    array and the array itself.

Returns: nothing
******************************************************************************/
static void free_chip_array
(
    IAS_GCP_IMAGE_CHIP_IO *chip_io  /* I: Contains image chip array to free */
)
{
    int i;

    /* Nothing to free if the image_chips pointer is NULL */
    if (!chip_io->image_chips)
        return;

    for (i = 0; i < chip_io->image_chip_count; i++)
    {
        free(chip_io->image_chips[i].chip);
        chip_io->image_chips[i].chip = NULL;
    }
    free(chip_io->image_chips);
    chip_io->image_chips = NULL;
    chip_io->image_chip_count = 0;
    chip_io->allocated_chip_count = 0;
}

/******************************************************************************
Name: read_image_chip_tarball

Purpose: Reads the image chips from a tarball (which may be compressed with
    gzip or bzip compression).  The image chips are stored in the image_chips
    array of the IAS_GCP_IMAGE_CHIP_IO structure.  libarchive is used to 
    read the tarball.  If that library is not available at compile time, this
    routine is built to just return an error.  This allows building the library
    on systems that do not have the library available.  Support for storing
    image chips in a directory should still function.

Notes:
    - The image chips are read into memory when the container is opened.  When
      ias_gcp_read_image_chip is called, it returns the chip from memory.

Returns: SUCCESS/ERROR
******************************************************************************/
static int read_image_chip_tarball
(
    IAS_GCP_IMAGE_CHIP_IO *chip_io  /* I/O: image chip container to read */
)
{
#ifdef HAVE_ARCHIVE_LIB
    /* This is only compiled if libarchive is available */
    struct archive *tarball;
    struct archive_entry *entry;
    int result;
    IMAGE_CHIP *chips_ptr;
    unsigned char *buffer = NULL;
    int64_t buffer_size = 0;

    /* Create a new archive object */
    tarball = archive_read_new();
    if (!tarball)
    {
        IAS_LOG_ERROR("Creating new archive object");
        return ERROR;
    }

    /* Enable support for all compression and tar file formats */
    if (archive_read_support_compression_all(tarball) != ARCHIVE_OK)
    {
        IAS_LOG_ERROR("Enabling archive compression support");
        archive_read_finish(tarball);
        return ERROR;
    }

    /* Enable support for reading all types of tar files */
    if (archive_read_support_format_all(tarball) != ARCHIVE_OK)
    {
        IAS_LOG_ERROR("Enabling tar format support");
        archive_read_finish(tarball);
        return ERROR;
    }

    /* Open the tar file, setting the block size to the largest expected (but
       unlikely) chip bytes */
    if (archive_read_open_filename(tarball, chip_io->container_name, 32768)
        != ARCHIVE_OK)
    {
        IAS_LOG_ERROR("Opening tar file %s", chip_io->container_name);
        archive_read_finish(tarball);
        return ERROR;
    }

    while ((result = archive_read_next_header(tarball, &entry)) == ARCHIVE_OK)
    {
        size_t bytes_read;
        const char *filename;
        int64_t file_size;

        /* Get the filename being read */
        filename = archive_entry_pathname(entry);
        if (!filename)
        {
            IAS_LOG_ERROR("Retrieving filename for entry");
            free(buffer);
            free_chip_array(chip_io);
            archive_read_finish(tarball);
            return ERROR;
        }

        IAS_LOG_DEBUG("Reading %s from tarball", archive_entry_pathname(entry));

        /* Get the size of the file and verify the buffer is large enough */
        file_size = archive_entry_size(entry);
        if (file_size > buffer_size)
        {
            /* Allocate a larger buffer */
            unsigned char *new_buffer;
            new_buffer = realloc(buffer, file_size);
            if (!new_buffer)
            {
                IAS_LOG_ERROR("Resizing memory allocation to %ld bytes for "
                    "the chip buffer", file_size);
                free(buffer);
                free_chip_array(chip_io);
                archive_read_finish(tarball);
                return ERROR;
            }
            buffer = new_buffer;
            buffer_size = file_size;
        }

        /* Read the data from the file */
        bytes_read = archive_read_data(tarball, buffer, buffer_size);

        /* Assuming here that ARCHIVE_RETRY is something that will only happen
           on damaged files or for files on tape.  Considering it an error
           until a case is found where it causes an issue. */
        if (bytes_read == ARCHIVE_WARN
            || bytes_read == ARCHIVE_RETRY
            || bytes_read == ARCHIVE_FATAL)
        {
            IAS_LOG_ERROR("Extracting %s from %s", filename,
                chip_io->container_name);
            free_chip_array(chip_io);
            archive_read_finish(tarball);
            return ERROR;
        }

        /* Verify the entire file was successfully read */
        if (bytes_read != file_size)
        {
            IAS_LOG_ERROR("Incorrect number of bytes read (%d) from %s for "
                "chip %s", (int)bytes_read, chip_io->container_name, filename);
            free(buffer);
            free_chip_array(chip_io);
            archive_read_finish(tarball);
            return ERROR;
        }

        /* Add the new chip to the chip array */
        if (add_chip(chip_io, filename, buffer, bytes_read) != SUCCESS)
        {
            IAS_LOG_ERROR("Adding image chip %s to chip array", filename);
            free(buffer);
            free_chip_array(chip_io);
            archive_read_finish(tarball);
            return ERROR;
        }
    }
    free(buffer);
    buffer = NULL;

    if (result != ARCHIVE_EOF)
    {
        /* If not at the end of the archive, we have an error.  Report the
           error from the archive library */
        IAS_LOG_ERROR("Reading archive %s: %s", chip_io->container_name,
            archive_error_string(tarball));
        free_chip_array(chip_io);
        archive_read_finish(tarball);
        return ERROR;
    }

    /* Shrink the memory allocation for the chips array to the number actually
       used */
    chips_ptr = realloc(chip_io->image_chips, chip_io->image_chip_count
        * sizeof(chip_io->image_chips[0]));
    if (!chips_ptr)
    {
        /* Failed shrinking the array.  This probably could just be ignored
           but if it fails, it probably is due to a corrupt heap, so best to
           provide an error as soon as possible. */
        IAS_LOG_ERROR("Shrinking memory allocation for the image chips");
        free_chip_array(chip_io);
        archive_read_finish(tarball);
        return ERROR;
    }
    chip_io->image_chips = chips_ptr;

    /* Free the archive */
    if (archive_read_finish(tarball) != ARCHIVE_OK)
    {
        IAS_LOG_ERROR("Freeing archive object");
        free_chip_array(chip_io);
        return ERROR;
    }

    return SUCCESS;
#else
    /* If not compiled with libarchive, this feature is not available */
    IAS_LOG_ERROR("Not compiled with support for image chips in a tarball");
    return ERROR;
#endif
}

/******************************************************************************
Name: write_image_chip_tarball

Purpose: Writes the image chips to a tarball.  Support is not provided to
    compress the tarball since uncompressed is sufficient for our needs.
    The image chips to be written are stored in the image_chips array of the
    IAS_GCP_IMAGE_CHIP_IO structure.  libarchive is used to write the tarball.
    If that library is not available at compile time, this routine is built to
    just return an error.  This allows building the library on systems that do
    not have the library available.  Support for storing image chips in a
    directory should still function.

Notes:
    - The ias_gcp_write_image_chip routine stores the image chip in the image
      chips array.  The tarball is not written to disk until the container
      is closed.

Returns: SUCCESS/ERROR
******************************************************************************/
static int write_image_chip_tarball
(
    const IAS_GCP_IMAGE_CHIP_IO *chip_io  /* I: image chip container to read */
)
{
#ifdef HAVE_ARCHIVE_LIB
    struct archive *tarball;
    struct archive_entry *entry;
    time_t current_time;
    int index;

    /* Create the archive object for writing */
    tarball = archive_write_new();
    if (!tarball)
    {
        IAS_LOG_ERROR("Creating archive object");
        return ERROR;
    }

    /* Set the format option */
    if (archive_write_set_format_pax_restricted(tarball) != ARCHIVE_OK)
    {
        IAS_LOG_ERROR("Setting format option");
        archive_write_finish(tarball);
        return ERROR;
    }

    /* Open the tarball file */
    if (archive_write_open_filename(tarball, chip_io->container_name)
        != ARCHIVE_OK)
    {
        IAS_LOG_ERROR("Opening tarball %s", chip_io->container_name);
        archive_write_finish(tarball);
        return ERROR;
    }

    /* Create the archive entry object */
    entry = archive_entry_new();
    if (!entry)
    {
        IAS_LOG_ERROR("Creating archive entry object");
        archive_write_finish(tarball);
        return ERROR;
    }

    /* Get the current time to use for the files */
    current_time = time(NULL);

    /* Loop through the image chips in the array */
    for (index = 0; index < chip_io->image_chip_count; index++)
    {
        IMAGE_CHIP *chip = &chip_io->image_chips[index];

        /* Set metadata for the entry.  None of these return errors */
        archive_entry_clear(entry);
        archive_entry_set_pathname(entry, chip->name);
        archive_entry_set_size(entry, chip->bytes);
        archive_entry_set_filetype(entry, AE_IFREG);
        archive_entry_set_perm(entry, 0644);
        archive_entry_set_mtime(entry, current_time, 0);

        if (archive_write_header(tarball, entry) != ARCHIVE_OK)
        {
            IAS_LOG_ERROR("Writing header to tar file %s for image chip %s",
                chip_io->container_name, chip->name);
            archive_entry_free(entry);
            archive_write_finish(tarball);
            return ERROR;
        }

        /* Write the image chip data to the tarball */
        if (archive_write_data(tarball, chip->chip, chip->bytes) 
            != chip->bytes)
        {
            IAS_LOG_ERROR("Writing image chip %s to the tar file %s",
                chip->name, chip_io->container_name);
            archive_entry_free(entry);
            archive_write_finish(tarball);
            return ERROR;
        }
    }
    archive_entry_free(entry);

    /* Close the tarball file */
    if (archive_write_close(tarball) != ARCHIVE_OK)
    {
        IAS_LOG_ERROR("Closing tar file %s : %s", chip_io->container_name,
            archive_error_string(tarball));
        archive_write_finish(tarball);
        return ERROR;
    }
    /* Free the archive object */
    if (archive_write_finish(tarball) != ARCHIVE_OK)
    {
        IAS_LOG_ERROR("Freeing archive");
        return ERROR;
    }

    return SUCCESS;
#else
    /* If not compiled with libarchive, this feature is not available */
    IAS_LOG_ERROR("Not compiled with support for image chips in a tarball");
    return ERROR;
#endif
}

/******************************************************************************
Name: ias_gcp_open_image_chip_container

Purpose: Opens an image chip container for reading or writing.  The container
    can be either a directory or a tarball (if libarchive is available).  When
    reading a tarball, all the image chips are read into memory in the open
    routine.

Notes:
    - The container type can be IAS_GCP_CONTAINER_SEARCH.  When it is, the
      provided container name can be a directory name that is searched for the
      standard image chip tarball name. If a tarball is found, it is assumed to
      be the desired container.  If no tarball is found, the directory must
      be the container.  This is used to allow applications to not really know
      which type of container is being used.

Returns:
    Pointer the IAS_GCP_IMAGE_CHIP_IO structure that was allocated, or NULL
    if an error occurs.
******************************************************************************/
IAS_GCP_IMAGE_CHIP_IO *ias_gcp_open_image_chip_container
(
    const char *container_name,     /* I: Name of container to open */
    IAS_GCP_CONTAINER_TYPE type,    /* I: Type of container */
    IAS_ACCESS_MODE access_mode     /* I: Access mode to use when opening */
)
{
    IAS_GCP_IMAGE_CHIP_IO *chip_io;
    char local_name[PATH_MAX];
    const char *name_ptr = container_name;

    /* If the container type indicates to search for the correct container,
       try to figure out what the name should be. */
    if (type == IAS_GCP_CONTAINER_SEARCH)
    {
        struct stat stat_buf;

        if (access_mode == IAS_WRITE)
        {
            IAS_LOG_ERROR("Searching for the image chip container not "
                "supported in write mode");
            return NULL;
        }

        /* Determine whether the container name is a directory or file */
        if (stat(container_name, &stat_buf) != 0)
        {
            IAS_LOG_ERROR("Unable to find %s", container_name);
            return NULL;
        }

        if (S_ISREG(stat_buf.st_mode))
        {
            /* It is a file, so no need to search, the caller must have been
               lazy and didn't want to figure out which they had */
            type = IAS_GCP_CONTAINER_TARBALL;
        }
        else if (S_ISDIR(stat_buf.st_mode))
        {
            int status;

            /* Directory, so check for the standard tarball names */
            status = snprintf(local_name, sizeof(local_name), "%s/%s",
                container_name, CHIP_TARBALL);
            if (status <= 0 || status >= sizeof(local_name))
            {
                IAS_LOG_ERROR("Buffer is not large enough");
                return NULL;
            }
            if (stat(local_name, &stat_buf) != 0)
            {
                /* Try the next name */
                status = snprintf(local_name, sizeof(local_name), "%s/%s",
                    container_name, CHIP_BZIP_TARBALL);
                if (status <= 0 || status >= sizeof(local_name))
                {
                    IAS_LOG_ERROR("Buffer is not large enough");
                    return NULL;
                }
                if (stat(local_name, &stat_buf) != 0)
                {
                    /* Neither type of tarball name has been found, so assume
                       it is the directory type */
                    type = IAS_GCP_CONTAINER_DIR;
                }
                else
                {
                    type = IAS_GCP_CONTAINER_TARBALL;
                    name_ptr = local_name;
                }
            }
            else
            {
                type = IAS_GCP_CONTAINER_TARBALL;
                name_ptr = local_name;
            }
        }
        else
        {
            IAS_LOG_ERROR("%s is not a directory or file", container_name);
            return NULL;
        }
    }

    /* Allocate the IO structure */
    chip_io = calloc(1, sizeof(*chip_io));
    if (!chip_io)
    {
        IAS_LOG_ERROR("Allocating memory for the IO structure");
        return NULL;
    }

    /* Populate the structure */
    chip_io->container_name = strdup(name_ptr);
    if (!chip_io->container_name)
    {
        IAS_LOG_ERROR("Duplicating container name string");
        free(chip_io);
        return NULL;
    }
    chip_io->mode = access_mode;
    chip_io->type = type;

    /* If opened in reading mode for a tarball, read the tarball now */
    if (access_mode == IAS_READ && type == IAS_GCP_CONTAINER_TARBALL)
    {
        if (read_image_chip_tarball(chip_io) != SUCCESS)
        {
            IAS_LOG_ERROR("Failed reading image chips from tarball %s",
                chip_io->container_name);
            free(chip_io->container_name);
            free(chip_io);
            return NULL;
        }
    }

    return chip_io;
}

/******************************************************************************
Name: ias_gcp_close_image_chip_container

Purpose: Closes an image chip container that was opened using 
    ias_gcp_open_image_chip_container.  The IAS_GCP_IMAGE_CHIP_IO is freed and
    may no longer be used after this routine is called.  When the container is
    opened in write mode for a tarball container, the image chips are written
    to the tarball file by this routine.

Returns: SUCCESS/ERROR
******************************************************************************/
int ias_gcp_close_image_chip_container
(
    IAS_GCP_IMAGE_CHIP_IO *chip_io  /* I: Chip container to close */
)
{
    int status = SUCCESS;

    /* Consider it an error if passed a NULL pointer */
    if (!chip_io)
    {
        IAS_LOG_ERROR("Null pointer provided");
        return ERROR;
    }

    /* If open for writing a tarball, write the image chips */
    if (chip_io->mode == IAS_WRITE 
        && chip_io->type == IAS_GCP_CONTAINER_TARBALL)
    {
        if (write_image_chip_tarball(chip_io) != SUCCESS)
        {
            IAS_LOG_ERROR("Failed trying to write the image chip tarball %s",
                chip_io->container_name);
            status = ERROR;
        }
    }

    /* Free the memory allocated to the container */
    free_chip_array(chip_io);
    free(chip_io->container_name);
    chip_io->container_name = NULL;
    free(chip_io);

    return status;
}

/******************************************************************************
Name: ias_gcp_get_image_chip_container_name

Purpose: Provides access to the image chip container name.  This is useful
    when the container is opened in the mode that searches for the image chip
    container to use since the name can not be determined otherwise.

Returns: Pointer to the character buffer with the container name.

******************************************************************************/
const char *ias_gcp_get_image_chip_container_name
(
    const IAS_GCP_IMAGE_CHIP_IO *chip_io  /* I: Chip container to close */
)
{
    return chip_io->container_name;
}

/******************************************************************************
Name: ias_gcp_get_image_chip_container_type

Purpose: Provides access to the type of container that is being used.  This
    is useful to determine the container type found when opened in search mode.
    Some applications may need to know what kind of container was found.

Returns: The container type
******************************************************************************/
IAS_GCP_CONTAINER_TYPE ias_gcp_get_image_chip_container_type
(
    const IAS_GCP_IMAGE_CHIP_IO *chip_io  /* I: Chip container to close */
)
{
    return chip_io->type;
}

/******************************************************************************
Name: swap_bytes

Purpose: Provides a routine to swap bytes when running on a big endian machine.
    The image chips are stored in little endian byte order and this routine
    is called to swap the bytes between the two formats.  If the machine
    uses big endian, this routine just copies the data between the buffers.

Returns: SUCCESS/ERROR
******************************************************************************/
static int swap_bytes
(
    void *input_buf,            /* I: Input buffer with original data */
    IAS_DATA_TYPE data_type,    /* I: Data type in the buffer */
    int buffer_element_count,   /* I: Count of elements in the buffer */
    void *output_buf            /* O: Buffer with swapped bytes */
)
{
    if (data_type == IAS_BYTE)
    {
        /* byte swapping unnecessary for byte data */
        memcpy(output_buf, input_buf, buffer_element_count);
    }
    else if ((data_type == IAS_I2) || (data_type == IAS_UI2))
    {
        /* swap byte order in 16 bit data case */
        const IAS_UI2_TYPE *in_buf_16 = input_buf;
        IAS_UI2_TYPE *out_buf_16 = output_buf;
        int i;

        for (i = 0; i < buffer_element_count; i++)
        {
            IAS_UI2_TYPE data_value = in_buf_16[i];
            SWAP_BYTES_16(data_value);
            out_buf_16[i] = data_value;
        }
    }
    else if (data_type == IAS_R4)
    {
        /* swap byte order in 32 bit data case */
        const IAS_R4_TYPE *in_buf_32 = input_buf;
        IAS_R4_TYPE *out_buf_32 = output_buf;
        int i;

        for (i = 0; i < buffer_element_count; i++)
        {
            /* Doing a bit different than 16 bit case, as using the local
               variable with the floating point causes a compile warning.
               Swapping directly in the buffer avoids a compile warning, so
               do this even though it's slightly less efficient, since this
               is such an unlikely case. */
            out_buf_32[i] = in_buf_32[i];
            SWAP_BYTES_32(out_buf_32[i]);
        }
    }
    else
    {
        /* This shouldn't happen since the caller should have already checked,
           but handling it just in case a bug is introduced */
        IAS_LOG_ERROR("Unsupported data type %d", data_type);
        return ERROR;
    }

    return SUCCESS;
}

/******************************************************************************
Name: convert_buffer_to_float

Purpose: Converts the data in the input buffer from one of the supported types
    to 32-bit floating point data.  If needed, byte swapping is performed.

Returns: SUCCESS/ERROR
******************************************************************************/
static int convert_buffer_to_float
(
    const void *input_buffer,       /* I: Buffer to convert */
    int buffer_elements,            /* I: Number of elements in the buffer */
    IAS_DATA_TYPE input_data_type,  /* I: Input buffer data type */
    float *output_buffer            /* O: Output buffer containing floats */
)
{
    int i;

    if (input_data_type == IAS_BYTE)
    {
        /* Create a byte data type buffer pointer to point at the
           max sized data buffer */
        const IAS_BYTE_TYPE *buf_byte = input_buffer;

        for (i = 0; i < buffer_elements; i++)
            output_buffer[i] = buf_byte[i];
    }
    else if (input_data_type == IAS_I2)
    {
        /* Create a signed 16-bit int data type buffer pointer to point
           at the max sized data buffer */
        const IAS_I2_TYPE *buf_i2 = input_buffer;

        for (i = 0; i < buffer_elements; i++)
        {
            IAS_I2_TYPE data_value = buf_i2[i];
            /* Swap byte order of the buffer since GCP chips on disk are
               in little endian format, in case on big endian system */
            SWAP_BYTES_16(data_value);
            output_buffer[i] = data_value;
        }
    }
    else if (input_data_type == IAS_UI2)
    {
        /* Create an unsigned 16-bit int data type buffer pointer to point
           at the max sized data buffer */
        const IAS_UI2_TYPE *buf_ui2 = input_buffer;

        for (i = 0; i < buffer_elements; i++)
        {
            IAS_UI2_TYPE data_value = buf_ui2[i];
            /* Swap byte order of the buffer since GCP chips on disk are
               in little endian format, in case on big endian system */
            SWAP_BYTES_16(data_value);
            output_buffer[i] = data_value;
        }
    }
    else if (input_data_type == IAS_R4)
    {
        /* Create a 4-byte floating point data type buffer pointer to point
           at the max sized data buffer */
        const IAS_R4_TYPE *buf_r4 = input_buffer;

        for (i = 0; i < buffer_elements; i++)
        {
            /* Swap byte order of the buffer since GCP chips on disk are
               in little endian format, in case on big endian system */
            /* Doing a bit different than 16 bit cases, as using the local
               variable with the floating point causes a compile warning.
               Swapping directly in the buffer avoids a compile warning, so
               do this even though it's slightly less efficient, since this
               is such an unlikely case. */
            output_buffer[i] = buf_r4[i];
            SWAP_BYTES_32(output_buffer[i]);
        }
    }
    else
    {
        /* This shouldn't happen since the caller should have already checked,
           but handling it just in case a bug is introduced */
        IAS_LOG_ERROR("Unsupported data type %d", input_data_type);
        return ERROR;
    }

    return SUCCESS;
}

/******************************************************************************
Name: ias_gcp_read_image_chip

Purpose: Reads an image chip from an open container.  If the container is a
    directory, the file with the chip name is read.  If the container is a
    tarball, the chip is retrieved from the image chips array that was
    populated from the tarball when the container was opened.

Returns: SUCCESS/ERROR
******************************************************************************/
int ias_gcp_read_image_chip
(
    const IAS_GCP_IMAGE_CHIP_IO *chip_io, /* I: Chip container to read */
    const char *chip_name,          /* I: Image chip name to read */
    int chip_size_lines,            /* I: Expected number of chip lines */
    int chip_size_samples,          /* I: Expected number of chip samples */
    IAS_DATA_TYPE chip_data_type,   /* I: Image chip data type */
    int convert_output_to_float,    /* I: flag to convert the output to
                                          floating point (1 = convert to float,
                                          0 = no conversion) */
    void *output_chip_buffer        /* O: Buffer of chip image data converted
                                          to float (R*4) if requested, 
                                          otherwise the same type contained in
                                          the file */
)
{
    void *chip_buffer;                  /* Pointer to the chip buffer to use */
    void *allocated_chip_buffer = NULL; /* Pointer to a temporary buffer that
                                           is needed under some conditions */
    int chip_bytes;
    int chip_total_samples = chip_size_lines * chip_size_samples;
    int bytes_per_pixel;
    int expected_chip_bytes;

    /* Only legal to call this routine in read mode */
    if (chip_io->mode != IAS_READ)
    {
        IAS_LOG_ERROR("Attempting to read image chips from a container not "
            "opened for reading");
        return ERROR;
    }

    /* Verify the chip data type is a supported type */
    switch (chip_data_type)
    {
        /* This is the list of legal types */
        case IAS_BYTE:
        case IAS_I2:
        case IAS_UI2:
        case IAS_R4:
            break;
        default:
            IAS_LOG_ERROR("Unsupported GCP chip data type %d", chip_data_type);
            return ERROR;
    }

    /* Calculate the expected number of bytes for the image chip */
    if (ias_misc_get_sizeof_data_type(chip_data_type, &bytes_per_pixel)
        != SUCCESS)
    {
        IAS_LOG_ERROR("Unsupported GCP chip data type %d", chip_data_type);
        return ERROR;
    }
    expected_chip_bytes = chip_total_samples * bytes_per_pixel;

    if (chip_io->type == IAS_GCP_CONTAINER_DIR)
    {
        /* Read the image chip from a file in the directory */
        int status;
        char filename[PATH_MAX];
        int chip_fd;
        struct stat file_stat;

        /* Build the filename */
        status = snprintf(filename, sizeof(filename), "%s/%s",
            chip_io->container_name, chip_name);
        if (status <= 0 || status >= sizeof(filename))
        {
            IAS_LOG_ERROR("Insufficient buffer size to build image chip name");
            return ERROR;
        }

        /* Open the image chip to read */
        chip_fd = open(filename, O_RDONLY);
        if (chip_fd == -1)
        {
            IAS_LOG_ERROR("Opening image chip %s", filename);
            return ERROR;
        }

        /* Get the size of the image chip file */
        if (fstat(chip_fd, &file_stat) < 0)
        {
            IAS_LOG_ERROR("Getting image chip file status for %s", filename);
            close(chip_fd);
            return ERROR;
        }
        chip_bytes = file_stat.st_size;

        /* Allocate memory for the chip */
        allocated_chip_buffer = malloc(chip_bytes);
        if (!allocated_chip_buffer)
        {
            IAS_LOG_ERROR("Allocating memory for an image chip");
            close(chip_fd);
            return ERROR;
        }

        /* Read the image chip */
        if (read(chip_fd, allocated_chip_buffer, chip_bytes) != chip_bytes)
        {
            IAS_LOG_ERROR("Reading image chip %s", filename);
            free(allocated_chip_buffer);
            close(chip_fd);
            return ERROR;
        }

        /* Close the chip file (no need to check the return value since only
           opened for reading) */
        close(chip_fd);

        chip_buffer = allocated_chip_buffer;
    }
    else
    {
        /* The container is a tarball, so get the image chip from the array
           of chips read from the tarball */
        int index;
        int found = 0;

        /* Find the image chip in the array.  This uses a brute force linear
           search, but it isn't a performance concern. */
        for (index = 0; index < chip_io->image_chip_count; index++)
        {
            if (strcmp(chip_name, chip_io->image_chips[index].name) == 0)
            {
                /* Found the chip */
                chip_buffer = chip_io->image_chips[index].chip;
                chip_bytes = chip_io->image_chips[index].bytes;
                found = 1;
                break;
            }
        }
        if (!found)
        {
            IAS_LOG_ERROR("Image chip %s does not exist in %s", chip_name,
                chip_io->container_name);
            return ERROR;
        }
    }

    /* Make sure the size of the chip matches the expected size */
    if (chip_bytes != expected_chip_bytes)
    {
        IAS_LOG_ERROR("Image chip %s has %d bytes instead of the expected"
            " %d bytes", chip_name, chip_bytes, expected_chip_bytes);
        free(allocated_chip_buffer);
        return ERROR;
    }

    if (convert_output_to_float)
    {
        /* Convert the buffer to float */
        if (convert_buffer_to_float(chip_buffer, chip_total_samples,
            chip_data_type, output_chip_buffer) != SUCCESS)
        {
            IAS_LOG_ERROR("convert_buffer_to_float routine failed (should not "
                "happen)");
            free(allocated_chip_buffer);
            return ERROR;
        }
    }
    else
    {
        /* No conversion requested, so copy the data to the output buffer,
           swapping the bytes if needed */
        if (swap_bytes(chip_buffer, chip_data_type, chip_total_samples,
                output_chip_buffer) != SUCCESS)
        {
            IAS_LOG_ERROR("swap_bytes routine failed (should not happen)");
            free(allocated_chip_buffer);
            return ERROR;
        }
    }

    free(allocated_chip_buffer);

    return SUCCESS;
}

/******************************************************************************
Name: ias_gcp_write_image_chip

Purpose: Writes an image chip to the previously opened container.  If the 
    container is a directory, the chip is written to a file.  If the container
    is a tarball, the chip is added to the image chips array which is written
    to the tarball when the container is closed.

Returns: SUCCESS/ERROR
******************************************************************************/
int ias_gcp_write_image_chip
(
    IAS_GCP_IMAGE_CHIP_IO *chip_io, /* I/O: Chip container to write */
    const char *chip_name,      /* I: Name of the image chip file */
    int chip_size_lines,        /* I: Number of chip lines in data buffer */
    int chip_size_samples,      /* I: Number of chip samples in data buffer */
    IAS_DATA_TYPE dtype,        /* I: Data type for byte swapping if needed */
    const void *buf             /* I: Image chip data buffer to write to file */
)
{
    const void *write_buf = buf;    /* Pointer to the buffer to write.  Can
                                       change if the bytes are swapped */
    void *allocated_buf = NULL;     /* Buffer allocated if bytes are swapped */
    int bytes_per_pixel;
    int chip_pixels = chip_size_lines * chip_size_samples;
    int chip_bytes;

    /* Only legal to call this routine in write mode */
    if (chip_io->mode != IAS_WRITE)
    {
        IAS_LOG_ERROR("Attempting to write image chips to a container not "
            "opened for writing");
        return ERROR;
    }

    /* Verify the chip data type is a supported type */
    switch (dtype)
    {
        /* This is the list of legal types */
        case IAS_BYTE:
        case IAS_I2:
        case IAS_UI2:
        case IAS_R4:
            break;
        default:
            IAS_LOG_ERROR("Unsupported GCP chip data type %d", dtype);
            return ERROR;
    }

    /* Calculate the number of bytes for the image chip */
    if (ias_misc_get_sizeof_data_type(dtype, &bytes_per_pixel)
        != SUCCESS)
    {
        IAS_LOG_ERROR("Unsupported GCP chip data type %d", dtype);
        return ERROR;
    }
    chip_bytes = chip_pixels * bytes_per_pixel;

#ifndef HAVE_LITTLE_ENDIAN
    /* Allocate a buffer for the copy of the data with swapped bytes */
    allocated_buf = malloc(chip_bytes);
    if (!allocated_buf)
    {
        IAS_LOG_ERROR("Allocating memory for byte swapping");
        return ERROR;
    }
    write_buf = allocated_buf;

    if (swap_bytes(buf, dtype, chip_pixels, write_buf) != SUCCESS)
    {
        IAS_LOG_ERROR("swap_bytes routine failed (should not happen)");
        free(allocated_buf);
        return ERROR;
    }
#endif

    /* If in directory mode, write the file immediately */
    if (chip_io->type == IAS_GCP_CONTAINER_DIR)
    {
        char filename[PATH_MAX];
        int status;
        int fd;

        /* Build the full filename */
        status = snprintf(filename, sizeof(filename), "%s/%s", 
            chip_io->container_name, chip_name);
        if (status <= 0 || status >= sizeof(filename))
        {
            IAS_LOG_ERROR("Filename buffer not large enough");
            free(allocated_buf);
            return ERROR;
        }
        
        /* Prepare the output file */
        fd = creat(filename, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
        if (fd < 0)
        {
            IAS_LOG_ERROR("Opening output GCP chip file %s for writing: %s",
                filename, strerror(errno));
            free(allocated_buf);
            return ERROR;
        }

        /* Write output buffer */
        if (write(fd, write_buf, chip_bytes) != chip_bytes)
        {
            IAS_LOG_ERROR("Writing output GCP chip data to %s", filename);
            close(fd);
            free(allocated_buf);
            return ERROR;
        }

        /* Close the file */
        if (close(fd) < 0)
        {
            IAS_LOG_ERROR("Closing output GCP chip %s", filename);
            free(allocated_buf);
            return ERROR;
        }
    }
    else
    {
        /* Writing to a tar file, so add it to the list */
        if (add_chip(chip_io, chip_name, write_buf, chip_bytes) != SUCCESS)
        {
            IAS_LOG_ERROR("Adding image chip %s to the chip buffer",
                chip_name);
            free(allocated_buf);
            return ERROR;
        }
    }
    free(allocated_buf);

    return SUCCESS;
}
