package ias_gcp_old;

use strict;
use warnings;
use Logger;
use ias_const;                           # defines SUCCESS and ERROR

our (@ISA, @EXPORT_OK);
BEGIN
{
    require Exporter;
    @ISA = qw(Exporter);

    # symbols to export on request
    @EXPORT_OK = qw(ias_gcp_read_old_gcplib ias_gcp_write_old_gcplib);
}


#############################################################################
# 
# NAME:           ias_gcp_read_old_gcplib
# 
# PURPOSE:        Read GCP metadata from disk. 
# 
# RETURNS:        1 and GCP data structure on success
#                 0 and undef on error
# 
#############################################################################
sub ias_gcp_read_old_gcplib($)
{
    my $gcp_filename = shift;              # GCP data file name

    my @gcp_data; # array of hashes

    # Get the default logger object from the library 
    my $error_log = get_default_log();

    # Verify filename is defined and is not blank 
    if ((!defined($gcp_filename)) || ($gcp_filename eq ""))
    {
        $error_log->fatal_msg("Empty Filename");
        return ERROR, undef;
    }

    # Open GCPLIB file
    my $status = open GCP_FILE, "$gcp_filename";
    if (!defined($status))
    {
        $error_log->fatal_msg("Opening $gcp_filename");
        return ERROR, undef;
    }

    my $expected_record_count = 0;
    my $line_number = 0;
    while (my $line = <GCP_FILE>)
    {
        my $gcp_record = {};
        chomp $line;

        $line_number++;

        # Look for the BEGIN data start marker
        if ($line =~ /^BEGIN$/)
        {
            # Read the next line for the record count
            $line = <GCP_FILE>;
            chomp $line;
            if ($line =~ /^\d+$/)
            {
                $expected_record_count = $line;
                next;
            }
            else
            {
                # Invalid record count line
                $error_log->fatal_msg("Invalid record count following BEGIN "
                    . "in $gcp_filename");
                return ERROR, undef;
            }
        }

        # Skip comments
        next if $line =~ /^#/;

        #split out data
        my ($point_id, $chip_name, $reference_line, $reference_sample,
            $latitude, $longitude, $projection_y, $projection_x, $elevation, 
            $pixel_size_x, $pixel_size_y, $chip_size_lines, $chip_size_samples, 
            $source, $chip_type, $projection, $zone, $date, $abs_rel_flag, 
            $chip_data_type)
        = split (/\s/, $line, 20); 

        # Add vars to gcp_record hash
        $gcp_record->{'point_id'}               = $point_id;
        $gcp_record->{'chip_name'}              = $chip_name;
        $gcp_record->{'reference_line'}         = $reference_line;
        $gcp_record->{'reference_sample'}       = $reference_sample;
        $gcp_record->{'latitude'}               = $latitude;
        $gcp_record->{'longitude'}              = $longitude;
        $gcp_record->{'projection_x'}           = $projection_x;
        $gcp_record->{'projection_y'}           = $projection_y;
        $gcp_record->{'elevation'}              = $elevation;
        $gcp_record->{'pixel_size_x'}           = $pixel_size_x;
        $gcp_record->{'pixel_size_y'}           = $pixel_size_y;
        $gcp_record->{'chip_size_lines'}        = $chip_size_lines;
        $gcp_record->{'chip_size_samples'}      = $chip_size_samples;
        $gcp_record->{'source'}                 = $source;
        $gcp_record->{'chip_type'}              = $chip_type;
        $gcp_record->{'projection'}             = $projection;
        $gcp_record->{'zone'}                   = $zone;
        $gcp_record->{'date'}                   = $date;
        $gcp_record->{'abs_rel_flag'}           = $abs_rel_flag;
        $gcp_record->{'chip_data_type'}         = $chip_data_type;

        # Verify the GCP entries
        $status = check_gcp($gcp_record, $error_log);
        if ($status != SUCCESS)
        {
            $error_log->fatal_msg("Checking GCP record from line number "
                . "$line_number in $gcp_filename");
            return ERROR, undef;
        }

        # Add gcp_record hash to gcp_data array
        push @gcp_data, $gcp_record;
    }

    # Verify gcp_data array of hashes has the number of expected records
    my $total = $#gcp_data + 1;
    if ($total != $expected_record_count)
    {
        $error_log->fatal_msg("Expected $expected_record_count in "
            . "$gcp_filename, but found $total");
        return ERROR, undef;
    }

    close GCP_FILE;

    return SUCCESS, @gcp_data;
}

#############################################################################
# 
# NAME:           ias_gcp_write_old_gcplib
# 
# PURPOSE:        The GCP Retrieval Module Detail Design describes a Perl script
#                 that acquires GCP data. This Perl module writes the GCP 
#                 metadata to disk. 
# 
# RETURNS:        1 on success
#                 0 on error
# 
# 
#############################################################################
sub ias_gcp_write_old_gcplib($$)
{
    my $gcp_filename = shift;              # GCP data file name
    my $gcp_data_reference = shift;        # Array reference

    # Get the default logger object from the library 
    my $error_log = get_default_log();

    # Verify gcp_data_reference is defined
    if (!defined($gcp_data_reference))
    {
        my $error = "gcp_data_reference is not defined";
        $error_log->fatal_msg(__FILE__, __LINE__, $error);
        return ERROR;
    }

    # Verify filename is defined and is not blank 
    if ((!defined($gcp_filename)) || ($gcp_filename eq ""))
    {
        my $error = "Empty Filename";
        $error_log->fatal_msg(__FILE__, __LINE__, $error);
        return ERROR;
    }

    my @gcp_data = @{$gcp_data_reference}; # Array of hashes containing GCP data

    # Verify gcp_data array of hashes has at least one record (hash)
    my $total = $#gcp_data + 1;
    if ($total == 0)
    {
        my $warn = "No GCPLIB records";
        $error_log->warn_msg(__FILE__, __LINE__, $warn);
    }

    # Open the file
    my $status = open OUT, ">$gcp_filename";
    if (!defined($status))
    {
        my $error = "Opening $gcp_filename";
        $error_log->fatal_msg(__FILE__, __LINE__, $error);
        return ERROR;
    }

    # Print the header 
    $status = print OUT "# Unique Ground Control Point ID - integer\n"
        . "# Name of the chip image - string\n"
        . "# Reference line - double\n"
        . "# Reference sample - double\n"
        . "# True latitude of reference point in degrees - double\n"
        . "# True longitude of reference point in degrees - double\n"
        . "# Projection x value of reference point in meters - double\n"
        . "# Projection y value of reference point in meters - double\n"
        . "# Elevation at the Ground Control Point in meters - double\n"
        . "# Pixel size x value in meters - double\n"
        . "# Pixel size y value in meters - double\n"
        . "# Chip size in lines - double\n"
        . "# Chip size in samples - double\n"
        . "# Ground Control Point source - string\n"
        . "# Ground Control Point type - string\n"
        . "# Chip projection - string\n"
        . "# Projection UTM zone - integer\n"
        . "# Acquisition date - string in form mm-dd-yyyy\n"
        . "# Absolute or relative flag - string, should be ABS or REL\n"
        . "# Image chip data type - string\n";
    if ($status == 0)
    {
        my $error = "Writing header comment to $gcp_filename";
        $error_log->fatal_msg(__FILE__, __LINE__, $error);
        close OUT;
        return ERROR;
    }

    # Print the data start marker
    $status = print OUT "BEGIN\n";
    if ($status == 0)
    {
        my $error = "Writing BEGIN to $gcp_filename";
        $error_log->fatal_msg(__FILE__, __LINE__, $error);
        close OUT;
        return ERROR;
    }

    # Print the number of GCPs
    $status = print OUT "$total\n";
    if ($status == 0)
    {
        my $error = "Writing number of GCPs to $gcp_filename";
        $error_log->fatal_msg(__FILE__, __LINE__, $error);
        close OUT;
        return ERROR;
    }

    for my $gcp_record ( @gcp_data )
    {
        # Verify the GCP entries
        $status = check_gcp($gcp_record, $error_log);
        if ($status != SUCCESS)
        {
            my $error = "Found while checking GCPs in $gcp_filename";
            $error_log->fatal_msg(__FILE__, __LINE__, $error);
            close OUT;
            return ERROR;
        }

        # Get values out of each hash
        my $point_id          = $gcp_record->{'point_id'};
        my $chip_name         = $gcp_record->{'chip_name'};
        my $reference_line    = $gcp_record->{'reference_line'};
        my $reference_sample  = $gcp_record->{'reference_sample'};
        my $latitude          = $gcp_record->{'latitude'};
        my $longitude         = $gcp_record->{'longitude'};
        my $projection_x      = $gcp_record->{'projection_x'};
        my $projection_y      = $gcp_record->{'projection_y'};
        my $elevation         = $gcp_record->{'elevation'};
        my $pixel_size_x      = $gcp_record->{'pixel_size_x'};
        my $pixel_size_y      = $gcp_record->{'pixel_size_y'};
        my $chip_size_lines   = $gcp_record->{'chip_size_lines'};
        my $chip_size_samples = $gcp_record->{'chip_size_samples'};
        my $source            = $gcp_record->{'source'};
        my $chip_type         = $gcp_record->{'chip_type'};
        my $projection        = $gcp_record->{'projection'};
        my $zone              = $gcp_record->{'zone'};
        my $date              = $gcp_record->{'date'};
        my $abs_rel_flag      = $gcp_record->{'abs_rel_flag'};
        my $chip_data_type    = $gcp_record->{'chip_data_type'};

        # Print out the values to the file
        my $status = printf OUT "%d %s %lf %lf %lf %lf %lf %lf %lf %lf %lf " .
            "%lf %lf %s %s %s %d %s %s %s\n", $point_id, $chip_name,
            $reference_line, $reference_sample, $latitude, $longitude,
            $projection_y, $projection_x, $elevation, $pixel_size_x,
            $pixel_size_y, $chip_size_lines, $chip_size_samples, $source, 
            $chip_type, $projection, $zone, $date, $abs_rel_flag, 
            $chip_data_type;
        if ($status == 0)
        {
            my $error = "Writing values to $gcp_filename";
            $error_log->fatal_msg(__FILE__, __LINE__, $error);
            close OUT;
            return ERROR;
        }
    }
    close OUT;

    return SUCCESS;
}

#############################################################################
# 
# NAME:           check_gcp
# 
# PURPOSE:        Verifies the format the fields in a GCP.
# 
# RETURNS:        1 on success
#                 0 on error
# 
#############################################################################
sub check_gcp
{
    my $gcp_record = shift;
    my $error_log = shift;

    my $point_id          = $gcp_record->{'point_id'};
    my $chip_name         = $gcp_record->{'chip_name'};
    my $reference_line    = $gcp_record->{'reference_line'};
    my $reference_sample  = $gcp_record->{'reference_sample'};
    my $latitude          = $gcp_record->{'latitude'};
    my $longitude         = $gcp_record->{'longitude'};
    my $projection_x      = $gcp_record->{'projection_x'};
    my $projection_y      = $gcp_record->{'projection_y'};
    my $elevation         = $gcp_record->{'elevation'};
    my $pixel_size_x      = $gcp_record->{'pixel_size_x'};
    my $pixel_size_y      = $gcp_record->{'pixel_size_y'};
    my $chip_size_lines   = $gcp_record->{'chip_size_lines'};
    my $chip_size_samples = $gcp_record->{'chip_size_samples'};
    my $source            = $gcp_record->{'source'};
    my $chip_type         = $gcp_record->{'chip_type'};
    my $projection        = $gcp_record->{'projection'};
    my $zone              = $gcp_record->{'zone'};
    my $date              = $gcp_record->{'date'};
    my $abs_rel_flag      = $gcp_record->{'abs_rel_flag'};
    my $chip_data_type    = $gcp_record->{'chip_data_type'};

    # Verify these variables have the correct type of values
    if ($point_id !~ /^\d+$/) # should be int
    {
        my $error = "point_id: invalid value $point_id";
        $error_log->fatal_msg(__FILE__, __LINE__, $error);
        return ERROR;
    }
    if ($chip_name !~ /^\w+$/) # should be string
    {
        my $error = "chip_name: invalid value $chip_name";
        $error_log->fatal_msg(__FILE__, __LINE__, $error);
        return ERROR;
    }
    if ($reference_line !~ /^-?\d+\.?\d*$/) # should be double
    {
        my $error = "reference_line: invalid value $reference_line";
        $error_log->fatal_msg(__FILE__, __LINE__, $error);
        return ERROR;
    }
    if ($reference_sample !~ /^-?\d+\.?\d*$/) # should be double
    {
        my $error = "reference_sample: invalid value $reference_sample";
        $error_log->fatal_msg(__FILE__, __LINE__, $error);
        return ERROR;
    }
    if ($latitude !~ /^-?\d+\.\d+$/) # should be double
    {
        my $error = "latitude: invalid value $latitude";
        $error_log->fatal_msg(__FILE__, __LINE__, $error);
        return ERROR;
    }
    if ($longitude !~ /^-?\d+\.\d+$/) # should be double
    {
        my $error = "longitude: invalid value $longitude";
        $error_log->fatal_msg(__FILE__, __LINE__, $error);
        return ERROR;
    }
    if ($projection_y !~ /^-?\d+\.?\d*$/) # should be double
    {
        my $error = "projection_y: invalid value $projection_y";
        $error_log->fatal_msg(__FILE__, __LINE__, $error);
        return ERROR;
    }
    if ($projection_x !~ /^-?\d+\.?\d*$/) # should be double
    {
        my $error = "projection_x: invalid value $projection_x";
        $error_log->fatal_msg(__FILE__, __LINE__, $error);
        return ERROR;
    }
    if ($elevation !~ /^-?\d+\.?\d*$/) # should be double
    {
        my $error = "elevation: invalid value $elevation";
        $error_log->fatal_msg(__FILE__, __LINE__, $error);
        return ERROR;
    }
    if ($pixel_size_x !~ /^-?\d+\.?\d*$/) # should be double
    {
        my $error = "pixel_size_x: invalid value $pixel_size_x";
        $error_log->fatal_msg(__FILE__, __LINE__, $error);
        return ERROR;
    }
    if ($pixel_size_y !~ /^-?\d+\.?\d*$/) # should be double
    {
        my $error = "pixel_size_y: invalid value $pixel_size_y";
        $error_log->fatal_msg(__FILE__, __LINE__, $error);
        return ERROR;
    }
    if ($chip_size_lines !~ /^-?\d+\.?\d*$/) # should be double
    {
        my $error = "chip_size_lines: invalid value $chip_size_lines";
        $error_log->fatal_msg(__FILE__, __LINE__, $error);
        return ERROR;
    }
    if ($chip_size_samples !~ /^-?\d+\.?\d*$/) # should be double
    {
        my $error = "chip_size_samples: invalid value $chip_size_samples";
        $error_log->fatal_msg(__FILE__, __LINE__, $error);
        return ERROR;
    }
    if ($source !~ /^\w+\+?$/) # should be string 
    {
        my $error = "source: invalid value $source";
        $error_log->fatal_msg(__FILE__, __LINE__, $error);
        return ERROR;
    }
    if ($chip_type !~ /^\w+$/) # should be string
    {
        my $error = "chip_type: invalid value $chip_type";
        $error_log->fatal_msg(__FILE__, __LINE__, $error);
        return ERROR;
    }
    if ($projection !~ /^\w+$/) # should be string
    {
        my $error = "projection: invalid value $projection";
        $error_log->fatal_msg(__FILE__, __LINE__, $error);
        return ERROR;
    }
    if ($zone !~ /^\d+$/) # should be int
    {
        my $error = "zone: invalid value $zone";
        $error_log->fatal_msg(__FILE__, __LINE__, $error);
        return ERROR;
    }
    if ($date !~ /^\d{2}-\d{2}-\d{4}$/) # should be date: MM-DD-YYYY
    {
        my $error = "date: invalid value $date";
        $error_log->fatal_msg(__FILE__, __LINE__, $error);
        return ERROR;
    }

    if ($abs_rel_flag !~ /^(ABS|REL)$/) # Only ABS and REL are valid
    {
        my $error = "abs_rel_flag: invalid value $abs_rel_flag";
        $error_log->fatal_msg(__FILE__, __LINE__, $error);
        return ERROR;
    }

    if ($chip_data_type !~ /^[\w\*]+$/) # should be string, can have '*'
    {
        my $error = "chip_data_type: invalid value $chip_data_type";
        $error_log->fatal_msg(__FILE__, __LINE__, $error);
        return ERROR;
    }

    return SUCCESS;
}

