/*****************************************************************************
NAME: ias_gcp_write_gcplib

PURPOSE: Creates a GCPLib file from an array of GCP Records.

RETURN VALUE: type = int

    Value    Description
    -------  ----------------------------------------------------------------
    SUCCESS  Successfully retrieved the GCP information.
    ERROR    Failure to get the GCP information.

*****************************************************************************/

/* Standard C Includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* IAS Includes */
#include "ias_types.h"
#include "ias_gcp.h"
#include "ias_logging.h"
#include "ias_miscellaneous.h"

int ias_gcp_write_gcplib
(
    const char *gcplib_filename,       /* I: Output GCP filename */
    const IAS_GCP_RECORD *gcp_records, /* I: Structure of chip information */
    int number_of_gcps,                /* I: Number of ground control points */
    int overall_gcp_version            /* I: Overall GCP version */
)
{
    FILE *output_fd;            /* Pointer to the GCPLib output file */
    int status;                 /* Status of return from function */
    int index;                  /* Loop control variables */
    const char *data_type;      /* To hold the string version of the type */
    
    output_fd = fopen(gcplib_filename, "w");
    if(output_fd == NULL)
    {
        IAS_LOG_ERROR("Opening output GCPLib file %s", gcplib_filename);
        return ERROR;
    }

    /* Output header information */
    status = fprintf(output_fd,
        "# Unique Ground Control Point (GCP) ID - integer\n"
        "# GCP Active Flag - string (Y/N)\n"
        "# GCP date updated - string in form mm-dd-yyyy\n"
        "# GCP update reason - string\n"
        "# GCP version number - integer\n"
        "# GCP WRS-2 path - integer\n"
        "# GCP WRS-2 row - integer\n"
        "# GCP source - string\n"
        "# GCP latitude of reference point in degrees - double\n"
        "# GCP longitude of reference point in degrees - double\n"
        "# GCP elevation of reference point in meters - double\n"
        "# GCP type -string (CONTROL or VALIDATION)\n"
        "# Image chip ID - string\n"
        "# Image chip Active Flag - string (Y/N)\n"
        "# Image chip date updated - string in form mm-dd-yyyy\n"
        "# Image chip version number - integer\n"
        "# Image chip source WRS-2 path - integer\n"
        "# Image chip source WRS-2 row - integer\n"
        "# Image chip reference line - double\n"
        "# Image chip reference sample - double\n"
        "# Image chip reference point projection X in meters - double\n"
        "# Image chip reference point projection Y in meters - double\n"
        "# Image chip pixel size x value in meters - double\n"
        "# Image chip pixel size y value in meters - double\n"
        "# Image chip size in lines - double\n"
        "# Image chip size in samples - double\n"
        "# Image chip sensor - string\n"
        "# Image chip selection method - string\n"
        "# Image chip Moravec rank if selection method is MORAVEC, "
            "zero otherwise - integer\n"
        "# Image chip projection - string\n"
        "# Image chip projection UTM zone - integer\n"
        "# Image chip type - string ABS (Absolute) or REL (relative)\n"
        "# Image chip acquisition date - string in form mm-dd-yyyy\n"
        "# Image chip data type - string\n"
        "VERSION = %d\n"
        "BEGIN\n", overall_gcp_version);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing header to GCPLib file");
        fclose(output_fd);
        return ERROR;
    }
    status = fprintf(output_fd, "%d\n", number_of_gcps);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing to GCPLib file");
        fclose(output_fd);
        return ERROR;
    }

    for (index = 0; index < number_of_gcps; index++)
    {
        /* Determine the string representation for the data type */
        status = ias_misc_convert_data_type_to_string(
            gcp_records[index].chip_data_type, &data_type);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Getting string for GCP data type %d for point %d",
                    gcp_records[index].chip_data_type, index);
            fclose(output_fd);
            return ERROR;
        }

        /* Write the record to the file */
        status = fprintf(output_fd,
            "%s %s %s %s %d %d %d %s %.7f %.7f %.3f %s "
            /* next is image chip ID through chip size */
            "%s %s %s %d %d %d %lf %lf %.3f %.3f %lf %lf %lf %lf "
            /* and chip sensor to the end */
            "%s %s %d %s %d %s %s %s\n",
            gcp_records[index].point_id,
            gcp_records[index].point_active,
            gcp_records[index].point_date_updated,
            gcp_records[index].update_reason,
            gcp_records[index].point_version,
            gcp_records[index].point_wrs_path,
            gcp_records[index].point_wrs_row,
            gcp_records[index].source,
            gcp_records[index].latitude,
            gcp_records[index].longitude,
            gcp_records[index].elevation,
            gcp_records[index].chip_type,
            gcp_records[index].chip_name,
            gcp_records[index].chip_active,
            gcp_records[index].chip_date_updated,
            gcp_records[index].chip_version,
            gcp_records[index].chip_wrs_path,
            gcp_records[index].chip_wrs_row,
            gcp_records[index].reference_line,
            gcp_records[index].reference_sample,
            gcp_records[index].projection_x,
            gcp_records[index].projection_y,
            gcp_records[index].pixel_size_x,
            gcp_records[index].pixel_size_y,
            gcp_records[index].chip_size_lines,
            gcp_records[index].chip_size_samples,
            gcp_records[index].chip_sensor,
            gcp_records[index].chip_select_method,
            gcp_records[index].moravec_rank,
            gcp_records[index].projection,
            gcp_records[index].zone,
            gcp_records[index].absolute_or_relative,
            gcp_records[index].date,
            data_type);
        if (status < 0)
        {
            IAS_LOG_ERROR("Writing to GCPLib file");
            fclose(output_fd);
            return ERROR;
        }
    }

    if (fclose(output_fd) != 0)
    {
        IAS_LOG_ERROR("Closing output GCPLib file %s", gcplib_filename);
        return ERROR;
    }

    return SUCCESS;
}
