#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include "cgreen/cgreen.h"
#include "ias_logging.h"
#include "ias_const.h"
#include "ias_gcp.h"

/* Enable this to generate new test data for the test_read_image unit test */
/* #define CREATE_READ_IMAGE_TEST_DATA 1 */

/* distance in meters from the center of the Earth to the equator */
const double semi_major_axis = 6378137.0000;
/* ratio describing polar flattening or the Earth's deviation from an 
    exact sphere (WGS84 standard) */
const double ellipticity = 0.00335281066474;

Ensure test_read_write_correlation()
{
    IAS_GCP_RESULTS *gcp;       /* GCP data structures (1 per pt) */
    IAS_GCP_RESULTS *gcp2;      /* GCP data structures (1 per pt) */
    const char *gcpdata_filename1 = "gcpdata_old_test_data.txt";
    const char *gcpdata_filename2 = "gcpdata_old_test_data.txt.out";
    int numgcp = 0;             /* number of GCPs used in correlation */
    int status;                 /* function return status */
    int match = TRUE;
    char buffer1[256] = {""};
    char buffer2[256] = {""};
    FILE *stream1 = NULL;
    FILE *stream2 = NULL;

    /* test reading correlation data */
    status = ias_gcp_read_correlation_results(gcpdata_filename1, 
        semi_major_axis, ellipticity, &gcp, &numgcp);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
        return;

    /* test writing correlation data */
    status = ias_gcp_write_correlation_results(gcpdata_filename2, gcp, numgcp);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
        return;

    /* test reading the written correlation data */
    status = ias_gcp_read_correlation_results(gcpdata_filename2, 
        semi_major_axis, ellipticity, &gcp2, &numgcp);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
        return;

    /* compare what was read the first time to what was read after
       file was written */
    assert_true(memcmp(gcp, gcp2, sizeof(*gcp)) == 0);
    free(gcp);
    free(gcp2);

    /* diff the two files to make sure they're the same */
    stream1 = fopen(gcpdata_filename1, "r");
    if (stream1 == NULL)
    {
        printf("Error opening %s\n", gcpdata_filename1);
        exit(EXIT_FAILURE);
    }
    stream2 = fopen(gcpdata_filename2, "r");
    if (stream2 == NULL)
    {
        printf("Error opening %s\n", gcpdata_filename2);
        exit(EXIT_FAILURE);
    }
    while(!feof(stream1) && !feof(stream2))
    {
        fgets(buffer1,sizeof(buffer1),stream1);
        fgets(buffer2,sizeof(buffer2),stream2);
        if(strcmp(buffer1,buffer2) != 0)
        {
            match = FALSE;
            printf("Mismatched buffers: buffer1 %s\n", buffer1);
            printf("Mismatched buffers: buffer2 %s\n", buffer2);
            break;
        }
    }
    fclose(stream1);
    fclose(stream2);
    stream1 = NULL;
    stream2 = NULL;
    assert_equal(match, TRUE);
}

Ensure test_read_correlation_error_checking()
{
    IAS_GCP_RESULTS *gcp;       /* GCP data structures (1 per pt) */
    IAS_GCP_RESULTS *gcp2;      /* GCP data structures (1 per pt) */
    IAS_GCP_RESULTS *gcp3;      /* GCP data structures (1 per pt) */
    const char *gcpdata_filename1 = "gcpdata_old_test_data_err_1.txt";
    const char *gcpdata_filename2 = "gcpdata_old_test_data_err_2.txt";
    const char *gcpdata_filename3 = "gcpdata_old_test_data_err_3.txt";
    int numgcp = 0;             /* number of GCPs used in correlation */
    int status;                 /* function return status */
    int log_level;

    log_level = ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* test reading correlation data, check missing required field */
    status = ias_gcp_read_correlation_results(gcpdata_filename1,
        semi_major_axis, ellipticity, &gcp, &numgcp);
    assert_not_equal(status, SUCCESS);

    ias_log_set_output_level(log_level);

    /* test reading correlation data, check optional field missing is ok */
    status = ias_gcp_read_correlation_results(gcpdata_filename2,
        semi_major_axis, ellipticity, &gcp2, &numgcp);
    assert_equal(status, SUCCESS);

    log_level = ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* test reading correlation data, check old style correlation file */
    status = ias_gcp_read_correlation_results(gcpdata_filename3,
        semi_major_axis, ellipticity, &gcp3, &numgcp);
    assert_not_equal(status, SUCCESS);

    free(gcp);
    free(gcp2);
    free(gcp3);
}

Ensure test_read_gcplib()
{
    int status;                    /* function return status */
    const char *gcplib_filename;   /* Name of the GCPLIB file */
    IAS_GCP_RECORD *gcp_lib;       /* Structure of chip information */
    int numgcp;                    /* number of ground control points */
    int total_gcps = 633;          /* Total taken from test file */
    int overall_gcp_version;       /* Overall GCP version number */

    gcplib_filename = "gcplib_new_test_data_1.txt";

    /* test ias_gcp_read_gcplib */
    status = ias_gcp_read_gcplib(gcplib_filename, &gcp_lib, &numgcp, 
        &overall_gcp_version);
    assert_equal(status, SUCCESS);
    assert_equal(numgcp, total_gcps);
    assert_equal(overall_gcp_version, 1);

    free(gcp_lib);
}

Ensure test_read_gcplib_filtered_season()
{
    int status;                    /* Function return status */
    const char *gcplib_filename;   /* Name of the GCPLIB file */
    IAS_GCP_RECORD *gcp_lib;       /* Structure of chip information */
    int numgcp;                    /* Number of ground control points */
    int overall_gcp_version;       /* Overall GCP version number */
    int version_verify = 1;        /* Version taken from file */
    char season[IAS_GCP_NUM_SEASONS][IAS_GCP_SEASON_LEN] 
             = {"WINTER", "SPRING", "SUMMER", "FALL"},
         chip_source[IAS_GCP_NUM_CHIP_SOURCES][IAS_GCP_SOURCE_SIZE] 
             = {"", "", "", "", "", "", "", "", "", ""};
    char chip_type[IAS_GCP_TYPE_SIZE] = "";
    int begin_date[] = {0, 0},
        end_date[] = {0, 0},
        total_gcps = 9;            /* Total taken from test file */

    gcplib_filename = "gcplib_new_test_data_2.txt";

    /* Test ias_gcp_read_gcplib_filtered */
    status = ias_gcp_read_gcplib_filtered(gcplib_filename, begin_date,
            end_date, season, chip_source, chip_type, &gcp_lib, &numgcp, 
            &overall_gcp_version);
    assert_equal(status, SUCCESS);
    assert_equal(numgcp, total_gcps);
    assert_equal(overall_gcp_version, version_verify);
    free(gcp_lib);

    char season2[IAS_GCP_NUM_SEASONS][IAS_GCP_SEASON_LEN] 
        = {"SUMMER", "", "", ""};
    total_gcps = 2;
    gcplib_filename = "gcplib_new_test_data_3.txt";

    /* Test ias_gcp_read_gcplib_filtered */
    status = ias_gcp_read_gcplib_filtered(gcplib_filename, begin_date, end_date,
            season2, chip_source, chip_type, &gcp_lib, &numgcp, 
            &overall_gcp_version);
    assert_equal(status, SUCCESS);
    assert_equal(numgcp, total_gcps);
    assert_equal(overall_gcp_version, version_verify);
    free(gcp_lib);
}

/* Test filtering by GCP source */
Ensure test_write_and_read_gcplib_filtered_source()
{
    int status;                 /* Status holder */
    int index;                  /* Loop variable */
    const char *gcplib_filename = "gcplib_test_gcp_source.txt";
    const char *write_gcplib_filename = "gcplib_test_write_results.txt";
    const char *verify_gcplib_filename = "gcplib_test_write_results.verify";
    char command[PATH_MAX];     /* System command string */
    IAS_GCP_RECORD *gcp_lib;    /* GCP library pointer */
    IAS_GCP_RECORD expected_lib[2]; /* Expected library */
    int num_read_gcps;          /* Number of GCPs read from file */
    char season[IAS_GCP_NUM_SEASONS][IAS_GCP_SEASON_LEN]
            = {"WINTER", "SPRING", "SUMMER", "FALL"};
    char chip_source[IAS_GCP_NUM_CHIP_SOURCES][IAS_GCP_SOURCE_SIZE]
            = {"GLS", "", "", "", "", "", "", "", "", ""};
    char chip_type[IAS_GCP_TYPE_SIZE] = "";
    int begin_date[] = {0, 0};
    int end_date[] = {0, 0};
    int expected_gcps = 2;      /* Number of GCP that should be returned */
    int overall_gcp_version;       /* Overall GCP version number */
    int version_verify = 1;        /* Version taken from file */

    /* Expected records */
    expected_lib[0] = (IAS_GCP_RECORD){"170400007", "Y", "11-27-2007", 
        "ORIGINAL", 1, 17, 40, "GLS", -8.070515, -82.451505, -6.000000, 
        "CONTROL", "0170400007_01", "Y", "11-27-2007", 1, 17, 40, 31.000000, 
        31.000000, 3150550.000000, 355920.000000, 10.000000, 10.000000, 
        64.000000, 64.000000, "ETM+", "UNKNOWN", -1, "UTM", 1, "REL", 
        "06-01-2005", IAS_BYTE};
    expected_lib[1] = (IAS_GCP_RECORD){"170400009", "Y", "11-27-2007", 
        "ORIGINAL", 1, 17, 40, "GLS", 28.470213, -82.451304, -6.000000, 
        "CONTROL", "0170400009_01", "Y", "11-27-2007", 1, 17, 40, 31.000000, 
        31.000000, 3150150.000000, 357920.000000, 10.000000, 10.000000, 
        64.000000, 64.000000, "ETM+", "UNKNOWN", -1, "UTM", 1, "REL", 
        "01-01-2005", IAS_BYTE};

    /* Read in GCP library */
    status = ias_gcp_read_gcplib_filtered(gcplib_filename, begin_date, end_date,
            season, chip_source, chip_type, &gcp_lib, &num_read_gcps,
            &overall_gcp_version);
    assert_equal(status, SUCCESS);
    assert_equal(num_read_gcps, expected_gcps);
    assert_equal(overall_gcp_version, version_verify);
    
    /* Loop through and verify the complete IAS_GCP_RECORD */
    for(index = 0; index < expected_gcps; index++)
    {
        IAS_GCP_RECORD read_record = gcp_lib[index];
        IAS_GCP_RECORD expected_record = expected_lib[index];

        /* Check that all values were read correctly using first record
        * First 5 comparisons*/
        assert_string_equal(read_record.point_id, expected_record.point_id);
        assert_string_equal(read_record.point_active, 
            expected_record.point_active);
        assert_string_equal(read_record.point_date_updated, 
            expected_record.point_date_updated);
        assert_string_equal(read_record.update_reason, 
            expected_record.update_reason);
        assert_equal(read_record.point_version, expected_record.point_version);
    
        /* Next 5 comparisons 5-10 */
        assert_equal(read_record.point_wrs_path, 
            expected_record.point_wrs_path);
        assert_equal(read_record.point_wrs_row, expected_record.point_wrs_row);
        assert_string_equal(read_record.source, expected_record.source);
        assert_double_equal(read_record.latitude, expected_record.latitude);
        assert_double_equal(read_record.longitude, expected_record.longitude);
    
        /* Next 5 comparisons 10-15 */
        assert_double_equal(read_record.elevation, expected_record.elevation);
        assert_string_equal(read_record.chip_type, expected_record.chip_type);
        assert_string_equal(read_record.chip_name, expected_record.chip_name);
        assert_string_equal(read_record.chip_active, 
            expected_record.chip_active);
        assert_string_equal(read_record.chip_date_updated, 
            expected_record.chip_date_updated);
    
        /* Next 5 comparisons 15-20 */
        assert_equal(read_record.chip_version, expected_record.chip_version);
        assert_equal(read_record.chip_wrs_path, expected_record.chip_wrs_path);
        assert_equal(read_record.chip_wrs_row, expected_record.chip_wrs_row);
        assert_double_equal(read_record.reference_line, 
            expected_record.reference_line);
        assert_double_equal(read_record.reference_sample, 
            expected_record.reference_sample);
    
        /* Next 5 comparisons 20-25 */
        assert_double_equal(read_record.projection_x, 
            expected_record.projection_x);
        assert_double_equal(read_record.projection_y, 
            expected_record.projection_y);
        assert_double_equal(read_record.pixel_size_x, 
            expected_record.pixel_size_x);
        assert_double_equal(read_record.pixel_size_y, 
            expected_record.pixel_size_y);
        assert_double_equal(read_record.chip_size_lines, 
            expected_record.chip_size_lines);
    
        /* Next 5 comparisons 25-30 */
        assert_double_equal(read_record.chip_size_samples,
            expected_record.chip_size_samples);
        assert_string_equal(read_record.chip_sensor, 
            expected_record.chip_sensor);
        assert_string_equal(read_record.chip_select_method,
            expected_record.chip_select_method);
        assert_equal(read_record.moravec_rank, expected_record.moravec_rank);
        assert_string_equal(read_record.projection, 
            expected_record.projection);
    
        /* Last 4 comparisons 30-34 */
        assert_equal(read_record.zone, expected_record.zone);
        assert_string_equal(read_record.absolute_or_relative, 
            expected_record.absolute_or_relative);
        assert_string_equal(read_record.date,expected_record.date);
        assert_equal(read_record.chip_data_type, 
            expected_record.chip_data_type);
    }

    /* Write out GCP library that was just read in and verified */
    status = ias_gcp_write_gcplib(write_gcplib_filename, gcp_lib,
        num_read_gcps, overall_gcp_version);
    assert_equal(status, SUCCESS);
    
    /* Prepare diff command */
    status = snprintf(command, sizeof(command), "diff %s %s ",
        verify_gcplib_filename, write_gcplib_filename);
    if (status < 0 || status >= sizeof(command))
    {
        assert_equal(0, 1);
        free(gcp_lib);
        return;
    }

    /* Execute diff to verify ias_gcp_write_gcplib */
    status = system(command);
    assert_equal(status, SUCCESS);

    /* Free the GCP library */
    free(gcp_lib);
}

/* Test filtering by GCP type */
Ensure test_read_gcplib_filtered_type()
{
    int status; 
    const char *gcplib_filename = "gcplib_test_gcp_type.txt";
    IAS_GCP_RECORD *gcp_lib;
    int numgcp;
    int overall_gcp_version;       /* Overall GCP version number */
    int version_verify = 1;        /* Version taken from file */

    char season[IAS_GCP_NUM_SEASONS][IAS_GCP_SEASON_LEN]
            = {"WINTER", "SPRING", "SUMMER", "FALL"};
    char chip_source[IAS_GCP_NUM_CHIP_SOURCES][IAS_GCP_SOURCE_SIZE]
            = {"", "", "", "", "", "", "", "", "", ""};
    char chip_type[IAS_GCP_TYPE_SIZE] = "VALIDATION";
    int begin_date[] = {0, 0};
    int end_date[] = {0,0};
    int expected_gcps = 2;        /* Number of GCP that should be returned */

    status = ias_gcp_read_gcplib_filtered(gcplib_filename, begin_date, end_date,
            season, chip_source, chip_type, &gcp_lib, &numgcp,
            &overall_gcp_version);
    assert_equal(status, SUCCESS);
    assert_equal(numgcp, expected_gcps);
    assert_equal(overall_gcp_version, version_verify);
    free(gcp_lib);
}

Ensure test_read_empty_gcplib()
{
    int status;                    /* function return status */
    const char *gcplib_filename;   /* Name of the GCPLIB file */
    IAS_GCP_RECORD *gcp_lib;       /* Structure of chip information */
    int numgcp;                    /* number of ground control points */
    int overall_gcp_version;       /* Overall GCP version number */
    int version_verify = 1;        /* Version taken from file */

    gcplib_filename = "empty_gcplib.txt";

    /* test ias_gcp_read_gcplib */
    status = ias_gcp_read_gcplib(gcplib_filename, &gcp_lib, &numgcp, 
        &overall_gcp_version);
    assert_equal(status, SUCCESS);
    assert_equal(numgcp, 0);
    assert_equal(overall_gcp_version, version_verify);
    free(gcp_lib);
}

void read_chips
(
    IAS_GCP_IMAGE_CHIP_IO *chip_io, /* I: Image chip container IO structure */
    const char *byte_filename,      /* I: Name of the BYTE chip file */
    const char *ui2_filename        /* I: Name of the UI*2 chip file */
)
{
    int ref_win_lines = 64;         /* size of the ref GCP chip (lines) */
    int ref_win_samples = 64;       /* size of the ref GCP chip (samples) */
    int num_words = ref_win_lines * ref_win_samples;
    int status;                     /* return value */
    int i;
    int log_level;
    IAS_R4_TYPE *ref_buf;           /* max sized reference correlation window */
    IAS_UI2_TYPE ui2_buf[num_words];

    /* allocate memory for chip buffer for the library read routine */
    ref_buf = malloc(num_words * sizeof(IAS_R4_TYPE));
    assert_true(ref_buf != NULL);
    if (ref_buf == NULL)
        return;

    /* test ias_gcp_read_image_chip with the BYTE chip */
    status = ias_gcp_read_image_chip(chip_io, byte_filename, ref_win_lines,
        ref_win_samples, IAS_BYTE, IAS_GCP_CONVERT_TO_FLOAT, ref_buf);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
    {
        free(ref_buf);
        return;
    }

    /* Check that the data in the buffer matches the values the file contains
       (written to the benchmarked file at an earlier time with the code
        below, using this same pattern) */
    for (i = 0; i < num_words; i++)
    {
        if ((IAS_BYTE_TYPE)ref_buf[i] != ((i * 9) % 255))
            assert_equal((IAS_BYTE_TYPE)ref_buf[i], ((i * 9) % 255));
    }

    /* Clear buffer */
    memset(ref_buf, 0, (num_words * sizeof(IAS_R4_TYPE)));

    /* test ias_gcp_read_image_chip with the UI*2 chip */
    status = ias_gcp_read_image_chip(chip_io, ui2_filename, ref_win_lines,
        ref_win_samples, IAS_UI2, IAS_GCP_NO_CONVERSION, ui2_buf);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
    {
        free(ref_buf);
        return;
    }

    /* Check that the data in the buffer matches the values the file contains
       (written to the benchmarked file at an earlier time with the code
        below, using this same pattern) */
    for (i = 0; i < num_words; i++)
    {
        if (ui2_buf[i] != ((i * 2948) % 65535))
            assert_equal(ui2_buf[i], ((i * 2948) % 65535));
    }

    /* Try to read a chip that doesn't exist */
    log_level = ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = ias_gcp_read_image_chip(chip_io, "no_chip", ref_win_lines,
        ref_win_samples, IAS_UI2, IAS_GCP_CONVERT_TO_FLOAT, ref_buf);
    ias_log_set_output_level(log_level);
    assert_equal(status, ERROR);

    free(ref_buf);
    ref_buf = NULL;

    /* Close the open container */
    status = ias_gcp_close_image_chip_container(chip_io);
    assert_equal(status, SUCCESS);

    /* Check for an error if closing a NULL pointer */
    log_level = ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = ias_gcp_close_image_chip_container(NULL);
    assert_equal(status, ERROR);
    ias_log_set_output_level(log_level);

}

Ensure test_read_image_chip_files()
{
    IAS_GCP_IMAGE_CHIP_IO *chip_io; /* Image chip container IO structure */
    const char *container_name = ".";
    const char *name;
    IAS_GCP_CONTAINER_TYPE container_type;

    chip_io = ias_gcp_open_image_chip_container(container_name,
        IAS_GCP_CONTAINER_DIR, IAS_READ);
    assert_not_equal(chip_io, NULL);
    if (!chip_io)
        return;

    /* Verify the expected container name was opened */
    name = ias_gcp_get_image_chip_container_name(chip_io);
    assert_not_equal(name, NULL);
    assert_equal(strcmp(name, container_name), 0);
    
    /* Verify the expected container type was found */
    container_type = ias_gcp_get_image_chip_container_type(chip_io);
    assert_equal(container_type, IAS_GCP_CONTAINER_DIR);

    read_chips(chip_io, "gcp_chip_byte", "gcp_chip_ui2");
}

Ensure test_read_image_chip_tarball()
{
    IAS_GCP_IMAGE_CHIP_IO *chip_io; /* Image chip container IO structure */
    const char *container_name = "image_chips.tbz";
    const char *name;
    IAS_GCP_CONTAINER_TYPE container_type;

    chip_io = ias_gcp_open_image_chip_container(container_name,
        IAS_GCP_CONTAINER_TARBALL, IAS_READ);
    assert_not_equal(chip_io, NULL);
    if (!chip_io)
        return;

    /* Verify the expected container name was opened */
    name = ias_gcp_get_image_chip_container_name(chip_io);
    assert_not_equal(name, NULL);
    assert_equal(strcmp(name, container_name), 0);
    
    /* Verify the expected container type was found */
    container_type = ias_gcp_get_image_chip_container_type(chip_io);
    assert_equal(container_type, IAS_GCP_CONTAINER_TARBALL);

    read_chips(chip_io, "image_chip2_byte", "image_chip2_ui2");
}

Ensure test_read_image_chip_search_for_tarball()
{
    IAS_GCP_IMAGE_CHIP_IO *chip_io; /* Image chip container IO structure */
    const char *name;
    IAS_GCP_CONTAINER_TYPE container_type;

    chip_io = ias_gcp_open_image_chip_container(".",
        IAS_GCP_CONTAINER_SEARCH, IAS_READ);
    assert_not_equal(chip_io, NULL);
    if (!chip_io)
        return;

    /* Verify the expected container name was opened */
    name = ias_gcp_get_image_chip_container_name(chip_io);
    assert_not_equal(name, NULL);
    assert_equal(strcmp(name, "./image_chips.tar"), 0);

    /* Verify the expected container type was found */
    container_type = ias_gcp_get_image_chip_container_type(chip_io);
    assert_equal(container_type, IAS_GCP_CONTAINER_TARBALL);

    read_chips(chip_io, "image_chip_byte", "image_chip_ui2");
}

Ensure test_write_image_chip_tarball()
{
    IAS_GCP_IMAGE_CHIP_IO *chip_io; /* Image chip container IO structure */
    IAS_GCP_IMAGE_CHIP_IO *read_chip_io; /* Image chip container IO structure */
    int log_level;
    int status;
    struct stat stat_buf;
    float read_chip_buffer[64 * 64];
    unsigned char write_chip_buffer_byte[64 * 64];
    unsigned short int ui2_buffer[64 * 64];
    int index;
    #define NO_DATA_TARBALL_NAME "write_no_data_test.tar"
    #define TARBALL_NAME "write_test.tar"
    #define DIR_NAME "write_test"

    /* This should fail since search mode is not supported for writes */
    log_level = ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    chip_io = ias_gcp_open_image_chip_container(".",
        IAS_GCP_CONTAINER_SEARCH, IAS_WRITE);
    ias_log_set_output_level(log_level);
    assert_equal(chip_io, NULL);

    /* Open a file for writing */
    unlink(NO_DATA_TARBALL_NAME);
    chip_io = ias_gcp_open_image_chip_container(NO_DATA_TARBALL_NAME,
        IAS_GCP_CONTAINER_TARBALL, IAS_WRITE);
    assert_not_equal(chip_io, NULL);

    /* check name and type */
    assert_equal(strcmp(ias_gcp_get_image_chip_container_name(chip_io),
        NO_DATA_TARBALL_NAME), 0);
    assert_equal(ias_gcp_get_image_chip_container_type(chip_io),
        IAS_GCP_CONTAINER_TARBALL);

    /* Close it without writing any chips */
    status = ias_gcp_close_image_chip_container(chip_io);
    assert_equal(status, SUCCESS);
    chip_io = NULL;

    /* Verify a file was created */
    status = stat(NO_DATA_TARBALL_NAME, &stat_buf);
    assert_equal(status, 0);

    /* Use the tar_chips directory as a source of chips to write */
    read_chip_io = ias_gcp_open_image_chip_container("tar_chips",
        IAS_GCP_CONTAINER_SEARCH, IAS_READ);
    assert_not_equal(read_chip_io, NULL);
    if (!read_chip_io)
        return;

    /* Open a file for writing */
    unlink(TARBALL_NAME);
    chip_io = ias_gcp_open_image_chip_container(TARBALL_NAME,
        IAS_GCP_CONTAINER_TARBALL, IAS_WRITE);
    assert_not_equal(chip_io, NULL);
    if (!chip_io)
        return;

    /* Read a chip */
    status = ias_gcp_read_image_chip(read_chip_io, "1320500012_01", 64, 64,
        IAS_BYTE, IAS_GCP_CONVERT_TO_FLOAT, read_chip_buffer);
    assert_equal(status, SUCCESS);

    /* Convert the chip back to byte data */
    for (index = 0; index < 64 * 64; index++)
        write_chip_buffer_byte[index] = (unsigned char)read_chip_buffer[index];

    /* Write the chip */
    status = ias_gcp_write_image_chip(chip_io, "1320500012_01", 64, 64,
        IAS_BYTE, write_chip_buffer_byte);
    assert_equal(status, SUCCESS);

    /* Copy a 16-bit chip */
    status = ias_gcp_read_image_chip(read_chip_io, "1620460032_01", 64, 64,
        IAS_UI2, IAS_GCP_NO_CONVERSION, ui2_buffer);
    assert_equal(status, SUCCESS);

    status = ias_gcp_write_image_chip(chip_io, "1620460032_01", 64, 64,
        IAS_UI2, ui2_buffer);
    assert_equal(status, SUCCESS);

    /* Close the input container */
    status = ias_gcp_close_image_chip_container(read_chip_io);
    assert_equal(status, SUCCESS);
    read_chip_io = NULL;

    /* Close the container being written to */
    status = ias_gcp_close_image_chip_container(chip_io);
    assert_equal(status, SUCCESS);
    chip_io = NULL;

    /* Verify the contents of the created tarball by untarring the data and
       and verifying the md5sum of the image chips */
    status = system("rm -rf temp_untar && mkdir -p temp_untar && cd temp_untar "
        "&& tar xf ../" TARBALL_NAME
        "&& md5sum --check ../tar_chips/md5sum_write_test.txt");
    assert_equal(status, 0);
    if (status == 0)
    {
        /* remove the directory if the test passed */
        system("rm -r temp_untar");
    }

    /* Make a new directory for testing writing to a directory */
    status = system("rm -rf " DIR_NAME);
    assert_equal(status, 0);
    if (status != 0)
        return;
    status = mkdir(DIR_NAME, 0777);
    assert_equal(status, 0);
    if (status != 0)
        return;

    /* Open the directory container */
    chip_io = ias_gcp_open_image_chip_container(DIR_NAME,
        IAS_GCP_CONTAINER_DIR, IAS_WRITE);
    assert_not_equal(chip_io, NULL);
    if (!chip_io)
        return;

    /* Write the chips there */
    status = ias_gcp_write_image_chip(chip_io, "1320500012_01", 64, 64,
        IAS_BYTE, write_chip_buffer_byte);
    assert_equal(status, SUCCESS);

    status = ias_gcp_write_image_chip(chip_io, "1620460032_01", 64, 64,
        IAS_UI2, ui2_buffer);
    assert_equal(status, SUCCESS);

    /* Close the container being written to */
    status = ias_gcp_close_image_chip_container(chip_io);
    assert_equal(status, SUCCESS);
    chip_io = NULL;

    /* Verify the contents of the directory by verifying the md5sum of the
       image chips */
    status = system("cd " DIR_NAME
        " && md5sum --check ../tar_chips/md5sum_write_test.txt");
    assert_equal(status, 0);
    if (status == 0)
    {
        system("rm -r " DIR_NAME);
    }
}

int main(int argc, char **argv)
{
    ias_log_initialize("GCP TEST");
    TestSuite *suite = create_test_suite();
    add_test(suite, test_read_write_correlation);
    add_test(suite, test_read_correlation_error_checking);
    add_test(suite, test_read_gcplib);
    add_test(suite, test_read_gcplib_filtered_season);
    add_test(suite, test_write_and_read_gcplib_filtered_source);
    add_test(suite, test_read_gcplib_filtered_type);
    add_test(suite, test_read_empty_gcplib);
    add_test(suite, test_read_image_chip_files);
    add_test(suite, test_read_image_chip_tarball);
    add_test(suite, test_read_image_chip_search_for_tarball);
    add_test(suite, test_write_image_chip_tarball);
    if (argc > 1) {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}
