#!/usr/bin/perl -w

# This program is just for testing the ias_gcp_read_gcplib and 
# ias_gcp_write_gcplib routines in the ias_gcp.pm perl module.
# It reads in some GCP Lib data using ias_gcp_read_gcplib, verifies
# the status returned from that module, calls ias_gcp_write_gcplib,
# and verifies the status returned from ias_gcp_write_gcplib.
use strict;
use ias_const;                  # Defines SUCCESS and ERROR
use Logger;
use ias_gcp      qw(    IAS_GCP_IMAGE_CHIPS_FILENAME
                        IAS_GCP_BZIP2
                        IAS_GCP_GZIP
                        IAS_GCP_NO_COMPRESSION
                        IAS_GCP_NO_MD5SUM
                        IAS_GCP_DO_MD5SUM
                        ias_gcp_read_gcplib 
                        ias_gcp_write_gcplib
                        ias_gcp_get_sizeof_data_type                 
                        ias_gcp_tar_image_chips
                        ias_gcp_create_tar_filename
                    );

# Declare program constant
use constant PROG      => "test_ias_gcp.pl";

# Set up error logger
my $error_log = Logger->new("STDOUT");
if (!$error_log)
{
    print STDERR PROG . ": ERROR: Failed to establish logger object.\n";
    exit FAILURE;
}
$error_log->set_module_name(PROG);
set_default_log($error_log);

# Read the GCP data
my $gcp_filename = "gcplib_new_test_data_1.txt";
my ($return, $overall_gcp_version, @gcp_data) = 
    ias_gcp_read_gcplib($gcp_filename);
if ($return != SUCCESS)
{
    print "Test failed\n";
    print "Call to ias_gcp_read_gcplib returned error\n";
    exit FAILURE;
}

# Write the GCP data
$gcp_filename = "gcplib_out_test_data.txt";
$return = ias_gcp_write_gcplib($gcp_filename, $overall_gcp_version, \@gcp_data);
if ($return == SUCCESS)
{
    my $rc = system("diff gcplib_new_test_data_1.txt gcplib_out_test_data.txt");
    if ($rc == 0)
    {
        print "Successful calls to ias_gcp_read_gcplib and ";
        print "ias_gcp_write_gcplib;\nGCP Lib files match\n";
    }
    else
    {
        print "Test failed\n";
        print "GCP Lib test files differ\n";
        $error_log->close_log();
        exit FAILURE;
    }
}
else
{
    print "Test failed\n";
    print "Call to ias_gcp_write_gcplib returned error\n";
    $error_log->close_log();
    exit FAILURE;
}

# Verify reading a file with an incorrect number of records fails
($return, $overall_gcp_version, @gcp_data) =
    ias_gcp_read_gcplib("GCPLib.bad_record_count");
if ($return == SUCCESS)
{
    print "ERROR: Did not detect a bad number of records\n";
    $error_log->close_log();
    exit FAILURE;
}

# Test the ias_gcp_tar_image_chips routine
if (test_ias_gcp_tar_image_chips() != SUCCESS)
{
    $error_log->fatal_msg("Test ias_gcp_tar_image_chips failed");
    $error_log->close_log();
    exit FAILURE;
}

$error_log->close_log();

exit SUCCESS;

##############################################################################
# Name: test_ias_gcp_tar_image_chips
#
# Description: Test the ias_gcp_tar_image_chips routine from the 
#              ias_gcp module
#
# Parameters: N/A
#
# Returns: SUCCESS/ERROR
##############################################################################
sub test_ias_gcp_tar_image_chips
{
    my $logger = get_default_log();             # Logger
    my $ut_dir = "./tar_chips/";                # Test directory
    my $tar_file;                               # Tar filename
    my $gcplib_file = $ut_dir."GCPLib";         # GCPLib file path
    
    # Prepare GCP hash
    my ($status, $overall_gcp_version, @records) =
        ias_gcp_read_gcplib($gcplib_file);
    if ($status != SUCCESS)
    {
        $logger->fatal_msg("Unable to read in the gcp_lib");
        return ERROR;
    }
        
    # Read in the image chips
    foreach my $gcp (@records)
    {
        my $image_data;                        # Image data
        my $file = $ut_dir.$gcp->{'chip_name'};# Image chip path
        my $fh;                                # File handle for reading chips

        # Determine how many bytes to read for the image chip        
        my $bytes_to_read = $gcp->{'chip_size_lines'} 
            * $gcp->{'chip_size_samples'} 
            * ias_gcp_get_sizeof_data_type($gcp->{'chip_data_type'});

        # Open file handle for reading
        if (!defined open($fh, "<", "$file"))
        {
            $logger->fatal_msg("Image chip $file could not be opened");
            return ERROR;
        }
    
        # Read the image data 
        my $bytes_read = sysread($fh, $image_data, $bytes_to_read);
        if (!defined $bytes_read)
        {
            $logger->fatal_msg("Error reading image chip file '$file': $!");
            close($fh);
            return ERROR;
        }
        close($fh);
        
        # Attached the image data to hash
        $gcp->{'image_chip'} = $image_data;
    }

    # Checking that no test file was created
    $tar_file = ias_gcp_create_tar_filename(IAS_GCP_IMAGE_CHIPS_FILENAME,
        IAS_GCP_BZIP2);
    if (!defined $tar_file)
    {   
        $logger->fatal_msg("Creating tarball filename");
        return ERROR;
    }

    $status = ias_gcp_tar_image_chips(\@records, $ut_dir.$tar_file, 
        IAS_GCP_BZIP2, IAS_GCP_NO_MD5SUM);
    if ($status != SUCCESS)
    {
        $logger->fatal_msg("Creating BZIP2 tarball ");
        return ERROR;   
    }

    # Check that tarball was created
    if (! -e $ut_dir."image_chips.tbz")
    {
        $logger->fatal_msg("BZIP2 Tarball does not exist");
        return ERROR;   
    }

    # Checking that file is created without test in the name
    $tar_file = ias_gcp_create_tar_filename(IAS_GCP_IMAGE_CHIPS_FILENAME,
        IAS_GCP_GZIP);
    if (!defined $tar_file)
    {   
        $logger->fatal_msg("Creating tarball filename");
        return ERROR;
    }

    $status = ias_gcp_tar_image_chips(\@records, $ut_dir.$tar_file, 
        IAS_GCP_GZIP, IAS_GCP_NO_MD5SUM);
    if ($status != SUCCESS)
    {
        $logger->fatal_msg("Creating GZIP tarball");
        return ERROR;   
    }

    # Check that tarball was created
    if (! -e $ut_dir."image_chips.tgz")
    {
        $logger->fatal_msg("GZIP Tarball does not exist");
        return ERROR;   
    }

    # Checking that file is created without compression
    $tar_file = ias_gcp_create_tar_filename(IAS_GCP_IMAGE_CHIPS_FILENAME,
        IAS_GCP_NO_COMPRESSION);
    if (!defined $tar_file)
    {   
        $logger->fatal_msg("Creating tarball filename");
        return ERROR;
    }

    $status = ias_gcp_tar_image_chips(\@records, $ut_dir.$tar_file, 
        IAS_GCP_NO_COMPRESSION, IAS_GCP_DO_MD5SUM);
    if ($status != SUCCESS)
    {
        $logger->fatal_msg("Creating tarball");
        return ERROR;   
    }

    # Check that tarball was created
    if (! -e $ut_dir."image_chips.tar")
    {
        $logger->fatal_msg("Tarball does not exist");
        return ERROR;   
    }

    # Verify the chips are in the tar file
    $status = system("mkdir $ut_dir/untar && cd $ut_dir/untar "
        . "&& tar xf ../image_chips.tar && md5sum --check ../image_chips.md5");
    if ($status != SUCCESS)
    {
        $logger->info_msg("Error checking the contents of the tar file");
        return ERROR;
    }

    # Clean up the output directory
    $status = system("rm -rf $ut_dir/untar && rm -f $ut_dir*.tbz $ut_dir*.tgz"
        . " $ut_dir*.md5 $ut_dir*.tar");
    if ($status != SUCCESS)
    {
        $logger->info_msg("Unable to clean up $ut_dir directory");
    }

    return SUCCESS;
}
