/************************************************************************

NAME:     READWRITE_TESTS

PURPOSE:  Use CGreen to perform unit testing on GEOM Grid IO Library to 
          write and read back the Grid and Grid band data.

*************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <sys/stat.h>

#include "ias_geom_grid.h"
#include "cgreen/cgreen.h"
#include "ias_logging.h"
#include "ias_types.h"
#include "ias_const.h"

/*************************************************************************
           WRITE TEST
**************************************************************************/
Ensure test_ias_geom_grid_write_read()
{
    int i, k, m;
    int status;
    int bandno, nband;
    /* Simulated grid data type  */
    IAS_GEOM_GRID_TYPE gdt;
    IAS_GEOM_GRID_TYPE gdt_read;

    /* Initialize the grid */
    status = ias_geom_grid_initialize(IAS_MAX_NBANDS, &gdt);
    assert_true(status == SUCCESS);

    /* Initialize the satellite attributes library */
    ias_sat_attr_initialize(IAS_L8);

    /* write some grid data */
    strcpy(gdt.gridname, "gridfile");
    gdt.nbands = 11;
    gdt.frame_type = IAS_PROJBOX;
    
    strcpy(gdt.units, "DDR");
    gdt.code = 1;
    strcpy(gdt.datum, "WGS84");
    gdt.spheroid = 2;
    gdt.zone =3;
    for (i = 0; i < IAS_PROJ_PARAM_SIZE; i++)
        gdt.projprms[i] = i + 1.1;
    gdt.corners.upleft.x = 1.2;
    gdt.corners.upleft.y = 2.2;
    gdt.corners.upright.x = 3.2;
    gdt.corners.upright.y = 3.2;
    gdt.corners.loleft.x = 1.2;
    gdt.corners.loleft.y = 2.2;
    gdt.corners.loright.x = 3.2;
    gdt.corners.loright.y = 3.2;
    gdt.center_pt.lat = 5.5;
    gdt.center_pt.lng = 6.8;
    gdt.proj_center_pt.x = 5.6;
    gdt.proj_center_pt.y = 6.9;
    gdt.rotate_angle = 6.9; 
    gdt.orient_angle = 6.9; 

    /* write some grid band data */
    for (i=0; i< gdt.nbands; i++)
    {
        gdt.gridbands[i].band_number = i;
        gdt.gridbands[i].gridtype = 1;
           
        gdt.gridbands[i].cell_lines = 2;
        gdt.gridbands[i].cell_samps = 2;
        gdt.gridbands[i].pixsize_line = 0.5;
        gdt.gridbands[i].pixsize_samp = 0.5;
        gdt.gridbands[i].output_image_lines = 2;
        gdt.gridbands[i].output_image_samps = 3;
        gdt.gridbands[i].num_in_lines = 2;
        gdt.gridbands[i].num_in_samps = 4;
        gdt.gridbands[i].num_out_lines = 2;
        gdt.gridbands[i].num_out_samps = 3;
        gdt.gridbands[i].forward_present = 1;
        gdt.gridbands[i].reverse_present = 1;
        gdt.gridbands[i].grid_rows = 2;
        gdt.gridbands[i].grid_cols = 3;
        gdt.gridbands[i].degree = 2;
    }

    bandno = 1;           
    nband = 0;
    status = ias_geom_grid_malloc(&bandno,&nband,&gdt);
    assert_true(status == SUCCESS);

    /* write some grid band pointer data */
    for (i=0; i< gdt.nbands; i++)
    {
        for (k = 0; k < gdt.gridbands[i].num_in_lines; k++) 
        {
            gdt.gridbands[i].in_lines[k] = 8 ;      
        }
        for (k = 0; k < gdt.gridbands[i].num_in_samps; k++) 
        {
            gdt.gridbands[i].in_samps[k] = 9;        
        }
        for (k = 0; k < gdt.gridbands[i].num_out_lines; k++) 
        {
            gdt.gridbands[i].out_lines[k] = 6.1;   
        }
        for (k = 0; k < gdt.gridbands[i].num_out_samps; k++) 
        {
            gdt.gridbands[i].out_samps[k] = 7.2;   
        }
        for (k = 0; k < (gdt.gridbands[i].degree+1)
                *(gdt.gridbands[i].degree+1); k++) 
        {
            gdt.gridbands[i].poly_lines[k] = 3.3;   
            gdt.gridbands[i].poly_samps[k] = 4.4;   
        }
        for (k = 0; k < gdt.gridbands[i].grid_rows * 
             gdt.gridbands[i].grid_cols; k++) 
        {
            for (m=0; m < COEFS_SIZE; m++)
            {    
                gdt.gridbands[i].forward_coeffs[k].a[m] = 5.5;
                gdt.gridbands[i].forward_coeffs[k].b[m] = 6.6;
                gdt.gridbands[i].reverse_coeffs[k].a[m] = 7.7;
                gdt.gridbands[i].reverse_coeffs[k].b[m] = 8.8;
            }
        }
    } 

    /* Write the simulated data to the file. */
    status = ias_write_geom_grid("TestGeomGridIOData.h5",&gdt);
    assert_true(status == SUCCESS);

    /* free band pointer memory */        
    ias_geom_grid_free(&bandno, &nband, &gdt); 

    /*    Read the simulated data from the file. */
    bandno = 1;
    nband = 0;
    status = ias_read_geom_grid("TestGeomGridIOData.h5",&bandno,&nband,
                                 &gdt_read);
    assert_true(status == SUCCESS); 

    /* Assert the file type is a geometric grid file */
    status = ias_grid_is_geom_grid_file("TestGeomGridIOData.h5");
    assert_equal(status, TRUE);

    /* Assert the read out grid header data */
    assert_string_equal(gdt_read.gridname, "gridfile");
    assert_equal(gdt_read.nbands, 11);
    assert_equal(gdt_read.frame_type, IAS_PROJBOX);

    
    assert_string_equal(gdt_read.units, "DDR");
    assert_equal(gdt_read.code, 1);
    assert_string_equal(gdt_read.datum, "WGS84");
    assert_equal(gdt_read.spheroid, 2);
    assert_equal(gdt_read.zone, 3);
    for (i = 0; i < IAS_PROJ_PARAM_SIZE; i++)
        assert_double_equal(gdt_read.projprms[i], i + 1.1);
    assert_double_equal(gdt_read.corners.upleft.x, 1.2);
    assert_double_equal(gdt_read.corners.upleft.y, 2.2);
    assert_double_equal(gdt_read.corners.upright.x, 3.2);
    assert_double_equal(gdt_read.corners.upright.y, 3.2);
    assert_double_equal(gdt_read.corners.loleft.x, 1.2);
    assert_double_equal(gdt_read.corners.loleft.y, 2.2);
    assert_double_equal(gdt_read.corners.loright.x, 3.2);
    assert_double_equal(gdt_read.corners.loright.y, 3.2);
    assert_double_equal(gdt_read.center_pt.lat, 5.5);
    assert_double_equal(gdt_read.center_pt.lng, 6.8);
    assert_double_equal(gdt_read.proj_center_pt.x, 5.6);
    assert_double_equal(gdt_read.proj_center_pt.y, 6.9);
    assert_double_equal(gdt_read.rotate_angle, 6.9); 
    assert_double_equal(gdt_read.orient_angle, 6.9); 

    /* Assert the read out grid band header data */
    for (i=0; i< gdt_read.nbands; i++)
    {
        assert_equal(gdt_read.gridbands[i].band_number, i);
        assert_equal(gdt_read.gridbands[i].gridtype, 1);
        
        assert_equal(gdt_read.gridbands[i].cell_lines, 2);
        assert_equal(gdt_read.gridbands[i].cell_samps, 2);
        assert_double_equal(gdt_read.gridbands[i].pixsize_line, 0.5);
        assert_double_equal(gdt_read.gridbands[i].pixsize_samp, 0.5);
        assert_equal(gdt_read.gridbands[i].output_image_lines, 2);
        assert_equal(gdt_read.gridbands[i].output_image_samps, 3);
        assert_equal(gdt_read.gridbands[i].num_in_lines, 2);
        assert_equal(gdt_read.gridbands[i].num_in_samps, 4);
        assert_equal(gdt_read.gridbands[i].num_out_lines, 2);
        assert_equal(gdt_read.gridbands[i].num_out_samps, 3);
        assert_equal(gdt_read.gridbands[i].forward_present, 1);
        assert_equal(gdt_read.gridbands[i].reverse_present, 1);
        assert_equal(gdt_read.gridbands[i].grid_rows, 2);
        assert_equal(gdt_read.gridbands[i].grid_cols, 3);
        assert_equal(gdt_read.gridbands[i].degree, 2);
    }

    /* Assert the read out grid band pointer data */
    for (i=0; i< gdt_read.nbands; i++)
    {
        for (k = 0; k < gdt_read.gridbands[i].num_in_lines; k++) 
        {
            assert_equal(gdt_read.gridbands[i].in_lines[k], 8) ;      
        } 
        for (k = 0; k < gdt_read.gridbands[i].num_in_samps; k++) 
        {
            assert_equal(gdt_read.gridbands[i].in_samps[k], 9);    
        }
        for (k = 0; k < gdt_read.gridbands[i].num_out_lines; k++) 
        {
            assert_double_equal(gdt_read.gridbands[i].out_lines[k],
                                6.1);   
        }
        for (k = 0; k < gdt_read.gridbands[i].num_out_samps; k++) 
        {
            assert_double_equal(gdt_read.gridbands[i].out_samps[k],     
                                    7.2); 
        }
        for (k = 0; k < (gdt_read.gridbands[i].degree+1)
                *(gdt_read.gridbands[i].degree+1); k++) 
        {
            assert_double_equal(
                gdt_read.gridbands[i].poly_lines[k], 3.3);   
            assert_double_equal(
                gdt_read.gridbands[i].poly_samps[k], 4.4);   
        }
        for (k = 0; k < gdt_read.gridbands[i].grid_rows * 
             gdt_read.gridbands[i].grid_cols; k++) 
        {
            for (m=0; m < COEFS_SIZE; m++)
            {    
                 assert_double_equal(
                 gdt_read.gridbands[i].forward_coeffs[k].a[m], 5.5);
                 assert_double_equal(
                 gdt_read.gridbands[i].forward_coeffs[k].b[m], 6.6);
                 assert_double_equal(
                 gdt_read.gridbands[i].reverse_coeffs[k].a[m], 7.7);
                 assert_double_equal(
                 gdt_read.gridbands[i].reverse_coeffs[k].b[m], 8.8);
            }
        } 
    } 

    /* free band pointer memory */        
    ias_geom_grid_free(&bandno, &nband, &gdt_read); 
}

int main(int argc, char **argv)
{
    TestSuite *suite = create_test_suite();
    add_test(suite, test_ias_geom_grid_write_read);
        
    return run_test_suite(suite, create_text_reporter());
}



