/************************************************************************

NAME:     READWRITE_TESTS

PURPOSE:  Use CGreen to perform unit testing on Grid IO Library to 
          write and read back the Grid and Grid band data.

**************************************************************************
                        Property of the U.S. Government
                            USGS EROS Data Center
**************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <sys/stat.h>

#include "ias_grid_io.h"
#include "ias_miscellaneous.h" /* Header for 2d array allocation */
#include "cgreen/cgreen.h"
#include "ias_logging.h"

#define TOLERANCE 0.000001     /* Tolerance used for comparing float data */

/* Check if the float value is approximately at a specified value */
#define IS_DOUBLE_EQUAL(value, target) \
        ((value > target - TOLERANCE) && (value < target + TOLERANCE))


Ensure test_grid_allocation()
{
    int status;
    IAS_GRID_TYPE grid;
    int band_index;
    int sca_index;
    int band_numbers[IAS_MAX_NBANDS];
    int band_count;
    int i;

    ias_sat_attr_initialize(IAS_L8);

    /* test full grid initialization */
    status = ias_grid_initialize(&grid, 0, 0);
    assert_true(status == SUCCESS);
    assert_true(grid.gridbands == NULL);
    assert_true(grid.nbands == 0);
    assert_true(grid.maximum_nscas == 0);

    status = ias_grid_initialize(&grid, IAS_MAX_NBANDS, IAS_MAX_NSCAS);
    assert_true(status == SUCCESS);
    assert_true(grid.gridbands != NULL);
    assert_true(grid.nbands == IAS_MAX_NBANDS);
    assert_true(grid.maximum_nscas == IAS_MAX_NSCAS);
    assert_true(grid.gridbands[0][0].in_lines == NULL);
    for (band_index = 0; band_index < grid.nbands; band_index++)
    {
        int band_number = ias_sat_attr_convert_band_index_to_number(band_index);
        assert_true(band_number != ERROR);

        int scas = ias_sat_attr_get_scas_per_band(band_number);
        assert_true(scas = grid.scas_per_band[band_index]);
        assert_true((grid.bands_present[band_index] == 0) 
                    && (grid.bands_available[band_index]== 0));
    }

    /* test full grid allocation */
    status = ias_grid_malloc(&grid, NULL, 0);
    assert_true(status == SUCCESS);
    assert_true(grid.gridbands != NULL);
    for (band_index = 0; band_index < grid.nbands; band_index++)
    {
        assert_true((grid.bands_present[band_index] == 1) 
                    && (grid.bands_available[band_index]== 1));
    }

    for (band_index = 0; band_index < grid.nbands; band_index++)
    {
        for (sca_index = 0; sca_index < grid.scas_per_band[band_index];
             sca_index++)
        {
            IAS_GRID_BAND_TYPE *gridband 
                    = &grid.gridbands[band_index][sca_index];
            assert_true((gridband->in_lines != NULL)
                        && (gridband->in_samps != NULL)
                        && (gridband->out_lines != NULL)
                        && (gridband->out_samps != NULL)
                        && (gridband->delta_line_oe != NULL)
                        && (gridband->delta_samp_oe != NULL)
                        && (gridband->sattoproj != NULL)
                        && (gridband->projtosat != NULL)
                        && (gridband->poly_lines != NULL)
                        && (gridband->poly_samps != NULL)
                        && (gridband->line_sensitivity != NULL)
                        && (gridband->samp_sensitivity != NULL));
        }
    }
    /* make sure the TIRS bands do not have space allocated for the SCAs it
       doesn't have */
    status = ias_sat_attr_get_sensor_band_numbers(IAS_TIRS, IAS_NORMAL_BAND,
                0, band_numbers, IAS_MAX_NBANDS, &band_count);
    assert_true(status == SUCCESS);
    for (i = 0; i < band_count; i++)
    {
        int band_number = band_numbers[i];
        band_index = ias_sat_attr_convert_band_number_to_index(band_number);

        for (sca_index = grid.scas_per_band[band_index]; 
             sca_index < grid.maximum_nscas; sca_index++)
        {
            assert_true(grid.gridbands[band_index][sca_index].in_lines
                    == NULL);
        }
    }

    /* test freeing the grid */
    status = ias_grid_free(&grid);
    assert_true(status == SUCCESS);
    assert_true(grid.gridbands == NULL);
    for (band_index = 0; band_index < IAS_MAX_NBANDS; band_index++)
        assert_true(!grid.bands_present[band_index]);

    /* test allocation of a subset of bands */
    status = ias_grid_initialize(&grid, IAS_MAX_NBANDS, IAS_MAX_NSCAS);
    assert_true(status == SUCCESS);
    band_numbers[0] = 2;
    band_numbers[1] = 4;
    status = ias_grid_malloc(&grid, band_numbers, 2);
    assert_true(status == SUCCESS);
    assert_true(grid.gridbands[0][0].in_lines == NULL);
    assert_true(grid.gridbands[1][0].in_lines != NULL);
    assert_true(grid.gridbands[2][0].in_lines == NULL);
    assert_true(grid.gridbands[3][0].in_lines != NULL);
    assert_true(grid.gridbands[4][0].in_lines == NULL);
    assert_true(grid.gridbands[5][0].in_lines == NULL);
    assert_true(grid.bands_present[0] == 0);
    assert_true(grid.bands_present[1] == 1);
    assert_true(grid.bands_present[2] == 0);
    assert_true(grid.bands_present[3] == 1);
    assert_true(grid.bands_available[0] == 0);
    assert_true(grid.bands_available[1] == 1);
    assert_true(grid.bands_available[2] == 0);
    assert_true(grid.bands_available[3] == 1);

    /* test freeing a single band */
    status = ias_grid_free_band(&grid, 2);
    assert_true(status == SUCCESS);
    assert_true(grid.gridbands[0][0].in_lines == NULL);
    assert_true(grid.gridbands[1][0].in_lines == NULL);
    assert_true(grid.gridbands[2][0].in_lines == NULL);
    assert_true(grid.gridbands[3][0].in_lines != NULL);
    assert_true(grid.gridbands[4][0].in_lines == NULL);
    assert_true(grid.gridbands[5][0].in_lines == NULL);
    assert_true(grid.bands_present[0] == 0);
    assert_true(grid.bands_present[1] == 0);
    assert_true(grid.bands_present[2] == 0);
    assert_true(grid.bands_present[3] == 1);
    assert_true(grid.bands_available[0] == 0);
    assert_true(grid.bands_available[1] == 1);
    assert_true(grid.bands_available[2] == 0);
    assert_true(grid.bands_available[3] == 1);

    /* free the grid */
    status = ias_grid_free(&grid);
    assert_true(status == SUCCESS);
}

/*************************************************************************
           READ AND WRITE TEST
**************************************************************************/
Ensure test_ias_grid_write_read()
{
    int i, j, k, m;
    int status;
    int nband;
    int band_numbers[IAS_MAX_NBANDS];
    /* Simulated grid data type  */
    IAS_GRID_TYPE gdt;
    IAS_GRID_TYPE gdt_read;
    int succeeded; /* 0 if the tests succeeded; 1 if they failed */

    ias_sat_attr_initialize(IAS_L8);

    /* Initialize the grid */
    status = ias_grid_initialize(&gdt, IAS_MAX_NBANDS, IAS_MAX_NSCAS);
    assert_true(status == SUCCESS);

    /* write some grid data */
    strcpy(gdt.gridname, "gridfile");
    gdt.maximum_nscas = 4;
    gdt.nbands = 11;
    for (i = 0; i < gdt.nbands; i++)
    {
        gdt.bands_present[i] = 1;
        gdt.bands_available[i] = 1;
        if (i < 9)
            gdt.scas_per_band[i] = 4;
        else
            gdt.scas_per_band[i] = 2;
    }
    /* let's make band index 5 not be present */
    gdt.bands_present[5] = 0;
    gdt.bands_available[5] = 0;

    gdt.dtype = 1;
    gdt.correction_type = IAS_PRECISION;
    gdt.satnum = 1;
    gdt.frame_type = IAS_PROJBOX;
    gdt.acquisition_type = IAS_EARTH;
    
    strcpy(gdt.units, "DDR");
    gdt.code = 1;
    strcpy(gdt.datum, "WGS84");
    gdt.spheroid = 2;
    gdt.zone =3;
    for (i = 0; i < IAS_PROJ_PARAM_SIZE; i++)
        gdt.projprms[i] = i + 1.1;
    gdt.corners.upleft.x = 1.2;
    gdt.corners.upleft.y = 2.2;
    gdt.corners.upright.x = 3.2;
    gdt.corners.upright.y = 3.2;
    gdt.corners.loleft.x = 1.2;
    gdt.corners.loleft.y = 2.2;
    gdt.corners.loright.x = 3.2;
    gdt.corners.loright.y = 3.2;
    gdt.center_pt.lat = 5.5;
    gdt.center_pt.lng = 6.8;
    gdt.proj_center_pt.x = 5.5;
    gdt.proj_center_pt.y = 6.8;
    gdt.rotate_angle = 3.5;
    gdt.orient_angle = 6.9; 

    /* write some grid band data */
    for (i = 0; i < gdt.nbands; i++)
    {
        for (j = 0; j< gdt.scas_per_band[i]; j++)
        {
            gdt.gridbands[i][j].band_index = i;
            gdt.gridbands[i][j].sca_number = j;

            gdt.gridbands[i][j].grid_cell_size.line = 2;
            gdt.gridbands[i][j].grid_cell_size.samp = 2;
            gdt.gridbands[i][j].inv_cell_lines = 0.5;
            gdt.gridbands[i][j].inv_cell_samps = 0.5;
            gdt.gridbands[i][j].pixsize = 2.0;
            gdt.gridbands[i][j].zspacing = 2.0;
            gdt.gridbands[i][j].lines = 2;
            gdt.gridbands[i][j].samps = 4;
            gdt.gridbands[i][j].ngrid_lines = 2;
            gdt.gridbands[i][j].ngrid_samps = 3;
            gdt.gridbands[i][j].nzplanes = 2;
            gdt.gridbands[i][j].zeroplane = 1;
            gdt.gridbands[i][j].degree = 2;
        }
    }

    /* write some grid active area corners data */
    for (i = 0; i < gdt.nbands; i++)
    {
        gdt.active_corners[i].upleft.x = i + 0.11;
        gdt.active_corners[i].upleft.y = i + 0.12;
        gdt.active_corners[i].upright.x = i + 0.21;
        gdt.active_corners[i].upright.y = i + 0.22;
        gdt.active_corners[i].loleft.x = i + 0.31;
        gdt.active_corners[i].loleft.y = i + 0.32;
        gdt.active_corners[i].loright.x = i + 0.41;
        gdt.active_corners[i].loright.y = i + 0.42;
    }

    /* build bands number list to malloc */
    nband = 0;
    for (i = 0; i < gdt.nbands; i++)
    {
        if (gdt.bands_present[i])
        {
            band_numbers[nband] = i + 1;
            nband++;
        }
    }
    status = ias_grid_malloc(&gdt,band_numbers,nband);
    assert_true(status == SUCCESS);

    /* write some grid band pointer data */
    for (i = 0; i < gdt.nbands; i++)
    {
        if (!gdt.bands_present[i])
            continue;

        for (j = 0; j< gdt.scas_per_band[i]; j++)
        {
            for (k = 0; k < gdt.gridbands[i][j].ngrid_lines; k++) 
            {
                gdt.gridbands[i][j].in_lines[k] = 8 ;      
            }
            for (k = 0; k < gdt.gridbands[i][j].ngrid_samps; k++) 
            {
                gdt.gridbands[i][j].in_samps[k] = 9;            
            }
            for (k = 0; k < gdt.gridbands[i][j].ngrid_lines 
                 *gdt.gridbands[i][j].ngrid_samps 
                 *gdt.gridbands[i][j].nzplanes ; k++) 
            {
                gdt.gridbands[i][j].out_lines[k] = 6.1;   
                gdt.gridbands[i][j].out_samps[k] = 7.2;   
            }
            for (k = 0; k < 2*gdt.gridbands[i][j].ngrid_lines
                 * gdt.gridbands[i][j].ngrid_samps; k++) 
            {
                gdt.gridbands[i][j].delta_line_oe[k] = 1.1;   
                gdt.gridbands[i][j].delta_samp_oe[k] = 2.2;   
            }
            for (k = 0; k < (gdt.gridbands[i][j].degree+1)
                    *(gdt.gridbands[i][j].degree+1) 
                    *gdt.gridbands[i][j].nzplanes ; k++) 
            {
                gdt.gridbands[i][j].poly_lines[k] = 3.3;   
                gdt.gridbands[i][j].poly_samps[k] = 4.4;   
            }
            for (k = 0; k < (gdt.gridbands[i][j].ngrid_lines - 1)
                 * (gdt.gridbands[i][j].ngrid_samps - 1)
                 * gdt.gridbands[i][j].nzplanes ; k++) 
            {
                for (m=0; m < COEFS_SIZE; m++)
                {    
                    gdt.gridbands[i][j].sattoproj[k].a[m] = 5.5;
                    gdt.gridbands[i][j].sattoproj[k].b[m] = 6.6;
                    gdt.gridbands[i][j].projtosat[k].a[m] = 7.7;
                    gdt.gridbands[i][j].projtosat[k].b[m] = 8.8;
                }
            }

            for (k = 0; k < gdt.gridbands[i][j].ngrid_lines
                 * gdt.gridbands[i][j].ngrid_samps
                 * gdt.gridbands[i][j].nzplanes; k++) 
            {
                gdt.gridbands[i][j].line_sensitivity[k].x = 3.1 + k;
                gdt.gridbands[i][j].line_sensitivity[k].y = 3.2 + k;
                gdt.gridbands[i][j].line_sensitivity[k].z = 3.3 + k;
                gdt.gridbands[i][j].samp_sensitivity[k].x = 3.7 + k;
                gdt.gridbands[i][j].samp_sensitivity[k].y = 3.8 + k;
                gdt.gridbands[i][j].samp_sensitivity[k].z = 3.9 + k;
            }
        }
    } 

    /* Write the simulated data to the file. */
    status = ias_grid_write("TestGridIOData.h5",&gdt);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
        return;

    /*    Read the simulated data from the file. */
    nband = 0;
    status = ias_grid_read("TestGridIOData.h5", NULL, 0, &gdt_read);
    assert_true(status == SUCCESS); 
    if (status != SUCCESS)
        return;

    /* Assert the file type is a grid file */
    status = ias_grid_is_grid_file("TestGridIOData.h5");
    assert_equal(status, TRUE);

    /* Assert the read out grid header data */
    assert_string_equal(gdt_read.gridname, gdt.gridname);
    assert_equal(gdt_read.maximum_nscas, gdt.maximum_nscas);
    assert_equal(gdt_read.nbands, gdt.nbands);
    for (i = 0; i < gdt_read.nbands; i++)
    {
        assert_equal(gdt_read.bands_present[i], gdt.bands_present[i]);
        assert_equal(gdt_read.bands_available[i], gdt.bands_available[i]);
        assert_equal(gdt_read.scas_per_band[i], gdt.scas_per_band[i]);
    }
    assert_equal(gdt_read.dtype, gdt.dtype);
    assert_equal(gdt_read.correction_type, gdt.correction_type);
    assert_equal(gdt_read.satnum, gdt.satnum);
    assert_equal(gdt_read.frame_type, gdt.frame_type);
    assert_equal(gdt_read.acquisition_type, gdt.acquisition_type);
    
    assert_string_equal(gdt_read.units, gdt.units);
    assert_equal(gdt_read.code, gdt.code);
    assert_string_equal(gdt_read.datum, gdt.datum);
    assert_equal(gdt_read.spheroid, gdt.spheroid);
    assert_equal(gdt_read.zone, gdt.zone);
    for (i = 0; i < IAS_PROJ_PARAM_SIZE; i++)
        assert_double_equal(gdt_read.projprms[i], gdt.projprms[i]);
    assert_double_equal(gdt_read.corners.upleft.x, gdt.corners.upleft.x);
    assert_double_equal(gdt_read.corners.upleft.y, gdt.corners.upleft.y);
    assert_double_equal(gdt_read.corners.upright.x, gdt.corners.upright.x);
    assert_double_equal(gdt_read.corners.upright.y, gdt.corners.upright.y);
    assert_double_equal(gdt_read.corners.loleft.x, gdt.corners.loleft.x);
    assert_double_equal(gdt_read.corners.loleft.y, gdt.corners.loleft.y);
    assert_double_equal(gdt_read.corners.loright.x, gdt.corners.loright.x);
    assert_double_equal(gdt_read.corners.loright.y, gdt.corners.loright.y);
    assert_double_equal(gdt_read.center_pt.lat, gdt.center_pt.lat);
    assert_double_equal(gdt_read.center_pt.lng, gdt.center_pt.lng);
    assert_double_equal(gdt_read.proj_center_pt.x, gdt.proj_center_pt.x);
    assert_double_equal(gdt_read.proj_center_pt.y, gdt.proj_center_pt.y);
    assert_double_equal(gdt_read.rotate_angle, gdt.rotate_angle);
    assert_double_equal(gdt_read.orient_angle, gdt.orient_angle); 

    /* Assert the read out grid band header data */
    for (i = 0; i < gdt_read.nbands; i++)
    {
        for (j = 0; j < gdt_read.scas_per_band[i]; j++)
        {
            assert_equal(gdt_read.gridbands[i][j].band_index, i);
            assert_equal(gdt_read.gridbands[i][j].sca_number, j);
            
            assert_equal(gdt_read.gridbands[i][j].grid_cell_size.line, 
                         gdt.gridbands[i][j].grid_cell_size.line);
            assert_equal(gdt_read.gridbands[i][j].grid_cell_size.samp, 
                         gdt.gridbands[i][j].grid_cell_size.samp);
            assert_double_equal(gdt_read.gridbands[i][j].inv_cell_lines, 
                         gdt.gridbands[i][j].inv_cell_lines);
            assert_double_equal(gdt_read.gridbands[i][j].inv_cell_samps, 
                         gdt.gridbands[i][j].inv_cell_samps);
            assert_double_equal(gdt_read.gridbands[i][j].pixsize, 
                         gdt.gridbands[i][j].pixsize);
            assert_double_equal(gdt_read.gridbands[i][j].zspacing, 
                         gdt.gridbands[i][j].zspacing);
            assert_equal(gdt_read.gridbands[i][j].lines, 
                         gdt.gridbands[i][j].lines);
            assert_equal(gdt_read.gridbands[i][j].samps, 
                         gdt.gridbands[i][j].samps);
            assert_equal(gdt_read.gridbands[i][j].ngrid_lines, 
                         gdt.gridbands[i][j].ngrid_lines);
            assert_equal(gdt_read.gridbands[i][j].ngrid_samps, 
                         gdt.gridbands[i][j].ngrid_samps);
            assert_equal(gdt_read.gridbands[i][j].nzplanes, 
                         gdt.gridbands[i][j].nzplanes);
            assert_equal(gdt_read.gridbands[i][j].zeroplane, 
                         gdt.gridbands[i][j].zeroplane);
            assert_equal(gdt_read.gridbands[i][j].degree, 
                         gdt.gridbands[i][j].degree);
        }
    }

    /* Assert the read out grid band pointer data */
    for (i = 0; i < gdt_read.nbands; i++)
    {
        if (!gdt_read.bands_present[i])
            continue;

        for (j = 0; j < gdt_read.scas_per_band[i]; j++)
        {
            for (k = 0; k < gdt_read.gridbands[i][j].ngrid_lines; k++) 
            {
                assert_equal(gdt_read.gridbands[i][j].in_lines[k], 
                             gdt.gridbands[i][j].in_lines[k]);
            } 
            for (k = 0; k < gdt_read.gridbands[i][j].ngrid_samps; k++) 
            {
                assert_equal(gdt_read.gridbands[i][j].in_samps[k],
                             gdt.gridbands[i][j].in_samps[k]);
            }
            for (k = 0; k < gdt_read.gridbands[i][j].ngrid_lines
                 *gdt_read.gridbands[i][j].ngrid_samps
                 *gdt_read.gridbands[i][j].nzplanes ; k++) 
            {
                assert_double_equal(gdt_read.gridbands[i][j].out_lines[k],
                                    gdt.gridbands[i][j].out_lines[k]);
                assert_double_equal(gdt_read.gridbands[i][j].out_samps[k],
                                    gdt.gridbands[i][j].out_samps[k]);
            }
            for (k = 0; k < 2 * gdt_read.gridbands[i][j].ngrid_lines
                 * gdt_read.gridbands[i][j].ngrid_samps; k++) 
            {
                assert_double_equal(gdt_read.gridbands[i][j].delta_line_oe[k],
                                    gdt.gridbands[i][j].delta_line_oe[k]);
                assert_double_equal(gdt_read.gridbands[i][j].delta_samp_oe[k],
                                    gdt.gridbands[i][j].delta_samp_oe[k]);
                    
            }
            for (k = 0; k < (gdt_read.gridbands[i][j].degree+1)
                    *(gdt_read.gridbands[i][j].degree+1)  
                    *gdt_read.gridbands[i][j].nzplanes ; k++) 
            {
                assert_double_equal(gdt_read.gridbands[i][j].poly_lines[k], 
                                    gdt.gridbands[i][j].poly_lines[k]);
                assert_double_equal(gdt_read.gridbands[i][j].poly_samps[k], 
                                    gdt.gridbands[i][j].poly_samps[k]);
                    
            }
            for (k = 0; k < (gdt_read.gridbands[i][j].ngrid_lines - 1)
                 * (gdt_read.gridbands[i][j].ngrid_samps - 1)
                 * gdt_read.gridbands[i][j].nzplanes ; k++) 
            {
                for (m = 0; m < COEFS_SIZE; m++)
                {    
                     assert_double_equal(
                         gdt_read.gridbands[i][j].sattoproj[k].a[m],
                         gdt.gridbands[i][j].sattoproj[k].a[m]);
                     assert_double_equal(
                         gdt_read.gridbands[i][j].sattoproj[k].b[m],
                         gdt.gridbands[i][j].sattoproj[k].b[m]);
                     assert_double_equal(
                         gdt_read.gridbands[i][j].projtosat[k].a[m],
                         gdt.gridbands[i][j].projtosat[k].a[m]);
                     assert_double_equal(
                         gdt_read.gridbands[i][j].projtosat[k].b[m],
                         gdt.gridbands[i][j].projtosat[k].b[m]);
               }
            } 

            /* The number of asserts handled by cgreen is being exceeded.
               This style will reduce the number of asserts.  Expand its use
               as needed. */
            succeeded = 1;
            for (k = 0; k < (gdt_read.gridbands[i][j].ngrid_lines)
                 * (gdt_read.gridbands[i][j].ngrid_samps)
                 * gdt_read.gridbands[i][j].nzplanes ; k++) 
            {
                 if (!IS_DOUBLE_EQUAL(
                     gdt_read.gridbands[i][j].line_sensitivity[k].x,
                     gdt.gridbands[i][j].line_sensitivity[k].x))
                 {
                     assert_double_equal(
                         gdt_read.gridbands[i][j].line_sensitivity[k].x,
                         gdt.gridbands[i][j].line_sensitivity[k].x);
                     succeeded = 0;
                 }
                 if (!IS_DOUBLE_EQUAL(
                     gdt_read.gridbands[i][j].line_sensitivity[k].y,
                     gdt.gridbands[i][j].line_sensitivity[k].y))
                 {
                     assert_double_equal(
                         gdt_read.gridbands[i][j].line_sensitivity[k].y,
                         gdt.gridbands[i][j].line_sensitivity[k].y);
                     succeeded = 0;
                 }
                 if (!IS_DOUBLE_EQUAL(
                     gdt_read.gridbands[i][j].line_sensitivity[k].z,
                     gdt.gridbands[i][j].line_sensitivity[k].z))
                 {
                     assert_double_equal(
                         gdt_read.gridbands[i][j].line_sensitivity[k].z,
                         gdt.gridbands[i][j].line_sensitivity[k].z);
                     succeeded = 0;
                 }
                 if (!IS_DOUBLE_EQUAL(
                     gdt_read.gridbands[i][j].samp_sensitivity[k].x,
                     gdt.gridbands[i][j].samp_sensitivity[k].x))
                 {
                     assert_double_equal(
                         gdt_read.gridbands[i][j].samp_sensitivity[k].x,
                         gdt.gridbands[i][j].samp_sensitivity[k].x);
                     succeeded = 0;
                 }
                 if (!IS_DOUBLE_EQUAL(
                     gdt_read.gridbands[i][j].samp_sensitivity[k].y,
                     gdt.gridbands[i][j].samp_sensitivity[k].y))
                 {
                     assert_double_equal(
                         gdt_read.gridbands[i][j].samp_sensitivity[k].y,
                         gdt.gridbands[i][j].samp_sensitivity[k].y);
                     succeeded = 0;
                 }
                 if (!IS_DOUBLE_EQUAL(
                     gdt_read.gridbands[i][j].samp_sensitivity[k].z,
                     gdt.gridbands[i][j].samp_sensitivity[k].z))
                 {
                     assert_double_equal(
                         gdt_read.gridbands[i][j].samp_sensitivity[k].z,
                         gdt.gridbands[i][j].samp_sensitivity[k].z);
                     succeeded = 0;
                 }
            } 
            assert_equal(succeeded, 1);
        }
    } 

    /* Assert the read out grid active area corners data */
    for (i = 0; i < gdt_read.nbands; i++)
    {
        if (!gdt_read.bands_present[i])
            continue;

        assert_double_equal(gdt_read.active_corners[i].upleft.x,
            gdt.active_corners[i].upleft.x);

        assert_double_equal(gdt_read.active_corners[i].upleft.y,
            gdt.active_corners[i].upleft.y);

        assert_double_equal(gdt_read.active_corners[i].upright.x,
            gdt.active_corners[i].upright.x);

        assert_double_equal(gdt_read.active_corners[i].upright.y,
            gdt.active_corners[i].upright.y);

        assert_double_equal(gdt_read.active_corners[i].loleft.x,
            gdt.active_corners[i].loleft.x);

        assert_double_equal(gdt_read.active_corners[i].loleft.y,
            gdt.active_corners[i].loleft.y);

        assert_double_equal(gdt_read.active_corners[i].loright.x,
            gdt.active_corners[i].loright.x);

        assert_double_equal(gdt_read.active_corners[i].loright.y,
            gdt.active_corners[i].loright.y);
    }

    /* free band pointer memory */        
    ias_grid_free(&gdt_read); 

    ias_grid_free(&gdt); 
}

int main(int argc, char **argv)
{
    TestSuite *suite = create_test_suite();
    add_test(suite, test_grid_allocation);
    add_test(suite, test_ias_grid_write_read);

    if (argc > 1)
        return run_single_test(suite, argv[1], create_text_reporter());
    return run_test_suite(suite, create_text_reporter());
}



