#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "ias_const.h"
#include "ias_satellite_attributes.h"
#include "../ias_model_io.h"
#include "ias_logging.h"
#include "cgreen/cgreen.h"

static int fill_model(IAS_LOS_MODEL *);
Ensure model_io_test(void);

/* Usage:
   ./model_io_test
   ./model_io_test [name of test to run]
*/
int main(int argc, char *argv[])
{
    TestSuite *suite = create_test_suite();
    add_test(suite, model_io_test);
    if (argc == 2)
        return run_single_test(suite, argv[1], create_text_reporter());
    return run_test_suite(suite, create_text_reporter());
}

Ensure model_io_test(void)
{
    IAS_LOS_MODEL *the_model = NULL;
    IAS_LOS_MODEL *new_model = NULL;
    IAS_SENSOR_MODEL *the_sensor;
    IAS_SENSOR_MODEL *new_sensor;
    int status;
    size_t nbytes1;
    size_t nbytes2;
    char *filename[3] = {"regression_test.h5", "regression_test_version_2.h5",
                         "regression_test_version_1.h5"};
    int remove_file[3] = {TRUE, FALSE, FALSE};
    char bad_filetype_filename[] = "regression_test_bad_filetype.h5";
    int sensor_index;
    int band_index;
    int sca_index;
    int index;

    /* disable all output since the tests are hitting error conditions */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* test some error conditions first */
    the_model = ias_model_read("bogus_file_name");
    if (the_model != NULL)
    {
        assert_true_with_message(0, "Error validating file name");
        ias_los_model_free(the_model);
        return;
    }

    ias_sat_attr_initialize(IAS_L8);

    the_model = ias_model_read(bad_filetype_filename);
    if (the_model != NULL)
    {
        assert_true_with_message(0, "Error validating file type");
        ias_los_model_free(the_model);
        return;
    }

    /* now start over with good test cases */
    ias_log_set_output_level(IAS_LOG_LEVEL_WARN);

    the_model = ias_los_model_allocate();
    if (the_model == NULL)
    {
        assert_true_with_message(0, "Error allocating the_model");
        return;
    }

    status = fill_model(the_model);
    if (status) /* 1 is returned if error filling model */
    {
        assert_true_with_message(0, "Error filling the_model structure");
        ias_los_model_free(the_model);
        return;
    }

    status = ias_model_write(filename[0], the_model);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error writing the_model");
        ias_los_model_free(the_model);
        return;
    }

    /* Read in the version 1-3 files to verify */
    for (index = 0; index < 3; index++)
    {
        /* Read the model */
        new_model = ias_model_read(filename[index]);
        if (new_model == NULL)
        {
            assert_true_with_message(0, "Error reading new_model");
            ias_los_model_free(the_model);
            ias_los_model_free(new_model);
            return;
        }

        the_sensor = &the_model->sensor;
        new_sensor = &new_model->sensor;

        /*------Compare the General LOS Model Information------*/
        assert_equal(the_model->satellite_id, new_model->satellite_id);
        assert_equal(the_model->satellite_number, new_model->satellite_number);
        assert_string_equal(the_model->sw_version, new_model->sw_version);
        assert_equal(the_model->wrs_path, new_model->wrs_path);
        assert_equal(the_model->wrs_row, new_model->wrs_row);
        assert_equal(the_model->acquisition_type, new_model->acquisition_type);
        assert_equal(the_model->correction_type, new_model->correction_type);
        assert_equal(the_sensor->band_count, new_sensor->band_count);

        /*------Compare the Earth Model------*/
        assert_double_equal(the_model->earth.ut1_utc_correction,
            new_model->earth.ut1_utc_correction);
        assert_double_equal(the_model->earth.pole_wander_x,
            new_model->earth.pole_wander_x);
        assert_double_equal(the_model->earth.pole_wander_y,
            new_model->earth.pole_wander_y);
        assert_double_equal(the_model->earth.semi_major_axis,
            new_model->earth.semi_major_axis);
        assert_double_equal(the_model->earth.semi_minor_axis,
            new_model->earth.semi_minor_axis);
        assert_double_equal(the_model->earth.eccentricity,
            new_model->earth.eccentricity);
        assert_double_equal(the_model->earth.earth_angular_velocity,
            new_model->earth.earth_angular_velocity);
        assert_double_equal(the_model->earth.speed_of_light,
            new_model->earth.speed_of_light);

        /*------Compare the Spacecraft Model Information------*/
        /* Compare the Attitude Model */
        assert_double_equal(the_model->spacecraft.attitude.utc_epoch_time[0],
            new_model->spacecraft.attitude.utc_epoch_time[0]);
        assert_double_equal(the_model->spacecraft.attitude.utc_epoch_time[1],
            new_model->spacecraft.attitude.utc_epoch_time[1]);
        assert_double_equal(the_model->spacecraft.attitude.utc_epoch_time[2],
            new_model->spacecraft.attitude.utc_epoch_time[2]);
        assert_double_equal(the_model->spacecraft.attitude.nominal_sample_time,
            new_model->spacecraft.attitude.nominal_sample_time);
        assert_equal(the_model->spacecraft.attitude.sample_count,
            new_model->spacecraft.attitude.sample_count);
        nbytes1 = the_model->spacecraft.attitude.sample_count 
            * sizeof(IAS_SC_ATTITUDE_RECORD);
        assert_false(memcmp(&the_model->spacecraft.attitude.sample_records[0],
            &new_model->spacecraft.attitude.sample_records[0], nbytes1));

        /* Compare the Ephemeris Model */
        assert_double_equal(the_model->spacecraft.ephemeris.utc_epoch_time[0],
            new_model->spacecraft.ephemeris.utc_epoch_time[0]);
        assert_double_equal(the_model->spacecraft.ephemeris.utc_epoch_time[1],
            new_model->spacecraft.ephemeris.utc_epoch_time[1]);
        assert_double_equal(the_model->spacecraft.ephemeris.utc_epoch_time[2],
            new_model->spacecraft.ephemeris.utc_epoch_time[2]);
        assert_double_equal(the_model->spacecraft.ephemeris.nominal_sample_time,
            new_model->spacecraft.ephemeris.nominal_sample_time);
        assert_equal(the_model->spacecraft.ephemeris.sample_count,
            new_model->spacecraft.ephemeris.sample_count);
        nbytes1 = the_model->spacecraft.ephemeris.sample_count 
            * sizeof(IAS_SC_EPHEMERIS_RECORD);
        assert_false(memcmp(&the_model->spacecraft.ephemeris.sample_records[0],
            &new_model->spacecraft.ephemeris.sample_records[0], nbytes1));

        /* Compare the Precision Model */
        assert_double_equal(
            the_model->spacecraft.precision.seconds_from_image_epoch,
            new_model->spacecraft.precision.seconds_from_image_epoch);
        assert_equal(the_model->spacecraft.precision.ephemeris_order,
            new_model->spacecraft.precision.ephemeris_order);
        assert_equal(the_model->spacecraft.precision.attitude_order, 
            new_model->spacecraft.precision.attitude_order);
        nbytes1 = sizeof(double) * IAS_PRECISION_MAX_POLY_COEFF;
        assert_false(memcmp(&the_model->spacecraft.precision.x_corr[0],
            &new_model->spacecraft.precision.x_corr[0], nbytes1));
        assert_false(memcmp(&the_model->spacecraft.precision.y_corr[0],
            &new_model->spacecraft.precision.y_corr[0], nbytes1));
        assert_false(memcmp(&the_model->spacecraft.precision.z_corr[0],
            &new_model->spacecraft.precision.z_corr[0], nbytes1));
        assert_false(memcmp(&the_model->spacecraft.precision.roll_corr[0],
            &new_model->spacecraft.precision.roll_corr[0], nbytes1));
        assert_false(memcmp(&the_model->spacecraft.precision.pitch_corr[0],
            &new_model->spacecraft.precision.pitch_corr[0], nbytes1));
        assert_false(memcmp(&the_model->spacecraft.precision.yaw_corr[0],
            &new_model->spacecraft.precision.yaw_corr[0], nbytes1));

        /*------Compare the Sensor Model Information------*/
        nbytes1 = TRANSFORMATION_MATRIX_DIM * TRANSFORMATION_MATRIX_DIM
            * sizeof(double);
        nbytes2 = TRANSFORMATION_MATRIX_DIM * sizeof(double);
        for (sensor_index = 0; sensor_index < IAS_MAX_SENSORS; sensor_index++)
        {
            IAS_VECTOR *orig_jitter_table
                = the_sensor->sensors[sensor_index].jitter_table;
            IAS_VECTOR *new_jitter_table
                = new_sensor->sensors[sensor_index].jitter_table;

            assert_equal(the_sensor->sensors[sensor_index].sensor_id,
                new_sensor->sensors[sensor_index].sensor_id);
            assert_false(memcmp(the_sensor->sensors[sensor_index].sensor2acs,
                new_sensor->sensors[sensor_index].sensor2acs, nbytes1));
            assert_false(memcmp(the_sensor->sensors[sensor_index].
                center_mass2sensor_offset,
                new_sensor->sensors[sensor_index].center_mass2sensor_offset,
                nbytes2));

            /* confirm the jitter table information */
            assert_true(new_jitter_table != NULL);

            assert_equal(new_sensor->sensors[sensor_index].jitter_table_count,
                the_sensor->sensors[sensor_index].jitter_table_count);
            assert_equal(
                new_sensor->sensors[sensor_index].jitter_entries_per_frame,
                the_sensor->sensors[sensor_index].jitter_entries_per_frame);

            assert_false(memcmp(new_jitter_table, orig_jitter_table,
                the_sensor->sensors[sensor_index].jitter_table_count
                * sizeof(IAS_VECTOR)));

            /* confirm the ssm information */
            if (sensor_index == IAS_OLI)
                assert_true(new_sensor->sensors[sensor_index].ssm_model
                == NULL);
            else
            {
                /* for the second sensor, make sure the ssm info is correct */
                IAS_SENSOR_SCENE_SELECT_MIRROR_MODEL *new_ssm
                    = new_sensor->sensors[sensor_index].ssm_model;
                IAS_SENSOR_SCENE_SELECT_MIRROR_MODEL *orig_ssm
                    = the_sensor->sensors[sensor_index].ssm_model;

                if (index <= 1)
                {
                    /* The format version 2 and version 3 file should have the
                       expected position_status */
                    assert_equal(new_ssm->position_status, 
                        orig_ssm->position_status);
                }
                else
                {
                    /* Format version 1 should have zero for the position
                       status */
                    assert_equal(new_ssm->position_status, 0);
                }

                if (index == 0)
                {
                    /* The format version 3 file should have the expected
                       position_quality */
                    assert_equal(new_ssm->position_quality, 
                        orig_ssm->position_quality);
                }
                else
                {
                    /* Format versions 1 & 2 should have ACTUAL for the
                     * estimate quality */
                    assert_equal(new_ssm->position_quality,
                                 IAS_SSM_POSITION_ACTUAL);
                }

                assert_true(new_ssm != NULL);
                assert_false(memcmp(orig_ssm->alignment_matrix,
                    new_ssm->alignment_matrix, nbytes1));
                assert_false(memcmp(orig_ssm->utc_epoch_time,
                    new_ssm->utc_epoch_time, 3 * sizeof(double)));
                assert_equal(new_ssm->ssm_record_count, 
                    orig_ssm->ssm_record_count);
                assert_true(new_ssm->records != NULL);
                assert_false(memcmp(orig_ssm->records, new_ssm->records,
                    orig_ssm->ssm_record_count* sizeof(IAS_SENSOR_SSM_RECORD)));
            }
        }

        /*------Compare the Instrument Model Information------*/
        /* Compare the Band Model */
        for (band_index = 0; band_index < the_sensor->band_count; band_index++)
        {
            IAS_SENSOR_BAND_MODEL *the_band = &the_sensor->bands[band_index];
            IAS_SENSOR_BAND_MODEL *new_band = &new_sensor->bands[band_index];

            assert_equal(the_band->frame_count, new_band->frame_count);
            assert_equal(the_band->sca_count, new_band->sca_count);
            /* just a quick check to make sure the pointer is set properly */
            assert_equal(the_band->sensor->sensor_id, 
                new_band->sensor->sensor_id);
            assert_double_equal(the_band->utc_epoch_time[0],
                new_band->utc_epoch_time[0]);
            assert_double_equal(the_band->utc_epoch_time[1],
                new_band->utc_epoch_time[1]);
            assert_double_equal(the_band->utc_epoch_time[2],
                new_band->utc_epoch_time[2]);
            if (the_band->frame_seconds_from_epoch)
            {
                assert_true(new_band->frame_seconds_from_epoch != NULL);

                /* a quick check to make sure the pointer is set properly */
                assert_double_equal(the_band->frame_seconds_from_epoch[0],
                    new_band->frame_seconds_from_epoch[0]);
            }
            assert_double_equal(the_band->sampling_char.integration_time,
                new_band->sampling_char.integration_time);
            assert_double_equal(the_band->sampling_char.sampling_time,
                new_band->sampling_char.sampling_time);
            assert_equal(the_band->sampling_char.lines_per_frame,
                new_band->sampling_char.lines_per_frame);
            assert_double_equal(the_band->sampling_char.settling_time,
                new_band->sampling_char.settling_time);
            assert_double_equal(the_band->sampling_char.along_ifov,
                new_band->sampling_char.along_ifov);
            assert_double_equal(the_band->sampling_char.across_ifov,
                new_band->sampling_char.across_ifov);
            assert_double_equal(the_band->sampling_char.maximum_detector_delay,
                new_band->sampling_char.maximum_detector_delay);
            assert_double_equal(
                the_band->sampling_char.time_codes_at_frame_start,
                new_band->sampling_char.time_codes_at_frame_start);

            /* Compare the SCA Model */
            for (sca_index = 0; sca_index < the_band->sca_count; sca_index++)
            {
                IAS_SENSOR_SCA_MODEL *the_sca = &the_band->scas[sca_index];
                IAS_SENSOR_SCA_MODEL *new_sca = &new_band->scas[sca_index];

                if (the_sca)
                {
                    int term;

                    assert_true(new_sca != NULL);
                    assert_equal(the_sca->detectors, new_sca->detectors);
                    assert_equal(the_sca->nominal_fill, new_sca->nominal_fill);

                    assert_false(memcmp(the_sca->l0r_detector_offsets,
                        new_sca->l0r_detector_offsets,
                        the_sca->detectors * sizeof(int)));
                    assert_false(memcmp(the_sca->detector_offsets_along_track,
                        new_sca->detector_offsets_along_track,
                        the_sca->detectors * sizeof(double)));
                    assert_false(memcmp(the_sca->detector_offsets_across_track,
                        new_sca->detector_offsets_across_track,
                        the_sca->detectors * sizeof(double)));

                    for (term = 0; term < IAS_LOS_LEGENDRE_TERMS; term++)
                    {
                        assert_double_equal(the_sca->sca_coef_x[term], 
                            new_sca->sca_coef_x[term]);
                        assert_double_equal(the_sca->sca_coef_y[term], 
                            new_sca->sca_coef_y[term]);
                    }
                }
            }
        }

        /* Compare the frame times */
        for (sensor_index = 0; sensor_index < IAS_MAX_SENSORS; sensor_index++)
        {
            assert_equal(the_sensor->frame_counts[sensor_index],
                new_sensor->frame_counts[sensor_index]);
            nbytes1 = the_sensor->frame_counts[sensor_index] * sizeof(double);
            assert_false(memcmp(
                the_sensor->frame_seconds_from_epoch[sensor_index],
                new_sensor->frame_seconds_from_epoch[sensor_index], nbytes1));
        }

        /* Remove the file if requested */
        if (remove_file[index])
            assert_false(remove(filename[index]));
        ias_los_model_free(new_model);
        new_model = NULL;
    }

    ias_los_model_free(the_model);
}

static int fill_model(IAS_LOS_MODEL *mod)
{
    int sensor_index;
    int sca_index;
    int band_index;
    int band_number;
    int det;
    int frame;
    int i, j;
    int term;
    double x;
    int image_frame_count;
    double *oli_frame_seconds_from_epoch;
    double *tirs_frame_seconds_from_epoch;
    IAS_SENSOR_MODEL *sensor = &mod->sensor;
    const IAS_BAND_ATTRIBUTES *ias_band_attributes;

    /*------General LOS Model Information------*/
    mod->satellite_id = IAS_L8;
    mod->satellite_number = 8;
    strcpy(mod->sw_version, "L0R v3.4");
    mod->wrs_path = 180;
    mod->wrs_row = 33;
    mod->acquisition_type = IAS_EARTH;
    mod->correction_type = IAS_PRECISION;

    /*------Earth Model------*/
    mod->earth.ut1_utc_correction = 1.74;
    mod->earth.pole_wander_x = 0.653;
    mod->earth.pole_wander_y = 1.654;
    mod->earth.semi_major_axis = 6493.92;
    mod->earth.semi_minor_axis = 87302.84;
    mod->earth.eccentricity = 0.12345;
    mod->earth.earth_angular_velocity = 4.87209;
    mod->earth.speed_of_light = 1.86e6;

    /*------Spacecraft Model Information------*/
    /* Attitude Model */
    IAS_SC_ATTITUDE_MODEL *attitude = &mod->spacecraft.attitude;
    attitude->utc_epoch_time[0] = 2008.8;
    attitude->utc_epoch_time[1] = 268.7;
    attitude->utc_epoch_time[2] = 4286.9;
    attitude->nominal_sample_time = x = 1.0;
    attitude->sample_count = 55;
    attitude->sample_records = malloc(sizeof(IAS_SC_ATTITUDE_RECORD) * 
            attitude->sample_count);
    if (!attitude->sample_records)
        return 1;
    for (i = 0; i < attitude->sample_count; i++, x += 1.0)
    {
        attitude->sample_records[i].seconds_from_epoch = x;
        attitude->sample_records[i].satellite_attitude.x = x+0.1;
        attitude->sample_records[i].satellite_attitude.y = x+0.2;
        attitude->sample_records[i].satellite_attitude.z = x+0.3;
        attitude->sample_records[i].precision_attitude.x = x+0.4;
        attitude->sample_records[i].precision_attitude.y = x+0.5;
        attitude->sample_records[i].precision_attitude.z = x+0.6;
    }

    /* Ephemeris Model */
    IAS_SC_EPHEMERIS_MODEL *ephemeris = &mod->spacecraft.ephemeris;
    ephemeris->utc_epoch_time[0] = 2007.7;
    ephemeris->utc_epoch_time[1] = 14.8;
    ephemeris->utc_epoch_time[2] = 66654.9;
    ephemeris->nominal_sample_time = x = 1.0;
    ephemeris->sample_count = 14;
    ephemeris->sample_records
        = malloc(sizeof(IAS_SC_EPHEMERIS_RECORD) * ephemeris->sample_count);
    if (!ephemeris->sample_records)
        return 1;
    for (i = 0; i < ephemeris->sample_count; i++, x += .0005)
    {
        ephemeris->sample_records[i].seconds_from_epoch = x;
        ephemeris->sample_records[i].eci_position.x = x + 0.002;
        ephemeris->sample_records[i].eci_position.y = x + 0.01;
        ephemeris->sample_records[i].eci_position.z = x + 0.02;
        x += 1.0;
        ephemeris->sample_records[i].eci_velocity.x = x + 0.002;
        ephemeris->sample_records[i].eci_velocity.y = x + 0.01;
        ephemeris->sample_records[i].eci_velocity.z = x + 0.02;
        x += 1.0;
        ephemeris->sample_records[i].ecef_position.x = x + 0.002;
        ephemeris->sample_records[i].ecef_position.y = x + 0.01;
        ephemeris->sample_records[i].ecef_position.z = x + 0.02;
        x += 1.0;
        ephemeris->sample_records[i].ecef_velocity.x = x + 0.002;
        ephemeris->sample_records[i].ecef_velocity.y = x + 0.01;
        ephemeris->sample_records[i].ecef_velocity.z = x + 0.02;
        x += 1.0;
        ephemeris->sample_records[i].precision_eci_position.x = x + 0.002;
        ephemeris->sample_records[i].precision_eci_position.y = x + 0.01;
        ephemeris->sample_records[i].precision_eci_position.z = x + 0.02;
        x += 1.0;
        ephemeris->sample_records[i].precision_eci_velocity.x = x + 0.002;
        ephemeris->sample_records[i].precision_eci_velocity.y = x + 0.01;
        ephemeris->sample_records[i].precision_eci_velocity.z = x + 0.02;
        x += 1.0;
        ephemeris->sample_records[i].precision_ecef_position.x = x + 0.002;
        ephemeris->sample_records[i].precision_ecef_position.y = x + 0.01;
        ephemeris->sample_records[i].precision_ecef_position.z = x + 0.02;
        x += 1.0;
        ephemeris->sample_records[i].precision_ecef_velocity.x = x + 0.002;
        ephemeris->sample_records[i].precision_ecef_velocity.y = x + 0.01;
        ephemeris->sample_records[i].precision_ecef_velocity.z = x + 0.02;
    }

    /* Precision Model */
    mod->spacecraft.precision.seconds_from_image_epoch = 87654.4;
    mod->spacecraft.precision.ephemeris_order = 3;
    mod->spacecraft.precision.attitude_order = 2;
    for (i = 0, x = 0.0; i < IAS_PRECISION_MAX_POLY_COEFF; i++, x += 2.0)
    {
        mod->spacecraft.precision.x_corr[i] = x;
        mod->spacecraft.precision.y_corr[i] = x + 0.1;
        mod->spacecraft.precision.z_corr[i] = x + 0.2;
        mod->spacecraft.precision.roll_corr[i] = x + 0.3;
        mod->spacecraft.precision.pitch_corr[i] = x + 0.4;
        mod->spacecraft.precision.yaw_corr[i] = x + 0.5;
    }

    /*------Sensor Model Information------*/
    x = 9.0;
    for (sensor_index = 0; sensor_index < IAS_MAX_SENSORS; sensor_index++)
    {
        IAS_VECTOR *jitter_table;

        sensor->sensors[sensor_index].sensor_id = sensor_index;
        sensor->sensors[sensor_index].sensor_present = 1;
        for (i = 0; i < TRANSFORMATION_MATRIX_DIM; i++)
        {
            for (j = 0; j < TRANSFORMATION_MATRIX_DIM; j++)
            {
                x += 0.00003;
                sensor->sensors[sensor_index].sensor2acs[i][j] = x;
            }
            x += 0.01;
            sensor->sensors[sensor_index].center_mass2sensor_offset[i] = x;
        }

        /* add some test jitter model data */
        sensor->sensors[sensor_index].jitter_table_count = 50;
        if (sensor_index == IAS_OLI)
            sensor->sensors[sensor_index].jitter_entries_per_frame = 2;
        else
            sensor->sensors[sensor_index].jitter_entries_per_frame = 1;

        /* allocate the jitter table */
        jitter_table = malloc(sizeof(*jitter_table)
                * sensor->sensors[sensor_index].jitter_table_count);
        if (!jitter_table)
        {
            assert_true(jitter_table != NULL);
            return 1;
        }
        sensor->sensors[sensor_index].jitter_table = jitter_table;

        /* populate the jitter table */
        for (i = 0; i < sensor->sensors[sensor_index].jitter_table_count; i++)
        {
            jitter_table[i].x = (double)i;
            jitter_table[i].y = (double)i + 100.4;
            jitter_table[i].z = (double)i + 200.8;
        }

        /* for the second sensor, add SSM information */
        if (sensor_index == IAS_TIRS)
        {
            IAS_SENSOR_SCENE_SELECT_MIRROR_MODEL *ssm;
            double z = 12.0;

            /* allocate the main SSM structure */
            ssm = malloc(sizeof(*ssm));
            if (!ssm)
            {
                assert_true(ssm != NULL);
                return 1;
            }
            sensor->sensors[sensor_index].ssm_model = ssm;
        
            /* populate the alignment matrix */
            for (i = 0; i < TRANSFORMATION_MATRIX_DIM; i++)
            {
                for (j = 0; j < TRANSFORMATION_MATRIX_DIM; j++)
                {
                    z += 0.00003;
                    ssm->alignment_matrix[i][j] = z;
                }
            }

            /* populate the epoch time */
            ssm->utc_epoch_time[0] = 1.0;
            ssm->utc_epoch_time[1] = 2.0;
            ssm->utc_epoch_time[2] = 3.0;
            ssm->position_status = 3;
            ssm->position_quality = 3;

            /* allocate a set of SSM records */
            ssm->ssm_record_count = 4;
            ssm->records 
                = malloc(ssm->ssm_record_count * sizeof(*ssm->records));
            if (!ssm->records)
            {
                assert_true(ssm->records != NULL);
                return 1;
            }
            z = 15.0;
            for (i = 0; i < ssm->ssm_record_count; i++)
            {
                ssm->records[i].seconds_from_epoch = z;
                ssm->records[i].mirror_angle = z + 0.5;
                z++;
            }
        }
    }

    /*------Instrument Model Information------*/
    /* Band Model */
    x = 13.1;
    sensor->band_count = ias_sat_attr_get_normal_band_count();

    for (band_index = 0; band_index < sensor->band_count; band_index++)
    {
        IAS_SENSOR_BAND_MODEL *band = &sensor->bands[band_index];

        band_number = ias_sat_attr_convert_band_index_to_number(band_index);
        ias_band_attributes = ias_sat_attr_get_band_attributes(band_number);

        band->band_present = 1;
        band->sca_count = ias_band_attributes->scas;
        band->sensor = &sensor->sensors[ias_band_attributes->sensor_id];

        band->utc_epoch_time[0] = 2009.0;
        band->utc_epoch_time[1] = 189.0;
        band->utc_epoch_time[2] = 3688.083;

        band->sampling_char.integration_time = 9.0 + band_index;
        band->sampling_char.sampling_time = 4.5 + band_index;
        band->sampling_char.lines_per_frame = 1 + band_index;
        band->sampling_char.settling_time = 0.5 + band_index;
        band->sampling_char.along_ifov = 0.76 + band_index;
        band->sampling_char.across_ifov = 0.86 + band_index;
        band->sampling_char.maximum_detector_delay = 0.6 + band_index;
        if (band_index == 0)
            band->sampling_char.time_codes_at_frame_start = 0;
        else
            band->sampling_char.time_codes_at_frame_start = 1;

        /* SCA Model */
        for (sca_index = 0; sca_index < band->sca_count; sca_index++)
        {
            IAS_SENSOR_SCA_MODEL *sca = &band->scas[sca_index];

            sca->detectors = ias_band_attributes->detectors_per_sca;
            sca->nominal_fill = band_index + sca_index;

            for (det = 0; det < sca->detectors; det++)
                sca->l0r_detector_offsets[det] = band_index + sca_index + det;

            for (det = 0; det < sca->detectors; det++)
            {
                x = 0.5 + band_index + sca_index + det;
                sca->detector_offsets_along_track[det] = x;
                sca->detector_offsets_across_track[det] = x + 1.0;
            }

            x = 13.1 + band_index + sca_index;
            for (term = 0; term < IAS_LOS_LEGENDRE_TERMS; term++)
                sca->sca_coef_x[term] = x++;
            for (term = 0; term < IAS_LOS_LEGENDRE_TERMS; term++)
                sca->sca_coef_y[term] = x++;
        }
    }

    image_frame_count = 636;
    oli_frame_seconds_from_epoch = malloc(sizeof(double) * image_frame_count);
    if (!oli_frame_seconds_from_epoch)
        return 1;
    tirs_frame_seconds_from_epoch = malloc(sizeof(double) * image_frame_count);
    if (!tirs_frame_seconds_from_epoch)
        return 1;

    x = 4.0;
    for (frame = 0; frame < image_frame_count; frame++)
    {
        x += 0.001;
        oli_frame_seconds_from_epoch[frame] = x;
        tirs_frame_seconds_from_epoch[frame] = x;
    }
    if (ias_sensor_set_frame_times(sensor, IAS_OLI,
                oli_frame_seconds_from_epoch, image_frame_count) != SUCCESS)
        return 1;
    if (ias_sensor_set_frame_times(sensor, IAS_TIRS,
                tirs_frame_seconds_from_epoch, image_frame_count) != SUCCESS)
        return 1;

    return 0;
}
