#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/wait.h>
#include <limits.h>                   /* for PATH_MAX */
#include "ias_parm.h"
#include "ias_parm_private.h"
#include "ias_parm_standard_parameters.h"
#include "cgreen/cgreen.h"
#include "ias_logging.h"
#include "ias_odl.h"
#include "ias_satellite_attributes.h" /* for IAS_MAX_NBANDS */

#define NUM_SEASONS 4

/* declare the test suite at the module level so it can be destroyed by
   forked child processes */
static TestSuite *suite;

static int band_list[IAS_MAX_NBANDS] = {IAS_MAX_NBANDS, IAS_MAX_NBANDS,
    IAS_MAX_NBANDS, IAS_MAX_NBANDS, IAS_MAX_NBANDS, IAS_MAX_NBANDS,
    IAS_MAX_NBANDS, IAS_MAX_NBANDS, IAS_MAX_NBANDS, IAS_MAX_NBANDS};
static int band_list_compare[IAS_MAX_NBANDS] = {IAS_MAX_NBANDS, IAS_MAX_NBANDS,
    IAS_MAX_NBANDS, IAS_MAX_NBANDS, IAS_MAX_NBANDS, IAS_MAX_NBANDS,
    IAS_MAX_NBANDS, IAS_MAX_NBANDS, IAS_MAX_NBANDS, IAS_MAX_NBANDS};


Ensure test_ias_parm_read()
{
    const char *odl_file_name = NULL;   /* INPUT : ODL file name to read */
    int status;                         /* function return status */
    int count;
    int combine_scas;
    double pccalpha;
    char work_order_id[IAS_WORKORDER_ID_SIZE];
    char l1g_image[PATH_MAX];

    odl_file_name = "test.odl";

    memset(l1g_image,'\0',PATH_MAX);

    /*------------ start parameter table definition ------------*/
    IAS_PARM_DECLARE_TABLE(parms, 9);

    int default_combine_scas[] = {0};
    IAS_PARM_ADD_INT(parms, COMBINE_SCAS, "flag that the SCAs should be "
        "combined into the output image",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        /* has a limited range of 0-1 */
        1, 0, 1,
        /* has one default value */
        1, default_combine_scas,
        &combine_scas, sizeof(combine_scas), 0);

    /* add a CC resampling alpha parameter */
    double default_pccalpha[] = {0.0};
    IAS_PARM_ADD_DOUBLE(parms, PCCALPHA, "Cubic convolution alpha parameter",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        /* has range from -1.0 to 0.0 */
        1, -1.0, 0.0,
        /* had one default value */
        1, default_pccalpha,
        &pccalpha, sizeof(pccalpha), 0);

    IAS_PARM_ADD_STRING(parms, WORK_ORDER_ID, "Work order ID",
        IAS_PARM_REQUIRED, 
        /* does not have limits */
        0, NULL, 
        /* default */
        1, NULL,
        work_order_id, sizeof(work_order_id), 0);

    /* add a L1G_IMAGE parameter */
    const char *default_l1g_image[] = {"L1G_image.hdf"};
    IAS_PARM_ADD_STRING(parms, L1G_IMAGE, "Name of the L1G image file",
        IAS_PARM_REQUIRED, 
        /* Does not have limits */
        0, NULL, 
        /* default */
        1, default_l1g_image,
        l1g_image, sizeof(l1g_image), 0);

    IAS_PARM_BAND_LIST(parms, IAS_PARM_REQUIRED, band_list,
                            sizeof(band_list), 1);

    /* This assignment is intentionally larger than the actual values
       in the odl as this is allowed  */
    char *algstring[50];

    const char *default_algorithms[] = {"impulse_noise_char_SKIP",
                    "saturated_pixel_char", "bias_removal_DUMP",
                    "bias_determination"};
    IAS_PARM_ADD_STRING_ARRAY(parms, ALGORITHMS, "Test String Array",
        IAS_PARM_REQUIRED,
        0, NULL,
        4, default_algorithms, 
        algstring, sizeof(algstring), 0);

    /* Parameter: Single int with array of valid ranges */
    int multiple_ranges[] = {2, 2, 4, 4, 1, 1};
    int delay_flag_test;
    int int_def_value = 4;
    IAS_PARM_ADD_INT_WITH_RANGES(parms, DELAY_FLAG, 
        "Integer Example With Valid Ranges",
        IAS_PARM_REQUIRED, IAS_PARM_NOT_ARRAY,
        6, multiple_ranges,
        1, &int_def_value,
        &delay_flag_test, sizeof(delay_flag_test), 0);

    /* ODTYPE is a non-array string with a list of constraints */
    char local_odtype[5] = "BYTE";
    const char *valid_odtype_options[] = {"BYTE", "I*2", "I*4", "R*4"};
    const char *default_odtype[] = {"I*4"};
    IAS_PARM_ADD_STRING(parms, ODTYPE,
        "Character string storage for the output data type",
        IAS_PARM_OPTIONAL,
        /* has a list of valid options */
        4, valid_odtype_options,
        /* has a default value */
        1, default_odtype,
        local_odtype, sizeof(local_odtype), 0);

    char local_sensor[12];
    IAS_PARM_SENSOR(parms, IAS_PARM_REQUIRED, local_sensor,
        sizeof(local_sensor), 1);
    
    /* calculate number of items in the ODL_param_list */
    count = IAS_PARM_GET_TABLE_SIZE(parms);
    assert_equal(count, 9);

    /* call the routine to verify that the file exists and can be read */
    status = ias_parm_provide_help(odl_file_name, parms, count, 
                                   IAS_INPUT_PARAMETERS);
    assert_false(status);
    if (status)
        return;

    /* test reading odl file */
    status = ias_parm_read(odl_file_name, parms, count);
    assert_equal(status, SUCCESS);

    /* test retrieved parameters */
    assert_equal(combine_scas, 1);
    assert_string_equal(l1g_image, "EO1A1350462005074110PY_HDF.L1G");
    assert_string_equal(work_order_id, "IX223");
    assert_double_equal(pccalpha, -0.5);
    assert_string_equal(algstring[0], "impulse_noise_char_SKIP");
    assert_string_equal(algstring[1], "saturated_pixel_char");
    assert_string_equal(algstring[2], "bias_removal_DUMP");
    assert_string_equal(algstring[3], "bias_determination");
    assert_string_equal(local_odtype, "I*2");
    assert_string_equal(local_sensor, "L8_OLITIRS");

    /* Free the pointer for the string array test */
    int i;
    for (i = 0; i < 4; i++)
        free(algstring[i]);


    /*------start parameter table definition with int range error ------*/
    IAS_PARM_DECLARE_TABLE(parms_int_error, 1);

    /* Parameter: Single int with array of invalid multiple ranges */
    int multiple_rangesint[] = {2, 2, 4, 5, 1, 1};
    int_def_value = 4;
    IAS_PARM_ADD_INT_WITH_RANGES(parms_int_error, DELAY_FLAG, 
        "Integer Example With Invalid Ranges",
        IAS_PARM_REQUIRED, IAS_PARM_NOT_ARRAY,
        6, multiple_rangesint,
        1, &int_def_value,
        &delay_flag_test, sizeof(delay_flag_test), 0);

    /* test reading odl file */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = ias_parm_read(odl_file_name, parms_int_error, count);
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);
    assert_equal(status, ERROR);


    /*------start parameter table definition with double range error ------*/
    IAS_PARM_DECLARE_TABLE(parms_double_error, 1);

    /* Parameter:  Array of doubles with array of invalid multiple ranges */
    double error_double_ranges[] = {1.0, 1.0, 10.2, 10.2, 32764.0, 32767.0};

    double minmax_output_dn[2];

    IAS_PARM_ADD_DOUBLE_WITH_RANGES(parms_double_error, MINMAX_OUTPUT_DN, 
        "Double example for With invalid Ranges Array",
        IAS_PARM_OPTIONAL, IAS_PARM_ARRAY,
        6, error_double_ranges,
        0, NULL,
        minmax_output_dn, sizeof(minmax_output_dn), 0);

    /* test reading odl file */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = ias_parm_read(odl_file_name, parms_double_error, count);
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);
    assert_equal(status, ERROR);

    
}


Ensure test_ias_parm_update()
{
    const char *odl_file_name = NULL;  /* INPUT : ODL file name to read */
    int     status;                    /* function return status */
    int     count;
    char l1g_image[PATH_MAX];
    char l1g_image_compare[PATH_MAX];
    int band_count;
    int band_count_compare;
    
    odl_file_name = "test2.odl";

    memset(l1g_image,'\0',PATH_MAX);
    memset(l1g_image_compare,'\0',PATH_MAX);

    /* create a test file for updating */
    system("cp ./test.odl ./test2.odl");

    /*------------ start parameter table definition ------------*/
    IAS_PARM_DECLARE_TABLE(parms, 2);
    
    /* add a L1G_IMAGE parameter */
    sprintf(l1g_image,"NEW_DEC_L1G.HDF");
    const char *default_l1g_image[] = {"L1G_image2.hdf"};
    IAS_PARM_ADD_STRING(parms, L1G_IMAGE, "Name of the L1G image file",
        IAS_PARM_REQUIRED, 0, NULL, 1, default_l1g_image,
        l1g_image, sizeof(l1g_image), 0);
    IAS_PARM_SET_COUNT_WRITE(parms, L1G_IMAGE, 1);

    int band_list[] = {1,2,3,7};
    IAS_PARM_BAND_LIST(parms, IAS_PARM_REQUIRED, band_list,
                            sizeof(band_list), 1);

    IAS_PARM_SET_COUNT_WRITE(parms, BAND_LIST, 4);

    /* calculate number of items in the ODL_param_list */
    count = IAS_PARM_GET_TABLE_SIZE(parms);
    assert_equal(count, 2);
    
    /* test updating odl file */
    status = ias_parm_update(odl_file_name, parms, count);
    assert_equal(status, SUCCESS);

    /*------------ start parameter table definition ------------*/
    IAS_PARM_DECLARE_TABLE(parms_compare, 2);

    /* add a L1G_IMAGE parameter */
    IAS_PARM_ADD_STRING(parms_compare, L1G_IMAGE, "Name of the L1G image file",
        IAS_PARM_REQUIRED, 0, NULL, 1, default_l1g_image,
        l1g_image_compare, sizeof(l1g_image_compare), 0);

    IAS_PARM_BAND_LIST(parms_compare, IAS_PARM_REQUIRED,
                            band_list_compare, sizeof(band_list_compare), 1);

    /* call the routine to provide help for the parameters if requested 
       also verifies the file exists */
    status = ias_parm_provide_help(odl_file_name, parms, count,
                                   IAS_INPUT_PARAMETERS);
    assert_false(status);
    if (status)
        return;

    /* test reading odl file */
    status = ias_parm_read(odl_file_name, parms_compare, count);
    assert_equal(status, SUCCESS);

    /* compare l1g_image values */
    assert_string_equal(l1g_image_compare, "NEW_DEC_L1G.HDF");

    /* compare band_list values */
    assert_equal(memcmp(band_list, band_list_compare, sizeof(band_list)), 0);

    band_count = IAS_PARM_GET_COUNT_READ(parms, BAND_LIST);
    band_count_compare = IAS_PARM_GET_COUNT_READ(parms_compare, BAND_LIST);
    assert_equal(band_count, band_count_compare);
}


Ensure test_ias_parm_help()
{
    int combine_scas;
    double pccalpha;
    char work_order_id[IAS_WORKORDER_ID_SIZE];
    char l1g_image[PATH_MAX];
    const char *filename1 = "test_help.txt";
    const char *filename2 = "reference_test_help.txt";
    int match = 1;
    char buffer1[256] = {""};
    char buffer2[256] = {""};
    FILE *help_stream = NULL;
    FILE *stream1 = NULL;
    FILE *stream2 = NULL;

    memset(l1g_image,'\0',PATH_MAX);

    IAS_PARM_DECLARE_TABLE(parms, 10);

    int default_combine_scas[] = {1};
    IAS_PARM_ADD_INT(parms, COMBINE_SCAS, "flag to indicate the SCAs should be "
        "combined into the output image",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        /* has a limited range of 0-1 */
        1, 0, 1,
        /* has one default value */
        1, default_combine_scas,
        &combine_scas, sizeof(combine_scas), 0);

    /* add a CC resampling alpha parameter */
    double default_pccalpha[] = {0.0};
    IAS_PARM_ADD_DOUBLE(parms, PCCALPHA, "Cubic convolution alpha parameter",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        /* has range from -1.0 to 0.0 */
        1, -1.0, 0.0,
        /* had one default value */
        1, default_pccalpha,
        &pccalpha, sizeof(pccalpha), 0);

    /* add a resampling option parameter */
    IAS_PARM_ADD_STRING(parms, WORK_ORDER_ID, "Work order ID",
        IAS_PARM_REQUIRED,
        /* doesn't have a list of valid options */
        0, NULL,
        /* doesn't have a default value */
        0, NULL,
        work_order_id, sizeof(work_order_id), 0);

    /* add a L1G_IMAGE parameter */
    const char *default_l1g_image[] = {"L1G_image.hdf"};
    IAS_PARM_ADD_STRING(parms, L1G_IMAGE, "Name of the L1G image file",
        IAS_PARM_REQUIRED, 0, NULL, 1, default_l1g_image,
        l1g_image, sizeof(l1g_image), 0);

    /* Parameter: String, array, no valids, has defaults */

    /* This assignment is intentionally larger than the actual values
       in the odl as noted in actual example  */
    char *algstring[30];

    const char *default_algorithms[] = {"impulse_noise_char_SKIP",
                    "saturated_pixel_char", "bias_removal_DUMP",
                    "bias_determination"};
    IAS_PARM_ADD_STRING_ARRAY(parms, ALGORITHMS, 
        "An algorithm used in the RPS Driver",
        IAS_PARM_REQUIRED,
        0, NULL,
        4, default_algorithms, 
        algstring, sizeof(algstring), 0);

    /* Parameter: Double, array, list of valids, with defaults */
    double valid_double_ranges[] = {1.5, 1.5, 10.2, 10.2, 30.4, 30.4};
    double double_range_value[] = {1.5, 10.2, 30.4};

    IAS_PARM_ADD_DOUBLE_WITH_RANGES(parms, RANGES_ARR_DOUBLE, 
        "Valid Double Ranges Array",
        IAS_PARM_OPTIONAL, IAS_PARM_ARRAY,
        /* Valids */
        6, valid_double_ranges,
        /* Defaults */
        3, double_range_value,
        double_range_value, sizeof(double_range_value), 0);

    /* Parameter: Integer, array, has valid range, includes defaults */

    IAS_PARM_BAND_LIST(parms, IAS_PARM_REQUIRED, band_list,
                            sizeof(band_list), 1);

    /* Parameter: Integer, array, has valid list, includes defaults */
    int valid_int_values[] = {5, 5, 10, 10, 20, 20, 30, 30};
    int test_value[5];
    int default_int[3] = {5, 5, 30};
    IAS_PARM_ADD_INT_WITH_RANGES(parms, RANGES_ARRAY, "Valid Ranges Array",
        IAS_PARM_OPTIONAL, IAS_PARM_ARRAY,
        /* Valids */
        8, valid_int_values,
        /* Defaults */
        3, default_int,
        test_value, sizeof(test_value), 0);

    /* Parameter: String, not array, has valid list, includes default */
    const char *string_valids[8] = {"PATH_MINBOX","MAXBOX","PATH_MAXBOX",
                "GEOBOX", "PROJBOX", "ULSIZE", "MINBOX", "PATH"};
    char test_string[12];

    const char *default_frame_type[] = {"PATH_MINBOX"};
    IAS_PARM_ADD_STRING(parms, FRAME_TYPE, 
        "Output image frame computation method",
        IAS_PARM_REQUIRED,
        /* list of valid options */
        8, string_valids,
        /* default value */
        1, default_frame_type,
        test_string, sizeof(test_string), 0);

    char local_sensor[12];
    IAS_PARM_SENSOR(parms, IAS_PARM_REQUIRED, local_sensor,
        sizeof(local_sensor), 1);

    pid_t fork_pid;
    
    /* fork a child process to call the help routine with the template option
       since the redirecting of stdout cannot be done in this process without
       disabling all future output to stdout after the help routine returns */
    fork_pid = fork();
    if (fork_pid == -1)
    {
        assert_true_with_message(0, "Failed to fork child");
        return;
    }
    else if (fork_pid == 0)
    {
        /* child process, so redirect stdout and create the output */
        int status;

        /* Open the file to write the help output to */
        help_stream = freopen(filename1, "w", stdout);
        assert_not_equal(help_stream, NULL);
        if (!help_stream)
            exit(-1);

        /* dump out help for the parameters */
        status = ias_parm_provide_help("--help", parms,
                    IAS_PARM_GET_TABLE_SIZE(parms), IAS_OMF_PARAMETERS);
        fclose(help_stream);

        /* destroy the cgreen resources in the child to eliminate memory leaks
           reported by valgrind */
        destroy_reporter(get_test_reporter());
        destroy_test_suite(suite);

        if (status == 1)
            exit(0);
        else
            exit(-1);
    }
    else
    {
        int status;
        pid_t returned_pid;

        /* parent process, so wait for child to finish */
        returned_pid = wait(&status);
        if (returned_pid != fork_pid)
        {
            assert_true_with_message(0, "Unexpected child returned from wait");
            return;
        }
        if (!WIFEXITED(status))
        {
            assert_true_with_message(0, "Unexpected status returned from "
                        "child");
            return;
        }
    }

    /* diff the two files to make sure they're the same */
    stream1 = fopen(filename1, "r");
    assert_not_equal(stream1, NULL);
    if (!stream1)
        return;
    stream2 = fopen(filename2, "r");
    assert_not_equal(stream2, NULL);
    if (!stream2)
        return;
    while(!feof(stream1) && !feof(stream2))
    {
        fgets(buffer1,sizeof(buffer1),stream1);
        fgets(buffer2,sizeof(buffer2),stream2);
        if(strcmp(buffer1,buffer2) != 0)
        {
            match = 0;
            break;
        }
    }
    fclose(stream1);
    fclose(stream2);
    stream1 = NULL;
    stream2 = NULL;
    assert_equal(match, 1);
}


Ensure test_ias_parm_template()
{
    int combine_scas;
    double pccalpha;
    char work_order_id[IAS_WORKORDER_ID_SIZE];
    char l1g_image[PATH_MAX];
    const char *filename1 = "test_template.txt";
    const char *filename2 = "reference_test_template.txt";
    int match = 1;
    char buffer1[256] = {""};
    char buffer2[256] = {""};
    FILE *stream1 = NULL;
    FILE *stream2 = NULL;

    memset(l1g_image,'\0',PATH_MAX);

    IAS_PARM_DECLARE_TABLE(parms, 10);

    int default_combine_scas[] = {1};
    IAS_PARM_ADD_INT(parms, COMBINE_SCAS, "flag to indicate the SCAs should be "
        "combined into the output image",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        /* has a limited range of 0-1 */
        1, 0, 1,
        /* has one default value */
        1, default_combine_scas,
        &combine_scas, sizeof(combine_scas), 0);

    /* add a CC resampling alpha parameter */
    double default_pccalpha[] = {0.0};
    IAS_PARM_ADD_DOUBLE(parms, PCCALPHA, "Cubic convolution alpha parameter",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        /* has range from -1.0 to 0.0 */
        1, -1.0, 0.0,
        /* had one default value */
        1, default_pccalpha,
        &pccalpha, sizeof(pccalpha), 0);

    /* add a resampling option parameter */
    IAS_PARM_ADD_STRING(parms, WORK_ORDER_ID, "Work order ID",
        IAS_PARM_REQUIRED,
        /* doesn't have a list of valid options */
        0, NULL,
        /* doesn't have a default value */
        0, NULL,
        work_order_id, sizeof(work_order_id), 0);

    /* add a L1G_IMAGE parameter */
    const char *default_l1g_image[] = {"L1G_image.hdf"};
    IAS_PARM_ADD_STRING(parms, L1G_IMAGE, "Name of the L1G image file",
        IAS_PARM_REQUIRED, 0, NULL, 1, default_l1g_image,
        l1g_image, sizeof(l1g_image), 0);

    /* Parameter: String, array, no valids, has defaults */
    /* This assignment is intentionally larger than the actual values
       in the odl as noted in actual example  */
    char *algstring[30];
    const char *default_algorithms[] = {"impulse_noise_char_SKIP",
                    "saturated_pixel_char", "bias_removal_DUMP",
                    "bias_determination"};
    IAS_PARM_ADD_STRING_ARRAY(parms, ALGORITHMS, 
        "An algorithm used in the RPS Driver",
        IAS_PARM_REQUIRED,
        0, NULL,
        4, default_algorithms, 
        algstring, sizeof(algstring), 0);

    /* Parameter: Double, array, list of valids, with defaults */
    double valid_double_ranges[] = {1.5, 1.5, 10.2, 10.2, 30.4, 30.4};
    double double_range_value[] = {1.5, 10.2, 30.4};

    IAS_PARM_ADD_DOUBLE_WITH_RANGES(parms, RANGES_ARR_DOUBLE, 
        "Valid Double Ranges Array",
        IAS_PARM_OPTIONAL, IAS_PARM_ARRAY,
        /* Valids */
        6, valid_double_ranges,
        /* Defaults */
        3, double_range_value,
        double_range_value, sizeof(double_range_value), 0);

    /* Parameter: Integer, array, has valid range, includes defaults */
    IAS_PARM_BAND_LIST(parms, IAS_PARM_REQUIRED, band_list,
                            sizeof(band_list), 1);

    /* Parameter: Integer, array, has valid list, includes defaults */
    int valid_int_values[] = {5, 5, 10, 10, 20, 20, 30, 30};
    int test_value[5];
    int default_int[3] = {5, 5, 30};
    IAS_PARM_ADD_INT_WITH_RANGES(parms, RANGES_ARRAY, "Valid Ranges Array",
        IAS_PARM_OPTIONAL, IAS_PARM_ARRAY,
        /* Valids */
        8, valid_int_values,
        /* Defaults */
        3, default_int,
        test_value, sizeof(test_value), 0);

    /* Parameter: String, not array, has valid list, includes default */
    const char *string_valids[8] = {"PATH_MINBOX","MAXBOX","PATH_MAXBOX",
                "GEOBOX", "PROJBOX", "ULSIZE", "MINBOX", "PATH"};
    char test_string[12];

    const char *default_frame_type[] = {"PATH_MINBOX"};
    IAS_PARM_ADD_STRING(parms, FRAME_TYPE, 
        "Output image frame computation method",
        IAS_PARM_REQUIRED,
        /* list of valid options */
        8, string_valids,
        /* default value */
        1, default_frame_type,
        test_string, sizeof(test_string), 0);

    char local_sensor[12];
    IAS_PARM_SENSOR(parms, IAS_PARM_REQUIRED,local_sensor,
        sizeof(local_sensor), 1);

    pid_t fork_pid;

    /* fork a child process to call the help routine with the template option
       since the redirecting of stdout cannot be done in this process without
       disabling all future output to stdout after the help routine returns */
    fork_pid = fork();
    if (fork_pid == -1)
    {
        assert_true_with_message(0, "Failed to fork child");
        return;
    }
    else if (fork_pid == 0)
    {
        /* child process, so redirect stdout and create the output */
        int status;
        FILE *template_stream = NULL;

        template_stream = freopen(filename1, "w", stdout);
        if (!template_stream)
            exit(-1);

        /* dump out a template for the parameters */
        status = ias_parm_provide_help("--template", parms, 
                  IAS_PARM_GET_TABLE_SIZE(parms), IAS_INPUT_PARAMETERS);

        fclose(template_stream);

        /* destroy the cgreen resources in the child to eliminate memory leaks
           reported by valgrind */
        destroy_reporter(get_test_reporter());
        destroy_test_suite(suite);

        if (status == 1)
            exit(0);
        else
            exit(-1);
    }
    else
    {
        int status;
        pid_t returned_pid;

        /* parent process, so wait for child to finish */
        returned_pid = wait(&status);
        if (returned_pid != fork_pid)
        {
            assert_true_with_message(0, "Unexpected child returned from wait");
            return;
        }
        if (!WIFEXITED(status))
        {
            assert_true_with_message(0, "Unexpected status returned from "
                        "child");
            return;
        }
    }

    /* diff the two files to make sure they're the same */
    stream1 = fopen(filename1, "r");
    assert_not_equal(stream1, NULL);
    if (!stream1)
        return;
    stream2 = fopen(filename2, "r");
    assert_not_equal(stream2, NULL);
    if (!stream2)
        return;
    while(!feof(stream1) && !feof(stream2))
    {
        fgets(buffer1,sizeof(buffer1),stream1);
        fgets(buffer2,sizeof(buffer2),stream2);
        if(strcmp(buffer1,buffer2) != 0)
        {
            match = 0;
            break;
        }
    }
    fclose(stream1);
    fclose(stream2);
    stream1 = NULL;
    stream2 = NULL;

    assert_equal(match, 1);
}


Ensure test_check_parameter_name_length()
{
    const char *odl_file_name = NULL;    /* Name of ODL file to read */

    char work_order_id[IAS_WORKORDER_ID_SIZE]; /* Work order ID string */

    int status;                          /* Function return status */
    int count;                           /* Number of items read */
    int default_row = 19;                /* Default row */
    int default_path = 29;               /* Default path */
    int path;                            /* WRS2 path to read */
    int row;                             /* WRS2 row to read */
    double pccalpha;                     /* Cubic convolution alpha
                                            parameter value to be read */
    double default_pccalpha = -0.5;      /* Default CC alpha value */

    odl_file_name = "test3.odl";

    /* Set up parameter table */
    IAS_PARM_DECLARE_TABLE(param_table, 4);

    /* Add the work order parameter */
    IAS_PARM_ADD_STRING(param_table, WORK_ORDER_ID, "Work order ID",
        IAS_PARM_REQUIRED,
        /* doesn't have a list of valid options */
        0, NULL,
        /* doesn't have a default value */
        0, NULL,
        work_order_id, sizeof(work_order_id), 0);

    /* Add the cubic convolution parameter */
    IAS_PARM_ADD_DOUBLE(param_table,
        L8_OLITIRS_RESAMPLING_CUBIC_CONVOLUTION_ALPHA,
        "Cubic convolution alpha parameter",
        IAS_PARM_OPTIONAL,
        IAS_PARM_NOT_ARRAY,
        /* has range from -1.0 to 0.0 */
        1, -1.0, 0.0,
        /* has 1 default value */
        1, &default_pccalpha,
        &pccalpha, sizeof(pccalpha), 0);

    /* Add the path and row parameters */
    IAS_PARM_ADD_INT(param_table, PATH_WRS, "WRS PATH",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        /* has a limited range of 1-233 */
        1, 1, 233,
        /* has one default value */
        1, &default_path,
        &path, sizeof(path), 0);

    IAS_PARM_ADD_INT(param_table, ROW_WRS, "WRS ROW",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        /* has a limited range of 1-248 */
        1, 1, 248,
        /* has one default value */
        1, &default_row,
        &row, sizeof(row), 0);

    /* calculate number of items in the ODL_param_list */
    count = IAS_PARM_GET_TABLE_SIZE(param_table);
    assert_equal(count, 4);

    /* Call the routine to provide help for the parameters if requested 
       also verifies the file exists */
    status = ias_parm_provide_help(odl_file_name, param_table, count,
                                   IAS_INPUT_PARAMETERS);
    assert_false(status);
    if (status != 0)
        return;

    /* Test read the ODL file.  It should fail because the cubic convolution
       parameter name is too long for the database */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = ias_parm_read(odl_file_name, param_table, count);
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);
    assert_equal(status, ERROR);

    /* Done */
}


/* test that default values work when an optional parameter is missing */
Ensure test_ias_parm_default_values()
{
    const char *odl_file_name = NULL;  /* INPUT : ODL file name to read */
    int     status;                    /* function return status */
    int     count;
    int path = 1;
    int row = 1;
    double pccalpha = -1.0;
    double min_corr = 0.0;
    char l0r_image[PATH_MAX];
    int i = 0;                         /* counter for loop */
    int satellite;

    odl_file_name = "test.odl";

    memset(l0r_image,'\0',PATH_MAX);

    /*------------ start parameter table definition ------------*/
    IAS_PARM_DECLARE_TABLE(parms, 11);

    /* Parameter 1 */
    int default_path[] = {29};
    IAS_PARM_ADD_INT(parms, PATH_WRS, "WRS PATH",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        /* has a limited range of 1-233 */
        1, 1, 233,
        /* has one default value */
        1, default_path,
        &path, sizeof(path), 0);

    /* Parameter 2 */ 
    int default_row[] = {19};
    IAS_PARM_ADD_INT(parms, ROW_WRS, "WRS ROW",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        /* has a limited range of 1-248 */
        1, 1, 248,
        /* has one default value */
        1, default_row,
        &row, sizeof(row), 0);

    /* Parameter 3:  add a CC resampling alpha parameter */
    double default_pccalpha[] = {0.0};
    IAS_PARM_ADD_DOUBLE(parms, PCCALPHA, "Cubic convolution alpha parameter",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        /* has range from -1.0 to 0.0 */
        1, -1.0, 0.0,
        /* has one default value */
        1, default_pccalpha,
        &pccalpha, sizeof(pccalpha), 0);

    /* Parameter 4:  Get the minimum correlation value. */
    double default_min_corr[] = {1.0};
    IAS_PARM_ADD_DOUBLE(parms, MIN_CORR, "Minimum correlation strength", 
         IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY, 
        /* does not have a range */
         0, 0, 0, 
        /* has one default value */
         1, default_min_corr, 
         &min_corr, sizeof(min_corr), 0);


    /* Parameter 5:  add a L0R_IMAGE parameter */
    const char *default_l0r_image[] = {"L0R_image.hdf"};
    IAS_PARM_ADD_STRING(parms, L0R_IMAGE, "Name of the L0R image file",
        IAS_PARM_OPTIONAL, 0, NULL, 1, default_l0r_image,
        l0r_image, sizeof(l0r_image), 0);

    /* Parameter 6:  Get the list of bands. */
    static int band_number_list[IAS_MAX_NBANDS];
    int default_band_list_table[] = {1, 2, 3, 4, 5, 6, 7, 8, 9, 10};
    IAS_PARM_ADD_INT(parms, BAND_LIST, "Bands for correlation",
        IAS_PARM_OPTIONAL, IAS_PARM_ARRAY, 1, 1, 10, 10,
        default_band_list_table, band_number_list, sizeof(band_number_list), 1);

    /*  Parameter 7:  Get the chip size values. */
    static double local_chip_size[2];
    double default_chip_size[2];
    default_chip_size[0] = 64;
    default_chip_size[1] = 64;
    IAS_PARM_ADD_DOUBLE(parms, CHIP_SIZE,
        "Correlation chip size (line sample)", IAS_PARM_OPTIONAL,
         IAS_PARM_ARRAY, 0, 0, 0, 2,
         default_chip_size, local_chip_size,
         sizeof(local_chip_size), 0);

    /* Parameter 8: Single int with array of valid ranges */
    int multiple_ranges[] = {2, 2, 5, 5, 1, 1};
    int delay_flag_test;
    IAS_PARM_ADD_INT_WITH_RANGES(parms, DELAY_FLAG, 
        "Integer Example With Valid Ranges",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        6, multiple_ranges,
        0, NULL,
        &delay_flag_test, sizeof(delay_flag_test), 0);

    /* Parameter 9:  Array of doubles with array of valid ranges */
    double mult_double_ranges[] = {1.0, 1.0, 10.2, 10.2, 32767.0, 32767.0};

    double minmax_output_dn[2];

    IAS_PARM_ADD_DOUBLE_WITH_RANGES(parms, MINMAX_OUTPUT_DN, 
        "Double example for With Valid Ranges Array",
        IAS_PARM_OPTIONAL, IAS_PARM_ARRAY,
        6, mult_double_ranges,
        0, NULL,
        minmax_output_dn, sizeof(minmax_output_dn), 0);

    /* Parameter 10 */
    IAS_PARM_SATELLITE(parms, &satellite, sizeof(satellite), 1);

    /* Parameter 11:  array of strings */
    char *seasons[NUM_SEASONS]; /* Temporary char ptrs for string arrays */
    for (i=0; i<NUM_SEASONS; i++)
        seasons[i] = NULL;
    const char *default_seasons[] = {"WINTER", "SPRING"};
    const char *valid_seasons[] = {"WINTER", "SPRING", "SUMMER", "FALL"};
    IAS_PARM_ADD_STRING_ARRAY(parms, SEASONS,
        "Seasons, used to filter GCPs used: WINTER, SPRING, SUMMER, FALL",
        IAS_PARM_OPTIONAL, 4, valid_seasons, 2, default_seasons,
        seasons, sizeof(seasons), 0);

    /* calculate number of items in the ODL_param_list */
    count = IAS_PARM_GET_TABLE_SIZE(parms);
    assert_equal(count, 11);

    /* call the routine to provide help for the parameters if requested 
       also verifies the file exists */
    status = ias_parm_provide_help(odl_file_name, parms, count,
                                   IAS_INPUT_PARAMETERS);
    assert_false(status);
    if (status != 0)
        return;

    /* test reading odl file */
    status = ias_parm_read(odl_file_name, parms, count);
    assert_equal(status, SUCCESS);

    /* verify that the count of each parameter is set as expected */
    assert_equal(parms[0]->count_read, 1);
    assert_equal(parms[1]->count_read, 1);
    assert_equal(parms[2]->count_read, 1);
    assert_equal(parms[3]->count_read, 1);
    assert_equal(parms[4]->count_read, 1);
    assert_equal(parms[5]->count_read, 10);
    assert_equal(parms[6]->count_read, 2);
    assert_equal(parms[7]->count_read, 1);
    assert_equal(parms[8]->count_read, 2);
    assert_equal(parms[9]->count_read, 1);
    assert_equal(parms[10]->count_read, 2);

    /* test retrieved parameters */
    assert_equal(path, 29);
    assert_equal(row, 9);
    assert_double_equal(pccalpha, -0.5);
    assert_double_equal(min_corr,  1.0);
    assert_string_equal(l0r_image, "L0R_image.hdf");
    for (i = 0; i < 10; i++)
    {
        assert_equal(band_number_list[i], (i + 1));
    }
    assert_double_equal(local_chip_size[0], 64);
    assert_double_equal(local_chip_size[1], 64);
    assert_equal(delay_flag_test, 1);
    assert_double_equal(minmax_output_dn[0], 1.0);
    assert_double_equal(minmax_output_dn[1], 32767.0);
    assert_equal(satellite, 8);
    assert_string_equal(seasons[0], "WINTER");
    assert_string_equal(seasons[1], "SPRING");

    /* Free the temp strings */
    for (i = 0; i < NUM_SEASONS; i++)
    {
        free(seasons[i]);
        seasons[i] = NULL;
    }
}

/* Test the load table option of the help routine */
Ensure test_ias_parm_loadtable()
{
    const char *filename1 = "test_loadtable.txt";
    const char *filename2 = "reference_test_loadtable.txt";
    int match = 1;
    char buffer1[256] = {""};
    char buffer2[256] = {""};
    FILE *stream1 = NULL;
    FILE *stream2 = NULL;

    IAS_PARM_DECLARE_TABLE(parms, 10);

    /* Parameter: Int, not an array, has range, has default */
    int combine_scas;
    int default_combine_scas[] = {1};
    IAS_PARM_ADD_INT(parms, COMBINE_SCAS, "flag that the SCAs should be "
        "combined into the output image",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        /* has a limited range of 0 to 1 */
        1, 0, 1,
        /* has a default value */
        1, default_combine_scas,
        &combine_scas, sizeof(combine_scas), 0);

    /* Parameter: Double, not an array, has range, has default */
    double pccalpha;
    double default_pccalpha[] = {0.0};
    IAS_PARM_ADD_DOUBLE(parms, PCCALPHA, "Cubic convolution alpha parameter",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        /* has range from -1.0 to 0.0 */
        1, -1.0, 0.0,
        /* has a default value */
        1, default_pccalpha,
        &pccalpha, sizeof(pccalpha), 0);

    /* Parameter: String, not an array, no valids, no default */
    char work_order_id[IAS_WORKORDER_ID_SIZE];
    IAS_PARM_ADD_STRING(parms, WORK_ORDER_ID, "Work order ID",
        IAS_PARM_REQUIRED,
        /* doesn't have a list of valid options */
        0, NULL,
        /* doesn't have a default value */
        0, NULL,
        work_order_id, sizeof(work_order_id), 0);

    /* Parameter: String, not an array, no valids, has default */
    char l1g_image[PATH_MAX];
    memset(l1g_image,'\0',PATH_MAX);
    const char *default_l1g_image[] = {"L1G_image.hdf"};
    IAS_PARM_ADD_STRING(parms, L1G_IMAGE, "Name of the L1G image file",
        IAS_PARM_REQUIRED, 
        0, NULL, 
        1, default_l1g_image,
        l1g_image, sizeof(l1g_image), 0);

    /* Parameter: String, array, no valids, has defaults */
    /* This assignment is intentionally larger than the actual values
       in the odl as noted in actual example  */
    char *algstring[30];

    const char *default_algorithms[] = {"impulse_noise_char_SKIP",
                    "saturated_pixel_char", "bias_removal_DUMP",
                    "bias_determination"};
    IAS_PARM_ADD_STRING_ARRAY(parms, ALGORITHMS, 
        "An algorithm used in the RPS Driver",
        IAS_PARM_REQUIRED,
        0, NULL,
        4, default_algorithms, 
        algstring, sizeof(algstring), 0);

    /* Parameter: Double, array, list of valids, with defaults */
    double valid_double_ranges[] = {1.5, 1.5, 10.2, 10.2, 30.4, 30.4};
    double double_range_value[] = {1.5, 10.2, 30.4};

    IAS_PARM_ADD_DOUBLE_WITH_RANGES(parms, RANGES_ARR_DOUBLE, 
        "Valid Double Ranges Array",
        IAS_PARM_OPTIONAL, IAS_PARM_ARRAY,
        /* Valids */
        6, valid_double_ranges,
        /* Defaults */
        3, double_range_value,
        double_range_value, sizeof(double_range_value), 0);

    /* Parameter: Integer, array, has valid range, includes defaults */
    static int band_number_list[IAS_MAX_NBANDS];
    IAS_PARM_BAND_LIST(parms, IAS_PARM_REQUIRED, band_number_list, 
                sizeof(band_number_list), 1);

    /* Parameter: Integer, array, has valid list, includes defaults */
    int valid_int_values[] = {5, 5, 10, 10, 20, 20, 30, 30};
    int test_value[5];
    int default_int[3] = {5, 5, 30};
    IAS_PARM_ADD_INT_WITH_RANGES(parms, RANGES_ARRAY, "Valid Ranges Array",
        IAS_PARM_OPTIONAL, IAS_PARM_ARRAY,
        /* Valids */
        8, valid_int_values,
        /* Defaults */
        3, default_int,
        test_value, sizeof(test_value), 0);

    /* Parameter: String, not array, has valid list, includes default */
    const char *string_valids[8] = {"PATH_MINBOX","MAXBOX","PATH_MAXBOX",
                "GEOBOX", "PROJBOX", "ULSIZE", "MINBOX", "PATH"};
    /* string_valids[8] = NULL;*/
    char test_string[12];

    const char *default_frame_type[] = {"PATH_MINBOX"};
    IAS_PARM_ADD_STRING(parms, FRAME_TYPE, 
        "Output image frame computation method",
        IAS_PARM_REQUIRED,
        /* list of valid options */
        8, string_valids,
        /* default value */
        1, default_frame_type,
        test_string, sizeof(test_string), 0);

    char local_sensor[12];
    IAS_PARM_SENSOR(parms, IAS_PARM_REQUIRED, local_sensor,
        sizeof(local_sensor), 1);

    pid_t fork_pid;

    /* fork a child process to call the help routine with the template option
       since the redirecting of stdout cannot be done in this process without
       disabling all future output to stdout after the help routine returns */
    fork_pid = fork();
    if (fork_pid == -1)
    {
        assert_true_with_message(0, "Failed to fork child");
        return;
    }
    else if (fork_pid == 0)
    {
        /* child process, so redirect stdout and create the output */
        int status;
        FILE *loadtable_stream = NULL;

        loadtable_stream = freopen(filename1, "w", stdout);
        if (!loadtable_stream)
            exit(-1);

        /* dump out a template for the parameters */
        status = ias_parm_provide_help("--loadtable", parms, 
                  IAS_PARM_GET_TABLE_SIZE(parms), IAS_INPUT_PARAMETERS);

        fclose(loadtable_stream);

        /* destroy the cgreen resources in the child to eliminate memory leaks
           reported by valgrind */
        destroy_reporter(get_test_reporter());
        destroy_test_suite(suite);

        if (status == 1)
            exit(0);
        else
            exit(-1);
    }
    else
    {
        int status;
        pid_t returned_pid;

        /* parent process, so wait for child to finish */
        returned_pid = wait(&status);
        if (returned_pid != fork_pid)
        {
            assert_true_with_message(0, "Unexpected child returned from wait");
            return;
        }
        if (!WIFEXITED(status))
        {
            assert_true_with_message(0, "Unexpected status returned from "
                        "child");
            return;
        }
    }

    /* diff the two files to make sure they're the same */
    stream1 = fopen(filename1, "r");
    assert_not_equal(stream1, NULL);
    if (!stream1)
        return;
    stream2 = fopen(filename2, "r");
    assert_not_equal(stream2, NULL);
    if (!stream2)
        return;
    while(!feof(stream1) && !feof(stream2))
    {
        fgets(buffer1,sizeof(buffer1),stream1);
        fgets(buffer2,sizeof(buffer2),stream2);
        if(strcmp(buffer1,buffer2) != 0)
        {
            match = 0;
            break;
        }
    }
    fclose(stream1);
    fclose(stream2);
    stream1 = NULL;
    stream2 = NULL;
    assert_equal(match, 1);
}

/* Test the check range function */
Ensure test_ias_parm_check_ranges()
{
    int combine_scas;
    int test_value;
    double pccalpha;
    char work_order_id[IAS_WORKORDER_ID_SIZE];
    char test_string[6];
    int index;
    int status;

    /*------------ start parameter table definition ------------*/
    IAS_PARM_DECLARE_TABLE(parms, 11);

    /* Parameter 1:  Single integer with single range */
    int default_combine_scas[] = {1};
    
    IAS_PARM_ADD_INT(parms, COMBINE_SCAS, "Int value not array, single range",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        /* has a limited range of 0-1 */
        1, 0, 10,
        /* has one default value */
        1, default_combine_scas,
        &combine_scas, sizeof(combine_scas), 0);

    /* This value should pass the test */
    combine_scas = 7;

    /* This must be set manually.  It will generally be set in ias_parm_read */
    parms[0]->count_read = 1;

    status = ias_parm_check_ranges(
                parms[0]->value,parms[0]->valid_values,
                parms[0]->count_read, parms[0]->number_valids,
                parms[0]->type, parms[0]->is_an_array);

    /* Verify that the return value is correct */
    assert_equal(status, SUCCESS);

    /* This value should fail the test */
    combine_scas = -1;

    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = ias_parm_check_ranges(
                parms[0]->value,parms[0]->valid_values,
                parms[0]->count_read, parms[0]->number_valids,
                parms[0]->type, parms[0]->is_an_array);
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

    /* Verify that the return value is correct */
    assert_equal(status, ERROR);

    /* Parameter 2: non array double with one range */
    double default_pccalpha[] = {0.0};
    IAS_PARM_ADD_DOUBLE(parms, PCCALPHA,"Double value not array, single range",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        /* has range from -1.0 to 0.0 */
        1, -1.0, 0.0,
        /* had one default value */
        1, default_pccalpha,
        &pccalpha, sizeof(pccalpha), 0);

    /* This value should pass the test */
    pccalpha = -.5;

    /* This must be set manually.  It will generally be set in ias_parm_read */
    parms[1]->count_read = 1;

    status = ias_parm_check_ranges(
                parms[1]->value,parms[1]->valid_values,
                parms[1]->count_read, parms[1]->number_valids,
                parms[1]->type, parms[1]->is_an_array);

    /* Verify that the return value is correct */
    assert_equal(status, SUCCESS);

    /* This value should fail the test */
    pccalpha = .5;
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = ias_parm_check_ranges(
                parms[1]->value,parms[1]->valid_values,
                parms[1]->count_read, parms[1]->number_valids,
                parms[1]->type, parms[1]->is_an_array);
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

    /* Verify that the return value is correct */
    assert_equal(status, ERROR);

    /* Parameter 3: String with no valid values */
    /* Should not error in check_range */
    strcpy (work_order_id, "I9999");
    IAS_PARM_ADD_STRING(parms, WORK_ORDER_ID, "Work order ID",
        IAS_PARM_REQUIRED,
        /* doesn't have a list of valid options */
        0, NULL,
        /* doesn't have a default value */
        0, NULL,
        work_order_id, sizeof(work_order_id), 0);

    /* This must be set manually.  It will generally be set in ias_parm_read */
    parms[2]->count_read = 1;

    status = ias_parm_check_ranges(
                parms[2]->value,parms[2]->valid_values,
                parms[2]->count_read, parms[2]->number_valids,
                parms[2]->type, parms[2]->is_an_array);

    /* Verify that the return value is correct */
    assert_equal(status, SUCCESS);

    /* Parameter 4: Array of ints with one range */
    static int band_number_list[IAS_MAX_NBANDS];
    IAS_PARM_BAND_LIST(parms, IAS_PARM_REQUIRED, band_number_list, 
                sizeof(band_number_list), 1);

    /* Set band numbers that will pass the test */
    for (index = 0; index < IAS_MAX_NBANDS; index++)
        band_number_list[index] = index+1;

    /* This must be set manually.  It will generally be set in ias_parm_read */
    parms[3]->count_read = 11;

    status = ias_parm_check_ranges(
                parms[3]->value,parms[3]->valid_values,
                parms[3]->count_read, parms[3]->number_valids, 
                parms[3]->type, parms[3]->is_an_array);

    /* Verify that the return value is correct */
    assert_equal(status, SUCCESS);

    /* Set one value that will fail the test */
    band_number_list[5] = 20;

    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = ias_parm_check_ranges(
                parms[3]->value,parms[3]->valid_values,
                parms[3]->count_read, parms[3]->number_valids, 
                parms[3]->type, parms[3]->is_an_array);
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

    /* Verify that the return value is correct */
    assert_equal(status, ERROR);

    /* Parameter 5: Single string with valid list */
    const char *string_valids[5] = {"value1","value2","value3","value4"};
    string_valids[4] = NULL;

    /* add a String with a list of valid options */
    IAS_PARM_ADD_STRING(parms, TEST_STRING, "Test String",
        IAS_PARM_REQUIRED,
        /* list of valid options */
        4, string_valids,
        /* doesn't have a default value */
        0, NULL,
        test_string, sizeof(test_string), 0);

    /* Set the a string that will pass the test */
    strcpy(test_string, "value2");

    /* This must be set manually.  It will generally be set in ias_parm_read */
    parms[4]->count_read = 1;

    status = ias_parm_check_ranges(
                parms[4]->value,parms[4]->valid_values,
                parms[4]->count_read, parms[4]->number_valids,
                parms[4]->type, parms[4]->is_an_array);

    /* Verify that the return value is correct */
    assert_equal(status, SUCCESS);

    /* Set a string that will fail the test */
    strcpy(test_string, "value6");

    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = ias_parm_check_ranges(
                parms[4]->value,parms[4]->valid_values,
                parms[4]->count_read, parms[4]->number_valids,
                parms[4]->type, parms[4]->is_an_array);
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

    /* Verify that the return value is correct */
    assert_equal(status, ERROR);

    /* Parameter 6: Array of strings and array of valid values */
    const char *season_valids[] = {"Spring","Summer","Autumn","Winter"};
    char *tmp_seasons[4];

    int i;
    for (i=0; i < 4; i++)
        tmp_seasons[i] = NULL;

    IAS_PARM_ADD_STRING_ARRAY(parms, SEASONS, "Test String Array",
        IAS_PARM_OPTIONAL,
        /* list of valid options */
        4, season_valids,
        0, NULL, 
        tmp_seasons, sizeof(tmp_seasons), 0);

    /* set up string array that will pass the test */
    tmp_seasons[0] = "Summer";
    tmp_seasons[1] = "Spring";
    tmp_seasons[2] = "Autumn";

    /* Set manually.  This will normally be set in ias_parm_read */
    parms[5]->count_read = 3;

    status = ias_parm_check_ranges(
                parms[5]->value,parms[5]->valid_values,
                parms[5]->count_read, parms[5]->number_valids,
                parms[5]->type, parms[5]->is_an_array);

    /* Verify that the return value is correct */
    assert_equal(status, SUCCESS);

    /* Set string that will fail the test */
    tmp_seasons[2] = "Fall";

    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = ias_parm_check_ranges(
                parms[5]->value,parms[5]->valid_values,
                parms[5]->count_read, parms[5]->number_valids,
                parms[5]->type, parms[5]->is_an_array);
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

    /* Verify that the return value is correct */
    assert_equal(status, ERROR);

    /* Parameter 7: Single int with array of valid ranges */
    int multiple_ranges[] = {1, 1, 10, 10, 30, 30};
    IAS_PARM_ADD_INT_WITH_RANGES(parms, RANGES_ARRAY, "Valid Ranges Array",
        IAS_PARM_OPTIONAL, IAS_PARM_NOT_ARRAY,
        6, multiple_ranges,
        0, NULL,
        &test_value, sizeof(test_value), 0);

    /* Set value that will pass the test */
    test_value = 1;

    /* This must be set manually.  It will generally be set in ias_parm_read */
    parms[6]->count_read = 1;

    status = ias_parm_check_ranges(
                parms[6]->value,parms[6]->valid_values,
                parms[6]->count_read, parms[6]->number_valids,
                parms[6]->type, parms[6]->is_an_array);

    /* Verify that the return value is correct */
    assert_equal(status, SUCCESS);

    /* Set value that will fail the test */
    test_value = 51;
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = ias_parm_check_ranges(
                parms[6]->value,parms[6]->valid_values,
                parms[6]->count_read, parms[6]->number_valids,
                parms[6]->type, parms[6]->is_an_array);
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

    /* Verify that the return value is correct */
    assert_equal(status, ERROR);

   /* Parameter 8:  Array of doubles with array of valid ranges */
    double mult_double_ranges[] = {1.5, 1.5, 10.2, 10.2, 30.4, 30.4};

    /* Set the array values to that they will pass the test */
    double double_range_value[] = {1.5, 1.5, 10.2, 30.4};

    IAS_PARM_ADD_DOUBLE_WITH_RANGES(parms, RANGES_ARR_DOUBLE, 
        "Valid Double Ranges Array",
        IAS_PARM_OPTIONAL, IAS_PARM_ARRAY,
        6, mult_double_ranges,
        0, NULL,
        double_range_value, sizeof(double_range_value), 0);

    /* Manually set the count.  Will usually be set in ias_parm_read */
    parms[7]->count_read = 4;
        
    status = ias_parm_check_ranges(
                parms[7]->value,parms[7]->valid_values,
                parms[7]->count_read, parms[7]->number_valids,
                parms[7]->type, parms[7]->is_an_array);

    /* Verify that the return value is correct */
    assert_equal(status, SUCCESS);

    /* Set a value that will fail the test */
    double_range_value[2] = 10.199;

    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = ias_parm_check_ranges(
                parms[7]->value,parms[7]->valid_values,
                parms[7]->count_read, parms[7]->number_valids,
                parms[7]->type, parms[7]->is_an_array);
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

    /* Verify that the return value is correct */
    assert_equal(status, ERROR);

    /* Parameter 9:  Int array with no range at all. */
    /* Verify that it doesn't fail */

    /* Set numbers that will pass the test */
    int number_list[6];
    for (index = 0; index < 6; index++)
        number_list[index] = index+1;

    IAS_PARM_ADD_INT(parms, NUM_LIST, "Number list with no valid range",
        IAS_PARM_OPTIONAL, IAS_PARM_ARRAY, 
        0, 0, 0,
        0, NULL, 
        number_list, sizeof(number_list), 1);

    /* This must be set manually.  It will generally be set in ias_parm_read */
    parms[3]->count_read = 6;

    status = ias_parm_check_ranges(
                parms[8]->value,parms[8]->valid_values,
                parms[8]->count_read, parms[8]->number_valids, 
                parms[8]->type, parms[8]->is_an_array);

    /* Verify that the return value is correct */
    assert_equal(status, SUCCESS);

    /* Parameter 10:  WRS Path using the standard definition */
    int local_wrs_path = 2000;
    IAS_PARM_WRS_PATH(parms, &local_wrs_path, sizeof(local_wrs_path), 1);
    parms[9]->count_read = 1;

    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = ias_parm_check_ranges(
                parms[9]->value,parms[9]->valid_values,
                parms[9]->count_read, parms[9]->number_valids, 
                parms[9]->type, parms[9]->is_an_array);
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

    assert_equal(status, ERROR);

    /* Parameter 11: Boolean values */
    int local_bool = 0;
    IAS_PARM_ADD_BOOLEAN(parms, TEST_BOOLEAN, "Boolean value testing", 1,
                         0, NULL, &local_bool, sizeof(local_bool), 1);
    parms[10]->count_read = 1;

    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = ias_parm_check_ranges(
                parms[10]->value, parms[10]->valid_values,
                parms[10]->count_read, parms[10]->number_valids,
                parms[10]->type, parms[10]->is_an_array);
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

    assert_equal(status, SUCCESS);

    /* Test an invalid boolean value */
    parms[10]->value.type_int[0] = 2;

    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = ias_parm_check_ranges(
                parms[10]->value, parms[10]->valid_values,
                parms[10]->count_read, parms[10]->number_valids,
                parms[10]->type, parms[10]->is_an_array);
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

    assert_equal(status, ERROR);
    
}

/* Test that a parameter value that is longer than the space provided errors
   like it should for an optional parameter */
Ensure test_invalid_size()
{
    int status;

    /* declare a parameter table with a parameter buffer that is too small to
       hold the value in the file */
    char work_order_id[3];
    IAS_PARM_DECLARE_TABLE(parms, 1);

    IAS_PARM_ADD_STRING(parms, WORK_ORDER_ID, "Work order ID",
        IAS_PARM_OPTIONAL, 
        /* does not have limits */
        0, NULL, 
        /* default */
        1, NULL,
        work_order_id, sizeof(work_order_id), 0);

    /* read the parameter - it should fail */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = ias_parm_read("test.odl", parms, 1);
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);
    assert_equal(status, ERROR);
}

int main(int argc, char **argv)
{
    suite = create_test_suite();
    add_test(suite, test_ias_parm_read);
    add_test(suite, test_ias_parm_update);
    add_test(suite, test_ias_parm_help);
    add_test(suite, test_ias_parm_template);
    add_test(suite, test_ias_parm_default_values);
    add_test(suite, test_ias_parm_loadtable);
    add_test(suite, test_ias_parm_check_ranges);
    add_test(suite, test_check_parameter_name_length);
    add_test(suite, test_invalid_size);
    if (argc > 1) {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}
