#ifndef IAS_RLUT_H
#define IAS_RLUT_H

#include "ias_types.h"

/* Dimensions for file attribute strings (from CPF DFCB).  These DO NOT
   include NULL termination -- that should be accounted for when strings
   with these dimensions are declared */
#define IAS_RLUT_DATE_LENGTH            26    /* Size of date string */
#define IAS_RLUT_STATUS_LENGTH          12    /* Size of status string */
#define IAS_RLUT_FILE_SOURCE_LENGTH    200    /* Size of file source string */
#define IAS_RLUT_DESCRIPTION_LENGTH   4000    /* Size of description string */

/* Main file attributes data structure declaration.  The +1 is for NULL
   termination.  Dates are assumed to be in the format
   'YYYY[-]MM[-]DD[Tt ]HH24[:]MI[:]SS'.*/
typedef struct ias_rlut_file_attributes
{
    char file_source[IAS_RLUT_FILE_SOURCE_LENGTH + 1];  /* Filename of baseline
                                                         RLUT used as source */
    char effective_date_begin[IAS_RLUT_DATE_LENGTH + 1];/* Effective start date
                                                         for this RLUT */
    char effective_date_end[IAS_RLUT_DATE_LENGTH + 1];  /* Effective stop date
                                                         for this RLUT */
    char status[IAS_RLUT_STATUS_LENGTH + 1];            /* ACTIVE, INACTIVE,
                                                         UNTESTED, TESTED,
                                                         VALIDATED, DENIED */
    char baseline_date[IAS_RLUT_DATE_LENGTH + 1];       /* Date the RLUT was
                                                         entered into baseline
                                                         production */
    char description[IAS_RLUT_DESCRIPTION_LENGTH + 1];  /* Text field intended
                                                         to describe rationale
                                                         for updated RLUT */
    int  version;                                       /* Version number of
                                                         effective date 
                                                         range */
    int  collection;                                   /* Collection number of
                                                         effective date 
                                                         range */
} IAS_RLUT_FILE_ATTRIBUTES;


/* Main linearization parameter data structure definition */
typedef struct ias_rlut_linearization_params
{
    double cutoff_threshold_low;    /* Low DN range cutoff */
    double cutoff_threshold_high;   /* High DN range cutoff */
    double remap_coeff0_low;        /* Low range DN remapping coeffs */
    double remap_coeff1_low;
    double remap_coeff2_low;
    double remap_coeff0_mid;        /* Mid range DN remapping coeffs */
    double remap_coeff1_mid;
    double remap_coeff2_mid;
    double remap_coeff0_high;       /* High range DN remapping coeffs */
    double remap_coeff1_high;
    double remap_coeff2_high;
} IAS_RLUT_LINEARIZATION_PARAMS;

/*------------------------------------------------------------------------------
 * This structure contains two 2-dimensional arrays of floating point values.
 * These array contain a single row for each of the detectors in an SCA,
 * 'dn[detector]' and 'correction[detector]'.  An application will read one of
 * these structures from the RLUT file that contains DN values which are used
 * as a lookup table.  When a DN value from a detector is located within a row,
 * the index location is used in the parallel array 'correction' to find either
 * the linearity correction that should be applied to the DN value received
 * from the spacecraft.
 *----------------------------------------------------------------------------*/
typedef struct ias_rlut_lookup_table
{
    float *input_dn;      /* Table of input DN lookup values.  Access using
                             dn[detector * num_values + index] to get the
                             input value */
    float *output_correction;/* Table of linearity or nonuniformity corrections.
                             Access using correction[detector * num_values
                              + index] to get the output value. */
    int band_num;         /* Band this structure contains data for */
    int sca_num;          /* SCA this structure contains data for */
    int num_detectors;    /* Num detectors per SCA, (1st index) */
    int num_values;       /* Num distinct DN values used, (2nd index) */
} IAS_RLUT_LOOKUP_TABLE;

/* Forward reference to RLUT IO data structure */
typedef struct IAS_RLUT_IO IAS_RLUT_IO;

/* Define the lookup table types that can be in the file.  This is used to
   select the correct group name for the lookup table to read/write. */
typedef enum 
{
    IAS_RLUT_LINEARITY_LOOKUP,
    IAS_RLUT_TIRS_SECONDARY_LOOKUP
} IAS_RLUT_LOOKUP_TYPE;

/*                          FUNCTION PROTOTYPES                 */
IAS_RLUT_IO *ias_rlut_open_file
(
    const char *rlut_filename,            /* I: Name of RLUT file */
    IAS_ACCESS_MODE access_mode           /* I: Requested access mode */
);


int ias_rlut_close_file
(
    IAS_RLUT_IO *rlut_file                /* I: RLUT IO structure */
);


int ias_rlut_write_file_attributes
(
    const IAS_RLUT_IO *rlut_file,             /* I: Open RLUT file */
    const IAS_RLUT_FILE_ATTRIBUTES *file_attr /* I: Attributes to write */
);


int ias_rlut_read_file_attributes
(
    const IAS_RLUT_IO *rlut_file,         /* I: Open HDF5 RLUT file */
    IAS_RLUT_FILE_ATTRIBUTES *file_attr   /* O: Attributes read from file */
);


int ias_rlut_write_linearization_params
(
    const IAS_RLUT_IO *rlut_file,    /* I: Open RLUT file */
    const IAS_RLUT_LINEARIZATION_PARAMS *linearization_params,
                                     /* I: Array of detector-specific
                                        linearization parameters for the
                                        current band/SCA */
    int band_number,                 /* I: Current index into RLUT
                                        band array */
    int sca_number,                  /* I: Current SCA number */
    int num_detectors                /* I: Number of detectors in the
                                        current band/SCA */
);


IAS_RLUT_LINEARIZATION_PARAMS *ias_rlut_read_linearization_params
(
    const IAS_RLUT_IO *rlut_file,    /* I: Open RLUT file */
    int band_number,                 /* I: Current band number */
    int sca_number,                  /* I: Current SCA number */
    int num_detectors                /* I: Number of detectors in the
                                        current band/SCA */
);


int ias_rlut_is_rlut_file
(
    const char *rlut_filename        /* I: Name of (potential) RLUT file */
);

IAS_RLUT_LOOKUP_TABLE *ias_rlut_alloc_lookup_table
(
    int band_num,            /* I: Band number of interest */
    int sca_num,             /* I: SCA number of interest */
    int detector_count,      /* I: Number of detectors to allocate */
    int num_values           /* I: Number of values per detector */
);

void ias_rlut_free_lookup_table
(
    IAS_RLUT_LOOKUP_TABLE *table /* I/O: An RLUT table */
);

IAS_RLUT_LOOKUP_TABLE *ias_rlut_read_lookup_table
(
    const IAS_RLUT_IO *rlut,    /* I/O: File descriptor opened for writing */
    IAS_RLUT_LOOKUP_TYPE type,  /* I: Type of lookup table to read */
    int band_num,               /* I: Band number of interest */
    int sca_num                 /* I: SCA number of interest */
);

int ias_rlut_write_lookup_table
(
    const IAS_RLUT_IO *rlut,    /* I/O: File descriptor opened for writing */
    IAS_RLUT_LOOKUP_TYPE type,  /* I: Type of lookup table to write */
    IAS_RLUT_LOOKUP_TABLE *table/* I: RLUT table */
);


int ias_rlut_verify_linearity_group
(
    const IAS_RLUT_IO *rlut,        /* I: Used to get the file id to check */
    int band_number,                /* I: Band number of group to check*/
    int sca_number                  /* I: SCA number of group to check */
);

int ias_rlut_verify_linearization_group
(
    const IAS_RLUT_IO *rlut,        /* I: Used to get the file id to check */
    int band_number,                /* I: Band number of group to check*/
    int sca_number                  /* I: SCA number of group to check */
);

int ias_rlut_verify_tirs_secondary_group
(
    const IAS_RLUT_IO *rlut,        /* I: Used to get the file id to check */
    int band_number,                /* I: Band number of group to check*/
    int sca_number                  /* I: SCA number of group to check */
);


#endif
