/*==============================================================================
NAME: ias_rlut_read_lookup_table

PURPOSE:
    Provide the RLUT IO support for reading a response linearization lookup
    table.

==============================================================================*/

/* System Includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* COTS Includes */
#include "hdf5.h"

/* IAS Includes */
#include "ias_const.h"
#include "ias_types.h"
#include "ias_logging.h"
#include "ias_rlut.h"
#include "ias_satellite_attributes.h"

/* Local Includes */
#include "ias_rlut_private.h"

/*------------------------------------------------------------------------------
NAME: ias_rlut_read_lookup_table

PURPOSE:
    This function will read either the set of linearity correction values from
    an RLUT file for a specified band and SCA and return those values in an
    IAS_RLUT_LOOKUP_TABLE structure.

    This function will call ias_rlut_alloc_lookup_table() to allocate a 'table'
    sized appropriately for the requested data that is contained within the
    RLUT file.  The 'table' parameter is assumed to be a NULL pointer upon the
    first call to this function.  The calling application is responsible for
    calling ias_rlut_free_lookup_table() when processing is complete to free
    the dynamically allocated memory pointed to by 'table' when finished.

RETURNS:
    A pointer to the lookup table is returned if successful
    NULL is returned if an error occurs
------------------------------------------------------------------------------*/
IAS_RLUT_LOOKUP_TABLE *ias_rlut_read_lookup_table
(
    const IAS_RLUT_IO *rlut, /* I/O: File descriptor opened for writing */
    IAS_RLUT_LOOKUP_TYPE type,/* I: Type of lookup table to read */
    int band_num,            /* I:   Band number of interest */
    int sca_num              /* I:   SCA number of interest */
)
{
    char group_name[PATH_MAX];
    herr_t hdf_status;                /* HDF5 status */
    hid_t hdf_dataset_id;              /* HDF5 Data Set ID */
    hid_t hdf_dataspace_id;
    hid_t group_id;
    hid_t data_type;
    hsize_t dims[2];
    IAS_RLUT_LOOKUP_TABLE *table;
    int detector_count;
    int sca_count;
    const char *base_group;

    /* Set the group name */
    if (type == IAS_RLUT_LINEARITY_LOOKUP)
        base_group = LINEARITY_LOOKUP_GROUP_NAME;
    else if (type == IAS_RLUT_TIRS_SECONDARY_LOOKUP)
        base_group = TIRS_SECONDARY_LINEARIZATION_GROUP_NAME;
    else
    {
        IAS_LOG_ERROR("Unsupported lookup table type: %d", type);
        return NULL;
    }

    /* Ensure the file is open */
    if ((rlut == NULL) || (rlut->file_id < 0))
    {
        IAS_LOG_ERROR("RLUT file is not open");
        return NULL;
    }

    /* Verify the SCA requested is valid */
    sca_count = ias_sat_attr_get_scas_per_band(band_num);
    if (sca_count == ERROR)
    {
        IAS_LOG_ERROR("Unable to retreive the SCA count for band number %d",
            band_num);
        return NULL;
    }
    if (sca_num < 1 || sca_num > sca_count)
    {
        IAS_LOG_ERROR("Invalid SCA number (%d) provided for band number %d",
            sca_num, band_num);
        return NULL;
    }

    /* Get the number of detectors for this band */
    detector_count = ias_sat_attr_get_detectors_per_sca(band_num);
    if (detector_count == ERROR)
    {
        IAS_LOG_ERROR("Unable to retreive the detector count for band number "
            "%d", band_num);
        return NULL;
    }

    /* Verify the dataset exists and open the group containing it */
    group_id = ias_rlut_get_group(rlut, base_group, band_num, sca_num,
        VERIFY_PRESENT, LOOKUP_NAME);
    if (group_id < 0)
    {
        IAS_LOG_ERROR("Opening group %s for band number %d, sca number %d "
            "in file %s", base_group, band_num, sca_num, rlut->filename);
        return NULL;
    }

    /* Just check the full second dataset since the rest of the path has
       already been verified */
    if (H5Lexists(group_id, CORRECTION_NAME, H5P_DEFAULT) != TRUE)
    {
        IAS_LOG_ERROR("Dataset %s does not exist in file %s", CORRECTION_NAME,
            rlut->filename);
        H5Gclose(group_id);
        return NULL;
    }

    /*--------------------------------------------------------------------------
     * Read the DN Lookup Values
     *------------------------------------------------------------------------*/

    /* Open the data set for reading */
    hdf_dataset_id = H5Dopen(group_id, LOOKUP_NAME, H5P_DEFAULT);
    if (hdf_dataset_id < 0)
    {
        IAS_LOG_ERROR("Opening data set %s", LOOKUP_NAME);
        H5Gclose(group_id);
        return NULL;
    }

    /* Confirm the data type is correct.  If it is not, issue a warning. */
    data_type = H5Dget_type(hdf_dataset_id);
    if (data_type < 0)
    {
        IAS_LOG_ERROR("Getting data type of dataset %s", LOOKUP_NAME);
        H5Dclose(hdf_dataset_id);
        H5Gclose(group_id);
        return NULL;
    }
    if (!H5Tequal(data_type, H5T_IEEE_F32LE))
    {
        IAS_LOG_ERROR("Unexpected data type for group %s, band number %d, "
            "SCA %d, dataset %s", base_group, band_num, sca_num, LOOKUP_NAME);
        H5Tclose(data_type);
        H5Dclose(hdf_dataset_id);
        H5Gclose(group_id);
        return NULL;
    }
    if (H5Tclose(data_type) < 0)
    {
        IAS_LOG_ERROR("Closing data type of dataset %s", LOOKUP_NAME);
        H5Dclose(hdf_dataset_id);
        H5Gclose(group_id);
        return NULL;
    }


    /* Get the data space for the data set */
    hdf_dataspace_id = H5Dget_space(hdf_dataset_id);
    if (hdf_dataspace_id < 0)
    {
        IAS_LOG_ERROR("Obtaining data space for data set: %s", LOOKUP_NAME);
        H5Dclose(hdf_dataset_id);
        H5Gclose(group_id);
        return NULL;
    }

    /* Obtain the dimensions of the data set */
    if (H5Sget_simple_extent_ndims(hdf_dataspace_id) != 2)
    {
        IAS_LOG_ERROR("Data set is not 2 dimensional: %s", LOOKUP_NAME);
        H5Sclose(hdf_dataspace_id);
        H5Dclose(hdf_dataset_id);
        H5Gclose(group_id);
        return NULL;
    }
    if (H5Sget_simple_extent_dims(hdf_dataspace_id, dims, NULL) < 0)
    {
        IAS_LOG_ERROR("Obtaining dimensions of data set: %s", LOOKUP_NAME);
        H5Sclose(hdf_dataspace_id);
        H5Dclose(hdf_dataset_id);
        H5Gclose(group_id);
        return NULL;
    }

    /* Make sure the number of detectors in the file matches the expected
       number */
    if (dims[0] != detector_count)
    {
        IAS_LOG_ERROR("Detector count read from file is %d when %d is expected"
            " for band number %d", (int)dims[0], detector_count, band_num);
        H5Sclose(hdf_dataspace_id);
        H5Dclose(hdf_dataset_id);
        H5Gclose(group_id);
        return NULL;
    }

    /* Allocate the IAS_RLUT_LOOKUP_TABLE to the dimensions of the tables */
    table = ias_rlut_alloc_lookup_table(band_num, sca_num, dims[0], dims[1]);
    if (table == NULL)
    {
        IAS_LOG_ERROR("Allocating memory for RLUT lookup table");
        H5Sclose(hdf_dataspace_id);
        H5Dclose(hdf_dataset_id);
        H5Gclose(group_id);
        return NULL;
    }

    /* Read the DN lookup table */
    hdf_status = H5Dread(hdf_dataset_id, H5T_NATIVE_FLOAT, H5S_ALL, H5S_ALL,
        H5P_DEFAULT, table->input_dn);
    if (hdf_status < 0)
    {
        IAS_LOG_ERROR("Reading DN lookup values for: %s", LOOKUP_NAME);
        H5Sclose(hdf_dataspace_id);
        H5Dclose(hdf_dataset_id);
        H5Gclose(group_id);
        ias_rlut_free_lookup_table(table);
        return NULL;
    }

    /* Close the data space for the DN lookup data set */
    hdf_status = H5Sclose(hdf_dataspace_id);
    if (hdf_status < 0)
    {
        IAS_LOG_ERROR("Closing data space for: %s", LOOKUP_NAME);
        H5Dclose(hdf_dataset_id);
        H5Gclose(group_id);
        ias_rlut_free_lookup_table(table);
        return NULL;
    }

    /* Close the data set for the DN lookup values */
    hdf_status = H5Dclose(hdf_dataset_id);
    if (hdf_status < 0)
    {
        IAS_LOG_ERROR("Closing data set: %s", LOOKUP_NAME);
        H5Gclose(group_id);
        ias_rlut_free_lookup_table(table);
        return NULL;
    }

    /*--------------------------------------------------------------------------
     * Read the Correction Values
     *------------------------------------------------------------------------*/

    /* Open the data set for reading */
    hdf_dataset_id = H5Dopen(group_id, CORRECTION_NAME, H5P_DEFAULT);
    if (hdf_dataset_id < 0)
    {
        IAS_LOG_ERROR("Opening data set %s", CORRECTION_NAME);
        H5Gclose(group_id);
        ias_rlut_free_lookup_table(table);
        return NULL;
    }

    /* Confirm the data type is correct.  If it is not, issue a warning. */
    data_type = H5Dget_type(hdf_dataset_id);
    if (data_type < 0)
    {
        IAS_LOG_ERROR("Getting data type of dataset %s", CORRECTION_NAME);
        H5Dclose(hdf_dataset_id);
        H5Gclose(group_id);
        ias_rlut_free_lookup_table(table);
        return NULL;
    }
    if (!H5Tequal(data_type, H5T_IEEE_F32LE))
    {
        IAS_LOG_WARNING("Unexpected data type for group %s, band number %d, "
            "SCA %d, dataset %s", base_group, band_num, sca_num, 
            CORRECTION_NAME);
    }
    if (H5Tclose(data_type) < 0)
    {
        IAS_LOG_ERROR("Closing data type of dataset %s", CORRECTION_NAME);
        H5Dclose(hdf_dataset_id);
        H5Gclose(group_id);
        ias_rlut_free_lookup_table(table);
        return NULL;
    }

    /* Get the data space for the data set */
    hdf_dataspace_id = H5Dget_space(hdf_dataset_id);
    if (hdf_dataspace_id < 0)
    {
        IAS_LOG_ERROR("Obtaining data space for data set: %s", CORRECTION_NAME);
        H5Dclose(hdf_dataset_id);
        H5Gclose(group_id);
        ias_rlut_free_lookup_table(table);
        return NULL;
    }

    /* Obtain the dimensions of the data set */
    if (H5Sget_simple_extent_ndims(hdf_dataspace_id) != 2)
    {
        IAS_LOG_ERROR("Data set is not 2 dimensional: %s", CORRECTION_NAME);
        H5Sclose(hdf_dataspace_id);
        H5Dclose(hdf_dataset_id);
        H5Gclose(group_id);
        ias_rlut_free_lookup_table(table);
        return NULL;
    }
    if (H5Sget_simple_extent_dims(hdf_dataspace_id, dims, NULL) < 0)
    {
        IAS_LOG_ERROR("Obtaining dimensions of data set: %s", CORRECTION_NAME);
        H5Sclose(hdf_dataspace_id);
        H5Dclose(hdf_dataset_id);
        H5Gclose(group_id);
        ias_rlut_free_lookup_table(table);
        return NULL;
    }

    /* Verify that the dimensions of the correction data set are the same as the
       DN lookup values.  If these values are not the same, the RLUT file has
       an error in it. */
    if ((dims[0] != table->num_detectors) ||
        (dims[1] != table->num_values))
    {
        IAS_LOG_ERROR("Dimension mismatch for data set %s:\n"
            "    DN lookup values:  %d x %d\n"
            "    Correction values: %Lu x %Lu\n",
            group_name, table->num_detectors, table->num_values, dims[0],
            dims[1]);
        H5Sclose(hdf_dataspace_id);
        H5Dclose(hdf_dataset_id);
        H5Gclose(group_id);
        ias_rlut_free_lookup_table(table);
        return NULL;
    }

    /* Read the correction lookup table */
    hdf_status = H5Dread(hdf_dataset_id, H5T_NATIVE_FLOAT, H5S_ALL, H5S_ALL,
        H5P_DEFAULT, table->output_correction);
    if (hdf_status < 0)
    {
        IAS_LOG_ERROR("Reading output correction values for: %s",
            CORRECTION_NAME);
        H5Sclose(hdf_dataspace_id);
        H5Dclose(hdf_dataset_id);
        H5Gclose(group_id);
        ias_rlut_free_lookup_table(table);
        return NULL;
    }

    /* Close the data space for the correction lookup data set */
    hdf_status = H5Sclose(hdf_dataspace_id);
    if (hdf_status < 0)
    {
        IAS_LOG_ERROR("Closing data space for: %s", CORRECTION_NAME);
        H5Dclose(hdf_dataset_id);
        H5Gclose(group_id);
        ias_rlut_free_lookup_table(table);
        return NULL;
    }

    /* Close the data set for the correction lookup values */
    hdf_status = H5Dclose(hdf_dataset_id);
    if (hdf_status < 0)
    {
        IAS_LOG_ERROR("Closing data set: %s", CORRECTION_NAME);
        H5Gclose(group_id);
        ias_rlut_free_lookup_table(table);
        return NULL;
    }

    /* Close the group */
    if (H5Gclose(group_id) < 0)
    {
        IAS_LOG_ERROR("Closing group for band number %d, sca number %d",
            band_num, sca_num);
        ias_rlut_free_lookup_table(table);
        return NULL;
    }

    return table;
}
