/*==============================================================================
NAME: ias_rlut_utilities.c

PURPOSE:
    Provide utilities of use to other parts of the RLUT IO library.
    LookUp Table.

==============================================================================*/

/* System Includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* COTS Includes */
#include "hdf5.h"

/* IAS Includes */
#include "ias_const.h"
#include "ias_types.h"
#include "ias_logging.h"
#include "ias_rlut.h"

/* Local Includes */
#include "ias_rlut_private.h"

/*------------------------------------------------------------------------------
NAME: ias_rlut_alloc_lookup_table

PURPOSE:
    This function allocates a new IAS_RLUT_LOOKUP_TABLE.

RETURNS:
    Pointer to the allocated structure or NULL if there is an error.

------------------------------------------------------------------------------*/
IAS_RLUT_LOOKUP_TABLE *ias_rlut_alloc_lookup_table
(
    int band_number,         /* I: Band number of interest */
    int sca_number,          /* I: SCA number of interest */
    int detector_count,      /* I: Number of detectors to allocate */
    int num_values           /* I: Number of values per detector */
)
{
    size_t bytes;
    IAS_RLUT_LOOKUP_TABLE *table;

    /* Allocate memory for the structure */
    table = malloc(sizeof(*table));
    if (table == NULL)
    {
        IAS_LOG_ERROR("Allocating an RLUT lookup table");
        return NULL;
    }

    /* Initialize the new table */
    table->input_dn = NULL;
    table->output_correction = NULL;
    table->num_detectors = detector_count;
    table->num_values = num_values;

    /* Copy the band number and SCA number to the structure */
    table->band_num = band_number;
    table->sca_num = sca_number;

    /* Allocate memory for the two arrays */
    bytes = detector_count * num_values * sizeof(float);
    table->input_dn = malloc(bytes);
    if (table->input_dn == NULL)
    {
        long int my_size = (long int)bytes;
        IAS_LOG_ERROR("Allocating %ld bytes of memory for DN lookup", my_size);
        free(table);
        return NULL;
    }
    table->output_correction = malloc(bytes);
    if (table->output_correction == NULL)
    {
        long int my_size = (long int)bytes;
        IAS_LOG_ERROR("Allocating %ld bytes of memory for output correction",
            my_size);
        free(table->input_dn);
        free(table);
        return NULL;
    }

    return table;
}

/*------------------------------------------------------------------------------
NAME: ias_rlut_free_lookup_table

PURPOSE:
    This function will free up all of the memory associated with an RLUT lookup
    table.

    The specified table should have been allocated with a call to
    ias_rlut_alloc_lookup_table().

RETURNS:
    N/A
------------------------------------------------------------------------------*/
void ias_rlut_free_lookup_table
(
    IAS_RLUT_LOOKUP_TABLE *table /* I/O: An allocated lookup table */
)
{
    if (table == NULL)
        return;
    free(table->input_dn);
    table->input_dn = NULL;
    free(table->output_correction);
    table->output_correction = NULL;
    free(table);
}

/*------------------------------------------------------------------------------
NAME: ias_rlut_verify_object_exists

PURPOSE:
    This function verifies that all the objects from a path exist in an HDF
    file.

RETURNS:
    SUCCESS if all the objects exist
    ERROR if any object is missing
------------------------------------------------------------------------------*/
int ias_rlut_verify_object_exists
(
    const IAS_RLUT_IO *rlut,/* I: File to check for object existence */
    const char *path        /* I: Path to the object to check */
)
{
    int len = strlen(path);
    char partial_path[len + 1];
    char *curr;

    /* Verify the object starts at the root */
    if (path[0] != '/')
    {
        IAS_LOG_ERROR("HDF object path '%s' does not start at root", path);
        return ERROR;
    }

    /* Copy the input path to one that can be modified */
    strcpy(partial_path, path);

    /* Check each object in turn since if any objects along the way are
       missing, HDF5 will throw a bunch of error messages */
    curr = &partial_path[1];
    while (curr != NULL)
    {
        curr = strchr(curr, '/');

        /* Temporarily end the path at the current '/', if one was found */
        if (curr)
            *curr = '\0';

        /* Confirm the path exists */
        if (H5Lexists(rlut->file_id, partial_path, H5P_DEFAULT) != TRUE)
        {
            IAS_LOG_ERROR("HDF object '%s' does not exist", partial_path);
            return ERROR;
        }

        /* Restore the '/' and advance past it */
        if (curr)
        {
            *curr = '/';
            curr++;
        }
    }

    return SUCCESS;
}

/*------------------------------------------------------------------------------
NAME: ias_rlut_create_path

PURPOSE:
    This function verifies all the groups exist in the requested path and 
    creates them if they don't exist.

RETURNS:
    SUCCESS if the requested path already existed or was able to be created
    ERROR if the path couldn't be created
------------------------------------------------------------------------------*/
int ias_rlut_create_path
(
    const IAS_RLUT_IO *rlut,/* I: File to check for object existence */
    const char *path        /* I: Path to the object to check */
)
{
    int len = strlen(path);
    char partial_path[len + 1];
    char *curr;

    /* Verify the object starts at the root */
    if (path[0] != '/')
    {
        IAS_LOG_ERROR("HDF object path '%s' does not start at root", path);
        return ERROR;
    }

    /* Copy the input path to one that can be modified */
    strcpy(partial_path, path);

    /* Check each object in turn since if any objects along the way are
       missing, HDF5 will throw a bunch of error messages */
    curr = &partial_path[1];
    while (curr != NULL)
    {
        curr = strchr(curr, '/');

        /* Temporarily end the path at the current '/', if one was found */
        if (curr)
            *curr = '\0';

        /* Check whether the path exists */
        if (H5Lexists(rlut->file_id, partial_path, H5P_DEFAULT) != TRUE)
        {
            /* It doesn't exist, so create it */
            hid_t group_id;

            group_id = H5Gcreate(rlut->file_id, partial_path, H5P_DEFAULT,
                H5P_DEFAULT, H5P_DEFAULT );
            if (group_id < 0)
            {
                IAS_LOG_ERROR("Creating RLUT group: %s", partial_path);
                return ERROR;
            }

            /* Close the group since it isn't directly needed */
            if (H5Gclose(group_id) < 0)
            {
                IAS_LOG_ERROR("Closing RLUT group: %s", partial_path);
                return ERROR;
            }
        }

        /* Restore the '/' and advance past it */
        if (curr)
        {
            *curr = '/';
            curr++;
        }
    }

    return SUCCESS;
}

/*------------------------------------------------------------------------------
NAME: ias_rlut_get_group

PURPOSE:
    Given a group base name, band number and sca number, this routine will open
    the group named "/base_name/Band##/SCA##".  It can create missing groups or
    verify all the groups are there.  If a dataset name is provided and it is
    verifying the groups are there, it will also verify the dataset is present.

RETURNS:
    The opened group id, or -1 if an error
------------------------------------------------------------------------------*/
hid_t ias_rlut_get_group
(
    const IAS_RLUT_IO *rlut,/* I: The file to create the group in */
    const char *base_name,  /* I: Base name for the group */
    int band_number,        /* I: Band number to create */
    int sca_number,         /* I: SCA number to create */
    int create_missing_groups,/* I: flag to create missing groups */
    const char *dataset     /* I: optional dataset to verify exists (set to
                                  NULL if nothing to check) */
)
{
    int status;
    char group_name[PATH_MAX];
    hid_t group_id;

    /* Construct the group name for the current band/SCA */
    status = snprintf(group_name, sizeof(group_name), "/%s/%s%02d/%s%02d",
        base_name, BAND_NAME, band_number, SCA_NAME, sca_number);
    if ((status < 0) || (status >= sizeof(group_name)))
    {
        IAS_LOG_ERROR("Creating group name for band %d SCA %d "
            "for the %s group", band_number, sca_number, group_name);
        return -1;
    }

    if (create_missing_groups)
    {
        /* Create the groups needed if they don't exist */
        if (ias_rlut_create_path(rlut, group_name) != SUCCESS)
        {
            IAS_LOG_ERROR("Creating group %s in file %s", group_name,
                rlut->filename);
            return -1;
        }
    }
    else
    {
        /* Not creating the groups, so make sure they all exist */
        if (ias_rlut_verify_object_exists(rlut, group_name) != SUCCESS)
        {
            IAS_LOG_ERROR("Group %s does not exist in file %s", group_name, 
                rlut->filename);
            return -1;
        }
    }

    /* Open the group for the current band/SCA */
    group_id = H5Gopen(rlut->file_id, group_name, H5P_DEFAULT);
    if (group_id < 0)
    {
        IAS_LOG_ERROR("Opening group %s in file %s", group_name,
            rlut->filename);
        return -1;
    }

    if (!create_missing_groups && dataset != NULL)
    {
        if (H5Lexists(group_id, dataset, H5P_DEFAULT) != TRUE)
        {
            IAS_LOG_ERROR("Dataset %s does not exist in group %s",
                dataset, group_name);
            H5Gclose(group_id);
            return -1;
        }
    }

    return group_id;
}

/*------------------------------------------------------------------------------
NAME: verify_group_and_datatype

PURPOSE:
    Given a parent group, band number and sca number, this routine will open
    the group named "/parent_group/Band##/SCA##".  If it exists and the datatype
    is the correct type it will Return SUCCESS. If it does not exist or datatype
    is incorrect or another error occurs it will return ERROR.

RETURNS:
    SUCCESS if group exists and datatype is correct
    ERROR if an error occurs,if group does not exist, or a datatype is incorrect
------------------------------------------------------------------------------*/
static int verify_group_and_datatype
(
    const IAS_RLUT_IO *rlut,        /* I: Used to get the file id to check */
    const char *parent_group,       /* I: Lookup parent_group to check */
    int band_number,                /* I: Band number of group to check*/
    int sca_number                  /* I: SCA number of group to check */
)
{
    int status;                     /* Status holder for errors */
    char group_name[PATH_MAX];      /* Buffer to hold lookup path */
    hid_t group_id;                 /* Group id to check datatype */
    char *curr;                     /* Point to path for iteration */
    int lookup_exists;              /* Holds boolean for lookup dataset*/
    int correction_exists;          /* Holds boolean for correction dataset */  

    /* Construct the group name for the current band/SCA */
    status = snprintf(group_name, sizeof(group_name), "/%s/%s%02d/%s%02d",
        parent_group, BAND_NAME, band_number, SCA_NAME, sca_number);
    
    /* Check that group_name was formatted correctly */
    if ((status < 0) || (status >= sizeof(group_name)))
    {
        IAS_LOG_ERROR("Could not format group name %s", group_name);
        return ERROR;
    }

    /* Iterate through the path by HDF5 standards */
    curr = &group_name[1];
    while (curr != NULL)
    {
        curr = strchr(curr, '/');

        /* Temporarily end the path at the current '/', if one was found */
        if (curr)
        {
            *curr = '\0';
        }

        /* Confirm the path exists */
        if (H5Lexists(rlut->file_id, group_name, H5P_DEFAULT) != TRUE)
        {
            IAS_LOG_ERROR("Group does not exist for %s", group_name);
            return ERROR;
        }

        /* Restore the '/' and advance past it */
        if (curr)
        {
            *curr = '/';
            curr++;
        }
    }
    
    /* Linearization parameters doesn't contain datasets so end here */
    if (strcmp(parent_group, LINEARIZATION_PARAMS_GROUP_NAME) == 0)
    {
        return SUCCESS;
    }

    /* Open the group for the current band/SCA */
    group_id = H5Gopen(rlut->file_id, group_name, H5P_DEFAULT);
    if (group_id < 0)
    {
        IAS_LOG_ERROR("Group does not exist for %s", group_name);
        return ERROR;
    }

    /* Check after opening that dataset LOOKUP_NAME exist */
    lookup_exists = H5Lexists(group_id, LOOKUP_NAME, H5P_DEFAULT);
    correction_exists = H5Lexists(group_id, CORRECTION_NAME, H5P_DEFAULT);

    /* Check the LOOKUP_NAME dataset */
    if (lookup_exists == TRUE)
    {
        hid_t lookup_datatype; /* Datatype of lookup dataset */

        /* Open LOOKUP_NAME  */
        hid_t hdf_lookup_id = H5Dopen(group_id, LOOKUP_NAME, H5P_DEFAULT);
        if (hdf_lookup_id < 0)
        {
            IAS_LOG_ERROR("Opening data set %s", LOOKUP_NAME);
            H5Gclose(group_id);
            return ERROR;
        }

        /* Get datatype for LOOKUP_NAME */
        lookup_datatype = H5Dget_type(hdf_lookup_id);
        if (lookup_datatype < 0)
        {
            IAS_LOG_ERROR("Getting data type of dataset %s", LOOKUP_NAME);
            H5Dclose(hdf_lookup_id);
            H5Gclose(group_id);
            return ERROR;
        }

        /* Close the lookup dataset since it isn't needed anymore */
        if (H5Dclose(hdf_lookup_id) < 0 )
        {
            IAS_LOG_ERROR("Could not close H5D for lookup");
            H5Tclose(lookup_datatype);
            H5Gclose(group_id);
            return ERROR;
        }

        /* Check to see if the datatype is correct for lookup */
        if (!H5Tequal(lookup_datatype, H5T_IEEE_F32LE))
        {
            IAS_LOG_ERROR("Unexpected data type for group %s, dataset %s",
               group_name, LOOKUP_NAME);
            H5Tclose(lookup_datatype);
            H5Gclose(group_id);
            return ERROR;
        }

        /* Close the lookup_datatype */
        if (H5Tclose(lookup_datatype) < 0)
        {
            IAS_LOG_ERROR("Could not close H5T");
            H5Gclose(group_id);
            return ERROR;
        }
    }
    else
    {
        IAS_LOG_ERROR("Dataset %s does not exist for %s", LOOKUP_NAME,
            group_name);
        H5Gclose(group_id);
        return ERROR;
    }
        
    /* Check the CORRECTION_NAME dataset */
    if (correction_exists == TRUE)
    {
        hid_t correction_datatype; /* Datatype of correction dataset */

        /* Open CORRECTION_NAME and get datatype */
        hid_t hdf_correction_id = H5Dopen(group_id, CORRECTION_NAME,
            H5P_DEFAULT);
        if (hdf_correction_id < 0)
        {
            IAS_LOG_ERROR("Opening data set %s", LOOKUP_NAME);
            H5Gclose(group_id);
            return ERROR;
        }
        
        /* Get datatype for CORRECTION_NAME */
        correction_datatype = H5Dget_type(hdf_correction_id);
        if (correction_datatype < 0)
        {
            IAS_LOG_ERROR("Getting data type of dataset %s", CORRECTION_NAME);
            H5Dclose(hdf_correction_id);
            H5Gclose(group_id);
            return ERROR;
        }

        /* Closing lookup, correction, and group */
        if (H5Dclose(hdf_correction_id) < 0)
        {
            IAS_LOG_ERROR("Could not close H5D for correction");
            H5Tclose(correction_datatype);
            H5Gclose(group_id);
            return ERROR;
        }

        /* Check to see if the datatype is correct for correction */
        if (!H5Tequal(correction_datatype, H5T_IEEE_F32LE))
        {
            IAS_LOG_ERROR("Unexpected data type for group %s, dataset %s",
                group_name, CORRECTION_NAME);
            H5Tclose(correction_datatype);
            H5Gclose(group_id);
            return ERROR;
        }

        /* Close the correction_datatype */
        if (H5Tclose(correction_datatype) < 0)
        {
            IAS_LOG_ERROR("Could not close H5T");
            H5Gclose(group_id);
            return ERROR;
        }
    }
    else
    {
        IAS_LOG_ERROR("Dataset %s does not exist for %s", CORRECTION_NAME,
            group_name);
        H5Gclose(group_id);
        return ERROR;

    }

    /* Close group */    
    if (H5Gclose(group_id) < 0)    
    {
        IAS_LOG_ERROR("Could not close H5G");
        return ERROR;
    }

    return SUCCESS;
}

/* -----------------------------------------------------------------------------
 *  NAME: ias_rlut_verify_tirs_secondary_group
 *  PURPOSE: Serves as pass through routine for ias_rlut_verify_group_datatype
 *           to verify the group LINEARITY_LOOKUP_GROUP_NAME.
 *
 * RETURNS: SUCCESS / ERROR
 * ---------------------------------------------------------------------------*/
int ias_rlut_verify_linearity_group
(
    const IAS_RLUT_IO *rlut,        /* I: Used to get the file id to check */
    int band_number,                /* I: Band number of group to check*/
    int sca_number                  /* I: SCA number of group to check */
)
{
    return verify_group_and_datatype(rlut,LINEARITY_LOOKUP_GROUP_NAME,
        band_number, sca_number);
}

/* -----------------------------------------------------------------------------
 *  NAME: ias_rlut_verify_tirs_secondary_group
 *  PURPOSE: Serves as pass through routine for ias_rlut_verify_group_datatype
 *           to verify the group LINEARIZATION_PARAMS_GROUP_NAME.
 *
 * RETURNS: SUCCESS / ERROR
 * ---------------------------------------------------------------------------*/
int ias_rlut_verify_linearization_group
(
    const IAS_RLUT_IO *rlut,        /* I: Used to get the file id to check */
    int band_number,                /* I: Band number of group to check*/
    int sca_number                  /* I: SCA number of group to check */
)
{
    return verify_group_and_datatype(rlut,
        LINEARIZATION_PARAMS_GROUP_NAME, band_number, sca_number);
}

/* -----------------------------------------------------------------------------
 *  NAME: ias_rlut_verify_tirs_secondary_group
 *  PURPOSE: Serves as pass through routine for ias_rlut_verify_group_datatype
 *           to verify the group TIRS_SECONDARY_LINEARIZATION_GROUP_NAME.
 *
 * RETURNS: SUCCESS / ERROR
 * ---------------------------------------------------------------------------*/
int ias_rlut_verify_tirs_secondary_group
(
    const IAS_RLUT_IO *rlut,        /* I: Used to get the file id to check */
    int band_number,                /* I: Band number of group to check*/
    int sca_number                  /* I: SCA number of group to check */
)
{
    return verify_group_and_datatype(rlut,
        TIRS_SECONDARY_LINEARIZATION_GROUP_NAME, band_number, sca_number);
}






