/*==============================================================================
NAME: ias_rlut_write_lookup_table

PURPOSE:
    Provide the RLUT IO support for writing a response linearization lookup
    table.

==============================================================================*/

/* System Includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* COTS Includes */
#include "hdf5.h"

/* IAS Includes */
#include "ias_const.h"
#include "ias_types.h"
#include "ias_logging.h"
#include "ias_rlut.h"
#include "ias_satellite_attributes.h"

/* Local Includes */
#include "ias_rlut_private.h"

/*------------------------------------------------------------------------------
 * The linearity lookup table information is organized as follows in the HDF5
 * file.  The correction uses a lookup table to locate a DN value from the
 * satellite and then use a parallel array to find the corresponding correction
 * value that matches that input value.
 * 
 * Each band and SCA has a different set of lookup and correction values for
 * both of the corrections.  These values are organized within the HDF5
 * file as follows:
 *
 *     /
 *     |-- LINEARITY_LOOKUP
 *         |-- BAND01
 *         |   |-- SCA01
 *         |   |   |-- DN_LUT   (Data set of DN lookup values)
 *         |   |   `-- Correction  (Data set of Linearity correction values)
 *         |   |-- SCA02
 *         |   |   |-- DN_LUT
 *         |   |   `-- Correction
 *         |   |
 *         |   |   ... all SCA's ...
 *         |   |
 *         |   `-- SCA14
 *         |       |-- DN_LUT
 *         |       `-- Correction
 *         |
 *         |   ... all normal bands ...
 *         |
 *         `-- Band11
 *             |-- SCA01
 *             |   |-- DN_LUT
 *             |   `-- Correction
 *             |-- SCA02
 *             |   |-- DN_LUT
 *             |   `-- Correction
 *             `-- SCA03
 *                 |-- DN_LUT
 *                 `-- Correction
 *
 *----------------------------------------------------------------------------*/


/*------------------------------------------------------------------------------
NAME: util_write_array

PURPOSE:
    This utility function will write a 2D array of values to a data set in an
    HDF5 file.  This function can be used to write the lookup DN values as well
    as the correction values.

RETURNS:
    This function will return SUCCESS if the values are successfully written to
    the file, ERROR if an error is encountered.
------------------------------------------------------------------------------*/
static int util_write_array
(
    hid_t hdf_group_id,   /* I: HDF5 group ID being written to */
    hsize_t dims[2],      /* I: Dimensions of the 'values' parameter */
    const char *hdf_path, /* I: Full path of the data set */
    float *values         /* I: 2D array to be written to the file */
)
{
    herr_t hdf_status;      /* HDF5 status */
    hid_t  hdf_dataset_id;
    hid_t  hdf_dataspace_id;

    /* Create a new HDF5 dataspace for the table of values.  The first argument
       indicates the number of dimensions in the data space.  The last argument
       is the maximum dimensions of the data space, but by passing in a NULL
       pointer, the maximum dimensions will be set to the same as the initial
       dimensions. */
    hdf_dataspace_id = H5Screate_simple(2, dims, NULL);
    if (hdf_dataspace_id < 0)
    {
        IAS_LOG_ERROR("Creating an HDF5 data space for %s", hdf_path);
        return ERROR;
    }

    /* Create the data set for the table of values within the HDF5 file */
    hdf_dataset_id = H5Dcreate(hdf_group_id, hdf_path, H5T_IEEE_F32LE,
        hdf_dataspace_id, H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);
    if (hdf_dataset_id < 0)
    {
        IAS_LOG_ERROR("Creating an HDF5 data space for %s", hdf_path);
        H5Sclose(hdf_dataspace_id);
        return ERROR;
    }

    /* Write the 2D array of values to the HDF5 file */
    hdf_status = H5Dwrite(hdf_dataset_id, H5T_NATIVE_FLOAT, H5S_ALL, H5S_ALL,
        H5P_DEFAULT, values);
    if (hdf_status < 0)
    {
        IAS_LOG_ERROR("Writing to HDF5 data space for %s", hdf_path);
        H5Dclose(hdf_dataset_id);
        H5Sclose(hdf_dataspace_id);
        return ERROR;
    }

    /* Close the data set we just wrote */
    hdf_status = H5Dclose(hdf_dataset_id);
    if (hdf_status < 0)
    {
        IAS_LOG_ERROR("Closing data set for %s", hdf_path);
        H5Sclose(hdf_dataspace_id);
        return ERROR;
    }

    /* Close the data space we created for the table */
    hdf_status = H5Sclose(hdf_dataspace_id);
    if (hdf_status < 0)
    {
        IAS_LOG_ERROR("Closing data space for %s", hdf_path);
        return ERROR;
    }

    return SUCCESS;
}

/*------------------------------------------------------------------------------
NAME: ias_rlut_write_lookup_table

PURPOSE:
    This function will write a set of DN lookup values and their corresponding
    correction values for either a Linearity correction or a Nonuniformity
    correction.

RETURNS:
    This function will return SUCCESS if the values are successfully written to
    the RLUT file, ERROR if an error is encountered.
------------------------------------------------------------------------------*/
int ias_rlut_write_lookup_table
(
    const IAS_RLUT_IO *rlut,  /* I/O: File descriptor opened for writing */
    IAS_RLUT_LOOKUP_TYPE type,/* I: Type of lookup table to write */
    IAS_RLUT_LOOKUP_TABLE *table /* I: lookup table */
)
{
    herr_t hdf_status;                /* HDF5 status */
    hid_t hdf_group_id;                /* HDF5 Group ID */
    hsize_t dims[2];
    int status;
    const char *root_name;

    /* Ensure the file is open */
    if ((rlut == NULL) || (rlut->file_id < 0))
    {
        IAS_LOG_ERROR("RLUT file is not open");
        return ERROR;
    }

    /* Set the group name */
    if (type == IAS_RLUT_LINEARITY_LOOKUP)
        root_name = LINEARITY_LOOKUP_GROUP_NAME;
    else if (type == IAS_RLUT_TIRS_SECONDARY_LOOKUP)
        root_name = TIRS_SECONDARY_LINEARIZATION_GROUP_NAME;
    else
    {
        IAS_LOG_ERROR("Unsupported lookup table type: %d", type);
        return ERROR;
    }

    /* Validate the lookup table */
    if (table == NULL)
    {
        IAS_LOG_ERROR("RLUT lookup table not allocated");
        return ERROR;
    }
    if (table->num_detectors !=
        ias_sat_attr_get_detectors_per_sca(table->band_num))
    {
        IAS_LOG_ERROR("Lookup table contains an invalid number of detectors "
            "for band number %d (%d provided, should be %d)", table->band_num,
            table->num_detectors, 
           ias_sat_attr_get_detectors_per_sca(table->band_num));
        return ERROR;
    }

    /* Initialize the dimensions of the data that we'll be writing */
    dims[0] = table->num_detectors;
    dims[1] = table->num_values;

    hdf_group_id =  ias_rlut_get_group(rlut, root_name, table->band_num,
        table->sca_num, CREATE_MISSING, NULL);
    if (hdf_group_id < 0)
    {
        IAS_LOG_ERROR("Opening %s group for band number %d, sca number %d",
            root_name, table->band_num, table->sca_num);
        return ERROR;
    }

    /* Write the DN lookup values to the HDF5 file */
    status = util_write_array(hdf_group_id, dims, LOOKUP_NAME, table->input_dn);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing DN values to %s group for band number %d, sca "
            "number %d", root_name, table->band_num, table->sca_num);
        H5Gclose(hdf_group_id);
        return ERROR;
    }

    /* Write the correction values to the HDF5 file */
    status = util_write_array(hdf_group_id, dims, CORRECTION_NAME,
        table->output_correction);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing output correction values to %s group for band "
            "number %d, sca number %d", root_name, table->band_num,
            table->sca_num);
        H5Gclose(hdf_group_id);
        return ERROR;
    }

    /* Close the open group */
    hdf_status = H5Gclose(hdf_group_id);
    if (hdf_status < 0)
    {
        IAS_LOG_ERROR("Closing %s group for band number %d, sca number %d",
            root_name, table->band_num, table->sca_num);
        return ERROR;
    }
    return SUCCESS;
}
