/***************************************************************************
 NAME:     rlut_io_tests

 PURPOSE:  CGreen unit testing of various RLUT IO library components

 RETURNS:  CGreen test status
****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "hdf5.h"
#include "hdf5_hl.h"
#include "ias_logging.h"
#include "ias_types.h"
#include "ias_const.h"
#include "ias_satellite_attributes.h"
#include "cgreen.h"
#include "../ias_rlut.h"
#include "../ias_rlut_private.h"


static char test_rlut_name[100] = "test_LC08RLUT_20120101_20120331_01_01.h5";


/***************************************************************************
 NAME:     set_parameter_test_values

 PURPOSE:  Internal routine to calculate a set of test values for the set of
           linearization parameters for a given detector

 RETURNS:  Nothing
****************************************************************************/
static void set_parameter_test_values
(
    IAS_RLUT_LINEARIZATION_PARAMS *params, /* I/O:  Array of linearization
                                              parameters for all detectors */
    int band_number,                       /* I: Current band number */
    int det_index                          /* I: Current detector index */
)
{
    params[det_index].cutoff_threshold_low = (100 * band_number)
        + (10 * det_index);
    params[det_index].cutoff_threshold_high = (100 * (band_number + 1))
        + (10 * (det_index + 1));
    params[det_index].remap_coeff0_low =
        0.0001 * (band_number + (10 * det_index));
    params[det_index].remap_coeff1_low =
        -0.0002 * ((band_number + 1) + (10 * (det_index + 1)));
    params[det_index].remap_coeff2_low =
        0.0003 * ((band_number + 2) + (10 * (det_index + 2)));
    params[det_index].remap_coeff0_mid =
        0.0001 * (band_number + (10 * det_index));
    params[det_index].remap_coeff1_mid =
        -0.0002 * ((band_number + 1) + (10 * (det_index + 1)));
    params[det_index].remap_coeff2_mid =
        0.0003 * ((band_number + 2) + (10 * (det_index + 2)));
    params[det_index].remap_coeff0_high =
        0.0001 * (band_number + (10 * det_index));
    params[det_index].remap_coeff1_high =
        -0.0002 * ((band_number + 1) + (10 * (det_index + 1)));
    params[det_index].remap_coeff2_high =
        0.0003 * ((band_number + 2) + (10 * (det_index + 2)));
}   /* END set_parameter_test_values */

/***************************************************************************
 NAME:     set_lookup_test_values

 PURPOSE:  Internal routine to calculate a set of test values for the set of
           lookup table values for a given detector.  The table passed in
           will need to have members freed by the caller.

 RETURNS:  SUCCESS/ERROR
****************************************************************************/
static int set_lookup_test_values
(
    IAS_RLUT_LOOKUP_TABLE *table  /* I/O: Lookup table to populate */
)
{
    int det_index;
    int val_index;

    table->input_dn = malloc(table->num_detectors * table->num_values
        * sizeof(*table->input_dn));
    assert_not_equal_with_message(table->input_dn, NULL, "Error allocating "
        "input DN lookup table");
    if (table->input_dn == NULL)
        return ERROR;

    table->output_correction = malloc(table->num_detectors * table->num_values
        * sizeof(*table->output_correction));
    assert_not_equal_with_message(table->output_correction, NULL,
        "Error allocating correction lookup table");
    if (table->output_correction == NULL)
    {
        free(table->output_correction);
        return ERROR;
    }

    for (det_index = 0; det_index < table->num_detectors; det_index++)
    {
        for (val_index = 0; val_index < table->num_values; val_index++)
        {
            int index = det_index * table->num_values + val_index;
            table->input_dn[index] = table->band_num * 0.1
                + table->sca_num * 0.01 + det_index + val_index;
            table->output_correction[index] = table->band_num * 0.2
                + table->sca_num * 0.02 + det_index * 2 + val_index * 2;
        }
    }

    return SUCCESS;
}

/***************************************************************************
 NAME:     write_parameters

 PURPOSE:  Internal function to write test linearization parameters to an
           RLUT file

 RETURNS:  Integer status code of SUCCESS or ERROR
****************************************************************************/
static int write_parameters
(
    const IAS_RLUT_IO *rlut,   /* I: Open RLUT file */
    int band_number,           /* I: Current band number */
    int sca_number,            /* I: Current SCA number */
    int num_detectors,         /* I: Number of detectors in the current SCA */
    int tirs_band_number       /* I: */
)
{
    IAS_RLUT_LINEARIZATION_PARAMS *write_params = NULL;
    IAS_RLUT_LOOKUP_TABLE parms;
    int det_index;
    int status;
    int return_status = SUCCESS;


    /* Allocate the write parameter data buffer */
    write_params =
        malloc(num_detectors * sizeof(IAS_RLUT_LINEARIZATION_PARAMS));
    assert_not_equal_with_message(write_params, NULL, "Error allocating "
        "write parameter data buffer");
    if (write_params == NULL)
        return ERROR;

    /* Create "fake" parameter values */
    for (det_index = 0; det_index < num_detectors; det_index++)
    {
        set_parameter_test_values(write_params, band_number, det_index);
    }

    /* Write the coefficients to the file */
    status = ias_rlut_write_linearization_params(rlut, write_params,
        band_number, sca_number, num_detectors);
    if ( status != SUCCESS )
        return_status = ERROR;

    /* Free the data buffer */
    free(write_params);

    /* Check the return status for the parameter write */
    assert_equal_with_message(status, SUCCESS, "Error writing band %d SCA "
        "%d linearization parameters to file", band_number, sca_number);

    /* Set up the TIRS secondary parameters */
    parms.band_num = tirs_band_number;
    parms.sca_num = 2;
    parms.num_detectors = 640;
    parms.num_values = 5;
    status = set_lookup_test_values(&parms);
    if (status != SUCCESS)
        return ERROR;

    /* Write the data */
    status = ias_rlut_write_lookup_table( rlut, IAS_RLUT_TIRS_SECONDARY_LOOKUP,
        &parms );
    if ( status != SUCCESS )
        return_status = ERROR;

    assert_equal_with_message( status, SUCCESS, "Error writing the "
        "TIRS secondary parameters");

    free(parms.input_dn);
    free(parms.output_correction);

    return return_status;
}   /* END write_parameters */



/***********************************************************************
 NAME:     read_parameters

 PURPOSE:  Internal function to read test linearization parameters from
           RLUT file

 RETURNS:  Integer status code of SUCCESS or ERROR
************************************************************************/
static int read_parameters
(
    const IAS_RLUT_IO *rlut,   /* I: Open RLUT file */
    int band_number,           /* I: Current band number */
    int sca_number,            /* I: Current SCA number */
    int num_detectors,         /* I: Number of detectors in the current SCA */
    int tirs_band_number       /* I: */
)
{
    IAS_RLUT_LINEARIZATION_PARAMS *test_params = NULL;
    IAS_RLUT_LINEARIZATION_PARAMS *read_params = NULL;
    IAS_RLUT_LOOKUP_TABLE *inparms;
    IAS_RLUT_LOOKUP_TABLE expected_table;
    int det_index;
    int status;
    int return_status = SUCCESS;
    size_t size_of_data;


    /* Allocate the test (expected) parameter data buffer */
    test_params = malloc(num_detectors * sizeof(IAS_RLUT_LINEARIZATION_PARAMS));
    assert_not_equal_with_message(test_params, NULL, "Error allocating "
        "test parameter data buffer");
    if (test_params == NULL)
        return ERROR;

    /* Create "fake" parameter values identical to those created in the
       write routine */
    for (det_index = 0; det_index < num_detectors; det_index++)
    {
        set_parameter_test_values(test_params, band_number, det_index);
    }

    /* Read the coefficients from the file */
    read_params = ias_rlut_read_linearization_params(rlut, band_number,
        sca_number, num_detectors);
    assert_not_equal_with_message(read_params, NULL, "Error reading band "
        "%d SCA %d linearization parameters from file", band_number,
        sca_number);
    if (read_params == NULL)
    {
        free(test_params);
        return ERROR;
    }

    /* Check the read back parameters vs the "test" parameters */
    status = memcmp(test_params, read_params,
        num_detectors * sizeof(IAS_RLUT_LINEARIZATION_PARAMS));

    /* Free the data buffers */
    free(read_params);
    free(test_params);

    /* Check the status of the parameter read */
    if (status != 0)
    {
        assert_true_with_message(FALSE, "Test parameter data buffer "
            "contents do not equal read parameter buffer contents");
        for (det_index = 0; det_index < num_detectors; det_index++)
        {
            if (memcmp(&read_params[det_index], &test_params[det_index],
                sizeof(IAS_RLUT_LINEARIZATION_PARAMS)) != 0)
            {
                assert_true_with_message(FALSE, "Band %d SCA %d read "
                    "parameters for detector %d not equal to corresponding "
                    "test parameters", band_number, sca_number,
                    (det_index + 1));
            }
        }
    }
    if (status != 0)
        return_status = ERROR;

    /* Read the TIRS secondary parameters */
    inparms = ias_rlut_read_lookup_table( rlut, IAS_RLUT_TIRS_SECONDARY_LOOKUP,
        tirs_band_number, 2 );
    if ( ! inparms )
    {
        assert_true_with_message( FALSE, "Error reading TIRS secondary "
            "parameters");
        return ERROR;
    }
    assert_equal( inparms->band_num, tirs_band_number );
    assert_equal( inparms->sca_num, 2 );
    assert_equal( inparms->num_detectors, 640 );
    assert_equal( inparms->num_values, 5 );

    /* Set up the expected values */
    expected_table.band_num = tirs_band_number;
    expected_table.sca_num = 2;
    expected_table.num_detectors = 640;
    expected_table.num_values = 5;
    status = set_lookup_test_values(&expected_table);
    if (status != SUCCESS)
        return ERROR;

    /* Compare the values */
    size_of_data = sizeof(float) * inparms->num_detectors *
        inparms->num_values;
    status = memcmp( inparms->input_dn, expected_table.input_dn, size_of_data );
    if ( status != 0 )
    {
        assert_true_with_message( FALSE, "TIRS secondary DN look up "
            "doesn't match expected results");
        return_status = ERROR;
    }
    status = memcmp( inparms->output_correction, 
        expected_table.output_correction, size_of_data );
    if ( status != 0 )
    {
        assert_true_with_message( FALSE, "TIRS secondary linearity "
            "correction doesn't match expected results");
        return_status = ERROR;
    }
    ias_rlut_free_lookup_table( inparms );
    free(expected_table.input_dn);
    free(expected_table.output_correction);

    return return_status;
}




/******************** UNIT TESTS ***************************************/
/* Test RLUT file open with IAS_WRITE and IAS_UPDATE access modes.
   IAS_WRITE is a supported access mode, IAS_UPDATE is not.  The
   unit tests for file attributes and linearization parameter
   processing (test_file_attributes_io and test_linearization_parameter_io)
   will test the IAS_READ access mode... */
Ensure test_open_close_file()
{
    IAS_RLUT_IO *rlut = NULL;
    int status;
    int file_format_version;
    herr_t hdf_status;


    /* Try creating an empty RLUT file */
    printf("Testing creation of new RLUT file %s...\n", test_rlut_name);

    /* Create the empty file */
    rlut = ias_rlut_open_file(test_rlut_name, IAS_WRITE);
    assert_not_equal_with_message(rlut, NULL, "Error creating RLUT file %s",
        test_rlut_name);

    /* Make sure the file type attribute exists */
    hdf_status = H5LTfind_attribute(rlut->file_id, IAS_RLUT_FILE_TYPE);
    if (hdf_status < 0)
    {
        assert_true_with_message(FALSE, "File type attribute not found");
        ias_rlut_close_file(rlut);
        return;
    }

    /* Check the file format version */
    hdf_status = H5LTfind_attribute(rlut->file_id, "RLUT Format Version");
    if (hdf_status < 0)
    {
        assert_true_with_message(FALSE, "Format version attribute not found");
        ias_rlut_close_file(rlut);
        return;
    }

    hdf_status = H5LTget_attribute_int(rlut->file_id, "/",
        "FILE_FORMAT_VERSION", &file_format_version);
    if (hdf_status < 0)
    {
        assert_true_with_message(FALSE, "Error retrieving file format "
            "information");
        ias_rlut_close_file(rlut);
        return;
    }
    if (file_format_version != IAS_RLUT_FORMAT_VERSION)
    {
         assert_true_with_message(FALSE, "Retrieved file format version %d "
             "does not equal %d", file_format_version,
             IAS_RLUT_FORMAT_VERSION);
         ias_rlut_close_file(rlut);
         return;
    }

    /* Close the RLUT file */
    status = ias_rlut_close_file(rlut);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error closing RLUT file %s",
            test_rlut_name);
        return;
    }

    /* Validate it's a legitimate RLUT file */
    int is_rlut_flag = ias_rlut_is_rlut_file(test_rlut_name);
    assert_equal_with_message(is_rlut_flag, 1, "RLUT file type not validated "
        "for file %s", test_rlut_name);

    /* Try opening the file in IAS_UPDATE mode.  Should create an ERROR
       status because the operation is not supported.  Don't log the
       error message, however...*/
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    rlut = ias_rlut_open_file(test_rlut_name, IAS_UPDATE);
    assert_equal_with_message(rlut, NULL, "RLUT file %s should not be "
        "accessible in an update mode");
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);
}   /* END test_open_close_file */



/* Test RLUT file attributes read/write capability */
Ensure test_file_attribute_io()
{
    IAS_RLUT_FILE_ATTRIBUTES write_file_attr;
    IAS_RLUT_FILE_ATTRIBUTES read_file_attr;
    IAS_RLUT_IO *rlut = NULL;
    int status;


    printf("Testing file attribute I/O...\n");

    /* Clear the file attributes data structures */
    memset(&write_file_attr, 0, sizeof(IAS_RLUT_FILE_ATTRIBUTES));

    /* Initialize file attributes test data values */
    strcpy(write_file_attr.file_source, "EROS Data Center");
    strcpy(write_file_attr.effective_date_begin, "2012-01-01T00:00:00");
    strcpy(write_file_attr.effective_date_end, "2012-03-31T00:00:00");
    strcpy(write_file_attr.status, "ACTIVE");
    strcpy(write_file_attr.baseline_date, "2012-01-04T00:00:00");

    sprintf(write_file_attr.description, "Test RLUT file used to validate "
            "HDF5 write/read capabilities");

    write_file_attr.version = 1;
    write_file_attr.collection = 1;

    /* Overwrite any existing test RLUT file with this file name */
    rlut = ias_rlut_open_file(test_rlut_name, IAS_WRITE);
    if (rlut == NULL)
    {
        assert_true_with_message(FALSE, "Error opening RLUT file %s",
            test_rlut_name);
        return;
    }

    /* Test ias_rlut_write_attributes function */
    printf("Writing file attributes to RLUT file...\n");
    status = ias_rlut_write_file_attributes(rlut, &write_file_attr);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "File attributes write to RLUT file "
            "%s failed", test_rlut_name);
        ias_rlut_close_file(rlut);
        return;
    }

    /* Close the file */
    status = ias_rlut_close_file(rlut);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error closing RLUT file %s",
            test_rlut_name);
        return;
    }

    /* Reopen the file, this time for reading */
    rlut = NULL;
    rlut = ias_rlut_open_file(test_rlut_name, IAS_READ);
    if (rlut== NULL)
    {
        assert_not_equal_with_message(FALSE, "Error opening RLUT file %s ",
        test_rlut_name);
        return;
    }

    /* Test ias_rlut_read_attributes function */
    printf("Reading file attributes from RLUT file...\n");
    memset(&read_file_attr, 0, sizeof(IAS_RLUT_FILE_ATTRIBUTES));
    status = ias_rlut_read_file_attributes(rlut, &read_file_attr);

    /* Close the RLUT file */
    ias_rlut_close_file(rlut);

    /* Check the return status */
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "File attributes read from RLUT file "
        "%s failed\n", test_rlut_name);
        return;
    }

    /* Check the attribute fields.  First, do an overall check */
    printf("Validating file attributes read...\n");
    status = memcmp(&read_file_attr, &write_file_attr,
        sizeof(IAS_RLUT_FILE_ATTRIBUTES));
    if (status != 0)
    {
        assert_true_with_message(FALSE, "Read file attributes do not "
        "equal written file attributes:");

        /* Check each attribute individually.  One or more may not
           be processed correctly */
        if (strcmp(write_file_attr.file_source, read_file_attr.file_source)
            != 0)
        {
            assert_true_with_message(FALSE, "Read file source attribute "
                "%s does not equal written file source attribute %s",
                read_file_attr.file_source, write_file_attr.file_source);
        }
        if (strcmp(write_file_attr.effective_date_begin,
            read_file_attr.effective_date_begin) != 0)
        {
            assert_true_with_message(FALSE, "Read effective begin date "
                "attribute %s does not equal written effective begin date "
                "attribute %s", read_file_attr.effective_date_begin,
                write_file_attr.effective_date_begin);
        }
        if (strcmp(write_file_attr.effective_date_end,
            read_file_attr.effective_date_end) != 0)
        {
            assert_true_with_message(FALSE, "Read effective end date "
                "attribute %s does not equal written effective end "
                "date attribute %s", read_file_attr.effective_date_end,
                write_file_attr.effective_date_begin);
        }
        if (strcmp(write_file_attr.status, read_file_attr.status) != 0)
        {
            assert_true_with_message(FALSE, "Read status attribute "
                "%s does not equal written status attribute %s",
                read_file_attr.status, write_file_attr.status);
        }
        if (strcmp(write_file_attr.baseline_date,
            read_file_attr.baseline_date) != 0)
        {
            assert_true_with_message(FALSE, "Read baseline date attribute "
                "%s does not equal written baseline date attribute %s",
                read_file_attr.baseline_date, write_file_attr.baseline_date);
        }
        if (strcmp(write_file_attr.description,
            read_file_attr.description) != 0)
        {
            assert_true_with_message(FALSE, "Read description attribute "
                "%s does not equal written description attribute %s",
                read_file_attr.description, write_file_attr.description);
        }
        if (write_file_attr.version != read_file_attr.version)
        {
            assert_true_with_message(FALSE, "Read version attribute %d "
                "does not equal written version attribute %d",
                read_file_attr.version, write_file_attr.version);
        }
        if (write_file_attr.collection != read_file_attr.collection)
        {
            assert_true_with_message(FALSE, "Read collection attribute %d "
                "does not equal written collection attribute %d",
                read_file_attr.collection, write_file_attr.collection);
        }
    }
}   /* END test_file_attribute_io */



/* Test linearization parameter IO */
Ensure test_linearization_parameter_io()
{
    IAS_RLUT_IO *rlut = NULL;
    int band_number1 = 8;
    int sca_number1 = 10;
    int num_detectors1 = 988;
    int band_number2 = 15;
    int sca_number2 = 2;
    int num_detectors2 = 640;
    int status;
    IAS_RLUT_LINEARIZATION_PARAMS *read_params;
    IAS_RLUT_LOOKUP_TABLE *tirs_read_params;


    printf("Testing linearization parameter I/O...\n");

    /* Overwrite any existing RLUT file */
    rlut = ias_rlut_open_file(test_rlut_name, IAS_WRITE);
    if (rlut == NULL)
    {
        assert_true_with_message(FALSE, "Error opening RLUT file %s", 
            test_rlut_name);
        return;
    }

    /* Test case 1:  Write SCA 10 of the OLI pan band (8) */
    printf("Writing test case 1:  OLI PAN band (band %d), SCA %d "
        "(%d detectors)\n", band_number1, sca_number1, num_detectors1);
    status = write_parameters(rlut, band_number1, sca_number1,
        num_detectors1, 10 );
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error writing test 1 parameters to "
            "RLUT file %s", test_rlut_name);
        ias_rlut_close_file(rlut);
        return;
    }

    /* Test case 2:  Write SCA 2 of the TIRS blind band (15) */
    printf("Writing test case 2:  TIRS blind band (band %d), SCA %d "
        "(%d detectors)\n", band_number2, sca_number2, num_detectors2);
    status = write_parameters(rlut, band_number2, sca_number2,
        num_detectors2, 11 );
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error writing test 2 parameters to "
            "RLUT file %s", test_rlut_name);
        ias_rlut_close_file(rlut);
        return;
    }

    /* Close the file */
    status = ias_rlut_close_file(rlut);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error closing RLUT file %s",
            test_rlut_name);
        return;
    }

    /* Reopen the file, this time for reading */
    rlut = NULL;
    rlut = ias_rlut_open_file(test_rlut_name, IAS_READ);
    if (test_rlut_name == NULL)
    {
        assert_true_with_message(FALSE, "Error (re)opening RLUT file %s",
             test_rlut_name);
        return;
    }

    /* Read the parameters and verify they were written correctly */
    printf("Reading/validating test case 1...\n");
    status = read_parameters(rlut, band_number1, sca_number1,
        num_detectors1, 10 );
    assert_equal_with_message(status, SUCCESS, "Error reading test case 1"
        "linearization parameters from RLUT file %s", test_rlut_name);

    printf("Reading/validating test case 2...\n");
    status = read_parameters(rlut, band_number2, sca_number2,
        num_detectors2, 11 );
    assert_equal_with_message(status, SUCCESS, "Error reading test case 2"
        "linearization parameters from RLUT file %s", test_rlut_name);

    /* Make sure that reading a band that exists, but an SCA that doesn't exist
       in the file detects an error but doesn't spew out HDF error messages
       (visually check that).   Also check with both a band and SCA that do
       not exist in the file. */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    read_params = ias_rlut_read_linearization_params(rlut, band_number1,
        sca_number1 + 1, num_detectors1);
    assert_equal_with_message(read_params, NULL, "Did not get an error reading"
        "band %d SCA %d linearization parameters from file like expected",
        band_number1, sca_number1 + 1);

    read_params = ias_rlut_read_linearization_params(rlut, band_number1 + 1,
        sca_number1, num_detectors1);
    assert_equal_with_message(read_params, NULL, "Did not get an error reading"
        " band %d SCA %d linearization parameters from file like expected",
        band_number1, sca_number1 + 1);

    tirs_read_params = ias_rlut_read_lookup_table( rlut,
        IAS_RLUT_TIRS_SECONDARY_LOOKUP, 10, 1 );
    assert_equal_with_message(tirs_read_params, NULL, "Did not get an error "
        "reading band %d SCA %d linearization parameters from file like "
        "expected", 10, 1);

    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

    /* Close the file */
    ias_rlut_close_file(rlut);

}   /* END test_linearization_parameter_io */



/************************* MAIN *******************************************/
int main(int argc, char **argv)
{
    /* Initialize the logging library */
    ias_log_initialize("rlut_io_tests");

    /* Initialize the satellite attributes library */
    if (ias_sat_attr_initialize(IAS_L8) != SUCCESS)
    {
        assert_true_with_message(FALSE,
            "Error initializing satellite attributes library");
        return EXIT_FAILURE;
    }

    /* Define the master test suite */
    TestSuite *suite = create_test_suite();

    add_test(suite, test_open_close_file);
    add_test(suite, test_file_attribute_io);
    add_test(suite, test_linearization_parameter_io);

    /* If a single test name is a command line parameter, run that test.
       Otherwise, run the entire suite */
    if (argc > 1)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}
