/*==============================================================================
NAME:
    rlut_lookup_table_tests

PURPOSE:
    This application performs unit tests on the alternate Response linearization
    LookUp Table I/O routines.

==============================================================================*/

/* System Includes */
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/wait.h>

/* IAS Includes */
#include "ias_logging.h"
#include "ias_satellite_attributes.h"
#include "ias_rlut.h"
#include "ias_rlut_private.h"

#include "cgreen/cgreen.h"

/* Local Includes */

#define NVALUES 20      /* Number of values in each row */

/*------------------------------------------------------------------------------
NAME: util_get_benchmark

PURPOSE:
    Return the path to the benchmark files in IAS_UNIT_TEST_DATA_DIR

RETURNS:
    This function return SUCCESS if the path is successfully returned, ERROR if
    an error is encountered.
------------------------------------------------------------------------------*/
int util_get_benchmark
(
    int file_format, /* I: The file format version (currently 2 or 3) */
    char path[PATH_MAX] /* O: Path to the benchmark files */
)
{
    char * pDir  = NULL;
    int    bytes;

    /* Get the IAS_UNIT_TEST_DATA_DIR directory */
    pDir = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (pDir == NULL)
    {
        assert_true_with_message(FALSE, "IAS_UNIT_TEST_DATA_DIR not defined");
        return ERROR;
    }

    /* Create the path to the benchmark RLUT file */
    if (file_format == 3)
        bytes = snprintf(path, PATH_MAX,
                         "%s/ias_lib/io/rlut/lookup_benchmark.h5", pDir);
    else if (file_format == 2)
        bytes = snprintf(path, PATH_MAX,
                         "%s/ias_lib/io/rlut/lookup_benchmark_format2.h5",
                         pDir);
    else
    {
        assert_true_with_message(FALSE,
            "Invalid file format of %d. Expected 2 or 3", file_format);
        return ERROR;
    }

    if ((bytes < 0) || (bytes >= PATH_MAX))
    {
        assert_true_with_message(FALSE,
            "Error path too small, requires %d characters", bytes);
        return ERROR;
    }
    return SUCCESS;
}


/*------------------------------------------------------------------------------
NAME: util_file_diff

PURPOSE:
    Run an h5diff on the specified files.

RETURNS:
    This function will return SUCCESS if the file diff application is
    successfully run and the files are identical, ERROR if an error is
    encountered or if the files are different.
------------------------------------------------------------------------------*/
int util_file_diff
(
    char * path1, /* I: Path to an RLUT file */
    char * path2  /* I: Path to an RLUT file */
)
{
    char         cmd[1024];
    char         h5diff[1024] = "h5diff";
    const char * pEnv;
    int          status;

    /* Create the command to run an h5diff comparison of two files. */
    if ((pEnv = getenv("HDF5_HOME")) != NULL)
    {
        sprintf(h5diff, "%s/bin/h5diff", pEnv);
        if (access(h5diff, X_OK) != 0)
            strcpy(h5diff, "h5diff");
    }
    sprintf(cmd, "%s --quiet %s %s", h5diff, path1, path2);

    status = system(cmd);
    if (WIFSIGNALED(status) && (WTERMSIG(status) || WTERMSIG(status)))
    {
        assert_true_with_message(0, "h5diff file comparison interrupted");
        return ERROR;
    }
    if (WEXITSTATUS(status) != 0)
    {
        assert_true_with_message(0, "RLUTs differ:\n  %s\n  %s", path1, path2);
        return ERROR;
    }
    return SUCCESS;
}


/*------------------------------------------------------------------------------
NAME: util_table

PURPOSE:
    This function will compute a value for each entry in a two dimensional
    lookup table.  If the fillFlag argument is TRUE, the table will be filled
    with these computed values.  If the fillFlag is false, each position within
    the table will be compared to the computed value.

RETURNS:
    If fillFlag is TRUE, this function will return a status of 0 to indicate
    success.  If fillFlag is FALSE, this function will return the number of
    differences that were found within the nDetectors x nValues table.
------------------------------------------------------------------------------*/
int util_table
(
    int fillFlag,   /* I: If TRUE table will be filled, otherwise examined */
    int isLookup,   /* I: TRUE for DN lookup values, else correction values */
    int band,       /* I: Band number that is being examined or filled */
    int sca,        /* I: SCA number that is being examined or filled */
    int nDetectors, /* I: Number of detectors in the band */
    int nValues,    /* I: Number of values in each detector row */
    float * table   /* IO: 2D array of nDetectors x nValues */
)
{
    float value;
    int   d;
    int   i;
    int   nDifferences = 0;

    for (d = 0; d < nDetectors; d++)
    {
        for (i = 0; i < nValues; i++)
        {
            value = band * 10000000 + sca * 100000 + d * 100 + i;
            if (isLookup)
                value /= 100.0;
            else
                value /= 10000000.0;

            if (fillFlag)
                table[d * nValues + i] = value;
            else if (value != table[d * nValues + i])
                nDifferences++;
        }
    }
    if (fillFlag)
        return 0;
    return nDifferences;
}


/*------------------------------------------------------------------------------
NAME: test_benchmark

PURPOSE:
    This function will open a benchmark RLUT and validate that the
    ias_rlut_open_file(), ias_rlut_alloc_lookup_table(),
    ias_rlut_read_lookup_table(), and ias_rlut_close_file() functions are all
    working as expected.
------------------------------------------------------------------------------*/
Ensure test_benchmark ()
{
    IAS_RLUT_LOOKUP_TABLE *table = NULL;
    IAS_RLUT_IO *fd = NULL;
    char path[PATH_MAX];
    int band;
    int format;
    int nDetectors;
    int nDiffs;
    int nScas;
    int sca;
    int totalDiffs = 0;

    /* Test reading both format 2 and 3 files */
    for (format=2; format <= 3; format++)
    {
        /* Get the path to the benchmark file */
        if (util_get_benchmark(format, path) != SUCCESS)
        {
            assert_true_with_message(FALSE,
                "Error obtaining path to benchmark file");
            return;
        }

        /* Open the file for reading */
        printf("Opening benchmark file: %s\n", path);
        fd = ias_rlut_open_file(path, IAS_READ);
        if (fd == NULL)
        {
            assert_true_with_message(FALSE, "Error opening benchmark file %s",
                path);
            return;
        }

        /* For every SCA in each band, read the linearity lookup table */
        printf("Verifying all tables can be successfully read\n");
        for (band = 1; band <= IAS_MAX_NBANDS; band += IAS_MAX_NBANDS - 1)
        {
            nDetectors = ias_sat_attr_get_detectors_per_sca(band);
            nScas = ias_sat_attr_get_scas_per_band(band);

            for (sca = 1; sca <= nScas; sca++)
            {
                /* Read the linearity correction values */
                table = ias_rlut_read_lookup_table(fd,
                    IAS_RLUT_LINEARITY_LOOKUP, band, sca);
                if (table == NULL)
                {
                    assert_true_with_message(FALSE,
                        "Error reading Linearity for band %d, sca %d",
                        band, sca);
                    continue;
                }

                /* Examine both the DN lookup values and the linearit
                   correction values and ensure that there are no differences
                   from what we expect. */
                nDiffs = util_table(FALSE, TRUE, band, sca, nDetectors, NVALUES,
                    table->input_dn);
                assert_equal_with_message(nDiffs, 0,
                    "%d differences in LIN DN lookup for band %d, sca %d",
                    nDiffs, band, sca);
                totalDiffs += nDiffs;

                nDiffs = util_table(FALSE, FALSE, band, sca, nDetectors,
                    NVALUES, table->output_correction);
                assert_equal_with_message(nDiffs, 0,
                    "%d differences in LIN correction for band %d, sca %d",
                    nDiffs, band, sca);
                totalDiffs += nDiffs;

                /* free the table */
                ias_rlut_free_lookup_table(table);
                table = NULL;
            }
        }

        /* Close the file */
        printf("Closing benchmark file: %s\n", path);
        if (ias_rlut_close_file(fd) != SUCCESS)
        {
            assert_true_with_message(FALSE, "Error closing benchmark file %s",
                path);
            return;
        }

        /* Print a message */
        if (totalDiffs == 0)
            printf("Successfully read entire format %d benchmark file\n\n",
                format);
        else
            printf("Found %d mismatched values in format %d benchmark file\n\n",
                totalDiffs, format);
    }

}


/*------------------------------------------------------------------------------
NAME: test_write_read

PURPOSE:
    This function will first write a lookup table to an RLUT file, close the
    file, open the file for reading and then ensure all of the values read
    match the values that were written to the file.
------------------------------------------------------------------------------*/
Ensure test_write_read ()
{
    IAS_RLUT_LOOKUP_TABLE *table = NULL;
    IAS_RLUT_IO *fd = NULL;
    char pathBenchmark[PATH_MAX];
    char pathTemporary[PATH_MAX];
    int band;
    int nDetectors;
    int nScas;
    int sca;
    int status;

    /* Get the path to the benchmark file */
    if (util_get_benchmark(3, pathBenchmark) != SUCCESS)
    {
        assert_true_with_message(FALSE,
            "Error obtaining path to benchmark file");
        return;
    }

    /* Create the path to the temporary file to be created */
    strcpy(pathTemporary, "./tmpRlut2.h5");

    /* Open the file for writing */
    printf("Creating temporary RLUT file: %s\n", pathTemporary);
    fd = ias_rlut_open_file(pathTemporary, IAS_WRITE);
    if (fd == NULL)
    {
        assert_true_with_message(FALSE, "Error opening temporary file %s",
            pathTemporary);
        return;
    }

    /* For every SCA for a couple bands, generate and write the linearity 
       lookup table */
    for (band = 1; band <= IAS_MAX_NBANDS; band += IAS_MAX_NBANDS - 1)
    {
        nDetectors = ias_sat_attr_get_detectors_per_sca(band);
        nScas = ias_sat_attr_get_scas_per_band(band);

        for (sca = 1; sca <= nScas; sca++)
        {
            /* Allocate an lookup table large enough for the current band and
               SCA. */
            table = ias_rlut_alloc_lookup_table(band, sca, nDetectors, NVALUES);
            if (table == NULL)
            {
                assert_true_with_message(FALSE,
                    "Error allocating lookup table");
                if (ias_rlut_close_file(fd) != SUCCESS)
                    assert_true_with_message(FALSE,
                        "Error closing temporary file %s", pathTemporary);
                return;
            }

            /* Create the linearity correction values */
            util_table(TRUE, TRUE, band, sca, nDetectors, NVALUES,
                table->input_dn);
            util_table(TRUE, FALSE, band, sca, nDetectors, NVALUES,
                table->output_correction);

            /* Write these values to the temporary file */
            status = ias_rlut_write_lookup_table(fd, IAS_RLUT_LINEARITY_LOOKUP,
                table);
            assert_equal_with_message(status, 0,
                "Error writing to temporary RLUT file: %s", pathTemporary);

            ias_rlut_free_lookup_table(table);
            table = NULL;
        }
    }

    /* Close the file */
    if (ias_rlut_close_file(fd) != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error closing temporary file %s",
            pathTemporary);
        return;
    }
    printf("Successfully created temporary RLUT file: %s\n", pathTemporary);

    /* Run h5diff on the temporary file just created and the benchmark file */
    printf("Comparing benchmark file to temporary file\n");
    status = util_file_diff(pathBenchmark, pathTemporary);
    if (status == SUCCESS)
        printf("Success: files match\n\n");
    else
    {
        assert_true_with_message(FALSE, "Benchmark and temporary files differ");
        printf("Error: files do not match\n\n");
    }

    /* Most of the file reading has been tested in the test_benchmark test, so
       here just test that attempting to read a band that doesn't exist doesn't
       spew HDF error messages */
    fd = ias_rlut_open_file(pathTemporary, IAS_READ);
    if (fd == NULL)
    {
        assert_true_with_message(FALSE, "Error opening temporary file %s",
            pathTemporary);
        return;
    }

    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    table = ias_rlut_read_lookup_table(fd, IAS_RLUT_LINEARITY_LOOKUP, 2, 1);
    assert_true(table == NULL);
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

    if (ias_rlut_close_file(fd) != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error closing temporary file %s",
            pathTemporary);
        return;
    }
}

int main (int argc, char * argv[])
{
    TestSuite *suite;

    ias_log_initialize("rlut_lookup_table_tests");
    if (ias_sat_attr_initialize(IAS_L8) != SUCCESS)
    {
        assert_true_with_message(FALSE,
            "Error initializing satellite attributes library");
        return EXIT_FAILURE;
    }

    suite = create_test_suite();
    add_test(suite, test_benchmark);
    add_test(suite, test_write_read);

    if (argc > 1)
        return run_single_test(suite, argv[1], create_text_reporter());
    return run_test_suite(suite, create_text_reporter());
}
