#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "oli_model.h"
#include "ias_los_model.h"
#include "ias_model_io.h"
#include "ias_satellite_attributes.h"
#include "model_wedge.h"
#include "calval_band_map.h"

static int calval_version = 2;
/* quick hack to make the calval band map version available.  Will later set
   this via a command line parameter if needed */

int oli_free_model(OLIMODEL *olimodel);

/************************************************************/
/* structures and routines for generating findlos test data */
/************************************************************/
int oli_findlos(
    BandNumber band,    /*I: Input point band number            */
    int sca,            /*I: Input point SCA number (0-based)   */
    double detector,    /*I: Input point detector number        */
    DetType dettype,    /*I: Type of detector model to use      */
    const OLI_SENSOR_MODEL *sensor,/*I: OLI sensor structure    */
    IAS_VECTOR *losv);      /*O: Line of sight vector               */

typedef struct findlos_inputs
{
    double detector;
    DetType dettype;
} FINDLOS_INPUTS;

void test_oli_findlos
(
    OLIMODEL *model
)
{
    int band;
    int sca;
    int output_index;
    FINDLOS_INPUTS inputs[] =
    {
        {1.5, NOMINAL},
        {1.5, ACTUAL},
        {1.5, EXACT},
        {1.5, MAXIMUM},
        {400.5, NOMINAL},
        {400.5, ACTUAL},
        {400.5, EXACT},
        {400.5, MAXIMUM},
    };
    int input_count = sizeof(inputs)/sizeof(FINDLOS_INPUTS);
    IAS_VECTOR output[input_count * NBANDS * NSCA];
    int index;


    /* run a set of points through the findlos routine and dump out an array
       of inputs for easy inclusion in the regression tests */
    printf("    FINDLOS_INPUTS inputs[] = {\n");
    output_index = 0;
    for (band = 0; band < NBANDS; band++)
    {
        int adjusted_band 
            = convert_calval_band_index_to_oli(calval_version, band);

        for (sca = 0; sca < NSCA; sca++)
        {
            for (index = 0; index < input_count; index++)
            {
                FINDLOS_INPUTS *input = &inputs[index];

                oli_findlos(band, sca, input->detector, input->dettype,
                        &model->sensor, &output[output_index]);

                printf("        {%d, %d, %f, %d},\n", adjusted_band, sca, 
                        input->detector, input->dettype);
                output_index++;
            }
        }
    }
    printf("    };\n");

    printf("    IAS_VECTOR outputs[] = {\n");
    output_index = 0;
    for (band = 0; band < NBANDS; band++)
    {
        for (sca = 0; sca < NSCA; sca++)
        {
            for (index = 0; index < input_count; index++)
            {
                IAS_VECTOR *vec = &output[output_index];
                printf("        {%.8e, %.8e, %.8e},\n", vec->x, vec->y, vec->z);
                output_index++;
            }
        }
    }
    printf("    };\n");
}

/*************************************************************/
/* structures and routines for generating findtime test data */
/*************************************************************/
int oli_findtime (
   double line,         /* I: Line from start of the image (0-rel)*/
   double sample,       /* I: Sample from start of the line (0-rel)*/
   BandNumber band,     /* I: Band in the image */
   int sca,             /* I: Current SCA in image */
   OLIMODEL *olimodel,  /* I: OLI model structure */
   DetType dettype,     /* I: Detector option NOMINAL or ACTUAL  */
   double *Time);       /* O: Time from start of image (seconds) */

typedef struct findtime_inputs
{
    double line;
    double sample;
    DetType dettype;
} FINDTIME_INPUTS;

/* routine to generate test data for the find time algorithm */
void test_oli_findtime
(
    OLIMODEL *model
)
{
    int band;
    int sca;
    int output_index;
    FINDTIME_INPUTS inputs[] =
    {
        {0.0, 1.5, NOMINAL},
        {0.0, 1.5, ACTUAL},
        {0.0, 1.5, EXACT},
        {0.0, 1.5, MAXIMUM},
        {200.1, 400.3, NOMINAL},
        {200.1, 400.3, ACTUAL},
        {200.1, 400.3, EXACT},
        {200.1, 400.3, MAXIMUM},
        {200.1, 1000, MAXIMUM},
    };
    int input_count = sizeof(inputs)/sizeof(FINDTIME_INPUTS);
    double output[input_count * NBANDS * NSCA];
    int output_status[input_count * NBANDS * NSCA];
    int index;
    int output_count;
    int items_per_line;

    /* run a set of points through the findlos routine and dump out an array
       of inputs for easy inclusion in the regression tests */
    printf("    FINDTIME_INPUTS inputs[] = {\n");
    output_index = 0;
    for (band = 0; band < NBANDS; band++)
    {
        int adjusted_band 
            = convert_calval_band_index_to_oli(calval_version, band);

        for (sca = 0; sca < NSCA; sca++)
        {
            for (index = 0; index < input_count; index++)
            {
                FINDTIME_INPUTS *input = &inputs[index];

                output_status[output_index] 
                    = oli_findtime(input->line, input->sample, band, sca, model,
                            input->dettype, &output[output_index]);

                printf("        {%d, %d, %f, %f, %d},\n", adjusted_band, sca, 
                        input->line, input->sample, input->dettype);
                output_index++;
            }
        }
    }
    printf("    };\n");

    printf("    double outputs[] = {\n    ");
    output_index = 0;
    output_count = 0;
    items_per_line = 4;
    for (band = 0; band < NBANDS; band++)
    {
        for (sca = 0; sca < NSCA; sca++)
        {
            for (index = 0; index < input_count; index++)
            {
                printf("%.8e, ", output[output_index]);
                output_index++;
                output_count++;
                if (output_count >= items_per_line)
                {
                    printf("\n    ");
                    output_count = 0;
                }
            }
        }
    }
    printf("    };\n");

    printf("    int output_status[] = {\n    ");
    output_index = 0;
    output_count = 0;
    items_per_line = 20;
    for (band = 0; band < NBANDS; band++)
    {
        for (sca = 0; sca < NSCA; sca++)
        {
            for (index = 0; index < input_count; index++)
            {
                printf("%d, ", output_status[output_index]);
                output_index++;
                output_count++;
                if (output_count >= items_per_line)
                {
                    printf("\n    ");
                    output_count = 0;
                }
            }
        }
    }
    printf("    };\n");
}

/************************************************************/
/* structures and routines for generating findatt test data */
/************************************************************/
int oli_findatt(
   OLI_ATTITUDE_MODEL *att, /* I: OLI attitude structure */
   double att_time,     /* I: Actual time (sec from epoch)  */
   double *roll,        /* O: Output roll(rads)         */
   double *pitch,       /* O: Output pitch(rads)    */
   double *yaw);            /* O: Output yaw(rads)          */

void test_oli_findatt
(
    OLIMODEL *model
)
{
    int time_count = 5; /* number of samples to check */
    double start_time = 0.0;
    double end_time = (model->att.samples - 1) * model->att.nominal_samp_time;
    double time_step = (end_time - start_time) / time_count;
    int output_count = time_count + 3;
    double roll[output_count];
    double pitch[output_count];
    double yaw[output_count];
    int status[output_count];
    int index;

    /* run the findatt algorithm on the steps included in the attitude data
       and one on either side */
    printf("    double inputs[] = {\n");
    for (index = 0; index < output_count; index++)
    {
        char buffer[40];
        double input_time = start_time + time_step * (index - 1);
        /* send the input time through an ascii conversion to make sure it is 
           the same as the regression test will use */
        sprintf(buffer, "%.8e", input_time);
        sscanf(buffer, "%le", &input_time);
        roll[index] = 0.0;
        pitch[index] = 0.0;
        yaw[index] = 0.0;
        status[index] = oli_findatt(&model->att, input_time, &roll[index],
                &pitch[index], &yaw[index]);
        printf("        %.8e,\n", input_time);
    }
    printf("    };\n");

    /* output the results */
    printf("    double output_roll[] = {\n");
    for (index = 0; index < output_count; index++)
    {
        printf("        %.8e,\n", roll[index]);
    }
    printf("    };\n");
    printf("    double output_pitch[] = {\n");
    for (index = 0; index < output_count; index++)
    {
        printf("        %.8e,\n", pitch[index]);
    }
    printf("    };\n");
    printf("    double output_yaw[] = {\n");
    for (index = 0; index < output_count; index++)
    {
        printf("        %.8e,\n", yaw[index]);
    }
    printf("    };\n");
    printf("    int output_status[] = {\n");
    for (index = 0; index < output_count; index++)
    {
        printf("        %d,\n", status[index]);
    }
    printf("    };\n");
}

/************************************************************/
/* structures and routines for generating movesat test data */
/************************************************************/
int l8_movesat(
   OLI_EPHEMERIS_MODEL *eph, /* I: OLI ephemeris structure */
   AcqType acq_type,    /* I: Image acquisition type */
   double eph_time, /* I: Delta time from the reference time */
   IAS_VECTOR *satpos,  /* O: New satellite position at "dtime" */
   IAS_VECTOR *satvel); /* O: New satellite velocity at "dtime" */

typedef struct movesat_outputs
{
    IAS_VECTOR satpos;
    IAS_VECTOR satvel;
} MOVESAT_OUTPUTS;

void test_movesat
(
    OLIMODEL *model
)
{
    int time_count = 5; /* number of samples to check */
    double start_time = 0.0;
    double end_time = (model->eph.samples - 1) * model->eph.nominal_samp_time;
    double time_step = (end_time - start_time) / time_count;
    int output_count = time_count + 2;
    MOVESAT_OUTPUTS outputs[output_count * 3];
    int acq_type_map[] = {IAS_EARTH, IAS_STELLAR, IAS_LUNAR};
    AcqType acq_type;
    int index;
    int output_index;

    /* run the movesat algorithm on the steps included in the ephemeris data
       and one on either side */
    output_index = 0;
    printf("    MOVESAT_INPUTS inputs[] = {\n");
    for (acq_type = EARTH_ACQ; acq_type <= LUNAR_ACQ; acq_type++)
    {
        for (index = 0; index < output_count; index++)
        {
            char buffer[40];
            double input_time = start_time + time_step * (index - 1);
            /* send the input time through an ascii conversion to make sure it
               is the same as the regression test will use */
            sprintf(buffer, "%.8e", input_time);
            sscanf(buffer, "%le", &input_time);
            l8_movesat(&model->eph, acq_type, input_time,
                    &outputs[output_index].satpos, 
                    &outputs[output_index].satvel);
            printf("        {%d, %.8e},\n", acq_type_map[acq_type],
                    input_time);
            output_index++;
        }
    }
    printf("    };\n");

    /* output the results */
    output_index = 0;
    printf("    MOVESAT_OUTPUTS outputs[] = {\n");
    for (acq_type = EARTH_ACQ; acq_type <= LUNAR_ACQ; acq_type++)
    {
        for (index = 0; index < output_count; index++)
        {
            MOVESAT_OUTPUTS *output = &outputs[output_index];
            printf("        {{%.8e, %.8e, %.8e},\n"
                    "         {%.8e, %.8e, %.8e}},\n",
                    output->satpos.x, output->satpos.y, output->satpos.z,
                    output->satvel.x, output->satvel.y, output->satvel.z);
            output_index++;
        }
    }
    printf("    };\n");
}

/******************************************************************/
/* structures and routines for generating forward model test data */
/******************************************************************/
int oli_forward_model(
        double line,            /*I: Input line number */
        double sample,          /*I: Input sample number */
        BandNumber band,        /*I: Input band number */
        int sca,                /*I: ALI SCA number */
        double tarelev,         /*I: Target elevation (at line/sample) */
        OLIMODEL *olimodel,     /*I/O: OLI model information */
        DetType dettype,        /*I: Detector type to project
                                     NOMINAL, ACTUAL, or EXACT */
        double *tarlatd,        /*O: Target latitude */
        double *tarlong);       /*O: Target longitude */

typedef struct model_inputs
{
    double line;
    double sample;
    DetType dettype;
} MODEL_INPUTS;

void test_forward_model
(
    OLIMODEL *model
)
{
    double tarelev = 1000.0;
    int band;
    int sca;
    int output_index;
    MODEL_INPUTS inputs[] =
    {
        {0.0, 1.5, NOMINAL},
        {0.0, 1.5, ACTUAL},
        {0.0, 1.5, EXACT},
        {0.5, 1.5, EXACT},
        {0.0, 1.5, MAXIMUM},
        {200.1, 400.3, NOMINAL},
        {200.1, 400.3, ACTUAL},
        {200.1, 400.3, EXACT},
        {201.1, 400.3, EXACT},
        {200.1, 400.3, MAXIMUM},
        {200.1, 1000, MAXIMUM},
    };
    int input_count = sizeof(inputs)/sizeof(MODEL_INPUTS);
    double output_tarlatd[input_count * NBANDS * NSCA];
    double output_tarlong[input_count * NBANDS * NSCA];
    int output_status[input_count * NBANDS * NSCA];
    int index;
    int output_count;
    int items_per_line;

    /* run a set of points through the findlos routine and dump out an array
       of inputs for easy inclusion in the regression tests */
    printf("    MODEL_INPUTS inputs[] = {\n");
    output_index = 0;
    for (band = 0; band < NBANDS; band++)
    {
        int adjusted_band 
            = convert_calval_band_index_to_oli(calval_version, band);

        for (sca = 0; sca < NSCA; sca++)
        {
            for (index = 0; index < input_count; index++)
            {
                MODEL_INPUTS *input = &inputs[index];

                output_status[output_index] 
                    = oli_forward_model(input->line, input->sample, band, sca,
                            tarelev, model, input->dettype,
                            &output_tarlatd[output_index],
                            &output_tarlong[output_index]);

                printf("        {%d, %d, %f, %f, %d},\n", adjusted_band, sca, 
                        input->line, input->sample, input->dettype);
                output_index++;
            }
        }
    }
    printf("    };\n");

    printf("    double output_tarlatd[] = {\n    ");
    output_index = 0;
    output_count = 0;
    items_per_line = 4;
    for (band = 0; band < NBANDS; band++)
    {
        for (sca = 0; sca < NSCA; sca++)
        {
            for (index = 0; index < input_count; index++)
            {
                printf("%.8e, ", output_tarlatd[output_index]);
                output_index++;
                output_count++;
                if (output_count >= items_per_line)
                {
                    printf("\n    ");
                    output_count = 0;
                }
            }
        }
    }
    printf("    };\n");

    printf("    double output_tarlong[] = {\n    ");
    output_index = 0;
    output_count = 0;
    items_per_line = 4;
    for (band = 0; band < NBANDS; band++)
    {
        for (sca = 0; sca < NSCA; sca++)
        {
            for (index = 0; index < input_count; index++)
            {
                printf("%.8e, ", output_tarlong[output_index]);
                output_index++;
                output_count++;
                if (output_count >= items_per_line)
                {
                    printf("\n    ");
                    output_count = 0;
                }
            }
        }
    }
    printf("    };\n");

    printf("    int output_status[] = {\n    ");
    output_index = 0;
    output_count = 0;
    items_per_line = 20;
    for (band = 0; band < NBANDS; band++)
    {
        for (sca = 0; sca < NSCA; sca++)
        {
            for (index = 0; index < input_count; index++)
            {
                printf("%d, ", output_status[output_index]);
                output_index++;
                output_count++;
                if (output_count >= items_per_line)
                {
                    printf("\n    ");
                    output_count = 0;
                }
            }
        }
    }
    printf("    };\n");
}

void test_oli_jitter
(
    OLIMODEL *model
)
{
    int num_jit = model->image.number_image_lines * 2;
    int i;

    printf("    double calval_frame_times[] = {\n    ");
    for (i = 0; i < model->image.number_image_lines; i++)
    {
        printf("%.16e,\n    ", model->image.time_from_epoch[i]);
    }
    printf("};\n");

    printf("    double output[] = {\n    ");
    for (i = 0; i < num_jit; i++)
    {
        printf("%.8e, ", model->image.pan_table[i].x);
        printf("%.8e, ", model->image.pan_table[i].y);
        printf("%.8e,\n    ", model->image.pan_table[i].z);
    }
    printf("};\n");

    printf("    double sat_attitude[] = {\n    ");
    for (i = 0; i < model->att.samples; i++)
    {
        printf("%.8e, ", model->att.records[i].sat_att.x);
        printf("%.8e, ", model->att.records[i].sat_att.y);
        printf("%.8e,\n    ", model->att.records[i].sat_att.z);
    }
    printf("};\n");

    printf("    double prec_attitude[] = {\n    ");
    for (i = 0; i < model->att.samples; i++)
    {
        printf("%.8e, ", model->att.records[i].prec_att.x);
        printf("%.8e, ", model->att.records[i].prec_att.y);
        printf("%.8e,\n    ", model->att.records[i].prec_att.z);
    }
    printf("};\n");
}

/******************************************************************/
/* structures and routines for generating calc position test data */
/******************************************************************/
int oli_calc_position(
        double line,            /*I: Line of interest   (target)        */
        double sample,          /*I: Sample of interest (target)        */
        BandNumber band,        /*I: Band the line/sample are in.       */
        int sca,                /*I: SCA the line/sample are in.        */
        DetType dettype,        /*I: Detector type                      */
        OLIMODEL *olimodel,     /*I: OLI model structure information    */
        IAS_VECTOR *satpos,     /*O: satellite position (meters) in Earth
                                     Centered Inertial (ECI) coordinates*/
        IAS_VECTOR *satvel,     /*O: sat velocity (meters/millisec) in ECI*/
        IAS_VECTOR *satatt,     /*O: sat attitude (roll,pitch,yaw) in radians */
        double *Time,           /*O: time of the target image point     */
        int  *year,             /*O: year of image point                */
        int  *day);             /*O: day of image point                 */

/* Using the same data structure defined for the forward model test */

void test_oli_calc_position
(
    OLIMODEL *model
)
{
    int band;
    int sca;
    int index;
    int output_index;
    MODEL_INPUTS inputs[] =
    {
        {0.0, 1.5, NOMINAL},
        {0.0, 1.5, ACTUAL},
        {0.0, 1.5, EXACT},
        {0.0, 1.5, MAXIMUM},
        {200.1, 400.3, NOMINAL},
        {200.1, 400.3, ACTUAL},
        {200.1, 400.3, EXACT},
        {200.1, 400.3, MAXIMUM},
        {200.1, 1000, MAXIMUM},
    };
    int input_count = sizeof(inputs)/sizeof(MODEL_INPUTS);
    IAS_VECTOR satpos[input_count * NBANDS * NSCA];
    IAS_VECTOR satvel[input_count * NBANDS * NSCA];
    IAS_VECTOR satatt[input_count * NBANDS * NSCA];
    int output_status[input_count * NBANDS * NSCA];
    int output_count;
    int items_per_line;
    double time[input_count * NBANDS * NSCA];
    int year[input_count * NBANDS * NSCA];
    int day[input_count * NBANDS * NSCA];

    printf("    MODEL_INPUTS inputs[] = {\n");
    output_index = 0;
    sca = index = 0;
    for (band = 0; band < NBANDS; band++)
    {
        int adjusted_band 
            = convert_calval_band_index_to_oli(calval_version, band);

        for (sca = 0; sca < NSCA; sca++)
        {
            for (index = 0; index < input_count; index++)
            {
                MODEL_INPUTS *input = &inputs[index];


                output_status[output_index] = oli_calc_position( input->line,
                    input->sample, band, sca, input->dettype, model,
                    &satpos[output_index], &satvel[output_index],
                    &satatt[output_index], &time[output_index],
                    &year[output_index], &day[output_index] );
                printf("        {%d, %d, %f, %f, %d},\n", adjusted_band, sca, 
                                input->line, input->sample, input->dettype);
                output_index++;
            }
        }
    }
    printf("    };\n");

    printf("    IAS_VECTOR satpos[] = {\n");
    output_index = 0;
    output_count = 0;
    items_per_line = 0;
    for (band = 0; band < NBANDS; band++)
    {
        for (sca = 0; sca < NSCA; sca++)
        {
            for (index = 0; index < input_count; index++)
            {
                IAS_VECTOR *vec = &satpos[output_index];
                printf("        {%.8e, %.8e, %.8e},\n", vec->x, vec->y, vec->z);
                output_index++;
            }
        }
    }
    printf("    };\n");

    printf("    IAS_VECTOR satvel[] = {\n");
    output_index = 0;
    output_count = 0;
    items_per_line = 0;
    for (band = 0; band < NBANDS; band++)
    {
        for (sca = 0; sca < NSCA; sca++)
        {
            for (index = 0; index < input_count; index++)
            {
                IAS_VECTOR *vec = &satvel[output_index];
                printf("        {%.8e, %.8e, %.8e},\n", vec->x, vec->y, vec->z);
                output_index++;
            }
        }
    }
    printf("    };\n");

    printf("    IAS_VECTOR satatt[] = {\n");
    output_index = 0;
    output_count = 0;
    items_per_line = 0;
    for (band = 0; band < NBANDS; band++)
    {
        for (sca = 0; sca < NSCA; sca++)
        {
            for (index = 0; index < input_count; index++)
            {
                IAS_VECTOR *vec = &satatt[output_index];
                printf("        {%.8e, %.8e, %.8e},\n", vec->x, vec->y, vec->z);
                output_index++;
            }
        }
    }
    printf("    };\n");

    printf("    double time[] = {\n        ");
    output_index = 0;
    output_count = 0;
    items_per_line = 4;
    for (band = 0; band < NBANDS; band++)
    {
        for (sca = 0; sca < NSCA; sca++)
        {
            for (index = 0; index < input_count; index++)
            {
                printf("%.8e, ", time[output_index]);
                output_index++;
                output_count++;
                if (output_count >= items_per_line)
                {
                    printf("\n        ");
                    output_count = 0;
                }
            }
        }
    }
    printf("    };\n");

    printf("    int year[] = {\n        ");
    output_index = 0;
    output_count = 0;
    items_per_line = 11;
    for (band = 0; band < NBANDS; band++)
    {
        for (sca = 0; sca < NSCA; sca++)
        {
            for (index = 0; index < input_count; index++)
            {
                printf("%d, ", year[output_index]);
                output_index++;
                output_count++;
                if (output_count >= items_per_line)
                {
                    printf("\n        ");
                    output_count = 0;
                }
            }
        }
    }
    printf("    };\n");

    printf("    int day[] = {\n        ");
    output_index = 0;
    output_count = 0;
    items_per_line = 13;
    for (band = 0; band < NBANDS; band++)
    {
        for (sca = 0; sca < NSCA; sca++)
        {
            for (index = 0; index < input_count; index++)
            {
                printf("%d, ", day[output_index]);
                output_index++;
                output_count++;
                if (output_count >= items_per_line)
                {
                    printf("\n        ");
                    output_count = 0;
                }
            }
        }
    }
    printf("    };\n");

    printf("    int output_status[] = {\n        ");
    output_index = 0;
    output_count = 0;
    items_per_line = 20;
    for (band = 0; band < NBANDS; band++)
    {
        for (sca = 0; sca < NSCA; sca++)
        {
            for (index = 0; index < input_count; index++)
            {
                printf("%d, ", output_status[output_index]);
                output_index++;
                output_count++;
                if (output_count >= items_per_line)
                {
                    printf("\n        ");
                    output_count = 0;
                }
            }
        }
    }
    printf("    };\n");
}

/****************************************************************************/
int main(int argc, char *argv[])
{
    const char *input_filename;
    const char *output_filename;
    OLIMODEL old_model;
    IAS_LOS_MODEL *model;
    int calval_version;
  
    if ((argc != 3) && (argc != 4))
    {
        printf("Error: wrong number of parameters, expected:\n"
               "   convert_model [-2] input_model output_model\n"
               "   where -2 is optional for calval band layout #2\n");
        return EXIT_FAILURE;
    }
    if ((argc == 4) && (strcmp(argv[1], "-2") == 0))
    {
        input_filename = argv[2];
        output_filename = argv[3];
        calval_version = 2;
    }
    else
    {
        input_filename = argv[1];
        output_filename = argv[2];
        calval_version = 1;
    }

    printf("Converting model in %s to new model in %s\n", input_filename,
            output_filename);

    ias_sat_attr_initialize(IAS_L8);

    /* read the old model file */
    if (oli_get_model(input_filename, &old_model) != SUCCESS)
    {
        printf("Error reading model from %s\n", input_filename);
        return EXIT_FAILURE;
    }

    model = ias_los_model_wedge_old_model(&old_model, calval_version);
    if (!model)
    {
        printf("Error converting old model to the new model\n");
        return EXIT_FAILURE;
    }

    /* enable the following block to generate test data for the find los 
       algorithm */
#if 0
    test_oli_findlos(&old_model);
#endif

    /* enable the following block to generate test data for the find time 
       algorithm */
#if 0
    test_oli_findtime(&old_model);
#endif

    /* enable the following block to generate test data for the find attitude 
       algorithm */
#if 0
    test_oli_findatt(&old_model);
#endif

    /* enable the following block to generate test data for the movesat
       algorithm */
#if 0
    test_movesat(&old_model);
#endif

    /* enable the following block to generate test data for the forward model */
#if 0
    test_forward_model(&old_model);
#endif
#if 0
    test_oli_calc_position(&old_model);
#endif
#if 0
    test_oli_jitter(&old_model);
#endif

    /* write the new model to a file */
    if (ias_model_write(output_filename, model) != SUCCESS)
    {
        printf("Error writing model file: %s\n", output_filename);
        return EXIT_FAILURE;
    }

    /* free both models */
    ias_los_model_free(model);
    oli_free_model(&old_model);

    return EXIT_SUCCESS;
}
