#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "tirs_model.h"
#include "ias_los_model.h"
#include "ias_model_io.h"
#include "ias_satellite_attributes.h"
#include "model_wedge.h"
#include "calval_band_map.h"

#define TIRS_START_BAND_INDEX 9
#define NTIRS_BANDS 2

int oli_free_model(OLIMODEL *olimodel);

/************************************************************/
/* structures and routines for generating findlos test data */
/************************************************************/
int tirs_findlos(
    BandNumber band,    /*I: Input point band number            */
    int sca,            /*I: Input point SCA number (0-based)   */
    double detector,    /*I: Input point detector number        */
    DetType dettype,    /*I: Type of detector model to use      */
    const OLI_SENSOR_MODEL *sensor,/*I: OLI sensor structure    */
    IAS_VECTOR *losv);      /*O: Line of sight vector               */

typedef struct findlos_inputs
{
    double detector;
    DetType dettype;
} FINDLOS_INPUTS;

void test_tirs_findlos
(
    OLIMODEL *model
)
{
    int band;
    int sca;
    int output_index;
    FINDLOS_INPUTS inputs[] =
    {
        {1.5, NOMINAL},
        {1.5, ACTUAL},
        {1.5, EXACT},
        {1.5, MAXIMUM},
        {400.5, NOMINAL},
        {400.5, ACTUAL},
        {400.5, EXACT},
        {400.5, MAXIMUM},
    };
    int input_count = sizeof(inputs)/sizeof(FINDLOS_INPUTS);
    IAS_VECTOR output[input_count * NTIRS_BANDS * NSCA];
    int index;


    /* run a set of points through the findlos routine and dump out an array
       of inputs for easy inclusion in the regression tests */
    printf("    FINDLOS_INPUTS inputs[] = {\n");
    output_index = 0;
    for (band = 0; band < NTIRS_BANDS; band++)
    {
        int adjusted_band = band + TIRS_START_BAND_INDEX;

        for (sca = 0; sca < NSCA; sca++)
        {
            for (index = 0; index < input_count; index++)
            {
                FINDLOS_INPUTS *input = &inputs[index];

                tirs_findlos(band, sca, input->detector, input->dettype,
                        &model->sensor, &output[output_index]);

                printf("        {%d, %d, %f, %d},\n", adjusted_band, sca, 
                        input->detector, input->dettype);
                output_index++;
            }
        }
    }
    printf("    };\n");

    printf("    IAS_VECTOR outputs[] = {\n");
    output_index = 0;
    for (band = 0; band < NTIRS_BANDS; band++)
    {
        for (sca = 0; sca < NSCA; sca++)
        {
            for (index = 0; index < input_count; index++)
            {
                IAS_VECTOR *vec = &output[output_index];
                printf("        {%.8e, %.8e, %.8e},\n", vec->x, vec->y, vec->z);
                output_index++;
            }
        }
    }
    printf("    };\n");
}

/*************************************************************/
/* structures and routines for generating findtime test data */
/*************************************************************/
int tirs_findtime (
   double line,         /* I: Line from start of the image (0-rel)*/
   double sample,       /* I: Sample from start of the line (0-rel)*/
   BandNumber band,     /* I: Band in the image */
   int sca,             /* I: Current SCA in image */
   OLIMODEL *olimodel,  /* I: OLI model structure */
   DetType dettype,     /* I: Detector option NOMINAL or ACTUAL  */
   double *Time);       /* O: Time from start of image (seconds) */

typedef struct findtime_inputs
{
    double line;
    double sample;
    DetType dettype;
} FINDTIME_INPUTS;

/* routine to generate test data for the find time algorithm */
void test_tirs_findtime
(
    OLIMODEL *model
)
{
    int band;
    int sca;
    int output_index;
    FINDTIME_INPUTS inputs[] =
    {
        {0.0, 1.5, NOMINAL},
        {0.0, 1.5, ACTUAL},
        {0.0, 1.5, EXACT},
        {0.0, 1.5, MAXIMUM},
        {200.1, 400.3, NOMINAL},
        {200.1, 400.3, ACTUAL},
        {200.1, 400.3, EXACT},
        {200.1, 400.3, MAXIMUM},
        {200.1, 1000, MAXIMUM},
    };
    int input_count = sizeof(inputs)/sizeof(FINDTIME_INPUTS);
    double output[input_count * NTIRS_BANDS * NSCA];
    int output_status[input_count * NTIRS_BANDS * NSCA];
    int index;
    int output_count;
    int items_per_line;

    /* run a set of points through the findlos routine and dump out an array
       of inputs for easy inclusion in the regression tests */
    printf("    FINDTIME_INPUTS inputs[] = {\n");
    output_index = 0;
    for (band = 0; band < NTIRS_BANDS; band++)
    {
        int adjusted_band = band + TIRS_START_BAND_INDEX;

        for (sca = 0; sca < NSCA; sca++)
        {
            for (index = 0; index < input_count; index++)
            {
                FINDTIME_INPUTS *input = &inputs[index];

                output_status[output_index] 
                    = tirs_findtime(input->line, input->sample, band, sca,
                            model, input->dettype, &output[output_index]);

                printf("        {%d, %d, %f, %f, %d},\n", adjusted_band, sca, 
                        input->line, input->sample, input->dettype);
                output_index++;
            }
        }
    }
    printf("    };\n");

    printf("    double outputs[] = {\n    ");
    output_index = 0;
    output_count = 0;
    items_per_line = 4;
    for (band = 0; band < NTIRS_BANDS; band++)
    {
        for (sca = 0; sca < NSCA; sca++)
        {
            for (index = 0; index < input_count; index++)
            {
                printf("%.8e, ", output[output_index]);
                output_index++;
                output_count++;
                if (output_count >= items_per_line)
                {
                    printf("\n    ");
                    output_count = 0;
                }
            }
        }
    }
    printf("    };\n");

    printf("    int output_status[] = {\n    ");
    output_index = 0;
    output_count = 0;
    items_per_line = 20;
    for (band = 0; band < NTIRS_BANDS; band++)
    {
        for (sca = 0; sca < NSCA; sca++)
        {
            for (index = 0; index < input_count; index++)
            {
                printf("%d, ", output_status[output_index]);
                output_index++;
                output_count++;
                if (output_count >= items_per_line)
                {
                    printf("\n    ");
                    output_count = 0;
                }
            }
        }
    }
    printf("    };\n");
}

/************************************************************/
/* structures and routines for generating findssm test data */
/************************************************************/
int tirs_findssm(
   TIRS_SCENE_SELECT_MIRROR *tirs_ssm, /* I: TIRS SSM structure */
   double img_time,         /* I: Actual time (sec from image epoch)*/
   IAS_VECTOR *in_los,      /* I: TIRS telescope LOS vector */
   IAS_VECTOR *out_los );   /* O: TIRS SSM LOS vector */

typedef struct findssm_inputs
{
    double time;
    IAS_VECTOR los;
} FINDSSM_INPUTS;

/* routine to generate test data for the find ssm algorithm */
void test_tirs_findssm
(
    OLIMODEL *model
)
{
    int band;
    int sca;
    int output_index;
    FINDTIME_INPUTS inputs[] =
    {
        {0.0, 1.5, NOMINAL},
        {0.0, 1.5, ACTUAL},
        {0.0, 1.5, EXACT},
        {0.0, 1.5, MAXIMUM},
        {200.1, 400.3, NOMINAL},
        {200.1, 400.3, ACTUAL},
        {200.1, 400.3, EXACT},
        {200.1, 400.3, MAXIMUM},
        {200.1, 1000, MAXIMUM},
    };
    int input_count = sizeof(inputs)/sizeof(FINDTIME_INPUTS);
    IAS_VECTOR output[input_count * NTIRS_BANDS * NSCA];
    int output_status[input_count * NTIRS_BANDS * NSCA];
    int status;
    int index;
    int output_count;
    int items_per_line;
    int total;

    printf("typedef struct findssm_inputs\n"
           "{\n"
           "    double time;\n"
           "    IAS_VECTOR los;\n"
           "} FINDSSM_INPUTS;\n\n");

    /* run a set of points through the findlos routine and dump out an array
       of inputs for easy inclusion in the regression tests */
    printf("    FINDSSM_INPUTS inputs[] = {\n");
    output_index = 0;
    for (band = 0; band < NTIRS_BANDS; band++)
    {
        for (sca = 0; sca < NSCA; sca++)
        {
            for (index = 0; index < input_count; index++)
            {
                FINDTIME_INPUTS *input = &inputs[index];
                double time;

                /* find the time */
                status = tirs_findtime(input->line, input->sample, band, sca,
                        model, input->dettype, &time);

                /* if the status was successful, find the los vector */
                if (status == SUCCESS)
                {
                    IAS_VECTOR slos;

                    status = tirs_findlos(band, sca, input->sample,
                            input->dettype, &model->sensor, &slos);

                    if (status == SUCCESS)
                    {
                        /* finally, call findssm */
                        output_status[output_index]
                            = tirs_findssm(&model->tirs_ssm, time, &slos,
                                    &output[output_index]);

                        printf("        {%f, {%.8e, %.8e, %.8e}},\n",
                                time, slos.x, slos.y, slos.z);
                        output_index++;
                    }
                }
            }
        }
    }
    /* add a couple illegal times as input */
    printf("        {%f, {%.8e, %.8e, %.8e}},\n", -2.0, 0.0, 0.0, 0.0);
    printf("        {%f, {%.8e, %.8e, %.8e}},\n", 4000.0, 0.0, 0.0, 0.0);
    output_status[output_index] = ERROR;
    output[output_index].x = 0.0;
    output[output_index].y = 0.0;
    output[output_index].z = 0.0;
    output_index++;
    output_status[output_index] = ERROR;
    output[output_index].x = 0.0;
    output[output_index].y = 0.0;
    output[output_index].z = 0.0;
    output_index++;
    printf("    };\n");

    printf("    IAS_VECTOR outputs[] = {\n    ");
    total = output_index;
    output_count = 0;
    items_per_line = 1;
    for (output_index = 0; output_index < total; output_index++)
    {
        printf("{%.8e, %.8e, %.8e}, ", output[output_index].x,
            output[output_index].y, output[output_index].z);
        output_count++;
        if (output_count >= items_per_line)
        {
            printf("\n    ");
            output_count = 0;
        }
    }
    printf("    };\n");

    printf("    int output_status[] = {\n    ");
    output_index = 0;
    output_count = 0;
    items_per_line = 20;
    for (output_index = 0; output_index < total; output_index++)
    {
        printf("%d, ", output_status[output_index]);
        output_count++;
        if (output_count >= items_per_line)
        {
            printf("\n    ");
            output_count = 0;
        }
    }
    printf("    };\n");
}

/******************************************************************/
/* structures and routines for generating forward model test data */
/******************************************************************/
int oli_forward_model(
        double line,            /*I: Input line number */
        double sample,          /*I: Input sample number */
        BandNumber band,        /*I: Input band number */
        int sca,                /*I: ALI SCA number */
        double tarelev,         /*I: Target elevation (at line/sample) */
        OLIMODEL *olimodel,     /*I/O: OLI model information */
        DetType dettype,        /*I: Detector type to project
                                     NOMINAL, ACTUAL, or EXACT */
        double *tarlatd,        /*O: Target latitude */
        double *tarlong);       /*O: Target longitude */

typedef struct model_inputs
{
    double line;
    double sample;
    DetType dettype;
} MODEL_INPUTS;

void test_forward_model
(
    OLIMODEL *model
)
{
    double tarelev = 1000.0;
    int band;
    int sca;
    int output_index;
    MODEL_INPUTS inputs[] =
    {
        {0.0, 1.5, NOMINAL},
        {0.0, 1.5, ACTUAL},
        {0.0, 1.5, EXACT},
        {0.0, 1.5, MAXIMUM},
        {200.1, 400.3, NOMINAL},
        {200.1, 400.3, ACTUAL},
        {200.1, 400.3, EXACT},
        {200.1, 400.3, MAXIMUM},
        {200.1, 1000, MAXIMUM},
    };
    int input_count = sizeof(inputs)/sizeof(MODEL_INPUTS);
    double output_tarlatd[input_count * NTIRS_BANDS * NSCA];
    double output_tarlong[input_count * NTIRS_BANDS * NSCA];
    int output_status[input_count * NTIRS_BANDS * NSCA];
    int index;
    int output_count;
    int items_per_line;

    /* run a set of points through the findlos routine and dump out an array
       of inputs for easy inclusion in the regression tests */
    printf("    MODEL_INPUTS inputs[] = {\n");
    output_index = 0;
    for (band = 0; band < NTIRS_BANDS; band++)
    {
        int adjusted_band = band + TIRS_START_BAND_INDEX;

        for (sca = 0; sca < NSCA; sca++)
        {
            for (index = 0; index < input_count; index++)
            {
                MODEL_INPUTS *input = &inputs[index];

                output_status[output_index] 
                    = oli_forward_model(input->line, input->sample, band, sca,
                            tarelev, model, input->dettype,
                            &output_tarlatd[output_index],
                            &output_tarlong[output_index]);

                printf("        {%d, %d, %f, %f, %d},\n", adjusted_band, sca, 
                        input->line, input->sample, input->dettype);
                output_index++;
            }
        }
    }
    printf("    };\n");

    printf("    double output_tarlatd[] = {\n    ");
    output_index = 0;
    output_count = 0;
    items_per_line = 4;
    for (band = 0; band < NTIRS_BANDS; band++)
    {
        for (sca = 0; sca < NSCA; sca++)
        {
            for (index = 0; index < input_count; index++)
            {
                printf("%.8e, ", output_tarlatd[output_index]);
                output_index++;
                output_count++;
                if (output_count >= items_per_line)
                {
                    printf("\n    ");
                    output_count = 0;
                }
            }
        }
    }
    printf("    };\n");

    printf("    double output_tarlong[] = {\n    ");
    output_index = 0;
    output_count = 0;
    items_per_line = 4;
    for (band = 0; band < NTIRS_BANDS; band++)
    {
        for (sca = 0; sca < NSCA; sca++)
        {
            for (index = 0; index < input_count; index++)
            {
                printf("%.8e, ", output_tarlong[output_index]);
                output_index++;
                output_count++;
                if (output_count >= items_per_line)
                {
                    printf("\n    ");
                    output_count = 0;
                }
            }
        }
    }
    printf("    };\n");

    printf("    int output_status[] = {\n    ");
    output_index = 0;
    output_count = 0;
    items_per_line = 20;
    for (band = 0; band < NTIRS_BANDS; band++)
    {
        for (sca = 0; sca < NSCA; sca++)
        {
            for (index = 0; index < input_count; index++)
            {
                printf("%d, ", output_status[output_index]);
                output_index++;
                output_count++;
                if (output_count >= items_per_line)
                {
                    printf("\n    ");
                    output_count = 0;
                }
            }
        }
    }
    printf("    };\n");
}

void test_tirs_jitter
(
    OLIMODEL *model
)
{
    int i;

    printf("    double calval_frame_times[] = {\n    ");
    for (i = 0; i < model->image.number_image_lines; i++)
    {
        printf("%.16e,\n    ", model->image.time_from_epoch[i]);
    }
    printf("};\n");

    printf("    double output[] = {\n    ");
    for (i = 0; i < model->image.number_image_lines; i++)
    {
        printf("%.8e, ", model->image.jit_table[i].x);
        printf("%.8e, ", model->image.jit_table[i].y);
        printf("%.8e,\n    ", model->image.jit_table[i].z);
    }
    printf("};\n");

    printf("    double sat_attitude[] = {\n    ");
    for (i = 0; i < model->att.samples; i++)
    {
        printf("%.8e, ", model->att.records[i].sat_att.x);
        printf("%.8e, ", model->att.records[i].sat_att.y);
        printf("%.8e,\n    ", model->att.records[i].sat_att.z);
    }
    printf("};\n");

    printf("    double prec_attitude[] = {\n    ");
    for (i = 0; i < model->att.samples; i++)
    {
        printf("%.8e, ", model->att.records[i].prec_att.x);
        printf("%.8e, ", model->att.records[i].prec_att.y);
        printf("%.8e,\n    ", model->att.records[i].prec_att.z);
    }
    printf("};\n");
}

/****************************************************************************/
int tirs_smooth_ssm_data(
   TIRS_SCENE_SELECT_MIRROR *tirs_ssm); /* I/O: TIRS SSM model group */

/* helper routine to dump out the SSM structures in terms of our definition */
void print_ssm_info(TIRS_SCENE_SELECT_MIRROR *tirs_ssm, const char *name)
{
    int index;

    printf("IAS_SENSOR_SSM_RECORD %s_records[%d];\n", name,
            tirs_ssm->number_encoder_angles);
    printf("IAS_SENSOR_SCENE_SELECT_MIRROR_MODEL %s_ssm_model = \n", name);
    printf("    {{{%f, %f, %f},\n    {%f, %f, %f},\n"
           "    {%f, %f, %f}},\n", tirs_ssm->tele2ssm[0][0],
           tirs_ssm->tele2ssm[0][1], tirs_ssm->tele2ssm[0][2],
           tirs_ssm->tele2ssm[1][0], tirs_ssm->tele2ssm[1][1],
           tirs_ssm->tele2ssm[1][2], tirs_ssm->tele2ssm[2][0],
           tirs_ssm->tele2ssm[2][1], tirs_ssm->tele2ssm[2][2]);
    printf("    {0.0, 0.0, 0.0},\n");
    printf("     %d, NULL};\n", tirs_ssm->number_encoder_angles);
    printf("%s_ssm_model.records = %s_records;\n", name, name);

    for (index = 0; index < tirs_ssm->number_encoder_angles; index++)
    {
        printf("%s_records[%d].seconds_from_epoch = %.8e;\n", name, index,
               tirs_ssm->records[index].time_from_epoch);
        printf("%s_records[%d].mirror_angle = %.8e;\n", name, index,
               tirs_ssm->records[index].ssm_angle);
    }

    printf("\n");
}

#define SSM_SMOOTH_RECORDS 300

void test_ssm_smoothing
(
    OLIMODEL *model
)
{
    TIRS_SCENE_SELECT_MIRROR tirs_ssm;
    TIRS_SSM_RECORD records[SSM_SMOOTH_RECORDS];
    int index;
    int status;

    /* build a set of input data */
    memset(&tirs_ssm, 0, sizeof(tirs_ssm));

    tirs_ssm.number_encoder_angles = SSM_SMOOTH_RECORDS;
    tirs_ssm.records = records;

    for (index = 0; index < SSM_SMOOTH_RECORDS; index++)
    {
        double jitter;

        records[index].time_from_epoch = index * 1.0/20.0;

        /* make the ssm angles with a little jitter */
        if (index & 0x01)
            jitter = 0.01;
        else
            jitter = -0.01;

        records[index].ssm_angle = 0.1 * index + jitter;
    }

    print_ssm_info(&tirs_ssm, "input");

    /* run it through the smoothing function */
    status = tirs_smooth_ssm_data(&tirs_ssm);
    if (status != SUCCESS)
    {
        printf("Error calling tirs_smooth_ssm_data\n");
        exit(-1);
    }

    /* output the expected results */
    print_ssm_info(&tirs_ssm, "expected");

}

/****************************************************************************/
typedef struct
{
    uint16_t ssm_epoch_days;
    uint32_t ssm_milliseconds;
    uint16_t ssm_microseconds;
    uint32_t ssm_1_encoder_position;
    uint32_t ssm_2_encoder_position;
    uint32_t ssm_3_encoder_position;
    uint32_t ssm_4_encoder_position;
}OLI_L0R_TIRS_SSM;

int tirs_check_ssm_data(
   IAS_L0R_TIRS_SSM_ENCODER *l0r_tirs_ssm,/* I/O: L0ra SSM telemetry records */
   int count,                             /* I: Number of SSM records in L0Ra */
   void *cpf);                            /*I: Calibration parameter file */

/* Allocate and fill a set of ssm encoder records */
#define SSM_RECORDS 553
IAS_L0R_TIRS_SSM_ENCODER *fill_ssm_test_data()
{
    IAS_L0R_TIRS_SSM_ENCODER *ssm_data;
    int index;

    ssm_data = malloc(SSM_RECORDS * sizeof(*ssm_data));
    if (!ssm_data)
    {
        printf("Error allocating memory for ssm_data\n");
        exit(-1);
    }
    for (index = 0; index < SSM_RECORDS; index++)
    {
        int pos;

        ssm_data[index].l0r_time.days_from_J2000 = 1243;
        ssm_data[index].l0r_time.seconds_of_day = 21996.285637 + index;

        for (pos = 0; pos < IAS_L0R_TIRS_SSM_ENCODER_SAMPLES; pos++)
            ssm_data[index].encoder_position[pos] = 16772555;
    }

    /* insert some errors into the data (matching the cal/val prototype data) */
    ssm_data[26].l0r_time.seconds_of_day = 21497.093;
    ssm_data[6].encoder_position[0] = 16674251;
    ssm_data[33].encoder_position[2] = 16772491;

    /* add a few more errors since calval wasn't exercising all the code */
    /* make sure the code to correct the first time codes gets run */
    ssm_data[1].l0r_time.days_from_J2000 = 1242;

    /* make sure the code to correct the first SSM positions gets run */
    ssm_data[1].encoder_position[0] = 16772455;

    return ssm_data;
}

/* helper routine to dump out the SSM structures in terms of our definition */
void print_ssm_l0r_records
(
    IAS_L0R_TIRS_SSM_ENCODER *ssm_data,
    int count,
    const char *name
)
{
    int index;

    printf("IAS_L0R_TIRS_SSM_ENCODER %s_records[%d];\n", name, SSM_RECORDS);

    for (index = 0; index < count; index++)
    {
        int pos;

        printf("%s_records[%d].l0r_time.days_from_J2000 = %d;\n", name, index,
               ssm_data[index].l0r_time.days_from_J2000);
        printf("%s_records[%d].l0r_time.seconds_of_day = %f;\n", name, index,
               ssm_data[index].l0r_time.seconds_of_day);
        for (pos = 0; pos < IAS_L0R_TIRS_SSM_ENCODER_SAMPLES; pos++)
        {
            printf("%s_records[%d].encoder_position[%d] = %d;\n", name, index,
                   pos, ssm_data[index].encoder_position[pos]);
        }
    }

    printf("\n");
}

void test_ssm_validation
(
    OLIMODEL *model
)
{
    IAS_L0R_TIRS_SSM_ENCODER *ssm_data;
    int status;

    /* fill the L0R SSM structure with the values from the calval prototype
       test data delivery */
    ssm_data = fill_ssm_test_data();

    /* output the original data as the input data for the test */
    print_ssm_l0r_records(ssm_data, SSM_RECORDS, "input");

    /* output the test data for use in the test code */
    status = tirs_check_ssm_data(ssm_data, SSM_RECORDS, NULL);
    if (status != SUCCESS)
    {
        printf("Error calling tirs_check_ssm_data\n");
        exit(-1);
    }

    /* output the results obtained */
    print_ssm_l0r_records(ssm_data, SSM_RECORDS, "expected");
}

/****************************************************************************/
/* routine to generate test data for the ias_sensor_set_ssm_from_l0r routine.
   Requires a hacked version of calval's tirs_get_ssm_from_l0r. */

int tirs_get_ssm_from_l0r(
   char *l0ra_root,        /* I: L0ra ancillary file name */
   char *l0ra_dir,         /* I: L0ra ancillary file name */
   void *cpf,            /*I: Calibration parameter file */
   OLIMODEL *olimodel,     /*I/O: OLI model information */
   IAS_L0R_TIRS_SSM_ENCODER **returned_records,
   int *returned_record_count
);

void test_set_ssm_from_l0r
(
    OLIMODEL *model
)
{
    int status;
    IAS_L0R_TIRS_SSM_ENCODER *returned_records;
    int record_count;

    /* call the calval code to do the equivalent of our set_ssm_from_l0r.
       calval code actually reads the L0R, so I hacked a temporary copy of
       their routine to allow it to return the ssm data read from the L0R.
       The copy of the routine was also hacked to not require reading a CPF
       since including that header here doesn't work without a lot of pain. 
       Hard code the name of the L0R delivered by calval with the ADD. */
    status = tirs_get_ssm_from_l0r(
        "LO80400330332000227EDC00",
        "/edclxs110/data1/tbeckman/add_testing/tirs_jitter_3_2/orig",
        NULL, model, &returned_records, &record_count);
    if (status != SUCCESS)
    {
        printf("Error calling tirs_get_ssm_from_l0r\n");
        exit(-1);
    }

    /* write out the L0R ssm records as input for our test */
    print_ssm_l0r_records(returned_records, record_count, "input");

    /* write out the epoch time */
    printf("double input_epoch[3] = {%f, %f, %f};\n",
        model->image.utc_epoch_time[0],
        model->image.utc_epoch_time[1],
        model->image.utc_epoch_time[2]);
    printf("double frame_times[2] = {%.8e, %.8e};\n",
        model->image.time_from_epoch[0],
        model->image.time_from_epoch[model->image.number_image_lines-1]);

    /* write out the important values from the model as the expected results */
    print_ssm_info(&model->tirs_ssm, "expected");
}

/****************************************************************************/
int main(int argc, char *argv[])
{
    const char *input_filename;
    const char *output_filename;
    OLIMODEL old_model;
    IAS_LOS_MODEL *model;
    int calval_version;
  
    if ((argc != 3) && (argc != 4))
    {
        printf("Error: wrong number of parameters, expected:\n"
               "   convert_model [-2] input_model output_model\n"
               "   where -2 is optional for calval band layout #2\n");
        return EXIT_FAILURE;
    }
    if ((argc == 4) && (strcmp(argv[1], "-2") == 0))
    {
        input_filename = argv[2];
        output_filename = argv[3];
        calval_version = 2;
    }
    else
    {
        input_filename = argv[1];
        output_filename = argv[2];
        calval_version = 1;
    }

    printf("Converting model in %s to new model in %s\n", input_filename,
            output_filename);

    ias_sat_attr_initialize(IAS_L8);

    /* read the old model file */
    if (oli_get_model(input_filename, &old_model) != SUCCESS)
    {
        printf("Error reading model from %s\n", input_filename);
        return EXIT_FAILURE;
    }

    model = ias_los_model_wedge_old_model(&old_model, calval_version);
    if (!model)
    {
        printf("Error converting old model to the new model\n");
        return EXIT_FAILURE;
    }

    /* enable the following block to generate test data for the find los 
       algorithm */
#if 0
    test_tirs_findlos(&old_model);
#endif

    /* enable the following block to generate test data for the find time 
       algorithm */
#if 0
    test_tirs_findtime(&old_model);
#endif

    /* enable the following block to generate test data for the find ssm 
       algorithm */
#if 0
    test_tirs_findssm(&old_model);
#endif

    /* enable the following block to generate test data for the forward model */
#if 0
    test_forward_model(&old_model);
#endif

    /* enable the following block to generate test data for the jitter table */
#if 0
    test_tirs_jitter(&old_model);
#endif

    /* enable the following block to generate test data for the ssm smoothing */
#if 0
    test_ssm_smoothing(&old_model);
#endif

    /* enable the following block to generate test data for the ssm
       validation */
#if 0
    test_ssm_validation(&old_model);
#endif

    /* enable the following block to generate test data for set_ssm_from_l0r */
#if 0
    test_set_ssm_from_l0r(&old_model);
#endif


    /* write the new model to a file */
    if (ias_model_write(output_filename, model) != SUCCESS)
    {
        printf("Error writing model file: %s\n", output_filename);
        return EXIT_FAILURE;
    }

    /* free both models */
    ias_los_model_free(model);
    oli_free_model(&old_model);

    return EXIT_SUCCESS;
}
