/* This is a throw away module to convert the old model structure that is still
   used by the cal/val prototype into the new model structure */

#include <stdlib.h>
#include <string.h>
#include "ias_logging.h"
#include "ias_satellite_attributes.h"
#include "ias_los_model.h"
#include "oli_model.h"
#include "model_wedge.h"
#include "calval_band_map.h"

/* returns a pointer to the converted model, or NULL on error */
IAS_LOS_MODEL *ias_los_model_wedge_old_model
(
    const OLIMODEL *old_model,
    int calval_version
)
{
    IAS_LOS_MODEL *model = NULL;
    IAS_SC_ATTITUDE_MODEL *attitude;
    IAS_SC_EPHEMERIS_MODEL *ephemeris;
    IAS_SENSOR_MODEL *sensor;
    int calval_band_index;
    int sca_index;
    int detector_index;
    int line;

    /* allocate a new model structure */
    model = ias_los_model_allocate();
    if (!model)
    {
        IAS_LOG_ERROR("Error allocating model structure");
        return NULL;
    }
    attitude = &model->spacecraft.attitude;
    ephemeris = &model->spacecraft.ephemeris;

    /* copy some basic info */
    model->satellite_number = old_model->satnum;
    model->wrs_path = old_model->wrs_path;
    model->wrs_row = old_model->wrs_row;
    switch (old_model->acq_type)
    {
        case EARTH_ACQ:
            model->acquisition_type = IAS_EARTH;
            break;
        case STELLAR_ACQ:
            model->acquisition_type = IAS_STELLAR;
            break;
        case LUNAR_ACQ:
            model->acquisition_type = IAS_LUNAR;
            break;
        case dacquis:
            model->acquisition_type = IAS_ACQUISITION_TYPE_ERROR;
            break;
    }
    /* no apparent matching field in the old model, so assume systematic */
    model->correction_type = IAS_SYSTEMATIC;
    /* hard-code the satellite ID and version information */
    model->satellite_id = IAS_L8;
    strcpy(model->sw_version, "1.0");

    /* copy the earth characteristics */
    model->earth.ut1_utc_correction = old_model->earth.ut1_utc_correction;
    model->earth.pole_wander_x = old_model->earth.pole_wander_x;
    model->earth.pole_wander_y = old_model->earth.pole_wander_y;
    model->earth.semi_major_axis = old_model->earth.semi_major;
    model->earth.semi_minor_axis = old_model->earth.semi_minor;
    model->earth.eccentricity = old_model->earth.eccentricity;
    model->earth.earth_angular_velocity
        = old_model->earth.earth_angular_velocity;
    model->earth.speed_of_light = old_model->earth.speed_of_light;

    /* copy the attitude info */
    memcpy(attitude->utc_epoch_time, old_model->att.utc_epoch_time,
            sizeof(attitude->utc_epoch_time));
    attitude->nominal_sample_time = old_model->att.nominal_samp_time;
    attitude->sample_count = old_model->att.samples;
    attitude->sample_records = malloc(attitude->sample_count
            * sizeof(IAS_SC_ATTITUDE_RECORD));
    if (!attitude->sample_records)
    {
        IAS_LOG_ERROR("Error allocating attitude sample records");
        free(model);
        return NULL;
    }
    /* WARNING - since this is throw away code, cheat a bit here by just doing
       a memcpy.  Even though the structure has different names for the
       members, the size and order agree */
    memcpy(attitude->sample_records, old_model->att.records, 
            attitude->sample_count * sizeof(IAS_SC_ATTITUDE_RECORD));

    /* copy the ephemeris info */
    memcpy(ephemeris->utc_epoch_time, old_model->eph.utc_epoch_time,
            sizeof(ephemeris->utc_epoch_time));
    ephemeris->nominal_sample_time = old_model->eph.nominal_samp_time;
    ephemeris->sample_count = old_model->eph.samples;
    ephemeris->sample_records 
        = malloc(ephemeris->sample_count * sizeof(IAS_SC_EPHEMERIS_RECORD));
    if (!ephemeris->sample_records)
    {
        IAS_LOG_ERROR("Error allocating ephemeris sample records");
        free(model);
        return NULL;
    }
    /* WARNING - since this is throw away code, cheat a bit here by just doing
       a memcpy.  Even though the structure has different names for the
       members, the size and order agree */
    memcpy(ephemeris->sample_records, old_model->eph.records, 
            ephemeris->sample_count * sizeof(IAS_SC_EPHEMERIS_RECORD));

    /* WARNING - since this is throw away code, cheat a bit here by just doing
       a memcpy.  Even though the structure has different names for the
       members, the size and order agree */
    memcpy(&model->spacecraft.precision, &old_model->prec, 
            sizeof(IAS_SC_PRECISION_MODEL));

    /* copy the sensor model info */
    sensor = &model->sensor;
    memcpy(sensor->sensors[IAS_OLI].sensor2acs, old_model->sensor.oli2acs,
            sizeof(sensor->sensors[IAS_OLI].sensor2acs));
    memcpy(sensor->sensors[IAS_OLI].center_mass2sensor_offset,
            old_model->sensor.cm2oli,
            sizeof(sensor->sensors[IAS_OLI].center_mass2sensor_offset));
    sensor->sensors[IAS_OLI].jitter_entries_per_frame = 2;
    sensor->sensors[IAS_OLI].jitter_table_count
        = old_model->image.number_image_lines
        * sensor->sensors[IAS_OLI].jitter_entries_per_frame;
    sensor->sensors[IAS_OLI].jitter_table
        = (IAS_VECTOR *)malloc(sensor->sensors[IAS_OLI].jitter_table_count
                * sizeof(IAS_VECTOR));
    if (!sensor->sensors[IAS_OLI].jitter_table)
    {
        IAS_LOG_ERROR("Error allocating jitter model table");
        free(model);
        return NULL;
    }
    for (line = 0; line < sensor->sensors[IAS_OLI].jitter_table_count; line++)
    {
        sensor->sensors[IAS_OLI].jitter_table[line].x
            = old_model->image.pan_table[line].x;
        sensor->sensors[IAS_OLI].jitter_table[line].y
            = old_model->image.pan_table[line].y;
        sensor->sensors[IAS_OLI].jitter_table[line].z
            = old_model->image.pan_table[line].z;
    }

    /* copy the band info into the new structure */
    for (calval_band_index = 0; calval_band_index < NBANDS; calval_band_index++)
    {
        IAS_SENSOR_BAND_MODEL *band;
        int oli_band_index;

        oli_band_index = convert_calval_band_index_to_oli(calval_version,
                calval_band_index);

        band = &sensor->bands[oli_band_index];

        /* copy the sampling characteristics */
        if (oli_band_index == 7)
        {
            band->sampling_char.integration_time
                = old_model->sensor.int_times[0];
            band->sampling_char.sampling_time = old_model->sensor.samp_times[0];
            band->sampling_char.along_ifov = old_model->sensor.along_ifov[0];
            band->sampling_char.across_ifov = old_model->sensor.across_ifov[0];
            band->sampling_char.lines_per_frame = 2;
            band->sampling_char.settling_time 
                = old_model->sensor.settle_times[0];
        }
        else
        {
            band->sampling_char.integration_time
                = old_model->sensor.int_times[1];
            band->sampling_char.sampling_time = old_model->sensor.samp_times[1];
            band->sampling_char.along_ifov = old_model->sensor.along_ifov[1];
            band->sampling_char.across_ifov = old_model->sensor.across_ifov[1];
            band->sampling_char.lines_per_frame = 1;
            band->sampling_char.settling_time 
                = old_model->sensor.settle_times[1];
        }
        /* OLI time codes are at the end of the frame */
        band->sampling_char.time_codes_at_frame_start = 0;
        band->sampling_char.maximum_detector_delay = 10.0;
        band->sampling_char.frame_delay = 1;

        for (sca_index = 0; sca_index < NSCA; sca_index++)
        {
            IAS_SENSOR_SCA_MODEL *sca = &band->scas[sca_index];

            /* copy the nominal fill */
            sca->nominal_fill 
                = old_model->image.nominal_fill[calval_band_index][sca_index];

            /* copy the legendre coefficients */
            sca->sca_coef_x[0] 
                = old_model->sensor.sca_coef_x[sca_index][calval_band_index][0];
            sca->sca_coef_x[1] 
                = old_model->sensor.sca_coef_x[sca_index][calval_band_index][1];
            sca->sca_coef_x[2] 
                = old_model->sensor.sca_coef_x[sca_index][calval_band_index][2];
            sca->sca_coef_x[3] = 0.0; 
            sca->sca_coef_y[0] 
                = old_model->sensor.sca_coef_y[sca_index][calval_band_index][0];
            sca->sca_coef_y[1] 
                = old_model->sensor.sca_coef_y[sca_index][calval_band_index][1];
            sca->sca_coef_y[2] 
                = old_model->sensor.sca_coef_y[sca_index][calval_band_index][2];
            sca->sca_coef_y[3] = 0.0;

            /* copy the detector offsets and fill */
            if (oli_band_index == 7)
            {
                /* pan band */
                for (detector_index = 0; detector_index < sca->detectors;
                        detector_index++)
                {
                    sca->detector_offsets_across_track[detector_index]
                        = old_model->sensor.det_delays_pan
                        [sca_index][0][detector_index];
                    sca->detector_offsets_along_track[detector_index]
                        = old_model->sensor.det_delays_pan
                        [sca_index][1][detector_index];
                    sca->l0r_detector_offsets[detector_index]
                        = old_model->image.l0r_fill_pan
                        [sca_index][detector_index];
                }
            }
            else if (oli_band_index <= 8)
            {
                /* MS bands */
                for (detector_index = 0; detector_index < sca->detectors;
                        detector_index++)
                {
                    sca->detector_offsets_across_track[detector_index]
                        = old_model->sensor.det_delays_ms
                        [sca_index][calval_band_index-1][0][detector_index];
                    sca->detector_offsets_along_track[detector_index]
                        = old_model->sensor.det_delays_ms
                        [sca_index][calval_band_index-1][1][detector_index];
                    sca->l0r_detector_offsets[detector_index]
                        = old_model->image.l0r_fill_ms
                        [sca_index][calval_band_index-1][detector_index];
                }
            }
            else
            {
                /* must have gotten the indexes wrong */
                IAS_LOG_ERROR("Unexpected band index: %d", calval_band_index);
                exit(-1);
            }
        }

        /* get epoch */
        band->frame_count = old_model->image.number_image_lines;
        band->utc_epoch_time[0] = old_model->image.utc_epoch_time[0];
        band->utc_epoch_time[1] = old_model->image.utc_epoch_time[1];
        band->utc_epoch_time[2] = old_model->image.utc_epoch_time[2];
        band->band_present = 1;
    }

    double *frame_seconds_from_epoch
        = malloc(old_model->image.number_image_lines * sizeof(double));
    if (!frame_seconds_from_epoch)
    {
        IAS_LOG_ERROR("Error allocating line times");
        free(model);
        return NULL;
    }

    memcpy(frame_seconds_from_epoch, old_model->image.time_from_epoch,
            sizeof(double) * old_model->image.number_image_lines);
    if (ias_sensor_set_frame_times(&model->sensor, IAS_OLI,
                frame_seconds_from_epoch, old_model->image.number_image_lines)
            != SUCCESS)
    {
        IAS_LOG_ERROR("Error setting line times");
        exit(-1);
    }
    sensor->sensors[IAS_OLI].sensor_present = 1;
    sensor->sensors[IAS_TIRS].sensor_present = 0;

    return model;
}
