/* This is a throw away module to convert the old model structure that is still
   used by the cal/val prototype into the new model structure */

#include <stdlib.h>
#include <string.h>
#include "ias_logging.h"
#include "ias_satellite_attributes.h"
#include "ias_los_model.h"
#include "tirs_model.h"
#include "model_wedge.h"
#include "calval_band_map.h"

/* returns a pointer to the converted model, or NULL on error */
IAS_LOS_MODEL *ias_los_model_wedge_old_model
(
    const OLIMODEL *old_model,
    int calval_version
)
{
    IAS_LOS_MODEL *model = NULL;
    IAS_SC_ATTITUDE_MODEL *attitude;
    IAS_SC_EPHEMERIS_MODEL *ephemeris;
    IAS_SENSOR_MODEL *sensor;
    int calval_band_index;
    int sca_index;
    int detector_index;
    IAS_SENSOR_SCENE_SELECT_MIRROR_MODEL *ssm;
    int index;
    int line;

    /* allocate a new model structure */
    model = ias_los_model_allocate();
    if (!model)
    {
        IAS_LOG_ERROR("Error allocating model structure");
        return NULL;
    }
    attitude = &model->spacecraft.attitude;
    ephemeris = &model->spacecraft.ephemeris;

    /* copy some basic info */
    model->satellite_number = old_model->satnum;
    model->wrs_path = old_model->wrs_path;
    model->wrs_row = old_model->wrs_row;
    switch (old_model->acq_type)
    {
        case EARTH_ACQ:
            model->acquisition_type = IAS_EARTH;
            break;
        case STELLAR_ACQ:
            model->acquisition_type = IAS_STELLAR;
            break;
        case LUNAR_ACQ:
            model->acquisition_type = IAS_LUNAR;
            break;
        case dacquis:
            model->acquisition_type = IAS_ACQUISITION_TYPE_ERROR;
            break;
    }
    /* no apparent matching field in the old model, so assume systematic */
    model->correction_type = IAS_SYSTEMATIC;
    /* hard-code the satellite ID and version information */
    model->satellite_id = IAS_L8;
    strcpy(model->sw_version, "1.0");

    /* copy the earth characteristics */
    model->earth.ut1_utc_correction = old_model->earth.ut1_utc_correction;
    model->earth.pole_wander_x = old_model->earth.pole_wander_x;
    model->earth.pole_wander_y = old_model->earth.pole_wander_y;
    model->earth.semi_major_axis = old_model->earth.semi_major;
    model->earth.semi_minor_axis = old_model->earth.semi_minor;
    model->earth.eccentricity = old_model->earth.eccentricity;
    model->earth.earth_angular_velocity
        = old_model->earth.earth_angular_velocity;
    model->earth.speed_of_light = old_model->earth.speed_of_light;

    /* copy the attitude info */
    memcpy(attitude->utc_epoch_time, old_model->att.utc_epoch_time,
            sizeof(attitude->utc_epoch_time));
    attitude->nominal_sample_time = old_model->att.nominal_samp_time;
    attitude->sample_count = old_model->att.samples;
    attitude->sample_records = malloc(attitude->sample_count
            * sizeof(IAS_SC_ATTITUDE_RECORD));
    if (!attitude->sample_records)
    {
        IAS_LOG_ERROR("Error allocating attitude sample records");
        free(model);
        return NULL;
    }
    /* WARNING - since this is throw away code, cheat a bit here by just doing
       a memcpy.  Even though the structure has different names for the
       members, the size and order agree */
    memcpy(attitude->sample_records, old_model->att.records, 
            attitude->sample_count * sizeof(IAS_SC_ATTITUDE_RECORD));

    /* copy the ephemeris info */
    memcpy(ephemeris->utc_epoch_time, old_model->eph.utc_epoch_time,
            sizeof(ephemeris->utc_epoch_time));
    ephemeris->nominal_sample_time = old_model->eph.nominal_samp_time;
    ephemeris->sample_count = old_model->eph.samples;
    ephemeris->sample_records 
        = malloc(ephemeris->sample_count * sizeof(IAS_SC_EPHEMERIS_RECORD));
    if (!ephemeris->sample_records)
    {
        IAS_LOG_ERROR("Error allocating ephemeris sample records");
        free(model);
        return NULL;
    }
    /* WARNING - since this is throw away code, cheat a bit here by just doing
       a memcpy.  Even though the structure has different names for the
       members, the size and order agree */
    memcpy(ephemeris->sample_records, old_model->eph.records, 
            ephemeris->sample_count * sizeof(IAS_SC_EPHEMERIS_RECORD));

    /* WARNING - since this is throw away code, cheat a bit here by just doing
       a memcpy.  Even though the structure has different names for the
       members, the size and order agree */
    memcpy(&model->spacecraft.precision, &old_model->prec, 
            sizeof(IAS_SC_PRECISION_MODEL));

    /* copy the sensor model info */
    sensor = &model->sensor;
    memcpy(sensor->sensors[IAS_TIRS].sensor2acs, old_model->sensor.oli2acs,
            sizeof(sensor->sensors[IAS_TIRS].sensor2acs));
    memcpy(sensor->sensors[IAS_TIRS].center_mass2sensor_offset,
            old_model->sensor.cm2oli,
            sizeof(sensor->sensors[IAS_TIRS].center_mass2sensor_offset));
    sensor->sensors[IAS_TIRS].jitter_entries_per_frame = 1;
    sensor->sensors[IAS_TIRS].jitter_table_count
        = old_model->image.number_image_lines;
    sensor->sensors[IAS_TIRS].jitter_table
        = (IAS_VECTOR *)malloc(sensor->sensors[IAS_TIRS].jitter_table_count
                * sizeof(IAS_VECTOR));
    if (!sensor->sensors[IAS_TIRS].jitter_table)
    {
        IAS_LOG_ERROR("Error allocating jitter model table");
        free(model);
        return NULL;
    }
    for (line = 0; line < sensor->sensors[IAS_TIRS].jitter_table_count; line++)
    {
        sensor->sensors[IAS_TIRS].jitter_table[line].x
            = old_model->image.jit_table[line].x;
        sensor->sensors[IAS_TIRS].jitter_table[line].y
            = old_model->image.jit_table[line].y;
        sensor->sensors[IAS_TIRS].jitter_table[line].z
            = old_model->image.jit_table[line].z;
    }

    /* copy the band info into the new structure.  Only copy the first two
       bands to bands 10 and 11 */
    for (calval_band_index = 0; calval_band_index <= 1; calval_band_index++)
    {
        IAS_SENSOR_BAND_MODEL *band;    /* output band */
        int tirs_band_index;

        tirs_band_index = calval_band_index + 9;
        band = &sensor->bands[tirs_band_index];

        /* copy the sampling characteristics */
        band->sampling_char.integration_time
            = old_model->sensor.int_times[1];
        band->sampling_char.sampling_time = old_model->sensor.samp_times[1];
        band->sampling_char.along_ifov = old_model->sensor.along_ifov[0];
        band->sampling_char.across_ifov = old_model->sensor.across_ifov[0];
        band->sampling_char.lines_per_frame = 1;
        band->sampling_char.settling_time  = 0.0;
        // Not in source model      = old_model->sensor.settle_times[1];
        /* TIRS time codes are at the start of the frame */
        band->sampling_char.time_codes_at_frame_start = 1;
        band->sampling_char.maximum_detector_delay = 10.0;
        band->sampling_char.frame_delay = 0;

        for (sca_index = 0; sca_index < NSCA; sca_index++)
        {
            IAS_SENSOR_SCA_MODEL *sca = &band->scas[sca_index];

            /* copy the nominal fill */
            sca->nominal_fill 
                = old_model->image.nominal_fill[calval_band_index][sca_index];

            /* copy the legendre coefficients */
            sca->sca_coef_x[0] 
                = old_model->sensor.sca_coef_x[sca_index][calval_band_index][0];
            sca->sca_coef_x[1] 
                = old_model->sensor.sca_coef_x[sca_index][calval_band_index][1];
            sca->sca_coef_x[2] 
                = old_model->sensor.sca_coef_x[sca_index][calval_band_index][2];
            sca->sca_coef_x[3]
                = old_model->sensor.sca_coef_x[sca_index][calval_band_index][3];
            sca->sca_coef_y[0] 
                = old_model->sensor.sca_coef_y[sca_index][calval_band_index][0];
            sca->sca_coef_y[1] 
                = old_model->sensor.sca_coef_y[sca_index][calval_band_index][1];
            sca->sca_coef_y[2] 
                = old_model->sensor.sca_coef_y[sca_index][calval_band_index][2];
            sca->sca_coef_y[3]
                = old_model->sensor.sca_coef_y[sca_index][calval_band_index][3];

            /* copy the detector offsets and fill */
            /* MS bands */
            for (detector_index = 0; detector_index < sca->detectors;
                    detector_index++)
            {
                sca->detector_offsets_across_track[detector_index]
                    = old_model->sensor.det_delays_ms[sca_index]
                        [calval_band_index][0][detector_index];
                sca->detector_offsets_along_track[detector_index]
                    = old_model->sensor.det_delays_ms[sca_index]
                        [calval_band_index][1][detector_index];
                sca->l0r_detector_offsets[detector_index]
                    = old_model->image.l0r_fill_ms[sca_index]
                        [calval_band_index][detector_index];
            }
        }

        /* get epoch */
        band->frame_count = old_model->image.number_image_lines;
        band->utc_epoch_time[0] = old_model->image.utc_epoch_time[0];
        band->utc_epoch_time[1] = old_model->image.utc_epoch_time[1];
        band->utc_epoch_time[2] = old_model->image.utc_epoch_time[2];
        band->band_present = 1;
    }

    double *frame_seconds_from_epoch
        = malloc(old_model->image.number_image_lines * sizeof(double));
    if (!frame_seconds_from_epoch)
    {
        IAS_LOG_ERROR("Error allocating line times");
        exit(-1);
    }

    memcpy(frame_seconds_from_epoch, old_model->image.time_from_epoch,
            sizeof(double) * old_model->image.number_image_lines);
    if (ias_sensor_set_frame_times(&model->sensor, IAS_TIRS,
                frame_seconds_from_epoch, old_model->image.number_image_lines)
            != SUCCESS)
    {
        IAS_LOG_ERROR("Error setting line times");
        exit(-1);
    }
    sensor->sensors[IAS_OLI].sensor_present = 0;
    sensor->sensors[IAS_TIRS].sensor_present = 1;

    /* allocate an SSM model for tirs */
    sensor->sensors[IAS_OLI].ssm_model = NULL;
    ssm = malloc(sizeof(*ssm));
    if (!ssm)
    {
        IAS_LOG_ERROR("Error allocating memory for the SSM model");
        exit(-1);
    }
    sensor->sensors[IAS_TIRS].ssm_model = ssm;

    /* copy the SSM information to the model */
    memcpy(ssm->alignment_matrix, old_model->tirs_ssm.tele2ssm,
            sizeof(ssm->alignment_matrix));
    ssm->ssm_record_count = old_model->tirs_ssm.number_encoder_angles;

    /* get the epoch time from the first band */
    ssm->utc_epoch_time[0] = old_model->image.utc_epoch_time[0];
    ssm->utc_epoch_time[1] = old_model->image.utc_epoch_time[1];
    ssm->utc_epoch_time[2] = old_model->image.utc_epoch_time[2];

    /* allocate memory the ssm records */
    ssm->records = malloc(sizeof(*ssm->records) * ssm->ssm_record_count);
    if (!ssm->records)
    {
        IAS_LOG_ERROR("Error allocating memory for the SSM records");
        exit(-1);
    }

    /* copy the ssm records */
    for (index = 0; index < ssm->ssm_record_count; index++)
    {
        ssm->records[index].seconds_from_epoch
            = old_model->tirs_ssm.records[index].time_from_epoch;
        ssm->records[index].mirror_angle
            = old_model->tirs_ssm.records[index].ssm_angle;
    }

    return model;
}
