#ifndef _OLI_STRUCTURES_H_
#define _OLI_STRUCTURES_H_

#include <hdf5.h>

//#include "math_quaternion.h"
//#include "xxx_const.h"
#include "ias_const.h"
#include "ias_structures.h"
#include "ias_l0r_constants.h"
#include "tirs_const.h"

//#include <hdf5_hl.h>

#define     NPREC       4

typedef struct EARTH_MODEL
{
   double ut1_utc_correction; /* UT1-UTC conversion */
   double pole_wander_x;      /* Earth's polar wander */
   double pole_wander_y;      /* Earth's polar wander */
   double semi_major;         /* Earth's semi-major axis */
   double semi_minor;         /* Earth's semi-minor axis */
   double eccentricity;       /* Earth's eccentricity */
   double earth_angular_velocity;  /* Earth's inertial rotational rate, or angular velocity */
   double speed_of_light;     /* Speed of light */
} EARTH_MODEL;

typedef struct OLI_IMAGE_MODEL
{  
   int number_image_lines;        /* Number of lines in imagery */
   double utc_epoch_time[3];      /* Year, day of year, seconds of day for start of image */
   double *time_from_epoch;       /* Time of each line present in image data */
   int l0r_fill_pan[NSCA][NUM_DETS_TIRS];             /* Fill inserted in L0R */
   int l0r_fill_ms[NSCA][NBANDS][NUM_DETS_TIRS];     /* Fill inserted in L0R */
   int nominal_fill[NBANDS][NSCA];/* Fill needed to ensure positive pixel locations in L0R */
   IAS_VECTOR *jit_table;             /* Jitter table */
} OLI_IMAGE_MODEL;

typedef struct OLI_SENSOR_MODEL
{
  double oli2acs[3][3];      /* Sensor to Attitude Control System transformation */
  double cm2oli[3];          /* Offset between SC center of mass and OLI sensor */
  double int_times[2];       /* Pan and MS integration times */ 
  double samp_times[2];      /* Pan and MS sampling times */ 
  int number_scas;           /* Number of SCAs in OLI instruments */
  int number_bands;          /* Number of bands in OLI instruments */
  double along_ifov[2];      /* Along track IFOV */
  double across_ifov[2];     /* Across track IFOV */
  int number_det_sca[NBANDS];/* Number of detectors in SCA */
  double sca_coef_x[NSCA][NBANDS][4];
  /* Along-track SCA LOS Legendre coeffs. Note that "_x" refers to
     the sensor coordinate system where the x direction is in the
     direction of satellite motion. */
  double sca_coef_y[NSCA][NBANDS][4];
  /* Across-track SCA LOS Legendre coeffs.  Note that "_y" refers to
     the sensor coordinate system where the y direction is
     perpendicular to the plane formed by the x axis (satellite
     motion) and the z axis pointing to the earth's center. */
  double det_delays_ms[NSCA][NBANDS][2][NUM_DETS_TIRS]; /* Across, along, per -> SCA, Band, detector */
  double det_delays_pan[NSCA][2][NUM_DETS_TIRS]; /* Across, along, per -> SCA, Band, detector */
  /* Detector offsets per SCA per Band for each detector.  Contains integer and
     fractional component */
} OLI_SENSOR_MODEL;

typedef struct OLI_TIME_RECORD
{
   double utc_epoch_time[3]; /* year, day of year, seconds of day */
   double nominal_samp_time;
} OLI_TIME_RECORD;

typedef struct OLI_ATTITUDE_RECORD
{
   double time_from_epoch;    /* seconds */
   IAS_VECTOR sat_att;            /* radians */
   IAS_VECTOR prec_att;           /* radians */
} OLI_ATTITUDE_RECORD;

typedef struct OLI_ATTITUDE_ANCILLARY_RECORD
{
   double time_from_epoch;   /* seconds */
   IAS_QUATERNION eci_quat;       
   IAS_QUATERNION ecf_quat;       
   IAS_VECTOR body_rates;        /* radians per second */    
   IAS_VECTOR rpy;               /* radians */
} OLI_ATTITUDE_ANCILLARY_RECORD;

typedef struct OLI_EPHEMERIS_RECORD
{
   double time_from_epoch;       /* radians */
   IAS_VECTOR eci_position;          /* X, Y, Z (meters) */
   IAS_VECTOR eci_velocity;          /* Vx, Vy, Vz (meters/sec) */
   IAS_VECTOR ecef_position;         /* X, Y, Z (meters) */
   IAS_VECTOR ecef_velocity;         /* Vx, Vy, Vz (meters/sec) */
   IAS_VECTOR prec_eci_position;     /* X, Y, Z (meters) */
   IAS_VECTOR prec_eci_velocity;     /* Vx, Vy, Vz (meters/sec) */
   IAS_VECTOR prec_ecef_position;    /* X, Y, Z (meters) */
   IAS_VECTOR prec_ecef_velocity;    /* Vx, Vy, Vz (meters/sec) */
} OLI_EPHEMERIS_RECORD;

typedef struct TIRS_SSM_RECORD
{
   double time_from_epoch;   /* Time of each SSM encoder angle */
   double ssm_angle;         /* Value of each SSM encoder angle */
} TIRS_SSM_RECORD;

typedef struct OLI_ATTITUDE_MODEL
{
   double utc_epoch_time[3]; /* year, day of year, seconds of day */
   double nominal_samp_time;
   int samples;
   OLI_ATTITUDE_RECORD *records;
} OLI_ATTITUDE_MODEL;

typedef struct OLI_ATTITUDE_ANCILLARY
{
   double utc_epoch_time[3]; 
   double nominal_samp_time;
   int samples;
   OLI_ATTITUDE_ANCILLARY_RECORD *records;
} OLI_ATTITUDE_ANCILLARY;

typedef struct OLI_EPHEMERIS_MODEL
{
   double utc_epoch_time[3]; /* year, day of year, seconds of day */
   double nominal_samp_time;
   int samples;
   OLI_EPHEMERIS_RECORD *records;
} OLI_EPHEMERIS_MODEL;

typedef struct TIRS_TIME_PARAMETERS
{
   double nominal_frame_time;
   double nominal_integration_time;
   double time_code_outlier_threshold;
   double encoder_outlier_threshold;
} TIRS_TIME_PARAMETERS;

typedef struct TIRS_SCENE_SELECT_MIRROR
{
   double tele2ssm[3][3];     /* Telescope to SSM alignment matrix */
   int number_encoder_angles; /* Number of SSM encoder angles */
   TIRS_SSM_RECORD *records;  /* Time and angle pairs */
} TIRS_SCENE_SELECT_MIRROR;

typedef struct TIRS_L0R_IRU_GYRO
{
    uint16_t data_validity_flags;
    double gyro_a_integrated_angle;
    double gyro_b_integrated_angle;
    double gyro_c_integrated_angle;
    double gyro_d_integrated_angle;
    double time_tag;
    uint16_t iru_config_flag;
} TIRS_L0R_IRU_GYRO;

typedef struct TIRS_L0R_ACS_QUATERNION
{
    uint32_t time_tag_sec;
    double time_tag_usec;
    double qi2b_1;
    double qi2b_2;
    double qi2b_3;
    double qi2b_4;
    double wb_x;
    double wb_y;
    double wb_z;
    double ab_x;
    double ab_y;
    double ab_z;
} TIRS_L0R_ACS_QUATERNION;

typedef struct TIRS_L0R_IRU_TIME
{
    uint32_t iru_epoch_sec;
    double iru_epoch_usec;
} TIRS_L0R_IRU_TIME;

typedef struct TIRS_L0R_EPHEMERIS
{
    uint32_t time_tag_sec;
    double time_tag_usec;
    double ecef_x_pos;
    double ecef_y_pos;
    double ecef_z_pos;
    double ecef_x_vel;
    double ecef_y_vel;
    double ecef_z_vel;
    double eci_x_pos;
    double eci_y_pos;
    double eci_z_pos;
    double eci_x_vel;
    double eci_y_vel;
    double eci_z_vel;
} TIRS_L0R_EPHEMERIS;

//typedef struct TIRS_L0R_SSM
//{
//    uint16_t ssm_epoch_days;
//    uint32_t ssm_milliseconds;
//    uint16_t ssm_microseconds;
//    uint32_t ssm_1_encoder_position;
//    uint32_t ssm_2_encoder_position;
//    uint32_t ssm_3_encoder_position;
//    uint32_t ssm_4_encoder_position;
//    uint32_t ssm_5_encoder_position;
//    uint32_t ssm_6_encoder_position;
//    uint32_t ssm_7_encoder_position;
//    uint32_t ssm_8_encoder_position;
//    uint32_t ssm_9_encoder_position;
//    uint32_t ssm_10_encoder_position;
//    uint32_t ssm_11_encoder_position;
//    uint32_t ssm_12_encoder_position;
//    uint32_t ssm_13_encoder_position;
//    uint32_t ssm_14_encoder_position;
//    uint32_t ssm_15_encoder_position;
//    uint32_t ssm_16_encoder_position;
//    uint32_t ssm_17_encoder_position;
//    uint32_t ssm_18_encoder_position;
//    uint32_t ssm_19_encoder_position;
//    uint32_t ssm_20_encoder_position;
//} TIRS_L0R_SSM;
//
typedef struct IAS_TIRS_L0R_FILE
{
    char name[IAS_L0R_FILE_NAME_LENGTH+1]; /* file name */
      char date[IAS_L0R_FILEDATE_LENGTH+1];  /* file creation date */
      char checksum[IAS_L0R_CHECKSUM_LENGTH+1]; /* calculated MD5 checksum */
      uint64_t size;                       /* size, in bytes, of the file */
}IAS_TIRS_L0R_FILE;

typedef struct IAS_TIRS_L0R_FILE_METADATA
{
    IAS_TIRS_L0R_FILE ancillary_file; /* information about the ancillary file */
    IAS_TIRS_L0R_FILE checksum_file;
    IAS_TIRS_L0R_FILE band_files[IAS_L0R_MAX_BAND_FILES]; /* information on each band
                                                             file */
    uint16_t interval_files;
    IAS_TIRS_L0R_FILE metadata_file;
}IAS_TIRS_L0R_FILE_METADATA;

typedef struct IAS_TIRS_L0R_INTERVAL_METADATA
{
    char ancillary_start_time[IAS_L0R_DATE_LENGTH+1];
    /* Year, Day of year and spacecraft start time
        associated with the first group ancillary data received */
    char ancillary_stop_time[IAS_L0R_DATE_LENGTH+1];
    /* Year, Day of year and spacecraft start time
        associated with the last group ancillary data received */
    uint32_t attitude_points; /*good spacecraft attitude points */
    uint32_t attitude_points_missing; /*spacecraft attitude points missing*/
    uint32_t attitude_points_rejected; /*spacecraft attitude points rejected*/
    char collection_type[IAS_L0R_COLLECTION_TYPE_LENGTH+1];
    IAS_CORNERS corners; /* corner points of the interval */
    IAS_CORNERS corners_thermal; /* scene corner points for the thermal band*/
    char cpf_name[IAS_L0R_CPF_NAME_LENGTH+1];  /* CPF used for processing */
    uint32_t crc_errors_oli; /* Number of frame CRC failures for OLI */
    uint32_t crc_errors_tirs; /* Number of frame CRC failures for TIRS */
    char data_type[IAS_L0R_DATA_TYPE_LENGTH+1]; /*data type identifier string */
    char date_acquired[IAS_L0R_DATE_LENGTH+1];
    /* Year, Day of year and GMT that this scene was imaged */
    uint32_t ephemeris_points; /* good ephemeris data points received */
    uint32_t ephemeris_points_missing; /* ephemeris data points found missing */
    uint32_t ephemeris_points_rejected;/* ephemeris data points found to fail */
    char hostname[IAS_L0R_HOSTNAME_LENGTH+1];
    /* machine where data was processed*/
    uint8_t interval_files;/* Total number of files included in this interval */
    uint32_t interval_frames_oli; /*number of OLI frames reported
         in this interval file */
    uint32_t interval_frames_tirs; /*number of TIRS frames reported
         in this interval file */
    uint8_t interval_number;/*Sequence of the interval within the ingest file */
    uint16_t interval_version;/*Processed version of the interval, starts at 0*/
    char is_version[IAS_SOFTWARE_VERSION_SIZE];
    /* Version number of the software installed on the IS when a metadata file
    * and its associated Level 0R files were generated */
    char landsat_cal_interval_id[IAS_L0R_INTERVAL_ID_LENGTH+1];
    /* Unique Landsat calibration interval identifier*/
    char landsat_interval_id[IAS_L0R_INTERVAL_ID_LENGTH+1];
    uint32_t frames_filled_oli;/*number of OLI frames that were filled*/
    uint32_t frames_filled_tirs;/*number of TIRs frames that were filled*/
    char nadir_offnadir[IAS_L0R_NADIR_LENGTH+1];
    /* Nadir or Off-Nadir condition of the interval */
    char quality_algorithm[IAS_L0R_ALGORITHM_NAME_LENGTH+1];
    /*Algorithm (date, name, and version number) used to calculate the
        quality for this interval*/
    float roll_angle; /*amount of spacecraft roll at interval center */
    uint8_t satellite; /* Landsat vehicle number that captured the data */
    char sensor_id[IAS_L0R_SENSOR_ID_LENGTH+1]; /* name of the sensor  */
    char spacecraft_id[IAS_L0R_SPACECRAFT_ID_LENGTH+1];
    /*name of the spacecraft*/
    char start_time_oli[IAS_L0R_DATE_LENGTH+1];
    /* Year, Day of year and GMT spacecraft start time of the first OLI frame
         of the interval */
    char start_time_tirs[IAS_L0R_DATE_LENGTH+1];
    /* Year, Day of year and GMT spacecraft start time of the first TIRS frame
        of the interval */
    char station_id[IAS_L0R_STATION_ID_LENGTH+1]; /* receiving group
        station code */
    char stop_time_oli[IAS_L0R_DATE_LENGTH+1];
    /* Year, Day of year and spacecraft start time of the last OLI frame
        of the interval */
    char stop_time_tirs[IAS_L0R_DATE_LENGTH+1];
    /* Year, Day of year and spacecraft start time of the last TIRS frame
        of the interval */
    uint32_t time_code_errors_oli;/*number of OLI time code
        errors in the interval */
    uint32_t time_code_errors_tirs;/*number of TIRS time code
        errors in the interval */
    uint8_t wrs_ending_row;
    /* WRS row number for the last reported scene in this interval */
    uint8_t wrs_scenes; /* total number of full & partial scenes
        in the interval  */
    uint8_t wrs_scenes_full;    /* number of full scenes in the interval */
    uint8_t wrs_scenes_partial; /* number of partial scenes in the interval */
    uint8_t wrs_starting_path;  /* starting WRS2 path for scenes */
    uint8_t wrs_starting_row;   /* WRS2 number row for the first scene */
    uint8_t wrs_type;           /* WRS which applies to the interval, WRS-2 */
}IAS_TIRS_L0R_INTERVAL_METADATA;

typedef struct IAS_TIRS_L0R_SCENE_METADATA
{
    int8_t ancillary_quality;     /*Anc quality 9 = Best, 0 = Worst, -1 = NA */
    uint16_t attitude_points;         /*good spacecraft attitude points */
    uint16_t attitude_points_missing; /*spacecraft attitude points missing*/
    uint16_t attitude_points_rejected;/*spacecraft attitude points rejected*/
    char collection_type[IAS_L0R_COLLECTION_TYPE_LENGTH+1];
    IAS_CORNERS corners;             /* scene corner points */
    IAS_CORNERS corners_thermal; /* scene corner points for the thermal band*/
    uint32_t crc_errors; /* Number of frame CRC failures both instruments */
    char date_acquired[IAS_L0R_TIME_LENGTH+1];
    /* timecode for the center frame */
    char day_night[IAS_L0R_DAY_NIGHT_LENGTH+1]; /* day/night condition
                                                 of the scene */
    uint16_t ephemeris_points;         /* good ephemeris data points received */
    uint16_t ephemeris_points_missing; /* ephemeris data points found missing */
    uint16_t ephemeris_points_rejected;/* ephemeris data points found to fail */
    char full_partial_scene[IAS_L0R_SCENE_FULL_PARTIAL_LENGTH+1];
    /* full or partial scene*/
    int32_t scene_center_shift;
    /* distance between the calculated center and the nominal center*/
    char hostname[IAS_L0R_HOSTNAME_LENGTH+1];
    int8_t image_quality;
    char landsat_scene_id[IAS_L0R_SCENE_ID_LENGTH+1]; /* unique ID used
                                                     for Landsat data */
    uint16_t missing_frames;
    char nadir_offnadir[IAS_L0R_NADIR_LENGTH+1];
    float roll_angle;        /* amount or spacecraft roll at the scene center */
    IAS_DBL_LAT_LONG scene_center; /* scene center point */
    uint32_t scene_start_frame_oli;  /* first OLI frame number in the scene */
    uint32_t scene_stop_frame_oli;   /* last OLI frame number in the scene */
    uint32_t scene_start_frame_tirs; /* first TIRS frame number in the scene */
    uint32_t scene_stop_frame_tirs;  /* last TIRS frame number in the scene */
    char start_time[IAS_L0R_TIME_LENGTH+1]; /* earliest spacecraft time of
                               any (OLI/TIRS) scene frame within the interval*/
    char stop_time[IAS_L0R_TIME_LENGTH+1]; /* latest spacecraft time of
                               any (OLI/TIRS) scene frame within the interval*/
    char subsetter_version_l0rp[IAS_SOFTWARE_VERSION_SIZE];
    /* Version of the subsetter that created this data.  This will only
       be applicable to L0Rp datasets */
    double sun_azimuth;   /* sun azimuth angle at the scene center */
    double sun_elevation; /* sun elevation angle at the scene center */
    uint16_t target_wrs_path; /* nearest WRS path to
                             the frame of sight scene center*/
    uint8_t target_wrs_row; /* nearest WRS row to the
                             frame of sight scene center */
    uint16_t time_code_errors; /* the number of time code errors in the scene */
    uint16_t wrs_path;
    uint8_t wrs_row;
    uint8_t wrs_scene_number;
}IAS_TIRS_L0R_SCENE_METADATA;

typedef struct OLI_PRECISION_MODEL
{
  double secs_from_image_epoch;   /* Seconds from epoch */
  int eph_order;            /* Number of ephemeris correction coefficients */
  double x[NPREC];          /* Ephemeric correction X direction */
  double y[NPREC];          /* Ephemeric correction Y direction */
  double z[NPREC];          /* Ephemeric correction Z direction */
  int att_order;            /* Number of attitude correction coefficients */
  double roll[NPREC];       /* Attitude correction X direction */
  double pitch[NPREC];      /* Attitude correction Y direction */
  double yaw[NPREC];        /* Attitude correction Z direction */
} OLI_PRECISION_MODEL;

#endif
