/******************************************************************************
Name: ias_los_model_get_min_max_detector_offsets

Purpose: Calculates the minimum and maximum detector offset for each of the
    SCAs in the requested band.

Returns: SUCCESS/ERROR

******************************************************************************/
#include <stdlib.h>
#include "ias_los_model.h"
#include "logging_channel.h" /* define debug logging channel */
#include "ias_logging.h"

int ias_los_model_get_min_max_detector_offsets
(
    const IAS_LOS_MODEL *model,         /* I: LOS model structure */
    int band_number,                    /* I: Band number */
    double min_det_offsets[IAS_MAX_NSCAS],/* O: Minimum detector offsets in
                                                each SCA */
    double max_det_offsets[IAS_MAX_NSCAS] /* O: Maximum detector offsets in 
                                                each SCA */
)
{
    int band_index;
    int sca_index;
    const IAS_SENSOR_BAND_MODEL *band_model;

    /* Convert the band number to an index */
    band_index = ias_sat_attr_convert_band_number_to_index(band_number);
    if (band_index == ERROR)
    {
        IAS_LOG_ERROR("Failed to convert band number %d to an index",
                band_number);
        return ERROR;
    }
    band_model = &model->sensor.bands[band_index];

    for (sca_index = 0; sca_index < band_model->sca_count; sca_index++)
    {
        const IAS_SENSOR_SCA_MODEL *sca_model = &band_model->scas[sca_index];
        int detector_index;
        double min_offset = 12.0; /* start with the min/max offsets at invalid
                                     values */
        double max_offset = -12.0;

        /* Find the minimum and maximum offsets in the detectors for the
           current SCA */
        for (detector_index = 0; detector_index < sca_model->detectors;
             detector_index++)
        {
            double offset;

            offset = sca_model->detector_offsets_along_track[detector_index];

            if (offset < min_offset)
                min_offset = offset;
            if (offset > max_offset)
                max_offset = offset;
        }

        /* Adjust the offset for the differences in collection types */
        if (model->acquisition_type != IAS_EARTH)
        {
            /* Lunar and stellar may scan at 1/12 nominal rate */
            min_offset = 12 * min_offset + 1;
            max_offset = 12 * max_offset + 1;
        }

        min_det_offsets[sca_index] = min_offset;
        max_det_offsets[sca_index] = max_offset;
    }

    return SUCCESS;
}
