/******************************************************************************
NAME: ias_sensor_set_ssm_from_l0r

PURPOSE: Set the Scene Select Mirror telemetry into the model from the L0R.
    Information needed include the sample times for each SSM angle and the SSM
    encoder angles.

RETURNS
    SUCCESS or ERROR
******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include <string.h>
#include "../logging_channel.h" /* define the debug logging channel */
#include "ias_logging.h"
#include "ias_cpf.h"
#include "ias_sensor_model.h"
#include "local_defines.h"

/* Constant value for expected TIRS scene select mirror electronics mode */
/* The SSM has multiple modes of operation. MCE (mechanical electronics) mode 4
   is what we expect it to be using normally. There is also a mode 10 that was
   added during instrument testing to address some unusual behavior being seen.
   It has been decided that mode 10 won't be used on orbit and even if it were,
   it doesn't really affect us... much. If it ends up being in mode 10 much, 
   we'll probably have to do something extra some day to beef up a few things. 
   There is also a mode 0, open loop operation, that was used when the SSM 
   encoder began to exhibit problems. So, if the SSM electronics isn't in 
   mode 4, a warning is appropriate at this time. If it is in mode 0 we need 
   to also check the status of the encoder. If it is powered and indexed then 
   valid encoder data will be available to measure the SSM position. If not, 
   then an external source of SSM position data is needed. */
 
#define SSM_MCE_NORMAL_MODE 4
#define SSM_MCE_OPENLOOP_MODE 0

/* define the size of the window to use for smoothing SSM encoder data */
/* We use a bigger window (more smoothing) in scenes that contain a
   SSM mode transition */
#define NORMAL_SSM_SMOOTH_WINDOW        5
#define MODESWITCH_SSM_SMOOTH_WINDOW    21

/* Define the bits that indicated the encoder is powered on and/or indexed */
#define SSM_POWERED                     0x08 
#define SSM_INDEXED                     0x01 

/* Defines the number of seconds before the start of a scene that a SSM encoder
   mode switch can exist to classify the scene as "SWITCHED" */
#define SWITCHED_MODE_GAP_SECONDS 10.0

/* Max line length in estimated positions file for reading header lines */
#define HEADER_LINE_LENGTH 256

/* Define the format version for the estimated ssm file */
#define FORMAT_VERSION 1

int ias_sensor_set_ssm_from_l0r
(
    int ssm_encoder_value,          /* I: User-specified SSM encoder value */
    TIRS_SENSOR_L0R_DATA *l0r_data, /* I: L0R data to create a model */
    IAS_CPF *cpf,                   /* I: Calibration parameter file */
    IAS_SENSOR_MODEL *sensor,       /* I/O: Sensor model to update with flags */
    IAS_SENSOR_ID sensor_id,        /* I: Sensor id to set SSM info into */
    const char *ssm_position_filename,/* I: SSM Position Filename */
    IAS_SSM_POSITION_QUALITY *ssm_position_quality/*I/O: SSM Positions quality 
                                                         flag */
)
{
    int l0r_record_count;           /* Number of records in L0R */
    double start_secs;              /* Image start time from spacecraft epoch */
    double end_secs;                /* Image end time from spacecraft epoch */
    double ssm_secs;                /* SSM record time from spacecraft epoch */
    IAS_L0R_TIRS_SSM_ENCODER *l0r_ssm_records;/* SSM records from l0r_data */
    int field;                      /* Loop counter */
    int index;                      /* Loop counter */
    int start_rec;                  /* First SSM record to include in model */
    int current_rec;                /* Current SSM record when looping */
    int scene_ssm_records;          /* SSM record count to include in model */
    int num_frames;                 /* Count of frames for a sensor */
    IAS_SENSOR_SSM_RECORD *ssm_data;/* Structure array to hold model SSM data */
    double sample_time;             /* The time between SSM samples */
    double ssm_scale;               /* SSM scale factor */
    double pi;                      /* Constant */
    double encoder_origin;          /* Encoder angle at nadir - from CPF
                                       (depends on MCE side) */
    unsigned char sideA = 0x01;     /* Electronics enabled mask - MCE side A */
    unsigned char sideB = 0x02;     /* Electronics enabled mask - MCE side B */
    int mce_vote = 0;               /* MCE majority vote variable */
    int mech_mode_warn = 0;         /* Flag to track if warning issued about
                                       unexpected mechanical mode yet */
    int mech_mode_0_cnt = 0;        /* Number of mode 0 ancillary records */
    int mech_mode_4_cnt = 0;        /* Number of mode 4 ancillary records */
    int ssm_position_warn = 0;      /* Flag to track if warning issued about 
                                       the SSM position yet */
    int need_estimated_ssm = 0;     /* Flag indicating number of records that
                                       require external SSM encoder data */
    int window_size = NORMAL_SSM_SMOOTH_WINDOW;
                                    /* SSM encoder smoothing window size */
    IAS_L0R_TIRS_SSM_ENCODER *record; /* Pointer to the current ssm_record */
    const struct IAS_CPF_TIRS_PARAMETERS *tirs_params; 
                                    /* Pointer to CPF TIRS parameters */
    IAS_SENSOR_SCENE_SELECT_MIRROR_MODEL *ssm_model;
                                    /* Cached pointer to the SSM model */
    IAS_SENSOR_ESTIMATED_SSM_ENCODER *est_ssm_data = NULL;
                                    /* Array of structures containing estimated
                                       SSM encoder data from external file */
    int number_est_encoder_positions = 0;/* Number of estimated SSM records
                                            found */
    int mode_switch_start = 0;  /* Record index for 10 seconds before the scene
                                   start index to check for mode switching */
    int mode_switch_in_interval = 0; /* Flag to indicate there is a mode switch
                                        in the scene */
    int mode_switch_in_scene = 0; /* Flag to indicate there is a mode switch
                                     in the scene */

    /* Make sure the sensor id is valid */
    if (sensor_id < 0 || sensor_id >= IAS_MAX_SENSORS)
    {
        IAS_LOG_ERROR("Invalid sensor id: %d", sensor_id);
        return ERROR;
    }

    /* Cache a pointer to the ssm model for the sensor */
    ssm_model = sensor->sensors[sensor_id].ssm_model;

    /* Get the pi constant */
    pi = ias_math_get_pi();

    /* Get the TIRS parameters from the CPF */
    tirs_params = ias_cpf_get_tirs_parameters(cpf);
    if (!tirs_params)
    {
        IAS_LOG_ERROR("Failed to retrieve TIRS parameters from the CPF");
        return ERROR;
    }

    l0r_ssm_records = l0r_data->ssm_records;
    l0r_record_count = l0r_data->num_ssm_records;

    /* Find a band from this sensor and copy the image epoch time to the SSM
       structure */
    for (index = 0; index < sensor->band_count; index++)
    {
        IAS_SENSOR_BAND_MODEL *band = &sensor->bands[index];

        if (band->band_present && (band->sensor->sensor_id == sensor_id))
        {
            ssm_model->utc_epoch_time[0] = band->utc_epoch_time[0];
            ssm_model->utc_epoch_time[1] = band->utc_epoch_time[1];
            ssm_model->utc_epoch_time[2] = band->utc_epoch_time[2];
            break;
        }
    }
    if (index >= sensor->band_count)
    {
        IAS_LOG_ERROR("Unable to find a band present for the sensor");
        return ERROR;
    }

    /* Logic to check MCE A/B side flags */
    for (index = 0; index < l0r_record_count; index++)
    {
        if ((l0r_ssm_records[index].elec_enabled_flags & sideA) > 0)
            mce_vote++;
        if ((l0r_ssm_records[index].elec_enabled_flags & sideB) > 0)
            mce_vote--;

        /* While looping through all the records, also check the SSM
           mechanical mode and warn if any mode other than the normal
           mode expected is found for any telemetry sample */
        if (l0r_ssm_records[index].ssm_mech_mode != SSM_MCE_NORMAL_MODE)
        {
            /* Issue warning message first time */
            if (!mech_mode_warn)
            {
                IAS_LOG_WARNING("MCE mode %d in use, not mode %d as expected",
                        l0r_ssm_records[index].ssm_mech_mode,
                        SSM_MCE_NORMAL_MODE);
                mech_mode_warn = 1;
            }
            if (l0r_ssm_records[index].ssm_mech_mode == SSM_MCE_OPENLOOP_MODE)
            {
                /* See if there is valid encoder data */
                if ((l0r_ssm_records[index].mc_encoder_flags & SSM_POWERED) == 0
                     || (l0r_ssm_records[index].mc_encoder_flags & SSM_INDEXED) 
                     == 0)
                {
                   if (!need_estimated_ssm)
                   {
                        IAS_LOG_WARNING("Operating in MCE mode 0 without "
                            "valid SSM encoder data!");
                   }
                   need_estimated_ssm++;
                }

                mech_mode_0_cnt++;
            }
        }
        else
        {
            mech_mode_4_cnt++;
        }

        /* Also check the SSM position and issue a warning if it is not nadir- 
           pointing for any telemetry sample */
        if (!ssm_position_warn && (l0r_ssm_records[index].ssm_position_sel !=
            IAS_L0R_TIRS_SSM_NADIR_POSITION))
        {
            IAS_LOG_WARNING("SSM position at index %d is %d, not %d (nadir "
                "position) as expected", index,
                l0r_ssm_records[index].ssm_position_sel, 
                IAS_L0R_TIRS_SSM_NADIR_POSITION);
            ssm_position_warn = 1;
        }
    }

    /* Set the encoder origin value based on the majority MCE electronics
       side being used */
    if (mce_vote < 0)
        encoder_origin = tirs_params->ssm_encoder_origin_side_b;
    else
        encoder_origin = tirs_params->ssm_encoder_origin_side_a;

    /* Use a larger smoothing window if a mode switch is present.  This allows
       the rapid SSM movement that happens right after some mode switches to
       still be processed through the grid.  Without this, the transition is
       so fast that it can't be handled by the grid. The check for > 1 is used
       to filter out a spurious single frame */
    if (mech_mode_0_cnt > 1 && mech_mode_4_cnt > 1)
    {
        mode_switch_in_interval = 1;
        window_size = MODESWITCH_SSM_SMOOTH_WINDOW;
    }

    /* Align the TIRS SSM encoder telemetry and trim back down to 20 samples */
    ias_sensor_align_ssm_data(l0r_ssm_records, l0r_record_count);

    /* Check the SSM data time codes and angle values for consistency */
    if (ias_sensor_check_ssm_encoder_data(l0r_ssm_records, l0r_record_count,
                encoder_origin, cpf, &l0r_data->num_bad_ssm_times,
                &l0r_data->num_bad_ssm_encoder_values) != SUCCESS)
    {
        IAS_LOG_ERROR("SSM telemetry is bad");
        return ERROR;
    }

    /* Convert the epoch time to J2000 seconds */
    if (ias_math_convert_year_doy_sod_to_j2000_seconds(
                ssm_model->utc_epoch_time, &start_secs) != SUCCESS)
    {
        IAS_LOG_ERROR("Failed to convert epoch year/doy/sod to J2000");
        return ERROR;
    }

    /* Find the SSM record at the start of the scene */
    start_rec = 0;
    for (field = 0; field < l0r_record_count; field++)
    {
        record = &l0r_ssm_records[field];

        ssm_secs = (double)record->l0r_time.days_from_J2000 * IAS_SEC_PER_DAY
            + record->l0r_time.seconds_of_day
            + tirs_params->ssm_encoder_time_offset;

        if (ssm_secs > start_secs)
            break;

        start_rec = field;
    }

    /* Back up one record, if possible, to account for the need to calculate
       the terrain sensitivity using the maximum offset which can result in
       needing SSM information from approximately 10 * sampling time before
       the start of the image.  A single record is enough to cover that time. */
    if (start_rec > 0)
        start_rec--;

    /* Initialize the count of SSM records to use for the scene to one since
       we always include the first record and start looking for the ending
       record one record after the starting record */
    scene_ssm_records = 1;

    /* Calculate the end time for the scene */
    num_frames = sensor->frame_counts[sensor_id];
    end_secs = start_secs
        + sensor->frame_seconds_from_epoch[sensor_id][num_frames - 1];
    
    /* Starting with the record after the starting record (since we have to
       include at least one record), count the number of records available
       until the record after the ending time */
    for (field = (start_rec + 1); field < l0r_record_count; field++)
    {
        record = &l0r_ssm_records[field];

        ssm_secs = (double)record->l0r_time.days_from_J2000 * IAS_SEC_PER_DAY
            + record->l0r_time.seconds_of_day
            + tirs_params->ssm_encoder_time_offset;

        scene_ssm_records++;

        if (ssm_secs > end_secs)
            break;
    }

    /* Add an extra record at the end, if possible, just in case */
    if ((start_rec + scene_ssm_records) < (l0r_record_count - 1))
    {
        scene_ssm_records++;
    }
    else if ((start_rec + scene_ssm_records) > l0r_record_count)
    {
        /* Make sure the scene_ssm_records + start_rec isn't greater than the
           l0r_record_count.  Note that this should never happen, but extra
           insurance doesn't hurt since a bug caused this to happen at one
           point. */
        IAS_LOG_WARNING("Number of scene_ssm_records (%i) + start_rec (%i) is "
            "> l0r_record_count (%i), adjusting accordingly",
            scene_ssm_records, start_rec, l0r_record_count);
        scene_ssm_records = l0r_record_count - start_rec;
    }

    /* Store number of SSM samples in model, there are
       IAS_L0R_TIRS_SSM_ENCODER_SAMPLES samples per record */
    ssm_model->ssm_record_count = IAS_L0R_TIRS_SSM_ENCODER_SAMPLES
        * scene_ssm_records;
    IAS_LOG_DEBUG("Number of SSM samples %d", ssm_model->ssm_record_count);

    /* Allocate space for the output SSM data */
    ssm_data = malloc(sizeof(*ssm_data) * ssm_model->ssm_record_count);
    if (ssm_data == NULL)
    {
        IAS_LOG_ERROR("Error allocating SSM data array");
        return ERROR;
    }

    /* Issue message if there is a user entered SSM encoder value */
    if (ssm_encoder_value > 0)
    {
        IAS_LOG_WARNING("Replacing all SSM encoder data with user-provided "
            "value of %d", ssm_encoder_value);
    }

    /* If we have determined that estimated SSM encoder data are required,
       look for and load that file. The check >1 has been used to filter out
       any spurious single frames */
    else if (need_estimated_ssm > 1 && (ssm_position_filename != NULL &&
        ssm_position_filename[0] != '\0'))
    {
        if (ias_sensor_load_estimated_ssm(start_secs, end_secs, 
            ssm_position_filename, &est_ssm_data,
            &number_est_encoder_positions) != SUCCESS)
        {
            IAS_LOG_ERROR("Loading the estimated SSM file");
            free(ssm_data);
            return ERROR;
        }
    }
    else if (need_estimated_ssm > 1 && (ssm_position_filename == NULL ||
        ssm_position_filename[0] == '\0'))
    {
        /* Note that an empty or NULL ssm_position filename is allowed
           since Ingest will not use the SSM estimated positions when 
           framing */
        IAS_LOG_WARNING("Estimated SSM encoder data required, however "
            "no estimated SSM encoder file provided.");
    }

    /* If there is a mode switch in the interval check if that switch happened
       within this scenes window */
    if (mode_switch_in_interval)
    {
        /* A mode switch between mode 4 and mode 0 causes gaps between SCAs for
           roughly 10 seconds, so if the switch happens 10 seconds before this 
           scene starts or anywhere in the scene, classify it as a mode swiched
           scene */
        for (mode_switch_start = start_rec; mode_switch_start > 0; 
             mode_switch_start--)
        {
            record = &l0r_ssm_records[mode_switch_start];

            ssm_secs = (double)record->l0r_time.days_from_J2000 
                * IAS_SEC_PER_DAY + record->l0r_time.seconds_of_day
                + tirs_params->ssm_encoder_time_offset;

            if (ssm_secs <= start_secs - SWITCHED_MODE_GAP_SECONDS)
                break;
        }

        /* Reset the mech mode counts to only count the records for this scene 
           to determine if there was a mode switch or not */
        mech_mode_4_cnt = 0;
        mech_mode_0_cnt = 0;

        /* Check for a mode switch from 0 to 4 or from 4 to 0 for this scene */
        for (field = mode_switch_start; field < start_rec + scene_ssm_records; 
             field++)
        {
            if (l0r_ssm_records[field].ssm_mech_mode != SSM_MCE_NORMAL_MODE)
            {
                if (l0r_ssm_records[field].ssm_mech_mode 
                    == SSM_MCE_OPENLOOP_MODE)
                    mech_mode_0_cnt++;
            }
            else
            {
                mech_mode_4_cnt++;         
            }

            if (mech_mode_0_cnt > 0 && mech_mode_4_cnt > 0)
            {
                mode_switch_in_scene = 1;
                break;        
            }
        }
    }

    /* Determine the correct SSM position status. The logic here uses >= 1
       because a single frame still indicates that a switch happened
       within the scene window */
    if (mode_switch_in_scene)
        ssm_model->position_status = IAS_SSM_POSITION_SWITCHED;
    else if (ssm_encoder_value > 0 || number_est_encoder_positions > 0)
        ssm_model->position_status = IAS_SSM_POSITION_ESTIMATED;
    else if (l0r_data->num_bad_ssm_encoder_values > 0)
        ssm_model->position_status = IAS_SSM_POSITION_DEFAULT;
    else
        ssm_model->position_status = IAS_SSM_POSITION_NOMINAL;

    /* Calculate the scaling factor to convert from SSM encoder counts to
       radians */
    ssm_scale = 2.0 * pi/(double)TIRS_SSM_OVERFLOW;

    /* Convert the SSM times to seconds from epoch */
    current_rec = start_rec;
    sample_time = 1.0 / IAS_L0R_TIRS_SSM_ENCODER_SAMPLES;
    for (field = 0 ; field < ssm_model->ssm_record_count;
            field += IAS_L0R_TIRS_SSM_ENCODER_SAMPLES)
    {
        int prev;                  /* Previous SSM encoder sample */
        double seconds_from_epoch; /* Cached copy of seconds from epoch
                                      for the 24-bit record */

        record = &l0r_ssm_records[current_rec++];

        /* Calculate the time at the current record */
        seconds_from_epoch
            = (double)record->l0r_time.days_from_J2000 * IAS_SEC_PER_DAY
            + record->l0r_time.seconds_of_day
            + tirs_params->ssm_encoder_time_offset - start_secs;

        ssm_data[field].seconds_from_epoch = seconds_from_epoch;

        /* Compute the offset times for each additional sample */
        for (index = 1; index < IAS_L0R_TIRS_SSM_ENCODER_SAMPLES; index++)
        {
            ssm_data[field + index].seconds_from_epoch =
                seconds_from_epoch + (sample_time * index);
        }

        /* See if the user entered a SSM encoder value to use */
        if (ssm_encoder_value > 0)
        {
            for (index = 0; index < IAS_L0R_TIRS_SSM_ENCODER_SAMPLES; index++)
            {
                record->encoder_position[index] = (uint32_t)ssm_encoder_value;
            }
        }
        /* See if we need to replace encoder data with estimated */
        else if (number_est_encoder_positions > 0 &&
            record->ssm_mech_mode == SSM_MCE_OPENLOOP_MODE &&
            ((record->mc_encoder_flags & SSM_POWERED) == 0 ||
            (record->mc_encoder_flags & SSM_INDEXED) == 0))
        {
            if (ias_sensor_get_estimated_ssm(number_est_encoder_positions,
                field, start_secs, est_ssm_data, ssm_data, ssm_position_quality,
                record) != SUCCESS)
            { 
                IAS_LOG_ERROR("Getting estimated SSM encoder data");
                free(ssm_data);
                free(est_ssm_data);
                return ERROR;
            }
        }

        /* Convert the angles */
        prev = record->encoder_position[0] & TIRS_SSM_BIT_MASK;
        ssm_data[field].mirror_angle = (double)prev * ssm_scale
            - encoder_origin;

        /* Make sure the angle is between -pi and pi */
        while (ssm_data[field].mirror_angle > pi)
            ssm_data[field].mirror_angle -= 2.0 * pi;
        while (ssm_data[field].mirror_angle < -pi)
            ssm_data[field].mirror_angle += 2.0 * pi;

        /* Convert the SSM encoder samples to angles */
        for (index = 1; index < IAS_L0R_TIRS_SSM_ENCODER_SAMPLES; index++)
        {
            int delta_ssm;          /* Change in SSM encoder reading */
            int current;            /* Current SSM encoder sample */
            int last24;             /* Previous 24-bit encoder sample */
            double mirror_angle;    /* Temporary mirror angle */

            /* Convert the encoder samples */
            last24 = prev & TIRS_SSM_BIT_MASK;
            current = record->encoder_position[index] & TIRS_SSM_BIT_MASK;
            delta_ssm = current - last24;
            if (delta_ssm > TIRS_SSM_LIMIT)
                delta_ssm -= TIRS_SSM_OVERFLOW;
            if (delta_ssm < -TIRS_SSM_LIMIT)
                delta_ssm += TIRS_SSM_OVERFLOW;

            current = prev + delta_ssm;
            if (current < 0)
                current += TIRS_SSM_OVERFLOW;

            /* Convert the ssm encoder count to radians and limit to be
               between -pi and pi */
            mirror_angle = (double)current * ssm_scale - encoder_origin;
            while (mirror_angle > pi)
                mirror_angle -= 2.0 * pi;
            while (mirror_angle < -pi)
                mirror_angle += 2.0 * pi;

            /* Save the mirror angle */
            ssm_data[field + index].mirror_angle = mirror_angle;

            prev = current;
        }
    }
    /* Set the positions quality for the model based on the 
       ssm_position_quality */
    ssm_model->position_quality = *ssm_position_quality;

    /* Save the SSM data in the model */
    ssm_model->records = ssm_data;

    /* See if we need to release estimated encoder data */
    if (est_ssm_data)
    { 
        free(est_ssm_data);
        est_ssm_data = NULL;
    }

    /* Smooth the SSM data to suppress encoder jitter */
    if (ias_sensor_smooth_ssm_data(window_size, ssm_model) != SUCCESS)
    {
        IAS_LOG_ERROR("Error smoothing SSM data");
        free(ssm_model->records);
        ssm_model->records = NULL;
        return ERROR;
    }

    return SUCCESS;
}

/******************************************************************************
NAME: ias_sensor_load_estimated_ssm

PURPOSE: Loads the estimated Scene Select Mirror data using the provided
         ssm_position_filename which contains the ssm positions. If the
         file cannot be found a warning will be issued and SUCCESS will be 
         returned with the number of encoder positions set to 0.

NOTE: The caller is responsible for freeing the memory allocated to 
      the est_ssm_data pointer. This will only need to be freed if the 
      number_estimated_encoder_positions > 0

RETURNS
    SUCCESS or ERROR
******************************************************************************/
int ias_sensor_load_estimated_ssm
(
    double start_time,             /* I: Seconds from  J2000 epoch time for
                                         which SSM data are needed - start */
    double stop_time,              /* I: Seconds from  J2000 epoch time for
                                         which SSM data are needed - stop */
    const char *ssm_position_filename,/* I: SSM Position Filename */
    IAS_SENSOR_ESTIMATED_SSM_ENCODER **est_ssm_data,/* O: Pointer to allocated 
                                         and populated estimated SSM encoder
                                         data structure array */
    int *number_estimated_encoder_positions/* O: The number of estimated encoder
                                                 positions */
)                                              
{
    char buffer[HEADER_LINE_LENGTH]; /* Buffer to store read header lines */
    FILE *ssm_fptr;               /* Estimated SSM file pointer */

    int format_found = 0;         /* Flag indicating format_version line
                                     found in ssm file */
    int header_end_found = 0;     /* Flag indicating the end of the header was
                                     found in ssm file */
    int num_header_lines = 0;     /* Number of lines read from the est ssm
                                     file before the actual positions */
    int num_records_found = 0;    /* Flag indicating num_records line found in
                                      ssm file */
    int num_records_expected = 0; /* Total number of records reported in ssm
                                     file */
    int num_records_read = 0;     /* Number of records read from ssm file */
    IAS_SENSOR_ESTIMATED_SSM_ENCODER *local_est_ssm_data = NULL;/* Local pointer
                                        to the estimated SSM data array */
    int status;                   /* Return status placeholder */
    int total_records_read = 0;   /* Total number of estimated records */

    /* Initialize the output data */
    *est_ssm_data = NULL;
    *number_estimated_encoder_positions = 0;

    /* Open file with estimated SSM encoder values */
    ssm_fptr = fopen(ssm_position_filename, "r");
    if (!ssm_fptr)
    {
        IAS_LOG_ERROR("Cannot open estimated ssm position file %s",
            ssm_position_filename);
        return ERROR;
    }

    /* First read the entire line in using fgets, so that we are guaranteed
       everything to the end of line is consumed, since we don't really
       care (or know) the format of each header line */
    while(fgets(buffer, HEADER_LINE_LENGTH, ssm_fptr) != NULL)
    {
        char parameter[HEADER_LINE_LENGTH];
        char value[HEADER_LINE_LENGTH];
        char delimiter[HEADER_LINE_LENGTH];

        /* Now parse out the information from the line that we DO care
           about */ 
        status = sscanf(buffer, "%s %s %s", parameter, delimiter, value);
        if (status != 3)
        {
            IAS_LOG_ERROR("Reading the SSM header information at line %d",
                num_header_lines);
            fclose(ssm_fptr);
            return ERROR;
        }
        if (strcmp(parameter, "FORMAT_VERSION") == 0)
        {
            if (atoi(value) != FORMAT_VERSION)
            {
                IAS_LOG_ERROR("Invalid format version (%s) read at line %d",
                    value, num_header_lines);
                fclose(ssm_fptr);
                return ERROR;
            }
            format_found = 1;
        }
        else if (strcmp(parameter, "NUM_RECORDS") == 0)
        {
            num_records_expected = atoi(value);
            num_records_found = 1;
        }
        else if (strncmp(parameter, "#", 1) == 0 && 
            strncmp(delimiter, "YEAR", 4) == 0)
        {
            header_end_found = 1;
        }
        num_header_lines++;

        /* Break if all header lines we need to check have been found */
        if (header_end_found && format_found && num_records_found)
        {
            break;
        }
    }

    /* Verify the important header lines were found */
    if (!header_end_found || !format_found || !num_records_found)
    {
        IAS_LOG_ERROR("Expected header lines not found");
        fclose(ssm_fptr);
        return ERROR;
    }

    /* If the number of records isn't at least one, it is an error since we
       decided estimated positions were needed and a file was provided.
       That file should not be empty of estimated positions. */
    if (num_records_expected <= 0)
    {
        IAS_LOG_ERROR("Invalid NUM_RECORDS (%d) found in %s. It must have at "
            "least one record.", num_records_expected, ssm_position_filename);
        fclose(ssm_fptr);
        return ERROR;
    }

    /* Allocate memory for the records that will be read */
    local_est_ssm_data = malloc(num_records_expected
        * sizeof(*local_est_ssm_data));
    if (!local_est_ssm_data)
    {
        IAS_LOG_WARNING("Allocating %d estimated SSM records",
            num_records_expected);
        fclose(ssm_fptr);
        return ERROR;
    }

    /* At this point we are ready to read the data */
    for (num_records_read = 0; num_records_read < num_records_expected;
        num_records_read++)
    {
        /* For quality read from ssm file */
        int ssm_quality;

        if (feof(ssm_fptr))
        {
            break;
        }

        /* Read this record's time, encoder position, and quality */
        status = fscanf(ssm_fptr, "%lf %lf %lf %u %d ",
            &local_est_ssm_data[num_records_read].ssm_time[0], 
            &local_est_ssm_data[num_records_read].ssm_time[1],
            &local_est_ssm_data[num_records_read].ssm_time[2], 
            &local_est_ssm_data[num_records_read].encoder_position,
            &ssm_quality);
        if (status != 5)
        {
            IAS_LOG_ERROR("Reading the SSM encoder positions at line %d",
                num_records_read + 1);
            fclose(ssm_fptr);
            free(local_est_ssm_data);
            return ERROR;
        }

        /* Convert estimated ssm time (year, doy, sod) to j2000 seconds */
        if (ias_math_convert_year_doy_sod_to_j2000_seconds( 
            local_est_ssm_data[num_records_read].ssm_time, 
            &local_est_ssm_data[num_records_read].ssm_j2000_seconds) != SUCCESS)
        {
            IAS_LOG_ERROR("Calculating J2000 seconds for ssm record %d",
                num_records_read);
            fclose(ssm_fptr);
            free(local_est_ssm_data);
            return ERROR;
        }

        /* Now set the quality in the struct */
        local_est_ssm_data[num_records_read].ssm_position_quality = ssm_quality;
    }
    total_records_read = num_records_read;

    /* Close the estimated SSM file */
    if (fclose(ssm_fptr) != 0)
    {
        IAS_LOG_ERROR("Closing the estimated SSM file %s",
            ssm_position_filename);
        free(local_est_ssm_data);
        return ERROR;
    }

    /* Check that we read the number of SSM records expected */
    if (total_records_read != num_records_expected)
    {
        IAS_LOG_ERROR("Number of records read (%i) from input file %s does not "
            "match expected number of records (%i)", total_records_read,
            ssm_position_filename, num_records_expected);
        free(local_est_ssm_data);
        return ERROR;
    }

    /* See if the SSM file intersects the current interval time */
    if (stop_time < local_est_ssm_data[0].ssm_j2000_seconds || start_time >
        local_est_ssm_data[total_records_read - 1].ssm_j2000_seconds)
    { 
        IAS_LOG_ERROR("Definitive SSM encoder data does not cover this "
            "scene");
        IAS_LOG_ERROR("Scene time range: %lf to %lf", start_time, stop_time);
        IAS_LOG_ERROR("Definitive SSM encoder data time range: %lf to %lf", 
            local_est_ssm_data[0].ssm_j2000_seconds, 
            local_est_ssm_data[total_records_read - 1].ssm_j2000_seconds);
        free(local_est_ssm_data);
        return ERROR;
    }

    IAS_LOG_INFO("Using estimated SSM encoder data from %s to replace "
        "missing mirror telemetry", ssm_position_filename);

    /* Output the ssm data and number of estimated positions */
    *est_ssm_data = local_est_ssm_data;
    *number_estimated_encoder_positions = total_records_read;

    return SUCCESS;
}

/******************************************************************************
NAME: ias_sensor_get_estimated_ssm

PURPOSE: Retrieves the estimated encoder position using the estimated SSM 
         data. If the time comes before the estimated data then the earliest
         encoder position from the estimated data will be used. If the time 
         comes after the estimated data then the latest encoder position from 
         the estimated data will be used. Otherwise the encoder position will
         be interpolated using the estimated SSM data.

RETURNS
    SUCCESS or ERROR
******************************************************************************/
int ias_sensor_get_estimated_ssm
( 
    int number_est_encoder_positions,/* I: Number of entries in the estimated 
                                           SSM data array */
    int field_index,                 /* I: SSM data starting field index */
    double start_seconds,            /* I: Starting time offset */
    const IAS_SENSOR_ESTIMATED_SSM_ENCODER *est_ssm_data,/* I: Array containing
                                                   estimated SSM encoder data */
    const IAS_SENSOR_SSM_RECORD *ssm_data, /* I: SSM data */
    IAS_SSM_POSITION_QUALITY *ssm_position_quality,/* I/O: SSM positions quality
                                                           flag */
    IAS_L0R_TIRS_SSM_ENCODER *record /* O: Pointer to SSM record */
)     
{
    int ssm_index = 0;       /* Array index into estimated SSM data */
    int index;               /* Generic loop index */

    /* Interpolate the estimated encoder position */
    for (index = 0; index < IAS_L0R_TIRS_SSM_ENCODER_SAMPLES; index++)
    {  
        double start_j2000 = ssm_data[field_index + index].seconds_from_epoch 
            + start_seconds; /* Starting index time from J2000 */

        if (start_j2000 > est_ssm_data[ssm_index].ssm_j2000_seconds)
        {
            /* Loop through estimated ssm data and find closest value */
            for (; ssm_index < number_est_encoder_positions; ssm_index++)
            {
                if (start_j2000 < est_ssm_data[ssm_index].ssm_j2000_seconds)
                    break;
            }
        }
        
        /* If we are past the end of the file, use the last estimated encoder 
           position for the entire record */
        if (ssm_index >= number_est_encoder_positions)
        {
            record->encoder_position[index] = est_ssm_data[
                number_est_encoder_positions - 1].encoder_position;

            /* Determine the overall lowest quality */
            *ssm_position_quality = IAS_MIN(
                    est_ssm_data[number_est_encoder_positions - 1].
                    ssm_position_quality, *ssm_position_quality);
            
        }    
        /* If we are before the beginning of the file, use the first 
           estimated encoder position */
        else if (ssm_index == 0) 
        {
            record->encoder_position[index] = est_ssm_data[0].encoder_position;

            /* Determine the overall lowest quality */
            *ssm_position_quality =  IAS_MIN(
                    est_ssm_data[0].ssm_position_quality, 
                    *ssm_position_quality);
        }
        else
        {
            double frac;          /* Linear interpolation fraction */
            double total_seconds; /* Time difference between estimated 
                                     SSM samples */

            /* Interpolate this sample */ 
            total_seconds = est_ssm_data[ssm_index].ssm_j2000_seconds 
                - est_ssm_data[ssm_index - 1].ssm_j2000_seconds;
            if (total_seconds <= 0.0)
            {
                IAS_LOG_ERROR("Time inconsistency detected in estimated "
                    "SSM file");
                return ERROR;
            }       
     
            frac = (ssm_data[field_index + index].seconds_from_epoch 
                + start_seconds - est_ssm_data[ssm_index - 1].ssm_j2000_seconds)
                / total_seconds;

            record->encoder_position[index] = floor((est_ssm_data[
                ssm_index - 1].encoder_position * (1.0 - frac) +
                est_ssm_data[ssm_index].encoder_position * frac) + 0.5);

            /* Check ssm_index - 1 estimated ssm positions quality first */
            *ssm_position_quality =  IAS_MIN(
                    est_ssm_data[ssm_index - 1].ssm_position_quality,
                    *ssm_position_quality);

            /* Now check ssm_index estimated ssm positions quality */
            *ssm_position_quality = IAS_MIN(
                    est_ssm_data[ssm_index].ssm_position_quality,
                    *ssm_position_quality);
        }
    }

    return SUCCESS;
}
