#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>

#include "ias_gcc_pragmas.h"
#include "ias_satellite_attributes.h"
#include "ias_los_model.h"
#include "ias_model_io.h"
#include "ias_cpf.h"
#include "ias_logging.h"
#include "cgreen/cgreen.h"
#include "../local_defines.h"

#define PAN_SCALE 2.0

#define HEADER_LINE_LENGTH 256

/* Disable the command line specified optimization
   It is re-enabled below
   We disable the optimization because the inclusion of headers which contain
   structures defining huge amounts of testing data, cause very very very long
   compiles.  And it is the optimization in the newer compilers which cause
   the lengthy compile time with the include files. */
IAS_DISABLE_OPTIMIZATION

/* Test the set frame times routine */
Ensure test_set_frame_times()
{
    IAS_LOS_MODEL *model;
    IAS_SENSOR_MODEL *sensor;
    int band_index;
    int oli_frames;
    double *oli_frame_seconds_from_epoch;
    int tirs_frames;
    double *tirs_frame_seconds_from_epoch;
    int frame;
    int status;
    int oli_band_count = 0;
    int tirs_band_count = 0;

    /* Initialize the satellite attributes */
    ias_sat_attr_initialize(IAS_L8);

    /* Allocate a model */
    model = ias_los_model_allocate();
    assert_not_equal(model, NULL);
    if (model == NULL)
        return;

    /* Allocate memory for the OLI times */
    oli_frames = 20;
    oli_frame_seconds_from_epoch = malloc(oli_frames * sizeof(double));
    assert_not_equal(oli_frame_seconds_from_epoch, NULL);
    if (oli_frame_seconds_from_epoch == NULL)
    {
        ias_los_model_free(model);
        return;
    }

    /* Initialize the OLI times */
    for (frame = 0; frame < oli_frames; frame++)
        oli_frame_seconds_from_epoch[frame] = frame + 1.5;

    /* Disable all output since the tests are hitting error conditions */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    sensor = &model->sensor;

    /* Test some error conditions first */
    status = ias_sensor_set_frame_times(sensor, -1,
        oli_frame_seconds_from_epoch, oli_frames);
    assert_equal(status, ERROR);

    status = ias_sensor_set_frame_times(sensor, IAS_MAX_SENSORS,
        oli_frame_seconds_from_epoch, oli_frames);
    assert_equal(status, ERROR);

    ias_log_set_output_level(IAS_LOG_LEVEL_WARN);

    /* Now actually set the OLI info */
    status = ias_sensor_set_frame_times(sensor, IAS_OLI,
        oli_frame_seconds_from_epoch, oli_frames);
    assert_equal(status, SUCCESS);

    /* Set up some TIRS info */
    tirs_frames = 10;
    tirs_frame_seconds_from_epoch = malloc(tirs_frames * sizeof(double));
    assert_not_equal(tirs_frame_seconds_from_epoch, NULL);

    /* Initialize the TIRS times */
    for (frame = 0; frame < tirs_frames; frame++)
        tirs_frame_seconds_from_epoch[frame] = frame + 2.5;

    /* Make sure an error happens if trying to set the OLI times a second
       time */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = ias_sensor_set_frame_times(sensor, IAS_OLI,
        tirs_frame_seconds_from_epoch, tirs_frames);
    ias_log_set_output_level(IAS_LOG_LEVEL_WARN);
    assert_equal(status, ERROR);

    /* Now really set the TIRS values */
    status = ias_sensor_set_frame_times(sensor, IAS_TIRS,
        tirs_frame_seconds_from_epoch, tirs_frames);
    assert_equal(status, SUCCESS); 

    /* Verify the frame time counts in the main structure are correct */
    assert_equal(sensor->frame_counts[IAS_OLI], oli_frames);
    assert_equal(sensor->frame_counts[IAS_TIRS], tirs_frames);

    /* Verify the model contains the right info for each of the bands */
    for (band_index = 0; band_index < sensor->band_count; band_index++)
    {
        IAS_SENSOR_BAND_MODEL *band = &sensor->bands[band_index];
        double *test_lines = band->frame_seconds_from_epoch;

        if (band->sensor->sensor_id == IAS_OLI)
        {
            oli_band_count++;
            assert_equal(band->frame_count, oli_frames);
            for (frame = 0; frame < oli_frames; frame++)
            {
                assert_double_equal(test_lines[frame], frame + 1.5);
            }
        }
        else if (band->sensor->sensor_id == IAS_TIRS)
        {
            tirs_band_count++;
            assert_equal(band->frame_count, tirs_frames);
            for (frame = 0; frame < tirs_frames; frame++)
            {
                assert_double_equal(test_lines[frame], frame + 2.5);
            }
        }
        else
        {
            /* This should never happen */
            assert_true(0);
        }
    }

    /* Make sure the right number of bands was found for each sensor */
    assert_equal(oli_band_count, 9);
    assert_equal(tirs_band_count, 2);

    ias_los_model_free(model);
}

/* Read in the correct model to use for testing. The OLI versus TIRS models
   are what were delivered with the OLI or TIRS ADD deliveries for
   create model. */
IAS_LOS_MODEL *read_test_model
(
    IAS_SENSOR_ID sensor        /* I */
)
{
    IAS_LOS_MODEL *model;
    const char *test_data_dir;
    char model_filename[PATH_MAX];
    int status = -1;

    /* Get the test data directory */
    test_data_dir = getenv("IAS_UNIT_TEST_DATA_DIR");
    assert_true(test_data_dir != NULL);
    if (test_data_dir == NULL)
        return NULL;

    /* Build the model filename */
    if (sensor == IAS_OLI)
    {
        status = snprintf(model_filename, sizeof(model_filename),
                "%s/ias_lib/los_model/sensor/oli_model.h5", test_data_dir);
    }
    if (sensor == IAS_TIRS)
    {
        status = snprintf(model_filename, sizeof(model_filename),
                "%s/ias_lib/los_model/sensor/tirs_model.h5", test_data_dir);
    }
    if (status < 0 || status >= sizeof(model_filename))
    {
        assert_true_with_message(0, "Error assembling model filename string");
        return NULL;
    }

    model = ias_model_read(model_filename);
    if (!model)
    {
        assert_true_with_message(0, "Error reading input model");
        return NULL;
    }

    return model;
}

typedef struct findlos_inputs
{
    int band;
    int sca;
    double detector;
    IAS_SENSOR_DETECTOR_TYPE dettype;
} FINDLOS_INPUTS;

Ensure test_find_los_vector()
{
    IAS_LOS_MODEL *model;
    /* The following values were obtained by running the cal/val prototype
       code */
    #include "find_los_vector_test_data.h"
    int count = sizeof(inputs)/sizeof(FINDLOS_INPUTS);
    int index;
    IAS_VECTOR output;
    int status;

    ias_sat_attr_initialize(IAS_L8);

    /* Get the test model */
    model = read_test_model(IAS_OLI);
    assert_true(model != NULL);
    if (!model)
        return;

    /* Disable all output since the tests are hitting error conditions */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* Verify illegal detector numbers are caught as errors, first with -1 */
    status = ias_sensor_find_los_vector(0, -1, IAS_ACTUAL_DETECTOR,
            &model->sensor.bands[inputs[0].band], &output);
    assert_equal(status, ERROR);

    /* And now with one too many detectors */
    status = ias_sensor_find_los_vector(0, 
            model->sensor.bands[inputs[0].band].scas[0].detectors, 
            IAS_ACTUAL_DETECTOR, &model->sensor.bands[inputs[0].band], &output);
    assert_equal(status, ERROR);

    /* Re-enable logging */
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

    for (index = 0; index < count; index++)
    {
        FINDLOS_INPUTS *input = &inputs[index];

        status = ias_sensor_find_los_vector(input->sca, input->detector,
                input->dettype, &model->sensor.bands[input->band], &output);
        if (status != SUCCESS)
        {
            assert_true_with_message(0, "Error finding LOS vector");
        }
        else
        {
            assert_double_equal_with_message(outputs[index].x, output.x,
                "failed x check at index %d: expected %.8e, got %.8e", index,
                outputs[index].x, output.x);
            assert_double_equal_with_message(outputs[index].y, output.y,
                "failed y check at index %d: expected %.8e, got %.8e", index,
                outputs[index].y, output.y);
            assert_double_equal_with_message(outputs[index].z, output.z,
                "failed z check at index %d: expected %.8e, got %.8e", index,
                outputs[index].z, output.z);
        }
    }

    ias_los_model_free(model);
}

Ensure test_tirs_find_los_vector()
{
    IAS_LOS_MODEL *model;
    /* The following values were obtained by running the cal/val prototype
       code */
    #include "find_tirs_los_vector_test_data.h"
    int count = sizeof(inputs)/sizeof(FINDLOS_INPUTS);
    int index;
    IAS_VECTOR output;
    int status;

    ias_sat_attr_initialize(IAS_L8);

    /* Get the test model */
    model = read_test_model(IAS_TIRS);
    assert_true(model != NULL);
    if (!model)
        return;

    /* Disable all output since the tests are hitting error conditions */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* Verify illegal detector numbers are caught as errors, first with -1 */
    status = ias_sensor_find_los_vector(0, -1, IAS_ACTUAL_DETECTOR,
            &model->sensor.bands[inputs[0].band], &output);
    assert_equal(status, ERROR);

    /* And now with one too many detectors */
    status = ias_sensor_find_los_vector(0, 
            model->sensor.bands[inputs[0].band].scas[0].detectors, 
            IAS_ACTUAL_DETECTOR, &model->sensor.bands[inputs[0].band], &output);
    assert_equal(status, ERROR);

    /* Re-enable logging */
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

    for (index = 0; index < count; index++)
    {
        FINDLOS_INPUTS *input = &inputs[index];

        status = ias_sensor_find_los_vector(input->sca, input->detector,
                input->dettype, &model->sensor.bands[input->band], &output);
        if (status != SUCCESS)
        {
            assert_true_with_message(0, "Error finding LOS vector");
        }
        else
        {
            assert_double_equal_with_message(outputs[index].x, output.x,
                "failed x check at index %d: expected %.8e, got %.8e", index,
                outputs[index].x, output.x);
            assert_double_equal_with_message(outputs[index].y, output.y,
                "failed y check at index %d: expected %.8e, got %.8e", index,
                outputs[index].y, output.y);
            assert_double_equal_with_message(outputs[index].z, output.z,
                "failed z check at index %d: expected %.8e, got %.8e", index,
                outputs[index].z, output.z);
        }
    }

    ias_los_model_free(model);
}

Ensure test_tirs_adjust_los_for_ssm()
{
    IAS_LOS_MODEL *model;
    /* The following values were obtained by running the cal/val prototype
       code */
    #include "tirs_adjust_los_for_ssm_test_data.h"
    int count = sizeof(inputs)/sizeof(FINDSSM_INPUTS);
    int index;
    int status;

    ias_sat_attr_initialize(IAS_L8);

    /* Get the test model */
    model = read_test_model(IAS_TIRS);
    assert_true(model != NULL);
    if (!model)
        return;

    /* Disable all output since the tests are hitting error conditions */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    for (index = 0; index < count; index++)
    {
        IAS_VECTOR output;
        FINDSSM_INPUTS *input = &inputs[index];

        status = ias_sensor_adjust_los_for_ssm(
                    model->sensor.sensors[IAS_TIRS].ssm_model,
                    input->time, &input->los, &output);
        if (status != output_status[index])
        {
            assert_true_with_message(0, "Error finding LOS vector");
            ias_los_model_free(model);
            return;           
        }

        if (status == SUCCESS)
        {
            assert_double_equal_with_message(outputs[index].x, output.x,
                "failed x check at index %d: expected %.8e, got %.8e", index,
                outputs[index].x, output.x);
            assert_double_equal_with_message(outputs[index].y, output.y,
                "failed y check at index %d: expected %.8e, got %.8e", index,
                outputs[index].y, output.y);
            assert_double_equal_with_message(outputs[index].z, output.z,
                "failed z check at index %d: expected %.8e, got %.8e", index,
                outputs[index].z, output.z);
        }
    }
    ias_log_set_output_level(IAS_LOG_LEVEL_WARN);

    ias_los_model_free(model);
}

typedef struct findtime_inputs
{
    int band;
    int sca;
    double line;
    double sample;
    IAS_SENSOR_DETECTOR_TYPE dettype;
} FINDTIME_INPUTS;

/* Verify find_time values, shared aspects of verification done by multiple
   tests, so broken out into a separate routine to avoid code duplication */
void verify_find_time_results
(
    IAS_LOS_MODEL *model,
    FINDTIME_INPUTS *inputs,
    double *outputs,
    int *output_status,
    int count
)
{
    int index;
    int status;
    double output;

    /* Disable all output since the tests are hitting error conditions */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* Make sure invalid sample numbers are caught as errors, first with a
       sample of -1 */
    status = ias_sensor_find_time(0.0, -1.0, inputs[0].band, 0,
        &model->sensor, IAS_NOMINAL_DETECTOR, &output);
    assert_equal(status, ERROR);

    /* And now test with one more SCA than expected */
    status = ias_sensor_find_time(0.0, 
        model->sensor.bands[inputs[0].band].scas[0].detectors, inputs[0].band,
        0, &model->sensor, IAS_NOMINAL_DETECTOR, &output);
    assert_equal(status, ERROR);

    for (index = 0; index < count; index++)
    {
        FINDTIME_INPUTS *input = &inputs[index];

        status = ias_sensor_find_time(input->line, input->sample, input->band,
                    input->sca, &model->sensor, input->dettype, &output);
        if (status != output_status[index])
        {
            /* Only use asserts if there is a failure due to the cgreen limit
               of 16384 asserts per run */
            assert_equal_with_message(status, output_status[index],
                "failed status - expected %d, got %d for index %d",
                output_status[index], status, index);
        }
        else if (status != ERROR)
        {
            /* Use a cgreen function to verify the doubles are equal to 8
               significant digits */
            int valid = doubles_are_equal(outputs[index], output);
            if (!valid)
            {
                assert_equal_with_message(valid, 1, 
                    "output error - expected %.8e, got %.8e for index %d",
                    outputs[index], output, index);
            }
        }
    }

    ias_log_set_output_level(IAS_LOG_LEVEL_WARN);
}

Ensure test_find_time()
{
    IAS_LOS_MODEL *model;
    /* The following values were obtained by running the cal/val prototype
       code */
    #include "find_time_test_data.h"

    int count = sizeof(inputs)/sizeof(FINDTIME_INPUTS);

    ias_sat_attr_initialize(IAS_L8);

    /* Get the test model */
    model = read_test_model(IAS_OLI);
    assert_true(model != NULL);
    if (!model)
        return;

    /* Verify the results with the normal settling times */
    verify_find_time_results(model, inputs, outputs, output_status, count);

    ias_los_model_free(model);
}

Ensure test_tirs_find_time()
{
    IAS_LOS_MODEL *model;
    /* The following values were obtained by running the cal/val prototype
       code */
    #include "tirs_find_time_test_data.h"

    int count = sizeof(inputs)/sizeof(FINDTIME_INPUTS);

    ias_sat_attr_initialize(IAS_L8);

    /* Get the test model */
    model = read_test_model(IAS_TIRS);
    assert_true(model != NULL);
    if (!model)
        return;

    /* Verify the results with the normal settling times */
    verify_find_time_results(model, inputs, outputs, output_status, count);

    ias_los_model_free(model);
}

/* Read the CPF used for testing */
IAS_CPF *read_test_cpf
(
    IAS_SENSOR_ID sensor            /* I: Sensor testing reading CPF for */
)
{
    IAS_CPF *cpf;
    char cpf_filename[PATH_MAX];
    int status;

    /* Build the CPF filename */
    status = snprintf(cpf_filename, sizeof(cpf_filename),
            "./L8CPF20090101_20090331.01");
    if (status < 0 || status >= sizeof(cpf_filename))
    {
        assert_true_with_message(0, "Error setting cpf filename string");
        return NULL;
    }

    /* Call a script to copy the CPF file locally, since in the TIRS case
       the file needs some manual tweaking of values */
    if (sensor == IAS_TIRS)
    {
        /* If reading the CPF for TIRS SSM testing, need to tweak the
           ssm_tolerance value in the CPF to use the relaxed value that CalVal
           used for their ADD test data. This is done to fully exercise the new
           SSM alignment algorithm. */
        status = system("./setup_cpf.csh tirs");
    }
    else
        status = system("./setup_cpf.csh");

    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error setting up CPF file for testing");
        return NULL;
    }

    /* Allocate the CPF data structure */
    cpf = ias_cpf_read(cpf_filename);
    assert_true(cpf != NULL);
    if (cpf == NULL)
    {
        assert_true_with_message(0, "Error reading CPF %s", cpf_filename);
        return NULL;
    }

    return cpf;
}

/* Test the set CPF routine */
Ensure test_set_cpf()
{
    IAS_LOS_MODEL *model;
    IAS_SENSOR_MODEL *sensor;
    IAS_CPF *cpf;
    IAS_SENSOR_L0R_DATA l0r_sensor_data;
    int band_index;
    int sca_index;
    int i;
    int status;

    /* Set the log level to warnings only to avoid the debugging output
       to clutter up the test output. */
    ias_log_set_output_level(IAS_LOG_LEVEL_WARN);

    /* Initialize the satellite attributes */
    ias_sat_attr_initialize(IAS_L8);

    /* Allocate a model */
    model = ias_los_model_allocate();
    assert_not_equal(model, NULL);
    if (model == NULL)
        return;

    sensor = &model->sensor;

    /* Read CPF */
    cpf = read_test_cpf(IAS_OLI);
    assert_not_equal(cpf, NULL);
    if (cpf == NULL)
    {
        ias_los_model_free(model);
        return;
    }

    /* Don't exit if failure; check individual values to assist debugging */
    status = ias_sensor_set_cpf(cpf, &l0r_sensor_data, sensor);
    assert_equal(status, SUCCESS);

    /* Nominal frame times */
    assert_double_equal(0.004236, l0r_sensor_data.oli_data.nominal_frame_time);
    assert_double_equal(0.014286,
            l0r_sensor_data.tirs_data.nominal_frame_time);

    /* OLI to ACS alignment matrix */
    assert_double_equal(0.99999852, sensor->sensors[0].sensor2acs[0][0]);
    assert_double_equal(0.0015036322, sensor->sensors[0].sensor2acs[0][1]);
    assert_double_equal(-0.00083058264, sensor->sensors[0].sensor2acs[0][2]);
    assert_double_equal(-0.0015045439, sensor->sensors[0].sensor2acs[1][0]);
    assert_double_equal(0.99999827, sensor->sensors[0].sensor2acs[1][1]);
    assert_double_equal(-0.0010981457, sensor->sensors[0].sensor2acs[1][2]);
    assert_double_equal(0.00082892998, sensor->sensors[0].sensor2acs[2][0]);
    assert_double_equal(0.0010993938, sensor->sensors[0].sensor2acs[2][1]);
    assert_double_equal(0.99999905, sensor->sensors[0].sensor2acs[2][2]);

    /* Center of mass to OLI offsets */
    assert_double_equal(2.556, sensor->sensors[0].center_mass2sensor_offset[0]);
    assert_double_equal(-0.669, 
                         sensor->sensors[0].center_mass2sensor_offset[1]);
    assert_double_equal(0.082, sensor->sensors[0].center_mass2sensor_offset[2]);

    /* TIRS to ACS alignment matrix */
    assert_double_equal(0.99999852, sensor->sensors[IAS_TIRS].sensor2acs[0][0]);
    assert_double_equal(0.0015036322,
                        sensor->sensors[IAS_TIRS].sensor2acs[0][1]);
    assert_double_equal(-0.00083058264,
                        sensor->sensors[IAS_TIRS].sensor2acs[0][2]);
    assert_double_equal(-0.0015045439,
                        sensor->sensors[IAS_TIRS].sensor2acs[1][0]);
    assert_double_equal(0.99999827, sensor->sensors[IAS_TIRS].sensor2acs[1][1]);
    assert_double_equal(-0.0010981457,
                        sensor->sensors[IAS_TIRS].sensor2acs[1][2]);
    assert_double_equal(0.00082892998,
                        sensor->sensors[IAS_TIRS].sensor2acs[2][0]);
    assert_double_equal(0.0010993938,
                        sensor->sensors[IAS_TIRS].sensor2acs[2][1]);
    assert_double_equal(0.99999905, sensor->sensors[IAS_TIRS].sensor2acs[2][2]);

    /* Center of mass to TIRS offsets */
    assert_double_equal(2.556,
                        sensor->sensors[IAS_TIRS].center_mass2sensor_offset[0]);
    assert_double_equal(-0.669, 
                        sensor->sensors[IAS_TIRS].center_mass2sensor_offset[1]);
    assert_double_equal(0.082,
                        sensor->sensors[IAS_TIRS].center_mass2sensor_offset[2]);

    /* The TIRS alignment matrix should be an identity matrix with the given
       input */
    for (i = 0; i < 3; i++)
    {
        IAS_SENSOR_SCENE_SELECT_MIRROR_MODEL *ssm_model
                = sensor->sensors[IAS_TIRS].ssm_model;
        int j;
        for (j = 0; j < 3; j++)
        {
            if (i == j)
                assert_double_equal(1.0, ssm_model->alignment_matrix[i][j]);
            else
                assert_double_equal(0.0, ssm_model->alignment_matrix[i][j]);
        }
    }

    for (band_index = 0; band_index < sensor->band_count; band_index++)
    {
        IAS_SENSOR_DETECTOR_SAMPLING_CHARACTERISTICS *samp_char;
        IAS_SENSOR_BAND_MODEL *band = &sensor->bands[band_index];
        samp_char = &band->sampling_char;

        /* Lines per frame */
        switch(band_index)
        {
            case 0:
            case 1:
            case 2:
            case 3:
            case 4:
            case 5:
            case 6:
            case 8:
            case 9:
            case 10:
                assert_equal(1, samp_char->lines_per_frame);
                break;
            case 7:
                assert_equal(2, samp_char->lines_per_frame);
                break;
            default:
                /* This should never happen */
                assert_true(0);
                break;
        }

        /* Sampling settling time is set to zero on TIRS bands instead of
           read from the CPF */
        if (band->sensor->sensor_id == IAS_OLI)
            assert_double_equal(0.00001, samp_char->settling_time);
        else
            assert_double_equal(0.0, samp_char->settling_time);

        /* Along and across track IFOV */
        switch(band_index)
        {
            case 0:
            case 1:
            case 2:
            case 3:
            case 4:
            case 5:
            case 6:
            case 8:
                assert_double_equal(0.0000426, samp_char->along_ifov);
                assert_double_equal(0.0000426, samp_char->across_ifov);
                break;
            case 7:
                assert_double_equal(0.00002125, samp_char->along_ifov);
                assert_double_equal(0.00002125, samp_char->across_ifov);
                break;
            case 9:
            case 10:
                assert_double_equal(1.4184e-4, samp_char->along_ifov);
                assert_double_equal(1.4184e-4, samp_char->across_ifov);
                break;
            default:
                /* This should never happen */
                assert_true(0);
                break;
        }

        /* Integration time */
        switch(band_index)
        {
            case 0:
            case 1:
            case 2:
            case 3:
            case 4:
            case 5:
            case 6:
            case 8:
                assert_double_equal(0.0036, samp_char->integration_time);
                break;
            case 7:
                assert_double_equal(0.00066, samp_char->integration_time);
                break;
            case 9:
            case 10:
                assert_double_equal(0.002, samp_char->integration_time);
                break;
            default:
                /* This should never happen */
                assert_true(0);
                break;
        }

        /* Maximum detector delay */
        assert_double_equal(10.0, samp_char->maximum_detector_delay);

        for (sca_index = 0; sca_index < sensor->bands[band_index].sca_count;
                sca_index++)
        {
            if (band_index == 0 && sca_index == 0)
            {
                IAS_SENSOR_SCA_MODEL *sca;
                sca = &sensor->bands[0].scas[0];

                /* SCA Nominal fill */
                assert_equal(2, sca->nominal_fill);

                /* Along track Legendre coefficients */
                assert_double_equal(0.0061299674, sca->sca_coef_x[0]);
                assert_double_equal(-0.00034091384, sca->sca_coef_x[1]);
                assert_double_equal(0.0000034868684, sca->sca_coef_x[2]);

                /* Across track Legendre coefficients */
                assert_double_equal(0.12485876, sca->sca_coef_y[0]);
                assert_double_equal(-0.010237616, sca->sca_coef_y[1]);
                assert_double_equal(0.000011593474, sca->sca_coef_y[2]);

                for (i = 0; i < 5; i++) /* detector */
                {
                    double a, b;
                    /* Along track offsets */
                    a=sensor->bands[0].scas[0].detector_offsets_along_track[i];
                    /* Across track offsets */
                    b=sensor->bands[0].scas[0].detector_offsets_across_track[i];

                    switch(i)
                    {
                        case 0:
                            assert_double_equal(0.000, a);
                            assert_double_equal(0.005, b);
                            break;
                        case 1:
                            assert_double_equal(2.022, a);
                            assert_double_equal(0.008, b);
                            break;
                        case 2:
                            assert_double_equal(0.000, a);
                            assert_double_equal(0.004, b);
                            break;
                        case 3:
                            assert_double_equal(2.022, a);
                            assert_double_equal(0.007, b);
                            break;
                        case 4:
                            assert_double_equal(0.000, a);
                            assert_double_equal(0.004, b);
                            break;
                    }
                }
            }
        }
    }

    /* Leap seconds */
    assert_equal(25, l0r_sensor_data.leap_seconds_data->leap_seconds_count);
    for (i = 0; i < l0r_sensor_data.leap_seconds_data->leap_seconds_count; i++)
    {
        switch(i)
        {
            /* Select a few to validate */
            case 0:
                assert_equal(1972,
                    l0r_sensor_data.leap_seconds_data->leap_years[0]);
                assert_equal(7,
                    l0r_sensor_data.leap_seconds_data->leap_months[0]);
                assert_equal(1, 
                    l0r_sensor_data.leap_seconds_data->leap_days[0]);
                assert_equal(1,
                    l0r_sensor_data.leap_seconds_data->num_leap_seconds[0]);
                break;
            case 5:
                assert_equal(1977,
                    l0r_sensor_data.leap_seconds_data->leap_years[5]);
                assert_equal(1,
                    l0r_sensor_data.leap_seconds_data->leap_months[5]);
                assert_equal(1, 
                    l0r_sensor_data.leap_seconds_data->leap_days[5]);
                assert_equal(1,
                    l0r_sensor_data.leap_seconds_data->num_leap_seconds[5]);
                break;
            case 12:
                assert_equal(1985,
                    l0r_sensor_data.leap_seconds_data->leap_years[12]);
                assert_equal(7,
                    l0r_sensor_data.leap_seconds_data->leap_months[12]);
                assert_equal(1,
                    l0r_sensor_data.leap_seconds_data->leap_days[12]);
                assert_equal(1,
                    l0r_sensor_data.leap_seconds_data->num_leap_seconds[12]);
                break;
            case 13:
                assert_equal(1988,
                    l0r_sensor_data.leap_seconds_data->leap_years[13]);
                assert_equal(1,
                    l0r_sensor_data.leap_seconds_data->leap_months[13]);
                assert_equal(1,
                    l0r_sensor_data.leap_seconds_data->leap_days[13]);
                assert_equal(1,
                    l0r_sensor_data.leap_seconds_data->num_leap_seconds[13]);
                break;
            case 22:
                assert_equal(2006,
                    l0r_sensor_data.leap_seconds_data->leap_years[22]);
                assert_equal(1,
                    l0r_sensor_data.leap_seconds_data->leap_months[22]);
                assert_equal(1,
                    l0r_sensor_data.leap_seconds_data->leap_days[22]);
                assert_equal(1,
                    l0r_sensor_data.leap_seconds_data->num_leap_seconds[22]);
                break;
            case 1: case 2: case 3: case 4: case 6: case 7: case 8: case 9:
            case 10: case 11: case 14: case 15: case 16: case 17: case 18:
            case 19: case 20: case 21: case 23: case 24:
                /* No validation (would get too long); avoid fall to default */
                break;
            default:
                /* This should never happen; gone beyond expected leap
                   second count */
                assert_true(0);
                break;
        }
    }

    ias_los_model_free(model);
    ias_cpf_free(cpf);
}

void cleanup_l0r_sensor_data
(
    IAS_SENSOR_L0R_DATA *l0r_sensor_data,   /* I: struct to free */
    int band_count                          /* I: number of bands */
)
{
    if (l0r_sensor_data->oli_data.frame_headers != NULL)
        free(l0r_sensor_data->oli_data.frame_headers);
    if (l0r_sensor_data->tirs_data.frame_headers != NULL)
        free(l0r_sensor_data->tirs_data.frame_headers);
    if (l0r_sensor_data->l0r_detector_offsets != NULL)
    {
        int i;
        for (i = 0; i < band_count; i++)
        {
            if ((l0r_sensor_data->bands_present[i] == 1) &&
                    (l0r_sensor_data->l0r_detector_offsets[i] != NULL))
            {
                free(l0r_sensor_data->l0r_detector_offsets[i]);
            }
        }
        free(l0r_sensor_data->l0r_detector_offsets);
    }
}

/* When the input sensor indicates OLI, this uses an OLI_TIRS combined image,
   delivered by CalVal with OLI create model ADD version 3.5a.
   When the input sensor indicates TIRS, this uses a TIRS-only image, delivered
   by CalVal with TIRS create model ADD version 3.4b. */
int read_test_l0r
(
    IAS_SENSOR_ID sensor,           /* I: sensor flag - different L0Rs */
    IAS_LOS_MODEL *model,
    IAS_SENSOR_L0R_DATA *l0r_sensor_data,
    IAS_SENSOR_SSM_RECORD **expected_ssm_records    /* O */
)
{
    const char *test_data_dir;
    char l0r_product_dir[PATH_MAX];
    char l0r_id[IAS_L0R_INTERVAL_ID_LENGTH + 1];
    L0RIO *l0r_ptr;
    int status;
    int num_frames;
    uint16_t *l0r_detector_offsets;
    int i;
    int band_num;
    int scas;
    int dets;
    int dets_total;
    #include "set_l0r_test_data.h"
    int input_count = sizeof(input_records)/sizeof(input_records[0]);
    *expected_ssm_records = expected_records;

    test_data_dir = getenv("IAS_UNIT_TEST_DATA_DIR");
    assert_true(test_data_dir != NULL);
    if (test_data_dir == NULL)
    {
        assert_true_with_message(0, "IAS_UNIT_TEST_DATA_DIR isn't set");
        return ERROR;
    }

    /* Initialize the L0R sensor data structure */
    memset(l0r_sensor_data, 0, sizeof(*l0r_sensor_data));

    if (sensor == IAS_OLI)
    {
        status = snprintf(l0r_product_dir, sizeof(l0r_product_dir),
                "%s/ias_lib/los_model/sensor/l0ra/", test_data_dir);
    }
    else
    {
        status = snprintf(l0r_product_dir, sizeof(l0r_product_dir),
                "%s/ias_lib/los_model/sensor/l0ra_tirs/", test_data_dir);
    }
    if (status < 0 || status >= sizeof(l0r_product_dir))
    {
        assert_true_with_message(0, "Error assembling L0R product "
                "directory string");
        return ERROR;
    }
    strcpy(l0r_id, "test");

    l0r_ptr = ias_l0r_open(l0r_id, l0r_product_dir, 0);
    assert_not_equal(l0r_ptr, NULL);
    if (l0r_ptr == NULL)
    {
        assert_true_with_message(0, "Error opening L0R %s/%s", l0r_product_dir,
                l0r_id);
        return ERROR;
    }

    status = ias_l0r_open_metadata(l0r_ptr, IAS_READ);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error opening L0R metadata for L0R %s/%s",
                l0r_product_dir, l0r_id);
        ias_l0r_close(l0r_ptr);
        return ERROR;
    }

    status = ias_l0r_open_header(l0r_ptr, IAS_READ);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error opening L0R header for L0R %s/%s",
                l0r_product_dir, l0r_id);
        ias_l0r_close(l0r_ptr);
        return ERROR;
    }

    if (sensor == IAS_OLI)
    {
        status = ias_l0r_get_oli_image_header(l0r_ptr,
                &l0r_sensor_data->image_header);
        assert_equal(status, SUCCESS);
        if (status != ERROR)
            l0r_sensor_data->image_header_flag = 1;
        else
            l0r_sensor_data->image_header_flag = 0;

        status = ias_l0r_get_oli_frame_headers_records_count(l0r_ptr,
                &num_frames);
        assert_equal(status, SUCCESS);
        if (status != SUCCESS)
        {
            assert_true_with_message(0, "Error getting OLI frame header "
                    "records count for L0R %s/%s", l0r_product_dir, l0r_id);
            ias_l0r_close(l0r_ptr);
            return ERROR;
        }
        l0r_sensor_data->oli_data.num_frames = num_frames;

        l0r_sensor_data->oli_data.frame_headers
            = malloc(sizeof(*l0r_sensor_data->oli_data.frame_headers)
                    * l0r_sensor_data->oli_data.num_frames);
        assert_not_equal(l0r_sensor_data->oli_data.frame_headers, NULL);
        if (l0r_sensor_data->oli_data.frame_headers == NULL)
        {
            assert_true_with_message(0, "Error allocating OLI frame headers");
            ias_l0r_close(l0r_ptr);
            return ERROR;
        }

        status = ias_l0r_get_oli_frame_headers(l0r_ptr, 0,
                l0r_sensor_data->oli_data.num_frames,
                l0r_sensor_data->oli_data.frame_headers);
        assert_equal(status, SUCCESS);
        if (status != SUCCESS)
        {
            assert_true_with_message(0, "Error getting OLI frame headers for "
                    "L0R %s/%s", l0r_product_dir, l0r_id);
            ias_l0r_close(l0r_ptr);
            cleanup_l0r_sensor_data(l0r_sensor_data, model->sensor.band_count);
            return ERROR;
        }

        l0r_sensor_data->oli_data.sensor_present = 1;
    }
    else
        l0r_sensor_data->oli_data.sensor_present = 0;

    status = ias_l0r_get_tirs_frame_headers_records_count(l0r_ptr, &num_frames);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error getting TIRS frame header "
                "records count for L0R %s/%s", l0r_product_dir, l0r_id);
        ias_l0r_close(l0r_ptr);
        cleanup_l0r_sensor_data(l0r_sensor_data, model->sensor.band_count);
        return ERROR;
    }
    l0r_sensor_data->tirs_data.num_frames = num_frames;

    l0r_sensor_data->tirs_data.frame_headers
        = malloc(sizeof(*l0r_sensor_data->tirs_data.frame_headers)
        * l0r_sensor_data->tirs_data.num_frames);
    assert_not_equal(l0r_sensor_data->tirs_data.frame_headers, NULL);
    if (l0r_sensor_data->tirs_data.frame_headers == NULL)
    {
        assert_true_with_message(0, "Error allocating TIRS frame headers");
        ias_l0r_close(l0r_ptr);
        cleanup_l0r_sensor_data(l0r_sensor_data, model->sensor.band_count);
        return ERROR;
    }

    status = ias_l0r_get_tirs_frame_headers(l0r_ptr, 0,
            l0r_sensor_data->tirs_data.num_frames,
            l0r_sensor_data->tirs_data.frame_headers);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error getting TIRS frame headers for "
                "L0R %s/%s", l0r_product_dir, l0r_id);
        ias_l0r_close(l0r_ptr);
        cleanup_l0r_sensor_data(l0r_sensor_data, model->sensor.band_count);
        return ERROR;
    }

    l0r_sensor_data->tirs_data.sensor_present = 1;

    l0r_sensor_data->l0r_detector_offsets = malloc(sizeof(l0r_detector_offsets)
            * model->sensor.band_count);
    assert_not_equal(l0r_sensor_data->l0r_detector_offsets, NULL);
    if (l0r_sensor_data->l0r_detector_offsets == NULL)
    {
        assert_true_with_message(0, "Error allocating detector offsets");
        ias_l0r_close(l0r_ptr);
        cleanup_l0r_sensor_data(l0r_sensor_data, model->sensor.band_count);
        return ERROR;
    }

    /* Set the TIRS SSM encoding records */
    l0r_sensor_data->tirs_data.ssm_records = input_records;
    l0r_sensor_data->tirs_data.num_ssm_records = input_count;

    for (i = 0; i < model->sensor.band_count; i++)
    {
        int band_present;   /* Used to check if band is in the L0R */

        band_num = ias_sat_attr_convert_band_index_to_number(i);

        /* Check if the band is in the image */
        status = ias_l0r_get_band_present(l0r_ptr, band_num, &band_present);
        assert_equal(status, SUCCESS);
        if (status != SUCCESS)
        {
            assert_true_with_message(0, "Error checking for presence of band "
                    "number %d in L0R %s/%s", band_num, l0r_product_dir,
                    l0r_id);
            ias_l0r_close(l0r_ptr);
            cleanup_l0r_sensor_data(l0r_sensor_data, model->sensor.band_count);
            return ERROR;
        }

        /* If the current band is not in the L0R, go to the next band */
        if (!band_present)
            continue;

        status = ias_l0r_open_band(l0r_ptr, band_num, IAS_READ);
        assert_equal(status, SUCCESS);
        if (status != SUCCESS)
        {
            assert_true_with_message(0, "Error opening band number %d for "
                    "L0R %s/%s", band_num, l0r_product_dir, l0r_id);
            ias_l0r_close(l0r_ptr);
            cleanup_l0r_sensor_data(l0r_sensor_data, model->sensor.band_count);
            return ERROR;
        }

        scas = model->sensor.bands[i].sca_count;
        dets = model->sensor.bands[i].scas[0].detectors;
        dets_total = scas * dets;

        l0r_detector_offsets = malloc(sizeof(*l0r_detector_offsets)
                * dets_total);
        assert_not_equal(l0r_detector_offsets, NULL);
        if (l0r_detector_offsets == NULL)
        {
            assert_true_with_message(0, "Error allocating detector offsets");
            ias_l0r_close(l0r_ptr);
            cleanup_l0r_sensor_data(l0r_sensor_data, model->sensor.band_count);
            return ERROR;
        }

        status = ias_l0r_get_top_detector_offsets(l0r_ptr, band_num,
                l0r_detector_offsets, dets_total);
        assert_equal(status, SUCCESS);
        if (status != SUCCESS)
        {
            assert_true_with_message(0, "Error getting detector offsets");
            ias_l0r_close(l0r_ptr);
            cleanup_l0r_sensor_data(l0r_sensor_data, model->sensor.band_count);
            return ERROR;
        }

        l0r_sensor_data->l0r_detector_offsets[i] = l0r_detector_offsets;
        l0r_detector_offsets = NULL;

        /* set the bands_present flag */
        l0r_sensor_data->bands_present[i] = 1;

        /* While ias_l0r_close will close the bands, closing the bands manually
           has been found to be helpful for debugging some issues. It is good
           improvement here anyway. */
        status = ias_l0r_close_band(l0r_ptr, band_num);
        assert_equal(status, SUCCESS);  /* error closing is not test failure */
    }

    ias_l0r_close(l0r_ptr);

    return SUCCESS;
}

/* test the set L0R routine */
/* The Cal/Val test data given with OLI Model Creation V3.5a did not contain
   complete TIRS data and what was there was not correct. Left some aspects
   of TIRS validation in this unit test where there were at least values
   (even if incorrect as compared to the OLI data in the same image) to still
   test the OLI/TIRS combined implementation aspects of the set_l0r code.
   Created a separate unit test to specifically test the TIRS frame times and
   SSM data with a different image, containing good TIRS data. */
Ensure test_set_l0r()
{
    #include "smooth_oli_frame_times_test_data.h"

    const struct IAS_CPF_OLI_PARAMETERS *cpf_oli;
    const struct IAS_CPF_TIRS_PARAMETERS *cpf_tirs;
    const struct IAS_CPF_EARTH_CONSTANTS *cpf_earth_consts;
    IAS_LOS_MODEL *model;
    IAS_SENSOR_MODEL *sensor;
    IAS_SENSOR_L0R_DATA l0r_sensor_data;
    IAS_CPF *cpf;
    /* pointer not used by this test, but a required param of a function */
    IAS_SENSOR_SSM_RECORD *unused_expected_results_ptr = NULL;
    int status;
    int band_num;
    int i, j, k;

    IAS_SSM_POSITION_QUALITY ssm_positions_quality;

    /* Set the log level to warnings only to avoid the debugging output
       to clutter up the test output. */
    ias_log_set_output_level(IAS_LOG_LEVEL_WARN);

    /* initialize the satellite attributes */
    ias_sat_attr_initialize(IAS_L8);

    /* allocate a model */
    model = ias_los_model_allocate();
    assert_not_equal(model, NULL);
    if (model == NULL)
        return;

    sensor = &model->sensor;

    /* Read L0R */
    status = read_test_l0r(IAS_OLI, model, &l0r_sensor_data,
            &unused_expected_results_ptr);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
    {
        ias_los_model_free(model);
        return;
    }

    cpf = read_test_cpf(IAS_OLI);
    assert_not_equal(cpf, NULL);
    if (cpf == NULL)
    {
        cleanup_l0r_sensor_data(&l0r_sensor_data, model->sensor.band_count);
        ias_los_model_free(model);
        return;
    }

    /* Need the nominal frame times */
    cpf_oli = ias_cpf_get_oli_parameters(cpf);
    assert_not_equal(cpf_oli, NULL);
    if (cpf_oli == NULL)
    {
        cleanup_l0r_sensor_data(&l0r_sensor_data, model->sensor.band_count);
        ias_los_model_free(model);
        ias_cpf_free(cpf);
        return;
    }

    /* The CPF info needs to be set into the model before setting the L0R
       info */
    status = ias_sensor_set_cpf(cpf, &l0r_sensor_data, sensor);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
    {
        cleanup_l0r_sensor_data(&l0r_sensor_data, model->sensor.band_count);
        ias_los_model_free(model);
        ias_cpf_free(cpf);
        return;
    }

    /* Convert milliseconds in CPF to seconds that the model frame time units
       will be in */
    l0r_sensor_data.oli_data.nominal_frame_time = cpf_oli->nominal_frame_time
        * 1.0e-3;

    cpf_tirs = ias_cpf_get_tirs_parameters(cpf);
    assert_not_equal(cpf_tirs, NULL);
    if (cpf_tirs == NULL)
    {
        cleanup_l0r_sensor_data(&l0r_sensor_data, model->sensor.band_count);
        ias_los_model_free(model);
        ias_cpf_free(cpf);
        return;
    }

    /* Convert milliseconds in CPF to seconds that the model frame time units
       will be in */
    l0r_sensor_data.tirs_data.nominal_frame_time = cpf_tirs->nominal_frame_time
        * 1.0e-3;

    /* The leap second information */
    cpf_earth_consts = ias_cpf_get_earth_const(cpf);
    assert_not_equal(cpf_earth_consts, NULL);
    if (cpf_earth_consts == NULL)
    {
        cleanup_l0r_sensor_data(&l0r_sensor_data, model->sensor.band_count);
        ias_los_model_free(model);
        ias_cpf_free(cpf);
        return;
    }

    l0r_sensor_data.leap_seconds_data = &cpf_earth_consts->leap_seconds_data;

    /* Set the L0R source info into the model. Don't exit if failure; continue
       to see detailed errors for assistance debugging. */
    /* The TIRS data is not valid for this image meant more for testing the
       OLI frame times and the combined OLI/TIRS routines we have which is
       different from Cal/Val's implementation. As such, the TIRS SSM data is
       not reliable for testing and does generate an error, so disable logging
       to avoid seeing that error on the testing output. */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = ias_sensor_set_l0r(0, cpf, &l0r_sensor_data, 
                                "est_ssm_encoder.data", sensor, 
                                &ssm_positions_quality);
    assert_equal(status, SUCCESS);
    assert_equal(ssm_positions_quality, IAS_SSM_POSITION_ACTUAL);
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

    /* Frame counts */
    for (i = 0; i < sensor->band_count; i++)
    {
        const IAS_BAND_ATTRIBUTES *band_attribs;

        band_num = ias_sat_attr_convert_band_index_to_number(i);
        band_attribs = ias_sat_attr_get_band_attributes(band_num);
        assert_not_equal(band_attribs, NULL);  /* continue testing all bands */

        if (sensor->bands[i].sensor->sensor_id == IAS_OLI)
            assert_equal(OLI_EXPECTED_NUM_FRAMES, sensor->bands[i].frame_count);
        else
        {
            /* For the TIRS bands, confirm the expected number of frames.
               This is different than OLI due to the different sampling rates
               and framing. */
            assert_equal(4270, sensor->bands[i].frame_count);
        }

        assert_equal(2014, (int)sensor->bands[i].utc_epoch_time[0]);
        assert_equal(163, (int)sensor->bands[i].utc_epoch_time[1]);
        if (sensor->bands[i].sensor->sensor_id == IAS_OLI)
        {
            assert_double_equal(59358.2491922012,
                    sensor->bands[i].utc_epoch_time[2]);
        }
        else
        {
            assert_double_equal(59357.01428598,
                    sensor->bands[i].utc_epoch_time[2]);
        }

        /* Integration time */
        switch(i)
        {
            case 0:
            case 1:
            case 2:
            case 3:
            case 4:
            case 5:
            case 6:
            case 8:
                /* the L0R has integration times of 0.0, but the CPF value
                   is set into it */
                assert_double_equal(0.0036, 
                        sensor->bands[i].sampling_char.integration_time);
                break;
            case 7:
                /* the L0R has integration times of 0.0, but the CPF value
                   is set into it */
                assert_double_equal(0.000660,
                        sensor->bands[i].sampling_char.integration_time);
                break;
            case 9:
            case 10:
                /* TIRS integration time comes from the L0R */
                assert_double_equal(0.00200,
                        sensor->bands[i].sampling_char.integration_time);
                break;
            default:
                /* this should never happen */
                assert_true(0);
                break;
        }

        if (sensor->bands[i].sensor->sensor_id == IAS_OLI)
        {
            assert_equal(sensor->frame_seconds_from_epoch[IAS_OLI],
                    sensor->bands[i].frame_seconds_from_epoch);
            if (band_attribs->band_type == IAS_PAN_BAND)
            {
                assert_double_equal(0.004235947 / PAN_SCALE,
                        sensor->bands[i].sampling_char.sampling_time);
            }
            else
            {
                assert_double_equal(0.004235947,
                        sensor->bands[i].sampling_char.sampling_time);
            }
        }
        else
        {
            assert_equal(sensor->frame_seconds_from_epoch[IAS_TIRS],
                    sensor->bands[i].frame_seconds_from_epoch);
            assert_double_equal(0.0142857143,
                    sensor->bands[i].sampling_char.sampling_time);
        }

        /* J is the SCA number, K is the detector offset index. Just checking
           some of them since would be a lot of values and a lot of time to
           check them all. If the first so many of them are correct, no
           problems should be present. */
        /* Ingest no longer does the even/odd detector alignment.
           Since the Detector_Offsets in the L0R image files is to report the
           even/odd detector alignments ingest has done, having them be all
           zeros now is correct. In theory, the Detector_Offsets could be
           removed from the L0R file and the code to deal with it removed, but
           it was decided to keep it around... just in case someone found a
           valid reason to have Ingest do the even/odd alignment in the
           future. Due to this change, this test was changed to ensure these
           values are all 0's. */
        for (j = 0; j < 2; j++)
        {
            for (k = 0; k < 5; k++)
            {
                if (sensor->bands[i].sensor->sensor_id == IAS_OLI)
                {
                    if (band_attribs->band_type == IAS_PAN_BAND)
                    {
                        assert_equal(0, sensor->bands[i].scas[j].
                                l0r_detector_offsets[k]);
                    }
                    else
                    {
                        assert_equal(0, sensor->bands[i].scas[j].
                                l0r_detector_offsets[k]);
                    }
                }
                else
                {
                    assert_equal(0,
                            sensor->bands[i].scas[j].l0r_detector_offsets[k]);
                }
            }
        }
    }

    /* Check the smoothed frame times */
    assert_equal(OLI_EXPECTED_NUM_FRAMES, sensor->frame_counts[IAS_OLI]);
    for (i = 0; i < sensor->frame_counts[IAS_OLI]; i++)
    {
        if (expected_frame_times[i] != 
                sensor->frame_seconds_from_epoch[IAS_OLI][i])
        {
            assert_double_equal(expected_frame_times[i],
                    sensor->frame_seconds_from_epoch[IAS_OLI][i]);
        }
    }
    assert_equal(9147, l0r_sensor_data.oli_data.num_bad_frame_times);

    cleanup_l0r_sensor_data(&l0r_sensor_data, model->sensor.band_count);
    ias_los_model_free(model);
    ias_cpf_free(cpf);
}

/* test the set L0R routine */
/* This version focuses on checking the TIRS smoothed frame time codes and SSM
   data with new Cal/Val test data given with the TIRS Create Model V3.4b ADD */
Ensure test_set_l0r_tirs()
{
    #include "smooth_tirs_frame_times_test_data.h"

    const struct IAS_CPF_TIRS_PARAMETERS *cpf_tirs;
    const struct IAS_CPF_EARTH_CONSTANTS *cpf_earth_consts;
    IAS_LOS_MODEL *model;
    IAS_SENSOR_MODEL *sensor;
    IAS_SENSOR_L0R_DATA l0r_sensor_data;
    IAS_SENSOR_SCENE_SELECT_MIRROR_MODEL *ssm_model;
    IAS_CPF *cpf;
    IAS_SENSOR_SSM_RECORD *expected_ssm_records;
    int status;
    int band_num;
    int i;

    char ssm_positions_filename[PATH_MAX] = "";
    IAS_SSM_POSITION_QUALITY ssm_positions_quality;

    /* Set the log level to warnings only to avoid the debugging output
       to clutter up the test output. */
    ias_log_set_output_level(IAS_LOG_LEVEL_WARN);

    /* Initialize the satellite attributes */
    ias_sat_attr_initialize(IAS_L8);

    /* Allocate a model */
    model = ias_los_model_allocate();
    assert_not_equal(model, NULL);
    if (model == NULL)
        return;

    sensor = &model->sensor;

    /* Read L0R */
    status = read_test_l0r(IAS_TIRS, model, &l0r_sensor_data,
            &expected_ssm_records);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
    {
        ias_los_model_free(model);
        return;
    }

    /* Read CPF */
    cpf = read_test_cpf(IAS_TIRS);
    assert_not_equal(cpf, NULL);
    if (cpf == NULL)
    {
        cleanup_l0r_sensor_data(&l0r_sensor_data, model->sensor.band_count);
        ias_los_model_free(model);
        return;
    }

    /* The CPF info needs to be set into the model before setting the L0R
       info */
    status = ias_sensor_set_cpf(cpf, &l0r_sensor_data, sensor);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
    {
        cleanup_l0r_sensor_data(&l0r_sensor_data, model->sensor.band_count);
        ias_los_model_free(model);
        ias_cpf_free(cpf);
        return;
    }

    /* Get the TIRS CPF parameter group */
    cpf_tirs = ias_cpf_get_tirs_parameters(cpf);
    assert_not_equal(cpf_tirs, NULL);
    if (cpf_tirs == NULL)
    {
        cleanup_l0r_sensor_data(&l0r_sensor_data, model->sensor.band_count);
        ias_los_model_free(model);
        ias_cpf_free(cpf);
        return;
    }

    /* Convert milliseconds in CPF to seconds that the model frame time units
       will be in */
    l0r_sensor_data.tirs_data.nominal_frame_time = cpf_tirs->nominal_frame_time
        * 1.0e-3;

    /* The leap second information */
    cpf_earth_consts = ias_cpf_get_earth_const(cpf);
    assert_not_equal(cpf_earth_consts, NULL);
    if (cpf_earth_consts == NULL)
    {
        cleanup_l0r_sensor_data(&l0r_sensor_data, model->sensor.band_count);
        ias_los_model_free(model);
        ias_cpf_free(cpf);
        return;
    }

    l0r_sensor_data.leap_seconds_data = &cpf_earth_consts->leap_seconds_data;

    /* Set the L0R source info into the model. Don't exit if failure; continue
       to see detailed errors for assistance debugging. */
    ias_log_set_output_level(IAS_LOG_LEVEL_ERROR);
    status = ias_sensor_set_l0r(0, cpf, &l0r_sensor_data,
        ssm_positions_filename, sensor, &ssm_positions_quality);
    assert_equal(ssm_positions_quality, IAS_SSM_POSITION_ACTUAL);
    assert_equal(status, SUCCESS);
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

    /* Frame counts */
    for (i = 0; i < sensor->band_count; i++)
    {
        const IAS_BAND_ATTRIBUTES *band_attribs;

        /* Skip the OLI bands not in this TIRS test image */
        if (sensor->bands[i].band_present == 0)
            continue;

        band_num = ias_sat_attr_convert_band_index_to_number(i);
        band_attribs = ias_sat_attr_get_band_attributes(band_num);
        assert_not_equal(band_attribs, NULL);  /* continue testing all bands */

        assert_equal(TIRS_EXPECTED_NUM_FRAMES, sensor->bands[i].frame_count);
        assert_equal(2003, (int)sensor->bands[i].utc_epoch_time[0]);
        assert_equal(148, (int)sensor->bands[i].utc_epoch_time[1]);
        assert_double_equal(65040.10076, sensor->bands[i].utc_epoch_time[2]);

        /* Integration time */
        switch(band_num)
        {
            case 10:
            case 11:
                /* TIRS integration time comes from the L0R */
                assert_double_equal(0.00500,
                        sensor->bands[i].sampling_char.integration_time);
                break;
            default:
                /* This should never happen */
                assert_true(0);
                break;
        }

        assert_equal(sensor->frame_seconds_from_epoch[IAS_TIRS],
                sensor->bands[i].frame_seconds_from_epoch);
        assert_double_equal(0.014286,
                sensor->bands[i].sampling_char.sampling_time);
    }

    /* Check the smoothed frame times */
    assert_equal(TIRS_EXPECTED_NUM_FRAMES, sensor->frame_counts[IAS_TIRS]);
    for (i = 0; i < sensor->frame_counts[IAS_TIRS]; i++)
    {
        if (expected_frame_times[i] !=
                sensor->frame_seconds_from_epoch[IAS_TIRS][i])
        {
            assert_double_equal(expected_frame_times[i],
                    sensor->frame_seconds_from_epoch[IAS_TIRS][i]);
        }
    }
    assert_equal(0, l0r_sensor_data.tirs_data.num_bad_frame_times);

    /* The following test is pretty much a duplicate of test_set_ssm_from_l0r()
       but keeping it here as well since this provides an overall test of
       ias_sensor_set_l0r() and makes sure the lower-level routines were
       executed and succeeded as part of the higher-level call */
    /* Cache a pointer to the SSM model */
    ssm_model = sensor->sensors[IAS_TIRS].ssm_model;

    /* Confirm TIRS had the correct number of data items corrected */
    assert_equal(l0r_sensor_data.tirs_data.num_bad_ssm_times, 3);
    assert_equal(l0r_sensor_data.tirs_data.num_bad_ssm_encoder_values, 2);
    assert_equal(ssm_model->ssm_record_count, 740);

    assert_double_equal(sensor->bands[10].utc_epoch_time[0],
            ssm_model->utc_epoch_time[0]);
    assert_double_equal(sensor->bands[10].utc_epoch_time[1],
            ssm_model->utc_epoch_time[1]);
    assert_double_equal(sensor->bands[10].utc_epoch_time[2],
            ssm_model->utc_epoch_time[2]);

    for (i = 0; i < ssm_model->ssm_record_count; i++)
    {
        IAS_SENSOR_SSM_RECORD *r = &ssm_model->records[i];
        IAS_SENSOR_SSM_RECORD *er = &expected_ssm_records[i];

        /* make sure the mirror angle agrees */
        if (!doubles_are_equal(r->mirror_angle, er->mirror_angle))
        {
            assert_true_with_message(0, "Index %d mirror angle results "
                    "disagree.  Got %.7e Expected %.7e", i, 
                    r->mirror_angle, er->mirror_angle);
        }
    }

    cleanup_l0r_sensor_data(&l0r_sensor_data, model->sensor.band_count);
    ias_los_model_free(model);
    ias_cpf_free(cpf);
}

/* Test the SSM smoothing routine */
Ensure test_ssm_smoothing()
{
    #include "smooth_ssm_test_data.h"

    int status;
    int index;

    /* Call the smoothing routine */
    status = ias_sensor_smooth_ssm_data(5, &input_ssm_model);
    assert_equal(status, SUCCESS);

    /* Verify the results */
    for (index = 0; index < expected_ssm_model.ssm_record_count; index++)
    {
        assert_double_equal(input_records[index].mirror_angle,
                            expected_records[index].mirror_angle);
    }

    /* Now check an error condition */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    input_ssm_model.ssm_record_count = 4;
    status = ias_sensor_smooth_ssm_data(5, &input_ssm_model);
    assert_equal(status, ERROR);

    ias_log_set_output_level(IAS_LOG_LEVEL_WARN);

    /* 5 records should pass */
    input_ssm_model.ssm_record_count = 5;
    status = ias_sensor_smooth_ssm_data(5, &input_ssm_model);
    assert_equal(status, SUCCESS);
}

/* Test the ias_sensor_align_ssm_data routine */
Ensure test_align_ssm_data()
{
    int index;
    #include "align_ssm_encoder_data.h"
    int count = sizeof(input_records)/sizeof(input_records[0]);

    ias_sensor_align_ssm_data(&input_records[0], count);

    /* Verify the results */
    for (index = 0; index < count; index++)
    {
        int i;
        for (i = 0;
                i < IAS_L0R_TIRS_TELEMETRY_SSM_ENCODER_POSITION_SAMPLE_COUNT;
                i++)
        {
            if (input_records[index].encoder_position[i]
                    != expected_records[index].encoder_position[i])
            {
                assert_double_equal(input_records[index].encoder_position[i],
                        expected_records[index].encoder_position[i]);
            }
        }
    }
}

/* Test the ias_sensor_check_ssm_encoder_data routine */
Ensure test_check_ssm_encoder_data()
{
    int status;
    int index;
    int num_bad_ssm_times;
    int num_bad_ssm_encoder_values;
    const struct IAS_CPF_TIRS_PARAMETERS *tirs_params;
    #include "check_ssm_encoder_data.h"
    int count = sizeof(input_records)/sizeof(input_records[0]);

    /* Initialize the satellite attributes */
    ias_sat_attr_initialize(IAS_L8);

    /* Get the CPF */
    IAS_CPF *cpf = read_test_cpf(IAS_TIRS);
    assert_true(cpf != NULL);
    if (cpf == NULL)
        return;

    tirs_params = ias_cpf_get_tirs_parameters(cpf);
    assert_true(tirs_params != NULL);
    if (tirs_params == NULL)
    {
        ias_cpf_free(cpf);
        return;
    }

    /* Call the routine */
    printf(".... expect a message about fixing 2 SSM encoder values "
           "with stuck bits\n");
    status = ias_sensor_check_ssm_encoder_data(input_records, count,
            tirs_params->ssm_encoder_origin_side_a, cpf, &num_bad_ssm_times,
            &num_bad_ssm_encoder_values);
    assert_equal(status, SUCCESS);

    /* Confirm the number of data items corrected */
    assert_equal(num_bad_ssm_times, 3);
    assert_equal(num_bad_ssm_encoder_values, 4);

    for (index = 0; index < count; index++)
    {
        IAS_L0R_TIRS_SSM_ENCODER *ir = &input_records[index];
        IAS_L0R_TIRS_SSM_ENCODER *er = &expected_records[index];

        /* Make sure each member is as expected */
        if (!doubles_are_equal(er->l0r_time.seconds_of_day,
                    ir->l0r_time.seconds_of_day)
                || (er->l0r_time.days_from_J2000
                    != ir->l0r_time.days_from_J2000)
                || (er->encoder_position[0] != ir->encoder_position[0])
                || (er->encoder_position[1] != ir->encoder_position[1])
                || (er->encoder_position[2] != ir->encoder_position[2])
                || (er->encoder_position[3] != ir->encoder_position[3]))
        {
            assert_true_with_message(0, "Index %d results disagree", index);
        }
    }

    ias_cpf_free(cpf);
}

/* Test the ias_sensor_set_ssm_from_l0r routine */
Ensure test_set_ssm_from_l0r()
{
    double round(double);   /* Prototype for rounding */
    int status;
    TIRS_SENSOR_L0R_DATA l0r_data;
    IAS_SENSOR_MODEL sensor_model;
    IAS_SENSOR_BAND_MODEL sensor_bands[1];
    IAS_SENSOR_SCENE_SELECT_MIRROR_MODEL ssm_model;
    const struct IAS_CPF_EARTH_CONSTANTS *earth_params;
    int index;
    double seconds_from_j2000;
    #include "set_ssm_from_l0r_test_data.h"
    int input_count = sizeof(input_records) / sizeof(input_records[0]);
    int expected_count = sizeof(expected_records) / sizeof(expected_records[0]);

    IAS_SSM_POSITION_QUALITY ssm_positions_quality;

    /* Initialize the ssm_position_select field */
    for (index = 0; index < input_count; index++)
    {
        input_records[index].ssm_position_sel = IAS_L0R_TIRS_SSM_NADIR_POSITION;
    }

    /* Initialize the satellite attributes */
    ias_sat_attr_initialize(IAS_L8);

    /* Get the CPF */
    IAS_CPF *cpf = read_test_cpf(IAS_TIRS);
    assert_true(cpf != NULL);
    if (cpf == NULL)
        return;

    earth_params = ias_cpf_get_earth_const(cpf);
    assert_true(earth_params != NULL);
    if (!earth_params)
    {
        ias_cpf_free(cpf);
        return;
    }

    seconds_from_j2000 = input_records[0].l0r_time.days_from_J2000
            * IAS_SEC_PER_DAY + input_records[0].l0r_time.seconds_of_day;
    status = ias_math_init_leap_seconds(seconds_from_j2000, 
        &earth_params->leap_seconds_data);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
    {
        ias_cpf_free(cpf);
        return;
    }

    /* Set up l0r_data structure */
    memset(&l0r_data, 0, sizeof(l0r_data));
    l0r_data.ssm_records = input_records;
    l0r_data.num_ssm_records = input_count;

    /* Set up the sensor_model structure */
    memset(sensor_bands, 0, sizeof(sensor_bands));
    memset(&sensor_model, 0, sizeof(sensor_model));
    memset(&ssm_model, 0, sizeof(ssm_model));

    sensor_bands[0].utc_epoch_time[0] = input_epoch[0];
    sensor_bands[0].utc_epoch_time[1] = input_epoch[1];
    sensor_bands[0].utc_epoch_time[2] = input_epoch[2];
    sensor_bands[0].band_present = 1;
    sensor_bands[0].sensor = &sensor_model.sensors[IAS_TIRS];
    sensor_model.bands = sensor_bands;
    sensor_model.band_count = 1;
    sensor_model.sensors[IAS_TIRS].sensor_id = IAS_TIRS;
    sensor_model.sensors[IAS_TIRS].sensor_present = 1;
    sensor_model.sensors[IAS_TIRS].ssm_model = &ssm_model;
    sensor_model.frame_counts[IAS_OLI] = 0;
    /* Mocking up the frame times to cover the entire image. The set SSM
       code uses nothing more of this frame info than to determine the time
       duration, so mocking up 2 frames that result in the same duration as
       the L0R used to generate the header file test data is sufficient. */
    sensor_model.frame_counts[IAS_TIRS] = 2;
    sensor_model.frame_seconds_from_epoch[IAS_TIRS] = frame_times;

    /* Call the routine */
    ssm_positions_quality = IAS_SSM_POSITION_ACTUAL;
    status = ias_sensor_set_ssm_from_l0r(0, &l0r_data, cpf, &sensor_model,
                IAS_TIRS, NULL, &ssm_positions_quality);
    assert_equal(status, SUCCESS);
    assert_equal(ssm_positions_quality, IAS_SSM_POSITION_ACTUAL);

    /* Confirm the number of data items corrected */
    assert_equal(l0r_data.num_bad_ssm_times, 6);
    assert_equal(l0r_data.num_bad_ssm_encoder_values, 4);
    assert_equal(ssm_model.ssm_record_count, expected_count);
    assert_equal(ssm_model.position_status, IAS_SSM_POSITION_DEFAULT);
    for (index = 0; index < ssm_model.ssm_record_count; index++)
    {
        IAS_SENSOR_SSM_RECORD *r = &ssm_model.records[index];
        IAS_SENSOR_SSM_RECORD *er = &expected_records[index];
        double expected_seconds;
        double our_seconds;

        /* Round the seconds to the nearest microsecond since any more
           precise comparison is foolish */
        expected_seconds = round(er->seconds_from_epoch * 1.0e6) * 1.0e-6;
        our_seconds = round(r->seconds_from_epoch * 1.0e6) * 1.0e-6;

        /* Make sure each of the members is as expected */
        if (!doubles_are_equal(our_seconds, expected_seconds)
            || !doubles_are_equal(r->mirror_angle,
                                  er->mirror_angle))
        {
            assert_true_with_message(0, "Index %d results disagree.  "
                "Got %.7e, %.7e Expected %.7e, %.7e", index, 
                r->seconds_from_epoch, r->mirror_angle,
                er->seconds_from_epoch, er->mirror_angle);
        }
    }

    free(sensor_model.sensors[IAS_TIRS].ssm_model->records);
    ias_cpf_free(cpf);
}

/* Test the ias_sensor_set_ssm_from_l0r routine */
/* Note that this identical to the test above, except for a different
   *.h that gets included to force looking for an estimated ssm file.
   Want to make sure the results are the same when the ssm file is not
   provided */
Ensure test_null_ssm_filename()
{
    double round(double);   /* Prototype for rounding */
    int status;
    TIRS_SENSOR_L0R_DATA l0r_data;
    IAS_SENSOR_MODEL sensor_model;
    IAS_SENSOR_BAND_MODEL sensor_bands[1];
    IAS_SENSOR_SCENE_SELECT_MIRROR_MODEL ssm_model;
    const struct IAS_CPF_EARTH_CONSTANTS *earth_params;
    int index;
    double seconds_from_j2000;
    #include "set_ssm_from_l0r_null_filename_test.h"
    int input_count = sizeof(input_records) / sizeof(input_records[0]);
    int expected_count = sizeof(expected_records) / sizeof(expected_records[0]);

    IAS_SSM_POSITION_QUALITY ssm_positions_quality;
    char *ssm_positions_filename = NULL;

    /* Initialize the ssm_position_select field */
    for (index = 0; index < input_count; index++)
    {
        input_records[index].ssm_position_sel = IAS_L0R_TIRS_SSM_NADIR_POSITION;
    }

    /* Initialize the satellite attributes */
    ias_sat_attr_initialize(IAS_L8);

    /* Get the CPF */
    IAS_CPF *cpf = read_test_cpf(IAS_TIRS);
    assert_true(cpf != NULL);
    if (cpf == NULL)
        return;

    earth_params = ias_cpf_get_earth_const(cpf);
    assert_true(earth_params != NULL);
    if (!earth_params)
    {
        ias_cpf_free(cpf);
        return;
    }

    seconds_from_j2000 = input_records[0].l0r_time.days_from_J2000
            * IAS_SEC_PER_DAY + input_records[0].l0r_time.seconds_of_day;
    status = ias_math_init_leap_seconds(seconds_from_j2000, 
        &earth_params->leap_seconds_data);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
    {
        ias_cpf_free(cpf);
        return;
    }
    /* Set up l0r_data structure */
    memset(&l0r_data, 0, sizeof(l0r_data));
    l0r_data.ssm_records = input_records;
    l0r_data.num_ssm_records = input_count;

    /* Set up the sensor_model structure */
    memset(sensor_bands, 0, sizeof(sensor_bands));
    memset(&sensor_model, 0, sizeof(sensor_model));
    memset(&ssm_model, 0, sizeof(ssm_model));

    sensor_bands[0].utc_epoch_time[0] = input_epoch[0];
    sensor_bands[0].utc_epoch_time[1] = input_epoch[1];
    sensor_bands[0].utc_epoch_time[2] = input_epoch[2];
    sensor_bands[0].band_present = 1;
    sensor_bands[0].sensor = &sensor_model.sensors[IAS_TIRS];
    sensor_model.bands = sensor_bands;
    sensor_model.band_count = 1;
    sensor_model.sensors[IAS_TIRS].sensor_id = IAS_TIRS;
    sensor_model.sensors[IAS_TIRS].sensor_present = 1;
    sensor_model.sensors[IAS_TIRS].ssm_model = &ssm_model;
    sensor_model.frame_counts[IAS_OLI] = 0;
    /* Mocking up the frame times to cover the entire image. The set SSM
       code uses nothing more of this frame info than to determine the time
       duration, so mocking up 2 frames that result in the same duration as
       the L0R used to generate the header file test data is sufficient. */
    sensor_model.frame_counts[IAS_TIRS] = 2;
    sensor_model.frame_seconds_from_epoch[IAS_TIRS] = frame_times;

    /* Call the routine */
    printf(".... expect a message about an estimated SSM encoder data file "
        "required, however no estimated SSM encoder file provided\n");
    ssm_positions_quality = IAS_SSM_POSITION_ACTUAL;
    status = ias_sensor_set_ssm_from_l0r(0, &l0r_data, cpf, &sensor_model,
                IAS_TIRS, ssm_positions_filename, &ssm_positions_quality);
    assert_equal(status, SUCCESS);
    assert_equal(ssm_positions_quality, IAS_SSM_POSITION_ACTUAL);

    /* Confirm the number of data items corrected */
    assert_equal(l0r_data.num_bad_ssm_times, 6);
    assert_equal(l0r_data.num_bad_ssm_encoder_values, 4);
    assert_equal(ssm_model.ssm_record_count, expected_count);
    assert_equal(ssm_model.position_status, IAS_SSM_POSITION_DEFAULT);
    for (index = 0; index < ssm_model.ssm_record_count; index++)
    {
        IAS_SENSOR_SSM_RECORD *r = &ssm_model.records[index];
        IAS_SENSOR_SSM_RECORD *er = &expected_records[index];
        double expected_seconds;
        double our_seconds;

        /* Round the seconds to the nearest microsecond since any more
           precise comparison is foolish */
        expected_seconds = round(er->seconds_from_epoch * 1.0e6) * 1.0e-6;
        our_seconds = round(r->seconds_from_epoch * 1.0e6) * 1.0e-6;

        /* Make sure each of the members is as expected */
        if (!doubles_are_equal(our_seconds, expected_seconds)
            || !doubles_are_equal(r->mirror_angle,
                                  er->mirror_angle))
        {
            assert_true_with_message(0, "Index %d results disagree.  "
                "Got %.7e, %.7e Expected %.7e, %.7e", index, 
                r->seconds_from_epoch, r->mirror_angle,
                er->seconds_from_epoch, er->mirror_angle);
        }
    }

    free(sensor_model.sensors[IAS_TIRS].ssm_model->records);
    ias_cpf_free(cpf);
}


typedef struct get_jit_inputs
{
    int band_index;
    int sca_index;
    double line;
    double sample;
} GET_JIT_INPUTS;

Ensure test_get_jitter()
{
    IAS_LOS_MODEL *model;
    int index;

    #include "get_jitter_test_data.h"
    int count = sizeof(inputs)/sizeof(GET_JIT_INPUTS);

    ias_sat_attr_initialize(IAS_L8);

    /* Get the test model */
    model = read_test_model(IAS_OLI);
    assert_true(model != NULL);
    if (!model)
        return;

    for (index = 0; index < count; index++)
    {
        GET_JIT_INPUTS *input = &inputs[index];
        int status;
        double roll, pitch, yaw;

        status = ias_sensor_get_jitter(input->line, input->sample,
                input->band_index, input->sca_index, &model->sensor,
                &roll, &pitch, &yaw);

        assert_equal_with_message(status, output_status[index],
                "failed status - expected %d, got %d for index %d",
                output_status[index], status, index);

        assert_double_equal_with_message(output_roll[index], roll,
                "output error - expected %.8e got %.8e for roll index %d",
                output_roll[index], roll, index);
        assert_double_equal_with_message(output_pitch[index], pitch,
                "output error - expected %.8e got %.8e for pitch index %d",
                output_pitch[index], pitch, index);
        assert_double_equal_with_message(output_yaw[index], yaw,
                "output error - expected %.8e got %.8e for yaw index %d",
                output_yaw[index], yaw, index);
    }

    ias_los_model_free(model);
}

Ensure test_convert_ssm_position()
{
    const char *position_string_status[6] = {"DEFAULT", "ESTIMATED", 
        "UNKNOWN", "NOMINAL", "SWITCHED", "UNKNOWN"};
    const int position_status[6] = {IAS_SSM_POSITION_DEFAULT,
        IAS_SSM_POSITION_ESTIMATED, IAS_SSM_POSITION_UNKNOWN,
        IAS_SSM_POSITION_NOMINAL, IAS_SSM_POSITION_SWITCHED,
        IAS_SSM_POSITION_UNKNOWN};
    const int return_status[6] = {SUCCESS, SUCCESS, ERROR, SUCCESS, SUCCESS,
        ERROR};
    int index;

    for (index = 0; index < 6; index++)
    {
        IAS_SENSOR_SSM_POSITION_STATUS return_position_status;
        const char *return_string;
        int status;
        int log_level;

        /* If the expected result is an error, turn off log messages */
        if (return_status[index] == ERROR)
            log_level = ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

        /* Convert each status to string */
        status = ias_sensor_convert_ssm_position_status_to_string(
            position_status[index], &return_string);
        assert_equal(status, return_status[index]);
        assert_string_equal(return_string, position_string_status[index]);

        /* Convert each string to a enumeration status */
        status = ias_sensor_convert_string_to_ssm_position_status(
            position_string_status[index], &return_position_status);
        assert_equal(status, return_status[index]);
        assert_equal(return_position_status, position_status[index]);

        /* Restore the logging level */
        if (return_status[index] == ERROR)
            ias_log_set_output_level(log_level);
    }
}

Ensure test_get_ssm_encoder_position()
{
    #include "set_est_encoder_data.h"

    int index; /* Generic Loop index */
    IAS_SENSOR_ESTIMATED_SSM_ENCODER *estimated_ptr; /* Encoder data pointer */
    IAS_L0R_TIRS_SSM_ENCODER *record_ptr;   /* Current record pointer */
    IAS_SENSOR_SSM_RECORD *ssm_ptr; /* Pointer to SSM data to estimate */
    IAS_L0R_TIRS_SSM_ENCODER record[5]; /* Holds all estimated positions */
    IAS_SSM_POSITION_QUALITY ssm_positions_quality = IAS_SSM_POSITION_ACTUAL;

    /* All of the verification data. The before and after tests will all contain
       the same value so no array is made to verify them, it will just check the
       same value every time */
    uint32_t verify_normal[20] = {
        16572653, 16572653, 16572654, 16572655, 16572655,
        16572656, 16572656, 16572657, 16572658, 16572658,
        16572659, 16573837, 16573837, 16573837, 16573837,
        16573837, 16573837, 16573837, 16573837, 16573837
    };
    uint32_t verify_before = 16572652;
    uint32_t verify_before_overlap[20] = {
        16572652, 16572652, 16572652, 16572652, 16572652, 
        16572652, 16572652, 16572652, 16572652, 16573218, 
        16573373, 16573450, 16573520, 16573585, 16573645, 
        16573699, 16573749, 16573795, 16573837, 16573876
    };

    uint32_t verify_after = 16574366;
    uint32_t verify_after_overlap[20] = {
        16574349, 16574350, 16574352, 16574353, 16574355, 
        16574356, 16574357, 16574359, 16574360, 16574362,
        16574366, 16574366, 16574366, 16574366, 16574366, 
        16574366, 16574366, 16574366, 16574366, 16574366
    };

    /* Set the estimated encoder data pointer */
    estimated_ptr = &est_ssm[0];

    /* Set up the normal test. All records should be interpolated */
    record_ptr = &record[0];
    ssm_ptr = &normal_records[0];
    if (ias_sensor_get_estimated_ssm(number_of_estimated_encoders,
        0, 0.0, estimated_ptr, ssm_ptr, &ssm_positions_quality, record_ptr) 
        != SUCCESS)
    {
        assert_true_with_message(0, "Running the normal case for "
            "ias_sensor_get_estimated_ssm()");
        return;
    }
    assert_equal(ssm_positions_quality, IAS_SSM_POSITION_FINAL);

    /* Setup the before test. All records should use the encoder position
       from the first record in the estimated encoder data */
    record_ptr = &record[1];
    ssm_ptr = &before_records[0];
    if (ias_sensor_get_estimated_ssm(number_of_estimated_encoders,
        0, 0.0, estimated_ptr, ssm_ptr, &ssm_positions_quality, record_ptr) 
        != SUCCESS)
    {
        assert_true_with_message(0, "Running the before case for "
            "ias_sensor_get_estimated_ssm()");
        return;
    }
    assert_equal(ssm_positions_quality, IAS_SSM_POSITION_FINAL);

    /* Setup the before overlap test. This will contain half records that use
       the first record from the estimated encoder data and then the next half
       will be interpolated */
    record_ptr = &record[2];
    ssm_ptr = &overlap_before_records[0];
    if (ias_sensor_get_estimated_ssm(number_of_estimated_encoders,
        0, 1000.0, estimated_ptr, ssm_ptr, &ssm_positions_quality, record_ptr) 
        != SUCCESS)
    {
        assert_true_with_message(0, "Running the before overlap case for "
            "ias_sensor_get_estimated_ssm()");
        return;
    }
    assert_equal(ssm_positions_quality, IAS_SSM_POSITION_FINAL);

    /* Setup the after test. All records are all after the last records time so
       they will use the last records encoder position */
    record_ptr = &record[3];
    ssm_ptr = &after_records[0];
    ssm_positions_quality = IAS_SSM_POSITION_PRELIMINARY;
    if (ias_sensor_get_estimated_ssm(number_of_estimated_encoders,
        0, 0.0, estimated_ptr, ssm_ptr, &ssm_positions_quality, record_ptr) 
        != SUCCESS)
    {
        assert_true_with_message(0, "Running the after case for "
            "ias_sensor_get_estimated_ssm()");
        return;
    }
    assert_equal(ssm_positions_quality, IAS_SSM_POSITION_PRELIMINARY);

    /* Setup the after overlap test. Half will be interpolated and the other
       half pass the last records time so they will use the last records 
       encoder position */
    record_ptr = &record[4];
    ssm_ptr = &overlap_after_records[0];
    ssm_positions_quality = IAS_SSM_POSITION_PREDICTED;
    if (ias_sensor_get_estimated_ssm(number_of_estimated_encoders,
        0, 0.0, estimated_ptr, ssm_ptr, &ssm_positions_quality, record_ptr) 
        != SUCCESS)
    {
        assert_true_with_message(0, "Running the after case for "
            "ias_sensor_get_estimated_ssm()");
        return;
    }
    assert_equal(ssm_positions_quality, IAS_SSM_POSITION_PREDICTED);

    /* Verify all of the encoder positions */
    for (index = 0; index < 20; index++)
    {
        assert_equal(record[0].encoder_position[index],
            verify_normal[index]);
        assert_equal(record[1].encoder_position[index],
            verify_before);
        assert_equal(record[2].encoder_position[index],
            verify_before_overlap[index]);
        assert_equal(record[3].encoder_position[index],
            verify_after);
        assert_equal(record[4].encoder_position[index],
            verify_after_overlap[index]);               
    }
}

Ensure test_get_ssm_quality()
{
    double start_time = 474084265.400012;
    double stop_time = 474084303.955721;
    IAS_SENSOR_ESTIMATED_SSM_ENCODER *est_ssm_data = NULL;
    int log_level;
    int number_est_encoder_positions = 0;
    int status;

    /* Set the log level to warnings only to avoid the debugging output
       to clutter up the test output. */
    ias_log_set_output_level(IAS_LOG_LEVEL_WARN);

    log_level = ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    
    /* Format version = 2 should error */
    status = ias_sensor_load_estimated_ssm(start_time, stop_time, 
        "est_ssm_encoder_bad_format_version.data", &est_ssm_data, 
        &number_est_encoder_positions);
    assert_equal(status, ERROR);

    /* Missing format version should error */
    status = ias_sensor_load_estimated_ssm(start_time, stop_time, 
        "est_ssm_encoder_missing_format_version.data", &est_ssm_data, 
        &number_est_encoder_positions);
    assert_equal(status, ERROR);

    /* Missing num records should error */
    status = ias_sensor_load_estimated_ssm(start_time, stop_time, 
        "est_ssm_encoder_missing_num_records.data", &est_ssm_data, 
        &number_est_encoder_positions);
    assert_equal(status, ERROR);

    /* Missing YEAR line should error */
    status = ias_sensor_load_estimated_ssm(start_time, stop_time, 
        "est_ssm_encoder_missing_year_line.data", &est_ssm_data, 
        &number_est_encoder_positions);
    assert_equal(status, ERROR);
}

Ensure test_load_ssm_file()
{
    double start_time;
    double stop_time;
    FILE *ssm_fptr;
    IAS_SENSOR_ESTIMATED_SSM_ENCODER *est_ssm_data = NULL;
    int number_est_encoder_positions = 0;
    int status;
    char buffer[256];
    double time[3];
    uint32_t position;
    int quality;
    /* CPF variables for initializing leap seconds */
    IAS_CPF *cpf;
    const IAS_CPF_EARTH_CONSTANTS *earth = 0;

    /* Initialize the satellite attributes */
    ias_sat_attr_initialize(IAS_L8);
    cpf = read_test_cpf(IAS_TIRS);

    /* Open file with estimated SSM encoder values */
    ssm_fptr = fopen("est_ssm_encoder.data", "r");
    /* Read the header lines */
    while(fgets(buffer, HEADER_LINE_LENGTH, ssm_fptr) != NULL)
    {
        char parameter[HEADER_LINE_LENGTH];
        char value[HEADER_LINE_LENGTH];
        char delimiter[HEADER_LINE_LENGTH];

        /* When we are the line right before the data lines break out of
           loop */
        status = sscanf(buffer, "%s %s %s", parameter, delimiter, value);
        if (strncmp(parameter, "#", 1) == 0 &&
            strncmp(delimiter, "YEAR", 4) == 0)
        {
            break;
        }
    }
    /* Pointing at the first data line, store value */
    status = fscanf(ssm_fptr, "%lf %lf %lf %u %d ",
        &time[0], &time[1] ,&time[2], &position, &quality);

    /* Initialize the leap seconds using the time from the first 
       line read in from the ssm encoder file */
    earth = ias_cpf_get_earth_const(cpf);
    ias_math_init_leap_seconds_from_UTC_time(time, &earth->leap_seconds_data);

    /* Get start time based on first line read in*/
    status = ias_math_convert_year_doy_sod_to_j2000_seconds(time,
        &start_time);
    assert_equal(status, SUCCESS);

    /* Now continue reading all lines in file */
    while(!feof(ssm_fptr))
    {
        status = fscanf(ssm_fptr, "%lf %lf %lf %u %d ",
            &time[0], &time[1] ,&time[2], &position, &quality);
        /* Get stop time based on last line read in*/
        status = ias_math_convert_year_doy_sod_to_j2000_seconds(time,
            &stop_time);
        assert_equal(status, SUCCESS);
    }
    fclose(ssm_fptr);

    /* Now call load estimated ssm to read the same file */
    status = ias_sensor_load_estimated_ssm(start_time, stop_time, 
        "est_ssm_encoder.data", &est_ssm_data, &number_est_encoder_positions);

    /* Verify number of positions read */
    assert_equal(number_est_encoder_positions, 243);
    /* Verify some records read, obviously if the file changes, it will impact
       these checks */
    /* Verify first record */
    assert_equal(est_ssm_data[0].ssm_time[0], 2010);
    assert_equal(est_ssm_data[0].ssm_time[1], 100);
    assert_equal(est_ssm_data[0].ssm_time[2], 30060);
    assert_equal(est_ssm_data[0].encoder_position, 16575739);
    assert_equal(est_ssm_data[0].ssm_position_quality, 1);

    /* Verify a record in the middle */
    assert_equal(est_ssm_data[105].ssm_time[0], 2010);
    assert_equal(est_ssm_data[105].ssm_time[1], 131);
    assert_equal(est_ssm_data[105].ssm_time[2], 30000);
    assert_equal(est_ssm_data[105].encoder_position, 16576513);
    assert_equal(est_ssm_data[105].ssm_position_quality, 0);

    /* Verify last record */
    int last_record = number_est_encoder_positions - 1;
    assert_equal(est_ssm_data[last_record].ssm_time[0], 2010);
    assert_equal(est_ssm_data[last_record].ssm_time[1], 268);
    assert_equal(est_ssm_data[last_record].ssm_time[2], 30000);
    assert_equal(est_ssm_data[last_record].encoder_position, 16576994);
    assert_equal(est_ssm_data[last_record].ssm_position_quality, 2);

    ias_cpf_free(cpf);
}

/* Re-enable the command line specified optimization */
IAS_RESTORE_OPTIMIZATION

int main(int argc, char **argv)
{
    TestSuite *suite = create_test_suite();
    add_test(suite, test_set_frame_times);
    add_test(suite, test_find_los_vector);
    add_test(suite, test_find_time);
    add_test(suite, test_set_cpf);
    add_test(suite, test_set_l0r);
    add_test(suite, test_set_l0r_tirs);
    add_test(suite, test_tirs_find_time);
    add_test(suite, test_tirs_find_los_vector);
    add_test(suite, test_tirs_adjust_los_for_ssm);
    add_test(suite, test_ssm_smoothing);
    add_test(suite, test_align_ssm_data);
    add_test(suite, test_check_ssm_encoder_data);
    add_test(suite, test_set_ssm_from_l0r);
    add_test(suite, test_null_ssm_filename);
    add_test(suite, test_get_jitter);
    add_test(suite, test_convert_ssm_position);
    add_test(suite, test_get_ssm_encoder_position);
    add_test(suite, test_get_ssm_quality);
    add_test(suite, test_load_ssm_file);
    ias_log_initialize("Sensor Test");
    if (argc > 1) {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}
