#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include "ias_satellite_attributes.h"
#include "ias_los_model.h"
#include "ias_model_io.h"
#include "ias_logging.h"
#include "cgreen/cgreen.h"

IAS_LOS_MODEL *read_test_model()
{
    IAS_LOS_MODEL *model;
    const char *test_data_dir;
    char model_filename[PATH_MAX];

    /* get the test data directory */
    test_data_dir = getenv("IAS_UNIT_TEST_DATA_DIR");
    assert_true(test_data_dir != NULL);

    /* build the model filename */
    snprintf(model_filename, sizeof(model_filename),
             "%s/ias_lib/los_model/spacecraft/oli_model.h5", test_data_dir);

    model = ias_model_read(model_filename);
    if (!model)
    {
        assert_true_with_message(0, "Error reading input model");
        return NULL;
    }

    return model;
}

Ensure test_find_attitude()
{
    IAS_LOS_MODEL *model;
    /* the following values were obtained by running the cal/val prototype
       code */
    #include "findatt_test_data.h"
    int count = sizeof(inputs)/sizeof(inputs[0]);
    int index;

    ias_sat_attr_initialize(IAS_L8);

    /* get the test model */
    model = read_test_model();
    assert_true(model != NULL);
    if (!model)
        return;

    /* disable all output since some tests are hitting error conditions */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    for (index = 0; index < count; index++)
    {
        double input_time = inputs[index];
        double roll;
        double pitch;
        double yaw;
        int status;

        status = ias_sc_model_find_attitude_at_time(&model->spacecraft.attitude,
                                             input_time, &roll, &pitch, &yaw);
        if (status != output_status[index])
        {
            /* only use asserts if there is a failure due to the cgreen limit
               of 16384 asserts per run */
            assert_equal_with_message(status, output_status[index],
                "failed status - expected %d, got %d for index %d",
                output_status[index], status, index);
        }
        else if (status != ERROR)
        {
            /* use a cgreen function to verify the doubles are equal to 8
               significant digits */
            int valid = doubles_are_equal(output_roll[index], roll);
            if (!valid)
            {
                assert_equal_with_message(valid, 1, 
                    "roll error - expected %.8e, got %.8e for index %d",
                    output_roll[index], roll, index);
            }
            valid = doubles_are_equal(output_pitch[index], pitch);
            if (!valid)
            {
                assert_equal_with_message(valid, 1, 
                    "pitch error - expected %.8e, got %.8e for index %d",
                    output_pitch[index], pitch, index);
            }
            valid = doubles_are_equal(output_yaw[index], yaw);
            if (!valid)
            {
                assert_equal_with_message(valid, 1, 
                    "yaw error - expected %.8e, got %.8e for index %d",
                    output_yaw[index], yaw, index);
            }
        }
    }

    ias_log_set_output_level(IAS_LOG_LEVEL_WARN);

    ias_los_model_free(model);
}

void compare_vector
(
    IAS_VECTOR *expected,
    IAS_VECTOR *got,
    const char *name,
    int index
)
{
    /* use a cgreen function to verify the doubles are equal to 8
       significant digits */
    int valid = doubles_are_equal(expected->x, got->x);
    if (!valid)
    {
        assert_true_with_message(valid, 
            "%s.x error - expected %.8e, got %.8e for index %d",
            name, expected->x, got->x, index);
    }
    valid = doubles_are_equal(expected->y, got->y);
    if (!valid)
    {
        assert_true_with_message(valid, 
            "%s.y error - expected %.8e, got %.8e for index %d",
            name, expected->y, got->y, index);
    }
    valid = doubles_are_equal(expected->z, got->z);
    if (!valid)
    {
        assert_true_with_message(valid, 
            "%s.z error - expected %.8e, got %.8e for index %d",
            name, expected->z, got->z, index);
    }
}

typedef struct movesat_inputs
{
    IAS_ACQUISITION_TYPE acq_type;
    double input_time;
} MOVESAT_INPUTS;

typedef struct movesat_outputs
{
    IAS_VECTOR satpos;
    IAS_VECTOR satvel;
} MOVESAT_OUTPUTS;

Ensure test_movesat()
{
    IAS_LOS_MODEL *model;
    /* the following values were obtained by running the cal/val prototype
       code */
    #include "movesat_test_data.h"
    int count = sizeof(inputs)/sizeof(inputs[0]);
    int index;

    ias_sat_attr_initialize(IAS_L8);

    /* get the test model */
    model = read_test_model();
    assert_true(model != NULL);
    if (!model)
        return;

    for (index = 0; index < count; index++)
    {
        IAS_VECTOR satpos;
        IAS_VECTOR satvel;

        ias_sc_model_get_position_and_velocity_at_time(
                &model->spacecraft.ephemeris,
                inputs[index].acq_type, inputs[index].input_time, &satpos,
                &satvel);

        compare_vector(&outputs[index].satpos, &satpos, "satpos", index);
        compare_vector(&outputs[index].satvel, &satvel, "satvel", index);
    }

    ias_los_model_free(model);
}

Ensure test_set_ancillary()
{
    IAS_LOS_MODEL *model;
    IAS_ANC_EPHEMERIS_DATA *ephemeris_data;
    IAS_ANC_ATTITUDE_DATA *attitude_data;
    IAS_SPACECRAFT_MODEL *sc;
    char anc_filename[PATH_MAX];
    const char *test_data_dir;
    int status;
    int i;

    /* Set the logging to warning so the library debugging messages don't
       clutter the test output. */
    ias_log_set_output_level(IAS_LOG_LEVEL_WARN);

    ias_sat_attr_initialize(IAS_L8);

    test_data_dir = getenv("IAS_UNIT_TEST_DATA_DIR");
    assert_true(test_data_dir != NULL);

    snprintf(anc_filename, sizeof(anc_filename),
            "%s/ias_lib/los_model/spacecraft/PREPROCESSING_ANC.h5",
            test_data_dir);

    /* allocate a model */
    model = ias_los_model_allocate();
    assert_not_equal(model, NULL);

    sc = &model->spacecraft;

    /* Read ancillary file */
    status = ias_ancillary_read(anc_filename, &attitude_data, &ephemeris_data);
    assert_equal(status, SUCCESS);

    /* Set the attitude and ephemeris values from the ancillary data file
       into the model structure */
    status = ias_sc_model_set_ancillary(ephemeris_data, attitude_data, sc);
    assert_equal(status, SUCCESS);

    /* check the values against known static values */
    assert_equal(2000, (int)sc->ephemeris.utc_epoch_time[0]);
    assert_equal(227, (int)sc->ephemeris.utc_epoch_time[1]);
    assert_double_equal(65256.949607, sc->ephemeris.utc_epoch_time[2]);
    assert_double_equal(1.0, sc->ephemeris.nominal_sample_time);
    assert_equal(400, sc->ephemeris.sample_count);
    assert_equal(2000, (int)sc->attitude.utc_epoch_time[0]);
    assert_equal(227, (int)sc->attitude.utc_epoch_time[1]);
    assert_double_equal(65292.621110, sc->attitude.utc_epoch_time[2]);
    assert_double_equal(0.02, sc->attitude.nominal_sample_time);
    assert_equal(16384, sc->attitude.sample_count);
    for (i = 0; i < 3; i++)
    {
        assert_double_equal((double)i,
                sc->ephemeris.sample_records[i].seconds_from_epoch);
        assert_double_equal(i * 0.02,
                sc->attitude.sample_records[i].seconds_from_epoch);
    }

    assert_double_equal(-2410302.326840,
            sc->ephemeris.sample_records[0].eci_position.x);
    assert_double_equal(3325854.318184,
            sc->ephemeris.sample_records[0].eci_position.y);
    assert_double_equal(5760372.853320,
            sc->ephemeris.sample_records[0].eci_position.z);
    assert_double_equal(-2412292.803273,
            sc->ephemeris.sample_records[1].eci_position.x);
    assert_double_equal(3331738.715247,
            sc->ephemeris.sample_records[1].eci_position.y);
    assert_double_equal(5756149.317551,
            sc->ephemeris.sample_records[1].eci_position.z);
    assert_double_equal(-2414278.639517,
            sc->ephemeris.sample_records[2].eci_position.x);
    assert_double_equal(3337616.316788,
            sc->ephemeris.sample_records[2].eci_position.y);
    assert_double_equal(5751916.971069,
            sc->ephemeris.sample_records[2].eci_position.z);

    assert_double_equal(-1992.679853,
            sc->ephemeris.sample_records[0].eci_velocity.x);
    assert_double_equal(5884.362911,
            sc->ephemeris.sample_records[0].eci_velocity.y);
    assert_double_equal(-4212.065879,
            sc->ephemeris.sample_records[0].eci_velocity.z);
    assert_double_equal(-1985.299958,
            sc->ephemeris.sample_records[1].eci_velocity.x);
    assert_double_equal(5875.292484,
            sc->ephemeris.sample_records[1].eci_velocity.y);
    assert_double_equal(-4228.206425,
            sc->ephemeris.sample_records[1].eci_velocity.z);
    assert_double_equal(-1984.141805,
            sc->ephemeris.sample_records[2].eci_velocity.x);
    assert_double_equal(5873.322095,
            sc->ephemeris.sample_records[2].eci_velocity.y);
    assert_double_equal(-4231.471749,
            sc->ephemeris.sample_records[2].eci_velocity.z);

    assert_double_equal(-1366592.862715,
            sc->ephemeris.sample_records[0].ecef_position.x);
    assert_double_equal(-3873610.687438,
            sc->ephemeris.sample_records[0].ecef_position.y);
    assert_double_equal(5760236.440601,
            sc->ephemeris.sample_records[0].ecef_position.z);
    assert_double_equal(-1370586.086062,
            sc->ephemeris.sample_records[1].ecef_position.x);
    assert_double_equal(-3878492.688391,
            sc->ephemeris.sample_records[1].ecef_position.y);
    assert_double_equal(5756012.738711,
            sc->ephemeris.sample_records[1].ecef_position.z);
    assert_double_equal(-1374577.073476,
            sc->ephemeris.sample_records[2].ecef_position.x);
    assert_double_equal(-3883366.447555,
            sc->ephemeris.sample_records[2].ecef_position.y);
    assert_double_equal(5751780.226376,
            sc->ephemeris.sample_records[2].ecef_position.z);

    assert_double_equal(-3709.115576,
            sc->ephemeris.sample_records[0].ecef_velocity.x);
    assert_double_equal(-4983.729487,
            sc->ephemeris.sample_records[0].ecef_velocity.y);
    assert_double_equal(-4212.232291,
            sc->ephemeris.sample_records[0].ecef_velocity.z);
    assert_double_equal(-3706.209444,
            sc->ephemeris.sample_records[1].ecef_velocity.x);
    assert_double_equal(-4972.231425,
            sc->ephemeris.sample_records[1].ecef_velocity.y);
    assert_double_equal(-4228.372438,
            sc->ephemeris.sample_records[1].ecef_velocity.z);
    assert_double_equal(-3705.608730,
            sc->ephemeris.sample_records[2].ecef_velocity.x);
    assert_double_equal(-4969.888399,
            sc->ephemeris.sample_records[2].ecef_velocity.y);
    assert_double_equal(-4231.637690,
            sc->ephemeris.sample_records[2].ecef_velocity.z);

    assert_double_equal(-2410302.326840,
            sc->ephemeris.sample_records[0].precision_eci_position.x);
    assert_double_equal(3325854.318184,
            sc->ephemeris.sample_records[0].precision_eci_position.y);
    assert_double_equal(5760372.853320,
            sc->ephemeris.sample_records[0].precision_eci_position.z);
    assert_double_equal(-2412292.803273,
            sc->ephemeris.sample_records[1].precision_eci_position.x);
    assert_double_equal(3331738.715247,
            sc->ephemeris.sample_records[1].precision_eci_position.y);
    assert_double_equal(5756149.317551,
            sc->ephemeris.sample_records[1].precision_eci_position.z);
    assert_double_equal(-2414278.639517,
            sc->ephemeris.sample_records[2].precision_eci_position.x);
    assert_double_equal(3337616.316788,
            sc->ephemeris.sample_records[2].precision_eci_position.y);
    assert_double_equal(5751916.971069,
            sc->ephemeris.sample_records[2].precision_eci_position.z);

    assert_double_equal(-1992.679853,
            sc->ephemeris.sample_records[0].precision_eci_velocity.x);
    assert_double_equal(5884.362911,
            sc->ephemeris.sample_records[0].precision_eci_velocity.y);
    assert_double_equal(-4212.065879,
            sc->ephemeris.sample_records[0].precision_eci_velocity.z);
    assert_double_equal(-1985.299958,
            sc->ephemeris.sample_records[1].precision_eci_velocity.x);
    assert_double_equal(5875.292484,
            sc->ephemeris.sample_records[1].precision_eci_velocity.y);
    assert_double_equal(-4228.206425,
            sc->ephemeris.sample_records[1].precision_eci_velocity.z);
    assert_double_equal(-1984.141805,
            sc->ephemeris.sample_records[2].precision_eci_velocity.x);
    assert_double_equal(5873.322095,
            sc->ephemeris.sample_records[2].precision_eci_velocity.y);
    assert_double_equal(-4231.471749,
            sc->ephemeris.sample_records[2].precision_eci_velocity.z);

    assert_double_equal(-1366592.862715,
            sc->ephemeris.sample_records[0].precision_ecef_position.x);
    assert_double_equal(-3873610.687438,
            sc->ephemeris.sample_records[0].precision_ecef_position.y);
    assert_double_equal(5760236.440601,
            sc->ephemeris.sample_records[0].precision_ecef_position.z);
    assert_double_equal(-1370586.086062,
            sc->ephemeris.sample_records[1].precision_ecef_position.x);
    assert_double_equal(-3878492.688391,
            sc->ephemeris.sample_records[1].precision_ecef_position.y);
    assert_double_equal(5756012.738711,
            sc->ephemeris.sample_records[1].precision_ecef_position.z);
    assert_double_equal(-1374577.073476,
            sc->ephemeris.sample_records[2].precision_ecef_position.x);
    assert_double_equal(-3883366.447555,
            sc->ephemeris.sample_records[2].precision_ecef_position.y);
    assert_double_equal(5751780.226376,
            sc->ephemeris.sample_records[2].precision_ecef_position.z);

    assert_double_equal(-3709.115576,
            sc->ephemeris.sample_records[0].precision_ecef_velocity.x);
    assert_double_equal(-4983.729487,
            sc->ephemeris.sample_records[0].precision_ecef_velocity.y);
    assert_double_equal(-4212.232291,
            sc->ephemeris.sample_records[0].precision_ecef_velocity.z);
    assert_double_equal(-3706.209444,
            sc->ephemeris.sample_records[1].precision_ecef_velocity.x);
    assert_double_equal(-4972.231425,
            sc->ephemeris.sample_records[1].precision_ecef_velocity.y);
    assert_double_equal(-4228.372438,
            sc->ephemeris.sample_records[1].precision_ecef_velocity.z);
    assert_double_equal(-3705.608730,
            sc->ephemeris.sample_records[2].precision_ecef_velocity.x);
    assert_double_equal(-4969.888399,
            sc->ephemeris.sample_records[2].precision_ecef_velocity.y);
    assert_double_equal(-4231.637690,
            sc->ephemeris.sample_records[2].precision_ecef_velocity.z);

    assert_double_equal(0.0001100676,
            sc->attitude.sample_records[0].satellite_attitude.x);
    assert_double_equal(0.000005936351,
            sc->attitude.sample_records[0].satellite_attitude.y);
    assert_double_equal(-0.04028832,
            sc->attitude.sample_records[0].satellite_attitude.z);
    assert_double_equal(0.0001087610,
            sc->attitude.sample_records[1].satellite_attitude.x);
    assert_double_equal(0.000005890525,
            sc->attitude.sample_records[1].satellite_attitude.y);
    assert_double_equal(-0.04028952,
            sc->attitude.sample_records[1].satellite_attitude.z);
    assert_double_equal(0.0001071401,
            sc->attitude.sample_records[2].satellite_attitude.x);
    assert_double_equal(0.000005635164,
            sc->attitude.sample_records[2].satellite_attitude.y);
    assert_double_equal(-0.04029083,
            sc->attitude.sample_records[2].satellite_attitude.z);

    assert_double_equal(0.0001100676,
            sc->attitude.sample_records[0].precision_attitude.x);
    assert_double_equal(0.000005936351,
            sc->attitude.sample_records[0].precision_attitude.y);
    assert_double_equal(-0.04028832,
            sc->attitude.sample_records[0].precision_attitude.z);
    assert_double_equal(0.0001087610,
            sc->attitude.sample_records[1].precision_attitude.x);
    assert_double_equal(0.000005890525,
            sc->attitude.sample_records[1].precision_attitude.y);
    assert_double_equal(-0.04028952,
            sc->attitude.sample_records[1].precision_attitude.z);
    assert_double_equal(0.0001071401,
            sc->attitude.sample_records[2].precision_attitude.x);
    assert_double_equal(0.000005635164,
            sc->attitude.sample_records[2].precision_attitude.y);
    assert_double_equal(-0.04029083,
            sc->attitude.sample_records[2].precision_attitude.z);

    free(ephemeris_data);
    free(attitude_data);
    ias_los_model_free(model);
}

int main(int argc, char **argv)
{
    TestSuite *suite = create_test_suite();
    add_test(suite, test_find_attitude);
    add_test(suite, test_movesat);
    add_test(suite, test_set_ancillary);
    if (argc > 1) {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}
