#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include "ias_satellite_attributes.h"
#include "ias_los_model.h"
#include "ias_model_io.h"
#include "ias_geo.h"
#include "ias_cpf.h"
#include "ias_logging.h"
#include "ias_math.h"
#include "cgreen/cgreen.h"

/* Model file types. There is a complete model file used by most of the tests.
   There is also a model file that was built skipping the build jitter table
   step, with no jitter table in it and the attitude data has not been modified
   from original yet, which is used to regression test the build jitter table
   routine. These macros make it easy for a test to get the right model file. */
#define READ_COMPLETE_MODEL                         0
#define READ_PARTIAL_MODEL_NO_JITTER_TABLE_BUILT    1

/* Provide a routine to initialize the leap seconds in the math library since
   they need to be initialized for some geo library calls */
static void initialize_leap_seconds
(
    int include_leap_second /* I: flag to initialize leap seconds to 1 */
)
{
    int status;

    /* Set up a leap second structure with just a single entry to make it
       easy */
    IAS_MATH_LEAP_SECONDS_DATA leap_data;
    int leap_years = 2006;
    int leap_months = 1;
    int leap_days = 1;
    int leap_seconds = 1;
    leap_data.leap_years = &leap_years;
    leap_data.leap_months = &leap_months;
    leap_data.leap_days = &leap_days;
    leap_data.num_leap_seconds = &leap_seconds;
    leap_data.leap_seconds_count = 1;

    /* If a leap second is needed, reset the leap year to be 2001 so the 
       ias_math_init_leap_seconds call sets the leap seconds to 1 (since 
       that will be before the seconds being passed into the init routine) */
    if (include_leap_second)
        leap_years = 2001;

    /* Initialize the leap seconds for a few days into 2001. */
    status = ias_math_init_leap_seconds(IAS_SEC_PER_DAY * 380.0, &leap_data);
    assert_equal(status, SUCCESS);
}

/* Test the model allocation and freeing work as expected */
Ensure test_los_model_allocation()
{
    IAS_LOS_MODEL *model;
    IAS_SENSOR_MODEL *sensor;
    int band_index;

    /* make sure the allocation fails if the satellite attributes are not
       initialized */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    model = ias_los_model_allocate();
    ias_log_set_output_level(IAS_LOG_LEVEL_WARN);
    assert_equal(model, NULL);

    ias_sat_attr_initialize(IAS_L8);

    model = ias_los_model_allocate();
    assert_not_equal(model, NULL);

    /* verify some things are set as expected in the model */
    assert_equal(model->satellite_id, IAS_L8);
    assert_equal(model->satellite_number, 8);
    assert_equal(model->spacecraft.attitude.sample_count, 0);
    assert_equal(model->spacecraft.attitude.sample_records, NULL);
    assert_equal(model->spacecraft.ephemeris.sample_count, 0);
    assert_equal(model->spacecraft.ephemeris.sample_records, NULL);
    assert_equal(model->sensor.band_count,
                 ias_sat_attr_get_normal_band_count());
    assert_not_equal(model->sensor.bands, NULL);

    /* check each of the bands */
    sensor = &model->sensor;
    for (band_index = 0; band_index < sensor->band_count; band_index++)
    {
        int sca_index;
        IAS_SENSOR_BAND_MODEL *band = &sensor->bands[band_index];
        const IAS_BAND_ATTRIBUTES *band_info 
                = ias_sat_attr_get_band_attributes(band_index + 1);

        assert_not_equal(band, NULL);
        assert_equal(band->sca_count, band_info->scas);
        assert_not_equal(band->scas, NULL);
        assert_equal(band->frame_seconds_from_epoch, NULL);

        /* check each of the SCAs */
        for (sca_index = 0; sca_index < band->sca_count; sca_index++)
        {
            IAS_SENSOR_SCA_MODEL *sca = &band->scas[sca_index];

            assert_not_equal(sca, NULL);
            assert_not_equal(sca->l0r_detector_offsets, NULL);
            assert_not_equal(sca->detector_offsets_along_track, NULL);
            assert_not_equal(sca->detector_offsets_across_track, NULL);
            assert_equal(sca->detectors, band_info->detectors_per_sca);
        }
    }

    /* free the model.  Nothing really to test here, but running with valgrind
       will confirm no memory leaks. */
    ias_los_model_free(model);
}

IAS_LOS_MODEL *read_test_model
(
    IAS_SENSOR_ID sensor_id,    /* I: The model type to read */
    int model_type_flag         /* I: Indicates what type of model file to
                                      read. See #def's at top of file. */
)
{
    IAS_LOS_MODEL *model = NULL;
    const char *test_data_dir;
    char model_filename[PATH_MAX];

    /* Get the test data directory */
    test_data_dir = getenv("IAS_UNIT_TEST_DATA_DIR");
    assert_true(test_data_dir != NULL);

    /* Make sure an expected model type was indicated */
    if ((model_type_flag != READ_COMPLETE_MODEL)
            && (model_type_flag != READ_PARTIAL_MODEL_NO_JITTER_TABLE_BUILT))
    {
        assert_true_with_message(0, "Error unrecognized model type requested");
        return NULL;
    }

    /* Build the model filename */
    if (sensor_id == IAS_OLI)
    {
        if (model_type_flag == READ_COMPLETE_MODEL)
        {
            snprintf(model_filename, sizeof(model_filename),
                    "%s/ias_lib/los_model/oli_model.h5", test_data_dir);
        }
        else
        {
            snprintf(model_filename, sizeof(model_filename),
                    "%s/ias_lib/los_model/pre_jit_oli_model.h5",
                    test_data_dir);
        }
    }
    else if (sensor_id == IAS_TIRS)
    {
        if (model_type_flag == READ_COMPLETE_MODEL)
        {
            snprintf(model_filename, sizeof(model_filename),
                    "%s/ias_lib/los_model/tirs_model.h5", test_data_dir);
        }
        else
        {
            snprintf(model_filename, sizeof(model_filename),
                    "%s/ias_lib/los_model/pre_jit_tirs_model.h5",
                    test_data_dir);
        }
    }
    else
    {
        assert_true_with_message(0, "Error bad sensor type for reading model");
        return NULL;
    }

    model = ias_model_read(model_filename);
    if (!model)
    {
        assert_true_with_message(0, "Error reading input model %s",
                model_filename);
        return NULL;
    }

    return model;
}

IAS_LOS_MODEL *read_lunar_test_model()
{
    IAS_LOS_MODEL *model = NULL;
    const char *test_data_dir;
    char model_filename[PATH_MAX];

    /* get the test data directory */
    test_data_dir = getenv("IAS_UNIT_TEST_DATA_DIR");
    assert_true(test_data_dir != NULL);

    /* build the model filename */
    snprintf(model_filename, sizeof(model_filename),
             "%s/ias_lib/los_model/lunar_model.h5", test_data_dir);

    model = ias_model_read(model_filename);
    if (!model)
    {
        assert_true_with_message(0, "Error reading input model");
        return NULL;
    }

    return model;
}

typedef struct model_inputs
{
    int band;
    int sca;
    double line;
    double sample;
    IAS_SENSOR_DETECTOR_TYPE dettype;
} MODEL_INPUTS;

Ensure test_forward_model()
{
    double tarlatd;
    double tarlong;
    double tarelev = 1000.0;

    IAS_LOS_MODEL *model;
    /* the following values were obtained by running the cal/val prototype
       code */
    #include "forward_model_test_data.h"
    int count = sizeof(inputs)/sizeof(MODEL_INPUTS);
    int index;

    ias_sat_attr_initialize(IAS_L8);

    /* get the test model */
    model = read_test_model(IAS_OLI, READ_COMPLETE_MODEL);
    assert_true(model != NULL);
    if (!model)
        return;

    /* disable all output since the tests are hitting error conditions */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    for (index = 0; index < count; index++)
    {
        MODEL_INPUTS *input = &inputs[index];
        int status;

        status = ias_los_model_input_line_samp_to_geodetic(input->line,
                input->sample, input->band, input->sca, tarelev, model,
                input->dettype, NULL, &tarlatd, &tarlong);

        if (status != output_status[index])
        {
            /* only use asserts if there is a failure due to the cgreen limit
               of 16384 asserts per run */
            assert_equal_with_message(status, output_status[index],
                    "failed status - expected %d, got %d for index %d",
                    output_status[index], status, index);
        }
        else if (status != ERROR)
        {
            /* use a cgreen function to verify the doubles are equal to 8
               significant digits */
            int valid = doubles_are_equal(output_tarlatd[index], tarlatd)
                && doubles_are_equal(output_tarlong[index], tarlong);
            if (!valid)
            {
                assert_equal_with_message(valid, 1, 
                        "output error - expected (%.8e, %.8e) "
                        "got (%.8e, %.8e) for index %d", output_tarlatd[index],
                        output_tarlong[index], tarlatd, tarlong, index);
            }
        }
    }

    ias_log_set_output_level(IAS_LOG_LEVEL_WARN);

    ias_los_model_free(model);
}

Ensure test_tirs_forward_model()
{
    double tarlatd;
    double tarlong;
    double tarelev = 1000.0;

    IAS_LOS_MODEL *model;
    /* the following values were obtained by running the cal/val prototype
       code */
    #include "tirs_forward_model_test_data.h"
    int count = sizeof(inputs)/sizeof(MODEL_INPUTS);
    int index;

    ias_sat_attr_initialize(IAS_L8);

    /* get the test model */
    model = read_test_model(IAS_TIRS, READ_COMPLETE_MODEL);
    assert_true(model != NULL);
    if (!model)
        return;

    /* disable all output since the tests are hitting error conditions */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    for (index = 0; index < count; index++)
    {
        MODEL_INPUTS *input = &inputs[index];
        int status;

        status = ias_los_model_input_line_samp_to_geodetic(input->line,
                input->sample, input->band, input->sca, tarelev, model,
                input->dettype, NULL, &tarlatd, &tarlong);

        if (status != output_status[index])
        {
            /* only use asserts if there is a failure due to the cgreen limit
               of 16384 asserts per run */
            assert_equal_with_message(status, output_status[index],
                    "failed status - expected %d, got %d for index %d",
                    output_status[index], status, index);
        }
        else if (status != ERROR)
        {
            /* use a cgreen function to verify the doubles are equal to 8
               significant digits */
            int valid = doubles_are_equal(output_tarlatd[index], tarlatd)
                && doubles_are_equal(output_tarlong[index], tarlong);
            if (!valid)
            {
                assert_equal_with_message(valid, 1, 
                        "output error - expected (%.8e, %.8e) "
                        "got (%.8e, %.8e) for index %d", output_tarlatd[index],
                        output_tarlong[index], tarlatd, tarlong, index);
            }
        }
    }

    ias_log_set_output_level(IAS_LOG_LEVEL_WARN);

    ias_los_model_free(model);
}

Ensure test_oli_calc_position()
{
    IAS_LOS_MODEL *model;
    /* the following values were obtained by running the cal/val prototype
       code.  Within it are:
       MODEL_INPUTS inputs[] = { ... };
       IAS_VECTOR satpos[] = { ... };
       IAS_VECTOR satvel[] = { ... };
       IAS_VECTOR satatt[] = { ... };
       double time[] = { ... };
       int year[] = { ... };
       int day[] = { ... };
       int output_status[] = { ... };
     */
    #include "oli_calc_position_test_data.h"
    int count = sizeof(inputs)/sizeof(MODEL_INPUTS);
    IAS_VECTOR position;
    IAS_VECTOR velocity;
    IAS_VECTOR attitude;
    double image_time;
    int image_year;
    int image_day;
    int index;

    ias_sat_attr_initialize(IAS_L8);

    /* get the test model */
    model = read_test_model(IAS_OLI, READ_COMPLETE_MODEL);
    assert_true(model != NULL);
    if (!model)
        return;

    /* disable all output since the tests are hitting error conditions */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    for (index = 0; index < count; index++)
    {
        MODEL_INPUTS *input = &inputs[index];
        int status;

        status = ias_los_model_get_satellite_state_vector_at_location(
                input->line, input->sample, input->band, input->sca,
                input->dettype, model, &position, &velocity, &attitude,
                &image_time, &image_year, &image_day );

        if (status != output_status[index])
        {
            /* only use asserts if there is a failure due to the cgreen limit
               of 16384 asserts per run */
            assert_equal_with_message(status, output_status[index],
                    "failed status - expected %d, got %d for index %d",
                    output_status[index], status, index);
        }
        else if (status != ERROR)
        {
            /* Check the satellite position.  Use a cgreen function to
               verify the doubles are equal to 8 significant digits. */
            int valid = doubles_are_equal(satpos[index].x, position.x)
                && doubles_are_equal(satpos[index].y, position.y)
                && doubles_are_equal(satpos[index].z, position.z);
            if (!valid)
            {
                assert_equal_with_message(valid, 1, "output error - \n"
                        "expected (%.8e, %.8e, %.8e)\n"
                        "     got (%.8e, %.8e, %.8e)\nfor index %d",
                        satpos[index].x, satpos[index].y, satpos[index].z,
                        position.x, position.y, position.z, index );
            }

            /* Check the satellite velocity.  Use a cgreen function to
               verify the doubles are equal to 8 significant digits. */
            valid = doubles_are_equal(satvel[index].x, velocity.x)
                && doubles_are_equal(satvel[index].y, velocity.y)
                && doubles_are_equal(satvel[index].z, velocity.z);
            if (!valid)
            {
                assert_equal_with_message(valid, 1, "output error - \n"
                        "expected (%.8e, %.8e, %.8e)\n"
                        "     got (%.8e, %.8e, %.8e)\nfor index %d",
                        satvel[index].x, satvel[index].y, satvel[index].z,
                        velocity.x, velocity.y, velocity.z, index );
            }

            /* Check the satellite attitude.  Use a cgreen function to
               verify the doubles are equal to 8 significant digits. */
            valid = doubles_are_equal(satatt[index].x, attitude.x)
                && doubles_are_equal(satatt[index].y, attitude.y)
                && doubles_are_equal(satatt[index].z, attitude.z);
            if (!valid)
            {
                assert_equal_with_message(valid, 1, "output error - \n"
                        "expected (%.8e, %.8e, %.8e)\n"
                        "     got (%.8e, %.8e, %.8e)\nfor index %d",
                        satatt[index].x, satatt[index].y, satatt[index].z,
                        attitude.x, attitude.y, attitude.z, index );
            }

            /* Check the time.  Use a cgreen function to
               verify the doubles are equal to 8 significant digits. */
            valid = doubles_are_equal(time[index], image_time);
            if (!valid)
            {
                assert_equal_with_message(valid, 1, "output error - \n"
                        "expected (%.8e)\n"
                        "     got (%.8e)\nfor index %d", time[index],
                        image_time, index );
            }

            /* Check the day and year */
            if ( day[index] != image_day || year[index] != image_year )
            {
                assert_true_with_message( 0, "output error - \n"
                        "expected (%d, %d)\n"
                        "     got (%d, %d)\nfor index %d", day[index],
                        year[index], image_day, image_year, index );
            }
        }
    }

    ias_log_set_output_level(IAS_LOG_LEVEL_WARN);

    ias_los_model_free(model);
}

Ensure test_set_cpf()
{
    IAS_LOS_MODEL *model;
    IAS_CPF *cpf;
    IAS_SENSOR_L0R_DATA l0r_sensor_data;
    const char *test_data_dir;
    char cpf_filename[PATH_MAX];
    int status;

    test_data_dir = getenv("IAS_UNIT_TEST_DATA_DIR");
    assert_not_equal(test_data_dir, NULL);

    snprintf(cpf_filename, sizeof(cpf_filename),
            "%s/ias_lib/cpf/L8CPF20090101_20090331.01", test_data_dir);

    ias_sat_attr_initialize(IAS_L8);

    model = ias_los_model_allocate();

    cpf = ias_cpf_read(cpf_filename);
    assert_not_equal(cpf, NULL);

    model->sensor.bands[0].utc_epoch_time[0] = 2008;
    model->sensor.bands[0].utc_epoch_time[1] = 227;
    model->sensor.bands[0].utc_epoch_time[2] = 0.0;

    status = ias_los_model_set_cpf(cpf, &l0r_sensor_data, model);
    assert_equal(status, SUCCESS);

    assert_double_equal(6378137.0, model->earth.semi_major_axis);
    assert_double_equal(6356752.3142, model->earth.semi_minor_axis);
    assert_double_equal(0.00669437999, model->earth.eccentricity);
    assert_double_equal(0.000072921158553, model->earth.earth_angular_velocity);
    assert_double_equal(299792458.0, model->earth.speed_of_light);

    ias_cpf_free(cpf);
    ias_los_model_free(model);
}

Ensure test_set_l0r()
{
    IAS_LOS_MODEL *model;
    IAS_LOS_MODEL_L0R_DATA l0r_data;
    IAS_CPF *cpf;
    char cpf_filename[PATH_MAX];
    char ssm_position_filename[PATH_MAX];
    const char *test_data_dir;
    int status;

    IAS_SSM_POSITION_QUALITY ssm_position_quality;

    test_data_dir = getenv("IAS_UNIT_TEST_DATA_DIR");
    assert_not_equal(test_data_dir, NULL);

    snprintf(cpf_filename, sizeof(cpf_filename),
            "%s/ias_lib/cpf/L8CPF20090101_20090331.01", test_data_dir);

    ias_sat_attr_initialize(IAS_L8);

    cpf = ias_cpf_read(cpf_filename);
    assert_not_equal(cpf, NULL);

    snprintf(ssm_position_filename, sizeof(ssm_position_filename),
             "%s/ias_base/gps/create_los_model/ENCODER_OFF/"
             "est_ssm_encoder.data", test_data_dir);

    /* get the test model */
    model = read_test_model(IAS_OLI, READ_COMPLETE_MODEL);
    assert_true(model != NULL);
    if (!model)
        return;

    l0r_data.wrs_path = 40;
    l0r_data.wrs_row = 33;
    memset(&l0r_data.l0r_sensor_data, 0, sizeof(l0r_data.l0r_sensor_data));
    l0r_data.l0r_sensor_data.bands_present[0] = 1;

    /* The getmjdcoords needs the band epoch time set to a valid date/time
       in the CPF */
    model->sensor.bands[0].utc_epoch_time[0] = 2008.0;
    model->sensor.bands[0].utc_epoch_time[1] = 182.0;
    model->sensor.bands[0].utc_epoch_time[2] = 65294.816;
    model->sensor.bands[0].band_present = 1;
    model->sensor.band_count = 11;

    status = ias_los_model_set_l0r(0, cpf, &l0r_data, ssm_position_filename,
        &ssm_position_quality, model);
    assert_equal(status, SUCCESS);

    /* Check the WRS path/row were set properly. Sensor level components
       are fully tested in the sensor model library test. */
    assert_equal(40, model->wrs_path);
    assert_equal(33, model->wrs_row);

    assert_equal(ssm_position_quality, IAS_SSM_POSITION_ACTUAL);

    ias_los_model_free(model);
    ias_cpf_free(cpf);
}

void setup_prec_params
(
    IAS_SC_PRECISION_MODEL *params
)
{
    params->seconds_from_image_epoch = 18474;
    params->ephemeris_order = 2;
    params->x_corr[0] = 0.1;
    params->x_corr[1] = 0.1;
    params->y_corr[0] = 0.2;
    params->y_corr[1] = 0.2;
    params->z_corr[0] = 0.3;
    params->z_corr[1] = 0.3;
    params->attitude_order = 2;
    params->roll_corr[0] = 0.4;
    params->roll_corr[1] = 0.4;
    params->pitch_corr[0] = 0.5;
    params->pitch_corr[1] = 0.5;
    params->yaw_corr[0] = 0.6;
    params->yaw_corr[1] = 0.6;
}

Ensure test_precision_params()
{
    IAS_LOS_MODEL *model;
    IAS_SC_PRECISION_MODEL prec_params;
    int last_rec;
    int status;

    #include "precision_params_test_data.h"

    ias_sat_attr_initialize(IAS_L8);

    /* get the test model */
    model = read_test_model(IAS_OLI, READ_COMPLETE_MODEL);
    assert_true(model != NULL);
    if (!model)
        return;

    /* Initialize the leap seconds for the geo library calls that the 
       routine calls */
    initialize_leap_seconds(0);

    setup_prec_params(&prec_params);

    status = ias_los_model_apply_precision_params(&prec_params, model);
    assert_equal(status, SUCCESS);

    /* Check the 1st, 2nd, and last precision attitude sample records
       for accuracy */
    assert_double_equal_with_message(att_output_1st_rec[0],
            model->spacecraft.attitude.sample_records[0].precision_attitude.x,
            "failed attitude precision correction - expected %.8e, got %.8e "
            "for precision attitude x, index 0", att_output_1st_rec[0],
            model->spacecraft.attitude.sample_records[0].precision_attitude.x);
    assert_double_equal_with_message(att_output_1st_rec[1],
            model->spacecraft.attitude.sample_records[0].precision_attitude.y,
            "failed attitude precision correction - expected %.8e, got %.8e "
            "for precision attitude y, index 0", att_output_1st_rec[1],
            model->spacecraft.attitude.sample_records[0].precision_attitude.y);
    assert_double_equal_with_message(att_output_1st_rec[2],
            model->spacecraft.attitude.sample_records[0].precision_attitude.z,
            "failed attitude precision correction - expected %.8e, got %.8e "
            "for precision attitude z, index 0", att_output_1st_rec[2],
            model->spacecraft.attitude.sample_records[0].precision_attitude.z);
    assert_double_equal_with_message(att_output_2nd_rec[0],
            model->spacecraft.attitude.sample_records[1].precision_attitude.x,
            "failed attitude precision correction - expected %.8e, got %.8e "
            "for precision attitude x, index 1", att_output_2nd_rec[0],
            model->spacecraft.attitude.sample_records[1].precision_attitude.x);
    assert_double_equal_with_message(att_output_2nd_rec[1],
            model->spacecraft.attitude.sample_records[1].precision_attitude.y,
            "failed attitude precision correction - expected %.8e, got %.8e "
            "for precision attitude y, index 1", att_output_2nd_rec[1],
            model->spacecraft.attitude.sample_records[1].precision_attitude.y);
    assert_double_equal_with_message(att_output_2nd_rec[2],
            model->spacecraft.attitude.sample_records[1].precision_attitude.z,
            "failed attitude precision correction - expected %.8e, got %.8e "
            "for precision attitude z, index 1", att_output_2nd_rec[2],
            model->spacecraft.attitude.sample_records[1].precision_attitude.z);
    last_rec = model->spacecraft.attitude.sample_count - 1;
    assert_double_equal_with_message(att_output_last_rec[0],
            model->spacecraft.attitude.sample_records[last_rec]
            .precision_attitude.x,
            "failed attitude precision correction - expected %.8e, got %.8e "
            "for precision attitude x, index %d", att_output_last_rec[0],
            model->spacecraft.attitude.sample_records[last_rec]
            .precision_attitude.x, last_rec);
    assert_double_equal_with_message(att_output_last_rec[1],
            model->spacecraft.attitude.sample_records[last_rec]
            .precision_attitude.y,
            "failed attitude precision correction - expected %.8e, got %.8e "
            "for precision attitude y, index %d", att_output_last_rec[1],
            model->spacecraft.attitude.sample_records[last_rec]
            .precision_attitude.y, last_rec);
    assert_double_equal_with_message(att_output_last_rec[2],
            model->spacecraft.attitude.sample_records[last_rec]
            .precision_attitude.z,
            "failed attitude precision correction - expected %.8e, got %.8e "
            "for precision attitude z, index %d", att_output_last_rec[2],
            model->spacecraft.attitude.sample_records[last_rec]
            .precision_attitude.z, last_rec);

    /* Check the 1st, 2nd, last precision ephemeris sample records
       for accuracy */
    /* Index 0 values */
    assert_double_equal_with_message(eph_output_1st_rec[0],
            model->spacecraft.ephemeris.sample_records[0]
            .precision_eci_position.x, "failed ephemeris precision correction "
            "- expected %.8e, got %.8e for precision ephemeris ECI position x, "
            "index 0", eph_output_1st_rec[0],
            model->spacecraft.ephemeris.sample_records[0]
            .precision_eci_position.x);
    assert_double_equal_with_message(eph_output_1st_rec[1],
            model->spacecraft.ephemeris.sample_records[0]
            .precision_eci_position.y, "failed ephemeris precision correction "
            "- expected %.8e, got %.8e for precision ephemeris ECI position y, "
            "index 0", eph_output_1st_rec[1],
            model->spacecraft.ephemeris.sample_records[0]
            .precision_eci_position.y);
    assert_double_equal_with_message(eph_output_1st_rec[2],
            model->spacecraft.ephemeris.sample_records[0]
            .precision_eci_position.z, "failed ephemeris precision correction "
            "- expected %.8e, got %.8e for precision ephemeris ECI position z, "
            "index 0", eph_output_1st_rec[2],
            model->spacecraft.ephemeris.sample_records[0]
            .precision_eci_position.z);
    assert_double_equal_with_message(eph_output_1st_rec[3],
            model->spacecraft.ephemeris.sample_records[0]
            .precision_eci_velocity.x, "failed ephemeris precision correction "
            "- expected %.8e, got %.8e for precision ephemeris ECI velocity x, "
            "index 0", eph_output_1st_rec[3],
            model->spacecraft.ephemeris.sample_records[0]
            .precision_eci_velocity.x);
    assert_double_equal_with_message(eph_output_1st_rec[4],
            model->spacecraft.ephemeris.sample_records[0]
            .precision_eci_velocity.y, "failed ephemeris precision correction "
            "- expected %.8e, got %.8e for precision ephemeris ECI velocity y, "
            "index 0", eph_output_1st_rec[4],
            model->spacecraft.ephemeris.sample_records[0]
            .precision_eci_velocity.y);
    assert_double_equal_with_message(eph_output_1st_rec[5],
            model->spacecraft.ephemeris.sample_records[0]
            .precision_eci_velocity.z, "failed ephemeris precision correction "
            "- expected %.8e, got %.8e for precision ephemeris ECI velocity z, "
            "index 0", eph_output_1st_rec[5],
            model->spacecraft.ephemeris.sample_records[0]
            .precision_eci_velocity.z);
    assert_double_equal_with_message(eph_output_1st_rec[6],
            model->spacecraft.ephemeris.sample_records[0]
            .precision_ecef_position.x, "failed ephemeris precision correction "
            "- expected %.8e, got %.8e for precision ephemeris ECEF position "
            "x, index 0", eph_output_1st_rec[6],
            model->spacecraft.ephemeris.sample_records[0]
            .precision_ecef_position.x);
    assert_double_equal_with_message(eph_output_1st_rec[7],
            model->spacecraft.ephemeris.sample_records[0]
            .precision_ecef_position.y, "failed ephemeris precision correction "
            "- expected %.8e, got %.8e for precision ephemeris ECEF position "
            "y, index 0", eph_output_1st_rec[7],
            model->spacecraft.ephemeris.sample_records[0]
            .precision_ecef_position.y);
    assert_double_equal_with_message(eph_output_1st_rec[8],
            model->spacecraft.ephemeris.sample_records[0]
            .precision_ecef_position.z, "failed ephemeris precision correction "
            "- expected %.8e, got %.8e for precision ephemeris ECEF position "
            "z, index 0", eph_output_1st_rec[8],
            model->spacecraft.ephemeris.sample_records[0]
            .precision_ecef_position.z);
    assert_double_equal_with_message(eph_output_1st_rec[9],
            model->spacecraft.ephemeris.sample_records[0]
            .precision_ecef_velocity.x, "failed ephemeris precision correction "
            "- expected %.8e, got %.8e for precision ephemeris ECEF velocity "
            "x, index 0", eph_output_1st_rec[9],
            model->spacecraft.ephemeris.sample_records[0]
            .precision_ecef_velocity.x);
    assert_double_equal_with_message(eph_output_1st_rec[10],
            model->spacecraft.ephemeris.sample_records[0]
            .precision_ecef_velocity.y, "failed ephemeris precision correction "
            "- expected %.8e, got %.8e for precision ephemeris ECEF velocity "
            "y, index 0", eph_output_1st_rec[10],
            model->spacecraft.ephemeris.sample_records[0]
            .precision_ecef_velocity.y);
    assert_double_equal_with_message(eph_output_1st_rec[11],
            model->spacecraft.ephemeris.sample_records[0]
            .precision_ecef_velocity.z, "failed ephemeris precision correction "
            "- expected %.8e, got %.8e for precision ephemeris ECEF velocity "
            "z, index 0", eph_output_1st_rec[11],
            model->spacecraft.ephemeris.sample_records[0]
            .precision_ecef_velocity.z);

    /* Index 1 values */
    assert_double_equal_with_message(eph_output_2nd_rec[0],
            model->spacecraft.ephemeris.sample_records[1]
            .precision_eci_position.x, "failed ephemeris precision correction "
            "- expected %.8e, got %.8e for precision ephemeris ECI position x, "
            "index 1", eph_output_2nd_rec[0],
            model->spacecraft.ephemeris.sample_records[1]
            .precision_eci_position.x);
    assert_double_equal_with_message(eph_output_2nd_rec[1],
            model->spacecraft.ephemeris.sample_records[1]
            .precision_eci_position.y, "failed ephemeris precision correction "
            "- expected %.8e, got %.8e for precision ephemeris ECI position y, "
            "index 1", eph_output_2nd_rec[1],
            model->spacecraft.ephemeris.sample_records[1]
            .precision_eci_position.y);
    assert_double_equal_with_message(eph_output_2nd_rec[2],
            model->spacecraft.ephemeris.sample_records[1]
            .precision_eci_position.z, "failed ephemeris precision correction "
            "- expected %.8e, got %.8e for precision ephemeris ECI position z, "
            "index 1", eph_output_2nd_rec[2],
            model->spacecraft.ephemeris.sample_records[1]
            .precision_eci_position.z);
    assert_double_equal_with_message(eph_output_2nd_rec[3],
            model->spacecraft.ephemeris.sample_records[1]
            .precision_eci_velocity.x, "failed ephemeris precision correction "
            "- expected %.8e, got %.8e for precision ephemeris ECI velocity x, "
            "index 1", eph_output_2nd_rec[3],
            model->spacecraft.ephemeris.sample_records[1]
            .precision_eci_velocity.x);
    assert_double_equal_with_message(eph_output_2nd_rec[4],
            model->spacecraft.ephemeris.sample_records[1]
            .precision_eci_velocity.y, "failed ephemeris precision correction "
            "- expected %.8e, got %.8e for precision ephemeris ECI velocity y, "
            "index 1", eph_output_2nd_rec[4],
            model->spacecraft.ephemeris.sample_records[1]
            .precision_eci_velocity.y);
    assert_double_equal_with_message(eph_output_2nd_rec[5],
            model->spacecraft.ephemeris.sample_records[1]
            .precision_eci_velocity.z, "failed ephemeris precision correction "
            "- expected %.8e, got %.8e for precision ephemeris ECI velocity z, "
            "index 1", eph_output_2nd_rec[5],
            model->spacecraft.ephemeris.sample_records[1]
            .precision_eci_velocity.z);
    assert_double_equal_with_message(eph_output_2nd_rec[6],
            model->spacecraft.ephemeris.sample_records[1]
            .precision_ecef_position.x, "failed ephemeris precision correction "
            "- expected %.8e, got %.8e for precision ephemeris ECEF position "
            "x, index 1", eph_output_2nd_rec[6],
            model->spacecraft.ephemeris.sample_records[1]
            .precision_ecef_position.x);
    assert_double_equal_with_message(eph_output_2nd_rec[7],
            model->spacecraft.ephemeris.sample_records[1]
            .precision_ecef_position.y, "failed ephemeris precision correction "
            "- expected %.8e, got %.8e for precision ephemeris ECEF position "
            "y, index 1", eph_output_2nd_rec[7],
            model->spacecraft.ephemeris.sample_records[1]
            .precision_ecef_position.y);
    assert_double_equal_with_message(eph_output_2nd_rec[8],
            model->spacecraft.ephemeris.sample_records[1]
            .precision_ecef_position.z, "failed ephemeris precision correction "
            "- expected %.8e, got %.8e for precision ephemeris ECEF position "
            "z, index 1", eph_output_2nd_rec[8],
            model->spacecraft.ephemeris.sample_records[1]
            .precision_ecef_position.z);
    assert_double_equal_with_message(eph_output_2nd_rec[9],
            model->spacecraft.ephemeris.sample_records[1]
            .precision_ecef_velocity.x, "failed ephemeris precision correction "
            "- expected %.8e, got %.8e for precision ephemeris ECEF velocity "
            "x, index 1", eph_output_2nd_rec[9],
            model->spacecraft.ephemeris.sample_records[1]
            .precision_ecef_velocity.x);
    assert_double_equal_with_message(eph_output_2nd_rec[10],
            model->spacecraft.ephemeris.sample_records[1]
            .precision_ecef_velocity.y, "failed ephemeris precision correction "
            "- expected %.8e, got %.8e for precision ephemeris ECEF velocity "
            "y, index 1", eph_output_2nd_rec[10],
            model->spacecraft.ephemeris.sample_records[1]
            .precision_ecef_velocity.y);
    assert_double_equal_with_message(eph_output_2nd_rec[11],
            model->spacecraft.ephemeris.sample_records[1]
            .precision_ecef_velocity.z, "failed ephemeris precision correction "
            "- expected %.8e, got %.8e for precision ephemeris ECEF velocity "
            "z, index 1", eph_output_2nd_rec[11],
            model->spacecraft.ephemeris.sample_records[1]
            .precision_ecef_velocity.z);

    /* Index last_rec values */
    last_rec = model->spacecraft.ephemeris.sample_count - 1;
    assert_double_equal_with_message(eph_output_last_rec[0],
            model->spacecraft.ephemeris.sample_records[last_rec]
            .precision_eci_position.x, "failed ephemeris precision correction "
            "- expected %.8e, got %.8e for precision ephemeris ECI position x, "
            "index %d", eph_output_last_rec[0],
            model->spacecraft.ephemeris.sample_records[last_rec]
            .precision_eci_position.x, last_rec);
    assert_double_equal_with_message(eph_output_last_rec[1],
            model->spacecraft.ephemeris.sample_records[last_rec]
            .precision_eci_position.y, "failed ephemeris precision correction "
            "- expected %.8e, got %.8e for precision ephemeris ECI position y, "
            "index %d", eph_output_last_rec[1],
            model->spacecraft.ephemeris.sample_records[last_rec]
            .precision_eci_position.y, last_rec);
    assert_double_equal_with_message(eph_output_last_rec[2],
            model->spacecraft.ephemeris.sample_records[last_rec]
            .precision_eci_position.z, "failed ephemeris precision correction "
            "- expected %.8e, got %.8e for precision ephemeris ECI position z, "
            "index %d", eph_output_last_rec[2],
            model->spacecraft.ephemeris.sample_records[last_rec]
            .precision_eci_position.z, last_rec);
    assert_double_equal_with_message(eph_output_last_rec[3],
            model->spacecraft.ephemeris.sample_records[last_rec]
            .precision_eci_velocity.x, "failed ephemeris precision correction "
            "- expected %.8e, got %.8e for precision ephemeris ECI velocity x, "
            "index %d", eph_output_last_rec[3],
            model->spacecraft.ephemeris.sample_records[last_rec]
            .precision_eci_velocity.x, last_rec);
    assert_double_equal_with_message(eph_output_last_rec[4],
            model->spacecraft.ephemeris.sample_records[last_rec]
            .precision_eci_velocity.y, "failed ephemeris precision correction "
            "- expected %.8e, got %.8e for precision ephemeris ECI velocity y, "
            "index %d", eph_output_last_rec[4],
            model->spacecraft.ephemeris.sample_records[last_rec]
            .precision_eci_velocity.y, last_rec);
    assert_double_equal_with_message(eph_output_last_rec[5],
            model->spacecraft.ephemeris.sample_records[last_rec]
            .precision_eci_velocity.z, "failed ephemeris precision correction "
            "- expected %.8e, got %.8e for precision ephemeris ECI velocity z, "
            "index %d", eph_output_last_rec[5],
            model->spacecraft.ephemeris.sample_records[last_rec]
            .precision_eci_velocity.z, last_rec);
    assert_double_equal_with_message(eph_output_last_rec[6],
            model->spacecraft.ephemeris.sample_records[last_rec]
            .precision_ecef_position.x, "failed ephemeris precision correction "
            "- expected %.8e, got %.8e for precision ephemeris ECEF position "
            "x, index %d", eph_output_last_rec[6],
            model->spacecraft.ephemeris.sample_records[last_rec]
            .precision_ecef_position.x, last_rec);
    assert_double_equal_with_message(eph_output_last_rec[7],
            model->spacecraft.ephemeris.sample_records[last_rec]
            .precision_ecef_position.y, "failed ephemeris precision correction "
            "- expected %.8e, got %.8e for precision ephemeris ECEF position "
            "y, index %d", eph_output_last_rec[7],
            model->spacecraft.ephemeris.sample_records[last_rec]
            .precision_ecef_position.y, last_rec);
    assert_double_equal_with_message(eph_output_last_rec[8],
            model->spacecraft.ephemeris.sample_records[last_rec]
            .precision_ecef_position.z, "failed ephemeris precision correction "
            "- expected %.8e, got %.8e for precision ephemeris ECEF position "
            "z, index %d", eph_output_last_rec[8],
            model->spacecraft.ephemeris.sample_records[last_rec]
            .precision_ecef_position.z, last_rec);
    assert_double_equal_with_message(eph_output_last_rec[9],
            model->spacecraft.ephemeris.sample_records[last_rec]
            .precision_ecef_velocity.x, "failed ephemeris precision correction "
            "- expected %.8e, got %.8e for precision ephemeris ECEF velocity "
            "x, index %d", eph_output_last_rec[9],
            model->spacecraft.ephemeris.sample_records[last_rec]
            .precision_ecef_velocity.x, last_rec);
    assert_double_equal_with_message(eph_output_last_rec[10],
            model->spacecraft.ephemeris.sample_records[last_rec]
            .precision_ecef_velocity.y, "failed ephemeris precision correction "
            "- expected %.8e, got %.8e for precision ephemeris ECEF velocity "
            "y, index %d", eph_output_last_rec[10],
            model->spacecraft.ephemeris.sample_records[last_rec]
            .precision_ecef_velocity.y, last_rec);
    assert_double_equal_with_message(eph_output_last_rec[11],
            model->spacecraft.ephemeris.sample_records[last_rec]
            .precision_ecef_velocity.z, "failed ephemeris precision correction "
            "- expected %.8e, got %.8e for precision ephemeris ECEF velocity "
            "z, index %d", eph_output_last_rec[11],
            model->spacecraft.ephemeris.sample_records[last_rec]
            .precision_ecef_velocity.z, last_rec);

    ias_los_model_free(model);
}

/* This is a rudimentary test for the get_moon_position_at_location routine.
   At the time that routine was written, the test data from calval also
   included the jitter model changes which we have not incorporated.  So, just
   a couple of values are being tested.  That should be sufficient since the
   main thing the routine does is find the moon position at a particular time.
   Most of the options to the routine are just passed on to the
   ias_sensor_find_time routine which already has an in-depth test. */
Ensure test_get_moon_position_at_location()
{
    IAS_LOS_MODEL *model;
    int status;
    double right_ascension;
    double declination;
    double distance;
    int size;
    int index;

    /* Define an array of input values to test with and the expected results.
       Note that these results were generated from our own model after having
       cal/val validate the results are reasonable. */
    struct test_parameters
    {
        /* input parameters to the routine */
        int band_index;
        int sca_index;
        double line;
        double sample;
        IAS_SENSOR_DETECTOR_TYPE dettype;
        /* expected results */
        double right_ascension;
        double declination;
        double distance;
        int status;
    };

    /* Define a couple values to test with.  Note that a couple should be
       enough since the main thing the first 6 parameters are used for are
       to just find the right time to get the moon position for. */
    struct test_parameters test_data[] =
    { {0, 0, 0.0, 0.0, IAS_NOMINAL_DETECTOR, -1.539300291156e+00,
          -4.0962988221e-01,  373441340.069170, SUCCESS},
    {4, 13, 300.0, 120.0, IAS_NOMINAL_DETECTOR, -1.539301511156e+00,
        -4.0964525640e-01,  373448713.042029, SUCCESS} };
    size = sizeof(test_data) / sizeof(struct test_parameters);

    ias_sat_attr_initialize(IAS_L8);

    /* get the test model */
    model = read_lunar_test_model();
    assert_true(model != NULL);
    if (!model)
        return;

    /* Initialize the leap seconds for the geo library calls that the 
       routine calls */
    initialize_leap_seconds(1);

    status = ias_geo_initialize_novas();
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
        return;

    /* call the get moon position at location routine for the different
       test data values */
    for (index = 0; index < size; index++)
    {
        struct test_parameters *data = &test_data[index];
        status = ias_los_model_get_moon_position_at_location(model,
                data->band_index, data->sca_index, data->line, data->sample,
                data->dettype, &right_ascension, &declination, &distance);
        assert_equal(status, data->status);
        assert_double_equal(right_ascension, data->right_ascension);
        assert_double_equal(declination, data->declination);
        assert_double_equal(distance, data->distance);
    }

    ias_los_model_free(model);
    ias_geo_shutdown_novas();
}

Ensure test_lunar_projection()
{
    IAS_LUNAR_PROJECTION *proj;
    IAS_LOS_MODEL *model;
    int status;
    int index;
    int cache;

    struct test_parameters
    {
        double line;
        double sample;
        double in_lat;
        double in_long;
        double out_lat;
        double out_long;
        double distance_scale;
    };
    /* The following test data was obtained by running the grid with the
       test model and capturing a few results */
    struct test_parameters test_data[] =
    {
        {0.000000, 0.000000, -4.2815574675e-01, -1.5557114447e+00,
            -4.2847493250e-01, -1.5542526767e+00, 9.9956969502e-01},
        {0.000000, 493.000000, -4.3555360722e-01, -1.5350497253e+00,
            -4.3586960964e-01, -1.5354800139e+00, 9.9956969502e-01},
        {12984.000000, 0.000000, -3.8700221227e-01, -1.5399579393e+00,
            -3.8667667648e-01, -1.5398737205e+00, 1.0004368556e+00},
        {12984.000000, 493.000000, -3.9418386560e-01, -1.5195739889e+00,
            -3.9386151592e-01, -1.5210406627e+00, 1.0004367878e+00},
        {2019.000000, 199.000000, -4.2476041577e-01, -1.5449037813e+00,
            -4.2498011076e-01, -1.5444271752e+00, 9.9970279187e-01}
    };
    int size = sizeof(test_data) / sizeof(test_data[0]);

    ias_sat_attr_initialize(IAS_L8);

    /* get the test model */
    model = read_lunar_test_model();
    assert_true(model != NULL);
    if (!model)
        return;

    /* Initialize the leap seconds for the geo library calls that the 
       routine calls */
    initialize_leap_seconds(1);

    status = ias_geo_initialize_novas();
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
        return;

    /* test uncached and cached to make sure they are the same */
    for (cache = 0; cache <=1; cache++)
    {
        /* create the lunar projection */
        proj = ias_los_model_create_lunar_projection(model, 0, 0,
                IAS_NOMINAL_DETECTOR, 1.0, cache);

        assert_true(proj != NULL);
        if (proj != NULL)
        {
            for (index = 0; index < size; index++)
            {
                struct test_parameters *data = &test_data[index];
                double distance_scale;
                double lat = data->in_lat;
                double lng = data->in_long;
                status = ias_los_model_transform_lunar_projection(proj,
                        data->line, data->sample, &lat, &lng, &distance_scale);

                assert_equal(status, SUCCESS);
                assert_double_equal(lat, data->out_lat);
                assert_double_equal(lng, data->out_long);
                assert_double_equal(distance_scale, data->distance_scale);
            }
            ias_los_model_free_lunar_projection(proj);
        }
    }

    ias_los_model_free(model);
    ias_geo_shutdown_novas();
}

Ensure test_get_moon_center()
{
    IAS_LUNAR_PROJECTION *proj;
    IAS_LOS_MODEL *model;
    int status;
    double moon_declination;
    double moon_right_ascension;

    ias_sat_attr_initialize(IAS_L8);

    /* get the test model */
    model = read_lunar_test_model();
    assert_true(model != NULL);
    if (!model)
        return;

    /* Initialize the leap seconds for the geo library calls that the 
       routine calls */
    initialize_leap_seconds(1);

    status = ias_geo_initialize_novas();
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
        return;

    /* create the lunar projection using SECONDS scale */
    proj = ias_los_model_create_lunar_projection(model, 0, 0,
        IAS_NOMINAL_DETECTOR, ias_math_get_arcsec_to_radian_conversion(), 0);

    assert_true(proj != NULL);
    if (proj != NULL)
    {
        /* Test getting scaled lunar center right ascension and declination. 
           Hardcoded values were calculated by hand using arcseconds */ 
        ias_los_model_get_moon_center(proj, &moon_declination,
            &moon_right_ascension);
        assert_double_equal(moon_declination, -84559.718378548932);
        assert_double_equal(moon_right_ascension, -317509.91515249148); 

        ias_los_model_free_lunar_projection(proj);
    }

    ias_los_model_free(model);
    ias_geo_shutdown_novas();
}

Ensure test_jitter_model()
{
    IAS_LOS_MODEL *model;
    IAS_CPF *cpf;
    char cpf_filename[PATH_MAX];
    const char *test_data_dir;
    int i;

    /* the following values were obtained by running the cal/val prototype
       code */
    #include "oli_jitter_table_test_data.h"

    ias_sat_attr_initialize(IAS_L8);

    test_data_dir = getenv("IAS_UNIT_TEST_DATA_DIR");
    assert_not_equal(test_data_dir, NULL);

    snprintf(cpf_filename, sizeof(cpf_filename),
            "%s/ias_lib/los_model/L8CPF20090101_20090331.01", test_data_dir);

    /* get the OLI test model - the pre jitter table version */
    /* This version of the model is generated by running create_los_model,
       but having it skip running the build_jitter_table routine. Apply the
       create_los_model_unit_test.patch in this los_model/tests directory to
       create_los_model.c to easily make a model without building the jitter
       table. */
    model = read_test_model(IAS_OLI, READ_PARTIAL_MODEL_NO_JITTER_TABLE_BUILT);
    assert_true(model != NULL);
    if (!model)
        return;

    /* since the jitter calculations are very sensitive, the frame time
       differences between calval and our code affect things and cause the
       testing to fail (we round our frame times to the nearest microsecond,
       calval does not). So load up the frame times calval used so our jitter
       calculations results can match theirs. */
    for (i = 0; i < model->sensor.frame_counts[IAS_OLI]; i++)
    {
        model->sensor.frame_seconds_from_epoch[IAS_OLI][i]
            = calval_frame_times[i];
    }

    cpf = ias_cpf_read(cpf_filename);
    assert_not_equal(cpf, NULL);

    ias_los_model_build_jitter_table(cpf, model);

    assert_equal_with_message(12100,
            model->sensor.sensors[IAS_OLI].jitter_table_count,
            "Expected 12100, got %d for OLI jitter table count",
            model->sensor.sensors[IAS_OLI].jitter_table_count);
    assert_equal(2, model->sensor.sensors[IAS_OLI].jitter_entries_per_frame);

    /* only check every 5th jitter table vector due to the cgreen limit of
       16384 asserts per run */
    for (i = 0; i < model->sensor.sensors[IAS_OLI].jitter_table_count; i += 5)
    {
        assert_double_equal_with_message(output[i*3],
                model->sensor.sensors[IAS_OLI].jitter_table[i].x,
                "Expected %.8e, got %.8e for OLI jitter table roll, index %d",
                output[i*3],
                model->sensor.sensors[IAS_OLI].jitter_table[i].x, i);
        assert_double_equal_with_message(output[i*3+1],
                model->sensor.sensors[IAS_OLI].jitter_table[i].y,
                "Expected %.8e, got %.8e for OLI jitter table pitch, index %d",
                output[i*3+1],
                model->sensor.sensors[IAS_OLI].jitter_table[i].y, i);
        assert_double_equal_with_message(output[i*3+2],
                model->sensor.sensors[IAS_OLI].jitter_table[i].z,
                "Expected %.8e, got %.8e for OLI jitter table yaw, index %d",
                output[i*3+2],
                model->sensor.sensors[IAS_OLI].jitter_table[i].z, i);
    }

    /* Check some of the low frequency attitude data left in the attitude
       model structure. Only check some of them due to the cgreen limit of
       16384 asserts per run. Check some different areas of the data and
       check them at differing intervals just to do as thorough of a check
       as possible without checking them all. */
    for (i = 200; (i < 250) && (i < model->spacecraft.attitude.sample_count);
            i += 7)
    {
        assert_double_equal_with_message(sat_attitude[i*3],
                model->spacecraft.attitude.sample_records[i]
                .satellite_attitude.x,
                "Expected %.8e, got %.8e for OLI low frequency attitude roll, "
                "index %d", sat_attitude[i*3],
                model->spacecraft.attitude.sample_records[i]
                .satellite_attitude.x);
        assert_double_equal_with_message(sat_attitude[i*3+1],
                model->spacecraft.attitude.sample_records[i]
                .satellite_attitude.y,
                "Expected %.8e, got %.8e for OLI low frequency attitude pitch, "
                "index %d", sat_attitude[i*3+1],
                model->spacecraft.attitude.sample_records[i]
                .satellite_attitude.y);
        assert_double_equal_with_message(sat_attitude[i*3+2],
                model->spacecraft.attitude.sample_records[i]
                .satellite_attitude.z,
                "Expected %.8e, got %.8e for OLI low frequency attitude yaw, "
                "index %d", sat_attitude[i*3+2],
                model->spacecraft.attitude.sample_records[i]
                .satellite_attitude.z);
        assert_double_equal_with_message(prec_attitude[i*3],
                model->spacecraft.attitude.sample_records[i]
                .precision_attitude.x,
                "Expected %.8e, got %.8e for OLI low frequency attitude "
                "precision roll, index %d", prec_attitude[i*3],
                model->spacecraft.attitude.sample_records[i]
                .precision_attitude.x);
        assert_double_equal_with_message(prec_attitude[i*3+1],
                model->spacecraft.attitude.sample_records[i]
                .precision_attitude.y,
                "Expected %.8e, got %.8e for OLI low frequency attitude "
                "precision pitch, index %d", prec_attitude[i*3+1],
                model->spacecraft.attitude.sample_records[i]
                .precision_attitude.y);
        assert_double_equal_with_message(prec_attitude[i*3+2],
                model->spacecraft.attitude.sample_records[i]
                .precision_attitude.z,
                "Expected %.8e, got %.8e for OLI low frequency attitude "
                "precision yaw, index %d", prec_attitude[i*3+2],
                model->spacecraft.attitude.sample_records[i]
                .precision_attitude.z);
    }
    for (i = 13937; (i < 14300) &&
            (i < model->spacecraft.attitude.sample_count); i += 11)
    {
        assert_double_equal_with_message(sat_attitude[i*3],
                model->spacecraft.attitude.sample_records[i]
                .satellite_attitude.x,
                "Expected %.8e, got %.8e for OLI low frequency attitude roll, "
                "index %d", sat_attitude[i*3],
                model->spacecraft.attitude.sample_records[i]
                .satellite_attitude.x);
        assert_double_equal_with_message(sat_attitude[i*3+1],
                model->spacecraft.attitude.sample_records[i]
                .satellite_attitude.y,
                "Expected %.8e, got %.8e for OLI low frequency attitude pitch, "
                "index %d", sat_attitude[i*3+1],
                model->spacecraft.attitude.sample_records[i]
                .satellite_attitude.y);
        assert_double_equal_with_message(sat_attitude[i*3+2],
                model->spacecraft.attitude.sample_records[i]
                .satellite_attitude.z,
                "Expected %.8e, got %.8e for OLI low frequency attitude yaw, "
                "index %d", sat_attitude[i*3+2],
                model->spacecraft.attitude.sample_records[i]
                .satellite_attitude.z);
        assert_double_equal_with_message(prec_attitude[i*3],
                model->spacecraft.attitude.sample_records[i]
                .precision_attitude.x,
                "Expected %.8e, got %.8e for OLI low frequency attitude "
                "precision roll, index %d", prec_attitude[i*3],
                model->spacecraft.attitude.sample_records[i]
                .precision_attitude.x);
        assert_double_equal_with_message(prec_attitude[i*3+1],
                model->spacecraft.attitude.sample_records[i]
                .precision_attitude.y,
                "Expected %.8e, got %.8e for OLI low frequency attitude "
                "precision pitch, index %d", prec_attitude[i*3+1],
                model->spacecraft.attitude.sample_records[i]
                .precision_attitude.y);
        assert_double_equal_with_message(prec_attitude[i*3+2],
                model->spacecraft.attitude.sample_records[i]
                .precision_attitude.z,
                "Expected %.8e, got %.8e for OLI low frequency attitude "
                "precision yaw, index %d", prec_attitude[i*3+2],
                model->spacecraft.attitude.sample_records[i]
                .precision_attitude.z);
    }

    ias_los_model_free(model);
}

Ensure test_tirs_jitter_model()
{
    IAS_LOS_MODEL *model;
    IAS_CPF *cpf;
    char cpf_filename[PATH_MAX];
    const char *test_data_dir;
    int i;

    /* the following values were obtained by running the cal/val prototype
       code */
    #include "tirs_jitter_table_test_data.h"

    ias_sat_attr_initialize(IAS_L8);

    test_data_dir = getenv("IAS_UNIT_TEST_DATA_DIR");
    assert_not_equal(test_data_dir, NULL);

    snprintf(cpf_filename, sizeof(cpf_filename),
            "%s/ias_lib/los_model/L8CPF20000701_20000930.87", test_data_dir);

    /* get the TIRS test model - the pre jitter table version */
    /* This version of the model is generated by running create_los_model,
       but having it skip running the build_jitter_table routine. Apply the
       create_los_model_unit_test.patch in this los_model/tests directory to
       create_los_model.c to easily make a model without building the jitter
       table. */
    model = read_test_model(IAS_TIRS, READ_PARTIAL_MODEL_NO_JITTER_TABLE_BUILT);
    assert_true(model != NULL);
    if (!model)
        return;

    /* since the jitter calculations are very sensitive, the frame time
       differences between calval and our code affect things and cause the
       testing to fail (we round our frame times to the nearest microsecond,
       calval does not). So load up the frame times calval used so our jitter
       calculations results can match theirs. */
    for (i = 0; i < model->sensor.frame_counts[IAS_TIRS]; i++)
    {
        model->sensor.frame_seconds_from_epoch[IAS_TIRS][i]
            = calval_frame_times[i];
    }

    cpf = ias_cpf_read(cpf_filename);
    assert_not_equal(cpf, NULL);

    ias_los_model_build_jitter_table(cpf, model);

    assert_equal_with_message(2291,
            model->sensor.sensors[IAS_TIRS].jitter_table_count,
            "Expected 2291, got %d for TIRS jitter table count",
            model->sensor.sensors[IAS_TIRS].jitter_table_count);
    assert_equal(1, model->sensor.sensors[IAS_TIRS].jitter_entries_per_frame);

    /* only check every 5th vector due to the cgreen limit of 16384 asserts
       per run */
    for (i = 0; i < model->sensor.sensors[IAS_TIRS].jitter_table_count; i += 5)
    {
        assert_double_equal_with_message(output[i*3],
                model->sensor.sensors[IAS_TIRS].jitter_table[i].x,
                "Expected %.8e, got %.8e for TIRS jitter table roll, index %d",
                output[i*3],
                model->sensor.sensors[IAS_TIRS].jitter_table[i].x, i);
        assert_double_equal_with_message(output[i*3+1],
                model->sensor.sensors[IAS_TIRS].jitter_table[i].y,
                "Expected %.8e, got %.8e for TIRS jitter table pitch, index %d",
                output[i*3+1],
                model->sensor.sensors[IAS_TIRS].jitter_table[i].y, i);
        assert_double_equal_with_message(output[i*3+2],
                model->sensor.sensors[IAS_TIRS].jitter_table[i].z,
                "Expected %.8e, got %.8e for TIRS jitter table yaw, index %d",
                output[i*3+2],
                model->sensor.sensors[IAS_TIRS].jitter_table[i].z, i);
    }

    /* Check some of the low frequency attitude data left in the attitude
       model structure. Only check some of them due to the cgreen limit of
       16384 asserts per run. Check some different areas of the data and
       check them at differing intervals just to do as thorough of a check
       as possible without checking them all. */
    for (i = 200; (i < 250) && (i < model->spacecraft.attitude.sample_count);
            i += 7)
    {
        assert_double_equal_with_message(sat_attitude[i*3],
                model->spacecraft.attitude.sample_records[i]
                .satellite_attitude.x,
                "Expected %.8e, got %.8e for TIRS low frequency attitude roll, "
                "index %d", sat_attitude[i*3],
                model->spacecraft.attitude.sample_records[i]
                .satellite_attitude.x, i);
        assert_double_equal_with_message(sat_attitude[i*3+1],
                model->spacecraft.attitude.sample_records[i]
                .satellite_attitude.y,
                "Expected %.8e, got %.8e for TIRS low frequency attitude "
                "pitch, index %d", sat_attitude[i*3+1],
                model->spacecraft.attitude.sample_records[i]
                .satellite_attitude.y, i);
        assert_double_equal_with_message(sat_attitude[i*3+2],
                model->spacecraft.attitude.sample_records[i]
                .satellite_attitude.z,
                "Expected %.8e, got %.8e for TIRS low frequency attitude yaw, "
                "index %d", sat_attitude[i*3+2],
                model->spacecraft.attitude.sample_records[i]
                .satellite_attitude.z, i);
        assert_double_equal_with_message(prec_attitude[i*3],
                model->spacecraft.attitude.sample_records[i]
                .precision_attitude.x,
                "Expected %.8e, got %.8e for TIRS low frequency attitude "
                "precision roll, index %d", prec_attitude[i*3],
                model->spacecraft.attitude.sample_records[i]
                .precision_attitude.x, i);
        assert_double_equal_with_message(prec_attitude[i*3+1],
                model->spacecraft.attitude.sample_records[i]
                .precision_attitude.y,
                "Expected %.8e, got %.8e for TIRS low frequency attitude "
                "precision pitch, index %d", prec_attitude[i*3+1],
                model->spacecraft.attitude.sample_records[i]
                .precision_attitude.y, i);
        assert_double_equal_with_message(prec_attitude[i*3+2],
                model->spacecraft.attitude.sample_records[i]
                .precision_attitude.z,
                "Expected %.8e, got %.8e for TIRS low frequency attitude "
                "precision yaw, index %d", prec_attitude[i*3+2],
                model->spacecraft.attitude.sample_records[i]
                .precision_attitude.z, i);
    }
    for (i = 13937; (i < 14300) &&
            (i < model->spacecraft.attitude.sample_count); i += 11)
    {
        assert_double_equal_with_message(sat_attitude[i*3],
                model->spacecraft.attitude.sample_records[i]
                .satellite_attitude.x,
                "Expected %.8e, got %.8e for TIRS low frequency attitude roll, "
                "index %d", sat_attitude[i*3],
                model->spacecraft.attitude.sample_records[i]
                .satellite_attitude.x, i);
        assert_double_equal_with_message(sat_attitude[i*3+1],
                model->spacecraft.attitude.sample_records[i]
                .satellite_attitude.y,
                "Expected %.8e, got %.8e for TIRS low frequency attitude "
                "pitch, index %d", sat_attitude[i*3+1],
                model->spacecraft.attitude.sample_records[i]
                .satellite_attitude.y, i);
        assert_double_equal_with_message(sat_attitude[i*3+2],
                model->spacecraft.attitude.sample_records[i]
                .satellite_attitude.z,
                "Expected %.8e, got %.8e for TIRS low frequency attitude yaw, "
                "index %d", sat_attitude[i*3+2],
                model->spacecraft.attitude.sample_records[i]
                .satellite_attitude.z, i);
        assert_double_equal_with_message(prec_attitude[i*3],
                model->spacecraft.attitude.sample_records[i]
                .precision_attitude.x,
                "Expected %.8e, got %.8e for TIRS low frequency attitude "
                "precision roll, index %d", prec_attitude[i*3],
                model->spacecraft.attitude.sample_records[i]
                .precision_attitude.x, i);
        assert_double_equal_with_message(prec_attitude[i*3+1],
                model->spacecraft.attitude.sample_records[i]
                .precision_attitude.y,
                "Expected %.8e, got %.8e for TIRS low frequency attitude "
                "precision pitch, index %d", prec_attitude[i*3+1],
                model->spacecraft.attitude.sample_records[i]
                .precision_attitude.y, i);
        assert_double_equal_with_message(prec_attitude[i*3+2],
                model->spacecraft.attitude.sample_records[i]
                .precision_attitude.z,
                "Expected %.8e, got %.8e for TIRS low frequency attitude "
                "precision yaw, index %d", prec_attitude[i*3+2],
                model->spacecraft.attitude.sample_records[i]
                .precision_attitude.z, i);
    }
    ias_los_model_free(model);
}

Ensure test_calc_scene_corners()
{
    IAS_LOS_MODEL *model;
    const char *test_data_dir;
    int status;
    struct IAS_CORNERS corners;
    int number_of_lines;
    int band_number = 9;  /* use band number 9 for tests */
    int band_index;

    ias_sat_attr_initialize(IAS_L8);

    test_data_dir = getenv("IAS_UNIT_TEST_DATA_DIR");
    assert_not_equal(test_data_dir, NULL);

    /* Read the model */
    model = read_test_model(IAS_OLI, READ_COMPLETE_MODEL);
    assert_true(model != NULL);
    if (!model)
        return;

    /* Get the band index for the band number */
    band_index = ias_sat_attr_convert_band_number_to_index(band_number);
    assert_not_equal(band_index, ERROR);
    if (band_index == ERROR)
        return;

    /* calculate the ending line number from the model info */
    number_of_lines =  model->sensor.bands[band_index].frame_count
        * model->sensor.bands[band_index].sampling_char.lines_per_frame;

    /* Call the routine to calculate the scene corners */
    status = ias_los_model_calc_scene_corners(model, band_number, 0,
            number_of_lines - 1, 0.0, &corners);
    assert_equal(status, SUCCESS);

    /* Verify the corners are correct */
    assert_double_equal(corners.upleft.x, -116.088996);
    assert_double_equal(corners.upleft.y, 35.490488);
    assert_double_equal(corners.upright.x, -114.044000);
    assert_double_equal(corners.upright.y, 35.100101);
    assert_double_equal(corners.loleft.x, -116.453899);
    assert_double_equal(corners.loleft.y, 34.137744);
    assert_double_equal(corners.loright.x, -114.442233);
    assert_double_equal(corners.loright.y, 33.746871);

    /* Call the routine to calculate the scene corners with the line number
       range 40 lines in from either end */
    status = ias_los_model_calc_scene_corners(model, band_number, 40,
            number_of_lines - 1 - 40, 0.0, &corners);
    assert_equal(status, SUCCESS);

    /* Verify the corners are correct */
    assert_double_equal(corners.upleft.x, -116.091812);
    assert_double_equal(corners.upleft.y, 35.480326);
    assert_double_equal(corners.upright.x, -114.047050);
    assert_double_equal(corners.upright.y, 35.089964);
    assert_double_equal(corners.loleft.x, -116.451105);
    assert_double_equal(corners.loleft.y, 34.147913);
    assert_double_equal(corners.loright.x, -114.439193);
    assert_double_equal(corners.loright.y, 33.756999);

    /* Call the routine to calculate the scene corners with an elevation of
       2000 meters */
    status = ias_los_model_calc_scene_corners(model, band_number, 0,
            number_of_lines - 1, 2000.0, &corners);
    assert_equal(status, SUCCESS);

    /* Verify the corners are correct */
    assert_double_equal(corners.upleft.x, -116.085683);
    assert_double_equal(corners.upleft.y, 35.490124);
    assert_double_equal(corners.upright.x, -114.047112);
    assert_double_equal(corners.upright.y, 35.100963);
    assert_double_equal(corners.loleft.x, -116.450790);
    assert_double_equal(corners.loleft.y, 34.136815);
    assert_double_equal(corners.loright.x, -114.445459);
    assert_double_equal(corners.loright.y, 33.747173);

    /* Disable log messages for the next tests which are expected to return
       errors */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* Test an illegal band number */
    status = ias_los_model_calc_scene_corners(model, -1, 0, number_of_lines - 1,
            0.0, &corners);
    assert_equal(status, ERROR);

    /* Test a band that isn't in the model to make sure it fails */
    status = ias_los_model_calc_scene_corners(model, 10, 0, number_of_lines - 1,
            0.0, &corners);
    assert_equal(status, ERROR);

    /* Test an invalid starting line number */
    status = ias_los_model_calc_scene_corners(model, 9, -1,
            number_of_lines - 1, 0.0, &corners);
    assert_equal(status, ERROR);

    /* Test an invalid ending line number */
    status = ias_los_model_calc_scene_corners(model, 9, 0, number_of_lines,
            0.0, &corners);
    assert_equal(status, ERROR);

    /* Return the logging level back to normal */
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

    /* Free the model */
    ias_los_model_free(model);
}

Ensure test_tirs_calc_scene_corners()
{
    IAS_LOS_MODEL *model;
    const char *test_data_dir;
    int status;
    struct IAS_CORNERS corners;
    int number_of_lines;
    int band_number = 10;  /* use band number 10 for tests */
    int band_index;

    ias_sat_attr_initialize(IAS_L8);

    test_data_dir = getenv("IAS_UNIT_TEST_DATA_DIR");
    assert_not_equal(test_data_dir, NULL);

    /* Read the model */
    model = read_test_model(IAS_TIRS, READ_COMPLETE_MODEL);
    assert_true(model != NULL);
    if (!model)
        return;

    /* Get the band index for the band number */
    band_index = ias_sat_attr_convert_band_number_to_index(band_number);
    assert_not_equal(band_index, ERROR);
    if (band_index == ERROR)
        return;

    /* calculate the ending line number from the model info */
    number_of_lines =  model->sensor.bands[band_index].frame_count
        * model->sensor.bands[band_index].sampling_char.lines_per_frame;

    /* Call the routine to calculate the scene corners */
    status = ias_los_model_calc_scene_corners(model, band_number, 0,
            number_of_lines - 1, 0.0, &corners);
    assert_equal(status, SUCCESS);

    /* Verify the corners are correct */
    assert_double_equal(corners.upleft.x, -116.412050);
    assert_double_equal(corners.upleft.y, 39.767553);
    assert_double_equal(corners.upright.x, -114.283957);
    assert_double_equal(corners.upright.y, 39.371859);
    assert_double_equal(corners.loleft.x, -116.825108);
    assert_double_equal(corners.loleft.y, 38.355671);
    assert_double_equal(corners.loright.x, -114.738665);
    assert_double_equal(corners.loright.y, 37.961904);

    /* Call the routine to calculate the scene corners with the line number
       range 40 lines in from either end */
    status = ias_los_model_calc_scene_corners(model, band_number, 40,
            number_of_lines - 1 - 40, 0.0, &corners);
    assert_equal(status, SUCCESS);

    /* Verify the corners are correct */
    assert_double_equal(corners.upleft.x, -116.422145);
    assert_double_equal(corners.upleft.y, 39.733328);
    assert_double_equal(corners.upright.x, -114.295085);
    assert_double_equal(corners.upright.y, 39.337740);
    assert_double_equal(corners.loleft.x, -116.815245);
    assert_double_equal(corners.loleft.y, 38.389944);
    assert_double_equal(corners.loright.x, -114.727849);
    assert_double_equal(corners.loright.y, 37.996083);

    /* Call the routine to calculate the scene corners with an elevation of
       2000 meters */
    status = ias_los_model_calc_scene_corners(model, band_number, 0,
            number_of_lines - 1, 2000.0, &corners);
    assert_equal(status, SUCCESS);

    /* Verify the corners are correct */
    assert_double_equal(corners.upleft.x, -116.408418);
    assert_double_equal(corners.upleft.y, 39.767811);
    assert_double_equal(corners.upright.x, -114.286987);
    assert_double_equal(corners.upright.y, 39.373356);
    assert_double_equal(corners.loleft.x, -116.822033);
    assert_double_equal(corners.loleft.y, 38.354242);
    assert_double_equal(corners.loright.x, -114.742173);
    assert_double_equal(corners.loright.y, 37.961717);

    /* Disable log messages for the next tests which are expected to return
       errors */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* Test an illegal band number */
    status = ias_los_model_calc_scene_corners(model, -1, 0, number_of_lines - 1,
            0.0, &corners);
    assert_equal(status, ERROR);

    /* Test a band that isn't in the model to make sure it fails */
    status = ias_los_model_calc_scene_corners(model, 9, 0, number_of_lines - 1,
            0.0, &corners);
    assert_equal(status, ERROR);

    /* Test an invalid starting line number */
    status = ias_los_model_calc_scene_corners(model, 10, -1,
            number_of_lines - 1, 0.0, &corners);
    assert_equal(status, ERROR);

    /* Test an invalid ending line number */
    status = ias_los_model_calc_scene_corners(model, 10, 0, number_of_lines,
            0.0, &corners);
    assert_equal(status, ERROR);

    /* Return the logging level back to normal */
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

    /* Free the model */
    ias_los_model_free(model);
}

Ensure test_get_min_max_detector_offsets()
{
    IAS_LOS_MODEL *model;
    const char *test_data_dir;
    int status;
    double min_det_offsets[IAS_MAX_NSCAS];
    double max_det_offsets[IAS_MAX_NSCAS];
    int index;
    int sca_count;
    double exp_b9_min_offsets[] = {-2.02800012, -2.00000000, -2.01999998,
        -3.98300004, -2.01399994, -3.97600007, -4.02299976, -3.97499990,
        -4.02400017, -0.00000000, -4.03100014, -0.00000000, -0.00000000,
        -2.00500011};
    double exp_b9_max_offsets[] = {4.05600023, 3.99799991, 4.04099989,
        3.98099995, 4.02899981, 3.97600007, 6.03499985, 3.97600007, 6.03900003,
        1.99100006, 6.04699993, 1.99699998, 4.04099989, 6.01300001};
    double exp_b10_min_offsets[] = {-0.00100000, -0.00100000, -0.00100000};
    double exp_b10_max_offsets[] = {0.00100000, 6.01200008, 0.00100000};

    ias_sat_attr_initialize(IAS_L8);

    test_data_dir = getenv("IAS_UNIT_TEST_DATA_DIR");
    assert_not_equal(test_data_dir, NULL);

    /* Read the model */
    model = read_test_model(IAS_OLI, READ_COMPLETE_MODEL);
    assert_true(model != NULL);
    if (!model)
        return;

    /* Call the routine to get the min/max detector offsets */
    status = ias_los_model_get_min_max_detector_offsets(model, 9,
            min_det_offsets, max_det_offsets);
    assert_equal(status, SUCCESS);

    sca_count = ias_sat_attr_get_scas_per_band(9);
    for (index = 0; index < sca_count; index++)
    {
        assert_double_equal(min_det_offsets[index], exp_b9_min_offsets[index]);
        assert_double_equal(max_det_offsets[index], exp_b9_max_offsets[index]);
    }

    /* Free the model */
    ias_los_model_free(model);
    model = NULL;

    /* Read the TIRS model */
    model = read_test_model(IAS_TIRS, READ_COMPLETE_MODEL);
    assert_true(model != NULL);
    if (!model)
        return;
    status = ias_los_model_get_min_max_detector_offsets(model, 10,
            min_det_offsets, max_det_offsets);
    assert_equal(status, SUCCESS);

    sca_count = ias_sat_attr_get_scas_per_band(10);
    for (index = 0; index < sca_count; index++)
    {
        assert_double_equal(min_det_offsets[index], exp_b10_min_offsets[index]);
        assert_double_equal(max_det_offsets[index], exp_b10_max_offsets[index]);
    }

    /* Disable log messages for the next tests which are expected to return
       errors */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* Test an illegal band number */
    status = ias_los_model_get_min_max_detector_offsets(model, 0, 
        min_det_offsets, max_det_offsets);
    assert_equal(status, ERROR);

    /* Return the logging level back to normal */
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

    /* Free the model */
    ias_los_model_free(model);
}
int main(int argc, char **argv)
{
    TestSuite *suite = create_test_suite();
    add_test(suite, test_los_model_allocation);
    add_test(suite, test_forward_model);
    add_test(suite, test_set_cpf);
    add_test(suite, test_set_l0r);
    add_test(suite, test_precision_params);
    add_test(suite, test_oli_calc_position);
    add_test(suite, test_tirs_forward_model);
    add_test(suite, test_get_moon_position_at_location);
    add_test(suite, test_get_moon_center);
    add_test(suite, test_lunar_projection);
    add_test(suite, test_jitter_model);
    add_test(suite, test_tirs_jitter_model);
    add_test(suite, test_calc_scene_corners);
    add_test(suite, test_tirs_calc_scene_corners);
    add_test(suite, test_get_min_max_detector_offsets);
    ias_log_initialize("modeltest");
    if (argc > 1) {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}
