/* Standard Library Includes */
#include <math.h>

/* IAS Library Includes */
#include "ias_logging.h"
#include "ias_angle_gen.h"
#include "ias_angle_gen_private.h"

/*******************************************************************************
 Name: ias_angle_gen_calculate_angles_rigor
 
 Purpose: Calculates the satellite viewing and solar illumination zenith and 
          azimuth angles for a specified L1T line/sample and height (from DEM)
          using the full viewing geometry including the ephemeris data and sun
          vectors provided in the metadata. If the SCA location falls in the 
          SCA overlap area it returns the average value of the combined SCAs.
 
 Return: 
    Type = integer
    SUCCESS / ERROR
 ******************************************************************************/
int ias_angle_gen_calculate_angles_rigor
(
    const IAS_ANGLE_GEN_METADATA *metadata,/* I: Metadata structure */
    double l1t_line,    /* I: Output space line coordinate */
    double l1t_samp,    /* I: Output space sample coordinate */
    const double *elev, /* I: Pointer to input elevation or NULL if
                              mean scene height should be used */
    int band_index,     /* I: Current band index */
    IAS_ANGLE_GEN_TYPE sat_or_sun_type, /* I: Angle calculation type */
    int *outside_image_flag,/* O: Flag indicating return was outside image */
    double *angle       /* O: Array containing zenith and azimuth angles */
)       
{
    int nsca_found;     /* Number of SCAS containing the point */
    int sca_index;      /* SCA index */
    double height;      /* Model height */
    double l1r_line[2]; /* Input space (L1R) line coordinate declared size 2
                           so it can support 2 SCAs */
    double l1r_samp[2]; /* Input space (L1R) sample coordinate declared size
                           2 so it can support 2 SCAs */

    /* Check that the band index is valid */
    if (!ias_angle_gen_valid_band_index(metadata, band_index))
    {
        IAS_LOG_ERROR("Band index %d is invalid", band_index);
        return ERROR;
    }

    /* Initialize the angles */
    angle[IAS_ANGLE_GEN_ZENITH_INDEX] = 0.0;
    angle[IAS_ANGLE_GEN_AZIMUTH_INDEX] = 0.0;
    *outside_image_flag = 0;

    /* Set the height to use */
    height = metadata->band_metadata[band_index].sca_metadata[0].mean_height;
    if (elev)
    { 
        height = *elev;
    }

    /* See which SCA(s) the point falls inside */
    nsca_found = ias_angle_gen_find_scas(&metadata->band_metadata[band_index],
        l1t_line, l1t_samp, elev, l1r_line, l1r_samp);

    if (nsca_found < 1)
    { 
        *outside_image_flag = 1; /* Set the flag to indicate outside of image */
        return SUCCESS;   /* Outside active image, return zero vectors */
    }

    if (nsca_found > 2)
    {
        IAS_LOG_ERROR("Too many SCAs found locating point in active image");
        return ERROR;
    }

    /* Check the located SCAs */
    for (sca_index = 0; sca_index < nsca_found; sca_index++)
    {
        IAS_VECTOR unit_vector;  /* Normalized vector */
        IAS_VECTOR rigor_vector; /* Rigorous viewing vector */

        /* Calculate the vectors rigorously */
        if (ias_angle_gen_calculate_vector(metadata, l1t_line, l1t_samp, 
            l1r_line[sca_index], height, band_index, sat_or_sun_type, 
            &rigor_vector) != SUCCESS)
        {
            IAS_LOG_ERROR("Error computing satellite and sun vectors");
            return ERROR;
        }

        /* Normalize the vector in case the polynomial fit results in non-unit
           vectors that will fail the following trig functions */
        if (ias_math_compute_unit_vector(&rigor_vector, &unit_vector) 
            != SUCCESS)
        {
            IAS_LOG_ERROR("Unable to normalize the rigorous vector");
            return ERROR;
        }

        /* Calculate the zenith and azimuth angles */
        angle[IAS_ANGLE_GEN_ZENITH_INDEX] += acos(unit_vector.z);
        angle[IAS_ANGLE_GEN_AZIMUTH_INDEX] += atan2(unit_vector.x, 
            unit_vector.y);
    }

    /* Average the angles */
    angle[IAS_ANGLE_GEN_ZENITH_INDEX] /= nsca_found;
    angle[IAS_ANGLE_GEN_AZIMUTH_INDEX] /= nsca_found;

    return SUCCESS;
}
