/* Standard Library Includes */
#include <stdlib.h>
#include <string.h>
#include <math.h>

/* IAS Library Includes */
#include "ias_angle_gen.h"
#include "ias_logging.h"
#include "ias_angle_gen_private.h"

/* Local Defines */
#define ANG_UNKNOWN (2 * IAS_ANGLE_GEN_ANG_RPC_COEF - 1)/* Total number of RPC 
                                                         coefficients per vector
                                                         component (x/y/z) */
#define SAT_SOLAR_COUNT 6 /* Count of terms in the combined satellite and solar 
                             vectors */

#define PARAMETER_CORRELATION_THRESHOLD 0.9999  /* Maximum allowable parameter 
                                                   correlation value */

#define SINGULAR_MATRIX_FLAG -1.0       /* Flag indicating a singular matrix */

/* Struct used to contain all of the mapping data for each L1T/L1R line and
   sample combination for the entire band */
typedef struct BAND_MAP_DATA
{
    double l1t_line;
    double l1t_samp;
    double l1r_line;
    double l1r_samp;
    double height;
    double vector_comp[SAT_SOLAR_COUNT];
} BAND_MAP_DATA;

/*******************************************************************************
Name: analyze_normals

Purpose: Analyze the normal equations to see if they need to be reduced. This 
         can happen in the denominator coefficients when the data points being
         fitted are all close to zero. 

Parameters:
    equation_ptr: Array expected to be ANG_UNKNOWN x ANG_UKNOWN size using an
                  array of pointers to each row.
    max_correlation: Max correlation output value. If the input matrix is 
                     singular then the value output is SINGULAR_MATRIX_FLAG.

Return: 
    Type = integer
    SUCCESS / ERROR
 ******************************************************************************/
static int analyze_normals
(
    double *equation_ptr[], /* I: Normal equations pointer */
    double *max_correlation /* O: Maximum correlation */
)              
{
    int i;     /* Generic loop index */
    int j;     /* Generic loop index */
    int count; /* Array element counter */
    double normals[ANG_UNKNOWN * ANG_UNKNOWN];   /* Copy of normal equations */
    double normal_inv[ANG_UNKNOWN * ANG_UNKNOWN];/* Inverted normal equations */
    double local_max_correlation = 0; /* Maximum correlation coefficient */

    /* Copy normal equations */
    for (i = 0, count = 0; i < ANG_UNKNOWN; i++)
        for (j = 0; j < ANG_UNKNOWN; j++, count++)
            normals[count] = equation_ptr[i][j];

    /* Determine if this is a singular matrix */
    if (ias_math_check_if_singular_matrix(normals, ANG_UNKNOWN))
    {
        *max_correlation = SINGULAR_MATRIX_FLAG;
        return SUCCESS;
    }

    /* Invert normal equations matrix */
    if (ias_math_invert_matrix(normals, normal_inv, ANG_UNKNOWN) != SUCCESS)
    {
        IAS_LOG_ERROR("Inverting the normal equations matrix");
        return ERROR;
    }

    /* Calculate the correlation coefficients */ 
    for (j = 0; j < ANG_UNKNOWN; j++)
    {
        i = j * ANG_UNKNOWN + j;
        normal_inv[i] = sqrt(normal_inv[i]);
    }

    /*  Normalize all the correlation coefficients */
    for (i = 0, count = 0; i < ANG_UNKNOWN; i++)
    {
        for (j = 0; j < ANG_UNKNOWN; j++, count++)
        {
            if (j == i) 
                continue;

            normal_inv[count] /= normal_inv[i * ANG_UNKNOWN + i]
                * normal_inv[j * ANG_UNKNOWN + j];

            if (fabs(normal_inv[count]) > local_max_correlation)
                local_max_correlation = fabs(normal_inv[count]);
        }
    }

    /* Output the max correlation coefficient */
    *max_correlation = local_max_correlation;
    return SUCCESS;
}

/*******************************************************************************
Name: ias_angle_gen_calculate_band_rpc

Purpose: Compute the rational polynomial functions that map L1T line/sample, 
         height, and L1R line/sample to satellite view vectors and solar 
         illumination vectors for the current band.

Return: 
    Type = integer
    SUCCESS / ERROR
 ******************************************************************************/
int ias_angle_gen_calculate_band_rpc
(
    const IAS_GRID_BAND_TYPE *gridband, /* I: Grid structure for current band */
    int band_index,                     /* I: Current band index */
    IAS_ANGLE_GEN_METADATA *metadata    /* O: Metadata structure */
)          
{
    int sca_index;              /* SCA index */
    int direction;              /* Vector component index */
    int index;                  /* Grid point index */
    int grid_pts;               /* Grid point counter */
    size_t grid_size;           /* Total number of grid points */
    double *numerator[SAT_SOLAR_COUNT];  /* Pointer to numerator coeffs */
    double *denominator[SAT_SOLAR_COUNT];/* Pointer to denominator coeffs */
    double mean_offset[SAT_SOLAR_COUNT]; /* Mean vector offsets */
    IAS_ANGLE_GEN_BAND *band_ptr;/* Band data pointer */
    BAND_MAP_DATA *map_data;     /* Mapping data for entire band */

    /* Check that the band index is valid */
    if (!ias_angle_gen_valid_band_index(metadata, band_index))
    {
        IAS_LOG_ERROR("Band index %d is invalid", band_index);
        return ERROR;
    }
 
    /* Set up the band pointer */
    band_ptr = &metadata->band_metadata[band_index];

    /* Determine the total number of grid points*/
    grid_size = (band_ptr->num_scas * gridband[0].nzplanes 
        * gridband[0].ngrid_lines * gridband[0].ngrid_samps)
        * sizeof(BAND_MAP_DATA);

    /* Allocate the mapping data */
    map_data = (BAND_MAP_DATA *) malloc(grid_size);
    if (!map_data)
    {
        IAS_LOG_ERROR("Allocating band map data");
        return ERROR;
    }

    /* Initialize the means */
    band_ptr->satellite.line_terms.l1t_mean_offset = 0.0;
    band_ptr->satellite.samp_terms.l1t_mean_offset = 0.0;
    band_ptr->satellite.mean_height = 0.0;
    band_ptr->satellite.line_terms.l1r_mean_offset = 0.0;
    band_ptr->satellite.samp_terms.l1r_mean_offset = 0.0;
    band_ptr->satellite.mean_offset.x = 0.0;
    band_ptr->satellite.mean_offset.y = 0.0;
    band_ptr->satellite.mean_offset.z = 0.0;
    band_ptr->solar.line_terms.l1t_mean_offset = 0.0;
    band_ptr->solar.samp_terms.l1t_mean_offset = 0.0;
    band_ptr->solar.mean_height = 0.0;
    band_ptr->solar.line_terms.l1r_mean_offset = 0.0;
    band_ptr->solar.samp_terms.l1r_mean_offset = 0.0;
    band_ptr->solar.mean_offset.x = 0.0;
    band_ptr->solar.mean_offset.y = 0.0;
    band_ptr->solar.mean_offset.z = 0.0;
    
    /* Accumulate each point in the grid */
    for (sca_index = 0, grid_pts = 0; sca_index < band_ptr->num_scas; 
        sca_index++)
    {
        int plane; /* Grid zplane index */

        for (plane = 0; plane < gridband[sca_index].nzplanes; plane++)
        {
            int line; /* Grid line index */

            for (line = 0; line < gridband[sca_index].ngrid_lines; line++)
            {
                int sample; /* Grid sample index */

                index = (plane * gridband[sca_index].ngrid_lines + line) 
                    * gridband[sca_index].ngrid_samps;

                for (sample = 0; sample < gridband[sca_index].ngrid_samps;
                     sample++, grid_pts++)
                {
                    IAS_VECTOR satellite_vector;/* Satellite pointing vector */
                    IAS_VECTOR solar_vector;    /* Solar pointing vector */
                    BAND_MAP_DATA *cur_map = &map_data[grid_pts];

                    /* Calculate the height, l1r and l1t line/sample */
                    cur_map->height = (plane - gridband[sca_index].zeroplane) 
                        * gridband[sca_index].zspacing;
                    cur_map->l1r_line = gridband[sca_index].in_lines[line];
                    cur_map->l1r_samp = gridband[sca_index].in_samps[sample] 
                        + sca_index * band_ptr->l1r_samps;
                    cur_map->l1t_line = 
                        gridband[sca_index].out_lines[index + sample];
                    cur_map->l1t_samp = 
                        gridband[sca_index].out_samps[index + sample];

                    /* Calculate the satellite vector */
                    if (ias_angle_gen_calculate_vector(metadata, 
                        cur_map->l1t_line, cur_map->l1t_samp, cur_map->l1r_line,
                        cur_map->height, band_index, IAS_ANGLE_GEN_SATELLITE,
                        &satellite_vector) != SUCCESS)
                    {
                        IAS_LOG_ERROR("Error computing satellite vectors");
                        free(map_data);
                        return ERROR;
                    }

                    /* Calculate the solar vector */
                    if (ias_angle_gen_calculate_vector(metadata, 
                        cur_map->l1t_line, cur_map->l1t_samp, cur_map->l1r_line,
                        cur_map->height, band_index, IAS_ANGLE_GEN_SOLAR, 
                        &solar_vector) != SUCCESS)
                    {
                        IAS_LOG_ERROR("Error computing sun vectors");
                        free(map_data);
                        return ERROR;
                    }

                    /* Save the vectors for future use */
                    cur_map->vector_comp[0] = satellite_vector.x;
                    cur_map->vector_comp[1] = satellite_vector.y;
                    cur_map->vector_comp[2] = satellite_vector.z;
                    cur_map->vector_comp[3] = solar_vector.x;
                    cur_map->vector_comp[4] = solar_vector.y;
                    cur_map->vector_comp[5] = solar_vector.z;

                    band_ptr->satellite.line_terms.l1t_mean_offset 
                        += cur_map->l1t_line;
                    band_ptr->satellite.samp_terms.l1t_mean_offset 
                        += cur_map->l1t_samp;
                    band_ptr->satellite.mean_height += cur_map->height;
                    band_ptr->satellite.line_terms.l1r_mean_offset 
                        += cur_map->l1r_line;
                    band_ptr->satellite.samp_terms.l1r_mean_offset 
                        += cur_map->l1r_samp;
                    band_ptr->satellite.mean_offset.x += satellite_vector.x;
                    band_ptr->satellite.mean_offset.y += satellite_vector.y;
                    band_ptr->satellite.mean_offset.z += satellite_vector.z;
                    band_ptr->solar.line_terms.l1t_mean_offset 
                        += cur_map->l1t_line;
                    band_ptr->solar.samp_terms.l1t_mean_offset 
                        += cur_map->l1t_samp;
                    band_ptr->solar.mean_height += cur_map->height;
                    band_ptr->solar.line_terms.l1r_mean_offset 
                        += cur_map->l1r_line;
                    band_ptr->solar.samp_terms.l1r_mean_offset 
                        += cur_map->l1r_samp;
                    band_ptr->solar.mean_offset.x += solar_vector.x;
                    band_ptr->solar.mean_offset.y += solar_vector.y;
                    band_ptr->solar.mean_offset.z += solar_vector.z;
                }
            }
        }
    }

    /* Divide by the number of points */
    band_ptr->satellite.line_terms.l1t_mean_offset /= grid_pts;
    band_ptr->satellite.samp_terms.l1t_mean_offset /= grid_pts;
    band_ptr->satellite.mean_height /= grid_pts;
    band_ptr->satellite.line_terms.l1r_mean_offset /= grid_pts;
    band_ptr->satellite.samp_terms.l1r_mean_offset /= grid_pts;
    band_ptr->satellite.mean_offset.x /= grid_pts;
    band_ptr->satellite.mean_offset.y /= grid_pts;
    band_ptr->satellite.mean_offset.z /= grid_pts;
    band_ptr->solar.line_terms.l1t_mean_offset /= grid_pts;
    band_ptr->solar.samp_terms.l1t_mean_offset /= grid_pts;
    band_ptr->solar.mean_height /= grid_pts;
    band_ptr->solar.line_terms.l1r_mean_offset /= grid_pts;
    band_ptr->solar.samp_terms.l1r_mean_offset /= grid_pts;
    band_ptr->solar.mean_offset.x /= grid_pts;
    band_ptr->solar.mean_offset.y /= grid_pts;
    band_ptr->solar.mean_offset.z /= grid_pts;

    mean_offset[0] = band_ptr->satellite.mean_offset.x;
    mean_offset[1] = band_ptr->satellite.mean_offset.y;
    mean_offset[2] = band_ptr->satellite.mean_offset.z;
    mean_offset[3] = band_ptr->solar.mean_offset.x;
    mean_offset[4] = band_ptr->solar.mean_offset.y;
    mean_offset[5] = band_ptr->solar.mean_offset.z;

    /* Set up pointers for the six cases */
    numerator[0] = band_ptr->satellite.x_terms.numerator;
    numerator[1] = band_ptr->satellite.y_terms.numerator;
    numerator[2] = band_ptr->satellite.z_terms.numerator;
    numerator[3] = band_ptr->solar.x_terms.numerator;
    numerator[4] = band_ptr->solar.y_terms.numerator;
    numerator[5] = band_ptr->solar.z_terms.numerator;
    denominator[0] = band_ptr->satellite.x_terms.denominator;
    denominator[1] = band_ptr->satellite.y_terms.denominator;
    denominator[2] = band_ptr->satellite.z_terms.denominator;
    denominator[3] = band_ptr->solar.x_terms.denominator;
    denominator[4] = band_ptr->solar.y_terms.denominator;
    denominator[5] = band_ptr->solar.z_terms.denominator;

    /* Direction loop */
    for (direction = 0; direction < SAT_SOLAR_COUNT; direction++)
    {
        double norm_equations[ANG_UNKNOWN][ANG_UNKNOWN]; /* Normal equations */
        double *equation_ptr[ANG_UNKNOWN];/* Pointers to the normal equations */
        double observe[ANG_UNKNOWN]; /* Coefficients for current observation */
        double norm_vector[ANG_UNKNOWN];/* Normal equations vector */
        double reduced_normals[ANG_UNKNOWN][ANG_UNKNOWN]; /* Temporary reduced 
                                                  version of normal equations */
        double *reduced_ptr[ANG_UNKNOWN]; /* Pointers to reduced normal 
                                             equations */
        double max_correlation;  /* Largest correlation between parameters */
        int num_reduced; /* Number of denominator parameters suppressed */
        int coef_index; /* Index of denominator coefficient */
        int point;                      /* Point loop variable */
        int iteration;                  /* Internal loop indices */

        /* Initialize normal equations */
        for (index = 0; index < ANG_UNKNOWN; index++)
        {
            norm_vector[index] = 0.0;
            equation_ptr[index] = &norm_equations[index][0];
            reduced_ptr[index] = &reduced_normals[index][0];
            for (iteration = 0; iteration < ANG_UNKNOWN; iteration++)
            {
                norm_equations[iteration][index] = 0.0;
            }
        }
        observe[0] = 1.0;       /* Coefficient of constant term */
        
        /* Point loop */
        for (point = 0; point < grid_pts; point++)
        {
            double vector_offset; /* Current vector component offset value */
            BAND_MAP_DATA *cur_map = &map_data[point];

            vector_offset = cur_map->vector_comp[direction]
                - mean_offset[direction];
            observe[1] = cur_map->l1t_line 
                - band_ptr->satellite.line_terms.l1t_mean_offset;
            observe[2] = cur_map->l1t_samp
                - band_ptr->satellite.samp_terms.l1t_mean_offset;
            observe[3] = cur_map->height - band_ptr->satellite.mean_height;
            observe[4] = cur_map->l1r_line 
                - band_ptr->satellite.line_terms.l1r_mean_offset;
            observe[5] = observe[1] * observe[1];
            observe[6] = observe[1] * observe[2];
            observe[7] = observe[2] * observe[2];
            observe[8] = (cur_map->l1r_samp 
                - band_ptr->satellite.samp_terms.l1r_mean_offset) * observe[4] 
                * observe[4];
            observe[9] = observe[4] * observe[4] * observe[4];
            observe[10] = -vector_offset * observe[1];
            observe[11] = -vector_offset * observe[2];
            observe[12] = -vector_offset * observe[3];
            observe[13] = -vector_offset * observe[4];
            observe[14] = -vector_offset * observe[5];
            observe[15] = -vector_offset * observe[6];
            observe[16] = -vector_offset * observe[7];
            observe[17] = -vector_offset * observe[8];
            observe[18] = -vector_offset * observe[9];

            /* Add this observation to the normal equations */
            for (index = 0; index < ANG_UNKNOWN; index++)
            {
                norm_vector[index] += vector_offset * observe[index];
                for (iteration = 0; iteration < ANG_UNKNOWN; iteration++)
                {
                    norm_equations[iteration][index] += observe[iteration] 
                        * observe[index];
                }
            }
        }

        /* Analyze the normal equations to see if they need to be reduced */
        /* This can happen in the denominator coefficients when the data points 
           being fitted are all close to zero. */
        num_reduced = 0;

        /* Calculate the maximum correlation between parameters */
        if (analyze_normals(equation_ptr, &max_correlation) != SUCCESS)
        {
            IAS_LOG_ERROR("Analyzing normals for the initial equation matrix");
            free(map_data);
            return ERROR;
        }

        /* See if the normal equation matrix is singular */
        if (max_correlation < 0.0)
        {
            /* Try removing denominator parameters. */
            for (coef_index = IAS_ANGLE_GEN_ANG_RPC_COEF; max_correlation < 0.0
                 && num_reduced < (IAS_ANGLE_GEN_ANG_RPC_COEF - 1);
                 num_reduced++, coef_index++)
            {
                for (index = 0; index < ANG_UNKNOWN; index++)
                {
                    norm_equations[index][coef_index] = 0.0;
                    norm_equations[coef_index][index] = 0.0;
                }
                norm_equations[coef_index][coef_index] = 1.0;
                norm_vector[coef_index] = 0.0;
                if (analyze_normals(equation_ptr, &max_correlation) != SUCCESS)
                {
                    IAS_LOG_ERROR("Analyzing normals for the singular matrix");
                    free(map_data);
                    return ERROR;
                }
            }
        }

        /* Run normal equations reduction loop until correlations are within 
           threshold */
        for (; max_correlation > PARAMETER_CORRELATION_THRESHOLD &&
                num_reduced < (IAS_ANGLE_GEN_ANG_RPC_COEF - 1); num_reduced++)
        {
            /* Start at the first denominator coefficient */
            int best_index = IAS_ANGLE_GEN_ANG_RPC_COEF;

            /* Try all the denominator coefficients */
            for (coef_index = IAS_ANGLE_GEN_ANG_RPC_COEF; 
                 coef_index < ANG_UNKNOWN; coef_index++)
            {
                double reduced_correlation; /* Largest parameter correlation in 
                                               reduced normals */

                /* Make a clean copy of the normal equations */
                for (index = 0; index < ANG_UNKNOWN; index++)
                    memcpy(reduced_ptr[index], equation_ptr[index], 
                        sizeof(double) * ANG_UNKNOWN);

                /* Suppress coefficient number coef_index */
                for (index = 0; index < ANG_UNKNOWN; index++)
                {
                    reduced_normals[index][coef_index] = 0.0;
                    reduced_normals[coef_index][index] = 0.0;
                }
                reduced_normals[coef_index][coef_index] = 1.0;

                /* Calculate the maximum correlation with this reduction */
                if (analyze_normals(reduced_ptr, &reduced_correlation) 
                    != SUCCESS)
                {
                    IAS_LOG_ERROR("Analyzing normals for the reduced equation "
                        "matrix");
                    free(map_data);
                    return ERROR;
                }

                /* If this is an improvement, make coef_index the best_index */
                if (reduced_correlation < max_correlation)
                {
                    max_correlation = reduced_correlation;
                    best_index = coef_index;
                }
            }

            /* Remove the parameter that reduced parameter correlation 
               the most */
            for (index = 0; index < ANG_UNKNOWN; index++)
            {
                norm_equations[index][best_index] = 0.0;
                norm_equations[best_index][index] = 0.0;
            }
            norm_equations[best_index][best_index] = 1.0;
            norm_vector[best_index] = 0.0;

            /* Re-evaluate the parameter correlation */
            if (analyze_normals(equation_ptr, &max_correlation) != SUCCESS)
            {
                IAS_LOG_ERROR("Analyzing normals after removing the max "
                    "reduction parameter from the equation matrix");
                free(map_data);
                return ERROR;
            }
        }

        /* Check to see if the normal equation analysis found a stable
           parameterization */
        if (max_correlation > PARAMETER_CORRELATION_THRESHOLD 
            || max_correlation < 0.0)
        {
            IAS_LOG_ERROR("Unstable normal equations detected");
            free(map_data);
            return ERROR;
        }

        /* Solve for the RPC coefficients */
        if (ias_math_solve_linear_equation(equation_ptr, ANG_UNKNOWN, 
            norm_vector) != SUCCESS)
        {
            IAS_LOG_ERROR("Solving for RPC coefficients");
            free(map_data);
            return ERROR;
        }

        /* Load the RPC coefficients into the metadata structure */
        for (index = 0; index < IAS_ANGLE_GEN_ANG_RPC_COEF; index++)
        {
            numerator[direction][index] = norm_vector[index];
            if (index < (IAS_ANGLE_GEN_ANG_RPC_COEF - 1)) 
            {
                denominator[direction][index] = norm_vector[
                    IAS_ANGLE_GEN_ANG_RPC_COEF + index];
            }
        }
    }

    /* Free the map data */
    free(map_data);
    
    return SUCCESS;
}
