/* Standard Library Includes */
#include <math.h>

/* IAS Library Includes */
#include "ias_logging.h"
#include "ias_const.h"
#include "ias_geo.h"
#include "ias_angle_gen.h"
#include "ias_angle_gen_private.h"

/*******************************************************************************
Name: ias_angle_gen_l1r_to_time

Purpose: Converts the L1R line to image time.

Returns: Image time
 ******************************************************************************/
static double ias_angle_gen_l1r_to_time
(
    const IAS_ANGLE_GEN_BAND *band_metadata, /* I: Band-specific metadata */
    double l1r_line                          /* I: Level 1R line coordinate */
)      
{
    return (band_metadata->image_start_time + band_metadata->seconds_per_line 
        * l1r_line);
}

/*******************************************************************************
Name: ias_angle_gen_geo_to_ecef

Purpose: Calculates the geocentric vector corresponding to the input  
         latitude, longitude, and height.

Returns: 
    Type = integer
    SUCCESS / ERROR
 ******************************************************************************/
static void ias_angle_gen_geo_to_ecef
(
    const IAS_ANGLE_GEN_METADATA *metadata, /* I: Projection information */
    double latitude,                        /* I: Latitude in radians */
    double longitude,                       /* I: Longitude in radians */
    double height,                          /* I: Height in meters */
    IAS_VECTOR *ecef_vector                 /* O: ECEF vector */
)             
{
    double wgs84_flattening; /* Flattening value calculated from axis */

    /* Calculate the WGS84 flattening */
    wgs84_flattening = (metadata->wgs84_major_axis 
        - metadata->wgs84_minor_axis) / metadata->wgs84_major_axis;

    /* Use the GEO library routine to perform the computations */
    ias_geo_convert_geod2cart(latitude, longitude, height, 
        metadata->wgs84_major_axis, wgs84_flattening, ecef_vector);
}

/*******************************************************************************
Name: ias_angle_gen_l1t_to_proj

Purpose: Uses the scene corners and pixel size to compute the map 
         projection X/Y coordinates for the input L1T line/sample

Returns: 
    Type = integer
    SUCCESS / ERROR
 ******************************************************************************/
static void ias_angle_gen_l1t_to_proj
(
    const IAS_ANGLE_GEN_METADATA *metadata, /* I: Metadata info */
    int band_index,       /* I: Band index */
    double l1t_line,      /* I: L1T line number */
    double l1t_samp,      /* I: L1T sample number */
    double *projection_x, /* O: Projection X coordinate */
    double *projection_y  /* O: Projection Y coordinate */
)           
{
    /* Convert L1T line/sample to projection X/Y */
    /* Note, this only works for north-up images */
    *projection_x = metadata->corners.upleft.x 
        + metadata->band_metadata[band_index].pixel_size * l1t_samp;
    *projection_y = metadata->corners.upleft.y 
        - metadata->band_metadata[band_index].pixel_size * l1t_line;
}

/*******************************************************************************
Name: ias_angle_gen_calculate_vector

Purpose: Calculate the satellite viewing and solar illumination vectors at a 
         specified L1T line/sample/height location

Return:
    Type = integer 
    SUCCESS / ERROR
 ******************************************************************************/
int ias_angle_gen_calculate_vector
(
    const IAS_ANGLE_GEN_METADATA *metadata,/* I: Metadata structure */
    double l1t_line,         /* I: Current L1T line number */
    double l1t_samp,         /* I: Current L1T sample number */
    double l1r_line,         /* I: Current L1R line number */
    double height,           /* I: Current L1T height */
    int band_index,          /* I: Current band index */
    IAS_ANGLE_GEN_TYPE sat_or_sun_type, /* I: Angle type */
    IAS_VECTOR *view         /* O: View vector */
)   
{
    double projection_x;               /* Projection X coordinate */
    double projection_y;               /* Projection Y coordinate */
    double latitude;                   /* Latitude */
    double longitude;                  /* Longitude */
    double sample_time;                /* Sample time from epoch */
    IAS_VECTOR normal_vector;          /* Normalized ecef vector */
    IAS_VECTOR ecef_vertical[3];       /* Vertical ECEF basis vectors */
    IAS_VECTOR ecef_vector;            /* ECEF vector */
    IAS_ANGLE_GEN_EPHEMERIS *ephemeris;/* Ephemeris pointer */

    /* Convert the line/sample to projection coordinates */
    ias_angle_gen_l1t_to_proj(metadata, band_index, l1t_line, l1t_samp, 
        &projection_x, &projection_y);

    /* Convert the projection coordinates to latitude/longitude */
    if (ias_geo_transform_coordinate(metadata->transformation, projection_x, 
        projection_y, &longitude, &latitude) != SUCCESS)
    {
        IAS_LOG_ERROR("Converting projection X/Y to lat/long for l1t line %lf "
            "and sample %lf", l1t_line, l1t_samp);
        return ERROR;
    }

    /* Calculate local vertical basis vectors */
    ecef_vertical[0].z = 0.0;
    ecef_vertical[0].x = -sin(longitude);
    ecef_vertical[0].y = cos(longitude);
    ecef_vertical[1].z = cos(latitude);
    ecef_vertical[2].z = sin(latitude);
    ecef_vertical[1].x = -ecef_vertical[2].z * ecef_vertical[0].y;
    ecef_vertical[1].y = -ecef_vertical[2].z * -ecef_vertical[0].x;
    ecef_vertical[2].x = ecef_vertical[1].z * ecef_vertical[0].y;
    ecef_vertical[2].y = ecef_vertical[1].z * -ecef_vertical[0].x;

    /* Calculate observation time from L1R line number */
    sample_time = ias_angle_gen_l1r_to_time(&metadata->band_metadata[
        band_index], l1r_line);
    
    /* Determine the correct ephemeris */
    ephemeris = metadata->solar_vector;
    if (sat_or_sun_type == IAS_ANGLE_GEN_SATELLITE)
    {
        ephemeris = metadata->ephemeris;
    }

    /* Interpolate the ecef vector */
    if (ias_angle_gen_interpolate_ephemeris(ephemeris, metadata->ephem_count, 
        sample_time, &ecef_vector) != SUCCESS)
    {
        IAS_LOG_ERROR("Interpolating the ephemeris");
        return ERROR;
    }  

    /* Adjust the ecef vector for the satellite */
    if (sat_or_sun_type == IAS_ANGLE_GEN_SATELLITE)
    {
        IAS_VECTOR ground_position;        /* Ground ECEF vector */

        /* Calculate the ground position vector */
        ias_angle_gen_geo_to_ecef(metadata, latitude, longitude, height, 
            &ground_position);

        ecef_vector.x -= ground_position.x;
        ecef_vector.y -= ground_position.y;
        ecef_vector.z -= ground_position.z;
    } 

    /* Calculate the ECEF vector */
    if (ias_math_compute_unit_vector(&ecef_vector, &normal_vector) != SUCCESS)
    {
        IAS_LOG_ERROR("Normalizing the ECEF vector");
        return ERROR;
    }

    /* Calculate the view vector */
    view->x = ias_math_compute_3dvec_dot(&normal_vector, &ecef_vertical[0]);
    view->y = ias_math_compute_3dvec_dot(&normal_vector, &ecef_vertical[1]);
    view->z = ias_math_compute_3dvec_dot(&normal_vector, &ecef_vertical[2]);

    return SUCCESS;
}
