/* Standard Library Includes */
#include <string.h>

/* IAS Library Includes */
#include "ias_logging.h"
#include "ias_miscellaneous.h"
#include "ias_grid.h"
#include "ias_angle_gen_private.h"

/* Local Defines */
#define EPHEM_OVERLAP   8.0 /* Seconds of ephemeris desired before and after */
#define NUM_UNKNOWN (2 * IAS_ANGLE_GEN_NUM_RPC_COEF - 1)/* Total number of RPC 
                                                           coefficients per 
                                                           direction (line/ 
                                                           sample) */

/*******************************************************************************
Name: ias_angle_gen_l1t_to_l1r_rpc

Purpose: Compute the rational polynomial functions that map L1T line/sample 
         and height to L1R line/sample, for the current band and SCA.

Return: 
    Type = integer
    SUCCESS / ERROR
 ******************************************************************************/
int ias_angle_gen_l1t_to_l1r_rpc
(
    const IAS_GRID_BAND_TYPE *gridband, /* I: Input grid for current band/SCA */
    IAS_ANGLE_GEN_IMAGE_RPC *coefficients /* I/O: Image coefficients */
)         
{
    int direction;              /* Direction index */
    int line;                   /* Grid line index */
    int sample;                 /* Grid sample index */
    int plane;                  /* Grid zplane index */
    int iteration;              /* Internal loop indices */
    int index;                  /* Grid point index */
    double height;              /* Grid plane height */
    double l1r_offset;          /* Offset value of current L1R coordinate */
    double *numerator;          /* Pointer to numerator coefficients */
    double *denominator;        /* Pointer to denominator coefficients */
    double observe[NUM_UNKNOWN];/* Coefficients for current observation */
    double norm_matrix[NUM_UNKNOWN][NUM_UNKNOWN];/* Normal equations matrix */
    double *norm_equation[NUM_UNKNOWN]; /* Pointers to the normal equations */
    double norm_vector[NUM_UNKNOWN];    /* Normal equations vector */
    double line_offset;         /* Local mean offset variable */
    double samp_offset;         /* Local mean offset variable */

    IAS_ANGLE_GEN_IMAGE_RPC_TERMS *line_terms = &coefficients->line_terms;
    IAS_ANGLE_GEN_IMAGE_RPC_TERMS *samp_terms = &coefficients->samp_terms;

    /* Initialize the L1R line/sample offset */
    l1r_offset = 0.0;

    /* Calculate the mean height */
    coefficients->mean_height = (double)(gridband->nzplanes - 1 - 2 
        * gridband->zeroplane) * gridband->zspacing / 2.0;

    /* Calculate mean L1R line/sample */
    for (line = 0, line_offset = 0; line < gridband->ngrid_lines; line++)
    { 
        line_offset += (double)gridband->in_lines[line];
    }
    line_terms->l1r_mean_offset = line_offset / (double)gridband->ngrid_lines;

    for (sample = 0, samp_offset = 0; sample < gridband->ngrid_samps; sample++)
    { 
        samp_offset += (double)gridband->in_samps[sample];
    }
    samp_terms->l1r_mean_offset = samp_offset / (double)gridband->ngrid_samps;

    /* Calculate mean L1T line/sample */
    for (plane = 0, samp_offset = 0, line_offset = 0; 
         plane < gridband->nzplanes; plane++)
    {
        for (line = 0; line < gridband->ngrid_lines; line++)
        {
            index = (plane * gridband->ngrid_lines + line) 
                * gridband->ngrid_samps;
            for (sample = 0; sample < gridband->ngrid_samps; sample++)
            {
                line_offset += gridband->out_lines[index + sample];
                samp_offset += gridband->out_samps[index + sample];
            }
        }
    }
    line_terms->l1t_mean_offset = line_offset / (double)(gridband->nzplanes 
        * gridband->ngrid_lines * gridband->ngrid_samps);
    samp_terms->l1t_mean_offset = samp_offset / (double)(gridband->nzplanes 
        * gridband->ngrid_lines * gridband->ngrid_samps);

    for (direction = 0; direction < 2; direction++)
    {
        /* Prepare the data pointers */
        if (direction == 0)
        {
            numerator = line_terms->numerator;
            denominator = line_terms->denominator;
        }
        else
        {
            numerator = samp_terms->numerator;
            denominator = samp_terms->denominator;
        }

        /* Initialize normal equations */
        for (iteration = 0; iteration < NUM_UNKNOWN; iteration++)
        {
            norm_vector[iteration] = 0.0;
            norm_equation[iteration] = &norm_matrix[iteration][0];
            for (index = 0; index < NUM_UNKNOWN; index++)
            {
                norm_matrix[index][iteration] = 0.0;
            }
        }

        observe[0] = 1.0; /* Coefficient of constant term */ 

        /* Loop through the Z-planes */
        for (plane = 0; plane < gridband->nzplanes; plane++)
        {
            height = (plane - gridband->zeroplane) * gridband->zspacing;
          
            for (line = 0; line < gridband->ngrid_lines; line++)
            {
                if (direction == 0)
                {
                    l1r_offset = gridband->in_lines[line] 
                        - line_terms->l1r_mean_offset;
                }
                for (sample = 0; sample < gridband->ngrid_samps; sample++)
                {
                    if (direction == 1) 
                    {
                        l1r_offset = gridband->in_samps[sample] 
                            - samp_terms->l1r_mean_offset;
                    }

                    index = (plane * gridband->ngrid_lines + line) 
                        * gridband->ngrid_samps + sample;
                    observe[1] = gridband->out_lines[index] 
                        - line_terms->l1t_mean_offset;
                    observe[2] = gridband->out_samps[index] 
                        - samp_terms->l1t_mean_offset;
                    observe[3] = height - coefficients->mean_height;
                    observe[4] = observe[1] * observe[2];
                    observe[5] = -l1r_offset * observe[1];
                    observe[6] = -l1r_offset * observe[2];
                    observe[7] = -l1r_offset * observe[3];
                    observe[8] = -l1r_offset * observe[4];

                    /* Add this observation to the normal equations */
                    for (iteration = 0; iteration < NUM_UNKNOWN; iteration++)
                    {
                        norm_vector[iteration] += l1r_offset 
                            * observe[iteration];
                        for (index = 0; index < NUM_UNKNOWN; index++)
                        {    
                            norm_matrix[index][iteration] += observe[index] 
                                * observe[iteration];
                        }
                    }
                }
            }
        }

        /* Solve for the RPC coefficients */
        if (ias_math_solve_linear_equation(norm_equation, NUM_UNKNOWN,
            norm_vector) != SUCCESS)
        {
            IAS_LOG_ERROR("Solving for RPC coefficients");
            return ERROR;
        }

        /* Load the RPC coefficients into the metadata structure */
        for (index = 0; index < IAS_ANGLE_GEN_NUM_RPC_COEF; index++)
        {
            numerator[index] = norm_vector[index];
            if (index < (IAS_ANGLE_GEN_NUM_RPC_COEF - 1)) 
            {
                denominator[index] = norm_vector[IAS_ANGLE_GEN_NUM_RPC_COEF 
                    + index];
            }
        }
    }

    return SUCCESS;
}

/*******************************************************************************
Name: ias_angle_gen_get_model_data

Purpose: Gets all of the data from the model that is needed to calculate 
         the coefficients.

Note: Allocates the memory for the ephemeris and solar vector members 
      of the IAS_ANGLE_GEN_METADATA structure along with the 
      IAS_GEO_PROJ_TRANSFORMATION in the metadata struct. This method 
      should be called after ias_angle_gen_get_grid_data because it needs
      the projection information to create the transformation and it
      does a memset to intialize all of the memory.

Returns: 
    Type = integer
    SUCCESS / ERROR
 ******************************************************************************/
static int ias_angle_gen_get_model_data
(
    const IAS_LOS_MODEL *model,      /* I: LOS model struct */
    IAS_ANGLE_GEN_METADATA *metadata /* O: Metadata struct */
)
{
    int index;              /* Generic loop variable */
    int start_index;        /* Start index in ephemeris */
    int stop_index;         /* Stop index in ephemeris */
    double start_time;      /* Ephemeris start index time */
    double time_offset;     /* Offset between epochs */
    double image_start;     /* Offset between ephemeris and image start */
    double image_stop;      /* Offset to end of image */
    const IAS_SC_EPHEMERIS_MODEL *ephem = &model->spacecraft.ephemeris;
    double declination;     /* Calculate declination */
    double right_ascension; /* Calculated ascension */
    double mid_time;        /* Calculated middle scene time */
    double sun_time[3];     /* Local sun epoch time */
    double julian_day;      /* Julian day of sun vector */
    int month;              /* Month of sun vector */
    int day;                /* Day (of month) of sun vector */

    /* Check that the model has been initialized */
    if (!model)
    {
        IAS_LOG_ERROR("Model has not been initialized");
        return ERROR;
    }

    /* Check the bands to find the earliest image start time (normally TIRS) */
    image_start = 2 * 86400.0;
    image_stop = -2 * 86400.0;

    for (index = 0; index < IAS_MAX_NBANDS; index++)
    {
        IAS_SENSOR_BAND_MODEL *cur_band = &model->sensor.bands[index];

        if (!cur_band->band_present)
        { 
            continue;
        }

        if (ias_math_get_time_difference(cur_band->utc_epoch_time, 
            ephem->utc_epoch_time, &time_offset) != SUCCESS)
        {
            IAS_LOG_ERROR("Calculating time offset between ephemeris and "
                "image");
            return ERROR;
        }

        /* See if this is the earliest image start or latest image stop */
        if (image_start > time_offset + cur_band->frame_seconds_from_epoch[0])
        {
             image_start = time_offset + cur_band->frame_seconds_from_epoch[0];
        }

        if (image_stop < time_offset + cur_band->frame_seconds_from_epoch[
            cur_band->frame_count - 1])
        {
             image_stop = time_offset + cur_band->frame_seconds_from_epoch[
                cur_band->frame_count - 1];
        }
    }

    /* See if we found valid image start/stop times */
    if (image_start > image_stop)
    {
        IAS_LOG_ERROR("Invalid image start/stop times");
        return ERROR;
    }

    /* Adjust the image start/stop to include ephemeris overlap */
    image_start -= EPHEM_OVERLAP;
    image_stop += EPHEM_OVERLAP;

    /* Scan the ephemeris data to find the corresponding indices */
    start_index = 0;
    while (start_index < ephem->sample_count && image_start 
        > ephem->sample_records[start_index].seconds_from_epoch)
    {
        start_index++;
    }

    stop_index = ephem->sample_count - 1;
    while (stop_index >= 0 && image_stop 
        < ephem->sample_records[stop_index].seconds_from_epoch)
    { 
        stop_index--;
    }

    if (stop_index < start_index)
    {
        IAS_LOG_ERROR("Ephemeris does not cover imagery");
        return ERROR;
    }

    /* Calculate the number of ephemeris points and allocate the structure */
    metadata->ephem_count = stop_index - start_index + 1;
    if (ias_angle_gen_initialize(metadata) != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing the angle generation struct");
        return ERROR;
    }

    /* Load the ephemeris structure from the model */
    metadata->ephem_epoch_time.year = ephem->utc_epoch_time[0];
    metadata->ephem_epoch_time.day = ephem->utc_epoch_time[1];
    metadata->ephem_epoch_time.seconds = ephem->utc_epoch_time[2]
        + ephem->sample_records[start_index].seconds_from_epoch;

    start_time = ephem->sample_records[start_index].seconds_from_epoch;
    for (index = start_index; index <= stop_index; index++)
    {
        IAS_ANGLE_GEN_EPHEMERIS *meta_ephem = &metadata->ephemeris[index 
            - start_index];
        IAS_SC_EPHEMERIS_RECORD *cur_record = &ephem->sample_records[index];

        meta_ephem->sample_time = cur_record->seconds_from_epoch - start_time; 
        meta_ephem->position = cur_record->precision_ecef_position;
    }

    /* Generate a vector for each point */
    for (index = 0; index < metadata->ephem_count; index++)
    {
        IAS_ANGLE_GEN_EPHEMERIS *solar = &metadata->solar_vector[index];
        IAS_VECTOR sun_vect;    /* Sun vector ECI true-of-date */
        IAS_VECTOR sun_tod;     /* Normalized sun vector ECI true-of-date */
        IAS_VECTOR sun_j2k;     /* Normalized sun vector ECI of epoch J2000 */

        /* Set up the time */
        solar->sample_time = metadata->ephemeris[index].sample_time;
        sun_time[0] = metadata->ephem_epoch_time.year;
        sun_time[1] = metadata->ephem_epoch_time.day;
        sun_time[2] = metadata->ephem_epoch_time.seconds;

        ias_math_add_seconds_to_year_doy_sod(solar->sample_time, sun_time);

        /* Get the month and day for the image time */
        if (ias_math_convert_doy_to_month_day(sun_time[1], sun_time[0], &month, 
            &day) != SUCCESS)
        {
            IAS_LOG_ERROR("Converting Year %f, DOY %f to a month and day", 
                sun_time[0], sun_time[1]);
            ias_angle_gen_free(metadata);
            return ERROR;
        }

        /* Get Julian day needed for calculating the Sun's position */
        julian_day = ias_math_compute_full_julian_date(sun_time[0], month, 
            day, sun_time[2]);

        /* Compute the ECITOD solar vector */
        if (ias_geo_get_solar_position_vector(julian_day, &sun_vect) != SUCCESS)
        {
            IAS_LOG_ERROR("Computing solar ECITOD vector at JD %lf", 
                julian_day);
            ias_angle_gen_free(metadata);
            return ERROR;
        }

        /* Normalize the sun vector */
        if (ias_math_compute_unit_vector(&sun_vect, &sun_tod) != SUCCESS)
        {
            IAS_LOG_ERROR("Normalizing the solar vector");
            ias_angle_gen_free(metadata);
            return ERROR;
        }

        /* Convert to ECI of epoch J2000 */
        if (ias_geo_transform_tod2j2k(model->earth.ut1_utc_correction, &sun_tod,
            sun_time, &sun_j2k) != SUCCESS)
        {
            IAS_LOG_ERROR("Converting the sun vector to ECI J2000");
            ias_angle_gen_free(metadata);
            return ERROR;
        }

        /* Convert ECI J2K to ECEF */
        if (ias_geo_eci2ecef(model->earth.pole_wander_x, 
            model->earth.pole_wander_y, model->earth.ut1_utc_correction, 
            &sun_j2k, &sun_j2k, sun_time, &solar->position, 
            &sun_tod) != SUCCESS)
        {
            IAS_LOG_ERROR("Converting the sun vector to ECEF");
            ias_angle_gen_free(metadata);
            return ERROR;
        }
    }        

    /* Start of earth sun distance calculation */
    mid_time = (image_stop - image_start) / 2;
    sun_time[0] = metadata->ephem_epoch_time.year;
    sun_time[1] = metadata->ephem_epoch_time.day;
    sun_time[2] = metadata->ephem_epoch_time.seconds;

    /* Add mid seconds to sun time */
    ias_math_add_seconds_to_year_doy_sod(mid_time, sun_time);

    /* Get the month and day for the sun mid time */
    if (ias_math_convert_doy_to_month_day(sun_time[1], sun_time[0], 
        &month, &day) != SUCCESS)
    {
        IAS_LOG_ERROR("Converting Year %f, DOY %f to a month and day", 
            sun_time[0], sun_time[1]);
        ias_angle_gen_free(metadata);
        return ERROR;
    }

    /* Get Julian day needed for calculating the Sun's position */
    julian_day = ias_math_compute_full_julian_date(sun_time[0], month, 
        day, sun_time[2]);

    /* Compute the Earth-sun distance in AU. Values for right ascension
       and declination are determined as part of this calculation, so 
       variables for them must be provided as parameters. However, their
       values are not used in this context. */
    if (ias_geo_compute_sun_position(julian_day, &right_ascension, 
               &declination, &metadata->earth_sun_distance) != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating Earth-Sun distance");
        ias_angle_gen_free(metadata);
        return ERROR;
    }
        
    return SUCCESS;
}

/*******************************************************************************
Name: ias_angle_gen_get_grid_data

Purpose: Gets all of the data from the grid that is needed to calculate 
         the coefficients.

Returns: 
    Type = integer
    SUCCESS / ERROR
 ******************************************************************************/
static int ias_angle_gen_get_grid_data
(
    double major_axis,               /* I: WGS84 Major Axis */
    double minor_axis,               /* I: WGS84 Minor Axis */
    const IAS_GRID_TYPE *grid,       /* I: Grid data struct */
    IAS_ANGLE_GEN_METADATA *metadata /* O: Metadata struct */
)
{
    int index; /* Loop Variable */

    /* Check that the grid has been read in */
    if (!grid)
    {
        IAS_LOG_ERROR("Grid has not been initialized");
        return ERROR;
    }

    /* Initialize the metadata */
    memset(metadata, 0, sizeof(IAS_ANGLE_GEN_METADATA));

   /* Initialize the WGS84 ellipsoid constants */
    metadata->wgs84_major_axis = major_axis;
    metadata->wgs84_minor_axis = minor_axis;

    /* Copy projection information from grid */
    strcpy(metadata->units, grid->units);
    metadata->projection.proj_code = grid->code;
    strcpy(metadata->datum, grid->datum);
    /* Check if the datum cotains the underscore */
    if (strcmp(grid->datum, "WGS_84") == 0)
    {
        strcpy(metadata->datum, "WGS84");
    }

    metadata->projection.spheroid = grid->spheroid;
    metadata->projection.zone = grid->zone;

    /* Copy the projection parameters */
    for (index = 0; index < IAS_PROJ_PARAM_SIZE; index++)
    {
        metadata->projection.parameters[index] = grid->projprms[index];
    }
    
    /* Copy the corner coordinates */
    metadata->corners = grid->corners;
    
    return SUCCESS;
}

/*******************************************************************************
Name: ias_angle_gen_populate_metadata
 
Purpose: Populates all of the metadata except for the data specific to each
         band.
 
Note: The NOVAs library must be initialized before calling this routine.

Returns: 
    Type = integer
    SUCCESS / ERROR
 ******************************************************************************/
int ias_angle_gen_populate_metadata
(
    double major_axis,               /* I: WGS84 Major Axis */
    double minor_axis,               /* I: WGS84 Minor Axis */
    const IAS_GRID_TYPE *grid,       /* I: Grid data struct */
    const IAS_LOS_MODEL *model,      /* I: LOS model struct */
    IAS_ANGLE_GEN_METADATA *metadata /* O: Metadata struct */
)
{
    /* Pull the nessary data from the grid */
    if (ias_angle_gen_get_grid_data(major_axis, minor_axis, grid, metadata) 
        != SUCCESS)
    {
        IAS_LOG_ERROR("Populating the metadata from the grid data");
        return ERROR;
    }

    /* Pull the ephemeris data from the los model */
    if (ias_angle_gen_get_model_data(model, metadata) != SUCCESS)
    {
        IAS_LOG_ERROR("Populating the metadata from the los model data");
        return ERROR;
    }

    return SUCCESS;
}

/*******************************************************************************
Name: ias_angle_gen_get_band_data

Purpose: Initializes the band data including calculating the per SCA image
         RPCs.

Returns: 
    Type = integer
    SUCCESS / ERROR
 ******************************************************************************/
int ias_angle_gen_get_band_data
(
    int band_index,                  /* I: Input band index */
    const IAS_GRID_TYPE *grid,       /* I: Input Grid */
    const IAS_LOS_MODEL *model,      /* I: Input LOS Model */
    IAS_ANGLE_GEN_METADATA *metadata /* O: Output metadata structure */
)   
{
    int band_number;                   /* Current band number */
    int sca;                           /* SCA loop variable */
    double time[3];                    /* EPOCH time placeholder */
    IAS_SENSOR_BAND_MODEL *model_band; /* Model band pointer */
    IAS_ANGLE_GEN_BAND *metadata_band; /* Metadata band pointer */
    
    /* Check that the grid has been initialized */
    if (!grid)
    {
        IAS_LOG_ERROR("Grid has not been initialized");
        return ERROR;
    }

    /* Check that the model has been initialized */
    if (!model)
    {
        IAS_LOG_ERROR("Model has not been initialized");
        return ERROR;
    }

    /* Check that the band index is valid */
    band_number = ias_sat_attr_convert_band_index_to_number(band_index);
    if (band_number == ERROR)
    {
        IAS_LOG_ERROR("Converting band index %d to band number", band_index);
        return ERROR;
    }

    /* Check that the band index is valid */
    if (band_index < 0 || band_index >= IAS_MAX_NBANDS)
    {
        IAS_LOG_ERROR("Band index %d is invalid", band_index);
        return ERROR;
    }

    /* Check that the band is present in the grid */
    if (!grid->bands_present[band_index])
    {
        IAS_LOG_ERROR("Band index %d not present in grid", band_index);
        return ERROR;
    }

    /* Initialize the pointers and helper variables */
    model_band = &model->sensor.bands[band_index];
    metadata_band = &metadata->band_metadata[band_index];
    metadata->band_present[band_index] = TRUE;
    metadata->num_bands++;

    /* Load the band specific fields from the grid */
    metadata_band->band_number = band_number;
    metadata_band->num_scas = grid->scas_per_band[band_index];
    metadata_band->l1t_lines = grid->gridbands[band_index][0].lines;
    metadata_band->l1t_samps = grid->gridbands[band_index][0].samps;
    metadata_band->pixel_size = grid->gridbands[band_index][0].pixsize;

    /* Load the band specific fields from the model */
    metadata_band->l1r_lines = model_band->frame_count
        * model_band->sampling_char.lines_per_frame;
    metadata_band->l1r_samps = model_band->scas[0].detectors;

    time[0] = metadata->ephem_epoch_time.year;
    time[1] = metadata->ephem_epoch_time.day;
    time[2] = metadata->ephem_epoch_time.seconds;

    if (ias_math_get_time_difference(model_band->utc_epoch_time, time, 
        &metadata_band->image_start_time) != SUCCESS)
    {
        IAS_LOG_ERROR("Calculating time difference between image and "
            "ephemeris epochs");
        return ERROR;
    }
    
    if (ias_misc_calculate_trim_box(band_index, 
        &grid->active_corners[band_index], grid,
        metadata_band->active_l1t_corner_lines, 
        metadata_band->active_l1t_corner_samps) != SUCCESS)
    {
        IAS_LOG_ERROR("Converting the active corners from lat/long to "
            "line/sample");
        return ERROR;
    }

    metadata_band->image_start_time += model_band->frame_seconds_from_epoch[0];
    metadata_band->seconds_per_line = (model_band->frame_seconds_from_epoch[
        model_band->frame_count - 1] - model_band->frame_seconds_from_epoch[0])
        / (model_band->frame_count - 1) 
        / (double)model_band->sampling_char.lines_per_frame;

    /* Loop through the SCAs */
    for (sca = 0; sca < metadata_band->num_scas; sca++)
    {
        IAS_ANGLE_GEN_IMAGE_RPC *cur_sca = &metadata_band->sca_metadata[sca];

        /* Generate the image RPCs */
        cur_sca->sca_number = grid->gridbands[band_index][sca].sca_number;
        if (ias_angle_gen_l1t_to_l1r_rpc(&grid->gridbands[band_index][sca], 
            cur_sca) != SUCCESS)
        {
            IAS_LOG_ERROR("Generating RPCs for Band %d SCA %d", band_number, 
                cur_sca->sca_number);
            return ERROR;
        }
    }
 
    return SUCCESS;
}
