/* IAS Library Includes */
#include "ias_logging.h"
#include "ias_const.h"
#include "ias_angle_gen_private.h"

/* Local Defines */
#define NUM_INTERP  4 /* Number of points to use in Lagrange interpolation */

/*******************************************************************************
Name: ias_angle_gen_interpolate_ephem

Purpose: Interpolate the provided ephemeris structure points at the provided
         input time.

Returns: 
     Type = integer
     SUCCESS / ERROR
 ******************************************************************************/
int ias_angle_gen_interpolate_ephemeris
(
    const IAS_ANGLE_GEN_EPHEMERIS *ephemeris, /* I: Metadata ephemeris points */
    int ephem_count,             /* I: Number of entries in ephem and time */
    double in_time,              /* I: Time from epoch to interpolate */
    IAS_VECTOR *vector           /* O: Output interpolated vector */
)  
{
    int point;                   /* Loop index */
    int index;                   /* Subindex at which to start interpolation */
    int used_count;              /* Number of points used */
    IAS_VECTOR ephem[NUM_INTERP];/* Local array of vectors for interpolation */
    double time[NUM_INTERP];     /* Local array of sample times */

    if (ephem_count < 1)
    {
        IAS_LOG_ERROR("Not enough points provided to interpolate");
        return ERROR;
    }

    /* If not enough points load all of them */
    if (ephem_count <= NUM_INTERP)
    {
        for (point = 0; point < ephem_count; point++)
        {
            ephem[point] = ephemeris[point].position;
            time[point] = ephemeris[point].sample_time;
        }
        used_count = ephem_count;
    }
    else
    {
        /* Find the samples that straddle the desired time */
        for (point = 0; point < ephem_count && ephemeris[point].sample_time 
             < in_time; point++);
      
        point -= (NUM_INTERP + 1) / 2;

        if (point < 0)
        { 
            point = 0;
        }
        if (point > (ephem_count - NUM_INTERP))
        { 
            point = ephem_count - NUM_INTERP;
        }

        /* Load the local array */
        for (index = 0; index < NUM_INTERP; index++)
        {
            ephem[index] = ephemeris[point + index].position;
            time[index] = ephemeris[point + index].sample_time;
        }
        used_count = NUM_INTERP;
    }

    /* Run the interpolation on the data window */
    ias_math_interpolate_lagrange_3dvec(ephem, time, used_count, in_time,
        vector);

    return SUCCESS;
}
