/* Standard Library Includes */
#include <stdlib.h>
#include <math.h>
#include <string.h>

/* IAS Library Includes */
#include "ias_logging.h"         
#include "ias_angle_gen.h"
#include "ias_angle_gen_private.h"
#include "ias_angle_gen_includes.h"

/* Local Defines */
#define SAMPLE_TIME_INDEX 0    /* Ephemeris array sample time index */
#define POSITION_X_INDEX 1     /* Ephemeris array position x index */
#define POSITION_Y_INDEX 2     /* Ephemeris array position y index */
#define POSITION_Z_INDEX 3     /* Ephemeris array position z index */
#define EPHEMERIS_ARRAY_SIZE 4 /* Ephemeris array size */

/*******************************************************************************
Name: convert_ephemeris_to_array
 
Purpose: Converts the angle generation ephemeris struct to an array.
 
Returns: 
    Type = integer
    SUCCESS / ERROR
 ******************************************************************************/
static int convert_ephemeris_to_array
(
    IAS_ANGLE_GEN_EPHEMERIS *ephemeris, /* I: Ephemeris metadata */
    int ephemeris_count,                /* I: Number of ephemeris points */
    double **ephemeris_array            /* O: Ephemeris output arrays */
)
{
    int index;

    /* Loop through each index in the array allocating the memory needed */
    for (index = 0; index < EPHEMERIS_ARRAY_SIZE; index++)
    {
        ephemeris_array[index] = (double *)malloc(sizeof(double) 
            * ephemeris_count);
        if (!ephemeris_array[index])
        {
            int loop;
            IAS_LOG_ERROR("Converting ephemeris struct to array");
            for (loop = 0; loop < index; loop++)
            {
                free(ephemeris_array[loop]);
            }
            return ERROR;
        }
    }

    /* Add the data to the arrays */
    for (index = 0; index < ephemeris_count; index++)
    {
        IAS_ANGLE_GEN_EPHEMERIS *cur_ephem = &ephemeris[index];
        ephemeris_array[SAMPLE_TIME_INDEX][index] = cur_ephem->sample_time;
        ephemeris_array[POSITION_X_INDEX][index] = cur_ephem->position.x;
        ephemeris_array[POSITION_Y_INDEX][index] = cur_ephem->position.y;
        ephemeris_array[POSITION_Z_INDEX][index] = cur_ephem->position.z;
    }
    
    return SUCCESS;
}

/*******************************************************************************
Name: write_single_parameter_array

Purpose: Prints out the provided value to file. If there needs to be a
         newline this routine will automatically add it according to the
         current index. This routine also accepts the precision as 2
         parameters and can print exponent values.

Note: This routine expects that the first value has already been printed to
      file.

Returns: 
    Type = integer
    SUCCESS / ERROR
 ******************************************************************************/
static int write_single_parameter_array
(    
    FILE *output_file_ptr,  /* I: File pointer */
    const double *buffer,   /* I: Buffer to print from */
    int buffer_size,        /* I: Buffer size */
    int integer_precision,  /* I: Integer precision format */
    int decimal_precision,  /* I: Decimal precision format */
    int num_spaces,         /* I: Number of spaces after the newline */
    int use_exponents_flag  /* I: Print value as exponent */
)
{
    int status;         /* Routine status return variable */
    int index;          /* Generic loop index */

    /* Writing out the band satellite x numerator coefficients */
    for (index = 0; index < buffer_size; index++)
    {
        double value = buffer[index];

        /* Check if newline needs to be added */
        if (!(index % 5) && index != 0)
        {
            /* Print the newline value to file */
            status = fprintf(output_file_ptr, "\n");
            if (status < 0)
            {
                IAS_LOG_ERROR("Writing newline to file");
                return ERROR;
            }

            /* Print the number of spaces after the newline */
            status = fprintf(output_file_ptr,"%*s", num_spaces, "");
            if (status < 0)
            {
                IAS_LOG_ERROR("Padding array values");  
                return ERROR;
            }
        }
        
        /* Print the value in the specified format */
        if (use_exponents_flag)
        {
            status = fprintf(output_file_ptr, "%*.*le", integer_precision, 
                decimal_precision, value);
        }
        else
        {
            status = fprintf(output_file_ptr, "%*.*f", integer_precision, 
                decimal_precision, value);
        }
        if (status < 0)
        {
            IAS_LOG_ERROR("Writing value to file at index %d", index);
            return ERROR;
        }
        
        /* Print the next comma to separate the values */
        if (index != buffer_size - 1)
        {
            status = fprintf(output_file_ptr, ", ");
            if (status < 0)
            {
                IAS_LOG_ERROR("Comma separating the values");
                return ERROR;
            }
        }
    }

    return SUCCESS;
}

/*******************************************************************************
Name: ias_angle_gen_write_ang_sca

Purpose: Write the angle SCA metadata to file. 

Returns: 
    Type = integer
    SUCCESS / ERROR
 ******************************************************************************/
static int ias_angle_gen_write_ang_sca
(
    const char *band_string, /* I: Band string formatted as BAND%02d */ 
    const IAS_ANGLE_GEN_IMAGE_RPC *cur_sca, /* I: SCA metadata pointer */
    FILE *output_file_ptr    /* O: Output file pointer */
)
{
    char band_sca_string[13]; /* Band & SCA combination string */
    int status;               /* Status placeholder */

    /* Create the band string */
    status = snprintf(band_sca_string, sizeof(band_sca_string), "%s_SCA%02d", 
        band_string, cur_sca->sca_number);
    if (status < 0 || status >= sizeof(band_sca_string))
    {
        IAS_LOG_ERROR("Constructing the band/sca string for SCA %d", 
            cur_sca->sca_number);
        return ERROR;
    } 

    status = fprintf(output_file_ptr, 
        "  %s_MEAN_HEIGHT = %6.3lf\n"
        "  %s_MEAN_L1R_LINE_SAMP = (%8.3lf, %8.3lf)\n" 
        "  %s_MEAN_L1T_LINE_SAMP = (%8.3lf, %8.3lf)\n"
        "  %s_LINE_NUM_COEF = (", 
        band_sca_string, cur_sca->mean_height, band_sca_string, 
        cur_sca->line_terms.l1r_mean_offset, 
        cur_sca->samp_terms.l1r_mean_offset, band_sca_string, 
        cur_sca->line_terms.l1t_mean_offset, 
        cur_sca->samp_terms.l1t_mean_offset, band_sca_string);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing the %s means", band_sca_string);
        return ERROR;
    }

    /* SCA line numerator coefficients */
    if (write_single_parameter_array(output_file_ptr, 
        cur_sca->line_terms.numerator, IAS_ANGLE_GEN_NUM_RPC_COEF,
        13, 6, 18, 1) != SUCCESS)
    {
        IAS_LOG_ERROR("Writing the %s_LINE_NUM_COEF parameter values", 
            band_sca_string);
        return ERROR;
    }

    status = fprintf(output_file_ptr, ")\n  %s_LINE_DEN_COEF = (", 
        band_sca_string);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing the %s_LINE_DEN_COEF parameter", 
            band_sca_string); 
        return ERROR;
    }

    /* SCA line denominator coefficients */
    if (write_single_parameter_array(output_file_ptr, 
        cur_sca->line_terms.denominator, IAS_ANGLE_GEN_NUM_RPC_COEF - 1,
        13, 6, 18, 1) != SUCCESS)
    {
        IAS_LOG_ERROR("Writing the %s_LINE_DEN_COEF parameter values",
            band_sca_string);
        return ERROR;
    }

    status = fprintf(output_file_ptr, ")\n  %s_SAMP_NUM_COEF = (", 
        band_sca_string);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing the %s_SAMP_NUM_COEF parameter", 
            band_sca_string);
        return ERROR;
    }

    /* SCA sample numerator coefficients */
    if (write_single_parameter_array(output_file_ptr, 
        cur_sca->samp_terms.numerator, IAS_ANGLE_GEN_NUM_RPC_COEF,
        13, 6, 18, 1) != SUCCESS)
    {        
        IAS_LOG_ERROR("Writing the %s_SAMP_NUM_COEF parameter values",
            band_sca_string);
        return ERROR;  
    }

    status = fprintf(output_file_ptr, ")\n  %s_SAMP_DEN_COEF = (", 
        band_sca_string);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing the %s_SAM_DEN_COEF parameter", band_sca_string);
        return ERROR;
    }

    /* SCA sample denominator coefficients */
    if (write_single_parameter_array(output_file_ptr, 
        cur_sca->samp_terms.denominator, IAS_ANGLE_GEN_NUM_RPC_COEF - 1,
        13, 6, 18, 1) != SUCCESS)
    {
        IAS_LOG_ERROR("Writing the %s_SAMP_DEN_COEF group values", 
            band_sca_string);
        return ERROR;
    }

    status = fprintf(output_file_ptr, ")\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing the %s_SAM_DEN_COEF group values closing", 
            band_sca_string);
        return ERROR;
    }
    
    return SUCCESS;
}

/*******************************************************************************
Name: ias_angle_gen_write_ang_band

Purpose: Write the angle coefficients band metadata to file. 

Returns: 
    Type = integer
    SUCCESS / ERROR
 ******************************************************************************/
static int ias_angle_gen_write_ang_band
(
    const IAS_ANGLE_GEN_BAND *cur_band, /* I: Band metadata pointer */
    FILE *output_file_ptr               /* O: Output file pointer */
)
{
    char band_string[7];    /* Current band string identifier */
    int status;             /* Status placeholder */
    int sca_index;          /* SCA loop variable */

    /* Create the band string */
    status = snprintf(band_string, sizeof(band_string), "BAND%02d", 
        cur_band->band_number);
    if (status < 0 || status >= sizeof(band_string))
    {
        IAS_LOG_ERROR("Constructing the band string for band number %d", 
            cur_band->band_number);
        return ERROR;
    } 

    status = fprintf(output_file_ptr, 
        "GROUP = RPC_%s\n"
        "  %s_NUMBER_OF_SCAS = %d\n"        
        "  %s_NUM_L1T_LINES = %d\n"
        "  %s_NUM_L1T_SAMPS = %d\n"
        "  %s_L1T_IMAGE_CORNER_LINES = (%12.6f, %12.6f, %12.6f, %12.6f)\n"
        "  %s_L1T_IMAGE_CORNER_SAMPS = (%12.6f, %12.6f, %12.6f, %12.6f)\n"
        "  %s_NUM_L1R_LINES = %d\n"       
        "  %s_NUM_L1R_SAMPS = %d\n" 
        "  %s_PIXEL_SIZE = %6.3lf\n"       
        "  %s_START_TIME = %10.6lf\n"
        "  %s_LINE_TIME = %8.9lf\n"       
        "  %s_MEAN_HEIGHT = %6.3lf\n"
        "  %s_MEAN_L1R_LINE_SAMP = (%8.3lf, %8.3lf)\n"
        "  %s_MEAN_L1T_LINE_SAMP = (%8.3lf, %8.3lf)\n"
        "  %s_MEAN_SAT_VECTOR = (%12.9lf, %12.9lf, %12.9lf)\n"
        "  %s_SAT_X_NUM_COEF = (", 
        band_string, band_string, cur_band->num_scas, band_string, 
        cur_band->l1t_lines, band_string, cur_band->l1t_samps, band_string, 
        cur_band->active_l1t_corner_lines[0], 
        cur_band->active_l1t_corner_lines[1],
        cur_band->active_l1t_corner_lines[2],
        cur_band->active_l1t_corner_lines[3], band_string, 
        cur_band->active_l1t_corner_samps[0],
        cur_band->active_l1t_corner_samps[1],
        cur_band->active_l1t_corner_samps[2],
        cur_band->active_l1t_corner_samps[3], band_string,  
        cur_band->l1r_lines, band_string, cur_band->l1r_samps,
        band_string, cur_band->pixel_size, band_string, 
        cur_band->image_start_time, band_string, cur_band->seconds_per_line, 
        band_string, cur_band->satellite.mean_height, band_string,
        cur_band->satellite.line_terms.l1r_mean_offset,
        cur_band->satellite.samp_terms.l1r_mean_offset, band_string, 
        cur_band->satellite.line_terms.l1t_mean_offset, 
        cur_band->satellite.samp_terms.l1t_mean_offset, band_string, 
        cur_band->satellite.mean_offset.x, 
        cur_band->satellite.mean_offset.y, 
        cur_band->satellite.mean_offset.z, band_string);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing the RPC_%s group", band_string); 
        return ERROR;
    }

    /* Writing out the band satellite x numerator coefficients */
    if (write_single_parameter_array(output_file_ptr,
        cur_band->satellite.x_terms.numerator, IAS_ANGLE_GEN_ANG_RPC_COEF,
        13, 6, 27, 1) != SUCCESS)
    {
        IAS_LOG_ERROR("Writing the %s_SAT_X_NUM_COEF parameter", 
            band_string);
        return ERROR;
    }
    
    status = fprintf(output_file_ptr, ")\n  %s_SAT_X_DEN_COEF = (", 
        band_string);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing the %s_SAT_X_DEN_COEF parameter", 
            band_string);
        return ERROR;
    }

    /* Writing out the band satellite x denominator coefficients */
    if (write_single_parameter_array(output_file_ptr,
        cur_band->satellite.x_terms.denominator, IAS_ANGLE_GEN_ANG_RPC_COEF - 1,
        13, 6, 27, 1) != SUCCESS)
    {
        IAS_LOG_ERROR("Writing the %s_SAT_X_DEN_COEF parameter values", 
            band_string);
        return ERROR;
    }

    status = fprintf(output_file_ptr, ")\n  %s_SAT_Y_NUM_COEF = (", 
        band_string);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing the %s_SAT_Y_NUM_COEF parameter", band_string);
        return ERROR;
    }

    /* Writing out the band satellite y numerator coefficients */
    if (write_single_parameter_array(output_file_ptr,
        cur_band->satellite.y_terms.numerator, IAS_ANGLE_GEN_ANG_RPC_COEF,
        13, 6, 27, 1) != SUCCESS)
    {
        IAS_LOG_ERROR("Writing the %s_SAT_Y_NUM_COEF parameter values", 
            band_string);
        return ERROR;
    }

    status = fprintf(output_file_ptr, ")\n  %s_SAT_Y_DEN_COEF = (", 
        band_string);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing the %s_SAT_Y_DEN_COEF parameter", band_string);
        return ERROR;
    }

    /* Writing out the band satellite y denominator coefficients */
    if (write_single_parameter_array(output_file_ptr,
        cur_band->satellite.y_terms.denominator, IAS_ANGLE_GEN_ANG_RPC_COEF - 1,
        13, 6, 27, 1) != SUCCESS)
    {
        IAS_LOG_ERROR("Writing the %s_SAT_Y_DEN_COEF parameter values", 
            band_string);
        return ERROR;
    }
    
    status = fprintf(output_file_ptr, ")\n  %s_SAT_Z_NUM_COEF = (", 
        band_string);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing the %s_SAT_Z_NUM_COEF parameter", band_string);
        return ERROR;
    } 

    /* Writing out the band satellite z numerator coefficients */
    if (write_single_parameter_array(output_file_ptr,
        cur_band->satellite.z_terms.numerator, IAS_ANGLE_GEN_ANG_RPC_COEF,
        13, 6, 27, 1) != SUCCESS)
    {
        IAS_LOG_ERROR("Writing the %s_SAT_Z_NUM_COEF parameter values", 
            band_string); 
        return ERROR;
    }

    status = fprintf(output_file_ptr, ")\n  %s_SAT_Z_DEN_COEF = (", 
        band_string);
    if (status < 0)
     {
        IAS_LOG_ERROR("Writing the %s_SAT_Z_DEN_COEF parameter", band_string);
        return ERROR;
    }

    /* Writing out the band satellite z denominator coefficients */
    if (write_single_parameter_array(output_file_ptr,
        cur_band->satellite.z_terms.denominator, IAS_ANGLE_GEN_ANG_RPC_COEF - 1,
        13, 6, 27, 1) != SUCCESS)
    {
        IAS_LOG_ERROR("Writing the %s_SAT_Z_DEN_COEF parameter values", 
            band_string);
        return ERROR;
    }

    status = fprintf(output_file_ptr, ")\n  %s_MEAN_SUN_VECTOR = "
        "(%12.9lf, %12.9lf, %12.9lf)\n"
        "  %s_SUN_X_NUM_COEF = (", band_string, cur_band->solar.mean_offset.x,
        cur_band->solar.mean_offset.y, cur_band->solar.mean_offset.z,
        band_string);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing the %s_MEAN_SUN_VECTOR parameter", band_string);
        return ERROR;
    } 

    /* Writing out the band solar x numerator coefficients */
    if (write_single_parameter_array(output_file_ptr,
        cur_band->solar.x_terms.numerator, IAS_ANGLE_GEN_ANG_RPC_COEF,
        13, 6, 27, 1) != SUCCESS)
    {
        IAS_LOG_ERROR("Writing the %s_SUN_X_NUM_COEF parameter values", 
            band_string);
        return ERROR;
    }

    status = fprintf(output_file_ptr, ")\n  %s_SUN_X_DEN_COEF = (", 
        band_string);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing the %s_SUN_X_DEN_COEF parameter", band_string);
        return ERROR;
    }

    /* Writing out the band solar x denominator coefficients */
    if (write_single_parameter_array(output_file_ptr,
        cur_band->solar.x_terms.denominator, IAS_ANGLE_GEN_ANG_RPC_COEF - 1,
        13, 6, 27, 1) != SUCCESS)
    {
        IAS_LOG_ERROR("Writing the %s_SUN_X_DEN_COEF parameter values", 
            band_string); 
        return ERROR;
    }

    status = fprintf(output_file_ptr, ")\n  %s_SUN_Y_NUM_COEF = (", 
        band_string);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing the %s_SUN_Y_NUM_COEF parameter", band_string);
        return ERROR;
    }

    /* Writing out the band solar y numerator coefficients */   
    if (write_single_parameter_array(output_file_ptr,
        cur_band->solar.y_terms.numerator, IAS_ANGLE_GEN_ANG_RPC_COEF,
        13, 6, 27, 1) != SUCCESS)
    {
        IAS_LOG_ERROR("Writing the %s_SUN_Y_NUM_COEF parameter values", 
            band_string);
        return ERROR;
    }

    status = fprintf(output_file_ptr, ")\n  %s_SUN_Y_DEN_COEF = (",
        band_string);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing the %s_SUN_Y_DEN_COEF parameter", band_string);
        return ERROR;
    }


    /* Writing out the band solar y denominator coefficients */
    if (write_single_parameter_array(output_file_ptr,
        cur_band->solar.y_terms.denominator, IAS_ANGLE_GEN_ANG_RPC_COEF - 1,
        13, 6, 27, 1) != SUCCESS)
    {
        IAS_LOG_ERROR("Writing the %s_SUN_Y_DEN_COEF parameter values", 
            band_string);
        return ERROR;
    }

    status = fprintf(output_file_ptr, ")\n  %s_SUN_Z_NUM_COEF = (", 
        band_string);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing the %s_SUN_Z_NUM_COEF parameter", band_string);
        return ERROR;
    }


    /* Writing out the band solar z numerator coefficients */
    if (write_single_parameter_array(output_file_ptr,
        cur_band->solar.z_terms.numerator, IAS_ANGLE_GEN_ANG_RPC_COEF,
        13, 6, 27, 1) != SUCCESS)
    {
        IAS_LOG_ERROR("Writing the %s_SUN_Z_NUM_COEF parameter values", 
            band_string);
        return ERROR;
    }

    status = fprintf(output_file_ptr, ")\n  %s_SUN_Z_DEN_COEF = (",
        band_string);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing the %s_SUN_Z_DEN_COEF parameter", band_string);
        return ERROR;
    }

    /* Writing out the band solar z denominator coefficients */
    if (write_single_parameter_array(output_file_ptr,
        cur_band->solar.z_terms.denominator, IAS_ANGLE_GEN_ANG_RPC_COEF - 1,
        13, 6, 27, 1) != SUCCESS)
    {
        IAS_LOG_ERROR("Writing the %s_SUN_Z_NUM_COEF parameter values", 
            band_string);
        return ERROR;
    }

    status = fprintf(output_file_ptr, ")\n  %s_SCA_LIST = (%d", band_string, 
        cur_band->sca_metadata[0].sca_number);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing the %s_SCA_LIST parameter", band_string);
        return ERROR;
    }

    /* Writing out the SCA list */
    for (sca_index = 1; sca_index < cur_band->num_scas; sca_index++)
    {
        status = fprintf(output_file_ptr, ", %d", 
            cur_band->sca_metadata[sca_index].sca_number);
        if (status < 0)
        {
            IAS_LOG_ERROR("Writing the %s_SCA_LIST parameter values", 
                band_string);
            return ERROR;
        }
    }

    status = fprintf(output_file_ptr, ")\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing the %s_SCA_LIST parameter closing", band_string);
        return ERROR;
    }

    /* SCA information per band */
    for (sca_index = 0; sca_index < cur_band->num_scas; sca_index++)
    {
        if (ias_angle_gen_write_ang_sca(band_string, 
            &cur_band->sca_metadata[sca_index], output_file_ptr) != SUCCESS)
        {
            IAS_LOG_ERROR("Writing the SCA metadata for SCA %d", 
                cur_band->sca_metadata[sca_index].sca_number);
            return ERROR;
        }        
    }

    status = fprintf(output_file_ptr, "END_GROUP = RPC_%s\n", band_string);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing the closing to the RPC_%s group", band_string); 
        return ERROR;
    }

    return SUCCESS;
}

/*******************************************************************************
Name: ias_angle_gen_write_ang

Purpose: Write angle coefficients metadata to file.

Returns: 
    Type = integer
    SUCCESS / ERROR
 ******************************************************************************/
int ias_angle_gen_write_ang
(
    const char *ang_filename,             /* I: ANG Filename */
    const IAS_ANGLE_GEN_METADATA *metadata/* I: Metadata struct to write from */
)
{
    FILE *output_file_ptr;  /* Output file pointer */
    int index;              /* Loop index */
    int status;             /* Status placeholder */
    char map_projection[4]; /* Local map projection string */
    double *ephemeris_array[EPHEMERIS_ARRAY_SIZE];
    int start_band;

    /* Create file */
    output_file_ptr = fopen(ang_filename, "w");
    if (!output_file_ptr)
    {
        IAS_LOG_ERROR("Opening the output file %s", ang_filename);
        return ERROR;
    }

    /* Determine the correct band to start at */
    for (start_band = 0; start_band < IAS_MAX_NBANDS; start_band++)
    {
        if (metadata->band_present[start_band]) 
            break;
    }

    /* Write data */
    /* File information group */
    status = fprintf(output_file_ptr, 
        "GROUP = FILE_HEADER\n"
        "  LANDSAT_SCENE_ID = \"%s\"\n"
        "  SPACECRAFT_ID = \"%s\"\n" 
        "  NUMBER_OF_BANDS = %d\n"
        "  BAND_LIST = (%d", metadata->landsat_scene_id, 
        metadata->spacecraft_id, metadata->num_bands, 
        metadata->band_metadata[start_band].band_number);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing FILE_HEADER group to file");
        fclose(output_file_ptr); 
        return ERROR;
    }

    /* Write out each band to file */
    for (index = start_band + 1; index < IAS_MAX_NBANDS; index++)
    {
        if (!metadata->band_present[index]) 
            continue;

        status = fprintf(output_file_ptr, ", %d", 
            metadata->band_metadata[index].band_number);
        if (status < 0)
        {
            IAS_LOG_ERROR("Write Band %d to file in FILE_HEADER group",
                metadata->band_metadata[index].band_number);
            fclose(output_file_ptr); 
            return ERROR;
        }
    }

    status = fprintf(output_file_ptr, ")\nEND_GROUP = FILE_HEADER\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing the closing of the FILE_HEADER group to file");
        fclose(output_file_ptr); 
        return ERROR;
    }

    /* Check the projection spheroid */
    if (metadata->projection.spheroid != WGS84_SPHEROID)
    {
        IAS_LOG_ERROR("Spheroid was not the expected WGS84 spheroid");
        fclose(output_file_ptr);
        return ERROR;
    }

    /* Projection information group */
    if (metadata->projection.proj_code == PS)
    {
        strcpy(map_projection, "PS");
    }
    else if (metadata->projection.proj_code == ALBERS)
    {
        strcpy(map_projection, "AEA");
    }
    else if (metadata->projection.proj_code == UTM)
    {
        strcpy(map_projection, "UTM");
    }
    else
    {
        IAS_LOG_ERROR("Determining PROJECTION CODE for PROJECTION group write");
        fclose(output_file_ptr); 
        return ERROR;
    }

    /* Write the beginning part of the projection group to the file */
    status = fprintf(output_file_ptr, 
        "GROUP = PROJECTION\n"
        "  ELLIPSOID_AXES = (%14.6lf, %14.6lf)\n"
        "  MAP_PROJECTION = \"%s\"\n"
        "  PROJECTION_UNITS = \"%s\"\n"    
        "  DATUM = \"%s\"\n"    
        "  ELLIPSOID = \"WGS84\"\n",
           metadata->wgs84_major_axis,
           metadata->wgs84_minor_axis, map_projection, metadata->units,
           metadata->datum);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing the PROJECTION group to file");
        fclose(output_file_ptr); 
        return ERROR;
    }

    /* Next write the UTM Zone */
    if (metadata->projection.proj_code == UTM)
    {
        status = fprintf(output_file_ptr, 
            "  UTM_ZONE = %d\n", metadata->projection.zone);
        if (status < 0)
        {
            IAS_LOG_ERROR("Writing the UTM_ZONE to file");
            fclose(output_file_ptr); 
            return ERROR;
        }
    }

    /* Now write the projection parameters */
    status = fprintf(output_file_ptr, "  PROJECTION_PARAMETERS = (");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing the Projection Parameters to file");
        fclose(output_file_ptr); 
        return ERROR;
    }
    status = write_single_parameter_array(output_file_ptr,  
        metadata->projection.parameters, IAS_PROJ_PARAM_SIZE, 5, 6, 27, 0);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Writing Projection Parameter %d to file in "
            "PROJECTION group", index);
        fclose(output_file_ptr); 
        return ERROR;
    }
    
    /* Corner information */
    status = fprintf(output_file_ptr, ")\n"
        "  UL_CORNER = (%11.3lf, %12.3lf)\n"
        "  UR_CORNER = (%11.3lf, %12.3lf)\n"
        "  LL_CORNER = (%11.3lf, %12.3lf)\n"
        "  LR_CORNER = (%11.3lf, %12.3lf)\n"    
        "END_GROUP = PROJECTION\n", metadata->corners.upleft.x,  
        metadata->corners.upleft.y , 
        metadata->corners.upright.x, 
        metadata->corners.upright.y, 
        metadata->corners.loleft.x,  
        metadata->corners.loleft.y , 
        metadata->corners.loright.x, 
        metadata->corners.loright.y);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing the closing to the PROJECTION group");
        fclose(output_file_ptr); 
        return ERROR;
    }

    /* Ephemeris data group */
    status = fprintf(output_file_ptr, 
        "GROUP = EPHEMERIS\n"
        "  EPHEMERIS_EPOCH_YEAR = %04d\n"
        "  EPHEMERIS_EPOCH_DAY = %03d\n"
        "  EPHEMERIS_EPOCH_SECONDS = %12.6lf\n"
        "  NUMBER_OF_POINTS = %d\n"       
        "  EPHEMERIS_TIME = (", 
        (int)floor(metadata->ephem_epoch_time.year + 0.5), 
        (int)floor(metadata->ephem_epoch_time.day + 0.5), 
        metadata->ephem_epoch_time.seconds, metadata->ephem_count);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing the EPHEMERIS group");
        fclose(output_file_ptr); 
        return ERROR;
    }

    /* Convert the ephemeris data to array */
    if (convert_ephemeris_to_array(metadata->ephemeris, metadata->ephem_count,
        ephemeris_array) != SUCCESS)
    {
        IAS_LOG_ERROR("Creating ephemeris array");
        fclose(output_file_ptr); 
        return ERROR;
    }
    
    /* Write out the ephemeris sample times */
    if (write_single_parameter_array(output_file_ptr, 
        ephemeris_array[SAMPLE_TIME_INDEX], metadata->ephem_count, 10, 6, 20, 0)
        != SUCCESS)
    {
        IAS_LOG_ERROR("Writing Projection Parameter %d to file in "
            "PROJECTION group", index);
        fclose(output_file_ptr);
        for (index = 0; index < EPHEMERIS_ARRAY_SIZE; index++)
            free(ephemeris_array[index]); 
        return ERROR;
    }

    /* ECEF coordinate information */
    status = fprintf(output_file_ptr, ")\n  EPHEMERIS_ECEF_X = (");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing EPHEMERIS_ECEF_X parameter to the "
            "EPHEMERIS group");
        fclose(output_file_ptr); 
        for (index = 0; index < EPHEMERIS_ARRAY_SIZE; index++)
            free(ephemeris_array[index]); 
        return ERROR;
    }

    /* Write out the ephemeris x position */
    if (write_single_parameter_array(output_file_ptr, 
        ephemeris_array[POSITION_X_INDEX], metadata->ephem_count, 15, 6, 22, 0)
        != SUCCESS)
    {
        IAS_LOG_ERROR("Writing EPHEMERIS_ECEF_X value at index %d to the "
            "EPHEMERIS group", index);
        fclose(output_file_ptr); 
        for (index = 0; index < EPHEMERIS_ARRAY_SIZE; index++)
            free(ephemeris_array[index]); 
        return ERROR;
    }

    status = fprintf(output_file_ptr, ")\n  EPHEMERIS_ECEF_Y = (");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing EPHEMERIS_ECEF_Y parameter to the "
            "EPHEMERIS group");
        fclose(output_file_ptr); 
        for (index = 0; index < EPHEMERIS_ARRAY_SIZE; index++)
            free(ephemeris_array[index]); 
        return ERROR;
    }

    /* Write out the ephemeris y position */
    if (write_single_parameter_array(output_file_ptr, 
        ephemeris_array[POSITION_Y_INDEX], metadata->ephem_count, 15, 6, 22, 0)
        != SUCCESS)
    {
        IAS_LOG_ERROR("Writing EPHEMERIS_ECEF_Y value at index %d to the "
            "EPHEMERIS group", index);
        fclose(output_file_ptr);
        for (index = 0; index < EPHEMERIS_ARRAY_SIZE; index++)
            free(ephemeris_array[index]);  
        return ERROR;
    }

    status = fprintf(output_file_ptr, ")\n  EPHEMERIS_ECEF_Z = (");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing EPHEMERIS_ECEF_Z parameter to the "
            "EPHEMERIS group");
        fclose(output_file_ptr);
        for (index = 0; index < EPHEMERIS_ARRAY_SIZE; index++)
            free(ephemeris_array[index]); 
        return ERROR;
    }
    
    /* Write out the ephemeris z position */
    if (write_single_parameter_array(output_file_ptr, 
        ephemeris_array[POSITION_Z_INDEX], metadata->ephem_count, 15, 6, 22, 0)
        != SUCCESS)
    {
        IAS_LOG_ERROR("Writing EPHEMERIS_ECEF_Z value at index %d to the "
            "EPHEMERIS group", index);
        fclose(output_file_ptr); 
        for (index = 0; index < EPHEMERIS_ARRAY_SIZE; index++)
            free(ephemeris_array[index]); 
        return ERROR;
    }

    /* Free up the emphemeris array */
    for (index = 0; index < EPHEMERIS_ARRAY_SIZE; index++)
    {
        free(ephemeris_array[index]);
        ephemeris_array[index] = NULL; 
    }

    status = fprintf(output_file_ptr, ")\nEND_GROUP = EPHEMERIS\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing the closing to the EPHEMERIS group");
        fclose(output_file_ptr); 
        return ERROR;
    }

    /* Solar vector data group */
    fprintf(output_file_ptr, "GROUP = SOLAR_VECTOR\n" 
        "  SOLAR_EPOCH_YEAR = %04d\n"
        "  SOLAR_EPOCH_DAY = %03d\n"
        "  SOLAR_EPOCH_SECONDS = %12.6lf\n"
        "  EARTH_SUN_DISTANCE = %10.8lf\n"
        "  NUMBER_OF_POINTS = %d\n"
        "  SAMPLE_TIME = (", 
        (int)floor(metadata->ephem_epoch_time.year + 0.5),
        (int)floor(metadata->ephem_epoch_time.day + 0.5), 
        metadata->ephem_epoch_time.seconds, metadata->earth_sun_distance,
        metadata->ephem_count);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing the SOLAR_VECTOR group");
        fclose(output_file_ptr); 
        return ERROR;
    }

    /* Convert the solar vector ephemeris struct to an array */
    if (convert_ephemeris_to_array(metadata->solar_vector,
        metadata->ephem_count, ephemeris_array) != SUCCESS)
    {
        IAS_LOG_ERROR("Creating ephemeris array");
        fclose(output_file_ptr); 
        return ERROR;
    }

    /* Write out the solar vector sample times */
    if (write_single_parameter_array(output_file_ptr, 
        ephemeris_array[SAMPLE_TIME_INDEX], metadata->ephem_count, 10, 6, 17, 0)
        != SUCCESS)
    {
        IAS_LOG_ERROR("Writing the SAMPLE_TIME value at index %d to the "
            "SOLAR_VECTOR group", index);
        fclose(output_file_ptr); 
        for (index = 0; index < EPHEMERIS_ARRAY_SIZE; index++)
            free(ephemeris_array[index]); 
        return ERROR;
    }

    /* Solar ECEF coordinate information */
    status = fprintf(output_file_ptr, ")\n  SOLAR_ECEF_X = (");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing SOLAR_ECEF_X parameter to the SOLAR_VECTOR "
            "group");
        fclose(output_file_ptr); 
        for (index = 0; index < EPHEMERIS_ARRAY_SIZE; index++)
            free(ephemeris_array[index]); 
        return ERROR;
    }

    /* Writing out the solar ECEF x coordinates */
    if (write_single_parameter_array(output_file_ptr, 
        ephemeris_array[POSITION_X_INDEX], metadata->ephem_count, 15, 8, 18, 1) 
        != SUCCESS)
    {
        IAS_LOG_ERROR("Writing the SOLAR_ECEF_X value at index %d to the "
            "SOLAR_VECTOR group", index);
        fclose(output_file_ptr); 
        for (index = 0; index < EPHEMERIS_ARRAY_SIZE; index++)
            free(ephemeris_array[index]); 
        return ERROR;
    }

    status = fprintf(output_file_ptr, ")\n  SOLAR_ECEF_Y = (");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing SOLAR_ECEF_Y parameter to the SOLAR_VECTOR "
            "group");
        fclose(output_file_ptr); 
        for (index = 0; index < EPHEMERIS_ARRAY_SIZE; index++)
            free(ephemeris_array[index]); 
        return ERROR;
    }

    /* Writing out the solar ECEF y coordinates */
    if (write_single_parameter_array(output_file_ptr, 
        ephemeris_array[POSITION_Y_INDEX], metadata->ephem_count, 15, 8, 18, 1)
        != SUCCESS)
    {
        IAS_LOG_ERROR("Writing the SOLAR_ECEF_Y value at index %d to the "
            "SOLAR_VECTOR group", index);
        fclose(output_file_ptr);
        for (index = 0; index < EPHEMERIS_ARRAY_SIZE; index++)
            free(ephemeris_array[index]);  
        return ERROR;
    }

    status = fprintf(output_file_ptr, ")\n  SOLAR_ECEF_Z = (");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing SOLAR_ECEF_Z parameter to the SOLAR_VECTOR "
            "group");
        fclose(output_file_ptr);
        for (index = 0; index < EPHEMERIS_ARRAY_SIZE; index++)
            free(ephemeris_array[index]);  
        return ERROR;
    }

    /* Writing out the solar ECEF z coordinates */
    if (write_single_parameter_array(output_file_ptr, 
        ephemeris_array[POSITION_Z_INDEX], metadata->ephem_count, 15, 8, 18, 1)
        != SUCCESS)
    {
        IAS_LOG_ERROR("Writing the SOLAR_ECEF_Z value at index %d to the "
            "SOLAR_VECTOR group", index);
        fclose(output_file_ptr); 
        for (index = 0; index < EPHEMERIS_ARRAY_SIZE; index++)
            free(ephemeris_array[index]); 
        return ERROR;
    }

    /* Free up the emphemeris array */
    for (index = 0; index < EPHEMERIS_ARRAY_SIZE; index++)
    {
        free(ephemeris_array[index]);
        ephemeris_array[index] = NULL; 
    }

    fprintf(output_file_ptr, ")\nEND_GROUP = SOLAR_VECTOR\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing the closing parameter to the SOLAR_VECTOR "
            "group");
        fclose(output_file_ptr); 
        return ERROR;
    }

    /* Band information */
    for (index = 0; index < IAS_MAX_NBANDS; index++)
    {
        if (!metadata->band_present[index]) 
            continue;

        if (ias_angle_gen_write_ang_band(&metadata->band_metadata[index],
            output_file_ptr) != SUCCESS)
        {
            IAS_LOG_ERROR("Writing the band metadata for band number %d", 
                metadata->band_metadata[index].band_number);
            fclose(output_file_ptr); 
            return ERROR;
        }
    }

    status = fprintf(output_file_ptr, "END\n");
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing the END group line");
        fclose(output_file_ptr); 
        return ERROR;     
    }

    /* Close file */
    if (fclose(output_file_ptr) != 0)
    {
        IAS_LOG_ERROR("Closing the ANG file");
        return ERROR;
    }

    return(SUCCESS);
}
