/* Standard Library Includes */
#include <limits.h>
#include <string.h>

/* IAS Library includes */
#include "ias_logging.h"
#include "ias_const.h"
#include "ias_grid_io.h" 
#include "ias_model_io.h"
#include "ias_geo.h"

/* Local Includes */
#include "cgreen/cgreen.h"
#include "../ias_angle_gen.h"
#include "../ias_angle_gen_private.h"

/* Local Variables */
static IAS_ANGLE_GEN_METADATA metadata;
static IAS_GRID_BAND_TYPE blank_gridband;
static IAS_GRID_TYPE grid;
static IAS_LOS_MODEL *model;    
static IAS_CPF *cpf;     
static const struct IAS_CPF_EARTH_CONSTANTS *cpf_earth; 

int ias_angle_gen_l1t_to_l1r_rpc
(
    const IAS_GRID_BAND_TYPE *gridband, /* I: Input grid for current band/SCA */
    IAS_ANGLE_GEN_IMAGE_RPC *coefficients /* I/O: Image coefficients */
);

static void setup_angle_test
(
)
{
    char *metadata_filename = "./verify/LC81420402013121LGN00_ANG.txt";
    char model_filename[PATH_MAX];
    char cpf_filename[PATH_MAX];
    char grid_filename[PATH_MAX];
    int status;

    /* Initialize the blank metadata to 0 */
    memset(&blank_gridband, 0, sizeof(IAS_GRID_BAND_TYPE));

    /* Setup the model, cpf, and grid filenames */
    status = snprintf(grid_filename, sizeof(grid_filename),
        "%s/ias_lib/misc/angle_generation/grid.h5",
        getenv("IAS_UNIT_TEST_DATA_DIR"));
    if (status < 0 || status > sizeof(grid_filename))
    {
        IAS_LOG_ERROR("Creating the grid filename");
        exit (EXIT_FAILURE);
    }

    status = snprintf(cpf_filename, sizeof(cpf_filename),
        "%s/ias_lib/misc/angle_generation/L8CPF20130401_20130630.06",
        getenv("IAS_UNIT_TEST_DATA_DIR"));
    if (status < 0 || status > sizeof(cpf_filename))
    {
        IAS_LOG_ERROR("Creating the cpf filename");
        exit (EXIT_FAILURE);
    }

    status = snprintf(model_filename, sizeof(model_filename),
        "%s/ias_lib/misc/angle_generation/precision_model.h5",
        getenv("IAS_UNIT_TEST_DATA_DIR"));
    if (status < 0 || status > sizeof(model_filename))
    {
        IAS_LOG_ERROR("Creating the model filename");
        exit (EXIT_FAILURE);
    }

    if (ias_sat_attr_initialize(IAS_L8) != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing satellite attributes library");
        exit(EXIT_FAILURE);
    }

    /* Read the enhanced metadata file */
    if (ias_angle_gen_read_ang(metadata_filename, &metadata) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading the enhanced metadata file %s.",
            metadata_filename);
        exit (EXIT_FAILURE);
    }
    
    if (ias_angle_gen_initialize_transformation(&metadata) != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing the transformation");
        ias_angle_gen_free(&metadata);
        ias_angle_gen_free_transformation(&metadata);
        exit (EXIT_FAILURE);
    }

    /* Read the full grid structure */
    if (ias_grid_read(grid_filename, NULL, 0, &grid) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading the grid file %s.", grid_filename);
        ias_angle_gen_free(&metadata);
        ias_angle_gen_free_transformation(&metadata);
        exit (EXIT_FAILURE);
    }

    /* Read the full model structure */
    model = ias_model_read(model_filename);
    if (!model)
    {
        IAS_LOG_ERROR ("Reading the model file %s.", model_filename);
        ias_grid_free (&grid);
        ias_angle_gen_free(&metadata);
        ias_angle_gen_free_transformation(&metadata);
        exit (EXIT_FAILURE);
    }

    /* Read the CPF */
    cpf = ias_cpf_read(cpf_filename);
    if (!cpf)
    {
        IAS_LOG_ERROR("Cannot read CPF file %s", cpf_filename);
        ias_grid_free (&grid);
        ias_los_model_free(model);
        ias_angle_gen_free(&metadata);
        ias_angle_gen_free_transformation(&metadata);
        exit (EXIT_FAILURE);
    }

    /* Get the needed group from the CPF. */
    cpf_earth = ias_cpf_get_earth_const(cpf);
    if (!cpf_earth)
    {
        IAS_LOG_ERROR("Retrieving Earth constants group from CPF file %s",
            cpf_filename);
        ias_angle_gen_free(&metadata);
        ias_angle_gen_free_transformation(&metadata);
        ias_cpf_free(cpf);
        ias_grid_free(&grid);
        ias_los_model_free(model);
        exit(EXIT_FAILURE);
    }
}

static void teardown_angle_test
(
)
{
    /* Free the ephemeris structure & transformation */
    ias_angle_gen_free(&metadata);
    ias_angle_gen_free_transformation(&metadata);

    /* Free the grid */
    ias_grid_free(&grid);

    /* Free the CPF */
    ias_cpf_free(cpf);

    /* Free the model */
    ias_los_model_free(model);
}

/* Start of the unit tests */
Ensure test_calc_angle_rpc
(
)
{
    int index;          /* Loop variable */ 
    int status;         /* Status placeholder */ 
    int val_index = 0;  /* Validation array index */
    double satang[2];   /* Satellite zenith and azimuth angles */    
    double sunang[2];   /* Solar zenith and azimuth angles */
    int outside_flag;

    /* Validation array of Sun / Sat values */
    double validation[20] = {
        .103693090, 1.69438485, .429203417, 2.10512764,
        .103636195, 1.69434908, .429196747, 2.10512747,
        .103579300, 1.69431327, .429190077, 2.10512731,
        .103522405, 1.69427740, .429183406, 2.10512715,
        .103465509, 1.69424150, .429176736, 2.10512698 };

    setup_angle_test();

    /* Loop through 5 line / sample combinations and calculate the angles */
    for (index = 1995; index < 2000; index++)
    {   
        /* Calculate the satellite angle */
        status = ias_angle_gen_calculate_angles_rpc(&metadata, index, index, 
            NULL, 5, IAS_ANGLE_GEN_SATELLITE, &outside_flag, satang);
        assert_equal(status, SUCCESS);
        assert_equal(outside_flag, 0);

        /* Calculate the sun angle */
        status = ias_angle_gen_calculate_angles_rpc(&metadata, index, index, 
            NULL, 5, IAS_ANGLE_GEN_SOLAR, &outside_flag, sunang);
        assert_equal(status, SUCCESS);
        assert_equal(outside_flag, 0);

        /* Verify both the sun and satellite angles */
        assert_double_equal(satang[0], validation[val_index++]);
        assert_double_equal(satang[1], validation[val_index++]);
        assert_double_equal(sunang[0], validation[val_index++]);
        assert_double_equal(sunang[1], validation[val_index++]);
    }

    /* Check that the outside flag is set for points outside image */
    status = ias_angle_gen_calculate_angles_rpc(&metadata, 0, 0, 
        NULL, 5, IAS_ANGLE_GEN_SOLAR, &outside_flag, sunang);
    assert_equal(status, SUCCESS);
    assert_equal(outside_flag, 1);

    /* Successful tests have been made, Now do the error checking */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* Check that the invalid band index returns and error */
    status = ias_angle_gen_calculate_angles_rpc(&metadata, index, index,
        NULL, 16, IAS_ANGLE_GEN_SOLAR, &outside_flag, sunang);
    assert_equal(status, ERROR);

    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

    teardown_angle_test();
}

Ensure test_calc_angle_rigor
(
)
{
    int index;          /* Loop variable */
    int status;         /* Status placeholder */
    int val_index = 0;  /* Validation array index */
    double satang[2];   /* Satellite zenith and azimuth angles */    
    double sunang[2];   /* Solar zenith and azimuth angles */
    int outside_flag;

    /* Validation array of Sun / Sat values */
    double validation[20] = {
        0.103701905127, 1.694353850915, 0.429203162126, 2.105126516457,
        0.103645012150, 1.694318167819, 0.429196491706, 2.105126353171,
        0.103588118752, 1.694282439896, 0.429189821288, 2.105126189749,
        0.103531224934, 1.694246667069, 0.429183150874, 2.105126026189,
        0.103474330696, 1.694210849267, 0.429176480463, 2.105125862492};

    setup_angle_test();

    /* Loop through 5 line / sample combinations and calculate the angles */
    for (index = 1995; index < 2000; index++)
    {   
        /* Calculate the satellite angle */
        status = ias_angle_gen_calculate_angles_rigor(&metadata, index, index,
            NULL, 5, IAS_ANGLE_GEN_SATELLITE, &outside_flag, satang);
        assert_equal(status, SUCCESS);
        assert_equal(outside_flag, 0);

        /* Calculate the sun angle */
        status = ias_angle_gen_calculate_angles_rigor(&metadata, index, index, 
            NULL, 5, IAS_ANGLE_GEN_SOLAR, &outside_flag, sunang);
        assert_equal(status, SUCCESS);
        assert_equal(outside_flag, 0);

        /* Verify both the sun and satellite angles */
        assert_double_equal(satang[0], validation[val_index++]);
        assert_double_equal(satang[1], validation[val_index++]);
        assert_double_equal(sunang[0], validation[val_index++]);
        assert_double_equal(sunang[1], validation[val_index++]);
    }

    /* Check that the outside flag is set for points outside image */
    status = ias_angle_gen_calculate_angles_rigor(&metadata, 0, 0, 
        NULL, 5, IAS_ANGLE_GEN_SOLAR, &outside_flag, sunang);
    assert_equal(status, SUCCESS);
    assert_equal(outside_flag, 1);

    /* Successful tests have been made, Now do the error checking */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    IAS_GEO_PROJ_TRANSFORMATION *temp_trans = metadata.transformation;
    metadata.transformation = NULL;

    /* Check that calculate vector returns an error */
    status = ias_angle_gen_calculate_angles_rigor(&metadata, index, index, 
        NULL, 5, IAS_ANGLE_GEN_SOLAR, &outside_flag, sunang);
    assert_equal(status, ERROR);
    metadata.transformation = temp_trans;

    /* Check that the invalid band index returns and error */
    status = ias_angle_gen_calculate_angles_rigor(&metadata, index, index,
        NULL, 16, IAS_ANGLE_GEN_SOLAR, &outside_flag, sunang);
    assert_equal(status, ERROR);

    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

    teardown_angle_test();
}

Ensure test_calc_vector
(
)
{
    int index;          /* Loop variable */ 
    int status;         /* Status placeholder */ 
    int val_index = 0;  /* Validation array index */
    IAS_VECTOR satvec;  /* Satellite vector */
    IAS_VECTOR sunvec;  /* Solar vector */
    double height;      /* Mean height */

    /* Validation array of Sun / Sat values */
    double validation[30] = {
        0.101943794931, -0.016322190929, 0.994656246529,
        0.358117580671, -0.211932970656, 0.909305347154,
        0.101885940022, -0.016321286199, 0.994662189310,
        0.358112334975, -0.211929834930, 0.909308143923,
        0.101828084165, -0.016320381060, 0.994668128793,
        0.358107089292, -0.211926699148, 0.909310940650,
        0.101770227359, -0.016319475511, 0.994674064979,
        0.358101843623, -0.211923563312, 0.909313737336,
        0.101712369605, -0.016318569552, 0.994679997867,
        0.358096597967, -0.211920427420, 0.909316533979
    };

    setup_angle_test();

    height = metadata.band_metadata[1].satellite.mean_height;

    /* Loop through 5 line / sample combinations and calculate the vectors */
    for (index = 1995; index < 2000; index++)
    {   
        /* Calculate the satellite vector */
        status = ias_angle_gen_calculate_vector(&metadata, index, index, index, 
            height, 0, IAS_ANGLE_GEN_SATELLITE, &satvec);
        assert_equal(status, SUCCESS);

        /* Calculate the sun vector */
        status = ias_angle_gen_calculate_vector(&metadata, index, index, index, 
            height, 0, IAS_ANGLE_GEN_SOLAR, &sunvec);
        assert_equal(status, SUCCESS);

        /* Validate both of the vectors */
        assert_double_equal(satvec.x,  validation[val_index++]);
        assert_double_equal(satvec.y,  validation[val_index++]);
        assert_double_equal(satvec.z,  validation[val_index++]);
        assert_double_equal(sunvec.x,  validation[val_index++]);
        assert_double_equal(sunvec.y,  validation[val_index++]);
        assert_double_equal(sunvec.z,  validation[val_index++]);
    }

    /* Successful tests have been made, Now do the error checking */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    IAS_GEO_PROJ_TRANSFORMATION *temp_trans = metadata.transformation;
    metadata.transformation = NULL;

    /* Check that ias_geo_transform_coordinate throws error */
    status = ias_angle_gen_calculate_vector(&metadata, -9000, -9000, -9000,
        height, 0, IAS_ANGLE_GEN_SATELLITE, &satvec);
    assert_equal(status, ERROR);
    metadata.transformation = temp_trans;

    /* Check that interpolation returns error */
    int points = metadata.ephem_count;
    metadata.ephem_count = -1;
    status = ias_angle_gen_calculate_vector(&metadata, -9000, -9000, -9000,
        height, 0, IAS_ANGLE_GEN_SATELLITE, &satvec);
    assert_equal(status, ERROR);
    metadata.ephem_count = points;

    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

    teardown_angle_test();
}

Ensure test_find_scas
(
)
{
    double l1r_line[2] = {0};   /* Input space (L1R) line coordinate */
    double l1r_samp[2] = {0};   /* Input space (L1R) sample coordinate */
    int index;                  /* Loop variable */ 
    int status;                 /* Status placeholder */ 
    int val_index = 0;          /* Validation array index */
    double height = 7000;       /* Height to search scas */
    
    /* Validation array for scas with no height */
    double validation[30] = {
        1944.814516, 0, 1038.596124, 0, 1945.596006, 0, 1039.853236, 0,
        1946.377491, 0, 1041.110367, 0, 1947.158971, 0, 1042.367518, 0,
        1947.940447, 0, 1043.624687, 0 };

    /* Validation array for scas with height */
    double validation_height[30] = {
        1945.426153, 0, 1027.813782, 0, 1946.207648, 0, 1029.076730, 0,
        1946.989139, 0, 1030.339697, 0, 1947.770625, 0, 1031.602683, 0,
        1948.552106, 0, 1032.865688, 0 };

    setup_angle_test();

    /* Loop through 5 line/sample combinations and find the scas */
    for (index = 1995; index < 2000; index++)
    {   
        /* Find the scas using no height */
        status = ias_angle_gen_find_scas(&metadata.band_metadata[6],
            index, index, NULL, l1r_line, l1r_samp);      
        assert_equal(status, 1);

        assert_double_equal(l1r_line[0], validation[val_index++]);
        assert_double_equal(l1r_line[1], validation[val_index++]);
        assert_double_equal(l1r_samp[0], validation[val_index++]);
        assert_double_equal(l1r_samp[1], validation[val_index++]);
    }

    /* Loop through 5 line/sample combinations and find the scas with height */
    for (index = 1995, val_index = 0; index < 2000; index++)
    {   
        /* Find the scas using height */
        status = ias_angle_gen_find_scas(&metadata.band_metadata[6],
            index, index, &height, l1r_line, l1r_samp);        
        assert_equal(status, 1);

        assert_double_equal(l1r_line[0], validation_height[val_index++]);
        assert_double_equal(l1r_line[1], validation_height[val_index++]);
        assert_double_equal(l1r_samp[0], validation_height[val_index++]);
        assert_double_equal(l1r_samp[1], validation_height[val_index++]);
    }

    teardown_angle_test();
}

Ensure test_map_l1t_to_l1r_rpc
(
)
{
    IAS_ANGLE_GEN_IMAGE_RPC rpc;  /* Coefficients structure */
    int status;                   /* Status placeholder */

    setup_angle_test();

    /* Convert the gridband data from l1t to l1r data and store it in the
     * structure */
    status = ias_angle_gen_l1t_to_l1r_rpc(&grid.gridbands[1][1], &rpc);
    assert_equal(status, SUCCESS);

    /* Validate all of the data */
    assert_double_equal(rpc.line_terms.l1r_mean_offset, 3508.923404);
    assert_double_equal(rpc.samp_terms.l1r_mean_offset, 257.214286);

    assert_double_equal(rpc.line_terms.l1t_mean_offset, 3026.788139);
    assert_double_equal(rpc.samp_terms.l1t_mean_offset, 1473.574593);

    assert_double_equal(rpc.mean_height, 4000);

    assert_double_equal(rpc.line_terms.numerator[0], .198743);
    assert_double_equal(rpc.line_terms.numerator[1], 1.01061227);
    assert_double_equal(rpc.line_terms.numerator[2], -.228447);
    assert_double_equal(rpc.line_terms.numerator[3], -.000436808891);

    assert_double_equal(rpc.line_terms.denominator[0], .000000149326981);
    assert_double_equal(rpc.line_terms.denominator[1], -.0000056557594);    
    assert_double_equal(rpc.line_terms.denominator[2], .0000000000294018098);   
    assert_double_equal(rpc.line_terms.denominator[3], -.000000000000446123830);

    assert_double_equal(rpc.samp_terms.numerator[0], -.361808750);
    assert_double_equal(rpc.samp_terms.numerator[1], .252871539);
    assert_double_equal(rpc.samp_terms.numerator[2], 1.003947);
    assert_double_equal(rpc.samp_terms.numerator[3], -.00399278724);

    assert_double_equal(rpc.samp_terms.denominator[0], -.00000182047293);
    assert_double_equal(rpc.samp_terms.denominator[1], -.00000651504190);    
    assert_double_equal(rpc.samp_terms.denominator[2], -.00000155903160);
    assert_double_equal(rpc.samp_terms.denominator[3], -.00000000000194965878);

    /* Successful tests have been made, Now do the error checking */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    
    /* Check that the blank grid throws an error */
    status = ias_angle_gen_l1t_to_l1r_rpc(&blank_gridband, &rpc);
    assert_equal(status, ERROR);

    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

    teardown_angle_test();
}

Ensure test_calc_band_rpc
(
)
{
    int status;           /* Status placeholder */
    int band_index = 4;   /* Current band number */
    IAS_ANGLE_GEN_BAND *band_ptr = &metadata.band_metadata[band_index];

    setup_angle_test();

    /* Calculate the coefficients per band */
    status = ias_angle_gen_calculate_band_rpc(grid.gridbands[band_index], 
        band_index, &metadata);
    assert_equal(status, SUCCESS);

    /* Validate the Satellite means */
    assert_double_equal(band_ptr->satellite.line_terms.l1t_mean_offset, 
        3734.209668);
    assert_double_equal(band_ptr->satellite.samp_terms.l1t_mean_offset, 
        3838.422710);
    assert_double_equal(band_ptr->satellite.mean_height, 4000.00);
    assert_double_equal(band_ptr->satellite.line_terms.l1r_mean_offset, 
        3508.923404);
    assert_double_equal(band_ptr->satellite.samp_terms.l1r_mean_offset, 
        3468.214286);
    assert_double_equal(band_ptr->satellite.mean_offset.x, -0.002244867467);
    assert_double_equal(band_ptr->satellite.mean_offset.y, -0.003800223202);
    assert_double_equal(band_ptr->satellite.mean_offset.z, 0.996292349998);

    /* Validate the Sun means */
    assert_double_equal(band_ptr->solar.line_terms.l1t_mean_offset, 
        3734.209668);
    assert_double_equal(band_ptr->solar.samp_terms.l1t_mean_offset, 
        3838.422710);
    assert_double_equal(band_ptr->solar.mean_height, 4000.00);
    assert_double_equal(band_ptr->solar.line_terms.l1r_mean_offset, 
        3508.923404);
    assert_double_equal(band_ptr->solar.samp_terms.l1r_mean_offset, 
        3468.214286);
    assert_double_equal(band_ptr->solar.mean_offset.x, 0.348555430888);
    assert_double_equal(band_ptr->solar.mean_offset.y, -0.206441707459);
    assert_double_equal(band_ptr->solar.mean_offset.z, 0.914182087219); 

    /* Validate the numerators and denominators */
    assert_double_equal(band_ptr->satellite.x_terms.numerator[0], 
        -0.000072994195914);
    assert_double_equal(band_ptr->satellite.x_terms.denominator[0], 
        0.000003784893460);
    assert_double_equal(band_ptr->solar.x_terms.numerator[0], 0.00001805833211);
    assert_double_equal(band_ptr->solar.x_terms.denominator[0], 
        0.000000113520152);

    assert_double_equal(band_ptr->satellite.y_terms.numerator[0], 
        -0.000399864024);
    assert_double_equal(band_ptr->satellite.y_terms.denominator[0], 
        -0.000000498931);
    assert_double_equal(band_ptr->solar.y_terms.numerator[0], -0.0000481694276);
    assert_double_equal(band_ptr->solar.y_terms.denominator[0], 0.000000000000);

    assert_double_equal(band_ptr->satellite.z_terms.numerator[0], 
        0.003664392511);
    assert_double_equal(band_ptr->satellite.z_terms.denominator[0], 
        -0.000004065133);
    assert_double_equal(band_ptr->solar.z_terms.numerator[0], 0.00007089947763);
    assert_double_equal(band_ptr->solar.z_terms.denominator[0], 
        -0.000009592974);

    /* Successful tests have been made, Now do the error checking */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    IAS_GEO_PROJ_TRANSFORMATION *temp_trans = metadata.transformation;
    metadata.transformation = NULL;

    /* Check that the invalid band index throws an error */
    status = ias_angle_gen_calculate_band_rpc(grid.gridbands[band_index], 
        13, &metadata);
    assert_equal(status, ERROR);

    /* Check that calculate vector returns with an error */
    status = ias_angle_gen_calculate_band_rpc(grid.gridbands[band_index], 
        band_index, &metadata);
    assert_equal(status, ERROR);

    metadata.transformation = temp_trans;
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

    teardown_angle_test();
}

Ensure test_interpolate_ephem
(
)
{
    int index;              /* Loop variable */ 
    int status;             /* Status placeholder */ 
    int val_index = 0;      /* Validation array index */
    IAS_VECTOR interpol;    /* Interpolated vector */ 

    double validation[15] = {
        580858.285727304756, 6079671.963714634068, 3579581.930633046664,
        581367.499910062528, 6080565.404811437242, 3577986.713977500331,
        581876.700091789244, 6081458.402702523395, 3576391.244148453698,
        582385.887790999957, 6082350.957778999582, 3574795.521904000081,
        582895.064526210772, 6083243.070431975648, 3573199.548002233729};

    double increase = .25; /* Time increase per loop */
    double sample_increase;

    setup_angle_test();

    sample_increase = metadata.ephemeris->sample_time + increase;

    /* Loop through 5 line/sample combos interpolating the ephemeris */
    for (index = 1995; index < 2000; index++)
    {  
        /* Interpolate the ephemeris */
        status = ias_angle_gen_interpolate_ephemeris(metadata.ephemeris, 
            metadata.ephem_count, sample_increase, &interpol);
        assert_equal(status, SUCCESS);

        /* Increase the time */
        sample_increase += increase;
        
        /* Validate the interpolated vector */
        assert_double_equal(interpol.x, validation[val_index++]);
        assert_double_equal(interpol.y, validation[val_index++]);
        assert_double_equal(interpol.z, validation[val_index++]);
    }

    /* Successful tests have been made, Now do the error checking */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    
    /* Check that interpolation fails if the number of points pass is 
       less than 1 */
    status = ias_angle_gen_interpolate_ephemeris(metadata.ephemeris, 
        0, sample_increase, &interpol);
    assert_equal(status, ERROR);

    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

    teardown_angle_test();
}

Ensure test_initialize_data
(
)
{
    IAS_ANGLE_GEN_METADATA data;
    int status;

    setup_angle_test();

    if (ias_math_init_leap_seconds_from_UTC_time(
        model->spacecraft.ephemeris.utc_epoch_time, 
        &cpf_earth->leap_seconds_data) != SUCCESS)
    {
        assert_equal(SUCCESS, ERROR);
        teardown_angle_test();
        return;
    }

    /* Connect to NOVAS for sun angle computation */
    if (ias_geo_initialize_novas() != SUCCESS)
    {
        assert_equal(SUCCESS, ERROR);
        teardown_angle_test();
        return;
    }

    status = ias_angle_gen_populate_metadata(cpf_earth->semi_major_axis, 
        cpf_earth->semi_minor_axis, &grid, model, &data);
    assert_equal(status, SUCCESS);
    
    /* Disconnect from NOVAS */
    if (ias_geo_shutdown_novas() != SUCCESS)
    {
        assert_equal(SUCCESS, ERROR);
        ias_angle_gen_free(&data);
        teardown_angle_test();
        return;
    }
    
    int band_index = 0;

    strcpy(data.landsat_scene_id, "LC81420402013121LGN00");
    strcpy(data.spacecraft_id, "LANDSAT_8");

    if (ias_angle_gen_initialize_transformation(&data) != SUCCESS)
    {
        assert_true_with_message(0, "Initializing the transformation");
        ias_angle_gen_free_transformation(&data);
        ias_angle_gen_free(&data);
        teardown_angle_test();
        return;
    }

    /* Loop through the bands looking for one to assign */
    while (band_index < 1)
    {
        status = ias_angle_gen_get_band_data(band_index, &grid, model, &data);
        assert_equal(status, SUCCESS);

        IAS_LOG_INFO("PROCESSING BAND %d", band_index + 1);
        /* Only consider bands that are available */
        if (grid.bands_present[band_index])
        {
            status = ias_angle_gen_calculate_band_rpc(grid.gridbands[
                band_index], band_index, &data);
            assert_equal(status, SUCCESS);
        }
        band_index++;
    }

    /* Write out the retrieved data */
    status = ias_angle_gen_write_ang("per_band_calc.txt", &data);
    assert_equal(status, SUCCESS);

    /* Verify the retrieved data */
    status = system("diff -q ./verify/per_band_calc.verify per_band_calc.txt");
    assert_equal(status, SUCCESS);

    ias_angle_gen_free_transformation(&data);
    ias_angle_gen_free(&data);

    /* Successful tests have been made, Now do the error checking */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    
    /* Check that passing NULL grid and model return ERRORS */
    status = ias_angle_gen_populate_metadata(cpf_earth->semi_major_axis, 
        cpf_earth->semi_minor_axis, &grid, NULL, &data);
    assert_equal(status, ERROR);

    status = ias_angle_gen_populate_metadata(cpf_earth->semi_major_axis, 
        cpf_earth->semi_minor_axis, NULL, model, &data);
    assert_equal(status, ERROR);

    /* Check that grid is NULL */
    status = ias_angle_gen_get_band_data(band_index, NULL, model, &data);
    assert_equal(status, ERROR);

    /* Check that model is NULL */
    status = ias_angle_gen_get_band_data(band_index, &grid, NULL, &data);
    assert_equal(status, ERROR);

    /* Bad satellite band conversion */
    status = ias_angle_gen_get_band_data(31, &grid, model, &data);
    assert_equal(status, ERROR);

    /* Bad band index */
    status = ias_angle_gen_get_band_data(12, &grid, model, &data);
    assert_equal(status, ERROR);

    /* Band not present in the grid */
    grid.bands_present[band_index] = 0;

    status = ias_angle_gen_get_band_data(band_index, &grid, model, &data);
    assert_equal(status, ERROR);

    grid.bands_present[band_index] = 1;
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

    teardown_angle_test();
}

int main(int argc, char **argv)
{
    int status; /* Status placeholder */
    ias_log_initialize("Angle Test");

    TestSuite *suite = create_test_suite();
    add_test(suite, test_map_l1t_to_l1r_rpc);
    add_test(suite, test_calc_vector);
    add_test(suite, test_calc_angle_rpc);
    add_test(suite, test_calc_angle_rigor);
    add_test(suite, test_find_scas);
    add_test(suite, test_calc_band_rpc);
    add_test(suite, test_interpolate_ephem);
    add_test(suite, test_initialize_data);

    if (argc > 1) {
        status = run_single_test(suite, argv[1], create_text_reporter());
        return status;
    }
    status = run_test_suite(suite, create_text_reporter());
    return status;
}
