/*************************************************************************

NAME: database_test.c

PURPOSE: Using Cgreen for testing database library routines 

Algorithm References: None

NOTES:
    1. Note this set of tests uses the connection returned by 
       ias_db_get_connect_info.  Set your environment variables appropriately 
       to point to the actual database you want to use.


**************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <time.h>
#include <sys/timeb.h>

#include "ias_logging.h"
#include "ias_db_get_connect_info.h"
#include "ias_db.h"
#include "ias_const.h"
#include "ias_satellite_attributes.h"
#include "cgreen/cgreen.h"

/* Global variables used by tests, and teardown */
struct ias_db_connection *conn = NULL;


static void initialize()
{
    ias_db_initialize_database_lib();
    ias_log_initialize("database_test");
}

static void cleanup() 
{
    if (conn) ias_db_close_connection(conn);
    ias_db_close_database_lib();
}

Ensure test_get_login_for_satellite_sensor()
{
    int status;
    char *schema;
    char *password;
    int log_level;

    /* Store info in an environment variable */
    status = setenv("IAS_DB_L20_HYPERSENSOR", "myschema/thispw", 1);
    assert_equal(status, 0);
    if (status != 0)
        return;

    /* Get the info from the newly created environment variable */
    status = ias_db_get_login_for_satellite_sensor("L20", "HYPERSENSOR",
        &schema, &password);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
        return;

    /* Verify the values */
    assert_string_equal(schema, "myschema");
    assert_string_equal(password, "thispw");
    free(schema);
    schema = NULL;
    free(password);
    password = NULL;

    /* Verify an error is detected with an environment variable that doesn't 
       exist */
    log_level = ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = ias_db_get_login_for_satellite_sensor("L2000", "ICANTEXIST",
        &schema, &password);
    assert_equal(status, ERROR);
    ias_log_set_output_level(log_level);
}

Ensure test_connection_info()
{
    char *db = NULL;
    char *user = NULL;
    char *pwd = NULL;
    char *host = NULL;
    int status;
    
    status = ias_db_get_connect_info(&db,&user,&pwd,&host);
    assert_equal(status, SUCCESS);    
    assert_true(db != NULL);
    assert_true(user != NULL);
    assert_true(pwd != NULL);
    assert_true(host == NULL);
    free(db);
    free(user);
    free(pwd);
    free(host);
}

Ensure test_common_connection_info()
{
    char *db = NULL;
    char *user = NULL;
    char *pwd = NULL;
    char *host = NULL;
    int status;

    /* Prepare the environment variables */
    assert_equal(0, setenv("IAS_DB_COM", "user_test/pass_test", 1));
    assert_equal(0, setenv("TWO_TASK", "two_task_db_test", 1));

    status = ias_db_get_common_connect_info(&db,&user,&pwd,&host);
    assert_equal(status, SUCCESS);
    assert_true(db != NULL);
    assert_true(user != NULL);
    assert_true(pwd != NULL);
    assert_true(host == NULL);
    assert_string_equal(db, "two_task_db_test");
    assert_string_equal(user, "user_test");
    assert_string_equal(pwd, "pass_test");

    /* Cleanup */
    free(db);
    free(user);
    free(pwd);
    free(host);
}

Ensure test_connect()
{
    char *db = NULL;
    char *user = NULL;
    char *pwd = NULL;
    char *host = NULL;
    int status;
    
    status = ias_db_get_connect_info(&db, &user, &pwd, &host);
    assert_equal(status, SUCCESS);    

    conn = ias_db_connect_to_database("NONE", "NONE", "NONE", "NONE");
    assert_true(conn == NULL);

    conn = ias_db_connect_to_database(db, user, pwd, host);
    assert_not_equal(conn, NULL);

    free(db);
    free(user);
    free(pwd);
    free(host);
}

Ensure test_query()
{
    char *db = NULL;
    char *user = NULL;
    char *pwd = NULL;
    char *host = NULL;
    int status;
    int nrows = 0;
    int ivalue = -1;    /* query field variables of various types: integer */
    double dvalue = -1; /*                                         double  */
    char svalue[80];    /*                                         string  */

    struct ias_db_query *query = NULL;

    status = ias_db_get_connect_info(&db, &user, &pwd, &host);
    conn = ias_db_connect_to_database(db, user, pwd, host);
    assert_not_equal(conn, NULL);
    free(db);
    free(user);
    free(pwd);
    free(host);

    char sql[240];
    sprintf(sql,
          "CREATE TABLE test_table \
           (\"ID1\" NUMBER(4,0), \"ID2\" NUMBER(2,0), \"DATE_ACQUIRED\" DATE, \
            \"SENSOR_ID\" VARCHAR2(8), \
            \"SOFTWARE_VERSION\" VARCHAR2(16), \"INTEGRATION_TIME\" NUMBER)"
          );
    query = ias_db_perform_query(conn, sql);
    assert_not_equal(query, NULL);
    assert_true(ias_db_query_was_successful(query));
    ias_db_query_close(query);

    status = ias_db_rollback_transaction(conn);
    assert_equal(status, ERROR);    

    status = ias_db_start_transaction(conn);
    assert_equal(status, SUCCESS);    
    status = ias_db_rollback_transaction(conn);
    assert_equal(status, SUCCESS);    

    status = ias_db_start_transaction(conn);
    assert_equal(status, SUCCESS);    

    query = ias_db_perform_query(conn, "insert into test_table values (1, 2, \
                            '01-DEC-09', null, 'Version 1.0', 0.006)");
    assert_not_equal(query, NULL);
    assert_true(ias_db_query_was_successful(query));
    nrows = ias_db_query_get_modified_rows(query);
    assert_equal(nrows, 1);    
    ias_db_query_close(query);
    query = NULL;
    /* Test after closing the query, should get a -1 back */
    nrows = ias_db_query_get_modified_rows(query);
    assert_equal(nrows, -1);    

    status = ias_db_commit_transaction(conn);
    assert_equal(status, SUCCESS);    

    query = ias_db_perform_query(conn, "select id3 from test_table");
    assert_not_equal(query, NULL);
    assert_false(ias_db_query_was_successful(query));
    ias_db_query_close(query);

    query = ias_db_prepare_query(conn, "select id1, id2, date_acquired, \
              sensor_id, software_version, integration_time from test_table");
    assert_not_equal(query, NULL);
    status = ias_db_set_rows_to_insert(query, 1);
    assert_equal(status, SUCCESS);    
    status = ias_db_exec_prepared_query(query);
    assert_equal(status, SUCCESS);    
    assert_true(ias_db_query_is_active(query));
    assert_true(ias_db_query_was_successful(query));

    assert_true(ias_db_query_next(query));
    status = ias_db_query_int_value(query, &ivalue, 0);
    assert_equal(status, SUCCESS);    
    assert_true(ivalue == 1);
    status = ias_db_query_double_value(query, &dvalue, 5);
    assert_equal(status, SUCCESS);    
    assert_true(dvalue == 0.006);
    ias_db_query_datetime_value(query, 2, svalue, sizeof(svalue));
    assert_equal(status, SUCCESS);    
    assert_string_equal(svalue, "2009-12-01T00:00:00");
    ias_db_query_string_value(query, 4, svalue, sizeof(svalue));
    assert_equal(status, SUCCESS);    
    assert_string_equal(svalue, "Version 1.0");
    assert_true(ias_db_query_is_null(query, 3));

    ias_db_query_close(query);

    query = ias_db_perform_query(conn, "DROP TABLE test_table");
    assert_not_equal(query, NULL);
    ias_db_query_close(query);

}


int main(int argc, char **argv)
{
    TestSuite *suite = create_test_suite();

    /* Set up the setup & teardown for all the tests */
    setup(suite, initialize);
    teardown(suite, cleanup);

    /* Actual tests here */
    add_test(suite, test_get_login_for_satellite_sensor);
    add_test(suite, test_connection_info);
    add_test(suite, test_common_connection_info);
    if (getenv("ENABLE_DB_TESTS"))
    {
        add_test(suite, test_connect);
        add_test(suite, test_query);
    }

    /* Initialize the satellite attributes library */
    ias_sat_attr_initialize(IAS_L8);

    return run_test_suite(suite, create_text_reporter());
}
