/*************************************************************************

NAME: database_test_insert.c

PURPOSE: Using Cgreen for testing database library insert routines 

Algorithm References: None


**************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <time.h>
#include <sys/timeb.h>
#include <sql.h>

#include "ias_logging.h"
#include "ias_const.h"
#include "ias_db.h"
#include "ias_db_get_connect_info.h"
#include "ias_db_insert.h"
#include "ias_db_query.h"
#include "ias_satellite_attributes.h"
#include "../local.h"
#include "cgreen/cgreen.h"

/* Uncomment this to see messages with timing information, useful for 
 * performance testing
#define TIMING_TEST
*/

/* Global variables used by more than one test */
struct ias_db_connection *conn;

/* Static sizes & insert values */
int nsca = 4;
int nband = 3;
int num_records;

int *band_nbrs = NULL;
int *sca_nbrs = NULL;


static void initialize()
{
    char *db = NULL;
    char *user = NULL;
    char *pwd = NULL;
    char *host = NULL;
    int status;
    struct ias_db_query *query;

    ias_log_initialize("database_test_insert");

    /* Initialize the satellite attributes library */
    ias_sat_attr_initialize(IAS_L8);
    
    ias_db_initialize_database_lib();

    status = ias_db_get_connect_info(&db,&user,&pwd,&host);
    assert_equal(status, SUCCESS);    
    assert_true(db != NULL);
    assert_true(user != NULL);
    assert_true(pwd != NULL);
    assert_true(host == NULL);

    conn = ias_db_connect_to_database(db, user, pwd, host);
    assert_not_equal(conn, NULL);
    free(db);
    free(user);
    free(pwd);
    free(host);

    char sql[640];
    sprintf(sql,
          "CREATE TABLE test_table \
           (\"BAND_ID\" NUMBER(4,0), CONSTRAINT \
            \"TEST_TABLE_CK\" CHECK (BAND_ID IN (1, 2, 3, 4)), \
            \"SCA_ID\" NUMBER(2,0), \
            \"DATE_ACQUIRED\" DATE, \
            \"SCENE_START_TIME\" TIMESTAMP(9), \
            \"PREDK_START_DATE\" DATE, \
            \"PREDK_START_TIME\" NUMBER(14,9), \
            \"SENSOR_ID\" VARCHAR2(8), \
            \"SOFTWARE_VERSION\" VARCHAR2(16), \
            \"INTEGRATION_TIME\" NUMBER)"
          );

    query = ias_db_perform_query(conn, sql);
    assert_not_equal(query, NULL);
    assert_true(ias_db_query_was_successful(query));
    ias_db_query_close(query);

   /* Create the err log table.  Catch & ignore 'already exists' errors.*/
    if (getenv("ENABLE_DML_ERROR_LOGGING"))
    {
        sprintf(sql, "DECLARE AEE EXCEPTION; PRAGMA EXCEPTION_INIT(AEE,-00955);"
            "BEGIN DBMS_ERRLOG.create_error_log('test_table'); "
            "EXCEPTION WHEN AEE THEN NULL; END;");
        query = ias_db_perform_query(conn, sql);
        assert_not_equal(query, NULL);
        assert_true(ias_db_query_was_successful(query));
        ias_db_query_close(query);
    }


    /* Set up arrays for values that vary */
    num_records = nsca * nband;
    band_nbrs = (int *)malloc(num_records * sizeof(int));
    sca_nbrs = (int *)malloc(num_records * sizeof(int));
    int i,j,k;
    for (i = 0, k = 0; i < nband; i++) 
    {
        for (j = 0; j < nsca; j++, k++) 
        {
            band_nbrs[k] = i + 1;
            sca_nbrs[k] = j + 1;
        }
    }

}

static void cleanup() 
{
    free(band_nbrs);
    free(sca_nbrs);

    struct ias_db_query *query;

    if (conn) 
    {
        query = ias_db_perform_query(conn, "drop table test_table");
        assert_not_equal(query, NULL);
        assert_true(ias_db_query_was_successful(query));
        ias_db_query_close(query);

        ias_db_close_connection(conn);
    }

    ias_db_close_database_lib();
}


/* Test using a table with variables */
Ensure test_ias_database_insert_table()
{
#ifdef TIMING_TEST
    struct timeb tp;               /* time structure, used to time 
                                      the insert operations            */
#endif
    double it_list[num_records];   /* list of integration times        */
    int i;
    int nrec;
    int status;
    int table_size;
    int band_id = 1;
    int string_length = 20;
    char sensor[20]= "OLITIRS";

    /* date/times in timestamp format */
    /*  timestamp format can hold up to 9 fractional digits in the seconds */
    char scene_start[30] = "2014:184:17:00:45.2345678";      
    char pre_dark_start[30] = "2010-068T17:00:45.123456789"; 

    char vstrings[num_records * string_length]; /* SW version strings */
    memset(vstrings, 0, num_records * string_length);
    char dstrings[num_records * string_length]; /* date acquired strings */
    memset(dstrings, 0, num_records * string_length);

    /* Initialize values to write to the DB */
    for (i=0; i<num_records; i++) 
    {
        sprintf(&vstrings[i*string_length], "Value%02d", i+1);
        sprintf(&dstrings[i*string_length], "2009-12-%02dT04:22:08", (i+1)%30);
        it_list[i] = (i+1) / 10.0;
    }

    nrec = nsca;
    /* Table 1 - all fields use single values */
    IAS_DB_INSERT_TABLE table1[] =
    {
        IAS_DB_TABLE_INT("band_id", band_id),
        IAS_DB_TABLE_INT("sca_id" , sca_nbrs[0]),
        IAS_DB_TABLE_DATETIME("date_acquired" , dstrings),
        IAS_DB_TABLE_TIMESTAMP("scene_start_time" , scene_start),
        IAS_DB_TABLE_DATETIME_NS_DOY("predk_start_date" , pre_dark_start),
        IAS_DB_TABLE_STRING("sensor_id" , sensor),
        IAS_DB_TABLE_STRING("software_version", vstrings),
        IAS_DB_TABLE_DOUBLE("integration_time", it_list[0]),
    };
    table_size  = sizeof(table1)/sizeof(IAS_DB_INSERT_TABLE);

#ifdef TIMING_TEST
    printf(" Run db insert tests (timing info follows)...\n");
    ftime(&tp);
    printf("Before insert for %d rows: %ld %d\n",nrec,(long)tp.time,tp.millitm);
#endif
    status = ias_db_insert_using_table(conn, "test_table",
                                       table1, table_size, nrec);
#ifdef TIMING_TEST
    ftime(&tp);
    printf("After insert for %d rows: %ld %d\n",nrec,(long)tp.time,tp.millitm);
#endif
    assert_equal(status, SUCCESS);    

    /* Now do a query to make sure everything comes back out OK */
    int q_nrec;
    int q_band_id[2];
    int q_sca_id[2];
    double q_integration_time[2];
    char q_acqdate[2][30];
    char q_scene_start[2][30];
    char q_pre_dark_start[60] = ""; /* just a single value here */
    char q_sensor[2][20];
    char q_version[2][20];

    q_nrec = 2; /* check 2 records, both should have the same values */
    /* Initialize variables before querying from the db */
    memset(q_acqdate, 0, q_nrec * 30);
    memset(q_scene_start, 0, q_nrec * 30);
    memset(q_sensor, 0, q_nrec * 20);
    memset(q_version, 0, q_nrec * 20);
    for (i=0; i < q_nrec; i++)
    {
        q_band_id[i] = 0;
        q_sca_id[i] = 0;
        q_integration_time[i] = 0;
    }

    IAS_DB_QUERY_TABLE qtable[] =
    {
        {"band_id", q_band_id, sizeof(q_band_id), IAS_QUERY_INT_ARRAY, q_nrec },
        {"sca_id", q_sca_id, sizeof(q_sca_id), IAS_QUERY_INT_ARRAY, q_nrec },
        {"date_acquired", q_acqdate, sizeof(q_acqdate[0]), 
            IAS_QUERY_STRING_ARRAY, q_nrec },
        {"scene_start_time", q_scene_start, sizeof(q_scene_start[0]), 
            IAS_QUERY_STRING_ARRAY, q_nrec },
        {"predk_start_date", q_pre_dark_start, sizeof(q_pre_dark_start), 
            IAS_QUERY_DATETIME_NS_DOY, 1 },
        {"sensor_id", q_sensor, sizeof(q_sensor[0]), 
            IAS_QUERY_STRING_ARRAY, q_nrec },
        {"software_version", q_version, sizeof(q_version[0]), 
            IAS_QUERY_STRING_ARRAY, q_nrec},
        {"integration_time", q_integration_time, sizeof(q_integration_time), 
            IAS_QUERY_DOUBLE_ARRAY, q_nrec },
    };

    table_size  = sizeof(qtable)/sizeof(IAS_DB_QUERY_TABLE);
    status = ias_db_query_using_table(conn, "test_table",
        qtable, table_size, 
        /* use a very long "where" clause to verify the library can handle it */
        "1 = 1                                                                " 
        "                                                                     " 
        "                                                                     " 
        "                                                                     " 
        "                                                                     ",
        q_nrec);
    assert_equal(status, SUCCESS);    
    /* Check values from record 1 */
    assert_string_equal(q_acqdate[0], "2009-12-01 04:22:08");
    assert_string_equal(q_scene_start[0], "2014-07-03 17:00:45.234567800");
    assert_string_equal(q_pre_dark_start, "2010-068T17:00:45.123456789");
    assert_string_equal(q_sensor[0], "OLITIRS");
    /* Check values from record 2 */
    assert_equal(q_band_id[0], q_band_id[1]);
    assert_equal(q_sca_id[0], q_sca_id[1]);
    assert_equal(q_integration_time[0], q_integration_time[1]);
    assert_string_equal(q_acqdate[0], q_acqdate[1]);
    assert_string_equal(q_scene_start[0], q_scene_start[1]);
    assert_string_equal(q_sensor[0], q_sensor[1]);
    assert_string_equal(q_version[0], q_version[1]);


    /* Table 2 - some fields use single values and some use arrays */
    IAS_DB_INSERT_TABLE table2[] =
    {
        IAS_DB_TABLE_INT("band_id", band_id),
        IAS_DB_TABLE_INT_ARRAY("sca_id" , sca_nbrs),
        IAS_DB_TABLE_DATETIME_ARRAY("date_acquired" , dstrings, string_length),
        IAS_DB_TABLE_STRING("sensor_id" , sensor),
        IAS_DB_TABLE_STRING_ARRAY("software_version", vstrings, string_length),
        IAS_DB_TABLE_DOUBLE_ARRAY("integration_time", it_list)
    };
    table_size  = sizeof(table2)/sizeof(IAS_DB_INSERT_TABLE);

#ifdef TIMING_TEST
    ftime(&tp);
    printf("Before insert for %d rows: %ld %d\n",nrec,(long)tp.time,tp.millitm);
#endif
    status = ias_db_insert_using_table(conn, "test_table",
                                       table2, table_size, nrec);
#ifdef TIMING_TEST
    ftime(&tp);
    printf("After insert for %d rows: %ld %d\n",nrec,(long)tp.time,tp.millitm);
#endif
    assert_equal(status, SUCCESS);    

}

/* Test using a table with variables, and some null values */
Ensure test_ias_database_insert_table_with_nulls()
{
#ifdef TIMING_TEST
    struct timeb tp;               /* time structure, used to time 
                                      the insert operations            */
#endif
    IAS_DB_NULL_TYPE null_flags[num_records]; /* null indicator flags */
    double it_list[num_records];   /* list of integration time values  */
    int i;
    int nrec;
    int status;
    int table_size;
    int band_id = 1;
    int string_length = 20;
    char sensor[20]= "OLITIRS";
    char vstrings[num_records * string_length];      /* SW version strings    */
    memset(vstrings, 0, num_records * string_length);
    char dstrings[num_records * string_length];      /* date acquired strings */
    memset(dstrings, 0, num_records * string_length);

    /* Initialize value to be written to the DB */
    for (i=0; i<num_records; i++) 
    {
        sprintf(&vstrings[i*string_length], "Value%02d", i+1);
        sprintf(&dstrings[i*string_length], "2009-12-%02dT00:00:00", (i+1)%30);
        null_flags[i]=0;
        it_list[i] = i / 10.0;
    }
    /* Indicate the value for some records is null */
    null_flags[1] = ias_db_get_null_data_flag();
    null_flags[3] = ias_db_get_null_data_flag();

    nrec = nsca;
    /* Table 1 - all fields use single values */
    IAS_DB_INSERT_TABLE table1[] =
    {
        IAS_DB_TABLE_INT("band_id", band_id),
        IAS_DB_TABLE_INT_WITH_NULLS("sca_id" , sca_nbrs[0], null_flags),
        IAS_DB_TABLE_DATETIME("date_acquired" , dstrings),
        IAS_DB_TABLE_STRING("sensor_id" , sensor),
        IAS_DB_TABLE_STRING_NULL("software_version", vstrings),
        IAS_DB_TABLE_DOUBLE_NULL("integration_time", it_list[0]),
    };
    table_size  = sizeof(table1)/sizeof(IAS_DB_INSERT_TABLE);

#ifdef TIMING_TEST
    ftime(&tp);
    printf("Before insert for %d rows: %ld %d\n",nrec,(long)tp.time,tp.millitm);
#endif
    status = ias_db_insert_using_table(conn, "test_table",
                                       table1, table_size, nrec);
#ifdef TIMING_TEST
    ftime(&tp);
    printf("After insert for %d rows: %ld %d\n",nrec,(long)tp.time,tp.millitm);
#endif
    assert_equal(status, SUCCESS);    

    /* Table 2 - some fields use single values and some use arrays */
    IAS_DB_INSERT_TABLE table2[] =
    {
        IAS_DB_TABLE_INT_NULL("band_id", band_id),
        IAS_DB_TABLE_INT_ARRAY_WITH_NULLS("sca_id" , sca_nbrs, null_flags),
        IAS_DB_TABLE_DATETIME_ARRAY_WITH_NULLS("date_acquired" , dstrings, 
                        string_length, null_flags),
        IAS_DB_TABLE_STRING_NULL("sensor_id" , sensor),
        IAS_DB_TABLE_STRING_ARRAY_WITH_NULLS("software_version", vstrings, 
                        string_length, null_flags),
        IAS_DB_TABLE_DOUBLE_ARRAY_NULL("integration_time", it_list)
    };
    table_size  = sizeof(table2)/sizeof(IAS_DB_INSERT_TABLE);

#ifdef TIMING_TEST
    ftime(&tp);
    printf("Before insert for %d rows: %ld %d\n",nrec,(long)tp.time,tp.millitm);
#endif
    status = ias_db_insert_using_table(conn, "test_table",
                                       table2, table_size, nrec);
#ifdef TIMING_TEST
    ftime(&tp);
    printf("After insert for %d rows: %ld %d\n",nrec,(long)tp.time,tp.millitm);
#endif
    assert_equal(status, SUCCESS);    

}

/* Test using a table with structure elements */
Ensure test_ias_database_insert_struct()
{
#ifdef TIMING_TEST
    struct timeb tp;               /* time structure, used to time 
                                      the insert operations            */
#endif
    int status;
    int table_size;
    int nrec = nsca;  /* write a subset of records */
    struct chardata { /* structure of characterization data   */
        int band_id;
        int sca_id[num_records];
        char date_acquired[20];
        char sensor_id[12];
        char software_version[num_records][16];
        double integration_time;
    } cd;
    struct chardata *cdptr = &cd; /* pointer to characterization data struct */
    int i;

    cd.band_id = 1;
    sprintf(cd.sensor_id, "OLI");
    sprintf(cd.date_acquired, "2009-12-%02dT00:00:00", 5);
    for (i=0; i<num_records; i++) 
    {
            cd.sca_id[i] = sca_nbrs[i];
            sprintf(cd.software_version[i], "Version 1.%02d", i+1);
    }

    IAS_DB_INSERT_TABLE table[] =
    {
        IAS_DB_TABLE_STRUCT_INT(cdptr, band_id),
        IAS_DB_TABLE_STRUCT_INT_ARRAY(cdptr, sca_id),
        IAS_DB_TABLE_STRUCT_DATETIME(cdptr, date_acquired),
        IAS_DB_TABLE_STRUCT_STRING(cdptr, sensor_id),
        IAS_DB_TABLE_STRUCT_STRING_ARRAY(cdptr, software_version, 16),
        IAS_DB_TABLE_STRUCT_DOUBLE(cdptr, integration_time)
    };
    table_size  = sizeof(table)/sizeof(IAS_DB_INSERT_TABLE);

#ifdef TIMING_TEST
    ftime(&tp);
    printf("Before insert for %d rows: %ld %d\n",num_records,(long)tp.time, 
        tp.millitm);
#endif
    status = ias_db_insert_using_table(conn, "test_table",
                                       table, table_size, nrec);
#ifdef TIMING_TEST
    ftime(&tp);
    printf("After insert for %d rows: %ld %d\n",num_records,(long)tp.time, 
        tp.millitm);
#endif
    assert_equal(status, SUCCESS);    
}

/* Test using a table with structure elements with some null values */
Ensure test_ias_database_insert_struct_with_nulls()
{
    struct ias_db_query *query;    /* query structure pointer          */
#ifdef TIMING_TEST
    struct timeb tp;               /* time structure, used to time 
                                      the insert operations            */
#endif

    IAS_DB_NULL_TYPE null_flags[num_records];
    int status;
    int nrec = num_records;
    int table_size;
    struct chardata {   /* A structure of characterization data */
        int band_id;
        int sca_id[num_records];
        char date_acquired[20];
        char sensor_id[12];
        char software_version[num_records][16];
        double integration_time;
    } cd;
    struct chardata *cdptr = &cd;  /* pointer to characterization data */
    int i;

    cd.integration_time = 0.15;
    cd.band_id = 1;
    sprintf(cd.sensor_id, "OLI");
    sprintf(cd.date_acquired, "2009-12-%02dT00:00:00", 5);
    for (i=0; i<num_records; i++) 
    {
            null_flags[i]=0;
            cd.sca_id[i] = sca_nbrs[i];
            sprintf(cd.software_version[i], "Version 1.%02d", i+1);
    }
    /* Indicate the value for some records is null */
    null_flags[0] = ias_db_get_null_data_flag();
    null_flags[3] = ias_db_get_null_data_flag();
    null_flags[6] = ias_db_get_null_data_flag();
    null_flags[8] = ias_db_get_null_data_flag();

    /* Table 1 - rows without nulls */
    IAS_DB_INSERT_TABLE table[] =
    {
        IAS_DB_TABLE_STRUCT_INT(cdptr, band_id),
        IAS_DB_TABLE_STRUCT_INT_ARRAY(cdptr, sca_id),
        IAS_DB_TABLE_STRUCT_DATETIME(cdptr, date_acquired),
        IAS_DB_TABLE_STRUCT_STRING(cdptr, sensor_id),
        IAS_DB_TABLE_STRUCT_STRING_ARRAY(cdptr, software_version, 16),
        IAS_DB_TABLE_STRUCT_DOUBLE(cdptr, integration_time)
    };
    table_size  = sizeof(table)/sizeof(IAS_DB_INSERT_TABLE);

#ifdef TIMING_TEST
    ftime(&tp);
    printf("Before insert for %d rows: %ld %d\n",num_records,(long)tp.time, 
        tp.millitm);
#endif
    status = ias_db_insert_using_table(conn, "test_table",
                                       table, table_size, nrec);
#ifdef TIMING_TEST
    ftime(&tp);
    printf("After insert for %d rows: %ld %d\n",num_records,(long)tp.time, 
        tp.millitm);
#endif
    assert_equal(status, SUCCESS);    

    /* Table 2 - rows with nulls */
    IAS_DB_INSERT_TABLE table2[] =
    {
        IAS_DB_TABLE_STRUCT_INT(cdptr, band_id),
        IAS_DB_TABLE_STRUCT_INT_ARRAY_WITH_NULLS(cdptr, sca_id, null_flags),
        IAS_DB_TABLE_STRUCT_DATETIME_NULL(cdptr, date_acquired),
        IAS_DB_TABLE_STRUCT_STRING_WITH_NULLS(cdptr, sensor_id, null_flags),
        IAS_DB_TABLE_STRUCT_STRING_ARRAY_WITH_NULLS(cdptr, software_version, 16,
            null_flags),
        IAS_DB_TABLE_STRUCT_DOUBLE_NULL(cdptr, integration_time)
    };
    table_size  = sizeof(table2)/sizeof(IAS_DB_INSERT_TABLE);

#ifdef TIMING_TEST
    ftime(&tp);
    printf("Before insert for %d rows: %ld %d\n",num_records,(long)tp.time, 
        tp.millitm);
#endif
    status = ias_db_insert_using_table(conn, "test_table",
                                       table2, table_size, nrec);
#ifdef TIMING_TEST
    ftime(&tp);
    printf("After insert for %d rows: %ld %d\n",num_records,(long)tp.time, 
        tp.millitm);
#endif
    assert_equal(status, SUCCESS);    

    /* Query the data back, ensure nulls are where they should be */
    query = ias_db_perform_query(conn, 
         "select BAND_ID, SCA_ID, DATE_ACQUIRED, SENSOR_ID, SOFTWARE_VERSION, \
          INTEGRATION_TIME from test_table");
    assert_not_equal(query, NULL);
    assert_true(ias_db_query_was_successful(query));
    /* table 1 data - all columns non-null */
    for (i=0; i< nrec; i++)
    {
        assert_true(ias_db_query_next(query));
        assert_false(ias_db_query_is_null(query, 0));
        assert_false(ias_db_query_is_null(query, 1));
        assert_false(ias_db_query_is_null(query, 2));
        assert_false(ias_db_query_is_null(query, 3));
        assert_false(ias_db_query_is_null(query, 4));
        assert_false(ias_db_query_is_null(query, 5));
    }

    /* table 2 data - various null columns */
    for (i=0; i< nrec; i++)
    {
        assert_true(ias_db_query_next(query));
        assert_false(ias_db_query_is_null(query, 0));
        assert_true(ias_db_query_is_null(query, 2));
        assert_true(ias_db_query_is_null(query, 5));
        if (null_flags[i] == ias_db_get_null_data_flag()) 
        {
            assert_true(ias_db_query_is_null(query, 1));
            assert_true(ias_db_query_is_null(query, 3));
            assert_true(ias_db_query_is_null(query, 4));
        } 
        else 
        {
            assert_false(ias_db_query_is_null(query, 1));
            assert_false(ias_db_query_is_null(query, 3));
            assert_false(ias_db_query_is_null(query, 4));
        }
    }
    ias_db_query_close(query);

}

/* Test IAS databse insert statement with DML Error Logging
   Note: If the DML Error Logging table (ERR$_<db_table_name>) 
         hasn't been created, then the below procedure needs to be 
         run manully in the local L8_IAS database to create the table:
              "BEGIN 
               DBMS_ERRLOG.create_error_log (dml_table_name => 
               '<db_table_name>', err_log_table_name => 
               'ERR$_<db_table_name>'); 
               END;"
         The detailed DML Error Logging message can be read in the "
         "ERR$_<db_table_name> table in the local L8_IAS database */
Ensure test_ias_database_insert_with_dml_error_logging()
{
#ifdef TIMING_TEST
    struct timeb tp;               /* time structure, used to time 
                                      the insert operations            */
#endif
    double it_list[num_records];   /* list of integration times        */
    int i;
    int nrec;
    int status;
    int table_size;
    int band_id = 12;
    int string_length = 20;
    char sensor[20]= "OLITIRS";

    /* date/times in timestamp format */
    /*  timestamp format can hold up to 9 fractional digits in the seconds */
    char scene_start[30] = "2014:184:17:00:45.2345678";      
    char pre_dark_start[30] = "2010-068T17:00:45.123456789"; 

    char vstrings[num_records * string_length]; /* SW version strings */
    memset(vstrings, 0, num_records * string_length);
    char dstrings[num_records * string_length]; /* date acquired strings */
    memset(dstrings, 0, num_records * string_length);

    /* Initialize values to write to the DB */
    for (i=0; i<num_records; i++) 
    {
        sprintf(&vstrings[i*string_length], "Value%02d", i+1);
        sprintf(&dstrings[i*string_length], "2009-12-%02dT04:22:08", (i+1)%30);
        it_list[i] = (i+1) / 10.0;
    }

    nrec = nsca;
    /* Table 1 - all fields use single values */
    IAS_DB_INSERT_TABLE table1[] =
    {
        IAS_DB_TABLE_INT("band_id", band_id),
        IAS_DB_TABLE_INT("sca_id" , sca_nbrs[0]),
        IAS_DB_TABLE_DATETIME("date_acquired" , dstrings),
        IAS_DB_TABLE_TIMESTAMP("scene_start_time" , scene_start),
        IAS_DB_TABLE_DATETIME_NS_DOY("predk_start_date" , pre_dark_start),
        IAS_DB_TABLE_STRING("sensor_id" , sensor),
        IAS_DB_TABLE_STRING("software_version", vstrings),
        IAS_DB_TABLE_DOUBLE("integration_time", it_list[0]),
    };
    table_size  = sizeof(table1)/sizeof(IAS_DB_INSERT_TABLE);

#ifdef TIMING_TEST
    printf(" Run db insert tests (timing info follows)...\n");
    ftime(&tp);
    printf("Before insert for %d rows: %ld %d\n",nrec,(long)tp.time,tp.millitm);
#endif

    /* Temporarily disable error logging */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    status = ias_db_insert_using_table(conn, "test_table",
                                       table1, table_size, nrec);
    /* Reenable error logging */
    ias_log_set_output_level(IAS_LOG_LEVEL_ERROR);

#ifdef TIMING_TEST
    ftime(&tp);
    printf("After insert for %d rows: %ld %d\n",nrec,(long)tp.time,tp.millitm);
#endif

    /* Band_id = 12 is out of the added constraint of band_id in (1, 2, 3, 4),
       the status will return ERROR. Besides print out error message, the
       DML error logging message can be checked in the ERR$_TEST_TABLE table 
       content in the local L8_IAS database */ 
    assert_equal(status, ERROR);    
}

Ensure test_db_cleanup()
{
    int status;                 /* function return value */
    int rec_count;              /* number of records found */

    /* check the radiometry tables */
    rec_count = ias_db_get_count(conn, "non_uniformity_band", 
                                     "radiometric_wo_common_id", "5");

    assert_equal_with_message(5, 5, 
          "Number records found in non_uniformity_band table: %d RAD ID: %d", 
          rec_count, 5);

    status = ias_db_delete_records(conn, "non_uniformity_band", 
                                 "radiometric_wo_common_id", "5");
    assert_equal(status, 0);

    assert_equal_with_message(5, 5, 
                "Delete records table: non_uniformity_band ID: %d", 5);

    /* do geometry tables here */
    
    rec_count = ias_db_get_count(conn, "focal_plane_calibration", 
                                     "geometric_wo_common_id", "5");

    assert_equal_with_message(5, 5, 
        "Number records found in focal_plane_calibration table: %d RAD ID: %d", 
          rec_count, 5);

    status = ias_db_delete_records(conn, "focal_plane_calibration", 
                                 "geometric_wo_common_id", "5");
    assert_equal(status, 0);
    assert_equal_with_message(5, 5, 
           "Delete records table: focal_plane_calibration ID: %d", 5);

} /* end of ensure test */

Ensure test_ias_database_parse()
{
    int status;
    SQL_TIMESTAMP_STRUCT timestamp;

    status = ias_db_parse_yyyy_mm_dd_time("2010-03-09T12:04:13", &timestamp);
    assert_equal(status, SUCCESS);
    assert_equal(timestamp.fraction, 0);

    status = ias_db_parse_yyyy_mm_dd_time("2010-03-09T12:04:13.50000", 
        &timestamp);
    assert_equal(status, SUCCESS);
    assert_equal(timestamp.fraction, 500000000);

    status = ias_db_parse_yyyy_mm_dd_time("2010-03-09T12:04:13.00005", 
        &timestamp);
    assert_equal(status, SUCCESS);
    assert_equal(timestamp.fraction, 50000);


} /* end of ensure test */

int main(int argc, char **argv)
{
    TestSuite *suite = create_test_suite();

    /* Set up the setup & teardown for all the tests */
    setup(suite, initialize);
    teardown(suite, cleanup);

    if (getenv("ENABLE_DB_TESTS") == NULL)
    {
        printf("DB Tests will not run, setenv ENABLE_DB_TESTS to enable....\n");
        return 0;
    }

    add_test(suite, test_db_cleanup);
    add_test(suite, test_ias_database_parse);
    add_test(suite, test_ias_database_insert_table);
    add_test(suite, test_ias_database_insert_table_with_nulls);
    add_test(suite, test_ias_database_insert_struct);
    add_test(suite, test_ias_database_insert_struct_with_nulls);

    /* Check if ENABLE_DML_ERROR_LOGGING is set, if it's set then the
       test_ias_database_insert_with_dml_error_logging will be run */ 
    if (getenv("ENABLE_DML_ERROR_LOGGING"))
        add_test(suite, test_ias_database_insert_with_dml_error_logging);
    else
    {
        printf("\n\t***** ENABLE_DML_ERROR_LOGGING is not set, therefore "
               "DML Error Logging will not be made *****\n");
    }

    if (argc > 1) 
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());

}
