/****************************************************************************
* Name: database_test_stored_proc
*
* Purpose: Cgreen test program for stored procs
*
* Notes:   
*
*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <time.h>
#include <sys/timeb.h>
#include <sql.h>
#include <sqlext.h>

#include "ias_logging.h"
#include "ias_const.h"
#include "ias_db_get_connect_info.h"
#include "ias_db.h"
#include "ias_db_stored_proc.h"
#include "ias_db_insert.h"
#include "ias_satellite_attributes.h"
#include "local.h"
#include "cgreen/cgreen.h"

/* Global variables used by tests, and teardown */
struct ias_db_connection *conn = NULL;


static void initialize()
{
    char *db = NULL;
    char *user = NULL;
    char *pwd = NULL;
    char *host = NULL;
    int status;
    struct ias_db_query *query;
    
    ias_log_initialize("database_test_stored_proc");

    /* Initialize the satellite attributes library */
    ias_sat_attr_initialize(IAS_L8);
    
    status = ias_db_get_connect_info(&db,&user,&pwd,&host);
    assert_equal(status, SUCCESS);    
    assert_true(db != NULL);
    assert_true(user != NULL);
    assert_true(pwd != NULL);
    assert_true(host == NULL);

    conn = ias_db_connect_with_transaction(db, user, pwd, host);
    assert_not_equal(conn, NULL);
    free(db);
    free(user);
    free(pwd);
    free(host);
    if (conn == NULL) return;

    char sql[640];
    /* Create a simple procedure with only input parameters */
    sprintf(sql,
        "CREATE OR REPLACE PROCEDURE TEST_SIMPLE_PROC "
        "    (arg1 IN VARCHAR2, "
        "     arg2 IN INTEGER DEFAULT 4) "
        "AS "
        "BEGIN "
        "    DBMS_OUTPUT.PUT_LINE('ARG1 = ' || ARG1); "
        "    DBMS_OUTPUT.PUT_LINE('ARG2 = ' || ARG2); "
        "END TEST_SIMPLE_PROC;"
          );
    query = ias_db_perform_query(conn, sql);
    assert_not_equal(query, NULL);
    assert_true(ias_db_query_was_successful(query));
    ias_db_query_close(query);

    /* Create a procedure (well, actually a function) with arguments of various
     * types, that returns a value and has input & output parameters */
    sprintf(sql,
        "CREATE OR REPLACE FUNCTION TEST_PROC "
        "    (arg1 IN NUMBER DEFAULT 20, "
        "    arg2 IN OUT VARCHAR2, "
        "    arg3 IN TIMESTAMP, "
        "    arg3a OUT TIMESTAMP, "
        "    arg4 OUT INTEGER, "
        "    arg5 IN INTEGER DEFAULT 4) "
        "RETURN INTEGER "
        "AS "
        "BEGIN "
        "    arg4 := arg5; "
        "    select sysdate into arg3a from dual; "
        "    arg2 := 'help me'; "
        "    commit;"
        "    RETURN arg1; "
        "END TEST_PROC;"
          );

    query = ias_db_perform_query(conn, sql);
    assert_not_equal(query, NULL);
    assert_true(ias_db_query_was_successful(query));
    ias_db_query_close(query);

}

static void cleanup() 
{
    struct ias_db_query *query;

    if (conn) 
    {
        query = ias_db_perform_query(conn, "drop function TEST_PROC");
        assert_not_equal(query, NULL);
        assert_true(ias_db_query_was_successful(query));
        ias_db_query_close(query);

        query = ias_db_perform_query(conn, "drop procedure TEST_SIMPLE_PROC");
        assert_not_equal(query, NULL);
        assert_true(ias_db_query_was_successful(query));
        ias_db_query_close(query);

        ias_db_disconnect_with_transaction(conn, IAS_DB_COMMIT);
    }
    ias_db_close_database_lib();
}


Ensure test_proc_stmt()
{
    int num_records = 1;
    int status;
    int table_size;

    /* The stored proc calls in this test should fail on constructing
     * the SQL statement, due to errors in how the table is set up.
     * Therefore we disable output of expected error messages. */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE); 

    /* Should always put the return value first */
    IAS_DB_STORED_PROC_TABLE proc_table_invalid_return[] = {
        {NULL, IAS_DB_PARAMETER_MODE_INPUT, IAS_STORED_PROC_INT, NULL, 0, NULL},
        {NULL, IAS_DB_PARAMETER_MODE_RETURN, IAS_STORED_PROC_INT, NULL, 0, NULL}
    };
    table_size  = sizeof(proc_table_invalid_return) /
        sizeof(IAS_DB_STORED_PROC_TABLE);
    status = ias_db_stored_proc_using_table(conn, "test_proc_xyz", 
        proc_table_invalid_return, table_size, num_records);
    assert_not_equal(status,SUCCESS);

    /* Should only have 1 return value */
    IAS_DB_STORED_PROC_TABLE proc_table_2_return[] = {
        {NULL, IAS_DB_PARAMETER_MODE_RETURN, IAS_STORED_PROC_INT, NULL, 0, 
            NULL},
        {NULL, IAS_DB_PARAMETER_MODE_RETURN, IAS_STORED_PROC_INT, NULL, 0, 
            NULL},
        {NULL, IAS_DB_PARAMETER_MODE_INPUT, IAS_STORED_PROC_INT, NULL, 0, 
            NULL},
    };
    table_size  = sizeof(proc_table_2_return) /
        sizeof(IAS_DB_STORED_PROC_TABLE);
    status = ias_db_stored_proc_using_table(conn, "test_proc_xyz", 
        proc_table_2_return, table_size, num_records);
    assert_not_equal(status,SUCCESS);
}

Ensure test_proc_call()
{

    /* Arguments to the stored procedure of various types */
    int arg1 = 99;
    char arg2[80] = "";
    char arg3[80] = "2010:151:08:30:00.150"; /* yyyy:ddd:hh:mm:ss.sss */
    char arg3a[80] = ""; 
    int arg4 = 0;
    int arg5 = 9;
    int retval=0;

    int status=0;
    IAS_DB_STORED_PROC_TABLE proc_table[] = {
        /* Always put the return value first */
        {NULL, IAS_DB_PARAMETER_MODE_RETURN, IAS_STORED_PROC_INT, &retval, 0, 
            NULL},
        {NULL, IAS_DB_PARAMETER_MODE_INPUT, IAS_STORED_PROC_INT, &arg1, 0, 
            NULL},
        {NULL, IAS_DB_PARAMETER_MODE_INPUTOUTPUT, IAS_STORED_PROC_STRING, arg2, 
            sizeof(arg2), NULL},
        {NULL, IAS_DB_PARAMETER_MODE_INPUT, IAS_STORED_PROC_TIMESTAMP, arg3, 
            sizeof(arg3), NULL},
        /* bind output timestamp as a string */
        {NULL, IAS_DB_PARAMETER_MODE_OUTPUT, IAS_STORED_PROC_STRING, arg3a, 
            sizeof(arg3a), NULL},
        {NULL, IAS_DB_PARAMETER_MODE_OUTPUT, IAS_STORED_PROC_INT, 
            &arg4, 0, NULL}, 
        {NULL, IAS_DB_PARAMETER_MODE_INPUT, IAS_STORED_PROC_INT, &arg5, 0,
            NULL},
    };
    int table_size  = sizeof(proc_table)/sizeof(IAS_DB_STORED_PROC_TABLE);

    /* First test, just call the function once */
    int num_records = 1;
    status = ias_db_stored_proc_using_table(conn, "test_proc", proc_table,
            table_size, num_records);
    assert_equal(status,SUCCESS);
    assert_equal(retval,arg1);
    assert_string_not_equal(arg3a,"");
    assert_equal(arg4,arg5);
    assert_string_equal(arg2,"help me");


    /* Second test, call multiple times in a batch */
    /* This should fail because arrays were not provided for output parameters*/
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);  /* disable logging */
    num_records = 20;
    status = ias_db_stored_proc_using_table(conn, "test_proc", proc_table,
            table_size, num_records);
    assert_not_equal(status,SUCCESS);
    ias_log_initialize("database_test_stored_proc"); /* reset logging */


    /* Third test, call multiple times in a batch */
    /*  to do so, set up arrays of argument values */
    int i;
    int string_length = 80;
    int arr_arg1[num_records];
    char arr_arg2[string_length*num_records];
    char arr_arg3[string_length*num_records]; /* yyyy:ddd:hh:mm:ss.sss */
    char arr_arg3a[string_length*num_records];
    int arr_arg4[num_records];
    int arr_arg5[num_records];
    int arr_retval[num_records];
    IAS_DB_NULL_TYPE null_flags[num_records];
    for (i=0; i<num_records; i++) 
    {
        arr_arg1[i] = i;
        sprintf(&arr_arg2[i*string_length], "myvalue%d", i);
        sprintf(&arr_arg3[i*string_length], "2010:151:08:30:00.150");
        strcpy(&arr_arg3a[i*string_length], "");
        arr_arg4[i] = 0;
        arr_arg5[i] = i+1;
        if (i%2) null_flags[i] = ias_db_get_null_data_flag();
    }
    IAS_DB_STORED_PROC_TABLE proc_table_array[] = {
        /* always put the return value first */
        {NULL, IAS_DB_PARAMETER_MODE_RETURN, IAS_STORED_PROC_INT_ARRAY, 
            arr_retval, 0, NULL},
        /* array of input values */
        {NULL, IAS_DB_PARAMETER_MODE_INPUT, IAS_STORED_PROC_INT_ARRAY, 
            arr_arg1, 0, NULL},
        /* array of input/output values */
        {NULL, IAS_DB_PARAMETER_MODE_INPUTOUTPUT, IAS_STORED_PROC_STRING_ARRAY, 
            arr_arg2, string_length, null_flags},
        /* use a single input value, will be replicated for each row */
        {NULL, IAS_DB_PARAMETER_MODE_INPUT, IAS_STORED_PROC_TIMESTAMP, 
            arg3, sizeof(arg3), NULL},
        /* bind output timestamp as a string */
        {NULL, IAS_DB_PARAMETER_MODE_OUTPUT, IAS_STORED_PROC_STRING_ARRAY, 
            arr_arg3a, string_length, NULL},
        /* array of output values */
        {NULL, IAS_DB_PARAMETER_MODE_OUTPUT, IAS_STORED_PROC_INT_ARRAY, 
            arr_arg4, 0, NULL}, 
        /* array of input values */
        {NULL, IAS_DB_PARAMETER_MODE_INPUT, IAS_STORED_PROC_INT_ARRAY, 
            arr_arg5, 0, NULL},
    };
    table_size  = sizeof(proc_table_array)/sizeof(IAS_DB_STORED_PROC_TABLE);
    status = ias_db_stored_proc_using_table(conn, "test_proc", proc_table_array,
            table_size, num_records);
    assert_equal(status,SUCCESS);
    /* Spot check the results */
    assert_equal(arr_retval[0],arr_arg1[0]);
    assert_equal(arr_retval[num_records-1],arr_arg1[num_records-1]);
    assert_string_equal(arr_arg2,"help me");
    assert_string_not_equal(arr_arg3a,"");
    assert_equal(arr_arg4[0],arr_arg5[0]);
    assert_equal(arr_arg4[num_records-1],arr_arg5[num_records-1]);

}

Ensure test_proc_call_novars()
{
    struct ias_db_query *dbq = NULL;
    int status = 0;

    /* Call with both parameters */
    dbq = ias_db_prepare_query(conn, 
            "{CALL TEST_SIMPLE_PROC('testing',1)}"); 
    assert_not_equal(dbq, NULL);
    status = ias_db_exec_prepared_query(dbq);
    assert_equal(status,SUCCESS);
    ias_db_query_close(dbq);

    /* Call with one default parameter */
    dbq = ias_db_prepare_query(conn, 
            "{CALL TEST_SIMPLE_PROC('testing')}"); 
    assert_not_equal(dbq, NULL);
    status = ias_db_exec_prepared_query(dbq);
    assert_equal(status,SUCCESS);
    ias_db_query_close(dbq);

}


int main(int argc, char **argv)
{
    TestSuite *suite = create_test_suite();

    /* Set up the setup & teardown for all the tests */
    setup(suite, initialize);
    teardown(suite, cleanup);

    if (getenv("ENABLE_DB_TESTS") == NULL)
    {
        printf("DB Tests will not run, setenv ENABLE_DB_TESTS to enable....\n");
        return 0;
    }

    /* Actual tests here */
    add_test(suite, test_proc_stmt);
    add_test(suite, test_proc_call_novars);
    add_test(suite, test_proc_call);

    if (argc > 1) 
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}
