/*******************************************************************************
Name: ias_geo_calculate_sun_angle

Purpose: Calculate the solar azimuth and elevation provided a geodetic latitude,
    longitude, height, and time.

Note: The NOVAS library must be initialized prior to using this function.
    Leap seconds must be initialized for the UT1 correction.

Returns: 
    Type = integer
    SUCCESS / ERROR
*******************************************************************************/
/* System included files */
#include <math.h>
#include <stdlib.h>
#include <string.h>

/* IAS Lib included files */
#include "ias_geo.h"
#include "ias_const.h"
#include "ias_cpf.h"
#include "ias_logging.h"

int ias_geo_calculate_sun_angle
(
    IAS_CPF *cpf, /* I: Earth CPF parameters */
    double latitude, /* I: Latitude in radians */
    double longitude, /* I: Longitude in radians */
    double height, /* I: Height in meters from ellipse */
    const double *time_utc, /* I UTC year, doy, and seconds of day */ 
    double *azimuth, /* O: Sun azimuth in radians from 0 north +east */
    double *elevation /* O: Sun elevation */
)
{
    const IAS_CPF_EARTH_CONSTANTS *cpf_earth;
    const double D2R = ias_math_get_radians_per_degree(); 

    double flattening; /* Flattening value calculated from axis */
    double julian_date; /* Time represented as Julian Date */
    double ut1_utc_correction; /* UT1-UTC correction (seconds) */
    double pole_wander_x; /* Earth's pole wander correction */
    double pole_wander_y; /* Earth's pole wander correction */

    double temp_lat = 0.0; /* Latitude used in calculations */
    double temp_long = 0.0; /* Longitude used in calculations */
    double radius = 0.0; /* Distance from origin used in calculations */ 
    double temp_azimuth = 0.0;

    int status = ERROR; /* Return value variable */

    IAS_VECTOR ecef_ground; /* Ground point ECEF vector */
    IAS_VECTOR ecij2k_pos; /* ECIJ2K sun position vector */
    IAS_VECTOR ecef_sun; /* ECEF sun position vector */
    IAS_VECTOR local_ground; /* Ground position vector used in calculations */
    IAS_VECTOR local_sun; /* Sun position vector used in calculations */
    IAS_VECTOR solar_position; /* Sun position vector used in calculations */
    IAS_VECTOR temp_vector; /* Temp vector to hold results between calculation*/
    IAS_VECTOR zero_vector; /* Zero vector for velocity */

    /* Verify pointer parameters are valid */
    if (cpf == NULL)
    {
        IAS_LOG_ERROR("CPF parameter is null.");
        return ERROR;
    }
    if (azimuth == NULL)
    {
        IAS_LOG_ERROR("Azimuth parameter is null.");
        return ERROR;
    }
    if (elevation == NULL)
    {
        IAS_LOG_ERROR("Elevation parameter is null.");
        return ERROR;
    }

    /* Initialize variables */
    memset(&zero_vector, 0, sizeof(zero_vector));
    cpf_earth = ias_cpf_get_earth_const(cpf);

    IAS_LOG_DEBUG("Calculating sun angle at time %lf, %lf, %lf.",
        time_utc[0], time_utc[1], time_utc[2]);

    /* Obtain the solar position and convert it to an ECEF vector */    
    /* Get the pole wander X/Y and UT1/UTC conversion */
    status = ias_geo_compute_getmjdcoords(time_utc, cpf, 
        &julian_date, &pole_wander_x, &pole_wander_y, &ut1_utc_correction);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving pole wander X/Y and UT1-UTC conversion");
        return ERROR;
    }

    IAS_LOG_DEBUG("Pole wander %lf, %lf, and UT1-UTC %lf.",
        pole_wander_x, pole_wander_y, ut1_utc_correction);

    /* Convert the Modified Julian Date to Full Julian Date */
    julian_date += IAS_MODIFIED_JULIAN_DATE;

    /* Compute the ECI true-of-date solar direction vector at the time. */
    status = ias_geo_get_solar_position_vector(julian_date, &solar_position);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Unable to get solar position vector for Julian date "
            "%lf.", julian_date);
        return ERROR;
    }

    /* Convert the ECITOD sun vector to ECI of epoch J2000 */
    status = ias_geo_transform_tod2j2k(ut1_utc_correction,
        &solar_position, time_utc, &ecij2k_pos);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Unable to convert the ECITOD sun vector to ECIJ2K.");
        return ERROR;
    }

   /* Convert the ECIJ2000 sun vector to ECEF, including the pole wander and 
        UT1-UTC corrections from the geometric model. Zero vector used for 
        velocity of the earth position instead of the sat velocity vector. */
    status = ias_geo_eci2ecef(pole_wander_x, pole_wander_y, ut1_utc_correction,
        &ecij2k_pos, &zero_vector, time_utc, &ecef_sun, &temp_vector);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Unable to convert the ECIJ2K sun vector to ECEF.");
        return ERROR;
    }

    IAS_LOG_DEBUG("ECEF sun vector %lf, %lf, %lf.", 
        ecef_sun.x, ecef_sun.y, ecef_sun.z);

    /* Obtain the ECEF vector for the geodetic location */
    /* Calculate the WGS84 flattening */
    flattening = (cpf_earth->semi_major_axis
        - cpf_earth->semi_minor_axis) / cpf_earth->semi_major_axis;

    /* Convert the geodetic location to a ground point vector. */
    ias_geo_convert_geod2cart(latitude, longitude, height,
        cpf_earth->semi_major_axis, flattening, &ecef_ground);
    status = ias_geo_convert_cart2sph(&ecef_ground, &temp_lat, &temp_long, 
        &radius);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Unable to convert to convert %lf, %lf, %lf to spherical "
            "coords.", ecef_ground.x, ecef_ground.y, ecef_ground.z);
        return ERROR;
    }

    /* Convert the geodetic location with radius to spherical coordinates to
       create the local coordinate system */
    ias_geo_convert_sph2cart(latitude, longitude, radius, &ecef_ground);
    IAS_LOG_DEBUG("ECEF ground point vector %lf, %lf, %lf, for lat/lon/height "
        "%lf, %lf, %lf", ecef_ground.x, ecef_ground.y, ecef_ground.z,
            latitude, longitude, height);
    
    /* Rotate the vectors to make the ground point the local coordinate system*/
    /* Convert the vector to spherical coordinates for rotation */
    status = ias_geo_convert_cart2sph(&ecef_ground, &temp_lat, &temp_long, 
        &radius);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Unable to convert to convert %lf, %lf, %lf to spherical "
            "coords.", ecef_ground.x, ecef_ground.y, ecef_ground.z);
        return ERROR;
    }
    IAS_LOG_DEBUG("ECEF ground spherical coords %lf, %lf, %lf",
        temp_lat, temp_long, radius);

    /* Rotate about the z axis to return X to the origin */
    ias_math_rotate_3dvec_around_z(&ecef_sun, temp_long, &local_sun);
    ias_math_rotate_3dvec_around_z(&ecef_ground, temp_long, &local_ground);
    IAS_LOG_DEBUG("Earth coords after rotate around z: %lf, %lf, %lf",
       local_ground.x, local_ground.y, local_ground.z);

    /* Rotate about the y axis to align the local coordinate system with the Z
        axis */
    ias_math_rotate_3dvec_around_y(&local_sun, ((90.0 * D2R) - temp_lat), 
        &ecef_sun);
    ias_math_rotate_3dvec_around_y(&local_ground, ((90.0 * D2R) - temp_lat), 
        &ecef_ground);
    IAS_LOG_DEBUG("Earth coords after rotate around y: %lf, %lf, %lf",
        ecef_ground.x, ecef_ground.y, ecef_ground.z);

    /* Shift the coordinates for the height, convert to km for the solar unit */
    ecef_ground.z -= radius; 
    ecef_sun.z -= (radius / 1000);

    /* After the rotations the local coordinate system is now X+ facing south,
        Y+ east,  Z+ zenith. */
    IAS_LOG_DEBUG("Final ground coords: %lf, %lf, %lf",
        ecef_ground.x, ecef_ground.y, ecef_ground.z);
    IAS_LOG_DEBUG("Final sun coords: %lf, %lf, %lf",
        ecef_sun.x, ecef_sun.y, ecef_sun.z);

    /* Convert the vector to spherical coordinates for the azimuth and elev. */
    ias_geo_convert_cart2sph(&ecef_sun, elevation, &temp_azimuth, &radius);
   
    /* Switch the azimuth to be measured from north as 0 from geo measurement */
    *azimuth = ((180.0 * D2R) - temp_azimuth); 

    /* Move the values into range */
    while (*azimuth > (180.0 * D2R)) 
    {
        *azimuth -= (360.0 * D2R);
    }

    while (*azimuth < (-180.0 * D2R))
    {
        *azimuth += (360.0 * D2R);
    }

    while (*elevation > (90.0 * D2R)) 
    {
        *elevation -= (180.0 * D2R);
    }

    while (*elevation < (-90.0 * D2R))
    {
        *elevation += (180.0 * D2R); 
    }
 
    return SUCCESS;
}

