/*******************************************************************************
NAME: ias_geo_convert_geodetic_to_geocentric_height

PURPOSE:  Calculate the geocentric latitude and the radius of a point above the
ellipse, given the geodetic latitude and the height.

RETURN VALUE:
Type = int 
Value           Description
-----           -----------
SUCCESS         Latitude and radius calculated
ERROR           Error calculating latitude and height 

*******************************************************************************/
#include <math.h>
#include "ias_geo.h"
#include "ias_logging.h"

int ias_geo_convert_geodetic_to_geocentric_height
(
    double geodetic_latitude,    /* I: Geodetic latitude (radians) */
    double height,               /* I: Elevation (meters) */
    const IAS_EARTH_CHARACTERISTICS *earth, /* I: Earth constants */
    double *geocentric_latitude, /* O: Geocentric latitude (radians) */
    double *radius               /* O: Satellite radius at point (meters) */
)
{
    double sine_geodetic_latitude; /* Sine of the geodetic latitude */
    double denominator;          /* Denominator in calculation of radius of
                                    curvature */
    double g1;                   /* N + h (radius of curvature plus geodetic 
                                    height (meters) */
    double g2;                   /* N + (1-e^2) + h (radius of curvature *
                                    (1 - eccentricity squared) + geodetic
                                    height (meters) */
    double x_coordinate;         /* ECR (geocentric) vector X coordinate */
    double z_coordinate;         /* ECR (geocentric) vector Z coordinate */

    /* Calculate the sine of the latitude */
    sine_geodetic_latitude = sin(geodetic_latitude);

    /* Find the denominator in the radius of curvature calcuation */
    denominator = sqrt(1.0 - earth->eccentricity * sine_geodetic_latitude *
        sine_geodetic_latitude);

    /* Check the denominator for divide by 0 */
    if (denominator == 0.0)
    {
        IAS_LOG_ERROR("Calculating ellipsoid radius of curvator; invalid "
            "denominator %f", denominator); 
        return ERROR;
    }

    g1 = earth->semi_major_axis / denominator + height;
    g2 = (earth->semi_major_axis * (1.0 - earth->eccentricity)) / denominator
        + height;

    x_coordinate = g1 * cos(geodetic_latitude);
    z_coordinate = g2 * sine_geodetic_latitude;

    /* Calculate the radius of the point */
    *radius = sqrt(x_coordinate * x_coordinate + z_coordinate * z_coordinate);

    /* Check radius */
    if (*radius == 0.0)
    {
        IAS_LOG_ERROR("Invalid radius %f", *radius); 
        return ERROR;
    }

    /* Calculate the geocentric latitude of the point */
    *geocentric_latitude = asin(z_coordinate / *radius);

    return SUCCESS;
}
