#include <stdio.h>
#include <math.h>
#include "ias_const.h"
#include "ias_structures.h"
#include "ias_geo.h"
#include "ias_logging.h"
#include "cgreen/cgreen.h"

/* Initialize the leap seconds to zero to match the cal/val data used to 
   generate the expected results */
static void initialize_leap_seconds()
{
    int status;

    /* Set up a leap second structure with just a single entry to make it
       easy */
    IAS_MATH_LEAP_SECONDS_DATA leap_data;
    int leap_years = 2006;
    int leap_months = 1;
    int leap_days = 1;
    int leap_seconds = 1;
    leap_data.leap_years = &leap_years;
    leap_data.leap_months = &leap_months;
    leap_data.leap_days = &leap_days;
    leap_data.num_leap_seconds = &leap_seconds;
    leap_data.leap_seconds_count = 1;

    /* Initialize the leap seconds.  Note that it will be zero with the
       values set up here.  This is done to match the cal/val test case we
       have. */
    status = ias_math_init_leap_seconds(1, &leap_data);
    assert_equal(status, SUCCESS);
}

/* Initialize the leap seconds in a way to get 2 leap seconds since some
   calval test data was run with that. */
static void initialize_leap_seconds2()
{
    int status;

    /* Set up a leap second structure with just a single entry to make it
       easy */
    IAS_MATH_LEAP_SECONDS_DATA leap_data;
    int leap_years = 2000;
    int leap_months = 2;
    int leap_days = 1;
    int leap_seconds = 2;
    leap_data.leap_years = &leap_years;
    leap_data.leap_months = &leap_months;
    leap_data.leap_days = &leap_days;
    leap_data.num_leap_seconds = &leap_seconds;
    leap_data.leap_seconds_count = 1;

    /* Initialize the leap seconds.  Note that 300 days into year 2000 is
       chosen to make sure it gets the 2 leap seconds defined above. */
    status = ias_math_init_leap_seconds(IAS_SEC_PER_DAY * 300.0, &leap_data);
    assert_equal(status, SUCCESS);
}

Ensure test_compute_orientation_matrices()
{
    /* The following values were extracted from the routine
       oli_calc_attitude by running the cal/val prototype code */
    #include "oli_calc_attitude.h"
    int count = sizeof(inputs)/sizeof(inputs[0]);
    int index;

    for (index = 0; index < count; index++)
    {
        int status;
        double attpert[3][3];
        double orb2ecf[3][3];
        IAS_VECTOR pos, vel;
        pos.x = inputs[index].position_x;
        pos.y = inputs[index].position_y;
        pos.z = inputs[index].position_z;
        vel.x = inputs[index].velocity_x;
        vel.y = inputs[index].velocity_y;
        vel.z = inputs[index].velocity_z;

        status = ias_geo_compute_orientation_matrices(&pos, &vel,
            inputs[index].roll, inputs[index].pitch, inputs[index].yaw,
            orb2ecf, attpert);
        if ( status != SUCCESS )
        {
            assert_equal_with_message( status, 0,
                "failed status with index = %d", index );
        }
        else
        {
            int valid;
            valid  = doubles_are_equal( exp_attpert[index].val_0_0,
                                        attpert[0][0] );
            valid += doubles_are_equal( exp_attpert[index].val_0_1,
                                        attpert[0][1] );
            valid += doubles_are_equal( exp_attpert[index].val_0_2,
                                        attpert[0][2] );
            valid += doubles_are_equal( exp_attpert[index].val_1_0,
                                        attpert[1][0] );
            valid += doubles_are_equal( exp_attpert[index].val_1_1,
                                        attpert[1][1] );
            valid += doubles_are_equal( exp_attpert[index].val_1_2,
                                        attpert[1][2] );
            valid += doubles_are_equal( exp_attpert[index].val_2_0,
                                        attpert[2][0] );
            valid += doubles_are_equal( exp_attpert[index].val_2_1,
                                        attpert[2][1] );
            valid += doubles_are_equal( exp_attpert[index].val_2_2,
                                        attpert[2][2] );
            /* Use the valid flag to reduce the number of asserts */
            if ( valid != 9 )
            {
                assert_equal_with_message( valid, 9,
                    "attpert[%d] failure %d", index );
            }

            valid = 0;
            valid  = doubles_are_equal( exp_orb2ecf[index].val_0_0,
                                        orb2ecf[0][0] );
            valid += doubles_are_equal( exp_orb2ecf[index].val_0_1,
                                        orb2ecf[0][1] );
            valid += doubles_are_equal( exp_orb2ecf[index].val_0_2,
                                        orb2ecf[0][2] );
            valid += doubles_are_equal( exp_orb2ecf[index].val_1_0,
                                        orb2ecf[1][0] );
            valid += doubles_are_equal( exp_orb2ecf[index].val_1_1,
                                        orb2ecf[1][1] );
            valid += doubles_are_equal( exp_orb2ecf[index].val_1_2,
                                        orb2ecf[1][2] );
            valid += doubles_are_equal( exp_orb2ecf[index].val_2_0,
                                        orb2ecf[2][0] );
            valid += doubles_are_equal( exp_orb2ecf[index].val_2_1,
                                        orb2ecf[2][1] );
            valid += doubles_are_equal( exp_orb2ecf[index].val_2_2,
                                        orb2ecf[2][2] );
            /* Use the valid flag to reduce the number of asserts */
            if ( valid != 9 )
            {
                assert_equal_with_message( valid, 9,
                    "orb2ecf[%d] failure", index );
            }
        }
    }
}

Ensure test_dms2deg()
{
    double dms = 75000000;
    double degrees;
    int status;

    status = ias_geo_convert_dms2deg(dms, &degrees ,"LAT");
    assert_equal(status, SUCCESS);
    assert_double_equal(degrees, 75.0);

    dms = -135045018;
    status = ias_geo_convert_dms2deg(dms, &degrees, "LON");
    assert_equal(status, SUCCESS);
    assert_double_equal(degrees, -135.755);
}

Ensure test_transform_precession()
{
    IAS_VECTOR mod;         /* mean-of-date */
    IAS_VECTOR j2k;         /* J2000 */
    IAS_VECTOR mod2;        /* second mean-of-date */
    double jd_tdb;          /* Julian date (Barycentric) */
    int status;

    /* test with a reference value from the calval code */
    mod.x = 5.549152772992e+03;
    mod.y = 3.375800609455e+03;
    mod.z = -3.761132173410e+03;
    jd_tdb = 2.4527882492868355e+06;
    status = ias_geo_transform_precession_mod2j2k(&mod, jd_tdb, &j2k);
    assert_equal(status, SUCCESS);

    significant_figures_for_assert_double_are(12);
    assert_double_equal(j2k.x, 5.550476308335e+03);
    assert_double_equal(j2k.y, 3.371576564066e+03);
    assert_double_equal(j2k.z, -3.762967672188e+03);
    significant_figures_for_assert_double_are(8);

    /* test the inverse routine to make sure we get back to the input */
    status = ias_geo_transform_precession_j2k2mod(&j2k, jd_tdb, &mod2);
    assert_equal(status, SUCCESS);
    significant_figures_for_assert_double_are(12);
    assert_double_equal(mod2.x, mod.x);
    assert_double_equal(mod2.y, mod.y);
    assert_double_equal(mod2.z, mod.z);
    significant_figures_for_assert_double_are(8);

    /* Test an independent value through j2k2mod */
    j2k.x = 4.071237028243e+03;
    j2k.y = 1.894260904552e+03;
    j2k.z = -6.014324435052e+03;
    jd_tdb = 2.4527882537891506e+06;
    status = ias_geo_transform_precession_j2k2mod(&j2k, jd_tdb, &mod2);
    assert_equal(status, SUCCESS);

    significant_figures_for_assert_double_are(12);
    assert_double_equal(mod2.x, 4.071783015361e+03);
    assert_double_equal(mod2.y, 1.897359788049e+03);
    assert_double_equal(mod2.z, -6.012977850885e+03);
    significant_figures_for_assert_double_are(8);

}

Ensure test_transform_nutation()
{
    IAS_VECTOR tod;         /* true-of-date */
    IAS_VECTOR mod;         /* mean-of-date */
    IAS_VECTOR tod2;        /* second true-of-date */
    double jd_tdb;          /* Julian date (Barycentric) */

    /* test with a reference value from the calval code */
    tod.x = 5.312245100101e+03;
    tod.y = 3.107357836477e+03;
    tod.z = -4.296574602283e+03;
    jd_tdb = 2.4527882502243356e+06;
    ias_geo_transform_nutation_tod2mod(&tod, jd_tdb, &mod);

    significant_figures_for_assert_double_are(12);
    assert_double_equal(mod.x, 5.312158495740e+03);
    assert_double_equal(mod.y, 3.107635961188e+03);
    assert_double_equal(mod.z, -4.296480523558e+03);
    significant_figures_for_assert_double_are(8);

    /* test the inverse routine to make sure we get back to the input */
    ias_geo_transform_nutation_mod2tod(&mod, jd_tdb, &tod2);
    significant_figures_for_assert_double_are(12);
    assert_double_equal(tod2.x, tod.x);
    assert_double_equal(tod2.y, tod.y);
    assert_double_equal(tod2.z, tod.z);
    significant_figures_for_assert_double_are(8);

    /* Test the mod2tod independently */
    mod.x = 3.194988176846e+06;
    mod.y = 3.291866263308e+06;
    mod.z = 5.387087402926e+06;
    jd_tdb = 2.4527882510460950e+06;
    ias_geo_transform_nutation_mod2tod(&mod, jd_tdb, &tod2);

    significant_figures_for_assert_double_are(12);
    assert_double_equal(tod2.x, 3.195379712683e+06);
    assert_double_equal(tod2.y, 3.291529241408e+06);
    assert_double_equal(tod2.z, 5.387061107617e+06);
    significant_figures_for_assert_double_are(8);

}

Ensure test_transform_tod2j2k()
{
    IAS_VECTOR ecitod_pos;  /* ECI true-of-date */
    IAS_VECTOR ecij2k_pos;  /* ECI J2000 */
    IAS_VECTOR ecitod_pos2; /* ECI true-of-date output*/
    double ut1_utc;
    double ephem_time[3];
    int status;
    int old_level;

    /* test with a reference value from the calval code */
    ut1_utc = -2.8901100158691406e-01;
    ecitod_pos.x = 7.7007300574826784e+06;
    ecitod_pos.y = -3.4719108799844891e+08;
    ecitod_pos.z = -1.4540810860492480e+08;
    ephem_time[0] = 2011.000000;
    ephem_time[1] = 167.000000;
    ephem_time[2] = 2.8968784413965102e+04;

    /* Test first without initializing the leap seconds */
    old_level = ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = ias_geo_transform_tod2j2k(ut1_utc, &ecitod_pos, ephem_time,
        &ecij2k_pos);
    ias_log_set_output_level(old_level);
    assert_equal(status, ERROR);

    /* Now initialize the leap seconds so we get 2 leap seconds like the 
       source calval data used and make sure it succeeds */
    initialize_leap_seconds2();
    status = ias_geo_transform_tod2j2k(ut1_utc, &ecitod_pos, ephem_time,
        &ecij2k_pos);
    assert_equal(status, SUCCESS);

    significant_figures_for_assert_double_are(12);
    assert_double_equal(ecij2k_pos.x, 6.6184588250776259e+06);
    assert_double_equal(ecij2k_pos.y, -3.4720882716210353e+08);
    assert_double_equal(ecij2k_pos.z, -1.4541904008940428e+08);
    significant_figures_for_assert_double_are(8);

    /* Make sure the reverse transformation gets us back to the starting
       values */
    status = ias_geo_transform_j2k2tod(ut1_utc, &ecij2k_pos, ephem_time,
        &ecitod_pos2);
    assert_equal(status, SUCCESS);

    significant_figures_for_assert_double_are(12);
    assert_double_equal(ecitod_pos2.x, ecitod_pos.x);
    assert_double_equal(ecitod_pos2.y, ecitod_pos.y);
    assert_double_equal(ecitod_pos2.z, ecitod_pos.z);
    significant_figures_for_assert_double_are(8);
}

Ensure test_transform_sidereal()
{
    /* the test values here were obtained from the calval code */
    IAS_VECTOR craft_pos;
    IAS_VECTOR craft_vel;
    IAS_VECTOR output_pos;
    IAS_VECTOR output_vel;
    IAS_VECTOR expected_satpos;
    IAS_VECTOR expected_satvel;
    double ephem_time[3];
    double ut1_utc;
    int old_level;
    int status;

    craft_pos.x = -8.0199568376053055e+05;
    craft_pos.y = -3.2903798985486268e+06;
    craft_pos.z = 6.2106211433689268e+06;
    craft_vel.x = -3.9040692649947755e+03;
    craft_vel.y = -5.5260507457419071e+03;
    craft_vel.z = -3.4237088818397810e+03;
    ut1_utc = -3.7553811073303223e-01;
    ephem_time[0] = 2003.000000;
    ephem_time[1] = 148.000000;
    ephem_time[2] = 6.4625197637006640e+04;

    expected_satpos.x = 2.1113892612408218e+06;
    expected_satpos.y = 2.6479864691988756e+06;
    expected_satpos.z = 6.2106211433689268e+06;
    expected_satvel.x = 5.6729220937511536e+03;
    expected_satvel.y = 3.5257378811302415e+03;
    expected_satvel.z = -3.4237088818397810e+03;

    /* Test first without initializing the leap seconds */
    old_level = ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = ias_geo_transform_sidereal_ecef2eci(&craft_pos, &craft_vel,
        ut1_utc, ephem_time, &output_pos, &output_vel);
    ias_log_set_output_level(old_level);
    assert_equal(status, ERROR);

    /* Now initialize the leap seconds */
    initialize_leap_seconds();

    /* transform sidereal from ecef to eci */
    status = ias_geo_transform_sidereal_ecef2eci(&craft_pos, &craft_vel,
        ut1_utc, ephem_time, &output_pos, &output_vel);
    assert_equal(status, SUCCESS);
    significant_figures_for_assert_double_are(12);
    assert_double_equal(output_pos.x, expected_satpos.x);
    assert_double_equal(output_pos.y, expected_satpos.y);
    assert_double_equal(output_pos.z, expected_satpos.z);
    assert_double_equal(output_vel.x, expected_satvel.x);
    assert_double_equal(output_vel.y, expected_satvel.y);
    assert_double_equal(output_vel.z, expected_satvel.z);
    significant_figures_for_assert_double_are(8);

    /* close the loop by going back the other way */
    status = ias_geo_transform_sidereal_eci2ecef(&expected_satpos,
        &expected_satvel, ut1_utc, ephem_time, &output_pos, &output_vel);
    assert_equal(status, SUCCESS);
    significant_figures_for_assert_double_are(12);
    assert_double_equal(output_pos.x, craft_pos.x);
    assert_double_equal(output_pos.y, craft_pos.y);
    assert_double_equal(output_pos.z, craft_pos.z);
    assert_double_equal(output_vel.x, craft_vel.x);
    assert_double_equal(output_vel.y, craft_vel.y);
    assert_double_equal(output_vel.z, craft_vel.z);
    significant_figures_for_assert_double_are(8);
}

Ensure test_transform_ecef2j2k()
{
    /* the test values here were obtained from the calval code */
    double xp = 9.2359995469450951e-03;
    double yp = 5.4391497373580933e-01;
    double ut1_utc = -3.7553811073303223e-01;
    IAS_VECTOR craft_pos = {-8.0589937257266312e+05, -3.2959202485220185e+06,
        6.2071852505678236e+06};
    IAS_VECTOR craft_vel = {-3.9039778036655675e+03, -5.5217958233198315e+03,
        -3.4307064440203471e+03};

    double ephem_time[3] = {2003.000000, 148.000000, 6.4626197637006640e+04};

    IAS_VECTOR expected_satpos = {2.1207597529798527e+06,
        2.6501776311321389e+06, 6.2065005769583937e+06};
    IAS_VECTOR expected_satvel = {5.6719540593724760e+03,
        3.5187705162779439e+03, -3.4324716562244193e+03};
    IAS_VECTOR output_pos;
    IAS_VECTOR output_vel;
    int old_level;
    int status;

    /* Test without initializing the leap seconds to make sure the error
       checking is correct */
    old_level = ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = ias_geo_transform_ecef2j2k(xp, yp, ut1_utc, &craft_pos,
        &craft_vel, ephem_time, &output_pos, &output_vel);
    ias_log_set_output_level(old_level);
    assert_equal(status, ERROR);

    /* Now initialize the leap seconds */
    initialize_leap_seconds();

    /* transform from ecef to eci */
    status = ias_geo_transform_ecef2j2k(xp, yp, ut1_utc, &craft_pos,
        &craft_vel, ephem_time, &output_pos, &output_vel);
    assert_equal(status, SUCCESS);
    significant_figures_for_assert_double_are(12);
    assert_double_equal(output_pos.x, expected_satpos.x);
    assert_double_equal(output_pos.y, expected_satpos.y);
    assert_double_equal(output_pos.z, expected_satpos.z);
    assert_double_equal(output_vel.x, expected_satvel.x);
    assert_double_equal(output_vel.y, expected_satvel.y);
    assert_double_equal(output_vel.z, expected_satvel.z);
    significant_figures_for_assert_double_are(8);
}

Ensure test_does_cross_180()
{
    int units;
    int crosses_180;
    int status;
    double lons[4];
    int i;

    /* Test longitudes that don't cross 180 degrees */
    lons[0] = 65.0;
    lons[1] = 67.0;
    lons[2] = 65.0;
    lons[3] = 67.0;

    status = ias_geo_get_units("DEGREES", &units);
    assert_equal(status, SUCCESS);

    crosses_180 = ias_geo_does_cross_180(units, lons);
    assert_equal(crosses_180, 0);

    /* Test longitudes that do cross 180 degrees */
    lons[0] = 179.0;
    lons[1] = -179.0;
    lons[2] = 179.0;
    lons[3] = -179.0;

    status = ias_geo_get_units("DEGREES", &units);
    assert_equal(status, SUCCESS);

    crosses_180 = ias_geo_does_cross_180(units, lons);
    assert_equal(crosses_180, 1);

    /* Rotate one corner on the opposite side of the boundary through all four
       corners to make sure they all work */
    for (i = 0; i < 4; i++)
    {
        lons[0] = -179.0;
        lons[1] = -179.0;
        lons[2] = -179.0;
        lons[3] = -179.0;

        lons[i] = 179.0;
        crosses_180 = ias_geo_does_cross_180(units, lons);
        assert_equal(crosses_180, 1);
    }
    for (i = 0; i < 4; i++)
    {
        lons[0] = 179.0;
        lons[1] = 179.0;
        lons[2] = 179.0;
        lons[3] = 179.0;

        lons[i] = -179.0;
        crosses_180 = ias_geo_does_cross_180(units, lons);
        assert_equal(crosses_180, 1);
    }

    /* Now check two corners in different orders */
    lons[0] = -179.0;
    lons[1] = -179.0;
    lons[2] = 179.0;
    lons[3] = 179.0;
    crosses_180 = ias_geo_does_cross_180(units, lons);
    assert_equal(crosses_180, 1);

    lons[0] = -179.0;
    lons[1] = 179.0;
    lons[2] = -179.0;
    lons[3] = 179.0;
    crosses_180 = ias_geo_does_cross_180(units, lons);
    assert_equal(crosses_180, 1);
}

Ensure test_add_once_around()
{

    double lon;
    int status;
    int units;
   
    lon = 65.0;
    status = ias_geo_get_units("DEGREES", &units);
    assert_equal(status, SUCCESS);

    status = ias_geo_add_once_around( units, &lon);
    assert_equal(status, 1);
    assert_double_equal(lon, 425.0);


    status = ias_geo_get_units("DMS", &units);
    lon  = 125010000.0;
    status = ias_geo_add_once_around( units, &lon);
    assert_double_equal(lon, 485010000.0);

    status = ias_geo_get_units("DMS", &units);
    lon  = -125001000.0;
    status = ias_geo_add_once_around( units, &lon);
    assert_double_equal(lon, 234059000.0);
}

Ensure test_check_start_end_date()
{
    int status;
    int old_level;

    /* Valid dates */
    status = ias_geo_check_start_end_date(20010501, 20071130);
    assert_equal(status, SUCCESS);

    /* Wrong order */
    old_level = ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = ias_geo_check_start_end_date(20071130, 20010501);
    ias_log_set_output_level(old_level);
    assert_equal(status, ERROR);

    /* Same day */
    status = ias_geo_check_start_end_date(20010501, 20010501);
    assert_equal(status, SUCCESS);

    /* Earliest valid year */
    status = ias_geo_check_start_end_date(19700101, 19700101);
    assert_equal(status, SUCCESS);

    /* Year too early */
    old_level = ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = ias_geo_check_start_end_date(19691201, 19700101);
    ias_log_set_output_level(old_level);
    assert_equal(status, ERROR);
}

Ensure test_convert_utc2times()
{
    /* Note that the test inputs and expected outputs were obtained from
       running the cal/val prototype with instrumentation added */
    double utc_time[3] = {2003, 148, 64621.197637007};
    double ut1_utc_correction =  -0.375538;
    double ut1;
    double tdb;
    double tt;
    int status;
    int old_level;
    double expected_ut1 = 2452788.247926182;
    double expected_tdb = 2452788.248673410;
    double expected_tt = 2452788.248673399;

    /* Set up a leap second structure with just a single entry to make it
       easy */
    IAS_MATH_LEAP_SECONDS_DATA leap_data;
    int leap_years = 2006;
    int leap_months = 1;
    int leap_days = 1;
    int leap_seconds = 1;
    leap_data.leap_years = &leap_years;
    leap_data.leap_months = &leap_months;
    leap_data.leap_days = &leap_days;
    leap_data.num_leap_seconds = &leap_seconds;
    leap_data.leap_seconds_count = 1;

    /* Call the utc2times converter before initializing the leap seconds.  It
       should get an error. */
    old_level = ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = ias_geo_convert_utc2times(ut1_utc_correction, utc_time, &ut1,
        &tdb, &tt);
    ias_log_set_output_level(old_level);
    assert_equal(status, ERROR);

    /* Now initialize the leap seconds.  Note that it will be zero with the
       values set up here.  This is done to match the cal/val test case we
       have.  The number of leap seconds involved doesn't really matter since
       the leap second adjustments are handling in a routine other than the
       one being tested. */
    status = ias_math_init_leap_seconds(1, &leap_data);
    assert_equal(status, SUCCESS);

    /* Call the time conversion again.  It should succeed. */
    status = ias_geo_convert_utc2times(ut1_utc_correction, utc_time, &ut1,
        &tdb, &tt);
    assert_equal(status, SUCCESS);

    /* For checking the output, adjust the number of significant figures the
       asserts should use since the normal 8 isn't nearly enough */
    significant_figures_for_assert_double_are(14);
    assert_double_equal(ut1, expected_ut1);
    assert_double_equal(tdb, expected_tdb);
    assert_double_equal(tt, expected_tt);
    significant_figures_for_assert_double_are(8);
}

Ensure test_transform_polar_motion()
{
    /* This is a simple routine to test the polar motion transformation
       routines.  The input data and expected results were obtained by
       instrumenting the cal/val prototype. */
    IAS_VECTOR orig_mean_vec =
        {-3.9040694182566572e+03, -5.526041717521e+03, -3.4237232790671233e+03};
    IAS_VECTOR expected_true_vec =
    {-3.9040692649947755e+03, -5.5260507457419071e+03, -3.4237088818397810e+03};
    IAS_VECTOR orig_true_vec =
        {-3.6106390977172086e+03, -5.252997267482e+03, -3.9629407873372243e+03};
    IAS_VECTOR expected_mean_vec =
    {-3.6106392751266053e+03, -5.2529868173199702e+03, -3.9629544776667494e+03};
    double xp;
    double yp;
    double jd_tdb;
    IAS_VECTOR result_vec;

    /* Test the polar motion mean to true routine */
    xp = 9.2359995469450951e-03;
    yp = 5.4391497373580933e-01;
    jd_tdb = 2.4527882487197062e+06;
    ias_geo_transform_polar_motion_mean_pole_to_true(&orig_mean_vec,
        xp, yp, jd_tdb, &result_vec);

    /* Check the results to 12 digits since the normal 8 is not really
       enough */
    significant_figures_for_assert_double_are(12);
    assert_double_equal(result_vec.x, expected_true_vec.x);
    assert_double_equal(result_vec.y, expected_true_vec.y);
    assert_double_equal(result_vec.z, expected_true_vec.z);
    significant_figures_for_assert_double_are(8);

    /* Test the polar motion true to mean routine */
    xp = 9.2359995469450951e-03;
    yp = 5.4391497373580933e-01;
    jd_tdb = 2.4527882496340577e+06;
    ias_geo_transform_polar_motion_true_pole_to_mean(&orig_true_vec, xp, yp,
        jd_tdb, &result_vec);

    /* Check the results to 12 digits since the normal 8 is not really
       enough */
    significant_figures_for_assert_double_are(12);
    assert_double_equal(result_vec.x, expected_mean_vec.x);
    assert_double_equal(result_vec.y, expected_mean_vec.y);
    assert_double_equal(result_vec.z, expected_mean_vec.z);
    significant_figures_for_assert_double_are(8);
}

Ensure test_eci2ecef()
{
    /* the test values here were obtained from the calval code */
    IAS_VECTOR craft_pos;
    IAS_VECTOR craft_vel;
    IAS_VECTOR output_pos;
    IAS_VECTOR output_vel;
    IAS_VECTOR expected_satpos;
    IAS_VECTOR expected_satvel;
    double ephem_time[3];
    double ut1_utc;
    double xp;
    double yp;
    int old_level;
    int status;

    craft_pos.x = 2.1886505753927529e+06;
    craft_pos.y = 2.6921881180430208e+06;
    craft_pos.z = 6.1648087911697980e+06;

    craft_vel.x = 5.6429478033750656e+03;
    craft_vel.y = 3.4828111094225696e+03;
    craft_vel.z = -3.5159599565255185e+03;

    ut1_utc = -3.7553811073303223e-01;
    ut1_utc = -3.7553811073303223e-01;
    xp = 9.2359995469450951e-03;
    yp = 5.4391497373580933e-01;

    ephem_time[0] = 2003.000000;
    ephem_time[1] = 148.000000;
    ephem_time[2] = 6.4638197637006640e+04;

    expected_satpos.x = -8.5273915085308347e+05;
    expected_satpos.y = -3.3618746847285721e+06;
    expected_satpos.z = 6.1655145917082233e+06;
    expected_satvel.x = -3.6573365425108068e+03;
    expected_satvel.y = -5.5325460126438502e+03;
    expected_satvel.z = -3.5142042604574658e+03;

    /* Test first without initializing the leap seconds */
    old_level = ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = ias_geo_eci2ecef(xp, yp, ut1_utc, &craft_pos, &craft_vel,
        ephem_time, &output_pos, &output_vel);
    ias_log_set_output_level(old_level);
    assert_equal(status, ERROR);

    /* Now initialize the leap seconds */
    initialize_leap_seconds();

    /* transform eci to ecef */
    status = ias_geo_eci2ecef(xp, yp, ut1_utc, &craft_pos, &craft_vel,
        ephem_time, &output_pos, &output_vel);
    assert_equal(status, SUCCESS);
    significant_figures_for_assert_double_are(12);
    assert_double_equal(output_pos.x, expected_satpos.x);
    assert_double_equal(output_pos.y, expected_satpos.y);
    assert_double_equal(output_pos.z, expected_satpos.z);
    assert_double_equal(output_vel.x, expected_satvel.x);
    assert_double_equal(output_vel.y, expected_satvel.y);
    assert_double_equal(output_vel.z, expected_satvel.z);
    significant_figures_for_assert_double_are(8);

    /* close the loop by going back the other way */
    status = ias_geo_ecef2eci(xp, yp, ut1_utc, &expected_satpos, 
        &expected_satvel, ephem_time, &output_pos, &output_vel);

    assert_equal(status, SUCCESS);
    significant_figures_for_assert_double_are(12);
    assert_double_equal(output_pos.x, craft_pos.x);
    assert_double_equal(output_pos.y, craft_pos.y);
    assert_double_equal(output_pos.z, craft_pos.z);
    assert_double_equal(output_vel.x, craft_vel.x);
    assert_double_equal(output_vel.y, craft_vel.y);
    assert_double_equal(output_vel.z, craft_vel.z);
    significant_figures_for_assert_double_are(8);
}

int main(int argc, char **argv)
{
    TestSuite *suite = create_test_suite();
    ias_log_initialize("geo_test");
    add_test(suite, test_compute_orientation_matrices);
    add_test(suite, test_dms2deg);
    add_test(suite, test_transform_precession);
    add_test(suite, test_transform_nutation);
    add_test(suite, test_transform_tod2j2k);
    add_test(suite, test_add_once_around);
    add_test(suite, test_does_cross_180);
    add_test(suite, test_transform_sidereal);
    add_test(suite, test_transform_ecef2j2k);
    add_test(suite, test_check_start_end_date);
    add_test(suite, test_convert_utc2times);
    add_test(suite, test_transform_polar_motion);
    add_test(suite, test_eci2ecef);
    if (argc > 1) {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}
