/* Standard Library Includes */
#include <limits.h>
#include <math.h>
#include <string.h>

/* IAS Library includes */
#include "ias_logging.h"
#include "ias_const.h"
#include "ias_cpf.h"

/* Local Includes */
#include "cgreen/cgreen.h"
#include "ias_geo.h"

/* Local Variables */
static IAS_CPF *cpf;     

/*Regression test of LC80280312013186LGN00 values from Earth Explorer L1T data*/
Ensure test_LC80280312013186LGN00()
{
    const double D2R = ias_math_get_radians_per_degree();
    const double R2D = ias_math_get_degrees_per_radian();

    const IAS_CPF_EARTH_CONSTANTS *earth = 0;

    double time_utc[3]; 

    double latitude = (41.75949 * D2R);
    double longitude = (-96.04587 * D2R);

    double azimuth = 0.0;
    double elevation = 0.0;
 
    int status = 0;
    
    time_utc[0] = 2013;
    time_utc[1] = 186;
    time_utc[2] = 59.6167929;
    time_utc[2] += (7 * 60);
    time_utc[2] += (17 * 60 * 60);

    /*Initialize the leap seconds*/
    earth = ias_cpf_get_earth_const(cpf);
    ias_math_init_leap_seconds_from_UTC_time(
        time_utc, &earth->leap_seconds_data);

    status = ias_geo_calculate_sun_angle(cpf,
        latitude, longitude, 381.0, time_utc, 
        &azimuth, &elevation);

    assert_equal(status, SUCCESS);

    /* Check that the values are within 8 significant digits */
    assert_double_equal((elevation * R2D), 64.544654);
    assert_double_equal((azimuth * R2D), 132.155590);
}

/*Regression test of LC81620312013181LGN00 values from Earth Explorer L1T data*/
Ensure test_LC81620312013181LGN00()
{
    const double D2R = ias_math_get_radians_per_degree();
    const double R2D = ias_math_get_degrees_per_radian();

    const IAS_CPF_EARTH_CONSTANTS *earth = 0;

    double time_utc[3];

    double latitude = (41.75971 * D2R);
    double longitude = (56.92105 * D2R);

    double azimuth = 0.0;
    double elevation = 0.0;

    int status = 0;

    time_utc[0] = 2013;
    time_utc[1] = 181;
    time_utc[2] = 67.2496737;
    time_utc[2] += (55 * 60);
    time_utc[2] += (6 * 60 * 60);

    /*Initialize the leap seconds*/
    earth = ias_cpf_get_earth_const(cpf);
    ias_math_init_leap_seconds_from_UTC_time(
        time_utc, &earth->leap_seconds_data);

    status = ias_geo_calculate_sun_angle(cpf,
        latitude, longitude, 85.0, time_utc,
        &azimuth, &elevation);

    assert_equal(status, SUCCESS);

    /* Check that the values are within 8 significant digits */
    assert_double_equal((elevation * R2D), 65.030998);
    assert_double_equal((azimuth * R2D), 131.971772);
}

Ensure test_night()
{
    const double D2R = ias_math_get_radians_per_degree();
    const double R2D = ias_math_get_degrees_per_radian();

    const IAS_CPF_EARTH_CONSTANTS *earth = 0;

    double time_utc[3];

    double latitude = (-79.9807 * D2R);
    double longitude = (-104.199 * D2R);

    double azimuth = 0.0;
    double elevation = 0.0;

    int status = 0;

    time_utc[0] = 2013;
    time_utc[1] = 119;
    time_utc[2] = 2.5422482;
    time_utc[2] += (38 * 60);
    time_utc[2] += (10 * 60 * 60);

    /*Initialize the leap seconds*/
    earth = ias_cpf_get_earth_const(cpf);
    ias_math_init_leap_seconds_from_UTC_time(
        time_utc, &earth->leap_seconds_data);

    status = ias_geo_calculate_sun_angle(cpf,
        latitude, longitude, 923.0, time_utc,
        &azimuth, &elevation);

    assert_equal(status, SUCCESS);

    /* Check that the values are within 8 significant digits */
    assert_double_equal((elevation * R2D), -20.007952);
    assert_double_equal((azimuth * R2D), 121.381263);
}

static void setup_test()
{
    char cpf_filename[PATH_MAX];
    int status;

    if (ias_sat_attr_initialize(IAS_L8) != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing satellite attributes library");
        exit(EXIT_FAILURE);
    }

    status = snprintf(cpf_filename, sizeof(cpf_filename),
        "%s/ias_lib/misc/geo/L8CPF20130401_20130627.08",
        getenv("IAS_UNIT_TEST_DATA_DIR"));
    if (status < 0 || status > sizeof(cpf_filename))
    {
        IAS_LOG_ERROR("Creating the cpf filename");
        exit(EXIT_FAILURE);
    }

    /* Read the CPF */
    cpf = ias_cpf_read(cpf_filename);
    if (!cpf)
    {
        IAS_LOG_ERROR("Cannot read CPF file %s", cpf_filename);
        exit(EXIT_FAILURE);
    }

    if (ias_geo_initialize_novas() != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing NOVAS library");
        exit(EXIT_FAILURE);
    }
}

static void teardown_test()
{
    /* Free the CPF */
    ias_cpf_free(cpf);
}

int main(int argc, char **argv)
{
    int status; /* Status placeholder */
    ias_log_initialize("Sun Angle Test");
    /*ias_log_set_output_level(IAS_LOG_LEVEL_DEBUG);*/

    TestSuite *suite = create_test_suite();
    add_test(suite, test_LC80280312013186LGN00);
    add_test(suite, test_LC81620312013181LGN00);
    add_test(suite, test_night);

    setup_test();
    if (argc > 1) 
    {
        status = run_single_test(suite, argv[1], create_text_reporter());
        teardown_test();
        return status;
    }

    status = run_test_suite(suite, create_text_reporter());
    teardown_test();
    return status;
}
