/***************************************************************************
 NAME:                   solarcal_oli_jpl_solar_interface_test

 PURPOSE:   CGreen unit tests of the jpl_solar_interface routines
            jpl_get_earth_sun_distance() and jpl_get_solar_position()
            to make sure the proper earth-sun distance and solar
            position vectors are returned for a specified Julian date

 RETURNS:   CGreen status

****************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "ias_l0r.h"
#include "ias_math.h"
#include "ias_structures.h"
#include "ias_miscellaneous.h"
#include "ias_geo.h"
#include "ias_logging.h"
#include "cgreen.h"


Ensure test_compute_sun_position()
{
    char acquisition_date[IAS_L0R_DATE_LENGTH];
    IAS_DATETIME  datetime;
    double jul_date;          /* Time information for Julian dates */
    double jul_date_out_of_range;  /* Time information for Julian dates */
    double secs_in_day;
    double distance;
    double right_ascension;
    double declination;
    int status;

    /* Set a test acquisition date of February 14, 2002 at 12:27:22
       in L0R format */
    memset(acquisition_date, 0, IAS_L0R_DATE_LENGTH);
    sprintf(acquisition_date, "2002:045:12:27:22.00000");
    memset(&datetime, 0, sizeof(IAS_DATETIME));

    status = ias_misc_parse_datetime_string(acquisition_date,
                                 IAS_DATETIME_L0R_FORMAT, &datetime);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("Cannot parse current date/time string %s\n",
               acquisition_date);
        return;
    }

    /* Calculate the total seconds corresponding to the time in the
       current day */
    secs_in_day = (datetime.hour * 3600) + (datetime.minute * 60)
                + datetime.second;

    /* Compute the full Julian date */
    jul_date = ias_math_compute_full_julian_date(datetime.year,
                    datetime.month, datetime.day_of_month, secs_in_day);
    if (status != SUCCESS)
    {
        printf("Cannot calculate Julian date for given acquisition date\n");
        return;
    }

    /* Make sure the call fails when the initialize routine hasn't been 
       called */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = ias_geo_compute_sun_position(jul_date, &right_ascension,
                    &declination, &distance);
    assert_true(status == ERROR);
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

    status = ias_geo_initialize_novas();
    assert_true(status == SUCCESS);

    /* Make sure a dates out of range of the ephemeris data are caught */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    jul_date_out_of_range = 2817562.019005; /* Beyond date range upper bound */
    status = ias_geo_compute_sun_position(jul_date_out_of_range,
            &right_ascension, &declination, &distance);
    assert_true(status == ERROR);
    jul_date_out_of_range = 2217562.019005; /* Before date range lower bound */
    status = ias_geo_compute_sun_position(jul_date_out_of_range,
            &right_ascension, &declination, &distance);
    assert_true(status == ERROR);
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

    /* Get the earth-sun distance using the L0R-formatted date/time */
    status = ias_geo_compute_sun_position(jul_date, &right_ascension,
                    &declination, &distance);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("Cannot obtain Earth-Sun distance using NOVAS library\n");
        return;
    }

    /* Check the results */
    assert_double_equal(jul_date, 2452320.019005);
    assert_double_equal(right_ascension, 21.859670);
    assert_double_equal(declination, -12.975859);
    assert_double_equal(distance, 0.987639428);

    status = ias_geo_shutdown_novas();
    assert_true(status == SUCCESS);
}

Ensure test_get_solar_position_vector()
{
    char acquisition_date[IAS_L0R_DATE_LENGTH];
    IAS_DATETIME  datetime;
    double jul_date;          /* Time information for Julian dates */
    double secs_in_day;
    IAS_VECTOR vector;
    int status;

    /* Set a test acquisition date of February 14, 2002 at 12:27:22
       in L0R format */
    memset(acquisition_date, 0, IAS_L0R_DATE_LENGTH);
    sprintf(acquisition_date, "2002:045:12:27:22.00000");
    memset(&datetime, 0, sizeof(IAS_DATETIME));

    status = ias_misc_parse_datetime_string(acquisition_date,
                                 IAS_DATETIME_L0R_FORMAT, &datetime);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("Cannot parse current date/time string %s\n",
               acquisition_date);
        return;
    }

    /* Calculate the total seconds corresponding to the time in the
       current day */
    secs_in_day = (datetime.hour * 3600) + (datetime.minute * 60)
                + datetime.second;

    /* Compute the full Julian date */
    jul_date = ias_math_compute_full_julian_date(datetime.year,
                    datetime.month, datetime.day_of_month, secs_in_day);
    if (status != SUCCESS)
    {
        printf("Cannot calculate Julian date for given acquisition date\n");
        return;
    }

    /* Make sure the call fails when the initialize routine hasn't been 
       called */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = ias_geo_get_solar_position_vector(jul_date, &vector);
    assert_true(status == ERROR);
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);
 
    status = ias_geo_initialize_novas();
    assert_true(status == SUCCESS);

    /* Get the earth-sun distance using the L0R-formatted date/time */
    status = ias_geo_get_solar_position_vector(jul_date, &vector);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("Cannot obtain Earth-Sun distance using NOVAS library\n");
        return;
    }

    /* Check the results */
    assert_double_equal(vector.x, 121958582.876289);
    assert_double_equal(vector.y, -76519146.532261);
    assert_double_equal(vector.z, -33175578.713717);

    status = ias_geo_shutdown_novas();
    assert_true(status == SUCCESS);
}

Ensure test_compute_moon_position()
{
    double jul_date;          /* Time information for Julian dates */
    double jul_date_out_of_range; /* Time information for Julian dates */
    double distance;
    double right_ascension;
    double declination;
    int status;

    /* set date to test */
    jul_date = 2455728.838331;

    /* Make sure the call fails when the initialize routine hasn't been 
       called */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = ias_geo_compute_moon_position(jul_date, &right_ascension,
                    &declination, &distance);
    assert_true(status == ERROR);
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

    status = ias_geo_initialize_novas();
    assert_true(status == SUCCESS);

    /* Make sure a date out of range of the ephemeris data is caught */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    jul_date_out_of_range = 2817562.019005; /* Beyond date range upper bound */
    status = ias_geo_compute_moon_position(jul_date_out_of_range,
            &right_ascension, &declination, &distance);
    assert_true(status == ERROR);
    jul_date_out_of_range = 2217562.019005; /* Before date range lower bound */
    status = ias_geo_compute_moon_position(jul_date_out_of_range,
            &right_ascension, &declination, &distance);
    assert_true(status == ERROR);
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

    /* Get the earth-moon distance */
    status = ias_geo_compute_moon_position(jul_date, &right_ascension,
                    &declination, &distance);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("Cannot obtain Earth-Moon distance using NOVAS library\n");
        return;
    }

    /* Check the results */
    assert_double_equal(right_ascension, 18.08772268279828);
    assert_double_equal(declination, -22.715346);
    assert_double_equal(distance, 376502.490102);

    status = ias_geo_shutdown_novas();
    assert_true(status == SUCCESS);
}

Ensure test_initialize()
{
    int status;

    /* Make sure the initialize routine returns an error if the environment
       variable for the NOVAS library isn't set. */
    assert_equal(unsetenv("JPLDE421"), 0);

    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    status = ias_geo_initialize_novas();
    assert_true(status == ERROR);

    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);
}

int main(int argc, char **argv)
{
    TestSuite *suite = create_test_suite();
    add_test(suite, test_initialize);
    add_test(suite, test_compute_sun_position);
    add_test(suite, test_get_solar_position_vector);
    add_test(suite, test_compute_moon_position);
    if (argc > 1) {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}
