/*******************************************************************************
NAME: ias_math_calculate_polygon_centroid

PURPOSE: Given the vertices of a polygon determine its center of mass. While
    calculating the center of mass, the polygon is automatically closed by
    wrapping around back to the first point.

RETURNS:
    X coordinate of center of mass
    Y coordinate of center of mass
    0 for success, 1 for failure

NOTE:
    The formula used is as follows
        Cx = (1/6A) * (summation i=0->n-1)(x[i] + x(i+1))(xi*y(i+1)-x(i+1)*y[i])
        Cy = (1/6A) * (summation i=0->n-1)(y[i] + y(i+1))(x[i]*y(i+1)-x(i+1)*yi)
        A = (1/2) * (summation i=0->n-1)(x[i]*y(i+1)-x(i+1)*y[i]
*******************************************************************************/

#include "ias_math.h"
#include "ias_logging.h"

int ias_math_calculate_polygon_centroid
(
    const double *x_points, /* I: Array of x points. Parallel to y_points. */
    const double *y_points, /* I: Array of y points. Parallel to x_points. */
    int num_points,   /* I: Number of provided points */
    double *x_center, /* O: X coordinate of polygon center */
    double *y_center  /* O: Y coordinate of polygon center */
)
{
    int i; /* Loop counter */
    double a;
    double temp_sum_a = 0;
    double temp_sum_x = 0;
    double temp_sum_y = 0;

    /* If we're not receiving a polygon, return error */
    if (num_points < 3)
    {
        IAS_LOG_ERROR("Not enough provided points to define a polygon");
        return ERROR;
    }

    for (i = 0; i < num_points; i++)
    {
        int point1 = i;
        int point2 = (i+1)%num_points;

        double t = x_points[point1] * y_points[point2] 
            - x_points[point2] * y_points[point1];

        temp_sum_a += t;
        temp_sum_x += (x_points[point1] + x_points[point2]) * t;
        temp_sum_y += (y_points[point1] + y_points[point2]) * t;
    }

    /* Compute final part of equation for centroid location */
    a = 0.5 * temp_sum_a;
    *x_center = (1/(6*a)) * temp_sum_x;
    *y_center = (1/(6*a)) * temp_sum_y;

    return SUCCESS;
}
