/******************************************************************************
NAME:           ias_math_complex.c

PURPOSE:
A package of routines that manipulate complex variable data structures.

******************************************************************************/
#include <stdlib.h>
#include <math.h>              /* Defines several math functions */
#include "ias_const.h"         /* Defines ERROR and SUCCESS */
#include "ias_math_complex.h"  /* Defines complex functions */

/******************************************************************************
NAME:           ias_math_complex_exponential

PURPOSE:
Computes the complex exponential of a complex number.

RETURN VALUE:
Type = void

******************************************************************************/

void ias_math_complex_exponential
(
    const IAS_COMPLEX *argument, /* I: Input complex argument      */
    IAS_COMPLEX *value     /* O: Output complex exponential value  */
)
{
    double temp;           /* Temporary variable                   */

    /* The complex exponential e^c is e^re * ( cos(im) + j sin(im) ) */
    temp = exp(argument->re);
    value->re = temp * cos(argument->im);
    value->im = temp * sin(argument->im);
}

/******************************************************************************
NAME:           ias_math_complex_multiply

PURPOSE:
Multiplies two complex numbers.

RETURN VALUE:
Type = void

******************************************************************************/

void ias_math_complex_multiply
(
    const IAS_COMPLEX *arg1, /* I: First input complex argument    */
    const IAS_COMPLEX *arg2, /* I: Second input complex argument   */
    IAS_COMPLEX *value     /* O: Output complex value              */
)
{
    /* The complex product a * b is (a.re*b.re - a.im*b.im)
         + j (a.re*b.im + a.im*b.re) */
    value->re = arg1->re * arg2->re - arg1->im * arg2->im;
    value->im = arg1->re * arg2->im + arg1->im * arg2->re;
}

/******************************************************************************
NAME:           ias_math_complex_divide

PURPOSE:
Divides two complex numbers.

RETURN VALUE:
Type = void

******************************************************************************/

void ias_math_complex_divide
(
    const IAS_COMPLEX *arg1, /* I: First input complex argument    */
    const IAS_COMPLEX *arg2, /* I: Second input complex argument   */
    IAS_COMPLEX *value     /* O: Output complex value              */
)
{
    double denominator;    /* Denominator for real and imaginary parts */

    /* The complex division a / b is ab / |b|^2
       In component notation:
       (a, b)      ac + bd      bc - ad
       ------ = (-----------, -----------)
       (c, d)    c**2 + d**2  c**2 + d**2      */

    denominator = arg2->re * arg2->re + arg2->im * arg2->im;
    value->re = (arg1->re * arg2->re + arg1->im * arg2->im) / denominator;
    value->im = (arg1->im * arg2->re - arg1->re * arg2->im) / denominator;
}

/******************************************************************************
NAME:           ias_math_complex_magnitude

PURPOSE:
Computes the magnitude of a complex number.

RETURN VALUE:
Type = double

Value           Description
-----           -----------
mag             Returns the magnitude of the complex number.

******************************************************************************/
double ias_math_complex_magnitude
(
    const IAS_COMPLEX *arg1 /* I: Input complex argument           */
)
{
    double mag;            /* Magnitude of input complex number    */

    /* Compute the modulus */
    mag = sqrt(arg1->re * arg1->re + arg1->im * arg1->im);

    return mag;
}

