/* Standard Includes */
#include <math.h>

/* IAS Library Includes */
#include "ias_math.h"
#include "ias_logging.h"

/*******************************************************************************
NAME: ias_math_merge_mean_and_stdev_samples

PURPOSE: Calculates the combined mean and standard deviations.

RETURNS: SUCCESS/ERROR
*******************************************************************************/
int ias_math_merge_mean_and_stdev_samples
(
    const int *counts_buffer, /* I: Counts per entry */
    const float *mean_buffer, /* I: Mean per entry */
    const float *stdev_buffer,/* I: Standard deviation per entry */
    int buffer_entry_count,   /* I: Total number of entries */
    float *combined_stdev,    /* O: Combined Standard deviation */
    float *combined_mean,     /* O: Combined Mean */
    float *combined_counts    /* O: Combined Count */
)
{
    int sum_counts = 0;                 /* Summation of counts */
    double sum_counts_and_means = 0;    /* Summation of counts and means */
    double mean_squared_entry = 0;      /* Mean squared per entry */
    double mean_squared = 0;            /* Mean squared of all entries */
    double stdev_squared_entry = 0;     /* Entry standard deviation squared */
    double product_count_and_sum_stdev_mean = 0;/* Summation of ith entry times
                                           sum of entry mean squared and the
                                           entry standard deviation squared */
    double sum_stdev_and_mean = 0;      /* Sum of standard deviation squared 
                                           and mean squared */
    double local_combined_mean = 0;     /* Local copy of combined_mean */
    int iteration;                      /* Detector loop variable */
   
    /* Check if counts_buffer is NULL */
    if (!counts_buffer)
    {
        IAS_LOG_ERROR("NULL pointer used for counts buffer");
        return ERROR;
    } 

    /* Check if mean_buffer is NULL */
    if (!mean_buffer)
    {
        IAS_LOG_ERROR("NULL pointer used for mean buffer");
        return ERROR;
    } 

    /* Check if stdev_buffer is NULL */
    if (!stdev_buffer)
    {
        IAS_LOG_ERROR("Null pointer used for standard deviation buffer");
        return ERROR;
    } 

    /* Running the calculations for mean, stdev, and counts */
    for (iteration = 0; iteration < buffer_entry_count; iteration++)
    {
        /* Summation of (count times mean) */
        sum_counts_and_means += counts_buffer[iteration]
            * mean_buffer[iteration];

        /* Summation of number of counts */
        sum_counts += counts_buffer[iteration];

        /* Standard deviation of ith stdev entry squared */
        stdev_squared_entry = stdev_buffer[iteration]
            * stdev_buffer[iteration];

        /* Mean of ith entry squared */
        mean_squared_entry = mean_buffer[iteration] * mean_buffer[iteration];

        /* Sum of standard deviation squared and mean squared */
        sum_stdev_and_mean = stdev_squared_entry + mean_squared_entry;
    
        /* Standard deviation and Mean times count of ith entry */
        product_count_and_sum_stdev_mean += sum_stdev_and_mean
            * counts_buffer[iteration];
    }
    
    /* Storing the combined mean and counts */
    local_combined_mean = sum_counts_and_means / (double) sum_counts;
    *combined_counts = (float) ((double) sum_counts
        / (double) buffer_entry_count);
    *combined_mean = (float) local_combined_mean;

    /* Mean squared */
    mean_squared = local_combined_mean * local_combined_mean;
   
    /* Storing the standard deviation. The equations is : the square root 
       of (summation of top left equation) minus (summation of counts times
       mean squared) divided by the summation of counts) */
    *combined_stdev =  (float) (sqrt((product_count_and_sum_stdev_mean - 
        (sum_counts * mean_squared)) / sum_counts));

    return SUCCESS;
}
