/*******************************************************************************
NAME: ias_math_point_in_polygon

PURPOSE: Given the vertices of a polygon determine if the point (x,y)
    falls within the polygon.  The point in polygon (PIP) method used 
    determines how many sides are crossed by a line formed by a point
    extending just above the highest y coordinate on the polygon.   The
    even-odd rule algorithm states that if the line crosses an odd
    number of sides of the polygon, The point lies inside the polygon.

RETURNS:
    TRUE The point was found in the polygon
    FALSE The point was not found in the polygon

NOTE:
    This algorthm will work for concave or convex polygons most of the time.
    The exception is for a concave polygon where the point's x-coordinate is 
    equal to an x-coordinate of one of the vertices.  The results of 
    this are erroneous and can not be trusted.  Until this is resolved, 
    this routine should be reserved for convex polygons only.

*******************************************************************************/
#include <stdlib.h>
#include "ias_math.h"
#include "ias_const.h"

int ias_math_point_in_polygon
(
    int vertice_count,      /* I: Number of vertices in x/y_vert */
    const double *x_vert,   /* I: Vertices of polygon */
    const double *y_vert,   /* I: Vertices of polygon */
    double x_coord,         /* I: X coordinate of point to check */
    double y_coord          /* I: Y coordinate of point to check */
)
{
    int point;
    int intflag = 0;    /* Flag denoting even (0) or odd (1) number of polygon
                           side intersections */

    /* A polygon must have at least 3 sides */
    if (vertice_count < 3)
    {
        return FALSE;
    }

    for (point = 0; point < vertice_count; point++)
    {
        /* Wrap around to the first point */
        int next_point = point + 1;
        if (next_point >= vertice_count)
            next_point -= vertice_count;

        if (((x_vert[point] > x_coord) != (x_vert[next_point] > x_coord))
            && (y_coord < (y_vert[next_point] - y_vert[point]) * (x_coord
            - x_vert[point]) / (x_vert[next_point] - x_vert[point])
            + y_vert[point]))
        {
            intflag = !intflag;
        }
    }

    return intflag;
}
