/******************************************************************************
NAME:           ias_math_simulated_annealing

PURPOSE:        
Source file for package of routines that perform the simulated annealing 
procedure. Derived from "Numerical Recipes in C". 

******************************************************************************/
#include <math.h>         /* Defines the math functions */
#include <stdlib.h>       /* Defines random number generation functions */

#include "ias_logging.h"  /* Defines the IAS logging routines */
#include "ias_math.h"     /* Define these functions */

#define GET_PSUM() \
    for (n = 0; n < ndim; n++) {\
    for (sum = 0.0, m = 0; m < mpts; m++) sum += p[m][n];\
    psum[n] = sum;}

static double ias_math_amotsa
(
    int npts,                      /* I: Number of points in profile */
    double **p,                    /* I/O: Simplex array of parameter vectors */
    double *y,                     /* I/O: Objective function vals for the
                                         simplex */
    double *psum,                  /* I: Array of parameter space sums */
    int ndim,                      /* I: Number of unknown parameters */
    double *pb,                    /* I/O: Array of "best" parameters */
    double *yb,                    /* I/O: Objective function value at pb */
    double (*eval_func)(int n, double a[], void *func_params),
                                   /* I: Pointer to the objective function */
    void *func_params,             /* I: Pointer to a structure for any 
                                         parameters that need to be passed to
                                         the eval function (can be NULL) */
    int ihi,                       /* I: Index of the "worst" point in the
                                         simplex */
    double *yhi,                   /* I/O: Objective function value at ihi */
    double fac,                    /* I: Extrapolation factor */
    double *ptry,                  /* O: Buffer for variable model parameters,
                                         modified but not used externally */
    double tt                      /* I: Annealing temperature parameter */
);

/******************************************************************************
NAME:           ias_math_ran1

PURPOSE:        
Replaces the "Numerical Recipes in C" ran1 function by using the C library 
"rand" function previously initialized via the srand function.  

RETURN VALUE:
Type = double
Value           Description
-----           -----------
double          Random number in the range (0.0,1.0)
                The value is exclusive at 0.0 and inclusive at 1.0.

******************************************************************************/
static double ias_math_ran1
(
    void
)
{
    double rnum;

    rnum = (double)(rand() + 1) / (double)(RAND_MAX + 1.0);
    return rnum;
}
  
/******************************************************************************
NAME:           ias_math_simulated_annealing

PURPOSE:        
Runs the simulated annealing iterations.  Derived from "Numerical Recipes in C"
routine amebsa.

RETURN VALUE:
Type = int
Value    Description
-----    -----------
SUCCESS  Successful completion
ERROR    Operation failed

******************************************************************************/
int ias_math_simulated_annealing
(
    int npts,                      /* I: Number of points in profile */
    double **p,                    /* I/O: Simplex array of parameter vectors */
    double *y,                     /* I/O: Objective function values for the 
                                         simplex */
    int ndim,                      /* I: Number of unknown parameters */
    double *pb,                    /* I/O: Array of "best" parameters */
    double *yb,                    /* I/O: Objective function value at pb */
    double ftol,                   /* I: Convergence tolerance */
    double (*eval_func)(int n, double a[], void *func_params),
                                   /* I: Pointer to the objective function */
    void *func_params,             /* I: Pointer to a structure for any 
                                         parameters that need to be passed to
                                         the eval function (can be NULL) */
    int *iter,                     /* I/O: Iteration limit/counter */
    double temptr                  /* I: Annealing temperature */
)
{
    int i;
    int ihi;
    int ilo;
    int j;
    int m;
    int n;
    int mpts;
    double rtol;
    double sum;
    double swap;
    double yhi;
    double ylo;
    double ynhi;
    double ysave;
    double yt;
    double ytry;
    double *psum;
    double tt;                     /* Annealing temperature parameter */
    double *ptry;                  /* Variable model parameters */

    mpts = ndim + 1;

    /* Allocate psum vector */
    psum = (double *)malloc(ndim * sizeof(double));
    if (psum == NULL)
    {
        IAS_LOG_ERROR("Error allocating memory for psum vector");
        return ERROR;
    }

    ptry = (double *)malloc(ndim * sizeof(double));
    if (ptry == NULL)
    {
        IAS_LOG_ERROR("Error allocating memory for ptry parameters");
        free(psum);
        return ERROR;
    }

    /* Initialize the global temperature variable tt */
    tt = -temptr;

    /* Sum up the simplex */
    GET_PSUM()

    /* Annealing loop */
    for (;;) 
    {
        /* Determine which point is the highest (worst), next highest, and 
           lowest (best) */
        ilo = 0;
        ihi = 1;

        /* Whenever we "look at" at vertex, it gets a random thermal 
           fluctuation */
        ynhi = ylo = y[0] + tt * log(ias_math_ran1());
        yhi = y[1] + tt * log(ias_math_ran1());
        if (ylo > yhi) 
        {
            ihi = 0;
            ilo = 1;
            ynhi = yhi;
            yhi = ylo;
            ylo = ynhi;
        }

        /* Loop over the points in the simplex */
        for (i = 2; i < mpts; i++) 
        {
            /* More thermal fluctuations */
            yt = y[i] + tt * log(ias_math_ran1());
            if (yt <= ylo) 
            {
                ilo = i;
                ylo = yt;
            }
            if (yt > yhi) 
            {
                ynhi = yhi;
                ihi = i;
                yhi = yt;
            } 
            else if (yt > ynhi) 
            {
                ynhi = yt;
            }
        }
        rtol = 2.0 * fabs(yhi - ylo) / (fabs(yhi) + fabs(ylo));

        /* Compute the fractional range from highest to lowest and return if
           satisfactory */
        if (rtol < ftol || *iter < 0) 
        {
            /* If returning, put the best point and value in slot 0 */
            swap = y[0];
            y[0] = y[ilo];
            y[ilo] = swap;
            for (n=0; n < ndim; n++) 
            {
                swap = p[0][n];
                p[0][n] = p[ilo][n];
                p[ilo][n] = swap;
            }
            break;
        }
        *iter -= 2;

        /* Begin a new iteration.
           First, extrapolate by a factor -1 through the face of the
           simplex across from the high point, i.e., reflect the simplex
           from the high point */
        ytry = ias_math_amotsa(npts, p, y, psum, ndim, pb, yb, eval_func,
            func_params, ihi, &yhi, -1.0, ptry, tt);
        if (ytry <= ylo) 
        {
            /* Gives a better result than the best point, so try an additional
               extrapolation by a factor of 2 */
            ytry = ias_math_amotsa(npts, p, y, psum, ndim, pb, yb, eval_func,
                func_params, ihi, &yhi, 2.0, ptry, tt);
        } 
        else if (ytry >= ynhi) 
        {
            /* The reflected point is worse than the second-highest, so look for
               an intermediate lower point, i.e., do a one-dimensional
               contraction. */
            ysave = yhi;
            ytry = ias_math_amotsa(npts, p, y, psum, ndim, pb, yb, eval_func,
                func_params, ihi, &yhi, 0.5, ptry, tt);
            if (ytry >= ysave) 
            {
                /* Can't seem to get rid of that high point.
                   Better contract around the lowest (best) point */
                for (i = 0; i < mpts; i++) 
                {
                    if (i != ilo) 
                    {
                        for (j = 0; j < ndim; j++) 
                        {
                            psum[j] = 0.5 * (p[i][j] + p[ilo][j]);
                            p[i][j] = psum[j];
                        }
                        y[i] = (*eval_func)(npts, psum, func_params);
                    }
                }
                *iter -= ndim;

                /* Recompute psum */
                GET_PSUM()
            }
        } 
        else 
        {
            /* Correct the evaluation count */
            ++(*iter);
        }
    }
    free(psum);
    free(ptry);

    return SUCCESS;
}

/******************************************************************************
NAME:           ias_math_amotsa

PURPOSE:        
Updates the simplex during the simulated annealing procedure.  Derived from
"Numerical Recipes in C".

RETURN VALUE:
Type = double
Value           Description
-----           -----------
double          Objective function value with thermal fluctuation

******************************************************************************/
static double ias_math_amotsa
(
    int npts,                      /* I: Number of points in profile */
    double **p,                    /* I/O: Simplex array of parameter vectors */
    double *y,                     /* I/O: Objective function vals for the 
                                         simplex */
    double *psum,                  /* I: Array of parameter space sums */
    int ndim,                      /* I: Number of unknown parameters */
    double *pb,                    /* I/O: Array of "best" parameters */
    double *yb,                    /* I/O: Objective function value at pb */
    double (*eval_func)(int n, double a[], void *func_params),
                                   /* I: Pointer to the objective function */
    void *func_params,             /* I: Pointer to a structure for any 
                                         parameters that need to be passed to
                                         the eval function (can be NULL) */
    int ihi,                       /* I: Index of the "worst" point in the 
                                         simplex */
    double *yhi,                   /* I/O: Objective function value at ihi */
    double fac,                    /* I: Extrapolation factor */
    double *ptry,                  /* I: Buffer for variable model parameters,
                                         modified but not used externally */
    double tt                      /* I: Annealing temperature parameter */
)
{
    int j;
    double fac1;
    double fac2;
    double yflu;
    double ytry;

    /* Extrapolates by a factor fac through the face of the simplex across from
       the high point, tries it, and replaces the high point if the new point
       is better. */
    fac1 = (1.0 - fac) / ndim;
    fac2 = fac1 - fac;

    for (j = 0; j < ndim; j++)
    {
        ptry[j] = psum[j] * fac1 - p[ihi][j] * fac2;
    }

    ytry = (*eval_func)(npts, ptry, func_params);

    if (ytry <= *yb) 
    {
        /* Save the best-ever */
        for (j = 0; j < ndim; j++)
        {
            pb[j] = ptry[j];
        }
        *yb = ytry;
    }

    /* We added a thermal fluctuation to all the current vertices, but we
       subtract it here, so as to give the simplex a thermal Brownian motion:
       It likes to accept any suggested change.  */
    yflu = ytry - tt * log(ias_math_ran1());
    if (yflu < *yhi)
    {
        y[ihi] = ytry;
        *yhi = yflu;
        for (j = 0; j < ndim; j++) 
        {
            psum[j] += ptry[j] - p[ihi][j];
            p[ihi][j] = ptry[j];
        }
    }

    return yflu;
}

