#include <string.h>
#include <stdlib.h>
#include <math.h>

#include "ias_logging.h"
#include "ias_const.h"
#include "ias_math.h"
#include "ias_math_complex.h"
#include "cgreen/cgreen.h"

IAS_MATH_LEAP_SECONDS_DATA leap_seconds_data;

void initialize_leap_seconds_struct(void)
{
    /* setup a few leap seconds to test with - totally fabricated data */
    memset(&leap_seconds_data, 0, sizeof(leap_seconds_data));

    leap_seconds_data.leap_seconds_count = 3;

    leap_seconds_data.leap_years = calloc(3, sizeof(int));
    leap_seconds_data.leap_months = calloc(3, sizeof(int));
    leap_seconds_data.leap_days = calloc(3, sizeof(int));
    leap_seconds_data.num_leap_seconds = calloc(3, sizeof(int));

    leap_seconds_data.leap_years[0] = 2000;
    leap_seconds_data.leap_months[0] = 4;
    leap_seconds_data.leap_days[0] = 24;
    leap_seconds_data.num_leap_seconds[0] = 1;

    leap_seconds_data.leap_years[1] = 2002;
    leap_seconds_data.leap_months[1] = 12;
    leap_seconds_data.leap_days[1] = 11;
    leap_seconds_data.num_leap_seconds[1] = 1;

    leap_seconds_data.leap_years[2] = 2003;
    leap_seconds_data.leap_months[2] = 7;
    leap_seconds_data.leap_days[2] = 2;
    leap_seconds_data.num_leap_seconds[2] = 1;
}

void free_leap_seconds_struct(void)
{
    free(leap_seconds_data.leap_months);
    free(leap_seconds_data.leap_years);
    free(leap_seconds_data.leap_days);
    free(leap_seconds_data.num_leap_seconds);
}

Ensure test_leap_seconds1()
{
    double seconds_from_j2000;
    double time[3];       /* Year, day of year (DOY), seconds of day (SOD) */
    int status;

    /* Only need to do this in the first leap seconds regression test. Once
       the struct has data in it, let all the other tests use it too. */
    initialize_leap_seconds_struct();

    /* 4 years, 190 days (considering leap years in 2000 & 2004) */
    /* Calendar date equivalent should be midnight between July 8/9, 2004 */
    /* Conversion to relevant to true J2000 epoch adds 12hours. Leap seconds
       initialization will report day 191 at noon because of this. */
    seconds_from_j2000
        /* 2000 (leap year) - 366 days, 2001-2003 - 3 years of 365 days */
        = 126230400
        /* 190 days into the year 2004 (leap year) */
        + 16416000;

    /* We set our time to midnight between July 8/9, 2004. Then adjust to be
       relevant to J2000 epoch, which is 1/1/00 at NOON, so this bumps us into
       the following day (July 9), at noon. */
    /* Should end up, after LS init, with epoch being 2004:191:43135.816
       and 3 leap seconds found. */
    status = ias_math_init_leap_seconds(seconds_from_j2000, &leap_seconds_data);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
    {
        free_leap_seconds_struct();
        return;
    }

    /* Should have removed the 3 leap seconds that occurred before the
       date/time being looked at. */
    /* Seconds should have gone to 43132.816 */
    status = ias_math_convert_j2000_seconds_to_year_doy_sod(seconds_from_j2000,
            time);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
    {
        free_leap_seconds_struct();
        return;
    }

    /* Validate the year, day of year, seconds of day values calculated */
    assert_equal_with_message((int)time[0], 2004, "J2000 conversion error - "
            "expected year 2004, got %d", (int)time[0]);
    assert_equal_with_message((int)time[1], 191, "J2000 conversion error - "
            "expected day of year 191, got %d", (int)time[1]);
    assert_double_equal_with_message(time[2], 43132.816, "J2000 conversion "
            "error - expected seconds of day 43132.816, got %lf", time[2]);

    /* Now try the conversion back. Should get the total seconds initialized
       above - 142,646,400 */
    status = ias_math_convert_year_doy_sod_to_j2000_seconds(time,
            &seconds_from_j2000);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
    {
        free_leap_seconds_struct();
        return;
    }

    /* Validate the seconds from J2000 */
    assert_double_equal_with_message(seconds_from_j2000, 142646400,
            "J2000 conversion error - expected J2000 seconds 142,646,400, "
            "got %lf", seconds_from_j2000);

    free_leap_seconds_struct();
}

Ensure test_leap_seconds2()
{
    double seconds_from_j2000;
    double time[3];       /* Year, day of year (DOY), seconds of day (SOD) */
    int status;

    initialize_leap_seconds_struct();

    /* 1 year, 27.25 days (considering leap year in 2000) */
    /* Calendar date equivalent should be 6am on January 27, 2001 */
    /* Conversion to relevant to true J2000 epoch adds 12hours. Leap seconds
       initialization will report day 27 at 6pm because of this. */
    seconds_from_j2000
        /* 2000 (leap year) - 366 days */
        = 31622400
        /* 27.25 days into the year 2001 */
        + 2354400;

    /* We set our time to 6am January 28, 2001. Then adjust to be relevant
       to J2000 epoch, which is 1/1/00 at NOON, so this bumps us to 6pm. */
    /* Should end up, after LS init, with epoch being 2001:028:64735.816
       and 1 leap second found. */
    status = ias_math_init_leap_seconds(seconds_from_j2000, &leap_seconds_data);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
    {
        free_leap_seconds_struct();
        return;
    }

    /* Should have removed the 1 leap second that occurred before the
       date/time being looked at. */
    /* Seconds should have gone to 64734.816 */
    status = ias_math_convert_j2000_seconds_to_year_doy_sod(seconds_from_j2000,
            time);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
    {
        free_leap_seconds_struct();
        return;
    }

    /* Validate the year, day of year, seconds of day values calculated */
    assert_equal_with_message((int)time[0], 2001, "J2000 conversion error - "
            "expected year 2001, got %d", (int)time[0]);
    assert_equal_with_message((int)time[1], 28, "J2000 conversion error - "
            "expected day of year 28, got %d", (int)time[1]);
    assert_double_equal_with_message(time[2], 64734.816, "J2000 conversion "
            "error - expected seconds of day 64734.816, got %lf", time[2]);

    /* Now try the conversion back. Should get the total seconds initialized
       above - 33,976,800 */
    status = ias_math_convert_year_doy_sod_to_j2000_seconds(time,
            &seconds_from_j2000);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
    {
        free_leap_seconds_struct();
        return;
    }

    /* Validate the seconds from J2000 */
    assert_double_equal_with_message(seconds_from_j2000, 33976800,
            "J2000 conversion error - expected J2000 seconds 33,976,800, "
            "got %lf", seconds_from_j2000);

    free_leap_seconds_struct();
}

Ensure test_leap_seconds3()
{
    double seconds_from_j2000;
    double time[3];       /* Year, day of year (DOY), seconds of day (SOD) */
    int status;

    initialize_leap_seconds_struct();

    /* 3 years, 2.75 days (considering leap year in 2000) */
    /* Calendar date equivalent should be 6pm on January 3, 2003 */
    /* Conversion to relevant to true J2000 epoch adds 12hours. Leap seconds
       initialization will report day 4 at 6am because of this. */
    seconds_from_j2000
        /* 2000 (leap year) - 366 days, 2001-2002 - 2 years of 365 days */
        = 94694400
        /* 2.75 days into the year 2003 */
        + 237600;

    /* We set our time to 6pm on January 3, 2003. Then adjust to be
       relevant to J2000 epoch, which is 1/1/00 at NOON, so this bumps us into
       the following day (January 4), at 6am. */
    /* Should end up, after LS init, with epoch being 2003:004:21535.816
       and 2 leap seconds found. */
    status = ias_math_init_leap_seconds(seconds_from_j2000, &leap_seconds_data);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
    {
        free_leap_seconds_struct();
        return;
    }

    /* Should have removed the 2 leap seconds that occurred before the
       date/time being looked at. */
    /* Seconds should have gone to 21533.816 */
    status = ias_math_convert_j2000_seconds_to_year_doy_sod(seconds_from_j2000,
            time);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
    {
        free_leap_seconds_struct();
        return;
    }

    /* Validate the year, day of year, seconds of day values calculated */
    assert_equal_with_message((int)time[0], 2003, "J2000 conversion error - "
            "expected year 2003, got %d", (int)time[0]);
    assert_equal_with_message((int)time[1], 4, "J2000 conversion error - "
            "expected day of year 4, got %d", (int)time[1]);
    assert_double_equal_with_message(time[2], 21533.816, "J2000 conversion "
            "error - expected seconds of day 21533.816, got %lf", time[2]);

    /* Now try the conversion back. Should get the total seconds initialized
       above - 94,932,000 */
    status = ias_math_convert_year_doy_sod_to_j2000_seconds(time,
            &seconds_from_j2000);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
    {
        free_leap_seconds_struct();
        return;
    }

    /* Validate the seconds from J2000 */
    assert_double_equal_with_message(seconds_from_j2000, 94932000,
            "J2000 conversion error - expected J2000 seconds 94,932,000, "
            "got %lf", seconds_from_j2000);

    free_leap_seconds_struct();
}

Ensure test_leap_seconds4()
{
    double seconds_from_j2000;
    double time[3];       /* Year, day of year (DOY), seconds of day (SOD) */
    int status;

    initialize_leap_seconds_struct();

    /* -2 years, -190 days */
    /* Calendar date equivalent should be midnight between June 24/25, 1997 */
    /* Conversion to relevant to true J2000 epoch adds 12hours. Leap seconds
       initialization will report day 176 at noon because of this. */
    seconds_from_j2000
        /* times a negative one to provide a negative value, indicating before
           J2000 epoch */
        = -1
        /* 1998-1999 - 2 years of 365 days */
        * (63072000
                /* 190 days back from end of the year 1999 */
                + 16416000);

    /* We set our time to midnight between June 24/25, 1997. Then adjust to be
       relevant to J2000 epoch, which is 1/1/00 at NOON, so this bumps us into
       the following day (June 25), at noon. */
    /* Should end up, after LS init, with epoch being 1997:176:43135.816
       and 0 leap seconds found. */
    status = ias_math_init_leap_seconds(seconds_from_j2000, &leap_seconds_data);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
    {
        free_leap_seconds_struct();
        return;
    }

    /* Don't need to remove any leap seconds, since none occurred before the
       date/time being looked at. */
    /* Seconds should have stayed the same at 43135.816 */
    status = ias_math_convert_j2000_seconds_to_year_doy_sod(seconds_from_j2000,
            time);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
    {
        free_leap_seconds_struct();
        return;
    }

    /* Validate the year, day of year, seconds of day values calculated */
    assert_equal_with_message((int)time[0], 1997, "J2000 conversion error - "
            "expected year 1997, got %d", (int)time[0]);
    assert_equal_with_message((int)time[1], 176, "J2000 conversion error - "
            "expected day of year 176, got %d", (int)time[1]);
    assert_double_equal_with_message(time[2], 43135.816, "J2000 conversion "
            "error - expected seconds of day 43135.816, got %lf", time[2]);

    /* Now try the conversion back. Should get the total seconds initialized
       above - -79,488,000 */
    status = ias_math_convert_year_doy_sod_to_j2000_seconds(time,
            &seconds_from_j2000);

    /* Validate the seconds from J2000 */
    assert_double_equal_with_message(seconds_from_j2000, -79488000,
            "J2000 conversion error - expected J2000 seconds -79,488,000, "
            "got %lf", seconds_from_j2000);

    free_leap_seconds_struct();
}

Ensure test_leap_seconds5()
{
    double seconds_from_j2000;
    double time[3];       /* Year, day of year (DOY), seconds of day (SOD) */
    int status;

    initialize_leap_seconds_struct();

    /* 7 years, 257.2 days (considering leap years in 2000 & 2004) */
    /* Calendar date equivalent should be noon on September 15, 2007 */
    /* Conversion to relevant to true J2000 epoch adds 12hours. Leap seconds
       initialization will report day 257.7 because of this. */
    seconds_from_j2000
        /* 2000 (leap year) - 366 days, 2001-2003 - 3 years of 365 days */
        /* 2004 (leap year) - 366 days, 2005-2006 - 2 years of 365 days */
        = 220924800
        /* 257.2 days into the year 2007 */
        + 22222080;

    /* We set our time to 4:48am, September 15, 2007. Then adjust to be
       relevant to J2000 epoch, which is 1/1/00 at NOON, so this bumps us to
       about 4:48pm. */
    /* Should end up, after LS init, with epoch being 2007:258:60415.816
       and 3 leap seconds found. */
    status = ias_math_init_leap_seconds(seconds_from_j2000, &leap_seconds_data);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
    {
        free_leap_seconds_struct();
        return;
    }

    /* Should have removed the 3 leap seconds that occurred before the
       date/time being looked at. */
    /* Seconds should have gone to 60412.816 */
    status = ias_math_convert_j2000_seconds_to_year_doy_sod(seconds_from_j2000,
            time);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
    {
        free_leap_seconds_struct();
        return;
    }

    /* Validate the year, day of year, seconds of day values calculated */
    assert_equal_with_message((int)time[0], 2007, "J2000 conversion error - "
            "expected year 2007, got %d", (int)time[0]);
    assert_equal_with_message((int)time[1], 258, "J2000 conversion error - "
            "expected day of year 258, got %d", (int)time[1]);
    assert_double_equal_with_message(time[2], 60412.816, "J2000 conversion "
            "error - expected seconds of day 60412.816, got %lf", time[2]);

    /* Now try the conversion back. Should get the total seconds initialized
       above - 243,146,880 */
    status = ias_math_convert_year_doy_sod_to_j2000_seconds(time,
            &seconds_from_j2000);

    /* Validate the seconds from J2000 */
    assert_double_equal_with_message(seconds_from_j2000, 243146880,
            "J2000 conversion error - expected J2000 seconds 243,146,880, "
            "got %lf", seconds_from_j2000);

    /* Same date as last test - no re-initialization of the leap seconds.
       Try another time stamp a minute after the last one, with the same
       leap seconds initialization settings. */
    /* 7 years, 257.2 days (considering leap years in 2000 & 2004) */
    /* Calendar date equivalent should be noon on September 15, 2007 */
    /* Conversion to relevant to true J2000 epoch adds 12hours. Leap seconds
       initialization will report day 257.7 because of this. */
    seconds_from_j2000
        /* 2000 (leap year) - 366 days, 2001-2003 - 3 years of 365 days */
        /* 2004 (leap year) - 366 days, 2005-2006 - 2 years of 365 days */
        = 220924800
        /* 257.2 days into the year 2007 */
        + 22222080
        /* plus one more minute, to simulate a second time conversion in
           same image */
        + 60;

    /* We set our time to 4:49am, September 15, 2007. Then adjust to be
       relevant to J2000 epoch, which is 1/1/00 at NOON, so this bumps us to
       about 4:49pm. */
    /* Should end up, after LS init, with epoch being 2007:258:60475.816
       and 3 leap seconds found. */
    status = ias_math_init_leap_seconds(seconds_from_j2000, &leap_seconds_data);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
    {
        free_leap_seconds_struct();
        return;
    }

    /* Should have removed the 3 leap seconds that occurred before the
       date/time being looked at. */
    /* Seconds should have gone to 60472.816 */
    status = ias_math_convert_j2000_seconds_to_year_doy_sod(seconds_from_j2000,
            time);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
    {
        free_leap_seconds_struct();
        return;
    }

    /* Validate the year, day of year, seconds of day values calculated */
    assert_equal_with_message((int)time[0], 2007, "J2000 conversion error - "
            "expected year 2007, got %d", (int)time[0]);
    assert_equal_with_message((int)time[1], 258, "J2000 conversion error - "
            "expected day of year 258, got %d", (int)time[1]);
    assert_double_equal_with_message(time[2], 60472.816, "J2000 conversion "
            "error - expected seconds of day 60472.816, got %lf", time[2]);

    /* Now try the conversion back. Should get the total seconds initialized
       above - 243,146,940 */
    status = ias_math_convert_year_doy_sod_to_j2000_seconds(time,
            &seconds_from_j2000);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
    {
        free_leap_seconds_struct();
        return;
    }

    /* Validate the seconds from J2000 */
    assert_double_equal_with_message(seconds_from_j2000, 243146940,
            "J2000 conversion error - expected J2000 seconds 243,146,940, "
            "got %lf", seconds_from_j2000);

    free_leap_seconds_struct();
}

Ensure test_leap_seconds6()
{
    double seconds_from_j2000;
    double seconds_from_j2000_2;
    int status;

    initialize_leap_seconds_struct();

    /* Chose two dates that result in different leap seconds adjustments.
       Make sure the second call to init fails. */

    /* 7 years, 257.2 days (considering leap years in 2000 & 2004) */
    /* Calendar date equivalent should be noon on September 15, 2007 */
    /* Conversion to relevant to true J2000 epoch adds 12hours. Leap seconds
       initialization will report day 257.7 because of this. */
    seconds_from_j2000
        /* 2000 (leap year) - 366 days, 2001-2003 - 3 years of 365 days */
        /* 2004 (leap year) - 366 days, 2005-2006 - 2 years of 365 days */
        = 220924800
        /* 257.2 days into the year 2007 */
        + 22222080;

    /* 3 years, 2.75 days (considering leap year in 2000) */
    /* Calendar date equivalent should be 6pm on January 3, 2003 */
    /* Conversion to relevant to true J2000 epoch adds 12hours. Leap seconds
       initialization will report day 4 at 6am because of this. */
    seconds_from_j2000_2
        /* 2000 (leap year) - 366 days, 2001-2002 - 2 years of 365 days */
        = 94694400
        /* 2.75 days into the year 2003 */
        + 237600;

    /* Should result in 3 leap seconds found. */
    status = ias_math_init_leap_seconds(seconds_from_j2000, &leap_seconds_data);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
    {
        free_leap_seconds_struct();
        return;
    }

    /* Disable all output logging since this test hits an error condition */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* Should result in 2 leap seconds found and an error. */
    status = ias_math_init_leap_seconds(seconds_from_j2000_2,
            &leap_seconds_data);
    assert_equal(status, ERROR);
    if (status != ERROR)
    {
        free_leap_seconds_struct();
        return;
    }

    /* Re-enable the logging */
    ias_log_set_output_level(IAS_LOG_LEVEL_WARN);

    free_leap_seconds_struct();
}

Ensure test_leap_seconds_init_from_UTC()
{
    int adjustment;
    double date[3];
    int status;

    initialize_leap_seconds_struct();

    /* initialize the leap seconds using a UTC time of September 15, 2007 */
    date[0] = 2007.0;
    date[1] = 258.0;
    date[2] = 20.0;
    status = ias_math_init_leap_seconds_from_UTC_time(date, &leap_seconds_data);
    assert_equal(status, SUCCESS);

    /* get the adjustment and confirm it is three */
    status = ias_math_get_leap_second_adjustment(&adjustment);
    assert_equal(status, SUCCESS);
    assert_equal(adjustment, 3);

    /* make sure a second call to init succeeds and still gets the same
       adjustment */
    status = ias_math_init_leap_seconds_from_UTC_time(date, &leap_seconds_data);
    assert_equal(status, SUCCESS);
    status = ias_math_get_leap_second_adjustment(&adjustment);
    assert_equal(status, SUCCESS);
    assert_equal(adjustment, 3);

    /* verify initializing it with a different value results in an error */
    date[0] = 2001.0;
    date[1] = 258.0;
    date[2] = 20.0;
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = ias_math_init_leap_seconds_from_UTC_time(date, &leap_seconds_data);
    assert_equal(status, ERROR);
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

    /* clear the leap seconds and make sure it indicates it isn't initialized
       now */
    ias_math_clear_leap_seconds();
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = ias_math_get_leap_second_adjustment(&adjustment);
    assert_equal(status, ERROR);
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

    /* now test initializing on either side of a leap second */
    date[0] = 2002.0;
    date[1] = 344.0;
    date[2] = 86399.0;
    status = ias_math_init_leap_seconds_from_UTC_time(date, &leap_seconds_data);
    assert_equal(status, SUCCESS);
    /* get the adjustment and confirm it is 1 */
    status = ias_math_get_leap_second_adjustment(&adjustment);
    assert_equal(status, SUCCESS);
    assert_equal(adjustment, 1);

    ias_math_clear_leap_seconds();
    date[0] = 2002.0;
    date[1] = 345.0;
    date[2] = 0.0;
    status = ias_math_init_leap_seconds_from_UTC_time(date, &leap_seconds_data);
    assert_equal(status, SUCCESS);
    /* get the adjustment and confirm it is 2 */
    status = ias_math_get_leap_second_adjustment(&adjustment);
    assert_equal(status, SUCCESS);
    assert_equal(adjustment, 2);

    /* note, no reason to actually do any conversion tests here since the
       conversions are no different when it has been initialized from a
       UTC time. */

    /* free the leap second structure */
    free_leap_seconds_struct();
}

/* additional test to make sure the repeated calling of the init routine
   doesn't have a problem */
Ensure test_leap_seconds_init()
{
    double seconds_from_j2000;
    int adjustment;
    double date[3];
    int status;

    initialize_leap_seconds_struct();

    /* first initialize things for one second before a leap second */
    date[0] = 2003.0;
    date[1] = 182.0;
    date[2] = 86399.0;
    status = ias_math_init_leap_seconds_from_UTC_time(date, &leap_seconds_data);
    assert_equal(status, SUCCESS);

    /* get the adjustment and confirm it is two */
    status = ias_math_get_leap_second_adjustment(&adjustment);
    assert_equal(status, SUCCESS);
    assert_equal(adjustment, 2);

    /* call it again to make sure it works with the same date */
    status = ias_math_init_leap_seconds_from_UTC_time(date, &leap_seconds_data);
    assert_equal(status, SUCCESS);

    /* convert the UTC date to J2000 */
    status = ias_math_convert_year_doy_sod_to_j2000_seconds(date,
            &seconds_from_j2000);
    assert_equal(status, SUCCESS);

    /* now initialize the leap seconds again with the J2000 seconds */
    status = ias_math_init_leap_seconds(seconds_from_j2000, &leap_seconds_data);
    assert_equal(status, SUCCESS);

    /* add one second to move into the next leap second and make sure the 
       initialization now fails */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    seconds_from_j2000 += 1.0;
    status = ias_math_init_leap_seconds(seconds_from_j2000, &leap_seconds_data);
    assert_equal(status, ERROR);
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

    /* free the leap second structure */
    free_leap_seconds_struct();
}

Ensure test_constants()
{
    double pi;
    double arc2rad;
    double radians_per_degree;
    double degrees_per_radian;

    /* test pi */
    pi = ias_math_get_pi();
    assert_double_equal(pi, 3.141592653589793238);

    /* do it again to make sure the caching doesn't mess something up */
    pi = ias_math_get_pi();
    assert_double_equal(pi, 3.141592653589793238);

    /* test radians per degree */
    radians_per_degree = ias_math_get_radians_per_degree();
    assert_double_equal(radians_per_degree,
                        (3.141592653589793238 / 180.0));

    /* and again for the caching */
    radians_per_degree = ias_math_get_radians_per_degree();
    assert_double_equal(radians_per_degree,
                        (3.141592653589793238 / 180.0));

    /* test degrees per radian */
    degrees_per_radian = ias_math_get_degrees_per_radian();
    assert_double_equal(degrees_per_radian,
                        (180.0 / 3.141592653589793238));

    /* and again for the caching */
    degrees_per_radian = ias_math_get_degrees_per_radian();
    assert_double_equal(degrees_per_radian,
                        (180.0 / 3.141592653589793238));

    /* test the arc2rad conversion */
    arc2rad = ias_math_get_arcsec_to_radian_conversion();
    assert_double_equal(arc2rad, 4.848136811e-6);

    /* and again for the caching */
    arc2rad = ias_math_get_arcsec_to_radian_conversion();
    assert_double_equal(arc2rad, 4.848136811e-6);
}

Ensure test_legendre()
{
    /* test with an x value of 2 and coefficients of 2,4,6,8 to
       make calculating by hand easy */
    double ret_value;
    double coefficients[] = {2, 4, 6, 8};

    ret_value = ias_math_eval_legendre(2, coefficients, 1);
    assert_double_equal(ret_value, 2.000000e+00);

    ret_value = ias_math_eval_legendre(2, coefficients, 2);
    assert_double_equal(ret_value, 1.000000e+01);

    ret_value = ias_math_eval_legendre(2, coefficients, 3);
    assert_double_equal(ret_value, 4.300000e+01);

    ret_value = ias_math_eval_legendre(2, coefficients, 4);
    assert_double_equal(ret_value, 1.790000e+02);

}

Ensure test_add_seconds_to_year_doy_sod()
{
    double date[3];

    /******  Tests to add seconds ********/

    /* test a general date */
    date[0] = 2007.0;
    date[1] = 1.0;
    date[2] = 0.0;
    ias_math_add_seconds_to_year_doy_sod(10.0, date);
    assert_double_equal(date[0], 2007.0);
    assert_double_equal(date[1], 1.0);
    assert_double_equal(date[2], 10.0);

    /* test adding more than a day */
    date[0] = 2007.0;
    date[1] = 1.0;
    date[2] = 20.5;
    ias_math_add_seconds_to_year_doy_sod(IAS_SEC_PER_DAY + 10.0, date);
    assert_double_equal(date[0], 2007.0);
    assert_double_equal(date[1], 2.0);
    assert_double_equal(date[2], 30.5);

    /* test adding more than a day to the second last day of the year */
    date[0] = 2007.0;
    date[1] = IAS_DAYS_IN_YEAR - 1;
    date[2] = 20.5;
    ias_math_add_seconds_to_year_doy_sod(IAS_SEC_PER_DAY + 10.0, date);
    assert_double_equal(date[0], 2007.0);
    assert_double_equal(date[1], IAS_DAYS_IN_YEAR);
    assert_double_equal(date[2], 30.5);

    /* test adding less than a day to the last day of the year */
    date[0] = 2007.0;
    date[1] = IAS_DAYS_IN_YEAR;
    date[2] = 20.5;
    ias_math_add_seconds_to_year_doy_sod(10.0, date);
    assert_double_equal(date[0], 2007.0);
    assert_double_equal(date[1], IAS_DAYS_IN_YEAR);
    assert_double_equal(date[2], 30.5);

    /* test rolling over to a new year */
    date[0] = 2007.0;
    date[1] = IAS_DAYS_IN_YEAR;
    date[2] = 20.5;
    ias_math_add_seconds_to_year_doy_sod(IAS_SEC_PER_DAY + 10.0, date);
    assert_double_equal(date[0], 2008.0);
    assert_double_equal(date[1], 1);
    assert_double_equal(date[2], 30.5);

    /* test support for extra day in leap year not rolling over */
    date[0] = 2008.0;
    date[1] = IAS_DAYS_IN_YEAR;
    date[2] = 20.5;
    ias_math_add_seconds_to_year_doy_sod(IAS_SEC_PER_DAY + 10.0, date);
    assert_double_equal(date[0], 2008.0);
    assert_double_equal(date[1], 366.0);
    assert_double_equal(date[2], 30.5);

    /******  Tests to subtract seconds ********/

    /* test a general date */
    date[0] = 2007.0;
    date[1] = 1.0;
    date[2] = 20.0;
    ias_math_add_seconds_to_year_doy_sod(-10.0, date);
    assert_double_equal(date[0], 2007.0);
    assert_double_equal(date[1], 1.0);
    assert_double_equal(date[2], 10.0);

    /* test subtracting enough to roll to a previous day */
    date[0] = 2007.0;
    date[1] = 2.0;
    date[2] = 20.5;
    ias_math_add_seconds_to_year_doy_sod(-30.0, date);
    assert_double_equal(date[0], 2007.0);
    assert_double_equal(date[1], 1.0);
    assert_double_equal(date[2], IAS_SEC_PER_DAY + 20.5 - 30.0);

    /* test subtracting enough to roll to a previous year from the first day
       of the current year */
    date[0] = 2007.0;
    date[1] = 1.0;
    date[2] = 20.5;
    ias_math_add_seconds_to_year_doy_sod(-30.0, date);
    assert_double_equal(date[0], 2006.0);
    assert_double_equal(date[1], IAS_DAYS_IN_YEAR);
    assert_double_equal(date[2], IAS_SEC_PER_DAY + 20.5 - 30.0);

    /* test subtracting enough to roll back a couple days to the previous
       year */
    date[0] = 2007.0;
    date[1] = 2.0;
    date[2] = 20.5;
    ias_math_add_seconds_to_year_doy_sod(-(IAS_SEC_PER_DAY + 30.0), date);
    assert_double_equal(date[0], 2006.0);
    assert_double_equal(date[1], IAS_DAYS_IN_YEAR);
    assert_double_equal(date[2], IAS_SEC_PER_DAY + 20.5 - 30.0);

    /* test rolling back to a leap year */
    date[0] = 2009.0;
    date[1] = 1.0;
    date[2] = 20.5;
    ias_math_add_seconds_to_year_doy_sod(-30.0, date);
    assert_double_equal(date[0], 2008.0);
    assert_double_equal(date[1], IAS_DAYS_IN_LEAP_YEAR);
    assert_double_equal(date[2], IAS_SEC_PER_DAY + 20.5 - 30.0);
}

Ensure test_compute_unit_vector()
{
    IAS_VECTOR vec;
    IAS_VECTOR unit_vector;
    int status;

    /* test a vector that should succeed */
    vec.x = 2.0;
    vec.y = 3.0;
    vec.z = 4.0;

    status = ias_math_compute_unit_vector(&vec, &unit_vector);
    assert_equal(status, SUCCESS);
    assert_double_equal(unit_vector.x, 0.371390676);
    assert_double_equal(unit_vector.y, 0.557086015);
    assert_double_equal(unit_vector.z, 0.742781353);

    /* make sure a zero length vector fails */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    vec.x = 0.0;
    vec.y = 0.0;
    vec.z = 0.0;
    status = ias_math_compute_unit_vector(&vec, &unit_vector);
    assert_equal(status, ERROR);

    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);
}

Ensure test_line_intersection()
{
    double x;
    double y;
    int status;
    IAS_LINE_SEGMENT line1;
    IAS_LINE_SEGMENT line2;

    /* Test intersecting lines */
    line1.x1 = -1.0;
    line1.x2 = 6941.0;
    line1.y1 = 2.0;
    line1.y2 = 2.0;
    line2.x1 = 6940.0;
    line2.x2 = 6940.0;
    line2.y1 = 0.0;
    line2.y2 = 4390.0;

    status = ias_math_find_line_segment_intersection(&line1, &line2, &x, &y);
    assert_equal(status, IAS_LINES_INTERSECT);
    assert_double_equal(x, 6940.0);
    assert_double_equal(y, 2.0);

    line2.x1 = 0.0;
    line2.x2 = 0.0;
    line2.y1 = 4390.0;
    line2.y2 = 0.0;
    status = ias_math_find_line_segment_intersection(&line1, &line2, &x, &y);
    assert_equal(status, IAS_LINES_INTERSECT);
    assert_double_equal(x, 0.0);
    assert_double_equal(y, 2.0);

    /* Test lines that do not intersect */
    line2.x1 = 6940.0;
    line2.x2 = 0.0;
    line2.y1 = 4391.0;
    line2.y2 = 4390.0;
    status = ias_math_find_line_segment_intersection(&line1, &line2, &x, &y);
    assert_equal(status, IAS_LINES_DO_NOT_INTERSECT);

    line1.x1 = -1.0;
    line1.x2 = 6941.0;
    line1.y1 = 3.0;
    line1.y2 = 3.0;
    line2.x1 = 10.0;
    line2.x2 = 10.0;
    line2.y1 = 0.0;
    line2.y2 = 2.9;
    status = ias_math_find_line_segment_intersection(&line1, &line2, &x, &y);
    assert_equal(status, IAS_LINES_DO_NOT_INTERSECT);

    /* Test parallel lines */
    line2.x1 = 10.0;
    line2.x2 = 50.0;
    line2.y1 = 2.0;
    line2.y2 = 2.0;
    status = ias_math_find_line_segment_intersection(&line1, &line2, &x, &y);
    assert_equal(status, IAS_LINES_PARALLEL);
}

Ensure test_quaternion_conversions()
{
    IAS_QUATERNION original_quaternion; /* Starting quaternion */
    double euler_matrix[3][3];          /* Euler matrix */ 
    IAS_QUATERNION final_quaternion;    /* This should match the original */
    int status;
    int i;
    int j;

    /* This expected euler matrix was found by running the same quaternion
       through the cal/val version of the routine */
    double expected_euler_matrix[3][3] =
        {
            {7.728663176e-01, 4.427573473e-01, -4.545805698e-01},
            {5.587760610e-01, -8.143516721e-01, 1.568457062e-01},
            {-3.007438722e-01, -3.752295209e-01, -8.767872249e-01}
        };

    /* Initialize original quaternion */
    original_quaternion.vector.x = 9.305919041830e-01; 
    original_quaternion.vector.y = 2.690581671096e-01; 
    original_quaternion.vector.z = -2.029150583186e-01; 
    original_quaternion.scalar = 0.142940; 

    /* Convert the original quaternion to an Euler matrix */
    ias_math_convert_quaternion_to_euler(&original_quaternion, euler_matrix);

    /* Verify the expected euler matrix was returned */
    for (i = 0; i < 3; i++)
    {
        for (j= 0; j < 3; j++)
        {
            assert_double_equal(euler_matrix[i][j],
                expected_euler_matrix[i][j]);
        }
    }

    /* Convert the intermediate matrix to a quaternion */
    status = ias_math_convert_euler_to_quaternion(1e-6, euler_matrix, 
        &final_quaternion);
    assert_equal(status, SUCCESS);

    /* Verify that the final quaternion is the same as the original */
    if (status == SUCCESS)
    {
        assert_double_equal(original_quaternion.vector.x, 
            final_quaternion.vector.x);
        assert_double_equal(original_quaternion.vector.y, 
            final_quaternion.vector.y);
        assert_double_equal(original_quaternion.vector.z, 
            final_quaternion.vector.z);
        assert_double_equal(original_quaternion.scalar, 
            final_quaternion.scalar);
    }

    /* Conjugating the original quaternion */ 
    ias_math_conjugate_quaternion(&original_quaternion, &final_quaternion);

    /* Verify that the final quaternion is the same as the original but 
       with the vector values negated */
    assert_double_equal(original_quaternion.vector.x, 
        -final_quaternion.vector.x);
    assert_double_equal(original_quaternion.vector.y, 
        -final_quaternion.vector.y);
    assert_double_equal(original_quaternion.vector.z, 
        -final_quaternion.vector.z);
    assert_double_equal(original_quaternion.scalar, final_quaternion.scalar);
}

Ensure test_point_in_polygon(void)
{
    int nsides;                 /* number of sides of the polygon */
    int status;                 /* output status of the call */
    double x_vertices[7];       /* x coordinates for a polygon */
    double y_vertices[7];       /* y coordinates for a polygon */
    double x_coordinate;        /* The x coordinate to be tested */
    double y_coordinate;        /* The y coordinate to be tested */

    /* set up a 5 sided polygon with no vertical sides */

    nsides = 5;

    /* point 1 */
    x_vertices[0] = 2.0;
    y_vertices[0] = 6.0;
    /* point 2 */
    x_vertices[1] = 7.0;
    y_vertices[1] = 8.0;
    /* point 3 */
    x_vertices[2] = 9.0;
    y_vertices[2] = 5.0;
    /* point 4 */
    x_vertices[3] = 7.0;
    y_vertices[3] = 2.0;
    /* point 5 */
    x_vertices[4] = 3.0;
    y_vertices[4] = 3.0;

    /* assign a point inside the polygon */
    x_coordinate = 5.0;
    y_coordinate = 4.0;

    status = ias_math_point_in_polygon(nsides, x_vertices, y_vertices,
        x_coordinate, y_coordinate);

    assert_true_with_message(status, 
        "The point (%f,%f) should be in the polygon",
        x_coordinate, y_coordinate);

    /* assign a point outside the polygon that crosses the polygon */
    x_coordinate = 2.5;
    y_coordinate = 3.0;

    status = ias_math_point_in_polygon(nsides, x_vertices, y_vertices,
        x_coordinate, y_coordinate);

    assert_false_with_message(status, 
        "The point (%f,%f) should not be in the polygon",
        x_coordinate, y_coordinate);


    /* assign a point outside the polygon that does not cross the polygon */
    x_coordinate = 3.0;
    y_coordinate = 1.0;

    status = ias_math_point_in_polygon(nsides, x_vertices, y_vertices,
        x_coordinate, y_coordinate);

    assert_false_with_message(status, 
        "The point (%f,%f) should not be in the polygon",
        x_coordinate, y_coordinate);

    /* assign a point inside the polygon that does not cross the polygon */
    x_coordinate = 7.0;
    y_coordinate = 5.0;

    status = ias_math_point_in_polygon(nsides, x_vertices, y_vertices,
        x_coordinate, y_coordinate);

    assert_true_with_message(status, 
        "The point (%f,%f) should be in the polygon",
        x_coordinate, y_coordinate);


    /* set up a 7 sided polygon with a vertical and horizontal side */
    nsides = 7;

    /* point 1 */
    x_vertices[0] = 3.0;
    y_vertices[0] = 6.0;
    /* point 2 */
    x_vertices[1] = 7.0;
    y_vertices[1] = 10.0;
    /* point 3 */
    x_vertices[2] = 12.0;
    y_vertices[2] = 8.0;
    /* point 4 */
    x_vertices[3] = 13.0;
    y_vertices[3] = 6.0;
    /* point 5 */
    x_vertices[4] = 13.0;
    y_vertices[4] = 4.0;
    /* point 6 */
    x_vertices[5] = 6.0;
    y_vertices[5] = 2.0;
    /* point 7 */
    x_vertices[6] = 3.0;
    y_vertices[6] = 2.0;

    /* assign a point inside the polygon */
    x_coordinate = 7.0;
    y_coordinate = 6.0;

    status = ias_math_point_in_polygon(nsides, x_vertices, y_vertices,
        x_coordinate, y_coordinate);

    assert_true_with_message(status, 
        "The point (%f,%f) should be in the polygon",
        x_coordinate, y_coordinate);

    /* assign a point inside the polygon */
    x_coordinate = 12.0;
    y_coordinate = 6.0;

    status = ias_math_point_in_polygon(nsides, x_vertices, y_vertices,
        x_coordinate, y_coordinate);

    assert_true_with_message(status, 
        "The point (%f,%f) should be in the polygon",
        x_coordinate, y_coordinate);

    /* assign a point outside the polygon that crosses the polygon */
    x_coordinate = 12.0;
    y_coordinate = 3.0;

    status = ias_math_point_in_polygon(nsides, x_vertices, y_vertices,
        x_coordinate, y_coordinate);

    assert_false_with_message(status, 
        "The point (%f,%f) should not be in the polygon",
        x_coordinate, y_coordinate);

    /* assign a point outside the polygon that crosses the polygon */
    x_coordinate = 7.0;
    y_coordinate = 2.0;

    status = ias_math_point_in_polygon(nsides, x_vertices, y_vertices,
        x_coordinate, y_coordinate);

    assert_false_with_message(status, 
        "The point (%f,%f) should not be in the polygon",
        x_coordinate, y_coordinate);

    /* assign a point outside the polygon that crosses the polygon */
    x_coordinate = 3.0;
    y_coordinate = 1.0;

    status = ias_math_point_in_polygon(nsides, x_vertices, y_vertices,
        x_coordinate, y_coordinate);

    assert_false_with_message(status, 
        "The point (%f,%f) should not be in the polygon",
        x_coordinate, y_coordinate);
}

Ensure test_calculate_polygon_centroid()
{
    double x[5];    /* Array to hold hard coded x points */
    double y[5];    /* Array to hold hard coded y points */

    double calculated_x; /* Hold calculated centroid  x coordinate*/
    double calculated_y; /* Hold calculated centroid  y coordinate*/

    double expected_x; /* Hold hard coded centroid  x coordinate*/
    double expected_y; /* Hold hard coded centroid y coordinate*/

    x[0] = 2;
    y[0] = 1;

    x[1] = 5;
    y[1] = 4;

    x[2] = 1;
    y[2] = -10;

    x[3] = -3;
    y[3] = 4;

    x[4] = -0;
    y[4] = 1;

    expected_x = 1;
    expected_y = -1.93495935;

    assert_equal(
        ias_math_calculate_polygon_centroid(
            x, y, 5, &calculated_x, &calculated_y),
            0);

    assert_double_equal(calculated_x, expected_x);
    assert_double_equal(calculated_y, expected_y);
}

Ensure test_complex()
{
    IAS_COMPLEX input_arg1 = { 0, 0 };
    IAS_COMPLEX input_arg2 = { 0, 0 };
    IAS_COMPLEX return_val;

    /* test ias_math_complex_exponential() with {0.0, 0.0} input, 
       result should be {1.0, 0.0} */
    ias_math_complex_exponential(&input_arg1, &return_val);
    assert_double_equal(return_val.re, 1.0);
    assert_double_equal(return_val.im, 0.0);

    /* test ias_math_complex_exponential() with {0.0, pi} input, 
       result should be {-1.0, 0.0} */
    input_arg1.im = ias_math_get_pi();
    ias_math_complex_exponential(&input_arg1, &return_val);
    assert_double_equal(return_val.re, -1.0);
    /* Check for the value being essentially zero.  assert_double_equal
       doesn't work well for comparing to zero since 8 signficant digits
       can be changed in the E-20 range depending on the compiler. */
    assert_true(fabs(return_val.im) < 1e-9);

    /* test ias_math_complex_exponential() with {0.0, pi} input, 
       result should be {-1.0, 0.0} */
    input_arg1.im = ias_math_get_pi();
    ias_math_complex_exponential(&input_arg1, &return_val);
    assert_double_equal(return_val.re, -1.0);
    assert_true(fabs(return_val.im) < 1e-9);

    /* test ias_math_complex_exponential() with {0.0, pi/2} input, 
       result should be {0.0, 1.0} */
    input_arg1.im = ias_math_get_pi() / 2;
    ias_math_complex_exponential(&input_arg1, &return_val);
    assert_true(fabs(return_val.re) < 1e-9);
    assert_double_equal(return_val.im, 1.0);

    /* test ias_math_complex_exponential() with {2.0, -pi/4} input, 
       result should be {5.224851674, -5.224851674} */

    input_arg1.re = 2.0;
    input_arg1.im = -1 * ias_math_get_pi() / 4;
    ias_math_complex_exponential(&input_arg1, &return_val);
    assert_double_equal(return_val.re, 5.224851674);
    assert_double_equal(return_val.im, -5.224851674);

    /* test ias_math_complex_multiply with {0.0,0.0} and {1.0,0.0}, 
       result should be {0.0,0.0} */
    input_arg1.re = 0.0;
    input_arg1.im = 0.0;
    input_arg2.re = 1.0;
    input_arg2.im = 0.0;
    ias_math_complex_multiply(&input_arg1, &input_arg2, &return_val);
    assert_double_equal(return_val.re, 0.0);
    assert_double_equal(return_val.im, 0.0);

    /* test ias_math_complex_multiply with {1.0,1.0} and {1.0,0.0}, 
       result should be {1.0,1.0} */
    input_arg1.re = 1.0;
    input_arg1.im = 1.0;
    input_arg2.re = 1.0;
    input_arg2.im = 0.0;
    ias_math_complex_multiply(&input_arg1, &input_arg2, &return_val);
    assert_double_equal(return_val.re, 1.0);
    assert_double_equal(return_val.im, 1.0);

    /* test ias_math_complex_multiply with {1.0,1.0} and {1.0,1.0}, 
       result should be {0.0,2.0} */
    input_arg1.re = 1.0;
    input_arg1.im = 1.0;
    input_arg2.re = 1.0;
    input_arg2.im = 1.0;
    ias_math_complex_multiply(&input_arg1, &input_arg2, &return_val);
    assert_double_equal(return_val.re, 0.0);
    assert_double_equal(return_val.im, 2.0);

    /* test ias_math_complex_multiply with {1.0,2.0} and {-1.0,2.0}, 
       result should be {-5.0,0.0} */
    input_arg1.re = 1.0;
    input_arg1.im = 2.0;
    input_arg2.re = -1.0;
    input_arg2.im = 2.0;
    ias_math_complex_multiply(&input_arg1, &input_arg2, &return_val);
    assert_double_equal(return_val.re, -5.0);
    assert_double_equal(return_val.im, 0.0);

    /* test ias_math_complex_multiply with {-2.0,1.0} and {-2.0,2.0}, 
       result should be {2.0,-6.0} */
    input_arg1.re = -2.0;
    input_arg1.im = 1.0;
    input_arg2.re = -2.0;
    input_arg2.im = 2.0;
    ias_math_complex_multiply(&input_arg1, &input_arg2, &return_val);
    assert_double_equal(return_val.re, 2.0);
    assert_double_equal(return_val.im, -6.0);


    /* test ias_math_complex_divide with {0.0,0.0} and {1.0,0.0}, 
       result should be {0.0,0.0} */
    input_arg1.re = 0.0;
    input_arg1.im = 0.0;
    input_arg2.re = 1.0;
    input_arg2.im = 0.0;
    ias_math_complex_divide(&input_arg1, &input_arg2, &return_val);
    assert_double_equal(return_val.re, 0.0);
    assert_double_equal(return_val.im, 0.0);

    /* test ias_math_complex_divide with {1.0,1.0} and {1.0,0.0}, 
       result should be {1.0,1.0} */
    input_arg1.re = 1.0;
    input_arg1.im = 1.0;
    input_arg2.re = 1.0;
    input_arg2.im = 0.0;
    ias_math_complex_divide(&input_arg1, &input_arg2, &return_val);
    assert_double_equal(return_val.re, 1.0);
    assert_double_equal(return_val.im, 1.0);

    /* test ias_math_complex_divide with {1.0,1.0} and {1.0,1.0}, 
       result should be {1.0,0.0} */
    input_arg1.re = 1.0;
    input_arg1.im = 1.0;
    input_arg2.re = 1.0;
    input_arg2.im = 1.0;
    ias_math_complex_divide(&input_arg1, &input_arg2, &return_val);
    assert_double_equal(return_val.re, 1.0);
    assert_double_equal(return_val.im, 0.0);

    /* test ias_math_complex_divide with {1.0,2.0} and {-1.0,2.0}, 
       result should be {0.6,0.8} */
    input_arg1.re = 1.0;
    input_arg1.im = 2.0;
    input_arg2.re = -1.0;
    input_arg2.im = 2.0;
    ias_math_complex_divide(&input_arg1, &input_arg2, &return_val);
    assert_double_equal(return_val.re, 0.6);
    assert_double_equal(return_val.im, -0.8);

    /* test ias_math_complex_divide with {-2.0,1.0} and {-2.0,2.0}, 
       result should be {0.75,0.25} */
    input_arg1.re = -2.0;
    input_arg1.im = 1.0;
    input_arg2.re = -2.0;
    input_arg2.im = 2.0;
    ias_math_complex_divide(&input_arg1, &input_arg2, &return_val);
    assert_double_equal(return_val.re, 0.75);
    assert_double_equal(return_val.im, 0.25);

    /* test ias_math_complex_magnitude with {0.0,0.0},
       result should be 0.0 */
    input_arg1.re = 0.0;
    input_arg1.im = 0.0;
    double result = ias_math_complex_magnitude(&input_arg1);
    assert_double_equal(result, 0.0);

    /* test ias_math_complex_magnitude with {1.0,0.0},
       result should be 1.0 */
    input_arg1.re = 1.0;
    input_arg1.im = 0.0;
    result = ias_math_complex_magnitude(&input_arg1);
    assert_double_equal(result, 1.0);

    /* test ias_math_complex_magnitude with {0.0,-1.0},
       result should be 1.0 */
    input_arg1.re = 0.0;
    input_arg1.im = -1.0;
    result = ias_math_complex_magnitude(&input_arg1);
    assert_double_equal(result, 1.0);

    /* test ias_math_complex_magnitude with {-3.0,4.0},
       result should be 5.0 */
    input_arg1.re = -3.0;
    input_arg1.im = 4.0;
    result = ias_math_complex_magnitude(&input_arg1);
    assert_double_equal(result, 5.0);
}

/* This is a simple function evaluated by the simulated annealing test */
double evaluate_function
(
    int n,
    double *a,
    void *params
)
{
    double total = 0.0;
    double fake_param = *(double *)params;
    int i;

    for (i = 0; i < 5; i++)
    {
        total += a[i] * a[i];
    }

    /* The fake parameter is supposed to be 1.0, so multiply by it to make sure
       it came through as expected */
    total *= fake_param;

    return total;
}

Ensure test_simulated_annealing()
{
    int number_points;
    int number_parameters;
    double best_value;
    double convergence_tolerance;
    double temperature;
    int annealing_iterations;
    double **simplex;
    double simplex_values[6];
    double best_fit_parameters[5];
    int i;
    int j;
    int status;
    double fake_param;

    /* Set up a scenario similar to one of the CalVal prototype scenarios, 
       execept for the evaluated function which is much simpler */
    number_points = 32;
    number_parameters = 5;
    best_value = 96.325074834884177;
    convergence_tolerance = 1.0e-6;
    temperature = 19.226542655123936;
    annealing_iterations = 300;

    /* Allocate simplex 2D array */
    simplex = (double **)malloc((number_parameters + 1) * sizeof(double *));
    if (simplex == NULL)
    {
        assert_true_with_message(0, "Error allocating simplex array");
        return;
    }

    for (i = 0; i < (number_parameters + 1); i++)
    {
        if ((simplex[i] = (double *)malloc(number_parameters * sizeof(double)))
            == NULL)
        {
            for (j = 0; j < i; j++)
            {
                free(simplex[j]);
            }
            free(simplex);
            assert_true_with_message(0, "Error allocating simplex subarrays");
            return;
        }
    }

    simplex[0][0] = 1089.8476314375;
    simplex[0][1] = 0.0;
    simplex[0][2] = 1370.0;
    simplex[0][3] = 1370.0;
    simplex[0][4] = 0.0;
    simplex[1][0] = 1139.8476314375;
    simplex[1][1] = 0.0;
    simplex[1][2] = 1370.0;
    simplex[1][3] = 1370.0;
    simplex[1][4] = 0.0;
    simplex[2][0] = 1089.8476314375;
    simplex[2][1] = 0.0;
    simplex[2][2] = 1370.0;
    simplex[2][3] = 1370.0;
    simplex[2][4] = 0.0;
    simplex[3][0] = 1089.8476314375;
    simplex[3][1] = 0.0;
    simplex[3][2] = 1370.0;
    simplex[3][3] = 1370.0;
    simplex[3][4] = 0.0;
    simplex[4][0] = 1089.8476314375;
    simplex[4][1] = 0.0;
    simplex[4][2] = 1370.0;
    simplex[4][3] = 1370.0;
    simplex[4][4] = 0.0;
    simplex[5][0] = 1089.8476314375;
    simplex[5][1] = 0.0;
    simplex[5][2] = 1370.0;
    simplex[5][3] = 1370.0;
    simplex[5][4] = 0.0;

    simplex_values[0] = 124.06648268468346;
    simplex_values[1] = 96.325074834884177;
    simplex_values[2] = 116.14483891067628;
    simplex_values[3] = 116.55318256135678;
    simplex_values[4] = 120.25951346336578;
    simplex_values[5] = 123.99816625256733;

    best_fit_parameters[0] = 1139.8476314375;
    best_fit_parameters[1] = 0.0;
    best_fit_parameters[2] = 1370.0;
    best_fit_parameters[3] = 1370.0;
    best_fit_parameters[4] = 0.0;

    fake_param = 1.0;

    status = ias_math_simulated_annealing(number_points, simplex, 
        simplex_values, number_parameters, best_fit_parameters, &best_value, 
        convergence_tolerance, &evaluate_function, &fake_param,
        &annealing_iterations, temperature);
    assert_equal(status, SUCCESS);

    assert_double_equal(best_value, 96.325075);
    assert_equal(annealing_iterations, -1);

    assert_double_equal(simplex[0][0], 1089.847631);
    assert_double_equal(simplex[0][1], 0.0);
    assert_double_equal(simplex[0][2], 1370.0);
    assert_double_equal(simplex[0][3], 1370.0);
    assert_double_equal(simplex[0][4], 0.0);
    assert_double_equal(simplex[1][0], 1089.830977);
    assert_double_equal(simplex[1][1], 0.0);
    assert_double_equal(simplex[1][2], 1370.0);
    assert_double_equal(simplex[1][3], 1370.0);
    assert_double_equal(simplex[1][4], 0.0);
    assert_double_equal(simplex[2][0], 1089.824888);
    assert_double_equal(simplex[2][1], 0.0);
    assert_double_equal(simplex[2][2], 1370.0);
    assert_double_equal(simplex[2][3], 1370.0);
    assert_double_equal(simplex[2][4], 0.0);
    assert_double_equal(simplex[3][0], 1089.826566);
    assert_double_equal(simplex[3][1], 0.0);
    assert_double_equal(simplex[3][2], 1370.0);
    assert_double_equal(simplex[3][3], 1370.0);
    assert_double_equal(simplex[3][4], 0.0);
    assert_double_equal(simplex[4][0], 1089.815846);
    assert_double_equal(simplex[4][1], 0.0);
    assert_double_equal(simplex[4][2], 1370.0);
    assert_double_equal(simplex[4][3], 1370.0);
    assert_double_equal(simplex[4][4], 0.0);
    assert_double_equal(simplex[5][0], 1089.830076);
    assert_double_equal(simplex[5][1], 0.0);
    assert_double_equal(simplex[5][2], 1370.0);
    assert_double_equal(simplex[5][3], 1370.0);
    assert_double_equal(simplex[5][4], 0.0);

    assert_double_equal(simplex_values[0], 116.144839);
    assert_double_equal(simplex_values[1], 4941531.559075);
    assert_double_equal(simplex_values[2], 4941518.286249);
    assert_double_equal(simplex_values[3], 4941521.943718);
    assert_double_equal(simplex_values[4], 4941498.577176);
    assert_double_equal(simplex_values[5], 4941529.593755);

    assert_double_equal(best_fit_parameters[0], 1139.847631);
    assert_double_equal(best_fit_parameters[1], 0.000000);
    assert_double_equal(best_fit_parameters[2], 1370.000000);
    assert_double_equal(best_fit_parameters[3], 1370.000000);
    assert_double_equal(best_fit_parameters[4], 0.000000);

    /* Free the simplex array */
    for (i = 0; i < (number_parameters + 1); i++)
    {
        free(simplex[i]);
    }
    free(simplex);

}

Ensure test_merge_stdev_mean()
{
    int index;                  /* Loop index */
    float combined_stdev;       /* Calculated stdev */
    float combined_mean;        /* Calculated mean */
    float combined_frames;      /* Average frames */
    int status;                 /* Status holder for return */
    int frames[5];              /* Frames array */
    float means[5];             /* Means array */
    float stdevs[5];            /* Stdevs array */

    /* Set up data */
    for (index = 1; index <= 5; index++)
    {
        means[index-1] = 1.0 + (index *.1);
        frames[index-1] =  10 * index;
        stdevs[index-1] = 2.5 * index;
    }

    /* Calculate mean, stdev, and average frames */
    status = ias_math_merge_mean_and_stdev_samples(frames, means, stdevs, 5, 
        &combined_stdev, &combined_mean, &combined_frames);
    assert_equal(status, SUCCESS);

    /* Check Results */
    assert_equal(combined_frames, 30);
    assert_double_equal(combined_stdev, 9.683261871338);
    assert_double_equal(combined_mean, 1.36666666666666666666667);

    /* Disable all output logging since the following tests will generate
       errors */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* Calculate mean, stdev, and average frames using NULL frames */
    status = ias_math_merge_mean_and_stdev_samples(NULL, means, 
        stdevs, 5, &combined_stdev, &combined_mean, &combined_frames);
    assert_equal(status, ERROR);

    /* Calculate mean, stdev, and average frames using NULL mean */
    status = ias_math_merge_mean_and_stdev_samples(frames, NULL, 
        stdevs, 5, &combined_stdev, &combined_mean, &combined_frames);
    assert_equal(status, ERROR);

    /* Calculate mean, stdev, and average frames using NULL stdev */
    status = ias_math_merge_mean_and_stdev_samples(frames, means, 
        NULL, 5, &combined_stdev, &combined_mean, &combined_frames);
    assert_equal(status, ERROR);

    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);
}

Ensure test_check_if_singular_matrix()
{
    int index;         /* Generic loop index */
    double *matrix[8]; /* Matrix pointer array */

    /* Singular matrices */
    double matrix1[2][2] = { {0, 0}, {0, 0} };
    double matrix2[2][2] = { {0, 0}, {0, 1} };
    double matrix3[2][2] = { {0, 0}, {1, 0} };
    double matrix4[2][2] = { {0, 1}, {0, 0} };

    /* Nonsingular matrices */
    double matrix5[2][2] = { {1, 0}, {0, 1} };
    double matrix6[2][2] = { {0, 1}, {1, 0} };
    double matrix7[2][2] = { {0, 1}, {1, 1} };
    double matrix8[2][2] = { {1, 0}, {1, 1} };

    /* Return Results */
    int results[8] = {1, 1, 1, 1, 0, 0, 0, 0};

    /* Initialize the matrix pointer array */
    matrix[0] = &matrix1[0][0];
    matrix[1] = &matrix2[0][0]; 
    matrix[2] = &matrix3[0][0];
    matrix[3] = &matrix4[0][0];
    matrix[4] = &matrix5[0][0];
    matrix[5] = &matrix6[0][0];
    matrix[6] = &matrix7[0][0];
    matrix[7] = &matrix8[0][0];

    /* Test the matrices for singularity */
    for (index = 0; index < 8; index++)
    {
        assert_equal(ias_math_check_if_singular_matrix(matrix[index], 2),
            results[index]);
    }
}

int main(int argc, char **argv)
{
    unsigned int seed = 5;     /* Seed for random number generator */
    srand(seed);

    TestSuite *suite = create_test_suite();
    /* Leap seconds tests are separated like this for a reason, due to needing
       to have the leap seconds newly initialized for each run and there is no
       re-initialize or clear initialization capability. Do not consolidate. */
    add_test(suite, test_leap_seconds1);
    add_test(suite, test_leap_seconds2);
    add_test(suite, test_leap_seconds3);
    add_test(suite, test_leap_seconds4);
    add_test(suite, test_leap_seconds5);
    add_test(suite, test_leap_seconds6);
    add_test(suite, test_leap_seconds_init_from_UTC);
    add_test(suite, test_leap_seconds_init);
    add_test(suite, test_constants);
    add_test(suite, test_legendre);
    add_test(suite, test_add_seconds_to_year_doy_sod);
    add_test(suite, test_compute_unit_vector);
    add_test(suite, test_line_intersection);
    add_test(suite, test_quaternion_conversions);
    add_test(suite, test_point_in_polygon);
    add_test(suite, test_calculate_polygon_centroid);
    add_test(suite, test_complex);
    add_test(suite, test_simulated_annealing);
    add_test(suite, test_merge_stdev_mean); 
    add_test(suite, test_check_if_singular_matrix);

    ias_log_initialize("mathtest");
    if (argc > 1) {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}
