#include <stdlib.h>
#include "gctp.h"
#include "ias_math.h"
#include "ias_logging.h"
#include "ias_grid.h"
#include "ias_geo.h"

/******************************************************************************
NAME: ias_misc_calculate_trim_box

PURPOSE: Converts the trimming bounding box from lat/long to line/sample
    locations in the output image.

RETURNS: SUCCESS or ERROR

******************************************************************************/
int ias_misc_calculate_trim_box
(
    int band_index,               /* I: Band index to use for calculating LUT */
    const IAS_CORNERS *corners,   /* I: Corners to base trimming on (lat/long)*/
    const IAS_GRID_TYPE *grid_ptr,/* I: Grid */
    double *line,                 /* O: Calculated line numbers for 
                                        (ul, ur, lr, ll)*/
    double *samp                  /* O: Calculated sample numbers for 
                                        (ul, ur, lr, ll) */
)
{
    double geo_parms[IAS_PROJ_PARAM_SIZE]; /* Geo projection parameters */
    double proj_x;               /* Projection X coordinate */
    double proj_y;               /* Projection Y coordinate */
    int output_lines;            /* Output number of lines */
    int output_samples;          /* Output number of samples */
    int target_units;            /* Projection units code */
    int index;                      /* Counter */
    int sca_index = 0;           /* Use the first SCA since they are all the
                                    same for the intents of this routine */
    double source_x[4];          /* corner samples of the output space */
    double source_y[4];          /* corner lines of the output space */
    double target_x[4];          /* corner samples of the input space */
    double target_y[4];          /* corner lines of the input space */
    double poly_x[4];            /* calculated x polynomial coefficients */
    double poly_y[4];            /* calculated y polynomial coefficients */
    int degree = 1;              /* Polynomial degree */
    IAS_PROJECTION geo_proj;
    IAS_PROJECTION target_proj;
    IAS_GEO_PROJ_TRANSFORMATION *degrees2proj = NULL;
    /* Projection transformation to go from the geographic projection in
       degrees to the selected output projection */

    /* Get the number of output lines and samples from the grid */
    output_lines = grid_ptr->gridbands[band_index][sca_index].lines;
    output_samples = grid_ptr->gridbands[band_index][sca_index].samps;

    /* Set up the source and target corners for the mapping polynomials */
    source_x[0] = grid_ptr->corners.upleft.x; 
    source_x[1] = grid_ptr->corners.upright.x;
    source_x[2] = grid_ptr->corners.loright.x;
    source_x[3] = grid_ptr->corners.loleft.x; 
    source_y[0] = grid_ptr->corners.upleft.y; 
    source_y[1] = grid_ptr->corners.upright.y;
    source_y[2] = grid_ptr->corners.loright.y;
    source_y[3] = grid_ptr->corners.loleft.y; 

    target_x[0] = 0.0;
    target_x[1] = output_samples - 1;
    target_x[2] = output_samples - 1;
    target_x[3] = 0.0;
    target_y[0] = 0.0;
    target_y[1] = 0.0;
    target_y[2] = output_lines - 1;
    target_y[3] = output_lines - 1;

    /* Calculate the polynomials to map from projection coordinates to
       line/sample */
    if (ias_geo_compute_proj2proj_poly(source_x, source_y, target_x, target_y,
        0 /* 3 terms */, poly_x, poly_y) != SUCCESS)
    {
        IAS_LOG_ERROR("Error in determining projection to line/sample poly.");
        return ERROR;
    }
    IAS_LOG_DEBUG("Poly X %f %f %f",poly_x[0],poly_x[1],poly_x[2]);
    IAS_LOG_DEBUG("Poly y %f %f %f",poly_y[0],poly_y[1],poly_y[2]);

    /* Create the projection transformation to go from geographic degrees to
       the target projection */
    if (ias_geo_get_units(grid_ptr->units, &target_units) != SUCCESS)
    {
        IAS_LOG_ERROR("Unsupported units ('%s') found in the grid",
            grid_ptr->units);
        return ERROR;
    }
    
    /* Initialize the projection parameters */
    for (index = 0; index < IAS_PROJ_PARAM_SIZE; index++)
    {
        geo_parms[index] = 0.0;
    }

    ias_geo_set_projection(GEO, NULLZONE, DEGREE, grid_ptr->spheroid,
        geo_parms, &geo_proj);
    ias_geo_set_projection(grid_ptr->code, grid_ptr->zone, target_units,
        grid_ptr->spheroid, grid_ptr->projprms, &target_proj);

    /* Create the transformation */
    degrees2proj = ias_geo_create_proj_transformation(&geo_proj, &target_proj);
    if (!degrees2proj)
    {
        IAS_LOG_ERROR("Error creating degrees to map projection "
                "transformation");
        return ERROR;
    }

    /* Convert each of the bounding box corners from lat/long to projection
       coordinates, then convert them to line/sample */
    if (ias_geo_transform_coordinate(degrees2proj, corners->upleft.x,
        corners->upleft.y, &proj_x, &proj_y) != SUCCESS)
    {
        IAS_LOG_ERROR("Error transforming the upper left coordinate");
        ias_geo_destroy_proj_transformation(degrees2proj);
        return ERROR;
    }
    ias_math_eval_poly(degree, poly_y, proj_x, proj_y, &line[0]);
    ias_math_eval_poly(degree, poly_x, proj_x, proj_y, &samp[0]);

    if (ias_geo_transform_coordinate(degrees2proj, corners->upright.x,
        corners->upright.y, &proj_x, &proj_y) != SUCCESS)
    {
        IAS_LOG_ERROR("Error transforming the upper right coordinate");
        ias_geo_destroy_proj_transformation(degrees2proj);
        return ERROR;
    }
    ias_math_eval_poly(degree, poly_y, proj_x, proj_y, &line[1]);
    ias_math_eval_poly(degree, poly_x, proj_x, proj_y, &samp[1]);

    if (ias_geo_transform_coordinate(degrees2proj, corners->loright.x,
        corners->loright.y, &proj_x, &proj_y) != SUCCESS)
    {
        IAS_LOG_ERROR("Error transforming the lower right coordinate");
        ias_geo_destroy_proj_transformation(degrees2proj);
        return ERROR;
    }
    ias_math_eval_poly(degree, poly_y, proj_x, proj_y, &line[2]);
    ias_math_eval_poly(degree, poly_x, proj_x, proj_y, &samp[2]);

    if (ias_geo_transform_coordinate(degrees2proj, corners->loleft.x, 
        corners->loleft.y, &proj_x, &proj_y) != SUCCESS)
    {
        IAS_LOG_ERROR("Error transforming the lower left coordinate");
        ias_geo_destroy_proj_transformation(degrees2proj);
        return ERROR;
    }
    ias_math_eval_poly(degree, poly_y, proj_x, proj_y, &line[3]);
    ias_math_eval_poly(degree, poly_x, proj_x, proj_y, &samp[3]);

    /* Free the projection transformation */
    ias_geo_destroy_proj_transformation(degrees2proj);
    degrees2proj = NULL;

    IAS_LOG_DEBUG("Trimming Bounding Box: UL %f %f, UR %f %f, LR %f %f, "
        "LL %f %f",line[0],samp[0],line[1],samp[1],line[2],samp[2],
        line[3],samp[3]);

    return SUCCESS;
}
