/******************************************************************************
Name: ias_misc_get_doq_reference_filename

Purpose: Given a WRS path and row, construct the name of the DOQ filename to
    use.  It first looks for one in the UTM projection and falls back to the
    SOM projection if the UTM file isn't available.  The IAS_IMG environment
    variable is used to define the directory to search for the reference files.

Returns:
    SUCCESS if the filename was successfully constructed and the file exists
    ERROR if no reference image could be found
******************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include "ias_logging.h"
#include "ias_miscellaneous.h"


int ias_misc_get_doq_reference_filename
(
    int wrs_path,       /* I: Target WRS path */
    int wrs_row,        /* I: Target WRS row */
    char *filename,     /* O: Buffer to store the filename in */
    int filename_size   /* I: Size of the filename buffer */
)
{
    const char *dir_ptr;
    int status;

    /* assign the DOQ file name */
    dir_ptr = getenv("IAS_IMG");
    if ((dir_ptr == NULL) || (strlen(dir_ptr) <= 0))
    {
        IAS_LOG_ERROR("Retrieving the IAS_IMG environment variable to "
                "build the DOQ path");
        return ERROR;
    }

    /* construct the input DOQ image filename, looking for one in
       the UTM projection */
    status = snprintf(filename, filename_size, 
            "%s/ias%03d%03d_UTM_doq.h5", dir_ptr, wrs_path, wrs_row);
    if (status < 0 || status >= filename_size)
    {
        /* there was an error constructing the DOQ file name */
        IAS_LOG_ERROR("Constructing the UTM DOQ file name");
        return ERROR;
    }

    /* Check whether the constructed filename exists.  If it does not
       try looking for a SOM image. */
    if (access(filename, R_OK) != 0)
    {
        /* construct the input DOQ image filename, looking for one in
           the SOM projection */
        status = snprintf(filename, filename_size, 
                "%s/ias%03d%03d_SOM_doq.h5", dir_ptr, wrs_path, wrs_row);
        if (status < 0 || status >= filename_size)
        {
            /* there was an error constructing the DOQ file name */
            IAS_LOG_ERROR("Constructing the SOM DOQ file name");
            return ERROR;
        }

        if (access(filename, R_OK) != 0)
        {
            IAS_LOG_ERROR("Unable to find DOQ reference image in directory %s"
                " for WRS path %d, row %d", dir_ptr, wrs_path, wrs_row);
            return ERROR;
        }
    }

    return SUCCESS;
}
