/******************************************************************************
NAME:           ias_misc_read_coregistered_single_band_l1g

PURPOSE:        Read any image and verify that it is coregistered by comparing
                the datatype, image_corners, and the scaled lines/samples 
                against what is expected.

RETURN VALUE: 
Type = SUCCESS/ERROR
******************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <unistd.h>
#include <math.h>    
#include "ias_const.h"
#include "ias_types.h"
#include "ias_logging.h"
#include "ias_miscellaneous.h"

int ias_misc_read_coregistered_single_band_l1g 
(
    const char *filename,                 /* I: Name of image file to read */
    const IAS_CORNERS *expected_corners,  /* I: Expected corners of image */
    int lines_in_image,                   /* I: Lines to be scaled */
    int samples_per_image_line,           /* I: Samples to be scaled  */
    double image_pixsize,                 /* I: Pixel size used in scaling */
    IAS_IMAGE *image_read,                /* O: Image to be read in */
    IAS_DATA_TYPE expected_type           /* I: Expected datatype of image */ 
)
{
    int status;                 /* Return status */
    double max_corner_diff;     /* Maximum allowed difference in the image and
                                   image corner points */
    int scaled_lines_in_image;  /* Number of lines in the image after the
                                   lines_in_image is scaled by the ratio of
                                   the image pixel size to image_read 
                                   pixel size*/
    int scaled_samps_in_image;  /* Samples across the image after the
                                   samples_per_image_line is scaled by the ratio
                                   of the image pixel size to image_read 
                                   pixel size */
    IAS_DATA_TYPE datatype_read;/* Datatype of image_read as read from image */

    /* Read in image and check return value */
    status = ias_misc_read_single_band_l1g(filename, 1, &datatype_read,
        image_read);
    if (status == ERROR)
    {
        IAS_LOG_ERROR("Reading image file %s", filename);
        return ERROR;
    }

    /* Verify data type is the expected type passed as parameter */
    if (datatype_read != expected_type)
    {
        const char *expected_type_string; 
        const char *datatype_read_string;

        /* Convert types to strings */
        ias_misc_convert_data_type_to_string(expected_type, 
            &expected_type_string);    
        ias_misc_convert_data_type_to_string(datatype_read, 
            &datatype_read_string); 

        IAS_LOG_ERROR("Not expected data type, type found was %s"
            " expected was %s", datatype_read_string, expected_type_string);
        free(image_read->data);
        image_read->data = NULL;
        return ERROR;        
    }

    /* Set the maximum difference between the image_read and image corner points
       to one-half of the pixsize */
    max_corner_diff = image_read->pixel_size_x * 0.5;
    
    /* Verify the image_read is valid */
    /* Make sure the corners of the image are within 0.5 pixels */
    if ((fabs(expected_corners->upleft.y - image_read->corners.upleft.y)
                > max_corner_diff)
            || (fabs(expected_corners->upleft.x - image_read->corners.upleft.x) 
                > max_corner_diff)
            || (fabs(expected_corners->upright.y - 
                    image_read->corners.upright.y) > max_corner_diff)
            || (fabs(expected_corners->upright.x - 
                    image_read->corners.upright.x) > max_corner_diff)
            || (fabs(expected_corners->loleft.y - image_read->corners.loleft.y) 
                > max_corner_diff)
            || (fabs(expected_corners->loleft.x - image_read->corners.loleft.x) 
                > max_corner_diff)
            || (fabs(expected_corners->loright.y - 
                    image_read->corners.loright.y) > max_corner_diff)
            || (fabs(expected_corners->loright.x - 
                    image_read->corners.loright.x) > max_corner_diff)
       )
    {
        IAS_LOG_ERROR("Image corners do not match");
        free(image_read->data);
        image_read->data = NULL;
        return ERROR;
    }
    
    /* Scale the expected lines and samples to account for different pixel
       sizes */
    scaled_lines_in_image = (int)(image_pixsize / image_read->pixel_size_x
            * lines_in_image - ((image_pixsize - image_read->pixel_size_x)
                / image_read->pixel_size_x) + 0.5);
    scaled_samps_in_image = (int)(image_pixsize / image_read->pixel_size_y
            * samples_per_image_line
            - ((image_pixsize - image_read->pixel_size_y)
                / image_read->pixel_size_y) + 0.5);

    /* Make sure the number of lines and sample agree between the coregistered
       images */
    if ((image_read->ns != scaled_samps_in_image )
            || (image_read->nl != scaled_lines_in_image))
    {
        IAS_LOG_ERROR("Image sizes do not match. Image read has %d, %d. "
            "Expected %d, %d (line, sample)", image_read->nl, image_read->ns,
            scaled_lines_in_image, scaled_samps_in_image);
        free(image_read->data);
        image_read->data = NULL;
        return ERROR;
    }

    return SUCCESS;
}
