/*****************************************************************************
 NAME:         ias_misc_retrieve_rlut

 PURPOSE:      Retrieve the RLUT filename and path from the CPS and then
               create a symbolic link to that path/filename.

 RETURNS:      This function returns a status of SUCCESS if the requested file
               is successfully retrieved from the CPS, otherwise ERROR is
               returned.
*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <unistd.h>

#include "ias_const.h"
#include "ias_logging.h"
#include "ias_miscellaneous.h"

#define  RLUT_LINE_SIZE  300
#define  SERVICE_FILE_NAME "&filename="
#define  SERVICE_RLUT "/dataservices/rest/rlut?operation=cacheRlutParmFile"

int ias_misc_retrieve_rlut
(
    const char *date_time,   /* I: Acquisition date-time */
    const char *service_url, /* I: CPS URL */
    char *path,              /* I/O: RLUT file to be retrieved */
    int max_path_size        /* I: Max size of the path buffer */
)
{
    FILE *fp;
    char buffer[RLUT_LINE_SIZE];
    char tmp_path[PATH_MAX];
    char url[PATH_MAX];
    char *ias_data_dir;
    char *newline;
    int size;
    int status;

    /* If the requested file path contains a '/' character, the path is
       interpreted as either an absolute path or a relative path to a file on
       disk.  If the file exists, return a SUCCESS status otherwise return an
       ERROR status to indicate the file does not exist. */
    if (strchr(path, '/') != NULL)
    {
        if (access(path, F_OK) == 0)
            return SUCCESS;
        IAS_LOG_ERROR("File does not exist: %s", path);
        return ERROR;
    }

    ias_data_dir = getenv("IAS_DATA_DIR");
    if (ias_data_dir == NULL)
    {
        IAS_LOG_ERROR("Retrieving environment variable IAS_DATA_DIR");
        return ERROR;
    }

    if (strcmp(path, "") == 0)
    {
        /* If no name has been specified at all, attempt to retrieve the
           nearest match for the scene's acquisition date and time. */
        size = snprintf(url, sizeof(url), "%s%s&date=%s", service_url,
            SERVICE_RLUT, date_time);
        if ((size < 0) || (size >= sizeof(url)))
        {
            IAS_LOG_ERROR("URL string truncated by snprintf()");
            return ERROR;
        }

        size = snprintf(tmp_path, sizeof(tmp_path), "RLUT%s", date_time);
        if ((size < 0) || (size >= sizeof(tmp_path)))
        {
            IAS_LOG_ERROR("File path truncated by snprintf()");
            return ERROR;
        }

        /* Retrieve the RLUT filename */
        status = ias_misc_retrieve_file_from_url(url, tmp_path);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Retrieving file from CPS: %s", url);
            return ERROR;
        }

        /* Open the retrieved file and read the RLUT file name from it.  When
           finished, the temporary file is closed and removed.  We simply need
           to read the name of the actual RLUT from this file. */
        fp = fopen(tmp_path, "r");
        if (fp == NULL)
        {
            IAS_LOG_ERROR("Opening file after retrieving from CPS: %s",
                path);
            return ERROR;
        }
        if (fgets(buffer, RLUT_LINE_SIZE, fp) == NULL)
        {
            IAS_LOG_ERROR("Reading from RLUT file: %s", path);
            fclose(fp);
            return ERROR;
        }
        fclose(fp);
        unlink(tmp_path);

        /* Remove a trailing newline if present
           The IAS service does not generate a response with a newline
           character, however during unit testing it is possible for a
           response to be created with a newline. */
        newline = strrchr(buffer, '\n');
        if (newline != NULL)
        {
            newline[0] = '\0';
        }
        newline = NULL;

        /* Save the new RLUT file name */
        size = snprintf(path, max_path_size, "%s", buffer);
        if ((size < 0) || (size >= max_path_size))
        {
            IAS_LOG_ERROR("File path truncated by snprintf()");
            return ERROR;
        }

        /* Remove any existing symbolic link in case DSW has failed before */
        unlink(path);

        /* Create a symbolic link in the current directory to the actual RLUT
           file that was cached by the CPS. */
        size = snprintf(tmp_path, sizeof(tmp_path), "%s/rlut/%s", ias_data_dir,
            path);
        if ((size < 0) || (size >= sizeof(tmp_path)))
        {
            IAS_LOG_ERROR("File path truncated by snprintf()");
            return ERROR;
        }

        if (symlink(tmp_path, path))
        {
            IAS_LOG_ERROR("Creating symlink %s -> %s", path, tmp_path);
            return ERROR;
        }

        return SUCCESS;
    }

    /* The work order is requesting a specific RLUT file.  Create the URL to
       cache this file. */
    size = snprintf(url, sizeof(url), "%s%s%s%s", service_url, SERVICE_RLUT,
        SERVICE_FILE_NAME, path);
    if ((size < 0) || (size >= sizeof(url)))
    {
        IAS_LOG_ERROR("URL string truncated by snprintf()");
        return ERROR;
    }

    /* Create a temporary file name that will be retrieved from the CPS.  This
       action will cause the RLUT to be cached in the ${IAS_DATA_DIR}/rlut
       directory. */
    size = snprintf(tmp_path, sizeof(tmp_path), "TMP%s", date_time);
    if ((size < 0) || (size >= sizeof(tmp_path)))
    {
        IAS_LOG_ERROR("File path truncated by snprintf()");
        return ERROR;
    }

    /* Retrieve the RLUT filename */
    status = ias_misc_retrieve_file_from_url(url, tmp_path);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Retrieving file from CPS: %s", url);
        return ERROR;
    }

    /* The temporary file used to get the RLUT cached is no longer needed */
    unlink(tmp_path);

    /* Remove any existing symbolic link */
    unlink(path);

    /* Create a symbolic link in the current directory to the actual RLUT file
       that was cached by the CPS. */
    size = snprintf(tmp_path, sizeof(tmp_path), "%s/rlut/%s", ias_data_dir,
        path);
    if ((size < 0) || (size >= sizeof(tmp_path)))
    {
        IAS_LOG_ERROR("File path truncated by snprintf()");
        return ERROR;
    }

    if (symlink(tmp_path, path))
    {
        IAS_LOG_ERROR("Creating symlink %s -> %s", path, tmp_path);
        return ERROR;
    }
    return SUCCESS;
} /* END -- ias_misc_retrieve_rlut */

