/*****************************************************************************
 NAME: ias_misc_write_envi_image

 PURPOSE: Called to create an ENVI image output file containing the data from
          the specified data buffer. Also creates the ENVI header by calling
          the ias_misc_write_envi_header function using the specified input.

 RETURN VALUE: Type = int
    Value    Description
    -------  ---------------------
    SUCCESS  Successful completion
    ERROR    Operation failed
*****************************************************************************/
/* IAS Includes */
#include "ias_const.h"
#include "ias_types.h"
#include "ias_logging.h"
#include "ias_miscellaneous.h"

int ias_misc_write_envi_image
(
    const char *image_filename,     /* I: Full path name of output image */
    const IAS_PROJECTION *proj_info,/* I: Optional projection information,
                                          set to NULL if not known or needed */
    const char *description,        /* I: Optional header description, set to 
                                          NULL if not known or needed */
    int number_of_lines,            /* I: Number of data lines */
    int number_of_samples,          /* I: Number of data samples */
    double upper_left_x,            /* I: Optional upper-left X coordinate, set
                                          to 0.0 if not known or needed 
                                          (requires proj_info) */
    double upper_left_y,            /* I: Optional upper-left Y coordinate, set
                                          to 0.0 if not known or needed 
                                          (requires proj_info) */
    double projection_distance_x,   /* I: Optional pixel size in X projection, 
                                          set to 0.0 if not known or needed
                                          (requires proj_info) */
    double projection_distance_y,   /* I: Optional pixel size in Y projection,
                                          set to 0.0 if not known or needed
                                          (requires proj_info) */
    const char *band_names,         /* I: Optional single string for all 
                                          band names, set to NULL if not known 
                                          or needed */
    IAS_DATA_TYPE data_type,        /* I: Output image datatype */
    const void *data_buffer         /* I: Image data */
)
{
    int count;                      /* Output data count */
    int data_count;                 /* Amount of data to write */
    FILE *output_fd;                /* File descriptor for the output file */
    int data_size;                  /* Data size of output files */
    int number_of_bands = 1;        /* Number of output bands */

    /* Write the ENVI header */
    if (ias_misc_write_envi_header(image_filename, proj_info, description,
        number_of_lines, number_of_samples, number_of_bands, upper_left_x, 
        upper_left_y, projection_distance_x, projection_distance_y, band_names, 
        data_type) != SUCCESS)
    {
        IAS_LOG_ERROR("Unable to create ENVI header");
        return ERROR;
    }

    /* Open the file */
    output_fd = fopen(image_filename, "wb");
    if (!output_fd)
    {
        IAS_LOG_ERROR("Unable to open %s for writing", image_filename);
        return ERROR;
    }

    /* Figure out how much data has to be written */
    data_count = number_of_lines * number_of_samples;

    /* Get the size of the data type */
    if (ias_misc_get_sizeof_data_type(data_type, &data_size) != SUCCESS)
    {
        IAS_LOG_ERROR("Getting the size of the datatype");
        fclose(output_fd);
        return ERROR;
    }

    /* Write the data to the file */
    count = fwrite(data_buffer, data_size, data_count, output_fd);
    if (count != data_count)
    {
        IAS_LOG_ERROR("Error writing %s: Expected output size [%d]:"
            " Wrote [%d] values", image_filename, data_count, count);
        fclose(output_fd);
        return ERROR;
    }

    /* Close the file */
    if (fclose(output_fd) != 0)
    {
        IAS_LOG_ERROR("Closing file %s", image_filename);
        return ERROR;
    }

    return SUCCESS;
}
