/**************************************************************************
 NAME:                   ias_report_header_test

 PURPOSE:      Performs CGreen unit testing of the routines implemented
               to create a standard IAS report header with RPS-specific
               information

 RETURNS:      CGreen unit status

**************************************************************************/
#include <libgen.h>         /* for basename() */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "ias_const.h"
#include "ias_l0r.h"
#include "../ias_miscellaneous.h"
#include "../ias_miscellaneous_private.h"
#include "ias_types.h"
#include "ias_satellite_attributes.h"
#include "cgreen.h"

Ensure test_ias_misc_initialize_report_header_no_l0r()
{
    const char algorithm_name[IAS_L0R_ALGORITHM_NAME_LENGTH] =
        "initialize_report_header_no_l0r";
    char work_order_id[IAS_WORKORDER_ID_SIZE];
    char *path = NULL;
    int status;

    IAS_REPORT_HEADER *test_header = NULL;


    printf("Testing initialization of report header when no l0r file "
           "is present...\n");

    /* Initialize the satellite attributes library */
    status = ias_sat_attr_initialize(IAS_L8);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
        return;

    /* Construct the full path to the L0R data files */
    path = getenv("IAS_UNIT_TEST_DATA_DIR");
    assert_true(path != NULL);
    if (path == NULL)
    {
        assert_true_with_message(0, "Cannot locate test directory");
        return;
    }

    /* Try to read the test data into the header file structure */
    strcpy(work_order_id, "I1002");

    /* Initialize the report header */
    test_header = ias_misc_initialize_report_header(NULL,
        work_order_id, algorithm_name);
    assert_true(test_header != NULL);
    if (test_header == NULL)
    {
        assert_true_with_message(0, "Cannot initialize common header "
            "information");
        return;
    }

    /* Done */
    ias_misc_free_report_header(test_header);
}

Ensure test_ias_misc_initialize_report_header()
{
    const char l0r_directory[PATH_MAX] = "ias_lib/misc/l0ra_location";
    const char l0r_filename[IAS_L0R_FILE_NAME_LENGTH] = "test";
    const char algorithm_name[IAS_L0R_ALGORITHM_NAME_LENGTH]
        = "initialize_report_header";

    char work_order_id[IAS_WORKORDER_ID_SIZE];
    char *path = NULL;
    char full_l0r_name[PATH_MAX];
    int status;

    IAS_REPORT_HEADER *test_header = NULL;


    printf("Testing initialization of report header with l0r present...\n");

    /* Initialize the satellite attributes library */
    status = ias_sat_attr_initialize(IAS_L8);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
        return;

    /* Construct the full path to the L0R data files */
    path = getenv("IAS_UNIT_TEST_DATA_DIR");
    assert_true(path != NULL);
    if (path == NULL)
    {
        assert_true_with_message(0, "Cannot locate test data directory");
        return;
    }
    sprintf(full_l0r_name, "%s/%s/%s", path, l0r_directory, l0r_filename);

    /* Try to read the test data into the header file structure */
    strcpy(work_order_id, "I1002");

    /* Initialize the report header */
    test_header = ias_misc_initialize_report_header(full_l0r_name,
        work_order_id, algorithm_name);
    assert_true(test_header != NULL);
    if (test_header == NULL)
    {
        assert_true_with_message(0, "Cannot initialize test IAS common report "
                "header information");
        return;
    }

    /* Done */
    ias_misc_free_report_header(test_header);
}

Ensure test_ias_misc_initialize_rps_report_header()
{
    char l0r_directory[PATH_MAX]
        = "ias_lib/misc/test_LO800O0000002000274EDC00";
    char l0r_filename[IAS_L0R_FILE_NAME_LENGTH + 1] =
       "LO800O0000002000274EDC00";
    const char *blackbody_l0r 
        = "ias_lib/misc/LT800B2213422013101LGN00/LT800B2213422013101LGN00";
    char *path = NULL;
    char full_l0r_name[PATH_MAX];

    int status;

    RPS_REPORT_HEADER rps_header;


    printf("Testing initialization of RPS report header section...\n");

    /* Initialize the satellite attributes library */
    status = ias_sat_attr_initialize(IAS_L8);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
        return;

    /* Initialize the report header data structure */
    memset(&rps_header, 0, sizeof(RPS_REPORT_HEADER));

    /* Construct the full path to the L0R data files */
    path = getenv("IAS_UNIT_TEST_DATA_DIR");
    assert_true(path != NULL);
    if (path == NULL)
    {
        assert_true_with_message(0, "Cannot locate test data directory");
        return;
    }
    status = snprintf(full_l0r_name, sizeof(full_l0r_name), "%s/%s/%s", path,
             l0r_directory, l0r_filename);
    if (status < 0 || status >= sizeof(full_l0r_name))
    {
        assert_true_with_message(0, "L0R name buffer is not large enough");
        return;
    }

    /* Try to read the test data into the header file structure */
    status = ias_misc_initialize_rps_report_header(full_l0r_name,
        &rps_header);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Cannot initialize test RPS report "
            "header information");
        return;
    }

    /* Check to make sure the "right" data got copied to the data
       structure */
    assert_equal(rps_header.ms_integration_time, 3600);
    assert_equal(rps_header.pan_integration_time, 660);
    assert_equal(rps_header.current_detector_select_table, 3);
    assert_string_equal(rps_header.acq_datetime.start_time_oli,
                        "2000:274:12:43:10.7365840");
    assert_string_equal(rps_header.acq_datetime.stop_time_oli,
                        "2000:274:12:44:11.3410360");
    assert_true(rps_header.acq_datetime.total_time_oli > 0.0);
    assert_equal(rps_header.num_oli_frames, 14308);
    assert_equal(rps_header.average_oli_frame_rate, 236);

    /* Test a blackbody collection to make sure the blackbody temperatures
       are initialized */
    status = snprintf(full_l0r_name, sizeof(full_l0r_name), "%s/%s", path,
             blackbody_l0r);
    if (status < 0 || status >= sizeof(full_l0r_name))
    {
        assert_true_with_message(0, "L0R name buffer is not large enough");
        return;
    }

    /* Initialize the report header data structure */
    memset(&rps_header, 0, sizeof(RPS_REPORT_HEADER));

    /* Try to read the test data into the header file structure */
    status = ias_misc_initialize_rps_report_header(full_l0r_name,
        &rps_header);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Cannot initialize test RPS report "
            "header information for blackbody collection");
        return;
    }

    /* Check to make sure the "right" data got copied to the data structure.
       OLI fields should have been left untouched and the TIRS fields populated
       with the values from the file. */
    assert_equal(rps_header.ms_integration_time, 0);
    assert_equal(rps_header.pan_integration_time, 0);
    assert_equal(rps_header.current_detector_select_table, 0);
    assert_string_equal(rps_header.acq_datetime.start_time_oli, "");
    assert_string_equal(rps_header.acq_datetime.stop_time_oli, "");
    assert_equal(rps_header.acq_datetime.total_time_oli, 0.0);
    assert_equal(rps_header.num_oli_frames, 0);
    assert_equal(rps_header.average_oli_frame_rate, 0);
    assert_string_equal(rps_header.acq_datetime.start_time_tirs,
        "2013:101:22:13:49.5249290");
    assert_string_equal(rps_header.acq_datetime.stop_time_tirs,
        "2013:101:22:14:49.5003820");
    assert_equal(rps_header.num_tirs_frames, 4201);
    assert_equal(rps_header.average_tirs_frame_rate, 70);
    assert_double_equal(rps_header.tirs_blackbody_1_start_temp, 21.055027);
    assert_double_equal(rps_header.tirs_blackbody_1_stop_temp, 20.972475);
    assert_double_equal(rps_header.tirs_blackbody_2_start_temp, 21.039793);
    assert_double_equal(rps_header.tirs_blackbody_2_stop_temp, 20.955898);
    assert_double_equal(rps_header.tirs_blackbody_3_start_temp, 20.458765);
    assert_double_equal(rps_header.tirs_blackbody_3_stop_temp, 20.40709);
    assert_double_equal(rps_header.tirs_blackbody_4_start_temp, 20.15345);
    assert_double_equal(rps_header.tirs_blackbody_4_stop_temp, 20.088663);
}

Ensure test_write_report_header()
{
    const char l0r_filename[IAS_L0R_FILE_NAME_LENGTH + 1] =
        "LO800O0000002000274EDC00";
    const char l0r_directory[PATH_MAX] =
        "ias_lib/misc/test_LO800O0000002000274EDC00";
    const char out_directory[PATH_MAX] = "./";
    const char out_basename[PATH_MAX]
        = "test_report_header_common.txt";
    const char algorithm_name[IAS_L0R_ALGORITHM_NAME_LENGTH + 1]
        = "write_report_header";

    char work_order_id[IAS_WORKORDER_ID_SIZE];
    char *path = NULL;
    char full_l0r_name[PATH_MAX];
    char full_outname[PATH_MAX];

    char *collection_type = NULL;
    char *cpf_name = NULL;
    char *sensor_id = NULL;

    IAS_REPORT_HEADER *test_header = NULL;
    RPS_REPORT_HEADER test_rps_header;

    FILE *out_fptr  = NULL;

    int status;


    /* Initialize the satellite attributes library */
    status = ias_sat_attr_initialize(IAS_L8);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
        return;

    /* Get the test data directory path */
    path = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (path == NULL)
    {
        assert_true_with_message(0, "Cannot get test data directory path");
        return;
    }

    /* Build the L0R filename and initialize the header data structure*/
    sprintf(full_l0r_name, "%s/%s/%s", path, l0r_directory, l0r_filename);
    sprintf(full_outname, "%s/%s", out_directory, out_basename); 

    strcpy(work_order_id, "I1002");

    /* Initialize the report header */
    test_header = ias_misc_initialize_report_header(full_l0r_name,
        work_order_id, algorithm_name);
    assert_true(test_header != NULL);
    if (test_header == NULL)
    {
        assert_true_with_message(0, "Cannot initialize test IAS report "
            "header");
        return;
    }
    /* Set the band and sca numbers to avoid later debug messages from being
       generated by write report header */
    ias_misc_set_header_band_number(test_header, 1);
    ias_misc_set_header_sca_number(test_header, 14);
    ias_misc_set_header_cpf_name(test_header, "L08CPF20090101_20090331.01");
    /* Initialize the RPS report header */
    memset(&test_rps_header, 0, sizeof(RPS_REPORT_HEADER));
    status = ias_misc_initialize_rps_report_header(full_l0r_name,
        &(test_rps_header));
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Cannot initialize test RPS report "
            "header");
        ias_misc_free_report_header(test_header);
        return;
    }

    /* Now write the header data to a file in the specified format */
    out_fptr = fopen(full_outname, "w");
    assert_true_with_message((out_fptr != NULL), "Cannot create output "
        "file %s", full_outname);;
    if (out_fptr == NULL)
    {
        ias_misc_free_report_header(test_header);
        return;
    }

    collection_type = ias_misc_get_header_collection_type(test_header);
    cpf_name = ias_misc_get_header_cpf_name(test_header);
    sensor_id = ias_misc_get_header_sensor_id(test_header);

    assert_true_with_message((collection_type != NULL && cpf_name != NULL
        && sensor_id != NULL), "Cannot obtain RPS-specific information "
        "from common header");
    if ((collection_type == NULL) || (cpf_name == NULL)
            || (sensor_id == NULL))
    {
        ias_misc_free_report_header(test_header);
        fclose(out_fptr);
        return;
    }

    status = ias_misc_write_report_header(out_fptr, test_header);
    assert_true_with_message((status == SUCCESS),
        "Common header write failed");
    if (status != SUCCESS)
    {
        ias_misc_free_report_header(test_header);
        fclose(out_fptr);
        return;
    }

    status = ias_misc_write_rps_report_header(collection_type, cpf_name,
        sensor_id, test_rps_header, out_fptr);
    assert_true_with_message((status == SUCCESS),
        "RPS header write failed");

    /* Done */
    ias_misc_free_report_header(test_header);
    fclose(out_fptr);
}

Ensure test_setget_routines()
{
    const char l0r_directory[PATH_MAX] = "ias_lib/misc/l0ra_location";
    const char l0r_filename[IAS_L0R_FILE_NAME_LENGTH] = "test";
    const char expected_acq_date[IAS_L0R_DATE_LENGTH] =
        "2012:12:21 10:00:00";
    char *test_acq_date = NULL;
    const char algorithm_name[IAS_L0R_ALGORITHM_NAME_LENGTH]
        = "test_getset_routines";

    char work_order_id[IAS_WORKORDER_ID_SIZE];
    char *path = NULL;
    char full_l0r_name[PATH_MAX];

    int band = 1;
    int sca = 14;
    int wrs_path;
    int wrs_row;
    int status;

    IAS_REPORT_HEADER *test_header = NULL;


    printf("Testing header get/set function sets...\n");

    /* Initialize the satellite attributes library */
    status = ias_sat_attr_initialize(IAS_L8);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
        return;

    /* Construct the full path to the L0R data files */
    path = getenv("IAS_UNIT_TEST_DATA_DIR");
    assert_true(path != NULL);
    if (path == NULL)
    {
        assert_true_with_message(0,
                "Cannot locate test data directory");
        return;
    }
    sprintf(full_l0r_name, "%s/%s/%s", path, l0r_directory, l0r_filename);

    /* Try to read the test data into the header file structure */
    strcpy(work_order_id, "I1002");

    /* Initialize the report header */
    test_header = ias_misc_initialize_report_header(full_l0r_name,
            work_order_id, algorithm_name);
    assert_true_with_message((test_header != NULL), "Cannot initialize "
            "IAS common report header information");
    if (test_header == NULL)
    {
        return;
    }

    /* Test some of the getter/setter routines */

    ias_misc_set_header_acquisition_date(test_header, expected_acq_date);
    test_acq_date = ias_misc_get_header_date_acquired(test_header);
    assert_not_equal(test_acq_date, NULL);
    if (test_acq_date != NULL)
        assert_string_equal(expected_acq_date, test_acq_date);

    ias_misc_set_header_pathrow(test_header, 1, 10);
    ias_misc_get_header_pathrow(test_header, &wrs_path, &wrs_row);
    assert_equal(wrs_path, 1);
    assert_equal(wrs_row, 10);

    ias_misc_set_header_band_number(test_header, band);
    ias_misc_set_header_sca_number(test_header, sca);
    int test_band = ias_misc_get_header_band_number(test_header);
    int test_sca  = ias_misc_get_header_sca_number(test_header);

    assert_equal(test_band, band);
    assert_equal(test_sca, sca);

    /* Free the test report header */
    ias_misc_free_report_header(test_header);
}

Ensure test_ias_misc_copy_report_header()
{
    char l0r_name[PATH_MAX] = "test_solar";
    char cpf_name[PATH_MAX] = "L8CPF20101001_20101231.01";
    char algorithm_name[IAS_L0R_ALGORITHM_NAME_LENGTH] = "test_copy_header";
    char acquisition_date[IAS_L0R_DATE_LENGTH] = "2010:320:17:00:45.3456";
    int band_number = 1;
    int sca_number = 1;
    int status;
    IAS_REPORT_HEADER *in_header = NULL;
    IAS_REPORT_HEADER *copy_header = NULL;


    printf("Testing standard header copy...\n");

    /* Allocate the input header */
    in_header = calloc(1, sizeof(IAS_REPORT_HEADER));
    if (in_header == NULL)
    {
        assert_true_with_message(0, "Error allocating input report header");
        return;
    }

    /* Populate the input header with some test data */
    ias_misc_set_header_l0r_filename(in_header, l0r_name);
    ias_misc_set_header_cpf_name(in_header, cpf_name);
    ias_misc_set_header_collection_type(in_header, "OLI_SOLAR");
    ias_misc_set_header_algorithm_name(in_header, algorithm_name);
    ias_misc_set_header_acquisition_date(in_header, acquisition_date);
    ias_misc_set_header_band_number(in_header, band_number);
    ias_misc_set_header_sca_number(in_header, sca_number);

    /* Copy the report header */
    copy_header = ias_misc_copy_report_header(in_header);
    if (copy_header == NULL)
    {
        assert_true_with_message(0, "Error copying test standard report header "
                "information");
        ias_misc_free_report_header(in_header);
        return;
    }

    /* Do a direct memory block comparison between the two report headers */
    status = memcmp(copy_header, in_header, sizeof(IAS_REPORT_HEADER));
    assert_equal_with_message(status, 0, "Local copy of header is not the "
        "same as the input header");

    /* Cleanup */
    ias_misc_free_report_header(in_header);
    ias_misc_free_report_header(copy_header);

}   /* END -- test_copy_report_header */

Ensure test_ias_misc_initialize_gps_report_header()
{
    const char l1g_filename[PATH_MAX] = "ias_lib/misc/l1t.h5";
    const char l0r_filename[PATH_MAX] = "test_L0R_name";
    const char algorithm_name[IAS_L0R_ALGORITHM_NAME_LENGTH]
        = "initialize_gps_report_header";
    char work_order_id[IAS_WORKORDER_ID_SIZE];
    char *unit_test_data_dir = NULL;
    char full_l1g_name[PATH_MAX];
    int status;
    IAS_REPORT_HEADER *gps_header = NULL;

    printf("Testing initialization of GPS report header...\n");

    /* Initialize the satellite attributes library */
    status = ias_sat_attr_initialize(IAS_L8);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
        return;

    /* Construct the full path to the L1G data file */
    unit_test_data_dir = getenv("IAS_UNIT_TEST_DATA_DIR");
    assert_true(unit_test_data_dir != NULL);
    if (unit_test_data_dir == NULL)
    {
        assert_true_with_message(0, "Cannot locate test data directory");
        return;
    }
    sprintf(full_l1g_name, "%s/%s", unit_test_data_dir, l1g_filename);

    /* Try to set the test data into the header file structure */
    strcpy(work_order_id, "9876");

    /* Initialize the report header */
    gps_header = ias_misc_initialize_gps_report_header(full_l1g_name,
            l0r_filename, work_order_id, algorithm_name);
    assert_true(gps_header != NULL);
    if (gps_header == NULL)
    {
        assert_true_with_message(0, "Cannot initialize test IAS GPS report "
                "header information");
        ias_misc_free_report_header(gps_header);
        return;
    }

    /* Check the values got populated as expected */
    assert_string_equal(gps_header->work_order_id, "9876");
    assert_string_equal(gps_header->algorithm_name,
            "initialize_gps_report_header");
    assert_string_equal(gps_header->ias_software_version,
            ias_misc_get_software_version());
    assert_string_equal(gps_header->processing_center,
            getenv("PROCESSING_CENTER"));
    assert_string_equal(gps_header->l0r_filename, l0r_filename);
    assert_string_equal(gps_header->l1g_filename, basename(full_l1g_name));
    assert_string_equal(gps_header->collection_type, "EARTH_IMAGING");
    assert_string_equal(gps_header->spacecraft_id, "L8");
    assert_string_equal(gps_header->sensor_id, "OLITIRS");
    assert_equal(gps_header->path, 173);
    assert_equal(gps_header->row, 60);
    assert_string_equal(gps_header->date_acquired, "2003/01/31");
    assert_double_equal(gps_header->roll_angle, -5.099584814161062e-4);

    /* Done */
    ias_misc_free_report_header(gps_header);
}

Ensure test_ias_misc_write_gps_report_header()
{
    const char l1g_filename[PATH_MAX] = "ias_lib/misc/l1t.h5";
    const char l0r_filename[PATH_MAX] = "test_L0R_name";
    const char algorithm_name[IAS_L0R_ALGORITHM_NAME_LENGTH]
        = "write_gps_report_header";
    char outname[PATH_MAX] = "test_gps_report_header.txt";
    char outname_verify_file[PATH_MAX]
        = "ias_lib/misc/test_gps_report_header.txt";
    char full_outname_verify_file[PATH_MAX];
    char work_order_id[IAS_WORKORDER_ID_SIZE];
    char *unit_test_data_dir = NULL;
    char full_l1g_name[PATH_MAX];
    char command[PATH_MAX];
    int status;
    IAS_REPORT_HEADER *gps_header = NULL;
    FILE *out_fptr  = NULL;

    printf("Testing writing of GPS report header...\n");

    /* Initialize the satellite attributes library */
    status = ias_sat_attr_initialize(IAS_L8);
    assert_equal(status, SUCCESS);
    if (status != SUCCESS)
        return;

    /* Construct the full path to the L1G data file */
    unit_test_data_dir = getenv("IAS_UNIT_TEST_DATA_DIR");
    assert_true(unit_test_data_dir != NULL);
    if (unit_test_data_dir == NULL)
    {
        assert_true_with_message(0, "Cannot locate test data directory");
        return;
    }
    sprintf(full_l1g_name, "%s/%s", unit_test_data_dir, l1g_filename);

    /* Try to set the test data into the header file structure */
    strcpy(work_order_id, "5678");

    /* Initialize the report header */
    gps_header = ias_misc_initialize_gps_report_header(full_l1g_name,
            l0r_filename, work_order_id, algorithm_name);
    assert_true(gps_header != NULL);
    if (gps_header == NULL)
    {
        assert_true_with_message(0, "Cannot initialize test IAS GPS report "
                "header information");
        ias_misc_free_report_header(gps_header);
        return;
    }

    /* Now write the header data to a file in the specified format */
    out_fptr = fopen(outname, "w");
    assert_true_with_message((out_fptr != NULL), "Cannot create output "
            "GPS report header file %s", outname);;
    if (out_fptr == NULL)
    {
        ias_misc_free_report_header(gps_header);
        return;
    }

    status = ias_misc_write_gps_report_header(out_fptr, gps_header);
    assert_true_with_message((status == SUCCESS), "GPS header write failed");
    if (status != SUCCESS)
    {
        ias_misc_free_report_header(gps_header);
        fclose(out_fptr);
        return;
    }

    /* Close the file and release the header memory */
    ias_misc_free_report_header(gps_header);
    fclose(out_fptr);

    /* Construct the full path to the verification report header file */
    sprintf(full_outname_verify_file, "%s/%s", unit_test_data_dir,
            outname_verify_file);

    /* Compare the report header file created to the saved one it should
       match. Use SED to remove the lines that vary from run to run - the line
       with the date/time and the line with the IAS version. */
    status = snprintf(command, PATH_MAX,
            "sed '/Processed at/'d %s | sed '/IAS Version/'d | "
            "sed '/Database schema/'d - | diff -q - %s",
            outname, full_outname_verify_file);
    if (status < 0 || PATH_MAX < status)
    {
        assert_true_with_message(0, "Error preparing system command");
        return;
    }
    status = system(command);
    if (status != 0)
    {
        assert_true_with_message(0, "Differences found in GPS report headers "
                "%s and %s", outname, full_outname_verify_file);
        return;
    }
}


int main(int argc, char *argv[])
{
    TestSuite *suite = create_test_suite();

    add_test(suite, test_ias_misc_initialize_report_header_no_l0r);
    add_test(suite, test_ias_misc_initialize_report_header);
    add_test(suite, test_ias_misc_initialize_rps_report_header);
    add_test(suite, test_setget_routines);
    add_test(suite, test_write_report_header);
    add_test(suite, test_ias_misc_copy_report_header);
    add_test(suite, test_ias_misc_initialize_gps_report_header);
    add_test(suite, test_ias_misc_write_gps_report_header);

    if (argc > 1)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}
