/*************************************************************************

NAME: miscellaneous_library_tests.c

PURPOSE: Using Cgreen for testing the 2d array memory allocation 
         and testing the version library files

**************************************************************************/
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <math.h>
#include <sys/stat.h>
#include "hdf5.h"
#include "hdf5_hl.h"
#include "gctp.h"
#include "ias_db.h"
#include "ias_db_insert.h"
#include "ias_db_get_connect_info.h"
#include "ias_miscellaneous.h"
#include "ias_types.h"
#include "ias_const.h"
#include "ias_geo.h"
#include "ias_l1g.h"
#include "ias_l0r.h"
#include "ias_gcp.h"
#include "ias_grid.h"
#include "ias_geo.h"
#include "ias_structures.h"
#include "ias_logging.h"
#include "ias_angle_gen.h"
#include "ias_grid_io.h"
#include "time.h"
#include "cgreen/cgreen.h"
#include "config.h"
#include "expected_trim_lut.h"

#define DATETIME_STR_LENGTH 30
#define SQL_QUERY_LENGTH 200
#define BUFFER_LENGTH 200

typedef struct ias_model
{
    int model1;             /* band number */
    double model2;          /* SCA number */
}IAS_MODEL;

typedef struct ias_grid
{
    int bandno;             /* band number */
    double sca_number;      /* SCA number */
    IAS_MODEL *test_model;
}IAS_GRID;

Ensure test_ias_misc_get_char_id()
{
    int status;
    char char_id[IAS_CHAR_ID_SIZE];  
    char second_char_id[IAS_CHAR_ID_SIZE];  
    char sensor_id[IAS_L0R_SENSOR_ID_LENGTH];
    char *user = NULL;
    char *pwd = NULL;
    char *dbname = NULL;
    char *dbhost = NULL;
    char sql[SQL_QUERY_LENGTH];       /* buffer for SQL commands */
    char buffer[BUFFER_LENGTH];       /* buffer for SQL return messages */
    struct ias_db_query *query = NULL;
    struct ias_db_connection *db;
    char work_order_id[IAS_WORKORDER_ID_SIZE] = {"L1949"};

    /* Initialize the satellite attributes library */
    ias_sat_attr_initialize(IAS_L8);

    status = ias_db_initialize_database_lib();
    assert_equal(status, 0);

    status = ias_db_get_connect_info(&dbname, &user, &pwd, &dbhost);
    assert_equal(status, 0);

    db = ias_db_connect_to_database(dbname, user, pwd, dbhost);
    if (db == NULL)
    {
        exit(1);
    }

    free(dbname);
    free(user);
    free(pwd);
    free(dbhost);
    dbname = NULL;
    user = NULL;
    pwd = NULL;
    dbhost = NULL;

    /* changed test to logic test as everyone's tables are populated 
       differently */
    printf("Perform get char id tests.....\n");

    /* Radiometric */
    status = ias_misc_get_char_id(db, "radiometric_work_order_common",
            "radiometric_wo_common_id", work_order_id, char_id);
    assert_equal(status, SUCCESS);

    /* Set up a query that will fail and deliver an ORACLE diagnostic
       message.  A simple way to do this is with a typo to the
       'radiometric_work_order_common' table name in the previous query, as
       the "mis-identified" table doesn't exist.  Don't disable logging
       here, as we want to see the additional diagnostic message... */
    status = ias_misc_get_char_id(db, "radiomeric_work_order_common",
        "radiometric_wo_common_id", work_order_id, char_id);
    assert_not_equal(status, SUCCESS);

    /* Geometric */
    status = ias_misc_get_char_id(db, "geometric_work_order_common",
            "geometric_wo_common_id", work_order_id, char_id);
    assert_equal(status, SUCCESS);

    /* make sure trying to get the char ID again gives the same ID and does
       not generate a new one */
    status = ias_misc_get_char_id(db, "geometric_work_order_common",
            "geometric_wo_common_id", work_order_id, second_char_id);
    assert_equal(status, SUCCESS);
    assert_string_equal(char_id, second_char_id);

    /* try to get a char ID from a column that has NULL.  Should return 
     * an error */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = ias_misc_get_char_id(db, "geometric_work_order_common",
            "sensor_id", work_order_id, second_char_id);
    assert_equal(status, ERROR);
    assert_string_equal("NULL", second_char_id);
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

    /* test to make sure sensor_id was not populated
       ******************************************************************
       This is included due to at an early development time, the database
       had been setup to default this field to a value, but then the
       default was removed. So this check is essentially testing the
       database defaults to ensure the field was not defaulted to any
       value, as the database should be operating like now.
       *******************************************************************/
    snprintf (sql, sizeof(sql), 
            "select sensor_id from geometric_work_order_common "
            "where work_order_id = '%s';", work_order_id);

    query = ias_db_perform_query(db, sql);
    if (!ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, buffer, sizeof(buffer));
        printf("Error selecting sensor_id: %s\n", buffer);
        ias_db_query_close(query);
        ias_db_close_connection(db);
        ias_db_close_database_lib();
        query = NULL;
        return;
    }

    /*  get the first row returned to get the sensor ID */
    if (ias_db_query_next(query))
    {
        /* a row was returned, so read the sensor_id */
        ias_db_query_string_value(query, 0, sensor_id, 
                IAS_L0R_SENSOR_ID_LENGTH);
    }

    /* verify the sensor_id was empty */
    assert_string_equal(sensor_id, "NULL");

    ias_db_query_close(query);

    /* test to make sure sensor_id was not populated */
    snprintf (sql, sizeof(sql), 
            "select sensor_id from radiometric_work_order_common "
            "where work_order_id = '%s';", work_order_id);

    query = ias_db_perform_query(db, sql);
    if (!ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, buffer, sizeof(buffer));
        printf("Error selecting sensor_id: %s\n", buffer);
        ias_db_query_close(query);
        ias_db_close_connection(db);
        ias_db_close_database_lib();
        query = NULL;
        return;
    }

    /*  get the first row returned to get the sensor ID */
    if (ias_db_query_next(query))
    {
        /* a row was returned, so read the sensor_id */
        ias_db_query_string_value(query, 0, sensor_id,
                IAS_L0R_SENSOR_ID_LENGTH);
    }

    /* verify the sensor_id was empty */
    assert_string_equal(sensor_id, "NULL");

    ias_db_query_close(query);
    query = NULL;

    ias_db_close_connection(db);
    ias_db_close_database_lib();
}

Ensure test_ias_misc_allocate_2d_array()
{

    IAS_GRID **rptr;
    int rows = 5;     /* Both nrows and ncols could be evaluated */
    int cols = 8;    /* or read in at run time */
    int row, col;
    int status;

    rptr = (IAS_GRID **)ias_misc_allocate_2d_array(rows, cols, 
                                                 sizeof(IAS_GRID)); 
    assert_true(rptr != NULL);

    for (row = 0; row < rows; row++)
    {
        for (col = 0; col < cols; col++)
        {
            rptr[row][col].bandno = col + row;
            rptr[row][col].sca_number = 0.1*(row + col);
        }
    }

    for (row = 0; row < rows; row++)
    {
        for (col = 0; col < cols; col++)
        {
            rptr[row][col].test_model = malloc(sizeof(IAS_MODEL));
            rptr[row][col].test_model->model1 = row+col;
            rptr[row][col].test_model->model2 = 2.0*(row + col);
        }
    }

    for (row = 0; row < rows; row++)
    {
        for (col = 0; col < cols; col++)
        {
            assert_equal(rptr[row][col].bandno, col + row);
            assert_double_equal(rptr[row][col].sca_number, 0.1*(row + col));
        }
    }

    for (row = 0; row < rows; row++)
    {
        for (col = 0; col < cols; col++)
        {
            assert_equal(rptr[row][col].test_model->model1, row+col);
            assert_double_equal(rptr[row][col].test_model->model2,
                                2.0*(row + col));

           free(rptr[row][col].test_model);
           rptr[row][col].test_model = NULL;  
        }
    }

    status = ias_misc_get_2d_array_size((void **)rptr, &row, &col);
    assert_equal(status, SUCCESS);
    assert_equal(row, rows);
    assert_equal(col, cols);

    status = ias_misc_free_2d_array((void **)rptr);
    assert_equal(status, SUCCESS);
}

Ensure test_ias_misc_get_software_version()
{
    const char *ver;
    int length;
    ver = ias_misc_get_software_version();
    length = strlen(ver);
    assert_true_with_message
        (length > 0 && length < IAS_SOFTWARE_VERSION_SIZE,
        "IAS software version: '%s' has an invalid length: %d",
            ver, length);
    assert_string_equal(ver, IAS_VERSION);
}

Ensure test_ias_misc_get_processing_system()
{
    char *env_var;
    IAS_PROCESSING_SYSTEM_TYPE processing_system_type;

    processing_system_type = ias_misc_get_processing_system();
    env_var = getenv("PROCESSING_SYSTEM");
    if (!env_var)
    {
       assert_true_with_message(0, "PROCESSING_SYSTEM environment variable"
           " not set");
    }
    else
    {
        /* If we are running as IAS test that we get 
           IAS_PROCESSING_SYSTEM_IAS */
        if (strncmp(env_var, "IAS", 3) == 0)
        {
            assert_true_with_message( processing_system_type == 
                IAS_PROCESSING_SYSTEM_IAS,
                "processing system not IAS_PROCESSING_SYSTEM_IAS");
        }
        /* If we are running as LPGS test that we get
           IAS_PROCESSING_SYSTEM_LPGS */
        if (strncmp(env_var, "LPGS", 4) == 0)
        {
            assert_true_with_message( processing_system_type == 
                IAS_PROCESSING_SYSTEM_LPGS,
                "processing system not IAS_PROCESSING_SYSTEM_LPGS");
        }
    }
}

Ensure test_ias_misc_read_coregistered_single_band_l1g()
{
    int lines = 20;
    int samples = 10;
    IAS_DATA_TYPE data_type = IAS_I2;
    L1G_BAND_IO *l1g_band;
    int number_of_scas = 1;      /* number of SCAs */
    L1GIO *img;
    IAS_L1G_BAND_METADATA bmd;
    short *image_data;
    short int data[200];
    int i, status;
    IAS_DATA_TYPE dtype;  
    IAS_IMAGE image;

    int x = 186;
    int y = 86;
    int z = 1;
    
    IAS_CORNERS corners;

    /* open the test image */
    img = ias_l1g_open_image("test.h5", IAS_WRITE);
    assert_true(img != NULL);

    /* open the test band */
    l1g_band = ias_l1g_open_band(img, 1, &data_type, &number_of_scas,
                                   &lines, &samples);
    assert_true(l1g_band != NULL);

    /* write some test image data */
    for (i = 0; i < lines * samples; i++)
        data[i] = 30 + i;

    /* write some band metadata */
    bmd.band_number = 1;
    strcpy(bmd.band_name, "RED");
    bmd.upper_left_y = 100.1;
    bmd.upper_left_x = 200.1;
    bmd.upper_right_y = 100.1;
    bmd.upper_right_x = 250.1;
    bmd.lower_left_y = 150.1;
    bmd.lower_left_x = 200.1;
    bmd.lower_right_y = 150.1;
    bmd.lower_right_x = 250.1;
    bmd.projection_distance_x = 10.0;
    bmd.projection_distance_y = 10.0;
    bmd.maximum_pixel_value = 50.0;
    bmd.minimum_pixel_value = 0.0;
    bmd.pixel_range_valid = 1;
    bmd.maximum_radiance = 50.5;
    bmd.minimum_radiance = 2.1;
    bmd.spectral_radiance_scaling_offset = 0.2;
    bmd.spectral_radiance_scaling_gain = 0.99;
    bmd.radiance_valid = 1;
    bmd.reflectance_scaling_offset = 0.1;
    bmd.reflectance_scaling_gain = 1.1;
    bmd.reflectance_valid = 1;
    strcpy(bmd.instrument_source, "OLI Red Band");
    status = ias_l1g_set_band_metadata(img, &bmd, 1);
    assert_equal(status, SUCCESS);

    /* test writing the imagery */
    status = ias_l1g_write_image(l1g_band, 0, 0, 0, lines, samples, data);
    assert_equal(status, SUCCESS);

    /* close the test band */
    status = ias_l1g_close_band(l1g_band);
    assert_equal(status, SUCCESS);

    /* close the test image */
    status = ias_l1g_close_image(img);
    assert_equal(status, SUCCESS);

    /* Populate values for IAS_CORNERS structure */
    corners.upleft.x = 200.1;
    corners.upleft.y = 100.1;
    corners.upright.x = 250.1;
    corners.upright.y = 100.1;
    corners.loleft.x = 200.1;
    corners.loleft.y = 150.1;
    corners.loright.x = 250.1;
    corners.loright.y = 150.1;

    /* test ias_misc_read_dem */
    status = ias_misc_read_coregistered_single_band_l1g("test.h5", &corners,
        x, y, z, &image, IAS_I2);
    assert_equal(status, SUCCESS);
    assert_equal(image.nl, 20);
    assert_equal(image.ns, 10);
    assert_double_equal(image.pixel_size_x, 10.0);
    assert_double_equal(image.pixel_size_y, 10.0);
    assert_double_equal(image.corners.upleft.x, 200.1);
    assert_double_equal(image.corners.upleft.y, 100.1);
    assert_double_equal(image.corners.upright.x, 250.1);
    assert_double_equal(image.corners.upright.y, 100.1);
    assert_double_equal(image.corners.loleft.x, 200.1);
    assert_double_equal(image.corners.loleft.y, 150.1);
    assert_double_equal(image.corners.loright.x, 250.1);
    assert_double_equal(image.corners.loright.y, 150.1);

    image_data = (short *)image.data;

    for (i = 0; i < lines * samples; i++)
        assert_equal(image_data[i], 30 + i);
    free(image.data);
    image.data = NULL;

    /* test ias_misc_read_single_band_l1g */
    status = ias_misc_read_single_band_l1g("test.h5", 1,  
                                                    &dtype, &image);
    assert_equal(status, SUCCESS);
    assert_equal(dtype, IAS_I2);
    assert_equal(image.nl, 20);
    assert_equal(image.ns, 10);
    assert_double_equal(image.pixel_size_x, 10.0);
    assert_double_equal(image.pixel_size_y, 10.0);
    assert_double_equal(image.corners.upleft.x, 200.1);
    assert_double_equal(image.corners.upleft.y, 100.1);
    assert_double_equal(image.corners.upright.x, 250.1);
    assert_double_equal(image.corners.upright.y, 100.1);
    assert_double_equal(image.corners.loleft.x, 200.1);
    assert_double_equal(image.corners.loleft.y, 150.1);
    assert_double_equal(image.corners.loright.x, 250.1);
    assert_double_equal(image.corners.loright.y, 150.1);

    /* Reassign the pointer */
    image_data = (short *)image.data;

    for (i = 0; i < lines * samples; i++)
        assert_equal(image_data[i], 30 + i);
    free(image.data);
    image.data = NULL;
}


Ensure test_ias_misc_parse_datetime_string()
{
    char datetime_string[DATETIME_STR_LENGTH];
    int test_year;
    int test_month;
    int test_day;
    int test_hour;
    int test_minute;
    int test_day_of_year;
    double test_second;
    int status;
    IAS_DATETIME datetime;
    IAS_DATETIME_FORMAT_TYPE format_type;

    /* Initialize */
    memset(datetime_string, 0, DATETIME_STR_LENGTH);
    memset(&datetime, 0, sizeof(IAS_DATETIME));
    format_type = IAS_DATETIME_CPF_FORMAT;

    test_year = 2005;
    test_day_of_year = 184;
    test_month = 7;
    test_day = 3;
    test_hour = 17;
    test_minute = 0;
    test_second = 45.63452;

    /* Parse the string */
    snprintf(datetime_string, sizeof(datetime_string), 
            "%04d-%02d-%02dT%02d:%02d:%8.5lf", test_year, test_month, 
            test_day, test_hour, test_minute, test_second);
    status = ias_misc_parse_datetime_string((const char *)datetime_string,
            format_type, &datetime);
    assert_true_with_message((status == SUCCESS), "Date/time parsing failed\n");

    /* Check the results */
    assert_equal(test_year, datetime.year);
    assert_equal(test_month, datetime.month);
    assert_equal(test_day, datetime.day_of_month);
    assert_equal(test_hour, datetime.hour);
    assert_equal(test_minute, datetime.minute);
    assert_double_equal(test_second, datetime.second);
    assert_equal(test_day_of_year, datetime.day_of_year);

    /* Same CPF datetime/string, only this time, no fractional seconds */
    test_second = 46.0;

    snprintf(datetime_string, sizeof(datetime_string), 
            "%04d-%02d-%02dT%02d:%02d:%02d", test_year, test_month, 
            test_day, test_hour, test_minute, (int)test_second);
    status = ias_misc_parse_datetime_string((const char *)datetime_string,
                format_type, &datetime);
    assert_true_with_message((status == SUCCESS), "Date/time parsing failed\n");

    /* Check the results */
    assert_equal(test_year, datetime.year);
    assert_equal(test_month, datetime.month);
    assert_equal(test_day, datetime.day_of_month);
    assert_equal(test_hour, datetime.hour);
    assert_equal(test_minute, datetime.minute);
    assert_double_equal(test_second, datetime.second);
    assert_equal(test_day_of_year, datetime.day_of_year);

    /* Test for no leading zeros */
    test_year = 2005;
    test_day_of_year = 184;
    test_month = 7;
    test_day = 3;
    test_hour = 7;
    test_minute = 1;
    test_second = 5.63452;

    snprintf(datetime_string, sizeof(datetime_string), "%d-%d-%dT%d:%d:%.5lf",
            test_year, test_month, test_day, test_hour, test_minute,
            test_second);
    status = ias_misc_parse_datetime_string((const char *)datetime_string,
            format_type, &datetime);
    assert_true_with_message((status == SUCCESS), "Date/time parsing failed\n");

    /* Check the results */
    assert_equal(test_year, datetime.year);
    assert_equal(test_month, datetime.month);
    assert_equal(test_day, datetime.day_of_month);
    assert_equal(test_hour, datetime.hour);
    assert_equal(test_minute, datetime.minute);
    assert_double_equal(test_second, datetime.second);
    assert_equal(test_day_of_year, datetime.day_of_year);

    /* Another test, this time with a L0R date format */
    memset(datetime_string, 0, DATETIME_STR_LENGTH);
    memset(&datetime, 0, sizeof(IAS_DATETIME));
    format_type = IAS_DATETIME_L0R_FORMAT;

    test_month = 12;
    test_day = 31;
    test_day_of_year = 366;
    test_year = 2012;
    test_hour = 16;
    test_minute = 15;
    test_second = 27.04520;
    snprintf(datetime_string, sizeof(datetime_string), 
            "%04d:%03d:%02d:%02d:%13.10lf", test_year, test_day_of_year, 
            test_hour, test_minute, test_second);

    status = ias_misc_parse_datetime_string((const char *)datetime_string,
                format_type, &datetime);
    assert_true(status == SUCCESS);

    /* Check the second test's results */
    assert_equal(test_year, datetime.year);
    assert_equal(test_day_of_year, datetime.day_of_year);
    assert_equal(test_hour, datetime.hour);
    assert_equal(test_minute, datetime.minute);
    assert_double_equal(test_second, datetime.second);;
    assert_equal(test_month, datetime.month);
    assert_equal(test_day, datetime.day_of_month);

    /* Error cases... */

    /* Disable the logging since remaining tests are for invalid cases to
       ensure failures happen as expected. */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* 1. Invalid IAS_DATETIME_FORMAT_TYPE identifier */
    memset(&datetime, 0, sizeof(IAS_DATETIME));
    format_type = IAS_DATETIME_TYPE_ERROR;

    status = ias_misc_parse_datetime_string((const char *)datetime_string,
            format_type, &datetime);
    assert_true_with_message((status == ERROR),
            "Invalid format type, string should not have been parsed\n");

    /* 2.  Invalid character in one of the numbers */
    format_type = IAS_DATETIME_L0R_FORMAT;
    datetime_string[6] = 'A';  /* Use the L0R format string in test #2,
                                  but change one of the numbers in the
                                  day of year to a letter */
    status = ias_misc_parse_datetime_string((const char *)datetime_string,
            format_type, &datetime);
    assert_true_with_message((status == ERROR),
            "Invalid character in string, parsing should have failed\n");

    /* 3. Invalid separator character */
    format_type = IAS_DATETIME_CPF_FORMAT;
    memset(datetime_string, 0, DATETIME_STR_LENGTH);

    snprintf(datetime_string, sizeof(datetime_string), 
            "%4d-%02d-%02d-%02d:%02d:%8.5lf", test_year, test_month, 
            test_day, test_hour, test_minute, test_second);
    status = ias_misc_parse_datetime_string((const char *)datetime_string,
            format_type, &datetime);
    assert_true_with_message((status == ERROR),
            "Invalid separator character, parsing should have failed\n");

    /* 4. Invalid month */
    memset(datetime_string, 0, DATETIME_STR_LENGTH);
    test_month = 18;
    snprintf(datetime_string, sizeof(datetime_string),
            "%04d-%02d-%02dT%02d:%02d:%8.5lf", test_year, test_month,
            test_day, test_hour, test_minute, test_second);
    status = ias_misc_parse_datetime_string((const char *)datetime_string,
            format_type, &datetime);
    assert_true_with_message((status == ERROR ), "Month range check failed\n");

    /* 5. Invalid leap year */
    memset(datetime_string, 0, DATETIME_STR_LENGTH);
    test_year = 2011;
    test_month = 2;
    test_day = 29;
    test_hour = 16;
    test_minute = 15;
    test_second = 27.04520;
    snprintf(datetime_string, sizeof(datetime_string),
            "%04d-%02d-%02dT%02d:%02d:%8.5lf", test_year, test_month,
            test_day, test_hour, test_minute, test_second);
    status = ias_misc_parse_datetime_string((const char *)datetime_string,
            format_type, &datetime);
    assert_true_with_message((status == ERROR ), "Leap year check failed\n");

    /* 6. Invalid hour */
    memset(datetime_string, 0, DATETIME_STR_LENGTH);
    test_year = 2011;
    test_month = 2;
    test_day = 28;
    test_hour = 24;
    test_minute = 15;
    test_second = 27.04520;
    snprintf(datetime_string, sizeof(datetime_string),
            "%04d-%02d-%02dT%02d:%02d:%8.5lf", test_year, test_month,
            test_day, test_hour, test_minute, test_second);
    status = ias_misc_parse_datetime_string((const char *)datetime_string,
            format_type, &datetime);
    assert_true_with_message((status == ERROR ), "hour check failed\n");

    /* 7. Invalid minute */
    memset(datetime_string, 0, DATETIME_STR_LENGTH);
    test_year = 2011;
    test_month = 2;
    test_day = 28;
    test_hour = 16;
    test_minute = 60;
    test_second = 27.04520;
    snprintf(datetime_string, sizeof(datetime_string),
            "%04d-%02d-%02dT%02d:%02d:%8.5lf", test_year, test_month,
            test_day, test_hour, test_minute, test_second);
    status = ias_misc_parse_datetime_string((const char *)datetime_string,
            format_type, &datetime);
    assert_true_with_message((status == ERROR ), "minute check failed\n");

    /* 8. Invalid second */
    memset(datetime_string, 0, DATETIME_STR_LENGTH);
    test_year = 2011;
    test_month = 2;
    test_day = 28;
    test_hour = 16;
    test_minute = 15;
    test_second = 60.00;
    snprintf(datetime_string, sizeof(datetime_string),
            "%04d-%02d-%02dT%02d:%02d:%8.5lf", test_year, test_month,
            test_day, test_hour, test_minute, test_second);
    status = ias_misc_parse_datetime_string((const char *)datetime_string,
            format_type, &datetime);
    assert_true_with_message((status == ERROR ), "second check failed\n");

    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);
}


/* test the ias_misc_check_file_type routine */
Ensure test_check_file_type()
{
    #define TEST_FILE "test.h5"
    #define FILE_TYPE "Test File"

    hid_t file_id;  /* HDF file id */
    int status;

    /* test on a file that doesn't exist */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = ias_misc_check_file_type("this_file_does_not_exist.h5", FILE_TYPE);
    assert_false(status);
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

    /* open the test file */
    file_id = H5Fcreate(TEST_FILE, H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT);
    assert_true(file_id >= 0);

    /* close the test file without writing an attribute to make sure it
       handles it gracefully */
    H5Fclose(file_id);
    file_id = -1;

    /* expect a failure, so turn off logging output */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* make sure the file type check returns 0 */
    status = ias_misc_check_file_type(TEST_FILE, FILE_TYPE);
    assert_false(status);

    /* reenable logging output */
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);

    /* open the test file */
    file_id = H5Fcreate(TEST_FILE, H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT);
    assert_true(file_id >= 0);

    /* write the attribute to the file */
    status = H5LTset_attribute_string(file_id, "/", IAS_FILE_TYPE_ATTRIBUTE, 
            FILE_TYPE);
    assert_true(status >= 0);

    /* close the file */
    H5Fclose(file_id);
    file_id = -1;

    /* verify the file type can be read */
    status = ias_misc_check_file_type(TEST_FILE, FILE_TYPE);
    assert_true(status);

    /* verify the wrong file type fails */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = ias_misc_check_file_type(TEST_FILE, "Wrong Type");
    assert_false(status);
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);
}


/* test the ias_misc_convert_data_type routines */
Ensure test_ias_misc_convert_data_type()
{
    int status;
    const char *dtype_string;
    IAS_DATA_TYPE dtype;

    for (dtype = IAS_BYTE; dtype < IAS_NDTYPES; dtype++)
    {
        IAS_DATA_TYPE new_dtype;

        status = ias_misc_convert_data_type_to_string(dtype, &dtype_string);
        assert_equal(status, SUCCESS);
        if (status != SUCCESS)
            return;

        switch(dtype)
        {
            case IAS_BYTE:
                assert_string_equal(dtype_string, "BYTE");
                break;
            case IAS_I2:
                assert_string_equal(dtype_string, "I*2");
                break;
            case IAS_UI2:
                assert_string_equal(dtype_string, "UI*2");
                break;
            case IAS_I4:
                assert_string_equal(dtype_string, "I*4");
                break;
            case IAS_UI4:
                assert_string_equal(dtype_string, "UI*4");
                break;
            case IAS_R4:
                assert_string_equal(dtype_string, "R*4");
                break;
            case IAS_R8:
                assert_string_equal(dtype_string, "R*8");
                break;
            case IAS_CHAR:
                assert_string_equal(dtype_string, "CHAR");
                break;
            case IAS_UCHAR:
                assert_string_equal(dtype_string, "UCHAR");
                break;
            default:
                assert_true_with_message(0, "Unknown data type %d", dtype);
        }

        /* Now test the conversion back the other direction */
        status = ias_misc_convert_string_to_data_type(dtype_string, &new_dtype);
        assert_equal(status, SUCCESS);
        if (status == SUCCESS)
            assert_equal(new_dtype, dtype);
    }
}

Ensure test_get_doq_reference_filename()
{
    int status;
    char filename[PATH_MAX];
    int old_log_level;
    const char *utm_filename = "./ias028029_UTM_doq.h5";
    const char *som_filename = "./ias028029_SOM_doq.h5";
    FILE *file;

    /* Make sure the temporary files do not exist */
    unlink(utm_filename);
    unlink(som_filename);

    /* Override the IAS_IMG environment variable to point to the current
       directory */
    status = setenv("IAS_IMG", ".", 1);
    assert_equal(status, 0);
    if (status != 0)
        return;

    /* Set the logging level to disable the logging since these tests should
       produce error messages */
    old_log_level = ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* Test with a file that shouldn't exist.  It should return an error. */
    status = ias_misc_get_doq_reference_filename(28, 29, filename, 
        sizeof(filename));
    assert_equal(status, ERROR);

    ias_log_set_output_level(old_log_level);

    /* Create a dummy file with the SOM name */
    file = fopen(som_filename, "w");
    assert_not_equal(file, NULL);
    if (file == NULL)
        return;
    fclose(file);

    /* Make sure the SOM filename is found */
    status = ias_misc_get_doq_reference_filename(28, 29, filename, 
        sizeof(filename));
    assert_equal(status, SUCCESS);
    assert_string_equal(filename, som_filename);

    /* Create a dummy file with the UTM name */
    file = fopen(utm_filename, "w");
    assert_not_equal(file, NULL);
    if (file == NULL)
        return;
    fclose(file);

    /* Make sure the UTM filename is found */
    status = ias_misc_get_doq_reference_filename(28, 29, filename, 
        sizeof(filename));
    assert_equal(status, SUCCESS);
    assert_string_equal(filename, utm_filename);

    /* Set the logging level to disable the logging since these tests should
       produce error messages */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* Test with a buffer that is too small */
    status = ias_misc_get_doq_reference_filename(28, 29, filename, 2);
    assert_equal(status, ERROR);

    /* Unset the IAS_IMG environment variable and confirm it is detected as
       an error */
    status = unsetenv("IAS_IMG");
    assert_equal(status, 0);
    if (status != 0)
        return;

    status = ias_misc_get_doq_reference_filename(28, 29, filename,
        sizeof(filename));
    assert_equal(status, ERROR);

    /* Set the environment variable to an empty string and make sure it is
       detected as an error */
    status = setenv("IAS_IMG", "", 1);
    assert_equal(status, 0);
    if (status != 0)
        return;

    status = ias_misc_get_doq_reference_filename(28, 29, filename,
        sizeof(filename));
    assert_equal(status, ERROR);

    ias_log_set_output_level(old_log_level);

    /* Clean up the temporary files */
    unlink(utm_filename);
    unlink(som_filename);
}

Ensure test_image_calculate_trim_box_and_lut()
{
    IAS_GRID_TYPE grid;                 /* Grid structure */
    char grid_filename[PATH_MAX];       /* Grid filename */
    int status;                         /* Status variable */
    int index;                          /* Loop variable */
    int val_index = 0;                  /* Validation loop variable */
    double line[4];                     /* Line corners */
    double samp[4];                     /* Sample Corners */
    int output_lines;                   /* Number of lines in grid */
    int output_samples;                 /* Number of samps in grid */
    IAS_MISC_LINE_EXTENT *trim_lut;     /* Lookup table pointer */    

    /* Validation values for trim box */
    double validation[10] = { 7.1361889005, 412.9081154269, 1317.3649656506, 
        6645.7587284763, 3257.3509633325, 6234.2876587727, 1926.8964971753, 
        6.4592377565, 7.1361889005, 412.9081154269 };

    if (ias_sat_attr_initialize(IAS_L8) != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing satellite attributes library");
        assert_equal(SUCCESS, ERROR);
        return;
    }

    /* Create the grid filename */
    status = snprintf(grid_filename, sizeof(grid_filename), "%s/%s",
        getenv("IAS_UNIT_TEST_DATA_DIR"), "ias_lib/misc/miscellaneous/"
        "ias_grid.h5");
    if (status < 0 || status >= sizeof(grid_filename))
    {
        IAS_LOG_ERROR("Unable to construct diff command");
        assert_equal(SUCCESS, ERROR);
        return;
    }

    /* Read the full grid structure */
    if (ias_grid_read(grid_filename, NULL, 0, &grid) != SUCCESS)
    {
        IAS_LOG_ERROR("Reading the grid file %s.", grid_filename);
        assert_equal(SUCCESS, ERROR);
        return;
    }

    status = ias_misc_calculate_trim_box(0, &grid.active_corners[0],
        &grid, line, samp);

    /* Validate all the corners */
    for (index = 0; index < 4; index++)
    {
        assert_double_equal(line[index], validation[val_index++]);
        assert_double_equal(samp[index], validation[val_index++]);
    }

    output_lines = grid.gridbands[0][0].lines;
    output_samples = grid.gridbands[0][0].samps;

    /* Create the lookup table */
    trim_lut = ias_misc_create_output_image_trim_lut(line, samp, 
        output_lines, output_samples);
    if (!trim_lut)
    {
        IAS_LOG_ERROR("Creating the lookup table");
        assert_equal(SUCCESS, ERROR);
        ias_grid_free(&grid);
        return;
    }
    
    /* Validate the start and end samples */
    for (index = 0, val_index = 0; index < output_lines; index++)
    {
        assert_equal_with_message(trim_lut[index].start_sample,
                expected_trim_lut[index].start_sample,
                "Index %d: start sample expected %d, got %d", index,
                trim_lut[index].start_sample, 
                expected_trim_lut[index].start_sample);
        assert_equal_with_message(trim_lut[index].end_sample,
                expected_trim_lut[index].end_sample,
                "Index %d: end sample expected %d, got %d", index,
                trim_lut[index].end_sample, 
                expected_trim_lut[index].end_sample);
    }
    
    free(trim_lut);
    ias_grid_free(&grid);
}

Ensure test_write_envi_image_and_header()
{
    const char *image_filename = "test_image.img";
    const char *verified_image_filename = "ver_test_image.img";
    const char *header_filename = "test_image.img.hdr";
    const char *verified_header_filename = "ver_test_image.img.hdr";
    const char *verified_header_filename2 = "ver_test_image.img.hdr2";
    IAS_DATA_TYPE data_type = IAS_UI2;
    IAS_UI2_TYPE data_buffer[10000];
    int number_of_lines = 100;
    int number_of_samps = 100;
    double upper_left_x = 14030.00;
    double upper_left_y = 50002.00;
    double proj_dist_x = 30.0;
    double proj_dist_y = 30.0;
    IAS_PROJECTION projection;
    double projection_parameters[IAS_PROJ_PARAM_SIZE];
    int index;
    char command[PATH_MAX];

    /* Set the data buffer values */
    for (index = 0; index < 10000; index++)
            data_buffer[index] = (IAS_UI2_TYPE)index;
  
    /* Create a UTM projection */
    for (index = 0; index < IAS_PROJ_PARAM_SIZE; index++)
        projection_parameters[index] = 0.0;
    
    ias_geo_set_projection(UTM, 19, DEGREE, WGS84_SPHEROID,
        projection_parameters, &projection);

    /* Test that the envi image and header has been created correctly */
    if (ias_misc_write_envi_image(image_filename, NULL, 
        "Testing ias_misc_write_envi_image", number_of_lines, number_of_samps,
        upper_left_x, upper_left_y, proj_dist_x, proj_dist_y, "Test Band",
        data_type, data_buffer) != SUCCESS)
    {
        assert_true_with_message(0, "Error writing image to file");
        return;
    }
   
    /* Verify the output image */
    index = snprintf(command, sizeof(command), "diff -q %s envi/%s", 
        image_filename, verified_image_filename);
    if (index < 0 || index >= sizeof(command))
    {   
        assert_true_with_message(0, "Error creating diff command");
        return;
    } 
    else if (system(command))
    {
        assert_true_with_message(0, "Error running command: %s", command);
        return;
    }

    /* Verify the output header */
    index = snprintf(command, sizeof(command), "diff -q %s envi/%s", 
        header_filename, verified_header_filename);
    if (index < 0 || index >= sizeof(command))
    {   
        assert_true_with_message(0, "Error creating diff command");
        return;
    } 
    else if (system(command))
    {
        assert_true_with_message(0, "Error running command: %s", command);
        return;
    }

    /* Test that the envi header has been created correctly */
    if (ias_misc_write_envi_header(image_filename, &projection, 
        "Testing ias_misc_write_envi_header", number_of_lines, number_of_samps,
        1, upper_left_x, upper_left_y, proj_dist_x, proj_dist_y, "Test Band2",
        data_type) != SUCCESS)
    {
        assert_true_with_message(0, "Error writing header to file");
        return;
    }
   
    /* Verify the output header */
    index = snprintf(command, sizeof(command), "diff -q %s envi/%s", 
        header_filename, verified_header_filename2);
    if (index < 0 || index >= sizeof(command))
    {   
        assert_true_with_message(0, "Error creating diff command");
        return;
    } 
    else if (system(command))
    {
        assert_true_with_message(0, "Error running command: %s", command);
        return;
    }
}


/* globals */

/* initialize to false if user database up, need to set ENABLE_DB_TESTS
   in user env to run tests */
static int test_database = FALSE;

int main(int argc, char **argv)
{
    char *enable_db_tests;

    /* See if the database tests should be enabled */
    enable_db_tests = getenv( "ENABLE_DB_TESTS" );
    if ( enable_db_tests )
    {
        test_database = TRUE;
        printf( "\n\tDatabase checks have been enabled\n" );
    }
    else
        printf( "\n\t***** Database checks will not be made *****\n" );
    


    TestSuite *suite = create_test_suite();

    add_test(suite, test_check_file_type);
    add_test(suite, test_ias_misc_allocate_2d_array);
    add_test(suite, test_ias_misc_get_software_version);
    add_test(suite, test_ias_misc_get_processing_system);
    add_test(suite, test_ias_misc_parse_datetime_string);
    add_test(suite, test_ias_misc_read_coregistered_single_band_l1g);
    add_test(suite, test_ias_misc_convert_data_type);
    add_test(suite, test_get_doq_reference_filename);
    add_test(suite, test_image_calculate_trim_box_and_lut);
    add_test(suite, test_write_envi_image_and_header);

    if (enable_db_tests)
        add_test(suite, test_ias_misc_get_char_id);

    if (argc > 1)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }

    return run_test_suite(suite, create_text_reporter());
}
