#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ias_const.h"
#include "ias_linked_list.h"
#include "ias_pixel_mask.h"
#include "../pm_local.h"
#include "ias_radiometry.h"

#include "cgreen.h"

#define TIRS_BLIND_BAND 15     /* Current blind band number */
#define TIRS_THERMAL1_BAND 10


Ensure test_pm_get_mask_index()
{
    PIXEL_MASK_TYPE mask = PM_NOVALUE;

    IAS_PIXEL_MASK **pm_array = NULL;

    int band_number = 0;
    int band_index = 0;
    int num_bands  = 3;
    int num_scas   = 3;
    int sca_index  = 0;
    int sca_number = 0;
    int mask_index = 0;
    int num_masks = 0;
    int num_detectors =  640;
    int num_pixels = 2;
    int ii;                      /* Local loop counter */
    int status = SUCCESS;


    printf("Testing mask array index identification...\n");

    /* Construct a 9-element pixel mask array.  This is effectively
       a simulated TIRS pixel mask array */
    num_masks = num_bands * num_scas;
    pm_array = calloc(num_masks, sizeof(IAS_PIXEL_MASK *));
    assert_not_equal_with_message(pm_array, NULL, "Cannot allocate base "
        "pixel mask array");
    if (pm_array == NULL)
        return;

    /* Create each pixel mask in the array */
    for (band_index = 0; band_index < num_bands; band_index++)
    {
        /* Use current TIRS band numbers */
        if (band_index == num_bands - 1)
            band_number = TIRS_BLIND_BAND;
        else
            band_number = TIRS_THERMAL1_BAND + band_index;

        for (sca_index = 0; sca_index < num_scas; sca_index++)
        {
            sca_number = sca_index + 1;

            mask_index = (band_index * num_scas) + sca_index;
            pm_array[mask_index] = ias_pm_create(band_number,
                sca_number, num_detectors, num_pixels);
            if (pm_array[mask_index] == NULL)
            {
                 assert_true_with_message(0, "Cannot create empty mask "
                     "for band %d SCA %d", band_number, sca_number);
                 for (ii = 0; ii < mask_index; ii++)
                     ias_pm_destroy(pm_array[mask_index]);
                 free(pm_array);
                 return;
            }
        }
    }

    /* Let's make a few masks have artifacts in them.  This will provide
       better confirmation that the correct mask index was found */

    /* Band 10, SCA 2: Detector 226 inoperable */
    status = ias_pm_add_pixels(pm_array[1], 225, 0, num_pixels,
        IAS_RPS_INOP_DETECTOR);
    assert_equal_with_message(status, SUCCESS, "Cannot add inoperable "
        "detector to mask");
    if (status != SUCCESS)
    {
        for (mask_index = 0; mask_index < num_masks; mask_index++)
            ias_pm_destroy(pm_array[mask_index]);
        free(pm_array);
        return;
    }

    /* Band 11, SCA 3:  Detectors 245 and 247, pixel 2:  impulse noise */
    status = ias_pm_add_pixels(pm_array[5], 244, 1, 1, IAS_RPS_IMPNOISE);
    assert_equal_with_message(status, SUCCESS, "Cannot add impulse noise "
        "to mask");
    if (status != SUCCESS)
    {
        for (mask_index = 0; mask_index < num_masks; mask_index++)
            ias_pm_destroy(pm_array[mask_index]);
        free(pm_array);
        return;
    }
    status = ias_pm_add_pixels(pm_array[5], 246, 1, 1, IAS_RPS_IMPNOISE);
    assert_equal_with_message(status, SUCCESS, "Cannot add impulse noise "
        "to mask");
    if (status != SUCCESS)
    {
        for (mask_index = 0; mask_index < num_masks; mask_index++)
            ias_pm_destroy(pm_array[mask_index]);
        free(pm_array);
        return;
    }

    /* Test 1:  Band 15, SCA 1.  Empty pixel mask.  Should return
       mask index of 6 (7th element in array) */
    band_number = TIRS_BLIND_BAND;
    sca_number  = 1;
    mask_index = ias_pm_get_mask_index(pm_array, num_masks, band_number,
        sca_number);
    assert_equal_with_message(mask_index, 6, "Did not find band %d SCA %d "
        "mask in array", band_number, sca_number);
    if (mask_index != 6)
    {
        for (mask_index = 0; mask_index < num_masks; mask_index++)
            ias_pm_destroy(pm_array[mask_index]);
        free(pm_array);
        return;
    }

    /* Test 2:  Band 10, SCA 2.  Inoperable detector 226.  Should return
       mask index of 1 (2nd array element) */
    band_number = TIRS_THERMAL1_BAND;
    sca_number  = 2;
    mask_index = ias_pm_get_mask_index(pm_array, num_masks, band_number,
        sca_number);
    assert_equal_with_message(mask_index, 1, "Did not find band %d SCA %d "
        "mask", band_number, sca_number);
    if (mask_index != 2)
    {
        for (mask_index = 0; mask_index < num_masks; mask_index++)
            ias_pm_destroy(pm_array[mask_index]);
        free(pm_array);
        return;
    }
    status = ias_pm_get_mask_at(pm_array[mask_index], 225, 1, &mask);
    if ((status != SUCCESS) || (mask != IAS_RPS_INOP_DETECTOR))
    {
        assert_true_with_message(FALSE, "Cannot locate flagged pixel for "
            "inoperable detector");
        for (mask_index = 0; mask_index < num_masks; mask_index++)
            ias_pm_destroy(pm_array[mask_index]);
        free(pm_array);
        return;
    }

    /* Test 3:  Band 11, SCA 3.  Impulse noise at 2nd pixel in detectors
       245 and 247.  Should return mask index of 5. */
    band_number = TIRS_THERMAL1_BAND + 1;
    sca_number  = 3;
    mask_index = ias_pm_get_mask_index(pm_array, num_masks, band_number,
        sca_number);
    assert_equal_with_message(mask_index, 5, "Did not find band %d SCA %d "
        "mask", band_number, sca_number);
    if (mask_index != 5)
    {
        for (mask_index = 0; mask_index < num_masks; mask_index++)
            ias_pm_destroy(pm_array[mask_index]);
        free(pm_array);
        return;
    }
    status = ias_pm_get_mask_at(pm_array[mask_index], 244, 1, &mask);
    if ((status != SUCCESS) || (mask != IAS_RPS_IMPNOISE))
    {
        assert_true_with_message(FALSE, "Cannot locate flagged pixel for "
            "impulse noise");
        for (mask_index = 0; mask_index < num_masks; mask_index++)
            ias_pm_destroy(pm_array[mask_index]);
        free(pm_array);
        return;
    }
    status = ias_pm_get_mask_at(pm_array[mask_index], 246, 1, &mask);
    if ((status != SUCCESS) || (mask != IAS_RPS_IMPNOISE))
    {
        assert_true_with_message(FALSE, "Cannot locate flagged pixel for "
            "impulse noise");
        for (mask_index = 0; mask_index < num_masks; mask_index++)
            ias_pm_destroy(pm_array[mask_index]);
        free(pm_array);
        return;
    }

    /* Test 4:  Totally wrong band number, 4.  Should return error code
       of -1 */
    band_number = 4;
    sca_number = 3;
    mask_index = ias_pm_get_mask_index(pm_array, num_masks, band_number,
        sca_number);
    assert_equal_with_message(mask_index, -1, "Should have returned "
        "ERROR for nonexistent band");
    if (mask_index != ERROR)
    {
        for (mask_index = 0; mask_index < num_masks; mask_index++)
            ias_pm_destroy(pm_array[mask_index]);
        free(pm_array);
        return;
    }

    /* Test 5:  Band 10, totally wrong SCA number, 10.  Should return
       error code of -1 */
    mask_index = 0;
    band_number = TIRS_THERMAL1_BAND;
    sca_number = 10;
    mask_index = ias_pm_get_mask_index(pm_array, num_masks, band_number,
        sca_number);
    assert_equal_with_message(mask_index, -1, "Should have returned "
        "ERROR for nonexistent SCA");

    /* Clean up */
    for (mask_index = 0; mask_index < num_masks; mask_index++)
        ias_pm_destroy(pm_array[mask_index]);
    free(pm_array);

}




int main(int argc, char *argv[])
{
    TestSuite *suite = create_test_suite();
    add_test(suite, test_pm_get_mask_index);

    if (argc > 1)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}
