#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ias_const.h"
#include "ias_linked_list.h"
#include "ias_pixel_mask.h"
#include "ias_logging.h"
#include "../pm_local.h"
#include "ias_radiometry.h"

#include "cgreen.h"

Ensure test_pm_is_detector_flagged()
{
    IAS_PIXEL_MASK           *pm_test = NULL;
    PIXEL_MASK_TYPE          mask;

    int num_lines = 100;
    int num_dets  = 100;
    int det_index;
    int status;

    int is_in_mask;

    /* Create an empty pixel mask for "band 1, SCA 1", with
       100 detectors x 100 lines */
    pm_test = ias_pm_create(1, 1, num_dets, num_lines);
    assert_true(pm_test != NULL);
    if (pm_test == NULL)
    {
        printf("Cannot create test pixel mask\n");
        return;
    }


    /* First test:  look for detector(s) with dropped frames in an
       empty pixel mask */
    mask = IAS_RPS_DROPLINE;
    for (det_index = 0; det_index < num_dets; det_index++)
    {
        is_in_mask
            = ias_pm_is_detector_flagged(pm_test, det_index, mask);
        assert_true(is_in_mask == FALSE);
        if(is_in_mask)
        {
            printf("Should be empty pixel mask, check algorithm logic\n");
            ias_pm_destroy(pm_test);
            return;
        }
    }

    /* Second test:  simulate the case of a 5 pixel sequence of dropped
       frames starting at line 64 in detector 25 */
    mask = IAS_RPS_DROPLINE;
    status = ias_pm_add_pixels((void *)pm_test, 24, 63, 5, mask);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("Cannot simulate dropped line in pixel mask\n");
        ias_pm_destroy(pm_test);
        return;
    }
    is_in_mask = ias_pm_is_detector_flagged(pm_test, 24, mask);
    assert_true(is_in_mask == TRUE);
    if (!is_in_mask)
    {
        printf("Should be a partial dropped line in pixel mask at detector "
               "25\n");
        ias_pm_destroy(pm_test);
        return;
    }

    /* Third test:  look for simulated impulse noise.  There shouldn't be
       any.  There should be a warning issued because detector 25 was
       flagged with dropped frames in test #2 */
    mask = IAS_RPS_IMPNOISE;
    for (det_index = 0; det_index < num_dets; det_index++)
    {
        is_in_mask = ias_pm_is_detector_flagged(pm_test, det_index, mask);
        assert_true(is_in_mask == FALSE);
        if (is_in_mask)
        {
            printf("Should not be any impulse noise artifacts flagged in "
                   "pixel mask\n");
            ias_pm_destroy(pm_test);
            return;
        }
    }

    /* This test is the result of the discovery of a logic issue during
       preliminary review of the routine... */

    /* Fourth test:  add simulated impulse noise and another artifact at
       detector 40, starting at frame 21 going for 40 pixels, then test
       just for impulse noise.
       The original implementation of the routine would return 'false' for
       this scenario.  The logic has been changed so that now the input test
       mask is bitwise ANDed with the pixel mask in each span.  If the 
       bitwise-AND is nonzero, the detector is flagged with the artifact(s)
       and the routine returns true.  Otherwise, the routine returns false.
       In this scenario, since we're testing for impulse noise, and
       impulse noise was one of the artifacts that was added, the routine
       should return true   */
    mask = IAS_RPS_IMPNOISE | IAS_RPS_ANALOG_SAT_LO;
    ias_pm_add_pixels((void *)pm_test, 39, 20, 40, mask);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("Cannot simulate combined artifacts for given span\n");
        ias_pm_destroy(pm_test);
        return;
    }
    mask = IAS_RPS_IMPNOISE;
    is_in_mask = ias_pm_is_detector_flagged(pm_test, 39, mask);
    assert_true(is_in_mask == TRUE);
    if (!is_in_mask)
    {
        printf("Cannot find specified artifact(s) within multiple artifact "
               "spans\n");
        ias_pm_destroy(pm_test);
        return;
    }

    /* Fifth test:  simulate detector 45 as inoperable */
    mask = IAS_RPS_INOP_DETECTOR;
    status = ias_pm_add_pixels((void *)pm_test, 44, 0, 100, mask);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("Cannot simulate inoperable detector 45\n");
        ias_pm_destroy(pm_test);
        return;
    }
    is_in_mask = ias_pm_is_detector_flagged(pm_test, 44, mask);
    assert_true(is_in_mask == TRUE);
    if (!is_in_mask)
    {
        printf("Detector 45 should be flagged as inoperable\n");
        ias_pm_destroy(pm_test);
        return;
    }

    /* Sixth test:  Assume detector 46 is out-of-spec.  Currently,
       out-of-spec detectors are not flagged in the pixel mask (to do so
       would require extending the mask to 16 bits).  For some purposes
       (for example, see comments in the 60s radiometric stability
       characterization modules char_bias_stability.c and/or
       char_gain_stabiity.c), out-of-spec detectors can be treated as though
       they were truly inoperable.

       So... flag this detector as "inoperable" to "fake" the flag as
       out-of-spec. BUT...  just flag the first pixel in the mask.  This
       will test the logic implemented in char_bias_stability and
       char_gain_stabilty to handle out-of-spec detectors (it's not easily
       determinable how many lines there are in the pixel mask, so just the
       first pixel in the mask was flagged).  If all works as
       expected, ias_pm_is_detector_flagged should indicate that it is... */
    mask = IAS_RPS_INOP_DETECTOR;
    status = ias_pm_add_pixels((void *)pm_test, 45, 0, 1, mask);
    assert_true_with_message((status == SUCCESS),
        "Cannot simulate inoperable detector 46\n");
    if (status != SUCCESS)
    {
        ias_pm_destroy(pm_test);
        return;
    }

    is_in_mask = ias_pm_is_detector_flagged(pm_test, 45, mask);
    assert_true_with_message((is_in_mask == TRUE),
         "Detector 46 should be flagged as inoperable\n");


    /* Seventh test:  Check detector 45 for dropped frames.  Since it
       only was flagged as inoperable, the routine should return
       false.  In addition, there should be a warning that the detector
       was flagged with an artifact(s) that aren't specified in the
       input mask */
    mask = IAS_RPS_DROPLINE;
    is_in_mask = ias_pm_is_detector_flagged(pm_test, 44, mask);
    assert_true(is_in_mask == FALSE);
    if (is_in_mask)
    {
        printf("Detector 45 should not be flagged for dropped frames, "
               "it's only inoperable\n");
        ias_pm_destroy(pm_test);
        return;
    }

    /* Final test:  add a block to detector 67 with impulse noise and
       saturation, 3 pixels in length, starting at line 23, then just
       check to see if anything is flagged.  As with test #4, the
       routine should return 'true'.  This simulates a "blind"
       check of the pixel mask by an application, where we don't really
       care WHAT was flagged, just that SOMETHING was flagged*/
    mask = IAS_RPS_IMPNOISE | IAS_RPS_ADCONV_SAT_HI;
    status = ias_pm_add_pixels((void *)pm_test, 66, 22, 3, mask);
    assert_true(status == SUCCESS);
    if (status != SUCCESS)
    {
        printf("Cannot add artifacts to detector 67\n");
        ias_pm_destroy(pm_test);
        return;
    }
    mask = ~PM_NOVALUE;
    is_in_mask = ias_pm_is_detector_flagged(pm_test, 66, mask);
    assert_true(is_in_mask == TRUE);
    if (!is_in_mask)
        printf("Global find should return true for detector 67\n");

    /* Clean up */
    ias_pm_destroy(pm_test);
}




int main(int argc, char *argv[])
{
    TestSuite *suite = create_test_suite();
    add_test(suite, test_pm_is_detector_flagged);

    ias_log_initialize("pm_is_detector_flagged_test");

    if (argc > 1)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}
