/***************************************************************************
 NAME:          pm_single_mask_test

 PURPOSE:       Performs CGreen unit testing of new functionality in
                the pixel mask library allowing reads/writes of a single
                pixel mask from/to a pixel mask file.
****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "pm_local.h"
#include "ias_const.h"
#include "ias_pixel_mask.h"
#include "ias_logging.h"
#include "cgreen.h"


Ensure pm_single_mask_test()
{
    /* name of pixel mask file */
    const char *pixel_mask_file_name = "testmasks.dat";

    IAS_PIXEL_MASK *pm1_read = NULL;     /* pixel mask #1 read from file */
    IAS_PIXEL_MASK *pm1 = NULL;          /* pixel mask #1 written to file */
    IAS_PIXEL_MASK *pm2_read = NULL;     /* pixel mask #2 read from file */
    IAS_PIXEL_MASK *pm2 = NULL;          /* pixel mask #2 written to file */
    IAS_PIXEL_MASK_IO *pm_file = NULL;   /* pixel mask file */
    IAS_PIXEL_MASK_FILE_HOUSEKEEPING exp_hk[2];
                                         /* expected housekeeping data
                                            associated with pixel masks
                                            #1 and #2 */
    IAS_PIXEL_MASK_ITERATOR *it1 = NULL; /* iterator used to get spans in
                                            pixel mask #1 */
    IAS_PIXEL_MASK_ITERATOR *it2 = NULL; /* iterator used to get spans in
                                            pixel mask #2 */
    IAS_PIXEL_MASK_ITERATOR *it_read = NULL;
                                         /* iterator used to get spans in
                                            pixel mask read from file */
    IAS_PIXEL_MASK_SPAN span1;           /* span information for pixel mask
                                            #1 */
    IAS_PIXEL_MASK_SPAN span2;           /* span information for pixel mask
                                            #2 */
    IAS_PIXEL_MASK_SPAN span_read;       /* span information for pixel mask
                                            read from file */

    int mask1_band_number = 1;             /* mask #1 band number */
    int mask1_sca_number = 2;              /* mask #1 SCA number */
    int mask1_number_of_detectors = 10;    /* mask #1 number of detectors */
    int mask1_number_of_pixels = 600;      /* mask #1 number of pixels */
    int mask2_band_number = 7;             /* mask #2 band number */
    int mask2_sca_number = 10;             /* mask #2 SCA number */
    int mask2_number_of_detectors = 20;    /* mask #2 number of detectors */
    int mask2_number_of_pixels = 1200;     /* mask #2 number of lines */
    int pm1_span_count = 0;                /* number of spans added to
                                              pixel mask #1 */
    int pm2_span_count = 0;                /* number of spans added to
                                              pixel mask #2 */
    int read_span_count = 0;               /* number of spans in version of
                                              pixel mask read from the
                                              file */
    int status;


    printf("Testing read/write of single pixel masks to file...\n");

    /* Set the logging level to suppress DEBUG messages, but print
       ERROR messages in case there's an issue needing to be fixed. */
    ias_log_set_output_level(IAS_LOG_LEVEL_INFO);
 
    /* Create the base pixel masks. */
    pm1 = ias_pm_create(mask1_band_number, mask1_sca_number,
        mask1_number_of_detectors, mask1_number_of_pixels);
    pm2 = ias_pm_create(mask2_band_number, mask2_sca_number,
        mask2_number_of_detectors, mask2_number_of_pixels);
    if ((pm1 == NULL) || (pm2 == NULL))
    {
        assert_true_with_message(0, "Error creating test pixel masks");
        if (pm1 != NULL)
        {
            ias_pm_destroy(pm1);
        }
        if (pm2 != NULL)
        {
            ias_pm_destroy(pm2);
        }
        return;
    }

    /* Add spans for detector 0 to pixel mask #1. */
    status = ias_pm_add_pixels(pm1, 0, 50, 1, 0x1);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error filling out pixel mask #1");
        ias_pm_destroy(pm1);
        ias_pm_destroy(pm2);
        return;
    }

    status = ias_pm_add_pixels(pm1, 0, 49, 1, 0x1);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error filling out pixel mask #1");
        ias_pm_destroy(pm1);
        ias_pm_destroy(pm2);
        return;
    }

    /* Add spans for detectors 0, 5, and 9 to pixel mask #2. */
    status = ias_pm_add_pixels(pm2, 0, 500, 100, 0x2);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error filling out pixel mask #2");
        ias_pm_destroy(pm1);
        ias_pm_destroy(pm2);
        return;
    }

    status = ias_pm_add_pixels(pm2, 5, 500, 91, 0x3);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error filling out pixel mask #2");
        ias_pm_destroy(pm1);
        ias_pm_destroy(pm2);
        return;
    }

    status = ias_pm_add_pixels(pm2, 9, 500, 99, 0x4);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "Error filling out pixel mask #2");
        ias_pm_destroy(pm1);
        ias_pm_destroy(pm2);
        return;
    }

    /* Create a pixel mask file, and write pixel mask #1 to it.  After
       the PIXEL_MASK_IO data structure is created, check its
       initialization.  */
    pm_file = ias_pm_open_pixel_mask(pixel_mask_file_name, IAS_WRITE);
    assert_not_equal_with_message(pm_file, NULL, "Error creating pixel "
        "mask IO structure");
    if (pm_file == NULL)
    {
        ias_pm_destroy(pm1);
        ias_pm_destroy(pm2);
        return;
    }
    assert_string_equal_with_message(pm_file->pixel_mask_file_name,
        pixel_mask_file_name, "Error copying pixel mask file name to "
        "data structure");
    assert_equal_with_message(pm_file->number_of_masks_present, 0,
        "Pixel mask IO data structure not properly initialized on "
        "open");
    assert_not_equal_with_message(pm_file->fptr, NULL, "Error opening "
        "pixel mask file");
    if ((pm_file->fptr == NULL) || (pm_file->number_of_masks_present != 0)
        || (strcmp(pm_file->pixel_mask_file_name, pixel_mask_file_name)) != 0)
    {
        ias_pm_destroy(pm1);
        ias_pm_destroy(pm2);
        ias_pm_close_pixel_mask(pm_file);
        return;
    }
    status = ias_pm_write_single_mask_to_file(pm_file, pm1);
    assert_equal_with_message(status, SUCCESS, "Error writing pixel mask #1 "
        "to %s", pixel_mask_file_name);
    if (status != SUCCESS)
    {
        ias_pm_close_pixel_mask(pm_file);
        ias_pm_destroy(pm1);
        ias_pm_destroy(pm2);
        return;
    }
    exp_hk[0] = pm_file->hk[0];

    /* Write pixel mask #2 to the file. */
    status = ias_pm_write_single_mask_to_file(pm_file, pm2);
    assert_equal_with_message(status, SUCCESS, "Error writing pixel mask #2 "
        "to %s", pixel_mask_file_name);
    if (status != SUCCESS)
    {
        ias_pm_close_pixel_mask(pm_file);
        ias_pm_destroy(pm2);
        return;
    }
    assert_equal_with_message(pm_file->number_of_masks_present, 2,
        "Number of masks present is %d, should be 2",
        pm_file->number_of_masks_present);
    if (pm_file->number_of_masks_present != 2)
    {
        ias_pm_close_pixel_mask(pm_file);
        ias_pm_destroy(pm1);
        ias_pm_destroy(pm2);
        return;
    }
    exp_hk[1] = pm_file->hk[1];

    /* Close the pixel mask file.  Check the status of the close. */
    status = ias_pm_close_pixel_mask(pm_file);
    pm_file = NULL;
    assert_equal_with_message(status, SUCCESS, "Error closing pixel mask "
        "file %s in write mode", pixel_mask_file_name);
    if (status != SUCCESS)
    {
        return;
    }

    /* Open the pixel mask file in read-only mode */
    pm_file = ias_pm_open_pixel_mask(pixel_mask_file_name, IAS_READ);
    if (pm_file == NULL)
    {
        assert_true_with_message(0, "Error opening %s for reading",
            pixel_mask_file_name);
        return;
    }

    /* verify the correct band/SCAs are found present */
    status = ias_pm_is_band_sca_present(pm_file, pm1->band, pm1->sca);
    assert_true(status);
    status = ias_pm_is_band_sca_present(pm_file, pm2->band, pm2->sca);
    assert_true(status);

    /* verify band/SCAs not in the file are not found */
    status = ias_pm_is_band_sca_present(pm_file, pm2->band, pm2->sca-1);
    assert_false(status);
    status = ias_pm_is_band_sca_present(pm_file, pm2->band, pm2->sca+1);
    assert_false(status);
    status = ias_pm_is_band_sca_present(pm_file, pm2->band-1, pm2->sca);
    assert_false(status);
    status = ias_pm_is_band_sca_present(pm_file, pm2->band+1, pm2->sca);
    assert_false(status);

    /* Attempt to read both masks */
    pm1_read = ias_pm_read_single_mask_from_file(pm_file, pm1->band,
        pm1->sca);
    assert_not_equal_with_message(pm1_read, NULL, "Error reading first mask "
        "from %s", pixel_mask_file_name);
    if (pm1_read == NULL)
    {
        ias_pm_close_pixel_mask(pm_file);
        ias_pm_destroy(pm1);
        ias_pm_destroy(pm2);
        return;
    }
    pm2_read = ias_pm_read_single_mask_from_file(pm_file, pm2->band,
        pm2->sca);
    assert_not_equal_with_message(pm2_read, NULL, "Error reading second mask "
        "from %s", pixel_mask_file_name);
    if (pm2_read == NULL)
    {
        ias_pm_close_pixel_mask(pm_file);
        ias_pm_destroy(pm1);
        ias_pm_destroy(pm1_read);
        ias_pm_destroy(pm2);
        return;
    }

    /* Verify pixel mask #1 was read correctly from the file. */
    assert_equal(pm1_read->band, pm1->band);
    assert_equal(pm1_read->sca, pm1->sca);
    assert_equal(pm1_read->num_of_detectors, pm1->num_of_detectors);
    assert_equal(pm1_read->num_of_pixels, pm1->num_of_pixels);

    it1 = ias_pm_get_iterator(pm1, IAS_PM_EXCLUDE, 0x0);
    it_read = ias_pm_get_iterator(pm1_read, IAS_PM_EXCLUDE, 0x0);
    if ((it1 == NULL) || (it_read == NULL))
    {
        assert_true_with_message(0, "Error obtaining iterators for pixel "
            "mask #1 comparison");
        ias_pm_close_pixel_mask(pm_file);
        if (it1 != NULL)
        {
            ias_pm_destroy_iterator(it1);
        }
        if (it_read != NULL)
        {
            ias_pm_destroy_iterator(it_read);
        }
        ias_pm_destroy(pm1);
        ias_pm_destroy(pm1_read);
        ias_pm_destroy(pm2);
        ias_pm_destroy(pm2_read);
        return;
    }

    while (ias_pm_get_next_span(it1, &span1))
    {
        pm1_span_count++;
        if (ias_pm_get_next_span(it_read, &span_read))
        {
            assert_equal(span_read.detector_index, span1.detector_index);
            assert_equal(span_read.starting_pixel_index,
                span1.starting_pixel_index);
            assert_equal(span_read.length_of_span, span1.length_of_span);
            assert_equal((int)span_read.pixel_mask, (int)span1.pixel_mask);
            read_span_count++;
        }
    }
    assert_equal(read_span_count, pm1_span_count);

    /* Destroy the iterators for pixel mask #1. */
    ias_pm_destroy_iterator(it1);
    ias_pm_destroy_iterator(it_read);
    ias_pm_destroy(pm1);
    ias_pm_destroy(pm1_read);
    it1 = NULL;
    it_read = NULL;
    pm1 = NULL;
    pm1_read = NULL;

    /* Verify pixel mask #2 was read correctly from the file. */
    assert_equal(pm2_read->band, pm2->band);
    assert_equal(pm2_read->sca, pm2->sca);
    assert_equal(pm2_read->num_of_detectors, pm2->num_of_detectors);
    assert_equal(pm2_read->num_of_pixels, pm2->num_of_pixels);

    it2 = ias_pm_get_iterator(pm2, IAS_PM_EXCLUDE, 0x0);
    it_read = ias_pm_get_iterator(pm2_read, IAS_PM_EXCLUDE, 0x0);
    if ((it2 == NULL) || (it_read == NULL))
    {
        assert_true_with_message(0, "Error obtaining iterators for pixel "
            "mask #2 comparison");
        ias_pm_close_pixel_mask(pm_file);
        if (it2 != NULL)
        {
            ias_pm_destroy_iterator(it2);
        }
        if (it_read != NULL)
        {
            ias_pm_destroy_iterator(it_read);
        }
        ias_pm_destroy(pm2);
        ias_pm_destroy(pm2_read);
        return;
    }

    read_span_count = 0;
    while (ias_pm_get_next_span(it2, &span2))
    {
        pm2_span_count++;
        if (ias_pm_get_next_span(it_read, &span_read))
        {
            assert_equal(span_read.detector_index, span2.detector_index);
            assert_equal(span_read.starting_pixel_index,
                span2.starting_pixel_index);
            assert_equal(span_read.length_of_span, span2.length_of_span);
            assert_equal((int)span_read.pixel_mask, (int)span2.pixel_mask);
            read_span_count++;
        }
    }
    assert_equal(read_span_count, pm2_span_count);

    /* Destroy the mask structures and iterators for pixel mask #2. */
    ias_pm_destroy_iterator(it2);
    ias_pm_destroy_iterator(it_read);
    ias_pm_destroy(pm2);
    ias_pm_destroy(pm2_read);
    it2 = NULL;
    it_read = NULL;
    pm2 = NULL;
    pm2_read = NULL;

    /* Compare the housekeeping data.  The pixel mask file needs to be kept
       open so that the housekeeping data block is valid.  */
    assert_equal(pm_file->number_of_masks_present, 2);

    assert_equal(pm_file->hk[0].band_number, exp_hk[0].band_number);
    assert_equal(pm_file->hk[0].sca_number, exp_hk[0].sca_number);
    assert_equal(pm_file->hk[0].mask_index, exp_hk[0].mask_index);
    assert_equal(pm_file->hk[0].starting_data_offset,
        exp_hk[0].starting_data_offset);
    assert_equal(pm_file->hk[0].mask_data_size, exp_hk[0].mask_data_size);

    assert_equal(pm_file->hk[1].band_number, exp_hk[1].band_number);
    assert_equal(pm_file->hk[1].sca_number, exp_hk[1].sca_number);
    assert_equal(pm_file->hk[1].mask_index, exp_hk[1].mask_index);
    assert_equal(pm_file->hk[1].starting_data_offset,
        exp_hk[1].starting_data_offset);
    assert_equal(pm_file->hk[1].mask_data_size, exp_hk[1].mask_data_size);

    /* Close the pixel mask file.  Check the status of the close. */
    status = ias_pm_close_pixel_mask(pm_file);
    assert_equal_with_message(status, SUCCESS, "Error closing pixel mask "
        "file %s in read mode");

}   /* END -- pm_single_mask_test */



int main(int argc, char *argv[])
{
    TestSuite *suite = create_test_suite();
    add_test(suite, pm_single_mask_test);

    if (argc > 1)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}
