#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "../pm_local.h"
#include "ias_const.h"
#include "ias_pixel_mask.h"
#include "ias_linked_list.h"
#include "ias_logging.h"
#include "cgreen/cgreen.h"
#include "ias_radiometry.h"

#define NUM_DETS 10
#define NUM_PIXS 600

Ensure pm_test( void );
Ensure subset_test( void );
Ensure add_pixel_test( void );
void test_span( IAS_PIXEL_MASK_SPAN, int, int, int, PIXEL_MASK_TYPE );

int main( int argc, char *argv[] )
{
    TestSuite *suite = create_test_suite();
    add_test( suite, pm_test );
    add_test( suite, subset_test );
    add_test( suite, add_pixel_test );

    if (argc > 1) {
        return run_single_test( suite, argv[1], create_text_reporter());
    }
    return run_test_suite( suite, create_text_reporter() );
}

Ensure pm_test( void )
{
    IAS_PIXEL_MASK *pm;
    IAS_PIXEL_MASK *pm2;
    IAS_PIXEL_MASK *pm_array[2];
    IAS_PIXEL_MASK **pm_arrayptr;
    struct IAS_PIXEL_MASK *pmi;
    PIXEL_MASK_TYPE mask;
    PIXEL_MASK_TYPE *pm_image;
    PIXEL_MASK_TYPE *cmp_image;
    PIXEL_MASK_TYPE *tr_image;
    PIXEL_MASK_TYPE *image_ptr;
    PIXEL_MASK_TYPE *pmi_ptr;
    IAS_PIXEL_MASK_ITERATOR *it;
    IAS_PIXEL_MASK_SPAN span;
    IAS_LINKED_LIST_NODE *base;
    int image_size;
    int detector_offset;
    int count;
    int i;
    int status;
    int deti;
    int det_save;
    int det_index;
    int pix_index;
    int pix[] = {0,50,100,105,190,200,208,218,220,232,350,500,530,
                 0,500,530,
                 50,100,105,190,200,208,218,220,232,
                 200,208,218,220,500,500,
                 200,208,218,220,500,500,
                 696,701,425,430,187,491,500,501,509,510};
    int len[] = {50,50,5,85,10,8,10,2,12,118,150,30,70,
                 500,30,70,
                 50,5,85,10,8,10,2,12,118,
                 8,10,2,12,30,30,
                 8,10,2,12,30,30,
                 5,5,5,5,15,9,1,8,1,9};
    PIXEL_MASK_TYPE m[] =
        {0x0,0x10,0x12,0x10,0x14,0x15,0x11,0x19,0x18,0x10,0x0,0x1,0x0,
         0x0,0x1,0x0,
         16,18,16,20,21,17,25,24,16,
         21,17,25,24,1,1,
         0x15,0x11,0x19,0x18,0x1,0x1,
         0x2,0x3,0x3,0x2,0x3,0x2,0x3,0x1,0x5,0x4};

    /* Initialize the logging library in case any library call issues a
    log message */
    ias_log_initialize( "pm_test" );

    pm = NULL;
    ias_log_set_output_level( IAS_LOG_LEVEL_DISABLE );
    /* Check band accessor function - Error */
    assert_equal( ias_pm_get_band_number(pm), ERROR );
    /* Check sca accessor function - Error */
    assert_equal( ias_pm_get_sca_number(pm), ERROR );
    /* Check number of detectors accessor function - Error */
    assert_equal( ias_pm_get_number_of_detectors(pm), ERROR );
    /* Check number of pixels accessor function - Error */
    assert_equal( ias_pm_get_number_of_pixels(pm), ERROR );
    ias_log_set_output_level( IAS_LOG_LEVEL_INFO );

    /* Create the pixel mask */
    pm = ias_pm_create( 1, 2, NUM_DETS, NUM_PIXS );
    assert_not_equal( pm, NULL );
    /* Check band accessor function - Normal */
    assert_equal( ias_pm_get_band_number(pm), 1 );
    /* Check sca accessor function - Normal */
    assert_equal( ias_pm_get_sca_number(pm), 2 );
    /* Check number of detectors accessor function - Normal */
    assert_equal( ias_pm_get_number_of_detectors(pm), NUM_DETS );
    assert_equal( ias_pm_get_number_of_pixels(pm), NUM_PIXS );
    
    pmi = pm;
    pm_array[0] = pm;

    /* Create an image to compare with.  This will be filled from the
    iterator test that returns all spans. */
    image_size = NUM_DETS * NUM_PIXS;
    cmp_image = malloc( sizeof( PIXEL_MASK_TYPE ) * image_size );
    if ( !cmp_image )
    {
        assert_true_with_message(0, "Error allocating cmp_image");
        return;
    }
    memset( cmp_image, 0, image_size );
    tr_image = malloc( sizeof( PIXEL_MASK_TYPE ) * image_size );
    if ( !tr_image )
    {
        assert_true_with_message(0, "Error allocating tr_image");
        free( cmp_image );
        return;
    }

    /* Make sure the base nodes in the detector look up table are all NULL */
    for ( i = 0; i < NUM_DETS; i++ )
    {
        base = pmi->detector_lut[i];
        assert_equal( base, NULL );
    }

    /* Try to add a span with a negative length (should error) */
    ias_log_set_output_level( IAS_LOG_LEVEL_DISABLE );
    status = ias_pm_add_pixels( pm, 4, 200, -4, 0x01 );
    ias_log_set_output_level( IAS_LOG_LEVEL_INFO );
    assert_equal( status, ERROR );

    /* Add the 200 span to the mask */
    assert_false( ias_pm_add_pixels( pm, 1, 200, 20, 0x1 ));

    /* Now the base node for detector 1 should not be NULL */
    base = pmi->detector_lut[1];
    assert_not_equal( base, NULL );
    count = ias_linked_list_count_nodes( base );
    assert_equal( count, 1 );

    /* Get an iterator for this pixel mask, use no value so that it
    returns all masked pixels */
    it = ias_pm_get_iterator( pm, IAS_PM_EXCLUDE, 0x0 );
    assert_not_equal( it, NULL );

    /* This adds a span completely before the 200 span */
    assert_false( ias_pm_add_pixels( pm, 1, 100,  5, 0x2 ));
    count = ias_linked_list_count_nodes( base );
    assert_equal( count, 2 );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 1, 100, 5, 0x2 );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 1, 200, 20, 0x1 );
    assert_false( ias_pm_get_next_span( it, &span ));
    ias_pm_destroy_iterator( it );

    /* Adding the next span should split the 200 span.
    The span overlaps the start of the 200 span */
    assert_false( ias_pm_add_pixels( pm, 1, 190, 18, 0x4 ));
    it = ias_pm_get_iterator( pm, IAS_PM_EXCLUDE, 0x0 );
    assert_not_equal( it, NULL );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 1, 100, 5, 0x2 );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 1, 190, 10, 0x4 );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 1, 200, 8, 0x5 );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 1, 208, 12, 0x1 );
    assert_false( ias_pm_get_next_span( it, &span ));
    ias_pm_destroy_iterator( it );

    /* This span will split the 208 span.  It overlaps the end of
    the span. */
    assert_false( ias_pm_add_pixels( pm, 1, 218, 14, 0x8 ));
    it = ias_pm_get_iterator( pm, IAS_PM_EXCLUDE, 0x0 );
    assert_not_equal( it, NULL );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 1, 100, 5, 0x2 );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 1, 190, 10, 0x4 );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 1, 200, 8, 0x5 );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 1, 208, 10, 0x1 );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 1, 218, 2, 0x9 );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 1, 220, 12, 0x8 );
    assert_false( ias_pm_get_next_span( it, &span ));
    ias_pm_destroy_iterator( it );

    /* Adding this span will overlap all existing spans */
    assert_false( ias_pm_add_pixels( pm, 1, 50, 300, 0x10 ));
    it = ias_pm_get_iterator( pm, IAS_PM_EXCLUDE, 0x0 );
    assert_not_equal( it, NULL );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 1, 50, 50, 0x10 );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 1, 100, 5, 0x12 );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 1, 105, 85, 0x10 );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 1, 190, 10, 0x14 );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 1, 200, 8, 0x15 );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 1, 208, 10, 0x11 );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 1, 218, 2, 0x19 );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 1, 220, 12, 0x18 );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 1, 232, 118, 0x10 );
    assert_false( ias_pm_get_next_span( it, &span ));
    ias_pm_destroy_iterator( it );

    /* Finally add a span after all existing spans */
    assert_false( ias_pm_add_pixels( pm, 1, 500, 30, 0x1 ));
    it = ias_pm_get_iterator( pm, IAS_PM_EXCLUDE, 0x0 );
    assert_not_equal( it, NULL );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 1, 50, 50, 0x10 );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 1, 100, 5, 0x12 );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 1, 105, 85, 0x10 );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 1, 190, 10, 0x14 );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 1, 200, 8, 0x15 );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 1, 208, 10, 0x11 );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 1, 218, 2, 0x19 );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 1, 220, 12, 0x18 );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 1, 232, 118, 0x10 );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 1, 500, 30, 0x1 );
    assert_false( ias_pm_get_next_span( it, &span ));
    ias_pm_destroy_iterator( it );

    /* Add a dropped line span for detector 5 */
    assert_false( ias_pm_add_pixels( pm, 5, 500, 30, 0x1 ));

    /* Get masks at specific entries */
    /* There should be no mask at detector 3 */
    assert_false( ias_pm_get_mask_at( pm, 3, 200, &mask ));
    assert_equal( mask, 0 );
    /* There should be no mask at detector 1 pixel 20 */
    assert_false( ias_pm_get_mask_at( pm, 1, 20, &mask ));
    assert_equal( mask, 0 );
    /* All others should return a mask */
    assert_false( ias_pm_get_mask_at( pm, 1, 199, &mask ));
    assert_equal( mask, 0x14 );
    assert_false( ias_pm_get_mask_at( pm, 1, 200, &mask ));
    assert_equal( mask, 0x15 );
    assert_false( ias_pm_get_mask_at( pm, 1, 201, &mask ));
    assert_equal( mask, 0x15 );
    assert_false( ias_pm_get_mask_at( pm, 5, 520, &mask ));
    assert_equal( mask, 0x1 );

    /* This test returns all spans for the image, masked or not */
    it = ias_pm_get_iterator( pm, IAS_PM_ALL, PM_NOVALUE );
    assert_not_equal( it, NULL );
    count = deti = 0;
    while ( ias_pm_get_next_span( it, &span ))
    {
        count++;
        detector_offset = span.detector_index * NUM_PIXS;
        switch ( span.detector_index )
        {
            case 1 :
                test_span( span, 1, pix[deti], len[deti], m[deti] );
                memset( cmp_image + detector_offset +  pix[deti],
                    m[deti], len[deti] );
                deti++;
                break;
            case 5 :
                test_span( span, 5, pix[deti], len[deti], m[deti] );
                memset( cmp_image + detector_offset +  pix[deti],
                    m[deti], len[deti] );
                deti++;
                break;
            case 0 :
            case 2 :
            case 3 :
            case 4 :
            case 6 :
            case 7 :
            case 8 :
            case 9 :
                test_span( span, span.detector_index, 0, NUM_PIXS, 0x0 );
                break;
            default:
                assert_true_with_message( 0,
                    "Should not have gotten here with detector number %d",
                    span.detector_index );
        }  /* END switch */
    }  /* END while */
    ias_pm_destroy_iterator( it );

    /* This test returns all unmasked spans except those that are only
    masked with 0x1: det1,500,30 and det5,500,30 */
    it = ias_pm_get_iterator( pm, IAS_PM_EXCLUDE, 0x1 );
    assert_not_equal( it, NULL );
    while ( ias_pm_get_next_span( it, &span ))
    {
        switch ( span.detector_index )
        {
            case 1 :
                test_span( span, 1, pix[deti], len[deti], m[deti] );
                deti++;
                break;
            default:
                assert_true_with_message( 0,
                    "Should not have gotten here with detector number %d",
                    span.detector_index );
        }  /* END switch */
    }  /* END while */
    ias_pm_destroy_iterator( it );

    /* This test returns all spans masked with 0x1 and 0x8 */
    det_save = deti;
    it = ias_pm_get_iterator( pm, IAS_PM_INCLUDE, 0x1|0x8 );
    assert_not_equal( it, NULL );
    while ( ias_pm_get_next_span( it, &span ))
    {
        test_span( span, span.detector_index, pix[deti], len[deti], m[deti] );
        deti++;
    }  /* END while */
    ias_pm_destroy_iterator( it );

    /* The next test is to get the entire image for this pixel mask */
    pm_image = malloc( sizeof( PIXEL_MASK_TYPE ) * image_size );
    assert_not_equal( pm_image, NULL );
    assert_false( ias_pm_get_image( pm, pm_image ));
    /* The image returned from ias_pm_get_image is transposed so transpose
       it before the comparison */
    pmi_ptr = pm_image;
    for ( pix_index = 0; pix_index < NUM_PIXS; pix_index++ )
    {
        image_ptr = &tr_image[pix_index];
        for ( det_index = 0; det_index < NUM_DETS; det_index++ )
        {
            image_ptr[det_index * NUM_PIXS] = *pmi_ptr;
            pmi_ptr++;
        }
    }
    assert_false( memcmp( tr_image, cmp_image, image_size ));

    /* Create a second pixel mask */
    pm2 = ias_pm_create( 7, 10, NUM_DETS*2, NUM_PIXS*2 );
    assert_not_equal( pm2, NULL );
    /* Check band accessor function - Normal */
    assert_equal( ias_pm_get_band_number(pm2), 7 );
    /* Check sca accessor function - Normal */
    assert_equal( ias_pm_get_sca_number(pm2), 10 );
    pmi = pm2;
    pm_array[1] = pm2;

    /* Add a test span */
    assert_false( ias_pm_add_pixels( pm2, 18, 500, 10, 0x1 ));
    /* Add a span that begins before the 500 span and overlaps at pixel 500 */
    assert_false( ias_pm_add_pixels( pm2, 18, 491, 10, 0x2 ));
    /* Add a span the begins on the last pixel of the 500 span */
    assert_false( ias_pm_add_pixels( pm2, 18, 509, 10, 0x4 ));
    it = ias_pm_get_iterator( pm2, IAS_PM_EXCLUDE, 0x0 );
    assert_not_equal( it, NULL );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 18, 491, 9, 0x2 );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 18, 500, 1, 0x3 );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 18, 501, 8, 0x1 );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 18, 509, 1, 0x5 );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 18, 510, 9, 0x4 );
    assert_false( ias_pm_get_next_span( it, &span ));
    ias_pm_destroy_iterator( it );

    /* Add another test span */
    assert_false( ias_pm_add_pixels( pm2, 11, 187, 15, 0x01 ));
    /* Add a second span that overlaps exactly the 187 span */
    assert_false( ias_pm_add_pixels( pm2, 11, 187, 15, 0x02 ));
    /* Test the new spans */
    it = ias_pm_get_iterator( pm2, IAS_PM_EXCLUDE, 0x0 );
    assert_not_equal( it, NULL );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 11, 187, 15, 0x3 );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 18, 491, 9, 0x2 );
    ias_pm_destroy_iterator( it );

    /* Add another test span */
    assert_false( ias_pm_add_pixels( pm2, 7, 425, 5, 0x01 ));
    /* Add a span that starts at pixel 425 and extends beyond previous span */
    assert_false( ias_pm_add_pixels( pm2, 7, 425, 10, 0x02 ));
    /* Test the new spans */
    it = ias_pm_get_iterator( pm2, IAS_PM_EXCLUDE, 0x0 );
    assert_not_equal( it, NULL );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 7, 425, 5, 0x3 );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 7, 430, 5, 0x2 );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 11, 187, 15, 0x3 );
    ias_pm_destroy_iterator( it );

    /* Add a final test span */
    assert_false( ias_pm_add_pixels( pm2, 0, 701, 5, 0x01 ));
    /* Add a span that starts before pixel 701 and ends exactly on the end
    of previous span */
    assert_false( ias_pm_add_pixels( pm2, 0, 696, 10, 0x02 ));
    /* Test the new spans */
    it = ias_pm_get_iterator( pm2, IAS_PM_EXCLUDE, 0x0 );
    assert_not_equal( it, NULL );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 0, 696, 5, 0x2 );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 0, 701, 5, 0x3 );
    assert_true( ias_pm_get_next_span( it, &span ));
    test_span( span, 7, 425, 5, 0x3 );
    ias_pm_destroy_iterator( it );

            
    /* Write the pixel mask to a file */
    assert_false( ias_pm_write_array_to_file( pm_array, 2, "pixel_mask" ));

    /* Now read the pixel mask back from the file */
    pm_arrayptr = ias_pm_read_array_from_file( "pixel_mask", &count );
    if ( count != 2 )
    {
        assert_true_with_message( 0,
            "read_array_from_file didn't return 2" );
        return;
    }
    if ( pm_arrayptr[0] == NULL || pm_arrayptr[1] == NULL )
    {
        assert_true_with_message( 0,
            "read_array_from_file returned a NULL pointer" );
        return;
    }

    /* Do the following tests to make sure the same pixel mask was
    read back in */
    /* This test returns all spans masked with 0x1 and 0x8 */
    deti = det_save;
    it = ias_pm_get_iterator( pm_arrayptr[0], IAS_PM_INCLUDE, 0x1|0x8 );
    assert_not_equal( it, NULL );
    while ( ias_pm_get_next_span( it, &span ))
    {
        test_span( span, span.detector_index, pix[deti], len[deti], m[deti] );
        deti++;
    }  /* END while */
    ias_pm_destroy_iterator( it );

    /* Test the other pixel mask */
    deti = 37;
    it = ias_pm_get_iterator( pm_arrayptr[1], IAS_PM_EXCLUDE, 0x0 );
    assert_not_equal( it, NULL );
    while ( ias_pm_get_next_span( it, &span ))
    {
        test_span( span, span.detector_index, pix[deti], len[deti], m[deti] );
        deti++;
    }  /* END while */
    ias_pm_destroy_iterator( it );

    /* Destroy the pixel masks */
    ias_pm_destroy( pm );
    ias_pm_destroy( pm2 );
    ias_pm_destroy( pm_arrayptr[0] );
    ias_pm_destroy( pm_arrayptr[1] );
    free( pm_arrayptr );
    free( cmp_image );
    free( tr_image );
    free( pm_image );
}  /* END pm_test */


void test_span
(
    IAS_PIXEL_MASK_SPAN span,
    int detector_index,
    int starting_pixel_index,
    int length_of_span,
    PIXEL_MASK_TYPE pixel_mask
)
{
    assert_equal( span.detector_index, detector_index );
    assert_equal( span.starting_pixel_index, starting_pixel_index );
    assert_equal( span.length_of_span, length_of_span );
    assert_equal( span.pixel_mask, pixel_mask );
}


Ensure subset_test( void )
{
    IAS_PIXEL_MASK *pm;
    PIXEL_MASK_TYPE *pm_image;
    PIXEL_MASK_TYPE *image_ptr;
    int det;
    int line;
    int status;
    int num_dets;
    int num_lines;
    IAS_PIXEL_MASK_ITERATOR *it;

    /* Initialize the logging library in case any library call issues a
    log message */
    ias_log_initialize( "pm_test" );

    /* Create the pixel mask */
    pm = ias_pm_create( 1, 2, NUM_DETS, NUM_PIXS );
    assert_not_equal( pm, NULL );

    /*----------------------------------------------------------------------
      TEST case 1: This test is to ensure the first detector is included.
                   It also tests the case where the artifact spans are
                   totally included within the subset image.  The artifacts
                   are 20 pixel spans starting at line 100 on the first
                   two detectors.  The subset image starts at line 99 and
                   ends at line 121 so the first and last line of the subset
                   shouldn't have an artifact.  The subset includes detector
                   3 which also has no artifacts.
    -----------------------------------------------------------------------*/
    num_dets = 4;
    num_lines = 22;
    for ( det = 0; det < 3; det++ )
        assert_false( ias_pm_add_pixels( pm, det, 100, 20, 0x1 ));

    pm_image = malloc( num_dets * num_lines * sizeof(PIXEL_MASK_TYPE));
    if ( !pm_image )
    {
        assert_true_with_message(0, "Error allocating subset memory");
        return;
    }

    status = ias_pm_get_image_subset( pm, 0, 99, num_dets, num_lines,
        pm_image );
    if ( status )
    {
        assert_true_with_message(0,
            "Error returned from ias_pm_get_image_subset");
        return;
    }

    for ( det = 0; det < num_dets; det++ )
    {
        image_ptr = pm_image + ( det * num_lines );
        for ( line = 0; line < num_lines; line++ )
        {
            /* All of det 3, and lines 0 and 21 of the others should be 0 */
            if ( det == 3 || line == 0 || line == 21 )
                assert_equal( image_ptr[line], 0x0 );
            else
                assert_equal( image_ptr[line], 0x1 );
        }
    }
    free( pm_image );

    /*----------------------------------------------------------------------
      TEST case 2: This test is to ensure the first line is included as well
                   as the last detector.  Also tested is the case where the
                   artifact spans overlap the last line of the subset.  The
                   artifacts are 20 pixel spans starting on line 0 for
                   detector 6, 5 det 7, and 9 det 8.  Detector 9 has no
                   artifacts.  The subset image starts at line 0 detector x
                   and ends at line 10 detector y.
    -----------------------------------------------------------------------*/
    num_dets = 4;
    num_lines = 10;
    assert_false( ias_pm_add_pixels( pm, 6,  0, 20, 0x2 ));
    assert_false( ias_pm_add_pixels( pm, 7,  5, 20, 0x3 ));
    assert_false( ias_pm_add_pixels( pm, 8,  9, 20, 0x4 ));

    pm_image = malloc( num_dets * num_lines * sizeof(PIXEL_MASK_TYPE));
    if ( !pm_image )
    {
        assert_true_with_message(0, "Error allocating subset memory");
        return;
    }

    status = ias_pm_get_image_subset( pm, 6, 0, num_dets, num_lines, pm_image );
    if ( status )
    {
        assert_true_with_message(0,
            "Error returned from ias_pm_get_image_subset");
        return;
    }

    det = 0; /* Detector 6 is detector 0 in the subset */
    image_ptr = pm_image + ( det * num_lines );
    for ( line = 0; line < num_lines; line++ )
        assert_equal( image_ptr[line], 0x2 );
    det = 1; /* Detector 7 is detector 1 in the subset */
    image_ptr = pm_image + ( det * num_lines );
    for ( line = 0; line < 5; line++ )
        assert_equal( image_ptr[line], 0x0 );
    for ( line = 5; line < num_lines; line++ )
        assert_equal( image_ptr[line], 0x3 );
    det = 2; /* Detector 8 is detector 2 in the subset */
    image_ptr = pm_image + ( det * num_lines );
    for ( line = 0; line < 9; line++ )
        assert_equal( image_ptr[line], 0x0 );
    for ( line = 9; line < num_lines; line++ )
        assert_equal( image_ptr[line], 0x4 );
    det = 3; /* Detector 9 is detector 3 in the subset */
    image_ptr = pm_image + ( det * num_lines );
    for ( line = 0; line < num_lines; line++ )
        assert_equal( image_ptr[line], 0x0 );
    free( pm_image );

    /*----------------------------------------------------------------------
      TEST case 3: This test is to ensure the last line is included as well
                   first and last detector.  Also tested is the case where
                   the artifact spans overlap the first line of the subset.
                   Detector 0 has artifact from 500 to 599
                   Detector 5 has artifact from 500 to 590
                   Detector 9 has artifact from 500 to 598
                   The subset is Detector 0-9 lines 590-599
    -----------------------------------------------------------------------*/
    num_dets = NUM_DETS;
    num_lines = 10;
    assert_false( ias_pm_add_pixels( pm, 0,  500, 100, 0x2 ));
    assert_false( ias_pm_add_pixels( pm, 5,  500,  91, 0x3 ));
    assert_false( ias_pm_add_pixels( pm, 9,  500,  99, 0x4 ));

    pm_image = malloc( num_dets * num_lines * sizeof(PIXEL_MASK_TYPE));
    if ( !pm_image )
    {
        assert_true_with_message(0, "Error allocating subset memory");
        return;
    }

    status = ias_pm_get_image_subset( pm, 0, 590, num_dets, num_lines,
        pm_image );
    if ( status )
    {
        assert_true_with_message(0,
            "Error returned from ias_pm_get_image_subset");
        return;
    }

    det = 0;
    image_ptr = pm_image + ( det * num_lines );
    for ( line = 0; line < num_lines; line++ )
        assert_equal( image_ptr[line], 0x2 );
    det = 5;
    image_ptr = pm_image + ( det * num_lines );
    assert_equal( image_ptr[0], 0x3 );
    for ( line = 1; line < num_lines; line++ )
        assert_equal( image_ptr[line], 0x0 );
    det = 9;
    image_ptr = pm_image + ( det * num_lines );
    for ( line = 0; line < num_lines-1; line++ )
        assert_equal( image_ptr[line], 0x4 );
    assert_equal( image_ptr[line], 0x0 );
    free( pm_image );

    /*----------------------------------------------------------------------
      TEST case 4: This test is to ensure that a subset that is completely
                   contained within a span is handled.
                   Detector 4 has artifact from 100 to 499
                   The subset is Detector 4 lines 200-250
    -----------------------------------------------------------------------*/
    num_dets = 1;
    num_lines = 50;
    assert_false( ias_pm_add_pixels( pm, 4, 50, 400, 0x4 ));

    pm_image = malloc( num_dets * num_lines * sizeof(PIXEL_MASK_TYPE));
    if ( !pm_image )
    {
        assert_true_with_message(0, "Error allocating subset memory");
        return;
    }

    status = ias_pm_get_image_subset( pm, 4, 200, num_dets, num_lines,
        pm_image );
    if ( status )
    {
        assert_true_with_message(0,
            "Error returned from ias_pm_get_image_subset");
        return;
    }

    image_ptr = pm_image;
    for ( line = 0; line < num_lines; line++ )
        assert_true_with_message( image_ptr[line] == 0x4, "line %d value %d",
                line, image_ptr[line] );

    /* ------------------------------------------
       Following test case added for new routine
       ias_pm_get_iterator_at_detector
       -----------------------------------------*/
    /* Test the newly created routine for the error case */
    ias_log_set_output_level( IAS_LOG_LEVEL_DISABLE );
    it = ias_pm_get_iterator_at_detector( pm, 0, 0, NUM_DETS );
    ias_log_set_output_level( IAS_LOG_LEVEL_INFO );
    assert_equal( it, NULL );

    /* Test that asking for detector 4 retrieves a span for detector 4 */
    it = ias_pm_get_iterator_at_detector( pm, 0, 0, 4 );
    if ( !it )
    {
        assert_true_with_message(0,
            "Error getting iterator for detector 4");
    }
    else
    {
        IAS_PIXEL_MASK_SPAN span;
        if ( ias_pm_get_next_span( it, &span ))
        {
            assert_equal( span.detector_index, 4 );
        }
        else
        {
            assert_true_with_message(0,
                "Error getting span for detector 4");
        }
        ias_pm_destroy_iterator( it );
    }

    /* Free the resources used */
    ias_pm_destroy( pm );
    free( pm_image );
}  /* END subset_test */

Ensure add_pixel_test( void )
{
    IAS_PIXEL_MASK *pm;
    IAS_PIXEL_MASK_ITERATOR *it;
    IAS_PIXEL_MASK_SPAN span;
    int span_count;

    /* Initialize the logging library */
    ias_log_initialize( "add_pixel_test" );

    /* Create the pixel mask */
    pm = ias_pm_create( 1, 2, NUM_DETS, NUM_PIXS );
    assert_not_equal( pm, NULL );

    /* Test 1: Verify a span can grow downward (toward pixel zero) */
   
    /* Setup the mask. */
    assert_false( ias_pm_add_pixels( pm, 0, 50, 1, 0x1 )); 
    assert_false( ias_pm_add_pixels( pm, 0, 49, 1, 0x1 )); 
    assert_false( ias_pm_add_pixels( pm, 0, 40, 9, 0x1 )); 
    assert_false( ias_pm_add_pixels( pm, 0, 38, 1, 0x1 )); 
    assert_false( ias_pm_add_pixels( pm, 0, 39, 1, 0x1 )); 

    it = ias_pm_get_iterator( pm, IAS_PM_INCLUDE, 0x1 );
    assert_not_equal( it, NULL );

    span_count = 0;
    while ( ias_pm_get_next_span( it, &span ))
    {
        span_count++;
    }
    if ( span_count == 1 )
    {
        assert_equal( span.starting_pixel_index, 38 );
        assert_equal( span.length_of_span, 13 );
        assert_equal( span.pixel_mask, 0x1 );
    }
    else
    {
        assert_true_with_message( 0, "Prepending to existing span failed" );
    }
    ias_pm_destroy_iterator( it );
    /* End Test 1 */

    /* Test 2: Verify a span a grow upward (toward the maximum pixel value) */

    assert_false( ias_pm_add_pixels( pm, 0, 51, 10, 0x1 ));
    assert_false( ias_pm_add_pixels( pm, 0, 62, 1, 0x1 ));
    assert_false( ias_pm_add_pixels( pm, 0, 61, 1, 0x1 ));

    it = ias_pm_get_iterator( pm, IAS_PM_INCLUDE, 0x1 );
    assert_not_equal( it, NULL );

    span_count = 0;
    while ( ias_pm_get_next_span( it, &span ))
    {
        span_count++;
    }
    if ( span_count == 1 )
    {
        assert_equal( span.starting_pixel_index, 38 );
        assert_equal( span.length_of_span, 25 );
        assert_equal( span.pixel_mask, 0x1 );
    }
    else
    {
        assert_true_with_message( 0, "Appending to existing span failed" );
    }
    ias_pm_destroy_iterator( it );
    /* End Test 2 */

    /* Test 3: Verify intersecting spans */

    assert_false( ias_pm_add_pixels( pm, 0, 55, 2, 0x1 ));
    assert_false( ias_pm_add_pixels( pm, 0, 30, 10, 0x1 ));

    it = ias_pm_get_iterator( pm, IAS_PM_INCLUDE, 0x1 );
    assert_not_equal( it, NULL );

    span_count = 0;
    while ( ias_pm_get_next_span( it, &span ))
    {
        span_count++;
    }
    if ( span_count == 1 )
    {
        assert_equal( span.starting_pixel_index, 30 );
        assert_equal( span.length_of_span, 33 );
        assert_equal( span.pixel_mask, 0x1 );
    }
    else
    {
        assert_true_with_message( 0, "Inserting intersecting span into "
            "existing span failed" );
    }
    ias_pm_destroy_iterator( it );
    /* End Test 3 */
    
    /* Test 4 - Updates modifying existing pixel mask */
    assert_false( ias_pm_add_pixels( pm, 0, 50, 1, 0x2 )); 
    assert_false( ias_pm_add_pixels( pm, 0, 49, 1, 0x2 )); 
    assert_false( ias_pm_add_pixels( pm, 0, 40, 9, 0x2 )); 
    assert_false( ias_pm_add_pixels( pm, 0, 38, 1, 0x2 )); 
    assert_false( ias_pm_add_pixels( pm, 0, 39, 1, 0x2 )); 
    assert_false( ias_pm_add_pixels( pm, 0, 51, 10, 0x2 ));
    assert_false( ias_pm_add_pixels( pm, 0, 62, 1, 0x2 ));
    assert_false( ias_pm_add_pixels( pm, 0, 61, 1, 0x2 ));
    assert_false( ias_pm_add_pixels( pm, 0, 55, 2, 0x2 ));
    assert_false( ias_pm_add_pixels( pm, 0, 30, 10, 0x2 )); 
    
    it = ias_pm_get_iterator( pm, IAS_PM_INCLUDE, 0x3 );
    assert_not_equal( it, NULL );

    span_count = 0;
    while ( ias_pm_get_next_span( it, &span ))
    {
        span_count++;
    }
    if ( span_count == 1 )
    {
        assert_equal( span.starting_pixel_index, 30 );
        assert_equal( span.length_of_span, 33 );
        assert_equal( span.pixel_mask, 0x3 );
    }
    else
    {   
        assert_true_with_message( 0, "Updating intersecting span into "
            "existing span failed" );
    }
    ias_pm_destroy_iterator( it );
    /* End Test 4 */

    /* Test 5 - check for preset bit in existing pixel mask */

    assert_false( ias_pm_add_pixels( pm, 0, 25, 7, 0x1 ));
    assert_false( ias_pm_add_pixels( pm, 0, 43, 1, 0x1 ));
    assert_false( ias_pm_add_pixels( pm, 0, 60, 6, 0x1 ));

    it = ias_pm_get_iterator( pm, IAS_PM_INCLUDE, 0x3 );
    assert_not_equal( it, NULL );

    span_count = 0;
    while ( ias_pm_get_next_span( it, &span ))
    {
        span_count++;
        if ( span_count == 1 )
        {
            assert_equal( span.starting_pixel_index, 25 );
            assert_equal( span.length_of_span, 5 );
            assert_equal( span.pixel_mask, 0x1 );
        }
        else if ( span_count == 2 )
        {
            assert_equal( span.starting_pixel_index, 30 );
            assert_equal( span.length_of_span, 33 );
            assert_equal( span.pixel_mask, 0x3 );
        }
        else if ( span_count == 3 )
        {
            assert_equal( span.starting_pixel_index, 63 );
            assert_equal( span.length_of_span, 3 );
            assert_equal( span.pixel_mask, 0x1 );
        }
    }
    if (span_count != 3)
    { 
        printf("count %d\n", span_count);
        assert_true_with_message( 0, "Setting bit again in pixel mask of "
            "existing span failed" );
    }
    ias_pm_destroy_iterator( it );

    /* End Test 5 */

    /* Free pixel mask */
    ias_pm_destroy( pm );

} /* END add_pixel_test */

