/*------------------------------------------------------------------------------
NAME:    linked_list_test

PURPOSE: Test routine for the linked list library.

------------------------------------------------------------------------------*/
#include <stdlib.h>
#include "cgreen/cgreen.h"
#include "ias_linked_list.h"

#define NUM 15
#define MID 10

/* Define a structure that includes the linked list node structure */
typedef struct test_struct
{
    int a;
    double b;
    IAS_LINKED_LIST_NODE linked_list_node;
} TEST_STRUCT;

/* Define a linked list delete macro.  This is not absolutely necessary
since one may call ias_linked_list_delete directly with the offset.  */
#define LINKED_LIST_DELETE( list ) \
    ias_linked_list_delete( (list), \
    offsetof( TEST_STRUCT, linked_list_node ));


/* Local prototypes */
TEST_STRUCT *allocate_node( void );
Ensure ll_test( void );

/*--------------------- Main -----------------------*/
int main( int argc, char *argv[] )
{
    TestSuite *suite = create_test_suite();
    add_test( suite, ll_test );
    return run_test_suite( suite, create_text_reporter() );
}  /* END main */

/* This is a convenience routine to allocate and initialize the test
structure */
TEST_STRUCT *allocate_node( void )
{
    TEST_STRUCT *node = malloc( sizeof( TEST_STRUCT ));
    assert_not_equal( node, NULL );
    if ( node )
        ias_linked_list_initialize_node( &node->linked_list_node );

    return node;
}


Ensure ll_test( void )
{
    IAS_LINKED_LIST_NODE base_node;  /* This is the head of the list.  It
                                        should not be deleted. */
    IAS_LINKED_LIST_NODE *scratch;   /* Pointer to a linked list node */
    TEST_STRUCT *current;
    TEST_STRUCT *node_to_free = NULL;
    TEST_STRUCT *first_node;
    TEST_STRUCT *last_node;
    int i;
    int list_count;
    int count;
    /* These arrays contain the values after the first round of inserts */
    int a1[] = {44,0,1,2,3,4,5,6,7,8,14,13,12,11,10,9,99};
    double b1[] = {44.25,0.0,1.25,2.5,3.75,5.0,6.25,7.5,8.75,10.0,17.5,
                 16.25,15.0,13.75,12.5,11.25,99.25};
    /* These arrays contain the values after removing the node with value
    of 6 in a. */
    int a2[] = {44,0,1,2,3,4,5,7,8,14,13,12,11,10,9,99};
    double b2[] = {44.25,0.0,1.25,2.5,3.75,5.0,6.25,8.75,10.0,17.5,
                 16.25,15.0,13.75,12.5,11.25,99.25};

    /* Initialize the linked list head */
    /* Base node should point to itself */
    ias_linked_list_initialize_node( &base_node );
    assert_equal( &base_node, base_node.next );
    assert_equal( base_node.next, base_node.prev );

    /* Get a pointer to the head node.  When attempting to get the head of
    an empty list NULL will be returned */
    scratch = ias_linked_list_get_head_node( &base_node );
    assert_equal( scratch, NULL );

    /* For convenience give scratch a value by pointing it to the base
    node.  This is so the get next node in the loop won't fail. */
    scratch = base_node.next;

    /* Allocate and add several nodes to the linked list */
    for ( i = 0; i < NUM; i++ )
    {
        current = allocate_node();
        assert_not_equal( current, NULL );
        if ( !current )
            return;
        /* Set the member values */
        current->a = i;
        current->b = i * 1.25;

        /* For the first part of the list insert after the current node */
        if ( i < MID )
        {
            /* Get the next node in the list, first time through it is the
            an empty list so the returned value will be NULL, all subsequent
            times it will be the node added in the previous iteration of
            the loop. */
            scratch = ias_linked_list_get_next_node( &base_node, scratch );
            if ( !scratch )
                scratch = base_node.next;

            /* Add the new node after the current one */
            ias_linked_list_insert_after( scratch,
                &current->linked_list_node );
        }
        else
        {
            /* Get the next node pointer if this is the first time here, it
            will be the previously added node.  All subsequent times get the
            previous pointer ... again it will be the previously added 
            node.  */
            static int f = 1;
            if ( f )
            {
                f = 0;
                scratch = ias_linked_list_get_next_node( &base_node, scratch);
            }
            else
            {
                scratch = ias_linked_list_get_prev_node( &base_node, scratch);
            }
            assert_not_equal( scratch, NULL );

            /* Add the new node before the current one */
            ias_linked_list_insert_before( scratch,
                &current->linked_list_node );
        }

    }  /* END for i loop */

    /* Add a node to the head of the list */
    first_node = allocate_node();
    assert_not_equal( first_node, NULL );
    if ( !first_node )
        return;
    /* Set the member values */
    first_node->a = 44;
    first_node->b = 44.25;
    ias_linked_list_add_head( &base_node,
        &first_node->linked_list_node );

    /* Add a node to the tail of the list */
    last_node = allocate_node();
    assert_not_equal( last_node, NULL );
    if ( !last_node )
        return;
    /* Set the member values */
    last_node->a = 99;
    last_node->b = 99.25;
    ias_linked_list_add_tail( &base_node,
        &last_node->linked_list_node );

    /* Go through the list and check that the list is in the correct order */
    count = 0;
    GET_OBJECT_FOR_EACH_ENTRY( current, &base_node, TEST_STRUCT,
        linked_list_node )
    {
        assert_equal( current->a, a1[count] );
        assert_double_equal( current->b, b1[count] );
        count++;
    }

    /* Remove the node with a == 6 and check the list again */
    GET_OBJECT_FOR_EACH_ENTRY( current, &base_node, TEST_STRUCT,
        linked_list_node )
    {
        if ( current->a == 6 )
        {
            ias_linked_list_remove_node( &current->linked_list_node );
            /* Note: cannot free the structure until it's used after its
            use in GET_OBJECT_FOR_EACH_ENTRY on the next iteration of the
            loop.  Save a pointer to it and remove it after the loop. */
            node_to_free = current;
        }
    }
    free( node_to_free );
    count = 0;
    GET_OBJECT_FOR_EACH_ENTRY( current, &base_node, TEST_STRUCT,
        linked_list_node )
    {
        assert_equal( current->a, a2[count] );
        assert_double_equal( current->b, b2[count] );
        count++;
    }

    /* Get and test the head node */
    scratch = ias_linked_list_get_head_node( &base_node );
    assert_not_equal( scratch, NULL );
    if ( scratch )
    {
        current = GET_OBJECT_OF( scratch, TEST_STRUCT, linked_list_node );
        assert_equal( current->a, 44 );
        assert_double_equal( current->b, 44.25 );
    }

    /* Make sure there are 16 nodes */
    list_count = ias_linked_list_count_nodes(&base_node );
    assert_equal( list_count, 16 );

    /* Now delete the entire list */
    LINKED_LIST_DELETE( &base_node );
    list_count = ias_linked_list_count_nodes(&base_node );
    assert_equal( list_count, 0 );

    /* Base node should now point to itself again */
    assert_equal( &base_node, base_node.next );
    assert_equal( base_node.next, base_node.prev );

}  /* END ll_test */



