#include <stdio.h>
#include <unistd.h>
#include <pthread.h>
#include "cgreen/cgreen.h"
#include "../ias_threadpool.h"
#include "../ias_threadsync.h"
#include "../ias_work_queue.h"
#include "ias_const.h"

static IAS_THREAD_MUTEX_TYPE mutex;
static IAS_THREAD_COND cond;
static int condition = 0;
static int total_produced[4] = {0};

/* These are boolean check functions that are passed to
   ias_thread_wait_on_condition.  The first one uses the passed parameter
   and the second one ignores it. */
static int check_function1
(
    void *check_parameter
)
{
    int *condition_parameter = (int *)check_parameter;
    if ( *condition_parameter == 0 )
        return TRUE;
    else
        return FALSE;
}

static int check_function2
(
    void *check_parameter
)
{
    if ( condition == 0 )
        return TRUE;
    else
        return FALSE;
}


/* This routine ``produces'' by incrementing counters.  It is designed to
   force two threads to wait when they produce #3 and #6.  Only 10 values
   should be produced. */
void *produce( void *arg )
{
    long id = (long)arg;
    int done = FALSE;

    /* Loop until the task is done */
    while ( 1 )
    {
        /* Lock the mutex increment the count */
        IAS_THREAD_LOCK_MUTEX( &mutex );
        condition++;
        total_produced[id-1]++;
        total_produced[3]++;

        /* The threads that produce 3 and 6 will wait on the condition.
           Call the first function using a valid condition. */
        if ( condition == 3 )
        {
            ias_thread_wait_on_condition( &mutex,
                &cond, check_function1, &condition );
            done = TRUE;
        }
        /* Call the second function with NULL */
        else if ( condition == 6 )
        {
            ias_thread_wait_on_condition( &mutex,
                &cond, check_function2, NULL );
            done = TRUE;
        }

        /* The remaining thread will produce the rest and set the condition
           variable to zero. */
        if ( condition == 10 )
        {
            condition = 0;
            done = TRUE;
        }

        IAS_THREAD_UNLOCK_MUTEX( &mutex );

        /* Return when done */
        if ( done )
        {
            ias_thread_broadcast_condition( &mutex, &cond );
            return 0;
        }
        /* Otherwise do some ``work'' to allow other threads the opportunity
           to produce as well. */
        usleep( 100 );
    }
}


Ensure test_wait_on_condition()
{
    int i;
    pthread_t my_thread[3];
    long t1 = 1, t2 = 2, t3 = 3;
    pthread_attr_t attr;
    int status;
    int sum;

    /* Create the mutex and condition variable */
    condition = 0;
    status = IAS_THREAD_CREATE_MUTEX( &mutex );
    if ( status )
    {
        assert_true_with_message( 0, "Error creating mutex" );
        return;
    }
    status = IAS_THREAD_CREATE_COND( &cond );
    if ( status )
    {
        assert_true_with_message( 0, "Error creating condition" );
        IAS_THREAD_DESTROY_MUTEX( &mutex );
        return;
    }

    /* Create three joinable threads */
    pthread_attr_init( &attr );
    pthread_attr_setdetachstate( &attr, PTHREAD_CREATE_JOINABLE );
    pthread_create( &my_thread[0], &attr, produce, (void *)t1 );
    pthread_create( &my_thread[1], &attr, produce, (void *)t2 );
    pthread_create( &my_thread[2], &attr, produce, (void *)t3 );

    /* wait on the threads */
    for ( i = 0; i < 3; i++ )
        pthread_join( my_thread[i], NULL );

    /* Find out how many values were produced */
    for ( i = 0, sum = 0; i < 3; i++ )
    {
        sum += total_produced[i];
    }

    /* Make sure there are only ten produced. */
    assert_equal( sum, total_produced[3] );
    assert_equal( sum, 10 );
}

Ensure test_get_thread_count()
{
    struct ias_threadpool *pool;
    int num_threads;

    pool = ias_threadpool_initialize(3); /* initialize for 3 threads */

    num_threads = ias_threadpool_get_thread_count(pool);

    assert_equal(num_threads, 3);

    ias_threadpool_destroy(pool);

}


int main( int argc, char *argv[] )
{
    TestSuite *suite = create_test_suite();
    add_test( suite, test_wait_on_condition );
    add_test( suite, test_get_thread_count );
    return run_test_suite( suite, create_text_reporter() );
}
