#!/usr/bin/env python3
""" IAS database module

    Exported Classes:
        DatabaseConnection - database connection class with connection methods

"""
import os
import cx_Oracle

""" The following disconnect flags are used when closing the database.
    When only querying a database it is best to use the rollback flag.
"""
DISCONNECT_COMMIT = 0
DISCONNECT_ROLLBACK = 1


class IASDBError(Exception):
    """ IAS database exception class. This is the generic exception thrown
        when an error occurs in the DatabaseConnection class.
    """
    pass


class DatabaseConnection:

    def __init__(self, db_schema=None, satellite=None, sensor=None):
        """ Creates a new instance of the database_connection class

            Optional arguments:
                These arguments are optional.  If provided, only db_schema
                or the satellite/sensor pair are needed.
                db_schema -- Schema information to set to defaults to the
                             COMMON schema
                satellite -- Satellite name (like "L8")
                sensor -- Sensor name (like "OLITIRS")

                satellite/sensor: Optional arguements that must be specified
                                  together. Will set the schema according
                                  to the satellite and sensor specified.

            Returns:
                DatabaseConnection object
        """
        # Initialize all of the instance variables
        self.username = ""
        self.password = ""
        self._connection = None

        # Set the database connection info if schema is passed
        if db_schema:
            self.set_connection_info(db_schema)

        # Set the database connection infor using the satellite and sensor
        if satellite is not None and sensor is not None:
            self.set_login_using_satellite_sensor(satellite, sensor)

    def __enter__(self):
        """ Called when using the "with" convention. No need to call
            connect() when using "with" it will automatically be called
            for ease of use.
        """
        if not self._connection:
            self.connect()
        return self

    def __exit__(self, type, value, traceback):
        """ Called when exiting the "with" convention. This will do all
            of the cleanup nessary if an exception occurs. If no exeception
            occurs and close has not been called it will rollback any changes.
        """
        if type == Exception:
            self._cleanup()
        elif self._connection:
            self.close(DISCONNECT_COMMIT)

    def set_connection_info(self, db_schema=None):
        """ Sets the database information based on the schema asked for.
            This method uses the environment variables to get the username
            and password for the schema specified.

            Arguments:
                db_schema -- Schema information to set to defaults to the
                             COMMON schema

            Returns:
                Void
        """

        # Default the schema to the common schema
        if db_schema is None:
            db_schema = 'COMMON'

        db_con_str = ''

        # Get the schema username and password using the environment variables
        if db_schema == 'COMMON' and os.getenv('IAS_DB_COM'):
            db_con_str = os.getenv('IAS_DB_COM')
        elif db_schema == 'L7_ETM' and os.getenv('IAS_DB_L7_ETM'):
            db_con_str = os.getenv('IAS_DB_L7_ETM')
        elif db_schema == 'L5_TM' and os.getenv('IAS_DB_L5_TM'):
            db_con_str = os.getenv('IAS_DB_L5_TM')
        elif db_schema == 'L4_TM' and os.getenv('IAS_DB_L4_TM'):
            db_con_str = os.getenv('L8_OLITIRS')
        elif db_schema == 'L8_OLITIRS' and os.getenv('IAS_DB_L8_OLITIRS'):
            db_con_str = os.getenv('IAS_DB_L8_OLITIRS')
        else:
            self.username = "/"
            self.password = ""
            return

        # Split the username and password into individual variables
        self.username, self.password = db_con_str.split('/')

    def set_login_using_satellite_sensor(self, satellite, sensor):
        """ Sets the database information based on the satellite and sensor
            asked for.

            Arguments:
                satellite -- Satellite name (like "L8")
                sensor -- Sensor name (like "OLITIRS")
        """

        # Build the environment variable string
        db_env = "IAS_DB_{0}_{1}".format(satellite, sensor)

        # Try and retrieve the connection string
        db_con_str = os.getenv(db_env)

        # Set the username and password
        self.username, self.password = db_con_str.split('/')

    def connect(self):
        """ Creates the database connection using the instance specified schema

            Note: SetDBConnectionInfo should be called before creating a
                  connection or the default username and password will be
                  assumed.

            Arguments: None

        """
        if self._connection:
            raise IASDBError("Connection already established, close the "
                             "existing connection before starting a new one")
        try:
            sid = os.getenv('ORACLE_SID')
            connection_string = self.username + "/" + self.password + '@' + sid
            self._connection = cx_Oracle.connect(connection_string)
        except (cx_Oracle.DatabaseError, OSError) as exc:
            raise IASDBError("Creating connection to database using "
                             "connection string: " + connection_string) \
                             from exc

    def close(self, commit):
        """ Closes the open database connection

            Arguments:
                commit -- Rollback or Commit flag
        """
        if not self._connection:
            raise IASDBError("No database connection has been established")
        try:
            if commit == DISCONNECT_ROLLBACK:
                self._connection.rollback()
            elif commit == DISCONNECT_COMMIT:
                self._connection.commit()
            else:
                raise IASDBError("Incorrect commit flag was passed to method "
                                 "only DISCONNECT_ROLLBACK and "
                                 "DISCONNECT_COMMIT are valid flags ")
            self._connection.close()
            self._connection = None
        except cx_Oracle.DatabaseError as exc:
            raise IASDBError("Could not close connection properly") from exc

    def execute(self, sql, params=None):
        """ Executes the sql argument passed using the existing connection

            Arguments:
                sql    - Input sql to execute
                params - Parameters to be processed as part of the sql

            Returns:
                Dataset on success
                IASDBError exception on error

            Note: The user is responsible for closing the returned cursor
                  object. The cursor will be automatically closed if
                  an exception occurs for this class.
        """
        if self._connection:
            # Catch any exceptions that might occur
            try:
                # Get the connection cursor
                cursor = self._connection.cursor()

                # Ensure the parameters pass is a dictionary
                if params and type(params) is not dict:
                    raise IASDBError("Incorrect type for parameters dictionary"
                                     " was expected but {0} was passed".format(
                                         typeof(params)))

                # Determine if parameters were passed as an argument
                if params and type(params) is dict:
                    cursor.execute(sql, params)
                else:
                    cursor.execute(sql)
                return cursor
            except cx_Oracle.DatabaseError as exc:
                raise IASDBError("Executing sql: " + sql + " - Parameters: " +
                                 str(params)) from exc
        else:
            raise IASDBError("No database connection established")

    def _cleanup(self):
        """ Internal class cleanup method
        """
        # Close the connection
        if self._connection:
            self._connection.rollback()
            self._connection.close()

        # Set the connection and cursor to None so they cannot be used
        self._connection = None

    def begin_transaction(self):
        """ Begins new transaction with connection
        """
        if not self._connection:
            raise IASDBError("Transaction started with no database connection")

        self._connection.begin()
        return self

    def commit(self):
        """ Commit the pending transaction to the database"""
        if not self._connection:
            raise IASDBError("Commit attempted with no database connection")

        self._connection.commit()

    def rollback(self):
        """ Rollback the pending transaction in the database"""
        if not self._connection:
            raise IASDBError("Rollback attempted with no database connection")

        self._connection.rollback()

    def get_cursor(self):
        """ Returns a cursor object related to the connection """
        if self._connection:
            return cx_Oracle.Cursor(self._connection)
        else:
            raise IASDBError("No connection has been established")
