#!/usr/bin/env python3
""" test_database

    Database unit test script
"""
import unittest
from subprocess import call
import os
import imp
import database as db


class TestDatabase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        """ Setup method ran before all unit tests"""
        db_cred = os.getenv("IAS_DB_L8_OLITIRS")
        if not db_cred:
            raise Exception("IAS_DB_L8_OLITIRS environment variable not set")

        status = call(["sqlplus", db_cred, "@create_db_table.sql"])
        if status != 0:
            raise Exception("Creating db tables")

    def test_enter_exit(self):
        """ Tests that the __enter__ and __exit__ behave as intended """
        myconn = db.DatabaseConnection("L8_OLITIRS")
        with myconn as conn:
            cur = conn.execute("Select * from DATABASE_UNIT_TEST")
            rows = cur.fetchall()
            self.assertEqual(rows[0][0], "UNIT_TEST_IAS")

        # Check that after exiting the with statement the connection
        # has been severed
        self.assertRaises(db.IASDBError, myconn.execute,
                          ("Select * from DATABASE_UNIT_TEST"), ())

    def test_connection_info(self):
        """ Tests that the set_connection_info method behaves as
            intended
        """
        myconn = db.DatabaseConnection()
        con_str = os.getenv("IAS_DB_COM")
        user, psswd = con_str.split('/')

        # Set the connection info
        myconn.set_connection_info()
        self.assertEqual(user, myconn.username)
        self.assertEqual(psswd, myconn.password)

        # Check another one
        con_str = os.getenv("IAS_DB_L8_OLITIRS")
        user, psswd = con_str.split('/')

        # Reset the connection info
        myconn.set_connection_info("L8_OLITIRS")
        self.assertEqual(user, myconn.username)
        self.assertEqual(psswd, myconn.password)

        myconn.set_connection_info("")
        self.assertEqual("/", myconn.username)
        self.assertEqual("", myconn.password)

    def test_connect_close_execute(self):
        """ Tests the connect, close, and execute methods """
        myconn = db.DatabaseConnection("L8_OLITIRS")
        self.assertEqual(myconn._connection, None)
        myconn.connect()
        self.assertNotEqual(myconn._connection, None)

        # Select all the rows from the table
        cur = myconn.execute("Select * from DATABASE_UNIT_TEST")
        rows = cur.fetchall()
        self.assertEqual(cur.rowcount, 1)
        self.assertEqual(rows[0][0], "UNIT_TEST_IAS")

        # Insert a new row
        cur = myconn.execute("INSERT INTO DATABASE_UNIT_TEST (\"NAME\") "
                             "VALUES ('UNIT_TEST_IAS2')")
        self.assertEqual(cur.rowcount, 1)
        cur = myconn.execute("Select * from DATABASE_UNIT_TEST")
        rows = cur.fetchall()
        self.assertEqual(cur.rowcount, 2)

        # Rollback the changes
        myconn.close(db.DISCONNECT_ROLLBACK)
        self.assertEqual(myconn._connection, None)
        myconn.connect()
        self.assertNotEqual(myconn._connection, None)

        # Select all the rows from the table, which should still be only
        # one row since the changes were rolled back
        cur = myconn.execute("Select * from DATABASE_UNIT_TEST")
        rows = cur.fetchall()
        self.assertEqual(cur.rowcount, 1)
        self.assertEqual(rows[0][0], "UNIT_TEST_IAS")

        # Insert a new row
        cur = myconn.execute("INSERT INTO DATABASE_UNIT_TEST (\"NAME\") "
                             "VALUES (:t1)", {'t1': 'UNIT_TEST_IAS2'})
        self.assertEqual(cur.rowcount, 1)
        cur = myconn.execute("Select * from DATABASE_UNIT_TEST")
        rows = cur.fetchall()
        self.assertEqual(cur.rowcount, 2)

        # Commit the results this time
        myconn.close(db.DISCONNECT_COMMIT)
        self.assertEqual(myconn._connection, None)
        myconn.connect()
        self.assertNotEqual(myconn._connection, None)

        # Select all the rows from the table, which should still be only
        # one row since the changes were rolled back
        cur = myconn.execute("Select * from DATABASE_UNIT_TEST order by name")
        rows = cur.fetchall()
        self.assertEqual(cur.rowcount, 2)
        self.assertEqual(rows[0][0], "UNIT_TEST_IAS")

        cur = myconn.execute("DELETE from DATABASE_UNIT_TEST where NAME = "
                             ":t1", {'t1': 'UNIT_TEST_IAS2'})
        self.assertEqual(cur.rowcount, 1)

        # Rollback the final time
        myconn.close(db.DISCONNECT_COMMIT)
        self.assertEqual(myconn._connection, None)

        # Check that after closing that the connection
        # has been severed
        self.assertRaises(db.IASDBError, myconn.execute,
                          ("Select * from DATABASE_UNIT_TEST"), ())

    @classmethod
    def tearDownClass(cls):
        """ Tear down method ran after all unit tests"""
        db_cred = os.getenv("IAS_DB_L8_OLITIRS")
        if not db_cred:
            raise Exception("IAS_DB_L8_OLITIRS environment variable not set")

        status = call(["sqlplus", db_cred, "@cleanup_db_table.sql"])
        if status != 0:
            raise Exception("Cleaning up db tables")


if __name__ == '__main__':
    unittest.main()
