#!/usr/bin/env python3
""" IAS CPF module

    Exported Classes:
        CPF

    Exported Constants:
        _FILENAME_SERVICE
        _DATE_SERVICE

    Dynamic Methods:
        All of the group accessor method are dynamically created using
        the _groups variable. Any group can be accessed using the following
        syntax:
            cpf.get_{group_name}
        This will return an object of type {group_name}. You can access
        all attributes on the object using the following syntax:
            obj = cpf.get_{group_name}
            obj.{attribute_name}

        Example using the file attributes(FILE_ATTRIBUTES) group:
            file_attr = cpf.get_file_attributes
            spacecraft = file_attr.spacecraft_name
            version = file_attr.version
"""
import urllib.request
import tempfile
import time
import os
import ias.odl.odl as par
from ias.log import logger as lg
from ias.io.cpf import oli_gains
from types import MethodType

# Create the service constants
_FILENAME_SERVICE = ("{0}/dataservices/rest/cpf?operation=getCalParmFile&"
                    "filename={1}")
_DATE_SERVICE = ("{0}/dataservices/rest/cpf?operation=getCalParmFile&"
                "nearMatch=true&date={1}")

# Groups in the CPF that will be loaded/created dynamically. If special
# handling is needed remove from this list and add a specific accessor method
_groups = {'FILE_ATTRIBUTES', 'UT1_TIME_PARAMETERS', 'CLOUD_COVER_ASSESSMENT',
          'EARTH_CONSTANTS', 'GEO_SYSTEM', 'TIRS_NONUNIFORMITY',
          'OLI_POST_RELATIVE_GAINS', 'TIRS_POST_RELATIVE_GAINS',
          'OLI_SATURATION_LEVEL', 'TIRS_SCA_PARAMETERS', 'LUNAR_IRRADIANCE',
          'TIRS_DETECTOR_RESPONSE_BLIND', 'IMPULSE_NOISE',
          'TIRS_TEMP_SENSITIVITY', 'ANCILLARY_ENG_CONV',
          'TIRS_PRE_RELATIVE_GAINS', 'OLI_SCA_PARAMETERS',
          'OLI_DETECTOR_OFFSETS', 'TIRS_ALIGN_CAL', 'TIRS_FRAMING_PARAMETERS',
          'HISTOGRAM_CHARACTERIZATION', 'LAMP_RADIANCE', 'OLI_NONUNIFORMITY',
          'TIRS_PRE_REL_GAINS_BLIND', 'TIRS_DETECTOR_NOISE', 'AVERAGE_BIAS',
          'TIRS_RADIANCE_RESCALE', 'TIRS_DETECTOR_RESPONSE',
          'LOS_MODEL_CORRECTION', 'REFLECTANCE_CONVERSION',
          'OLI_PRE_RELATIVE_GAINS', 'TIRS_POST_REL_GAINS_BLIND',
          'TIRS_DETECTOR_OFFSETS', 'TIRS_PARAMETERS', 'OLI_FOCAL_PLANE',
          'ORBIT_PARAMETERS', 'DIFFUSER_RADIANCE', 'OLI_TEMP_SENSITIVITY',
          'OLI_RADIANCE_RESCALE', 'TIRS_DETECTOR_STATUS',
          'GCP_CORRELATION', 'TIRS_THERMAL_CONSTANTS',
          'TIRS_RELATIVE_GAINS', 'TIRS_ABSOLUTE_GAINS', 'ATTITUDE_PARAMETERS',
          'FOCAL_PLANE_CAL', 'B2B_ASSESSMENT', 'ANCILLARY_QA_THRESHOLDS',
          'TIRS_REL_GAINS_BLIND', 'TIRS_SATURATION_LEVEL',
          'TIRS_ABSOLUTE_GAINS_BLIND', 'OLI_DETECTOR_STATUS',
          'OLI_DETECTOR_NOISE', 'OLI_FRAMING_PARAMETERS', 'OLI_PARAMETERS',
          'TIRS_FOCAL_PLANE', 'I2I_ASSESSMENT'}


class CPF():
    """ Class used to retrieve, read, and parse CPF files """

    def __init__(self, filename=None, cpf_name=None, date=None):
        """ Creates new instance of a CPF object

            NOTE:
                If filename is set, the cpf will be loaded from file
                If cpf_name is set, the cpf will be loaded by name from
                the service
                If date is set, the cpf will be loaded by date from the
                service
        """
        self.data = {}
        self.init_flags = dict()

        # Set where the CPF will be read from
        self.filename = filename
        self.cpf_name = cpf_name
        if date is not None:
            date = time.strftime(
                '%Y-%m-%dT00:00:00', time.strptime(date, "%d-%b-%Y"))
        self.date = date

        # Initialize the ODL parser
        self.lines = None
        self.parser = par.ODLParser()

        # Create all of the accessor methods dynamically from the list of
        # groups
        for name in _groups:
            setattr(self, "get_{0}".format(name.lower()),
                    MethodType(create_cpf_object(name), self))

    def _initialize(self, group):
        """ Private method used to check if a group has already been
            initialized

            Returns: CPF object
        """

        # If already initialized return
        if group in self.init_flags:
            return self

        # Set the initialization flag for this group
        self.init_flags[group] = True
        return self.load(group)

    def _load_from_file(self, group):
        """ Reads the CPF from file

            Returns:
                CPF object
        """

        # Check that the filename attribute was set
        if self.filename is None:
            raise CPFError("CPF filename was not set")

        with open(self.filename) as f:
            self.lines = f.readlines()

        # Parse the CPF and update the current dictionary
        self.parser.load(iter(self.lines), group)
        self.data.update(self.parser.contents)

        return self

    def _load_from_service(self, service, group):
        """ Reads the CPF from file

            Arguments:
                service: Service url used to retrieve cpf

            Returns:
                CPF object
        """
        try:
            # Save the CPF to a temporary file and read it in
            with tempfile.NamedTemporaryFile("w+t") as f:
                f.write(urllib.request.urlopen(service).read().decode('utf-8'))
                f.flush()
                f.seek(0)
                self.lines = f.readlines()
                self.parser.load(iter(self.lines), group)

            # Update the current dictionary
            self.data.update(self.parser.contents)
        except Exception as e:
            raise CPFServiceError("Unable to retreive cpf using service call "
                                  "({0})".format(service))
        return self

    def load(self, group=None):
        """ Loads CPF using the passed parameter

            Returns:
                CPF object
        """

        # If CPF is already loaded into memory parse group and return
        if self.lines:
            self.parser.load(iter(self.lines), group)
            self.data.update(self.parser.contents)
            return self

        # Check if load from file was requested
        if self.filename is not None:
            return self._load_from_file(group)

        # All load methods from here on out use the service so retrieve
        # the correct one to use from the environment variable
        ias_services = os.getenv("IAS_SERVICES")
        if not ias_services:
            raise CPFServiceError("IAS_SERVICES environment variable not set")

        # Create the service url
        service = None
        if self.cpf_name is not None:
            service = (_FILENAME_SERVICE.format(ias_services, self.cpf_name))
        if self.date is not None:
            service = (_DATE_SERVICE.format(ias_services, self.date))

        return self._load_from_service(service, group)

    def get_oli_relative_gains(self, sat_attr, combined=None):
        """ Initializes the CPF and returns the oli relative gains object

            Parameters:
                combined: Combine all scas for each band
        """
        return oli_gains.RELATIVE_GAINS(
            self._initialize("OLI_RELATIVE_GAINS"), sat_attr, combined)

    def get_oli_absolute_gains(self, sat_attr):
        """ Initializes the CPF and returns the oli absolute gains object """
        return oli_gains.ABSOLUTE_GAINS(self._initialize("OLI_ABSOLUTE_GAINS"),
                                        sat_attr)


def create_cpf_object(name):
    """ Method closure used to create dynamic assessor functions using
        the input group name. Using the input ("name") this function will
        return a method that will retrieve the group from the cpf.

        Returns: create_object method
    """
    def create_object(self):
        """ Method to dynamically create the CPF group object and save it
            for future use

            Returns: CPF group object
        """

        # Store off the function name using the closure
        class_name = name

        # Check if the object already exists
        class_exists = getattr(self, class_name, None)
        if class_exists:
            return class_exists

        # Create a new class object
        new_class = type(class_name, (object,), {})
        self._initialize(class_name)

        # Add the attributes to the object
        for k, v in self.data[class_name].items():
            if 'value' in v:
                setattr(new_class, k.lower(), v['value'])

        # Save the object to the current CPF object for later use
        setattr(self, class_name, new_class)
        return new_class

    return create_object


class CPFServiceError(Exception):
    pass


class CPFError(Exception):
    pass
