#!/usr/bin/env python3
import argparse
import urllib.request
import tempfile
import os
from ias.log import logger as lg
from ias.misc.const import SUCCESS, ERROR
from ias.io.cpf.cpf import CPF

# Create the service constants
_FILENAME_SERVICE = ("{0}/dataservices/rest/cpf?operation=getCalParmFile&"
                    "filename={1}")
_DATE_SERVICE = ("{0}/dataservices/rest/cpf?operation=getCalParmFile&"
                "nearMatch=true&date={1}")
_DATETYPE = 1
_FILETYPE = -1

def retrieve_cpf(intype, incpf, output_name, outcpf=None):
    """ Static Method used to retrieve a CPF using the CPF name or CPF date

        Parameters:
            intype: Type of CPF retrieval to perform: "FILE" or "DATE"
            incpf:  CPF Name or Date to use in retrieval
            output_name: Output CPF name to use, or None if the name should be
                         determined from the file metadata
            outcpf: Optional parameter to return the name of the CPF if 
                    output_name is None. It should be an empty list.
    """
    use_cpf_name = False

    # Get the correct service to use
    ias_services = os.getenv("IAS_SERVICES")
    if not ias_services:
        raise ValueError("IAS_SERVICES environment variable not set")

    # Parse the first argument
    type = _DATETYPE
    if intype == "FILE":
        type = _FILETYPE

    # Build the correct service url
    service = None
    if type == _FILETYPE:
        service = (_FILENAME_SERVICE.format(ias_services, incpf))
    else:
        service = (_DATE_SERVICE.format(ias_services, incpf))

    if output_name is None:
        use_cpf_name = True
        output_name = incpf

    # Save the CPF to file after requesting it from the service
    try:
        with open(output_name, "w") as f:
            f.write(urllib.request.urlopen(service).read().decode('utf-8'))
    except Exception as e:
        lg.Logger.get_default_logger().error(
            "Unable to retrieve the requested cpf using "
            "{0}".format(incpf))
        return ERROR

    # Move the file to the correct filename
    if (use_cpf_name):
        cpf = CPF(filename=output_name)
        attr = cpf.get_file_attributes()
        os.rename(output_name, attr.file_name)
        output_name = attr.file_name

    # If needed, populate the outcpf list with the cpf name
    if (outcpf is not None):
        outcpf.append(output_name)

    lg.Logger.get_default_logger().info(
        "CPF saved as {0}".format(output_name))

    return SUCCESS

# Runs static method retrieve_cpf if called from command line
if __name__ == "__main__":
    argparser = argparse.ArgumentParser(description='Retrieve CPF')
    argparser.add_argument(
        metavar='CPF', dest='cpf', help='CPF Name or CPF Date')
    argparser.add_argument(
        '--output_name', dest='output_name', required=False,
        help='CPF output filename')
    argparser.add_argument(
        '--type', dest='type', required=True,
        help="Type of retrieval to be used", choices=['FILE', 'DATE'])
    args = argparser.parse_args()

    exit(retrieve_cpf(args.type, args.cpf, args.output_name))
