#!/usr/bin/env python3
""" test_cpf

    Verifys that each getter method works for each group in the provided CPF.
    This also verifies that the CPF was read correctly.
"""
import os
import unittest
import imp
from subprocess import call
import cpf as ias_cpf


GROUPS = {'FILE_ATTRIBUTES', 'UT1_TIME_PARAMETERS', 'CLOUD_COVER_ASSESSMENT',
          'EARTH_CONSTANTS', 'GEO_SYSTEM', 'TIRS_NONUNIFORMITY',
          'OLI_POST_RELATIVE_GAINS', 'TIRS_POST_RELATIVE_GAINS',
          'OLI_SATURATION_LEVEL', 'TIRS_SCA_PARAMETERS', 'LUNAR_IRRADIANCE',
          'TIRS_DETECTOR_RESPONSE_BLIND', 'IMPULSE_NOISE',
          'TIRS_TEMP_SENSITIVITY', 'ANCILLARY_ENG_CONV',
          'TIRS_PRE_RELATIVE_GAINS', 'OLI_SCA_PARAMETERS',
          'OLI_DETECTOR_OFFSETS', 'TIRS_ALIGN_CAL', 'TIRS_FRAMING_PARAMETERS',
          'HISTOGRAM_CHARACTERIZATION', 'LAMP_RADIANCE', 'OLI_NONUNIFORMITY',
          'TIRS_PRE_REL_GAINS_BLIND', 'TIRS_DETECTOR_NOISE', 'AVERAGE_BIAS',
          'TIRS_RADIANCE_RESCALE', 'TIRS_DETECTOR_RESPONSE',
          'LOS_MODEL_CORRECTION', 'REFLECTANCE_CONVERSION',
          'OLI_PRE_RELATIVE_GAINS', 'TIRS_POST_REL_GAINS_BLIND',
          'TIRS_DETECTOR_OFFSETS', 'TIRS_PARAMETERS', 'OLI_FOCAL_PLANE',
          'ORBIT_PARAMETERS', 'DIFFUSER_RADIANCE', 'OLI_TEMP_SENSITIVITY',
          'OLI_RADIANCE_RESCALE', 'TIRS_DETECTOR_STATUS',
          'GCP_CORRELATION', 'TIRS_THERMAL_CONSTANTS',
          'TIRS_RELATIVE_GAINS', 'TIRS_ABSOLUTE_GAINS', 'ATTITUDE_PARAMETERS',
          'FOCAL_PLANE_CAL', 'B2B_ASSESSMENT', 'ANCILLARY_QA_THRESHOLDS',
          'TIRS_REL_GAINS_BLIND', 'TIRS_SATURATION_LEVEL',
          'TIRS_ABSOLUTE_GAINS_BLIND', 'OLI_DETECTOR_STATUS',
          'OLI_DETECTOR_NOISE', 'OLI_FRAMING_PARAMETERS', 'OLI_PARAMETERS',
          'TIRS_FOCAL_PLANE', 'I2I_ASSESSMENT'}


class TestCPFParser(unittest.TestCase):
    """ CPF testing Class"""

    @classmethod
    def setUpClass(cls):
        print("Setting up verification files")
        if call(["mkdir", "-p", "verify"]) != 0:
            raise Exception("Creating verify directory")

        if call(["./parse_verify_files.py"]) != 0:
            raise Exception("Creating verification files")
        print("Verification files were created")

    def test_cpf(self):
        """ Tests all of the functionality of the CPF class"""
        # Create the CPF object
        cpf_filename = (os.environ['IAS_UNIT_TEST_DATA_DIR']
                        + "/ias_lib/pylib/ias/io/cpf/L8CPF20130701_20130919.05")
        cpf = ias_cpf.CPF(filename = cpf_filename)

        # The CPF should not have parsed
        self.assertEqual(cpf.data, {})

        # Validate the total number of fields
        total_fields = 0

        for group in GROUPS:
            # Load the verification data object from the verify directory
            verify = imp.load_source(
                "{0}".format(group.lower()), "./verify/{0}.py".format(
                    group.lower()))
            verify_obj = getattr(verify, group)

            # Get the CPF object by calling the get method
            myobject = getattr(cpf, "get_{0}".format(group.lower()))()

            # Get all of the CPF attributes (only the actual variables)
            attr_list = [a for a in dir(verify_obj) if not
                         a.startswith('__') and not callable(
                            getattr(verify_obj, a))]

            # Validate each attribute
            for attr in attr_list:
                # Get the value of the current attribute from the verify object
                # as well as the read in CPF
                verify_attr = str(getattr(verify_obj, attr))
                read_attr = str(getattr(myobject, attr))

                # Compare the attributes as strings since they should be
                # identical
                self.assertEqual(verify_attr, read_attr, "Group {0} failed "
                                 "on attr {1}".format(group, attr))
                total_fields += 1

        print("Total Fields verified = {0}".format(total_fields))
        self.assertEqual(5081, total_fields)

if __name__ == '__main__':
    unittest.main()
