#!/usr/bin/env python3
""" Parameter file and OMF parsing module
"""
import re
from sys import exit
from ias.log import logger as lg
from ias.odl import odl
from ias.io.parameter_file import standard_parameters as std

# Global variables to mark OMF and ODL parameters
OMF = 1
PARAMETERS = 0

class Parameters():
    """ The Parameters class is returned with the requested parameter after
        the file is parsed """
    def __init():
        pass

class Parameter():
    """ Support class to contain all the needed information within a single
        class """

    def __init__(self, parm_name, variable_name, required=None, help=None,
                 default=None, lower_range=None, upper_range=None,
                 data_type=None, valid_options=None, is_list=None):
        """ Parameter Constructer

            Parameters:
                    Name: Parameter Name
                    Variable Name: Name of the variable that it will be stored
                    Required: Flag to indicate the parameter is required
                    Help: Help information string
                    Default: The default value for the parameter
                    Low Range: The lower range of the parameter
                    Upper Range: The upper range of the parameter
                    Datatype: The data type of the parameter
                    Valid Options: The valid options of the parameter (list)
                    Is List: Indicates whether the parameter value is a list

            Example:
                Parameter(parm_name="parm_name", variable_name="name",
                    required=True, help="Name of the parameter",
                    default="TestName", lower_range=None, upper_range=None,
                    data_type=str, valid_options=("TestName", "TestName2",
                    "Test3"))

        """

        # Initialize all the object variables
        self.parm_name = parm_name
        self.variable_name = variable_name
        self.default = default
        self.data_type = data_type
        self.is_list = False

        # Both the upper and lower range must be valid
        self.has_range = False
        self.lower_range = lower_range
        self.upper_range = upper_range

        if ((upper_range is None and lower_range is not None) or
                (upper_range is not None and lower_range is None)):
            raise ValueError("Both the lower and upper range must be present")

        if not (upper_range is None or lower_range is None):
            self.has_range = True

            # Check that the range is valid
            if lower_range >= upper_range:
                raise Exception(
                    "{0} has a invalid range {1} to {2}".format(
                        parm_name, lower_range, upper_range))

        # Set the optional parameters
        if required is None:
            self.required = False
        else:
            self.required = required
        if not help:
            self.help = ""
        else:
            self.help = help

        # Check if this should be converted to a list
        if is_list is not None:
            self.is_list = is_list

        # Set the valid options
        self.valid_options = valid_options
        self.has_options = False

        # Make sure the valid options is a list
        if (valid_options is not None):
            if (type(valid_options) is list):
                self.has_options = True
            else:
                raise Exception("Valid options must be a list of options")

        # Check if a range and valid options have both been specified
        if self.has_range and self.has_options:
            raise Exception("Both range and valid options were specified "
                            "only 1 can be used at a time")

        # Set the error flags to false
        self.found = False
        self.bad_range = False
        self.bad_option = False
        self.bad_type = False
        self.not_list = False


class ParameterParser():
    """ Class used to parse an odl file and initialize parameters"""

    def __init__(self):
        """ Default Constructor """
        self.parameters = list()
        self.omf_parms = list()

        # Store the ODL parser to be used for the OMF and parameter 
        # file parsing
        self.odl = odl.ODLParser()

    def provide_help(self, option, type):
        """ Generate the help or template output if requested

            Parameters:
                option: --help or --template or a filename
                type: PARAMETER or OMF indicator

            Returns:
                True if help was provided, False otherwise
        """

        # Make sure there aren't leading/trailing spaces
        option = option.strip()

        # Check if help was requested
        if "--help" == option or "-h" == option:
            self._help(type)
            return True

        # Check if a template was requested
        if "--template" == option or "-t" == option:
            self._template(type)
            return True

        # Help not provided
        return False


    def parse_parameters(self, filename, type=None, target=None):
        """ This method is called when a parmeter file needs to be parsed
            only parameters that have been specified will be stored and used

            Parameters:
                filename: File with the parameters to read
                type: normal parameter file/OMF switch
                target: optional target object to add parameters to.  If not
                        provided, a new object will be returned.

            Note: If "--help" or "-h" is passed as the filename then the help
                  will be output and the parser will exit the program
                  successfully. If "--template" or "-t" is passed as the
                  filename then a template will be output and the parser will
                  again exit successfully.
        """

        # Initialize the return parameter structure
        if not target:
            return_parms = Parameters()
        else:
            return_parms = target

        # Cleanup the filename as a precaution
        filename = filename.strip()

        # Read the ODL file
        self.odl.read_odl(filename)

        # Get the groups from the file.  Only one should be present.
        groups = self.odl.get_groups()
        if len(groups) > 1:
            raise FormatError("More than a single group in the file")
        group = groups[0]

        # Search through the dictionary for the parameters
        use_usage_flag = False
        for parm in self.parameters:
            # Get the parameter value
            value = self.odl.get_parameter_value(group, parm.parm_name)

            if value != None:
                # Convert to a list if requested
                if parm.is_list:
                    if not isinstance(value, list):
                        value = list(str(value))
                # But if it is a list and a list was not requested, it is an
                # error
                elif isinstance(value, list):
                    use_usage_flag = True
                    parm.not_list = True
                    parm.found = True
                    continue

                # Convert to the correct type if specified
                if parm.data_type:
                    try:
                        if isinstance(value, list):
                            value = [parm.data_type(i) for i in value]
                        else:
                            value = parm.data_type(value)
                    except ValueError:
                        use_usage_flag = True
                        parm.bad_type = True

                # Validate the value is within the correct range
                if parm.has_range:
                    if isinstance(value, list):
                        # Test each value in the a list
                        for test_value in value:
                            if (test_value < parm.lower_range or
                                test_value > parm.upper_range):
                                use_usage_flag = True
                                parm.bad_range = True
                                break
                    elif value < parm.lower_range or value > parm.upper_range:
                        use_usage_flag = True
                        parm.bad_range = True

                if parm.has_options:
                    if isinstance(value, list):
                        for test_value in value:
                            if test_value not in parm.valid_options:
                                use_usage_flag = True
                                parm.bad_option = True
                                break
                    elif value not in parm.valid_options:
                        use_usage_flag = True
                        parm.bad_option = True

                # Set as an attribute of the returned object
                setattr(return_parms, parm.variable_name, value)
                parm.found = True

            elif parm.default is not None:
                setattr(return_parms, parm.variable_name, parm.default)
                parm.found = True

            elif parm.required:
                use_usage_flag = True

        # Error while parsing, create the usage message and exit
        if use_usage_flag:
            self._usage(self.parameters, return_parms)

        # Return the parsed parameters
        return return_parms

    def add_parameter(self, parm_name, variable_name, required=None,
                      help=None, default=None, lower_range=None,
                      upper_range=None, data_type=None,
                      valid_options=None, is_list=None):
        """ Add ODL parameter to be added to the list of parsed parameters """

        # Cast options to a list if passed which could be a tuple depending
        # how the user passed the options
        if valid_options is not None:
            valid_options = list(valid_options)

        # Create the parameter
        parm = Parameter(parm_name, variable_name, required, help, default,
                         lower_range, upper_range, data_type, valid_options,
                         is_list)

        # Add to the parameters list
        self.parameters.append(parm)

    def add_standard_parameter(self, parm_name, variable_name, required=None,
                               help=None, default=None, lower_range=None,
                               upper_range=None, data_type=None,
                               valid_options=None, is_list=None):

        """ Add standard ODL parameter to the list of parsed parameters """

        # Cast options to a list if passed
        if valid_options is not None:
            valid_options = list(valid_options)

        # Find the standard parameter in the standard parameter dictionary
        # Let the user override any parameter
        if parm_name in std.standard_parms:
            parm = std.standard_parms[parm_name]
            if required is None:
                required = parm[std.REQUIRED]
            if help is None:
                help = parm[std.HELP]
            if default is None:
                default = parm[std.DEFAULT]
            if lower_range is None:
                lower_range = parm[std.LORANGE]
            if upper_range is None:
                upper_range = parm[std.UPRANGE]
            if data_type is None:
                data_type = parm[std.DATATYPE]
        else:
            raise ParameterNoStandardError(
                "Parameter {0} was not found in standard "
                "parameter list".format(parm_name))

        # Create the parameter
        parm = Parameter(parm_name, variable_name, required, help, default,
                         lower_range, upper_range, data_type, valid_options,
                         is_list)

        # Add to the parameters list
        self.parameters.append(parm)

    def _usage(self, requested_parms, read_parms):
        """ Outputs all errors with parameters """
        log = lg.Logger.get_default_logger()
        raiseException = None
        for parm in requested_parms:
            if not parm.found:
                log.error(
                    "KEYWORD identifier: {0} is missing a "
                    "VALUE".format(parm.parm_name))
                if not raiseException:
                    raiseException = ParameterRequiredError
            elif parm.bad_range:
                log.error(
                    "{0} value ({1}) does not fall in the correct range of "
                    "{2} to {3}".format(parm.parm_name, getattr(
                        read_parms, parm.variable_name), parm.lower_range,
                        parm.upper_range))
                if not raiseException:
                    raiseException = ParameterRangeError
            elif parm.bad_option:
                log.error(
                    "{0} value ({1}) is not in the list of valid options "
                    "{2}".format(parm.parm_name, getattr(
                        read_parms, parm.variable_name), parm.valid_options))
                if not raiseException:
                    raiseException = ParameterOptionError
            elif parm.bad_type:
                log.error(
                    "{0} value ({1}) does not have the correct type "
                    "{2}".format(parm.parm_name, getattr(read_parms,
                        parm.variable_name), str(parm.data_type)))
                if not raiseException:
                    raiseException = ParameterTypeError
            elif parm.not_list:
                log.error("{0} is a list when it should not be"
                          .format(parm.parm_name))
                if not raiseException:
                    raiseException = ParameterTypeError
        raise raiseException()

    def _help(self, parm_type):
        """ Provides help for the parameters provided """

        # Check which type of parameters help was requested for
        if parm_type == OMF:
            print("\n### OMF Parameters ###\n")
        else:
            print("\n### Application Input Parameters ###\n")

        # Loop over all of the parameters
        for parm in self.parameters:
            print("Name: {0}".format(parm.parm_name))
            print("    Description: {0}".format(parm.help))

            # Print if the parm is required
            if parm.required:
                print("    Required: Yes")
            else:
                print("    Required: No")

            # Print the parameter range
            if parm.has_range:
                print("    Range: {0} to {1}".format(parm.lower_range,
                      parm.upper_range))
            print("    Default = {0}".format(parm.default))

            # Print the parameter options
            if parm.has_options:
                print("    Legal Values: {0}".format(
                    ", ".join(parm.valid_options)))

    def _template(self, parm_type):
        """ Outputs a template for the list of request parameters """

        # Check which type of parameters help was requested for
        if parm_type == OMF:
            print("OBJECT = OMF")
        else:
            print("OBJECT = PARAMETERS")

        # Loop over all of the parameters printing out necessary information
        for parm in self.parameters:
            # Print out the help
            if parm.help:
                print("    /* {0} */".format(parm.help))

            # Required is always True or False
            if parm.required:
                print("    /*    Required: Yes */")
            else:
                print("    /*    Required: No */")

            # If the parameter has a range print it out
            if parm.has_range:
                print("    /*    Range: {0} to {1} */".format(parm.lower_range,
                      parm.upper_range))

            # Print the parameter options
            if parm.has_options:
                print("    /*    Legal Values: {0} */".format(
                    ", ".join(parm.valid_options)))

            # Leave the value blank if no default was specified
            if parm.default:
                print("    {0} = {1}".format(parm.parm_name, parm.default))
            else:
                print("    {0} =".format(parm.parm_name))

        # Print the ending lines
        if parm_type == OMF:
            print("END_OBJECT = OMF")
        else:
            print("END_OBJECT = PARAMETERS")
        print("END")


class FormatError(Exception):
    pass


class ParameterTypeError(Exception):
    pass


class ParameterRangeError(Exception):
    pass


class ParameterOptionError(Exception):
    pass


class ParameterRequiredError(Exception):
    pass


class ParameterNoStandardError(Exception):
    pass
