#!/usr/bin/env python3
""" Simple wrapper module for the python logging module

    Classes Exported:
        Logger - Wrapper class for logging module
"""
import logging
from logging import CRITICAL, ERROR, WARNING, INFO, DEBUG, NOTSET
import sys
import traceback
import os


def log_exceptions(function):
    """ Logger class exception decorator

        Description: This will wrap main functions in a try-except and log
        any exceptions that occur.

        Example:

        @log_exceptions
        def main():
    """
    # Using a closure create a new function that wraps the passed in function
    # in a try-except. When this is called all exceptions will be logged.
    def exception_logger(*args, **kwargs):
        try:
            # Return the passed in functions with the args
            return function(*args, **kwargs)
        except Exception as error:
            # Get the default logger
            logger = Logger.get_default_logger()

            # Change the format of the logger so this can manually unwind the
            # exception trace and provide the filename and line number
            format = logging.Formatter("%(asctime)19s  %(name)s %(process)7s "
                                       "%(local_file)-20s %(local_lineno)7s  "
                                       "%(levelname)s "
                                       "%(message)s", "%Y-%m-%d %H:%M:%S")
            if logger.console_handler:
                logger.console_handler.setFormatter(format)
            if logger.file_handler:
                logger.file_handler.setFormatter(format)

            # Get the traceback info, drop the first entry which is this
            # decorator that shouldn't be in the traceback
            exc_type, exc_value, exc_traceback = sys.exc_info()
            tb = traceback.extract_tb(exc_traceback)[1:]

            # Reverse the traceback so that the deepest part of the call tree
            # is printed first to match log output from C applications
            tb.reverse()

            # For each call in the traceback, issue a message
            first = True
            for call in tb:
                # Build the message to output
                filename = os.path.basename(call[0])
                line_number = call[1]
                func_name = call[2]
                if first:
                    message = func_name + " raised: " + str(exc_value)
                    first = False
                else:
                    message = "   ... called by: " + func_name

                # Output the error message
                logger.error(message, extra={'local_file': filename,
                             'local_lineno': line_number})
            sys.exit(1)

    # Return the closure function
    return exception_logger


class Logger(logging.Logger):
    """ Logger class that utilizes the base python Logging library by
        setting up the console and file handlers and formatting output
        to be standard.
    """

    default_instance = None
    instances = {}

    def __init__(self, logger_name=None, filename=None, use_stdout=None):
        """ Creates Logger object and adds console handler along with
            an optional file handler

            Arguments:
                logger_name: Name of the logger
                filename: Name of the file to log to
                use_stdout: Flag indicating output logging to standard
                            output.

            Note: The logger flushes the buffer after each log call.
        """
        self.logger_name = logger_name
        self.console_handler = None
        self.file_handler = None

        # Check if stdout is requested
        if use_stdout is None:
            use_stdout = True

        # Set the default log level
        ias_log_message_level = INFO

        # Check if the environment log level variable was set
        log_level = os.getenv("IAS_LOG_LEVEL")
        if log_level:
            if log_level == "DEBUG":
                ias_log_message_level = DEBUG
            elif log_level == "INFO":
                ias_log_message_level = INFO
            elif log_level == "WARN":
                ias_log_message_level = WARN
            elif log_level == "ERROR":
                ias_log_message_level = ERROR
            elif log_level == "DISABLE":
                ias_log_message_level = CRITICAL
            else:
                raise Exception("Invalid IAS_LOG_LEVEL environment variable "
                                "{0}".format(log_level))

        # Initialize object by using super
        super(Logger, self).__init__(logger_name)
        self.setLevel(ias_log_message_level)

        # Set the output format
        format = logging.Formatter("%(asctime)19s  %(name)s %(process)7s "
                                   "%(filename)-20s %(lineno)7s %(levelname)s  "
                                   "%(message)s", "%Y-%m-%d %H:%M:%S")

        # Add the console handler
        if use_stdout:
            console_handler = logging.StreamHandler(sys.stdout)
            console_handler.setLevel(NOTSET)
            console_handler.setFormatter(format)
            self.addHandler(console_handler)
            self.console_handler = console_handler

        # Add the file handler
        if filename:
            file_handler = logging.FileHandler(filename)
            file_handler.setLevel(NOTSET)
            file_handler.setFormatter(format)
            self.addHandler(file_handler)
            self.file_handler = file_handler

        # Add a default instance and add the logger to the list of loggers
        if not Logger.default_instance:
            Logger.default_instance = self
        if logger_name:
            Logger.instances[logger_name] = self

    @staticmethod
    def get_default_logger(logger_name=None):
        """ Returns the default logger or named logger """
        if logger_name and logger_name in Logger.instances:
            # Return named logger
            return Logger.instances[logger_name]
        if Logger.default_instance:
            # Return the default logger
            return Logger.default_instance
        else:
            # Create a new logger and return it
            return Logger(sys.argv[0])

    @staticmethod
    def set_default_logger(logger):
        """ Sets the default logger """
        Logger.default_instance = logger

    def close(self):
        """ Closes all the file handlers and cleans up the logger"""

        # Remove the console handler
        if self.console_handler:
            self.console_handler.close()
            self.removeHandler(self.console_handler)
            self.console_hander = None

        # Remove the file handler
        if self.file_handler:
            self.file_handler.close()
            self.removeHandler(self.file_handler)
            self.file_hander = None

        # Remove the instance
        if self.logger_name:
            del Logger.instances[self.logger_name]

        # Remove as the default logger (if necessary)
        if self == Logger.default_instance:
            Logger.default_instance = None
