#!/usr/bin/env python3

from enum import Enum

class BadBand(Exception):
    pass
class BadSatelliteNumber(Exception):
    pass

class SensorId(Enum):
    OLI = 0
    TIRS = 1

class SatelliteId(Enum):
    L8 = 0


class BandAttr():
    """ Class for per-band attributes """

    def __init__(self, number, sensor_id, name, scas, detectors_per_sca,
                 pixel_resolution, lines_per_frame = 1):
        self.number = number
        self.sensor_id = sensor_id
        self.name = name
        self.scas = scas
        self.detectors_per_sca = detectors_per_sca
        self.pixel_resolution = pixel_resolution
        self.lines_per_frame = lines_per_frame

class SatAttrBase():
    """ Base class for satellite attributes """

    def __init__(self, satellite_id, satellite_number, sensor_name):
        self.bands = []
        self.band_number_to_index = {}
        self.satellite_id = satellite_id
        self.satellite_number = satellite_number
        self.sensor_name = sensor_name

    def get_satellite_id(self):
        return self.satellite_id

    def get_satellite_name(self):
        if self.satellite_id == SatelliteId.L8:
            return "L8"
        return "UNKNOWN"

    def get_satellite_sensor_name(self):
        return self.sensor_name

    def convert_band_number_to_index(self, band_number):
        if band_number in self.band_number_to_index:
            return self.band_number_to_index[band_number]
        raise BadBand("Band number {0} not defined".format(band_number))

    def convert_band_index_to_number(self, band_index):
        if band_index < 0 or band_index >= len(self.bands):
            raise BadBand("Band index {0} not defined".format(band_index))
        return self.bands[band_index].number

    def get_band_count(self):
        return len(self.bands)

    def get_band_name(self, band_number):
        band_index = self.convert_band_number_to_index(band_number)
        return self.bands[band_index].name

    def get_scas_per_band(self, band_number):
        band_index = self.convert_band_number_to_index(band_number)
        return self.bands[band_index].scas

    def get_detectors_per_sca(self, band_number):
        band_index = self.convert_band_number_to_index(band_number)
        return self.bands[band_index].detectors_per_sca

    def get_pixel_resolution(self, band_number):
        band_index = self.convert_band_number_to_index(band_number)
        return self.bands[band_index].pixel_resolution

    def get_lines_per_frame(self, band_number):
        band_index = self.convert_band_number_to_index(band_number)
        return self.bands[band_index].lines_per_frame

    def get_sensor_id(self, band_number):
        band_index = self.convert_band_number_to_index(band_number)
        return self.bands[band_index].sensor_id

    def band_numbers(self, sensor_id = None):
        """ band number iterator with optional sensor id filter """
        for band in self.bands:
            if sensor_id == None or band.sensor_id == sensor_id:
                yield band.number

    def add_band(self, band):
        band.index = len(self.bands)
        self.bands.append(band)
        self.band_number_to_index[band.number] = band.index


class L8SatAttr(SatAttrBase):
    """ Landsat 8 satellite attributes """
    _OLI_SCAS = 14
    _OLI_DETECTORS_PER_MS_SCA = 494
    _OLI_DETECTORS_PER_PAN_SCA = 988
    _OLI_MS_PIXEL_RESOLUTION = 30
    _OLI_PAN_PIXEL_RESOLUTION = 15
    _TIRS_SCAS = 3
    _TIRS_DETECTORS_PER_SCA = 640
    _TIRS_PIXEL_RESOLUTION = 100

    def __init__(self, satellite_id):
        super(L8SatAttr, self).__init__(satellite_id, 8, "OLITIRS")
        self.add_band(BandAttr(1, SensorId.OLI, "OLI_Coastal_Aerosol",
                      self._OLI_SCAS, self._OLI_DETECTORS_PER_MS_SCA,
                      self._OLI_MS_PIXEL_RESOLUTION))
        self.add_band(BandAttr(2, SensorId.OLI, "OLI_Blue", self._OLI_SCAS,
                      self._OLI_DETECTORS_PER_MS_SCA,
                      self._OLI_MS_PIXEL_RESOLUTION))
        self.add_band(BandAttr(3, SensorId.OLI, "OLI_Green", self._OLI_SCAS,
                      self._OLI_DETECTORS_PER_MS_SCA,
                      self._OLI_MS_PIXEL_RESOLUTION))
        self.add_band(BandAttr(4, SensorId.OLI, "OLI_Red", self._OLI_SCAS,
                      self._OLI_DETECTORS_PER_MS_SCA,
                      self._OLI_MS_PIXEL_RESOLUTION))
        self.add_band(BandAttr(5, SensorId.OLI, "OLI_NIR", self._OLI_SCAS,
                      self._OLI_DETECTORS_PER_MS_SCA,
                      self._OLI_MS_PIXEL_RESOLUTION))
        self.add_band(BandAttr(6, SensorId.OLI, "OLI_SWIR1", self._OLI_SCAS,
                      self._OLI_DETECTORS_PER_MS_SCA,
                      self._OLI_MS_PIXEL_RESOLUTION))
        self.add_band(BandAttr(7, SensorId.OLI, "OLI_SWIR2", self._OLI_SCAS,
                      self._OLI_DETECTORS_PER_MS_SCA,
                      self._OLI_MS_PIXEL_RESOLUTION))
        self.add_band(BandAttr(8, SensorId.OLI, "OLI_PAN", self._OLI_SCAS,
                      self._OLI_DETECTORS_PER_PAN_SCA,
                      self._OLI_PAN_PIXEL_RESOLUTION, lines_per_frame = 2))
        self.add_band(BandAttr(9, SensorId.OLI, "OLI_CIRRUS", self._OLI_SCAS,
                      self._OLI_DETECTORS_PER_MS_SCA,
                      self._OLI_MS_PIXEL_RESOLUTION))
        self.add_band(BandAttr(10, SensorId.TIRS, "TIRS_THERMAL1",
                      self._TIRS_SCAS, self._TIRS_DETECTORS_PER_SCA,
                      self._TIRS_PIXEL_RESOLUTION))
        self.add_band(BandAttr(11, SensorId.TIRS, "TIRS_THERMAL2",
                      self._TIRS_SCAS, self._TIRS_DETECTORS_PER_SCA,
                      self._TIRS_PIXEL_RESOLUTION))


_cache = {}
_default_attr = None
_sat_number_to_id = {8: SatelliteId.L8}

def get_satellite_id_from_number(satellite_number):
    """ Convert the a satellite number to the matching satellite ID """
    return _sat_number_to_id[satellite_number]

def get(satellite_id):
    """ Get the satellite attributes for the requested satellite ID.  The
        attributes are cached so later calls for the same satellite ID can
        return the already created version. """
    global _cache
    global _default_attr
    if satellite_id in _cache:
        sat_attr = _cache[satellite_id]
    else:
        if satellite_id == SatelliteId.L8:
            sat_attr = L8SatAttr(satellite_id)
        else:
            raise BadSatelliteNumber("Unsupported satellite id: {0}"
                                     .format(satellite_id))
        _cache[satellite_id] = sat_attr
        if _default_attr is None:
            _default_attr = sat_attr
    return sat_attr

def set_default(satellite_id):
    """ Set the default satellite attributes """
    global _default_attr
    _default_attr = get(satellite_id)
    return _default_attr

def get_default():
    """ Get the default satellite attributes """
    return _default_attr

