#!/usr/bin/env python3
""" test_misc

    Unit test for band_spec.py and misc.py modules
"""
import unittest
from ias.log import logger as lg
import imp
import datetime as dt
import satellite_attributes as sat_attr
import const
import band_specs


class TestMisc(unittest.TestCase):

    def test_sat_attr(self):
        """ Tests all of the band specs for bands 1 to 11 """
        landsat_band_specs = band_specs.BandSpecs(8)
        specs = (
            [1, "CA", 443, 2, 433, 453, 15, 10, 10, 15, 40, 190, 555, 130,
             290],
            [2, "Blue", 482, 5, 450, 515, 25, 20, 20, 25, 40, 190, 581, 130,
             360],
            [3, "Green", 562, 5, 525, 600, 25, 20, 20, 25, 30, 194, 544, 100,
             390],
            [4, "Red", 655, 5, 630, 680, 25, 20, 15, 20, 22, 150, 462, 90,
             340],
            [5, "NIR", 865, 5, 845, 885, 25, 20, 15, 20, 14, 150, 281, 90,
             460],
            [6, "SWIR1", 1610, 10, 1560, 1660, 40, 30, 30, 40, 4, 32, 71.3,
             100, 540],
            [7, "SWIR2", 2200, 10, 2100, 2300, 50, 40, 40, 50, 1.7, 11, 24.3,
             100, 510],
            [8, "Pan", 590, 10, 500, 680, 50, 40, 40, 50, 23, 156, 515, 80,
             230],
            [9, "Cirrus", 1375, 5, 1360, 1390, 15, 10, 10, 15, 6, 999, 88.5,
             50, 50],
            [10, "TIRS1", 10800, 200, 10300, 11300, 400, 300, 300, 400, 9.60,
             12.69, 20.5, 0.40, 0.27],
            [11, "TIRS2", 12000, 200, 11500, 12500, 400, 300, 300, 400, 8.93,
             11.53, 17.8, 0.40, 0.29])

        i = -1
        for band in landsat_band_specs.band_specs:
            i += 1
            self.assertEqual(band.band_number, specs[i][0])
            self.assertEqual(band.band_color, specs[i][1])
            self.assertEqual(band.center_wavelength, specs[i][2])
            self.assertEqual(band.center_wavelength_tolerance, specs[i][3])
            self.assertEqual(band.lower_edge, specs[i][4])
            self.assertEqual(band.upper_edge, specs[i][5])
            self.assertEqual(band.lower_edge_slope_c1, specs[i][6])
            self.assertEqual(band.lower_edge_slope_c2, specs[i][7])
            self.assertEqual(band.upper_edge_slope_c1, specs[i][8])
            self.assertEqual(band.upper_edge_slope_c2, specs[i][9])
            self.assertEqual(band.ltypical, specs[i][10])
            self.assertEqual(band.lhigh, specs[i][11])
            self.assertEqual(band.lmax, specs[i][12])
            self.assertEqual(band.snr_at_ltyp, specs[i][13])
            self.assertEqual(band.snr_at_lhigh, specs[i][14])

    def test_constants(self):
        """ Verifys all of the constant values """

        # Test that bad satellite number do not work
        with self.assertRaises(LookupError):
            sat = sat_attr.set_default(sat_attr.get_satellite_id_from_number(9))

        sat = sat_attr.set_default(sat_attr.get_satellite_id_from_number(8))
        MAX_NBANDS = 11
        BAND_NUMBER = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]
        BAND_INDEX = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        NUM_OF_SCAS = [14, 14, 14, 14, 14, 14, 14, 14, 14, 3, 3]
        NUM_OF_DETECTORS = [494, 494, 494, 494, 494, 494, 494, 988, 494, 640,
                            640]
        BAND_NAMES = ['OLI_Coastal_Aerosol', 'OLI_Blue', 'OLI_Green',
                      'OLI_Red', 'OLI_NIR', 'OLI_SWIR1', 'OLI_SWIR2',
                      'OLI_PAN', 'OLI_CIRRUS', 'TIRS_THERMAL1',
                      'TIRS_THERMAL2']
        OLI = sat_attr.SensorId.OLI
        TIRS = sat_attr.SensorId.TIRS
        SENSOR_TYPE = [OLI, OLI, OLI, OLI, OLI, OLI, OLI, OLI, OLI, TIRS, TIRS]
        SUCCESS = 0
        ERROR = 1

        self.assertEqual(ERROR, const.ERROR)
        self.assertEqual(SUCCESS, const.SUCCESS)

        for band_number in sat.band_numbers():
            band_index = sat.convert_band_number_to_index(band_number)
            band_number = sat.convert_band_index_to_number(band_index)

            self.assertEqual(SENSOR_TYPE[band_index],
                             sat.get_sensor_id(band_number))
            self.assertEqual(BAND_NUMBER[band_index],
                             sat.convert_band_index_to_number(band_index))
            self.assertEqual(BAND_INDEX[band_index], band_index)
            self.assertEqual(NUM_OF_SCAS[band_index],
                             sat.get_scas_per_band(band_number))
            self.assertEqual(NUM_OF_DETECTORS[band_index],
                             sat.get_detectors_per_sca(band_number))
            self.assertEqual(BAND_NAMES[band_index],
                             sat.get_band_name(band_number))


if __name__ == '__main__':
    unittest.main()
