#!/usr/bin/env python3
""" Standard ODL reading library
"""
import re


def _boolify(s):
    """ Turns string True and False into actual representation """
    if s.lower() == 'true':
        return True
    if s.lower() == 'false':
        return False
    raise ValueError()


def _autoconvert(s):
    """ Converts the passed string value to the correct type """
    if type(s) is not str:
        return s
    try:
        try:
            if "." in s:
                return float(s)
            else:
                return int(s)
        except ValueError:
            pass
        return _boolify(s)
    except ValueError:
        pass

    if type(s) is str:
        return s.strip().replace("\"", "")

    return s


class ODLParser():
    """ Parses ODL from file and from memory """

    def __init__(self):
        """Default Constructor"""

        self.contents = {}

    def read_odl(self, filename, group=None):
        """ Reads the odl file into memory and calls the load method """

        # Read the file into memory
        lines = None
        with open(filename) as f:
            lines = f.readlines()
            lines = iter(lines)

        # Parse the odl file
        self.load(lines, group)

    def get_groups(self):
        """ Get the list of groups available after parsing """
        groups = list(self.contents.keys())
        groups.remove('IAS_ODL_ORDER')
        return groups

    def get_parameter_names(self, group_name):
        """ Get the list of parameters in a group """
        parameters = list(self.contents[group_name].keys())
        parameters.remove('IAS_ODL_ORDER')
        return parameters

    def get_parameter_value(self, group_name, parm_name):
        """ Get the value for a parameter """
        if parm_name in self.contents[group_name]:
            return self.contents[group_name][parm_name]['value']
        else:
            return None

    def load(self, lines, group_to_parse=None):
        """ Parses the ODL file from memory """
        self.contents = {}
        self.contents['IAS_ODL_ORDER'] = []
        # All the regex used to match odl lines
        list_match = re.compile("(?:\s*(\w+)\s*=\s*\((.+)\s*\Z)|(?:\s*(\w+)\s*"
                                "=\s*\(\s*\Z)")
        begin_match = re.compile("\s*(GROUP|OBJECT)\s*=\s*(\S+)\s*\Z")
        end_match = re.compile("\s*END_(GROUP|OBJECT)\s*=\s*(\S+)\s*\Z")
        value_match = re.compile("\s*(\w+)\s*=\s*(.+)\Z")
        comment_match = re.compile("\s*\/\*.*\*\/\s*\Z")

        group_name = "ROOT"
        middle_of_group = False
        group_found = False
        file_ended = False

        # Loop over all of the lines provided
        for line in lines:
            line = line.strip()

            # Ignore blank lines
            if line == "":
                continue

            # If the file ended and there are still non-empty lines being read,
            # it is an error
            if file_ended:
                raise Exception("Data found past the end of the END statement")

            # If the line has just "END", the end of the file has been reached
            if line == "END":
                file_ended = True
                continue

            # Ignore comment lines
            if comment_match.match(line):
                continue

            # If the line ends with a '=' it needs to be merged with the
            # next line.  If the next line is empty, this will not work, but
            # that shouldn't happen.
            if line.endswith("="):
                line = line + next(lines).strip()

            # Match the beginning group
            if "GROUP" in line or "OBJECT" in line:
                is_match = begin_match.match(line)
                if is_match:
                    if middle_of_group:
                        raise Exception("Group {0} began without group "
                                         "{1} ending".format(
                                            is_match.group(2),
                                            group_name))
                    group_name = is_match.group(2)
                    middle_of_group = True
                    # Check if a single group needs parsed only parse
                    # that group (ignoring all others)
                    if group_to_parse and group_name == group_to_parse:
                        group_found = True

                    if group_name in self.contents:
                        raise Exception("Group {0} appears more than once"
                                        .format(group_name))

                    self.contents['IAS_ODL_ORDER'].append(group_name)
                    self.contents[group_name] = {}
                    self.contents[group_name]['IAS_ODL_ORDER'] = []

                    continue

                # Match the ending group
                is_match = end_match.match(line)
                if is_match:
                    if (group_name != is_match.group(2)):
                        raise Exception("Mismatched groups", group_name,
                                        is_match.group(2))
                    middle_of_group = False
                    # Check if a single group needs parsed exit if end
                    # group found
                    if group_to_parse and group_name == group_to_parse:
                        return
                    group_name = "ROOT"
                    continue

            # If a single group needs parsed ignore all other groups
            if group_to_parse and not group_found:
                continue

            # If not in a group by this point, it is an error
            if not middle_of_group:
                raise Exception("Error parsing file: Not in group for line '"
                                + line + "'")

            # Check whether the parameter matches a list or non-list parameter
            is_list = True
            parm_match = list_match.match(line)
            if not parm_match:
                parm_match = value_match.match(line)
                is_list = False

            if parm_match:
                # Get the parameter name and value from the match
                parm_name = parm_match.group(1)
                parm_value = parm_match.group(2)

                # It is illegal to have the same parameter to show up in the
                # group more than once
                if parm_name in self.contents[group_name]:
                    raise Exception("Parameter {0} appears more than once"
                                    " in group {1}".format(parm_name,
                                    group_name))

                self.contents[group_name][parm_name] = {}
                self.contents[group_name]['IAS_ODL_ORDER'].append(parm_name)

                if is_list:
                    # Handle list parsing
                    self.contents[group_name][parm_name]['type'] = "LIST"

                    # Check for an empty match
                    if not parm_value:
                        parm_value = ''

                    # Continue reading if values on the next line
                    while (not re.search(".+\)\s*\Z", line)):
                        line = next(lines)
                        parm_value = parm_value + " " + line.strip()

                    parm_value = parm_value.replace(')', "")
                    if "\"" in parm_value and "," not in parm_value:
                        self.contents[group_name][parm_name]['type'] = (
                            "QUOTED_LIST")

                    # This code will not handle a list of strings with quotes
                    # inside the strings below is a start for the fix
                    # parm_value = re.compile(r'''\s*(".+"),\s*''').split(
                    #      parm_value)
                    parm_value = parm_value.split(',')
                    parm_value[:] = [_autoconvert(x) for x in parm_value]

                    self.contents[group_name][parm_name]['value'] = parm_value
                    continue

                else:
                    # Handle normal parsing
                    self.contents[group_name][parm_name]['type'] = 'VALUE'

                    # Check if this is a quoted string
                    if (re.search("\"(.+)", parm_value)):
                        self.contents[group_name][parm_name]['type'] = 'QUOTES'
                        vmatch = re.match("\"(.*)\"", parm_value)
                        if (vmatch):
                            parm_value = vmatch.group(1)
                        else:
                            combined_line = parm_value

                            while (not re.search("\"(.+)\"", combined_line)):
                                line = next(lines)
                                line = line.strip()
                                combined_line = combined_line + " " + line
                            clmatch = re.search("\"(.+)\"", combined_line)
                            parm_value = ' '.join(clmatch.group(1).split())

                    self.contents[group_name][parm_name]['value'] = (
                        _autoconvert(parm_value))
                    continue

            # Unrecognized line, report an error
            raise Exception("Unable to parse the line that reads: '"
                            + line + "'")

        # If the file didn't end properly it is an error
        if not file_ended:
            raise Exception("File not properly terminated with 'END'")
