#ifndef IAS_RADIOMETRY_H
#define IAS_RADIOMETRY_H
/*---------------------------------------------------------------------------
NAME:    ias_radiometry.h

PURPOSE: Implements the shared radiometry library routines.
-----------------------------------------------------------------------------*/
#include <stdint.h>

#include "ias_cpf.h"
#include "ias_db.h"
#include "ias_l0r.h"
#include "ias_miscellaneous.h"
#include "ias_pixel_mask.h"
#include "ias_satellite_attributes.h"
#include "ias_los_model.h"
#include "ias_rps_db_bias_model_calibration.h"
#include "ias_rps_bias_model_calibration.h"
#include "ias_rps_structures.h"
#include "ias_rps_histogram_statistics_characterization.h"
#include "ias_rps_impulse_noise_characterization.h"
#include "ias_rps_saturated_pixel_characterization.h"
#include "ias_rps_dropped_frame_characterization.h"
#include "ias_rps_frame_averages.h"
#include "ias_rps_detector_status_identification.h"
#include "ias_rps_3q_response_linearization.h"
#include "ias_rps_lookup_response_linearization.h"
#include "ias_rps_tirs_dark_response_determination.h"
#include "ias_rps_identify_diffuser_panel.h"
#include "ias_rps_calculate_sca_offsets.h"

/* NOTE: The pixel mask value in the library is an unsigned char.  The
         following definitions use up all those bits therefore no additions
         can be made without a corresponding change to the pixel mask
         library. */
#define IAS_RPS_NOVALUE             0x00
#define IAS_RPS_DROPLINE            0x01
#define IAS_RPS_OUTOFSPEC_DETECTOR  0x02
#define IAS_RPS_IMPNOISE            0x04
#define IAS_RPS_ADCONV_SAT_LO       0x08
#define IAS_RPS_ADCONV_SAT_HI       0x10
#define IAS_RPS_ANALOG_SAT_LO       0x20
#define IAS_RPS_ANALOG_SAT_HI       0x40
#define IAS_RPS_INOP_DETECTOR       0x80

/* Diffuser panel identifier strings */
#define IAS_RPS_OLI_DIFFUSER_WORKING  "WORKING"
#define IAS_RPS_OLI_DIFFUSER_PRISTINE "PRISTINE"

/* Various OLI/TIRS collection type strings */
#define IAS_RPS_OLI_SOLAR_COLLECT           "OLI_SOLAR"
#define IAS_RPS_OLI_SOLAR_TIMESWEEP_COLLECT "OLI_SOLAR_INTEGRATION_TIME_SWEEP"
#define IAS_RPS_OLI_SHUTTER_COLLECT         "OLI_SHUTTER"
#define IAS_RPS_TIRS_BLACKBODY_COLLECT      "TIRS_BLACKBODY"
#define IAS_RPS_TIRS_DEEPSPACE_COLLECT      "TIRS_DEEPSPACE"
#define IAS_RPS_EARTH_IMAGING_COLLECT       "EARTH_IMAGING"

/* RPS Level process ids */
#define IAS_RPS_LEVEL_CONVERT_FLOAT_NOSHIFT '0'
#define IAS_RPS_LEVEL_CONVERT_FLOAT_CORRECT 'C'
#define IAS_RPS_LEVEL_BIAS_REMOVAL          'B'
#define IAS_RPS_LEVEL_RESPONSE_LINEAR       'L'
#define IAS_RPS_LEVEL_LOOKUP_RESPONSE_LINEAR 'A'
#define IAS_RPS_LEVEL_TEMP_SENSITIVITY_BIAS 'T'
#define IAS_RPS_LEVEL_TEMP_SENSITIVITY_GAIN 'K'
#define IAS_RPS_LEVEL_GAIN_APPLICATION      'G'
#define IAS_RPS_LEVEL_SCA_DISCONTINUITY     'D'
#define IAS_RPS_LEVEL_RESIDUAL_STRIPING     'S'
#define IAS_RPS_LEVEL_SATURATED_PIXEL       'P'
#define IAS_RPS_LEVEL_INOPERABLE_DETECTORS  'I'
#define IAS_RPS_LEVEL_REFLECTANCE           'R'
#define IAS_RPS_LEVEL_SKIP                  '.'
#define IAS_RPS_LEVEL_TIRS_2ND_LINEARIZATION '2'
#define IAS_RPS_LEVEL_GAIN_OFFSET_REMOVAL   'O'

/* RPS band linesets */
#define IAS_RPS_ODD_PAN_LINESET      0
#define IAS_RPS_EVEN_PAN_LINESET     1
#define IAS_RPS_NORMAL_BAND_LINESET  2

/* Enumeration defining the diffuser panel types */
typedef enum ias_rps_oli_diffuser_type
{
    IAS_RPS_OLI_DIFFUSER_TYPE_WORKING,
    IAS_RPS_OLI_DIFFUSER_TYPE_PRISTINE,
    IAS_RPS_OLI_DIFFUSER_TYPE_NOTDEPLOYED
} IAS_RPS_OLI_DIFFUSER_TYPE;



/* Prototypes for routines that don't have their own header file */
int ias_rps_calculate_filtered_mean
(
    const double *data_buffer,     /* I: Data buffer */
    const int *invalid_sample_flags,
                                   /* I: Data validity flag for all
                                      samples (0=valid sample,
                                      nonzero=invalid sample) */
    int num_samples,               /* I: Total number of samples*/
    double *valid_mean,            /* O: Valid sample mean */
    int *valid_count               /* O: Number of "valid" data samples used
                                      in the calculations (pass in NULL if
                                      this output is not desired) */
);

int ias_rps_calculate_filtered_stdev
(
    const double *data_buffer,     /* I: Data buffer */
    const int *invalid_sample_flags,
                                   /* I: Data validity flag for all
                                      samples (0=valid sample,
                                      nonzero=invalid sample) */
    int num_samples,               /* I: Total number of samples */
    double valid_mean,             /* I: Valid sample mean */
    double *stdev                  /* O: Sample standard deviation */
);

int ias_rps_calculate_filtered_rms
(
    const double *data_buffer,    /* I: Data buffer */
    const int *invalid_sample_flags,
                                  /* I: Data validity flag for all
                                     samples (0=valid sample,
                                     nonzero=invalid sample) */
    int num_samples,              /* I: Total number of samples*/
    double valid_mean,            /* I: Sample mean */
    double *valid_rms,            /* O: Sample RMS */
    int *valid_count              /* O: Number of "valid" data samples used
                                     in the calculations (pass in NULL if
                                     this output is not desired) */
);


void ias_rps_calculate_radiance
(
    const double *data_buffer,      /* I: Data buffer to search */
    int num_dets_per_sca,           /* I: Number of detectors in an SCA */
    double abs_gain,                /* I: Absolute band gain */
    const double *rg_data_ptr,      /* I: Pointer to relative gain data
                                       buffer */
    double *radiance_buffer         /* O: Detector radiances */
);

char *ias_rps_get_report_file_name
(
    const IAS_REPORT *report_info,  /* I: Rreport info structure */ 
    int band_number,                /* I: Band number */
    int sca_number,                 /* I: SCA number */
    const char *file_base_name      /* I: Base file name */
);

void ias_rps_convert_detector_delays2spatial_offsets
(
    const double *detector_delays,  /* I: Detector delays from CPF */
    int number_detectors,           /* I: Number of values in the arrays */
    int *spacial_offsets,           /* O: Converted offset values */
    int *largest_offset             /* O: Largest offset, if NULL is passed
                                          the value isn't returned */
);

int ias_rps_get_invalid_data_flags
(
    IAS_PIXEL_MASK *current_pm,          /* I: Pixel mask for the current
                                            band/SCA */
    PIXEL_MASK_TYPE artifact_flags_to_get,
                                         /* I: Artifact flag(s) to get.
                                            Multiple artifacts can be
                                            specified by using the
                                            bitwise OR operator (e.g
                                            IAS_RPS_INOP_DETECTOR |
                                            IAS_RPS_OUTOFSPEC_DETECTOR).
                                            By using ~IAS_RPS_NO_VALUE,
                                            all currently known artifact
                                            flags can be retrieved. */
    int buffer_size,                     /* I: Size of the flag buffer which
                                            is number of detectors */
    int *flag_buffer                     /* O: Populated flag buffer, all
                                            detectors for the current band/SCA
                                            are set -- previous contents
                                            overwritten */
);

#endif
