/*****************************************************************************
 NAME:    ias_rps_3q_response_linearization

 PURPOSE: Adjusts the DN values for the OLI/TIRS image bands to compensate
          for non-linearity in the detector response.

 RETURNS: Nothing

 REFERENCES:  Additional detailed information describing the algorithm can be
              found in the OLI and/or TIRS Response Linearization ADDs
*****************************************************************************/
#include <stdlib.h>                /* For NULL definition */
#include "ias_logging.h"
#include "ias_radiometry.h"
#include "ias_rlut.h"


int ias_rps_3q_response_linearization
(
    float *data_buffer,            /* I/O: On input it is the DN from sensor
                                      On output it is the adjusted DN
                                      with non-linearity compensation */
    const IAS_RLUT_LINEARIZATION_PARAMS *params,
                                   /* I: Array of linearization parameters
                                      for all detectors in the current
                                      band/SCA */
    int number_detectors,          /* I: Number of detectors for this SCA */
    int number_lines               /* I: Number of lines for this SCA */
)
{
    const IAS_RLUT_LINEARIZATION_PARAMS *param_ptr = NULL;
                                   /* Local pointer to input linearization
                                      parameter buffer */
    int det_index;                 /* Detector loop index */
    int line;                      /* Line loop index */

    float *data_value;             /* This is used to point to the address
                                      of the data value being modified */
    float value;                   /* Current image pixel value */

    double low_coeff0;             /*---------------------------------------*/
    double low_coeff1;             /*                                       */
    double low_coeff2;             /*                                       */
    double mid_coeff0;             /*   These single value variables are    */
    double mid_coeff1;             /*   used for convenience and for a      */
    double mid_coeff2;             /*   small increase in efficiency since  */
    double high_coeff0;            /*   they will be accessed many times    */
    double high_coeff1;            /*   through the for detector loop       */
    double high_coeff2;            /*                                       */
    float low_threshold;           /*                                       */
    float high_threshold;          /*---------------------------------------*/


    /* Sanity check */
    if ((data_buffer == NULL) || (params == NULL))
    {
        IAS_LOG_ERROR("Missing input image data buffer and/or "
            "linearization parameter data buffer");
        return ERROR;
    }

    /* Loop through the detectors.  An outer detector loop is used here
       since the same thresholds and remapping coefficients are used for
       all the detector samples. */
    for (det_index = 0; det_index < number_detectors; det_index++)
    {
        param_ptr = &params[det_index];
 
        /* Get local copies of the coefficients and thresholds */
        low_coeff0 = param_ptr->remap_coeff0_low;
        low_coeff1 = param_ptr->remap_coeff1_low;
        low_coeff2 = param_ptr->remap_coeff2_low;

        mid_coeff0 = param_ptr->remap_coeff0_mid;
        mid_coeff1 = param_ptr->remap_coeff1_mid;
        mid_coeff2 = param_ptr->remap_coeff2_mid;

        high_coeff0 = param_ptr->remap_coeff0_high;
        high_coeff1 = param_ptr->remap_coeff1_high;
        high_coeff2 = param_ptr->remap_coeff2_high;

        low_threshold = param_ptr->cutoff_threshold_low;
        high_threshold = param_ptr->cutoff_threshold_high;

        /* Loop through the lines */
        data_value = &data_buffer[det_index];
        for (line = 0; line < number_lines; line++)
        {
            value = *data_value;

            /* Determine the remapping parameters and apply the correct
               set */
            if (value < low_threshold)
            {
                *data_value = low_coeff0 + (value
                    * (low_coeff1 + value * low_coeff2));
            }
            else if (value < high_threshold)
            {
                *data_value = mid_coeff0 + (value
                    * (mid_coeff1 + value * mid_coeff2));
            }
            else
            {
                *data_value = high_coeff0 + (value
                    * (high_coeff1 + value * high_coeff2));
            }

            /* Point to the next data value in line */
            data_value += number_detectors;

        }  /* END for line loop */
    }  /* END for detector loop */

    return SUCCESS;
}  /* END ias_rps_3q_response_linearization */
