#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "gsl/gsl_fit.h"
#include "ias_bpf.h"
#include "ias_bpf_database.h"
#include "ias_cpf.h"
#include "ias_db.h"
#include "ias_logging.h"
#include "ias_pixel_mask.h"
#include "ias_radiometry.h"
#include "ias_rps_bias_model_calibration.h"
#include "ias_rps_db_bias_model_calibration.h"
#include "ias_rps_histogram_statistics_characterization.h"
#include "ias_const.h"

/*------------------------------------------------------------------------------
   Forward declaration of private routines in this file
------------------------------------------------------------------------------*/
static int calculate_frame_variation_coefficients(
    const char *post_id,
    const IAS_BMC_WORKORDERS *pre_workorders,
    const int previous_workorders_n0,
    const IAS_BAND_ATTRIBUTES *band_attributes,
    const IAS_BAND_ATTRIBUTES *vrp_attributes,
    const int sca_number, 
    const int fill_value,
    struct ias_db_connection *db_connection,
    IAS_BPF *bpf,
    IAS_RPS_DB_BAND_RESIDUAL *band_residuals);

static int calculate_dark_response_coefficients(
    IAS_CPF *cpf, 
    const char *post_id,
    const IAS_BMC_WORKORDERS *pre_workorders,
    const int previous_workorders_n1,
    IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS **band_characterization,
    IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS **vrp_characterization,
    const IAS_BAND_ATTRIBUTES *band_attributes,
    const IAS_BAND_ATTRIBUTES *vrp_attributes,
    const int sca_number, 
    struct ias_db_connection *db_connection,
    IAS_BPF *bpf,
    IAS_RPS_DB_BAND_RESIDUAL *band_residuals);

static int calculate_mean(
    const char *post_workorder,
    const IAS_BMC_WORKORDERS *pre_workorders,
    const IAS_BAND_ATTRIBUTES *band_attributes,
    const int sca_number, 
    IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS **band_characterization,
    const int previous_workorder_count,
    struct ias_db_connection *db_connection,
    IAS_BPF *bpf);

static void free_characterization(
    const int previous_workorder_count, 
    IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS ***band_characterization);

static int generate_file_info(
    const char *mission_id,
    const char *post_workorder,
    const IAS_BMC_WORKORDERS *pre_workorders,
    struct ias_db_connection *db_connection,
    IAS_BPF *bpf);

static int load_characterization(
    const char *post_workorder,
    const IAS_BMC_WORKORDERS *pre_workorders,
    struct ias_db_connection *db_connection,
    const IAS_BAND_ATTRIBUTES *band_attributes,
    const int sca_number,
    IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS ***band_characterization,
    IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS ***vrp_characterization,
    const char *collection_type);


/*------------------------------------------------------------------------------
                            Public Routines 
------------------------------------------------------------------------------*/
/*-----------------------------------------------------------------------------
NAME: ias_rps_bias_model_calibration

PURPOSE: Calculate bias parameters and store them into the database.

NOTE: Satellite attributes must be initialized before calling this routine.

RETURNS: An integer indicating SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int ias_rps_bias_model_calibration(
    /*I: CPF used in processing */
    IAS_CPF* cpf,
    /*I: Database connection*/
    struct ias_db_connection *db_connection,
    /*I: Interval ID of the mission data*/
    const char *mission_id,
    /*I: Workorder for the last shutter collect*/
    const char *post_workorder, 
    /*I: The number of prior collections to use for pre/post and A0 */
    const int prev_collections_requested_n0,
    /*I: The number of prior collections to use for C1 and A1 */
    const int prev_collections_requested_n1,
    /*I: Image fill value used to detect filled frames*/
    const int fill_value,
    /*I/O: The BPF structure containing Bias Model Parameters, previously 
        allocated */
    IAS_BPF *bpf,
    /*O: The number of previous collections actually used in processing*/
    int *collections_used,
    /*O: The residuals of the linear fit.  If NULL is passed, the default
         Ingest case, the residuals are stored in the database.  If a
         pre-allocated structure is passed, the IAS case, the resuduals are
         copied into the structure and the residuals won't be stored in the
         database. */
    IAS_RPS_DB_BIAS_MODEL_RESIDUALS *output_residuals)
{
    const IAS_BAND_ATTRIBUTES *band_attributes = NULL;
    const IAS_BAND_ATTRIBUTES *vrp_attributes = NULL;
    const IAS_SATELLITE_ATTRIBUTES *sat_attributes = NULL;

    int band_number = 0; /* band loop counter */
    int max_collections = 0;
    int sca_index = 0; /* sca loop counter */
    int status = SUCCESS; /* Status if an error has occurred */

    IAS_BMC_WORKORDERS *previous_workorders = NULL;
    IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS **band_characterization =NULL;
    IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS **vrp_characterization = NULL;
    IAS_RPS_DB_BAND_RESIDUAL *band_residuals = NULL;

    if (collections_used == NULL)
    {
        IAS_LOG_ERROR("NULL pointer provided to the integer for previous "
            "collections used");
        return ERROR;
    }

    *collections_used = 0; /*Initialize collections used to 0*/

    /* Check the parameters for the number of previous collections to use in 
        generating the Bias Model Parameters */
    if (prev_collections_requested_n0 < 1 ||
        prev_collections_requested_n1 < 1)
    {
        IAS_LOG_ERROR("At least 1 previous collection must be specified");
        return ERROR;
    }
   
    if (prev_collections_requested_n0 < prev_collections_requested_n1)
    {
        max_collections = prev_collections_requested_n1;
    } 
    else
    {
        max_collections = prev_collections_requested_n0; 
    } 

    if (bpf == NULL)
    {
        IAS_LOG_ERROR("BPF structure provided is NULL");
        return ERROR;
    }

    sat_attributes = ias_sat_attr_get_attributes();

    if (sat_attributes == NULL)
    {
        IAS_LOG_ERROR("Unable to obtain sat attributes");
        return ERROR;
    }

    IAS_LOG_DEBUG("Obtaining previous workorders for %s", post_workorder);

    /* Obtain the workorders for the previous collections, the order of the
    workorders is newest first to allow for both the N0 and N1 lists */
    previous_workorders = ias_rps_db_get_bias_model_previous_workorders(
        cpf, db_connection, post_workorder, max_collections);

    if (previous_workorders == NULL)
    {
        IAS_LOG_ERROR("Unable to obtain work orders for previous "
            "collections");
        return ERROR;
    }

    if (previous_workorders->workorder_count == 0)
    {
        ias_rps_db_free_bias_model_workorders(&previous_workorders);
        IAS_LOG_INFO("No collections available prior to RWOC ID %s",
            post_workorder);
        return SUCCESS;
    }

    IAS_LOG_DEBUG("%d previous collections have been loaded",
        previous_workorders->workorder_count);
    *collections_used = previous_workorders->workorder_count;

    /* Get the correct sensor name into the BPF structure. */
    ias_bpf_set_file_attributes_sensor_name(bpf,
        ias_sat_attr_get_sensor_name(IAS_OLI));

    /* Loop through all bands, create bias parameters for applicable OLI bands*/
    for (band_number = 1;
        status == SUCCESS && band_number <= sat_attributes->total_bands;
        band_number++)
    {
        band_attributes = ias_sat_attr_get_band_attributes(band_number);

        if (band_attributes == NULL)
        {
            IAS_LOG_ERROR("Unable to get attributes for band %d", band_number);
            status = ERROR;
            break;
        }

        /* Bias parameters are calculated for the Normal OLI bands, skip
        the others */
        if ((band_attributes->band_classification & IAS_NORMAL_BAND) == 0 ||
            (band_attributes->sensor_id != IAS_OLI))
        {
            IAS_LOG_DEBUG("Skipping band number %d",
                band_attributes->band_number);
            continue;
        }
    
        /* If this is called with a residuals output structure, it is being
           run stand alone.  Output a processing message. */
        if ( output_residuals )
        {
            IAS_LOG_INFO("Processing band %d", band_number );
        }

        vrp_attributes = ias_sat_attr_get_band_attributes(
            band_attributes->vrp_band_number);
      
        if (vrp_attributes == NULL)
        {
            IAS_LOG_ERROR("Unable to get attributes for band %d VRPs", 
                band_number);
            status = ERROR;
            break;
        }
   
        /*Set the band number*/
        IAS_LOG_DEBUG("calling ias_bpf_set_bias_model_band_number to set "
            "band number to %d", band_number);
        if (ias_bpf_set_bias_model_band_number(bpf, 
            band_attributes->band_number) != SUCCESS)
        {
            IAS_LOG_ERROR("Unable to store the band number in the "
                "IAS BPF");
            status = ERROR;
            break;
        }
    
        /* Set the spectral type */
        if (ias_bpf_set_bias_model_spectral_type(bpf, 
            band_attributes->band_number, 
            band_attributes->spectral_type) != SUCCESS)
        {
            IAS_LOG_ERROR("Unable to store the spectral type in the "
                "IAS BPF");
            status = ERROR;
            break;
        }

        /* Set the band residuals if the data has requested to be returned */
        if ( output_residuals )
        {
            band_residuals = &output_residuals->band_residuals
                [band_attributes->band_index];
        }
    
        for (sca_index = 0; status == SUCCESS && 
            sca_index < band_attributes->scas; 
            sca_index++)
        {
            int sca_number = sca_index + 1;  
    
            if (SUCCESS != load_characterization(post_workorder, 
                    previous_workorders, db_connection, band_attributes, 
                    sca_number, &band_characterization, &vrp_characterization,
                    IAS_RPS_OLI_SHUTTER_COLLECT))
            {
                IAS_LOG_ERROR("Unable to load characterization for band %d", 
                    band_attributes->band_number);
                status = ERROR;
                break;
            }

            IAS_LOG_DEBUG("Calculating pre/post means");

            /* Calculate the pre and post-collection group means */
            if (SUCCESS != calculate_mean(post_workorder, 
                    previous_workorders, band_attributes, sca_number, 
                    band_characterization, prev_collections_requested_n0, 
                    db_connection, bpf))
            {
                IAS_LOG_ERROR("Unable to calculate mean for band %d", 
                    band_attributes->band_number);
                status = ERROR;
                break;
            }

            if (SUCCESS != calculate_frame_variation_coefficients(
                    post_workorder,
                    previous_workorders, prev_collections_requested_n0, 
                    band_attributes, vrp_attributes, 
                    sca_number, fill_value, db_connection, bpf,
                    band_residuals))
            {
                IAS_LOG_ERROR("Unable to calculate a0 for band %d", 
                    band_attributes->band_number);
                status = ERROR;
                break;
            }

            if (SUCCESS != calculate_dark_response_coefficients(cpf,
                    post_workorder, previous_workorders, 
                    prev_collections_requested_n1,
                    &band_characterization[1], /* skip over index 0, the post*/ 
                    &vrp_characterization[1], /*skip over index 0, the post*/
                    band_attributes, vrp_attributes, sca_number, 
                    db_connection, bpf, band_residuals))
            {
                IAS_LOG_ERROR("Unable to calculate a1 for band %d", 
                    band_attributes->band_number);
                status = ERROR;
                break;
            }
        }/* End of SCA loop */
    }/* End of band loop */

    IAS_LOG_DEBUG("Populating file information");

    /* Populate misc information in the BPF */
    if (status == SUCCESS && SUCCESS != generate_file_info(mission_id, 
        post_workorder, previous_workorders, db_connection, bpf))
    {
        IAS_LOG_ERROR("Unable to set file info in the BPF");
        status = ERROR;
    } 

    IAS_LOG_DEBUG("Freeing resources");

    /* reusable function to release characterization from load*/
    free_characterization(previous_workorders->workorder_count,
        &band_characterization);
    free_characterization(previous_workorders->workorder_count, 
        &vrp_characterization);

    /* Cleanup the array of previous RWOC IDs */
    IAS_LOG_DEBUG("Clean up previous workorder IDs");
    ias_rps_db_free_bias_model_workorders(&previous_workorders);

    IAS_LOG_DEBUG("Bias Model Calibration complete with status %d", status);   

    return status;
}

int ias_rps_bias_model_calibration_tirs(
    /*I: CPF populated structure*/
    IAS_CPF *cpf,
    /*I: Established database connection*/
    struct ias_db_connection *db_connection,
    /*I: Mission ID of the post collection being processed*/
    const char *mission_id,
    /*I: Radiometric Work Order ID of the post collections being processed*/
    const char *post_workorder,
    /*I: The number of previous collections to be used with the post*/
    const int previous_collections,
    /*I: TBD is this needed since only detector averages are used?*/
    const int fill_value,
    /*O: BPF structure populated with the TIRS bias model parameters*/
    struct IAS_BPF *bpf,
    /*O: The actual number of collections used to create the bias parameters*/
    int *collections_used)
{ 
    int band_number = 0; /* band number iterator for the bands processed */
    int sca_index = 0; /*sca index iterator, process 1 sca at a time */
    int status = SUCCESS; /* Status if an error has occurred */

    const IAS_BAND_ATTRIBUTES *band_attributes = NULL;
    const IAS_SATELLITE_ATTRIBUTES *sat_attributes = NULL;

    IAS_BMC_WORKORDERS *previous_workorders = NULL;
    IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS **band_characterization =NULL;

    /*VRP isn't used with TIRS but it is still used in the interface to a reused
        function*/
    IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS **vrp_characterization =NULL;

    IAS_LOG_DEBUG("Starting to create TIRS bias model parameters");

    sat_attributes = ias_sat_attr_get_attributes();

    if (sat_attributes == NULL)
    {
        IAS_LOG_ERROR("Unable to obtain sat attributes");
        return ERROR;
    }

    IAS_LOG_DEBUG("Obtaining previous workorders for %s", post_workorder);

    /* Obtain the workorders for the previous collections */
    previous_workorders = ias_rps_db_get_bias_model_previous_workorders_tirs(
        cpf, db_connection, post_workorder, previous_collections);

    if (previous_workorders == NULL)
    {
        IAS_LOG_ERROR("Unable to obtain work orders for previous "
            "collections");
        return ERROR;
    }

    if (previous_workorders->workorder_count == 0)
    {
        ias_rps_db_free_bias_model_workorders(&previous_workorders);
        IAS_LOG_INFO("No collections available prior to RWOC ID %s",
            post_workorder);
        return SUCCESS;
    }

    IAS_LOG_DEBUG("%d previous collections have been loaded",
        previous_workorders->workorder_count);
    *collections_used = previous_workorders->workorder_count;

    /* Get the correct sensor name into the BPF file attributes
       information. */
    ias_bpf_set_file_attributes_sensor_name(bpf,
        ias_sat_attr_get_sensor_name(IAS_TIRS));

    /*Loop through all bands, create bias parameters for applicable TIRS bands*/
    for (band_number = 1;
        status == SUCCESS && band_number <= sat_attributes->total_bands;
        band_number++)
    {
        band_attributes = ias_sat_attr_get_band_attributes(band_number);

        if (band_attributes == NULL)
        {
            IAS_LOG_ERROR("Unable to get attributes for band %d", band_number);
            status = ERROR;
            break;
        }

        /* Bias parameters are calculated for the Normal TIRS bands, skip
        the others */
        if ((band_attributes->band_classification & IAS_NORMAL_BAND) == 0 ||
            (band_attributes->sensor_id != IAS_TIRS))
        {
            IAS_LOG_DEBUG("Skipping band number %d",
                band_attributes->band_number);
            continue;
        }

        /*Set the band number*/
        IAS_LOG_DEBUG("calling ias_bpf_set_bias_model_band_number to set "
            "band number to %d", band_number);
        if (ias_bpf_set_bias_model_band_number(bpf,
            band_attributes->band_number) != SUCCESS)
        {
            IAS_LOG_ERROR("Unable to store the band number in the "
                "IAS BPF");
            status = ERROR;
            break;
        }

        /* Set the spectral type */
        if (ias_bpf_set_bias_model_spectral_type(bpf,
            band_attributes->band_number,
            band_attributes->spectral_type) != SUCCESS)
        {
            IAS_LOG_ERROR("Unable to store the spectral type in the "
                "IAS BPF");
            status = ERROR;
            break;
        }

        IAS_LOG_DEBUG("Entering SCA loop");

        for (sca_index = 0; status == SUCCESS &&
            sca_index < band_attributes->scas;
            sca_index++)
        {
            int sca_number = sca_index + 1;

            if (SUCCESS != load_characterization(post_workorder,
                    previous_workorders, db_connection, band_attributes,
                    sca_number, &band_characterization, &vrp_characterization,
                    IAS_RPS_TIRS_DEEPSPACE_COLLECT))
            {
                IAS_LOG_ERROR("Unable to load characterization for band %d",
                    band_attributes->band_number);
                status = ERROR;
                break;
            }

            IAS_LOG_DEBUG("Calculating TIRS pre/post means for SCA number %d", 
                sca_number);

            if (SUCCESS != calculate_mean(post_workorder, previous_workorders,
                band_attributes, sca_number, band_characterization, 
                previous_collections, db_connection, bpf))
            {
                IAS_LOG_ERROR("Unable to calculate mean for band %d",
                    band_attributes->band_number);
                status = ERROR;
                break;
            }
        } /* End of SCA loop */
    } /* End of band loop */

    IAS_LOG_DEBUG("Generate the misc items for the BPF");

    /* Populate misc information in the BPF */
    if (status == SUCCESS && SUCCESS != generate_file_info(mission_id,
        post_workorder, previous_workorders, db_connection, bpf))
    {
        IAS_LOG_ERROR("Unable to set file info in the BPF");
        status = ERROR;
    }

    IAS_LOG_DEBUG("Freeing resources");

    /* reusable function to release characterization from load*/
    free_characterization(previous_workorders->workorder_count,
        &band_characterization);
 
    /* Cleanup the array of previous RWOC IDs */
    IAS_LOG_DEBUG("Clean up previous workorder IDs");
    ias_rps_db_free_bias_model_workorders(&previous_workorders);

    IAS_LOG_DEBUG("Bias Model Calibration complete with status %d", status);

    return status;
}

/*------------------------------------------------------------------------------
                            Private Routines 
------------------------------------------------------------------------------*/
/*-----------------------------------------------------------------------------
NAME: ias_rps_calculate_frame_variation_coefficients
    
PURPOSE: Perform a linear regression of the per-band, per-SCA, per-frame 
    averages of the imaging detector data and the per-band, per-SCA, per-frame 
    averages of the VRP data to obtain the a0 bias model parameter.

RETURNS: An integer indicating SUCCESS or ERROR
-----------------------------------------------------------------------------*/
static int calculate_frame_variation_coefficients(
    /*I: Post collection workorder ID */
    const char *post_id,
    /*I: Array of previous collection workorder IDs*/
    const IAS_BMC_WORKORDERS *pre_workorders,
    /*I: The number of workorders to use*/
    const int previous_shutter_count,
    /*I: Attributes of the band being processed */
    const IAS_BAND_ATTRIBUTES *band_attributes,
    /*I: Attributes for the VRP band being processed */
    const IAS_BAND_ATTRIBUTES *vrp_attributes,
    /*I: 1 based sca number */
    const int sca_number,
    /*I: value used in fill frames */
    const int frame_fill_value,
    /*I: database connection */
    struct ias_db_connection *db_connection,
    /*I/O: Bias parameters*/
    IAS_BPF *bpf,
    /*O: residuals structure or NULL */
    IAS_RPS_DB_BAND_RESIDUAL *band_residuals)
{
    /*Define the output parameters to gsl_fit_linear*/
    double a0 = 0.0; /*second output coefficient*/
    double c0 = 0.0; /*first output coefficient*/
    double cov00 = 0.0; /*first output in covariance matrix*/
    double cov01 = 0.0; /*second output in covariance matrix*/
    double cov11 = 0.0; /*third output in covariance matrix*/
    double chisq = 0.0; /*the sum of squares of the residuals*/
    double mean = 0.0; 
    double residual = 0.0;

    double* x = NULL;
    double* y = NULL;

    int i = 0; /* Loop iterator */
    int collections_used = 0; /*Actual number of collections used*/
    int line_counter = 0;
    int odd_even = 0;
    int pan_flag = 0;
    int sca_index = sca_number - 1;
    int status = SUCCESS;
    /*the number of points provided to the gsl_linear_fit*/
    int n = 0;
    int xi = 0; /* The x index */
    int yi = 0; /* The y index */

    IAS_RPS_DB_FRAME_AVERAGES **frame_averages = NULL;
    IAS_RPS_DB_FRAME_AVERAGES **vrp_averages = NULL;

    if (band_attributes == NULL)
    {
        IAS_LOG_ERROR("Band attributes provided are NULL");
        return ERROR;
    }
  
    if (vrp_attributes == NULL)
    {
        IAS_LOG_ERROR("VRP attributes provided are NULL");
        return ERROR;
    }

    if (previous_shutter_count > pre_workorders->workorder_count)
    {
        IAS_LOG_WARNING("The number of previous collections requested is "
            "greater than the data provided. %d collections will be used.",
            pre_workorders->workorder_count);
        collections_used = pre_workorders->workorder_count;
    }
    else
    {
        collections_used = previous_shutter_count;
    }

    /* Allocate space for all the pre/post workorders */
    frame_averages = (IAS_RPS_DB_FRAME_AVERAGES**) calloc(
        collections_used, sizeof(IAS_RPS_DB_FRAME_AVERAGES*));

    if (frame_averages == NULL)
    {
        IAS_LOG_ERROR("Unable to perform allocation for frame averages");
        status = ERROR;        
    }

    vrp_averages = (IAS_RPS_DB_FRAME_AVERAGES**) calloc(
        collections_used, sizeof(IAS_RPS_DB_FRAME_AVERAGES*));

    if (vrp_averages == NULL)
    {
        IAS_LOG_ERROR("Unable to perform allocation for VRP frame averages");
        status = ERROR;        
    }
 
    /* load all the frame averages for the pre/post collections */
    for (i = 0; status == SUCCESS && i < collections_used; i++)
    {
        IAS_LOG_DEBUG("Obtaining frame averages for %s", 
            pre_workorders->workorder_ids[i]);

        /* Obtain the averages from the database */
        frame_averages[i] = ias_rps_db_get_frame_averages(db_connection, 
            pre_workorders->workorder_ids[i], band_attributes->band_number, 
            sca_number); 

        vrp_averages[i] = ias_rps_db_get_frame_averages(db_connection,
            pre_workorders->workorder_ids[i], vrp_attributes->band_number, 
            sca_number);

        if (frame_averages[i] == NULL || vrp_averages[i] == NULL)
        {
            IAS_LOG_ERROR("Frame averages obtained are NULL");
            status = ERROR;
            break;
        }

        /* Verify there are the same number of frames for both the normal
            and VRP bands*/
        if (frame_averages[i]->num_lines != vrp_averages[i]->num_lines)
        {
            IAS_LOG_ERROR("Normal and VRP frame count mismatch");
            status = ERROR;
            break;
        }

        IAS_LOG_DEBUG("%d lines obtained for workorder %s", 
            frame_averages[i]->num_lines, pre_workorders->workorder_ids[i]);
 
        n += frame_averages[i]->num_lines;
    }
   
    /* Allocate space for the points provided to gsl_fit_linear */
    x = (double*) calloc(n, sizeof(double));

    if (x == NULL)
    {
        IAS_LOG_ERROR("Unable to perform allocation for x coordinates");
        status = ERROR;        
    }

    y = (double*) calloc(n, sizeof(double)); 

    if (y == NULL)
    {
        IAS_LOG_ERROR("Unable to perform allocation for y coordinates");
        status = ERROR;        
    }

    if (band_attributes->band_type == IAS_PAN_BAND)
    {
        IAS_LOG_DEBUG("Pan flag set");
        pan_flag = 1;
    }

    for (odd_even = 0; odd_even <= pan_flag; odd_even++)
    {
        double diff_sum = 0.0; /*sum of the difference from the mean*/
        double sum = 0.0;
        xi = 0;
        yi = 0;

        /*load the values for the gsl */
        for (i = 0; status == SUCCESS && i < collections_used; 
            i++)
        {
            /* Calculate the frames using the PAN settings */
            for (line_counter = odd_even; 
                line_counter < frame_averages[i]->num_lines; 
                line_counter += (1 + pan_flag))
            {
                /* If the value is complete fill, skip this frame */
                if (vrp_averages[i]->line_average[line_counter] == 
                        frame_fill_value ||
                    frame_averages[i]->line_average[line_counter] == 
                        frame_fill_value)
                { 
                    IAS_LOG_DEBUG("Skipping fill line at %d", line_counter);
                    continue;
                }

                sum += frame_averages[i]->line_average[line_counter];
                x[xi++] = vrp_averages[i]->line_average[line_counter];
                y[yi++] = frame_averages[i]->line_average[line_counter];
            }
        } /* end work order loop */
 
        if (yi == 0)
        {
            mean = 0;
        }
        else
        {
            mean = sum / (double)yi;
        }

        /*Calculate the differences from the sum, used for the residual 
        squared parameter */
        for (i = 0; status == SUCCESS && i < collections_used; i++)
        {
            /* Calculate the frames using the PAN settings */
            for (line_counter = odd_even;
                line_counter < frame_averages[i]->num_lines;
                line_counter += (1 + pan_flag))
            {
                /* If the value is complete fill, skip this frame */
                if (vrp_averages[i]->line_average[line_counter] ==
                        frame_fill_value ||
                    frame_averages[i]->line_average[line_counter] ==
                        frame_fill_value)
                {
                    IAS_LOG_DEBUG("Skipping fill line at %d", line_counter);
                    continue;
                }
                
                diff_sum += 
                    (frame_averages[i]->line_average[line_counter] - mean)
                    * (frame_averages[i]->line_average[line_counter] - mean);
            }
        } 
    
        IAS_LOG_DEBUG("Calling gsl routine");
    
        /* Call to gsl_fit_linear of the GNU Scientific Library
         This function computes the best-fit linear regression
         of the model Y = c_0 + c_1 X */
        if (status == SUCCESS)
        {
            if (0 != gsl_fit_linear(x, 1, y, 1, xi, 
                &c0, &a0, &cov00, &cov01, &cov11, &chisq))
            {
                IAS_LOG_ERROR("Linear fit error");
                status = ERROR;
                break;
            }
        }
   
        if (isnan(a0) || isinf(a0))
        {
            IAS_LOG_ERROR("A0 is an invalid value.");
            status = ERROR;
            break;
        }
    
        if (isnan(c0) || isinf(c0))
        {
            IAS_LOG_ERROR("C0 is an invalid value.");
            status = ERROR;
            break;
        } 
 
        /*Log the model output*/
        IAS_LOG_DEBUG("# best fit: Y = %f + %f X", c0, a0);
 
        /*If there is no difference the residual is 0*/ 
        if (diff_sum == 0)
        {
            residual = 0; 
        }
        else
        {
            residual = 1 - (chisq / diff_sum);
        }
 
        IAS_LOG_DEBUG("# r0 = %lf", residual);
    
        IAS_LOG_DEBUG("Setting the a0 coefficient in the bpf to %lf for "
            "band number %d sca index %d of %d", 
            a0, band_attributes->band_number, sca_index, band_attributes->scas);
       
        if (status == SUCCESS && ias_bpf_set_bias_model_a0_coefficient(
            bpf, band_attributes->band_number, sca_index, 
            /*Use the macro to define which lines are even or odd*/
            IAS_IS_EVEN_LINE(odd_even), a0) 
            != SUCCESS)
        {
            IAS_LOG_ERROR("Unable to store the bias coefficient in the "
                "IAS BPF");
            status = ERROR;
            break;
        }

        /* Store the residual in the database if the band_residuals is NULL */
        if ( ! band_residuals )
        {
            /* output residual by a db routine */
            if (SUCCESS != ias_rps_db_set_bias_model_residual_a0(
                db_connection, residual, post_id,
                band_attributes->band_number,
                sca_number, 
                /*Use the macro to define which lines are even or odd*/
                /*0,1 for pan else 2*/
                (2 - (pan_flag + IAS_IS_EVEN_LINE(odd_even))))) 
            {
                IAS_LOG_ERROR("Unable to store the a0 residual in the "
                    "database");
                status = ERROR;
                break;
            }
        }
        else /* Copy the data into the output structure */
        {
            int lineset = 2 - (pan_flag + IAS_IS_EVEN_LINE(odd_even));
            int ndx     = lineset;
            if (!pan_flag)
                ndx = 0;
            band_residuals->lineset[ndx] = lineset;
            band_residuals->r_squared_a0[ndx][sca_index] = residual;
        }
    } /* End of odd_even loop */

    /* free up the double arrays */
    free(x);
    x = NULL;

    free(y);
    y = NULL;

    /* free the space allocated for the frame averages structures */
    for (i = 0; i < collections_used; i++)
    {  
        ias_rps_db_free_frame_averages(&frame_averages[i]);
        ias_rps_db_free_frame_averages(&vrp_averages[i]);
    }

    free(*frame_averages);
    frame_averages = NULL;

    free(*vrp_averages);
    vrp_averages = NULL;

    if (status != SUCCESS)
    {
        IAS_LOG_DEBUG("calculate_frame_variation_coefficients failed");
    }
    else 
    {
        IAS_LOG_DEBUG("calculate_frame_variation_coefficients successful");
    } 
 
    return status;
}

/*------------------------------------------------------------------------------
NAME: calculate_dark_response_coefficients

PURPOSE: Calculate a1 and c1 coefficients

RETURNS: An integer indicating SUCCESS or ERROR
------------------------------------------------------------------------------*/
static int calculate_dark_response_coefficients(
    /*I: CPF used for processing */
    IAS_CPF *cpf,
    /*I: Post collection workorder ID*/
    const char *post_id,
    /*I: Array of previous collection workorder IDs*/
    const IAS_BMC_WORKORDERS *pre_workorders,
    /*I: The number of previous workorders to use, N1 */
    const int previous_workorders_n1,
    /*I: Histogram statistics for the band */
    IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS **band_characterization,
    /*I: Histogram statistics for the band VRP detectors */
    IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS **vrp_characterization,
    /*I: Attributes for the band being processed */
    const IAS_BAND_ATTRIBUTES *band_attributes,
    /*I: Attributes for the bands VRPs being processed */
    const IAS_BAND_ATTRIBUTES *vrp_attributes,
    /*I: The 1 based sca number to process */
    const int sca_number, 
    /*I: database connection */
    struct ias_db_connection *db_connection,
    /*I/O: Bias parameters*/
    IAS_BPF *bpf,
    /*O: residuals structure or NULL */
    IAS_RPS_DB_BAND_RESIDUAL *band_residuals
)
{
    const struct IAS_BPF_BIAS_MODEL *bias_model = NULL;
    const struct IAS_CPF_DETECTOR_STATUS *detector_status = NULL;

    /*Define the output parameters to gsl_fit_linear*/
    double a0 = 0.0;
    double cov00 = 0.0; /*first output in covariance matrix*/
    double cov01 = 0.0; /*second output in covariance matrix*/
    double cov11 = 0.0; /*third output in covariance matrix*/
    double chisq = 0.0; /*the sum of squares of the residuals*/
    double mean = 0.0;
    double vrp_average = 0.0; /*average of the VRP detector avgs for a collect*/
    double vrp_sum = 0.0; /*sum of the VRP detector averages for a collection */

    double parms[IAS_BPF_MAX_MODEL_PARAM_VALUES];
    double *pre_avg = NULL;
    double *post_avg = NULL;
    double *a1 = NULL;
    double *c1 = NULL;
    double *residuals = NULL; /*residuals from calculating the bias parameters*/

    double **x = NULL; 
    double **y = NULL;

    int collections_used = 0; /*Actual number of collections used*/
    int detector_index = 0;
    int i = 0; /* Loop iterator */
    int odd_even = 0;
    int pan_flag = 0;
    int sca_index = sca_number - 1;
    int status = SUCCESS;
    int temp = 0;
    int vrp_count = 0;
    int vrp_index = 0;

    if (band_attributes == NULL || vrp_attributes == NULL)
    {
        IAS_LOG_ERROR("Attributes for bands provided are NULL");
        return ERROR;
    }

    if (band_characterization == NULL || vrp_characterization == NULL)
    {
        IAS_LOG_ERROR("Histogram statistics provided are NULL");
        return ERROR;
    }

    IAS_LOG_DEBUG("Calculating a1 for band %d SCA number %d",
        band_attributes->band_number, sca_number);

    if (previous_workorders_n1 > pre_workorders->workorder_count)
    {
        IAS_LOG_WARNING("The number of previous collections requested is "
            "greater than the data provided. %d collections will be used.",
            pre_workorders->workorder_count);
        collections_used = pre_workorders->workorder_count;
    }
    else
    {
        collections_used = previous_workorders_n1;
    }

    bias_model = ias_bpf_get_bias_model(bpf);

    if (bias_model == NULL)
    {
        IAS_LOG_ERROR("Unable to obtain the bias model");
        return ERROR;
    }

    detector_status = ias_cpf_get_det_status(cpf);

    if (detector_status == NULL)
    {
        IAS_LOG_ERROR("Unable to obtain the detector status");
        return ERROR;
    }

    /*Allocate space to hold the points used by the GSL */
    x = (double**) calloc(band_attributes->detectors_per_sca,
        sizeof(double*));
    y = (double**) calloc(band_attributes->detectors_per_sca,
        sizeof(double*));

    /* Allocate space to read the current parameters from the model */
    pre_avg = (double*) calloc(band_attributes->detectors_per_sca,
        sizeof(double));
    post_avg = (double*) calloc(band_attributes->detectors_per_sca,
        sizeof(double));
    a1 = (double*) calloc(band_attributes->detectors_per_sca,
        sizeof(double));
    c1 = (double*) calloc(band_attributes->detectors_per_sca,
        sizeof(double));
    residuals = (double*) calloc(band_attributes->detectors_per_sca,
        sizeof(double));

    /* Check if any of the allocations failed, if so set the status to ERROR
        to perform cleanup and exist with an error status */
    if (x == NULL || y == NULL || pre_avg == NULL || post_avg == NULL || 
        a1 == NULL || c1 == NULL || residuals == NULL)
    {
        IAS_LOG_ERROR("Unable to allocate space");
        status = ERROR;
    }

    for (detector_index = 0; status == SUCCESS &&
        detector_index < band_characterization[i]->num_detectors;
        detector_index++)
    {
        /* Allocate space for the points provided to gsl_fit_linear */
        x[detector_index] = (double*) calloc(collections_used, sizeof(double));
        y[detector_index] = (double*) calloc(collections_used, sizeof(double));

        if (x[detector_index] == NULL || y[detector_index] == NULL)
        {
            IAS_LOG_ERROR("Unable to allocate space");
            status = ERROR;
        }
    }

    /*Do an odd/even loop for the pan band split lines */
    if (band_attributes->band_type == IAS_PAN_BAND) 
    {
        pan_flag = 1;
    }

    for (odd_even = 0; odd_even <= pan_flag && status != ERROR; odd_even++)
    {
        /* load the values for the gsl, each point is the detector average vs 
            the VRP detector average average*/
        for (i = 0; i < collections_used && status != ERROR; i++)
        {
            vrp_count = 0;
            vrp_sum = 0.0;
    
            if (vrp_characterization[i]->num_detectors==0)
            {
                IAS_LOG_ERROR("Divide by zero");
                status = ERROR;
                break;
            }

            if (vrp_characterization[i]->mean[odd_even] == NULL ||
                band_characterization[i]->mean[odd_even] == NULL)
            {
                IAS_LOG_ERROR("NULL pointer obtained in means");
                status = ERROR;
                break;
            }
    
            /* Verify there are the expected number of detectors in the 
                characterization data */
            if (status == SUCCESS && band_characterization[i]->num_detectors != 
                band_attributes->detectors_per_sca)
            {
                IAS_LOG_ERROR("Number of detectors in characterization, %d, "
                    "does not match the number of detectors in attributes %d",
                    band_characterization[i]->num_detectors,
                    band_attributes->detectors_per_sca);
                status = ERROR;
                break;
            }

            if (detector_status->inoperable[band_attributes->band_index]
                [sca_index] == NULL)
            {
                IAS_LOG_ERROR("Inoperable list NULL for band index %d, "
                    "sca index %d", 
                    vrp_attributes->band_number, sca_index);
                status = ERROR;
                break;
            }
 
            /* Obtain the averages from the database */
            for (vrp_index = 0; 
                vrp_index < vrp_characterization[i]->num_detectors;
                vrp_index++)
            {
                int inoperable_count = detector_status->inoperable_count
                    [vrp_attributes->band_index][sca_index];
                int skip = 0;
                 
                for (temp = 0; temp < inoperable_count; temp++)
                {
                    if (detector_status->inoperable
                        [vrp_attributes->band_index][sca_index][temp] == 
                        (vrp_index + 1))
                    {
                        IAS_LOG_DEBUG("Skipping VRP at band %d, "
                            "sca %d, detector %d", 
                            vrp_attributes->band_index, 
                            sca_index, vrp_index);
                         skip = 1;
                         break;
                    }
                }
  
                /* Skip the detector if the mask was set */ 
                if (skip == 0)
                { 
                    vrp_sum += 
                        vrp_characterization[i]->mean[odd_even][vrp_index];
                    vrp_count++;
                }
            }
      
            /*If an error has occurred or there are no values used in the mean*/
            if (status != SUCCESS || vrp_count == 0)
            {
                vrp_average = 0;
            }
            else 
            {
                vrp_average = vrp_sum / (double)vrp_count;

                /* populate the point arrays used with the GSL */
                for (detector_index = 0; status == SUCCESS &&
                    detector_index < band_characterization[i]->num_detectors;
                    detector_index++)
                {
                    /* Each point is the VRP average to the detectors averages*/
                    x[detector_index][i] = vrp_average;
                    y[detector_index][i] = band_characterization[i]->
                        mean[odd_even][detector_index];
                }
            }
        } /*End of workorder loop */
    
        /*Get the existing detector values in the model to update the others*/
        if (status == SUCCESS)
        {
            if (SUCCESS != ias_bpf_get_model_parameters(
                bias_model, band_attributes->spectral_type,
                band_attributes->band_number, sca_number,
                band_attributes->detectors_per_sca,
                /*Use the macro to define which lines are even or odd*/
                IAS_IS_EVEN_LINE(odd_even), pre_avg, post_avg, a1, c1, &a0))
            {
                IAS_LOG_ERROR("Unable to get current parameters");
                status = ERROR;
            }
        }

        IAS_LOG_DEBUG("Performing linear fit per detector for %d collections",
            collections_used);
     
 
        /* Calculate the A1 and C1 values per detector */
        for (detector_index = 0; status == SUCCESS &&
            detector_index < band_attributes->detectors_per_sca;
            detector_index++)
        {
            double diff_sum = 0; 
            double sum = 0;
            int workorder = 0;

            /* Call to gsl_fit_linear of the GNU Scientific Library
            This function computes the best-fit linear regression
            of the model Y = c_0 + c_1 X */
            if (0 != gsl_fit_linear(x[detector_index], 1, y[detector_index], 1, 
                collections_used,
                &parms[IAS_BPF_DETECTOR_C1], 
                &parms[IAS_BPF_DETECTOR_A1], 
                &cov00, &cov01, &cov11, &chisq))
            {
                IAS_LOG_ERROR("Linear fit error");
                status = ERROR;
                break;
            }

            /*Verify that the A1/C1 values are not NaN or Inf*/
            if (isnan(parms[IAS_BPF_DETECTOR_A1]) || 
                isinf(parms[IAS_BPF_DETECTOR_A1]))
            {
                IAS_LOG_ERROR("A1 is an invalid value.");
                status = ERROR;
                break;
            } 
     
            if (isnan(parms[IAS_BPF_DETECTOR_C1]) || 
                isinf(parms[IAS_BPF_DETECTOR_C1]))
            { 
                IAS_LOG_ERROR("C1 is an invalid value.");
                status = ERROR;
                break;
            }
       
            /* Assigning the other detector values for the set routine */ 
            parms[IAS_BPF_DETECTOR_PRE_AVG] = pre_avg[detector_index];
            parms[IAS_BPF_DETECTOR_POST_AVG] = post_avg[detector_index];
    
            if (SUCCESS != ias_bpf_set_bias_model_detector_coefficients(
                bpf, band_attributes->band_number, sca_index, detector_index, 
                odd_even, parms, IAS_BPF_MAX_MODEL_PARAM_VALUES))
            {
                IAS_LOG_ERROR("Unable to set bias model per detector values");
                status = ERROR;
                break;
            }
       
            /* Compute the residual */ 
            for (workorder = 0; workorder < collections_used; workorder++)
            {
                sum += y[detector_index][workorder];
            }

            mean = sum / collections_used;

            for (workorder = 0; workorder < collections_used; workorder++)
            {
                diff_sum += (y[detector_index][workorder] - mean) * 
                    (y[detector_index][workorder] - mean);
            }

            /*If there is no difference in the sum squared the residual is 0*/
            if (diff_sum == 0)
            {
                residuals[detector_index] = 0;
            }
            else
            {
                residuals[detector_index] = 1.0 - (chisq / diff_sum);
            }
        } /* End detector loop */
 
        /* Store the residual in the database if the band_residuals is NULL */
        if ( ! band_residuals )
        {
            /*output residual by a db routine */
            if (SUCCESS != ias_rps_db_set_bias_model_residual_a1(
                db_connection, residuals, post_id, band_attributes, sca_number,
                /*Use the macro to define which lines are even or odd*/
                /*0,1 for pan else 2*/
                (2 - (pan_flag + IAS_IS_EVEN_LINE(odd_even))))) 
            {
                IAS_LOG_ERROR("Unable to store the a1 residual");
                status = ERROR;
                break;
            } 
        }
        else /* Copy the data into the output structure */
        {
            int lineset = 2 - (pan_flag + IAS_IS_EVEN_LINE(odd_even));
            int ndx     = lineset;
            if (!pan_flag)
                ndx = 0;
            band_residuals->lineset[ndx] = lineset;
            memcpy(band_residuals->r_squared_a1[ndx][sca_index], residuals,
                (band_attributes->detectors_per_sca * sizeof(double)));
        }
    }/* End odd_even loop */

    /* Clean up the resources used by this routine */
    free(pre_avg);
    pre_avg = NULL;
    free(post_avg);
    post_avg = NULL;
    free(c1);
    c1 = NULL;
    free(a1);
    a1 = NULL;

    for (i = 0; i < collections_used; i++)
    {
        free(x[i]);
        x[i] = NULL;
        free(y[i]);
        y[i] = NULL; 
    }

    free(x);
    x = NULL;
    free(y);
    y = NULL;

    return status;
}

/*------------------------------------------------------------------------------
NAME: calculate_mean

PURPOSE: Calculate the mean of the pre and post collections, Sa and Sb bias 
    parameters.

RETURNS: An integer indicating SUCCESS or ERROR
------------------------------------------------------------------------------*/
static int calculate_mean(
    /*I: Workorder ID for the post collection*/
    const char *post_workorder,
    /*I: Array of previous collection workorder IDs*/
    const IAS_BMC_WORKORDERS *pre_workorders,
    /*I: Attributes for the band being processed */
    const IAS_BAND_ATTRIBUTES *band_attributes,
    /*I: 1 based sca number */
    const int sca_number,
    /*I: Array of band characterization for all work orders, post and pre*/
    IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS **band_characterization,
    /*I: number of previous collections to use*/
    const int previous_collections,
    /*I: database connection */
    struct ias_db_connection *db_connection,
    /*I/O: Bias parameters*/
    IAS_BPF *bpf)
{
    /*place the post at last index*/ 
    const int post_wo_index = 0; /* List is descending, 0 is post*/
    const int sca_index = sca_number - 1;

    double means[IAS_BPF_MAX_MODEL_PARAM_VALUES]; /*pre/post collection means*/
    int collections_used = 0; /*Number of collections to actually use*/
    int detector_index = 0;  /*Detector loop counter*/
    int odd_even = 0;
    int param_values = 0;
    int pan_flag = 0;
    int status = SUCCESS; /*Error occurred flag*/
    int wo_index = 0; /*Work Order loop counter*/

    if (bpf == NULL)
    {
        IAS_LOG_ERROR("BPF parameter is NULL");
        return ERROR;
    }

    if (pre_workorders == NULL)
    {
        IAS_LOG_ERROR("previous workorders pointer is NULL");
        return ERROR;
    } 

    if (previous_collections > pre_workorders->workorder_count)
    {
        IAS_LOG_WARNING("The number of previous collections requested is "
            "greater than the data provided. %d collections will be used.",
            pre_workorders->workorder_count);
        collections_used = pre_workorders->workorder_count;
    }
    else
    {
        collections_used = previous_collections;
    }

    IAS_LOG_DEBUG("Calculating mean using %d collections", collections_used);
  
    if (band_attributes->band_type == IAS_PAN_BAND)
    {
        IAS_LOG_DEBUG("Pan flag set");
        pan_flag = 1;
    }

    for (odd_even = 0; odd_even <= pan_flag; odd_even++)
    {
        for (detector_index = 0; status == SUCCESS && 
            detector_index < band_attributes->detectors_per_sca;
            detector_index++)
        {
            double sum = 0.0;
    
            for (wo_index = post_wo_index + 1;
                status == SUCCESS && wo_index <= collections_used;
                wo_index++)
            { 
                sum += band_characterization[wo_index]->
                    mean[odd_even][detector_index];
            }
    
            /* Only a single post collection is used, so the mean does not need
            to be calculated as is done for the pre average */
            means[IAS_BPF_DETECTOR_PRE_AVG] = sum / (double)collections_used;
            means[IAS_BPF_DETECTOR_POST_AVG] = band_characterization
                [post_wo_index]->mean[odd_even][detector_index];
   
            if (band_attributes->sensor_id == IAS_OLI)
            {
                param_values = IAS_BPF_MAX_MODEL_PARAM_VALUES;
            }  
            else
            {
                param_values = IAS_BPF_THERMAL_VALUES;
            }

            if (SUCCESS != ias_bpf_set_bias_model_detector_coefficients(
                bpf, band_attributes->band_number, sca_index, detector_index,
                odd_even, means, param_values))
            {
                IAS_LOG_ERROR("Unable to set coefficients for band %d, sca %d, "
                    "detector %d", 
                    band_attributes->band_number, sca_number, detector_index);
    
                status = ERROR;
                break;
            }
        }/*end per detector loop*/
    }/*End odd/even loop*/

    if (status == SUCCESS) 
    {
        IAS_LOG_DEBUG("calculate_mean finished successfully");
    }
    else
    {
        IAS_LOG_DEBUG("calculate_mean failed");
    }
 
    return status;
}

/*------------------------------------------------------------------------------
NAME: free_characterization

DESCRIPTION: Release the histogram characterization data provided
------------------------------------------------------------------------------*/
static void free_characterization(
    /*I: Count of histogram statistics, both array are the same size */    
    const int count, 
    /*I: Address of the array of histogram statistic pointers for the bands */
    IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS ***characterization_results)
{
    int i = 0;      

    /* Free each set of histogram statistics from the array */
    for (i = 0; i < count; i++)
    {
        if ((characterization_results) != NULL && 
            (*characterization_results) != NULL &&
            (*characterization_results)[i] != NULL)
        {
            ias_rps_db_free_histogram_results((*characterization_results)[i]);
            (*characterization_results)[i] = NULL;
        }
    }

    /* free the arrays and set them to NULL */
    free(*characterization_results);
    *characterization_results = NULL;
}

/*------------------------------------------------------------------------------
NAME: generate_file_info

PURPOSE: Populate misc fields in the BPF

RETURNS: ERROR if an error occurred, else SUCCESS
------------------------------------------------------------------------------*/
static int generate_file_info(
    /*I: Interval id of the mission data*/
    const char *mission_id,
    /*I: Workorder ID for the post collection*/
    const char *post_workorder,
    /*I: Array of previous collection workorder IDs*/
    const IAS_BMC_WORKORDERS *pre_workorders,
    /*I: database connection */
    struct ias_db_connection *db_connection,
    /*I/O: Bias parameters*/
    IAS_BPF *bpf)
{
    int orbit = 0; /* Orbit value */

    if (pre_workorders == NULL) 
    {
        IAS_LOG_ERROR("Null pointer encountered");
        return ERROR;
    }

    if (SUCCESS != ias_rps_db_bmc_populate_start_stop(db_connection, 
        post_workorder, pre_workorders->workorder_ids[0], bpf))
    {
        IAS_LOG_ERROR("Unable to populate start/stop time");
        return ERROR;
    }

    IAS_LOG_DEBUG("BPF start/stop times populated");

    /* Find the orbit using the landsat interval id */
    if (SUCCESS != ias_rps_db_get_bias_model_orbit(
        db_connection, mission_id, &orbit))
    {
        IAS_LOG_ERROR("Unable to get the orbit number");
        return ERROR;
    }

    IAS_LOG_DEBUG("Orbit %d", orbit);

    /* load orbit in BPF struct */
    ias_bpf_set_orbit_parameters_begin_orbit_number(bpf, orbit);

    /* Database constraints require this field to not be empty */
    ias_bpf_set_file_attributes_file_source(bpf, "None");

    /* Assign the version number to the BPF */
    if (SUCCESS != ias_bpf_db_populate_version(db_connection, bpf))
    {
        IAS_LOG_ERROR("Unable to populate the version in the BPF");
        return ERROR;
    } 

    return SUCCESS;
}

/*------------------------------------------------------------------------------
NAME: load_characterization

PURPOSE: Read the characterization data from the database for later use.

RETURNS: A pointer to an allocated set of histogram statistics, NULL if error
------------------------------------------------------------------------------*/
static int load_characterization(
    /*I: Workorder ID for the post collection*/
    const char *post_workorder,
    /*I: Array of previous collection workorder IDs*/
    const IAS_BMC_WORKORDERS *pre_workorders,
    /*I: Database connection */
    struct ias_db_connection *db_connection,
    /*I Attributes for the band being processed */
    const IAS_BAND_ATTRIBUTES *band_attributes,
    /*I: 1 based sca for which to load data*/
    const int sca_number, 
    /*O: Histogram data for the band being processed*/
    IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS ***band_characterization_ptr,
    /*O: Histogram data for the band VRP detectors, not populated for TIRS*/
    IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS ***vrp_characterization_ptr,
    const char *collection_type /* I: collection type name */
)
{
    int status = SUCCESS;
    int wo_index = 0;

    /* Use local arrays to simplify derefing the pointers passed in provided */
    IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS **band_characterization =NULL;
    IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS **vrp_characterization = NULL;

    /*Allocate pointers for the characterization results, add 1 for the
       post collection work order*/
    *band_characterization_ptr = calloc(pre_workorders->workorder_count + 1,
        sizeof(IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS*));

    /* After the allocation assign the array pointers for easier use */
    band_characterization = *band_characterization_ptr;
    if (band_characterization == NULL)
    {
        IAS_LOG_ERROR("Allocation failed");
        status = ERROR;
    }

    if (band_attributes->sensor_id != IAS_TIRS)
    {
        IAS_LOG_DEBUG("VRPs being loaded");

        *vrp_characterization_ptr = calloc(pre_workorders->workorder_count + 1,
            sizeof(IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS*));

        /* After the allocation assign the array pointers for easier use */
        vrp_characterization = *vrp_characterization_ptr;

        if (vrp_characterization == NULL)
        {
             IAS_LOG_ERROR("Allocation failed"); 
             status = ERROR;
        }
    }

    /* Load the histogram statistics for each work order */
    for (wo_index = 0;
        status == SUCCESS && wo_index < pre_workorders->workorder_count + 1;
        wo_index++)
    {
        /*get the histogram statistics from the database, the post workorder is
        stored at index 0 of the histogram results*/
        if (wo_index == 0)
        {
            band_characterization[wo_index] =
                ias_rps_db_get_histogram_statistics(db_connection,
                    post_workorder, band_attributes->band_number, sca_number,
                    collection_type);

            if (band_attributes->sensor_id != IAS_TIRS)
            { 
                vrp_characterization[wo_index] = 
                    ias_rps_db_get_histogram_statistics(db_connection, 
                        post_workorder, band_attributes->vrp_band_number, 
                        sca_number, collection_type);
            }
        }
        else
        {
            IAS_LOG_DEBUG("Obtaining workorder for index %d", wo_index);

            band_characterization[wo_index] =
                ias_rps_db_get_histogram_statistics(db_connection,
                    pre_workorders->workorder_ids[wo_index - 1], 
                    band_attributes->band_number, sca_number, collection_type);

            if (band_attributes->sensor_id != IAS_TIRS)
            {
                vrp_characterization[wo_index] =
                    ias_rps_db_get_histogram_statistics(db_connection,
                        pre_workorders->workorder_ids[wo_index - 1],
                        band_attributes->vrp_band_number, sca_number, 
                        collection_type);
            }
        }

        /* Log the error and move out of the routine if the characterization
            results were unable to be obtained */
        if (band_characterization[wo_index] == NULL)
        {
            IAS_LOG_ERROR("Unable to obtain characterization results for "
                "band %d, sca number %d", 
                band_attributes->band_number, sca_number);
            status = ERROR;
            break;
        }

        if ((band_attributes->sensor_id != IAS_TIRS) &&
            vrp_characterization[wo_index] == NULL)
        {
            IAS_LOG_ERROR("Unable to obtain VRP characterization results for "
                "band %d, sca number %d",
                band_attributes->vrp_band_number, sca_number);
            status = ERROR;
            break;
        }
    } /*End work order loop*/

    if (status != SUCCESS)
    {
        IAS_LOG_DEBUG("Unable to load histograms");
    }

    IAS_LOG_DEBUG("Finished loading histogram statistics for band number %d", 
        band_attributes->band_number);

    return status;
}

