/***************************************************************************
 NAME:      ias_rps_calculate_filtered_statistics

 PURPOSE:   Calculates basic statistics using a specified filter value
            to exclude samples from the calculations

 ROUTINES:  ias_rps_calculate_filtered_mean
            ias_rps_calculate_filtered_stdev
            ias_rps_calculate_filtered_rms
****************************************************************************/
#include <math.h>
#include <float.h>         /* For DBL_EPSILON machine precision limit */
#include "ias_logging.h"
#include "ias_radiometry.h"


/*-----------------------------------------------------------------------
 NAME:     ias_rps_calculate_filtered_mean

 PURPOSE:  Calculates the mean of a given data vector from all samples
           excluding those flagged as "invalid"

 RETURNS:  Integer status code of SUCCESS or ERROR

 NOTES:    The optional output *valid_count in ias_rps_calculate_filtered_mean
           returns the total number of "valid" samples used in the
           calculations.  It is expected that this value would be the same
           as the number of valid samples used to calculate the standard
           deviation, given the same input data vector, total number of
           samples and set of invalid flags.
-------------------------------------------------------------------------*/
int ias_rps_calculate_filtered_mean
(
    const double *data_buffer,     /* I: Data buffer */
    const int *invalid_flags_buffer,
                                   /* I: Data validity flag for all
                                      samples (0=valid sample,
                                      nonzero=invalid sample) */
    int num_samples,               /* I: Total # of samples*/
    double *valid_mean,            /* O: Valid sample mean */
    int *valid_count               /* O: # of data samples used in the
                                      calculations (pass in NULL if this
                                      output is not desired) */
)
{
    int tracking_count = 0;        /* Valid sample counter */
    int sample_index;

    double sum = 0.0;              /* Cumulative sum of samples */

    /* Sanity check:  do we have a "reasonable" number of samples? */
    if (num_samples < 1)
    {
        IAS_LOG_ERROR("Invalid number of input data samples %d",
                      num_samples);
        return ERROR;
    }

    /* Inputs look good, so do the calculation.  Exclude flagged sample
       values */
    for (sample_index = 0; sample_index < num_samples; sample_index++)
    {
        if (invalid_flags_buffer[sample_index])
            continue;

        sum += data_buffer[sample_index];
        tracking_count++;
    }

    /* Do we have a reasonable amount of non-zero data */
    if (tracking_count < 1)
    {
        IAS_LOG_ERROR("Too few valid data samples counted:  %d",
                      tracking_count);
        return ERROR;
    }

    /* We've got valid data, calculate the mean */
    *valid_mean = sum / tracking_count;

    /* If the number samples used is desired (count pointer is NOT NULL),
       set it, too */
    if (valid_count != NULL)
        *valid_count = tracking_count;

    return SUCCESS;
}




/*-----------------------------------------------------------------------
 NAME:     ias_rps_calculate_filtered_stdev

 PURPOSE:  Calculates the std deviation of a given data vector from all
           samples excluding those flagged as "invalid"

 RETURNS:  Integer status code of SUCCESS or ERROR
-------------------------------------------------------------------------*/
int ias_rps_calculate_filtered_stdev
(
    const double *data_buffer,     /* I: Data buffer */
    const int *invalid_flags_buffer,
                                   /* I: Data validity flag for all
                                      samples (0=valid sample,
                                      nonzero=invalid sample) */
    int num_samples,               /* I: Total # of samples */
    double valid_mean,             /* I: Valid sample mean */
    double *stdev                  /* O: Sample standard deviation */
)
{
    int data_count = 0;            /* Valid sample counter */
    int sample_index;                 /* 0-based detector index */

    double resid = 0.0;            /* Sum of residuals */
    double sum_sqr = 0.0;          /* Cumulative sum of squared samples */
    double diff;                   /* Difference between a sample value and
                                      the mean */
    double var;                    /* Sample variance */

    /* Sanity check:  do we have a "reasonable" amount of data? */
    if (num_samples < 1)
    {
        IAS_LOG_ERROR("Invalid number of input data samples");
        return ERROR;
    }
    
    /* Inputs appear to be good.  Do the calculation, excluding samples
       flagged as "invalid"  */
    for (sample_index = 0; sample_index < num_samples; sample_index++) 
    {
        if (invalid_flags_buffer[sample_index])
            continue;

        data_count++;
        diff = data_buffer[sample_index] - valid_mean;
        resid += diff;
        sum_sqr += (diff * diff);
    }

    /* If we have virtually no good data, log an error */
    if (data_count < 2)
    {
        IAS_LOG_ERROR("Too few valid data samples counted:  %d",
                      data_count);
        return ERROR;
    }
    else
    {
        /* Calculate the variance.  The formula is taken from eq
           (14.1.8) in "Numerical Recipes, The Art of Scientific
           Computing (vol 3)", pg. 724 (Ch 14.  Statistical
           Description of Data).  It's intended to reduce roundoff
           effects in the final result*/
        var = (sum_sqr - ((resid*resid)/data_count)) / (data_count - 1);

        /* If the variance is "very close" to 0 (or negative), log a
           warning and set the standard deviation to 0.  Since testing for
           "exact" 0 isn't feasible for floating point values,
           use the machine's lower limit as an "effective" 0 */
        if (var < 0)
        {
            IAS_LOG_WARNING("Negative variance %lf, setting standard "
                "deviation to 0", var);
            *stdev = 0.0;
        }
        else if (var < DBL_EPSILON)
        {
            IAS_LOG_WARNING("Variance %lf effectively 0, setting standard "
                "deviation to 0", var);
            *stdev = 0.0;
        }
        else
            *stdev = sqrt(var);
    }

    return SUCCESS;
}




/*-----------------------------------------------------------------------
 NAME:     ias_rps_calculate_filtered_rms

 PURPOSE:  Calculates RMS value for a set of samples given their mean
           excluding samples flagged as "invalid"

 RETURNS:  Integer status code of SUCCESS or ERROR

 NOTES:    The number of valid data samples is returned as an optional
           output if its pointer parameter is not NULL
-------------------------------------------------------------------------*/
int ias_rps_calculate_filtered_rms
(
    const double *data_buffer,    /* I: Data buffer */
    const int *invalid_flags_buffer,
                                  /* I: Data validity flag for all
                                     samples (0=valid sample,
                                     nonzero=invalid sample) */
    int num_samples,              /* I: Total # of samples*/
    double valid_mean,            /* I: Sample mean */
    double *valid_rms,            /* O: Sample RMS */
    int *valid_count              /* O: Optional count on the number of
                                     valid samples (pass in NULL if this
                                     output is not desired) */
)
{
    int count = 0;
    int sample_index;                /* 0-based detector index */

    double diff;                  /* Difference between sample value and
                                     mean */
    double sum_sqr = 0.0;         /* Sum of squared differences */


    /* Sanity check */
    if (num_samples < 1)
    {
        IAS_LOG_ERROR("Invalid number of input samples %d", num_samples);
        return ERROR;
    }

    /* Do the calculations */
    for (sample_index = 0; sample_index < num_samples; sample_index++)
    {
        if (invalid_flags_buffer[sample_index])
            continue;

        diff = data_buffer[sample_index] - valid_mean;
        sum_sqr += (diff * diff);
        count++;
    }

    /* Do we have a valid data count ? */
    if (count < 1)
    {
        IAS_LOG_ERROR("Insufficient valid data found for RMS calculations");
        return ERROR;
    }

    /* Calculate the RMS value */
    *valid_rms = sqrt(sum_sqr / count);

    /* Get the total number of valid samples if that output is desired */
    if (valid_count != NULL)
        *valid_count = count;

    return SUCCESS;
}
