/*----------------------------------------------------------------------------
NAME:    ias_rps_calculate_sca_offsets.c

PURPOSE: Contains two routines for calculating SCA offsets -- one uses the
         Legendre coefficients, orbital and earth parameters from the CPF
         and L0R data; the other uses the model.  Routines are:

             void ias_rps_legendre_calculate_sca_offsets
             int ias_rps_model_calculate_sca_offsets
----------------------------------------------------------------------------*/
#include <string.h>
#include <math.h>

#include "ias_const.h"
#include "ias_cpf.h"
#include "ias_logging.h"
#include "ias_math.h"
#include "ias_los_model.h"
#include "ias_rps_calculate_sca_offsets.h"


/**************************************************************************
 NAME:     ias_rps_legendre_calculate_sca_offsets

 PURPOSE:  Determines an "adjusted" frame offset for all SCAs in a band and
           the frame offset for that band, based on the image nominal frame
           time, the image scene-center latitude and elevation, and
           various Earth constants.  Consideration of these additional
           factors should result in offset values that refine band and SCA
           alignment in an image.

 RETURNS:  Nothing.
***************************************************************************/
void ias_rps_legendre_calculate_sca_offsets
(
    const double *legendre_coeffs, /* I: 1D array containing Legendre
                                         coefficients for all SCAs in the
                                         current band.  Layout of this array
                                         follows that of CPF, indexed into by
                                         [sca_index * IAS_LOS_LEGENDRE_TERMS] */
    int    number_scas,            /* I: Number of SCAs */
    double scene_center_latitude,  /* I: Scene center latitude in degrees */
    double nominal_scene_elevation,/* I: Scene elevation in km */
    double nominal_frame_time,     /* I: Sensor-specific nominal frame
                                         time.  Calling routine is expected
                                         to adjust this value appropriately
                                         for the pan band.  The unit for this
                                         in the CPF is in milliseconds;
                                         for the purposes of this algorithm
                                         it needs to be seconds */
    double eccentricity,           /* I: Earth's eccentricity */
    double semi_major_axis,        /* I: Earth's semi-major axis */
    double semi_minor_axis,        /* I: Earth's semi-minor axis */
    double angular_velocity,       /* I: Earth's angular velocity */
    double inclination_angle,      /* I: Earth's inclination angle */
    double nominal_orbit_radius,   /* I: Earth's nominal orbit radius */
    double orbital_period,         /* I: Earth's orbital period */
    int    *band_offset,           /* O: Adjusted offset for the current
                                         band */
    int    *sca_offsets            /* O: Buffer containing adjusted per-SCA
                                         offsets for the current band */
)
{
    double total_earth_theta[number_scas]; /* Total Earth-relative angle
                                              between adjacent SCAs */
    double earth_theta_first[number_scas]; /* Earth-relative angle for first
                                              detector in right SCA */
    double earth_theta_last[number_scas];  /* Earth-relative angle for last
                                              detector in left SCA */

    /* Conversion factors between radians and degrees */
    double radians_per_degree = ias_math_get_radians_per_degree();

    /* Calculate intermediate quantities used in the various calculations */
    double semi_axis_ratio = semi_minor_axis / semi_major_axis;
    double scenter_lat_angle_radians =
        scene_center_latitude * radians_per_degree;
    double cos_scenter_lat = cos(scenter_lat_angle_radians);
    double cos_sq_scenter_lat = cos_scenter_lat * cos_scenter_lat;
    double sin_sq_scenter_lat = 1. - cos_sq_scenter_lat;

    /* Calculate the radius of the earth at the specified scene center
       latitude angle.  The radius of the earth should be in km */
    double numerator = cos_sq_scenter_lat
        + (pow(semi_axis_ratio, 4) * sin_sq_scenter_lat);
    double denominator = 1. - (eccentricity * sin_sq_scenter_lat);
    double earth_radius = semi_major_axis * sqrt((numerator / denominator));

    /* Calculate an adjustment factor based on the nominal orbit radius,
       the earth radius calculated above, and the nominal scene elevation.
       This factor will be used in converting the sensor-relative SCA
       angles to the corresponding Earth-relative angles.  */
    double radius_adjustment = nominal_orbit_radius
        / (earth_radius + nominal_scene_elevation);

    /* Correct the angular velocity to account for the Earth's
       inclination angle */
    double cos_inclination_angle =
        cos((inclination_angle * radians_per_degree));
    double adj_angular_velocity = ((2. * ias_math_get_pi()) / orbital_period)
        - (angular_velocity * cos_inclination_angle);

    int sca_index;

    /* Initialize the first SCA's total earth angle to 0 */
    total_earth_theta[0] = 0;

    for (sca_index = 0; sca_index < number_scas; sca_index++)
    {
        /* Locate the correct set of Legendre coefficients for the
           current SCA */
        int coeff1_offset = sca_index * IAS_LOS_LEGENDRE_TERMS;

        /* Determine the sensor-specific SCA angles using the available
           Legendre coefficients.  The 'odd' coefficients are subtracted
           when calculating the SCA angle for the first detector (the
           first detector represents the starting value of -1 for the
           range of values over which the Legendre polynomials are
           calculated; similarly, the last detector represents the
           ending value of +1 for the range) */
        double sensor_theta_first = ias_math_eval_legendre(-1.0,
            &legendre_coeffs[coeff1_offset], IAS_LOS_LEGENDRE_TERMS);

        double sensor_theta_last = ias_math_eval_legendre(1.0,
            &legendre_coeffs[coeff1_offset], IAS_LOS_LEGENDRE_TERMS);

        /* Convert the sensor-relative SCA angles to the corresponding
           Earth-relative angles in degrees */
        earth_theta_first[sca_index] =
            (asin(radius_adjustment * sin(sensor_theta_first)))
             - sensor_theta_first;
        earth_theta_last[sca_index] =
            (asin(radius_adjustment * sin(sensor_theta_last)))
             - sensor_theta_last;
    }

    /* Calculate the total earth angle for all but the first
       SCA.  It's the difference between the last detector's angle in
       the left SCA and the first detector's angle in the right
       SCA */
    for (sca_index = 0; sca_index < number_scas - 1; sca_index++)
    {
        total_earth_theta[sca_index + 1] = earth_theta_first[sca_index + 1]
            - earth_theta_last[sca_index];
    }

    /* Find the current band offset, which is the number of frames
       between the bore site and the last detector of the first SCA */
    *band_offset = (int)((earth_theta_last[0]
        / adj_angular_velocity / nominal_frame_time) + 0.5);

    /*  Calculate the adjusted SCA offset.  For each SCA, we need to
        sum up the offsets from all the SCAs previous to it */
    for (sca_index = 0; sca_index < number_scas; sca_index++)
    {
        int i;
        double sum = 0.0;
        for (i = 0; i <= sca_index; i++)
        {
            sum += (total_earth_theta[i] / adj_angular_velocity
                / nominal_frame_time);
        }
        if (sum < 0.0)
            sca_offsets[sca_index] = (int)(sum - 0.5);
        else
            sca_offsets[sca_index] = (int)(sum + 0.5);
    }
}   /* END -- ias_rps_calculate_sca_offsets */
