/*---------------------------------------------------------------------------
NAME:    ias_rps_convert_detector_delays2spatial_offsets

PURPOSE: To convert the floating point detector delays from the CPF into 
         normalized spatial detector offsets.

         The detector delays stored in the CPF are floating point and give
         the amount of delay a given detector has to seeing the same point
         in the image.  For use in the RPS these delays need to be
         translated to an offset into the image.  This is to reduce the
         overhead of calculating this all the time and eliminating negative
         offsets.  For example the CPF may have values such as:

            +0.000, -2.021, +0.000, +2.021, +0.000, +4.021, +0.000, +2.021

         and would be converted to:

             4, 6, 4, 2, 4, 0, 4, 2

         note the negative delay has the largest offset and the largest
         delay has an offset of 0.

RETURNS: None
----------------------------------------------------------------------------*/
#include <math.h>

#include "ias_radiometry.h"

void ias_rps_convert_detector_delays2spatial_offsets
(
    const double *detector_delays,  /* I: Detector delays from CPF */
    int number_detectors,           /* I: Number of values in the arrays */
    int *spacial_offsets,           /* O: Converted offset values */
    int *largest_offset             /* O: Largest offset, if NULL is passed
                                          the value isn't returned */
)
{
    int det;
    int max_delay = (int)floor( detector_delays[0] + 0.5 );
    uint16_t max_offset = 0;

    /* Find the largest detector delay in the input array */
    for ( det = 1; det < number_detectors; det++ )
    {
        int value = (int)floor( detector_delays[det] + 0.5 );
        if ( value > max_delay )
            max_delay = value;
    }

    /* Convert the detector delays to spatial offsets */
    for ( det = 0; det < number_detectors; det++ )
    {
        spacial_offsets[det] =
            -(floor( detector_delays[det] + 0.5 ) - max_delay );
        if ( spacial_offsets[det] > max_offset )
            max_offset = spacial_offsets[det];
    }

    if ( largest_offset )
        *largest_offset = max_offset;
}
