#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ias_bpf.h"
#include "ias_bpf_database.h"
#include "ias_db.h"
#include "ias_db_insert.h"
#include "ias_db_query.h"
#include "ias_radiometry.h"
#include "ias_logging.h"
#include "ias_rps_db_bias_model_calibration.h"


/*----------------------------------------------------------------------------
NAME: ias_rps_db_get_bias_model_previous_workorders

PURPOSE: Query the database to obtain the work order IDs for the previous 
    collections to use in calculating the bias parameters. The order of the 
    workorders is obtained newest first.

RETURNS: A pointer to a populated workorder structure, NULL on error
-----------------------------------------------------------------------------*/
IAS_BMC_WORKORDERS* ias_rps_db_get_bias_model_previous_workorders(
    /*I: CPF used in processing */
    IAS_CPF *cpf,
    /*I: Database connection*/
    struct ias_db_connection *db_connection,
    /*I: The common work order ID of the most recent collection */
    const char *post_collection_work_order,
    /*I: The number of previous collections */
    const int previous_collections_requested)
{
    const int where_clause_size = 512;
    const int sql_statement_size = 1024; /*SQL buffer size*/

    char sql[sql_statement_size]; /*Buffer for SQL commands*/
    char common_where_clause[where_clause_size];

    int collections_available;
    int status = SUCCESS;
    int wo_index = 0;

    const struct IAS_CPF_OLI_PARAMETERS *oli_parms = NULL;
    struct ias_db_query *query = NULL; /*DB query structure*/
    IAS_BMC_WORKORDERS *workorders = NULL;

    IAS_LOG_DEBUG("Selecting previous workorders based on collection %s and "
        "type %s", post_collection_work_order, IAS_RPS_OLI_SHUTTER_COLLECT); 

    oli_parms = ias_cpf_get_oli_parameters(cpf);
    if (oli_parms == NULL)
    {
        IAS_LOG_ERROR("Unable to obtain OLI CPF Parameters");
        return NULL;
    }

    /* Obtain the common part of the where clause to use in queries */
    status = ias_rps_db_create_oli_bias_model_where_clause(common_where_clause, 
        sizeof(common_where_clause), "bias_frame_averages", 
        oli_parms->nominal_integration_time_ms,
        oli_parms->nominal_integration_time_pan);

   /* Query to obtain the number of collections available to handle cases
      where the full number is unavailable. The query is limited to return at
      maximum the number of previous collections that will actually be used. 
      Without limiting the number of results the entire table is checked
      which dramatically impacts performance.*/
   status = snprintf(sql, sizeof(sql),
        "SELECT COUNT(RADIOMETRIC_WO_COMMON_ID)\n"
        "FROM RADIOMETRIC_WORK_ORDER_COMMON\n"
        "WHERE %s\n"
        "AND DATE_ACQUIRED < (SELECT DATE_ACQUIRED\n"
        "  FROM RADIOMETRIC_WORK_ORDER_COMMON\n"
        "  WHERE RADIOMETRIC_WO_COMMON_ID = %s)\n"
        "AND ROWNUM <= %d;\n",
        common_where_clause, 
        post_collection_work_order, previous_collections_requested);

    if (status >= sizeof(sql) || status < 0)
    {
        IAS_LOG_ERROR("Unable to create query string");
        return NULL;
    }

    query = ias_db_perform_query(db_connection, sql);
    if (!ias_db_query_was_successful(query))
    {
        IAS_LOG_ERROR("Unable to execute query: %s",
            sql);
        return NULL;
    }

    if (!ias_db_query_next(query))
    {
        IAS_LOG_ERROR("No data returned using RWOC ID %s", 
            post_collection_work_order);
        ias_db_query_close(query);
        return NULL;
    }

    if (SUCCESS != ias_db_query_int_value(query, &collections_available, 0))
    {
        IAS_LOG_ERROR("Unable to get the count of previous collections.");
        ias_db_query_close(query);
        return NULL;
    }

    ias_db_query_close(query);
    query = NULL;

    if (collections_available < 1)
    {
        IAS_LOG_ERROR("No previous collections found.");
        return NULL;
    }

    /* Allocate the space for the workorders structure */
    workorders = (IAS_BMC_WORKORDERS*) calloc(1, sizeof(IAS_BMC_WORKORDERS));
    if (workorders == NULL)
    {
        IAS_LOG_ERROR("Unable to allocate workorders structure");
        return NULL;
    }

    /* Require at least 2 previous collections to be available, this is the 
       minimum possible to create A1/C1 parameters*/ 
    if (collections_available < 2)
    {
        IAS_LOG_WARNING("Insufficient previous collections, %d, are available "
            "for processing", collections_available);
        workorders->workorder_count = 0; 
        return workorders;
    }

    IAS_LOG_DEBUG("Previous collections found %d", collections_available);

    if (collections_available < previous_collections_requested)
    {
        IAS_LOG_WARNING("%d previous collections do not exist for "
            "work order %s. %d collections found and will be used instead.",
            previous_collections_requested, post_collection_work_order,
            collections_available);
        workorders->workorder_count = collections_available;
    }
    else
    {
        workorders->workorder_count = previous_collections_requested;
    }

    /* Allocate space for each work order ID */
    workorders->workorder_ids = 
        (char **)calloc(workorders->workorder_count, sizeof(char *));
    if (workorders->workorder_ids == NULL)
    {
        IAS_LOG_ERROR("Allocating work order ID string buffer");
        ias_rps_db_free_bias_model_workorders(&workorders);
        return NULL;
    }

    workorders->workorder_buffer = (char *)calloc(
        (IAS_CHAR_ID_SIZE) * workorders->workorder_count, sizeof(char)); 

    if (workorders->workorder_buffer == NULL)
    {
        IAS_LOG_ERROR("Unable to allocate string");
        ias_rps_db_free_bias_model_workorders(&workorders);
        return NULL;
    }

    IAS_LOG_DEBUG("Space for previous workorders allocated. Building query");

    /* Query to obtain the RWOC IDs for the previous collections. The query is
       limited to return at maximum the number of previous collections that will
       actually be used. Without limiting the number of results the entire 
       table is checked which dramatically impacts performance. */
    status = snprintf(sql, sizeof(sql),
        "SELECT RADIOMETRIC_WO_COMMON_ID\n"
        "FROM RADIOMETRIC_WORK_ORDER_COMMON rwoc\n"
        "WHERE %s\n"
        "AND DATE_ACQUIRED < (SELECT DATE_ACQUIRED\n"
        "  FROM RADIOMETRIC_WORK_ORDER_COMMON\n"
        "  WHERE RADIOMETRIC_WO_COMMON_ID = %s)\n"
        "ORDER BY DATE_ACQUIRED desc\n"
        "FETCH FIRST %d ROWS ONLY;\n",
        common_where_clause, 
        post_collection_work_order, workorders->workorder_count);

    if (status >= sizeof(sql) || status < 0)
    {
        IAS_LOG_ERROR("Unable to create query string");
        return NULL;
    }

    query = ias_db_perform_query(db_connection, sql);
    if (!ias_db_query_was_successful(query))
    {
        IAS_LOG_ERROR("Unable to acquired collections using RWOC ID %s",
            post_collection_work_order);
        return NULL;
    }

    for (wo_index = 0; wo_index < workorders->workorder_count; wo_index++)
    {
        if (!ias_db_query_next(query))
        {
            IAS_LOG_ERROR("No data returned for index %d using RWOC ID %s",
                wo_index, post_collection_work_order);
            ias_db_query_close(query);
            return NULL;
        }

        /* Assign the pointers to the individual workorders */
        workorders->workorder_ids[wo_index] =
            &workorders->workorder_buffer[wo_index * IAS_CHAR_ID_SIZE];
   
        ias_db_query_string_value(query, 0, 
            workorders->workorder_ids[wo_index], IAS_CHAR_ID_SIZE);
 
        if (strlen(workorders->workorder_ids[wo_index]) < 1)
        {
            IAS_LOG_ERROR("No ID returned for work order %d", wo_index);
            ias_db_query_close(query);
            return NULL;
        }

        IAS_LOG_INFO("Previous collect %d, workorder ID:  %s",
            (wo_index + 1), workorders->workorder_ids[wo_index]);
    }

    IAS_LOG_DEBUG("Query complete");
    ias_db_query_close(query); 

    IAS_LOG_DEBUG("Done loading workorders");

    return workorders;
}

/*----------------------------------------------------------------------------
NAME: ias_rps_db_get_bias_model_previous_workorders_tirs

PURPOSE: Query the database to obtain the work order IDs for the previous 
    collections to use in calculating the bias parameters. The order of the 
    workorders is obtained newest first.

RETURNS: A pointer to a populated workorder structure, NULL on error
-----------------------------------------------------------------------------*/
IAS_BMC_WORKORDERS* ias_rps_db_get_bias_model_previous_workorders_tirs(
    /*I: CPF used in processing */
    IAS_CPF *cpf,
    /*I: Database connection*/
    struct ias_db_connection *db_connection,
    /*I: The common work order ID of the most recent collection */
    const char *post_collection_work_order,
    /*I: The number of previous collections */
    const int previous_collections_requested)
{
    const int sql_statement_size = 1024; /*SQL buffer size*/
    const int where_clause_size = 512;

    char sql[sql_statement_size]; /*Buffer for SQL commands*/
    char common_where_clause[where_clause_size];

    int collections_available;
    int status = SUCCESS;
    int wo_index = 0;

    struct ias_db_query *query = NULL; /*DB query structure*/
    IAS_BMC_WORKORDERS *workorders = NULL;

    IAS_LOG_DEBUG("Selecting previous workorders based on collection %s and "
        "type %s", post_collection_work_order,
        IAS_RPS_TIRS_DEEPSPACE_COLLECT);

    /* Obtain the common portion of the where clause to use in queries. */
    status = ias_rps_db_create_tirs_bias_model_where_clause(
        common_where_clause, sizeof(common_where_clause), 
        "bias_frame_averages");

    /* Query to obtain the number of collections available to handle cases
       where the full number is unavailable. The query is limited to return at
       maximum the number of previous collections that will actually be used. 
       Without limiting the number of results the entire table is checked
       which dramatically impacts performance. */
    status = snprintf(sql, sizeof(sql),
        "SELECT COUNT(RADIOMETRIC_WO_COMMON_ID)\n"
        "FROM RADIOMETRIC_WORK_ORDER_COMMON\n"
        "WHERE %s\n"
        "AND DATE_ACQUIRED < (SELECT DATE_ACQUIRED\n"
        "  FROM RADIOMETRIC_WORK_ORDER_COMMON\n"
        "  WHERE RADIOMETRIC_WO_COMMON_ID = %s)\n"
        "AND ROWNUM <= %d;\n",
        common_where_clause,
        post_collection_work_order, previous_collections_requested);

    if (status >= sizeof(sql) || status < 0)
    {
        IAS_LOG_ERROR("Unable to create query string");
        return NULL;
    }

    query = ias_db_perform_query(db_connection, sql);
    if (!ias_db_query_was_successful(query))
    {
        IAS_LOG_ERROR("Unable to execute query: %s", sql);
        return NULL;
    }

    if (!ias_db_query_next(query))
    {
        IAS_LOG_ERROR("No data returned for RWOC ID %s",
            post_collection_work_order);
        ias_db_query_close(query);
        return NULL;
    }

    if (SUCCESS != ias_db_query_int_value(query, &collections_available, 0))
    {
        IAS_LOG_ERROR("Unable to get the count of previous collections.");
        IAS_LOG_DEBUG("SQL:\n %s", sql);
        ias_db_query_close(query);
        return NULL;
    }

    ias_db_query_close(query);
    query = NULL;

    /* Allocate the space for the workorders structure */
    workorders = (IAS_BMC_WORKORDERS*) calloc(1, sizeof(IAS_BMC_WORKORDERS));

    if (workorders == NULL)
    {
        IAS_LOG_ERROR("Unable to allocate workorders structure");
        return NULL;
    }

    /* Require at least 1 previous collections to be available */
    if (collections_available < 1)
    {
        IAS_LOG_WARNING("Insufficient previous collections, %d, are available "
            "for processing", collections_available);
        workorders->workorder_count = 0;
        return workorders;
    }

    IAS_LOG_DEBUG("Previous collections found %d", collections_available);

    if (collections_available < previous_collections_requested)
    {
        IAS_LOG_WARNING("%d previous collections do not exist for "
            "work order %s. %d collections found and will be used instead.",
            previous_collections_requested, post_collection_work_order,
            collections_available);

        workorders->workorder_count = collections_available;
    }
    else
    {
        workorders->workorder_count = previous_collections_requested;
    }

    /* Allocate space for each work order ID */
    workorders->workorder_ids =
        (char**) calloc(workorders->workorder_count,  sizeof(char*));

    workorders->workorder_buffer = (char*) calloc(
        IAS_CHAR_ID_SIZE * workorders->workorder_count, sizeof(char));

    if (workorders->workorder_buffer == NULL)
    {
        IAS_LOG_ERROR("Unable to allocate string");
        ias_rps_db_free_bias_model_workorders(&workorders);
        return NULL;
    }

    /* Assign the pointers to the individual workorders */
    for (wo_index = 0; wo_index < workorders->workorder_count; wo_index++)
    {
        workorders->workorder_ids[wo_index] =
            &workorders->workorder_buffer[wo_index * IAS_CHAR_ID_SIZE];
    }

    IAS_LOG_DEBUG("Space for previous workorders allocated. Building query");

    /* Query to obtain the RWOC IDs. The query is limited to return at maximum
       the number of previous collections that will actually be used. Without 
       limiting the number of results the entire table is checked which
       dramatically impacts performance. */
    status = snprintf(sql, sizeof(sql),
        "SELECT RADIOMETRIC_WO_COMMON_ID\n"
        "FROM RADIOMETRIC_WORK_ORDER_COMMON\n"
        "WHERE %s\n"
        "AND DATE_ACQUIRED < (SELECT DATE_ACQUIRED\n"
        "  FROM RADIOMETRIC_WORK_ORDER_COMMON\n"
        "  WHERE RADIOMETRIC_WO_COMMON_ID = %s)\n"
        "ORDER BY DATE_ACQUIRED desc\n"
        "FETCH FIRST %d ROWS ONLY;\n",
        common_where_clause,
        post_collection_work_order, workorders->workorder_count);

    if (status >= sizeof(sql) || status < 0)
    {
        IAS_LOG_ERROR("Unable to create query string");
        return NULL;
    }

    query = ias_db_perform_query(db_connection, sql);
    if (!ias_db_query_was_successful(query))
    {
        IAS_LOG_ERROR("Unable to acquired collections using RWOC ID %s",
            post_collection_work_order);
        return NULL;
    }

    IAS_LOG_DEBUG("Query complete");

    for (wo_index = 0; wo_index < workorders->workorder_count; wo_index++)
    {
        if (!ias_db_query_next(query))
        {
            IAS_LOG_ERROR("No data returned for index %d using RWOC ID %s",
                wo_index, post_collection_work_order);
            ias_db_query_close(query);
            return NULL;
        }

        /* Assign the pointers to the individual workorders */
        workorders->workorder_ids[wo_index] =
            &workorders->workorder_buffer[wo_index * IAS_CHAR_ID_SIZE];

        ias_db_query_string_value(query, 0,
            workorders->workorder_ids[wo_index], IAS_CHAR_ID_SIZE);

        if (strlen(workorders->workorder_ids[wo_index]) < 1)
        {
            IAS_LOG_ERROR("No ID returned for work order %d", wo_index);
            ias_db_query_close(query);
            return NULL;
        }

        IAS_LOG_INFO("Previous collect %d, workorder ID:  %s",
            (wo_index + 1), workorders->workorder_ids[wo_index]);
    }

    IAS_LOG_DEBUG("Done loading workorders");

    ias_db_query_close(query);

    return workorders;
}

/*----------------------------------------------------------------------------
NAME: ias_rps_db_free_bias_model_workorders

PURPOSE: Release memory allocated for a workorders struct.

RETURNS: An integer indicating SUCCESS or ERROR
-----------------------------------------------------------------------------*/
void ias_rps_db_free_bias_model_workorders(
    /* Address of a pointer to a workorder struct*/
    IAS_BMC_WORKORDERS **workorders)
{
    IAS_BMC_WORKORDERS *wo_ptr = *workorders;

    IAS_LOG_DEBUG("Releasing structure");

    free(wo_ptr->workorder_buffer);
    free(wo_ptr->workorder_ids);
    free(wo_ptr);

    *workorders = NULL;
}

/*----------------------------------------------------------------------------
NAME: ias_rps_db_get_bias_model_orbit

PURPOSE: Obtain the orbit used in the BPF

RETURNS: An integer indicating SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int ias_rps_db_get_bias_model_orbit(
    /*I: database connection */
    struct ias_db_connection *dbase,
    /*I: landsat interval ID, mission id to find orbit */
    const char interval_id[],
    /*O: orbit*/
    int *orbit)
{
    const int sql_statement_size = 1024; /*SQL buffer size*/
    char sql[sql_statement_size]; /*Buffer for SQL commands*/
    struct ias_db_query *query = NULL; /*DB query structure*/
    int status = SUCCESS;

    /* Find the orbit using the landsat interval id */
    status = snprintf(sql, sizeof(sql),
        "select ORBIT "
        "from LMD_MISSION "
        "where LANDSAT_CAL_INTERVAL_ID = '%s'",
        interval_id);
    if (status >= sizeof(sql) || status < 0)
    {
        IAS_LOG_ERROR("Unable to create SQL string");
        return ERROR;
    }

    query = ias_db_perform_query(dbase, sql);
    if (!ias_db_query_was_successful(query))
    {
        IAS_LOG_ERROR("Unable to find orbit using interval Id %s",
            interval_id);
        return ERROR;
    }

    if (!ias_db_query_next(query))
    {
        IAS_LOG_ERROR("No data returned from query %s", sql);
        ias_db_query_close(query);
        return ERROR;
    }

    if (ias_db_query_int_value(query, orbit, 0) != SUCCESS)
    {
        IAS_LOG_ERROR("Unable to retrieve orbit information");
        ias_db_query_close(query);
        return ERROR;
    }

    ias_db_query_close(query);
    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME: ias_rps_db_get_bias_model_residual_a0

PURPOSE: Obtain the r0 residual from the database.

RETURNS: An integer indicating SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int ias_rps_db_get_bias_model_residual_a0(
    /*I: database connection */
    struct ias_db_connection *dbase,
    /*I: Work order being processed */
    const char *work_order_id,
    /*I: Band to write to the db, 1 based value */
    const int band_number,
    /*I: Sca to write to the db, 1 based value */
    const int sca_number,
    /*I: Lineset 0/1 = PAN Odd/Even, 2=All */
    const int lineset,
    /*O: Residual squared value */
    double *r_squared
)
{
    const int sql_statement_size = 1024; /* SQL buffer size */
    char sql[sql_statement_size]; /* Buffer for sql commands */
    int status = SUCCESS;

    struct ias_db_query *query = NULL; /* DB query structure */

    memset(sql, 0, sizeof(sql) + 1);

    /* Find the r0 value */
    status = snprintf(sql, sizeof(sql),
        "SELECT r_square "
        "FROM bias_model_residuals "
        "WHERE radiometric_wo_common_id = '%s' "
        "AND BAND_ID = %i "
        "AND SENSOR_CHIP_ASSEMBLY_ID = %i "
        "AND LINESET = %i "
        "AND BPF_PARAMETER_ID = "
        "(SELECT BPF_PARAMETER_ID FROM BPF_PARAMETER WHERE NAME = "
        "'A0_Coefficient')",
        work_order_id, band_number, sca_number, lineset);
    if (status >= sizeof(sql) || status < 0)
    {
        IAS_LOG_ERROR("Unable to create SQL string");
        return ERROR;
    }
   
    query = ias_db_perform_query(dbase, sql);
    if (!ias_db_query_was_successful(query))
    {
        IAS_LOG_ERROR("Unable to perform query %s.", sql);
        return ERROR;
    }

    if (!ias_db_query_next(query))
    {
        IAS_LOG_ERROR("No data returned from query %s", sql);
        ias_db_query_close(query);
        return ERROR;
    }

    if (ias_db_query_double_value(query, r_squared, 0) != SUCCESS)
    {
        IAS_LOG_ERROR("Unable to retrieve r0 value");
        ias_db_query_close(query);
        return ERROR;
    }

    ias_db_query_close(query);

    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME: ias_rps_db_get_bias_model_residual_a1

PURPOSE: Obtain r1 values from the database.

RETURNS: An integer indicating SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int ias_rps_db_get_bias_model_residual_a1(
    /*I: database connection */
    struct ias_db_connection *dbase,
    /*I: Work order being processed */
    const char *work_order_id,
    /*I: Attributes for the band being processed */
    const IAS_BAND_ATTRIBUTES *band_attributes,
    /*I: Sca to write to the db, 1 based value */
    int sca_number,
    /*I: Lineset 0/1 = PAN Odd/Even lines, 2=All */
    int lineset,
    /*I: Residual squared value ascending order by detector number */
    double *r_squared
)
{
    const int sql_statement_size = 1024; /* SQL buffer size */
    char sql[sql_statement_size]; /* Buffer for sql commands */

    int i = 0;
    int status = SUCCESS;

    struct ias_db_query *query = NULL; /* DB query structure */

    memset(sql, 0, sizeof(sql) + 1);

    /* Find the r1 values*/
    status = snprintf(sql, sizeof(sql),
        "SELECT r_square, bpf_parameter_id "
        "FROM bias_model_residuals "
        "WHERE radiometric_wo_common_id = '%s' "
        "AND band_id = %i "
        "AND sensor_chip_assembly_id = %i "
        "AND lineset = %i "
        "AND bpf_parameter_id in "
        "(SELECT bpf_parameter_id FROM bpf_parameter WHERE name like "
        "'D___') "
        "ORDER BY bpf_parameter_id ASC",
        work_order_id, band_attributes->band_number, sca_number, lineset);
    if (status >= sizeof(sql) || status < 0)
    {
        IAS_LOG_ERROR("Unable to create SQL string");
        return ERROR;
    }

    query = ias_db_perform_query(dbase, sql);
    if (!ias_db_query_was_successful(query))
    {
        IAS_LOG_ERROR("Unable to obtain r1 value");
        return ERROR;
    }
    
    for (i = 0; i < band_attributes->detectors_per_sca; i++)
    {
        if (!ias_db_query_next(query))
        {
            IAS_LOG_ERROR("No data returned from query %s", sql);
            ias_db_query_close(query);
            return ERROR;
        }
 
        if (ias_db_query_double_value(query, &r_squared[i], 0) != SUCCESS)
        {
            IAS_LOG_ERROR("Unable to retrieve r1 value");
            ias_db_query_close(query);
            return ERROR;
        }
    }

    ias_db_query_close(query);

    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME: ias_rps_db_bmc_populate_start_stop
 
PURPOSE: Populate the start and stop dates for the BPF
 
RETURNS: An integer indicating SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int ias_rps_db_bmc_populate_start_stop(
    /*I: database connection */
    struct ias_db_connection *database,
    /*I: Workorder ID for the post collection*/
    const char *post_workorder,
    /*I: Workorder ID for the previous collection*/
    const char *pre_workorder,
    /*I/O: Bias parameters*/
    IAS_BPF *bpf
)
{
    const int sql_statement_size = 300; /*SQL buffer size*/
    char sql[sql_statement_size]; /*Buffer for SQL commands*/
    struct ias_db_query *query = NULL; /*DB query structure*/
    char date_acquired[IAS_L0R_DATE_LENGTH]; /*Date Acquired buffer*/
    int status = 0; /* return value variable */

    memset(sql, 0, sizeof(sql) + 1);

    /*Find the effective begin date using the last pre collected id*/
    status = snprintf(sql, sizeof(sql),
        "select DATE_ACQUIRED "
        "from RADIOMETRIC_WORK_ORDER_COMMON "
        "where RADIOMETRIC_WO_COMMON_ID = %s",
        pre_workorder);

    if (status >= sizeof(sql) || status < 0)
    {
        IAS_LOG_ERROR("Unable to create query string");
        return ERROR;
    }

    query = ias_db_perform_query(database, sql);
    if (!ias_db_query_was_successful(query))
    {
        IAS_LOG_ERROR("Unable to find date acquired using RWOC ID %s", 
            pre_workorder);
        return ERROR;
    }
    if (!ias_db_query_next(query))
    {
        IAS_LOG_ERROR("No data returned using RWOC ID %s", pre_workorder);
        ias_db_query_close(query);
        return ERROR;
    }

    ias_db_query_datetime_value(query, 0, date_acquired, IAS_L0R_DATE_LENGTH);
    IAS_LOG_DEBUG("PreId DATE_ACQUIRED is %s", date_acquired);

    /*load date in BPF struct*/
    ias_bpf_set_file_attributes_effective_date_begin(bpf, date_acquired);

    ias_db_query_close(query);

    /*Find the effective end date using the first post collected id*/
    status = snprintf(sql, sizeof(sql),
        "select DATE_ACQUIRED "
        "from RADIOMETRIC_WORK_ORDER_COMMON "
        "where RADIOMETRIC_WO_COMMON_ID = %s",
        post_workorder);

    if (status >= sizeof(sql) || status < 0)
    {
        IAS_LOG_ERROR("Unable to create query string");
        return ERROR;
    }

    query = ias_db_perform_query(database, sql);
    if (!ias_db_query_was_successful(query))
    {
        IAS_LOG_ERROR("Unable to find date acquired using RWOC ID %s", 
            post_workorder);
        return ERROR;
    }
    if (!ias_db_query_next(query))
    {
        IAS_LOG_ERROR("No data returned using RWOC ID %s", post_workorder);
        ias_db_query_close(query);
        return ERROR;
    }
    ias_db_query_datetime_value(query, 0, date_acquired, IAS_L0R_DATE_LENGTH);
    IAS_LOG_DEBUG("PostId DATE_ACQUIRED is %s", date_acquired);

    /*load date in BPF struct*/
    ias_bpf_set_file_attributes_effective_date_end(bpf, date_acquired);

    ias_db_query_close(query);

    IAS_LOG_DEBUG("Start/Stop populated");

    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME: ias_rps_db_set_bias_model_residual_a0

PURPOSE: Populate the database with the information provided to the routine.

RETURNS: An integer indicating SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int ias_rps_db_set_bias_model_residual_a0(
    /*I: database connection */
    struct ias_db_connection *dbase,
    /*I: Residual squared value */
    const double r_squared,
    /*I: Work order being processed */
    const char *work_order_id,
    /*I: Band to write to the db, 1 based value */
    const int band_number,
    /*I: Sca to write to the db, 1 based value */
    const int sca_number,
    /*I: Lineset 0/1 = PAN Odd/Even, 2=All */ 
    const int lineset
)
{
    const int sql_statement_size = 512; /* SQL buffer size */
    char sql[sql_statement_size]; /* Buffer for sql commands */
    int status = SUCCESS;
 
    struct ias_db_query *query = NULL; /* DB query structure */

    memset(sql, 0, sizeof(sql) + 1);

    /* Build the SQL statement */
    status = snprintf(sql, sizeof(sql),
        "INSERT INTO BIAS_MODEL_RESIDUALS "
        "( "
        "RADIOMETRIC_WO_COMMON_ID, "
        "BAND_ID, "
        "SENSOR_CHIP_ASSEMBLY_ID, "
        "LINESET, "
        "R_SQUARE, "
        "BPF_PARAMETER_ID "
        ") "
        "values "
        "(%s, %i, %i, %i, %f, "
        "(SELECT BPF_PARAMETER_ID FROM BPF_PARAMETER WHERE NAME = "
        " 'A0_Coefficient'))",
        work_order_id,
        band_number,
        sca_number,
        lineset,
        r_squared);
    
    if (status >= sizeof(sql) || status < 0) 
    {
        IAS_LOG_ERROR("Unable to create SQL string");
        return ERROR;
    }

    /* Perform the query */
    query = ias_db_perform_query(dbase, sql);
    if (!ias_db_query_was_successful(query))
    {
        IAS_LOG_ERROR("Unable to store the rSquared data in the "
            "BIAS_MODEL_RESIDUALS with query %s", sql);
        ias_db_query_close(query);
        return ERROR;
    }

    ias_db_query_close(query);

    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME: ias_rps_db_set_bias_model_residual_a1

PURPOSE: Populate the database with the information provided to the routine.

RETURNS: An integer indicating SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int ias_rps_db_set_bias_model_residual_a1(
    /*I: database connection */
    struct ias_db_connection *dbase,
    /*I: Residual squared value ascending order by detector number */
    double *residuals_squared,
    /*I: Work order being processed */
    const char *work_order_id,
    /*I: Attributes for the band being processed */
    const IAS_BAND_ATTRIBUTES *band_attributes,
    /*I: Sca to write to the db, 1 based value */
    int sca_number,
    /*I: Lineset 0/1 = PAN Odd/Even lines, 2=All */
    int lineset
)
{
    char *rwoc_id = NULL; /*Non-const copy of post work order id*/
    int *parameter_ids = NULL; /*BPF parameter ids*/
    int band_id = band_attributes->band_number;
    int status = SUCCESS;
    int table_size = 0;

    parameter_ids = calloc(band_attributes->detectors_per_sca, sizeof(int));
 
    if (parameter_ids == NULL)
    {
        IAS_LOG_ERROR("Unable to allocate array for parameter ids");
        return ERROR;
    }

    rwoc_id = strdup(work_order_id);
    if (rwoc_id == NULL)
    {
        IAS_LOG_ERROR("Unable to duplicate work order string");
        free(parameter_ids);
        return ERROR;
    }

    /* NOTE If the order of the parameter IDs for the detectors is not 
       sequential and in ascending order this query will not work */
    IAS_DB_QUERY_TABLE parameter_table[] =
    {
        {"BPF_PARAMETER_ID", parameter_ids, sizeof(parameter_ids[0]), 
            IAS_QUERY_INT_ARRAY, band_attributes->detectors_per_sca},
    };
    table_size  = sizeof(parameter_table)/sizeof(IAS_DB_QUERY_TABLE);
    status = ias_db_query_using_table(dbase, "BPF_PARAMETER", parameter_table,
        table_size, "NAME like 'D___' ORDER BY BPF_PARAMETER_ID ASC",
        band_attributes->detectors_per_sca);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("The query to obtain the bpf parameter ids failed");
        free(rwoc_id);
        free(parameter_ids);
        return ERROR;
    }

    IAS_DB_INSERT_TABLE table[]=
    {
        IAS_DB_TABLE_STRING("radiometric_wo_common_id", rwoc_id),
        IAS_DB_TABLE_INT("band_id", band_id),
        IAS_DB_TABLE_INT("sensor_chip_assembly_id", sca_number),
        IAS_DB_TABLE_INT("lineset", lineset),
        IAS_DB_TABLE_DOUBLE_ARRAY("R_SQUARE", residuals_squared),
        IAS_DB_TABLE_INT_ARRAY("bpf_parameter_id", parameter_ids)
    };

    IAS_LOG_DEBUG("Inserting R1 for RWOC %s band %d sca %d lineset %d", 
        rwoc_id, band_id, sca_number, lineset);

    /* Get the number of columns inserted into the table and insert the
       data into the database */
    table_size = sizeof(table) / sizeof(IAS_DB_INSERT_TABLE);
    status = ias_db_insert_using_table(dbase, "BIAS_MODEL_RESIDUALS", 
        table, table_size, band_attributes->detectors_per_sca);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Inserting A1 residuals squared for RWOC %s", 
            work_order_id);
        free(rwoc_id);
        free(parameter_ids);
        return ERROR;
    }

    free(rwoc_id);
    free(parameter_ids);
    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME: *ias_rps_bias_residuals_allocate

PURPOSE: Allocate a structure to hold the residuals from bias model
         calibration.

RETURNS: A pointer to the allocated structure or NULL on error
-----------------------------------------------------------------------------*/
IAS_RPS_DB_BIAS_MODEL_RESIDUALS *ias_rps_bias_residuals_allocate(void)
{
    int status;
    int i;
    int band_list[IAS_MAX_NBANDS];
    int num_bands;
    IAS_RPS_DB_BIAS_MODEL_RESIDUALS *residuals = calloc( 1,
        sizeof(IAS_RPS_DB_BIAS_MODEL_RESIDUALS));
    if ( ! residuals )
    {
        IAS_LOG_ERROR("Error allocating the residuals");
        return NULL;
    }

    /* Get a list and count of the normal OLI bands */
    status = ias_sat_attr_get_sensor_band_numbers( IAS_OLI, IAS_NORMAL_BAND,
        IAS_SECONDARY_BAND, band_list, IAS_MAX_NBANDS, &num_bands);
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Retrieving the list of normal bands");
        free( residuals );
        return NULL;
    }

    /* Loop through the bands allocating memory */
    for ( i = 0; i < num_bands; i++ )
    {
        IAS_RPS_DB_BAND_RESIDUAL *ptr;
        int band_index;
        int nscas;
        int band_number = band_list[i];
        int n_loops = 1;
        int n;
        const IAS_BAND_ATTRIBUTES *band_attributes =
            ias_sat_attr_get_band_attributes( band_number );
        if ( !band_attributes )
        {
            IAS_LOG_ERROR("Getting band attributes for band %d", band_number );
            ias_rps_bias_residuals_free( residuals );
            return NULL;
        }

        band_index = band_attributes->band_index;
        nscas = band_attributes->scas;
        /* Get a convenience pointer to the band structure */
        ptr = &residuals->band_residuals[band_index];

        /* Adjust the number of loops for the pan band and loop allocating
           one set for all non-pan and 2 sets for pan */
        if ( band_attributes->band_type == IAS_PAN_BAND )
            n_loops = 2;
        for ( n = 0; n < n_loops; n++ )
        {
            int sca;

            ptr->band_number = band_number;

            /* Allocate the a0 residual - there is one per sca */
            ptr->r_squared_a0[n] = malloc( sizeof(double) * nscas );
            if ( ! ptr->r_squared_a0[n] )
            {
                IAS_LOG_ERROR("Allocating a0 residuals for band %d",
                    band_number );
                ias_rps_bias_residuals_free( residuals );
                return NULL;
            }

            /* Allocate the a1 residuals - there are one per detector so
               allocate the SCA pointers here */
            ptr->r_squared_a1[n] = calloc( nscas, sizeof(double *));
            if ( ! ptr->r_squared_a1[n] )
            {
                IAS_LOG_ERROR("Allocating a1 residuals for band %d",
                    band_number );
                ias_rps_bias_residuals_free( residuals );
                return NULL;
            }

            /* and the detectors here */
            for ( sca = 0; sca < nscas; sca++ )
            {
                ptr->r_squared_a1[n][sca] =
                    malloc( band_attributes->detectors_per_sca *
                    sizeof(double));
                if ( ! ptr->r_squared_a1[n][sca] )
                {
                    IAS_LOG_ERROR("Allocating a1 sca residuals for band %d "
                        "SCA %d", band_number, sca + 1 );
                    ias_rps_bias_residuals_free( residuals );
                    return NULL;
                }
            }
        }
    }  /* END for band loop */

    return residuals;
}  /* END ias_rps_bias_residuals_allocate */

/*----------------------------------------------------------------------------
NAME: ias_rps_bias_residuals_free

PURPOSE: Frees the memory that was allocated by ias_rps_bias_residuals_allocate

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int ias_rps_bias_residuals_free
(
    IAS_RPS_DB_BIAS_MODEL_RESIDUALS *residuals
)
{
    int index;
    int band_list[IAS_MAX_NBANDS];
    int num_bands;
    int status;

    if ( residuals == NULL )
        return SUCCESS;

    /* Get a list of the normal OLI bands */
    status = ias_sat_attr_get_sensor_band_numbers( IAS_OLI, IAS_NORMAL_BAND,
        IAS_SECONDARY_BAND, band_list, IAS_MAX_NBANDS, &num_bands);
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Retrieving the list of normal bands");
        return ERROR;
    }

    for ( index = 0; index < num_bands; index++ )
    {
        int sca;
        int band_number = band_list[index];
        const IAS_BAND_ATTRIBUTES *band_attributes =
            ias_sat_attr_get_band_attributes( band_number );
        int band_index = band_attributes->band_index;
        IAS_RPS_DB_BAND_RESIDUAL *ptr = &residuals->band_residuals[band_index];
        int pan = 0;

        if ( band_attributes == NULL )
        {
            IAS_LOG_ERROR("Getting the band attributes for band %d",
                band_number );
            return ERROR;
        }

        if ( band_attributes->band_type == IAS_PAN_BAND )
            pan = 1;

        free( ptr->r_squared_a0[0] );
        if ( pan )
            free( ptr->r_squared_a0[1] );

        for ( sca = 0; sca < band_attributes->scas; sca++ )
        {
            free ( ptr->r_squared_a1[0][sca] );
            if ( pan )
                free ( ptr->r_squared_a1[1][sca] );
        }

        free( ptr->r_squared_a1[0] );
        if ( pan )
            free( ptr->r_squared_a1[1] );
    }

    free( residuals );
    return SUCCESS;
}  /* END ias_rps_bias_residuals_free */


/*-----------------------------------------------------------------------
 NAME:      ias_rps_db_create_oli_bias_model_where_clause

 PURPOSE:   Creates a where clause for an OLI bias model calibration
            query

 RETURNS:   Integer status code of SUCCESS or ERROR
-------------------------------------------------------------------------*/
int ias_rps_db_create_oli_bias_model_where_clause
(
    char       *where_clause_string,          /* O: Populated common part
                                                 of the total 'where'
                                                 clause */
    int        sizeof_where_clause_string,    /* I: Total size of the common
                                                 'where' clause (including
                                                 NULL terminator) */
    const char *source_table_name,            /* I: Name of source table */
    double     integration_time_oli_ms,       /* I: OLI MS band integration
                                                 time */
    double     integration_time_oli_pan       /* I: OLI PAN band integration
                                                 time */
)
{
    int status = snprintf(where_clause_string, sizeof_where_clause_string,
        "COLLECTION_TYPE='%s' AND VALID='Y' "
        "AND INTEGRATION_TIME_OLI_MS=%lf "
        "AND INTEGRATION_TIME_OLI_PAN=%lf "
        "AND RADIOMETRIC_WO_COMMON_ID IN "
        "  (SELECT DISTINCT RADIOMETRIC_WO_COMMON_ID "
        "   FROM %s) ",
        IAS_RPS_OLI_SHUTTER_COLLECT, integration_time_oli_ms,
        integration_time_oli_pan, source_table_name);
    if ((status < 0) || (status >= sizeof_where_clause_string))
    {
        IAS_LOG_ERROR("Constructing OLI bias model where clause string");
        return ERROR;
    }

    return SUCCESS;
}   /* END -- ias_rps_db_create_oli_bias_model_where_clause */


/*-----------------------------------------------------------------------
 NAME:      ias_rps_db_create_tirs_bias_model_where_clause

 PURPOSE:   Creates the where clause for a TIRS bias model calibration
            query

 RETURNS:   Integer status code of SUCCESS or ERROR
-------------------------------------------------------------------------*/
int ias_rps_db_create_tirs_bias_model_where_clause
(
    char       *where_clause_string,          /* O: Populated common part
                                                 of the total 'where'
                                                 clause */
    int        sizeof_where_clause_string,    /* I: Total size of the common
                                                 'where' clause (including
                                                 NULL terminator) */
    const char *source_table_name             /* I: Name of source table */
)
{
    int status = snprintf(where_clause_string, sizeof_where_clause_string,
        "COLLECTION_TYPE='%s' AND VALID='Y' "
        "AND RADIOMETRIC_WO_COMMON_ID IN "
        "  (SELECT DISTINCT RADIOMETRIC_WO_COMMON_ID "
        "   FROM %s) ",
        IAS_RPS_TIRS_DEEPSPACE_COLLECT, source_table_name);
    if ((status < 0) || (status >= sizeof_where_clause_string))
    {
        IAS_LOG_ERROR("Constructing TIRS bias model where clause string");
        return ERROR;
    }

    return SUCCESS;
}   /* END -- ias_rps_db_create_tirs_bias_model_where_clause */

