#include <string.h>
#include "ias_logging.h"
#include "ias_pixel_mask.h"
#include "ias_radiometry.h"


/*-------------------------------------------------------------------------
 NAME:     add_detector_list_to_mask

 PURPOSE:  Given an inoperable detector/out-of-spec detector list and the
           corresponding pixel mask value, adds the affected detector(s)
           to the pixel mask

 RETURNS:  Integer status of SUCCESS or ERROR
--------------------------------------------------------------------------*/
static int add_detector_list_to_mask
(
    const int *detector_status_list, /* I: List of inoperable or out-of-spec
                                        detectors */
    int number_of_detectors_in_list, /* I: Number of detectors in the
                                        inoperable/out-of-spec list */
    PIXEL_MASK_TYPE status_mask,     /* I: Status to be identified
                                        (IAS_RPS_INOP_DETECTOR,
                                        IAS_RPS_OUTOFSPEC_DETECTOR) */
    int band_number,                 /* I: Band number, 1-based */
    int sca_number,                  /* I: SCA number, 1-based */
    int number_of_sca_detectors,     /* I: Number of detectors in the SCA */
    int num_lines,                   /* I: Number of lines in the data */
    IAS_PIXEL_MASK *pixel_mask       /* I/O: Contains pixel artifacts */
)
{
    char status_list_type[12];
    int list_index;

    /* Create a string representing the status to be identified.  The
       current allowed choices are "inoperable" or "out-of-spec" */
    if (status_mask == IAS_RPS_INOP_DETECTOR)
    {
        strcpy(status_list_type, "inoperable");
    }
    else if (status_mask == IAS_RPS_OUTOFSPEC_DETECTOR)
    {
        strcpy(status_list_type, "out-of-spec");
    }
    else
    {
        IAS_LOG_ERROR("Invalid status mask %c", status_mask);
        return ERROR;
    }

    for (list_index = 0;
         list_index < number_of_detectors_in_list;
         list_index++)
    {
        int detector_number;
        int status;

        detector_number = detector_status_list[list_index];
        if (detector_number > number_of_sca_detectors || detector_number < 1)
        {
            IAS_LOG_ERROR("Invalid detector number %d found in band %d "
                "SCA %d %s detector list", detector_number, band_number,
                sca_number, status_list_type);
            return ERROR;
        }

        /* Subtract one from the detector number to make it an index,
           and add the detector to the mask */
        status = ias_pm_add_pixels(pixel_mask, detector_number - 1,
            0, num_lines, status_mask);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Adding detector %d with '%s' status to band %d"
                "SCA %d pixel mask", detector_number, status_list_type,
                band_number, sca_number);
            return ERROR;
        }
    }

    return SUCCESS;
}   /* END -- add_detector_list_to_mask */


/*----------------------------------------------------------------------------
NAME:    ias_rps_inoperable_detector_identification

PURPOSE: Identify the inoperable detectors from the input parameter list
         and set the pixel mask with the artifact.  Implemented as a wrapper
         function calling the internal 'add_detector_list_to_mask'
         function.

RETURNS: Integer status of SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int ias_rps_inoperable_detector_identification
(
    const int *inop_detector_list,     /* I: List of inoperable detectors
                                          from CPF (1-based) */
    int number_of_inoperable_detectors,/* I: Number of inoperable detectors in
                                          the list */
    int band_number,                   /* I: Band number, 1-based */
    int sca_number,                    /* I: SCA number, 1-based */
    int number_of_sca_detectors,       /* I: Number of detectors in the SCA */
    int number_of_lines,               /* I: Number of lines in the data */
    IAS_PIXEL_MASK *pixel_mask         /* I/O: Contains pixel artifacts */
)
{
    int status = add_detector_list_to_mask(inop_detector_list,
        number_of_inoperable_detectors, IAS_RPS_INOP_DETECTOR,
        band_number, sca_number, number_of_sca_detectors, number_of_lines,
        pixel_mask);

    return status;
}   /* END -- ias_rps_inoperable_detector_identification */


/*----------------------------------------------------------------------------
NAME:    ias_rps_outofspec_detector_identification

PURPOSE: Identify the out-of-spec detectors from the input parameter list
         and set the pixel mask with the artifact.  Implemented as a wrapper
         function calling the internal 'add_detector_list_to_mask' function.

RETURNS: Integer status of SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int ias_rps_outofspec_detector_identification
(
    const int *outof_spec_list,        /* I: List of out-of-spec detectors */
    int number_of_outofspec_detectors, /* I: Number of out-of-spec dets */
    int band_number,                   /* I: Current (1-based) band number */
    int sca_number,                    /* I: Current (1-based) SCA number */
    int number_of_sca_detectors,       /* I: Number of detectors in an SCA */
    int number_of_lines,               /* I: Number of lines of image data */
    IAS_PIXEL_MASK *pixel_mask         /* I/O: Mask with current set of
                                          identified artifacts */
)
{
    int status = add_detector_list_to_mask(outof_spec_list,
        number_of_outofspec_detectors, IAS_RPS_OUTOFSPEC_DETECTOR,
        band_number, sca_number, number_of_sca_detectors, number_of_lines,
        pixel_mask);

    return status;
}   /* END -- ias_rps_outfspec_detector_identification */
