/**************************************************************************
 NAME:     ias_rps_dropped_frame_characterization

 PURPOSE:  Source file implementing the routines needed to perform dropped
           frame characterization as a function in the IAS Radiometric
           Library

 ALGORITHM REFERENCES:  L0R line header descriptions are provided
                        in LDCM-DFCB-002, "Level 0 Reformatted Archive (L0Ra)
                        Data Format Control Book (DFCB)"
****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <limits.h>
#include "ias_l0r.h"
#include "ias_logging.h"
#include "ias_miscellaneous.h"
#include "ias_pixel_mask.h"
#include "ias_radiometry.h"
#include "ias_satellite_attributes.h"
#include "ias_const.h"



/*========================================================================
                         INTERNAL FUNCTIONS
==========================================================================*/
/*------------------------------------------------------------------------
 NAME:     get_frame_status_field

 PURPOSE:  Returns the frame_status value from the current frame header

 RETURNS:  Line status value if nonzero
           ERROR if line status value is 0

---------------------------------------------------------------------------*/
static int get_frame_status_field
(
    const void *l0r_frame_headers,   /* I: Frame header array */
    IAS_SENSOR_ID band_sensor_id,    /* I: Sensor identifier for the
                                        current band */
    int index,                       /* I: Current index into frame header
                                        array */
    uint16_t *frame_status           /* O: Composite value of frame_status
                                        field */
)
{
    /* Get the status field for the current frame header.
       Since we're dealing with a void pointer representing the
       array of frame headers, it needs to be recast to the proper
       type for the given sensor.

       NOTE:  The current field name for frame status information in both
              IAS_L0R_OLI_FRAME_HEADER and IAS_L0R_TIRS_FRAME_HEADER is
              frame_status.  For consistency and context, I'm calling this
              frame_status */
    if (band_sensor_id == IAS_OLI)
    {
        *frame_status =
            ((const IAS_L0R_OLI_FRAME_HEADER *)l0r_frame_headers)[index]
                .frame_status;
    }
    else if (band_sensor_id == IAS_TIRS)
    {
        *frame_status =
            ((const IAS_L0R_TIRS_FRAME_HEADER *)l0r_frame_headers)[index]
                .frame_status;
    }
    else
    {
        IAS_LOG_ERROR("Invalid sensor ID %d found--no known frame "
            "header information available", band_sensor_id);
        return ERROR;
    }

    return SUCCESS;
}   /* END get_frame_status_field */




/*========================================================================
                          RADIOMETRIC LIBRARY FUNCTIONS
==========================================================================*/
/*-------------------------------------------------------------------------
 NAME:            ias_rps_dropped_frame_characterization

 PURPOSE:  Main routine implementing dropped frame characterization

 RETURNS:  Integer status code of SUCCESS or ERROR
---------------------------------------------------------------------------*/
int ias_rps_dropped_frame_characterization
(
    const void *l0r_frame_headers,       /* I: Array of L0R frame headers
                                            for all video data lines */
    IAS_SENSOR_ID band_sensor_id,        /* I: Sensor ID for the current
                                            band/SCA */
    int  band_number,                    /* I: 1-based band number */
    int  number_detectors,               /* I: # of detectors per SCA */
    int  number_frames,                  /* I: Number of frames of image data */
    IAS_PIXEL_MASK  *pixel_mask          /* I: Pixel mask for the current
                                            band/SCA (could be
                                            populated with other artifact
                                            information)
                                            O: Pixel mask populated with
                                            dropped frame locations */
)
{
    int local_crc_error_count = 0;       /* Number of frames with CRC errors
                                            in the current band/SCA */
    int local_dropped_frame_count = 0;   /* Number of dropped frames in the
                                            current band/SCA */
    int det_index;                       /* Detector loop counter */
    int frame_index;                     /* Line counter */
    int status;
    uint16_t frame_status;               /* Composite frame status value in
                                            frame header */
    int current_span_count;              /* Temporary copy of the current
                                            span count */
    int span_count = 0;
    int start_span_index = 0;            /* Starting line index for a
                                            dropped frame span */
    int lines_per_frame;                 /* lines of imagery per frame for the
                                            current band */

    /* Get the number of lines per frame for this band */
    lines_per_frame = ias_sat_attr_get_lines_per_frame(band_number);
    if (lines_per_frame == ERROR)
    {
        IAS_LOG_ERROR("Obtaining the lines per frame for band number %d",
            band_number);
        return ERROR;
    }

    /* Check the frame_status value in the frame status information */
    for (frame_index = 0; frame_index < number_frames; frame_index++)
    {
        int line_index = frame_index * lines_per_frame;

        /* Get the frame status value */
        status = get_frame_status_field(l0r_frame_headers, band_sensor_id,
            frame_index, &frame_status);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Obtaining frame status for frame %d",
                (frame_index + 1));
            return ERROR;
        }

        /* Check frame_status for a CRC error */
        if ((frame_status & IAS_L0R_CRC_STATUS) != IAS_L0R_CRC_STATUS)
            local_crc_error_count++;

        /* Check frame_status for fill -- if there, the frame will be assumed
           to be dropped */
        if ((frame_status & IAS_L0R_FILLED) == IAS_L0R_FILLED)
        {
            /* Count this frame as "dropped".  Also flag it if it is the
               starting frame of a span.  Go to the next frame to
               see if it's part of this span, and keep going to subsequent
               frames until this span ends (i.e. the current frame
               is not a dropped frame) */
            local_dropped_frame_count++;
            if (span_count == 0)
                start_span_index = line_index;
            span_count += lines_per_frame;
            continue;
        }

        /* We have a complete dropped frame span.  Mark these frames in the
           pixel mask as "dropped" for all detectors; Reset the span count to 
           0 and keep checking the remaing frames */
        if (span_count != 0)
        {
            for (det_index = 0; det_index < number_detectors; det_index++)
            {
                /* use a copy of the span count in case it is adjusted */
                current_span_count = span_count;

                /* Write the span to the pixel mask */
                status = ias_pm_add_pixels(pixel_mask,
                    det_index, start_span_index, current_span_count,
                    IAS_RPS_DROPLINE);
                if (status != SUCCESS)
                {
                    IAS_LOG_ERROR("Writing dropped frame span of length %d "
                        "for detector %d", current_span_count,
                    (det_index + 1));
                    return ERROR;
                }
            }
            span_count = 0;
        }
    }

    /* A span including the last frame (if there is one) won't get written
       in the above loop.  Write it here */
    if (span_count != 0)
    {
        for (det_index = 0; det_index < number_detectors; det_index++)
        {
            /* use a copy of the span count in case it is adjusted */
            current_span_count = span_count;

            /* Write the span to the pixel mask */
            status = ias_pm_add_pixels(pixel_mask,
                det_index, start_span_index, current_span_count,
                IAS_RPS_DROPLINE);
            if (status != SUCCESS)
            {
                IAS_LOG_ERROR("Writing final dropped frame span of length "
                    "%d for detector %d", current_span_count,
                    (det_index + 1));
                return ERROR;
            }
        }
    }

    return SUCCESS;
}   /* END ias_rps_dropped_frame_characterization */
