/**************************************************************************
 NAME:      ias_rps_frame_averages.c

 PURPOSE:   Implementation of the ias_rps_frame_averages routines.

 ROUTINES:  ias_rps_calculate_frame_averages
            ias_rps_db_allocate_frame_averages
            ias_rps_db_free_frame_averages
            ias_rps_db_store_frame_averages

****************************************************************************/
#include <math.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include "ias_db_insert.h"
#include "ias_db_query.h"
#include "ias_logging.h"
#include "ias_pixel_mask.h"
#include "ias_radiometry.h"

/*----------------------------------------------------------------------------
NAME: ias_rps_calculate_frame_averages

PURPOSE: Routine used to compute the cross-track averages for a line of data.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int ias_rps_calculate_frame_averages(
    const void *input_buffer,      /* I: One SCA of data */
    const IAS_DATA_TYPE data_type, /* I: Data type may be float or uint16_t */
    const int band_number,         /* I: Band number, 1-based */
    const int sca_number,          /* I: SCA number, 1-based */
    const int num_detectors,       /* I: Number of detectors in the SCA */
    const int starting_line_number,/* I: Number of the first line, 1 based */
    const int num_lines,           /* I: Number of lines in the data */
    const int frames_to_skip_top,  /* I: Number of frames to skip at the 
                                       top (beginning) of image */
    const int frames_to_skip_bottom,  /* I: Number of frames to skip at the 
                                       bottom (end) of image */
    IAS_PIXEL_MASK *pixel_mask, /* I: Contains pixel artifacts */
    /* I/O: Pointer to results of the calculation. Allocate structure using: 
       ias_rps_db_allocate_frame_averages */
    IAS_RPS_DB_FRAME_AVERAGES *averages
)
{
    int line = 0; /*line iterator*/
    int detector = 0; /*detector iterator*/
    uint16_t *uint16_t_ptr = NULL;/* correctly. Depending on input data type */
    float *float_ptr = NULL;    /* Pointers used to access the input buffer */

    PIXEL_MASK_TYPE artifacts_to_ignore; /* Mask values of pixels to ignore */
    PIXEL_MASK_TYPE *line_mask; /* pixel mask used with the frame average calc*/

    /* The only data types acceptable are float and uint_16 */
    if ( data_type != IAS_R4 && data_type != IAS_UI2 )
    {
        IAS_LOG_ERROR("Unsupported data type: %d", data_type );
        return ERROR;
    }

    if (averages == NULL)
    {
        IAS_LOG_ERROR("NULL pointer received");
        return ERROR;
    }

    /*assign the pointers based on the type provided */
    if (data_type == IAS_R4)
    {
        float_ptr = (float*)input_buffer;
    }
    else
    {
        uint16_t_ptr = (uint16_t*)input_buffer;
    }       

    artifacts_to_ignore = IAS_RPS_DROPLINE| IAS_RPS_IMPNOISE |
        IAS_RPS_ADCONV_SAT_LO | IAS_RPS_ADCONV_SAT_HI | IAS_RPS_ANALOG_SAT_LO |
        IAS_RPS_ANALOG_SAT_HI | IAS_RPS_INOP_DETECTOR;

    line_mask = malloc(sizeof(PIXEL_MASK_TYPE) * num_detectors * num_lines);
    if (line_mask == NULL)
    {
        IAS_LOG_ERROR("Unable to allocate memory for the mask");
        return ERROR;
    }

    /* to support operations by line efficiently the entire mask must be loaded
        to avoid a performance impact of loading per line */
    if (SUCCESS != ias_pm_get_image_subset(pixel_mask, 0, 0,
        num_detectors, num_lines, line_mask))
    {
       IAS_LOG_ERROR("Unable to obtain mask for line %d", line);
       free(line_mask);
       return ERROR;
    }

    /*compute the frame averages*/
    for (line = 0; line < num_lines; line++)
    { 
        double sum = 0.0;
        int elements = num_detectors;

        /* tally based on the data type provided */ 
        if (data_type == IAS_R4)
        {
            for (detector = 0; detector < num_detectors; detector++)
            {
                /*Check if the pixel should be included by the mask*/
                if ( line_mask[(detector * num_lines) + line] & 
                    artifacts_to_ignore )
                { 
                    /*reduce the number of elements used in the average*/
                    elements--;
                    continue;
                }
    
                sum += float_ptr[(num_detectors * line) + detector];
            }
        }
        else
        {
            for (detector = 0; detector < num_detectors; detector++)
            {
                /*Check if the pixel should be included by the mask*/
                if ( line_mask[(detector * num_lines) + line] &
                    artifacts_to_ignore )
                {
                    /*reduce the number of elements used in the average*/
                    elements--;
                    continue;
                }

                sum += uint16_t_ptr[(num_detectors * line) + detector];
            }
        }

        averages->line_number[line] = starting_line_number + line;
 
        /* Assign the average to zero if divide by 0 would occur */ 
        if (elements == 0)
        { 
            averages->line_average[line] = 0;
        }
        else 
        {
            averages->line_average[line] = sum / elements;
        }
    }

    free(line_mask);
    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME: ias_rps_db_allocate_frame_averages

PURPOSE: Allocate storage for the cross track averages.

RETURNS: Allocated structure for the frame averages. NULL is returned if the
    allocation is unable to take place.
-----------------------------------------------------------------------------*/
IAS_RPS_DB_FRAME_AVERAGES* ias_rps_db_allocate_frame_averages
(
    const int number_of_lines /* I: Number of array elements to allocate. */
)
{
    IAS_RPS_DB_FRAME_AVERAGES *averages = 
        calloc(1, sizeof(IAS_RPS_DB_FRAME_AVERAGES));
    
    if (averages == NULL)
    {
        IAS_LOG_ERROR("Unable to allocate IAS_RPS_DB_FRAME_AVERAGES");
        return NULL;
    }

    //allocate the averages for the number of averages in the structure
    averages->line_number = calloc(number_of_lines, sizeof(int));
    if (averages->line_number == NULL)
    {
        IAS_LOG_ERROR("Unable to allocate IAS_RPS_DB_FRAME_AVERAGES");
        free(averages);
        return NULL;
    }

    averages->line_average = calloc(number_of_lines, sizeof(double));

    if (averages->line_average == NULL)
    {
        IAS_LOG_ERROR("Unable to allocate IAS_RPS_DB_FRAME_AVERAGES");
        free(averages->line_number);
        free(averages);
        return NULL;
    }

    averages->num_lines = number_of_lines;
    return averages; 
}

/*----------------------------------------------------------------------------
NAME: ias_rps_db_free_frame_averages

PURPOSE: Release storage used by the IAS_RPS_DB_FRAME_AVERAGES structure.

NOTE: The address of the pointer is used to set the pointer to NULL after being 
    released.
-----------------------------------------------------------------------------*/
void ias_rps_db_free_frame_averages
(
    IAS_RPS_DB_FRAME_AVERAGES **averages_ptr/*I:address of the pointer to free*/
)
{
    IAS_RPS_DB_FRAME_AVERAGES *averages = *averages_ptr;

    /* if the structure is already null there is nothing to be released */
    if (averages == NULL) 
    {
        return;
    }

    if (averages->line_number != NULL) 
    {
        free(averages->line_number);
        averages->line_number = NULL;
    }

    if (averages->line_average != NULL)
    {
        free(averages->line_average);
        averages->line_average = NULL;
    }

    free(*averages_ptr);
    *averages_ptr = 0;
}

/*----------------------------------------------------------------------------
NAME: ias_rps_db_store_frame_averages

PURPOSE: Store the averages in the database as provided in the input structure.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int ias_rps_db_store_frame_averages
(
    void *message /* I: Pointer to IAS_RPS_DB_FRAME_AVERAGES
                         structure.  A void pointer is used since it is
                         called from the threading routines and a void
                         pointer is used to support passing messages to
                         different routines */
)
{
    int status = SUCCESS;    /* function return status */
    int number_db_columns;   /* Number of database columns being inserted */

    IAS_RPS_DB_FRAME_AVERAGES *averages;

    if (message == NULL)
    {
        IAS_LOG_ERROR("NULL pointer received for data");
        return ERROR;        
    }

    /* cast the void pointer to the type used */
    averages = (IAS_RPS_DB_FRAME_AVERAGES*)message;

    /* Create the table */
    IAS_DB_INSERT_TABLE table[]=
    {
        IAS_DB_TABLE_STRING( "radiometric_wo_common_id", 
            averages->rad_common_id),
        IAS_DB_TABLE_INT( "band_id", averages->band_number),
        IAS_DB_TABLE_INT( "sensor_chip_assembly_id", averages->sca),
        IAS_DB_TABLE_INT_ARRAY( "line_number", averages->line_number),
        IAS_DB_TABLE_DOUBLE_ARRAY( "bf_average", averages->line_average),
    };

    /* Get the number of columns inserted into the table and insert the
       data into the database */
    number_db_columns = sizeof(table) / sizeof(IAS_DB_INSERT_TABLE);
    status = ias_db_insert_using_table( averages->dbase, 
         "bias_frame_averages", table, number_db_columns, 
        averages->num_lines);

    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Unable to insert the line averages for %d, "
            "sca %d", averages->band_number, averages->sca );
    }

    return status;
}

/*----------------------------------------------------------------------------
NAME: ias_rps_db_get_frame_averages

PURPOSE: Populate the structure with data obtained from the database.

NOTE: The database, common work order id, band number, and sca number are 
    required to be populated in the input struct.

RETURNS: An allocated IAS_RPS_DB_FRAME_AVERAGES*, NULL if error.
-----------------------------------------------------------------------------*/
IAS_RPS_DB_FRAME_AVERAGES* ias_rps_db_get_frame_averages
(
    struct ias_db_connection *dbase, /*I: database connection */
    const char *rad_common_id, /*I: common work order ID */
    const int band_number,  /*I: Band number, 1-based */
    const int sca_number    /*I: SCA number, 1-based */ 
)
{
    const int max_where_clause = 128;
    char where_clause[max_where_clause];
    int num_lines = 0;
    int status = SUCCESS;    /* function return status */
    int table_size = 0;
    IAS_RPS_DB_FRAME_AVERAGES *averages = NULL;

    if (dbase == NULL)
    {
        IAS_LOG_ERROR("NULL pointer received for database connection");
        return NULL;
    }

    /* Obtain the number of lines that will be in the structure */
    IAS_DB_QUERY_TABLE count_table[] =
    {
        {"COUNT(*)", &num_lines, sizeof(num_lines),
            IAS_QUERY_INT, 1}
    };

    snprintf(where_clause, max_where_clause,
        "RADIOMETRIC_WO_COMMON_ID = %s AND BAND_ID = %d "
        "AND SENSOR_CHIP_ASSEMBLY_ID = %d",
        rad_common_id, band_number, sca_number);

    table_size = sizeof(count_table)/sizeof(IAS_DB_QUERY_TABLE);
    status = ias_db_query_using_table(dbase, "BIAS_FRAME_AVERAGES",
        count_table, table_size, where_clause, 1);

    if (num_lines < 1) 
    {
        IAS_LOG_ERROR("No entries exist for work order %s, band %d, sca %d",
            rad_common_id, band_number, sca_number);
        return NULL;
    }

    /* Allocate the structure to hold the database results */
    averages = ias_rps_db_allocate_frame_averages(num_lines);

    if (averages == NULL)
    {
        IAS_LOG_ERROR("Unable to allocate frame averages");
        return NULL;
    }

    averages->dbase = dbase;
    strncpy(averages->rad_common_id, rad_common_id, 
        sizeof(averages->rad_common_id));
    averages->rad_common_id[sizeof(averages->rad_common_id) - 1] = '\0';
    averages->band_number = band_number;
    averages->sca = sca_number;

    /* Load the data into the recently allocated space from the database */
    IAS_DB_QUERY_TABLE qtable[] =
    {
        {"line_number", averages->line_number, sizeof(averages->line_number[0]),
            IAS_QUERY_INT_ARRAY, averages->num_lines}, 
        {"bf_average", averages->line_average, 
            sizeof(averages->line_average[0]), IAS_QUERY_DOUBLE_ARRAY, 
            averages->num_lines},
    };

    /* Read the data by ascending line numbers to put the smallest number 
        at index 0 */
    snprintf(where_clause, max_where_clause,
        "RADIOMETRIC_WO_COMMON_ID = %s AND BAND_ID = %d "
        "AND SENSOR_CHIP_ASSEMBLY_ID = %d ORDER BY LINE_NUMBER ASC",
        rad_common_id, band_number, sca_number);
    if (status < 0 || status >= max_where_clause)
    {
        IAS_LOG_ERROR("Unable to create the where clause for query");
        free(averages);
        return NULL;
    }

    table_size  = sizeof(qtable)/sizeof(IAS_DB_QUERY_TABLE);
    status = ias_db_query_using_table(averages->dbase, "BIAS_FRAME_AVERAGES",
        qtable, table_size, where_clause, averages->num_lines);
 
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Unable to perform query on the BIAS_FRAME_AVERAGES "\
            "table");
        free(averages);
        return NULL;
    }
   
    return averages;
}
