/************************************************************************
 NAME:       ias_rps_get_invalid_data_flags

 PURPOSE:    Takes the pixel mask values for the current band/SCA and writes
             them to an array

 RETURNS:    Integer status code of SUCCESS or ERROR

 NOTES:      This routine creates a 1D array of mask values for all of the
             detectors in a band/SCA.  Applications needing to perform
             detector-level validity checking (i.e. is a detector flagged
             with any artifact) can do this checking more efficiently using
             this array.  "Valid" data values are indicated by a flag
             value of 0; "invalid" values are indicated by a nonzero flag
             value.

             The flag buffer was chosen to be of type 'int' to allow future
             use of an expanded pixel mask without modification here.
**************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ias_pixel_mask.h"
#include "ias_logging.h"
#include "ias_radiometry.h"


int ias_rps_get_invalid_data_flags
(
    IAS_PIXEL_MASK *current_pm,          /* I: Pixel mask for the current
                                            band/SCA */
    PIXEL_MASK_TYPE artifact_flags_to_get,
                                         /* I: Artifact flag(s) to get.
                                            Multiple artifacts can be
                                            specified by using the
                                            bitwise OR operator (e.g
                                            IAS_RPS_INOP_DETECTOR |
                                            IAS_RPS_OUTOFSPEC_DETECTOR).
                                            By using ~IAS_RPS_NO_VALUE,
                                            all currently known artifact
                                            flags can be retrieved. */
    int buffer_size,                     /* I: Size of the flag buffer which
                                            is number of detectors */
    int *flag_buffer                     /* O: Populated flag buffer, all
                                            detectors for the current band/SCA
                                            are set -- previous contents
                                            overwritten */
)
{
    IAS_PIXEL_MASK_ITERATOR *it;
    IAS_PIXEL_MASK_SPAN span;
    int num_detectors;
    int detector_index;

    /* Ensure buffer size is correct for this band/SCA */
    num_detectors = ias_pm_get_number_of_detectors( current_pm );
    if ( num_detectors != buffer_size )
    {
        IAS_LOG_ERROR("Flag buffer is not the correct size, %d, for this "
            "date.  Should be %d", buffer_size, num_detectors );
        return ERROR;
    }

    /* initialize the flag buffer */
    memset( flag_buffer, 0, num_detectors * sizeof(*flag_buffer));

    /* Create an iterator for the specified artifact(s) to flag. */
    it = ias_pm_get_iterator(current_pm, IAS_PM_INCLUDE,
        artifact_flags_to_get);
    if (it == NULL)
    {
        IAS_LOG_ERROR("Getting mask iterator");
        return ERROR;
    }

    /* Walk the mask.  Mark any artifact as invalid */
    while (ias_pm_get_next_span(it, &span))
    {
        detector_index = span.detector_index;
        flag_buffer[detector_index] = 1;
    }

    /* Destroy the iterator. */
    ias_pm_destroy_iterator(it);

    return SUCCESS;
}
