/*----------------------------------------------------------------------------
NAME:    ias_rps_histogram_statistics_characterization

PURPOSE: General purpose routine to compute per-SCA statistics: min, max,
         mean, standard deviation, skewness kurtosis, neighbor pixel
         correlation and mean squared response.  There is also the ability
         to have an array of structures returned so that cummulative
         statistics can be calculated from separate calls to the algorithm.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <math.h>
#include <float.h>

#include "ias_const.h"
#include "ias_db_query.h"
#include "ias_types.h"
#include "ias_math.h"
#include "ias_logging.h"
#include "ias_miscellaneous.h"
#include "ias_pixel_mask.h"
#include "ias_satellite_attributes.h"
#include "ias_db_insert.h"
#include "ias_radiometry.h"

/* This is the structure used for keeping track of report information */
typedef struct hist_data
{
    int num_values;
    double min;
    double max;
    double mean;
    double stdev;
    double skewness;
    double kurtosis;
    int num_valid_frames;
    double correlation;
    double mean_sq_resp;
} HIST_DATA;

/* Prototypes for report generation */
static FILE *create_report_file( int, int, const char *, IAS_REPORT * );
static void close_report_file( FILE * );
static int write_report_file( FILE *report, int det, int pan, HIST_DATA * );


int ias_rps_histogram_statistics_characterization
(
    void *input_buffer,         /* I: One SCA of data */
    IAS_DATA_TYPE data_type,    /* I: Data type may be float or uint16_t */
    int band,                   /* I: Band number, 1-based */
    int sca,                    /* I: SCA number, 1-based */
    int num_detectors,          /* I: Number of detectors in the SCA */
    int num_lines,              /* I: Number of lines in the data */
    int *detector_offsets,      /* I: Integer detector offsets with most
                                      negative offset shifted to zero */
    int frames_to_skip_top,     /* I: Number of frames to skip at the 
                                       top (beginning) of image */
    int frames_to_skip_bottom,  /* I: Number of frames to skip at the 
                                       bottom (end) of image */
    IAS_PIXEL_MASK *pixel_mask, /* I: Contains pixel artifacts */
    IAS_REPORT *report_info,    /* I: General information for the test
                                       report.  If no report is desired pass
                                       NULL. */
    /* I/O: Pointer to results of the characterization.  If results are
       desired pass, an allocated structure or pass NULL if database output
       is not desired.  Allocate structure using:
       ias_rps_db_allocate_histogram_results */
    IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS *characterization_results,

    /* I/O: Results of some calculations to allow combining histogram
            statistics from multiple runs.  If they are not desired pass
            NULL.  The memory sould be allocated as one block of memory as
            (IAS_HISTOGRAM_INTERMEDIATE_RESULTS * num_detectors) and the
            memory accessed as intermediate_results[detector] */
    IAS_RPS_HISTOGRAM_INTERMEDIATE_RESULTS *intermediate_results
)
{
    PIXEL_MASK_TYPE artifacts_to_ignore;/* Mask values of pixels to ignore */
    PIXEL_MASK_TYPE *left_mask; /* Line of pixel masks from image */
    PIXEL_MASK_TYPE *right_mask;/* Line of pixel masks from image */
    PIXEL_MASK_TYPE *mask_swap; /* Pointer to allow left and right mask swap */
    int status;                 /* Routine return status */
    int left_det_num;           /* Detector number for the left detector */
    int right_det_num;          /* Detector number for the right detector */
    int line;                   /* Loop counter for each line */
    float *float_ptr = NULL;    /* Pointers used to access the current */
                                /* left detector data buffer correctly. */
    uint16_t *uint16_t_ptr = NULL;   /* Depending on input data type */
    float *neighbor_float_ptr = NULL;/* Pointers used to access the neighbor */
                                     /* right detector data buffer correctly.*/
    uint16_t *neighbor_uint16_t_ptr = NULL;/*Depending on input data type */
    int line_stride;            /* Stride of the array used to increment the
                                   input data pointers. */
    int pan_band;               /* Flag to indicate this is pan band */
    int neighbor_is_valid = FALSE; /* Flag indicating if neighbor pixel is
                                       valid */
    int starting_line;          /* This is the line to start on, it includes
                                   the number of lines to skip at the
                                   beginning. */
    int ending_line;            /* This is the line to end on, it includes
                                   the number of lines to skip at the end */
    int step_size;              /* Either 2 for pan band or 1 for others */
    int number_line_loops;      /* Either 2 for pan band or 1 for others */
    int line_index;             /* Plain old loop counter */
    FILE *report_file = NULL;   /* File handle for the report */
    const IAS_BAND_ATTRIBUTES *band_attributes;
    HIST_DATA *output_report_data = NULL; /* Report file information */


    /* The remainder of the following variables are the calculated
       statistical measures and intermediate results */
    double kurtosis;
    double maximum;
    double mean;
    double mean_squared_response;
    double minimum;
    double neighbor_correlation;
    double one_over_n;
    double skewness;
    double standard_deviation;
    double kurtosis_denom;
    double skewness_denom;
    double std_dev_sq;
    double sum_of_adjacent_products;
    /* Depending on the type of data and the number of lines, the following
       two intermediate values can become quite large.  If, in practice, this
       becomes a problem they can be changed to long doubles.  There is a
       small performance hit -- on the order of 4 seconds wall clock time
       for a standard WRS scene size of data. */
    double sum_of_cubes;
    double sum_of_quartics;
    double sum_of_squares;
    double sum_of_squares_with_valid_neigh;
    double sum_of_values;
    int number_valid_adjacent_products;
    int number_valid_values;


    /* The only data types acceptable are float and uint_16 */
    if ( data_type != IAS_R4 && data_type != IAS_UI2 )
    {
        IAS_LOG_ERROR("Unsupported data type: %d", data_type );
        return ERROR;
    }

    /* Determine if we're operating on the pan band */
    pan_band = FALSE;
    band_attributes = ias_sat_attr_get_band_attributes( band );
    if ( band_attributes->band_type == IAS_PAN_BAND )
        pan_band = TRUE;

    /* Get a file handle for the report, this also allocates memory */
    if ( report_info )
    {
        output_report_data = malloc( sizeof(HIST_DATA) * 2 );
        if ( !output_report_data )
        {
            IAS_LOG_ERROR("Allocating report data");
            return ERROR;
        }
        report_file = create_report_file( band, sca, 
                report_info->rps_level, report_info );
        if ( !report_file )
            return ERROR;
    }

    /* Create a convenience mask of artifacts to ignore */
    artifacts_to_ignore = IAS_RPS_DROPLINE|IAS_RPS_IMPNOISE|
                          IAS_RPS_ADCONV_SAT_LO|IAS_RPS_ADCONV_SAT_HI|
                          IAS_RPS_ANALOG_SAT_LO|IAS_RPS_ANALOG_SAT_HI;

    /* Allocate the left and right lines of pixel masks */
    left_mask = malloc( sizeof(PIXEL_MASK_TYPE) * num_lines );
    right_mask = malloc( sizeof(PIXEL_MASK_TYPE) * num_lines );
    if ( !left_mask || !right_mask )
    {
        IAS_LOG_ERROR("Allocating pixel mask memory");
        free( output_report_data );
        free( left_mask );
        free( right_mask );
        close_report_file( report_file );
        return ERROR;
    }
   
    /* For efficiency and ease of use we want an array of pixel masks.  For
       the same reason the right data buffer was filled, fill the right mask. */
    status = ias_pm_get_image_subset( pixel_mask, 0, 0, 1, num_lines,
        right_mask );
    if ( status )
    {
        IAS_LOG_ERROR("Retrieving pixel mask subset");
        free( output_report_data );
        free( left_mask );
        free( right_mask );
        close_report_file( report_file );
        return ERROR;
    }

    /* Loop through the detectors */
    for ( left_det_num = 0, right_det_num = 1; left_det_num < num_detectors;
        left_det_num++, right_det_num++ )
    {
        /* Swap the left and right mask pointers */
        mask_swap = left_mask;
        left_mask = right_mask;
        right_mask = mask_swap;

        /* Fill the ``no longer valid'' right mask with the next detectors
           mask values. */
        if ( right_det_num != num_detectors )
        {
            status = ias_pm_get_image_subset( pixel_mask, right_det_num, 0, 1,
                num_lines, right_mask );
            if ( status )
            {
                IAS_LOG_ERROR("Retrieving pixel mask subset");
                free( output_report_data );
                free( left_mask );
                free( right_mask );
                close_report_file( report_file );
                return ERROR;
            }
        }

        /* At this point we have two valid data and mask buffers.  Image
           statistics will be calculated on the left buffer.  The right
           buffer is only used for the neighboring pixel correlation.
           
           This was designed as a single pass (as in one pass through the
           data).  Typically moment calculations are two-pass algorithms but
           at the time of implementation, a requirement existed for the caller
           to be able to combine multiple intermediate results.  To this end
           the ``standard'' formulae for the moments were expanded; like
           terms collected and simplified.  The terms needed for the
           intermediate results were identified allowing for their collection
           in a single data pass.
           
           With the exception of the pan band, only loop once and step
           through the lines one at a time.  The pan band will execute the
           loop twice first the odd lines, then the even lines. */
        step_size = 1;
        number_line_loops = 1;
        if ( pan_band )
        {
            step_size = 2;
            number_line_loops = 2;
        }

        /* Set the start and end line based on the input frames_to_skip_top
           and frames_to_skip_bottom values */
        starting_line = frames_to_skip_top;
        ending_line = (num_lines - 1) - frames_to_skip_bottom;

        /* This shouldn't happen but, if it ever does, correct it to point
           to the first line and output a warning. */
        if ( starting_line < 0 )
        {
            IAS_LOG_WARNING("Starting overlap line of %d was calculated and "
                "corrected to 0 for Band %d, SCA %d", starting_line, 
                band, sca );
            starting_line = 0;
        }

        /* This shouldn't happen but, if it ever does, correct it to point
           to the last line and output a warning. */
        if ( ending_line < 0 )
        {
            IAS_LOG_WARNING("Ending overlap line of %d was calculated and "
                "corrected to %d for Band %d, SCA %d", ending_line,
                num_lines - 1, band, sca );
            ending_line = num_lines - 1;
        }

        /* This shouldn't happen but, if it ever does, fatal error */
        if ( starting_line > num_lines - 1 || ending_line < starting_line )
        {
            IAS_LOG_ERROR("Incorrected staring_line or ending _line was "
                          "calculated and for Band %d, SCA %d", band, sca );
            return ERROR;
        }

        /* Determine the array span.  This could be one or two lines of data
           depending on the step size. */
        line_stride = num_detectors * step_size;

        /* This loop is for the pan band exception.  Increment the starting
           line number on the second iteration to do the even pan band lines. */
        for ( line_index = 0; line_index < number_line_loops;
              line_index++, starting_line++ )
        {
            /* The first phase is to loop through the lines and calculate the
               intermediate statistics. */
            number_valid_values = number_valid_adjacent_products = 0;
            sum_of_values = sum_of_squares = sum_of_cubes = sum_of_quartics = 
                0.0;
            sum_of_adjacent_products = sum_of_squares_with_valid_neigh = 0.0;
            maximum = -DBL_MAX;
            minimum =  DBL_MAX;

            /* Point to the first line of this detector */
            if ( data_type == IAS_R4 )
            {
                float_ptr = ( (float *)input_buffer
                    + ( starting_line * num_detectors )) + left_det_num;
                if (right_det_num != num_detectors)
                {
                    neighbor_float_ptr = ((float *)input_buffer + 
                            (starting_line + detector_offsets[right_det_num] - 
                             detector_offsets[left_det_num]) * 
                             num_detectors) + right_det_num;
                }
            }
            else
            {
                uint16_t_ptr = ( (uint16_t *)input_buffer
                    + ( starting_line * num_detectors )) + left_det_num;
                if (right_det_num != num_detectors)
                {
                    neighbor_uint16_t_ptr = ( (uint16_t *)input_buffer +
                            (starting_line + detector_offsets[right_det_num] - 
                             detector_offsets[left_det_num]) * 
                             num_detectors) + right_det_num;
                }
            }

            for ( line = starting_line; line <= ending_line; line += step_size )
            {
                double pixel_value;
                double neighbor_value;
                int right_detector_line = 0;
                /* This is used for the running calculation of pixel_value,
                   square, cube and quartic. */
                double product;

                if (right_det_num != num_detectors)
                {
                    right_detector_line = line + 
                        detector_offsets[right_det_num] - 
                        detector_offsets[left_det_num];
                }

                /* Determine if this pixel is not valid and go to the next
                   iteration if it's not valid. */
                if ( left_mask[line] & artifacts_to_ignore )
                {
                    /* The data pointers need to be updated if the rest of
                       the loop is skipped. */
                    if ( data_type == IAS_R4 )
                    {
                        float_ptr += line_stride;
                        if (right_det_num != num_detectors)
                            neighbor_float_ptr += line_stride;
                    }                        
                    else
                    {
                        uint16_t_ptr += line_stride;
                        if (right_det_num != num_detectors)
                            neighbor_uint16_t_ptr += line_stride;
                    }
                    continue;
                }
                number_valid_values++;

                /* Check to see if there is a neighbor to the right.  If
                   there is, check it for validity, and if it is valid, set
                   the flag and count it. */
                if ( right_det_num < num_detectors )
                {
                    /* The order of operations is important here.  If the
                       right detector is not within the starting and ending
                       range we want the if to short-circuit before
                       accessing the mask. */
                    if ( right_detector_line < starting_line
                         || right_detector_line > ending_line
                         || ( right_mask[right_detector_line]
                              & artifacts_to_ignore ))
                    {
                        neighbor_is_valid = FALSE;
                    }
                    else
                    {
                        neighbor_is_valid = TRUE;
                        number_valid_adjacent_products++;
                    }
                }
                else
                {
                    neighbor_is_valid = FALSE;
                }

                /* Extract the pixel value and neighbor value, if the
                   neighbor is valid, from the proper representation of
                   input buffer.  Also to reduce the number of if blocks in
                   this for loop, move the pointer to the next line.  Note
                   that the line count may be two. */
                if ( data_type == IAS_R4 )
                {
                    pixel_value = (double)( *float_ptr );
                    if ( neighbor_is_valid )
                        neighbor_value = (double)( *neighbor_float_ptr );
                    else
                        neighbor_value = 0.0;
                    float_ptr += line_stride;
                    if (right_det_num != num_detectors)
                        neighbor_float_ptr += line_stride;
                }
                else
                {
                    pixel_value = (double)( *uint16_t_ptr );
                    if ( neighbor_is_valid )
                        neighbor_value = (double)( *neighbor_uint16_t_ptr );
                    else
                        neighbor_value = 0.0;
                    uint16_t_ptr += line_stride;
                    if (right_det_num != num_detectors)
                        neighbor_uint16_t_ptr += line_stride;
                }
                sum_of_values += pixel_value;

                if( pixel_value > maximum )
                    maximum = pixel_value;
                if( pixel_value < minimum )
                    minimum = pixel_value;

                product = pixel_value * pixel_value;
                sum_of_squares += product;
                if ( neighbor_is_valid )
                {
                    sum_of_squares_with_valid_neigh += product;
                    sum_of_adjacent_products += pixel_value * neighbor_value;
                }
                sum_of_cubes    += product * pixel_value;
                sum_of_quartics += product * product;
            }  /* END for line loop */

            /* Phase two is to use the intermediate values to ...
               Calculate the mean and standard deviation statistics as well as
               some shortcut intermediate values for simplicity and
               efficiency */
            if ( number_valid_values > 0 )
            {
                one_over_n = 1.0 / number_valid_values;
                mean = one_over_n * sum_of_values;
                std_dev_sq = one_over_n * ( sum_of_squares
                    + mean * ( -2.0 * sum_of_values
                    + mean * number_valid_values ));
            }
            else
            {
                one_over_n = 0.0;
                maximum = minimum = mean = std_dev_sq = 0.0;
            }

            /* Prevent divide by zero or an extremely small value.
               When the standard deviation is set to 0, set the skewness to 0
               and set the kurtosis to the flag value 99999, so it's known
               the result shouldn't be considered valid.  */
            if ( std_dev_sq < DBL_EPSILON )
            {
                standard_deviation = skewness = 0.0;
                kurtosis = 99999.0;
            } 
            else
            {
                kurtosis_denom = 1.0 /
                    ( number_valid_values * std_dev_sq * std_dev_sq );
                standard_deviation = sqrt( std_dev_sq );
                skewness_denom = 1.0 /
                    ( number_valid_values * std_dev_sq * standard_deviation );

                /* Calculate the higher moments */
                skewness = skewness_denom * ( sum_of_cubes
                    + mean * ( -3.0 * sum_of_squares
                    + mean * (  3.0 * sum_of_values
                    + mean * -number_valid_values )));
                kurtosis =   kurtosis_denom * ( sum_of_quartics
                    + mean * ( -4.0 * sum_of_cubes
                    + mean * (  6.0 * sum_of_squares
                    + mean * ( -4.0 * sum_of_values
                    + mean * number_valid_values )))) - 3.0;
            }

            /* Calculate the neighbor correlation statistics */
            if ( number_valid_adjacent_products > 0 )
            {
                one_over_n = 1.0 / number_valid_adjacent_products;
                neighbor_correlation = one_over_n * sum_of_adjacent_products;
                mean_squared_response = one_over_n *
                    sum_of_squares_with_valid_neigh;
            }
            else
            {
                neighbor_correlation = mean_squared_response = 0.0;
            }

            /* Save the statistics */
            if ( characterization_results )
            {
                characterization_results->detector[line_index][left_det_num] =
                    left_det_num+1;
                characterization_results->
                    num_values[line_index][left_det_num] =
                    number_valid_values;
                characterization_results->
                    min[line_index][left_det_num] = minimum;
                characterization_results->
                    max[line_index][left_det_num] = maximum;
                characterization_results->
                    mean[line_index][left_det_num] = mean;
                characterization_results->stdev[line_index][left_det_num] =
                    standard_deviation;
                characterization_results->
                    skewness[line_index][left_det_num] = skewness;
                characterization_results->
                    kurtosis[line_index][left_det_num] = kurtosis;
                characterization_results->
                    correlation[line_index][left_det_num] =
                    neighbor_correlation;
                characterization_results->
                    mean_sq_resp[line_index][left_det_num] =
                    mean_squared_response;
            }

            /* Optional report */
            if ( report_info )
            {
                output_report_data[line_index].num_values = number_valid_values;
                output_report_data[line_index].min = minimum;
                output_report_data[line_index].max = maximum;
                output_report_data[line_index].mean = mean;
                output_report_data[line_index].stdev = standard_deviation;
                output_report_data[line_index].skewness = skewness;
                output_report_data[line_index].kurtosis = kurtosis;
                output_report_data[line_index].num_valid_frames =
                    number_valid_adjacent_products;
                output_report_data[line_index].correlation = 
                    neighbor_correlation;
                output_report_data[line_index].mean_sq_resp = 
                    mean_squared_response;
                /* If dealing with the pan band don't write the report until
                   the second pass (odd line pass), for all other bands
                   write the report since there is only one pass. */
                if (( pan_band && line_index == 1 ) || !pan_band )
                {
                    if ( write_report_file( report_file, left_det_num, pan_band,
                        output_report_data ))
                    {
                        free( output_report_data );
                        free( left_mask );
                        free( right_mask );
                        close_report_file( report_file );
                        return ERROR;
                    }
                }
            }

            /* If no database or report are desired return the statistics to
               the caller. */
            if ( intermediate_results )
            {
                intermediate_results[left_det_num].minimum[line_index] =
                    (int)minimum;
                intermediate_results[left_det_num].maximum[line_index] =
                    (int)maximum;
                intermediate_results[left_det_num].mean[line_index] = mean;
                intermediate_results[left_det_num].
                    standard_deviation[line_index] = standard_deviation;
                intermediate_results[left_det_num].skewness[line_index] =
                    skewness;
                intermediate_results[left_det_num].kurtosis[line_index] =
                    kurtosis;
                intermediate_results[left_det_num].
                    number_valid_values[line_index] = number_valid_values;
                intermediate_results[left_det_num].sum_of_values[line_index] =
                    sum_of_values;
                intermediate_results[left_det_num].sum_of_squares[line_index] =
                    sum_of_squares;
                intermediate_results[left_det_num].sum_of_cubes[line_index] =
                    sum_of_cubes;
                intermediate_results[left_det_num].
                    sum_of_quartics[line_index] = sum_of_quartics;
                intermediate_results[left_det_num].
                    sum_of_adjacent_products[line_index] =
                    sum_of_adjacent_products;
                intermediate_results[left_det_num].
                    sum_of_squares_with_valid_neigh[line_index] =
                    sum_of_squares_with_valid_neigh;
                intermediate_results[left_det_num].
                    number_valid_adjacent_products[line_index] =
                    number_valid_adjacent_products;
            }

        }  /* END pan band exception for line_index loop */

    }  /* END for detector loop */

    /* Free memory */
    free( output_report_data );
    free( left_mask );
    free( right_mask );
    /* Don't free the database memory here.  If memory has been allocated
       the database storage routine will free it when called.  Since that
       may happen in another thread it would be an error to do so here. */

    close_report_file( report_file );

    return SUCCESS;
}  /* END ias_rps_histogram_statistics_characterization */


/*-------------------------------------------------------------------------
   Report Section
-------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------
   NAME:    create_report_file
   PURPOSE: Open the report file and allocate temporary storage variables
-------------------------------------------------------------------------*/
static FILE *create_report_file
(
    int band_number,              /* I: 1-based band number */
    int sca_number,               /* I: 1-based SCA number */
    const char *processing_level, /* I: processing label */
    IAS_REPORT *report_info       /* I: IAS report structure pointer */
)
{
    FILE *report_file = NULL;
    char algorithm_name[] = "histogram_statistics_characterization";
    char *file_name;
    int status;
    IAS_REPORT_HEADER *copy_of_report_header;
    char *file_base_name;
    int length;

    if (processing_level)
    {
        /* Build the base report file name.  Add one to the length for the NULL
           terminator and one for "_" between algorithm_name and 
           processing_level */
        length = strlen(algorithm_name) + strlen(processing_level) + 2;
        file_base_name = malloc(sizeof(char) * length);
        if (!file_base_name)
        {
            IAS_LOG_ERROR("Allocating memory for file_base_name");
            return NULL;        
        }
        else
        {
            sprintf(file_base_name, "%s_%s", algorithm_name,
                     processing_level);
        }
    }
    else
    {
        length = strlen(algorithm_name) + 1;
        file_base_name = malloc(sizeof(char) * length);
        strcpy(file_base_name, algorithm_name);
    }

    /* Get a report file name */
    file_name = ias_rps_get_report_file_name( report_info, band_number, 
        sca_number, file_base_name );
    if ( !file_name )
    {
        IAS_LOG_ERROR("Getting an RPS report file name");
        free(file_base_name);
        return NULL;
    }

    /* Free the file_base_name memory */
    free(file_base_name);
    file_base_name = NULL;

    /* Open the report file */
    report_file = fopen( file_name, "w" );
    if ( !report_file )
    {
        IAS_LOG_ERROR("Opening report file %s", file_name );
        free( file_name );
        return NULL;
    }
    free( file_name );
    file_name = NULL;

    /* Get the local copy of the report header since the "standard" header
       includes non-standard things such as band, SCA and algorithm name.
       In a threaded environment the input header cannot be manipulated. */
    copy_of_report_header = ias_misc_copy_report_header( report_info->header );
    if ( !copy_of_report_header )
    {
        IAS_LOG_ERROR("Creating local copy of report header");
        fclose( report_file );
        return NULL;
    }

    /* Initialize the band/SCA/application name information in the
       "standard" IAS report header */
    ias_misc_set_header_band_number( copy_of_report_header, band_number );
    ias_misc_set_header_sca_number( copy_of_report_header, sca_number );
    ias_misc_set_header_algorithm_name( copy_of_report_header,
        "Histogram Statistics Characterization" );

    status = ias_misc_write_report_header( report_file,
        copy_of_report_header );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Writing out the standard header");
        ias_misc_free_report_header( copy_of_report_header );
        fclose( report_file );
        return NULL;
    }

    /* Write the additional information wanted by the algorithm, the 
       Frame Rate for OLI and TIRS maybe different */
    const IAS_BAND_ATTRIBUTES *ias_band_attributes =
        ias_sat_attr_get_band_attributes( band_number );
    if (ias_band_attributes->spectral_type == IAS_SPECTRAL_THERMAL)
    {
        status = fprintf( report_file, "RPS Level: %s          RLUT Version: %s"
            "          Calculated TIRS Frame Rate: %10.7f\n\n",
            report_info->rps_level, report_info->rlut_version,
            report_info->tirs_frame_rate );
        if ( status < 0 )
        {
            IAS_LOG_ERROR("Writing to report file %s", file_name );
            ias_misc_free_report_header( copy_of_report_header );
            fclose( report_file );
            return NULL;
        }
    }
    else
    {
        status = fprintf( report_file, "RPS Level: %s          RLUT Version: %s"
            "          Calculated OLI Frame Rate: %10.7f\n\n",
            report_info->rps_level, report_info->rlut_version,
            report_info->oli_frame_rate );
        if ( status < 0 )
        {
            IAS_LOG_ERROR("Writing to report file %s", file_name );
            ias_misc_free_report_header( copy_of_report_header );
            fclose( report_file );
            return NULL;
        }
    }

    ias_misc_free_report_header( copy_of_report_header );
    return report_file;
}


/*-------------------------------------------------------------------------
   NAME:    close_report_file
   PURPOSE: Close the report file
-------------------------------------------------------------------------*/
static void close_report_file
(
    FILE *report_file
)
{
    if ( !report_file )
        return;
    fclose( report_file );
}


/* These are the column headings for the report */
static const char data_header1[] =
    " Detector Valid                                     "
    "               Standard                                      "
    "Valid     Neighbor    Mean Squared";
static const char data_header2[] =
    "  Number  Frames     Minimum         Maximum        "
    "Mean           Deviation        Skewness        Kurtosis"
    "   Neighbors  Correlation    Response";

/*-------------------------------------------------------------------------
   NAME:    write_report_file
   PURPOSE: Write the results to the report file
-------------------------------------------------------------------------*/
static int write_report_file( FILE *report, int det, int pan, HIST_DATA
    *output_report_data )
{
    int status;
    int i;
    int num_loops;

    /* Default to the normal band even/odd value */
    char even_odd_line = ' '; 

    if ( pan )
    {
        num_loops = 2;
    }
    else
    {
        num_loops = 1;
    }

    /* Print a nice little header before the data */
    if ( det == 0 )
    {
        status = fprintf( report, "%s\n%s\n", data_header1, data_header2 );
        if ( status < 0 )
        {
            IAS_LOG_ERROR("Writing report header");
            return ERROR;
        }
    }
    for ( i = 0; i < num_loops; i++ )
    {
        if (pan)
        {
            if (IAS_IS_EVEN_LINE(i))
            {
                even_odd_line = 'E';
            }
            else
            {
                even_odd_line = 'O';
            }
        }

        status = fprintf( report, "%c%5d   %5d    %12.7lf    %12.7lf   "
            "%12.7lf    %12.7lf    %12.9lf    %12.9lf    %5d    %10.3lf    "
            "%10.3lf\n",
            even_odd_line, det+1, output_report_data[i].num_values,
            output_report_data[i].min,
            output_report_data[i].max,
            output_report_data[i].mean,
            output_report_data[i].stdev,
            output_report_data[i].skewness,
            output_report_data[i].kurtosis,
            output_report_data[i].num_valid_frames,
            output_report_data[i].correlation,
            output_report_data[i].mean_sq_resp );

        if ( status < 0 )
        {
            IAS_LOG_ERROR("Writing report data");
            return ERROR;
        }
            
    }
    return SUCCESS;
}


/*-------------------------------------------------------------------------
   Database Section
-------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------
   NAME:    ias_rps_db_allocate_histogram_results
   PURPOSE: Allocates the data structure for database storage
-------------------------------------------------------------------------*/
IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS 
    *ias_rps_db_allocate_histogram_results
(
    int number_of_detectors, /* I: Number of array elements to allocate */
    int pan_band             /* I: Flag to indicate if this is for pan band */
)
{
    IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS *db_results;
    int status = SUCCESS;
    int i;
    int stop_condition = 1;

    db_results = malloc(
        sizeof(IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS));
    if ( !db_results )
        return NULL;

    db_results->processing_level = NULL;
    strcpy(db_results->collection_type, "");

    if ( pan_band )
    {
        db_results->is_odd_even_data_split = TRUE;
        stop_condition = 2;
    }
    else
    {
        db_results->is_odd_even_data_split = FALSE;
        db_results->detector[1] = NULL;
        db_results->num_values[1] = NULL;
        db_results->min[1] = NULL;
        db_results->max[1] = NULL;
        db_results->mean[1] = NULL;
        db_results->stdev[1] = NULL;
        db_results->skewness[1] = NULL;
        db_results->kurtosis[1] = NULL;
        db_results->correlation[1] = NULL;
        db_results->mean_sq_resp[1] = NULL;
    }

    for ( i = 0; i < stop_condition; i++ )
    {
        db_results->detector[i] = 
            malloc( number_of_detectors * sizeof(int));
        db_results->num_values[i] =
            malloc( number_of_detectors * sizeof(int));
        db_results->min[i] =
            malloc( number_of_detectors * sizeof(float));
        db_results->max[i] =
            malloc( number_of_detectors * sizeof(float));
        db_results->mean[i] =
            malloc( number_of_detectors * sizeof(float));
        db_results->stdev[i] =
            malloc( number_of_detectors * sizeof(float));
        db_results->skewness[i] =
            malloc( number_of_detectors * sizeof(float));
        db_results->kurtosis[i] =
            malloc( number_of_detectors * sizeof(float));
        db_results->correlation[i] =
            malloc( number_of_detectors * sizeof(float));
        db_results->mean_sq_resp[i] =
            malloc( number_of_detectors * sizeof(float));
        if ( !db_results->detector[i] ||
             !db_results->num_values[i] ||
             !db_results->min[i] ||
             !db_results->max[i] ||
             !db_results->mean[i] ||
             !db_results->stdev[i] ||
             !db_results->skewness[i] ||
             !db_results->kurtosis[i] ||
             !db_results->correlation[i] ||
             !db_results->mean_sq_resp[i] )
        {
            status = ERROR;
        }
    }

    if ( status == ERROR )
    {
        ias_rps_db_free_histogram_results( db_results );
        IAS_LOG_ERROR("Allocating histogram statistics database structure");
    }

    return db_results;
}

/*-------------------------------------------------------------------------
   NAME:    ias_rps_store_histogram_stats_in_database
   PURPOSE: Stores the histogram statistics in the database
-------------------------------------------------------------------------*/
int ias_rps_store_histogram_stats_in_database
(
    void *message /* I: Pointer to IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS
                         structure.  A void pointer is used since it is
                         called from the threading routines and a void
                         pointer is used to support passing messages to
                         different routines */
)
{
    IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS *db_results =
        (IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS *)message;
    int lineset;             /* Lineset value for database column */
    int loop_stop_value = 3; /* Normal case stopping value when combined
                                with lineset or two will loop only once */
    int line_index;          /* Index into the database arrays.  In the
                                normal case, and for the pan band even
                                lines, this is 0.  For the pan band odd
                                numbered lines it's 1 */
    int status = SUCCESS;    /* function return status */
    int number_db_columns;   /* Number of database columns being inserted */
    const char *hist_stats_table_name; /* Correct table name for histogram 
                                          statistics */  
    
    /* Get the correct histogram statistics table name according to the
       collection type provided. */
    hist_stats_table_name = ias_rps_get_histogram_statistics_table_name
                            (db_results->collection_type);

    /* When the pan band is encountered there are two sets of data, one for
       the even rows (line_index 1) and one for the odd (line_index 0).
       Lineset is the database column that indicates whether the data is
       from even lines (lineset 1), odd lines (lineset 0) or all lines
       (lineset 2). */

    /* Set the loop bounds */
    if ( db_results->is_odd_even_data_split )
    {
        /* For the pan band iterate twice */
        loop_stop_value = 2;
    }
    else
    {
        /* For normal bands iterate once */
        loop_stop_value = 1;
    }

    for (line_index = 0; line_index < loop_stop_value; line_index++)
    {
        /* Determine the lineset value. */
        if (db_results->is_odd_even_data_split)
        {
            /* For the pan band, even lines have a lineset = 1 and odd lines
               have a lineset = 0 */
            if (IAS_IS_EVEN_LINE(line_index))
                lineset = 1;
            else
                lineset = 0;
        }
        else
        {   
            /* For normal bands the lineset is 2 */
            lineset = 2;
        }        

        /* Create the table */
        IAS_DB_INSERT_TABLE table[]=
        {
            IAS_DB_TABLE_STRING( "radiometric_wo_common_id",
                db_results->rad_common_id ),
            IAS_DB_TABLE_INT( "band_id", db_results->band_number ),
            IAS_DB_TABLE_INT( "sensor_chip_assembly_id", db_results->sca ),
            IAS_DB_TABLE_INT_ARRAY( "detector_id",
                db_results->detector[line_index] ),
            IAS_DB_TABLE_STRING( "processing_level",
                db_results->processing_level ),
            IAS_DB_TABLE_INT( "lineset", lineset ),
            IAS_DB_TABLE_FLOAT_ARRAY( "scene_pixel_min", 
                db_results->min[line_index] ),
            IAS_DB_TABLE_FLOAT_ARRAY( "scene_pixel_max", 
                db_results->max[line_index] ),
            IAS_DB_TABLE_FLOAT_ARRAY( "scene_pixel_stdev",
                db_results->stdev[line_index] ),
            IAS_DB_TABLE_FLOAT_ARRAY( "scene_pixel_mean",
                db_results->mean[line_index] ),
            IAS_DB_TABLE_FLOAT_ARRAY( "scene_pixel_skewness",
                db_results->skewness[line_index] ),
            IAS_DB_TABLE_FLOAT_ARRAY( "scene_pixel_kurtosis",
                db_results->kurtosis[line_index] ),
            IAS_DB_TABLE_INT_ARRAY( "nbr_of_frames",
                db_results->num_values[line_index] ),
            IAS_DB_TABLE_FLOAT_ARRAY( "mean_squared_response",
                db_results->mean_sq_resp[line_index] ),
            IAS_DB_TABLE_FLOAT_ARRAY( "adjacent_correlation",
                db_results->correlation[line_index] )
        };

        /* Get the number of columns inserted into the table and insert the
           data into the database */
        number_db_columns = sizeof(table) / sizeof(IAS_DB_INSERT_TABLE);
        status = ias_db_insert_using_table( db_results->dbase,
            hist_stats_table_name, table, number_db_columns,
            db_results->num_detectors );
        if ( status )
        {
            IAS_LOG_ERROR("Inserting histogram statistics for band %d, "
                "sca %d", db_results->band_number, db_results->sca );
        }
    }

    ias_rps_db_free_histogram_results( db_results );
    return status;
}

/*----------------------------------------------------------------------------
NAME: ias_rps_db_get_histogram_statistics

PURPOSE: Populate the structure with data obtained from the database.

NOTE: The database, common work order id, band number, and sca number are 
    required to be populated in the input struct and collection type is
    required to know the correct histogram statistics table.

RETURNS: An allocated IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS*, NULL if 
    error.
-----------------------------------------------------------------------------*/
IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS* 
    ias_rps_db_get_histogram_statistics
(
    struct ias_db_connection *dbase, /* I: database connection */
    const char *rad_common_id,       /* I: common work order ID */
    const int band_number,           /* I: Band number, 1-based */
    const int sca_number,            /* I: SCA number, 1-based */
    const char *collection_type      /* I: Collection type data */
)
{
    const int max_where_clause = 128;
    char where_clause[max_where_clause];
    int count = 0;
    int line_index = 0;
    int lineset = 0;
    int loop_stop_value = 0;
    int num_lines = 0;
    int pan_band = FALSE;
    int status = SUCCESS;    /* function return status */
    int table_size = 0;
    const char *hist_stats_table_name;

    const IAS_BAND_ATTRIBUTES *band_attributes = NULL;
    IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS *statistics = NULL;

    if (dbase == NULL)
    {
        IAS_LOG_ERROR("NULL pointer received for database connection");
        return NULL;
    }

    IAS_LOG_DEBUG("Obtaining histogram statistics for common work order "
        "%s, band %d, sca %d", rad_common_id, band_number, sca_number);

    /* Obtain the number of lines that will be in the structure */
    /* Build a second query to obtain the single processing level used with 
        the data */
    IAS_DB_QUERY_TABLE qtable_count[] =
    {
        {"COUNT(*)", &num_lines, sizeof(num_lines), IAS_QUERY_INT, 1}
    };

    /* Get the correct histogram statistics table name according to the
       collection type provided. */
    hist_stats_table_name = 
        ias_rps_get_histogram_statistics_table_name(collection_type);
    status = snprintf(where_clause, max_where_clause, 
        "RADIOMETRIC_WO_COMMON_ID = '%s' AND BAND_ID = %d "
        "AND SENSOR_CHIP_ASSEMBLY_ID = %d GROUP BY LINESET",
         rad_common_id, band_number, sca_number);
    if (status < 0 || status >= max_where_clause)
    {
        IAS_LOG_ERROR("Unable to obtain a detector count from the database"); 
        return NULL;
    }
    table_size  = sizeof(qtable_count)/sizeof(IAS_DB_QUERY_TABLE);
    
    status = ias_db_query_using_table(dbase, hist_stats_table_name, 
        qtable_count, table_size, where_clause, 1);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("The query to obtain the number of statistic entries "
            "failed");
        return NULL;
    }

    /* Verify there are histogram statistics available to obtain */
    if (num_lines < 1)
    {
        IAS_LOG_ERROR("No entries exist for work order %s, band %d, sca %d",
            rad_common_id, band_number, sca_number);
        return NULL;
    }

    band_attributes = ias_sat_attr_get_band_attributes( band_number );
    if (band_attributes == NULL)
    {
        IAS_LOG_ERROR("Unable to obtain band attributes for band number %d", 
            band_number);
        return NULL;
    }
    if (band_attributes->band_type == IAS_PAN_BAND)
    { 
        pan_band = TRUE;
    }

    /* Allocate the structure to hold the database results */
    statistics = ias_rps_db_allocate_histogram_results(num_lines, pan_band);

    if (statistics == NULL) 
    {
        IAS_LOG_ERROR("Unable to allocate structure for histogram statistics");
        return NULL;
    }
 
    statistics->dbase = dbase;
    strncpy(statistics->rad_common_id, rad_common_id, 
        sizeof(statistics->rad_common_id));
    statistics->rad_common_id[sizeof(statistics->rad_common_id)- 1] = '\0';
    statistics->band_number = band_number;
    statistics->sca = sca_number;
    statistics->num_detectors = num_lines;
    strncpy(statistics->collection_type, collection_type,
        sizeof(statistics->collection_type));
    statistics->collection_type[sizeof(statistics->collection_type) - 1] = '\0';
    statistics->processing_level = calloc(IAS_RPS_PROCESSING_LEVEL_LENGTH, 1);

    if (statistics->processing_level == NULL)
    {
        IAS_LOG_ERROR("Unable to allocate space for the processing level "
            "string");
        ias_rps_db_free_histogram_results(statistics);
        return NULL;
    }

    /* Set the loop bounds for the PAN and normal bands. */
    if ( pan_band )
    {
        /* For the pan band iterate twice */
        loop_stop_value = 2;
    }
    else
    {
        /* For normal bands iterate once. */
        loop_stop_value = 1;
    }

    /* Execute the query to obtain the data per lineset */
    for (line_index = 0; line_index < loop_stop_value; line_index++)
    {
        IAS_LOG_DEBUG("Line Index loop %d of %d", 
            (line_index + 1), (loop_stop_value));
 
        /* Determine the lineset value. The lineset is used to indecate the 
           odd/even pan lines or the line as a normal frame. */
        if (pan_band)
        {
            /* For the pan band, even lines have a lineset = 1 and odd lines
               have a lineset = 0 */
            if (IAS_IS_EVEN_LINE(line_index))
                lineset = 1;
            else
                lineset = 0;
        }
        else
        {   
            /* For normal bands the lineset is 2 */
            lineset = 2;
        }        

        /* Build the query to obtain the histogram statistics data */
        IAS_DB_QUERY_TABLE qtable[] =
        {
            {"detector_id", statistics->detector[line_index],
                sizeof(statistics->detector[0][0]),
                IAS_QUERY_INT_ARRAY, num_lines},
            {"scene_pixel_min", statistics->min[line_index],
                sizeof(statistics->min[0][0]),
                IAS_QUERY_FLOAT_ARRAY, num_lines},
            {"scene_pixel_max", statistics->max[line_index],
                sizeof(statistics->max[0][0]),
                IAS_QUERY_FLOAT_ARRAY, num_lines},
            {"scene_pixel_stdev", statistics->stdev[line_index],
                sizeof(statistics->stdev[0][0]),
                IAS_QUERY_FLOAT_ARRAY, num_lines}, 
            {"scene_pixel_mean", statistics->mean[line_index],
                sizeof(statistics->mean[0][0]),
                IAS_QUERY_FLOAT_ARRAY, num_lines},
            {"scene_pixel_skewness", statistics->skewness[line_index],
                sizeof(statistics->skewness[0][0]), IAS_QUERY_FLOAT_ARRAY, 
                num_lines},
            {"scene_pixel_kurtosis", statistics->kurtosis[line_index],
                sizeof(statistics->kurtosis[0][0]), IAS_QUERY_FLOAT_ARRAY, 
                num_lines},
            {"mean_squared_response", statistics->mean_sq_resp[line_index],
                sizeof(statistics->mean_sq_resp[0][0]), IAS_QUERY_FLOAT_ARRAY, 
                num_lines},
            {"adjacent_correlation", statistics->correlation[line_index],
                sizeof(statistics->correlation[0][0]), IAS_QUERY_FLOAT_ARRAY, 
                num_lines},
            {"nbr_of_frames", statistics->num_values[line_index],
                sizeof(statistics->num_values[0][0]), IAS_QUERY_INT_ARRAY, 
                num_lines},
        };

        status = snprintf(where_clause, max_where_clause, 
            "RADIOMETRIC_WO_COMMON_ID = %s AND LINESET = %d AND BAND_ID = %d "
            "AND SENSOR_CHIP_ASSEMBLY_ID = %d ORDER BY DETECTOR_ID", 
            rad_common_id, lineset, band_number, sca_number);
        
        if (status < 0 || status >= max_where_clause)
        {
            IAS_LOG_ERROR("Unable to create where clause for the query"); 
            ias_rps_db_free_histogram_results(statistics);
            return NULL;
        }

        table_size  = sizeof(qtable)/sizeof(IAS_DB_QUERY_TABLE);
        status = ias_db_query_using_table(statistics->dbase, 
            hist_stats_table_name, qtable, table_size, where_clause, 
            num_lines);
 
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Unable to obtain %d rows of histogram statistics by "
                "where clause: %s", num_lines, where_clause);
            ias_rps_db_free_histogram_results(statistics);
            return NULL;
        }

        IAS_LOG_DEBUG("Querying for processing level where %s", where_clause);

        /* Build a second query to obtain the single processing level used with 
        the data */
        IAS_DB_QUERY_TABLE qtable_single[] =
        {
            {"COUNT(*)", &count, sizeof(count), IAS_QUERY_INT, 1},
            {"MAX(processing_level)", statistics->processing_level, 
                IAS_RPS_PROCESSING_LEVEL_LENGTH, IAS_QUERY_STRING, 1}
        };

        table_size  = sizeof(qtable_single)/sizeof(IAS_DB_QUERY_TABLE);
        status = ias_db_query_using_table(statistics->dbase,
            hist_stats_table_name, qtable_single, table_size, where_clause, 1);

        if (count < 1 || status != SUCCESS)
        {
            IAS_LOG_ERROR("Unable to obtain the processing level");
            ias_rps_db_free_histogram_results(statistics);
            return NULL;
        }

    }/* End of lineset loop */

    IAS_LOG_DEBUG("Done obtaining histogram statistics");

    return statistics;
}

/*-------------------------------------------------------------------------
   NAME:    ias_rps_db_free_histogram_results
   PURPOSE: Frees the structure
-------------------------------------------------------------------------*/
void ias_rps_db_free_histogram_results
(
    IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS *db_results /* I: structure
                                                                    to free */
)
{
    free( db_results->detector[0] );
    free( db_results->num_values[0] );
    free( db_results->min[0] );
    free( db_results->max[0] );
    free( db_results->mean[0] );
    free( db_results->stdev[0] );
    free( db_results->skewness[0] );
    free( db_results->kurtosis[0] );
    free( db_results->correlation[0] );
    free( db_results->mean_sq_resp[0] );
    free( db_results->detector[1] );
    free( db_results->num_values[1] );
    free( db_results->min[1] );
    free( db_results->max[1] );
    free( db_results->mean[1] );
    free( db_results->stdev[1] );
    free( db_results->skewness[1] );
    free( db_results->kurtosis[1] );
    free( db_results->correlation[1] );
    free( db_results->mean_sq_resp[1] );
    free( db_results->processing_level );
    free( db_results );
}

/*------------------------------------------------------------------------
 NAME:     ias_rps_get_histogram_statistics_table_name

 PURPOSE:  Accepts the collection type and returns the 
           correct histogram statistics table name

 RETURNS:  Returns the correct histogram statistics table name
---------------------------------------------------------------------------*/
const char * ias_rps_get_histogram_statistics_table_name
(
    const char *collection_type
)
{
    if (strcmp(collection_type, IAS_RPS_EARTH_IMAGING_COLLECT) == 0)
    {
        return "HISTOGRAM_STATISTICS_EARTH";
    }
    else
    {
        return "HISTOGRAM_STATISTICS_CAL";
    }    
}
