#ifndef IAS_RPS_HISTOGRAM_STATISTICS_CHARACTERIZATION_H
#define IAS_RPS_HISTOGRAM_STATISTICS_CHARACTERIZATION_H

/* Maximum string length that can be stored in the IAS database for the 
processing level string */
#define IAS_RPS_PROCESSING_LEVEL_LENGTH 20 

typedef struct ias_rps_db_histogram_characterization_results
{
    struct ias_db_connection *dbase;
    char rad_common_id[IAS_CHAR_ID_SIZE];  
    int  band_number;
    int  sca;
    int  num_detectors;
    int  is_odd_even_data_split;
    char *processing_level; /* This is allocated with strdup and must be
                               freed  before freeing the structure. */

    /* The following are dynamically allocated.  If errors occur before the
       database routine is called, they must be freed.  Do not free the memory
       once the storage routine is called since it will be freed there.
       Since the pan band histogram statistics are calculated separately for
       the odd and even lines of the image, two sets of pointers are needed
       for the statictics. */
    int   *detector[2];
    int   *num_values[2];
    float *min[2];
    float *max[2];
    float *mean[2];
    float *stdev[2];
    float *skewness[2];
    float *kurtosis[2];
    float *correlation[2];
    float *mean_sq_resp[2];
    char  collection_type[IAS_L0R_COLLECTION_TYPE_LENGTH+1];    
} IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS;

typedef struct ias_rps_histogram_intermediate_results
{
    char processing_level[IAS_RPS_PROCESSING_LEVEL_LENGTH + 1];
    double minimum[2];
    double maximum[2];
    double mean[2];
    double standard_deviation[2];
    double skewness[2];
    double kurtosis[2];

    /* Cumulative results for standard statistics */
    int  number_valid_values[2];
    double sum_of_values[2];
    double sum_of_squares[2];
    double sum_of_cubes[2];
    double sum_of_quartics[2];

    /* Cumulative results for the adjacent pixel correlation */
    int number_valid_adjacent_products[2];
    double sum_of_adjacent_products[2];
    double sum_of_squares_with_valid_neigh[2];
} IAS_RPS_HISTOGRAM_INTERMEDIATE_RESULTS;

int ias_rps_histogram_statistics_characterization
(
    void *input_buffer,         /* I: One SCA of data */
    IAS_DATA_TYPE data_type,    /* I: Data type may be float or uint16_t */
    int band,                   /* I: Band number, 1-based */
    int sca,                    /* I: SCA number, 1-based */
    int num_detectors,          /* I: Number of detectors in the SCA */
    int num_lines,              /* I: Number of lines in the data */
    int *detector_offsets,      /* I: Integer detector offsets with most
                                      negative offset shifted to zero */
    int frames_to_skip_top,     /* I: Number of frames to skip at the 
                                       top (beginning) of image */
    int frames_to_skip_bottom,  /* I: Number of frames to skip at the 
                                       bottom (end) of image */
    IAS_PIXEL_MASK *pixel_mask, /* I: Contains pixel artifacts */
    IAS_REPORT *report_info,    /* I: General information for the test
                                       report.  If no report is desired pass
                                       NULL. */
    /* I/O: Pointer to results of the characterization.  If results
       are desired pass an allocated structure or NULL if database output is
       not desired.  Allocate structure using:
       ias_rps_db_allocate_histogram_results */
    IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS *characterization_results,

    /* I/O: Results of some calculations to allow combining histogram
            statistics from multiple runs.  If they are not desired pass
            NULL.  The memory sould be allocated as one block of memory as
            (IAS_HISTOGRAM_INTERMEDIATE_RESULTS * num_detectors) and the
            memory accessed as intermediate_results[detector] */
    IAS_RPS_HISTOGRAM_INTERMEDIATE_RESULTS *intermediate_results
);

IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS 
    *ias_rps_db_allocate_histogram_results
(
    int number_of_detectors, /* I: Number of array elements to allocate */
    int pan_band             /* I: Flag to indicate if this is for pan band */
);

void ias_rps_db_free_histogram_results
(
    IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS *db_results /* I: structure
                                                                    to free */
);

int ias_rps_store_histogram_stats_in_database
(
    void *message /* I: Pointer to IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS
                         structure.  A void pointer is used since it is
                         called from the threading routines and a void
                         pointer is used to support passing messages to
                         different routines */
);

IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS*
    ias_rps_db_get_histogram_statistics
(
    struct ias_db_connection *dbase, /* I: database connection */
    const char *rad_common_id,       /* I: common work order ID */
    const int band_number,           /* I: Band number, 1-based */
    const int sca,                   /* I: SCA number, 1-based */
    const char *collection_type      /* I: Collection type data */
);

const char * ias_rps_get_histogram_statistics_table_name
(
    const char *collection_type /* I: Collection type */
);

#endif
