/****************************************************************************
 NAME:     ias_rps_identify_diffuser_panel

 PURPOSE:  Reads the OLI telemetry data containing the diffuser resolver
           position information around the image start time to determine
           which diffuser panel was deployed during the solar collect.

 RETURNS:  Integer status code of SUCCESS or ERROR

 NOTES:    At this time, the only characterizations calling this routine
           are the standalone RPS solar-based characterizations
           (i.e. char_nonuniformity, solarcal_oli, and char_rad_stability).
           These characterizations 'verify' that the collect is an OLI solar
           collect as part of their processing.  The implementation for this
           routine assumes the collect is a valid solar collect.
*****************************************************************************/
#include <stdlib.h>
#include "ias_const.h"
#include "ias_cpf.h"
#include "ias_l0r.h"
#include "ias_logging.h"
#include "ias_math.h"
#include "ias_radiometry.h"
#include "ias_structures.h"
#include "ias_types.h"


int ias_rps_identify_diffuser_panel
(
    const char   *l0r_filename,            /* I: Name of L0R data file */
    const struct IAS_CPF_EARTH_CONSTANTS *earth_constants,
                                           /* I: CPF Earth constants data */
    IAS_DATETIME  starting_date,           /* I: Starting date/time
                                              components */
    uint16_t *diffuser_resolver_position,  /* O: Diffuser position telemetry
                                              value (optional-- pass in
                                              NULL if it isn't needed) */
    int *panel_type                        /* O: Enumerated value
                                              representing the
                                              deployed panel (possible
                                              values are
                                        IAS_RPS_OLI_DIFFUSER_TYPE_PRISTINE
                                        and
                                        IAS_RPS_OLI_DIFFUSER_TYPE_WORKING) */
)
{
    L0RIO *l0r = NULL;                         /* Pointer to L0R image
                                                  data */
    IAS_L0R_OLI_TELEMETRY *oli_telem = NULL;   /* OLI telemetry data */
    double year_doy_sod[3];
    double seconds;                            /* Seconds-from-J2000
                                                  representing the image
                                                  start time */
    int num_telemetry_records = 0;             /* Number of OLI telemetry
                                                  records */
    int status;
    int index;                                 /* Telemetry data loop
                                                  counter */
    uint16_t resolver_position;
    int resolver_found;


    /* Convert the datetime components to the corresponding
       year/day-of-year/second-of-day components */
    year_doy_sod[0] = starting_date.year;
    year_doy_sod[1] = starting_date.day_of_year;
    year_doy_sod[2] = (starting_date.hour * 3600.0)
        + (starting_date.minute * 60.0) + starting_date.second;

    /* Initialize the math library leap seconds functionality */
    status = ias_math_init_leap_seconds_from_UTC_time(year_doy_sod,
        &earth_constants->leap_seconds_data);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Initializing math library leap-seconds for the "
            "specified starting date");
        return ERROR;
    }

    /* Convert the starting date/time components to an equivalent
       seconds-of-day */
    year_doy_sod[0] = starting_date.year;
    year_doy_sod[1] = starting_date.day_of_year;
    year_doy_sod[2] = (starting_date.hour * 3600)
        + (starting_date.minute * 60) + (starting_date.second);

    /* Convert the UTC seconds-of-day to seconds-of-day from J2000, then
       "reset" the leap year information */
    status = ias_math_convert_year_doy_sod_to_j2000_seconds(year_doy_sod,
        &seconds);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Converting starting date/time to seconds from J2000 "
            "epoch");
        return ERROR;
    }
    ias_math_clear_leap_seconds();

    /* Open the L0R file. The ias_l0r_open_from_filename function uses
       the full L0R filename, assuming it consists of a path plus a base
       name. */
    l0r = ias_l0r_open_from_filename(l0r_filename, IAS_L0R_COMPRESSION_ON);
    if (l0r == NULL)
    {
        IAS_LOG_ERROR("Unable to open L0R file %s", l0r_filename);
        return ERROR;
    }

    /* Open the ancillary dataset for reading */
    status = ias_l0r_open_ancillary(l0r, IAS_READ);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Cannot open ancillary data for reading from L0R %s",
            l0r_filename);
        ias_l0r_close(l0r);
        return ERROR;
    }

    /* Get the total number of OLI telemetry records */
    status = ias_l0r_get_ancillary_oli_telemetry_records_count(l0r,
        &num_telemetry_records);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Cannot obtain number of ancillary OLI telemetry "
            "records in L0R %s", l0r_filename);
        ias_l0r_close_ancillary(l0r);
        ias_l0r_close(l0r);
        return ERROR;
    }
    else if (num_telemetry_records < 1)
    {
        IAS_LOG_ERROR("Insufficient number of OLI telemetry records in "
            "L0R data file %s", l0r_filename);
        ias_l0r_close_ancillary(l0r);
        ias_l0r_close(l0r);
        return ERROR;
    }

    /* Allocate a data buffer to contain all the telemetry records */
    oli_telem = malloc(num_telemetry_records * sizeof(IAS_L0R_OLI_TELEMETRY));
    if (oli_telem == NULL)
    {
        IAS_LOG_ERROR("Unable to allocate memory for OLI telemetry");
        ias_l0r_close_ancillary(l0r);
        ias_l0r_close(l0r);
        return ERROR;
    }

    /* Get all of the available telemetry records */
    status = ias_l0r_get_ancillary_oli_telemetry(l0r, 0,
        num_telemetry_records, oli_telem);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Cannot obtain OLI ancillary telemetry group "
            "data from L0R %s", l0r_filename);
        ias_l0r_close_ancillary(l0r);
        ias_l0r_close(l0r);
        free(oli_telem);
        return ERROR;
    }

    /* Close access to the ancillary data and L0R image data */
    ias_l0r_close_ancillary(l0r);
    ias_l0r_close(l0r);

    resolver_found = 0;
    resolver_position = oli_telem[0].oli_telemetry_header_4
                .diffuser_resolver_position;
    for (index = 0; index < num_telemetry_records; index++)
    {
        double record_time_from_J2000 =
            IAS_L0R_CONVERT_TIME_TO_SECONDS_SINCE_J2000(
                oli_telem[index].oli_telemetry_header_4.l0r_time);

        /* Get the resolver position from the current index since we want
           to use the nearest available to the image start time */
        resolver_position = oli_telem[index].oli_telemetry_header_4
                .diffuser_resolver_position;

        /* If the record time is after the image start time, use this value */
        if (record_time_from_J2000 >= seconds)
        {
            /* Only consider the resolver to be found if the record index isn't
               the starting record.  If it is the starting record, it might 
               be after the imagery was finished collecting and might not
               overlap. */
            if (index > 0)
                resolver_found = 1;
            break;
        }
    }

    /* If we haven't found a telemetry value near the estimated image start
       time, warn that we used the closest available */
    if (!resolver_found)
    {
        IAS_LOG_WARNING("No diffuser resolver position telemetry found "
            "overlapping with the estimated image start time %10.3lf seconds, "
            "using closest available", seconds);
    }

    /* Clean up */
    free(oli_telem);
    oli_telem = NULL;

    /* If the actual panel telemetry value is wanted, save it */
    if (diffuser_resolver_position != NULL)
    {
        *diffuser_resolver_position = resolver_position;
    }

    /* Determine the panel ID based on the expected range of diffuser
       resolver position values */
    if ((resolver_position >= IAS_RPS_OLI_DIFFUSER_POSITION_PRISTINE_LOW)
        && (resolver_position
            <= IAS_RPS_OLI_DIFFUSER_POSITION_PRISTINE_HIGH))
    {
        *panel_type = IAS_RPS_OLI_DIFFUSER_TYPE_PRISTINE;
    }
    else if ((resolver_position
              >= IAS_RPS_OLI_DIFFUSER_POSITION_WORKING_LOW)
        && (resolver_position
              <= IAS_RPS_OLI_DIFFUSER_POSITION_WORKING_HIGH))
    {
        *panel_type = IAS_RPS_OLI_DIFFUSER_TYPE_WORKING;
    }
    else
    {
        IAS_LOG_ERROR("Diffuser resolver position %u does not correspond "
           "to either the working panel or the pristine panel",
           resolver_position);
        return ERROR;
    }

    return SUCCESS;
}   /* END -- ias_rps_identify_oli_diffuser_panel */
