/*----------------------------------------------------------------------------
NAME:    ias_rps_impulse_noise_characterization

PURPOSE: Identify the locations of impulse noise in the data.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <math.h>

#include "ias_const.h"
#include "ias_math.h"
#include "ias_logging.h"
#include "ias_pixel_mask.h"
#include "ias_db_insert.h"
#include "ias_radiometry.h"

#define TWODARRAY( a, s, r, c )  *(((a) + ((r) * (s))) + (c))
#define REPORT_BURST_SIZE 100

/* Structure to hold data for the report generation routines */
struct report_struct
{
    int *detectors;
    int *frames;
    int *value1;
    int *value2;
    int *value3;
};

/* Prototypes for local report generation routines */
static FILE *open_report_file( const IAS_REPORT *, int, int,
    struct report_struct * );
static int close_report_file( FILE *, int, struct report_struct * );
static int write_report( FILE *, int, int, int, uint16_t, uint16_t,
    uint16_t, int, struct report_struct * );

/* Characterize impulse noise */
int ias_rps_impulse_noise_characterization
(
    const uint16_t *l0r_data,   /* I: One SCA of level 0 data in DN */
    int band,                   /* I: Band number, 1-based */
    int sca,                    /* I: SCA number, 1-based */
    int num_detectors,          /* I: Number of detectors in the SCA */
    int num_lines,              /* I: Number of lines in the data */
    int median_filter_width,    /* I: Filter width from CPF */
    const double *detector_noise,/* I: Per detector noise from CPF */
    int impulse_noise_limit,    /* I: Limit of minimum detectable impulse
                                       noise */
    IAS_PIXEL_MASK *pixel_mask, /* I: Contains inop detector and dropped
                                       frame locations
                                   O - Impulse noise locations */
    const IAS_REPORT *report_info,/* I: General information for the test
                                       report.  If no report is desired pass
                                       NULL. */
    IAS_RPS_DB_IMPULSE_NOISE_CHAR_RESULTS *db_results
                                /* I/O: General information on the database
                                       connection.  If database output is
                                       not desired pass NULL. */
)
{
    int line;                   /* Loop counter for the lines */
    int det;                    /* Loop counter for the detectors */
    int num_extra_lines;        /* Number of extra lines to append to data */
    int current;                /* Index of pixel being tested */
    int prev;                   /* Index of the pixel before current */
    int next;                   /* Index of the pixel after current */
    int oldest;                 /* Index of the oldest pixel in the array
                                   of filter pixels. */
    uint16_t *filter_pixels;    /* This buffer contains the pixels to be
                                   filtered */
    int filter_pixel_index;     /* Index into the l0r_data array for next
                                   filter pixel. */
    int found;                  /* Flag used for controlling dropped frame
                                   search loop. */
    uint16_t *pre_data_lines;   /* Lines appended to beginning of data */
    uint16_t *post_data_lines;  /* Lines appended to end of data */
    int post_pixel_index;       /* Index into the post-data pixels */
    int candidate_index;        /* Index into the l0r_data used to look for
                                   pre- and post-data pixels. */
    int pixel_count;            /* Count of pre- and post-data pixels found */
    PIXEL_MASK_TYPE *temp_mask; /* Line of pixel masks from image */
    int median_value;           /* Median value of the filter pixels */
    int pixel_median_diff;      /* Absolute difference between the current
                                   pixel and the median value */
    int neighbors_diff;         /* Absolute difference between the current 
                                   pixel's two neighboring pixels */
    float threshold;            /* Noise threshold.  If pixel is greater
                                   than this, it is identified as IN. */
    float pre_calc_gt_twice;    /* Intermediate calculation of:
                                   IN limit / ( 2 * detector noise ) */
    float pre_calc_lte_twice;   /* Intermediate calculation of:
                                   IN limit * detector noise */
    float twice_detector_noise; /* Pre-calculated value */
    int total_impulse_noise_count;/* Count of impulse noise pixels found */
    FILE *report_file = NULL;   /* File handle for the report */
    int return_status = SUCCESS;/* Status to return to caller */
    int generate_report = TRUE;    /* Convenience flag */
    int store_in_database = TRUE;  /* Convenience flag */
    struct report_struct *report_data = NULL;
    int status;                    /* Return status value */

    /* This algorithm expects an odd median filter width so make sure it is
       and return an ERROR otherwise. */
    if ( median_filter_width % 2 == 0 )
    {
        IAS_LOG_ERROR("Median filter width of %d was passed for band %d "
            "SCA %d.  Filter width must be odd.", median_filter_width, band,
            sca );
        return ERROR;
    }

    /* Set the flags false if NULL was passed in either of the report or
       database structures. */
    if ( report_info == NULL )
        generate_report = FALSE;
    if ( db_results == NULL )
        store_in_database = FALSE;


    /* Open the report file and write the standards header */
    if ( generate_report )
    {
        report_data = malloc(sizeof(struct report_struct));
        if ( !report_data )
        {
            IAS_LOG_ERROR("Allocating report structure");
            return ERROR;
        }

        report_file = open_report_file( report_info, band, sca, report_data );
        if ( !report_file )
        {
            free( report_data );
            return ERROR;
        }
    }

    /* Calculate the number of lines of data to append and allocate the
       pre- and post-data buffers. */
    num_extra_lines = (int)floor( median_filter_width / 2.0 );
    if ( num_extra_lines < 1 )
    {
        IAS_LOG_ERROR("Median filter width is %d giving %d extra lines",
            median_filter_width, num_extra_lines );
        close_report_file( report_file, 0, report_data );
        free( report_data );
        return ERROR;
    }
    pre_data_lines = malloc( sizeof(uint16_t) * num_extra_lines );
    if ( !pre_data_lines )
    {
        IAS_LOG_ERROR( "Allocating pre-data buffer" );
        close_report_file( report_file, 0, report_data );
        free( report_data );
        return ERROR;
    }
    post_data_lines = malloc( sizeof(uint16_t) * num_extra_lines );
    if ( !post_data_lines )
    {
        IAS_LOG_ERROR( "Allocating post-data buffer" );
        free( pre_data_lines );
        close_report_file( report_file, 0, report_data );
        free( report_data );
        return ERROR;
    }

    /* Allocate the array for the filter pixels */
    filter_pixels = malloc( sizeof(uint16_t) * median_filter_width );
    if ( !filter_pixels )
    {
        IAS_LOG_ERROR("Allocating the filter pixels array");
        free( pre_data_lines );
        free( post_data_lines );
        close_report_file( report_file, 0, report_data );
        free( report_data );
        return ERROR;
    }

    /* Allocate the temporary lines of pixel masks */
    temp_mask = malloc( sizeof(PIXEL_MASK_TYPE) * num_lines );
    if (!temp_mask)
    {
        IAS_LOG_ERROR("Allocating pixel mask memory");
        free( pre_data_lines );
        free( post_data_lines );
        free( filter_pixels );
        close_report_file( report_file, 0, report_data );
        free( report_data );
        return ERROR;
    }

    /* Loop through each detector of the input data */
    total_impulse_noise_count = 0;
    for ( det = 0; det < num_detectors; det++ )
    {
        int impulse_noise_count = 0;
        /* Grab the pixel mask information for all frames of data for
           the current detector */
        status = ias_pm_get_image_subset(pixel_mask, det, 0, 1,
             num_lines, temp_mask);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Attempting to get pixel mask at detector %d"
                    " line 0+1", det+1 );
            free( pre_data_lines );
            free( post_data_lines );
            free( filter_pixels );
            close_report_file( report_file, impulse_noise_count, report_data );
            free( report_data );
            free ( temp_mask );
            return ERROR;
        }

        /* Now loop through the first line of the input image, ignore the
           the whole detector lines if it's an inoperative detector */
        if (temp_mask[0] & IAS_RPS_INOP_DETECTOR)
        {
            if ( store_in_database )
                db_results->in_counts[det] = 0;
            continue;
        }

        /* Pre-calculate several invariant values for this detector.  This
           is done here instead of re-calculating them in the inner loop. */
        twice_detector_noise = 2 * detector_noise[det];
        pre_calc_gt_twice = impulse_noise_limit / twice_detector_noise;
        pre_calc_lte_twice = impulse_noise_limit * detector_noise[det];

        /* Copy the extra pre- and post-data pixels into arrays.
           For pre-data copy the first num_extra_lines starting at the
           second pixel */
        candidate_index = 1;
        pixel_count = 0;
        while ( pixel_count < num_extra_lines )
        {
            if ( temp_mask[candidate_index] & IAS_RPS_DROPLINE )
            {
                candidate_index++;
                continue;
            }
            pre_data_lines[pixel_count] = TWODARRAY( l0r_data, num_detectors,
                candidate_index, det );
            pixel_count++;
            candidate_index++;
        }

        /* For post-data copy the last num_extra_lines ending at the
           second to the last pixel */
        candidate_index = num_lines - 2;
        pixel_count = num_extra_lines - 1;
        while ( pixel_count >= 0 )
        {
            if ( temp_mask[candidate_index] & IAS_RPS_DROPLINE )
            {
                candidate_index--;
                continue;
            }
            post_data_lines[pixel_count] = TWODARRAY( l0r_data,
                num_detectors, candidate_index, det );
            pixel_count--;
            candidate_index--;
        }

        /* Set up the filter array with initial values.
           This copies in the pre-data pixels. */
        memcpy( filter_pixels, pre_data_lines, num_extra_lines *
            sizeof(uint16_t) );
        /* And this the rest of the filter values. */
        for ( line = 0, current = num_extra_lines;
                line <= num_extra_lines; line++ )
        {
            /* See if the current line is a dropped frame and skip it */
            if ( temp_mask[line] & IAS_RPS_DROPLINE )
            {
                continue;
            }
            filter_pixels[current++] = TWODARRAY( l0r_data, num_detectors,
                    line, det );
        }

        /* Set up the initial indices.  Choosing multiple indices to
           eliminate the need to do modulus arithmetic.  These all get
           incremented early in the next loop so they are set to values
           that, when incremented, will be correct for the first time
           through the loop.  It doesn't matter if prev goes negative in the
           case of the filter width being 3 since it will be incremented
           before its first use. */
        current = num_extra_lines - 1;
        prev = current - 1;
        next = current + 1;
        oldest = median_filter_width - 1;

        /* Now loop through the lines of the input image.  The index to
           retrieve pixels from the data needs to be set one less than
           desired since it will be incremented just after the dropped frame
           check. */
        filter_pixel_index = num_extra_lines - 1;
        for ( line = 0; line < num_lines; line++ )
        {
            /* See if the current line is a dropped frame and skip it */
            if ( temp_mask[line] & IAS_RPS_DROPLINE )
            {
                continue;
            }

            /* This index must not be updated until after the check for
               dropped frame.  This is because the index will have already
               stepped beyond the dropped frame. */
            filter_pixel_index++;

            /* Check to see if the next pixel for the filter is a dropped
               frame but only within the actual image data.  If this is a
               dropped frame skip it */
            found = FALSE;
            while ( !found && filter_pixel_index < num_lines )
            {
                if ( temp_mask[filter_pixel_index] & IAS_RPS_DROPLINE )
                {
                    filter_pixel_index++;
                }
                else
                    found = TRUE;
            }

            /* If the next filter pixel is off the end of the l0r_data we
               need to use the post-data fill from here on out. */
            if ( filter_pixel_index >= num_lines )
            {
                post_pixel_index = filter_pixel_index - num_lines;
                filter_pixels[oldest] = post_data_lines[post_pixel_index];
            }
            /* otherwise */
            else
            {
                filter_pixels[oldest] = TWODARRAY( l0r_data,
                        num_detectors, filter_pixel_index, det );
            }

            /* Adjust the indices.  Increment each one and, if they reach
               filter width, set to zero.  The comparison and set to zero is
               the most efficient way to maintain the indices of the 
               rotating array. */
            if ( ++current == median_filter_width )
                current = 0;
            if ( ++prev == median_filter_width )
                prev = 0;
            if ( ++next == median_filter_width )
                next = 0;
            if ( ++oldest == median_filter_width )
                oldest = 0;

            /* Up to this point everything has been about housekeeping.  The
               guts of the algorithm begin here. */

            /* Get the median value of the pixels */
            median_value = ias_math_find_median_unsigned(
                median_filter_width, filter_pixels );

            /* Calculate the two absolute differences */
            pixel_median_diff = median_value - filter_pixels[current];
            if ( pixel_median_diff < 0 )
                pixel_median_diff *= -1;
            neighbors_diff = filter_pixels[next] - filter_pixels[prev];
            if ( neighbors_diff < 0 )
                neighbors_diff *= -1;

            /* Calculate the threshold value based on the difference in the
               neighbor pixels. */
            if ( neighbors_diff > twice_detector_noise )
                threshold = neighbors_diff * pre_calc_gt_twice;
            else
                threshold = pre_calc_lte_twice;

            /* Determine if the pixel is impulse noise */
            if ( pixel_median_diff > threshold )
            {
                impulse_noise_count++;
                total_impulse_noise_count++;
                ias_pm_add_pixels( pixel_mask, det, line, 1,
                    IAS_RPS_IMPNOISE );
                if ( generate_report )
                {
                    if ( write_report( report_file, total_impulse_noise_count,
                        det+1, line, filter_pixels[current],
                        filter_pixels[prev], filter_pixels[next], FALSE,
                        report_data )
                        != SUCCESS )
                    {
                        free( pre_data_lines );
                        free( post_data_lines );
                        free( filter_pixels );
                        close_report_file( report_file, 
                            total_impulse_noise_count, report_data);
                        free( report_data );
                        free( temp_mask );
                        return ERROR;
                    }
                }
            }

        }  /* END for line */

        if ( store_in_database )
        {
            db_results->in_counts[det] = impulse_noise_count;
        }

    }  /* END for det loop */

    /* Clean up... */
    free( pre_data_lines );
    free( post_data_lines );
    free( filter_pixels );
    free( temp_mask );

    if ( generate_report )
    {
        if ( close_report_file( report_file, total_impulse_noise_count,
            report_data ))
        {
            free( report_data );
            return ERROR;
        }
    }

    free( report_data );
    return return_status;
}  /* END ias_rps_impulse_noise_characterization */


/* Open the report file and allocate temporary storage variables */
/*----------------------------------------------------------------------------
NAME:    open_report_file

PURPOSE: Open the report file and allocate temporary storage variables

RETURNS: File descriptor of the open file
         NULL on ERROR
-----------------------------------------------------------------------------*/
static FILE *open_report_file
(
    const IAS_REPORT *report_info,
    int band,
    int sca,
    struct report_struct *report_data
)
{
    FILE *report_file = NULL;
    char file_base_name[] = "impulse_noise_characterization";
    char *file_name;
    IAS_REPORT_HEADER *copy_of_report_header;
    int status;

    /* Get a report file name */
    file_name = ias_rps_get_report_file_name( report_info, band, sca,
        file_base_name );
    if ( !file_name )
    {
        IAS_LOG_ERROR("Getting an RPS report file name");
        return report_file;
    }

    /* These are dynamically allocated for threading purposes.  Allocating
       REPORT_BURST_SIZE so that we only do I/O in bursts instead of at each
       instance of impulse noise. */
    report_data->detectors = malloc( sizeof(int) * REPORT_BURST_SIZE );
    report_data->frames = malloc( sizeof(int) * REPORT_BURST_SIZE );
    report_data->value1 = malloc( sizeof(int) * REPORT_BURST_SIZE );
    report_data->value2 = malloc( sizeof(int) * REPORT_BURST_SIZE );
    report_data->value3 = malloc( sizeof(int) * REPORT_BURST_SIZE );
    if ( !report_data->detectors || !report_data->frames
        || !report_data->value1 || !report_data->value2
        || !report_data->value3 )
    {
        IAS_LOG_ERROR("Allocating report arrays");
        free( file_name );
        return report_file;
    }

    /* Open the file */
    report_file = fopen( file_name, "w" );
    if ( !report_file )
    {
        IAS_LOG_ERROR("Opening report file");
        return report_file;
    }
    free( file_name );
    file_name = NULL;

    /* Get the local copy of the report header since the "standard" header
       includes non-standard things such as band, SCA and algorithm name.
       In a threaded environment the input header cannot be manipulated. */
    copy_of_report_header = ias_misc_copy_report_header( report_info->header );
    if ( !copy_of_report_header )
    {
        IAS_LOG_ERROR("Creating local copy of report header");
        fclose(report_file);
        return NULL;
    }

    /* Initialize the band/SCA/application name information in the
       "standard" IAS report header */
    ias_misc_set_header_band_number( copy_of_report_header, band);
    ias_misc_set_header_sca_number( copy_of_report_header, sca);
    ias_misc_set_header_algorithm_name( copy_of_report_header,
        "Impulse Noise Characterization");

    status = ias_misc_write_report_header(report_file, copy_of_report_header);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Cannot write IAS header info to file %s", file_name);
        ias_misc_free_report_header( copy_of_report_header );
        fclose(report_file);
        return NULL;
    }

    ias_misc_free_report_header( copy_of_report_header );

    /* Write out the column titles */
    status = fprintf(report_file, 
                 "\nDet#  Line#     Impulse Noise Previous Neighbor "
                 "Next Neighbor\n");
    if ( status < 0 )
    {
        IAS_LOG_ERROR("Writing out report column titles");
        ias_misc_free_report_header( copy_of_report_header );
        fclose(report_file);
        return NULL;
    }

    return report_file;
}

/*----------------------------------------------------------------------------
NAME:    write_report

PURPOSE: Local routine to Write the data to the report.  The actual write is
         not done during each call, instead data is cached for burst writes of 
         REPORT_BURST_SIZE at a time.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
static int write_report
(
    FILE *report_file,
    int noise_count,
    int detector,
    int frame,
    uint16_t noise_value,
    uint16_t previous_neighbor,
    uint16_t next_neighbor,
    int flush,
    struct report_struct *report_data
)
{
    int index;
    int n;
    int i;

    index = ( noise_count - 1 ) % REPORT_BURST_SIZE;
    if ( !flush )
    {
        report_data->detectors[index] = detector;
        report_data->frames[index] = frame;
        report_data->value1[index] = noise_value;
        report_data->value2[index] = previous_neighbor;
        report_data->value3[index] = next_neighbor;
    }
    else if ( flush || index == REPORT_BURST_SIZE - 1 )
    {
        for ( i = 0; i <= index; i++ )
        {
            n = fprintf( report_file, "%3d  %5d  %13d  %13d  %12d\n",
                report_data->detectors[i], report_data->frames[i],
                report_data->value1[i], report_data->value2[i],
                report_data->value3[i] );
            if ( n < 0 )
            {
                IAS_LOG_ERROR("Writing to report file: n = %d", n );
                return ERROR;
            }
        }
    }

    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME:    close_report_file

PURPOSE: Local routine to close the report file.  This calls write_report
         with flush to make sure all information has been printed.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
static int close_report_file
(
    FILE *report_file,
    int noise_count,
    struct report_struct *report_data
)
{
    int status = SUCCESS;
    if ( noise_count > 0 )
    {
        status = write_report( report_file, noise_count, -1, -1,
            0, 0, 0, TRUE, report_data );
    }

    if ( report_data )
    {
        free( report_data->detectors );
        free( report_data->frames );
        free( report_data->value1 );
        free( report_data->value2 );
        free( report_data->value3 );
    }

    if ( fclose( report_file ))
    {
        IAS_LOG_ERROR("Closing report file");
        return ERROR;
    }
    return status;
}

/*----------------------------------------------------------------------------
        Database routines
-----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------
NAME:    ias_rps_store_impulse_noise_counts_in_database

PURPOSE: Store the counts of impulse noise in the database, the input 
         parameter message should be allocated by calling the
         ias_rps_db_allocate_impulse_noise_results routine.  On successful
         completion this routine frees message.  On ERROR the data is not
         freed.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int ias_rps_store_impulse_noise_counts_in_database
(
    void *message       /* I: Pointer to the
                               IAS_RPS_DB_IMPULSE_NOISE_CHAR_RESULTS
                           structure defined in
                               ias_rps_db_allocate_impulse_noise_results
                           A void pointer is used since it is called from the
                           threading routines and a void pointer is used to
                           support passing messages to different routines */
)
{
    IAS_RPS_DB_IMPULSE_NOISE_CHAR_RESULTS *db_results =
        (IAS_RPS_DB_IMPULSE_NOISE_CHAR_RESULTS *)message;
    int status;
    int number_db_columns;   /* Number of database columns being inserted */
    int number_db_records;   /* Number of database records being inserted */
    int det;                 /* loop variable for detector */
    int *detector_numbers;

    /* The impulse noise results will likely be sparse.  Squeeze out the
       detectors that don't have impulse noise events.  We'll also need an
       array of detector numbers. */
    detector_numbers = malloc( db_results->num_detectors * sizeof(int));
    if ( !detector_numbers )
    {
        IAS_LOG_ERROR("Error allocating detector number array");
        ias_rps_db_free_impulse_noise_results( db_results );
        return ERROR;
    }
    number_db_records = 0;
    for ( det = 0; det < db_results->num_detectors; det++ )
    {
        if ( db_results->in_counts[det] > 0 )
        {
            detector_numbers[number_db_records] = det + 1;
            db_results->in_counts[number_db_records] =
                db_results->in_counts[det];
            number_db_records++;
        }
    }

    /* If there were no records return since it is an error to call the
       insert with 0 records */
    if ( number_db_records == 0 )
    {
        free( detector_numbers );
        ias_rps_db_free_impulse_noise_results( db_results );
        return SUCCESS;
    }

    /* Create the impulse noise table */
    IAS_DB_INSERT_TABLE table[] =
    {
        IAS_DB_TABLE_STRING( "radiometric_wo_common_id",
                db_results->rad_common_id ),
        IAS_DB_TABLE_INT( "band_id", db_results->band ),
        IAS_DB_TABLE_INT( "sensor_chip_assembly_id", db_results->sca ),
        IAS_DB_TABLE_INT_ARRAY( "detector_id", detector_numbers ),
        IAS_DB_TABLE_INT_ARRAY( "nbr_pixels", db_results->in_counts )
    };

    /* Get the number of columns inserted into the table and insert the
       data into the database */
    number_db_columns = sizeof(table) / sizeof(IAS_DB_INSERT_TABLE);

    status = ias_db_insert_using_table( db_results->dbase,
        "impulse_noise", table, number_db_columns,
        number_db_records );
    if ( status != SUCCESS )
    {
        IAS_LOG_ERROR("Inserting impulse noise for Band %d, SCA %d",
            db_results->band, db_results->sca );
    }
    
    free( detector_numbers );
    ias_rps_db_free_impulse_noise_results( db_results );
    return status;
}


/*----------------------------------------------------------------------------
NAME:    ias_rps_db_allocate_impulse_noise_results

PURPOSE: Allocate memory for impulse noise db structure

RETURNS: pointer to the IAS_RPS_DB_IMPULSE_NOISE_CHAR_RESULTS structure
         NULL on ERROR
-----------------------------------------------------------------------------*/
IAS_RPS_DB_IMPULSE_NOISE_CHAR_RESULTS *ias_rps_db_allocate_impulse_noise_results
(
    int number_of_detectors /* I: Number of detector array elements to
                                   allocate */
)
{
    IAS_RPS_DB_IMPULSE_NOISE_CHAR_RESULTS *db_results = malloc(
        sizeof(IAS_RPS_DB_IMPULSE_NOISE_CHAR_RESULTS) );
    if ( !db_results )
    {
        IAS_LOG_ERROR("Allocating the database information structure");
        return NULL;
    }

    db_results->in_counts = malloc( sizeof(int) * number_of_detectors );
    if ( !db_results->in_counts )
    {
        IAS_LOG_ERROR("Allocating database storage memory");
        free( db_results );
        return NULL;
    }

    return db_results;
}

/*----------------------------------------------------------------------------
NAME:    ias_rps_db_free_impulse_noise_results

PURPOSE: Free the allocated impulse noise results db structure

RETURNS: None
-----------------------------------------------------------------------------*/
void ias_rps_db_free_impulse_noise_results
(
    IAS_RPS_DB_IMPULSE_NOISE_CHAR_RESULTS *db_results
                               /* I: Structure to be freed */
)
{
    free( db_results->in_counts );
    free( db_results );
}
