/*****************************************************************************
 NAME:    ias_rps_lookup_response_linearization

 PURPOSE: Adjusts the DN values for the OLI/TIRS image bands to compensate
          for lookup response linearization correction.

 RETURNS: SUCCESS or ERROR

 REFERENCES:  Additional detailed information describing the algorithm can be
              found in the Response Linearization ADDs
*****************************************************************************/
#include <stdlib.h>                /* For NULL definition */
#include "ias_logging.h"
#include "ias_radiometry.h"
#include "ias_rlut.h"

/*****************************************************************************
    Forward declaration of private routines in this file
*****************************************************************************/

static int index_search
(
    float search_array[],   /* I: Array to be searched */
    float inval,            /* I: Value to search for in array */
    int array_size          /* I: Length of array to search */
);

/* Begin main function */
int ias_rps_lookup_response_linearization
(
    float *data_buffer,             /* I/O: On input it is the DN from sensor
                                      On output it is the adjusted DN
                                      with lookup response linearization
                                      correction */
    const IAS_RLUT_LOOKUP_TABLE *params,
                                    /* I: Array of lookup linearization 
                                      parameters for all detectors in the 
                                      current band/SCA */
    int number_lines                /* I: Number of lines for this SCA */
)
{
    int det;                        /* Detector loop index */
    int line;                       /* Line loop index */

    float adjustment;               /* The adjustment value from RLUT */
    int index;                      /* The index of the rlut array used to
                                       retrieve the DN */
    int number_detectors;           /* Number of detectors */
    unsigned int out_of_range_count = 0;

    /* Sanity check */
    if ((data_buffer == NULL) || (params == NULL))
    {
        IAS_LOG_ERROR("Missing input image data buffer and/or "
            "linearization parameter data buffer");
        return ERROR;
    }

    /* Grab the number of detectors from the rlut parameters */
    number_detectors = params->num_detectors;

    /* Loop through the detectors.  An outer detector loop is used here
       since the same thresholds and remapping coefficients are used for
       all the detector samples. */
    for (det = 0; det < number_detectors; det++)
    {
        float *value_ptr;              /* This is used to point to the address
                                       of the data value being modified */
        /* Loop through the lines */
        value_ptr = &data_buffer[det];
        int offset = det * params->num_values;
        float low_value = params->input_dn[offset];
        float high_value = params->input_dn[offset + params->num_values - 1];
        float *dn_ptr = &(params->input_dn[offset]);

        for (line = 0; line < number_lines; line++)
        {
            float value = *value_ptr;       /* Current image pixel value */

            /* Determine if the value is within range but not exactly at the 
               max range by comparing to the last value in the array.  
               If it's valid, start the algorithm.  If not, no adjustment
               will be made. */
            if ( value < high_value && value > low_value )
            {
                /* Find the index to use as bracket for the algorithm */
                index = index_search( dn_ptr, value, params->num_values);
                if (index < 0)
                {
                    /* Since the input value was verified, a negative index 
                       indicates a problem with the rlut values */
                    IAS_LOG_ERROR(
                            "Value was not found in lookup RLUT search");
                    return ERROR;
                }

                /* The index returned from the range search represents
                   the array value from zero for that detector 
                   alone.  the algorithm arrays contain all detectors and 
                   values.  Add the (number of detectors * num_values)
                   to that index for the full array index. */
                index += offset;

                /* If the consecutive values in the DN array are the same value,
                   an error here indicates an error in the RLUT array */
                if (params->input_dn[index] == params->input_dn[index + 1])
                {
                    IAS_LOG_ERROR(
                            "Invalid consecutive values in RLUT index found");
                    return ERROR;
                }

                /* Perform interpolation to determine the adjustment value */
                adjustment = params->output_correction[index] + 
                    (params->output_correction[index + 1] - 
                    params->output_correction[index]) *
                    (value - params->input_dn[index])/
                    (params->input_dn[index + 1] - params->input_dn[index]);
            
                value += adjustment;

                /* Assign to output value */
                *value_ptr = value;

            }   /* End of Valid pixel value condition */
            else
            {
                out_of_range_count++;
                IAS_LOG_DEBUG("Value %f is outside the range of the RLUT: "
                    "%f - %f", value, low_value, high_value );
            }

            /* Point to the next data value in line */
            value_ptr += number_detectors;

        }  /* END for line loop */
    }  /* END for detector loop */

    if ( out_of_range_count )
    {
        /* It is considered normal to get out of range values, especially for
           dark images with radiances near zero. So only issue this message
           as a debug message. */
        IAS_LOG_DEBUG("%u out of range values encountered",
            out_of_range_count );
    }

    return SUCCESS;
}  /* END ias_rps_lookup_response_linearization */

/*****************************************************************************
    NAME: index_search

    PURPOSE:  This routine does a binary search of an array for indexes 
        that bracket a give input value.  Returned is the largest index 
        value where:

                search_array[index] <= inval.

    RETURNS:  The resulting index if successful.
              -1 if the index is not found.
     
*****************************************************************************/
static int index_search
(
    float search_array[],   /* I: Array to be searched */
    float inval,            /* I: Value to search for in array */
    int array_size          /* I: Length of array to search */
)
{
    int middle_position;    /* Used in binary search loop */

    int result = -1.0;      /* Output index, -1 indicates value not found */
    float low = 0;          /* Used in binary search loop */
    float high = array_size - 1;  /* Used in binary search loop */

    /* Begin looping through the array.
       If the result was found, it will be assigned a value which is not -1, 
       as the index of an array can not be negative.
       If Low is greater than High, the value can not be found */
    while ((result == -1) && (low <= high))
    {
        /* Find the half position */
        middle_position = (low + high) / 2;

        if (inval >= search_array[middle_position])
        {
            if (inval < search_array[middle_position + 1])
            {
                result = middle_position;
            }
            else    /* The value is higher in list */
            {
                low = middle_position + 1;
            }
        }
        else    /* The value is lower in the list */
        {
            high = middle_position - 1;
        }
    }

    return result;
}

