/*----------------------------------------------------------------------------
NAME:    ias_rps_saturated_pixel_characterization

PURPOSE: Identify the locations of saturation pixels in the data.

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <math.h>

#include "ias_const.h"
#include "ias_math.h"
#include "ias_logging.h"
#include "ias_miscellaneous.h"
#include "ias_pixel_mask.h"
#include "ias_db_insert.h"
#include "ias_radiometry.h"
#include "ias_satellite_attributes.h"

/* This structure is used to pass parameters to the output report */
typedef struct output_report
{
    int  det;
    int  line;
    int  pixel_value;
} OUTPUT_REPORT;


/* MACRO for low and high analog pixels */
#define LOW 0
#define HIGH 1

/* Prototypes for report generation */
static FILE *create_report_file( int, int, int, const IAS_REPORT * );
static void close_report_file( FILE * );
static int write_report_file( FILE *, OUTPUT_REPORT **, int, int );

/* Characterize saturated pixel */
int ias_rps_saturated_pixel_characterization
(
    const uint16_t *l0r_data,   /* I: One SCA of level 0 data in DN */
    int band,                   /* I: Band number, 1-based */
    int sca,                    /* I: SCA number, 1-based */
    int num_detectors,          /* I: Number of detectors in the SCA */
    int num_lines,              /* I: Number of lines in the data */
    const int *digital_low_saturated, /* I: Digital low saturated level */
    const int *digital_high_saturated,/* I: Digital high saturated level */
    const int *analog_low_saturated,  /* I: Analog low saturated level */
    const int *analog_high_saturated, /* I: Analog high saturated level */
    IAS_PIXEL_MASK *pixel_mask, /* I: Contains inop detector and other
                                       ignore pixel locations
                                   O - Saturated pixel locations */
    const IAS_REPORT *report_info, /* I: General information for the test
                                       report.  If no report is desired pass
                                       NULL. */
    IAS_RPS_DB_SATURATED_CHARACTERIZATION_RESULTS *db_results
                                /* I/O - Pointer to results of the 
                                       saturated pixel characterization. */
)
{
    int line;                    /* Loop counter for the lines */
    int det;                     /* Loop counter for the detectors */

    uint16_t sample;             /* This buffer contains the sample to be
                                    saturated judged */
    PIXEL_MASK_TYPE artifacts_to_ignore; /* Mask values of pixels to ignore */
    PIXEL_MASK_TYPE *temp_mask;  /* Line of pixel masks from image */
    FILE *report_file = NULL;    /* File handle for the report */
    int return_status = SUCCESS; /* Status to return to caller */
    int generate_report = TRUE;  /* Convenience flag */
    int db_writing = TRUE;       /* Convenience flag */
    int total_analog_high_pixels;/* Counts of total analog high pixels */ 
    int total_analog_low_pixels; /* Counts of total analog low pixels */
    int low_index;               /* Index for analog low pixels */
    int high_index;              /* Index for analog high pixels */
    int mem_low_size;            /* Number used for memory allocated for 
                                    analog low output structure  */
    int mem_high_size;           /* Number used for memory allocated for
                                    analog high output structure */
    int status;                  /* Function return status */

    int start_range;             /* Start of saturated pixel range */
    OUTPUT_REPORT *output_report_data[2]; /* Array for output report data
                                             for both LOW and HIGH
                                             saturation types */
    PIXEL_MASK_TYPE sat_range_type; /* Current saturation range type */
    PIXEL_MASK_TYPE current_saturation; /* Saturation value for current line */

    mem_low_size = 100;  /* Temporary size for the analog low saturated pixel 
                            counts */
    mem_high_size = 100; /* Temporary size for the analog high saturated pixel 
                            counts */

    /* Initialize output_report_data array */
    output_report_data[LOW] = NULL;
    output_report_data[HIGH] = NULL;

    /* Set the flags false if NULL was passed in the report structures. */
    if (report_info == NULL)
        generate_report = FALSE;
    if (db_results == NULL)
        db_writing = FALSE;

    /* Open the report file and write the standards header */
    if (generate_report)
    {
        /* Allocate memory for output report data pointers */
        output_report_data[LOW] = malloc(mem_low_size * 
                                          sizeof(OUTPUT_REPORT));
        output_report_data[HIGH] = malloc(mem_high_size * 
                                          sizeof(OUTPUT_REPORT));
        if (!output_report_data[LOW] || !output_report_data[HIGH])
        {
            free(output_report_data[LOW]);
            free(output_report_data[HIGH]);
            IAS_LOG_ERROR("Allocating memory for output report data pointers");
            return ERROR;
        }

        /* Create report file */
        report_file = create_report_file(band, sca, num_detectors,
            report_info);
        if ( !report_file )
        {
            free(output_report_data[LOW]);
            free(output_report_data[HIGH]);
            return ERROR;
        }
    }

    /* Create a convenience mask of artifacts to ignore */
    artifacts_to_ignore = IAS_RPS_DROPLINE|IAS_RPS_IMPNOISE;

    /* Allocate the temporary lines of pixel masks */
    temp_mask = malloc( sizeof(PIXEL_MASK_TYPE) * num_lines );
    if (!temp_mask)
    {
        IAS_LOG_ERROR("Allocating pixel mask memory");
        if (generate_report)
        {
            free(output_report_data[LOW]);
            free(output_report_data[HIGH]);
            close_report_file(report_file);
        }
        return ERROR;
    }

    /* Initialize the total pixel counts */
    total_analog_high_pixels = 0;
    total_analog_low_pixels = 0;

    /* Loop through each detector of the input data */
    for (det = 0; det < num_detectors; det++)
    {
        /* Assign detector number for database outputs */
        if (db_writing)
        {
            db_results->detector[det] = det + 1;
        }

        /* Grab the pixel mask information for all frames of data for
           the current detector */
        status = ias_pm_get_image_subset(pixel_mask, det, 0, 1,
             num_lines, temp_mask);
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Retrieving pixel mask subset for detector %d",
                          (det + 1));
            if (generate_report)
            {
                free(output_report_data[HIGH]);
                free(output_report_data[LOW]);
                close_report_file(report_file);
            }
            free(temp_mask);
            return ERROR;
        }

        /* Initialize the saturated pixel counts to zeros */
        if (db_writing)
        {
            db_results->counts_digital_high[det] = 0;
            db_results->counts_digital_low[det] = 0;
            db_results->counts_analog_high[det] = 0;
            db_results->counts_analog_low[det] = 0;
        }

        /* Initialize saturation range tracking variables */
        start_range = 0;
        sat_range_type = IAS_RPS_NOVALUE;

        /* Now loop through the lines of the input image.   */
        for (line = 0; line < num_lines; line++)
        {
            /* See if the current line is a dropped frame, fill frame,
               or saturated noise and skip it */
            if (temp_mask[line] & artifacts_to_ignore)
            {
                continue;
            }

            /* Initialize the saturation value for the current line */
            current_saturation = IAS_RPS_NOVALUE;

            /* Up to this point everything has been about housekeeping.  The
               guts of the algorithm begin here. */
            sample = *((l0r_data + (line * num_detectors)) + det);

            /* Digital high pixel case: populate digital high saturated pixel
               counts to database structure, and output positions to the pixel
               mask */
            if (sample == digital_high_saturated[det])
            {
                if (db_writing)
                {
                    db_results->counts_digital_high[det]++;
                }
                current_saturation = IAS_RPS_ADCONV_SAT_HI;
            }

            /* Digital low pixel case: populate digital low saturated pixel 
               counts to database structure, and output positions to the pixel 
               mask */
            if (sample == digital_low_saturated[det])
            {
                if (db_writing)
                {
                    db_results->counts_digital_low[det]++;
                }
                current_saturation = IAS_RPS_ADCONV_SAT_LO;
            }

            /* populate analog low saturated pixel counts to database 
               structure, and output positions to the pixel mask, also
               output positions to the output report structure */
            /* Get the analog low pixels when the pixel value is larger
               than digital_low_saturated[det] and smaller than
               analog_low_saturated[det] */
            if (sample > digital_low_saturated[det]
                && sample < analog_low_saturated[det])
            {
                if (db_writing)
                {
                    db_results->counts_analog_low[det]++;
                }
                total_analog_low_pixels++;
                current_saturation = IAS_RPS_ANALOG_SAT_LO;
                
                if (generate_report)
                {
                    low_index = total_analog_low_pixels - 1;
                    if (low_index >= mem_low_size)
                    {
                        mem_low_size += low_index;
                        output_report_data[LOW] =  realloc(
                                  output_report_data[LOW],
                                  mem_low_size * sizeof(OUTPUT_REPORT));
                        if (output_report_data[LOW] == NULL)
                        {
                            free(temp_mask);
                            free(output_report_data[LOW]);
                            free(output_report_data[HIGH]);
                            close_report_file(report_file);
                            IAS_LOG_ERROR("Reallocating memory for output" 
                                      " report data pointers");
                        }
                    }
                    output_report_data[LOW][low_index].det = det;
                    output_report_data[LOW][low_index].line = line;
                    output_report_data[LOW][low_index].pixel_value =
                                   sample;
                }
            }

            /* populate analog high saturated pixel counts to database 
               structure, and output positions to the pixel mask, also
               output positions to the output report structure */
            /* Get the analog high pixels when the pixel value is larger
               than analog_high_saturated[det] and smaller than
               digital_high_saturated[det] */
            if (sample > analog_high_saturated[det]
                && sample < digital_high_saturated[det])
            {
                if (db_writing)
                {
                    db_results->counts_analog_high[det]++;
                }
                total_analog_high_pixels++;
                current_saturation = IAS_RPS_ANALOG_SAT_HI;
                
                if (generate_report)
                {
                    high_index = total_analog_high_pixels - 1;
                    if (high_index >= mem_high_size)
                    {
                        mem_high_size += high_index;
                        output_report_data[HIGH] = realloc(
                               output_report_data[HIGH], 
                               mem_high_size * sizeof(OUTPUT_REPORT));
                        if (!output_report_data[HIGH])
                        {
                            free(temp_mask);
                            free(output_report_data[LOW]);
                            free(output_report_data[HIGH]);
                            close_report_file(report_file);
                            IAS_LOG_ERROR("Reallocating memory for output" 
                                      " report data pointers");
                        }
                    }
                    output_report_data[HIGH][high_index].det = det;
                    output_report_data[HIGH][high_index].line = line;
                    output_report_data[HIGH][high_index].pixel_value =
                                   sample;
                }
            }

            /* If a new range of saturation values has started, add the current
               saturation value range to the pixel mask */
            if(current_saturation != sat_range_type)
            {
                /* Only ranges of saturation values need to be added to the
                   pixel mask. Ignore IAS_RPS_NOVALUE range. */
                if(sat_range_type != IAS_RPS_NOVALUE)
                {
                    ias_pm_add_pixels(pixel_mask, det, start_range,
                        (line - start_range), sat_range_type);
                }
                
                /* Re-initial the range tracking variables for the new range */
                start_range = line;
                sat_range_type = current_saturation;
            }
        }  /* END for line */

        /* If neccessary, add the final saturation value range to the
           pixel mask before moving on to the next detector */
        if(sat_range_type != IAS_RPS_NOVALUE)
        {
            ias_pm_add_pixels(pixel_mask, det, start_range,
                (num_lines - start_range), sat_range_type);
        }

    }  /* END for det loop */

    /* Output report file and clean up */
    if (generate_report)
    {
        if (write_report_file(report_file, output_report_data, 
            total_analog_high_pixels, total_analog_low_pixels) == ERROR)
        {
            IAS_LOG_ERROR("Writing report file");
            return_status = ERROR;
        }
        close_report_file(report_file);

        /* free the allocated memory */
        output_report_data[LOW] = realloc(output_report_data[LOW], 0);
        output_report_data[HIGH] = realloc(output_report_data[HIGH], 0);
    }
    free(temp_mask);

    return return_status;
}  /* END ias_rps_saturated_pixel_characterization */


/*----------------------------------------------------------------------------
NAME: create_report_file

PURPOSE: Open the report file and allocate temporary storage variables.

RETURNS: Success or Error
-----------------------------------------------------------------------------*/
static FILE *create_report_file
(
    int band,                   /* 1-based band number */
    int sca,                    /* 1-based sca number */
    int num_detectors,          /* number of detectors */
    const IAS_REPORT *report_info /* IAS report structure pointer */
)
{
    FILE *report_file = NULL;
    char file_base_name[] = "saturated_pixel_characterization";
    char *file_name;
    int status;
    IAS_REPORT_HEADER *copy_of_report_header;

    /* Get output report file name */
    file_name = ias_rps_get_report_file_name(report_info, band, sca, 
                                             file_base_name);
    if (file_name == NULL)
    {
        IAS_LOG_ERROR("Getting report filename %s", file_name);
        return report_file;        
    }

    /* Open the report file */
    report_file = fopen(file_name, "w");
    if ( !report_file )
    {
        IAS_LOG_ERROR("Opening report file %s", file_name);
        return report_file;
    }
    free(file_name);
    file_name = NULL;

    /* Get the local copy of the report header since the "standard" header
       includes non-standard things such as band, SCA and algorithm name.
       In a threaded environment the input header cannot be manipulated. */
    copy_of_report_header = ias_misc_copy_report_header( report_info->header );
    if ( !copy_of_report_header )
    {
        IAS_LOG_ERROR("Creating local copy of report header");
        fclose(report_file);
        return NULL;
    }

    /* Initialize the band/SCA/application name information in the
       "standard" IAS report header */
    ias_misc_set_header_band_number( copy_of_report_header, band);
    ias_misc_set_header_sca_number( copy_of_report_header, sca);
    ias_misc_set_header_algorithm_name( copy_of_report_header,
        "Saturated Pixel Characterization");

    status = ias_misc_write_report_header(report_file, copy_of_report_header);
    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Cannot write IAS header info to file %s", file_name);
        ias_misc_free_report_header( copy_of_report_header );
        fclose(report_file);
        return NULL;
    }

    ias_misc_free_report_header( copy_of_report_header );
    return report_file;
}

/*----------------------------------------------------------------------------
NAME: write_report_file

PURPOSE: Write the data to the report.  The actual write is done at one time.

RETURNS: Success or Error
-----------------------------------------------------------------------------*/
static int write_report_file
(
    FILE *report_file,              /* Output file pointer */
    OUTPUT_REPORT **report,         /* Output report structure pointer */  
    int total_high_counts,          /* Total high saturated pixel counts */
    int total_low_counts            /* Total low saturated pixel counts */
)
{
    int status;
    int index;
   
    /* Print out the total analog low pixel counts and the pixel positions */ 
    status = fprintf(report_file, " There are %d analog low"
                     " saturated pixels\n", total_low_counts);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing report data");
        return ERROR;
    }

    if (total_low_counts != 0)
    {
        status = fprintf(report_file, 
                 "Det#   Line#   DN \n");
        if ( status < 0 )
        {
            IAS_LOG_ERROR("Writing report data");
            return ERROR;
        }

        for (index = 0; index < total_low_counts; index++)
        {
            status = fprintf( report_file, "%5d %5d %5d\n",
                report[LOW][index].det + 1, report[LOW][index].line + 1,
                report[LOW][index].pixel_value );
            if (status < 0)    
            {
                IAS_LOG_ERROR("Writing report data");
                return ERROR;
            }
        }   
    }

    /* Print out the total analog high pixel counts and the pixel positions */ 
    status = fprintf(report_file, " There are %d analog high"
                     " saturated pixels\n", total_high_counts);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing report data");
        return ERROR;
    }
    if (total_high_counts != 0)
    {
        status = fprintf(report_file, 
                 "Det#   Line#   DN \n");
        if (status < 0)
        {
            IAS_LOG_ERROR("Writing report data");
            return ERROR;
        }

        for (index = 0; index < total_high_counts; index++)
        {
            status = fprintf( report_file, "%5d %5d %5d\n",
                report[HIGH][index].det + 1, report[HIGH][index].line + 1,
                report[HIGH][index].pixel_value );
            if (status < 0)    
            {
                IAS_LOG_ERROR("Writing report data");
                return ERROR;
            }
        }   
    }   
    return SUCCESS;
}

/*----------------------------------------------------------------------------
NAME: close_report_file

PURPOSE: Close the report file.

RETURNS: None
-----------------------------------------------------------------------------*/
static void close_report_file
(
     FILE *report_file                /* Output file pointer */
)
{
    if (!report_file)
        return;

    fclose(report_file);
    return;
}

/*-------------------------------------------------------------------------
   Database Section (all are public routines) 
-------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------
NAME:  ias_rps_db_allocate_saturated_results  

PURPOSE: Allocate memory for saturated results db structure

RETURNS: pointer to the IAS_RPS_DB_SATURATED_CHARACTERIZATION_RESULTS structure
-----------------------------------------------------------------------------*/
IAS_RPS_DB_SATURATED_CHARACTERIZATION_RESULTS 
    *ias_rps_db_allocate_saturated_results
(
    int number_of_detectors /* I: Number of array elements to allocate */
)
{
    IAS_RPS_DB_SATURATED_CHARACTERIZATION_RESULTS *db_results;

    db_results = malloc(
        sizeof(IAS_RPS_DB_SATURATED_CHARACTERIZATION_RESULTS));
    if (!db_results)
    {
        ias_rps_db_free_saturated_results(db_results);
        IAS_LOG_ERROR("Allocating memory for DB results structure");
        return NULL;
    }

    db_results->num_detectors = number_of_detectors;
    db_results->detector = malloc(number_of_detectors * sizeof (int));
    db_results->counts_digital_high = 
                      malloc(number_of_detectors * sizeof (int));
    db_results->counts_digital_low = 
                      malloc(number_of_detectors * sizeof (int));
    db_results->counts_analog_high = 
                      malloc(number_of_detectors * sizeof (int));
    db_results->counts_analog_low = 
                      malloc(number_of_detectors * sizeof (int));
    if (!db_results->detector || !db_results->counts_digital_high
        || !db_results->counts_digital_low || !db_results->counts_analog_high 
        || !db_results->counts_analog_low)
    {
        ias_rps_db_free_saturated_results(db_results);
        IAS_LOG_ERROR("Allocating memory for DB results elements");
        return NULL;
    }
    return db_results;
}

/*----------------------------------------------------------------------------
NAME:  ias_rps_store_saturated_pixels_in_database 

PURPOSE: Storing the saturated results into db table, the message needs to be 
         memory allocated first by calling 
         ias_rps_db_allocate_saturated_results routine

RETURNS: SUCCESS or ERROR
-----------------------------------------------------------------------------*/
int ias_rps_store_saturated_pixels_in_database
(
    void *message /* I: Pointer to IAS_RPS_DB_SATURATED_CHARACTERIZATION_RESULTS
                         structure.  A void pointer is used since it is
                         called from the threading routines and a void
                         pointer is used to support passing messages to
                         different routines */
)
{
    IAS_RPS_DB_SATURATED_CHARACTERIZATION_RESULTS *db_results =
        (IAS_RPS_DB_SATURATED_CHARACTERIZATION_RESULTS *)message;
    int status = SUCCESS;    /* function return status */
    int number_db_columns;   /* Number of database columns being inserted */
    int number_db_records;   /* Number of database records being inserted */
    int det;                 /* loop variable for detector */
    
    /* Initialize the number of database records variable */
    number_db_records = 0;

    /* Remove the saturated pixels for one detector if all four kinds of counts
       are zero */
    for ( det = 0; det < db_results->num_detectors; det++ )
    {
        /* The database results are most likely sparse arrays, this squeezes 
           out the all non-saturated pixel events so they aren't stored in the 
           database */
        if ((db_results->counts_digital_high[det] != 0) || 
           (db_results->counts_digital_low[det] != 0) ||
           (db_results->counts_analog_high[det] != 0) ||
           (db_results->counts_analog_low[det] != 0))
        {
            db_results->detector[number_db_records] 
                = db_results->detector[det]; 
            db_results->counts_digital_high[number_db_records]
                =db_results->counts_digital_high[det]; 
            db_results->counts_digital_low[number_db_records]
                =db_results->counts_digital_low[det]; 
            db_results->counts_analog_high[number_db_records]
                =db_results->counts_analog_high[det]; 
            db_results->counts_analog_low[number_db_records]
                =db_results->counts_analog_low[det]; 
            number_db_records++;
        }
    }

    /* If there were no records return since it is an error to call the
       insert with 0 records */
    if (number_db_records == 0)
    {
        ias_rps_db_free_saturated_results(db_results);
        return SUCCESS;
    }

    /* Create the saturated pixel table */
    IAS_DB_INSERT_TABLE table[]=
    {
        IAS_DB_TABLE_STRING( "radiometric_wo_common_id",
                db_results->rad_common_id ),
        IAS_DB_TABLE_INT( "band_id", db_results->band ),
        IAS_DB_TABLE_INT( "sensor_chip_assembly_id", db_results->sca ),
        IAS_DB_TABLE_INT_ARRAY( "detector_id",
                db_results->detector ),
        IAS_DB_TABLE_INT_ARRAY( "digital_low_sat_pixels", 
                           db_results->counts_digital_low ),
        IAS_DB_TABLE_INT_ARRAY( "digital_high_sat_pixels", 
                           db_results->counts_digital_high ),
        IAS_DB_TABLE_INT_ARRAY( "analog_low_sat_pixels", 
                           db_results->counts_analog_low ),
        IAS_DB_TABLE_INT_ARRAY( "analog_high_sat_pixels", 
                           db_results->counts_analog_high )
    };

    /* Get the number of columns inserted into the table and insert the
       data into the database */
    number_db_columns = sizeof(table) / sizeof(IAS_DB_INSERT_TABLE);

    status = ias_db_insert_using_table( db_results->dbase,
        "saturated_pixels", table, number_db_columns,
        number_db_records );
    if ( status )
    {
        IAS_LOG_ERROR("Inserting saturated pixel for band %d, "
            "sca %d", db_results->band, db_results->sca );
    }
    
    ias_rps_db_free_saturated_results(db_results);

    return status;
}

/*----------------------------------------------------------------------------
NAME:  ias_rps_db_free_saturated_results

PURPOSE: Free the allocated saturated results db structure

RETURNS: None
-----------------------------------------------------------------------------*/
void ias_rps_db_free_saturated_results
(
    IAS_RPS_DB_SATURATED_CHARACTERIZATION_RESULTS *db_results /* I: structure
                                                                 to free */
)
{
    free(db_results->detector);
    free(db_results->counts_digital_high);
    free(db_results->counts_digital_low);
    free(db_results->counts_analog_high);
    free(db_results->counts_analog_low);
    free(db_results);
}

