/**************************************************************************
 NAME:              ias_rps_tirs_dark_response_determination

 PURPOSE:   Module implementing routines that can be used for per-scene
            or offline determination of TIRS dark response

***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>

#include "ias_logging.h"
#include "ias_pixel_mask.h"
#include "ias_radiometry.h"

#include "ias_rps_histogram_statistics_characterization.h"
#include "ias_rps_tirs_dark_response_determination.h"


/*-------------------------------------------------------------------------
 NAME:              ias_rps_tirs_dark_response_determination

 PURPOSE:  Determines the TIRS dark response for a given band/SCA

 RETURNS:  Integer status code of SUCCESS or ERROR

 ALGORITHM REFERENCES:  TIRS Dark Response Determination ADD, V3.0a

---------------------------------------------------------------------------*/
int ias_rps_tirs_dark_response_determination
(
    IAS_RPS_HISTOGRAM_INTERMEDIATE_RESULTS *blind_stats,
                                           /* I: Histogram statistics for
                                              the current band/SCA */
    double *dark_response,                 /* I/O: CPF baseline dark response
                                              for the current band/SCA
                                              (as input and output IF
                                              use_cpf_baseline_response
                                              flag is set to 1); otherwise,
                                              adjusted dark response for
                                              the current band/SCA
                                              (output) */
    double *sca_dark_response_difference,  /* I/O: Mean dark difference
                                              for the current SCA (input
                                              for imaging bands, output
                                              for blind band) */
    double abs_gain,                       /* Absolute gain for the current
                                              band/SCA */
    const double *relative_gains,          /* Detector relative gains for
                                              the current band/SCA */
    int is_blind_band,                     /* I: Flag indicating this band
                                              is the blind band */
    int number_detectors,                  /* I: Number of detectors for
                                              the current band/SCA */
    IAS_PIXEL_MASK *pmask                  /* I: Pixel mask for the current
                                              band/SCA */
)
{
    double detector_gain = 0.0;            /* Absolute gains for each
                                              detector in the current
                                              band/SCA */
    double adjustment_factor = 0.0;        /* Factor that corrects the
                                              observed baseline dark
                                              response */
    double dark_response_difference = 0.0; /* "Local" copy of
                                              SCA dark response difference
                                              parameter */
    double response_sum = 0.0;             /* Running sum counter for
                                              detector dark responses */
    double diff = 0.0;                     /* Current difference */

    int detector_count = 0;                /* Operable detector count */
    int det_index = 0;                     /* Current detector */

    PIXEL_MASK_TYPE mask = 0;              /* Mask value to use if searching
                                              pixel mask */


    /* Depending on the band number, the algorithm has to do the
       following:
       a.  For the blind band, compute and return the mean dark difference
           for the specified SCA
       b.  For all bands, take the specified mean dark difference and 
           compute adjustment factors (based on the detector gain) that
           yield the adjusted dark response when applied to the baseline
           dark response */
    if (is_blind_band)
    {
        /* Make sure the blind band histogram statistics means are
           available */
        if (blind_stats == NULL)
        {
            IAS_LOG_ERROR("No blind band histogram statistics means "
                "available");
            return ERROR;
        }

        /* Set the search mask just to look for inoperable detectors */
        mask = IAS_RPS_INOP_DETECTOR;

        for (det_index = 0; det_index < number_detectors; det_index++)
        {
            /* Calculate the absolute detector gain */
            detector_gain = abs_gain * relative_gains[det_index];

            /* Is the detector flagged as inoperable in the pixel mask?
               If it is, skip it */
            if (ias_pm_is_detector_flagged(pmask, det_index, mask))
            {
               continue;
            }

            /* Calculate the adjusted dark response difference for
               the detector */
            diff =
                dark_response[det_index] - blind_stats[det_index].mean[0];

            response_sum += (diff / detector_gain);
            detector_count++;
        }

        /* Calculate the SCA dark response difference */
        if (detector_count == 0)
        {
            IAS_LOG_ERROR("No data from operable detectors for calculating "
                "adjusted dark response difference");
            return ERROR;
        }
        *sca_dark_response_difference = response_sum / detector_count;
    }

    /* Pull the SCA dark response difference into a local variable
       to optimize its access */
    dark_response_difference = *sca_dark_response_difference;

    for (det_index = 0; det_index < number_detectors; det_index++)
    {
        /* Calculate detector gain */
        detector_gain = abs_gain * relative_gains[det_index];

        /* Calculate the amount by which to adjust the baseline dark
           response */
        adjustment_factor = (detector_gain * dark_response_difference);

        /* Adjust the baseline dark response */
        dark_response[det_index] += adjustment_factor;
    }

    /* Done */
    return SUCCESS;
}




/*-------------------------------------------------------------------------
 NAME:         ias_rps_tirs_dark_response_write_to_group_file

 PURPOSE:  Writes the adjusted dark response to an ASCII data file that can
           later be merged with an existing CPF to create a "test" CPF

 RETURNS:  Integer status code of SUCCESS or ERROR

---------------------------------------------------------------------------*/
int ias_rps_tirs_dark_response_write_to_group_file
(
    const double *adjusted_dark_response,  /* I: Adjusted dark response
                                              array for the current
                                              band/SCA */
    int is_blind_band,                     /* I: Flag indicating whether
                                              blind band is being
                                              processed now */
    int band_number,                       /* I: TIRS band number */
    int sca_number,                        /* I: Current SCA number */
    int number_detectors                   /* I: Total number of detectors
                                              in the current band/SCA */
)
{
    char group_filename[PATH_MAX];         /* Name of band/SCA group
                                              file (with path) */
    char groupfile_basename[] = "cpf_dark_response_group";
    char nonblind_group_header[] = "GROUP = TIRS_DETECTOR_RESPONSE";
    char blind_group_header[] = "GROUP = TIRS_DETECTOR_RESPONSE_BLIND";
    char nonblind_endgroup_header[] = "END_GROUP = TIRS_DETECTOR_RESPONSE";
    char blind_endgroup_header[] = 
        "END_GROUP = TIRS_DETECTOR_RESPONSE_BLIND";
    char parameter_basename[] = "Baseline_Dark_Response";
    char *group_header_selector = NULL;
    char *endgroup_header_selector = NULL;

    int detector_index = 0;                /* Detector loop counter */
    int status = SUCCESS;

    FILE *group_fptr = NULL;

    /* Create the file name (based on the band number and the SCA number */
    status = snprintf(group_filename, sizeof(group_filename),
        "%s_B%02d_SCA%02d.odl", groupfile_basename,
        band_number, sca_number);
    if ((status < 0) || (status >= sizeof(group_filename)))
    {
        IAS_LOG_ERROR("Constructing CPF group file name");
        return ERROR;
    }

    group_fptr = fopen(group_filename, "w");
    if (group_fptr == NULL)
    {
        IAS_LOG_ERROR("Creating new group file %s", group_filename);
        return ERROR;
    }

    /* Write the "GROUP = X" header information string, depending on
       whether it's an imaging band or the blind band */
    if (is_blind_band)
    {
        group_header_selector = blind_group_header;
        endgroup_header_selector = blind_endgroup_header;
    }
    else
    {
        group_header_selector = nonblind_group_header;
        endgroup_header_selector = nonblind_endgroup_header;
    }

    status = fprintf(group_fptr, "%s\n", group_header_selector);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing group start header");
        fclose(group_fptr);
        remove(group_filename);
        return ERROR;
    }

    /* Write the parameter name with the correct band and SCA */
    status = fprintf(group_fptr, " %s_B%02d_SCA%02d = ( ",
        parameter_basename, band_number, sca_number);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing band %d SCA %d adjusted dark response "
            "parameter name to group file", band_number, sca_number);
        fclose(group_fptr);
        remove(group_filename);
        return ERROR;
    }

    /* Write the values to the group file */
    for (detector_index = 0;
         detector_index < number_detectors;
         detector_index++)
    {
        if ((detector_index % 5) == 0)
        {
            status = fprintf(group_fptr, "\n  %11.6lf,",
                adjusted_dark_response[detector_index]);
        }
        else if (detector_index == (number_detectors - 1))
        {
            status = fprintf(group_fptr, "%11.6lf )\n",
                adjusted_dark_response[detector_index]);
        }
        else
        {
            status = fprintf(group_fptr, "%11.6lf,",
                adjusted_dark_response[detector_index]);
        }

        /* Did the print work */
        if (status < 0)
        {
            IAS_LOG_ERROR("Writing adjusted dark response parameters "
                "for band %d, SCA %d, detector %d", band_number,
                sca_number, (detector_index + 1));
            return ERROR;
        }
    }

    /* Write the end group header and end statement */
    status = fprintf(group_fptr, "%s\nEND\n", endgroup_header_selector);
    if (status < 0)
    {
        IAS_LOG_ERROR("Writing end group header");
        fclose(group_fptr);
        return ERROR;
    }

    /* Close the group file and check the final status */
    status = fclose(group_fptr);
    if (status != 0)
    {
        IAS_LOG_ERROR("Closing file %s", group_filename);
    }

    /* Done */
    return status;
}
