/*******************************************************************************
 NAME: bias_model_cal_test_tirs

 PURPOSE: Performs CGREEN unit testing on the bias frame averages

 RETURNS: CGREEN unit status

 NOTES: Unit test based on the verification method specified in the bias
        model calibration ADD.
*******************************************************************************/
#include <math.h>
#include <string.h>

#include "cgreen.h"
#include "ias_const.h"
#include "ias_bpf.h"
#include "ias_bpf_database.h"
#include "ias_db.h"
#include "ias_db_get_connect_info.h"
#include "ias_db_query.h"
#include "ias_rps_frame_averages.h"
#include "ias_logging.h"
#include "ias_radiometry.h"
#include "ias_rps_bias_model_calibration.h"
#include "../ias_rps_db_bias_model_calibration.h"
#include "ias_satellite_attributes.h"
#include "ias_pixel_mask.h"

static char *test_data_path = NULL;

#define DOUBLE_TOLERANCE 1E-6

#define FILE_COLLECTIONS_COUNT 2

#define FILE_DETECTOR_COUNT 640
#define FILE_SCA_COUNT 3
#define FILE_BAND_COUNT 3
#define FILE_FRAME_COUNT 255
#define FILE_TOTAL_DETECTOR_MEANS \
    (FILE_DETECTOR_COUNT * FILE_SCA_COUNT * FILE_BAND_COUNT)
#define FILE_TOTAL_VRP_MEANS\
    (FILE_VRP_COUNT * FILE_SCA_COUNT * FILE_BAND_COUNT)
#define FILE_PIXEL_COUNT (FILE_TOTAL_DETECTOR_MEANS * FILE_FRAME_COUNT)


#define PRE_VALUE 473
#define POST_VALUE 4058

/* List of the band order stored in the data files for testing*/
typedef enum 
{
    TIRS_THERMAL_1_BAND, TIRS_THERMAL_2_BAND
}RAW_FILE_BANDS;

/* Structure to hold the test data from data files */
typedef struct test_data
{
    double means[FILE_COLLECTIONS_COUNT]\
        [FILE_BAND_COUNT][FILE_SCA_COUNT][FILE_DETECTOR_COUNT];
}TEST_DATA;

void free_test_data(TEST_DATA **test_data)
{
    if (test_data == NULL) return;
    if (*test_data == NULL) return;

    IAS_LOG_DEBUG("Releasing test data");

    free(*test_data);
    *test_data = NULL;
}

/* Read in the test data provided by cal/val into a structure */
TEST_DATA* get_test_data()
{
    int band = 0;
    int collection = 0;
    int detector = 0;
    int sca = 0;
    

    TEST_DATA *test_data;    

    IAS_LOG_DEBUG("Loading test data");

    test_data = (TEST_DATA*) calloc(1, sizeof(TEST_DATA));

    if (test_data == NULL)
    {
        IAS_LOG_ERROR("Memory allocation failed");
        return NULL;
    }

    for (collection = 0; collection < FILE_COLLECTIONS_COUNT; collection++)
    {
        for (band = 0; band < FILE_BAND_COUNT; band++)
        {
            for (sca = 0; sca < FILE_SCA_COUNT; sca++)
            {
                for (detector = 0; detector < FILE_DETECTOR_COUNT; detector++)
                {
                    if (collection == 0)
                    {
                        test_data->means[collection][band][sca][detector] = 
                            POST_VALUE;
                    }
                    else
                    {
                        test_data->means[collection][band][sca][detector] = 
                            PRE_VALUE;
                    }
                }
            }
        }
    }

    return test_data;
}

static struct ias_db_connection *db_open()
{
    struct ias_db_connection *db = NULL;
    char *db_name = NULL;
    char *user_name = NULL;
    char *passwd = NULL;
    char *host = NULL;

    int status = SUCCESS;


    /* Initialize the database library */
    status = ias_db_initialize_database_lib();
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error initializing database "
           "library");
        return NULL;
    }

    /* Get the connection information */
    status = ias_db_get_connect_info(&db_name, &user_name, &passwd, &host);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error obtaining DB connection "
            "information");
        ias_db_close_database_lib();
        return NULL;
    }

    /* Try to connect */
    db = ias_db_connect_to_database(db_name, user_name, passwd, host);
    free(db_name);
    free(user_name);
    free(passwd);
    free(host);
    if (db == NULL)
    {
        assert_true_with_message(FALSE, "Error connecting to DB %s as user %s"
           "at host %s", db_name, user_name, host);
        return NULL;
    }
    return db;
}

static void db_close(struct ias_db_connection *db)
{
    ias_db_close_connection(db);
    ias_db_close_database_lib();
}


static int get_workorder_id(
    /*I: database connection*/
    struct ias_db_connection *database,
    /*O: work order ID */
    char work_order_id[])
{
    int existing_records = 0;
    int status = 0;
    int table_size = 0;

    long int order_number = 1;

   /* Obtain the largest IS work order ID currently used
        SELECT COUNT(*), MAX(WORK_ORDER_ID)
        FROM RADIOMETRIC_WORK_ORDER_COMMON
        WHERE WORK_ORDER_ID LIKE 'IS%'; */
    IAS_DB_QUERY_TABLE qtable[] =
    {
        {"COUNT(*)", &existing_records, sizeof(existing_records),
             IAS_QUERY_INT, 1},
        {"max(to_number(SUBSTR(WORK_ORDER_ID, 1, (length(WORK_ORDER_ID)-2))))",
            work_order_id, IAS_WORKORDER_ID_LENGTH, IAS_QUERY_STRING, 1}
    };

    table_size  = sizeof(qtable)/sizeof(IAS_DB_QUERY_TABLE);
    status = ias_db_query_using_table(database,
        "RADIOMETRIC_WORK_ORDER_COMMON", qtable, table_size,
        "WORK_ORDER_ID LIKE '%IS'", 1);

    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Unable to obtain maximum work order ID");
        db_close(database);
        return ERROR;
    }

    IAS_LOG_DEBUG("Obtained MAX ID %s", work_order_id);

    /* convert the work order id to a numerical value to do an increment */
    if (existing_records > 0)
    {
        order_number = atol(work_order_id);
        order_number++;
    }

    status = snprintf(work_order_id, IAS_WORKORDER_ID_LENGTH, "%ldIS",
        order_number);

    if (status < 0 || status > IAS_WORKORDER_ID_LENGTH)
    {
        IAS_LOG_ERROR("Unable to create the work order ID");
        db_close(database);
        return ERROR;
    }

    IAS_LOG_DEBUG("Using work order string %s", work_order_id);

    return SUCCESS;
}

/*
* ROUTINE: insert_test_data
* 
* DESCRIPTION: Inserts the detector mean for all detectors
*
*/
static int insert_test_data(
    /*I: database connection */
    struct ias_db_connection *database,
    /*I: test data to store*/
    const TEST_DATA *test_data,
    /*O: common work order ID of the last workorder*/
    char *common_wo_id)
{
    const int sql_statement_size = 2048; 

    char ingest_processing_level[IAS_RPS_PROCESSING_LEVEL_LENGTH] = "ING1";
    char sql[sql_statement_size]; 
    char work_order_id[IAS_WORKORDER_ID_SIZE];

    int band_number = 0;
    int collection = 0;
    int detector = 0;
    int frame_count = 0;
    int i = 0;
    int pan_flag = 0;
    int sca_index = 0;
    int sca_number = 0;
    int split_line = 0;
    int status = SUCCESS;

    float correlation = 0.0;
    float min = 0.0;
    float mean_squared = 0.0;
    float max = 4096;
    float std_dev = 0.0;
    float skewness = 0.0;
    float kurtosis = 0.0;

    const IAS_BAND_ATTRIBUTES *band_attributes = NULL;
    const IAS_SATELLITE_ATTRIBUTES *sat_attributes = NULL;
    struct ias_db_query *query = NULL; // DB query structure
    IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS *statistics = NULL;
    IAS_RPS_DB_FRAME_AVERAGES *frame_averages = NULL;

    sat_attributes = ias_sat_attr_get_attributes();

    if (common_wo_id == NULL)
    {
        IAS_LOG_ERROR("NULL RWOC ID received");
        return ERROR;
    }

    for (collection = 0; 
        collection < FILE_COLLECTIONS_COUNT; collection++)
    {
        int collection_index = 0;
        int mission_id = 0;
        int table_size = 0;

        /*Handle the +1 condition for the post collection*/
        if (collection >= FILE_COLLECTIONS_COUNT) 
        {
            /* An additional collection needs to be inserted for the post */
            collection_index = FILE_COLLECTIONS_COUNT;
        }
        else
        {
            collection_index = collection;
        }

        if (get_workorder_id(database, work_order_id) != SUCCESS)
        {
            IAS_LOG_ERROR("Unable to obtain workoder ID");
            return ERROR;
        }
    
        /*obtain a common ID to use in the frame statistics table*/
        if (ias_misc_get_char_id(database, "radiometric_work_order_common",
            "radiometric_wo_common_id", work_order_id, common_wo_id) != SUCCESS)
        {
            IAS_LOG_ERROR("Unable to obtain radiometric workoder common ID");
            return ERROR;
        }
    
        /* Build the SQL statement */
        snprintf(sql, sql_statement_size,
            "UPDATE RADIOMETRIC_WORK_ORDER_COMMON "
            "SET DATE_ACQUIRED = DATE_PROCESSED, "
            "COLLECTION_TYPE = 'TIRS_DEEPSPACE', "
            "VALID = 'Y', "
            "LANDSAT_INTERVAL_ID = 'LC8BMCTEST%s' "
            "WHERE RADIOMETRIC_WO_COMMON_ID = %s",
            common_wo_id, common_wo_id);
    
        /* Perform the query */
        query = ias_db_perform_query(database, sql);
        if (!ias_db_query_was_successful(query))
        {
            IAS_LOG_ERROR("Unable to update RWOC ID %s", common_wo_id);
            return ERROR;
        }

        IAS_DB_QUERY_TABLE qtable[] =
        {
            {"max(LMD_MISSION_ID) + 1",
                &mission_id, sizeof(mission_id), IAS_QUERY_INT, 1}
        };
    
        table_size  = sizeof(qtable)/sizeof(IAS_DB_QUERY_TABLE);
        status = ias_db_query_using_table(database,
            "LMD_MISSION", qtable, table_size, "", 1);
    
        if (status != SUCCESS)
        {
            IAS_LOG_ERROR("Unable to obtain maximum work order ID");
            db_close(database);
            return ERROR;
        }

        /* Build the SQL statement */
        status = snprintf(sql, sql_statement_size,
            "INSERT INTO LMD_MISSION (LMD_MISSION_ID, "
            "LANDSAT_CAL_INTERVAL_ID, ORBIT, INTERVAL_FILES, NADIR_OFFNADIR, "
            "WRS_PATH, WRS_STARTING_ROW, WRS_ENDING_ROW, "
            "FULL_PARTIAL_INTERVAL, DATE_ACQUIRED, DATE_ENTERED, DATA_TYPE, "
            "DATA_CATEGORY_ORIGINAL, DATA_CATEGORY, STATION_ID, SENSOR_ID, "
            "MOE_INTERVAL_ID, MISSING_FILES, COLLECTION_TYPE)"
            "VALUES (%d, 'LC8BMCTEST%s', 1, 1, 'NADIR', 0, 0, 0, 'FULL'," 
            "TO_DATE('2011-08-08 08:08:08', 'YYYY-MM-DD HH24:MI:SS'),"
            "TO_DATE('2011-08-08 08:08:08', 'YYYY-MM-DD HH24:MI:SS'),"
            "'TIRS_L0RA', 'NOMINAL', 'NOMINAL', 'EDC', 'TIRS', 'LC8BT%s', 0,"
            "'TIRS_DEEPSPACE')",
            mission_id, common_wo_id, common_wo_id);

        if (status < 0 || status >= sql_statement_size)
        {
            IAS_LOG_ERROR("Unable to create sql statement");
            return ERROR;
        }
        else
        {
            /*Set the return value back to SUCCESS*/
            status = SUCCESS;
        }

        /* Perform the query */
        query = ias_db_perform_query(database, sql);
        if (!ias_db_query_was_successful(query))
        {
            IAS_LOG_ERROR("Unable to perform insert with statement %s",
                sql);
            return ERROR;
        }

        /* Insert histogram statistics for every band */
        for (band_number = 1; band_number <= sat_attributes->total_bands;
            band_number++)
        {
            band_attributes = ias_sat_attr_get_band_attributes(band_number);
     
            if (band_attributes == NULL)
            {
                IAS_LOG_ERROR("Unable to obtain sat attributes for band "
                    "number %d", band_number);
                return ERROR;
            }

            switch (band_attributes->band_type)
            {
                case IAS_THERMAL1_BAND:  /* 10.8um thermal TIRS band */
                    break;
                case IAS_THERMAL2_BAND: /* 12um thermal TIRS band */
                    break;
                default:
                    /*TIRS blind & OLI bands are not applicable to this test */
                    continue;
            }
    
            for (sca_number = 1; sca_number <= band_attributes->scas; 
                sca_number++)
            {
                sca_index = sca_number - 1;

                statistics = ias_rps_db_allocate_histogram_results(
                    band_attributes->detectors_per_sca, pan_flag);
                if (statistics == NULL) 
                {
                    IAS_LOG_ERROR("Unable to allocate statistics");
                    return ERROR;
                }
    
                statistics->dbase = database;
                strncpy(statistics->rad_common_id, common_wo_id, 
                    IAS_CHAR_ID_SIZE);

                statistics->band_number = band_number;
                statistics->sca = sca_number;
                statistics->num_detectors = band_attributes->detectors_per_sca;
                statistics->is_odd_even_data_split = pan_flag;
                statistics->processing_level = strdup(ingest_processing_level);
                if (statistics->processing_level == NULL) 
                {
                    return ERROR;
                }

                /* assign values for the statistics for later comparison */
                for (split_line = 0; split_line <= pan_flag; split_line++)
                {
 
                    for (detector = 0;
                        detector < band_attributes->detectors_per_sca; 
                        detector++)
                    {
                        /* The detectors are stored in the database as 1 based 
                            IDs*/
                        statistics->detector[split_line][detector] = detector+1;
                        statistics->num_values[split_line][detector] = 
                            band_attributes->detectors_per_sca;
                        statistics->min[split_line][detector] = min;
                        statistics->max[split_line][detector] = max;
                        statistics->stdev[split_line][detector] = std_dev;
                        statistics->skewness[split_line][detector] = skewness;
                        statistics->kurtosis[split_line][detector] = kurtosis;
                        statistics->correlation[split_line][detector] = 
                            correlation;
                        statistics->mean_sq_resp[split_line][detector]=
                            mean_squared;

                        statistics->mean[split_line][detector] = 
                            test_data->means[collection_index][0][sca_index][detector];
                    }/* End detector loop */
                }/* End split line loop */
    
                if (ias_rps_store_histogram_stats_in_database(statistics) !=
                    SUCCESS)
                {
                    IAS_LOG_ERROR("Unable to store statistics for band %d, "
                        "sca %d", band_number, sca_number);
                    return ERROR;
                }

                /* statistics are freed by the store routine */
                statistics = NULL;

                frame_count = FILE_FRAME_COUNT;
  
                frame_averages =
                    ias_rps_db_allocate_frame_averages(frame_count);
                frame_averages->dbase = database;
                strncpy(frame_averages->rad_common_id, common_wo_id, 
                    IAS_CHAR_ID_LENGTH);
                frame_averages->band_number = band_number;
                frame_averages->sca = sca_number;
                frame_averages->num_lines = frame_count;

                for (i = 0; i < frame_count; i++)
                {
                    frame_averages->line_number[i] = i + 1;
                    frame_averages->line_average[i] = 1;
                } /* End of frame loop */
    
                if (SUCCESS != ias_rps_db_store_frame_averages(frame_averages))
                {
                    IAS_LOG_ERROR("Unable to store frame averages");
                    ias_rps_db_free_frame_averages(&frame_averages);
                    return ERROR;
                }

                IAS_LOG_DEBUG("Frame averages stored for band %d", band_number);
    
                ias_rps_db_free_frame_averages(&frame_averages);
            }/*End SCA Loop*/
        }/*End band Loop*/
    }/*End of collection loop */

    IAS_LOG_DEBUG("Test data loaded into the database");

    return status;
}/*End  insert_test_data*/

/*----------------------------------------------------------------------------
NAME: bias_model_verification

PURPOSE: Test the ability to calculate the bias model parameters from 
    histogram statistics from the database.
-----------------------------------------------------------------------------*/
Ensure bias_model_verification(void)
{
    const int previous_collections_n0 = 1;
    const struct IAS_BPF_BIAS_MODEL *bias_model = NULL;
    const IAS_SATELLITE_ATTRIBUTES *sat_attributes = NULL;

    char common_wo_id[IAS_CHAR_ID_SIZE];
    char mission_id[IAS_L0R_INTERVAL_ID_LENGTH];
    char cpf_name[] = "ias_lib/rps/bias_model_calibration/L8CPFTEST.01";
    char full_path_file[PATH_MAX];

    int band_number = 0;
    int collections_used = 0;
    int sca_index = 0;
    int status = SUCCESS;

    struct ias_db_connection *db_connection = db_open();

    IAS_BPF *bpf = NULL; 
    IAS_CPF *cpf = NULL;
    TEST_DATA *test_data = NULL;

    sat_attributes = ias_sat_attr_get_attributes();

    if (sat_attributes == NULL)
    {
        IAS_LOG_ERROR("Unable to obtain satellite attributes");
        assert_true(0);
        return;
    }

    if (db_connection == NULL)
    {
        IAS_LOG_ERROR("Unable to establish database connection");
        assert_true(0);
        return;
    }

    /* Build the full path for the CPF and read the file using the lib */
    sprintf(full_path_file, "%s/%s", test_data_path, cpf_name);
    cpf = ias_cpf_read(full_path_file);

    if (cpf == NULL)
    {
        IAS_LOG_ERROR("Unable to read CPF");
        assert_true(0);
        return;
    }

    test_data = get_test_data();
    if (test_data == NULL)
    {
        ias_cpf_free(cpf);
        cpf = NULL;
        IAS_LOG_ERROR("Unable to load test data");
        assert_true(0);
        return;
    }

    memset(common_wo_id, 0, IAS_CHAR_ID_SIZE);
 
    /*Insert the values for the workorder*/
    if (insert_test_data(db_connection, test_data, common_wo_id) != SUCCESS)
    {
        ias_cpf_free(cpf);
        cpf = NULL;
        free_test_data(&test_data);
        IAS_LOG_ERROR("Unable to insert RWOC %s", common_wo_id);
        assert_true(0);
        return;
    }

    bpf = ias_bpf_allocate();
    if (bpf == NULL)
    {
        ias_cpf_free(cpf);
        cpf = NULL;
        free_test_data(&test_data);
        IAS_LOG_ERROR("Unable to allocate the bpf");
        assert_true(0);
        return;
    }

    /* examine the bias parameters */
    bias_model = ias_bpf_get_bias_model(bpf);

    sprintf(mission_id, "LC8BMCTEST%s", common_wo_id);

    /* With the values inserted calculate the bias parameters */
    if (ias_rps_bias_model_calibration_tirs(
        cpf, db_connection, mission_id, common_wo_id, previous_collections_n0,
        0, bpf, &collections_used) != SUCCESS)
    {
        IAS_LOG_ERROR("Unable to generate bias parameters");
        assert_true(0); 
        return;
    }

    /* Check the bias parameters calculated to those from the test data */
    for (band_number = 1; band_number <= sat_attributes->total_bands;
        band_number++)
    {
        const IAS_BAND_ATTRIBUTES *band_attributes =
            ias_sat_attr_get_band_attributes(band_number);

        double *pre_acquisition_shutter_average = NULL;
        double *post_acquisition_shutter_average = NULL;
        double *mean_shutter_average = NULL;
        int detector = 0;
        int even_odd = 0;

        /* Bias parameters are calculated for the Normal TIRS bands, skip
        the others */
        if ((band_attributes->band_classification & IAS_NORMAL_BAND) == 0 ||
            (band_attributes->sensor_id != IAS_TIRS))
        {
            IAS_LOG_DEBUG("Skipping band number %d", 
                band_attributes->band_number);
            continue;
        }

        if (status == SUCCESS)
        {
            int check_failed = 0;

            pre_acquisition_shutter_average = 
                calloc(band_attributes->detectors_per_sca, sizeof(double));
            post_acquisition_shutter_average = 
                calloc(band_attributes->detectors_per_sca, sizeof(double));
            mean_shutter_average = 
                calloc(band_attributes->detectors_per_sca, sizeof(double));

            if (pre_acquisition_shutter_average == NULL || 
                post_acquisition_shutter_average == NULL ||
                mean_shutter_average == NULL)
            {
                IAS_LOG_ERROR("Unable to perform allocation");
                status = ERROR;
                break;
            }
            
            for (sca_index = 0; sca_index < band_attributes->scas; sca_index++)
            {
                int sca_number = sca_index + 1;

                IAS_LOG_DEBUG("Verifying band number %d, sca index %d of %d", 
                    band_number, sca_index, band_attributes->scas);
    
                if (bias_model == NULL)
                {
                    assert_true(0);
                    break;
                }
   
                    switch (band_attributes->band_type)
                    {
                        case IAS_THERMAL1_BAND:  /* 10.8um thermal TIRS band */
                            break;
                        case IAS_THERMAL2_BAND: /* 12um thermal TIRS band */
                            break;
                        default:
                            IAS_LOG_ERROR("Unexpected band type %d",
                                band_attributes->band_type);
                            status = ERROR;
                            continue;
                    }

                    IAS_LOG_DEBUG("Obtaining bias model parameters for band "
                        "number %d of %d, sca index %d of %d", 
                        band_number, sat_attributes->total_bands, sca_index, 
                        band_attributes->scas);

                    if (SUCCESS != ias_bpf_get_model_parameters(
                        bias_model, band_attributes->spectral_type,
                        band_attributes->band_number, 
                        sca_number, 
                        band_attributes->detectors_per_sca, 
                        even_odd,              
                        pre_acquisition_shutter_average,
                        post_acquisition_shutter_average,
                        NULL, NULL, NULL))
                    {
                        assert_true(0);
                        status = ERROR;
                    }
       
                    /* verify the values from the bias model are correct */
                    for (detector = 0; 
                        detector < band_attributes->detectors_per_sca; 
                        detector++)
                    {

                        double difference = 
                            pre_acquisition_shutter_average[detector] - 
                            test_data->means[0][0]\
                                [sca_index][detector];

                        if (DOUBLE_TOLERANCE < fabs(difference))
                        {
                            IAS_LOG_ERROR("band number %d, sca number %d: "
                                "%lf != %lf", 
                                band_attributes->band_number,
                                sca_number,
                                pre_acquisition_shutter_average[detector],
                                test_data->means[0][0]\
                                    [sca_index][detector]);
                            check_failed++;
                        }

                        difference = 
                            post_acquisition_shutter_average[detector] - 
                            test_data->means[1][0]\
                                [sca_index][detector];

                        if (DOUBLE_TOLERANCE < fabs(difference))
                        {
                            IAS_LOG_ERROR("band number %d, sca number %d: "
                                "%lf != %lf",
                                band_attributes->band_number,
                                sca_number,
                                post_acquisition_shutter_average[detector],
                                test_data->means[1][0]\
                                    [sca_index][detector]);
                            check_failed++;
                        }
                    }/* End of detector loop */
            }/* End SCA loop */

            if (check_failed != 0)
            {
                IAS_LOG_ERROR("Band number %d failed with %d checks not "
                    "successfull", band_number, check_failed);
                assert_true(0);
            }
            else
            {
                IAS_LOG_DEBUG("Band number %d success", band_number);
                assert_true(1);
            }

            free(pre_acquisition_shutter_average);
            free(post_acquisition_shutter_average);
            free(mean_shutter_average);
    
            pre_acquisition_shutter_average = NULL;
            post_acquisition_shutter_average = NULL;
            mean_shutter_average = NULL;
        } /* End of if condition */
    } /* End of band for loop */

    ias_bpf_free(bpf);
    bpf = NULL;
  
    ias_cpf_free(cpf);
    cpf = NULL;

    //free_test_data(&test_data);
    db_close(db_connection);     

    IAS_LOG_DEBUG("Bias Model Cal test complete");
}

int main(int argc, char *argv[])
{
    /* Construct the test suite */
    TestSuite *suite = create_test_suite();

    /* Set the log level for regression test execution of unit tests */
    ias_log_set_output_level(IAS_LOG_LEVEL_WARN);

    /* Get the test data directory */
    test_data_path = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (test_data_path == NULL)
    {
        printf("Cannot locate IAS test data directory\n");
        return 1;
    }

    if (ias_sat_attr_initialize(IAS_L8) != SUCCESS)
    {
        IAS_LOG_ERROR("Unable to load the satellite attributes used for "
            "processing.");
        return ERROR;
    }
    
    /* Use the env variable to determine if the database related tests should 
    be run */
    if (getenv("ENABLE_DB_TESTS") == NULL)
    {
        printf("DB Tests will not run, setenv ENABLE_DB_TESTS to enable....\n");
        return SUCCESS;
    }

    add_test(suite, bias_model_verification);

    return run_test_suite(suite, create_text_reporter());
}

