/*******************************************************************************
 NAME: bias_model_cal_test

 PURPOSE: Performs CGREEN unit regression testing on the bias frame averages

 RETURNS: CGREEN unit status

*******************************************************************************/
#include <math.h>
#include <string.h>

#include "cgreen.h"
#include "ias_const.h"
#include "ias_bpf.h"
#include "ias_bpf_database.h"
#include "ias_db.h"
#include "ias_db_get_connect_info.h"
#include "ias_db_query.h"
#include "ias_rps_frame_averages.h"
#include "ias_logging.h"
#include "ias_radiometry.h"
#include "ias_rps_bias_model_calibration.h"
#include "../ias_rps_db_bias_model_calibration.h"
#include "ias_satellite_attributes.h"
#include "ias_pixel_mask.h"

static char *test_data_path = NULL;

#define DOUBLE_TOLERANCE 1E-6

/* For normal unit testing use a limited number of collections
as the full 40 previous collections takes too long to process*/
#define FILE_COLLECTIONS_COUNT 3

#define FILE_NORMAL_BAND_COUNT 8
#define FILE_DETECTOR_COUNT 494
#define FILE_PAN_DETECTOR_COUNT 988
#define FILE_VRP_DETECTOR_COUNT 12
#define FILE_VRP_PAN_DETECTOR_COUNT 24
#define FILE_SCA_COUNT 14
#define FILE_FRAME_COUNT 500

#define LINESETS 2

/* Structure to hold the test data from data files */
typedef struct test_data
{
    float a0[FILE_NORMAL_BAND_COUNT][FILE_SCA_COUNT];
    float a0_pan[FILE_SCA_COUNT][LINESETS];
    float c1[FILE_NORMAL_BAND_COUNT][FILE_SCA_COUNT][FILE_DETECTOR_COUNT];
    float c1_pan[LINESETS][FILE_SCA_COUNT][FILE_PAN_DETECTOR_COUNT];
    float a1[FILE_NORMAL_BAND_COUNT][FILE_SCA_COUNT][FILE_DETECTOR_COUNT];
    float a1_pan[LINESETS][FILE_SCA_COUNT][FILE_PAN_DETECTOR_COUNT];

    float r0[FILE_NORMAL_BAND_COUNT][FILE_SCA_COUNT]; 
    float r0_pan[FILE_SCA_COUNT][LINESETS];
    float r1[FILE_NORMAL_BAND_COUNT][FILE_SCA_COUNT][FILE_DETECTOR_COUNT];
    float r1_pan[LINESETS][FILE_SCA_COUNT][FILE_PAN_DETECTOR_COUNT];
 
    float Sa[FILE_NORMAL_BAND_COUNT][FILE_SCA_COUNT][FILE_DETECTOR_COUNT];
    float Sb[FILE_NORMAL_BAND_COUNT][FILE_SCA_COUNT][FILE_DETECTOR_COUNT];

    float Sa_pan[LINESETS][FILE_SCA_COUNT][FILE_PAN_DETECTOR_COUNT];
    float Sb_pan[LINESETS][FILE_SCA_COUNT][FILE_PAN_DETECTOR_COUNT];

    float means[FILE_COLLECTIONS_COUNT][FILE_NORMAL_BAND_COUNT][FILE_SCA_COUNT]
        [FILE_DETECTOR_COUNT];
    float means_pan[FILE_COLLECTIONS_COUNT][FILE_SCA_COUNT]
        [FILE_PAN_DETECTOR_COUNT][LINESETS];
    float means_vrp[FILE_COLLECTIONS_COUNT][FILE_NORMAL_BAND_COUNT]
        [FILE_SCA_COUNT][FILE_VRP_DETECTOR_COUNT];
    float means_vrp_pan[FILE_COLLECTIONS_COUNT][FILE_SCA_COUNT]
        [FILE_VRP_PAN_DETECTOR_COUNT][LINESETS];

    double image_frame_average[FILE_COLLECTIONS_COUNT][FILE_NORMAL_BAND_COUNT]
        [FILE_SCA_COUNT][FILE_FRAME_COUNT];
    double image_frame_average_pan[FILE_COLLECTIONS_COUNT][FILE_SCA_COUNT]
        [FILE_FRAME_COUNT * LINESETS];
    double vrp_frame_average[FILE_COLLECTIONS_COUNT][FILE_NORMAL_BAND_COUNT]
        [FILE_SCA_COUNT][FILE_FRAME_COUNT];
    double vrp_frame_average_pan[FILE_COLLECTIONS_COUNT][FILE_SCA_COUNT]
        [FILE_FRAME_COUNT * LINESETS];
}TEST_DATA;

void free_test_data(TEST_DATA **test_data)
{
    if (test_data == NULL) return;
    if (*test_data == NULL) return;

    IAS_LOG_DEBUG("Releasing test data");

    free(*test_data);
    *test_data = NULL;
}

int populate_buffer
(
    const char *full_file_name, /*I: full path of the file to read*/
    const size_t type_size, /*I: size of an element being read*/
    const size_t num_elements, /*I: number of elements to read*/
    void *buffer /*O: buffer to populate*/
)
{
    FILE *fp;
    int status = SUCCESS;

    fp = fopen(full_file_name, "r");
    if (fp != NULL)
    {
        size_t items_read = fread(buffer, type_size, num_elements, fp);
        fclose(fp);
        if (items_read != num_elements)
        {
            long unsigned my_sizes[2];
            my_sizes[0] = (long unsigned) items_read;
            my_sizes[1] = (long unsigned) num_elements;
            IAS_LOG_ERROR("Unable to read full dataset. "
                "%lu read of the %lu requested", my_sizes[0], my_sizes[1]);
            status = ERROR;
        }
    }
    else
    {
        IAS_LOG_ERROR("Unable to open %s", full_file_name);
        status = ERROR;
    }

    return status;
}

/* Read in the test data provided by cal/val into a structure */
TEST_DATA* get_test_data()
{
    char file_name_a0[] = "ias_lib/rps/bias_model_calibration/a0.dat";
    char file_name_a0_pan[] = "ias_lib/rps/bias_model_calibration/a0_pan.dat";
    char file_name_c1[] = "ias_lib/rps/bias_model_calibration/c1.dat";
    char file_name_c1_pan[] = "ias_lib/rps/bias_model_calibration/c1_pan.dat";
    char file_name_a1[] = "ias_lib/rps/bias_model_calibration/a1.dat";
    char file_name_a1_pan[] = "ias_lib/rps/bias_model_calibration/a1_pan.dat";
    char file_name_r0[] = "ias_lib/rps/bias_model_calibration/r0.dat";
    char file_name_r0_pan[] = "ias_lib/rps/bias_model_calibration/r0_pan.dat";
    char file_name_r1[] = "ias_lib/rps/bias_model_calibration/r1.dat";
    char file_name_r1_pan[] = "ias_lib/rps/bias_model_calibration/r1_pan.dat";

    char file_name_Sa[] = "/ias_lib/rps/bias_model_calibration/Sa.dat";
    char file_name_Sa_pan[] = "/ias_lib/rps/bias_model_calibration/Sa_pan.dat";
    char file_name_Sb[] = "/ias_lib/rps/bias_model_calibration/Sb.dat";
    char file_name_Sb_pan[] = "/ias_lib/rps/bias_model_calibration/Sb_pan.dat";

    char file_name_detector_means[] = 
        "ias_lib/rps/bias_model_calibration/detector_means.dat";
    char file_name_detector_means_pan[] =
        "/ias_lib/rps/bias_model_calibration/detector_means_pan.dat";
    char file_name_detector_means_vrp[] = 
        "ias_lib/rps/bias_model_calibration/detector_means_vrp.dat";
    char file_name_detector_means_vrp_pan[] =
        "ias_lib/rps/bias_model_calibration/detector_means_vrp_pan.dat";
    char file_name_frame_averages[] =
        "ias_lib/rps/bias_model_calibration/frame_averages.dat";
    char file_name_frame_averages_pan[] =
        "ias_lib/rps/bias_model_calibration/frame_averages_pan.dat";
    char file_name_frame_averages_vrp[] =
        "ias_lib/rps/bias_model_calibration/frame_averages_vrp.dat";
    char file_name_frame_averages_vrp_pan[] =
        "ias_lib/rps/bias_model_calibration/frame_averages_vrp_pan.dat";
    char full_path_file[PATH_MAX];

    TEST_DATA *test_data;    

    IAS_LOG_DEBUG("Loading test data from file.");

    test_data = (TEST_DATA*) calloc(1, sizeof(TEST_DATA));

    if (test_data == NULL)
    {
        IAS_LOG_ERROR("Memory allocation failed");
        return NULL;
    }

    sprintf(full_path_file, "%s/%s", test_data_path, file_name_a0);
    if (SUCCESS != populate_buffer(full_path_file, sizeof(float), 
       (FILE_NORMAL_BAND_COUNT * FILE_SCA_COUNT), test_data->a0))
    {
        IAS_LOG_ERROR("Unable to populate data from %s", full_path_file);
        free(test_data);
        return NULL;
    }

    sprintf(full_path_file, "%s/%s", test_data_path, file_name_a0_pan);
    if (SUCCESS != populate_buffer(full_path_file, sizeof(float),
       (LINESETS * FILE_SCA_COUNT), test_data->a0_pan))
    {
        IAS_LOG_ERROR("Unable to populate data from %s", full_path_file);
        free(test_data);
        return NULL;
    }

    sprintf(full_path_file, "%s/%s", test_data_path, file_name_r0);
    if (SUCCESS != populate_buffer(full_path_file, sizeof(float), 
        (FILE_NORMAL_BAND_COUNT * FILE_SCA_COUNT), test_data->r0))
    {
        IAS_LOG_ERROR("Unable to populate data from %s", full_path_file);
        free(test_data);
        return NULL; 
    } 

    sprintf(full_path_file, "%s/%s", test_data_path, file_name_r0_pan);
    if (SUCCESS != populate_buffer(full_path_file, sizeof(float), 
        (FILE_SCA_COUNT * LINESETS), test_data->r0_pan))
    {
        IAS_LOG_ERROR("Unable to populate data from %s", full_path_file);
        free(test_data);
        return NULL; 
    } 

    sprintf(full_path_file, "%s/%s", test_data_path, file_name_c1);
    if (SUCCESS != populate_buffer(full_path_file, sizeof(float), 
        FILE_DETECTOR_COUNT * FILE_SCA_COUNT * FILE_NORMAL_BAND_COUNT, 
        test_data->c1))
    {
        IAS_LOG_ERROR("Unable to populate data from %s", full_path_file);
        free(test_data);
        return NULL; 
    } 

    sprintf(full_path_file, "%s/%s", test_data_path, file_name_c1_pan);
    if (SUCCESS != populate_buffer(full_path_file, sizeof(float),
        FILE_PAN_DETECTOR_COUNT * FILE_SCA_COUNT * LINESETS,
        test_data->c1_pan))
    {
        IAS_LOG_ERROR("Unable to populate data from %s", full_path_file);
        free(test_data);
        return NULL;
    }

    sprintf(full_path_file, "%s/%s", test_data_path, file_name_a1);
    if (SUCCESS != populate_buffer(full_path_file, sizeof(float), 
        FILE_DETECTOR_COUNT * FILE_SCA_COUNT * FILE_NORMAL_BAND_COUNT, 
        test_data->a1))
    {
        IAS_LOG_ERROR("Unable to populate data from %s", full_path_file);
        free(test_data);
        return NULL; 
    } 

    sprintf(full_path_file, "%s/%s", test_data_path, file_name_a1_pan);
    if (SUCCESS != populate_buffer(full_path_file, sizeof(float),
        LINESETS * FILE_PAN_DETECTOR_COUNT * FILE_SCA_COUNT,
        test_data->a1_pan))
    {
        IAS_LOG_ERROR("Unable to populate data from %s", full_path_file);
        free(test_data);
        return NULL;
    }

    sprintf(full_path_file, "%s/%s", test_data_path, file_name_r1);
    if (SUCCESS != populate_buffer(full_path_file, sizeof(float),
        FILE_DETECTOR_COUNT * FILE_SCA_COUNT * FILE_NORMAL_BAND_COUNT,
        test_data->r1))
    {
        IAS_LOG_ERROR("Unable to populate data from %s", full_path_file);
        free(test_data);
        return NULL;
    }

    sprintf(full_path_file, "%s/%s", test_data_path, file_name_r1_pan);
    if (SUCCESS != populate_buffer(full_path_file, sizeof(float),
        FILE_PAN_DETECTOR_COUNT * FILE_SCA_COUNT * LINESETS, test_data->r1_pan))
    {
        IAS_LOG_ERROR("Unable to populate data from %s", full_path_file);
        free(test_data);
        return NULL;
    }

    sprintf(full_path_file, "%s/%s", test_data_path, file_name_Sa);
    if (SUCCESS != populate_buffer(full_path_file, sizeof(float),
        FILE_DETECTOR_COUNT * FILE_SCA_COUNT * FILE_NORMAL_BAND_COUNT, 
        test_data->Sa))
    {
        IAS_LOG_ERROR("Unable to populate data from %s", full_path_file);
        free(test_data);
        return NULL;
    }

    sprintf(full_path_file, "%s/%s", test_data_path, file_name_Sa_pan);
    if (SUCCESS != populate_buffer(full_path_file, sizeof(float),
        FILE_PAN_DETECTOR_COUNT * FILE_SCA_COUNT * LINESETS, 
        test_data->Sa_pan))
    {
        IAS_LOG_ERROR("Unable to populate data from %s", full_path_file);
        free(test_data);
        return NULL;
    }

    sprintf(full_path_file, "%s/%s", test_data_path, file_name_Sb);
    if (SUCCESS != populate_buffer(full_path_file, sizeof(float),
        FILE_DETECTOR_COUNT * FILE_SCA_COUNT * FILE_NORMAL_BAND_COUNT, 
        test_data->Sb))
    {
        IAS_LOG_ERROR("Unable to populate data from %s", full_path_file);
        free(test_data);
        return NULL;
    }

    sprintf(full_path_file, "%s/%s", test_data_path, file_name_Sb_pan);
    if (SUCCESS != populate_buffer(full_path_file, sizeof(float),
        FILE_PAN_DETECTOR_COUNT * FILE_SCA_COUNT * LINESETS,              
        test_data->Sb_pan))
    {
        IAS_LOG_ERROR("Unable to populate data from %s", full_path_file);
        free(test_data);
        return NULL;
    }

    sprintf(full_path_file, "%s/%s", test_data_path, file_name_detector_means);
    if (SUCCESS != populate_buffer(full_path_file, sizeof(float),
        FILE_COLLECTIONS_COUNT * FILE_NORMAL_BAND_COUNT * FILE_SCA_COUNT *
             FILE_DETECTOR_COUNT, test_data->means))
    {
        IAS_LOG_ERROR("Unable to populate data from %s", full_path_file);
        free(test_data);
        return NULL;
    }

    sprintf(full_path_file, "%s/%s", test_data_path, 
        file_name_detector_means_pan);
    if (SUCCESS != populate_buffer(full_path_file, sizeof(float),
        FILE_COLLECTIONS_COUNT * FILE_SCA_COUNT * LINESETS *
             FILE_PAN_DETECTOR_COUNT, test_data->means_pan))
    {
        IAS_LOG_ERROR("Unable to populate data from %s", full_path_file);
        free(test_data);
        return NULL;
    }

    sprintf(full_path_file, "%s/%s", test_data_path, 
        file_name_detector_means_vrp);
    if (SUCCESS != populate_buffer(full_path_file, sizeof(float),
        (FILE_COLLECTIONS_COUNT * FILE_VRP_DETECTOR_COUNT * FILE_SCA_COUNT * 
            FILE_NORMAL_BAND_COUNT),
        test_data->means_vrp))
    {
        IAS_LOG_ERROR("Unable to populate data from %s", full_path_file);
        free(test_data);
        return NULL;
    }

    sprintf(full_path_file, "%s/%s", test_data_path, 
        file_name_detector_means_vrp_pan);
    if (SUCCESS != populate_buffer(full_path_file, sizeof(float),
        (FILE_COLLECTIONS_COUNT * LINESETS * FILE_VRP_PAN_DETECTOR_COUNT 
            * FILE_SCA_COUNT), test_data->means_vrp_pan))
    {
        IAS_LOG_ERROR("Unable to populate data from %s", full_path_file);
        free(test_data);
        return NULL;
    }

    sprintf(full_path_file, "%s/%s", test_data_path, file_name_frame_averages);
    if (SUCCESS != populate_buffer(full_path_file, sizeof(double),
        (FILE_COLLECTIONS_COUNT * FILE_NORMAL_BAND_COUNT * FILE_SCA_COUNT 
            * FILE_FRAME_COUNT), 
        test_data->image_frame_average))
    {
        IAS_LOG_ERROR("Unable to populate data from %s", full_path_file);
        free(test_data);
        return NULL;
    }

    sprintf(full_path_file, "%s/%s", test_data_path, 
        file_name_frame_averages_pan);
    if (SUCCESS != populate_buffer(full_path_file, sizeof(double),
        (FILE_COLLECTIONS_COUNT * FILE_SCA_COUNT * LINESETS * FILE_FRAME_COUNT),
        test_data->image_frame_average_pan))
    {
        IAS_LOG_ERROR("Unable to populate data from %s", full_path_file);
        free(test_data);
        return NULL;
    }

    sprintf(full_path_file, "%s/%s", test_data_path,
        file_name_frame_averages_vrp);
    if (SUCCESS != populate_buffer(full_path_file, sizeof(double),
        (FILE_COLLECTIONS_COUNT * FILE_NORMAL_BAND_COUNT * FILE_SCA_COUNT 
            * FILE_FRAME_COUNT),
        test_data->vrp_frame_average))
    {
        IAS_LOG_ERROR("Unable to populate data from %s", full_path_file);
        free(test_data);
        return NULL;
    }

    sprintf(full_path_file, "%s/%s", test_data_path,
        file_name_frame_averages_vrp_pan);
    if (SUCCESS != populate_buffer(full_path_file, sizeof(double),
        (FILE_COLLECTIONS_COUNT * FILE_SCA_COUNT * LINESETS * FILE_FRAME_COUNT),
        test_data->vrp_frame_average_pan))
    {
        IAS_LOG_ERROR("Unable to populate data from %s", full_path_file);
        free(test_data);
        return NULL;
    }

    IAS_LOG_DEBUG("Test data read");

    return test_data;
}

static struct ias_db_connection *db_open()
{
    struct ias_db_connection *db = NULL;
    char *db_name = NULL;
    char *user_name = NULL;
    char *passwd = NULL;
    char *host = NULL;

    int status = SUCCESS;


    /* Initialize the database library */
    status = ias_db_initialize_database_lib();
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error initializing database "
           "library");
        return NULL;
    }

    /* Get the connection information */
    status = ias_db_get_connect_info(&db_name, &user_name, &passwd, &host);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error obtaining DB connection "
            "information");
        ias_db_close_database_lib();
        return NULL;
    }

    /* Try to connect */
    db = ias_db_connect_to_database(db_name, user_name, passwd, host);
    free(db_name);
    free(user_name);
    free(passwd);
    free(host);
    if (db == NULL)
    {
        assert_true_with_message(FALSE, "Error connecting to DB %s as user %s"
           "at host %s", db_name, user_name, host);
        return NULL;
    }
    return db;
}

static void db_close(struct ias_db_connection *db)
{
    ias_db_close_connection(db);
    ias_db_close_database_lib();
}


static int get_workorder_id(
    /*I: database connection*/
    struct ias_db_connection *database,
    /*O: work order ID */
    char work_order_id[])
{
    const int sql_statement_size = 2048;

    char sql[sql_statement_size];

    int status = 0;
    int order_number = 1;

    struct ias_db_query *query = NULL; // DB query structure

    /* Build the SQL statement to get the next sequence to create the work
       order ID.  This is a bit of a hack to use a sequence generator that
       is available to both IAS and Ingest, but it works well enough. */
    snprintf(sql, sql_statement_size,
        "SELECT DBSEQ_RADIOMETRIC_WO_COMMON_PK.nextval FROM dual");

    /* Perform the query */
    query = ias_db_perform_query(database, sql);
    if (!ias_db_query_was_successful(query))
    {
        IAS_LOG_ERROR("Query failed: %s", sql);
        return ERROR;
    }

    if (!ias_db_query_next(query))
    {
        IAS_LOG_ERROR("Failed to obtain job ID for IS work order.");
         db_close(database);
        return ERROR;
    }

    if (ias_db_query_int_value(query, &order_number, 0) != SUCCESS)
    {
        IAS_LOG_ERROR("Failed to obtain ID for lmd_mission");
        db_close(database);
        return ERROR;
    }

    IAS_LOG_DEBUG("Obtained MAX ID %d", order_number);

    /* Make a fake work order ID in the ingest format, but with "BT" for
       bias test */
    status = snprintf(work_order_id, IAS_WORKORDER_ID_LENGTH, "%dBT",
        order_number);

    if (status < 0 || status > IAS_WORKORDER_ID_LENGTH)
    {
        IAS_LOG_ERROR("Unable to create the work order ID");
        db_close(database);
        return ERROR;
    }

    IAS_LOG_DEBUG("Using work order string %s", work_order_id);

    return SUCCESS;
}

/*
* ROUTINE: insert_test_data
* 
* DESCRIPTION: Inserts the detector mean for all detectors
*
*/
static int insert_test_data(
    /*I: database connection */
    struct ias_db_connection *database,
    /*I: test data to store*/
    const TEST_DATA *test_data,
    /*O: common work order ID of the last workorder*/
    char *common_wo_id)
{
    const int sql_statement_size = 2048; 

    char ingest_processing_level[IAS_RPS_PROCESSING_LEVEL_LENGTH] = "ING1";
    char sql[sql_statement_size]; 
    char work_order_id[IAS_WORKORDER_ID_SIZE];

    int collection = 0;
    int detector = 0;
    int frame_count = 0;
    int split_line = 0;
    int status = SUCCESS;

    float correlation = 0.0;
    float min = 0.0;
    float mean_squared = 0.0;
    float max = 4096;
    float std_dev = 0.0;
    float skewness = 0.0;
    float kurtosis = 0.0;

    const IAS_BAND_ATTRIBUTES *band_attributes = NULL;
    const IAS_SATELLITE_ATTRIBUTES *sat_attributes = NULL;
    struct ias_db_query *query = NULL; // DB query structure
    IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS *statistics = NULL;
    IAS_RPS_DB_FRAME_AVERAGES *frame_averages = NULL;

    sat_attributes = ias_sat_attr_get_attributes();

    if (common_wo_id == NULL)
    {
        IAS_LOG_ERROR("NULL RWOC ID received");
        return ERROR;
    }

    /*The FILE_COLLECTIONS_COUNT is the previous collections*/
    for (collection = 0; collection < FILE_COLLECTIONS_COUNT; collection++)
    {
        int band_index = 0;
        int mission_id = 0;
        int sca_index = 0;

        if (get_workorder_id(database, work_order_id) != SUCCESS)
        {
            IAS_LOG_ERROR("Unable to obtain workoder ID");
            return ERROR;
        }
    
        /*obtain a common ID to use in the frame statistics table*/
        if (ias_misc_get_char_id(database, "radiometric_work_order_common",
            "radiometric_wo_common_id", work_order_id, common_wo_id) != SUCCESS)
        {
            IAS_LOG_ERROR("Unable to obtain radiometric workoder common ID");
            return ERROR;
        }

        // Build the SQL statement
        snprintf(sql, sql_statement_size,
        "UPDATE RADIOMETRIC_WORK_ORDER_COMMON "
        "SET DATE_ACQUIRED = SYSDATE, "
        "  SENSOR_ID = 'OLI', "
        "  SOFTWARE_VERSION_IS = '9.9.9', COLLECTION_TYPE = 'OLI_SHUTTER', "
        "  INTEGRATION_TIME_OLI_PAN = .660, "
        "  INTEGRATION_TIME_OLI_MS = 3.600, "
        "  WORK_ORDER_ID = '%s', "
        "  VALID = 'Y', "
        "  DATE_PROCESSED = SYSDATE "
        "WHERE RADIOMETRIC_WO_COMMON_ID = %s\n ",
        work_order_id, common_wo_id);

        /* Perform the query */
        query = ias_db_perform_query(database, sql);
        if (!ias_db_query_was_successful(query))
        {
            IAS_LOG_ERROR("Failed to insert RWOC ID %s: %s", 
                common_wo_id, ias_db_connect_last_error(database));
            printf("%s\n", sql);
            return ERROR;
        }

        snprintf(sql, sql_statement_size, "SELECT DBSEQ_LMD_MISSION_PK.nextval "
            "FROM dual");
        query = ias_db_perform_query(database, sql);
        if (!ias_db_query_was_successful(query))
        {
            IAS_LOG_ERROR("Failed to obtain ID for lmd_mission");
            printf("%s\n", sql);
            db_close(database);
            return ERROR;
        }

        if (!ias_db_query_next(query))
        {
            IAS_LOG_ERROR("Failed to obtain ID for lmd_mission");
            db_close(database);
            return ERROR;
        }

        if (ias_db_query_int_value(query, &mission_id, 0) != SUCCESS)
        {
            IAS_LOG_ERROR("Failed to obtain ID for lmd_mission");
            db_close(database);
            return ERROR;
        }

        IAS_LOG_DEBUG("lmd_mission_id %d", mission_id); 

        /* Build the SQL statement */
        status = snprintf(sql, sql_statement_size,
            "INSERT INTO LMD_MISSION (LMD_MISSION_ID, "
            "LANDSAT_CAL_INTERVAL_ID, ORBIT, INTERVAL_FILES, NADIR_OFFNADIR, "
            "WRS_PATH, WRS_STARTING_ROW, WRS_ENDING_ROW, "
            "FULL_PARTIAL_INTERVAL, DATE_ACQUIRED, DATE_ENTERED, DATA_TYPE, "
            "DATA_CATEGORY_ORIGINAL, DATA_CATEGORY, STATION_ID, SENSOR_ID, "
            "MOE_INTERVAL_ID, MISSING_FILES, "
            "COLLECTION_TYPE)"
            "VALUES (%d, 'LC8BMCTEST%s', 1, 1, 'NADIR', 0, 0, 0, 'FULL'," 
            "TO_DATE('2011-08-08 08:08:08', 'YYYY-MM-DD HH24:MI:SS'),"
            "TO_DATE('2011-08-08 08:08:08', 'YYYY-MM-DD HH24:MI:SS'),"
            "'OLI_L0RA', 'NOMINAL', 'NOMINAL', 'EDC', 'OLI', 'B%s', 0,"
            "'OLI_SHUTTER')",
            mission_id, common_wo_id, common_wo_id);

        if (status < 0 || status >= sql_statement_size)
        {
            IAS_LOG_ERROR("Unable to create sql statement");
            return ERROR;
        }
        else
        {
            /*Set the return value back to SUCCESS*/
            status = SUCCESS;
        }

        /* Perform the query */
        query = ias_db_perform_query(database, sql);
        if (!ias_db_query_was_successful(query))
        {
            IAS_LOG_ERROR("Unable to perform insert with statement %s",
                sql);
            return ERROR;
        }

        /* Insert histogram statistics for every band */
        for (band_index = 0; band_index < sat_attributes->total_bands;
            band_index++)
        {
            int band_number = band_index + 1;
            int file_band_index = band_index;
            int pan_flag = 0;
            int raw_detector_count = 0; 

            band_attributes = ias_sat_attr_get_band_attributes(band_number);
            file_band_index = band_attributes->normal_band_number-1;

            if (band_attributes == NULL)
            {
                IAS_LOG_ERROR("Unable to obtain sat attributes for band "
                    "number %d", band_number);
                return ERROR;
            }

            if (band_attributes->sensor_id != IAS_OLI)
            {
                continue;
            }

            /* Only normal and VRP bands are used for bias model parameters */
            if (band_attributes->band_classification == IAS_NORMAL_BAND)
            {
                raw_detector_count = FILE_DETECTOR_COUNT;
            }
            else if (band_attributes->band_classification == IAS_VRP_BAND)
            {
                raw_detector_count = FILE_VRP_DETECTOR_COUNT;
            }
            else
            {
                IAS_LOG_DEBUG("Skipping data for band %d", band_index);
                continue;
            }

            //Account for pan band stored separately
            if (band_attributes->normal_band_number == 9)
            {
                file_band_index = FILE_NORMAL_BAND_COUNT - 1;
            }

            IAS_LOG_DEBUG("Inserting test data for RWOC %s band %d",
                common_wo_id, band_attributes->band_number);

            if (band_attributes->band_type == IAS_PAN_BAND)
            {
                pan_flag = 1;
            }
   
            for (sca_index = 0; sca_index < band_attributes->scas; sca_index+=1)
            {
                statistics = ias_rps_db_allocate_histogram_results(
                    band_attributes->detectors_per_sca, pan_flag);
                if (statistics == NULL) 
                {
                    IAS_LOG_ERROR("Unable to allocate statistics");
                    return ERROR;
                }

                statistics->dbase = database;
                strncpy(statistics->rad_common_id, common_wo_id, 
                    IAS_CHAR_ID_SIZE);

                strcpy(statistics->collection_type, "OLI_SHUTTER");
                statistics->band_number = band_attributes->band_number;
                statistics->sca = sca_index+1;
                statistics->num_detectors = band_attributes->detectors_per_sca;
                statistics->is_odd_even_data_split = pan_flag;
                statistics->processing_level = strdup(ingest_processing_level);
                if (statistics->processing_level == NULL) 
                {
                    return ERROR;
                }
 
                for (split_line = 0; split_line <= pan_flag; split_line++)
                for (detector = 0;
                    detector < band_attributes->detectors_per_sca;
                    detector++)
                {
                    /* The detectors are stored in the database as 1 based 
                        IDs*/
                    statistics->detector[split_line][detector] = detector+1;
                    statistics->num_values[split_line][detector] =
                        FILE_FRAME_COUNT;
                    statistics->min[split_line][detector] = min;
                    statistics->max[split_line][detector] = max;
                    statistics->stdev[split_line][detector] = std_dev;
                    statistics->skewness[split_line][detector] = skewness;
                    statistics->kurtosis[split_line][detector] = kurtosis;
                    statistics->correlation[split_line][detector] =
                        correlation;
                    statistics->mean_sq_resp[split_line][detector]=
                        mean_squared;
                   
                    /* VRP is stored separately */ 
                    if (band_attributes->band_classification & IAS_VRP_BAND)
                    {
                        if (pan_flag)
                        {
                            statistics->mean[split_line][detector] =
                                test_data->means_vrp_pan[collection]
                                    [sca_index][detector][split_line];
                        }
                        else
                        {
                            statistics->mean[split_line][detector] =
                                test_data->means_vrp[collection]
                                    [file_band_index][sca_index][detector];
                        }
                    }
                    else
                    {
                        if (pan_flag)
                        {
                            statistics->mean[split_line][detector] =
                                test_data->means_pan[collection]
                                    [sca_index][detector][split_line];
                        }   
                        else
                        {
                            statistics->mean[split_line][detector] =
                                test_data->means[collection][file_band_index]
                                    [sca_index][detector];
                        }
                    }
                } /*End of detector loop */

                if (ias_rps_store_histogram_stats_in_database(statistics) !=
                    SUCCESS)
                {
                    IAS_LOG_ERROR("Unable to store statistics for band %d, "
                        "sca %d", band_number, sca_index+1);
                    return ERROR;
                }

                /* statistics are freed by the store routine */
                statistics = NULL;
            } /*End of SCA Loop*/
 
            IAS_LOG_DEBUG("Inserting frame averages for band number %d", 
                band_attributes->band_number);
            
            for (sca_index = 0; sca_index < band_attributes->scas; sca_index+=1)
            {
                int frame_index = 0;

                frame_count = FILE_FRAME_COUNT;

                if (pan_flag)
                { 
                    frame_count *= 2;
                }

                frame_averages =
                    ias_rps_db_allocate_frame_averages(frame_count);
                frame_averages->dbase = database;
                strncpy(frame_averages->rad_common_id, common_wo_id, 
                    IAS_CHAR_ID_LENGTH);
                frame_averages->band_number = band_attributes->band_number;
                frame_averages->sca = sca_index+1;
                frame_averages->num_lines = frame_count;

                for (frame_index = 0; frame_index < frame_count; frame_index+=1)
                {
                    frame_averages->line_number[frame_index] = frame_index + 1;

                    /* VRP is stored separately */
                    if (band_attributes->band_classification & IAS_VRP_BAND)
                    {
                        if (band_attributes->band_type != IAS_PAN_BAND)
                        {
                            frame_averages->line_average[frame_index] =
                                test_data->vrp_frame_average[collection]
                                    [file_band_index][sca_index][frame_index];
                        }
                        else
                        {
                            frame_averages->line_average[frame_index] =
                                test_data->vrp_frame_average_pan[collection]
                                    [sca_index][frame_index];
                        }

                    }
                    else
                    {
                        if (band_attributes->band_type != IAS_PAN_BAND)
                        {
                            frame_averages->line_average[frame_index] =  
                                test_data->image_frame_average[collection]
                                    [file_band_index][sca_index][frame_index];
                        }
                        else
                        {
                            frame_averages->line_average[frame_index] =
                                test_data->image_frame_average_pan[collection]
                                    [sca_index][frame_index];
                        }
                    }
                } /* End of frame loop */

                if (SUCCESS != ias_rps_db_store_frame_averages(frame_averages))
                {
                    IAS_LOG_ERROR("Unable to store frame averages for "
                        "band index %d.", band_index);

                    ias_rps_db_free_frame_averages(&frame_averages);
                    return ERROR;
                }
 
                ias_rps_db_free_frame_averages(&frame_averages);
                frame_averages = NULL;

            }/*End SCA Loop*/
        }/*End band Loop*/
    }/*End of collection loop */

    IAS_LOG_DEBUG("Test data loaded into the database");

    return status;
}/*End  insert_test_data*/

/*----------------------------------------------------------------------------
NAME: bias_model_verification

PURPOSE: Test the ability to calculate the bias model parameters from 
    histogram statistics from the database.
-----------------------------------------------------------------------------*/
Ensure bias_model_verification(void)
{
    const int previous_collections_n0 = 1;
    const int previous_collections_n1 = 2;
    const struct IAS_BPF_BIAS_MODEL *bias_model = NULL;
    const IAS_SATELLITE_ATTRIBUTES *sat_attributes = NULL;

    double r0 = 0.0;
    double r1[FILE_DETECTOR_COUNT * LINESETS];

    char common_wo_id[IAS_CHAR_ID_SIZE];
    char mission_id[IAS_L0R_INTERVAL_ID_LENGTH];
    char cpf_name[] = "ias_lib/rps/bias_model_calibration/L8CPFTEST.01";
    char full_path_file[PATH_MAX];

    int band_index = 0;
    int collections_used = 0;
    int sca_index = 0;
    int status = SUCCESS;

    struct ias_db_connection *db_connection = db_open();

    IAS_BPF *bpf = NULL; 
    IAS_CPF *cpf = NULL;
    TEST_DATA *test_data = NULL;

    sat_attributes = ias_sat_attr_get_attributes();

    if (sat_attributes == NULL)
    {
        IAS_LOG_ERROR("Unable to obtain satellite attributes");
        assert_true(0);
        return;
    }

    if (db_connection == NULL)
    {
        IAS_LOG_ERROR("Unable to establish database connection");
        assert_true(0);
        return;
    }

    /* Build the full path for the CPF and read the file using the lib */
    sprintf(full_path_file, "%s/%s", test_data_path, cpf_name);
    cpf = ias_cpf_read(full_path_file);

    if (cpf == NULL)
    {
        IAS_LOG_ERROR("Unable to read CPF");
        assert_true(0);
        return;
    }

    test_data = get_test_data();
    if (test_data == NULL)
    {
        ias_cpf_free(cpf);
        cpf = NULL;
        IAS_LOG_ERROR("Unable to load test data");
        assert_true(0);
        return;
    }

    memset(common_wo_id, 0, IAS_CHAR_ID_SIZE);
 
    /*Insert the values for the workorder*/
    if (status == SUCCESS && 
        insert_test_data(db_connection, test_data, common_wo_id) != SUCCESS)
    {
        ias_cpf_free(cpf);
        cpf = NULL;
        free_test_data(&test_data);
        IAS_LOG_ERROR("Unable to insert RWOC %s", common_wo_id);
        assert_true(0);
        status = ERROR;
        return;
    }

    bpf = ias_bpf_allocate();
    if (bpf == NULL)
    {
        ias_cpf_free(cpf);
        cpf = NULL;
        free_test_data(&test_data);
        IAS_LOG_ERROR("Unable to allocate the bpf");
        assert_true(0);
        status = ERROR;
        return;
    }

    /* examine the bias parameters */
    bias_model = ias_bpf_get_bias_model(bpf);

    sprintf(mission_id, "LC8BMCTEST%s", common_wo_id);

    /* With the values inserted calculate the bias parameters */
    if (status == SUCCESS &&
        ias_rps_bias_model_calibration(cpf,
        db_connection, mission_id, common_wo_id,
        previous_collections_n0, previous_collections_n1, 0, bpf, 
        &collections_used, NULL) != SUCCESS)
    {
        IAS_LOG_ERROR("Unable to generate bias parameters");
        free_test_data(&test_data);
        status = ERROR;
        assert_true(0); 
        return;
    }

    /* Check the bias parameters calculated to those from the test data */
    for (band_index = 0; band_index < sat_attributes->total_bands;
        band_index++)
    {
        const int band_number = band_index + 1;
        const IAS_BAND_ATTRIBUTES *band_attributes =
            ias_sat_attr_get_band_attributes(band_number);

        double a0_coefficient = 0;
        double c1_coefficient[FILE_DETECTOR_COUNT * LINESETS];
        double a1_coefficient[FILE_DETECTOR_COUNT * LINESETS];
     
        double *pre_acquisition_shutter_average = NULL;
        double *post_acquisition_shutter_average = NULL;
        double *mean_shutter_average = NULL;
  
        int file_band_index = band_index; 
        int detector = 0;
        int even_odd = 0;
        int pan_flag = 0;

        if (band_attributes == NULL)
        {
            IAS_LOG_ERROR("Unable to obtain band attributes for band index %d",
                band_index);
            assert_true(0);
            return;
        }

        /* Bias parameters are calculated for the Normal OLI bands, skip
        the others */
        if ((band_attributes->band_classification != IAS_NORMAL_BAND) ||
            (band_attributes->sensor_id != IAS_OLI))
        {
            IAS_LOG_DEBUG("Skipping band number %d", 
                band_attributes->band_number);
            continue;
        }

        IAS_LOG_DEBUG("Verifying band number %d", band_attributes->band_number);

        //Account for the 4 "raw" PAN bands
        if (band_attributes->band_number == 9)
        {
            file_band_index = FILE_NORMAL_BAND_COUNT - 1;
        }

        if (status != SUCCESS)
        {
            break; 
        }

        int check_failed = 0;

        pre_acquisition_shutter_average = 
            calloc(band_attributes->detectors_per_sca, sizeof(double));
        post_acquisition_shutter_average = 
            calloc(band_attributes->detectors_per_sca, sizeof(double));
        mean_shutter_average = 
            calloc(band_attributes->detectors_per_sca, sizeof(double));

        if (pre_acquisition_shutter_average == NULL || 
            post_acquisition_shutter_average == NULL ||
            mean_shutter_average == NULL)
        {
            IAS_LOG_ERROR("Unable to perform allocation");
            status = ERROR;
            break;
        }

        if (band_attributes->band_type == IAS_PAN_BAND) pan_flag = 1;
       
        for (even_odd = 0; even_odd <= pan_flag; even_odd++)
        for (sca_index = 0; sca_index < band_attributes->scas; sca_index++)
        {
            int sca_number = sca_index + 1;

            if (!pan_flag)
            {
                even_odd = 2;
            }

            IAS_LOG_DEBUG("Verifying band number %d, sca index %d of %d", 
                band_number, sca_index, band_attributes->scas);

            if (bias_model == NULL)
            {
                assert_true(0);
                break;
            }
  
            IAS_LOG_DEBUG("Obtaining bias model parameters for band "
                "number %d of %d, sca index %d of %d", 
                band_number, sat_attributes->total_bands, sca_index, 
                band_attributes->scas);

            if (SUCCESS != ias_bpf_get_model_parameters(
                bias_model, band_attributes->spectral_type,
                band_attributes->band_number, 
                sca_number, 
                band_attributes->detectors_per_sca, 
                even_odd,              
                pre_acquisition_shutter_average,
                post_acquisition_shutter_average,
                a1_coefficient,     
                c1_coefficient, 
                &a0_coefficient))
            {
                assert_true(0);
                status = ERROR;
            }

            if (SUCCESS != ias_rps_db_get_bias_model_residual_a0(
                db_connection, common_wo_id,
                band_attributes->band_number, sca_number,
                even_odd, &r0))
            {
                IAS_LOG_ERROR("Unable to obtain r0 for band index %d, "
                    "sca index %d.",
                    band_attributes->band_index, sca_index);
                check_failed++;
            } 
 
            /* verify the values from the bias model are correct */
            if (pan_flag)
            {
                if (fabs(test_data->a0_pan[sca_index][even_odd] -
                    a0_coefficient) > DOUBLE_TOLERANCE)
                {
                    IAS_LOG_ERROR("Computed A0 value %lf != expected "
                        "value %lf", a0_coefficient,
                        test_data->a0_pan[sca_index][even_odd]);
                     check_failed++;
                }
                else
                {
                    IAS_LOG_DEBUG("SUCCESS: A0 value %lf == expected "
                        "value %lf", a0_coefficient,
                        test_data->a0_pan[sca_index][even_odd]);
                }

                if (DOUBLE_TOLERANCE <
                    fabs(test_data->r0_pan[sca_index][even_odd] - r0))
                {
                    IAS_LOG_ERROR("r0 computed for Pan band lineset %d sca %d "
                        "value %lf != expected value %lf",
                       even_odd, sca_index, r0, 
                       test_data->r0_pan[sca_index][even_odd]);
                    check_failed++;
                }
            }
            else
            {
                if (fabs(test_data->a0[file_band_index][sca_index] - 
                    a0_coefficient) > DOUBLE_TOLERANCE)
                {
                    IAS_LOG_ERROR("Computed A0 value for band index %d, %lf "
                        "!= expected value %lf", file_band_index,  
                        a0_coefficient, 
                        test_data->a0[file_band_index][sca_index]);
                     check_failed++;
                }
                else
                {
                    IAS_LOG_DEBUG("SUCCESS: A0 value for band index %d, %lf == "
                        "expected value %lf", file_band_index, a0_coefficient, 
                        test_data->a0[file_band_index][sca_index]);
                }

                if (DOUBLE_TOLERANCE <
                    fabs(test_data->r0[file_band_index][sca_index] - r0))
                {
                    IAS_LOG_ERROR("r0 computed for band %d sca %d "
                        "value %lf != expected value %lf",
                       band_attributes->band_number, sca_index,
                       r0, test_data->r0[file_band_index][sca_index]);
                       check_failed++;
                }
            }

            if (SUCCESS != ias_rps_db_get_bias_model_residual_a1(
                db_connection, common_wo_id, band_attributes,
                sca_number, even_odd, r1))
            {
                IAS_LOG_ERROR("Unable to obtain a1 for band "
                    "index %d, sca index %d.",
                    band_attributes->band_index, sca_index);
                check_failed++;
            }
   
            for (detector = 0; 
                detector < band_attributes->detectors_per_sca; 
                detector++)
            {
                int file_sca_index = sca_index;

                if (pan_flag)
                {
                    if (fabs(test_data->Sa_pan[even_odd][file_sca_index]
                        [detector] -
                        (float)pre_acquisition_shutter_average[detector]) 
                        > DOUBLE_TOLERANCE)
                    {
                        IAS_LOG_ERROR("Precollection shutter average for band "
                        "number %d detector %d sca %d odd/even %d value %lf "
                        "!= expected value %lf",
                        band_attributes->band_number, detector,
                        sca_index, even_odd, 
                        pre_acquisition_shutter_average[detector],
                        test_data->Sa_pan[even_odd][file_sca_index][detector]);
                        check_failed++;
                    }

                    if (fabs(test_data->Sb_pan[even_odd][file_sca_index]
                        [detector] -
                        (float)post_acquisition_shutter_average[detector]) 
                        > DOUBLE_TOLERANCE)
                    {
                        IAS_LOG_ERROR("Postcollection shutter average for band "
                        "number %d detector %d sca %d odd/even %d value %lf "
                        "!= expected value %lf",
                        band_attributes->band_number, detector,
                        sca_index, even_odd, 
                        post_acquisition_shutter_average[detector],
                        test_data->Sb_pan[even_odd][file_sca_index][detector]);
                        check_failed++;
                    }

                    if (fabs(test_data->a1_pan[even_odd][file_sca_index]
                        [detector] -
                        (float)a1_coefficient[detector]) > DOUBLE_TOLERANCE)
                    {
                        IAS_LOG_ERROR("A1 Computed for band number %d detector "
                        "%d sca %d odd/even %d value %lf != expected " 
                        "value %lf",
                        band_attributes->band_number, detector,
                        sca_index, even_odd, a1_coefficient[detector],
                        test_data->a1_pan[even_odd][file_sca_index][detector]);
                        check_failed++;
                    }

                    if (DOUBLE_TOLERANCE <
                        fabs((float)c1_coefficient[detector] -
                        test_data->c1_pan[even_odd][sca_index][detector]))
                    {
                        IAS_LOG_ERROR("C1 Computed for band %d detector "
                        "%d sca %d odd/even %d value %lf != expected "
                        "value %lf",
                        band_attributes->band_number, detector,
                        sca_index, even_odd, c1_coefficient[detector],
                        test_data->c1[even_odd][sca_index][detector]);
                        check_failed++;
                    }

                    if (DOUBLE_TOLERANCE < fabs(r1[detector] -
                        test_data->r1_pan[even_odd][sca_index][detector]))
                    {
                        IAS_LOG_ERROR("r1 computed for band %d sca %d "
                            "odd/even %d value %lf != %lf",
                            band_attributes->band_number, sca_index, 
                            even_odd, r1[detector],
                            test_data->r1_pan[even_odd][sca_index][detector]);
                        check_failed++;
                    }

                }
                else
                {
                    if (fabs(test_data->Sa[file_band_index][file_sca_index]
                        [detector] -
                        (float)pre_acquisition_shutter_average[detector]) 
                        > DOUBLE_TOLERANCE)
                    {
                        IAS_LOG_ERROR("pre collection shutter averages for "
                        "band number %d detector %d sca %d odd/even %d value "
                        "%lf != expected value %lf",
                        band_attributes->band_number, detector,
                        sca_index, even_odd, 
                        pre_acquisition_shutter_average[detector],
                        test_data->Sa[file_band_index][file_sca_index]
                            [detector]);
                        check_failed++;
                    }

                    if (DOUBLE_TOLERANCE < 
                        ((float)post_acquisition_shutter_average[detector] -
                        test_data->Sb[file_band_index][sca_index][detector]))
                    {
                        IAS_LOG_ERROR("C1 Computed for band %d detector "
                        "%d sca %d odd/even %d value %lf != expected "
                        "value %lf",
                        band_attributes->band_number, detector,
                        sca_index, even_odd, 
                        post_acquisition_shutter_average[detector],
                        test_data->Sb[file_band_index][sca_index][detector]);
                        check_failed++;
                    }

                    if (fabs(test_data->a1[file_band_index][file_sca_index]
                        [detector] -
                        (float)a1_coefficient[detector]) > DOUBLE_TOLERANCE)
                    {
                        IAS_LOG_ERROR("A1 Computed for band number %d detector "
                        "%d sca %d odd/even %d value %lf != expected " 
                        "value %lf",
                        band_attributes->band_number, detector, 
                        sca_index, even_odd, a1_coefficient[detector], 
                        test_data->a1[file_band_index][file_sca_index]
                            [detector]);
                        check_failed++;    
                    }

                    if (DOUBLE_TOLERANCE < 
                        fabs((float)c1_coefficient[detector] -
                        test_data->c1[file_band_index][sca_index][detector]))
                    {
                        IAS_LOG_ERROR("C1 Computed for band %d detector "
                        "%d sca %d odd/even %d value %lf != expected "
                        "value %lf",
                        band_attributes->band_number, detector,
                        sca_index, even_odd, c1_coefficient[detector],
                        test_data->c1[file_band_index][sca_index][detector]);
                        check_failed++;
                    }

                    if (DOUBLE_TOLERANCE < fabs(r1[detector] -
                        test_data->r1[file_band_index][sca_index][detector]))
                    {
                        IAS_LOG_ERROR("r1 computed for band %d sca %d "
                        "detector %d value %lf != %lf",
                        band_attributes->band_number, sca_index, 
                        detector, r1[detector], 
                        test_data->r1[file_band_index][sca_index][detector]);
                        check_failed++;
                    }
                }
            }/* End of detector loop */
        }/* End SCA loop */

        if (check_failed != 0)
        {
            IAS_LOG_DEBUG("Band number %d failed", band_number);
            assert_true(0);
        }
        else
        {
            IAS_LOG_DEBUG("Band number %d success", band_number);
            assert_true(1);
        }

        free(pre_acquisition_shutter_average);
        free(post_acquisition_shutter_average);
        free(mean_shutter_average);

        pre_acquisition_shutter_average = NULL;
        post_acquisition_shutter_average = NULL;
        mean_shutter_average = NULL;
    } /* End of band for loop */

    ias_bpf_free(bpf);
    bpf = NULL;
  
    ias_cpf_free(cpf);
    cpf = NULL;

    free_test_data(&test_data);
    db_close(db_connection);     

    IAS_LOG_DEBUG("Bias Model Cal test complete");
} /* End of bias_model_verification */

int main(int argc, char *argv[])
{
    /* Construct the test suite */
    TestSuite *suite = create_test_suite();

    /* Set the log level for regression test execution of unit tests */
    ias_log_set_output_level(IAS_LOG_LEVEL_WARN);

    /* Get the test data directory */
    test_data_path = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (test_data_path == NULL)
    {
        printf("Cannot locate IAS test data directory\n");
        return 1;
    }

    if (ias_sat_attr_initialize(IAS_L8) != SUCCESS)
    {
        IAS_LOG_ERROR("Unable to load the satellite attributes used for "
            "processing.");
        return ERROR;
    }
    
    /* Use the env variable to determine if the database related tests should 
    be run */
    if (getenv("ENABLE_DB_TESTS") == NULL)
    {
        printf("DB Tests will not run, setenv ENABLE_DB_TESTS to enable....\n");
        return SUCCESS;
    }

    IAS_LOG_DEBUG("Running Bias Model Parameter verification");   

    add_test(suite, bias_model_verification);

    return run_test_suite(suite, create_text_reporter());
}

