/**********************************************************************
 NAME:      calculate_filtered_statistics_test

 PURPOSE:   Use CGREEN to perform unit testing of the Radiometric
            Library algorithms used to calculate summary statistics 
            (e.g. mean, stdev, RMS) with filtering used to exclude
            invalid sample values

 ROUTINES:  test_calculate_filter_0
            test_calculate_filter_258
            test_calculate_filter_245
***********************************************************************/
#include <float.h>
#include <math.h>
#include <string.h>

#include "ias_radiometry.h"
#include "ias_satellite_attributes.h"
#include "cgreen.h"


#define NSAMPS  10
static double data_vector[NSAMPS] =
    {2.5, 2.45, 2.6, 2.55, 2.88, 2.52, 2.58, 2.61, 2.45, 2.57};


/* Test with no invalid data */
Ensure test_calculate_filter_none_invalid()
{
    int status;
    int flag_vector[NSAMPS];
    double returned_mean;
    double returned_stdev;
    double returned_rms;
    double expected_mean;
    double expected_stdev;
    double expected_rms;
    int returned_count;
    int expected_count;

    /* Initialization */
    expected_mean  = 2.571;
    expected_stdev = 0.122606;
    expected_rms   = 0.1163142;
    expected_count = 10;

    memset(flag_vector, 0, sizeof(flag_vector));

    returned_count = 0;
    status = ias_rps_calculate_filtered_mean(data_vector, flag_vector,
        NSAMPS, &returned_mean, &returned_count);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error performing filtered mean "
            "calculation");
        return;
    }
    assert_equal(returned_count, expected_count);

    status = ias_rps_calculate_filtered_stdev(data_vector, flag_vector,
        NSAMPS, returned_mean, &returned_stdev);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error performing standard deviation "
            "calculation");
        return;
    }

    returned_count = 0;
    status = ias_rps_calculate_filtered_rms(data_vector, flag_vector,
        NSAMPS, returned_mean, &returned_rms, &returned_count);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error performing filtered RMS "
            "calculation");
        return;
    }
    assert_equal(returned_count, expected_count);

    /* Check the results */
    assert_double_equal( returned_mean, expected_mean );
    assert_double_equal( returned_stdev, expected_stdev );
    assert_double_equal( returned_rms, expected_rms );
}



/* Test with two data values set invalid */
Ensure test_calculate_filter_with_invalid()
{
    int status;
    int flag_vector[NSAMPS];
    double returned_mean;
    double returned_stdev;
    double returned_rms;
    double expected_mean;
    double expected_stdev;
    double expected_rms;
    int returned_count;
    int expected_count;

    /* Initialization */
    expected_mean = 2.60125;
    expected_stdev = 0.118736;
    expected_rms = 0.1110672;
    expected_count = 8;

    memset(flag_vector, 0, sizeof(flag_vector));
    flag_vector[1] = 1;
    flag_vector[8] = 1;

    status = ias_rps_calculate_filtered_mean(data_vector, flag_vector,
        NSAMPS, &returned_mean, &returned_count);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error running filtered mean "
            "calculation");
        return;
    }
    assert_equal(returned_count, expected_count);

    status = ias_rps_calculate_filtered_stdev(data_vector, flag_vector,
        NSAMPS, returned_mean, &returned_stdev);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error running filtered standard "
            "deviation calculation");
        return;
    }

    returned_count = 0;
    status = ias_rps_calculate_filtered_rms(data_vector, flag_vector,
        NSAMPS, returned_mean, &returned_rms, &returned_count);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error running filtered RMS "
             "calculation");
        return;
    }
    assert_equal(returned_count, expected_count);

    /* Check the results */
    assert_double_equal( returned_mean, expected_mean );
    assert_double_equal( returned_stdev, expected_stdev );
    assert_double_equal( returned_rms, expected_rms );
}




int main(int argc, char *argv[])
{
    TestSuite *suite = create_test_suite();

    add_test(suite, test_calculate_filter_none_invalid);
    add_test(suite, test_calculate_filter_with_invalid);

    if (argc > 1)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}

