/**********************************************************************
                calculate_radiance_test

 PURPOSE:      Use CGREEN to perform unit testing of the algorithm that
               calculates the corresponding radiance values for a data
               vector of DN values

***********************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <float.h>
#include "ias_miscellaneous.h"
#include "ias_radiometry.h"
#include "ias_satellite_attributes.h"
#include "cgreen.h"

static char *test_data_path = NULL;

Ensure test_calculate_radiance()
{
    char test_data_dir[] = "ias_lib/rps";
    char hist_stats_basename[] = "simulated_hist_stats.txt";
    const char rad_basename[] = "derived_radiances_no_inopdetectors.txt";
    char full_hist_stats_filename[PATH_MAX];
    char full_rad_filename[PATH_MAX];
    const IAS_BAND_ATTRIBUTES *band_attributes = NULL;
    double abs_gain = 126.5;
    double *relative_gains = NULL;    /* Relative gains for the current
                                         band/SCA */
    double *hist_stats_buffer = NULL; /* Histogram statistics for the
                                         current band/SCA */
    double *radiance_buffer = NULL;   /* Radiances for all detectors in the
                                         current band */
    double *expected_radiance;        /* Benchmark radiance values */
    FILE *data_fptr = NULL;
    int dd;                            /* Detector loop counter */
    int num_band_detectors = 0;

    sprintf(full_hist_stats_filename, "%s/%s/%s", test_data_path,
        test_data_dir, hist_stats_basename);
    sprintf(full_rad_filename, "%s/%s/%s", test_data_path,
        test_data_dir, rad_basename);

    band_attributes = ias_sat_attr_get_band_attributes(1);
    assert_true(band_attributes != NULL);
    if (band_attributes == NULL)
    {
        printf("Cannot get band 1 attributes\n");
        return;
    }
    num_band_detectors = 
        band_attributes->scas * band_attributes->detectors_per_sca;

    /* Allocate the data buffers */
    relative_gains = malloc(band_attributes->detectors_per_sca
        * sizeof(*relative_gains));
    hist_stats_buffer = malloc(band_attributes->detectors_per_sca
        * sizeof(*hist_stats_buffer));
    radiance_buffer = malloc(num_band_detectors * sizeof(*radiance_buffer));
    expected_radiance = malloc(num_band_detectors *
        sizeof(*expected_radiance));

    assert_true( relative_gains != NULL  && hist_stats_buffer != NULL
         && radiance_buffer != NULL && expected_radiance != NULL);

    if ( relative_gains == NULL || hist_stats_buffer == NULL
         || radiance_buffer == NULL || expected_radiance == NULL )
    {
        printf("Data buffer allocation failed\n");
        free(relative_gains);
        free(hist_stats_buffer);
        free(radiance_buffer);
        free(expected_radiance);
        return;
    }

    data_fptr = fopen( full_rad_filename, "r");
    assert_true( data_fptr != NULL );
    if ( !data_fptr )
    {
        printf("Error opening file %s\n", full_rad_filename );
        free(relative_gains);
        free(hist_stats_buffer);
        free(radiance_buffer);
        free(expected_radiance);
        return;
    }
    for ( dd = 0; dd < band_attributes->detectors_per_sca; dd++ )
    {
        int count;

        count = fscanf( data_fptr, "%*d %lf", &expected_radiance[dd] );
        if ( count != 1 )
        {
            assert_equal( count, 1 );
            free(relative_gains);
            free(hist_stats_buffer);
            free(radiance_buffer);
            free(expected_radiance);
            fclose( data_fptr );
            return;
        }
    }
    fclose( data_fptr );
    data_fptr = NULL;

    data_fptr = fopen(full_hist_stats_filename, "r");
    assert_true( data_fptr != NULL );
    if (data_fptr == NULL)
    {
        printf("Cannot open simulated hist statistics data file %s\n",
               full_hist_stats_filename);
        free(relative_gains);
        free(hist_stats_buffer);
        free(radiance_buffer);
        free(expected_radiance);
        return;
    }

    /* Read the histogram statistics and relative gains from a text file.
       Only the first SCA is tested here. */
    for (dd = 0; dd < band_attributes->detectors_per_sca; dd++)
    {
        int count;

        count = fscanf(data_fptr, "%*d%lf%lf", &hist_stats_buffer[dd],
            &relative_gains[dd]);
        if ( count != 2 )
        {
            assert_equal( count, 2 );
            free(relative_gains);
            free(hist_stats_buffer);
            free(radiance_buffer);
            free(expected_radiance);
            fclose( data_fptr );
            return;
        }
    }
    fclose(data_fptr);

    /* Calculate the radiances for all detectors in the current SCA */
    ias_rps_calculate_radiance(hist_stats_buffer,
            band_attributes->detectors_per_sca,
            abs_gain, relative_gains, radiance_buffer);

    /* Check results */
    for ( dd = 0; dd < band_attributes->detectors_per_sca; dd++ )
        assert_double_equal( radiance_buffer[dd], expected_radiance[dd] );

    /* Cleanup */
    free(radiance_buffer);
    free(relative_gains);
    free(hist_stats_buffer);
    free(expected_radiance);
}


int main(int argc, char *argv[])
{
    IAS_SATELLITE_ID satellite_id;
    int status;

    /* First, get the test data directory location and construct the
       input and output file names */
    test_data_path = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (test_data_path == NULL)
    {
        printf("Cannot find test data directory\n");
        return 1;
    }

    /* Initialize the satellite attributes library */
    satellite_id = ias_sat_attr_get_satellite_id_from_satellite_number(8);
    if (satellite_id != IAS_L8)
    {
        printf("Cannot obtain satellite ID from sensor string\n");
        return 1;
    }

    status = ias_sat_attr_initialize(satellite_id);
    if (status != SUCCESS)
    {
        printf("Cannot initialize satellite attributes library\n");
        return 1;
    }

    TestSuite *suite = create_test_suite();

    add_test(suite, test_calculate_radiance);

    if (argc > 1)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}
