#include <stdio.h>
#include <string.h>
#include <limits.h>

#include "ias_const.h"
#include "ias_cpf.h"
#include "ias_miscellaneous.h"
#include "ias_satellite_attributes.h"
#include "ias_logging.h"
#include "ias_model_io.h"
#include "cgreen.h"

#include "../ias_rps_calculate_sca_offsets.h"

#define NUMBER_OLI_SCAS    14
#define NUMBER_TIRS_SCAS    3

static char *test_data_path = NULL;
static char cpf_dir[] = "ias_lib/rps/calculate_sca_offsets";
static char cpf_name[] = "L8CPF20111001_20111231.01";


/*---------------------------------------------------------------------
 NAME:     load_legendre_coeff_array

 PURPOSE:  Populates an allocated array with legendre coefficients for
           all SCAs in the specified band

 RETURNS:  Nothing
-----------------------------------------------------------------------*/
static void load_legendre_coeff_array
(
    const struct IAS_CPF_FOCAL_PLANE *focal_plane,
                                /* I: Focal plane group */
    int    band_index,          /* I: Index to current OLI/TIRS band */
    int    number_scas,         /* I: Number of SCAs in current band */
    double *legendre_coeffs     /* O: Populated array of Legendre
                                   coefficients for the current band.
                                   Layout is equal to that of
                                   &focal_plane->along_los_legendre[b][0][0],
                                   where b is the current band index */
)
{
    int sca_index;

    for (sca_index = 0; sca_index < number_scas; sca_index++)
    {
        int i;
        for (i = 0; i < IAS_LOS_LEGENDRE_TERMS; i++)
        {
            int coeff_offset = sca_index * IAS_LOS_LEGENDRE_TERMS + i;
            legendre_coeffs[coeff_offset] =
                focal_plane->along_los_legendre[band_index][sca_index][i];
        }
    }
}


/*----------------------------------------------------------------------
 NAME:     run_offset_calculation

 PURPOSE:  Wrapper routine used to call ias_rps_legendre_calculate_sca_offsets

 RETURNS:  Nothing
------------------------------------------------------------------------*/
static void run_offset_calculation
(
    const char *cpf_name,       /* I: CPF file name */
    const char *sensor_id,      /* I: Sensor ID string ("OLI", "TIRS") */
    int        band_number,     /* I: Current band */
    int        number_scas,     /* I: Number of SCAs in current band */
    int        *band_offset,    /* O: Band offset in frames */
    int        **sca_offsets    /* O: SCA offsets in frames */
)
{
    IAS_CPF  *cpf;
    const struct IAS_CPF_EARTH_CONSTANTS *earth_const      = NULL;
    const struct IAS_CPF_FOCAL_PLANE     *focal_plane      = NULL;
    const struct IAS_CPF_ORBIT_PARAMETERS *orbit_params    = NULL;
    const struct IAS_CPF_OLI_PARAMETERS  *oli_params       = NULL;
    const struct IAS_CPF_TIRS_PARAMETERS *tirs_params      = NULL;

    double nominal_frame_time;
    double *legendre_coeffs = NULL;
    double scene_center_latitude = 44.3;   /* Scene center lat for
                                              Brookings, SD */
    double elevation = 0.5;                /* Scene center elevation for
                                              Brookings, SD */


    /* Read the CPF contents into memory */
    printf("Reading CPF contents...\n");
    cpf = ias_cpf_read(cpf_name);
    if (cpf == NULL)
    {
        assert_true_with_message(FALSE, "Error reading CPF %s\n", cpf_name);
        return;
    }

    /* Get the "common" parameter groups */
    earth_const = ias_cpf_get_earth_const(cpf);
    focal_plane = ias_cpf_get_focal_plane(cpf);
    orbit_params = ias_cpf_get_orbit(cpf);
    if ((earth_const == NULL) || (focal_plane == NULL)
        || (orbit_params == NULL))
    {
        assert_true_with_message(FALSE, "Error retrieving required CPF "
            "parameter group(s)\n");
        ias_cpf_free(cpf);
        return;
    }

    /* Now get the sensor specific parameter group.  Pull the nominal
       frame time from that group, and be sure to convert it to seconds
       (as it's in units of milliseconds in the CPF) */
    if (strcmp(sensor_id, "OLI") == 0)
    {
        oli_params = ias_cpf_get_oli_parameters(cpf);
        if (oli_params == NULL)
        {
            assert_true_with_message(FALSE, "Error retrieving required "
                "OLI parameters");
            ias_cpf_free(cpf);
            return;
        }
        nominal_frame_time = 0.001 * oli_params->nominal_frame_time;

        /* Cut the frame time in half for the PAN band */
        if (band_number == 8)
        {
            nominal_frame_time /= 2.0;
        }
    }
    else if (strcmp(sensor_id, "TIRS") == 0)
    {
        tirs_params = ias_cpf_get_tirs_parameters(cpf);
        if (tirs_params == NULL)
        {
            assert_true_with_message(FALSE, "Error retrieving required "
                "TIRS parameters");
            ias_cpf_free(cpf);
            return;
        }
        nominal_frame_time = 0.001 * tirs_params->nominal_frame_time;
    }
    else
    {
        assert_true_with_message(FALSE, "Invalid sensor ID string %s, "
            "should be either OLI or TIRS", sensor_id);
        ias_cpf_free(cpf);
        return;
    }

    /* Allocate the Legendre coefficients data buffer, then load it */
    legendre_coeffs = malloc(number_scas * IAS_LOS_LEGENDRE_TERMS *
        sizeof(double));
    if (legendre_coeffs == NULL)
    {
        assert_true_with_message(FALSE, "Error allocating Legendre "
            "coefficient data buffer");
        ias_cpf_free(cpf);
        return;
    }
    load_legendre_coeff_array(focal_plane, band_number-1, number_scas,
        legendre_coeffs);

    /* Call the library routine */
    ias_rps_legendre_calculate_sca_offsets(legendre_coeffs, number_scas,
        scene_center_latitude, elevation, nominal_frame_time,
        earth_const->eccentricity,
        0.001 * earth_const->semi_major_axis,
        0.001 * earth_const->semi_minor_axis,
        earth_const->earth_angular_velocity,
        orbit_params->inclination_angle,
        orbit_params->nominal_orbit_radius,
        orbit_params->orbital_period,
        band_offset,
        *sca_offsets);

    /* Clean up */
    free(legendre_coeffs);
    ias_cpf_free(cpf);
}   /* END -- run_offset_calculation */



Ensure test_calculate_tirs_sca_offsets()
{
    char cpf_filename[PATH_MAX];

    int expected_band10_offset = 341;
    int expected_band10_sca_offsets[NUMBER_TIRS_SCAS] = {0, -635, -1};
    int *sca_offsets;
    int band_offset;
    int sca_index;
    int number_diffs_more_than_two_frames = 0;
    int status;


    printf("Checking band 10 offsets...\n");

    /* Construct the CPF filename */
    status = snprintf(cpf_filename, sizeof(cpf_filename), "%s/%s/%s",
        test_data_path, cpf_dir, cpf_name);
    if ((status < 0) || (status >= sizeof(cpf_filename)))
    {
        assert_true_with_message(FALSE, "Error creating CPF file name");
        return;
    }

    /* Allocate the SCA offsets data buffer */
    sca_offsets = malloc(NUMBER_TIRS_SCAS * sizeof(int));
    if (sca_offsets == NULL)
    {
        assert_true_with_message(FALSE, "Error allocating TIRS SCA offset "
            "data buffer");
        return;
    }

    /* Run the function */
    run_offset_calculation(cpf_filename, "TIRS", 10, NUMBER_TIRS_SCAS,
        &band_offset, &sca_offsets);

    /* Check the results */
    assert_equal_with_message(band_offset, expected_band10_offset,
        "Band 10 offset %d does not match expected band 10 offset %d",
        band_offset, expected_band10_offset);

    /* Due to roundoff effects,etc, there may be a one frame difference
       between the offsets calculated by the current implementation and 
       the offsets generated from the prototype code.  So instead of
       testing for exact equality, check for a maximum one-frame difference
       between the prototype and test results.  If the difference is two or
       more frames, the test fails */
    for (sca_index = 0; sca_index < NUMBER_TIRS_SCAS; sca_index++)
    {
        int frame_diff = abs(sca_offsets[sca_index]
            - expected_band10_sca_offsets[sca_index]);
        if (frame_diff >= 2)
            number_diffs_more_than_two_frames++;
    }
    assert_equal_with_message(number_diffs_more_than_two_frames, 0,
        "At least one Band 10 SCA has offset differences of 2 or more "
        "frames");

    /* Cleanup */
    free(sca_offsets);

}   /* END -- test_calculate_tirs_sca_offsets */



Ensure test_calculate_oli_sca_offsets()
{
    char cpf_filename[PATH_MAX];

    int expected_band1_sca_offsets[NUMBER_OLI_SCAS] = {0, -288, 18, -273,
        29, -265, 33, -263, 32, -268, 24, -280, 10, -298};
    int expected_band8_sca_offsets[NUMBER_OLI_SCAS] = {0, -332, 36, -302,
        58, -285, 68, -281, 65, -291, 49, -315, 20, -352};
    int expected_band1_offset = 144;
    int expected_band8_offset = 166;
    int *sca_offsets;
    int band_offset;
    int sca_index;
    int number_diffs_more_than_two_frames = 0;
    int status;


    /* Construct the CPF filename */
    status = snprintf(cpf_filename, sizeof(cpf_filename), "%s/%s/%s",
        test_data_path, cpf_dir, cpf_name);
    if ((status < 0) || (status >= sizeof(cpf_filename)))
    {
        assert_true_with_message(FALSE, "Error creating CPF file name");
        return;
    }

    /* Allocate the SCA offsets data buffer */
    sca_offsets = malloc(NUMBER_OLI_SCAS * sizeof(int));
    if (sca_offsets == NULL)
    {
        assert_true_with_message(FALSE, "Error allocating TIRS SCA offset "
            "data buffer");
        return;
    }

    /****************** BAND 1 TEST ************************************/
    printf("Checking band 1 offsets...\n");

    /* Run the function for band 1 */
    run_offset_calculation(cpf_filename, "OLI", 1, NUMBER_OLI_SCAS,
        &band_offset, &sca_offsets);

    /* Check the band 1 results */
    assert_equal_with_message(band_offset, expected_band1_offset,
        "Band 1 offset %d does not match expected band 1 offset %d",
        band_offset, expected_band1_offset);

    /* Due to roundoff effects,etc, there may be a one frame difference
       between the offsets calculated by the current implementation and 
       the offsets generated from the prototype code.  So instead of
       testing for exact equality, check for a maximum one-frame difference
       between the prototype and test results.  If the difference is two or
       more frames, the test fails */
    for (sca_index = 0; sca_index < NUMBER_OLI_SCAS; sca_index++)
    {
        int frame_diff = abs(sca_offsets[sca_index]
            - expected_band1_sca_offsets[sca_index]);
        if (frame_diff >= 2)
            number_diffs_more_than_two_frames++;
    }
    assert_equal_with_message(number_diffs_more_than_two_frames, 0,
        "At least one Band 1 SCA has offset differences of 2 or more "
        "frames");

    /****************** BAND 8 TEST ************************************/
    printf("Checking band 8 offsets...\n");

    /* Run the function for band 1 */
    run_offset_calculation(cpf_filename, "OLI", 8, NUMBER_OLI_SCAS,
        &band_offset, &sca_offsets);

    /* Check the band 1 results */
    assert_equal_with_message(band_offset, expected_band8_offset,
        "Band 8 offset %d does not match expected band 8 offset %d",
        band_offset, expected_band8_offset);

    /* Due to roundoff effects,etc, there may be a one frame difference
       between the offsets calculated by the current implementation and 
       the offsets generated from the prototype code.  So instead of
       testing for exact equality, check for a maximum one-frame difference
       between the prototype and test results.  If the difference is two or
       more frames, the test fails */
    for (sca_index = 0; sca_index < NUMBER_OLI_SCAS; sca_index++)
    {
        int frame_diff = abs(sca_offsets[sca_index]
            - expected_band8_sca_offsets[sca_index]);
        if (frame_diff >= 2)
            number_diffs_more_than_two_frames++;
    }
    assert_equal_with_message(number_diffs_more_than_two_frames, 0,
        "At least one Band 8 SCA has offset differences of 2 or more "
        "frames");

    /* Cleanup */
    free(sca_offsets);
}   /* END -- test_calculate_oli_sca_offsets */


int main(int argc, char *argv[])
{
    int status;

    /* Get the test directory location */
    test_data_path = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (test_data_path == NULL)
    {
        printf("Cannot find IAS test data directory\n");
        exit(1);
    }

    /* Initialize the satellite attributes */
    status = ias_sat_attr_initialize(IAS_L8);
    if (status != SUCCESS)
    {
        printf("Cannot initialize IAS satellite attribute\n");
        exit(2);
    }

    /* Construct the test suite */
    TestSuite *suite = create_test_suite();
    add_test(suite, test_calculate_oli_sca_offsets);
    add_test(suite, test_calculate_tirs_sca_offsets);

    /* Run the test suite */
    if (argc > 1)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}   /* END -- main */
