#include <stdio.h>
#include <string.h>
#include <math.h>
#include <limits.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "ias_pixel_mask.h"
#include "ias_logging.h"
#include "ias_radiometry.h"
#include "ias_satellite_attributes.h"

#include "cgreen.h"

#include "ias_rps_histogram_statistics_characterization.h"
#include "../ias_rps_tirs_dark_response_determination.h"


static char *test_data_path = NULL;
static char test_data_directory[] =
    "ias_lib/rps/tirs_dark_response_determination";

/*-------------------------------------------------------------------------
 NAME:                 check_group_files

 PURPOSE:  Internal routine used to compare an adjusted dark response
           output file produced by the TIRS dark response determination
           algorithm to a "reference" output file

 RETURNS:  Integer status code of SUCCESS or ERROR.  ERROR is only
           returned if the test and/or reference file cannot be read

--------------------------------------------------------------------------*/
static int check_group_files(const char *test_group_filename,
    const char *reference_group_filename, double pass_threshold)
{
    char test_buff[1000] = "";             /* String buffer used for
                                              "test" data file read */
    char reference_buff[1000] = "";           /* String buffer used for
                                              "reference" data file read */

    FILE *test_stats_fptr = NULL;          /* "Test" data file */
    FILE *reference_stats_fptr = NULL;     /* "reference data file for
                                              comparison */

    double diff = 0.0;                     /* Absolute difference between
                                              corresponding test and
                                              reference data */
    double test_response[5] = {0.0, 0.0, 0.0, 0.0, 0.0};
                                           /* Array of test dark responses
                                              from a line in the test
                                              output file */
    double reference_response[5] = {0.0, 0.0, 0.0, 0.0, 0.0};
                                           /* Array of reference dark responses
                                              from a line in the reference
                                              output file */

    int ii = 0;                            /* Loop counter */
    int jj = 0;                            /* Loop counter */
    int response_diffs = 0;                /* # of differences in dark
                                              response (for a given detector
                                              between the "test" data and the
                                              "reference" data) exceeding
                                              the specified threshold
                                              value*/
    int status = SUCCESS;


    /* Open the group files */
    test_stats_fptr = fopen(test_group_filename, "r");
    reference_stats_fptr = fopen(reference_group_filename, "r");

    assert_not_equal_with_message(test_stats_fptr, NULL, "Cannot open test "
        "group file %s", test_group_filename);
    assert_not_equal_with_message(reference_stats_fptr, NULL,
        "Cannot open reference group file %s", reference_group_filename);
    if ((test_stats_fptr == NULL) || (reference_stats_fptr == NULL))
    {
        if (test_stats_fptr)
            fclose(test_stats_fptr);
        if (reference_stats_fptr)
            fclose(reference_stats_fptr);
        return ERROR;
    }

    /* Check the GROUP header and parameter names to make sure the
       test one is correctly written */
    fgets(test_buff, 1000, test_stats_fptr);
    fgets(reference_buff, 1000, reference_stats_fptr);
    assert_string_equal_with_message(test_buff, reference_buff,
        "GROUP header statement difference");

    fgets(test_buff, 1000, test_stats_fptr);
    fgets(reference_buff, 1000, reference_stats_fptr);
    assert_string_equal_with_message(test_buff, reference_buff,
        "Parameter name difference");

    /* Check each line of data.  As currently implemented,
       there are 5 dark response values per line.  For 640
       detectors, that translates into 128 lines of dark response
       values */
    for (ii = 0; ii < 128; ii++)
    {
        status = fscanf(test_stats_fptr,
            "%lf,%lf,%lf,%lf,%lf%*[ ,)]\n",
            &test_response[0], &test_response[1],
            &test_response[2], &test_response[3],
            &test_response[4]);
        if (status < 0)
        {
            assert_true_with_message(FALSE, "Cannot read test line %d",
                (ii + 1));
            fclose(test_stats_fptr);
            fclose(reference_stats_fptr);
            return ERROR;
        }

        status = fscanf(reference_stats_fptr,
            "%lf,%lf,%lf,%lf,%lf%*[ ,)]\n",
            &reference_response[0], &reference_response[1],
            &reference_response[2], &reference_response[3],
            &reference_response[4]);
        if (status < 0)
        {
            assert_true_with_message(FALSE, "Cannot read reference line %d",
                (ii + 1));
            fclose(test_stats_fptr);
            fclose(reference_stats_fptr);
            return ERROR;
        }

        /* Check the absolute difference for each set of dark responses.
           Count those that are greater than the passing threshold */
        for (jj = 0; jj < 5; jj++)
        {
            diff = fabs(test_response[jj] - reference_response[jj]);
            if (diff > pass_threshold)
            {
                response_diffs++;
            }
        }
    }

    assert_equal_with_message(response_diffs, 0,
        "%d differences of %lf or greater detected", pass_threshold);

    /* Check the END_GROUP and END statements */
    fgets(test_buff, 1000, test_stats_fptr);
    fgets(reference_buff, 1000, reference_stats_fptr);
    assert_string_equal_with_message(test_buff, reference_buff,
        "END_GROUP header statement difference");

    fgets(test_buff, 1000, test_stats_fptr);
    fgets(reference_buff, 1000, reference_stats_fptr);
    assert_string_equal_with_message(test_buff, reference_buff,
        "END closing statement difference");

    /* Close the (blind band) files */
    fclose(test_stats_fptr);
    fclose(reference_stats_fptr);

    /* Done */
    return SUCCESS;
}



Ensure test_calculate_adjusted_response_blind()
{
    IAS_RPS_HISTOGRAM_INTERMEDIATE_RESULTS *blind_stats = NULL;

    char dr_filename[PATH_MAX];            /* Test baseline dark response
                                              file name (test data in .txt
                                              file) */
    char gain_filename[PATH_MAX];          /* Test detector gains file
                                              name (test data in .txt
                                              file) */
    char scene_mean_filename[PATH_MAX];    /* Test scene mean histogram
                                              statistics file name
                                              (test data in .txt file) */
    char ref_adr_group_filename[PATH_MAX] = "";   /* Name of reference
                                                     ODL-formatted group
                                                     file */
    char test_adr_group_filename[PATH_MAX] = "";  /* Name of test
                                                     ODL-formatted group
                                                     file */

    /* The structure of these ASCII data files consists of one value per
       detector */
    FILE *dr_fptr = NULL;                  /* Baseline dark response
                                              data file */
    FILE *gain_fptr = NULL;                /* Absolute detector gains
                                              data file */
    FILE *means_fptr = NULL;               /* Image hist stats means
                                              data file */

    IAS_PIXEL_MASK *pmask = NULL;          /* Pixel mask */

    int is_blind_band = 0;                 /* Blind band flag */
    int blind_band_number = 15;            /* Current TIRS blind
                                              band number */
    int band_number = 15;
    int sca_number = 1;
    int number_detectors = 640;
    int number_rows = 2;
    int det_index = 0;
    int status = SUCCESS;
    int dr_status = SUCCESS;
    int gain_status = SUCCESS;
    int means_status = SUCCESS;

    double *dark_response = NULL;              /* Dark response */
    double *relative_gains = NULL;             /* Relative detector gains */
    double sca_dark_response_difference = 0.0; /* Mean dark response
                                                  difference for the
                                                  current SCA */
    double prototype_dark_response_difference = 1.29237;
    double abs_gain = 1.0;                     /* Absolute gain used in
                                                  prototype code */

    struct stat buff;


    printf("Testing calculation of adjusted blind band dark response...\n");

    /* Open the data files for reading */
    sprintf(dr_filename, "%s/%s/dr_B%02d_SCA%02d.txt", test_data_path,
        test_data_directory, blind_band_number, sca_number);
    sprintf(gain_filename, "%s/%s/rg_B%02d_SCA%02d.txt", test_data_path,
        test_data_directory, blind_band_number, sca_number);
    sprintf(scene_mean_filename, "%s/%s/sc_B%02d_SCA%02d.txt",
        test_data_path, test_data_directory, blind_band_number, sca_number);

    dr_fptr = fopen(dr_filename, "r");
    gain_fptr = fopen(gain_filename, "r");
    means_fptr = fopen(scene_mean_filename, "r");

    assert_not_equal_with_message(dr_fptr, NULL, "Cannot open %s to read",
        dr_filename);
    assert_not_equal_with_message(gain_fptr, NULL, "Cannot open %s to read",
        gain_filename);
    assert_not_equal_with_message(means_fptr, NULL, "Cannot open %s to "
        "read", scene_mean_filename);

    if ((dr_fptr == NULL) || (gain_fptr == NULL) || (means_fptr == NULL))
    {
        if (dr_fptr)
            fclose(dr_fptr);
        if (gain_fptr)
            fclose(gain_fptr);
        if (means_fptr)
            fclose(means_fptr);
        return;
    }

    /* Allocate required data buffers.*/
    blind_stats = calloc(number_detectors,
        sizeof(IAS_RPS_HISTOGRAM_INTERMEDIATE_RESULTS));
    assert_not_equal_with_message(blind_stats, NULL, "Cannot allocate "
        "histogram statistics intermediate results data buffer");
    if (blind_stats == NULL)
    {
        fclose(dr_fptr);
        fclose(gain_fptr);
        fclose(means_fptr);
        return;
    }

    dark_response = calloc(number_detectors, sizeof(*dark_response));
    relative_gains = calloc(number_detectors, sizeof(*relative_gains));

    assert_not_equal_with_message(dark_response, NULL, "Memory allocation "
        "of dark response buffer failed, cannot continue testing");
    assert_not_equal_with_message(relative_gains, NULL, "Memory allocation "
        "of detector gains failed, cannot continue processing");

    if ((dark_response == NULL) || (relative_gains == NULL))
    {
        free(dark_response);
        free(relative_gains);
        free(blind_stats);
        fclose(dr_fptr);
        fclose(gain_fptr);
        fclose(means_fptr);
        return;
    }

    /* Read in the data */
    for (det_index = 0; det_index < number_detectors; det_index++)
    {
        dr_status = fscanf(dr_fptr, "%lf\n", &dark_response[det_index]);
        gain_status = fscanf(gain_fptr, "%lf\n", &relative_gains[det_index]);
        means_status = fscanf(means_fptr, "%lf\n",
            &(blind_stats[det_index].mean[0]));



        if ((dr_status < 0) || (gain_status < 0) || (means_status < 0))
        {
            assert_true_with_message(FALSE, "Gain read for "
                "detector %d failed, cannot continue processing",
                (det_index + 1));
            assert_true_with_message(FALSE,
                "Dark response read for detector %d failed, cannot "
                "continue processing", (det_index + 1));
            assert_true_with_message(FALSE, "Blind means read "
                " for detector %d failed, cannot continue processing",
                (det_index + 1));           
            free(dark_response);
            free(relative_gains);
            fclose(dr_fptr);
            fclose(gain_fptr);
            fclose(means_fptr);
            return;
        }
    }

    /* Close the files, we're done with them */
    fclose(dr_fptr);
    fclose(gain_fptr);
    fclose(means_fptr);

    /* Generate a pixel mask with no artifacts in it */
    pmask = ias_pm_create(blind_band_number, sca_number,
        number_detectors, number_rows);
    assert_not_equal_with_message(pmask, NULL, "Pixel mask creation "
        "failed, cannot continue processing");
    if (pmask == NULL)
    {
        free(dark_response);
        free(relative_gains);
        free(blind_stats);
        return;
    }

    /* Call the algorithm */
    is_blind_band = (band_number == blind_band_number);
    status = ias_rps_tirs_dark_response_determination(blind_stats,
        dark_response, &sca_dark_response_difference,
        abs_gain, relative_gains, is_blind_band, number_detectors, pmask);
    assert_equal_with_message(status, SUCCESS, "Should have calculated "
        "blind band adjusted dark response");
    if (status != SUCCESS)
    {
        free(dark_response);
        free(relative_gains);
        free(blind_stats);
        return;
    }

    printf("Implementation estimated SCA dark response difference:  %lf\n",
        sca_dark_response_difference);
    printf("Prototype estimated SCA dark response difference:       %lf\n",
        prototype_dark_response_difference);

    /* Free the data buffers */
    free(blind_stats);
    free(relative_gains);
    ias_pm_destroy(pmask);

    /* Write the results to a group file */
    printf("Writing band %d results to group file...\n", blind_band_number);

    status = ias_rps_tirs_dark_response_write_to_group_file(
        dark_response, is_blind_band, band_number,
        sca_number, number_detectors);
    assert_equal_with_message(status, SUCCESS, "Error writing to group "
        "file");
    if (status != SUCCESS)
    {
        free(dark_response);
        return;
    }

    /* Free the dark response data buffer */
    free(dark_response);

    /* Check to make sure the test file is there */
    sprintf(test_adr_group_filename,
        "cpf_dark_response_group_B%02d_SCA%02d.odl",
        blind_band_number, sca_number);

    status = stat(test_adr_group_filename,&buff);
    assert_equal_with_message(status, 0, "Test group file %s not found",
        test_adr_group_filename);
    if (status != 0)
    {
        return;
    }

    sprintf(ref_adr_group_filename, "%s/%s/cpf_adrgroup_B%02d_SCA%02d.odl",
         test_data_path, test_data_directory, blind_band_number,
         sca_number);

    status = check_group_files(test_adr_group_filename,
        ref_adr_group_filename, 1e-4);
    assert_equal_with_message(status, SUCCESS, "Cannot complete check");

    /* Done */
}



Ensure test_calculate_adjusted_response_image()
{
    IAS_RPS_HISTOGRAM_INTERMEDIATE_RESULTS *blind_stats = NULL;

    char dr_filename[PATH_MAX] = "";       /* Test baseline dark response
                                              file name (test data in .txt
                                              file) */
    char gain_filename[PATH_MAX] = "";     /* Test detector gains file
                                              name (test data in .txt
                                              file) */

    char ref_adr_group_filename[PATH_MAX] = "";   /* Name of reference
                                                     ODL-formatted group
                                                     file */
    char test_adr_group_filename[PATH_MAX] = "";  /* Name of test
                                                     ODL-formatted group
                                                     file */

    /* The structure of these ASCII data files consists of one value per
       detector */
    FILE *dr_fptr = NULL;                      /* Baseline dark response
                                                  data file */
    FILE *gain_fptr = NULL;                    /* Absolute detector gains
                                                  data file */

    IAS_PIXEL_MASK *pmask = NULL;              /* Pixel mask */

    int is_blind_band = 0;                     /* Blind band flag */
    int blind_band_number = 15;                /* Current TIRS blind
                                                  band number */
    int band_number = 10;
    int sca_number = 1;
    int number_detectors = 640;
    int number_rows = 2;
    int det_index = 0;
    int status = SUCCESS;
    int dr_status = SUCCESS;
    int gain_status = SUCCESS;

    double *dark_response = NULL;              /* Baseline dark response */
    double *relative_gains = NULL;             /* Relative detector gains */
    double sca_dark_response_difference = 1.29237;       
                                               /* Prototype mean dark response
                                                  difference for SCA 1 */
    double abs_gain = 1.0;

    struct stat buff;


    printf("Testing calculation of adjusted band %02d dark response...\n",
        band_number);

    /* Open the data files for reading */
    sprintf(dr_filename, "%s/%s/dr_B%02d_SCA%02d.txt", test_data_path,
        test_data_directory, band_number, sca_number);
    sprintf(gain_filename, "%s/%s/rg_B%02d_SCA%02d.txt", test_data_path,
        test_data_directory, band_number, sca_number);

    dr_fptr = fopen(dr_filename, "r");
    gain_fptr = fopen(gain_filename, "r");

    assert_not_equal_with_message(dr_fptr, NULL, "Cannot open %s to read",
        dr_filename);
    assert_not_equal_with_message(gain_fptr, NULL, "Cannot open %s to read",
        gain_filename);

    if ((dr_fptr == NULL) || (gain_fptr == NULL))
    {
        if (dr_fptr)
            fclose(dr_fptr);
        if (gain_fptr)
            fclose(gain_fptr);
        return;
    }

    /* Allocate required data buffers.*/
    dark_response = calloc(number_detectors, sizeof(*dark_response));
    relative_gains = calloc(number_detectors, sizeof(*relative_gains));

    assert_not_equal_with_message(dark_response, NULL, "Memory allocation "
        "of dark response buffer failed, cannot continue testing");
    assert_not_equal_with_message(relative_gains, NULL, "Memory allocation "
        "of detector gains failed, cannot continue processing");

    if ((dark_response == NULL) || (relative_gains == NULL))
    {
        free(dark_response);
        free(relative_gains);
        fclose(dr_fptr);
        fclose(gain_fptr);
        return;
    }

    /* Read in the data */
    for (det_index = 0; det_index < number_detectors; det_index++)
    {
        dr_status = fscanf(dr_fptr, "%lf\n", &dark_response[det_index]);
        gain_status = fscanf(gain_fptr, "%lf\n", &relative_gains[det_index]);
        if ((dr_status < 0) || (gain_status < 0))
        {
            assert_true_with_message(FALSE, "Gain read for "
                "detector %d failed, cannot continue processing",
                (det_index + 1));
            assert_true_with_message(FALSE, "Dark response read "
                "for detector %d failed, cannot continue processing",
                (det_index + 1));
            free(dark_response);
            free(relative_gains);
            fclose(dr_fptr);
            fclose(gain_fptr);
            return;
        }
    }

    /* Close the files, we're done with them */
    fclose(dr_fptr);
    fclose(gain_fptr);

    /* Generate a pixel mask with no artifacts in it */
    pmask = ias_pm_create(blind_band_number, sca_number,
        number_detectors, number_rows);
    assert_not_equal_with_message(pmask, NULL, "Pixel mask creation "
        "failed cannot continue processing");
    if (pmask == NULL)
    {
        free(dark_response);
        free(relative_gains);
        return;
    }

    /* Calculate the adjusted dark response */
    is_blind_band = (band_number == blind_band_number);
    status = ias_rps_tirs_dark_response_determination(blind_stats,
        dark_response, &sca_dark_response_difference,
        abs_gain, relative_gains, is_blind_band, number_detectors,
        pmask);
    assert_equal_with_message(status, SUCCESS, "Should have calculated "
        "image band adjusted dark response");
    if (status != SUCCESS)
    {
        free(dark_response);
        free(relative_gains);
        return;
    }

    /* Free uneeded data buffers */
    free(relative_gains);
    ias_pm_destroy(pmask);

    /* Write the results to a group file */
    printf("Writing band %d results to group file...\n", band_number);

    status = ias_rps_tirs_dark_response_write_to_group_file(
        dark_response, is_blind_band, band_number,
        sca_number, number_detectors);
    assert_equal_with_message(status, SUCCESS, "Error writing to group "
        "file");
    if (status != SUCCESS)
    {
        free(dark_response);
        return;
    }

    /* Free the dark response data buffer */
    free(dark_response);

    /* Check to make sure the test file is there */
    sprintf(test_adr_group_filename,
        "cpf_dark_response_group_B%02d_SCA%02d.odl",
        band_number, sca_number);

    status = stat(test_adr_group_filename,&buff);
    assert_equal_with_message(status, 0, "Test group file %s not found",
        test_adr_group_filename);
    if (status != 0)
    {
        return;
    }

    /* Open the "reference" and test group files */
    sprintf(ref_adr_group_filename, "%s/%s/cpf_adrgroup_B%02d_SCA%02d.odl",
         test_data_path, test_data_directory, band_number, sca_number);

    status = check_group_files(test_adr_group_filename,
        ref_adr_group_filename, 1e-3);
    assert_equal_with_message(status, SUCCESS, "Cannot complete check");

    /* Done */
}




int main(int argc, char *argv[])
{
    test_data_path = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (test_data_path == NULL)
    {
        printf("Cannot locate IAS test data directory\n");
        exit(100);
    }

    TestSuite *suite = create_test_suite();

    add_test(suite, test_calculate_adjusted_response_blind);
    add_test(suite, test_calculate_adjusted_response_image);

    if (argc > 1)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}
