/**************************************************************************
 NAME:      radiometric_library_dropped_frame_test

 PURPOSE:   Implement CGreen unit testing for the RPS library function
            ias_rps_dropped_frame_characterization

 RETURNS:   CGreen unit status

 ROUTINES:  cleanup                               (internal routine)
            compare_files_after_first_header_line (internal routine)
            test_read_frame_header_status_bit
            test_dropped_frame_characterization
***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "ias_const.h"
#include "ias_l0r.h"
#include "ias_miscellaneous.h"
#include "ias_pixel_mask.h"
#include "../ias_radiometry.h"
#include "ias_satellite_attributes.h"
#include "ias_types.h"
#include "cgreen.h"


static char *test_data_path = NULL;



/********************* INTERNAL ROUTINES **********************************/
static void cleanup(
    IAS_L0R_OLI_FRAME_HEADER *fh,
    IAS_PIXEL_MASK *pm)
{
    if (pm != NULL)
        ias_pm_destroy(pm);

    free(fh);
}   /* END -- cleanup */



/******************** Main dropped frame characterization test ************/
Ensure test_dropped_frame_characterization()
{
    IAS_PIXEL_MASK_SPAN span;
    char full_l0r_path[PATH_MAX];
    char full_l0r_name[PATH_MAX];
    char data_dir[] = "ias_lib/rps/dropped_frame_characterization";
    char l0r_name[] = "LO80390360362003148EDC00";
    L0RIO *l0r_image = NULL;
    IAS_L0R_OLI_FRAME_HEADER *fh = NULL;
    IAS_PIXEL_MASK *pm = NULL;
    IAS_PIXEL_MASK_ITERATOR *it = NULL;
    int span_start_index[5] = {1, 4, 7, 21, 6044};
    int det_index;
    int dropped_frame_error_count = 0;         /* Dropped frame
                                                  identification error
                                                  counter */
    int status;
    int band_number = 1;
    int sca_number = 1;
    int num_detectors = 494;
    int num_frames_expected = 6050;
    int num_frames;


    printf("Testing dropped frame characterization...\n");

    /* Open the test L0R image */
    sprintf(full_l0r_path, "%s/%s", test_data_path, data_dir);
    sprintf(full_l0r_name, "%s/%s", full_l0r_path, l0r_name);
    l0r_image = ias_l0r_open(l0r_name, full_l0r_path, IAS_READ);
    if (l0r_image == NULL)
    {
        assert_true_with_message(FALSE, "Error opening image file %s",
            l0r_name);
        return;
    }

    /* Open the L0R header */
    status = ias_l0r_open_header(l0r_image, IAS_READ);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error opening L0R frame header "
            "data to read");
        ias_l0r_close(l0r_image);
        return;
    }

    /* Get the number of frames */
    status = ias_l0r_get_oli_frame_headers_records_count(l0r_image,
        &num_frames);
    if ((num_frames == 0) || (num_frames != num_frames_expected))
        status = ERROR;
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error finding number of OLI frame "
            "headers");
        ias_l0r_close_header(l0r_image);
        ias_l0r_close(l0r_image);
        return;
    }

    /* Allocate the frame header and detector offset data buffers.
       For simplicity, assume there is no detector offset */
    fh = calloc(num_frames, sizeof(IAS_L0R_OLI_FRAME_HEADER));
    if (fh == NULL)
    {
        assert_true_with_message(FALSE, "Error allocating OLI frame header "
            "data buffer of %d bytes",
            num_frames * sizeof(IAS_L0R_OLI_FRAME_HEADER));
        ias_l0r_close_header(l0r_image);
        ias_l0r_close(l0r_image);
        return;
    }

    /* Get the frame headers */
    status = ias_l0r_get_oli_frame_headers(l0r_image, 0, num_frames, fh);
    ias_l0r_close_header(l0r_image);
    ias_l0r_close(l0r_image);
    assert_equal_with_message(status, SUCCESS, "Error retrieving OLI frame "
        "headers for scene %s", l0r_name);
    if (status != SUCCESS)
    {
        free(fh);
        return;
    }

    /* Allocate an empty pixel mask */
    pm = ias_pm_create(band_number, sca_number, num_detectors, num_frames);
    if (pm == NULL)
    {
        assert_true_with_message(FALSE, "Error allocating empty pixel "
            "mask");
        free(fh);
        return;
    }

    /* Call the characterization */
    status = ias_rps_dropped_frame_characterization((const void *)fh,
        IAS_OLI, band_number, num_detectors, num_frames, pm);
    assert_equal_with_message(status, SUCCESS, "Error completing dropped "
        "frame characterization");
    if (status != SUCCESS)
    {
        cleanup(fh, pm);
        return;
    }

    /* Check the pixel mask */
    it = ias_pm_get_iterator(pm, IAS_PM_EXCLUDE, 0x00);
    assert_true(it != NULL);
    if (it == NULL)
    {
        cleanup(fh, pm);
        return;
    }

    for (det_index = 0; det_index < num_detectors;  det_index++)
    {
        /* Check the first span--1 frame, starting at frame 2 */
        status = ias_pm_get_next_span(it, &span);
        if (status == 1)
        {
            if ((span.starting_pixel_index != span_start_index[0])
                || (span.length_of_span != 1)
                || (span.detector_index != det_index)
                || (span.pixel_mask != IAS_RPS_DROPLINE))
            {
                dropped_frame_error_count++;
            }
        }
        else
            dropped_frame_error_count++;

        /* Check the second span--1 frame, starting at frame 5 */
        status = ias_pm_get_next_span(it, &span);
        if (status == 1)
        {
            if ((span.starting_pixel_index != span_start_index[1])
                || (span.length_of_span != 1)
                || (span.detector_index != det_index)
                || (span.pixel_mask != IAS_RPS_DROPLINE))
            {
                dropped_frame_error_count++;
            }
        }
        else
            dropped_frame_error_count++;

        /* Check the third span--5 frames, starting at frame 8 */
        status = ias_pm_get_next_span(it, &span);
        if (status == 1)
        {
            if ((span.starting_pixel_index != span_start_index[2])
                || (span.length_of_span != 5)
                || (span.detector_index != det_index)
                || (span.pixel_mask != IAS_RPS_DROPLINE))
            {
                dropped_frame_error_count++;
            }
        }
        else
            dropped_frame_error_count++;

        /* Check the fourth span-- 8 frames, starting at frame 22 */
        status = ias_pm_get_next_span(it, &span);
        if (status == 1)
        {
            if ((span.starting_pixel_index != span_start_index[3])
                || (span.length_of_span != 8)
                || (span.detector_index != det_index)
                || (span.pixel_mask != IAS_RPS_DROPLINE))
            {
                dropped_frame_error_count++;
            }
        }
        else
            dropped_frame_error_count++;

        /* Check the final span--6 frames, starting at frame 6044.  Skip
           detector index 1 since that one has an offset in that pushes the
           span outside the image */
        status = ias_pm_get_next_span(it, &span);
        if (status == 1)
        {
            int expected_span = 6;

            if ((span.starting_pixel_index != span_start_index[4])
                || (span.length_of_span != expected_span)
                || (span.detector_index != det_index)
                || (span.pixel_mask != IAS_RPS_DROPLINE))
            {
                dropped_frame_error_count++;
            }
        }
        else
            dropped_frame_error_count++;
    }

    /* Make sure no errors were found */
    assert_equal(dropped_frame_error_count , 0);

    /* Cleanup */
    ias_pm_destroy_iterator(it);
    cleanup(fh, pm);
}   /* END -- test_dropped_frame_characterization */

/******** Test Pan band to make sure it handles 2 lines per frame  *********/
/* Note: This is essentially a copy of the previous test with adjustments
   made to test the different number of lines per frame present in the PAN
   band. */
Ensure test_pan_band_dropped_frame_characterization()
{
    IAS_PIXEL_MASK_SPAN span;
    char full_l0r_path[PATH_MAX];
    char full_l0r_name[PATH_MAX];
    char data_dir[] = "ias_lib/rps/dropped_frame_characterization";
    char l0r_name[] = "LO80390360362003148EDC00";
    L0RIO *l0r_image = NULL;
    IAS_L0R_OLI_FRAME_HEADER *fh = NULL;
    IAS_PIXEL_MASK *pm = NULL;
    IAS_PIXEL_MASK_ITERATOR *it = NULL;
    int det_index;
    int dropped_frame_error_count = 0;         /* Dropped frame
                                                  identification error
                                                  counter */
    int status;
    int band_number = 8;
    int sca_number = 1;
    int num_detectors = 988;
    int num_frames_expected = 6050;
    int num_frames;
    int lines_per_frame = 2;
    int span_start_index[5] = {1 * lines_per_frame, 4 * lines_per_frame,
        7 * lines_per_frame, 21 * lines_per_frame, 6044 * lines_per_frame};
    int num_lines;


    printf("Testing PAN band dropped frame characterization...\n");

    /* Open the test L0R image */
    sprintf(full_l0r_path, "%s/%s", test_data_path, data_dir);
    sprintf(full_l0r_name, "%s/%s", full_l0r_path, l0r_name);
    l0r_image = ias_l0r_open(l0r_name, full_l0r_path, IAS_READ);
    if (l0r_image == NULL)
    {
        assert_true_with_message(FALSE, "Error opening image file %s",
            l0r_name);
        return;
    }

    /* Open the L0R header */
    status = ias_l0r_open_header(l0r_image, IAS_READ);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error opening L0R frame header "
            "data to read");
        ias_l0r_close(l0r_image);
        return;
    }

    /* Get the number of frames */
    status = ias_l0r_get_oli_frame_headers_records_count(l0r_image,
        &num_frames);
    if ((num_frames == 0) || (num_frames != num_frames_expected))
        status = ERROR;
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error finding number of OLI frame "
            "headers");
        ias_l0r_close_header(l0r_image);
        ias_l0r_close(l0r_image);
        return;
    }

    /* Allocate the frame header and detector offset data buffers.
       For simplicity, assume there is no detector offset */
    fh = calloc(num_frames, sizeof(IAS_L0R_OLI_FRAME_HEADER));
    if (fh == NULL)
    {
        assert_true_with_message(FALSE, "Error allocating OLI frame header "
            "data buffer of %d bytes",
            num_frames * sizeof(IAS_L0R_OLI_FRAME_HEADER));
        ias_l0r_close_header(l0r_image);
        ias_l0r_close(l0r_image);
        return;
    }

    /* Get the frame headers */
    status = ias_l0r_get_oli_frame_headers(l0r_image, 0, num_frames, fh);
    ias_l0r_close_header(l0r_image);
    ias_l0r_close(l0r_image);
    assert_equal_with_message(status, SUCCESS, "Error retrieving OLI frame "
        "headers for scene %s", l0r_name);
    if (status != SUCCESS)
    {
        free(fh);
        return;
    }

    /* Convert frames to lines */
    num_lines = num_frames * lines_per_frame;

    /* Allocate an empty pixel mask */
    pm = ias_pm_create(band_number, sca_number, num_detectors, num_lines);
    if (pm == NULL)
    {
        assert_true_with_message(FALSE, "Error allocating empty pixel "
            "mask");
        free(fh);
        return;
    }

    /* Call the characterization */
    status = ias_rps_dropped_frame_characterization((const void *)fh,
        IAS_OLI, band_number, num_detectors, num_frames, pm);
    assert_equal_with_message(status, SUCCESS, "Error completing dropped "
        "frame characterization");
    if (status != SUCCESS)
    {
        cleanup(fh, pm);
        return;
    }

    /* Check the pixel mask */
    it = ias_pm_get_iterator(pm, IAS_PM_EXCLUDE, 0x00);
    assert_true(it != NULL);
    if (it == NULL)
    {
        cleanup(fh, pm);
        return;
    }
    for (det_index = 0; det_index < num_detectors;  det_index++)
    {
        /* Check the first span--1 frame, starting at frame 2 */
        status = ias_pm_get_next_span(it, &span);
        if (status == 1)
        {
            if ((span.starting_pixel_index != span_start_index[0])
                || (span.length_of_span != 1 * lines_per_frame)
                || (span.detector_index != det_index)
                || (span.pixel_mask != IAS_RPS_DROPLINE))
            {
                dropped_frame_error_count++;
            }
        }
        else
            dropped_frame_error_count++;

        /* Check the second span--1 frame, starting at frame 5 */
        status = ias_pm_get_next_span(it, &span);
        if (status == 1)
        {
            if ((span.starting_pixel_index != span_start_index[1])
                || (span.length_of_span != 1 * lines_per_frame)
                || (span.detector_index != det_index)
                || (span.pixel_mask != IAS_RPS_DROPLINE))
            {
                dropped_frame_error_count++;
            }
        }
        else
            dropped_frame_error_count++;

        /* Check the third span--5 frames, starting at frame 8 */
        status = ias_pm_get_next_span(it, &span);
        if (status == 1)
        {
            if ((span.starting_pixel_index != span_start_index[2])
                || (span.length_of_span != 5 * lines_per_frame)
                || (span.detector_index != det_index)
                || (span.pixel_mask != IAS_RPS_DROPLINE))
            {
                dropped_frame_error_count++;
            }
        }
        else
            dropped_frame_error_count++;

        /* Check the fourth span-- 8 frames, starting at frame 22 */
        status = ias_pm_get_next_span(it, &span);
        if (status == 1)
        {
            if ((span.starting_pixel_index != span_start_index[3])
                || (span.length_of_span != 8 * lines_per_frame)
                || (span.detector_index != det_index)
                || (span.pixel_mask != IAS_RPS_DROPLINE))
            {
                dropped_frame_error_count++;
            }
        }
        else
            dropped_frame_error_count++;

        /* Check the final span--6 frames, starting at frame 6044. */
        status = ias_pm_get_next_span(it, &span);
        if (status == 1)
        {
            int expected_span = 6 * lines_per_frame;

            if ((span.starting_pixel_index != span_start_index[4])
                || (span.length_of_span != expected_span)
                || (span.detector_index != det_index)
                || (span.pixel_mask != IAS_RPS_DROPLINE))
            {
                dropped_frame_error_count++;
            }
        }
        else
            dropped_frame_error_count++;

    }

    /* Make sure no errors were found */
    assert_equal(dropped_frame_error_count , 0);

    /* Cleanup */
    ias_pm_destroy_iterator(it);
    cleanup(fh, pm);
}   /* END -- test_dropped_frame_characterization */


int main(int argc, char *argv[])
{
    int status;


    /* Get the test data directory */
    test_data_path = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (test_data_path == NULL)
    {
        printf("Cannot locate IAS test data directory\n");
        return 1;
    }

    /* Initialize the satellite attributes library */
    status = ias_sat_attr_initialize(IAS_L8);
    if (status != SUCCESS)
    {
        printf("Cannot initialize Satellite Attributes Library\n");
        return 1;
    }

    /* Build the test suite */
    TestSuite *suite = create_test_suite();
    add_test(suite, test_dropped_frame_characterization);
    add_test(suite, test_pan_band_dropped_frame_characterization);

    /* Run the test(s) */
    if (argc > 1)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}
