/**************************************************************************
 NAME:      frame_averages_test

 PURPOSE:   Performs CGREEN unit testing on the frame average calculation

 RETURNS:   CGREEN unit status

 ROUTINES:  test_database_io
            test_averages_calculation
            main

 NOTES:  
****************************************************************************/
#include <math.h>
#include <string.h>

#include "cgreen.h"
#include "ias_db.h"
#include "ias_db_get_connect_info.h"
#include "ias_db_query.h"
#include "ias_logging.h"
#include "ias_radiometry.h"
#include "ias_rps_frame_averages.h"
#include "ias_pixel_mask.h"
#include "ias_satellite_attributes.h"

static struct ias_db_connection *db_open()
{
    struct ias_db_connection *db = NULL;
    char *db_name = NULL;
    char *user_name = NULL;
    char *passwd = NULL;
    char *host = NULL;
    int status;

    /* Initialize the satellite attributes library */
    ias_sat_attr_initialize(IAS_L8);

    /* Initialize the database library */
    status = ias_db_initialize_database_lib();
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error initializing database "
           "library");
        return NULL;
    }

    /* Get the connection information */
    status = ias_db_get_connect_info(&db_name, &user_name, &passwd, &host);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error obtaining DB connection "
            "information");
        ias_db_close_database_lib();
        return NULL;
    }

    /* Try to connect */
    db = ias_db_connect_to_database(db_name, user_name, passwd, host);
    free(db_name);
    free(user_name);
    free(passwd);
    free(host);
    if (db == NULL)
    {
        assert_true_with_message(FALSE, "Error connecting to DB %s as user %s"
           "at host %s", db_name, user_name, host);
        return NULL;
    }
    return db;
}

static void db_close(struct ias_db_connection *db)
{
    ias_db_close_connection(db);
    ias_db_close_database_lib();
}

/*
* NAME: test_database_io
*
* PURPOSE: Test the ability to store the frame averages in the database.
* 
*/
Ensure test_database_io()
{
    const int number_of_lines = 12;
    const int band_number = 4;
    const int sca_number = 5;
    char work_order_id[IAS_WORKORDER_ID_SIZE];
    int line = 0;
    int order_number = 0;
    int table_size = 0; 
    int status = ERROR;

    IAS_RPS_DB_FRAME_AVERAGES *averages = NULL;
    IAS_RPS_DB_FRAME_AVERAGES *read_averages = NULL;

    averages = ias_rps_db_allocate_frame_averages(number_of_lines);
    assert_not_equal(averages, NULL);

    /* assign values for the averages in a unique fashion for later comparison*/
    for (line = 0; line < number_of_lines; line++)
    {
        averages->line_number[line] = line;
        averages->line_average[line] = line;
    }

    averages->dbase = db_open();
    if (averages->dbase == NULL)
    {
        ias_rps_db_free_frame_averages(&averages);
        assert_equal(averages, NULL);
        return;
    } 

    /* Obtain the largest IS work order ID currently used
        SELECT MAX(WORK_ORDER_ID)
        FROM RADIOMETRIC_WORK_ORDER_COMMON
        WHERE WORK_ORDER_ID LIKE 'IS%'; */
    IAS_DB_QUERY_TABLE qtable[] =
    {
       {"MAX(WORK_ORDER_ID)", work_order_id, sizeof(work_order_id), 
            IAS_QUERY_STRING, 1}
    };
    table_size  = sizeof(qtable)/sizeof(IAS_DB_QUERY_TABLE);
    status = ias_db_query_using_table(averages->dbase,
        "RADIOMETRIC_WORK_ORDER_COMMON", qtable, table_size, 
        "WORK_ORDER_ID LIKE 'IS%'", 1);

    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Unable to obtain maximum work order ID");
        db_close(averages->dbase);
        assert_true(0);
        return;
    }

    /* convert the work order id to a numerical value to do an increment */
    if (strlen(work_order_id) > 0)
    { 
        order_number = atoi(&work_order_id[2]);
        order_number++;
    }
    
    status = snprintf(work_order_id, sizeof(work_order_id), "IS%04d", 
        order_number);
    if (status < 0 || status >= sizeof(work_order_id))
    {
        IAS_LOG_ERROR("Unable to create the work order ID");
        db_close(averages->dbase);
        assert_true(0);
        return;
    }
    
    /* obtain a common ID to use in the frame averages table */
    status = ias_misc_get_char_id(
        averages->dbase, "radiometric_work_order_common", 
        "radiometric_wo_common_id", work_order_id, averages->rad_common_id);

    if (status == ERROR)
    {
        IAS_LOG_ERROR("Unable to obtain a radiometric_wo_common_id");
        assert_not_equal(status, ERROR);
        db_close(averages->dbase);
        return;
    }

    averages->band_number = band_number;
    averages->sca = sca_number; 

    assert_equal(SUCCESS, ias_rps_db_store_frame_averages(averages));

    read_averages = ias_rps_db_get_frame_averages(
        averages->dbase, averages->rad_common_id, band_number, sca_number);

    assert_not_equal(read_averages, NULL);
 
    /* verify the number of lines read is as expected */ 
    assert_equal(read_averages->num_lines, number_of_lines);
  
    /* Verify the averages read match those written */ 
    if (read_averages != NULL)
    {
        for (line = 0; line < number_of_lines; line++)
        {
            if (averages->line_average[line] !=
                read_averages->line_average[line])
            {
                IAS_LOG_ERROR("Results do not match at line %d."
                    " Wrote %lf read %lf", line, averages->line_average[line], 
                    read_averages->line_average[line]); 
                assert_equal(averages->line_average[line], 
                    read_averages->line_average[line]);
            }
        }

    }

    db_close(averages->dbase);

    /* cleanup data from the test */
    ias_rps_db_free_frame_averages(&averages);
    assert_equal(averages, NULL);
}

/*
*NAME: test_averages_calculation
*
*PURPOSE: Test the ability to calculate the cross-track averages on a simple set
*   set of test data
*
*NOTE: Values obtained from the Verification Methods section of the 
*   Hist_Stats_Char ADD
*/
Ensure test_averages_calculation()
{
    const int band_number = 1;
    const int sca_number = 1;
    const int detectors_per_line = 8;
    const int number_of_lines = 12;
    const int image_size = number_of_lines * detectors_per_line;
    const uint16_t pixels[] = 
    {
        601, 305, 413, 2, 100,  89, 200, 354,
        601, 363, 466, 2, 100, 198, 200, 437,
        601, 281, 283, 2, 100, 375, 200, 999,
        601, 294, 999, 2, 900, 607, 200, 611,
        4,     4,   4, 2, 4,     4,   4,   4,
        601, 888, 279, 2, 100, 980, 200, 298,
        599, 278, 888, 2, 100, 980, 400, 504,
        599, 396, 493, 2, 100, 835, 400, 361,
        599, 999, 401, 2, 100, 607, 400, 262,
        599, 265, 335, 2, 888, 375, 400, 432,
        4,     4,   4, 2, 4,     4,   4,   4,
        599, 980, 423, 2, 100,  89, 400, 471
    };

    const double results[] = 
    {
        294.57142857, 337.85714286, 306.66666667, 
        535.50000000, 0.0, 478.00, 
        535.571429, 454.857143,  481.142857,    
        470.571429, 0.0, 437.428571
    };

    int i = 0;

    IAS_RPS_DB_FRAME_AVERAGES *averages = NULL;
    IAS_PIXEL_MASK *pixel_mask = NULL;

    averages = ias_rps_db_allocate_frame_averages(number_of_lines);
    assert_not_equal(averages, NULL);

    /*create and populate the pixel mask*/
    pixel_mask = ias_pm_create(band_number, sca_number, detectors_per_line, 
        image_size);
    assert_not_equal(pixel_mask, NULL);

    /* Set the pixel mask for the artifacts in the test data */ 
    for (i = 0; i < detectors_per_line; i++)
    {
        /* frames at index 4 and 10 are populated with fill */
        assert_equal(SUCCESS, ias_pm_add_pixels(pixel_mask, 
            i, 4, 1, IAS_RPS_DROPLINE));
        assert_equal(SUCCESS, ias_pm_add_pixels(pixel_mask,
            i, 10, 1, IAS_RPS_DROPLINE));
    }

    /* The detector at index 3 is dead */
    assert_equal(SUCCESS, ias_pm_add_pixels(pixel_mask,
            3, 0, number_of_lines, IAS_RPS_DROPLINE));

    /* mark some impulse noise locations */ 
    ias_pm_add_pixels( pixel_mask, 7, 2, 1, IAS_RPS_IMPNOISE );
    ias_pm_add_pixels( pixel_mask, 2, 3, 1, IAS_RPS_IMPNOISE );

    /*obtain the cross-track averages*/
    ias_rps_calculate_frame_averages(pixels, IAS_UI2, band_number, 
        sca_number, detectors_per_line, 
        1, /*start at line 1*/ 
        number_of_lines, 
        0, 0, /*don't skip any lines at the start or end*/
        pixel_mask, averages);

    assert_equal(averages->num_lines, number_of_lines);

    for (i = 0; i < number_of_lines; i++)
    {
        assert_double_equal(results[i], averages->line_average[i]);
    } 

    /*free resources created in the test*/
    ias_pm_destroy(pixel_mask);
    pixel_mask = NULL; 

    ias_rps_db_free_frame_averages(&averages);
    assert_equal(averages, NULL);
}

int main(int argc, char *argv[])
{
    /* Construct the test suite */
    TestSuite *suite = create_test_suite();
    add_test(suite, test_averages_calculation);
   
    /* Use the env variable to determine if the database related tests should 
    be run */ 
    if (getenv("ENABLE_DB_TESTS") == NULL)
    {
        printf("DB Tests will not run, setenv ENABLE_DB_TESTS to enable....\n");
    }
    else
    {
        add_test(suite, test_database_io);
    }

    /* Run the test(s) */
    if (argc > 1)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}

