/**************************************************************************
 NAME:      histogram_statistics_io_test

 PURPOSE:   Performs CGREEN unit testing on the frame average calculation

 RETURNS:   CGREEN unit status

 ROUTINES:  test_database_io
            main

 NOTES:  
****************************************************************************/
#include <math.h>
#include <string.h>

#include "cgreen.h"
#include "ias_db.h"
#include "ias_db_get_connect_info.h"
#include "ias_db_query.h"
#include "ias_logging.h"
#include "ias_radiometry.h"
#include "ias_pixel_mask.h"

/* difference between two double values to consider them equal */
const double double_tolerance = 1E-9;

static struct ias_db_connection *db_open()
{
    struct ias_db_connection *db = NULL;
    char *db_name = NULL;
    char *user_name = NULL;
    char *passwd = NULL;
    char *host = NULL;
    int status;


    /* Initialize the database library */
    status = ias_db_initialize_database_lib();
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error initializing database "
           "library");
        return NULL;
    }

    /* Get the connection information */
    status = ias_db_get_connect_info(&db_name, &user_name, &passwd, &host);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error obtaining DB connection "
            "information");
        ias_db_close_database_lib();
        return NULL;
    }

    /* Try to connect */
    db = ias_db_connect_to_database(db_name, user_name, passwd, host);
    free(db_name);
    free(user_name);
    free(passwd);
    free(host);
    if (db == NULL)
    {
        assert_true_with_message(FALSE, "Error connecting to DB %s as user %s"
           "at host %s", db_name, user_name, host);
        return NULL;
    }
    return db;
}

static void db_close(struct ias_db_connection *db)
{
    if (db != NULL)
    {
        ias_db_close_connection(db);
    }
    ias_db_close_database_lib();
}

/*
* NAME: test_database_io
*
* PURPOSE: Test the ability to store the histogram statistics in the database.
* 
*/
void test_database_io(char *collection_type)
{
    const int number_of_detectors = 12;
    char work_order_id[IAS_WORKORDER_ID_SIZE];
    int band_number = 0;
    char common_wo_id[IAS_CHAR_ID_SIZE];  
    char ingest_processing_level[] = "ING1";
    int detector = 0;
    int existing_records = 0;
    int split_line = 0;
    int pan_flag = 0;
    int order_number = 0;
    int sca_number = 0;
    int status = ERROR;
    int table_size = 0;

    float correlation = 0.0;
    float kurtosis = 0.0;
    float min = 0.0;
    float max = 0.0;
    float mean = 0.0;
    float mean_squared = 0.0;
    float skewness = 0.0;
    float std_dev = 0.0;

    const const IAS_BAND_ATTRIBUTES *band_attributes = NULL;
    const IAS_SATELLITE_ATTRIBUTES *sat_attributes = NULL; 
    struct ias_db_connection *database = NULL;
    IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS *statistics = NULL;
    IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS *stored_statistics = NULL;
    IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS *read_statistics = NULL;

    sat_attributes = ias_sat_attr_get_attributes();
    if (sat_attributes == NULL) 
    {
        assert_not_equal(sat_attributes, NULL);
        return;
    }

    database = db_open();
    if (database == NULL)
    {
          assert_not_equal(database, NULL);
         return;
    }

    /* Obtain the largest IS work order ID currently used
        SELECT COUNT(*), MAX(WORK_ORDER_ID)
        FROM RADIOMETRIC_WORK_ORDER_COMMON
        WHERE WORK_ORDER_ID LIKE 'IS%'; */
    IAS_DB_QUERY_TABLE qtable[] =
    {
        {"COUNT(*)", &existing_records, sizeof(existing_records),
             IAS_QUERY_INT, 1},
        {"MAX(WORK_ORDER_ID)", work_order_id, sizeof(work_order_id),
            IAS_QUERY_STRING, 1}
    };

    table_size  = sizeof(qtable)/sizeof(IAS_DB_QUERY_TABLE);
    status = ias_db_query_using_table(database,
    "RADIOMETRIC_WORK_ORDER_COMMON", qtable, table_size,
    "WORK_ORDER_ID LIKE 'IS%'", 1);

    if (status != SUCCESS)
    {
        IAS_LOG_ERROR("Unable to obtain maximum work order ID");
        db_close(database);
        assert_true(0);
        return;
    }

    /* convert the work order id to a numerical value to do an increment */
    if (existing_records > 0)
    {
        order_number = atoi(&work_order_id[2]);
        order_number++;
    }

    status = snprintf(work_order_id, sizeof(work_order_id), "IS%04d",
        order_number);

    if (status < 0 || status >= sizeof(work_order_id))
    {
        IAS_LOG_ERROR("Unable to create the work order ID");
        db_close(database);
        assert_true(0);
        return;
    }

    IAS_LOG_DEBUG("Using work order string %s", work_order_id);

    /* obtain a common ID to use in the frame statistics table */
    status = ias_misc_get_char_id(database, 
        "radiometric_work_order_common",
        "radiometric_wo_common_id", work_order_id, common_wo_id);

    if (status == ERROR)
    {
        IAS_LOG_ERROR("Unable to obtain the common work order ID");
        db_close(database);
        assert_true(0);
        return;
    }

    /* Test reading and writing over each band of data */
    for (band_number = 1; band_number <= sat_attributes->total_bands; 
        band_number++)
    {
        band_attributes = ias_sat_attr_get_band_attributes(band_number);
        if ( band_attributes->band_type == IAS_PAN_BAND ) pan_flag = 1;
        else pan_flag = 0;

        for (sca_number = 1; sca_number < band_attributes->scas; sca_number++)
        {
            
            statistics = ias_rps_db_allocate_histogram_results(
                number_of_detectors, pan_flag);
            assert_not_equal(statistics, NULL);
            if (statistics == NULL) return;
          
            stored_statistics = ias_rps_db_allocate_histogram_results(
                number_of_detectors, pan_flag);
            assert_not_equal(stored_statistics, NULL);
    
            if (stored_statistics == NULL) 
            {
                ias_rps_db_free_histogram_results(statistics);
                statistics = NULL;
                return;
            }
             
            statistics->band_number = band_number;
            statistics->sca = sca_number;
            statistics->num_detectors = number_of_detectors;
            statistics->is_odd_even_data_split = pan_flag;
            statistics->processing_level 
                = strdup(ingest_processing_level);
            assert_not_equal(statistics->processing_level, NULL);
            if (statistics->processing_level == NULL) return;
        
            stored_statistics->band_number = band_number;
            stored_statistics->sca = sca_number;
            stored_statistics->num_detectors = number_of_detectors;
            stored_statistics->is_odd_even_data_split = pan_flag;
            stored_statistics->processing_level 
                = strdup(ingest_processing_level);
            assert_not_equal(stored_statistics->processing_level, NULL);
            if (stored_statistics->processing_level == NULL) return;
        
            /* assign values for the statistics for later comparison */
            for (split_line = 0; 
                split_line < stored_statistics->is_odd_even_data_split + 1;
                split_line++)
            {
                for (detector = 0; detector < number_of_detectors; detector++)
                {
                    /* Modify the values stored to provide some variability */
                    correlation += 0.1;
                    kurtosis += 0.2;
                    min += 0.3;
                    max += 0.4;
                    mean += 0.35;
                    mean_squared += 0.25;
                    skewness += 0.15;
                    std_dev += 0.05;

                    /* The detectors are stored in the database as 1 based IDs*/
                    statistics->detector[split_line][detector] = detector + 1;
                    statistics->num_values[split_line][detector] = detector;
                    statistics->min[split_line][detector] = min;
                    statistics->max[split_line][detector] = max;
                    statistics->mean[split_line][detector] = mean;
                    statistics->stdev[split_line][detector] = std_dev;
                    statistics->skewness[split_line][detector] = skewness;
                    statistics->kurtosis[split_line][detector] = kurtosis;
                    statistics->correlation[split_line][detector] = correlation;
                    statistics->mean_sq_resp[split_line][detector]=mean_squared;
            
                    /* The detectors are stored in the database as 1 based IDs*/
                    stored_statistics->detector[split_line][detector]= 
                        detector + 1;
                    stored_statistics->num_values[split_line][detector] = 
                        detector;
                    stored_statistics->min[split_line][detector] = min;
                    stored_statistics->max[split_line][detector] = max;
                    stored_statistics->mean[split_line][detector] = mean;
                    stored_statistics->stdev[split_line][detector] = std_dev;
                    stored_statistics->skewness[split_line][detector]=skewness;
                    stored_statistics->kurtosis[split_line][detector]=kurtosis;
                    stored_statistics->correlation[split_line][detector] = 
                        correlation;
                    stored_statistics->mean_sq_resp[split_line][detector] = 
                        mean_squared;
                }
            }
        
            statistics->dbase = database;
            stored_statistics->dbase = database;
            strncpy(statistics->rad_common_id, common_wo_id, 
                sizeof(statistics->rad_common_id));
            statistics->rad_common_id[sizeof(statistics->rad_common_id) - 1] 
                = '\0';
             
            strcpy(statistics->collection_type,collection_type);
            if (ias_rps_store_histogram_stats_in_database(statistics) !=
                SUCCESS)
            {
                IAS_LOG_ERROR("Unable to store statistics for band %d, sca %d, "
                    "work order %s", band_number, sca_number, 
                    statistics->rad_common_id); 
                assert_true(0);
            }
        
            /* statistics are freed by the store routine */
            statistics = NULL;
        
            read_statistics = ias_rps_db_get_histogram_statistics(database,
                common_wo_id, band_number, sca_number, collection_type);
            if (read_statistics == NULL)
            {
                assert_true(0);
            }
         
            /* Verify the statistics read match those written */ 
            if (stored_statistics != NULL && read_statistics != NULL)
            {
                for (split_line = 0; 
                    split_line < stored_statistics->is_odd_even_data_split + 1;
                    split_line++)
                {
                    for (detector = 0; detector < number_of_detectors; 
                        detector++)
                    {
                        if (stored_statistics->detector[split_line][detector] !=
                            read_statistics->detector[split_line][detector])
                        {
                            IAS_LOG_ERROR("Results do not match at detector %d."
                                " Wrote %d read %d", detector, 
                                stored_statistics->\
                                    detector[split_line][detector],
                                read_statistics->\
                                    detector[split_line][detector]); 
                            assert_equal(
                                stored_statistics->\
                                    detector[split_line][detector], 
                                read_statistics->\
                                    detector[split_line][detector]);
                        }
            
                        if (stored_statistics->num_values[split_line][detector]
                            !=read_statistics->num_values[split_line][detector])
                        {
                            IAS_LOG_ERROR("Results do not match at detector %d."
                                " Wrote %d read %d", detector, 
                            stored_statistics->num_values[split_line][detector],
                            read_statistics->num_values[split_line][detector]);
                            assert_true(0);
                        }
            
                        if (fabs(stored_statistics->min[split_line][detector] - 
                            read_statistics->min[split_line][detector]) > 
                            double_tolerance)
                        {
                            IAS_LOG_ERROR("Results do not match at detector %d."
                                " Wrote %lf read %lf", detector,
                                stored_statistics->min[split_line][detector],
                                read_statistics->min[split_line][detector]);
                            assert_true(0);
                        }
                     
                        if(fabs(stored_statistics->max[split_line][detector] -
                            read_statistics->max[split_line][detector]) >
                            double_tolerance)
                        {
                            IAS_LOG_ERROR("Results do not match at detector %d."
                                " Wrote %lf read %lf", detector,
                                stored_statistics->max[split_line][detector],
                                read_statistics->max[split_line][detector]);
                            assert_true(0);
                        }
            
                        if(fabs(stored_statistics->mean[split_line][detector] -
                            read_statistics->mean[split_line][detector]) >
                            double_tolerance)
                        {
                            IAS_LOG_ERROR("Results do not match at detector %d."
                                " Wrote %lf read %lf", detector,
                            stored_statistics->mean[split_line][detector],
                            read_statistics->mean[split_line][detector]);
                            assert_true(0);
                        }
                         
                        if(fabs(stored_statistics->stdev[split_line][detector]
                            - read_statistics->stdev[split_line][detector]) >
                                double_tolerance)
                        {
                            IAS_LOG_ERROR("Results do not match at detector %d."
                                " Wrote %lf read %lf", detector,
                                stored_statistics->stdev[split_line][detector],
                                read_statistics->stdev[split_line][detector]);
                            assert_true(0);
                        } 
               
                        if(fabs(
                            stored_statistics->skewness[split_line][detector] -
                            read_statistics->skewness[split_line][detector]) >
                            double_tolerance)
                        {
                            IAS_LOG_ERROR("Skewness results do not match for "
                                "band %d, sca %d detector %d", band_number, 
                                sca_number, detector);
                            assert_true(0);
                        }
            
                        if(fabs(
                            stored_statistics->kurtosis[split_line][detector] -
                            read_statistics->kurtosis[split_line][detector]) >
                            double_tolerance)
                        {
                            IAS_LOG_ERROR("Results do not match at detector %d."
                                " Wrote %lf read %lf", detector,
                                stored_statistics->\
                                    kurtosis[split_line][detector],
                                read_statistics->\
                                    kurtosis[split_line][detector]);
                            assert_true(0);
                        }
                        
                        if(fabs(stored_statistics->\
                            correlation[split_line][detector] -
                            read_statistics->correlation[split_line][detector])
                            > double_tolerance)
                        {
                            IAS_LOG_ERROR("Results do not match at detector %d."
                                " Wrote %lf read %lf", detector, 
                                stored_statistics->\
                                    correlation[split_line][detector],
                                read_statistics->\
                                    correlation[split_line][detector]);
                            assert_true(0);
                        }
            
                        if(fabs(stored_statistics->\
                            mean_sq_resp[split_line][detector] -
                            read_statistics->mean_sq_resp[split_line][detector])
                            > double_tolerance)
                        {
                            IAS_LOG_ERROR("Results do not match at detector %d."
                                " Wrote %lf read %lf", detector, 
                                stored_statistics->\
                                    mean_sq_resp[split_line][detector],
                                read_statistics->\
                                    mean_sq_resp[split_line][detector]);
                            assert_true(0);
                        }
                        
                        if(strncmp(stored_statistics->processing_level, 
                            read_statistics->processing_level, 
                            IAS_RPS_PROCESSING_LEVEL_LENGTH) != 0)
                        {
                            IAS_LOG_ERROR("Written processing level %s does "
                                "not equal read processing level %s", 
                                stored_statistics->processing_level, 
                                read_statistics->processing_level);
                            assert_true(0);
                        }
                    }
                }
            }
            else
            {
                IAS_LOG_ERROR("Unable to perform comparison of data");
                assert_true(0);
            }
        
            /* cleanup data from the test */
            if (read_statistics != NULL)
            {
                ias_rps_db_free_histogram_results(read_statistics);
                read_statistics = NULL;
            }
        
            /* cleanup data from the test */
            if (stored_statistics != NULL)
            {
                ias_rps_db_free_histogram_results(stored_statistics);
                stored_statistics = NULL;
            }
        }
    }

    db_close(database);

    IAS_LOG_DEBUG("Unit test complete");
}

/*
 * NAME: test_database_io_earth
 *
 * PURPOSE: Test the ability to store the histogram statistics in 
 * HISTORAM_STATISTICS_EARTH table in the database.
 * 
 */
Ensure test_database_io_earth()
{
   test_database_io(IAS_RPS_EARTH_IMAGING_COLLECT);
}

/*
 * NAME: test_database_io_cal
 *
 * PURPOSE: Test the ability to store the histogram statistics in
 * HISTOGRAM_STATISTICS_CAL table in the database.
 * 
 */
Ensure test_database_io_cal()
{
   test_database_io(IAS_RPS_OLI_SOLAR_COLLECT);
   test_database_io(IAS_RPS_OLI_SOLAR_TIMESWEEP_COLLECT);
   test_database_io(IAS_RPS_OLI_SHUTTER_COLLECT);
   test_database_io(IAS_RPS_TIRS_BLACKBODY_COLLECT);
   test_database_io(IAS_RPS_TIRS_DEEPSPACE_COLLECT);
   test_database_io(IAS_RPS_EARTH_IMAGING_COLLECT);
}


/*
 * main function
 */
int main(int argc, char *argv[])
{
    /* Construct the test suite */
    TestSuite *suite = create_test_suite();

    /* Set the log level to warning for regular regression test execution of 
    unit tests */
    ias_log_set_output_level(IAS_LOG_LEVEL_WARN);

    if (ias_sat_attr_initialize(IAS_L8) != SUCCESS)
    {
        IAS_LOG_ERROR("Unable to load the satellite attributes used for "
            "processing.");
        return ERROR;
    }
   
    /* Use the env variable to determine if the database related tests should 
    be run */ 
    if (getenv("ENABLE_DB_TESTS") == NULL)
    {
        printf("DB Tests will not run, setenv ENABLE_DB_TESTS to enable....\n");
        return SUCCESS;
    }
    else
    {
        add_test(suite, test_database_io_earth);
        add_test(suite, test_database_io_cal);
    }

    return run_test_suite(suite, create_text_reporter());
}

