/**************************************************************************
 NAME:     identify_diffuser_panel_test

 PURPOSE:  Performs CGreen unit testing of the routines implemented
           to identify the diffuser panel used during an OLI solar collect

 RETURNS:  CGreen unit status

**************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "ias_const.h"
#include "ias_cpf.h"
#include "ias_l0r.h"
#include "ias_logging.h"
#include "ias_types.h"
#include "ias_satellite_attributes.h"
#include "cgreen.h"
#include "../ias_radiometry.h"


static char *test_data_path = NULL;


Ensure test_ias_rps_identify_diffuser_panel()
{
    char l0r_filename[PATH_MAX];
    char cpf_filename[PATH_MAX];
    char start_date[IAS_L0R_DATE_LENGTH];
    char cpf_dir[]  = ".";
    char cpf_name[] = "L8CPF20011001_20011231.01";
    char l0r_directory[] = "ias_lib/rps/identify_diffuser_panel";
    char l0r_name[] = "LO800O0000012001276EDC00";
    IAS_CPF *cpf = NULL;
    const struct IAS_CPF_EARTH_CONSTANTS *ec = NULL;
    IAS_DATETIME start_datetime;
    int test_panel_type;
    int expected_panel_type = IAS_RPS_OLI_DIFFUSER_TYPE_WORKING;
    uint16_t expected_diffuser_resolver_position = 8752;
    uint16_t test_diffuser_resolver_position;
    int status;


    printf("Checking identification of OLI diffuser panel from test "
        "telemetry dataset...\n");

    /* Turn off logging */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* Initialize the satellite attributes library */
    status = ias_sat_attr_initialize(IAS_L8);
    if (status != SUCCESS)
    {
        assert_true_with_message(FALSE, "Error initializing Satellite "
            "Attributes Library");
        return;
    }

    /* Open the CPF */
    status = snprintf(cpf_filename, sizeof(cpf_filename),
        "%s/%s", cpf_dir, cpf_name);
    if ((status < 0) || (status >= sizeof(cpf_filename)))
    {
        assert_true_with_message(FALSE, "Error creating test CPF file name");
        return;
    }
    cpf = ias_cpf_read(cpf_filename);
    assert_not_equal_with_message(cpf, NULL, "Error reading CPF file %s",
        cpf_filename);
    if (cpf == NULL)
    {
        return;
    }
    ec = ias_cpf_get_earth_const(cpf);
    if (ec == NULL)
    {
        assert_true_with_message(FALSE, "Error reading EARTH_CONSTANTS "
            "group data");
        ias_cpf_free(cpf);
        return;
    }

    /* Create the a string representing the image start/stop time */
    sprintf(start_date, "2001:276:12:40:59.6825240");
    status = ias_misc_parse_datetime_string((const char *)start_date,
       IAS_DATETIME_L0R_FORMAT, &start_datetime);
    assert_equal_with_message(status, SUCCESS, "Error parsing test "
        "datetime string %s", start_date);
    if (status != SUCCESS)
    {
        ias_cpf_free(cpf);
        return;
    }

    /* Run the algorithm */
    status = snprintf(l0r_filename, sizeof(l0r_filename),
        "%s/%s/%s", test_data_path, l0r_directory, l0r_name);
    if ((status < 0) || (status >= sizeof(l0r_filename)))
    {
        assert_true_with_message(FALSE, "Error creating test L0R file name");
        ias_cpf_free(cpf);
        return;
    }
    status = ias_rps_identify_diffuser_panel(l0r_filename, ec,
        start_datetime, &test_diffuser_resolver_position,
        &test_panel_type);
    assert_equal_with_message(status, SUCCESS, "Error from panel "
        "identification routine");
    if (status != SUCCESS)
    {
        ias_cpf_free(cpf);
        return;
    }

    /* Check the results */
    assert_equal_with_message(test_diffuser_resolver_position,
        expected_diffuser_resolver_position, "Expected resolver position %d "
        "not equal to returned resolver position %d",
        expected_diffuser_resolver_position,
        test_diffuser_resolver_position);
    assert_equal_with_message(test_panel_type, expected_panel_type,
        "Expected panel ID %d not equal to returned panel ID %d",
        expected_panel_type, test_panel_type);

    /* Clean up */
    ias_cpf_free(cpf);

    /* Re-enable logging */
    ias_log_set_output_level(IAS_LOG_LEVEL_DEBUG);

}   /* END -- test_ias_rps_identify_diffuser_panel */



int main(int argc, char *argv[])
{
    /* Check for the existence of the test data directory.  If it doesn't
       exist, we can't go any further */
    test_data_path = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (test_data_path == NULL)
    {
        printf("Cannot find main IAS unit test data directory\n");
        exit(EXIT_FAILURE);
    }

    /* Create the test suite */
    TestSuite *suite = create_test_suite();
    add_test(suite, test_ias_rps_identify_diffuser_panel);

    /* Run the test(s) */
    if (argc > 1)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}
