
#include "ias_pixel_mask.h"
#include "cgreen/cgreen.h"
#include "../ias_radiometry.h"
#include "ias_logging.h"

Ensure inoperable_detector_test(void);
Ensure outofspec_detector_test(void);

int main( int argc, char *argv[] )
{
    TestSuite *suite = create_test_suite();
    add_test(suite, inoperable_detector_test);
    add_test(suite, outofspec_detector_test);
    return run_test_suite(suite, create_text_reporter());
}


Ensure outofspec_detector_test(void)
{
    int status;
    IAS_PIXEL_MASK *pm;
    IAS_PIXEL_MASK_ITERATOR *it = NULL;
    IAS_PIXEL_MASK_SPAN span;
    int num_detectors = 8;
    int num_lines = 20;
    int band = 3;
    int sca = 2;
    int outofspec_detectors[12] = {3, 6, 8, 12};
    int i;


    /* Create a pixel mask */
    pm = ias_pm_create(band, sca, num_detectors, num_lines);
    assert_not_equal(pm, NULL);

    status = ias_rps_outofspec_detector_identification(outofspec_detectors,
        3, band, sca, num_detectors, num_lines, pm);
    if (status != SUCCESS)
    {
        assert_true_with_message(0, "ERROR executing "
            "ias_rps_outofspec_detector_identification");
        ias_pm_destroy_iterator(it);
        return;
    }

    /* Get an iterator for out-of-spec detectors */
    it = ias_pm_get_iterator(pm, IAS_PM_INCLUDE, IAS_RPS_OUTOFSPEC_DETECTOR);
    assert_not_equal(it, NULL);

    /* Get the spans, there should be three for detectors 3, 6, and 8 */
    i = 0;
    while (ias_pm_get_next_span(it, &span))
    {
        assert_equal(span.detector_index, outofspec_detectors[i] - 1);
        assert_equal(span.starting_pixel_index, 0);
        assert_equal(span.length_of_span, num_lines);
        assert_equal(span.pixel_mask, IAS_RPS_OUTOFSPEC_DETECTOR);
        i++;
    }
    assert_equal(i, 3);
    ias_pm_destroy_iterator(it);

    /* Get span(s) excluding inoperable detectors--there shouldn't be any */
    it = ias_pm_get_iterator(pm, IAS_PM_EXCLUDE, IAS_RPS_OUTOFSPEC_DETECTOR);
    assert_not_equal(it, NULL);
    status = ias_pm_get_next_span(it, &span);
    assert_equal(status, 0);

    /* Iterator clean up */
    ias_pm_destroy_iterator(it);

    /* Test for error conditions */
    outofspec_detectors[0] = 12;
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = ias_rps_outofspec_detector_identification(outofspec_detectors,
        3, band, sca, num_detectors, num_lines, pm);
    assert_true_with_message(status, "Detector out of range not caught");
    ias_log_set_output_level(IAS_LOG_LEVEL_WARN);

    /* Final cleanup */
    ias_pm_destroy(pm);
}


Ensure inoperable_detector_test(void)
{
    int status;
    IAS_PIXEL_MASK *pm;
    IAS_PIXEL_MASK_ITERATOR *it = NULL;
    IAS_PIXEL_MASK_SPAN span;
    int num_detectors = 8;
    int num_lines = 20;
    int band = 2;
    int sca = 3;
    int inop_detectors[12] = { 3, 6, 8, 12 };
    int i;

    pm = ias_pm_create( band, sca, num_detectors, num_lines );
    assert_not_equal( pm, NULL );

    status = ias_rps_inoperable_detector_identification( inop_detectors,
        3, band, sca, num_detectors, num_lines, pm );
    if ( status != SUCCESS )
    {
        assert_true_with_message( status, "ERROR executing "
            "ias_rps_inoperable_detector_identification" );
        ias_pm_destroy_iterator( it );
        return;
    }

    /* Get an iterator for inoperable detectors */
    it = ias_pm_get_iterator( pm, IAS_PM_INCLUDE, IAS_RPS_INOP_DETECTOR );
    assert_not_equal( it, NULL );

    /* Get the spans, there should be three for detectors 3, 6 and 8 */
    i = 0;
    while ( ias_pm_get_next_span( it, &span ))
    {
        assert_equal( span.detector_index, inop_detectors[i] - 1 );
        assert_equal( span.starting_pixel_index, 0 );
        assert_equal( span.length_of_span, num_lines );
        assert_equal( span.pixel_mask, IAS_RPS_INOP_DETECTOR );
        i++;
    }
    assert_equal( i, 3 );
    ias_pm_destroy_iterator( it );

    /* Get an iterator that excludes the inoperable detectors */
    it = ias_pm_get_iterator( pm, IAS_PM_EXCLUDE, IAS_RPS_INOP_DETECTOR );
    assert_not_equal( it, NULL );

    /* Attempt to get a span, there should be none */
    status = ias_pm_get_next_span( it, &span );
    assert_equal( status, 0 );
    ias_pm_destroy_iterator( it );

    /* Test for error conditions */
    inop_detectors[0] = 12;
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = ias_rps_inoperable_detector_identification( inop_detectors,
        3, band, sca, num_detectors, num_lines, pm );
    assert_true_with_message( status, "Detector out of range not caught" );
    ias_log_set_output_level(IAS_LOG_LEVEL_WARN);

    ias_pm_destroy( pm );
}
