/*----------------------------------------------------------------------------
NAME:    lookup_response_linearization_test

PURPOSE: Test the OLI/TIRS lookup response linearization algorithm 
         implementation.

USAGE:   lookup_response_linearization_test [test name]

         When run without a command line argument all tests will be run.
         Pass the name of a test to run the single test -- useful for
         debugging.
----------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "cgreen.h"
#include "ias_logging.h"
#include "../ias_radiometry.h"
#include "ias_rlut.h"


#define OLI_NLINES 10220
#define OLI_NDETS  494
#define TIRS_NLINES 35
#define TIRS_NDETS  2
#define TIRS_NVALS  12

#include "tirs_lookup_linearization_data.h"

/* Path to test data and file names */
static char *test_data_path = NULL;
static char data_dir[] = "ias_lib/rps/lookup_response_linearization";

/* Test RLUT file.  This file is a copy of the production RLUT and modifided
   for Tracker 6001 which modified the RLUT and RLUT I/O to combine the 3Q
   and lookup tables into a single file. */
static char rlut_basename[] = "eval_L8RLUT20130211_20431231v07.h5";

/* Binary data files containing test input data and expected correction results
   used in the cal/val prototype code */
static char data_file[] = "ERL_allints_f1b1.dat";
static char results_file[] = "Valdata_ALT_lintest_f1b1.dat";

/*--------------------------------------------------------------------------
   test_algorithm_directly

   This test executes the algorithm with a subset of the Cal/Val prototype
   input and output.  There is a README in the test directory that explains
   where the input files come from.  That README points to another that
   explains how the data were derived.
---------------------------------------------------------------------------*/
Ensure oli_test_algorithm_directly()
{
    FILE *inf;                  /* File pointer for all reads */
    char file_name[PATH_MAX];   /* File name to open */
    IAS_RLUT_IO *test_rlut = NULL;
                                /* Open RLUT file */
    IAS_RLUT_LOOKUP_TABLE *lin_params = NULL;
                                /* Linearization parameters */
    float *data_buffer = NULL;  /* Buffer containing the test input data */
    float *expected_results;    /* Cal/Val expected results */
    int band_number;            /* Current band number */
    int sca_number;             /* Current SCA number */
    int n;                      /* Number of values read from file */
    int num_values;             /* Number of values for malloc and reads */
    int i;                      /* Just your ordinary loop counter */
    int difference_count;       /* Count of differences between the 
                                   calculated value and expected results */
    int status;                 /* Return status */


    printf("Testing lookup linearization correction on simulated "
            "OLI data...\n");

    /* Open and read the input image data buffer */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_data_path,
        data_dir, data_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building input data file name");
        return;
    }
    inf = fopen(file_name, "r");
    if (inf == NULL)
    {
        assert_true_with_message(FALSE, "Error opening input image data "
            "file %s", file_name);
        return;
    }
    num_values = OLI_NLINES * OLI_NDETS;
    data_buffer = malloc(num_values * sizeof(float));
    if (!data_buffer)
    {
        assert_true_with_message(FALSE, "Error allocating memory for input "
            "image data buffer");
        return;
    }
    n = fread(data_buffer, sizeof(float), num_values, inf);
    fclose(inf);
    if (n != num_values)
    {
        assert_true_with_message(FALSE, "Unable to read data file" );
        free(data_buffer);
        return;
    }

    /* Open and read the RLUT file */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_data_path,
        data_dir, rlut_basename);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building RLUT file name");
        free(data_buffer);
        return;
    }
    test_rlut = ias_rlut_open_file(file_name, IAS_READ);
    assert_not_equal_with_message(test_rlut, NULL, "Error opening RLUT "
        "file %s to read", file_name);
    if (test_rlut == NULL)
    {
        free(data_buffer);
        return;
    }

    /* Set the band and sca number consistent with CalVal expected results */
    band_number = 2;
    sca_number = 2;

    /* Read the linearity parameters */
    lin_params = ias_rlut_read_lookup_table(test_rlut,
        IAS_RLUT_LINEARITY_LOOKUP, band_number, sca_number);
    assert_not_equal_with_message(lin_params, NULL, "Error reading band %d "
            "SCA %d linearization parameters", band_number, sca_number);
    if (lin_params == NULL)
    {
        free(data_buffer);
        ias_rlut_close_file(test_rlut);
        return;
    }

    /* Call the algorithm */
    status = ias_rps_lookup_response_linearization(data_buffer,
            (const IAS_RLUT_LOOKUP_TABLE *)lin_params, OLI_NLINES);
    assert_equal_with_message(status, SUCCESS, "Error in lookup "
            "response linearization algorithm implementation");
    if (status != SUCCESS)
    {
        free(data_buffer);
        ias_rlut_free_lookup_table(lin_params);
        lin_params = NULL;
        ias_rlut_close_file(test_rlut);
        return;
    }
    ias_rlut_free_lookup_table(lin_params);
    lin_params = NULL;

    /* We are finished with the rlut file, close it */
    ias_rlut_close_file(test_rlut);

    /* Open the expected results file and read its contents */
    status = snprintf(file_name, sizeof(file_name), "%s/%s/%s", test_data_path,
        data_dir, results_file);
    if ((status < 0) || (status >= sizeof(file_name)))
    {
        assert_true_with_message(FALSE, "Error building results file name");
        free(data_buffer);
        return;
    }
    num_values = OLI_NLINES * OLI_NDETS;
    expected_results = malloc(num_values * sizeof(float));
    if (expected_results == NULL)
    {
        assert_true_with_message( FALSE, "Error allocating memory for the "
           "expected results data buffer");
        free(data_buffer);
        return;
    }
    inf = fopen(file_name, "r");
    if (inf == NULL)
    {
        assert_true_with_message(FALSE, "Opening expected results file %s",
            file_name);
        free(expected_results);
        free(data_buffer);
        return;
    }
    n = fread(expected_results, sizeof(float), num_values, inf);
    fclose(inf);
    if (n != num_values)
    {
        assert_true_with_message(FALSE, "Unable to read results file %s",
            file_name);
        free(expected_results);
        free(data_buffer);
        return;
    }

    /* Compare the results.  First, do a buffer-to-buffer comparison; if
       the buffers have the same contents, we're done.  If they don't...
       Loop through the results and count the differences.  This comparison
       isn't done with asserts since there is no need for 6400 assert
       statements which cost message buffer space.  One assert for pass
       fail with a count is good enough. */
    num_values = OLI_NLINES * OLI_NDETS;
    difference_count = 0;

    status = memcmp(data_buffer, expected_results, num_values * sizeof(float));
    assert_equal_with_message(status, 0, "Buffer comparison failed:  "
        "possible differences found between the test corrected data and "
        "expected values");
    if (status != 0)
    {
        for (i = 0; i < num_values; i++)
        {
            if (data_buffer[i] != expected_results[i])
            {
                difference_count++;
            }
        }
        if (difference_count == 0)
        {
            assert_true_with_message(FALSE, "Anomalous failure of memcmp() "
                "function--no differences found out of %d values", num_values);
        }
        else    /* Most likely reason for intial failure */
        {
            assert_true_with_message(FALSE, "Found %d differences out of "
                "%d values ", difference_count, num_values);
        }
    }

    /* Cleanup */
    free(expected_results);
    free(data_buffer);
}  /* END oli_test_algorithm_directly */


Ensure tirs_test_algorithm_directly()
{
    int i;                      /* Just your ordinary loop counter */
    int difference_count;
    int status;
    IAS_RLUT_LOOKUP_TABLE params;

    params.num_detectors = TIRS_NDETS;
    params.num_values = TIRS_NVALS;
    params.input_dn = input_dn;
    params.output_correction = output_correction;


    /* Call the algorithm */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = ias_rps_lookup_response_linearization( data_buffer,
            &params, TIRS_NLINES );
    assert_equal_with_message(status, SUCCESS, "Error in secondary "
            "response linearization algorithm implementation");
    if (status != SUCCESS)
    {
        return;
    }

    difference_count = 0;
    for ( i = 0; i < TIRS_NDETS * TIRS_NLINES; i++ )
    {
        if ( data_buffer[i] != expected_results[i] )
            difference_count++;
    }
    assert_equal( difference_count, 0 );
}  /* END tirs_test_algorithm_directly */


int main(int argc, char *argv[])
{
    /* Get the test data directory */
    test_data_path = getenv("IAS_UNIT_TEST_DATA_DIR");
    if (test_data_path == NULL)
    {
        printf("Cannot find test data directory location\n");
        return 1;
    }

    /* Initialize the satellite attributes */
    int status = ias_sat_attr_initialize(IAS_L8);
    if (status != SUCCESS)
    {
        printf("Cannot initialize IAS satellite attributes information\n");
        return 1;
    }

    /* Build the test suite */
    TestSuite *suite = create_test_suite();
    add_test(suite, oli_test_algorithm_directly);
    add_test(suite, tirs_test_algorithm_directly);

    /* Run either a single test or the entire suite */
    if (argc == 2)
    {
        return run_single_test(suite, argv[1], create_text_reporter());
    }
    return run_test_suite(suite, create_text_reporter());
}   /* END unit test main */
