#include <stdio.h>
#include <string.h>
#include <math.h>

#include "ias_miscellaneous.h"
#include "ias_pixel_mask.h"
#include "ias_satellite_attributes.h"
#include "ias_db.h"
#include "ias_db_get_connect_info.h"
#include "ias_db_insert.h"
#include "ias_logging.h"
#include "cgreen/cgreen.h"
#include "../ias_radiometry.h"


Ensure impulse_noise_test( void );
Ensure histogram_stats_test( void );
int db_store_histogram_stats( IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS * );
int db_check_histogram_stats( IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS * );
int db_store_impulse_noise( IAS_RPS_DB_IMPULSE_NOISE_CHAR_RESULTS * );
int db_check_impulse_noise( IAS_RPS_DB_IMPULSE_NOISE_CHAR_RESULTS * );
struct ias_db_connection *db_open( void );
void db_close( struct ias_db_connection * );

static int test_database = FALSE;

int main( int argc, char *argv[] )
{
    char *enable_db_tests;

    /* See if the database tests should be enabled */
    enable_db_tests = getenv( "ENABLE_DB_TESTS" );
    if ( enable_db_tests )
    {
        test_database = TRUE;
        printf("\n\tDatabase checks have been enabled\n\n");
    }
    else
    {
        printf("\n\t********* Database checks will not be made *********\n");
        printf("\tTo enable the checks execute: setenv ENABLE_DB_TESTS\n\n");
    }

    TestSuite *suite = create_test_suite();
    add_test( suite, impulse_noise_test );
    add_test( suite, histogram_stats_test );

    if (argc > 1)
    {
        return run_single_test( suite, argv[1], create_text_reporter() );
    }
    return run_test_suite( suite, create_text_reporter() );
}


/*------------------------------------------------------------------
  NAME: impulse_noise_test
  PURPOSE: Test the impulse noise algorithm
------------------------------------------------------------------*/
Ensure impulse_noise_test( void )
{
    IAS_REPORT report_info;
    IAS_REPORT_HEADER *report_header = NULL;
    char work_order_id[IAS_WORKORDER_ID_SIZE];
    char routine_name[] = "Impulse noise characterization";
    IAS_PIXEL_MASK *pm1;
    IAS_PIXEL_MASK *pm2;
    PIXEL_MASK_TYPE mask;
    int det;
    int status;
    uint16_t *buffer;
    uint16_t l0r_data[12][8] =
    {
        { 400,650,399,2  ,402,400,2  ,400 },
        { 399,400,400,2  ,412,408,2  ,401 },
        { 500,399,399,2  ,399,400,2  ,394 },
        { 400,600,401,2  ,401,399,2  ,400 },
        { 4  ,4  ,4  ,2  ,4  ,4  ,2  ,4   },
        { 401,400,401,2  ,400,200,2  ,400 },
        { 401,399,300,2  ,280,400,2  ,401 },
        { 550,520,399,2  ,400,500,2  ,600 },
        { 550,402,400,2  ,280,150,2  ,420 },
        { 399,400,401,50 ,401,400,2  ,399 },
        { 4  ,4  ,80 ,2  ,4  ,4  ,2  ,4   },
        { 400,400,400,2  ,400,60 ,2  ,400 }
    };
    double detector_noise[] =
        { 0.78, 0.81, 0.86, 0.82, 0.79, 0.85, 0.91, 0.80 };
    int i, j;
    IAS_RPS_DB_IMPULSE_NOISE_CHAR_RESULTS *db_results = NULL;

    /* Allocate database results structure if needed */
    if ( test_database )
    {
        db_results = ias_rps_db_allocate_impulse_noise_results( 8 );
        if ( !db_results )
            assert_true_with_message( 0,
                "Error allocating impulse noise memory database tests will "
                "be skipped");
    }

    /* Create two pixel masks */
    pm1 = ias_pm_create( 1, 1, 8, 12 );
    assert_not_equal( pm1, NULL );
    pm2 = ias_pm_create( 1, 1, 8, 12 );
    assert_not_equal( pm2, NULL );

    /* Add the inoperable detectors */
    assert_equal( SUCCESS, ias_pm_add_pixels( pm1, 4-1, 1-1, 12,
        IAS_RPS_INOP_DETECTOR));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm1, 7-1, 1-1, 12,
        IAS_RPS_INOP_DETECTOR));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm2, 4-1, 1-1, 12,
        IAS_RPS_INOP_DETECTOR));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm2, 7-1, 1-1, 12,
        IAS_RPS_INOP_DETECTOR));
    /* Add the dropped frames */
    for ( det = 1-1; det <= 8-1; det++ )
    {
        assert_equal( SUCCESS, ias_pm_add_pixels( pm1, det, 5-1, 1,
            IAS_RPS_DROPLINE ));
        assert_equal( SUCCESS, ias_pm_add_pixels( pm2, det, 5-1, 1,
            IAS_RPS_DROPLINE ));
    }
    for ( det = 1-1; det <= 8-1; det++ )
    {
        assert_equal( SUCCESS, ias_pm_add_pixels( pm1, det, 11-1, 1,
            IAS_RPS_DROPLINE ));
        assert_equal( SUCCESS, ias_pm_add_pixels( pm2, det, 11-1, 1,
            IAS_RPS_DROPLINE ));
    }

    /* Point to the data */
    buffer = &l0r_data[0][0];

    /* Set up a standard header */
    setenv("PCS_WOID", "I4582", 1);
    strcpy(work_order_id, "I4582");
    report_header = ias_misc_initialize_report_header("",
        work_order_id, routine_name);
    if (!report_header)
    {
        assert_true_with_message( 0, "Error initializing report header");
        return;
    }
    ias_misc_set_header_pathrow(report_header, 0, 0);
    report_info.header = report_header;
    report_info.directory = NULL;

    /* Calling with an even median width filter is an error - test it */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);
    status = ias_rps_impulse_noise_characterization( buffer, 1, 1, 8, 12,
        4, // median_filter_width
        detector_noise,
        8, // int impulse_noise_limit
        pm1, NULL, db_results );
    assert_equal( status, ERROR );

    /* Using pm1 run the routine with filter width of 3 */
    status = ias_rps_impulse_noise_characterization( buffer, 1, 1, 8, 12,
        3, // median_filter_width
        detector_noise,
        8, // int impulse_noise_limit
        pm1, &report_info, db_results );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0,
            "ias_rps_impulse_noise_characterization returned ERROR");
        ias_misc_free_report_header(report_header);
        return;
    }

    /* Using pm2 run the routine with filter width of 5 */
    status = ias_rps_impulse_noise_characterization( buffer, 1, 1, 8, 12,
        5, // median_filter_width
        detector_noise,
        8, // int impulse_noise_limit
        pm2, NULL, NULL );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0,
            "ias_rps_impulse_noise_characterization returned ERROR");
        ias_misc_free_report_header(report_header);
        return;
    }

    /* Make sure the correct number of impulse noise locations were
       identified */
    int count1 = 0;
    int count2 = 0;
    for ( i = 0; i < 8; i++ )
    {
        for ( j = 0; j < 12; j++ )
        {
            ias_pm_get_mask_at( pm1, i, j, &mask );
            if ( mask & IAS_RPS_IMPNOISE )
                count1++;
            ias_pm_get_mask_at( pm2, i, j, &mask );
            if ( mask & IAS_RPS_IMPNOISE )
                count2++;
        }
    }
    assert_equal( count1, 14 );
    assert_equal( count2, 10 );

    /* Check each of the pixel locations that should be impulse noise */
    int in_det1[] = {0,1,1,1,2,4,4,4,4,5,5,5,5,7};
    int in_line1[] = {2,0,3,7,6,0,6,7,8,0,1,5,1,7};
    for ( i = 0; i < 14; i++ )
    {
        ias_pm_get_mask_at( pm1, in_det1[i], in_line1[i], &mask );
        assert_true_with_message( mask & IAS_RPS_IMPNOISE,
            "i = %d  Detector %d  Line %d  Mask %d", in_det1[i],
                in_line1[i], mask );
    }

    int in_det2[]  = {0,1,1,1,2,4,4,5,5,7};
    int in_line2[] = {2,0,3,7,6,6,8,1,5,7};
    for ( i = 0; i < 10; i++ )
    {
        ias_pm_get_mask_at( pm2, in_det2[i], in_line2[i], &mask );
        assert_true_with_message( mask & IAS_RPS_IMPNOISE,
            "i = %d  Detector %d  Line %d  Mask %d", in_det2[i],
                in_line2[i], mask );
    }

    ias_pm_destroy( pm1 );
    ias_pm_destroy( pm2 );

    ias_misc_free_report_header( report_header );

    /* Test the database only if the allocation of memory was successful */
    if ( db_results )
    {
        db_results->band = 1;
        db_results->sca = 1;
        db_results->num_detectors = 8;
        if ( db_store_impulse_noise( db_results ) != SUCCESS)
        {
            assert_true_with_message( 0, "Error storing impulse noise "
                "results to the database");
        }
        else
        {
            db_check_impulse_noise( db_results );
        }
        db_results = NULL;

        /* Test again with zero impulse noise locations */
        db_results = ias_rps_db_allocate_impulse_noise_results( 20 );
        if ( db_results )
        {
            int i;
            for ( i = 0; i < 20; i++ )
                db_results->in_counts[i] = 0;
            db_results->band = 2;
            db_results->sca = 1;
            db_results->num_detectors = 20;
            if ( db_store_impulse_noise( db_results ) != SUCCESS)
            {
                assert_true_with_message( 0, "Error attempting to store "
                    " zero impulse noise results");
            }
        }
    }

}  /* END impulse_noise_test */


/*------------------------------------------------------------------
  NAME:    histogram_stats_test
  PURPOSE: Test the histogram statistics algorithm
------------------------------------------------------------------*/
Ensure histogram_stats_test( void )
{
    IAS_PIXEL_MASK *pm;
    IAS_PIXEL_MASK *pan_pm;
    IAS_REPORT report_info;
    IAS_REPORT_HEADER *report_header = NULL;
    IAS_RPS_HISTOGRAM_INTERMEDIATE_RESULTS *intermediate_results;
    IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS *db_results;
    char work_order_id[IAS_WORKORDER_ID_SIZE];
    char routine_name[] = "Histogram tests";
    char rps_level[] = "RPS1";
    char rlut_version[] = "V1.0.0";
    char l0r[] = "LO82220010052011265EDC00";
    char proc_level[] = "RPS2";
    int status;
    void *buffer;
    uint16_t l0r_data[14][8];
    FILE *inf;
    unsigned int temp, det, line;
    float l1r_data[14][8];
    float pan_data[2002][8];
    int detector_offsets[] = { 0,  2,  0,  2,  0,  2,  0,  2 }; 
    int pan_detector_offsets[] = { 0,  0,  0,  0,  0,  0,  0,  0,
                                        0,  0,  0,  0,  0,  0,  0,  0 }; 

    /* Suppress the error messages from the library */
    ias_log_set_output_level(IAS_LOG_LEVEL_DISABLE);

    /* Initialize the attributes lib */
    status = ias_sat_attr_initialize( IAS_L8 );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error initializing the attributes"
            "library" );
        return;
    }

/* This test section is for timing purposes.  By setting the defines
   a large dataset can be tested for efficiency. */
//#define TEST_DATA_SET
#ifdef TEST_DATA_SET
    #define LINES2DO 90000
    #define NUM_LINES2DO 100
    if ( NUM_LINES2DO < 0 )
    {
        assert_true_with_message( 0, "No lines input on parameter");
        return;
    }
    float bigbuff[LINES2DO];
    inf = fopen( "big.txt", "r" );
    if ( !inf )
    {
        assert_true_with_message( 0, "Error opening: testim11.txt");
        return;
    }
    /* Set up a standard header */
    setenv( "PCS_WOID", "I4582", 1 );
    strcpy( work_order_id, "I4582" );
    report_header = ias_misc_initialize_report_header( l0r, work_order_id,
        routine_name );
    if ( !report_header )
    {
        assert_true_with_message( 0, "Error initializing report header" );
        return;
    }
    ias_misc_set_header_pathrow( report_header, 10, 10 );
    report_info.header = report_header; *
    /* Additional data needed for histogram stats */
    report_info.l0r_file_name = l0r;
    report_info.rps_level = rps_level;
    report_info.rlut_version = rlut_version;
    report_info.oli_frame_rate = 15.0 / 6000.0;
    report_info.tirs_frame_rate = 30.0 / 6000.0;

    for ( line = 0; line < NUM_LINES2DO; line++ )
        fscanf( inf, "%f", &bigbuff[line] );
    pm = ias_pm_create( 1, 1, 1, NUM_LINES2DO );
    assert_not_equal( pm, NULL );
    buffer = (void *)bigbuff;
    for ( line = 0; line < NUM_LINES2DO; line++ )
    {
        status = ias_rps_histogram_statistics_characterization(
                buffer, IAS_R4, 1, 1, 1, NUM_LINES2DO, 0, 0, pm, NULL,
                NULL, NULL );
    }
    fclose( inf );
    ias_misc_free_report_header( report_header );
    return;
#endif

    inf = fopen( "testim11.txt", "r" );
    if ( !inf )
    {
        assert_true_with_message( 0, "Error opening: testim11.txt");
        return;
    }
    for ( det = 0; det < 8; det++ )
    {
        for ( line = 0; line < 14; line++ )
        {
            fscanf( inf, "%ud", &temp );
            l0r_data[line][det] = (uint16_t)temp;
            l1r_data[line][det] = (float)temp;
        }
    }
    fclose( inf );

    /* Create the pixel mask */
    pm = ias_pm_create( 1, 1, 8, 14 );
    assert_not_equal( pm, NULL );
    /* Note: The pixel mask is a 0-based detector and line structure.  The
       Cal/Val test code is 1-based.  In all the following pixel mask
       initializations, I've used the 1-based values from the test code and
       subtracted one to ensure the masks are correctly populated. */
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 1-1, 5-1, 1,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 1-1, 11-1, 1,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 1-1, 13-1, 2,
        IAS_RPS_DROPLINE ));

    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 2-1, 1-1, 2,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 2-1, 7-1, 1,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 2-1, 8-1, 1,
        IAS_RPS_IMPNOISE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 2-1, 11-1, 1,
        IAS_RPS_IMPNOISE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 2-1, 13-1, 1,
        IAS_RPS_DROPLINE ));

    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 3-1, 4-1, 1,
        IAS_RPS_IMPNOISE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 3-1, 5-1, 1,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 3-1, 7-1, 1,
        IAS_RPS_IMPNOISE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 3-1, 11-1, 1,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 3-1, 13-1, 2,
        IAS_RPS_DROPLINE ));

    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 4-1, 1-1, 2,
        IAS_RPS_DROPLINE ));

    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 5-1, 5-1, 1,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 5-1, 11-1, 1,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 5-1, 10-1, 1,
        IAS_RPS_IMPNOISE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 5-1, 13-1, 2,
        IAS_RPS_DROPLINE ));

    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 6-1, 7-1, 1,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 6-1, 13-1, 1,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 6-1, 1-1, 2,
        IAS_RPS_DROPLINE ));

    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 7-1, 5-1, 1,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 7-1, 11-1, 1,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 7-1, 13-1, 2,
        IAS_RPS_DROPLINE ));

    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 8-1, 7-1, 1,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 8-1, 13-1, 1,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 8-1, 5-1, 1,
        IAS_RPS_IMPNOISE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 8-1, 1-1, 2,
        IAS_RPS_DROPLINE ));

    /* Set up a standard header */
    setenv( "PCS_WOID", "I4582", 1 );
    strcpy( work_order_id, "I4582" );

    report_header = ias_misc_initialize_report_header( l0r, work_order_id,
        routine_name );
    if ( !report_header )
    {
        assert_true_with_message( 0, "Error initializing report header" );
        return;
    }
    ias_misc_set_header_pathrow( report_header, 10, 10 );

    report_info.header = report_header;
    report_info.directory = NULL;
    /* Additional data needed for histogram stats */
    report_info.rps_level = rps_level;
    report_info.rlut_version = rlut_version;
    report_info.oli_frame_rate = 15.0 / 12000.0;
    report_info.tirs_frame_rate = 30.0 / 12000.0;

    /* Get a structure for the database */
    db_results = ias_rps_db_allocate_histogram_results( 8, FALSE );
    if ( !db_results )
    {
        assert_true_with_message( 0, "Error allocating database memory" );
        ias_misc_free_report_header( report_header );
        return;
    }

    /* IAS_R4 or IAS_UI2 */
    buffer = (void *)l0r_data;
    status = ias_rps_histogram_statistics_characterization(
        buffer, IAS_UI2, 1, 1, 8, 14, detector_offsets, 0, 0, pm, &report_info,
        db_results, NULL );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error from first "
            "ias_rps_histogram_statistics_characterization" );
        ias_misc_free_report_header( report_header );
        return;
    }

    /* Test the database routine but only if it was specified on the command
       line.  Reason for this is that not all developer schemas will match
       what histogram stats and the local sql script sets up for column
       names. */
    if ( test_database )
    {
        db_results->band_number = 1;
        db_results->sca = 1;
        db_results->processing_level = malloc( sizeof(proc_level) + 1 );
        strcpy( db_results->processing_level, proc_level );
        db_results->num_detectors = 8;
        db_results->is_odd_even_data_split = FALSE;
        if ( db_store_histogram_stats( db_results ) != SUCCESS)
        {
            assert_true_with_message( 0, "Error storing results to the"
                "database");
        }
        else
        {
            db_check_histogram_stats( db_results );
        }
    }
    else
    {
        printf("\t***** Database checks skipped *****\n");
        ias_rps_db_free_histogram_results( db_results );
    }

    buffer = (void *)l1r_data;
    status = ias_rps_histogram_statistics_characterization(
        buffer, IAS_R4, 1, 1, 8, 14, detector_offsets, 0, 0, pm, NULL,
        NULL, NULL );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error from second "
            "ias_rps_histogram_statistics_characterization" );
        ias_misc_free_report_header( report_header );
        return;
    }

    intermediate_results =
        malloc( sizeof(IAS_RPS_HISTOGRAM_INTERMEDIATE_RESULTS) * 8 );

    buffer = (void *)l0r_data;
    status = ias_rps_histogram_statistics_characterization(
        buffer, IAS_UI2, 1, 1, 8, 14, detector_offsets, 0, 0, pm, NULL,
        NULL, intermediate_results );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error from first "
            "ias_rps_histogram_statistics_characterization" );
        ias_misc_free_report_header( report_header );
        return;
    }

    /* Spot check some of the results, due to the use of detector_offsets,
       some of the original skipped pixels at the top of the input test array
       are set to IAS_RPS_DROPLINE, therefore, the results can be kept as
       those expected in ADD */
    assert_equal( intermediate_results[7].number_valid_values[0], 9);
    assert_double_equal( intermediate_results[7].mean[0], 414.4444444444444);
    assert_double_equal( intermediate_results[7].maximum[0], 611.0);
    assert_double_equal( intermediate_results[7].standard_deviation[0],
        102.18622554609807);
    assert_double_equal( intermediate_results[7].sum_of_adjacent_products[0],
        0.0);

    /* Test using the same data but make sure none of the pixels are valid 
       for detector three, detector 7 should be the same as above again. */
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 4-1, 1-1, 14, 
        IAS_RPS_DROPLINE ));
    status = ias_rps_histogram_statistics_characterization(
        buffer, IAS_UI2, 1, 1, 8, 14, detector_offsets, 0, 0, pm, NULL,
        NULL, intermediate_results );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error from all pixels invalid "
            "ias_rps_histogram_statistics_characterization" );
        ias_misc_free_report_header( report_header );
        return;
    }

    /* Detector three should be all zero */
    assert_equal( intermediate_results[3].number_valid_values[0], 0);
    assert_double_equal( intermediate_results[3].mean[0], 0.0);
    assert_double_equal( intermediate_results[3].maximum[0], 0.0);
    assert_double_equal( intermediate_results[3].minimum[0], 0.0);
    assert_double_equal( intermediate_results[3].standard_deviation[0], 0.0);
    assert_double_equal( intermediate_results[3].sum_of_adjacent_products[0],
        0.0);

    /* Detector seven should be the same as above */
    assert_equal( intermediate_results[7].number_valid_values[0], 9);
    assert_double_equal( intermediate_results[7].mean[0], 414.4444444444444);
    assert_double_equal( intermediate_results[7].maximum[0], 611.0);
    assert_double_equal( intermediate_results[7].standard_deviation[0],
        102.18622554609807);
    assert_double_equal( intermediate_results[7].sum_of_adjacent_products[0],
        0.0);

    free( intermediate_results );
    ias_pm_destroy( pm );
    ias_misc_free_report_header( report_header );

    inf = fopen( "testim12.txt", "r" );
    if ( !inf )
    {
        assert_true_with_message( 0, "Error opening: testim12.txt");
        return;
    }
    for ( det = 0; det < 8; det++ )
    {
        for ( line = 0; line < 2002; line++ )
        {
            fscanf( inf, "%ud", &temp );
            pan_data[line][det] = (float)temp;
        }
    }
    fclose( inf );

/*-------------------------------------------------------------------------
                    PAN BAND TESTING
---------------------------------------------------------------------------*/

    /* Create the pixel mask */
    pm = ias_pm_create( 5, 1, 8, 2002 );
    assert_not_equal( pm, NULL );
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 1-1, 5-1, 1,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 1-1, 11-1, 1,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 1-1, 2001-1, 2,
        IAS_RPS_DROPLINE ));

    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 2-1, 1-1, 2,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 2-1, 7-1, 1,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 2-1, 8-1, 1,
        IAS_RPS_IMPNOISE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 2-1, 11-1, 1,
        IAS_RPS_IMPNOISE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 2-1, 13-1, 1,
        IAS_RPS_DROPLINE ));

    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 3-1, 4-1, 1,
        IAS_RPS_IMPNOISE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 3-1, 5-1, 1,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 3-1, 7-1, 1,
        IAS_RPS_IMPNOISE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 3-1, 11-1, 1,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 3-1, 2001-1, 2,
        IAS_RPS_DROPLINE ));

    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 4-1, 1-1, 2, 
        IAS_RPS_DROPLINE ));

    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 5-1, 5-1, 1,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 5-1, 11-1, 1,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 5-1, 10-1, 1,
        IAS_RPS_IMPNOISE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 5-1, 2001-1, 2,
        IAS_RPS_DROPLINE ));

    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 6-1, 7-1, 1,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 6-1, 13-1, 1,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 6-1, 1-1, 2,
        IAS_RPS_DROPLINE ));

    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 7-1, 5-1, 1,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 7-1, 11-1, 1,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 7-1, 2001-1, 2,
        IAS_RPS_DROPLINE ));

    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 8-1, 7-1, 1,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 8-1, 13-1, 1,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 8-1, 5-1, 1,
        IAS_RPS_IMPNOISE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pm, 8-1, 1-1, 2, 
        IAS_RPS_DROPLINE ));

    /* Set up a standard header */
    setenv( "PCS_WOID", "I8888", 1 );
    strcpy( work_order_id, "I8888" );

    report_header = ias_misc_initialize_report_header( l0r, work_order_id,
        routine_name );
    if ( !report_header )
    {
        assert_true_with_message( 0, "Error initializing report header" );
        return;
    }
    ias_misc_set_header_pathrow( report_header, 10, 10 );

    /*report_header.l0r_filename[0] = '\0';
    report_header.path = 10;
    report_header.row = 10;
    status = ias_misc_initialize_report_header( &report_header, NULL );
    strcpy( report_header.l0r_filename, l0r );
    report_info.routine_name = routine_name;
    report_info.l0r_file_name = l0r; */

    report_info.header = report_header;
    /* Additional data needed for histogram stats */
    report_info.rps_level = rps_level;
    report_info.rlut_version = rlut_version;
    report_info.oli_frame_rate = 14.0 / 6000.0;
    report_info.tirs_frame_rate = 28.0 / 6000.0;

    buffer = (void *)pan_data;
    status = ias_rps_histogram_statistics_characterization(
        buffer, IAS_R4, 5, 1, 8, 2002, pan_detector_offsets, 0, 0, pm, 
        &report_info, NULL, NULL );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error from third "
            "ias_rps_histogram_statistics_characterization" );
        ias_misc_free_report_header( report_header );
        return;
    }

    ias_misc_free_report_header( report_header );

    /* Create the pixel mask for the pan band */
    pan_pm = ias_pm_create( 8, 1, 8, 2002 );
    assert_not_equal( pan_pm, NULL );
    assert_equal( SUCCESS, ias_pm_add_pixels( pan_pm, 1-1, 5-1, 1,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pan_pm, 1-1, 11-1, 1,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pan_pm, 1-1, 2001-1, 2,
        IAS_RPS_DROPLINE ));

    assert_equal( SUCCESS, ias_pm_add_pixels( pan_pm, 2-1, 1-1, 2,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pan_pm, 2-1, 7-1, 1,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pan_pm, 2-1, 8-1, 1,
        IAS_RPS_IMPNOISE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pan_pm, 2-1, 11-1, 1,
        IAS_RPS_IMPNOISE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pan_pm, 2-1, 13-1, 1,
        IAS_RPS_DROPLINE ));

    assert_equal( SUCCESS, ias_pm_add_pixels( pan_pm, 3-1, 4-1, 1,
        IAS_RPS_IMPNOISE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pan_pm, 3-1, 5-1, 1,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pan_pm, 3-1, 7-1, 1,
        IAS_RPS_IMPNOISE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pan_pm, 3-1, 11-1, 1,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pan_pm, 3-1, 2001-1, 2,
        IAS_RPS_DROPLINE ));

    assert_equal( SUCCESS, ias_pm_add_pixels( pan_pm, 4-1, 1-1, 2,
        IAS_RPS_DROPLINE ));

    assert_equal( SUCCESS, ias_pm_add_pixels( pan_pm, 5-1, 5-1, 1,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pan_pm, 5-1, 11-1, 1,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pan_pm, 5-1, 10-1, 1,
        IAS_RPS_IMPNOISE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pan_pm, 5-1, 2001-1, 2,
        IAS_RPS_DROPLINE ));

    assert_equal( SUCCESS, ias_pm_add_pixels( pan_pm, 6-1, 7-1, 1,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pan_pm, 6-1, 13-1, 1,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pan_pm, 6-1, 1-1, 2,
        IAS_RPS_DROPLINE ));

    assert_equal( SUCCESS, ias_pm_add_pixels( pan_pm, 7-1, 5-1, 1,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pan_pm, 7-1, 11-1, 1,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pan_pm, 7-1, 2001-1, 2,
        IAS_RPS_DROPLINE ));

    assert_equal( SUCCESS, ias_pm_add_pixels( pan_pm, 8-1, 7-1, 1,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pan_pm, 8-1, 13-1, 1,
        IAS_RPS_DROPLINE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pan_pm, 8-1, 5-1, 1,
        IAS_RPS_IMPNOISE ));
    assert_equal( SUCCESS, ias_pm_add_pixels( pan_pm, 8-1, 1-1, 2,
        IAS_RPS_DROPLINE ));

    /* Set up a standard header */
    setenv( "PCS_WOID", "I8888", 1 );

    strcpy( work_order_id, "I8888" );

    report_header = ias_misc_initialize_report_header( l0r, work_order_id,
        routine_name );
    if ( !report_header )
    {
        assert_true_with_message( 0, "Error initializing report header" );
        return;
    }
    ias_misc_set_header_pathrow( report_header, 10, 10 );

    /*report_header.l0r_filename[0] = '\0';
    report_header.path = 10;
    report_header.row = 10;
    status = ias_misc_initialize_report_header( &report_header, NULL );
    strcpy( report_header.l0r_filename, l0r );
    report_info.routine_name = routine_name;
    report_info.l0r_file_name = l0r; */

    report_info.header = report_header;
    /* Additional data needed for histogram stats */
    report_info.rps_level = rps_level;
    report_info.rlut_version = rlut_version;
    report_info.oli_frame_rate = 14.0 / 12000.0;
    report_info.tirs_frame_rate = 28.0 / 12000.0;

    buffer = (void *)pan_data;
    status = ias_rps_histogram_statistics_characterization(
        buffer, IAS_R4, 8, 1, 8, 2002, pan_detector_offsets, 0, 0, pan_pm, 
        &report_info, NULL, NULL );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error from third "
            "ias_rps_histogram_statistics_characterization" );
        ias_misc_free_report_header( report_header );
        return;
    }

    /* Compare the output files with baseline files.  This not only ensures
       the output is being done but also compares the values for the three
       cases listed in the ADD. */
    char output_line[80];
    FILE *pipe = popen("./compare_hist_results.csh", "r");
    if ( !pipe )
    {
        assert_true_with_message( 0, "Error comparing expected results");
        ias_misc_free_report_header( report_header );
        return;
    }
    while( fgets( output_line, 80, pipe ))
    {
        /* The script returns a string with the status as the first
           character.  The rest of the message is the file containing the
           error. */
        if ( output_line[0] != '0' )
        {
            output_line[0] = ' ';
            assert_true_with_message( 0, "Error in: %s", output_line);
        }
    }
    pclose( pipe );
    ias_pm_destroy( pm );
    ias_pm_destroy( pan_pm );
    ias_misc_free_report_header( report_header );
}  /* END histogram_stats_test */

/*------------------------------------------------------------------
  NAME:    db_open

  PURPOSE: Open a database connection

  RETURNS: database connection
           NULL on error
------------------------------------------------------------------*/
struct ias_db_connection *db_open( void )
{
    char *database_name = NULL;
    char *user_name = NULL;
    char *password = NULL;
    char *host = NULL;
    struct ias_db_connection *db = NULL;
    int status;

    status = ias_db_initialize_database_lib( );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error from "
            "ias_db_initialize_database_lib" );
        return NULL;
    }
    status = ias_db_get_connect_info( &database_name, &user_name,
        &password, &host );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error from "
            "ias_db_get_connect_info" );
        free( database_name );
        free( user_name );
        free( password );
        free( host );
        ias_db_close_database_lib();
        return NULL;
    }

    db = ias_db_connect_to_database( database_name, user_name, password,
        host );
    free( database_name );
    free( user_name );
    free( password );
    free( host );
    if ( !db )
    {
        assert_true_with_message( 0, "Error from "
            "ias_db_connect_to_database" );
        ias_db_close_database_lib();
        return NULL;
    }

    return db;
}  /* END db_open */

/*------------------------------------------------------------------
  NAME:    db_close

  PURPOSE: Close the database connection
------------------------------------------------------------------*/
void db_close( struct ias_db_connection *db )
{
    ias_db_close_connection( db );
    ias_db_close_database_lib();
}

/*------------------------------------------------------------------
  NAME:    db_store_histogram_stats
  PURPOSE: This routine stores the histogram statistics results
           in the database.  It also saves a file of the results
           to be used for ensuring database results were stored
           correctly.
------------------------------------------------------------------*/
int db_store_histogram_stats
(
    IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS *db_results
)
{
    int status;
    struct ias_db_connection *db;
    char db_error_message[IAS_SQL_RTN_MSGLEN+1];/* Query error message buffer */
    struct ias_db_query *query = NULL;    /* IAS DB query handle data */
    char query_string[] = "select radiometric_wo_common_id from "
        "radiometric_work_order_common where work_order_id='HS8899'";
    char insert_string[] = "insert into radiometric_work_order_common "
        "(radiometric_wo_common_id, work_order_id) "
        "values (8899, 'HS8899');";
    char rad_common_id[IAS_CHAR_ID_SIZE];  
    char delete_string[256];
    FILE *ouf;
    int det;

    db = db_open();
    if ( !db )
        return 1;

    status = ias_db_start_transaction( db );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error from "
            "ias_db_start_transaction" );
        db_close( db );
        return 1;
    }

    /* Get the rad common ID */
    query = ias_db_perform_query(db, query_string);
    if ( !ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, db_error_message,
            sizeof(db_error_message));
        assert_true_with_message( 0, "Error from "
            "ias_db_perform_query: %s", db_error_message );
        ias_db_query_close(query);
        db_close( db );
        return 1;
    }
    else
    {
        /* If it isn't there create it */
        if (!ias_db_query_next(query))
        {
            /* Need to close the previous query before doing the insert */
            ias_db_query_close(query);
            query = ias_db_perform_query(db, insert_string );
            if ( ias_db_query_was_successful( query ))
            {
                strcpy(rad_common_id, "8899");
            }
            else
            {
                ias_db_query_get_error_message( query, db_error_message,
                    sizeof(db_error_message));
                assert_true_with_message( 0, "Error from "
                    "ias_db_perform_query: %s", db_error_message );
                ias_db_query_close(query);
                db_close( db );
                return 1;
            }
        }
        /* If it was there use the ID */
        else
        {
            ias_db_query_string_value(query, 0, rad_common_id, 
                sizeof(rad_common_id));
            if ( strcmp(rad_common_id, "") == 0 )
            {
                assert_true_with_message( 0, "Empty string returned from "
                    "ias_db_query_string_value, couldn't get ID" );
                ias_db_query_close(query);
                db_close( db );
                return 1;
            }
        }
    }
    ias_db_query_close(query);

    /* Now clean out the histogram stats table */
    sprintf( delete_string, "delete from histogram_statistics where "
        "radiometric_wo_common_id=%s", rad_common_id );
    query = ias_db_perform_query(db, delete_string);
    if ( !ias_db_query_was_successful(query))
    {
        assert_true_with_message( 0, "Error from "
            "ias_db_perform_query, deleting records from histogram_statistics");
        ias_db_query_close(query);
        db_close( db );
        return 1;
    }
    ias_db_query_close(query);
    
    /* Change the results for the new rad_common_id */
    strncpy(db_results->rad_common_id, rad_common_id, 
        sizeof(db_results->rad_common_id));
    db_results->rad_common_id[sizeof(db_results->rad_common_id) - 1] = '\0';

    ouf = fopen( "expected_hs8899.txt", "w" );
    if ( ouf )
    {
        fprintf( ouf, "%d\n",db_results->num_detectors );
        for ( det = 0; det < db_results->num_detectors; det++ )
        {
            fprintf( ouf, "%s %d %d %d %s 2 %f %f %f %f %f %f %d %f %f\n",
                db_results->rad_common_id, 
                db_results->band_number, 
                db_results->sca, 
                db_results->detector[0][det],
                db_results->processing_level,
                db_results->min[0][det], 
                db_results->max[0][det], 
                db_results->stdev[0][det], 
                db_results->mean[0][det], 
                db_results->skewness[0][det],
                db_results->kurtosis[0][det], 
                db_results->num_values[0][det], 
                db_results->correlation[0][det],
                db_results->mean_sq_resp[0][det]);
        }
        fclose( ouf );
    }


    /* Store the stats in the database */
    db_results->dbase = db;
    status = ias_rps_store_histogram_stats_in_database( db_results );
    if ( status != SUCCESS )
    {
        ias_db_rollback_transaction( db );
        assert_true_with_message( 0, "Error from "
            "ias_rps_store_histogram_stats_in_database" );
    }
    else
    {
        ias_db_commit_transaction( db );
    }

    db_close( db );
    return status;
}  /* END db_store_histogram_stats */

/*------------------------------------------------------------------
  NAME:    db_check_histogram_stats
  PURPOSE: This routine retrieves results from the database and 
           checks them against the values in the expected results
           file.
------------------------------------------------------------------*/
int db_check_histogram_stats
(
    IAS_RPS_DB_HISTOGRAM_CHARACTERIZATION_RESULTS *db_results
)
{
    FILE *inf_db;
    FILE *inf_exp;
    int status;
    int det;
    char db_processing_level[25];
    char exp_processing_level[25];
    float db_correlation;
    float exp_correlation;
    float db_kurtosis;
    float exp_kurtosis;
    float db_max;
    float exp_max;
    float db_mean;
    float exp_mean;
    float db_mean_sq_resp;
    float exp_mean_sq_resp;
    float db_min;
    float exp_min;
    float db_skewness;
    float exp_skewness;
    float db_stdev;
    float exp_stdev;
    int db_line;
    int exp_line;
    int db_band;
    int exp_band;
    int db_num_values;
    int exp_num_values;
    int db_detector;
    int exp_detector;
    char db_rad_common_id[IAS_CHAR_ID_SIZE];  
    char exp_rad_common_id[IAS_CHAR_ID_SIZE];  
    int db_sca;
    int exp_sca;
    int num_detectors;
    int n;

    status = system( "./retrieve_stats.csh" );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error executing database retrieve"
            " script");
        return 1;
    }

    inf_db = fopen( "./hs8899.txt", "r" );
    if ( !inf_db )
    {
        assert_true_with_message( 0, "Couldn't open database results "
            "file: ./hs8899.txt");
        return 1;
    }
    inf_exp = fopen( "expected_hs8899.txt", "r" );
    if ( !inf_exp )
    {
        assert_true_with_message( 0, "Couldn't open expected results");
        fclose( inf_db );
        return 1;
    }

    n = fscanf( inf_exp, "%d\n", &num_detectors );
    if ( n != 1 )
    {
        assert_true_with_message( 0, "Error reading number of detectors"
            "from expected results" );
        fclose( inf_db );
        fclose( inf_exp );
        return 1;
    }

    status = 0;
    for ( det = 0; det < num_detectors; det++ )
    {
        n = fscanf( inf_db, "%s %d %d %d %s %d %f %f %f %f %f %f %d %f %f\n",
            db_rad_common_id, &db_band, &db_sca, &db_detector,
            db_processing_level, &db_line, &db_min, &db_max, &db_stdev,
            &db_mean, &db_skewness, &db_kurtosis, &db_num_values,
            &db_correlation, &db_mean_sq_resp);
        if ( n != 15 )
        {
            assert_true_with_message( 0, "Error reading from database"
                "dump file, n = %d  det = %d", n, det);
            status = 1;
            break;
        }
        n = fscanf( inf_exp, "%s %d %d %d %s %d %f %f %f %f %f %f %d %f %f\n",
            exp_rad_common_id, &exp_band, &exp_sca, &exp_detector,
            exp_processing_level, &exp_line, &exp_min, &exp_max, &exp_stdev,
            &exp_mean, &exp_skewness, &exp_kurtosis, &exp_num_values,
            &exp_correlation, &exp_mean_sq_resp);
        if ( n != 15 )
        {
            assert_true_with_message( 0, "Error reading from expected"
                "results file, n = %d", n);
            status = 1;
            break;
        }

        assert_string_equal( db_rad_common_id, exp_rad_common_id);
        assert_equal( db_band, exp_band);
        assert_equal( db_sca, exp_sca);
        assert_equal( db_detector, exp_detector);
        assert_string_equal( db_processing_level, exp_processing_level );
        if ( fabs( db_min - exp_min ) > 1e-6 )
        {
            assert_true_with_message( 0, "%f should match %f", db_min,
                exp_min );
        }
        if ( fabs( db_max - exp_max ) > 1e-6 )
        {
            assert_true_with_message( 0, "%f should match %f", db_max,
                exp_max );
        }
        if ( fabs( db_stdev - exp_stdev ) > 1e-6 )
        {
            assert_true_with_message( 0, "%f should match %f", db_stdev,
                exp_stdev );
        }
        if ( fabs( db_mean - exp_mean ) > 1e-6 )
        {
            assert_true_with_message( 0, "%f should match %f", db_mean,
                exp_mean );
        }
        if ( fabs( db_skewness - exp_skewness ) > 1e-6 )
        {
            assert_true_with_message( 0, "%f should match %f", db_skewness,
                exp_skewness );
        }
        if ( fabs( db_kurtosis - exp_kurtosis ) > 1e-6 )
        {
            assert_true_with_message( 0, "%f should match %f", db_kurtosis,
                exp_kurtosis );
        }
        assert_equal( db_num_values, exp_num_values);
        if ( fabs( db_correlation - exp_correlation ) > 1e-6 )
        {
            assert_true_with_message( 0, "%f should match %f",
                db_correlation, exp_correlation );
        }
        if ( fabs( db_mean_sq_resp - exp_mean_sq_resp ) > 1e-6 )
        {
            assert_true_with_message( 0, "%f should match %f",
                db_mean_sq_resp, exp_mean_sq_resp );
        }
    }

    /* Close the results files and remove them */
    fclose( inf_db );
    fclose( inf_exp );
    remove( "./hs8899.txt" );
    remove( "./expected_hs8899.txt" );

    return status;
}  /* END db_check_histogram_stats */

/*------------------------------------------------------------------
  NAME:    db_store_impulse_noise
  PURPOSE: This routine stores the impulse noise results
           in the database.  It also saves a file of the results
           to be used for ensuring database results were stored
           correctly.
------------------------------------------------------------------*/
int db_store_impulse_noise
(
    IAS_RPS_DB_IMPULSE_NOISE_CHAR_RESULTS *db_results
)
{
    int status;
    struct ias_db_connection *db;
    char db_error_message[IAS_SQL_RTN_MSGLEN+1];/* Query error message buffer */
    struct ias_db_query *query = NULL;    /* IAS DB query handle data */
    char query_string[] = "select radiometric_wo_common_id from "
        "radiometric_work_order_common where work_order_id='IN9899'";
    char insert_string[] = "insert into radiometric_work_order_common "
        "(radiometric_wo_common_id, work_order_id) "
        "values (9899, 'IN9899');";
    char rad_common_id[IAS_CHAR_ID_SIZE];  
    char delete_string[256];
    FILE *ouf;
    int det;
    int number_nonzero_counts;

    db = db_open();
    if ( !db )
        return 1;

    status = ias_db_start_transaction( db );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error from "
            "ias_db_start_transaction" );
        db_close( db );
        return 1;
    }

    /* Get the rad common ID */
    query = ias_db_perform_query(db, query_string);
    if ( !ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, db_error_message,
            sizeof(db_error_message));
        assert_true_with_message( 0, "Error from "
            "ias_db_perform_query: %s", db_error_message );
        ias_db_query_close(query);
        db_close( db );
        return 1;
    }
    else
    {
        /* If it isn't there create it */
        if (!ias_db_query_next(query))
        {
            /* Need to close the previous query before doing the insert */
            ias_db_query_close(query);
            query = ias_db_perform_query(db, insert_string );
            if ( ias_db_query_was_successful( query ))
            {
                strcpy(rad_common_id, "9899");
            }
            else
            {
                ias_db_query_get_error_message( query, db_error_message,
                    sizeof(db_error_message));
                assert_true_with_message( 0, "Error from "
                    "ias_db_perform_query: %s", db_error_message );
                ias_db_query_close(query);
                db_close( db );
                return 1;
            }
        }
        /* If it was there use the ID */
        else
        {
            ias_db_query_string_value(query, 0, rad_common_id, 
                sizeof(rad_common_id));
            if ( strcmp(rad_common_id, "") == 0 )
            {
                assert_true_with_message( 0, "Empty string returned from "
                    "ias_db_query_string_value, couldn't get ID" );
                ias_db_query_close(query);
                db_close( db );
                return 1;
            }
        }
    }
    ias_db_query_close(query);

    /* Now clean out the impulse noise table */
    sprintf( delete_string, "delete from impulse_noise where "
        "radiometric_wo_common_id=%s", rad_common_id );
    query = ias_db_perform_query(db, delete_string);
    if ( !ias_db_query_was_successful(query))
    {
        assert_true_with_message( 0, "Error from "
            "ias_db_perform_query, deleting records from "
            "impulse_noise");
        ias_db_query_close(query);
        db_close( db );
        return 1;
    }
    ias_db_query_close(query);
    
    /* Change the results for the new rad_common_id */
    strncpy(db_results->rad_common_id, rad_common_id, 
        sizeof(db_results->rad_common_id));
    db_results->rad_common_id[sizeof(db_results->rad_common_id) - 1] = '\0';

    /* Create a file of expected results */
    number_nonzero_counts = 0;
    for ( det = 0; det < db_results->num_detectors; det++ )
    {
        if ( db_results->in_counts[det] > 0 )
            number_nonzero_counts++;
    }

    ouf = fopen( "expected_in9899.txt", "w" );
    if ( ouf )
    {
        fprintf( ouf, "%d\n", number_nonzero_counts );
        for ( det = 0; det < db_results->num_detectors; det++ )
        {
            if ( db_results->in_counts[det] > 0 )
            {
                fprintf( ouf, "%s %d %d %d %d\n",
                    db_results->rad_common_id, 
                    db_results->band, 
                    db_results->sca, 
                    det + 1,
                    db_results->in_counts[det]);
            }
        }
        fclose( ouf );
    }


    /* Store the stats in the database */
    db_results->dbase = db;
    status = ias_rps_store_impulse_noise_counts_in_database( db_results );
    if ( status != SUCCESS )
    {
        ias_db_rollback_transaction( db );
        assert_true_with_message( 0, "Error from "
            "ias_rps_store_histogram_stats_in_database" );
    }
    else
    {
        ias_db_commit_transaction( db );
    }

    db_close( db );
    return status;
}  /* END db_store_impulse_noise */

/*------------------------------------------------------------------
  NAME:    db_check_impulse_noise
  PURPOSE: This routine retrieves results from the database and 
           checks them against the values in the expected results
           file.
------------------------------------------------------------------*/
int db_check_impulse_noise
(
    IAS_RPS_DB_IMPULSE_NOISE_CHAR_RESULTS *db_results
)
{
    FILE *inf_db;
    FILE *inf_exp;
    int status;
    int det;
    int db_band;
    int exp_band;
    int db_detector;
    int exp_detector;
    char db_rad_common_id[IAS_CHAR_ID_SIZE];  
    char exp_rad_common_id[IAS_CHAR_ID_SIZE];  
    int db_sca;
    int exp_sca;
    int db_in_counts;
    int exp_in_counts;
    int num_detectors;
    int n;

    status = system( "./retrieve_in.csh" );
    if ( status != SUCCESS )
    {
        assert_true_with_message( 0, "Error executing database retrieve"
            " script");
        return 1;
    }

    inf_db = fopen( "./in9899.txt", "r" );
    if ( !inf_db )
    {
        assert_true_with_message( 0, "Couldn't open database results "
            "file: ./in9899.txt");
        return 1;
    }
    inf_exp = fopen( "expected_in9899.txt", "r" );
    if ( !inf_exp )
    {
        assert_true_with_message( 0, "Couldn't open expected results");
        fclose( inf_db );
        return 1;
    }

    n = fscanf( inf_exp, "%d\n", &num_detectors );
    if ( n != 1 )
    {
        assert_true_with_message( 0, "Error reading number of detectors"
            "from expected results" );
        fclose( inf_db );
        fclose( inf_exp );
        return 1;
    }

    status = 0;
    for ( det = 0; det < num_detectors; det++ )
    {
        n = fscanf( inf_db, "%s %d %d %d %d\n",
            db_rad_common_id, &db_band, &db_sca, &db_detector,
            &db_in_counts );
        if ( n != 5 )
        {
            assert_true_with_message( 0, "Error reading from database"
                "dump file, n = %d  det = %d", n, det);
            status = 1;
            break;
        }
        n = fscanf( inf_exp, "%s %d %d %d %d\n",
            exp_rad_common_id, &exp_band, &exp_sca, &exp_detector,
            &exp_in_counts);
        if ( n != 5 )
        {
            assert_true_with_message( 0, "Error reading from expected"
                "results file, n = %d", n);
            status = 1;
            break;
        }

        assert_string_equal( db_rad_common_id, exp_rad_common_id);
        assert_equal( db_band, exp_band);
        assert_equal( db_sca, exp_sca);
        assert_equal( db_detector, exp_detector);
        assert_equal( db_in_counts, exp_in_counts);
    }

    /* Close the results files and remove them */
    fclose( inf_db );
    fclose( inf_exp );
    remove( "./in9899.txt" );
    remove( "./expected_in9899.txt" );

    return status;
}  /* END db_check_impulse_noise */
