#include <stdio.h>
#include <string.h>

#include "ias_miscellaneous.h"
#include "ias_pixel_mask.h"
#include "ias_satellite_attributes.h"
#include "ias_db_get_connect_info.h"
#include "ias_db_insert.h"
#include "cgreen/cgreen.h"
#include "../ias_radiometry.h"

Ensure saturated_pixel_test(void);
/* database prototypes */
int db_store_saturated_pixel(IAS_RPS_DB_SATURATED_CHARACTERIZATION_RESULTS *);
int db_check_saturated_pixel(IAS_RPS_DB_SATURATED_CHARACTERIZATION_RESULTS *);
struct ias_db_connection *db_open(void);
void db_close(struct ias_db_connection *);

static int test_database = FALSE;

int main(int argc, char *argv[])
{
    char *enable_db_tests;

    /* See if the database tests should be enabled */
    enable_db_tests = getenv("ENABLE_DB_TESTS");
    if (enable_db_tests)
    {
        test_database = TRUE;
        printf("\n\tDatabase checks have been enabled\n");
    }
    else
        printf("\n\t***** Database checks will not be made *****\n");

    TestSuite *suite = create_test_suite();
    add_test(suite, saturated_pixel_test);

    return run_test_suite(suite, create_text_reporter());
}


Ensure saturated_pixel_test(void)
{
    IAS_PIXEL_MASK *pm;
    IAS_REPORT report_info;
    IAS_REPORT_HEADER *report_header = NULL;
    IAS_RPS_DB_SATURATED_CHARACTERIZATION_RESULTS *db_results;
    char work_order_id[IAS_WORKORDER_ID_SIZE];
    char routine_name[] = "Saturated tests";
    int status;
    void *buffer;
    uint16_t l0r_data[14][8];
    FILE *inf;
    unsigned int temp, det, line;

    int num_detectors = 8;

    int image_data_truncation_setting = 1;
    int digital_low_saturated[8]; 
                            /* I - Digital low saturated level from CPF */
    int digital_high_saturated[8];
                            /* I - Digital high saturated level from CPF */
    int analog_low_saturated[8];  
                            /* I - Analog low saturated level from CPF */
    int analog_high_saturated[8]; 
                            /* I - Analog high saturated level from CPF */
    

    /* Initialize the attributes lib */
    status = ias_sat_attr_initialize(IAS_L8);
    if (status)
    {
        assert_true_with_message(0,
            "ias_sat_attr_initialize returned ERROR");
        return;
    }

    /* Loop through the detector */
    for (det = 0; det < num_detectors; det++)
    {
         digital_low_saturated[det] = 0;
         digital_high_saturated[det] = 4095;
         analog_low_saturated[det] = 2;
         analog_high_saturated[det] = 4000;

        if (image_data_truncation_setting == 0)
        {
             digital_low_saturated[det] >>= 2;
             digital_high_saturated[det] >>= 2;
             analog_low_saturated[det] >>= 2;
             analog_high_saturated[det] >>= 2;
        }
    }

    /* Read in the input testing data (the same as listed in ADD) */
    inf = fopen("testsa.txt", "r");
    if (!inf)
    {
        assert_true_with_message(0, "Error opening: testsa.txt");
        return;
    }
    for (det = 0; det < 8; det++)
    {
        for (line = 0; line < 14; line++)
        {
            fscanf( inf, "%ud", &temp );
            l0r_data[line][det] = (uint16_t)temp;
        }
    }
    fclose(inf);

    /* Create the pixel mask */
    pm = ias_pm_create(1, 1, 8, 14);
    assert_not_equal(pm, NULL);

    /* Add the dropped frames */
    assert_equal( SUCCESS,  ias_pm_add_pixels(pm, 4-1, 4-1, 1,
        IAS_RPS_DROPLINE));

    /* Add the impulse noise */
    assert_equal( SUCCESS, ias_pm_add_pixels(pm, 3-1, 3-1, 1,
        IAS_RPS_IMPNOISE));
    assert_equal( SUCCESS, ias_pm_add_pixels(pm, 5-1, 8-1, 1,
        IAS_RPS_IMPNOISE));
    assert_equal( SUCCESS, ias_pm_add_pixels(pm, 6-1, 9-1, 1,
        IAS_RPS_IMPNOISE));
    assert_equal( SUCCESS, ias_pm_add_pixels(pm, 5-1, 10-1, 1,
        IAS_RPS_IMPNOISE));

    setenv( "PCS_WOID", "I4582", 1 );
    strcpy( work_order_id, "I4582" );
    report_header = ias_misc_initialize_report_header( "", work_order_id,
        routine_name );
    if ( !report_header )
    {
        assert_true_with_message( 0, "Error initializing report header" );
        return;
    }
    ias_misc_set_header_pathrow( report_header, 0, 0 );

    report_info.header = report_header;
    report_info.directory = NULL;

    /* Get a structure for the database */
    db_results = ias_rps_db_allocate_saturated_results(8);
    if (!db_results)
    {
        assert_true_with_message(0, "Error allocating database memory");
        ias_misc_free_report_header( report_header );
        return;
    }

    /* IAS_UI2 */
    buffer = (void *)l0r_data;

    status = ias_rps_saturated_pixel_characterization(
                buffer, 1, 1, 8, 14, digital_low_saturated,
                digital_high_saturated, analog_low_saturated, 
                analog_high_saturated, pm, &report_info, db_results);

    if (status)
    {
        assert_true_with_message( 0, "Error from first "
            "ias_rps_saturated_pixel_characterization" );
        ias_misc_free_report_header( report_header );
        return;
    }


    /* Test the database routine but only if it was specified on the command
       line.  Reason for this is that not all developer schemas will match
       what saturated pixels and the local sql script sets up for column
       names. */
    if (test_database)
    {
        db_results->band = 1;
        db_results->sca = 1;
        if ( db_store_saturated_pixel( db_results ))
        {
            assert_true_with_message(0, "Error storing results to the"
                " database");
        }
        else
        {
            db_check_saturated_pixel(db_results);
        }

        /* Test storing with zero saturated pixels */
        db_results = ias_rps_db_allocate_saturated_results(20);
        assert_true(db_results != NULL);
        if (db_results)
        {
            db_results->band = 1;
            db_results->sca = 1;
            /* set saturated pixel counts to zero for all detectors */
            for(det = 0; det < db_results->num_detectors; det++)
            {
                db_results->detector[det] = det + 1;
                db_results->counts_digital_high[det] = 0;
                db_results->counts_digital_low[det] = 0;
                db_results->counts_analog_high[det] = 0;
                db_results->counts_analog_low[det] = 0;
            }

            if(db_store_saturated_pixel(db_results) != SUCCESS)
            {
                assert_true_with_message(0, "Error attempting to store zero "
                    "saturated pixel results");
            }
        }
    }
    else
    {
        printf("\t***** Database checks skipped *****\n");
        ias_rps_db_free_saturated_results(db_results);
    }

    ias_pm_destroy(pm);
    ias_misc_free_report_header( report_header );
}  /* END saturated_pixel_test */


struct ias_db_connection *db_open(void)
{
    char *database_name = NULL;
    char *user_name = NULL;
    char *password = NULL;
    char *host = NULL;
    struct ias_db_connection *db = NULL;
    int status;

    status = ias_db_initialize_database_lib( );
    if (status)
    {
        assert_true_with_message(0, "Error from "
            "ias_db_initialize_database_lib");
        return NULL;
    }
    status = ias_db_get_connect_info(&database_name, &user_name,
        &password, &host);
    if (status)
    {
        assert_true_with_message(0, "Error from "
            "ias_db_get_connect_info");
        free(database_name);
        free(user_name);
        free(password);
        free(host);
        ias_db_close_database_lib();
        return NULL;
    }

    db = ias_db_connect_to_database(database_name, user_name, password,
        host);
    free(database_name);
    free(user_name);
    free(password);
    free(host);
    if (!db)
    {
        assert_true_with_message(0, "Error from "
            "ias_db_connect_to_database");
        ias_db_close_database_lib();
        return NULL;
    }

    return db;
}

void db_close(struct ias_db_connection *db)
{
    ias_db_close_connection(db);
    ias_db_close_database_lib();
}

int db_store_saturated_pixel
(
    IAS_RPS_DB_SATURATED_CHARACTERIZATION_RESULTS *db_results
)
{
    int status;
    struct ias_db_connection *db;
    char db_error_message[256];  /* Query error message buffer */
    struct ias_db_query *query = NULL;    /* IAS DB query handle data */
    char query_string[] = "select radiometric_wo_common_id from "
        "radiometric_work_order_common where work_order_id='HS8899'";
    char insert_string[] = "insert into radiometric_work_order_common "
        "(radiometric_wo_common_id, work_order_id) "
        "values (8899, 'HS8899');";
    char rad_common_id[IAS_CHAR_ID_SIZE];
    char delete_string[256];
    FILE *ouf;
    int det;

    db = db_open();
    if ( !db )
        return 1;

    status = ias_db_start_transaction(db);
    if (status)
    {
        assert_true_with_message(0, "Error from "
            "ias_db_start_transaction");
        db_close(db);
        return 1;
    }

    /* Get the rad common ID */
    query = ias_db_perform_query(db, query_string);
    if ( !ias_db_query_was_successful(query))
    {
        ias_db_query_get_error_message(query, db_error_message,
            sizeof(db_error_message));
        assert_true_with_message( 0, "Error from "
            "ias_db_perform_query: %s", db_error_message );
        ias_db_query_close(query);
        db_close(db);
        return 1;
    }
    else
    {
        /* If it isn't there create it */
        if (!ias_db_query_next(query))
        {
            /* Need to close the previous query before doing the insert */
            ias_db_query_close(query);
            query = ias_db_perform_query(db, insert_string);
            if ( ias_db_query_was_successful(query))
            {
                strcpy(rad_common_id, "8899");
            }
            else
            {
                ias_db_query_get_error_message(query, db_error_message,
                    sizeof(db_error_message));
                assert_true_with_message(0, "Error from "
                    "ias_db_perform_query: %s", db_error_message);
                ias_db_query_close(query);
                db_close(db);
                return 1;
            }
        }
        /* If it was there use the ID */
        else
        {
            ias_db_query_string_value(query, 0, rad_common_id, 
                sizeof(rad_common_id));
            if ( strcmp(rad_common_id, "") == 0 )
            {
                assert_true_with_message( 0, "Empty string returned from "
                    "ias_db_query_string_value, couldn't get ID" );
                ias_db_query_close(query);
                db_close( db );
                return 1;
            }
        }
    }
    ias_db_query_close(query);

    /* Now clean out the saturated pixels table */
    sprintf(delete_string, "delete from saturated_pixels where "
        "radiometric_wo_common_id=%s", rad_common_id);
    query = ias_db_perform_query(db, delete_string);
    if ( !ias_db_query_was_successful(query))
    {
        assert_true_with_message(0, "Error from "
            "ias_db_perform_query, deleting records from saturated_pixels");
        ias_db_query_close(query);
        db_close(db);
        return 1;
    }
    ias_db_query_close(query);
    
    /* Change the results for the new rad_common_id */
    strncpy(db_results->rad_common_id, rad_common_id, 
        sizeof(db_results->rad_common_id));
    db_results->rad_common_id[sizeof(db_results->rad_common_id) - 1] = '\0';

    ouf = fopen("expected_saturated_hs8899.txt", "w");
    if ( ouf )
    {
        fprintf(ouf, "%d\n",db_results->num_detectors);
        for (det = 0; det < db_results->num_detectors; det++)
        {
            fprintf(ouf, "%s %d %d %d %d %d %d %d\n",
                db_results->rad_common_id, 
                db_results->band, 
                db_results->sca, 
                db_results->detector[det],
                db_results->counts_digital_high[det],
                db_results->counts_digital_low[det],
                db_results->counts_analog_high[det],
                db_results->counts_analog_low[det]);
        }
        fclose( ouf );
    }


    /* Store the stats in the database */
    db_results->dbase = db;
    status = ias_rps_store_saturated_pixels_in_database(db_results);
    if (status)
    {
        ias_db_rollback_transaction(db);
        assert_true_with_message( 0, "Error from "
            "ias_rps_store_saturated_pixels_in_database" );
    }
    else
    {
        ias_db_commit_transaction(db);
    }

    db_close(db);
    return status;
}

int db_check_saturated_pixel
(
    IAS_RPS_DB_SATURATED_CHARACTERIZATION_RESULTS *db_results
)
{
    FILE *inf_db;
    FILE *inf_exp;
    int status;
    int det;
    int db_counts_digital_high;
    int exp_counts_digital_high;
    int db_counts_digital_low;
    int exp_counts_digital_low;
    int db_counts_analog_high;
    int exp_counts_analog_high;
    int db_counts_analog_low;
    int exp_counts_analog_low;
    int db_detector;
    int exp_detector;
    char db_rad_common_id[IAS_CHAR_ID_SIZE];
    char exp_rad_common_id[IAS_CHAR_ID_SIZE];
    int db_band;
    int exp_band;
    int db_sca;
    int exp_sca;
    int num_detectors;
    int n;

    status = system("./retrieve_saturated.csh");
    if (status)
    {
        assert_true_with_message(0, "Error executing database retrieve"
            " script");
        return 1;
    }

    inf_db = fopen("./hs8899.txt", "r");
    if ( !inf_db )
    {
        assert_true_with_message(0, "Couldn't open database results "
            "file: ./hs8899.txt");
        return 1;
    }

    inf_exp = fopen("expected_saturated_hs8899.txt", "r");
    if ( !inf_exp )
    {
        assert_true_with_message(0, "Couldn't open expected results");
        fclose(inf_db);
        return 1;
    }

    n = fscanf(inf_exp, "%d\n", &num_detectors);
    if (n != 1)
    {
        assert_true_with_message(0, "Error reading number of detectors"
            "from expected results");
        fclose(inf_db);
        fclose(inf_exp);
        return 1;
    }

    status = 0;
    for (det = 0; det < num_detectors; det++)
    {
        n = fscanf(inf_db, "%s %d %d %d %d %d %d %d\n",
            db_rad_common_id, &db_band, &db_sca, &db_detector,
            &db_counts_digital_high, &db_counts_digital_low,
            &db_counts_analog_high, &db_counts_analog_low);
        if (n != 8)
        {
            assert_true_with_message(0, "Error reading from database"
                "dump file, n = %d  det = %d", n, det);
            status = 1;
            break;
        }
        n = fscanf(inf_exp, "%s %d %d %d %d %d %d %d\n",
            exp_rad_common_id, &exp_band, &exp_sca, &exp_detector,
            &exp_counts_digital_low, &exp_counts_digital_high,
            &exp_counts_analog_low, &exp_counts_analog_high);
        if (n != 8)
        {
            assert_true_with_message(0, "Error reading from expected"
                "results file, n = %d", n);
            status = 1;
            break;
        }

        assert_string_equal(db_rad_common_id, exp_rad_common_id);
        assert_equal(db_band, exp_band);
        assert_equal(db_sca, exp_sca);
        assert_equal(db_detector, exp_detector);
        assert_equal(db_counts_digital_high, exp_counts_digital_high);
        assert_equal(db_counts_digital_low, exp_counts_digital_low);
        assert_equal(db_counts_analog_high, exp_counts_analog_high);
        assert_equal(db_counts_analog_low, exp_counts_analog_low);
    }

    /* Close the results files and remove them */
    fclose(inf_db);
    fclose(inf_exp);
    remove("./hs8899.txt");
    remove("./expected_saturated_hs8899.txt");

    return status;
}

