#!/usr/bin/perl -w
#------------------------------------------------------------------------------
# NAME: test_dxl.pl
#
# PURPOSE: Test DXL script
#
# PARAMETERS: NONE
#
# OUTPUT: SUCCESS - test completed successfully
#         FAILURE - test failed
#------------------------------------------------------------------------------

use strict;

use ias_const;
use ias_db;
use ias_processing_system;

# function prototypes
sub close_database($);
sub cleanup_copied_files($$$);
sub clean_load_dist_dir_table($);
sub clean_load_processing_instances_table($);


# first check to see if database tests are enabled. If not, quit.
if (!defined $ENV{'ENABLE_DB_TESTS'} )
{
   print "Database tests not enabled. Exiting\n";
   exit SUCCESS;
}

# Clean up the work orders
if (cleanup_for_exit() != SUCCESS)
{
    print "Failed to clean up data\n";
    exit FAILURE;
}

# do some initialization
if (ias_init_system() == ERROR)
{
    print STDERR "Error initializing ias system: $ias_error_msg\n";
    cleanup_for_exit();
    exit FAILURE;
}

# Load the unit test processing instances for "project"
if (clean_load_processing_instances_table(FALSE) != SUCCESS)
{
    print "Failed to load PROCESSING_INSTANCE table\n";
    cleanup_for_exit();
    exit FAILURE;
}

print "Loading Data\n";
if (system("./data/load_work_orders.sh") != 0)
{
    print "Failed loading data\n";
    cleanup_for_exit();
    exit FAILURE;
}

my $l0r_test_storage_location = "./data/l0rp";
my $l1_test_storage_location = "./data/l1";
my $albers_test_storage_location = "./data/l1_lcmap";

# Make sure any files from a previous failed run are cleaned up.
if (cleanup_copied_files($l0r_test_storage_location,
            $l1_test_storage_location,
            $albers_test_storage_location) != SUCCESS)
{
    print "Error removing test files\n";
    cleanup_for_exit();
    exit FAILURE;
}

# Load the unit test "project"
if (clean_load_dist_dir_table(FALSE) != SUCCESS)
{
    print "Failed to load DISTRIBUTION_DIRECTORY table\n";
    cleanup_for_exit();
    exit FAILURE;
}

print "Running Test 1\n";
if (system("../transfer_product ./data/test1") != 0)
{
    print "Test 1 failed\n";
    cleanup_for_exit();
    exit FAILURE;
}

unless (-e
    "$l0r_test_storage_location/2011/024/033/LC80240332001117EDC00_L0R.tar.gz")
{
    print "Test 1 L0r package copy failed\n";
    cleanup_for_exit();
    exit FAILURE;
}

unless (-e
    "$l0r_test_storage_location/2011/024/033/LC80240332001117EDC00_L0R_MD5.txt")
{
    print "Test 1 L0r checksum copy failed\n";
    cleanup_for_exit();
    exit FAILURE;
}

unless (-e
    "$l1_test_storage_location/2011/024/033/LC80240332001117EDC00.tar.gz")
{
    print "Test 1 L1t package copy failed\n";
    cleanup_for_exit();
    exit FAILURE;
}

unless (-e
    "$l1_test_storage_location/2011/024/033/LC80240332001117EDC00_MD5.txt")
{
    print "Test 1 L1t checksum copy failed\n";
    cleanup_for_exit();
    exit FAILURE;
}

# clean up copied files
if (cleanup_copied_files($l0r_test_storage_location,
            $l1_test_storage_location,
            $albers_test_storage_location) != SUCCESS)
{
    print "Error removing test files\n";
    cleanup_for_exit();
    exit FAILURE;
}

# test a second dataset that only has an l0r product
print "Running Test 2\n";
if (system("../transfer_product ./data/test2") != 0)
{
    print "Test 2 failed\n";
    cleanup_for_exit();
    exit FAILURE;
}

# check that copy worked
unless (-e
    "$l0r_test_storage_location/2013/024/033/LC80240332001117EDC00_L0R.tar.gz")
{
    print "Test 2 L0r package copy failed\n";
    cleanup_for_exit();
    exit FAILURE;
}

unless (-e
    "$l0r_test_storage_location/2013/024/033/LC80240332001117EDC00_L0R_MD5.txt")
{
    print "Test 2 L0r checksum copy failed\n";
    cleanup_for_exit();
    exit FAILURE;
}

# clean up copied files
if (cleanup_copied_files($l0r_test_storage_location,
            $l1_test_storage_location,
            $albers_test_storage_location) != SUCCESS)
{
    print "Error removing test files\n";
    cleanup_for_exit();
    exit FAILURE;
}

# test a third dataset that only has an l1 product
print "Running Test 3\n";
if (system("../transfer_product ./data/test3") != 0)
{
    print "Test 3 failed\n";
    cleanup_for_exit();
    exit FAILURE;
}

unless (-e
    "$l1_test_storage_location/2011/024/033/LC00000000000000EDC00.tar.gz")
{
    print "Test 3 L1t package copy failed\n";
    cleanup_for_exit();
    exit FAILURE;
}

unless (-e
    "$l1_test_storage_location/2011/024/033/LC00000000000000EDC00_MD5.txt")
{
    print "Test 3 L1t checksum copy failed\n";
    cleanup_for_exit();
    exit FAILURE;
}

# clean up copied files
if (cleanup_copied_files($l0r_test_storage_location,
            $l1_test_storage_location,
            $albers_test_storage_location) != SUCCESS)
{
    print "Error removing test files\n";
    cleanup_for_exit();
    exit FAILURE;
}

# test a fourth dataset that is an Albers product
print "Running Test 4\n";
if (system("../transfer_product ./data/test4") != 0)
{
    print "Test 4 failed\n";
    cleanup_for_exit();
    exit FAILURE;
}

unless (-e
    "$albers_test_storage_location/2013/165/034/"
  . "LC08_L1GT_165034_20130619_20000622_01_A1.tar.gz")
{
    print "Test 4 L1t package copy failed\n";
    cleanup_for_exit();
    exit FAILURE;
}

unless (-e
    "$albers_test_storage_location/2013/165/034/"
  . "LC08_L1GT_165034_20130619_20000622_01_A1_MD5.txt")
{
    print "Test 4 L1t checksum copy failed\n";
    cleanup_for_exit();
    exit FAILURE;
}

# clean up copied files
if (cleanup_copied_files($l0r_test_storage_location,
            $l1_test_storage_location,
            $albers_test_storage_location) != SUCCESS)
{
    print "Error removing test files\n";
    cleanup_for_exit();
    exit FAILURE;
}

# Clean up the work orders
if (cleanup_for_exit() != SUCCESS)
{
    print "Failed to clean up data\n";
    exit FAILURE;
}

print "Tests completed successfully\n";

exit SUCCESS;

# end main script

#------------------------------------------------------------------------------
# Name: cleanup_copied_files
# Purpose: clean up files populated by the tests.
# Returns: SUCCESS or ERROR
# Parameters:  L0R storage directory
#              L1 storage directory
#              Albers storage directory
#------------------------------------------------------------------------------
sub cleanup_copied_files($$$)
{
    my $l0r_test_storage_location = $_[0];
    my $l1_test_storage_location = $_[1];
    my $albers_test_storage_location = $_[2];
    my $find_command;
    my @find_output;
    my @filenames = ("LC80240332001117EDC00_L0R_MD5.txt",
                     "LC80240332001117EDC00_L0R.tar.gz",
                     "LC00000000000000EDC00_MD5.txt",
                     "LC00000000000000EDC00.tar.gz",
                     "LC80240332001117EDC00_MD5.txt",
                     "LC80240332001117EDC00.tar.gz",
                     "LC08_L1GT_165034_20130619_20000622_01_A1.tar.gz",
                     "LC08_L1GT_165034_20130619_20000622_01_A1_MD5.txt");

    $find_command = "find \"$l0r_test_storage_location\" "
                  . "\"$l1_test_storage_location\" "
                  . "\"$albers_test_storage_location\" -name "
                  .  join (" -or -name ", @filenames);
    @find_output = `$find_command`;
    if (($? >> 8 ) != 0)
    {
        print "Failed to find files to delete\n";
        return ERROR;
    }
    foreach my $file (@find_output)
    {
        chomp($file);
        if (!unlink($file))
        {
            print "Failed to delete $file\n";
            return ERROR;
        }
    }

    return SUCCESS;
}

#------------------------------------------------------------------------------
# Name: clean_load_dist_dir_table
# Purpose: Cleans and (optionally) load the DISTRIBUTION_DIRECTORY table with
#          unit test values;
# Returns: SUCCESS or ERROR
#------------------------------------------------------------------------------
sub clean_load_dist_dir_table($)
{
    my $clean_only = $_[0];

    my $sth;
    my $sql;

    # get the username and password for the WO common database
    my ($user_id, $password) = ias_db_GetDBConnectionInfo("COMMON");

    # connect to the database
    my $dbh = ias_db_ConnectToDB($user_id, $password);
    if (!$dbh)
    {
        print "Error connecting to database\n";
        return ERROR;
    }

    $sql = "DELETE FROM distribution_directory WHERE project  " .
           "LIKE 'UNIT_TEST%' OR PROCESSING_INSTANCE = 'UNIT_TEST'";
    if (! ($sth = $dbh->prepare($sql)))
    {
        print "Error preparing sql\n";
        $dbh->disconnect;
        return ERROR;
    }

    # Get the L0Rp location first
    if (! $sth->execute())
    {
        print "Error executing sql : $DBI::errstr\n";
        $sth->finish;
        $dbh->disconnect;
        return ERROR;
    }
    $sth->finish();

    # Early exit if requested to cleanup only
    if ($clean_only)
    {
        $dbh->commit();
        $dbh->disconnect();
        return SUCCESS;
    }

    # Make sure the instance is defined
    my $instance = $ENV{IAS_INSTANCE_NAME};
    if (!defined $instance)
    {
        print "Error IAS_INSTANCE_NAME not set to valid value";
        $sth->finish;
        $dbh->disconnect;
        return ERROR;
    }

    $sql = "Insert into DISTRIBUTION_DIRECTORY (PROJECT,SENSOR,TIER,DIRECTORY, "
         . "PROCESSING_INSTANCE) values (?,'OLI_TIRS',?,?,?)";
    if (! ($sth = $dbh->prepare($sql)))
    {
        print "Error executing sql : $DBI::errstr\n";
        $dbh->disconnect;
        return ERROR;
    }
    # Add L1 Realtime location
    if (!($sth->execute(("UNIT_TEST", "RT",
        "../../" . $l1_test_storage_location, $instance))))
    {
        print "Error executing sql : $DBI::errstr\n";
        $sth->finish;
        $dbh->disconnect;
        return ERROR;
    }

    # Now add the L1 Albers location
    my $albers_instance = "UNIT_TEST";
    if (!($sth->execute(("LCMAP", "A1",
        "../../" . $albers_test_storage_location, $albers_instance))))
    {
        print "Error executing sql : $DBI::errstr\n";
        $sth->finish;
        $dbh->disconnect;
        return ERROR;
    }

    # Add L0RP location
    if (!($sth->execute(("UNIT_TEST", "L0RP",
        "../../" . $l0r_test_storage_location,
        $instance))))
    {
        print "Error executing sql : $DBI::errstr\n";
        $sth->finish;
        $dbh->disconnect;
        return ERROR;
    }

    $sth->finish();
    $dbh->commit();
    $dbh->disconnect();
    return SUCCESS;
}

#------------------------------------------------------------------------------
# Name: clean_load_processing_instances_table
# Purpose: Cleans and (optionally) load the PROCESSING_INSTANCE table with
#          unit test values for the albers unit test
# Returns: SUCCESS or ERROR
#------------------------------------------------------------------------------
sub clean_load_processing_instances_table($)
{
    my $clean_only = $_[0];

    my $sth;
    my $sql;

    # get the username and password for the WO common database
    my ($user_id, $password) = ias_db_GetDBConnectionInfo("COMMON");

    # connect to the database
    my $dbh = ias_db_ConnectToDB($user_id, $password);
    if (!$dbh)
    {
        print "Error connecting to database\n";
        return ERROR;
    }

    $sql = "DELETE FROM PROCESSING_INSTANCES WHERE INSTANCE_NAME " .
           "= 'UNIT_TEST'";
    if (! ($sth = $dbh->prepare($sql)))
    {
        print "Error preparing sql\n";
        $dbh->disconnect;
        return ERROR;
    }
    if (! $sth->execute())
    {
        print "Error executing sql : $DBI::errstr\n";
        $sth->finish;
        $dbh->disconnect;
        return ERROR;
    }
    $sth->finish();

    # Early exit if requested to cleanup only
    if ($clean_only)
    {
        $dbh->commit();
        $dbh->disconnect();
        return SUCCESS;
    }

    $sql = "Insert into PROCESSING_INSTANCES (INSTANCE_NAME,COLLECTION_NUMBER,"
         . "SETUP_FILE,ENABLED,INVENTORY_DPKG,SUBSETTER_DPKG,SUBSETTER_VIEW,"
         . "PRIORITY_ADJUSTMENT) values ('UNIT_TEST',1,'unit_test_setup_file'," 
         . "'N','dbpkg_l0rp_l1_data','DBPKG_SUBSETTER',"
         . "'SUBSETTER_REQUEST_VIEW',0)";
    if (! ($sth = $dbh->prepare($sql)))
    {
        print "Error executing sql : $DBI::errstr\n";
        $dbh->disconnect;
        return ERROR;
    }

    if (! $sth->execute())
    {
        print "Error executing sql : $DBI::errstr\n";
        $sth->finish;
        $dbh->disconnect;
        return ERROR;
    }

    $sth->finish();
    $dbh->commit();
    $dbh->disconnect();
    return SUCCESS;
}

#------------------------------------------------------------------------------
# Name: cleanup_for_exit
# Purpose: Cleans up the data that was inserted into the database
# Returns: SUCCESS or ERROR
#------------------------------------------------------------------------------
sub cleanup_for_exit
{
    my $ret_status = SUCCESS;

    print "Cleaning up Data\n";

    # Clean up the work orders
    if (system("./data/delete_work_orders.sh") != 0)
    {
        print "Failed to clean up data\n";
        $ret_status = ERROR;
    }

    # Clean the unit test "project"
    if (clean_load_dist_dir_table(TRUE) != SUCCESS)
    {
        print "Failed to clean DISTRIBUTION_DIRECTORY table\n";
        $ret_status = ERROR;
    }
    if (clean_load_processing_instances_table(TRUE) != SUCCESS)
    {
        print "Failed to clean PROCESSING_INSTANCES table\n";
        $ret_status = ERROR;
    }

    return $ret_status;
}
