#!/usr/bin/perl
use warnings;
use strict;

####
# verify_tc002_inv_data.pl
#
# Usage: verify_tc002_inv_data.pl <scene_id> [environment (it|st)]
#
# <scene_id> is the LDCM scene ID of the scene to verify the inventory
# database values for.
#
# This script will query the inventory for the data items that are entered to
# it by the LPGS (and underlying IAS) processing system. Those items are
# verified against a benchmarked file of the same information, expected to be
# located in the $IAS_IT_TEST_DATA_DIR/test_case_002 directory
####

use lib '../tools';     # to locate the get_inv_data module
use get_inv_data;       # shared module to get the inventory data
use ias_const;          # definition of ERROR

# Build a hash of the fields we are interested in checking for this test
# from the inventory metadata populated by LPGS
my %inventoryFieldsToCheck = (
        L0RP_LOCATION => 1,
        L0RP_CHECKSUM_FILE => 1,     # hash
        L0RP_FILE => 1,     # hash
        HOSTNAME => 1,
        SUBSETTER_VERSION_L0RP => 1,
        SCENE_BANDS_L0RP => 1
        );

# Get the scene ID
my $numArgs = $#ARGV + 1;
die ("Please provide an input scene ID.\n") if ($numArgs < 1);

my $landsatSceneId = $ARGV[0];

my $environment = "it";
if ($numArgs == 2 && $ARGV[1] eq "st")
{
    $environment = "st";
}

my %inventorySceneData = ();
my $outFile = $landsatSceneId . "_inventory_data.txt";

# Call the routine provided by the get_inv_data module to query for all data
# LPGS (and IAS) put into the inventory
my $status = get_data_by_scene_id($landsatSceneId, \%inventorySceneData);
if ($status == ERROR)
{
    print "Error getting data from inventory: $get_inv_data::error_message\n";
    exit 1;
}

# Write the inventory data to a file for verification
open (FILE, ">$outFile") || die "Error opening file $outFile for writing\n";

my $l0rpLocation = '';
my $l0rpFileName = '';
my $checksumFileDate = '';
my $l0rpDataFileDate = '';
my $l0rpDataFileSize = '';
my $l0rpDataProductFileSize = '';
my $hostname = '';
my $subsetterVer = '';

foreach my $invField (sort keys %inventorySceneData)
{
    if (exists $inventoryFieldsToCheck{$invField})
    {
        # Print the field value - handle some differently. L0RP_LOCATION needs
        # to be generalized to work with diff, the L0RP_CHECKSUM_FILE and
        # L0RP_DATA_FILE fields are actually hashes that need their contents
        # all printed
        if ($invField =~ /L0RP_LOCATION/)
        {
            # Save the L0RP_LOCATION to print to the screen for the tester
            $l0rpLocation = $inventorySceneData{$invField};

            # Save it for the inventory dump file diff too, but generalize the
            # storage node part of the path so that doesn't cause diffs.
            # Geralize path when on the dev system too so diff can work across
            # systems.
            my $l0rpLocationForFile = $inventorySceneData{$invField};
            $l0rpLocationForFile =~ s/^\/sno\d\d\/$environment\/l0rp\//\//;
            $l0rpLocationForFile
                =~ s/^\/pcachedev\/lpgs\/\w*\/online_cache\//\//;

            # Print the field name and generalized value
            print FILE "$invField: ";
            print FILE "$l0rpLocationForFile\n";
        }
        elsif ($invField =~ /L0RP_FILE/)
        {
            foreach my $innerField (sort keys %{$inventorySceneData{$invField}})
            {
                # Don't need to check the file type field
                next if ($innerField =~ /FILE_TYPE/);
                # Can't compare the checksum, as it will be different due to
                # internal processing date information that will differ each
                # time
                next if ($innerField =~ /FILE_CHECKSUM/);

                if ($innerField =~ /FILE_DATE/)
                {
                    # Save to print to screen later
                    $l0rpDataFileDate
                        = $inventorySceneData{$invField}{$innerField};
                }
                elsif ($innerField =~ /PRODUCT_FILE_SIZE/)
                {
                    # File size differs each run (date contents?)
                    # Save to print to screen later
                    $l0rpDataProductFileSize
                        = $inventorySceneData{$invField}{$innerField};
                }
                elsif ($innerField =~ /FILE_SIZE/)
                {
                    # File size differs each run (date contents?)
                    # Save to print to screen later
                    $l0rpDataFileSize
                        = $inventorySceneData{$invField}{$innerField};
                }
                else
                {
                    if ($innerField =~ /FILE_NAME/)
                    {
                        $l0rpFileName
                            = $inventorySceneData{$invField}{$innerField};
                    }

                    # Print the field name and value
                    print FILE "L0RP_$innerField: ";
                    print FILE $inventorySceneData{$invField}{$innerField}
                          . "\n";
                }
            }
        }
        elsif ($invField =~ /L0RP_CHECKSUM_FILE/)
        {
            foreach my $innerField (sort keys %{$inventorySceneData{$invField}})
            {
                # A few fields that aren't relevant for this file type or
                # don't need to check
                next if ($innerField =~ /FILE_TYPE/);
                next if ($innerField =~ /FILE_CHECKSUM/);
                next if ($innerField =~ /PRODUCT_FILE_SIZE/);

                if ($innerField =~ /FILE_DATE/)
                {
                    # Save to print to screen later
                    $checksumFileDate
                        = $inventorySceneData{$invField}{$innerField};
                }
                else
                {
                    # Print the field name and value
                    print FILE "L0RP_CHECKSUM_$innerField: ";
                    print FILE $inventorySceneData{$invField}{$innerField}
                          . "\n";
                }
            }
        }
        else
        {
            if ($invField =~ /HOSTNAME/)
            {
                # Save to print to screen later
                $hostname = $inventorySceneData{$invField};
            }
            elsif ($invField =~ /SUBSETTER_VERSION_L0RP/)
            {
                # Save to print to screen later
                $subsetterVer = $inventorySceneData{$invField};
            }
            else
            {
                # Print the field name and value
                print FILE "$invField: ";
                print FILE $inventorySceneData{$invField} . "\n";
            }
        }
    }
}
close FILE;

# Verify the inventory dump file matches the benchmarked one
my $sysCmd = "diff -q " . $ENV{IAS_IT_TEST_DATA_DIR}
           . "/test_case_002/$outFile $outFile";
my $rc = system($sysCmd);
if ($rc != 0)
{
    print "Diff failed: $sysCmd\n";
    exit 1;
}

# Everything was successful, so remove the file, tell the tester about the
# results, and print out the L0RP_LOCATION for the tester
unlink $outFile;

print "\nInventory data automated verification successful\n\n";

print "Inventory data for manual verification:\n";
if ($l0rpLocation ne '')
{
    print "L0RP_LOCATION: $l0rpLocation\n";
}
else
{
    print "Error retrieving the L0RP_LOCATION\n";
}

# Print other values that cannot be automatically verfied, for manual checking
print "L0RP_PACKAGE_NAME: $l0rpFileName\n";
print "SUBSETTER_VERSION_L0RP: $subsetterVer\n";
print "L0RP_CHECKSUM_FILE_DATE: $checksumFileDate\n";
print "L0RP_PACKAGE_FILE_DATE: $l0rpDataFileDate\n";
print "L0RP_PACKAGE_FILE_SIZE: $l0rpDataFileSize\n";
print "L0RP_PRODUCT_FILE_SIZE: $l0rpDataProductFileSize\n";
print "HOSTNAME: $hostname\n";

exit 0;
