#!/bin/csh -f
#***************************************************************************
# NAME:     verify_tc003_valid_flag
#
#  PURPOSE:  Verifies that the radiometric work order common ''Valid'' flags
#            are set correctly
#  RETURNS:
#        Type:  Integer
#        Value: 0 / 1
#
#  USAGE:    ./verify_tc003_valid_flag.csh G|R product_request_WOs.txt
#                  optional {product_request_secondary_WOs.txt}
#****************************************************************************
# Create a log file
if ( -e test_case_id ) then
    set tcnum = `cat test_case_id`
    set log = tc${tcnum}.log
    set woid_file = tc${tcnum}_woids.txt
else
    set script = $0:t
    set log = ${script:r}.out
    set woid_file = work_order_id.txt
endif

# Must have one argument
if ( ${#argv} < 2 ) then
    echo "" |& tee -a $log
    echo "    Usage: $0 G|R product_request_WOs.txt {optional" \
        "product_request_secondary_WOs.txt}" |& tee -a $log
    echo "" |& tee -a $log
    exit 1
else if (${#argv} == 2) then
    set expected_invalid_records = 0
    set product_req_ids = `awk '{print $1}' $2`
else
    # The number of invalid records should match the number of secondary
    # scene ran through, because it will set the previous records to invalid
    set expected_invalid_records = `wc -l < secondary_product_request_WOs.txt`
    set product_req_ids = `awk '{print $1}' $2; awk '{print $1}' $3`
endif

# Retrieve the work order ids
set woids = ""
set first_id = 1
foreach pid ($product_req_ids)
    set id = `sqlplus -s ${IAS_DB_COM}` << END
        set head off;
        SELECT work_order_id FROM work_orders
            WHERE product_request_id='$pid';
        exit;
END

    if ( $first_id ) then
        set woids = `echo "$woids '$id'"`
        set first_id = 0
    else
        set woids = `echo "$woids, '$id'"`
    endif
end


# Check whether to query RADIOMETRIC_WORK_ORDER_COMMON or
# GEOMETRIC_WORK_ORDER_COMMON table for valid flags
set table_flag = $1

if ( $table_flag == G ) then
    set table = "GEOMETRIC_WORK_ORDER_COMMON"
else if ( $table_flag == R ) then
    set table = "RADIOMETRIC_WORK_ORDER_COMMON"
else
    echo "" |& tee -a $log
    echo "Table does not exist."
    echo "    Usage: $0 G|R product_request_WOs.txt {optional" \
        "product_request_secondary_WOs.txt}" |& tee -a $log
    echo "" |& tee -a $log
    exit 1
endif

# Using the list of work order ids count the number of valid and
# invalid work orders and verify that it is correct
set valid_wos = `sqlplus -s ${IAS_DB_L8_OLITIRS}` << END
    set head off;
    SELECT count(*) FROM $table
    WHERE work_order_id in ($woids)
    AND VALID = 'Y';
    exit;
END

set invalid_wos = `sqlplus -s ${IAS_DB_L8_OLITIRS}` << END
    set head off;
    SELECT count(*) FROM $table
    WHERE work_order_id in ($woids)
    AND VALID = 'N';
    exit;
END

echo ""
if ( $valid_wos != 13 || $invalid_wos != $expected_invalid_records ) then
    echo "ERROR: The number of valid records is incorrect"
    echo ""
    echo "There were $valid_wos Valid records where (13) were expected"
    echo "There were $invalid_wos Invalid records where"\
            "($expected_invalid_records) were expected"
    echo ""
    exit 1

else
    echo "SUCCESS: All validity flags have been verified"
    echo ""
    echo "There were $valid_wos Valid records and $invalid_wos Invalid records"
    echo ""
    exit 0
endif
