#!/usr/bin/perl
use warnings;
use strict;

# Description:
#    This script checks file and directory permissions.  It is used for LPGS
#    test case 007.
#
# Usage:
#    verify_permissions.pl <Work Order ID> <test directory>

use File::Find;    # Recursively traverse directory trees
use File::stat;    # Get file attributes like permissions

use constant SUCCESS => 0;
use constant ERROR => 1;

#command line parameters
my $work_order_id;  # <work order id> - Work Order Id to verify
my $test_directory; # <test directory> - Directory with configuration data

# configuration
my $permissions_filename = "permissions.config";

# variables
my $work_order_directory = $ENV{IAS_WO};
my $data_directory = $ENV{IAS_DATA_DIR};
my @all_permissions;
my $status;

#### Subroutine Declarations ####
sub check_File_Permissions();
sub load_permissions($);

#### MAIN ####

# Verify two required command line parameters were provided
if ($#ARGV < 1)
{
    die("Usage: verify_permissions.pl <WO ID> <test directory>");
}

$work_order_id = $ARGV[0];
$test_directory = $ARGV[1];

# Verify the work order directory exists
if (!(-d "$work_order_directory/$work_order_id"))
{
    die("Error: Invalid IAS_WO environment variable or Work Order ID.");
}

# Verify the IAS data directory exists
if (!(-d "$data_directory"))
{
    die("Error: IAS_DATA_DIR environment variable not set correctly.");
}

# Verify the test directory path exists
if (!(-d $test_directory))
{
    die("Error: Invalid test directory.\nUsage: verify_permissions.pl " .
        "<WO ID> <test directory>");
}

# Verify permissions configuration file exists
if (!(-f "$test_directory/$permissions_filename"))
{
    die("$test_directory/$permissions_filename not found.");
}

$status = load_permissions("$test_directory/$permissions_filename");
if ($status != SUCCESS)
{
    die("Error: loading permission configuration file.");
}

# Check permissions in the specified directory including subdirectories
find(\&check_File_Permissions, $data_directory);

#### SUBROUTINES ####

##############################################################################
# Name: check_File_Permissions
# Description: This is the "wanted" subroutine for the Perl File::Find.
#   It verifies the permissions of a single file or directory.
##############################################################################
sub check_File_Permissions()
{
    my $filename = $_;
    my $file_attributes;
    my $name;
    my $permissions;
    my $search_type;

    # In the wo directory, only look at the specified work order
    if ((index($File::Find::name, '/wo/') != -1) &&
        (index($File::Find::name, $work_order_id) == -1))
    {
        return;
    }

    # There is a possibility that a "dead" symbolic link could be
    # encountered.  In that case, the attributes won't be defined.
    # So, if the attributes aren't defined, just issue an error
    # message and return.  This will prevent the script from terminating
    # prematurely.
    $file_attributes = stat($File::Find::name);
    if (not defined $file_attributes)
    {
        printf "Attributes for %s are not defined\n", $File::Find::name;
        return;
    }

    # If it doesn't match the regular symbolic link, file, and directory
    # permissions, see if it matches the special cases in the config file
    if (((-l $filename) && ($file_attributes->mode & 00777) != 0775) ||
       ((-f $filename) && ($file_attributes->mode & 00777) != 0664) ||
       ((-d $filename) && ($file_attributes->mode & 00777) != 0775))
   {
       foreach my $permission (@all_permissions)
       {
           $name = $permission->{"name"};
           $permissions = oct($permission->{"permissions"});
           $search_type = $permission->{"search_type"};
            if ($search_type eq 'name')
           {
               if ($filename eq $name)
               {
                   if (($file_attributes->mode & 00777) != $permissions)
                   {
                       printf "%s has permissions %04o but should have %04o\n",
                           $File::Find::name, $file_attributes->mode & 00777,
                           $permissions;
                   }
               }
           }
           else # search type is 'substring'
           {
               if (index($filename, $name) != -1)
               {
                   if (($file_attributes->mode & 00777) != $permissions)
                   {
                       printf "%s has permissions %04o but should have %04o\n",
                           $File::Find::name, $file_attributes->mode & 00777,
                           $permissions;
                   }
               }
           }
       }
   }
}

##############################################################################
# Name: load_permissions
# Description: Accepts the permission configuration file as a parameter.
#   Loads the non-standard permissions for symbolic links, regular files, and
#   directories.
# Returns ERROR or SUCCESS
##############################################################################
sub load_permissions($)
{
    my ($filename) = @_;

    if (!open(FH, "<$filename"))
    {
        print "ERROR: couldn't open permissions config file: $filename\n";
        return ERROR;
    }

    while(<FH>)
    {
        chomp;

        # Skip comments and blank lines
        if ($_ =~ /^\s*(#|$)/) { next; }

        # Trim whitespace from the beginning and end of the line
        $_ =~ s/(^\s+|\s+$)//g;

        # Get the columns of the line
        my ($name, $permissions, $search_type) = split(/\|/, $_);

        my $row= {
            name => $name,
            permissions => $permissions,
            search_type => $search_type
        };

        # Create hash for all files' permissions
        push(@all_permissions, $row);
    }

    close(FH);

    return SUCCESS;
}
