package db_utils;

##############################################################################
#
# This module contains a set of database utilities for accessing the IAS
# database.
#
##############################################################################

use strict;
use vars qw(@ISA @EXPORT @EXPORT_OK %EXPORT_TAGS $VERSION);
use Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(
            db_GetConnectionInfo
            db_Connect
            db_Disconnect
            db_FetchAll
            db_FetchRow
            db_FetchOne
);
$VERSION = '0.01';

use DBI;

##############################################################################
# Name: db_GetConnectionInfo
#
# Description: Get the database username and password.
#
# Parameters:
#   database schema
#
# Returns:
#   username and password
##############################################################################
sub db_GetConnectionInfo(@)
{
    my ($schema) = @_;

    # default to Oracle OS authentication
    my $username = "/";
    my $password = "";
    my $conn_str;

    # default to the common schema if non was provided
    $schema = "COMMON" if not defined $schema;

    if ($schema eq "COMMON" && (defined $ENV{IAS_DB_COM}))
    {
        $conn_str = $ENV{IAS_DB_COM};
    }
    elsif ($schema eq "L8_OLITIRS" && (defined $ENV{IAS_DB_L8_OLITIRS}))
    {
        $conn_str = $ENV{IAS_DB_L8_OLITIRS};
    }
    elsif ($schema eq "L7_ETM" && (defined $ENV{IAS_DB_L7_ETM}))
    {
        $conn_str = $ENV{IAS_DB_L7_ETM};
    }
    elsif ($schema eq "L5_TM" && (defined $ENV{IAS_DB_L5_TM}))
    {
        $conn_str = $ENV{IAS_DB_L5_TM};
    }
    elsif ($schema eq "L4_TM" && (defined $ENV{IAS_DB_L4_TM}))
    {
        $conn_str = $ENV{IAS_DB_L4_TM};
    }
    else
    {
        # return the default OS auth values
        return ($username, $password);
    }

    ($username, $password) = split '/', $conn_str;
    return ($username, $password);
}

##############################################################################
# Name: db_Connect
#
# Description: Creates a connection to the database.
#
# Parameters:
#   DB username
#   DB password
#   Hash reference with connection attributes (optional - AutoCommit and
#       PrintError disabled by default)
#
# Returns:
#   An open database handle if successful
#   undef if not successful
##############################################################################
sub db_Connect(@)
{
    my ($username, $password, $attr_ref) = @_;
    my $sid = $ENV{ORACLE_SID};

    # if connect attributes weren't provided, disable AutoCommit, PrintError,
    # and RaiseError by default.
    if (not defined $attr_ref)
    {
        $attr_ref = {AutoCommit => 0, PrintError => 0, RaiseError => 0};
    }

    my $dbh = DBI->connect("DBI:Oracle:$sid", $username, $password, $attr_ref);
    if (!$dbh)
    {
        return undef;
    }

    return $dbh;
}

##############################################################################
# Name: db_Disconnect
#
# Description: Closes the database handle. 
#
# Parameters:
#   a database handle
#
# Returns:
#   true value when successful, false value when not successful
##############################################################################
sub db_Disconnect($)
{
    my ($dbh) = @_;
    return $dbh->disconnect();
}

##############################################################################
# Name: db_FetchAll
#
# Description: Fetch the entire result for the SQL statement
#
# Parameters:
#   a database handle
#   SQL statement
#   array reference contains bind param values (optional)
#
# Returns:
#   array of row arrays if successful
#   undef if not successful
##############################################################################
sub db_FetchAll(@)
{
    my ($dbh, $sql, $key_field, $bind_ref) = @_;

    # key field must be uppercase
    $key_field = uc($key_field);

    if (not defined $key_field)
    {
        return undef;
    }

    # parse and validate the statement
    my $sth = $dbh->prepare($sql);
    if (!$sth)
    {
        return undef;
    }

    # bind parameter values if they were provided
    if (defined $bind_ref)
    {
        my $size = @{$bind_ref};
        for (my $i = 0; $i < $size; $i++)
        {
            $sth->bind_param($i + 1, $bind_ref->[$i]);
        }
    }

    # execute the query
    if (!$sth->execute())
    {
        return undef;
    }
    
    # retrieve the results
    my $result = $sth->fetchall_hashref($key_field);

    return $result;
}

##############################################################################
# Name: db_FetchRow
#
# Description: Fetch the first row returned by SQL statement.
#
# Parameters:
#   a database handle
#   SQL statement
#   array reference contains bind param values (optional)
#
# Returns:
#   row array if successful
#   undef if not successful
##############################################################################
sub db_FetchRow(@)
{
    my ($dbh, $sql, $bind_ref) = @_;

    # parse and validate the statement
    my $sth = $dbh->prepare($sql);
    if (!$sth)
    {
        return undef;
    }
    
    # bind parameter values if they were provided
    if (defined $bind_ref)
    {
        my $size = @{$bind_ref};
        for (my $i = 0; $i < $size; $i++)
        {
            $sth->bind_param($i + 1, $bind_ref->[$i]);
        }
    }

    # execute the query
    if (!$sth->execute())
    {
        return undef;
    }
    
    # retrieve the results
    return $sth->fetchrow_hashref();
}

##############################################################################
# Name: db_FetchOne
#
# Description: Fetch the value of the first column of the first row returned 
#   by SQL statement.
#
# Parameters:
#   a database handle
#   SQL statement
#   array reference contains bind param values (optional)
#
# Returns:
#   scalar value if successful
#   undef if not successful
##############################################################################
sub db_FetchOne(@)
{
    my ($dbh, $sql, $bind_ref) = @_;

    # parse and validate the statement
    my $sth = $dbh->prepare($sql);
    if (!$sth)
    {
        return undef;
    }

    # bind parameter values if they were provided
    if (defined $bind_ref)
    {
        my $size = @{$bind_ref};
        for (my $i = 0; $i < $size; $i++)
        {
            $sth->bind_param($i + 1, $bind_ref->[$i]);
        }
    }

    # execute the query
    if (!$sth->execute())
    {
        return undef;
    }
    
    my @row;
    if (@row = $sth->fetchrow_array())
    {
        return $row[0];
    }
    else
    {
        return undef;
    }
}
