#!/usr/bin/perl
###############################################################################
# Name: verify_WOs.pl
#
# Description: This script verifies the requirements associated with LPGS
#   Test Case 009: Volume, Metrics, and Latency. The metrics requirements are 
#   checked by verifing each required metric is stored with the database. To
#   verify the latency requirement the script takes the time PWC starts 
#   processing the WO to the actual completion time and adds the worst case
#   times for subsetter_interface and PWS based on their polling intervals. 
#   Volume requirements are verified by calculating the total time required 
#   by the number of products in the requirement and determine the existing 
#   processing node and maximum load setup will meet that requirement.  If 
#   the year 3 volume requirements (IP4104 and IP4105) are not currently 
#   met, the number of additional processing nodes required is printed.
#   
# Requirements Verified:
#   IP1942 - The LPGS shall have the capability to produce 550 L0Rp data 
#            products per day, subject to availability requirements.
#   IP1943 - The LPGS shall have the capability to produce 900 Level 1 data 
#            products per day, subject to availability requirements.
#   IP1967 - The LPGS shall collect metrics on the size of each scene output 
#            product.
#   IP2377 - The LPGS shall collect metrics on the LPGS processing start and 
#            completion times.
#   IP2378 - The LPGS shall collect metrics on the LPGS processing start and 
#            stop times of each step within the processing flows.
#   IP2379 - The LPGS shall collect metrics on the time of order submittal 
#            during LPGS processing.
#   IP2382 - The LPGS shall collect metrics on the size of each input product.
#   IP2383 - The LPGS shall collect metrics on the number of orders received.
#   IP2384 - The LPGS shall collect metrics on the LPGS work orders that 
#            complete successfully.
#   IP2387 - The LPGS shall include priority information in job related 
#            processing metrics.
#   IP3780 - The LPGS shall make a Level 1 data product available for 
#            collected data within the time allocated by the USGS Landsat 
#            Data Continuity Mission (LDCM) Data Processing and Archive 
#            System (DPAS) Subsystem Performance Requirements Allocations 
#            (LDCM-REF-028), consistent with availability requirements.
#   IP3895 - The LPGS shall allow MDS to pull metrics.
#   IP4104 - The LPGS shall have the capability to expand to produce at least
#             600 L0Rp data products per day in the third year after launch.
#   IP4105 - The LPGS shall have the capability to expand to produce at least
#            1650 Level 1 data products per day in the third year after launch.
#
# Usage:
#   verify_WOs.pl --infile=<order_list> [--logfile=<log_filename>]
#       <order_list> is a file contains a newline deliniated list of product
#           request Ids. The file is the output file from the 
#           tram_mass_loader.pl script.
#       <log_filename> is an optional parameter.  If provided, more detailed 
#           messages will be written to the specified log file.
##############################################################################

use warnings;
use strict;

use lib '../tools';
use get_inv_data;
use db_utils;

use ias_const;
use ias_wo_const;
use IO::File;
use Getopt::Long;
use POSIX;

# Subroutine declarations
sub log_msg(@);

# Processing time allocated to LPGS. Defined in LDCM-REF-028
use constant MAX_LPGS_PROCESS_HOURS => 4.00;

# Scenes per day requirement
my $L1_VOLUME_REQ_1 = 900;  # IP 1943
my $L1_VOLUME_REQ_2 = 1650; # IP 4105
my $L0_VOLUME_REQ_1 = 550;  # IP 1942
my $L0_VOLUME_REQ_2 = 600;  # IP 4104

# Production system expected maximum load setup
use constant EXPECTED_PROD_PROCESS_NODES  => 5;
use constant EXPECTED_PROD_MAX_LOAD       => 6;

my $wo_filename;
my $log_filename;
my $verify_l0;
my $verify_l1;
my $l0_found = FALSE;
my $l1_found = FALSE;
my $total_l0_products = 0;
my $total_l1_products = 0;

my $result = GetOptions(
    "infile=s"  => \$wo_filename,
    "logfile=s" => \$log_filename
);

# Define the usage string
sub usage { print "Usage: verify_WOs.pl --infile=s --logfile=s\n"; }

if (not defined $wo_filename)
{
    print "Missing input file.\n";
    usage();
    exit 1;
}

# verify the WO list file exists
if (! -e $wo_filename)
{
    print "ERROR: $wo_filename does not exist\n";
    exit 1;
}

# open the input file contains the scene list
my $fh = IO::File->new($wo_filename, "r");
if (not defined $fh)
{
    print "ERROR: opening $wo_filename\n";
    exit 1;
}

# read the list of product requests
my @pr_list = ();
while (<$fh>)
{
    chomp;
    if ($_ =~ /^[\s]*$/) { next; }; # skip blank lines

    push(@pr_list, $_);
}
$fh->close();

# open the log file
my $log_fh = undef;
if (defined $log_filename)
{
    $log_fh = IO::File->new($log_filename, "w");
    if (not defined $log_fh)
    {
        print "ERROR: opening log file $log_filename\n";
        exit 1;
    }
}

# get the total number of requests submitted
my $total_in_products = scalar(@pr_list);

# get the database username/password and open a DB connection
my ($username, $password) = db_GetConnectionInfo();
my $dbh = db_Connect($username, $password);
if (not defined $dbh)
{
    print "Error: connecting to the LPGS database\n";
    exit 1;
}

# Get the max LPGS overhead before a request begins processing.  This will be
# add to the total processing time for each scene to calculate the latency.
# The overhead accounts for 2 waits on the DSI (one requesting data from the 
# subsetter and one for retrieving the subsetter finished message) and 1 wait
# on the PWS to start the work order. Divide by 3600 to covert seconds to hours
my $sql = "SELECT ((2 * dsi_polling_interval "
       . "        + work_order_scheduler_interval) / 3600) wait_overhead "
       . "  FROM sys_parms";
my $wait_overhead = db_FetchOne($dbh, $sql);
if (not defined $wait_overhead)
{
    print "Error: loading wait overhead.\n";
    print "    -> $DBI::errstr\n";
    db_Disconnect($dbh);
    exit 1;
}

my $total_l1_process_time = 0;
my $total_l0_process_time = 0;
my $total_db_products = 0;
my $error_count = 0;
for my $product_request_id (@pr_list)
{
    log_msg("Verifying $product_request_id...\n", TRUE);
    $total_db_products++;

    my $work_order_id;
    my $landsat_scene_id;

    # Load the work order for this product request
    $sql = "SELECT work_order_id, state, priority, scene_id, procedure_id, "
         . "   to_char(date_entered,'YYYY-MM-DD HH24:MI:SS') enter_date, "
         . "   to_char(req_start_date,'YYYY-MM-DD HH24:MI:SS') start_date, "
         . "   to_char(act_compl_date,'YYYY-MM-DD HH24:MI:SS') comp_date "
         . "  FROM work_orders "
         . " WHERE product_request_id = '$product_request_id'";

    my $wo_ref = db_FetchRow($dbh, $sql);
    if (not defined $wo_ref)
    {
        log_msg("Error: loading work_orders info for $product_request_id:\n", 
                TRUE);
        log_msg("    -> $DBI::errstr\n", TRUE);
        $error_count++;
        next;
    }
    $work_order_id = $wo_ref->{"WORK_ORDER_ID"};
    $landsat_scene_id = $wo_ref->{"SCENE_ID"};

    # Load the script metrics information
    $sql = "SELECT wo.script_id, wo.seq_nbr, wo.compl_status, "
         . "       decode(a.status, 'START', 'Y', 'N') started, "
         . "       decode(b.status, 'SUCCESS', 'Y', 'N') success, "
         . "       decode(c.status, 'ABORTED', 'Y', 'N') aborted "
         . "  FROM wo_scripts wo, "
         . "       (SELECT component, status "
         . "          FROM work_order_log "
         . "         WHERE work_order_id = '$work_order_id' "
         . "           AND status = 'START') a, "
         . "       (SELECT component, status "
         . "          FROM work_order_log " 
         . "         WHERE work_order_id = '$work_order_id' "
         . "           AND status = 'SUCCESS') b, "
         . "       (SELECT component, status "
         . "          FROM work_order_log " 
         . "         WHERE work_order_id = '$work_order_id' "
         . "           AND status = 'ABORTED') c "
         . " WHERE wo.work_order_id = '$work_order_id' "
         . "   AND wo.script_id = a.component(+) "
         . "   AND wo.script_id = b.component(+) "
         . "   AND wo.script_id = c.component(+) "
         . " ORDER BY seq_nbr ASC";

    my $wo_scripts_ref = db_FetchAll($dbh, $sql, "SCRIPT_ID");
    if (not defined $wo_scripts_ref)
    {
        log_msg("Error: loading wo_scripts info for $work_order_id:\n", TRUE);
        log_msg("    -> $DBI::errstr\n", TRUE);
        $error_count++;
        next;
    }

    # Need to add the subsetter interface and transfer product components to
    # the work order log metrics info collected. While they are not really part
    # of the LPGS script set, the overall processing control logs metrics for
    # the start and end of these interfacing components to help with overall
    # metrics gathering. As such, confirmation of the metrics requirements can
    # easily be accomplished by simply checking these in a similar manner as
    # the LPGS scripts statuses are checked.
    # IP 3895 - The LPGS shall allow MDS to pull metrics.
    $sql = "SELECT wo.component script_id, 0 seq_nbr, "
         . "       to_char('S') compl_status, "
         . "       decode(a.status, 'START', 'Y', 'N') started, "
         . "       decode(b.status, 'SUCCESS', 'Y', 'N') success, "
         . "       decode(c.status, 'ABORTED', 'Y', 'N') aborted "
         . "FROM work_order_log wo, "
         . "     (SELECT component, status "
         . "        FROM work_order_log "
         . "       WHERE work_order_id = '$work_order_id' "
         . "         AND component = 'SUBSETTER_INTERFACE' "
         . "         AND status = 'START') a, "
         . "     (SELECT component, status "
         . "        FROM work_order_log " 
         . "       WHERE work_order_id = '$work_order_id' "
         . "         AND component = 'SUBSETTER_INTERFACE' "
         . "         AND status = 'SUCCESS') b, "
         . "     (SELECT component, status "
         . "        FROM work_order_log " 
         . "       WHERE work_order_id = '$work_order_id' "
         . "         AND component = 'SUBSETTER_INTERFACE' "
         . "         AND status = 'ABORTED') c "
         . "WHERE wo.work_order_id = '$work_order_id' "
         . "  AND wo.component = 'SUBSETTER_INTERFACE' "
         . "  AND wo.component = a.component(+) "
         . "  AND wo.component = b.component(+) "
         . "  AND wo.component = c.component(+)";

    my $subsetter_int_wo_ref = db_FetchAll($dbh, $sql, "SCRIPT_ID");
    if (not defined $subsetter_int_wo_ref)
    {
        log_msg("Error: loading subsetter interface info for $work_order_id:\n",
                TRUE);
        log_msg("    -> $DBI::errstr\n", TRUE);
        $error_count++;
        next;
    }

    # Now get the transfer product metrics entries from the log table
    $sql = "SELECT wo.component script_id, 24 seq_nbr, "
         . "       to_char('S') compl_status, "
         . "       decode(a.status, 'START', 'Y', 'N') started, "
         . "       decode(b.status, 'SUCCESS', 'Y', 'N') success, "
         . "       decode(c.status, 'ABORTED', 'Y', 'N') aborted "
         . "FROM work_order_log wo, "
         . "     (SELECT component, status "
         . "        FROM work_order_log "
         . "       WHERE work_order_id = '$work_order_id' "
         . "         AND component = 'TRANSFER_PRODUCT' "
         . "         AND status = 'START') a, "
         . "     (SELECT component, status "
         . "        FROM work_order_log " 
         . "       WHERE work_order_id = '$work_order_id' "
         . "         AND component = 'TRANSFER_PRODUCT' "
         . "         AND status = 'SUCCESS') b, "
         . "     (SELECT component, status "
         . "        FROM work_order_log " 
         . "       WHERE work_order_id = '$work_order_id' "
         . "         AND component = 'SUBSETTER_INTERFACE' "
         . "         AND status = 'ABORTED') c "
         . "WHERE wo.work_order_id = '$work_order_id' "
         . "  AND wo.component = 'TRANSFER_PRODUCT' "
         . "  AND wo.component = a.component(+) "
         . "  AND wo.component = b.component(+) "
         . "  AND wo.component = c.component(+)";

    my $txfr_int_wo_ref = db_FetchAll($dbh, $sql, "SCRIPT_ID");
    if (not defined $txfr_int_wo_ref)
    {
        log_msg("Error: loading transfer product interface info for " .
                "$work_order_id:\n", TRUE);
        log_msg("    -> $DBI::errstr\n", TRUE);
        $error_count++;
        next;
    }

    # Combine all the scripts into one hash now
    my %combined_scripts = (
            %{$wo_scripts_ref},
            %{$subsetter_int_wo_ref},
            %{$txfr_int_wo_ref});

    # Load the processing time (in hours) for the work order id.
    $sql = "SELECT (wo.act_compl_date - a.start_date) * 24 process_time "
         . "  FROM work_orders wo, "
         . "       (SELECT work_order_id, min(date_statused) start_date "
         . "          FROM work_order_log "
         . "         WHERE work_order_id = '$work_order_id' "
         . "           AND component = 'DSW' "
         . "         GROUP BY work_order_id) a "
         . " WHERE wo.work_order_id = a.work_order_id "
         . "   AND wo.work_order_id = '$work_order_id' ";

    my $process_time = db_FetchOne($dbh, $sql);
    if (not defined $process_time)
    {
        log_msg("Error: loading process_time info for $work_order_id:\n",TRUE);
        log_msg("    -> $DBI::errstr\n", TRUE);
        $error_count++;
        next;
    }

    # Load the inventory data for this request
    my $inv_data = {};
    if (get_data_by_scene_id($landsat_scene_id, $inv_data) != SUCCESS)
    {
        log_msg("Error: loading inventory data for $landsat_scene_id\n", TRUE);
        $error_count++;
        next;
    }

    # Get the product type. This will be used to determine which output files
    # will be verified and which daily processing requirement will be checked.
    if ($wo_ref->{"PROCEDURE_ID"} eq "L1T Product"
            || $wo_ref->{"PROCEDURE_ID"} eq "L1GT Product TIRS")
    {
        $total_l1_products++;
        $l1_found = TRUE;

        $verify_l1 = TRUE;
        
        # The L1T/L1GT procedures will also update the L0R product information
        $verify_l0 = TRUE;  
    }
    elsif ($wo_ref->{"PROCEDURE_ID"} eq "L0R Product")
    {
        $total_l0_products++;
        $l0_found = TRUE;

        $verify_l1 = FALSE;
        $verify_l0 = TRUE;
    }
    else
    {
        $verify_l1 = FALSE;
        $verify_l0 = FALSE;

        my $proc_id = $wo_ref->{"PROCEDURE_ID"};
        log_msg("WARN: Unexpected procedure_id found: $proc_id\n", TRUE);
    }

    # Verify the date_entered value is set
    #   IP 2379 - LPGS shall collect metrics on the time of order submittal
    #   during LPGS processing.
    log_msg("  Verify date_entered value...");
    if (defined $wo_ref->{"ENTER_DATE"})
    {
        log_msg("PASS (" . $wo_ref->{"ENTER_DATE"} . ")\n");
    }
    else
    {
        log_msg("FAIL\n");
        $error_count++;
    }

    # Verify the start and complete dates
    #   IP 2377 - The LPGS shall collect metrics on the LPGS processing
    #   start and completion times.
    log_msg("  Verify start date value...");
    if (defined $wo_ref->{"START_DATE"})
    {
        log_msg("PASS (" . $wo_ref->{"START_DATE"} . ")\n");
    }
    else
    {
        log_msg("FAIL\n");
        $error_count++;
    }
    log_msg("  Verify completion date value...");
    if (defined $wo_ref->{"COMP_DATE"})
    {
        log_msg("PASS (" . $wo_ref->{"COMP_DATE"} . ")\n");
    }
    else
    {
        log_msg("FAIL\n");
        $error_count++;
    }

    # Verify the priority is set
    #   IP 2387 - The LPGS shall include priority information in job related
    #   processing metrics.
    log_msg("  Verify priority value...");
    if (defined $wo_ref->{"PRIORITY"})
    {
        log_msg("PASS (" . $wo_ref->{"PRIORITY"} . ")\n");
    }
    else
    {
        log_msg("FAIL\n");
        $error_count++;
    }

    # Verify the state is set to 'C'
    #   IP 2384 - The LPGS shall collect metrics on the LPGS work orders that
    #   complete successfully.
    log_msg("  Verify state set to complete...");
    if (defined $wo_ref->{"STATE"} && $wo_ref->{"STATE"} eq DB_WO_COMPLETED)
    {
        log_msg("PASS (" . $wo_ref->{"STATE"} . ")\n");
    }
    else
    {
        log_msg("FAIL\n");
        $error_count++;
    }

    # Verify the processing time meets requirements
    #   IP 3780 - The LPGS shall make a Level 1 data product available for 
    #   collected data within the time allocated by the USGS Landsat Data 
    #   Continuity Mission (LDCM) Data Processing and Archive System (DPAS) 
    #   Subsystem Performance Requirements Allocations (LDCM-REF-028), 
    #   consistent with availability requirements.
    log_msg("  Verify processing time requirement...");
    my $total_time_in_lpgs = $wait_overhead + $process_time;
    if ($total_time_in_lpgs <= MAX_LPGS_PROCESS_HOURS)
    {
        log_msg(sprintf("PASS (%02.4f)\n", $total_time_in_lpgs));
    }
    else
    {
        log_msg("FAIL ($total_time_in_lpgs)\n");
        $error_count++;
    }

    # Sum the total_time_in_lpgs for all requests. This will be used to 
    # verify the volume requirement
    if ($verify_l1)
    {
        $total_l1_process_time += $total_time_in_lpgs;
    }
    else
    {
        $total_l0_process_time += $total_time_in_lpgs;
    }

    # Verify each script_id in the expected script list has a start/success
    # value in the work order log. This is where we check the subsetter
    # interface and transfer product components similarly, due to metrics
    # collection requirements.
    #   IP 2378 - The LPGS shall collect metrics on the LPGS processing start
    #   and stop times of each step within the processing flows.
    log_msg("  Verify wo_script metrics were captured...\n");
    for my $script (keys %combined_scripts)
    {
        my $script_ref = $combined_scripts{$script};

        log_msg("    Verify completion status set for $script...");
        if ($script_ref->{"COMPL_STATUS"} eq "S")
        {
            log_msg("PASS\n");
        }
        else
        {
            log_msg("FAIL\n");
            $error_count++;
        }

        log_msg("    Verify start metric was captured for $script...");
        if ($script_ref->{"STARTED"} eq "Y")
        {
            log_msg("PASS\n");
        }
        else
        {
            log_msg("FAIL\n");
            $error_count++;
        }

        log_msg("    Verify success metric was captured  for $script...");
        if ($script_ref->{"SUCCESS"} eq "Y")
        {
            log_msg("PASS\n");
        }
        else
        {
            log_msg("FAIL\n");
            $error_count++;
        }
    }

    # Verify the file size for L1 product is recorded in the inventory
    #   IP 1967 - The LPGS shall collect metrics on the size of each scene
    #   output product.
    if ($verify_l1)
    {
        log_msg("  Verify L1 product size is set...");
        if (defined $inv_data->{"L1_FILE"}
             && defined $inv_data->{"L1_FILE"}{"FILE_SIZE"}
             && $inv_data->{"L1_FILE"}{"FILE_SIZE"} > 0)
        {
            log_msg("PASS (" . 
                $inv_data->{"L1_FILE"}{"FILE_SIZE"}.")\n");
        }
        else
        {
            log_msg("FAIL\n");
            $error_count++;
        }
    }

    # Verify the file size for L0RP product is recorded in the inventory
    #   IP 2382 - The LPGS shall collect metrics on the size of each scene
    #   input product.
    if ($verify_l0)
    {
        log_msg("  Verify L0RP product size is set...");
        if (defined $inv_data->{"L0RP_FILE"}
             && defined $inv_data->{"L0RP_FILE"}{"FILE_SIZE"}
             && $inv_data->{"L0RP_FILE"}{"FILE_SIZE"} > 0)
        {
            log_msg("PASS (" .
                $inv_data->{"L0RP_FILE"}{"FILE_SIZE"} . ")\n");
        }
        else
        {
            log_msg("FAIL\n");
            $error_count++;
        }
    }
   
    log_msg("Verifying $product_request_id successful\n", TRUE);
}

# Disconnect from the database
db_Disconnect($dbh);

# Verify there was a work order record for each product request in the input 
# list.
#   IP 2383 - The LPGS shall collect metrics  on the number of orders received
log_msg("Verify work_orders record exists for each product request...");
if ($total_in_products == $total_db_products)
{
    log_msg("PASS\n");
}
else
{
    log_msg("FAIL (Expected: $total_in_products,Found: $total_db_products)\n");
    $error_count++;
}

# Verify scenes per day processing
#  IP 1942 -The LPGS shall have the capability to produce 550 L0Rp data 
#   products per day, subject to availability requirements. 
#  IP 4104 - The LPGS shall have the capability to expand to produce at least 
#   600 L0Rp data products per day in the third year after launch.
#    OR
#  IP 1943 - The LPGS shall have the capability to produce 900 Level 1 data 
#   products per day, subject to availability requirements.
#  IP 4105 - The LPGS shall have the capability to expand to produce at least 
#   1650 Level 1 data products per day in the third year after launch.
my $concurrent_wos = EXPECTED_PROD_PROCESS_NODES * EXPECTED_PROD_MAX_LOAD;

# If the input file contained L0 scenes, calculate the L0 processing time
my $l0_average_time = 0;
my $l0_time_required = 0;
my $l0_process_time = 0;
if ($l0_found)
{
    $l0_average_time = $total_l0_process_time / $total_l0_products;
    $l0_time_required = $l0_average_time * $L0_VOLUME_REQ_1;
    $l0_process_time = $l0_time_required / $concurrent_wos;
}

# If the input file contained L1 scenes, calculate the L1 processing time
my $l1_average_time = 0;
my $l1_time_required = 0;
my $l1_process_time = 0;
if ($l1_found)
{
    $l1_average_time = $total_l1_process_time / $total_l1_products;
    $l1_time_required = $l1_average_time * $L1_VOLUME_REQ_1;
    $l1_process_time = $l1_time_required / $concurrent_wos;
}


log_msg("LPGS L0 actual total processing time: $total_l0_process_time hours\n");
log_msg("LPGS L1 actual total processing time: $total_l1_process_time hours\n");
log_msg("LPGS L0 actual average time per request: $l0_average_time hours\n");
log_msg("LPGS L1 actual average time per request: $l1_average_time hours\n");
log_msg("LPGS L0 expected processing time (IP1942): $l0_process_time hours\n");
log_msg("LPGS L1 expected processing time (IP1943): $l1_process_time hours\n");

# Verify the IP 1942 and IP 1943 requirements. The combined procesing time
# for the L0 and L1 requests should be less than or equal to 24 hours. 
log_msg("Verify volume requirements IP1942 and IP1943... ");
if ($l0_process_time + $l1_process_time <= 24.0)
{
    log_msg("PASS\n");
}
else
{
    log_msg("FAIL\n");
    $error_count++;
}

# Calculate the processing time for the year 3 requirements
if ($l0_found)
{
    $l0_time_required = $l0_average_time * $L0_VOLUME_REQ_2;
    $l0_process_time = $l0_time_required / $concurrent_wos;
}
if ($l1_found)
{
    $l1_time_required = $l1_average_time * $L1_VOLUME_REQ_2;
    $l1_process_time = $l1_time_required / $concurrent_wos;
}

log_msg("LPGS L0 expected processing time (IP4104): $l0_process_time hours\n");
log_msg("LPGS L1 expected processing time (IP4105): $l1_process_time hours\n");

log_msg("Verify volume requirements IP4104 and IP4105... ");
if ($l0_process_time + $l1_process_time <= 24.0)
{
    log_msg("PASS\n");
}
else
{
    # Calculate the number of additional concurrent processes required
    # to meet IP 4104 and IP 4105.

    # If 4 times current processing power, LPGS fails to meet these 
    # requirements
    my $max_concurrent_wos = $concurrent_wos * 4;

    my $additional_procs = -1;
    for (my $tmp = $concurrent_wos; $tmp <= $max_concurrent_wos; $tmp++)
    {
        $l0_process_time = $l0_time_required / $tmp;
        $l1_process_time = $l1_time_required / $tmp;
       
        # Check if requirements would be meet. If it would save the number
        # of additional concurrent processes needed.
        if ($l0_process_time + $l1_process_time <= 24.0)
        {
            $additional_procs = $tmp - $concurrent_wos;
            last;
        }
    }

    if ($additional_procs > -1)
    {
        my $servers = ceil($additional_procs / EXPECTED_PROD_MAX_LOAD);
        log_msg("$servers additional processing nodes required to meet IP4104 "
            . "and IP4105 requirements\n");

        print "$servers additional processing nodes required to meet IP4104 "
            . "and IP4105 requirements\n";
    }
    else
    {
        log_msg("FAIL\n");
        $error_count++;
    }
}


if ($error_count > 0)
{
    log_msg("Verification failed.\n", TRUE);
    exit 1;
}

log_msg("Verification success.\n", TRUE);

# Close the log file if open
if (defined $log_fh)
{
    $log_fh->close();
}

exit 0;

###################
# Subroutines
###################
sub log_msg(@)
{
    my ($msg, $echo) = @_;

    if (defined $log_fh)
    {
        print $log_fh $msg;
    }

    if (defined $echo && $echo == TRUE)
    {   
        print $msg;
    }
}
