#!/usr/bin/perl
##############################################################################
# Name: verify_invalid_WO.pl
#
# Description: Verify metrics are collected for aborted work orders. 
#
# Requirements Verified:
#   IP2385 - The LPGS shall collect metrics on the LPGS work orders that had 
#            an error during processing.
#
# Usage:
#   verify_invalid_WO.pl <WO
##############################################################################

use warnings;
use strict;

use db_utils;
use ias_const;
use ias_wo_const;
use IO::File;

my $error_count = 0;

# Get the WO filename
if (not defined $ARGV[0])
{
    print "Missing filename\n";
    print "verify_invalid_WO.pl <WO Filename>\n";
    exit 1;
}
my $wo_filename = $ARGV[0];

# Open the input file
my $fh = IO::File->new($wo_filename, "r");
if (not defined $fh)
{
    print "Error: opening input file\n";
    exit 1;
}

# Pull the first product request id from the input file
my $pr_id = undef;
while (<$fh>)
{
    chomp;
    if ($_ =~ /^[\s]*$/) { next; }; # skip blank lines

    $pr_id = $_;
    last;
}
$fh->close();

# Verify a product request was found
if (not defined $pr_id)
{
    print "Error: loading product request id\n";
    exit 1;
}

print "Verifying $pr_id \n";

# Get the database connection
my ($username, $password) = db_GetConnectionInfo();
my $dbh = db_Connect($username, $password);
if (not defined $dbh)
{
    print "Error: connecting to the LPGS database\n";
    exit 1;
}

# Load the WO state
my $sql = "SELECT work_order_id, state "
        . "  FROM work_orders "
        . " WHERE product_request_id = '$pr_id'";
my $wo_row = db_FetchRow($dbh, $sql);

my $pr_state = $wo_row->{"STATE"};
my $wo_id = $wo_row->{"WORK_ORDER_ID"};

# Load the anomaly information
$sql = "SELECT anomaly_id "
     . "  FROM anomalies "
     . " WHERE product_request_id = '$pr_id'";
my $anomaly_id = db_FetchOne($dbh, $sql);

# Load the last script completion status
$sql = "SELECT compl_status "
     . "  FROM wo_scripts "
     . " WHERE work_order_id = '$wo_id' "
     . "   AND seq_nbr = (SELECT max(seq_nbr) "
     . "                    FROM wo_scripts "
     . "                   WHERE work_order_id = '$wo_id' "
     . "                     AND state = 'C') ";
my $last_compl_status = db_FetchOne($dbh, $sql);

db_Disconnect($dbh);

# Verify the state is set to Aborted
print "Verify state recorded...";
if (defined $pr_state && $pr_state eq DB_WO_ABORTED)
{
    print "PASS\n";
}
else
{
    print "FAIL\n";
    $error_count++;
}

# Verify an anomaly was created
print "Verify anomaly created...";
if (defined $anomaly_id)
{
    print "PASS\n";
}
else
{
    print "FAIL\n";
    $error_count++;
}

# Verify last completion status is (F)ailed
print "Verify completion status set to (F)ailed...";
if (defined $last_compl_status && $last_compl_status eq "F")
{
    print "PASS\n";
}
else
{
    print "FAIL\n";
    $error_count++;
}

# Check for errors
if ($error_count > 0)
{
    print "Verification failed\n";
    exit 1;
}
else
{
    print "Verification passed\n";
    exit 0;
}
