#!/bin/csh -f
#***************************************************************************
# NAME:      verify_tc018_rad_wo_common.csh
#
#  PURPOSE:  Verifies that the radiometric work order common records are
#            cleaned up properly.
#  RETURNS:
#        Type:  Integer
#        Value: 0 / 1
#
#  USAGE:    ./verify_tc018_rad_wo_common.csh product_request_WOs.txt \
#                                             product_request_WOs_reload.txt
#****************************************************************************
# Create a log file
if ( -e test_case_id ) then
    set tcnum = `cat test_case_id`
    set log = tc${tcnum}.log
    set woid_file = tc${tcnum}_woids.txt
else
    set script = $0:t
    set log = ${script:r}.out
    set woid_file = work_order_id.txt
endif

# Must have two arguments
if ( ${#argv} != 2 ) then
    echo "" |& tee -a $log
    echo "    Usage: $0 product_request_WOs.txt "\
         "product_request_WOs_reload.txt"  |& tee -a $log
    echo "" |& tee -a $log
    exit 1
endif

# Retrieve the first group of work order ids
set temp_ids = `awk '{print $3}' $1`
set first_id = 1
foreach id ($temp_ids)
    if ( $first_id ) then
        set woids1 = `echo "'$id'"`
        set first_id = 0
    else
        set woids1 = `echo "$woids1, '$id'"`
    endif
end
# Retrieve the second group of work order ids
set temp_ids = `awk '{print $3}' $2`
set first_id = 1
foreach id ($temp_ids)
    if ( $first_id ) then
        set woids2 = `echo "'$id'"`
        set first_id = 0
    else
        set woids2 = `echo "$woids2, '$id'"`
    endif
end

# Retrieve the first group of radiometric wo common Ids
set temp_ids = `awk '{print $4}' $1`
set first_id = 1
foreach id ($temp_ids)
    if ( $first_id ) then
        set radids1 = $id
        set first_id = 0
    else
        set radids1 = `echo "$radids1, $id"`
    endif
end
# Retrieve the second group of radiometric wo common Ids
set temp_ids = `awk '{print $4}' $2`
set first_id = 1
foreach id ($temp_ids)
    if ( $first_id ) then
        set radids2 = $id
        set first_id = 0
    else
        set radids2 = `echo "$radids2, $id"`
    endif
end

# Using the list of work order ids count the number of valid and 
# invalid work orders and verify that it is correct
set invalid_wos = `sqlplus -s ${IAS_DB_L8_OLITIRS}` << END
    set head off; 
    SELECT count(*) FROM RADIOMETRIC_WORK_ORDER_COMMON
    WHERE work_order_id in ($woids1)
    AND VALID = 'N';
    exit;
END

set valid_wos = `sqlplus -s ${IAS_DB_L8_OLITIRS}` << END
    set head off; 
    SELECT count(*) FROM RADIOMETRIC_WORK_ORDER_COMMON
    WHERE work_order_id in ($woids2)
    AND VALID = 'Y';
    exit;
END


set invalid_histo_stats = `sqlplus -s ${IAS_DB_L8_OLITIRS}` << END
    set head off; 
    SELECT count(*) FROM HISTOGRAM_STATISTICS_EARTH WHERE
      RADIOMETRIC_WO_COMMON_ID IN ($radids1);
    exit;
END

set valid_histo_stats = `sqlplus -s ${IAS_DB_L8_OLITIRS}` << END
    set head off; 
    SELECT count(*) FROM HISTOGRAM_STATISTICS_EARTH WHERE
      RADIOMETRIC_WO_COMMON_ID IN ($radids2);
    exit;
END

set invalid_histo_summ = `sqlplus -s ${IAS_DB_L8_OLITIRS}` << END
    set head off; 
    SELECT count(*) FROM HIST_SCENE_SUMMARY_STATS WHERE
      RADIOMETRIC_WO_COMMON_ID IN ($radids1);
    exit;
END

set valid_histo_summ = `sqlplus -s ${IAS_DB_L8_OLITIRS}` << END
    set head off; 
    SELECT count(*) FROM HIST_SCENE_SUMMARY_STATS WHERE
      RADIOMETRIC_WO_COMMON_ID IN ($radids2);
    exit;
END

set invalid_sat_pixels = `sqlplus -s ${IAS_DB_L8_OLITIRS}` << END
    set head off; 
    SELECT count(*) FROM SATURATED_PIXELS WHERE
      RADIOMETRIC_WO_COMMON_ID IN ($radids1);
    exit;
END

set valid_sat_pixels = `sqlplus -s ${IAS_DB_L8_OLITIRS}` << END
    set head off; 
    SELECT count(*) FROM SATURATED_PIXELS WHERE
      RADIOMETRIC_WO_COMMON_ID IN ($radids2);
    exit;
END

echo ""
# Note the counts for valid records (except work orders) are approximate to
# avoid making this test very very brittle
if ( $valid_wos != 12 || $invalid_wos != 0 || $invalid_histo_stats != 0 \
      || $valid_histo_stats < 500000 || $invalid_histo_summ != 0 \
      || $valid_histo_summ < 1000 || $invalid_sat_pixels != 0 \
      || $valid_sat_pixels < 400) then
    echo "ERROR: The number of on or more record types is incorrect"
    echo ""
    echo "There were $invalid_wos Invalid records where"\
            "(0) were expected"
    echo "There were $valid_wos Valid records where (12) were expected"
    echo "There were $invalid_histo_stats invalid histrogram stat records "\
         "where (0) were expected"
    echo "There were $valid_histo_stats valid histrogram stat records where"\
         " at least (500000) were expected"
    echo "There were $invalid_histo_summ invalid histrogram summary records "\
         "where (0) were expected"
    echo "There were $valid_histo_summ valid histrogram summary records where"\
         " at least (1000) were expected"
    echo "There were $invalid_sat_pixels invalid saturated pixel records "\
         "where (0) were expected"
    echo "There were $valid_sat_pixels valid saturated pixel records where"\
         " at least (400) were expected"
    echo ""
    exit 1

else
    echo "SUCCESS: All records have been verified"
    echo ""
    echo "$invalid_wos Invalid work order records"
    echo "$valid_wos Valid work order records"
    echo "$invalid_histo_stats Invalid histrogram stat records"
    echo "$valid_histo_stats Valid histrogram stat records"
    echo "$invalid_histo_summ Invalid histrogram summary records"
    echo "$valid_histo_summ Valid histrogram summary records"
    echo "$invalid_sat_pixels Invalid saturated pixels records"
    echo "$valid_sat_pixels Valid saturated pixels records"
    echo ""
    exit 0
endif
