package get_inv_data;
use warnings;
use strict;

use Exporter;
use DBI;
use ias_const;

use vars qw(@ISA @EXPORT @EXPORT_OK %EXPORT_TAGS $VERSION);
$VERSION = 0.1;
@ISA = qw(Exporter);
@EXPORT = qw(
        get_data_by_scene_id
    );

# set when module call returns an error
our $error_message = "";

##############################################################################
#   get_data_by_scene_id
#
# Retrieves product output file information for a given $landsat_scene_id. The
# information is return in $hash_ref hash reference parameter.
#
# Returns SUCCESS or ERROR. If ERROR, $error_message will be set
##############################################################################
sub get_data_by_scene_id($$$;$)
{
    my ($landsat_scene_id, $processing_instance, $hash_ref, $albers) = @_;

    # reset the global error message
    $error_message = "";

    my $dbh = get_inv_db_conn();
    if (not defined $dbh)
    {
        return ERROR;
    }

    # Retrieve the correct inventory schema
    my $sql = "Select inventory_schema, collection_number from processing_instances " .
              "where instance_name = '$processing_instance'";
    my $sth = $dbh->prepare($sql);
    if (!$sth)
    {
        $error_message = "Error preparing main query; $DBI::errstr";
        $dbh->disconnect();
        return ERROR;
    }
    if (!$sth->execute())
    {
        $error_message = "Error executing main query; $DBI::errstr";
        $sth->finish();
        $dbh->disconnect();
        return ERROR;
    }

    my ($inv_schema, $collection_number) = $sth->fetchrow_array();
    $sth->finish();

    my $tirs_image_quality = ", tirs_sp.image_quality as TIRS_IMAGE_QUALITY ";
    my $tirs_condition = " AND tirs_sp.sensor_id = 'TIRS' ";
    my $tirs_ssm_model = ", tirs_sp.tirs_ssm_model ";
    my $oli_image_quality = ", oli_sp.image_quality as OLI_IMAGE_QUALITY";
    my $oli_condition = " AND oli_sp.sensor_id = 'OLI' ";

    # Remove the TIRS parameters for OLI-only and OLI paramters for TIRS-only
    if ($landsat_scene_id =~ /^LO/)
    {
        $tirs_image_quality = "";
        $tirs_condition = "";
        $tirs_ssm_model = "";
    }
    elsif ($landsat_scene_id =~ /^LT/)
    {
        $oli_image_quality = "";
        $oli_condition = "";
    }

    $sql = "SELECT s.lmd_scene_id, s.l1_location, s.l0rp_location, "
            . "       s.vertical_lon_from_pole, s.cpf_name, "
            . "       s.data_type, s.elevation_source, s.false_easting, "
            . "       s.false_northing, s.geometric_rmse_model_y, "
            . "       s.geometric_rmse_model_x, s.grid_cell_size_panchromatic,"
            . "       s.grid_cell_size_reflective, s.grid_cell_size_thermal, "
            . "       s.processing_software_version, s.map_projection, "
            . "       s.orientation, "
            . "       s.panchromatic_lines, s.panchromatic_samples,"
            . "       s.reflective_lines, "
            . "       s.reflective_samples, s.resampling_option, "
            . "       s.rlut_file_name, s.scene_bands_l1, s.scene_status, "
            . "       s.subsetter_version_l1, s.thermal_lines, "
            . "       s.thermal_samples, s.true_scale_lat, s.utm_zone, "
            . "       s.scene_bands_l0rp, s.subsetter_version_l0rp, "
            . "       s.cloud_cover, s.cloud_cover_land, s.snow_ice_cover,"
            . "       ph.hostname, sc.corner_ll_lat, sc.corner_ll_lon, "
            . "       sc.corner_lr_lat, sc.corner_lr_lon, sc.corner_ul_lat, "
            . "       sc.corner_ul_lon, sc.corner_ur_lat, sc.corner_ur_lon, "
            . "       sc.coordinate_type scene_coord_type, "
            . "       sc.scene_center_lat, sc.scene_center_lon, "
            . "       s.geometric_rmse_model, s.ground_control_points_model, "
            . "       s.output_format, s.ground_control_points_version "
            . $tirs_ssm_model . $tirs_image_quality . $oli_image_quality
            . "  FROM $inv_schema.lmd_scene s, "
            . "       $inv_schema.lmd_processing_host ph, "
            . "       $inv_schema.lmd_l1_scene_coordinate sc, "
            . "       $inv_schema.lmd_sensor_scene_parameter tirs_sp, "
            . "       $inv_schema.lmd_sensor_scene_parameter oli_sp "
            . " WHERE s.lmd_scene_id = ph.lmd_scene_id(+) "
            . "   AND s.lmd_scene_id = sc.lmd_scene_id(+) "
            . "   AND s.lmd_scene_id = tirs_sp.lmd_scene_id(+) "
            . "   AND s.lmd_scene_id = oli_sp.lmd_scene_id(+) "
            . "   AND s.landsat_scene_id = '$landsat_scene_id'"
            . $tirs_condition . $oli_condition;

    $sth = $dbh->prepare($sql);
    if (!$sth)
    {
        $error_message = "Error preparing main query; $DBI::errstr";
        $dbh->disconnect();
        return ERROR;
    }

    if (!$sth->execute())
    {
        $error_message = "Error executing main query; $DBI::errstr";
        $sth->finish();
        $dbh->disconnect();
        return ERROR;
    }

    my $row_hash;
    if ($row_hash = $sth->fetchrow_hashref())
    {
        for my $key (keys %{$row_hash})
        {
            $hash_ref->{$key} = $row_hash->{$key};
        }
    }
    else
    {
        $error_message = "No result for $landsat_scene_id";
        $sth->finish();
        $dbh->disconnect();
        return ERROR;
    }

    # Check for additional lmd_scene records
    if ($row_hash = $sth->fetchrow_hashref())
    {
        $error_message = "Multiple results for $landsat_scene_id found";
        $sth->finish();
        $dbh->disconnect();
        return ERROR;
    }
    $sth->finish();

    # Load the L0RP data file information
    my $lmd_scene_id = $hash_ref->{"LMD_SCENE_ID"};
    $sql = "SELECT file_name, file_date, file_size, file_checksum, "
         . "       file_type, product_file_size "
         . "  FROM $inv_schema.lmd_l0rp_data_file "
         . " WHERE lmd_scene_id = $lmd_scene_id";

    $sth = $dbh->prepare($sql);
    if (!$sth)
    {
        $error_message = "Error preparing l0rp data query; $DBI::errstr";
        $dbh->disconnect();
        return ERROR;
    }

    if (!$sth->execute())
    {
        $error_message = "Error executing l0rp data query; $DBI::errstr";
        $sth->finish();
        $dbh->disconnect();
        return ERROR;
    }

    my $l0rp_data_hash;
    if ($l0rp_data_hash = $sth->fetchall_hashref("FILE_TYPE"))
    {
        for my $file_type (keys %{$l0rp_data_hash})
        {
            $hash_ref->{"$file_type"."_FILE"} = $l0rp_data_hash->{$file_type};
        }
    }
    else
    {
        $error_message = "No l0rp data results for $landsat_scene_id";
        $sth->finish();
        $dbh->disconnect();
        return ERROR;
    }
    $sth->finish();

    # Load the L1 data file information
    $sql = "SELECT d.file_name, d.file_date, d.file_size, d.file_checksum, "
         . "       d.file_type, d.product_file_size "
	 . ($collection_number >= 1 ? ", s.landsat_product_id " : "") 
         . "  FROM $inv_schema.lmd_l1_data_file d, $inv_schema.lmd_scene s "
         . " WHERE d.lmd_scene_id = $lmd_scene_id "
         . "   AND s.lmd_scene_id = $lmd_scene_id";

    $sth = $dbh->prepare($sql);
    if (!$sth)
    {
        $error_message = "Error preparing l1 data query; $DBI::errstr";
        $dbh->disconnect();
        return ERROR;
    }

    if (!$sth->execute())
    {
        $error_message = "Error executing l1 data query; $DBI::errstr";
        $sth->finish();
        $dbh->disconnect();
        return ERROR;
    }

    my $l1_data_hash;
    if ($l1_data_hash = $sth->fetchall_hashref("FILE_TYPE"))
    {
        for my $file_type (keys %{$l1_data_hash})
        {
            $hash_ref->{"$file_type"."_FILE"} = $l1_data_hash->{$file_type};
        }
    }
    else
    {
        $error_message = "No l1 data results for $landsat_scene_id";
        $sth->finish();
        $dbh->disconnect();
        return ERROR;
    }
    $sth->finish();

    # Load the bpf file information
    $sql = "SELECT bpf_name, sensor_id "
         . "FROM $inv_schema.lmd_sensor_scene_parameter "
         . "WHERE lmd_scene_id = $lmd_scene_id";

    $sth = $dbh->prepare($sql);
    if (!$sth)
    {
        $error_message = "Error preparing bpf query; $DBI::errstr";
        $dbh->disconnect();
        return ERROR;
    }

    if (!$sth->execute())
    {
        $error_message = "Error executing bpf query; $DBI::errstr";
        $sth->finish();
        $dbh->disconnect();
        return ERROR;
    }

    my $bpf_hash;
    if ($bpf_hash = $sth->fetchall_hashref("SENSOR_ID"))
    {
        my $bpf = {};
        for my $sensor_id (keys %{$bpf_hash})
        {
            my $bpf_name = "BPF_NAME";
            $bpf->{$sensor_id} = $bpf_hash->{$sensor_id}{$bpf_name};
        }
        $hash_ref->{"BPF_NAMES"} = $bpf;
    }
    else
    {
        $error_message = "No bpf results for $landsat_scene_id";
        $sth->finish();
        $dbh->disconnect();
        return ERROR;
    }
    $sth->finish();

    if ($collection_number < 0)
    {
        # Load the cloud cover algorithms information
	$sql = "SELECT cloud_cover_algorithm, cloud_cover_algorithm_weight "
	    . "  FROM $inv_schema.lmd_cloud_cover_algorithm "
	    . " WHERE lmd_scene_id = $lmd_scene_id";

	$sth = $dbh->prepare($sql);
	if (!$sth)
	{
	    $error_message = "Error preparing cloud cover query; $DBI::errstr";
	    $dbh->disconnect();
	    return ERROR;
	}

	if (!$sth->execute())
	{
	    $error_message = "Error executing cloud cover query; $DBI::errstr";
	    $sth->finish();
	    $dbh->disconnect();
	    return ERROR;
	}

	my $cloud_cover_hash;
	if ($cloud_cover_hash = $sth->fetchall_hashref("CLOUD_COVER_ALGORITHM"))
	{
	    my $ccw = {};
	    for my $alg (keys %{$cloud_cover_hash})
	    {
		my $weight_col = "CLOUD_COVER_ALGORITHM_WEIGHT";
		$ccw->{$alg} = $cloud_cover_hash->{$alg}{$weight_col};
	    }
	    $hash_ref->{"CLOUD_COVER_WEIGHTS"} = $ccw;
	}
	else
	{
	    $error_message = "No cloud cover results for $landsat_scene_id";
	    $sth->finish();
	    $dbh->disconnect();
	    return ERROR;
	}
	$sth->finish();
    }

    if ($albers)
    {
        # Load the albers metadata information
        $sql = "SELECT date_l1_generated, file_size, l1_location, "
             . "landsat_product_id, map_projection, project "
             . "FROM $inv_schema.lmd_scene_project_data "
             . "WHERE lmd_scene_id = $lmd_scene_id";

        $sth = $dbh->prepare($sql);
        if (!$sth)
        {
            $error_message = "Error preparing albers project metadata query; "
                . "$DBI::errstr";
            $dbh->disconnect();
            return ERROR;
        }

        if (!$sth->execute())
        {
            $error_message = "Error executing albers project metadata query; "
                . "$DBI::errstr";
            $sth->finish();
            $dbh->disconnect();
            return ERROR;
        }

        my $albers_hash;
        if ($albers_hash = $sth->fetchrow_hashref)
        {
            for my $key (keys %{$albers_hash})
            {
                $hash_ref->{$key."_AEA"} = $albers_hash->{$key};
            }

            #$hash_ref->{"PROJECT"} = $albers_hash;
        }
        else
        {
            $error_message = "No albers project results for $landsat_scene_id";
            $sth->finish();
            $dbh->disconnect();
            return ERROR;
        }
        $sth->finish();
    }

    $dbh->disconnect();

    return SUCCESS;
}

##############################################################################
# Name: get_inv_db_conn
#
# Purpose: Creates and returns a open database handle to the COMMON schema.
#
##############################################################################
sub get_inv_db_conn()
{
    my $con_str = $ENV{"IAS_DB_COM"};
    if (not defined $con_str)
    {
        $error_message = "\$IAS_DB_COM not set";
        return undef;
    }

    my $sid = $ENV{"ORACLE_SID"};
    if (not defined $sid)
    {
        $error_message = "\$ORACLE_SID not set";
        return undef;
    }

    my ($username, $password) = split("/", $con_str);
    if (not defined $password)
    {
        $error_message = "Invalid value for \$IAS_DB_COM";
        return undef;
    }

    my $dbh = DBI->connect("DBI:Oracle:$sid", $username, $password,
                           {AutoCommit => 0, PrintError => 0});
    if (!$dbh)
    {
        $error_message = "Error connecting to database $DBI::errstr";
        return undef;
    }

    return $dbh;
}

1;
