#!/bin/csh -fe
#
# Name: set_wo_common_state.csh
#
# Description: The purpose of this script is to update the
#              radiometric_work_order_common or geometric_work_order_common
#              valid column to a known state for the indicated scene or interval
#              IDs at the start of a test case.
#
# Parameters:
#             Subsystem - The system to update either RPS or GPS
#
#             Valid Flag - The flag to update the subsytem with either 'Y'
#                          or 'N'
#
#             Update Eval Flag - Flag to indicate the evalution database should
#                                be updated as well.
#
# Usage: set_wo_common_state.csh <SUBSYSTEM> <VALID_FLAG> <UPDATE_EVAL_FLAG>
#               LANDSAT_SCENE_ID|LANDSAT_INTERVAL_ID ...
#---------------------------------------------------------------------------

alias usage 'echo "" |& tee -a $log; echo "    Usage: set_wo_common_state ' \
            '<SUBSYSTEM> <VALID_FLAG> <UPDATE_EVAL_FLAG> ' \
            'LANDSAT_SCENE_ID|LANDSAT_INTERVAL_ID ... "; echo "";' \
            'echo "        Valid Values:"; echo "";'\
            'echo "            SUBSYSTEM       : RPS or GPS";'\
            'echo "            VALID_FLAG      : Y or N";'\
            'echo "            UPDATE_EVAL_FLAG: Y or N" |& tee -a $log; '\
            'echo "" |& tee -a $log'

# Create a log file
if (-e test_case_id) then
    set id = `cat test_case_id`
    set log = tc${id}.log
else
    set script = $0:t
    set log = ${script:r}.out
endif

# Must have at least three arguments
if (${#argv} < 4) then
    usage
    exit 1
endif

# Pick the correct table to update
set subsystem = $argv[1]
shift
if ($subsystem == "RPS") then
    set table = "RADIOMETRIC_WORK_ORDER_COMMON"
else if ($subsystem == "GPS") then
    set table = "GEOMETRIC_WORK_ORDER_COMMON"
else
    echo "" |& tee -a $log
    echo "Error: $subsystem is not valid for this script" |& tee -a $log
    usage
    exit 1
endif

# Make sure the valid value is legal
set valid_state = $argv[1]
shift
if ($valid_state != "Y" && $valid_state != "N") then
    echo "" |& tee -a $log
    echo "Error: $valid_state is not valid for this script" |& tee -a $log
    usage
    exit 1
endif

# Make sure the eval flag is legal
set update_eval = $argv[1]
shift
if ($update_eval != "Y" && $update_eval != "N") then
    echo "" |& tee -a $log
    echo "Error: $update_eval is not valid for this script" |& tee -a $log
    usage
    exit 1
endif

# Issue a message
echo "" |& tee -a $log
echo "Setting table $table valid flag to $valid_state for:" |& tee -a $log
echo "" |& tee -a $log

# Update each of the scene or interval IDs provided.  Note that the query
# uses a shortcut to match either type of ID instead of figuring out which type
# of ID is provided since that is sufficient.

while (${#argv})
    /bin/echo -n $argv[1] |& tee -a $log

    # Making sure to handle the version number in ST
    set scene_id_base = `echo $argv[1] | cut -c 1-19`%
    set interval_id_base = `echo $argv[1] |cut -c 1-22`%

    (sqlplus -s $IAS_DB_L8_OLITIRS | tee -a $log) <<END
        update $table set valid='$valid_state'
        where (landsat_scene_id like '$scene_id_base' OR
        landsat_interval_id like '$interval_id_base');
        exit;
END

    # Update eval if requested
    if ($update_eval == "Y") then
        /bin/echo -n $argv[1]" in the EVAL database." |& tee -a $log

        (sqlplus -s $IAS_DB_L8_OLITIRS_EVAL | tee -a $log) <<END
        update $table set valid='$valid_state'
        where (landsat_scene_id like '$scene_id_base' OR
        landsat_interval_id like '$interval_id_base');
        exit;
END

    endif

    shift
end

exit 0
