#!/usr/bin/perl
############################################################################## 
# Description:
#    This script submits a test order through the TRAM Mass Loader. 
#
# Usage:
#   tram_mass_loader.pl <input order file> <output file>
#    
#   <input order file> contains the order information use when submitting the 
#       order to the Mass Loader. See below for information on the input file
#       format and parameters.
#
#   <output file> file name for the output. The Product Request Ids for each
#       unit will be written the output file with one product request per line.
#       The order of the product request will match the order of the units 
#       from the input file. So, the first unit will be the first line in the
#       output file, the second unit the second line, etc....
#
# Input Order File format and parameters:
#   The input file is broken out into three sections. The [service] section
#   allows you to override the default MassLoader endpoint set in the 
#   IAS_MASSLOADER_ENDPOINT environment variable. To override, provide as 
#   "serviceEndpoint" value in this section. If the IAS_MASSLOADER_ENDPOINT is
#   not set, the "serviceEndpoint" is required. 
#
#   The [order] section defines order level values.  The "registrationId" 
#   parameter allows you to override the default IAS_MASSLOADER_USERID 
#   environment variable. If IAS_MASSLOADER_USERID is not set, the 
#   "registrationId" is required. The get a registrationId you'll need to 
#   create an account use EarthExplorer.  If testing with "MassLoaderdevsys" 
#   create an account in EarthExplorer devsys. If testing through 
#   MassLoaderdevmast create an account in EarthExplorer devmast. The "roleId"
#   is an optional parameter. "requestor" specified the TRAM node. This is
#   required.  Use "L1_GENERATION" for testing. "externalRefNumber" is a
#   required parameter.  The externalRefNumber is useful when looking up
#   errors in TRAM's logs. "priority" is the LPGS priority. TRAM does not
#   use this in scheduling but it is provided to LPGS through the DCO. The
#   "orderComment" is optional.a
#
#   The [units] section allows a list of units. Provide one unit per line
#   and the unit parameters are pipe-delimited.  The format is:
#   <Landsat Scene Id>|<TRAM product code|<Unit Comment>|<Recipe ID>
#   The scene Id and product code are required.  The recipe Id is not used by
#   LDCM. If the comment and recipe id are not provided still include the 
#   pipe delimiters:
#   LT81510282000189SGS00|T285||
#
#   Comments can be added using the pound (#) sign at the beginning of a line.
#
# Sample Input:
# [service]
# serviceEndpoint=http://edclxs151/MassLoaderdevsys/MassLoader
# [order]
# registrationId=<EarthExporer Registration Id> 
# roleId=LPGS
# requestor=L1_GENERATION
# externalRefNumber=ABC123
# priority=9
# orderComment=Order Level Comment
#
# [units]
# LT81510302000189SGS00|L8T01|Test Case 9 - TIRS|
# LC80240192001117EDC00|L8C01|Test Case 9 - L1T|
#
##############################################################################
use warnings;
use strict;

use IO::File;
use LWP::UserAgent;
use HTTP::Request;
use HTTP::Request::Common;
use ias_const;

#### Subroutine Declarations ####
sub loadWoFile($$);
sub buildSoapRequest($);
sub trim($);

#### MAIN ####

# Use more permissive hostname verification
$ENV{PERL_LWP_SSL_VERIFY_HOSTNAME} = 0;

# Get the input file
if (not defined $ARGV[0])
{
    die("Error: Missing order file.");
}
my $wo_filename = $ARGV[0];

if (not defined $ARGV[1])
{
    die("Error: Missing output file.");
}
my $output_filename = $ARGV[1];

# Verify the input file exists
if (! -e $wo_filename)
{
    die("Error: $wo_filename does not exist.");
}

my $order_hash = {};
if (loadWoFile($wo_filename, $order_hash) != SUCCESS)
{
    die("Error: loading order information from $wo_filename");
}

# Verify required values are present
if (not defined $order_hash->{"serviceEndpoint"})
{
    die("Error: missing serviceEndpoint");
}
if (not defined $order_hash->{"registrationId"})
{
    die("Error: missing registrationId");
}
if (not defined $order_hash->{"requestor"})
{
    die("Error: missing requestor");
}

# Set default values for missing parameters
if (not defined $order_hash->{"priority"})
{
    $order_hash->{"priority"} = 9;
}
if (not defined $order_hash->{"roleId"})
{
    $order_hash->{"roleId"} = "LPGS_SCRIPT";
}
if (not defined $order_hash->{"orderComment"})
{
    $order_hash->{"orderComment"} = "";
}
if (not defined $order_hash->{"externalRefNumber"})
{
    $order_hash->{"externalRefNumber"} = "1234567890";
}

# Verify the scene parameters
for my $scene (@{$order_hash->{"units"}})
{
    if (not defined $scene->{"sceneId"})
    {
        die("Error: missing sceneId in unit");
    }
    if (not defined $scene->{"productName"})
    {
        die("Error: missing productName in unit");
    }
    
    # set default values for option parameters
    if (not defined $scene->{"unitComment"})
    {
        $scene->{"unitComment"} = "";
    }
    if (not defined $scene->{"recipeId"})
    {
        $scene->{"recipeId"} = "";
    }
}

# Build the XML
my $soapEnv =  buildSoapRequest($order_hash);

# Get the service endpoint
my $endpoint = $order_hash->{"serviceEndpoint"};

# Build the request objects
my $userAgent = LWP::UserAgent->new();
my $request = HTTP::Request->new(POST => $endpoint);
$request->header(SOAPAction => '"uri:gov.usgs.tram.massloader/submitOrder"');
$request->content($soapEnv);
$request->content_type("text/xml; charset=utf-8");

# Submit the order to TRAM
my $response = $userAgent->request($request);

my $orderNumber;
if ($response->code == 200)
{
    # Parse the order number from the response
    my $xmlResp = $response->content;
    if ($xmlResp =~ /<orderNumber>([0-9]+)<\/orderNumber>/)
    {
        $orderNumber = $1;
    }
    else
    {
        print "Error: parsing orderNumber from TRAM response\n";
        print "    -> $xmlResp\n";
        exit 1;
    }
}
else
{
    # Try to parse the error message from the response. If that fails
    # print the whole response XML.
    my $xmlResp = $response->content;
    if ($xmlResp =~ /<faultstring>([a-zA-Z0-9-_: ]+)<\/faultstring>/)
    {
        print "Error: submitting order to TRAM\n";
        print "    -> " . $1 . "\n";
        exit 1;
    }
    else
    {
        print "Error: submitting order to TRAM\n";
        print "    -> " . $response->content . "\n";
        exit 1;
    }
}

# Write the product request ids to the output file
my $out_fh = IO::File->new($output_filename, "w");
my $i = 1;
for my $scene (@{$order_hash->{"units"}})
{
    print $out_fh sprintf("%s_%05d\n", $orderNumber, $i);
    $i++;
}
$out_fh->close();

exit 0;


#### SUBROUTINES ####

##############################################################################
# Name: loadWoFile
# Description: Accepts a $filename and a $hash_ref.  Loads the work order
#   information from $filename into $hash_ref.
# Return: SUCCESS or ERROR
##############################################################################
sub loadWoFile($$)
{
    my ($filename, $hash_ref) = @_;

    # Open the input file contains the order details
    my $fh = IO::File->new($filename, 'r');
    if (not defined $fh)
    {
        die("Error: opening $filename failed");
    }

    # create a new unit array in the order hash reference
    my $units_list = [];

    my $section = "";
    while (<$fh>)
    {
        chomp;

        # skip comments and blank lines
        if ($_ =~ /^\s*(#|$)/) { next; }

        # check for a new section
        if ($_ =~ /^\[([\w\s]+)\]$/)
        {
            $section = $1;
            next;
        }

        if ($section eq "order" || $section eq "service")
        {
            my ($key, $value) = split("=", $_);
            $key = trim($key);
            $value = trim($value);
            $hash_ref->{$key} = $value;
        }
        elsif ($section eq "units")
        {
            my ($scene_id, $product, $comment, $recipe) = split(/\|/, $_);
            
            my $unit_hash = {
                "sceneId" => $scene_id,
                "productName" => $product,
                "unitComment" => $comment,
                "recipeId" => $recipe
            };

            push(@{$units_list}, $unit_hash);
        }
        else
        {
            print "Error: invalid section header [$section] in $filename\n";
            $fh->close();
            return ERROR;
        }
    }
    $fh->close();

    $hash_ref->{"units"} = $units_list;

    # If the serviceEndpoint and registrationId values are not set check
    # environment variables.
    if (not defined $hash_ref->{"serviceEndpoint"})
    {
        if (defined $ENV{"IAS_MASSLOADER_ENDPOINT"})
        {
            $hash_ref->{"serviceEndpoint"} = $ENV{"IAS_MASSLOADER_ENDPOINT"};
        }
    }

    if (not defined $hash_ref->{"registrationId"})
    {
        if (defined $ENV{"IAS_MASSLOADER_USERID"})
        {
            $hash_ref->{"registrationId"} = $ENV{"IAS_MASSLOADER_USERID"};
        }
    }

    return SUCCESS; 
}

##############################################################################
# Name: buildSoapRequest
# Description: Builds the SOAP request example from the work hash loaded by
#   loadWoFile.
# Return: XML request as string
##############################################################################
sub buildSoapRequest($)
{
    my ($order_hashref) = @_;

    # build the units XML
    my $units_arrayref = $order_hashref->{"units"};

    my $unitsXml = "";
    for my $unit (@{$units_arrayref})
    {
        $unitsXml .= <<UNITXML;
<scene>
    <sceneId>$unit->{"sceneId"}</sceneId>
    <productName>$unit->{"productName"}</productName>
    <unitComment>$unit->{"unitComment"}</unitComment>
    <recipeId>$unit->{"recipeId"}</recipeId>
    <parameters/>
</scene>
UNITXML
    }
    
    # build the order XML
    my $orderFormat = <<ORDERFMT;
<order>
    <registrationId>%s</registrationId>
    <roleId>%s</roleId>
    <requestor>%s</requestor>
    <externalRefNumber>%s</externalRefNumber>
    <priority>%s</priority>
    <orderComment>%s</orderComment>
    <scenes>
        %s
    </scenes>
</order>
ORDERFMT

    my $orderXml = sprintf($orderFormat,
        $order_hashref->{"registrationId"},
        $order_hashref->{"roleId"},
        $order_hashref->{"requestor"},
        $order_hashref->{"externalRefNumber"},
        $order_hashref->{"priority"},
        $order_hashref->{"orderComment"},
        $unitsXml
    );

    # combine order/units with the SOAP envelope
    my $soapXml = <<SOAPENV;
<?xml version='1.0' encoding='UTF-8'?>
<soapenv:Envelope xmlns:soapenv="http://schemas.xmlsoap.org/soap/envelope/"
                  xmlns:uri="uri:gov.usgs.tram.massloader">
    <soapenv:Header/>
    <soapenv:Body>
        <uri:submitOrder>
            $orderXml
        </uri:submitOrder>
    </soapenv:Body>
</soapenv:Envelope>
SOAPENV

    return $soapXml;
}

sub trim($)
{
    my ($str) = @_;
    $str =~ s/^\s+//;
    $str =~s/\s+$//;
    return $str;
}
