#!/usr/bin/perl
################################################################################
# verify_inv_data.pl
#
# Usage: verify_inv_data.pl <scene_id>
#
# <scene_id> is the Landsat scene ID of the scene to verify the inventory
# database values for.
#
# Note - the following fields are in the product_data.txt inventory load
#        file from transfer_product, but aren't returned by the inventory
#        query results because they are not set from these values in the
#        database. Rather, these fields are set by the database package by
#        pulling them from the scene id. So they are not verified here.
#
#           date_acquired
#           target_wrs_path
#           target_wrs_row
#
# This script will query the inventory for the data items that are entered to
# it by the LPGS (and underlying IAS) processing system. Those items are
# verified against a benchmarked file of the same information, expected to be
# located in the $IAS_IT_TEST_DATA_DIR/test_case_### directory
###############################################################################
use warnings;
use strict;
use lib '../tools/';    # Using the get_inv_data module
use get_inv_data;       # Shared module to get the inventory data
use ias_const;          # Definition of ERROR
use Logger;             # Logging module
use File::Basename;     # Basename routine
use Getopt::Long;

# Declare program constant
use constant PROG => basename($0);

# Build a hash of the fields we don't want to check from the inventory
# query results
my %inventoryFieldsToSkip = (
        SUBSETTER_VERSION_L0RP => 1,        # Checked by LPGS TC002
        SUBSETTER_VERSION_L1 => 1,          # Not available to verify in L1T pkg
        LMD_SCENE_ID => 1,                  # No need to check
        BROWSE_QUALITY_METADATA_FILE => 1,  # Not generated by LPGS
        BROWSE_REFL_METADATA_FILE => 1,     # Not generated by LPGS
        BROWSE_THERMAL_METADATA_FILE => 1,  # Not generated by LPGS
        BROWSE_LEVEL_1_QUALITY_FULL_RESOLUTION_FILE => 1,
        BROWSE_LEVEL_1_REFLECTANCE_FULL_RESOLUTION_FILE => 1,
        BROWSE_LEVEL_1_THERMAL_FULL_RESOLUTION_FILE => 1
);

my $numArgs = $#ARGV + 1;
if ($numArgs < 1)
{
    usage();
}

my $environment;
my $processing_instance;
my $albers;
GetOptions("processing-instance=s" => \$processing_instance,
           "environment=s" => \$environment,
           "albers" => \$albers);

# Set up the logger
my $logger = Logger->new("STDOUT");
if (!$logger)
{
    print STDERR PROG . ": ERROR:  Failed to establish logger object for "
        . "STDOUT.\n";
    exit FAILURE;
}
$logger->set_module_name(PROG);
set_default_log($logger);

# Get the test case id from caller
my $test_case_id;
if ( -e "./test_case_id" )
{
    $test_case_id = `cat ./test_case_id`;
    chomp($test_case_id);
}
else
{
    $logger->fatal_msg("The file test_case_id does not exist");
    exit FAILURE;
}

# Get the scene ID
my $landsatSceneId = $ARGV[0];

# Check what environment we are using
if (!defined $environment)
{
    $environment = "it";
}
elsif (uc($environment) eq "ST")
{
    $environment = "st";
}

# Check the processing instance to use
if (!defined $processing_instance)
{
    $processing_instance = $ENV{IAS_INSTANCE_NAME};
    if (!defined $processing_instance)
    {
        $logger->fatal_msg("The IAS_INSTANCE_NAME env variable is not set");
        exit FAILURE;
    }
}

# Grab the software version
my $softwareVer = `ias_version`;
if (!defined $softwareVer)
{
    $logger->fatal_msg("Could not retrieve software version through the ias".
        " version utility "); 
    exit FAILURE;
}
chomp($softwareVer);
$softwareVer = "LPGS_$softwareVer";

my %inventorySceneData = ();
my $outFile = $landsatSceneId . "_" . lc($processing_instance) . "_inventory_data.txt";
$outFile =~ s/-//;

# Call the routine provided by the get_inv_data module to query for all data
# LPGS (and IAS) put into the inventory
my $status = get_data_by_scene_id($landsatSceneId, $processing_instance,
                                  \%inventorySceneData, $albers);
if ($status == ERROR)
{
    $logger->fatal_msg("Getting data from inventory:".
        " $get_inv_data::error_message\n");
    exit FAILURE;
}

# Write the inventory data to a file for verification
$status = open (FILE, ">$outFile");
if (!defined $status)
{
    $logger->fatal_msg("Opening file $outFile for writing\n");
    exit FAILURE;
}

# values that need to be checked manually, so will be printed to the screen
my $l0rpLocation = '';
my $l0rpDataFileDate = '';
my $l0rpChecksumFileDate = '';
my $l0rpDataFileSize = '';
my $l0rpDataProductFileSize = '';
my $l1DataFileDate = '';
my $l1ChecksumFileDate = '';
my $hostname = '';
my $lpgsVer;
my $l1Location = '';
my $l1FileName = '';
my $l1DataFileSize = '';
my $l1DataProductFileSize = '';
my $landsatProductId = '';
my $l1_product = 0;
my $l0_product = 0;
my %albers_hash = ();

foreach my $invField (sort keys %inventorySceneData)
{
    if (not exists $inventoryFieldsToSkip{$invField})
    {
        # Print the field values - handle some differently
        if ($invField =~ /L1_LOCATION/ && $invField !~ /_AEA$/ &&
            defined($inventorySceneData{$invField}))
        {
            my $environmentuc = uc($environment);

            # Save the L1_LOCATION to print to the screen for the tester
            $l1Location = $inventorySceneData{$invField};

            # Save it for the inventory dump file diff too, but generalize
            # the storage node part of the path so that doesn't cause diffs.
            # Generalize path when on the dev system too so diff can work
            # across systems.
            my $l1LocationForFile = $inventorySceneData{$invField};
            $l1LocationForFile =~ s/^\/sno\d\d\/$environment\/l1\//\//;
            $l1LocationForFile =~ 
                s/^\/hsm\/lsat1\/$environmentuc\/collection\d\d\//\//;
            $l1LocationForFile
                =~ s/^\/pcachedev\/lpgs\/\w*\/online_cache\//\//;

            # Print the field name and generalized value
            print FILE "$invField: ";
            print FILE "$l1LocationForFile\n";
        }
        elsif ($invField =~ /DATA_TYPE/)
        {
            print FILE "$invField: $inventorySceneData{$invField}\n";

            if ($inventorySceneData{$invField} =~ /L1T/
                || $inventorySceneData{$invField} =~ /L1GT/)
            {
                $l1_product = 1;
            }
        }
        elsif ($invField =~ /BPF_NAMES/)
        {
            foreach my $innerField (sort keys %{$inventorySceneData{$invField}})
            {
                # Print the field name and value
                print FILE "BPF_NAME_$innerField: ";
                print FILE $inventorySceneData{$invField}{$innerField} . "\n";
            }
        }
        elsif ($invField =~ /CLOUD_COVER_WEIGHTS/)
        {
            foreach my $innerField (sort keys %{$inventorySceneData{$invField}})
            {
                # Print the field name and value
                print FILE "CLOUD_COVER_WEIGHTS_$innerField: ";
                print FILE $inventorySceneData{$invField}{$innerField} . "\n";
            }
        }
        elsif ($invField =~ /L0RP_FILE/)
        {
            $l0_product = 1;
            foreach my $innerField (sort keys %{$inventorySceneData{$invField}})
            {
                # Don't need to check the file type field
                next if ($innerField =~ /FILE_TYPE/);
                # Can't compare the checksum, as it will be different due to
                # internal processing date information that will differ each
                # time
                next if ($innerField =~ /FILE_CHECKSUM/);

                if ($innerField =~ /FILE_DATE/)
                {
                    # Save to print to screen later
                    $l0rpDataFileDate
                        = $inventorySceneData{$invField}{$innerField};
                }
                elsif ($innerField =~ /PRODUCT_FILE_SIZE/)
                {
                    # File size differs each run (date contents?)
                    # Save to print to screen later
                    $l0rpDataProductFileSize
                        = $inventorySceneData{$invField}{$innerField};
                }
                elsif ($innerField =~ /FILE_SIZE/)
                {
                    # File size differs each run (date contents?)
                    # Save to print to screen later
                    $l0rpDataFileSize
                        = $inventorySceneData{$invField}{$innerField};
                }
                else
                {
                    # Print the field name and value
                    print FILE "L0RP_$innerField: ";
                    print FILE $inventorySceneData{$invField}{$innerField}
                          . "\n";
                }
            }
        }
        elsif ($invField =~ /L0RP_CHECKSUM_FILE/)
        {
            foreach my $innerField (sort keys %{$inventorySceneData{$invField}})
            {
                # A few fields that aren't relevant for this file type or
                # don't need to check
                next if ($innerField =~ /FILE_TYPE/);
                next if ($innerField =~ /FILE_CHECKSUM/);
                next if ($innerField =~ /PRODUCT_FILE_SIZE/);

                if ($innerField =~ /FILE_DATE/)
                {
                    # Save to print to screen later
                    $l0rpChecksumFileDate
                        = $inventorySceneData{$invField}{$innerField};

                }
                else
                {
                    # Print the field name and value
                    print FILE "L0RP_CHECKSUM_$innerField: ";
                    print FILE $inventorySceneData{$invField}{$innerField}
                          . "\n";
                }
            }
        }
        elsif ($invField =~ /L1_FILE/)
        {
            foreach my $innerField (sort keys %{$inventorySceneData{$invField}})
            {
                # Don't need to check the file type field
                next if ($innerField =~ /FILE_TYPE/);
                # Can't compare the checksum, as it will be different due to
                # internal processing date information that will differ each
                # time
                next if ($innerField =~ /FILE_CHECKSUM/);

                if ($innerField =~ /FILE_DATE/)
                {
                    # Save to print to screen later
                    $l1DataFileDate
                        = $inventorySceneData{$invField}{$innerField};
                }
                elsif ($innerField =~ /PRODUCT_FILE_SIZE/)
                {
                    # File size differs each run (date contents?)
                    # Save to print to screen later
                    $l1DataProductFileSize
                        = $inventorySceneData{$invField}{$innerField};
                }
                elsif ($innerField =~ /FILE_SIZE/)
                {
                    # File size differs each run (date contents?)
                    # Save to print to screen later
                    $l1DataFileSize
                        = $inventorySceneData{$invField}{$innerField};
                }
                elsif ($innerField =~ /LANDSAT_PRODUCT_ID/)
                {
                    # Save to print to screen later
                    $landsatProductId
                        = $inventorySceneData{$invField}{$innerField};
                } 
                else
                {
                    if ($innerField =~ /FILE_NAME/)
                    {
                        $l1FileName
                            = $inventorySceneData{$invField}{$innerField};
                    }

                    # Print the field name and value
                    print FILE "L1_$innerField: ";
                    print FILE $inventorySceneData{$invField}{$innerField}
                          . "\n";
                }
            }
        }
        elsif ($invField =~ /L1_CHECKSUM_FILE/)
        {
            foreach my $innerField (sort keys %{$inventorySceneData{$invField}})
            {
                # A few fields that aren't relevant for this file type or
                # don't need to check
                next if ($innerField =~ /FILE_TYPE/);
                next if ($innerField =~ /FILE_CHECKSUM/);
                next if ($innerField =~ /PRODUCT_FILE_SIZE/);

                if ($innerField =~ /FILE_DATE/)
                {
                    # Save to print to screen later
                    $l1ChecksumFileDate
                        = $inventorySceneData{$invField}{$innerField};
                }
                else
                {
                    # Print the field name and value
                    print FILE "L1_CHECKSUM_$innerField: ";
                    print FILE $inventorySceneData{$invField}{$innerField}
                          . "\n";
                }
            }
        }
        elsif ($invField =~ /L0RP_LOCATION/)
        {
            # Save to print to screen later
            $l0rpLocation = $inventorySceneData{$invField};
        }
        elsif ($invField =~ /HOSTNAME/)
        {
            # Save to print to screen later
            $hostname = $inventorySceneData{$invField};
        }
        elsif ($invField =~ /PROCESSING_SOFTWARE_VERSION/)
        {
            $lpgsVer = $inventorySceneData{$invField};
        }
        elsif ($invField =~ /_AEA$/)
        {
            # Remove the _AEA that was appended to the field
            my $field = substr($invField, 0, -4);
            $albers_hash{$field} = $inventorySceneData{$invField};
        }
        else
        {
            # Print the field name and value
            print FILE "$invField: ";
            if (defined $inventorySceneData{$invField})
            {
                print FILE $inventorySceneData{$invField} . "\n";
            }
            else
            {
                print FILE "undef\n";
            }
        }
    }
}
if (!close FILE)
{
    $logger->fatal_msg("Unable to close inventory output file");
    exit FAILURE;
}

if (!$albers)
{
    # Verify the inventory dump file matches the benchmarked one
    # The dump file is  piped through
    # sed to replace the generation date with PRODDATE, so output is
    # consistent whenever the file was made. The first group in the sed
    # matches all the characters up to the gen date. Then the the date
    # is matched outside of a group. The 2nd group consists of the rest
    # of the filename and the file size.
    my $sysCmd = "cat $outFile | "
              . " sed 's/\\([[:alnum:]_]\\{26\\}\\)[[:digit:]]\\{8\\}"
               ."\\(_.*\\)/\\1PRODDATE\\2/'"
               . " | diff -q " . $ENV{IAS_IT_TEST_DATA_DIR}
               . "/test_case_$test_case_id/$outFile -";
    my $rc = system($sysCmd);
    if ($rc != 0)
    {
        $logger->fatal_msg("Diff failed: $sysCmd");
        exit FAILURE;
    }
    
    # Software version checks
    if (!defined $lpgsVer)
    {
        $logger->fatal_msg("Database software version could not be retrieved");
        exit FAILURE;
    }
    
    if ($lpgsVer ne $softwareVer)
    {
        $logger->fatal_msg("Database software version $lpgsVer does " 
            . "not match software version $softwareVer");
        exit FAILURE;
    }
    print "\nInventory data automated verification successful\n\n";
}

# Everything was successful, so remove the file, tell the tester about the
# results, and print out the L1_LOCATION for the tester
unlink $outFile;


if (!$albers)
{
    print "Inventory data for manual verification:\n";
}
else
{
    print "\nInventory data for reference:\n";
}

if ($l1Location ne '')
{
    print "L1_LOCATION: $l1Location\n";
}
else
{
    print "Error retrieving the L1_LOCATION\n" if (!$albers);
}

# Print other values that cannot be automatically verfied, for manual checking.
# The L1 info is split strangely below since a bug was fixed in this area and
# providing the output in a completely logical order would have required
# updating a number of test cases.
if ($l1_product)
{
    print "L1_PACKAGE_NAME: $l1FileName\n";
    print "PROCESSING_SOFTWARE_VERSION: $lpgsVer\n";
    print "LANDSAT_PRODUCT_ID: " . (defined $landsatProductId ? $landsatProductId : "N/A")  . "\n";
}
if ($l0_product)
{
    print "L0RP_LOCATION: $l0rpLocation\n";
    print "L0RP_PACKAGE_FILE_DATE: $l0rpDataFileDate\n";
    print "L0RP_CHECKSUM_FILE_DATE: $l0rpChecksumFileDate\n";
    print "L0RP_PACKAGE_FILE_SIZE: $l0rpDataFileSize\n";
    print "L0RP_PRODUCT_FILE_SIZE: $l0rpDataProductFileSize\n";
}
if ($l1_product)
{
    print "L1_PACKAGE_FILE_DATE: $l1DataFileDate\n";
    print "L1_CHECKSUM_FILE_DATE: $l1ChecksumFileDate\n";
    print "L1_PACKAGE_FILE_SIZE: $l1DataFileSize\n";
    print "L1_PRODUCT_FILE_SIZE: $l1DataProductFileSize\n";
    print "HOSTNAME: $hostname\n";
}
if ($albers)
{
    print "\nAlbers LMD_SCENE_PROJECT_DATA for manual verification:\n";
    foreach my $key(keys %albers_hash)
    {
        print $key . ": " . $albers_hash{$key} . "\n";
    }
    print "\n";
}

exit SUCCESS;

##############################################################################
# Name: usage
# 
# Description: Prints out the usage message for the application
#
##############################################################################
sub usage
{
    print "\nUsage: ". PROG . " [scene_id]";
    exit FAILURE;
}
