#!/usr/bin/perl -w
use strict;

####
# verify_test_results.pl
#
# Usage: verify_test_results.pl <work order id> <test directory> <db info>
#
# <work order id> should be a valid work order with a work order directory in
# $IAS_WO.
#
# The <test directory> contains configuration file (test.config) listing what
# will be verified: 
#   - [databasefiles] section lists each table to be checked. Each table is
#     listed in line ('\n') delimited format. Each line must contain the
#     table type (RADIOMETRIC or GEOMETRIC) and the table name in pipe
#     delimited format. Example: GEOMETRIC|IMAGE_CHARACTERIZATION
#     There is also an optional sorting parameter containing an order by 
#     clause. Example: GEOMETRIC|IMAGE_CHARACTERIZATION|REFERENCE_IMAGE ASC
#   - [reportfiles] section contains a list of IAS generated report files to
#     verify. Reports are listed in line ('\n') delimited format. Filenames
#     should match the files generated by IAS.  There is an option second 
#     parameter for a header cutoff string.  In the IAS generated report, any
#     text before the first occurance of the cutoff string will be removed 
#     before verifying against verification file. If provided, the cutoff 
#     string is the second parameter in a pipe delimited list.  
#     Example: i2i_char.res|L0R Reference Image.  If the cutoff string is not
#     provided, a basic file diff without running the IAS report through sed
#     will be performed.
#   - [hdffiles] section contains a list of IAS generated HDF5 image files to 
#     verify. The HDF5 files should be listed in line ('\n') delimited format.
#     The filenames should match the files generated by IAS.  
#
# All sections in test.config are optional.  If any are missing, verification 
# of that file type will be skipped. 
#
# The <test directory> should also contain verification files for each table
# and report file listed in test.config.  The database files should be 
# named with format '<TABLE_NAME>.db.verify'. 
#
# When verifying IAS generated reports and hdf5 files, the script assumes the
# files being verified are in the $IAS_WO/<work order id> directory.  The
# verification files should use the same name as the IAS reports and HDF5
# files.  
#
# <db info> should be Oracle username and password in format username/password.
# Example: value of environment variable $IAS_DB_L8_OLITIRS.  Db info is only
# required if there are [databasefiles] in test.config.
#
# Template files may be used for string replace of some IAS environment 
# variables. See @filter_vars array for a list of variables that may be used in
# a template. To create a template, append '.tpl' extension to a verfication
# file in the <test directory>.  Before the diff is performed, all instances of
# variables defined in @filter_vars will be replaced with the value of the 
# Linux environment variable of the same name. 
####

use Cwd qw(abs_path);
use File::Basename;
use DBI;
use Time::Piece;

use constant SUCCESS => 0;
use constant ERROR => 1;

# command line parameters
my $work_order_id;  # <work order id> - Work Order Id to verify
my $test_dir_path;  # <test directory> - Directory containing config and 
                    #                    verfication data
my $db_info;        # <database connection info> 
my $eval_db_info;   # optional EVAL database info

# configuration 
my $config_filename = "test.config";

# List of variables that can be used in verification data templates. Each
# variable must be a defined environment variable.   
my @filter_vars = ("IAS_IT_TEST_DATA_DIR", "IAS_WO", "IAS_IMG", "TODAYS_DATE", "IAS_SYSTEM_USER_ID");

# List columns to exclude from results
# Hash key should be the tablename and the value an array of columns
my %exclude_columns = (
    "GEOMETRIC_WORK_ORDER_COMMON" => ["DATE_PROCESSED",
                                      "SOFTWARE_VERSION_IS",
                                      "SOFTWARE_VERSION_LPGS",
                                      "SOFTWARE_VERSION_IAS"],
    "RADIOMETRIC_WORK_ORDER_COMMON" => ["DATE_PROCESSED",
                                        "SOFTWARE_VERSION_IS",
                                        "SOFTWARE_VERSION_LPGS",
                                        "SOFTWARE_VERSION_IAS"],
);

# variables
my @database_tables;
my @report_files;
my @hdf_files;
my @landsat_product_id_files;
my @cleanup_files;
my @template_files;
my $wo_full_path;
my $error_cnt = 0;
my $wo_dir = $ENV{IAS_WO};
my $dump_sql_file;
my $script_path = dirname(abs_path($0));
my $dbh;
my $result;

### Subroutine declarations
sub usage();
sub db_connect($);
sub db_close();
sub dump_table($$$$$);
sub get_columns($);
sub process_template($);
sub load_config($);
sub check_templates();
sub get_landsat_product_id_files($);

##### Process parameters ######################################################

# Verify two required command line parameters were provided
if ($#ARGV < 1)
{
    usage();
    exit 1;
}

$work_order_id = $ARGV[0]; # <work order id>
$test_dir_path = $ARGV[1]; # <test directory>

# Verify the test directory path exists
if ( !(-d $test_dir_path) )
{
    print "Invalid test directory path \n";
    usage();
    exit 1;
}

# Verify config file exists
if ( !(-f "$test_dir_path/$config_filename") )
{
    print "$config_filename not found\n";
    exit 1;
}

# Verify the work order directory exists
if ( !(-d "$wo_dir/$work_order_id") )
{
    print "Invalid work order id \n";
    usage();
    exit 1;
}

$wo_full_path = "$wo_dir/$work_order_id";

##### Verify Test Results #####################################################

$result = load_config("$test_dir_path/$config_filename");
if ($result != SUCCESS)
{
    print "ERROR loading configuration file\n";
    exit 1;
}

# If there's nothing to verify, exit the script
my $files_to_verify = scalar(@database_tables) + scalar(@report_files) 
                    + scalar(@hdf_files) + scalar(@landsat_product_id_files);
if ($files_to_verify == 0)
{
    print "Nothing to verify. Exiting.\n";
    exit;
}

# Process any template files
$result = check_templates();
if ($result != SUCCESS)
{
    print "ERROR processing templates\n";
    exit 1;
}

# Only attempt to connect to the database if there are database tables to 
# verify.
if (scalar(@database_tables) > 0)
{
    # Verify DB connection info was provided and pull it from command line args
    if ($#ARGV < 2)
    {
        print "ERROR: <database info> is required to verify [databasefiles]\n";
        usage();
        exit 1;
    }
    $db_info = $ARGV[2];       # <database info>
    if ($#ARGV >= 3)
    {
        $eval_db_info = $ARGV[3];
    }
   
     # Initiate the database connection
    $result = db_connect($db_info);
    if ($result != SUCCESS)
    {
        print "ERROR: could not connect to database\n"; 
        exit ERROR;
    }

    # Dump the database tables and verify the results
    foreach my $line (@database_tables)
    {
        my $type = $line->{"type"};
        my $table = $line->{"table"};
        my $sortclause = $line->{"sortclause"};
        my $save_dbh;

        # Handle the rare case where EVAL_ is used in the table name to 
        # indicate it is in the eval schema
        if ($table =~ /^EVAL_(.*)/)
        {
            $table = $1;
            $save_dbh = $dbh;

            if (db_connect($eval_db_info) != SUCCESS)
            {
                print "ERROR: could not connect to eval database\n"; 
                exit ERROR;
            }
        }

        my $rc = dump_table($work_order_id, $table, $type, 
                        "$table.db.results", $sortclause);
        if ($rc != SUCCESS)
        {
            print "ERROR: dumping data for table $type|$table\n";
            
            # Report this error in the results
            $error_cnt++; 

            next;
        }

        # If the eval database was connected to, disconnect and restore the
        # original
        if (defined $save_dbh)
        {
            db_close();
            $dbh = $save_dbh;
            undef $save_dbh;
        }

        # save the results file name so the file can be removed later
        push(@cleanup_files, "$table.db.results");

        my $cmd = "diff -q $table.db.results "
                . "$test_dir_path/$table.db.verify";
    
        print "Running: $cmd\n";

        $rc = system($cmd);
        if ($rc != 0)
        {   
            print "Diff failed: $table\n";
            $error_cnt++;
        }

        print "\n";
    }

    # Done with the database
    db_close();
}

# Verify report files generated by IAS
foreach my $file (@report_files)
{
    my $filename = $file->{"filename"};
    my $cutoff_str = $file->{"cutoff_str"};
    my $cmd;

    if (defined $cutoff_str)
    { 
        if ( $filename eq "omf" || $filename eq "product_data.txt")
        {
            # Special handling of OMF and product_data.txt comparisons (which
            # should be generalized).  Change the comma delimited list to a
            # pipe delimited one for egrep
            $cutoff_str =~ s/,/|/g;

            # Set the source name for the OMF case
            my $source_filename = $filename;
            if ($filename eq "omf")
            {
                $source_filename = $work_order_id . ".omf";
            }

            # Use egrep to remove the line(s) specified
            $cmd = "egrep -v \"($cutoff_str)\" $wo_full_path/$source_filename";
            $cmd .= " | diff - $test_dir_path/$filename";

            my $print_cmd = "egrep -v \"($cutoff_str)\" \$IAS_WO/";
            $print_cmd .= "$work_order_id/$source_filename | ";
            $print_cmd .= "diff - $test_dir_path/$filename";
            print "Running: $print_cmd\n";
        }
        else
        {
            # Use sed to remove info that changes with each report run (time,
            # work order id, etc)
            $cmd = "sed '/$cutoff_str/,\$!'d $wo_full_path/$filename ";
            $cmd .= "| diff -q - $test_dir_path/$filename";

            my $print_cmd = "sed '/$cutoff_str/,\$!'d \$IAS_WO/$work_order_id";
            $print_cmd .= "/$filename | diff -q - $test_dir_path/$filename";
            print "Running: $print_cmd\n";
        }
    }
    else
    { 
        # These are files that shouldn't change with each run so diff files
        # without running IAS file through sed
        $cmd = "diff -q $wo_full_path/$filename $test_dir_path/$filename";

        my $print_cmd = "diff -q \$IAS_WO/$work_order_id/$filename ";
        $print_cmd .= "$test_dir_path/$filename";
        print "Running: $print_cmd\n";
    }

    my $rc = system($cmd);
    if ($rc != 0)
    {
        print "Diff failed: $filename\n";
        $error_cnt++;
    }

    print "\n";
}

# Verify the IAS files names with the landsat_product_id
if (scalar(@landsat_product_id_files) > 0)
{
    # Get the corresponding files in the wo directory 
    my %work_orders = %{get_landsat_product_id_files($wo_full_path)};
    # Get the corresponding files in the expected results directory
    my %expected_results = %{get_landsat_product_id_files($test_dir_path)};
    
    foreach my $file(@landsat_product_id_files)
    {
        my $filename = $file->{"filename"};
        my $cmd;
        my $print_cmd;

        my $work_order_file = $work_orders{$filename};
        my $expected_results_file = $expected_results{$filename};

        if ($filename !~ /MTL/)
        {
            $cmd = "diff -q $wo_full_path/$work_order_file ".
                "$test_dir_path/$expected_results_file";

            $print_cmd = "diff -q \$IAS_WO/$work_order_id/$work_order_file";
            $print_cmd .= " $test_dir_path/$expected_results_file";  
            print "Running: $print_cmd\n";
        }
        else
        {
            # First remove any lines in MTL file that change during each
            # processing run
            $cmd = "sed '/FILE_NAME\\|REQUEST_ID\\|FILE_DATE\\|" .
                   "LANDSAT_PRODUCT_ID\\|SOFTWARE_VERSION/d' " .
                    "$wo_full_path/$work_order_file" ;
            $cmd .= "| diff - $test_dir_path/$expected_results_file";
            $print_cmd = "sed '/FILE_NAME\\|REQUEST_ID\\|FILE_DATE\\|".
                 "LANDSAT_PRODUCT_ID\\|SOFTWARE_VERSION/d' " .
                 "\$IAS_WO/$work_order_id";
            $print_cmd .= "/$work_order_file | diff - ".
                "$test_dir_path/$work_order_file";
            print "Running: $print_cmd\n";
        }
        my $rc = system($cmd);
        if ($rc != 0)
        {
            print "Diff failed\n";
            $error_cnt++;
        } 
        print "\n";
    }
}

# Verify HDF5 files generated by IAS
foreach my $hdf_file (@hdf_files)
{
    # Get the list of datasets in the comparison file
    my @contents = `h5dump -n $test_dir_path/$hdf_file`;
    my @expected_datasets = ();
    foreach (@contents) 
    {
        /^\s*dataset\s+(.+)/ and do
        {
            push @expected_datasets, $1;
        };
    }

    # Get the list of datasets in the wo file
    @contents = `h5dump -n $wo_full_path/$hdf_file`;
    my @found_datasets = ();
    foreach (@contents) 
    {
        /^\s*dataset\s+(.+)/ and do
        {
            push @found_datasets, $1;
        };
    }
    # Make sure the two files contain the same datasets
    if (join(',', sort @expected_datasets)
        ne join(',', sort @found_datasets))
    {
        print " Diff failed: $hdf_file (does not contain expected datasets)\n";
        $error_cnt++;
        next;
    }

    # Use h5diff to verify the HDF5 files generated by IAS
    my $cmd = "h5diff -c -q $test_dir_path/$hdf_file $wo_full_path/$hdf_file";
    
    my $print_cmd = "h5diff -c -q \$IAS_WO/$work_order_id/$hdf_file " .
        "$test_dir_path/$hdf_file";

    print "Running: $print_cmd\n";

    # Compare each dataset 
    foreach my $dataset (@expected_datasets)
    {
        # Use special handling for File & General Metadata
        # Dump the metadata & delete lines with IAS & Ingest SW version
        if ($dataset eq "/File Metadata" or
            $dataset eq "/General Metadata") 
        {
            my $hdf_file_base = basename($hdf_file);

            my $metadata_ref = "metadata_ref_$hdf_file_base";
            my $metadata_wo = "metadata_${work_order_id}_$hdf_file_base";
            system(qq{h5ls -l -w 10 -d "$test_dir_path/$hdf_file/$dataset" | 
                grep -v Software.Version > $metadata_ref});
            system(qq{h5ls -l -w 10 -d "$wo_full_path/$hdf_file/$dataset" | 
                grep -v Software.Version > $metadata_wo});

            my $rc = system(qq{cmp -s $metadata_ref $metadata_wo});
            if ($rc != 0)
            {
                print qq{ Diff failed: $hdf_file "$dataset"\n};
                $error_cnt++;
            }
            else
            {
                system("rm -f $metadata_ref");
                system("rm -f $metadata_wo");
            }
        } 
        elsif ($dataset =~ /^\/B[0-9]{2}/)
        {
            my $clean_dataset = $dataset;
            $clean_dataset =~ s/\///g;

            my $hdf_file_base = basename($hdf_file);

            my $tmp_wo_file = "$hdf_file_base-$clean_dataset.wo";
            my $tmp_base_file = "$hdf_file_base-$clean_dataset.base";

            # Dump dataset from the WO file
            my $h5dump_cmd = "h5dump -b NATIVE -d $dataset -o $tmp_wo_file "
                . "\$IAS_WO/$work_order_id/$hdf_file > /dev/null";
            my $rc = system($h5dump_cmd);
            if ($rc != 0)
            {
                print "Failed h5dump on WO file: $hdf_file \n";
                print qq{ Diff failed: $hdf_file "$dataset"\n};
                $error_cnt++;

                system("rm -f $tmp_wo_file");
                next;
            }

            # Dump dataset from the baseline file
            $h5dump_cmd = "h5dump -b NATIVE -d $dataset -o $tmp_base_file "
                . "$test_dir_path/$hdf_file > /dev/null";
            $rc = system($h5dump_cmd);
            if ($rc != 0)
            {
                print "Failed h5dump on baseline file: $hdf_file \n";
                print qq{ Diff failed: $hdf_file "$dataset"\n};
                $error_cnt++;

                system("rm -f $tmp_wo_file");
                system("rm -f $tmp_base_file");
                next;
            }

            # execute the diff on the dump files
            if (system("diff -q $tmp_wo_file $tmp_base_file"))
            {
                print qq{ Diff failed: $hdf_file "$dataset"\n};
                $error_cnt++;
            }
            else
            {
                system("rm -f $tmp_wo_file");
                system("rm -f $tmp_base_file");
            }
        }
        # All other datasets, use plain h5diff
        else 
        {
            # Capture output, in case a dataset is flagged as not comparable
            my $output = qx{$cmd "$dataset"};
            if ($? != 0 or $output =~ /Not comparable/)
            {
                print qq{ Diff failed: $hdf_file "$dataset"\n};
                $error_cnt++;
            }
        }

    }
    print "\n";
}

# Check for errors
if ($error_cnt > 0)
{
    print "Running diffs failed\n";
    exit 1;
}
else
{
    # Test was successful. Remove any files marked for cleanup.
    foreach my $file (@cleanup_files)
    {
        my $rc = system("rm", "-rf", $file);
        if ($rc != 0)
        {   
            print "WARN: could not remove file: $file\n";
        }
    }

    print "Running diffs passed\n";
    exit 0;
}

#### Subroutines ##############################################################

## usage
sub usage()
{
    print <<USAGE
Usage: verify_test_results.pl <work order id> <test directory> <database info>
  <work order id> = Work Order Id
  <test directory> = Directory containing test.config and verification data
  <database info> - Database connection information (i.e \$IAS_DB_L8_OLITIRS)
USAGE
}

## db_connect
sub db_connect($)
{
    my ($dbinfo) = @_;
    my ($user, $pass) = split(/\//, $dbinfo);
    my $sid = $ENV{"ORACLE_SID"};

    if (!defined $sid)
    {
        print "ERROR: ORACLE_SID not defined\n";
        return ERROR;
    }

    $dbh = DBI->connect("DBI:Oracle:$sid", $user, $pass,
                         {AutoCommit => 0, PrintError => 0});

    if (!$dbh)
    {
        print "ERROR: Error connecting to the database: ".$DBI::errstr."\n";
        return ERROR;
    }

    return SUCCESS;
}

## db_close
sub db_close()
{
    $dbh->disconnect();
}

## dump_table
# Dump database table information for a single work order to a file. 
sub dump_table($$$$$)
{
    my ($work_order_id, $tablename, $table_type, $outfilename, $sort) = @_;

    my $source_table;
    my $link_column;
    my @columns;

    # Use table type to figure which table and column to use to find data
    if ($table_type eq "GEOMETRIC")
    {
        $source_table = "GEOMETRIC_WORK_ORDER_COMMON";
        $link_column = "GEOMETRIC_WO_COMMON_ID";
    }
    elsif ($table_type eq "RADIOMETRIC")
    {
        $source_table = "RADIOMETRIC_WORK_ORDER_COMMON";
        $link_column = "RADIOMETRIC_WO_COMMON_ID";
    }

    # Get list of column in the table
    @columns = get_columns($tablename);

    # Verify columns were found for this table
    if (scalar(@columns) == 0)
    {
        print "ERROR: no columns found for table: $tablename\n";
        return ERROR;
    }

    my $column_str = join(",", @columns);

    my $sql = "SELECT $column_str "
            . "  FROM $tablename "
            . " WHERE $link_column = "
            . "       (SELECT $link_column "
            . "          FROM $source_table "
            . "         WHERE work_order_id = '$work_order_id')";

    if (defined $sort)
    {
        $sql .= " ORDER BY $sort";
    }

    my $sth = $dbh->prepare($sql);
    if (!$sth)
    {
        print "ERROR: parsing dump sql ($sql): ";
        return ERROR;
    }

    $sth->execute();

    open(OUTF, ">$outfilename");

    # Write the column header
    print OUTF join("|", @columns) . "\n";

    # Write table data
    while (my @data = $sth->fetchrow_array())
    {
        # Set any undefined values to empty string. This prevents uninitialized
        # value in join warnings.
        foreach my $val (@data)
        {
            if (not defined $val)
            {
                $val = '';
            }
        }

        print OUTF join("|", @data) . "\n";
    }
    $sth->finish();

    close(OUTF);

    return SUCCESS;
}

## get_columns
# Return array of columns for a given table name.
# Doesn't return columns contain large binary data or columns with any type of
# unique constraint. 
sub get_columns($)
{
    my ($tablename) = @_;
    my @columns;

    $tablename = uc($tablename);

    # Check for exclude columns
    my $exclude_columns_clause = "";
  
    # If this table has columns to be excluded 
    if (exists $exclude_columns{$tablename})
    { 
        # Get array of columns and add an exclude clause for each column
        my $ex_columns = $exclude_columns{$tablename};
        foreach my $col (@$ex_columns)
        {
            $exclude_columns_clause .= " AND column_name <> '$col' ";
        }
    }

    # This query gets the distinct column names for the given table,
    # based on all the tables of that name that this schema
    # has access to (has been given a grant to see).
    # If the table exists in two schemas with different columns,
    # the inclusive set of all column names from both tables is returned.
    # This is probably not what you want (trying to use all these columms will 
    # probably result in an error), but it's impossible to predict which table 
    # you're really interested in, and this situation indicates that something 
    # is not correct in the database.
    my $sql = "SELECT column_name " 
            . "  FROM all_tab_columns "
            . " WHERE table_name = '$tablename' "
            . "   AND data_type NOT IN ('CLOB', 'BLOB', 'RAW') "
            . "   $exclude_columns_clause " 
            . " GROUP BY column_name"
            . " ORDER BY min(column_id)";

    my $sth = $dbh->prepare($sql);

    $sth->execute();

    my @tmp_columns;
    while (my ($column) = $sth->fetchrow_array())
    {
        push(@tmp_columns, $column);
    }
    $sth->finish();

    # Get a list of unique key columns to exclude from the list
    $sql = "SELECT unique(cc.column_name) "
         . "  FROM all_constraints c, "
         . "       all_cons_columns cc "
         . " WHERE c.constraint_name = cc.constraint_name "
         . "   AND c.table_name = '$tablename' "
         . "   AND c.constraint_type IN ('P', 'U')";

    $sth = $dbh->prepare($sql);
    $sth->execute();

    my %exclude_columns;
    while (my ($column) = $sth->fetchrow_array())
    {
        $exclude_columns{$column} = 1;
    }
    $sth->finish();

    # Build the column list
    foreach my $column (@tmp_columns)
    {
        push(@columns, $column) if (not defined $exclude_columns{$column});
    }

    return @columns;
}

## process_template
# Run search and replace on template file.  Write the output to filename minus
# .tpl extension.
sub process_template($)
{
    my ($filename) = @_;

    # Remove .tpl extension for output filename
    my $outfile = $filename;
    $outfile =~ s/.tpl$//;

    if (!open(FH, "<$filename"))
    {
        print "ERROR: opening template: $filename\n";
        return ERROR;
    }

    if (!open(OFH, ">$outfile"))
    {
        print "ERROR: opening output file: $outfile\n";
        close(FH);
        return ERROR;
    }

    # Read the entire template file.  Right now the verification files are
    # not large enough to cause any problems reading the entire file. If the
    # files become too large, this may need to be changed to reading the file
    # one line at a time.  
    my $filedata;
    {
        local($/);
        $filedata = <FH>;
    }

    # Do search and replace for each variable in the filter list
    foreach my $var (@filter_vars)
    {
        my $search = "\\\$$var";
        my $replace;
        if ($var eq "TODAYS_DATE")
        {
            # Get todays date
            my $time = localtime();
            $replace = sprintf("%02d-%02d-%4d", $time->mon, $time->mday,
                $time->year);
        }
        else
        {
            $replace = $ENV{$var};
        }
        if ( not defined $replace )
        {
            print "ERROR: template variable not defined: $var\n";
            return ERROR;
        }
        
        $filedata =~ s/$search/$replace/g;
    }
   
    # Write data to output file 
    print OFH $filedata;

    close(OFH);
    close(FH);
    
    return SUCCESS;
}

## load_config
# Accepts the configuration filename as a parameter.  Loads the global test
# verification variables and arrays.
# Returns ERROR or SUCCESS
sub load_config($)
{
    my ($filename) = @_;
    my $section;

    if (!open(FH, "<$filename"))
    {
        print "ERROR: couldn't open config file: $filename\n";
        return ERROR;
    }

    while(<FH>)
    {
        chomp;

        # Skip comments and blank lines
        if ($_ =~ /^\s*(#|$)/) { next; }

        # Trim whitespace from the beginning and end of the line
        $_ =~ s/(^\s+|\s+$)//g;

        # Check for new section
        if ($_ =~ /^\[([\w\s]+)\]$/)
        {
            $section = $1;
            next;
        }

        # Create hash for each object (report, database, hdf5)
        if ($section eq "reportfiles")
        {
            my ($filename, $cutoff_str) = split(/\|/, $_);

            my $file = {
                filename => $filename,
                cutoff_str => $cutoff_str
            };

            push(@report_files, $file);
        }
        elsif ($section eq "databasefiles")
        {
            my ($type, $table, $sortclause) = split(/\|/, $_);

            my $file = {
                type => $type,
                table => $table,
                sortclause => $sortclause
            };

            push(@database_tables, $file);
        }
        elsif ($section eq "hdffiles")
        {
            my $filename = $_;
            push(@hdf_files, $filename);
        }
        elsif ($section eq "LandsatProductId")
        {
            my ($filename, $cutoff_str) = split(/\|/, $_);

            my $file = {
                filename => $filename,
                cutoff_str => $cutoff_str
            };
            push(@landsat_product_id_files, $file);
        }
        else
        {
            print "Invalid section header [$section] in $config_filename\n";
            return ERROR;
        }
    }

    close(FH);
    
    return SUCCESS;
}

## check_templates
# Checks for and processes database and report templates
sub check_templates()
{
    # Process database templates
    foreach my $dbtable (@database_tables)
    {
        my $table = $dbtable->{"table"};
        my $template = "$test_dir_path/$table.db.verify.tpl";

        if ( -f $template )
        {
            push(@template_files, $template);

            my $rc = process_template($template);
            if ($rc == SUCCESS)
            {
                # Remove .tpl extension to get file generated by 
                # process_template and add it to the list of files to be
                # cleaned up if the test completes successfully
                $template =~ s/.tpl$//;
                push(@cleanup_files, $template);
            }
            else
            {
                print "ERROR processing template: $template\n";
                return ERROR;
            }
        }
    }

    # Process report templates
    foreach my $report (@report_files)
    {
        my $filename = $report->{"filename"};
        my $template = "$test_dir_path/$filename.tpl";

        if ( -f $template )
        {
            push(@template_files, $template);
            
            my $rc = process_template($template);
            if($rc == SUCCESS)
            {
                # Remove .tpl extension to get file generated by 
                # process_template and add it to the list of files to be 
                # cleaned up if the test completes successfully
                $template =~ s/.tpl$//;
                push(@cleanup_files, $template);
            }
            else
            {
                print "ERROR processing template: $template\n";
                return ERROR;
            }
        }
    }

    # No template support for HDF files

    return SUCCESS;
}

# Gets the files from path provided that are named with the landsat product id
sub get_landsat_product_id_files($)
{
    my $path = shift;
    my %wo_files = ();

    # Get all the files in the work order path
    opendir(WO, $path) or die ("Can't open $path\n");
    my @work_order_files = readdir(WO);
    closedir(WO);

    # Go through all the files to find the ones that are named with the
    # Landsat product ID
    foreach my $file (@work_order_files)
    {
        # Skip any hidden files
        next if($file =~ /^\./);

        # Build landsat_product_id format string
        my $landsat_product_id_format =
            "(\\w{4})_(\\w{4})\_(\\d{6})_(\\d{8})_(\\d{8})_(\\d{2})_".
            "(\\w{2})_(\\w{2,}\.\\w{3})";
        if ($file =~ /($landsat_product_id_format)/)
        {
            # If there are multiple files in the directory with the same 
            # extension:
            # ie an order was reran and now there are 2 files named with the
            # landsat product ID with B1.TIF extension, etc
            if (exists($wo_files{$9}))
            {
                # Create time piece objects in order to do subtraction
                # Parse out times of file already stored in hash
                my $process_date = substr($wo_files{$9}, 26, 8);
                my $first_time = Time::Piece->strptime($process_date, "%Y%m%d");
                my $second_time = Time::Piece->strptime($6, "%Y%m%d");
                # If the second one is more recent than the one already stored
                # in hash, replace
                if (($second_time - $first_time) > 0)
                {
                    $wo_files{$9} = $file;
                }
            }
            else
            {
                $wo_files{$9} = $file;
            }
        }
    }

    return \%wo_files;
}

