#!/usr/bin/perl -w

# Distribute LPGS software to processing nodes
#  List of processing nodes is retrieved from the database (processing_nodes 
#    table)
#  A list of files or all of $IASBUILD may be distributed.  The directory
#  structure is assumed to be the same on the remote host.
# COTS can be distributed by this script, for example:
#  lpgs_distribute_execs.pl -d /home/scmit/it/COTS64/ -f /home/scmit/it/COTS64/

use strict;
use Getopt::Long;
use DBI;
use Sys::Hostname;

my $help;
my $dist_path;
my @files_to_distribute;
GetOptions("help!"       => \$help,
           "dpath=s"     => \$dist_path,
           "file=s"      => \@files_to_distribute
           ) or die "Unable to process command line options\n";

if (defined($help))
{
    print "Usage: $0 [-d distribution path] [-f file] ...\n";
    exit 0;
}
# Verify our destination
unless (defined($dist_path))
{
    if (defined($ENV{IASBUILD}))
    {
        $dist_path = $ENV{IASBUILD};
    }
    else
    {
        die "No destination provided.  Specify -d argument or set IASBUILD\n";
    }
}

my $two_task = $ENV{ 'TWO_TASK' } or die("TWO_TASK is not set");
my $comm_db_un = $ENV{ 'IAS_DB_COM' } or die("IAS_DB_COM is not set");
my ($user_id, $password) = split '/', $comm_db_un;

# Connect to Oracle.
my $dbh = DBI->connect( "dbi:Oracle:$two_task", $user_id, ,$password, 
    { AutoCommit => 0 } ) ||
    die( "Couldn't connect to database!" );

# Get the list of hosts from the processing nodes table
my @hosts = ();
my $sth = $dbh->prepare( qq{ SELECT hostname FROM processing_nodes });
$sth->execute();

while ( my $hostname = $sth->fetchrow() ) 
{
    # Skip the local host, only distribute to remote hosts
    next if (uc($hostname) eq uc(hostname()));
    push @hosts, $hostname;
}

$sth->finish();
$dbh->disconnect;

# First make sure the dist path exists on the remote host(s)
my $output;
foreach my $hostname ( @hosts ) 
{
    $output = qx{ ssh $hostname mkdir -p $dist_path 2>&1 };
    if ($?)
    {
        print "Error creating $dist_path on $hostname: $output\n";
        exit 1;
    }
}

# If no command-line arguments were supplied, distribute all of $IASBUILD
if ($#files_to_distribute < 0)
{
    # Verify the source location
    die "ERROR: \$IASBUILD is not defined.  Nothing to distribute.\n"
        unless defined($ENV{IASBUILD});

    foreach my $hostname ( @hosts ) 
    {
        print "Distributing \$IASBUILD to $dist_path on $hostname...\n";
        # add slash to rsync SRC to prevent putting into a subdir of dist_path
        $output = qx{ rsync -a --exclude=/src/ $ENV{IASBUILD}/ $hostname:$dist_path 2>&1 };
        if ($?)
        {
            print "Error copying to $hostname: $output\n";
            exit 1;
        }
    }
}

# Distribute files (command-line args) to all hosts
#  Use the same path on the remote host, unless $dist_path is defined, then
#  replace $IASBUILD with $dist_path
while ( my $dist_file = pop( @files_to_distribute ) ) 
{
    if (! -e $dist_file)
    {
        warn "WARNING: $dist_file does not exist.  Skipping.\n";
        next;
    }
    # Get the remote file name: replace IASBUILD prefix with the remote
    # dist path 
    my $remote_file = $dist_file;
    $remote_file =~ s/^$ENV{IASBUILD}/$dist_path\//;

    foreach my $hostname ( @hosts ) 
    {
        print "Distributing $dist_file to $hostname:$dist_file...\n";
        $output = qx{ rsync -a $dist_file $hostname:$remote_file 2>&1 };
        if ($?)
        {
            print "Error copying file: $output\n";
            exit 1;
        }
    }
}

exit 0;
