/******************************************************************************
               Level 1 Product Generation System (LPGS) Software
******************************************************************************
 *
 * UNIT NAME:   EventLog.h
 *
 * PURPOSE:
 * This is the include file for the EventLog class
 *
 *****************************************************************************/

#ifndef EVENTLOG_H
#define EVENTLOG_H

/* System Includes */
#include <stdio.h>
#include <stdlib.h>
#include <string>
#include <time.h>
#include <unistd.h>
#include <iostream>

/* Qt Includes */
#include <QtCore>
#include <QtGui>

/* Local Includes */
#include "ui_EventLogBase.h"
#include "IASDatabaseErrorDialog.h"
#include "DateRange.h"
#include "TreeWidgetUtilities.h"
#include "ColoredTreeWidgetItem.h"
#include "IASSqlQuery.h"

// defines for database column names
#define sDATE_TIME      "EVENT_DATE"
#define nDATE_TIME      0
#define sPR_ID          "PRODUCT_REQUEST_ID"
#define nPR_ID          1
#define sWO_ID          "WORK_ORDER_ID"
#define nWO_ID          2
#define sSCRIPT_NAME    "SCRIPT_ID"   
#define nSCRIPT_NAME    3
#define sTASK_NAME      "PROGRAM_ID" 
#define nTASK_NAME      4 
#define sMSG            "MESSAGE_TEXT"
#define nMSG            5
#define sSEQ_NUM        "SEQ_NBR" 
#define nSEQ_NUM        6
#define sSEV            "SEVERITY"     
#define nSEV            7
#define sMSG_NUM        "MESSAGE_ID" 
#define sEVENT_COMMENT  "EVENT_COMMENT"

// defines for database table names
#define sEVENTS_TABLE   "events"
#define sMESSAGES_TABLE "messages"

class EventLog : public QDialog, public Ui::EventLogBase
{ 
    Q_OBJECT

    public:
        EventLog(QWidget *parent = 0, Qt::WindowFlags fl = 0);

        EventLog(QString strPRID = "",
                 QWidget *parent = 0,
                 Qt::WindowFlags fl = 0);

    private:
        
            // create empty list view, add columns, set alignments, etc.
            // called once
        void buildEventListView();
            // build the where clause for querying the database, the clause
            // is based on selections made by the operator such as dates,
            // message types, and possible a certain task name, product 
            // request id, etc.
        QString buildWhereClause();
            // This function calculates the number of days from current
            // day to start query date, and the number of days from 
            // current date to stop query date.  A TRUE will be returned
            // if the dates are valid                        
        bool calcNumberofDays(bool notify,
                              int *daysToStart,
                              int *daysToEnd);
            // This function clears the event list and fills it with the 
            // data in the events RecordSet.
        void fillEventList();
            // get list of events from events table in database
        void getEvents();
            // return column (number) selected combo box
        int  getSelectedColumn();
            // get the text for the selected item from the column that is
            // selected at the top of the form
        QString getSelectedColumnText();
            // get list of items from database to fill the list of possible
            // selections the operator selects it
        void getSelectItems();
            // set the selected list view item to the value that was 
            // prevoiusly saved.
        void reselectLastItemSelected();
            // update the Date and Time at top of GUI
        void updateDateTime();
            // set the Detailed Message label to the value for the selected
            // list view item
        void updateDetailedMessage();
            // update the number of items in event list returned at top of GUI
        void updateNumReturned(QString);
            // update label to show the number of items selected
        void updateSelect(QString);

            // Object to print event list to
        QPrinter *printer;
            // all events read from database
        QListView* lvEvents;
            // string used for processing in various functions.
        QString strTemp;
            // storage for creating SQL commands
        QString strSQLCommand;
            // the sequence number of the currently selected item in the event 
            // list view it is saved so that the currently selected item can be
            // restored to again be the selected item when the refresh button
            // is selected
        QString strSelectedSequenceNumber;
            // Holds the events from the query;
        IASSqlQuery eventsQuery;
            // when true the background and text colors are reversed.  Normal
            // would be red, yellow, and green lettering for error, warnings,
            // and status, with a white backgound.  If swapColors is selected,
            // the lettering is white and background is colored.
        bool swapColors;
            // structure holding column information for the event log
            // twInfo is defined in TreeWidgetUtilities.h    
        static twInfo elInfo[];    
            // string holding the table names needed to query database
        QString strTables; 
            // string containing the SQL join string
        QString strJoin;
            // number of columns in the event log, not all columns returned
            // by the query are displayed in the event list. (detailed 
            // message, severity, and sequence number are not displayed)
        int numColumns;

    private slots: 
            // setup the siganls/slots for the GUI 
        void setupSignalSlots();
            // copy the selected item into the list of all items selected
        void addItem();
            // slot, all (types) of events is selected, will clear the other
            // message types check boxs if they are checked
        void allSelected();
            // clear list of all selected items
        void clearSelectedList();
            // slot operator selected errors, all is cleared if selected
        void errorsSelected();
            // functions saves the sequence number of the currently selected 
            // item when the operator selects it
        void eventOperatorSelected(QTreeWidgetItem *);
            // This function updates the detailed message for selected item
        void eventSelectionChanged();
            // print the event list to a selected printer
        void print();
            // query on selected item
        void queryOnItem();
            // requery the database and update the list
        void refreshList();
            // The All checkbox will be cleared, if the Status checkbox is
            // selected
        void statusSelected();
            // swap text and background colors in Event List 
        void swapEventListColors();
            // if a different column has been selected, update the select list
            // with appropriate values from the database
        void updateSelectList();
            // The All checkbox will be cleared, if the Warnings checkbox is
            // selected
        void warningsSelected();
};

#endif /* EVENTLOG_H */

