#ifndef PRODUCTREQUESTTREEWIDGET_H
#define PRODUCTREQUESTTREEWIDGET_H
/******************************************************************************
               Level 1 Product Generation System (LPGS) Software
 ******************************************************************************
 *
 * UNIT NAME:   ProductRequestTreeWidget.h
 *
 * PURPOSE:
 * The ProductRequestTreeWidget class extends the generic QTreeWidget class by
 * providing a list of colored ProductRequestTreeWidgetItems, along with 
 * context-sensitive popup menus that can be invoked with a right mouse
 * click.  From these menus, the operator is given access to several
 * options (which depend on what state the product request is in.)
 *
 *****************************************************************************/
/* System Includes */

/* Qt Includes */
#include <QtCore>
#include <QtGui>

/* Local Includes */
#include "DateRange.h"

/* Must match the columnTitles and alignment variables
   This must also match the select columns from the work_orders select in the
   refresh method */
enum productRequestColumns
{
    PRODUCT_REQUEST_ID = 0,
    LANDSAT_SCENE_ID,
    PROCESSING_INSTANCE,
    STATE,
    WORK_ORDER_ID,
    PRIORITY,
    PROCESSING_NODE,
    CANCELLATION_STATUS,
    DATE_REQUESTED,
    SATELLITE,
    SENSOR,
    PROCEDURE_ID,
    PR_COLUMN_COUNT
};

/* Product request partial row data. Only the columns of data that are used are
   currently included to avoid carrying around unneeded data */
typedef struct
{
    QString state;
    QString work_order_id;
    QString product_request_id;
} PRODUCT_REQUEST_ROW;

class ProductRequestTreeWidget : public QTreeWidget 
{
    Q_OBJECT

    public:
        ProductRequestTreeWidget(QWidget *parent = 0,
                                 DateRange *drMain = 0);
    
    signals:
        void countChanged();

    public slots:
        void refresh();
        void openUNIXShell();
        QString getCurrentProductRequestId();
       
    private slots:
        void copyPRIdToClipboard(QTreeWidgetItem *);
        void displayAnomalyInformation();
        void displayWorkOrderScripts(bool);
        void displayEventLog();
        void retryProductTransfer();
        void retryL0RRequest();
        void refreshContextMenu();
        void showUserParameters();
        void failCancelRequest();
        void reprioritize(bool);
        void sendEmail();
        void resumeRequest();
        void viewSelectedWorkOrderLog(bool);
        void setSatelliteFilter(const QString &);
        void setSensorFilter(const QString &);
        void setInstanceFilter(const QString &);
        void setProcedureTypeFilter(const QString &);
        QList<PRODUCT_REQUEST_ROW> getSelectedRows(const QString stateFilter);
        QStringList buildColumnList(const QList<PRODUCT_REQUEST_ROW> &rows, 
            const productRequestColumns joinColumn);

    private:
        void buildContextMenus();

        QMenu *popmPriority;          // "Reprioritize" popup menu
        QMenu *popmWorkOrderScripts;  // Work order id's for the PR id
        QMenu *popmWorkOrdersLogs;

        QAction *actionUserParameters;
        QAction *actionDisplayEventLog;

        QAction *actionWorkOrderScripts;
        QAction *actionWorkOrderLogs;
        QAction *actionOpenUnixShell;
        QAction *actionChangePriority;

        QAction *actionSeparator1;
        QAction *actionRetryProductTransfer;
        QAction *actionRetryL0RRequest;
        QAction *actionAnomalyInformation;

        QAction *actionSeparator2;
        QAction *actionResume;
        QAction *actionCancelPR;

        QAction *actionSeparator3;
        QAction *actionSendEmail;

        QStringList strlWorkOrderIDs; // List containing work order ids

        DateRange *dr; 

        QString satelliteFilter;
        QString sensorFilter;
        QString procedureTypeFilter;
        QString instanceFilter;
        bool resizeColumns;
};

#endif /* PRODUCTREQUESTLISTVIEW_H */

