/******************************************************************************
               Level 1 Product Generation System (LPGS) Software
 ******************************************************************************
 *
 * UNIT NAME:   AnomalyDetail.cpp
 *
 * PURPOSE:
 * The Anomaly Detail dialog shows the operator various information about
 * a particular anomaly.  This includes:
 *
 * 1) the anomaly number
 * 2) the assignee / responsible party
 * 3) the anomaly description
 * 4) the resolution of the anomaly
 *
 *****************************************************************************/

/* System Includes */

/* Qt Includes */

/* Local Includes */
#include "IASSqlQuery.h"

#include "AnomalyDetail.h"

#include "IASErrorDialog.h"

AnomalyDetail::AnomalyDetail(
    QString strAnomalyId,   // I: Anomaly to view details for
    QWidget *parent,        // I: Parent widget
    Qt::WindowFlags fl)     // I: Additional Qt flags
: QDialog(parent, fl)
{
    setupUi(this);

    setModal(true);

    lblAnomalyId->setText(strAnomalyId);

    connect(pbOK, SIGNAL(clicked()), this, SLOT(accept()));
    connect(pbCancel, SIGNAL(clicked()), this, SLOT(reject()));
    connect(leAssignee, SIGNAL(textChanged(const QString &)),
            this, SLOT(enableApplyButton()));
    connect(teDescription, SIGNAL(textChanged()),
            this, SLOT(enableApplyButton()));
    connect(teResolution, SIGNAL(textChanged()),
            this, SLOT(enableApplyButton()));
    connect(pbApply, SIGNAL(clicked()), this, SLOT(applyChanges()));

    // Grab the details for this anomaly from the database.
    QString sql = "select assignee, problem_description, problem_resolution"
                  " from anomalies"
                  " where anomaly_id = '" + strAnomalyId + "'";

    IASSqlQuery query;
    bool result = query.exec(sql, __FILE__, __LINE__);
    
    if (result && query.first())
    {
        leAssignee->setText(query.value("assignee").toString());
        teDescription->setPlainText(
            query.value("problem_description").toString());
        teResolution->setPlainText(
            query.value("problem_resolution").toString());

        pbApply->setEnabled(FALSE);
    }
}

/*****************************************************************************/

void AnomalyDetail::enableApplyButton()
{
    // This function is invoked any time one of the QLineEdit or
    // QMultiLineEdit widgets within this dialog has its contents
    // changed; it enables the Apply button so that it can be pressed.

    pbApply->setEnabled(TRUE);
}

/*****************************************************************************/

void AnomalyDetail::applyChanges()
{
    // This function just pulls all of the various information out of
    // the line edit / multiline edit widgets and inserts it into the
    // ANOMALIES table for the current anomaly ID.

    QString strAssignee = leAssignee->text();

    // Write out the contents of the Description and Resolution widgets to the
    // anomalites table

    QString strDescription = teDescription->toPlainText();
    QString strResolution = teResolution->toPlainText();

    QString sql = "update anomalies"
                  " set assignee = :assignee,"
                      " problem_description = :description,"
                      " problem_resolution = :resolution"
                  " where anomaly_id = :anomaly_id";
    // Update the anomaly detail.
    QSqlQuery query;
    query.prepare(sql);
    query.bindValue(":assignee", strAssignee);
    query.bindValue(":description", strDescription);
    query.bindValue(":resolution", strResolution);
    query.bindValue(":anomaly_id", lblAnomalyId->text());
    if (!query.exec())
    {
        IASErrorDialog::databaseError(0, "A database error occurred while "
                                      "trying to execute an SQL statement",
                                   __FILE__, __LINE__, &query);
    }
    else
    {
        pbApply->setEnabled(FALSE);
    }
}

/*****************************************************************************/

void AnomalyDetail::accept()
{
    // This function gets called when the "OK" button is pressed.  If
    // changes were made to any of the fields, apply them to the database.
    // Follow this up by calling the accept() method for QDialog.

    if (pbApply->isEnabled())
        applyChanges();

    QDialog::accept();
}

