/******************************************************************************
               Level 1 Product Generation System (LPGS) Software
*******************************************************************************
*
* UNIT NAME:   AnomalyInformation.cpp
*
* PURPOSE:
* This class provides the implementation for the main Anomaly Information
* window.  Here, the LPGS operators can view a list of anomalies, filter
* them, and view details about the work orders associated with them.  They
* can approve and/or fail products, and they can generate new work orders.
*
******************************************************************************/

/* System Includes */
#include <iostream>

/* Qt Includes */
#include <QtCore>
#include <QtSql>

/* IAS Qt Includes */
#include <IASQtConst.h>
#include <iasQtWoConst.h>

/* Local Includes */
#include "AnomalyDetail.h"
#include "DateInput.h"
#include "DatabaseUtils.h"
#include "IASErrorDialog.h"
#include "Utilities.h"
#include "WorkOrderParameters.h"
#include "AnomalyInformation.h"
#include "IASUtilities.h"
#include "IASSqlQuery.h"

struct columns
{
    QString columnName;
    QString heading;
    int alignment;
};

struct states
{
    QString strFilterName;
    QString strStates;
};

// The first array is a list of column names, QTreeWidget headings,
// and alignments for the information in the Anomalies tree widget.

enum anomalyInfoColumns{
    ANOMALY_ID = 0,
    PRODUCT_REQUEST_ID,
    PR_STATE,
    ASSIGNEE,
    PR_DATE_ENTERED,
    DATE_CLOSED
};
// The list above and below need to match order.
columns anomolyInfoList[] =
{
    {"ANOMALY_ID",         "Anomaly ID",         Qt::AlignLeft},
    {"PRODUCT_REQUEST_ID", "Product Request ID", Qt::AlignLeft},
    {"STATE",              "State",              Qt::AlignLeft},
    {"ASSIGNEE",           "Assignee",           Qt::AlignLeft},
    {"DATE_ENTERED",       "Date Entered",       Qt::AlignLeft},
    {"DATE_CLOSED",        "Date Closed",        Qt::AlignLeft}
};

// Below is a list of column names, QTreeWidget headings, and
// alignments for the information in the Work Orders tree widget.
// The contents of this tree widget change depending on which anomaly
// is currently selected; if the operator chooses a new anomaly
// from the top tree widget, the bottom is re-populated with information
// about all of the work orders associated with that anomaly.

enum woColumns{
    WORK_ORDER_ID = 0,
    REQUEST_TYPE,
    WO_STATE,
    WO_DATE_ENTERED,
    WO_PATH
};
// The list above and below need to match order.
columns woInfoList[] =
{
    {"WORK_ORDER_ID", "Work Order ID",   Qt::AlignLeft},
    {"REQUEST_TYPE",  "Request Type",    Qt::AlignLeft},
    {"STATE",         "State",           Qt::AlignLeft},
    {"DATE_ENTERED",  "Date Entered",    Qt::AlignLeft},
    {"WO_PATH",       "Work Order Path", Qt::AlignLeft}
};
/* This list also needs to be maintained with the above two.  It is used
   for printing the anomaly information. */
twInfo aiInfo[] =
{
    {"Work Order Id", Qt::AlignLeft, -1, false, true, true, 75,  ""},
    {"Request Type",  Qt::AlignLeft, -1, false, true, true, 90,  ""},
    {"State",         Qt::AlignLeft, -1, false, true, true, 130, ""},
    {"Date Entered",  Qt::AlignLeft, -1, false, true, true, 115, ""},
    {"WO Path",       Qt::AlignLeft, -1, false, true, true, 250, ""}
};

// The following list associates the filter names in the dialog's
// combo box with the states that they are associated with.
// When the operator selects a new filter from the combo box, the
// code will grab the associated state list and use it in the WHERE
// clause for the anomaly query.  In other words, if the operator
// has selected the "Closed" filter, the query will take the form
// "SELECT .. FROM ANOMALIES WHERE STATE IN ( 3, 4 ) AND .."

states filterList[5];

QHash<int, QString> AnomalyInformation::stateHash;

/****************************************************************************/

AnomalyInformation::AnomalyInformation(
    QString prId,    // I: Product Request ID
    QWidget *parent) // I: Parent widget
: QDialog(parent)
{
    setupUi(this);

    setupFilterList();

    // Initialize the state value / description mapping dictionary
    if (stateHash.isEmpty())
    {
        // Set up a hash that associates the anomaly's state number in the
        // database to its human-readable name.  This will be used to
        // translate from the database value to its readable value.

        // This should match what's used above (in filterList) and should
        // match what's declared in ias_lib/perllib/ias_const.pm and
        // AnomalyConsts.h.
        stateHash[DB_ANOMALY_NEW] = UI_ANOMALY_NEW_STR;
        stateHash[DB_ANOMALY_ANALYSIS] = UI_ANOMALY_ANALYSIS_STR;
        stateHash[DB_ANOMALY_APPROVED] = UI_ANOMALY_APPROVED_STR;
        stateHash[DB_ANOMALY_FAILED] = UI_ANOMALY_FAILED_STR;
        stateHash[DB_ANOMALY_RUN_HELD] = UI_ANOMALY_RUN_HELD_STR;
        stateHash[DB_ANOMALY_RUN_PAUSED] = UI_ANOMALY_RUN_PAUSED_STR;
        stateHash[DB_ANOMALY_RUNNING] = UI_ANOMALY_RUNNING_STR;
    }
    printer = new QPrinter;
    specifiedPRId = prId;

    signalMapper = new QSignalMapper(this);

    // Create the popup menu(s) and the screen's menubar.
    popmWO = new QMenu(twWOs);

    popmAnomaly = new QMenu(twWOs);
    popmAnomaly->addAction("Anomaly Detail",
        this, SLOT(showAnomalyDetail()));
    popmAnomaly->addAction("Print Anomaly Information",
        this, SLOT(printAnomaly()));

    actionViewModifyParameters = popmWO->addAction("View / Modify Parameters",
        this, SLOT(showWOParms()));
    actionCreateNewWO = popmWO->addAction("Create New Work Order",
        this, SLOT(createNewWO()));
    actionBackupWODir = popmWO->addAction("Backup Work Order Directory",
        this, SLOT(backupWODir()));
    actionMenuSep1 = popmWO->addSeparator();
    actionActivateWO =
        popmWO->addAction("Activate Work Order", this, SLOT(activateWO()));
    actionRetryIngest =
        popmWO->addAction("Retry Ingest Data", this, SLOT(retryIngestData()));
    actionResumeWO =
        popmWO->addAction("Resume Work Order", this, SLOT(resumeWO()));
    actionApproveProduct = popmWO->addAction("Approve Product Request",
        this, SLOT(approveProduct()));
    actionMenuSep2 = popmWO->addSeparator();
    actionFailProduct = popmWO->addAction("Fail Product Request",
        this, SLOT(failCancelRequest()));

    connect(leAnomalyTitle, SIGNAL(textChanged(const QString &)),
            this, SLOT(setChangeEnabled(const QString &)));

    twiAINumColumns = ARRAY_SIZE(aiInfo);
    for (int column = 0; column < twiAINumColumns; column++)
    {
        twiAnomolyInfo.push_back(aiInfo[column]);
    }

    // Set up the Anomaly and Work Order tree widgets.
    setupListViews();

    // Set the value of the "To" date to the current date.  The value
    // of the "From" date is initialized to 1/1/1998 in the DateInput
    // constructor.  By default, then, this should give us the entire
    // date range that LPGS has been in production.  Initially set the
    // first item in the combo box as selected and call "filterChanged"
    // to set up the filter states and populate the tree widgets.

    diToDate->setDate(QDate::currentDate());

    cboFilters->setCurrentIndex(0);

    twWOs->setContextMenuPolicy(Qt::CustomContextMenu);
    twAnomalies->setContextMenuPolicy(Qt::CustomContextMenu);

    connect(pbDismiss, SIGNAL(clicked()), this, SLOT(accept()));
    connect(twWOs, SIGNAL(customContextMenuRequested(const QPoint &)),
            this, SLOT(showWOPopupMenu(const QPoint &)));
    connect(pbChangeTitle, SIGNAL(clicked()),
            this, SLOT(changeAnomalyTitle()));
    connect(pbRefresh, SIGNAL(clicked()), this, SLOT(refresh()));
    connect(cboFilters, SIGNAL(activated(int)),
            this, SLOT(filterChanged()));
    connect(twAnomalies, SIGNAL(itemSelectionChanged()),
            this, SLOT(updateAnomalyDetail()));
    connect(twAnomalies, SIGNAL(customContextMenuRequested(const QPoint &)),
            this, SLOT(showAnomalyPopupMenu(const QPoint &)));

    filterChanged();
}

void AnomalyInformation::setupFilterList()
{
    QStringList tmp;

    tmp << QString::number(DB_ANOMALY_NEW)
        << QString::number(DB_ANOMALY_ANALYSIS)
        << QString::number(DB_ANOMALY_RUN_HELD)
        << QString::number(DB_ANOMALY_RUN_PAUSED)
        << QString::number(DB_ANOMALY_RUNNING);
    filterList[0].strFilterName = "Open";
    filterList[0].strStates = tmp.join(", ");

    tmp.clear();
    tmp << QString::number(DB_ANOMALY_APPROVED)
        << QString::number(DB_ANOMALY_FAILED);
    filterList[1].strFilterName = "Closed";
    filterList[1].strStates = tmp.join(", ");

    filterList[2].strFilterName = "Approved";
    filterList[2].strStates = QString::number(DB_ANOMALY_APPROVED);

    filterList[3].strFilterName = "Failed";
    filterList[3].strStates = QString::number(DB_ANOMALY_FAILED);

    tmp.clear();
    tmp << QString::number(DB_ANOMALY_NEW)
        << QString::number(DB_ANOMALY_ANALYSIS)
        << QString::number(DB_ANOMALY_APPROVED)
        << QString::number(DB_ANOMALY_FAILED)
        << QString::number(DB_ANOMALY_RUN_HELD)
        << QString::number(DB_ANOMALY_RUN_PAUSED)
        << QString::number(DB_ANOMALY_RUNNING);
    filterList[4].strFilterName = "All";
    filterList[4].strStates = tmp.join(", ");
}

/****************************************************************************/

AnomalyInformation::~AnomalyInformation()
{
    delete printer;
}

/****************************************************************************/

void AnomalyInformation::setupListViews()
{
    // This function sets up the QTreeWidget objects associated with this
    // window.  There are two of them; the top one ("twAnomalies") is
    // a listing of all of the anomalies that match the search criteria.
    // The bottom on ("twWOs") is a list of all of the work orders that
    // have been generated for the selected anomaly.  The bottom tree
    // widget is not populated until an anomaly is actually selected from
    // the top list (depending on how the operator invokes this screen,
    // an anomaly may be auto-selected for them.)

    QStringList columnNames;

    // LIST 1: The Anomalies tree widget

    ailNumColumns = ARRAY_SIZE(anomolyInfoList);
    twAnomalies->setColumnCount(ailNumColumns);

    for (int curColumn = 0; curColumn < ailNumColumns; ++curColumn)
    {
        columnNames << anomolyInfoList[curColumn].heading;
    }
    twAnomalies->setHeaderLabels(columnNames);

    // LIST 2: The Work Orders tree widget

    columnNames.clear();
    int numColumns = ARRAY_SIZE(woInfoList);
    twWOs->setColumnCount(numColumns);

    for (int curColumn = 0; curColumn < numColumns; ++curColumn)
    {
        columnNames << woInfoList[curColumn].heading;
    }
    twWOs->setHeaderLabels(columnNames);

}

/****************************************************************************/

void AnomalyInformation::updateAnomalies()
{
    // This function updates the information in the Anomalies tree widget
    // by performing a query on the ANOMALIES table in the database.
    // To do this, it uses the date range specified by the operator in
    // the DateInput widgets, as well as the anomaly states that are
    // currently in effect for the selected filter.

    QString strQuery; // Query to retrieve information from ANOMALIES table

    IASSqlQuery query;

    // If the operator has an item in the tree widget currently selected,
    // keep track of where the selection was.  This way, when we
    // clear and re-populate the tree widget, we can put the cursor back
    // where it was before as a convenience.
    QString strAnomalyID;
    if (twAnomalies->currentItem())
    {
        strAnomalyID = twAnomalies->currentItem()->text(ANOMALY_ID);
    }

    // Dynamically build a SELECT statement from the given column names.
    strQuery = "SELECT ";
    for (int curColumn = 0; curColumn < ailNumColumns; ++curColumn) 
    {
        if (curColumn > 0)
            strQuery += ", ";
        strQuery += anomolyInfoList[curColumn].columnName;
    }

    // Okay, we've got all of the proper column names in the SELECT
    // statement .. now it's time to add the qualifiers so that we only
    // retrieve results that are in the desired states and in the
    // specified date range.

    strQuery += " FROM anomalies"
                " WHERE state IN (" + strFilterStates + ")"
                " AND (date_entered >= " + diFromDate->oracleFormat() +
                  " AND date_entered <= " + diToDate->oracleFormat(true) + ")";

    // The query is built; execute it.
    if (!query.exec(strQuery, __FILE__, __LINE__))
    {
        return;
    }

    // Clear will fire itemSelectionChanged(), so we disable it for now.
    disconnect(twAnomalies, SIGNAL(itemSelectionChanged()),
            this, SLOT(updateAnomalyDetail()));
    // Clear the tree widget out; we're repopulating it from scratch.
    twAnomalies->clear();
    leAnomalyTitle->setText("");
    pbChangeTitle->setEnabled(false);
    // Reenable signal/slot connection
    connect(twAnomalies, SIGNAL(itemSelectionChanged()),
        this, SLOT(updateAnomalyDetail()));
    if (!query.isActive())
    {
        return;
    }

    QTreeWidgetItem *curr_item = NULL;
    while (query.next())
    {
        // For each returned row, populate the QTreeWidget with the
        // results.  Note that column #2, which is the anomaly state,
        // must be translated from the state number to the state name.

        QTreeWidgetItem *lviAnomaly = new QTreeWidgetItem(twAnomalies);
        lviAnomaly->setText(ANOMALY_ID, query.value("ANOMALY_ID").toString());
        if (strAnomalyID.compare(query.value("ANOMALY_ID").toString()) == 0)
        {
            curr_item = lviAnomaly;
        }
        lviAnomaly->setText(PRODUCT_REQUEST_ID,
            query.value("PRODUCT_REQUEST_ID").toString());
        if (stateHash.contains(query.value("STATE").toInt()))
            lviAnomaly->setText(
                PR_STATE, stateHash.value(query.value("STATE").toInt()));
        else
            lviAnomaly->setText(PR_STATE, "Undefined (" +
                query.value("STATE").toString() + ")");
        lviAnomaly->setText(ASSIGNEE, query.value("ASSIGNEE").toString());
        lviAnomaly->setText(PR_DATE_ENTERED,
                            DatabaseUtils::convertDateforAlphaSort(
                                query.value("DATE_ENTERED").toString()));
        lviAnomaly->setText(DATE_CLOSED,
                            DatabaseUtils::convertDateforAlphaSort(
                                query.value("DATE_CLOSED").toString()));
    }

    if (curr_item != NULL)
    {
        twAnomalies->setCurrentItem(curr_item); 
    }
    else
    {

        // If we didn't find the item that was previously selected, fall back on
        // the product request parameter that was furnished to this routine.
        QTreeWidgetItem *activeItem = NULL;
        if (specifiedPRId != "NULL")
        {
            activeItem = twAnomalies->topLevelItem(0);
            if (activeItem)
            {
                do
                {
                    if (activeItem->text(PRODUCT_REQUEST_ID) == specifiedPRId)
                    {
                        // Found a match to the one we want
                        // so quit out of the loop
                        twAnomalies->setCurrentItem(activeItem);
                        return;
                    }
                } while ((activeItem = twAnomalies->itemBelow(activeItem)));
            }
        }
        activeItem = twAnomalies->topLevelItem(0);
        if (activeItem)
        {
            twAnomalies->setCurrentItem(activeItem);
        }
    }
}

/****************************************************************************/

void AnomalyInformation::updateAnomalyDetail()
{
    // This function is similar to the one prior except that it updates
    // the Work Orders tree widget instead of the Anomalies tree widget.

    IASSqlQuery query;

    // Extract the anomaly ID and Product Request ID of the current
    // selection in the Anomalies tree widget.  We know for a fact that
    // there is a current selection in that tree widget... otherwise,
    // we would not be here!

    QString strAnomalyID = twAnomalies->currentItem()->text(ANOMALY_ID);
    QString strPRId = twAnomalies->currentItem()->text(PRODUCT_REQUEST_ID);

    // If possible, try to reselect the Work Order ID that was previously
    // selected.
    QString strWOID;
    if (twWOs->currentItem())
    {
        strWOID = twWOs->currentItem()->text(WORK_ORDER_ID);
    }

    // Pull the anomaly title out of the database and use it to
    // set the value of the QLineEdit widget.

    QString strQuery = "SELECT title"
                       " FROM anomalies"
                       " WHERE anomaly_id = '" + strAnomalyID + "'";
    if (!query.exec(strQuery, __FILE__, __LINE__))
    {
        return;
    }

    query.next(); /* Advance to the returned record */

    leAnomalyTitle->setText(query.value("title").toString());

    // At this point, the operator has not made any modifications to
    // the anomaly title, so set the state of the "Change Title"
    // button to Disabled.  There is a signal/slot connection set up
    // to enable this button if the operator changes the title.

    pbChangeTitle->setEnabled(false);

    // Now that all the ground work is done, it's time to actually get
    // the work order information.  Construct a query to retrieve the
    // information from the WORK_ORDERS table that will return matches
    // for all work orders associated with this Product Request.

    strQuery = "SELECT work_order_id, request_type, date_entered,"
                     " state, wo_path"
               " FROM work_orders"
               " WHERE product_request_id = '" + strPRId + "'";

    if (!query.exec(strQuery, __FILE__, __LINE__))
    {
        return;
    }

    // Here again we clear the tree widget and re-populate from scratch.
    twWOs->clear();

    QString requestType;
    QTreeWidgetItem *curr_item = NULL;
    while (query.next())
    {
        // For each row returned, insert a line into the Work Orders
        // tree widget.  As with the Anomalies tree widget, we have to do
        // some translation from the database values in order to
        // provide more human-friendly results.

        QTreeWidgetItem *lviWOs = new QTreeWidgetItem(twWOs);

        requestType = query.value("REQUEST_TYPE").toString();

        if (requestType == "D")
            requestType = "Diagnostic";
        else
            requestType = "Standard";

        lviWOs->setText(WORK_ORDER_ID, query.value("WORK_ORDER_ID").toString());
        if (strWOID.compare(query.value("WORK_ORDER_ID").toString()) == 0)
        {
            curr_item = lviWOs;
        }
        lviWOs->setText(REQUEST_TYPE, requestType);

        lviWOs->setText(WO_DATE_ENTERED, DatabaseUtils::convertDateforAlphaSort(
            query.value("DATE_ENTERED").toString()));

        lviWOs->setText(WO_STATE,
                        stateToString(query.value("STATE").toString()));

        lviWOs->setText(WO_PATH, query.value("WO_PATH").toString());
    }
    if (curr_item != NULL)
    {
        twWOs->setCurrentItem(curr_item);
    }
    else
    {
        curr_item = twWOs->topLevelItem(0);
        if (curr_item)
        {
            twWOs->setCurrentItem(curr_item);
        }
    }
}

/****************************************************************************/

void AnomalyInformation::showWOPopupMenu(const QPoint &point)
{
    QString state; // State selected request is in
    QString sqlCmd;
    QString strPRId;
    QString strWOId;
    QString tmpWOId;
    QString tmpState;
    bool haveRunningWO;
    bool haveCompletedWO;
    QMap<QString, bool> replacedWO;
    QMap<QString, bool> runningWO;

    /* If we are not positioned on data, just return */
    if (twWOs->itemAt(point) == NULL)
        return;

    // Pull the current Product Request ID and anomaly type out of
    // the currently-selected item in the tree widget.
    strPRId = twAnomalies->currentItem()->text(PRODUCT_REQUEST_ID);
    /* Now get the selected WO ID */
    strWOId = twWOs->currentItem()->text(WORK_ORDER_ID);

    /* Need to gather some information about the work order, such as, if it
       has been replaced or not to help figure out the correct menu options. */
    sqlCmd = "SELECT work_order_id,"
                   " state,"
                   " wo_replaced"
             " FROM work_orders"
             " WHERE product_request_id = '" + strPRId + "'";

    IASSqlQuery query;
    if (!query.exec(sqlCmd, __FILE__, __LINE__))
    {
        return;
    }

    haveRunningWO = false;
    haveCompletedWO = false;
    while (query.next())
    {
        tmpWOId = query.value("WORK_ORDER_ID").toString();

        if (query.value("WO_REPLACED").toInt() == 1)
        {
            replacedWO[tmpWOId] = true;
        }
        else
        {
            replacedWO[tmpWOId] = false;
        }

        tmpState = query.value("STATE").toString();
        /* For states that are not "processing" states. (Not doing anything) */
        if (tmpState == DB_WO_RUNNING || tmpState == DB_WO_PENDING ||
            tmpState == DB_WO_L0R_NEW || tmpState == DB_WO_L0R_REQUESTED ||
            tmpState == DB_WO_L0R_READY)
        {
            runningWO[tmpWOId] = true;
            haveRunningWO = true;
        }
        else
        {
            runningWO[tmpWOId] = false;
        }

        if (tmpState == DB_WO_COMPLETE_WO  || tmpState == DB_WO_ABORT_WO
            || tmpState == DB_WO_RELEASE_COMPLETE
            || tmpState == DB_WO_RELEASE_ABORT || tmpState == DB_WO_COMPLETED
            || tmpState == DB_WO_ABORTED || tmpState == DB_WO_READY_TO_SHIP)
        {
            haveCompletedWO = true;
        }

        // The current state is needed when building the menus below
        if (tmpWOId == strWOId)
        {
            state = tmpState;
        }
    }

    // Hide any existing items in the context-sensitive popup menus.
    // These will change, depending on what state a work order is actually in.
    actionCreateNewWO->setVisible(false);
    actionBackupWODir->setVisible(false);
    actionMenuSep1->setVisible(false);
    actionActivateWO->setVisible(false);
    actionRetryIngest->setVisible(false);
    actionResumeWO->setVisible(false);
    actionApproveProduct->setVisible(false);
    actionMenuSep2->setVisible(false);
    actionFailProduct->setVisible(false);

    if (!replacedWO[strWOId] && !haveRunningWO && !haveCompletedWO)
    {
        if (state == DB_WO_HALTED)
        {
            /* If a work order has been halted because of a pause
               flag, give the operator an option to resume processing. */
            actionMenuSep1->setVisible(true);
            actionResumeWO->setVisible(true);
        }
        else if (state == DB_WO_DIAGNOSTIC_COMPLETE)
        {
            /* If the work order has completed processing, the operator
               should be able to approve the product so that it can
               proceed to the remainder of the LPGS processing flow. */
            actionMenuSep1->setVisible(true);
            actionApproveProduct->setVisible(true);
        }
        else if (state == DB_WO_INGEST_FAILED)
        {
            /* If the work order has failed during ingest, the operator
               should be able to retry the ingest. */
            actionMenuSep1->setVisible(true);
            actionRetryIngest->setVisible(true);
        }
        else if (state == DB_WO_REQUEST_FAILED
                 || state == DB_WO_SUBSETTER_FAILED)
        {
            /* Nothing to add for these here */
        }
        else 
        {
            /* Can only activare WOs that have not been replaced and
               not currently the active WO. */
            actionMenuSep1->setVisible(true);
            actionActivateWO->setVisible(true);
        }
    }

    /* For states that are not "processing" states. (Not doing anything) */
    if (state == DB_WO_INGEST_FAILED  || state == DB_WO_HALTED ||
        state == DB_WO_DIAGNOSTIC_COMPLETE || state == DB_WO_TRANSFER_FAILED)
    {
        /* If the selected WO is not running, then we can back it up. */
        actionBackupWODir->setVisible(true);
    }

    if (!haveRunningWO && !haveCompletedWO)
    {
        /* If we don't have a running or completed WO then we can create a 
           new one */
        actionCreateNewWO->setVisible(true);

        /* Give the operator the option to outright fail the product.
           (so long as it isn't currently running or completed) */
        actionMenuSep2->setVisible(true);
        actionFailProduct->setVisible(true);
    }

    popmWO->popup(twWOs->viewport()->mapToGlobal(point));
}

/****************************************************************************/

void AnomalyInformation::showAnomalyPopupMenu(const QPoint &point)
{
    /* If we are not positioned on data, just return */
    if (twAnomalies->itemAt(point) == NULL)
        return;

    popmAnomaly->popup(twAnomalies->viewport()->mapToGlobal(point));
}

/****************************************************************************/

void AnomalyInformation::printAnomaly()
{
    QString strAnomalyID = twAnomalies->currentItem()->text(ANOMALY_ID);
    QString strPRId = twAnomalies->currentItem()->text(PRODUCT_REQUEST_ID);
    QString strState = twAnomalies->currentItem()->text(PR_STATE);
    QString strAssignee = twAnomalies->currentItem()->text(ASSIGNEE);
    QString strDateEntered = twAnomalies->currentItem()->text(PR_DATE_ENTERED);
    QString strDateClosed = twAnomalies->currentItem()->text(DATE_CLOSED);

    IASSqlQuery query;

    const int margin = 10;
    int pageNo = 1;
    int yPos = 0;
    int pageWidth = 0;

    QString strLine;
    QPainter painter;

    printer->setPageSize(QPrinter::Letter);
    printer->setColorMode(QPrinter::GrayScale);

    QPrintDialog dialog(printer, this);
    if (dialog.exec() == QDialog::Accepted)
    {
        if (!painter.begin(printer))
            return;

        painter.setFont(QFont("Times", 10, QFont::Bold));
        QFontMetrics fm = painter.fontMetrics();
        pageWidth = painter.device()->width();

        strLine = "\t\t\t\tLPGS Anomaly Information";
        painter.drawText(margin, margin + yPos, pageWidth,
            fm.lineSpacing(), Qt::TextExpandTabs, strLine);
        yPos += fm.height();

        strLine = "Anomaly ID: " + strAnomalyID;
        painter.drawText(margin, margin + yPos, pageWidth,
        fm.lineSpacing(), Qt::TextExpandTabs, strLine);
        yPos += fm.height();

        strLine = "Product Request ID: " + strPRId;
        painter.drawText(margin, margin + yPos, pageWidth,
                fm.lineSpacing(), Qt::TextExpandTabs, strLine);
        yPos += fm.height();

        strLine = "State: " + strState;
        painter.drawText(margin, margin + yPos, pageWidth,
                fm.lineSpacing(), Qt::TextExpandTabs, strLine);
        yPos += fm.height();

        strLine = "Assignee: " + strAssignee;
        painter.drawText(margin, margin + yPos, pageWidth,
                fm.lineSpacing(), Qt::TextExpandTabs, strLine);
        yPos += fm.height();

        strLine = "Date Entered: " + strDateEntered;
        painter.drawText(margin, margin + yPos, pageWidth,
                fm.lineSpacing(), Qt::TextExpandTabs, strLine);
        yPos += fm.height();

        strLine = "Date Closed: " + strDateClosed;
        painter.drawText(margin, margin + yPos, pageWidth,
                fm.lineSpacing(), Qt::TextExpandTabs, strLine);
        yPos += 2 * fm.height();

        if (!query.exec("select problem_description, problem_resolution"
                        " from anomalies"
                        " where anomaly_id = '" + strAnomalyID + "'",
                        __FILE__, __LINE__ ))
        {
            return;
        }
        query.next(); /* Advance to the returned record */

        strLine = "Problem Description:";
        painter.drawText(margin, margin + yPos, pageWidth,
                fm.lineSpacing(), Qt::TextExpandTabs, strLine);
        yPos += fm.height();

        strLine = query.value("problem_description").toString();
        QRect descriptionBox =
            painter.boundingRect(margin, margin + yPos, pageWidth,
                fm.lineSpacing(), Qt::TextWordWrap | Qt::TextExpandTabs |
                    Qt::AlignLeft, strLine);
        painter.drawText(descriptionBox, Qt::TextWordWrap | Qt::TextExpandTabs |
                    Qt::AlignLeft, strLine);
        yPos += descriptionBox.height() + fm.height();

        strLine = "Problem Resolution:";
        painter.drawText(margin, margin + yPos, pageWidth,
                fm.lineSpacing(), Qt::TextExpandTabs, strLine);
        yPos += fm.height();

        strLine = query.value("problem_resolution").toString();
        QRect resolutionBox =
            painter.boundingRect(margin, margin + yPos, pageWidth,
                fm.lineSpacing(), Qt::TextWordWrap | Qt::TextExpandTabs |
                    Qt::AlignLeft, strLine);
        painter.drawText(resolutionBox, Qt::TextWordWrap | Qt::TextExpandTabs |
                    Qt::AlignLeft, strLine);
        yPos += resolutionBox.height() + fm.height();

        TreeWidgetUtilities::printTreeWidget(twWOs, painter, printer,
            margin, yPos, pageNo, twiAnomolyInfo, twiAINumColumns);
    }
}

/****************************************************************************/

void AnomalyInformation::backupWODir()
{
    QString strWOId = twWOs->currentItem()->text(WORK_ORDER_ID);
    QString strOWD = getenv("OWD");
    QString strSourceDir = getenv("IAS_WO");
    QString strFilename = strWOId + ".tar";

    if (strOWD.isEmpty())
        return;

    if (strSourceDir.isEmpty())
        return;

    QString strDestDir = QFileDialog::getExistingDirectory(this, 
        "Select Destination Directory For Backup", strOWD,
        QFileDialog::DontUseNativeDialog | QFileDialog::ShowDirsOnly);

    if (strDestDir.isEmpty())
        return;

    QString strCommand = "cd " + strSourceDir + "; /bin/tar -cf " +
        strDestDir + "/" + strFilename + " " + strWOId + " &";

    system(strCommand.toAscii().data());

    QMessageBox::information(this, "Anomaly Being Saved",
        "The data directories associated with this Work Order\n"
        "are being saved to " + strDestDir + "/" + strFilename + ".",
        "&OK");
}

/****************************************************************************/

void AnomalyInformation::showAnomalyDetail()
{
    // This function displays the Anomaly Detail dialog for the
    // currently-selected anomaly.

    if (twAnomalies->currentItem() == NULL)
        return;

    QString strAnomalyId = twAnomalies->currentItem()->text(ANOMALY_ID);

    AnomalyDetail anomalyDetail(strAnomalyId, this);
    anomalyDetail.exec();

    refresh();
}

/****************************************************************************/

void AnomalyInformation::showWOParms()
{
    // This function brings up the Work Order Parameters dialog for
    // the currently-selected work order.  This dialog allows the
    // operators to view and/or modify the parameters (and pause flag)
    // for all of the scripts associated with the work order.

    IASSqlQuery query;

    QString strWOId = twWOs->currentItem()->text(WORK_ORDER_ID);
    QString strPRId = twAnomalies->currentItem()->text(PRODUCT_REQUEST_ID);

    /* We need to figure out if the "OK" and "Apply" buttons on the
       WorkOrderParameters popup should be in a state to allow the update
       of parameters or  should not allow change. */
    QString strQuery = "SELECT current_wo_id"
                       " FROM anomalies"
                       " WHERE product_request_id = '" + strPRId + "'";
    if (!query.exec(strQuery, __FILE__, __LINE__))
    {
        return;
    }

    query.next(); /* Advance to the returned record */

    /* Turn off the Apply and OK buttons by default */
    QString currWOId = query.value("CURRENT_WO_ID").toString();

    strQuery = "SELECT state"
               " FROM work_orders"
               " WHERE work_order_id = '" + strWOId + "'";

    if (!query.exec(strQuery, __FILE__, __LINE__))
    {
        return;
    }

    query.next(); /* Advance to the returned record */

    QString state = query.value("STATE").toString();

    // If the work order is the current WO for this product request and the 
    // WO is in a stopped state, allow modifications to the WO parameters
    bool editFlag = false;
    if (currWOId == strWOId && (state == DB_WO_HALTED || 
                                state == DB_WO_INACTIVE))
    {
        editFlag = true;
    }

    WorkOrderParameters woParms(editFlag, strPRId, strWOId, this);
    woParms.exec();

    // If any parameters were modified, change the request type to diagnostic.
    if (woParms.parametersModified())
    {
        QSqlDatabase::database().transaction();

        strQuery = "UPDATE work_orders "
                   "   SET request_type = 'D' "
                   " WHERE work_order_id = '" + strWOId + "'";

        if (!query.exec(strQuery, __FILE__, __LINE__))
        {
            QSqlDatabase::database().rollback();
            return;
        }

        QSqlDatabase::database().commit();

        refresh();
    }
}

/****************************************************************************/

void AnomalyInformation::activateWO()
{
    QSqlDatabase database = QSqlDatabase::database();

    QString strWOId = twWOs->currentItem()->text(WORK_ORDER_ID);
    QString strAnomalyID = twAnomalies->currentItem()->text(ANOMALY_ID);

    // First, make sure that the work order that the operator is
    // trying to activate is the current one associated with the
    // anomaly (otherwise, activation doesn't make sense.)

    IASSqlQuery checkQuery;
    if (!checkQuery.exec("SELECT current_wo_id"
                         " FROM anomalies"
                         " WHERE anomaly_id = '" + strAnomalyID + "'",
                         __FILE__, __LINE__))
    {
        return;
    }
    checkQuery.next(); /* Advance to the returned record */

    if (strWOId != checkQuery.value("CURRENT_WO_ID").toString())
    {
        QMessageBox::warning(this, "Not Current WO",
            strWOId + " is not the current Work Order.", "&OK");
        return;
    }

    database.transaction();

    IASSqlQuery query; // must be after the creation of the transaction

    if (!query.exec("UPDATE work_orders"
                    " SET state = '" + DB_WO_L0R_NEW + "'"
                    " WHERE work_order_id = '" + strWOId + "'",
                    __FILE__, __LINE__))
    {
        database.rollback();
        return;
    }

    database.commit();

    refresh();
}

/****************************************************************************/

void AnomalyInformation::retryIngestData()
{
    QSqlDatabase database = QSqlDatabase::database();

    QString strWOId = twWOs->currentItem()->text(WORK_ORDER_ID);
    QString strAnomalyID = twAnomalies->currentItem()->text(ANOMALY_ID);

    // First, make sure that the work order that the operator is
    // trying to retry is the current one associated with the
    // anomaly (otherwise, retry ingest data doesn't make sense.)

    IASSqlQuery checkQuery;
    if (!checkQuery.exec("SELECT current_wo_id"
                         " FROM anomalies"
                         " WHERE anomaly_id = '" + strAnomalyID + "'",
                         __FILE__, __LINE__))
    {
        return;
    }
    checkQuery.next(); /* Advance to the returned record */

    if (strWOId != checkQuery.value("CURRENT_WO_ID").toString())
    {
        QMessageBox::warning(this, "Not Current WO",
            strWOId + " is not the current Work Order.", "&OK");
        return;
    }

    database.transaction();

    IASSqlQuery query; // must be after the creation of the transaction

    if (!query.exec("UPDATE work_orders"
                    " SET state = '" + DB_WO_L0R_READY + "'"
                    " WHERE work_order_id = '" + strWOId + "'",
                    __FILE__, __LINE__))
    {
        database.rollback();
        return;
    }

    database.commit();

    refresh();
}

/****************************************************************************/

void AnomalyInformation::createNewWO()
{
    QSqlDatabase database = QSqlDatabase::database();

    /* Start a transaction for the inserting of the new work order */
    database.transaction();

    IASSqlQuery query;
    IASSqlQuery subQuery;

    QString strPRId = twAnomalies->currentItem()->text(PRODUCT_REQUEST_ID);
    QString strOldWOId = twWOs->currentItem()->text(WORK_ORDER_ID);

    QString strQuery = "SELECT *"
                       " FROM work_orders"
                       " WHERE work_order_id = '" + strOldWOId + "'";
    if (!query.exec(strQuery, __FILE__, __LINE__))
    {
        database.rollback();
        return;
    }

    query.next(); /* Advance to the returned record */

    QSqlRecord woRecord = query.record();
    int numColumns = woRecord.count();

    QStringList listWOValues;
    /* Place the column results into a list for modification later */
    for (int columnPos = 0; columnPos < numColumns; columnPos++)
    {
        if (woRecord.field(columnPos).type() == QVariant::DateTime)
        {
            listWOValues.append("TO_DATE('"
                + query.value(columnPos).toDateTime()
                              .toString("yyyy-MM-dd hh:mm:ss")
                + "', 'YYYY-MM-DD HH24:MI:SS')");
        }
        else
        {
            listWOValues.append(query.value(columnPos).toString());
        }
    }

    /* Save for later to fix for the new WO */
    QString strWoPath = query.value("WO_PATH").toString();
    QString strWoReqType = query.value("REQUEST_TYPE").toString();

    // The first thing that needs to be done is to determine the ID
    // of the new work order.  This is done by selecting the next
    // value from the wo_seq sequence, and prepending the L1_SYSTEM_ID on
    // the front.  This will yield a value of the form "L821."

    /* Get the L1_SYSTEM_ID from the environment. */
    QString strNewWOId = getenv("L1_SYSTEM_ID");
    if (strNewWOId.isEmpty())
    {
        QMessageBox::critical(0, "Missing Environment",
            "Could not get L1_SYSTEM_ID from environment.\n"
            "This variable is required for system startup or shutdown",
            QMessageBox::Ok, 0);

        database.rollback();
        return;
    }

    /* Make sure all the others are set as replaced */
    strQuery = "UPDATE work_orders"
               " SET wo_replaced = 1"
               " WHERE product_request_id = '" + strPRId + "'";

    if (!query.exec(strQuery, __FILE__, __LINE__))
    {
        database.rollback();
        return;
    }

    /* Get the new work order ID */
    strQuery = "SELECT wo_seq.nextval FROM dual";
    if (!query.exec(strQuery, __FILE__, __LINE__))
    {
        database.rollback();
        return;
    }

    query.next(); /* Advance to the returned record */

    strNewWOId += query.value(0).toString();

    /* Now update the work order ID in the list */
    listWOValues[woRecord.indexOf("WORK_ORDER_ID")] = strNewWOId;

    /* Empty the processing node to ensure that the new order gets
       picked up and processed on a node. */
    listWOValues[woRecord.indexOf("PROCESSING_NODE")] = "";

    // The value of REQUEST_TYPE is pulled from the existing work order; 
    // If the work order has not been changed it will be a Standard work order 
    // (as opposed to Diagnostic if the parameters have been changed)
    listWOValues[woRecord.indexOf("REQUEST_TYPE")] = strWoReqType;

    // The value of STATE is "I" (Inactive WO)
    listWOValues[woRecord.indexOf("STATE")] = DB_WO_INACTIVE;

    // Make sure the new one is not set to replaced.
    listWOValues[woRecord.indexOf("WO_REPLACED")] = "0";

    /* Fix the wo_path for the new WO */
    strWoPath.replace(strOldWOId, strNewWOId);
    listWOValues[woRecord.indexOf("WO_PATH")] = strWoPath;

    // At this point, we have all of the information that is required
    // to generate a new work_orders entry.  All of the remaining
    // items in the listWOValues list are empty (which is fine.)
    // Insert the new row into the work_orders table.
    QString strUpdate = "INSERT INTO work_orders VALUES (";
    for (int columnPos = 0; columnPos < numColumns; columnPos++)
    {
        if (woRecord.field(columnPos).type() == QVariant::DateTime)
        {
            strUpdate += listWOValues[columnPos];
        }
        else
        {
            strUpdate += "'" + listWOValues[columnPos] + "'";
        }

        if (columnPos != numColumns - 1)
            strUpdate += ", ";
    }
    strUpdate += ")";

    if (!query.exec(strUpdate, __FILE__, __LINE__))
    {
        database.rollback();
        return;
    }

    // Onto the second phase of work order generation.  We have the
    // work_orders entry, but there are no associated entries in
    // wo_scripts or wo_parms.  What we need to do is basically copy
    // all of the existing wo_scripts and wo_parms that have the old
    // work order ID and create copies for the new work order ID.

    // First, get all of the wo_scripts rows for the old work order.
    if (!query.exec("SELECT *"
                     " FROM wo_scripts"
                     " WHERE work_order_id = '" + strOldWOId + "'",
                     __FILE__, __LINE__))
    {
        database.rollback();
        return;
    }

    QSqlRecord scriptRecord;
    QStringList listScripts;
    while (query.next())
    {
        // Insert the new wo_scripts rows into the table.  Note that
        // we set the status of every script to "P" (Pending) and
        // clear out the completion status and failure reasons.

        scriptRecord = query.record();

        listScripts.clear();
        for (int columnPos = 0; columnPos < scriptRecord.count(); columnPos++)
        {
            listScripts.append(query.value(columnPos).toString());
        }

        /* Update specific part of the record */
        listScripts[scriptRecord.indexOf("WORK_ORDER_ID")] = strNewWOId;
        listScripts[scriptRecord.indexOf("STATE")] = "P";
        listScripts[scriptRecord.indexOf("COMPL_STATUS")] = "";
        listScripts[scriptRecord.indexOf("FAILURE_REASON")] = "";

        strUpdate = "INSERT INTO wo_scripts VALUES " +
            commaSeparatedList(listScripts);

        if (!subQuery.exec(strUpdate, __FILE__, __LINE__))
        {
            database.rollback();
            return;
        }
    }

    // Finally, we need to copy all of the work order parameters for
    // the old work order into the WO_PARMS table for the new work
    // order.  Here, we just directly copy everything; no values need
    // to be tweaked or reset.  Typically, the operator will use the
    // parameter editor to change certain parameters before activation.

    if (!query.exec("SELECT *"
                    " FROM wo_parms"
                    " WHERE work_order_id = '" + strOldWOId + "'",
                    __FILE__, __LINE__))
    {
        database.rollback();
        return;
    }

    QSqlRecord parmsRecord;
    QStringList listParms;
    while (query.next())
    {
        parmsRecord = query.record();

        listParms.clear();
        for (int columnPos = 0; columnPos < parmsRecord.count(); columnPos++)
        {
            listParms.append(query.value(columnPos).toString());
        }

        listParms[parmsRecord.indexOf("WORK_ORDER_ID")] = strNewWOId;

        strUpdate = "INSERT INTO wo_parms VALUES " +
            commaSeparatedList(listParms);

        if (!subQuery.exec(strUpdate, __FILE__, __LINE__))
        {
            database.rollback();
            return;
        }
    }

    // Set the current work order ID for this anomaly to be the
    // one that we just created.

    QString strCurrAnomaly = twAnomalies->currentItem()->text(ANOMALY_ID);

    if (!query.exec("UPDATE anomalies"
                    " SET current_wo_id = '" + strNewWOId + "',"
                        " state = " + QString::number(DB_ANOMALY_RUN_HELD) +
                    " WHERE anomaly_id = '" + strCurrAnomaly + "'",
                    __FILE__, __LINE__))
    {
        database.rollback();
        return;
    }

    database.commit();

    refresh();
}

/****************************************************************************/

void AnomalyInformation::approveProduct()
{
    // This function approves the product that was generated by the
    // current work order and allows it to finish LPGS processing.

    QString strWOId = twWOs->currentItem()->text(WORK_ORDER_ID);
    QString strAnomalyID = twAnomalies->currentItem()->text(ANOMALY_ID);

    int option = QMessageBox::question(this, "Approve Work Order?",
        "Approve work order " + strWOId + "?",
        QMessageBox::Yes | QMessageBox::No, QMessageBox::No);

    if (option == QMessageBox::No)
    {
        return;
    }

    QSqlDatabase database = QSqlDatabase::database();

    /* Start a transaction for the updating the database */
    database.transaction();

    IASSqlQuery query;

    /* Set the state of the work order to DB_WO_READY_TO_SHIP */
    if (!query.exec("UPDATE work_orders"
                    " SET state = '" + DB_WO_READY_TO_SHIP + "'"
                    " WHERE work_order_id = '" + strWOId + "'",
                    __FILE__, __LINE__ ))
    {
        database.rollback();
        return;
    }

    /* Set the state of the anomaly to "Closed-Approved." */
    if (!query.exec("UPDATE anomalies"
                    " SET state = " + QString::number(DB_ANOMALY_APPROVED) +
                        ", date_closed = SYSDATE"
                    " WHERE anomaly_id = '" + strAnomalyID + "'",
                    __FILE__, __LINE__ ))
    {
        database.rollback();
        return;
    }

    database.commit();

    refresh();
}

/****************************************************************************/

void AnomalyInformation::failCancelRequest()
{
    QStringList strWOIds; /* Work order Ids list */ 
    QStringList strPRIds; /* Product Request Ids list */

    /* Append the current item to the string lists */
    strWOIds.append(twWOs->currentItem()->text(WORK_ORDER_ID));
    strPRIds.append(twAnomalies->currentItem()->text(PRODUCT_REQUEST_ID));

    failCancelProductRequest(this, strPRIds, strWOIds);

    refresh();
}

/****************************************************************************/

void AnomalyInformation::filterChanged()
{
    // This function is called when the operator changes the anomaly
    // filter at the top of the screen (i.e., they switch from "Open"
    // to "Failed".)

    QString strFilter = cboFilters->currentText();

    int numFilters = ARRAY_SIZE(filterList);

    for ( int currFilter = 0; currFilter < numFilters; ++currFilter )
    {
        if ( strFilter == filterList[currFilter].strFilterName )
        {
            strFilterStates = filterList[currFilter].strStates;
            break;
        }
    }

    refresh();
}

/****************************************************************************/

void AnomalyInformation::setChangeEnabled(
    const QString &) // I: Information sent by signal (discarded)
{
    // This slot is called when the contents of the QLineEdit widget
    // for the anomaly title are modified; it enables the Apply button
    // so that the changes can be made.

    if (twAnomalies->currentItem())
        pbChangeTitle->setEnabled(true);
}

/****************************************************************************/

void AnomalyInformation::changeAnomalyTitle()
{
    // This function updates the title of the anomaly to the value
    // that has been entered in the provided QLineEdit widget.

    IASSqlQuery query;

    QString strUpdate = "UPDATE anomalies"
                        " SET title = '" + leAnomalyTitle->text() + "'"
                        " WHERE anomaly_id = '" +
                            twAnomalies->currentItem()->text(ANOMALY_ID) + "'";

    if (!query.exec(strUpdate, __FILE__, __LINE__))
    {
        return;
    }

    pbChangeTitle->setEnabled(false);
}

/****************************************************************************/

void AnomalyInformation::refresh()
{
    updateAnomalies();

    if (!twAnomalies->currentItem())
        twWOs->clear();
}

/****************************************************************************/

void AnomalyInformation::resumeWO()
{
    // This function resumes a work order that is in a state of
    // "WO Halted."  Typically, a work order enters this state
    // because the operator has requested a pause upon its
    // completion.  Work orders are resumed by setting the state
    // of the Product Request to "WO Resumable" (for informational
    // purposes for the operator) and setting the state of the
    // work order to "RM" so that PWS picks it up on its next
    // trip through its processing loop.

    QString errorMessage;
    QString woID = twWOs->currentItem()->text(WORK_ORDER_ID);
    if (IASUtilities::resetWOPending(woID, errorMessage, true))
    {
        /* Since things have changed, the display should be updated */
        refresh();
    }
}
