/******************************************************************************
               Level 1 Product Generation System (LPGS) Software
 ******************************************************************************
 *
 * UNIT NAME:   DatabaseUtils.cpp
 *
 * PURPOSE:
 * The DatabaseUtils class contains a collection of database related utility
 * functions.
 *
 *****************************************************************************/

// System includes

// Qt includes
#include <QtCore>
#include <QtGui>
#include <QtSql>

// Local includes
#include "DatabaseUtils.h"

QStringList DatabaseUtils::getConstraint(
    const QString strTableName,  
    const QString strColumnName, 
    datatype type)
{
    // This function gets the limits on a column by querying the data
    // dictionary for its SEARCH_CONSTRAINT, and returning the values
    // in a QStringList.  If the datatype is number, the first value
    // will be the minimun value and the second a maximum value.  If
    // the constraint is not found, an empty QStringList will be returned.

    char removeChars[]="()' "; // characters to remove from constraint list"
    int index;                 // index used to locate a word
    int rchar;
    QString strConstraint;     // used to hold constraint
    QString strSQLCommand;     // SQL command used to retrieve constraints
    QString strTemp;
    QString strKeyword;        // keyword in constraint,
                               // BETWEEN (number) IN (varchar)
    QString strSep;            // seporator in constraint list
    QStringList strlLimits;    // constraints list to return

    // create SQL command
    strSQLCommand = "SELECT uc.search_condition FROM all_constraints uc,"
        "(SELECT constraint_name FROM all_cons_columns"
        " WHERE column_name = '" + strColumnName + "') ucc"
        " WHERE uc.constraint_name = ucc.constraint_name"
        " AND uc.table_name = '" + strTableName + "'";

    // set up keywords and separator values
    if (type == varchar)
    {
        strKeyword = QString("IN");
        strSep = QString(",");
    }
    else
    {
        strKeyword = QString("BETWEEN");
        strSep = QString("AND");
    }

    // query database for constraints
    QSqlQuery query;

    // if query for constraint not successful, return an empty list
    if (!query.exec(strSQLCommand))
    {
        return strlLimits;
    }

    // search through list of constraints, and find the constraint containing
    // the limits value between x and y
    int row = 0;
    while (query.next())
    {
        QString searchCon = query.value(0).toString();

        // if a match is found
        if (searchCon.contains(strKeyword, Qt::CaseInsensitive))
        {
            // remove string contents from beginning to end of word BETWEEN
            index = searchCon.indexOf(strKeyword, 0, Qt::CaseInsensitive);
            strConstraint = searchCon.remove(0, index + strKeyword.length());

            // remove extra characters from constraint
            for (rchar = 0; rchar < int(strlen(removeChars)); rchar++)
            {
                index = strConstraint.indexOf(removeChars[rchar], 0);

                // while characters are fount (-1 = not found)
                while (index != -1)
                {
                    strTemp = strConstraint.remove(index, 1);
                    strConstraint = strTemp;
                    index = strConstraint.indexOf(removeChars[rchar], 0);
                }
            }

            // create a QStringList by splitting the QString on the word AND
            strlLimits = strConstraint.split(strSep, QString::SkipEmptyParts);

            if(type == number)
            {
                if(strlLimits.count() < 2)
                {
                    strlLimits = strConstraint.split(strSep.toLower(), 
                        QString::SkipEmptyParts);
                }
            }

            // return QStringList containing the limits;
            return strlLimits;
        }

        row++;
    }

    // no constraint found, return an empty QStringList
    return strlLimits;
}

QString DatabaseUtils::convertDateforAlphaSort(const QString strDateTime)
{
    // This function takes a date in the following format:
    // "DD-MON-YEAR HH:MM:SS" and converts is to a date that will correctly
    // sort by date when sorted alphabetically.  Output format is :
    // "YEAR-MM-DD HH:MM:SS"

    if (strDateTime.isEmpty() || strDateTime.isNull())
        return QString("");

    QDateTime dateTime;
    QString defaultFormat = "dd-MMM-yyyy hh:mm:ss";

    // attempt to parse the default format
    dateTime = QDateTime::fromString(strDateTime, defaultFormat);

    // if the default format was not successful, try the ISO DateTime
    if (!dateTime.isValid())
    {
        dateTime = QDateTime::fromString(strDateTime, Qt::ISODate); 
        
        // if the ISO format also failed, return empty string
        if (!dateTime.isValid())
            return QString("");
    }

    return dateTime.toString("yyyy-MM-dd hh:mm:ss");
}

QString DatabaseUtils::friendlyColumnName(const QString strColumnName)
{
    // convert a database column name to a more people friendly value.  The
    // underscores are removed, and replaces by spaces, the first character
    // of each word is capitalized, and the remainder of each word is set
    // to lower case

    QChar chCap;
    QString strTemp;
    QString strData;
    QString strOut;
    QStringList::Iterator it;

    // covert column name to lower case characters
    strTemp = strColumnName.toLower();

    // split the column name apart using the '_' character
    QStringList strlData = strTemp.split('_', QString::SkipEmptyParts);

    // loop through each word and capitalize the first character, and making
    // lower case the remainder
    for (it = strlData.begin(); it != strlData.end(); ++it)
    {
        strData = (*it);

        // upper case the first character
        chCap = strData.at(0).toUpper();

        // replace first character in word with upper case character
        strOut += strData.replace(0, 1, chCap);

        // add space between words
        strOut += " ";
    }

    return strOut.trimmed();
}

QString DatabaseUtils::databaseColumnName(const QString strColumnName)
{
    // convert a people friendly column name back to a database column name
    QString temp = strColumnName.toUpper();
    temp = temp.replace(QChar(' '), QChar('_'));
    return temp; 
}

