/******************************************************************************
               Level 1 Product Generation System (LPGS) Software
 ******************************************************************************
 *
 * UNIT NAME:   DateInput.cpp
 *
 * PURPOSE:
 * This is the definition file for the DateInput class.  It uses a combo box
 * and two spinboxes to capture a valid date from a user.
 *
 *****************************************************************************/
 
/* System Includes */
#include <iostream>

/* Qt Includes */
#include <QtCore>
#include <QtGui>

/* Local Includes */
#include "DateInput.h"

DateInput::DateInput(QWidget* parent, Qt::WindowFlags f)
  : QWidget(parent, f)
{
     // This is the constructor for the DateInput class.  It first calls
     // the QWidget constructor since the DateInput object is a widget.

    setSizePolicy(QSizePolicy((QSizePolicy::Expanding), 
        (QSizePolicy::Expanding)));

    hbox = new QHBoxLayout(this); 
    hbox->setSpacing(6);
    hbox->setMargin(0);

    cboDateMonth = new QComboBox(this);
    cboDateMonth->addItem(tr("January"));
    cboDateMonth->addItem(tr("February"));
    cboDateMonth->addItem(tr("March"));
    cboDateMonth->addItem(tr("April"));
    cboDateMonth->addItem(tr("May"));
    cboDateMonth->addItem(tr("June"));
    cboDateMonth->addItem(tr("July"));
    cboDateMonth->addItem(tr("August"));
    cboDateMonth->addItem(tr("September"));
    cboDateMonth->addItem(tr("October"));
    cboDateMonth->addItem(tr("November"));
    cboDateMonth->addItem(tr("December"));
    hbox->addWidget(cboDateMonth);

    spinDateDay = new QSpinBox(this);
    spinDateDay->setMinimum(0);
    spinDateDay->setMaximum(32);
    spinDateDay->setEnabled(true);
    spinDateDay->setValue(1);
    hbox->addWidget(spinDateDay);
    setTabOrder(cboDateMonth, spinDateDay);

    spinDateYear = new QSpinBox(this);
    spinDateYear->setEnabled(true);
    hbox->addWidget(spinDateYear);
    setTabOrder(spinDateDay, spinDateYear);

    // signals and slots connections
    connect(spinDateDay, SIGNAL(valueChanged(const QString&)), 
        this, SLOT(DayChange()));
    connect(cboDateMonth,SIGNAL(activated(int)),
        this, SLOT(MonthChange()));

    spinDateYear->setMinimum(1998);
    spinDateYear->setMaximum(9999);
    spinDateYear->setValue(1998);

    timeVisible = false;
}

//***********************************************************************

DateInput::~DateInput()
{
    // no need to delete child widgets, Qt does it all for us
}

//***********************************************************************

void DateInput::setTimeToStart()
{
    if (timeVisible) 
    {
        spinDateHour->setValue(0);
        spinDateMinute->setValue(0);
    }
}

void DateInput::setTimeToEnd()
{
    if (timeVisible) 
    {
        spinDateHour->setValue(23);
        spinDateMinute->setValue(59);
    }
}

void DateInput::showTime(bool endOfDay) // I: Set time to the end of day?
{
    spinDateHour = new SpinBox(this);
    spinDateHour->setEnabled(true);
    hbox->addWidget(spinDateHour);
    setTabOrder(spinDateYear, spinDateHour);

    /* Having the label expand make things look very bad. */
    QLabel *tempLabel = new QLabel(":", this);
    tempLabel->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);
    hbox->addWidget(tempLabel);

    spinDateMinute = new SpinBox(this);
    spinDateMinute->setEnabled(true);
    hbox->addWidget(spinDateMinute);
    setTabOrder(spinDateHour, spinDateMinute);

    spinDateHour->setMinimum(0);
    spinDateHour->setMaximum(23);
    spinDateMinute->setMinimum(0);
    spinDateMinute->setMaximum(59);

    if (endOfDay)
    {
        spinDateHour->setValue(23);
        spinDateMinute->setValue(59);
    } 
    else    
    {
        spinDateHour->setValue(0);
        spinDateMinute->setValue(0);
    }

    timeVisible = TRUE;
}

//***********************************************************************

QDate DateInput::getDate()
{
     // This function returns a QDate representation of the date that is
     // currently entered into the Date Input object.  After retrieving the
     // date, the DateInput's isValid member contains TRUE if the date is
     // valid and FALSE if it is not.
    QDate temp;
    
    isValid = temp.setYMD(spinDateYear->value(),
        cboDateMonth->currentIndex() + 1,
        spinDateDay->value());

    return temp;
}

//***********************************************************************

QDateTime DateInput::getDateTime(bool endOfDay)
{
     // This function returns a QDateTime representation of the date that is
     // currently entered into the Date Input object. If mytime is invalid,
     // the time will be set to midnight by the QDateTime constructor.
    QTime mytime;
    if (timeVisible)
        mytime.setHMS(spinDateHour->value(),
            spinDateMinute->value(), 0);
    else
        if (endOfDay)
            mytime.setHMS(23, 59, 59);
        else
            mytime.setHMS(0, 0, 0);
    QDateTime temp(getDate(), mytime);
    return temp;
}
//***********************************************************************

void DateInput::setDate(const QDate newDate) 
{
     // This function sets the display of the DateInput object to show
     // the date that is passed in.
    setDate(newDate.year(), newDate.month(), newDate.day());
}

//***********************************************************************

void DateInput::setDate(const int year, const int month, const int day)
{
     // This function sets the display of the DateInput object to show
     // the date that is passed in.
    setYear(year);
    setMonth(month);
    setDay(day);
}

//***********************************************************************

void DateInput::setDay(const int day)
{
     // This function sets the display of the Day of the DateInput object 
     // to show the day that is passed in.
    spinDateDay->setValue(day);
}

//***********************************************************************

void DateInput::setMonth(const int month)
{
     // This function sets the display of the Month of the DateInput object 
     // to show the Month that is passed in.
     cboDateMonth->setCurrentIndex(month - 1);
}

//***********************************************************************

void DateInput::setYear(const int year)
{
     // This function sets the display of the Year of the DateInput object 
     // to show the year that is passed in.
     spinDateYear->setValue(year);
}

//***********************************************************************

int DateInput::day()
{
     // This function returns an integer representation of the day set in
     // the Day portion of the DateInput's Date. 

    return spinDateDay->value();
}

//***********************************************************************

int DateInput::month()
{
     // This function returns an integer representation of the month set in
     // the month portion of the DateInput's Date. 

    return cboDateMonth->currentIndex() + 1;
}

//***********************************************************************

int DateInput::year()
{
     // This function returns an integer representation of the year set in
     // the year portion of the DateInput's Date. 
    return spinDateYear->value();
}

//***********************************************************************

void DateInput::DayChange()
{
     // This is a slot that is called when the number in the Day box is
     // changed.  It checks to see if the number has gone out of range
     // for the selected month and year.  If it has, the other two pieces
     // of the date are updated as necessary and the day is either changed
     // to the last day of the previous month, or the first day of the next
     // month.

    QDate tempDate;
    int month;

    if (tempDate.isValid(spinDateYear->value(), cboDateMonth->currentIndex()+1,
        spinDateDay->value()))
    {
        return;
    }
    else // invalid start date
    {
        // if value > 0, we must have exceed valid days in month, 
        // hence roll over to next month, and reset day to 1
        if (spinDateDay->value() > 0)
        {
            if (cboDateMonth->currentIndex() == 11)
            {
                cboDateMonth->setCurrentIndex(0);
                spinDateYear->setValue(spinDateYear->value() + 1);        
            }
            else
            {
                cboDateMonth->setCurrentIndex(cboDateMonth->currentIndex()+1);
            }

            spinDateDay->setValue(1);
        }
        else // rolled down to next month
        {
            // set month to previous month
            if (cboDateMonth->currentIndex() == 0)
            {
                month = 12;
                spinDateYear->setValue(spinDateYear->value() - 1);
            }
            else
            {
                month = cboDateMonth->currentIndex();
            }

            cboDateMonth->setCurrentIndex(month - 1);
            int day = 31;
            // set day to max days, and until a valid days is 
            // reached, keep decrementing
            while (!tempDate.isValid(spinDateYear->value(), month, day))
            {
                day--;
            }

            // set to new last day of previous month
            spinDateDay->setValue(day);
        }
    }

    return;
}

//***********************************************************************

void DateInput::MonthChange()
{
     // This is a slot that is called when the month in the month box is
     // changed.  It checks to see if the number in the day box has gone 
     // out of range for the selected month and year.  If it has, the other
     // two pieces of the date are updated as necessary.

    QDate tempDate;
    
    if (tempDate.isValid(spinDateYear->value(), cboDateMonth->currentIndex()+1, 
        spinDateDay->value()))
    {
        return;
    }
    else
    {
        int day = spinDateDay->value();
        
        // set day to current day, keep decrementing until valid
        while (!tempDate.isValid(spinDateYear->value(), 
            cboDateMonth->currentIndex() + 1, day))
        {
            day--;
        }

        // set to new last day of previous month
        spinDateDay->setValue(day);
    }
    
    return;
}

//***********************************************************************

int DateInput::isValidDate()
{
    // This function determines if the date currently stored within
    // this class instance is valid, and returns TRUE if valid, 
    // FALSE if not

    // use QDate::isValid function to check valid status of data stored in
    // this DateIput instance. (arguments are isValid(year,month,day)
    // add 1 to month because values stored start at 0 
    return (QDate::isValid((int) spinDateYear->value(), 
        (int) cboDateMonth->currentIndex() + 1, 
        (int) spinDateDay->value()));
}

//***********************************************************************

QString DateInput::toString(bool endOfDay) // I: Go to the *end* of the day?
{
    const QString strMonth[] =
    { "",
        "JAN", "FEB", "MAR", "APR", "MAY", "JUN",
        "JUL", "AUG", "SEP", "OCT", "NOV", "DEC"
    };

    QString strOracleDate;

    QString strDay;
    strDay.sprintf( "%02d", spinDateDay->value() );

    strOracleDate = strDay + "-";
    strOracleDate += strMonth[cboDateMonth->currentIndex() + 1];
    strOracleDate += "-";

    QString strYear;
    strYear.setNum( spinDateYear->value() );
    strOracleDate += strYear;

    if (timeVisible)
    {
        // If the time widgets for this particular DateInput widget
        // are visible, then use the value that they specify and
        // ignore the passed-in endOfDay flag.
        strOracleDate += " " + spinDateHour->text() + ":"
                         + spinDateMinute->text() + ":00";
    }
    else
    {
        if (endOfDay)
            strOracleDate += " 23:59:59";
        else
            strOracleDate += " 00:00:00";
    }

    return strOracleDate;
}

//***********************************************************************

QString DateInput::oracleFormat(bool endOfDay) // I: Go to the *end* of the day
{
    QString strOracleDate = "to_date('" + toString(endOfDay)
                            + "', 'DD-MON-YYYY HH24:MI:SS')";

    return strOracleDate;
}

