/******************************************************************************
               Level 1 Product Generation System (LPGS) Software
 ******************************************************************************
 *
 * UNIT NAME:   DateRange.cpp
 *
 * PURPOSE:
 * 
 * This class provides the operator the ability to select a date range.  There
 * are start and stop dates.  The dates are selectable via drop down combo 
 * boxes (months) or spin boxes (day and year).  The day and year can also be 
 * directly entered.  The class includes two date input instances, some labels 
 * for identification, and a reset button used to reset the stop date to the 
 * current date and time and the start date to the specified number of days
 * (value of daysBetween) before the stop date.
 *
 *****************************************************************************/

/* System Includes */
#include <iostream>

/* Qt Includes */
#include <QtCore>
#include <QtGui>

/* Local Includes */
#include "DateRange.h"

//*****************************************************************************

DateRange::DateRange(QWidget* parent, Qt::WindowFlags fl,
    int daysDiff,       // days between start and stop dates
    int direction)      // direction to display
  : QWidget(parent, fl)
{
    // This is the constructor for the DateInput class.  This constructor 
    // takes two significant arguements, daysDiff is the number of days
    // between the start date and stop date when the default settings are
    // used, or when the reset button is depressed.  The stop date is set to 
    // the current date, and the start day is set to daysDiff days prior to
    // the current date.  The second argument is the overall orientation of
    // the object, horizontal or vertical
    
    // set value of days between the start date and the stop date
    setDaysBetween(daysDiff);
    setWindowTitle("Date Select");
    
    grid = new QGridLayout(this); 
    grid->setSpacing(6);
    grid->setMargin(11);

    hbox = new QHBoxLayout(); 
    hbox->setSpacing(6);
    hbox->setMargin(0);

    vbox = new QVBoxLayout(); 
    vbox->setSpacing(6);
    vbox->setMargin(0);
    
    // create new start and stop date input instances
    // create in this order (diStart, diEnd) to get the right tab order
    diStart = new DateInput(this);
    diEnd = new DateInput(this);

    // create start and stop text labels
    lblStartDatetxt = new QLabel(this);
    lblStartDatetxt->setText(tr("Start Date"));
    lblStartDatetxt->setAlignment(Qt::AlignCenter);
     
    lblStopDatetxt = new QLabel(this);
    lblStopDatetxt->setText("Stop Date");
    lblStopDatetxt->setAlignment(Qt::AlignCenter);
    
    lblResetText = new QLabel(this);
    lblResetText->setText("Reset to Current");

    // create pushbutton to reset dates
    pbReset = new QPushButton(this);
    pbReset->setText("Reset");

    // if horizontal direction selected for date range
    if(direction == horizontal)
    {
        resize(363, 92); 

        vbox->addWidget(lblStartDatetxt);
    
        vbox->addWidget(diStart);
        hbox->addLayout(vbox);

        QVBoxLayout *vboxSpacer = new QVBoxLayout(); 
        vboxSpacer->setSpacing(6);
        vboxSpacer->setMargin(0);

        QSpacerItem* spacer = new QSpacerItem(20, 20,
            QSizePolicy::Expanding, QSizePolicy::Minimum);

        vboxSpacer->addItem(spacer);
        hbox->addLayout(vboxSpacer);

        vbox_2 = new QVBoxLayout(); 
        vbox_2->setSpacing(6);
        vbox_2->setMargin(0);

        vbox_2->addWidget(lblStopDatetxt);

        vbox_2->addWidget(diEnd);
        hbox->addLayout(vbox_2);

        vbox_3 = new QVBoxLayout(); 
        vbox_3->setSpacing(6);
        vbox_3->setMargin(0);

        vbox_3->addWidget(lblResetText);

        vbox_3->addWidget(pbReset);
        hbox->addLayout(vbox_3);
        grid->addLayout(hbox, 0, 0);
    }
    else  // vertical direction selected for date range
    {
        resize(312, 169); 
    
        vbox_2 = new QVBoxLayout(); 
        vbox_2->setSpacing(6);
        vbox_2->setMargin(0);
        vbox_2->addWidget(lblStartDatetxt);
        
        vbox_2->addWidget(diStart);
        vbox->addLayout(vbox_2);

        vbox_3 = new QVBoxLayout(); 
        vbox_3->setSpacing(6);
        vbox_3->setMargin(0);

        vbox_3->addWidget(lblStopDatetxt);
        vbox_3->addWidget(diEnd);
        vbox->addLayout(vbox_3);

        hbox = new QHBoxLayout(); 
        hbox->setSpacing(6);
        hbox->setMargin(0);

        hbox->addWidget(lblResetText);

        hbox->addWidget(pbReset);
        vbox->addLayout(hbox);

        grid->addLayout(vbox, 0, 0);
    }

    // signals and slots connections
    connect(pbReset, SIGNAL(clicked()), this, SLOT(setDateToCurrent()));

    // set date to current date and time
    setDateToCurrent();
}

//*****************************************************************************

DateRange::~DateRange()
{
    // no need to delete child widgets, Qt does it all for us
}

//*****************************************************************************

void DateRange::showTimes()
{
    diStart->showTime( FALSE );
    diEnd->showTime( TRUE );
}

int DateRange::daystoStart()
{
    // This function returns the number of days between the current 
    // date/time and the selected start date
    
    // get current date and call days to function using the start date
    return(QDate::currentDate().daysTo(diStart->getDate()));
}

//*****************************************************************************

int DateRange::daystoEnd()
{
    // This function returns the number of days between the current 
    // date/time and the selected end date
    
    // get current date and call days to function using the endig date
    return(QDate::currentDate().daysTo(diEnd->getDate()));
}

//*****************************************************************************

bool DateRange::isValid( bool notify )
{
    // This function returns valid if both the starting and ending dates
    // are valid.
    
    // if either start date or stop date are invalid,
    //      or the end date (and time) is before start date (and time)
    //          return FALSE
    if(!diStart->isValidDate() 
        || !diEnd->isValidDate() 
        || (diStart->getDateTime(false) > diEnd->getDateTime(true)))
    {
        // if we are to issue a warning message
        if( notify == TRUE)
            issueInvalidWarning();
        return FALSE;
    }
    
    return TRUE;
}

//*****************************************************************************

int DateRange::getDaysBetween()
{
    // return days between start and stop dates
    return(diStart->getDate().daysTo(diEnd->getDate()));
}

//*****************************************************************************

void DateRange::setDateToCurrent()
{
    // This function sets the stop time the current date and time and 
    // sets the start date (beginning) to the current date - daysBetween
    // days.  
    diEnd->setDate(QDate::currentDate());
    diEnd->setTimeToEnd();
    
    // set date to daysBetween days prior to the stop date
    diStart->setDate(diEnd->getDate().addDays(-daysBetween));
    diStart->setTimeToStart();
}

//*****************************************************************************

void DateRange::setDateToCurrent( const int days )
{
    // This function sets the stop time the current date and time and 
    // sets the start date (beginning) to the current date - days
    // days.  
    setDaysBetween(days);
    setDateToCurrent();
}

//*****************************************************************************

QString DateRange::getDateQueryString()
{
    // This functions returns the portion of a SQL statement that will include
    // all entries after midnight the day before the start date, and up to 
    // midnight the stop day
    
    // add one to daysToEnd because TRUNC() will truncate the time to 00:00:00 
    // on the end date, which will actuall not get you any events on the ending
    // date.  This will get you everything up to and including midnight of the
    // ending date

    QString tempstr = " BETWEEN " + diStart->oracleFormat( FALSE ) + " AND "
                      "" + diEnd->oracleFormat( TRUE );
    return tempstr;
}

//*****************************************************************************

void DateRange::setDaysBetween(const int days)
{
    // set value between ending and starting days.  This value is used to set
    // the start day when function setDateToCurrent( void ) is called.
    if(days >= 0)
        daysBetween = days;
    else 
        daysBetween = -days;
}

//*****************************************************************************

void DateRange::setStartDate(const QDate date)
{
    // This functions sets the start date to the date passed in
    diStart->setDate(date);
}

//*****************************************************************************

void DateRange::setStopDate(const QDate date)
{
    // This function sets the stop date to the date passed in
    diEnd->setDate(date);
}

//*****************************************************************************

void DateRange::setStartDay(const int day)
{
    // This function sets the start day to the day passed in
    diStart->setDay(day);
}

//*****************************************************************************

void DateRange::setStopDay(const int day)
{
    // This function set the stop day to the day passed in
    diEnd->setDay(day);
}

//*****************************************************************************

void DateRange::setStartYear(const int year)
{
    // This function sets the start year to the year passed in
    diStart->setYear(year);
}

//*****************************************************************************

void DateRange::setStopYear(const int year)
{
    // This function set the stop year to the year passed in
    diEnd->setYear(year);
}

//*****************************************************************************

void DateRange::setStartMonth(const int month)
{
    // This function sets the start month to the month passed i
    diStart->setMonth(month);
}

//*****************************************************************************

void DateRange::setStopMonth(const int month)
{
    // This function set the stop month to the month passed in
    diEnd->setMonth(month);
}

//*****************************************************************************

QDate DateRange::getStartDate()
{
    // This functions returns the start date
    return (diStart->getDate());
}

//*****************************************************************************

QDate DateRange::getStopDate()
{
    // This function sets the stop date to the date passed in
    return (diEnd->getDate());
}

//*****************************************************************************

int DateRange::getStartDay()
{
    // This function sets the start day to the day passed in
    return (diStart->day());
}

//*****************************************************************************

int DateRange::getStopDay()
{
    // This function set the stop day to the day passed in
    return (diEnd->day());
}

//*****************************************************************************

int DateRange::getStartYear()
{
    // This function returns the start year as an int
    return (diStart->year());
}

//*****************************************************************************

int DateRange::getStopYear()
{
    // This function returns the stop year as an int 
    return (diEnd->year());
}

//*****************************************************************************

int DateRange::setStartMonth()
{
    // This function returns the start month as an int
    return (diStart->month());
}

//*****************************************************************************

int DateRange::getStopMonth()
{
    // This function returns the stop month as an int
    return (diEnd->month());
}

//*****************************************************************************

QString DateRange::oracleFormatEnd()
{
    // This function returns the start month as an int
    return (diEnd->oracleFormat());
}

//*****************************************************************************

QString DateRange::oracleFormatStart()
{
    // This function returns the stop month as an int
    return (diEnd->oracleFormat());
}

//*****************************************************************************

void DateRange::issueInvalidWarning()
{
    static bool warningDisplayed = FALSE;
    QString strErrorHdr = "Date Range Selection";
    QString strRetryMsg = "Please enter a valid date and retry.";
    
    // if warning already displayed, don't redisplay
    if(warningDisplayed == FALSE)
    {
        if(!diStart->isValidDate())
        {
            warningDisplayed = TRUE;
            QMessageBox::warning(this, strErrorHdr,
                "You have entered an Invalid Start Date for Query.\n" +
                strRetryMsg,
                QMessageBox::Ok | QMessageBox::Default,0);
        }
        else if(!diEnd->isValidDate())
        {
            warningDisplayed = TRUE;
            QMessageBox::warning(this, strErrorHdr,
                "You have entered an Invalid Ending Date for Query.\n" +
                strRetryMsg,
                QMessageBox::Ok | QMessageBox::Default,0);
        }
        else if (daystoEnd() < daystoStart())
        {
            warningDisplayed = TRUE;
            QMessageBox::warning(this, strErrorHdr,
                "The starting day must be before the ending day.\n" +
                strRetryMsg,
                QMessageBox::Ok | QMessageBox::Default,0);
        }
        else if (diStart->getDateTime(false) > diEnd->getDateTime(true))
        {
            warningDisplayed = TRUE;
            QMessageBox::warning(this, strErrorHdr,
                "The starting time must be before the ending time.\n" +
                strRetryMsg,
                QMessageBox::Ok | QMessageBox::Default,0);
        }
        else
        {
            warningDisplayed = TRUE;
            QMessageBox::information(this, strErrorHdr,
                "Selected Dates are valid.\n",
                QMessageBox::Ok | QMessageBox::Default);
        }
        warningDisplayed = FALSE;
    }
}

//*****************************************************************************

QString DateRange::toString()
{
    // This function outputs a QString (in Oracle Format) that contains the 
    // period covered by the DateRange
    return("Start: " + diStart->toString(FALSE) + " Stop: " 
        + diEnd->toString(TRUE));
}

