/******************************************************************************
               Level 1 Product Generation System (LPGS) Software
 ******************************************************************************
 *
 * UNIT NAME:   DiskUsage.cpp
 *
 * PURPOSE:
 * This is the definition for the DiskUsage class that uses progress bars
 * to show the percentage of disk space on different disk volumes that is
 * being used.  
 *
 *****************************************************************************/

/* System Includes */
#include <iostream>
#include <stdio.h>

/* Qt Includes */
#include <QtCore>
#include <QtGui>

/* Local Includes */
#include "DiskUsage.h"
#include "IASSqlQuery.h"

DiskUsage::DiskUsage(QWidget *parent, Qt::WindowFlags fl) 
  : QDialog(parent, fl)
{
    setupUi(this);

    // Calls the base class constructor to instantiate the disk space
    // dialog, then adds the needed progress bars to the dialog, and then sets 
    // all the values by calling refresh.

    numDiskUsageVolumes = 0;
    usedPosition = 0;
    availablePosition = 0;
    percentPosition = 0;

    gridDiskUsage = new QGridLayout(frameDiskUsage); 
    gridDiskUsage->setSpacing(6);
    gridDiskUsage->setMargin(11);

    // The following QColorGroup and QPalette commands are for setting
    // up the look of the QProgress bars that will show the disk usage
    // in percentages.
    pal.setColor(QPalette::Active, QPalette::Foreground, Qt::black);
    pal.setColor(QPalette::Active, QPalette::Button, QColor(192, 192, 192));
    pal.setColor(QPalette::Active, QPalette::Light, Qt::black);
    pal.setColor(QPalette::Active, QPalette::Midlight, QColor(223, 223, 223));
    pal.setColor(QPalette::Active, QPalette::Dark, QColor(96, 96, 96));
    pal.setColor(QPalette::Active, QPalette::Mid, QColor(128, 128, 128));
    pal.setColor(QPalette::Active, QPalette::Text, Qt::black);
    pal.setColor(QPalette::Active, QPalette::BrightText, Qt::white);
    pal.setColor(QPalette::Active, QPalette::ButtonText, Qt::black);
    pal.setColor(QPalette::Active, QPalette::Base, Qt::white);
    pal.setColor(QPalette::Active, QPalette::Background, Qt::white);
    pal.setColor(QPalette::Active, QPalette::Shadow, Qt::black);
    pal.setColor(QPalette::Active, QPalette::Highlight, QColor(0, 0, 128));
    pal.setColor(QPalette::Active, QPalette::HighlightedText, Qt::white);
    
    pal.setColor(QPalette::Inactive, QPalette::Foreground, Qt::black);
    pal.setColor(QPalette::Inactive, QPalette::Button, QColor(192, 192, 192));
    pal.setColor(QPalette::Inactive, QPalette::Light, Qt::white);
    pal.setColor(QPalette::Inactive, QPalette::Midlight, QColor(220, 220, 220));
    pal.setColor(QPalette::Inactive, QPalette::Dark, QColor(96, 96, 96));
    pal.setColor(QPalette::Inactive, QPalette::Mid, QColor(128, 128, 128));
    pal.setColor(QPalette::Inactive, QPalette::Text, Qt::black);
    pal.setColor(QPalette::Inactive, QPalette::BrightText, Qt::white);
    pal.setColor(QPalette::Inactive, QPalette::ButtonText, Qt::black);
    pal.setColor(QPalette::Inactive, QPalette::Base, Qt::white);
    pal.setColor(QPalette::Inactive, QPalette::Background, Qt::white);
    pal.setColor(QPalette::Inactive, QPalette::Shadow, Qt::black);
    pal.setColor(QPalette::Inactive, QPalette::Highlight, QColor(0, 0, 128));
    pal.setColor(QPalette::Inactive, QPalette::HighlightedText, Qt::white);
    
    pal.setColor(QPalette::Disabled, QPalette::Foreground,
        QColor(128, 128, 128));
    pal.setColor(QPalette::Disabled, QPalette::Button, QColor(192, 192, 192));
    pal.setColor(QPalette::Disabled, QPalette::Light, Qt::white);
    pal.setColor(QPalette::Disabled, QPalette::Midlight, QColor(220, 220, 220));
    pal.setColor(QPalette::Disabled, QPalette::Dark, QColor(96, 96, 96));
    pal.setColor(QPalette::Disabled, QPalette::Mid, QColor(128, 128, 128));
    pal.setColor(QPalette::Disabled, QPalette::Text, Qt::black);
    pal.setColor(QPalette::Disabled, QPalette::BrightText, Qt::white);
    pal.setColor(QPalette::Disabled, QPalette::ButtonText, 
        QColor(128, 128, 128));
    pal.setColor(QPalette::Disabled, QPalette::Base, Qt::white);
    pal.setColor(QPalette::Disabled, QPalette::Background, Qt::white);
    pal.setColor(QPalette::Disabled, QPalette::Shadow, Qt::black);
    pal.setColor(QPalette::Disabled, QPalette::Highlight, QColor(0, 0, 128));
    pal.setColor(QPalette::Disabled, QPalette::HighlightedText, Qt::white);

    connect(pbDismiss, SIGNAL(clicked()), this, SLOT(accept()));
    connect(pbRefresh, SIGNAL(clicked()), this, SLOT(refresh()));

    refresh();
}

//*****************************************************************************

DiskUsage::~DiskUsage()
{
    // destructor deletes all pointers to the labels and progress bars
    // that are currently used.

    for (int counter = 0; counter < numDiskUsageVolumes; counter++)
    {
        delete diskSpaceInfo[counter].pbar;
        delete diskSpaceInfo[counter].lbl;
        delete diskSpaceInfo[counter].lblSizeInfo;
    }

    delete gridDiskUsage;
}

//*****************************************************************************

void DiskUsage::refreshLocalStruct()
{
    // This function sets up the local struct again with the info needed
    // to get the necessary disk space.

    int counter;
    IASSqlQuery query;
    QStringList strlDfCommandResults;

    // de-allocate memory if numDiskUsageVolumes is greater than 0
    // The constructor sets the numDiskUsageVolumes to 0, so the first
    // time this is called, this loop is not exected.
    for (counter = 0; counter < numDiskUsageVolumes; counter++)
    {
        delete diskSpaceInfo[counter].pbar;
        delete diskSpaceInfo[counter].lbl;
        delete diskSpaceInfo[counter].lblSizeInfo;

        diskSpaceInfo[counter].pbar = NULL;
        diskSpaceInfo[counter].lbl = NULL;
        diskSpaceInfo[counter].lblSizeInfo = NULL;
        diskSpaceInfo[counter].strVolumeName = "";
    }

    GUIObjects tempGUIInfo [] = 
    {
        {NULL, NULL, NULL, "/", 0, "", ""},
        {NULL, NULL, NULL, "/usr", 0, "", ""},
    };

    numDiskUsageVolumes = sizeof(tempGUIInfo) / sizeof(tempGUIInfo[0]);

    for (counter = 0; counter < numDiskUsageVolumes; counter++)
    {
        diskSpaceInfo[counter] = tempGUIInfo[counter];
        strlDfCommandResults = getDiskUsageValues(
                                   diskSpaceInfo[counter].strVolumeName);

        diskSpaceInfo[counter].percentUsed =
            strlDfCommandResults[percentPosition]
                .left(strlDfCommandResults[percentPosition].length() - 1)
                .toInt();

        diskSpaceInfo[counter].strMbUsed = strlDfCommandResults[usedPosition];
        diskSpaceInfo[counter].strMbAvailable =
            strlDfCommandResults[availablePosition];
    }
    
    if (query.exec("SELECT distinct def_path FROM def_directories "
                   "WHERE file_type = 'PREQ'", __FILE__, __LINE__))
    {
        // Add the raid group disk usage info to the struct.
        for (counter = numDiskUsageVolumes; query.next();)
        {
            QString volumeName;
            int baseEnd;
            int hasBeenChecked = 0;

            // Check the max number of volumns to display has not been reached
            if (counter == MAX_DISK_VOLUMES)
            {
                QMessageBox::warning(this, "Max Display Volumes Reached",
                    "The maximum number of volumes that can be displayed has "
                    "been reached. No additional volumes will be added to the "
                    "display.");
                break;
            }

            // Format the volume name to the base directory
            volumeName = query.value("def_path").toString();
            baseEnd = volumeName.indexOf('/', 1);
            if (baseEnd > 1)
            {
                volumeName = volumeName.mid(0, baseEnd);
            }

            // Check the volume names for duplicates
            for (int i = 0; i < counter; i++)
            {
                if (diskSpaceInfo[i].strVolumeName == volumeName)
                {
                    hasBeenChecked = 1;
                    break;
                }
            }

            // Check to see if this has been checked already
            if (hasBeenChecked)
            {
                continue;
            }

            strlDfCommandResults = getDiskUsageValues(volumeName);

            diskSpaceInfo[counter].strVolumeName = volumeName;

            diskSpaceInfo[counter].percentUsed =
                strlDfCommandResults[percentPosition]
                    .left(strlDfCommandResults[percentPosition].length() - 1)
                    .toInt();

            diskSpaceInfo[counter].strMbUsed =
                strlDfCommandResults[usedPosition];
            diskSpaceInfo[counter].strMbAvailable =
                strlDfCommandResults[availablePosition];

            counter++;
        }
    }

    // Update the number of disk space volumes by adding the number of 
    // raid groups that were defined and selected.
    numDiskUsageVolumes = counter;
}

//*****************************************************************************

void DiskUsage::refresh()
{
    // This function refreshes the list of disk volumes and the associated
    // disk volume info for each disk volume depending on the results of a df 
    // command.  Before recreating widgets, it calls refreshLocalStruct
    // which refreshes the values and cleans up any pointers that will no
    // longer be used.

    int row;
    int diskVolume;
    QString strSizeInfo;
    QString strShortName;

    refreshLocalStruct();

    for (row = 0, diskVolume = 0; diskVolume < numDiskUsageVolumes;
         row += 2, diskVolume++)
    {
        diskSpaceInfo[diskVolume].lbl = new QLabel(frameDiskUsage);

        strShortName = diskSpaceInfo[diskVolume].strVolumeName;
        strShortName.truncate(10);
        diskSpaceInfo[diskVolume].lbl->setText(strShortName);
        diskSpaceInfo[diskVolume].lbl->setAlignment(
            Qt::AlignVCenter | Qt::AlignRight);
        diskSpaceInfo[diskVolume].lbl->setSizePolicy(QSizePolicy(
            QSizePolicy::Minimum, QSizePolicy::Minimum));

        gridDiskUsage->addWidget(diskSpaceInfo[diskVolume].lbl, row, 0);
        diskSpaceInfo[diskVolume].lbl->show();

        diskSpaceInfo[diskVolume].pbar = new QProgressBar(frameDiskUsage);
        diskSpaceInfo[diskVolume].pbar->setValue(
            diskSpaceInfo[diskVolume].percentUsed);
        diskSpaceInfo[diskVolume].pbar->setSizePolicy(QSizePolicy(
            QSizePolicy::Expanding, QSizePolicy::Fixed));
        diskSpaceInfo[diskVolume].pbar->setMinimumSize(QSize(390, 0));
        diskSpaceInfo[diskVolume].pbar->setPalette(pal);

        gridDiskUsage->addWidget(diskSpaceInfo[diskVolume].pbar, row, 1);
        diskSpaceInfo[diskVolume].pbar->show();

        diskSpaceInfo[diskVolume].lblSizeInfo = new QLabel(frameDiskUsage);

        strSizeInfo = "Used: " + diskSpaceInfo[diskVolume].strMbUsed 
            + "       Available: " 
            + diskSpaceInfo[diskVolume].strMbAvailable + "\n";
    
        diskSpaceInfo[diskVolume].lblSizeInfo->setText(strSizeInfo);
        diskSpaceInfo[diskVolume].lblSizeInfo->setAlignment(
            Qt::AlignTop | Qt::AlignLeft);
        diskSpaceInfo[diskVolume].lblSizeInfo->setSizePolicy(QSizePolicy(
            QSizePolicy::Expanding, QSizePolicy::Fixed));

        gridDiskUsage->addWidget(diskSpaceInfo[diskVolume].lblSizeInfo,
                                 row + 1, 1);
        diskSpaceInfo[diskVolume].lblSizeInfo->show();
    }
}

QStringList DiskUsage::getDiskUsageValues(QString strDiskVolumeToGet)
{
    // This function returns a qstringlist with the columns returned by a
    // df command separated out into the different QStringList items.

    QStringList strlTempList;
    char tempChar[2000];
    QString strDiskUsedUnit = "KB";
    QString diskAvailableUnit = "KB";
    float diskUsedSize;
    float diskAvailableSize;
    int counter;

    FILE *pipe;
    QString cmd;

#ifdef LINUX
    cmd = "df -TP " + strDiskVolumeToGet;
    pipe = popen(cmd.toAscii().data(), "r");
    usedPosition = 3;
    availablePosition = 4;
    percentPosition = 5;
#else
    cmd "df -k " + strDiskVolumeToGet;
    pipe = popen(cmd.toAscii().data(), "r");
    usedPosition = 2;
    availablePosition = 3;
    percentPosition = 4;
#endif

    // Create a list of "ERROR" values to return if an error occurs
    QStringList errorList;
    for (counter = 0; counter <= percentPosition; counter++)
    {
        errorList += "ERROR";
    }

    if (pipe == NULL)
    {
        QMessageBox::critical(this, "Error", "Error finding the disk usage "
            "information for " + strDiskVolumeToGet);
        return errorList;
    }

    // Read header line of df output
    char *result = fgets(tempChar, 2000, pipe);
    if (result == NULL)
    {
        QMessageBox::critical(this, "Error", "Error reading output from the "
            "df command for " + strDiskVolumeToGet);
        return errorList;
    }

    // Read the disk usage info from df comand
    result = fgets(tempChar, 2000, pipe);
    if (result == NULL)
    {
        QMessageBox::critical(this, "Error", "Error reading output from the "
            "df command for " + strDiskVolumeToGet);
        return errorList;
    }

    if (pclose(pipe) == -1)
    {
        QMessageBox::warning(this, "Warning", "Error closing pipe for df "
            "command");
    }

    QString tempStr = tempChar;
    strlTempList = tempStr.split(' ', QString::SkipEmptyParts);

    // Make sure split command produced enough values
    if (strlTempList.count() >=  (percentPosition + 1))
    {
        diskUsedSize = strlTempList[usedPosition].toFloat();
        diskAvailableSize = strlTempList[availablePosition].toFloat();

        for (counter = 0; counter < 3; counter++)
        {
            if (diskUsedSize > 1024)
            {
                diskUsedSize = diskUsedSize/1024;        
                switch (counter)
                {
                    case 0: strDiskUsedUnit = "MB";
                        break;
                    case 1: strDiskUsedUnit = "GB";
                        break;
                    case 2: strDiskUsedUnit = "TB";
                        break;
                }
            }

            if (diskAvailableSize > 1024)
            {
                diskAvailableSize = diskAvailableSize/1024;
                switch (counter)
                {
                    case 0: diskAvailableUnit = "MB";
                        break;
                    case 1: diskAvailableUnit = "GB";
                        break;
                    case 2: diskAvailableUnit = "TB";
                        break;
                }
            }
        } // end for counter = 0.....

        strlTempList[usedPosition].sprintf("%0.2f ", diskUsedSize);
        strlTempList[usedPosition] += strDiskUsedUnit;
        strlTempList[availablePosition].sprintf("%0.2f ", diskAvailableSize);
        strlTempList[availablePosition] += diskAvailableUnit;
    }
    else // the df command did not find the requested volume
    {
        return errorList;
    }

    return strlTempList;
}

