/******************************************************************************
               Level 1 Product Generation System (LPGS) Software
******************************************************************************
 *
 * UNIT NAME:   EventLog.cpp
 *
 * PURPOSE:
 * This class provides access to items in the events and messages tables.
 * The type of event to include in the list (Status, Warnings, and Errors) can
 * be selected, date range, and also messages relating to a specific item 
 * (product request id, task name, work order id, or scripts
 * id).  The data to generate the event list view is stored in a structure
 * elInfo.  
 *
 *****************************************************************************/
// System Includes 

// Qt Includes 
#include <QtCore>
#include <QtGui>

// Local Includes 
#include "DatabaseUtils.h"
#include "EventLog.h"
#include "Preferences.h"
#include "IASUtilities.h"
#include "IASSqlQuery.h"

extern Preferences *preferences;

// See ListViewUtilities.h for lvInfo type definition.  If you change the
// columns below the #defines in EventLog.h will need to change as well
twInfo EventLog::elInfo[] =
{
    {"Date/Time ", Qt::AlignLeft, -1, TRUE, TRUE, TRUE, 115, sDATE_TIME},
    {"Product Request ID", Qt::AlignLeft, -1, TRUE, TRUE, TRUE, 115, sPR_ID},
    {"Work Order ID ", Qt::AlignCenter, 140, TRUE, TRUE, TRUE, 90, sWO_ID},
    {"Script ID ", Qt::AlignCenter, -1, TRUE, TRUE, TRUE, 70, sSCRIPT_NAME},
    {"Task ID ", Qt::AlignCenter, -1, TRUE, TRUE, TRUE, 50, sTASK_NAME},
    {"Brief Message ", Qt::AlignLeft, 500, TRUE, TRUE, TRUE, 400, sMSG},
    {"SequenceNumber", Qt::AlignLeft, 0, TRUE, FALSE, FALSE, -1, sSEQ_NUM},
    {"Severity ", Qt::AlignLeft, 0, TRUE, FALSE, FALSE, -1, sSEV},
};
 
//*****************************************************************************

EventLog::EventLog(QWidget *parent, Qt::WindowFlags fl) 
  : QDialog(parent, fl), printer(0)
{
    setupUi(this);

    numColumns = ARRAY_SIZE(elInfo);
    
    // set number of days to show, reset to end today, and start 
    // the user-defined number of  days ago.  Subtract 1 because
    // ending is included 
    dr->setDateToCurrent(preferences->strEventLogDuration.toInt() - 1);

    // set swap text and background colors to FALSE
    swapColors = FALSE;

    // clear selected sequence number since no items have been selected yet
    strSelectedSequenceNumber = "";    

    // set items that can be selected to the column titles of those rows 
    cobColumnSelect->addItem(elInfo[nPR_ID].strColumnTitle);
    cobColumnSelect->addItem(elInfo[nWO_ID].strColumnTitle);
    cobColumnSelect->addItem(elInfo[nSCRIPT_NAME].strColumnTitle);
    cobColumnSelect->addItem(elInfo[nTASK_NAME].strColumnTitle);

    // set here to ensure auto-completion works 
    cobValueSelect->setEditable(TRUE);
    cobValueSelect->setAutoCompletion(TRUE);

    // update status bar
    updateSelect("0");  
    updateNumReturned("0");
    updateDateTime();

    // build empty list to hold events
    buildEventListView();

    // update event list view with selected events
    refreshList();

    // update list of available items to select from, i.e., list of:
    // product requests, task order ids, work orders, script ids, etc.
    updateSelectList();

    setupSignalSlots();
}

//*****************************************************************************
EventLog::EventLog(QString strPRID, QWidget *parent, Qt::WindowFlags fl) 
  : QDialog(parent, fl), printer(0)
{
    setupUi(this);

    strTables = " events, messages ";
    strJoin = " WHERE events.message_id = messages.message_id ";

    numColumns = ARRAY_SIZE(elInfo);

    // set number of days to show, reset to end today, and start 
    // the user-defined number of  days ago.  Subtract 1 because
    // ending is included 
    dr->setDateToCurrent(preferences->strEventLogDuration.toInt() - 1);

    // set swap text and background colors to FALSE
    swapColors = FALSE;

    // clear selected sequence number since no items have been selected yet
    strSelectedSequenceNumber = ""; 

    // set items that can be selected to the column titles of those rows 
    cobColumnSelect->addItem(elInfo[nPR_ID].strColumnTitle);
    cobColumnSelect->addItem(elInfo[nWO_ID].strColumnTitle);
    cobColumnSelect->addItem(elInfo[nSCRIPT_NAME].strColumnTitle);
    cobColumnSelect->addItem(elInfo[nTASK_NAME].strColumnTitle);

    // set here to ensure auto-completion works 
    cobValueSelect->setEditable(TRUE);
    cobValueSelect->setAutoCompletion(TRUE);
    
    // update status bar
    updateSelect("0");  
    updateNumReturned("0");
    updateDateTime();

    // refresh list of select items, if they were selected, they would have
    // been removed from the select list combo-box
    getSelectItems();

    // ensure selection is not an empty string
    if (!strPRID.isEmpty())
    {
        // remove the item from the select object.
        cobValueSelect->removeItem(cobValueSelect->findText(strPRID));

        // insert selection into list
        lwSelectedList->addItem(strPRID);

        // make last selection current the item, and ensure it's visible
        lwSelectedList->setCurrentRow(lwSelectedList->count() - 1);

        IASSqlQuery query;
        strSQLCommand =
            "SELECT trunc(min(event_date)) AS start_date, "
            "trunc(max(event_date)) AS stop_date "
            "FROM events "
            "WHERE product_request_id = '" + strPRID + "'";
        if (query.exec(strSQLCommand, __FILE__, __LINE__))
        {
            if (query.next())
            {
                dr->setStartDate(query.value("start_date").toDate());
                dr->setStopDate(query.value("stop_date").toDate());
            }
        }
    }
    // update the number of items in the list (number selected (copied)
    updateSelect(QString::number(lwSelectedList->count(), 10));

    // build empty list to hold events
    buildEventListView();

    // update event list view with selected events
    refreshList();

    setupSignalSlots();
}

//*****************************************************************************
void EventLog::setupSignalSlots()
{
    connect(cobColumnSelect, SIGNAL(activated(QString)), 
        this, SLOT(updateSelectList()));
    connect(chbAll, SIGNAL(clicked()), this, SLOT(allSelected()));
    connect(chbErrors, SIGNAL(clicked()), this, SLOT(errorsSelected()));
    connect(chbWarnings, SIGNAL(clicked()), this, SLOT(warningsSelected()));
    connect(chbStatus, SIGNAL(stateChanged(int)),this, SLOT(statusSelected()));
    connect(cobValueSelect, SIGNAL(activated(QString)), this, SLOT(addItem()));
    connect(twEventList, SIGNAL(itemDoubleClicked(QTreeWidgetItem*, int)),
        this, SLOT(queryOnItem()));
    connect(twEventList, SIGNAL(itemClicked(QTreeWidgetItem*, int)), 
        this, SLOT(eventOperatorSelected(QTreeWidgetItem*)));
    connect(twEventList, SIGNAL(itemSelectionChanged()), 
        this, SLOT(eventSelectionChanged()));
    connect(pbAddItem, SIGNAL(clicked()), this, SLOT(addItem()));
    connect(pbSwapColors, SIGNAL(clicked()), 
        this, SLOT(swapEventListColors()));
    connect(pbClearSelectedList, SIGNAL(clicked()), 
        this, SLOT(clearSelectedList()));
    connect(pbRefresh, SIGNAL(clicked()), this, SLOT(refreshList()));
    connect(pbPrint, SIGNAL(clicked()), this, SLOT(print()));
    connect(pbDismiss, SIGNAL(clicked()), this, SLOT(accept()));
    connect(pbRefresh_2, SIGNAL(clicked()), this, SLOT(refreshList()));

    // enabling the custom context menu so the right mouse click can
    // be mapped to the addItem function
    twEventList->setContextMenuPolicy(Qt::CustomContextMenu);
    connect(twEventList, SIGNAL(customContextMenuRequested(const QPoint&)),
        this, SLOT(addItem()));
}

//*****************************************************************************
void EventLog::refreshList()
{
    // This function refreshes the list of items in the event log  
    
    // if operator selected errors, warnings, and status, clear them and 
    // set all to checked
    if (chbStatus->isChecked() == TRUE && chbWarnings->isChecked() == TRUE 
        && chbErrors->isChecked() == TRUE) 
    {
        chbAll->setChecked(TRUE);
        chbStatus->setChecked(FALSE);
        chbWarnings->setChecked(FALSE);
        chbErrors->setChecked(FALSE);
    }    // else if operator has selected nothing, select all
    else if (chbStatus->isChecked() == FALSE 
             && chbWarnings->isChecked() == FALSE 
             && chbErrors->isChecked() == FALSE
             && chbAll->isChecked() == FALSE) 
    {
        chbAll->setChecked(TRUE);
    }

    // check selected date range, issue warning message if invalid
    if (dr->isValid(TRUE) == FALSE)
        return;

    // Query database and get a list of items from events table.  Place events 
    // in event list
    getEvents(); 
    
    // Reselect same items as was seleted before it is is in the list
    reselectLastItemSelected();

    // update time on status bar
    updateDateTime();

    // update number of records returned
    updateNumReturned(QString::number(twEventList->topLevelItemCount(), 10));

    // query database and retrieve new list of select items only if they
    // do not have any items selected now
    if (lwSelectedList->count() == 0)
        getSelectItems();
}

//*****************************************************************************

void EventLog::addItem()
{
    // This function (slot) is used to copy the string from item selected in
    // the listbox and add to list of items to search for in database query
    QLineEdit *le;

    le = cobValueSelect->lineEdit();

    // if no item are selected in the Column Value Select combo-box, 
    //     copy specified item from event log listing in list view
    if (cobValueSelect->currentIndex() < 0)
    {
        // if combo box is edited, and is not Empty,
        //   use this value to query database
        if (le->isModified() && !le->text().isEmpty())
        {
            strTemp = le->text();
            le->setText("");
        }

        // if there are not items selected in list copy from event list
        else if (twEventList->currentItem() != 0)
            strTemp = getSelectedColumnText();
        else
            return;
            
        // remove item copied from the Column Value Select combo-box
        for (int index = 0; index < cobValueSelect->count(); index++)
        {
            if (strTemp.compare(cobValueSelect->itemText(index)) == 0)
            {
                cobValueSelect->removeItem(index);
                break;
            }
        }
    }
    else // item selected in combo list, copy it
    {   
        // string is the data displayed in drop down text box
        strTemp = cobValueSelect->currentText();

        // if not an empty string
        if (!strTemp.isEmpty())
        {
            // remove item copied from select list from that list
            cobValueSelect->removeItem(cobValueSelect->currentIndex());

            // clear selected item, when one item is removed, the 
            //  next item is automatically selected
            cobValueSelect->setCurrentIndex(-1);
        }
    }

    // ensure selection is not an empty string
    if (!strTemp.isEmpty())
    {
        bool found = FALSE;    // found indicates item not found in list

        // check new item against other items in list (already selected)
        // do no copy duplicates into list
        for (int index = 0; index < int(lwSelectedList->count()); index++)
        {
            // new item match item in list already?
            if (!strTemp.compare(lwSelectedList->item(index)->text()))
            {
                found = TRUE;
                break;
            }
        }

        // if item was not found in list, insert it into list
        if (!found)
        {
            // insert selection into list
            lwSelectedList->addItem(strTemp);

            // make last selection current the item, and ensure it's visible
            lwSelectedList->setCurrentRow(lwSelectedList->count() - 1);

            // scroll list to show last item in list
            //lwSelectedList->centerCurrentItem();
        }
    }

    // update the number of items in the list (number selected (copied)
    updateSelect(QString::number(lwSelectedList->count(), 10));
}     

//*****************************************************************************

void EventLog::clearSelectedList()
{
    // This function (slot) clears the list of items selected and the sequence
    // number of the curently selected item, so the last item in the list will 
    // be selected next time
    
    // clear list of items that were selected 
    lwSelectedList->clear();

    // set selected sequence number to an emptry string
    strSelectedSequenceNumber = "";

    // refresh list of select items, if they were selected, they would have
    // been removed from the select list combo-box
    getSelectItems();

    // clear number of itmes, selected
    updateSelect(QString::number(lwSelectedList->count(), 10));
}

//*****************************************************************************

void EventLog::updateSelectList()
{
    // This function will clear the selected items, and fill the options for
    // selections based on the column the operator has selected
 
    // clear any curently selected items, since they are no longer appropriate
    // because a different column has been selected
    lwSelectedList->clear();

    // set selected sequence number to an emptry string
    strSelectedSequenceNumber = "";

    // refresh list of select items, if they were selected, they would have
    // been removed from the select list combo-box
    getSelectItems();

    // clear number of itmes, selected
    updateSelect(QString::number(lwSelectedList->count(), 10));

    // set label for combo box select item to include the columns's name
    lblColumnItemSelect->setText("Select " 
        + elInfo[getSelectedColumn()].strColumnTitle);
}

//*****************************************************************************

void EventLog::buildEventListView()
{
    // This functions builds\creates the event list view (table), by adding the
    // columns and setting the alignment, etc.
    QTreeWidgetItem *header = new QTreeWidgetItem();
    for (int column = 0; column < numColumns; column++)
    {
        header->setText(column, elInfo[column].strColumnTitle);
        header->setTextAlignment(column, elInfo[column].columnAlignment);
    }

    twEventList->setHeaderItem(header);
    twEventList->setSelectionBehavior(QAbstractItemView::SelectRows);
    twEventList->setRootIsDecorated(false);
    twEventList->setSortingEnabled(true);

    for (int col = 0; col < twEventList->columnCount(); col++)
        twEventList->resizeColumnToContents(col);

    twEventList->setFocus();
}
 
//*****************************************************************************

void EventLog::updateDateTime()
{
    // This function updates the time displayed at the top of the Event Log

    // set the datetime label value to the current date and time
    lblDateTime->setText(QDateTime::currentDateTime().toString());

    // adjust label size to fit contents
    lblDateTime->adjustSize();
}

//*****************************************************************************

void EventLog::updateNumReturned( 
    QString cntReturn)// I: cntReturn is a string containing the number of items
                      //    returned in the query, (number of rows in RecordSet)
{
    // update the label on the Event Log to indicate number of rows returned
    lblNumberReturned->setText("Number Returned: " + cntReturn);

    // call adjustSize to ensure the whole label is visible
    lblNumberReturned->adjustSize();
}

//*****************************************************************************

void EventLog::updateSelect(
    QString cntSelect) // I:  The number of items selected
{
    // This function updates the label at the top of the Event Log GUI 
    // indicating the number of items the operator has selected to query on,
    // such as specific product_request_ids

    // update the label on the GUI
    lblNumberSelected->setText("Number Selected: " + cntSelect);

    // call adjustSize to ensure the whole label is visible
    lblNumberSelected->adjustSize();
}

//*****************************************************************************

int EventLog::getSelectedColumn()
{
    // This function returns a number (int) that indicates which column is
    // selected above in the select list.  If a match is not found, -1 is
    // returned

    // get current selection
    strTemp = cobColumnSelect->currentText();

    // loop through all columns in structure to determine which column the
    // operator has selected in the Column Select combo box 
    for (int column = 0; column < numColumns; column++)
        if (strTemp.compare(elInfo[column].strColumnTitle) == 0)
            return (column);
   
    // return -1 if not found
    return -1;
}

//*****************************************************************************

QString EventLog::getSelectedColumnText()
{
    // This function returns the string contained in the selected listview 
    // item for the column selected in the list above on the GUI
    return twEventList->currentItem()->text(getSelectedColumn());
}

//*****************************************************************************

void EventLog::allSelected()
{
    // This fuction checks to see if all types of messages are selected, and
    // if they are all selected, it sets the other check boxs to unchecked

    if (chbAll->isChecked() == TRUE)
    {
        chbStatus->setChecked(FALSE);
        chbWarnings->setChecked(FALSE);
        chbErrors->setChecked(FALSE);
    }
}

//****************************************************************************

void EventLog::errorsSelected()
{
    // This slot is called when errors are selected.  If all is selected it
    // will be cleared

    // If errors is checked, and if all is selected, all is cleared
    if (chbErrors->isChecked() == TRUE)
        if (chbAll->isChecked() == TRUE)
            chbAll->setChecked(FALSE);
}

//****************************************************************************

void EventLog::warningsSelected()
{
    // This slot is called when warnings is selected.  If all is selected
    // (checked) it will be cleared

    // If warnings is checked, and if all is selected, all is cleared
    if (chbWarnings->isChecked() == TRUE)
        if (chbAll->isChecked() == TRUE)
            chbAll->setChecked(FALSE);
}

//****************************************************************************

void EventLog::statusSelected()
{
    // This slot is called when status is selected.  If all is selected
    // (checked) it will be cleared

    // if status is checked, and if all is checked, all is cleared
    if (chbStatus->isChecked() == TRUE)
        if (chbAll->isChecked() == TRUE)
            chbAll->setChecked(FALSE);
}

//****************************************************************************

void EventLog::print()
{
    // This function is used to print the event log to a printer
    const int margin = 10;
    int pageNo = 1;         // start at page 1
    int yPos = 0;           // y position for each line

    int column;
    QString strPrinter;     // selected printer
    QPainter p;             // object to paint to printer

    if (printer == NULL)
        printer = new QPrinter(); // printer object

    printer->setPageSize(QPrinter::Letter);
    printer->setOrientation(QPrinter::Landscape);
    printer->setColorMode(QPrinter::GrayScale);

    // get default printer
    strPrinter = getenv( "LPDEST");
    if (!strPrinter.isEmpty())
        printer->setPrinterName(strPrinter);

    QPrintDialog printDialog(printer, this);
    if (printDialog.exec() == QDialog::Accepted)
    {
        // if operator doesn't select OK to continue, exit without printing
        if (!p.begin(printer))
            return;                            

        p.setFont(QFont("courier", 8));
        QFontMetrics fm = p.fontMetrics();

        // clear str
        strTemp.truncate(0);

        // print header
        strTemp.insert(0, "\t\t\t\t\tLPGS EVENT LOG PRINTOUT");

        p.drawText(margin, margin + yPos, printer->width(),
            fm.lineSpacing(), Qt::TextExpandTabs | Qt::TextDontClip, strTemp);
        yPos += 24;     

        // clear str 
        strTemp.truncate(0);        

        // print out the current date and time
        strTemp.insert(0,lblDateTime->text());
        strTemp.append("    ");
        strTemp.append(lblNumberReturned->text());
        strTemp.append("    ");
        strTemp.append(dr->toString());
        p.drawText(margin, margin + yPos, printer->width(), 
            fm.lineSpacing(), Qt::TextExpandTabs | Qt::TextDontClip, strTemp);
        yPos += 24;

        strTemp.truncate(0); // clear str 

        // if any items are selected, print them
        if (lwSelectedList->count() > 0)
        {
            strTemp.insert(0,"Selected Options");

            // print out selected column title
            column = getSelectedColumn(); 

            strTemp.append(" Selected :  ");
            strTemp.append(elInfo[column].strColumnTitle);
            strTemp.append("     ");
            strTemp.append(lblNumberSelected->text());
            strTemp.append("     Message Types:  ");

            // print out the type of messages they selected 
            if (chbAll->isChecked() == TRUE) 
                strTemp.append("All\n");
            else
            {
                if (chbStatus->isChecked() == TRUE)
                    strTemp.append("Status\n");
                if (chbWarnings->isChecked() == TRUE)
                    strTemp.append("Warnings\n");
                if (chbErrors->isChecked() == TRUE)
                    strTemp.append("Errors\n");
            }

            p.drawText(margin, margin + yPos, printer->width(),
                fm.lineSpacing(), Qt::TextExpandTabs | Qt::TextDontClip, 
                strTemp);
            yPos += 24;     

            // print out items operator selected (specific product request ids,
            // task ids,  etc.)
            for (int item = 0; item < lwSelectedList->count(); item++)
            {
                strTemp.truncate(0);
                strTemp.insert(0, "                             ");
                strTemp.append(lwSelectedList->item(item)->text());
                p.drawText(margin, margin + yPos, printer->width(),
                    fm.lineSpacing(), Qt::TextExpandTabs | Qt::TextDontClip, 
                    strTemp);
                yPos += 12;     
            }
        } 
        else
        {
            strTemp.append("Message Types:  ");
        
            // print out the type of messages they selected 
            if (chbAll->isChecked() == TRUE) 
                strTemp.append("All\n");
            else
            {
                if (chbStatus->isChecked() == TRUE)
                    strTemp.append("Status\n");
                if (chbWarnings->isChecked() == TRUE)
                    strTemp.append("Warnings\n");
                if (chbErrors->isChecked() == TRUE)
                    strTemp.append("Errors\n");
            }
            
            p.drawText(margin, margin + yPos, printer->width(),
                fm.lineSpacing(), Qt::TextExpandTabs | Qt::TextDontClip, 
                strTemp);
        }
        yPos += 12; 
        
        // print out detailed messages
        strTemp = "Detailed Message:  \n" + leDetailedMsg->text();
        p.drawText(margin, margin + yPos, printer->width(),
            fm.lineSpacing(), Qt::TextExpandTabs | Qt::TextDontClip, strTemp);
        
        yPos += 36;  

        // print out event tree widget
        vector<twInfo> elInfoVector;
        for (int i = 0; i < numColumns; i++)
            elInfoVector.push_back(elInfo[i]);
        
        TreeWidgetUtilities::printTreeWidget(twEventList, p, printer, margin, 
            yPos, pageNo, elInfoVector, numColumns);
    } 
}

//****************************************************************************

void EventLog::getEvents()
{
    // This function is used to query the database and get the specified 
    // events for display.  fillEventList() is called to fill the events
    // listview

    if (dr->isValid() == FALSE)
        return;

    // build SQL comand 
    strSQLCommand = "SELECT ";

    // loop through the columns and get the database column names to append
    // to SQL command
    for (int column = 0; column < numColumns; column++) 
    {
        strSQLCommand += elInfo[column].strDbColumn;
        // if this is the last column, do not add a command after the column
        if(column == numColumns - 1)
            strSQLCommand += " ";
        else
            strSQLCommand += ", ";
    }

    strSQLCommand += " FROM events, messages "
                     " WHERE " + buildWhereClause()
                     + " AND events.MESSAGE_ID = messages.MESSAGE_ID";
    // execute Query
    if (eventsQuery.exec(strSQLCommand, __FILE__, __LINE__))
        // fill events list view with data from the RecordSet
        fillEventList();
    else
        twEventList->clear();
}

//*****************************************************************************

void EventLog::getSelectItems()
{
    // This function builds the select list items, it queries the database
    // based on values selected and the column selected and returns a list
    // of select list items used to fill a combo box which can be used to
    // limit the number of items returned in the database

    IASSqlQuery query;

    if (dr->isValid() == FALSE)
        return;

    // clear the current contents from the select list
    cobValueSelect->clear();
    
    QString dbColumn = elInfo[getSelectedColumn()].strDbColumn;
    // create SQL command using the currently selected column
    if (chbAll->isChecked())
    {
        strSQLCommand = "SELECT DISTINCT " +  dbColumn
            + " FROM events"
            + " WHERE" + buildWhereClause();
    }
    else // Need to join the messages table to get message severity
    {
        strSQLCommand = "SELECT DISTINCT " +  dbColumn
            + " FROM events, messages"
            + " WHERE events.message_id = messages.message_id"
                + " AND " + buildWhereClause();
    }
    strSQLCommand += " ORDER BY " + dbColumn;

    // query database
    if (query.exec(strSQLCommand, __FILE__, __LINE__))
    {
        /* insert the results into the combo box */
        while (query.next())
        {
            cobValueSelect->addItem(query.value(0).toString());
        }
    }

    // ensure no items are selected
    cobValueSelect->setCurrentIndex(-1);
}

//****************************************************************************

void EventLog::reselectLastItemSelected()
{
    // This function searches through items in list, and finds the item that 
    // was selected before refresh was pressed, and ensures that the same item
    // is selected again.  If the previous item is not found, i.e. no longer
    // among the returned list, due to different selections, then the first
    // child in the list is selected, and the current selected sequence number
    // is updated.
    
    bool found = false;
    
    // if the selected sequence number is not empty
    if (!strSelectedSequenceNumber.isEmpty())
    {
        // loop through all items in list and find the one containing the 
        // previously selected sequence number
        QTreeWidgetItemIterator it(twEventList);
        while (*it)
        {
            if ((*it)->text(nSEQ_NUM) == strSelectedSequenceNumber)
            {
                found = true;
                (*it)->setSelected(true);
                break;
            }
            ++it;
        }
    }

    // if the item was not found, or one has not been selected, set the
    // current item to the first item in the list
    if (found == false)
    {
        if (twEventList->topLevelItemCount() > 0)
        {
            twEventList->setCurrentItem(twEventList->topLevelItem(0));
        }
    }
}

//****************************************************************************

QString EventLog::buildWhereClause()
{
    // This functions builds a where clause based on operator selections.  The
    // items include type of messages, (errors, warnings, status), starting and
    // ending dates, and selected column values (product reques id, task name,
    // script id, work order id) for use in database queries

    QString strWhereClause(""); // storage for where clause

    if (dr->isValid() == FALSE)
        return strWhereClause;

    strWhereClause = " EVENT_DATE " + dr->getDateQueryString();

    if (lwSelectedList->count() > 0)
    {  
        for (int item = 0; item < int(lwSelectedList->count()); item++)
        {  
            if (item == 0)
            {
                if (lwSelectedList->count() > 1)
                    strWhereClause += " AND ( ";
                else
                    strWhereClause += " AND ";
            }
            else
            {
                strWhereClause += " OR ";
            }
            strWhereClause += elInfo[getSelectedColumn()].strDbColumn;
            strWhereClause += " = '";
            strWhereClause += lwSelectedList->item(item)->text();
            strWhereClause += "'";
        }

        if (lwSelectedList->count() > 1)
            strWhereClause += " ) ";
    }

    // if all is not selected, 
    //     must generate SQL statement to return only the message types selected
    if (!chbAll->isChecked())
    {
        int msgTypes = 0;   
        // number of messages types selected (warnings, errors, status)
        bool start = FALSE; 

        // determine how many message types to return
        if (chbErrors->isChecked()) 
            msgTypes++;
        
        if (chbWarnings ->isChecked()) 
            msgTypes++;
        
        if (chbStatus->isChecked()) 
            msgTypes++;
        
        if (msgTypes > 0)
            strWhereClause += " AND ";
        
        if (chbErrors->isChecked())
        {
            if (msgTypes > 1)
                strWhereClause +="( ";
             
            strWhereClause += " SEVERITY = 'E' ";
            start = TRUE;
        }
        
        if (chbWarnings->isChecked())
        {
            if (msgTypes > 1)
            {
                if (start == TRUE)
                    strWhereClause += " OR ";
                else
                {
                    strWhereClause += " ( ";
                    start = TRUE;
                }
            }
            strWhereClause += " SEVERITY = 'W' ";
        }
        
        if (chbStatus->isChecked())
        {
            if (msgTypes > 1)
                strWhereClause += " OR ";
            
            strWhereClause += "SEVERITY = 'I' ";
        }
    
        if (msgTypes > 1)
             strWhereClause += " ) ";
    } // if all is checked, no changes required

    return strWhereClause;
}

//****************************************************************************

void EventLog::eventSelectionChanged()
{
    // This function updates the detailed message to the event comment from
    // for the selected item in the event list.  

    QTreeWidgetItem *selectedItem = twEventList->currentItem();
    if (selectedItem != NULL)
    {
        IASSqlQuery query;
        strSQLCommand = "SELECT event_comment FROM events WHERE seq_nbr = '"
            + selectedItem->text(nSEQ_NUM) + "'";
        leDetailedMsg->setText("");
        // query database
        if (query.exec(strSQLCommand, __FILE__, __LINE__))
        {
            // check to see if query returned results
            if (!query.next())
                return;
            // set detailed message
            leDetailedMsg->setText(query.value("event_comment").toString());
        }
    }
}

//****************************************************************************

void EventLog::eventOperatorSelected(QTreeWidgetItem *twiSelectedItem)
{
    // This function keeps track of the QTreeWidgetItem that the operator
    // selected.  It only tracks items selected by the operator, not by
    // the log itself.  When the operator has not selected an event in the
    // listview, the last item in the listview will be selected, or when the
    // last operator selected item is not found in the list, the last item 
    // will be selected.
    
    // if the list is not empty
    if (twEventList->topLevelItemCount() > 0)
    {
        // save sequence number of currently selected item
        if (twiSelectedItem != NULL)
            strSelectedSequenceNumber = twiSelectedItem->text(nSEQ_NUM);
    }
}

//****************************************************************************

void EventLog::queryOnItem()
{
    // This function adds the item that is selected in event list to list
    // of the selected items and performs a query on the currently selected
    // items. This slot is connected to the event list ListViewItem 
    // double_click() slot.  When the operator double_clicks on an item
    // in event listview, it will add the data from the selected column of
    // of selected item (if column not empty) and add it to the list of
    // items selected to query, and then execute a query.
    
    // If the selected listviewitem contains data, continue
    if (getSelectedColumnText().length() > 0)
    {
        // add selected item to selected items to query on
        addItem();

        // if item was added, refresh the list
        if (lwSelectedList->count() > 0)
            refreshList();
    }
}

//****************************************************************************

void  EventLog::swapEventListColors()
{
    // This function causes the background and text colors to be swapped in
    // the Event List View.  The list will be refreshed with new colors
    
    // invert the value of swapColors
    swapColors = !swapColors;
    // clear current list of events    
    fillEventList();
}
//*****************************************************************************

void  EventLog::fillEventList()
{
    // This function clears the event list and fills it with the data in the
    // events RecordSet.

    QColor qcForeground; // Foreground color of list item
    QColor qcBackground; // Background color of list item
    ColoredTreeWidgetItem *listViewItem;

    // clear events list view
    twEventList->clear();

    // Force the position to before the first record since we may not have
    // just executed the query.
    eventsQuery.seek(-1);
    while (eventsQuery.next())
    {
        // get message severity and ensure that it is in
        // upper case for comparison
        QString strSeverity = eventsQuery.value(nSEV).toString();
            
        // choose text color based on message severity
        // "I"  status (Green)
        // "W"  warning (Yello) 
        // "E"  error (Red)
        if (strSeverity.compare(QString("I")) == 0)
            qcForeground =  Qt::darkGreen;
        else if (strSeverity.compare(QString("W")) == 0)
            qcForeground = Qt::darkYellow;
        else if (strSeverity.compare(QString("E")) == 0)
            qcForeground = Qt::red;


        // if swap colors is true, swap text and background colors
        if (swapColors == TRUE)
        {
            qcBackground = qcForeground;
            qcForeground = Qt::white;
        }
        else
        {
            // set background color (white)
            qcBackground  = Qt::white;
        }

        listViewItem = new ColoredTreeWidgetItem(qcForeground, qcBackground,
                twEventList);

        for (int columnPos = 0; columnPos < numColumns; columnPos++)
        {
            if (columnPos == nSEQ_NUM)
                // Add in this manner to afford numeric sorting on the sequence
                // column.
                listViewItem->setData(columnPos, Qt::DisplayRole,
                    eventsQuery.value(columnPos).toInt());
            else
                listViewItem->setText(columnPos,
                    eventsQuery.value(columnPos).toString());
        }
        listViewItem->setColorProperties(qcForeground, qcBackground);

        // set text for date to a value that can be correctly sorted 
        // alphabetically using the build in sort in Qt list view
        listViewItem->setColoredText(nDATE_TIME,
            DatabaseUtils::convertDateforAlphaSort(
                listViewItem->text(nDATE_TIME)));
    }
}

