/******************************************************************************
               Level 1 Product Generation System (LPGS) Software
 ******************************************************************************
 *
 * UNIT NAME:   FileViewer.cpp
 *
 * PURPOSE:
 * This is the implementation of the FileViewer class.  This QDialog-based
 * class implements a window that allows the operator to view the contents
 * of order ODL files, work order logs, etc.
 *
 *****************************************************************************/

/* System Includes */
#include <iostream>

/* Qt Includes */
#include <QtCore>
#include <QtGui>

/* Local Includes */
#include "FileViewer.h"

//*****************************************************************************

FileViewer::FileViewer(
    QString strFilenameIn,  // I: Name of file to view in widget
    QString startDir,       // I: Directory to start file searches from
    bool autoRefresh,       // I: Automatically refresh file content?
    int refreshInterval,    // I: Number of seconds between auto refreshes
    QWidget* parent,        // I: Parent widget
    Qt::WindowFlags fl) 
  : QDialog(parent, fl)
{
    setupUi(this);

    strFilename_ = strFilenameIn;
    startDir_ = startDir;

    chbAutoRefresh->setChecked(autoRefresh);

    // Set up the widgets needed to implement the automatic refresh.
    // The default interval will be 10 seconds.
    refreshTimer_ = new QTimer(this);
    QString strInterval;
    strInterval.sprintf("%d", refreshInterval);
    leAutoRefresh->setText(strInterval);

    // Set up all of the required connections for the File Viewer dialog.
    connect(pbRefresh, SIGNAL(clicked()), this, SLOT(refresh()));
    connect(pbFind, SIGNAL(clicked()), this, SLOT(findForward()));
    connect(pbBack, SIGNAL(clicked()), this, SLOT(findBackward()));
    connect(pbDismiss, SIGNAL(clicked()), this, SLOT(accept()));
    connect(pbFileOpen, SIGNAL(clicked()), this, SLOT( fileOpen()));
    connect(leFind, SIGNAL(returnPressed()), this, SLOT(findForward()));
    connect(chbAutoRefresh, SIGNAL(toggled(bool)),
         this, SLOT(autoRefreshToggled(bool)));
    connect(refreshTimer_, SIGNAL(timeout()), this, SLOT(timerRefresh()));

    // Call refresh() to initially populate the contents of the window,
    // and position the cursor at the beginning of the file.  If the
    // dialog was constructed to automatically refresh by default, set
    // up the timer for that operation as well.
    refresh();

    if (autoRefresh)
        refreshTimer_->start(refreshInterval * 1000);

    teFileViewer->moveCursor(QTextCursor::Start);
}
//*****************************************************************************

void FileViewer::timerRefresh()
{

    refreshTimer_->stop();
    refresh();
    // After the refresh is complete, reinstate the timer.
    bool ok;
    int interval = leAutoRefresh->text().toInt(&ok);
    if (ok)
        refreshTimer_->start(interval * 1000);
    else
        refreshTimer_->start(); // If invalid, use current interval setting.
}

//*****************************************************************************

void FileViewer::refresh()
{
    // Record the current vertical position of the scrollbar so that
    // we can set it back to the same value after we do the refresh.
    // This yields a "smoother" refresh.
    int scrollbarPos = teFileViewer->verticalScrollBar()->value();

    // Check and see if the vertical scrollbar is currently positioned
    // at the very bottom of the listview.  This is so that we can
    // place it back there after the refresh in the event that the
    // file's contents have changed (i.e., it has been added to) .. if
    // that happens, the scrollbar position previously noted may no
    // longer be valid.
    bool maxxed = (teFileViewer->verticalScrollBar()->value()
                   == teFileViewer->verticalScrollBar()->maximum()) 
               && (teFileViewer->verticalScrollBar()->maximum() != 0);

    // Open the specified file, read it into a QTextStream object, and
    // dump it from the text stream into the QTextEdit widget all at
    // once.  This is more efficient than doing it line-by-line.
    QFile file(strFilename_);
    if (file.open(QIODevice::ReadOnly)) 
    {
        QTextStream ts(&file);
        teFileViewer->setText(ts.readAll());
    }
    file.close();

    // If the scrollbar was positioned at the bottom of the list before
    // the refresh, go to the bottom of the list now.  Otherwise, just
    // reset its position value to the previously-recorded value.
    teFileViewer->moveCursor(QTextCursor::End);
    if (maxxed)
        scrollbarPos = teFileViewer->verticalScrollBar()->maximum();

    teFileViewer->verticalScrollBar()->setValue(scrollbarPos);
}

//*****************************************************************************

void FileViewer::fileOpen()
{
    // Pop open a QFileDialog so that the operator can interactively
    // choose a file to view.
    QString strFilename = QFileDialog::getOpenFileName(this, "Choose a file "
        "to view", startDir_, QString(), 0, QFileDialog::DontUseNativeDialog);

    // If a valid filename was selected (in other words, the operator did
    // not hit the Cancel button), replace the existing filename with
    // the new one and refresh the dialog to update the contents.
    if (!strFilename.isEmpty()) 
    {
        strFilename_ = strFilename;
        teFileViewer->verticalScrollBar()->setValue(0);
        refresh();
        teFileViewer->moveCursor(QTextCursor::Start);
    }
}

//*****************************************************************************

void FileViewer::autoRefreshToggled(bool on)
{
    if (on) 
    {
        // If the operator turned on the Auto Refresh feature, retrieve
        // the number of seconds from the line entry widget and start
        // up the QTimer object, using that number as the refresh interval.
        // Note that the argument to start() is in milliseconds.
        bool ok;
        int interval = leAutoRefresh->text().toInt(&ok);
        if (ok)
            refreshTimer_->start(interval * 1000);
        else
            refreshTimer_->start(); // If invalid, use current interval setting.
    }
    else
    {
        // Otherwise, stop the (presumably running) timer.
        refreshTimer_->stop();
    }
}

//*****************************************************************************

void FileViewer::findForward()
{
    // Use QTextEdit's built-in find() function in order to search
    // forward for the selected text in the file.
    QFlags<QTextDocument::FindFlag> flags;

    // Check for Case Sensitivity
    if (chbCaseSensitive->isChecked())
        flags |= QTextDocument::FindCaseSensitively;
    
    if (teFileViewer->find(leFind->text(), flags) == false) 
    {
        if (QMessageBox::information(this, "Search String Not Found",
                "The specified search string was not found.  Would you\n"
                "like to restart the search from the top of the file?",
                QMessageBox::Yes | QMessageBox::No) == QMessageBox::Yes)
        {
            // Move the cursor to the beginning of the document
            teFileViewer->moveCursor(QTextCursor::Start);

            findForward();
            return;
        }
    }
}

//*****************************************************************************

void FileViewer::findBackward()
{
    // Use QTextEdit's built-in find() function in order to search
    // backward for the selected text in the file.
    QFlags<QTextDocument::FindFlag> flags;

    // Find Backwards
    flags |= QTextDocument::FindBackward;

    // Check for Case Sensitivity
    if (chbCaseSensitive->isChecked())
        flags |= QTextDocument::FindCaseSensitively;

    if (teFileViewer->find(leFind->text(), flags) == false) 
    {
        if (QMessageBox::information(this, "Search String Not Found",
                "The specified search string was not found.  Would you\n"
                "like to restart the search from the bottom of the file?",
                QMessageBox::Yes | QMessageBox::No) == QMessageBox::Yes)
        {
            // Move the cursor to the end of the document
            teFileViewer->moveCursor(QTextCursor::End);

            findBackward();
            return;
        }
    }
}
