/******************************************************************************
               Level 1 Product Generation System (LPGS) Software
 ******************************************************************************
 *
 * UNIT NAME:   GUIConfiguration.cpp
 *
 * PURPOSE:
 * This dialog allows the user to view and modify the GUI configuration
 * parameters. Any changes made are saved to the lpgsuirc configuration file.
 *
 *****************************************************************************/

/* System Includes */
#include <iostream>

/* Qt Includes */
#include <QtCore>
#include <QtGui>

/* Local Includes */
#include "GUIConfiguration.h"
#include "Preferences.h"
#include "Utilities.h"
#include "IASUtilities.h"

extern Preferences *preferences;

static QString strFontList[] = 
{
    "6x10",
    "6x12",
    "6x13",
    "6x13bold",
    "6x9",
    "7x13",
    "7x13bold",
    "7x14",
    "7x14bold",
    "8x13",
    "8x13bold",
    "8x16",
    "9x15",
    "9x15bold",
    "10x20",
    "12x24",
};

static struct 
{
    QString strColorScheme;
    QString strColorSwitch;
} schemeList[] = 
{
    { "White on Black",     "-fg white -bg black" },
    { "Black on White",     "-fg black -bg white" },
    { "Green on Black",     "-fg green -bg black" },
    { "Jackrabbit",         "-fg yellow -bg blue4" },
    { "Husker",             "-fg white -bg red4" },
    { "Packer",             "-fg yellow -bg green4" },
    { "Viking",             "-fg yellow -bg purple4" },
};

GUIConfiguration::GUIConfiguration(QWidget *parent)
: QDialog(parent)
{
    setupUi(this);

    // The first thing that we need to do is make our own local copy
    // of the global Preferences object.  Since we'll be screwing around
    // with the values in it, and since we don't want any changes to
    // take effect until the operator actually hits Apply or OK, we need
    // to have someplace to store them until they're ready.
    localPreferences = new Preferences();

    // Insert the group and filter information into their associated
    // combo box, and call groupChanged() and filterChanged() in order
    // to select the first item in the list and populate the list
    // boxes accordingly.
    QStringListIterator stateIterator(stateList());
    while (stateIterator.hasNext())
    {
        QString stateText = stateIterator.next();
        lbStates->addItem(stateText);
        lbAvailStates->addItem(stateText);
    }

    QStringListIterator groupIterator(localPreferences->groupNames());
    while (groupIterator.hasNext())
    {
        cboGroups->addItem(groupIterator.next());
    }
    cboGroups->setCurrentIndex(0);
    groupChanged(cboGroups->currentText());

    QStringListIterator filterIterator(localPreferences->filterNames());
    while (filterIterator.hasNext())
    {
        cboFilters->addItem(filterIterator.next());
    }
    cboFilters->setCurrentIndex(0);
    filterChanged(cboFilters->currentText());

    // If there is only one filter, the operator should NOT be able to delete it
    if(cboFilters->count() > 1)
    {
        pbDeleteFilter->setEnabled(TRUE);
    }
    else
    {
        pbDeleteFilter->setEnabled(FALSE);
    }

    chbEnableColors->setChecked(localPreferences->areColorsEnabled());

    // Populate the timer and duration QLineEdits.
    leRefreshTimer->setText(localPreferences->strRefreshTimer);
    lePRDuration->setText(localPreferences->strPRDuration);
    leEventLogDuration->setText(localPreferences->strEventLogDuration);

    // For each toolbar, set the state of the checkbox accordingly,
    // depending on whether or not the preferences indicate that the
    // operator wants to be shown it by default.
    chbCommonTasks->setChecked(localPreferences->commonTasksVisible);
    chbFindRequest->setChecked(
        localPreferences->findRequestVisible);
    chbStateFilters->setChecked(localPreferences->stateFiltersVisible);
    chbDateRange->setChecked(localPreferences->dateRangeVisible);
    chbInterfaceStatuses->setChecked(localPreferences->interfaceStatusesVisible);

    // Insert all of the above-listed font names into the font combo box.
    for (unsigned int i = 0; i < ARRAY_SIZE(strFontList); ++i)
    {
        cboFont->addItem(strFontList[i]);
        if (strFontList[i] == preferences->strFont)
            cboFont->setCurrentIndex(i);
    }

    // Insert all of the above-listed color scheme names into the
    // combo box for the color schemes.
    for (unsigned int scheme = 0; scheme < ARRAY_SIZE(schemeList); ++scheme)
    {
        cboColorScheme->addItem(schemeList[scheme].strColorScheme);
        if (schemeList[scheme].strColorScheme == preferences->strColorScheme)
            cboColorScheme->setCurrentIndex(scheme);
    }

    leSize->setText(preferences->strSize);
    leScrollbackLines->setText(preferences->strScrollbackLines);

    QString strWidth;
    strWidth.setNum(preferences->width);
    leWidth->setText(strWidth);

    QString strHeight;
    strHeight.setNum(preferences->height);
    leHeight->setText(strHeight);

    // Setup connections
    connect(pbOk, SIGNAL(clicked()), this, SLOT(accept()));
    connect(pbCancel, SIGNAL(clicked()), this, SLOT(reject()));
    connect(pbBackgroundColor, SIGNAL(clicked()), 
        this, SLOT(selectBackgroundColor()));
    connect(pbTextColor, SIGNAL(clicked()), this, SLOT(selectTextColor()));
    connect(pbRemoveGroup, SIGNAL(clicked()), 
        this, SLOT(removeStateFromList()));
    connect(pbAddGroup, SIGNAL(clicked()), this, SLOT(addStateToList()));
    connect(chbEnableColors, SIGNAL(stateChanged(int)), 
        this, SLOT(enableColors(int)));
    connect(pbNewGroup, SIGNAL(clicked()), this, SLOT(addNewGroup()));
    connect(pbNewFilter, SIGNAL(clicked()), this, SLOT(addNewFilter()));
    connect(pbAddStateToFilter, SIGNAL(clicked()),
        this, SLOT(addStateToFilterList()));
    connect(pbRemoveStateFromFilter, SIGNAL(clicked()),
        this, SLOT(removeStateFromFilterList()));

    connect(cboGroups, SIGNAL(currentIndexChanged(const QString&)), 
        this, SLOT(groupChanged(const QString&)));
    connect(cboFilters, SIGNAL(currentIndexChanged(const QString&)),
        this, SLOT(filterChanged(const QString&)));

    connect(pbApply, SIGNAL(clicked()), this, SLOT(applyChanges()));
    connect(pbUseCurrent, SIGNAL(clicked()), this, SLOT(setCurrentGeometry()));
    connect(pbDeleteGroup, SIGNAL(clicked()), this, SLOT(deleteGroup()));
    connect(pbDeleteFilter, SIGNAL(clicked()), this, SLOT(deleteFilter()));
}

GUIConfiguration::~GUIConfiguration()
{
    delete localPreferences;
}

/*****************************************************************************/

void GUIConfiguration::enableColors(int flag)
{
    localPreferences->setColorsEnabled(flag);
}

/*****************************************************************************/

void GUIConfiguration::addNewGroup()
{
    // This function allows the operator to enter the name of a new
    // state group.  It prompts for the name via QInputDialog,
    // inserts the new name into the combo box, and informs the local
    // Preferences object about the new group.
    bool entryOk;
    QString groupName = QInputDialog::getText(this, "New Group",
        "Enter name of new group:", QLineEdit::Normal, "", &entryOk);

    if (!entryOk)
        return;

    groupName = groupName.trimmed();

    if(localPreferences->groupNames().indexOf(groupName) >= 0 )
    {
        QMessageBox::warning(this, "Duplicate Group", 
            "There is already a group called \"" + groupName
            + "\".\nPlease enter a different name.");
        addNewGroup();
        return;
    }
    else if(groupName.isNull() || groupName.isEmpty())
    {
        QMessageBox::warning(this, "No Name Entered", 
            "You must enter a name for the new group." 
            "\nPlease enter a new group name.");
        addNewGroup();
        return;
    }

    cboGroups->addItem(groupName);
    cboGroups->setCurrentIndex(cboGroups->findText(groupName));

    localPreferences->addGroup(groupName);
    localPreferences->backgroundColor[groupName] = QColor(255, 255, 255);
    localPreferences->textColor[groupName] = QColor(0, 0, 0);
}

/*****************************************************************************/

void GUIConfiguration::deleteGroup()
{
    QString strGroupToDelete = cboGroups->currentText();
    if (QMessageBox::warning(this, "Delete Group",
            "Are you sure you want to delete the Group \""
            + strGroupToDelete + "\"?",
            QMessageBox::Yes,QMessageBox::No)==QMessageBox::Yes)
    {
        localPreferences->deleteGroup(strGroupToDelete);
        cboGroups->removeItem(cboGroups->currentIndex());
        cboGroups->setCurrentIndex(0);
        groupChanged(cboGroups->currentText());
    }
}

/*****************************************************************************/

void GUIConfiguration::addNewFilter()
{
    // This function allows the operator to enter the name of a new
    // state filter.  It prompts for the name via QInputDialog,
    // inserts the new name into the combo box, and informs the local
    // Preferences object about the new filter.
    bool entryOk;
    QString strFilterName = QInputDialog::getText(this, "New Filter",
        "Enter name of new filter:", QLineEdit::Normal, "", &entryOk);

    if (!entryOk)
        return;

    strFilterName = strFilterName.trimmed();

    if (localPreferences->filterNames().indexOf(strFilterName) >= 0)
    {
        QMessageBox::warning(this, "Duplicate Filter", 
            "There is already a filter called \"" + strFilterName
            + "\".\nPlease enter a different name.");
        
        addNewFilter();
        return;
    }
    else if (strFilterName.isNull() || strFilterName.isEmpty())
    {
        QMessageBox::warning(this, "No Name Entered", 
            "You must enter a name for the new filter" 
            "\nPlease enter a new filter name.");
        
        addNewFilter();
        return;
    }

    cboFilters->addItem(strFilterName);
    cboFilters->setCurrentIndex(cboFilters->findText(strFilterName));

    if (cboFilters->count() > 1)
        pbDeleteFilter->setEnabled(TRUE);

    localPreferences->addFilter(strFilterName);
}

/*****************************************************************************/

void GUIConfiguration::deleteFilter()
{
    // This functions deletes the filter that is currently selected from
    // the local preferences object.
    QString strFilterToDelete = cboFilters->currentText();
    
    if (QMessageBox::warning(this, "Delete Filter",
        "Are you sure you want to delete the filter \""
        + strFilterToDelete + "\"?",
        QMessageBox::Yes,QMessageBox::No) == QMessageBox::Yes)
    {
        //Delete from local preferences
        localPreferences->deleteFilter(strFilterToDelete);
    
        //Remove from combo box
        cboFilters->removeItem(cboFilters->currentIndex());

        //Set the current item to the first in the list and make sure the other
        //gui components are updated with the new selection.
        cboFilters->setCurrentIndex(0);
        if (cboFilters->count() < 2)
            pbDeleteFilter->setEnabled(FALSE);

        filterChanged(cboFilters->currentText());
    }
}

/*****************************************************************************/

void GUIConfiguration::addStateToList()
{
    // Add the currently-selected product request state to
    // the list of states in the current group.  (Do nothing
    // if there is no state selected, or if the selected
    // state is already in the group list.

    if (lbStates->currentRow() < 0)
        return; // Nothing is selected!

    QListWidgetItem *item = lbStates->currentItem();

    QList<QListWidgetItem*> matches = lbGroups->findItems(item->text(),
                                                          Qt::MatchExactly);
    if (!matches.isEmpty())
        return; // Item is already in list!

    lbGroups->addItem(item->text());

    localPreferences->group[stringToState(item->text())] =
        cboGroups->currentText();
}

/*****************************************************************************/

void GUIConfiguration::addStateToFilterList()
{
    // This is called when the operator clicks the right-arrow button to add
    // a Product Request state to the group of states associated with a 
    // current filter.  It starts by performing the same checks that are
    // performed above (is anything actually selected in the left-hand list,
    // and is it already a member of this group?)

    if (lbAvailStates->currentRow() < 0)
        return; // Nothing is selected!

    QListWidgetItem *item = lbAvailStates->currentItem();
    
    QList<QListWidgetItem*> matches = lbFilters->findItems(item->text(),
                                                           Qt::MatchExactly);
    if (!matches.isEmpty())
        return; // Item is already in list!

    // So far, so good .. insert the selected state into the group.
    lbFilters->addItem(item->text());

    // Now we've got to do a little bit of work.  The filter states are
    // stored internally in the LPGS Preferences object as comma-separated
    // lists of state numbers.  For example, if I've got a "Visual QA"
    // group that consists of the "L1R QA", "L1G QA", and "Final QA"
    // states, that list is stored (internally) as "9,10,11".
    //
    // So what we've got to do is loop through each state that is in the
    // group for this filter, look up its state number, and then finally
    // create a comma-separated list of them and store that as the
    // filterStates string for this particular filter.
    QStringList stateList;
    for (int currState = 0; currState < int(lbFilters->count()); ++currState) 
    {
        stateList += stringToState(lbFilters->item(currState)->text());
    }

    localPreferences->setFilterStates(cboFilters->currentText(),
                                      stateList.join(","));
}

/*****************************************************************************/

void GUIConfiguration::removeStateFromList()
{
    // Remove the currently-selected product request state
    // from the list of states in the current group.  (Do
    // nothing if no state is actually selected.
    if (lbGroups->currentRow() < 0)
        return;

    QListWidgetItem *item = lbGroups->currentItem();

    localPreferences->group[stringToState(item->text())] = DEFAULT_GROUP;

    // takeItem will remove the item from the groups list. The item is no
    // longer managed by Qt and must be explicitly deleted
    delete lbGroups->takeItem(lbGroups->currentRow());
}

/*****************************************************************************/

void GUIConfiguration::removeStateFromFilterList()
{
    // Pretty much the same story; make sure that a state is actually
    // selected and if it is, remove it from the list.

    if (lbFilters->currentRow() < 0)
        return;
    
    // takeItem will remove the item from the groups list. The item is no
    // longer managed by Qt and must be explicitly deleted
    delete lbFilters->takeItem(lbFilters->currentRow());

    // Now we have to do the exact same thing that was done in the
    // addStateToFilterList() function; namely, we have to assemble a
    // comma-separated list of the items that remain and then use
    // that to reset the state list in the Preferences object.
    QStringList stateList;
    for (int currState = 0; currState < int(lbFilters->count()); ++currState)
    {
        stateList += stringToState(lbFilters->item(currState)->text());
    }

    localPreferences->setFilterStates(cboFilters->currentText(),
                                      stateList.join(","));
}

/*****************************************************************************/

void GUIConfiguration::groupChanged(const QString &group)
{
    // This function is called when the operator selects a new group
    // name from the combo box.  It clears the current contents of
    // the Group States listbox and repopulates it with the appropriate
    // states for the new group (as retrieved from the preferences.)

    lbGroups->clear();

    QStringListIterator stateIterator(stateList());
    while (stateIterator.hasNext())
    {
        QString stateText = stateIterator.next();
        QString stateValue = stringToState(stateText);
        if (localPreferences->group[stateValue] == group)
            lbGroups->addItem(stateText);
    }

    QPalette palette = frameBackgroundColor->palette();
    palette.setColor(backgroundRole(), 
        localPreferences->backgroundColor[group]);
    frameBackgroundColor->setPalette(palette);
    frameBackgroundColor->setAutoFillBackground(true);

    palette = frameTextColor->palette();
    palette.setColor(backgroundRole(), 
        localPreferences->textColor[group]);
    frameTextColor->setPalette(palette);
    frameTextColor->setAutoFillBackground(true);

    if (group == DEFAULT_GROUP)
        pbDeleteGroup->setEnabled(false);
    else
        pbDeleteGroup->setEnabled(true);
}

/*****************************************************************************/

void GUIConfiguration::filterChanged(const QString &filter)
{
    // This function is called when the operator selects a new filter
    // name from the combo box.  It clears the current contents of
    // the Filter States listbox and repopulates it with the appropriate
    // states for the new filter (as retrieved from the preferences.)
    lbFilters->clear();

    QString strStates = localPreferences->filterStates(filter);
    strStates.remove("'"); // Remove the single quoates

    QStringList stateList = strStates.split(",", QString::SkipEmptyParts);
    QStringListIterator iterator(stateList);
    while (iterator.hasNext())
    {
        lbFilters->addItem(stateToString(iterator.next()));
    }
}

/*****************************************************************************/

void GUIConfiguration::selectBackgroundColor()
{
    // This function opens up a QColorDialog that allows the operator
    // to change the color of the background for the current group.  Once
    // a new color is picked, the color of the box is changed to
    // appropriately reflect the new value.
    QString groupName = cboGroups->currentText();

    QPalette palette = frameBackgroundColor->palette();
    
    QColor color = QColorDialog::getColor(
        palette.color(backgroundRole()), this);

    // If color is invalid, then the user pressed the cancel button
    if (color.isValid())
    {
        palette.setColor(backgroundRole(), color);
        frameBackgroundColor->setPalette(palette);
        frameBackgroundColor->setAutoFillBackground(true);
    
        localPreferences->backgroundColor[groupName] = color;
    }
}

/*****************************************************************************/

void GUIConfiguration::selectTextColor()
{
    // This function opens up a QColorDialog that allows the operator
    // to change the color of the text for the current group.  Once
    // a new color is picked, the text color of the box is changed
    // to appropriately reflect the new value.
    QString groupName = cboGroups->currentText();

    QPalette palette = frameTextColor->palette();

    QColor color = QColorDialog::getColor(
        palette.color(backgroundRole()), this);

    // If color is invalid, then the user pressed the cancel button
    if (color.isValid())
    {
        palette.setColor(backgroundRole(), color);
        frameTextColor->setPalette(palette);
        frameTextColor->setAutoFillBackground(true);

        localPreferences->textColor[groupName] = color;
    }
}

/*****************************************************************************/

void GUIConfiguration::setCurrentGeometry()
{
    // This function sets the Width and Height line edit values to
    // be the actual current width and height of the parent (main)
    // window, so that the current size will be used in all future
    // invocations of the GUI.
    QString strWidth;
    strWidth.setNum(parentWidget()->width());
    leWidth->setText(strWidth);

    QString strHeight;
    strHeight.setNum(parentWidget()->height());
    leHeight->setText(strHeight);
}

/*****************************************************************************/

void GUIConfiguration::applyChanges()
{
    // The main purpose of this function is to take all of the values
    // from the GUI widgets within the dialog, retrieve their values,
    // and then stuff them back into the local preferences object.
    // Once this is done, the local preferences object is copied over
    // the global ("real") preferences object, at which point the
    // changes should take effect.

    localPreferences->commonTasksVisible = chbCommonTasks->isChecked();
    localPreferences->findRequestVisible =
        chbFindRequest->isChecked();
    localPreferences->stateFiltersVisible = chbStateFilters->isChecked();
    localPreferences->dateRangeVisible = chbDateRange->isChecked();
    localPreferences->interfaceStatusesVisible =
        chbInterfaceStatuses->isChecked();
    localPreferences->strRefreshTimer = leRefreshTimer->text();
    localPreferences->strPRDuration = lePRDuration->text();
    localPreferences->strEventLogDuration = leEventLogDuration->text();

    localPreferences->strFont = cboFont->currentText();
    localPreferences->strSize = leSize->text();
    localPreferences->strScrollbackLines = leScrollbackLines->text();
    localPreferences->strColorScheme = cboColorScheme->currentText();
    localPreferences->strColorSwitch =
        schemeList[cboColorScheme->currentIndex()].strColorSwitch;

    localPreferences->width = leWidth->text().toInt();
    localPreferences->height = leHeight->text().toInt();

    parentWidget()->resize(localPreferences->width, localPreferences->height);

    // As in the constructor for Preferences, put together a string
    // of xterm options to implement the operator's desired behavior.
    localPreferences->strXtermOptions =
        "-font " + localPreferences->strFont +
        " -geometry " + localPreferences->strSize +
        " -sl " + localPreferences->strScrollbackLines +
        " " + localPreferences->strColorSwitch;

    *preferences = *localPreferences;

    // Now write the preference changes to file
    if (!preferences->writePreferencesFile())
    {
        QMessageBox::warning(this, "Failed to write preferences", 
            "There was an error writing the preferences out to the "
            "resource file.");
    }
}

/*****************************************************************************/

void GUIConfiguration::accept()
{
    applyChanges();
    QDialog::accept();
}

