/******************************************************************************
               Level 1 Product Generation System (LPGS) Software
 ******************************************************************************
 *
 * UNIT NAME:   LPGSController.cpp
 *
 * PURPOSE:
 * The LPGSController class provides facilities that the GUI uses to
 * start up, shut down, and query the status of the LPGS system itself.
 *
 *****************************************************************************/

// System Includes
#include <stdlib.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <pwd.h>
#include <unistd.h>
#include <iostream>

// Qt Includes
#include <QtCore>
#include <QtGui>

// Local Includes
#include "IASQtSystemStatus.h"
#include "LPGSController.h"
#include "MainWindow.h"
#include "IASSqlQuery.h"

LPGSController::LPGSController()
{
    IASSqlQuery query;

    shutdownType = LPGS_SHUTDOWN_NONE;

    strSystemId = getenv("L1_SYSTEM_ID");
    if (strSystemId.isEmpty())
    {
        QMessageBox::critical(0, "Missing Environment",
            "Could not get L1_SYSTEM_ID from environment.\n"
            "This variable is required for system startup or shutdown",
            QMessageBox::Ok, 0);
        return;
    }

    strLpgsHome = getenv("LPGSHOME");
    if (strLpgsHome.isEmpty())
    {
        QMessageBox::critical(0, "Missing Environment",
            "Could not get LPGSHOME from environment.\n"
            "This variable is required for system startup or shutdown",
            QMessageBox::Ok, 0);
        return;
    }

    if (!strLpgsHome.endsWith('/'))
    {
        strLpgsHome += '/';
    }

    strSystemId = strSystemId.toUpper();
    strSystemId = strSystemId.at(0);

    QString sqlCmd = "SELECT task_id, proc_id"
                     " FROM tasks"
                     " WHERE system_id = '" + strSystemId + "'";
    query.exec(sqlCmd, __FILE__, __LINE__);

    strlTasks.clear();
    strlTaskProcIds.clear();

    while (query.next())
    {
        strlTasks << query.value("TASK_ID").toString();
        strlTaskProcIds << query.value("PROC_ID").toString();
    }

    if (strlTasks.count() == 0)
    {
        QMessageBox::critical(0, "Missing Configuration",
            "The tasks table in the database does not contain any records.",
            QMessageBox::Ok, 0);
        return;
    }
}

/****************************************************************************/

void LPGSController::startupLPGS()
{
    int status;          // Status returned by wait()

    // Bring up a warning dialog box that asks the operator to confirm
    // if they really want to start the system up.  Note that the
    // parent widget to the QMessageBox is NULL rather than "this"
    // because the LPGSController is a QObject, not a QWidget.
    if (QMessageBox::warning(0, "System Startup", "Start the LPGS system?",
        QMessageBox::Yes | QMessageBox::Cancel, 
        QMessageBox::Yes) != QMessageBox::Yes)
    {
        return;
    }

    if (strLpgsHome.isEmpty())
    {
        QMessageBox::critical(0, "Error", "$LPGSHOME is not set!");
        return;
    }

    status = system("cd $LPGSHOME && PSI &");
    if (status)
    {
        QMessageBox::critical(0, "Error", "Error starting PSI");
        return;
    }
}

/****************************************************************************/
bool LPGSController::checkDupShutdownDirective()
{
    QString sqlStmt;
    IASSqlQuery query;

    // Find if any directives are already entered into the database
    // ------------------------------------------------------------
    sqlStmt="SELECT count(*) FROM ipc_directives ";
    sqlStmt += "WHERE directive LIKE 'shutdown%' AND ";
    sqlStmt += "SYSTEM_ID = '" + strSystemId + "'";

    if (!query.exec(sqlStmt,__FILE__,__LINE__))
    {
        return TRUE;
    }
    query.next();

    // Check if the table has any rows
    // -------------------------------
    if (query.value(0).toInt() > 0)
    {
        QMessageBox::warning(0, "Error",
            "A shutdown directive exists -- cannot override");
        return TRUE;
    }
    return FALSE;
}

void LPGSController::gracefulShutdownLPGS()
{
    /* First check to see if we already gave a shutdown command. */
    if (checkDupShutdownDirective())
        return;

    /* Ask for confirmation and then proceed. */
    if (confirmShutdown("a graceful") == LPGS_PROCEED)
    {
        IASSqlQuery query;
        if (query.exec("INSERT INTO ipc_directives"
                       " (recipient, directive, system_id) "
                       "VALUES ('PSI', 'shutdown gracefully', '"
                       + strSystemId + "')", __FILE__, __LINE__))
        {
            if (query.numRowsAffected() != 1)
            {
                QMessageBox::warning(0, "DB Error",
                    "Failed to insert shutdown command");
                return;
            }

            shutdownType = LPGS_SHUTDOWN_GRACEFUL;
        }
    }
}

/****************************************************************************/

void LPGSController::expeditedShutdownLPGS()
{
    /* First check to see if we already gave a shutdown command. */
    if (checkDupShutdownDirective())
        return;

    /* Ask for confirmation and then proceed. */
    if (confirmShutdown("an expedited") == LPGS_PROCEED)
    {
        IASSqlQuery query;

        if (query.exec("INSERT INTO ipc_directives"
                       " (recipient, directive, system_id) "
                       "VALUES ('PSI', 'shutdown immediately', '"
                       + strSystemId + "')", __FILE__, __LINE__))
        {
            if (query.numRowsAffected() != 1)
            {
                QMessageBox::warning(0, "DB Error",
                    "Failed to insert shutdown command");
                return;
            }

            shutdownType = LPGS_SHUTDOWN_EXPEDITED;
        }
    }
}

/****************************************************************************/

int LPGSController::confirmShutdown(const QString shutdownType)
{
    QString shutdownMessage = "Perform ";
    shutdownMessage.append(shutdownType);
    shutdownMessage.append(" shutdown of the LPGS system?");

    // Bring up a warning dialog box that asks the operator to confirm
    // if they really want to shut the system down.  Note that the
    // parent widget to the QMessageBox is NULL rather than "this"
    // because the LPGSController is a QObject, not a QWidget.

    if (QMessageBox::warning(0, "System Shutdown", shutdownMessage,
        QMessageBox::Yes | QMessageBox::Cancel, 
        QMessageBox::Yes) == QMessageBox::Yes)
    {
        return LPGS_PROCEED;
    }
    else
    {
        return LPGS_CANCEL;
    }
}

/****************************************************************************/

int LPGSController::getShutdownType()
{
    return shutdownType;
}

/****************************************************************************/

void LPGSController::setShutdownType(int newShutdownType) // I: Shutdown type
{
    shutdownType = newShutdownType;
}

/****************************************************************************/

int LPGSController::getSystemStatus()
{
    IASSqlQuery query;
    int currTask;
    int numProcessesRunning = 0;

    QString sqlCmd = "SELECT task_id, proc_id"
                     " FROM tasks"
                     " WHERE system_id = '" + strSystemId + "'";
    query.exec(sqlCmd, __FILE__, __LINE__);

    strlTasks.clear();
    strlTaskProcIds.clear();

    while (query.next())
    {
        strlTasks << query.value("TASK_ID").toString();
        strlTaskProcIds << query.value("PROC_ID").toString();
    }

    for (currTask = 0; currTask < strlTasks.count(); ++currTask)
    {
        if (SystemStatus::isTaskRunning(strlTasks[currTask],
                                        strlTaskProcIds[currTask]))
        {
            numProcessesRunning++;
        }
    }

    if (numProcessesRunning == 0)
    {
        return LPGS_STATUS_DOWN;
    }
    else if (numProcessesRunning == strlTasks.count())
    {
        return LPGS_STATUS_UP;
    }
    else
    {
        return LPGS_STATUS_PARTIAL;
    }
}

