/******************************************************************************
               Level 1 Product Generation System (LPGS) Software
*******************************************************************************
*
* UNIT NAME:   MainWindow.cpp
*
* PURPOSE:
* This is the main window of the LPGS GUI.  It contains the Product Request
* list view, the toolbars, the status bar, and the pull-down menu
* structure.  From here, all other portions of the GUI are dispatched.
*
******************************************************************************/

/* System Includes */
#include <errno.h>
#include <signal.h>
#include <stdlib.h>
#include <unistd.h>
#include <iostream>
#include <time.h>

/* Qt Includes */
#include <QtCore>
#include <QtGui>

/* IAS Qt Includes */
#include <iasQtWoConst.h>

/* Local Includes */
#include "AnomalyInformation.h"
#include "GUIConfiguration.h"
#include "MainWindow.h"
#include "NodeConfiguration.h"
#include "Preferences.h"
#include "IASProcessExecution.h"
#include "SystemConfiguration.h"
#include "IASConfigureInstances.h"
#include "IASUtilities.h"
#include "IASSqlQuery.h"
#include "EventLog.h"

#define MAXHOSTNAMELEN 256
#define PS_BUFFLEN     512
#define PLUGIN_FILENAME "lpgspluginrc"

/* Define some color styles to use for the system status in the status bar */
const QString UP_COLOR_STYLE = "background-color : #44ff44; color : black;";
const QString DOWN_COLOR_STYLE = "background-color : #ff4444; color : black;";
const QString PARTIAL_COLOR_STYLE =
    "background-color : #ffff44; color : black;";
const QString BAD_COLOR_STYLE = "background-color : red; color : black;";

// See TreeWidgetUtilities.h for twInfo type definition
twInfo MainWindow::prInfo[] =
{
    {"Product Request ID",Qt::AlignLeft,-1,FALSE,TRUE,TRUE,110,""},
    {"Landsat Scene ID", Qt::AlignLeft,-1,FALSE,TRUE,TRUE,110,""},
    {"State",Qt::AlignLeft,-1,FALSE,TRUE,TRUE,110,""},
    {"Curr WO ID",Qt::AlignLeft,-1,FALSE,TRUE,TRUE,60,""},
    {"Pri",Qt::AlignLeft,-1,FALSE,TRUE,TRUE,20,""},
    {"Node",Qt::AlignLeft,-1,FALSE,TRUE,TRUE,80,""},
    {"Cncld",Qt::AlignLeft,-1,FALSE,TRUE,TRUE,30,""},
    {"Date Received",Qt::AlignLeft,-1,FALSE,TRUE,TRUE,110,""},
    {"Satellite",Qt::AlignLeft,-1,FALSE,TRUE,TRUE,55,""},
    {"Sensor",Qt::AlignLeft,-1,FALSE,TRUE,TRUE,50,""},
    {"Product Type",Qt::AlignLeft,-1,FALSE,TRUE,TRUE,130,""},
};

extern Preferences *preferences;

MainWindow::MainWindow(QWidget* parent, Qt::WindowFlags fl)
    : QMainWindow(parent, fl)
{
    setupUi(this);

    // Set dialog pointers to NULL so later on their creation can be verified
    systemProcesses = NULL;
    diskUsage = NULL;
    systemTaskStatus = NULL;
    sqlEditor = NULL;
    strHostName = NULL;
    printer = NULL;
    controller = new LPGSController();
    signalMapper = new QSignalMapper(this);

    // Resize to MainWindow to user prefereces
    resize(preferences->width, preferences->height);

    setupMenus();
    setupToolBars();
    setupListView();

    // Link the filters to the ProductRequestTreeWidget
    connect(cboSatelliteFilter, SIGNAL(activated(const QString &)),
        twPRs, SLOT(setSatelliteFilter(const QString &)));
    connect(cboSensorFilter, SIGNAL(activated(const QString &)),
        twPRs, SLOT(setSensorFilter(const QString &)));
    connect(cboProcedureTypeFilter, SIGNAL(activated(const QString &)),
        twPRs, SLOT(setProcedureTypeFilter(const QString &)));
    connect(cboInstanceFilter, SIGNAL(activated(const QString &)),
        twPRs, SLOT(setInstanceFilter(const QString &)));

    // The timer is stopped when any of the filters, besides WO state, are
    // changed. This is requested by the operators because a refresh
    // can be a slow process at times. Clicking the refresh button
    // will restart the timer.
    connect(cboSatelliteFilter, SIGNAL(activated(const QString &)),
        this, SLOT(stopTimer()));
    connect(cboSensorFilter, SIGNAL(activated(const QString &)),
        this, SLOT(stopTimer()));
    connect(cboProcedureTypeFilter, SIGNAL(activated(const QString &)),
        this, SLOT(stopTimer()));
    connect(cboInstanceFilter, SIGNAL(activated(const QString &)),
        this, SLOT(stopTimer()));
    connect(signalMapper, SIGNAL(mapped(const QString &)),
        this, SLOT(executePlugin(const QString &)));

    /* Connect the ProductRequestTreeWidget to the record count */
    connect(twPRs, SIGNAL(countChanged()), this, SLOT(updateRecordCount()));

    // add setup status bar 
    dateTimeLabel = new QLabel(statusBar());
    statusLabel = new QLabel(statusBar());
    numRecordsLabel = new QLabel(statusBar());
    numRecordsLabel->setText("Pancakes for all");

    statusBar()->addWidget(dateTimeLabel);
    statusBar()->addWidget(statusLabel);
    statusBar()->addWidget(new QLabel(statusBar()), true);
    statusBar()->addWidget(numRecordsLabel);

    refreshTimer = new QTimer(this);
    connect(refreshTimer, SIGNAL(timeout()), this, SLOT(refreshEvent()));

    int defaultRefreshRate = preferences->strRefreshTimer.toInt();
    refreshTimer->setSingleShot(true);
    refreshTimer->start(defaultRefreshRate * 1000);

    IASUtilities::addPlugins(menuBar(), signalMapper, PLUGIN_FILENAME);

    alreadyRefreshing = false; // initialize the refresh test variable
    refreshEvent();
}

MainWindow::~MainWindow()
{
    delete systemProcesses;
    delete diskUsage;
    delete systemTaskStatus;
    delete controller;
    delete sqlEditor;
    delete strHostName;
    delete printer;
}

/****************************************************************************/

void MainWindow::refreshStateFilter()
{
    // This function refreshes the list of filters in the state filter toolbar
    // and is called when it is possible that a new filter has been added or
    // one has been deleted. It is mainly called after any execution of the
    // GUI Configuration Dialog since that is where such changes can occur.

    // Store the value of what should be selected upon rebuild.  If nothing
    // is selected, in other words if this is the first time this list is
    // being created, then by default the first item is selected.
    QString strStateToSelect = cboStateFilter->currentText();

    // Clear out the existing entries in the filter combo boxes.
    cboStateFilter->clear();

    IASSqlQuery query;

    // Populate the combo box with the state filter names
    QStringListIterator filterIterator(preferences->filterNames());
    while (filterIterator.hasNext())
    {
        cboStateFilter->addItem(filterIterator.next());
    }
    cboStateFilter->setCurrentIndex(0);
    preferences->currFilter = cboStateFilter->currentText();
    
    // If something was selected before,
    //     loop through the combo box items and search for the desired filter.
    //     If the filter is found, set it as the currentItem.
    if (!strStateToSelect.isEmpty() && !strStateToSelect.isNull())
    {
        for (int compare = 0; compare < cboStateFilter->count(); compare++)
        {
            if (cboStateFilter->itemText(compare) == strStateToSelect)
            {
                cboStateFilter->setCurrentIndex(compare);
                preferences->currFilter = cboStateFilter->currentText();
                break;
            }
        }
    }
}

/****************************************************************************/

void MainWindow::setupStaticfilters()
{
    IASSqlQuery query;

    /* Save the currently selected options. */ 
    QString strSatelliteToSelect = cboSatelliteFilter->currentText();
    QString strSensorToSelect = cboSensorFilter->currentText();
    QString strProcedureTypeToSelect = cboProcedureTypeFilter->currentText();
    QString strInstanceToSelect = cboInstanceFilter->currentText();

    cboSatelliteFilter->clear();
    cboSensorFilter->clear();
    cboProcedureTypeFilter->clear();
    cboInstanceFilter->clear();

    // Populate the combo box with the satellite filter names
    cboSatelliteFilter->addItem("All Satellites");

    // We don't care if the query fails because we have the "All" option
    // already inserted.
    query.exec("SELECT DISTINCT satellite FROM satellite_sensor_cfg",
               __FILE__, __LINE__);

    while (query.next())
    {
        cboSatelliteFilter->addItem(query.value("SATELLITE").toString());
    }

    // If something was selected before, 
    //     loop through the combo box items and search for the desired filter.
    //     If the filter is found, set it as the currentItem.
    if (!strSatelliteToSelect.isEmpty() && !strSatelliteToSelect.isNull())
    {
        for (int compare = 0;compare < cboSatelliteFilter->count(); compare++)
        {
            if (cboSatelliteFilter->itemText(compare) == strSatelliteToSelect)
            {
                cboSatelliteFilter->setCurrentIndex(compare);
                break;
            }
        }
    }

    // Populate the combo box with the sensor filter names
    cboSensorFilter->addItem("All Sensors");

    // We don't care if the query fails because we have the "All" option
    // already inserted.
    query.exec("SELECT DISTINCT sensor FROM satellite_sensor_cfg",
               __FILE__, __LINE__);

    while (query.next())
    {
        cboSensorFilter->addItem(query.value("SENSOR").toString());
    }

    // If something was selected before, 
    //     loop through the combo box items and search for the desired filter.
    //     If the filter is found, set it as the currentItem.
    if (!strSensorToSelect.isEmpty() && !strSensorToSelect.isNull())
    {
        for(int compare = 0; compare < cboSensorFilter->count(); compare++)
        {
            if (cboSensorFilter->itemText(compare) == strSensorToSelect)
            {
                cboSensorFilter->setCurrentIndex(compare);
                break;
            }
        }
    }

    // Populate the combo box with the product type filter names
    cboProcedureTypeFilter->addItem("All Procedures");

    // We don't care if the query fails because we have the "All" option
    // already inserted.
    query.exec("SELECT DISTINCT module_id FROM product_code_modules",
               __FILE__, __LINE__);

    while (query.next())
    {
        cboProcedureTypeFilter->addItem(query.value("MODULE_ID").toString());
    }

    // If something was selected before, 
    //     loop through the combo box items and search for the desired filter.
    //     If the filter is found, set it as the currentItem.
    if (!strProcedureTypeToSelect.isEmpty() &&
        !strProcedureTypeToSelect.isNull())
    {
        for (int compare = 0;
             compare < cboProcedureTypeFilter->count();
             compare++)
        {
            if (cboProcedureTypeFilter->itemText(compare) 
                        == strProcedureTypeToSelect)
            {
                cboProcedureTypeFilter->setCurrentIndex(compare);
                break;
            }
        }
    }

    // Populate the combo box with the instance filter names
    cboInstanceFilter->addItem("All Instances");

    // We don't care if the query fails because we have the "All" option
    // already inserted.
    query.exec("SELECT DISTINCT instance_name FROM processing_instances",
               __FILE__, __LINE__);

    while (query.next())
    {
        cboInstanceFilter->addItem(query.value("INSTANCE_NAME").toString());
    }

    // If something was selected before,
    //     loop through the combo box items and search for the desired filter.
    //     If the filter is found, set it as the currentItem.
    if (!strInstanceToSelect.isEmpty() && !strInstanceToSelect.isNull())
    {
        for(int compare = 0; compare < cboInstanceFilter->count(); compare++)
        {
            if (cboInstanceFilter->itemText(compare) == strInstanceToSelect)
            {
                cboInstanceFilter->setCurrentIndex(compare);
                break;
            }
        }
    }
}

/****************************************************************************/

void MainWindow::setupMenus()
{
    /* Setup signal handlers for the LPGS menu */
    connect(actionSystemStartup, SIGNAL(triggered(bool)), 
        controller, SLOT(startupLPGS()));
    connect(actionGraceful, SIGNAL(triggered(bool)), 
        controller, SLOT(gracefulShutdownLPGS()));
    connect(actionExpedited, SIGNAL(triggered(bool)),
        controller, SLOT(expeditedShutdownLPGS()));
    connect(actionPrint, SIGNAL(triggered(bool)),
        this, SLOT(printRequests()));
    connect(actionQuit, SIGNAL(triggered(bool)),
        qApp, SLOT(closeAllWindows()));

    /* Setup signal handlers for the Edit menu */
    connect(actionSystemConfiguration, SIGNAL(triggered(bool)),
        this, SLOT(displaySystemConfiguration()));
    connect(actionNodeConfiguration, SIGNAL(triggered(bool)),
        this, SLOT(displayNodeConfiguration()));
    connect(actionGuiConfiguration, SIGNAL(triggered(bool)),
        this, SLOT(displayGUIConfiguration()));
    connect(actionInstanceConfiguration, SIGNAL(triggered(bool)),
        this, SLOT(displayInstanceConfiguration()));


    /* Setup signal handlers for the View menu items */
    connect(actionSystemTaskStatus, SIGNAL(triggered(bool)),
        this, SLOT(displaySystemTaskStatus()));
    connect(actionDiskUsage, SIGNAL(triggered(bool)),
        this, SLOT(displayDiskUsage()));
    connect(actionSystemProcesses, SIGNAL(triggered(bool)),
        this, SLOT(displaySystemProcesses()));
    connect(actionLPGSSoftwareLog, SIGNAL(triggered(bool)),
        this, SLOT(displaySoftwareLog()));
    connect(actionEventsLog, SIGNAL(triggered(bool)),
        this, SLOT(displayEventLog()));

    /* Setup signal handlers for toggling toolbars */
    connect(actionCommonTasks, SIGNAL(toggled(bool)),
        this, SLOT(toggleCommonTasks(bool)));
    connect(actionFindRequest, SIGNAL(toggled(bool)),
        this, SLOT(toggleFindRequest(bool)));
    connect(actionStateFilters, SIGNAL(toggled(bool)),
        this, SLOT(toggleStateFilters(bool)));
    connect(actionDateSelect, SIGNAL(toggled(bool)),
        this, SLOT(toggleDateSelect(bool)));
    connect(actionInterfaceStatuses, SIGNAL(toggled(bool)),
        this, SLOT(toggleInterfaceStatuses(bool)));

    connect(menuToolbars, SIGNAL(aboutToShow()),
        this, SLOT(refreshMenuToolbars()));

    /* Determine which statusbar items should be enabled */
    if (preferences->statusbarDateVisible)
        actionDateToggle->setChecked(true);
    if (preferences->statusbarRecordsReturnedVisible)
        actionRecordsReturnedToggle->setChecked(true);
    if (preferences->statusbarSystemStatusVisible)
        actionSystemStatusToggle->setChecked(true);

    /* Setup signal handlers for toggling statubar items */
    connect(actionDateToggle, SIGNAL(toggled(bool)),
        this, SLOT(toggleDate(bool)));
    connect(actionRecordsReturnedToggle, SIGNAL(toggled(bool)),
        this, SLOT(toggleRecordsReturned(bool)));
    connect(actionSystemStatusToggle, SIGNAL(toggled(bool)),
        this, SLOT(toggleSystemTaskStatus(bool)));

    /* Setup signal handlers for the Tools menu items */
    connect(actionAnomalyInformation, SIGNAL(triggered(bool)),
        this, SLOT(displayAnomalyInformation()));
    connect(actionSQLPlusWindow, SIGNAL(triggered(bool)),
        this, SLOT(openSQLWindow()));
    connect(actionSQLEditor, SIGNAL(triggered(bool)),
        this, SLOT(displayRecordSetViewer()));
    connect(actionUNIXShell, SIGNAL(triggered(bool)),
        this, SLOT(openUNIXShell()));
}

/****************************************************************************/

void MainWindow::setupListView()
{
    twPRs = new ProductRequestTreeWidget(this, drMain);
    setCentralWidget(twPRs);

    // set validator for product request line edit entry to new 
    // ProductRequestIncrementalSearch instance
    prValidator = new ProductRequestIncrementalSearch(twPRs, twPRs);
    leFindRequest->setValidator(prValidator);
}

/****************************************************************************/

void MainWindow::setupToolBars()
{
    /* Check preferences to determine which toolbar should be displayed 
       by default */
    if (!preferences->commonTasksVisible)
        toolbarCommonTasks->hide();
    if (!preferences->findRequestVisible)
        toolbarFindRequest->hide();
    if (!preferences->stateFiltersVisible)
        toolbarStateFilters->hide();
    if (!preferences->dateRangeVisible)
        toolbarDateSelect->hide();
    if (!preferences->interfaceStatusesVisible)
        toolbarInterfaceStatuses->hide();

    /* This one is actualy a label in the status bar. */
    if (!preferences->statusbarRecordsReturnedVisible)
        numRecordsLabel->hide();

    /* Most of the Common Tasks signal are setup in setupMenus.  Only
       the refresh signal needs to be setup here */
    connect(actionRefresh, SIGNAL(triggered(bool)),
        this, SLOT(refreshEvent()));
   
    /* Widgets cannot be added to a toolbar within Designer so the Incremental
       Search toolbar must be setup programmatically.  
       The toolbar provides a line edit box that performs an
       incremental search of all currently-displayed Product Requests.
       additional code to set the validator for the line edit is located in
       setupListView(); because it the validator can not be set for an object
       that does not exist yet */
    toolbarFindRequest->addWidget(new QLabel("Find Request: ",
        toolbarFindRequest));

    leFindRequest = new QLineEdit(toolbarFindRequest);
    toolbarFindRequest->addWidget(leFindRequest);
    
    // The third toolbar provides the operator with a combo box that
    // contains various "quick filter" options (i.e., "All Active".)
    // Another dialog allows more sophisticated filtering, but the
    // purpose of this toolbar is to allow the operator to quickly
    // view common sets of LPGS Product Request states.
    toolbarStateFilters->addWidget(new QLabel("Filter: ", toolbarStateFilters));
    
    // cboStateFilter is a private member of the MainWindow class
    cboStateFilter = new QComboBox(toolbarStateFilters);
    toolbarStateFilters->addWidget(cboStateFilter);
    
    // cboSatelliteFilter is a private member of the MainWindow class
    cboSatelliteFilter = new QComboBox(toolbarStateFilters);
    toolbarStateFilters->addWidget(cboSatelliteFilter);
    
    // cboSensorFilter is a private member of the MainWindow class
    cboSensorFilter = new QComboBox(toolbarStateFilters);
    toolbarStateFilters->addWidget(cboSensorFilter);
    
    // cboProcedureTypeFilter is a private member of the MainWindow class
    cboProcedureTypeFilter = new QComboBox(toolbarStateFilters);
    toolbarStateFilters->addWidget(cboProcedureTypeFilter);

    // cboInstanceFilter is a private member of the MainWindow class
    cboInstanceFilter = new QComboBox(toolbarStateFilters);
    toolbarStateFilters->addWidget(cboInstanceFilter);

    // Populate the list of filters in the combo boxes
    refreshStateFilter();
    setupStaticfilters();
    
    connect(cboStateFilter, SIGNAL(activated( const QString &)),
        this, SLOT(changeStateFilter(const QString &)));
    
    // add a tool bar object that allows the operator to select a date range for
    // the product request.  By default this object is not displayed
    int defaultDuration = preferences->strPRDuration.toInt();
    drMain = new DateRange(toolbarDateSelect, 0, defaultDuration, horizontal);
    drMain->showTimes();
    toolbarDateSelect->addWidget(drMain);

    interfaceStatus = new InterfaceStatus(toolbarInterfaceStatuses);
    toolbarInterfaceStatuses->addWidget(interfaceStatus);

    insertToolBarBreak(toolbarDateSelect); /* Separate the tool bars. */
    insertToolBarBreak(toolbarInterfaceStatuses); /* Separate the tool bars. */
}
/****************************************************************************/

void MainWindow::refreshMenuToolbars()
{
    // This slot is tied to the menuTools's aboutToShow event.  This keeps the
    // checked status in synch with the visibility of the toolbar items.
    actionCommonTasks->setChecked(toolbarCommonTasks->isVisible());
    actionFindRequest->setChecked(toolbarFindRequest->isVisible());
    actionStateFilters->setChecked(toolbarStateFilters->isVisible());
    actionDateSelect->setChecked(toolbarDateSelect->isVisible());
    actionInterfaceStatuses->setChecked(toolbarInterfaceStatuses->isVisible());
}

/****************************************************************************/

void MainWindow::displayDiskUsage()
{
    // If diskUsage has not been shown yet, create it, otherwise just
    // refresh the contents to give a current display.

    if (diskUsage == NULL)
        diskUsage = new DiskUsage(this);
    else
        diskUsage->refresh();

    diskUsage->show();
}

/****************************************************************************/

void MainWindow::displayEventLog()
{
    EventLog eventLog(this);

    eventLog.exec();
}

void MainWindow::displayGUIConfiguration()
{
    GUIConfiguration guiConfiguration(this);
    guiConfiguration.setModal(true);
    guiConfiguration.exec();

    // Update Main Window objects that may need to change as a result of changed
    // settings in the GUI Configuration dialog.
    drMain->setDaysBetween(preferences->strPRDuration.toInt());

    refreshStateFilter();
    changeStateFilter(cboStateFilter->currentText());
}

/****************************************************************************/

void MainWindow::displaySystemConfiguration()
{
    SystemConfiguration systemConfiguration(this);
    systemConfiguration.exec();
}

/****************************************************************************/

void MainWindow::displaySystemProcesses()
{
    // If systemProcesses has not been shown yet, create it, otherwise just
    // refresh the contents to give a current display.

    if (systemProcesses == NULL)
        systemProcesses = new SystemProcesses(this);
    else
        systemProcesses->refresh();

    systemProcesses->show();
}

/****************************************************************************/

void MainWindow::displaySystemTaskStatus()
{
    // If systemTaskStatus has not been shown yet, create it, otherwise just
    // refresh the contents to give a current display.

    if (systemTaskStatus == NULL)
        systemTaskStatus = new SystemTaskStatus(this);
    else
        systemTaskStatus->refresh();

    systemTaskStatus->show();
}

/****************************************************************************/

void MainWindow::displayRecordSetViewer()
{
    // This function displays an empty RecordSetViewer object (class) which
    // will allow the operator to enter SQL commands, execute them, and
    // view their return values.  If the command is a query, it will return
    // the rows and columns of the query in a treewidget.  Any Oracle error
    // messages will also be displayed

    // If the SQL editor has not been created do so
    if (sqlEditor == NULL)
        sqlEditor = new RecordSetViewer(this);

    sqlEditor->show();
}

/****************************************************************************/

void MainWindow::displayAnomalyInformation()
{
    AnomalyInformation anomalyInformation("NULL", this);

    anomalyInformation.exec();
}

/****************************************************************************/

void MainWindow::displayNodeConfiguration()
{
    // Bring up the Node Configuration dialog
    NodeConfiguration nodeConfiguration(this);
    nodeConfiguration.exec();
}

void MainWindow::displayInstanceConfiguration()
{
    IASConfigureInstances configureInstances(this);
    configureInstances.exec();
}

/****************************************************************************/

void MainWindow::displaySoftwareLog()
{
    system("xterm -T \"LPGS Software Log - $LPGS_LOG/LPGS-Log\" "
           "-fg white -bg black -font 8x13 -geometry 262x15 "
           "-e tail -f -n250 $LPGS_LOG/LPGS-Log &");
}

/****************************************************************************/

QString MainWindow::getDeletablePRSpace()
{
    // This function calculates how much space is occupied by completed product
    // requests whose space is deletable. A QString of the number followed by
    // the closest unit is returned.  (Mb, Kb, Gb, etc...)
    QString strDeletableSpace = "AN UNKNOWN AMOUNT";
    float totalSpace = 0;
    QString strTotalSpaceUnit = "Bytes";
    QDir woPath;

    IASSqlQuery query;

    QString sqlCmd = "SELECT wo_path"
                     " FROM work_orders"
                     " WHERE state IN ('" + DB_WO_COMPLETED + "', '"
                                          + DB_WO_ABORTED + "')"
                         " AND deleted = 0";

    if (!query.exec(sqlCmd, __FILE__, __LINE__))
    {
        strDeletableSpace = QString::number(totalSpace, 'f', 2)
            + " " + strTotalSpaceUnit;

        return strDeletableSpace;
    }

    while (query.next())
    {
        woPath.setPath(query.value("WO_PATH").toString());
        if (woPath.exists())
            totalSpace += getDirectorySize(woPath);
    }

    // Get closest unit.  (i.e. Kb, Mb, etc...)
    for (int unit = 0; unit < 4 && totalSpace > 1024; unit++)
    {
        totalSpace = totalSpace / 1024;        
        switch (unit)
        {
            case 0: 
                strTotalSpaceUnit = "Kb";
                break;
            case 1: 
                strTotalSpaceUnit = "Mb";
                break;
            case 2: 
                strTotalSpaceUnit = "Gb";
                break;
            case 3: 
                strTotalSpaceUnit = "Tb";
                break;
        }
    }
    strDeletableSpace = QString::number(totalSpace, 'f', 2)
        + " " + strTotalSpaceUnit;

    return strDeletableSpace;
}

/****************************************************************************/

float MainWindow::getDirectorySize(const QDir path)
{
    //Returns the size, in bytes, of the directory passed in
    float directorySize = 0;
    
    QString filename;
    QFileInfoList list = path.entryInfoList();
    QFileInfo fi;

    QListIterator<QFileInfo> it(list);
    while (it.hasNext())
    {
        fi = it.next();
        filename = fi.fileName();
        
        //DO not calculate space if the filename is "." or ".." because
        //these are relative paths that will cause recursion to occur
        //that would look at every single directory on the entire system.
        if (!(filename == "." || filename == ".." ))
        {
            if (fi.isDir())
                directorySize += getDirectorySize(QDir(fi.filePath()));
            else
                directorySize += fi.size();
        }
    }

    return directorySize;
}

/****************************************************************************/

void MainWindow::openSQLWindow()
{
    QString cmd = "xterm -T \"LPGS Database\" " + preferences->strXtermOptions 
        + " -e sqlplus $IAS_DB_COM  &";

    system(cmd.toAscii().data());
}

/****************************************************************************/

// The below functions are nothing more than slots that get called
// when the user toggles one of the menu items under the "Toolbars"
// submenu (i.e., when they hide or unhide the Incremental Search tool.)

void MainWindow::toggleCommonTasks(bool showIt)
{
    if (showIt)
        toolbarCommonTasks->show();
    else
        toolbarCommonTasks->hide();
}

/****************************************************************************/

void MainWindow::toggleFindRequest(bool showIt)
{
    if (showIt)
        toolbarFindRequest->show();
    else
        toolbarFindRequest->hide();
}

/****************************************************************************/

void MainWindow::toggleStateFilters(bool showIt)
{
    if (showIt)
        toolbarStateFilters->show();
    else
        toolbarStateFilters->hide();
}

/****************************************************************************/

void MainWindow::toggleDateSelect(bool showIt)
{
    if (showIt)
        toolbarDateSelect->show();
    else
        toolbarDateSelect->hide();
}

/****************************************************************************/

void MainWindow::toggleInterfaceStatuses(bool showIt)
{
    if (showIt)
        toolbarInterfaceStatuses->show();
    else
        toolbarInterfaceStatuses->hide();
}
/****************************************************************************/

void MainWindow::toggleDate(bool showIt)
{
    if (showIt)
        dateTimeLabel->show();
    else
        dateTimeLabel->hide();
}

/****************************************************************************/

void MainWindow::toggleSystemTaskStatus(bool showIt)
{
    if (showIt)
        statusLabel->show();
    else
        statusLabel->hide();
}

/****************************************************************************/

void MainWindow::toggleRecordsReturned(bool showIt)
{
    if (showIt)
        numRecordsLabel->show();
    else
        numRecordsLabel->hide();
}

/****************************************************************************/

void MainWindow::updateDateTime()
{
    time_t now = time(NULL);
    QString dateTime = ctime(&now);
    dateTimeLabel->setText(dateTime.trimmed());
}

/****************************************************************************/

void MainWindow::refreshEvent()
{
    updateDateTime();

    /* This will return true if it can set the value to true, otherwise it
       will return false  */
    if (!alreadyRefreshing.testAndSetAcquire(false, true))
        return;
    
    static int lastStatus = -1;
    static int lastDay = -1;
    static int hasBeenWarned = false;

    // Check the current date and see if it has changed since the last
    // refresh (i.e., if it's past midnight.)  If so, update the date
    // range to use the current (new) day.  If this is not done, then
    // if the operators leave the GUI running overnight, new product
    // requests will not show up in the morning since the "To" date in
    // the date range widget is still set to yesterday.
    time_t now = time(NULL);
    struct tm *lt = localtime(&now);

    if (lt->tm_mday != lastDay)
        drMain->setDateToCurrent();

    lastDay = lt->tm_mday;

    int status = controller->getSystemStatus();
    if (status == LPGS_STATUS_UP)
    {
        statusLabel->setStyleSheet(UP_COLOR_STYLE);
        statusLabel->setText("LPGS System Status: Up");
        hasBeenWarned = false;

        // Based on Status, certain menu items are enabled/disabled
        actionSystemStartup->setEnabled(false);
        actionGraceful->setEnabled(true);
        actionExpedited->setEnabled(true);
    }
    else if (status == LPGS_STATUS_DOWN)
    {
        statusLabel->setStyleSheet(DOWN_COLOR_STYLE);
        statusLabel->setText("LPGS System Status: Down");
        controller->setShutdownType(LPGS_SHUTDOWN_NONE);
        hasBeenWarned = false;

        // Based on Status, certain menu items are enabled/disabled
        actionSystemStartup->setEnabled(true);
        actionGraceful->setEnabled(false);
        actionExpedited->setEnabled(false);
    }
    else if (status == LPGS_STATUS_PARTIAL)
    {
        if (lastStatus == LPGS_STATUS_UP &&
             controller->getShutdownType() == LPGS_SHUTDOWN_NONE &&
             hasBeenWarned == false)
        {
            QMessageBox::critical(this, "LPGS Task(s) Missing!",
                "One or more the the LPGS system tasks has gone down.\n"
                "Please review the software log to determine the cause.",
                "&OK"); 
            hasBeenWarned = true;
        }

        QString strStatus = "LPGS System Status: Partially Up";

        // Based on Status, certain menu items are enabled/disabled
        actionSystemStartup->setEnabled(false);
        actionGraceful->setEnabled(false);
        actionExpedited->setEnabled(false);
        switch (controller->getShutdownType())
        {
            case LPGS_SHUTDOWN_GRACEFUL:
                strStatus += " (Graceful Shutdown in Progress)";
                break;

            case LPGS_SHUTDOWN_EXPEDITED:
                strStatus += " (Expedited Shutdown in Progress)";
                break;

            default:
                strStatus += " (System Tasks Missing)";
                actionGraceful->setEnabled(true);
                actionExpedited->setEnabled(true);
                break;
        }

        statusLabel->setStyleSheet(PARTIAL_COLOR_STYLE);
        statusLabel->setText(strStatus);
    }
    else
    {
        /* LPGS_STATUS_UNKNOWN */
        statusLabel->setStyleSheet(BAD_COLOR_STYLE);
        statusLabel->setText("LPGS System Status: Unknown (!?)");
    }

    twPRs->refresh();

    interfaceStatus->refresh();

    lastStatus = status;

    /* Reset the test value back to false. */
    alreadyRefreshing = false;

    refreshTimer->start(preferences->strRefreshTimer.toInt() * 1000);
}

/****************************************************************************/

void MainWindow::updateRecordCount()
{
    numRecordsLabel->setText(QString::number(twPRs->topLevelItemCount())
                             + " Records Returned");
}

/****************************************************************************/

void MainWindow::changeStateFilter(const QString &strFilterName)
{
    preferences->currFilter = strFilterName;
    refreshEvent();
}

/****************************************************************************/

void MainWindow::openUNIXShell()
{
    // If there is a currently-selected item in the Product Request
    // treewidget, call the openUNIXShell() slot in that widget so that
    // the default behavior applies (the shell window is brought up
    // in the proper directory.)  Otherwise, just open up a new
    // shell window in the current working directory.

    if (twPRs->currentItem())
    {
        twPRs->openUNIXShell();
    }
    else
    {
        QString cmd = "xterm -T \"LPGS\" "+preferences->strXtermOptions+" &";
        system(cmd.toAscii().data());
    }
}

/****************************************************************************/

void MainWindow::printRequests()
{
    const int margin = 10;
    int pageNo = 1;
    int yPos = 0;
    QString strLine;
    QPainter painter;

    if (printer == NULL)
        printer = new QPrinter();

    printer->setPageSize(QPrinter::Letter);
    printer->setColorMode(QPrinter::GrayScale);

    QPrintDialog printDialog(printer, this);
    if (printDialog.exec() == QDialog::Accepted)
    {
        if (!painter.begin(printer))
            return;

        painter.setFont(QFont("Times", 10, QFont::Bold));
        QFontMetrics fm = painter.fontMetrics();

        strLine = "\t\t\t\tLPGS Product Requests";
        painter.drawText(margin, margin + yPos, printer->width(),
            fm.lineSpacing(), Qt::TextExpandTabs | Qt::TextDontClip, strLine);
        yPos += 24;

        strLine = "Date Range: " + drMain->toString();
        painter.drawText(margin, margin + yPos, printer->width(),
            fm.lineSpacing(), Qt::TextExpandTabs | Qt::TextDontClip, strLine);
        yPos += 24;

        // Copy the column struct array into a vector for printing
        int numColumns = ARRAY_SIZE(prInfo);

        vector<twInfo> prInfoVector;
        for (int i = 0; i < numColumns; i++)
            prInfoVector.push_back(prInfo[i]);
        
        TreeWidgetUtilities::printTreeWidget(twPRs, painter, printer,
            margin, yPos, pageNo, prInfoVector, numColumns);
    }
}

/****************************************************************************/
void MainWindow::stopTimer()
{
    refreshTimer->stop();
}

/****************************************************************************/
void MainWindow::executePlugin(const QString& cmd)
{
    IASProcessExecution pe(this);
    QString command = cmd;
    
    if (!pe.executeDetached(command))
    {
        QString msg = tr("Failed to execute plugin.\n") + command;
        QMessageBox::critical(this, "Plugin Error", msg);
    }
}

