/******************************************************************************
               Level 1 Product Generation System (LPGS) Software
 ******************************************************************************
 *
 * UNIT NAME:   NodeConfiguration.cpp
 *
 * PURPOSE:
 * The Node Configuration dialog allows the operator to view the available
 * LPGS processing nodes, set the state of a node to UP or DOWN, and set
 * the maximum number of concurrent work orders for each node.
 *
 *****************************************************************************/

/* System Includes */

/* Qt Includes */
#include <QtCore>
#include <QtGui>

/* Local Includes */
#include "NodeConfiguration.h"
#include "IASSqlQuery.h"
#include "IASErrorDialog.h"

NodeConfiguration::NodeConfiguration(
    QWidget *parent,    // I: Parent widget
    Qt::WindowFlags fl) // I: Additional Qt window flags
: QDialog(parent, fl),
  numNodes(0)
{
    setupUi(this);

    QLabel *lblTmp;         // Temporary label pointer for columns

    // Setup the column headers
    QStringList columns;
    columns << "Hostname" << "Host State" << "SW Status" << "SW Status Date"
            << "Max Load" << "Current Load" << "CPUs Per Order";

    for (int currColumn = 0; currColumn < columns.size(); ++currColumn)
    {
        lblTmp = new QLabel(columns[currColumn]);
        glayoutNodeConfig->addWidget(lblTmp, 0, currColumn);
    }

    refreshNodes();

    // Setup the signal/slots
    QObject::connect(pbRefresh, SIGNAL(clicked()), this, SLOT(refreshNodes()));
    QObject::connect(pbApply, SIGNAL(clicked()), this, SLOT(applyChanges()));
    QObject::connect(pbCancel, SIGNAL(clicked()), this, SLOT(reject()));
    QObject::connect(pbOK, SIGNAL(clicked()), this, SLOT(accept()));

    // Disable the Apply button initially because no changes have been
    // made yet. If any of the node configuration changes, the Apply
    // button will be enabled.
    pbApply->setEnabled(false);
}

void NodeConfiguration::refreshNodes()
{
    QString sql = "SELECT hostname, state, status, status_date, "
                  "       max_load, current_load, cpus_per_order"
                  "  FROM processing_nodes";

    IASSqlQuery query;
    if (!query.exec(sql, __FILE__, __LINE__))
    {
        return;
    }

    clearNodes();

    QIntValidator *maxLoadVal = new QIntValidator(0, 999, this);
    QIntValidator *maxCpuVal = new QIntValidator(1, 99, this);

    int currNode = 0;
    while (query.next())
    {
        // check max nodes reached (MAX_NODES)
        if (currNode >= MAX_NODES)
        {
            QMessageBox::critical(this, "Maximum nodes exceeded",
                "The number of nodes exceeds maximum number allowed by "
                "the Node Configuration. Additional nodes will not be added.");
            break;
        }

        QString hostname = query.value("HOSTNAME").toString();
        QString state = query.value("STATE").toString();
        QString statusDate = query.value("STATUS_DATE").toString();
        QString status = query.value("STATUS").toString();
        QString maxLoad = query.value("MAX_LOAD").toString();
        QString currLoad = query.value("CURRENT_LOAD").toString();
        QString cpusPerOrder = query.value("CPUS_PER_ORDER").toString();

        leHostname[currNode] = new QLineEdit(hostname);
        leHostname[currNode]->setMinimumWidth(100);
        leHostname[currNode]->setReadOnly(true);

        cboState[currNode] = new QComboBox();
        cboState[currNode]->addItem("UP");
        cboState[currNode]->addItem("DOWN");
    
        if (state == "UP")
            cboState[currNode]->setCurrentIndex(0);
        else
            cboState[currNode]->setCurrentIndex(1);
   
        connect(cboState[currNode], SIGNAL(activated(int)),
                this, SLOT(valueChanged()));

        leStatusDate[currNode] = new QLineEdit(statusDate);
        leStatusDate[currNode]->setMinimumWidth(170);
        leStatusDate[currNode]->setReadOnly(true);

        leStatus[currNode] = new QLineEdit(status);
        leStatus[currNode]->setMinimumWidth(80);
        if (status == "RUNNING")
        {
            leStatus[currNode]->setStyleSheet(
                "background-color : #44ff44; color : black;");
        }
        else
        {
            leStatus[currNode]->setStyleSheet(
                "background-color : #ff4444; color : black;");
        }
        leStatus[currNode]->setReadOnly(true);

        leMaxLoad[currNode] = new QLineEdit(maxLoad);
        leMaxLoad[currNode]->setMaxLength(3);
        leMaxLoad[currNode]->setMaximumWidth(75);
        leMaxLoad[currNode]->setValidator(maxLoadVal);
        connect(leMaxLoad[currNode], SIGNAL(textChanged(const QString &)),
            this, SLOT(valueChanged()));

        leCurrLoad[currNode] = new QLineEdit(currLoad);
        leCurrLoad[currNode]->setReadOnly(true);
        leCurrLoad[currNode]->setMaximumWidth(75);

        leCpusPerOrder[currNode] = new QLineEdit(cpusPerOrder);
        leCpusPerOrder[currNode]->setMaxLength(2);
        leCpusPerOrder[currNode]->setMaximumWidth(75);
        leCpusPerOrder[currNode]->setValidator(maxCpuVal);
        connect(leCpusPerOrder[currNode], SIGNAL(textChanged(const QString &)),
            this, SLOT(valueChanged()));
        
        glayoutNodeConfig->addWidget(leHostname[currNode], currNode+1, 0);
        glayoutNodeConfig->addWidget(cboState[currNode], currNode+1, 1);
        glayoutNodeConfig->addWidget(leStatus[currNode], currNode+1, 2);
        glayoutNodeConfig->addWidget(leStatusDate[currNode], currNode+1, 3);
        glayoutNodeConfig->addWidget(leMaxLoad[currNode], currNode+1, 4);
        glayoutNodeConfig->addWidget(leCurrLoad[currNode], currNode+1, 5);
        glayoutNodeConfig->addWidget(leCpusPerOrder[currNode], currNode+1, 6);
        
        currNode++;
    }

    numNodes = currNode;

    // The GUI values will now match the database. Disable the Apply button
    pbApply->setEnabled(false);
}

void NodeConfiguration::clearNodes()
{
    for (int i = 0; i < numNodes; i++)
    {
        glayoutNodeConfig->removeWidget(leHostname[i]);
        glayoutNodeConfig->removeWidget(cboState[i]);
        glayoutNodeConfig->removeWidget(leStatus[i]);
        glayoutNodeConfig->removeWidget(leStatusDate[i]);
        glayoutNodeConfig->removeWidget(leMaxLoad[i]);
        glayoutNodeConfig->removeWidget(leCurrLoad[i]);
        glayoutNodeConfig->removeWidget(leCpusPerOrder[i]);

        delete leHostname[i];
        delete cboState[i];
        delete leStatus[i];
        delete leStatusDate[i];
        delete leMaxLoad[i];
        delete leCurrLoad[i];
        delete leCpusPerOrder[i];

        leHostname[i] = NULL;
        cboState[i] = NULL;
        leStatus[i] = NULL;
        leStatusDate[i] = NULL;
        leMaxLoad[i] = NULL;
        leCurrLoad[i] = NULL;
        leCpusPerOrder[i] = NULL;
    }
}

void NodeConfiguration::valueChanged()
{
    pbApply->setEnabled(true);
}

void NodeConfiguration::applyChanges()
{
    QString sql = "UPDATE processing_nodes "
                  "   SET max_load = :max_load, "
                  "       state = :state, "
                  "       cpus_per_order = :cpus_per_order"
                  " WHERE hostname = :hostname";

    QSqlDatabase::database().transaction();

    QSqlQuery query;
    if (!query.prepare(sql))
    {
        IASErrorDialog::databaseError(this, "Update processing nodes failed",
            __FILE__, __LINE__, &query);
        return;
    }

    for (int currNode = 0; currNode < numNodes; ++currNode) 
    {
        query.bindValue(":max_load", leMaxLoad[currNode]->text());
        query.bindValue(":state", cboState[currNode]->currentText());
        query.bindValue(":hostname", leHostname[currNode]->text());
        query.bindValue(":cpus_per_order", leCpusPerOrder[currNode]->text());

        bool result = query.exec();

        // On error, display a message and rollback
        if (result == false)
        {
            IASErrorDialog::databaseError(this, 
                "Update processing nodes failed", __FILE__, __LINE__, &query);
            QSqlDatabase::database().rollback();
            return;
        }
    }
    QSqlDatabase::database().commit();

    pbApply->setEnabled(false);
}

void NodeConfiguration::accept()
{
    applyChanges();
    QDialog::accept();
}

