/******************************************************************************
               Level 1 Product Generation System (LPGS) Software
 ******************************************************************************
 *
 * UNIT NAME:   Preferences.cpp
 *
 * PURPOSE:
 * The Preferences class stores various operator preferences for things
 * such as colors, sorting and filtering options, fonts, etc.  It also
 * provides the rest of the GUI with a convenient mechanism for accessing
 * them.
 *
 *****************************************************************************/

/* System Includes */
#include <stdlib.h>
#include <iostream>

/* Local Includes */
#include "Utilities.h"
#include "Preferences.h"

/* Use constants for the names of the preference parameters. */
const QString PREF_WIDTH = "width";
const QString PREF_HEIGHT = "height";
const QString PREF_COLOR_CODING = "colorCoding";
const QString PREF_COMMON_TASKS_VISIBLE = "commonTasksVisible";
const QString PREF_FIND_REQUEST_VISIBLE = "findRequestVisible";
const QString PREF_STATE_FILTERS_VISIBLE = "stateFiltersVisible";
const QString PREF_DATE_RANGE_VISIBLE = "dateRangeVisible";
const QString PREF_INTERFACE_STATUSES_VISIBLE = "interfaceStatusesVisible";
const QString PREF_STATUSBAR_DATE_VISIBLE = "statusbarDateVisible";
const QString PREF_STATUSBAR_RECORDS_RETURNED_VISIBLE =
              "statusbarRecordsReturnedVisible";
const QString PREF_STATUSBAR_SYSTEM_STATUS_VISIBLE =
              "statusbarSystemStatusVisible";
const QString PREF_REFRESH_TIMER = "refreshTimer";
const QString PREF_PRODUCT_REQUEST_DURATION = "productRequestDuration";
const QString PREF_EVENT_LOG_DURATION = "eventLogDuration";
const QString PREF_FONT = "font";
const QString PREF_SIZE = "size";
const QString PREF_SCROLLBACK_LINES = "scrollbackLines";
const QString PREF_COLOR_SCHEME = "colorScheme";
const QString PREF_COLOR_SWITCH = "colorSwitch";
const QString PREF_GROUP_NAME = "groupName";
const QString PREF_GROUP_STATES = "groupStates";
const QString PREF_GROUP_BACKGROUND_COLOR = "groupBackgroundColor";
const QString PREF_GROUP_FOREGROUND_COLOR = "groupForegroundColor";

const QString PREF_FILTER_NAME = "filterName";
const QString PREF_FILTER_STATES = "filterStates";

/* Deprecated names that can eventually be removed. */
const QString PREF_INCREMENTAL_SEARCH_VISIBLE = "incrementalSearchVisible";
const QString PREF_STATE_FILTER_VISIBLE = "stateFilterVisible";

Preferences *preferences;

/*****************************************************************************/

Preferences::Preferences()
{
    // Initialize the list of states with the names from Utilities.cpp.
    // This information is being (essentially) duplicated because in
    // Utilities, it is stored as an array of QStrings and here it is
    // advantageous to store them as a QStringList.

    /* Default the width and height values to half the screen */
    width = QApplication::desktop()->screenGeometry().width()/2;
    height = QApplication::desktop()->screenGeometry().height()/2;

    // Now that we've assembled the name of the preferences file, read
    // it in and parse out the parameter values.
    if (!readPreferencesFile())
    {
        QMessageBox::critical(0, "lpgsuirc File Missing",
            "Can't find the " + strPreferencesFilename + " file "
                "terminating!",
            QMessageBox::Ok, QMessageBox::Ok);

        exit(EXIT_FAILURE);
    }

    // Assemble a list of options to "xterm" that will implement the
    // user's desired color scheme, geometry, and scrollback lines.
    // This list will look like "-font 8x13 -geometry 80x30 -sl 1024."

    strXtermOptions = "-font " + strFont + " -geometry " + strSize +
                      " -sl " + strScrollbackLines + " " + strColorSwitch;
}

/*****************************************************************************/

/*****************************************************************************/

// Below are a bunch of accessor functions intended to let the LPGS GUI
// "get at" some of the private variables contained within the class.

QStringList Preferences::groupNames()
{
    return backgroundColor.keys();
}

QStringList Preferences::filterNames()
{
    return strlOrderedFilterNames;
}

QString Preferences::filterStates(
    const QString &filterName)
{
    return mapFilterNameToStates[filterName];
}

QString Preferences::quotedFilterStates(
    const QString &filterName)
{
    QStringList subParts = mapFilterNameToStates[filterName].split(",",
                               QString::SkipEmptyParts);

    QString strQuotedList = "";
    for (int i = 0; i < subParts.count(); i++)
    {
        if (i > 0)
            strQuotedList += ",";

        strQuotedList += "'" + subParts[i] + "'";
    }

    return strQuotedList;
}

void Preferences::addGroup(
    const QString &strGroupName) // I: Name of new group to add
{
    backgroundColor[strGroupName] = "";
    textColor[strGroupName] = "";
}

void Preferences::deleteGroup(
    const QString &strGroupName) // I: Name of group to delete
{
    backgroundColor.remove(strGroupName);
    textColor.remove(strGroupName);

    MapStatesToGroupsIterator iterator(group);
    while (iterator.hasNext())
    {
        iterator.next();
        if (iterator.value() == strGroupName)
            group[iterator.key()] = DEFAULT_GROUP;
    }
}

void Preferences::addFilter(
    const QString &strFilterName) // I: Name of new filter to add
{
    strlOrderedFilterNames += strFilterName;
    mapFilterNameToStates[strFilterName] = "";
}

void Preferences::deleteFilter(
    const QString &strFilterName) // I: Name of filter to delete
{
    strlOrderedFilterNames.removeAll(strFilterName);
    mapFilterNameToStates.remove(strFilterName);
}

void Preferences::setFilterStates(
    const QString &strFilterName, // I: Filter number to modify
    const QString &strStates)  // I: List of states for specified filter
{
    mapFilterNameToStates[strFilterName] = strStates;
}

/*****************************************************************************/

bool Preferences::areColorsEnabled()
{
    return colorsEnabled;
}

/*****************************************************************************/

void Preferences::setColorsEnabled(
    bool flag )     // I: Display Product Requests in color?
{
    colorsEnabled = flag;
}

/*****************************************************************************/

bool Preferences::readPreferencesFile()
{
    QString line;
    QString currentFilterName;
    QFile filePreferences;

    int scrWidth = QApplication::desktop()->screenGeometry().width();
    int scrHeight = QApplication::desktop()->screenGeometry().height();

    /* Get the name of the preferences file.  The default file will always
       be named "lpgsuirc", and is stored under the OWD directory for the
       specific LPGS instance. */

    /* Try the users home directory first. */
    QString preferencesFilename = getenv("HOME");
    if (preferencesFilename.isEmpty())
    {
        return false;
    }
    preferencesFilename += "/.lpgsuirc";

    /* Always save to the users home directory. */
    strPreferencesFilename = preferencesFilename;

    filePreferences.setFileName(preferencesFilename);
    if (!filePreferences.open(QIODevice::ReadOnly)) 
    {
        /* That failed so try the installation. */
        preferencesFilename = getenv("OWD");
        if (preferencesFilename.isEmpty())
        {
            return false;
        }
        preferencesFilename += "/etc/lpgsuirc";

        filePreferences.setFileName(preferencesFilename);
        if (!filePreferences.open(QIODevice::ReadOnly)) 
        {
            return false;
        }
    }

    QTextStream tsPrefs(&filePreferences);
    currFilter = "";

    QString currGroup = DEFAULT_GROUP;

    // Parse out the values from each line in the preferences
    // file.  The file has a Windows ".INI" style format, where
    // the parameter name is followed by an equals sign and the
    // value of that parameter.  Stuff each read value into the
    // member variable where it belongs.

    while (!tsPrefs.atEnd()) 
    {
        line = tsPrefs.readLine();

        /* Comments are only maintained at the beginning of the file. */
        if (line.startsWith("#"))
        {
            strlComments += line;

            continue;
        }

        // Get the parameter value from the line
        QStringList parts = line.split("=", QString::SkipEmptyParts);
        
        if (line.startsWith(PREF_COLOR_CODING))
        {
            colorsEnabled = parts[1].toInt();

            continue;
        }

        if (line.startsWith(PREF_WIDTH))
        {
            width = parts[1].toInt();

            /* Cap the width to the screen width */
            if (width > scrWidth)
                width = scrWidth;

            /* Try to prevent some insane values */
            if (width < 100)
                width = scrWidth/2;

            continue;
        }

        if (line.startsWith(PREF_HEIGHT))
        {
            height = parts[1].toInt();

            /* Cap the height to the screen height */
            if (height > scrHeight)
                height = scrHeight;

            /* Try to prevent some insane values */
            if (height < 100)
                height = scrHeight/2;

            continue;
        }

        if (line.startsWith(PREF_COMMON_TASKS_VISIBLE))
        {
            commonTasksVisible = parts[1].toInt();

            continue;
        }

        if (line.startsWith(PREF_FIND_REQUEST_VISIBLE) ||
            line.startsWith(PREF_INCREMENTAL_SEARCH_VISIBLE))
            /* PREF_INCREMENTAL_SEARCH_VISIBLE is deprecated and will eventually
               be removed. */
        {
            findRequestVisible = parts[1].toInt();

            continue;
        }

        if (line.startsWith(PREF_STATE_FILTERS_VISIBLE) ||
            line.startsWith(PREF_STATE_FILTER_VISIBLE))
            /* PREF_STATE_FILTER_VISIBLE is deprecated and will eventually
               be removed. */
        {
            stateFiltersVisible = parts[1].toInt();

            continue;
        }

        if (line.startsWith(PREF_DATE_RANGE_VISIBLE))
        {
            dateRangeVisible = parts[1].toInt();

            continue;
        }

        if (line.startsWith(PREF_INTERFACE_STATUSES_VISIBLE))
        {
            interfaceStatusesVisible = parts[1].toInt();

            continue;
        }

        if (line.startsWith(PREF_STATUSBAR_DATE_VISIBLE))
        {
            statusbarDateVisible = parts[1].toInt();

            continue;
        }

        if (line.startsWith(PREF_STATUSBAR_RECORDS_RETURNED_VISIBLE))
        {
            statusbarRecordsReturnedVisible = parts[1].toInt();

            continue;
        }

        if (line.startsWith(PREF_STATUSBAR_SYSTEM_STATUS_VISIBLE))
        {
            statusbarSystemStatusVisible = parts[1].toInt();

            continue;
        }

        if (line.startsWith(PREF_REFRESH_TIMER))
        {
            strRefreshTimer = parts[1];

            continue;
        }

        if (line.startsWith(PREF_PRODUCT_REQUEST_DURATION))
        {
            strPRDuration = parts[1];

            continue;
        }

        if (line.startsWith(PREF_EVENT_LOG_DURATION))
        {
            strEventLogDuration = parts[1];

            continue;
        }

        if (line.startsWith(PREF_FONT))
        {
            strFont = parts[1];

            continue;
        }

        if (line.startsWith(PREF_SIZE))
        {
            strSize = parts[1];

            continue;
        }

        if (line.startsWith(PREF_SCROLLBACK_LINES))
        {
            strScrollbackLines = parts[1];

            continue;
        }

        if (line.startsWith(PREF_COLOR_SCHEME))
        {
            strColorScheme = parts[1];

            continue;
        }

        if (line.startsWith(PREF_COLOR_SWITCH))
        {
            strColorSwitch = parts[1];

            continue;
        }

        if (line.startsWith(PREF_GROUP_NAME)) 
        {
            currGroup = parts[1];

            strlOrderedGroupNames += currGroup;

            continue;
        }

        if (line.startsWith(PREF_GROUP_STATES)) 
        {
            QStringList stateList = 
                parts[1].split(",", QString::SkipEmptyParts);

            QStringListIterator iterator(stateList);
            while (iterator.hasNext())
            {
                group[iterator.next()] = currGroup;
            }

            continue;
        }

        if (line.startsWith(PREF_GROUP_BACKGROUND_COLOR)) 
        {
            QString strColor = parts[1];
            QString strRed = strColor.mid(1, 2);
            QString strGreen = strColor.mid(3, 2);
            QString strBlue = strColor.mid(5, 2);

            bool ok;

            backgroundColor[currGroup] = QColor(
                strRed.toInt(&ok, 16),
                strGreen.toInt(&ok, 16),
                strBlue.toInt(&ok, 16));

            continue;
        }

        if (line.startsWith(PREF_GROUP_FOREGROUND_COLOR)) 
        {
            QString strColor = parts[1];
            QString strRed = strColor.mid(1, 2);
            QString strGreen = strColor.mid(3, 2);
            QString strBlue = strColor.mid(5, 2);

            bool ok;

            textColor[currGroup] = QColor(
                strRed.toInt(&ok, 16),
                strGreen.toInt(&ok, 16),
                strBlue.toInt(&ok, 16));

            continue;
        }

        if (line.startsWith(PREF_FILTER_NAME)) 
        {
            currentFilterName = parts[1];

            strlOrderedFilterNames += currentFilterName;

            if (currFilter == "") // Only set to the first one
                currFilter = currentFilterName;

            continue;
        }

        if (line.startsWith(PREF_FILTER_STATES))
        {
            mapFilterNameToStates[currentFilterName] = parts[1];

            continue;
        }
    }

    filePreferences.close();

    return true;
}

/*****************************************************************************/

bool Preferences::writePreferencesFile()
{
    QFile filePreferences(strPreferencesFilename);

    if (filePreferences.open(QIODevice::WriteOnly)) 
    {
        QTextStream tsPrefs(&filePreferences);

        /* Write out all of the comments. */
        QStringListIterator commentIterator(strlComments);
        while (commentIterator.hasNext())
        {
            tsPrefs << commentIterator.next() << endl;
        }

        tsPrefs << endl << endl;

        /* Write out all of the general parameters.  These include the
           refresh timer and Event Log/Product Request durations, default
           toolbars, and X terminal options. */

        tsPrefs << "[general]" << endl << endl;

        tsPrefs << PREF_WIDTH << "=" << width << endl;
        tsPrefs << PREF_HEIGHT << "=" << height << endl;
        tsPrefs << PREF_REFRESH_TIMER << "=" << strRefreshTimer << endl;
        tsPrefs << PREF_PRODUCT_REQUEST_DURATION << "="
                << strPRDuration << endl;
        tsPrefs << PREF_EVENT_LOG_DURATION << "="
                << strEventLogDuration << endl;
        tsPrefs << PREF_COMMON_TASKS_VISIBLE << "="
                << commonTasksVisible << endl;
        tsPrefs << PREF_FIND_REQUEST_VISIBLE << "="
                << findRequestVisible << endl;
        tsPrefs << PREF_STATE_FILTERS_VISIBLE << "="
                << stateFiltersVisible << endl;
        tsPrefs << PREF_DATE_RANGE_VISIBLE << "=" << dateRangeVisible << endl;
        tsPrefs << PREF_INTERFACE_STATUSES_VISIBLE << "=" <<
            interfaceStatusesVisible << endl;
        tsPrefs << PREF_STATUSBAR_DATE_VISIBLE << "="
                << statusbarDateVisible << endl;
        tsPrefs << PREF_STATUSBAR_RECORDS_RETURNED_VISIBLE << "="
                << statusbarRecordsReturnedVisible << endl;
        tsPrefs << PREF_STATUSBAR_SYSTEM_STATUS_VISIBLE << "="
                << statusbarSystemStatusVisible << endl;
        tsPrefs << PREF_FONT << "=" << strFont << endl;
        tsPrefs << PREF_SIZE << "=" << strSize << endl;
        tsPrefs << PREF_SCROLLBACK_LINES << "=" << strScrollbackLines << endl;
        tsPrefs << PREF_COLOR_SCHEME << "=" << strColorScheme << endl;
        tsPrefs << PREF_COLOR_SWITCH << "=" << strColorSwitch << endl;
    
        if (colorsEnabled)
            tsPrefs << PREF_COLOR_CODING << "=1" << endl;
        else
            tsPrefs << PREF_COLOR_CODING << "=0" << endl;
        tsPrefs << endl << endl;

        // Next, write out all of the color groups.  Each color
        // group has four lines, the first being the name of the
        // group as displayed by the GUI, the second being a comma
        // separated list of the LPGS states in that group, the
        // third being the background color in "#RRGGBB" format
        // and the fourth being the text color (in the same format.)

        tsPrefs << "[colorGroups]" << endl;

        QStringListIterator groupIterator(strlOrderedGroupNames);
        while (groupIterator.hasNext())
        {
            QString groupName = groupIterator.next();

            tsPrefs << endl;
            tsPrefs << PREF_GROUP_NAME << "=" << groupName << endl;
            tsPrefs << PREF_GROUP_STATES << "=";

            // Write out a comma-separated list of the states.  If
            // this is the first state in the list.

            QStringListIterator stateIterator(group.keys());
            bool first = true;
            while (stateIterator.hasNext())
            {
                QString stateValue = stateIterator.next();
                if (group[stateValue] == groupName) 
                {
                    if (!first)
                        tsPrefs << ",";
                    else
                        first = false;

                    tsPrefs << stateValue;
                }
            }
            tsPrefs << endl;

            tsPrefs << PREF_GROUP_BACKGROUND_COLOR << "="
                    << backgroundColor[groupName].name() << endl;
            tsPrefs << PREF_GROUP_FOREGROUND_COLOR << "="
                    << textColor[groupName].name() << endl;
        }

        // Finally, dump out the state filter information.  These
        // appear similar to the color groups, except that there are
        // only two lines per filter (the filter name and the list
        // of states that belong to that filter.)

        tsPrefs << endl << endl;
        tsPrefs << "[quickFilters]" << endl;

        QStringListIterator filterIterator(strlOrderedFilterNames);
        while (filterIterator.hasNext())
        {
            QString filterName = filterIterator.next();

            tsPrefs << endl;
            tsPrefs << PREF_FILTER_NAME << "=" << filterName << endl;
            tsPrefs << PREF_FILTER_STATES << "="
                    << mapFilterNameToStates[filterName] << endl;
        }

        tsPrefs.flush();

        if (tsPrefs.status() != QTextStream::Ok)
        {
            return false;
        }

        filePreferences.flush();

        if (filePreferences.error() != QFile::NoError)
        {
            return false;
        }

        filePreferences.close();

        return true;
    }
    else
    {
        return false;
    }
}

