/*****************************************************************************
               Level 1 Product Generation System (LPGS) Software
******************************************************************************

  UNIT NAME: ProductRequestIncrementalSearch.cpp

  PURPOSE: To provide an incremental search in the MainWindow Product Request
           Listview.  The search allows the operator to enter a product
           request id into the line edit, and it will automatically select the
           first line in the listview that matches the entered portion of the
           product request id.  If the user enters enough characters for the
           underscore to be present, the underscore will be automatically
           insertied into the correct location.
 
*****************************************************************************/
// System Includes 

// Qt Includes 

// Local Includes 
#include "ProductRequestIncrementalSearch.h"

// Constants used internally
const int PR_BEFORE_UNDERSCORE = 13;
const int PR_AFTER_UNDERSCORE  = 14;
const int PR_LENGTH            = 19;

//*****************************************************************************
ProductRequestIncrementalSearch::ProductRequestIncrementalSearch(
    ProductRequestTreeWidget* twPR, QWidget* parent) 
  : QValidator(parent)
{
    ptwPR = twPR;
}
//*****************************************************************************
QValidator::State ProductRequestIncrementalSearch::validate(
    QString &strInput, // I:  string to validate
    int &pos) const    // I/0: current cursor position
{
    // This function is called for every key stroke entered in the
    // line edit widget.  If the output state value is Invalid it is
    // called a second time after the character is echoed to the line
    // edit widget.

    QString strPRID;  // holds the product reqeuest id of item in tree widget, 
                      // while comparing them to the entered value
    QValidator::State outputState; // output state from validator

    QRegExp rx("\\d{13}_\\d{5}");
    QRegExpValidator validator(rx, parent());

    int startPosition;
    int startLength;
    int currentLength;
    QString currentString;

    static int lastLength = strInput.length();
    static QString lastString = strInput;

    startPosition = pos;
    startLength = strInput.length();

    currentString = strInput;

    // First remove any non-digit characters
    currentString.remove(QRegExp("[\\D]"));

    // if the string is empty, don't bother with validation
    if (currentString.isEmpty())
    {
        strInput = "";
        return QValidator::Intermediate;
    }

    // Determine the length
    currentLength = currentString.length();

    // Add the underscore back in if needed
    if (currentLength > PR_BEFORE_UNDERSCORE)
    {
        currentString.insert(PR_BEFORE_UNDERSCORE, "_");
    }

    // Ensure that the product request id entered is not longer than its 
    // maximum length
    if (currentLength >= PR_LENGTH)
    {
        currentString = currentString.left(PR_LENGTH);
    }

    // We may have changed the length above so get it again
    currentLength = currentString.length();

    if (currentLength - lastLength < 0)
    {
        // We deleted something
    }
    else if (currentLength - lastLength > 0)
    {
        // We added something
        if (startPosition == PR_AFTER_UNDERSCORE)
        {
            // We added in front of the underscore, but need to move the
            // position
            pos++;
        }
    }
    else if (currentLength == lastLength)
    {
        // We remained the same, but should check and see if we tried to
        // delete around the underscore and advance the position
        if (startLength < lastLength)
        {
            // We hit the delete key or the backspace key
            // But since we can't determine which, do nothing
        }
        else if (currentString != lastString)
        {
            // We modified the string and should try to see if we need to
            // move the position because we attempted to add around the
            // underscore
            if (startPosition == PR_AFTER_UNDERSCORE)
            {
                // We tried to add something before the underscore but it got
                // placed after the underscore
                pos++;
            }
        }
    }

    lastLength = currentLength;
    lastString = strInput;
    strInput = currentString;

    outputState = validator.validate(currentString, pos);

    // The primary task of validating the product request entry has been 
    // performed at this point.  Next we want to use this customized version
    // to do an incremental search in the product requests list view for the
    // product request entered thus far.

    // The loop below looks for the first matching product request ID When a 
    // match is found it makes sure the item is visible and then marks the item
    // as selected.
    
    // This searches in the order shown in the list view, NOT in a sorted order
    // of product request ID values.  This is by design.  If the operator knows
    // a product request ID for a date and wants to veiw other product requests
    // near that date then he/she would sort the list view by date column and 
    // use this to find the first product request ID in the list with the list
    // view ordered by date.  This may be a less efficient search but may 
    // provide desired functionality.

    // create iterator to search through product request list view
    QTreeWidgetItemIterator twIterator(ptwPR);  

    while (*twIterator)
    {
        //  Column 0 (zero) contains the product request id string.
        strPRID = (*twIterator)->text(0);
        if (strPRID.startsWith(currentString))
        {
            ptwPR->scrollToItem(*twIterator);
            ptwPR->setCurrentItem(*twIterator);
            break;
        }

        ++twIterator;
    }
    
    // return the current validator output state
    return outputState;
}

