/******************************************************************************
               Level 1 Product Generation System (LPGS) Software
 ******************************************************************************
 *
 * UNIT NAME:   ProductRequestTreeWidget.h
 *
 * PURPOSE:
 * The ProductRequestTreeWidget class extends the generic QTreeWidget class by
 * providing a list of colored ProductRequestTreeWidgetItems, along with
 * context-sensitive popup menus that can be invoked with a right mouse
 * click.  From these menus, the operator is given access to several
 * options (which depend on what state the product request is in.)
 *
 *****************************************************************************/

// System Includes
#include <stdlib.h>
#include <unistd.h>
#include <iostream>
#include <signal.h>
#include <errno.h>

// Qt Includes 
#include <QtCore>
#include <QtGui>

// IAS Qt Includes
#include <iasQtWoConst.h>

// Local Includes
#include "AnomalyInformation.h"
#include "ColoredTreeWidgetItem.h"
#include "DatabaseUtils.h"
#include "IASErrorDialog.h"
#include "ErrorEmail.h"
#include "EventLog.h"
#include "FileViewer.h"
#include "Preferences.h"
#include "ProductRequestTreeWidget.h"
#include "SystemProcesses.h"
#include "UserParameters.h"
#include "Utilities.h"
#include "WorkOrderScripts.h"
#include "IASUtilities.h"
#include "IASSqlQuery.h"

// Constant to define retrieving all product request rows
const QString ALL_ROWS("ALL"); 

extern Preferences *preferences;

//*****************************************************************************

ProductRequestTreeWidget::ProductRequestTreeWidget(
    QWidget *parent,    // I: Parent widget
    DateRange *drMain)  // I: Date range for widget
  : QTreeWidget(parent)
{
    // Save pointer to date range object
    dr = drMain;

    // Set so that it resizes the first time.
    resizeColumns = true;
  
    // Setup QTreeWidget configuration options 
    setSizePolicy(QSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding));
    setSortingEnabled(true);
    setSelectionBehavior(QAbstractItemView::SelectRows);
    setSelectionMode(QAbstractItemView::ExtendedSelection);
    setRootIsDecorated(false);
    setContextMenuPolicy(Qt::PreventContextMenu);

    // Set the initial sorting    
    sortItems(0, Qt::AscendingOrder);

    // Initialize the context menu's
    buildContextMenus();

    // Below is a list of the column titles for the
    // main list view, in their default order.
    // Must match the productRequestColumns enumeration
    QStringList columnTitles;
    columnTitles << "Product Request ID"
                 << "Landsat Scene ID"
                 << "Instance"
                 << "State"
                 << "Current WO ID"
                 << "Priority"
                 << "Node"
                 << "Cancelled"
                 << "Date Received"
                 << "Satellite"
                 << "Sensor"
                 << "Procedure";

    // The following list specifies the horizontal alignment for the
    // information in each of the columns in the list view.  For some
    // of the columns, things look best if they are left-justified
    // within the column (for example, the Product Request ID.)  For
    // others, the information looks best if it is centered within the
    // column (for example, the 'Y' or 'N' flag for VQA needed.)
    // There should be one entry below for every column name above.
    // Must match the productRequestColumns enumeration
    int alignment[] = 
    {
        Qt::AlignLeft,      // Product Request ID
        Qt::AlignLeft,      // Landsat Scene ID
        Qt::AlignLeft,      // Processing Instance
        Qt::AlignLeft,      // State
        Qt::AlignLeft,      // Current WO ID
        Qt::AlignCenter,    // Priority
        Qt::AlignCenter,    // Node
        Qt::AlignCenter,    // Cancelled
        Qt::AlignLeft,      // Date Received
        Qt::AlignCenter,    // Satellite
        Qt::AlignCenter,    // Sensor
        Qt::AlignCenter,    // Procedure
    };

    // Generate the column header item
    QTreeWidgetItem *header = new QTreeWidgetItem();
    for (int i = 0; i < columnTitles.size(); ++i) 
    {
        header->setText(i, columnTitles[i]);
        header->setTextAlignment(i, alignment[i]);
    }
    setHeaderItem(header);

    // Initialize the filter values
    satelliteFilter = "All Satellites";
    sensorFilter = "All Sensors";
    procedureTypeFilter = "All Procedures";
    instanceFilter = "All Instances";
    setFocus();
    
    connect(this, SIGNAL(itemSelectionChanged()), 
        this, SLOT(refreshContextMenu()));
    connect(this, SIGNAL(itemClicked(QTreeWidgetItem *, int)),
        this, SLOT(copyPRIdToClipboard(QTreeWidgetItem *)));
}

//*****************************************************************************

void ProductRequestTreeWidget::buildContextMenus()
{
    popmWorkOrderScripts = new QMenu(this);
    popmWorkOrdersLogs = new QMenu(this);
    popmPriority = new QMenu(this);

    // Build the priority menu
    for (int currPriority = 1; currPriority < 10; currPriority++)
    {
        popmPriority->addAction(QString::number(currPriority),
            this, SLOT(reprioritize(bool)));
    }

    // Add the lower priority
    QMenu *prioritySubMenu = new QMenu("10-99", popmPriority);
    popmPriority->addMenu(prioritySubMenu);

    // Add the additional 10-99 sub menus
    for (int currPriority = 1; currPriority < 10; currPriority++)
    {
        QMenu *prioritySubSubMenu = new QMenu(QString::number(currPriority * 10)
            + " - " + QString::number(currPriority * 10 + 9), prioritySubMenu);

        for (int priority = 0; priority <= 9; priority++)
        {
            prioritySubSubMenu->addAction(QString::number((currPriority * 10)
                + priority), this, SLOT(reprioritize(bool)));
        }
        prioritySubMenu->addMenu(prioritySubSubMenu);
    }

    // Creating actions and adding them to the widget using addAction,
    // automatically creates a context menu where the shortcut keys work.
    // But since our menu options are based on the state of the selected
    // product request, we need to turn them on and off at certain points.
    // Initially (below) they are all turned off.  Also we need to set the
    // context policy at certain points so that an empty/incorrect menu and
    // options are not available.  One quirky thing, is that to turn off the
    // shortcut keys, you have to hide the action in addition to preventing
    // the context menu from appearing.  That is why we have a few chunks of
    // code that hide all of the actions.

    // Add the User Parameters menu option
    actionUserParameters = new QAction("User Parameters...", this);
    actionUserParameters->setShortcut(QKeySequence("Ctrl+U"));
    actionUserParameters->setShortcutContext(Qt::WidgetShortcut);
    addAction(actionUserParameters);
    connect(actionUserParameters, SIGNAL(triggered()),
            this, SLOT(showUserParameters()));
    actionUserParameters->setVisible(false);

    // Add the Display Event Log menu option
    actionDisplayEventLog = new QAction("Display Event Log", this);
    actionDisplayEventLog->setShortcut(QKeySequence("Ctrl+E"));
    actionDisplayEventLog->setShortcutContext(Qt::WidgetShortcut);
    addAction(actionDisplayEventLog);
    connect(actionDisplayEventLog, SIGNAL(triggered()),
            this, SLOT(displayEventLog()));
    actionDisplayEventLog->setVisible(false);

    // Add the Work Order Scripts menu option
    actionWorkOrderScripts = new QAction("Work Order Scripts...", this);
    actionWorkOrderScripts->setMenu(popmWorkOrderScripts);
    addAction(actionWorkOrderScripts);
    actionWorkOrderScripts->setVisible(false);

    // Add the Work Order log menu option
    actionWorkOrderLogs = new QAction("Work Order Log...", this);
    actionWorkOrderLogs->setMenu(popmWorkOrdersLogs);
    addAction(actionWorkOrderLogs);
    actionWorkOrderLogs->setVisible(false);

    // Add the Unix shell menu option
    actionOpenUnixShell = new QAction("Open UNIX Shell", this);
    actionOpenUnixShell->setShortcut(QKeySequence("Ctrl+S"));
    actionOpenUnixShell->setShortcutContext(Qt::WidgetShortcut);
    addAction(actionOpenUnixShell);
    connect(actionOpenUnixShell, SIGNAL(triggered()),
            this, SLOT(openUNIXShell()));
    actionOpenUnixShell->setVisible(false);

    // Add the Reprioritize menu option 
    actionChangePriority = new QAction("Reprioritize", this);
    actionChangePriority->setMenu(popmPriority);
    addAction(actionChangePriority);
    actionChangePriority->setVisible(false);

    actionSeparator1 = new QAction(this);
    actionSeparator1->setSeparator(true);
    addAction(actionSeparator1);
    actionSeparator1->setVisible(false);

    // Add the Retry Product Transfer menu option
    actionRetryProductTransfer = new QAction("Retry Product Transfer", this);
    actionRetryProductTransfer->setShortcut(QKeySequence("Ctrl+T"));
    actionRetryProductTransfer->setShortcutContext(Qt::WidgetShortcut);
    addAction(actionRetryProductTransfer);
    connect(actionRetryProductTransfer, SIGNAL(triggered()),
            this, SLOT(retryProductTransfer()));
    actionRetryProductTransfer->setVisible(false);

    // Add the Retry L0R Request menu option
    actionRetryL0RRequest = new QAction("Retry L0R Request", this);
    actionRetryL0RRequest->setShortcut(QKeySequence("Ctrl+L"));
    actionRetryL0RRequest->setShortcutContext(Qt::WidgetShortcut);
    addAction(actionRetryL0RRequest);
    connect(actionRetryL0RRequest, SIGNAL(triggered()), this, 
        SLOT(retryL0RRequest()));
    actionRetryL0RRequest->setVisible(false);

    // Add the Anomaly Information menu option
    actionAnomalyInformation = new QAction("Anomaly Information...", this);
    actionAnomalyInformation->setShortcut(QKeySequence("Ctrl+A"));
    actionAnomalyInformation->setShortcutContext(Qt::WidgetShortcut);
    addAction(actionAnomalyInformation);
    connect(actionAnomalyInformation, SIGNAL(triggered()),
            this, SLOT(displayAnomalyInformation()));
    actionAnomalyInformation->setVisible(false);

    actionSeparator2 = new QAction(this);
    actionSeparator2->setSeparator(true);
    addAction(actionSeparator2);
    actionSeparator2->setVisible(false);

    // Add the Resume Work order menu option
    actionResume = new QAction("Resume Work Order", this);
    actionResume->setShortcut(QKeySequence("Ctrl+R"));
    actionResume->setShortcutContext(Qt::WidgetShortcut);
    addAction(actionResume);
    connect(actionResume, SIGNAL(triggered()), this, SLOT(resumeRequest()));
    actionResume->setVisible(false);

    // Add the Fail Product Request menu option
    actionCancelPR = new QAction("Fail Product Request", this);
    actionCancelPR->setShortcut(QKeySequence("Ctrl+F"));
    actionCancelPR->setShortcutContext(Qt::WidgetShortcut);
    addAction(actionCancelPR);
    connect(actionCancelPR, SIGNAL(triggered()),
            this, SLOT(failCancelRequest()));
    actionCancelPR->setVisible(false);

    // Add the email selection for failed work orders menu option
    actionSeparator3 = new QAction(this);
    actionSeparator3->setSeparator(true);
    addAction(actionSeparator3);
    actionSeparator3->setVisible(false);

    actionSendEmail = new QAction("Send Email", this);
    actionSendEmail->setShortcut(QKeySequence("Ctrl+M"));
    actionSendEmail->setShortcutContext(Qt::WidgetShortcut);
    addAction(actionSendEmail);
    connect(actionSendEmail, SIGNAL(triggered()), this, SLOT(sendEmail()));
    actionSendEmail->setVisible(false);
}

//*****************************************************************************

void ProductRequestTreeWidget::refresh()
{
    IASSqlQuery query;
    int vScrollPos;
    int hScrollPos;
    QList<ColoredTreeWidgetItem *> selectedTreeWidgetItems;
    QList<QTreeWidgetItem *> listOfItems;

    // Compute the position of the item that is currently selected in
    // the list view.  This is so that after the database is re-queried
    // and the contents have been refreshed, we can go back and re-select
    // the item that the operator already had selected.  This is common
    // courtesy; it can also help avoid disasters stemming from the
    // operators bringing up a popup menu with the right mouse button
    // and then applying an option to an incorrect product request 
    // because there was a refresh during the menu operation

    // if dates selected in DateRange are invalid return
    if (!dr->isValid(true)) 
        return;

    vScrollPos = verticalScrollBar()->value();
    hScrollPos = horizontalScrollBar()->value();
    QStringList selectedItems = buildColumnList(getSelectedRows(ALL_ROWS),
        PRODUCT_REQUEST_ID);

    // Dynamically build the where clause for the sql command that populates
    // the listview. Start with the state and the date range from there add
    // on controller, sensor, and product type. If either of the last three
    // are not defined or they contain the word All then add 1=1 which 
    // effectively does nothing which is what filtering for All does.
    QString whereClause = "WHERE state IN (" + 
        preferences->quotedFilterStates(preferences->currFilter) + ")" + 
        " AND date_entered " + dr->getDateQueryString();

    if (!satelliteFilter.isEmpty() && 
        !(satelliteFilter.contains("All", Qt::CaseInsensitive) != 0))
    {
        whereClause += " AND satellite LIKE '" + satelliteFilter + "%'";
    }
 
    if (!sensorFilter.isEmpty() && 
        !(sensorFilter.contains("All", Qt::CaseInsensitive) != 0))
    {   
        whereClause += " AND sensor LIKE '" + sensorFilter + "%'";
    }

    if (!instanceFilter.isEmpty() &&
        !(instanceFilter.contains("All Instances", Qt::CaseInsensitive) != 0))
    {
        whereClause += " AND processing_instance = '" + instanceFilter + "'";
    }

    if (!procedureTypeFilter.isEmpty() && 
        !(procedureTypeFilter.contains("All", Qt::CaseInsensitive) != 0))
    {   
        whereClause += " AND procedure_id = '" + procedureTypeFilter + "'";
    }

    // discriminate against the anomaly orders
    whereClause += " AND wo_replaced != 1";

    // The order of the select is important, it must match the
    // productRequestColumns enumeration 
    QString sql =  "SELECT product_request_id,"
                         " scene_id,"
                         " processing_instance,"
                         " state,"
                         " work_order_id,"
                         " priority,"
                         " processing_node,"
                         " decode (cancellation_status, 1, 'Yes', 'No')"
                             " as cancellation_status,"
                         " date_entered,"
                         " satellite,"
                         " sensor,"
                         " procedure_id"
                   " FROM work_orders " + whereClause;

    // Display the wait cursor while we do a bunch of work
    QApplication::setOverrideCursor(Qt::WaitCursor);

    if (!query.exec(sql, __FILE__, __LINE__))
    {
        // restore the cursor 
        QApplication::restoreOverrideCursor();
        return;
    }

    setUpdatesEnabled(false);
    setSortingEnabled(false);

    clear();

    // Disable the menu options here because they will be reset later,
    // and because it is possible to have options available to a product
    // request that it should not have.  Disabling them fixes that.
    setContextMenuPolicy(Qt::PreventContextMenu);
    actionUserParameters->setVisible(false);
    actionDisplayEventLog->setVisible(false);
    actionChangePriority->setVisible(false);
    actionWorkOrderScripts->setVisible(false);
    actionWorkOrderLogs->setVisible(false);
    actionOpenUnixShell->setVisible(false);
    actionChangePriority->setVisible(false);
    actionSeparator1->setVisible(false);
    actionRetryProductTransfer->setVisible(false);
    actionRetryL0RRequest->setVisible(false);
    actionAnomalyInformation->setVisible(false);
    actionSeparator2->setVisible(false);
    actionResume->setVisible(false);
    actionCancelPR->setVisible(false);
    actionSeparator3->setVisible(false);
    actionSendEmail->setVisible(false);

    bool colorPref = preferences->areColorsEnabled();
    QString groupName; // Used for finding the right colors
    while (query.next())
    {
        ColoredTreeWidgetItem *currentTreeWidgetItem;
        QStringList columnValues;

        // Because we have all the columns in the select matching the columns
        // on the interface and enumeration we can just grab all of them at
        // once 
        columnValues = query.toStringList();

        // Get the correct group name before modifying the state to a string 
        if (colorPref) 
        {
            groupName = preferences->group[columnValues[STATE]];
        }
        columnValues[STATE] = stateToString(columnValues[STATE]);

        // Create a colored or B&W tree widget item
        if (colorPref) 
        {
            currentTreeWidgetItem = new ColoredTreeWidgetItem(
                preferences->textColor[groupName],
                preferences->backgroundColor[groupName], columnValues);
        } 
        else 
        {
            currentTreeWidgetItem = new ColoredTreeWidgetItem(
                Qt::black, Qt::white, columnValues);
        }
        
        // If any rows were selected before then check if this row was one
        // of the selected rows so it can be selected again later 
        if (selectedItems.size() > 0)
        {
            if (selectedItems.contains(columnValues[PRODUCT_REQUEST_ID], 
                Qt::CaseInsensitive))
            {
                selectedTreeWidgetItems.append(currentTreeWidgetItem);
            }
        }

        listOfItems.append(currentTreeWidgetItem);
    }
    addTopLevelItems(listOfItems);

    // Only resize all the columns to fit the data contained in the
    // QTreeWidget the first time.
    if (resizeColumns)
    {
        for (int i = 0; i < columnCount(); i++)
            resizeColumnToContents(i);

        resizeColumns = false;
    }

    setSortingEnabled(true);

    // Reset the selection to the way it was before the refresh began.
    // Note that this isn't perfect.  If new product requests come into
    // the system, the selection may be off a bit.  However, when you're
    // talking about an environment where items in the list can vanish
    // and appear arbitrarily, there's not much more that can be done.
    if (!selectedTreeWidgetItems.isEmpty())
    {
        setCurrentItem(selectedTreeWidgetItems.first());
    }

    // Set all the previous selected rows back to selected 
    foreach (ColoredTreeWidgetItem *item, selectedTreeWidgetItems)
    {
        item->setSelected(true);
    }
   
    verticalScrollBar()->setValue(vScrollPos);
    horizontalScrollBar()->setValue(hScrollPos);

    setUpdatesEnabled(true);

    // restore the cursor
    QApplication::restoreOverrideCursor();

    emit countChanged();
}

/*****************************************************************************/

void ProductRequestTreeWidget::refreshContextMenu()
{
    // This function creates the context sensitive menu that is displayed when
    // an item is right-clicked.  It is created separately from the right-click
    // signal since the keyboard shortcuts can change.

    // Retrieve all of the selected product request ids
    QList<PRODUCT_REQUEST_ROW> rows = getSelectedRows(ALL_ROWS);
    
    // Hide everything that is state dependent
    actionUserParameters->setVisible(false);
    actionDisplayEventLog->setVisible(false);
    actionChangePriority->setVisible(false);

    actionWorkOrderScripts->setVisible(false);
    actionWorkOrderLogs->setVisible(false);
    actionOpenUnixShell->setVisible(false);

    actionChangePriority->setVisible(false);

    actionSeparator1->setVisible(false);
    actionRetryProductTransfer->setVisible(false);
    actionAnomalyInformation->setVisible(false);
    actionRetryL0RRequest->setVisible(false);

    actionSeparator2->setVisible(false);
    actionResume->setVisible(false);
    actionCancelPR->setVisible(false);

    actionSeparator3->setVisible(false);
    actionSendEmail->setVisible(false);

    // If no item is selected, return from the function
    if (rows.isEmpty() || rows.first().product_request_id.isEmpty())
    {
        return;
    }

    // Turn the context menu back on 
    setContextMenuPolicy(Qt::ActionsContextMenu);

    // Clear the WO ID holding menus since they are repopulated
    popmWorkOrderScripts->clear();
    popmWorkOrdersLogs->clear();

    // Check all of the rows for single/multiple options
    if (rows.size() > 0)
    {
        bool failProductRequestFlag = false;
        bool retryProductTransferFlag = false;
        bool retryL0RRequestFlag = false;
        bool retryWorkOrderFlag = false;

        // Iterate through each of the selected rows and determine which
        // menu options need to be enabled
        foreach (PRODUCT_REQUEST_ROW row, rows)
        {
            // Check if product transfer can be retried
            if (!retryProductTransferFlag && row.state == DB_WO_TRANSFER_FAILED)
            {
                retryProductTransferFlag = true;
                actionSeparator1->setVisible(true);
                actionRetryProductTransfer->setVisible(true);
            } 

            // Check if L0R request can be retried 
            if (!retryL0RRequestFlag && row.state == DB_WO_REQUEST_FAILED)
            {
                retryL0RRequestFlag = true;
                actionSeparator1->setVisible(true);
                actionRetryL0RRequest->setVisible(true);
            }
            
            // Check if WO can be resumed 
            if (!retryWorkOrderFlag && row.state == DB_WO_HALTED)
            {
                // If the current work order has not been replaced 
                // allow the operator to resume it.
                if (!workOrderHasBeenReplaced(row.product_request_id, 
                    row.work_order_id))
                {
                    retryWorkOrderFlag = true;
                    actionSeparator2->setVisible(true);
                    actionResume->setVisible(true);
                }
            }

            // Check if the cancel option should be enabled 
            if (!failProductRequestFlag 
                && ((rows.size() > 1 && displayCancelMultiplePRMenu(row.state)) 
                || (rows.size() == 1 && displayCancelPRMenu(row.state))))
            {            
                failProductRequestFlag = true;
                actionSeparator2->setVisible(true);
                actionCancelPR->setVisible(true);
            }
        }
    }
    
    // Enable the single row menus
    if (rows.size() == 1)
    {
        PRODUCT_REQUEST_ROW row = rows.first(); // Get the only row selected 
      
        // Enable the ones which are always visible.
        actionUserParameters->setVisible(true);
        actionDisplayEventLog->setVisible(true);
        actionChangePriority->setVisible(true);

        // Load the work orders associated with this product request 
        strlWorkOrderIDs = workOrderID(row.product_request_id);

        // Start adding them back
        if (displayWOScriptsMenu(row.state))
        {
            if (strlWorkOrderIDs.count() > 0)
            {
                for (QStringList::Iterator it = strlWorkOrderIDs.begin();
                     it != strlWorkOrderIDs.end(); ++it)
                {
                    popmWorkOrderScripts->addAction((*it),
                        this, SLOT(displayWorkOrderScripts(bool)));
                }
            }
            
            actionWorkOrderScripts->setVisible(true);
        }
       
        if (displayWOLogMenu(row.state))
        {
            if (strlWorkOrderIDs.count() > 0)
            {
                for (QStringList::Iterator it = strlWorkOrderIDs.begin(); 
                     it != strlWorkOrderIDs.end(); ++it)
                {
                    popmWorkOrdersLogs->addAction((*it), 
                        this, SLOT(viewSelectedWorkOrderLog(bool)));
                }

                actionWorkOrderLogs->setVisible(true);
            }
        }

        if (displayOpenUnixShellMenu(row.state))
        {
            actionOpenUnixShell->setVisible(true);
        }

        if (displayChangePriorityMenu(row.state))
        {
            actionChangePriority->setVisible(true);
        }

        if (displayAnomalyInformationMenu(row.product_request_id, row.state))
        {
            // The states that make it here always have an anomaly
            actionSeparator1->setVisible(true);
            actionAnomalyInformation->setVisible(true);
        }

        // Enable the send email option
        if (row.state == DB_WO_HALTED)
        {
            actionSeparator3->setVisible(true);
            actionSendEmail->setVisible(true);
        }        
    }
}

/*****************************************************************************/

void ProductRequestTreeWidget::openUNIXShell()
{
    QString strWOId = currentItem()->text(WORK_ORDER_ID).trimmed();
    QString strPath;
    IASSqlQuery query;

    QString sqlCmd = "SELECT wo_path"
                     " FROM work_orders"
                     " WHERE work_order_id = '" + strWOId + "'";

    if (query.exec(sqlCmd, __FILE__, __LINE__))
    {
        query.next();
        strPath = query.value("WO_PATH").toString();
    }
    else
    {
        return;
    }

    // Change to the proper directory and start up an xterm
    // so that the operator can work there.  Exit directly after the
    // xterm is closed.
    QString cmd =  "cd " + strPath + " ; xterm -T \"LPGS\" " +
             preferences->strXtermOptions + " &";
    system(cmd.toAscii());
}

/*****************************************************************************/

void ProductRequestTreeWidget::reprioritize(bool)
{ 
    // All that is being done here is an update to the product_requests
    // table to reset the priority of the selected product request to
    // the value associated with the activated menu item.

    QAction *sender = static_cast<QAction*>(QObject::sender());
    
    QString strPRId = currentItem()->text(PRODUCT_REQUEST_ID).trimmed();
    QString strPriority = sender->text();

    QSqlDatabase::database().transaction();

    IASSqlQuery query;

    // Change all the WOs associated with this PR ID 
    QString sqlCmd = "UPDATE work_orders"
                     " SET priority = " + strPriority +
                     " WHERE product_request_id = '" + strPRId + "'";

    if (!query.exec(sqlCmd, __FILE__, __LINE__))
    {
        QSqlDatabase::database().rollback();
        return;
    }
            
    QSqlDatabase::database().commit();

    refresh();
}

//****************************************************************************

void ProductRequestTreeWidget::resumeRequest()
{
    QStringList resumeWOList; // List of work order ids to resume
    QStringList resumePRList; // List of request ids to resume
    QString message = "Resume the selected Product Requests that are in "
        "the \"Halted\" state?";
    bool needsRefresh = false; // Refresh menu indicator

    // Loop through all of the selectedRows (Single or Multiple) that
    // are in the HALTED state. Then check if they can be resumed and
    // try to reset them 
    foreach (PRODUCT_REQUEST_ROW row, getSelectedRows(DB_WO_HALTED))
    {
        // If the current work order has not been replaced then
        // allow the operator to resume it.
        if (!workOrderHasBeenReplaced(row.product_request_id, 
            row.work_order_id))
        {
            resumeWOList.append(row.work_order_id);
            resumePRList.append(row.product_request_id);
        }
    }

    // Check if ids should be included in message box
    if (resumePRList.size() < 10)
    {
        message = "Resume the Product Requests: \n" 
            + resumePRList.join("\n") + "?";
    }

    if (QMessageBox::warning(this, "Resume Product Request?", message, 
        QMessageBox::Yes | QMessageBox::No) != QMessageBox::Yes) 
    {
        return;
    }

    // Resume each of the work orders in the work order list
    foreach (QString work_order_id, resumeWOList)
    {
        QString errorMessage;
        if (IASUtilities::resetWOPending(work_order_id, errorMessage, true))
            needsRefresh = true;
    }
    
    // Refresh the menu if any work order has been reset
    if (needsRefresh)
        refresh();
}

//*****************************************************************************

void ProductRequestTreeWidget::copyPRIdToClipboard(QTreeWidgetItem *twiPR)
{
    // This function copies the product request id of the selected product
    // request to the clipboard.

    if (twiPR != NULL)
    {
        QClipboard *cb = QApplication::clipboard();

        switch (currentColumn())
        {
            case WORK_ORDER_ID:
                // If the WO ID column is chosen the copy that to the
                // clipboard. 
                cb->setText(twiPR->text(WORK_ORDER_ID).trimmed(),
                            QClipboard::Selection);
                break;
            default:
                // Original behaviour was to only copy the PR ID to the
                // clipboard everytime. 
                cb->setText(twiPR->text(PRODUCT_REQUEST_ID).trimmed(),
                            QClipboard::Selection);
                break;
        }
    }
}

//*****************************************************************************

void ProductRequestTreeWidget::displayAnomalyInformation()
{
    QString prId = currentItem()->text(PRODUCT_REQUEST_ID).trimmed();

    AnomalyInformation anomalyInfo(prId, this);
    anomalyInfo.exec();
}

/*****************************************************************************/

void ProductRequestTreeWidget::displayWorkOrderScripts(bool)
{
    QAction *sender = static_cast<QAction*>(QObject::sender());

    // Displays the Work Order Scripts dialog based upon the PR ID that is
    // selected.
    QString strPRID = currentItem()->text(PRODUCT_REQUEST_ID).trimmed();
    QString strWOID = sender->text();

    WorkOrderScripts workOrderScripts(strPRID, strWOID, this);
    workOrderScripts.exec();
}

/*****************************************************************************/

void ProductRequestTreeWidget::retryProductTransfer()
{
    // For WO's that are in a DB_WO_TRANSFER_FAILED state, attempt to
    // retry the transfer.  All that this really involves is resetting
    // the state of the WO to DB_WO_READY_TO_SHIP.  The capture_order will
    // retry the request when its next polling interval expires.
    QList<PRODUCT_REQUEST_ROW> transFailedList = getSelectedRows(
        DB_WO_TRANSFER_FAILED);
    QStringList strPRIds = buildColumnList(transFailedList, PRODUCT_REQUEST_ID);
    QStringList strWOIds = buildColumnList(transFailedList, WORK_ORDER_ID);
    QString message = "Retry transferring the selected Product Requests that "
        "are in the \"Transfer Failed\" state?";
    
    // Check if ids should be included in message box
    if (strPRIds.size() < 10)
    {
        message = "Retry transferring the Product Requests: \n" 
            + strPRIds.join("\n") + "?";
    }

    if (QMessageBox::warning(this, "Retry Product Transfer?", message, 
        QMessageBox::Yes | QMessageBox::No) != QMessageBox::Yes) 
    {
        return;
    }

    QString sqlCmd = "UPDATE work_orders"
                     " SET state = '" + DB_WO_READY_TO_SHIP + "'"
                     " WHERE work_order_id in ('" + strWOIds.join("','") + "')";
    QSqlDatabase::database().transaction();
    IASSqlQuery query;

    if (query.exec(sqlCmd, __FILE__, __LINE__))
    {
        QSqlDatabase::database().commit();
    }
    else
    {
        QSqlDatabase::database().rollback();
    }
    refresh();
}

/*****************************************************************************/

void ProductRequestTreeWidget::retryL0RRequest()
{
    // For WO's that are in a DB_WO_REQUEST_FAILED state, attempt to
    // retry the request. All that this really involves is resetting
    // the state of the WO to DB_WO_L0R_NEW. The capture_order will
    // retry the request when its next polling interval expires.
    QList<PRODUCT_REQUEST_ROW> requestFailedList = getSelectedRows(
        DB_WO_REQUEST_FAILED);
    QStringList strPRIds = buildColumnList(requestFailedList,
        PRODUCT_REQUEST_ID);
    QStringList strWOIds = buildColumnList(requestFailedList, WORK_ORDER_ID);
    QString message = "Retry requesting the L0R from subsetter for the "
        "selected Product Requests in the \"L0R Request Failed\" state?";
    
    // Check if ids should be included in message box
    if (strPRIds.size() < 10)
    {
        message = "Retry requesting the L0R from subsetter for Product "
            "Requests: \n" + strPRIds.join("\n") + "?";    
    }

    if (QMessageBox::warning(this, "Retry L0R Request?", message,
        QMessageBox::Yes | QMessageBox::No) != QMessageBox::Yes) 
    {
        return;
    }

    QString sqlCmd = "UPDATE work_orders"
                     " SET state = '" + DB_WO_L0R_NEW + "'"
                     " WHERE work_order_id in ('" + strWOIds.join("','") + "')";
    QSqlDatabase::database().transaction();
    IASSqlQuery query;

    if (query.exec(sqlCmd, __FILE__, __LINE__))
    {
        QSqlDatabase::database().commit();
    }
    else
    {
        QSqlDatabase::database().rollback();
    }
    refresh();
}

/*****************************************************************************/

void ProductRequestTreeWidget::failCancelRequest()
{
    // Cancel the selected product request.  Most of the actual work involved
    // in doing this is done behind the scenes by the LPGS daemons.  All that
    // really needs to be done by the OUI is a database update 

    QStringList failPRIds; // List of product request Ids to fail
    QStringList failWOIds; // List of work order Ids to fail
    QList<PRODUCT_REQUEST_ROW> rows = getSelectedRows(ALL_ROWS);

    // Loop through all of the selected rows and add the product request
    // id and work order id to the lists if they should be failed/cancelled
    foreach (PRODUCT_REQUEST_ROW row, rows)
    {
        bool displayCancelMenuFlag; // Display cancel option flag 

        // If more than 1 row use the multiple cancel menu mapping
        if (rows.size() > 1)
        {
            displayCancelMenuFlag = displayCancelMultiplePRMenu(row.state);
        }
        else
        {
            displayCancelMenuFlag = displayCancelPRMenu(row.state);
        }

        if (!displayCancelMenuFlag)
            continue;
    
        // Check passed this row should be failed/cancelled
        failPRIds.append(row.product_request_id);
        failWOIds.append(row.work_order_id);
    }
    
    // Fail the list of orders
    if (failPRIds.size() > 0)
        failCancelProductRequest(this, failPRIds, failWOIds);
    
    refresh();
}

/*****************************************************************************/

void ProductRequestTreeWidget::showUserParameters()
{
    // Display the User Parameters dialog for the selected product request.

    QString strPRId = currentItem()->text(PRODUCT_REQUEST_ID).trimmed();
    QString strWOId = currentItem()->text(WORK_ORDER_ID).trimmed();
    UserParameters userParameters(strPRId, strWOId, this, 0);

    userParameters.exec();
}

/*****************************************************************************/

void ProductRequestTreeWidget::viewSelectedWorkOrderLog(bool)
{
    QAction *sender = static_cast<QAction*>(QObject::sender());
    QString strWOId = sender->text();

    // This functions allows the operator to view the current WO log
    // for the selected product request id.  The WO log is located by
    // retrieving the wo_path from the work_orders table for the selected
    // PR/WO ID.

    // Get selected WO ID from list view
    QString strPath;

    QString sql = "SELECT wo_path"
                  " FROM work_orders "
                  " WHERE work_order_id = '" + strWOId + "'";

    IASSqlQuery query;

    if (query.exec(sql, __FILE__, __LINE__))
    {
        query.next();
        strPath = query.value("WO_PATH").toString();
    }
    else    
        return;

    // If the wo_path is empty, do not display
    if (strPath.isEmpty())
        return;

    // Create a complete path to the work order log.
    QString strWorkOrderLog = strPath + "/save/" + strWOId + ".log";

    QFile fileWorkOrderLog(strWorkOrderLog);

    if (! fileWorkOrderLog.exists())
    {
        QMessageBox::warning(this, "Can not find work order log",
            "The file \"" + strWorkOrderLog + "\" does not exist",
            QMessageBox::Ok);
        return;
    }

    FileViewer *fileviewer = new FileViewer(strWorkOrderLog, strPath);

    fileviewer->setAttribute(Qt::WA_DeleteOnClose);
    fileviewer->show();
}

//****************************************************************************

void ProductRequestTreeWidget::displayEventLog()
{
    QString strPRId = currentItem()->text(PRODUCT_REQUEST_ID).trimmed();
    EventLog eventLog(strPRId, this);

    eventLog.exec();
}

//****************************************************************************
void ProductRequestTreeWidget::setSatelliteFilter(const QString &filterText)
{
     satelliteFilter = filterText;    
}
//****************************************************************************
void ProductRequestTreeWidget::setSensorFilter(const QString &filterText)
{
     sensorFilter = filterText;    
}
//****************************************************************************
void ProductRequestTreeWidget::setInstanceFilter(const QString &filterText)
{
     instanceFilter = filterText;
}
//****************************************************************************
void ProductRequestTreeWidget::setProcedureTypeFilter(const QString &filterText)
{
     procedureTypeFilter = filterText;    
}

//****************************************************************************
void ProductRequestTreeWidget::sendEmail()
{
    // This provides an email widget for the operator with text
    // containing the WO directory of the failed product.
    // Get the WO directory path and bring up the email dialog

    QString strPRId = currentItem()->text(PRODUCT_REQUEST_ID).trimmed();
    QString strWOId = currentItem()->text(WORK_ORDER_ID).trimmed();
    QString strPath;
    QString sqlCmd;
    IASSqlQuery query;

    sqlCmd = "SELECT wo_path"
             " FROM work_orders"
             " WHERE work_order_id = '" + strWOId + "'";

    if (query.exec(sqlCmd, __FILE__, __LINE__))
    {
        query.next();
        strPath = query.value("WO_PATH").toString();
    }
    else
        return;
    
    ErrorEmail errorEmail(strPath);
    errorEmail.exec();
}

//****************************************************************************
QString ProductRequestTreeWidget::getCurrentProductRequestId() 
{
    QTreeWidgetItem *selected = currentItem();

    if (selected != NULL)
        return selected->text(PRODUCT_REQUEST_ID).trimmed();

    return QString();
}

//****************************************************************************
QList<PRODUCT_REQUEST_ROW> ProductRequestTreeWidget::getSelectedRows
(
    const QString stateFilter // I: State to filter from selected rows
)
{
    QList<PRODUCT_REQUEST_ROW> list; // Return product request list

    // Loop through each selected row   
    foreach (QTreeWidgetItem* item, selectedItems())
    {  
        PRODUCT_REQUEST_ROW row; // Current row 
      
        // Check if current row matches the filter state 
        if (stateFilter != ALL_ROWS && stateFilter 
            != stringToState(item->text(STATE)))
            continue;

        // Add the row to the return list
        row.state = stringToState(item->text(STATE).trimmed());
        row.work_order_id = item->text(WORK_ORDER_ID).trimmed();
        row.product_request_id = item->text(PRODUCT_REQUEST_ID).trimmed();
        list.append(row);
    }

    return list;
}

/****************************************************************************/
QStringList ProductRequestTreeWidget::buildColumnList
(
    const QList<PRODUCT_REQUEST_ROW> &rows, // I: List to join 
    const productRequestColumns joinColumn  // I: Column to join into list
)
{
    QStringList list; // Return list of strings

    // Loop through each of the rows append the correct column to the return 
    // string list 
    foreach (PRODUCT_REQUEST_ROW row, rows)
    {
        // Only the columns in the PRODUCT_REQUEST_ROW struct are supported
        switch (joinColumn)
        {
            case PRODUCT_REQUEST_ID: 
                list.append(row.product_request_id);
                break;
            case STATE: 
                list.append(row.state); 
                break;
            case WORK_ORDER_ID: 
                list.append(row.work_order_id); 
                break;
            default: 
                break;
        }
    }

    return list;
}
