/******************************************************************************
               Level 1 Product Generation System (LPGS) Software
******************************************************************************
 *
 * UNIT NAME:   RecordSetViewer.cpp
 *
 * PURPOSE:
 * This class creates a dialog which allows the operator to see and modify the
 * contents of a RecordSet.  All of the contents of the RecordSet will be 
 * viewable, and the SQL command will be editable.  The operator will be able
 * to requery the database with the new values.
 *
 *****************************************************************************/
// System Includes 

// Qt Includes 
#include <QtCore>
#include <QtGui>

// Local Includes
#include "RecordSetViewer.h"
#include "DatabaseUtils.h"
#include "IASSqlQuery.h"
#include "IASUtilities.h"

//*****************************************************************************

RecordSetViewer::RecordSetViewer(QWidget *parent, Qt::WindowFlags fl) 
    : QDialog(parent, fl)
{
    setupUi(this);

    // initialize the printer
    printer = NULL;

    // clear the result tree widget header 
    tw->setHeaderLabel("");

    // set keyboard focus to the multi-line sql command entry widget
    mleSQLCommand->setFocus();

    connect(pbDismiss, SIGNAL(clicked()), this, SLOT(accept()));
    connect(pbRefresh, SIGNAL(clicked()), this, SLOT(refresh()));
    connect(pbClearSQLCmd, SIGNAL(clicked()), this, SLOT(clearSQLCommand()));
    connect(pbPrint, SIGNAL(clicked()), this, SLOT(print()));
    connect(cbFriendlyColumnNames, SIGNAL(toggled(bool)), 
        this, SLOT(switchColumnHeaders()));
    connect(pbExecute, SIGNAL(clicked()), this, SLOT(refresh()));
    connect(pbCommitCmd, SIGNAL(clicked()), this, SLOT(commit()));
}

//*****************************************************************************

RecordSetViewer::~RecordSetViewer()
{
    delete printer;
}

//*****************************************************************************

bool RecordSetViewer::askForCommit()
{
    QMessageBox::StandardButton response =
        QMessageBox::warning(this, "Database Modifications Not Committed",
            "Database modifications have not been committed.\n"
            "Would you like to commit them?",
            QMessageBox::Yes | QMessageBox::No | QMessageBox::Cancel, 
            QMessageBox::Cancel);

    switch (response)
    {
        case QMessageBox::Yes:
            commit();
            return (lblReturnValue->text() == "SUCCESS");
            break;

        case QMessageBox::No:
            if (QSqlDatabase::database().rollback())
            {
                pbCommitCmd->setEnabled(false);
                return true;
            }
            else
            {
                QSqlError error = QSqlDatabase::database().lastError();
                QMessageBox::warning (this, "Rollback Error",
                "Oracle reported an error while rolling back this transaction:"
                "\n" + error.databaseText().trimmed());
                // Once the error is acknowledged, let them close the window
                return true;
            }
            break;

        case QMessageBox::Cancel:
            // fall through to default
        default:
            return false;
            break;
    }
}

//*****************************************************************************

void RecordSetViewer::closeEvent(QCloseEvent *event)
{
    bool ok;
    if (inTransaction(ok))
    {
        if (askForCommit())
        {
            event->accept();
        }
        else
        {
            event->ignore();
        }
    }
    else
    {
        event->accept();
    }
}

//*****************************************************************************

void RecordSetViewer::accept()
{
    bool ok;
    if (inTransaction(ok))
    {
        if (!askForCommit())
        {
            return;
        }
    }

    QDialog::accept();
}

//*****************************************************************************

void RecordSetViewer::refresh()
{
    // This function re-executes the SQL command (query) and updates the
    // contents of the Record Set Viewer.  If the SQL command was edited,
    // these new values are used.

    // clear rows from result tree widget    
    tw->clear();
    tw->setColumnCount(0);

    // Reset the previous status line edits
    lblValidStatusVal->setText("");
    lblNumRowsVal->setText("");
    lblNumColumnsVal->setText("");
    lblReturnValue->setText("");
    lblOracleErrNumVal->setText("");
    lblOracleErrMsgVal->setText("");
    lblErrMsgVal->setText("");

    // Prepare the SQL Query
    QString strQuery = mleSQLCommand->toPlainText().simplified();
    /* remove any ';' from the tail end */
    strQuery.replace(QRegExp(";$"), "");

    /* if the query can be committed, start a transaction */
    if (isUpdateable(strQuery))
    {
       QSqlDatabase::database().transaction();
    }

    /* you must create the QSqlQuery object after the transaction
       I don't know why, but it doesn't work if you don't ?????????? */
    QSqlQuery queryObj;

    // Run and verify the results of the query
    if (queryObj.exec(strQuery))
    {
        lblReturnValue->setText("SUCCESS");
        lblValidStatusVal->setText("VALID SQL COMMAND");
    }
    else
    {
        lblReturnValue->setText("FAILURE");
        lblValidStatusVal->setText("SQL COMMAND NOT ALLOWED");

        // Get the last error information
        QSqlError error = queryObj.lastError();
        lblOracleErrNumVal->setNum(error.number());
        lblErrMsgVal->setText(error.driverText().trimmed());
        lblOracleErrMsgVal->setText(error.databaseText().trimmed());

        return;
    }

    lblValidStatusVal->adjustSize();

    QSqlRecord record = queryObj.record();
    int numColumns = record.count();

    // Set the number of columns in the query result
    lblNumColumnsVal->setNum(numColumns);

    // Setup the tree widget headers
    QStringList headers;
    for (int col = 0; col < numColumns; col++)
    {
        // check if user friendly column names should be used
        if (cbFriendlyColumnNames->isChecked())
            headers += DatabaseUtils::friendlyColumnName(record.fieldName(col));
        else
            headers += record.fieldName(col);
    }

    // set the header labels
    tw->setHeaderLabels(headers);

    // Add the Query results to the results tree widget
    QTreeWidgetItem *item = NULL;
    int resultCnt = 0;
    while (queryObj.next())
    {
        item = new QTreeWidgetItem(tw);
        for (int col = 0; col < numColumns; col++)
        {
            if (!queryObj.isNull(col))
                item->setText(col, queryObj.value(col).toString());
        }
        resultCnt++;
    }

    // set number of rows returned
    lblNumRowsVal->setNum(resultCnt);

    // results the columns to fit the content
    for (int col = 0; col < numColumns; col++)
        tw->resizeColumnToContents(col);

    // if columns exist, set last one selected, done mostly to remove
    // compiler warnings
    if (item != NULL)
        item->setSelected(true);

    // if the command is not a query, enable the commit button 
    if (!queryObj.isSelect())
    {
        bool ok;
        pbCommitCmd->setEnabled(inTransaction(ok));
    }

    // set keyboard focus to the multi-line sql command entry widget
    mleSQLCommand->setFocus();
}

//*****************************************************************************

void RecordSetViewer::switchColumnHeaders()
{
    // This functions updates the list view column names to display either
    // the friendly (modied) versions or the column names straight from
    // the datebase

    QTreeWidgetItem *header = tw->headerItem();

    QStringList labels;
    for (int col = 0; col < tw->columnCount(); col++)
    {
        // check if user friendly column names should be used
        if (cbFriendlyColumnNames->isChecked())
            labels += DatabaseUtils::friendlyColumnName(header->text(col));
        else
            labels += DatabaseUtils::databaseColumnName(header->text(col));
    }

    // set the header labels
    tw->setHeaderLabels(labels);
}

//*****************************************************************************

void RecordSetViewer::clearSQLCommand()
{
    // clears the SQL command from the SQL Command line edit entry widget

    lblValidStatusVal->clear();
    mleSQLCommand->clear();

    // set keyboard focus to the text edit sql command entry widget
    mleSQLCommand->setFocus();
}

//*****************************************************************************

void RecordSetViewer::commit()
{
    // slot connected to commit button or RecordSetViewer     
    bool return_status = QSqlDatabase::database().commit();

    if (return_status)
    {
        // set displayed return value
        lblReturnValue->setText("SUCCESS");
        pbCommitCmd->setEnabled(false);
    }
    else
    {
        // set displayed return value
        lblReturnValue->setText("FAILURE");

        // Get the last error information
        QSqlError error = QSqlDatabase::database().lastError();
        lblOracleErrNumVal->setNum(error.number());
        lblErrMsgVal->setText(error.driverText().trimmed());
        lblOracleErrMsgVal->setText(error.databaseText().trimmed());
    }
    lblValidStatusVal->clear();

}

//*****************************************************************************

void RecordSetViewer::print()
{
    // print contents of RecordSetViewer Dialog     
    const int margin = 10;
    int pageNo = 1;                 // start at page 1
    int yPos = 0;                   // y position for each line
    QString strTemp;
    QString strPrinter;             // selected printer
    QPainter p;                     // object to paint to printer

    if (printer == NULL)
        printer = new QPrinter();         // printer object

    printer->setPageSize(QPrinter::Letter);
    printer->setOrientation(QPrinter::Landscape);
    printer->setColorMode(QPrinter::GrayScale);

    // get default printer
    strPrinter = getenv("LPDEST");
    if(!strPrinter.isEmpty())
        printer->setPrinterName(strPrinter);

    QPrintDialog printDialog(printer, this);
    if (printDialog.exec() == QDialog::Accepted)
    { 
        // if operator doesn't select OK to continue, exit without printing
        if(!p.begin(printer))
            return;

        p.setFont(QFont("courier", 8));
        QFontMetrics fm = p.fontMetrics();

        // print header
        strTemp.insert(0,"\t\t\t\t\tSQL EDITOR PRINTOUT");

        p.drawText(margin, margin + yPos, printer->width(),
            fm.lineSpacing(), Qt::TextExpandTabs | Qt::TextDontClip, strTemp);

        yPos += 2 *  fm.lineSpacing();     

        QApplication::setOverrideCursor(Qt::WaitCursor);

        // print out event list view
        TreeWidgetUtilities::printTreeWidget(tw, p, printer, margin, 
            yPos, pageNo);

        QApplication::restoreOverrideCursor();  
    }
}

//*****************************************************************************

bool RecordSetViewer::isUpdateable(
    const QString strCommand)   // I: SQL command to verify if allowed or not
{
    // This function determines if the commit button should be enabled or not
    int index;
    int i = 0;

    // create a list of allowable SQL commands
    char cmd_list[][20]={"^UPDATE","^INSERT", "^DELETE"};

    // convert SQL command to upper case so it is easier to check for
    // valid SQL command. NOTE:  SQL will accept mixed case comands
    QString strUpperCmd = strCommand.toUpper();

    // create a list of regular expressions from list of valid commands
    QRegExp regExpType(cmd_list[i]);

    // verify valid sql command 
    do { 
        regExpType.setPattern(cmd_list[i]);
        index = strUpperCmd.indexOf(regExpType);
        // if command not found
        if (index == -1)
            i++;
        else
            break;
    } while (index == -1 && i < int(ARRAY_SIZE(cmd_list)));

    // if the valid command was not found return false
    if(index == -1)
        return false; 
    else
        return true;
}

//*****************************************************************************

bool RecordSetViewer::inTransaction(bool &queryok)
{
    IASSqlQuery query;
    if (query.exec("SELECT dbms_transaction.local_transaction_id from DUAL",
        __FILE__, __LINE__))
    {
        if (query.next())
        {
            queryok = true;
            return !query.isNull(0);
        }
    }
    queryok = false;
    // If that doesn't work, assume we are in a transaction (just in case we
    // are).
    return true;
}

//*****************************************************************************
