/******************************************************************************
               Level 1 Product Generation System (LPGS) Software
 ******************************************************************************
 *
 * UNIT NAME:   SystemConfiguration.cpp
 *
 * PURPOSE:
 * This is the definition of the SystemConfiguration class used to view and
 * modify certain system configuration parameters for LPGS.
 *
 *****************************************************************************/

/* System Includes */
#include <iostream>
#include <math.h>

/* Qt Includes */
#include <QtCore>
#include <QtGui>

/* Local Includes */
#include "DatabaseUtils.h"
#include "IASErrorDialog.h"
#include "IASSqlQuery.h"
#include "SystemConfiguration.h"

//*************************************************************************

SystemConfiguration::SystemConfiguration( 
    QWidget* parent,    // I: The address to this widget's parent 
    Qt::WindowFlags fl) // I: Used by QT
: QDialog(parent, fl)
{
    setupUi(this);

    // This is the constructor for the SystemConfiguration Object.  It
    // initializes all of the proper variables needed by the object within
    // its scope, retrieves the necessary values from the databse, and displays
    // the values in the GUI for the user.

    // Set up the local structs
    initLocalStructs();

    // Populate the form with the configuration values from the database
    refreshValues(); 

    // Set validator to prevent invalid characters in the form
    //   The min,max values were set based on the database constraints on the
    //   sys_parms table.
    QIntValidator *pollingIntervalVal = new QIntValidator(10, 99999, this);
    leWorkOrderSchedulerInterval->setValidator(pollingIntervalVal);
    leSubsetterInterfaceInterval->setValidator(pollingIntervalVal);
    leShutdownPollingInterval->setValidator(pollingIntervalVal);
    leResourceMonitorPollingInterval->setValidator(pollingIntervalVal);
    leFileDeletionInterval->setValidator(pollingIntervalVal);
    leCaptureOrderInterval->setValidator(pollingIntervalVal);

    leDXLMaxConcurrentTxfr->setValidator(new QIntValidator(1, 99, this));

    QIntValidator *woProductSizeVal = new QIntValidator(this);
    woProductSizeVal->setBottom(0);
    leWorkOrderProductSize->setValidator(woProductSizeVal);
    
    connect(pbCancel, SIGNAL(clicked()), this, SLOT(reject()));
    connect(pbApply, SIGNAL(clicked()), this, SLOT(applyChanges()));
    connect(leWorkOrderSchedulerInterval, SIGNAL(textChanged(QString)), 
        this, SLOT(valueChanged()));
    connect(leFileDeletionInterval, SIGNAL(textChanged(QString)), 
        this, SLOT(valueChanged()));
    connect(leShutdownPollingInterval, SIGNAL(textChanged(QString)), 
        this, SLOT(valueChanged()));
    connect(leResourceMonitorPollingInterval, SIGNAL(textChanged(QString)), 
        this, SLOT(valueChanged()));
    connect(leFirstDiskAdvisoryLimit, SIGNAL(textChanged(QString)), 
        this, SLOT(valueChanged()));
    connect(leSecondDiskAdvisoryLimit, SIGNAL(textChanged(QString)), 
        this, SLOT(valueChanged()));
    connect(leDXLMaxConcurrentTxfr, SIGNAL(textChanged(QString)),
        this, SLOT(valueChanged()));
    connect(leCaptureOrderInterval, SIGNAL(textChanged(QString)),
        this, SLOT(valueChanged()));
    connect(leSubsetterInterfaceInterval, SIGNAL(textChanged(QString)),
        this, SLOT(valueChanged()));
    connect(pbOK, SIGNAL(clicked()), this, SLOT(applyChanges()));
    connect(pbOK, SIGNAL(clicked()), this, SLOT(accept()));
    connect(leWorkOrderProductSize, SIGNAL(textChanged(QString)), 
        this, SLOT(valueChanged()));
}

//***********************************************************************

void SystemConfiguration::initLocalStructs()
{
    // This function initializes the values of the class's structs that
    // will hold the information about the widgets that are on the dialog.
    // This information is then used to select the proper fields from the
    // database and populate the dialog with the proper fields.  This should
    // only be called at the instantiation of the SystemConfiguration object.

    // See SystemConfiguration.h for the declaration and meaning of the
    // leInfo struct.
    leInfo temp_leInfo[]=
    {
        {leWorkOrderSchedulerInterval, lblWorkOrderSchedulerInterval,
         "WORK_ORDER_SCHEDULER_INTERVAL", number},
        {leFileDeletionInterval,lblFileDeletionInterval,
         "DELETION_INTERVAL", number},
        {leShutdownPollingInterval, lblShutdownPollingInterval,
         "PSI_POLLING_INTERVAL", number},
        {leResourceMonitorPollingInterval, lblResourceMonitorPollingInterval,
         "RESOURCE_MONITOR_INTERVAL", number},
        {leCaptureOrderInterval, lblCaptureOrderInterval,
         "DCO_POLLING_INTERVAL", number},
        {leSubsetterInterfaceInterval, lblSubsetterInterfaceInterval,
         "DSI_POLLING_INTERVAL", number},
        {leWorkOrderProductSize,lblWorkOrderProductSize,
         "EST_AVG_WO_SIZE", number},
        {leDXLMaxConcurrentTxfr,lblDXLMaxConcurrentTxfr,
         "DXL_MAX_CONCURRENT_TXFR", number},
 
    };

    // See SystemConfiguration.h for the declaration and meaning of the
    // lepInfo struct.
    lepInfo temp_lepInfo[]=
    {
        {leFirstDiskAdvisoryLimit, lblFirstDiskAdvisoryLimit,
         "DISK_FIRST_ADVISORY_LIMIT", number},
        {leSecondDiskAdvisoryLimit, lblSecondDiskAdvisoryLimit,
         "DISK_SECOND_ADVISORY_LIMIT", number},
    };

    // Set class variables
    num_allLineEdits = sizeof(temp_leInfo) / sizeof(temp_leInfo[0]);
    num_allLineEditPercents = sizeof(temp_lepInfo) / sizeof(temp_lepInfo[0]);

    // Initialize the class structs allLineEdits and allLineEditPercents
    for(int i = 0; i < num_allLineEdits; i++)
    {
        allLineEdits[i] = temp_leInfo[i];
        allLineEdits[i].leObject->setAlignment(Qt::AlignRight);
    }

    for(int j = 0; j < num_allLineEditPercents; j++)
    {
        allLineEditPercents[j] = temp_lepInfo[j];
        allLineEditPercents[j].leObject->setAlignment(Qt::AlignRight);
    }
}

//***********************************************************************

void SystemConfiguration::valueChanged()
{
    // This is a slot that is called every time a value is changed on the
    // GUI. This function enables the apply button and bolds the label 
    // associated with the changed value. 

    int counter;
    QFont labelFont; // Will retrieve the current font of the object and
                     // change to bold for anything that has changed.

    pbApply->setEnabled(TRUE); // Enable the apply button since a value has
                               // been changed now.

    // Update the proper values in the structs
    for (counter = 0; counter < num_allLineEdits; counter++)
    {
        if (allLineEdits[counter].leObject->isModified())
        {
            labelFont = allLineEdits[counter].lblObject->font();
            labelFont.setBold(TRUE);
            allLineEdits[counter].lblObject->setFont(labelFont);
        }
    }

    for (counter = 0; counter < num_allLineEditPercents; counter++)
    {
        if (allLineEditPercents[counter].leObject->isModified())
        {
            labelFont = allLineEditPercents[counter].lblObject->font();
            labelFont.setBold(TRUE);
            allLineEditPercents[counter].lblObject->setFont(labelFont);
        }
    }
}

//***********************************************************************

void SystemConfiguration::refreshValues()
{
    // This function will query the database for configuration values stored in
    // the sys_parms table and populate the form with the returned values.

    int i;
    QString strSqlSelect = "SELECT ";
    QString strSqlFrom = " FROM sys_parms";

    // Get the LineEdit values
    for (i = 0; i < num_allLineEdits; i++)
    {
        if (i > 0)
            strSqlSelect += ", ";

        strSqlSelect += allLineEdits[i].strDbColumn;
    }

    strSqlSelect += strSqlFrom;

    IASSqlQuery query;
    if (!query.exec(strSqlSelect, __FILE__, __LINE__))
    {
        return;
    }

    query.next(); // Position on the returned record
    for (i = 0; i < num_allLineEdits; i++)
    {
        allLineEdits[i].leObject->blockSignals(true);
        allLineEdits[i].leObject->setText(query.value(i).toString());
        allLineEdits[i].leObject->blockSignals(false);
    }

    // Get the LineEditPercent values
    strSqlSelect = "SELECT ";
    for (i = 0; i < num_allLineEditPercents; i++)
    {
        if (i > 0)
            strSqlSelect += ", ";

        strSqlSelect += allLineEditPercents[i].strDbColumn;
    }

    strSqlSelect += strSqlFrom;
    
    if (!query.exec(strSqlSelect, __FILE__, __LINE__))
    {
        return;
    }

    query.next();
    for (i = 0; i < num_allLineEditPercents; i++)
    {
        allLineEditPercents[i].leObject->blockSignals(true);
        allLineEditPercents[i].leObject->setText(query.value(i).toString());
        allLineEditPercents[i].leObject->blockSignals(false);
    }
}

//***********************************************************************

void SystemConfiguration::applyChanges()
{
    // This function will save configuration changes made in the GUI to the 
    // database.

    int counter;
    bool constraintViolated = false;
    bool hasErrors = false;
    QString strSqlUpdate, strLowerBound, strUpperBound;
    QStringList strlConstraints;
    QFont labelFont; // Will hold the font of the label and
                     // change it to Not Bold after applying changes

    //Update the proper values in the Database for only what has changed
    for (counter = 0; counter < num_allLineEdits; counter++)
    {
        if (allLineEdits[counter].leObject->isModified())
        {
            QString value = allLineEdits[counter].leObject->text();

            //If supposed to be a number, check to make sure it is a valid
            //number
            if (allLineEdits[counter].dtType == number 
                && !validNumber(value))
            {
                QMessageBox::warning (this, "Constraint Error", 
                    allLineEdits[counter].lblObject->text()
                    + " can not be updated\nThe value entered must be"
                    " a valid number.");
                hasErrors = true;
            }
            else
            {
                strSqlUpdate = "UPDATE sys_parms set " + 
                    allLineEdits[counter].strDbColumn + " = '" + 
                    value + "'";

                IASSqlQuery query;

                if (query.exec(strSqlUpdate, __FILE__, __LINE__))
                {
                    allLineEdits[counter].leObject->setModified(false);
                    labelFont = allLineEdits[counter].lblObject->font();
                    labelFont.setBold(FALSE);
                    allLineEdits[counter].lblObject->setFont(labelFont);
                }
                else
                {
                    if (allLineEdits[counter].dtType == number)
                    {
                        constraintViolated = checkConstraint("SYS_PARMS",
                            allLineEdits[counter].strDbColumn, value, 
                            allLineEdits[counter].lblObject->text(), 
                            number, FALSE);
                    } 
                    else
                    {
                        constraintViolated = FALSE;
                    }

                    if (!constraintViolated)
                    {
                        IASErrorDialog::databaseError(this, "A database error "
                            "occurred while updating SYS_PARMS", 
                            __FILE__,__LINE__);
                    }
 
                    hasErrors = true;
                }
            }
        }
    }

    float firstValue = leFirstDiskAdvisoryLimit->text().toFloat();
    float secondValue = leSecondDiskAdvisoryLimit->text().toFloat();

    for (counter = 0; counter < num_allLineEditPercents; counter++)
    {
        if (allLineEditPercents[counter].leObject->isModified())
        {
            QString value = allLineEditPercents[counter].leObject->text();

            if (allLineEditPercents[counter].strDbColumn 
                    == "DISK_FIRST_ADVISORY_LIMIT")
            {
                if (value.toFloat() > secondValue)
                {
                    QMessageBox::warning(this, "Advisory Limit Error",
                        "First disk advisory limit must be less than the "
                        "second disk advisory limit");
                    hasErrors = true;
                    continue;
                }
            }
            else if (allLineEditPercents[counter].strDbColumn 
                    == "DISK_SECOND_ADVISORY_LIMIT")
            {
                if (value.toFloat() < firstValue)
                {
                    QMessageBox::warning(this, "Advisory Limit Error",
                        "Second disk advisory limit must be greater than the "
                        "first disk advisory limit");
                    hasErrors = true;
                    continue;
                }
            }
            

            if (allLineEditPercents[counter].dtType == number 
                && !validNumber(value))
            {
                QMessageBox::warning (this, "Constraint Error", 
                    allLineEditPercents[counter].lblObject->text()
                    + " can not be updated\nThe value entered must be"
                    " a valid number.");
                hasErrors = true;
            }
            else    //Valid Number
            {
                strSqlUpdate = "UPDATE sys_parms set " 
                    + allLineEditPercents[counter].strDbColumn
                    + "= '" + value + "'";

                IASSqlQuery query;
                bool result = query.exec(strSqlUpdate, __FILE__, __LINE__);
                if (!result)
                {
                    //IF this is supposed to be a number, make sure any range
                    //constraint was not violated before assuming a different
                    //type of database error.
                    if (allLineEditPercents[counter].dtType == number)
                    {
                        constraintViolated = checkConstraint("SYS_PARMS",
                            allLineEditPercents[counter].strDbColumn, 
                            value, 
                            allLineEditPercents[counter].lblObject->text(),
                            number,TRUE);
                    }
                    else
                    {
                        constraintViolated = FALSE;
                    }
                

                    if (!constraintViolated)
                    {
                        IASErrorDialog::databaseError(this, "A database error "
                            "occurred while updating SYS_PARMS",
                            __FILE__, __LINE__);
                    }

                    hasErrors = true;
                }
                else
                {
                    allLineEditPercents[counter].leObject->setModified(false);
                    labelFont = allLineEditPercents[counter].lblObject->font();
                    labelFont.setBold(FALSE);
                    allLineEditPercents[counter].lblObject->setFont(labelFont);
                }
            } // Valid Number
        }
    }

    if (!hasErrors)
    {
        // All changes were saved. Disable the apply button
        pbApply->setEnabled(FALSE);
    }
}

//***********************************************************************

bool SystemConfiguration::checkConstraint(
    const QString strTableName,  // The table name that the column is in
    const QString strColumnName, // The column name to get constraint for
    const QString newValue,      // The new value that the user entered
    const QString strLabelText,  // The label text for this column
    datatype dtType,             // The datatype that the constraint is
    bool lineEditPercent)        // I: if TRUE, object is a line edit percent
{
    // This function checks for a range of constraints on the table and column
    // passed in.  If the constraint is violated, a message box appears and
    // TRUE is returned.
    // This function is only called if a database error occurs.
    QStringList strlConstraints;
    bool constraintViolated = FALSE;
    
    strlConstraints.clear();
    strlConstraints =
        DatabaseUtils::getConstraint(strTableName, strColumnName, dtType);
    
    if (strlConstraints.count() == 2 )
    {
        if (newValue.toFloat() < strlConstraints[0].toFloat() 
            || newValue.toFloat() > strlConstraints[1].toFloat())
        {
            if (lineEditPercent)
            {
                // if object is a line edit percent, it must be handled 
                // differently, because while the values are displayed as
                // an int with units of percent, the values are stored in the
                // database as floats [.01 to 1.00]

                char strMin[7];
                char strMax[7];
                sprintf(strMin, "%.6g", strlConstraints[0].toFloat() * 100);
                sprintf(strMax, "%.6g", strlConstraints[1].toFloat() * 100);
                QMessageBox::warning(this, "Constraint Error",
                    strLabelText + " can not be updated.\n"
                    "The value entered must be between "
                    + QString(strMin) + " and " + QString(strMax) + ".");
            }
            else
            {
                QMessageBox::warning(this, "Constraint Error",
                    strLabelText + " can not be updated.\n"
                    "The value entered must be between "
                    + strlConstraints[0] + " and " + strlConstraints[1] + ".");
            }
            constraintViolated = TRUE;
        }
        else 
        {
            constraintViolated = FALSE;
        }
    }
    else
    {
        constraintViolated = FALSE;
    }

    return constraintViolated;  
}

//***********************************************************************

bool SystemConfiguration::validNumber(const QString strNumber)
{
    // This function returns TRUE if every character of the QString passed
    // in is a numeric character.

    bool returnValue;
    strNumber.toFloat(&returnValue);

    return returnValue;
}

