/******************************************************************************
               Level 1 Product Generation System (LPGS) Software
 ******************************************************************************
 *
 * UNIT NAME:   SystemProcesses.cpp
 *
 * PURPOSE:
 * This is the declaration file for the SystemProcesses class which provides
 * a window displaying the system processes and relevant information about
 * those processes to the user. 
 *
 *****************************************************************************/

/* System Includes */
#include <ctype.h>
#include <iostream>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/procfs.h>
#include <stdlib.h>
#include <dirent.h>
#include <pwd.h>
#include <fcntl.h>
#include <unistd.h>
#include <assert.h>
#include <signal.h>

/* Qt Includes */
#include <QtCore>
#include <QtGui>

/* Local Includes */
#include "SystemProcesses.h"
#include "IASUtilities.h"

//*****************************************************************************

SystemProcesses::SystemProcesses(QWidget *parent) // pointer to parent widget
: QDialog(parent)
{
    // Set up the user interface.
    setupUi(this);

    // This function constructs a SystemProcesses Dialog by calling the base
    // class's constructor and then populating the public listview with the
    // system processes and relevant information.

    twInfo tempLVI[] =
    {
        {"Process",Qt::AlignLeft,250,TRUE,TRUE,TRUE,-1,NULL},
        {"PID",Qt::AlignLeft,-1,TRUE,TRUE,TRUE,-1,NULL},
        {"User",Qt::AlignLeft,-1,TRUE,TRUE,TRUE,-1,NULL},
        {"PS",Qt::AlignLeft,-1,TRUE,TRUE,TRUE,-1,NULL},
        {"RSS",Qt::AlignLeft,-1,TRUE,TRUE,TRUE,-1,NULL},
    };

    numColumns = ARRAY_SIZE(tempLVI);

    twSystemProcesses->setRootIsDecorated(true);

    int column;
    for (column = 0; column < numColumns; column++)
    {
        twiSystemProcesses.push_back(tempLVI[column]);
    }

    TreeWidgetUtilities::buildTreeWidget(twSystemProcesses,
                                         twiSystemProcesses,
                                         numColumns);

    twSystemProcesses->setAllColumnsShowFocus(TRUE);
    pagesize = getpagesize();

    popmProcess = new QMenu(twSystemProcesses);

    connect(twSystemProcesses,
            SIGNAL(customContextMenuRequested(const QPoint&)),
            this, SLOT(showPopupMenu(const QPoint&)));

    popmKillProcess = new QMenu(twSystemProcesses);
    popmKillProcess->setTitle("Kill Process..");
    popmKillProcess->addAction("TERM");
    popmKillProcess->addSeparator();
    popmKillProcess->addAction("INT");
    popmKillProcess->addAction("QUIT");
    popmKillProcess->addAction("ABRT");
    popmKillProcess->addAction("KILL");

    popmProcess->addMenu(popmKillProcess);

    connect(popmKillProcess, SIGNAL(activated(int)),
            this, SLOT(killProcess(int)));

    connect(pbDismiss, SIGNAL(clicked()), this, SLOT(accept()));
    connect(pbRefresh, SIGNAL(clicked()), this, SLOT(refresh()));

    head = NULL;

    refresh();
}

SystemProcesses::~SystemProcesses()
{
    // Free up any memory that may be used
    deleteTree();
}

//*****************************************************************************

void SystemProcesses::refresh()
{
    // refreshes the list by deleting and recreating the tree, clearing the
    // public listview, and re-populating the listview with the new tree

    twSystemProcesses->clear();
    if (head != NULL)
    {
        deleteTree();
        head = NULL;
    }

    buildProcessTree();

    LPGS_PROC_NODE *nodeptr = findPid(1);
    buildTree(nodeptr, NULL, twSystemProcesses);

    /* Resize the columns so values are visible */
    for (int column = 0; column < twSystemProcesses->columnCount(); column++)
    {
        twSystemProcesses->resizeColumnToContents(column);
    }
}

//*****************************************************************************

void SystemProcesses::killProcess(int) // I: The index of the menu item
                                       //    that was selected,
                                       //    -TERM, -INT, etc...
{
    // Based on the selection from the popmKillProcess menu, the proper kill
    // process signal is sent to the operating system after the user verifies
    // that they wish to kill the process they selected.
    QAction *sender = static_cast<QAction*>(QObject::sender());

    pid_t processID;
    int killSignal;

    if (sender->text() == "TERM")
        killSignal = SIGTERM;
    else if (sender->text() == "INT")
        killSignal = SIGINT;
    else if (sender->text() == "QUIT")
        killSignal = SIGQUIT;
    else if (sender->text() == "ABRT")
        killSignal = SIGABRT;
    else if (sender->text() == "KILL")
        killSignal = SIGKILL;
    else
        killSignal = SIGTERM;

    processID = twSystemProcesses->currentItem()->text(1).toInt();

    if (QMessageBox::warning(this, "Kill Process",
        "Are you sure you want to kill the selected process?",
        QMessageBox::Yes,QMessageBox::No) == QMessageBox::Yes)
    {
        // USER Has selected yes that they want to kill the process
        if (kill(processID, killSignal) == 0)
        {
            // Give the system a second to kill the process before refreshing
            sleep(1);
            refresh();
        }
        else // The kill signal was unsuccessful
        {
            if (errno == EPERM)
            {
                QMessageBox::warning(this, "Kill Process Error", 
                    "You do not have the permission to kill this process");
            }
            else if (errno == ESRCH)
            {
                QMessageBox::warning(this, "Kill Process Error", 
                    "The process ID could not be found. Try clicking on\n"
                    "the refresh button to make sure thie process still "
                    "exists.");
            }
            else
            {
                QMessageBox::warning(this, "Kill Process Error", 
                    "The process could not be killed due to an unexpected "
                    "error.");
            }
        }
    }
}

//*****************************************************************************

void SystemProcesses::showPopupMenu(
    const QPoint &point) // I: Point within view that was clicked on
{
    // Displays the popup menu related to the process that was right-clicked on

    popmProcess->popup(twSystemProcesses->viewport()->mapToGlobal(point));
}

//*****************************************************************************

void SystemProcesses::buildTree(LPGS_PROC_NODE *nodeptr,
    QTreeWidgetItem *parent,
    QTreeWidget *tree)
{
    /* Traverses through the stored tree data strcuture that contains the
       process information and adds each process to the list view. */

    QString strLabel;
    QString strPid;
    QString strPSize;
    QString strRSize;
    struct passwd *pw;
    QTreeWidgetItem *listItem;
    QStringList itemText;

    strLabel = nodeptr->strName + " " + nodeptr->strArgs;
    strLabel.truncate(60);

    strPid.setNum(nodeptr->pid);

    if (nodeptr->procsize >= 10000) 
    {
        strPSize.setNum((double) nodeptr->procsize / 1024, 'g', 4);
        strPSize += "M";
    } 
    else 
    {
        strPSize.setNum(nodeptr->procsize);
        strPSize += "K";
    }

    if (nodeptr->rss >= 10000) 
    {
        strRSize.setNum((double) nodeptr->rss / 1024, 'g', 4);
        strRSize += "M";
    } 
    else 
    {
        strRSize.setNum(nodeptr->rss);
        strRSize += "K";
    }

    pw = getpwuid(nodeptr->uid);
    assert(pw);

    itemText << strLabel;
    itemText << strPid;
    itemText << pw->pw_name;
    itemText << strPSize;
    itemText << strRSize;

    if (parent == NULL)
    {
        listItem = new QTreeWidgetItem(tree, itemText);
        listItem->setExpanded(true);
    }
    else
    {
        listItem = new QTreeWidgetItem(parent, itemText);
    }

    if (nodeptr->num_children > 0) 
    {
        for (int i = 0; i < nodeptr->num_children; ++i)
            buildTree(nodeptr->children[i], listItem, tree);

        listItem->setExpanded(true);
    }
}

//*****************************************************************************

void SystemProcesses::addChild(pid_t ppid, LPGS_PROC_NODE *nodeptr)
{
    // Add a child process to the child list of an existing process.

    LPGS_PROC_NODE *curptr = findPid(ppid);

    if (curptr == NULL)
    {
        return;
    }
    else 
    {
        curptr->children.append(nodeptr);
        curptr->num_children += 1;
    }
}

//*****************************************************************************

LPGS_PROC_NODE *SystemProcesses::findPid(pid_t pid) 
{
    // Given an input PID, return a pointer to the linked list node
    // that corresponds to that process ID (or a NULL pointer if no
    // such node exists in the list.)

    LPGS_PROC_NODE *curptr = head;

    while (curptr != NULL) 
    {
        if (curptr->pid == pid)
        {
            return curptr;
        }
        curptr = curptr->next;
    }

    return NULL;
}

//*****************************************************************************

#ifdef LINUX

// NON-PORTABLE CODE: The code that reads the information present in the
// /proc pseudo-filesystem is inherently nonportable. Under IRIX, we can
// look at all of the files in the "/proc/pinfo" directory and use an
// ioctl to populate a prpsinfo structure in order to get information
// about a running process (i.e., parent PID, resident set size, etc.)
//
// Under Linux, however, "/proc/pinfo" doesn't exist, and neither does
// the associated ioctl. However, what *does* exist is a set of files
// named "/proc/<pid>/stat", which gives you almost all of the info you
// could possibly want about a running process. So instead of using an
// ioctl, we just parse the contents of this file.

void SystemProcesses::buildProcessTree()
{
    // Builds the stored tree structure of all the processes that
    // are running in the system.

    DIR *dirp;            // Directory pointer for "/proc"
    FILE *fp;             // File I/O pointer for "stat" and "cmdline"
    QString strFilename;  // Name of file to open
    struct dirent *res;   // Directory entry for "/proc" contents
    struct stat st;       // Information returned by stat() syscall

    char dummy_char;      // Dummy character for unneeded proc stats
    char dummy_str[1024]; // Dummy string for unneeded proc stats
    int dummy_int;        // Dummy int for unneeded proc stats

    LPGS_PROC_NODE *newnode;       // Pointer to new linked list node
    LPGS_PROC_NODE *curptr = NULL; // Pointer used to walk linked list

    if (chdir("/proc") < 0)
    {
        QMessageBox::warning(this, "Change Directory Failure",
            "chdir failed: " + QString(strerror(errno)));
        return;
    }

    dirp = opendir("/proc");
    if (dirp == NULL)
    {
        QMessageBox::warning(this, "Open Directory Failure",
            "opendir failed: " + QString(strerror(errno)));
        return;
    }

    // Iterate through each file in the "/proc" directory.
    while ((res = readdir(dirp)))
    {
        // There are more files in the "/proc" directory than just
        // process files (for example, "/proc/cpuinfo.") For this
        // reason, check to see if the first character in the filename
        // is a digit.  If so, assume that this file represents a
        // running process.  If not, skip it.
        if (!isdigit(res->d_name[0]))
            continue;

        // First call stat() on the file. This gives us, among other
        // things, the owner of the file, which corresponds to the
        // effective user ID that is executing the process.
        if (stat(res->d_name, &st) < 0) 
        {
            /* We don't need to issue an error because most-likely the
               process terminated.  Just move on to the next process. */
            continue;
        }

        // Create a new list node.
        newnode = new LPGS_PROC_NODE();
        if (newnode == NULL)
        {
            QMessageBox::warning(this, "Memory Allocation Failure",
                "new failed: " + QString(strerror(errno)));
            continue;
        }

        // Build the path to the "/proc/<PID>/cmdline" file, which
        // contains the exact command-line invocation that was used
        // to invoke the process. This is what we'll report.

        strFilename = "/proc/";
        strFilename += res->d_name;
        strFilename += "/cmdline";

        fp = fopen(strFilename.toAscii().data(), "r");
        if (fp == NULL)
        {
            /* We don't need to issue an error because most-likely the
               process terminated.  Just move on to the next process. */
            delete newnode;
            continue;
        }

        fgets(dummy_str, sizeof dummy_str, fp);
        newnode->strName = dummy_str;

        fclose(fp);

        // Now build the path to the "/proc/<PID>/stat" file, which
        // has all of the information that we need about the parent
        // PID and the memory usage.

        strFilename = "/proc/";
        strFilename += res->d_name;
        strFilename += "/stat";

        fp = fopen(strFilename.toAscii().data(), "r");
        if (fp == NULL)
        {
            /* We don't need to issue an error because most-likely the
               process terminated.  Just move on to the next process. */
            delete newnode;
            continue;
        }

        // Parse the "stat" file. The format of this file is described
        // in the Linux man page for proc(5). Note that there's a lot
        // of information here that we don't need, so it just gets
        // stuffed into a dummy variable and ignored.  This is kind of
        // ugly, but it gets the job done.
        fscanf(fp, "%d %s %c %d %d %d %d %d %u %u %u %u %u "
                   "%d %d %d %d %d %d %u %u %d %ld %ld",
               &newnode->pid,
               dummy_str,
               &dummy_char,
               &newnode->ppid,
               &dummy_int, &dummy_int, &dummy_int, &dummy_int, &dummy_int,
               &dummy_int, &dummy_int, &dummy_int, &dummy_int, &dummy_int,
               &dummy_int, &dummy_int, &dummy_int, &dummy_int, &dummy_int,
               &dummy_int, &dummy_int, &dummy_int,
               &newnode->procsize,
               &newnode->rss);

        // The virtual memory size is reported in bytes, and the
        // resident set size is reported in pages, so to get the proper
        // RSS we need to multiply it by the page size (determined
        // earlier.)  Then both need to get converted to kilobytes.

        newnode->procsize /= 1024;
        newnode->rss = newnode->rss * pagesize / 1024;

        newnode->uid = st.st_uid;
        newnode->next = head;
        newnode->num_children = 0;
        head = newnode;

        fclose(fp);
    }

    closedir(dirp);

    // Now we've built a linked list of all of the processes running on
    // the system.  What remains to be done is to go through each one,
    // figure out what its parent process is, and add a link to it in
    // that parent process' children array.  Once that is done, we will
    // be able to start at the node entry for process ID 1 (init) and
    // recurse our way down to every process on the system.

    curptr = head;
    while (curptr != NULL)
    {
        if (curptr->ppid > 0)
            addChild(curptr->ppid, curptr);

        curptr = curptr->next;
    }
}

#endif

//*****************************************************************************

#ifdef IRIX

// NON-PORTABLE CODE: The IRIX version of the /proc parser does the same
// thing that the Linux version does, but is a bit different since it
// (as has already been mentioned) uses an ioctl to extract process
// information about the files in the "/proc/pinfo" directory.

// WARNING: IRIX specific code has not been used for years and no IRIX systems
// were available during testing.  The code is being left in because it still
// looks valid based the man pages for IRIX 6.5.

void SystemProcesses::buildProcessTree()
{
    // Builds the stored tree structure of all the processes that
    // are running in the system.

    int fd;                  /* File descriptor of open process */
    prpsinfo_t prpsinfo;     /* Process info table              */
    DIR *dirp;               /* POSIX directory structure       */
    struct dirent *res;      /* POSIX directory entry           */
    LPGS_PROC_NODE *newnode; /* Pointer to new tree node        */
    LPGS_PROC_NODE *curptr;  /* Current position in list        */

    if (chdir("/proc/pinfo") < 0) 
    {
        perror("chdir");
        exit(EXIT_FAILURE);
    }

    dirp = opendir("/proc/pinfo");
    if (dirp == NULL)
    {
        perror("opendir");
        exit(EXIT_FAILURE);
    }

    while (res = readdir(dirp)) 
    {
        // readdir() will return the current and parent directories
        // ("." and "..") in its list.  Ignore these; we are only
        // interested in actual process files.

        if (res->d_name[0] == '.')
            continue;

        // Do not treat an open-file failure as a critical error; it
        // is possible that a process might have terminated between
        // the time that we read the directory listing and the time
        // we open the file.  This would result in the file no longer
        // being there.  Therefore, if there are problems opening a
        // process image, just skip on to the next one.

        fd = ::open(res->d_name, O_RDONLY);
        if (fd < 0)
            continue;

        if (ioctl(fd, PIOCPSINFO, &prpsinfo) < 0) 
        {
            perror("ioctl");
            exit(EXIT_FAILURE);
        }
    
        // Allocate memory for a new node in the linked list, and copy
        // selected information from the process table into it (such
        // as the parent process ID, the executable name, etc.)  Set
        // the link information (next node, number of children. etc)
        // to their proper initial values.

        newnode = new LPGS_PROC_NODE();
        if (newnode == NULL) 
        {
            perror("malloc");
            exit(EXIT_FAILURE);
        }

        newnode->pid = prpsinfo.pr_pid;
        newnode->ppid = prpsinfo.pr_ppid; 
        newnode->strName = prpsinfo.pr_fname;
        newnode->strArgs = prpsinfo.pr_psargs;
        newnode->uid = prpsinfo.pr_uid;
        newnode->procsize = prpsinfo.pr_size * pagesize / 1024;
        newnode->rss = prpsinfo.pr_rssize * pagesize / 1024;

        newnode->next = head;
        newnode->num_children = 0;
        head = newnode;

        // NOTE: The process viewer dialog inherits from QDialog, which
        // inherits from QWidget, which has a member function named
        // close()!  If we do not properly scope the function call
        // below with "::", we end up destroying our own widget!

        ::close(fd);
    }

    closedir(dirp);

    // Now we've built a linked list of all of the processes running on
    // the system.  What remains to be done is to go through each one,
    // figure out what its parent process is, and add a link to it in
    // that parent process' children array.  Once that is done, we will
    // be able to start at the node entry for process ID 1 (init) and
    // recurse our way down to every process on the system.

    curptr = head;
    while (curptr != NULL) 
    {
        if (curptr->ppid > 0)
            addChild(curptr->ppid, curptr);
        curptr = curptr->next;
    }
}

#endif

//*****************************************************************************

void SystemProcesses::deleteTree()
{
    // Deletes the entire contents of the stored tree strucutre and frees up
    // all of the memory it was using.

    LPGS_PROC_NODE *current;
    LPGS_PROC_NODE *next;

    current = head;
    while (current) 
    {
        next = current->next;
        delete current;
        current = next;
    }
}

//*****************************************************************************

