/******************************************************************************
               Level 1 Product Generation System (LPGS) Software
******************************************************************************
*
* UNIT NAME:   SystemTaskStatus.cpp
*
* PURPOSE:
* This is the class implementation for the SystemTaskStatus class used to
* display a list view of the status of the LPGS systems.
* 
*****************************************************************************/

/* System Includes */

/* Qt Includes */

/* Local Includes */
#include "IASSqlQuery.h"
#include "IASQtSystemStatus.h"
#include "SystemTaskStatus.h"
#include "ColoredTreeWidgetItem.h"

/* Define some color contants to use with the ColoredTreeWidgetItem */
const QColor TASK_UP_COLOR(68, 255, 68);
const QColor TASK_DOWN_COLOR(255, 68, 68);

//***********************************************************************

SystemTaskStatus::SystemTaskStatus(
    QWidget *parent,  // I: Pointer to parent widget
    Qt::WindowFlags fl)        // I: Used by Qt.
: QDialog(parent, fl)
{
    // Set up the user interface.
    setupUi(this);

    // This constructor instantiates the SystemTaskStatus object by calling
    // the base class's constructor.  It sets up the system status dialog
    // and gets the values from the database for the statuses of the 
    // different systems.

    twSystemTaskStatus->clear();
    twSystemTaskStatus->setRootIsDecorated(false);
    taskItems.clear();

    strSystemId = getenv("L1_SYSTEM_ID");
    if (strSystemId.isEmpty())
    {
        QMessageBox::critical(0, "Missing Environment",
            "Could not get L1_SYSTEM_ID from environment.\n"
            "This variable is required for system startup or shutdown",
            QMessageBox::Ok, 0);
        return;
    }

    strSystemId = strSystemId.toUpper();
    strSystemId = strSystemId.at(0);

    connect(pbDismiss, SIGNAL(clicked()), this, SLOT(accept()));
    connect(pbRefresh, SIGNAL(clicked()), this, SLOT(refresh()));

    refresh(); // load the current status of each task
}

//***********************************************************************

void SystemTaskStatus::refresh()
{
    // This function clears the treewidget for the SystemTaskStatus object and
    // executes the query to get the values needed again for the systems
    // and creates treewidget items for each system.

    IASSqlQuery query;

    QString sqlCmd = "SELECT task_id, proc_id, status_date"
                     " FROM tasks"
                     " WHERE system_id = '" + strSystemId + "'"
                     " ORDER BY task_id";

    if (query.exec(sqlCmd, __FILE__, __LINE__))
    {
        QStringList columnData;
        while (query.next())
        {
            QString taskName = query.value("TASK_ID").toString();
            QString processId = query.value("PROC_ID").toString();
            QString statusDate = query.value("STATUS_DATE").toString();

            /* Determine the status text and color */
            QString taskStatus = "DOWN";
            QColor statusColor = TASK_DOWN_COLOR;

            if (SystemStatus::isTaskRunning(taskName, processId))
            {
                taskStatus = "UP";
                statusColor = TASK_UP_COLOR;
            }

            QMap<QString, ColoredTreeWidgetItem*>::const_iterator itr =
                taskItems.find(taskName);
            if (itr == taskItems.end())
            {
                /* Add a new tree widget item populated with the results of
                   the query */
                columnData.clear();
                columnData << taskName;
                columnData << taskStatus;
                columnData << processId;
                columnData << statusDate;

                taskItems[taskName] = new ColoredTreeWidgetItem(
                    Qt::black, statusColor, twSystemTaskStatus, columnData);
            }
            else
            {
                /* Update the colors */
                itr.value()->setColorProperties(Qt::black, statusColor);
                /* Update the status text */
                itr.value()->setColoredText(1, taskStatus);
                /* Update the process ID */
                itr.value()->setColoredText(2, processId);
                /* Update the status date */
                itr.value()->setColoredText(3, statusDate);
            }
        }
    }

    // Find the width of the contents of each column
    int tableWidth = 0;
    for (int column = 0; column < twSystemTaskStatus->columnCount(); column++)
    {
        twSystemTaskStatus->resizeColumnToContents(column);
        tableWidth += twSystemTaskStatus->columnWidth(column);
    }

    // Set a minimum size on the tree widget large enough to show all column 
    // contents.
    twSystemTaskStatus->setMinimumWidth(tableWidth);
}

