/******************************************************************************
*
* NAME:  TreeWidgetUtilities.cpp
*
* PURPOSE:  This is the implementation file for the TreeWidgetUtilities
*           class.  TreeWidgetUtilities contains a set of useful static
*           fuctions using QTreeWidgets.  These function include printing a
*           QTreeWidget, writing a QTreeWidget, and building a QTreeWidget.
*
*****************************************************************************/

/* System Includes */
#include <iostream>

/* Local Includes */
#include "TreeWidgetUtilities.h"

//***********************************************************************

void TreeWidgetUtilities::printTreeWidget(
    QTreeWidget *twTreeWidget, // I: Pointer to the TreeWidget to be printed
    QPainter &pntPainter,      // I/O: The painting device
    QPrinter *pPrinter,   // I/O: The printer that pntPainter is painting on
    const int margin, // I: The margin that the user would like to have
    int &yPos, // I/O: The current y position on the printer page when called
    int &pageNo, // I/O: Current page number the printer is on
    const vector<twInfo> &twMetadata, // I: Struct containing tree widget 
                                      //    metadata
    const int NumberofColumns, // I: number of elements in twMetadata array
    const PrintOptions desiredPrinting) // I: Printing options 
{
    // This function prints out a QTreeWidget using the printer and painter
    // provided. Before calling this function, pntPainter must be painting
    // to pPrinter and margin, yPos, and pageNo must be initialized.
    // How it prints depends on data stored in the twMetadata struct about
    // the tree widget, such column alignment, column width, etc...

    int columnCounter; // used when iterating through each column
    int xPos;          // tracks the current X position on the printer
    int colXPos;       // tracks of the current X position within the
                       // current column that is printing.
    int startPos;      // Holds the starting X position of the current column
    int chCounter;     // Used to iterate through each character of the string
                       // that is currently printing.
    int charWidth;     // Holds the pixel width of the current font set
    int colWidth;      // Holds the expected width of the column being printed

    QFont oldPainterFont = pntPainter.font(); // save the current font 
    QString strColumnHeader, strColumnText;

    QTreeWidgetItemIterator it(twTreeWidget); 

    pntPainter.setFont(QFont("courier", 8));
    charWidth = 5;  // width of each character for the courier 8pt font
    QFontMetrics fm = pntPainter.fontMetrics();

    xPos = 0;
    startPos = 0;
    colXPos = 0;

    printColumnHeaders(twTreeWidget, pntPainter, pPrinter, margin, yPos,
                       twMetadata, NumberofColumns);

    yPos += 24;

    while (*it)
    {
        if (!printThisItem(*it , desiredPrinting))
        {
            ++it;
            continue;
        }

        // if there is not enough room on this page, goto next page
        if (margin + yPos > pPrinter->height() - margin)
        {
            pPrinter->newPage();             // no more room on this page
            pageNo++;
            yPos = 0;                       // back to top of page

            printColumnHeaders(twTreeWidget, pntPainter, pPrinter, margin,
                               yPos, twMetadata, NumberofColumns);

            yPos += 24;
            xPos = 0;
            startPos = 0;
            colXPos = 0;
        }

        //For each tree widget Item, print it out.

        for (columnCounter = 0; columnCounter < NumberofColumns;
                columnCounter++)
        {
            if (twMetadata[columnCounter].doPrint)
            {
                if (twMetadata[columnCounter].printWidth > 0)
                { 
                    colWidth = twMetadata[columnCounter].printWidth; 
                }
                else
                { 
                    colWidth = twTreeWidget->columnWidth(columnCounter); 
                }

                // get the column text for this column
                strColumnText = (*it)->text(columnCounter);

                for (chCounter = 0; chCounter < (int)strColumnText.length() &&
                        colXPos < colWidth; chCounter++)
                {
                    if (xPos > pPrinter->width())
                    {
                        xPos = 0;
                        startPos = 0;
                        yPos += 12;
                    }

                    pntPainter.drawText(margin + xPos, margin + yPos,
                                        pPrinter->width(), 70, 
                                        Qt::TextExpandTabs | Qt::TextWordWrap,
                                        QString(strColumnText[chCounter]));
                    xPos += charWidth;
                    colXPos += charWidth;
                }

                startPos += colWidth;
                xPos = startPos;
                colXPos = 0;
            }
        }


        yPos += 12;
        xPos = 0;
        startPos = 0;
        colXPos = 0;

        ++it;
    }

    pntPainter.setFont(oldPainterFont); // restore the original font
}

//***********************************************************************

void TreeWidgetUtilities::printTreeWidget(
    QTreeWidget *twTreeWidget, // I: Pointer to the QTreeWidget to be printed
    QPainter &pntPainter, // I/O: The desired painting device to use
    QPrinter *pPrinter, // I/O: The printer that pntPainter is paining on
    const int margin, // I: The desired margin for the paper, in pixels
    int &yPos, // I/O: The y position of the painting pen
    int &pageNo, // I/O: The page number that the printer is on.
    const PrintOptions desiredPrinting) //I: Printing options for the TreeWidget
{
    // This function prints out a QTreeWidget using the printer and painter
    // provided. Before calling this function, pntPainter must be painting to 
    // pPrinter and margin, yPos, and pageNo must be initialized.
    //
    // This function differs from the above printTreeWidget function by the
    // fact that it does not use a struct containing data about the tree widget
    // to print it out.  Instead, it iterates through the entire tree widget and
    // gets data such as column width, alignment, etc.. by retrieving it
    // from how the tree widget is actually displayed on the screen at the
    // time this function is invoked.

    int columnCounter;  // used to iterate the columns in the TreeWidget
    int xPos;       // used to track the current X position on the printer
    int colXPos;    // Keeps track of the current X position within the
                    // current column that is printing.
    int startPos;   // Hold the X position of the start of the current
                    // column that is printing.
    int chCounter;  // Used to iterate through each character of the string
                    // that is currently printing.
    int charWidth;  // Holds the pixel width of the current font set
    int colWidth;   // Hold the expected width of the column currently being
                    // printed

    int NumberofColumns = twTreeWidget->columnCount();

    QFont oldPainterFont = pntPainter.font(); //save the current font 
    QString strColumnHeader, strColumnText;

    QTreeWidgetItemIterator it(twTreeWidget); 

    pntPainter.setFont(QFont("courier", 8));
    charWidth = 5;  // width of each character for the courier 8pt font
    QFontMetrics fm = pntPainter.fontMetrics();

    xPos = 0;
    startPos = 0;
    colXPos = 0;

    printColumnHeaders(twTreeWidget, pntPainter, pPrinter, margin, yPos);

    yPos += 24;
    xPos = 0;
    startPos = 0;
    colXPos = 0;

    while (*it)
    {
        if (!printThisItem(*it, desiredPrinting))
        {
            ++it;
            continue;
        }


        // if there is not enough room on this page, goto next page
        if (margin + yPos > pPrinter->height() - margin)
        {
            pPrinter->newPage(); // no more room on this page
            pageNo++;
            yPos = 0;            // back to top of page

            printColumnHeaders(twTreeWidget, pntPainter, pPrinter, 
                margin, yPos);

            yPos += 24;
            xPos = 0;
            startPos = 0;
            colXPos = 0;
        }

        //For each tree widget Item, print it out.
        for (columnCounter = 0; columnCounter < NumberofColumns; 
             columnCounter++)
        {
            colWidth = twTreeWidget->columnWidth(columnCounter);

            // get the column text for this column
            strColumnText = (*it)->text(columnCounter);

            for (chCounter = 0; chCounter < (int)strColumnText.length() &&
                    colXPos < colWidth; chCounter++)
            {
                if (xPos > pPrinter->width())
                {
                    xPos = 0;
                    startPos = 0;
                    yPos += 12;
                }

                pntPainter.drawText(margin + xPos, margin + yPos,
                                    pPrinter->width(), 70, //fm.lineSpacing(),
                                    Qt::TextExpandTabs | Qt::TextWordWrap,
                                    QString(strColumnText[chCounter]));
                xPos += charWidth;
                colXPos += charWidth;
            }

            startPos += colWidth;
            xPos = startPos;
            colXPos = 0;
        }

        yPos += 12;
        xPos = 0;
        startPos = 0;
        colXPos = 0;

        ++it;
    }

    pntPainter.setFont(oldPainterFont); // restore the original font
}

//***********************************************************************

void TreeWidgetUtilities::writeTreeWidgetToFile(
    QTreeWidget *twTreeWidget, // I: The QTreeWidget to write to file
    QFile &file, // I/O: Location file
    const vector<twInfo> &twMetadata, // I: struct containing QTreeWidget info
    const int NumberofColumns, //I: Number of elements in twMetadata array
    const PrintOptions desiredPrinting) //I: Printing options 
{
    // This function writes the conents of the QTreeWidget passed in to the
    // provided QFile. Before invoking, file must be open and writable

    int columnCounter;
    QString strColumnText, lineToWrite;

    QStringList strlMaxLength = longestTextFields(twTreeWidget, twMetadata,
                                NumberofColumns);

    int insertPoint;
    QTreeWidgetItemIterator it(twTreeWidget);

    lineToWrite = "";   //Make sure string is empty.
    insertPoint = 0;    //Start at position 0 for the line.

    for (columnCounter = 0; columnCounter < NumberofColumns; columnCounter++)
    {
        if (twMetadata[columnCounter].doPrint)
        {
            // get column header and add to end of current line
            lineToWrite.insert(insertPoint,
                               twMetadata[columnCounter].strColumnTitle);
            insertPoint += strlMaxLength[columnCounter].toInt() + 2;
        }
    }

    //Add the new line character to the end of the string
    lineToWrite += "\n";
    //Write out the current line that was just built
    file.write(lineToWrite.toLocal8Bit());

    //Iterate through the tree widget now.
    while (*it)
    {
        if (!printThisItem(*it , desiredPrinting))
        {
            ++it;
            continue;
        }

        lineToWrite = "";
        insertPoint = 0;

        for (columnCounter = 0; columnCounter < NumberofColumns; 
             columnCounter++)
        {
            if (twMetadata[columnCounter].doPrint)
            {
                //GET The text to print out for the current column
                lineToWrite.insert(insertPoint,
                                   (*it)->text(columnCounter));
                insertPoint += strlMaxLength[columnCounter].toInt() + 2 ;
            }
        }

        //Add the new line character to the end of the string
        lineToWrite += "\n";

        //Write out the current line 
        file.write(lineToWrite.toLocal8Bit());

        ++it;
    }
}

//***********************************************************************

void TreeWidgetUtilities::writeTreeWidgetToFile(
    QTreeWidget *twTreeWidget, // I: The TreeWidget to write to file
    QFile &file, // I/O: Location file
    const PrintOptions desiredPrinting) //I: Printing options 
{
    // This function writes the QTreeWidget to the QFile. Before invoking, file
    // must be open and writable.
    // It differs from the above writeTreeWidgetToFile only in that a struct
    // describing the TreeWidget does not have to be passed in.  It writes
    // all data from the entire TreeWidget to file and uses column names
    // and information based on how the TreeWidget is displayed on the
    // screen at the time this function is invoked.

    int columnCounter;
    int NumberofColumns = twTreeWidget->columnCount();
    QString strColumnText, lineToWrite;

    QStringList strlMaxLength = longestTextFields(twTreeWidget);

    int insertPoint;
    QTreeWidgetItemIterator it(twTreeWidget);

    lineToWrite = "";   //Make sure string is empty.
    insertPoint = 0;    //Start at position 0 for the line.

    for (columnCounter = 0; columnCounter < NumberofColumns; 
         columnCounter++)
    {
        // get column header and add to end of current line
        lineToWrite.insert(insertPoint, 
            twTreeWidget->headerItem()->text(columnCounter));

        insertPoint += strlMaxLength[columnCounter].toInt() + 2;
    }

    //Add the new line character to the end of the string
    lineToWrite += "\n";
    //Write out the current line that was just built
    file.write(lineToWrite.toLocal8Bit());

    //Iterate through the tree widget now.
    while (*it)
    {
        if (!printThisItem(*it , desiredPrinting))
        {
            ++it;
            continue;
        }

        lineToWrite = "";
        insertPoint = 0;

        for (columnCounter = 0; columnCounter < NumberofColumns; 
             columnCounter++)
        {
            lineToWrite.insert(insertPoint, (*it)->text(columnCounter));
            insertPoint += strlMaxLength[columnCounter].toInt() + 2 ;
        }

        //Add the new line character to the end of the string
        lineToWrite += "\n";
        //Write out the current line that was just built
        file.write(lineToWrite.toLocal8Bit());

        ++it;
    }
}

//***********************************************************************

QStringList TreeWidgetUtilities::longestTextFields(
    QTreeWidget *twTreeWidget, // I: The TreeWidget to write to file
    const vector<twInfo> &twMetadata, // I: Struct containing QTreeWidget info
    const int NumberofColumns, // I: Number of elements in twMetadata array
    const PrintOptions desiredPrinting) //I: Printing options
{
    // This function returns a QStringList with each item of the list
    // containing the length of the longest text field in twTreeWidget of
    // that column.

    QString strTemp;
    QTreeWidgetItemIterator it(twTreeWidget);
    QStringList strlMaxSizes;
    int counter;

    //Set the max length to the length of the column titles at first.
    for (counter = 0; counter < NumberofColumns; counter++)
    {
        strlMaxSizes += 
            QString::number(twMetadata[counter].strColumnTitle.length());
    }

    while (*it)
    {
        if (!printThisItem(*it , desiredPrinting))
        {
            ++it;
            continue;
        }

        for (counter = 0; counter < NumberofColumns; counter++)
        {
            strTemp = (*it)->text(counter);

            if ((int)strTemp.length() > strlMaxSizes[counter].toInt())
            { 
                strlMaxSizes[counter] = QString::number(strTemp.length()); 
            }
        }

        ++it;
    }

    return strlMaxSizes;
}

//***********************************************************************

QStringList TreeWidgetUtilities::longestTextFields(
    QTreeWidget *twTreeWidget,  // I: QTreeWidget to check
    const PrintOptions desiredPrinting) // I: Printing options
{
    // This function returns a QStringList with each item of the list
    // containing the length of the longest text field in twTreeWidget of
    // that column.
    // It differs from the above longestTextFields function only in that
    // it does not accept as an argument a pointer to a struct containing
    // data about each of the columns of the TreeWidget.  Instead this
    // information is retrieved by getting that data as it is displayed
    // on the screen at the time of invocation.

    QString strTemp;
    QTreeWidgetItemIterator it(twTreeWidget);
    QStringList strlMaxSizes;
    int counter;
    int NumberofColumns = twTreeWidget->columnCount();

    //Set the max length to the length of the column titles at first.
    for (counter = 0; counter < NumberofColumns; counter++)
    {
        strlMaxSizes += 
           QString::number(twTreeWidget->headerItem()->text(counter).length());
    }

    while (*it)
    {
        if (!printThisItem(*it , desiredPrinting))
        {
            ++it;
            continue;
        }

        for (counter = 0; counter < NumberofColumns; counter++)
        {
            strTemp = (*it)->text(counter);

            if ((int)strTemp.length() > strlMaxSizes[counter].toInt())
            { 
                strlMaxSizes[counter] = QString::number(strTemp.length()); 
            }
        }

        ++it;
    }

    return strlMaxSizes;
}

//***********************************************************************

void TreeWidgetUtilities::buildTreeWidget(
    QTreeWidget *twTreeWidget, // I/O: TreeWidget to build
    const vector<twInfo> &twMetadata, //I: Struct containing QTreeWidget info
    const int NumberofColumns) //I: Number of elements in twMetadata array
{
    // This function sets up the TreeWidget twTreeWidget that is passed in.
    // It used information about each column stored in twMetadata.

    twTreeWidget->setColumnCount(NumberofColumns);

    QStringList strlist;

    // Add each column to the QTreeWidget
    for (int i = 0; i < NumberofColumns; i++) 
    {
        strlist << twMetadata[i].strColumnTitle;
    }

    if (strlist.size() > 0)
    {
        twTreeWidget->setHeaderLabels(strlist);
    }
}

//***********************************************************************

void TreeWidgetUtilities::printColumnHeaders(
    QTreeWidget *twTreeWidget, // I/O: QTreeWidget to build
        QPainter &pntPainter, // I/O: The painting device being used
        QPrinter *pPrinter, // I/O: The printer that pntPainter is using 
        const int margin, // I: The desired margin for the paper
        int &yPos, // I/O: The y position on the page for printing
        const vector<twInfo> &twMetadata, // I: Struct with QTreeWidget info
        const int NumberofColumns) // I: Number of elements in twMetadata array
{
    // This function prints out the column headers for the current list
    // view to the printer that pntPainter is painting on at the time
    // of invocation.  The struct twMetadata contains all of the column
    // information such as column alignment, width, etc...

    int colWidth;
    QString strColumnHeader;
    int xPos = 0;
    int startPos = 0;
    int colXPos = 0;
    int charWidth = 5;

    QFontMetrics fm = pntPainter.fontMetrics();

    //PRINT COLUMN HEADERS
    for (int columnCounter = 0; columnCounter < NumberofColumns; 
         columnCounter++)
    {
        if (twMetadata[columnCounter].doPrint)
        {
            if (twMetadata[columnCounter].printWidth > 0)
            { 
                colWidth = twMetadata[columnCounter].printWidth; 
            }
            else
            { 
                colWidth = twTreeWidget->columnWidth(columnCounter); 
            }

            // get column header
            strColumnHeader = twMetadata[columnCounter].strColumnTitle;

            for (int chCounter = 0; chCounter < (int)strColumnHeader.length() 
                    && colXPos < colWidth; chCounter++)
            {
                if (xPos > pPrinter->width())
                {
                    xPos = 0;
                    startPos = 0;
                    yPos += 12;
                }

                pntPainter.drawText(margin + xPos, margin + yPos,
                                    pPrinter->width(), 70, 
                                    Qt::TextExpandTabs | Qt::TextWordWrap,
                                    QString(strColumnHeader[chCounter]));
                xPos += charWidth;
                colXPos += charWidth;
            }

            startPos += colWidth;
            xPos = startPos;
            colXPos = 0;
        }
    }
}

//***********************************************************************

void TreeWidgetUtilities::printColumnHeaders(
    QTreeWidget *twTreeWidget, // I: The TreeWidget to printer headers for
    QPainter &pntPainter, // I/O: The painting device being used
    QPrinter *pPrinter, // I/O: The printer being printed on by pntPainter
    const int margin, // I: The desired margin for the paper
    int &yPos) // I/O: The y position on the page
{
    // This function prints out the column headers for the current list
    // view to the printer that pntPainter is painting on at the time
    // of invocation.
    // It differs from the above function of printColumnHeaders by the fact
    // that all of the information about how to print each column is
    // retrieved from the way the tree widget is displayed at the time of
    // invocation, rather than using a struct to hold the information.
    // This is a useful function if you don't care how a tree widget
    // is printed and you just want it printed as it is shown on the screen
    // and all columns to print.

    int colWidth;
    QString strColumnHeader;
    int xPos = 0;
    int startPos = 0;
    int colXPos = 0;
    int charWidth = 5;
    int NumberofColumns = twTreeWidget->columnCount();

    QFontMetrics fm = pntPainter.fontMetrics();

    //PRINT COLUMN HEADERS
    for (int columnCounter = 0; columnCounter < NumberofColumns; 
         columnCounter++)
    {
        colWidth = twTreeWidget->columnWidth(columnCounter);

        // get column header
        strColumnHeader = twTreeWidget->headerItem()->text(columnCounter);

        for (int chCounter = 0; chCounter < (int)strColumnHeader.length() &&
                colXPos < colWidth; chCounter++)
        {
            if (xPos > pPrinter->width())
            {
                xPos = 0;
                startPos = 0;
                yPos += 12;
            }

            pntPainter.drawText(margin + xPos, margin + yPos,
                                pPrinter->width(), 70, //fm.lineSpacing(),
                                Qt::TextExpandTabs | Qt::TextWordWrap,
                                QString(strColumnHeader[chCounter]));
            xPos += charWidth;
            colXPos += charWidth;
        }

        startPos += colWidth;
        xPos = startPos;
        colXPos = 0;
    }
}
//****************************************************************************
QTreeWidgetItem *TreeWidgetUtilities::getTreeWidgetItembyNum(
    QTreeWidget *treeWidget, // I: QTreeWidget 
    int index)               // I: Index of desired item
{
    // This function returns a pointer to the QTreeWidgetItem indicated by the
    // index in the arguemnt.

    QTreeWidgetItemIterator it(treeWidget);
    for (int count = 0; *it; ++it, count++)
    {
        if (count == index)
        { return (*it); }
    }

    return NULL;
}
//****************************************************************************
QTreeWidgetItem *TreeWidgetUtilities::getLastChild(
    QTreeWidget *treeWidget)  // I: QTreeWidget 
{
    int lastIndex = treeWidget->topLevelItemCount() - 1;

    // This function returns a pointer to the last item in a QtreeWidget
    return (getTreeWidgetItembyNum(treeWidget, lastIndex));
}

//****************************************************************************
bool TreeWidgetUtilities::printThisItem(
    QTreeWidgetItem *currItem, 
    PrintOptions desiredPrinting)
{
    QTreeWidgetItem *itemPointer;

    if (desiredPrinting == CurrentState)
    {
        if (currItem->parent() == 0)
        { 
            return true; 
        }

        itemPointer = currItem;

        while (itemPointer->parent())
        {
            if (itemPointer->parent()->isExpanded() == false)
            { 
                return false; 
            }

            itemPointer = itemPointer->parent();
        }

        return true;
    }
    else if (desiredPrinting == Selected && currItem->isSelected())
    { 
        return true; 
    }
    else if (desiredPrinting == SelectedPlusChildren)
    {
        if (currItem->isSelected() || currItem->parent()->isSelected())
        { 
            return true; 
        }
    }
    else if (desiredPrinting == SelectedPlusAllChildren)
    {
        if (currItem->isSelected())
        { 
            return true; 
        }

        itemPointer = currItem;

        while (itemPointer->parent())
        {
            if (itemPointer->parent()->isSelected())
            { 
                return true; 
            }

            itemPointer = itemPointer->parent();
        }
    }
    else if (desiredPrinting == AllPlusChildren 
            && (currItem->parent() == 0 || currItem->parent()->parent() == 0))
    { 
        return true; 
    }
    else if (desiredPrinting == AllPlusAllChildren)
    { 
        return true; 
    }

    return false;
}

/*****************************************************************************/

