/******************************************************************************
               Level 1 Product Generation System (LPGS) Software
******************************************************************************
 *
 * UNIT NAME:   UserParameters.cpp
 *
 * PURPOSE:
 * This class creates a tabbed dialog which allows the operator to see ordering
 * information about a particular product request id.  The information is 
 * retrieved from the work_orders and scenes tables.
 *
 * The class creates two tabs: General Information and L0R Data.
 *
 *****************************************************************************/

// Qt Includes 
#include <QtCore>
#include <QtGui>

// Local Includes 
#include "UserParameters.h"
#include "IASSqlQuery.h"
#include "DatabaseUtils.h"
#include "IASSchemaConfig.h"

//***************************************************************************** 

UserParameters::UserParameters(
    QString prId,
    QString woId,
    QWidget* parent, 
    Qt::WindowFlags fl)
: QDialog(parent, fl)
{
    setupUi(this);
    
    strPRID = prId;
    strWOID = woId;

    // set caption to class name and product request id passed in
    setWindowTitle("User Parameters :  " + strPRID);

    // set error caption
    strErrorCaption = "User Parameters Error :  " + strPRID;

    // set max label height
    maxLabelHeight = 20;
    labelSpacing = 5;
    
    // set flags indicating whether tab exists already or not
    // The General Tab widget and layout were added using Designer
    generalInfoTab.tabExists = TRUE;
    L0RDataTab.tabExists = FALSE;

    // fill/create tabs
    buildUserParametersDialog();

    connect(pbPrint, SIGNAL(clicked()), this, SLOT(print()));
    connect(pbDismiss, SIGNAL(clicked()), this, SLOT(accept()));
}

//***************************************************************************** 

UserParameters::~UserParameters()
{
    // destructor for UserParameters
}

//*****************************************************************************

void UserParameters::buildUserParametersDialog()
{
    // This function calls functions that fill/create the tabs
    // Query the database and determine if product is a L1G type product

    QString strProcedureId("");
    
    // create SQL command
    strSQLCommand = "SELECT procedure_id FROM work_orders"
                    " WHERE work_order_id ='" + strWOID + "'";
    
    // Get the product type from the datebase
    IASSqlQuery query;
    bool result = query.exec(strSQLCommand, __FILE__, __LINE__);

    if (result && query.first())
        strProcedureId = query.value("PROCEDURE_ID").toString();
    else
        return;
    
    // build the general information tab
    buildGeneralInfoTab();

    // build the L0R data tab
    buildL0RDataTab();
}

//****************************************************************************

void UserParameters::buildGeneralInfoTab()
{
    // This function creates the general information tab
    // query database for data
    query = TRUE;
    
    // set number of rows to indicate none exist
    generalInfoTab.numberOfRows = 0;
    
    // set string that is used as base in creating widget names
    generalInfoTab.strNameBase = "generalInfo";

    // set string used to create widget's (tab's) title
    generalInfoTab.strTabTitle = "General Info";    
    
    // set pointer to form layout added in Designer for the General Tab
    generalInfoTab.layout = flayoutGeneralTab;
    
    // set tab widget added in Designer
    generalInfoTab.tab = generalTabWidget;
    
    // set SQL command used to retrieve information to display in tab
    strSQLCommand = "SELECT product_request_id, work_order_id,"
                          " procedure_id, state, priority,"
                          " scene_id, wo_path"
                    " FROM work_orders"
                    " WHERE work_order_id = '" + strWOID + "'";

    // set name of table in struct, only used in error reporting
    generalInfoTab.strTable = "WORK_ORDERS";

    // call buildTab() to fill tab
    buildTab(&generalInfoTab);
}

//****************************************************************************

void UserParameters::buildL0RDataTab()
{
    // This functions used used to create the L0R Data tab in the User
    // Parameters Dialog

    // query database for data
    query = TRUE;

    // set number of rows to indicate none exist
    L0RDataTab.numberOfRows = 0;
    
    // set string that is used as base in creating widget names 
    L0RDataTab.strNameBase = "l0rdata";
    
    // set string used to create widget's (tab's) title
    L0RDataTab.strTabTitle = "L0R Data";    
    
    // set SQL command used to retrieve information to display in tab
    strSQLCommand = "SELECT TARGET_WRS_PATH, TARGET_WRS_ROW, DATE_ACQUIRED"
                    " FROM scenes WHERE work_order_id ='" + strWOID + "'";

    // set name of table in struct, only used in error reporting
    L0RDataTab.strTable = "scenes";
        
    // call buildTab() to create and fill tab
    buildTab(&L0RDataTab);
}

//****************************************************************************

void UserParameters::buildTab(UserParametersTab *tab)
{
    // this function will create labels that hold information about the
    // specified Product Request found by querying the product_request
    // table in the database.  A label that holds the database column
    // name (people friendly version) and a label that holds the column
    // value will be created.  All non-NULL columns in the database row
    // will be displayed.  NULL columns will not be displayed
   
    int count = 0; 
    int resultCount = 0;
    
    // if tab doesn't exist, create a new tab
    if (tab->tabExists == FALSE)
    {
        tab->tab = new QWidget(this);
        tab->layout = new QFormLayout();
        tab->tab->setLayout(tab->layout);

        if (tab->tab == 0)
        {
            return;
        }

        tab->tab->setContentsMargins(5, 5, 5, 5);
        tabWidget->addTab(tab->tab, "&" + tab->strTabTitle);
        tab->tabExists = TRUE;
    }

    if (query == TRUE) // execute query and get data from the (database)
    {
        IASSqlQuery query;
        bool result = query.exec(strSQLCommand, __FILE__, __LINE__);

        if (!result)
        {
            return;
        }

        resultCount = query.record().count();
        if (!query.next())
        {
            return;
        }

        // loop through columns in the search result, for each column that
        // contains a value, build a label with the column name followed by a
        // label containing the column value
        for (int column = 0, displayRow = tab->numberOfRows; 
             column < resultCount; column++)
        {
            QString value = query.value(column).toString();

            // Convert to column index to a column name
            QSqlRecord record = query.record();
            QString columnName = record.fieldName(column);

            if (value.length() > 0)
            {
                QString labelText =
                    DatabaseUtils::friendlyColumnName(columnName);

                tab->item[displayRow].lblTxt =
                    new QLabel(labelText, tab->tab);

                tab->item[displayRow].lblTxt->setGeometry(
                    QRect(30,40,110,16)); 
                tab->item[displayRow].lblTxt->setSizePolicy( 
                    QSizePolicy::Minimum, QSizePolicy::Minimum);
                tab->item[displayRow].lblTxt->setMaximumSize(
                    QSize(32767,maxLabelHeight));
                tab->item[displayRow].lblTxt->setMinimumSize(
                    QSize(0,maxLabelHeight));
                tab->item[displayRow].lblTxt->setFrameShadow(QFrame::Plain);
                tab->item[displayRow].lblTxt->setFrameShape(QFrame::NoFrame);

                // create a new label to hold the database column value

                if (columnName.compare("STATE") == 0)
                {
                    // if column is state, display string value, 
                    //  not the int value
                    tab->item[displayRow].leVal =
                        new QLineEdit(stateToString(value), tab->tab);
                    tab->item[displayRow].leVal->setReadOnly(true);
                    tab->item[displayRow].leVal->setGeometry(
                        QRect(160,40,210,20)); 
                    tab->item[displayRow].leVal->setMaximumSize(
                        QSize(32767,maxLabelHeight));
                }
                else
                {
                    // add label containing column value from database
                    tab->item[displayRow].leVal =
                        new QLineEdit(value,tab->tab);
                    tab->item[displayRow].leVal->setReadOnly(true);
                    tab->item[displayRow].leVal->setGeometry(
                        QRect(160,40,210,20)); 
                    tab->item[displayRow].leVal->setMaximumSize(
                       QSize(32767,maxLabelHeight));
                }

                tab->layout->addRow(tab->item[displayRow].lblTxt, 
                    tab->item[displayRow].leVal);

                displayRow++;
                count++;

            }
        }
    }
    else // get data from QStringLists 
    {
        // loop through the items in the qstring lists, for each index that 
        // contains a value, build a label with the identifier followed by a 
        // label containing the column value
        for (int index = 0, displayRow = tab->numberOfRows; 
             index < strlIdentifiers.count(); index++)
        {
            // if the list has data in it
            if (strlValues[index].length() > 0)
            {
                tab->item[displayRow].lblTxt = new QLabel(
                    strlIdentifiers[index], tab->tab);
                tab->item[displayRow].lblTxt->setGeometry(
                    QRect(30,40,110,16)); 
                tab->item[displayRow].lblTxt->setSizePolicy( 
                    QSizePolicy::Minimum, QSizePolicy::Minimum);
                tab->item[displayRow].lblTxt->setMaximumSize(
                    QSize(32767,maxLabelHeight));
                tab->item[displayRow].lblTxt->setMinimumSize(
                    QSize(0,maxLabelHeight));
                tab->item[displayRow].lblTxt->setFrameShadow(
                    QFrame::Plain);
                tab->item[displayRow].lblTxt->setFrameShape(
                    QFrame::NoFrame);

                // add label containing value (QString) from QStringList
                //  at index
                tab->item[displayRow].leVal = 
                    new QLineEdit(strlValues[index],tab->tab);
                tab->item[displayRow].leVal->setReadOnly(true);
                tab->item[displayRow].leVal->setGeometry(
                    QRect(160,40,210,20)); 
                tab->item[displayRow].leVal->setMaximumSize(
                    QSize(32767,maxLabelHeight));
            }

            tab->layout->addRow(tab->item[displayRow].lblTxt, 
                tab->item[displayRow].leVal);

            displayRow++;
            count++;
        }
    }

    // update field to contain number of rows in tab
    tab->numberOfRows = tab->numberOfRows + count;
}

//*****************************************************************************

void UserParameters::print()
{
    // print the contents of the User Parameters tab'd dialog
    QString strPrinter;             // selected printer
    QPainter painter;

    printer = new QPrinter();        // printer object

    yPos = 0;                   // y position for each line, start at 0
    margin = 10;

    printer->setPageSize(QPrinter::Letter);
    printer->setOrientation(QPrinter::Portrait);
    printer->setColorMode(QPrinter::GrayScale);
        
    // get default printer
    strPrinter = getenv("LPDEST");
    if(!strPrinter.isEmpty())
        printer->setPrinterName(strPrinter);

    // if operator wants to print
    QPrintDialog printDialog(printer, this);
    if (printDialog.exec() == QDialog::Accepted)
    {
        // if operator doesn't select OK to continue, exit without printing
        if (!painter.begin(printer))
            return;                            
        
        // set font
        painter.setFont(QFont("courier", 8));
        
        // need width/height of printer surface
        QFontMetrics fm = painter.fontMetrics();
        
        // print header
        strTemp = "LPGS User Paramaters for " + strPRID + "   Printed Date: " 
            + QDateTime::currentDateTime().toString();
            
        painter.drawText(margin, margin + yPos, printer->width(),
            fm.lineSpacing(), Qt::TextExpandTabs | Qt::TextDontClip, strTemp);
        yPos+= 2 * fm.lineSpacing();     
        
        // print General Info Tab
        if (generalInfoTab.tabExists)
            printTab(&generalInfoTab, yPos, &painter, &fm);
    
        // print L0R Data Tab
        if (L0RDataTab.tabExists)
            printTab(&L0RDataTab, yPos, &painter, &fm);
    } // operator selected to not print
}

//*****************************************************************************

void UserParameters::printTab(UserParametersTab *tab, 
    int &yPos,              // I: vertical position on printed page
    QPainter *p,            // I: pointer to painter for printer
    QFontMetrics *fm)       // I: font metrics such as size
{       
    // This function prints out the contents of a tab
    
    int indent = 200; // how far to indent the second row
        
    QString strline = "______________________________";
    
    // set header
    strTemp = strline + " " + tab->strTabTitle + " " + strline ;

    // print header
    p->drawText( margin, margin + yPos, printer->width(),
        fm->lineSpacing(), Qt::TextExpandTabs | Qt::TextDontClip, strTemp );
    
    // move cursor
    yPos += 2 *  fm->lineSpacing();
        
    // print out items in the Tab
    for(int item = 0; item < tab->numberOfRows; item++)
    {
        strTemp = tab->item[item].lblTxt->text();
        if (strTemp.trimmed().compare("Band Combination") == 0 )
        {
            p->drawText( margin, margin + yPos, printer->width(),
                fm->lineSpacing(), Qt::TextExpandTabs | Qt::TextDontClip, 
                strTemp);
            
            QStringList strlBandComb = 
                tab->item[item].leVal->text().split("\n", 
                    QString::SkipEmptyParts);

            for (int i = 0; i < strlBandComb.count(); i++)
            {
                // draw text
                p->drawText(margin + indent , margin + yPos,
                    printer->width(), fm->lineSpacing(), 
                    Qt::TextExpandTabs | Qt::TextDontClip,
                    strlBandComb[i]);

                // move cursor down one line
                yPos += fm->lineSpacing();
            }
        }
        else
        {
            p->drawText(margin, margin + yPos, printer->width(),
                fm->lineSpacing(), Qt::TextExpandTabs | Qt::TextDontClip, 
                strTemp );
            
            strTemp = tab->item[item].leVal->text();
            p->drawText(margin + indent, margin + yPos, printer->width(),
                fm->lineSpacing(), Qt::TextExpandTabs | Qt::TextDontClip, 
                strTemp);
            
            yPos +=  fm->lineSpacing();
        }   
    }
}

//*****************************************************************************

