/******************************************************************************
               Level 1 Product Generation System (LPGS) Software
*******************************************************************************
*
* UNIT NAME:   WorkOrderParameters.cpp
*
* PURPOSE:
* Provides a dialog that allows the operator to view and/or modify the
* parameters associated with a work order.  The operator chooses which
* script to view/modify through a combo box on top of the GUI, and a list
* of parameters allows them to examine or modify the values.
*
******************************************************************************/

/* System Includes */
#include <iostream>

/* Qt Includes */

#include "IASErrorDialog.h"
#include "IASSqlQuery.h"
#include "IASUtilities.h"
#include "IASSchemaConfig.h"
#include "IASParameterDelegate.h"

/* Local Includes */
#include "WorkOrderParameters.h"

WorkOrderParameters::WorkOrderParameters(
    bool editFlag,      // I: Params editable or not? 
    QString strPRId,    // I: Product Request Identifier
    QString strWOId,    // I: Work Order Identifier
    QWidget *parent,    // I: Parent widget
    Qt::WindowFlags fl) // I: Window flags
: QDialog(parent, fl)
{
    setupUi(this);

    setModal(true);

    editableValues = editFlag;

    // Initialize the parmsModified flag
    parmsModified = false;

    if (editableValues)
    {
        pbOK->setEnabled(true);
        /* For LDCM there currently is no reason to switch procedures, so
           we will also disable it here.  Landsat will need to do something
           different. */
        cboProductType->setEnabled(false);
        grpBoxPauseFlag->setEnabled(true);
    }
    else
    {
        pbOK->setEnabled(false);
        cboProductType->setEnabled(false);
        grpBoxPauseFlag->setEnabled(false);
    }

    strWOId_ = strWOId;
    strPRId_ = strPRId;

    getWorkOrderInstanceInformation();

    setupParmsTable();
    
    populateProcedureSelection();
    populateScriptSelection();

    // Connect signals and slots
    QItemSelectionModel* selModel = tblParms->selectionModel();
    connect(selModel,
            SIGNAL(currentChanged(const QModelIndex&, const QModelIndex&)),
            this,
            SLOT(setCurrentParmValue(const QModelIndex&, const QModelIndex&)));
    selModel = NULL;

    connect(parameterDataModel, SIGNAL(itemDataChanged(const QModelIndex&)),
            this, SLOT(updateParms(const QModelIndex&)));

    connect(pbOK, SIGNAL(clicked()), this, SLOT(accept()));
    connect(pbCancel, SIGNAL(clicked()), this, SLOT(reject()));

    connect(cboScriptName, SIGNAL(activated(const QString &)),
            this, SLOT(scriptChanged(const QString &)));

    /* Only need to capture the signal from the on button. */
    connect(rbPauseOn, SIGNAL(toggled(bool)), this, SLOT(pauseToggled(bool)));

    /* Initiate a transaction */
    QSqlDatabase::database().transaction();
}

/*****************************************************************************/

void WorkOrderParameters::setCurrentParmValue(
    const QModelIndex& cIndex, /* I: Current model index */
    const QModelIndex&)        /* I: Previous model index - NOT USED */
{
    if (!cIndex.isValid())
    {
        return;
    }

    IASParameterItem *pItem = parameterDataModel->getItem(cIndex.row(), 3);

    currentParmValue = pItem->data().toString();
}


/*****************************************************************************/

void WorkOrderParameters::getWorkOrderInstanceInformation()
{
    IASSqlQuery query;

    QString sqlCmd = "SELECT processing_instance, satellite, sensor "
                     "FROM work_orders "
                     "WHERE work_order_id = '" + strWOId_ + "'";

    if (!query.exec(sqlCmd, __FILE__, __LINE__))
    {
        return;
    }

    // Get the work order information
    query.next();
    instance = query.value("PROCESSING_INSTANCE").toString();
    satellite = query.value("SATELLITE").toString();
    sensor = query.value("SENSOR").toString();

    if (instance.isEmpty() || satellite.isEmpty() || sensor.isEmpty())
    {
        QMessageBox::warning(this, "Error",
            "Unable to retrieve the processing instance, satellite, and sensor "
            "for work order id " + strWOId_ + ".",
            QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);
    }
}


/*****************************************************************************/

void WorkOrderParameters::updateParms(const QModelIndex& pIndex)
{
    // This function checks the parameters updated value against the max
    // and min (if there is one) and updates the wo_parms table with the
    // new value.  We need to verify if the fields value was really updated
    // since the valueUpdated signal is emitted whether the value
    // was actually changed or not.

    if (!pIndex.isValid())
    {
        return;
    }

    IASParameterItem *pItem = NULL;
    int row = pIndex.row();
    int column = pIndex.column();

    pItem = parameterDataModel->getItem(row, column);

    if ((pItem->data().toString().length() == 0) && // check for null vs empty
        (currentParmValue.length() == 0)) 
    {
        return;
    }
    else
    {
        if(pItem->data().toString() == currentParmValue) // no change
            return;
    }

    // Validate the parameter unless it is from a combobox.  Values in a
    // combobox are all assumed to be valid since they were read from the
    // database and there is only a discrete set of values to choose from.
    if (pItem->type() != IASParameterItem::TextSelection) // Not a combobox value
    {
        if (!validateParm(row)) // Out of range, reset value to original
        {
            QVariant data = currentParmValue;

            // make sure the correct item is accessed
            pItem = parameterDataModel->getItem(row, 3); /* 3 - Parm Value */
            pItem->setData(data);

            tblParms->setCurrentIndex(parameterDataModel->index(row, column));

            return;
        }
    }

    // If the database update is successful reset the current value
    // to the newly updated value
    if (updateParmsTable(row))
    {
        currentParmValue = pItem->data().toString();
    }
}

/*****************************************************************************/

bool WorkOrderParameters::validateParm(int row)
{
    // Checks whether the string can be converted to a valid number
    // and checks the value of the parameter against the max/min in the
    // parm_edits table (if there is one).

    bool ok;
    IASParameterItem *pItem = NULL;

    pItem = parameterDataModel->getItem(row, 0); /* 0 - Parm Name */
    QString parmName = pItem->data().toString();
    pItem = parameterDataModel->getItem(row, 2); /* 2 - Type */
    QString parmType = pItem->data().toString().toUpper();
    pItem = parameterDataModel->getItem(row, 3); /* 3 - Parm Value */
    QString parmValue = pItem->data().toString();
    pItem = parameterDataModel->getItem(row, 4); /* 4 - Required */
    QString parmRequired = pItem->data().toString();

    // allow parameters which are not required to be empty strings.
    if (parmValue == "" && parmRequired == "FALSE")
    {
        return true;
    }

    if (parmValue == "")
    {
        QMessageBox::warning(this, "Error", 
            "The Parameter selected is required",
            QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);

        return false;
    }

    // make sure the value is valid and return false if not
    if ((parmType == "INT") || (parmType == "LONG"))
    {
        parmValue.toLong(&ok);
        if(!ok)
        {
            QMessageBox::warning(this, "Error", 
                "The Parameter value is not a valid integer number",
                QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);

            return false;
        }
    }
    else if ((parmType == "FLOAT") || (parmType == "DOUBLE"))
    {
        parmValue.toDouble(&ok); 
        if (!ok)
        {
            QMessageBox::warning(this, "Error", 
                "The Parameter value is not a valid floating point number",
                QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);

            return false;
        }
    }
    else 
    {
        return true;
    }

    // If there is a max/min associated with the parameter check if it is
    // in range.  Return false if not so the out of range value will be reset.
    IASSqlQuery query;
    QString sqlStmt = "SELECT min_value, max_value"
                      " FROM parm_edits"
                      " WHERE parm_name='" + parmName + "'"
                      " AND processing_instance = '" + instance + "'"
                      " AND satellite = '" + satellite + "'"
                      " AND sensor = '" + sensor + "'";

    if (query.exec(sqlStmt, __FILE__, __LINE__))
    {
        query.next();
    }

    if (!query.isNull(0)) // There is a max/min
    {
        bool validVal = true;

        if ((parmType == "INT") || (parmType == "LONG"))
        {
            if((parmValue.toInt(&ok) < query.value("MIN_VALUE").toInt(&ok)) ||
               (parmValue.toInt(&ok) > query.value("MAX_VALUE").toInt(&ok)))
            {
                validVal = false;
            }
        }
        else if ((parmType == "FLOAT") || (parmType == "DOUBLE"))
        {
            if ((parmValue.toDouble(&ok) 
                    < query.value("MIN_VALUE").toDouble(&ok)) ||
               (parmValue.toDouble(&ok) 
                    > query.value("MAX_VALUE").toDouble(&ok)))
            {
                validVal = false;
            }
        }

        // If the value is not valid report the min/max and type
        // information for the parameter and return false to prevent
        // update with this value
        if (!validVal)
        {
            QMessageBox::warning(this, "Error",
                "The Parameter value is invalid, should be:\nType = " +
                parmType + "\n" + 
                "Min = " + query.value("MIN_VALUE").toString() + "\n" +
                "Max = " + query.value("MAX_VALUE").toString(),
                QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);

            return false;
        }
    }

    return true;
}

/*****************************************************************************/

bool WorkOrderParameters::updateParmsTable(int row)
{
    // Updates the value of a parameter in the wo_parms table.

    IASSqlQuery query;
    IASParameterItem *pItem = NULL;

    pItem = parameterDataModel->getItem(row, 0); /* 0 - Parm Name */
    QString parmName = pItem->data().toString();
    pItem = parameterDataModel->getItem(row, 1); /* 1 - Parm Occurance */
    QString parmOccur = pItem->data().toString();
    pItem = parameterDataModel->getItem(row, 2); /* 2 - Type */
    QString parmType = pItem->data().toString();
    pItem = parameterDataModel->getItem(row, 3); /* 3 - Parm Value */
    QString parmValue = pItem->data().toString();
    pItem = parameterDataModel->getItem(row, 5); /* 5 - Attribute */
    QString attribute = pItem->data().toString();

    QString sqlStmt = "UPDATE wo_parms"
                      " SET parm_value = ";
    if (parmValue.length() == 0)
    {
        sqlStmt += "NULL";
    }
    else if ((parmType.toUpper() != "INT") && (parmType.toUpper() != "LONG"))
    {
        sqlStmt += "'" + parmValue;

        // include the algorithm attribute if present
        if ((attribute != "RUN") && (attribute != ""))
        {
            sqlStmt += "_" + attribute;
        }

        sqlStmt += "'";
    }
    else
    {
        sqlStmt += parmValue;
    }

    sqlStmt += " WHERE work_order_id = '" + strWOId_ + "'"
                   " AND script_id = '" + currentScriptID + "'"
                   " AND parm_name = '" + parmName + "'"
                   " AND parm_occurrence = " + parmOccur;

    if (!query.exec(sqlStmt, __FILE__, __LINE__))
    {
        return false;
    }

    // parameters have been modified
    parmsModified = true;

    return true;
}

/*****************************************************************************/

void WorkOrderParameters::populateProcedureSelection()
{
    IASSqlQuery query;
    QString strProcedureId = "";
    QString strModuleId = "";
    bool found;

    cboProductType->clear();

    QString sqlCmd = "SELECT procedure_id"
                     " FROM work_orders"
                     " WHERE work_order_id = '" + strWOId_ + "'";

    if (query.exec(sqlCmd, __FILE__, __LINE__))
    {
        query.next();
        strProcedureId = query.value("PROCEDURE_ID").toString();
    }

    /* Need to know sensor as MSS does not have a L1Gt product type */
    sqlCmd = "SELECT module_id"
             " FROM modules"
             " WHERE mod_type = 'PROCEDURE'"
             " AND processing_instance = '" + instance + "'"
             " AND satellite = '" + satellite + "'"
             " AND sensor = '" + sensor + "'";

    found = false;
    if (query.exec(sqlCmd, __FILE__, __LINE__))
    {
        while (query.next())
        {
            strModuleId = query.value("MODULE_ID").toString();
            cboProductType->addItem(strModuleId);

            if (strModuleId == strProcedureId)
            {
                found = true;
            }
        }
    }

    /* Just populate the list with the STD product for now */
    if (!found)
    {
        /* Have a new product type -- so put first in
           list, add the others, and set to current
           product type */
        cboProductType->addItem(strProcedureId);
    } 

    cboProductType->setCurrentIndex(cboProductType->findText(strProcedureId));
}

/*****************************************************************************/

void WorkOrderParameters::populateScriptSelection()
{
    IASSqlQuery query;

    QString sqlCmd = "SELECT UNIQUE script_id, seq_nbr"
                     " FROM wo_scripts"
                     " WHERE work_order_id = '" + strWOId_ + "'"
                     " ORDER BY seq_nbr" ;

    if (query.exec(sqlCmd, __FILE__, __LINE__))
    {
        cboScriptName->clear();

        while (query.next())
        {
            cboScriptName->addItem(query.value("SCRIPT_ID").toString());
        }

        cboScriptName->setCurrentIndex(0);

        scriptChanged(cboScriptName->currentText());
    }
}

/*****************************************************************************/

void WorkOrderParameters::setupParmsTable()
{
    QFont f;
    QStringList labels;

    parameterDataModel = new IASParameterModel;
    parameterDataModel->setColumnCount(6);

    labels << "Parm Name" << "Element" << "Type" << "Parm Value"
           << "Required" << "Attribute";
    parameterDataModel->setHorizontalHeaderLabels(labels);

    // Set up the delegate for the table.
    tblParms->setItemDelegate(new IASParameterDelegate);

    // Set up the edit triggers.
    tblParms->setEditTriggers(QAbstractItemView::AllEditTriggers);

    // Set the rows to alternate colors.
    tblParms->setAlternatingRowColors(true);

    // Make headings bold and set focus and selection
    f.setBold(true);
    tblParms->horizontalHeader()->setFont(f);

    // Only select one row at a time.
    tblParms->setSelectionMode(QAbstractItemView::SingleSelection);
    tblParms->setSelectionBehavior(QAbstractItemView::SelectRows);

    // Get rid of the vertical header.
    tblParms->verticalHeader()->hide();

    // Set the model
    tblParms->setModel(parameterDataModel);

    // Adjust the columns to the new data.
    tblParms->resizeColumnsToContents();

    // Set last column to be stretchable
    tblParms->horizontalHeader()->setStretchLastSection(true);
}

/*****************************************************************************/

void WorkOrderParameters::scriptChanged(const QString &scriptName)
{
    // This function is called every time the operator changes the active
    // script by selecting a different item in the combo box.  It does
    // two main things: first, it determines the value of the pause flag
    // for this script and sets the radio buttons accordingly.  More
    // importantly, it extracts all of the parameters (name, occurrence,
    // and value) from the WO_PARMS table and inserts them into the
    // list view object.

    IASSqlQuery query;
    QString strPauseFlag = "";
    IASParameterItem *pItem = NULL;
    IASParameterItemRow *pItemRow = NULL;

    currentScriptID = scriptName;

    tblParms->setSortingEnabled(true);
    tblParms->setSelectionMode(QAbstractItemView::NoSelection);

    QString strQuery = "SELECT pause_flag"
                       " FROM wo_scripts"
                       " WHERE script_id = '" + currentScriptID + "'"
                       " AND work_order_id = '" + strWOId_ + "'";

    // Get the value of the pause flag for this script from the database,
    // so that we know how to set up the radio button(s) initially.  If
    // the pause flag is already enabled, select the "Yes" radio button
    // (and deselect the "No" button.)  If the pause flag is not enabled,
    // then do the exact opposite ..
    if (query.exec(strQuery, __FILE__, __LINE__))
    {
        query.next();
        strPauseFlag = query.value("PAUSE_FLAG").toString();
    }

    /* stop the signals */
    rbPauseOff->blockSignals(true);
    rbPauseOn->blockSignals(true);
    /* configure the flag for this script */
    if (strPauseFlag == "0") 
    {
        // A value of "0" in the field indicates no pause.
        rbPauseOff->setChecked(true);
        rbPauseOn->setChecked(false);
    } 
    else    
    {
        rbPauseOff->setChecked(false);
        rbPauseOn->setChecked(true);
    }
    /* restore the signals */
    rbPauseOff->blockSignals(false);
    rbPauseOn->blockSignals(false);

    // Delete any rows currently in the table
    parameterDataModel->deleteData();

    // Extract all of the work order parameters for this script for
    // this work order and insert them into the list view.
    strQuery = "SELECT parm_name, parm_occurrence, parm_value"
               " FROM wo_parms"
               " WHERE script_id = '" + currentScriptID + "'"
                   " AND work_order_id = '" + strWOId_ + "'"
               " ORDER BY parm_name DESC, parm_occurrence DESC";

    QVariant data;
    if (query.exec(strQuery, __FILE__, __LINE__))
    {
        while (query.next())
        {
            pItemRow = new IASParameterItemRow;

            /* 0 - Parm Name */
            pItem = new IASParameterItem;
            data = query.value("PARM_NAME").toString();
            pItem->setData(data);
            pItemRow->append(pItem);

            /* 1 - Element */
            pItem = new IASParameterItem;
            data = query.value("PARM_OCCURRENCE").toString();
            pItem->setData(data);
            pItemRow->append(pItem);

            /* 2 - Type */
            pItem = new IASParameterItem;
            data = "";
            pItem->setData(data);
            pItemRow->append(pItem);

            /* 3 - Parm Value */
            pItem = new IASParameterItem;
            data = query.value("PARM_VALUE").toString();
            pItem->setData(data);
            pItemRow->append(pItem);

            /* 4 - Required */
            pItem = new IASParameterItem;
            data = "";
            pItem->setData(data);
            pItemRow->append(pItem);

            /* 5 - Attribute */
            pItem = new IASParameterItem;
            data = "";
            pItem->setData(data);
            pItemRow->append(pItem);

            parameterDataModel->appendRow(pItemRow);
        }
    }

    // Update the Type/Required and Parameter Value fields seperately to
    // minimize the number of queries.
    insertTypeRequired();
    insertParmValueItem();

    // Select the first parameter and display the parameter
    // description.  We want to keep the focus in the scripts table
    tblParms->setCurrentIndex(parameterDataModel->index(0, 3));

    /* Update the description object.  Just use the specified index for both
       parameters. */
    setCurrentParmValue(parameterDataModel->index(0, 2),
                        parameterDataModel->index(0, 2));

    // Adjust the columns to the new data.
    tblParms->resizeColumnsToContents();

    // Set last column to be stretchable
    tblParms->horizontalHeader()->setStretchLastSection(true);
}

//***********************************************************************

void WorkOrderParameters::insertTypeRequired()
{
    // Inserts values into the parameter "type" and "required" fields
    // from the def_parms database table.  Number of queries is cut down
    // by eliminating duplicate queries of parameters with multiple
    // occurrences.

    IASSqlQuery query;
    IASParameterItem *pItem = NULL;

    QString parmName = ""; // Current and last parameter name.
    QString parmOccur = "";

    for (int row = 0; row < parameterDataModel->rowCount(); row++)
    {
        pItem = parameterDataModel->getItem(row, 0); // Parm Name
        parmName = pItem->data().toString();

        pItem = parameterDataModel->getItem(row, 1); // Parm Occurance
        parmOccur = pItem->data().toString();

        QString sqlStmt =
            "SELECT parm_type, required"
            " FROM def_parms"
            " WHERE module_id = "
                    "(SELECT procedure_id"
                    " FROM work_orders"
                    " WHERE work_order_id = '" + strWOId_ + "')"
                " AND sub_module_id = '" + currentScriptID + "'"
                " AND parm_name = '" + parmName + "'"
                " AND parm_occurrence = " + parmOccur +
                " AND processing_instance = '" + instance + "'"
                " AND satellite = '" + satellite + "'"
                " AND sensor = '" + sensor + "'";


        if (!query.exec(sqlStmt, __FILE__, __LINE__))
        {
            return;
        }
        query.next();

        // Insert the parameter type and required flag in the parameters table
        if (query.isValid())
        {
            QVariant data;

            /* 2 - Type */
            data = query.value("PARM_TYPE").toString();
            parameterDataModel->setItemData(row, 2, data);

            /* 4 - Required */
            if(query.value("REQUIRED").toInt())
            {
                data = "TRUE";
            }
            else
            {
                data = "FALSE";
            }
            parameterDataModel->setItemData(row, 4, data);
        }
    }
}

//***********************************************************************

void WorkOrderParameters::insertParmValueItem()
{
    // Checks to see if there are a fixed set of parameter values
    // and whether the value read from the database should be mapped to
    // a different string to be shown to the user.  A combobox will be
    // inserted if there are a fixed set of values for the parameter.
    // Also, if the value includes an algorithm attribute, this is 
    // not displayed in the parm_value column, but rather is displayed
    // in a separate combobox.

    IASSqlQuery query;
    QStringList ouiValues;
    QStringList parmValueSegments; // algorithm name, and potentially
                                   // algorithm attribute as well
    IASSqlQuery queryAttributes;   // Query for parm_attributes table
    QString sqlStmtAttributes;     // SQL to query parm_attributes table
    QStringList attributes;        // list of all algorithm attributes 
    QString parmValue;             // parm value (sometimes an algorithm name)
    QString attribute;             // the attribute in the parmValue field 

    IASParameterItem *pItem = NULL;

    // Get the list of possible attribute values.
    sqlStmtAttributes = "SELECT oui_value FROM "
        "parm_list where parm_name = 'ATTRIBUTE'"
        " AND processing_instance = '" + instance + "'"
        " AND satellite = '" + satellite + "'"
        " AND sensor = '" + sensor + "'";

    if (!queryAttributes.exec(sqlStmtAttributes, __FILE__, __LINE__))
    {
        return;
    }

    // get next parameter from database cursor
    while (queryAttributes.next())
    {
        attributes << queryAttributes.value("OUI_VALUE").toString();
    }

    QString parmName = "";
    QString parmRequired = "";
    QVariant data;

    for (int row = 0; row < parameterDataModel->rowCount(); row++)
    {
        pItem = parameterDataModel->getItem(row, 4); /* 4 - Required */
        parmRequired = pItem->data().toString();
        pItem = parameterDataModel->getItem(row, 0); /* 0 - Parm Name */
        // Only query the parm_list table if the parameter name is different
        // from the previous row (to cut down on queries)
        if ((row == 0) || (pItem->data().toString() != parmName))
        {
            parmName = pItem->data().toString();

            ouiValues.clear();

            QString sqlStmt = "SELECT oui_value, program_value"
                              " FROM parm_list"
                              " WHERE parm_name = '" +  parmName + "'"
                              " AND processing_instance = '" + instance + "'"
                              " AND satellite = '" + satellite + "'"
                              " AND sensor = '" + sensor + "'";

            // Create a stringlist with the possible values for the parameter
            if (!query.exec(sqlStmt, __FILE__, __LINE__))
            {
                return;
            }

            while (query.next())
            {
                ouiValues += query.value("OUI_VALUE").toString();
            }
        }

        // Initialize the parmValue value. 
        pItem = parameterDataModel->getItem(row, 3);
        parmValue = pItem->data().toString();

        attribute = ""; // Force to empty string.

        // If the Parm name is ALGORITHMS, get a list of parm attributes
        if (parmName == "ALGORITHMS")
        {
            // Find if an attribute is at the end of the parmValue
            parmValueSegments = parmValue.split(QRegExp("_"));

            // This last element is an attribute if it's in the attribute list
            if (attributes.contains(parmValueSegments.last()))
            {
                // The last element is an attribute
                attribute = parmValueSegments.last();

                // Remove the last element (the attribute) from the list
                parmValueSegments.pop_back();

                // Form the algorithm name from the rest of the parmValue 
                // segments
                parmValue = parmValueSegments.join ("_");
            }
            else
            {
                // The last element is not an attribute
                attribute = "RUN";
            }
        }

        // Create a combo box in the cell if there is a set of values for the
        // parameter. Otherwise use the line edit that is already there.  The
        // present value of the parameter is already in the parm value column
        pItem = parameterDataModel->getItem(row, 3); /* 3 - Parm Value */
        data = parmValue;
        pItem->setData(data);

        if (!ouiValues.isEmpty())
        {
            // Discriminate against the ALGORITHMS.
            // Don't allow any modifications to them.
            if (parmName == "ALGORITHMS")
            {
                pItem->setType(IASParameterItem::Text);
            }
            else
            {
                if (parmRequired == "FALSE")
                {
                    if (!ouiValues.contains("")) // add it if not present
                    {
                        ouiValues.prepend(""); // add an empty selection
                    }
                }
                else
                {
                    if (ouiValues.contains("")) // if present remove it
                    {
                        ouiValues.removeAll("");
                    }
                }

                /* Sort the options with a special routine if they are for
                   the BAND_LIST */
                if (parmName == "BAND_LIST")
                {
                    qSort(ouiValues.begin(), ouiValues.end(),
                          IASUtilities::numericLessThan);
                }

                if (editableValues)
                {
                    pItem->setEditable(true);
                }
                else
                {
                    pItem->setEditable(false);
                }
                pItem->setType(IASParameterItem::TextSelection);
                pItem->setOptions(ouiValues);
            }
        }
        else
        {
            if (editableValues)
            {
                pItem->setEditable(true);
            }
            else
            {
                pItem->setEditable(false);
            }
        }

        pItem = parameterDataModel->getItem(row, 5); /* 5 - Attribute */
        data = attribute;
        pItem->setData(data);

        // Also do this for the attribute column if appropriate
        if (parmName == "ALGORITHMS")
        {
            if (editableValues)
            {
                pItem->setEditable(true);
            }
            else
            {
                pItem->setEditable(false);
            }
            pItem->setType(IASParameterItem::TextSelection);
            pItem->setOptions(attributes);
        }
    }
}


/*****************************************************************************/

void WorkOrderParameters::pauseToggled(bool checked)
{
    IASSqlQuery query;
    QString state = "0";

    if (checked)
    {
        state = "1";
    }

    if (query.exec("UPDATE wo_scripts"
               " SET pause_flag = '" + state + "'"
               " WHERE work_order_id = '" + strWOId_ + "'"
               " AND script_id = '" + currentScriptID + "'",
               __FILE__, __LINE__))
    {
        parmsModified = true;
    }
}

/*****************************************************************************/

void WorkOrderParameters::accept()
{
    if(!QSqlDatabase::database().commit())
    {
        QMessageBox::warning(this, "Error", 
            "Commit of database changes failed",
            QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);
    }

    QDialog::accept();
}

/*****************************************************************************/

void WorkOrderParameters::reject()
{
    if(!QSqlDatabase::database().rollback())
    {
        QMessageBox::warning(this, "Error", 
            "Rollback of database changes failed",
            QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);
    }

    // Any changes that were made have been discarded. Reset the parmsModified 
    // flag to false.
    parmsModified = false;

    QDialog::reject();
}

/*****************************************************************************/

bool WorkOrderParameters::parametersModified()
{
    return parmsModified;
}
